#***************************************************************************
#
# VooDoo cIRCle - an IRC (ro)bot
# Port to PERL Copyright (C) 2006 by Varun Poduval
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  
# 02111-1307, USA.
#
# 
#****************************************************************************

use strict;
use warnings;

package c_server;

sub init { 
	my ($class, $host, $port, $nick) = @_;
	bless {
		host => $host, 
		port => $port, 
		nick => $nick,
		isupport => {},
		user_mode_prefix => {},
		chan_modes_class_a => '',    #!< String with list of CLASS A channel modes (e.g. "beIR")
		chan_modes_class_b => '',    #!< String with list of CLASS B channel modes (e.g. "k")
		chan_modes_class_c => '',    #< String with list of CLASS C channel modes (e.g. "l")
		chan_modes_class_d => ''   
		}, $class; 
}

1;

package c_input;

sub init {
	my ($class, $channel) = @_;
	bless { 
		channel => $channel, 
		params => []
		}, $class;
}

sub set_next_parameter {
	my ($self, $param) = @_;
	push @{ $self->{'params'} }, $param;
}

sub get_channel {
	my $self = shift;
	return $self->[0];
}

sub get_param {
	my ($self, $i) = @_;
	return $self->[1+$i]; # because $channel is at 0
}

1;

package c_user;

sub init {
	my ($class, $nick, $ident, $host, $full_name, $name, $online,
		$host_unknown, $host_bot, $irc_op, $mode, $host_masks,
		$fullname_masks, $meta, $groups) = @_;
	bless {
		nick => $nick, 
		ident => $ident, 
  		host => $host,
		full_name => $full_name, 
		name => $name, 
		online => $online,
		host_unknown => $host_unknown, 
		host_bot => $host_bot, 
		irc_op => $irc_op,
        mode => $mode,
		host_masks => $host_masks, 
		fullname_masks => $fullname_masks, 
		meta => $meta,
		groups => $groups
	}, $class
}

1;

package c_channel;

sub init {
	my ($class, $name, $bans, $exceptions, $invites, $reops) = @_;
	bless {
		name => $name, 
		bans => $bans, 
		exceptions => $exceptions,
		invites => $invites, 
		reops => $reops,
		users => [],
		key => {},               #!< Key for the channel, or an empty string
		limit => {},             #!< Limit for the channel as string, or an empty string for none
		mode => {},             #!< String with modes (e.g. "nt") NB: There is no '+' at the beginning!
		topic => {}             #< Channel's topic
	}, $class;
}

sub add_user {
	my ($self, $user) = @_;
	foreach my $e ( @{ $self->{'users'} } ) {
		if ($e->{nick} eq $user->{nick}) {
			return;
		}
	}
	push @{ $self->{'users'} }, $user;
}

sub find_users_name {
	my ($self, $name, $want_online, $want_offline, $ignore_case) = @_;
	if(!$want_online && !$want_offline) {
		print "Warning: c_channel::find_users_name() three-state logic issue: 
Do we want online, offline, or both? You want none!\n";
		return ();
	}
	my @res = ();
	foreach my $e ( @{ $self->{'users'} } ) {
		if ((($want_online && $e->online) || ($want_offline && !$e->online)) 
&& ((!$ignore_case && $e->name eq $name) || ($ignore_case && 
lc($e->name) eq lc($name)))) {
			push @res, $e;
		}
	}
	return @res;
}

sub find_user_nick {
	my ($self, $nick, $ignore_case) = @_;
	foreach my $e ( @{ $self->{'users'} } ) {
		if((!$ignore_case && $e->nick eq $nick) || ($ignore_case && 
lc($e->nick) eq lc($nick))) {
			return $e;
		}
	}
	return ();
}

sub find_irc_op {
	my ($self) = shift;
	return grep $_->irc_op, @{ $self->{users} };
}

sub find_unknown {
	my ($self) = shift;
	return grep $_->host_unknown, @{ $self->{users} };
}

sub find_host_bot {
	my ($self) = shift;
	return grep $_->host_bot, @{ $self->{users} };
}

sub find_mode {
	my ($self, $mode, $plus) = @_;
		if(length($mode)!=1) {
			print "Error: c_channel::find_mode() argument \$mode should be 
string one-character long. It is: \"".$mode."\"\n";
			return ();
		}
	my @res = ();
	foreach my $e ( @{ $self->{'users'} } ) {
		if (($plus && index($e->mode,$mode)) || (!$plus && 
(!index($e->mode,$mode)))) {
			push @res, $e;
		}
	}
	return @res;
}

1;

package c_filesys_access;

sub init {
	my ($class, $all_users, $user_name, $owner, $read, $delete,
        $notify_owner, $notify_user, $secure, $all_on_channel,
        $also_unknown, $notify_owner_message, $notify_user_message) = 
@_;
	bless {
		all_users => $all_users, 
		user_name => $user_name, 
		owner => $owner,
		read => $read, 
		delete => $delete, 
		notify_owner => $notify_owner,
		notify_user => $notify_user, 
		all_on_channel => $all_on_channel, 
		also_unknown => $also_unknown,
		notify_owner_message => $notify_owner_message, 
		notify_user_message => $notify_user_message
	}, $class; 
}
	
1;

package c_filesys_event;

sub init {
	my ($class, $event_has_read, $event_owner_notified, $event_user_name,
        $user_notified, $event_timestamp) = @_;
	bless {
		event_has_read => $event_has_read, 
		event_owner_notified => $event_owner_notified, 
		event_user_name => $event_user_name,
		user_notified => $user_notified, 
		event_timestamp => $event_timestamp
	}, $class; 
}

1;

package c_filesys_object;

sub init {
	my ($class, $type, $internal_name, $public_name, $time, $published,
        $complete, $expiration) = @_;
	bless {
		type => $type, 
		internal_name => $internal_name, 
		public_name => $public_name,
		time => $time, 
		published => $published,
		complete => $complete, 
		expiration => $expiration,
		access => [],
		events => []		
	}, $class; 
}

sub add_access {
	my ($self, $access) = @_;
	push @{ $self->{'access'} }, $access;
}

sub add_event {
	my ($self, $event) = @_;
	push @{ $self->{'events'} }, $event;
}

1;

package c_output;

sub new {
	my ($class, $lines, $commited) = shift;
	bless {
		lines => $lines,
		commited => $commited,
	}, $class;
}

sub put {
	my ($self, $priority_class, $raw) = @_;
	my $x="PUT ".$priority_class." ".$raw;
	$self->{'lines'} .= $x."\n";
}

sub put_lines {
	my ($self, $priority_class, $prefix, $lines, $allow_empty) = @_;
	my @lns = split("\n",$lines);
	foreach my $str (@lns) {
		$str=~ s{\r}{};
        if ((!$allow_empty) && ($str eq "")) {
			next;
		}
		my $x="PUT ".$priority_class." ".$prefix.$str;
		$self->{'lines'} .= $x."\n";
	}
}

sub log {
	my ($self, $string) = @_;
	my $x="LOG ".$string;
	$x =~ s/(["\000..\037\177..\377"])/\\$1/g;
	$self->{'lines'}.=$x."\n";
}

sub execute {
	my ($self, $procedure, $arguments) = @_;
	my $x="EXECUTE ".$procedure."(";
	my $first=1;
	foreach my $e ( @{ $self->{'arguments'} } ) {
		if (!$first) {
			$x.=",";
		}
		undef($first);
        $x.="\"";
        $x.=quotemeta($e);
        $x.="\"";
	}
	$x.=")";
	
	$self->{'lines'}.=$x."\n";
}

sub filesys {
	my ($self, $object) = @_;
	
	my $r = "FILESYS InternalName ".@{ 
$self->{'object'}->{'internal_name'} };
	$self->{'lines'}.=$r."\n";
	
	$r="FILESYS SET type ".@{ $self->{'object'}->{'type'} };
	$self->{'lines'}.=$r."\n";

	$r="FILESYS SET PublicName ".@{ $self->{'object'}->{'public_name'} };
	$self->{'lines'}.=$r."\n";

	$r="FILESYS SET Time ".@{ $self->{'object'}->{'time'} };
	$self->{'lines'}.=$r."\n";

	$r="FILESYS SET Published ".( @{ 
$self->{'object'}->{'published?"1":"0"'} } );
	$self->{'lines'}.=$r."\n";

	$r="FILESYS SET Complete ".( @{ 
$self->{'object'}->{'complete?"1":"0"'} } );
	$self->{'lines'}.=$r."\n";

	$r="FILESYS SET Expiration ".@{ $self->{'object'}->{'expiration'} };
	$self->{'lines'}.=$r."\n";
	
	foreach my $e ( @{ $self->{'object'}->{'access'} } ) {

		$r="FILESYS ACCESS_BEGIN";
		$self->{'lines'}.=$r."\n";

		$r="FILESYS SET AllUsers ".($e->all_users?"1":"0");
		$self->{'lines'}.=$r."\n";

		$r="FILESYS SET UserName ".$e->user_name;
		$self->{'lines'}.=$r."\n";

		$r="FILESYS SET Owner ".($e->owner?"1":"0");
		$self->{'lines'}.=$r."\n";

		$r="FILESYS SET Read ".($e->read?"1":"0");
		$self->{'lines'}.=$r."\n";

		$r="FILESYS SET Delete ".($e->delete?"1":"0");
		$self->{'lines'}.=$r."\n";

		$r="FILESYS SET NotifyOwner ".($e->notify_owner?"1":"0");
		$self->{'lines'}.=$r."\n";
		
		$r="FILESYS SET NotifyUser ".($e->notify_user?"1":"0");
		$self->{'lines'}.=$r."\n";

		$r="FILESYS SET Secure ".($e->secure?"1":"0");
		$self->{'lines'}.=$r."\n";

		$r="FILESYS SET AllOnChannel ".$e->all_on_channel;
		$self->{'lines'}.=$r."\n";

		$r="FILESYS SET AlsoUnknown ".($e->also_unknown?"1":"0");
		$self->{'lines'}.=$r."\n";

		$r="FILESYS SET NotifyOwnerMessage ".$e->notify_owner_message;
		$self->{'lines'}.=$r."\n";

		$r="FILESYS SET NotifyUserMessage ".$e->notify_user_message;
		$self->{'lines'}.=$r."\n";

		$r="FILESYS ACCESS_END";
		$self->{'lines'}.=$r."\n";
	}
	
	$r="FILESYS COMMIT";
	$self->{'lines'}.=$r."\n";
}

sub cancel {
	my ($self, $string) = @_;
	undef( $self->{'lines'} );
}

sub commit {
	my $self = shift;
	if ( $self->{'commited'} ) {
		print "Warning: c_output::commit() You have already done commit - 
this can confuse bot!\n";
	}
	$self->{'commited'} = 1;
	print "\n#####\n".$self->{'lines'}."#####\n";
	$self->{'lines'}="";
}
	
1;