<?php

/***************************************************************************

VooDoo cIRCle - an IRC (ro)bot
Copyright (C) 2004-2005  Marian VooDooMan Meravy
                         (ghostvoodooman@users.sourceforge.net)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************/

/*!
    \brief Server connection informations
    \version 1
    \author VooDooMan
    \date 2005
*/
class c_server {
    var $host;                  //!< Host we are connected to
    var $port;                  //!< Port we are connected to
    var $nick;                  //!< Nickname we are using
    
    var $isupport;              //!< Associative array of RPL_ISSUPORT 005 numeric data
    var $user_mode_prefix;      //!< Associative array of user mode prefix (e.g. "o" == "@")
    var $chan_modes_class_a;    //!< String with list of CLASS A channel modes (e.g. "beIR")
    var $chan_modes_class_b;    //!< String with list of CLASS B channel modes (e.g. "k")
    var $chan_modes_class_c;    //!< String with list of CLASS C channel modes (e.g. "l")
    var $chan_modes_class_d;    //!< String with list of CLASS D channel modes (e.g. "mnti")
    
    /*!
        \brief Initialization
        \version 1
        \author VooDooMan
        \date 2005
        \param host Host of IRC server
        \param port Port of IRC server
        \param nick Nick of bot
    */
    function init($host,$port,$nick) {
        $this->host=$host;
        $this->port=$port;
        $this->nick=$nick;
    }
}

/*!
    \brief Input parameters given by SCRIPT command
    \version 1
    \author VooDooMan
    \date 2005
*/
class c_input {
    var $channel;       //!< Name of channel (if applicable)
    var $params;        //!< Array of input parameters
    
    /*!
        \brief Initialization: clean-up
        \version 1
        \author VooDooMan
        \date 2005
        \param channel Name of channel
    */
    function init($channel) {
        $this->channel=$channel;
        $this->params=array();
    }
  
    /*!
        \brief Adds a new parameter to the array
        \version 1
        \author VooDooMan
        \date 2005
        \param value Value as next parameter
        \warning For internal use only
    */  
    function set_next_parameter($value) {
        array_push($this->params,$value);
    }
    
    /*!
        \brief Gets channel from SCRIPT command
        \version 1
        \author VooDooMan
        \date 2005
        \return Returns channel name
    */
    function get_channel() {
        return $this->channel;
    }
    
    /*!
        \brief Gets parameter
        \version 1
        \author VooDooMan
        \date 2005
        \param index Zero-based index of parameter
        \return Returns parameter at given index
        \retval NULL Returns NULL if index is out of the range
    */
    function get_param($index)
    {
        if($index>=count($this->params))
            return NULL;
        return $this->params[$index];
    }
}

/*!
    \brief Stores informations on one user
    \version 1
    \author VooDooMan
    \date 2005
*/
class c_user {
    var $nick;              //!< Nick name
    var $ident;             //!< Ident of user
    var $host;              //!< Host of user
    var $full_name;         //!< Full name of user
    var $name;              //!< Name of user as in logic.txt
    var $online;            //!< TRUE if user is online
    var $host_unknown;      //!< TRUE if user is flagged as "host_unknown"
    var $host_bot;          //!< TRUE if user is flagged as "host_bot"
    var $irc_op;            //!< TRUE if user has an IRC operator status
    var $mode;              //!< Mode of user
    var $host_masks;        //!< Array of host masks
    var $fullname_masks;    //!< Array of fullname masks
    var $meta;              //!< Associative array of meta data
    var $groups;            //!< Array of strings: groups which is the user member of
    
    /*!
        \brief Initialization
        \version 1
        \author VooDooMan
        \date 2005
        \param nick Nick name
        \param ident Ident of user
        \param host Host of user
        \param full_name Full name of user
        \param name Name of user as in logic.txt
        \param online TRUE if user is online
        \param host_unknown TRUE if user is flagged as "host_unknown"
        \param host_bot TRUE if user is flagged as "host_bot"
        \param irc_op TRUE if user has an IRC operator status
        \param mode Mode of user
        \param host_masks Array of host masks
        \param fullname_masks Array of fullname masks
        \param meta Associative array of meta data
        \param groups Array of strings: groups which is the user member of
    */
    function init($nick, $ident, $host, $full_name, $name, $online,
        $host_unknown, $host_bot, $irc_op, $mode, $host_masks,
        $fullname_masks, $meta, $groups)
    {
        $this->nick=$nick;
        $this->ident=$ident;
        $this->host=$host;
        $this->full_name=$full_name;
        $this->name=$name;
        $this->online=$online;
        $this->host_unknown=$host_unknown;
        $this->host_bot=$host_bot;
        $this->irc_op=$irc_op;
        $this->mode=$mode;
        $this->host_masks=$host_masks;
        $this->fullname_mask=$fullname_masks;
        $this->meta=$meta;
        $this->groups=$groups;
    }
}

/*!
    \brief Stores informations on one channel
    \version 1
    \author VooDooMan
    \date 2005
*/
class c_channel {
    var $name;              //!< Channel name
    var $users;             //!< Array of c_user
    var $bans;              //!< Array of ban masks
    var $exceptions;        //!< Array of exceptions (to bans)
    var $invites;           //!< Array of invites
    var $reops;             //!< Array of reop hints (IRCnet-specific)
    
    var $key;               //!< Key for the channel, or an empty string
    var $limit;             //!< Limit for the channel as string, or an empty string for none
    var $mode;              //!< String with modes (e.g. "nt") NB: There is no '+' at the beginning!
    var $topic;             //!< Channel's topic
    
    /*!
        \brief Initialization
        \version 1
        \author VooDooMan
        \date 2005
        \param name Name of channel
        \param bans Array of ban masks
        \param exceptions Array of exceptions (to bans)
        \param invites Array of invites
        \param reops Array of reop hints (IRCnet-specific)
    */
    function init($name,$bans,$exceptions,$invites,$reops)
    {
        $this->name=$name;
        $this->users=array();
        $this->bans=$bans;
        $this->exceptions=$exceptions;
        $this->invites=$invites;
        $this->reops=$reops;
    }
    
    /*!
        \brief Adds an user to to array
        \version 1
        \author VooDooMan
        \date 2005
        \param user c_user object to add
        \warning For internal use only
    */
    function add_user($user)
    {
        foreach($this->users as $e)
            if($e->nick==$user->nick)
                return;
        array_push($this->users,$user);
    }
    
    /*!
        \brief Finds user(s) as name on this channel
        \version 1
        \author VooDooMan
        \date 2005
        \param name Name of user as in logic.txt
        \param want_online Set this to TRUE if you want to get online users
        \param want_offline Set this to TRUE if you want to get offline users
        \param ignore_case Set this to TRUE for case-insensitivity
        \return Returns array of found users, array of c_user
    */
    function find_users_name($name,$want_online,$want_offline,$ignore_case)
    {
        if(!$want_online && !$want_offline) {
            echo "Warning: ".__CLASS__."::".__FUNCTION__."() three-state logic issue: Do we want online, offline, or both? You want none!\n";
            return array(); // optimization: it is not needed to continue
        }
        $res=array();
        foreach($this->users as $e) {
            if((($want_online && $e->online) || ($want_offline && !$e->online)) && ((!$ignore_case && $e->name==$name) || ($ignore_case && strtolower($e->name)==strtolower($name))))
                array_push($res,$e);
        }
        return $res;
    }

    /*!
        \brief Finds user as nick on this channel
        \version 1
        \author VooDooMan
        \date 2005
        \param nick Nick of user
        \param ignore_case Set this TRUE for case-insensitive comparsion
        \return Returns c_user or NULL
        \retval NULL Returns NULL if no such nick has been found on this channel
    */
    function find_user_nick($nick,$ignore_case)
    {
        foreach($this->users as $e) {
            if((!$ignore_case && $e->nick==$nick) || ($ignore_case && strtolower($e->nick)==strtolower($nick)))
                return $e;
        }
        return NULL;
    }
    
    /*!
        \brief Finds IRC operators on this channel
        \version 1
        \author VooDooMan
        \date 2005
        \return Returns array of c_user
    */
    function find_irc_op()
    {
        $res=array();
        foreach($this->users as $e) {
            if($e->irc_op)
                array_push($res,$e);
        }
        return $res;
    }
    
    /*!
        \brief Finds unknown users on this channel ("host_unknown")
        \version 1
        \author VooDooMan
        \date 2005
        \return Returns array of c_user
    */
    function find_unknown()
    {
        $res=array();
        foreach($this->users as $e) {
            if($e->host_unknown)
                array_push($res,$e);
        }
        return $res;
    }
    
    /*!
        \brief Finds this bot on this channel ("host_bot")
        \version 1
        \author VooDooMan
        \date 2005
        \return Returns array of c_user
        \warning If there are two instances of this bot, or host/fullname masks matches user with "host_bot" flag, there will be more users found!
    */
    function find_host_bot()
    {
        $res=array();
        foreach($this->users as $e) {
            if($e->host_bot)
                array_push($res,$e);
        }
        return $res;
    }
    
    /*!
        \brief Finds users with/out some mode on this channel
        \version 1
        \author VooDooMan
        \date 2005
        \param mode Mode required to include, or exclude (use "@" for op, "+" for voice)
        \param plus If set to TRUE: find users with that mode; FALSE: find users without that mode
        \return Returns array of c_user
    */
    function find_mode($mode,$plus)
    {
        if(strlen($mode)!=1) {
            echo "Error: c_channel::find_mode() argument \$mode should be string one-character long. It is: \"".$mode."\"\n";
            return array();
        }
        $res=array();
        foreach($this->users as $e) {
            if(($plus && strpos($e->mode,$mode)!==FALSE) || (!$plus && strpos($e->mode,$mode)===FALSE))
                array_push($res,$e);
        }
        return $res;
    }
}

/*!
    \brief Stores access rights of one user for filesystem object
    \version 1
    \author VooDooMan
    \date 2005
*/
class c_filesys_access {
    var $all_users;             //!< Boolean: is for all users?
    var $user_name;             //!< Name of user
    var $owner;                 //!< Boolean: OWNER: is user owner?
    var $read;                  //!< Boolean: READ: can read?
    var $delete;                //!< Boolean: DELETE: can delete?
    var $notify_owner;          //!< Boolean: NOTIFY_OWNER: should be owner notified when someone reads the object?
    var $notify_user;           //!< Boolean: NOTIFY_USER: should be user notified to read the object?
    var $secure;                //!< Boolean: is SECUREd object?
    var $all_on_channel;        //!< Name of channel for all users, or empty
    var $also_unknown;          //!< Boolean: also for unknown users?
    var $notify_owner_message;  //!< String: Message for owner used as notification
    var $notify_user_message;   //!< String: Message for user used as notification
    
    /*!
        \brief Initialization
        \version 1
        \author VooDooMan
        \date 2005
        \param all_users Boolean: is for all users?
        \param user_name Name of user
        \param owner Boolean: OWNER: is user owner?
        \param read Boolean: READ: can read?
        \param delete Boolean: DELETE: can delete?
        \param notify_owner Boolean: NOTIFY_OWNER: should be owner notified when someone reads the object?
        \param notify_user Boolean: NOTIFY_USER: should be user notified to read the object?
        \param secure Boolean: is SECUREd object?
        \param all_on_channel Name of channel for all users, or empty
        \param also_unknown Boolean: also for unknown users?
        \param notify_owner_message String: Message for owner used as notification
        \param notify_user_message String: Message for user used as notification
    */
    function init($all_users, $user_name, $owner, $read, $delete,
        $notify_owner, $notify_user, $secure, $all_on_channel,
        $also_unknown, $notify_owner_message, $notify_user_message)
    {
        $this->all_users=$all_users;
        $this->user_name=$user_name;
        $this->owner=$owner;
        $this->read=$read;
        $this->delete=$delete;
        $this->notify_owner=$notify_owner;
        $this->notify_user=$notify_user;
        $this->secure=$secure;
        $this->all_on_channel=$all_on_channel;
        $this->also_unknown=$also_unknown;
        $this->notify_owner_message=$notify_owner_message;
        $this->notify_user_message=$notify_user_message;
    }
}

/*!
    \brief Stores event for filesystem object
    \version 1
    \author VooDooMan
    \date 2005
*/
class c_filesys_event {
    var $event_has_read;            //!< Boolean: whether has read
    var $event_owner_notified;      //!< Boolean: whether the owner has been notified
    var $event_user_name;           //!< Name of user for event
    var $user_notified;             //!< Boolean: whether user has been notified
    var $event_timestamp;           //!< Time of event
    
    /*!
        \brief Initialization
        \version 1
        \author VooDooMan
        \date 2005
        \param event_has_read Boolean: whether has read
        \param event_owner_notified Boolean: whether the owner has been notified
        \param event_user_name Name of user for event
        \param user_notified Boolean: whether user has been notified
        \param event_timestamp Time of event
    */
    function init($event_has_read, $event_owner_notified, $event_user_name,
        $user_notified, $event_timestamp)
    {
        $this->event_has_read=$event_has_read;
        $this->event_owner_notified=$event_owner_notified;
        $this->event_user_name=$event_user_name;
        $this->user_notified=$user_notified;
        $this->event_timestamp=$event_timestamp;
    }
}

/*!
    \brief Stores filesystem object
    \version 1
    \author VooDooMan
    \date 2005
    \todo Make some *clever* manipulating methods
*/
class c_filesys_object {
    var $type;              //!< Type of object ("ft_message", or "ft_file")
    var $internal_name;     //!< Internal name of object (physical name of file)
    var $public_name;       //!< Public name of object
    var $time;              //!< Time of creation
    var $published;         //!< Boolean: is it published?
    var $complete;          //!< Boolean: is file completely received?
    var $expiration;        //!< Time of expiration of object
    
    var $access;            //!< Array of c_filesys_access
    var $events;            //!< Array of c_filesys_event
    
    /*!
        \brief Initialization
        \version 1
        \author VooDooMan
        \date 2005
        \param type Type of object ("ft_message", or "ft_file")
        \param internal_name Internal name of object (physical name of file)
        \param public_name Public name of object
        \param time Time of creation
        \param published Boolean: is it published?
        \param complete Boolean: is file completely received?
        \param expiration Time of expiration of object
    */
    function init($type, $internal_name, $public_name, $time, $published,
        $complete, $expiration)
    {
        $this->type=$type;
        $this->internal_name=$internal_name;
        $this->public_name=$public_name;
        $this->time=$time;
        $this->published=$published;
        $this->complete=$complete;
        $this->expiration=$expiration;
        $this->access=array();
        $this->events=array();
    }
    
    /*!
        \brief Adds access right
        \version 1
        \author VooDooMan
        \date 2005
        \param access c_filesys_access object to add
    */
    function add_access($access) {
        array_push($this->access,$access);
    }
    
    /*!
        \brief Adds event
        \version 1
        \author VooDooMan
        \date 2005
        \param event c_filesys_event object to add
    */
    function add_event($event) {
        array_push($this->events,$event);
    }
}

/*!
    \brief Generates output
    \version 1
    \author VooDooMan
    \date 2005
*/
class c_output {
    var $lines;     //!< Holds lines of output
    
    var $commited;
    
    /*!
        \brief Initialization
        \version 1
        \author VooDooMan
        \date 2005
    */
    function init() {
        $this->lines="";
        $this->commited=FALSE;
    }
    
    /*!
        \brief Puts raw message to IRC server
        \version 1
        \author VooDooMan
        \date 2005
        \param priority_class Priority class (e.g. "LOW", "HIGH", "CRITICAL")
        \param raw Raw message
        \bug Processing by bot will fail if string contains CR or LF (consider using put_lines())
    */
    function put($priority_class,$raw)
    {
        $x="PUT ".$priority_class." ".$raw;
        $this->lines.=$x."\n";
    }
    
    /*!
        \brief Puts raw message(s)/line(s) to IRC server
        \version 1
        \author VooDooMan
        \date 2005
        \param priority_class Priority class (e.g. "LOW", "HIGH", "CRITICAL")
        \param prefix For example "PRIVMSG #channel :"
        \param lines String with multiple lines separated by "\n" (they will be prefixed with $prefix)
        \param allow_empty Set to true if to allow empty lines; if set to false, empty lines will be excluded
    */
    function put_lines($priority_class,$prefix,$lines,$allow_empty)
    {
        $lns=explode("\n",$lines);
        foreach($lns as $element) {
            $str=str_replace("\r","",$element);
            if(!$allow_empty && $str=="")
                continue;
            $x="PUT ".$priority_class." ".$prefix.$str;
            $this->lines.=$x."\n";
        }
    }
    
    /*!
        \brief Generates a log message
        \version 1
        \author VooDooMan
        \date 2005
        \param string Log message
    */
    function log($string)
    {
        $x="LOG ".addcslashes($string,"\000..\037\177..\377");
        $this->lines.=$x."\n";
    }
    
    /*!
        \brief Generates execution of procedure
        \version 1
        \author VooDooMan
        \date 2005
        \param procedure Name of procedure
        \param arguments Array of strings as arguments
        \bug Processing by bot will fail if procedure and/or arguments' elements contain CR or LF
    */
    function execute($procedure,$arguments)
    {
        $x="EXECUTE ".$procedure."(";
        $first=TRUE;
        foreach($arguments as $e) {
            if(!$first)
                $x.=",";
            $first=FALSE;
            $x.="\"";
            $x.=addslashes($e);
            $x.="\"";
        }
        $x.=")";
        
        $this->lines.=$x."\n";
    }
    
    /*!
        \brief Modifies filesystem object
        \version 1
        \author VooDooMan
        \date 2005
        \param object c_filesys_object object
        \warning In notify messages string "hello\nworld\x07!" should be represented as "hello\\nworld\\x07!"
        \bug Event modifying not supported yet
    */
    function filesys($object)
    {
        $r="FILESYS InternalName ".$object->internal_name;
        $this->lines.=$r."\n";

        $r="FILESYS SET type ".$object->type;
        $this->lines.=$r."\n";

        $r="FILESYS SET PublicName ".$object->public_name;
        $this->lines.=$r."\n";

        $r="FILESYS SET Time ".$object->time;
        $this->lines.=$r."\n";

        $r="FILESYS SET Published ".($object->published?"1":"0");
        $this->lines.=$r."\n";

        $r="FILESYS SET Complete ".($object->complete?"1":"0");
        $this->lines.=$r."\n";

        $r="FILESYS SET Expiration ".$object->expiration;
        $this->lines.=$r."\n";

        foreach($object->access as $e) {
            // begin of access rights
            $r="FILESYS ACCESS_BEGIN";
            $this->lines.=$r."\n";

            $r="FILESYS SET AllUsers ".($e->all_users?"1":"0");
            $this->lines.=$r."\n";

            $r="FILESYS SET UserName ".$e->user_name;
            $this->lines.=$r."\n";

            $r="FILESYS SET Owner ".($e->owner?"1":"0");
            $this->lines.=$r."\n";

            $r="FILESYS SET Read ".($e->read?"1":"0");
            $this->lines.=$r."\n";

            $r="FILESYS SET Delete ".($e->delete?"1":"0");
            $this->lines.=$r."\n";

            $r="FILESYS SET NotifyOwner ".($e->notify_owner?"1":"0");
            $this->lines.=$r."\n";
            
            $r="FILESYS SET NotifyUser ".($e->notify_user?"1":"0");
            $this->lines.=$r."\n";

            $r="FILESYS SET Secure ".($e->secure?"1":"0");
            $this->lines.=$r."\n";

            $r="FILESYS SET AllOnChannel ".$e->all_on_channel;
            $this->lines.=$r."\n";

            $r="FILESYS SET AlsoUnknown ".($e->also_unknown?"1":"0");
            $this->lines.=$r."\n";

            // warning: string "hello\nworld\x07!" should be represented as "hello\\nworld\\x07!"
            $r="FILESYS SET NotifyOwnerMessage ".$e->notify_owner_message;
            $this->lines.=$r."\n";

            // warning: string "hello\nworld\x07!" should be represented as "hello\\nworld\\x07!"
            $r="FILESYS SET NotifyUserMessage ".$e->notify_user_message;
            $this->lines.=$r."\n";

            // end of access rights
            $r="FILESYS ACCESS_END";
            $this->lines.=$r."\n";
        }
        
        // event modifying not supported yet

        // commit
        $r="FILESYS COMMIT";
        $this->lines.=$r."\n";
    }
    
    /*!
        \brief Cancels all pending operations
        \version 1
        \author VooDooMan
        \date 2005
    */
    function cancel($string)
    {
        $this->lines="";
    }
    
    /*!
        \brief Commits (posts) all pending operations
        \version 1
        \author VooDooMan
        \date 2005
    */
    function commit()
    {
        if($this->commited) {
            echo "Warning: ".__CLASS__."::".__FUNCTION__."() You have already done commit - this can confuse bot!\n";
        }
        $this->commited=TRUE;
        echo "\n#####\n".$this->lines."#####\n"; // NB: we got \n at the end of $this->lines
        $this->lines="";
    }
}

?>