/*
 * Copyright (c) 1997-1999
 * Silicon Graphics Computer Systems, Inc.
 *
 * Copyright (c) 1999
 * Boris Fomitchev
 *
 * This material is provided "as is", with absolutely no warranty expressed
 * or implied. Any use is at your own risk.
 *
 * Permission to use or copy this software for any purpose is hereby granted
 * without fee, provided the above notices are retained on all copies.
 * Permission to modify the code and to distribute modified code is granted,
 * provided the above notices are retained, and a notice that the code was
 * modified is included with the above copyright notice.
 *
 */

/*
    WARNING!
    This file is part of STLport library. It WAS MODIFIED, this
    is NOT original file from original STLport distribution.

    It was modified for purposes of VooDoo cIRCle, an irc bot,
    to allow heap debugging (memory leaks / resource type mismatch
    while deallocation). It was modified by Marian VooDooMan Meravy
    on 2007-07-29. ghostvoodooman -aT= users (dot) sourceforge.net

    Original sources can be downloaded from the official STLport
    site: http://www.stlport.org/
*/

/* $Id: auto_ptr.h,v 1.14 2008/08/23 15:10:42 ghostvoodooman Exp $ */

/*!
    \file
    \brief 'auto_ptr' class template implementation
    \version $Id: auto_ptr.h,v 1.14 2008/08/23 15:10:42 ghostvoodooman Exp $
*/

#ifndef _STLP_AUTO_PTR_H_VDC__      // to not collide with Borland's implementation
# define _STLP_AUTO_PTR_H_VDC__

#if defined(__INTEL_COMPILER)
#   pragma warning(push)
#   pragma warning(disable:444)     // remark #444: destructor for base class "my_auto_ptr::__ptr_base" (declared at line 53) is not virtual
#endif

//_STLP_BEGIN_NAMESPACE

namespace my_auto_ptr {

#define __CONST_CAST(to,what)           const_cast<to>(what)
#define __REINTERPRET_CAST(to,what)     reinterpret_cast<to>(what)
#define _STLP_VERBOSE_ASSERT(x,y)       /* */

// implementation primitive
class __ptr_base {
public:
  void* _M_p;
  void  __set(const void* p) { _M_p = __CONST_CAST(void*,p); }
  void  __set(void* p) { _M_p = p; }
};

#if defined(_MSC_VER)
//  BEGIN: disable MSVC warning: warning C4512: 'my_auto_ptr::auto_ptr_ref<_Tp>' : assignment operator could not be generated
#   pragma warning( push )
#   pragma warning( disable : 4512 )
#endif

template <class _Tp> class auto_ptr_ref {
public:
  __ptr_base& _M_r;
  _Tp* const _M_p;

  auto_ptr_ref(__ptr_base& __r, _Tp* __p) : _M_r(__r), _M_p(__p) {  }

  _Tp* release() const { _M_r.__set((void*)0); return _M_p; }

};

#if defined(_MSC_VER)
//  END:   disable MSVC warning: warning C4512: 'my_auto_ptr::auto_ptr_ref<_Tp>' : assignment operator could not be generated
#   pragma warning( pop )
#endif

template<class _Tp> class auto_ptr :  public __ptr_base {
public:
  typedef _Tp element_type;
  typedef auto_ptr<_Tp>           _Self;

  _Tp* release() {
    _Tp* __px = this->get();
    this->_M_p = 0;
    return __px;
  }

  void reset(_Tp* __px=0) {
    _Tp* __pt = this->get();
    if (__px != __pt)
      MY_DELETE(__pt);
    this->__set(__px);
  }

  _Tp* get() const { return __REINTERPRET_CAST(_Tp*,__CONST_CAST(void*,_M_p)); }

# if !defined (_STLP_NO_ARROW_OPERATOR)
  _Tp* operator->() const {
    _STLP_VERBOSE_ASSERT(get()!=0, _StlMsg_AUTO_PTR_NULL)
    return get();
  }
# endif
  _Tp& operator*() const  {
    _STLP_VERBOSE_ASSERT(get()!=0, _StlMsg_AUTO_PTR_NULL)
    return *get();
  }

  auto_ptr() { this->_M_p = 0; }

  explicit auto_ptr(_Tp* __px) { this->__set(__px); }

#if defined (_STLP_MEMBER_TEMPLATES)
# if !defined (_STLP_NO_TEMPLATE_CONVERSIONS)
  template<class _Tp1> auto_ptr(auto_ptr<_Tp1>& __r) {
    _Tp* __conversionCheck = __r.release();
    this->__set(__conversionCheck);
  }
# endif
  template<class _Tp1> auto_ptr<_Tp>& operator=(auto_ptr<_Tp1>& __r) {
    _Tp* __conversionCheck = __r.release();
    reset(__conversionCheck);
    return *this;
  }
#endif /* _STLP_MEMBER_TEMPLATES */

  auto_ptr(_Self& __r) { this->__set(__r.release()); }

  _Self& operator=(_Self& __r)  {
    reset(__r.release());
    return *this;
  }

  ~auto_ptr() { /* boris : reset(0) might be better */ MY_DELETE(this->get()); }

  auto_ptr(auto_ptr_ref<_Tp> __r) {
    this->__set(__r.release());
  }

  _Self& operator=(auto_ptr_ref<_Tp> __r) {
    reset(__r.release());
    return *this;
  }

# if defined(_STLP_MEMBER_TEMPLATES) && !defined(_STLP_NO_TEMPLATE_CONVERSIONS)
  template<class _Tp1> operator auto_ptr_ref<_Tp1>() {
    return auto_ptr_ref<_Tp1>(*this, this->get());
  }
  template<class _Tp1> operator auto_ptr<_Tp1>() {
    return auto_ptr<_Tp1>(release());
  }
# else
  operator auto_ptr_ref<_Tp>()
  { return auto_ptr_ref<_Tp>(*this, this->get()); }
# endif

};

//_STLP_END_NAMESPACE

} // namespace my_auto_ptr

#if defined(__INTEL_COMPILER)
#   pragma warning(pop)             // remark #444: destructor for base class "my_auto_ptr::__ptr_base" (declared at line 53) is not virtual
#endif

#endif /* _STLP_AUTO_PTR_H */

// Local Variables:
// mode:C++
// End:

