/***************************************************************************
                          binmod.cpp  -  description
                             -------------------
    begin                : Fri Aug 25 2006
    copyright            : (C) 2006-2007 by Marian VooDooMan Meravy
    email                : ghostvoodooman -aT= users (dot) sourceforge.net
 ***************************************************************************/

/* $Id: binmod.cpp,v 1.35 2008/08/23 15:10:38 ghostvoodooman Exp $ */

/***************************************************************************

VooDoo cIRCle - an IRC (ro)bot
Copyright (C) 2006-2008 by Marian VooDooMan Meravy
  (ghostvoodooman -aT= users (dot) sourceforge.net)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************/

/*!
    \file
    \brief Provides interface for binary modules
    \version $Id: binmod.cpp,v 1.35 2008/08/23 15:10:38 ghostvoodooman Exp $
*/

#if defined(HAVE_CONFIG_H)
#   include <config.h>
#endif

#include "utils.h"

#include <string>
#include <vector>
#include <stack>
#include <list>
#include <map>
#include <deque>

#if defined(_WIN32) || defined(_WIN64) || defined(__CYGWIN__)
#   include <windows.h>
#endif

#ifdef HAVE_DLOPEN
#   include <dlfcn.h>
#endif

using namespace std;

#include "utils/memleak.h"
#include "utils/assert.h"

#include "params.h"
#include "compat.h"
#include "function.h"
#include "eval.h"
#include "var_keeper.h"
#include "binmod.h"
#include "var/c_var_base.h"
#include "var/c_var_bool.h"
#include "var/c_var_int.h"
#include "var/c_var_string.h"
#include "var/c_var_array.h"
#include "var/c_var_unset.h"
#include "internal_state.h"
#include "logic/logic_internal.h"

#include "vdcmod.h"

#include "utils/precomp.h"

#if defined(_MSC_VER) || defined(__BORLANDC__)
#   pragma hdrstop
#endif

static void bind_func(bool ex, vdcmod_mod_id_t mod_id, const char* type_mask, const char* user_mask, const char* channel_mask, const char* line_mask, int before, vdcmod_dispatcher_func dispatcher)
{
    vector<c_module*>::iterator i1;
    for(i1=modules.begin(); i1!=modules.end(); ++i1) {
        if((vdcmod_mod_id_t)(*i1)==mod_id) {
            vector<string> error_log;
            c_binder b(ex,type_mask,user_mask,channel_mask,line_mask,(before!=0)?true:false,dispatcher,*i1,error_log,"'bind' event call from binary module");
            if(error_log.empty())
                (*i1)->binds.push_back(b);
            else if((*i1)->get_logger()) {
                vector<string>::iterator i3;
                for(i3=error_log.begin(); i3!=error_log.end(); ++i3)
                    (*i1)->get_logger()->log_print1(E_LOG_DEBUG,LOG_SCOPE_MODULE,"",*i3);
            }
        }
    }
}

static void unbind_func(bool ex, vdcmod_mod_id_t mod_id, const char* type_mask, const char* user_mask, const char* channel_mask, const char* line_mask, int before, vdcmod_dispatcher_func dispatcher)
{
    vector<c_module*>::iterator i1;
    for(i1=modules.begin(); i1!=modules.end(); ++i1) {
        if((vdcmod_mod_id_t)(*i1)==mod_id) {
            vector<string> error_log;
            c_binder b(ex,type_mask,user_mask,channel_mask,line_mask,(before!=0)?true:false,dispatcher,*i1,error_log,"'unbind' event call from binary module");
            if(error_log.empty()) {
                size_t unbound=0;
        unbind_again:
                vector<c_binder>::iterator i3;
                for(i3=(*i1)->binds.begin(); i3!=(*i1)->binds.end(); ++i3) {
                    if((*i3)==b) {
                        (*i1)->binds.erase(i3);
                        ++unbound;
                        goto unbind_again;
                    }
                }
                if((*i1)->get_logger()) {
                    string log;
                    if(0==unbound)
                        log="WARNING: None buond event was unbound";
                    if(1<unbound)
                        log=(string)"WARNING: "+tos(unbound)+" buond event were unbound";
                    if(!log.empty()) {
                        log+="; Additional info: ";
                        log+="'unbind' event call from binary module";
                        (*i1)->get_logger()->log_print1(E_LOG_DEBUG,LOG_SCOPE_MODULE,"",log);
                    }
                }
            } else if((*i1)->get_logger()) {
                vector<string>::iterator i3;
                for(i3=error_log.begin(); i3!=error_log.end(); ++i3)
                    (*i1)->get_logger()->log_print1(E_LOG_DEBUG,LOG_SCOPE_MODULE,"",*i3);
            }
            break;
        }
    }
}

extern "C" void mod_bind_func(vdcmod_mod_id_t mod_id, const char* type_mask, const char* user_mask, const char* channel_mask, const char* line_mask, int before, vdcmod_dispatcher_func dispatcher)
{
    bind_func(false,mod_id,type_mask,user_mask,channel_mask,line_mask,before,dispatcher);
}

extern "C" void mod_unbind_func(vdcmod_mod_id_t mod_id, const char* type_mask, const char* user_mask, const char* channel_mask, const char* line_mask, int before, vdcmod_dispatcher_func dispatcher)
{
    unbind_func(false,mod_id,type_mask,user_mask,channel_mask,line_mask,before,dispatcher);
}

extern "C" void mod_bind_ex_func(vdcmod_mod_id_t mod_id, const char* type_mask, const char* user_mask, const char* channel_mask, const char* line_mask, int before, vdcmod_dispatcher_func dispatcher)
{
    bind_func(true,mod_id,type_mask,user_mask,channel_mask,line_mask,before,dispatcher);
}

extern "C" void mod_unbind_ex_func(vdcmod_mod_id_t mod_id, const char* type_mask, const char* user_mask, const char* channel_mask, const char* line_mask, int before, vdcmod_dispatcher_func dispatcher)
{
    unbind_func(true,mod_id,type_mask,user_mask,channel_mask,line_mask,before,dispatcher);
}

extern "C" void mod_get_superglobals_func(vdcmod_mod_id_t mod_id, c_vdcmod_data_t** superglobals)
{
    if(!superglobals)
        return;
    *superglobals=NULL;
    vector<c_module*>::iterator i1;
    for(i1=modules.begin(); i1!=modules.end(); ++i1) {
        if((vdcmod_mod_id_t)(*i1)==mod_id) {
            c_binmodule* m=dynamic_cast<c_binmodule*>(*i1);
            if(m) {
                c_vdcmod_data_t* res=NULL;
                m->convert_vars(super_global,res);
                *superglobals=res;
            }
            break;
        }
    }
}

extern "C" void mod_free_structure_func(vdcmod_mod_id_t mod_id, c_vdcmod_data_t* struc)
{
    if(!struc)
        return;
    vector<c_module*>::iterator i1;
    for(i1=modules.begin(); i1!=modules.end(); ++i1) {
        if((vdcmod_mod_id_t)(*i1)==mod_id) {
            c_binmodule* m=dynamic_cast<c_binmodule*>(*i1);
            if(m) {
                m->free_vars(struc);
            }
            break;
        }
    }
}

extern "C" void mod_process_script_out_func(vdcmod_mod_id_t mod_id, const char* text)
{
    if(!text)
        return;
    vector<c_module*>::iterator i1;
    for(i1=modules.begin(); i1!=modules.end(); ++i1) {
        if((vdcmod_mod_id_t)(*i1)==mod_id) {
            c_binmodule* m=dynamic_cast<c_binmodule*>(*i1);
            if(m) {
                m->process_script_output("php_2",0,text);
            }
            break;
        }
    }
}

extern "C" int mod_eval_func(vdcmod_mod_id_t mod_id, const char* exp, c_vdcmod_data_t** result)
{
    int res=0;
    if(result)
        *result=NULL;
    if(!exp || !result)
        return res;

    vector<c_module*>::iterator i1;
    for(i1=modules.begin(); i1!=modules.end(); ++i1) {
        if((vdcmod_mod_id_t)(*i1)==mod_id) {
            c_binmodule* m=dynamic_cast<c_binmodule*>(*i1);
            if(m) {
                var_keeper->set_current_session(m->get_session());
                c_temp_local local;
                c_var_array* vars=NULL;
                unsigned new_local_id=local.get_params(vars);
                string debug="<vdcmod_eval_func>:1 -> ";
                debug+=exp;
                string code_str=exp;
                {
                    string next_expression;

                    stack<c_eval_iteration> st;

                    c_eval_iteration ste(m,vars,debug);
                    ste.proc_eval=true;
                    ste.state=0;
                    ste.return_to_state=0;
                    ste.save_return_value=true;
                    ste.local_id=new_local_id;
                    st.push(ste);

                    if(!eval_function(m,code_str,vars,m->r_functions,debug,st)) {
                        // nothing
                    } else {
                        eval_loop(st);
                        if(st.top().return_value) {
                            m->convert_var(st.top().return_value,*result);
                        }
                        res=st.top().exception_reported ? 0 : 1;
                    }
                    var_keeper->set_current_session(SESSION_RESERVED_ID);
                }
            }
            break;
        }
    }

    return res;
}

c_binmodule::c_binmodule()
    : c_module()
{
    mod_hnd=NULL;
    init_func=NULL;
    done_func=NULL;
}

c_binmodule::c_binmodule(const c_binmodule& orig)
    : c_module(orig)
{
    mod_hnd=orig.mod_hnd;
    init_func=orig.init_func;
    done_func=orig.done_func;
}

c_binmodule::~c_binmodule()
{
    unload();
}

c_module::e_rc c_binmodule::load(const string& mod_name, const string& so_name)
{
    if(so_name.empty())
        return RC_E_INV_MODNAME;
    // other sanity checks for 'so_name' are in c_module::load()
    c_module::e_rc rc=c_module::load(mod_name,so_name);
    if(RC_OK==rc) {
        string so=dir_prefix;
        so+=so_name;
        if(!lib_load(so)) {
            logger->log_print1(E_LOG_DEBUG,LOG_SCOPE_GLOBAL | LOG_SCOPE_MODULE,"","Failed to load module's .so/.dll file",true);
            unload();
            return RC_E_SO_ERROR;
        }
        char tmp[1024*2];
        *tmp='\0';
        getcwd(tmp,sizeof(tmp)-1);
        string dir=tmp;
        dir+=FILE_SLASH;
        dir+=dir_prefix;

        c_vdcmod_init* init=MY_NEW c_vdcmod_init;
        memset(init,0,sizeof(*init));
        init->version=VDCMOD_API_VERSION;
        init->size=sizeof(*init);
        init->mod_id=(vdcmod_mod_id_t)this;
        init->bind=&mod_bind_func;
        init->unbind=&mod_unbind_func;
        init->get_superglobals=&mod_get_superglobals_func;
        init->free_struct=&mod_free_structure_func;
        init->eval_func=&mod_eval_func;
        init->process_script_out=&mod_process_script_out_func;
        init->vdc_host_sizeof_long=sizeof(long);
        init->vdc_host_sizeof_long_long=sizeof(MY_LONG_LONG);
        init->bind_ex=&mod_bind_ex_func;
        init->unbind_ex=&mod_unbind_ex_func;

        c_vdcmod_init_ret* ret=MY_NEW c_vdcmod_init_ret;
        memset(ret,0,sizeof(*ret));

        (*init_func)(mod_name.c_str(),dir.c_str(),init,ret);

        MY_DELETE(init);
        MY_DELETE(ret);
    } else
        return rc;
    return RC_OK;
}

void c_binmodule::unload()
{
    if(done_func)
        (*done_func)();
    lib_unload();
    c_module::unload();
}

bool c_binmodule::execute(string func, vector<string>& string_vars, const string& debug)
{
    return c_module::execute(func,string_vars,debug);
}

bool c_binmodule::execute(string func, c_var_array* caller_vars, const string& debug, bool core_module_entrypoint)
{
    return c_module::execute(func,caller_vars,debug,core_module_entrypoint);
}

bool c_binmodule::execute(string func, const vector<c_var_base*>& args_, const string& debug)
{
    return c_module::execute(func,args_,debug);
}

bool c_binmodule::call_startup()
{
    return c_module::call_startup();
}

bool c_binmodule::call_bound_dispatch_events(c_var_array* dispatch, const vector<pair<vector<c_binder>::iterator,c_binder_data> >& matches)
{
    return c_module::call_bound_dispatch_events(dispatch,matches);
}

bool c_binmodule::get_matches(const string& type, const string& user, const string& channel, const string& line, bool before, vector<pair<vector<c_binder>::iterator,c_binder_data> >& matches)
{
    return c_module::get_matches(type,user,channel,line,before,matches);
}

bool c_binmodule::lib_load(const string& fn)
{
    if(mod_hnd)
        return true;

    mod_hnd=NULL;
    init_func=NULL;
    done_func=NULL;
#if defined(_WIN32) || defined(_WIN64) || defined(__CYGWIN__)
    mod_hnd=LoadLibrary(fn.c_str());
#else
#   ifdef HAVE_DLOPEN
    mod_hnd=dlopen(fn.c_str(),RTLD_LAZY);
#   else
    return false;
#   endif
#endif

    if(!mod_hnd)
        return false;

    void* f;

#if defined(_WIN32) || defined(_WIN64) || defined(__CYGWIN__)
    f=(void*)GetProcAddress(mod_hnd,"vdcmod_init");
    memcpy(&init_func,&f,sizeof(init_func));
#else
#   ifdef HAVE_DLOPEN
    dlerror(); // to clear any pending errors
    f=dlsym(mod_hnd,"vdcmod_init");
    memcpy(&init_func,&f,sizeof(init_func));
#   else
    // HUH?
#   endif
#endif
    if(!init_func) {
        lib_unload();
        return false;
    }

#if defined(_WIN32) || defined(_WIN64) || defined(__CYGWIN__)
    f=(void*)GetProcAddress(mod_hnd,"vdcmod_done");
    memcpy(&done_func,&f,sizeof(done_func));
#else
#   ifdef HAVE_DLOPEN
    dlerror(); // to clear any pending errors
    f=dlsym(mod_hnd,"vdcmod_done");
    memcpy(&done_func,&f,sizeof(done_func));
#   else
    // HUH?
#   endif
#endif
    if(!done_func) {
        init_func=NULL;
        lib_unload();
        return false;
    }

    return true;
}

void c_binmodule::lib_unload()
{
    init_func=NULL;
    done_func=NULL;
    if(!mod_hnd)
        return;
#if defined(_WIN32) || defined(_WIN64) || defined(__CYGWIN__)
    FreeLibrary(mod_hnd);
    mod_hnd=NULL;
#else
#   ifdef HAVE_DLOPEN
    dlclose(mod_hnd);
    mod_hnd=NULL;
#   else
    // HUH?
    mod_hnd=NULL;
#   endif
#endif
    mod_hnd=NULL;
}

/*!
    \brief Stores data for iterative walking in variable tree
    \author VooDooMan
    \version 1
    \date 2006-11-11
*/
class c_array_dump_iteration
{
public:
    tree<c_tree_node, c_single_element_allocator<tree_node_<c_tree_node> > >::iterator iter;       //!< Current position
    int depth;                              //!< Starting depth
    //#int lastd;                              //!< Last depth

    c_vdcmod_data_t* parent;                //!< Parent data (base)
    c_vdcmod_data_t* curr;                  //!< Current data (initalized to base)
    bool has_children;                      //!< If this has children

    /*!
        \brief Constructor
        \author VooDooMan
        \version 1
        \date 2006-11-11
        \param i Initializes iterator with this value
        \param d Initializes starting depth with this value
    */
    c_array_dump_iteration(tree<c_tree_node, c_single_element_allocator<tree_node_<c_tree_node> > >::iterator i, int d)
    {
        iter=i;
        depth=d;
        //#lastd=d;

        curr=NULL;

        has_children=false;
    }

    /*!
        \brief End of sub tree, closes open data
        \author VooDooMan
        \version 1
        \date 2006-11-11
    */
    void eof_subtree()
    {
        if(DATA_TYPE_ARRAY==curr->type) {
            // all members are clean already
        }
        if(DATA_TYPE_ARRAY_ENTRY==curr->type) {
            curr->u_data.c_array_entry.next=NULL;
        }
    }
};

void c_binmodule::_converted_vars_dump(const string& fn, c_vdcmod_data_t* result)
{
    FILE* f=fopen(fn.c_str(),"w");
    if(!f)
        return;
    stack<pair<int, c_vdcmod_data_t*> > st;
    st.push(make_pair(0,result));

    int iters=0;
    int max_iters=1000;

    for(;;) {
        if(st.empty() || iters && st.size()==1)
            break;

        if(iters++>=max_iters) {
            fprintf(f,"\n=== MAX ITERATIONS REACHED (%s) ===\n",tos(max_iters).c_str());
            break;
        }

        string indent;
        for(size_t ii=0; ii<st.size()-1; ++ii)
            indent+="    ";

        if(NULL==st.top().second) {
            while(!st.empty() && !st.top().second) {
                indent="";
                for(size_t ii=0; ii<st.size()-1; ++ii)
                    indent+="    ";

                fprintf(f,"%s%s\n",indent.c_str(),"NULL");
                fprintf(f,"\n");
                st.pop();
            }
            continue;
        }
        switch(st.top().second->type) {
            case DATA_TYPE_UNSET:
                {
                    fprintf(f,"%s%s\n",indent.c_str(),"UNSET");
                    st.pop();
                }
                break;
            case DATA_TYPE_BOOL:
                {
                    fprintf(f,"%s%s\n",indent.c_str(),"BOOL");
                    fprintf(f,"%s%s\n",indent.c_str(),st.top().second->u_data.c_bool.data? "TRUE" : "FALSE");
                    st.pop();
                }
                break;
            case DATA_TYPE_INT:
                {
                    fprintf(f,"%s%s\n",indent.c_str(),"INT");
                    fprintf(f,"%s%s\n",indent.c_str(),tos(st.top().second->u_data.c_int.data).c_str());
                    fprintf(f,"%s%s\n",indent.c_str(),tos(st.top().second->u_data.c_big_int.data).c_str());
                    st.pop();
                }
                break;
            case DATA_TYPE_STRING:
                {
                    fprintf(f,"%s%s\n",indent.c_str(),"STRING");
                    fprintf(f,"%s'%s'\n",indent.c_str(),st.top().second->u_data.c_str.data);
                    st.pop();
                }
                break;
            case DATA_TYPE_ARRAY:
                {
                    fprintf(f,"%s%s\n\n",indent.c_str(),"ARRAY");
                    st.push(make_pair(1000,st.top().second->u_data.c_array.value));

                    if(!st.top().second) {
                        fprintf(f,"%s%s\n",indent.c_str(),"NULL");
                        fprintf(f,"\n");
                        st.pop();
                        st.pop();
                    }
                }
                break;
            case DATA_TYPE_ARRAY_ENTRY:
                {
                    if(1000==st.top().first) {
                        fprintf(f,"%s%s\n",indent.c_str(),"ARRAY -> IDX");
                        st.top().first++;
                        st.push(make_pair(1000,st.top().second->u_data.c_array_entry.index));
                        break;
                    }
                    if(1001==st.top().first) {
                        fprintf(f,"%s%s\n",indent.c_str(),"ARRAY -> VAL");
                        st.top().first++;
                        st.push(make_pair(1000,st.top().second->u_data.c_array_entry.value));
                        break;
                    }
                    if(1002==st.top().first) {
                        fprintf(f,"%s%s\n\n",indent.c_str(),"ARRAY -> NEXT");
                        st.top().first++;
                        if(!st.top().second->u_data.c_array_entry.next) {
                            fprintf(f,"%s%s\n",indent.c_str(),"NULL");
                            fprintf(f,"\n");
                            st.pop();
                            st.pop();
                        } else {
                            c_vdcmod_data_t* next=st.top().second->u_data.c_array_entry.next;
                            st.pop();
                            st.push(make_pair(1000,next));
                        }
                        break;
                    }
                    // else
                    st.pop();
                    st.pop();
                }
                break;
            default:
                fprintf(f,"%s%s%s\n\n",indent.c_str(),"INVALID ENTRY TYPE: 0x",toshex(st.top().second->type).c_str());
        }
    }

    fclose(f);
    f=NULL;
}

void c_binmodule::convert_vars(c_var_array* vars, c_vdcmod_data_t*& result)
{
    result=NULL;
    if(!vars)
        return;

    tree<c_tree_node, c_single_element_allocator<tree_node_<c_tree_node> > >::iterator iter_i1;
    if(var_keeper->var_lookup(vars,iter_i1))
        goto gotcha1;
    for(iter_i1=var_keeper->variables.begin(); iter_i1!=var_keeper->variables.end(); ++iter_i1) {
        if((*iter_i1).data2.second==vars)
            break;
    }
    if(var_keeper->variables.end()==iter_i1) {
        // huh?
        return;
    }
gotcha1:

    c_vdcmod_data_t* root=MY_NEW c_vdcmod_data_t;
    memset(root,0,sizeof(*root));
    root->type=DATA_TYPE_ARRAY;

    root->u_data.c_array.value=NULL;

    c_vdcmod_data_t* first=MY_NEW c_vdcmod_data_t;
    memset(first,0,sizeof(*first));
    first->type=DATA_TYPE_ARRAY_ENTRY;

    root->u_data.c_array.value=first;

    result=root;

    dealloc1.push_back(make_pair(result,root));
    dealloc1.push_back(make_pair(result,first));

    int iter_depth=var_keeper->variables.depth(iter_i1)+1;

    stack<c_array_dump_iteration> st;
    st.push(c_array_dump_iteration(iter_i1,iter_depth));
    st.top().parent=first;
    st.top().has_children=true;

    c_vdcmod_data_t* parent=NULL;

    bool first_iteration=true;
    bool make_sub_array_only=false;
    int initial_depth=-1;
    bool should_increment=false;
    bool another_scope=false;

    int lastd=var_keeper->variables.depth(iter_i1);

    for(;;) {

        if(st.empty())
            return;

        if(!st.top().curr) {
            st.top().curr=st.top().parent;
        }

        if(first_iteration)
            lastd++;

        st.top().iter++;

        if(-1==initial_depth)
            initial_depth=var_keeper->variables.depth(st.top().iter);

again:
        make_sub_array_only=false;

        tree<c_tree_node, c_single_element_allocator<tree_node_<c_tree_node> > >::iterator i1=st.top().iter;

        bool eof=false;
        if(var_keeper->variables.end()==st.top().iter)
            eof=true;
        else if(var_keeper->variables.depth(st.top().iter)<initial_depth)
            eof=true;

        if(eof) {
            while(!st.empty()) {
                st.top().eof_subtree();
                st.pop();
            }
            return;
        }

        if(var_keeper->variables.depth(st.top().iter)>lastd) {

            should_increment=false;

            if(!parent && !another_scope) {
                should_increment=true;
                st.top().iter--;
                make_sub_array_only=true;
                goto make_sub_array;
            }

            for(;;) {
made_sub_array_only:
                if(should_increment)
                    st.top().iter++;
                break;
            }

            if(another_scope && !parent) {
                another_scope=false;

                c_vdcmod_data_t* curr=MY_NEW c_vdcmod_data_t;
                dealloc1.push_back(make_pair(result,curr));
                memset(curr,0,sizeof(*curr));
                curr->type=DATA_TYPE_ARRAY;

                parent=st.top().curr;
                parent->u_data.c_array_entry.next=curr;
                parent=curr;
            }

            another_scope=false;

            c_vdcmod_data_t* curr=parent;

            st.top().has_children=true;
            st.push(c_array_dump_iteration(st.top().iter,lastd+1));

            st.top().parent=curr;
            st.top().curr=curr;

            lastd++;

            if(!should_increment) {
                continue;
            } else {
                should_increment=false;
                goto again;
            }
            // NOTREACH
        } else if(var_keeper->variables.depth(st.top().iter)<lastd) {

            parent=NULL;
            another_scope=true;

            while(var_keeper->variables.depth(st.top().iter)<lastd) {
                st.pop();
                if(st.empty())
                    return;
                st.top().iter.skip_children();
                st.top().iter++;

                lastd--;
            }
            goto again;
        }

make_sub_array:

        another_scope=false;

        if(!first_iteration && DATA_TYPE_ARRAY_ENTRY==st.top().curr->type && st.top().curr->u_data.c_array_entry.index) {
            if(st.top().has_children) {
                c_vdcmod_data_t* value=MY_NEW c_vdcmod_data_t;
                dealloc1.push_back(make_pair(result,value));
                memset(value,0,sizeof(*value));
                value->type=(e_data_type_t)0;

                st.top().curr->u_data.c_array_entry.next=value;
                st.top().curr=value;
            }
        }
        first_iteration=false;

        if(DATA_TYPE_ARRAY==st.top().curr->type && !st.top().curr->u_data.c_array.value) {
            c_vdcmod_data_t* value=MY_NEW c_vdcmod_data_t;
            dealloc1.push_back(make_pair(result,value));
            memset(value,0,sizeof(*value));
            value->type=(e_data_type_t)0;

            st.top().curr->u_data.c_array.value=value;
            st.top().curr=value;
        }
        if(DATA_TYPE_ARRAY==st.top().curr->type && st.top().curr->u_data.c_array.value)
            st.top().curr=st.top().curr->u_data.c_array.value;

        c_index_base* key=(*st.top().iter).data2.first;
        c_var_base* val=(*st.top().iter).data2.second;

        if(DATA_TYPE_ARRAY_ENTRY==st.top().curr->type && st.top().curr!=first) {
            st.top().curr->u_data.c_array_entry.next=MY_NEW c_vdcmod_data_t;
            dealloc1.push_back(make_pair(result,st.top().curr->u_data.c_array_entry.next));
            memset(st.top().curr->u_data.c_array_entry.next,0,sizeof(*(st.top().curr->u_data.c_array_entry.next)));
            st.top().curr=st.top().curr->u_data.c_array_entry.next;
        }

        st.top().curr->type=DATA_TYPE_ARRAY_ENTRY;

        // key
        if(!st.top().curr->u_data.c_array_entry.index) {
            st.top().curr->u_data.c_array_entry.index=MY_NEW c_vdcmod_data_t;
            dealloc1.push_back(make_pair(result,st.top().curr->u_data.c_array_entry.index));
            memset(st.top().curr->u_data.c_array_entry.index,0,sizeof(*(st.top().curr->u_data.c_array_entry.index)));
            c_vdcmod_data_t* x=st.top().curr->u_data.c_array_entry.index;

            if(key->type()==c_index_base::IDX_INT) {
                c_index_int* i=dynamic_cast<c_index_int*>(key);
                x->type=DATA_TYPE_INT;
                // BUG: NEXT LINE: degrading number (warning okay)
#if defined(_MSC_VER)
#   pragma warning(push)
#   pragma warning(disable:4244)    // warning C4244: '=' : conversion from '__int64' to 'long', possible loss of data
#endif
                x->u_data.c_int.data=i->get();
#if defined(_MSC_VER)
#   pragma warning(pop)
#endif
                x->u_data.c_big_int.data=i->get();
            } else if(key->type()==c_index_base::IDX_STR) {
                c_index_str* i=dynamic_cast<c_index_str*>(key);
                x->type=DATA_TYPE_STRING;
                char* s=MY_NEW char[i->get().length()+1];
                strcpy(s,i->get().c_str());
                dealloc2.push_back(make_pair(result,s));
                x->u_data.c_str.data=s;
            } else {
                x->type=DATA_TYPE_UNSET;
            }
        }

        // val
        c_vdcmod_data_t* x;
        {
            st.top().curr->u_data.c_array_entry.value=MY_NEW c_vdcmod_data_t;
            dealloc1.push_back(make_pair(result,st.top().curr->u_data.c_array_entry.value));
            memset(st.top().curr->u_data.c_array_entry.value,0,sizeof(*(st.top().curr->u_data.c_array_entry.value)));
            x=st.top().curr->u_data.c_array_entry.value;

            if(!val || typeid(*val)==typeid(c_var_unset)) {
                x->type=DATA_TYPE_UNSET;
            } else if(typeid(*val)==typeid(c_var_int)) {
                c_var_int* i=dynamic_cast<c_var_int*>(val);
                x->type=DATA_TYPE_INT;
                // BUG: NEXT LINE: degrading number (warning okay)
#if defined(_MSC_VER)
#   pragma warning(push)
#   pragma warning(disable:4244)    // warning C4244: '=' : conversion from '__int64' to 'long', possible loss of data
#endif
                x->u_data.c_int.data=i->get();
#if defined(_MSC_VER)
#   pragma warning(pop)
#endif
                x->u_data.c_big_int.data=i->get();
            } else if(typeid(*val)==typeid(c_var_string)) {
                c_var_string* i=dynamic_cast<c_var_string*>(val);
                x->type=DATA_TYPE_STRING;
                char* s=MY_NEW char[i->get().length()+1];
                strcpy(s,i->get().c_str());
                dealloc2.push_back(make_pair(result,s));
                x->u_data.c_str.data=s;
            } else if(typeid(*val)==typeid(c_var_bool)) {
                c_var_bool* i=dynamic_cast<c_var_bool*>(val);
                x->type=DATA_TYPE_BOOL;
                x->u_data.c_bool.data=i->get();
            } else if(typeid(*val)==typeid(c_var_array)) {
                // implemented below
            } else {
                x->type=DATA_TYPE_UNSET;
            }
        }

        if(typeid(*val)==typeid(c_var_array)) {
            x->type=DATA_TYPE_ARRAY;
            st.top().curr->u_data.c_array_entry.next=NULL;
            parent=x;

            // this is handled at start of the for(;;) loop

            if(make_sub_array_only)
                goto made_sub_array_only;

            goto made_sub_array_only;
        }
        if(make_sub_array_only)
            goto made_sub_array_only;
        continue;
    }
}

void c_binmodule::convert_var(c_var_base* data, c_vdcmod_data_t*& result)
{
    if(data->get_var_type()!=VARTYPE_ARRAY) {
        c_vdcmod_data_t* res=MY_NEW c_vdcmod_data_t;
        result=res;

        dealloc1.push_back(make_pair(result,res));
        res->type=DATA_TYPE_UNSET;
        if(typeid(*data)==typeid(c_var_int)) {
            res->type=DATA_TYPE_INT;
            c_var_int* i=dynamic_cast<c_var_int*>(data);
            // BUG: NEXT LINE: degrading number (warning okay)
#if defined(_MSC_VER)
#   pragma warning(push)
#   pragma warning(disable:4244)    // warning C4244: '=' : conversion from '__int64' to 'long', possible loss of data
#endif
            res->u_data.c_int.data=i->get();
#if defined(_MSC_VER)
#   pragma warning(pop)
#endif
            res->u_data.c_big_int.data=i->get();
        }
        if(typeid(*data)==typeid(c_var_string)) {
            res->type=DATA_TYPE_STRING;
            c_var_string* s=dynamic_cast<c_var_string*>(data);
            res->u_data.c_str.data=MY_NEW char[s->get().length()+1];
            dealloc2.push_back(make_pair(result,res->u_data.c_str.data));
            strcpy(res->u_data.c_str.data,s->get().c_str());
        }
        if(typeid(*data)==typeid(c_var_bool)) {
            res->type=DATA_TYPE_BOOL;
            c_var_bool* b=dynamic_cast<c_var_bool*>(data);
            res->u_data.c_bool.data=b->get()?'\x1':'\x0';
        }
    } else {
        c_var_array* a=dynamic_cast<c_var_array*>(data);
        convert_vars(a,result);
    }
}

void c_binmodule::free_vars(c_vdcmod_data_t* struc)
{
    vector<pair<c_vdcmod_data_t*,c_vdcmod_data_t*> >::iterator i1;
again1:
    for(i1=dealloc1.begin(); i1!=dealloc1.end(); ++i1) {
        if((*i1).first==struc) {
            MY_DELETE((*i1).second);
            dealloc1.erase(i1);
            goto again1;
        }
    }
    vector<pair<c_vdcmod_data_t*,char*> >::iterator i2;
again2:
    for(i2=dealloc2.begin(); i2!=dealloc2.end(); ++i2) {
        if((*i2).first==struc) {
            MY_DELETE_A((*i2).second);
            dealloc2.erase(i2);
            goto again2;
        }
    }
}
