/***************************************************************************
                          binmod.h  -  description
                             -------------------
    begin                : Fri Aug 25 2006
    copyright            : (C) 2006-2007 by Marian VooDooMan Meravy
    email                : ghostvoodooman -aT= users (dot) sourceforge.net
 ***************************************************************************/

/* $Id: binmod.h,v 1.33 2008/08/23 15:10:38 ghostvoodooman Exp $ */

/***************************************************************************

VooDoo cIRCle - an IRC (ro)bot
Copyright (C) 2006-2008 by Marian VooDooMan Meravy
  (ghostvoodooman -aT= users (dot) sourceforge.net)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************/

/*!
    \file
    \brief Provides interface for binary modules
    \version $Id: binmod.h,v 1.33 2008/08/23 15:10:38 ghostvoodooman Exp $
*/

#ifndef BINMOD_H__
#define BINMOD_H__

#if defined(HAVE_CONFIG_H)
#   include <config.h>
#endif

#include <string>
#include <vector>
#include <stack>
#include <list>
#include <map>
#include <deque>

#ifdef HAVE_DLOPEN
#   include <dlfcn.h>
#endif

using namespace std;

#include "utils.h"      // because we want to enable mem leak detection

#include "params.h"
#include "compat.h"

#include "vdcmod.h"
#include "shared.h"

#if defined(_WIN32) || defined(_WIN64)
#   define MY_MOD_HANDLE    HINSTANCE
#else
#   define MY_MOD_HANDLE    void*
#endif

#include "utils/assert.h"

/*!
    \brief Stores one binary module
    \author VooDooMan
    \version 1
    \date 2005, 2006-08-25
*/
class c_binmodule : public c_module
{
protected:

    MY_MOD_HANDLE mod_hnd;          //!< Module handle
    vdcmod_init_func init_func;     //!< Init function
    vdcmod_done_func done_func;     //!< Done function

    /*!
        \brief Loads a dynamic link library / shared object
        \version 1
        \author VooDooMan
        \date 2006-08-26
        \param fn File name
        \retval true Okay
        \retval false Error
    */
    bool lib_load(const string& fn);

    /*!
        \brief Unloads a dynamic link library / shared object
        \version 1
        \author VooDooMan
        \date 2006-08-26
    */
    void lib_unload();

    vector<pair<c_vdcmod_data_t*,c_vdcmod_data_t*> > dealloc1;  //!< Should be MY_DELETE'd
    vector<pair<c_vdcmod_data_t*,char*> > dealloc2;             //!< Should be MY_DELETE_A'd

public:

    c_binmodule();

    c_binmodule(const c_binmodule& orig);
    virtual ~c_binmodule();

    /*!
        \brief Loads module
        \author VooDooMan
        \version 2
        \date 2005, 2006-08-25
        \param mod_name Name of module
        \param so_name Name of shared object/DLL
    */
    virtual c_module::e_rc load(const string& mod_name, const string& so_name);

    /*!
        \brief Unloads module
        \author VooDooMan
        \version 2
        \date 2005, 2006-08-25
    */
    virtual void unload();

    /*!
        \brief Executes function
        \author VooDooMan
        \version 2
        \date 2005, 2006-08-25
        \param func Name of function
        \param string_vars List of string arguments
        \param debug Debug string
        \retval true OK
        \retval false Procedure/function not found
    */
    virtual bool execute(string func, vector<string>& string_vars, const string& debug);

    /*!
        \brief Executes function
        \author VooDooMan
        \version 2
        \date 2005, 2006-08-25
        \param func Name of function
        \param caller_vars Caller's variables
        \param debug Debug string
        \param core_module_entrypoint 'true' if this is core module entry point (should be always 'false')
        \retval true OK
        \retval false Procedure/function not found
    */
    virtual bool execute(string func, c_var_array* caller_vars, const string& debug, bool core_module_entrypoint = false);

    /*!
        \brief Executes function
        \author VooDooMan
        \version 2
        \date 2005, 2006-08-25
        \param func Name of function
        \param args_ Argument list
        \param debug Debug string
        \retval true OK
        \retval false Procedure/function not found
    */
    virtual bool execute(string func, const vector<c_var_base*>& args_, const string& debug);

    virtual bool call_startup();

    static bool call_bound_dispatch_events(c_var_array* dispatch, const vector<pair<vector<c_binder>::iterator,c_binder_data> >& matches);
    virtual bool get_matches(const string& type, const string& user, const string& channel, const string& line, bool before, vector<pair<vector<c_binder>::iterator,c_binder_data> >& matches);

    /*!
        \brief Deallocated variables (after call to "convert_vars")
        \version 1
        \author VooDooMan
        \date 2006-08-27
        \param struc Root variable
    */
    void free_vars(c_vdcmod_data_t* struc);

    /*!
        \brief Convers variables for binary module
        \version 1
        \author VooDooMan
        \date 2006-08-27
        \param vars Variables
        \param result Receives result (should be deallocated by call to "free_vars")
    */
    void convert_vars(c_var_array* vars, c_vdcmod_data_t*& result);

    /*!
        \brief Convers variable for binary module
        \version 1
        \author VooDooMan
        \date 2006-09-24
        \param data Variable
        \param result Receives result (should be deallocated by call to "free_vars")
    */
    void convert_var(c_var_base* data, c_vdcmod_data_t*& result);

    /*!
        \brief Dumps converted variables for binary module
        \version 1
        \author VooDooMan
        \date 2006-01-12
        \param fn File name
        \param result Converted variables
    */
    void _converted_vars_dump(const string& fn, c_vdcmod_data_t* result);
};

#endif

