/***************************************************************************
                          bitarray.cpp  -  description
                             -------------------
    begin                : Mon Feb 5 2007
    copyright            : (C) 2007 by Marian VooDooMan Meravy
    email                : ghostvoodooman -aT= users (dot) sourceforge.net
 ***************************************************************************/

/* $Id: bitarray.cpp,v 1.21 2008/11/24 12:09:23 ghostvoodooman Exp $ */

/***************************************************************************

VooDoo cIRCle - an IRC (ro)bot
Copyright (C) 2007-2008 by Marian VooDooMan Meravy
  (ghostvoodooman -aT= users (dot) sourceforge.net)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************/

/*!
    \file
    \brief Provides fast binary array implementation
    \version $Id: bitarray.cpp,v 1.21 2008/11/24 12:09:23 ghostvoodooman Exp $
*/

//---------------------------------------------------------------------------

#if defined(HAVE_CONFIG_H)
#   include <config.h>
#endif

#include "utils/memleak.h"
#include "utils/assert.h"
#include "utils.h"

#include "params.h"

#include "bitarray.h"

#include "utils/precomp.h"

#if defined(_MSC_VER) || defined(__BORLANDC__)
#   pragma hdrstop
#endif

// ---

/*!
    \brief Constructor
    \author VooDooMan
    \version 1
    \date 2007-02-05
*/
c_bitarray_except::c_bitarray_except()
{
    MY_ASSERT_NOTREACH();
}

/*!
    \brief Constructor
    \author VooDooMan
    \version 1
    \date 2007-02-05
*/
c_bitarray::c_bitarray()
{
    size=C_BITARRAY_INCREMENT;
    array=(unsigned char*)my_malloc(size);
    if(!array)
        throw c_bitarray_except();
    memset(array,0,size);
    first_zero_bit=0;
    first_zero_bit_valid=true;
}

/*!
    \brief Destructor
    \author VooDooMan
    \version 1
    \date 2007-02-05
*/
c_bitarray::~c_bitarray()
{
    if(!array)
        return;
    size=0;
    my_free(array);
}

/*!
    \brief Sets bit to value of 1
    \author VooDooMan
    \version 1
    \date 2007-02-05
    \param pos Offset (bit number)
*/
void c_bitarray::set(unsigned long pos)
{
    size_t offset=(size_t)pos/8;
    if(offset>=size)
        more(offset);
    *(array+offset)|=1U << (pos%8U);
    if(first_zero_bit_valid && first_zero_bit==pos)
        first_zero_bit_valid=false;
}

/*!
    \brief Sets bit to value of 0
    \author VooDooMan
    \version 1
    \date 2007-02-05
    \param pos Offset (bit number)
*/
void c_bitarray::clear(unsigned long pos)
{
    size_t offset=(size_t)pos/8;
    if(offset>=size) {
        // do nothing: everything outside of range has value of 0 (virtually)
        return;
    }
    *(array+offset)&=0xffU ^ (1U << (pos%8U));
    if(first_zero_bit_valid && first_zero_bit>pos) {
        first_zero_bit=pos;
    }
}

/*!
    \brief Allocates more space (reallocates buffer)
    \author VooDooMan
    \version 1
    \date 2007-02-05
    \param offset At least this number of bytes (ensures this offset will be present in the array)
    \warning Not MT-safe!
*/
void c_bitarray::more(size_t offset)
{
    if(offset<size)
        return;
    size_t new_size=(offset+1)/C_BITARRAY_INCREMENT;
    new_size++;
    new_size*=C_BITARRAY_INCREMENT;
    unsigned char* new_array=(unsigned char*)my_realloc(array,new_size);
    if(!new_array)
        throw c_bitarray_except();
    array=new_array;
    unsigned char* begin=array+size;
    memset(begin,0,new_size-size);
    size=new_size;
}

/*!
    \brief Finds first bit of value 0
    \author VooDooMan
    \version 1
    \date 2007-02-05
    \return Returns offset of that bit
    \warning Not MT-safe!
*/
size_t c_bitarray::find_first_zero_bit()
{
    if(first_zero_bit_valid)
        return first_zero_bit;

    for(register unsigned long i1=0; i1<size; ++i1) {
        if(0xff == array[i1])
            continue;
        if(!(array[i1] & 1))
            first_zero_bit=i1*8 + 0;
        else if(!(array[i1] & (1 << 1)))
            first_zero_bit=i1*8 + 1;
        else if(!(array[i1] & (1 << 2)))
            first_zero_bit=i1*8 + 2;
        else if(!(array[i1] & (1 << 3)))
            first_zero_bit=i1*8 + 3;
        else if(!(array[i1] & (1 << 4)))
            first_zero_bit=i1*8 + 4;
        else if(!(array[i1] & (1 << 5)))
            first_zero_bit=i1*8 + 5;
        else if(!(array[i1] & (1 << 6)))
            first_zero_bit=i1*8 + 6;
        else if(!(array[i1] & (1 << 7)))
            first_zero_bit=i1*8 + 7;
        first_zero_bit_valid=true;
        return first_zero_bit;
    }

    first_zero_bit=size*8;
    first_zero_bit_valid=true;
    return first_zero_bit;
}

