/***************************************************************************
                          bitarray.h  -  description
                             -------------------
    begin                : Mon Feb 5 2007
    copyright            : (C) 2007 by Marian VooDooMan Meravy
    email                : ghostvoodooman -aT= users (dot) sourceforge.net
 ***************************************************************************/

/* $Id: bitarray.h,v 1.18 2008/03/11 21:11:26 ghostvoodooman Exp $ */

/***************************************************************************

VooDoo cIRCle - an IRC (ro)bot
Copyright (C) 2007-2008 by Marian VooDooMan Meravy
  (ghostvoodooman -aT= users (dot) sourceforge.net)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************/

/*!
    \file
    \brief Provides fast binary array implementation
    \version $Id: bitarray.h,v 1.18 2008/03/11 21:11:26 ghostvoodooman Exp $
*/

//---------------------------------------------------------------------------

#ifndef BITARRAY_H__
#define BITARRAY_H__

#if defined(HAVE_CONFIG_H)
#   include <config.h>
#endif

#include "utils.h"      // because we want to enable mem leak detection

#include "params.h"

/*!
    \brief Exception class for unsuccessful memory re/allocation (fatal error)
    \author VooDooMan
    \version 1
    \date 2007-02-05
*/
class c_bitarray_except
{
public:

    c_bitarray_except();
};

/*!
    \def C_BITARRAY_INCREMENT
    \brief Number of bytes used for reallocating array at once
*/
#define C_BITARRAY_INCREMENT                (512)

/*!
    \brief Binary array implementation
    \author VooDooMan
    \version 1
    \date 2007-02-05
*/
class c_bitarray
{
    unsigned char* array;   //!< Binary array buffer
    size_t size;            //!< Size (in bytes) of binary array buffer
    size_t first_zero_bit;  //!< Offset of first zero bit in the array
    bool first_zero_bit_valid;      //!< If "first_zero_bit" has valid value; else: will be computed when required

    void more(size_t offset);

public:

    c_bitarray();
    ~c_bitarray();

    void set(unsigned long pos);
    void clear(unsigned long pos);

    size_t find_first_zero_bit();
};

#endif

