/***************************************************************************
                          compat.cpp  -  description
                             -------------------
    begin                : Thu Jul 31 2005
    copyright            : (C) 2005-2007 by Marian VooDooMan Meravy
    email                : ghostvoodooman -aT= users (dot) sourceforge.net
 ***************************************************************************/

/* $Id: compat.cpp,v 1.38 2008/03/11 21:11:26 ghostvoodooman Exp $ */

/***************************************************************************

VooDoo cIRCle - an IRC (ro)bot
Copyright (C) 2005-2008 by Marian VooDooMan Meravy
  (ghostvoodooman -aT= users (dot) sourceforge.net)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the

GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************/

/*!
    \file
    \brief Provides compatibility functions
    \version $Id: compat.cpp,v 1.38 2008/03/11 21:11:26 ghostvoodooman Exp $
*/

//---------------------------------------------------------------------------

#if defined(HAVE_CONFIG_H)
#   include <config.h>
#endif

#if defined(__GNUC__)
#   ifndef _FILE_OFFSET_BITS
#       define _FILE_OFFSET_BITS        64
#   else
#       if _FILE_OFFSET_BITS < 64
#           undef _FILE_OFFSET_BITS
#           define _FILE_OFFSET_BITS    64
#       endif
#   endif
#   ifndef _GNU_SOURCE
#       define _GNU_SOURCE
#   endif
#   include <stdio.h>
#endif

#include <string>
#include <vector>
#include <stack>
#include <list>
#include <map>

#include <stdio.h>
#include <errno.h>

#include "utils/memleak.h"
#include "utils/assert.h"
#include "utils.h"
#include "params.h"

#include "compat.h"

#include "log.h"

#include "utils/precomp.h"

#if defined(_MSC_VER) || defined(__BORLANDC__)
#   pragma hdrstop
#endif

using namespace std;

/*!
    \brief Portable large file 'fseek()' implementation
    \version 1
    \date 2008-03-07
    \author VooDooMan
    \param file File descriptor
    \param offset Offset to seek to
    \param whence 'SEEK_SET', 'SEEK_CUR', or 'SEEK_END'
    \param my_errno Returns errno
    \return Returns 0 (zero) on success, other on error (see 'my_errno')
*/
int my_fseek(FILE* file, MY_LONG_LONG offset, int whence, int& my_errno)
{
    int rc=-1;
    errno= my_errno= 0;
#if defined(_MSC_VER)
    rc=_fseeki64(file,static_cast<__int64>(offset),whence);
    my_errno=errno;
    errno=0;
#elif defined(__MINGW32__)
    // Dev-C++
    rc=fseeko64(file,offset,whence);
    my_errno=errno;
    errno=0;
#elif defined(__BORLANDC__)
    // FIXME: Does Borland have 64-bit version? fallback to 32-bit
    if(sizeof(offset)>sizeof(long) && offset!=static_cast<MY_LONG_LONG>(static_cast<long>(offset))) {
        my_errno=EINVAL;
        return -1;
    }
    rc=fseek(file,static_cast<long>(offset),whence);
    my_errno=errno;
    errno=0;
#elif defined(HAVE_FTELLO) && defined(HAVE_FSEEKO)
    if(sizeof(offset)>sizeof(off_t) && offset!=static_cast<MY_LONG_LONG>(static_cast<off_t>(offset))) {
        my_errno=EINVAL;
        return -1;
    }
    rc=fseeko(file,offset,whence);
    my_errno=errno;
    errno=0;
#else
    if(sizeof(offset)>sizeof(long) && offset!=static_cast<MY_LONG_LONG>(static_cast<long>(offset))) {
        my_errno=EINVAL;
        return -1;
    }
    rc=fseek(file,static_cast<long>(offset),whence);
    my_errno=errno;
    errno=0;
#endif
    return rc;
}

/*!
    \brief Portable large file 'ftell()' implementation
    \version 1
    \date 2008-03-07
    \author VooDooMan
    \param file File descriptor
    \param my_errno Returns errno
    \return Returns current file position, -1 on error (see 'my_errno')
*/
MY_LONG_LONG my_ftell(FILE* file, int& my_errno)
{
    MY_LONG_LONG rc=-1;
    errno= my_errno= 0;
#if defined(_MSC_VER)
    rc=_ftelli64(file);
    my_errno=errno;
    errno=0;
#elif defined(__MINGW32__)
    // Dev-C++
    rc=ftello64(file);
    my_errno=errno;
    errno=0;
#elif defined(__BORLANDC__)
    // FIXME: Does Borland have 64-bit version? fallback to 32-bit
    rc=ftell(file);
    my_errno=errno;
    errno=0;
#elif defined(HAVE_FTELLO) && defined(HAVE_FSEEKO)
    rc=ftello(file);
    my_errno=errno;
    errno=0;
#else
    rc=ftell(file);
    my_errno=errno;
    errno=0;
#endif
    return rc;
}

/*!
    \brief Portable large file size of file detection
    \version 1
    \date 2008-03-07
    \author VooDooMan
    \param file File descriptor
    \param my_errno Returns errno
    \return Returns size of file in bytes, if 'my_errno' is 0 (zero); if 'my_errno' is not 0 (zero), then returns -1
*/
MY_LONG_LONG my_filesize(FILE* file, int& my_errno)
{
    MY_LONG_LONG save=my_ftell(file,my_errno);
    if(0!=my_errno)
        return -1;

    my_fseek(file,0,SEEK_END,my_errno);
    if(0!=my_errno)
        return -1;

    MY_LONG_LONG size=my_ftell(file,my_errno);
    if(0!=my_errno) {
        int dummy;
        my_fseek(file,save,SEEK_SET,dummy);
        return -1;
    }

    my_fseek(file,save,SEEK_SET,my_errno);
    if(0!=my_errno)
        return -1;

    return size;
}

