/***************************************************************************
                          compress.cpp  -  description
                             -------------------
    begin                : Mon Nov 7 2005
    copyright            : (C) 2005-2007 by Marian VooDooMan Meravy
    email                : ghostvoodooman -aT= users (dot) sourceforge.net
 ***************************************************************************/

/* $Id: compress.cpp,v 1.39 2008/11/24 12:09:23 ghostvoodooman Exp $ */

/***************************************************************************

VooDoo cIRCle - an IRC (ro)bot
Copyright (C) 2005-2008 by Marian VooDooMan Meravy
  (ghostvoodooman -aT= users (dot) sourceforge.net)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************/

/*!
    \file
    \brief Routines for compression
    \version $Id: compress.cpp,v 1.39 2008/11/24 12:09:23 ghostvoodooman Exp $
*/

#if defined(HAVE_CONFIG_H)
#   include <config.h>
#endif

#include <string>

#if defined(_MSC_VER) || defined(__BORLANDC__)
#   include <zlib/include/zlib.h>
#else
#   include <zlib.h>
#endif

#if (defined(_MSC_VER) || defined(__BORLANDC__)) && !defined(HAVE_ZLIB_COMPRESSBOUND)
#   define HAVE_ZLIB_COMPRESSBOUND  1
#endif

using namespace std;

#include "utils/memleak.h"
#include "utils/assert.h"
#include "utils.h"
#include "params.h"
#include "compat.h"
#include "compress.h"
#include "utils/precomp.h"

#if defined(_MSC_VER) || defined(__BORLANDC__)
#   pragma hdrstop
#endif

/*!
    \author VooDooMan
    \version 1
    \date 2005-11-07
    \brief Gets error message from zlib's error code
    \param rc Return code from zlib
    \return Returns error string [human friendly - I hope ;-)]
*/
string compress_zlib_get_str_rc(int rc)
{
    switch(rc) {
        case Z_OK:
            return (string)"OK = Z_OK";
        case Z_STREAM_END:
            return (string)"OK = Z_STREAM_END";
        case Z_NEED_DICT:
            return (string)"OK = Z_NEED_DICT";

        case Z_ERRNO:
            return (string)"ER = Z_ERRNO";
        case Z_STREAM_ERROR:
            return (string)"ER = Z_STREAM_ERROR";
        case Z_DATA_ERROR:
            return (string)"ER = Z_DATA_ERROR";
        case Z_MEM_ERROR:
            return (string)"ER = Z_MEM_ERROR";
        case Z_BUF_ERROR:
            return (string)"ER = Z_BUF_ERROR";
        case Z_VERSION_ERROR:
            return (string)"ER = Z_VERSION_ERROR";
    }
    char tmp[128];
    sprintf(tmp,"?? = %d",rc);
    return (string)tmp;
}

/*!
    \author VooDooMan
    \version 2
    \date 2005-11-07, 2007-8-02
    \brief Compress with zlib
    \param in_buf Input buffer to compress
    \param in_len Length of input buffer in bytes
    \param out_len Returns lenght in bytes of output buffer
    \param level Compression level
    \param rc Return code from zlib
    \return Returns NULL for error (use "rc" parameter), or compressed buffer in length of returned "out_len" bytes
    \warning Result (if not NULL) must be deallocated by "MY_DELETE_A"
*/
char* compress_zlib_compress(void *in_buf, unsigned long in_len, unsigned long& out_len, int level, int& rc)
{
#ifdef HAVE_ZLIB_COMPRESSBOUND
    out_len=compressBound(in_len);
#else
    out_len=in_len+3*1024;
#endif
    char* out=MY_NEW char[out_len+1024]; // i'm affraid of buffer overflow ;-)
    uLongf o=out_len;
    rc=compress2((Bytef*)out,&o,(Bytef*)in_buf,in_len,level);
    out_len=o;
    if(rc!=Z_OK) {
        MY_DELETE_A(out);
        out=NULL;
    }
    return out;
}

/*!
    \author VooDooMan
    \version 2
    \date 2005-11-07, 2007-08-2
    \brief Uncompress with zlib
    \param in_buf Input buffer to uncompress
    \param in_len Length of input buffer in bytes
    \param out_len Returns lenght in bytes of output buffer
    \param rc Return code from zlib
    \return Returns NULL for error (use "rc" parameter), or uncompressed buffer in length of returned "out_len" bytes
    \warning Result (if not NULL) must be deallocated by "MY_DELETE_A"
*/
char* compress_zlib_decompress(void *in_buf, size_t in_len, unsigned long& out_len, int& rc)
{
    char* dest=MY_NEW char[out_len+1024]; // i'm affraid of buffer overflow ;-)
    uLongf o=out_len;
    rc=uncompress(reinterpret_cast<Bytef*>(dest),&o,static_cast<Bytef*>(in_buf),static_cast<uLong>(in_len));
    out_len=o;
    if(rc!=Z_OK) {
        MY_DELETE_A(dest);
        dest=NULL;
    }
    return dest;
}

/*!
    \brief Returns zlib version string
*/
string compress_get_zlib_version()
{
    return ZLIB_VERSION;
}

