/***************************************************************************
                          conf.cpp  -  description
                             -------------------
    begin                : Thu Dec 9 2004
    copyright            : (C) 2004-2007 by Marian VooDooMan Meravy
    email                : ghostvoodooman -aT= users (dot) sourceforge.net
 ***************************************************************************/

/* $Id: conf.cpp,v 1.56 2008/08/23 15:10:39 ghostvoodooman Exp $ */

/***************************************************************************

VooDoo cIRCle - an IRC (ro)bot
Copyright (C) 2004-2008 by Marian VooDooMan Meravy
  (ghostvoodooman -aT= users (dot) sourceforge.net)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************/

/*!
    \file
    \brief Configuration file reader
    \version $Id: conf.cpp,v 1.56 2008/08/23 15:10:39 ghostvoodooman Exp $
*/

//---------------------------------------------------------------------------

#if defined(HAVE_CONFIG_H)
#   include <config.h>
#endif

#include <string>
#include <vector>
#include <stack>
#include <list>
#include <map>

#include "utils/memleak.h"
#include "utils/assert.h"
#include "utils.h"

#include "conf.h"
#include "ssl.h"
#include "log.h"

#include "params.h"

#include "shared.h"
#include "modules.h"
#include "internal_state.h"

#include "utils/precomp.h"

#if defined(_MSC_VER) || defined(__BORLANDC__)
#   pragma hdrstop
#endif

//---------------------------------------------------------------------------
#ifdef __BORLANDC__
#pragma package(smart_init)
#endif

using namespace std;

bool pass_init_=false;      //!< Indicates wheter pass.txt configuration file was loaded

int conf_umask=077;         //!< umask value for pre-start

c_pass conf_pass;           // pass.txt parsed file

c_ssl_conf ssl_conf;        //!< Configuration of SSL read from file "ssl.txt"

/*!
    \brief Loads ssl.txt configuration file
    \author VooDooMan
    \version 2
    \date 2004, 2008-03-03
    \retval true For error
    \retval false Okay.
*/
bool conf_ssl_init()
{
    return ssl_conf.rehash();
}

/*!
    \brief Loads pass.txt configuration file
    \author VooDooMan
    \version 1
    \date 2005-08-02
    \retval true For error
    \retval false Okay.
*/
bool c_pass::init()
{
    if(pass_init_)
        return false;

    FILE* f=fopen("pass.txt","r");
    if(f==NULL)
        return true;

    string in_section="";
    map<string,string> data;

    bool empty=false;

    bool got_eof=false;
    do {
        char ln[10240+1];
        if(!feof(f)) {
            if(fgets(ln,1024*10-1,f)==NULL) {
                got_eof=true;
            }
        } else
            got_eof=true;

        if(got_eof)
            strcpy(ln,"");

        ln[10240]=0;
        if(ln[0] && ln[strlen(ln)-1]=='\r')
            ln[strlen(ln)-1]=0;
        if(ln[0] && ln[strlen(ln)-1]=='\n')
            ln[strlen(ln)-1]=0;
        if(ln[0] && ln[strlen(ln)-1]=='\r')
            ln[strlen(ln)-1]=0;
        if(ln[0] && ln[strlen(ln)-1]=='\n')
            ln[strlen(ln)-1]=0;

        if(got_eof || (strlen(ln) && ln[0]=='[')) {
            if(!data.empty()) {
                map<string,map<string, string> >::iterator i=conf_pass.sections.find(in_section);
                if(i!=conf_pass.sections.end()) {
                    string log="OOPS: Erasing previous duplicate section \"";
                    log+=in_section;
                    log+="\"";
                    log_bot(log);
                    log_debug(log);
                    conf_pass.sections.erase(i);
                }
                pair<string,map<string,string> > p;
                p.first=in_section;
                p.second=data;
                conf_pass.sections.insert(p);
                data.clear();
            }
            if(got_eof)
                break;
            in_section=ln;
            if(in_section.length() && in_section[0]=='[')
                in_section.erase(0,1);
            if(in_section.length() && in_section[in_section.length()-1]==']')
                in_section.erase(in_section.length()-1,1);
            continue;
        }

        if(strlen(ln)==0)
            continue;

        string id_, val_;
        int pos=0;
        for(unsigned int i1=0; i1<strlen(ln); ++i1) {
            if(pos==0 && ln[i1]=='=') {
                pos++;
                continue;
            }
            if(pos==0)
                id_+=ln[i1];
            if(pos==1)
                val_+=ln[i1];
        }

        if(in_section.empty()) {
            empty=true;

            in_section=id_;
            id_="remote_password";

            map<string,map<string, string> >::iterator i=sections.find(in_section);
            if(i!=conf_pass.sections.end()) {
                string log="OOPS: Erasing previous duplicate section \"";
                log+=in_section;
                log+="\"";
                log_bot(log);
                log_debug(log);
                sections.erase(i);
            }

            pair<string,string> p;

            p.first=id_;
            p.second=val_;
            data.insert(p);

            p.first="local_password";
            p.second="";
            data.insert(p);

            p.first="use_ssl";
            p.second="0";
            data.insert(p);

            p.first="disabled";
            p.second="0";
            data.insert(p);

            p.first="botnet12_compatibility";
            p.second="0";
            data.insert(p);

            p.first="zlib_level";
            p.second="5";
            data.insert(p);

            {
                pair<string,map<string,string> > p;
                p.first=in_section;
                p.second=data;
                conf_pass.sections.insert(p);
            }
            data.clear();

            in_section="";
            continue;
        }

        pair<string,string> p;
        p.first=id_;
        p.second=val_;

        map<string,string>::iterator i=data.find(id_);
        if(i!=data.end()) {
            string log="OOPS: Erasing previous duplicate key \"";
            log+=id_;
            log+="\" with value \"";
            log+=val_;
            log+="\" in section \"";
            log+=in_section;
            log+="\"";
            log_bot(log);
            log_debug(log);
            data.erase(i);
        }
        data.insert(p);
    } while(!got_eof);
    fclose(f);
    pass_init_=true;

    if(empty) {
        string log="Writing new pass.txt file.";
        log_bot(log);
        backup();
    }

    return false;
}

/*!
    \brief Returns password of remote user encoded in MD5
    \author VooDooMan
    \version 1
    \date 2005-08-02
    \retval "" for no entry
    \retval "disbaled" for disabled access
*/
string c_pass::get_remote_password_md5(string user)
{
    if(init()) {
        string log="OOPS: Cannot load pass.txt";
        log_bot(log);
        log_debug(log);
        return (string)"";
    }
    map<string,map<string,string> >::iterator i=sections.find(user);
    if(i==sections.end()) {
        string log="OOPS: Cannot get password for user \"";
        log+=user;
        log+="\" form pass.txt";
        log_debug(log);
        return (string)"disabled";
    }
    map<string,string>::iterator i2=(*i).second.find("disabled");
    if(i2!=(*i).second.end() && (*i2).second.compare("0")) // !="0"
        return (string)"disabled";

    i2=(*i).second.find("remote_password");
    if(i2!=(*i).second.end())
        return (*i2).second;

    return (string)"";
}

/*!
    \brief Saves new configuration file
    \author VooDooMan
    \version 1
    \date 2005-08-02
*/
void c_pass::backup()
{
    if(init()) {
        string log="OOPS: Cannot load pass.txt - cannot backup.";
        log_bot(log);
        log_debug(log);
        return;
    }

    FILE* f=fopen("pass.txt","w");
    if(!f) {
        string log="OOPS: Error writing new pass.txt file.";
        log_debug(log);
        log_bot(log);
        return;
    }
    map<string,map<string,string> >::iterator i1;
    for(i1=sections.begin(); i1!=sections.end(); ++i1) {
        fprintf(f,"[%s]\n",(*i1).first.c_str());
        map<string,string>::iterator i2;
        for(i2=(*i1).second.begin(); i2!=(*i1).second.end(); ++i2)
            fprintf(f,"%s=%s\n",(*i2).first.c_str(),(*i2).second.c_str());
        fprintf(f,"\n");
    }
    fclose(f);
}

/*!
    \brief Invalidates cache
    \author VooDooMan
    \version 1
    \date 2005-08-02
*/
void c_pass::rehash()
{
    conf_pass.sections.clear();
    pass_init_=false;

    conf_ssl_init();
}

/*!
    \brief Sets new password
    \author VooDooMan
    \version 1
    \date 2005-08-02
*/
void c_pass::set_remote_password_md5(string user, string md5pass)
{
    if(init()) {
        string log="OOPS: Cannot load pass.txt - cannot set new password.";
        log_bot(log);
        log_debug(log);
        return;
    }
    map<string,map<string,string> >::iterator i=sections.find(user);
    if(i==sections.end()) {
        map<string,string> data;

        pair<string,string> p;

        p.first="remote_password";
        p.second=md5pass;
        data.insert(p);

        p.first="local_password";
        p.second="";
        data.insert(p);

        p.first="use_ssl";
        p.second="0";
        data.insert(p);

        p.first="disabled";
        p.second="1";
        data.insert(p);

        p.first="botnet12_compatibility";
        p.second="0";
        data.insert(p);

        {
            pair<string,map<string,string> > p;
            p.first=user;
            p.second=data;
            conf_pass.sections.insert(p);
        }
        data.clear();

        string log="Created new user \"";
        log+=user;
        log+="\"";
        log_bot(log);
        return;
    }

    map<string,string>::iterator i2=(*i).second.find("remote_password");
    if(i2==(*i).second.end()) {
        pair<string,string> p;
        p.first="remote_password";
        p.second=md5pass;

        (*i).second.insert(p);
        return;
    }
    (*i2).second=md5pass;
}

/*!
    \brief Sets new password
    \author VooDooMan
    \version 1
    \date 2005-08-03
*/
void c_pass::set_local_password_md5(string user, string md5pass)
{
    if(init()) {
        string log="OOPS: Cannot load pass.txt - cannot set new password.";
        log_bot(log);
        log_debug(log);
        return;
    }
    map<string,map<string,string> >::iterator i=sections.find(user);
    if(i==sections.end()) {
        map<string,string> data;

        pair<string,string> p;

        p.first="remote_password";
        p.second="";
        data.insert(p);

        p.first="local_password";
        p.second=md5pass;
        data.insert(p);

        p.first="use_ssl";
        p.second="0";
        data.insert(p);

        p.first="disabled";
        p.second="1";
        data.insert(p);

        p.first="botnet12_compatibility";
        p.second="0";
        data.insert(p);

        {
            pair<string,map<string,string> > p;
            p.first=user;
            p.second=data;
            conf_pass.sections.insert(p);
        }
        data.clear();

        string log="Created new user \"";
        log+=user;
        log+="\"";
        log_bot(log);
        return;
    }

    map<string,string>::iterator i2=(*i).second.find("local_password");
    if(i2==(*i).second.end()) {
        pair<string,string> p;
        p.first="local_password";
        p.second=md5pass;

        (*i).second.insert(p);
        return;
    }
    (*i2).second=md5pass;
}

/*!
    \brief Returns true if backward compatibility to BOTNET protocol 12 and below authentication is allowed
    \author VooDooMan
    \version 1
    \date 2005-08-02
    \param user Name of user
    \retval true For enabled compatibility
*/
bool c_pass::get_botnet12_compatibility(string UNUSEDARG(user))
{
    return false;
}

/*!
    \brief Sets if backward compatibility to BOTNET protocol 12 and below authentication is allowed
    \author VooDooMan
    \version 1
    \date 2005-08-03
    \param user Name of user
    \param set Set to "true" to allow
*/
void c_pass::set_botnet12_compatibility(string UNUSEDARG(user), bool UNUSEDARG(set))
{
}

/*!
    \brief Returns password of local user encoded in MD5 (for use with "user")
    \author VooDooMan
    \version 1
    \date 2005-08-02
    \retval "" for no entry
*/
string c_pass::get_local_password_md5(string user)
{
    if(init()) {
        string log="OOPS: Cannot load pass.txt";
        log_bot(log);
        log_debug(log);
        return (string)"";
    }
    map<string,map<string,string> >::iterator i=sections.find(user);
    if(i==sections.end()) {
        string log="OOPS: Cannot get password for user \"";
        log+=user;
        log+="\" form pass.txt";
        log_debug(log);
        return (string)"";
    }
    map<string,string>::iterator i2=(*i).second.find("disabled");
    if(i2!=(*i).second.end() && (*i2).second.compare("0")) // !="0"
        return (string)"";

    i2=(*i).second.find("local_password");
    if(i2!=(*i).second.end())
        return (*i2).second;

    return (string)"";
}

/*!
    \brief Returns true if user is not allowed to log in
    \author VooDooMan
    \version 1
    \date 2005-08-03
    \param user Name of user
*/
bool c_pass::get_disabled_user(string user)
{
    if(init()) {
        string log="OOPS: Cannot load pass.txt - returning true/disabled user";
        log_bot(log);
        log_debug(log);
        return true;
    }
    map<string,map<string,string> >::iterator i=sections.find(user);
    if(i==sections.end()) {
        string log="OOPS: Cannot get user \"";
        log+=user;
        log+="\"'s \"disabled\" form pass.txt - returning true/disabled user";
        log_debug(log);
        return true;
    }
    map<string,string>::iterator i2=(*i).second.find("disabled");

    if(i2!=(*i).second.end() && (*i2).second.compare("0")) // !="0"
        return true;

    if(i2==(*i).second.end()) {
        string log="OOPS: Cannot get user \"";
        log+=user;
        log+="\"'s \"disabled\" form pass.txt - returning true/disabled user";
        log_debug(log);
        return false;
    }

    return false;
}

/*!
    \brief Sets if user is not allowed to log in
    \author VooDooMan
    \version 1
    \date 2005-08-03
    \param user Name of user
    \param disabled Set to "true" to disable
*/
void c_pass::set_disabled_user(string user, bool disabled)
{
    if(init()) {
        string log="OOPS: Cannot load pass.txt";
        log_bot(log);
        log_debug(log);
        return;
    }
    map<string,map<string,string> >::iterator i=sections.find(user);
    if(i==sections.end()) {
        string log="OOPS: Cannot get user \"";
        log+=user;
        log+="\" form pass.txt";
        log_debug(log);
        return;
    }
    map<string,string>::iterator i2=(*i).second.find("disabled");
    if(i2!=(*i).second.end()) {
        (*i2).second=disabled?"1":"0";
    } else {
        pair<string,string> p;
        p.first="disabled";
        p.second=disabled?"1":"0";
        (*i).second.insert(p);
    }
}

/*!
    \brief Returns true if user is SSL only user
    \author VooDooMan
    \version 1
    \date 2005-08-03
    \param user Name of user
*/
bool c_pass::get_ssl_user(string user)
{
    if(init()) {
        string log="OOPS: Cannot load pass.txt - returning false/not to use SSL";
        log_bot(log);
        log_debug(log);
        return false;
    }
    map<string,map<string,string> >::iterator i=sections.find(user);
    if(i==sections.end()) {
        string log="OOPS: Cannot get user \"";
        log+=user;
        log+="\"'s \"use_ssl\" form pass.txt - returning false/not to use SSL";
        log_debug(log);
        return false;
    }
    map<string,string>::iterator i2=(*i).second.find("use_ssl");
    if(i2!=(*i).second.end() && (*i2).second.compare("0")) // !="0"
        return true;

    if(i2==(*i).second.end()) {
        string log="OOPS: Cannot get user \"";
        log+=user;
        log+="\"'s \"use_ssl\" form pass.txt - returning false/not to use SSL";
        log_debug(log);
        return false;
    }

    return false;
}

/*!
    \brief Sets if user is SSL only user
    \author VooDooMan
    \version 1
    \date 2005-08-03
    \param user Name of user
    \param use_ssl Set to "true" to use SSL
*/
void c_pass::set_ssl_user(string user, bool use_ssl)
{
    if(init()) {
        string log="OOPS: Cannot load pass.txt";
        log_bot(log);
        log_debug(log);
        return;
    }
    map<string,map<string,string> >::iterator i=sections.find(user);
    if(i==sections.end()) {
        string log="OOPS: Cannot get user \"";
        log+=user;
        log+="\" form pass.txt";
        log_debug(log);
        return;
    }
    map<string,string>::iterator i2=(*i).second.find("use_ssl");
    if(i2!=(*i).second.end()) {
        (*i2).second=use_ssl?"1":"0";
    } else {
        pair<string,string> p;
        p.first="use_ssl";
        p.second=use_ssl?"1":"0";
        (*i).second.insert(p);
    }
}

/*!
    \brief Returns zlib compression level
    \author VooDooMan
    \version 1
    \date 2005-11-07
    \param user Name of user (bot)
*/
int c_pass::get_zlib_level(string user)
{
    if(init()) {
        string log="OOPS: Cannot load pass.txt - returning \"1\" for zlib_level";
        log_bot(log);
        log_debug(log);
        return 1;
    }
    map<string,map<string,string> >::iterator i=sections.find(user);
    if(i==sections.end()) {
        string log="OOPS: Cannot get user \"";
        log+=user;
        log+="\" form pass.txt - returning \"1\" for zlib_level";
        log_debug(log);
        return 1;
    }
    map<string,string>::iterator i2=(*i).second.find("zlib_level");
    if(i2!=(*i).second.end())
        return stoi((*i2).second);

    string log="OOPS: Cannot get user \"";
    log+=user;
    log+="\"'s \"zlib_level\" form pass.txt - returning \"1\" for zlib_level";
    log_debug(log);
    return 1;
}

/*!
    \brief Sets zlib compression level for user (bot)
    \author VooDooMan
    \version 1
    \date 2005-11-07
    \param user Name of user (bot)
    \param level Level of compression
*/
void c_pass::set_zlib_level(string user, int level)
{
    if(init()) {
        string log="OOPS: Cannot load pass.txt";
        log_bot(log);
        log_debug(log);
        return;
    }
    map<string,map<string,string> >::iterator i=sections.find(user);
    if(i==sections.end()) {
        string log="OOPS: Cannot get user \"";
        log+=user;
        log+="\" form pass.txt";
        log_debug(log);
        return;
    }
    map<string,string>::iterator i2=(*i).second.find("zlib_level");
    if(i2!=(*i).second.end()) {
        (*i2).second=tos(level);
    } else {
        pair<string,string> p;
        p.first="zlib_level";
        p.second=tos(level);
        (*i).second.insert(p);
    }
}

/*!
    \brief Reads "conf.txt" and initializes values that are needed before "real" configuration will be read (before everything else)
    \version 3
    \author VooDooMan
    \date 2006-12-27, 2007-09-06, 2007-12-22
    \warning Not thread-safe (see comments)
*/
void conf_init_pre_start()
{
    static bool done=false;
    if(done)
        return;
    done=true;

    // init to default values
    conf_umask=077;

    FILE* f=fopen("conf.txt","r");
    if(f) {
        while(!feof(f)) {
            char ln[1024+1];
            *ln='\0';
            fgets(ln,1024,f);
            ln[1024]='\0';
            string key=ln, val=ln;
            if(key.find('=')==string::npos)
                continue;
            key.erase(key.find('='),key.length()-key.find('='));
            val.erase(0,val.find('=')+1);
            if(val.length() && val[val.length()-1]=='\n')
                val.erase(val.length()-1,1);
            if(!key.compare("umask")) {
                if(!val.compare("x") || !val.compare("X")) {
                    conf_umask=-1;
                } else {
                    int v;
                    int n=sscanf(val.c_str(),"%o",&v);
                    if(1==n)
                        conf_umask=v;
                }
            }
            if(!key.compare("log_cumul")) {
                int on=stoi(val);
                log_cumul=TO_BOOL(on);
            }
        }
        fclose(f);
    }

#if !defined(_WIN32) && !defined(_WIN64)
    if(-1==conf_umask) {
#       if 0    // TODO: autoconf check: HAVE_GETUMASK (?)
        conf_umask=getumask(); // GNU extension
#       else
        // this makes this function not thread-safe
        conf_umask=umask(0);
        umask(conf_umask);
#       endif
    }
#endif  // !defined(_WIN32)
}

