/***************************************************************************
                          c_conf.cpp  -  description
                             -------------------
    begin                : Thu May 10 2007
    copyright            : (C) 2004-2007 by Marian VooDooMan Meravy
    email                : ghostvoodooman -a.T= users (dot) sourceforge.net
 ***************************************************************************/

/* $Id: c_conf.cpp,v 1.22 2008/11/24 12:09:23 ghostvoodooman Exp $ */

/***************************************************************************

VooDoo cIRCle - an IRC (ro)bot
Copyright (C) 2004-2008 by Marian VooDooMan Meravy
  (ghostvoodooman -a.T= users (dot) sourceforge.net)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************/

/*!
    \file
    \brief Class 'c_conf'
    \version $Id: c_conf.cpp,v 1.22 2008/11/24 12:09:23 ghostvoodooman Exp $
*/

#if defined(HAVE_CONFIG_H)
#   include <config.h>
#endif

#include "../params.h"
#include "../utils/memleak.h"
#include "../utils/assert.h"
#include "../compat.h"
#include "../mod/c_module.h"
#include "../log.h"

#include "c_conf.h"
#include "../modules.h"
#include "../utils/precomp.h"

#if defined(_MSC_VER) || defined(__BORLANDC__)
#   pragma hdrstop
#endif

c_conf::c_conf(const c_conf& orig)
{
    keys=orig.keys;
    fn=orig.fn;
    init_ok=orig.init_ok;
    parsed=orig.parsed;

    log_echo_debug_=orig.log_echo_debug_;
    log_echo_irc_=orig.log_echo_irc_;
    log_echo_socket_=orig.log_echo_socket_;
    log_echo_bot_=orig.log_echo_bot_;
    log_echo_identd_=orig.log_echo_identd_;
    log_echo_botnet_=orig.log_echo_botnet_;
    log_echo_botnet_debug_=orig.log_echo_botnet_debug_;
    log_echo_ssl_=orig.log_echo_ssl_;
    log_echo_broadcast_=orig.log_echo_broadcast_;
    log_echo_clients_=orig.log_echo_clients_;

    log_debug_=orig.log_debug_;
    log_irc_=orig.log_irc_;
    log_socket_=orig.log_socket_;
    log_bot_=orig.log_bot_;
    log_identd_=orig.log_identd_;
    log_botnet_=orig.log_botnet_;
    log_botnet_debug_=orig.log_botnet_debug_;
    log_ssl_=orig.log_ssl_;
    log_broadcast_=orig.log_broadcast_;
    log_clients_=orig.log_clients_;

    // do NOT copy-construct this class! just copy the pointer!
    logger_=orig.logger_;

    eval_stack_size=orig.eval_stack_size;
    eval_script_timeout=orig.eval_script_timeout;

    is_core_conf=orig.is_core_conf;
}

void c_conf::clear()
{
    keys.clear();
    parsed=false;
}

bool c_conf::log_echo_debug()
{
    return parsed && log_echo_debug_;
}

bool c_conf::log_log_debug()
{
    return parsed && log_debug_;
}

bool c_conf::log_echo_irc()
{
    return parsed && log_echo_irc_;
}

bool c_conf::log_log_irc()
{
    return parsed && log_irc_;
}

bool c_conf::log_echo_socket()
{
    return parsed && log_echo_socket_;
}

bool c_conf::log_log_socket()
{
    return parsed && log_socket_;
}

bool c_conf::log_echo_bot()
{
    return parsed && log_echo_bot_;
}

bool c_conf::log_log_bot()
{
    return parsed && log_bot_;
}

bool c_conf::log_echo_identd()
{
    return parsed && log_echo_identd_;
}

bool c_conf::log_log_identd()
{
    return parsed && log_identd_;
}

bool c_conf::log_echo_botnet()
{
    return parsed && log_echo_botnet_;
}

bool c_conf::log_log_botnet()
{
    return parsed && log_botnet_;
}

bool c_conf::log_echo_botnet_debug()
{
    return parsed && log_echo_botnet_debug_;
}

bool c_conf::log_log_botnet_debug()
{
    return parsed && log_botnet_debug_;
}

bool c_conf::log_echo_ssl()
{
    return parsed && log_echo_ssl_;
}

bool c_conf::log_log_ssl()
{
    return parsed && log_ssl_;
}

bool c_conf::log_echo_broadcast()
{
    return parsed && log_echo_broadcast_;
}

bool c_conf::log_log_broadcast()
{
    return parsed && log_broadcast_;
}

bool c_conf::log_echo_clients()
{
    return parsed && log_echo_clients_;
}

bool c_conf::log_log_clients()
{
    return parsed && log_clients_;
}

unsigned c_conf::get_eval_stack_size()
{
    return eval_stack_size;
}

my_time_t c_conf::get_eval_script_timeout()
{
    return eval_script_timeout;
}

/*!
    \brief Loads conf.txt configuration file
    \author VooDooMan
    \version 2
    \date 2004, 2005-10-21
    \retval true For error
    \retval false Okay.
*/
bool c_conf::rehash()
{
    keys.clear();
    parsed=false;

    if(!init_ok || fn.empty())
        return true;

    FILE* f=fopen(fn.c_str(),"r");
    if(f==NULL)
        return true;

    char ln[10240+1];
    while(!feof(f)) {
        if(fgets(ln,1024*10-1,f)==NULL)
            break;
        ln[10240]=0;
        if(ln[0] && ln[strlen(ln)-1]=='\r')
            ln[strlen(ln)-1]=0;
        if(ln[0] && ln[strlen(ln)-1]=='\n')
            ln[strlen(ln)-1]=0;
        if(ln[0] && ln[strlen(ln)-1]=='\r')
            ln[strlen(ln)-1]=0;
        if(ln[0] && ln[strlen(ln)-1]=='\n')
            ln[strlen(ln)-1]=0;
        string id_, val_;
        int pos=0;
        for(unsigned int i1=0; i1<strlen(ln); ++i1) {
            if(pos==0 && ln[i1]=='=') {
                pos++;
                continue;
            }
            if(pos==0)
                id_+=ln[i1];
            if(pos==1)
                val_+=ln[i1];
        }
        keys.insert(make_pair(id_,val_));
    }
    fclose(f);
    parsed=true;

    rehash_magicks();

    return false;
}

/*!
    \brief Rehashes, and re-reads conf.txt configuration file into keys variable and sets init to true
    \author VooDooMan
    \version 1
    \date 2004
*/
bool c_conf::init(const string& file_name, c_logger* logger, bool is_core)
{
    fn=file_name;
    logger_=logger;
    init_ok=true;
    is_core_conf=is_core;
    return !rehash();
}

string c_conf::convert_old_time_offset_if_needed(const string& orig)
{
    bool convert=false;
    if(orig.length()!=sizeof("+HHMM")-1)
        convert=true;
    else if('+'!=orig[0] && '-'!=orig[0])
        convert=true;
    else if(orig[1]<'0' || orig[1]>'9')
        convert=true;
    else if(orig[2]<'0' || orig[2]>'9')
        convert=true;
    else if(orig[3]<'0' || orig[3]>'9')
        convert=true;
    else if(orig[4]<'0' || orig[4]>'9')
        convert=true;

    if(!convert)
        return orig;

    int off=0;
    sscanf(orig.c_str(),"%d",&off);
    string res="+";
    if(off<0) {
        res="-";
        off=-off;
    }
    char tmp[16];
    snprintf(tmp,sizeof(tmp)-1,"%02d00",off);
    tmp[sizeof(tmp)-1]='\0';
    res+=tmp;

    return res;
}

/*!
    \brief Gets variable from conf.txt configuration file
    \author VooDooMan
    \version 3
    \date 2004, 2005-10-21, 2008-06-08
    \param key Key string to find in configuration file
    \return Value for key
*/
string c_conf::getvar(const string& key)
{
    if(!init_ok)
        return (string)"";
    if(!parsed)
        rehash();

    map<string,string>::iterator i=keys.find(key);
    if(keys.end()==i) {
        if(!is_core_conf) {
            c_conf* cc=get_core_conf();
            if(cc) {
                return cc->getvar(key);
            }
        }
        string log="OOPS: Request for key \"";
        log+=key;
        log+="\" from \"";
        log+=fn;
        log+="\" - not found, returning an empty string (in case of boolean or integer key type, this will be interpreted as it were set to '0')";
        if(logger_)
            logger_->log_print1(E_LOG_DEBUG,LOG_SCOPE_DEFAULT,"",log);

        return (string)"";
    }

    if(!key.compare("time_offset"))
        return convert_old_time_offset_if_needed((*i).second);

    return (*i).second;
}

/*!
    \brief Gets variable from conf.txt configuration file (as an integer)
    \author VooDooMan
    \version 1
    \date 2008-03-12
    \param key Key string to find in configuration file
    \return Value for key asn an integer
*/
MY_LONG_LONG c_conf::getvar_int(const string& key)
{
    if(!init_ok)
        return 0;
    if(!parsed)
        rehash();

    string val;
    MY_LONG_LONG res;

    map<string,string>::iterator i=keys.find(key);
    if(keys.end()==i) {
        if(!is_core_conf) {
            c_conf* cc=get_core_conf();
            if(cc)
                return cc->getvar_int(key);
        }
        string log="OOPS: Request for key \"";
        log+=key;
        log+="\" from \"";
        log+=fn;
        log+="\" - not valid integer value, returning a zero integer.";
        if(logger_)
            logger_->log_print1(E_LOG_DEBUG,LOG_SCOPE_DEFAULT,"",log);

        return 0;
    }
    val=(*i).second;
    if(sscanf(val.c_str(),"%" MY_LONG_LONG_STR "d",&res)!=1) {
        string log="OOPS: Request for key \"";
        log+=key;
        log+="\" from \"";
        log+=fn;
        log+="\" - not valid integer value, returning a zero integer.";
        if(logger_)
            logger_->log_print1(E_LOG_DEBUG,LOG_SCOPE_DEFAULT,"",log);

        return 0;
    }
    return res;
}

void c_conf::rehash_magicks()
{
    log_echo_debug_=getvar_int("log_echo_debug")!=0;
    log_echo_irc_=getvar_int("log_echo_irc")!=0;
    log_echo_socket_=getvar_int("log_echo_socket")!=0;
    log_echo_bot_=getvar_int("log_echo_bot")!=0;
    log_echo_identd_=getvar_int("log_echo_identd")!=0;
    log_echo_botnet_=getvar_int("log_echo_botnet")!=0;
    log_echo_botnet_debug_=getvar_int("log_echo_botnet_debug")!=0;
    log_echo_ssl_=getvar_int("log_echo_ssl")!=0;
    log_echo_broadcast_=getvar_int("log_echo_broadcast")!=0;
    log_echo_clients_=getvar_int("log_echo_clients")!=0;

    log_debug_=getvar_int("log_debug")!=0;
    log_irc_=getvar_int("log_irc")!=0;
    log_socket_=getvar_int("log_socket")!=0;
    log_bot_=getvar_int("log_bot")!=0;
    log_identd_=getvar_int("log_identd")!=0;
    log_botnet_=getvar_int("log_botnet")!=0;
    log_botnet_debug_=getvar_int("log_botnet_debug")!=0;
    log_ssl_=getvar_int("log_ssl")!=0;
    log_broadcast_=getvar_int("log_broadcast")!=0;
    log_clients_=getvar_int("log_clients")!=0;

    if(is_core_conf) {
        eval_stack_size=static_cast<unsigned>(getvar_int("eval_stack_size"));
        if(0==eval_stack_size)
            eval_stack_size=200;

        eval_script_timeout=getvar_int("eval_script_timeout");
        if(0==eval_script_timeout)
            eval_script_timeout=10;
    } else {
        eval_stack_size=static_cast<unsigned>(getvar_int("eval_stack_size"));
        eval_script_timeout=getvar_int("eval_script_timeout");
        c_conf* cc=get_core_conf();
        if(cc) {
            if(0==eval_stack_size)
                eval_stack_size=static_cast<unsigned>(cc->getvar_int("eval_stack_size"));
            if(0==eval_script_timeout)
                eval_script_timeout=cc->getvar_int("eval_script_timeout");
        }
        if(0==eval_stack_size)
            eval_stack_size=200;
        if(0==eval_script_timeout)
            eval_script_timeout=10;
    }
}

string core_conf_getvar(const string& key)
{
    extern c_module* core_mod;

    if(!core_mod)
        core_mod=get_module("core",false);

    c_conf* core_conf=NULL;
    if(core_mod)
        core_conf=core_mod->get_conf();

    string res;
    if(core_conf)
        res=core_conf->getvar(key);

    return res;
}

MY_LONG_LONG core_conf_getvar_int(const string& key)
{
    extern c_module* core_mod;

    if(!core_mod)
        core_mod=get_module("core",false);

    c_conf* core_conf=NULL;
    if(core_mod)
        core_conf=core_mod->get_conf();

    MY_LONG_LONG res=0;
    if(core_conf)
        res=core_conf->getvar_int(key);

    return res;
}

