/***************************************************************************
                          c_conf.h  -  description
                             -------------------
    begin                : Thu May 10 2007
    copyright            : (C) 2004-2007 by Marian VooDooMan Meravy
    email                : ghostvoodooman -a.T= users (dot) sourceforge.net
 ***************************************************************************/

/* $Id: c_conf.h,v 1.16 2008/06/08 08:43:13 ghostvoodooman Exp $ */

/***************************************************************************

VooDoo cIRCle - an IRC (ro)bot
Copyright (C) 2004-2008 by Marian VooDooMan Meravy
  (ghostvoodooman -a.T= users (dot) sourceforge.net)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************/

/*!
    \file
    \brief Class 'c_conf'
    \version $Id: c_conf.h,v 1.16 2008/06/08 08:43:13 ghostvoodooman Exp $
*/

#ifndef C_CONF_H__
#define C_CONF_H__

#if defined(HAVE_CONFIG_H)
#   include <config.h>
#endif

#include <string>
#include <vector>
#include <stack>
#include <list>
#include <map>
#include <algorithm>

#include "../utils/memleak.h"

#include "../params.h"

#include "../log/c_logger.h"
#include "../compat.h"

class c_logger;

/*!
    \brief Stores parsed "conf.txt" file
    \author VooDooMan
    \version 1
    \date 2005
*/
class c_conf
{
    map<string,string> keys;                    //!< First: key, second: value
    string fn;                                  //!< Name of file
    bool init_ok;                               //!< If initialization was okay
    bool parsed;                                //!< If we were parsed it

    c_logger* logger_;                          //!< Logger for this module

    bool is_core_conf;                          //!< true if this config is for core module

    void rehash_magicks();                      //!< Rehasesh cached keys (frequently used keys; for optimization)

    bool log_echo_debug_,                       //!< "log_echo_debug" key value
        log_echo_irc_,                          //!< "log_echo_irc" key value
        log_echo_socket_,                       //!< "log_echo_socket" key value
        log_echo_bot_,                          //!< "log_echo_bot" key value
        log_echo_identd_,                       //!< "log_echo_ident" key value
        log_echo_botnet_,                       //!< "log_echo_botnet" key value
        log_echo_botnet_debug_,                 //!< "log_echo_botnet_debug" key value
        log_echo_ssl_,                          //!< "log_echo_ssl" key value
        log_echo_broadcast_,                    //!< "log_echo_broadcast" key value
        log_echo_clients_,                      //!< "log_echo_clients" key value

        log_debug_,                             //!< "log_debug" key value
        log_irc_,                               //!< "log_irc" key value
        log_socket_,                            //!< "log_socket" key value
        log_bot_,                               //!< "log_bot" key value
        log_identd_,                            //!< "log_ident" key value
        log_botnet_,                            //!< "log_botnet" key value
        log_botnet_debug_,                      //!< "log_botnet_debug" key value
        log_ssl_,                               //!< "log_ssl" key value
        log_broadcast_,                         //!< "log_broadcast" key value
        log_clients_;                           //!< "log_clients" key value

    unsigned eval_stack_size;                   //!< "eval_stack_size" key value
    my_time_t eval_script_timeout;              //!< "eval_script_timeout" key value

public:
    c_conf()
        : init_ok(false), parsed(false), logger_(NULL)
    { ; }

    c_conf(const c_conf& orig);
    bool init(const string& file_name, c_logger* logger, bool is_core);
    void clear();
    bool rehash();
    string getvar(const string& key);
    MY_LONG_LONG getvar_int(const string& key);
    bool log_echo_debug();
    bool log_log_debug();
    bool log_echo_irc();
    bool log_log_irc();
    bool log_echo_socket();
    bool log_log_socket();
    bool log_echo_bot();
    bool log_log_bot();
    bool log_echo_identd();
    bool log_log_identd();
    bool log_echo_botnet();
    bool log_log_botnet();
    bool log_echo_botnet_debug();
    bool log_log_botnet_debug();
    bool log_echo_ssl();
    bool log_log_ssl();
    bool log_echo_broadcast();
    bool log_log_broadcast();
    bool log_echo_clients();
    bool log_log_clients();
    unsigned get_eval_stack_size();
    my_time_t get_eval_script_timeout();

    static string convert_old_time_offset_if_needed(const string& orig);
};

string core_conf_getvar(const string& key);
MY_LONG_LONG core_conf_getvar_int(const string& key);

#endif

