/***************************************************************************
                          dcc.cpp  -  description
                             -------------------
    begin                : Thu Dec 9 2004
    copyright            : (C) 2004-2007 by Marian VooDooMan Meravy
    email                : ghostvoodooman -aT= users (dot) sourceforge.net
 ***************************************************************************/

/* $Id: dcc.cpp,v 1.88 2008/06/29 17:37:49 ghostvoodooman Exp $ */

/***************************************************************************

VooDoo cIRCle - an IRC (ro)bot
Copyright (C) 2004-2008 by Marian VooDooMan Meravy
  (ghostvoodooman -aT= users (dot) sourceforge.net)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************/

/*!
    \file
    \brief Handles DCC / telnet communication
    \version $Id: dcc.cpp,v 1.88 2008/06/29 17:37:49 ghostvoodooman Exp $
*/

#if defined(HAVE_CONFIG_H)
#   include <config.h>
#endif

#include <string>
#include <vector>
#include <stack>
#include <list>
#include <map>
#include <algorithm>
#include <iostream>

#include <stdlib.h>

#include <time.h>
#include <stdio.h>
#if defined(_WIN32) || defined(_WIN64)
#   include <io.h>
#endif
#include <sys/stat.h>

#include "utils/memleak.h"
#include "utils/assert.h"
#include "utils.h"

#include "params.h"

#include "dcc/dcc_internal.h"
#include "misc/shutdown.h"

#include "utils/precomp.h"

#if defined(_MSC_VER) || defined(__BORLANDC__)
#   pragma hdrstop
#endif

using namespace std;

//---------------------------------------------------------------------------
#ifdef __BORLANDC__
#pragma package(smart_init)
#endif

vector<c_dcc_awaiting_user> dcc_awaiting_users;     //!< Stores users who request dcc connection

c_socket* dcc_killing_user_socket=NULL;             //!< Socket of DCC/telnet user who is killing the bot (.die/.restart), or NULL; (will be sent log to it)
c_ssl* dcc_killing_ssl_hnd=NULL;                    //!< SSL handle of killing user's socket
bool dcc_killing_ssl=false;                         //!< true if killing user is using SSL
string dcc_killing_user_eol;                        //!< EOL string for user (DCC/telnet) who is killing the bot (.die/.restart)
string dcc_killing_user_lang;                       //!< Language code string for user (DCC/telnet) who is killing the bot (.die/.restart)
my_time_t last_new_instance_join;                      //!< After .upgrade command, there is timestamp of last new instance's join some channel

log_client_id_t dcc_last_client_id=0;               //!< Last client ID for logs

vector<c_dcc_client> dcc_clients;               //!< List of DCC clients
vector<c_dcc_server> dcc_servers;               //!< List of DCC servers
vector<c_telnet_server> telnet_servers;         //!< List of telnet servers

vector<c_dcc_send_file> dcc_awaiting_send;      //!< List of files that are awaiting to send (DCC SEND) to remote host

vector<c_dcc_client_to_resume> dcc_to_resume;   //!< Temporary array for user who wants to resume

vector<c_user_to_lag> users_to_lag;             //!< Array of DCC requests statistics (for flood protection)

vector<c_dcc_request_whois> dcc_request_whois;  //!< Array, that stores DCC requests that are pending to establish, and waiting for whois result

void dcc_check_for_filesystem(string user_name);

vector<c_user_to_lock_out> dcc_users_to_lock_out;   //!< Users to lock out due to bad log in

void dcc_broadcast(const string& msg);

/*!
    \brief Sends a message to DCC CHAT / telnet connection and updates statistics
    \author VooDooMan
    \version 2
    \date 2004, 2006-07-10
    \param socket Socket handle
    \param msg Message to send
    \param ec Returns socket I/O error code
    \param ssl NULL for non-ssl, SSL handle for SSL
*/
void dcc_send_msg(c_socket& socket, string msg, int& ec, c_ssl* ssl)
{
    if(!ssl) {
        ec=0;
        size_t sent=sock_send(socket,msg.c_str(),msg.length(),ec);
        if(!ec)
            stats_dcc_chat_bytes_sent(sent);
    } else {
        ssl->write(msg.c_str(),msg.length());
        stats_dcc_chat_bytes_sent(msg.length());
        sock_set_last_io(socket,my_time(NULL),'s');
    }
}

/*!
    \brief Closes listening sockets
    \author VooDooMan
    \version 2
    \date 2005, 2007-08-20
*/
void dcc_close_servers()
{
    {
        vector<c_dcc_server>::iterator i1;
        while((i1=dcc_servers.begin())!=dcc_servers.end()) {
            if((*i1).active)
                sock_close((*i1).socket,__FILE__,__LINE__);
            dcc_servers.erase(i1);
        }
    }
    {
        vector<c_telnet_server>::iterator i1;
        while((i1=telnet_servers.begin())!=telnet_servers.end()) {
            if((*i1).active)
                sock_close((*i1).socket,__FILE__,__LINE__);
            telnet_servers.erase(i1);
        }
    }
}

/*!
    \brief Tries to upgrade bot
    \author VooDooMan
    \version 2
    \date 2005, 2007-08-20
*/
void dcc_upgrade(vector<c_dcc_client>::iterator i, string& eol, string new_file_name)
{
    string path="upgrade" FILE_SLASH;
    path+=new_file_name;
    {
        FILE* f=fopen(path.c_str(),"rb");
        if(f==NULL) {
            string disp=lang_get_string(1,(*i).lang,630)+eol;
            int ec=0;
            dcc_send_msg((*i).socket,disp,ec,(*i).ssl);
            return;
        } else {
            fclose(f);
        }
    }
    dcc_killing_user_eol=eol;
    dcc_killing_user_socket=&((*i).socket);
    dcc_killing_ssl=(*i).is_ssl_client;
    if(dcc_killing_ssl)
        dcc_killing_ssl_hnd=(*i).ssl;

    char* tmp=MY_NEW char[(*i).as_in_logic.length()+new_file_name.length()+128];
    sprintf(tmp,".upgrade command for \"%s\" by %s",path.c_str(),(*i).as_in_logic.c_str());
    log_bot(tmp);
    sprintf(tmp,"UPGRADE by %s",(*i).as_in_logic.c_str());
    dcc_broadcast(tmp);
    MY_DELETE_A(tmp);
    tmp=NULL;

    logic_on_internal_event("@upgrade@",(*i).as_in_logic,path,"","",PRIVATE_EVENT_SEVERITY_CODE_INFORMATIONAL,"","");

    dcc_close_servers();

    c_module* core_mod=get_module("core",true);
    logic_partyline_backup(core_mod,(*i).as_in_logic);

    string msg="Entering UPGRADE state. All server sockets have been closed.";
    string msg2=" Spawning new instance.";
    string log=msg+msg2;
    log_bot(log);
    dcc_broadcast(msg);

    string arg="--upgrade=";
    string tmp2=binary_name;
    while(tmp2.find(FILE_SLASH,0)!=string::npos)
        tmp2.erase(0,tmp2.find(FILE_SLASH,0)+1);
    arg+=tmp2;
    arg+=" --wait-for-released-lock";

    string tmp3=(string)"\""+path+"\"";
    cout << "Executing in " << "." FILE_SLASH << ": " << tmp3 << arg << endl;
    c_exec_handle* h=exec_async_exec2(tmp3.c_str(),arg.c_str(),"." FILE_SLASH);

    if(!h || h->error) {
        string disp=lang_get_string(1,(*i).lang,631)+eol;
        int ec=0;
        dcc_send_msg((*i).socket,disp,ec,(*i).ssl);
        bot_cancel_upgrade();
        char msg1[]="UPGRADE panic! Error spawning new instance. You should use .restart command now.";
        log_bot(msg1);

        logic_on_internal_event("@upgrade@",(*i).as_in_logic,path,"","",PRIVATE_EVENT_SEVERITY_CODE_ERROR,msg1,"");
    } else {
        string disp=lang_get_string(1,(*i).lang,632)+eol;
        int ec=0;
        dcc_send_msg((*i).socket,disp,ec,(*i).ssl);
        char msg1[]="UPGRADE: New instance spawned, waiting for it to come in.";
        log_bot(msg1);

        logic_on_internal_event("@upgrade@",(*i).as_in_logic,path,"","",PRIVATE_EVENT_SEVERITY_CODE_INFORMATIONAL,msg1,"");

        last_new_instance_join=my_time(NULL);
        dcc_killing_user_socket=&(*i).socket;
        dcc_killing_ssl=(*i).is_ssl_client;
        if(dcc_killing_ssl)
            dcc_killing_ssl_hnd=(*i).ssl;
        bot_shutdown_action(E_SD_UPGRADE,(*i).as_in_logic);
        dcc_killing_user_lang=(*i).lang;

        log_free_locks();
    }

    exec_detach_process(h);
}

/*!
    \brief Called in UPGRADE state, to inform user who initiated .upgrade command that another instance is on channel
    \author VooDooMan
    \version 1
    \date 2005
    \param nick Nick of new instance
    \param ident Ident of new instance
    \param host Host of new instance
    \param fullname Full name of new instance
    \param channel Channnel that new instance joined
*/
void dcc_upgrade_bot_join(string nick, string ident, string host, string fullname, string channel)
{
    last_new_instance_join=my_time(NULL);
    string eol=dcc_killing_user_eol;
    c_socket* s=dcc_killing_user_socket;
    string lang=dcc_killing_user_lang;
    if(s && s->cmp()) {
        string disp=lang_get_string(1,lang,633)+eol;
        disp+="    ";
        disp+=channel+eol;
        disp+="    ";
        disp+=nick;
        disp+="!";
        disp+=ident;
        disp+="@";
        disp+=host;
        disp+=" * ";
        disp+=fullname;
        disp+=eol;
        int ec=0;
        dcc_send_msg(*s,disp,ec,dcc_killing_ssl_hnd);
    }
}

/*!
    \brief Called in UPGRADE state, to inform user who initiated .upgrade command that another instance on some channel gets some mode
    \author VooDooMan
    \version 1
    \date 2005
    \param nick Nick of new instance
    \param ident Ident of new instance
    \param host Host of new instance
    \param fullname Full name of new instance
    \param channel Channnel that new instance joined
    \param mode New mode
    \param src Source of MODE command
*/
void dcc_upgrade_bot_mode(string nick, string ident, string host, string fullname, string channel, string mode, string src)
{
    string eol=dcc_killing_user_eol;
    c_socket* s=dcc_killing_user_socket;
    string lang=dcc_killing_user_lang;
    if(s && s->cmp()) {
        string disp=lang_get_string(1,lang,634)+eol;
        disp+="    ";
        disp+=channel;
        disp+=" <";
        disp+=src;
        disp+="> ";
        disp+=mode;
        disp+=eol;
        disp+="    ";
        disp+=nick;
        disp+="!";
        disp+=ident;
        disp+="@";
        disp+=host;
        disp+=" * ";
        disp+=fullname;
        disp+=eol;
        int ec=0;
        dcc_send_msg(*s,disp,ec,dcc_killing_ssl_hnd);
    }
}

/*!
    \brief Called in UPGRADE state, to inform user who initiated .upgrade command that another instance changed its nick
    \author VooDooMan
    \version 1
    \date 2005
    \param nick Nick of new instance
    \param ident Ident of new instance
    \param host Host of new instance
    \param fullname Full name of new instance
    \param new_nick New nick
*/
void dcc_upgrade_bot_nick(string nick, string ident, string host, string fullname, string new_nick)
{
    string eol=dcc_killing_user_eol;
    c_socket* s=dcc_killing_user_socket;
    string lang=dcc_killing_user_lang;
    if(s && s->cmp()) {
        string disp=lang_get_string(1,lang,635);
        disp+=" ";
        disp+=new_nick;
        disp+=eol;
        disp+="    ";
        disp+=nick;
        disp+="!";
        disp+=ident;
        disp+="@";
        disp+=host;
        disp+=" * ";
        disp+=fullname;
        disp+=eol;
        int ec=0;
        dcc_send_msg(*s,disp,ec,dcc_killing_ssl_hnd);
    }
}

/*!
    \brief Retrieves number of concurrent DCC trtansfers
    \author VooDooMan
    \version 1
    \date 2004
    \return Returns number of concurrent DCC trtansfers
*/
int dcc_get_num_of_transfers()
{
    int num=0;
    vector<c_dcc_client>::iterator i1;
    for(i1=dcc_clients.begin(); i1!=dcc_clients.end(); ++i1)
        if((*i1).send || (*i1).send_from_filesystem)
            num++;

    return num;
}

/*!
    \brief Sends file to remote host
    \author VooDooMan
    \version 1
    \date 2004
    \param from_dcc Is this function invoked from DCC CHAT?
    \param public_name Public name of the file
    \param nick Nickname of user
    \param dcc_group DCC group of the connection that should be used
*/
int dcc_send_file(bool from_dcc, string public_name, string nick, int dcc_group)
{
    string ident=irc_get_ident(nick);
    string host=irc_get_host(nick);
    string full_name=irc_get_fullname(nick);

    string user_name=logic_find_user(nick,ident,host,full_name,irc_is_ircop(nick));

    if(!logic_access_to_filesystem(user_name))
        return DCC_ACCESS_DENIED;

    vector<c_file> files;
    filesys_dcc_filelist(user_name,files,false);

    c_file f;

    bool got=false;

    vector<c_file>::iterator i1;
    for(i1=files.begin(); i1!=files.end(); ++i1) {
        if(!(*i1).public_name.compare(public_name)) {
            got=true;
            f=*i1;
        }
    }

    if(!got)
        return DCC_NO_SUCH_FILE;

    got=false;
    vector<c_access>::iterator i2;
    for(i2=f.access.begin(); i2!=f.access.end(); ++i2)
        if((*i2).all_users || !(*i2).user_name.compare(user_name) || (*i2).also_unknown)
            if((*i2).read) {
                if(!from_dcc && (*i2).secure) {
                    string msg=":";
                    msg+=lang_get_string(1,"en",333);
                    irc_putserv(c_tracked_msg_to_server("PRIVMSG").add_nick(nick).add_piece(msg),false,LOW_PRIORITY);
                    return DCC_ACCESS_DENIED;
                }
                got=true;
                break;
            }
    if(!got)
        return DCC_ACCESS_DENIED;

    bool using_ipv6=false;
    in_addr ip;
    memset(&ip,0,sizeof(ip));
    in_addr6_ ip6;
    memset(&ip6,0,sizeof(ip6));
    in_addr ip_serv;
    memset(&ip_serv,0,sizeof(ip_serv));
    in_addr6_ ip6_serv;
    string bind_ip6;
    memset(&ip6_serv,0,sizeof(ip6_serv));

    bool want_ipv6=false;
again_ipv6:
    got=false;
    vector<c_dcc_server>::iterator i3;
    for(i3=dcc_servers.begin(); i3!=dcc_servers.end(); ++i3) {
        if(!(*i3).active)
            continue;
        if((*i3).group==dcc_group) {
            if(!want_ipv6 && !(*i3).using_ipv6) {
                using_ipv6=false;
                int type=sock_get_address_type(host);
                if(type==4)
                    my_inet_pton(AF_INET,host.c_str(),&ip);
                if(type==0) {
                    e_resolv_result res=threads_resolve4(host.c_str(),ip,RT_RESOLVE_SCHEDULE);
                    if(RR_SCHEDULED==res)
                        return DCC_AGAIN;
                    if(RR_FAILURE==res)
                        return DCC_IO_ERROR;
                }
                if(type==6) {
                    want_ipv6=true;
                    goto again_ipv6;
                }
                {
#if defined(_WIN32) || defined(_WIN64)
                    if(!ip.S_un.S_addr)
#else
                    if(!ip.s_addr)
#endif
                        return DCC_IO_ERROR;
                }
#if defined(_WIN32) || defined(_WIN64)
                ip_serv.S_un.S_addr=(*i3).dcc_addr;
#else
                ip_serv.s_addr=(*i3).dcc_addr;
#endif
            } else {
                using_ipv6=true;
                int type=sock_get_address_type(host);
                if(type==0) {
                    e_resolv_result res=threads_resolve6(host.c_str(),ip6,RT_RESOLVE_SCHEDULE);
                    if(RR_SCHEDULED==res)
                        return DCC_AGAIN;
                    if(RR_FAILURE==res)
                        return DCC_IO_ERROR;
                }
                if(type==6)
                    my_inet_pton(AF_INET6,host.c_str(),&ip6);
                {
                    if(!ip6.bytes[0] && !ip6.bytes[1] && !ip6.bytes[2] && !ip6.bytes[3]
                        && !ip6.bytes[4] && !ip6.bytes[5] && !ip6.bytes[6] && !ip6.bytes[7]
                        && !ip6.bytes[8] && !ip6.bytes[9] && !ip6.bytes[10] && !ip6.bytes[11]
                        && !ip6.bytes[12] && !ip6.bytes[13] && !ip6.bytes[14] && !ip6.bytes[15])
                        return DCC_IO_ERROR;
                }
                memcpy(&ip6_serv,&(*i3).dcc_addr6,sizeof(ip6_serv));
                bind_ip6=(*i3).bind_ip;
                if(bind_ip6.find("%",0)!=string::npos)
                    bind_ip6.erase(bind_ip6.find("%",0),bind_ip6.length()-bind_ip6.find("%",0));
            }
            got=true;
            break;
        }
    }

    if(!got)
        return DCC_IO_ERROR;

    c_dcc_send_file send;
    send.user_name_as_in_logic=user_name;
    send.resume_pos=0;
    send.public_name=public_name;
    send.start_time=my_time(NULL);
    send.using_ipv6=using_ipv6;
    if(!using_ipv6) {
        send.remote_ip=ip;
        memset(&send.remote_ip6,0,sizeof(send.remote_ip6));
    } else {
#if defined(_WIN32) || defined(_WIN64)
        send.remote_ip.S_un.S_addr=static_cast<unsigned int>(-1);
#else
        send.remote_ip.s_addr=static_cast<unsigned int>(-1);
#endif
        memcpy(&send.remote_ip6,&ip6,sizeof(send.remote_ip6));
    }

    if(!using_ipv6) {
        // ip_serv; // it is already in network byte order
    }

    send.file_name="." FILE_SLASH "filesystem" FILE_SLASH;
    send.file_name+=f.internal_name;

    FILE* ff=fopen(send.file_name.c_str(),"rb");
    if(!ff)
        return DCC_IO_ERROR;

    int my_errno;
    send.size=my_filesize(ff,my_errno);
    fclose(ff);
    if(-1==send.size || 0!=my_errno)
        send.size=0;

    dcc_awaiting_send.push_back(send);

    string inv=":\x01"; // PRIVMSG
    inv+="DCC SEND ";
    inv+=public_name;
    inv+=" ";

    string ip_string;

    char tmp[64];
    if(!using_ipv6) {
        char tmp[128];
        *tmp='\0';
        my_inet_ntop(AF_INET,&ip_serv,tmp,sizeof(tmp)-1);
        ip_string=tmp;
        uint32 ip;
        memcpy(&ip,&ip_serv,sizeof(ip));
#if !defined(WORDS_BIGENDIAN)
        ip=swap_bytes(ip);
#endif
        sprintf(tmp,"%u",static_cast<unsigned>(ip));
        inv+=tmp;
    } else {
        inv+=bind_ip6;
        ip_string=bind_ip6;
    }
    inv+=" ";

    inv+=tos((*i3).port);
    inv+=" ";
    sprintf(tmp,"%s",tos(send.size).c_str());
    inv+=tmp;
    inv+="\x01";

    string inv2=":DCC Send "; // NOTICE
    inv2+=public_name;
    inv2+=" (";
    inv2+=ip_string;
    inv2+=")";
    irc_putserv(c_tracked_msg_to_server("NOTICE").add_nick(nick).add_piece(inv2),false,HIGH_PRIORITY);

    irc_putserv(c_tracked_msg_to_server("PRIVMSG").add_nick(nick).add_piece(inv),false,HIGH_PRIORITY);

    return 0;
}

/*!
    \brief Called to set the READ-"ed" flag and notify OWNER(s) if needed
    \author VooDooMan
    \version 1
    \date 2004
    \param public_name Public name of the file
    \param nick Nick name of user
*/
void dcc_file_has_been_read(string public_name, string nick)
{
    if(dcc_get_num_of_transfers()>=stoi(core_conf_getvar("dcc_max_transfers"))) {
        return;
    }

    string ident=irc_get_ident(nick);
    string host=irc_get_host(nick);
    string full_name=irc_get_fullname(nick);

    string user_name=logic_find_user(nick,ident,host,full_name,irc_is_ircop(nick));

    if(!logic_access_to_filesystem(user_name))
        return;

    vector<c_file> files;
    filesys_dcc_filelist(user_name,files,false);

    c_file f;

    bool got=false;

    vector<c_file>::iterator i1;
    for(i1=files.begin(); i1!=files.end(); ++i1) {
        if(!(*i1).public_name.compare(public_name)) {
            got=true;
            f=*i1;
        }
    }

    if(!got)
        return;

    got=false;
    vector<c_access>::iterator i2;
    for(i2=f.access.begin(); i2!=f.access.end(); ++i2)
        if((*i2).all_users || !(*i2).user_name.compare(user_name))
            if((*i2).read) {
                if((*i2).notify_owner) {
                    vector<c_event>::iterator i3;
                    bool has_read=false;
                    for(i3=f.events.begin(); i3!=f.events.end(); ++i3) {
                        if(!(*i3).user_name.compare(user_name)) {
                            if((*i3).has_read)
                                has_read=true;
                            break;
                        }
                    }
                    if(!has_read) {
                        c_event e;
                        e.clear();
                        e.has_read=true;
                        e.owner_notified=false;
                        e.user_name=user_name;
                        e.user_notified=false;
                        e.first=my_time(NULL);

                        f.events.push_back(e);
                        filesys_dcc_set_file_attrs(f);
                    }
                    break;
                }
                got=true;
                break;
            }
    if(!got)
        return;
}

/*!
    \brief Gets MD5-encoded password for user
    \author VooDooMan
    \version 3
    \date 2004, 2005-08-02
    \param user Name of user
    \return Returns MD5-encoded password, or empty string is password has not been set yet, or string "disabled" if access is disabled
    \warning Name of user is case-sensitive
*/
string dcc_get_password(string user)
{
    return conf_pass.get_remote_password_md5(user);
}

/*!
    \brief Sets password for user
    \author VooDooMan
    \version 3
    \date 2005, 2005-08-02
    \param user Name of user
    \param md5pass MD5-encoded password
    \warning Name of user is case-sensitive
*/
void dcc_set_password(string user, string md5pass)
{
    e_shutdown_reason sr=bot_get_shutdown_state();
    MY_ASSERT(E_SD_UPGRADE!=sr);
    if(E_SD_UPGRADE==sr)
        return;

    conf_pass.set_remote_password_md5(user,md5pass);
}

/*!
    \brief Called when 0x01-quoted PRIVMSG has been received to initialise DCC CHAT or DCC SEND connection
    \author VooDooMan
    \version 2
    \date 2004, 2005-08-01
    \param str Message content to parse
    \param nick Nick name of sender
    \param host_mask Host mask of sender
    \param comment Full request string
*/
void irc_quoted_callback(const string& str, const string& nick, const string& host_mask, const string& comment)
{
    if(str.find("DCC ")==string::npos)
        return;

    string user_name;
    {
        if(irc_get_host(nick).empty()) {
            bool got=false;
            vector<c_dcc_request_whois>::iterator i1;
        l1:
            for(i1=dcc_request_whois.begin(); i1!=dcc_request_whois.end(); ++i1) {
                if((*i1).start+120<my_time(NULL)) {
                    dcc_request_whois.erase(i1);
                    goto l1;
                }
            }
            for(i1=dcc_request_whois.begin(); i1!=dcc_request_whois.end(); ++i1) {
                if(!(*i1).nick.compare(nick) && (*i1).got_whois) {
                    user_name=logic_find_user(nick,(*i1).ident,(*i1).host,(*i1).full_name,(*i1).irc_op);
                    got=true;
                    break;
                }
            }

            if(!got) {
                c_dcc_request_whois r;
                r.clear();
                my_time(&r.start);
                r.str=str;
                r.nick=nick;
                r.host_mask=host_mask;
                r.comment=comment;
                dcc_request_whois.push_back(r);

                irc_putserv(c_tracked_msg_to_server("WHOIS").add_nick(nick),true,LOW_PRIORITY);
                return;
            }
        } else
            user_name=logic_find_user(nick,irc_get_ident(nick), irc_get_host(nick), irc_get_fullname(nick), irc_is_ircop(nick));

        logic_on_internal_event("@dcc_request@",user_name,nick,host_mask,"",PRIVATE_EVENT_SEVERITY_CODE_INFORMATIONAL,"Received DCC invitation.",str);

        vector<c_user_to_lag>::iterator i1;
again:
        for(i1=users_to_lag.begin(); i1!=users_to_lag.end(); ++i1) {
            if((*i1).last_try+sto_time(core_conf_getvar("ctcp_lag_user_seconds"))<my_time(NULL)) {
                users_to_lag.erase(i1);
                goto again;
            }
        }
        bool got=false;
        for(i1=users_to_lag.begin(); i1!=users_to_lag.end(); ++i1) {
            if(!(*i1).user_name.compare(user_name)) {
                (*i1).cnt++;
                (*i1).last_try=my_time(NULL);
                got=true;
                break;
            }
        }
        if(got) {
            if((*i1).cnt>stoi(core_conf_getvar("ctcp_lag_user_msgs"))) {
                my_time_t x=(*i1).last_try;
                my_time_t now=my_time(NULL);
                while(x>0 && (*i1).cnt>0 && my_difftime(now,x)>sto_time(core_conf_getvar("ctcp_lag_user_seconds"))) {
                    x-=sto_time(core_conf_getvar("ctcp_lag_user_seconds"));
                    (*i1).cnt--;
                }
                if((*i1).cnt>stoi(core_conf_getvar("ctcp_lag_user_msgs"))) {
                    if(!(*i1).notified) {
                        (*i1).notified=true;
                        string msg=":";
                        string lang="en"; // TODO: !!!
                        msg+=lang_get_string(1,lang,446);
                        irc_putserv(c_tracked_msg_to_server("PRIVMSG").add_nick(nick).add_piece(msg),false,LOW_PRIORITY);

                        logic_on_internal_event("@dcc_request_flood@",user_name,nick,host_mask,"",PRIVATE_EVENT_SEVERITY_CODE_WARNING,"Ignoring DCC invitation. DCC flood detected.",str);
                    }
                    return;
                }
                if((*i1).cnt==0) {
                    (*i1).cnt=1;
                    (*i1).notified=false;
                }
            }
        } else {
            c_user_to_lag u;
            u.cnt=1;
            u.last_try=my_time(NULL);
            u.user_name=user_name;
            u.notified=false;
            users_to_lag.push_back(u);
        }
    }

    if((str.length()>=14 && str[0]=='D' && str[1]=='C' && str[2]=='C' && str[3]==' ' && str[4]=='C' && str[5]=='H' && str[6]=='A' && str[7]=='T' && str[8]==' ' && str[9]=='c' && str[10]=='h' && str[11]=='a' && str[12]=='t' && str[13]==' ') ||
       (str.length()>=14 && str[0]=='D' && str[1]=='C' && str[2]=='C' && str[3]==' ' && str[4]=='C' && str[5]=='H' && str[6]=='A' && str[7]=='T' && str[8]==' ' && str[9]=='C' && str[10]=='H' && str[11]=='A' && str[12]=='T' && str[13]==' ') ||
       (str.length()>=9  && str[0]=='D' && str[1]=='C' && str[2]=='C' && str[3]==' ' && str[4]=='S' && str[5]=='E' && str[6]=='N' && str[7]=='D' && str[8]==' ')) {
        bool send=str[4]=='S';

        string user_name_as_in_logic=user_name;

        {
            c_user u;
            vector<c_channel> chs;
            logic_partyline_get_user(user_name_as_in_logic,u,chs);
            if(send && !u.access_to_filesystem) {
                logic_on_internal_event("@dcc_request_access@",user_name,nick,host_mask,"",PRIVATE_EVENT_SEVERITY_CODE_WARNING,"Ignoring DCC invitation. User has no access to the filesystem.",str);
                return;
            }
            if(!send && !u.access_to_partyline) {
                logic_on_internal_event("@dcc_request_access@",user_name,nick,host_mask,"",PRIVATE_EVENT_SEVERITY_CODE_WARNING,"Ignoring DCC invitation. User has no access to the partyline.",str);
                return;
            }
        }

        string file_name;

        int begin_pos=14;

        if(send) {
            if(dcc_get_num_of_transfers()>=stoi(core_conf_getvar("dcc_max_transfers"))) {
                string msg=":";
                string lang="en"; // TODO: !!!
                msg+=lang_get_string(1,lang,445);
                irc_putserv(c_tracked_msg_to_server("PRIVMSG").add_nick(nick).add_piece(msg),false,LOW_PRIORITY);

                logic_on_internal_event("@dcc_request_flood@",user_name,nick,host_mask,"",PRIVATE_EVENT_SEVERITY_CODE_WARNING,"Ignoring DCC invitation. Maximum number of concurrent DCC SEND connections exceeded.",str);
                return;
            }

            int i1=9;
            for(; str[i1]!=0; ++i1) {
                if(str[i1]==0x20)
                    break;
                file_name+=str[i1];
            }
            if(str[i1]==0x20)
                i1++;
            begin_pos=i1;
        }

        string ip_, port_;
        int i1=begin_pos;
        for(; str[i1]!=0; ++i1) {
            if(str[i1]==0x20)
                break;
            ip_+=str[i1];
        }
        if(str[i1]!=0)
            i1++;
        for(; str[i1]!=0; ++i1) {
            if(str[i1]==0x20)
                break;
            port_+=str[i1];
        }

        if(str[i1]!=0)
            i1++;

        string size;

        if(send) {
            while(str[i1]!=0 && str[i1]==0x20)
                i1++;
            for(; str[i1]!=0; ++i1) {
                if(str[i1]==0x20)
                    break;
                size+=str[i1];
            }
        }

        uint32 ip=0;
        string ip_addr;
        in_addr6_ ip6;
        memset(&ip6,0,sizeof(ip6));
        bool got_ipv6=false;
        long i_size=stol(size);
        if(i_size<0)
            i_size=0;
        unsigned short port=stous(port_);

        if(ip_.find(":",0)==string::npos && ip_.find(".",0)==string::npos) {
            got_ipv6=false;
            uint32 ip=stoui(ip_);
            ip_addr=long2ip((uint32)ip);
        }
        if(ip_.find(":",0)==string::npos && ip_.find(".",0)!=string::npos) {
            got_ipv6=false;
            if(my_inet_pton(AF_INET,ip_.c_str(),&ip)==1) {
                ip_addr=ip_;
            } else
                return; // error
        }
        if(ip_.find(":",0)!=string::npos) {
            got_ipv6=true;
            if(my_inet_pton(AF_INET6,ip_.c_str(),&ip6)==1) {
                ip_addr=ip_;
            } else
                return; // error
        }

        string dcc_bind_ip=core_conf_getvar("dcc_bind_ip");
        if(dcc_bind_ip.empty())
            dcc_bind_ip="0.0.0.0";
        string dcc_bind_ip6=core_conf_getvar("dcc_bind_ip6");
        if(dcc_bind_ip6.empty())
            dcc_bind_ip6="::";

        if(stoi(core_conf_getvar("dcc_bind_ip_dynamic"))!=0) {
            int type=sock_get_address_type(irc_host);
            if(type==4)
                dcc_bind_ip=irc_host;
            if(type==0) {
                in_addr ip;
                memset(&ip,0,sizeof(ip));
                threads_resolve4(irc_host.c_str(),ip,RT_RESOLVE_SCHEDULE);
                // ^ if it fails, or has not beed resolved yet
                // (e.g. it is not in cache), 0.0.0.0 will be used
                char tmp[128+1];
                tmp[0]=0;
                my_inet_ntop(AF_INET,&ip,tmp,128);
                dcc_bind_ip=tmp;
            }
        }

        if(stoi(core_conf_getvar("dcc_bind_ip6_dynamic"))!=0) {
            int type=sock_get_address_type(irc_host);
            if(type==6)
                dcc_bind_ip6=irc_host;
            if(type==0) {
                in_addr6_ ip;
                memset(&ip,0,sizeof(ip));
                threads_resolve6(irc_host.c_str(),ip,RT_RESOLVE_SCHEDULE);
                // ^ if it fails, or has not beed resolved yet
                // (e.g. it is not in cache), "::" will be used
                char tmp[128+1];
                tmp[0]=0;
                my_inet_ntop(AF_INET6,&ip,tmp,128);
                dcc_bind_ip6=tmp;
            }
        }

        if(send) {
            MY_LONG_LONG got_bytes;
            bool resume=filesys_dcc_check_for_resume(user_name_as_in_logic,file_name,got_bytes);
            if(resume) {
                string msg=":\x01" "DCC RESUME ";
                msg+=file_name;
                msg+=" ";
                msg+=port_;
                msg+=" ";
                msg+=tos(got_bytes);
                msg+="\x01";
                irc_putserv(c_tracked_msg_to_server("PRIVMSG").add_nick(nick).add_piece(msg),false,LOW_PRIORITY);

                c_dcc_client_to_resume res;
                res.comment=comment;
                res.file_name=file_name;
                res.file_size=i_size;
                res.using_ipv6=got_ipv6;
                if(!got_ipv6) {
                    res.ip_addr4=ip_;
                    res.ip_4=ip;

                    res.ip_addr6="";
                    memset(&res.ip_6,0,sizeof(res.ip_6));
                } else {
                    res.ip_addr4="";
                    res.ip_4=0;

                    res.ip_addr6=ip_;
                    res.ip_6=ip6;
                }
                res.nick=nick;
                res.port=port;
                res.start_time=my_time(NULL);
                res.resume_position=got_bytes;
                res.user_name=user_name_as_in_logic;

                dcc_to_resume.push_back(res);
                return;
            } else {
                int ec=filesys_check_add_file_raw(file_name,user_name_as_in_logic);
                if(ec==FILESYS_ALREADY_EXISTS) {
                    string msg=":";

                    string lang="en"; // TODO: !!!
                    string m=lang_subst(lang_get_string(1,lang,444),file_name);

                    msg+=m;
                    irc_putserv(c_tracked_msg_to_server("PRIVMSG").add_nick(nick).add_piece(msg),false,LOW_PRIORITY);
                    return;
                }
            }
        }

        {
            bool spoofing=false;
            in_addr6_ supplied6;
            in_addr6_ originator6;
            in_addr supplied4;
            in_addr originator4;
            if(got_ipv6) {
                e_resolv_result res1=threads_resolve6(ip_addr.c_str(),supplied6,RT_RESOLVE_SCHEDULE);
                e_resolv_result res2=threads_resolve6(irc_get_host(nick).c_str(),originator6,RT_RESOLVE_SCHEDULE);
                if(RR_SCHEDULED==res1 || RR_SCHEDULED==res2) {
                    c_dcc_request_whois r;
                    r.clear();
                    my_time(&r.start);
                    r.str=str;
                    r.nick=nick;
                    r.host_mask=host_mask;
                    r.comment=comment;
                    r.only_dns=true;
                    r.resolv6_1=ip_addr;
                    r.resolv6_2=irc_get_host(nick);
                    r.got_whois=true;
                    dcc_request_whois.push_back(r);
                    return;
                }
                unsigned char sum_supplied=0;
                unsigned char sum_originator=0;
                for(size_t i1=0; i1<sizeof(supplied6); ++i1) {
                    sum_supplied|=((unsigned char*)&supplied6)[i1];
                    sum_originator|=((unsigned char*)&originator6)[i1];
                    if(((unsigned char*)&supplied6)[i1]!=((unsigned char*)&originator6)[i1]) {
                        spoofing=true;
                        break;
                    }
                }
                if(!sum_supplied || !sum_originator || RR_OK!=res1 || RR_OK!=res2)
                    spoofing=true; // some of addresses are zero
            } else {
                e_resolv_result res1=threads_resolve4(ip_addr.c_str(),supplied4,RT_RESOLVE_SCHEDULE);
                e_resolv_result res2=threads_resolve4(irc_get_host(nick).c_str(),originator4,RT_RESOLVE_SCHEDULE);
                if(RR_SCHEDULED==res1 || RR_SCHEDULED==res2) {
                    c_dcc_request_whois r;
                    r.clear();
                    my_time(&r.start);
                    r.str=str;
                    r.nick=nick;
                    r.host_mask=host_mask;
                    r.comment=comment;
                    r.only_dns=true;
                    r.resolv4_1=ip_addr;
                    r.resolv4_2=irc_get_host(nick);
                    r.got_whois=true;
                    dcc_request_whois.push_back(r);
                    return;
                }
                unsigned char sum_supplied=0;
                unsigned char sum_originator=0;
                for(size_t i1=0; i1<sizeof(supplied4); ++i1) {
                    sum_supplied|=((unsigned char*)&supplied4)[i1];
                    sum_originator|=((unsigned char*)&originator4)[i1];
                    if(((unsigned char*)&supplied4)[i1]!=((unsigned char*)&originator4)[i1]) {
                        spoofing=true;
                        break;
                    }
                }
                if(!sum_supplied || !sum_originator || RR_OK!=res1 || RR_OK!=res2)
                    spoofing=true; // some of addresses are zero
            }
            if(spoofing) {
                string msg="Warning: DCC IP address spoofing from user ";
                msg+=host_mask;
                msg+=" with supplied IPv";
                msg+=(got_ipv6?"6":"4");
                msg+=" address \"";
                msg+=ip_;
                msg+="\" translated to ";
                char tmp[128+1];
                tmp[0]=tmp[128]=0;
                if(got_ipv6)
                    my_inet_ntop(AF_INET6,(const void*)&supplied6,tmp,128);
                else
                    my_inet_ntop(AF_INET,(const void*)&supplied4,tmp,128);
                msg+=tmp;
                msg+=", but host of sender of CTCP message was decoded/and/or/DNS-translated to IPv";
                msg+=(got_ipv6?"6":"4");
                msg+=" address ";
                tmp[0]=tmp[128]=0;
                if(got_ipv6)
                    my_inet_ntop(AF_INET6,(const void*)&originator6,tmp,128);
                else
                    my_inet_ntop(AF_INET,(const void*)&originator4,tmp,128);
                msg+=tmp;
                log_bot(msg);
                if(stoi(core_conf_getvar("dcc_disable_reverse_ip_check"))==0)
                    return;
            }
        }

        string comment_="DCC client initialized with string ";
        comment_+=comment;

        c_dcc_client c;

        if(!got_ipv6)
            c.connect_id=threads_connect(dcc_bind_ip,ip_addr,port,comment_,true);
        else
            c.connect_id=threads_connect(dcc_bind_ip6,ip_addr,port,comment_,true);
        c.socket.clear();

        if(!got_ipv6)
            c.local=dcc_bind_ip;
        else
            c.local=dcc_bind_ip6;
        c.remote=ip_addr;
        c.local_port=0;
        c.remote_port=port;

        c.nick=nick;
        c.ident=irc_get_ident(nick);
        c.host=irc_get_host(nick);
        c.fullname=irc_get_fullname(nick);

        c.resume_pos=0;

        c.chat=!send;
        c.send=send;
        c.send_from_filesystem=false;

        c.server_socket.clear();
        c.telnet=false;
        c.got_passwd=0;
        c.new_user=0;
        c.edit_user=0;
        c.edit_pass=0;
        c.filesystem=0;
        c.as_in_logic=user_name_as_in_logic;
        c.got_user_name=(c.as_in_logic.empty()?false:true);

        if(send) {
            c.dcc_send.size=i_size;
            c.dcc_send.original_name=file_name;
            c.dcc_send.is_file_server_proto=false;
        }

        c.client_id=++dcc_last_client_id;
        log_clients(c.client_id,comment_);
        log_clients(c.client_id,((string)"Connecting to [")+ip_addr+"]:"+tos(port));

        my_time(&c.session_start);
        if(c.chat)
            c.connect_time=dcc_push_bad_login(c.remote);
        dcc_clients.push_back(c);
        dcc_check_limit();
    }

    if(str.length()>=11 && str[0]=='D' && str[1]=='C' && str[2]=='C' && str[3]==' ' && str[4]=='A' && str[5]=='C' && str[6]=='C' && str[7]=='E' && str[8]=='P' && str[9]=='T' && str[10]==' ') {
        if(dcc_get_num_of_transfers()>=stoi(core_conf_getvar("dcc_max_transfers"))) {
            string msg=":";
            string lang="en"; // TODO: !!!
            msg+=lang_get_string(1,lang,445);
            irc_putserv(c_tracked_msg_to_server("PRIVMSG").add_nick(nick).add_piece(msg),false,HIGH_PRIORITY);

            logic_on_internal_event("@dcc_request_flood@",user_name,nick,host_mask,"",PRIVATE_EVENT_SEVERITY_CODE_WARNING,"Ignoring DCC invitation. Maximum number of concurrent DCC SEND connections exceeded.",str);
            return;
        }

        string file_name;

        int begin_pos=11;

        bool send=true;
        if(send) {
            int i1=11;
            for(; str[i1]!=0; ++i1) {
                if(str[i1]==0x20)
                    break;
                file_name+=str[i1];
            }
            if(str[i1]==0x20)
                i1++;
            begin_pos=i1;
        }

        string user_name_as_in_logic=user_name;

        {
            c_user u;
            vector<c_channel> chs;
            logic_partyline_get_user(user_name_as_in_logic,u,chs);

            if(send && !u.access_to_filesystem) {
                logic_on_internal_event("@dcc_request_access@",user_name,nick,host_mask,"",PRIVATE_EVENT_SEVERITY_CODE_WARNING,"Ignoring DCC invitation. User has no access to the filesystem.",str);
                return;
            }
            if(!send && !u.access_to_partyline) {
                logic_on_internal_event("@dcc_request_access@",user_name,nick,host_mask,"",PRIVATE_EVENT_SEVERITY_CODE_WARNING,"Ignoring DCC invitation. User has no access to the partyline.",str);
                return;
            }
        }

        string port_, position_;
        int i1=begin_pos;
        for(; str[i1]!=0; ++i1) {
            if(str[i1]==0x20)
                break;
            port_+=str[i1];
        }
        if(str[i1]!=0)
            i1++;
        for(; str[i1]!=0; ++i1) {
            if(str[i1]==0x20)
                break;
            position_+=str[i1];
        }

        if(str[i1]!=0)
            i1++;

        unsigned short port=stous(port_);

        {
            bool got=false;
            vector<c_dcc_client_to_resume>::iterator i1;
            for(i1=dcc_to_resume.begin(); i1!=dcc_to_resume.end(); ++i1)
                // BEGIN bug in mIRC workaround
                //if(!(*i1).file_name.compare(file_name) && (*i1).port==port && !(*i1).nick.compare(nick) && !(*i1).user_name.compare(user_name_as_in_logic)) {
                if(!(*i1).nick.compare(nick) && !(*i1).user_name.compare(user_name_as_in_logic)) {
                    file_name=(*i1).file_name;
                    port=(*i1).port;
                    position_=tos((unsigned)(*i1).resume_position);
                // END   bug in mIRC workaround
                    got=true;
                    break;
                }

            if(!got) {
                // huh! error!
                log_debug(GET_STR_DEBUG_INFO_EX("OOPS: Occurred oops: cannot find entry for DCC RESUME! huh?!"));
                return;
            }

            c_dcc_client_to_resume res=*i1;
            dcc_to_resume.erase(i1);

            MY_LONG_LONG i_size2;
            {
                size_t i_size;
                memset(&i_size,0xff,sizeof(i_size));
                i_size2=i_size;
                if(res.file_size<i_size2)
                    i_size2=res.file_size;
            }
            string ip_addr;
            if(!res.using_ipv6)
                ip_addr=res.ip_addr4;
            else
                ip_addr=res.ip_addr6;

            string dcc_bind_ip=core_conf_getvar("dcc_bind_ip");
            if(dcc_bind_ip.empty())
                dcc_bind_ip="0.0.0.0";
            string dcc_bind_ip6=core_conf_getvar("dcc_bind_ip6");
            if(dcc_bind_ip6.empty())
                dcc_bind_ip6="::";

            string comment_="DCC client initialised with string ";
            comment_+=res.comment;

            c_dcc_client c;

            if(!res.using_ipv6)
                c.connect_id=threads_connect(dcc_bind_ip,ip_addr,port,comment_,true);
            else
                c.connect_id=threads_connect(dcc_bind_ip6,ip_addr,port,comment_,true);
            c.socket.clear();

            if(!res.using_ipv6)
                c.local=dcc_bind_ip;
            else
                c.local=dcc_bind_ip6;
            c.remote=ip_addr;
            c.local_port=0;
            c.remote_port=port;

            c.nick=nick;
            c.ident=irc_get_ident(nick);
            c.host=irc_get_host(nick);
            c.fullname=irc_get_fullname(nick);

            c.resume_pos=stoll(position_);

            c.chat=!send;
            c.send=send;
            c.send_from_filesystem=false;

            c.server_socket.clear();
            c.telnet=false;
            c.got_passwd=0;
            c.new_user=0;
            c.edit_user=0;
            c.edit_pass=0;
            c.filesystem=0;
            c.as_in_logic=user_name_as_in_logic;
            c.got_user_name=(c.as_in_logic.empty()?false:true);

            c.dcc_send.size=i_size2;
            c.dcc_send.original_name=file_name;
            c.dcc_send.is_file_server_proto=false;

            c.client_id=++dcc_last_client_id;
            log_clients(c.client_id,comment_);
            log_clients(c.client_id,((string)"Connecting to [")+ip_addr+"]:"+tos(port));

            my_time(&c.session_start);
            if(c.chat)
                c.connect_time=dcc_push_bad_login(c.remote);
            dcc_clients.push_back(c);
            dcc_check_limit();
        }
    }

    if(str.length()>=11 && str[0]=='D' && str[1]=='C' && str[2]=='C' && str[3]==' ' && str[4]=='R' && str[5]=='E' && str[6]=='S' && str[7]=='U' && str[8]=='M' && str[9]=='E' && str[10]==' ') {
        if(dcc_get_num_of_transfers()>=stoi(core_conf_getvar("dcc_max_transfers"))) {
            string msg=":";
            string lang="en"; // TODO: !!!
            msg+=lang_get_string(1,lang,445);
            irc_putserv(c_tracked_msg_to_server("PRIVMSG").add_nick(nick).add_piece(msg),false,HIGH_PRIORITY);

            logic_on_internal_event("@dcc_request_flood@",user_name,nick,host_mask,"",PRIVATE_EVENT_SEVERITY_CODE_WARNING,"Ignoring DCC invitation. Maximum number of concurrent DCC SEND connections exceeded.",str);
            return;
        }

        string file_name;

        int begin_pos=11;

        string user_name_as_in_logic=user_name;

        {
            c_user u;
            vector<c_channel> chs;
            logic_partyline_get_user(user_name_as_in_logic,u,chs);
            if(!u.access_to_filesystem) {
                logic_on_internal_event("@dcc_request_access@",user_name,nick,host_mask,"",PRIVATE_EVENT_SEVERITY_CODE_WARNING,"Ignoring DCC invitation. User has no access to the filesystem.",str);
                return;
            }
        }

        string port_, position_;
        int i1=begin_pos;

        bool q=str[i1]=='"';
        for(; str[i1]!=0; ++i1) {
            if(str[i1]==0x20 && !q)
                break;
            if(str[i1]=='"')
                q=false;
            file_name+=str[i1];
        }
        if(file_name.length()>0 && file_name[0]=='"')
            file_name.erase(0,1);
        if(file_name.length()>0 && file_name[file_name.length()-1]=='"')
            file_name.erase(file_name.length()-1,1);

        if(str[i1]!=0)
            i1++;

        for(; str[i1]!=0; ++i1) {
            if(str[i1]==0x20)
                break;
            port_+=str[i1];
        }
        if(str[i1]!=0)
            i1++;
        for(; str[i1]!=0; ++i1) {
            if(str[i1]==0x20)
                break;
            position_+=str[i1];
        }

        if(str[i1]!=0)
            i1++;

        MY_LONG_LONG pos=stoll(position_);

        string msg=":\x01" "DCC ACCEPT ";
        msg+=(string)"\""+file_name+"\"";
        msg+=" ";
        msg+=port_;
        msg+=" ";
        msg+=position_;
        msg+="\x01";
        irc_putserv(c_tracked_msg_to_server("PRIVMSG").add_nick(nick).add_piece(msg),false,HIGH_PRIORITY);

        {
            vector<c_dcc_send_file>::iterator i1;
            for(i1=dcc_awaiting_send.begin(); i1!=dcc_awaiting_send.end(); ++i1) {
                if(!(*i1).user_name_as_in_logic.compare(user_name_as_in_logic) && !(*i1).public_name.compare(file_name)) {
                    (*i1).resume_pos=pos;
                    break;
                }
            }
        }
    }
}

/*!
    \brief Extracts one line from buffer
    \author VooDooMan
    \version 2
    \date 2004, 2005-12-10
    \param buff Buffer
    \param keep true if to keep in buffer
    \return Extracted line
*/
string dcc_get_line(string& buff, bool keep)
{
    if(buff.find('\n')==string::npos)
        return (string)"";
    string res;
    unsigned int i1;
    for(i1=0; i1<buff.length(); ++i1) {
        if(buff[i1]=='\n') {
            if(!keep)
                buff[i1]='X'; // overwrite (potentially) sensitive data before it gets released to the system
            break;
        }
        res+=buff[i1];
        if(!keep)
            buff[i1]='X'; // overwrite (potentially) sensitive data before it gets released to the system
    }

    if(!keep)
        buff.erase(0,i1+1);

    if(res.length()>0 && res[res.length()-1]=='\r')
        res.erase(res.length()-1,1);
    if(res.length()>0 && res[res.length()-1]=='\n')
        res.erase(res.length()-1,1);
    if(res.length()>0 && res[res.length()-1]=='\r')
        res.erase(res.length()-1,1);
    if(res.length()>0 && res[res.length()-1]=='\n')
        res.erase(res.length()-1,1);

    // erase backspace's
l:
    for(i1=0; i1<res.length(); ++i1)
        if(res[i1]==8) {
            if(i1==0) {
                res.erase(0,1);
                goto l;
            }
            res.erase(i1-1,2);
            goto l;
        }

    return res;
}

/*!
    \brief Check wheter user can unassign procedure for event (has access_usage_procedure right)
    \author VooDooMan
    \version 1
    \date 2004
    \param i Descriptor of remote connection
    \param eol End-of-line string
    \param proc_name Name of procedure
    \returns Returns true if can unassign
*/
bool dcc_can_unassign(vector<c_dcc_client>::iterator i, string& UNUSEDARG(eol), string proc_name)
{
    if(proc_name.empty())
        return true;

    if(proc_name.find("(",0)!=string::npos)
        proc_name.erase(proc_name.find("(",0),proc_name.length()-proc_name.find("(",0));

    vector<string>::iterator i1;
    for(i1=(*i).access_usage_proc.begin(); i1!=(*i).access_usage_proc.end(); ++i1) {
        if(!(*i1).compare("*"))
            return true;

        string name_only=*i1;
        if(name_only.find("(",0)!=string::npos)
            name_only.erase(name_only.find("(",0),name_only.length()-name_only.find("(",0));

        if(!name_only.compare(proc_name))
            return true;
    }

    return false;
}

/*!
    \brief Checks for flood on DCC/telnet chat/command connection
    \author VooDooMan
    \version 1
    \date 2004, 2005-12-10
    \param i Descriptor of remote connection
    \param eol EOL sequence
*/
bool dcc_is_flood(vector<c_dcc_client>::iterator i, const string& eol)
{
    my_time_t now;
    my_time(&now);
    // clean up flood buffers
    vector<c_flood_history>::iterator i4;
l_flood2:
    for(i4=(*i).msg_flood_hist.begin(); i4!=(*i).msg_flood_hist.end(); ++i4)
        if(now-(*i4).time>(*i).partyline_msg_flood.seconds) {
            (*i).msg_flood_hist.erase(i4);
            goto l_flood2;
        }

    string c=dcc_get_line((*i).in_buff,true);
    if(c.empty())
        return false;

    c_flood_history entry;
    entry.time=now;
    entry.bytes=c.length();
    entry.msg=c;

    (*i).msg_flood_hist.push_back(entry);
    unsigned int entries=static_cast<unsigned int>((*i).msg_flood_hist.size());
    if((*i).partyline_msg_flood.lines!=0 && entries>=(*i).partyline_msg_flood.lines) {
        dcc_get_line((*i).in_buff); // discard
        if((*i).last_flood+(*i).partyline_msg_flood.seconds<my_time(NULL)) {
            string disp=lang_subst(lang_get_string(1,(*i).lang,207),tos(static_cast<unsigned long>((*i).partyline_msg_flood.seconds)))+eol;
            int ec=0;
            dcc_send_msg((*i).socket,disp,ec,(*i).ssl);
            (*i).last_flood=my_time(NULL);
            return true;
        }
    }
    return false;
}

/*!
    \brief Gets/resets user's access rights
    \author VooDooMan
    \version 1
    \date 2004
    \param i Descriptor of remote connection
*/
void dcc_loop_get_rights(vector<c_dcc_client>::iterator i)
{
    (*i).chs.clear();
    logic_partyline_get_user((*i).as_in_logic,(*i).u,(*i).chs);
    logic_partyline_access(
        (*i).as_in_logic,
        (*i).access_to_users,
        (*i).access_to_groups,
        (*i).access_grant_groups,
        (*i).access_to_proc_groups,
        (*i).access_grant_proc_groups,
        (*i).access_grant_proc,
        (*i).access_to_channel,
        (*i).access_grant_channel,
        (*i).access_grant_partyline,
        (*i).access_to_procs,
        (*i).access_to_backup,
        (*i).access_grant_backup,
        (*i).access_to_rehash,
        (*i).access_grant_rehash,
        (*i).access_to_plususer,
        (*i).access_grant_plususer,
        (*i).access_to_plusproc,
        (*i).access_grant_plusproc,
        (*i).partyline_msg_flood,
        (*i).access_to_replication,
        (*i).access_grant_replication,
        (*i).access_to_chan_defs,
        (*i).access_to_restart,
        (*i).access_grant_restart,
        (*i).access_to_die,
        (*i).access_grant_die,
        (*i).access_to_filesystem,
        (*i).access_grant_filesystem,
        (*i).access_to_private,
        (*i).access_grant_private,
        (*i).access_to_can_send_all_users,
        (*i).access_grant_can_send_all_users,
        (*i).access_to_can_send_unknown_users,
        (*i).access_grant_can_send_unknown_users,
        (*i).access_usage_proc,
        (*i).access_usage_proc_groups,
        (*i).access_to_upgrade,
        (*i).access_grant_upgrade,
        (*i).access_to_apply,
        (*i).access_grant_apply,
        (*i).access_to_users_also_terminators
        );
}

/*!
    \brief Stores temporary data for counting amount of sessions
    \author VooDooMan
    \version 1
    \date 2005
*/
class c_session
{
public:
    int num;
    vector<c_dcc_client*> c;
};

/*!
    \brief Check for maximum connection limit rule, and drop some
    \author VooDooMan
    \version 1
    \date 2005
*/
void dcc_check_limit()
{
    if(stoi(core_conf_getvar("dcc_max_sessions_per_ip"))==0)
        return;

    map<string,c_session> sessions;

    vector<c_dcc_client>::iterator i1;
    for(i1=dcc_clients.begin(); i1!=dcc_clients.end(); ++i1) {
        if(TCID_INVALID!=(*i1).connect_id)
            continue;
        if(!(*i1).replication_partner && (*i1).chat && !(*i1).remote.empty()) {
            pair<string,c_session> p;
            p.first=(*i1).remote;
            p.second.num=1;
            p.second.c.clear();
            p.second.c.push_back(&(*i1));

            map<string,c_session>::iterator i2;
            bool got=false;
            for(i2=sessions.begin(); i2!=sessions.end(); ++i2) {
                if(!(*i2).first.compare(p.first)) {
                    got=true;
                    (*i2).second.num++;
                    break;
                }
            }
            if(!got)
                sessions.insert(p);
        }
    }
    map<string,c_session>::iterator i2;
    for(i2=sessions.begin(); i2!=sessions.end(); ++i2) {
        if((*i2).second.num>stoi(core_conf_getvar("dcc_max_sessions_per_ip"))) {
            c_dcc_client* max_idle=NULL;

            vector<c_dcc_client*>::iterator i3;
            for(i3=(*i2).second.c.begin(); i3!=(*i2).second.c.end(); ++i3) {
                if(TCID_INVALID!=(*i3)->connect_id)
                    continue;
                if(max_idle==NULL) {
                    max_idle=*i3;
                    continue;
                }
                if((*i3)->last_input<max_idle->last_input)
                    max_idle=*i3;
            }
            if(max_idle==NULL) // huh?! just in case... ;-)
                continue;

            string lang=max_idle->lang;
            if(lang.empty())
                lang="en";

            char dcc_eol[]="\n";
            char telnet_eol[]="\r\n";
            string eol=max_idle->telnet?telnet_eol:dcc_eol;

            if(dcc_killing_user_socket && max_idle->socket==*dcc_killing_user_socket) {
            } else {
                vector<c_dcc_client>::iterator i4;
                for(i4=dcc_clients.begin(); i4!=dcc_clients.end(); ++i4) {
                    if(TCID_INVALID!=(*i4).connect_id)
                        continue;
                    if((*i4).socket==max_idle->socket) {
                        log_clients((*i4).client_id,"Connection count limit in effect, closing connection");

                        string disp=lang_get_string(1,lang,620)+eol;
                        int ec=0;
                        dcc_send_msg(max_idle->socket,disp,ec,max_idle->ssl);

                        //if(max_idle->is_ssl_client && max_idle->ssl) {
                        if(max_idle->ssl) {
                            ssl_put_to_garbage(max_idle->ssl);
                            max_idle->ssl=NULL;
                            max_idle->socket.clear();
                        } else {
                            sock_linger_push(max_idle->socket);
                            max_idle->socket.clear();
                        }

                        (*i4).destroy();
                        dcc_clients.erase(i4);
                        break;
                    }
                }
            }
        }
    }
}

/*!
    \brief Called after .apply. Applies new configuration to all connected users on DCC/telnet (also BOTNET) connections.
    \author VooDooMan
    \version 1
    \date 2005
    \param who This should contain who caused .apply (to inform connected users)
*/
void dcc_apply(string who)
{
    vector<c_dcc_client>::iterator i;
    char dcc_eol[]="\n";
    char telnet_eol[]="\r\n";
    for(i=dcc_clients.begin(); i!=dcc_clients.end(); ++i) {
        if(TCID_INVALID!=(*i).connect_id)
            continue;
        if((*i).got_passwd!=3)
            continue;
        dcc_loop_get_rights(i);
        if((*i).send)
            continue;
        if((*i).send_from_filesystem)
            continue;
        if((*i).replication_partner)
            continue;
        if((*i).ssl_connection)
            continue;
        if(!(*i).chat)
            continue;
        string eol=(*i).telnet?telnet_eol:dcc_eol;
        string disp=lang_subst(lang_get_string(1,(*i).lang,687),who)+eol;
        int ec=0;
        dcc_send_msg((*i).socket,disp,ec,(*i).ssl);
    }
}

/*!
    \brief Removes users awaiting for DCC CHAT with same IP address (security issue: two or more clients connecting from sam IP - via proxy - rare, but possible)
    \author VooDooMan
    \version 1
    \date 2004, 2005-12-09
*/
void dcc_cleanup_same_ip()
{
    vector<uint32> ip4;
    vector<in_addr6_> ip6;

    vector<c_dcc_awaiting_user>::iterator i;
again_:
    for(i=dcc_awaiting_users.begin(); i!=dcc_awaiting_users.end(); ++i) {
        if((*i).start_time+20<my_time(NULL)) {
            dcc_awaiting_users.erase(i);
            goto again_;
        }
        if(!(*i).using_ipv6) {
#if defined(_WIN32) || defined(_WIN64)
            if((*i).addr4.S_un.S_addr)
                ip4.push_back((*i).addr4.S_un.S_addr);
#else
            if((*i).addr4.s_addr)
                ip4.push_back((*i).addr4.s_addr);
#endif
        }
        if((*i).using_ipv6) {
            if( (*i).addr6.bytes[0] ||
                (*i).addr6.bytes[1] ||
                (*i).addr6.bytes[2] ||
                (*i).addr6.bytes[3] ||
                (*i).addr6.bytes[4] ||
                (*i).addr6.bytes[5] ||
                (*i).addr6.bytes[6] ||
                (*i).addr6.bytes[7] ||
                (*i).addr6.bytes[8] ||
                (*i).addr6.bytes[9] ||
                (*i).addr6.bytes[10] ||
                (*i).addr6.bytes[11] ||
                (*i).addr6.bytes[12] ||
                (*i).addr6.bytes[13] ||
                (*i).addr6.bytes[14] ||
                (*i).addr6.bytes[15])
                    ip6.push_back((*i).addr6);
        }
    }

    {
        sort(ip4.begin(),ip4.end());
        vector<uint32>::iterator i1;
        uint32 last=0;
        for(i1=ip4.begin(); i1!=ip4.end(); ++i1) {
            if(*i1==last) {
        ip4_again:
                for(i=dcc_awaiting_users.begin(); i!=dcc_awaiting_users.end(); ++i) {
#if defined(_WIN32) || defined(_WIN64)
                    if((*i).addr4.S_un.S_addr==last) {
#else
                    if((*i).addr4.s_addr==last) {
#endif
                        dcc_awaiting_users.erase(i);
                        goto ip4_again;
                    }
                }
            }
            last=*i1;
        }
    }

    {
        sort(ip6.begin(),ip6.end());
        vector<in_addr6_>::iterator i1;
        in_addr6_ last;
        for(size_t i2=0; i2<16; ++i2)
            last.bytes[i2]=0;
        for(i1=ip6.begin(); i1!=ip6.end(); ++i1) {
            if(*i1==last) {
        ip6_again:
                for(i=dcc_awaiting_users.begin(); i!=dcc_awaiting_users.end(); ++i) {
                    if((*i).addr6==last) {
                        dcc_awaiting_users.erase(i);
                        goto ip6_again;
                    }
                }
            }
            last=*i1;
        }
    }
}

/*!
    \brief Message for partyline channel distribution
    \author VooDooMan
    \version 1
    \date 2004
    \param channel Partyline channel
    \param message Partyline message
*/
void dcc_partyline_message(string channel, string message)
{
    vector<c_dcc_client>::iterator i;
    for(i=dcc_clients.begin(); i!=dcc_clients.end(); ++i) {
        if(TCID_INVALID!=(*i).connect_id)
            continue;
        if(!(*i).replication_partner && ((*i).chat || (*i).telnet) && (*i).got_passwd==3 && !(*i).partyline_channel.compare(channel)) {
            string m=message;
            if((*i).telnet)
                m+="\r\n";
            else
                m+="\n";
            if(!(*i).is_ssl_client) {
                int ec=0;
                size_t sent=sock_send((*i).socket,m.c_str(),m.length(),ec);
                stats_dcc_chat_bytes_sent(sent);
            } else {
                if((*i).ssl) {
                    (*i).ssl->write(m.c_str(),m.length());
                    stats_dcc_chat_bytes_sent(m.length());
                }
            }
        }
    }
}

/*!
    \brief Message for partyline user's pong
    \author VooDooMan
    \version 1
    \date 2005
    \param time_ UNIX timestamp of original ping
    \param user Name of user as in "logic.txt"
    \param origin_bot Name of bot from which is event
*/
void dcc_user_pong(string time_, string user, string origin_bot)
{
    vector<c_dcc_client>::iterator i;
    for(i=dcc_clients.begin(); i!=dcc_clients.end(); ++i) {
        if(TCID_INVALID!=(*i).connect_id)
            continue;
        if(!(*i).replication_partner && ((*i).chat || (*i).telnet) && (*i).got_passwd==3 && !(*i).as_in_logic.compare(user)) {
            string eol;
            if((*i).telnet)
                eol+="\r\n";
            else
                eol+="\n";

            string disp=lang_subst(lang_get_string(1,(*i).lang,648),origin_bot)+eol;
            my_time_t now=my_time(NULL);
            my_time_t origin=sto_time(time_);
            disp=lang_subst(disp,tos((long)my_difftime(now,origin))); // not +eol !!!
            int ec=0;
            dcc_send_msg((*i).socket,disp,ec,(*i).ssl);
        }
    }
}

/*!
    \brief Message for partyline user's pong (BOT_VERSION reply)
    \author VooDooMan
    \version 1
    \date 2005
    \param user Name of user as in "logic.txt"
    \param origin_bot Name of bot from which is event
    \param name Version string #1: Name of bot (should be "VooDoo_cIRCle")
    \param number Version string #2: Numeric version string
    \param comment Version string #3: Comment
*/
void dcc_user_pong(string user, string origin_bot, string name, string number, string comment)
{
    vector<c_dcc_client>::iterator i;
    for(i=dcc_clients.begin(); i!=dcc_clients.end(); ++i) {
        if(TCID_INVALID!=(*i).connect_id)
            continue;
        if(!(*i).replication_partner && ((*i).chat || (*i).telnet) && (*i).got_passwd==3 && !(*i).as_in_logic.compare(user)) {
            string eol;
            if((*i).telnet)
                eol+="\r\n";
            else
                eol+="\n";

            string disp=lang_get_string(1,(*i).lang,692);
            disp=lang_subst(disp,origin_bot);
            disp=lang_subst(disp,name);
            disp=lang_subst(disp,number);
            disp=lang_subst(disp,comment);
            disp+=eol;
            int ec=0;
            dcc_send_msg((*i).socket,disp,ec,(*i).ssl);
        }
    }
}

/*!
    \brief Binds DCC server
    \author VooDooMan
    \version 4
    \date 2004, 2005-10-02, 2005-12-09
    \param ipv Version of IP protocol: 4, or 6
    \param group DCC group
    \param bind_ip IPv4/IPv6 bind address
    \param bind_port Port to listen on
    \param server_type Type of server (e.g. "chat", "send", or an empty string)
*/
void dcc_make_server(int ipv, int group, const string& bind_ip, unsigned short bind_port, string server_type)
{
    if(ipv==4) {
        if(bind_ip.length()>32)
            return;

        uint32 ip=0;
        my_inet_pton(AF_INET,bind_ip.c_str(),&ip);
        in_addr addr4;
        memcpy(&addr4,&ip,sizeof(addr4));

        int ec=0;

        c_dcc_server s;

        string comment="DCC server ";
        comment+=bind_ip;
        comment+=":";
        comment+=tos(bind_port);
        c_socket dcc_server;
        if(!bind_ip.empty()) {
            dcc_server=sock_bind(addr4,bind_port,ec,comment.c_str());

            if(dcc_server.cmp()==false) {
                char* errstr=NULL;
                log_socket(ec,errstr=sock_error(ec),GET_STR_DEBUG_INFO_EX("Error while calling sock_bind()"),&dcc_server);
                MY_DELETE_A(errstr);
                return;
            }

            s.active=true;
        } else {
            s.active=false;
            s.dynamic=true;
        }

        s.group=group;
        s.using_ipv6=false;
        s.dcc_addr=ip;
        if(s.dynamic)
            memset(&s.dcc_addr,0,sizeof(s.dcc_addr));
        s.port=bind_port;
        s.bind_ip=bind_ip;
        s.socket=dcc_server;
        if(server_type.find("\"",0)!=string::npos)
            server_type.erase(0,1);
        if(server_type.find("\"",0)!=string::npos)
            server_type.erase(server_type.find("\"",0),1);
        s.server_type=server_type;

        dcc_servers.push_back(s);
    }
    if(ipv==6) {
        string comment="DCC server [";
        comment+=bind_ip;
        comment+="]:";
        comment+=tos(bind_port);

        c_dcc_server s;

        int ec=0;
        c_socket dcc_server;
        if(!bind_ip.empty()) {
            dcc_server=sock_bind6(const_cast<char*>(bind_ip.c_str()),bind_port,ec,comment.c_str());

            if(dcc_server.cmp()==false) {
                char* errstr=NULL;
                log_socket(ec,errstr=sock_error(ec),GET_STR_DEBUG_INFO_EX("Error while calling sock_bind6()"),&dcc_server);
                MY_DELETE_A(errstr);
                return;
            }

            s.active=true;
        } else {
            s.active=false;
            s.dynamic=true;
        }

        s.group=group;
        s.using_ipv6=true;
        s.dcc_addr=(uint32)-1;
        my_inet_pton(AF_INET6,const_cast<char*>(bind_ip.c_str()),&s.dcc_addr6);
        s.port=bind_port;
        s.bind_ip=bind_ip;
        s.socket=dcc_server;
        if(server_type.find("\"",0)!=string::npos)
            server_type.erase(0,1);
        if(server_type.find("\"",0)!=string::npos)
            server_type.erase(server_type.find("\"",0),1);
        s.server_type=server_type;

        dcc_servers.push_back(s);
    }
}

/*!
    \brief Binds telnet server
    \author VooDooMan
    \version 5
    \date 2004, 2005-10-02, 2005-12-09, 2006-07-10
    \param ipv Version of IP protocol: 4, or 6
    \param bind_ip IPv4 bind address, or IPv6 address
    \param bind_port Port to listen on
    \param pure_ssl true if it is SSL server
*/
void dcc_make_telnet_server(int ipv, const string& bind_ip, unsigned short bind_port, bool pure_ssl)
{
    if(ipv==4) {
        if(bind_ip.length()>32)
            return;

        uint32 ip=0;
        my_inet_pton(AF_INET,bind_ip.c_str(),&ip);
        in_addr addr4;
        memcpy(&addr4,&ip,sizeof(addr4));

        int ec=0;

        c_telnet_server s;

        s.pure_ssl=pure_ssl;

        s.addr=ip;

        string comment="Telnet server ";
        comment+=bind_ip;
        comment+=":";
        comment+=tos(bind_port);
        c_socket telnet_server;

        if(!bind_ip.empty()) {
            telnet_server=sock_bind(addr4,bind_port,ec,comment.c_str());

            if(telnet_server.cmp()==false) {
                char* errstr=NULL;
                log_socket(ec,errstr=sock_error(ec),GET_STR_DEBUG_INFO_EX("Error while calling sock_bind()"),&telnet_server);
                MY_DELETE_A(errstr);
                return;
            }

            s.active=true;
        } else {
            s.active=false;
            s.dynamic=true;
        }

        s.using_ipv6=ipv!=4;
        s.telnet_addr=ip;
        s.port=bind_port;
        s.bind_ip=bind_ip;
        s.socket=telnet_server;

        telnet_servers.push_back(s);
    }
    if(ipv==6) {
        string comment="Telnet server [";
        comment+=bind_ip;
        comment+="]:";
        comment+=tos(bind_port);

        c_telnet_server s;

        s.pure_ssl=pure_ssl;

        int ec=0;
        c_socket telnet_server;
        if(!bind_ip.empty()) {
            telnet_server=sock_bind6(const_cast<char*>(bind_ip.c_str()),bind_port,ec,comment.c_str());

            if(telnet_server.cmp()==false) {
                char* errstr=NULL;
                log_socket(ec,errstr=sock_error(ec),GET_STR_DEBUG_INFO_EX("Error while calling sock_bind6()"),&telnet_server);
                MY_DELETE_A(errstr);
                return;
            }

            s.active=true;
        } else {
            s.active=false;
            s.dynamic=true;
        }

        s.using_ipv6=ipv!=4;
        s.telnet_addr=(uint32)-1;
        s.port=bind_port;
        s.bind_ip=bind_ip;
        s.socket=telnet_server;

        telnet_servers.push_back(s);
    }
}

/*!
    \brief Constructs response to "dcc" or "dcc <group_number>" PRIVMSG
    \author VooDooMan
    \version 1
    \date 2004
    \param msg Message that has been received from user (usually "dcc" or "dcc <group_number>"
    \param dcc_group Returns extracted DCC group number from message msg, zero if msg was just "msg"
    \param ip_ver Returns 4 or 6
    \return Returns DCC CHAT request string to send to requesting user as PRIVMSG
*/
string dcc_get_string(string msg, int& dcc_group, int& ip_ver)
{
    ip_ver=4;
    if( (msg.length()>=3 && msg[0]=='d' && msg[1]=='c' && msg[2]=='c') || (msg.length()>=4 && msg[0]=='.' && msg[1]=='d' && msg[2]=='c' && msg[3]=='c')) {
        string group;
        unsigned int i1=3;
        if('.'==msg[0])
            i1++;
        for(; i1<msg.length(); ++i1)
            group+=msg[i1];
        if(!group.compare(""))
            group="0";
        dcc_group=stoi(group);
        vector<c_dcc_server>::iterator i2;
        for(i2=dcc_servers.begin(); i2!=dcc_servers.end(); ++i2) {
            if(!(*i2).active)
                continue;
            if((*i2).group==dcc_group)
                break;
        }
        if(i2==dcc_servers.end())
            return (string)"";

        uint32 ip_=0;
        if(!(*i2).using_ipv6) {
            // it is already in network byte order
            memcpy(&ip_,&(*i2).dcc_addr,4);
#ifndef WORDS_BIGENDIAN
            ip_=swap_bytes(ip_);
#endif
        }

        string res="DCC CHAT chat ";
        char tmp[64];
        if(!(*i2).using_ipv6) {
            sprintf(tmp,"%u",static_cast<unsigned>(ip_));
            res+=tmp;
        } else {
            ip_ver=6;
            string ip6=(*i2).bind_ip;
            if(ip6.find("%",0)!=string::npos)
                ip6.erase(ip6.find("%",0),ip6.length()-ip6.find("%",0));
            res+=ip6;
        }
        res+=" ";
        res+=tos((*i2).port);
        //res+=" ";
        return res;
    }
    return (string)"";
}


/*!
    \brief Notifies user about filesystem events
    \author VooDooMan
    \version 1
    \date 2004
    \param user Name of the user from "logic.txt" file
    \param reason One of the following values: FILESYS_NOT_PUBLISHED - user has uploaded file(s) and they don't have set flags for publishing; FILESYS_GOT_FILE - user has notification about new file to filesystem; FILESYS_GOT_INCOMPLETE - user should be notified for incomplete uploaded file
    \param incomplete_file If reason==FILESYS_GOT_INCOMPLETE there is public name of incomplete received file to filesystem via DCC SEND (undefined in other cases)
*/
void dcc_notify(string user, int reason, string incomplete_file)
{
    char dcc_eol[]="\n";
    char telnet_eol[]="\r\n";

    bool got=false;

    vector<c_dcc_client>::iterator i;
    for(i=dcc_clients.begin(); i!=dcc_clients.end(); ++i) {
        if(TCID_INVALID!=(*i).connect_id)
            continue;
        if((*i).got_passwd!=3)
            continue;
        if(!(*i).chat)
            continue;
        if((*i).as_in_logic.compare(user))
            continue;
        string eol=(*i).telnet?telnet_eol:dcc_eol;

        got=true;

        string disp;
        switch(reason) {
            case FILESYS_NOT_PUBLISHED:
                disp=lang_get_string(1,(*i).lang,274)+eol;
                break;
            case FILESYS_GOT_FILE:
                dcc_check_for_filesystem((*i).as_in_logic);
                break;
            case FILESYS_GOT_INCOMPLETE:
                disp=lang_get_string(1,(*i).lang,400)+eol;
                disp+=incomplete_file;
                disp+=eol;
                break;
        }

        if(!disp.empty()) {
            int ec=0;
            dcc_send_msg((*i).socket,disp,ec,(*i).ssl);
        }
    }

    if(!got && reason==FILESYS_NOT_PUBLISHED) {
        string nick=irc_get_nick(user);
        if(nick.empty())
            return;

        string inv="PRIVMSG ";
        inv+=nick;
        inv+=" :";

        int dcc_group=0; // TODO: !!!

        uint32 dcc_addr=0;
        unsigned short dcc_port=0;
        vector<c_dcc_server>::iterator i1;
        for(i1=dcc_servers.begin(); i1!=dcc_servers.end(); ++i1) {
            if(!(*i1).active)
                continue;
            if((*i1).group==dcc_group) {
                dcc_addr=(*i1).dcc_addr;
                dcc_port=(*i1).port;
                break;
            }
        }

        if(dcc_addr && dcc_port) {
            irc_await_dcc_chat(nick,user,dcc_group);

            uint32 ip_=0;
            // it is already in newtork byte order
            memcpy(&ip_,&dcc_addr,4);

            string res="DCC CHAT chat ";
            char tmp[64];
            sprintf(tmp,"%u",static_cast<unsigned>(ip_));
            res+=tmp;
            res+=" ";
            res+=tos(dcc_port);

            inv+="\x01";
            inv+=res;
            inv+="\x01";

            irc_put(inv,LOW_PRIORITY);
        }
    }
    if(!got && reason==FILESYS_GOT_INCOMPLETE) {
        string nick=irc_get_nick(user);
        if(nick.empty())
            return;

        string lang="en"; // TODO: !!!

        string msg="PRIVMSG ";
        msg+=nick;
        msg+=" :";
        msg+=lang_get_string(1,lang,400);
        irc_put(msg,LOW_PRIORITY);

        msg="PRIVMSG ";
        msg+=nick;
        msg+=" :";
        msg+=incomplete_file;
        irc_put(msg,LOW_PRIORITY);
    }
}

/*!
    \brief Parses message, replaces new line signs
    \author VooDooMan
    \version 1
    \date 2004
    \param msg Message to parse
    \param eol End-of-line sequence
    \return Returns parsed message
*/
string dcc_put_multiple_lines(string msg, string eol)
{
    string result;

    string m;
    for(unsigned int i1=0; i1<msg.length(); ++i1) {
        if(msg[i1]=='\\') {
            if(!(i1+1<msg.length() && msg[i1+1]=='n'))
                m+=msg[i1];
        } else
            m+=msg[i1];
        if(i1==msg.length()-1 || (i1+1<msg.length() && msg[i1]=='\\' && msg[i1+1]=='n')) {
            if((i1+1<msg.length() && msg[i1]=='\\' && msg[i1+1]=='n'))
                i1++;

            if(!m.empty()) {
                //result+="*** ";
                result+=m;
                result+=eol;
            }
            m="";
        }
    }
    result=(string)"***"+eol+result+"***"+eol;
    return result;
}

/*!
    \brief Check filesystem for new files for user
    \author VooDooMan
    \version 1
    \date 2004
    \param user_name Name of the user from "logic.txt" file
*/
void dcc_check_for_filesystem(string user_name)
{
    char dcc_eol[]="\n";
    char telnet_eol[]="\r\n";

    vector<c_dcc_client>::iterator i;
    for(i=dcc_clients.begin(); i!=dcc_clients.end(); ++i) {
        if(TCID_INVALID!=(*i).connect_id)
            continue;
        if((*i).got_passwd!=3)
            continue;
        if(!(*i).chat)
            continue;
        if((*i).as_in_logic.compare(user_name))
            continue;
        string eol=(*i).telnet?telnet_eol:dcc_eol;

        string disp;
        {
            vector<c_dcc_notify> notifies;
            filesys_dcc_check_for_notifies((*i).as_in_logic,(*i).u,(*i).chs,"",notifies,true,(*i).lang,eol);
            filesys_dcc_drop_notifies((*i).as_in_logic,false);
            vector<c_dcc_notify>::iterator i1;
            bool unpublished=false;
            //bool incomplete=false;
            string incomplete_files;
            for(i1=notifies.begin(); i1!=notifies.end(); ++i1) {
                if(!unpublished && (*i1).unpublished) {
                    string msg=lang_get_string(1,(*i).lang,274); // 331
                    disp+=dcc_put_multiple_lines(msg,eol);
                    unpublished=true;
                }
                if((*i1).incomplete) {
                    if(!incomplete_files.empty())
                        incomplete_files+=", ";
                    incomplete_files+="\"";
                    incomplete_files+=(*i1).name;
                    incomplete_files+="\"";
                }
                if((*i1).notify_owner || (*i1).notify_user || (*i1).secure_notify_owner || (*i1).secure_notify_user) {
                    string msg=(*i1).message;
                    disp+=dcc_put_multiple_lines(msg,eol);
                }
            }
            if(!incomplete_files.empty()) {
                disp+=dcc_put_multiple_lines(lang_get_string(1,(*i).lang,402),eol);
                disp+=dcc_put_multiple_lines(incomplete_files,eol);
            }
        }

        if(!disp.empty()) {
            int ec=0;
            dcc_send_msg((*i).socket,disp,ec,(*i).ssl);
        }
    }
}

/*!
    \brief Called from logic.cpp after rehashing. Applies new configuration to all connected users on DCC/telnet (also BOTNET) connections.
    \author VooDooMan
    \version 1
    \date 2004
    \param who This should contain who caused rehasing (to inform connected users)
*/
void dcc_rehashed(string who)
{
    vector<c_dcc_client>::iterator i;
    char dcc_eol[]="\n";
    char telnet_eol[]="\r\n";
    for(i=dcc_clients.begin(); i!=dcc_clients.end(); ++i) {
        if(TCID_INVALID!=(*i).connect_id)
            continue;
        if((*i).got_passwd!=3)
            continue;
        dcc_loop_get_rights(i);
        if((*i).send)
            continue;
        if((*i).send_from_filesystem)
            continue;
        if((*i).replication_partner)
            continue;
        if((*i).ssl_connection)
            continue;
        if(!(*i).chat)
            continue;
        string eol=(*i).telnet?telnet_eol:dcc_eol;
        string disp=lang_subst(lang_get_string(1,(*i).lang,610),who)+eol;
        int ec=0;
        dcc_send_msg((*i).socket,disp,ec,(*i).ssl);
    }
}

/*!
    \brief Broatcasts the message to all connected users on DCC/telnet connection (used for kill message)
    \author VooDooMan
    \version 2
    \date 2004, 2007-08-20
    \param msg Message
*/
void dcc_broadcast(const string& msg)
{
    map<string,string> subst;
    subst.insert(make_pair("%message;",msg));
    subst.insert(make_pair("%timestamp;",make_nice_time(my_time(NULL))));

    vector<c_dcc_client>::iterator i;
    char dcc_eol[]="\n";
    char telnet_eol[]="\r\n";
    for(i=dcc_clients.begin(); i!=dcc_clients.end(); ++i) {
        if(TCID_INVALID!=(*i).connect_id)
            continue;
        if((*i).got_passwd!=3)
            continue;
        if((*i).send)
            continue;
        if((*i).send_from_filesystem)
            continue;
        if((*i).replication_partner)
            continue;
        if((*i).ssl_connection)
            continue;
        if(!(*i).chat)
            continue;
        string eol=(*i).telnet?telnet_eol:dcc_eol;
        string disp=lang_subst(lang_get_string(1,(*i).lang,773),subst);
        disp+=eol;
        int ec=0;
        dcc_send_msg((*i).socket,disp,ec,(*i).ssl);
        sock_send_cache();
        if(!(*i).is_ssl_client)
            sock_flush_later((*i).socket);
    }
}

/*!
    \brief Broatcasts the message to users that matches mask on DCC/telnet connection
    \author VooDooMan
    \version 2
    \date 2005, 2007-08-20
    \param user_mask User mask to match
    \param msg Message
    \param sender Sending bot, or an empty string for local message
    \param bot_mask Receiving bot mask, or an empty string for local message
*/
void dcc_broadcast(string user_mask, string msg, string sender, string bot_mask)
{
    string nt=make_nice_time(my_time(NULL));

    vector<c_dcc_client>::iterator i;
    char dcc_eol[]="\n";
    char telnet_eol[]="\r\n";
    for(i=dcc_clients.begin(); i!=dcc_clients.end(); ++i) {
        if(TCID_INVALID!=(*i).connect_id)
            continue;
        if((*i).got_passwd!=3)
            continue;
        if((*i).send)
            continue;
        if((*i).send_from_filesystem)
            continue;
        if((*i).replication_partner)
            continue;
        if((*i).ssl_connection)
            continue;
        if(!(*i).chat)
            continue;
        if(match(user_mask.c_str(),(*i).as_in_logic.c_str()))
            continue;
        string eol=(*i).telnet?telnet_eol:dcc_eol;

        // (string user_mask, string msg, string sender, string bot_mask)
        string message;
        message=lang_get_string(1,(*i).lang,693);

        message=lang_subst(message,nt);

        if(sender.empty())
            sender=lang_get_string(1,(*i).lang,694);

        message=lang_subst(message,sender);

        string to=user_mask;
        if(!bot_mask.empty())
            to+=(string)"@"+bot_mask;
        message=lang_subst(message,to);

        message=lang_subst(message,msg);

        string disp=message+eol;
        int ec=0;
        dcc_send_msg((*i).socket,disp,ec,(*i).ssl);
        sock_send_cache();
        if(!(*i).is_ssl_client)
            sock_flush_later((*i).socket);
    }
}

/*!
    \brief Flushes all pending DCC/telnet connection caches
    \author VooDooMan
    \version 1
    \date 2004
*/
void dcc_flush_all()
{
    sock_send_cache();
    vector<c_dcc_client>::iterator i;
    for(i=dcc_clients.begin(); i!=dcc_clients.end(); ++i)
        if(TCID_INVALID==(*i).connect_id && !(*i).is_ssl_client)
            sock_flush_later((*i).socket);
}

/*!
    \brief Closes all DCC/telnet connections
    \author VooDooMan
    \version 1
    \date 2004
*/
void dcc_close_all()
{
    vector<c_dcc_client>::iterator i;
    for(i=dcc_clients.begin(); i!=dcc_clients.end(); ++i) {
        if(TCID_INVALID!=(*i).connect_id) {
            if(!(*i).is_ssl_client)
                sock_close((*i).socket,__FILE__,__LINE__);
            else
                MY_DELETE((*i).ssl);
        }
    }
    dcc_clients.clear();
}

/*!
    \brief Sets logged-in flag for remote bot (for .stat command), as callback from botnet.cpp
    \author VooDooMan
    \param remote Socket of remote bot
*/
void dcc_set_bot_logged_in(c_socket* remote)
{
    vector<c_dcc_client>::iterator i;
    for(i=dcc_clients.begin(); i!=dcc_clients.end(); ++i) {
        if(TCID_INVALID!=(*i).connect_id)
            continue;
        if(*remote==(*i).socket) {
            (*i).got_passwd=3;
            break;
        }
    }
}

/*!
    \brief Updates bind host to listening dynamic servers
    \author VooDooMan
    \date 2005-12-09
    \version 1
    \param irc_host Our host, as reported from irc server
*/
void dcc_update_dynamic_servers(string irc_host)
{
    {
        vector<c_dcc_server>::iterator i1;
        for(i1=dcc_servers.begin(); i1!=dcc_servers.end(); ++i1) {
            if((*i1).dynamic && (*i1).bind_host.compare(irc_host)) { // !=
                in_addr addr4;
                e_resolv_result res=threads_resolve4(irc_host.c_str(),addr4,RT_RESOLVE_SCHEDULE);
                uint32 ip;
                memcpy(&ip,&addr4,sizeof(ip));
                if(RR_OK==res) {
                    if((*i1).dcc_addr!=ip) {
                        if((*i1).active) {
                            sock_close((*i1).socket,__FILE__,__LINE__);
                            (*i1).socket.clear();
                            (*i1).active=false;
                        }
                        int ec=0;

                        char tmp[128+1];
                        tmp[0]=0;
                        my_inet_ntop(AF_INET,&ip,tmp,sizeof(tmp)-1);

                        (*i1).dcc_addr=ip;
                        (*i1).bind_host=irc_host;

                        string comment="DCC server ";
                        comment+=tmp;
                        comment+=":";
                        comment+=tos((*i1).port);
                        if(strlen(tmp)) {
                            (*i1).socket=sock_bind(addr4,(*i1).port,ec,comment.c_str());

                            if((*i1).socket.cmp()==false) {
                                char* errstr=NULL;
                                log_socket(ec,errstr=sock_error(ec),GET_STR_DEBUG_INFO_EX("Error while calling sock_bind()"),&(*i1).socket);
                                MY_DELETE_A(errstr);
                                return;
                            }

                            (*i1).active=true;
                        } else {
                            (*i1).active=false;
                            (*i1).dynamic=true;
                        }

                        (*i1).using_ipv6=false;
                        (*i1).bind_ip=tmp;
                    }
                }
            }
        }
    }
    {
        vector<c_telnet_server>::iterator i1;
        for(i1=telnet_servers.begin(); i1!=telnet_servers.end(); ++i1) {
            if((*i1).dynamic && (*i1).bind_host.compare(irc_host)) { // !=
                in_addr addr4;
                e_resolv_result res=threads_resolve4(irc_host.c_str(),addr4,RT_RESOLVE_SCHEDULE);
                uint32 ip;
                memcpy(&ip,&addr4,sizeof(ip));
                if(RR_OK==res) {
                    if((*i1).telnet_addr!=ip) {
                        if((*i1).active) {
                            sock_close((*i1).socket,__FILE__,__LINE__);
                            (*i1).socket.clear();
                            (*i1).active=false;
                        }
                        int ec=0;

                        char tmp[128+1];
                        tmp[0]=0;
                        my_inet_ntop(AF_INET,&ip,tmp,sizeof(tmp)-1);

                        (*i1).telnet_addr=ip;
                        (*i1).bind_host=irc_host;

                        string comment;
                        if(!(*i1).pure_ssl)
                            comment="Telnet server ";
                        else
                            comment="SSL server ";
                        comment+=tmp;
                        comment+=":";
                        comment+=tos((*i1).port);
                        if(strlen(tmp)) {
                            (*i1).socket=sock_bind(addr4,(*i1).port,ec,comment.c_str());

                            if((*i1).socket.cmp()==false) {
                                char* errstr=NULL;
                                log_socket(ec,errstr=sock_error(ec),GET_STR_DEBUG_INFO_EX("Error while calling sock_bind()"),&(*i1).socket);
                                MY_DELETE_A(errstr);
                                return;
                            }

                            (*i1).active=true;
                        } else {
                            (*i1).active=false;
                            (*i1).dynamic=true;
                        }

                        (*i1).using_ipv6=false;
                        (*i1).bind_ip=tmp;
                    }
                }
            }
        }
    }
}

