/***************************************************************************
                          dcc.h  -  description
                             -------------------
    begin                : Thu Dec 9 2004
    copyright            : (C) 2004-2007 by Marian VooDooMan Meravy
    email                : ghostvoodooman -aT= users (dot) sourceforge.net
 ***************************************************************************/

/* $Id: dcc.h,v 1.51 2008/03/25 12:04:18 ghostvoodooman Exp $ */

/***************************************************************************

VooDoo cIRCle - an IRC (ro)bot
Copyright (C) 2004-2008 by Marian VooDooMan Meravy
  (ghostvoodooman -aT= users (dot) sourceforge.net)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************/

/*!
    \file
    \brief Handles DCC / telnet communication
    \version $Id: dcc.h,v 1.51 2008/03/25 12:04:18 ghostvoodooman Exp $
*/

#ifndef DCC_H__
#define DCC_H__

#if defined(HAVE_CONFIG_H)
#   include <config.h>
#endif

#include <string>
#include <vector>
#include <stack>
#include <list>
#include <map>

#include <time.h>

#include "utils.h"      // because we want to enable mem leak detection

#include "params.h"

#include "logic.h"
#include "sock.h"
#include "filesys.h"
#include "threads.h"

#include "utils/assert.h"

/*!
    \def DCC_NO_SUCH_FILE
    \brief No such file
*/
#define DCC_NO_SUCH_FILE        1
/*!
    \def DCC_ACCESS_DENIED
    \brief Access denied
*/
#define DCC_ACCESS_DENIED       2
/*!
    \def DCC_IO_ERROR
    \brief I/O error while accessing file
*/
#define DCC_IO_ERROR            3
/*!
    \def DCC_AGAIN
    \brief Should be called again (because IP address has not been resolved yet)
*/
#define DCC_AGAIN               4

/*!
    \brief Stores awaiting user on DCC request to process
    \author VooDooMan
    \version 2
    \date 2004
*/
class c_dcc_awaiting_user
{
public:
    bool using_ipv6;                        //!< Are we awaitng IPv6 connection?
    char user_name_as_in_logic[128];        //!< Name of user as in "logic.txt"
    in_addr addr4;                          //!< IPv4 address of remote user
    in_addr6_ addr6;                        //!< IPv6 address of remote user
    e_resolv_result resolv_result;          //!< Result from address resolution
    string dcc_host;                        //!< IPv4 address to resolve, if resolv_result==RR_SCHEDULED
    my_time_t start_time;                      //!< Time of user's initialization

    string nick;                            //!< Nick of user at initaition
    string ident;                           //!< Ident of user at initaition
    string host;                            //!< Host of user at initaition
    string fullname;                        //!< Full name of user at initaition

    c_dcc_awaiting_user()
    {
        memset(user_name_as_in_logic,0,sizeof(user_name_as_in_logic));
        memset(&addr4,0,sizeof(addr4));
        memset(&addr6,0,sizeof(addr6));
        using_ipv6=false;
        start_time=0;
        nick=ident=host=fullname="";

        resolv_result=RR_OK;
        dcc_host="";
    }
};

void dcc_make_server(int ipv, int group, const string& bind_ip, unsigned short bind_port, string server_type);
void dcc_make_telnet_server(int ipv, const string& bind_ip, unsigned short bind_port, bool pure_ssl = false);
string dcc_get_string(string msg, int& dcc_group, int& ip_ver);

string dcc_get_password(string user);

void dcc_partyline_message(string channel, string message);
void dcc_user_pong(string time_, string user, string origin_bot);
void dcc_user_pong(string user, string origin_bot, string name, string number, string comment);

void dcc_notify(string user, int reason, string incomplete_file);

int dcc_send_file(bool from_dcc, string public_name, string nick, int dcc_group);

void dcc_file_has_been_read(string public_name, string nick);

void dcc_loop();

void dcc_rehashed(string who);

void dcc_upgrade_bot_join(string nick, string ident, string host, string fullname, string channel);
void dcc_upgrade_bot_mode(string nick, string ident, string host, string fullname, string channel, string mode, string src);
void dcc_upgrade_bot_nick(string nick, string ident, string host, string fullname, string new_nick);

void dcc_broadcast(string user_mask, string msg, string sender, string bot_mask);

void dcc_set_bot_logged_in(c_socket* remote);

void dcc_update_dynamic_servers(string irc_host);

void dcc_apply(string who);

void irc_quoted_callback(const string& str, const string& nick, const string& host_mask, const string& comment);

void dcc_broadcast(const string& msg);
void dcc_flush_all();
void dcc_close_all();
void dcc_close_servers();
void dcc_send_msg(c_socket& socket, string msg, int& ec, c_ssl* ssl);

#endif
