/***************************************************************************
                          dcc_internal.h  -  description
                             -------------------
    begin                : Thu Dec 9 2004
    copyright            : (C) 2004-2007 by Marian VooDooMan Meravy
    email                : ghostvoodooman -a.T= users (dot) sourceforge.net
 ***************************************************************************/

/* $Id: dcc_internal.h,v 1.16 2008/06/29 17:37:53 ghostvoodooman Exp $ */

/***************************************************************************

VooDoo cIRCle - an IRC (ro)bot
Copyright (C) 2004-2008 by Marian VooDooMan Meravy
  (ghostvoodooman -a.T= users (dot) sourceforge.net)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************/

/*!
    \file
    \brief Handles DCC / telnet communication
    \version $Id: dcc_internal.h,v 1.16 2008/06/29 17:37:53 ghostvoodooman Exp $
*/

#ifndef DCC_INTERNAL_H__
#define DCC_INTERNAL_H__

#if defined(HAVE_CONFIG_H)
#   include <config.h>
#endif

#include <string>
#include <vector>
#include <stack>
#include <list>
#include <map>
#include <algorithm>

#include <stdlib.h>

#include <time.h>
#include <stdio.h>
#if defined(_WIN32) || defined(_WIN64)
#   include <io.h>
#endif
#include <sys/stat.h>

#include "../params.h"
#include "../utils/memleak.h"
#include "../utils/assert.h"
#include "../compat.h"

#include "../utils.h"

#include "../params.h"

#include "../sock.h"
#include "../log.h"
#include "../irc.h"
#include "../lang.h"
#include "../md5.h"
#include "../conf.h"
#include "../botnet.h"
#include "../shared.h"
#include "../filesys.h"
#include "../stats.h"
#include "../ssl.h"
#include "../exec.h"
#include "../compat.h"
#include "../variable.h"
#include "../modules.h"
#include "../threads.h"
#include "../compat.h"
#include "../logic.h"
#include "../eval.h"
#include "../match.h"
#include "../dcc.h"
#include "../irc/c_online_channel.h"
#include "../dcc/c_dcc_request_whois.h"
#include "../dcc/c_user_to_lock_out.h"

using namespace std;

//---------------------------------------------------------------------------
#ifdef __BORLANDC__
#pragma package(smart_init)
#endif

/*!
    \author VooDooMan
    \version 2
    \date 2004
    \brief Stores informations about DCC SEND-ing file
*/
class c_dcc_send_file
{
public:
    bool using_ipv6;                                //!< Is this connection an IPv6 connection?
    in_addr remote_ip;                              //!< IPv4 address of remote host
    in_addr6_ remote_ip6;                           //!< IPv6 address of remote host
    string file_name;                               //!< Internal name of file
    my_time_t start_time;                              //!< Timestamp of establishing connection
    MY_LONG_LONG size;                              //!< Size of the file
    string user_name_as_in_logic;                   //!< Name of the user form "logic.txt"
    string public_name;                             //!< Public name of the file
    MY_LONG_LONG resume_pos;

    c_dcc_send_file()
    {
        using_ipv6=false;
#if defined(_WIN32) || defined(_WIN64)
        remote_ip.S_un.S_addr=0;
#else
        remote_ip.s_addr=0;
#endif

        memset(&remote_ip6,0,sizeof(remote_ip6));
        file_name="";
        start_time=0;
        size=0;
        user_name_as_in_logic="";
        public_name="";
        resume_pos=0;
    }
};

/*!
    \brief Stores DCC state of user
    \author VooDooMan
    \version 1
    \date 2004
*/
class c_dcc_client
{
public:
    c_socket server_socket;                         //!< Socket handle of DCC server connection
    threaded_connect_id connect_id;                 //!< Thread's connect ID

    my_time_t session_start;                           //!< Timestamp of while when socket was open (for pre-auth timeout)

    log_client_id_t client_id;                      //!< ID of client (for logs)

    bool is_ssl_client;                             //!< true for SSL server

    string remote;                                  //!< Remote host/IP
    string local;                                   //!< Local IP
    unsigned short local_port;                      //!< Local port (or zero if it is outgoing)
    unsigned short remote_port;                     //!< Remote port (or zero if it is incomming)

    string nick;                                    //!< Nick of user at initiation of connection
    string ident;                                   //!< Ident of user at initiation of connection
    string host;                                    //!< Host of user at initiation of connection
    string fullname;                                //!< Full name of user at initiation of connection

    bool chat;                                      //!< Is this connection a DCC-CHAT-type connection?
    bool send;                                      //!< Is this connection a DCC-SEND-type connection?
    bool send_from_filesystem;                      //!< Is this connection a DCC-SEND-type connection? (sending file to remote user)
    bool telnet;                                    //!< Is this connection a telnet server?
    bool replication_partner;                       //!< Is this a replication partner?
    c_socket socket;                                //!< Socket handle of DCC connection

    bool got_user_name;                             //!< If we know user name as in "logic.txt", will be set to false on requesting nick name if "dcc_always_want_nick" from "conf.txt" file is non-zero.

    bool ssl_connection;                            //!< Is this a SSL connection?
    c_ssl* ssl;                                     //!< If there is a SSL connection, descriptor is here

    c_user u;                                       //!< User's descriptor
    vector<c_channel> chs;                          //!< User's channels

    bool access_to_can_send_all_users;              //!< Can this user send message / file to all users?
    bool access_grant_can_send_all_users;           //!< Can this user grant right to send message / file to all users?

    MY_LONG_LONG resume_pos;

    c_dcc_send_file send_file;                      //!< File that is sending

    string hostname;                                //!< Contains host of user
    string as_in_logic;                             //!< Contains name of user (as in logic.txt file)
    string supposed_to_be;                          //!< Contains name of user (as in logic.txt file) - set, before "as_in_logic" is cleaned

    //---

    string in_buff;                                 //!< Contains buffer for reading from DCC connection
    my_time_t last_input;                              //!< Timestamp of last socket read (non-zero bytes)

    string lang;                                    //!< Contains code of language which to speak with user

    int got_passwd;                                 //!< Has user enterred password yet?

    my_time_t last_filesys_lookup;                     //!< Timestamp of last filesystem lookup for notifies

    vector<string> access_to_users;                 //!< List of users that user has access to
    vector<string> access_to_users_also_terminators;    //!< List of users that user has access to (also terminator users)
    vector<string> access_to_groups;                //!< List of groups that user has access to
    vector<string> access_grant_groups;             //!< List of groups that user can add/exclude users to

    vector<string> access_to_procs;
    vector<string> access_to_proc_groups;           //!< List of groups of procedures that user has access to

    vector<string> access_grant_proc;               //!< List of procedures that user can grant access to
    vector<string> access_grant_proc_groups;        //!< List of groups of procedures that user can grant access to

    vector<string> access_usage_proc;               //!< List of procedures that user can add/exclude events to
    vector<string> access_usage_proc_groups;        //!< List of procedure groups that user can add/exclude events to

    vector<string> access_to_channel;               //!< List of channels that user has access to
    vector<string> access_grant_channel;            //!< List of channels that user can add/exclude users to/from
    bool access_grant_partyline;                    //!< Can user grant other user access to partyline?
    bool access_to_plususer;                        //!< Can this user (or bot) create new user?
    bool access_grant_plususer;                     //!< Can this user (or bot) grant creatoin new user to other user?

    bool access_to_plusproc;                        //!< Can this user (or bot) create new procedure?
    bool access_grant_plusproc;                     //!< Can this user (or bot) grant creatoin new procedure to other user?

    bool access_to_replication;                     //!< Can this user modify replication parameters?
    bool access_grant_replication;                  //!< Can this user grant access to modify replication parameters?

    bool access_to_restart;                         //!< Can this user use .restart command?
    bool access_grant_restart;                      //!< Can this user grant access to .restart command to other user?
    bool access_to_die;                             //!< Can this user use .die command?
    bool access_grant_die;                          //!< Can this user grant access to .die command to other user?

    bool access_to_filesystem;                      //!< Can this user use .filesystem command?
    bool access_grant_filesystem;                   //!< Can this user grant access to .filesystem command to other user?

    bool access_to_can_send_unknown_users;          //!< Can this user send message / file to unknown users?
    bool access_grant_can_send_unknown_users;       //!< Can this user grant right to send message / file to unknown?

    bool access_to_upgrade;                         //!< Has user access to .upgrade command?
    bool access_grant_upgrade;                      //!< Can this user grant accesst to .upgrade command to other user?

    bool access_to_apply;                           //!< Has user access to .apply command?
    bool access_grant_apply;                        //!< Can this user grant accesst to .apply command to other user?

    c_flood partyline_msg_flood;                    //!< This user: message flood on partylien channels
    vector<c_flood_history> msg_flood_hist;         //!< History for partline flood detection
    my_time_t last_flood;                              //!< Timestamp of last flood warning message sent

    c_flood user_partyline_msg_flood;               //!< Current-edited user: message flood on partylien channels

    int new_user;                                   //!< Number of state in adding of new user
    string user_name;                               //!< Name of new user
    vector<string> user_hosts;                      //!< List of hosts of new user
    vector<c_dcc_host> user_dcc_hosts;              //!< List of DCC hosts of new user
    c_dcc_host user_dcc_host;                       //!< Current-edited DCC host of new user
    vector<string> user_fullnames;                  //!< List of full names of new user
    bool user_access_to_partyline;                  //!< Has new user access to partyline?
    bool user_access_grant_partyline;               //!< Has new user access to grant access to partyline to other user?
    vector<string> user_access_to_groups;           //!< List of groups that new user has access to
    vector<string> user_access_grant_groups;        //!< List of groups that new user can add/exclude users to
    vector<string> user_access_to_proc_groups;      //!< List of groups of procedures that new user has access to
    vector<string> user_access_grant_proc_groups;   //!< List of groups of procedures that new user has access to
    vector<string> user_access_usage_proc_groups;   //!< List of groups of procedures that new user can assign events to
    vector<string> user_access_to_channel;          //!< List of channels that new user has access to
    vector<string> user_access_grant_channel;       //!< List of channels that new user can add/exclude users to/from
    c_channel user_channel;                         //!< Current-edited channel configuration for new user
    vector<c_channel> user_channels;                //!< List of channels for new user
    int num_not_configured_channels;                //!< Number of non-configured channels for new user
    vector<string> user_member_of;                  //!< List of groups which is new user member of
    bool user_access_to_plususer;                   //!< Can current-edited user (or bot) create new user?
    bool user_access_grant_plususer;                //!< Can current-edite user (or bot) grant creatoin new user to other user?

    bool user_access_to_plusproc;                   //!< Can current-edited user (or bot) create new procedure?
    bool user_access_grant_plusproc;                //!< Can current-edite user (or bot) grant creatoin new procedure to other user?

    bool user_access_to_replication;                //!< Can current-edited user modify replication parameters?
    bool user_access_grant_replication;             //!< Can current-edited user grant access to modify replication parameters?

    bool user_access_to_can_send_unknown_users;     //!< Can current-edited user send message / file to unknown?
    bool user_access_grant_can_send_unknown_users;  //!< Can current-edited user grant right to send message / file to unknown?

    map<string,string> user_meta;                   //!< Meta data for current-edited user

    int edit_user;                                  //!< Number of state in editting of user
    int edit_channel_num;                           //!< Number of channel which is edited for edited user
    bool edit_user_from_template;                   //!< Is this a new user from template?

    int edit_pass;                                  //!< Number of state: .pass command
    string edit_pasc_user;                          //!< Name of user

    int filesystem;                                 //!< Number of state in managing of the filesystem
    tm filesystem_expiration;                       //!< Current edited expiration of object in the filesystem

    int msg;                                        //!< Number of state in send message
    string msg_content;                             //!< Stores content of the message
    my_time_t msg_expiration;                          //!< Expiration of the message

    int private_edit;                               //!< Context state of .private command
    c_private private_;                             //!< Current-edited private query messages

    int replication;                                //!< Context state of .replication command
    bool b_repl_user;                               //!< Are we editing replication for user?
    c_user repl_user;                               //!< Current-edited replication properties if we are editing an user
    bool b_repl_user_template;                      //!< Are we editing replication for user template?
    c_user repl_user_template;                      //!< Current-edited replication properties if we are editing an user template
    bool b_repl_chandef;                            //!< Are we editing replication for channel definitions?
    c_channel_def repl_chandef;                     //!< Current-edited replication properties if we are editing channel definitions
    bool b_repl_proc;                               //!< Are we editing replication for procedure?
    c_procedure repl_proc;                          //!< Current-edited replication properties if we are editing a procedure
    bool b_repl_func;                               //!< Are we editing replication for function?
    c_function* repl_func;                          //!< Current-edited replication properties if we are editing a function
    bool b_repl_private;                            //!< Are we editing replication for private query events?
    c_private repl_private;                         //!< Current-edited replication properties if we are editing private query events
    c_replication repl_add;                         //!< Current-edited object's new replication parameters

    int terminator;                                 //!< Context state of .terminator command
    string terminator_user;                         //!< Name of current-edited user in .terminator command

    int language;                                   //!< Context state of .lang command

    int dynbans;                                    //!< Context state of .dynbans command
    string dynbanc_channel;                         //!< Name of channel for .dynbans command

    /*!
        \brief Stores data about file from the filesystem
        \author VooDooMan
        \version 1
        \date 2004
    */
    struct c_fs
    {
        string name;                                //!< Name of the file
        c_file file;                                //!< Properties of current-edited file

        c_access new_access;                        //!< Current-edited access rights

        c_fs()
        {
            name="";
            file.clear();
            new_access.clear();
        }
    } fs;                                           //!< Current-edited file from the filesystem

    int edit_proc;                                  //!< Number of procedure-editing state
    string proc_name;                               //!< Name of current-edited procedure
    string proc_to_delete;                          //!< Name and full declaration of procedure to MY_DELETE((after replace of procedure by .+proc to existing procedure))
    string proc_decl;                               //!< Declaration of current-edite procedure
    bool plus_proc;                                 //!< Are we processing a ".+proc" command?
    list<string> proc_commands;                     //!< List of commands for current-edited procedure
    int edit_proc_addbefore_line;                   //!< Number of line of current-edited procedure to add before (-1 for none)
    int edit_proc_addafter_line;                    //!< Number of line of current-edited procedure to add after (-1 for none)
    int edit_proc_replace_line;                     //!< Number of line of current-edited procedure to replace (-1 for none)
    string proc_error;                              //!< Syntax error string in current-edited procedure
    vector<string> proc_groups;                     //!< List of groups which is current-edited procedure member of
    bool it_is_func;                                //!< true for function, false for procedure
    vector<c_replication> proc_repl;                //!< List of replications
    vector<string> remote_call;                     //!< List of bots with remote call privilege

    int edit_rproc;                                 //!< Number of remote procedure callability editing state
    vector<c_rproc> rproc;                          //!< List of remote bots that can call this current-editing procedure remotely (BOTNET)

    bool access_to_backup;                          //!< Can this user use .backup command?
    bool access_grant_backup;                       //!< Can this user grant/deny access to .backup command to other user?
    bool access_to_rehash;                          //!< Can this user use .rehash command?
    bool access_grant_rehash;                       //!< Can this user grant/deny access to .rehash command to other user?

    bool user_access_to_backup;                     //!< Can edited/new user use .backup command?
    bool user_access_grant_backup;                  //!< Can edited/new user grant/deny access to .backup command to other user?
    bool user_access_to_rehash;                     //!< Can edited/new user use .rehash command?
    bool user_access_grant_rehash;                  //!< Can edited/new user grant/deny access to .rehash command to other user?

    bool user_access_to_restart;                    //!< Can edited/new user use .restart command?
    bool user_access_grant_restart;                 //!< Can edited/new user grant/deny access to .restart command to other user?
    bool user_access_to_die;                        //!< Can edited/new user use .die command?
    bool user_access_grant_die;                     //!< Can edited/new user grant/deny access to .die command to other user?

    bool user_access_to_filesystem;                 //!< Can edited/new user use .filesystem command?
    bool user_access_grant_filesystem;              //!< Can edited/new user grant/deny access to .filesystem command to other user?

    bool user_access_to_upgrade;                    //!< Has user access to .upgrade command?
    bool user_access_grant_upgrade;                 //!< Can this user grant accesst to .upgrade command to other user?

    bool user_access_to_apply;                      //!< Has user access to .apply command?
    bool user_access_grant_apply;                   //!< Can this user grant accesst to .apply command to other user?

    bool access_to_private;                         //!< Has user access to private query definitions?
    bool access_grant_private;                      //!< Has user right to grant to other user access to private query definitions?
    bool user_access_to_private;                    //!< Has current-edited user access to private query definitions?
    bool user_access_grant_private;                 //!< Has current-edited user right to grant to other user access to private query definitions?

    c_dynamic_rule user_channel_allow_dynamic;      //!< Current-edited user's current-edited dynamic rule

    int edit_chan;                                  //!< Number of state of editing channel definitions
    vector<string> access_to_chan_defs;             //!< List of channels that user has access to their definitions
    c_channel_def user_channel_def;                 //!< Current-edited channel definition

    char telnet_buffer[5*1024];                     //!< Input buffer for telnet connection
    size_t telnet_buffer_pos;                       //!< Position in input buffer for telnet connection
    bool telnet_botnet_called;                      //!< Have we called botnet function to process telnet input buffer?

    string partyline_channel;                       //!< Name of partyline channel that user is on, or an empty string for none

    bool user_access_to_can_send_all_users;         //!< Can current-edited user send message / file to all users?
    bool user_access_grant_can_send_all_users;      //!< Can current-edited user grant right to send message / file to all users?

    my_time_t connect_time;                            //!< Time when client has connected to us

    /*!
        \brief Stores data about temporary file being uploaded via DCC SEND
        \author VooDooMan
        \version 1
        \date 2004
    */
    struct c_dcc_client_send
    {
        char tmp_file_name[1024];               //!< Internal name of file
        FILE* f;                                //!< File handle
        MY_LONG_LONG size;                      //!< Size of the file
        MY_LONG_LONG received;                  //!< Number of bytes that has been received
        my_time_t last_packet;                     //!< Timestamp of last packet that has been received
        bool response_sent;                     //!< Has we sent response? (4 bytes of received size)
        string original_name;                   //!< Original name of the file on the remote side
        bool is_file_server_proto;              //!< Is this connection a DCC file server protocol?
    } dcc_send;                                 //!< Stores data about temporary file being uploaded via DCC SEND

    c_dcc_client()
    {
        local=remote="";
        local_port=remote_port=0;
        dcc_send.f=NULL;
        memset(dcc_send.tmp_file_name,0,sizeof(dcc_send.tmp_file_name));
        dcc_send.size=0;
        dcc_send.received=0;
        dcc_send.last_packet=0;
        dcc_send.response_sent=false;
        dcc_send.original_name="";

        user_access_to_can_send_all_users=false;
        user_access_grant_can_send_all_users=false;

        user_access_to_can_send_unknown_users=false;
        user_access_grant_can_send_unknown_users=false;

        access_to_can_send_all_users=false;
        access_grant_can_send_all_users=false;

        access_to_can_send_unknown_users=false;
        access_grant_can_send_unknown_users=false;

        access_to_can_send_all_users=false;
        access_grant_can_send_all_users=false;

        replication=0;
        b_repl_user=false;
        b_repl_user_template=false;
        b_repl_chandef=false;
        b_repl_proc=false;
        b_repl_private=false;

        private_edit=0;
        private_.clear();

        send_from_filesystem=false;

        filesystem=0;

        resume_pos=0;

        msg=0;
        msg_content="";
        msg_expiration=0;

        language=0;

        dcc_send.is_file_server_proto=false;

        server_socket.clear();
        chat=false;
        send=false;
        telnet=false;
        replication_partner=false;
        ssl_connection=false;
        socket.clear();
        hostname="";
        as_in_logic="";
        in_buff="";
        last_input=my_time(NULL);
        lang="";
        got_passwd=0;
        access_to_users.clear();
        access_to_users_also_terminators.clear();
        access_to_groups.clear();
        access_grant_groups.clear();
        access_to_proc_groups.clear();
        access_grant_proc_groups.clear();
        access_usage_proc.clear();
        access_usage_proc_groups.clear();
        access_grant_proc.clear();
        access_to_channel.clear();
        access_grant_channel.clear();
        access_grant_partyline=false;

        access_to_procs.clear();
        access_to_proc_groups.clear();

        access_grant_proc.clear();
        access_grant_proc_groups.clear();

        access_usage_proc.clear();
        access_usage_proc_groups.clear();

        access_to_plususer=false;
        access_grant_plususer=false;
        access_to_plusproc=false;
        access_grant_plusproc=false;
        access_to_replication=false;
        access_grant_replication=false;
        access_to_restart=false;
        access_grant_restart=false;
        access_to_die=false;
        access_grant_die=false;
        access_to_filesystem=false;
        access_grant_filesystem=false;
        access_to_upgrade=access_grant_upgrade=false;
        access_to_apply=access_grant_apply=false;
        user_access_to_upgrade=user_access_grant_upgrade=false;
        user_access_to_apply=user_access_grant_apply=false;
        user_access_to_filesystem=false;
        user_access_grant_filesystem=false;
        access_to_private=false;
        access_grant_private=false;
        user_access_to_private=false;
        user_access_grant_private=false;
        msg_flood_hist.clear();
        new_user=0;
        user_name="";
        user_hosts.clear();
        user_dcc_hosts.clear();
        user_fullnames.clear();
        user_meta.clear();
        user_access_to_partyline=false;
        user_access_grant_partyline=false;
        user_access_to_groups.clear();
        user_access_grant_groups.clear();
        user_access_to_proc_groups.clear();
        user_access_grant_proc_groups.clear();
        user_access_usage_proc_groups.clear();
        user_access_to_channel.clear();
        user_access_grant_channel.clear();
        user_channels.clear();
        user_channel.clear();
        num_not_configured_channels=0;
        user_member_of.clear();
        user_access_to_plususer=false;
        user_access_grant_plususer=false;
        user_access_to_plusproc=false;
        user_access_grant_plusproc=false;
        user_access_to_replication=false;
        user_access_grant_replication=false;
        edit_user=0;
        edit_channel_num=0;
        edit_proc=0;
        it_is_func=false;
        plus_proc=false;
        edit_pass=0;
        proc_name="";
        proc_to_delete="";
        proc_decl="";
        proc_commands.clear();
        edit_proc_addbefore_line=0;
        edit_proc_addafter_line=0;
        edit_proc_replace_line=0;
        proc_error="";
        proc_groups.clear();
        access_to_backup=false;
        access_grant_backup=false;
        access_to_rehash=false;
        access_grant_rehash=false;
        user_access_to_backup=false;
        user_access_grant_backup=false;
        user_access_to_rehash=false;
        user_access_grant_rehash=false;
        user_access_to_restart=false;
        user_access_grant_restart=false;
        user_access_to_die=false;
        user_access_grant_die=false;
        edit_chan=false;
        access_to_chan_defs.clear();
        memset(telnet_buffer,0,sizeof(telnet_buffer));
        telnet_buffer_pos=0;
        telnet_botnet_called=false;
        partyline_channel="";

        last_filesys_lookup=0;

        terminator=0;

        nick=ident=host=fullname="";

        got_user_name=false;

        edit_user_from_template=false;

        connect_id=TCID_INVALID;

        last_flood=0;

        got_passwd=
            new_user=
            edit_user=
            edit_channel_num=
            edit_pass=
            filesystem=
            msg=
            private_edit=
            replication=
            terminator=
            language=
            dynbans=
            edit_proc=
            edit_rproc=
            edit_chan=
            0;
        is_ssl_client=false;

        ssl=NULL;
        repl_func=NULL;

        dcc_send.f=NULL;

        session_start=0;

        connect_time=0;
    }

    ~c_dcc_client()
    {
        if(dcc_send.f)
            fclose(dcc_send.f);
        dcc_send.f=NULL;

        // overwrite (potentially) sensitive data before it gets released to the system
        for(size_t i1=0; i1<in_buff.length(); ++i1)
            in_buff[i1]='X';
        memset(telnet_buffer,0,sizeof(telnet_buffer));
        //
    }

    void destroy()
    {
        // overwrite (potentially) sensitive data before it gets released to the system
        for(size_t i1=0; i1<in_buff.length(); ++i1)
            in_buff[i1]='X';
        memset(telnet_buffer,0,sizeof(telnet_buffer));
        //
        MY_DELETE(repl_func);
        repl_func=NULL;
    }
};

/*!
    \brief Stores DCC server's parameters
    \author VooDooMan
    \version 2
    \date 2004
*/
class c_dcc_server
{
public:
    int group;                  //!< Number of "dcc_group"
    bool using_ipv6;            //!< Are we listening for IPv6 connections?
    uint32 dcc_addr;            //!< 4-byte IPv4 address of bound socket, or -1 for IPv6 address
    in_addr6_ dcc_addr6;        //!< 16-byte IPv6 address of bound socket
    unsigned short port;        //!< Port number of bound socket
    string bind_ip;             //!< IPv4/IPv6 address of bound socket (string)
    c_socket socket;            //!< Handle of socket
    string server_type;         //!< server_type that will be sent

    bool active;                //!< Is this server active?
    bool dynamic;               //!< Has this server dynamically assigned IP?
    string bind_host;           //!< DNS name/IP we try to bind to

    c_dcc_server()
    {
        group=0;
        using_ipv6=false;
        dcc_addr=0;
        memset(&dcc_addr6,0,sizeof(dcc_addr6));
        port=0;
        bind_ip="";
        socket.clear();

        active=false;
        dynamic=false;
        bind_host="";
    }
};

/*!
    \brief Stores telnet server's parameters
    \author VooDooMan
    \version 2
    \date 2004, 2006-07-19
*/
class c_telnet_server
{
public:
    uint32 telnet_addr;         //!< 4-byte IPv4 address of bound socket, or -1 for IPv6 bound address
    unsigned short port;        //!< Port number of bound socket
    string bind_ip;             //!< IP address of bound socket (string)
    c_socket socket;            //!< Handle of socket

    bool using_ipv6;            //!< Is this an IPv6 socket?

    uint32 addr;                //!< 4-byte IPv4 address of bound socket, or -1 for IPv6 address

    bool active;                //!< Is this server active?
    bool dynamic;               //!< Has this server dynamically assigned IP?
    string bind_host;           //!< DNS name/IP we try to bind to

    bool pure_ssl;              //!< true for ssl server

    c_telnet_server()
    {
        telnet_addr=0;
        port=0;
        bind_ip="";
        using_ipv6=false;
        socket.clear();

        active=false;
        dynamic=false;
        bind_host="";

        pure_ssl=false;
    }
};

/*!
    \brief Stores temporary resume information for DCC
    \author VooDooMan
    \version 2
    \date 2004
*/
class c_dcc_client_to_resume
{
public:
    bool using_ipv6;                            //!< Is this connection an IPv6 connection?
    string ip_addr4;                            //!< IPv4 address of remote host
    unsigned int ip_4;                          //!< IPv4 address of remote host
    string ip_addr6;                            //!< IPv6 address of remote host
    in_addr6_ ip_6;                             //!< IPv6 address of remote host
    unsigned short port;                        //!< Port
    string user_name;                           //!< Usaer name from "logic.txt"
    string nick;                                //!< Nick name
    string file_name;                           //!< Name of file
    MY_LONG_LONG file_size;                     //!< Size of file
    MY_LONG_LONG resume_position;               //!< Position to resume from

    my_time_t start_time;                          //!< Start time

    string comment;                             //!< Full string of request

    c_dcc_client_to_resume()
    {
        ip_addr4="";
        ip_4=0;
        ip_addr6="";
        memset(&ip_6,0,sizeof(ip_6));
        port=0;
        user_name="";
        nick="";
        file_name="";
        file_size=0;
        resume_position=0;
        start_time=0;
        comment="";
    }
};

/*!
    \brief Stores information about usage of DCC requests for users. Used for flood protection
    \author VooDooMan
    \version 1
    \date 2004
*/
class c_user_to_lag
{
public:
    string user_name;                           //!< Name of user from "logic.txt"
    int cnt;                                    //!< Number of DCC requests
    my_time_t last_try;                            //!< Timestamp of last try from user
    bool notified;                              //!< Has been user notified about its flood?

    c_user_to_lag() {
        user_name="";
        cnt=0;
        last_try=0;
        notified=false;
    }
};

void dcc_check_for_filesystem(string user_name);
void dcc_send_msg(c_socket& socket, string msg, int& ec, c_ssl* ssl);
string dcc_get_line(string& buff, bool keep = false);
bool dcc_can_unassign(vector<c_dcc_client>::iterator i, string& eol, string proc_name);
void dcc_cleanup_same_ip();
void dcc_check_limit();
void dcc_set_password(string user, string md5pass);
void dcc_loop_get_rights(vector<c_dcc_client>::iterator i);
bool dcc_is_flood(vector<c_dcc_client>::iterator i, const string& eol);
void dcc_loop_new_user(vector<c_dcc_client>::iterator i, string& eol);
void dcc_loop_edit_user1(vector<c_dcc_client>::iterator i, string& eol);
void dcc_loop_edit_user2(vector<c_dcc_client>::iterator i, string& eol);
void dcc_loop_edit_rproc(vector<c_dcc_client>::iterator i, string& eol);
void dcc_loop_edit_proc(vector<c_dcc_client>::iterator i, string& eol);
void dcc_loop_edit_chan(vector<c_dcc_client>::iterator i, string& eol);
string dcc_loop_filesystem_disp_rights(string lang, c_access& access, string& eol);
void dcc_loop_msg(vector<c_dcc_client>::iterator i, string& eol);
void dcc_loop_private(vector<c_dcc_client>::iterator i, string& eol);
void dcc_loop_replication(vector<c_dcc_client>::iterator i, string& eol);
void dcc_loop_terminator(vector<c_dcc_client>::iterator i, string& eol);
void dcc_loop_language(vector<c_dcc_client>::iterator i, string& eol);
void dcc_loop_edit_dynbans(vector<c_dcc_client>::iterator i, string& eol);
void dcc_loop_edit_pass(vector<c_dcc_client>::iterator i, string& eol);
void dcc_upgrade(vector<c_dcc_client>::iterator i, string& eol, string new_file_name);
void dcc_loop_filesystem(vector<c_dcc_client>::iterator i, string& eol);
void dcc_broadcast(const string& msg);
my_time_t dcc_push_bad_login(const string& remote);
void dcc_pop_bad_login(const string& remote, my_time_t timestamp);

#endif

