/***************************************************************************
                          dcc_loop_filesystem.cpp  -  description
                             -------------------
    begin                : Thu Dec 9 2004
    copyright            : (C) 2004-2007 by Marian VooDooMan Meravy
    email                : ghostvoodooman -a.T= users (dot) sourceforge.net
 ***************************************************************************/

/* $Id: dcc_loop_filesystem.cpp,v 1.16 2008/07/02 09:35:48 ghostvoodooman Exp $ */

/***************************************************************************

VooDoo cIRCle - an IRC (ro)bot
Copyright (C) 2004-2008 by Marian VooDooMan Meravy
  (ghostvoodooman -a.T= users (dot) sourceforge.net)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************/

/*!
    \file
    \brief Handles DCC / telnet communication
    \version $Id: dcc_loop_filesystem.cpp,v 1.16 2008/07/02 09:35:48 ghostvoodooman Exp $
*/

#if defined(HAVE_CONFIG_H)
#   include <config.h>
#endif

#include "../utils/memleak.h"
#include "../utils/assert.h"
#include "../utils.h"
#include "../params.h"

#include "dcc_internal.h"

extern deque<c_user> r_terminators;
extern deque<c_user> r_user_templates;
extern deque<c_user> r_users;

#include "../utils/precomp.h"

/*!
    \brief Returns access rights to file
    \author VooDooMan
    \version 1
    \date 2004
    \param lang Language code
    \param eol End-of-line string
    \param access Access descriptor
    \return User-friendly list of access rights
*/
string dcc_loop_filesystem_disp_rights(string lang, c_access& access, string& eol)
{
    string disp;

    if(access.all_users)
        disp=lang_get_string(1,lang,294)+eol;
    else
        if(!access.all_on_channel.empty()) {
            lang_subst(lang_get_string(1,lang,295),(string)"*"+access.all_on_channel);
            if(access.also_unknown)
                disp+=(string)" "+lang_get_string(1,lang,474);
            disp+=eol;
        } else
            disp+=lang_subst(lang_get_string(1,lang,295),access.user_name)+eol;
    disp+="   ";
    bool plus=false;
    if(access.owner) {
        disp+=lang_get_string(1,lang,296);
        plus=true;
    }
    if(access.read) {
        if(plus)
            disp+=lang_get_string(1,lang,301);
        disp+=lang_get_string(1,lang,297);
        plus=true;
    }
    if(access.del) {
        if(plus)
            disp+=lang_get_string(1,lang,301);
        disp+=lang_get_string(1,lang,298);
        plus=true;
    }
    if(access.notify_owner) {
        if(plus)
            disp+=lang_get_string(1,lang,301);
        disp+=lang_get_string(1,lang,299);
        plus=true;
    }
    if(access.notify_user) {
        if(plus)
            disp+=lang_get_string(1,lang,301);
        disp+=lang_get_string(1,lang,300);
        plus=true;
    }
    if(access.secure) {
        if(plus)
            disp+=lang_get_string(1,lang,301);
        disp+=lang_get_string(1,lang,330);
        plus=true;
    }
    if(!plus) {
        disp+=lang_get_string(1,lang,322);
    }
    disp+=eol;

    return disp;
}

/*!
    \brief Should be called in a loop to process filesystem commands in DCC CHAT/telnet connections
    \author VooDooMan
    \version 1
    \date 2004
    \param i Descriptor of remote connection
    \param eol End-of-line string
*/
void dcc_loop_filesystem(vector<c_dcc_client>::iterator i, string& eol)
{
    if((*i).filesystem==0)
        return;

    string disp;
    int ec=0;
    bool got=false;

    if((*i).filesystem==1) {
        vector<c_dcc_notify> notify;
        filesys_dcc_check_for_notifies((*i).as_in_logic,(*i).u,(*i).chs,"",notify,true,(*i).lang,eol);
        if(!notify.empty()) {
            filesys_dcc_drop_notifies((*i).as_in_logic,false);

            string files;
            string msgs;
            string incomplete_files;

            vector<c_dcc_notify>::iterator i1;
            for(i1=notify.begin(); i1!=notify.end(); ++i1) {
                if((*i1).unpublished) {
                    if(!files.empty())
                        files+=", ";
                    files+="\"";
                    files+=(*i1).name;
                    files+="\"";
                }
                if((*i1).incomplete) {
                    if(!incomplete_files.empty())
                        incomplete_files+=", ";
                    incomplete_files+="\"";
                    incomplete_files+=(*i1).name;
                    incomplete_files+="\"";
                }
                if((*i1).notify_owner || (*i1).notify_user || (*i1).secure_notify_owner || (*i1).secure_notify_user)
                    msgs+=(*i1).message;
            }

            disp=lang_get_string(1,(*i).lang,279)+eol;
            disp+=files;
            disp+=eol;
            disp+=lang_get_string(1,(*i).lang,401)+eol;
            disp+=incomplete_files;
            disp+=eol;
            disp+=lang_get_string(1,(*i).lang,324)+eol;
            disp+=lang_get_string(1,(*i).lang,280)+eol;
            disp+=lang_get_string(1,(*i).lang,281)+eol;
            disp+=msgs;
            disp+=lang_get_string(1,(*i).lang,469)+eol;
            ec=0;
            dcc_send_msg((*i).socket,disp,ec,(*i).ssl);
        }

        (*i).filesystem=2;
        return;
    }

    if((*i).filesystem==2) {
        disp=lang_get_string(1,(*i).lang,690)+eol;
        ec=0;
        dcc_send_msg((*i).socket,disp,ec,(*i).ssl);

        (*i).filesystem=1000;
    }

    string p=dcc_get_line((*i).in_buff);

    string c=p;
    string cmd, param1, param2;
    int pos=0;
    for(unsigned int i1=0; i1<c.length(); ++i1) {
        if((pos==0 || pos==1 || pos==2) && c[i1]==0x20) {
            pos++;
            continue;
        }
        if(pos==0)
            cmd+=c[i1];
        if(pos==1)
            param1+=c[i1];
        if(pos==2)
            param2+=c[i1];
    }

    if((*i).filesystem==1000) {
        bool got_cmd=false;
        if(!cmd.compare(".editfile")) {
            got_cmd=true;
            if((*i).access_to_filesystem) {
                vector<c_file> files;
                filesys_dcc_filelist((*i).as_in_logic,files,true);

                string files_;

                got=false;
                string fn;

                vector<c_file>::iterator i1;
                for(i1=files.begin(); i1!=files.end(); ++i1) {
                        if(!(*i1).public_name.compare(param1)) {
                            got=true;
                            fn=(*i1).public_name;
                            break;
                        }
                        if(!files_.empty())
                            files_+=", ";
                        files_+="\"";
                        files_+=(*i1).public_name;
                        files_+="\"";
                }

                if(!got) {
                    disp=lang_get_string(1,(*i).lang,279)+eol;
                    disp+=files_;
                    disp+=eol;
                    disp+=lang_get_string(1,(*i).lang,324)+eol;
                    disp+=lang_get_string(1,(*i).lang,280)+eol;
                    disp+=lang_get_string(1,(*i).lang,281)+eol;
                    disp+=lang_get_string(1,(*i).lang,282)+eol;
                    ec=0;
                    dcc_send_msg((*i).socket,disp,ec,(*i).ssl);
                } else {
                    (*i).fs.file.clear();
                    if(filesys_dcc_get_file((*i).as_in_logic,fn,(*i).fs.file)) {
                        disp.clear();
                        vector<c_access>::iterator i1;
                        for(i1=(*i).fs.file.access.begin(); i1!=(*i).fs.file.access.end(); ++i1)
                            disp+=dcc_loop_filesystem_disp_rights((*i).lang,*i1,eol);
                        ec=0;
                        dcc_send_msg((*i).socket,disp,ec,(*i).ssl);

                        (*i).fs.name=fn;
                        (*i).filesystem=2000;
                    } else {
                        disp=lang_get_string(1,(*i).lang,291)+eol;
                        ec=0;
                        dcc_send_msg((*i).socket,disp,ec,(*i).ssl);
                        (*i).filesystem=1000;
                    }
                    return;
                }
            } else {
                disp=lang_get_string(1,(*i).lang,275)+eol;
                ec=0;
                dcc_send_msg((*i).socket,disp,ec,(*i).ssl);
            }
        }
        if(!cmd.compare(".getfile")) {
            disp.clear();
            got_cmd=true;
            int dcc_group=stoi(param2);
            int res=dcc_send_file(true,param1,irc_get_nick((*i).as_in_logic),dcc_group);
            switch(res) {
                case DCC_NO_SUCH_FILE:
                    disp=lang_get_string(1,"en",334);
                    break;
                case DCC_ACCESS_DENIED:
                    disp=lang_get_string(1,"en",335);
                    break;
                case DCC_IO_ERROR:
                    disp=lang_get_string(1,"en",336);
                    break;
            }
            if(!disp.empty()) {
                disp+=eol;
                ec=0;
                dcc_send_msg((*i).socket,disp,ec,(*i).ssl);
            }
            return;
        }
        if(!cmd.compare(".finish")) {
            got_cmd=true;
            disp=lang_get_string(1,(*i).lang,325)+eol;
            ec=0;
            dcc_send_msg((*i).socket,disp,ec,(*i).ssl);

            (*i).filesystem=0;
            return;
        }
        if(!cmd.compare(".del")) {
            got_cmd=true;
            if(param1.empty() || !param2.empty()) {
                disp=lang_get_string(1,(*i).lang,403)+eol;
                ec=0;
                dcc_send_msg((*i).socket,disp,ec,(*i).ssl);
            } else {
                int fs_ec=filesys_dcc_delete((*i).as_in_logic,param1);
                switch(fs_ec) {
                    case DCC_NO_SUCH_FILE:
                        disp=lang_get_string(1,(*i).lang,334)+eol;
                        break;
                    case DCC_ACCESS_DENIED:
                        disp=lang_get_string(1,(*i).lang,404)+eol;
                        break;
                    case 0:
                        disp=lang_get_string(1,(*i).lang,405)+eol;
                }
                if(!disp.empty()) {
                    ec=0;
                    dcc_send_msg((*i).socket,disp,ec,(*i).ssl);
                }
            }
            (*i).filesystem=1000;
            return;
        }
        if(!cmd.compare(".list")) {
            got_cmd=true;
            vector<c_file> files;
            filesys_dcc_filelist((*i).as_in_logic,files,false);
            vector<c_file>::iterator i1;
            string msg;
            for(i1=files.begin(); i1!=files.end(); ++i1) {
                if(!((*i1).is_owner((*i).as_in_logic) || (*i1).can_read((*i).as_in_logic) || (*i1).can_delete((*i).as_in_logic)))
                    continue;
                if((*i1).is_owner((*i).as_in_logic))
                    msg+=lang_get_string(1,(*i).lang,296);
                if((*i1).can_read((*i).as_in_logic))
                    if(!msg.empty()) {
                        msg+=" ";
                        msg+=lang_get_string(1,(*i).lang,297);
                    }
                    else
                        msg+=lang_get_string(1,(*i).lang,297);
                if((*i1).can_delete((*i).as_in_logic))
                    if(!msg.empty()) {
                        msg+=" ";
                        msg+=lang_get_string(1,(*i).lang,298);
                    }
                    else
                        msg+=lang_get_string(1,(*i).lang,298);
                if((*i1).file_type==ft_file)
                    if(!msg.empty()) {
                        msg+=" ";
                        msg+=lang_get_string(1,(*i).lang,471);
                    }
                    else
                        msg+=lang_get_string(1,(*i).lang,471);
                if((*i1).file_type==ft_message)
                    if(!msg.empty()) {
                        msg+=" ";
                        msg+=lang_get_string(1,(*i).lang,472);
                    }
                    else
                        msg+=lang_get_string(1,(*i).lang,472);
                msg+=" = ";
                msg+=(*i1).public_name+eol;
            }
            if(msg.empty())
                disp=lang_get_string(1,(*i).lang,470)+eol;
            else
                disp=msg;
            ec=0;
            dcc_send_msg((*i).socket,disp,ec,(*i).ssl);
        }
        if(!cmd.compare(".help")) {
            got_cmd=true;
            disp=lang_get_string(1,(*i).lang,691)+eol;
            ec=0;
            dcc_send_msg((*i).socket,disp,ec,(*i).ssl);
        }
        if(!got_cmd && !cmd.empty()) {
            (*i).filesystem=2;
            return;
        }
        return;
    }

    // .editfile
    if((*i).filesystem==2000) {
        disp=lang_subst(lang_get_string(1,(*i).lang,283),(*i).fs.name)+eol;
        disp+=lang_get_string(1,(*i).lang,284)+eol;
        if((*i).fs.file.is_owner((*i).as_in_logic))
            disp+=lang_get_string(1,(*i).lang,285)+eol;
        else
            disp+=lang_get_string(1,(*i).lang,286)+eol;
        if((*i).fs.file.can_read((*i).as_in_logic))
            disp+=lang_get_string(1,(*i).lang,287)+eol;
        else
            disp+=lang_get_string(1,(*i).lang,288)+eol;
        if((*i).fs.file.can_delete((*i).as_in_logic))
            disp+=lang_get_string(1,(*i).lang,289)+eol;
        else
            disp+=lang_get_string(1,(*i).lang,290)+eol;
        disp+=lang_get_string(1,(*i).lang,291)+eol;
        ec=0;
        dcc_send_msg((*i).socket,disp,ec,(*i).ssl);

        (*i).filesystem++;
        return;
    }

    if((*i).filesystem==2001) {
        if(p.empty())
            return;

        if(!cmd.compare(".attr")) {
            if((*i).fs.file.is_owner((*i).as_in_logic)) {
                (*i).filesystem=3000;
            } else {
                disp=lang_get_string(1,(*i).lang,286)+eol;
                ec=0;
                dcc_send_msg((*i).socket,disp,ec,(*i).ssl);
                (*i).filesystem=2000;
            }
            return;
        }

        if(!cmd.compare(".show")) {
            disp.clear();
            vector<c_access>::iterator i1;
            for(i1=(*i).fs.file.access.begin(); i1!=(*i).fs.file.access.end(); ++i1)
                disp+=dcc_loop_filesystem_disp_rights((*i).lang,*i1,eol);
            ec=0;
            dcc_send_msg((*i).socket,disp,ec,(*i).ssl);
            (*i).filesystem=2000;
            return;
        }

        if(!cmd.compare(".expiration")) {
            vector<c_access>::iterator i1;
            (*i).filesystem=7000;
            return;
        }

        if(!cmd.compare(".finish")) {
            disp=lang_get_string(1,(*i).lang,319)+eol;
            ec=0;
            dcc_send_msg((*i).socket,disp,ec,(*i).ssl);

            (*i).filesystem=6000;
            return;
        }
        (*i).filesystem--;
        return;
    }

    // .attr
    if((*i).filesystem==3000) {
        disp=lang_get_string(1,(*i).lang,293)+eol;
        vector<c_access>::iterator i1;
        for(i1=(*i).fs.file.access.begin(); i1!=(*i).fs.file.access.end(); ++i1)
            disp+=dcc_loop_filesystem_disp_rights((*i).lang,*i1,eol);
        disp+=lang_get_string(1,(*i).lang,302)+eol;
        disp+=lang_get_string(1,(*i).lang,303)+eol;
        ec=0;
        dcc_send_msg((*i).socket,disp,ec,(*i).ssl);
        (*i).filesystem++;
        return;
    }
    if((*i).filesystem==3001) {
        bool all_users=!param1.compare("*");
        got=false;
        bool got_cmd=false;
        if(!cmd.compare(".remove")) {
            got_cmd=true;
            if(all_users) {
                vector<c_access>::iterator i1;
again2:
                for(i1=(*i).fs.file.access.begin(); i1!=(*i).fs.file.access.end(); ++i1) {
                    if((*i1).user_name.compare((*i).as_in_logic)) {
                        (*i).fs.file.access.erase(i1);
                        got=true;
                        goto again2;
                    }
                }
            } else {
                vector<c_access>::iterator i1;
again1:
                for(i1=(*i).fs.file.access.begin(); i1!=(*i).fs.file.access.end(); ++i1) {
                    if(!(*i1).user_name.compare(param1)) {
                        (*i).fs.file.access.erase(i1);
                        got=true;
                        goto again1;
                    }
                }
                if(!got) {
                    disp=lang_get_string(1,(*i).lang,304)+eol;
                    ec=0;
                    dcc_send_msg((*i).socket,disp,ec,(*i).ssl);
                    (*i).filesystem=3000;
                } else
                    (*i).filesystem=3000;
            }
            (*i).filesystem=3000;
            return;
        }
        if(!cmd.compare(".add")) {
            got_cmd=true;
            (*i).fs.new_access.clear();
            (*i).filesystem=5000;
        }
        if(!cmd.compare(".change")) {
            got_cmd=true;
            (*i).fs.new_access.clear();
            (*i).fs.new_access.all_users=all_users;
            if(!all_users)
                (*i).fs.new_access.user_name=param1;
            else
                (*i).fs.new_access.user_name="";
            vector<c_access>::iterator i1;
            got=false;
            for(i1=(*i).fs.file.access.begin(); i1!=(*i).fs.file.access.end(); ++i1) {
                if(all_users && (*i1).all_users) {
                    (*i).fs.new_access.del=(*i1).del;
                    (*i).fs.new_access.notify_owner=(*i1).notify_owner;
                    (*i).fs.new_access.notify_user=(*i1).notify_user;
                    (*i).fs.new_access.owner=(*i1).owner;
                    (*i).fs.new_access.read=(*i1).read;
                    got=true;
                    break;
                }
                if(!all_users && !(*i1).user_name.compare(param1)) {
                    (*i).fs.new_access.del=(*i1).del;
                    (*i).fs.new_access.notify_owner=(*i1).notify_owner;
                    (*i).fs.new_access.notify_user=(*i1).notify_user;
                    (*i).fs.new_access.owner=(*i1).owner;
                    (*i).fs.new_access.read=(*i1).read;
                    got=true;
                    break;
                }
            }
            if(!got && !all_users) {
                disp=lang_get_string(1,(*i).lang,304)+eol;
                ec=0;
                dcc_send_msg((*i).socket,disp,ec,(*i).ssl);
                (*i).filesystem=3000;
            } else
                (*i).filesystem=4000;
        }
        if(!cmd.compare(".finish")) {
            got_cmd=true;
            disp=lang_get_string(1,(*i).lang,319)+eol;
            ec=0;
            dcc_send_msg((*i).socket,disp,ec,(*i).ssl);

            (*i).filesystem=6000;
            return;
        }
        if(!cmd.compare(".addchan")) {
            got_cmd=true;
            got=false;
            vector<string>::iterator i1;
            for(i1=(*i).access_to_channel.begin(); i1!=(*i).access_to_channel.end(); ++i1)
                if(!cmp_strings_case_insensitive(*i1,param1)) {
                    got=true;
                    break;
                }
            if(got) {
                (*i).fs.new_access.clear();
                (*i).fs.new_access.all_on_channel=param1;
                c_user u;
                vector<c_channel> chs;
                logic_partyline_get_user((*i).as_in_logic,u,chs);
                vector<c_channel>::iterator i2;
                bool can=(*i).access_to_can_send_unknown_users;
                for(i2=chs.begin(); i2!=chs.end(); ++i2)
                    if(!cmp_strings_case_insensitive((*i2).channel_name,p)) {
                        can&=(*i2).can_send_unknown_users;
                        break;
                    }
                if(!(*i).access_to_can_send_unknown_users)
                    can=false;
                if(!can) {
                    (*i).filesystem=4000;
                } else {
                    (*i).filesystem=8000;
                }
            } else {
                (*i).fs.new_access.all_on_channel="";
                (*i).fs.new_access.also_unknown=false;
                disp=lang_get_string(1,(*i).lang,323)+eol;
                ec=0;
                dcc_send_msg((*i).socket,disp,ec,(*i).ssl);
                return;
            }
            return;
        }
        if(!got_cmd && !cmd.empty()) {
            (*i).filesystem=3000;
            return;
        }
        return;
    }

    // .change
    if((*i).filesystem==4000) {
        disp=lang_get_string(1,(*i).lang,305)+eol;
        disp+=dcc_loop_filesystem_disp_rights((*i).lang,(*i).fs.new_access,eol);
        disp+=lang_get_string(1,(*i).lang,306)+eol;
        ec=0;
        dcc_send_msg((*i).socket,disp,ec,(*i).ssl);
        (*i).filesystem++;
        return;
    }
    // .change && .add
    if((*i).filesystem==4001) {
        if(p.empty())
            return;

        if(p.compare("y") && p.compare("n") && p.compare("Y") && p.compare("N")) {
            disp=lang_get_string(1,(*i).lang,306)+eol; // !!!
            ec=0;
            dcc_send_msg((*i).socket,disp,ec,(*i).ssl);
        } else {
            if(!p.compare("n") || !p.compare("N")) {
                (*i).fs.new_access.owner=false; // !!!
                disp=lang_get_string(1,(*i).lang,307)+eol; // !!!
                ec=0;
                dcc_send_msg((*i).socket,disp,ec,(*i).ssl);
                (*i).filesystem++;
            } else {
                (*i).fs.new_access.owner=true; // !!!
                disp=lang_get_string(1,(*i).lang,307)+eol; // !!!
                ec=0;
                dcc_send_msg((*i).socket,disp,ec,(*i).ssl);
                (*i).filesystem++;
            }
        }
        return;
    }
    if((*i).filesystem==4002) {
        if(p.empty())
            return;
        if(p.compare("y") && p.compare("n") && p.compare("Y") && p.compare("N")) {
            disp=lang_get_string(1,(*i).lang,307)+eol; // !!!
            ec=0;
            dcc_send_msg((*i).socket,disp,ec,(*i).ssl);
        } else {
            if(!p.compare("n") || !p.compare("N")) {
                (*i).fs.new_access.read=false; // !!!
                disp=lang_get_string(1,(*i).lang,308)+eol; // !!!
                ec=0;
                dcc_send_msg((*i).socket,disp,ec,(*i).ssl);
                (*i).filesystem++;
            } else {
                (*i).fs.new_access.read=true; // !!!
                disp=lang_get_string(1,(*i).lang,308)+eol; // !!!
                ec=0;
                dcc_send_msg((*i).socket,disp,ec,(*i).ssl);
                (*i).filesystem++;
            }
        }
        return;
    }
    if((*i).filesystem==4003) {
        if(p.empty())
            return;
        if(p.compare("y") && p.compare("n") && p.compare("Y") && p.compare("N")) {
            disp=lang_get_string(1,(*i).lang,308)+eol; // !!!
            ec=0;
            dcc_send_msg((*i).socket,disp,ec,(*i).ssl);
        } else {
            if(!p.compare("n") || !p.compare("N")) {
                (*i).fs.new_access.del=false; // !!!
                disp=lang_get_string(1,(*i).lang,309)+eol; // !!!
                ec=0;
                dcc_send_msg((*i).socket,disp,ec,(*i).ssl);
                (*i).filesystem++;
            } else {
                (*i).fs.new_access.del=true; // !!!
                disp=lang_get_string(1,(*i).lang,309)+eol; // !!!
                ec=0;
                dcc_send_msg((*i).socket,disp,ec,(*i).ssl);
                (*i).filesystem++;
            }
        }
        return;
    }
    if((*i).filesystem==4004) {
        if(p.empty())
            return;
        if(p.compare("y") && p.compare("n") && p.compare("Y") && p.compare("N")) {
            disp=lang_get_string(1,(*i).lang,309)+eol; // !!!
            ec=0;
            dcc_send_msg((*i).socket,disp,ec,(*i).ssl);
        } else {
            if(!p.compare("n") || !p.compare("N")) {
                (*i).fs.new_access.notify_owner=false; // !!!
                disp=lang_get_string(1,(*i).lang,310)+eol; // !!!
                ec=0;
                dcc_send_msg((*i).socket,disp,ec,(*i).ssl);
                (*i).filesystem++;
            } else {
                (*i).fs.new_access.notify_owner=true; // !!!
                disp=lang_get_string(1,(*i).lang,310)+eol; // !!!
                ec=0;
                dcc_send_msg((*i).socket,disp,ec,(*i).ssl);
                (*i).filesystem++;
            }
        }
        return;
    }

    // SECURE flag
    if((*i).filesystem==4005) {
        if(p.empty())
            return;
        if(p.compare("y") && p.compare("n") && p.compare("Y") && p.compare("N")) {
            disp=lang_get_string(1,(*i).lang,310)+eol; // !!!
            ec=0;
            dcc_send_msg((*i).socket,disp,ec,(*i).ssl);
        } else {
            if(!p.compare("n") || !p.compare("N")) {
                (*i).fs.new_access.notify_user=false; // !!!
                if(!(*i).fs.new_access.notify_user) {
                    (*i).fs.new_access.notify_user_message="";
                    (*i).filesystem+=2;
                }
                disp=lang_get_string(1,(*i).lang,326)+eol; // !!!
                ec=0;
                dcc_send_msg((*i).socket,disp,ec,(*i).ssl);
                (*i).filesystem++;
            } else {
                (*i).fs.new_access.notify_user=true; // !!!
                disp=lang_get_string(1,(*i).lang,326)+eol; // !!!
                ec=0;
                dcc_send_msg((*i).socket,disp,ec,(*i).ssl);
                (*i).filesystem++;
            }
        }
        return;
    }
    if((*i).filesystem==4006) {
        if(p.empty())
            return;
        if(p.compare("y") && p.compare("n") && p.compare("Y") && p.compare("N")) {
            disp=lang_get_string(1,(*i).lang,326)+eol; // !!!
            ec=0;
            dcc_send_msg((*i).socket,disp,ec,(*i).ssl);
        } else {
            if(!p.compare("n") || !p.compare("N")) {
                (*i).fs.new_access.secure=false; // !!!
                if(!(*i).fs.new_access.notify_owner || (*i).fs.file.file_type==ft_message) {
                    (*i).fs.new_access.notify_owner_message="";
                    (*i).filesystem+=3;
                    return;
                }
                disp=lang_get_string(1,(*i).lang,327)+eol; // !!!
                disp+=lang_get_string(1,(*i).lang,349)+eol;
                disp+=lang_get_string(1,(*i).lang,350)+eol;
                disp+=lang_get_string(1,(*i).lang,353)+eol;
                ec=0;
                dcc_send_msg((*i).socket,disp,ec,(*i).ssl);
                (*i).filesystem++;
            } else {
                (*i).fs.new_access.secure=true; // !!!
                if(!(*i).fs.new_access.notify_owner) {
                    (*i).fs.new_access.notify_owner_message="";
                    (*i).filesystem+=3;
                    return;
                }
                disp=lang_get_string(1,(*i).lang,327)+eol; // !!!
                disp+=lang_get_string(1,(*i).lang,349)+eol;
                disp+=lang_get_string(1,(*i).lang,350)+eol;
                disp+=lang_get_string(1,(*i).lang,353)+eol;
                ec=0;
                dcc_send_msg((*i).socket,disp,ec,(*i).ssl);
                (*i).filesystem++;
            }
        }
        return;
    }
    // OWNER message edit
    if((*i).filesystem==4007) {
        if(p.empty())
            return;
        if(!p.compare(".")) {
            disp=lang_get_string(1,(*i).lang,328)+eol; // !!!
            ec=0;
            dcc_send_msg((*i).socket,disp,ec,(*i).ssl);
            if(!(*i).fs.new_access.notify_owner /*|| (*i).fs.file.file_type==ft_message*/) {
                (*i).filesystem+=2;
                return;
            }
            (*i).filesystem++;
            disp=lang_get_string(1,(*i).lang,329)+eol; // !!!
            disp+=lang_get_string(1,(*i).lang,349)+eol;
            disp+=lang_get_string(1,(*i).lang,351)+eol;
            disp+=lang_get_string(1,(*i).lang,352)+eol;
            disp+=lang_get_string(1,(*i).lang,353)+eol;
            dcc_send_msg((*i).socket,disp,ec,(*i).ssl);
            return;
        } else {
            (*i).fs.new_access.notify_owner_message+=p;
            (*i).fs.new_access.notify_owner_message+="\n";
        }
        return;
    }
    // USER message edit
    if((*i).filesystem==4008) {
        if(p.empty())
            return;
        if(!p.compare(".")) {
            disp=lang_get_string(1,(*i).lang,328)+eol; // !!!
            ec=0;
            dcc_send_msg((*i).socket,disp,ec,(*i).ssl);
            (*i).filesystem++;
            return;
        } else {
            (*i).fs.new_access.notify_user_message+=p;
            (*i).fs.new_access.notify_user_message+="\n";
        }
        return;
    }

    // done
    if((*i).filesystem==4009) {
        disp=lang_get_string(1,(*i).lang,311)+eol; // !!!
        disp+=dcc_loop_filesystem_disp_rights((*i).lang,(*i).fs.new_access,eol);
        disp+=lang_get_string(1,(*i).lang,312)+eol; // !!!
        ec=0;
        dcc_send_msg((*i).socket,disp,ec,(*i).ssl);
        (*i).filesystem++;
        return;
    }
    if((*i).filesystem==4010) {
        if(p.empty())
            return;
        if(p.compare("y") && p.compare("n") && p.compare("Y") && p.compare("N")) {
            disp=lang_get_string(1,(*i).lang,311)+eol; // !!!
            disp+=dcc_loop_filesystem_disp_rights((*i).lang,(*i).fs.new_access,eol);
            disp+=lang_get_string(1,(*i).lang,312)+eol; // !!!
            ec=0;
            dcc_send_msg((*i).socket,disp,ec,(*i).ssl);
        } else {
            if(!p.compare("n") || !p.compare("N")) {
                disp=lang_get_string(1,(*i).lang,314)+eol; // !!!
                ec=0;
                dcc_send_msg((*i).socket,disp,ec,(*i).ssl);
                (*i).filesystem=2000;
            } else {
                if((*i).fs.new_access.user_name.length()>1 && (*i).fs.new_access.user_name[0]=='*') {
                    // from .addchan
                    string chan;
                    if((*i).fs.new_access.all_on_channel.empty())
                        chan=(*i).fs.new_access.user_name;
                    else
                        chan=(*i).fs.new_access.all_on_channel;
                    chan.erase(0,1);
                    vector<string> users;
                    map<string,string> vars;
                    vector<c_channel>::iterator i2;
                    for(i2=r_channels.begin(); i2!=r_channels.end(); ++i2)
                        if(!cmp_strings_case_insensitive((*i2).channel_name,chan) && !(*i2).host_unknown)
                            users.push_back((*i2).username);

                    vector<string>::iterator i3;
                    for(i3=users.begin(); i3!=users.end(); ++i3) {
                        vector<c_access>::iterator i1;
again5:
                        for(i1=(*i).fs.file.access.begin(); i1!=(*i).fs.file.access.end(); ++i1) {
                            if((*i1).user_name.compare((*i).as_in_logic) && !(*i1).user_name.compare(*i3)) {
                                (*i).fs.file.access.erase(i1);
                                goto again5;
                            }
                        }
                    }
                    for(i3=users.begin(); i3!=users.end(); ++i3) {
                        (*i).fs.new_access.user_name=*i3;
                        (*i).fs.file.access.push_back((*i).fs.new_access);
                    }
                } else
                    if(!(*i).fs.new_access.all_on_channel.empty()) {
                        (*i).fs.file.access.push_back((*i).fs.new_access);
                    } else
                        if((*i).fs.new_access.all_users) {
                            vector<c_access>::iterator i1;
again4:
                            for(i1=(*i).fs.file.access.begin(); i1!=(*i).fs.file.access.end(); ++i1) {
                                if((*i1).user_name.compare((*i).as_in_logic)) {
                                    (*i).fs.file.access.erase(i1);
                                    goto again4;
                                }
                            }
                            (*i).fs.file.access.push_back((*i).fs.new_access);
                        } else {
                            vector<c_access>::iterator i1;
again3:
                            for(i1=(*i).fs.file.access.begin(); i1!=(*i).fs.file.access.end(); ++i1) {
                                if(!(*i1).user_name.compare((*i).fs.new_access.user_name)) {
                                    (*i).fs.file.access.erase(i1);
                                    goto again3;
                                }
                            }
                            (*i).fs.file.access.push_back((*i).fs.new_access);
                        }

                disp=lang_get_string(1,(*i).lang,313)+eol; // !!!
                ec=0;
                dcc_send_msg((*i).socket,disp,ec,(*i).ssl);
                (*i).filesystem=3000;
            }
        }
        return;
    }

    // .add
    if((*i).filesystem==5000) {
        disp=lang_get_string(1,(*i).lang,315)+eol;
        deque<c_user>::iterator i1;
        for(i1=r_users.begin(); i1!=r_users.end(); ++i1)
            disp+=(*i1).name+eol;
        disp+=lang_get_string(1,(*i).lang,316)+eol;
        disp+=lang_get_string(1,(*i).lang,303)+eol;
        ec=0;
        dcc_send_msg((*i).socket,disp,ec,(*i).ssl);
        (*i).filesystem++;
        return;
    }
    if((*i).filesystem==5001) {
        if(p.empty())
            return;
        bool all_users=!p.compare("*");
        (*i).fs.new_access.clear();
        (*i).fs.new_access.all_users=all_users;
        if(!all_users) {
            c_user u;
            vector<c_channel> chs;
            logic_partyline_get_user(p,u,chs);
            bool unknown_user=u.host_unknown;
            logic_partyline_get_user((*i).as_in_logic,u,chs);
            vector<c_channel>::iterator i2;
            bool can=(*i).access_to_can_send_unknown_users;
            for(i2=chs.begin(); i2!=chs.end(); ++i2)
                if(!cmp_strings_case_insensitive((*i2).channel_name,p)) {
                    can&=(*i2).can_send_unknown_users;
                    break;
                }
            if(unknown_user && !can) {
                disp=lang_get_string(1,(*i).lang,480)+eol;
                ec=0;
                dcc_send_msg((*i).socket,disp,ec,(*i).ssl);
                (*i).filesystem--;
                return;
            }
            (*i).fs.new_access.user_name=p;
        } else
            (*i).fs.new_access.user_name="";

        if(!all_users) {
            deque<c_user>::iterator i1;
            string user;
            for(i1=r_users.begin(); i1!=r_users.end(); ++i1) {
                if(!(*i1).name.compare(p)) {
                    user=p;
                    break;
                }
            }
            if(user.empty()) {
                disp=lang_get_string(1,(*i).lang,304)+eol;
                ec=0;
                dcc_send_msg((*i).socket,disp,ec,(*i).ssl);
                (*i).filesystem--;
                return;
            }
        }
        vector<c_access>::iterator i1;
        got=false;
        for(i1=(*i).fs.file.access.begin(); i1!=(*i).fs.file.access.end(); ++i1) {
            if(all_users && (*i1).all_users) {
                (*i).fs.new_access.del=(*i1).del;
                (*i).fs.new_access.notify_owner=(*i1).notify_owner;
                (*i).fs.new_access.notify_user=(*i1).notify_user;
                (*i).fs.new_access.owner=(*i1).owner;
                (*i).fs.new_access.read=(*i1).read;
                got=true;
                break;
            }
            if(!all_users && !(*i1).user_name.compare(p)) {
                (*i).fs.new_access.del=(*i1).del;
                (*i).fs.new_access.notify_owner=(*i1).notify_owner;
                (*i).fs.new_access.notify_user=(*i1).notify_user;
                (*i).fs.new_access.owner=(*i1).owner;
                (*i).fs.new_access.read=(*i1).read;
                got=true;
                break;
            }
        }
        if(got) {
            if(all_users) {
                if(!(*i).access_to_can_send_all_users) {
                    disp=lang_get_string(1,(*i).lang,479)+eol;
                    ec=0;
                    dcc_send_msg((*i).socket,disp,ec,(*i).ssl);
                    return;
                }

                disp=lang_get_string(1,(*i).lang,317)+eol;
                ec=0;
                dcc_send_msg((*i).socket,disp,ec,(*i).ssl);
                (*i).filesystem=5500;
                return;
            } else {
                disp=lang_subst(lang_get_string(1,(*i).lang,318),(*i).fs.new_access.user_name)+eol;
                ec=0;
                dcc_send_msg((*i).socket,disp,ec,(*i).ssl);
                (*i).filesystem=5500;
                return;
            }
        }
        (*i).filesystem=5200;
        return;
    }

    if((*i).filesystem==5200) {
        disp=lang_get_string(1,(*i).lang,306)+eol;
        ec=0;
        dcc_send_msg((*i).socket,disp,ec,(*i).ssl);
        (*i).filesystem=4001;
        return;
    }

    // replace? (.add)
    if((*i).filesystem==5500) {
        if(p.empty())
            return;
        if(p.compare("y") && p.compare("n") && p.compare("Y") && p.compare("N")) {
            if((*i).fs.new_access.all_users) {
                disp=lang_get_string(1,(*i).lang,317)+eol;
                ec=0;
                dcc_send_msg((*i).socket,disp,ec,(*i).ssl);
                return;
            } else {
                disp=lang_subst(lang_get_string(1,(*i).lang,318),(*i).fs.new_access.user_name)+eol;
                ec=0;
                dcc_send_msg((*i).socket,disp,ec,(*i).ssl);
                return;
            }
        } else {
            if(!p.compare("n") || !p.compare("N")) {
                (*i).filesystem=2000;
            } else {
                (*i).filesystem=5200;
            }
        }
        return;
    }

    // .finish (save changes?)
    if((*i).filesystem==6000) {
        if(p.empty())
            return;
        if(p.compare("y") && p.compare("n") && p.compare("Y") && p.compare("N")) {
            disp=lang_get_string(1,(*i).lang,319)+eol; // !!!
            ec=0;
            dcc_send_msg((*i).socket,disp,ec,(*i).ssl);
        } else {
            if(!p.compare("n") || !p.compare("N")) {
                disp=lang_get_string(1,(*i).lang,321)+eol; // !!!
                ec=0;
                dcc_send_msg((*i).socket,disp,ec,(*i).ssl);
                (*i).filesystem=1;
            } else {
                (*i).fs.file.published=true;

                vector<c_access>::iterator i1;

                bool read=true;
                bool notify_user=true;
                bool old_secure=false;
                bool notify_owner=false;

                for(i1=(*i).fs.file.access.begin(); i1!=(*i).fs.file.access.end(); ++i1) {
                    read|=(*i1).read;
                    notify_user|=(*i1).notify_user;
                    notify_owner|=(*i1).notify_owner;
                    old_secure|=(*i1).secure;
                }

                bool forced_secure;

                if(filesys_dcc_set_file_attrs((*i).as_in_logic,(*i).fs.file,forced_secure)) {
                    disp.clear();
                    if(!old_secure && forced_secure)
                        disp+=lang_get_string(1,(*i).lang,346)+eol;

                    disp+=lang_get_string(1,(*i).lang,320)+eol;
                    ec=0;
                    dcc_send_msg((*i).socket,disp,ec,(*i).ssl);
                } else {
                    disp=lang_get_string(1,(*i).lang,292)+eol;
                    ec=0;
                    dcc_send_msg((*i).socket,disp,ec,(*i).ssl);
                }
                (*i).filesystem=1;
            }
        }
        return;
    }

    if((*i).filesystem==5001) {
        if(p.empty())
            return;
        bool all_users=!p.compare("*");
        (*i).fs.new_access.clear();
        (*i).fs.new_access.all_users=all_users;
        if(!all_users)
            (*i).fs.new_access.user_name=p;
        else
            (*i).fs.new_access.user_name="";

        if(!all_users) {
            deque<c_user>::iterator i1;
            string user;
            for(i1=r_users.begin(); i1!=r_users.end(); ++i1) {
                if(!(*i1).name.compare(p)) {
                    user=p;
                    break;
                }
            }
            if(user.empty()) {
                disp=lang_get_string(1,(*i).lang,304)+eol;
                ec=0;
                dcc_send_msg((*i).socket,disp,ec,(*i).ssl);
                (*i).filesystem--;
                return;
            }
        }
        vector<c_access>::iterator i1;
        got=false;
        for(i1=(*i).fs.file.access.begin(); i1!=(*i).fs.file.access.end(); ++i1) {
            if(all_users && (*i1).all_users) {
                (*i).fs.new_access.del=(*i1).del;
                (*i).fs.new_access.notify_owner=(*i1).notify_owner;
                (*i).fs.new_access.notify_user=(*i1).notify_user;
                (*i).fs.new_access.owner=(*i1).owner;
                (*i).fs.new_access.read=(*i1).read;
                got=true;
                break;
            }
            if(!all_users && !(*i1).user_name.compare(p)) {
                (*i).fs.new_access.del=(*i1).del;
                (*i).fs.new_access.notify_owner=(*i1).notify_owner;
                (*i).fs.new_access.notify_user=(*i1).notify_user;
                (*i).fs.new_access.owner=(*i1).owner;
                (*i).fs.new_access.read=(*i1).read;
                got=true;
                break;
            }
        }
        if(got) {
            if(all_users) {
                disp=lang_get_string(1,(*i).lang,317)+eol;
                ec=0;
                dcc_send_msg((*i).socket,disp,ec,(*i).ssl);
                (*i).filesystem=5500;
                return;
            } else {
                disp=lang_subst(lang_get_string(1,(*i).lang,318),(*i).fs.new_access.user_name)+eol;
                ec=0;
                dcc_send_msg((*i).socket,disp,ec,(*i).ssl);
                (*i).filesystem=5500;
                return;
            }
        }
        (*i).filesystem=5200;
        return;
    }

    if((*i).filesystem==5200) {
        disp=lang_get_string(1,(*i).lang,306)+eol;
        ec=0;
        dcc_send_msg((*i).socket,disp,ec,(*i).ssl);
        (*i).filesystem=4001;
        return;
    }

    if((*i).filesystem==7000) {
        disp=lang_get_string(1,(*i).lang,433)+eol;
        my_time_t ex=(*i).fs.file.expiration;
        if(ex==0)
            disp+=lang_get_string(1,(*i).lang,434)+eol;
        else {
            disp+=make_nice_time(ex)+eol;
        }

        disp+=lang_get_string(1,(*i).lang,435)+eol;
        ec=0;
        dcc_send_msg((*i).socket,disp,ec,(*i).ssl);

        (*i).filesystem++;
        return;
    }
    if((*i).filesystem==7001) {
        string p=c;
        if(p.empty())
            return;

        if(p.compare("y") && p.compare("n") && p.compare("Y") && p.compare("N")) {
            disp=lang_get_string(1,(*i).lang,435)+eol; // !!!
            ec=0;
            dcc_send_msg((*i).socket,disp,ec,(*i).ssl);
        } else {
            if(!p.compare("n") || !p.compare("N")) {
                (*i).filesystem=2000;
                return;
            } else {
                disp=lang_get_string(1,(*i).lang,436)+eol; // !!!
                ec=0;
                dcc_send_msg((*i).socket,disp,ec,(*i).ssl);
                (*i).filesystem++;
                return;
            }
        }
    }
    if((*i).filesystem==7002) {
        string p=c;
        if(p.empty())
            return;

        int x=stoi(p);
        if(x==0) {
            (*i).fs.file.expiration=0;
            disp=lang_get_string(1,(*i).lang,441)+eol; // !!!
            ec=0;
            dcc_send_msg((*i).socket,disp,ec,(*i).ssl);
            (*i).filesystem=2000;
            return;
        }
        if(x<1970) {
            disp=lang_get_string(1,(*i).lang,436)+eol; // !!!
            ec=0;
            dcc_send_msg((*i).socket,disp,ec,(*i).ssl);
            return;
        }
        (*i).filesystem_expiration.tm_year=x;
        disp=lang_get_string(1,(*i).lang,437)+eol; // !!!
        ec=0;
        dcc_send_msg((*i).socket,disp,ec,(*i).ssl);
        (*i).filesystem++;
        return;
    }
    if((*i).filesystem==7003) {
        string p=c;
        if(p.empty())
            return;

        int x=stoi(p);
        if(x<1 || x>12) {
            disp=lang_get_string(1,(*i).lang,437)+eol; // !!!
            ec=0;
            dcc_send_msg((*i).socket,disp,ec,(*i).ssl);
            return;
        }
        (*i).filesystem_expiration.tm_mon=x;
        disp=lang_get_string(1,(*i).lang,438)+eol; // !!!
        ec=0;
        dcc_send_msg((*i).socket,disp,ec,(*i).ssl);
        (*i).filesystem++;
        return;
    }
    if((*i).filesystem==7004) {
        string p=c;
        if(p.empty())
            return;

        int x=stoi(p);
        if(x<1 || x>31) {
            disp=lang_get_string(1,(*i).lang,438)+eol; // !!!
            ec=0;
            dcc_send_msg((*i).socket,disp,ec,(*i).ssl);
            return;
        }
        (*i).filesystem_expiration.tm_mday=x;
        disp=lang_get_string(1,(*i).lang,439)+eol; // !!!
        ec=0;
        dcc_send_msg((*i).socket,disp,ec,(*i).ssl);
        (*i).filesystem++;
        return;
    }
    if((*i).filesystem==7005) {
        string p=c;
        if(p.empty())
            return;

        int x=stoi(p);
        if(x<0 || x>23) {
            disp=lang_get_string(1,(*i).lang,439)+eol; // !!!
            ec=0;
            dcc_send_msg((*i).socket,disp,ec,(*i).ssl);
            return;
        }
        (*i).filesystem_expiration.tm_hour=x;
        disp=lang_get_string(1,(*i).lang,440)+eol; // !!!
        ec=0;
        dcc_send_msg((*i).socket,disp,ec,(*i).ssl);
        (*i).filesystem++;
        return;
    }
    if((*i).filesystem==7006) {
        string p=c;
        if(p.empty())
            return;

        int x=stoi(p);
        if(x<0 || x>59) {
            disp=lang_get_string(1,(*i).lang,440)+eol; // !!!
            ec=0;
            dcc_send_msg((*i).socket,disp,ec,(*i).ssl);
            return;
        }
        (*i).filesystem_expiration.tm_min=x;
        (*i).filesystem_expiration.tm_isdst=-1;
        (*i).filesystem_expiration.tm_sec=0;
        (*i).filesystem_expiration.tm_wday=0;
        (*i).filesystem_expiration.tm_yday=0;

        (*i).filesystem_expiration.tm_mon-=1;
        (*i).filesystem_expiration.tm_year-=1900;

        my_time_t ex=mktime(&(*i).filesystem_expiration);
        if(ex==-1) {
            disp=lang_get_string(1,(*i).lang,442)+eol; // !!!
            disp+=lang_get_string(1,(*i).lang,436)+eol; // !!!
            ec=0;
            dcc_send_msg((*i).socket,disp,ec,(*i).ssl);
            (*i).filesystem=7002;
            return;
        }
        if(ex<my_time(NULL)) {
            disp=lang_get_string(1,(*i).lang,443)+eol; // !!!
            disp+=lang_get_string(1,(*i).lang,436)+eol; // !!!
            ec=0;
            dcc_send_msg((*i).socket,disp,ec,(*i).ssl);
            (*i).filesystem=7002;
            return;
        }

        (*i).fs.file.expiration=ex;

        disp=lang_get_string(1,(*i).lang,433)+eol;
        if(ex==0)
            disp+=lang_get_string(1,(*i).lang,434)+eol;
        else {
            disp+=make_nice_time(ex)+eol;
        }

        ec=0;
        dcc_send_msg((*i).socket,disp,ec,(*i).ssl);
        (*i).filesystem=2000;
        return;
    }

    // also unknown users?
    if((*i).filesystem==8000) {
        disp=lang_get_string(1,(*i).lang,476)+eol;
        ec=0;
        dcc_send_msg((*i).socket,disp,ec,(*i).ssl);
        (*i).filesystem++;
        return;
    }
    if((*i).filesystem==8001) {
        string p=c;
        if(!p.compare(""))
            return;

        if(p.compare("y") && p.compare("n") && p.compare("Y") && p.compare("N")) {
            disp=lang_get_string(1,(*i).lang,476)+eol;
            ec=0;
            dcc_send_msg((*i).socket,disp,ec,(*i).ssl);
        } else {
            if(!p.compare("n") || !p.compare("N")) {
                (*i).fs.new_access.also_unknown=false;
            } else {
                (*i).fs.new_access.also_unknown=true;;
            }
            disp=lang_get_string(1,(*i).lang,339)+eol;
            ec=0;
            dcc_send_msg((*i).socket,disp,ec,(*i).ssl);
            (*i).filesystem=4000;
            return;
        }
        return;
    }
}

