/***************************************************************************
                          dcc_loop_replication.cpp  -  description
                             -------------------
    begin                : Thu Dec 9 2004
    copyright            : (C) 2004-2007 by Marian VooDooMan Meravy
    email                : ghostvoodooman -a.T= users (dot) sourceforge.net
 ***************************************************************************/

/* $Id: dcc_loop_replication.cpp,v 1.16 2008/07/02 09:35:49 ghostvoodooman Exp $ */

/***************************************************************************

VooDoo cIRCle - an IRC (ro)bot
Copyright (C) 2004-2008 by Marian VooDooMan Meravy
  (ghostvoodooman -a.T= users (dot) sourceforge.net)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************/

/*!
    \file
    \brief Handles DCC / telnet communication
    \version $Id: dcc_loop_replication.cpp,v 1.16 2008/07/02 09:35:49 ghostvoodooman Exp $
*/

#if defined(HAVE_CONFIG_H)
#   include <config.h>
#endif

#include "../utils/memleak.h"
#include "../utils/assert.h"
#include "../utils.h"
#include "../params.h"

#include "dcc_internal.h"

extern deque<c_user> r_terminators;
extern deque<c_user> r_user_templates;
extern deque<c_user> r_users;

#include "../utils/precomp.h"

/*!
    \brief Should be called in a loop to process .replication commands in DCC CHAT/telnet connections
    \author VooDooMan
    \version 1
    \date 2004
    \param i Descriptor of remote connection
    \param eol End-of-line string
*/
void dcc_loop_replication(vector<c_dcc_client>::iterator i, string& eol)
{
    if((*i).replication==0)
        return;

    string disp;
    int ec=0;
    bool got=false;

    if((*i).replication==1) {
        (*i).b_repl_user=false;
        (*i).b_repl_user_template=false;
        (*i).b_repl_chandef=false;
        (*i).b_repl_proc=false;
        (*i).b_repl_private=false;
        (*i).b_repl_func=false;

        disp=lang_get_string(1,(*i).lang,421)+eol;

        {
            deque<c_user>::iterator i1;
            for(i1=r_users.begin(); i1!=r_users.end(); ++i1)
                disp+=(string)"USER = "+(*i1).name+eol;
        }
        {
            deque<c_user>::iterator i1;
            for(i1=r_user_templates.begin(); i1!=r_user_templates.end(); ++i1)
                disp+=(string)"USER TEMPLATE = "+(*i1).name+eol;
        }
        {
            c_module* core_mod=get_module("core",true);
            if(core_mod) {
                vector<c_procedure>::iterator i1;
                for(i1=core_mod->r_procedures.begin(); i1!=core_mod->r_procedures.end(); ++i1)
                    disp+=(string)"PROCEDURE = "+(*i1).name+eol;
            }
        }
        {
            c_module* core_mod=get_module("core",true);
            if(core_mod) {
                map<string,c_function*>::iterator i1;
                for(i1=core_mod->r_functions.functions.begin(); i1!=core_mod->r_functions.functions.end(); ++i1)
                    disp+=(string)"FUNCTION = "+(*i1).second->declaration+eol;
            }
        }
        {
            vector<c_channel_def>::iterator i1;
            for(i1=r_channel_defs.begin(); i1!=r_channel_defs.end(); ++i1)
                disp+=(string)"CHANDEF = "+(*i1).channel_name+eol;
        }
        disp+=(string)"PRIVATE = private"+eol;

        disp+=lang_get_string(1,(*i).lang,422)+eol;

        ec=0;
        dcc_send_msg((*i).socket,disp,ec,(*i).ssl);
        (*i).replication++;
        return;
    }

    if((*i).replication==2) {
        string p=dcc_get_line((*i).in_buff);
        if(!p.compare(""))
            return;

        if(!p.compare(".")) {
            (*i).replication=0;
            return;
        }

        got=false;

        {
            deque<c_user>::iterator i1;
            for(i1=r_users.begin(); i1!=r_users.end(); ++i1)
                if(!(*i1).name.compare(p)) {
                    got=true;
                    (*i).b_repl_user=true;
                    (*i).repl_user=*i1;
                    (*i).replication=1000;
                    return;
                }
        }
        {
            deque<c_user>::iterator i1;
            for(i1=r_user_templates.begin(); i1!=r_user_templates.end(); ++i1)
                if(!(*i1).name.compare(p)) {
                    got=true;
                    (*i).b_repl_user_template=true;
                    (*i).repl_user_template=*i1;
                    (*i).replication=1000;
                    return;
                }
        }
        {
            c_module* core_mod=get_module("core",true);
            if(core_mod) {
                vector<c_procedure>::iterator i1;
                for(i1=core_mod->r_procedures.begin(); i1!=core_mod->r_procedures.end(); ++i1)
                    if(!(*i1).name.compare(p)) {
                        got=true;
                        (*i).b_repl_proc=true;
                        (*i).repl_proc=*i1;
                        (*i).replication=3000;
                        return;
                    }
            }
        }
        {
            c_module* core_mod=get_module("core",true);
            if(core_mod) {
                map<string,c_function*>::iterator i1;
                for(i1=core_mod->r_functions.functions.begin(); i1!=core_mod->r_functions.functions.end(); ++i1)
                    if(!(*i1).second->declaration.compare(p)) {
                        got=true;
                        (*i).b_repl_func=true;
                        MY_DELETE((*i).repl_func);
                        (*i).repl_func=MY_NEW c_function(*(*i1).second);
                        (*i).replication=6000;
                        return;
                    }
            }
        }
        {
            vector<c_channel_def>::iterator i1;
            for(i1=r_channel_defs.begin(); i1!=r_channel_defs.end(); ++i1)
                if(!(*i1).channel_name.compare(p)) {
                    got=true;
                    (*i).b_repl_chandef=true;
                    (*i).repl_chandef=*i1;
                    (*i).replication=4000;
                    return;
                }
        }

        if(!p.compare("private")) {
            got=true;
            (*i).b_repl_private=true;
            (*i).repl_private.replication=r_private.replication;
            (*i).replication=5000;
            return;
        }

        disp=lang_get_string(1,(*i).lang,423)+eol;
        ec=0;
        dcc_send_msg((*i).socket,disp,ec,(*i).ssl);
        (*i).replication=1;
        return;
    }

    // user or user template
    if((*i).replication==1000) {
        disp=lang_get_string(1,(*i).lang,424)+eol;

        int line=0;

        vector<c_replication>::iterator i1;
        for(i1=(*i).repl_user.replication.begin(); i1!=(*i).repl_user.replication.end(); ++i1) {
            disp+=lang_get_string(1,(*i).lang,593);
            if(line<100)
                disp+="0";
            if(line<10)
                disp+="0";
            disp+=tos(++line);
            disp+=": ";
            disp+=(*i1).partner;
            disp+=" = ";
            if((*i1).type==REPL_PUSH)
                disp+="PUSH";
            if((*i1).type==REPL_PULL)
                disp+="PULL";
            if((*i1).type==REPL_PUSHPULL)
                disp+="PUSHPULL";
            disp+=eol;
        }
        if(line==0)
            disp+=lang_get_string(1,(*i).lang,741)+eol;

        disp+=lang_get_string(1,(*i).lang,425)+eol;

        ec=0;
        dcc_send_msg((*i).socket,disp,ec,(*i).ssl);

        (*i).replication++;
        return;
    }
    if((*i).replication==1001) {
        string p=dcc_get_line((*i).in_buff);
        if(!p.compare(""))
            return;

        if(!p.compare(".")) {
            disp=lang_get_string(1,(*i).lang,426)+eol;
            ec=0;
            dcc_send_msg((*i).socket,disp,ec,(*i).ssl);
            (*i).replication++;
            return;
        }

        if(!p.compare(".add")) {
            disp=lang_get_string(1,(*i).lang,428)+eol;

            deque<c_user>::iterator i1;
            for(i1=r_users.begin(); i1!=r_users.end(); ++i1)
                disp+=(*i1).name+eol;

            disp+=lang_get_string(1,(*i).lang,429)+eol;

            ec=0;
            dcc_send_msg((*i).socket,disp,ec,(*i).ssl);
            (*i).replication=10000;
            return;
        }

        string p1, p2;
        {
            int pos=0;
            for(unsigned int i1=0; i1<p.length(); ++i1) {
                if(p[i1]==0x20) {
                    pos=1;
                    continue;
                }
                if(pos==0)
                    p1+=p[i1];
                if(pos==1)
                    p2+=p[i1];
            }
        }
        if(!p1.compare(".remove")) {
            int line=0;

            vector<c_replication>::iterator i1;
            for(i1=(*i).repl_user.replication.begin(); i1!=(*i).repl_user.replication.end(); ++i1)
                if(++line==stoi(p2)) {
                    (*i).repl_user.replication.erase(i1);
                    break;
                }
            (*i).replication=1000;
            return;
        }

        (*i).replication=1000;
    }
    if((*i).replication==1002) {
        string p=dcc_get_line((*i).in_buff);
        if(!p.compare(""))
            return;

        if(p.compare("y") && p.compare("n") && p.compare("Y") && p.compare("N")) {
            disp=lang_get_string(1,(*i).lang,426)+eol;
            ec=0;
            dcc_send_msg((*i).socket,disp,ec,(*i).ssl);
        } else {
            if(!p.compare("y") || !p.compare("Y")) {
                c_user u;
                vector<c_channel> chs;
                if(!(*i).b_repl_user_template) {
                    logic_partyline_get_user((*i).repl_user.name,u,chs);
                } else {
                    logic_partyline_get_user_template((*i).repl_user.name,u,chs);
                    u.is_template=true;
                }
                logic_partyline_remove_user((*i).repl_user.name);
                my_time(&(*i).repl_user.last_changed);
                logic_partyline_add_user((*i).repl_user,chs);

                disp=lang_get_string(1,(*i).lang,427)+eol;
                ec=0;
                dcc_send_msg((*i).socket,disp,ec,(*i).ssl);
                (*i).replication=1;
                return;
            } else {
                disp=lang_get_string(1,(*i).lang,162)+eol;
                ec=0;
                dcc_send_msg((*i).socket,disp,ec,(*i).ssl);
                (*i).replication=1;
                return;
            }
        }
    }

    // procedure
    if((*i).replication==3000) {
        disp=lang_get_string(1,(*i).lang,424)+eol;

        int line=0;

        vector<c_replication>::iterator i1;
        for(i1=(*i).repl_proc.replication.begin(); i1!=(*i).repl_proc.replication.end(); ++i1) {
            disp+=lang_get_string(1,(*i).lang,593);
            if(line<100)
                disp+="0";
            if(line<10)
                disp+="0";
            disp+=tos(++line);
            disp+=": ";
            disp+=(*i1).partner;
            disp+=" = ";
            if((*i1).type==REPL_PUSH)
                disp+="PUSH";
            if((*i1).type==REPL_PULL)
                disp+="PULL";
            if((*i1).type==REPL_PUSHPULL)
                disp+="PUSHPULL";
            disp+=eol;
        }
        if(line==0)
            disp+=lang_get_string(1,(*i).lang,741)+eol;

        disp+=lang_get_string(1,(*i).lang,425)+eol;

        ec=0;
        dcc_send_msg((*i).socket,disp,ec,(*i).ssl);

        (*i).replication++;
        return;
    }
    if((*i).replication==3001) {
        string p=dcc_get_line((*i).in_buff);
        if(!p.compare(""))
            return;

        if(!p.compare(".")) {
            disp=lang_get_string(1,(*i).lang,426)+eol;
            ec=0;
            dcc_send_msg((*i).socket,disp,ec,(*i).ssl);
            (*i).replication++;
            return;
        }

        if(!p.compare(".add")) {
            disp=lang_get_string(1,(*i).lang,428)+eol;

            deque<c_user>::iterator i1;
            for(i1=r_users.begin(); i1!=r_users.end(); ++i1)
                disp+=(*i1).name+eol;

            disp+=lang_get_string(1,(*i).lang,429)+eol;

            ec=0;
            dcc_send_msg((*i).socket,disp,ec,(*i).ssl);
            (*i).replication=10000;
            return;
        }

        string p1, p2;
        {
            int pos=0;
            for(unsigned int i1=0; i1<p.length(); ++i1) {
                if(p[i1]==0x20) {
                    pos=1;
                    continue;
                }
                if(pos==0)
                    p1+=p[i1];
                if(pos==1)
                    p2+=p[i1];
            }
        }
        if(!p1.compare(".remove")) {
            int line=0;

            vector<c_replication>::iterator i1;
            for(i1=(*i).repl_proc.replication.begin(); i1!=(*i).repl_proc.replication.end(); ++i1)
                if(++line==stoi(p2)) {
                    (*i).repl_proc.replication.erase(i1);
                    break;
                }
            (*i).replication=3000;
            return;
        }

        (*i).replication=3000;
    }
    if((*i).replication==3002) {
        string p=dcc_get_line((*i).in_buff);
        if(!p.compare(""))
            return;

        if(p.compare("y") && p.compare("n") && p.compare("Y") && p.compare("N")) {
            disp=lang_get_string(1,(*i).lang,426)+eol;
            ec=0;
            dcc_send_msg((*i).socket,disp,ec,(*i).ssl);
        } else {
            if(!p.compare("y") || !p.compare("Y")) {
                c_module* core_mod=get_module("core",true);
                if(core_mod) {
                    vector<c_procedure>::iterator i1;
                    for(i1=core_mod->r_procedures.begin(); i1!=core_mod->r_procedures.end(); ++i1) {
                        if(!(*i1).name.compare((*i).repl_proc.name)) {
                            (*i1).replication.clear();
                            vector<c_replication>::iterator i2;
                            for(i2=(*i).repl_proc.replication.begin(); i2!=(*i).repl_proc.replication.end(); ++i2)
                                (*i1).replication.push_back(*i2);
                            break;
                        }
                    }
                }

                disp=lang_get_string(1,(*i).lang,427)+eol;
                ec=0;
                dcc_send_msg((*i).socket,disp,ec,(*i).ssl);
                (*i).replication=1;
                return;
            } else {
                disp=lang_get_string(1,(*i).lang,162)+eol;
                ec=0;
                dcc_send_msg((*i).socket,disp,ec,(*i).ssl);
                (*i).replication=1;
                return;
            }
        }
    }

    // chandef
    if((*i).replication==4000) {
        disp=lang_get_string(1,(*i).lang,424)+eol;

        int line=0;

        vector<c_replication>::iterator i1;
        for(i1=(*i).repl_chandef.replication.begin(); i1!=(*i).repl_chandef.replication.end(); ++i1) {
            disp+=lang_get_string(1,(*i).lang,593);
            if(line<100)
                disp+="0";
            if(line<10)
                disp+="0";
            disp+=tos(++line);
            disp+=": ";
            disp+=(*i1).partner;
            disp+=" = ";
            if((*i1).type==REPL_PUSH)
                disp+="PUSH";
            if((*i1).type==REPL_PULL)
                disp+="PULL";
            if((*i1).type==REPL_PUSHPULL)
                disp+="PUSHPULL";
            disp+=eol;
        }
        if(line==0)
            disp+=lang_get_string(1,(*i).lang,741)+eol;

        disp+=lang_get_string(1,(*i).lang,425)+eol;

        ec=0;
        dcc_send_msg((*i).socket,disp,ec,(*i).ssl);

        (*i).replication++;
        return;
    }
    if((*i).replication==4001) {
        string p=dcc_get_line((*i).in_buff);
        if(!p.compare(""))
            return;

        if(!p.compare(".")) {
            disp=lang_get_string(1,(*i).lang,426)+eol;
            ec=0;
            dcc_send_msg((*i).socket,disp,ec,(*i).ssl);
            (*i).replication++;
            return;
        }

        if(!p.compare(".add")) {
            disp=lang_get_string(1,(*i).lang,428)+eol;

            deque<c_user>::iterator i1;
            for(i1=r_users.begin(); i1!=r_users.end(); ++i1)
                disp+=(*i1).name+eol;

            disp+=lang_get_string(1,(*i).lang,429)+eol;

            ec=0;
            dcc_send_msg((*i).socket,disp,ec,(*i).ssl);
            (*i).replication=10000;
            return;
        }

        string p1, p2;
        {
            int pos=0;
            for(unsigned int i1=0; i1<p.length(); ++i1) {
                if(p[i1]==0x20) {
                    pos=1;
                    continue;
                }
                if(pos==0)
                    p1+=p[i1];
                if(pos==1)
                    p2+=p[i1];
            }
        }
        if(!p1.compare(".remove")) {
            int line=0;

            vector<c_replication>::iterator i1;
            for(i1=(*i).repl_chandef.replication.begin(); i1!=(*i).repl_chandef.replication.end(); ++i1)
                if(++line==stoi(p2)) {
                    (*i).repl_chandef.replication.erase(i1);
                    break;
                }
            (*i).replication=4000;
            return;
        }

        (*i).replication=4000;
    }
    if((*i).replication==4002) {
        string p=dcc_get_line((*i).in_buff);
        if(!p.compare(""))
            return;

        if(p.compare("y") && p.compare("n") && p.compare("Y") && p.compare("N")) {
            disp=lang_get_string(1,(*i).lang,426)+eol;
            ec=0;
            dcc_send_msg((*i).socket,disp,ec,(*i).ssl);
        } else {
            if(!p.compare("y") || !p.compare("Y")) {
                vector<c_channel_def>::iterator i1;
                for(i1=r_channel_defs.begin(); i1!=r_channel_defs.end(); ++i1) {
                    if(!(*i1).channel_name.compare((*i).repl_chandef.channel_name)) {
                        (*i1).replication.clear();
                        vector<c_replication>::iterator i2;
                        for(i2=(*i).repl_chandef.replication.begin(); i2!=(*i).repl_chandef.replication.end(); ++i2)
                            (*i1).replication.push_back(*i2);
                        break;
                    }
                }

                disp=lang_get_string(1,(*i).lang,427)+eol;
                ec=0;
                dcc_send_msg((*i).socket,disp,ec,(*i).ssl);
                (*i).replication=1;
                return;
            } else {
                disp=lang_get_string(1,(*i).lang,162)+eol;
                ec=0;
                dcc_send_msg((*i).socket,disp,ec,(*i).ssl);
                (*i).replication=1;
                return;
            }
        }
    }

    // private
    if((*i).replication==5000) {
        disp=lang_get_string(1,(*i).lang,424)+eol;

        int line=0;

        vector<c_replication>::iterator i1;
        for(i1=(*i).repl_private.replication.begin(); i1!=(*i).repl_private.replication.end(); ++i1) {
            disp+=lang_get_string(1,(*i).lang,593);
            if(line<100)
                disp+="0";
            if(line<10)
                disp+="0";
            disp+=tos(++line);
            disp+=": ";
            disp+=(*i1).partner;
            disp+=" = ";
            if((*i1).type==REPL_PUSH)
                disp+="PUSH";
            if((*i1).type==REPL_PULL)
                disp+="PULL";
            if((*i1).type==REPL_PUSHPULL)
                disp+="PUSHPULL";
            disp+=eol;
        }
        if(line==0)
            disp+=lang_get_string(1,(*i).lang,741)+eol;

        disp+=lang_get_string(1,(*i).lang,425)+eol;

        ec=0;
        dcc_send_msg((*i).socket,disp,ec,(*i).ssl);

        (*i).replication++;
        return;
    }
    if((*i).replication==5001) {
        string p=dcc_get_line((*i).in_buff);
        if(!p.compare(""))
            return;

        if(!p.compare(".")) {
            disp=lang_get_string(1,(*i).lang,426)+eol;
            ec=0;
            dcc_send_msg((*i).socket,disp,ec,(*i).ssl);
            (*i).replication++;
            return;
        }

        if(!p.compare(".add")) {
            disp=lang_get_string(1,(*i).lang,428)+eol;

            deque<c_user>::iterator i1;
            for(i1=r_users.begin(); i1!=r_users.end(); ++i1)
                disp+=(*i1).name+eol;

            disp+=lang_get_string(1,(*i).lang,429)+eol;

            ec=0;
            dcc_send_msg((*i).socket,disp,ec,(*i).ssl);
            (*i).replication=10000;
            return;
        }

        string p1, p2;
        {
            int pos=0;
            for(unsigned int i1=0; i1<p.length(); ++i1) {
                if(p[i1]==0x20) {
                    pos=1;
                    continue;
                }
                if(pos==0)
                    p1+=p[i1];
                if(pos==1)
                    p2+=p[i1];
            }
        }
        if(!p1.compare(".remove")) {
            int line=0;

            vector<c_replication>::iterator i1;
            for(i1=(*i).repl_private.replication.begin(); i1!=(*i).repl_private.replication.end(); ++i1)
                if(++line==stoi(p2)) {
                    (*i).repl_private.replication.erase(i1);
                    break;
                }
            (*i).replication=5000;
            return;
        }

        (*i).replication=5000;
    }
    if((*i).replication==5002) {
        string p=dcc_get_line((*i).in_buff);
        if(!p.compare(""))
            return;

        if(p.compare("y") && p.compare("n") && p.compare("Y") && p.compare("N")) {
            disp=lang_get_string(1,(*i).lang,426)+eol;
            ec=0;
            dcc_send_msg((*i).socket,disp,ec,(*i).ssl);
        } else {
            if(!p.compare("y") || !p.compare("Y")) {
                r_private.replication.clear();
                vector<c_replication>::iterator i2;
                for(i2=(*i).repl_private.replication.begin(); i2!=(*i).repl_private.replication.end(); ++i2)
                    r_private.replication.push_back(*i2);

                disp=lang_get_string(1,(*i).lang,427)+eol;
                ec=0;
                dcc_send_msg((*i).socket,disp,ec,(*i).ssl);
                (*i).replication=1;
                return;
            } else {
                disp=lang_get_string(1,(*i).lang,162)+eol;
                ec=0;
                dcc_send_msg((*i).socket,disp,ec,(*i).ssl);
                (*i).replication=1;
                return;
            }
        }
    }

    // function
    if((*i).replication==6000) {
        disp=lang_get_string(1,(*i).lang,424)+eol;

        int line=0;

        vector<c_replication>::iterator i1;
        for(i1=(*i).repl_func->replications.begin(); i1!=(*i).repl_func->replications.end(); ++i1) {
            disp+=lang_get_string(1,(*i).lang,593);
            if(line<100)
                disp+="0";
            if(line<10)
                disp+="0";
            disp+=tos(++line);
            disp+=": ";
            disp+=(*i1).partner;
            disp+=" = ";
            if((*i1).type==REPL_PUSH)
                disp+="PUSH";
            if((*i1).type==REPL_PULL)
                disp+="PULL";
            if((*i1).type==REPL_PUSHPULL)
                disp+="PUSHPULL";
            disp+=eol;
        }
        if(line==0)
            disp+=lang_get_string(1,(*i).lang,741)+eol;

        disp+=lang_get_string(1,(*i).lang,425)+eol;

        ec=0;
        dcc_send_msg((*i).socket,disp,ec,(*i).ssl);

        (*i).replication++;
        return;
    }
    if((*i).replication==6001) {
        string p=dcc_get_line((*i).in_buff);
        if(!p.compare(""))
            return;

        if(!p.compare(".")) {
            disp=lang_get_string(1,(*i).lang,426)+eol;
            ec=0;
            dcc_send_msg((*i).socket,disp,ec,(*i).ssl);
            (*i).replication++;
            return;
        }

        if(!p.compare(".add")) {
            disp=lang_get_string(1,(*i).lang,428)+eol;

            deque<c_user>::iterator i1;
            for(i1=r_users.begin(); i1!=r_users.end(); ++i1)
                disp+=(*i1).name+eol;

            disp+=lang_get_string(1,(*i).lang,429)+eol;

            ec=0;
            dcc_send_msg((*i).socket,disp,ec,(*i).ssl);
            (*i).replication=10000;
            return;
        }

        string p1, p2;
        {
            int pos=0;
            for(unsigned int i1=0; i1<p.length(); ++i1) {
                if(p[i1]==0x20) {
                    pos=1;
                    continue;
                }
                if(pos==0)
                    p1+=p[i1];
                if(pos==1)
                    p2+=p[i1];
            }
        }
        if(!p1.compare(".remove")) {
            int line=0;

            vector<c_replication>::iterator i1;
            for(i1=(*i).repl_func->replications.begin(); i1!=(*i).repl_func->replications.end(); ++i1)
                if(++line==stoi(p2)) {
                    (*i).repl_func->replications.erase(i1);
                    break;
                }
            (*i).replication=6000;
            return;
        }

        (*i).replication=6000;
    }
    if((*i).replication==6002) {
        string p=dcc_get_line((*i).in_buff);
        if(!p.compare(""))
            return;

        if(p.compare("y") && p.compare("n") && p.compare("Y") && p.compare("N")) {
            disp=lang_get_string(1,(*i).lang,426)+eol;
            ec=0;
            dcc_send_msg((*i).socket,disp,ec,(*i).ssl);
        } else {
            if(!p.compare("y") || !p.compare("Y")) {
                c_module* core_mod=get_module("core",true);
                if(core_mod) {
                    map<string,c_function*>::iterator i1;
                    for(i1=core_mod->r_functions.functions.begin(); i1!=core_mod->r_functions.functions.end(); ++i1) {
                        if(!(*i1).second->declaration.compare((*i).repl_func->declaration)) {
                            (*i1).second->replications.clear();
                            vector<c_replication>::iterator i2;
                            for(i2=(*i).repl_func->replications.begin(); i2!=(*i).repl_func->replications.end(); ++i2)
                                (*i1).second->replications.push_back(*i2);
                            break;
                        }
                    }
                }

                disp=lang_get_string(1,(*i).lang,427)+eol;
                ec=0;
                dcc_send_msg((*i).socket,disp,ec,(*i).ssl);
                (*i).replication=1;
                return;
            } else {
                disp=lang_get_string(1,(*i).lang,162)+eol;
                ec=0;
                dcc_send_msg((*i).socket,disp,ec,(*i).ssl);
                (*i).replication=1;
                return;
            }
        }
    }

    // .add (universal)
    if((*i).replication==10000) {
        string p=dcc_get_line((*i).in_buff);
        if(!p.compare(""))
            return;

        got=false;
        deque<c_user>::iterator i1;
        for(i1=r_users.begin(); i1!=r_users.end(); ++i1) {
            if(!(*i1).name.compare(p)) {
                got=true;
                break;
            }
        }

        if(!got) {
            disp=lang_get_string(1,(*i).lang,428)+eol;

            deque<c_user>::iterator i1;
            for(i1=r_users.begin(); i1!=r_users.end(); ++i1)
                disp+=(*i1).name+eol;

            disp+=lang_get_string(1,(*i).lang,429)+eol;

            ec=0;
            dcc_send_msg((*i).socket,disp,ec,(*i).ssl);
            (*i).replication=(*i).replication;
            return;
        }

        (*i).repl_add.partner=p;
        (*i).repl_add.type=REPL_INVALID;

        disp=lang_get_string(1,(*i).lang,430)+eol;
        ec=0;
        dcc_send_msg((*i).socket,disp,ec,(*i).ssl);
        (*i).replication++;
        return;
    }
    if((*i).replication==10001) {
        string p=dcc_get_line((*i).in_buff);
        if(!p.compare(""))
            return;

        if(!p.compare("PUSH"))
            (*i).repl_add.type=REPL_PUSH;
        if(!p.compare("PULL"))
            (*i).repl_add.type=REPL_PULL;
        if(!p.compare("PUSHPULL"))
            (*i).repl_add.type=REPL_PUSHPULL;

        if(!p.compare(".")) {
            (*i).replication+=2;
            return;
        }

        (*i).repl_add.type==REPL_INVALID?(*i).replication--:(*i).replication++;

        if((*i).repl_add.type==REPL_INVALID) {
            disp=lang_get_string(1,(*i).lang,430)+eol;
            ec=0;
            dcc_send_msg((*i).socket,disp,ec,(*i).ssl);
            return;
        }

        return;
    }
    if((*i).replication==10002) {
        c_user u;
        vector<c_channel> chs;
        logic_partyline_get_user((*i).repl_add.partner,u,chs);
        if(!u.replication_partner) {
            u.replication_partner=true;
            logic_partyline_remove_user(u.name);
            my_time(&u.last_changed);
            logic_partyline_add_user(u,chs);

            disp=lang_get_string(1,(*i).lang,431)+eol;
            ec=0;
            dcc_send_msg((*i).socket,disp,ec,(*i).ssl);
        } else {
            disp=lang_get_string(1,(*i).lang,432)+eol;
            ec=0;
            dcc_send_msg((*i).socket,disp,ec,(*i).ssl);
        }
        (*i).replication++;
        return;
    }
    if((*i).replication==10003) {
        if((*i).b_repl_user) {
            (*i).repl_user.replication.push_back((*i).repl_add);
            (*i).replication=1000;
        }
        if((*i).b_repl_proc) {
            (*i).repl_proc.replication.push_back((*i).repl_add);
            (*i).replication=3000;
        }
        if((*i).b_repl_chandef) {
            (*i).repl_chandef.replication.push_back((*i).repl_add);
            (*i).replication=4000;
        }
        if((*i).b_repl_private) {
            (*i).repl_private.replication.push_back((*i).repl_add);
            (*i).replication=5000;
        }
        if((*i).b_repl_func) {
            (*i).repl_func->replications.push_back((*i).repl_add);
            (*i).replication=6000;
        }
        return;
    }
}

