/***************************************************************************
                          debug1.cpp  -  description
                             -------------------
    begin                : Mon Dec 25 2006
    copyright            : (C) 2006-2007 by Marian VooDooMan Meravy
    email                : ghostvoodooman -aT= users (dot) sourceforge.net
 ***************************************************************************/

/* $Id: debug1.cpp,v 1.38 2008/08/23 15:10:39 ghostvoodooman Exp $ */

/***************************************************************************

VooDoo cIRCle - an IRC (ro)bot
Copyright (C) 2006-2008 by Marian VooDooMan Meravy
  (ghostvoodooman -aT= users (dot) sourceforge.net)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************/

/*!
    \file
    \brief Implements evaluate tests (self test of evaluating procedures and functions of internal scripting)
    \version $Id: debug1.cpp,v 1.38 2008/08/23 15:10:39 ghostvoodooman Exp $
*/

#if defined(HAVE_CONFIG_H)
#   include <config.h>
#endif

#include <string>
#include <vector>
#include <stack>
#include <list>
#include <map>
#include <iostream>

#include "utils/memleak.h"
#include "utils/assert.h"

#include "params.h"
#include "compat.h"
#include "utils.h"
#include "variable.h"
#include "var_keeper.h"
#include "modules.h"
#include "eval.h"
#include "var/c_var_base.h"
#include "var/c_var_bool.h"
#include "var/c_var_int.h"
#include "var/c_var_string.h"
#include "var/c_var_array.h"
#include "var/c_var_unset.h"
#include "stats.h"

#include "utils/precomp.h"

void report_heap_debug()
{
#ifdef _DEBUG
    MY_RPT0(MY_RPT_INFO,"===== BEGIN: HEAP DEBUG INFO =====");
#else
#if REPORT_LEAKS==1
    printf("===== BEGIN: HEAP DEBUG INFO =====\n");
#endif
#endif
    list<string> lines;
    stats_display(lines,"",true);
    list<string>::iterator stats_i;
    for(stats_i=lines.begin(); stats_i!=lines.end(); ++stats_i) {
#ifdef _DEBUG
        MY_RPT0(MY_RPT_INFO,(*stats_i).c_str());
#else
#if REPORT_LEAKS==1
        printf("%s\n",(*stats_i).c_str());
#endif
#endif
    }
#ifdef _DEBUG
    MY_RPT0(MY_RPT_INFO,"===== END:   HEAP DEBUG INFO =====");
#else
#if REPORT_LEAKS==1
    printf("===== END:   HEAP DEBUG INFO =====\n");
#endif
#endif
}

#if defined(_MSC_VER)
#   pragma warning(push)
#   pragma warning(disable:4706)    // warning C4706: assignment within conditional expression
#endif

/*!
    \brief Performs evaluation self-test suite
    \author VooDooMan
    \retval true All tests passed
    \retval false There were some errors
*/
bool debug_do_evaluate_tests()
{
    //_CrtSetAllocHook(debug_alloc_hook);

    BEGIN_DEBUG_TIME_MEASUREMENT(true,1,"Evaluate tests");

    {
#if (defined(_WIN32) || defined(_WIN64)) && !defined(__CYGWIN__)
        mkdir("modules");
        mkdir("modules" FILE_SLASH "self_test");
#else
        mkdir("." FILE_SLASH "modules",0777);
        mkdir("." FILE_SLASH "modules" FILE_SLASH "self_test",0777);
#endif
        FILE* f=fopen("modules" FILE_SLASH "self_test" FILE_SLASH "logic.txt","w");
        if(f) {
            fprintf(f,"%s","#version 3\n"

                "function f9() {\n"
                "    $testarray=array(1,2,3,4,5,6,7,8,9);\n"
                "    $res=\"\";\n"
                "    $i1=0;\n"
                "    for( ; ; ++$i1) {\n"
                "        {\n"
                "            foreach($testarray as $e) {\n"
                "                if(++$i1>2)\n"
                "                {\n"
                "                    break;\n"
                "                }\n"
                "                $res.=\"a$i1, \";\n"
                "            }\n"
                "            if($i1>10)\n"
                "                break;\n"
                "            $res.=\"b$i1, \";\n"
                "        }\n"
                "    }\n"
                "    return $res;\n"
                "}\n\n"

                "function test_no_effect_expression_1() { ; /* test */ ;/*test*/; /* this is just a test whether it emmits a warning */ }\n\n"

                "function a1() {\n"
                "    $ar=array('test1' => array(0, 1, 2), 'test2' => array(3, 4, 5));\n"
                "    foreach($ar as $element)\n"
                "        a1_b($element[0]);\n"
                "}\n"

                "function a1_b($x) {    log_print($x);    }\n"

                "function f15(&$test) {\n"
                "    $test=10;\n"
                "}\n"

                "function f13() {\n"
                "    $res=\"\";\n"
                "    $ar=array();\n"
                "    $ar[0]=0;\n"
                "    $ar[1]=1;\n"
                "    $ar[]=2;\n"
                "    $ar[]=3;\n"
                "    foreach($ar as &$e)\n"
                "        $e=13;\n"
                "    foreach($ar as $e)\n"
                "        $res.=$e.\", \";\n"
                "    return $res;\n"
                "}\n"

                "\n"

                "function test_apostrophe() { $x='a\\' abc\\\\def\\n-\\n\\-'; return $x; }\n\n"

                "function test__multiple_line_str_1() {\n"
                "    $eval='test1\n"
                "           /* test2 */\n"
                "    );';\n"
                "    return $eval;\n"
                "}\n\n"

                "function test__multiple_line_str_2() {\n"
                "    $eval=\"test1\n"
                "           /* test2 */\n"
                "    );\";\n"
                "    return $eval;\n"
                "}\n\n"

                "function test__bug001() {\n"
                "    $out=preg_split('/[\\\\s,]+/',\n"
                "        \"<b>example1a: </b><b>example1b: </b><div align=left>this is a test1a</div><b>example2a: </b><b>example2b: </b><div align=left>this is a test2a</div><div align=left>this is a test2b</div>\",\n"
                "        -2,PREG_SPLIT_OFFSET_CAPTURE);\n"
                "}\n\n"

                "function test_apostrophe2() { $x='/test\\x/'; return $x; }\n\n"

                "function or_opers_bug_test() {\n"
                "    $res='';\n"
                "    $val=array(\"nick\" => \"1\", \"ident\" => \"2\", \"host\" => 3);\n"
                "    $val[\"nick\"]=NULL;\n"
                "    $val[\"ident\"]=NULL;\n"
                "    $val[\"host\"]=NULL;\n"
                "    if(!isset($val) || !isset($val[\"nick\"]) || !isset($val[\"ident\"]) || !isset($val[\"host\"]))\n"
                "        $res.=\"OK\";\n"
                "    $res.=\"OK\";\n"
                "    return $res;\n"
                "}\n\n"

                "function and_opers_bug_test() {\n"
                "    $res='';\n"
                "    $val=array(\"nick\" => \"1\", \"ident\" => \"2\", \"host\" => 3);\n"
                "    if(!isset($val) && !isset($val[\"nick\"]) && !isset($val[\"ident\"]) && !isset($val[\"host\"]))\n"
                "        $res.=\"1\";\n"
                "    $val[\"ident\"]=NULL;\n"
                "    if(!isset($val) && !isset($val[\"nick\"]) && !isset($val[\"ident\"]) && !isset($val[\"host\"]))\n"
                "        $res.=\"2\";\n"
                "    $val[\"ident\"]='x';\n"
                "    $val[\"host\"]=NULL;\n"
                "    if(!isset($val) && !isset($val[\"nick\"]) && !isset($val[\"ident\"]) && !isset($val[\"host\"]))\n"
                "        $res.=\"3\";\n"
                "    $val[\"host\"]='y';\n"
                "    if(isset($val) && isset($val[\"nick\"]) && isset($val[\"ident\"]) && isset($val[\"host\"]))\n"
                "        $res.=\"4\";\n"
                "    $val[\"ident\"]=NULL;\n"
                "    $val[\"host\"]='x';\n"
                "    if(isset($val) && isset($val[\"nick\"]) && isset($val[\"ident\"]) && isset($val[\"host\"]))\n"
                "        $res.=\"5\";\n"
                "    $res.=\"OK\";\n"
                "    return $res;\n"
                "}\n\n"

                "function xx1() {\n"
                "    $b[\"test1\"][\"x1\"]=\"@\";\n"
                "    $ch=array(\"channel\" => \"x1\", \"mode\" => \"0\");\n"
                "    $botname=\"test1\";\n"
                "    if(isset($b[$botname][$ch[\"channel\"]/i]) && strpos($ch[\"mode\"],$b[$botname][$ch[\"channel\"]/i])===FALSE) {\n"
                "        return TRUE;\n"
                "    }\n"
                "    return FALSE;\n"
                "}\n"

                "function f14() {\n"
                "    $res=\"\";\n"
                "    $ar[][][]=0;\n"
                "    $ar[0][0][]=1;\n"
                "    $ar[0][0][]=2;\n"
                "    $ar[0][0][]=3;\n"
                "    foreach($ar[0][0] as $e)\n"
                "        $res.=$e.\", \";\n"
                "    foreach($ar[0][0] as &$e)\n"
                "        $e=14;\n"
                "    foreach($ar[0][0] as $e)\n"
                "        $res.=$e.\", \";\n"
                "    return $res;\n"
                "}\n"

                "function f2($a,$b,$c)\n"
                "{\n"
                "    f1($a+10,$b?2:2,$c);\n"
                "    f1($a+10,$b?2:2?2:2,$c);\n"
                "    f1($a+10,$b,$c);\n"
                "    f1($a+10,$b,$c);\n"
                "    \n"
                "    f1($a+10,f1($b,0,$c),$c);\n"
                "    f1($a+10,f1($b,0,$c)?2:2,$c);\n"
                "    f1($a+10,f1($b,0,$c)?2:2?2:2,$c);\n"
                "    \n"
                "    f1($a+10,$b,$c);\n"
                "    f1($a+10,$b,$c);\n"
                "    f1($a+10,$b,$c);\n"
                "    \n"
                "    $c=3;\n"
                "    f1($a+10,$b,$c);\n"
                "    return $c;\n"
                "}\n\n"

                "function abcxxx() { do {$res.=\"\\$i2=\".$i2.\", \";}\n"
                "    while($i2++<20);\n }\n\n"

                "function f4($test) {\n"
                "    $res=\"\";\n"
                "    if($test) {\n"
                "        $res.=\"1\";\n"
                "    } else {\n"
                "        $res.=\"2\";\n"
                "    }\n"
                "    if($test)\n"
                "        $res.=\"3\";\n"
                "    else\n"
                "        $res.=\"4\";\n"
                "    if(!$test)\n"
                "        $res.=\"a\";\n"
                "    else if(!$test)\n"
                "        {$res.=\"b\"; if(1) $res.=\"0\"; else $res.=\"0\";}\n"
                "    else if(!$test)\n"
                "        $res.=\"x\";\n"
                "    else if(!$test)\n"
                "        {$res.=\"%\";}\n"
                "    else $res.=\"*\";\n"
                "    if($test) {\n"
                "        $res.=\"1\";\n"
                "    } else {\n"
                "        $res.=\"2\";\n"
                "    }\n"
                "    if($test)\n"
                "        $res.=\"3\";\n"
                "    else\n"
                "        $res.=\"4\";\n"
                "    if(!$test) {\n"
                "        $res.=\"a\";\n"
                "    } else {\n"
                "        $res.=\"b\";\n"
                "    }\n"
                "    if(!$test)\n"
                "        $res.=\"c\";\n"
                "    else\n"
                "        $res.=\"d\";\n"
                "    if($test)\n"
                "        $res.=\"!\";\n"
                "    else {\n"
                "        $res.=\"@\";\n"
                "    }\n"
                "    if($test)\n"
                "        $res.=\"#\";\n"
                "    else {\n"
                "        $res.=\"\\$\";\n"
                "    }\n"
                "    if(!$test) {\n"
                "        $res.=\"A\";\n"
                "    } else\n"
                "        $res.=\"B\";\n"
                "    if(!$test) {\n"
                "        $res.=\"C\";\n"
                "    } else\n"
                "        $res.=\"D\";\n"
                "    {\n"
                "        if($test) {\n"
                "            $res.=\"1\";\n"
                "            {\n"
                "                $res.=\"2\";\n"
                "                if($test)\n"
                "                    $res.=\"3\";\n"
                "                else\n"
                "                    $res.=\"4\";\n"
                "            }\n"
                "        }\n"
                "        if(!$test) {\n"
                "            $res.=\"a\";\n"
                "        } else if($test) {\n"
                "            $res.=\"b\";\n"
                "        } else if(!$test) {\n"
                "            $res.=\"x\";\n"
                "        }\n"
                "        if(!$test)\n"
                "            $res.=\"a\";\n"
                "        else if($test)\n"
                "            $res.=\"b\";\n"
                "        else if(!$test)\n"
                "            $res.=\"x\";\n"
                "        if(!$test)\n"
                "            $res.=\"a\";\n"
                "        else if(!$test)\n"
                "            $res.=\"b\";\n"
                "        else if(!$test)\n"
                "            $res.=\"x\";\n"
                "    }\n"
                "    if(!$test)\n"
                "        return $res;\n"
                "    else {\n"
                "        $res.=\"E\";\n"
                "        return $res;\n"
                "    }\n"
                "}\n\n"

                "function f12() {\n"
                "    $res=\"\";\n"
                "    for($i1=0; $i1<5; ++$i1) {\n"
                "        $res.=$i1.\", \";\n"
                "        if(0==$i1)\n"
                "            continue;\n"
                "        $res.=$i1.\", \";\n"
                "        if(1==$i1)\n"
                "            continue;\n"
                "        $res.=$i1.\", \";\n"
                "        if(2==$i1)\n"
                "            continue;\n"
                "        $res.=$i1.\", \";\n"
                "        if(3==$i1)\n"
                "            continue;\n"
                "        $res.=$i1.\", \";\n"
                "    }\n"
                "    $res.=\"done #1; \";\n"
                "    while($i1++<100) {\n"
                "        $res.=$i1.\", \";\n"
                "        if(7==$i1)\n"
                "            continue;\n"
                "        $res.=$i1.\", \";\n"
                "        if(10==$i1)\n"
                "            break;\n"
                "    }\n"
                "    return $res;\n"
                "}\n\n"

                "function f10($x) {\n"
                "    $res=\"\";\n"
                "    switch($x) {\n"
                "        case 1: $res.=1; break;\n"
                "        case 2: { $res.=2; } break;\n"
                "        case 3: {\n"
                "                    switch($x) {\n"
                "                        case 1: $res.=1; break;\n"
                "                        case 2: $res.=2; break;\n"
                "                        case 3: $res.=3; break;\n"
                "                    }\n"
                "                }\n"
                "        case 4: { $res.=\"3+\"; break; }\n"
                "        case 5: $res.=5; break;\n"
                "        default: $res.=\"default\"; break;\n"
                "        case 6: $res.=6; break;\n"
                "    }\n"
                "    switch($x) {\n"
                "        case 1: $res.=1; break;\n"
                "        case 2: { $res.=2; } break;\n"
                "        case 3: {\n"
                "                    switch($x) {\n"
                "                        case 1: $res.=1; break;\n"
                "                        case 2: $res.=2; break;\n"
                "                        case 3: $res.=3; break;\n"
                "                    }\n"
                "                    break;\n"
                "                }\n"
                "        case 4: { $res.=\"3+\"; break; }\n"
                "        case 5: $res.=5; break;\n"
                "        case 6: $res.=6; break;\n"
                "   }\n"
                "   $res.=\"!\";\n"
                "   return $res;\n"
                "}\n\n"

                "function f8() {\n"
                "    $res=\"\";\n"
                "    $x=array(\"a\" => 1, \"b\" => 2, \"c\" => 3);\n"
                "    foreach($x) $res.=\"x rules!, \";\n"
                "    foreach($x as $e) $res.=$e.\", \";\n"
                "    foreach($x as $e) { $res.=$e.\", \"; }\n"
                "    foreach($x as $key => $value) $res.=\"$key == $value, \";\n"
                "    foreach($x as $key => $value) $value=0;\n"
                "    foreach($x as $key => $value) $res.=\"$key == $value, \";\n"
                "    foreach($x as $key => &$value) $value=0;\n"
                "    foreach($x as $key => $value) $res.=\"$key == $value, \";\n"
                "    foreach($x as $key => &$value) {\n"
                "        for($i1=0; $i1<4;) {\n"
                "            switch(++$i1) {\n"
                "                case 1: $res.=\"a=\".$x[\"a\"].\", \"; $value=$i1+10; break;\n"
                "                case 2: $res.=\"b=\".$x[\"b\"].\", \"; $value=$i1+10; break;\n"
                "                case 3: $res.=\"c=\".$x[\"c\"].\", \"; $value=$i1+10; break;\n"
                "            }\n"
                "        }\n"
                "    }\n"
                "    foreach($x as $e) { $res.=$e.\", \"; }\n"
                "    return $res;\n"
                "}\n\n"

                "function f6() {\n"
                "    $res=\"\";\n"
                "    $i1=0;\n"
                "    while($i1++<10)\n"
                "        $res.=\"\\$i1=\".$i1.\", \";\n"
                "    $res.=\"done #1, \";\n"
                "    while($i1++<20) {\n"
                "        $res.=\"\\$i1=\".$i1.\", \";\n"
                "    }\n"
                "    $res.=\"done #2, \";\n"
                "    while($i1++<10)\n"
                "        $res.=\"\\$i1=\".$i1.\", \";\n"
                "    $res.=\"done #3, \";\n"
                "    while($i1++<10) {\n"
                "        $res.=\"\\$i1=\".$i1.\", \";\n"
                "    }\n"
                "    $res.=\"done #4, \";\n"
                "    $i2=0;\n"
                "    while($i1++<30)\n"
                "        while($i2++<10)\n"
                "            $res.=\"\\$i1=\".$i1.\", \\$i2=\".$i2.\", \";\n"
                "    $res.=\"done #5, \";\n"
                "    do {$res.=\"\\$i2=\".$i2.\", \";}\n"
                "    while($i2++<20);\n"
                "    $res.=\"done #6, \";\n"
                "    do { $res.=\"\\$i2=\".$i2.\", \";\n"
                "    } while($i2++<30);\n"
                "    $res.=\"done #7, \";\n"
                "    $i1=$i2=0;\n"
                "    while($i1++<30) {\n"
                "        while($i2++<10)\n"
                "            $res.=\"\\$i1=\".$i1.\", \\$i2=\".$i2.\", \";\n"
                "    }\n"
                "    $res.=\"done #8, \";\n"
                "    $i2=0;\n"
                "    while($i1++<30) \n"
                "        while($i2++<10) {\n"
                "            $res.=\"\\$i1=\".$i1.\", \\$i2=\".$i2.\", \";\n"
                "        }\n"
                "    $res.=\"done #9, \";\n"
                "    $i2=0;\n"
                "    do {\n"
                "        do {\n"
                "            $res.=\"\\$i1=\".$i1.\", \\$i2=\".$i2.\", \";\n"
                "        } while($i2++<10);\n"
                "    } while($i1++<30);\n"
                "    $res.=\"done #10.\";\n"
                "    return $res;\n"
                "}\n\n"

                "function xxx_test($test) {\n"
                "    if(!$test)\n"
                "        $res.=\"a\";\n"
                "    else \n if(!$test)\n"
                "        {\n $res.=\"b\"; \n if(1) \n $res.=\"0\"; \n else \n $res.=\"0\"; \n }\n"
                "    else \n if(!$test)\n"
                "       $test=$test;\n"
                "}\n\n"

                "function f5($cnt) {\n"
                "    $res=\"\";\n"
                "    for($i1=0; $i1<$cnt; ++$i1)\n"
                "        $res.=$i1;\n"
                "    $res.=\"Done: $cnt times.\";\n"
                "    for( ; $i1<$cnt+10; ++$i1)\n"
                "    {\n"
                "        $res.=$i1;\n"
                "    }\n"
                "    $res.=\"Done: $cnt times.\";\n"
                "    return $res;\n"
                "}\n\n"

                "function f1($a,$b,&$c)\n"
                "{\n"
                "    $c=$a+$b+$c;\n"
                "    return $a;\n"
                "}\n\n"

                "function f3($str)\n"
                "{\n"
                "    return strtoupper($str);\n"
                "}\n\n"

                "function case_rfc1459_1()\n"
                "{\n"
                "    return strtolower(\"[]\\\\^\",CASE_RFC1459)==\"{}|~\";\n"
                "}\n\n"

                "function case_ascii_1()\n"
                "{\n"
                "    return strtolower(\"[]\\\\^\",CASE_ASCII)==\"[]\\\\^\";\n"
                "}\n\n"

                "function case_rfc1459_2()\n"
                "{\n"
                "    return strtoupper(\"{}|~\",CASE_RFC1459)==\"[]\\\\^\";\n"
                "}\n\n"

                "function case_ascii_2()\n"
                "{\n"
                "    return strtoupper(\"[]\\\\^\",CASE_ASCII)==\"[]\\\\^\";\n"
                "}\n\n"

                "function f7() {\n"
                "    $res=\"\";\n"
                "    for($i1=0; $i1<5; ++$i1){\n"
                "        for($i2=0; $i2<10; ++$i2){\n"
                "            $res.=\"\\$i1=$i1, \\$i2=$i2\";\n"
                "        }\n"
                "    }\n"
                "    $res.=\"done.\";\n"
                "    return $res;\n"
                "}\n\n"

                "function f11() {\n"
                "    $test=TRUE;\n"
                "    $res=\"\";\n"
                "    if($test)\n"
                "        $res.=\"1, \";\n"
                "    else if(!$test)\n"
                "        $res.=\"2, \";\n"
                "    else if($test)\n"
                "        $res.=\"3, \";\n"
                "    return $res;\n"
                "}\n\n"

                "\n"

                "function f16() {\n"
                "    $res=\"\";\n"
                "    $ar=array(0, 1, 2);\n"
                "    if(isset($ar[0]))\n"
                "        $res.=\"ok1, \";\n"
                "    if(!isset($ar[3]))\n"
                "        $res.=\"ok2, \";\n"
                "    $ar[0]=NULL;\n"
                "    if(!isset($ar[0]))\n"
                "        $res.=\"ok3, \";\n"
                "    return $res;\n"
                "}\n"

                "\n"

                "function f17($limit)\n"
                "{\n"
                "    $res=\"\";\n"
                "    if($limit!==\"x\")\n"
                "        $ar=explode(\"|-\",\"abc|def|-ghi|-|-jkl\",$limit);\n"
                "    else\n"
                "        $ar=explode(\"|-\",\"abc|def|-ghi|-|-jkl\");\n"
                "    $res=\"\";\n"
                "    foreach($ar as $e)\n"
                "        $res.=$e.\", \";\n"
                "    return $res;\n"
                "}\n"

                "\n"

                "function f18() {\n"
                "    $res=\"\";\n"
                "    $ar=array(1, \"abcd\", 2);\n"
                "    $res.=count($ar).\", \";\n"
                "    $res.=count($ar[1]).\", \";\n"
                "    $res.=strlen($ar[1]).\", \";\n"
                "    $res.=ord($ar[1]).\", \";\n"
                "    $res.=chr(33).\", \";\n"
                "    return $res;\n"
                "}\n"

                "\n"

                "function f19() {\n"
                "    $res=\"\";\n"
                "    $first=TRUE;\n"
                "    goto test_label;\n"
                "    $res.=\"!\";\n"
                "  test_label2:\n"
                "    $first=FALSE;\n"
                "  test_label:\n"
                "    $res.=\"A\";\n"
                "    if($first)\n"
                "        goto test_label2;\n"
                "    return $res;\n"
                "}\n"

                "function f20() {\n"
                "    $x=0;\n"
                "    $x/=$x;\n"
                "}\n"

                "function f21() {\n"
                "    $res=\"\";\n"
                "    global $testing_global;\n"
                "    if(isset($testing_global))\n"
                "        $res.=\"A\";\n"
                "    $testing_global=TRUE;\n"
                "    if(isset($testing_global))\n"
                "        $res.=\"B\";\n"
                "    return $res;\n"
                "}\n"

                "function a2_a($dispatch)\n"
                "{\n"
                "    $ar=explode(\" \",$dispatch[\"line\"]);\n"
                "    switch($ar[1]) {\n"
                "        case \"op\":\n"
                "        case \"+o\":\n"
                "            if(count($ar)!=4) {\n"
                "                admin_msg(\"@telnet@\",$dispatch[\"user\"],\"Usage: .control +o #channel nick\");\n"
                "            } else\n"
                "                op(LOW,$ar[2],$ar[3]);\n"
                "            break;\n"
                "    }\n"
                "}\n\n"

                "function b1() {\n"
                "    $res=\"\";\n"
                "    $ar=array(1,2,3,4,5,6,7);\n"
                "    foreach($ar as $e) {\n"
                "        $res.=$e;\n"
                "        continue;\n"
                "        $res.=$e;\n"
                "    }\n"
                "    return $res;\n"
                "}\n\n"

                "function c2($x) {\n"
                "    non_existing_function_call(2);\n"
                "}\n\n"

                "function c1() {\n"
                "    $xxx=1234;\n"
                "    c2(1);\n"
                "}\n\n"

                "function d2(&$x) {\n"
                "    non_existing_function_call(2);\n"
                "}\n\n"

                "function d1() {\n"
                "    $xxx=1234;\n"
                "    d2(1);\n"
                "}\n\n"

                "function e2(&$x) {\n"
                "    non_existing_function_call(2);\n"
                "}\n\n"

                "function e1() {\n"
                "    $xxx=1234;\n"
                "    $arg=1;\n"
                "    e2($arg);\n"
                "}\n\n"

                "procedure g1()\n"
                "{\n"
                "    $x2=set 10\n"
                "    $x=eval \"return 2+3+$x2;\"\n"
                "    LOG \"2+3+10 == \".$x\n"
                "}\n\n"

                "procedure h2($x)\n"
                "{\n"
                "    LOG $x\n"
                "    execute h1($x)\n"
                "    LOG \"Error: code should not go there!\"\n"
                "}\n\n"

                "function h1($x)\n"
                "{\n"
                "    $x++;\n"
                "    if($x>200)\n"
                "        return;\n"
                "    h2($x);\n"
                "}\n\n"

                "procedure j2($x)\n"
                "{\n"
                "    LOG $x\n"
                "    execute j1($x)\n"
                "}\n\n"

                "function j1($x)\n"
                "{\n"
                "    $x=(int)$x;\n"
                "    $x++;\n"
                "    if($x>200)\n"
                "        return;\n"
                "    j2($x);\n"
                "}\n\n"

                "function k2($a,$b)\n"
                "{\n"
                "    log_print($a);\n"
                "    log_print($b);\n"
                "}\n\n"

                "procedure k1($x)\n"
                "{\n"
                "    execute k2($x,$y)\n"
                "}\n\n"

                //"function m2($a1,$a2,$a3,$a4,$a5,$a6);\n"

                "function m1($dispatch) {\n"
                "    m2(\"php_2\",3,$dispatch[\"channel\"],$dispatch[\"meta\"][\"source\"],$dispatch[\"meta\"][\"source_nick\"],$dispatch[\"meta\"][\"msg\"]);\n"
                "}\n\n"

                "function m0() {\n"
                "    $dispatch[\"channel\"]='chan';    $dispatch[\"meta\"][\"source\"]='src';     $dispatch[\"meta\"][\"source_nick\"]='nick';    $dispatch[\"meta\"][\"msg\"]='msg';\n"
                "    m1($dispatch);\n"
                "}\n\n"

                "function n0() {\n"
                "    $x[2][3]=1;\n"
                "    n1($x);\n"
                "}\n\n"

                "function test__preg_grep() {\n"
                "    $array=array('abc tes', 'tes', 'XX' => 'how do you TeSt the case?', 23, 'how do you test?');\n"
                "    $res[]=preg_grep('/test/',$array);\n"
                "    $res[]=preg_grep('/test/i',$array);\n"
                "    $res[]=preg_grep('/test/',$array,PREG_GREP_INVERT);\n"
                "    $res[]=preg_grep('/test/i',$array,PREG_GREP_INVERT);\n"
                "\n"
                "    $ec1=preg_last_error();\n"
                "    $ec2=preg_last_error($errstr);\n"
                "    $correct=array(\n"
                "                   0 => array(3 => 'how do you test?'),\n"
                "                   1 => array('XX' => 'how do you TeSt the case?',\n"
                "                              3 => 'how do you test?'),\n"
                "                   2 => array('XX' => 'how do you TeSt the case?',\n"
                "                              0 => 'abc tes',\n"
                "                              1 => 'tes',\n"
                "                              2 => '23'),\n"
                "                   3 => array(0 => 'abc tes',\n"
                "                              1 => 'tes',\n"
                "                              2 => '23')\n"
                "                   );\n"
                "    return $res === $correct && $ec1===PREG_NO_ERROR && $ec2===PREG_NO_ERROR && $errstr==='No error';\n"
                "}\n\n"

                "function test__preg_grep_bad_pattern1() {\n"
                "    $array=array('abc tes', 'tes', 'XX' => 'how do you TeSt the case?', 23, 'how do you test?');\n"
                "    $res[]=preg_grep('test/',$array);\n"
                "    return 1;\n"
                "}\n\n"

                "function test__preg_grep_bad_pattern2() {\n"
                "    $array=array('abc tes', 'tes', 'XX' => 'how do you TeSt the case?', 23, 'how do you test?');\n"
                "    $res[]=preg_grep('/test/',$array);\n"
                "    $res[]=preg_grep('/test/www',$array);\n"
                "    return 2;\n"
                "}\n\n"

                "function test__preg_grep_bad_pattern3() {\n"
                "    $array=array('abc tes', 'tes', 'XX' => 'how do you TeSt the case?', 23, 'how do you test?');\n"
                "    $res[]=preg_grep('/te[st/',$array);\n"
                "    return 3;\n"
                "}\n\n"

                "function test__preg_match_all__test1() {\n"
                "    $rc=preg_match_all(\"|<[^>]+>(.*)</[^>]+>|U\",\n"
                "        \"<b>example1a: </b><b>example1b: </b><div align=left>this is a test1a</div><b>example2a: </b><b>example2b: </b><div align=left>this is a test2a</div><div align=left>this is a test2b</div>\",\n"
                "    $out, PREG_SET_ORDER);\n"
                "\n"
                "   $correct=array(\n"
                "       array('<b>example1a: </b>','example1a: '),\n"
                "       array('<b>example1b: </b>','example1b: '),\n"
                "       array('<div align=left>this is a test1a</div>','this is a test1a'),\n"
                "       array('<b>example2a: </b>','example2a: '),\n"
                "       array('<b>example2b: </b>','example2b: '),\n"
                "       array('<div align=left>this is a test2a</div>','this is a test2a'),\n"
                "       array('<div align=left>this is a test2b</div>','this is a test2b')\n"
                "       );\n"
                "   return $rc===7 && $out===$correct;\n"
                "}\n\n"

                "function test__preg_match_all__test2() {\n"
                "    $rc=preg_match_all(\"|<[^>]+>(.*)</[^>]+>|U\",\n"
                "        \"<b>example1a: </b><b>example1b: </b><div align=left>this is a test1a</div><b>example2a: </b><b>example2b: </b><div align=left>this is a test2a</div><div align=left>this is a test2b</div>\",\n"
                "    $out, PREG_SET_ORDER | PREG_OFFSET_CAPTURE);\n"
                "\n"
                "   $correct=array(\n"
                "       array(array('<b>example1a: </b>',0),array('example1a: ',3)),\n"
                "       array(array('<b>example1b: </b>',18),array('example1b: ',21)),\n"
                "       array(array('<div align=left>this is a test1a</div>',36),array('this is a test1a',52)),\n"
                "       array(array('<b>example2a: </b>',74),array('example2a: ',77)),\n"
                "       array(array('<b>example2b: </b>',92),array('example2b: ',95)),\n"
                "       array(array('<div align=left>this is a test2a</div>',110),array('this is a test2a',126)),\n"
                "       array(array('<div align=left>this is a test2b</div>',148),array('this is a test2b',164))\n"
                "       );\n"
                "   return $rc===7 && $out===$correct;\n"
                "}\n\n"

                "function test__preg_match_all__test3() {\n"
                "    $rc=preg_match_all(\"|<[^>]+>(.*)</[^>]+>|U\",\n"
                "        \"<b>example1a: </b><b>example1b: </b><div align=left>this is a test1a</div><b>example2a: </b><b>example2b: </b><div align=left>this is a test2a</div><div align=left>this is a test2b</div>\"\n"
                "    );\n"
                "\n"
                "    return $rc===7;\n"
                "}\n\n"

                "function test__preg_match_all__test4() {\n"
                "    $rc=preg_match_all(\"|<[^>]+>(.*)</[^>]+>|U\",\n"
                "        \"<b>example1a: </b><b>example1b: </b><div align=left>this is a test1a</div><b>example2a: </b><b>example2b: </b><div align=left>this is a test2a</div><div align=left>this is a test2b</div>\",\n"
                "    $out);\n"
                "\n"
                "    $correct=array(\n"
                "        array(\n"
                "            '<b>example1a: </b>',\n"
                "            '<b>example1b: </b>',\n"
                "            '<div align=left>this is a test1a</div>',\n"
                "            '<b>example2a: </b>',\n"
                "            '<b>example2b: </b>',\n"
                "            '<div align=left>this is a test2a</div>',\n"
                "            '<div align=left>this is a test2b</div>'\n"
                "        ),\n"
                "        array(\n"
                "            'example1a: ',\n"
                "            'example1b: ',\n"
                "            'this is a test1a',\n"
                "            'example2a: ',\n"
                "            'example2b: ',\n"
                "            'this is a test2a',\n"
                "            'this is a test2b'\n"
                "        )\n"
                "    );\n"
                "   return $rc===FALSE && !isset($out);\n"
                "}\n\n"

                "function test__preg_match__test1() {\n"
                "    $rc=preg_match(\"|<[^>]+>(.*)</[^>]+>|U\",\n"
                "        \"<b>example1a: </b><b>example1b: </b><div align=left>this is a test1a</div><b>example2a: </b><b>example2b: </b><div align=left>this is a test2a</div><div align=left>this is a test2b</div>\",\n"
                "    $out);\n"
                "\n"
                "    $correct=array(\n"
                "        '<b>example1a: </b>',\n"
                "        'example1a: '\n"
                "    );\n"
                "    return $rc===1 && $out===$correct;\n"
                "}\n\n"

                "function test__preg_match__test2() {\n"
                "    $rc=preg_match(\"|<[^>]+>(.*)</[^>]+>|U\",\n"
                "        \"<b>example1a: </b><b>example1b: </b><div align=left>this is a test1a</div><b>example2a: </b><b>example2b: </b><div align=left>this is a test2a</div><div align=left>this is a test2b</div>\",\n"
                "    $out, PREG_OFFSET_CAPTURE);\n"
                "\n"
                "    $correct=array(\n"
                "      array(\n"
                "        '<b>example1a: </b>',\n"
                "        0\n"
                "      ),\n"
                "      array(\n"
                "        'example1a: ',\n"
                "        3\n"
                "      )\n"
                "    );\n"
                "    return $rc===1 && $out===$correct;\n"
                "}\n\n"

                "function test__preg_match__test3() {\n"
                "    $rc=preg_match(\"|<[^>]+>(.*)</[^>]+>|U\",\n"
                "        \"<b>example1a: </b><b>example1b: </b><div align=left>this is a test1a</div><b>example2a: </b><b>example2b: </b><div align=left>this is a test2a</div><div align=left>this is a test2b</div>\"\n"
                "    );\n"
                "\n"
                "    return $rc===1;\n"
                "}\n\n"

                "function test__preg_match__test4() {\n"
                "    $rc=preg_match(\"|<[^>]+>(.*)</[^>]+>|U\",\n"
                "        \"<b>example1a: </b><b>example1b: </b><div align=left>this is a test1a</div><b>example2a: </b><b>example2b: </b><div align=left>this is a test2a</div><div align=left>this is a test2b</div>\",\n"
                "    $out);\n"
                "\n"
                "    $correct=array(\n"
                "        '<b>example1a: </b>',\n"
                "        'example1a: '\n"
                "    );\n"
                "    return $rc===1 && $out===$correct;\n"
                "}\n\n"

                "function test__preg_match__test5() {\n"
                "    $rc=preg_match(\"|<[^>]+>(.*)</[^>]+>|U\",\n"
                "        \"<b>example1a: </b><b>example1b: </b><div align=left>this is a test1a</div><b>example2a: </b><b>example2b: </b><div align=left>this is a test2a</div><div align=left>this is a test2b</div>\",\n"
                "    $out, PREG_OFFSET_CAPTURE);\n"
                "\n"
                "    $correct=array(\n"
                "      array(\n"
                "           '<b>example1a: </b>',\n"
                "           0\n"
                "       ),\n"
                "       array(\n"
                "           'example1a: ',\n"
                "           3\n"
                "       )\n"
                "     );\n"
                "    return $rc===1 && $out===$correct;\n"
                "}\n\n"

                "function test__preg_match__test6() {\n"
                "    $rc=preg_match(\"|<[^>]+>(.*)</[^>]+>|U\",\n"
                "        \"<b>example1a: </b><b>example1b: </b><div align=left>this is a test1a</div><b>example2a: </b><b>example2b: </b><div align=left>this is a test2a</div><div align=left>this is a test2b</div>\",\n"
                "    NULL, PREG_SET_ORDER | PREG_OFFSET_CAPTURE);\n"
                "\n"
                "    return $rc;\n"
                "}\n\n"

                "function test__preg_match__test7() {\n"
                "    $rc=preg_match(\"|<[^>]+>(.*)</[^>]+>|U\",\n"
                "        \"<b>example1a: </b><b>example1b: </b><div align=left>this is a test1a</div><b>example2a: </b><b>example2b: </b><div align=left>this is a test2a</div><div align=left>this is a test2b</div>\",\n"
                "    $out, NULL);\n"
                "\n"
                "    $correct==array(\n"
                "        '<b>example1a: </b>',\n"
                "        'example1a: '\n"
                "    );\n"
                "    return $rc===1 && $out===$correct;\n"
                "}\n\n"

                "function test__preg_match__test8() {\n"
                "    $rc=preg_match(\"|<[^>]+>(.*)</[^>]+>|U\",\n"
                "        \"blah blah blah\",\n"
                "    $out, 0);\n"
                "\n"
                "    return $rc===0 && $out===array();\n"
                "}\n\n"

                "function test__preg_split__test1() {\n"
                "    $out=preg_split('/[\\\\s,]+/',\n"
                "        \"<b>example1a: </b><b>example1b: </b><div align=left>this is a test1a</div><b>example2a: </b><b>example2b: </b><div align=left>this is a test2a</div><div align=left>this is a test2b</div>\");\n"
                "\n"
                "   return $out===array(\n"
                "       '<b>example1a:',\n"
                "       '</b><b>example1b:',\n"
                "       '</b><div',\n"
                "       'align=left>this',\n"
                "       'is',\n"
                "       'a',\n"
                "       'test1a</div><b>example2a:',\n"
                "       '</b><b>example2b:',\n"
                "       '</b><div',\n"
                "       'align=left>this',\n"
                "       'is',\n"
                "       'a',\n"
                "       'test2a</div><div',\n"
                "       'align=left>this',\n"
                "       'is',\n"
                "       'a',\n"
                "       'test2b</div>'\n"
                "   );\n"
                "}\n\n"

                "function test__preg_split__test2() {\n"
                "    $out=preg_split('/[\\\\s,]+/',\n"
                "        \"<b>example1a: </b><b>example1b: </b><div align=left>this is a test1a</div><b>example2a: </b><b>example2b: </b><div align=left>this is a test2a</div><div align=left>this is a test2b</div>\",\n"
                "        0, PREG_SPLIT_OFFSET_CAPTURE);\n"
                "\n"
                "   return $out===array(\n"
                "     array(\n"
                "       '<b>example1a:',\n"
                "       0\n"
                "     ),\n"
                "     array(\n"
                "       '</b><b>example1b:',\n"
                "       14\n"
                "     ),\n"
                "     array(\n"
                "       '</b><div',\n"
                "       32\n"
                "     ),\n"
                "     array(\n"
                "       'align=left>this',\n"
                "       41\n"
                "     ),\n"
                "     array(\n"
                "       'is',\n"
                "       57\n"
                "     ),\n"
                "     array(\n"
                "       'a',\n"
                "       60\n"
                "     ),\n"
                "     array(\n"
                "       'test1a</div><b>example2a:',\n"
                "       62\n"
                "     ),\n"
                "     array(\n"
                "       '</b><b>example2b:',\n"
                "       88\n"
                "     ),\n"
                "     array(\n"
                "       '</b><div',\n"
                "       106\n"
                "     ),\n"
                "     array(\n"
                "       'align=left>this',\n"
                "       115\n"
                "     ),\n"
                "     array(\n"
                "       'is',\n"
                "       131\n"
                "     ),\n"
                "     array(\n"
                "       'a',\n"
                "       134\n"
                "     ),\n"
                "     array(\n"
                "       'test2a</div><div',\n"
                "       136\n"
                "     ),\n"
                "     array(\n"
                "       'align=left>this',\n"
                "       153\n"
                "     ),\n"
                "     array(\n"
                "       'is',\n"
                "       169\n"
                "     ),\n"
                "     array(\n"
                "       'a',\n"
                "       172\n"
                "     ),\n"
                "     array(\n"
                "       'test2b</div>',\n"
                "       174\n"
                "     )\n"
                "   );\n"
                "}\n\n"

                "function test__preg_split__test3() {\n"
                "    $out=preg_split('/[\\\\s,]+/',\n"
                "        \"<b>example1a: </b><b>example1b: </b><div align=left>this is a test1a</div><b>example2a: </b><b>example2b: </b><div align=left>this is a test2a</div><div align=left>this is a test2b</div>\",\n"
                "        10);\n"
                "\n"
                "    return $out===array(\n"
                "       '<b>example1a:',\n"
                "       '</b><b>example1b:',\n"
                "       '</b><div',\n"
                "       'align=left>this',\n"
                "       'is',\n"
                "       'a',\n"
                "       'test1a</div><b>example2a:',\n"
                "       '</b><b>example2b:',\n"
                "       '</b><div',\n"
                "       'align=left>this is a test2a</div><div align=left>this is a test2b</div>'\n"
                "   );\n"
                "}\n\n"

                "function test__preg_split__test4() {\n"
                "    $out=preg_split('/[\\\\s,]+/',\n"
                "        \"<b>example1a: </b><b>example1b: </b><div align=left>this is a test1a</div><b>example2a: </b><b>example2b: </b><div align=left>this is a test2a</div><div align=left>this is a test2b</div>\",\n"
                "        10,PREG_SPLIT_OFFSET_CAPTURE);\n"
                "\n"
                "    return $out===array(\n"
                "                     array(\n"
                "                       '<b>example1a:',\n"
                "                       0\n"
                "                     ),\n"
                "                     array(\n"
                "                       '</b><b>example1b:',\n"
                "                       14\n"
                "                     ),\n"
                "                     array(\n"
                "                       '</b><div',\n"
                "                       32\n"
                "                     ),\n"
                "                     array(\n"
                "                       'align=left>this',\n"
                "                       41\n"
                "                     ),\n"
                "                     array(\n"
                "                       'is',\n"
                "                       57\n"
                "                     ),\n"
                "                     array(\n"
                "                       'a',\n"
                "                       60\n"
                "                     ),\n"
                "                     array(\n"
                "                       'test1a</div><b>example2a:',\n"
                "                       62\n"
                "                     ),\n"
                "                     array(\n"
                "                       '</b><b>example2b:',\n"
                "                       88\n"
                "                     ),\n"
                "                     array(\n"
                "                       '</b><div',\n"
                "                       106\n"
                "                     ),\n"
                "                     array(\n"
                "                       'align=left>this is a test2a</div><div align=left>this is a test2b</div>',\n"
                "                       115\n"
                "                     )\n"
                "                   );\n"
                "}\n\n"

                "function test__preg_split__test5() {\n"
                "    $out=preg_split('/[\\\\s,]+/',\n"
                "        \"<b>example1a: </b><b>example1b: </b><div align=left>this is a test1a</div><b>example2a: </b><b>example2b: </b><div align=left>this is a test2a</div><div align=left>this is a test2b</div>\",\n"
                "        -2);\n"
                "\n"
                "    return $out===array('<b>example1a: </b><b>example1b: </b><div align=left>this is a test1a</div><b>example2a: </b><b>example2b: </b><div align=left>this is a test2a</div><div align=left>this is a test2b</div>');\n"
                "}\n\n"

                "function test__preg_split__test6() {\n"
                "    $out=preg_split('/[\\\\s,]+/',\n"
                "        \"<b>example1a: </b><b>example1b: </b><div align=left>this is a test1a</div><b>example2a: </b><b>example2b: </b><div align=left>this is a test2a</div><div align=left>this is a test2b</div>\",\n"
                "        -2,PREG_SPLIT_OFFSET_CAPTURE);\n"
                "\n"
                "    return $out===array(array('<b>example1a: </b><b>example1b: </b><div align=left>this is a test1a</div><b>example2a: </b><b>example2b: </b><div align=left>this is a test2a</div><div align=left>this is a test2b</div>',0));\n"
                "}\n\n"

                "function test__preg_split__test7() {\n"
                "    $out=preg_split('/[\\\\s,]+/',\n"
                "        \"<b>example1a: </b><b>example1b: </b><div align=left>this is a test1a</div><b>example2a: </b><b>example2b: </b><div align=left>this is a test2a</div><div align=left>this is a test2b</div>\",\n"
                "        0,PREG_SPLIT_OFFSET_CAPTURE);\n"
                "\n"
                "    return $out===array(\n"
                "                     array(\n"
                "                       '<b>example1a:',\n"
                "                       0\n"
                "                     ),\n"
                "                     array(\n"
                "                       '</b><b>example1b:',\n"
                "                       14\n"
                "                     ),\n"
                "                     array(\n"
                "                       '</b><div',\n"
                "                       32\n"
                "                     ),\n"
                "                     array(\n"
                "                       'align=left>this',\n"
                "                       41\n"
                "                     ),\n"
                "                     array(\n"
                "                       'is',\n"
                "                       57\n"
                "                     ),\n"
                "                     array(\n"
                "                       'a',\n"
                "                       60\n"
                "                     ),\n"
                "                     array(\n"
                "                       'test1a</div><b>example2a:',\n"
                "                       62\n"
                "                     ),\n"
                "                     array(\n"
                "                       '</b><b>example2b:',\n"
                "                       88\n"
                "                     ),\n"
                "                     array(\n"
                "                       '</b><div',\n"
                "                       106\n"
                "                     ),\n"
                "                     array(\n"
                "                       'align=left>this',\n"
                "                       115\n"
                "                     ),\n"
                "                     array(\n"
                "                       'is',\n"
                "                       131\n"
                "                     ),\n"
                "                     array(\n"
                "                       'a',\n"
                "                       134\n"
                "                     ),\n"
                "                     array(\n"
                "                       'test2a</div><div',\n"
                "                       136\n"
                "                     ),\n"
                "                     array(\n"
                "                       'align=left>this',\n"
                "                       153\n"
                "                     ),\n"
                "                     array(\n"
                "                       'is',\n"
                "                       169\n"
                "                     ),\n"
                "                     array(\n"
                "                       'a',\n"
                "                       172\n"
                "                     ),\n"
                "                     array(\n"
                "                       'test2b</div>',\n"
                "                       174\n"
                "                     )\n"
                "                   );\n"
                "}\n\n"

                "function test__preg_replace__test1() {\n"
                "    $string = 'April 15, 2003';\n"
                "    $pattern = '/(\\w+) (\\d+), (\\d+)/i';\n"
                "    $replacement = '${1}1,$3';\n"
                "    $out=preg_replace($pattern, $replacement, $string);\n"
                "    return $out==='April1,2003';\n"
                "}\n\n"

                "function test__preg_replace__test2() {\n"
                "    $string = 'April 15, 2003';\n"
                "    $pattern = '/(\\w+) (\\d+), (\\d+)/i';\n"
                "    $replacement = '${1}1,$2';\n"
                "    $out=preg_replace($pattern, $replacement, $string, 2);\n"
                "    return $out==='April1,2003';\n"
                "}\n\n"

                "function test__preg_replace__test3() {\n"
                "    $string = 'April 15, 2003';\n"
                "    $pattern = '/(\\w+) (\\d+), (\\d+)/i';\n"
                "    $replacement = '${1}1,$3';\n"
                "    $out=preg_replace($pattern, $replacement, $string, 2, $cnt);\n"
                "    return $out==='April1,2003' && $cnt===2;\n"
                "}\n\n"

                "function test__preg_replace__test4() {\n"
                "    $string = 'April 15, 2003';\n"
                "    $pattern = '/(\\w+) (\\d+), (\\d+)/i';\n"
                "    $replacement = '${1}1,$3';\n"
                "    $out=preg_replace($pattern, $replacement, $string, 2);\n"
                "    return TRUE;\n"
                "}\n\n"

                "function test__preg_replace__test5() {\n"
                "    $out=preg_replace(array('/\\d/', '/\\s/'), array('Aa',2), array('xp 4 to'), -1 , $cnt);\n"
                "    return $out===array('xp2Aa2to') && $cnt===3;\n"
                "}\n\n"

                "function test__preg_replace__test6() {\n"
                "    $out=preg_replace(array('/\\d/', '/\\s/'), array('',2), array('xp 4 to'), -1 , $cnt);\n"
                "    return $out===array('xp22to') && $cnt===3;\n"
                "}\n\n"

                "function test__preg_replace__test7() {\n"
                "    $out=preg_replace(array('/\\d/', '/\\s/'), array('Aa',2), array('xp 4 to','xp 4 to','xp 4 to','xp 4 to','xp 4 to'), -1 , $cnt);\n"
                "    return $out===array('xp2Aa2to','xp2Aa2to','xp2Aa2to','xp2Aa2to','xp2Aa2to') && $cnt===15;\n"
                "}\n\n"

                "function test__preg_replace__test8() {\n"
                "    $out=preg_replace(array('/\\d/', '/\\s/'), array('Aa'), array('xp 4 to','xp 4 to','xp 4 to','xp 4 to','xp 4 to'), -1 , $cnt);\n"
                "    return $out===array('xpAato','xpAato','xpAato','xpAato','xpAato') && $cnt===15;\n"
                "}\n\n"

                "function test__preg_split__test9() {\n"
                "    $out=preg_split('/\\\\n\\.\\\\n\\.\\\\n/',\n"
                "        \"1\\n2\\n3\\n.\\na\\nb\\nc\\n.\\n.\\n4\\n5\\n6\\n.\\nd\\ne\\nf\\n.\\n.\\n\");\n"
                "\n"
                "    return $out===array(\n"
                "        \"1\\n2\\n3\\n.\\na\\nb\\nc\",\n"
                "        \"4\\n5\\n6\\n.\\nd\\ne\\nf\",\n"
                "        ''\n"
                "    );\n"
                "}\n\n"

                "function test__bool_xor_prec14() {\n"
                "    $a=FALSE ^ TRUE;\n"
                "    return $a; // TRUE\n"
                "}\n\n"

                "function test__bool_and_prec20() {\n"
                "    $a=TRUE and FALSE;\n"
                "    return $a; // TRUE\n"
                "}\n\n"

                "function test__bool_or_prec22() {\n"
                "    $a=FALSE or TRUE;\n"
                "    return $a; // FALSE\n"
                "}\n\n"

                "function test__bool_and_prec20_2() {\n"
                "    return $a=TRUE and FALSE; // TRUE\n"
                "}\n\n"

                "function preg_split_empty_string_bug() {\n"
                "    $str=\"a1\\nb1\\nc1\\nd1\\na2\\nb2\\nc2\\n\\na3\\nb3\\nc3\\nd3\";\n"
                "    $parts=preg_split('/\\\\n/',$str);\n"
                "    $i2=0;\n"
                "    foreach($parts as $i3) {\n"
                "        if($i2 % 4 == 0)\n"
                "            $p[\"name\"]=$i3;\n"
                "        else if($i2 % 4 == 1)\n"
                "            $p[\"type\"]=$i3;\n"
                "        else if($i2 % 4 == 2)\n"
                "            $p[\"reason\"]=$i3;\n"
                 "        else if($i2 % 4 == 3) {\n"
                "            $p[\"time\"]=$i3;\n"
                "            $state[\"channels\"][]=$p;\n"
                "        }\n"
                "        ++$i2;\n"
                "    }\n"
                "    return $state===array(\n"
                "       'channels' => array(\n"
                "           0 => array(\n"
                "               'name' => 'a1',\n"
                "               'type' => 'b1',\n"
                "               'reason' => 'c1',\n"
                "               'time' => 'd1'\n"
                "               ),\n"
                "           1 => array(\n"
                "               'name' => 'a2',\n"
                "               'type' => 'b2',\n"
                "               'reason' => 'c2',\n"
                "               'time' => ''\n"
                "               ),\n"
                "           2 => array(\n"
                "               'name' => 'a3',\n"
                "               'type' => 'b3',\n"
                "               'reason' => 'c3',\n"
                "               'time' => 'd3'\n"
                "               )\n"
                "           )\n"
                "       );\n"
                "}\n\n"

                "function test_foreach_nulled_reference_bug() {\n"
                "    $res='';\n"
                "    $x=array();\n"
                "    $x['test1']='1';\n"
                "    $x['test2']='2';\n"
                "    foreach($x as $k => &$v) {\n"
                "        if($k=='test1')\n"
                "            $v=NULL;\n"
                "    }\n"
                "    foreach($x as $k => $v)\n"
                "        $res.=$v;\n"
                "    $res.=';'.count($x);\n"
                "    return $res;\n"
                "}\n\n"

                "function bind_ex_test_1_dispatch($d) {\n"
                "    global $bind_ex_test_1_data;\n"
                "    $bind_ex_test_1_data[]=$d;\n"
                "}\n\n"

                "function bind_ex_test_1() {\n"
                "    global $bind_ex_test_1_data, $bind_ex_test_1_eval;\n"
                "    $bind_ex_test_1_data=array();\n"
                "    $bind_ex_test_1_eval='bind_ex(\n"
                "            \\'@preg@@match@/@test1$/\\', // $type */\n"
                "            \\'@exact@testing user\\', // $user */\n"
                "            \\'@exact_ci_ascii@#mYcHanneL\\', // $channel */\n"
                "            \\'@mask_ci_ascii@*bAr*\\', // $line */\n"
                "            TRUE, // $before */\n"
                "            \\'bind_ex_test_1_dispatch\\' // $func_name */\n"
                "    ); bind_ex(\n"
                "            \\'@preg@@match@/@test1$/\\', // $type */\n"
                "            \\'@exact@testing user\\', // $user */\n"
                "            \\'@exact_ci_ascii@#mYcHanneL\\', // $channel */\n"
                "            \\'@preg@@match_all@@set_order@/o/\\', // $line */\n"
                "            TRUE, // $before */\n"
                "            \\'bind_ex_test_1_dispatch\\' // $func_name */\n"
                "    ); bind_ex(\n"
                "            \\'@preg@@match@/@test1$/\\', // $type */\n"
                "            \\'@exact@testing user\\', // $user */\n"
                "            \\'@exact_ci_ascii@#mYcHanneL\\', // $channel */\n"
                "            \\'@preg@@match_all@@set_order@@offset_capture@/o/\\', // $line */\n"
                "            TRUE, // $before */\n"
                "            \\'bind_ex_test_1_dispatch\\' // $func_name */\n"
                "    ); bind_ex(\n"
                "            \\'@preg@@match@/@test1$/\\', // $type */\n"
                "            \\'@exact@testing user\\', // $user */\n"
                "            \\'@exact_ci_ascii@@exact_ci_rfc@#mYcHanneL\\', // $channel */\n"
                "            \\'@preg@@match_all@@set_order@/o/\\', // $line */\n"
                "            TRUE, // $before */\n"
                "            \\'bind_ex_test_1_dispatch\\' // $func_name */\n"
                "    ); bind_ex(\n"
                "            \\'@preg@@match@/@test1$/\\', // $type */\n"
                "            \\'@exact@testing user\\', // $user */\n"
                "            \\'@exact_ci_ascii@#mYcHanneL\\', // $channel */\n"
                "            \\'@preg@@match_all@@set_order@@match@/o/\\', // $line */\n"
                "            TRUE, // $before */\n"
                "            \\'bind_ex_test_1_dispatch\\' // $func_name */\n"
                "    ); bind_ex(\n"
                "            \\'@preg@@match@/@test1$/\\', // $type */\n"
                "            \\'@exact@testing user\\', // $user */\n"
                "            \\'@exact_ci_ascii@#mYcHanneL\\', // $channel */\n"
                "            \\'/o/\\', // $line */\n"
                "            TRUE, // $before */\n"
                "            \\'bind_ex_test_1_dispatch\\' // $func_name */\n"
                "    ); bind_ex(\n"
                "            \\'@preg@/@test1$/\\', // $type */\n"
                "            \\'@exact@testing user\\', // $user */\n"
                "            \\'@exact_ci_ascii@#mYcHanneL\\', // $channel */\n"
                "            \\'@preg@@match_all@@set_order@/o/\\', // $line */\n"
                "            TRUE, // $before */\n"
                "            \\'bind_ex_test_1_dispatch\\' // $func_name */\n"
                "    ); bind_ex(\n"
                "            \\'@preg@@exact@/@test1$/\\', // $type */\n"
                "            \\'@exact@testing user\\', // $user */\n"
                "            \\'@exact_ci_ascii@#mYcHanneL\\', // $channel */\n"
                "            \\'@preg@@match_all@@set_order@/o/\\', // $line */\n"
                "            TRUE, // $before */\n"
                "            \\'bind_ex_test_1_dispatch\\' // $func_name */\n"
                "    ); bind_ex(\n"
                "            \\'@mask@@@*@*\\', // $type */\n"
                "            \\'@mask@*\\', // $user */\n"
                "            \\'@mask@*\\', // $channel */\n"
                "            \\'@mask@*\\', // $line */\n"
                "            TRUE, // $before */\n"
                "            \\'bind_ex_test_1_dispatch\\' // $func_name */\n"
                "    );';\n"
                "    log_print($eval);\n"
                "    eval($bind_ex_test_1_eval);\n"
                "}\n\n"

                "function bind_ex_test_1b() {\n"
                "    global $bind_ex_test_1_data, $bind_ex_test_1_eval;\n"
                "    dispatch('self@test1','testing user','#myCHANNel','foobar',array()); // should pass\n"
                "    dispatch('self@test1','testing_user','#myCHANNel','foobar',array()); // should not pass\n"
                "    dispatch('self@test','testing user','#myCHANNel','foobar',array()); // should not pass\n"
                "    log_print(var_dump($bind_ex_test_1_data));\n"
                "    $bind_ex_test_1_eval='unbind_ex(\n"
                "            \\'@preg@@match@/@test1$/\\', // $type */\n"
                "            \\'@exact@testing user\\', // $user */\n"
                "            \\'@exact_ci_ascii@#mYcHanneL\\', // $channel */\n"
                "            \\'@mask_ci_ascii@*bAr*\\', // $line */\n"
                "            TRUE, // $before */\n"
                "            \\'bind_ex_test_1_dispatch\\' // $func_name */\n"
                "    ); unbind_ex(\n"
                "            \\'@preg@@match@/@test1$/\\', // $type */\n"
                "            \\'@exact@testing user\\', // $user */\n"
                "            \\'@exact_ci_ascii@#mYcHanneL\\', // $channel */\n"
                "            \\'@preg@@match_all@@set_order@/o/\\', // $line */\n"
                "            TRUE, // $before */\n"
                "            \\'bind_ex_test_1_dispatch\\' // $func_name */\n"
                "    ); unbind_ex(\n"
                "            \\'@preg@@match@/@test1$/\\', // $type */\n"
                "            \\'@exact@testing user\\', // $user */\n"
                "            \\'@exact_ci_ascii@#mYcHanneL\\', // $channel */\n"
                "            \\'@preg@@match_all@@set_order@@offset_capture@/o/\\', // $line */\n"
                "            TRUE, // $before */\n"
                "            \\'bind_ex_test_1_dispatch\\' // $func_name */\n"
                "    ); unbind_ex(\n"
                "            \\'@preg@@match@/@test1$/\\', // $type */\n"
                "            \\'@exact@testing user\\', // $user */\n"
                "            \\'@exact_ci_ascii@@exact_ci_rfc@#mYcHanneL\\', // $channel */\n"
                "            \\'@preg@@match_all@@set_order@/o/\\', // $line */\n"
                "            TRUE, // $before */\n"
                "            \\'bind_ex_test_1_dispatch\\' // $func_name */\n"
                "    ); unbind_ex(\n"
                "            \\'@preg@@match@/@test1$/\\', // $type */\n"
                "            \\'@exact@testing user\\', // $user */\n"
                "            \\'@exact_ci_ascii@#mYcHanneL\\', // $channel */\n"
                "            \\'@preg@@match_all@@set_order@@match@/o/\\', // $line */\n"
                "            TRUE, // $before */\n"
                "            \\'bind_ex_test_1_dispatch\\' // $func_name */\n"
                "    ); unbind_ex(\n"
                "            \\'@preg@@match@/@test1$/\\', // $type */\n"
                "            \\'@exact@testing user\\', // $user */\n"
                "            \\'@exact_ci_ascii@#mYcHanneL\\', // $channel */\n"
                "            \\'/o/\\', // $line */\n"
                "            TRUE, // $before */\n"
                "            \\'bind_ex_test_1_dispatch\\' // $func_name */\n"
                "    ); unbind_ex(\n"
                "            \\'@preg@/@test1$/\\', // $type */\n"
                "            \\'@exact@testing user\\', // $user */\n"
                "            \\'@exact_ci_ascii@#mYcHanneL\\', // $channel */\n"
                "            \\'@preg@@match_all@@set_order@/o/\\', // $line */\n"
                "            TRUE, // $before */\n"
                "            \\'bind_ex_test_1_dispatch\\' // $func_name */\n"
                "    ); unbind_ex(\n"
                "            \\'@mask@@@*@*\\', // $type */\n"
                "            \\'@mask@*\\', // $user */\n"
                "            \\'@mask@*\\', // $channel */\n"
                "            \\'@mask@*\\', // $line */\n"
                "            TRUE, // $before */\n"
                "            \\'bind_ex_test_1_dispatch\\' // $func_name */\n"
                "    );';\n"
                "    eval($bind_ex_test_1_eval);\n"
                "}\n\n"

                "function bind_ex_test_1c() {\n"
                "    global $bind_ex_test_1_data, $bind_ex_test_1_eval;\n"
                "    dispatch('self@test1','testing user','#myCHANNel','foobar',array()); // should not pass\n"
                "    dispatch('self@test1','testing_user','#myCHANNel','foobar',array()); // should not pass\n"
                "    dispatch('self@test','testing user','#myCHANNel','foobar',array()); // should not pass\n"
                "    log_print(var_dump($bind_ex_test_1_data));\n"
                "    return $bind_ex_test_1_data===array(\n"
                "        0 => array(\n"
                "            'type' => 'self\x40test1',\n"
                "            'user' => 'testing user',\n"
                "            'channel' => '#myCHANNel',\n"
                "            'line' => 'foobar',\n"
                "            'before' => TRUE,\n"
                "            'meta' => array(\n"
                "                'type_matches' => array(\n"
                "                    0 => '\x40test1'\n"
                "                )\n"
                "            )\n"
                "        ),\n"
                "        1 => array(\n"
                "            'type' => 'self\x40test1',\n"
                "            'user' => 'testing user',\n"
                "            'channel' => '#myCHANNel',\n"
                "            'line' => 'foobar',\n"
                "            'before' => TRUE,\n"
                "            'meta' => array(\n"
                "                'type_matches' => array(\n"
                "                    0 => '\x40test1'\n"
                "                ),\n"
                "                'line_matches' => array(\n"
                "                    0 => array(\n"
                "                        0 => 'o'\n"
                "                    ),\n"
                "                    1 => array(\n"
                "                        0 => 'o'\n"
                "                    )\n"
                "                )\n"
                "            )\n"
                "        ),\n"
                "        2 => array(\n"
                "            'type' => 'self\x40test1',\n"
                "            'user' => 'testing user',\n"
                "            'channel' => '#myCHANNel',\n"
                "            'line' => 'foobar',\n"
                "            'before' => TRUE,\n"
                "            'meta' => array(\n"
                "                'type_matches' => array(\n"
                "                    0 => '\x40test1'\n"
                "                ),\n"
                "                'line_matches' => array(\n"
                "                    0 => array(\n"
                "                        0 => array(\n"
                "                            0 => 'o',\n"
                "                            1 => 1\n"
                "                        )\n"
                "                    ),\n"
                "                    1 => array(\n"
                "                        0 => array(\n"
                "                            0 => 'o',\n"
                "                            1 => 2\n"
                "                        )\n"
                "                    )\n"
                "                )\n"
                "            )\n"
                "        ),\n"
                "        3 => array(\n"
                "            'type' => 'self\x40test1',\n"
                "            'user' => 'testing user',\n"
                "            'channel' => '#myCHANNel',\n"
                "            'line' => 'foobar',\n"
                "            'before' => TRUE,\n"
                "            'meta' => array(\n"
                "                'type_matches' => array(\n"
                "                    0 => '\x40test1'\n"
                "                ),\n"
                "                'line_matches' => array(\n"
                "                    0 => array(\n"
                "                        0 => 'o'\n"
                "                    ),\n"
                "                    1 => array(\n"
                "                        0 => 'o'\n"
                "                    )\n"
                "                )\n"
                "            )\n"
                "        ),\n"
                "        4 => array(\n"
                "            'type' => 'self\x40test1',\n"
                "            'user' => 'testing user',\n"
                "            'channel' => '#myCHANNel',\n"
                "            'line' => 'foobar',\n"
                "            'before' => TRUE,\n"
                "            'meta' => array(\n"
                "            )\n"
                "        ),\n"
                "        5 => array(\n"
                "            'type' => 'self\x40test1',\n"
                "            'user' => 'testing_user',\n"
                "            'channel' => '#myCHANNel',\n"
                "            'line' => 'foobar',\n"
                "            'before' => TRUE,\n"
                "            'meta' => array(\n"
                "            )\n"
                "        ),\n"
                "        6 => array(\n"
                "            'type' => 'self\x40test',\n"
                "            'user' => 'testing user',\n"
                "            'channel' => '#myCHANNel',\n"
                "            'line' => 'foobar',\n"
                "            'before' => TRUE,\n"
                "            'meta' => array(\n"
                "            )\n"
                "        )\n"
                "    );\n"
                "}\n\n"

                "\n"

                "#eof\n"

            );
            fclose(f);
            f=NULL;
        }
        f=fopen("modules" FILE_SLASH "self_test" FILE_SLASH "conf.txt","w");
        if(f) {
            fprintf(f,"%s",
                "log_echo_debug=1\n"
                "log_echo_irc=1\n"
                "log_echo_socket=1\n"
                "log_echo_bot=1\n"
                "log_echo_identd=1\n"
                "log_echo_botnet=1\n"
                "log_echo_botnet_debug=1\n"
                "log_echo_ssl=1\n"
                "log_echo_broadcast=1\n"
                "log_debug=1\n"
                "log_irc=1\n"
                "log_socket=1\n"
                "log_bot=1\n"
                "log_identd=1\n"
                "log_botnet=1\n"
                "log_botnet_debug=1\n"
                "log_ssl=1\n"
                "log_broadcast=1\n"
                "eval_script_timeout=1000\n"
                "log_rot=0\n"
                );
            fclose(f);
            f=NULL;
        }
    }

    c_var_array* dummy;
    unsigned master_local_id=var_keeper->register_local(dummy,NULL,__FILE__,__my_func__,__LINE__);

    c_module* test=MY_NEW c_module();
    test->load("self_test","");

    var_keeper->set_current_session(test->get_session());

    var_keeper->set_current_local(master_local_id);

    //c_var_array* vars=MY_NEW c_var_array();
    c_var_array* vars=test->global;
    vars->set_persistent(true);

    //c_var_array* v2;

    {
        string debug="etc...";

        c_var_base* v1=MY_NEW c_var_string("abc",true);

        c_index_base* mystr=MY_NEW c_index_str("$mystr");

        vars->set(NULL,mystr,v1,false,debug);

        c_var_base* v3=MY_NEW c_var_string("Index",true);

        c_index_base* idx=MY_NEW c_index_str("$idx");

        vars->set(NULL,idx,v3,false,debug);

        c_var_array* v2=MY_NEW c_var_array();
        v2->set_persistent(true);
        v2->set_read_only(true);

        c_index_base* myindex=MY_NEW c_index_str("myindex");

        c_var_string* def=MY_NEW c_var_string("def",true);
        v2->set(NULL,myindex,def,false,debug);

        c_var_int* num5=MY_NEW c_var_int(5,true);
        c_index_base* inum5=MY_NEW c_index_str("inum5");
        v2->set(NULL,inum5,num5,false,debug);

        c_var_int* num10=MY_NEW c_var_int(10,true);
        c_index_base* inum10=MY_NEW c_index_str("$num10");
        vars->set(NULL,inum10,num10,false,debug);

        {
            c_var_int* num10=MY_NEW c_var_int(10,true);
            c_index_base* inum10=MY_NEW c_index_str("num10");
            v2->set(NULL,inum10,num10,false,debug);
        }

        {
            c_var_int* num20=MY_NEW c_var_int(20,true);
            c_index_base* inum20=MY_NEW c_index_str("num20");
            v2->set(NULL,inum20,num20,false,debug);
        }

        {
            c_var_int* num30=MY_NEW c_var_int(30,true);
            c_index_base* inum30=MY_NEW c_index_str("$num30");
            vars->set(NULL,inum30,num30,false,debug);
        }

        {
            c_var_int* num=MY_NEW c_var_int(50,true);
            c_index_base* inum=MY_NEW c_index_str("$num50");
            vars->set(NULL,inum,num,false,debug);
        }

        {
            c_var_int* num=MY_NEW c_var_int(60,true);
            c_index_base* inum=MY_NEW c_index_str("$num60");
            vars->set(NULL,inum,num,false,debug);
        }

        {
            c_var_int* num=MY_NEW c_var_int(70,true);
            num->set_read_only(true);
            c_index_base* inum=MY_NEW c_index_str("$num70");
            vars->set(NULL,inum,num,false,debug);
        }

        {
            c_var_int* num40=MY_NEW c_var_int(40,true);
            c_index_base* inum40=MY_NEW c_index_str("$num40");
            vars->set(NULL,inum40,num40,false,debug);
        }

        {
            c_var_int* rovar=MY_NEW c_var_int(10,true);
            vars->set_ro("$rovar",rovar);
        }

        c_index_base* smyarray=MY_NEW c_index_str("$myarray");

        vars->set(NULL,smyarray,v2,false,debug);

        c_index_base* mynum=MY_NEW c_index_str("$mynum");
        c_var_int* mynum_v=MY_NEW c_var_int(10,true);
        vars->set(NULL,mynum,mynum_v,false,debug);

        {
            // "Internal event ".$time_string." of type ".$type.", flags1=\"".$flags1."\", flags2=\"".$flags2."\", flags3=\"".$flags3."\", flags4=\"".$flags4."\", with severity (".$severity_numeric.") ".$severity_string.", msg1=\"".$msg1."\", msg2=\"".$msg2."\", using IRC server ".$server.":".$port
            {
                // $time_string
                c_index_base* v1=MY_NEW c_index_str("$time_string");
                c_var_string* v2=MY_NEW c_var_string("00",true);
                vars->set(NULL,v1,v2,false,debug);
            }
            {
                // $type
                c_index_base* v1=MY_NEW c_index_str("$type");
                c_var_string* v2=MY_NEW c_var_string("@log@",true);
                vars->set(NULL,v1,v2,false,debug);
            }
            {
                // $flags1
                c_index_base* v1=MY_NEW c_index_str("$flags1");
                c_var_string* v2=MY_NEW c_var_string("f1",true);
                vars->set(NULL,v1,v2,false,debug);
            }
            {
                // $flags2
                c_index_base* v1=MY_NEW c_index_str("$flags2");
                c_var_string* v2=MY_NEW c_var_string("f2",true);
                vars->set(NULL,v1,v2,false,debug);
            }
            {
                // $flags3
                c_index_base* v1=MY_NEW c_index_str("$flags3");
                c_var_string* v2=MY_NEW c_var_string("f3",true);
                vars->set(NULL,v1,v2,false,debug);
            }
            {
                // $flags4
                c_index_base* v1=MY_NEW c_index_str("$flags4");
                c_var_string* v2=MY_NEW c_var_string("f4",true);
                vars->set(NULL,v1,v2,false,debug);
            }
            {
                // $severity_numeric
                c_index_base* v1=MY_NEW c_index_str("$severity_numeric");
                c_var_string* v2=MY_NEW c_var_string("1",true);
                vars->set(NULL,v1,v2,false,debug);
            }
            {
                // $severity_string
                c_index_base* v1=MY_NEW c_index_str("$severity_string");
                c_var_string* v2=MY_NEW c_var_string("severity1",true);
                vars->set(NULL,v1,v2,false,debug);
            }
            {
                // $msg1
                c_index_base* v1=MY_NEW c_index_str("$msg1");
                c_var_string* v2=MY_NEW c_var_string("msg1",true);
                vars->set(NULL,v1,v2,false,debug);
            }
            {
                // $msg2
                c_index_base* v1=MY_NEW c_index_str("$msg2");
                c_var_string* v2=MY_NEW c_var_string("msg2",true);
                vars->set(NULL,v1,v2,false,debug);
            }
            {
                // $server
                c_index_base* v1=MY_NEW c_index_str("$server");
                c_var_string* v2=MY_NEW c_var_string("server",true);
                vars->set(NULL,v1,v2,false,debug);
            }
            {
                // $port
                c_index_base* v1=MY_NEW c_index_str("$port");
                c_var_string* v2=MY_NEW c_var_string("6667",true);
                vars->set(NULL,v1,v2,false,debug);
            }
            {
                // $num2_
                c_index_base* v1=MY_NEW c_index_str("$num2_");
                c_var_string* v2=MY_NEW c_var_string("2",true);
                vars->set(NULL,v1,v2,false,debug);
            }
            {
                // $num3_
                c_index_base* v1=MY_NEW c_index_str("$num3_");
                c_var_int* v2=MY_NEW c_var_int(3,true);
                vars->set(NULL,v1,v2,false,debug);
            }
        }
    }

    var_keeper->_dump(test,__LINE__,"#1");

#define EXPS    (300)

    static struct
    {
        char exp[1024];
        c_var_base* res;
        c_function* fn;
        bool error;
        c_micro_symbol_table* symbol_table;
    } expressions[EXPS];

    memset(expressions,0,sizeof(expressions));

    string global="global $mystr, $idx, $num10, $num30, $num50, $num60, $num70, "
        "$num40, $myarray, $mynum, $time_string, $type, "
        "$flags1, $flags2, $flags3, $flags4, $severity_numeric, $severity_string, "
        "$msg1, $msg2, $server, $port, $num2_, $num3_, $my_num, $y2, $y3; ";

    strcpy(expressions[0].exp,"return \"hello \\\" \\\\ \".$mystr.$myarray[\"my\".$idx/i][2]");
    expressions[0].res=MY_NEW c_var_string("hello \" \\ abcf");

    strcpy(expressions[1].exp,"return \"hello \\\" \\\\ \".$mystr.(($myarray[\"my\".$idx/i])[2])");
    expressions[1].res=MY_NEW c_var_string("hello \" \\ abcf");

    strcpy(expressions[2].exp,"return (\"h\".$mystr)[0]");
    expressions[2].res=MY_NEW c_var_string("h");

    strcpy(expressions[3].exp,"return (\"hello \\\" \\\\ \".$mystr).(($mystr[1])[0])");
    expressions[3].res=MY_NEW c_var_string("hello \" \\ abcb");

    strcpy(expressions[4].exp,"return (($mystr[1])[0])");
    expressions[4].res=MY_NEW c_var_string("b");

    strcpy(expressions[5].exp,"return (\"hello \\\" \\\\ \".$mystr)[1]");
    expressions[5].res=MY_NEW c_var_string("e");

    strcpy(expressions[6].exp,"return (((\"a\"[0][0].((((\"b\"))))).(((((((((((\"c\")))))))))))))");
    expressions[6].res=MY_NEW c_var_string("abc");

    strcpy(expressions[7].exp,"return (\"a\".(((($mystr[1/i].\"x\"))))).(((((((((((\"c\")))))))))))");
    expressions[7].res=MY_NEW c_var_string("abxc");

    strcpy(expressions[8].exp,"return (\"abcef\"[4]).\"b\"");
    expressions[8].res=MY_NEW c_var_string("fb");

    strcpy(expressions[9].exp,"return (\"a\".\"b\".$myarray[\"myindex\"])");
    expressions[9].res=MY_NEW c_var_string("abdef");

    strcpy(expressions[10].exp,"return -(-$mynum)");
    expressions[10].res=MY_NEW c_var_int(10);

    strcpy(expressions[11].exp,"return \"Internal event \".$time_string.\" of type \".$type.\", flags1=\\\"\".$flags1.\"\\\", flags2=\\\"\".$flags2.\"\\\", flags3=\\\"\".$flags3.\"\\\", flags4=\\\"\".$flags4.\"\\\", with severity (\".$severity_numeric.\") \".$severity_string.\", msg1=\\\"\".$msg1.\"\\\", msg2=\\\"\".$msg2.\"\\\", using IRC server \".$server.\":\".$port");
    expressions[11].res=MY_NEW c_var_string("Internal event 00 of type @log@, flags1=\"f1\", flags2=\"f2\", flags3=\"f3\", flags4=\"f4\", with severity (1) severity1, msg1=\"msg1\", msg2=\"msg2\", using IRC server server:6667");

    strcpy(expressions[12].exp,"return (\"a\"[\"0\"][\"0\"].((\"b\").\"c\"))");
    expressions[12].res=NULL;

    strcpy(expressions[13].exp,"return \"abc\"[(int)$num2_]");
    expressions[13].res=MY_NEW c_var_string("c");

    strcpy(expressions[14].exp,"return \"abc\"[(int)$num2_-1]");
    expressions[14].res=MY_NEW c_var_string("b");

    strcpy(expressions[15].exp,"return \"abc\"[-((int)$num2_)+3]");
    expressions[15].res=MY_NEW c_var_string("b");

    strcpy(expressions[16].exp,"return \"abc\"[-((int)$num2_)-1000+3+1000]");
    expressions[16].res=MY_NEW c_var_string("b");

    strcpy(expressions[17].exp,"return \"abc\"[-((int)$num2_)-1000+3*100+3-300+1000]");
    expressions[17].res=MY_NEW c_var_string("b");

    strcpy(expressions[18].exp,"return \"abc\"[-((int)$num2_)-1000+3*-100+3+300+1000]");
    expressions[18].res=MY_NEW c_var_string("b");

    strcpy(expressions[19].exp,"return \"abc\"[-((int)$num2_)-1000+3*(int)\"-100\"+3+300+1000]");
    expressions[19].res=MY_NEW c_var_string("b");

    strcpy(expressions[20].exp,"return \"abc\"[-((int)$num2_)-1000+3*(int)\"-002\"+3+006+1000]");
    expressions[20].res=MY_NEW c_var_string("b");

    strcpy(expressions[21].exp,"return \"abc\"[-((int)$num2_)-1000+3*(int)\"-0x2\"+3+0x6+1000]");
    expressions[21].res=MY_NEW c_var_string("b");

    strcpy(expressions[22].exp,"return \"abc\"[-((-((int)$num2_))-1000+10/2-105%10+1000)]");
    expressions[22].res=MY_NEW c_var_string("c");

    strcpy(expressions[23].exp,"return \"abc\"[1|2-1]");
    expressions[23].res=MY_NEW c_var_string("b");

    strcpy(expressions[24].exp,"return \"abc\"[(1|2)-1]");
    expressions[24].res=MY_NEW c_var_string("c");

    strcpy(expressions[25].exp,"return ++$num3_+5");
    expressions[25].res=MY_NEW c_var_int(4+5);

    strcpy(expressions[26].exp,"return ++3");
    expressions[26].res=NULL;

    strcpy(expressions[27].exp,"return $myarray[\"inum5\"]");
    expressions[27].res=MY_NEW c_var_int(5);

    strcpy(expressions[28].exp,"return ++$myarray[\"inum5\"]");
    expressions[28].res=MY_NEW c_var_int(6);

    strcpy(expressions[29].exp,"return $myarray[\"inum5\"]");
    expressions[29].res=MY_NEW c_var_int(6);

    strcpy(expressions[30].exp,"return ++($myarray[\"inum5\"])");
    expressions[30].res=NULL;

    strcpy(expressions[31].exp,"return (++$num10)+2*++123");
    expressions[31].res=NULL;

    strcpy(expressions[32].exp,"return $num10");
    expressions[32].res=MY_NEW c_var_int(10);

    strcpy(expressions[33].exp,"return (++$myarray[\"num10\"])+2*++123");
    expressions[33].res=NULL;

    strcpy(expressions[34].exp,"return $myarray[\"num10\"]");
    expressions[34].res=MY_NEW c_var_int(10);

    strcpy(expressions[35].exp,"return ++$myarray[\"num10\"]");
    expressions[35].res=MY_NEW c_var_int(11);

    strcpy(expressions[36].exp,"return $myarray[\"num10\"]");
    expressions[36].res=MY_NEW c_var_int(11);

    strcpy(expressions[37].exp,"return $myarray[\"num20\"]");
    expressions[37].res=MY_NEW c_var_int(20);

    strcpy(expressions[38].exp,"return $myarray[\"num20\"]+++10");
    expressions[38].res=MY_NEW c_var_int(30);

    strcpy(expressions[39].exp,"return $myarray[\"num20\"]");
    expressions[39].res=MY_NEW c_var_int(21);

    strcpy(expressions[40].exp,"return $myarray[\"num20\"]+-22");
    expressions[40].res=MY_NEW c_var_int(-1);

    strcpy(expressions[41].exp,"return $myarray[\"num20\"]-+22");
    expressions[41].res=MY_NEW c_var_int(-1);

    strcpy(expressions[42].exp,"return !$myarray[\"num20\"]");
    expressions[42].res=MY_NEW c_var_bool(false);

    strcpy(expressions[43].exp,"return !$myarray[\"num20\"]+10+100*!-1");
    expressions[43].res=MY_NEW c_var_int(10);

    strcpy(expressions[44].exp,"return ~$myarray[\"num20\"]");
    expressions[44].res=MY_NEW c_var_int(~21);

    strcpy(expressions[45].exp,"return ~\"123\"");
    expressions[45].res=MY_NEW c_var_int(~123);

    strcpy(expressions[46].exp,"return !\"123\"");
    expressions[46].res=MY_NEW c_var_bool(false);

    strcpy(expressions[47].exp,"return !\"0\"");
    expressions[47].res=MY_NEW c_var_bool(true);

    strcpy(expressions[48].exp,"return !\"abc\"");
    expressions[48].res=MY_NEW c_var_bool(false);

    strcpy(expressions[49].exp,"return !$myarray[\"num20\"]+++10+100*!-1++");
    expressions[49].res=NULL;

    strcpy(expressions[50].exp,"return $myarray[\"num20\"]");
    expressions[50].res=MY_NEW c_var_int(21);

    strcpy(expressions[51].exp,"return !$myarray[\"num20\"]+++10+100*!-$num30++");
    expressions[51].res=MY_NEW c_var_int(10);

    strcpy(expressions[52].exp,"return !$myarray[\"num20\"]+++10+100*!-($num30++)");
    expressions[52].res=MY_NEW c_var_int(10);

    strcpy(expressions[53].exp,"return -1++");
    expressions[53].res=NULL;

    strcpy(expressions[54].exp,"return -1--");
    expressions[54].res=NULL;

    strcpy(expressions[55].exp,"return $num30++");
    expressions[55].res=MY_NEW c_var_int(32);

    strcpy(expressions[56].exp,"return ++$num30");
    expressions[56].res=MY_NEW c_var_int(34);

    strcpy(expressions[57].exp,"return ++$num30++");
    expressions[57].res=NULL;

    strcpy(expressions[58].exp,"return $num30");
    expressions[58].res=MY_NEW c_var_int(34);

    strcpy(expressions[59].exp,"return \"1\"-\"2\"");
    expressions[59].res=MY_NEW c_var_int(-1);

    strcpy(expressions[60].exp,"return (3).\"2\"");
    expressions[60].res=MY_NEW c_var_string("32");

    strcpy(expressions[61].exp,"return 1-\"2\"");
    expressions[61].res=MY_NEW c_var_int(-1);

    strcpy(expressions[62].exp,"return \"3\".(2)");
    expressions[62].res=MY_NEW c_var_string("32");

    strcpy(expressions[63].exp,"return \"1\"-2");
    expressions[63].res=MY_NEW c_var_int(-1);

    strcpy(expressions[64].exp,"return (int)\"64\"");
    expressions[64].res=MY_NEW c_var_int(64);

    strcpy(expressions[65].exp,"return (string)65");
    expressions[65].res=MY_NEW c_var_string("65");

    strcpy(expressions[66].exp,"return 1+!");
    expressions[66].res=NULL;

    strcpy(expressions[67].exp,"return 1+(int)");
    expressions[67].res=NULL;

    strcpy(expressions[68].exp,"return 1+");
    expressions[68].res=NULL;

    strcpy(expressions[69].exp,"return ++");
    expressions[69].res=NULL;

    strcpy(expressions[70].exp,"return <<");
    expressions[70].res=NULL;

    strcpy(expressions[71].exp,"return 3<<\"1\"");
    expressions[71].res=MY_NEW c_var_int(6);

    strcpy(expressions[72].exp,"return \"3\">>\"1\"");
    expressions[72].res=MY_NEW c_var_int(1);

    strcpy(expressions[73].exp,"return \"3\">>1");
    expressions[73].res=MY_NEW c_var_int(1);

    strcpy(expressions[74].exp,"return >=");
    expressions[74].res=NULL;

    strcpy(expressions[75].exp,"return \"3\">=1");
    expressions[75].res=MY_NEW c_var_bool(true);

    strcpy(expressions[76].exp,"return \"3\"<1");
    expressions[76].res=MY_NEW c_var_bool(false);

    strcpy(expressions[77].exp,"return \"3\">=\"3\"");
    expressions[77].res=MY_NEW c_var_bool(true);

    strcpy(expressions[78].exp,"return \"-3\"==-3");
    expressions[78].res=MY_NEW c_var_bool(true);

    strcpy(expressions[79].exp,"return -3!=\"-3\"");
    expressions[79].res=MY_NEW c_var_bool(false);

    strcpy(expressions[80].exp,"return -3!==\"-3\"");
    expressions[80].res=MY_NEW c_var_bool(true);

    strcpy(expressions[81].exp,"return -3===\"-3\"");
    expressions[81].res=MY_NEW c_var_bool(false);

    strcpy(expressions[82].exp,"return \"-3\"===\"-3\"");
    expressions[82].res=MY_NEW c_var_bool(true);

    strcpy(expressions[83].exp,"return -5===-5");
    expressions[83].res=MY_NEW c_var_bool(true);

    strcpy(expressions[84].exp,"return \"a\"||0");
    expressions[84].res=MY_NEW c_var_bool(true);

    strcpy(expressions[85].exp,"return \"a\"||10");
    expressions[85].res=MY_NEW c_var_bool(true);

    strcpy(expressions[86].exp,"return 0&&\"a\"");
    expressions[86].res=MY_NEW c_var_bool(false);

    strcpy(expressions[87].exp,"return \"1000\"&&2");
    expressions[87].res=MY_NEW c_var_bool(true);

    strcpy(expressions[88].exp,"return \"a\"&&10");
    expressions[88].res=MY_NEW c_var_bool(true);

    strcpy(expressions[89].exp,"return \"a\" && 10");
    expressions[89].res=MY_NEW c_var_bool(true);

    strcpy(expressions[90].exp,"return \"a\"[0)");
    expressions[90].res=NULL;

    strcpy(expressions[91].exp,"return  ");
    expressions[91].res=MY_NEW c_var_unset();

    strcpy(expressions[92].exp,"return 0 && $num50++");
    expressions[92].res=MY_NEW c_var_bool(false);

    strcpy(expressions[93].exp,"return $num50");
    expressions[93].res=MY_NEW c_var_int(50);

    strcpy(expressions[94].exp,"return 1 || 10 + (100 * (0 && $num50++) )");
    expressions[94].res=MY_NEW c_var_bool(true);

    strcpy(expressions[95].exp,"return $num50");
    expressions[95].res=MY_NEW c_var_int(50);

    strcpy(expressions[96].exp,"return $num50-- || 10 + (100 * (1 && $num50++) )");
    expressions[96].res=MY_NEW c_var_bool(true);

    strcpy(expressions[97].exp,"return $num50");
    expressions[97].res=MY_NEW c_var_int(49);

    strcpy(expressions[98].exp,"return $num50-- || 10 + (100 * (0 && $num50++) )");
    expressions[98].res=MY_NEW c_var_bool(true);

    strcpy(expressions[99].exp,"return $num50");
    expressions[99].res=MY_NEW c_var_int(48);

    strcpy(expressions[100].exp,"return ( --$num50 || 10 + (100 * (0 && ++$num50) ) ) + 20");
    expressions[100].res=MY_NEW c_var_int(21);

    strcpy(expressions[101].exp,"return $num50");
    expressions[101].res=MY_NEW c_var_int(47);

    strcpy(expressions[102].exp,"return ( $num50-- - 48 || 10 + (100 * (0 && ++$num50) ) ) + 20");
    expressions[102].res=MY_NEW c_var_int(21);

    strcpy(expressions[103].exp,"return $num50");
    expressions[103].res=MY_NEW c_var_int(46);

    strcpy(expressions[104].exp,"return ( --$num50 - 46+1 || 10 + (100 * (0 && ++$num50) ) ) + 20");
    expressions[104].res=MY_NEW c_var_int(21);

    strcpy(expressions[105].exp,"return $num50");
    expressions[105].res=MY_NEW c_var_int(45);

    strcpy(expressions[106].exp,"return ( --$num50 - 45+1 && 10 + (100 * (0 && ++$num50) ) ) + 20");
    expressions[106].res=MY_NEW c_var_int(20);

    strcpy(expressions[107].exp,"return $num50");
    expressions[107].res=MY_NEW c_var_int(44);

    strcpy(expressions[108].exp,"return $num60");
    expressions[108].res=MY_NEW c_var_int(60);

    strcpy(expressions[109].exp,"return ($num60 || ++$num60) ? $num60 : ++$num60 + 1");
    expressions[109].res=MY_NEW c_var_int(60);

    strcpy(expressions[110].exp,"return $num60");
    expressions[110].res=MY_NEW c_var_int(60);

    strcpy(expressions[111].exp,"return ( ($num60 || ++$num60) ? $num60 : ++$num60 ) + 1");
    expressions[111].res=MY_NEW c_var_int(61);

    strcpy(expressions[112].exp,"return $num60");
    expressions[112].res=MY_NEW c_var_int(60);

    strcpy(expressions[113].exp,"return ($num60 || ++$num60) ? $num60 : ++$num60");
    expressions[113].res=MY_NEW c_var_int(60);

    strcpy(expressions[114].exp,"return $num60");
    expressions[114].res=MY_NEW c_var_int(60);

    strcpy(expressions[115].exp,"return ($num60 || ++$num60) ? $num60 : $num60++");
    expressions[115].res=MY_NEW c_var_int(60);

    strcpy(expressions[116].exp,"return $num60");
    expressions[116].res=MY_NEW c_var_int(60);

    strcpy(expressions[117].exp,"return $num60 - $num60 ? 0 : $num60++ ? --$num60 + 1 : 0");
    expressions[117].res=MY_NEW c_var_int(61);

    strcpy(expressions[118].exp,"return $num60");
    expressions[118].res=MY_NEW c_var_int(60);

    strcpy(expressions[119].exp,"return $num60 - $num60 ? 0 : $num60++ +1 ? --$num60 + 1 : 0");
    expressions[119].res=MY_NEW c_var_int(61);

    strcpy(expressions[120].exp,"return $num60");
    expressions[120].res=MY_NEW c_var_int(60);

    strcpy(expressions[121].exp,"return $num60 - $num60 ? 0 : $num60++ +1 ? --$num60 + 1 * $some_unset_variable : 0");
    expressions[121].res=NULL;

    strcpy(expressions[122].exp,"return $num60");
    expressions[122].res=MY_NEW c_var_int(60);

    strcpy(expressions[123].exp,"return $num60 - $num60 ? 0 : ( 1 ? 0 : $num60++ +1) ? --$num60 + 1 * 2 : 0");
    expressions[123].res=MY_NEW c_var_int(0);

    strcpy(expressions[124].exp,"return $num60");
    expressions[124].res=MY_NEW c_var_int(60);

    strcpy(expressions[125].exp,"return $my_num=$num60");
    expressions[125].res=MY_NEW c_var_int(60);

    strcpy(expressions[126].exp,"return $my_num=$num60 + 1");
    expressions[126].res=MY_NEW c_var_int(61);

    strcpy(expressions[127].exp,"return $my_num+=$num60");
    expressions[127].res=MY_NEW c_var_int(61+60);

    strcpy(expressions[128].exp,"return $num70+=$num60");
    expressions[128].res=MY_NEW c_var_int(70 /* +60 */);

    strcpy(expressions[129].exp,"return $y2=$y");
    expressions[129].res=MY_NEW c_var_unset();

    strcpy(expressions[130].exp,"$x=0; return $x+=$num70+=$y=$num60++");
    expressions[130].res=MY_NEW c_var_int(70 /* +60 */);

    strcpy(expressions[131].exp,"return --$num60");
    expressions[131].res=MY_NEW c_var_int(60);

    strcpy(expressions[132].exp,"return $y");
    expressions[132].res=MY_NEW c_var_unset();

    strcpy(expressions[133].exp,"return $y2=$y");
    expressions[133].res=MY_NEW c_var_unset();

    strcpy(expressions[134].exp,"return $y2");
    expressions[134].res=MY_NEW c_var_unset();

    strcpy(expressions[135].exp,"return $y3=$num70+$y2");
    expressions[135].res=NULL;

    strcpy(expressions[136].exp,"return $y3");
    expressions[136].res=MY_NEW c_var_unset();

    strcpy(expressions[137].exp,"return $num60++,$num60");
    expressions[137].res=MY_NEW c_var_int(61);

    strcpy(expressions[138].exp,"return -f2(1,2,3)");
    expressions[138].res=MY_NEW c_var_int(-16);

    strcpy(expressions[139].exp,"return array(\"myindex\" => \"def\", \"inum5\" => 5, \"num10\" => 10, \"num20\" => 20)");
    expressions[139].res=NULL;

    strcpy(expressions[140].exp,"return f3(\"abcDEF\")");
    expressions[140].res=MY_NEW c_var_string("ABCDEF");

    strcpy(expressions[141].exp,"return f4(1)");
    expressions[141].res=MY_NEW c_var_string("13*13bd!#BD123bbE");

    strcpy(expressions[142].exp,"return $y2=60");
    expressions[142].res=MY_NEW c_var_int(60);

    strcpy(expressions[143].exp,"return 0 && 5, 7");
    expressions[143].res=MY_NEW c_var_int(7);

    strcpy(expressions[144].exp,"return 0 && 5*2, $y2++, 1 || $y2-- || 6 && 5, 13");
    expressions[144].res=MY_NEW c_var_int(13);

    strcpy(expressions[145].exp,"return $y2");
    expressions[145].res=MY_NEW c_var_int(61);

    strcpy(expressions[146].exp,"return 0 && $y2=100, $y2++, 0 && 1 || $y2-- || 6 && 5, 13");
    expressions[146].res=MY_NEW c_var_int(13);

    strcpy(expressions[147].exp,"return $y2");
    expressions[147].res=MY_NEW c_var_int(61);

    strcpy(expressions[148].exp,"return 0 && $y2=100, $y2++, $y2++ && 0 || $y2-- || 6 && 5, 13");
    expressions[148].res=MY_NEW c_var_int(13);

    strcpy(expressions[149].exp,"return $num50=53, $y2");
    expressions[149].res=MY_NEW c_var_int(62);

    strcpy(expressions[150].exp,"return (1 || 10 + (100 * (0 && $num50++) )) +1");
    expressions[150].res=MY_NEW c_var_int(2);

    strcpy(expressions[151].exp,"return $num50");
    expressions[151].res=MY_NEW c_var_int(53);

    strcpy(expressions[152].exp,"return 1+(1 || 10 + (100 * (0 && $num50++) ))");
    expressions[152].res=MY_NEW c_var_int(2);

    strcpy(expressions[153].exp,"return $num50");
    expressions[153].res=MY_NEW c_var_int(53);

    strcpy(expressions[154].exp,"return 1+ (1 || 10 + (100 * (0 && $num50++) )) +1");
    expressions[154].res=MY_NEW c_var_int(3);

    strcpy(expressions[155].exp,"return $num50");
    expressions[155].res=MY_NEW c_var_int(53);

    strcpy(expressions[156].exp,"return \"\\$time_string=$time_string\"");
    expressions[156].res=MY_NEW c_var_string("$time_string=00");

    strcpy(expressions[157].exp,"return f5(10)");
    expressions[157].res=MY_NEW c_var_string("0123456789Done: 10 times.10111213141516171819Done: 10 times.");

    strcpy(expressions[158].exp,"return f6()");
    expressions[158].res=MY_NEW c_var_string("$i1=1, $i1=2, $i1=3, $i1=4, $i1=5, $i1=6, $i1=7, $i1=8, $i1=9, $i1=10, done #1, $i1=12, $i1=13, $i1=14, $i1=15, $i1=16, $i1=17, $i1=18, $i1=19, $i1=20, done #2, done #3, done #4, $i1=24, $i2=1, $i1=24, $i2=2, $i1=24, $i2=3, $i1=24, $i2=4, $i1=24, $i2=5, $i1=24, $i2=6, $i1=24, $i2=7, $i1=24, $i2=8, $i1=24, $i2=9, $i1=24, $i2=10, done #5, $i2=17, $i2=18, $i2=19, $i2=20, done #6, $i2=21, $i2=22, $i2=23, $i2=24, $i2=25, $i2=26, $i2=27, $i2=28, $i2=29, $i2=30, done #7, $i1=1, $i2=1, $i1=1, $i2=2, $i1=1, $i2=3, $i1=1, $i2=4, $i1=1, $i2=5, $i1=1, $i2=6, $i1=1, $i2=7, $i1=1, $i2=8, $i1=1, $i2=9, $i1=1, $i2=10, done #8, done #9, $i1=32, $i2=0, $i1=32, $i2=1, $i1=32, $i2=2, $i1=32, $i2=3, $i1=32, $i2=4, $i1=32, $i2=5, $i1=32, $i2=6, $i1=32, $i2=7, $i1=32, $i2=8, $i1=32, $i2=9, $i1=32, $i2=10, done #10.");

    strcpy(expressions[159].exp,"return f7()");
    expressions[159].res=MY_NEW c_var_string("$i1=0, $i2=0$i1=0, $i2=1$i1=0, $i2=2$i1=0, $i2=3$i1=0, $i2=4$i1=0, $i2=5$i1=0, $i2=6$i1=0, $i2=7$i1=0, $i2=8$i1=0, $i2=9$i1=1, $i2=0$i1=1, $i2=1$i1=1, $i2=2$i1=1, $i2=3$i1=1, $i2=4$i1=1, $i2=5$i1=1, $i2=6$i1=1, $i2=7$i1=1, $i2=8$i1=1, $i2=9$i1=2, $i2=0$i1=2, $i2=1$i1=2, $i2=2$i1=2, $i2=3$i1=2, $i2=4$i1=2, $i2=5$i1=2, $i2=6$i1=2, $i2=7$i1=2, $i2=8$i1=2, $i2=9$i1=3, $i2=0$i1=3, $i2=1$i1=3, $i2=2$i1=3, $i2=3$i1=3, $i2=4$i1=3, $i2=5$i1=3, $i2=6$i1=3, $i2=7$i1=3, $i2=8$i1=3, $i2=9$i1=4, $i2=0$i1=4, $i2=1$i1=4, $i2=2$i1=4, $i2=3$i1=4, $i2=4$i1=4, $i2=5$i1=4, $i2=6$i1=4, $i2=7$i1=4, $i2=8$i1=4, $i2=9done.");

    strcpy(expressions[160].exp,"return f8()");
    expressions[160].res=MY_NEW c_var_string("x rules!, x rules!, x rules!, 1, 2, 3, 1, 2, 3, a == 1, b == 2, c == 3, a == 1, b == 2, c == 3, a == 0, b == 0, c == 0, a=0, b=0, c=0, a=13, b=11, c=0, a=13, b=13, c=12, 13, 13, 13, ");

    strcpy(expressions[161].exp,"return f9()");
    expressions[161].res=MY_NEW c_var_string("a1, a2, b3, b5, b7, b9, ");

    strcpy(expressions[162].exp,"return f4(TRUE)");
    expressions[162].res=MY_NEW c_var_string("13*13bd!#BD123bbE");

    strcpy(expressions[163].exp,"return f10(1)");
    expressions[163].res=MY_NEW c_var_string("11!");

    strcpy(expressions[164].exp,"return f10(2)");
    expressions[164].res=MY_NEW c_var_string("22!");

    strcpy(expressions[165].exp,"return f10(3)");
    expressions[165].res=MY_NEW c_var_string("33+3!");

    strcpy(expressions[166].exp,"return f10(6)");
    expressions[166].res=MY_NEW c_var_string("default6!");

    strcpy(expressions[167].exp,"return f10(7)");
    expressions[167].res=MY_NEW c_var_string("default!");

    strcpy(expressions[168].exp,"return f11()");
    expressions[168].res=MY_NEW c_var_string("1, ");

    strcpy(expressions[169].exp,"return f12()");
    expressions[169].res=MY_NEW c_var_string("0, 1, 1, 2, 2, 2, 3, 3, 3, 3, 4, 4, 4, 4, 4, done #1; 6, 6, 7, 8, 8, 9, 9, 10, 10, ");

    strcpy(expressions[170].exp,"return f13()");
    expressions[170].res=MY_NEW c_var_string("13, 13, 13, 13, ");

    strcpy(expressions[171].exp,"return f14()");
    expressions[171].res=MY_NEW c_var_string("0, 1, 2, 3, 14, 14, 14, 14, ");

    strcpy(expressions[172].exp,"global $my_test; return $my_test=100");
    expressions[172].res=MY_NEW c_var_int(100);

    strcpy(expressions[173].exp,"global $my_test; return f15($my_test)");
    expressions[173].res=MY_NEW c_var_unset();

    strcpy(expressions[174].exp,"global $my_test; return $my_test");
    expressions[174].res=MY_NEW c_var_int(10);

    strcpy(expressions[175].exp,"return f16()");
    expressions[175].res=MY_NEW c_var_string("ok1, ok2, ok3, ");

    strcpy(expressions[176].exp,"return f17(\"x\")");
    expressions[176].res=MY_NEW c_var_string("abc|def, ghi, , jkl, ");

    strcpy(expressions[177].exp,"return f17(0)");
    expressions[177].res=MY_NEW c_var_string("abc|def|-ghi|-|-jkl, ");

    strcpy(expressions[178].exp,"return f17(1)");
    expressions[178].res=MY_NEW c_var_string("abc|def|-ghi|-|-jkl, ");

    strcpy(expressions[179].exp,"return f17(2)");
    expressions[179].res=MY_NEW c_var_string("abc|def, ghi|-|-jkl, ");

    strcpy(expressions[180].exp,"return f17(3)");
    expressions[180].res=MY_NEW c_var_string("abc|def, ghi, |-jkl, ");

    strcpy(expressions[181].exp,"return f17(4)");
    expressions[181].res=MY_NEW c_var_string("abc|def, ghi, , jkl, ");

    strcpy(expressions[182].exp,"return f18()");
    expressions[182].res=MY_NEW c_var_string("3, 1, 4, 97, !, ");

    strcpy(expressions[183].exp,"return count(NULL)");
    expressions[183].res=MY_NEW c_var_int(0);

    strcpy(expressions[184].exp,"return count($some_unset_variable)");
    expressions[184].res=MY_NEW c_var_int(0);

    strcpy(expressions[185].exp,"return f19()");
    expressions[185].res=MY_NEW c_var_string("AA");

    strcpy(expressions[186].exp,"return 100/0");
    expressions[186].res=NULL;

    strcpy(expressions[187].exp,"return 123%0");
    expressions[187].res=NULL;

    strcpy(expressions[188].exp,"return f20()");
    expressions[188].res=NULL;

    strcpy(expressions[189].exp,"return f21()");
    expressions[189].res=MY_NEW c_var_string("B");

    strcpy(expressions[190].exp,"return f21()");
    expressions[190].res=MY_NEW c_var_string("AB");

    strcpy(expressions[191].exp,"return \"abc$num10 and \".'$num10'");
    expressions[191].res=MY_NEW c_var_string("abc10 and $num10");

    strcpy(expressions[192].exp,"return eval(\"\\\"\\$num10 and \\\".'$num10'\")");
    expressions[192].res=MY_NEW c_var_bool(false);

    strcpy(expressions[193].exp,"return eval(\"return \\\"\\$num10 and \\\".'$num10';\")");
    expressions[193].res=MY_NEW c_var_string("10 and 10");

    strcpy(expressions[194].exp,"return eval(\"return \\\"\\$num10 and \\\".'$num10'\")");
    expressions[194].res=MY_NEW c_var_bool(false);

    strcpy(expressions[195].exp,"return eval(\"\\\"\\$num10 and \\\".'$num10';\")");
    expressions[195].res=MY_NEW c_var_unset();

    strcpy(expressions[196].exp,"global $evaltest; return $evaltest=eval(\"\\\"\\$num10 and \\\".'$num10';\")");
    expressions[196].res=MY_NEW c_var_unset();

    strcpy(expressions[197].exp,"global $evaltest; return isset($evaltest)");
    expressions[197].res=MY_NEW c_var_bool(false);

    strcpy(expressions[198].exp,"global $evaltest; return $evaltest=eval(\"return \\\"\\$num10 and \\\".'$num10';\")");
    expressions[198].res=MY_NEW c_var_string("10 and 10");

    strcpy(expressions[199].exp,"global $evaltest; return isset($evaltest)");
    expressions[199].res=MY_NEW c_var_bool(true);

    strcpy(expressions[200].exp,"return a1()");
    expressions[200].res=MY_NEW c_var_unset();

    strcpy(expressions[201].exp,"global $nick; return $nick='test'");
    expressions[201].res=MY_NEW c_var_string("test");

    strcpy(expressions[202].exp,"global $got; return $got=TRUE");
    expressions[202].res=MY_NEW c_var_bool(true);

    strcpy(expressions[203].exp,"global $got, $nick; return log_print(\"nick='\".$nick.\" = '\".$nick.\"' = \".($got?\"yes\":\"no\"))");
    expressions[203].res=MY_NEW c_var_unset();

    strcpy(expressions[204].exp,"return $abcdefg_123='xyz'");
    expressions[204].res=MY_NEW c_var_string("xyz");

    strcpy(expressions[205].exp,"return a2_a(array(\"user\" => 'xxx', \"line\" => \".control +o #\"))");
    expressions[205].res=MY_NEW c_var_unset();

    strcpy(expressions[206].exp,"return b1()");
    expressions[206].res=MY_NEW c_var_string("1234567");

    strcpy(expressions[207].exp,"return c1()");
    expressions[207].res=NULL;

    strcpy(expressions[208].exp,"return d1()");
    expressions[208].res=NULL;

    strcpy(expressions[209].exp,"return e1()");
    expressions[209].res=NULL;

    strcpy(expressions[210].exp,"return g1()");
    expressions[210].res=NULL;

    strcpy(expressions[211].exp,"return h1(0)");
    expressions[211].res=NULL;

    strcpy(expressions[212].exp,"return j1(0)");
    expressions[212].res=NULL;

    strcpy(expressions[213].exp,"return k1(0)");
    expressions[213].res=NULL;

    strcpy(expressions[214].exp,"return m0()");
    expressions[214].res=NULL;

    strcpy(expressions[215].exp,"return n0()");
    expressions[215].res=NULL;

    strcpy(expressions[216].exp,"return $y2=60");
    expressions[216].res=MY_NEW c_var_int(60);

    strcpy(expressions[217].exp,"return 0 || 0 || 0 || $y2++");
    expressions[217].res=MY_NEW c_var_bool(true);

    strcpy(expressions[218].exp,"return $y2");
    expressions[218].res=MY_NEW c_var_int(61);

    strcpy(expressions[219].exp,"return (1 ? 'A' . 'A' : 1 ? 'B' . 'B' : 'b' . 'b') . 'END'");
    expressions[219].res=MY_NEW c_var_string("AAEND");

    strcpy(expressions[220].exp,"return (0 ? 'A' . 'A' : 1 ? 'B' . 'B' : 'b' . 'b') . 'END'");
    expressions[220].res=MY_NEW c_var_string("BBEND");

    strcpy(expressions[221].exp,"return (0 ? 'A' . 'A' : 0 ? 'B' . 'B' : 'b' . 'b') . 'END'");
    expressions[221].res=MY_NEW c_var_string("bbEND");

    strcpy(expressions[222].exp,"global $rovar; return $rovar++");
    expressions[222].res=MY_NEW c_var_int(10);

    strcpy(expressions[223].exp,"global $rovar; return $rovar++");
    expressions[223].res=MY_NEW c_var_int(10);

    strcpy(expressions[224].exp,"global $rovar; return ++$rovar");
    expressions[224].res=MY_NEW c_var_int(11);

    strcpy(expressions[225].exp,"global $rovar; return ++$rovar");
    expressions[225].res=MY_NEW c_var_int(11);

    strcpy(expressions[226].exp,"return case_ascii_1()");
    expressions[226].res=MY_NEW c_var_bool(true);

    strcpy(expressions[227].exp,"return case_rfc1459_1()");
    expressions[227].res=MY_NEW c_var_bool(true);

    strcpy(expressions[228].exp,"return case_ascii_2()");
    expressions[228].res=MY_NEW c_var_bool(true);

    strcpy(expressions[229].exp,"return case_rfc1459_2()");
    expressions[229].res=MY_NEW c_var_bool(true);

    strcpy(expressions[230].exp,"return xx1()");
    expressions[230].res=MY_NEW c_var_bool(true);

    strcpy(expressions[231].exp,"return or_opers_bug_test()");
    expressions[231].res=MY_NEW c_var_string("OKOK");

    strcpy(expressions[232].exp,"return and_opers_bug_test()");
    expressions[232].res=MY_NEW c_var_string("4OK");

    strcpy(expressions[233].exp,"return $x123 ? 'abc'");
    expressions[233].res=NULL;

    strcpy(expressions[234].exp,"return 0 ? 1 : 0 ? 2 : 0 ? 3 : 0 ? 4 : 5");
    expressions[234].res=MY_NEW c_var_int(5);

    strcpy(expressions[235].exp,"return 0 ? -1 : 0 ? 1 : ( 0 ? 2 : ( 0 ? 3 : 0 ? 4 : 5 ) )");
    expressions[235].res=MY_NEW c_var_int(5);

    strcpy(expressions[236].exp,"return ($num50, $num60) + 1");
    expressions[236].res=MY_NEW c_var_int(62);

    strcpy(expressions[237].exp,"global $botnet_test; return $botnet_test=-10");
    expressions[237].res=MY_NEW c_var_int(-10);

    strcpy(expressions[238].exp,"return test__preg_grep()");
    expressions[238].res=MY_NEW c_var_bool(true);

    strcpy(expressions[239].exp,"return test__preg_grep_bad_pattern1()");
    expressions[239].res=NULL;

    strcpy(expressions[240].exp,"return test__preg_grep_bad_pattern2()");
    expressions[240].res=NULL;

    strcpy(expressions[241].exp,"return test__preg_grep_bad_pattern3()");
    expressions[241].res=NULL;

    strcpy(expressions[243].exp,"return test_apostrophe()");
    expressions[243].res=MY_NEW c_var_string("a' abc\\def\\n-\\n\\-");

    strcpy(expressions[244].exp,"return test_apostrophe2()");
    expressions[244].res=MY_NEW c_var_string("/test\\x/");

    strcpy(expressions[245].exp,"return test__preg_match_all__test1()");
    expressions[245].res=MY_NEW c_var_bool(true);

    strcpy(expressions[246].exp,"return test__preg_match_all__test2()");
    expressions[246].res=MY_NEW c_var_bool(true);

    strcpy(expressions[247].exp,"return test__preg_match_all__test3()");
    expressions[247].res=NULL;

    strcpy(expressions[248].exp,"return test__preg_match_all__test4()");
    expressions[248].res=MY_NEW c_var_bool(true);

    strcpy(expressions[249].exp,"return test__preg_match__test1()");
    expressions[249].res=MY_NEW c_var_bool(true);

    strcpy(expressions[250].exp,"return test__preg_match__test2()");
    expressions[250].res=MY_NEW c_var_bool(true);

    strcpy(expressions[251].exp,"return test__preg_match__test3()");
    expressions[251].res=MY_NEW c_var_bool(true);

    strcpy(expressions[252].exp,"return test__preg_match__test4()");
    expressions[252].res=MY_NEW c_var_bool(true);

    strcpy(expressions[253].exp,"return test__preg_match__test5()");
    expressions[253].res=MY_NEW c_var_bool(true);

    strcpy(expressions[254].exp,"return test__preg_match__test6()");
    expressions[254].res=NULL;

    strcpy(expressions[255].exp,"return test__preg_match__test7()");
    expressions[255].res=NULL;

    strcpy(expressions[256].exp,"return test__preg_match__test8()");
    expressions[256].res=MY_NEW c_var_bool(true);

    strcpy(expressions[257].exp,"return test__preg_split__test1()");
    expressions[257].res=MY_NEW c_var_bool(true);

    strcpy(expressions[258].exp,"return test__preg_split__test2()");
    expressions[258].res=MY_NEW c_var_bool(true);

    strcpy(expressions[259].exp,"return test__preg_split__test3()");
    expressions[259].res=MY_NEW c_var_bool(true);

    strcpy(expressions[260].exp,"return test__preg_split__test4()");
    expressions[260].res=MY_NEW c_var_bool(true);

    strcpy(expressions[261].exp,"return test__preg_split__test5()");
    expressions[261].res=MY_NEW c_var_bool(true);

    strcpy(expressions[262].exp,"return test__preg_split__test6()");
    expressions[262].res=MY_NEW c_var_bool(true);

    strcpy(expressions[263].exp,"return test__bug001()");
    expressions[263].res=MY_NEW c_var_unset();

    strcpy(expressions[264].exp,"return test__preg_split__test7()");
    expressions[264].res=MY_NEW c_var_bool(true);

    strcpy(expressions[265].exp,"return test__preg_replace__test1()");
    expressions[265].res=MY_NEW c_var_bool(true);

    strcpy(expressions[266].exp,"return test__preg_replace__test2()");
    expressions[266].res=MY_NEW c_var_bool(true);

    strcpy(expressions[267].exp,"return test__preg_replace__test3()");
    expressions[267].res=MY_NEW c_var_bool(true);

    strcpy(expressions[268].exp,"return test__preg_replace__test4()");
    expressions[268].res=MY_NEW c_var_bool(true);

    strcpy(expressions[269].exp,"return test__preg_replace__test5()");
    expressions[269].res=MY_NEW c_var_bool(true);

    strcpy(expressions[270].exp,"return test__preg_replace__test6()");
    expressions[270].res=MY_NEW c_var_bool(true);

    strcpy(expressions[271].exp,"return test__preg_replace__test7()");
    expressions[271].res=MY_NEW c_var_bool(true);

    strcpy(expressions[272].exp,"return test__preg_replace__test8()");
    expressions[272].res=MY_NEW c_var_bool(true);

    strcpy(expressions[273].exp,"return test__preg_split__test9()");
    expressions[273].res=MY_NEW c_var_bool(true);

    strcpy(expressions[274].exp,"return test__bool_xor_prec14()");
    expressions[274].res=MY_NEW c_var_bool(true);

    strcpy(expressions[275].exp,"return test__bool_and_prec20()");
    expressions[275].res=MY_NEW c_var_bool(true);

    strcpy(expressions[276].exp,"return test__bool_or_prec22()");
    expressions[276].res=MY_NEW c_var_bool(false);

    strcpy(expressions[277].exp,"return test__bool_and_prec20_2()");
    expressions[277].res=MY_NEW c_var_bool(false);

    strcpy(expressions[278].exp,"return preg_split_empty_string_bug()");
    expressions[278].res=MY_NEW c_var_bool(true);

    strcpy(expressions[279].exp,"return test_foreach_nulled_reference_bug()");
    expressions[279].res=MY_NEW c_var_string(";2");

    strcpy(expressions[280].exp,"return test__multiple_line_str_1()");
    expressions[280].res=MY_NEW c_var_string("test1\n           /* test2 */\n    );");

    strcpy(expressions[281].exp,"return test__multiple_line_str_2()");
    expressions[281].res=MY_NEW c_var_string("test1\n           /* test2 */\n    );");

    strcpy(expressions[282].exp,"return bind_ex_test_1()");
    expressions[282].res=MY_NEW c_var_unset();

    strcpy(expressions[283].exp,"return bind_ex_test_1b()");
    expressions[283].res=MY_NEW c_var_unset();

    strcpy(expressions[284].exp,"return bind_ex_test_1c()");
    expressions[284].res=MY_NEW c_var_bool(true);

    expressions[285].exp[0]=0;
    expressions[285].res=NULL;

    BEGIN_DEBUG_TIME_MEASUREMENT(true,3,"Compiling tests");
    int start_from_index=0;
    int end_with_index_less_than=EXPS; // EXPS

    int i1;
    for(i1=start_from_index; i1<end_with_index_less_than; ++i1) {
        if(!*expressions[i1].exp)
            continue;
        bool error=false;
        bool need_more=false;
        string error_msg;
        string cmd="#version 2";
        expressions[i1].fn=MY_NEW c_function();
        expressions[i1].fn->parse_command(test,cmd,error,error_msg,need_more,"...","");
        cmd="function __test__() {";
        cmd+=global;
        cmd+=expressions[i1].exp;
        cmd+=";}";
        MY_RPT2(MY_RPT_INFO,"Compile #%s: %s",tos(i1).c_str(),expressions[i1].exp);
        expressions[i1].fn->parse_command(test,cmd,error,error_msg,need_more,"...","");
        expressions[i1].error=error || need_more;
        expressions[i1].symbol_table=NULL;
#define COMPILER_DEBUGGING      0
        if(expressions[i1].error) {
            expressions[i1].fn->destroy(true,true);
            MY_DELETE(expressions[i1].fn);
            expressions[i1].fn=NULL;

#if COMPILER_DEBUGGING!=0
            MY_ASSERT(NULL==expressions[i1].res);
#endif
        } else {
#if COMPILER_DEBUGGING!=0
            MY_ASSERT(NULL!=expressions[i1].res);
#endif
        }
    }

    END_DEBUG_TIME_MEASUREMENT(3,0);

    {
        c_index_str idx("$myarray");
        expressions[139].res=vars->get(NULL,&idx,false,"...")->copy();
    }

    vars->set_all_as_persistent();

    int num_all=0;
    int num_error=0;

    for(i1=start_from_index; i1<end_with_index_less_than; ++i1) {
        if(!*expressions[i1].exp)
            continue;

        mod_loop(false);

        var_keeper->set_current_session(test->get_session());
        num_all++;
        MY_RPT2(MY_RPT_INFO,"=====\nExpression #%s: %s",tos(i1).c_str(),expressions[i1].exp);

        {
            string exp=expressions[i1].exp;
            c_var_base* res=NULL;
            {
                string next_expression="";
                MY_RPT2(MY_RPT_INFO,"=== Expression #%s: %s",tos(i1).c_str(),expressions[i1].exp);

                if(!expressions[i1].error) {
                    c_var_array* local_vars=NULL;
                    unsigned local_id;
                    var_keeper->set_current_local(local_id=var_keeper->register_local(local_vars,NULL,__FILE__,__my_func__,__LINE__));

                    stack<c_eval_iteration> st;
                    c_eval_iteration ste(test,local_vars,"...");
                    ste.state=1000;
                    ste.return_to_state=0;
                    ste.fn=expressions[i1].fn;
                    ste.fn->create_symbols();
                    ste.fn_destroy=true;
                    ste.instruct_id=0;
                    ste.return_value=NULL;
                    ste.self_test=true;
                    ste.symbol_table_=ste.fn->symbol_table;
                    ste.local_id=local_id;

                    st.push(ste);

#if 0
                    if(102==i1) {
                        char tmp[256];
                        sprintf(tmp,"selftest.dump.%s.before.txt",tos(i1).c_str());
                        var_keeper->_dump((string)tmp);
#ifdef _DEBUG
                        //
                        int x=0;
                        x++;
#endif
                    }
#endif

                    BEGIN_DEBUG_TIME_MEASUREMENT(true,4,"Executing test");

                    st.top().local_id=local_id;
                    eval_loop(st);
                    var_keeper->free_local(local_id,true);

                    END_DEBUG_TIME_MEASUREMENT(4,0);

#if 0
                    if(102==i1) {
                        char tmp[256];
                        sprintf(tmp,"selftest.dump.%s.after.txt",tos(i1).c_str());
                        var_keeper->_dump((string)tmp);
#ifdef _DEBUG
                        //
                        int x=0;
                        x++;
#endif
                    }
#endif

                    res=st.top().return_value;
                }
            }

            if(res) {
                if(expressions[i1].res && res->is_identical(test,expressions[i1].res,"",exp)) {
                    MY_RPT0(MY_RPT_INFO,"Got correct result!");
                } else {
                    num_error++;
                    //var_keeper->_dump("dump.4.txt",test);
                    MY_RPT0(MY_RPT_INFO,"NB ERROR !!!!!!!!!!!!!!! Got incorrect result!");
                    // DEBUG: set breakpoint here
                    int dummy=0;
                    dummy++;
                }
            } else {
                if(!expressions[i1].res) {
                    MY_RPT0(MY_RPT_INFO,"Got correct (NULL) result!");
                } else {
                    num_error++;
                    MY_RPT0(MY_RPT_INFO,"NB ERROR !!!!!!!!!!!!!!! Got incorrect (NULL) result!");
                    if(expressions[i1].error) {
                        MY_RPT0(MY_RPT_INFO,"NB ERROR !!!!!!!!!!!!!!! Function script compiler have not compiled this expression!");
                    }
                    // DEBUG: set breakpoint here
                    int dummy=0;
                    dummy++;
                }
            }
        }

#if defined(_MSC_VER) && defined(_DEBUG)
        if(!_CrtCheckMemory()) {
            MY_RPT1(MY_RPT_INFO,"Memory corruption detected in function %s()",__my_func__);
            // DEBUG: set breakpoint here
            int dummy=0;
            dummy++;
        }
#endif

        expressions[i1].fn=NULL;
    }

    report_heap_debug();

    MY_RPT2(MY_RPT_INFO,"%s errors in %s tests (so far)",tos(num_error).c_str(),tos(num_all).c_str());

    var_keeper->set_current_local(master_local_id);

    {
        ++num_all;

        MY_RPT0(MY_RPT_INFO,"\n=====\nTesting encode/decode of BOTNET remote call variables\n=====");
        char* data;
        size_t size;
        vector<c_var_base*> args;
        args.push_back(vars);
        args.push_back(MY_NEW c_var_string("test"));
        args.push_back(MY_NEW c_var_array());
        args.push_back(MY_NEW c_var_string("test"));
        string func_name="test";
        bool problem=true;
        if(!var_encode_for_botnet(test,data,size,args,func_name)) {
            MY_RPT0(MY_RPT_INFO,"Error: var_encode_for_botnet() returned false");
        } else {
            MY_RPT1(MY_RPT_INFO,"Encoded to %s bytes",tos(size).c_str());
            vector<c_var_base*> new_args;
            string fn;
            if(!var_decode_for_botnet(test,data,size,new_args,fn)) {
                MY_RPT0(MY_RPT_INFO,"Error: var_decode_for_botnet() returned false");
            } else {
                if(args.size()!=new_args.size()) {
                    MY_RPT0(MY_RPT_INFO,"Error: Argument count mismatch");
                } else if(func_name.compare(fn)) { // !=
                    MY_RPT0(MY_RPT_INFO,"Error: Function name mismatch");
                } else {
                    problem=false;
                    vector<c_var_base*>::iterator i1, i2;
                    int ii=1;
                    for(i1=args.begin(), i2=new_args.begin(); i1!=args.end(); ++i1, ++i2, ++ii) {
                        if(!(*i1)->is_identical(test,*i2,"test","test")) {
                            MY_RPT1(MY_RPT_INFO,"Error: Argument #%s mismatch",tos(ii).c_str());
                            problem=true;
                        }
                    }
                }
                vector<c_var_base*>::iterator i1=new_args.begin();
                while(i1!=new_args.end()) {
                    MY_DELETE(*i1);
                    new_args.erase(i1);
                    i1=new_args.begin();
                }
                if(!problem) {
                    MY_RPT0(MY_RPT_INFO,"Ok: decoded to identical data");
                } else {
                    MY_RPT0(MY_RPT_INFO,"NB: NOT ok!");
                }
            }
            if(problem)
                ++num_error;

            MY_DELETE_A(data);
        }
        vector<c_var_base*>::iterator i1=args.begin();
        i1++;
        MY_DELETE(*i1);
    }

    for(i1=0; i1<EXPS; ++i1) {
        MY_DELETE(expressions[i1].res);
        expressions[i1].res=NULL;

        if(expressions[i1].fn) {
            /**/
            expressions[i1].fn->destroy(true,true);
            MY_DELETE(expressions[i1].fn);
            /**/
            expressions[i1].fn=NULL;
        }
        if(expressions[i1].symbol_table) {
            MY_DELETE(expressions[i1].symbol_table);
            expressions[i1].symbol_table=NULL;
        }
    }

    END_DEBUG_TIME_MEASUREMENT(1,0);

    var_keeper->_dump(test,__LINE__,"#2-after");

    BEGIN_DEBUG_TIME_MEASUREMENT(true,2,"Destroying variables");

    session_id_t sess=test->get_session();
    while(var_keeper->purge(sess))
        ;

    var_keeper->_dump(test,__LINE__,"#3-before-self_test-module-unload");

    unload_module("self_test");
    test=NULL;

    var_keeper->_dump(test,__LINE__,"#4-after-self_test-module-unload");

    MY_RPT0(MY_RPT_INFO,"=== Garbaging variable tree, this might take a while, be patient");

    var_keeper->free_local(master_local_id,false);

    while(var_keeper->purge(sess))
        ;
    size_t d_cnt=0;

    while((d_cnt=var_keeper->do_garbage(5000))) {
        MY_RPT1(MY_RPT_INFO,"Destroyed %s root nodes in limit of 5 seconds",tos(d_cnt).c_str());
    }

    var_keeper->_dump(test,__LINE__,"#5-after-garbage-collection");

    MY_RPT0(MY_RPT_INFO,"=== Finished garbaging variable tree");

    END_DEBUG_TIME_MEASUREMENT(2,0);

    MY_RPT0(MY_RPT_INFO,"\n=====\nAll tests done.\n=====");

#if defined(_MSC_VER) && defined(_DEBUG)
    if(!_CrtCheckMemory()) {
        MY_RPT1(MY_RPT_INFO,"Memory corruption detected in function %s()",__my_func__);
        // DEBUG: set breakpoint here
        int dummy=0;
        dummy++;
    }
#endif

    MY_RPT2(MY_RPT_INFO,"%s errors in %s tests (total)",tos(num_error).c_str(),tos(num_all).c_str());
    MY_RPT0(MY_RPT_INFO,"=====");

#ifndef _DEBUG
    cout << num_error << " errors in " << num_all << " tests" << endl;
#endif

    var_keeper->_dump(test,__LINE__,"#6-after-unload");

    report_heap_debug();

    return 0==num_error;
}

#if defined(_MSC_VER)
#   pragma warning(pop)
#endif

