/***************************************************************************
                          eval.cpp  -  description
                             -------------------
    begin                : Tue Feb 4 2006
    copyright            : (C) 2006-2007 by Marian VooDooMan Meravy
    email                : ghostvoodooman -a.T= users (dot) sourceforge.net
 ***************************************************************************/

/* $Id: eval.cpp,v 1.49 2008/07/02 09:35:46 ghostvoodooman Exp $ */

/***************************************************************************

VooDoo cIRCle - an IRC (ro)bot
Copyright (C) 2006-2008 by Marian VooDooMan Meravy
  (ghostvoodooman -a.T= users (dot) sourceforge.net)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************/

/*!
    \file
    \brief Evaluates procedure/function/expression
    \version $Id: eval.cpp,v 1.49 2008/07/02 09:35:46 ghostvoodooman Exp $
*/

#if defined(HAVE_CONFIG_H)
#   include <config.h>
#endif

#include <string>
#include <vector>
#include <stack>
#include <list>
#include <map>
#include <deque>

#include <stdlib.h>

using namespace std;

#include "utils/memleak.h"
#include "utils/assert.h"
#include "utils.h"
#include "params.h"
#include "compat.h"
#include "variable.h"
#include "log.h"
#include "utils.h"
#include "shared.h"
#include "function.h"
#include "logic.h"
#include "md5.h"
#include "modules.h"
#include "variable.h"
#include "irc.h"
#include "match.h"
#include "botnet.h"
#include "stats.h"
#include "dcc.h"
#include "threads.h"
#include "identd.h"
#include "proc.h"
#include "variable.h"
#include "eval.h"
#include "microcode.h"
#include "func/c_expression.h"
#include "microcode/c_micro_symbol_variable_mapping.h"
#include "microcode/c_micro_symbol_variable_table.h"
#include "func/c_opers.h"
#include "irc/c_irc_server.h"
#include "irc/c_online_channel.h"
#include "var/c_var_base.h"
#include "var/c_var_bool.h"
#include "var/c_var_int.h"
#include "var/c_var_string.h"
#include "var/c_var_array.h"
#include "var/c_var_unset.h"
#include "runtime/c_delay_link.h"
#include "runtime/c_delay_command.h"
#include "mod/c_module.h"
#include "proc/c_proc_exc.h"
#include "misc/shutdown.h"
#include "internal_state.h"
#include "function.h"
#include "func/eval_func.h"

#include "utils/precomp.h"

#if defined(_MSC_VER) || defined(__BORLANDC__)
#   pragma hdrstop
#endif

void eval_evaluate_proc(stack<c_eval_iteration>& st);

/*!
    \brief Performs tasks before popping of stack
    \param st Stack
*/
MY_STATIC void eval_stack_epilog(stack<c_eval_iteration>& st)
{
    microcode_perform_updates(st);
    if(st.top().parent && st.top().symbol_table_ && st.top().parent->symbol_table_)
        if(st.top().symbol_table_->get_var_table() && st.top().parent->symbol_table_->get_var_table())
            st.top().symbol_table_->get_var_table()->commit_changes(*st.top().parent->symbol_table_->get_var_table());
    if(st.top().proc_eval && st.top().fn)
        st.top().fn->destroy_symbols();
    MY_DELETE(st.top().fn);
    st.top().fn=NULL;
    st.top().fn_destroy=false;
    bool exception_reported=st.top().exception_reported;
    if(st.top().symbol_table_to_destroy==st.top().symbol_table_)
        st.top().symbol_table_=NULL;
    MY_DELETE(st.top().symbol_table_to_destroy);
    st.top().symbol_table_to_destroy=NULL;

    if(st.top().parent) {
        var_keeper->set_current_local(st.top().parent->local_id);
    }
    /*
    if(st.top().return_value)
        st.top().return_value=st.top().return_value->copy();
    */

    if(exception_reported && NULL!=st.top().micro_error) {
        c_eval_runtime_error e(st.top().this_module_,st.top().micro_error,st.top().error_debug_string,__FILE__,__my_func__);
        e.report(true);
    }

    if(!st.top().parent || st.top().parent->local_id!=st.top().local_id) {
        if(st.top().this_module_->get_session()!=SESSION_CORE_ENTRY_ID) {
            if(st.top().local_id)
                var_keeper->free_local(st.top().local_id,true);
            if(st.top().vars_)
                var_keeper->free_global(st.top().vars_,true);
        }
    }
}

/*!
    \brief Sets pesistency flag to all globals
    \version 1
    \date 2006-04-23
    \author VooDooMan
    \param this_module Current module
*/
void protect_globals(c_module* UNUSEDARG(this_module) = NULL)
{
    return;
}

bool eval_microcode(stack<c_eval_iteration>& st, c_expression* exp, c_micro_error*& error, var_sym_index_t& sym_return)
{
    sym_return=VAR_SYM_INDEX_INVALID;
    stack<c_eval_iteration>::size_type stack_size=st.size();
    if(!exp->perform(st.top().this_module_,st.top().vars_,error,sym_return,st,st.top().symbol_table_)) {
        c_eval_runtime_error e(st.top().this_module_,error,exp->debug,__FILE__,__my_func__);
        e.report(true);
        st.top().exception_reported=true;
        return false;
    }
    if(stack_size<st.size()) {
        // we got new sub-iteration (e.g. user function/procedure call)
        return true;
    }
    if(VAR_SYM_INDEX_INVALID==sym_return)
        sym_return=exp->result;
    return false;
}

void eval_loop(stack<c_eval_iteration>& st);

my_time_t script_eval_start;               //!< Timestamp of start of script evaluation
my_time_t script_eval_last_irc_loop;       //!< Timestamp of last call to 'irc_loop_heavy_time_consuming_operation()'

/*!
    \brief Reports fatal error that led to stop the script execution
    \version 1
    \date 2008-06-15
    \author VooDooMan
    \param st Evaluation stack
    \param error_str Error string (message)
    \param filn C++ file name ("__FILE__")
    \param funn C++ function name ("__my_func__")
*/
MY_STATIC void eval_report_runtime_error(stack<c_eval_iteration>& st, const string& error_str, const char* filn, const char* funn)
{
    if(st.top().exception_reported)
        return; // not doubled!

    bool report_as_function= NULL!=st.top().fn;
    if(report_as_function) {
        // try function first

        map<long,c_fn_cmd*>::iterator i=st.top().fn->commands.find(st.top().id);
        if(st.top().fn->commands.end()==i) {
            // fallback to procedure
            report_as_function=false;
        } else {
            // a function
            c_eval_runtime_error e(st.top().this_module_,(*i).second->get_debug_string(),error_str,filn,funn);
            e.report(true);
            st.top().exception_reported=true;
        }
    }
    if(!report_as_function) {
        // a procedure
        c_eval_runtime_error e(st.top().this_module_,st.top().proc_cmd.get_debug(),error_str,filn,funn);
        e.report(true);
        st.top().exception_reported=true;
    }
}

/*!
    \brief Checks limits of script (timeout and maximul number of stack frames)
    \version 2
    \date 2006-02-26, 2008-06-15
    \author VooDooMan
    \param st Stack
    \param start Timestamp of script's run
    \param eval_script_timeout Script execution timeout in seconds
    \param eval_stack_size Maximum number of stack frames for script
    \param stack_overflow [in/out] If there was maximum stack frames limit already reported
*/
void eval_check_limits(stack<c_eval_iteration>& st, my_time_t& start, const my_time_t eval_script_timeout, unsigned eval_stack_size, bool& stack_overflow)
{
    if(static_cast<MY_LONG_LONG>(start)+eval_script_timeout<=static_cast<MY_LONG_LONG>(my_time(NULL))) {
        //st.top().result_e_exp_exception=true;
        //st.top().exception1=EXP_E_TIMEOUT;
        if(!st.top().exception_reported) {
            if(st.top().this_module_ && st.top().this_module_->get_conf()
                && st.top().this_module_->get_conf()->getvar_int("sys_startup_script_timeout_protect")!=0) {
                    c_logger* l=st.top().this_module_->get_logger();
                    if(!l)
                        l=get_core_logger();
                    if(l) {
                        my_time_t how_long=my_time(NULL)-start;
                        string log=((string)"WARNING: Ignoring script evaluation timeout after ")+tos(how_long)+" seconds of execution, according to module's \"conf.txt\" \"sys_startup_script_timeout_protect\" rule relaxation";

                        l->log_print1(E_LOG_BOT,LOG_SCOPE_MODULE,"",log);
                        l->log_print1(E_LOG_DEBUG,LOG_SCOPE_MODULE,"",log);
                    }

                    start=my_time(NULL); // reset
            } else {
                eval_report_runtime_error(st,c_eval_runtime_error::get_script_timeout_error_message(eval_script_timeout),__FILE__,__my_func__);
            }
        }
    }
    if(st.size()>eval_stack_size || stack_overflow) {
        //st.top().result_e_exp_exception=true;
        //st.top().exception1=EXP_E_STACK_OVERFLOW;
        stack_overflow=true;
        eval_report_runtime_error(st,c_eval_runtime_error::get_script_stack_overflow_error_message(eval_stack_size),__FILE__,__my_func__);
    }

    if(my_difftime(my_time(NULL),script_eval_last_irc_loop)>=5) {
        irc_loop_heavy_time_consuming_operation("Still executing script...");
        script_eval_last_irc_loop=my_time(NULL);
    }
}

#if defined(_MSC_VER)
#   pragma warning(push)
#   pragma warning(disable:4702)    // warning C4702: unreachable code
#endif

void eval_proc(stack<c_eval_iteration>& st, my_time_t UNUSEDARG(start), const my_time_t eval_script_timeout, unsigned eval_stack_size, bool& stack_overflow)
{
    static bool in_exception=false;
    if(in_exception)
        return;

    if(st.top().exception_reported) {
        bool exception_reported=st.top().exception_reported;
        c_micro_error* micro_error=st.top().micro_error;
        string error=st.top().error;
        if(st.top().fn_destroy) {
            microcode_perform_updates(st);
            protect_globals(st.top().this_module_);

            if(st.top().fn_destroy_mc)
                st.top().fn->clear_microcode();
            st.top().fn->destroy(true);
        }

        eval_stack_epilog(st);
        c_eval_iteration top=st.top();
        st.pop();
        st.top().state=st.top().return_to_state;
        microcode_perform_updates(st);
        protect_globals(top.this_module_);
        if(top.vars_!=st.top().vars_) {
            var_keeper->free_global(top.vars_,true);
            top.vars_=NULL;
        }
        if(top.symbol_table_!=st.top().symbol_table_) {
            MY_DELETE(top.symbol_table_);
            top.symbol_table_=NULL;
        }

        st.top().exception_reported=exception_reported;
        st.top().state=0;
        st.top().result=VAR_SYM_INDEX_INVALID;
        st.top().func_call_ok=true;
        st.top().is_return_=true;
        st.top().processed=true;

        st.top().error=error;
        st.top().micro_error=micro_error;

        return;
    }

    string proc_name=st.top().proc.name;

    if(4000==st.top().state) {
        st.top().proc_iter=st.top().proc.commands.begin();
        st.top().state=4001;
    } else {
        if(st.top().proc.commands.end()==st.top().proc_iter) {
            microcode_perform_updates(st);
            protect_globals(st.top().this_module_);
            if(st.top().vars_) {
                var_keeper->free_global(st.top().vars_,true);
                st.top().vars_=NULL;
            }

            bool exception_reported=st.top().exception_reported;
            if(st.top().fn_destroy) {
                if(st.top().fn_destroy_mc)
                    st.top().fn->clear_microcode();
                st.top().fn->destroy(true);
            }

            eval_stack_epilog(st);
            c_eval_iteration top=st.top();
            st.pop();
            st.top().state=st.top().return_to_state;
            microcode_perform_updates(st);
            protect_globals(st.top().this_module_);
            if(top.vars_!=st.top().vars_) {
                var_keeper->free_global(top.vars_,true);
                top.vars_=NULL;
            }
            if(top.symbol_table_!=st.top().symbol_table_) {
                MY_DELETE(top.symbol_table_);
                top.symbol_table_=NULL;
            }

            st.top().exception_reported=exception_reported;
            st.top().state=0;
            st.top().result=VAR_SYM_INDEX_INVALID;
            st.top().func_call_ok=true;
            st.top().is_return_=true;
            st.top().processed=true;
            return;
        }
        if(st.top().proc.commands.end()==st.top().proc_iter) {
        end:

            bool exception_reported=st.top().exception_reported;
            if(st.top().fn_destroy) {
                microcode_perform_updates(st);
                protect_globals(st.top().this_module_);

                if(st.top().fn_destroy_mc)
                    st.top().fn->clear_microcode();
                st.top().fn->destroy(true);
            }

            if(st.size()!=1) {
                eval_stack_epilog(st);
                c_eval_iteration top=st.top();
                st.pop();
                st.top().state=st.top().return_to_state;
                microcode_perform_updates(st);
                protect_globals(top.this_module_);
                if(top.vars_!=st.top().vars_) {
                    var_keeper->free_global(top.vars_,true);
                    top.vars_=NULL;
                }
            }
            st.top().exception_reported=exception_reported;
            st.top().state=0;
            st.top().result=VAR_SYM_INDEX_INVALID;
            st.top().func_call_ok=true;
            st.top().is_return_=true;
            st.top().processed=true;
            return;
        }

        goto cont;
    }

    st.top().return_to_state=4001;

    st.top().do_redirect=false;
    for(;;) {
        if(st.top().do_redirect) {
            st.top().proc_iter=st.top().redirect_checkpoint_iter;
            st.top().do_redirect=false;
        }

        for(; st.top().proc_iter!=st.top().proc.commands.end(); ++st.top().proc_iter) {
    again:
            {
                string debug=(*st.top().proc_iter).get_debug();
                int command=(*st.top().proc_iter).get_command();
                if(_op==command || _deop==command || _voice==command || _devoice==command) {
                    if(irc_got_op(logic_eval(st.top().this_module_,(*st.top().proc_iter).get_channel(),st.top().vars_,st.top().proc_cmd.get_debug()))) {

                        char prefix='+';
                        char action='0';
                        if(_op==command)
                            action='o';
                        if(_voice==command)
                            action='v';
                        if(_deop==command) {
                            action='o';
                            prefix='-';
                        }
                        if(_devoice==command) {
                            action='v';
                            prefix='-';
                        }
                        logic_user_mode_changing(logic_eval(st.top().this_module_,(*st.top().proc_iter).get_nick(),st.top().vars_,st.top().proc_cmd.get_debug()),logic_eval(st.top().this_module_,(*st.top().proc_iter).get_channel(),st.top().vars_,debug),prefix,action);
                    }
                }

                stack<c_eval_iteration>::size_type size=st.size();

                st.top().proc_cmd=*(st.top().proc_iter);
                try {
                    eval_evaluate_proc(st);
                } catch(c_proc_exc &e) {
                    c_eval_runtime_error ee(st.top().this_module_,e,__FILE__,__my_func__);
                    ee.report(true);
                    // workaround with fake exception
                    st.top().exception_reported=true;
                }

                eval_check_limits(st,script_eval_start,eval_script_timeout,eval_stack_size,stack_overflow);

                if(st.top().exception_reported) {
                    int state=st.top().return_to_state;
                    bool exception_reported=st.top().exception_reported;

                    if(st.size()==1) {
                        return;
                    }
                    bool is_return=st.top().is_return_;
                    bool skip_block=st.top().skip_block;
                    bool redirect_checkpoint=st.top().redirect_checkpoint;
                    bool after_redirect=st.top().after_redirect;
                    int goto_line=st.top().goto_line;

                    if(st.top().fn_destroy) {
                        microcode_perform_updates(st);
                        protect_globals(st.top().this_module_);

                        if(st.top().fn_destroy_mc)
                            st.top().fn->clear_microcode();
                        st.top().fn->destroy(true);
                    }

                    eval_stack_epilog(st);
                    c_eval_iteration top=st.top();
                    st.pop();
                    st.top().state=st.top().return_to_state;
                    microcode_perform_updates(st);
                    protect_globals(top.this_module_);
                    if(top.vars_!=st.top().vars_) {
                        var_keeper->free_global(top.vars_,true);
                        top.vars_=NULL;
                    }
                    if(top.symbol_table_!=st.top().symbol_table_) {
                        MY_DELETE(top.symbol_table_);
                        top.symbol_table_=NULL;
                    }

                    st.top().exception_reported=exception_reported;
                    st.top().state=state;

                    st.top().is_return_=is_return;
                    st.top().skip_block=skip_block;
                    st.top().redirect_checkpoint=redirect_checkpoint;
                    st.top().after_redirect=after_redirect;
                    st.top().goto_line=goto_line;
                    return;
                }

                if(st.top().state<3000 || st.top().state>4999) {
                    return;
                }
                if(st.size()!=size) {
                    st.top().proc_iter=st.top().proc.commands.begin();
                    goto again;
                }
            }
        cont:
            if(st.top().r_checkpoint)
                st.top().redirect_checkpoint_iter=st.top().proc_iter;
            if(!st.top().after_redirect) {
                if(st.top().is_return_) {
                    st.top().proc.lock_cnt--;

                    st.top().state=st.top().return_to_state;
                    st.top().state=0;
                    st.top().result=VAR_SYM_INDEX_INVALID;
                    st.top().func_call_ok=true;
                    st.top().is_return_=true;
                    st.top().processed=true;
                    goto end;
                }
                if(st.top().goto_line!=-1) {
                    st.top().proc_iter=st.top().proc.commands.begin();
                    while(st.top().proc_iter!=st.top().proc.commands.end() && (*st.top().proc_iter).get_line()!=st.top().goto_line)
                        st.top().proc_iter++;
                    st.top().goto_line=-1;
                } else {
                    if(_eval==(*st.top().proc_iter).get_command()) {
                        // set eval's return value
                        c_var_base* rv=st.top().return_value;
                        if(!rv)
                            rv=MY_NEW c_var_unset();
                        string ret_var=st.top().proc_cmd.get_str_arg(1);

                        {
                            c_index_base* idx=MY_NEW c_index_str(ret_var);
                            c_var_base* val=rv;
                            try {
                                c_index_base* idx2=idx->copy();
                                st.top().vars_->unset(st.top().this_module_,idx2,st.top().proc_cmd.get_debug());
                                c_var_base* test=st.top().vars_->get(st.top().this_module_,idx,false,"");
                                if(test) {
                                    // it is read-only, so delete
                                    MY_DELETE(val);
                                    val=NULL;
                                } else {
                                    // ok to set
                                    st.top().vars_->set(st.top().this_module_,idx,val,false,st.top().proc_cmd.get_debug());
                                }
                            } catch(c_var_exception e) {
                                c_eval_runtime_error ee(st.top().this_module_,e,__FILE__,__my_func__);
                                ee.report(true);
                                if(val)
                                    MY_DELETE(val);
                            }
                        }
                    }
                    st.top().proc_iter++;
                }
                if(st.top().proc.commands.end()==st.top().proc_iter) {
                    goto end;
                }
            } else
                st.top().do_redirect=true;

            goto again;
        }
        if(st.top().do_redirect)
            continue;
        break;
    }

    bool exception_reported=st.top().exception_reported;
    if(st.top().fn_destroy) {
        microcode_perform_updates(st);
        protect_globals(st.top().this_module_);

        if(st.top().fn_destroy_mc)
            st.top().fn->clear_microcode();
        st.top().fn->destroy(true);
    }

    eval_stack_epilog(st);
    c_eval_iteration top=st.top();
    st.pop();
    if(!st.empty()) {
        st.top().state=st.top().return_to_state;
        microcode_perform_updates(st);
    }
    protect_globals(top.this_module_);
    if(st.empty() || top.vars_!=st.top().vars_) {
        var_keeper->free_global(top.vars_,true);
        top.vars_=NULL;
    }
    if(st.empty() || top.symbol_table_!=st.top().symbol_table_) {
        MY_DELETE(top.symbol_table_);
        top.symbol_table_=NULL;
    }

    if(!st.empty()) {
        st.top().exception_reported=exception_reported;
        st.top().state=0;
        st.top().result=VAR_SYM_INDEX_INVALID;
        st.top().func_call_ok=true;
        st.top().is_return_=true;
        st.top().processed=true;
    }
}

#if defined(_MSC_VER)
#   pragma warning(pop)
#endif

/*!
    \brief Resets evaluation timeout (start of evaluation timestamp)
    \version 1
    \date 2006-02-26
    \author VooDooMan
*/
void eval_cleanup_timeout()
{
    script_eval_start=my_time(NULL);
}

/*!
    \brief Processes and evaluates procedures and functions
    \version 1
    \date 2006-01, 2006-02
    \author VooDooMan
    \param st Stack with one initialised frame as input
*/
void eval_loop(stack<c_eval_iteration>& st)
{
    c_global_keeper::c_session_keeper sess_keeper;
    var_keeper->set_current_session(st.top().this_module_->get_session());

    (void)var_keeper->get_current_local();
    if(st.top().local_id)
        var_keeper->set_current_local(st.top().local_id);
    c_conf* conf=st.top().this_module_->get_conf();
    my_time_t eval_script_timeout=10;
    unsigned eval_stack_size=200;
    if(conf) {
        eval_script_timeout=conf->get_eval_script_timeout();
        eval_stack_size=conf->get_eval_stack_size();
    }

    bool stack_overflow=false;
    script_eval_start=my_time(NULL);
    script_eval_last_irc_loop=my_time(NULL);

    for(;;) {
        if(st.empty())
            break;

        eval_check_limits(st,script_eval_start,eval_script_timeout,eval_stack_size,stack_overflow);

        if(st.top().exception_reported) {
            break;
        }

        if(st.top().micro_error) {
            c_eval_runtime_error e(st.top().this_module_,st.top().micro_error->get_msg(),"",__FILE__,__my_func__);
            e.report(true);
            MY_DELETE(st.top().micro_error);
            st.top().micro_error=NULL;
            st.top().exception_reported=true;
            break;
        }

        if((0==st.top().state || st.top().is_return_) && st.size()==1) {
            break;
        }

        if(!st.top().state)
            st.top().state=st.top().return_to_state;

        if(7000==st.top().state) {
            // evaluating microcode stream
            stack<c_eval_iteration>::size_type stack_size=st.size();
            var_sym_index_t sym_return;
            if(eval_microcode(st,st.top().c->get_exp(),st.top().micro_error,sym_return)) {
                continue;
            }
            if(stack_size<st.size()) {
                // we got new sub-iteration (e.g. user function/procedure call)
                continue;
            }
            microcode_perform_updates(st);

            microcode_perform_updates(st);
            c_var_base* return_value=st.top().return_value;
            var_sym_index_t retval_store_to=st.top().retval_store_to;
            eval_stack_epilog(st);
            bool exception_reported=st.top().exception_reported;
            eval_stack_epilog(st);
            c_eval_iteration top=st.top();
            st.pop();
            st.top().exception_reported=exception_reported;
            if(return_value && retval_store_to!=VAR_SYM_INDEX_INVALID) {
                return_value=return_value->copy();
                bool dummy_b=false;
                c_micro_error* error=NULL;
                if(st.top().symbol_table_->get_var_table())
                    if(!st.top().symbol_table_->get_var_table()->set(retval_store_to,return_value,dummy_b,error,MICRO_FUNC_CALL,false,st.top().this_module_,"","")) {
                        MY_DELETE(error);
                    }
            }
            st.top().state=st.top().return_to_state;
            microcode_perform_updates(st);
            protect_globals(top.this_module_);
            if(top.vars_!=st.top().vars_) {
                var_keeper->free_global(top.vars_,true);
                top.vars_=NULL;
            }
            if(top.symbol_table_!=st.top().symbol_table_) {
                MY_DELETE(top.symbol_table_);
                top.symbol_table_=NULL;
            }

            st.top().state=st.top().return_to_state;
            continue;
        }
        if((st.top().state<1000 && st.top().state!=0) || (st.top().state==0 && !st.top().processed)) {
            st.top().state=3001;
        }
        if(st.top().state>=1000 && st.top().state<=1999) {
            eval_function(st,script_eval_start,eval_script_timeout,eval_stack_size,stack_overflow);
            
            var_keeper->set_current_session(st.top().this_module_->get_session());

            eval_check_limits(st,script_eval_start,eval_script_timeout,eval_stack_size,stack_overflow);

            // log exception to logs

            if(!st.top().func_call_ok)
                st.top().exception_reported=true;

            if(st.top().processed) {
                microcode_perform_updates(st);
                protect_globals(st.top().this_module_);

                if(st.top().fn_destroy) {
                    if(st.top().fn_destroy_mc)
                        st.top().fn->clear_microcode();
                    st.top().fn->destroy(true);
                }
                if(st.size()==1)
                    break;

                c_var_base* return_value=st.top().return_value;
                var_sym_index_t retval_store_to=st.top().retval_store_to;
                bool exception_reported=st.top().exception_reported;
                {
                    //eval_stack_epilog(st); // do NOT!
                    // just this only:
                    if(st.top().parent && st.top().symbol_table_ && st.top().parent->symbol_table_)
                        if(st.top().symbol_table_->get_var_table())
                            st.top().symbol_table_->get_var_table()->commit_changes(*st.top().parent->symbol_table_->get_var_table());
                }
                eval_stack_epilog(st);
                c_eval_iteration top=st.top();
                st.pop();
                st.top().exception_reported=exception_reported;
                if(return_value) {
                    return_value=return_value->copy();
                }
                st.top().return_value=return_value;
                if(return_value && retval_store_to!=VAR_SYM_INDEX_INVALID) {
                    return_value=return_value->copy(); // again!
                    bool dummy_b=false;
                    c_micro_error* error=NULL;
                    if(st.top().symbol_table_->get_var_table())
                        if(!st.top().symbol_table_->get_var_table()->set(retval_store_to,return_value,dummy_b,error,MICRO_FUNC_CALL,false,st.top().this_module_,"","")) {
                            MY_DELETE(error);
                        }
                }
                st.top().state=st.top().return_to_state;
                microcode_perform_updates(st);
                protect_globals(top.this_module_);
                if(top.vars_!=st.top().vars_ && !top.proc_eval) {
                    var_keeper->free_global(top.vars_,true);
                    top.vars_=NULL;
                }
                st.top().state=st.top().return_to_state;

                if(st.top().local_id)
                    var_keeper->set_current_local(st.top().local_id);
            }
            continue;
        }
        if(st.top().state>=3000 && st.top().state<4000) {
            if(3000==st.top().state) {
                size_t size=st.size();
                try {
                    eval_evaluate_proc(st);
                } catch(c_proc_exc &e) {
                    c_eval_runtime_error ee(st.top().this_module_,e,__FILE__,__my_func__);
                    ee.report(true);
                    st.top().exception_reported=true;
                }

                var_keeper->set_current_session(st.top().this_module_->get_session());

                eval_check_limits(st,script_eval_start,eval_script_timeout,eval_stack_size,stack_overflow);

                // log exception to logs

                if(size==st.size()) {
                    int state=st.top().return_to_state;
                    bool exception_reported=st.top().exception_reported;

                    if(st.top().fn_destroy) {
                        microcode_perform_updates(st);
                        protect_globals(st.top().this_module_);

                        if(st.top().fn_destroy_mc)
                            st.top().fn->clear_microcode();
                        st.top().fn->destroy(true);
                    }
                    st.top().vars_=NULL;
                    microcode_perform_updates(st);
                    protect_globals(st.top().this_module_);
                    MY_DELETE(st.top().symbol_table_);
                    st.top().symbol_table_=NULL;

                    if(st.size()==1) {
                        break;
                    }
                    bool is_return=st.top().is_return_;
                    bool skip_block=st.top().skip_block;
                    bool redirect_checkpoint=st.top().redirect_checkpoint;
                    bool after_redirect=st.top().after_redirect;
                    int goto_line=st.top().goto_line;

                    eval_stack_epilog(st);
                    c_eval_iteration top=st.top();
                    st.pop();
                    st.top().state=st.top().return_to_state;
                    microcode_perform_updates(st);
                    protect_globals(top.this_module_);
                    if(top.vars_!=st.top().vars_) {
                        var_keeper->free_global(top.vars_,true);
                        top.vars_=NULL;
                    }
                    if(top.symbol_table_!=st.top().symbol_table_) {
                        MY_DELETE(top.symbol_table_);
                        top.symbol_table_=NULL;
                    }

                    st.top().exception_reported=exception_reported;
                    st.top().state=state;

                    st.top().is_return_=is_return;
                    st.top().skip_block=skip_block;
                    st.top().redirect_checkpoint=redirect_checkpoint;
                    st.top().after_redirect=after_redirect;
                    st.top().goto_line=goto_line;
                }

                continue;
            }
            if(3001==st.top().state) {
                var_sym_index_t retval=st.top().result;
                string ret_var=st.top().proc_eval_ret_var;
                bool error=!st.top().func_call_ok;
                bool is_return=st.top().is_return_;

                c_var_base* return_value=NULL;
                c_micro_error* er=NULL;
                bool dummy;
                if(retval!=VAR_SYM_INDEX_DISCARD)
                    if(st.top().symbol_table_ && st.top().symbol_table_->get_var_table())
                        if(retval!=VAR_SYM_INDEX_INVALID && st.top().symbol_table_ && !st.top().symbol_table_->get_var_table()->get(retval,return_value,dummy,er))
                            error=true;

                if(error) {
                    st.top().func_call_ok=false;
                }

                {
                    int state=st.top().return_to_state;
                    bool exception_reported=st.top().exception_reported;

                    if(st.top().fn_destroy) {
                        microcode_perform_updates(st);
                        protect_globals(st.top().this_module_);

                        if(st.top().fn_destroy_mc)
                            st.top().fn->clear_microcode();
                        st.top().fn->destroy(true);
                    }
                    if(st.size()==1) {
                        break;
                    }

                    eval_stack_epilog(st);
                    c_eval_iteration top=st.top();
                    st.pop();
                    st.top().state=st.top().return_to_state;
                    microcode_perform_updates(st);
                    protect_globals(top.this_module_);
                    if(top.vars_!=st.top().vars_) {
                        var_keeper->free_global(top.vars_,true);
                        top.vars_=NULL;
                    }
                    if(top.symbol_table_!=st.top().symbol_table_) {
                        MY_DELETE(top.symbol_table_);
                        top.symbol_table_=NULL;
                    }

                    st.top().exception_reported=exception_reported;
                    st.top().state=state;
                    st.top().error=top.error;
                }

                if(error) {
                    st.top().func_call_ok=false;
                }

                if(is_return) {
                    st.top().return_value=return_value;
                    st.top().func_call_ok=true;

                    st.top().iterative_result=EXP_I_NEXT_ITERATION;
                    st.top().processed=true;

                    continue;
                } else {
                    if(!ret_var.empty()) {
                        st.top().return_value=MY_NEW c_var_bool(false);
                        //retval
                        if(st.top().symbol_table_->get_var_table())
                            if(st.top().symbol_table_ && !st.top().symbol_table_->get_var_table()->set(st.top().retval,st.top().return_value,false,st.top().micro_error,MICRO_RETURN,false,st.top().this_module_,"","")) {
                                st.top().func_call_ok=false;
                                continue;
                            }
                    } else {
                        st.top().result=VAR_SYM_INDEX_INVALID;
                    }
                }
                continue;
            }
            if(3999==st.top().state && st.size()==1) {
                break;
            }
            if(3999==st.top().state) {
                bool exception_reported=st.top().exception_reported;

                {
                    microcode_perform_updates(st);
                    protect_globals(st.top().this_module_);
                    MY_DELETE(st.top().symbol_table_);
                    st.top().symbol_table_=NULL;
                    var_keeper->free_global(st.top().vars_,true);
                    st.top().vars_=NULL;
                }

                if(st.size()==1) {
                    break;
                }

                if(st.top().fn_destroy) {
                    microcode_perform_updates(st);
                    protect_globals(st.top().this_module_);

                    if(st.top().fn_destroy_mc)
                        st.top().fn->clear_microcode();
                    st.top().fn->destroy(true);
                }

                eval_stack_epilog(st);
                c_eval_iteration top=st.top();
                st.pop();
                st.top().state=st.top().return_to_state;
                microcode_perform_updates(st);
                protect_globals(top.this_module_);
                if(top.vars_!=st.top().vars_) {
                    var_keeper->free_global(top.vars_,true);
                    top.vars_=NULL;
                }
                if(top.symbol_table_!=st.top().symbol_table_) {
                    MY_DELETE(top.symbol_table_);
                    top.symbol_table_=NULL;
                }

                st.top().exception_reported=exception_reported;
                st.top().state=4001;
                st.top().error=top.error;
            }
            continue;
        }
        if(st.top().state>=4000 && st.top().state<5000) {
            if(!st.top().symbol_table_) {
                c_micro_symbol_variable_table* vars=MY_NEW c_micro_symbol_variable_table(0);
                c_micro_symbol_variable_mapping* symbols=MY_NEW c_micro_symbol_variable_mapping();
                st.top().symbol_table_=MY_NEW c_micro_symbol_table(symbols,vars);
                st.top().symbol_table_to_destroy=st.top().symbol_table_;
            }
            eval_proc(st,script_eval_start,eval_script_timeout,eval_stack_size,stack_overflow);

            var_keeper->set_current_session(st.top().this_module_->get_session());

            continue;
        }
        if(0==st.top().state && (st.top().return_to_state>=3000 && st.top().return_to_state<4000)) {
            st.top().state=st.top().return_to_state;
            continue;
        }

        if(st.top().state>=1000 && st.top().state<=1999)
            break; // return to procedure evaluation/execution
    }

    while(st.size()) {
        if(st.top().self_test) {
            if(st.size()==1 && st.top().return_value)
                st.top().return_value=st.top().return_value->copy();
        }
        if(st.top().fn_destroy) {
            microcode_perform_updates(st);
            protect_globals(st.top().this_module_);

            if(st.top().fn_destroy_mc)
                st.top().fn->clear_microcode();
            st.top().fn->destroy(true,true);
        }

        if(st.top().proc_eval && st.top().fn)
            st.top().fn->destroy_symbols();

        size_t st_size=st.size();
        if(st_size!=1) {
            eval_stack_epilog(st);
        } else {
            protect_globals(st.top().this_module_);
            eval_stack_epilog(st);
        }
        c_eval_iteration top=st.top();
        if(st_size!=1) {
            st.pop();
        } else {
            if(st.top().fn_destroy) {
                if(st.top().fn_destroy_mc)
                    st.top().fn->clear_microcode();
                st.top().fn->destroy(true,true);
            }
            MY_DELETE(st.top().fn);
            st.top().fn=NULL;

            if(!st.top().self_test) {
                top.symbol_table_=NULL;
                top.vars_=NULL;
            }

            break;
        }

        st.top().state=st.top().return_to_state;
        microcode_perform_updates(st);
        protect_globals(top.this_module_);
        if(top.vars_!=st.top().vars_) {
            var_keeper->free_global(top.vars_,true);
            top.vars_=NULL;
        }
    }
    if(!st.empty()) {
        if(st.top().self_test) {
            return;
        }

        MY_DELETE(st.top().symbol_table_to_destroy);
        st.top().symbol_table_to_destroy=NULL;

        {
            if(st.top().save_return_value) {
                if(st.top().return_value) {
                    session_id_t old=var_keeper->get_current_session();
                    var_keeper->set_current_session(SESSION_RESERVED_ID);
                    st.top().return_value=st.top().return_value->copy();
                    var_keeper->set_current_session(old);
                }
            }
            while(var_keeper->purge(st.top().this_module_->get_session())) {
                ;
            }
            //var_keeper->_dump("dump.txt",st.top().this_module_);
        }
    }
}

void eval_proc(stack<c_eval_iteration>& st)
{
    eval_loop(st);
}

