/***************************************************************************
                          exec.cpp  -  description
                             -------------------
    begin                : Thu Dec 9 2004
    copyright            : (C) 2004-2007 by Marian VooDooMan Meravy
    email                : ghostvoodooman -aT= users (dot) sourceforge.net
 ***************************************************************************/

/* $Id: exec.cpp,v 1.55 2008/03/11 21:11:28 ghostvoodooman Exp $ */

/***************************************************************************

VooDoo cIRCle - an IRC (ro)bot
Copyright (C) 2004-2008 by Marian VooDooMan Meravy
  (ghostvoodooman -aT= users (dot) sourceforge.net)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************/

/*!
    \file
    \brief Handles child processes. Layer for platform abstraction
    \version $Id: exec.cpp,v 1.55 2008/03/11 21:11:28 ghostvoodooman Exp $
*/

#if defined(HAVE_CONFIG_H)
#   include <config.h>
#endif

#include <string>
#include <vector>
#include <stack>
#include <list>
#include <map>

#include <stdlib.h>
#if defined(HAVE_MALLOC_H)
#   include <malloc.h>
#endif
#include <string>
#if !defined(_WIN32) && !defined(_WIN64)
#   include <sys/types.h>
#   include <sys/stat.h>
#   include <sys/poll.h>
#   include <stdio.h>
#else
#   include <windows.h>
#endif
#include <errno.h>

using namespace std;

#include "utils/memleak.h"
#include "utils/assert.h"
#include "utils.h"

#include "exec.h"

#include "params.h"

#include "utils/precomp.h"

#if defined(_MSC_VER) || defined(__BORLANDC__)
#   pragma hdrstop
#endif

//---------------------------------------------------------------------------

/*!
    \brief Reads console output (stdout) from executed command
    \author VooDooMan
    \version 1
    \date 2004
    \param h Handle returned by exec_async_exec
    \param buffer Returns address of buffer, without NUL terminator (RAW)
    \warning buffer MUST be deallocated by free()
    \param buffer_size Returns lpBuffer length in bytes
    \return Returns number of read bytes (=buffer_size)
*/
int exec_read_output(c_exec_handle* h, char*& buffer, int& buffer_size)
{
    if(!h->out_buffer) {
        buffer=(char*)my_malloc(1);
        return buffer_size=0;
    }
    buffer=h->out_buffer;
    buffer_size=h->out_buffer_size;
    h->out_buffer=NULL;
    h->out_buffer_size=0;
    return buffer_size;
}

/*!
    \brief Executes command asynchronously
    \author VooDooMan
    \version 1
    \date 2004
    \param cmd Executable
    \param params Command-line parameters
    \param dir Full path name of working directory of new process
    \return Returns handle to new process
*/
c_exec_handle* exec_async_exec(const string& cmd, const string& params, const string& dir)
{
#if defined(_WIN32) || defined(_WIN64)
    _STARTUPINFOA SI;
    memset(&SI,0,sizeof(SI));
    GetStartupInfoA(&SI);
    SI.dwFlags=STARTF_USESHOWWINDOW | STARTF_FORCEOFFFEEDBACK | STARTF_USESTDHANDLES;
    SI.wShowWindow=SW_HIDE;
    string ln=cmd;
    if(ln.length()>0 && ln[0]!='"' && ln.find(0x20,0)!=string::npos)
        ln=(string)"\""+ln+"\"";
    ln+=" ";
    ln+=params;

    c_exec_handle* h=MY_NEW c_exec_handle;
    memset(&(h->PI),0,sizeof(h->PI));

    {
        HANDLE hOutputReadTmp,hOutputRead,hOutputWrite;
        HANDLE hInputWriteTmp,hInputRead,hInputWrite;
        HANDLE hErrorWrite;
        SECURITY_ATTRIBUTES sa;
        memset(&sa,0,sizeof(sa));
        sa.nLength=sizeof(sa);
        sa.lpSecurityDescriptor=NULL;
        sa.bInheritHandle=TRUE;
        CreatePipe(&hOutputReadTmp,&hOutputWrite,&sa,0);
        DuplicateHandle(GetCurrentProcess(),hOutputWrite,GetCurrentProcess(),&hErrorWrite,0,TRUE,DUPLICATE_SAME_ACCESS);
        CreatePipe(&hInputRead,&hInputWriteTmp,&sa,0);
        DuplicateHandle(GetCurrentProcess(),hOutputReadTmp,GetCurrentProcess(),&hOutputRead,0,FALSE,DUPLICATE_SAME_ACCESS);
        DuplicateHandle(GetCurrentProcess(),hInputWriteTmp,GetCurrentProcess(),&hInputWrite,0,FALSE,DUPLICATE_SAME_ACCESS);
        CloseHandle(hOutputReadTmp);
        CloseHandle(hInputWriteTmp);

        h->hOutputWrite=hOutputWrite;
        h->hOutputRead=hOutputRead;

        h->hInputRead=hInputRead;
        h->hInputWrite=hInputWrite;

        h->hErrorWrite=hErrorWrite;

        SI.hStdOutput=h->hOutputWrite;
        SI.hStdInput=h->hInputRead;
        SI.hStdError=h->hErrorWrite;
    }
    char fully_qualified_path[_MAX_PATH+128+1];
    LPTSTR* dummy=NULL;
    if(GetFullPathName(dir.c_str(),sizeof(fully_qualified_path)-1,fully_qualified_path,dummy))
        h->error=!CreateProcessA(NULL,(LPSTR)ln.c_str(),NULL,NULL,true,NORMAL_PRIORITY_CLASS,NULL,(LPSTR)fully_qualified_path,&SI,&(h->PI));
    else
        h->error=!CreateProcessA(NULL,(LPSTR)ln.c_str(),NULL,NULL,true,NORMAL_PRIORITY_CLASS,NULL,NULL,&SI,&(h->PI));

    h->os_hnd=h->PI.hProcess;
    return h;
#else
    if(cmd.length()==0) {
        c_exec_handle* h=MY_NEW c_exec_handle;
        h->os_hnd=NULL;
        h->error=true;
        h->out_buffer=NULL;
        h->out_buffer_size=0;
        return h;
    }
    string ln;
    if(dir.length()) {
        ln+="cd ";
        if(dir.length() && dir[0]!='"')
            ln+="\"";
        ln+=dir;
        if(dir.length() && dir[0]!='"')
            ln+="\"";
        ln+="; ";
    }
    ln+=cmd;
    if(params.length()) {
        ln+=" ";
        ln+=params;
    }
    FILE* hnd=popen(ln.c_str(),"r");
    c_exec_handle* h=MY_NEW c_exec_handle;
    h->os_hnd=hnd;
    h->error=!hnd;
    h->out_buffer=NULL;
    h->out_buffer_size=0;
    return h;
#endif
}

/*!
    \brief Checks wheter process is still active
    \param h Handle of process
    \return Returns true if process is still active, false if has exited
*/
bool exec_process_active(c_exec_handle* h)
{
    if(h==NULL)
        return false;
    if(h->os_hnd==NULL)
        return false;
#if defined(_WIN32) || defined(_WIN64)
    DWORD dwExitCode;
    GetExitCodeProcess(h->os_hnd,&dwExitCode);
    if(dwExitCode!=STILL_ACTIVE) {
        HANDLE hPipeRead=h->hOutputRead;

        DWORD nBytesRead;

        int buffer_size=GetFileSize(hPipeRead,NULL);
        if(buffer_size>0) {
            h->out_buffer=(char*)my_malloc(buffer_size);
            h->out_buffer_size=buffer_size;
            ReadFile(hPipeRead,h->out_buffer,h->out_buffer_size,&nBytesRead,NULL);
        } else {
            // hmm...
            h->out_buffer=(char*)my_malloc(1);
            h->out_buffer_size=0;
        }
        exec_process_end(h);
        h->os_hnd=NULL;
        h->exit_code=dwExitCode;
    }
    return dwExitCode==STILL_ACTIVE;
#else
    pollfd to_poll;
    memset(&to_poll,0,sizeof(to_poll));
    to_poll.fd=fileno(h->os_hnd);
    to_poll.events=POLLIN|POLLPRI|POLLHUP|POLLERR;
    /* Retval will always be greater than 0 or -1 in this case.
       Since we're doing it while blocking */
    if(poll(&to_poll,1,1)<0) {
        h->out_buffer=(char*)my_malloc(1);
        h->out_buffer_size=0;
        exec_process_end(h);
        h->os_hnd=NULL;
        h->exit_code=-1;
        return false;
    }
    if((to_poll.revents&POLLIN)==POLLIN || (to_poll.revents&POLLPRI)==POLLPRI) {
        int size=128;
        char* tmp=MY_NEW char[size+1];
    again:
        size=fread(tmp,1,size,h->os_hnd);
        if(size>0) {
            tmp[size]=0;
            if(!h->out_buffer) {
                h->out_buffer=(char*)my_malloc(size);
                h->out_buffer_size=size;
                memcpy(h->out_buffer,tmp,size);
            } else {
                int pos=h->out_buffer_size;
                h->out_buffer_size+=size;
                h->out_buffer=(char*)my_realloc(h->out_buffer,h->out_buffer_size);
                memcpy(&(h->out_buffer[pos]),tmp,size);
            }
            size=128;
            goto again;
        }
        MY_DELETE_A(tmp);
    }
    if((to_poll.revents&POLLHUP)==POLLHUP || (to_poll.revents&POLLERR)==POLLERR) {
        h->exit_code=pclose(h->os_hnd);
        h->os_hnd=NULL;
        int pos=h->out_buffer_size;
        h->out_buffer_size+=1;
        h->out_buffer=(char*)my_realloc(h->out_buffer,h->out_buffer_size);
        h->out_buffer[pos]=0;
        return false;
    }
    return true;
#endif
    // NOTREACH
}

/*!
    \brief Gets exit code of terminated process
    \param h Handle of process
    \return Returns exit code
*/
DWORD exec_get_exit_code(c_exec_handle* h)
{
    if(!h)
        return (DWORD)-1;
    if(h->os_hnd==NULL)
        return h->exit_code;
    if(exec_process_active(h))
        return (DWORD)-2;
#if !defined(_WIN32) && !defined(_WIN64)
    h->exit_code=pclose(h->os_hnd);
    h->os_hnd=NULL;
    return h->exit_code;
#else
    DWORD dwExitCode;
    GetExitCodeProcess(h->os_hnd,&dwExitCode);
    if(dwExitCode==STILL_ACTIVE)
        return (DWORD)-2;
    return dwExitCode;
#endif
}

/*!
    \brief Deallocates resorces created by exec_async_exec
    \author VooDooMan
    \version 2
    \date 2005
    \param h Handle of process
    \return Returns exit code of process
*/
int exec_process_end(c_exec_handle* h)
{
#if defined(_WIN32) || defined(_WIN64)
    if(!h)
        return 0;
    if(!h->os_hnd)
        return h->exit_code;

    DWORD dwExitCode;
    GetExitCodeProcess(h->os_hnd,&dwExitCode);

    if(h->hOutputWrite)
        CloseHandle(h->hOutputWrite);
    h->hOutputWrite=NULL;

    if(h->hOutputRead)
        CloseHandle(h->hOutputRead);
    h->hOutputRead=NULL;

    if(h->hInputWrite)
        CloseHandle(h->hInputWrite);
    h->hInputWrite=NULL;

    if(h->hInputRead)
        CloseHandle(h->hInputRead);
    h->hInputRead=NULL;

    if(h->hErrorWrite)
        CloseHandle(h->hErrorWrite);
    h->hErrorWrite=NULL;

    try { // for case it is closed externally
        CloseHandle(h->PI.hProcess);
    } catch(...) {}
    try { // for case it is closed externally
        CloseHandle(h->PI.hThread);
    } catch(...) {}
    /*bool ok=*/TerminateProcess(h->os_hnd,0);

    return dwExitCode;
#else
    if(!h)
        return -1;
    if(h->os_hnd)
        h->exit_code=pclose(h->os_hnd);
    h->os_hnd=NULL;
    if(h->out_buffer) {
        my_free(h->out_buffer);
        h->out_buffer=NULL;
        h->out_buffer_size=0;
    }
    return h->exit_code;
#endif
}

/*!
    \brief Hardly terminates the child process
    \author VooDooMan
    \version 1
    \date 2004
    \param h Process descriptor
    \param error_code Error code to end the process with
*/
void exec_terminate_process(c_exec_handle* h, int error_code)
{
#if defined(_WIN32) || defined(_WIN64)
    TerminateProcess(h->os_hnd,error_code);
    CloseHandle(h->PI.hProcess);
    CloseHandle(h->PI.hThread);
    h->os_hnd=NULL;
    if(h->out_buffer) {
        my_free(h->out_buffer);
        h->out_buffer=NULL;
        h->out_buffer_size=0;
    }
#else
    if(h->os_hnd) {
        h->exit_code=pclose(h->os_hnd);
        h->os_hnd=NULL;
        if(h->out_buffer) {
            my_free(h->out_buffer);
            h->out_buffer=NULL;
            h->out_buffer_size=0;
        }
    }
#endif
}

/*!
    \brief Executes the process without capturing output console (stdout)
    \author VooDooMan
    \version 1
    \date 2004
    \param cmd Path to process, cannot be NULL
    \param params Parameters to command line, can be NULL
    \param dir Working direstory of new process, can be NULL
    \return Returns handle to child process (need to be deallocated by MY_DELETE!)
    \warning Returned handle need to be deallocated by MY_DELETE!
*/
c_exec_handle* exec_async_exec2(const string& cmd, const string& params, const string& dir)
{
#if defined(_WIN32) || defined(_WIN64)
    _STARTUPINFOA SI;
    memset(&SI,0,sizeof(SI));
    GetStartupInfoA(&SI);
    SI.dwFlags=STARTF_USESHOWWINDOW | STARTF_FORCEOFFFEEDBACK | STARTF_USESTDHANDLES;
    SI.wShowWindow=SW_HIDE;
    string ln=cmd;
    if(ln.length()>0 && ln[0]!='"' && ln.find(0x20,0)!=string::npos)
        ln=(string)"\""+ln+"\"";
    if(params.length()) {
        ln+=" ";
        ln+=params;
    }

    c_exec_handle* h=MY_NEW c_exec_handle;
    memset(&(h->PI),0,sizeof(h->PI));

    {
        h->hOutputWrite=NULL;
        h->hOutputRead=NULL;

        h->hInputRead=NULL;
        h->hInputWrite=NULL;

        h->hErrorWrite=NULL;
    }
    char fully_qualified_path[_MAX_PATH+128+1];
    LPTSTR* dummy=NULL;
    if(GetFullPathName(dir.c_str(),sizeof(fully_qualified_path)-1,fully_qualified_path,dummy))
        h->error=!CreateProcessA(NULL,(LPSTR)ln.c_str(),NULL,NULL,false,NORMAL_PRIORITY_CLASS,NULL,(LPSTR)fully_qualified_path,&SI,&(h->PI));
    else
        h->error=!CreateProcessA(NULL,(LPSTR)ln.c_str(),NULL,NULL,false,NORMAL_PRIORITY_CLASS,NULL,NULL,&SI,&(h->PI));

    h->os_hnd=h->PI.hProcess;
    return h;
#else
    if(cmd.length()==0) {
        c_exec_handle* h=MY_NEW c_exec_handle;
        h->os_hnd=NULL;
        h->error=true;
        h->out_buffer=NULL;
        h->out_buffer_size=0;
        return h;
    }
    string ln;
    if(dir.length()) {
        ln+="cd ";
        if(dir.length() && dir[0]!='"')
            ln+="\"";
        ln+=dir;
        if(dir.length() && dir[0]!='"')
            ln+="\"";
        ln+="; ";
    }
    ln+=cmd;
    if(params.length()) {
        ln+=" ";
        ln+=params;
    }
    FILE* hnd=popen(ln.c_str(),"r");
    c_exec_handle* h=MY_NEW c_exec_handle;
    h->os_hnd=hnd;
    h->error=!hnd;
    h->out_buffer=NULL;
    h->out_buffer_size=0;
    return h;
#endif
}

/*!
    \brief Detaches the running sub-process: releases resources while keeping the process running
    \author VooDooMan
    \version 1
    \date 2008-03-07
    \param h Process descriptor
*/
void exec_detach_process(c_exec_handle* h)
{
    h->os_hnd=NULL;
    MY_DELETE(h);
}

