/***************************************************************************
                          exec.h  -  description
                             -------------------
    begin                : Thu Dec 9 2004
    copyright            : (C) 2004-2007 by Marian VooDooMan Meravy
    email                : ghostvoodooman -aT= users (dot) sourceforge.net
 ***************************************************************************/

/* $Id: exec.h,v 1.46 2008/03/11 21:11:28 ghostvoodooman Exp $ */

/***************************************************************************

VooDoo cIRCle - an IRC (ro)bot
Copyright (C) 2004-2008 by Marian VooDooMan Meravy
  (ghostvoodooman -aT= users (dot) sourceforge.net)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************/

/*!
    \file
    \brief Handles child processes. Layer for platform abstraction
    \version $Id: exec.h,v 1.46 2008/03/11 21:11:28 ghostvoodooman Exp $
*/

#ifndef EXEC_H__
#define EXEC_H__

#if defined(HAVE_CONFIG_H)
#   include <config.h>
#endif

#include <string>
#include <vector>
#include <stack>
#include <list>
#include <map>

using namespace std;

#include <stdio.h>

#include "utils/memleak.h"
#include "utils/assert.h"
#include "utils.h"
#include "params.h"

#if defined(_WIN32) || defined(_WIN64)
#   define _WINSOCKAPI_
#   include <windows.h>
#   define popen _popen
#   define pclose _pclose
#endif

#include "utils/assert.h"

/*!
    \brief Stores handle to process and aditional resources
    \author VooDooMan
    \version 1
    \date 2004
*/
struct c_exec_handle
{
#if defined(_WIN32) || defined(_WIN64)
    HANDLE os_hnd;                              //!< Handle of child process
    HANDLE hOutputWrite;                        //!< Handle of stdout for writing
    HANDLE hOutputRead;                         //!< Handle of stdout for reading
    HANDLE hInputWrite;                         //!< Handle of stdin for writing
    HANDLE hInputRead;                          //!< Handle of stdin for reading
    HANDLE hErrorWrite;                         //!< Handle of stderr

    _PROCESS_INFORMATION PI;                    //!< Process information structure
#else
    FILE* os_hnd;                               //!< Handle of pipe
#endif
    char* out_buffer;                           //!< Output buffer from console
    int out_buffer_size;                        //!< Current size of output buffer
    int exit_code;                              //!< Exit code of process
    bool error;                                 //!< true for error in execution

    c_exec_handle()
    {
        memset(this,0,sizeof(*this));
    }
};

c_exec_handle* exec_async_exec(const string& cmd, const string& params, const string& dir);
c_exec_handle* exec_async_exec2(const string& cmd, const string& params, const string& dir);

bool exec_process_active(c_exec_handle* h);

int exec_read_output(c_exec_handle* h, char*& buffer, int& buffer_size);

int exec_process_end(c_exec_handle* h);

void exec_terminate_process(c_exec_handle* h, int error_code);

#if !defined(_WIN32) && !defined(_WIN64)
#   define DWORD int
#endif

DWORD exec_get_exit_code(c_exec_handle* h);
void exec_detach_process(c_exec_handle* h);

#endif
