/***************************************************************************
                          filesys.cpp  -  description
                             -------------------
    begin                : Thu Dec 9 2004
    copyright            : (C) 2004-2007 by Marian VooDooMan Meravy
    email                : ghostvoodooman -a.T= users (dot) sourceforge.net
 ***************************************************************************/

/* $Id: filesys.cpp,v 1.58 2008/06/29 17:37:49 ghostvoodooman Exp $ */

/***************************************************************************

VooDoo cIRCle - an IRC (ro)bot
Copyright (C) 2004-2008 by Marian VooDooMan Meravy
  (ghostvoodooman -a.T= users (dot) sourceforge.net)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************/

/*!
    \file
    \brief Handles filesystem
    \version $Id: filesys.cpp,v 1.58 2008/06/29 17:37:49 ghostvoodooman Exp $
*/

#if defined(HAVE_CONFIG_H)
#   include <config.h>
#endif

#include <string>
#include <vector>
#include <stack>
#include <list>
#include <map>

#include <stdlib.h>
#include <time.h>
#include <stdio.h>
#if defined(_WIN32) || defined(_WIN64)
#    include <io.h>
#else
#    include <unistd.h>
#endif

#include "utils/memleak.h"
#include "utils/assert.h"
#include "utils.h"
#include "filesys.h"
#include "dcc.h"
#include "conf.h"
#include "irc.h"
#include "logic.h"
#include "lang.h"
#include "compat.h"
#include "modules.h"
#include "params.h"
#include "var/c_var_string.h"
#include "misc/shutdown.h"
#include "conf/c_conf.h"

#include "utils/precomp.h"

#if defined(_MSC_VER) || defined(__BORLANDC__)
#   pragma hdrstop
#endif

using namespace std;

vector<c_file> files;                       //!< List of all files in the filesystem

bool filesys_init_done=false;               //!< Have we initialised the filesystem yet?

my_time_t filesys_last_flush=0;                //!< Timestamp of last call to filesys_flush()

/*!
    \brief Parses line from index file (key=value)
    \author VooDooMan
    \version 1
    \date 2004
    \param ln Line to parse
    \param key Returns key (key=value)
    \param val Returns value (key=value)
    \param end Returns true if comment was found
    \return Returns true if succeeded (false for empty line, etc...)
*/
bool filesys_get_key(char* ln, string& key, string& val, bool& end)
{
    key="";
    val="";
    end=false;
    if(strlen(ln)) {
        if(ln[0]=='#') {
            end=true;
            return false;
        }
        int pos=0;
        for(unsigned int i1=0; i1<strlen(ln); ++i1) {
            if(pos==0 && ln[i1]=='=') {
                pos++;
                continue;
            }
            if(pos==0)
                key+=ln[i1];
            if(pos>0)
                val+=ln[i1];
        }
        return true;
    } else
        return false;
}

/*!
    \brief Initializes the filesystem
    \author VooDooMan
    \version 1
    \date 2004
*/
void filesys_init()
{
    if(filesys_init_done)
        return;
    filesys_init_done=true;

    files.clear();
    FILE* f=fopen("." FILE_SLASH "filesystem" FILE_SLASH "index.txt","r");
    if(!f) {
        f=fopen("." FILE_SLASH "filesystem" FILE_SLASH "index.txt","w");
        if(!f) {
            return;
        }
        fclose(f);
        return;
    }
    char ln[1024+16];

    c_file file;
    c_access access;
    c_event event;

    while(!feof(f)) {
        if(fgets(ln,1024-1,f)==NULL)
            break;

        while(ln[0] && ln[strlen(ln)-1]=='\n')
            ln[strlen(ln)-1]=0;
        while(ln[0] && ln[strlen(ln)-1]=='\r')
            ln[strlen(ln)-1]=0;

        string key, val;
        bool end;
        if(!filesys_get_key(ln,key,val,end))
            continue;

        if(!key.compare("FILE_BEGIN")) {
            access.clear();
            file.clear();
            event.clear();
            continue;
        }

        if(!key.compare("FILE_END")) {
            files.push_back(file);
            access.clear();
            file.clear();
            event.clear();
            continue;
        }

        if(!key.compare("InternalName")) {
            file.internal_name=val;
            continue;
        }

        if(!key.compare("PublicName")) {
            file.public_name=val;
            continue;
        }

        if(!key.compare("Published")) {
            file.published=!val.compare("1");
            continue;
        }

        if(!key.compare("Complete")) {
            file.complete=!val.compare("1");
            continue;
        }

        if(!key.compare("Time")) {
            file.ftime=sto_time(val);
            continue;
        }

        if(!key.compare("Expiration")) {
            file.expiration=sto_time(val);
            continue;
        }

        if(!key.compare("Type")) {
            file.file_type=ft_invalid;
            if(!val.compare("file"))
                file.file_type=ft_file;
            if(!val.compare("message"))
                file.file_type=ft_message;
            continue;
        }

        if(!key.compare("ACCESS_BEGIN")) {
            access.clear();
            continue;
        }

        if(!key.compare("ACCESS_END")) {
            file.access.push_back(access);
            access.clear();
            continue;
        }

        if(!key.compare("UserName")) {
            access.user_name=val;
            continue;
        }

        if(!key.compare("AllOnChannel")) {
            access.all_on_channel=val;
            continue;
        }

        if(!key.compare("AlsoUnknown")) {
            access.also_unknown=!val.compare("1");
            continue;
        }

        if(!key.compare("AllUsers")) {
            access.all_users=!val.compare("1");
            access.user_name="";
            continue;
        }

        if(!key.compare("Owner")) {
            access.owner=!val.compare("1");
            continue;
        }

        if(!key.compare("Read")) {
            access.read=!val.compare("1");
            continue;
        }

        if(!key.compare("Delete")) {
            access.del=!val.compare("1");
            continue;
        }

        if(!key.compare("NotifyOwner")) {
            access.notify_owner=!val.compare("1");
            continue;
        }

        if(!key.compare("NotifyUser")) {
            access.notify_user=!val.compare("1");
            continue;
        }

        if(!key.compare("Secure")) {
            access.secure=!val.compare("1");
            continue;
        }

        if(!key.compare("EVENT_BEGIN")) {
            event.clear();
            continue;
        }

        if(!key.compare("EVENT_END")) {
            file.events.push_back(event);
            event.clear();
            continue;
        }

        if(!key.compare("EventHasRead")) {
            event.has_read=!val.compare("1");
            continue;
        }

        if(!key.compare("EventOwnerNotified")) {
            event.owner_notified=!val.compare("1");
            continue;
        }

        if(!key.compare("EventUserName")) {
            event.user_name=val;
            continue;
        }

        if(!key.compare("EventTimestamp")) {
            event.first=sto_time(val);
            continue;
        }

        if(!key.compare("NotifyOwnerMessage")){
            access.notify_owner_message+=val;
            continue;
        }

        if(!key.compare("NotifyUserMessage")){
            access.notify_user_message+=val;
            continue;
        }

        if(!key.compare("UserNotified")){
            event.user_notified=!val.compare("1");
            continue;
        }
    }
    fclose(f);

    filesys_flush(); // for killing of expired files
}

/*!
    \brief Writes new index file from memory
    \author VooDooMan
    \version 1
    \date 2004
*/
void filesys_flush()
{
    if(!filesys_init_done)
        return;

    my_time(&filesys_last_flush);

    e_shutdown_reason sr=bot_get_shutdown_state();
    if(E_SD_UPGRADE==sr)
        return;

    FILE* f=fopen("." FILE_SLASH "filesystem" FILE_SLASH "index.txt","w");
    if(!f)
        return;

    vector<c_file>::iterator i1;
again:
    for(i1=files.begin(); i1!=files.end(); ++i1) {
        if((*i1).expiration!=0 && (*i1).expiration<=my_time(NULL)) {
            string fn="." FILE_SLASH "filesystem" FILE_SLASH;
            fn+=(*i1).internal_name;
            unlink(fn.c_str());
            (*i1).expired=true;
            files.erase(i1);
            goto again;
        }
        fprintf(f,"#\nFILE_BEGIN=\nInternalName=%s\nPublicName=%s\nTime=%s\nType=",(*i1).internal_name.c_str(),(*i1).public_name.c_str(),tos((*i1).ftime).c_str());
        if((*i1).file_type==ft_invalid)
            fprintf(f,"invalid");
        if((*i1).file_type==ft_file)
            fprintf(f,"file");
        if((*i1).file_type==ft_message)
            fprintf(f,"message");
        fprintf(f,"\nPublished=%d\nComplete=%d\nExpiration=%s\n",BOOL2INT((*i1).published),BOOL2INT((*i1).complete),tos((*i1).expiration).c_str());
        fprintf(f,"#\n");
        vector<c_access>::iterator i2;
        for(i2=(*i1).access.begin(); i2!=(*i1).access.end(); ++i2) {
            fprintf(f,"ACCESS_BEGIN=\n");
            fprintf(f,"AllUsers=%d\nUserName=%s\nOwner=%d\nRead=%d\nDelete=%d\nNotifyOwner=%d\nNotifyUser=%d\nSecure=%d\n",BOOL2INT((*i2).all_users),(*i2).user_name.c_str(),BOOL2INT((*i2).owner),BOOL2INT((*i2).read),BOOL2INT((*i2).del),BOOL2INT((*i2).notify_owner),BOOL2INT((*i2).notify_user),BOOL2INT((*i2).secure));
            vector<string>::iterator i3;
            fprintf(f,"AllOnChannel=%s\n",(*i2).all_on_channel.c_str());
            fprintf(f,"AlsoUnknown=%d\n",BOOL2INT((*i2).also_unknown));
            string msg_orig=(*i2).notify_owner_message;
            unsigned int i1;
            string msg;
            for(i1=0; i1<msg_orig.length(); ++i1) {
                if(msg_orig[i1]=='\n') {
                    msg+="\\n";
                    continue;
                }
                msg+=msg_orig[i1];
            }
            string partial;
            for(i1=0; i1<msg.length(); ++i1) {
                partial+=msg[i1];
                if((i1!=0 && i1%128==0) || i1==msg.length()-1) {
                    fprintf(f,"NotifyOwnerMessage=%s\n",partial.c_str());
                    partial="";
                }
            }
            msg_orig=(*i2).notify_user_message;
            msg="";
            for(i1=0; i1<msg_orig.length(); ++i1) {
                if(msg_orig[i1]=='\n') {
                    msg+="\\n";
                    continue;
                }
                msg+=msg_orig[i1];
            }
            partial="";
            for(i1=0; i1<msg.length(); ++i1) {
                partial+=msg[i1];
                if((i1!=0 && i1%128==0) || i1==msg.length()-1) {
                    fprintf(f,"NotifyUserMessage=%s\n",partial.c_str());
                    partial="";
                }
            }
            fprintf(f,"ACCESS_END=\n#\n");
        }

        // pack events
pack_continue:
        vector<c_event>::iterator i3;
        for(i3=(*i1).events.begin(); i3!=(*i1).events.end(); ++i3) {
            vector<c_event>::iterator i4;
            for(i4=(*i1).events.begin(); i4!=(*i1).events.end(); ++i4) {
                if(i3==i4)
                    continue;
                if(!(*i4).user_name.compare((*i3).user_name)) {
                    (*i3).has_read=(*i4).has_read || (*i3).has_read;
                    (*i3).owner_notified=(*i4).owner_notified || (*i3).owner_notified;
                    (*i3).user_notified=(*i4).user_notified || (*i3).user_notified;

                    if((*i4).first>(*i3).first)
                        (*i3).first=(*i4).first;

                    (*i1).events.erase(i4);
                    goto pack_continue;
                }
            }
        }

        for(i3=(*i1).events.begin(); i3!=(*i1).events.end(); ++i3) {
            fprintf(f,"EVENT_BEGIN=\n");
            fprintf(f,"EventHasRead=%d\nEventOwnerNotified=%d\nEventUserName=%s\nUserNotified=%d\nEventTimestamp=%s\n",BOOL2INT((*i3).has_read),BOOL2INT((*i3).owner_notified),(*i3).user_name.c_str(),BOOL2INT((*i3).user_notified),tos((*i3).first).c_str());
            fprintf(f,"EVENT_END=\n#\n");
        }

        fprintf(f,"FILE_END=\n###\n");
    }

    fclose(f);
}

/*!
    \brief Adds a new file to filesystem, if the same public name exists, and owner argument is its owner, it will be replaced
    \author VooDooMan
    \version 1
    \date 2004
    \param internal_name Internal name of the file
    \param original_name Original name of the file
    \param owner Owner of the file (from "logic.txt")
    \param complete Set to true if file is complete. If false, it will be tried to resume (DCC RESUME extension to standard DCC protocol)
    \param after_resume Set this to true if file was updated deu to resume in DCC SEND (DCC RESUME extension)
*/
void filesys_add_file_raw(string internal_name, string original_name, string owner, bool complete, bool after_resume)
{
    filesys_init();

    string sender_file_name=original_name;

    for(unsigned int i2=0; i2<original_name.length(); ++i2) {
        bool ok=false;
        if(original_name[i2]>='a' && original_name[i2]<='z')
            ok=true;
        if(original_name[i2]>='A' && original_name[i2]<='Z')
            ok=true;
        if(original_name[i2]>='0' && original_name[i2]<='9')
            ok=true;
        if(original_name[i2]=='_' || original_name[i2]=='-' || original_name[i2]=='.')
            ok=true;
        if(!ok)
            original_name[i2]='_';
    }

    c_file f;
    bool got_file=false;

again:
    if(!after_resume) {
        vector<c_file>::iterator i1;
        for(i1=files.begin(); i1!=files.end(); ++i1) {
            if(!(*i1).public_name.compare(original_name)) {
                if((*i1).is_owner(owner)) {
                    string fn="." FILE_SLASH "filesystem" FILE_SLASH;
                    fn+=(*i1).internal_name;
                    unlink(fn.c_str());
                    files.erase(i1);
                    goto again;
                }
            }
        }
    } else {
        vector<c_file>::iterator i1;
        bool is_owner=false;
        for(i1=files.begin(); i1!=files.end(); ++i1) {
            if((*i1).public_name.compare(sender_file_name))
                continue;
            vector<c_access>::iterator i2;
            for(i2=(*i1).access.begin(); i2!=(*i1).access.end(); ++i2) {
                if((*i2).owner && !owner.compare((*i2).user_name)) {
                    is_owner=true;
                    break;
                }
            }
            got_file=true;
            break;
        }
        if(!is_owner)
            return;

        f=*i1;
    }

    if(original_name.length()>1024-128)
        return;

    f.file_type=ft_file;
    my_time(&f.ftime);
    if(!got_file) {
        f.sender_file_name=sender_file_name;
        f.complete=complete;
        f.internal_name=internal_name;
        f.public_name=original_name;
        f.published=false;

        c_access a;
        a.all_users=false;
        a.user_name=owner;
        a.owner=true;
        a.read=true;
        a.del=true;

        f.access.push_back(a);

        files.push_back(f);
    } else {
        vector<c_file>::iterator i1;
        for(i1=files.begin(); i1!=files.end(); ++i1) {
            if((*i1).public_name.compare(sender_file_name))
                continue;
            f.complete=complete;
            (*i1)=f;
            break;
        }
    }

    if(filesys_last_flush+60<my_time(NULL))
        filesys_flush();

    if(complete)
        dcc_notify(owner,FILESYS_NOT_PUBLISHED,"");
    else
        dcc_notify(owner,FILESYS_GOT_INCOMPLETE,sender_file_name);
}

/*
    \brief Builds a message for notifying user how to download file via ".getfile"
    \author VooDooMan
    \version 2
    \date 2004
    \param a Access rights with messages
    \param public_name Public name of file
    \param eol EOL sequence
    \param lang Language to use for user
    \return Returns message for user
*/
string filesys_build_getfile_message(c_access a, string public_name, string lang, string eol)
{
    string msg;
    if(!a.notify_user_message.empty())  {
        string m;
        int i2=0;
        for(unsigned int i1=0; i1<a.notify_user_message.length(); ++i1) {
            if(a.notify_user_message[i1]=='\n')
                i2=i1;
            if(i1-i2>=128 && a.notify_user_message[i1]!='\n') {
                m+=eol;
                i2=i1;
            }
            m+=a.notify_user_message[i1];
        }
        string m2;
        for(unsigned int i3=0; i3<m.length(); ++i3) {
            if(i3+1<m.length() && m[i3]=='%' && m[i3+1]=='%') {
                i3++;
                m2+=public_name;
                continue;
            }
            if((i3+1<m.length() && m[i3]=='%' && m[i3+1]!='%') || (i3+1==m.length() && m[i3]=='%')) {
                m2+=" .getfile "; // SPACE as first character! To avoid bot loops!
                m2+=public_name;
                continue;
            }
            m2+=m[i3];
        }
        msg+=m2;
    }
    msg+=eol;
    msg+="---";
    msg+=eol;
    msg.append(lang_get_string(1,lang,655)); // Use command:
    msg+=eol;
    msg+=" .getfile "; // SPACE as first character! To avoid bot loops!
    msg+=public_name;
    msg+=eol;
    msg.append(lang_get_string(1,lang,656)); // To get the file.
    msg+=eol;
    msg+="---";
    msg+=eol;
    return msg;
}

/*
    \brief Returns time string
    \author VooDooMan
    \version 1
    \date 2005
    \param t Time to convert to text
    \return Returns time string of "t"
*/
string filesys_get_time_str(my_time_t t)
{
    char time_[512];
    strcpy(time_,make_nice_time(t,false).c_str());
    int ii;
    for(ii=0; time_[ii]!='\r' && time_[ii]!='\n' && time_[ii]!=0; ++ii)
        ;
    time_[ii]=0;
    return (string)time_;
}

/*!
    \brief Checks for notifications about new object in the filesystem
    \author VooDooMan
    \version 1
    \date 2004
    \param user User name from "logic.txt"
    \param notify Returns list of notifications
    \param secured Set to true if client is connected via DCC CHAT or telnet (has been authentificated)
    \param channel Name of channel that is the user on
    \param u User descriptor
    \param chs User's channels
    \param eol EOL sequence to use
    \param lang Language to use for user
    \return Returns true if any notification was found
*/
bool filesys_dcc_check_for_notifies(string user, c_user& u, vector<c_channel>& UNUSEDARG(chs), string channel, vector<c_dcc_notify>& notify, bool secured, string lang, string eol)
{
    filesys_init();

    notify.clear();

    bool host_unknown=false;
    /*c_user u;
    vector<c_channel> chs;
    logic_partyline_get_user(user,u,chs);*/
    host_unknown=u.host_unknown;

    vector<c_file>::iterator i1;
    for(i1=files.begin(); i1!=files.end(); ++i1) {
        if(!(*i1).published && (*i1).complete) {
            vector<c_access>::iterator i2;
            for(i2=(*i1).access.begin(); i2!=(*i1).access.end(); ++i2) {
                if((*i2).owner && ((*i2).all_users || !(*i2).user_name.compare(user))) {
                    c_dcc_notify n;
                    n.unpublished=true;
                    n.name=(*i1).public_name.c_str();
                    notify.push_back(n);
                    break;
                }
            }
            continue;
        }
        if(!(*i1).complete) {
            vector<c_access>::iterator i2;
            for(i2=(*i1).access.begin(); i2!=(*i1).access.end(); ++i2) {
                if((*i2).owner && ((*i2).all_users || !(*i2).user_name.compare(user))) {
                    c_dcc_notify n;
                    n.incomplete=true;
                    n.name=(*i1).public_name.c_str();
                    notify.push_back(n);
                    break;
                }
            }
            continue;
        }
        vector<c_access>::iterator i2;
        for(i2=(*i1).access.begin(); i2!=(*i1).access.end(); ++i2) {
            if(!(*i2).also_unknown && host_unknown && (*i2).user_name.compare(user))
                continue;
            string on_channel;
            if(!irc_cmp_strings_case_insensitive((*i2).all_on_channel,channel))
                on_channel=(*i2).all_on_channel;
            if((*i2).all_users || !(*i2).user_name.compare(user) || !on_channel.empty()) {
                if((*i2).notify_user) {
                    bool got=false;
                    vector<c_event>::iterator i3;
                    for(i3=(*i1).events.begin(); i3!=(*i1).events.end(); ++i3) {
                        if(!(*i3).has_read && !(*i3).user_name.compare(user) && !(*i3).user_notified) {
                            c_dcc_notify n;
                            bool secure=(*i2).secure;
                            n.notify_user=!secure;
                            n.secure_notify_user=secure;
                            n.name=(*i1).public_name.c_str();
                            n.message=(*i2).notify_user_message;
                            if((*i1).file_type==ft_message && ((secured && (*i2).secure) || !(*i2).secure) && !host_unknown) {
                                c_event e;
                                e.has_read=true;
                                e.owner_notified=false;
                                e.user_name=user;
                                e.user_notified=true;
                                e.first=my_time(NULL);
                                (*i1).events.push_back(e);

                                string msg=eol;
                                msg+="---";
                                msg+=eol;
                                msg.append(lang_get_string(1,lang,657)); // Message from
                                msg+=" ";
                                vector<c_access>::iterator i3;
                                for(i3=(*i1).access.begin(); i3!=(*i1).access.end(); ++i3)
                                    if((*i3).owner) {
                                        msg+=(*i3).user_name;
                                        break;
                                    }
                                msg+=eol;
                                lang_get_string(1,lang,658); // On
                                msg+=" ";
                                msg+=filesys_get_time_str((*i1).ftime);
                                msg+=eol;
                                msg.append(lang_get_string(1,lang,659)); // Subject:
                                msg+=" ";
                                msg+=(*i1).public_name;
                                msg+=eol;
                                string fn="." FILE_SLASH "filesystem" FILE_SLASH;
                                fn+=(*i1).internal_name;
                                FILE* f=fopen(fn.c_str(),"rb");
                                if(f) {
                                    char buff[128+1];
                                    while(!feof(f)) {
                                        size_t len=fread(buff,1,sizeof(buff)-1,f);
                                        buff[len]=0;
                                        for(unsigned int i2=0; i2<len; ++i2)
                                            if(buff[i2]=='\n')
                                                msg+=eol;
                                            else
                                                msg.push_back(buff[i2]);
                                        //msg+=buff;
                                        //msg+=eol;
                                    }
                                    fclose(f);
                                    msg+="---";
                                    msg+=eol;
                                    n.message=msg;
                                }
                            }
                            if(n.message.length()) {
                                string msg=eol;
                                msg+="---";
                                msg+=eol;
                                msg.append(lang_get_string(1,lang,660)); // File from
                                msg+=" ";
                                vector<c_access>::iterator i3;
                                for(i3=(*i1).access.begin(); i3!=(*i1).access.end(); ++i3)
                                    if((*i3).owner) {
                                        msg+=(*i3).user_name;
                                        break;
                                    }
                                msg+=eol;
                                msg.append(lang_get_string(1,lang,658)); // On
                                msg+=" ";
                                msg+=filesys_get_time_str((*i1).ftime);
                                msg+=eol;
                                msg.append(lang_get_string(1,lang,661)); // File name:
                                msg+=" ";
                                msg+=(*i1).public_name;
                                msg+=eol;
                                msg.append(lang_get_string(1,lang,662)); // Message from owner:
                                msg+=eol;
                                msg+=filesys_build_getfile_message(*i2,(*i1).public_name,lang,eol);
                                n.message=msg;
                                notify.push_back(n);
                                got=true;
                                break;
                            }
                            if(!secured && (*i2).secure) {
                                n.message="";
                                notify.push_back(n);
                                got=true;
                                break;
                            }
                        }
                        if(!(*i3).user_name.compare(user) && (*i3).user_notified) {
                            got=true;
                            break;
                        }
                    }
                    if(!got) {
                        c_dcc_notify n;
                        bool secure=(*i2).secure;
                        n.notify_user=((secured && (*i2).secure) || !(*i2).secure);
                        n.secure_notify_user=secure;
                        n.name=(*i1).public_name.c_str();
                        n.message=(*i2).notify_user_message;
                        if((*i1).file_type==ft_message && ((secured && (*i2).secure) || !(*i2).secure)) {
                            if(!host_unknown) {
                                c_event e;
                                e.has_read=true;
                                e.owner_notified=false;
                                e.user_name=user;
                                e.user_notified=true;
                                e.first=my_time(NULL);
                                (*i1).events.push_back(e);
                            }

                            string msg=eol;
                            msg+="---";
                            msg+=eol;
                            msg.append(lang_get_string(1,lang,657)); // Message from
                            msg+=" ";
                            vector<c_access>::iterator i3;
                            for(i3=(*i1).access.begin(); i3!=(*i1).access.end(); ++i3)
                                if((*i3).owner) {
                                    msg+=(*i3).user_name;
                                    break;
                                }
                            msg+=eol;
                            msg.append(lang_get_string(1,lang,658)); // On
                            msg+=" ";
                            msg+=filesys_get_time_str((*i1).ftime);
                            msg+=eol;
                            msg.append(lang_get_string(1,lang,659)); // Subject:
                            msg+=" ";
                            msg+=(*i1).public_name;
                            msg+=eol;
                            string fn="." FILE_SLASH "filesystem" FILE_SLASH;
                            fn+=(*i1).internal_name;
                            FILE* f=fopen(fn.c_str(),"rb");
                            if(f) {
                                char buff[128+1];
                                while(!feof(f)) {
                                    size_t len=fread(buff,1,sizeof(buff)-1,f);
                                    buff[len]=0;
                                    for(unsigned int i2=0; i2<len; ++i2)
                                        if(buff[i2]=='\n')
                                            msg+=eol;
                                        else
                                            msg.push_back(buff[i2]);
                                    //msg+=buff;
                                    //msg+=eol;
                                }
                                fclose(f);
                                msg+="---";
                                msg+=eol;
                                n.message=msg;
                            }
                        }
                        if((*i1).file_type==ft_file && ((secured && (*i2).secure) || !(*i2).secure)) {
                            if(!host_unknown) {
                                c_event e;
                                e.has_read=false;
                                e.owner_notified=false;
                                e.user_name=user;
                                e.user_notified=true;
                                e.first=my_time(NULL);
                                (*i1).events.push_back(e);
                            }

                            string msg=eol;
                            msg+="---";
                            msg+=eol;
                            msg.append(lang_get_string(1,lang,660)); // File from:
                            msg+=" ";
                            vector<c_access>::iterator i3;
                            for(i3=(*i1).access.begin(); i3!=(*i1).access.end(); ++i3)
                                if((*i3).owner) {
                                    msg+=(*i3).user_name;
                                    break;
                                }
                            msg+=eol;
                            msg.append(lang_get_string(1,lang,658)); // On
                            msg+=" ";
                            msg+=filesys_get_time_str((*i1).ftime);
                            msg+=eol;
                            msg+="File name: ";
                            msg+=(*i1).public_name;
                            msg+=eol;
                            msg.append(lang_get_string(1,lang,662)); // Message from owner:
                            msg+=eol;
                            msg+=filesys_build_getfile_message(*i2,(*i1).public_name,lang,eol);
                            n.message=msg;
                        }
                        if(n.message.length()) {
                            notify.push_back(n);
                        }
                        if(!secured && (*i2).secure) {
                            n.message="";
                            notify.push_back(n);
                            break;
                        }
                    }
                }
            }
            if((*i2).notify_owner && (*i1).is_owner(user)) {
                vector<c_event>::iterator i3;
                for(i3=(*i1).events.begin(); i3!=(*i1).events.end(); ++i3) {
                    if(!(*i3).owner_notified) {
                        vector<c_event>::iterator i4;
                        for(i4=(*i1).events.begin(); i4!=(*i1).events.end(); ++i4) {
                            if((*i4).has_read && !(*i4).owner_notified) {
                                (*i4).owner_notified=true;

                                c_dcc_notify n;
                                bool secure=(*i2).secure;
                                n.notify_owner=!secure;
                                n.secure_notify_owner=secure;
                                n.name=(*i1).public_name;
                                n.message=(*i2).notify_owner_message;
                                if((*i1).file_type==ft_message && ((secured && (*i2).secure) || !(*i2).secure)) {
                                    string msg=lang_get_string(1,lang,663); // User
                                    msg+=" ";
                                    char tmp[64];
                                    strncpy(tmp,(*i4).user_name.c_str(),sizeof(tmp));
                                    tmp[sizeof(tmp)-1]=0;
                                    msg+=tmp;
                                    msg+=" ";
                                    msg.append(lang_get_string(1,lang,664)); // has read your message with subject
                                    msg+=" ";
                                    strncpy(tmp,(*i1).public_name.c_str(),sizeof(tmp));
                                    tmp[sizeof(tmp)-1]=0;
                                    msg+=tmp;
                                    msg+=(string)"; ";
                                    msg.append(lang_get_string(1,lang,658)); // On
                                    msg+=" ";
                                    msg+=filesys_get_time_str((*i4).first);
                                    msg+=eol;
                                    n.message=msg;
                                }
                                if((*i1).file_type==ft_file && ((secured && (*i2).secure) || !(*i2).secure)) {
                                    string msg=n.message;
                                    if(!msg.empty())  {
                                        string m;
                                        {
                                            int i2=0;
                                            for(unsigned int i1=0; i1<msg.length(); ++i1) {
                                                if(msg[i1]=='\n')
                                                    i2=i1;
                                                if(i1-i2>=128 && msg[i1]!='\n') {
                                                    m+=eol;
                                                    i2=i1;
                                                }
                                                m+=msg[i1];
                                            }
                                        }
                                        string m2;
                                        for(unsigned int i3=0; i3<m.length(); ++i3) {
                                            if(i3+1<m.length() && m[i3]=='%' && m[i3+1]=='%') {
                                                i3++;
                                                m2+=(*i4).user_name;
                                                continue;
                                            }
                                            if((i3+1<m.length() && m[i3]=='%' && m[i3+1]!='%') || (i3+1==m.length() && m[i3]=='%')) {
                                                m2+=(*i1).public_name;
                                                continue;
                                            }
                                            m2+=m[i3];
                                        }
                                        msg=m2;
                                    }
                                    if(!msg.empty() && msg[msg.length()-1]!=eol[eol.length()-1])
                                        msg+=eol;
                                    msg.append(lang_get_string(1,lang,658)); // On
                                    msg+=" ";
                                    msg+=filesys_get_time_str((*i4).first);
                                    msg+=eol;
                                    n.message=msg;
                                }
                                if(!secured && (*i2).secure) {
                                    n.message="";
                                    notify.push_back(n);
                                    break;
                                }
                                notify.push_back(n);
                            }
                        }
                    }
                }
            }
        }
    }

    if(filesys_last_flush+60<my_time(NULL))
        filesys_flush();

    return !notify.empty();
}


/*!
    \brief Gets list of files for user from the filesystem
    \author VooDooMan
    \version 1
    \date 2004
    \param user User name from "logic.txt"
    \param files_ Returns list of files
    \param must_be_owner Set to true if we wish to get files that the user is owner of
    \return Returns true if any file was found
*/
bool filesys_dcc_filelist(string user, vector<c_file>& files_, bool must_be_owner)
{
    filesys_init();

    files_.clear();

    vector<c_file>::iterator i1;
    for(i1=files.begin(); i1!=files.end(); ++i1) {
        vector<c_access>::iterator i2;
        for(i2=(*i1).access.begin(); i2!=(*i1).access.end(); ++i2) {
            if(must_be_owner) {
                if((*i2).owner && !(*i2).user_name.compare(user)) {
                    files_.push_back(*i1);
                    break;
                }
            } else {
                if((*i2).read && ((*i2).all_users || !(*i2).user_name.compare(user) || (*i2).also_unknown)) {
                    files_.push_back(*i1);
                    break;
                }
            }
        }
    }

    return !files_.empty();
}


/*!
    \brief Gets file for user from the filesystem
    \author VooDooMan
    \version 1
    \date 2004
    \param user User name from "logic.txt"
    \param public_name Public name of the file
    \param file Returns file descriptor
    \return Returns true if file was found and user has access to it
*/
bool filesys_dcc_get_file(string user, string public_name, c_file& file)
{
    filesys_init();

    file.clear();

    bool got=false;

    vector<c_file>::iterator i1;
    for(i1=files.begin(); i1!=files.end(); ++i1) {
        if((*i1).public_name.compare(public_name))
            continue;
        if((*i1).expired)
            continue;
        vector<c_access>::iterator i2;
        for(i2=(*i1).access.begin(); i2!=(*i1).access.end(); ++i2) {
            if((*i2).owner && ((*i2).all_users || !(*i2).user_name.compare(user))) {
                file=*i1;
                got=true;
                break;
            }
        }
        if(got)
            break;
    }

    return got;
}

/*!
    \brief Sets new file attributes
    \author VooDooMan
    \version 1
    \date 2004
    \param user User name from "logic.txt"
    \param file File descriptor
    \param forced_secure Returns true if file were forced as SECUREd, due to excess in long notification messages
    \return Returns true if file was found and user has access to it
*/
bool filesys_dcc_set_file_attrs(string user, c_file& file, bool& forced_secure)
{
    filesys_init();

    bool ok=false;

    vector<c_file>::iterator i1;
    for(i1=files.begin(); i1!=files.end(); ++i1) {
        if((*i1).public_name.compare(file.public_name))
            continue;
        (*i1).expiration=file.expiration;
        vector<c_access>::iterator i2;
        for(i2=(*i1).access.begin(); i2!=(*i1).access.end(); ++i2) {
            if((*i2).owner && ((*i2).all_users || !(*i2).user_name.compare(user))) {
                ok=true;
                break;
            }
        }
        if(ok) {
            files.erase(i1);

            forced_secure=false;

            vector<c_access>::iterator iii;
            for(iii=file.access.begin(); iii!=file.access.end(); ++iii)
                for(int i1=0; i1<2; ++i1) {
                    if(forced_secure)
                        break;
                    string content;
                    if(i1==0)
                        content=(*iii).notify_owner_message;
                    if(i1==1)
                        content=(*iii).notify_user_message;
                    unsigned int ii, i3;
                    MY_LONG_LONG i2;

                    i2=1;
                    i3=0;
                    for(ii=0; ii<content.length(); ++ii) {
                        if(content[ii]=='\n') {
                            i2++;
                            i3=0;
                        }
                        i3++;
                        if(i3>=128) {
                            i2++;
                            i3=0;
                        }
                    }
                    if(i2>core_conf_getvar_int("msg_force_secure_lines")) {
                        vector<c_access>::iterator i1;
                        for(i1=file.access.begin(); i1!=file.access.end(); ++i1)
                            (*i1).secure=true;
                        forced_secure=true;
                    }
                }

            files.push_back(file);
            break;
        }
    }

    if(filesys_last_flush+60<my_time(NULL))
        filesys_flush();

    return ok;
}

/*!
    \brief Drops/clears notifications about new object in the filesystem
    \author VooDooMan
    \version 1
    \date 2004
    \param user User name from "logic.txt"
    \param only_non_secure Set to true if client is NOT connected via DCC CHAT or telnet (has NOT been authentificated)
*/
void filesys_dcc_drop_notifies(string user, bool only_non_secure)
{
    filesys_init();

    {
        // Check: do NOT drop notify for unknown user!!!
        c_user u;
        vector<c_channel> chs;
        logic_partyline_get_user(user,u,chs);
        if(u.host_unknown)
            return;
    }

    vector<c_file>::iterator i1;
    for(i1=files.begin(); i1!=files.end(); ++i1) {
        vector<c_access>::iterator i2;
        for(i2=(*i1).access.begin(); i2!=(*i1).access.end(); ++i2) {
            if((*i2).all_users || !(*i2).user_name.compare(user)) {
                if((*i2).notify_user && ((only_non_secure && !(*i2).secure) || !only_non_secure)) {
                    bool got=false;
                    vector<c_event>::iterator i3;
                    for(i3=(*i1).events.begin(); i3!=(*i1).events.end(); ++i3) {
                        if(!(*i3).has_read && !(*i3).user_name.compare(user)) {
                            (*i3).user_notified=true;
                            got=true;
                        }
                    }
                    if(!got) {
                        c_event e;
                        e.has_read=false;
                        e.owner_notified=false;
                        e.user_name=user;
                        e.user_notified=true;
                        e.first=my_time(NULL);
                        (*i1).events.push_back(e);
                    }
                }
                /*if(((*i2).notify_owner && ((only_non_secure && !(*i2).secure) || !only_non_secure))) {
                    vector<c_event>::iterator i3;
                    for(i3=(*i1).events.begin(); i3!=(*i1).events.end(); ++i3) {
                        if(!(*i3).owner_notified && (*i2).owner && !(*i2).user_name) {
                            (*i3).owner_notified=true;
                        }
                    }
                }*/
            }
        }
    }

    if(filesys_last_flush+60<my_time(NULL))
        filesys_flush();
}

/*!
    \brief Adds a message to the filesystem
    \author VooDooMan
    \version 1
    \date 2004
    \param from User name from "logic.txt"
    \param msg File/message descriptor
    \param content Message text
    \param force_secure Returns true if SECURE flag was explicitly set (for case that message is too long - flood protection)
    \return Returns internal name of file
*/
string filesys_dcc_add_message(string from, c_file& msg, string content, bool& force_secure)
{
    filesys_init();

    force_secure=false;

    unsigned int ii, i3;
    MY_LONG_LONG i2;

    i2=1;
    i3=0;
    for(ii=0; ii<content.length(); ++ii) {
        if(content[ii]=='\n') {
            i2++;
            i3=0;
        }
        i3++;
        if(i3>=128) {
            i2++;
            i3=0;
        }
    }
    if(i2>core_conf_getvar_int("msg_force_secure_lines")) {
        vector<c_access>::iterator i1;
        for(i1=msg.access.begin(); i1!=msg.access.end(); ++i1)
            (*i1).secure=true;
        force_secure=true;
    }

    for(ii=0; ii<msg.public_name.length(); ++ii)
        if(msg.public_name[ii]==' ')
            msg.public_name[ii]='_';

again:
    vector<c_file>::iterator i1;
    for(i1=files.begin(); i1!=files.end(); ++i1) {
        if(!(*i1).public_name.compare(msg.public_name)) {
            string n="_";
            n+=msg.public_name;
            msg.public_name=n;
            goto again;
        }
    }

    if(msg.public_name.length()>1024-128)
        return (string)"";

    msg.file_type=ft_message;
    my_time(&msg.ftime);

    char t[128];
    my_time_t now=my_time(NULL);
    strcpy(t,make_nice_time(now,false).c_str());

    string internal=t;
    internal+="_msg_";
    internal+=from;
    for(ii=0; ii<internal.length(); ++ii) {
        if(internal[ii]==' ' || internal[ii]=='\n')
            internal[ii]='_';
        bool okay=false;
        if(internal[ii]>='a' && internal[ii]<='z')
            okay=true;
        if(internal[ii]>='A' && internal[ii]<='Z')
            okay=true;
        if(internal[ii]>='0' && internal[ii]<='9')
            okay=true;
        if(!okay)
            internal[ii]='_';
    }
    internal+=".dat";

    msg.internal_name=internal;
    msg.published=true;
    msg.complete=true;

    string fn="." FILE_SLASH "filesystem" FILE_SLASH;
    fn+=internal;

    FILE* f=fopen(fn.c_str(),"wb");
    if(f) {
        fwrite(content.c_str(),1,content.length(),f);
        fclose(f);

        files.push_back(msg);
    }

    if(filesys_last_flush+60<my_time(NULL))
        filesys_flush();

    return internal;
}

/*!
    \brief Sets flag READ-"ed" for file from the filesystem
    \author VooDooMan
    \version 1
    \date 2004
    \param user Name of user from "logic.txt"
    \param public_name Public name of the file
*/
void filesys_set_file_was_read(string user, string public_name)
{
    filesys_init();

    vector<c_file>::iterator i1;
    for(i1=files.begin(); i1!=files.end(); ++i1) {
        if(!(*i1).public_name.compare(public_name)) {
            vector<c_access>::iterator i2;
            for(i2=(*i1).access.begin(); i2!=(*i1).access.end(); ++i2) {
                if(!(*i2).user_name.compare(user)) {
                    (*i2).read=true;
                }
            }
            break;
        }
    }

    if(filesys_last_flush+60<my_time(NULL))
        filesys_flush();
}

/*!
    \brief Sets new attributes for file in the filesystem
    \author VooDooMan
    \version 1
    \date 2004
    \param file File descriptor
    \return Returns true if succeeded
*/
bool filesys_dcc_set_file_attrs(c_file& file)
{
    filesys_init();

    bool ok=false;

    vector<c_file>::iterator i1;
    for(i1=files.begin(); i1!=files.end(); ++i1) {
        if((*i1).public_name.compare(file.public_name))
            continue;
        files.erase(i1);
        files.push_back(file);
        ok=true;
        break;
    }

    if(filesys_last_flush+60<my_time(NULL))
        filesys_flush();

    return ok;
}

/*!
    \brief Checks if the file in the filesystem can be resumed
    \author VooDooMan
    \version 1
    \date 2004
    \param user User from logic.txt who requests for resume
    \param file_name Public name of file
    \param got_bytes Returns (if succeeded) size of file in bytes that actually file has (if function returns false, this argument is undefined)
    \return Returns true if okay to resume
*/
// returns true for resume
bool filesys_dcc_check_for_resume(string user, string file_name, MY_LONG_LONG& got_bytes)
{
    filesys_init();

    vector<c_file>::iterator i1;
    for(i1=files.begin(); i1!=files.end(); ++i1) {
        if((*i1).file_type!=ft_file || (*i1).sender_file_name.compare(file_name) || (*i1).complete)
            continue;
        if((*i1).expired)
            continue;
        vector<c_access>::iterator i2;
        for(i2=(*i1).access.begin(); i2!=(*i1).access.end(); ++i2) {
            if((*i2).owner && !user.compare((*i2).user_name)) {
                char tmp[1024];
                strcpy(tmp,"." FILE_SLASH "filesystem" FILE_SLASH);
                strcat(tmp,(*i1).internal_name.c_str());
                FILE* f=fopen(tmp,"rb");
                if(!f)
                    return false;
                int my_errno;
                got_bytes=my_filesize(f,my_errno);
                fclose(f);
                if(-1==got_bytes || 0!=my_errno) {
                    got_bytes=0;
                    return false;
                }
                return true;
            }
        }
    }
    return false;
}

/*!
    \brief Retrieves informations about resume
    \author VooDooMan
    \version 1
    \date 2004
    \param user Name of user from logic.txt who requests resume
    \param file_name Public name of file
    \param internal_name Returns internal name of file in the filesystem
*/
// returns an empty string in internal_name on error
void filesys_dcc_get_resume_info(string user, string file_name, string& internal_name)
{
    filesys_init();

    vector<c_file>::iterator i1;
    for(i1=files.begin(); i1!=files.end(); ++i1) {
        if((*i1).file_type!=ft_file || (*i1).sender_file_name.compare(file_name) || (*i1).complete)
            continue;
        if((*i1).expired)
            continue;
        vector<c_access>::iterator i2;
        for(i2=(*i1).access.begin(); i2!=(*i1).access.end(); ++i2) {
            if((*i2).owner && !user.compare((*i2).user_name)) {
                internal_name=(*i1).internal_name;
                return;
            }
        }
    }
    internal_name=""; // an error
}

/*!
    \brief Deletes the file from the filesystem
    \author VooDooMan
    \version 1
    \date 2004
    \param user Name of user from logic.txt who wants to MY_DELETE(the fil)e
    \param file_name Public name of the file
    \retval 0 For successful delete
    \retval DCC_NO_SUCH_FILE The file is not present in the filesystem
    \retval DCC_ACCESS_DENIED Insufficient privileges for deletion
*/
int filesys_dcc_delete(string user, string file_name)
{
    filesys_init();

    bool del_it=false;
    bool got=false;

    vector<c_file>::iterator i1;
    for(i1=files.begin(); i1!=files.end(); ++i1) {
        if((*i1).file_type!=ft_file || (*i1).public_name.compare(file_name))
            continue;
        got=true;
        vector<c_access>::iterator i2;
        for(i2=(*i1).access.begin(); i2!=(*i1).access.end(); ++i2) {
            if((*i2).owner && !user.compare((*i2).user_name))
                del_it=true;
            if((*i2).del && !user.compare((*i2).user_name))
                del_it=true;
        }
        if(del_it)
            break;
    }

    if(del_it) {
        char tmp[1024];
        strcpy(tmp,"." FILE_SLASH "filesystem" FILE_SLASH);
        strcat(tmp,(*i1).internal_name.c_str());
        unlink(tmp);
        files.erase(i1);
    }

    if(filesys_last_flush+60<my_time(NULL))
        filesys_flush();

    if(!got)
        return DCC_NO_SUCH_FILE;

    return del_it?0:DCC_ACCESS_DENIED;
}

/*!
    \brief Check if the file with the same name already exists
    \author VooDooMan
    \version 1
    \date 2004
    \param public_name Original name of the file (aka public name in the filesystem terminology)
    \param user_name Owner of the file (from "logic.txt") / who is calling DCC SEND
    \retval 0 Okay, file doesn't exists
    \retval FILESYS_REPLACED File with such public name exists, argument owner matchs the OWNER, thus will be replaced, or continued due to resume
    \retval FILESYS_ALREADY_EXISTS File with such public name exists, argument owner DOESN'T matchs the OWNER, thus access is denied
*/
int filesys_check_add_file_raw(string public_name, string user_name)
{
    filesys_init();

    c_file f;
    //bool got_file=false;

    vector<c_file>::iterator i1;
    for(i1=files.begin(); i1!=files.end(); ++i1) {
        if(!(*i1).public_name.compare(public_name)) {
            if((*i1).is_owner(user_name))
                return FILESYS_REPLACED;
            else
                return FILESYS_ALREADY_EXISTS;
        }
    }

    return 0;
}

/*!
    \brief Returns array of filesystem for an external script
    \author VooDooMan
    \version 1
    \date 2004
*/
string filesys_get_script(string script_type)
{
    filesys_init();

    string res="";
    if(!script_type.compare("php")) {
        res+="    $filesys_files=array();\n";
        vector<c_file>::iterator i1;
        for(i1=files.begin(); i1!=files.end(); ++i1) {
            string x="    $x=array(); $x[\"type\"]=\"";
            switch((*i1).file_type) {
                case ft_invalid:
                    x+="ft_invalid";
                    break;
                case ft_file:
                    x+="ft_file";
                    break;
                case ft_message:
                    x+="ft_message";
                    break;
                default:
                    x+="ft_invalid";
                    break;
            }
            x+="\"; $x[\"InternalName\"]=\"";
            x+=(*i1).internal_name;
            x+="\"; $x[\"PublicName\"]=\"";
            x+=(*i1).public_name;
            x+="\"; $x[\"Time\"]=";
            x+=tos((*i1).ftime);
            x+="; $x[\"Published\"]=";
            x+=tos(BOOL2INT((*i1).published));
            x+="; $x[\"Complete\"]=";
            x+=tos(BOOL2INT((*i1).complete));
            x+="; $x[\"Expiration\"]=";
            x+=tos((*i1).expiration);

            x+="; $x[\"access\"]=array(); $x[\"events\"]=array();\n";

            vector<c_access>::iterator i2;
            for(i2=(*i1).access.begin(); i2!=(*i1).access.end(); ++i2) {
                x+="      $xa=array(); ";
                x+="$xa[\"AllUsers\"]=";
                x+=tos(BOOL2INT((*i2).all_users));
                x+="; $xa[\"UserName\"]=\"";
                x+=(*i2).user_name;
                x+="\"; $xa[\"Owner\"]=";
                x+=tos(BOOL2INT((*i2).owner));
                x+="; $xa[\"Read\"]=";
                x+=tos(BOOL2INT((*i2).read));
                x+="; $xa[\"Delete\"]=";
                x+=tos(BOOL2INT((*i2).del));
                x+="; $xa[\"NotifyOwner\"]=";
                x+=tos(BOOL2INT((*i2).notify_owner));
                x+="; $xa[\"NotifyUser\"]=";
                x+=tos(BOOL2INT((*i2).notify_user));
                x+="; $xa[\"Secure\"]=";
                x+=tos(BOOL2INT((*i2).secure));
                x+=";\n      $xa[\"AllOnChannel\"]=\"";
                x+=(*i2).all_on_channel;
                x+="\"; $xa[\"AlsoUnknown\"]=";
                x+=tos(BOOL2INT((*i2).also_unknown));
                x+="; $xa[\"NotifyOwnerMessage\"]=";
                x+=c_var_string::script_esc(script_type,(*i2).notify_owner_message);
                x+="; $xa[\"NotifyUserMessage\"]=";
                x+=c_var_string::script_esc(script_type,(*i2).notify_user_message);
                x+=";\n";
                x+="      array_push($x[\"access\"],$xa); unset($xa);\n";
            }

            vector<c_event>::iterator i3;
            for(i3=(*i1).events.begin(); i3!=(*i1).events.end(); ++i3) {
                x+="      $xe=array(); ";
                x+="$xe[\"EventHasRead\"]=";
                x+=tos(BOOL2INT((*i3).has_read));
                x+="; $xe[\"EventOwnerNotified\"]=";
                x+=tos(BOOL2INT((*i3).owner_notified));
                x+="; $xe[\"EventUserName\"]=\"";
                x+=(*i3).user_name;
                x+="\"; $xe[\"UserNotified\"]=";
                x+=tos(BOOL2INT((*i3).user_notified));
                x+="; $xe[\"EventTimestamp\"]=";
                x+=tos((*i3).first);
                x+=";\n      array_push($x[\"events\"],$xe); unset($xe);\n";
            }

            x+="    array_push($filesys_files,$x); unset($x);\n\n";
            res+=x;
        }
    }
    if(!script_type.compare("php_2")) {
        res+="    $filesys_files=array();\n";
        vector<c_file>::iterator i1;
        for(i1=files.begin(); i1!=files.end(); ++i1) {
            string x="    $x=new c_filesys_object; $x->init(\"";
            switch((*i1).file_type) {
                case ft_invalid:
                    x+="ft_invalid";
                    break;
                case ft_file:
                    x+="ft_file";
                    break;
                case ft_message:
                    x+="ft_message";
                    break;
                default:
                    x+="ft_invalid";
                    break;
            }
            x+="\",\"";
            x+=(*i1).internal_name;
            x+="\",\"";
            x+=(*i1).public_name;
            x+="\",";
            x+=tos((*i1).ftime);
            x+=",";
            x+=(*i1).published?"TRUE":"FALSE";
            x+=",";
            x+=(*i1).complete?"TRUE":"FALSE";
            x+=",";
            x+=tos((*i1).expiration);
            x+=");\n";

            vector<c_access>::iterator i2;
            for(i2=(*i1).access.begin(); i2!=(*i1).access.end(); ++i2) {
                x+="    $access=new c_filesys_access; $access->init(";

                x+=(*i2).all_users?"TRUE":"FALSE";
                x+=",\"";
                x+=(*i2).user_name;
                x+="\",";
                x+=(*i2).owner?"TRUE":"FALSE";
                x+=",";
                x+=(*i2).read?"TRUE":"FALSE";
                x+=",";
                x+=(*i2).del?"TRUE":"FALSE";
                x+=",";
                x+=(*i2).notify_owner?"TRUE":"FALSE";
                x+=",";
                x+=(*i2).notify_user?"TRUE":"FALSE";
                x+=",";
                x+=(*i2).secure?"TRUE":"FALSE";
                x+=",\"";
                x+=(*i2).all_on_channel;
                x+="\",";
                x+=(*i2).also_unknown?"TRUE":"FALSE";
                x+=",";
                x+=c_var_string::script_esc(script_type,(*i2).notify_owner_message);
                x+=",";
                x+=c_var_string::script_esc(script_type,(*i2).notify_user_message);
                x+=");\n";
                x+="    $x->add_access($access); unset($access);\n";
            }

            vector<c_event>::iterator i3;
            for(i3=(*i1).events.begin(); i3!=(*i1).events.end(); ++i3) {
                x+="    $event=new c_filesys_event; $event->init(";
                x+=(*i3).has_read?"TRUE":"FALSE";
                x+=",";
                x+=(*i3).owner_notified?"TRUE":"FALSE";
                x+=",\"";
                x+=(*i3).user_name;
                x+="\",";
                x+=(*i3).user_notified?"TRUE":"FALSE";
                x+=",";
                x+=tos((*i3).first);
                x+=");\n    $x->add_event($event); unset($event);\n";
            }

            x+="    $filesys_files[\"";
            x+=(*i1).internal_name;
            x+="\"]=$x; unset($x);\n\n";
            res+=x;
        }
    }
    if(!script_type.compare("perl")) {
        res+="    my %filesys_files;\n    my $access;\n    my $event;\n";
        vector<c_file>::iterator i1;
        for(i1=files.begin(); i1!=files.end(); ++i1) {
            string x="    $x=c_filesys_object->init(\"";
            switch((*i1).file_type) {
                case ft_invalid:
                    x+="ft_invalid";
                    break;
                case ft_file:
                    x+="ft_file";
                    break;
                case ft_message:
                    x+="ft_message";
                    break;
                default:
                    x+="ft_invalid";
                    break;
            }
            x+="\",\"";
            x+=(*i1).internal_name;
            x+="\",\"";
            x+=(*i1).public_name;
            x+="\",";
            x+=tos((*i1).ftime);
            x+=",";
            x+=(*i1).published?"1":"0";
            x+=",";
            x+=(*i1).complete?"1":"0";
            x+=",";
            x+=tos((*i1).expiration);
            x+=");\n";

            vector<c_access>::iterator i2;
            for(i2=(*i1).access.begin(); i2!=(*i1).access.end(); ++i2) {
                x+="    $access=c_filesys_access->init(";

                x+=(*i2).all_users?"1":"0";
                x+=",\"";
                x+=(*i2).user_name;
                x+="\",";
                x+=(*i2).owner?"1":"0";
                x+=",";
                x+=(*i2).read?"1":"0";
                x+=",";
                x+=(*i2).del?"1":"0";
                x+=",";
                x+=(*i2).notify_owner?"1":"0";
                x+=",";
                x+=(*i2).notify_user?"1":"0";
                x+=",";
                x+=(*i2).secure?"1":"0";
                x+=",\"";
                x+=(*i2).all_on_channel;
                x+="\",";
                x+=(*i2).also_unknown?"1":"0";
                x+=",";
                x+=c_var_string::script_esc(script_type,(*i2).notify_owner_message);
                x+=",";
                x+=c_var_string::script_esc(script_type,(*i2).notify_user_message);
                x+=");\n";
                x+="    $x->add_access($access); undef($access);\n";
            }

            vector<c_event>::iterator i3;
            for(i3=(*i1).events.begin(); i3!=(*i1).events.end(); ++i3) {
                x+="    $event=c_filesys_event->init(";
                x+=(*i3).has_read?"1":"0";
                x+=",";
                x+=(*i3).owner_notified?"1":"0";
                x+=",\"";
                x+=(*i3).user_name;
                x+="\",";
                x+=(*i3).user_notified?"1":"0";
                x+=",";
                x+=tos((*i3).first);
                x+=");\n    $x->add_event($event); undef($event);\n";
            }

            x+="    $filesys_files{\"";
            x+=(*i1).internal_name;
            x+="\"}=$x; undef($x);\n\n";
            res+=x;
        }
    }
    return res;
}

/*!
    \brief Gets file from the filesystem
    \author VooDooMan
    \version 1
    \date 2005
    \param internal_name Internal name of the file
    \param file Returns file descriptor
    \return Returns true if file was found
*/
bool filesys_logic_get_file(string internal_name, c_file& file)
{
    filesys_init();

    file.clear();

    bool got=false;

    vector<c_file>::iterator i1;
    for(i1=files.begin(); i1!=files.end(); ++i1) {
        if((*i1).internal_name.compare(internal_name))
            continue;
        if((*i1).expired)
            continue;
        file=*i1;
        got=true;
        break;
    }

    return got;
}

/*!
    \brief Sets file attributes for the file in filesystem
    \author VooDooMan
    \version 1
    \date 2005
    \param internal_name Internal name of the file
    \param file File descriptor to set
    \return Returns true if file was found
*/
bool filesys_logic_set_file(string internal_name, c_file& file)
{
    filesys_init();

    bool got=false;

    vector<c_file>::iterator i1;
    for(i1=files.begin(); i1!=files.end(); ++i1) {
        if((*i1).internal_name.compare(internal_name))
            continue;
        if((*i1).expired)
            continue;
        *i1=file;
        got=true;
        break;
    }

    return got;
}

