/***************************************************************************
                          c_block.h  -  description
                             -------------------
    begin                : Thu May 10 2007
    copyright            : (C) 2004-2007 by Marian VooDooMan Meravy
    email                : ghostvoodooman -a.T= users (dot) sourceforge.net
 ***************************************************************************/

/* $Id: c_block.h,v 1.15 2008/11/24 12:09:23 ghostvoodooman Exp $ */

/***************************************************************************

VooDoo cIRCle - an IRC (ro)bot
Copyright (C) 2004-2008 by Marian VooDooMan Meravy
  (ghostvoodooman -a.T= users (dot) sourceforge.net)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************/

/*!
    \file
    \brief Class 'c_block'
    \version $Id: c_block.h,v 1.15 2008/11/24 12:09:23 ghostvoodooman Exp $
*/

#ifndef C_BLOCK_H__
#define C_BLOCK_H__

#if defined(HAVE_CONFIG_H)
#   include <config.h>
#endif

#include <string>
#include <vector>
#include <stack>
#include <list>
#include <map>
#include <algorithm>

#include "../utils/memleak.h"

#include "../params.h"

#include "c_if_branch.h"

class c_expression;

/*!
    \brief Block of commands between '{' and '}' used in "stack" STL class
    \version 1
    \date 2006-01
    \author VooDooMan
*/
class c_block
{
public:
    enum e_type
    {
        _BLOCK_INVALID = 0, //!< Invalid block
        BLOCK_LEFT_BRACE,   //!< Left brace
        BLOCK_FOR,          //!< "for" command
        BLOCK_IF,           //!< "if" command
        BLOCK_SWITCH,       //!< "switch" command
        BLOCK_DO,           //!< "do" command
        BLOCK_WHILE,        //!< "while" command
        BLOCK_FOREACH,      //!< "foreach" command
    };                      //!< Type of block

    string _debug;          //!< Debug string

    // in case of BLOCK_FOR
    c_expression* for_exp_iter; //!< Iterative expression, if "_type" == BLOCK_FOR
    long for_iter_jump;         //!< ID to jump to next iteration, if "_type" == BLOCK_FOR
    // in case of BLOCK_IF
    long if_id;                 //!< ID of "if" command, if "_type" == BLOCK_IF
    bool ignore_next;           //!< If there were "else", if "_type" == BLOCK_IF
    vector<c_if_branch> if_bunch;    //!< Bunch of ID's, if "_type" == BLOCK_IF
    long if_bunch_jump_over;    //!< ID to jump over whole "if" (if condition were evaluated as true), if "_type" == BLOCK_IF
    // in case of BLOCK_SWITCH
    long switch_id;             //!< ID of "switch" command, if "_type" == BLOCK_SWITCH
    // in case of BLOCK_DO
    long do_id;                 //!< ID of "do" command, if "_type" == BLOCK_DO
    // in case of BLOCK_WHILE
    long while_again_id;        //!< ID command of next iteration, if "_type" == BLOCK_WHILE
    // in case of BLOCK_FOREACH
    long foreach_iter_jump;     //!< ID command of next iteration, if "_type" == BLOCK_FOREACH
    long foreach_init_id;       //!< ID command of initialization "foreach" loop, if "_type" == BLOCK_FOREACH

    // for inline
    bool was_inline;            //!< If last command was inlined (e.g. "if" followed by command, not left brace)
    bool was_next_expression;   //!< If last command was inlined and next expression was there (e.g. "if" followed by command, not left brace)

    e_type _type;               //!< Type of block

    long _id;                   //!< ID of block

    bool next_resolve_if;       //!< "if" should be resolved next time

    /*!
        \brief Constructor
        \version 1
        \date 2006-01
        \author VooDooMan
        \param debug Debug string
        \param id Unique ID of command
        \param type Type of block
    */
    c_block(e_type type, long id, const string& debug);
};

#endif

