/***************************************************************************
                          c_command.h  -  description
                             -------------------
    begin                : Thu May 10 2007
    copyright            : (C) 2004-2007 by Marian VooDooMan Meravy
    email                : ghostvoodooman -a.T= users (dot) sourceforge.net
 ***************************************************************************/

/* $Id: c_command.h,v 1.14 2008/03/11 21:11:34 ghostvoodooman Exp $ */

/***************************************************************************

VooDoo cIRCle - an IRC (ro)bot
Copyright (C) 2004-2008 by Marian VooDooMan Meravy
  (ghostvoodooman -a.T= users (dot) sourceforge.net)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************/

/*!
    \file
    \brief Class 'c_command'
    \version $Id: c_command.h,v 1.14 2008/03/11 21:11:34 ghostvoodooman Exp $
*/

#ifndef C_COMMAND_H__
#define C_COMMAND_H__

#if defined(HAVE_CONFIG_H)
#   include <config.h>
#endif

#include <string>
#include <vector>
#include <stack>
#include <list>
#include <map>
#include <algorithm>

#include "../utils/memleak.h"

#include "../params.h"

#include "c_expression.h"

/*!
    \brief Stores one expression in function
    \author VooDooMan
    \version 1
    \date 2005
*/
class c_command
{
public:
    enum e_type
    {
        CMD_LEFT_BRACKET,           //!< Left bracket
        CMD_RIGHT_BRACKET,          //!< Right bracket
        CMD_EXP,                    //!< Expression
        CMD_IF_EXP,                 //!< if(expression) construct
        CMD_ELSE,                   //!< else construct
        CMD_EXP_SUB,                //!< Sub expression for if_error() and wait_if_error()
        CMD_LABEL,                  //!< goto label
        CMD_GOTO,                   //!< goto command
        CMD_SWITCH,                 //!< switch construct
        CMD_CASE,                   //!< case command
        CMD_BREAK,                  //!< break command
        CMD_CONTINUE,               //!< continue command
        CMD_FOR,                    //!< for() construct
        CMD_WHILE,                  //!< while() construct
        CMD_DO,                     //!< do construct
        CMD_FOREACH_1,              //!< foreach($x) construct
        CMD_FOREACH_2,              //!< foreach($x as $y) construct
        CMD_FOREACH_3,              //!< foreach($x as $key => $val) construct
        CMD_GLOBAL,                 //!< global command
    } type;                         //!< Type of expression

    string exp_;                    //!< Expression/argument as string

    // type == CMD_EXP
    c_expression* exp;              //!< if type == CMD_EXP: expression

    // type == CMD_EXP_SUB
    c_expression* sub;              //!< If type == CMD_EXP_SUB: sub-expression
    string sub_;                    //!< If type == CMD_EXP_SUB: Subexpression as string

    // type == CMD_FOR
    c_expression* condition;        //!< If type == CMD_FOR: conditional expression
    c_expression* iterative_exp;    //!< If type == CMD_FOR: iterative expression

    // type == CMD_FOREACH_1, CMD_FOREACH_2, CMD_FOREACH_3
    c_expression* what_exp;         //!< If type == CMD_FOREACH_1, CMD_FOREACH_2, CMD_FOREACH_3: $x array expression

    // type == CMD_FOREACH_2
    string as;                      //!< If type == CMD_FOREACH_2: $y value variable

    // type == CMD_FOREACH_3
    string key;                     //!< If type == CMD_FOREACH_3: $key variable
    string value;                   //!< If type == CMD_FOREACH_3: $val variable
    bool ref_value;                 //!< If type == CMD_FOREACH_2, CMD_FOREACH_3: true if &$y/&$val is reference

    string debug;                   //!< Debug string

    bool semicolon;                 //!< true if there was semicolon at the end of expression

    c_command();
};

#endif

