/***************************************************************************
                          c_expression.h  -  description
                             -------------------
    begin                : Thu May 10 2007
    copyright            : (C) 2004-2007 by Marian VooDooMan Meravy
    email                : ghostvoodooman -a.T= users (dot) sourceforge.net
 ***************************************************************************/

/* $Id: c_expression.h,v 1.15 2008/08/23 15:10:43 ghostvoodooman Exp $ */

/***************************************************************************

VooDoo cIRCle - an IRC (ro)bot
Copyright (C) 2004-2008 by Marian VooDooMan Meravy
  (ghostvoodooman -a.T= users (dot) sourceforge.net)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************/

/*!
    \file
    \brief Class 'c_expression'
    \version $Id: c_expression.h,v 1.15 2008/08/23 15:10:43 ghostvoodooman Exp $
*/

#ifndef C_EXPRESSION_H__
#define C_EXPRESSION_H__

#if defined(HAVE_CONFIG_H)
#   include <config.h>
#endif

#include <string>
#include <vector>
#include <stack>
#include <list>
#include <map>
#include <algorithm>

#include "../utils/memleak.h"

#include "../params.h"

#include "expression/c_tree_microcode.h"
#include "../microcode/c_micro_error.h"
#include "../runtime/c_eval_iteration.h"
#include "../microcode/c_micro_symbol_table.h"

class c_sub_exp_oper;
class c_sub_exp_func;
class c_sub_exp_parenthesis;

/*!
    \brief Expression
    \version 2
    \author VooDooMan
    \date 2005, 2006-03-22
*/
class c_expression
{
    bool ok;                        //!< If it was parsed okay

    friend class c_micro_symbol_variable_mapping;

    /*!
        \brief Clears and deallocates microcode vector
        \author VooDooMan
        \version 1
        \date 2006-03-22
    */
    void clear_microcode();

    /*!
        \brief Adds a new microcode instruction
        \author VooDooMan
        \version 1
        \date 2006-03-22
        \param instr Instruction
        \return Returns ID of new instruction in stream
    */
    instruct_id_t add_micro(c_micro* instr);
    friend class c_sub_exp_oper;
    friend class c_sub_exp_func;
    friend class c_sub_exp_index;
    friend class c_sub_exp_variable;
    friend class c_sub_exp_numeric_immediate;
    friend class c_sub_exp_const;
    friend class c_sub_exp_eval_str;
    friend class c_sub_exp_immediate_str;
    friend class c_sub_exp_oper_ternary;
    friend class c_sub_exp_oper_ternary_false;
    friend class c_sub_exp_parenthesis;
    friend class c_sub_exp_ternary_true_end_marker;
    friend class c_sub_exp_oper_ternary_true;

    map<instruct_id_t, c_micro*> microcode;     //!< Microcode instructions
    instruct_id_t next_id;          //!< Next instruction ID

    /*
        \brief Updates result of shortened evaluation
        \version 2
        \author VooDooMan
        \date 2006-09-29, 2007-10-11
        \param this_module This module
        \param error Returns error, if result is 'false'
        \param symbols Symbol table
        \param res Result index
        \param instruct_id Next instruction
        \param current_instruct_id Causing instruction
    bool update_conditional(c_module* this_module, c_micro_error*& error, c_micro_symbol_table* symbols, var_sym_index_t res, instruct_id_t instruct_id, instruct_id_t current_instruct_id);
    */

    /*!
        \brief Updates result of shortened evaluation
        \version 2
        \author VooDooMan
        \date 2006-09-29, 2007-10-11
        \param this_module This module
        \param error Returns error, if result is 'false'
        \param symbols Symbol table
        \param micro Conditional microcode instruction
    */
    bool update_conditional(c_module* this_module, c_micro_error*& error, c_micro_symbol_table* symbols, c_micro* micro);

public:
    string debug;                   //!< Debug string

    string next_exp;                //!< Next expression

    var_sym_index_t result;         //!< ID of resulting variable

    /*!
        \brief Constructor
        \author VooDooMan
        \version 1
        \date 2005
        \param this_module Module
        \param exp Expression
        \param debug_ Debug string
        \param symbols Symbol table
        \param next_expression Next expression
    */
    c_expression(c_module* this_module, const string& exp, const string& debug_, c_micro_symbol_variable_mapping& symbols, const string& next_expression);
    c_expression() { ok=false; next_id=0; }

    c_expression(const c_expression& orig);
    c_expression& operator=(const c_expression& orig);

    /*!
        \brief Destructor
        \author VooDooMan
        \version 1
        \date 2006-03-22
    */
    virtual ~c_expression();

    /*!
        \brief Call this before destruction
        \author VooDooMan
        \version 1
        \date 2006-03-22
    */
    void clean_up();

    /*!
        \brief Assigns expression
        \author VooDooMan
        \version 1
        \date 2005
        \param this_module Module
        \param exp Expression
        \param debug_ Debug string
        \param symbols Symbol table
        \param next_expression Next expression
        \param debug_prefix Prefix for error
    */
    void assign(c_module* this_module, const string& exp, const string& debug_, c_micro_symbol_variable_mapping& symbols, const string& next_expression, const string& debug_prefix);

    /*!
        \brief Returns true if there is error
        \author VooDooMan
        \version 1
        \date 2005
    */
    bool got_error() {
        return !ok;
    }

    /*!
        \brief Performs expression
        \author VooDooMan
        \version 2
        \date 2005, 2006-04-03
        \param this_module Module
        \param vars Variables
        \param error Returns true for error
        \param sym_return Variable ID of result value
        \param st Evaluation stack
        \param symbols Symbol table
        \retval 'true' Okay
        \retval 'false' Error
    */
    bool perform(c_module* this_module, c_var_array* vars, c_micro_error*& error, var_sym_index_t& sym_return, stack<c_eval_iteration>& st, c_micro_symbol_table* symbols);

    /*!
        \brief Adds all inputs to vector
        \author VooDooMan
        \version 1
        \date 2007-03-18
        \param inputs Vector to add to
    */
    void add_inputs(vector<var_sym_index_t>& inputs);

    /*!
        \brief Optimizes output by removing result variables taht are not referenced as input
        \date 2007-03-18
        \author VooDooMan
        \version 1
        \param inputs All input variables
    */
    void optimize_output(const vector<var_sym_index_t>& inputs);

    instruct_id_t get_next_id();
};

#endif

