/***************************************************************************
                          c_fn_cmd.h  -  description
                             -------------------
    begin                : Thu May 10 2007
    copyright            : (C) 2004-2007 by Marian VooDooMan Meravy
    email                : ghostvoodooman -a.T= users (dot) sourceforge.net
 ***************************************************************************/

/* $Id: c_fn_cmd.h,v 1.14 2008/03/11 21:11:34 ghostvoodooman Exp $ */

/***************************************************************************

VooDoo cIRCle - an IRC (ro)bot
Copyright (C) 2004-2008 by Marian VooDooMan Meravy
  (ghostvoodooman -a.T= users (dot) sourceforge.net)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************/

/*!
    \file
    \brief Class 'c_fn_cmd'
    \version $Id: c_fn_cmd.h,v 1.14 2008/03/11 21:11:34 ghostvoodooman Exp $
*/

#ifndef C_FN_CMD_H__
#define C_FN_CMD_H__

#if defined(HAVE_CONFIG_H)
#   include <config.h>
#endif

#include <string>
#include <vector>
#include <stack>
#include <list>
#include <map>
#include <algorithm>

#include "../utils/memleak.h"

#include "../params.h"

#include "../shared.h"

/*!
    \brief Base class for function's command
    \version 1
    \date 2006-01
    \author VooDooMan
*/
class c_fn_cmd
{
protected:
    long _id;               //!< Unique ID of command
    string _debug;          //!< Debug string

public:

    /*!
        \brief Constructor
        \version 1
        \date 2006-01
        \author VooDooMan
        \param debug Debug string
        \param id Unique ID of command
    */
    c_fn_cmd(const string& debug, long id)
        : _id(id), _debug(debug)
    { ; }

    /*!
        \brief Gets ID of command
        \version 1
        \date 2006-01
        \author VooDooMan
        \return Returns ID of command
    */
    long get_id() { return _id; }

    /*!
        \brief Gets debug string
        \version 1
        \date 2006-01
        \author VooDooMan
        \return Returns debug string
    */
    const string& get_debug_string() { return _debug; }

    /*!
        \brief Performs sanity check
        \version 1
        \date 2006-01
        \author VooDooMan
        \retval true Error
        \retval false OK
        \param msg Returns error message if return value is 'true'
    */
    virtual bool sanity_check(string& msg) = 0;

    /*!
        \brief Creates copy of itself
        \version 1
        \date 2006-01
        \author VooDooMan
        \return Returns new copy instance
    */
    virtual c_fn_cmd* copy() = 0;

    virtual ~c_fn_cmd() { ; }

    /*!
        \brief Returns type of command
        \version 1
        \date 2006-04-06
        \author VooDooMan
    */
    virtual e_cmd_type get_type() = 0;

    c_fn_cmd(const c_fn_cmd& orig);

    /*!
        \brief Deallocates microcode (if applicable)
        \version 1
        \date 2006-05-09
        \author VooDooMan
    */
    virtual void clean_up() = 0;

    /*!
        \brief Adds all inputs to vector
        \author VooDooMan
        \version 1
        \date 2007-03-18
        \param inputs Vector to add to
    */
    virtual void add_inputs(vector<var_sym_index_t>& inputs) = 0;

    /*!
        \brief Optimizes output by removing result variables taht are not referenced as input
        \date 2007-03-18
        \author VooDooMan
        \version 1
        \param inputs All input variables
    */
    virtual void optimize_output(const vector<var_sym_index_t>& inputs) = 0;
};

#endif

