/***************************************************************************
                          c_fn_cmd_break.h  -  description
                             -------------------
    begin                : Thu May 10 2007
    copyright            : (C) 2004-2007 by Marian VooDooMan Meravy
    email                : ghostvoodooman -a.T= users (dot) sourceforge.net
 ***************************************************************************/

/* $Id: c_fn_cmd_break.h,v 1.14 2008/03/11 21:11:35 ghostvoodooman Exp $ */

/***************************************************************************

VooDoo cIRCle - an IRC (ro)bot
Copyright (C) 2004-2008 by Marian VooDooMan Meravy
  (ghostvoodooman -a.T= users (dot) sourceforge.net)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************/

/*!
    \file
    \brief Class 'c_fn_cmd_break'
    \version $Id: c_fn_cmd_break.h,v 1.14 2008/03/11 21:11:35 ghostvoodooman Exp $
*/

#ifndef C_FN_CMD_BREAK_H__
#define C_FN_CMD_BREAK_H__

#if defined(HAVE_CONFIG_H)
#   include <config.h>
#endif

#include <string>
#include <vector>
#include <stack>
#include <list>
#include <map>
#include <algorithm>

#include "../utils/memleak.h"

#include "../params.h"

#include "c_fn_cmd.h"
#include "c_expression.h"

/*!
    \brief Implements "break" command
    \version 1
    \date 2006-01
    \author VooDooMan
*/
class c_fn_cmd_break : public c_fn_cmd
{
    long begin_iteration_command_id;        //!< ID of command that belongs to this "break"
    bool it_ic_foreach;                     //!< Are we breaking "foreach"?
    bool it_is_switch;                      //!< Are we breaking "switch"?
public:
    c_fn_cmd* copy();

    /*!
        \brief Constructor
        \version 1
        \date 2006-01
        \author VooDooMan
        \param debug Debug string
        \param id Unique ID of command
    */
    c_fn_cmd_break(const string& debug, long id);

    bool sanity_check(string& msg);

    /*!
        \brief Gets ID of commnad that belongs to this "break"
        \version 1
        \date 2006-01
        \author VooDooMan
    */
    long get_begin_iteration_command_id();

    /*!
        \brief Sets ID of commnad that belongs to this "break"
        \version 1
        \date 2006-01
        \author VooDooMan
        \param id ID
    */
    void set_begin_iteration_command_id(long id);

    /*!
        \brief Sets flag that we are breaking "foreach"
        \version 1
        \date 2006-01
        \author VooDooMan
    */
    void set_for_foreach();

    /*!
        \brief Sets flag that we are breaking "switch"
        \version 1
        \date 2006-01
        \author VooDooMan
    */
    void set_for_switch();

    /*!
        \brief Gets flag that we are breaking "foreach"
        \version 1
        \date 2006-01
        \author VooDooMan
    */
    bool is_for_foreach();

    /*!
        \brief Gets flag that we are breaking "switch"
        \version 1
        \date 2006-01
        \author VooDooMan
    */
    bool is_for_switch();

    e_cmd_type get_type();

    /*!
        \brief Destructor
        \version 1
        \date 2006-05-09
        \author VooDooMan
    */
    virtual ~c_fn_cmd_break();

    /*!
        \brief Copy constructor
        \version 1
        \date 2006-05-09
        \author VooDooMan
        \param orig Original
    */
    c_fn_cmd_break(const c_fn_cmd_break& orig);

    /*!
        \brief Assign operator
        \version 1
        \date 2006-05-09
        \author VooDooMan
        \param orig Original
    */
    c_fn_cmd_break& operator=(const c_fn_cmd_break& orig);

    void clean_up();

    void add_inputs(vector<var_sym_index_t>& inputs);
    void optimize_output(const vector<var_sym_index_t>& inputs);
};

#endif

