/***************************************************************************
                          c_fn_cmd_exp.h  -  description
                             -------------------
    begin                : Thu May 10 2007
    copyright            : (C) 2004-2007 by Marian VooDooMan Meravy
    email                : ghostvoodooman -a.T= users (dot) sourceforge.net
 ***************************************************************************/

/* $Id: c_fn_cmd_exp.h,v 1.14 2008/03/11 21:11:36 ghostvoodooman Exp $ */

/***************************************************************************

VooDoo cIRCle - an IRC (ro)bot
Copyright (C) 2004-2008 by Marian VooDooMan Meravy
  (ghostvoodooman -a.T= users (dot) sourceforge.net)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************/

/*!
    \file
    \brief Class 'c_fn_cmd_exp'
    \version $Id: c_fn_cmd_exp.h,v 1.14 2008/03/11 21:11:36 ghostvoodooman Exp $
*/

#ifndef C_FN_CMD_EXP_H__
#define C_FN_CMD_EXP_H__

#if defined(HAVE_CONFIG_H)
#   include <config.h>
#endif

#include <string>
#include <vector>
#include <stack>
#include <list>
#include <map>
#include <algorithm>

#include "../utils/memleak.h"

#include "../params.h"

#include "c_fn_cmd.h"
#include "c_expression.h"

/*!
    \brief Implements common expression
    \version 1
    \date 2006-01
    \author VooDooMan
*/
class c_fn_cmd_exp : public c_fn_cmd
{
    c_expression* _exp;         //!< Expression to execute
    string _sub_exp;            //!< Sub-expression
public:

    e_cmd_type get_type();

    c_fn_cmd* copy();

    /*!
        \brief Constructor
        \version 1
        \date 2006-01
        \author VooDooMan
        \param debug Debug string
        \param id Unique ID of command
        \param exp Expression
    */
    c_fn_cmd_exp(c_expression* exp, const string& debug, long id);

    /*!
        \brief Constructor
        \version 1
        \date 2006-01
        \author VooDooMan
        \param debug Debug string
        \param id Unique ID of command
        \param exp Expression
        \param sub_exp Sub-expression
    */
    c_fn_cmd_exp(c_expression* exp, const string& sub_exp, const string& debug, long id);

    /*!
        \brief Gets expression
        \version 1
        \date 2006-01
        \author VooDooMan
    */
    c_expression* get_exp();

    /*!
        \brief Gets sub-expression
        \version 1
        \date 2006-01
        \author VooDooMan
    */
    const string& get_sub_exp();

    bool sanity_check(string& msg);

    /*!
        \brief Destructor
        \version 1
        \date 2006-05-09
        \author VooDooMan
    */
    virtual ~c_fn_cmd_exp();

    /*!
        \brief Copy constructor
        \version 1
        \date 2006-05-09
        \author VooDooMan
        \param orig Original
    */
    c_fn_cmd_exp(const c_fn_cmd_exp& orig);

    /*!
        \brief Assign operator
        \version 1
        \date 2006-05-09
        \author VooDooMan
        \param orig Original
    */
    c_fn_cmd_exp& operator=(const c_fn_cmd_exp& orig);

    void clean_up();

    void add_inputs(vector<var_sym_index_t>& inputs);
    void optimize_output(const vector<var_sym_index_t>& inputs);
};

#endif

