/***************************************************************************
                          c_fn_cmd_foreach.cpp  -  description
                             -------------------
    begin                : Thu May 10 2007
    copyright            : (C) 2004-2007 by Marian VooDooMan Meravy
    email                : ghostvoodooman -a.T= users (dot) sourceforge.net
 ***************************************************************************/

/* $Id: c_fn_cmd_foreach.cpp,v 1.14 2008/03/11 21:11:36 ghostvoodooman Exp $ */

/***************************************************************************

VooDoo cIRCle - an IRC (ro)bot
Copyright (C) 2004-2008 by Marian VooDooMan Meravy
  (ghostvoodooman -a.T= users (dot) sourceforge.net)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************/

/*!
    \file
    \brief Class 'c_fn_cmd_foreach'
    \version $Id: c_fn_cmd_foreach.cpp,v 1.14 2008/03/11 21:11:36 ghostvoodooman Exp $
*/

#if defined(HAVE_CONFIG_H)
#   include <config.h>
#endif

#include "../params.h"
#include "../utils/memleak.h"
#include "../utils/assert.h"
#include "../compat.h"

#include "c_fn_cmd_foreach.h"
#include "../utils/precomp.h"

#if defined(_MSC_VER) || defined(__BORLANDC__)
#   pragma hdrstop
#endif

c_fn_cmd* c_fn_cmd_foreach::copy()
{
    c_fn_cmd_foreach* x=MY_NEW c_fn_cmd_foreach(MY_NEW c_expression(*_what),_key,_as,_as_ref,_debug,_id);
    x->end_jump_id=end_jump_id;
    x->first_cmd_id=first_cmd_id;
    x->next_iteration_id=next_iteration_id;
    return x;
}

c_fn_cmd_foreach::c_fn_cmd_foreach(c_expression* what, const string& key, const string& as, bool as_ref, const string& debug, long id)
    : c_fn_cmd(debug,id)
{
    _what=what;
    _key=key;
    _as=as;
    _as_ref=as_ref;
    next_iteration_id=first_cmd_id=end_jump_id=JUMP_ID_INVALID;
}

bool c_fn_cmd_foreach::sanity_check(string& msg)
{
    CHECK_ID(_id);
    CHECK_ID(end_jump_id);
    CHECK_ID(first_cmd_id);
    CHECK_ID(next_iteration_id);
    CHECK_EXP(_what);
    return false;
}

void c_fn_cmd_foreach::set_end_id(long id)
{
    end_jump_id=id;
}

long c_fn_cmd_foreach::get_end_id()
{
    return end_jump_id;
}

void c_fn_cmd_foreach::set_first_cmd_id(long id)
{
    first_cmd_id=id;
}

long c_fn_cmd_foreach::get_first_cmd_id()
{
    return first_cmd_id;
}

void c_fn_cmd_foreach::set_next_iteration_id(long id)
{
    next_iteration_id=id;
}

long c_fn_cmd_foreach::get_next_iteration_id()
{
    return next_iteration_id;
}

c_expression* c_fn_cmd_foreach::get_what()
{
    return _what;
}

const string& c_fn_cmd_foreach::get_key()
{
    return _key;
}

const string& c_fn_cmd_foreach::get_as()
{
    return _as;
}

bool c_fn_cmd_foreach::get_as_ref()
{
    return _as_ref;
}

e_cmd_type c_fn_cmd_foreach::get_type()
{
    return CMD_FOREACH;
}

c_fn_cmd_foreach::~c_fn_cmd_foreach()
{
    clean_up();
    _what->clean_up();
    MY_DELETE(_what);
}

c_fn_cmd_foreach::c_fn_cmd_foreach(const c_fn_cmd_foreach& orig)
    : c_fn_cmd(orig)
{
    _what=MY_NEW c_expression(*orig._what);
    _key=orig._key;
    _as=orig._as;
    _as_ref=orig._as_ref;
    end_jump_id=orig.end_jump_id;
    first_cmd_id=orig.first_cmd_id;
    next_iteration_id=orig.next_iteration_id;
}

c_fn_cmd_foreach& c_fn_cmd_foreach::operator=(const c_fn_cmd_foreach& orig)
{
    _id=orig._id;
    _debug=orig._debug;

    _what=MY_NEW c_expression(*orig._what);
    _key=orig._key;
    _as=orig._as;
    _as_ref=orig._as_ref;
    end_jump_id=orig.end_jump_id;
    first_cmd_id=orig.first_cmd_id;
    next_iteration_id=orig.next_iteration_id;

    return *this;
}

void c_fn_cmd_foreach::clean_up()
{
    _what->clean_up();
}

void c_fn_cmd_foreach::add_inputs(vector<var_sym_index_t>& inputs)
{
    _what->add_inputs(inputs);
}

void c_fn_cmd_foreach::optimize_output(const vector<var_sym_index_t>& inputs)
{
    _what->optimize_output(inputs);
}

