/***************************************************************************
                          c_fn_cmd_foreach.h  -  description
                             -------------------
    begin                : Thu May 10 2007
    copyright            : (C) 2004-2007 by Marian VooDooMan Meravy
    email                : ghostvoodooman -a.T= users (dot) sourceforge.net
 ***************************************************************************/

/* $Id: c_fn_cmd_foreach.h,v 1.14 2008/03/11 21:11:36 ghostvoodooman Exp $ */

/***************************************************************************

VooDoo cIRCle - an IRC (ro)bot
Copyright (C) 2004-2008 by Marian VooDooMan Meravy
  (ghostvoodooman -a.T= users (dot) sourceforge.net)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************/

/*!
    \file
    \brief Class 'c_fn_cmd_foreach'
    \version $Id: c_fn_cmd_foreach.h,v 1.14 2008/03/11 21:11:36 ghostvoodooman Exp $
*/

#ifndef C_FN_CMD_FOREACH_H__
#define C_FN_CMD_FOREACH_H__

#if defined(HAVE_CONFIG_H)
#   include <config.h>
#endif

#include <string>
#include <vector>
#include <stack>
#include <list>
#include <map>
#include <algorithm>

#include "../utils/memleak.h"

#include "../params.h"

#include "c_fn_cmd.h"
#include "../func/c_expression.h"

/*!
    \brief Implements "foreach()" command
    \version 1
    \date 2006-01
    \author VooDooMan
*/
class c_fn_cmd_foreach : public c_fn_cmd
{
    c_expression* _what;    //!< Expression
    string _key;            //!< "key" variable (can be empty)
    string _as;             //!< "as" variable (can be empty)
    bool _as_ref;           //!< If "as" is referenced
    long end_jump_id;       //!< ID of command after loop
    long first_cmd_id;      //!< ID of first command in a loop
    long next_iteration_id; //!< ID of next iteration command
public:
    c_fn_cmd* copy();

    /*!
        \brief Constructor
        \version 1
        \date 2006-01
        \author VooDooMan
        \param debug Debug string
        \param id Unique ID of command
        \param what Expression to iterate
        \param key 'Key' variable (can be empty)
        \param as 'As' variable (can be empty)
        \param as_ref If 'as' variable is reference
    */
    c_fn_cmd_foreach(c_expression* what, const string& key, const string& as, bool as_ref, const string& debug, long id);

    bool sanity_check(string& msg);

    /*!
        \brief Sets ID of commnad after loop
        \version 1
        \date 2006-01
        \author VooDooMan
        \param id ID
    */
    void set_end_id(long id);

    /*!
        \brief Gets ID of commnad after loop
        \version 1
        \date 2006-01
        \author VooDooMan
    */
    long get_end_id();

    /*!
        \brief Sets ID of first commnad in a loop
        \version 1
        \date 2006-01
        \author VooDooMan
        \param id ID
    */
    void set_first_cmd_id(long id);

    /*!
        \brief Gets ID of first commnad in a loop
        \version 1
        \date 2006-01
        \author VooDooMan
    */
    long get_first_cmd_id();

    /*!
        \brief Sets ID of commnad for next iteration
        \version 1
        \date 2006-01
        \author VooDooMan
        \param id ID
    */
    void set_next_iteration_id(long id);

    /*!
        \brief Gets ID of commnad for next iteration
        \version 1
        \date 2006-01
        \author VooDooMan
    */
    long get_next_iteration_id();

    /*!
        \brief Gets expression
        \version 1
        \date 2006-01
        \author VooDooMan
    */
    c_expression* get_what();

    /*!
        \brief Gets "key" variable
        \version 1
        \date 2006-01
        \author VooDooMan
    */
    const string& get_key();

    /*!
        \brief Gets "as" variable
        \version 1
        \date 2006-01
        \author VooDooMan
    */
    const string& get_as();

    /*!
        \brief Gets "as" variable is referenced flag
        \version 1
        \date 2006-01
        \author VooDooMan
        \retval true It is referenced
        \retval false It is not referenced
    */
    bool get_as_ref();

    e_cmd_type get_type();

    /*!
        \brief Destructor
        \version 1
        \date 2006-05-09
        \author VooDooMan
    */
    virtual ~c_fn_cmd_foreach();

    /*!
        \brief Copy constructor
        \version 1
        \date 2006-05-09
        \author VooDooMan
        \param orig Original
    */
    c_fn_cmd_foreach(const c_fn_cmd_foreach& orig);

    /*!
        \brief Assign operator
        \version 1
        \date 2006-05-09
        \author VooDooMan
        \param orig Original
    */
    c_fn_cmd_foreach& operator=(const c_fn_cmd_foreach& orig);

    void clean_up();

    void add_inputs(vector<var_sym_index_t>& inputs);
    void optimize_output(const vector<var_sym_index_t>& inputs);
};

#endif

