/***************************************************************************
                          c_fn_cmd_goto.h  -  description
                             -------------------
    begin                : Thu May 10 2007
    copyright            : (C) 2004-2007 by Marian VooDooMan Meravy
    email                : ghostvoodooman -a.T= users (dot) sourceforge.net
 ***************************************************************************/

/* $Id: c_fn_cmd_goto.h,v 1.14 2008/03/11 21:11:36 ghostvoodooman Exp $ */

/***************************************************************************

VooDoo cIRCle - an IRC (ro)bot
Copyright (C) 2004-2008 by Marian VooDooMan Meravy
  (ghostvoodooman -a.T= users (dot) sourceforge.net)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************/

/*!
    \file
    \brief Class 'c_fn_cmd_goto'
    \version $Id: c_fn_cmd_goto.h,v 1.14 2008/03/11 21:11:36 ghostvoodooman Exp $
*/

#ifndef C_FN_CMD_GOTO_H__
#define C_FN_CMD_GOTO_H__

#if defined(HAVE_CONFIG_H)
#   include <config.h>
#endif

#include <string>
#include <vector>
#include <stack>
#include <list>
#include <map>
#include <algorithm>

#include "../utils/memleak.h"

#include "../params.h"

#include "c_fn_cmd.h"

/*!
    \brief Implements "goto" command
    \version 1
    \date 2006-01
    \author VooDooMan
*/
class c_fn_cmd_goto : public c_fn_cmd
{
    string _tag;            //!< "goto"'s label tag
    long label_id;          //!< ID of command as target of "goto"
public:
    c_fn_cmd* copy();

    /*!
        \brief Constructor
        \version 1
        \date 2006-01
        \author VooDooMan
        \param debug Debug string
        \param id Unique ID of command
        \param tag goto's tag
    */
    c_fn_cmd_goto(const string& tag, const string& debug, long id);

    bool sanity_check(string& msg);

    /*!
        \brief Gets "goto" label tag
        \version 1
        \date 2006-01
        \author VooDooMan
    */
    const string& get_tag();

    /*!
        \brief Gets ID of target of "goto" command
        \version 1
        \date 2006-01
        \author VooDooMan
    */
    long get_label_id();

    /*!
        \brief Sets ID of target of "goto" command
        \version 1
        \date 2006-01
        \author VooDooMan
        \param id ID
    */
    void set_label_id(long id);

    e_cmd_type get_type();

    /*!
        \brief Destructor
        \version 1
        \date 2006-05-09
        \author VooDooMan
    */
    virtual ~c_fn_cmd_goto();

    /*!
        \brief Copy constructor
        \version 1
        \date 2006-05-09
        \author VooDooMan
        \param orig Original
    */
    c_fn_cmd_goto(const c_fn_cmd_goto& orig);

    /*!
        \brief Assign operator
        \version 1
        \date 2006-05-09
        \author VooDooMan
        \param orig Original
    */
    c_fn_cmd_goto& operator=(const c_fn_cmd_goto& orig);

    void clean_up();

    void add_inputs(vector<var_sym_index_t>& inputs);
    void optimize_output(const vector<var_sym_index_t>& inputs);
};

#endif

