/***************************************************************************
                          c_fn_cmd_switch.cpp  -  description
                             -------------------
    begin                : Thu May 10 2007
    copyright            : (C) 2004-2007 by Marian VooDooMan Meravy
    email                : ghostvoodooman -a.T= users (dot) sourceforge.net
 ***************************************************************************/

/* $Id: c_fn_cmd_switch.cpp,v 1.14 2008/03/11 21:11:36 ghostvoodooman Exp $ */

/***************************************************************************

VooDoo cIRCle - an IRC (ro)bot
Copyright (C) 2004-2008 by Marian VooDooMan Meravy
  (ghostvoodooman -a.T= users (dot) sourceforge.net)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************/

/*!
    \file
    \brief Class 'c_fn_cmd_switch'
    \version $Id: c_fn_cmd_switch.cpp,v 1.14 2008/03/11 21:11:36 ghostvoodooman Exp $
*/

#if defined(HAVE_CONFIG_H)
#   include <config.h>
#endif

#include "../params.h"
#include "../utils/memleak.h"
#include "../utils/assert.h"
#include "../compat.h"

#include "c_fn_cmd_switch.h"
#include "../func/c_expression.h"
#include "../utils/precomp.h"

#if defined(_MSC_VER) || defined(__BORLANDC__)
#   pragma hdrstop
#endif

c_fn_cmd* c_fn_cmd_switch::copy()
{
    c_fn_cmd_switch* x=MY_NEW c_fn_cmd_switch(MY_NEW c_expression(*_exp),_debug,_id);
    x->cases_id=cases_id;
    x->no_match_id_jump=no_match_id_jump;
    return x;
}

c_fn_cmd_switch::c_fn_cmd_switch(c_expression* exp, const string& debug, long id)
    : c_fn_cmd(debug,id)
{
    _exp=exp;
}

bool c_fn_cmd_switch::sanity_check(string& msg)
{
    CHECK_ID(_id);
    CHECK_ID(no_match_id_jump);
    CHECK_EXP(_exp);
    return false;
}

void c_fn_cmd_switch::push_case_id(long id)
{
    cases_id.push_back(id);
}

c_expression* c_fn_cmd_switch::get_exp()
{
    return _exp;
}

vector<long> c_fn_cmd_switch::get_cases_id()
{
    return cases_id;
}

long c_fn_cmd_switch::get_no_match_id_jump()
{
    return no_match_id_jump;
}

void c_fn_cmd_switch::set_no_match_id_jump(long id)
{
    no_match_id_jump=id;
}

e_cmd_type c_fn_cmd_switch::get_type()
{
    return CMD_SWITCH;
}

c_fn_cmd_switch::~c_fn_cmd_switch()
{
    clean_up();
    _exp->clean_up();
    MY_DELETE(_exp);
}

c_fn_cmd_switch::c_fn_cmd_switch(const c_fn_cmd_switch& orig)
    : c_fn_cmd(orig)
{
    _exp=MY_NEW c_expression(*orig._exp);
    cases_id=orig.cases_id;
    no_match_id_jump=orig.no_match_id_jump;
}

c_fn_cmd_switch& c_fn_cmd_switch::operator=(const c_fn_cmd_switch& orig)
{
    _id=orig._id;
    _debug=orig._debug;

    _exp=MY_NEW c_expression(*orig._exp);
    cases_id=orig.cases_id;
    no_match_id_jump=orig.no_match_id_jump;

    return *this;
}

void c_fn_cmd_switch::clean_up()
{
    _exp->clean_up();
}

void c_fn_cmd_switch::add_inputs(vector<var_sym_index_t>& inputs)
{
    _exp->add_inputs(inputs);
}

void c_fn_cmd_switch::optimize_output(const vector<var_sym_index_t>& UNUSEDARG(inputs))
{
    // never optimize for 'switch'
    //_exp->optimize_output(inputs);
}

