/***************************************************************************
                          c_function.cpp  -  description
                             -------------------
    begin                : Thu May 10 2007
    copyright            : (C) 2004-2007 by Marian VooDooMan Meravy
    email                : ghostvoodooman -a.T= users (dot) sourceforge.net
 ***************************************************************************/

/* $Id: c_function.cpp,v 1.19 2008/11/24 12:09:23 ghostvoodooman Exp $ */

/***************************************************************************

VooDoo cIRCle - an IRC (ro)bot
Copyright (C) 2004-2008 by Marian VooDooMan Meravy
  (ghostvoodooman -a.T= users (dot) sourceforge.net)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************/

/*!
    \file
    \brief Class 'c_function'
    \version $Id: c_function.cpp,v 1.19 2008/11/24 12:09:23 ghostvoodooman Exp $
*/

#if defined(HAVE_CONFIG_H)
#   include <config.h>
#endif

#include "../params.h"
#include "../utils/memleak.h"
#include "../utils/assert.h"
#include "../compat.h"

#include "../microcode/c_micro_symbol_variable_table.h"
#include "../microcode/c_micro_symbol_table.h"
#include "c_function.h"
#include "../microcode/c_micro_symbol_variable_mapping.h"
#include "c_fn_cmd_do.h"
#include "c_fn_cmd_while.h"
#include "c_fn_cmd_jump.h"
#include "c_fn_cmd_dummy.h"
#include "c_fn_cmd_if.h"
#include "c_fn_cmd_exp.h"
#include "c_fn_cmd_for_cond.h"
#include "c_fn_cmd_foreach.h"
#include "c_fn_cmd_foreach_next_iter.h"
#include "c_fn_cmd_switch.h"
#include "c_fn_cmd_break.h"
#include "c_fn_cmd_continue.h"
#include "c_fn_cmd_goto.h"
#include "c_fn_cmd_global.h"
#include "c_fn_cmd_label.h"
#include "c_fn_cmd_case.h"
#include "c_fn_cmd_case_default.h"
#include "c_fn_cmd_right_brace.h"
#include "c_fn_cmd_do_while.h"
#include "c_fn_cmd_left_brace.h"
#include "c_fn_cmd_end_of_func.h"
#include "../utils/precomp.h"

#if defined(_MSC_VER) || defined(__BORLANDC__)
#   pragma hdrstop
#endif

string parse_get_func(string& str, bool& error, bool& need_more);
string explode_command(string& str, bool& error, bool& need_more, c_command::e_type& type, bool& semicolon, string& sub_exp);

void c_function::create_symbols()
{
    if(!symbols)
        symbols=MY_NEW c_micro_symbol_variable_mapping();
}

my_time_t c_function::get_last_changed()
{
    return last_changed;
}

void c_function::set_last_changed(my_time_t when)
{
    last_changed=when;
}

int c_function::get_minimal_botnet_version()
{
    return minimal_botnet_version;
}

bool c_function::is_okay()
{
    return done;
}

string c_function::get_name()
{
    return name;
}

vector<c_func_arg> c_function::get_args()
{
    return args;
}

int c_function::get_num_args()
{
    return num_args;
}

void c_function::set_dump(const string& func_dump)
{
    dump=func_dump;
}

void c_function::clear_microcode()
{
    map<long,c_fn_cmd*>::iterator i1;
    for(i1=commands.begin(); i1!=commands.end(); ++i1) {
        (*i1).second->clean_up();
        MY_DELETE((*i1).second);
    }
    commands.clear();
}

void c_function::set_symbols(c_micro_symbol_variable_mapping* s)
{
    symbols=s;
}

c_micro_symbol_table* c_function::make_symbol_table(c_micro_symbol_table* orig)
{
    if(symbol_table)
        return symbol_table;
    c_micro_symbol_variable_table* vars;
    if(!orig) {
        if(!symbols)
            symbols=MY_NEW c_micro_symbol_variable_mapping();
        vars=MY_NEW c_micro_symbol_variable_table(symbols->get_capacity());
    } else
        vars=MY_NEW c_micro_symbol_variable_table(*orig->var_table_);
    c_micro_symbol_table* table;
    if(!orig)
        table=MY_NEW c_micro_symbol_table(symbols,vars);
    else
        table=MY_NEW c_micro_symbol_table(*orig);

    symbol_table=table;

    return table;
}

bool c_function::set(string decl, string& error)
{
    error="";

    if(!attached_symbol_table)
        reset();
    else {
        c_micro_symbol_table* symt=symbol_table;
        c_micro_symbol_variable_mapping* syms=symbols;
        symbol_table=NULL;
        symbols=NULL;
        reset();
        if(symbols)
            MY_DELETE(symbols);
        symbol_table=symt;
        symbols=syms;
        attached_symbol_table=true;
    }

    int ctx=0;
    c_func_arg arg;
    int arg_num=1;
    bool has_arguments=false;
    for(size_t i1=0; i1<decl.length(); ++i1) {
        if(0x20==decl[i1])
            continue;
        bool is_alphanum=false;
        if((decl[i1]>='a' && decl[i1]<='z') ||
           (decl[i1]>='A' && decl[i1]<='Z') ||
           (decl[i1]>='0' && decl[i1]<='9') ||
           '_'==decl[i1])
           is_alphanum=true;
        if(0==ctx && is_alphanum) { // name of function
            name+=(string)""+decl[i1];
            continue;
        }
        if(0==ctx && '('==decl[i1]) {
            if(name.empty()) {
                error="Missing name of function";
                return false;
            }
            if(c_functions::is_built_in_function(name)) {
                error="Function with name '";
                error+=name;
                error+="' is reserved as built-in function";
                clear();
                return false;
            }
            ctx++;
            continue;
        }
        if(1==ctx && '&'==decl[i1]) { // argument reference prefix
            arg.reference=true;
            has_arguments=true;
            continue;
        }
        if(1==ctx && '$'==decl[i1]) { // argument '$' prefix
            arg.arg_expression="$";
            has_arguments=true;
            ctx++;
            continue;
        }
        if(2==ctx && is_alphanum) { // still argument
            arg.arg_expression+=decl[i1];
            continue;
        }
        if(2==ctx && ','==decl[i1]) { // end of argument
            if(arg.arg_expression.length()<2) {
                // it is an empty string, or just an "$"
                error="In declaration of function '";
                error+=name;
                error+="' argument #";
                error+=tos(arg_num);
                error+=" is erroneous";
                return false;
            }
            arg.value_sym=symbols->add(arg.arg_expression);
            args.push_back(arg);
            arg.clear();
            ctx=1;
            arg_num++;
            continue;
        }
        if((1==ctx || 2==ctx) && ')'==decl[i1]) { // end of argument
            if(has_arguments && arg.arg_expression.length()<2) {
                // it is an empty string, or just an "$"
                error="In declaration of function '";
                error+=name;
                error+="' argument #";
                error+=tos(arg_num);
                error+=" is erroneous";
                return false;
            }
            if(has_arguments) {
                arg.value_sym=symbols->add(arg.arg_expression);
                args.push_back(arg);
                arg_num++;
            }
            ctx++;
            continue;
        }
        if(3==ctx) {
            error="Invalid declaration of function '";
            error+=name;
            error+="' - some character(s) trail(s) after ending ')' character";
            return false;
        }
        error="Invalid function declaration";
        return false;
    }
    declaration=decl;
    num_args=arg_num-1;
    return true;
}

bool c_function::foreach_epilog(bool& error, string& error_msg, const string& debug, bool force)
{
    if(!block_stack.empty() && c_block::BLOCK_FOREACH==block_stack.back()._type) {
        if(block_stack.back().was_next_expression || force) {
            long id=block_stack.back().foreach_init_id;
            map<long,c_fn_cmd*>::iterator i1=commands.find(id);
            if(i1!=commands.end()) {
                c_fn_cmd_foreach* c=NULL;
                if((*i1).second && typeid(*(*i1).second)==typeid(c_fn_cmd_foreach))
                    c=dynamic_cast<c_fn_cmd_foreach*>((*i1).second);
                if(c) {
                    if(c->get_end_id()!=JUMP_ID_INVALID) {
                        error_msg="Internal error (foreach)";
                        error_msg+=(string)"; "+debug;
                        error_msg=GET_STR_DEBUG_INFO_EX(error_msg);
                        error=true;
                        return false;
                    } else
                        c->set_end_id(new_id+2);

                    c->set_next_iteration_id(new_id);
                    {
                        c_fn_cmd_foreach_next_iter* cmd=MY_NEW c_fn_cmd_foreach_next_iter(c->get_first_cmd_id(),debug,new_id);
                        commands.insert(make_pair(new_id,cmd));
                        new_id++;
                    }
                    {
                        c_fn_cmd_jump* cmd=MY_NEW c_fn_cmd_jump(c->get_first_cmd_id(),debug,new_id);
                        commands.insert(make_pair(new_id,cmd));
                        new_id++;
                    }
                    {
                        c_fn_cmd_dummy* cmd=MY_NEW c_fn_cmd_dummy(new_id);
                        commands.insert(make_pair(new_id,cmd));
                        new_id++;
                    }
                } else {
                    error_msg="Internal compile error (dynamic_cast<c_fn_cmd_foreach*> == NULL";
                    error_msg+=(string)", should be "+typeid(*(*i1).second).name()+")";
                    error_msg+=(string)"; "+debug;
                    error_msg=GET_STR_DEBUG_INFO_EX(error_msg);
                    error=true;
                    return false;
                }
            } else {
                error_msg="Internal compile error (not found matching 'foreach')";
                error_msg+=(string)"; "+debug;
                error_msg=GET_STR_DEBUG_INFO_EX(error_msg);
                error=true;
                return false;
            }
            block_stack.pop_back();
            return true;
        } else
            block_stack.back().was_next_expression=true;
    }
    return false;
}

bool c_function::do_epilog(bool& error, string& error_msg, const string& debug, bool force)
{
    if(!block_stack.empty() && c_block::BLOCK_DO==block_stack.back()._type) {
        if(block_stack.back().was_next_expression || force) {
            long id=block_stack.back().do_id;
            {
                c_fn_cmd_dummy* cmd=MY_NEW c_fn_cmd_dummy(new_id);
                commands.insert(make_pair(new_id,cmd));
            }

            map<long,c_fn_cmd*>::iterator i1=commands.find(id);
            if(i1!=commands.end()) {
                c_fn_cmd_do* c=NULL;
                if((*i1).second && typeid(*(*i1).second)==typeid(c_fn_cmd_do))
                    c=dynamic_cast<c_fn_cmd_do*>((*i1).second);
                if(c) {
                    if(c->get_end_loop_id()!=JUMP_ID_INVALID) {
                        /*
                        error_msg="Internal error (do)";
                        error_msg+=(string)"; "+debug;
                        error_msg=GET_STR_DEBUG_INFO_EX(error_msg);
                        error=true;
                        return false;
                        */
                    } else
                        c->set_end_loop_id(new_id);
                } else {
                    error_msg="Internal compile error (dynamic_cast<c_fn_cmd_do*> == NULL";
                    error_msg+=(string)", should be "+typeid(*(*i1).second).name()+")";
                    error_msg+=(string)"; "+debug;
                    error_msg=GET_STR_DEBUG_INFO_EX(error_msg);
                    error=true;
                    return false;
                }
            } else {
                error_msg="Internal compile error (not found matching 'do')";
                error_msg+=(string)"; "+debug;
                error_msg=GET_STR_DEBUG_INFO_EX(error_msg);
                error=true;
                return false;
            }
            new_id++;
            block_stack.pop_back();
            return true;
        } else
            block_stack.back().was_next_expression=true;
    }
    return false;
}

bool c_function::while_epilog(bool& error, string& error_msg, const string& debug, bool force)
{
    if(!block_stack.empty() && c_block::BLOCK_WHILE==block_stack.back()._type) {
        if(block_stack.back().was_next_expression || force) {
            long id=block_stack.back().while_again_id;
            map<long,c_fn_cmd*>::iterator i1=commands.find(id);
            c_fn_cmd_while* c=NULL;
            if(i1!=commands.end()) {
                c=NULL;
                if((*i1).second && typeid(*(*i1).second)==typeid(c_fn_cmd_while))
                    c=dynamic_cast<c_fn_cmd_while*>((*i1).second);
                if(c) {
                    if(c->get_end_loop_id()!=JUMP_ID_INVALID) {
                        error_msg="Internal error (while)";
                        error_msg+=(string)"; "+debug;
                        error_msg=GET_STR_DEBUG_INFO_EX(error_msg);
                        error=true;
                        return false;
                    }
                    c->set_end_loop_id(new_id+1);
                } else {
                    error_msg="Internal compile error (dynamic_cast<c_fn_cmd_while*> == NULL";
                    error_msg+=(string)", should be "+typeid(*(*i1).second).name()+")";
                    error_msg+=(string)"; "+debug;
                    error_msg=GET_STR_DEBUG_INFO_EX(error_msg);
                    error=true;
                    return false;
                }
            } else {
                error_msg="Internal compile error (not found matching 'while')";
                error_msg+=(string)"; "+debug;
                error_msg=GET_STR_DEBUG_INFO_EX(error_msg);
                error=true;
                return false;
            }
            if(c) {
                c_fn_cmd_jump* cmd=MY_NEW c_fn_cmd_jump(c->get_id(),c->get_debug_string(),new_id);
                commands.insert(make_pair(new_id,cmd));
                new_id++;
            }
            {
                c_fn_cmd_dummy* cmd=MY_NEW c_fn_cmd_dummy(new_id);
                commands.insert(make_pair(new_id,cmd));
                new_id++;
            }
            block_stack.pop_back();
            return true;
        } else
            block_stack.back().was_next_expression=true;
    }
    return false;
}

void c_function::resolve_if(bool& error, string& error_msg, const string& debug)
{
    // finish
    if(if_bunch.size()) {
        c_fn_cmd_dummy* cmd=MY_NEW c_fn_cmd_dummy(new_id);
        commands.insert(make_pair(new_id,cmd));

        {
            vector<c_if_branch>::iterator i1;
            for(i1=if_bunch.begin(); i1!=if_bunch.end(); ++i1) {
                map<long,c_fn_cmd*>::iterator i2;
                if((*i1).if_id!=JUMP_ID_INVALID) {
                    i2=commands.find((*i1).if_id);
                    if(i2==commands.end()) {
                        error=true;
                        error_msg="Internal compile error: could not find 'if' ID (ID is invalid?) - 'if'";
                        error_msg+=(string)"; "+debug;
                        error_msg=GET_STR_DEBUG_INFO_EX(error_msg);
                        return;
                    }
                    c_fn_cmd_if* i=NULL;
                    if((*i2).second && typeid(*(*i2).second)==typeid(c_fn_cmd_if))
                        i=dynamic_cast<c_fn_cmd_if*>((*i2).second);
                    if(!i) {
                        error=true;
                        error_msg="Internal compile error (dynamic_cast<c_fn_cmd_if*> == NULL";
                        error_msg+=(string)", should be "+typeid(*(*i2).second).name()+")";
                        error_msg+=(string)"; "+debug;
                        error_msg=GET_STR_DEBUG_INFO_EX(error_msg);
                        return;
                    }
                    if(i->get_else_jump_id()==JUMP_ID_INVALID || i->get_else_jump_id()==JUMP_ID_NA)
                        i->set_else_jump_id(new_id);
                }
                if((*i1).jump_id!=JUMP_ID_INVALID) {
                    i2=commands.find((*i1).jump_id);
                    if(i2==commands.end()) {
                        error=true;
                        error_msg="Internal compile error: could not find jump ID (ID is invalid?) - 'if'";
                        error_msg+=(string)"; "+debug;
                        error_msg=GET_STR_DEBUG_INFO_EX(error_msg);
                        return;
                    }
                    c_fn_cmd_jump* j=NULL;
                    if((*i2).second && typeid(*(*i2).second)==typeid(c_fn_cmd_jump))
                        j=dynamic_cast<c_fn_cmd_jump*>((*i2).second);
                    if(!j) {
                        error=true;
                        error_msg="Internal compile error (dynamic_cast<c_fn_cmd_jump*> == NULL";
                        error_msg+=(string)", should be "+typeid(*(*i2).second).name()+")";
                        error_msg+=(string)"; "+debug;
                        error_msg=GET_STR_DEBUG_INFO_EX(error_msg);
                        return;
                    }
                    if(j->get_jump_to_id()==JUMP_ID_INVALID || j->get_jump_to_id()==JUMP_ID_NA)
                        j->set_jump_to_id(new_id);
                }
            }
            if_bunch.clear();
        }

        new_id++;
    }
}

bool c_function::for_epilog(bool& error, string& error_msg, const string& debug)
{
    if(block_stack.empty() || c_block::BLOCK_FOR!=block_stack.back()._type)
        return false;

    long iterative_exp_id=new_id;

    {
        // iterative expression
        c_fn_cmd_exp* cmd=MY_NEW c_fn_cmd_exp(block_stack.back().for_exp_iter,debug,new_id);
        commands.insert(make_pair(new_id,cmd));
        new_id++;
    }
    {
        // jump up
        c_fn_cmd_jump* cmd=MY_NEW c_fn_cmd_jump(block_stack.back().for_iter_jump,debug,new_id);
        commands.insert(make_pair(new_id,cmd));
        new_id++;
    }
    {
        // resolve FALSE condition
        c_fn_cmd_dummy* cmd=MY_NEW c_fn_cmd_dummy(new_id);
        commands.insert(make_pair(new_id,cmd));
        map<long,c_fn_cmd*>::iterator i1=commands.find(block_stack.back().for_iter_jump);
        if(i1!=commands.end()) {
            c_fn_cmd_for_cond* cond=NULL;
            if((*i1).second && typeid(*(*i1).second)==typeid(c_fn_cmd_for_cond))
                cond=dynamic_cast<c_fn_cmd_for_cond*>((*i1).second);
            if(cond) {
                cond->set_false_jump_id(new_id);
                cond->set_iterative_exp_id(iterative_exp_id);
            } else {
                error=true;
                error_msg="Internal compile error (dynamic_cast<c_fn_cmd_for_cond*> == NULL";
                error_msg+=(string)", should be "+typeid(*(*i1).second).name()+")";
                error_msg+=(string)"; "+debug;
                error_msg=GET_STR_DEBUG_INFO_EX(error_msg);
                return false;
            }
        } else {
            error=true;
            error_msg="Internal compile error (not found matching 'for')";
            error_msg+=(string)"; "+debug;
            error_msg=GET_STR_DEBUG_INFO_EX(error_msg);
            return false;
        }
        new_id++;
    }
    block_stack.pop_back();
    return true;
}

bool c_function::all_epilogs(bool& error, string& error_msg, const string& debug)
{
    for(;;) {
        bool got=false;

        if(!got && for_epilog(error,error_msg,debug))
            got=true;

        if(error)
            return false;

        if(!got && foreach_epilog(error,error_msg,debug,true))
            got=true;

        if(error)
            return false;

        // 'do' epilog resolved elsewhere

        if(error)
            return false;

        if(!got && while_epilog(error,error_msg,debug,true))
            got=true;

        if(error)
            return false;

        if(!got && switch_epilog(error,error_msg,debug))
            got=true;

        if(error)
            return false;

        if(!got)
            break;
    }
    return true;
}

bool c_function::switch_epilog(bool& error, string& error_msg, const string& debug)
{
    if(block_stack.size() && c_block::BLOCK_SWITCH==block_stack.back()._type) {
        // finish switch
        // find last switch() ID
        bool got=false;
        deque<c_block>::reverse_iterator i2;
        for(i2=block_stack.rbegin(); i2!=block_stack.rend(); ++i2) {
            if(c_block::BLOCK_SWITCH==(*i2)._type) {
                got=true;
                break;
            }
        }
        if(!got) {
            error=true;
            error_msg="Syntax error (case statement)";
            error_msg+=(string)"; "+debug;
            error_msg=GET_STR_DEBUG_INFO_EX(error_msg);
            return false;
        }
        long id=(*i2).switch_id;
        //
        map<long,c_fn_cmd*>::iterator i1=commands.find(id);
        if(i1!=commands.end()) {
            c_fn_cmd_switch* c=NULL;
            if((*i1).second && typeid(*(*i1).second)==typeid(c_fn_cmd_switch))
                c=dynamic_cast<c_fn_cmd_switch*>((*i1).second);
            if(c) {
                c->set_no_match_id_jump(new_id);
                c_fn_cmd_dummy* cmd=MY_NEW c_fn_cmd_dummy(new_id);
                commands.insert(make_pair(new_id,cmd));
                new_id++;
            } else {
                error=true;
                error_msg="Internal compile error (dynamic_cast<c_fn_cmd_switch*> == NULL";
                error_msg+=(string)", should be "+typeid(*(*i1).second).name()+")";
                error_msg+=(string)"; "+debug;
                error_msg=GET_STR_DEBUG_INFO_EX(error_msg);
                return false;
            }
        } else {
            error=true;
            error_msg="Internal compile error (not found matching 'switch')";
            error_msg+=(string)"; "+debug;
            error_msg=GET_STR_DEBUG_INFO_EX(error_msg);
            return false;
        }
        block_stack.pop_back();
        return true;
    }
    return false;
}

c_function::e_parse_result c_function::parse_else(const string& cmd, bool& got_cmd, bool& error, string& error_msg, const string& debug)
{
    if(!cmd.compare("else;")) {
        got_cmd=true;
        // resolve ELSE condition

        block_stack.back().ignore_next=true;

        if(!block_stack.size() || c_block::BLOCK_IF!=block_stack.back()._type) {
            error=true;
            error_msg="Syntax error ('else')";
            error_msg+=(string)"; "+debug;
            error_msg=GET_STR_DEBUG_INFO_EX(error_msg);
            return PARSE_FALSE;
        }

        {
            // if we were here (if was evaluated as TRUE),
            //     then we should jump over 'if' bunch
            c_fn_cmd_jump* cmd=MY_NEW c_fn_cmd_jump(JUMP_ID_INVALID,debug,new_id);
            commands.insert(make_pair(new_id,cmd));
            c_if_branch b;
            b.jump_id=new_id;
            if_bunch.push_back(b);
        }

        long if_id=block_stack.back().if_id;

        map<long,c_fn_cmd*>::iterator i1=commands.find(if_id);
        if(i1!=commands.end()) {
            c_fn_cmd_if* c_if=NULL;
            if((*i1).second && typeid(*(*i1).second)==typeid(c_fn_cmd_if))
                c_if=dynamic_cast<c_fn_cmd_if*>((*i1).second);
            if(c_if) {
                c_if->set_else_jump_id(new_id+1);
            } else {
                error=true;
                error_msg="Internal compile error (dynamic_cast<c_fn_cmd_if*> == NULL";
                error_msg+=(string)", should be "+typeid(*(*i1).second).name()+")";
                error_msg+=(string)"; "+debug;
                error_msg=GET_STR_DEBUG_INFO_EX(error_msg);
                return PARSE_FALSE;
            }
        } else {
            error=true;
            error_msg="Internal compile error (not found matching 'if')";
            error_msg+=(string)"; "+debug;
            error_msg=GET_STR_DEBUG_INFO_EX(error_msg);
            return PARSE_FALSE;
        }
        new_id++;
        {
            c_fn_cmd_dummy* cmd=MY_NEW c_fn_cmd_dummy(new_id);
            commands.insert(make_pair(new_id,cmd));
            new_id++;
        }
        return PARSE_CONT;
    }
    return PARSE_FALL_THRU;
}

#define DO_ERROR() \
    { \
        error=true; \
        return PARSE_FALSE; \
    }

#define CAST(TYPE) \
    TYPE* c=NULL; \
    if((*i1).second && typeid(*(*i1).second)==typeid(TYPE)) \
        c=dynamic_cast<TYPE*>((*i1).second); \
    if(!c) { \
        error=true; \
        error_msg=(string)"Internal error: dynamic_cast<" + #TYPE + "*> == NULL, " \
        "should be " + typeid(*(*i1).second).name() + "; "; \
        error_msg+=(*i1).second->get_debug_string(); \
        error_msg=GET_STR_DEBUG_INFO_EX(error_msg); \
        return PARSE_FALSE; \
    }

#define CAST2(TYPE) \
    TYPE* c2=NULL; \
    if((*i2).second && typeid(*(*i2).second)==typeid(TYPE)) \
        c2=dynamic_cast<TYPE*>((*i2).second); \
    if(!c2) { \
        error=true; \
        error_msg=(string)"Internal error: dynamic_cast<" + #TYPE + "*> == NULL, " \
        "should be " + typeid(*(*i2).second).name() + "; "; \
        error_msg+=(*i2).second->get_debug_string(); \
        error_msg=GET_STR_DEBUG_INFO_EX(error_msg); \
        return PARSE_FALSE; \
    }

#define CAST3(TYPE) \
    TYPE* c2=NULL; \
    if((*i3).second && typeid(*(*i3).second)==typeid(TYPE)) \
        c2=dynamic_cast<TYPE*>((*i3).second); \
    if(!c2) { \
        error=true; \
        error_msg=(string)"Internal error: dynamic_cast<" + #TYPE + "*> == NULL, " \
        "should be " + typeid(*(*i3).second).name() + "; "; \
        error_msg+=(*i3).second->get_debug_string(); \
        error_msg=GET_STR_DEBUG_INFO_EX(error_msg); \
        return PARSE_FALSE; \
    }

c_function::e_parse_result c_function::parse_break_continue(const string& cmd, bool& got_cmd, bool& error, string& error_msg, const string& debug)
{
    if(!cmd.compare("break;") || !cmd.compare("continue;")) {
        got_cmd=true;
        // find last breakable/continuable command
        bool its_break=!cmd.compare("break;");

        long cmd_id=JUMP_ID_INVALID;
        bool got=false;
        deque<c_block>::reverse_iterator i2;
        for(i2=block_stack.rbegin(); i2!=block_stack.rend(); ++i2) {
            switch((*i2)._type) {
                case c_block::BLOCK_DO:
                    got=true;
                    {
                        map<long,c_fn_cmd*>::iterator i3=commands.find((*i2)._id);
                        if(i3==commands.end()) {
                            error_msg="Internal error: could not find ID ('do'); ";
                            error_msg+=debug;
                            error_msg+="; ";
                            error_msg+=(*i2)._debug;
                            error_msg=GET_STR_DEBUG_INFO_EX(error_msg);
                            error=true;
                            return PARSE_FALSE;
                        }
                        CAST3(c_fn_cmd_do);
                        cmd_id=c2->get_id();
                    }
                    break;
                case c_block::BLOCK_FOR:
                    got=true;
                    {
                        map<long,c_fn_cmd*>::iterator i3=commands.find((*i2)._id);
                        if(i3==commands.end()) {
                            error_msg="Internal error: could not find ID ('for'); ";
                            error_msg+=debug;
                            error_msg+="; ";
                            error_msg+=(*i2)._debug;
                            error_msg=GET_STR_DEBUG_INFO_EX(error_msg);
                            error=true;
                            return PARSE_FALSE;
                        }
                        CAST3(c_fn_cmd_for_cond);
                        cmd_id=(*i2).for_iter_jump;
                    }
                    break;
                case c_block::BLOCK_FOREACH:
                    got=true;
                    {
                        map<long,c_fn_cmd*>::iterator i3=commands.find((*i2)._id);
                        if(i3==commands.end()) {
                            error_msg="Internal error: could not find ID ('foreach'); ";
                            error_msg+=debug;
                            error_msg+="; ";
                            error_msg+=(*i2)._debug;
                            error_msg=GET_STR_DEBUG_INFO_EX(error_msg);
                            error=true;
                            return PARSE_FALSE;
                        }
                        CAST3(c_fn_cmd_foreach);
                        cmd_id=c2->get_id();
                    }
                    break;
                case c_block::BLOCK_SWITCH:
                    got=true;
                    {
                        map<long,c_fn_cmd*>::iterator i3=commands.find((*i2)._id);
                        if(i3==commands.end()) {
                            error_msg="Internal error: could not find ID ('switch'); ";
                            error_msg+=debug;
                            error_msg+="; ";
                            error_msg+=(*i2)._debug;
                            error_msg=GET_STR_DEBUG_INFO_EX(error_msg);
                            error=true;
                            return PARSE_FALSE;
                        }
                        CAST3(c_fn_cmd_switch);
                        cmd_id=c2->get_id();
                    }
                    break;
                case c_block::BLOCK_WHILE:
                    got=true;
                    {
                        map<long,c_fn_cmd*>::iterator i3=commands.find((*i2)._id);
                        if(i3==commands.end()) {
                            error_msg="Internal error: could not find ID ('while'); ";
                            error_msg+=debug;
                            error_msg+="; ";
                            error_msg+=(*i2)._debug;
                            error_msg=GET_STR_DEBUG_INFO_EX(error_msg);
                            error=true;
                            return PARSE_FALSE;
                        }
                        CAST3(c_fn_cmd_while);
                        cmd_id=c2->get_id();
                    }
                    break;
                default:
                    break;
            }
            if(got)
                break;
        }
        if(!got) {
            error=true;
            error_msg=(string)"Syntax error: illegal '"+(its_break?"break":"continue")+"'";
            error_msg+=(string)"; "+debug;
            error_msg=GET_STR_DEBUG_INFO_EX(error_msg);
            return PARSE_FALSE;
        }
        if(its_break) {
            c_fn_cmd_break* cmd=MY_NEW c_fn_cmd_break(debug,new_id);
            cmd->set_begin_iteration_command_id(cmd_id);
            if(c_block::BLOCK_FOREACH==(*i2)._type)
                cmd->set_for_foreach();
            if(c_block::BLOCK_SWITCH==(*i2)._type)
                cmd->set_for_switch();
            commands.insert(make_pair(new_id,cmd));
            new_id++;
        } else {
            c_fn_cmd_continue* cmd=MY_NEW c_fn_cmd_continue(debug,new_id);
            if(c_block::BLOCK_FOR==(*i2)._type) {
                cmd->set_begin_iteration_command_id((*i2).for_iter_jump);
            } else
                cmd->set_begin_iteration_command_id(cmd_id);
            commands.insert(make_pair(new_id,cmd));
            new_id++;
        }
        return PARSE_CONT;
    }
    return PARSE_FALL_THRU;
}

c_function::e_parse_result c_function::parse_foreach_1(string& cmd, bool& got_cmd, bool& error, string& error_msg, const string& debug, const c_command::e_type type, c_module* this_module)
{
    if(c_command::CMD_FOREACH_1==type) {
        got_cmd=true;
        if(cmd.find("foreach(",0)!=string::npos)
            cmd.erase(cmd.find("foreach(",0),strlen("foreach("));
        if(cmd.length() && ')'==cmd[cmd.length()-1])
            cmd.erase(cmd.length()-1,1);
        string what;
        string as; // "as"
        string as_element;
        string key_oper; // "=>"
        string value;
        bool ref_value=false;
        int ctx=0;
        bool in_str=false;
        int in_brace=0;
        for(size_t i1=0; i1<cmd.length(); ++i1) {
            char ch=cmd[i1];
            if('"'==ch) {
                in_str=!in_str;
            }
            if('['==ch && !in_str)
                in_brace++;
            if(']'==ch && in_brace && !in_str)
                in_brace--;
            if(0==ctx && ( (ch>='a' && ch<='z') || (ch>='A' && ch<='Z') || (ch>='0' && ch<='9') || '_'==ch || '$'==ch || '"'==ch || in_str || ']'==ch || in_brace)) {
                what+=(string)""+ch;
                continue;
            }
            if(0==ctx && 0x20==ch) {
                if(what.empty())
                    continue;
                ctx++;
                continue;
            }

            if(1==ctx && (ch>='a' && ch<='z')) {
                as+=(string)""+cmd[i1];
                continue;
            }
            if(1==ctx && 0x20==cmd[i1]) {
                if(as.empty())
                    continue;
                if(as.compare("as")) {
                    error=true;
                    error_msg="Invalid syntax: expecting 'as'";
                    error_msg+=(string)"; "+debug;
                    error_msg=GET_STR_DEBUG_INFO_EX(error_msg);
                    return PARSE_FALSE;
                }
                ctx++;
                continue;
            }

            if(2==ctx && ( (ch>='a' && ch<='z') || (ch>='A' && ch<='Z') || (ch>='0' && ch<='9') || '_'==ch || '$'==ch)) {
                as_element+=(string)""+ch;
                continue;
            }
            if(2==ctx && 0x20==ch) {
                if(as_element.empty())
                    continue;
                ctx++;
                continue;
            }
            if(2==ctx && '='==ch) {
                if(as_element.empty()) {
                    error=true;
                    error_msg="Invalid syntax";
                    error_msg+=(string)"; "+debug;
                    error_msg=GET_STR_DEBUG_INFO_EX(error_msg);
                    return PARSE_FALSE;
                }
                ctx++;
                // fall through
            }
            if(2==ctx && '&'==ch) {
                if(ref_value || !as_element.empty()) {
                    error=true;
                    error_msg="Invalid syntax";
                    error_msg+=(string)"; "+debug;
                    error_msg=GET_STR_DEBUG_INFO_EX(error_msg);
                    return PARSE_FALSE;
                }
                ref_value=true;
                continue;
            }

            if(3==ctx && ('='==ch || '>'==ch)) {
                key_oper+=(string)""+ch;
                continue;
            }
            if(3==ctx && 0x20==ch) {
                if(key_oper.empty())
                    continue;
                if(key_oper.compare("=>")) {
                    error=true;
                    error_msg="Invalid syntax: expecting '=>'";
                    error_msg+=(string)"; "+debug;
                    error_msg=GET_STR_DEBUG_INFO_EX(error_msg);
                    return PARSE_FALSE;
                }
                ctx++;
                continue;
            }
            if(3==ctx && ('$'==ch || '&'==ch)) {
                if(key_oper.compare("=>")) {
                    error=true;
                    error_msg="Invalid syntax: expecting '=>'";
                    error_msg+=(string)"; "+debug;
                    error_msg=GET_STR_DEBUG_INFO_EX(error_msg);
                    return PARSE_FALSE;
                }
                ctx++;
                // fall through
            }

            if(4==ctx && ( (ch>='a' && ch<='z') || (ch>='A' && ch<='Z') || (ch>='0' && ch<='9') || '_'==ch || '$'==ch)) {
                value+=(string)""+ch;
                continue;
            }
            if(4==ctx && 0x20==ch) {
                if(value.empty())
                    continue;
                ctx++;
                continue;
            }
            if(4==ctx && '&'==ch) {
                if(ref_value || !value.empty()) {
                    error=true;
                    error_msg="Invalid syntax";
                    error_msg+=(string)"; "+debug;
                    error_msg=GET_STR_DEBUG_INFO_EX(error_msg);
                    return PARSE_FALSE;
                }
                ref_value=true;
                continue;
            }

            error=true;
            error_msg="Invalid syntax";
            error_msg+=(string)"; "+debug;
            error_msg=GET_STR_DEBUG_INFO_EX(error_msg);
            return PARSE_FALSE;
        }
        c_command c;

        c.what_exp=MY_NEW c_expression();
        c.what_exp->assign(this_module,what,debug,*symbols,"","");
        if(c.what_exp->got_error()) {
            c.what_exp->clean_up();
            MY_DELETE(c.what_exp);
            error=true;
            error_msg="Compile error";
            error_msg+=(string)"; "+debug;
            error_msg=GET_STR_DEBUG_INFO_EX(error_msg);
            return PARSE_FALSE;
        }

        {
            string key, as;

            if(as_element.empty()) {
            } else if(key_oper.empty()) {
                as=as_element;
            } else {
                key=as_element;
                as=value;
            }

            if(!key.empty()) {
                symbols->add(key);
            }
            if(!as.empty()) {
                symbols->add(as);
            }

            c_block blck(c_block::BLOCK_FOREACH,new_id,debug);
            blck.foreach_init_id=new_id;
            blck.foreach_iter_jump=new_id+1;
            blck.was_next_expression=false;
            block_stack.push_back(blck);

            {
                c_fn_cmd_foreach* cmd=MY_NEW c_fn_cmd_foreach(c.what_exp,key,as,ref_value,debug,new_id);
                cmd->set_first_cmd_id(blck.foreach_iter_jump);
                commands.insert(make_pair(new_id,cmd));
            }

            new_id++;

            {
                c_fn_cmd_dummy* cmd=MY_NEW c_fn_cmd_dummy(new_id);
                commands.insert(make_pair(new_id,cmd));
                new_id++;
            }
        }
        return PARSE_CONT;
    }
    return PARSE_FALL_THRU;
}

bool c_function::parse_command(c_module* this_module, string& command, bool& error, string& error_msg, bool& need_more, const string& debug, const string& debug_prefix)
{
    typedef vector<string> vs;
    typedef vector<string>::iterator vsi;
    need_more=false;
    if(command.empty()) {
        need_more=true;
        return true;
    }
    if('#'==command[0] && 1==version) {
        vs args;
        string first=explode(command,0x20,args);
        bool got=false;
        if(!first.compare("#version")) {
            got=true;
            if(line!=1) {
                error=true;
                error_msg="'#version' can be only on the first line of script";
                error_msg+=(string)"; "+debug;
                error_msg=GET_STR_DEBUG_INFO_EX(error_msg);
                return false;
            }
            vsi i1=args.begin()+1;
            if(args.end()==i1) {
                error=true;
                error_msg="Incorrect use of '#version'";
                error_msg+=(string)"; "+debug;
                error_msg=GET_STR_DEBUG_INFO_EX(error_msg);
                return false;
            }

            while(!(*i1).empty() && (
                '\r'==(*i1)[(*i1).length()-1]
                || '\n'==(*i1)[(*i1).length()-1]
                )
                    )
                    (*i1).erase((*i1).length()-1,1);

            if(!(*i1).compare("1"))
                version=1;
            else if(!(*i1).compare("2"))
                version=2;
            else {
                error=true;
                error_msg="Unsupported version of file according to '#version'";
                error_msg+=(string)"; "+debug;
                error_msg=GET_STR_DEBUG_INFO_EX(error_msg);
                return false;
            }
            if(i1+1!=args.end()) {
                error=true;
                error_msg="Incorrect argument count to '#version'";
                error_msg+=(string)"; "+debug;
                error_msg=GET_STR_DEBUG_INFO_EX(error_msg);
                return false;
            }
            line++;
            need_more=true;
            command="";
            return true;
        }
        if(!got) {
            error=true;
            error_msg="Unknown '#'";
            error_msg+=(string)"; "+debug;
            error_msg=GET_STR_DEBUG_INFO_EX(error_msg);
            return false;
        }
    }
    string new_command;
    if(version==2) {
        size_t i1;
        {
            bool in_esc=false;
            char in_str=0;
            for(i1=0; i1<command.length(); ++i1) {
                if(!in_str && !in_comment && command.find("/*",i1)==i1) {
                    in_comment=true;
                    i1++;
                    continue;
                }
                if(!in_str && !in_comment && command.find("//",i1)==i1) {
                    for(; i1<command.length(); ++i1) {
                        if('\n'==command[i1]) {
                            ++i1;
                            break;
                        }
                    }
                    if(i1>=command.length()) {
                        command=new_command;
                        need_more=true;
                        return true;
                    }
                    continue;
                }
                if(in_comment && command.find("*/",i1)==i1) {
                    in_comment=false;
                    i1++;
                    continue;
                }
                if(in_comment) {
                    continue;
                }
                if('\\'==command[i1]) {
                    if(in_esc) {
                        in_esc=false;
                        new_command+=(string)""+(command[i1]);
                        continue;
                    }
                    in_esc=true;
                    new_command+=(string)""+(command[i1]);
                    continue;
                }
                if(in_esc) {
                    new_command+=(string)""+(command[i1]);
                    in_esc=false;
                    continue;
                }
                if('\''!=in_str && '"'==command[i1]) {
                    new_command+=(string)""+(command[i1]);
                    if('"'==in_str)
                        in_str=0;
                    else
                        in_str='"';
                    continue;
                }
                if('"'!=in_str && '\''==command[i1]) {
                    new_command+=(string)""+(command[i1]);
                    if('\''==in_str)
                        in_str=0;
                    else
                        in_str='\'';
                    continue;
                }
                if(!in_str) {
                    if(
                        '\x20'==command[i1] || '\r'==command[i1] ||
                        '\n'==command[i1] || '\t'==command[i1]
                    ) {
                        if(!new_command.empty() && '\x20'!=new_command[new_command.length()-1])
                            new_command+="\x20";
                      continue;
                    }
                }
                new_command+=(string)""+(command[i1]);
            }
            if(in_comment || in_str) {
                need_more=true;
                return true;
            }
            if(new_command.empty()) {
                need_more=true;
                return true;
            }
        }
        vs args;
        string first=explode(new_command,'\x20',args);
        bool got=false;
        if(!first.compare("function")) {
            got=true;
            if(!name.empty()) {
                error=true;
                error_msg="Invalid usage of 'function' keyword";
                error_msg+=(string)"; "+debug;
                error_msg=GET_STR_DEBUG_INFO_EX(error_msg);
                return false;
            }

            string com=new_command;
            com.erase(0,sizeof("function")-1);
            string func_def=parse_get_func(com,error,need_more);
            if(need_more)
                return true;
            if(error) {
                error_msg="Invalid function definition";
                error_msg+=(string)"; "+debug;
                error_msg=GET_STR_DEBUG_INFO_EX(error_msg);
                return false;
            }
            if(!set(func_def,error_msg)) {
                error=true;
                error_msg=GET_STR_DEBUG_INFO_EX(error_msg);
                return false;
            }
        }
        if(!name.empty() && !new_command.empty() && '#'==new_command[0]) {
            vsi i1=args.begin();
            new_command=*i1;
            got=false;
            if(i1!=args.end())
                i1++;
            else {
                error_msg="Incorrect syntax";
                error_msg+=(string)"; "+debug;
                error=true;
                error_msg=GET_STR_DEBUG_INFO_EX(error_msg);
                return false;
            }
            if(!new_command.compare("#last_changed")) {
                if(args.end()==i1) {
                    error_msg="Incorrect syntax";
                    error_msg+=(string)"; "+debug;
                    error_msg=GET_STR_DEBUG_INFO_EX(error_msg);
                    error=true;
                    return false;
                }
                my_time_t when=sto_time((*i1));
                set_last_changed(when);
                got=true;
            }
            if(!new_command.compare("#remote_bot_call")) {
                if(args.end()==i1) {
                    error_msg="Incorrect syntax";
                    error_msg+=(string)"; "+debug;
                    error_msg=GET_STR_DEBUG_INFO_EX(error_msg);
                    error=true;
                    return false;
                }
                string bot=*i1;
                if(!bot.empty())
                    add_attr_remote_call(bot);
                got=true;
            }
            if(!new_command.compare("#replication")) {
                if(args.end()==i1) {
                    error_msg="Incorrect syntax";
                    error_msg+=(string)"; "+debug;
                    error_msg=GET_STR_DEBUG_INFO_EX(error_msg);
                    error=true;
                    return false;
                }
                string bot=*i1;
                i1++;
                if(args.end()==i1) {
                    error_msg="Incorrect syntax";
                    error_msg+=(string)"; "+debug;
                    error_msg=GET_STR_DEBUG_INFO_EX(error_msg);
                    error=true;
                    return false;
                }
                string type=*i1;
                c_replication r;
                r.partner=bot;
                r.type=REPL_INVALID;
                if(!type.compare("push"))
                    r.type=REPL_PUSH;
                if(!type.compare("pull"))
                    r.type=REPL_PULL;
                if(!type.compare("pushpull"))
                    r.type=REPL_PUSHPULL;
                if(REPL_INVALID!=r.type) {
                    add_attr(r);
                }
                got=true;
            }
            if(!new_command.compare("#member_of_group")) {
                if(args.end()==i1) {
                    error_msg="Incorrect syntax";
                    error_msg+=(string)"; "+debug;
                    error_msg=GET_STR_DEBUG_INFO_EX(error_msg);
                    error=true;
                    return false;
                }
                string gr=*i1;
                if(!gr.empty())
                    add_attr_memeber_of_group(gr);
                got=true;
            }
            if(got) {
                command.clear();
                need_more=true;
                return true;
            } else {
                error_msg="Incorrect syntax";
                error_msg+=(string)"; "+debug;
                error_msg=GET_STR_DEBUG_INFO_EX(error_msg);
                error=true;
                return false;
            }
        }
        while(!name.empty() && !new_command.empty()) {
            if(new_command.find("function ")==0) {
                string::size_type i1=new_command.find(')');
                new_command.erase(0,i1+1);

                while(!new_command.empty() && '\x20'==new_command[0])
                    new_command.erase(0,1);

                if(new_command.empty()) {
                    command=new_command;
                    need_more=true;
                    return true;
                }
            }

            while(!new_command.empty() && '\x20'==new_command[0])
                new_command.erase(0,1);
            while(!new_command.empty() && '\x20'==new_command[new_command.length()-1])
                new_command.erase(new_command.length()-1,1);

            got=true;
            c_command::e_type type;
            bool semicolon=false;
            string sub_exp;
            string cmd=explode_command(new_command,error,need_more,type,semicolon,sub_exp);
            if(error) {
                error_msg="Incorrect syntax";
                error_msg+=(string)"; "+debug;
                error_msg=GET_STR_DEBUG_INFO_EX(error_msg);
                error=true;
                return false;
            }
            if(need_more) {
                if(new_command.find("function ")==0) {
                    string::size_type i1=new_command.find(')');
                    new_command.erase(0,i1+1);
                }
                command=new_command;
                return true;
            }

            if(cmd.empty() && c_command::CMD_EXP==type)
                continue;

            bool got_cmd=false;

            if(block_stack.size() && block_stack.back().next_resolve_if) {
                resolve_if(error,error_msg,debug);
                if(error)
                    return false;
            }

            bool next_resolve_if=false;

            if(was_right_brace && cmd.compare("else;") && !if_bunch.empty()) {
                resolve_if(error,error_msg,debug);
                if(error)
                    return false;
            } else if(!was_right_brace && cmd.compare("else;") && !cmd.empty() && !if_bunch.empty() && !block_stack.empty() && block_stack.back().was_inline) {
                next_resolve_if=true;
            }

            was_right_brace=false;

            if(commands.empty()) {
                got_cmd=true;
            }

            if(block_stack.size() && c_block::BLOCK_IF==block_stack.back()._type) {
                if(cmd.compare("else;") && c_command::CMD_LEFT_BRACKET!=type) { // != "else;"
                    if(!block_stack.back().ignore_next) {
                        resolve_if(error,error_msg,debug);
                        if(error)
                            return false;
                    }
                    block_stack.back().ignore_next=false;
                }
            }

            for(;;) {
                bool stop=false;
                if(!block_stack.empty() && block_stack.back().was_inline) {
                    switch(block_stack.back()._type) {
                        case c_block::BLOCK_FOR:
                            for_epilog(error,error_msg,debug);
                            if(error)
                                return false;
                            break;
                        case c_block::BLOCK_WHILE:
                            while_epilog(error,error_msg,debug,true);
                            if(error)
                                return false;
                            break;
                        case c_block::BLOCK_FOREACH:
                            foreach_epilog(error,error_msg,debug,true);
                            if(error)
                                return false;
                            break;
                        case c_block::BLOCK_DO:
                            // 'do' epilog resolved elsewhere
                            break;
                        case c_block::BLOCK_IF:
                            if(cmd.compare("else;") && c_command::CMD_LEFT_BRACKET!=type) { // != "else;"
                                block_stack.pop_back();
                            } else
                                stop=true;
                            break;
                        default:
                            stop=true;
                            break;
                    }
                } else
                    break;
                if(stop)
                    break;
            }

            if(!block_stack.empty() &&
                (
                c_block::BLOCK_FOR==block_stack.back()._type ||
                c_block::BLOCK_WHILE==block_stack.back()._type ||
                c_block::BLOCK_FOREACH==block_stack.back()._type ||
                c_block::BLOCK_IF==block_stack.back()._type
                ) && type!=c_command::CMD_LEFT_BRACKET) {
                    block_stack.back().was_inline=true;
              }

            switch(parse_else(cmd,got_cmd,error,error_msg,debug)) {
                case PARSE_FALSE:
                    return false;
                case PARSE_CONT:
                    continue;
                case PARSE_FALL_THRU:
                    break;
                case PARSE_TRUE:
                    command=new_command;
                    return true;
            }

            switch(parse_break_continue(cmd,got_cmd,error,error_msg,debug)) {
                case PARSE_FALSE:
                    return false;
                case PARSE_CONT:
                    continue;
                case PARSE_FALL_THRU:
                    break;
                case PARSE_TRUE:
                    command=new_command;
                    return true;
            }

            if(cmd.find("goto ")==0) {
                got_cmd=true;
                cmd.erase(0,strlen("goto "));

                if(cmd.length() && ';'==cmd[cmd.length()-1])
                    cmd.erase(cmd.length()-1,1);

                c_fn_cmd_goto* cmd_=MY_NEW c_fn_cmd_goto(cmd,debug,new_id);
                commands.insert(make_pair(new_id,cmd_));
                new_id++;

                continue;
            }
            if(cmd.find("global ")==0) {
                got_cmd=true;
                cmd.erase(0,strlen("global "));
                if(cmd.length())
                    cmd.erase(cmd.length()-1,1); // trailing ';'

                vector<string> globals;
                {
                    vector<string> args;
                    explode(cmd,',',args,false);
                    vector<string>::iterator i1;
                    for(i1=args.begin(); i1!=args.end(); ++i1) {
                        string arg=*i1;
                        while(arg.length() && 0x20==arg[arg.length()-1])
                            arg.erase(arg.length()-1,1);
                        while(arg.length() && 0x20==arg[0])
                            arg.erase(0,1);
                        if(arg.empty())
                            continue;
                        globals.push_back(arg);
                    }
                }

                c_fn_cmd_global* cmd_=MY_NEW c_fn_cmd_global(globals,debug,new_id);
                commands.insert(make_pair(new_id,cmd_));
                new_id++;

                continue;
            }
        do_again:
            if(cmd.find("do ")==0) {
                got_cmd=true;
                cmd.erase(0,strlen("do "));

                c_fn_cmd_do* cmd_=MY_NEW c_fn_cmd_do(debug,new_id);
                commands.insert(make_pair(new_id,cmd_));

                c_block blck(c_block::BLOCK_DO,new_id,debug);
                blck.do_id=new_id;
                block_stack.push_back(blck);

                new_id++;

                if(!cmd.compare(";"))
                    continue;
                goto do_again;
            }
            if(c_command::CMD_FOR==type) {
                got_cmd=true;
                if(cmd.find("for(",0)!=string::npos)
                    cmd.erase(cmd.find("for(",0),strlen("for("));
                if(cmd.length() && ')'==cmd[cmd.length()-1])
                    cmd.erase(cmd.length()-1,1);
                vector<string> args;
                cmd+=" "; // because we want an empty string as 3rd element in vector
                          //     in case of an empty iterative expression
                explode(cmd,';',args,true);
                if(args.size()!=3) {
                    error=true;
                    error_msg="Incorrect syntax";
                    error_msg+=(string)"; "+debug;
                    error_msg=GET_STR_DEBUG_INFO_EX(error_msg);
                    return false;
                }
                vsi i1=args.begin();
                c_command c;
                c.type=c_command::CMD_FOR;
                string exp=*i1;
                i1++;
                string condition=*i1;
                i1++;
                string iterative_exp=*i1;

                c.exp=MY_NEW c_expression();
                c.exp->assign(this_module,exp,debug,*symbols,"","for() construct initialization expression ");
                if(c.exp->got_error()) {
                    c.exp->clean_up();
                    MY_DELETE(c.exp);
                    error=true;
                    error_msg="Compile error";
                    error_msg+=(string)"; "+debug;
                    error_msg=GET_STR_DEBUG_INFO_EX(error_msg);
                    return false;
                }

                while(!condition.empty() && ('\x20'==condition[0] || '\t'==condition[0]))
                    condition.erase(0,1);
                if(condition.empty()) {
                    condition="TRUE";
                    var_oops(this_module,"Compile warning","for() construct conditional expression: an empty expression, assuming 'TRUE'",debug);
                }
                c.condition=MY_NEW c_expression();
                c.condition->assign(this_module,condition,debug,*symbols,"","for() construct conditional expression ");
                if(c.condition->got_error()) {
                    c.exp->clean_up();
                    MY_DELETE(c.exp);
                    c.condition->clean_up();
                    MY_DELETE(c.condition);
                    error=true;
                    error_msg="Compile error";
                    error_msg+=(string)"; "+debug;
                    error_msg=GET_STR_DEBUG_INFO_EX(error_msg);
                    return false;
                }
                c.iterative_exp=MY_NEW c_expression();
                c.iterative_exp->assign(this_module,iterative_exp,debug,*symbols,"","for() construct iterative expression ");
                if(c.iterative_exp->got_error()) {
                    c.exp->clean_up();
                    MY_DELETE(c.exp);
                    c.condition->clean_up();
                    MY_DELETE(c.condition);
                    c.iterative_exp->clean_up();
                    MY_DELETE(c.iterative_exp);
                    error=true;
                    error_msg="Compile error";
                    error_msg+=(string)"; "+debug;
                    error_msg=GET_STR_DEBUG_INFO_EX(error_msg);
                    return false;
                }

                {
                    // init exp
                    c_fn_cmd_exp* exp=MY_NEW c_fn_cmd_exp(c.exp,debug,new_id);
                    commands.insert(make_pair(new_id,exp));
                    new_id++;
                }
                c_block blck(c_block::BLOCK_FOR,new_id,debug);
                {
                    // cond exp
                    c_fn_cmd_for_cond* exp=MY_NEW c_fn_cmd_for_cond(c.condition,debug,new_id);
                    commands.insert(make_pair(new_id,exp));
                    blck.for_iter_jump=new_id;
                    new_id++;
                }
                blck.for_exp_iter=c.iterative_exp;
                block_stack.push_back(blck);

                continue;
            }

            e_parse_result rc;
            rc=parse_foreach_1(cmd,got_cmd,error,error_msg,debug,type,this_module);
            switch(rc) {
                case PARSE_FALSE:
                    return false;
                case PARSE_CONT:
                    continue;
                case PARSE_FALL_THRU:
                    break;
                case PARSE_TRUE:
                    command=new_command;
                    return true;
            }

            bool erase_trailing_semicolon;
            rc=parse_commands1(this_module,cmd,debug,new_id,type,got_cmd,error,error_msg,sub_exp,erase_trailing_semicolon);
            if(erase_trailing_semicolon && !new_command.empty() && ';'==new_command[new_command.length()-1])
                new_command.erase(0,1);
            switch(rc) {
                case PARSE_FALSE:
                    return false;
                case PARSE_CONT:
                    continue;
                case PARSE_FALL_THRU:
                    break;
                case PARSE_TRUE:
                    command=new_command;
                    return true;
            }

            rc=parse_commands2(this_module,cmd,debug,new_id,type,got_cmd,error,error_msg,semicolon,next_resolve_if,need_more,debug_prefix);
            switch(rc) {
                case PARSE_FALSE:
                    return false;
                case PARSE_CONT:
                    continue;
                case PARSE_FALL_THRU:
                    break;
                case PARSE_TRUE:
                    command=new_command;
                    return true;
            }

            if(!got_cmd) {
                error=true;
                error_msg="Invalid syntax";
                error_msg+=(string)"; "+debug;
                error_msg=GET_STR_DEBUG_INFO_EX(error_msg);
                return false;
            }
        }
        if(!got) {
            error=true;
            error_msg="Invalid syntax";
            error_msg+=(string)"; "+debug;
            error_msg=GET_STR_DEBUG_INFO_EX(error_msg);
            return false;
        }
    }
    command=new_command;
    need_more=true;
    return true;
}

bool c_function::get_replication(const string& partner, int type)
{
    vector<c_replication>::iterator i1;
    for(i1=replications.begin(); i1!=replications.end(); ++i1) {
        if(!partner.compare((*i1).partner)) {

            if(type==REPL_PUSH && (*i1).type==REPL_PUSH)
                return true;
            if(type==REPL_PULL && (*i1).type==REPL_PULL)
                return true;
            if(type==REPL_PUSH && (*i1).type==REPL_PUSHPULL)
                return true;
            if(type==REPL_PULL && (*i1).type==REPL_PUSHPULL)
                return true;

            if(type==REPL_PUSHPULL && (*i1).type==REPL_PUSH)
                return true;
            if(type==REPL_PUSHPULL && (*i1).type==REPL_PULL)
                return true;
            if(type==REPL_PUSHPULL && (*i1).type==REPL_PUSHPULL)
                return true;
        }
    }
    return false;
}

void c_function::dump_to_file(FILE* f)
{
    fprintf(f,"%s",dump.c_str());
}

string c_function::dump_to_string()
{
    return dump;
}

void c_function::add_attr(const c_replication& replication)
{
    replications.push_back(replication);
}

void c_function::add_attr_remote_call(const string& remote_bot)
{
    remote_bot_call.push_back(remote_bot);
}

void c_function::add_attr_memeber_of_group(const string& group)
{
    groups.push_back(group);
}

void c_function::get_prototype(vector<c_func_arg>& args_)
{
    args_=args;
}

c_function& c_function::operator=(const c_function& orig)
{
    //clear(false);

    name=orig.name;
    num_args=orig.num_args;
    args=orig.args;
    version=orig.version;
    line=orig.line;
    blocks=orig.blocks;
    done=orig.done;

    commands.clear();
    map<long,c_fn_cmd*>::const_iterator i1;
    for(i1=orig.commands.begin(); i1!=orig.commands.end(); ++i1)
        commands.insert(make_pair((*i1).first,(*i1).second->copy()));

    in_old_smtp=orig.in_old_smtp;
    cmd_smtp=orig.cmd_smtp;

    dump=orig.dump;

    minimal_botnet_version=orig.minimal_botnet_version;

    in_comment=orig.in_comment;

    new_id=orig.new_id;
    block_stack=orig.block_stack;
    if_bunch=orig.if_bunch;
    if_bunch_jump_over=orig.if_bunch_jump_over;
    groups=orig.groups;
    replications=orig.replications;
    remote_bot_call=orig.remote_bot_call;
    declaration=orig.declaration;

    last_changed=orig.last_changed;

    lock_cnt=orig.lock_cnt;

    was_right_brace=orig.was_right_brace;

    symbols=NULL;

    if(!orig.attached_symbol_table) {
        if(orig.symbol_table)
            symbol_table=MY_NEW c_micro_symbol_table(*orig.symbol_table);
        else
            symbol_table=NULL;

        if(symbol_table)
            symbols=symbol_table->var_mapping_;
        else
            symbols=NULL;
    } else {
        symbol_table=orig.symbol_table;
        symbols=orig.symbols;
    }

    attached_symbol_table=orig.attached_symbol_table;

    return *this;
}

c_function::~c_function()
{
    map<long,c_fn_cmd*>::iterator i1;
    for(i1=commands.begin(); i1!=commands.end(); ++i1)
        MY_DELETE((*i1).second);
    commands.clear();

    if(!attached_symbol_table) {
        /*
        MY_DELETE(symbols);
        symbols=NULL;
        */
        MY_DELETE(symbol_table);
    }
    symbol_table=NULL;
}

c_function::c_function(const c_function& orig)
{
    name=orig.name;
    num_args=orig.num_args;
    args=orig.args;
    version=orig.version;
    line=orig.line;
    blocks=orig.blocks;
    done=orig.done;

    commands.clear();
    map<long,c_fn_cmd*>::const_iterator i1;
    for(i1=orig.commands.begin(); i1!=orig.commands.end(); ++i1)
        commands.insert(make_pair((*i1).first,(*i1).second->copy()));

    in_old_smtp=orig.in_old_smtp;
    cmd_smtp=orig.cmd_smtp;

    dump=orig.dump;

    minimal_botnet_version=orig.minimal_botnet_version;

    in_comment=orig.in_comment;

    new_id=orig.new_id;
    block_stack=orig.block_stack;
    if_bunch=orig.if_bunch;
    if_bunch_jump_over=orig.if_bunch_jump_over;
    groups=orig.groups;
    replications=orig.replications;
    remote_bot_call=orig.remote_bot_call;
    declaration=orig.declaration;

    last_changed=orig.last_changed;

    lock_cnt=orig.lock_cnt;

    was_right_brace=orig.was_right_brace;

    symbols=NULL;

    if(!orig.attached_symbol_table) {
        if(orig.symbol_table)
            symbol_table=MY_NEW c_micro_symbol_table(*orig.symbol_table);
        else
            make_symbol_table(NULL);

        if(symbol_table)
            symbols=symbol_table->var_mapping_;
        else
            symbols=NULL;
    } else {
        symbol_table=orig.symbol_table;
        symbols=orig.symbols;
    }

    attached_symbol_table=orig.attached_symbol_table;
}

void c_function::clear(bool create_symbols)
{
    attached_symbol_table=false;
    symbol_table=NULL;
    name="";
    num_args=0;
    args.clear();
    map<long,c_fn_cmd*>::iterator i1;
    for(i1=commands.begin(); i1!=commands.end(); ++i1)
        MY_DELETE((*i1).second);
    commands.clear();
    version=1;
    line=1;
    blocks=0;
    done=false;
    replications.clear();
    remote_bot_call.clear();
    last_changed=0;
    groups.clear();
    in_old_smtp=0;
    dump="";
    minimal_botnet_version=1001;
    in_comment=false;
    lock_cnt=0;
    block_stack.clear();
    new_id=1;
    if_bunch.clear();
    if_bunch_jump_over=JUMP_ID_INVALID;
    was_right_brace=false;

    if(symbols)
        MY_DELETE(symbols);
    if(create_symbols) {
        symbols=MY_NEW c_micro_symbol_variable_mapping();
    } else {
        symbols=NULL;
    }
    declaration="";
}

c_function::c_function()
{
    symbols=NULL;
    symbol_table=NULL;
    clear(false);
}

void c_function::reset()
{
    symbol_table=NULL;
    name="";
    num_args=0;
    args.clear();
    map<long,c_fn_cmd*>::iterator i1;
    for(i1=commands.begin(); i1!=commands.end(); ++i1)
        MY_DELETE((*i1).second);
    commands.clear();
    //version=1; // this NOT!
    //line=1; // and this TOO!
    blocks=0;
    done=false;
    replications.clear();
    remote_bot_call.clear();
    last_changed=0;
    groups.clear();
    in_old_smtp=0;
    dump="";
    minimal_botnet_version=1001;
    in_comment=false;
    lock_cnt=0;
    block_stack.clear();
    new_id=1;
    if_bunch.clear();
    if_bunch_jump_over=JUMP_ID_INVALID;
    was_right_brace=false;

    if(symbols)
        MY_DELETE(symbols);
    symbols=MY_NEW c_micro_symbol_variable_mapping();
    declaration="";

    attached_symbol_table=false;
}

void c_function::destroy_symbols()
{
    if(!attached_symbol_table) {
        MY_DELETE(symbols);
        symbols=NULL;
        if(symbol_table)
            symbol_table->var_mapping_=NULL;
    } else {
        symbols=NULL;
    }
}

void c_function::destroy(bool also_symbols, bool also_cleanup_code)
{
    map<long,c_fn_cmd*>::iterator i1;
    for(i1=commands.begin(); i1!=commands.end(); ++i1) {
        if(also_cleanup_code)
            (*i1).second->clean_up();
        MY_DELETE((*i1).second);
    }
    commands.clear();
    if(also_symbols) {
        destroy_symbols();
        MY_DELETE(symbols);
        if(symbol_table && !attached_symbol_table)
            symbol_table->var_mapping_=NULL;
    }
    symbols=NULL;
    declaration="";
}

c_function::c_function(c_micro_symbol_table* symbol_table_to_attach)
{
    symbols=NULL;
    symbol_table=NULL;
    clear(false);

    symbol_table=symbol_table_to_attach;
    symbols=symbol_table->var_mapping_;
    attached_symbol_table=true;
}

void c_function::optimize_output()
{
    vector<var_sym_index_t> inputs;

    map<long,c_fn_cmd*>::iterator i1;
    for(i1=commands.begin(); i1!=commands.end(); ++i1) {
        (*i1).second->add_inputs(inputs);
    }

    // add mapped variables
    map<var_sym_index_t, string>::iterator i2;
    for(i2=symbols->mapping.begin(); i2!=symbols->mapping.end(); ++i2) {
        if(!(*i2).second.empty())
            inputs.push_back((*i2).first);
    }

    for(i1=commands.begin(); i1!=commands.end(); ++i1) {
        (*i1).second->optimize_output(inputs);
    }
}

c_micro_symbol_table* c_function::get_symbol_table()
{
    return symbol_table;
}

c_function::e_parse_result c_function::parse_commands1(c_module* this_module, string& cmd, const string& debug, long& new_id, c_command::e_type type, bool& got_cmd, bool& error, string& error_msg, string& sub_exp, bool &erase_trailing_semicolon)
{
    erase_trailing_semicolon=false;

    if(c_command::CMD_LABEL==type) {
        got_cmd=true;
        c_fn_cmd_label* cmd_=MY_NEW c_fn_cmd_label(cmd,debug,new_id);
        commands.insert(make_pair(new_id,cmd_));
        new_id++;
        return PARSE_CONT;
    }
    if(c_command::CMD_SWITCH==type) {
        got_cmd=true;
        c_command c;
        c.type=c_command::CMD_SWITCH;
        if(cmd.find("switch")==0)
            cmd.erase(0,strlen("switch"));
        if(cmd.length() && ';'==cmd[cmd.length()-1])
            cmd.erase(cmd.length()-1,1);
        c.exp=MY_NEW c_expression();
        c.exp->assign(this_module,cmd,debug,*symbols,"","");
        if(c.exp->got_error()) {
            c.exp->clean_up();
            MY_DELETE(c.exp);
            error=true;
            error_msg="Compile error";
            error_msg+=(string)"; "+debug;
            error_msg=GET_STR_DEBUG_INFO_EX(error_msg);
            return PARSE_FALSE;
        }
        c.debug=debug;

        {
            c_fn_cmd_switch* cmd=MY_NEW c_fn_cmd_switch(c.exp,debug,new_id);
            commands.insert(make_pair(new_id,cmd));

            c_block blck(c_block::BLOCK_SWITCH,new_id,debug);
            blck.switch_id=new_id;
            block_stack.push_back(blck);

            new_id++;
        }

        return PARSE_CONT;
    }
    if(c_command::CMD_CASE==type) {
        got_cmd=true;
        c_command c;
        c.type=c_command::CMD_CASE;
        if(cmd.find("case ")==0)
            cmd.erase(0,strlen("case "));
        bool default_case=false;
        if(cmd.find("default")==0) {
            default_case=true;
            cmd="";
        }
        if(!cmd.empty() && ('\x20'==cmd[0] || '\t'==cmd[0]))
            cmd.erase(0,1);
        if(!default_case) {
            c.exp=MY_NEW c_expression();
            c.exp->assign(this_module,cmd,debug,*symbols,"","");
            c.exp_=cmd;
            if(c.exp->got_error()) {
                c.exp->clean_up();
                MY_DELETE(c.exp);
                error=true;
                error_msg="Compile error";
                error_msg+=(string)"; "+debug;
                error_msg=GET_STR_DEBUG_INFO_EX(error_msg);
                return PARSE_FALSE;
            }
        } else {
            c.exp=NULL;
        }

        // add new case
        if(!cmd.empty() && !default_case) {
            // case xxx:
            c_fn_cmd_case* cmd=MY_NEW c_fn_cmd_case(c.exp,debug,new_id);
            commands.insert(make_pair(new_id,cmd));
        } else {
            // default:
            c_fn_cmd_case_default* cmd=MY_NEW c_fn_cmd_case_default(debug,new_id);
            commands.insert(make_pair(new_id,cmd));
        }

        // find last switch() ID
        bool got=false;
        deque<c_block>::reverse_iterator i2;
        for(i2=block_stack.rbegin(); i2!=block_stack.rend(); ++i2) {
            if(c_block::BLOCK_SWITCH==(*i2)._type) {
                got=true;
                break;
            }
        }
        if(!got) {
            error=true;
            error_msg="Syntax error (case statement)";
            error_msg+=(string)"; "+debug;
            error_msg=GET_STR_DEBUG_INFO_EX(error_msg);
            return PARSE_FALSE;
        }
        long id=(*i2).switch_id;
        //

        map<long,c_fn_cmd*>::iterator i1=commands.find(id);
        if(i1!=commands.end()) {
            c_fn_cmd_switch* c=NULL;
            if((*i1).second && typeid(*(*i1).second)==typeid(c_fn_cmd_switch))
                c=dynamic_cast<c_fn_cmd_switch*>((*i1).second);
            if(c) {
                c->push_case_id(new_id);
            } else {
                error=true;
                error_msg="Internal compile error (dynamic_cast<c_fn_cmd_switch*> == NULL";
                error_msg+=(string)", should be "+typeid(*(*i1).second).name()+")";
                error_msg+=(string)"; "+debug;
                error_msg=GET_STR_DEBUG_INFO_EX(error_msg);
                return PARSE_FALSE;
            }
        } else {
            error=true;
            error_msg="Internal compile error (not found matching 'switch')";
            error_msg+=(string)"; "+debug;
            error_msg=GET_STR_DEBUG_INFO_EX(error_msg);
            return PARSE_FALSE;
        }
        new_id++;
        return PARSE_CONT;
    }
    if(c_command::CMD_EXP_SUB==type) {
        got_cmd=true;
        c_command c;
        c.type=c_command::CMD_EXP_SUB;
        if(sub_exp.length())
            sub_exp.erase(sub_exp.length()-1,1);
        c.exp=MY_NEW c_expression();
        c.exp->assign(this_module,cmd,debug,*symbols,sub_exp,"");
        if(c.exp->got_error()) {
            c.exp->clean_up();
            MY_DELETE(c.exp);
            error=true;
            error_msg="Compile error";
            error_msg+=(string)"; "+debug;
            error_msg=GET_STR_DEBUG_INFO_EX(error_msg);
            return PARSE_FALSE;
        }
        c.sub_=sub_exp;
        c.sub=MY_NEW c_expression();
        c.sub->assign(this_module,sub_exp,debug,*symbols,"","");
        if(c.sub->got_error()) {
            c.exp->clean_up();
            MY_DELETE(c.exp);
            c.sub->clean_up();
            MY_DELETE(c.sub);
            error=true;
            error_msg="Compile error";
            error_msg+=(string)"; "+debug;
            error_msg=GET_STR_DEBUG_INFO_EX(error_msg);
            return PARSE_FALSE;
        }
        c.sub->clean_up();
        MY_DELETE(c.sub);
        c.sub=NULL;
        c.debug=debug;

        {
            // common exp
            c_fn_cmd_exp* cmd=MY_NEW c_fn_cmd_exp(c.exp,c.sub_,debug,new_id);
            commands.insert(make_pair(new_id,cmd));
            new_id++;
        }

        return PARSE_CONT;
    }

    if(c_command::CMD_WHILE==type) {
        got_cmd=true;
        c_command c;
        c.type=c_command::CMD_WHILE;
        if(cmd.find("while")==0)
            cmd.erase(0,strlen("while"));
        if(cmd.length() && ';'==cmd[cmd.length()-1]) {
            cmd.erase(cmd.length()-1,1);
            erase_trailing_semicolon=true;
        }
        c.exp=MY_NEW c_expression();
        c.exp->assign(this_module,cmd,debug,*symbols,"","");
        if(c.exp->got_error()) {
            c.exp->clean_up();
            MY_DELETE(c.exp);
            error=true;
            error_msg="Compile error";
            error_msg+=(string)"; "+debug;
            error_msg=GET_STR_DEBUG_INFO_EX(error_msg);
            return PARSE_FALSE;
        }

        if(block_stack.empty() || c_block::BLOCK_DO!=block_stack.back()._type) {
            // it is "while(...) { ... }"
            c_fn_cmd_while* cmd=MY_NEW c_fn_cmd_while(c.exp,debug,new_id);
            commands.insert(make_pair(new_id,cmd));

            c_block blck(c_block::BLOCK_WHILE,new_id,debug);
            blck.while_again_id=new_id;
            block_stack.push_back(blck);

            new_id++;
        } else {
            // it is "do { ... } while(...)"

            // set up end of loop
            map<long,c_fn_cmd*>::iterator i1=commands.find(block_stack.back().do_id);
            if(i1!=commands.end()) {
                c_fn_cmd_do* cmd=NULL;
                if((*i1).second && typeid(*(*i1).second)==typeid(c_fn_cmd_do))
                    cmd=dynamic_cast<c_fn_cmd_do*>((*i1).second);
                if(cmd) {
                    c_fn_cmd_do_while* dw=MY_NEW c_fn_cmd_do_while(c.exp,block_stack.back().do_id,debug,new_id);
                    cmd->set_end_loop_id(new_id+1);
                    commands.insert(make_pair(new_id,dw));
                    new_id++;
                    c_fn_cmd_dummy* dummy=MY_NEW c_fn_cmd_dummy(new_id);
                    commands.insert(make_pair(new_id,dummy));
                    new_id++;
                } else {
                    error=true;
                    error_msg="Internal compile error (dynamic_cast<c_fn_cmd_do*> == NULL";
                    error_msg+=(string)", should be "+typeid(*(*i1).second).name()+")";
                    error_msg+=(string)"; "+debug;
                    error_msg=GET_STR_DEBUG_INFO_EX(error_msg);
                    return PARSE_FALSE;
                }
            } else {
                error=true;
                error_msg="Internal compile error (not found matching 'do')";
                error_msg+=(string)"; "+debug;
                error_msg=GET_STR_DEBUG_INFO_EX(error_msg);
                return PARSE_FALSE;
            }
            do_epilog(error,error_msg,debug,true);
            if(error)
                return PARSE_FALSE;
        }

        return PARSE_CONT;
    }

    return PARSE_FALL_THRU;
}

#undef CAST
#undef CAST2
#undef CAST3

#define CAST(TYPE) \
    TYPE* c=NULL; \
    if((*i1).second && typeid(*(*i1).second)==typeid(TYPE)) \
        c=dynamic_cast<TYPE*>((*i1).second); \
    if(!c) { \
        error=true; \
        error_msg=(string)"Internal error: dynamic_cast<" + #TYPE + "*> == NULL, " \
        "should be " + typeid(*(*i1).second).name() + "; "; \
        error_msg+=(*i1).second->get_debug_string(); \
        error_msg=GET_STR_DEBUG_INFO_EX(error_msg); \
        return PARSE_FALSE; \
    }

#define CAST2(TYPE) \
    TYPE* c2=NULL; \
    if((*i2).second && typeid(*(*i2).second)==typeid(TYPE)) \
        c2=dynamic_cast<TYPE*>((*i2).second); \
    if(!c2) { \
        error=true; \
        error_msg=(string)"Internal error: dynamic_cast<" + #TYPE + "*> == NULL, " \
        "should be " + typeid(*(*i2).second).name() + "; "; \
        error_msg+=(*i2).second->get_debug_string(); \
        error_msg=GET_STR_DEBUG_INFO_EX(error_msg); \
        return PARSE_FALSE; \
    }

#define CAST3(TYPE) \
    TYPE* c2=NULL; \
    if((*i3).second && typeid(*(*i3).second)==typeid(TYPE)) \
        c2=dynamic_cast<TYPE*>((*i3).second); \
    if(!c2) { \
        error=true; \
        error_msg=(string)"Internal error: dynamic_cast<" + #TYPE + "*> == NULL, " \
        "should be " + typeid(*(*i3).second).name() + "; "; \
        error_msg+=(*i3).second->get_debug_string(); \
        error_msg=GET_STR_DEBUG_INFO_EX(error_msg); \
        return PARSE_FALSE; \
    }

c_function::e_parse_result c_function::parse_commands2(c_module* this_module, string& cmd, const string& debug, long& new_id, c_command::e_type type, bool& got_cmd, bool& error, string& error_msg, bool semicolon, bool next_resolve_if, bool& need_more, const string& debug_prefix)
{
    if(!cmd.empty() && ';'!=cmd[cmd.length()-1] && c_command::CMD_LEFT_BRACKET==type) {
        // nothing
    } else if(!cmd.empty() && ';'==cmd[cmd.length()-1] && c_command::CMD_EXP==type) {
        got_cmd=true;
        //cmd.erase(cmd.length()-1,1);
        c_command c;
        c.type=type;
        c.exp=MY_NEW c_expression();
        c.exp->assign(this_module,cmd,debug,*symbols,"",debug_prefix);
        if(c.exp->got_error()) {
            c.exp->clean_up();
            MY_DELETE(c.exp);
            error=true;
            error_msg="Compile error";
            error_msg+=(string)"; "+debug;
            error_msg=GET_STR_DEBUG_INFO_EX(error_msg);
            return PARSE_FALSE;
        }
        c.debug=debug;
        c.semicolon=semicolon;

        {
            // common exp
            c_fn_cmd_exp* cmd=MY_NEW c_fn_cmd_exp(c.exp,debug,new_id);
            commands.insert(make_pair(new_id,cmd));
            new_id++;
        }

        if(next_resolve_if)
            block_stack.back().next_resolve_if=true;

        return PARSE_CONT;
    } else if(cmd.length() && ';'==cmd[cmd.length()-1] && c_command::CMD_IF_EXP==type) {
        got_cmd=true;
        cmd.erase(cmd.length()-1,1);
        c_command c;
        c.type=type;
        c.exp=MY_NEW c_expression();
        c.exp->assign(this_module,cmd,debug,*symbols,"","");
        if(c.exp->got_error()) {
            c.exp->clean_up();
            MY_DELETE(c.exp);
            error=true;
            error_msg="Compile error";
            error_msg+=(string)"; "+debug;
            error_msg=GET_STR_DEBUG_INFO_EX(error_msg);
            return PARSE_FALSE;
        }
        c.debug=debug;
        c.semicolon=semicolon;

        {
            c_if_branch b;
            b.if_id=new_id;
            if_bunch.push_back(b);

            // if exp
            c_fn_cmd_if* cmd=MY_NEW c_fn_cmd_if(c.exp,debug,new_id);
            commands.insert(make_pair(new_id,cmd));

            c_block blck(c_block::BLOCK_IF,new_id,debug);
            blck.if_id=new_id;
            blck.ignore_next=true;
            block_stack.push_back(blck);

            new_id++;
        }

        return PARSE_CONT;
    } else if(cmd.length()==0 && c_command::CMD_LEFT_BRACKET==type) {
        got_cmd=true;
        c_block blck(c_block::BLOCK_LEFT_BRACE,-1,"");

        {
            blck.if_bunch=if_bunch;
            blck.if_bunch_jump_over=if_bunch_jump_over;
            if_bunch.clear();
            if_bunch_jump_over=JUMP_ID_INVALID;
        }

        block_stack.push_back(blck);
        blocks++;
    } else if(cmd.length()==0 && c_command::CMD_RIGHT_BRACKET==type) {
        got_cmd=true;
        was_right_brace=true;
        blocks--;
        if(blocks<0 || block_stack.empty()) {
            error=true;
            error_msg="Invalid syntax";
            error_msg+=(string)"; "+debug;
            error_msg=GET_STR_DEBUG_INFO_EX(error_msg);
            return PARSE_FALSE;
        }
        {
            c_fn_cmd_right_brace* cmd=MY_NEW c_fn_cmd_right_brace("}",new_id);
            commands.insert(make_pair(new_id,cmd));
            new_id++;

            while(block_stack.size()) {
                resolve_if(error,error_msg,debug);
                if(error)
                    return PARSE_FALSE;
                if_bunch=block_stack.back().if_bunch;
                if_bunch_jump_over=block_stack.back().if_bunch_jump_over;
                block_stack.back().if_bunch.clear();
                block_stack.back().if_bunch_jump_over=JUMP_ID_INVALID;

                if(c_block::BLOCK_LEFT_BRACE==block_stack.back()._type) {
                    block_stack.pop_back();
                    break;
                } else if(c_block::BLOCK_IF==block_stack.back()._type) {
                    resolve_if(error,error_msg,debug);
                }
                if(error)
                    return PARSE_FALSE;
                size_t old_size=block_stack.size();
                if(!all_epilogs(error,error_msg,debug))
                    return PARSE_FALSE;
                if(block_stack.size()==old_size)
                    block_stack.pop_back();
            }
            if(!all_epilogs(error,error_msg,debug))
                return PARSE_FALSE;
        }

        if(error)
            return PARSE_FALSE;
        if(0!=blocks)
            return PARSE_CONT;

        {
            c_fn_cmd_end_of_func* cmd=MY_NEW c_fn_cmd_end_of_func("END OF FUNCTION",new_id);
            commands.insert(make_pair(new_id,cmd));
            new_id++;
        }
        {
            // process goto's and labels, and check all id's for validity
            {
                // resolve goto's
                for(long id=1; ; ++id) {
                    map<long,c_fn_cmd*>::iterator i1=commands.find(id);
                    if(i1==commands.end())
                        break;
                    if((*i1).second && typeid(*(*i1).second)==typeid(c_fn_cmd_goto)) {
                        CAST(c_fn_cmd_goto);
                        if(c->get_label_id()==JUMP_ID_INVALID) {
                            for(long id2=1; ; ++id2) {
                                map<long,c_fn_cmd*>::iterator i2=commands.find(id2);
                                if(i2==commands.end())
                                    break;
                                if((*i2).second && typeid(*(*i2).second)==typeid(c_fn_cmd_label)) {
                                    CAST2(c_fn_cmd_label);
                                    if(!c->get_tag().compare(c2->get_tag())) {
                                        c->set_label_id(c2->get_id());
                                        break;
                                    }
                                    continue;
                                }
                            }
                        }
                        continue;
                    }
                }
            }

            for(long id=1; ; ++id) {
                map<long,c_fn_cmd*>::iterator i1=commands.find(id);
                if(i1==commands.end()) {
                    error=true;
                    error_msg=(string)"Internal error broken command chain, missing ID==";
                    error_msg+=tos(id);
                    error_msg=GET_STR_DEBUG_INFO_EX(error_msg);
                    return PARSE_FALSE;
                }
                if((*i1).second && typeid(*(*i1).second)==typeid(c_fn_cmd)) {
                    CAST(c_fn_cmd);
                    error=true;
                    error_msg=(string)"Internal error (c_fn_cmd): base class cannot be used; ";
                    error_msg+=(*i1).second->get_debug_string();
                    error_msg=GET_STR_DEBUG_INFO_EX(error_msg);
                    return PARSE_FALSE;
                }

                if((*i1).second && typeid(*(*i1).second)==typeid(c_fn_cmd_for_cond)) {
                    CAST(c_fn_cmd_for_cond);
                    if(c->sanity_check(error_msg))
                        DO_ERROR();
                    continue;
                }
                if((*i1).second && typeid(*(*i1).second)==typeid(c_fn_cmd_left_brace)) {
                    CAST(c_fn_cmd_left_brace);
                    if(c->sanity_check(error_msg))
                        DO_ERROR();
                    continue;
                }
                if((*i1).second && typeid(*(*i1).second)==typeid(c_fn_cmd_right_brace)) {
                    CAST(c_fn_cmd_right_brace);
                    if(c->sanity_check(error_msg))
                        DO_ERROR();
                    continue;
                }
                if((*i1).second && typeid(*(*i1).second)==typeid(c_fn_cmd_exp)) {
                    CAST(c_fn_cmd_exp);
                    if(c->sanity_check(error_msg))
                        DO_ERROR();
                    continue;
                }
                if((*i1).second && typeid(*(*i1).second)==typeid(c_fn_cmd_jump)) {
                    CAST(c_fn_cmd_jump);
                    if(c->sanity_check(error_msg))
                        DO_ERROR();
                    continue;
                }
                if((*i1).second && typeid(*(*i1).second)==typeid(c_fn_cmd_dummy)) {
                    CAST(c_fn_cmd_dummy);
                    if(c->sanity_check(error_msg))
                        DO_ERROR();
                    continue;
                }
                if((*i1).second && typeid(*(*i1).second)==typeid(c_fn_cmd_if)) {
                    CAST(c_fn_cmd_if);
                    if(c->sanity_check(error_msg))
                        DO_ERROR();
                    continue;
                }
                if((*i1).second && typeid(*(*i1).second)==typeid(c_fn_cmd_switch)) {
                    CAST(c_fn_cmd_switch);
                    if(c->sanity_check(error_msg))
                        DO_ERROR();
                    continue;
                }
                if((*i1).second && typeid(*(*i1).second)==typeid(c_fn_cmd_case)) {
                    CAST(c_fn_cmd_case);
                    if(c->sanity_check(error_msg))
                        DO_ERROR();
                    continue;
                }
                if((*i1).second && typeid(*(*i1).second)==typeid(c_fn_cmd_case_default)) {
                    CAST(c_fn_cmd_case_default);
                    if(c->sanity_check(error_msg))
                        DO_ERROR();
                    continue;
                }
                if((*i1).second && typeid(*(*i1).second)==typeid(c_fn_cmd_label)) {
                    CAST(c_fn_cmd_label);
                    if(c->sanity_check(error_msg))
                        DO_ERROR();
                    continue;
                }
                if((*i1).second && typeid(*(*i1).second)==typeid(c_fn_cmd_goto)) {
                    CAST(c_fn_cmd_goto);
                    if(c->sanity_check(error_msg))
                        DO_ERROR();
                    continue;
                }
                if((*i1).second && typeid(*(*i1).second)==typeid(c_fn_cmd_global)) {
                    CAST(c_fn_cmd_global);
                    if(c->sanity_check(error_msg))
                        DO_ERROR();
                    continue;
                }
                if((*i1).second && typeid(*(*i1).second)==typeid(c_fn_cmd_break)) {
                    CAST(c_fn_cmd_break);
                    if(c->sanity_check(error_msg))
                        DO_ERROR();
                    continue;
                }
                if((*i1).second && typeid(*(*i1).second)==typeid(c_fn_cmd_continue)) {
                    CAST(c_fn_cmd_continue);
                    if(c->sanity_check(error_msg))
                        DO_ERROR();
                    continue;
                }
                if((*i1).second && typeid(*(*i1).second)==typeid(c_fn_cmd_do)) {
                    CAST(c_fn_cmd_do);
                    if(c->sanity_check(error_msg))
                        DO_ERROR();
                    continue;
                }
                if((*i1).second && typeid(*(*i1).second)==typeid(c_fn_cmd_while)) {
                    CAST(c_fn_cmd_while);
                    if(c->sanity_check(error_msg))
                        DO_ERROR();
                    continue;
                }
                if((*i1).second && typeid(*(*i1).second)==typeid(c_fn_cmd_do_while)) {
                    CAST(c_fn_cmd_do_while);
                    if(c->sanity_check(error_msg))
                        DO_ERROR();
                    continue;
                }
                if((*i1).second && typeid(*(*i1).second)==typeid(c_fn_cmd_foreach)) {
                    CAST(c_fn_cmd_foreach);
                    if(c->sanity_check(error_msg))
                        DO_ERROR();
                    continue;
                }
                if((*i1).second && typeid(*(*i1).second)==typeid(c_fn_cmd_end_of_func)) {
                    CAST(c_fn_cmd_end_of_func);
                    if(c->sanity_check(error_msg))
                        DO_ERROR();
                    break;
                }
                if((*i1).second && typeid(*(*i1).second)==typeid(c_fn_cmd_foreach_next_iter)) {
                    CAST(c_fn_cmd_foreach_next_iter);
                    if(c->sanity_check(error_msg))
                        DO_ERROR();
                    break;
                }
                error_msg="Internal error: unimplemented class (";
                error_msg+=typeid(*(*i1).second).name();
                error_msg+="); ";
                error_msg+=(*i1).second->get_debug_string();
                DO_ERROR();
            }
        }
        if(0==blocks) {
            need_more=false;
            done=true;
            if(!attached_symbol_table) {
                MY_DELETE(symbol_table);
                symbol_table=NULL;
                make_symbol_table(NULL);
            }
            optimize_output();
            return PARSE_TRUE;
        }
    } else {
        error=true;
        error_msg="Invalid syntax";
        error_msg+=(string)"; "+debug;
        error_msg=GET_STR_DEBUG_INFO_EX(error_msg);
        return PARSE_FALSE;
    }

    return PARSE_FALL_THRU;
}

