/***************************************************************************
                          c_function.h  -  description
                             -------------------
    begin                : Thu May 10 2007
    copyright            : (C) 2004-2007 by Marian VooDooMan Meravy
    email                : ghostvoodooman -a.T= users (dot) sourceforge.net
 ***************************************************************************/

/* $Id: c_function.h,v 1.19 2008/11/24 12:09:23 ghostvoodooman Exp $ */

/***************************************************************************

VooDoo cIRCle - an IRC (ro)bot
Copyright (C) 2004-2008 by Marian VooDooMan Meravy
  (ghostvoodooman -a.T= users (dot) sourceforge.net)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************/

/*!
    \file
    \brief Class 'c_function'
    \version $Id: c_function.h,v 1.19 2008/11/24 12:09:23 ghostvoodooman Exp $
*/

#ifndef C_FUNCTION_H__
#define C_FUNCTION_H__

#if defined(HAVE_CONFIG_H)
#   include <config.h>
#endif

#include <string>
#include <vector>
#include <stack>
#include <list>
#include <map>
#include <algorithm>

#include "../utils/memleak.h"

#include "../params.h"

#include "../func/c_block.h"
#include "c_fn_cmd.h"
#include "c_func_arg.h"
#include "../botnet/c_replication.h"
#include "c_command.h"

class c_micro_symbol_table;
struct c_eval_iteration;

/*!
    \brief Stores function
    \author VooDooMan
    \version 1
    \date 2005
*/
class c_function
{
    enum e_parse_result
    {
        PARSE_FALSE,
        PARSE_CONT,
        PARSE_FALL_THRU,
        PARSE_TRUE,
    };
    e_parse_result parse_else(const string& cmd, bool& got_cmd, bool& error, string& error_msg, const string& debug);
    e_parse_result parse_break_continue(const string& cmd, bool& got_cmd, bool& error, string& error_msg, const string& debug);
    e_parse_result parse_foreach_1(string& cmd, bool& got_cmd, bool& error, string& error_msg, const string& debug, const c_command::e_type type, c_module* this_module);
    e_parse_result parse_commands1(c_module* this_module, string& cmd, const string& debug, long& new_id, c_command::e_type type, bool& got_cmd, bool& error, string& error_msg, string& sub_exp, bool &erase_trailing_semicolon);
    e_parse_result parse_commands2(c_module* this_module, string& cmd, const string& debug, long& new_id, c_command::e_type type, bool& got_cmd, bool& error, string& error_msg, bool semicolon, bool next_resolve_if, bool& need_more, const string& debug_prefix);

    string name;                    //!< Name of function, without arguments
    int num_args;                   //!< Number of arguments
    vector<c_func_arg> args;        //!< Arguments

    c_micro_symbol_variable_mapping* symbols;       //!< Variable names to index mapping table
    c_micro_symbol_table* symbol_table;
    bool attached_symbol_table;     //!< While 'true', we will not deallocate 'symbol_table', as it was attached

    friend class c_module;
    friend class c_micro_func_call;
    friend void logic_loop();
    friend bool debug_do_evaluate_tests();
    friend bool eval_function(c_module* this_module, const string& code_str, c_var_array* vars, c_functions& functions, const string& debug, stack<c_eval_iteration>& st);
    friend void eval_evaluate_proc(stack<c_eval_iteration>& st);
    friend class c_proc_command;

    int version;                    //!< Version of script
    int line;                       //!< Line while compilation of function
    int blocks;                     //!< Number of blocks while compilation of function
    bool done;                      //!< true if compile ended

    int in_old_smtp;                //!< Used for conversion from procedure (SMTP command)
    string cmd_smtp;                //!< Used for conversion from procedure (SMTP command)

    string dump;                    //!< Dump of function

    int minimal_botnet_version;     //!< Minimal BOTNET version

    bool in_comment;                //!< true if we are compiling/ignoring comment

    long new_id;

    bool was_right_brace;

    /*!
        \brief Sets declaration of function
        \author VooDooMan
        \version 1
        \date 2005
        \param decl Function declaration
        \param error Returns error string
        \return Returns false for error
    */
    bool set(string decl, string& error);

    /*!
        \brief Resets for next function
        \author VooDooMan
        \version 1
        \date 2005
    */
    void reset();

    deque<c_block> block_stack;

    vector<c_if_branch> if_bunch;
    long if_bunch_jump_over;

    bool foreach_epilog(bool& error, string& error_msg, const string& debug, bool force = false);
    bool do_epilog(bool& error, string& error_msg, const string& debug, bool force = false);
    bool while_epilog(bool& error, string& error_msg, const string& debug, bool force = false);
    void resolve_if(bool& error, string& error_msg, const string& debug);
    bool for_epilog(bool& error, string& error_msg, const string& debug);
    bool switch_epilog(bool& error, string& error_msg, const string& debug);
    bool all_epilogs(bool& error, string& error_msg, const string& debug);

public:
    vector<string> groups;                  //!< Group membership of function
    vector<c_replication> replications;     //!< Replication rules
    vector<string> remote_bot_call;         //!< Remote bot call rules
    string declaration;                     //!< Full declaration of function

    map<long,c_fn_cmd*> commands;           //!< List of commands

    my_time_t last_changed;                    //!< Last changed attribute

    int lock_cnt;                           //!< If we are executing this function

    /*!
        \brief Cleans up
        \author VooDooMan
        \version 2
        \date 2005, 2006-05-05
        \param create_symbols 'true' if we want to allocate symbol table
    */
    void clear(bool create_symbols = true);

    /*!
        \brief Creates a new symbol table
        \author VooDooMan
        \version 1
        \date 2006-05-05
    */
    void create_symbols();

    /*!
        \brief Constructor
        \author VooDooMan
        \version 1
        \date 2006
    */
    c_function();

    /*!
        \brief Constructor, while "attaches" already prepared symbol table
        \author VooDooMan
        \version 1
        \date 2006-12-22
        \param symbol_table_to_attach Symbol table to attach (it will never be deallocated implicitely)
    */
    c_function(c_micro_symbol_table* symbol_table_to_attach);

    /*!
        \brief Parses next command
        \author VooDooMan
        \version 1
        \date 2005
        \param this_module Module name
        \param command Command (next input must be appended to this on next call)
        \param error Returns true for error
        \param error_msg Returns error string
        \param need_more Returns true if more input is required
        \param debug Debug string
        \param debug_prefix Prefix for error
    */
    bool parse_command(c_module* this_module, string& command, bool& error, string& error_msg, bool& need_more, const string& debug, const string& debug_prefix);

    /*!
        \brief Gets replication for specified bot and requested type
        \author VooDooMan
        \version 1
        \date 2005
        \param partner Name of remote bot
        \param type Requested replication type
        \return Returns true if replication should be done
    */
    bool get_replication(const string& partner, int type);

    /*!
        \brief Dumps function to file
        \author VooDooMan
        \version 1
        \date 2005
        \param f File descriptor
    */
    void dump_to_file(FILE* f);

    /*!
        \brief Dumps function to string
        \author VooDooMan
        \version 1
        \date 2005
        \return Retruns dump
    */
    string dump_to_string();

    /*!
        \brief Adds replication rule
        \author VooDooMan
        \version 1
        \date 2005
        \param replication Replication rule definition
    */
    void add_attr(const c_replication& replication);

    /*!
        \brief Adds remote call rule
        \author VooDooMan
        \version 1
        \date 2005
        \param remote_bot Name of remote bot to add
    */
    void add_attr_remote_call(const string& remote_bot);

    /*!
        \brief Adds group membership
        \author VooDooMan
        \version 1
        \date 2005
        \param group Name of group
    */
    void add_attr_memeber_of_group(const string& group);

    /*!
        \brief Gets last changed attribute
        \author VooDooMan
        \version 1
        \date 2005
        \return Returns last changed attribute
    */
    my_time_t get_last_changed();

    /*!
        \brief Sets last changed attribute
        \author VooDooMan
        \version 1
        \date 2005
        \param when Last changed attribute to set
    */
    void set_last_changed(my_time_t when);

    /*!
        \brief Gets minimal BOTNET version
        \author VooDooMan
        \version 1
        \date 2005
        \return Returns minimal BOTNET version
    */
    int get_minimal_botnet_version();

    /*!
        \brief Gets argument list of function
        \author VooDooMan
        \version 1
        \date 2005
        \param args_ Returns arguments
    */
    void get_prototype(vector<c_func_arg>& args_);

    /*!
        \brief Returns true if function was compiled okay
        \author VooDooMan
        \version 1
        \date 2005
    */
    bool is_okay();

    /*!
        \brief Returns name of function (without declaration)
        \author VooDooMan
        \version 1
        \date 2005
    */
    string get_name();

    /*!
        \brief Gets argument list of function
        \author VooDooMan
        \version 1
        \date 2005
        \return Returns arguments
    */
    vector<c_func_arg> get_args();

    /*!
        \brief Gets number of arguments of function
        \author VooDooMan
        \version 1
        \date 2005
    */
    int get_num_args();

    /*!
        \brief Sets function's dump
        \author VooDooMan
        \version 1
        \date 2005
        \param func_dump Returns dump
    */
    void set_dump(const string& func_dump);

    c_function(const c_function& orig);

    void destroy_symbols();

    virtual ~c_function();

    c_function& operator=(const c_function& orig);

    /*!
        \brief Returns newly created symbol table and allocates space for them
        \date 2006-04-03
        \author VooDooMan
        \version 1
    */
    c_micro_symbol_table* make_symbol_table(c_micro_symbol_table* orig);

    /*!
        \brief Destroys completely function, and deallocates memory (used for helping temporary pseudo-functions)
        \date 2006-04-03, 2006-05-05
        \author VooDooMan
        \version 2
        \param also_symbols 'true' if we want to destroy also symbol table
        \param also_cleanup_code If to call clean_up()
    */
    void destroy(bool also_symbols = true, bool also_cleanup_code = true);

    /*!
        \brief Cleans-up microcode (for module unloading)
        \date 2006-05-09
        \author VooDooMan
        \version 1
    */
    void clear_microcode();

    /*!
        \brief Sets symbols (debug only)
        \date 2006-08-16
        \author VooDooMan
        \version 1
    */
    void set_symbols(c_micro_symbol_variable_mapping* s);

    /*!
        \brief Optimizes output by removing result variables that are not referenced as input
        \date 2007-03-18
        \author VooDooMan
        \version 1
    */
    void optimize_output();

    /*!
        \brief Retrieves symbol table
        \author VooDooMan
        \date 2008-06-06
        \version 1
    */
    c_micro_symbol_table* get_symbol_table();
};

#endif

