/***************************************************************************
                          eval_func.cpp  -  description
                             -------------------
    begin                : Thu May 10 2007
    copyright            : (C) 2004-2007 by Marian VooDooMan Meravy
    email                : ghostvoodooman -a.T= users (dot) sourceforge.net
 ***************************************************************************/

/* $Id: eval_func.cpp,v 1.22 2008/11/24 12:09:23 ghostvoodooman Exp $ */

/***************************************************************************

VooDoo cIRCle - an IRC (ro)bot
Copyright (C) 2004-2008 by Marian VooDooMan Meravy
  (ghostvoodooman -a.T= users (dot) sourceforge.net)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************/

/*!
    \file
    \brief Implements evaluation of function
    \version $Id: eval_func.cpp,v 1.22 2008/11/24 12:09:23 ghostvoodooman Exp $
*/

#if defined(HAVE_CONFIG_H)
#   include <config.h>
#endif

#include "../params.h"
#include "../utils/memleak.h"
#include "../utils/assert.h"
#include "../compat.h"

#include "../runtime/c_eval_iteration.h"
#include "c_function.h"
#include "c_fn_cmd_for_cond.h"
#include "c_fn_cmd_exp.h"
#include "c_fn_cmd_break.h"
#include "c_fn_cmd_continue.h"
#include "c_fn_cmd_do.h"
#include "c_fn_cmd_do_while.h"
#include "c_fn_cmd_end_of_func.h"
#include "c_fn_cmd_while.h"
#include "c_fn_cmd_case_default.h"
#include "c_fn_cmd_for_cond.h"
#include "c_fn_cmd_global.h"
#include "c_fn_cmd_goto.h"
#include "c_fn_cmd_jump.h"
#include "../mod/c_module.h"
#include "../log/c_logger.h"
#include "../var_keeper.h"
#include "../microcode/c_micro_symbol_variable_mapping.h"
#include "../microcode/c_micro_symbol_variable_table.h"
#include "../var/c_var_unset.h"
#include "../var/c_var_string.h"
#include "../var/c_var_int.h"
#include "../eval.h"
#include "../utils/precomp.h"

#if defined(_MSC_VER) || defined(__BORLANDC__)
#   pragma hdrstop
#endif

MY_STATIC void log_fn_eval_error(c_var_keeper::e_log_severity UNUSEDARG(severity), const string& msg, void* data)
{
    c_module* m=(c_module*)data;
    if(m && m->get_logger())
        m->get_logger()->log_print1(E_LOG_DEBUG,LOG_SCOPE_DEFAULT,"",msg);
}

#if defined(_MSC_VER)
#   pragma warning(push)
#   pragma warning(disable:4702)    // warning C4702: unreachable code
#endif

#define FUNC_CALL_ERROR(MSG,DEBUG_STR) \
    { \
        string d="FATAL: "; \
        d+=MSG; \
        if(!DEBUG_STR.empty()) { \
            d+=" ("; \
            d+=DEBUG_STR; \
            d+=")"; \
        } \
        d+=(string)"; "+GET_STR_DEBUG_INFO; \
        if(st.top().this_module_ && st.top().this_module_->get_logger()) { \
            st.top().this_module_->get_logger()->log_print1(E_LOG_DEBUG,LOG_SCOPE_MODULE,"",d); \
        } \
        st.top().exception_reported=true; \
        st.top().state=1000; \
        st.top().func_call_ok=false; \
        var_keeper->set_current_session(SESSION_RESERVED_ID); \
        return; \
    }

#define FUNC_CALL_ERROR_RETURN(MSG,DEBUG_STR) \
    { \
        string d="FATAL: "; \
        d+=MSG; \
        if(!DEBUG_STR.empty()) { \
            d+=" ("; \
            d+=DEBUG_STR; \
            d+=")"; \
        } \
        d+=(string)"; "+GET_STR_DEBUG_INFO; \
        if(st.top().this_module_ && st.top().this_module_->get_logger()) { \
            st.top().this_module_->get_logger()->log_print1(E_LOG_DEBUG,LOG_SCOPE_MODULE,"",d); \
        } \
        st.top().exception_reported=true; \
        st.top().state=1000; \
        st.top().func_call_ok=false; \
        var_keeper->set_current_session(SESSION_RESERVED_ID); \
        return EVAL_RETURN; \
    }

#define FUNC_CALL_CAST_RETURN(TYPE,ITER,DEBUG_STR) \
    TYPE* c=NULL; \
    if((*ITER).second) \
        c=dynamic_cast<TYPE*>((*ITER).second);

#define FUNC_CALL_CAST(TYPE,ITER,DEBUG_STR) \
    TYPE* c=NULL; \
    if((*ITER).second) \
        c=dynamic_cast<TYPE*>((*ITER).second);

#define FUNC_CALL_CAST_SWITCH(TYPE,ITER,DEBUG_STR) \
    st.top().c=NULL; \
    if((*ITER).second) \
        st.top().c=dynamic_cast<TYPE*>((*ITER).second);

#define FUNC_CALL_CAST_FOREACH(TYPE,ITER,DEBUG_STR) \
    st.top().c2=NULL; \
    if((*ITER).second) \
        st.top().c2=dynamic_cast<TYPE*>((*ITER).second);

#define FUNC_CALL_CAST_CASE(TYPE,ITER,DEBUG_STR) \
    st.top().c3=NULL; \
    if((*ITER).second) \
        st.top().c3=dynamic_cast<TYPE*>((*ITER).second);

enum e_eval_result
{
    EVAL_AGAIN,
    EVAL_RETURN,
    EVAL_BREAK,
    EVAL_CONT,
    EVAL_FALL_THRU,
};

enum e_eval_jump
{
    EVAL_JUMP_NONE,
    EVAL_JUMP_1001,
    EVAL_JUMP_1002,
    EVAL_JUMP_1003,
    EVAL_JUMP_1004,
    EVAL_JUMP_1005,
    EVAL_JUMP_1006,
    EVAL_JUMP_1007,
    EVAL_JUMP_1008,
};

MY_STATIC e_eval_result eval_for_cond(stack<c_eval_iteration>& st, const e_eval_jump jump)
{
    if(EVAL_JUMP_1001==jump)
        goto again_state_1001;

    {
        FUNC_CALL_CAST_RETURN(c_fn_cmd_for_cond,st.top().i1_fn,(*st.top().i1_fn).second->get_debug_string());

        if(st.top().loop_was_broken) {
            st.top().loop_was_broken=false;
            st.top().loop_was_continued=false;
            st.top().id=c->get_false_jump_id();
            return EVAL_AGAIN;
        }
        if(st.top().loop_was_continued) {
            st.top().loop_was_broken=false;
            st.top().loop_was_continued=false;
            st.top().id=c->get_iterative_exp_id();
            return EVAL_AGAIN;
        }
        st.top().loop_was_broken=false;
        st.top().loop_was_continued=false;

        {
            st.top().return_to_state=1001;
            st.top().instruct_id=0;
            var_sym_index_t sym_return;
            st.top().last_exp=c;
            if(eval_microcode(st,c->get_exp(),st.top().micro_error,sym_return)) {
                return EVAL_RETURN;
            }
            st.top().result=sym_return;
        }
    }
again_state_1001:

    c_micro_error* error=st.top().micro_error;
    if(error)
        st.top().error_debug_string=st.top().last_exp->get_debug_string();
    bool is_return=st.top().is_return_;
    c_var_base* return_value=st.top().return_value;
    if(is_return && !error && !return_value) {
        bool dummy;
        if(st.top().result!=VAR_SYM_INDEX_DISCARD)
            st.top().symbol_table_->get_var_table()->get(st.top().result,return_value,dummy,error);
    }
    c_var_base* res=NULL;
    if(!error) {
        bool dummy;
        if(st.top().result!=VAR_SYM_INDEX_DISCARD)
            st.top().symbol_table_->get_var_table()->get(st.top().result,res,dummy,error);
    }

    FUNC_CALL_CAST_RETURN(c_fn_cmd_for_cond,st.top().i1_fn,(*st.top().i1_fn).second->get_debug_string());

    if(error || st.top().exception_reported) {
        if(error)
            st.top().error_debug_string=st.top().last_exp->get_debug_string();
        st.top().micro_error=error;
        st.top().func_call_ok=false;
        return EVAL_RETURN;
    }
    if(is_return) {
        st.top().return_value=return_value;
        st.top().func_call_ok=true;

        st.top().state=st.top().return_to_state;
        st.top().iterative_result=EXP_I_NEXT_ITERATION;
        st.top().processed=true;

        return EVAL_RETURN;
    }
    bool result=false;
    if(res)
        result=res->is_true();

    if(!result) {
        st.top().id=c->get_false_jump_id();
        return EVAL_AGAIN;
    }
    return EVAL_BREAK;
}

MY_STATIC e_eval_result eval_exp(stack<c_eval_iteration>& st, const e_eval_jump jump)
{
    if(EVAL_JUMP_1002==jump)
        goto again_state_1002;

    {
        st.top().loop_was_broken=false;
        st.top().loop_was_continued=false;
        {
#if defined(_MSC_VER)
#   pragma warning(push)
#   pragma warning(disable:4702)    // warning C4702: unreachable code
#endif
            FUNC_CALL_CAST_RETURN(c_fn_cmd_exp,st.top().i1_fn,(*st.top().i1_fn).second->get_debug_string());
#if defined(_MSC_VER)
#   pragma warning(pop)
#endif
            st.top().c_exp=c;
        }

        {
            st.top().return_to_state=1002;
            st.top().instruct_id=0;
        again_state_1002:
            var_sym_index_t sym_return;
            st.top().last_exp=st.top().c_exp;
            if(eval_microcode(st,st.top().c_exp->get_exp(),st.top().micro_error,sym_return)) {
                return EVAL_RETURN;
            }
        }
    }

    c_micro_error* error=st.top().micro_error;
    if(error)
        st.top().error_debug_string=st.top().last_exp->get_debug_string();
    bool is_return=st.top().is_return_;
    c_var_base* return_value=st.top().return_value;
    if(is_return && !error && !return_value) {
        bool dummy;
        if(st.top().result!=VAR_SYM_INDEX_DISCARD)
            st.top().symbol_table_->get_var_table()->get(st.top().result,return_value,dummy,error);
    }

    if(error)
        st.top().error_debug_string=st.top().last_exp->get_debug_string();

#if defined(_MSC_VER)
#   pragma warning(push)
#   pragma warning(disable:4702)    // warning C4702: unreachable code
#endif
    //FUNC_CALL_CAST_RETURN(c_fn_cmd_exp,st.top().i1_fn,(*st.top().i1_fn).second->get_debug_string());
#if defined(_MSC_VER)
#   pragma warning(pop)
#endif

    if(error || st.top().exception_reported) {
        st.top().micro_error=error;
        st.top().func_call_ok=false;
        return EVAL_RETURN;
    }
    if(is_return) {
        st.top().return_value=return_value;
        st.top().func_call_ok=true;

        st.top().state=st.top().return_to_state;
        st.top().iterative_result=EXP_I_NEXT_ITERATION;
        st.top().processed=true;

        return EVAL_RETURN;
    }
    return EVAL_BREAK;
}

MY_STATIC e_eval_result eval_if(stack<c_eval_iteration>& st, const e_eval_jump jump)
{
    if(EVAL_JUMP_1003==jump)
        goto again_state_1003;

    {
        st.top().loop_was_broken=false;
        st.top().loop_was_continued=false;
        {
#if defined(_MSC_VER)
#   pragma warning(push)
#   pragma warning(disable:4702)    // warning C4702: unreachable code
#endif
            FUNC_CALL_CAST_RETURN(c_fn_cmd_if,st.top().i1_fn,(*st.top().i1_fn).second->get_debug_string());
#if defined(_MSC_VER)
#   pragma warning(pop)
#endif
            st.top().c_if=c;
        }

        {
            st.top().return_to_state=1003;
            st.top().instruct_id=0;
        again_state_1003:
            var_sym_index_t sym_return;
            st.top().last_exp=st.top().c_if;
            if(eval_microcode(st,st.top().c_if->get_exp(),st.top().micro_error,sym_return)) {
                return EVAL_RETURN;
            }
        }
    }

    c_micro_error* error=st.top().micro_error;
    if(error)
        st.top().error_debug_string=st.top().last_exp->get_debug_string();
    bool is_return=st.top().is_return_;
    c_var_base* return_value=st.top().return_value;
    if(is_return && !error && !return_value) {
        bool dummy;
        if(st.top().result!=VAR_SYM_INDEX_DISCARD)
            st.top().symbol_table_->get_var_table()->get(st.top().result,return_value,dummy,error);
    }
    if(error)
        st.top().error_debug_string=st.top().last_exp->get_debug_string();
    c_var_base* res=NULL;
    if(!error) {
        bool dummy;
        if(st.top().result!=VAR_SYM_INDEX_DISCARD)
            st.top().symbol_table_->get_var_table()->get(st.top().result,res,dummy,error);
    }

    if(error)
        st.top().error_debug_string=st.top().last_exp->get_debug_string();

#if defined(_MSC_VER)
#   pragma warning(push)
#   pragma warning(disable:4702)    // warning C4702: unreachable code
#endif
    FUNC_CALL_CAST_RETURN(c_fn_cmd_if,st.top().i1_fn,(*st.top().i1_fn).second->get_debug_string());
#if defined(_MSC_VER)
#   pragma warning(pop)
#endif

    if(error || st.top().exception_reported) {
        st.top().micro_error=error;
        st.top().func_call_ok=false;
        return EVAL_RETURN;
    }
    if(is_return) {
        st.top().return_value=return_value;
        st.top().func_call_ok=true;

        st.top().state=st.top().return_to_state;
        st.top().iterative_result=EXP_I_NEXT_ITERATION;
        st.top().processed=true;

        return EVAL_RETURN;
    }

    if(st.top().skip_block) {
        st.top().id=c->get_else_jump_id();
        return EVAL_AGAIN;
    }
    return EVAL_BREAK;
}

MY_STATIC e_eval_result eval_switch(stack<c_eval_iteration>& st, const e_eval_jump jump)
{
    if(EVAL_JUMP_1005==jump)
        goto again_state_1005;
    else if(EVAL_JUMP_1004==jump)
        goto again_state_1004;

    {
#if defined(_MSC_VER)
#   pragma warning(push)
#   pragma warning(disable:4702)    // warning C4702: unreachable code
#endif
        FUNC_CALL_CAST_SWITCH(c_fn_cmd_switch,st.top().i1_fn,(*st.top().i1_fn).second->get_debug_string());
#if defined(_MSC_VER)
#   pragma warning(pop)
#endif

        if(st.top().loop_was_broken) {
            st.top().loop_was_broken=false;
            st.top().loop_was_continued=false;
            st.top().id=st.top().c->get_no_match_id_jump();
            return EVAL_AGAIN;
        }
        st.top().loop_was_broken=false;
        st.top().loop_was_continued=false;

        {
            st.top().return_to_state=1004;
            st.top().instruct_id=0;
        again_state_1004:
            var_sym_index_t sym_return;
            st.top().last_exp=st.top().c;
            if(eval_microcode(st,st.top().c->get_exp(),st.top().micro_error,sym_return)) {
                return EVAL_RETURN;
            }
            st.top().result=sym_return;
        }
    }

    {
        c_micro_error* error=st.top().micro_error;
        if(error)
            st.top().error_debug_string=st.top().last_exp->get_debug_string();
        bool is_return=st.top().is_return_;
        c_var_base* return_value=st.top().return_value;
        if(is_return && !error && !return_value) {
            bool dummy;
            // never optimize for 'switch'!
            //if(st.top().result!=VAR_SYM_INDEX_DISCARD)
            st.top().symbol_table_->get_var_table()->get(st.top().result,return_value,dummy,error);
            //return_value=st.top().return_value;
        }
        var_sym_index_t res=st.top().result;

        st.top().res_1004=res;
        st.top().error_1004=error;
        st.top().is_return_1004=is_return;
        st.top().return_value=return_value;
    }

#if defined(_MSC_VER)
#   pragma warning(push)
#   pragma warning(disable:4702)    // warning C4702: unreachable code
#endif
    FUNC_CALL_CAST_SWITCH(c_fn_cmd_switch,st.top().i1_fn,(*st.top().i1_fn).second->get_debug_string());
#if defined(_MSC_VER)
#   pragma warning(pop)
#endif

    if(st.top().error_1004 || st.top().exception_reported) {
        st.top().micro_error=st.top().error_1004;
        st.top().func_call_ok=false;
        return EVAL_RETURN;
    }
    if(st.top().is_return_1004) {
        st.top().func_call_ok=true;

        st.top().state=st.top().return_to_state;
        st.top().iterative_result=EXP_I_NEXT_ITERATION;
        st.top().processed=true;

        return EVAL_RETURN;
    }

    st.top().sw_res=NULL;
    {
        bool dummy;
        // never optimize for 'switch'!
        //if(st.top().res_1004!=VAR_SYM_INDEX_DISCARD)
        if(!st.top().symbol_table_->get_var_table()->get(st.top().res_1004,st.top().sw_res,dummy,st.top().micro_error)) {
            st.top().func_call_ok=false;
            return EVAL_RETURN;
        }
        if(st.top().sw_res)
            st.top().sw_res=st.top().sw_res->copy();
    }

    {
        // iterate cases id's
        st.top().sw_c=st.top().c;
        st.top().cases=st.top().c->get_cases_id();
        for(st.top().ci=st.top().cases.begin(); st.top().ci!=st.top().cases.end(); ++st.top().ci) {
            st.top().i1_fn2=st.top().fn->commands.find(*st.top().ci);
            if(st.top().i1_fn2==st.top().fn->commands.end())
                break;

            if(typeid(*(*st.top().i1_fn2).second)==typeid(c_fn_cmd_case_default)) {
                if(st.top().sw_res && !st.top().sw_res->get_persistent()) {
                    MY_DELETE(st.top().sw_res);
                    st.top().sw_res=NULL;
                }
                st.top().id=*st.top().ci;
                return EVAL_AGAIN;
            }

            {
#if defined(_MSC_VER)
#   pragma warning(push)
#   pragma warning(disable:4702)    // warning C4702: unreachable code
#endif
                FUNC_CALL_CAST_CASE(c_fn_cmd_case,st.top().i1_fn2,(*st.top().i1_fn2).second->get_debug_string());
#if defined(_MSC_VER)
#   pragma warning(pop)
#endif

                {
                    st.top().return_to_state=1005;
                    st.top().instruct_id=0;
                again_state_1005:
                    var_sym_index_t sym_return;
                    st.top().last_exp=st.top().c3;
                    if(eval_microcode(st,st.top().c3->get_exp(),st.top().micro_error,sym_return)) {
                        return EVAL_RETURN;
                    }
                    st.top().result=sym_return;
                }
            }

            c_micro_error* error=st.top().micro_error;
            if(error)
                st.top().error_debug_string=st.top().last_exp->get_debug_string();
            bool is_return=st.top().is_return_;
            c_var_base* return_value=st.top().return_value;
            if(is_return && !error && !return_value) {
                bool dummy;
                // never optimize for 'switch'!
                //if(st.top().result!=VAR_SYM_INDEX_DISCARD)
                st.top().symbol_table_->get_var_table()->get(st.top().result,return_value,dummy,error);
                //return_value=st.top().return_value;
            }
            if(error)
                st.top().error_debug_string=st.top().last_exp->get_debug_string();
            c_var_base* res=NULL;
            if(!error) {
                bool dummy;
                // never optimize for 'switch'!
                //if(st.top().result!=VAR_SYM_INDEX_DISCARD)
                st.top().symbol_table_->get_var_table()->get(st.top().result,res,dummy,error);
            }

            if(error)
                st.top().error_debug_string=st.top().last_exp->get_debug_string();

#if defined(_MSC_VER)
#   pragma warning(push)
#   pragma warning(disable:4702)    // warning C4702: unreachable code
#endif
            //FUNC_CALL_CAST_RETURN(c_fn_cmd_case,st.top().i1_fn2,(*st.top().i1_fn).second->get_debug_string());
#if defined(_MSC_VER)
#   pragma warning(pop)
#endif

            if(error || st.top().exception_reported) {
                st.top().micro_error=error;
                st.top().func_call_ok=false;
                MY_DELETE(st.top().sw_res);
                st.top().sw_res=NULL;
                return EVAL_RETURN;
            }
            if(is_return) {
                st.top().return_value=return_value;
                st.top().func_call_ok=true;

                st.top().state=st.top().return_to_state;
                st.top().iterative_result=EXP_I_NEXT_ITERATION;
                st.top().processed=true;

                MY_DELETE(st.top().sw_res);
                st.top().sw_res=NULL;

                return EVAL_RETURN;
            }

            if(!res)
                res=MY_NEW c_var_unset();
            bool got=st.top().sw_res->is_equal(st.top().this_module_,res,"",(*st.top().i1_fn2).second->get_debug_string());
            res=NULL;
            if(got) {
                MY_DELETE(st.top().sw_res);
                st.top().sw_res=NULL;
                st.top().id=*st.top().ci;
                return EVAL_AGAIN;
            }
        }
        // not found match to any 'case', nor 'default'
        MY_DELETE(st.top().sw_res);
        st.top().sw_res=NULL;
        st.top().id=st.top().sw_c->get_no_match_id_jump();
        return EVAL_AGAIN;
    }
    // NOTREACH
}

MY_STATIC e_eval_result eval_while(stack<c_eval_iteration>& st, const e_eval_jump jump)
{
    if(EVAL_JUMP_1006==jump)
        goto again_state_1006;

    {
        FUNC_CALL_CAST_RETURN(c_fn_cmd_while,st.top().i1_fn,(*st.top().i1_fn).second->get_debug_string());

        if(st.top().loop_was_broken) {
            st.top().loop_was_broken=false;
            st.top().loop_was_continued=false;
            st.top().id=c->get_end_loop_id();
            return EVAL_AGAIN;
        }
        st.top().loop_was_broken=false;
        st.top().loop_was_continued=false;

        {
            st.top().return_to_state=1006;
            st.top().instruct_id=0;
            var_sym_index_t sym_return;
            st.top().last_exp=st.top().c;
            if(eval_microcode(st,c->get_exp(),st.top().micro_error,sym_return)) {
                return EVAL_RETURN;
            }
        }
    }
again_state_1006:

    c_micro_error* error=st.top().micro_error;
    if(error)
        st.top().error_debug_string=st.top().last_exp->get_debug_string();
    bool is_return=st.top().is_return_;
    c_var_base* return_value=st.top().return_value;
    if(is_return && !error && !return_value) {
        bool dummy;
        if(st.top().result!=VAR_SYM_INDEX_DISCARD)
            st.top().symbol_table_->get_var_table()->get(st.top().result,return_value,dummy,error);
        //return_value=st.top().return_value;
    }

    if(error)
        st.top().error_debug_string=st.top().last_exp->get_debug_string();

    FUNC_CALL_CAST_RETURN(c_fn_cmd_while,st.top().i1_fn,(*st.top().i1_fn).second->get_debug_string());

    if(error || st.top().exception_reported) {
        st.top().micro_error=error;
        st.top().func_call_ok=false;
        return EVAL_RETURN;
    }
    if(is_return) {
        st.top().return_value=return_value;
        st.top().func_call_ok=true;

        st.top().state=st.top().return_to_state;
        st.top().iterative_result=EXP_I_NEXT_ITERATION;
        st.top().processed=true;

        return EVAL_RETURN;
    }

    c_var_base* res=NULL;
    if(!error) {
        bool dummy;
        if(c->get_exp()->result!=VAR_SYM_INDEX_DISCARD)
            st.top().symbol_table_->get_var_table()->get(c->get_exp()->result,res,dummy,error);
    }

    if(error)
        st.top().error_debug_string=st.top().last_exp->get_debug_string();

    bool result=(res?res->is_true():false);

    if(!result) {
        st.top().id=c->get_end_loop_id();
        return EVAL_AGAIN;
    }
    return EVAL_BREAK;
}

MY_STATIC e_eval_result eval_do_while(stack<c_eval_iteration>& st, const e_eval_jump jump)
{
    if(EVAL_JUMP_1007==jump)
        goto again_state_1007;

    {
        FUNC_CALL_CAST_RETURN(c_fn_cmd_do_while,st.top().i1_fn,(*st.top().i1_fn).second->get_debug_string());
        if(st.top().loop_was_broken) {
            st.top().loop_was_broken=false;
            st.top().loop_was_continued=false;
            // just fall thru the code
            return EVAL_CONT;
        }
        st.top().loop_was_broken=false;
        st.top().loop_was_continued=false;

        {
            st.top().return_to_state=1007;
            st.top().instruct_id=0;
            var_sym_index_t sym_return;
            st.top().last_exp=st.top().c;
            if(eval_microcode(st,c->get_exp(),st.top().micro_error,sym_return)) {
                return EVAL_RETURN;
            }
            st.top().result=sym_return;
        }
    }
again_state_1007:

    c_micro_error* error=st.top().micro_error;
    if(error)
        st.top().error_debug_string=st.top().last_exp->get_debug_string();
    bool is_return=st.top().is_return_;
    c_var_base* return_value=st.top().return_value;
    if(is_return && !error && !return_value) {
        bool dummy;
        if(st.top().result!=VAR_SYM_INDEX_DISCARD)
            st.top().symbol_table_->get_var_table()->get(st.top().result,return_value,dummy,error);
    }
    if(error)
        st.top().error_debug_string=st.top().last_exp->get_debug_string();
    c_var_base* res=NULL;
    if(!error) {
        bool dummy;
        if(st.top().result!=VAR_SYM_INDEX_DISCARD)
            st.top().symbol_table_->get_var_table()->get(st.top().result,res,dummy,error);
    }
    if(error)
        st.top().error_debug_string=st.top().last_exp->get_debug_string();

    FUNC_CALL_CAST_RETURN(c_fn_cmd_do_while,st.top().i1_fn,(*st.top().i1_fn).second->get_debug_string());

    if(error || st.top().exception_reported) {
        st.top().micro_error=error;
        st.top().func_call_ok=false;
        return EVAL_RETURN;
    }
    if(is_return) {
        st.top().return_value=return_value;
        st.top().func_call_ok=true;

        st.top().state=st.top().return_to_state;
        st.top().iterative_result=EXP_I_NEXT_ITERATION;
        st.top().processed=true;

        return EVAL_RETURN;
    }
    bool result=(res?res->is_true():false);
    if(!result) {
        // just fall thru the code
        return EVAL_CONT;
    } else {
        st.top().id=c->get_again_jump_id();
        return EVAL_AGAIN;
    }

    // NOTREACH
}

MY_STATIC e_eval_result eval_foreach(stack<c_eval_iteration>& st, const e_eval_jump jump)
{
    if(EVAL_JUMP_1008==jump)
        goto again_state_1008;

    FUNC_CALL_CAST_FOREACH(c_fn_cmd_foreach,st.top().i1_fn,(*st.top().i1_fn).second->get_debug_string());

    if(st.top().loop_was_broken) {
        st.top().loop_was_broken=false;
        st.top().loop_was_continued=false;
        st.top().id=st.top().c2->get_end_id();
        return EVAL_AGAIN;
    }
    if(st.top().loop_was_continued) {
        st.top().loop_was_broken=false;
        st.top().loop_was_continued=false;
        st.top().id=st.top().c2->get_next_iteration_id();
        return EVAL_AGAIN;
    }

    st.top().loop_was_broken=false;
    st.top().loop_was_continued=false;

    st.top().b_fn.clear();
    st.top().b_fn.end_loop_id=st.top().c2->get_end_id();
    st.top().b_fn.debug=st.top().c2->get_debug_string();
    {
        // first iteration
        st.top().b_fn.foreach_iterator=0;

        // "what" expression evaluation
        {
            {
                st.top().return_to_state=1008;
                st.top().instruct_id=0;
                var_sym_index_t sym_return;
                st.top().last_exp=st.top().c2;
                if(eval_microcode(st,st.top().c2->get_what(),st.top().micro_error,sym_return)) {
                    return EVAL_RETURN;
                }
                st.top().result=sym_return;
            }
    again_state_1008:

            c_micro_error* error=st.top().micro_error;
            if(error)
                st.top().error_debug_string=st.top().last_exp->get_debug_string();
            bool is_return=st.top().is_return_;
            c_var_base* return_value=st.top().return_value;
            if(is_return && !error && !return_value) {
                bool dummy;
                if(st.top().result!=VAR_SYM_INDEX_DISCARD)
                    st.top().symbol_table_->get_var_table()->get(st.top().result,return_value,dummy,error);
                //return_value=st.top().return_value;
            }
            if(error)
                st.top().error_debug_string=st.top().last_exp->get_debug_string();
            c_var_base* res=NULL;
            if(!error) {
                bool dummy;
                if(st.top().result!=VAR_SYM_INDEX_DISCARD)
                    st.top().symbol_table_->get_var_table()->get(st.top().result,res,dummy,error);
            }
            if(error)
                st.top().error_debug_string=st.top().last_exp->get_debug_string();
            var_sym_index_t what=st.top().result;

            //FUNC_CALL_CAST_RETURN(c_fn_cmd_foreach,st.top().i1_fn,(*st.top().i1_fn).second->get_debug_string());

            if(error || st.top().exception_reported) {
                st.top().micro_error=error;
                st.top().func_call_ok=false;
                return EVAL_RETURN;
            }
            if(is_return) {
                st.top().return_value=return_value;
                st.top().func_call_ok=true;

                st.top().state=st.top().return_to_state;
                st.top().iterative_result=EXP_I_NEXT_ITERATION;
                st.top().processed=true;

                return EVAL_RETURN;
            }

            if(!res) {
                st.top().error_debug_string=st.top().last_exp->get_debug_string();
                c_eval_runtime_error ee(st.top().this_module_,st.top().error_debug_string,GET_STR_DEBUG_INFO_EX("UNSET/NULL variable used as an array to iterate"),__FILE__,__my_func__);
                ee.report(true);
                st.top().func_call_ok=false;
                return EVAL_RETURN;
            }
            st.top().b_fn.what=what;
            st.top().b_fn.value=st.top().c2->get_as();
            st.top().b_fn.key=st.top().c2->get_key();
            st.top().b_fn.ref_value=st.top().c2->get_as_ref();

            if(!st.top().b_fn.value.empty()) {
                // invalidate $element first!
                var_sym_index_t value=st.top().symbol_table_->get_variable_mapping()->get_var_index(st.top().b_fn.value);
                if(VAR_SYM_INDEX_INVALID==value) {
                    c_eval_runtime_error ee(st.top().this_module_,st.top().foreach_stack.top().debug,GET_STR_DEBUG_INFO_EX("Could not get ID for 'value/element' variable"),__FILE__,__my_func__);
                    ee.report(true);
                    st.top().func_call_ok=false;
                    return EVAL_RETURN;
                }
                if(!st.top().symbol_table_->get_var_table()->invalidate(value,st.top().micro_error,MICRO_NA)) {
                    st.top().func_call_ok=false;
                    return EVAL_RETURN;
                }
            }
        }
    }
    st.top().foreach_stack.push(st.top().b_fn);

    st.top().fall_thru=true;
    // do NOT continue! just fall thru

    return EVAL_FALL_THRU;
}

MY_STATIC e_eval_result eval_foreach_iter(stack<c_eval_iteration>& st)
{
    st.top().loop_was_broken=false;
    st.top().loop_was_continued=false;

    if(st.top().foreach_stack.empty())
        FUNC_CALL_ERROR_RETURN("Internal error, 'foreach_stack' is empty",(*st.top().i1_fn).second->get_debug_string());

    {
        if(!st.top().fall_thru)
            st.top().foreach_stack.top().foreach_iterator++;
        st.top().fall_thru=false;
        if(st.top().foreach_stack.top().ref_value) {
            st.top().refs.pop(st.top().vars_);
        }

        // not first iteration => unset $element/$key/$value
        if(st.top().foreach_stack.top().key.empty() && st.top().foreach_stack.top().value.empty()) {
            // nop
        } else if(!st.top().foreach_stack.top().key.empty() && !st.top().foreach_stack.top().value.empty()) {
        } else {
            // CMD_FOREACH_2
        }
    }

    if(st.top().foreach_stack.top().key.empty() && st.top().foreach_stack.top().value.empty()) {
        bool dummy;
        c_var_base* what=NULL;
        if(st.top().foreach_stack.top().what!=VAR_SYM_INDEX_DISCARD)
            if(!st.top().symbol_table_->get_var_table()->get(st.top().foreach_stack.top().what,what,dummy,st.top().micro_error)) {
                st.top().func_call_ok=false;
                return EVAL_RETURN;
            }
        c_var_array* a=NULL;
        if(what)
            a=dynamic_cast<c_var_array*>(what);
        if(!a) {
            st.top().error_debug_string=st.top().last_exp->get_debug_string();
            c_eval_runtime_error ee(st.top().this_module_,st.top().error_debug_string,GET_STR_DEBUG_INFO_EX("UNSET/NULL variable used as an array to iterate"),__FILE__,__my_func__);
            ee.report(true);
            st.top().func_call_ok=false;
            return EVAL_RETURN;
        }
        a->clear_unset();
        if(st.top().foreach_stack.top().foreach_iterator>=var_keeper->array_num_nonunset_children(st.top().this_module_,a)) {
            st.top().id=st.top().foreach_stack.top().end_loop_id;
            var_keeper->free_reference(st.top().foreach_stack.top().reference,log_fn_eval_error,st.top().this_module_);
            st.top().foreach_stack.pop();
            return EVAL_AGAIN;
        }
    } else if(!st.top().foreach_stack.top().key.empty() && !st.top().foreach_stack.top().value.empty() && !st.top().foreach_stack.top().ref_value) {
        bool dummy=true;
        c_var_base* what=NULL;
        if(st.top().foreach_stack.top().what!=VAR_SYM_INDEX_DISCARD) {
            if(!st.top().symbol_table_->get_var_table()->get(st.top().foreach_stack.top().what,what,dummy,st.top().micro_error)) {
                st.top().func_call_ok=false;
                return EVAL_RETURN;
            }
        }
        c_var_array* a=NULL;
        if(what)
            a=dynamic_cast<c_var_array*>(what);
        if(!a) {
            c_eval_runtime_error ee(st.top().this_module_,st.top().foreach_stack.top().debug,GET_STR_DEBUG_INFO_EX("Iterated variable is not an array anymore"),__FILE__,__my_func__);
            ee.report(false);
            st.top().func_call_ok=false;
            return EVAL_RETURN;
        }
        a->clear_unset();

        {
            // assign iterator to $key
            string as_key=st.top().foreach_stack.top().key;
            var_sym_index_t key=st.top().symbol_table_->get_variable_mapping()->get_var_index(as_key);
            if(VAR_SYM_INDEX_INVALID==key) {
                c_eval_runtime_error ee(st.top().this_module_,st.top().foreach_stack.top().debug,GET_STR_DEBUG_INFO_EX("Could not get ID for 'key' variable"),__FILE__,__my_func__);
                ee.report(true);
                st.top().func_call_ok=false;
                return EVAL_RETURN;
            }
            if(st.top().foreach_stack.top().foreach_iterator>=var_keeper->array_num_nonunset_children(st.top().this_module_,a)) {
                st.top().id=st.top().foreach_stack.top().end_loop_id;
                var_keeper->free_reference(st.top().foreach_stack.top().reference,log_fn_eval_error,st.top().this_module_);
                st.top().foreach_stack.pop();
                return EVAL_AGAIN;
            } else {
                c_tree_node* x=var_keeper->array_get_nonunset_child(st.top().this_module_,a,st.top().foreach_stack.top().foreach_iterator);
                c_var_base* key_val=NULL;
                if((*x).data2.first->type()==c_index_base::IDX_STR)
                    key_val=MY_NEW c_var_string(dynamic_cast<c_index_str*>((*x).data2.first)->get());
                if((*x).data2.first->type()==c_index_base::IDX_INT)
                    key_val=MY_NEW c_var_int(dynamic_cast<c_index_int*>((*x).data2.first)->get());
                if(key_val) {
                    {
                        c_var_base* key_val=NULL;
                        if(key!=VAR_SYM_INDEX_DISCARD) {
                            if(!st.top().symbol_table_->get_var_table()->get(key,key_val,dummy,st.top().micro_error,true)) {
                                st.top().func_call_ok=false;
                                return EVAL_RETURN;
                            }
                        }
                        if(key_val)
                            key_val->set_read_only(false);
                    }

                    if(!st.top().symbol_table_->get_var_table()->set(key,key_val,dummy,st.top().micro_error,MICRO_NA,false,st.top().this_module_,"",st.top().foreach_stack.top().debug)) {
                        st.top().func_call_ok=false;
                        return EVAL_RETURN;
                    }
                }
            }
        }

        {
            // assign value to $value
            string as_value=st.top().foreach_stack.top().value;
            var_sym_index_t value=st.top().symbol_table_->get_variable_mapping()->get_var_index(as_value);
            if(VAR_SYM_INDEX_INVALID==value) {
                c_eval_runtime_error ee(st.top().this_module_,st.top().foreach_stack.top().debug,GET_STR_DEBUG_INFO_EX("Could not get ID for 'value/element' variable"),__FILE__,__my_func__);
                ee.report(true);
                st.top().func_call_ok=false;
                return EVAL_RETURN;
            }
            if(st.top().foreach_stack.top().foreach_iterator>=var_keeper->array_num_nonunset_children(st.top().this_module_,a)) {
                st.top().id=st.top().foreach_stack.top().end_loop_id;
                var_keeper->free_reference(st.top().foreach_stack.top().reference,log_fn_eval_error,st.top().this_module_);
                st.top().foreach_stack.pop();
                return EVAL_AGAIN;
            } else {
                c_tree_node* x=var_keeper->array_get_nonunset_child(st.top().this_module_,a,st.top().foreach_stack.top().foreach_iterator);
                c_var_base* val=var_keeper->get(x->data2.second ? x->data2.second->get_id() : VAR_ID_INVALID,log_fn_eval_error,st.top().micro_error);
                if(!st.top().foreach_stack.top().ref_value) {
                    if(val) {
                        val=val->copy();
                        val->set_read_only(false); // because we could iterate array with read-only values
                    } else
                        val=MY_NEW c_var_unset();
                } else {
                    c_eval_runtime_error ee(st.top().this_module_,st.top().foreach_stack.top().debug,GET_STR_DEBUG_INFO_EX("Could not satisfy reference to 'value/element' variable, please report this serious problem! Setting UNSET variable to 'value/element' variable as temporary solution to the problem"),__FILE__,__my_func__);
                    ee.report(false);
                    if(!val)
                        val=MY_NEW c_var_unset();
                }

                if(!st.top().symbol_table_->get_var_table()->set(value,val,dummy,st.top().micro_error,MICRO_NA,false,st.top().this_module_,"",st.top().foreach_stack.top().debug)) {
                    st.top().func_call_ok=false;
                    return EVAL_RETURN;
                }
            }
        }
    } else {
        // CMD_FOREACH_2

        string as_value=st.top().foreach_stack.top().value;
        var_sym_index_t value=st.top().symbol_table_->get_variable_mapping()->get_var_index(as_value);
        if(VAR_SYM_INDEX_INVALID==value) {
            c_eval_runtime_error ee(st.top().this_module_,st.top().foreach_stack.top().debug,GET_STR_DEBUG_INFO_EX("Could not get ID for 'value/element' variable"),__FILE__,__my_func__);
            ee.report(true);
            st.top().func_call_ok=false;
            return EVAL_RETURN;
        }

        // assign value to $element
        bool dummy=true;
        c_var_base* what=NULL;
        if(st.top().foreach_stack.top().what!=VAR_SYM_INDEX_DISCARD) {
            if(!st.top().symbol_table_->get_var_table()->get(st.top().foreach_stack.top().what,what,dummy,st.top().micro_error)) {
                st.top().func_call_ok=false;
                return EVAL_RETURN;
            }
        }
        c_var_array* a=NULL;
        if(what)
            a=dynamic_cast<c_var_array*>(what);
        if(!a) {
            c_eval_runtime_error ee(st.top().this_module_,st.top().foreach_stack.top().debug,GET_STR_DEBUG_INFO_EX("Iterated variable is not an array anymore, terminating loop"),__FILE__,__my_func__);
            ee.report(false);
            st.top().id=st.top().foreach_stack.top().end_loop_id;
            var_keeper->free_reference(st.top().foreach_stack.top().reference,log_fn_eval_error,st.top().this_module_);
            st.top().foreach_stack.pop();
            return EVAL_AGAIN;
        }
        a->clear_unset();

        if(st.top().foreach_stack.top().foreach_iterator>=var_keeper->array_num_nonunset_children(st.top().this_module_,a)) {
            st.top().id=st.top().foreach_stack.top().end_loop_id;
            var_keeper->free_reference(st.top().foreach_stack.top().reference,log_fn_eval_error,st.top().this_module_);
            st.top().foreach_stack.pop();
            return EVAL_AGAIN;
        } else {
            c_tree_node* x=var_keeper->array_get_nonunset_child(st.top().this_module_,a,st.top().foreach_stack.top().foreach_iterator);

            if(!st.top().foreach_stack.top().key.empty()) {
                var_sym_index_t key=st.top().symbol_table_->get_variable_mapping()->get_var_index(st.top().foreach_stack.top().key);
                c_var_base* key_val=NULL;
                if((*x).data2.first->type()==c_index_base::IDX_STR)
                    key_val=MY_NEW c_var_string(dynamic_cast<c_index_str*>((*x).data2.first)->get());
                if((*x).data2.first->type()==c_index_base::IDX_INT)
                    key_val=MY_NEW c_var_int(dynamic_cast<c_index_int*>((*x).data2.first)->get());
                if(key_val) {
                    {
                        c_var_base* key_val=NULL;
                        if(key!=VAR_SYM_INDEX_DISCARD)
                            if(!st.top().symbol_table_->get_var_table()->get(key,key_val,dummy,st.top().micro_error,true)) {
                                st.top().func_call_ok=false;
                                return EVAL_RETURN;
                            }
                        if(key_val)
                            key_val->set_read_only(false);
                    }

                    if(!st.top().symbol_table_->get_var_table()->set(key,key_val,dummy,st.top().micro_error,MICRO_NA,false,st.top().this_module_,"",st.top().foreach_stack.top().debug)) {
                        st.top().func_call_ok=false;
                        return EVAL_RETURN;
                    }
                }
            }

            c_var_base* val=var_keeper->get(x->data2.second ? x->data2.second->get_id() : VAR_ID_INVALID,log_fn_eval_error,st.top().micro_error);
            if(!st.top().foreach_stack.top().ref_value) {
                if(val)
                    val=val->copy();
                else
                    val=MY_NEW c_var_unset();

                if(!st.top().symbol_table_->get_var_table()->invalidate(value,st.top().micro_error,MICRO_NA)) {
                    st.top().func_call_ok=false;
                    return EVAL_RETURN;
                }
            } else {
                if(!st.top().symbol_table_->get_var_table()->invalidate(value,st.top().micro_error,MICRO_NA)) {
                    st.top().func_call_ok=false;
                    return EVAL_RETURN;
                }

                if(!val)
                    val=MY_NEW c_var_unset();
                c_var_base* orig=val;
                val=MY_NEW c_var_unset();
                var_keeper->free_reference(st.top().foreach_stack.top().reference,log_fn_eval_error,st.top().this_module_);
                var_id_t r=var_keeper->add_reference(val->get_id(),orig->get_id(),log_fn_eval_error,st.top().this_module_);
                st.top().foreach_stack.top().reference=r;
            }

            if(!st.top().symbol_table_->get_var_table()->set(value,val,dummy,st.top().micro_error,MICRO_NA,st.top().foreach_stack.top().ref_value,st.top().this_module_,"",st.top().foreach_stack.top().debug)) {
                st.top().func_call_ok=false;
                return EVAL_RETURN;
            }
        }
    }
    return EVAL_BREAK;
}

void eval_function(stack<c_eval_iteration>& st, my_time_t& start, const my_time_t eval_script_timeout, unsigned eval_stack_size, bool& stack_overflow)
{
    my_time_t last_check=my_time(NULL);

    switch(st.top().state) {
        case 1001:
            goto again_state_1001;
        case 1002:
            goto again_state_1002;
        case 1003:
            goto again_state_1003;
        case 1004:
            goto again_state_1004;
        case 1005:
            goto again_state_1005;
        case 1006:
            goto again_state_1006;
        case 1007:
            goto again_state_1007;
        case 1008:
            goto again_state_1008;
    }

    for(;;) {
        ++st.top().id;
    again:
        st.top().i1_fn=st.top().fn->commands.find(st.top().id);
        if(st.top().i1_fn==st.top().fn->commands.end())
            break;

        {
            if(last_check!=my_time(NULL)) {
                eval_check_limits(st,start,eval_script_timeout,eval_stack_size,stack_overflow);

                if(st.top().exception_reported) {
                    st.top().func_call_ok=false;
                    return;
                }

                last_check=my_time(NULL);
            }
        }

#if 1==DEBUG_PROCFUNC_CMDS_ECHO
        cout << (*st.top().i1_fn).second->get_debug_string() << endl;
#endif

        switch((*st.top().i1_fn).second->get_type()) {
            case CMD_LEFT_BRACE:
            case CMD_RIGHT_BRACE:
            case CMD_CASE:
            case CMD_CASE_DEFAULT:
            case CMD_LABEL:
            case CMD_NOP:
                break;

            case CMD_FOR_COND:
                switch(eval_for_cond(st,EVAL_JUMP_NONE)) {
                    case EVAL_AGAIN:
                        goto again;
                    case EVAL_RETURN:
                        return;
                    case EVAL_BREAK:
                        break;
                    case EVAL_CONT:
                        continue;
                    case EVAL_FALL_THRU:
                        MY_ASSERT_NOTREACH();
                        break;
                }
                break;
again_state_1001:
                switch(eval_for_cond(st,EVAL_JUMP_1001)) {
                    case EVAL_AGAIN:
                        goto again;
                    case EVAL_RETURN:
                        return;
                    case EVAL_BREAK:
                        break;
                    case EVAL_CONT:
                        continue;
                    case EVAL_FALL_THRU:
                        MY_ASSERT_NOTREACH();
                        break;
                }
                break;
            case CMD_EXP:
                switch(eval_exp(st,EVAL_JUMP_NONE)) {
                    case EVAL_AGAIN:
                        goto again;
                    case EVAL_RETURN:
                        return;
                    case EVAL_BREAK:
                        break;
                    case EVAL_CONT:
                        continue;
                    case EVAL_FALL_THRU:
                        MY_ASSERT_NOTREACH();
                        break;
                }
                break;
again_state_1002:
                switch(eval_exp(st,EVAL_JUMP_1002)) {
                    case EVAL_AGAIN:
                        goto again;
                    case EVAL_RETURN:
                        return;
                    case EVAL_BREAK:
                        break;
                    case EVAL_CONT:
                        continue;
                    case EVAL_FALL_THRU:
                        MY_ASSERT_NOTREACH();
                        break;
                }
                break;
            case CMD_JMP:
                {
                    st.top().loop_was_broken=false;
                    st.top().loop_was_continued=false;
#if defined(_MSC_VER)
#   pragma warning(push)
#   pragma warning(disable:4702)    // warning C4702: unreachable code
#endif
                    FUNC_CALL_CAST(c_fn_cmd_jump,st.top().i1_fn,(*st.top().i1_fn).second->get_debug_string());
#if defined(_MSC_VER)
#   pragma warning(pop)
#endif
                    st.top().id=c->get_jump_to_id();
                    goto again;
                }
                break;
            case CMD_IF:
                switch(eval_if(st,EVAL_JUMP_NONE)) {
                    case EVAL_AGAIN:
                        goto again;
                    case EVAL_RETURN:
                        return;
                    case EVAL_BREAK:
                        break;
                    case EVAL_CONT:
                        continue;
                    case EVAL_FALL_THRU:
                        MY_ASSERT_NOTREACH();
                        break;
                }
                break;
again_state_1003:
                switch(eval_if(st,EVAL_JUMP_1003)) {
                    case EVAL_AGAIN:
                        goto again;
                    case EVAL_RETURN:
                        return;
                    case EVAL_BREAK:
                        break;
                    case EVAL_CONT:
                        continue;
                    case EVAL_FALL_THRU:
                        MY_ASSERT_NOTREACH();
                        break;
                }
                break;
            case CMD_SWITCH:
                switch(eval_switch(st,EVAL_JUMP_NONE)) {
                    case EVAL_AGAIN:
                        goto again;
                    case EVAL_RETURN:
                        return;
                    case EVAL_BREAK:
                        break;
                    case EVAL_CONT:
                        continue;
                    case EVAL_FALL_THRU:
                        MY_ASSERT_NOTREACH();
                        break;
                }
                break;
again_state_1004:
                switch(eval_switch(st,EVAL_JUMP_1004)) {
                    case EVAL_AGAIN:
                        goto again;
                    case EVAL_RETURN:
                        return;
                    case EVAL_BREAK:
                        break;
                    case EVAL_CONT:
                        continue;
                    case EVAL_FALL_THRU:
                        MY_ASSERT_NOTREACH();
                        break;
                }
                break;
again_state_1005:
                switch(eval_switch(st,EVAL_JUMP_1005)) {
                    case EVAL_AGAIN:
                        goto again;
                    case EVAL_RETURN:
                        return;
                    case EVAL_BREAK:
                        break;
                    case EVAL_CONT:
                        continue;
                    case EVAL_FALL_THRU:
                        MY_ASSERT_NOTREACH();
                        break;
                }
                break;
            case CMD_GOTO:
                {
                    st.top().loop_was_broken=false;
                    st.top().loop_was_continued=false;
#if defined(_MSC_VER)
#   pragma warning(push)
#   pragma warning(disable:4702)    // warning C4702: unreachable code
#endif
                    FUNC_CALL_CAST(c_fn_cmd_goto,st.top().i1_fn,(*st.top().i1_fn).second->get_debug_string());
#if defined(_MSC_VER)
#   pragma warning(pop)
#endif
                    st.top().id=c->get_label_id();
                    goto again;
                }
                break;
            case CMD_GLOBAL:
                {
                    microcode_perform_updates(st);
                    st.top().loop_was_broken=false;
                    st.top().loop_was_continued=false;
#if defined(_MSC_VER)
#   pragma warning(push)
#   pragma warning(disable:4702)    // warning C4702: unreachable code
#endif
                    FUNC_CALL_CAST(c_fn_cmd_global,st.top().i1_fn,(*st.top().i1_fn).second->get_debug_string());
#if defined(_MSC_VER)
#   pragma warning(pop)
#endif

                    vector<string> globals=c->get_var_names();
                    vector<string>::iterator i1;
                    for(i1=globals.begin(); i1!=globals.end(); ++i1) {
                        c_var_base* g=NULL;
                        if(super_global)
                            g=super_global->get(*i1,false);
                        if(!g) {
                            g=st.top().this_module_->global->get(*i1,false);
                            // it is already dereferenced, so do NOT do this:
                            /*
                            if(g) {
                                c_var_base* gg=var_keeper->get(g->get_id(),log_fn_eval_error,st.top().this_module_);
                                if(gg && g!=gg) {
                                    c_index_base* idx=MY_NEW c_index_str(*i1);
                                    st.top().this_module_->global->set(st.top().this_module_,idx,gg,false,"global");
                                    g=gg;
                                }
                            }
                            */
                        }
                        c_var_base* orig=g;
                        if(!orig) {
                            orig=MY_NEW c_var_unset();
                            c_index_base* idx=MY_NEW c_index_str(*i1);
                            st.top().this_module_->global->set(st.top().this_module_,idx,orig,false,c->get_debug_string());
                            g=orig;
                        }
                        g=MY_NEW c_var_unset();
                        orig->set_persistent();
                        var_sym_index_t new_one=st.top().symbol_table_->get_variable_mapping()->get_var_index(*i1);
                        if(VAR_SYM_INDEX_INVALID==new_one)
                            new_one=st.top().symbol_table_->get_var_table()->add_dynamic(st.top().micro_error);
                        if(st.top().micro_error)
                            return;
                        c_var_base* test=NULL;
                        bool test_lvalue=false;
                        if(new_one!=VAR_SYM_INDEX_DISCARD)
                            if(!st.top().symbol_table_->get_var_table()->get(new_one,test,test_lvalue,st.top().micro_error,true))
                                return;
                        if(!test || !test->get_read_only()) {
                            c_var_base* save=g;
                            if(!st.top().symbol_table_->get_var_table()->set(new_one,g,true,st.top().micro_error,MICRO_NA,false,st.top().this_module_,"",c->get_debug_string()))
                                return;
                            g=save;
                        }
                        if(!st.top().symbol_table_->get_var_table()->set_lvalue(new_one,true,st.top().micro_error,MICRO_NA))
                            return;
                        (void)var_keeper->add_reference(g->get_id(),orig->get_id(),log_fn_eval_error,st.top().this_module_);

                        c_index_base* idx=MY_NEW c_index_str(*i1);
                        st.top().vars_->set(st.top().this_module_,idx,g,false,c->get_debug_string());
                    }
                }
                break;
            case CMD_BREAK:
                {
#if defined(_MSC_VER)
#   pragma warning(push)
#   pragma warning(disable:4702)    // warning C4702: unreachable code
#endif
                    FUNC_CALL_CAST(c_fn_cmd_break,st.top().i1_fn,(*st.top().i1_fn).second->get_debug_string());
#if defined(_MSC_VER)
#   pragma warning(pop)
#endif
                    st.top().loop_was_broken=true;
                    st.top().loop_was_continued=false;
                    st.top().id=c->get_begin_iteration_command_id();
                    if(c->is_for_foreach() && st.top().foreach_stack.size()) {
                        var_keeper->free_reference(st.top().foreach_stack.top().reference,log_fn_eval_error,st.top().this_module_);
                        st.top().foreach_stack.pop();
                    }
                    goto again;
                }
                break;
            case CMD_CONTINUE:
                {
                    FUNC_CALL_CAST(c_fn_cmd_continue,st.top().i1_fn,(*st.top().i1_fn).second->get_debug_string());
                    st.top().loop_was_broken=false;
                    st.top().loop_was_continued=true;
                    st.top().id=c->get_begin_iteration_command_id();
                    goto again;
                }
                break;
            case CMD_DO:
                {
                    FUNC_CALL_CAST(c_fn_cmd_do,st.top().i1_fn,(*st.top().i1_fn).second->get_debug_string());
                    if(st.top().loop_was_broken) {
                        st.top().loop_was_broken=false;
                        st.top().loop_was_continued=false;
                        st.top().id=c->get_end_loop_id();
                        goto again;
                    }
                    st.top().loop_was_broken=false;
                    st.top().loop_was_continued=false;
                }
                break;
            case CMD_WHILE:
                switch(eval_while(st,EVAL_JUMP_NONE)) {
                    case EVAL_AGAIN:
                        goto again;
                    case EVAL_RETURN:
                        return;
                    case EVAL_BREAK:
                        break;
                    case EVAL_CONT:
                        continue;
                    case EVAL_FALL_THRU:
                        MY_ASSERT_NOTREACH();
                        break;
                }
                break;
again_state_1006:
                switch(eval_while(st,EVAL_JUMP_1006)) {
                    case EVAL_AGAIN:
                        goto again;
                    case EVAL_RETURN:
                        return;
                    case EVAL_BREAK:
                        break;
                    case EVAL_CONT:
                        continue;
                    case EVAL_FALL_THRU:
                        MY_ASSERT_NOTREACH();
                        break;
                }
                break;
            case CMD_DO_WHILE:
                switch(eval_do_while(st,EVAL_JUMP_NONE)) {
                    case EVAL_AGAIN:
                        goto again;
                    case EVAL_RETURN:
                        return;
                    case EVAL_BREAK:
                        break;
                    case EVAL_CONT:
                        continue;
                    case EVAL_FALL_THRU:
                        MY_ASSERT_NOTREACH();
                        break;
                }
                break;
again_state_1007:
                switch(eval_do_while(st,EVAL_JUMP_1007)) {
                    case EVAL_AGAIN:
                        goto again;
                    case EVAL_RETURN:
                        return;
                    case EVAL_BREAK:
                        break;
                    case EVAL_CONT:
                        continue;
                    case EVAL_FALL_THRU:
                        MY_ASSERT_NOTREACH();
                        break;
                }
                break;
            case CMD_FOREACH:
                switch(eval_foreach(st,EVAL_JUMP_NONE)) {
                    case EVAL_AGAIN:
                        goto again;
                    case EVAL_RETURN:
                        return;
                    case EVAL_BREAK:
                        break;
                    case EVAL_CONT:
                        continue;
                    case EVAL_FALL_THRU:
                        goto foreach_fall_thru;
                }
                break;
again_state_1008:
                switch(eval_foreach(st,EVAL_JUMP_1008)) {
                    case EVAL_AGAIN:
                        goto again;
                    case EVAL_RETURN:
                        return;
                    case EVAL_BREAK:
                        break;
                    case EVAL_CONT:
                        continue;
                    case EVAL_FALL_THRU:
                        goto foreach_fall_thru;
                }
            case CMD_FOREACH_ITER:
foreach_fall_thru:
                switch(eval_foreach_iter(st)) {
                    case EVAL_AGAIN:
                        goto again;
                    case EVAL_RETURN:
                        return;
                    case EVAL_BREAK:
                        break;
                    case EVAL_CONT:
                        continue;
                    case EVAL_FALL_THRU:
                        MY_ASSERT_NOTREACH();
                        break;
                }
                break;
            case CMD_EOF:
                {
                    microcode_perform_updates(st);
#if defined(_MSC_VER)
#   pragma warning(push)
#   pragma warning(disable:4702)    // warning C4702: unreachable code
#endif
                    //FUNC_CALL_CAST(c_fn_cmd_end_of_func,st.top().i1_fn,(*st.top().i1_fn).second->get_debug_string());
#if defined(_MSC_VER)
#   pragma warning(pop)
#endif
                    st.top().loop_was_broken=false;
                    st.top().loop_was_continued=false;
                    goto end;
                }
                break;
            default:
                {
                    string error_msg="Internal error: unimplemented class (";
                    error_msg+=typeid(*(*st.top().i1_fn).second).name();
                    error_msg+="); ";
                    error_msg+=(*st.top().i1_fn).second->get_debug_string();
                    if(st.top().this_module_ && st.top().this_module_->get_logger())
                        st.top().this_module_->get_logger()->log_print1(E_LOG_DEBUG,LOG_SCOPE_MODULE,"",GET_STR_DEBUG_INFO_EX(error_msg));
                    st.top().func_call_ok=false;
                    return;
                }
        }
    }
end:
    microcode_perform_updates(st);
    st.top().func_call_ok=true;
    st.top().is_return_=true;
    st.top().processed=true;
    return;
}

#if defined(_MSC_VER)
#   pragma warning(pop)
#endif

