/***************************************************************************
                          c_oper_desc.cpp  -  description
                             -------------------
    begin                : Thu May 10 2007
    copyright            : (C) 2004-2007 by Marian VooDooMan Meravy
    email                : ghostvoodooman -a.T= users (dot) sourceforge.net
 ***************************************************************************/

/* $Id: c_oper_desc.cpp,v 1.12 2008/03/11 21:11:38 ghostvoodooman Exp $ */

/***************************************************************************

VooDoo cIRCle - an IRC (ro)bot
Copyright (C) 2004-2008 by Marian VooDooMan Meravy
  (ghostvoodooman -a.T= users (dot) sourceforge.net)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************/

/*!
    \file
    \brief Class 'c_oper_desc'
    \version $Id: c_oper_desc.cpp,v 1.12 2008/03/11 21:11:38 ghostvoodooman Exp $
*/

#if defined(HAVE_CONFIG_H)
#   include <config.h>
#endif

#include <string>
#include <vector>
#include <stack>
#include <list>
#include <map>
#include <algorithm>

#include "../../utils/memleak.h"
#include "../../params.h"

#include "../../compat.h"      // for __my_func__

#include "c_oper_desc.h"

#include "../../utils/precomp.h"

#if defined(_MSC_VER) || defined(__BORLANDC__)
#   pragma hdrstop
#endif

c_oper_desc g_oper_precedence[] = {  //!< Operator precedence table
    //           .precedence     .oper       .associativity      .type
    c_oper_desc( 1,              "/i",       EVAL_NON_ASSOC,     OPER_UNARY),

    c_oper_desc( 2,              "(",        EVAL_LEFT,          OPER_UNARY),

    c_oper_desc( 3,              "[",        EVAL_LEFT,          OPER_UNARY),

    c_oper_desc( 4,              "--",       EVAL_NON_ASSOC,     OPER_UNARY),
    c_oper_desc( 4,              "++",       EVAL_NON_ASSOC,     OPER_UNARY),

    c_oper_desc( 5,              "~",        EVAL_RIGHT,         OPER_UNARY),
    c_oper_desc( 5,              "(int)",    EVAL_RIGHT,         OPER_UNARY),
    c_oper_desc( 5,              "(string)", EVAL_RIGHT,         OPER_UNARY),
    c_oper_desc( 5,              "(bool)",   EVAL_RIGHT,         OPER_UNARY),

    c_oper_desc( 6,              "!",        EVAL_RIGHT,         OPER_UNARY),

    // 7 are unary "+" and "-"

    c_oper_desc( 8,              "*",        EVAL_LEFT,          OPER_BINARY),
    c_oper_desc( 8,              "/",        EVAL_LEFT,          OPER_BINARY),
    c_oper_desc( 8,              "%",        EVAL_LEFT,          OPER_BINARY),

    c_oper_desc( 9,              "+",        EVAL_LEFT,          OPER_BINARY_OR_UNARY),
    c_oper_desc( 9,              "-",        EVAL_LEFT,          OPER_BINARY_OR_UNARY),
    c_oper_desc( 9,              ".",        EVAL_LEFT,          OPER_BINARY),

    c_oper_desc( 10,             "<<",       EVAL_LEFT,          OPER_BINARY),
    c_oper_desc( 10,             ">>",       EVAL_LEFT,          OPER_BINARY),

    c_oper_desc( 11,             "<",        EVAL_NON_ASSOC,     OPER_BINARY),
    c_oper_desc( 11,             "<=",       EVAL_NON_ASSOC,     OPER_BINARY),
    c_oper_desc( 11,             ">",        EVAL_NON_ASSOC,     OPER_BINARY),
    c_oper_desc( 11,             ">=",       EVAL_NON_ASSOC,     OPER_BINARY),

    c_oper_desc( 12,             "==",       EVAL_NON_ASSOC,     OPER_BINARY),
    c_oper_desc( 12,             "!=",       EVAL_NON_ASSOC,     OPER_BINARY),
    c_oper_desc( 12,             "===",      EVAL_NON_ASSOC,     OPER_BINARY),
    c_oper_desc( 12,             "!==",      EVAL_NON_ASSOC,     OPER_BINARY),

    c_oper_desc( 13,             "&",        EVAL_LEFT,          OPER_BINARY),

    c_oper_desc( 14,             "^",        EVAL_LEFT,          OPER_BINARY),

    c_oper_desc( 15,             "|",        EVAL_LEFT,          OPER_BINARY),

    c_oper_desc( 16,             "&&",       EVAL_LEFT,          OPER_BINARY),

    c_oper_desc( 17,             "||",       EVAL_LEFT,          OPER_BINARY),

    c_oper_desc( 18,             "?",        EVAL_LEFT,          OPER_TERNARY),
    c_oper_desc( 18,             ":",        EVAL_LEFT,          OPER_TERNARY_FALSE), // helper

    c_oper_desc( 19,             "=",        EVAL_RIGHT,         OPER_BINARY),
    c_oper_desc( 19,             "+=",       EVAL_RIGHT,         OPER_BINARY),
    c_oper_desc( 19,             "-=",       EVAL_RIGHT,         OPER_BINARY),
    c_oper_desc( 19,             "*=",       EVAL_RIGHT,         OPER_BINARY),
    c_oper_desc( 19,             "/=",       EVAL_RIGHT,         OPER_BINARY),
    c_oper_desc( 19,             ".=",       EVAL_RIGHT,         OPER_BINARY),
    c_oper_desc( 19,             "%=",       EVAL_RIGHT,         OPER_BINARY),
    c_oper_desc( 19,             "&=",       EVAL_RIGHT,         OPER_BINARY),
    c_oper_desc( 19,             "|=",       EVAL_RIGHT,         OPER_BINARY),
    c_oper_desc( 19,             "^=",       EVAL_RIGHT,         OPER_BINARY),
    c_oper_desc( 19,             "<<=",      EVAL_RIGHT,         OPER_BINARY),
    c_oper_desc( 19,             ">>=",      EVAL_RIGHT,         OPER_BINARY),
    c_oper_desc( 19,             "~=",       EVAL_RIGHT,         OPER_BINARY),

    c_oper_desc( 20,             "and",      EVAL_LEFT,          OPER_BINARY),

    c_oper_desc( 21,             "xor",      EVAL_LEFT,          OPER_BINARY),

    c_oper_desc( 22,             "or",       EVAL_LEFT,          OPER_BINARY),

    c_oper_desc( 23,             "=>",       EVAL_NON_ASSOC,     OPER_BINARY),

    c_oper_desc( 24,             ",",        EVAL_LEFT,          OPER_BINARY),

    c_oper_desc( 25,             "return",   EVAL_RIGHT,         OPER_UNARY),

    // end of array marker
    c_oper_desc( -1,             "",         EVAL_INVALID,       OPER_UNARY),
};

size_t g_max_oper_len=0;
static int g_max_precedence=0;

void c_oper_desc::assign_oper_len()
{
    oper_len=strlen(oper);

    if(get_prec()>g_max_precedence)
        g_max_precedence=get_prec();
}

int c_oper_desc::get_max_prec()
{
    return g_max_precedence;
}

e_assoc c_oper_desc::get_assoc_from_precedece(int prec)
{
    if(OPER_VAR==prec)              // magick: variable/function call
        return EVAL_NON_ASSOC;
    if(OPER_PARENTHESIS==prec)      // magick: parenthesis
        return EVAL_LEFT;
    if(7==prec)                     // magick: unary "=" and "-"
        return EVAL_RIGHT;

    static map<int,e_assoc> assoc;

    if(assoc.empty()) {

        for(int i1=0; g_oper_precedence[i1].precedence!=-1; ++i1) {
            if(assoc.find(g_oper_precedence[i1].precedence)==assoc.end())
                assoc.insert(make_pair(g_oper_precedence[i1].precedence,g_oper_precedence[i1].associativity));
        }

        return get_assoc_from_precedece(prec);
    }

    map<int,e_assoc>::iterator i2=assoc.find(prec);
    MY_ASSERT(i2!=assoc.end());
    return (*i2).second;
}

