/***************************************************************************
                          c_oper_desc.h  -  description
                             -------------------
    begin                : Thu May 10 2007
    copyright            : (C) 2004-2007 by Marian VooDooMan Meravy
    email                : ghostvoodooman -a.T= users (dot) sourceforge.net
 ***************************************************************************/

/* $Id: c_oper_desc.h,v 1.13 2008/08/23 15:10:43 ghostvoodooman Exp $ */

/***************************************************************************

VooDoo cIRCle - an IRC (ro)bot
Copyright (C) 2004-2008 by Marian VooDooMan Meravy
  (ghostvoodooman -a.T= users (dot) sourceforge.net)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************/

/*!
    \file
    \brief Class 'c_oper_desc'
    \version $Id: c_oper_desc.h,v 1.13 2008/08/23 15:10:43 ghostvoodooman Exp $
*/

#ifndef C_OPER_DESC_H__
#define C_OPER_DESC_H__

#if defined(HAVE_CONFIG_H)
#   include <config.h>
#endif

#include <string>
#include <vector>
#include <stack>
#include <list>
#include <map>
#include <algorithm>

using namespace std;

#include "../../utils/memleak.h"

#include "../../params.h"

#include "oper_assoc.h"

enum
{
    OPER_VAR = 26,
    OPER_PARENTHESIS,
    OPER_LAST_PREC,
};

enum e_oper_type
{
    OPER_UNARY = 1,
    OPER_BINARY,
    OPER_TERNARY,
    OPER_TERNARY_FALSE,
    OPER_BINARY_OR_UNARY,
};

#if defined(_MSC_VER) && _MSC_VER>=1400
#   pragma warning( push )

#   pragma warning( disable : 4510 )    // warning C4510: 'c_oper_desc' : default constructor could not be generated
#   pragma warning( disable : 4512 )    // warning C4512: 'c_oper_desc' : assignment operator could not be generated
#   pragma warning( disable : 4610 )    // warning C4610: struct 'c_oper_desc' can never be instantiated - user defined constructor required
#endif

/*!
    \brief Used for table of precedence of operators
    \author VooDooMan
    \version 1
    \date 2005
*/
struct c_oper_desc
{
    const int precedence;                       //!< Numeric precedence
    const char* oper;                           //!< Operator as text
    const e_assoc associativity;                //!< Associativity
    const e_oper_type type;                     //!< Type of operator
    size_t oper_len;                            //!< Lexical length of operator

    c_oper_desc(int precedence_, const char* oper_, e_assoc associativity_, e_oper_type type_)
        : precedence(precedence_)
        , oper(oper_)
        , associativity(associativity_)
        , type(type_)
    { ; }

    void assign_oper_len();                     //!< Initializes length

    MY_FORCE_INLINE e_assoc get_assoc() const
    {
        return associativity;
    }

    MY_FORCE_INLINE int get_prec() const
    {
        return precedence;
    }

    MY_FORCE_INLINE size_t get_len() const
    {
        return oper_len;
    }

    MY_FORCE_INLINE e_oper_type get_type() const
    {
        return type;
    }

    static int get_max_prec();
    static e_assoc get_assoc_from_precedece(int prec);
};

#if defined(_MSC_VER) && _MSC_VER>=1400
#   pragma warning( pop )
#endif

#endif

