#if defined(HAVE_CONFIG_H)
#   include <config.h>
#endif

#include <string>
#include <vector>
#include <stack>
#include <list>
#include <map>
#include <algorithm>

#include "../../utils/memleak.h"

#include "../../params.h"

#include "c_tree_builder.h"
#include "c_oper_desc.h"
#include "oper_prec.h"

#include "../../utils/precomp.h"

#if defined(_MSC_VER) || defined(__BORLANDC__)
#   pragma hdrstop
#endif

c_tree_builder::c_tree_builder(const string &expression)
{
    bool empty_expression_only_semicolon=false;
    {
        string e=expression;
        while(!e.empty() && '\x20'==e[0])
            e.erase(0,1);
        while(!e.empty() && '\x20'==e[e.length()-1])
            e.erase(e.length()-1,1);

        if(!e.compare(";") || e.empty()) {
            empty_expression_only_semicolon=true;
        }
        if(!e.empty() && ';'==e[e.length()-1])
            e.erase(e.length()-1,1);

        init(e);
    }

    tree<c_sub_exp*>::iterator cur=m_exp_tree.set_head(root = MY_NEW c_sub_exp_root());
    m_parenthesis.push_back(c_parenthesis('R',cur));
    if(!add_next_and_validate(root,1))
        return;

    if(empty_expression_only_semicolon) {
        c_sub_exp* el;
        m_exp_tree.insert(m_exp_tree.end(),el=MY_NEW c_sub_exp_root_end_dummy());
        if(!add_next_and_validate(el,1)) // to force emit warning
            return;
        return;
    }

    enum
    {
        VAR_PREC = 26,
    };

    char in_str=0;
    bool in_str_backslash=false;

    string::size_type sz=m_exp.length();
    string::size_type i1;
    for(i1=0; i1<sz; ++i1) {
        if('"'==m_exp[i1]) {
            if('"'==in_str && (0==in_str || '"'==in_str)) {
                if(m_parenthesis.size()>1) {
                    if(!in_str_backslash) {
                        char ch=m_parenthesis.back().m_type;
                        while('"'!=ch) {
                            pop(cur);
                            ch=m_parenthesis.back().m_type;
                        }
                        pop(cur);

                        in_str=0;

                        continue;
                    }
                } else {
                    error(E_TBE_PARETHESIS,i1);
                    return;
                }
            } else {
                c_sub_exp_eval_str* el;
                m_parenthesis.push_back(c_parenthesis('"',cur=m_exp_tree.append_child(cur,el=MY_NEW c_sub_exp_eval_str())));
                if(!add_next_and_validate(el,i1))
                    return;
                in_str='"';
                continue;
            }
            if(!in_str_backslash)
                continue;
        }
        if('\''==m_exp[i1] && (0==in_str || '\''==in_str)) {
            if('\''==in_str) {
                if(m_parenthesis.size()>1) {
                    if(!in_str_backslash) {
                        char ch=m_parenthesis.back().m_type;
                        while('\''!=ch) {
                            pop(cur);
                            ch=m_parenthesis.back().m_type;
                        }
                        pop(cur);

                        in_str=0;

                        continue;
                    }
                } else {
                    error(E_TBE_PARETHESIS,i1);
                    return;
                }
            } else {
                c_sub_exp_immediate_str* el;
                m_parenthesis.push_back(c_parenthesis('\'',cur=m_exp_tree.append_child(cur,el=MY_NEW c_sub_exp_immediate_str())));
                if(!add_next_and_validate(el,i1))
                    return;
                in_str='\'';
                continue;
            }
            if(!in_str_backslash)
                continue;
        }
        if(0!=in_str) {
            if(in_str_backslash) {
                in_str_backslash=false;
                if('"'==in_str) {
                    static_cast<c_sub_exp_immediate_str*>(*cur)->append('\\');
                    static_cast<c_sub_exp_immediate_str*>(*cur)->append(m_exp[i1]);
                    continue;
                } else {    // "'"
                    static_cast<c_sub_exp_immediate_str*>(*cur)->append('\\');
                    static_cast<c_sub_exp_immediate_str*>(*cur)->append(m_exp[i1]);
                    continue;
                }
            }
            if('\\'==m_exp[i1]) {
                in_str_backslash=true;
                continue;
            }
            if('"'==in_str) {
                static_cast<c_sub_exp_eval_str*>(*cur)->append(m_exp[i1]);
                if('$'==m_exp[i1])
                    dynamic_cast<c_sub_exp_eval_str*>(*(m_parenthesis.back().m_cur))->have_dollar();
            } else    // "'"
                static_cast<c_sub_exp_immediate_str*>(*cur)->append(m_exp[i1]);
            continue;
        }

        // from now on, we are not in the string

        if(';'==m_exp[i1])
            break;

        if(' '==m_exp[i1])
            continue;

        if('('==m_exp[i1]) {
            bool fall_down=false;
            {
                string::size_type save=i1;
                int oper_idx;
                if(get_oper(m_exp,i1,oper_idx)) {
                    if(g_oper_precedence[oper_idx].get_prec()==5) {
                        // "~", "{int}", "(string)", or "(bool)"
                        fall_down=true;
                    }
                }
                i1=save;
            }
            if(!fall_down) {
                c_sub_exp* el=MY_NEW c_sub_exp_parenthesis();
                if(!add_next_and_validate(el,i1))
                    return;

                cur=m_exp_tree.append_child(cur,el);

                m_parenthesis.push_back(c_parenthesis('(',cur));

                m_parenthesis.back().m_parent=el;

                continue;
            }
        }
        if(')'==m_exp[i1]) {
            if(m_parenthesis.size()>1) {
                while(m_parenthesis.size()>1 && m_parenthesis.back().m_type!='f' && m_parenthesis.back().m_type!='(') {
                    char ch=m_parenthesis.back().m_type;
                    /*while('o'==ch) {
                        pop(cur);
                        ch=m_parenthesis.back().m_type;
                    }*/
                    while('C'==ch) {
                        commit_conditionals(cur);
                        ch=m_parenthesis.back().m_type;
                    }
                    while('c'==ch) {
                        commit_conditionals(cur);
                        ch=m_parenthesis.back().m_type;
                    }
                    while(':'==ch) {
                        // auto unroll
                        c_sub_exp_ternary_false_end_marker* m=MY_NEW c_sub_exp_ternary_false_end_marker();
                        if(!add_next_and_validate(m,i1))
                            return;
                        m->set_true(m_parenthesis.back().m_ternary);
                        m_exp_tree.append_child(cur,m);
                        m_parenthesis.back().m_ternary_true->set_jump(m);

                        pop(cur);
                        ch=m_parenthesis.back().m_type;
                    }
                    if('?'==ch) {
                        error(E_TBE_MISPLACED_OPERATOR,i1);
                        return;
                    }
                }

                if(m_parenthesis.size()<=1) {
                    error(E_TBE_PARETHESIS,i1);
                    return;
                }

                c_sub_exp_parenthesis* inner_p=dynamic_cast<c_sub_exp_parenthesis*>(*(m_parenthesis.back().m_cur));
                c_sub_exp_func* inner_f=dynamic_cast<c_sub_exp_func*>(*(m_parenthesis.back().m_cur));
                pop(cur);

                if(inner_p) {
                    c_sub_exp* el=MY_NEW c_sub_exp_end_parenthesis();
                    m_extra.push_back(el);
                    if(!add_next_and_validate(el,i1))
                        return;
                }
                if(inner_f) {
                    c_sub_exp* el=MY_NEW c_sub_exp_func_end_args();
                    m_extra.push_back(el);
                    if(!add_next_and_validate(el,i1))
                        return;
                }

                c_sub_exp_parenthesis* p_outer=dynamic_cast<c_sub_exp_parenthesis*>(*(m_parenthesis.back().m_cur));
                if(p_outer && inner_p)
                    p_outer->add_sub(inner_p);
                if(p_outer && inner_f)
                    p_outer->add_sub(inner_f);

                continue;
            } else {
                error(E_TBE_PARETHESIS,i1);
                return;
            }
        }
        if('['==m_exp[i1]) {
            c_sub_exp* el=MY_NEW c_sub_exp_index();
            if(!add_next_and_validate(el,i1))
                return;

            cur=m_exp_tree.append_child(cur,el);

            m_parenthesis.push_back(c_parenthesis('[',cur));

            m_parenthesis.back().m_parent=el;

            continue;
        }
        if(']'==m_exp[i1]) {
            if(m_parenthesis.size()>1) {
                while(m_parenthesis.size()>1 && m_parenthesis.back().m_type!='[') {
                    char ch=m_parenthesis.back().m_type;
                    /*while('o'==ch) {
                        pop(cur);
                        ch=m_parenthesis.back().m_type;
                    }*/
                    while('C'==ch) {
                        commit_conditionals(cur);
                        ch=m_parenthesis.back().m_type;
                    }
                    while('c'==ch) {
                        commit_conditionals(cur);
                        ch=m_parenthesis.back().m_type;
                    }
                    while(':'==ch) {
                        // auto unroll
                        c_sub_exp_ternary_false_end_marker* m=MY_NEW c_sub_exp_ternary_false_end_marker();
                        if(!add_next_and_validate(m,i1))
                            return;
                        m->set_true(m_parenthesis.back().m_ternary);
                        m_exp_tree.append_child(cur,m);
                        m_parenthesis.back().m_ternary_true->set_jump(m);

                        pop(cur);
                        ch=m_parenthesis.back().m_type;
                    }
                    if('?'==ch) {
                        error(E_TBE_MISPLACED_OPERATOR,i1);
                        return;
                    }
                }

                if(m_parenthesis.size()<=1) {
                    error(E_TBE_PARETHESIS,i1);
                    return;
                }

                {
                    c_sub_exp* el=MY_NEW c_sub_exp_end_index();
                    m_extra.push_back(el);
                    if(!add_next_and_validate(el,i1))
                        return;
                }

                pop(cur);

                continue;
            } else {
                error(E_TBE_PARETHESIS,i1);
                return;
            }
        }
        if('$'==m_exp[i1]) {
            ostringstream varname;
            varname << '$';
            string::size_type i2;
            for(i2=i1+1; i2<sz; ++i2) {
                if(
                    (m_exp[i2]>='a' && m_exp[i2]<='z') ||
                    (m_exp[i2]>='A' && m_exp[i2]<='Z') ||
                    (m_exp[i2]>='0' && m_exp[i2]<='9') ||
                    ('_'==m_exp[i2])
                    ) {
                        varname << m_exp[i2];
                } else
                    break;
            }
            if(varname.str().length()==0) {
                error(E_TBE_BAD_VAR_NAME,i2);
                return;
            }

            bool ternary=false;
            if(m_parenthesis.size()>1) {
                char ch=m_parenthesis.back().m_type;
                if('?'==ch || ':'==ch)
                    ternary=true;
            }

            if(!ternary) {
                c_sub_exp* el=MY_NEW c_sub_exp_variable(varname.str());
                if(!add_next_and_validate(el,i2))
                    return;
                i1=i2-1;

                m_exp_tree.append_child(cur,el);
            } else {
                c_sub_exp* el=MY_NEW c_sub_exp_variable(varname.str());
                if(!add_next_and_validate(el,i2))
                    return;
                i1=i2-1;

                m_exp_tree.append_child(cur,el);
            }

            continue;
        }
        if('?'==m_exp[i1]) {
            int oper_idx;
            if(get_oper(m_exp,i1,oper_idx)) {
                c_sub_exp_oper_ternary* el=MY_NEW c_sub_exp_oper_ternary(oper_idx);
                if(!add_next_and_validate(el,i1))
                    return;

                m_exp_tree.append_child(cur,el);
                c_sub_exp_oper_ternary_true* el2=MY_NEW c_sub_exp_oper_ternary_true();
                if(!add_next_and_validate(el2,i1))
                    return;
                m_parenthesis.push_back(c_parenthesis('?',cur=m_exp_tree.append_child(cur,el2)));

                m_parenthesis.back().m_ternary=el;

                --i1;
                continue;
            }
        }
        if(':'==m_exp[i1]) {
            c_sub_exp_oper_ternary* ternary=NULL;
            c_sub_exp_ternary_true_end_marker* ternary_true=NULL;
            c_sub_exp_ternary_false_end_marker* ternary_jump=NULL;

            int oper_idx;
            if(get_oper(m_exp,i1,oper_idx)) {

                if(m_parenthesis.size()>1) {
                    char ch=m_parenthesis.back().m_type;
                    if('?'==ch) {
                        // unroll

                        ternary=m_parenthesis.back().m_ternary;
                        ternary_true=MY_NEW c_sub_exp_ternary_true_end_marker();
                        if(!add_next_and_validate(ternary_true,i1))
                            return;
                        ternary_true->set_true(ternary);
                        ternary_jump=m_parenthesis.back().m_ternary_jump;

                        m_exp_tree.append_child(cur,ternary_true);

                        pop(cur);
                        ch=m_parenthesis.back().m_type;
                    } else {
                        error(E_TBE_MISPLACED_OPERATOR,i1);
                        return;
                    }
                }

                c_sub_exp* el=MY_NEW c_sub_exp_oper_ternary_false(oper_idx);
                if(!add_next_and_validate(el,i1))
                    return;
                m_parenthesis.push_back(c_parenthesis(':',cur=m_exp_tree.append_child(cur,el)));

                m_parenthesis.back().m_ternary=ternary;
                m_parenthesis.back().m_ternary_true=ternary_true;
                m_parenthesis.back().m_ternary_jump=ternary_jump;

                --i1;
                continue;
            }
        }
        {
            int oper_idx;
            if(get_oper(m_exp,i1,oper_idx)) {
                size_t ternaries=0;
                char last_type=m_parenthesis.back().m_type;
                if(','==g_oper_precedence[oper_idx].oper[0] && ':'==last_type) {
                    size_t i1;
                    for(i1=m_parenthesis.size(); i1>0; --i1) {
                        if(':'==m_parenthesis[i1-1].m_type) {
                            ++ternaries;
                            continue;
                        }
                        break;
                    }
                    if(i1>0 && 'f'==m_parenthesis[i1-1].m_type)
                        last_type='f';
                }
                if('f'==last_type && ','==g_oper_precedence[oper_idx].oper[0]) {
                    // function argument separator, rather than binary operator

                    char ch=0;
                    if(ternaries>0)
                        ch=':';
                    while(':'==ch) {
                        // auto unroll
                        c_sub_exp_ternary_false_end_marker* m=MY_NEW c_sub_exp_ternary_false_end_marker();
                        if(!add_next_and_validate(m,i1))
                            return;
                        m->set_true(m_parenthesis.back().m_ternary);
                        m_exp_tree.append_child(cur,m);
                        m_parenthesis.back().m_ternary_true->set_jump(m);

                        pop(cur);
                        ch=m_parenthesis.back().m_type;
                    }

                    c_sub_exp_func_arg_separator* el=MY_NEW c_sub_exp_func_arg_separator();
                    if(!add_next_and_validate(el,i1))
                        return;
                    tree<c_sub_exp*>::iterator i=m_exp_tree.append_child(cur,el);

                    --i1;
                    continue;
                }

                c_sub_exp_oper* el=MY_NEW c_sub_exp_oper(oper_idx);
                if(!add_next_and_validate(el,i1))
                    return;

                bool first_c=false;

                if(el->get_prec()==16 || el->get_prec()==17 || el->get_prec()==20 || el->get_prec()==22) { // "&&", "||", "and", "or"
                    if(('C'==m_parenthesis.back().m_type || 'c'==m_parenthesis.back().m_type) && m_parenthesis.back().m_last_conditional_prec<el->get_prec()) {
                        commit_conditionals(cur);
                    }
                    if(('C'!=m_parenthesis.back().m_type && 'c'!=m_parenthesis.back().m_type) || ('C'==m_parenthesis.back().m_type && 'c'==m_parenthesis.back().m_type) && m_parenthesis.back().m_last_conditional_prec>el->get_prec()) {
                        switch(el->get_prec()) {
                            case 16:
                            case 17:
                                m_parenthesis.push_back(c_parenthesis('C',m_parenthesis.back().m_cur,el->get_prec()));
                                break;
                            case 20:
                            case 22:
                                m_parenthesis.push_back(c_parenthesis('c',m_parenthesis.back().m_cur,el->get_prec()));
                                break;
                        }
                        first_c=true;
                    }
                    m_parenthesis.back().m_all_conditionals.push_back(el);
                }
                while(el->get_prec()>17 && 'C'==m_parenthesis.back().m_type) { // "||"
                    commit_conditionals(cur);
                }
                while(el->get_prec()>22 && 'c'==m_parenthesis.back().m_type) { // "or"
                    commit_conditionals(cur);
                }

                bool ternary=false;
                char ch=m_parenthesis.back().m_type;
                if('?'==ch || ':'==ch)
                    ternary=true;
                if('?'==ch && el->get_prec()==18) { // ":"
                    c_sub_exp_ternary_true_end_marker* m=MY_NEW c_sub_exp_ternary_true_end_marker();
                    if(!add_next_and_validate(m,i1))
                        return;
                    m_exp_tree.append_child(cur,m);
                    m_parenthesis.back().m_ternary_true=m;
                }

                while((':'==ch || '?'==ch) && ternary && el->get_prec()>=18) {
                    if(ternary && el->get_prec()>=18) { // 18 == ternary
                        if('?'==ch && typeid(*el)!=typeid(c_sub_exp_oper_ternary_false)) {
                            error(E_TBE_INVALID_EXPRESSION,i1);
                            return;
                        }
                        c_sub_exp_ternary_false_end_marker* m=MY_NEW c_sub_exp_ternary_false_end_marker();
                        if(!add_next_and_validate(m,i1))
                            return;
                        m->set_true(m_parenthesis.back().m_ternary);
                        m_exp_tree.append_child(cur,m);
                        m_parenthesis.back().m_ternary_true->set_jump(m);

                        pop(cur); // auto unroll

                        ch=m_parenthesis.back().m_type;
                    }
                }

                tree<c_sub_exp*>::iterator i=m_exp_tree.append_child(cur,el);

                if(first_c)
                    m_parenthesis.back().m_first_conditional=i;

                --i1;

                continue;
            }
        }
        if(m_exp[i1]>='0' && m_exp[i1]<='9') {
            // constatnt numeric immediate
            enum
            {
                DEC = 1,
                OCTAL,
                HEX,
            };
            int kind=DEC;
            if('0'==m_exp[i1])
                kind=OCTAL;
            if('x'==m_exp[i1+1] || 'X'==m_exp[i1+1])
                kind=HEX;
            ostringstream varname;
            string::size_type i2;
            if(DEC==kind) {
                for(i2=i1; i2<sz; ++i2) {
                    if(
                        (m_exp[i2]>='0' && m_exp[i2]<='9')
                        ) {
                            varname << m_exp[i2];
                    } else
                        break;
                }
                i1=i2;
            } else if(HEX==kind) {
                varname << "0x";
                for(i2=i1+2; i2<sz; ++i2) {
                    if(
                        (
                            m_exp[i2]>='0' && m_exp[i2]<='9' ||
                            m_exp[i2]>='a' && m_exp[i2]<='f' ||
                            m_exp[i2]>='A' && m_exp[i2]<='F'
                        )
                        ) {
                            varname << m_exp[i2];
                    } else
                        break;
                }
                i1=i2;
            } else {
                // OCTAL
                varname << "0";
                for(i2=i1+1; i2<sz; ++i2) {
                    if(
                        (m_exp[i2]>='0' && m_exp[i2]<='7')
                        ) {
                            varname << m_exp[i2];
                    } else
                        break;
                }
                i1=i2;
            }
            --i1;
            MY_LONG_LONG num=0;
            if(DEC==kind) {
                if(sscanf(varname.str().c_str(),"%" MY_LONG_LONG_STR "d",&num)!=1) {
                    error(E_TBE_BAD_NUMBER,i2);
                    return;
                }
            } else if(HEX==kind) {
                if(sscanf(varname.str().c_str(),"%" MY_LONG_LONG_STR "x",&num)!=1) {
                    error(E_TBE_BAD_NUMBER,i2);
                    return;
                }
            } else {
                // OCTAL
                if(sscanf(varname.str().c_str(),"%" MY_LONG_LONG_STR "o",&num)!=1) {
                    error(E_TBE_BAD_NUMBER,i2);
                    return;
                }
            }

            bool ternary=false;
            if(m_parenthesis.size()>1) {
                char ch=m_parenthesis.back().m_type;
                if('?'==ch || ':'==ch)
                    ternary=true;
            }

            if(!ternary) {
                c_sub_exp* el=MY_NEW c_sub_exp_numeric_immediate(num);
                if(!add_next_and_validate(el,i2))
                    return;
                m_exp_tree.append_child(cur,el);
            } else {
                c_sub_exp* el=MY_NEW c_sub_exp_numeric_immediate(num);
                if(!add_next_and_validate(el,i2))
                    return;
                m_exp_tree.append_child(cur,el);
            }
            continue;
        }
        {
            // function or built-in constant
            ostringstream varname;
            string::size_type i2;
            for(i2=i1; i2<sz; ++i2) {
                if(
                    (m_exp[i2]>='a' && m_exp[i2]<='z') ||
                    (m_exp[i2]>='A' && m_exp[i2]<='Z') ||
                    (m_exp[i2]>='0' && m_exp[i2]<='9') ||
                    ('_'==m_exp[i2])
                    ) {
                        varname << m_exp[i2];
                } else
                    break;
            }
            if(varname.str().length()==0) {
                error(E_TBE_BAD_VAR_NAME,i2);
                return;
            }
            bool func=true;
            i1=i2;
            if(i1+1>=sz)
                func=false;
            else {
                if(m_exp[i1+1-1]!='(')
                    func=false;
            }

            if(func) {
                c_sub_exp* el=MY_NEW c_sub_exp_func(varname.str());
                if(!add_next_and_validate(el,i2))
                    return;
                cur=m_exp_tree.append_child(cur,el);

                m_parenthesis.push_back(c_parenthesis('f',cur));
                m_parenthesis.back().m_parent=el;
                ++i1; // skip first '('
            } else {
                bool ternary=false;
                if(m_parenthesis.size()>1) {
                    char ch=m_parenthesis.back().m_type;
                    if('?'==ch || ':'==ch)
                        ternary=true;
                }

                if(!ternary) {
                    c_sub_exp* el=MY_NEW c_sub_exp_const(varname.str());
                    if(!add_next_and_validate(el,i2))
                        return;
                    m_exp_tree.append_child(cur,el);
                } else {
                    c_sub_exp* el=MY_NEW c_sub_exp_const(varname.str());
                    if(!add_next_and_validate(el,i2))
                        return;
                    m_exp_tree.append_child(cur,el);
                }
            }
            --i1;
            continue;
        }
    }
    while(m_parenthesis.size()>1) {
        char ch=m_parenthesis.back().m_type;
        /*while('o'==ch) {
            pop(cur);
            ch=m_parenthesis.back().m_type;
        }*/
        while('C'==ch) {
            commit_conditionals(cur);
            ch=m_parenthesis.back().m_type;
        }
        while('c'==ch) {
            commit_conditionals(cur);
            ch=m_parenthesis.back().m_type;
        }
        while(':'==ch) {
            // auto unroll
            c_sub_exp_ternary_false_end_marker* m=MY_NEW c_sub_exp_ternary_false_end_marker();
            if(!add_next_and_validate(m,i1))
                return;
            m->set_true(m_parenthesis.back().m_ternary);
            m_exp_tree.append_child(cur,m);
            m_parenthesis.back().m_ternary_true->set_jump(m);

            pop(cur);
            ch=m_parenthesis.back().m_type;
        }
        if('?'==ch) {
            error(E_TBE_INVALID_EXPRESSION,0);
            return;
        }
        if('('==ch) {
            error(E_TBE_PARETHESIS,sz);
            return;
        }
        if('f'==ch) {
            error(E_TBE_PARETHESIS,sz);
            return;
        }
    }
    if(m_parenthesis.size()!=1) {
        error(E_TBE_PARETHESIS,sz);
    }

    {
        c_sub_exp_root_end_dummy* el;
        m_exp_tree.insert(m_exp_tree.end(),el=MY_NEW c_sub_exp_root_end_dummy());
        if(!add_next_and_validate(el,i1))
            return;
    }
}

void c_tree_builder::error(e_compiler_error error, string::size_type column)
{
    m_error=error;
    m_e_column=column;
}

bool c_tree_builder::has_error()
{
    return m_error!=_E_TBE_NO_ERROR;
}

string c_tree_builder::get_human_error()
{
    switch(m_error) {
        case _E_TBE_NO_ERROR:
            return (string)"<no_error>";
        case E_TBE_PARETHESIS:
            return concat_column("parenthesis mismatch");
        case E_TBE_BAD_VAR_NAME:
            return concat_column("invalid variable name");
        case E_TBE_BAD_NUMBER:
            return concat_column("bad number");
        case E_TBE_MISPLACED_OPERATOR:
            return concat_column("misplaced operator");
        case E_TBE_INVALID_EXPRESSION:
            return concat_column("invalid expression");

        case E_TBE_SANITY_CHECK_FAIL:
            return m_messages.back();

        case E_TGE_INTERNAL_ERROR:
            return concat_column("internal error (please report this problem");
        case E_TGE_L_VALUE_REQUIRED:
            return concat_column("L-value required for x++, x--, ++x, or --x operators");
        case E_TGE_INVALID_EXPRESSION:
            return concat_column("invalid expression");
        case E_TGE_EXPRESSION_SANITY_CHECK_FAILED:
            return (string)"expresion sanity check failed";
        case E_TGE_INVALID_BASE_VARIABLE_FOR_SUBSCRIPT:
            return (string)"invalid base variable for '[...]' subscript";
        case E_TBE_INVALID_ESCAPE_SEQUENCE:
            return (string)"invalid escape sequence";
    }
    return (string)"<other_error__this_is_bug__please_report_this_problem>";
}

string c_tree_builder::concat_column(const string& msg)
{
    ostringstream out;
    out << msg;
    out << " at column #";
    out << m_e_column + 1;
    return out.str();
}

void c_tree_builder::init(const string& expression)
{
    m_error=_E_TBE_NO_ERROR;
    m_e_column=0;

    m_exp=expression;
}

string::size_type c_tree_builder::get_error_column()
{
    return m_e_column;
}

bool c_tree_builder::get_oper(const string& exp, string::size_type& start, int& oper_idx)
{
    if(0==g_max_oper_len) {
        int i3=0;
        while(g_oper_precedence[i3].oper[0]) {
            g_oper_precedence[i3].assign_oper_len();
            if(g_oper_precedence[i3].oper_len>g_max_oper_len)
                g_max_oper_len=g_oper_precedence[i3].oper_len;
            ++i3;
        }
    }
    string::size_type len=exp.length();

    int best_match_idx=-1;
    string::size_type best_match_adjust=0;

    int i3=0;
    while(g_oper_precedence[i3].oper[0]) { // ignoree end of array marker
        size_t loper=g_oper_precedence[i3].get_len();

        if(start+static_cast<string::size_type>(loper)<=len &&
            strncmp(
                g_oper_precedence[i3].oper,
                static_cast<const char*>(&(exp.c_str()[start])),
                loper)==0) {
                    if(loper>static_cast<size_t>(best_match_adjust)) {
                        bool bad_match=false;
                        if(
                            g_oper_precedence[i3].get_prec()==20 || // "and"
                            g_oper_precedence[i3].get_prec()==21 || // "xor"
                            g_oper_precedence[i3].get_prec()==22 || // "or"
                            g_oper_precedence[i3].get_prec()==25    // "return"
                            ) {
                                if(start+static_cast<string::size_type>(loper)+1<=len) {
                                    if(exp.c_str()[start+static_cast<string::size_type>(loper)]!=' ')
                                        bad_match=true; // because it can be e.g. "ord(", not "or "
                                }
                        }
                        if(!bad_match) {
                            best_match_idx=i3;
                            best_match_adjust=loper;
                        }
                    }
        }
        i3++;
    }
    if(-1!=best_match_idx) {
        oper_idx=best_match_idx;
        start+=best_match_adjust;
        return true;
    }
    return false;
}

void c_tree_builder::emit_human_tree(string& out)
{
    ostringstream o;

    tree<c_sub_exp*>::iterator i1;
    for(i1=m_exp_tree.begin(); i1!=m_exp_tree.end(); ++i1) {
        int depth=m_exp_tree.depth(i1);
        string human;
        human=(*i1)->emit_human_desc(static_cast<string::size_type>(depth*2));
        o << human;
        o << "\n";
    }

    out+=o.str();
}

string c_tree_builder::get_exp()
{
    return m_exp;
}

void c_tree_builder::add_element(c_sub_exp* parent, c_sub_exp* el)
{
    MY_ASSERT(parent);
    MY_ASSERT(el);

    return;
}

c_tree_builder::~c_tree_builder()
{
    destroy();
}

void c_tree_builder::commit_conditionals(tree<c_sub_exp*>::iterator& cur)
{
    c_sub_exp_oper* el1=NULL;
    c_sub_exp_const* el2=NULL;
    c_sub_exp_conditional_end_marker* marker=NULL;

    tree<c_sub_exp*>::iterator i2;

    if(!m_parenthesis.back().m_all_conditionals.empty()) {
        vector<c_sub_exp_oper*>::iterator i1=m_parenthesis.back().m_all_conditionals.end();
        --i1;
        c_sub_exp_oper* o=*i1;

        el1=MY_NEW c_sub_exp_oper(o->get_oper_index());
        m_parenthesis.back().m_all_conditionals.push_back(el1);

        el2=MY_NEW c_sub_exp_const(o->get_prec() == 16 ? "TRUE" : "FALSE"); // 16: "&&", else 17: "||"

        marker=MY_NEW c_sub_exp_conditional_end_marker();
        for(i1=m_parenthesis.back().m_all_conditionals.begin(); i1!=m_parenthesis.back().m_all_conditionals.end(); ++i1) {
            marker->add_conditional(*i1);
            (*i1)->set_all_conditionals(m_parenthesis.back().m_all_conditionals);

            vector<c_sub_exp_oper*>::iterator i3;
            for(i3=m_parenthesis.back().m_all_conditionals.begin(); i3!=m_parenthesis.back().m_all_conditionals.end(); ++i3)
                (*i1)->add_same_prec_opers(*i3);
        }
        i2=m_parenthesis.back().m_first_conditional;
    }

    if(el1) {
        m_exp_tree.append_child(cur,el1);
        add_element(*(m_parenthesis.back().m_cur),el1);
        marker->add_conditional(el1);

        m_exp_tree.append_child(cur,el2);
        add_element(*(m_parenthesis.back().m_cur),el2);
    }

    pop(cur);

    if(el1) {
        m_exp_tree.append_child(cur,marker);
        add_element(*(m_parenthesis.back().m_cur),marker);

        marker->set_return_to(*i2);
    }
}

void c_tree_builder::pop(tree<c_sub_exp*>::iterator& cur)
{
    m_parenthesis.pop_back();
    cur=m_parenthesis.back().m_cur;
}

void c_tree_builder::destroy()
{
    vector<c_sub_exp*> p;
    vector<c_sub_exp*>::iterator i2;

    tree<c_sub_exp*>::iterator i1;
    for(i1=m_exp_tree.begin(); i1!=m_exp_tree.end(); ++i1)
        p.push_back(*i1);

    for(i2=m_extra.begin(); i2!=m_extra.end(); ++i2)
        p.push_back(*i2);

    sort(p.begin(),p.end());
    i2=unique(p.begin(),p.end());
    p.erase(i2,p.end());

    for(i2=p.begin(); i2!=p.end(); ++i2)
        MY_DELETE(*i2);
}

bool c_tree_builder::add_next_and_validate(c_sub_exp* el, string::size_type col)
{
    m_extra.push_back(el); // to prevent mem leaks upon error in compilation/sanity check

    if(!m_flow.empty()) {
        const c_exp_sanity* sanity=m_flow.back()->can_follow(*el);

        m_flow.push_back(el);

        const char* msg;
        const char* file;
        int line;
        c_exp_sanity::e_can_follow rc=sanity->get(msg,file,line);

        ostringstream log;

        switch(rc) {
            case c_exp_sanity::CF_OKAY:
                return true;
            case c_exp_sanity::CF_WARN:
                // fall thru
            case c_exp_sanity::CF_USER_FATAL_ERROR:
                // fall thru
            case c_exp_sanity::CF_OOPS_BUG:
                {
                    log << msg;
                    log << " at column #";
                    log << tos(col);
                    log << "; in C++ file:line ";
                    log << file;
                    log << ":";
                    log << tos(line);

                    m_messages.push_back(log.str());

                    if(c_exp_sanity::CF_WARN==rc)
                        return true;

                    error(E_TBE_SANITY_CHECK_FAIL,col);

                    return false;
                }
        }
    } else
        m_flow.push_back(el);

    return true;
}

const vector<string>& c_tree_builder::get_warnings()
{
    return m_messages;
}

