#ifndef C_TREE_BUILDER_H__
#define C_TREE_BUILDER_H__

#if defined(HAVE_CONFIG_H)
#   include <config.h>
#endif

#include <string>
#include <vector>
#include <stack>
#include <list>
#include <map>
#include <algorithm>

#include "../../utils/memleak.h"

#include "../../params.h"

#include "oper_prec.h"
#include "c_tree_shared.h"
#include "../../tree/tree.hh"

enum e_compiler_error
{
    _E_TBE_NO_ERROR = 0,
    E_TBE_PARETHESIS,
    E_TBE_BAD_VAR_NAME,
    E_TBE_BAD_NUMBER,
    E_TBE_MISPLACED_OPERATOR,
    E_TBE_INVALID_EXPRESSION,

    E_TBE_SANITY_CHECK_FAIL,

    E_TGE_INTERNAL_ERROR,
    E_TGE_L_VALUE_REQUIRED,
    E_TGE_INVALID_EXPRESSION,
    E_TGE_EXPRESSION_SANITY_CHECK_FAILED,
    E_TGE_INVALID_BASE_VARIABLE_FOR_SUBSCRIPT,
    E_TBE_INVALID_ESCAPE_SEQUENCE,
};

class c_tree_builder
{
protected:
    string m_exp;
    tree<c_sub_exp*> m_exp_tree;
    vector<c_sub_exp*> m_flow;
    vector<c_sub_exp*> m_extra;
    e_compiler_error m_error;
    string::size_type m_e_column;

    vector<string> m_messages;

    c_sub_exp* root;

    void error(e_compiler_error error, string::size_type column);

    string concat_column(const string& msg);

    bool get_oper(const string& exp, string::size_type& start, int& oper_idx);

    void destroy();

private:
    void init(const string& expression);

    bool add_next_and_validate(c_sub_exp* el, string::size_type col);

    struct c_parenthesis
    {
        tree<c_sub_exp*>::iterator m_cur;
        char m_type;

        c_sub_exp* m_parent;
        c_sub_exp* m_left;
        bool m_was_oper;

        c_sub_exp_oper_ternary* m_ternary;
        c_sub_exp_ternary_true_end_marker* m_ternary_true;
        c_sub_exp_ternary_false_end_marker* m_ternary_jump;

        // type=='C'
        int m_last_conditional_prec; // last prec of conditional, "&&" or "||" was the last one?
        tree<c_sub_exp*>::iterator m_first_conditional;
        vector<c_sub_exp_oper*> m_all_conditionals;
        MY_FORCE_INLINE c_parenthesis(char type, tree<c_sub_exp*>::iterator cur, int last_conditional_prec)
            : m_cur(cur)
            , m_type(type)
            , m_parent(NULL)
            , m_left(NULL)
            , m_was_oper(false)
            , m_ternary(NULL)
            , m_ternary_true(NULL)
            , m_ternary_jump(NULL)
            , m_last_conditional_prec(last_conditional_prec)
        { MY_ASSERT('C'==m_type || 'c'==m_type); }

        // type== others
        MY_FORCE_INLINE c_parenthesis(char type, tree<c_sub_exp*>::iterator cur)
            : m_cur(cur)
            , m_type(type)
            , m_parent(NULL)
            , m_left(NULL)
            , m_was_oper(false)
            , m_ternary(NULL)
            , m_ternary_true(NULL)
            , m_ternary_jump(NULL)
        { MY_ASSERT('C'!=m_type); }

    };

    vector<c_parenthesis> m_parenthesis;

    void add_element(c_sub_exp* parent, c_sub_exp* el);

    void commit_conditionals(tree<c_sub_exp*>::iterator& cur);

    void pop(tree<c_sub_exp*>::iterator& cur);

public:

    c_tree_builder(const string& exp);
    virtual ~c_tree_builder();

    bool has_error();

    string get_human_error();
    string::size_type get_error_column();

    void emit_human_tree(string& out);

    string get_exp();

    const vector<string>& get_warnings();
};

#endif

