#if defined(HAVE_CONFIG_H)
#   include <config.h>
#endif

#include <string>
#include <vector>
#include <stack>
#include <list>
#include <map>
#include <algorithm>

#include "../../utils/memleak.h"

#include "../../params.h"

#include "c_tree_builder.h"
#include "c_oper_desc.h"
#include "oper_prec.h"

#include "c_tree_groupper.h"
#include "c_tree_microcode.h"

#include "../../utils/precomp.h"

#if defined(_MSC_VER) || defined(__BORLANDC__)
#   pragma hdrstop
#endif

c_tree_groupper::c_tree_groupper(const string& exp)
: c_tree_builder(exp)
{
    if(has_error())
        return;
}

void c_tree_groupper::process()
{
    m_elim_tree=m_exp_tree;

    {
        tree<c_sub_exp*>::iterator ii1;
        for(ii1=m_elim_tree.begin(); ii1!=m_elim_tree.end(); ++ii1)
            (*ii1)->set_depth(m_elim_tree.depth(ii1));
    }

    bool var_only=true;
    m_prec=OPER_VAR;

    vector<c_sub_exp*> group;

    int unsucessful_iter_cnt=0;

    int iter_cnt=0;

    for(;;) {
        tree<c_sub_exp*>::iterator first=m_elim_tree.begin();
        tree<c_sub_exp*>::iterator last=m_elim_tree.end();
        --last;

        tree<c_sub_exp*>::iterator i1;

        ++iter_cnt;

        int want_depth=-1;

        for(;;) {
            e_assoc assoc=g_oper_precedence[0].get_assoc_from_precedece(1);

        again_prec:

            bool got=false;

            tree<c_sub_exp*>::iterator this_first;
            tree<c_sub_exp*>::iterator this_last;
            if(assoc!=EVAL_RIGHT) {
                // left or non-associative
                this_first=first;
                this_last=last;
            } else {
                // right
                this_first=last;
                this_last=first;
            }

            int last_prec=m_prec;

            for(i1=this_first; i1!=this_last; (assoc!=EVAL_RIGHT ? ++i1 : --i1)) {

        process_again:

                if(last_prec!=m_prec) {
                    last_prec=m_prec;
                    assoc=g_oper_precedence[0].get_assoc_from_precedece(m_prec);
                    goto again_prec;
                }

                if(!m_to_process.empty() && m_to_process.back()!=*i1)
                    continue;

                if(-1==m_prec) {
                    c_sub_exp* o=NULL;
                    for(; i1!=m_elim_tree.end(); ++i1) {
                        o=dynamic_cast<c_sub_exp_oper*>(*i1);
                        if(!o) {
                            o=dynamic_cast<c_sub_exp_index*>(*i1);
                            if(!o) {
                                o=dynamic_cast<c_sub_exp_func*>(*i1);
                                if(!o) {
                                    continue;
                                }
                            }
                        }
                        if(o) {
                            if((*i1)->complete(m_elim_tree)) {
                                if((*i1)->get_prec()==25) {
                                    if(!dynamic_cast<c_sub_exp_oper*>(*i1)->get_return_complete())
                                        break;
                                }
                                continue;
                            }
                            break;
                        }
                    }
                    if(m_elim_tree.end()==i1)
                        break;
                    m_prec=o->get_prec();
                    last_prec=m_prec;
                    assoc=g_oper_precedence[0].get_assoc_from_precedece(m_prec);
                    goto again_prec;
                }

                if(!m_elim_tree.is_valid(i1))
                    break;

                if((*i1)->get_prec()!=m_prec && m_prec!=7)
                    continue;

                if(-1!=want_depth && (*i1)->get_depth()<want_depth)
                    continue;

                const type_info& ti=typeid(**i1);

                if((*i1)->complete(m_elim_tree) && m_elim_tree.number_of_children(i1)==0) {
                    if((*i1)->get_prec()==25) {
                        if(dynamic_cast<c_sub_exp_oper*>(*i1)->get_return_complete())
                            continue;
                    } else
                        continue;
                }

                vector<c_sub_exp*> old_to_process=m_to_process;
                m_to_process.clear();

                if(7==m_prec) {
                    if(typeid(c_sub_exp_oper)==ti) { // unary operators
                        if(OPER_BINARY_OR_UNARY==dynamic_cast<c_sub_exp_oper*>(*i1)->get_oper().type) {
                            // check for unary "+" or "-"
                            tree<c_sub_exp*>::iterator left=i1;
                            --left;
                            int pl=m_elim_tree.is_valid(left) ? (*left)->get_prec() : 0;
                            bool binary_plus_or_minus=false;
                            if(m_elim_tree.is_valid(left) && (*left)->get_prec()==9) {
                                binary_plus_or_minus= OPER_BINARY_OR_UNARY==dynamic_cast<c_sub_exp_oper*>(*left)->get_oper().type;
                                if((*left)->complete(m_elim_tree))
                                    binary_plus_or_minus=false;
                            }
                            if(pl!=OPER_VAR && pl!=OPER_PARENTHESIS && pl!=5 && pl!=4 || binary_plus_or_minus) {
                                bool unary=true;
                                if(m_elim_tree.is_valid(left) && !binary_plus_or_minus) {
                                    if((*left)->get_depth()!=(*i1)->get_depth()) {
                                        if((*left)->get_depth()>(*i1)->get_depth())
                                            unary=false;
                                    } else {
                                        if((*i1)->get_prec()==pl)
                                            unary=false;
                                        if((*left)->complete(m_elim_tree))
                                            unary=false;
                                    }
                                }
                                if(unary) {
                                    // unary
                                    c_sub_exp* parent=*i1;
                                    bool processed=process_oper_unary(i1);
                                    if(has_error())
                                        return;
                                    if(processed) {
                                        got=true;
                                        group.push_back(parent);
                                        break;
                                    } else
                                        continue;
                                }
                            }
                        }
                    }
                    if(!var_only)
                        continue;
                }

                if(typeid(c_sub_exp_dummy)==ti) {
                    continue;
                }

                if(typeid(c_sub_exp_const)==ti) {
                    if(!(*i1)->complete(m_elim_tree)) {
                        (*i1)->set_complete();
                        got=true;
                        group.push_back(*i1);
                        if(!var_only)
                            break;
                        else
                            continue;
                    }
                    continue;
                }

                if(typeid(c_sub_exp_eval_str)==ti) {
                    if(!(*i1)->complete(m_elim_tree)) {
                        (*i1)->set_complete();
                        got=true;
                        group.push_back(*i1);
                        if(!var_only)
                            break;
                        else
                            continue;
                    }
                    continue;
                }

                if(typeid(c_sub_exp_func)==ti) {
                    if(var_only)
                        continue;
                    if(!(*i1)->complete(m_elim_tree) && children_complete(i1)) {
                        (*i1)->set_complete();
                        got=true;
                        group.push_back(*i1);

                        include_function_args(i1);
                        if(has_error())
                            return;
                        break;
                    }
                    if(!m_to_process.empty())
                        goto again_prec;
                    m_prec=-1;
                    goto again_prec;
                }

                if(typeid(c_sub_exp_immediate_str)==ti) {
                    if(!(*i1)->complete(m_elim_tree)) {
                        (*i1)->set_complete();
                        got=true;
                        group.push_back(*i1);
                        if(!var_only)
                            break;
                        else
                            continue;
                    }
                    continue;
                }

                if(typeid(c_sub_exp_index)==ti) {
                    if(has_error())
                        return;
                    if(!set_index_base(i1))
                        continue;
                    if((*i1)->complete(m_elim_tree)) {
                        if(children_complete(i1)) {
                            int num=m_elim_tree.number_of_children(i1)-1;
                            tree<c_sub_exp*>::sibling_iterator right=m_elim_tree.end(); // "[]" case
                            if(num>=0)
                                right=m_elim_tree.child(i1,num);

                            (*i1)->set_complete();
                            got=true;

                            if(m_elim_tree.is_valid(right))
                                dynamic_cast<c_sub_exp_index*>(*i1)->add_index(*right);

                            tree<c_sub_exp*>::sibling_iterator i2;
                            c_sub_exp_oper* lo=NULL;
                            i2=right;
                            while(m_elim_tree.is_valid(i2)) {
                                if(typeid(**i2)==typeid(c_sub_exp_oper))
                                    lo=dynamic_cast<c_sub_exp_oper*>(*i2);
                                ++i2;
                            }
                            if(lo || m_elim_tree.is_valid(right))
                                dynamic_cast<c_sub_exp_index*>(*i1)->set_result(lo!=NULL ? lo : *right);

                            tree<c_sub_exp*>::sibling_iterator base=i1;
                            --base;
                            // if 'base' is not valid, then we are chaining indexes
                            bool valid_base=m_elim_tree.is_valid(base);
                            if(valid_base) {
                                group.push_back(*base);
                            }

                            group.push_back(*i1);
                            if(valid_base)
                                *base=*i1;
                            if(valid_base)
                                include_children(*i1,i1,base,NULL);

                            if(valid_base)
                                m_elim_tree.erase(base);

                            if(has_error())
                                return;

                            break;
                        }
                        goto process_again;
                    }
                    continue;
                }

                if(typeid(c_sub_exp_numeric_immediate)==ti) {
                    if(!(*i1)->complete(m_elim_tree)) {
                        (*i1)->set_complete();
                        got=true;
                        group.push_back(*i1);
                        if(!var_only)
                            break;
                        else
                            continue;
                    }
                    continue;
                }

                if(typeid(c_sub_exp_oper)==ti) {
                    bool do_again=false;
                    c_sub_exp* parent=*i1;
                    bool processed=false;
                    switch(dynamic_cast<c_sub_exp_oper*>(*i1)->get_oper().type) {
                        case OPER_UNARY:
                            processed=process_oper_unary(i1);
                            if(has_error())
                                return;
                            if(processed && parent->get_prec()!=1) // 1 is "/i" operator
                                group.push_back(parent);
                            break;
                        case OPER_BINARY:
                        case OPER_BINARY_OR_UNARY:
                            if((*i1)->get_prec()==9 && ('+'==dynamic_cast<c_sub_exp_oper*>(*i1)->get_oper().oper[0] || '-'==dynamic_cast<c_sub_exp_oper*>(*i1)->get_oper().oper[0]))
                                processed=process_oper_binary_or_unary(group,i1);
                            else
                                processed=process_oper_binary(group,i1);
                            if(has_error())
                                return;
                            if(processed) {
                                c_sub_exp_oper* o=dynamic_cast<c_sub_exp_oper*>(parent);
                                if(processed && o->get_oper().get_prec()!=23) // 22: "=>" operator: handled in a special way
                                    if(o->get_oper().get_prec()!=16 && o->get_oper().get_prec()!=17) // 16: "&&", 17: "||": handled in a special way
                                        group.push_back(parent);
                                if(o->get_oper().get_prec()==16 || o->get_oper().get_prec()==17) { // 16: "&&", 17: "||"
                                    // to force putting it all into one group
                                    processed=false;
                                    do_again=true;
                                }
                            }
                            break;
                        default:
                            {
                                enum {
                                    ERROR_ID_2 = 2,
                                };
                                error(E_TGE_INTERNAL_ERROR,ERROR_ID_2);
                                MY_ASSERT_NOTREACH();
                                return;
                            }
                    }
                    if(processed) {
                        got=true;
                        break;
                    }
                    if(!m_to_process.empty())
                        goto again_prec;
                    if(do_again) {
                        m_prec=-1;
                        goto again_prec;
                    }
                    continue;
                }

                if(typeid(c_sub_exp_oper_ternary)==ti) {
                    if(!(*i1)->complete(m_elim_tree)) {
                        tree<c_sub_exp*>::iterator i2;
                        for(i2=m_elim_tree.begin(); i2!=m_elim_tree.end(); ++i2)
                            if(*i2==*i1)
                                break;
                        tree<c_sub_exp*>::iterator i3=m_elim_tree.previous_sibling(i2);
                        if(!m_elim_tree.is_valid(i3)) {
                            // chained ternaries
                            i3=m_elim_tree.parent(i2);
                        }
                        if(completed(EVAL_LEFT,i2)) {
                            c_sub_exp_oper_ternary* ternary=dynamic_cast<c_sub_exp_oper_ternary*>(*i1);

                            tree<c_sub_exp*>::iterator parent=m_elim_tree.parent(i1);
                            while(typeid(**parent)==typeid(c_sub_exp_parenthesis))
                                parent=m_elim_tree.parent(parent);
                            if(typeid(**parent)==typeid(c_sub_exp_oper_ternary_false)) {
                                tree<c_sub_exp*>::sibling_iterator i4=parent;
                                --i4;
                                --i4;
                                ternary->set_other_ternary(dynamic_cast<c_sub_exp_oper_ternary*>(*i4));
                            }

                            ternary->add_condition(*i3);

                            group.push_back(*i1);
                            (*i1)->set_complete();
                            got=true;

                            tree<c_sub_exp*>::sibling_iterator left=i1;
                            --left;
                            m_elim_tree.erase(left);

                            break;
                        }
                    }

                    if(!m_to_process.empty())
                        goto again_prec;

                    continue;
                }

                if(typeid(c_sub_exp_oper_ternary_true)==ti) {
                    if(!(*i1)->complete(m_elim_tree)) {
                        if(children_complete(i1)) {
                            c_sub_exp_oper_ternary_true* tt=dynamic_cast<c_sub_exp_oper_ternary_true*>(*i1);

                            tree<c_sub_exp*>::sibling_iterator t=i1;
                            --t;

                            bool chained=false;
                            tree<c_sub_exp*>::iterator parent=m_elim_tree.parent(i1);
                            while(typeid(**parent)==typeid(c_sub_exp_parenthesis))
                                parent=m_elim_tree.parent(parent);
                            if(typeid(**parent)!=typeid(c_sub_exp_oper_ternary_false))
                                parent=i1;
                            else
                                chained=true;

                            int num=m_elim_tree.number_of_children(i1)-1-1;

                            tree<c_sub_exp*>::iterator last=m_elim_tree.child(i1,num);
                            tt->set_result(*last);

                            tree<c_sub_exp*>::sibling_iterator i2=parent;
                            --i2;
                            if(chained)
                                --i2;
                            tt->set_ternary(dynamic_cast<c_sub_exp_oper_ternary*>(*i2));

                            group.push_back(*i1);
                            (*i1)->set_complete();
                            got=true;
                            include_children(*i1,i1,m_elim_tree.end(),&group);
                            if(has_error())
                                return;

                            break;
                        }
                    }

                    if(!m_to_process.empty())
                        goto again_prec;

                    continue;
                }

                if(typeid(c_sub_exp_oper_ternary_false)==ti) {
                    if(!(*i1)->complete(m_elim_tree)) {
                        if(children_complete(i1)) {
                            group.push_back(*i1);
                            (*i1)->set_complete();
                            got=true;

                            tree<c_sub_exp*>::iterator parent=m_elim_tree.parent(i1);
                            while(typeid(**parent)==typeid(c_sub_exp_parenthesis))
                                parent=m_elim_tree.parent(parent);
                            if(typeid(**parent)!=typeid(c_sub_exp_oper_ternary_false))
                                parent=i1;

                            c_sub_exp_oper_ternary_false* tf=dynamic_cast<c_sub_exp_oper_ternary_false*>(*i1);

                            int num=m_elim_tree.number_of_children(i1)-1-1;

                            tree<c_sub_exp*>::iterator last=m_elim_tree.child(i1,num);
                            tf->set_result(*last);

                            tree<c_sub_exp*>::sibling_iterator i2=parent;
                            --i2;
                            --i2;
                            tf->set_ternary(dynamic_cast<c_sub_exp_oper_ternary*>(*i2));

                            include_children(*i1,i1,m_elim_tree.end(),&group);
                            if(has_error())
                                return;

                            i2=i1;
                            --i2; // ternary true
                            m_elim_tree.erase(i2);
                            m_elim_tree.erase(i1);

                            break;
                        }
                    }

                    if(!m_to_process.empty())
                        goto again_prec;

                    continue;
                }

                if(typeid(c_sub_exp_parenthesis)==ti) {
                    if(m_elim_tree.number_of_children(i1)==0 && dynamic_cast<c_sub_exp_parenthesis*>(*i1)->get_eval().size()!=0) {
                        want_depth=(*i1)->get_depth();
                        continue;
                    }
                    if(children_complete(i1)) {
                        tree<c_sub_exp*>::iterator right=i1;
                        ++right;

                        dynamic_cast<c_sub_exp_parenthesis*>(*i1)->set_result(*right);
                        group_children(i1,group);
                        m_elim_tree.erase_children(i1);
                        (*i1)->set_complete();
                        got=true;
                        group.push_back(*i1);
                        break;
                    }

                    if(!m_to_process.empty())
                        goto again_prec;

                    continue;
                }

                if(typeid(c_sub_exp_variable)==ti) {
                    if(!(*i1)->complete(m_elim_tree)) {
                        group.push_back(*i1);
                        (*i1)->set_complete();
                        got=true;
                        if(!var_only)
                            break;
                        else
                            continue;
                    }
                    continue;
                }

                if(typeid(c_sub_exp_contitional_node_helper)==ti) {
                    continue;
                }

                if(typeid(c_sub_exp_conditional_end_marker)==ti) {
                    if(dynamic_cast<c_sub_exp_conditional_end_marker*>(*i1)->all_complete(*this,m_elim_tree) && !(*i1)->complete(m_elim_tree)) {
                        group.push_back(*i1);
                        (*i1)->set_complete();
                        m_elim_tree.erase(i1);
                        got=true;
                        break;
                    }
                    continue;
                }

                if(typeid(c_sub_exp_root)==ti) {
                    continue;
                }

                enum {
                    ERROR_ID_1 = 1,
                };
                error(E_TGE_INTERNAL_ERROR,ERROR_ID_1);
                MY_ASSERT_NOTREACH();
                return;
            }

            if(var_only) {
                var_only=false;
                if(!group.empty()) {
                    if(!sanity_check(group)) {
                        error(E_TGE_EXPRESSION_SANITY_CHECK_FAILED,0);
                        return;
                    }
                    m_groups.push_back(group);
                    group.clear();
                }

                m_prec=-1;

                break;
            }

            if(!got && !group.empty()) {
                // conditionals group
                got=true;
            }

            if(m_elim_tree.size()==2)
                goto finished;
            for(i1=m_elim_tree.begin(); i1!=m_elim_tree.end(); ++i1) {
                if((*i1)->get_prec()==25) {
                    tree<c_sub_exp*>::sibling_iterator i2=i1;
                    ++i2;
                    if(m_elim_tree.is_valid(i2))
                        break; // not complete
                } else
                    if(!(*i1)->complete(m_elim_tree))
                        break;
            }
            if(m_elim_tree.end()==i1)
                goto finished;

            m_to_process.clear();

            if(!got) {
                if(++m_prec==OPER_LAST_PREC) {
                    if(1==unsucessful_iter_cnt) {
                        //depth_increment=0;
                        unsucessful_iter_cnt=0;
                        m_prec=-1;
                        break;
                    }
                    if(unsucessful_iter_cnt>=2) {
                        error(E_TGE_INVALID_EXPRESSION,0);
                        return;
                    }
                    m_prec=1;
                    want_depth=-1;
                }
                continue;
            } else {
                if(!group.empty()) {
                    if(!sanity_check(group)) {
                        error(E_TGE_EXPRESSION_SANITY_CHECK_FAILED,0);
                        return;
                    }
                    m_groups.push_back(group);
                    group.clear();
                }

                unsucessful_iter_cnt=0;

                m_prec=-1;
                break;
            }
            // NOTREACH
        }
    }

finished:
    if(!sanity_check(group)) {
        error(E_TGE_EXPRESSION_SANITY_CHECK_FAILED,0);
        return;
    }

    if(!group.empty())
        m_groups.push_back(group);
}

c_tree_groupper::~c_tree_groupper()
{
}

string c_tree_groupper::emit_human_evaluation()
{
    if(has_error())
        return (string)"Compile error\n";

    ostringstream o;

    int cnt=0;

    vector<vector<c_sub_exp*> >::iterator i1;
    for(i1=m_groups.begin(); i1!=m_groups.end(); ++i1) {
        o << "Group #" << ++cnt << "\n";
        vector<c_sub_exp*>::iterator i2;
        for(i2=(*i1).begin(); i2!=(*i1).end(); ++i2) {
            string out=(*i2)->emit_human_desc(2);
            o << out.c_str();
        }
    }

    return o.str();
}

int c_tree_groupper::get_deepest()
{
    int deepest=0;

    tree<c_sub_exp*>::iterator i1;
    for(i1=m_elim_tree.begin(); i1!=m_elim_tree.end(); ++i1) {
        int d=m_elim_tree.depth(i1);
        if(d>deepest)
            deepest=d;
        (*i1)->set_depth(d);
    }

    return deepest;
}

bool c_tree_groupper::process_oper_unary(const tree<c_sub_exp*>::iterator& where)
{
    c_sub_exp_oper* oper=dynamic_cast<c_sub_exp_oper*>(*where);

    if((*where)->get_prec()==4) {
        // ++ or --
        tree<c_sub_exp*>::iterator i2;
        for(i2=m_elim_tree.begin(); i2!=m_elim_tree.end(); ++i2)
            if(*i2==*where)
                break;
        tree<c_sub_exp*>::iterator left=m_elim_tree.previous_sibling(i2);
        if(!m_elim_tree.is_valid(left)) {
            // chained
            left=m_elim_tree.parent(i2);
        }

        // try post-fix
        if(m_elim_tree.is_valid(left) && (typeid(**left)==typeid(c_sub_exp_variable) || typeid(**left)==typeid(c_sub_exp_index))) {
            dynamic_cast<c_sub_exp_oper*>(*where)->set_postfix(true);

            if(!completed(EVAL_LEFT,where,true))
                return false;

            oper->add_eval(*left);
            oper->set_complete();
            tree<c_sub_exp*>::iterator i2;
            for(i2=m_elim_tree.begin(); i2!=m_elim_tree.end(); ++i2)
                if(*i2==*left) {
                    m_elim_tree.erase_children(i2);
                    m_elim_tree.erase(i2);
                    break;
                }
            return true;
        } else {
            // try pre-fix
        }

        tree<c_sub_exp*>::iterator right=m_elim_tree.next_sibling(i2);
        if(!m_elim_tree.is_valid(right)) {
            error(E_TGE_L_VALUE_REQUIRED,0);
            return false;
        }

        if(typeid(**right)==typeid(c_sub_exp_variable) || typeid(**right)==typeid(c_sub_exp_index)) {
            dynamic_cast<c_sub_exp_oper*>(*where)->set_postfix(false);

            if(!completed(EVAL_RIGHT,where,true))
                return false;

            oper->add_eval(*right);
            oper->set_complete();
            tree<c_sub_exp*>::iterator i2;
            for(i2=m_elim_tree.begin(); i2!=m_elim_tree.end(); ++i2)
                if(*i2==*right) {
                    m_elim_tree.erase_children(i2);
                    m_elim_tree.erase(i2);
                    break;
                }
            return true;
        } else {
            error(E_TGE_L_VALUE_REQUIRED,0);
            return false;
        }
    }
    if((*where)->get_prec()==1) {
        // "/i"
        tree<c_sub_exp*>::iterator i2;
        for(i2=m_elim_tree.begin(); i2!=m_elim_tree.end(); ++i2)
            if(*i2==*where)
                break;
        tree<c_sub_exp*>::iterator left=m_elim_tree.previous_sibling(i2);
        if(!m_elim_tree.is_valid(left)) {
            // chained
            left=m_elim_tree.parent(i2);
        }
        left=m_elim_tree.parent(left);

        if(m_elim_tree.is_valid(left) && typeid(**left)==typeid(c_sub_exp_index)) {
            c_sub_exp_index* idx=dynamic_cast<c_sub_exp_index*>(*left);
            idx->set_case_insensitive();

            m_elim_tree.erase(where);
            return true;
        }
        error(E_TBE_MISPLACED_OPERATOR,0);
        return false;
    }
    if((*where)->get_prec()==25) {
        // "return ..."
        tree<c_sub_exp*>::iterator i2;
        for(i2=m_elim_tree.begin(); i2!=m_elim_tree.end(); ++i2)
            if(*i2==*where)
                break;
        tree<c_sub_exp*>::iterator right=i2;
        ++right;
        if(!m_elim_tree.is_valid(right) || typeid(**right)==typeid(c_sub_exp_root_end_dummy)) {
            // okay to process, w/o argument
            dynamic_cast<c_sub_exp_oper*>(*where)->set_return_complete();
        } else {
            if(!completed(EVAL_RIGHT,where,false))
                return false;
            dynamic_cast<c_sub_exp_oper*>(*where)->set_return_complete();
        }
        for(i2=m_elim_tree.begin(); i2!=m_elim_tree.end(); ++i2) {
            if(*i2==*where) {
                right=m_elim_tree.next_sibling(where);
                if(m_elim_tree.is_valid(right))
                    dynamic_cast<c_sub_exp_oper*>(*i2)->add_eval(*right);
                m_elim_tree.erase(i2);
                break;
            }
        }
        return true;
    }

    if(!completed(EVAL_RIGHT,where,true))
        return false;

    tree<c_sub_exp*>::iterator right=m_elim_tree.next_sibling(where);
    if(!m_elim_tree.is_valid(right)) {
        error(E_TGE_INVALID_EXPRESSION,0);
        return false;
    }

    oper->add_eval(*right);

    oper->set_complete();
    if(m_elim_tree.is_valid(right)) {
        tree<c_sub_exp*>::iterator i2;
        for(i2=m_elim_tree.begin(); i2!=m_elim_tree.end(); ++i2)
            if(*i2==*right) {
                m_elim_tree.erase_children(i2);
                m_elim_tree.erase(i2);
                break;
            }
    }
    return true;
}

bool c_tree_groupper::process_oper_binary(vector<c_sub_exp*>& group, const tree<c_sub_exp*>::iterator& where)
{
    int prec=dynamic_cast<c_sub_exp_oper*>(*where)->get_prec();
    if(16==prec || 17==prec) { // "&&" or "||"
        return process_oper_conditional(group,where);
    }
    tree<c_sub_exp*>::iterator i2;
    for(i2=m_elim_tree.begin(); i2!=m_elim_tree.end(); ++i2)
        if(*i2==*where)
            break;
    tree<c_sub_exp*>::iterator left=m_elim_tree.previous_sibling(i2);

    if(!(*left)->complete(m_elim_tree)) {
        return false;
    }

    c_sub_exp* right=get_right(where,true,false);
    {
        if(!right)
            return false;
    }

    c_sub_exp_oper* oper=dynamic_cast<c_sub_exp_oper*>(*where);

    bool okay_array=false;

    if(oper->get_prec()==23) { // 23: operator "=>": handled in a special way
        if(m_elim_tree.is_valid(left)) {
            tree<c_sub_exp*>::iterator parent=m_elim_tree.parent(where);
            if(m_elim_tree.is_valid(parent)) {
                c_sub_exp_func* ar=dynamic_cast<c_sub_exp_func*>(*parent);
                if(ar) {
                    if(!ar->get_func_name().compare("array")) {
                        ar->add_arg(oper);
                        okay_array=true;
                    }
                }
            }
        }
    } else
        okay_array=true;

    if(!okay_array) {
        error(E_TBE_MISPLACED_OPERATOR,0);
        return false;
    }

    if(dynamic_cast<c_sub_exp_root*>(*left)) {
        error(E_TGE_INVALID_EXPRESSION,0);
        return false;
    }
    if(dynamic_cast<c_sub_exp_root_end_dummy*>(right)) {
        error(E_TGE_INVALID_EXPRESSION,0);
        return false;
    }

    oper->add_eval(*left);
    oper->add_eval(right);
    oper->set_complete();

    m_elim_tree.erase(left);
    for(i2=m_elim_tree.begin(); i2!=m_elim_tree.end(); ++i2)
        if(*i2==right) {
            m_elim_tree.erase(i2);
            break;
        }
    return true;
}

bool c_tree_groupper::process_oper_binary_or_unary(vector<c_sub_exp*>& group, const tree<c_sub_exp*>::iterator& where)
{
    tree<c_sub_exp*>::iterator i2;
    for(i2=m_elim_tree.begin(); i2!=m_elim_tree.end(); ++i2)
        if(*i2==*where)
            break;
    tree<c_sub_exp*>::iterator left=m_elim_tree.previous_sibling(i2);
    if(!m_elim_tree.is_valid(left)) {
        left=m_elim_tree.parent(i2);
    }
    c_sub_exp_oper* lo=dynamic_cast<c_sub_exp_oper*>(*left);
    c_sub_exp* lv=dynamic_cast<c_sub_exp_parenthesis*>(*left);
    if(!lv)
        lv=dynamic_cast<c_sub_exp_variable*>(*left);
    if(!lv)
        lv=dynamic_cast<c_sub_exp_numeric_immediate*>(*left);
    if(!lv)
        lv=dynamic_cast<c_sub_exp_const*>(*left);
    if(!lv)
        lv=dynamic_cast<c_sub_exp_func*>(*left);
    if(!lv)
        lv=dynamic_cast<c_sub_exp_eval_str*>(*left);
    if(!lv)
        lv=dynamic_cast<c_sub_exp_immediate_str*>(*left);
    if(!lv)
        lv=dynamic_cast<c_sub_exp_index*>(*left);

    enum
    {
        UNARY = 0x00,
        BINARY = 0x01,
        SURE = 0x02,
    };

    unsigned kind;
    if((*left)->get_depth()==(*where)->get_depth())
        kind=BINARY;
    else
        kind=UNARY | SURE;

    if(lo!=NULL && lo->get_oper().get_prec()>(*where)->get_prec())
        kind=UNARY | SURE;
    if(lo!=NULL && lo->get_oper().get_prec()==8) { // "*", "/", "%"
        if(lo->get_eval().size()==0)
            kind=UNARY | SURE;
        else
            kind=BINARY | SURE;
    }

    if(lv!=NULL && (lv->get_prec()==2 || lv->get_prec()==3)) { // "{"
        if(*(m_elim_tree.parent(i2))==*left)
            kind=UNARY | SURE;
        else
            kind=BINARY | SURE;
    }

    if(typeid(**left)==typeid(c_sub_exp_func_arg_separator)) {
        // it is function's argument such as 'fn('test',-2);', i.e. unary '-2'
        kind=UNARY | SURE;
    }

    if(!(kind & SURE)) {
        if(lv!=NULL)
            kind=BINARY | SURE;
    }

    if(!(kind & SURE)) {
        if(lv!=NULL && typeid(*lv)==typeid(c_sub_exp_parenthesis) && typeid(*lv)==typeid(c_sub_exp_index))
            kind=BINARY | SURE;
    }

    if(!(kind & SURE)) {
        if(lv!=NULL || lo!=NULL && lo->get_oper().get_type()==OPER_BINARY)
            kind=BINARY | SURE;
    }

    if(!(kind & SURE)) {
        if(lv!=NULL || lo!=NULL && lo->get_oper().get_type()==OPER_UNARY) {
            if(lo->get_eval().size()==0)
                kind=UNARY | SURE;
            else
                kind=BINARY | SURE;
        }
    }

    if(!(kind & SURE)) {
        if(lv!=NULL || lo!=NULL && lo->get_oper().get_type()==OPER_BINARY_OR_UNARY) {
            if(lo->get_eval().size()==0)
                kind=UNARY | SURE;
            else
                kind=BINARY | SURE;
        }
    }

    if(lo!=NULL && lo->get_prec()==25)  // "return" operator
        kind=UNARY | SURE;  // this rules over "SURE"

    if(kind & BINARY)
        return process_oper_binary(group,where);
    else
        return process_oper_unary(where);
    // NOTREACH
}

bool c_tree_groupper::process_oper_conditional(vector<c_sub_exp*>& group, const tree<c_sub_exp*>::iterator& where)
{
    tree<c_sub_exp*>::iterator i2;
    for(i2=m_elim_tree.begin(); i2!=m_elim_tree.end(); ++i2)
        if(*i2==*where)
            break;

    c_sub_exp_oper* oper=dynamic_cast<c_sub_exp_oper*>(*where);

    size_t sz=oper->get_eval().size();
    switch(sz) {
        case 0:
            {
                tree<c_sub_exp*>::iterator left=m_elim_tree.previous_sibling(i2);
                if(!m_elim_tree.is_valid(left)) {
                    left=m_elim_tree.parent(i2);
                }

                if((*left)->get_prec()==oper->get_prec()) {
                    if(!(*left)->complete(m_elim_tree))
                        return false;

                    c_sub_exp* left_arg=dynamic_cast<c_sub_exp_oper*>(*left)->get_eval()[1];

                    oper->add_eval(left_arg);
                    group.push_back(left_arg);

                    m_elim_tree.erase(left);

                    group.push_back(oper);

                    goto process_right;
                }

                if(dynamic_cast<c_sub_exp_root*>(*left)) {
                    error(E_TGE_INVALID_EXPRESSION,0);
                    return false;
                }

                if(!(*left)->complete(m_elim_tree)) {
                    m_prec=(*left)->get_prec();
                    m_to_process.push_back(*where);
                    m_to_process.push_back(*left);
                    return false;
                }

                {
                    tree<c_sub_exp*>::iterator left2=m_elim_tree.parent(left);
                    if(m_elim_tree.is_valid(left2)) {
                        c_sub_exp_oper* o2=dynamic_cast<c_sub_exp_oper*>(*left2);
                        if(o2 && OPER_UNARY==o2->get_oper().type && o2->get_oper().precedence<oper->get_prec() && !o2->complete(m_elim_tree)) {
                            return false;
                        }
                    }
                }

                oper->add_eval(*left);
                group.push_back(*left);

                m_elim_tree.erase(left);

                group.push_back(oper);
            }
            // fall thru
        case 1:
            {
            process_right:
                if(!completed(EVAL_RIGHT,where,true))
                    return 0==sz ? true : false;

                tree<c_sub_exp*>::iterator right=m_elim_tree.next_sibling(where);
                if(!m_elim_tree.is_valid(right)) {
                    error(E_TGE_INVALID_EXPRESSION,0);
                    return false;
                }

                if(!m_elim_tree.is_valid(right)) {
                    error(E_TGE_INVALID_EXPRESSION,0);
                    return false;
                }

                if(!(*right)->complete(m_elim_tree)) {
                    m_to_process.push_back(*where);
                    m_to_process.push_back(*right);

                    if(0==sz)
                        return true;    // really true! not false!
                    else
                        return false;
                }

                if(dynamic_cast<c_sub_exp_root*>(*right)) {
                    error(E_TGE_INVALID_EXPRESSION,0);
                    return false;
                }

                if(m_elim_tree.number_of_children(right)!=0) {
                    tree<c_sub_exp*>::iterator right2=m_elim_tree.child(right,0);
                    c_sub_exp_oper* o2=dynamic_cast<c_sub_exp_oper*>(*right2);
                    if(o2 && OPER_UNARY==o2->get_oper().type && o2->get_oper().precedence<oper->get_prec() && !o2->complete(m_elim_tree)) {
                        // post-fix ++ or --
                        if(0==sz)
                            return true;    // really true! not false!
                        else
                            return false;
                    }
                }

                oper->add_eval(*right);
                if(oper->get_eval().size()==2) {
                    oper->set_complete();

                    m_elim_tree.erase(right);

                    (void)completed(EVAL_RIGHT,where,false); // to force processing conditional end marker

                    return true;
                }
                if(0==sz)
                    return true;    // really true! not false!
                else
                    return false;
            }
            // NOTREACH
    }

    if(oper->get_eval().size()>=2)
        return true;

    MY_ASSERT_NOTREACH();
    return true;
}

bool c_tree_groupper::children_complete(tree<c_sub_exp*>::iterator where)
{
    if(m_elim_tree.number_of_children(where)==0)
        return true;

    tree<c_sub_exp*>::iterator i1=m_elim_tree.child(where,0);

    c_sub_exp* r;
    r=get_right(i1,false,true);
    if(!r)
        return false;
    if(r==*where)
        return true;
    if(r==*i1)
        return true;
    return false;
}

void c_tree_groupper::include_children(c_sub_exp* parent,tree<c_sub_exp*>::iterator where, tree<c_sub_exp*>::iterator protect, vector<c_sub_exp*>* group)
{
    const type_info& ti=typeid(*parent);

    c_sub_exp_oper_ternary_true* tt=NULL;
    if(typeid(c_sub_exp_oper_ternary_true)==ti)
        tt=dynamic_cast<c_sub_exp_oper_ternary_true*>(parent);

    c_sub_exp_oper_ternary_false* tf=NULL;
    if(typeid(c_sub_exp_oper_ternary_false)==ti)
        tf=dynamic_cast<c_sub_exp_oper_ternary_false*>(parent);

    vector<c_sub_exp*> to_erase;
    to_erase.push_back(*where);
    int num=m_elim_tree.number_of_children(where);
    if(0!=num) {
        tree<c_sub_exp*>::sibling_iterator i3=m_elim_tree.child(where,0);
        for(; m_elim_tree.is_valid(i3); ++i3) {
            if(tt)
                tt->add_sub(*i3);
            if(tf)
                tf->add_eval(*i3);

            if(group)
                group->push_back(*i3);

            to_erase.push_back(*i3);
        }
    }
    vector<c_sub_exp*>::iterator i1;
    tree<c_sub_exp*>::iterator i2;
    for(i1=to_erase.begin(); i1!=to_erase.end(); ++i1) {
        for(i2=m_elim_tree.begin(); i2!=m_elim_tree.end(); ++i2)
            if(*i1==*i2) {
                if(i2==protect)
                    continue;
                m_elim_tree.erase_children(i2);
                break;
            }
    }
}

bool c_tree_groupper::sanity_check(const vector<c_sub_exp*>& group)
{
    vector<c_sub_exp*>::const_iterator i1;
    for(i1=group.begin(); i1!=group.end(); ++i1) {
        if(!(*i1)->complete(m_elim_tree) && (*i1)->get_prec()!=16 && (*i1)->get_prec()!=17) // "&&", "||"
            return false;
    }
    return true;
}

bool c_tree_groupper::set_index_base(tree<c_sub_exp*>::iterator where)
{
    tree<c_sub_exp*>::iterator i2;
    for(i2=m_elim_tree.begin(); i2!=m_elim_tree.end(); ++i2)
        if(*i2==*where)
            break;
    tree<c_sub_exp*>::iterator left=m_elim_tree.previous_sibling(i2);

    {
        tree<c_sub_exp*>::iterator left2=left;
        ++left2;
        if(left2!=where && typeid(**left2)==typeid(c_sub_exp_index))
            return false;
    }

    if(!m_elim_tree.is_valid(left)) {
        // chained
        left=m_elim_tree.parent(i2);
    }

    c_sub_exp* lv=dynamic_cast<c_sub_exp_parenthesis*>(*left);
    if(lv) {
        if(!(children_complete(left) && dynamic_cast<c_sub_exp_parenthesis*>(lv)->complete(m_elim_tree))) {
            return false;
        }
    }
    if(!lv)
        lv=dynamic_cast<c_sub_exp_variable*>(*left);
    if(!lv)
        lv=dynamic_cast<c_sub_exp_func*>(*left);
    if(!lv)
        lv=dynamic_cast<c_sub_exp_eval_str*>(*left);
    if(!lv)
        lv=dynamic_cast<c_sub_exp_immediate_str*>(*left);
    if(!lv)
        lv=dynamic_cast<c_sub_exp_index*>(*left);

    if(!lv) {
        bool err=false;

        left=m_elim_tree.previous_sibling(left);
        if(!m_elim_tree.is_valid(left)) {
            // chained
            left=m_elim_tree.parent(i2);
        } else
            err=true;

        if(!m_elim_tree.is_valid(left)) {
            err=true;
        }
        if(!err)
            lv=dynamic_cast<c_sub_exp_index*>(*left);

        if(!lv) {
            error(E_TGE_INVALID_BASE_VARIABLE_FOR_SUBSCRIPT,0);
            return false;
        }
    }

    dynamic_cast<c_sub_exp_index*>(*where)->set_base(lv);

    add_base_of_subscript_as_excluded(dynamic_cast<c_sub_exp_variable*>(lv));

    return true;
}

void c_tree_groupper::group_children(tree<c_sub_exp*>::iterator where, vector<c_sub_exp*>& group)
{
    tree<c_sub_exp*>::iterator i1=where;
    c_sub_exp_parenthesis* p=dynamic_cast<c_sub_exp_parenthesis*>(*where);
    int depth=(*i1)->get_depth();
    vector<c_sub_exp*> to_delete;
    ++i1;
    while((*i1)->get_depth()>depth) {
        group.push_back(*i1);
        if(p)
            p->add_sub(*i1);
        to_delete.push_back(*i1);
        ++i1;
    }
    vector<c_sub_exp*>::iterator i2;
    for(i2=to_delete.begin(); i2!=to_delete.end(); ++i2)
        for(i1=m_elim_tree.begin(); i1!=m_elim_tree.end(); ++i1)
            if(*i2==*i1) {
                m_elim_tree.erase(i1);
                break;
            }
    group.push_back(*where);
}

bool c_tree_groupper::completed(e_assoc direction, tree<c_sub_exp*>::iterator start, bool only_first)
{
    tree<c_sub_exp*>::iterator i1=start;

    switch(direction) {
        case EVAL_RIGHT:
            {
                c_sub_exp* r=get_right(i1,only_first,false);
                if(r && r->complete(m_elim_tree))
                    return true;
                if(r==*start)
                    return true;
                return false;
            }
        case EVAL_LEFT:
            for(;;) {
                i1=m_elim_tree.previous_sibling(i1);
                if(!m_elim_tree.is_valid(i1))
                    return true;

                if(!(*i1)->complete(m_elim_tree)) {
                    m_to_process.push_back(*start);
                    m_to_process.push_back(*i1);
                    return false;
                }
                // always only first one!
                return true;
            }
            // NOTREACH
        default:
            {
                MY_ASSERT_NOTREACH();
            }
            break;
    }

    return true;
}

c_sub_exp* c_tree_groupper::get_right(tree<c_sub_exp*>::iterator start, bool only_one, bool same_depth)
{
    tree<c_sub_exp*>::sibling_iterator i1=start;
    tree<c_sub_exp*>::iterator success;
    bool bsuccess=false;
    int prec=(*i1)->get_prec();

    bool checking_right_assoc=false;
    tree<c_sub_exp*>::iterator ra_success;
    bool ra_bsuccess=false;

    stack<tree<c_sub_exp*>::iterator> st;
    stack<int> stp;

    {
        for(;;++i1) {
    again:
            if(checking_right_assoc) {
                if(!m_elim_tree.is_valid(i1)) {
                    if(ra_bsuccess && *start==*success) {
                        return *ra_success;
                    }
                    if(!(*success)->complete(m_elim_tree) || m_elim_tree.number_of_children(success)!=0) {
                        m_prec=(*success)->get_prec();
                        m_to_process.push_back(*start);
                        m_to_process.push_back(*success);
                        return NULL;
                    }
                    return *success;
                }
                const type_info& ti=typeid(**i1);
                if(typeid(c_sub_exp_oper)==ti) {
                    if(dynamic_cast<c_sub_exp_oper*>(*i1)->get_prec()==prec && !(*i1)->complete(m_elim_tree)) {
                        success=i1;
                        continue;
                    }
                    if(dynamic_cast<c_sub_exp_oper*>(*i1)->get_prec()<prec && !(*i1)->complete(m_elim_tree)) {
                        m_prec=(*i1)->get_prec();
                        m_to_process.push_back(*start);
                        m_to_process.push_back(*i1);
                        return NULL;
                    }
                }
                if(!ra_bsuccess) {
                    ra_success=i1;
                    ra_bsuccess=true;
                }
                if(typeid(c_sub_exp_index)==ti) {
                    if(!(*i1)->complete(m_elim_tree) || m_elim_tree.number_of_children(i1)!=0) {
                        m_prec=(*i1)->get_prec();
                        m_to_process.push_back(*start);
                        m_to_process.push_back(*i1);
                        return NULL;
                    }
                    continue;
                }
                if(typeid(c_sub_exp_oper_ternary)==ti || typeid(c_sub_exp_oper_ternary_true)==ti || typeid(c_sub_exp_oper_ternary_false)==ti) {
                    if(!(*i1)->complete(m_elim_tree)) {
                        m_prec=(*i1)->get_prec();
                        m_to_process.push_back(*start);
                        m_to_process.push_back(*i1);
                        return NULL;
                    }
                    m_prec=(*success)->get_prec();
                    m_to_process.push_back(*start);
                    m_to_process.push_back(*success);
                    return NULL;
                }
                if(typeid(c_sub_exp_func)==ti) {
                    if(!(*i1)->complete(m_elim_tree) || m_elim_tree.number_of_children(i1)!=0) {
                        m_prec=(*i1)->get_prec();
                        m_to_process.push_back(*start);
                        m_to_process.push_back(*i1);
                        return NULL;
                    }
                }
                continue;
            }

            if(!m_elim_tree.is_valid(i1)) {
                if(!st.empty()) {
                    i1=st.top();
                    st.pop();
                    prec=stp.top();
                    stp.pop();
                    if(same_depth) {
                        if(!(*i1)->complete(m_elim_tree)) {
                            m_prec=(*i1)->get_prec();
                            m_to_process.push_back(*start);
                            m_to_process.push_back(*i1);
                            return NULL;
                        }
                    }
                    if(!(*i1)->complete(m_elim_tree)) {
                        m_prec=(*i1)->get_prec();
                        m_to_process.push_back(*start);
                        m_to_process.push_back(*i1);
                        return NULL;
                    }
                    goto start;
                }
                if(bsuccess) {
                    if(!only_one && !(*success)->complete(m_elim_tree)) {
                        m_prec=(*success)->get_prec();
                        m_to_process.push_back(*start);
                        m_to_process.push_back(*success);
                        return NULL;
                    }
                }
                if(same_depth)
                    return *start;
                if(bsuccess)
                    return *success;
                i1=start;
                tree<c_sub_exp*>::sibling_iterator i2=i1;
                if(!m_elim_tree.is_valid(++i2))
                    return *i1;
                else
                    return *i2;
            }

            {
                const type_info& ti=typeid(**i1);

                if(typeid(c_sub_exp_func)==ti) {
                    if(!bsuccess)
                        success=i1;
                    bsuccess=true;

                    if(!(*i1)->complete(m_elim_tree) && *start!=*i1) {
                        m_prec=(*i1)->get_prec();
                        m_to_process.push_back(*start);
                        m_to_process.push_back(*i1);
                        return NULL;
                    }

                    st.push(i1);
                    stp.push(prec);
                    i1=m_elim_tree.child(i1,0);
                    if(m_elim_tree.is_valid(i1))
                        prec=(*i1)->get_prec();
                    goto again;
                }

                if(typeid(c_sub_exp_parenthesis)==ti) {
                    if(!bsuccess)
                        success=i1;
                    bsuccess=true;

                    if(!(*i1)->complete(m_elim_tree) && *start!=*i1) {
                        m_prec=(*i1)->get_prec();
                        m_to_process.push_back(*start);
                        m_to_process.push_back(*i1);
                        return NULL;
                    }

                    if(m_elim_tree.number_of_children(i1)!=0) {
                        stp.push(prec);
                        i1=m_elim_tree.child(i1,0);
                        prec=(*i1)->get_prec();
                        goto again;
                    }

                    if(bsuccess && !(*success)->complete(m_elim_tree)) {
                        m_prec=(*success)->get_prec();
                        m_to_process.push_back(*start);
                        m_to_process.push_back(*success);
                        return NULL;
                    }

                    continue;
                }

                if(typeid(c_sub_exp_oper)==ti || typeid(c_sub_exp_index)==ti || typeid(c_sub_exp_oper_ternary)==ti || typeid(c_sub_exp_oper_ternary_true)==ti || typeid(c_sub_exp_oper_ternary_false)==ti || typeid(c_sub_exp_conditional_end_marker)==ti) {
                    c_sub_exp_oper* o=dynamic_cast<c_sub_exp_oper*>(*i1);
                    if(typeid(c_sub_exp_oper)==ti) {
                        // check for right associativity
                        if(o->get_oper().get_assoc()==EVAL_RIGHT && o->get_oper().get_type()==OPER_BINARY) {
                            if(!(*i1)->complete(m_elim_tree)) {
                                prec=(*i1)->get_prec();
                                success=i1;
                                bsuccess=true;
                                checking_right_assoc=true;
                                continue;
                            }
                        }
                    }
                    if((*i1)->get_prec()<=prec && *start!=*i1) {
                        if(bsuccess && typeid(c_sub_exp_conditional_end_marker)==ti) {
                            m_prec=(*i1)->get_prec();
                            m_to_process.push_back(*i1);
                        }
                        if(bsuccess && only_one && (*i1)->get_prec()==prec) {
                            return *success;
                        }
                        if(*i1!=*start && bsuccess && !(*success)->complete(m_elim_tree)) {
                            m_prec=(*success)->get_prec();
                            m_to_process.push_back(*start);
                            m_to_process.push_back(*success);
                            return NULL;
                        }
                        if(!(*i1)->complete(m_elim_tree)) {
                            m_prec=(*i1)->get_prec();
                            m_to_process.push_back(*start);
                            m_to_process.push_back(*i1);
                            return NULL;
                        }
                        if(typeid(c_sub_exp_index)==ti || typeid(c_sub_exp_oper_ternary)==ti || typeid(c_sub_exp_oper_ternary_true)==ti || typeid(c_sub_exp_oper_ternary_false)==ti) {
                            if(m_elim_tree.number_of_children(i1)>0) {
                                m_prec=(*i1)->get_prec();
                                m_to_process.push_back(*start);
                                m_to_process.push_back(*i1);
                                return NULL;
                            }
                        }
                        if(bsuccess && only_one)
                            return *success;

                        if(only_one) {
                            success=i1;
                            bsuccess=true;
                            continue;
                        }
                    } else {
                        prec=(*i1)->get_prec();
                        if(!only_one) {
                            if(*i1==*start) {
                                if(!bsuccess) {
                                    success=i1;
                                    bsuccess=true;
                                }
                            } else {
                                if(bsuccess && !(*success)->complete(m_elim_tree)) {
                                    m_prec=(*success)->get_prec();
                                    m_to_process.push_back(*start);
                                    m_to_process.push_back(*success);
                                    return NULL;
                                }
                            }
                            if(!(*i1)->complete(m_elim_tree) && *start!=*i1) {
                                m_prec=(*i1)->get_prec();
                                m_to_process.push_back(*start);
                                m_to_process.push_back(*i1);
                                return NULL;
                            } else
                                continue;
                            // NOTREACH
                        }
                        if(bsuccess && only_one)
                            return *success;
                        if(!(*i1)->complete(m_elim_tree)) {
                            if(*i1!=*start) {
                                m_prec=(*i1)->get_prec();
                                m_to_process.push_back(*start);
                                m_to_process.push_back(*i1);
                                return NULL;
                            }
                            continue;
                        }
                        if(bsuccess && only_one)
                            return *success;
                        success=i1;
                        bsuccess=true;
                        continue;
                    }
                    continue;
                }
                if(!bsuccess && *i1!=*start) {
                    success=i1;
                    bsuccess=true;
                } else if(bsuccess) {
                    if(!only_one && !(*success)->complete(m_elim_tree)) {
                        m_prec=(*success)->get_prec();
                        m_to_process.push_back(*start);
                        m_to_process.push_back(*success);
                        return NULL;
                    }
                }
            }
    start:
            continue;
        }
    }
    // NOTREACH
}

void c_tree_groupper::include_function_args(const tree<c_sub_exp*>::iterator& parent)
{
    int num=m_elim_tree.number_of_children(parent);
    if(0==num)
        return;

    tree<c_sub_exp*>::sibling_iterator i1=m_elim_tree.child(parent,0);

    for(; m_elim_tree.is_valid(i1); ++i1) {
        if(typeid(**i1)==typeid(c_sub_exp_func_arg_separator))
            continue;
        dynamic_cast<c_sub_exp_func*>(*parent)->add_arg(*i1);
    }

    m_elim_tree.erase_children(parent);
}

