#ifndef C_TREE_GROUPPER_H__
#define C_TREE_GROUPPER_H__

#if defined(HAVE_CONFIG_H)
#   include <config.h>
#endif

#include <string>
#include <vector>
#include <stack>
#include <list>
#include <map>
#include <algorithm>

#include "../../utils/memleak.h"

#include "../../params.h"

#include "c_tree_builder.h"

class c_tree_groupper : public c_tree_builder
{
    tree<c_sub_exp*> m_elim_tree;
    int get_deepest();

    int m_prec;
    vector<c_sub_exp*> m_to_process;

    bool process_oper_unary(const tree<c_sub_exp*>::iterator& where);
    bool process_oper_binary(vector<c_sub_exp*>& group, const tree<c_sub_exp*>::iterator& where);
    bool process_oper_binary_or_unary(vector<c_sub_exp*>& group, const tree<c_sub_exp*>::iterator& where);
    bool process_oper_conditional(vector<c_sub_exp*>& group, const tree<c_sub_exp*>::iterator& where);

    bool children_complete(tree<c_sub_exp*>::iterator where);
    void include_children(c_sub_exp* parent, tree<c_sub_exp*>::iterator where, tree<c_sub_exp*>::iterator protect, vector<c_sub_exp*>* group);

    void group_children(tree<c_sub_exp*>::iterator where, vector<c_sub_exp*>& group);

    bool sanity_check(const vector<c_sub_exp*>& group);

    bool set_index_base(tree<c_sub_exp*>::iterator where);
    c_sub_exp* get_right(tree<c_sub_exp*>::iterator start, bool only_one = true, bool same_depth = true);
    bool completed(e_assoc direction, tree<c_sub_exp*>::iterator start, bool only_first = false);

    friend class c_sub_exp_conditional_end_marker;

    void include_function_args(const tree<c_sub_exp*>::iterator& parent);

protected:
    vector<vector<c_sub_exp*> > m_groups;

    virtual void add_base_of_subscript_as_excluded(c_sub_exp_variable* base) = 0;

    virtual void process();

public:
    c_tree_groupper(const string& exp);
    virtual ~c_tree_groupper();

    string emit_human_evaluation();

    string emit_human_microcode(string::size_type indent);
};

#endif

