#if defined(HAVE_CONFIG_H)
#   include <config.h>
#endif

#include <string>
#include <vector>
#include <stack>
#include <list>
#include <map>
#include <algorithm>

#include "../../utils/memleak.h"

#include "../../params.h"

#include "c_tree_builder.h"
#include "c_oper_desc.h"
#include "oper_prec.h"

#include "c_tree_microcode.h"
#include "../../microcode/c_micro_symbol_variable_mapping.h"

#include "../c_expression.h"
#include "../../microcode/c_micro_log_or.h"
#include "../../microcode/c_micro_log_and.h"

#include "../../utils/precomp.h"

#if defined(_MSC_VER) || defined(__BORLANDC__)
#   pragma hdrstop
#endif

c_tree_microcode::c_tree_microcode(const string& exp, c_expression& expression, c_micro_symbol_variable_mapping& symbols)
: c_tree_groupper(exp)
, m_expression(expression)
, m_symbols(symbols)
, m_first_variable(VAR_SYM_INDEX_INVALID)
{
    if(has_error())
        return;
}

void c_tree_microcode::process()
{
    if(has_error())
        return;

    c_tree_groupper::process();

    if(has_error())
        return;

    compile_to_microcode();

    if(has_error())
        return;
}

c_tree_microcode::~c_tree_microcode()
{
}

void c_tree_microcode::compile_to_microcode()
{
    // first level
    vector<vector<c_sub_exp*> >::iterator i1;
    vector<c_sub_exp*>::iterator i2;
    list<c_sub_exp*>::iterator i4;

    list<c_sub_exp*> st;

    for(i1=m_groups.begin(); i1!=m_groups.end(); ++i1) {
        // compile variables
        for(i4=m_stream.begin(); i4!=m_stream.end(); ++i4) {
            (*i4)->compile(*this);
        }

        if((*i1).begin()!=(*i1).end() && typeid(*(*i1).back())==typeid(c_sub_exp_func) && !dynamic_cast<c_sub_exp_func*>((*i1).back())->get_func_name().compare("array")) {
            ((*i1).back())->compile(*this);
            m_stream.insert(m_stream.end(),st.begin(),st.end());
            m_stream.push_back((*i1).back());
        } else {
            st.assign((*i1).begin(),(*i1).end());
            m_stream.insert(m_stream.end(),(*i1).begin(),(*i1).end());
        }

        // compile variables
        for(i4=m_stream.begin(); i4!=m_stream.end(); ++i4) {
            (*i4)->compile(*this);
        }

        my_unique(m_stream);
        emit_microcode(m_expression,m_symbols);

again:
        // n-th level
        // compile variables
        for(i4=m_stream.begin(); i4!=m_stream.end(); ++i4) {
            (*i4)->compile(*this);
        }
        for(i4=m_stream.begin(); i4!=m_stream.end(); ++i4) {
            size_t sz=m_stream.size();
            if(!(*i4)->emit_evaluation_order(m_stream,i4)) {
                if(m_stream.size()==sz)
                    continue;
                goto again;
            }
        }

        // make it unique
        my_unique(st);

        // compile variables
        for(i4=m_stream.begin(); i4!=m_stream.end(); ++i4) {
            (*i4)->compile(*this);
        }

        my_unique(m_stream);
        emit_microcode(m_expression,m_symbols);
    }

    // make it unique
    my_unique(m_stream);

    // exclude base of indexes
    exclude_bases_of_subscriptps();

    emit_microcode(m_expression,m_symbols);

    collect_conditionals();

    if(!m_stream.empty())
        m_first_variable=m_stream.front()->get_var_id();
}

string c_tree_microcode::emit_human_microcode()
{
    size_t ii=0;
    ostringstream o;

    list<c_sub_exp*>::iterator i2;
    for(i2=m_stream.begin(); i2!=m_stream.end(); ++i2)
        o << "#" << ++ii << ": " << (*i2)->emit_human_microcode(0).c_str();

    return o.str();
}

var_sym_index_t c_tree_microcode::add_var(const string& var_name)
{
    return m_symbols.add(var_name);
}

var_sym_index_t c_tree_microcode::add_temp_var()
{
    return m_symbols.add();
}

var_sym_index_t c_tree_microcode::add_const_int_var(MY_LONG_LONG num)
{
    return m_symbols.add(tos(num),&m_expression);
}

var_sym_index_t c_tree_microcode::add_const_str_var(const string& str)
{
    string s=(string)"'"+str+"'";
    return m_symbols.add(s,&m_expression);
}

var_sym_index_t c_tree_microcode::add_const_bool_var(bool val)
{
    string s=val ? "TRUE" : "FALSE";
    return m_symbols.add(s,&m_expression);
}

var_sym_index_t c_tree_microcode::add_const_null_var()
{
    string s="NULL";
    return m_symbols.add(s,&m_expression);
}

var_sym_index_t c_tree_microcode::add_eval_str_var(const string& str)
{
    string s=(string)"\""+str+"\"";
    return m_symbols.add(s,&m_expression);
}

void c_tree_microcode::emit_microcode(c_expression& expression, c_micro_symbol_variable_mapping& symbols)
{
    list<c_sub_exp*>::iterator i1;
    for(i1=m_stream.begin(); i1!=m_stream.end(); ++i1) {
        if(m_stream_microcode_emitted.find(*i1)!=m_stream_microcode_emitted.end())
            continue;
        (*i1)->emit_microcode(expression,symbols);
        m_stream_microcode_emitted.insert(make_pair(*i1,true));
    }
}

void c_tree_microcode::add_base_of_subscript_as_excluded(c_sub_exp_variable* UNUSEDARG(base))
{
}

void c_tree_microcode::exclude_bases_of_subscriptps()
{
    list<c_sub_exp*>::iterator i1;
    vector<c_sub_exp_variable*>::iterator i2;
again:
    for(i1=m_stream.begin(); i1!=m_stream.end(); ++i1) {
        for(i2=m_excluded.begin(); i2!=m_excluded.end(); ++i2) {
            if(*i1==*i2) {
                m_stream.erase(i1);
                goto again;
            }
        }
    }
}

void c_tree_microcode::collect_conditionals()
{
    list<c_sub_exp*>::iterator i1;
    for(i1=m_stream.begin(); i1!=m_stream.end(); ++i1) {
        if(typeid(c_sub_exp_oper)==typeid(**i1)) {
            c_sub_exp_oper* o=dynamic_cast<c_sub_exp_oper*>(*i1);

            vector<c_micro_log_or*>& ors=o->get_all_conditionals_micro_or();
            vector<c_micro_log_and*>& ands=o->get_all_conditionals_micro_and();

            if(!ors.empty()) {
                vector<var_sym_index_t> outputs;

                vector<c_micro_log_or*>::iterator i2;
                for(i2=ors.begin(); i2!=ors.end(); ++i2) {
                    outputs.push_back((*i2)->get_output());
                }

                // optimize for execution/run time
                sort(outputs.begin(),outputs.end());
                vector<var_sym_index_t>::iterator i3=unique(outputs.begin(),outputs.end());
                outputs.erase(i3,outputs.end());
                //

                for(i2=ors.begin(); i2!=ors.end(); ++i2) {
                    (*i2)->set_all_outputs(outputs);
                }
            }

            if(!ands.empty()) {
                vector<var_sym_index_t> outputs;

                vector<c_micro_log_and*>::iterator i2;
                for(i2=ands.begin(); i2!=ands.end(); ++i2) {
                    outputs.push_back((*i2)->get_output());
                }

                // optimize for execution/run time
                sort(outputs.begin(),outputs.end());
                vector<var_sym_index_t>::iterator i3=unique(outputs.begin(),outputs.end());
                outputs.erase(i3,outputs.end());
                //

                for(i2=ands.begin(); i2!=ands.end(); ++i2) {
                    (*i2)->set_all_outputs(outputs);
                }
            }
        }
    }
}

void c_tree_microcode::my_unique(list<c_sub_exp*>& lst)
{
    list<c_sub_exp*>::iterator i3, i4;
again5:
    if(!lst.empty()) {
        for(i4=lst.end(), --i4; ; --i4) {
            for(i3=lst.begin(); i3!=i4; ++i3) {
                if(i3==i4)
                    continue;
                if(*i4==*i3) {
                    lst.erase(i3);
                    goto again5;
                }
            }
            if(lst.begin()==i4)
                break;
        }
    }
}

var_sym_index_t c_tree_microcode::get_first_variable()
{
    return m_first_variable;
}

var_sym_index_t c_tree_microcode::add_const_built_in(const string& name)
{
    return m_symbols.add(name,&m_expression);
}

