#ifndef C_TREE_MICROCODE_H__
#define C_TREE_MICROCODE_H__

#if defined(HAVE_CONFIG_H)
#   include <config.h>
#endif

#include <string>
#include <vector>
#include <stack>
#include <list>
#include <map>
#include <algorithm>

#include "../../utils/memleak.h"

#include "../../params.h"

#include "oper_prec.h"
#include "c_tree_groupper.h"
#include "c_tree_shared.h"

class c_expression;
class c_micro_symbol_variable_mapping;

#if defined(_MSC_VER) && _MSC_VER>=1400
#   pragma warning(push)
#   pragma warning(disable: 4512)   // warning C4512: 'c_tree_microcode' : assignment operator could not be generated
#endif

class c_sub_exp_variable;
class c_tree_groupper;

class c_tree_microcode : public c_tree_groupper
{
    list<c_sub_exp*> m_stream;
    map<c_sub_exp*, bool> m_stream_microcode_emitted;
    c_expression& m_expression;
    c_micro_symbol_variable_mapping& m_symbols;

    void compile_to_microcode();
    void emit_microcode(c_expression& expression, c_micro_symbol_variable_mapping& symbols);
    friend class c_sub_exp_variable;
    vector<c_sub_exp_variable*> m_excluded;
    void exclude_bases_of_subscriptps();

    void collect_conditionals();

    void my_unique(list<c_sub_exp*>& lst);

    var_sym_index_t m_first_variable;

protected:

    void add_base_of_subscript_as_excluded(c_sub_exp_variable* base);
    friend class c_tree_groupper;

public:
    c_tree_microcode(const string& exp, c_expression& expression, c_micro_symbol_variable_mapping& symbols);
    virtual ~c_tree_microcode();

    string emit_human_microcode();

    var_sym_index_t add_var(const string& var_name);
    var_sym_index_t add_temp_var();
    var_sym_index_t add_const_int_var(MY_LONG_LONG num);
    var_sym_index_t add_const_str_var(const string& str);
    var_sym_index_t add_const_bool_var(bool val);
    var_sym_index_t add_const_null_var();
    var_sym_index_t add_eval_str_var(const string& str);
    var_sym_index_t add_const_built_in(const string& name);

    virtual void process();

    var_sym_index_t get_first_variable();
};

#if defined(_MSC_VER) && _MSC_VER>=1400
#   pragma warning(pop)
#endif

#endif

