#ifndef C_TREE_SHARED_H__
#define C_TREE_SHARED_H__

#if defined(HAVE_CONFIG_H)
#   include <config.h>
#endif

#include <string>
#include <vector>
#include <stack>
#include <list>
#include <map>
#include <algorithm>

#include "../../utils/memleak.h"

#include "../../params.h"

#include "c_oper_desc.h"
#include "oper_prec.h"
#include "../../microcode_shared.h"

#include "../../tree/tree.hh"

class c_expression;
class c_micro_symbol_variable_mapping;

class c_tree_microcode;

#if defined(_MSC_VER) && _MSC_VER>=1400
#   pragma warning(push)
#   pragma warning(disable: 4512)   // warning C4512: 'c_sub_exp_oper' : assignment operator could not be generated
#endif

class c_exp_sanity
{
public:
    enum e_can_follow
    {
        CF_USER_FATAL_ERROR,
        CF_OKAY,
        CF_WARN,
        CF_OOPS_BUG,
    };

    c_exp_sanity(e_can_follow rc, const char* msg, int line);

    e_can_follow get(const char* &msg, const char* &file, int& line) const;

private:
    e_can_follow m_rc;
    const char* m_msg;
    int m_line;
};

class c_sub_exp_root;
class c_sub_exp_root_end_dummy;
class c_sub_exp_parenthesis;
class c_sub_exp_index;
class c_sub_exp_variable;
class c_sub_exp_numeric_immediate;
class c_sub_exp_const;
class c_sub_exp_func;
class c_sub_exp_eval_str;
class c_sub_exp_immediate_str;
class c_sub_exp_oper;
class c_sub_exp_oper_ternary;
class c_sub_exp_oper_ternary_true;
class c_sub_exp_oper_ternary_false;
class c_sub_exp_contitional_node_helper;
class c_sub_exp_end_parenthesis;
class c_sub_exp_end_index;
class c_sub_exp_func_end_args;
class c_sub_exp_func_arg_separator;
class c_sub_exp_conditional_end_marker;
class c_sub_exp_dummy;
class c_sub_exp_ternary_true_end_marker;
class c_sub_exp_ternary_false_end_marker;

class c_sub_exp
{
protected:
    bool m_evaluation_order_emited;

    var_sym_index_t m_var_id;

    int m_depth;

public:
    c_sub_exp()
        : m_evaluation_order_emited(false)
        , m_var_id(VAR_SYM_INDEX_INVALID)
        , m_depth(0)
    { ; }

    virtual ~c_sub_exp() = 0;

    virtual string emit_human_desc(string::size_type indent) = 0;

    virtual int get_prec() = 0;

    void reset_evaluation_order_emited();

    virtual bool emit_evaluation_order(list<c_sub_exp*>& st, list<c_sub_exp*>::iterator where);

    virtual void compile(c_tree_microcode& tm_callback) = 0;

    virtual string emit_human_microcode(string::size_type indent) = 0;

    bool compiled();

    void set_depth(int depth);

    int get_depth();

    virtual void emit_microcode(c_expression& expression, c_micro_symbol_variable_mapping& symbols) = 0;

    var_sym_index_t get_var_id()
    {
        return m_var_id;
    }

    virtual bool complete(tree<c_sub_exp*>& t) = 0;
    virtual void set_complete() { ; }

    virtual const c_exp_sanity* can_follow(c_sub_exp& right) = 0;

    virtual const c_exp_sanity* can_precede(c_sub_exp_root* left) = 0;
    virtual const c_exp_sanity* can_precede(c_sub_exp_root_end_dummy* left) = 0;
    virtual const c_exp_sanity* can_precede(c_sub_exp_parenthesis* left) = 0;
    virtual const c_exp_sanity* can_precede(c_sub_exp_index* left) = 0;
    virtual const c_exp_sanity* can_precede(c_sub_exp_variable* left) = 0;
    virtual const c_exp_sanity* can_precede(c_sub_exp_numeric_immediate* left) = 0;
    virtual const c_exp_sanity* can_precede(c_sub_exp_const* left) = 0;
    virtual const c_exp_sanity* can_precede(c_sub_exp_func* left) = 0;
    virtual const c_exp_sanity* can_precede(c_sub_exp_eval_str* left) = 0;
    virtual const c_exp_sanity* can_precede(c_sub_exp_immediate_str* left) = 0;
    virtual const c_exp_sanity* can_precede(c_sub_exp_oper* left) = 0;
    virtual const c_exp_sanity* can_precede(c_sub_exp_oper_ternary* left) = 0;
    virtual const c_exp_sanity* can_precede(c_sub_exp_oper_ternary_true* left) = 0;
    virtual const c_exp_sanity* can_precede(c_sub_exp_oper_ternary_false* left) = 0;
    virtual const c_exp_sanity* can_precede(c_sub_exp_contitional_node_helper* left) = 0;
    virtual const c_exp_sanity* can_precede(c_sub_exp_end_parenthesis* left) = 0;
    virtual const c_exp_sanity* can_precede(c_sub_exp_end_index* left) = 0;
    virtual const c_exp_sanity* can_precede(c_sub_exp_func_end_args* left) = 0;
    virtual const c_exp_sanity* can_precede(c_sub_exp_func_arg_separator* left) = 0;
    virtual const c_exp_sanity* can_precede(c_sub_exp_conditional_end_marker* left) = 0;
    virtual const c_exp_sanity* can_precede(c_sub_exp_dummy* left) = 0;
    virtual const c_exp_sanity* can_precede(c_sub_exp_ternary_true_end_marker* left) = 0;
    virtual const c_exp_sanity* can_precede(c_sub_exp_ternary_false_end_marker* left) = 0;
};

class c_sub_exp_root : public c_sub_exp
{
    vector<c_sub_exp*> m_eval;

public:
    c_sub_exp_root()
        : c_sub_exp()
    { ; }

    virtual ~c_sub_exp_root() { ; }

    virtual string emit_human_desc(string::size_type indent);

    void add_sub(c_sub_exp* sub);

    virtual int get_prec();

    virtual void compile(c_tree_microcode& UNUSEDARG(tm_callback)) { ; }

    virtual string emit_human_microcode(string::size_type indent);

    virtual void emit_microcode(c_expression& UNUSEDARG(expression), c_micro_symbol_variable_mapping& UNUSEDARG(symbols)) { ; }

    virtual bool complete(tree<c_sub_exp*>& UNUSEDARG(t)) { return true; }

    virtual const c_exp_sanity* can_follow(c_sub_exp& right);

    virtual const c_exp_sanity* can_precede(c_sub_exp_root* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_root_end_dummy* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_parenthesis* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_index* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_variable* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_numeric_immediate* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_const* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_func* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_eval_str* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_immediate_str* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_oper* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_oper_ternary* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_oper_ternary_true* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_oper_ternary_false* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_contitional_node_helper* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_end_parenthesis* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_end_index* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_func_end_args* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_func_arg_separator* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_conditional_end_marker* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_dummy* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_ternary_true_end_marker* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_ternary_false_end_marker* left);
};

class c_sub_exp_root_end_dummy : public c_sub_exp
{
public:
    c_sub_exp_root_end_dummy()
        : c_sub_exp()
    { ; }

    virtual ~c_sub_exp_root_end_dummy() { ; }

    virtual string emit_human_desc(string::size_type UNUSEDARG(indent)) { return (string)"EXPRESSION_END\n"; }

    virtual int get_prec() { return 0; }

    virtual void compile(c_tree_microcode& UNUSEDARG(tm_callback)) { ; }

    virtual string emit_human_microcode(string::size_type UNUSEDARG(indent)) { return (string)"EXPRESSION_END\n"; }

    virtual void emit_microcode(c_expression& UNUSEDARG(expression), c_micro_symbol_variable_mapping& UNUSEDARG(symbols)) { ; }

    virtual bool complete(tree<c_sub_exp*>& UNUSEDARG(t)) { return true; }

    virtual const c_exp_sanity* can_follow(c_sub_exp& right);

    virtual const c_exp_sanity* can_precede(c_sub_exp_root* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_root_end_dummy* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_parenthesis* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_index* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_variable* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_numeric_immediate* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_const* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_func* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_eval_str* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_immediate_str* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_oper* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_oper_ternary* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_oper_ternary_true* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_oper_ternary_false* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_contitional_node_helper* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_end_parenthesis* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_end_index* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_func_end_args* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_func_arg_separator* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_conditional_end_marker* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_dummy* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_ternary_true_end_marker* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_ternary_false_end_marker* left);
};

class c_sub_exp_parenthesis : public c_sub_exp
{
    vector<c_sub_exp*> m_eval;
    c_sub_exp* m_result;

public:
    c_sub_exp_parenthesis()
        : c_sub_exp()
        , m_result(NULL)
    { ; }

    virtual ~c_sub_exp_parenthesis() { ; }

    virtual string emit_human_desc(string::size_type indent);
    virtual string emit_human_desc_no_recursion(string::size_type indent);

    void add_sub(c_sub_exp* sub);

    virtual int get_prec();

    virtual bool complete(tree<c_sub_exp*>& t);

    virtual bool emit_evaluation_order(list<c_sub_exp*>& st, list<c_sub_exp*>::iterator where);

    virtual void compile(c_tree_microcode& tm_callback);

    virtual string emit_human_microcode(string::size_type indent);
    virtual string emit_human_microcode_no_recursion(string::size_type indent);

    virtual void emit_microcode(c_expression& expression, c_micro_symbol_variable_mapping& symbols);

    const vector<c_sub_exp*>& get_eval() const
    {
        return m_eval;
    }

    void set_result(c_sub_exp* result);
    c_sub_exp* get_result();

    virtual const c_exp_sanity* can_follow(c_sub_exp& right);

    virtual const c_exp_sanity* can_precede(c_sub_exp_root* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_root_end_dummy* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_parenthesis* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_index* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_variable* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_numeric_immediate* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_const* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_func* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_eval_str* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_immediate_str* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_oper* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_oper_ternary* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_oper_ternary_true* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_oper_ternary_false* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_contitional_node_helper* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_end_parenthesis* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_end_index* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_func_end_args* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_func_arg_separator* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_conditional_end_marker* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_dummy* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_ternary_true_end_marker* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_ternary_false_end_marker* left);
};

class c_sub_exp_index : public c_sub_exp
{
    vector<c_sub_exp*> m_idx;
    c_sub_exp* m_base_exp;
    bool m_case_insensitive;
    c_sub_exp* m_result;

public:
    c_sub_exp_index()
        : c_sub_exp()
        , m_base_exp(NULL)
        , m_case_insensitive(false)
        , m_result(NULL)
    { ; }

    virtual ~c_sub_exp_index() { ; }

    virtual string emit_human_desc(string::size_type indent);

    void add_index(c_sub_exp* idx);

    virtual int get_prec();

    virtual bool complete(tree<c_sub_exp*>& UNUSEDARG(t));

    virtual bool emit_evaluation_order(list<c_sub_exp*>& st, list<c_sub_exp*>::iterator where);

    virtual void compile(c_tree_microcode& tm_callback);

    virtual string emit_human_microcode(string::size_type indent);

    virtual void emit_microcode(c_expression& expression, c_micro_symbol_variable_mapping& symbols);

    void set_base(c_sub_exp* base);
    c_sub_exp* get_base();

    void set_case_insensitive();

    void set_result(c_sub_exp* result);
    c_sub_exp* get_result();

    virtual const c_exp_sanity* can_follow(c_sub_exp& right);

    virtual const c_exp_sanity* can_precede(c_sub_exp_root* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_root_end_dummy* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_parenthesis* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_index* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_variable* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_numeric_immediate* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_const* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_func* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_eval_str* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_immediate_str* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_oper* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_oper_ternary* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_oper_ternary_true* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_oper_ternary_false* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_contitional_node_helper* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_end_parenthesis* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_end_index* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_func_end_args* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_func_arg_separator* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_conditional_end_marker* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_dummy* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_ternary_true_end_marker* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_ternary_false_end_marker* left);
};

class c_sub_exp_variable : public c_sub_exp
{
    string m_name;
    bool m_complete;

public:
    c_sub_exp_variable(const string& name)
        : c_sub_exp()
        , m_name(name)
        , m_complete(false)
    { ; }

    virtual ~c_sub_exp_variable() { ; }

    virtual string emit_human_desc(string::size_type indent);

    string get_varname();

    virtual int get_prec();

    virtual bool emit_evaluation_order(list<c_sub_exp*>& st, list<c_sub_exp*>::iterator where);

    virtual void compile(c_tree_microcode& tm_callback);

    virtual string emit_human_microcode(string::size_type indent);

    virtual void emit_microcode(c_expression& expression, c_micro_symbol_variable_mapping& symbols);

    virtual bool complete(tree<c_sub_exp*>& UNUSEDARG(t)) { return m_complete; }
    virtual void set_complete() { m_complete=true; };

    virtual const c_exp_sanity* can_follow(c_sub_exp& right);

    virtual const c_exp_sanity* can_precede(c_sub_exp_root* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_root_end_dummy* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_parenthesis* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_index* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_variable* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_numeric_immediate* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_const* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_func* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_eval_str* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_immediate_str* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_oper* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_oper_ternary* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_oper_ternary_true* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_oper_ternary_false* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_contitional_node_helper* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_end_parenthesis* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_end_index* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_func_end_args* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_func_arg_separator* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_conditional_end_marker* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_dummy* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_ternary_true_end_marker* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_ternary_false_end_marker* left);
};

class c_sub_exp_numeric_immediate : public c_sub_exp
{
    MY_LONG_LONG m_num;
    bool m_complete;

public:
    c_sub_exp_numeric_immediate(MY_LONG_LONG num)
        : c_sub_exp()
        , m_num(num)
        , m_complete(false)
    { ; }

    virtual ~c_sub_exp_numeric_immediate() { ; }

    virtual string emit_human_desc(string::size_type indent);

    MY_LONG_LONG get_num();

    virtual int get_prec();

    virtual bool emit_evaluation_order(list<c_sub_exp*>& st, list<c_sub_exp*>::iterator where);

    virtual void compile(c_tree_microcode& tm_callback);

    virtual string emit_human_microcode(string::size_type indent);

    virtual void emit_microcode(c_expression& expression, c_micro_symbol_variable_mapping& symbols);

    virtual bool complete(tree<c_sub_exp*>& UNUSEDARG(t)) { return m_complete; }
    virtual void set_complete() { m_complete=true; };

    virtual const c_exp_sanity* can_follow(c_sub_exp& right);

    virtual const c_exp_sanity* can_precede(c_sub_exp_root* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_root_end_dummy* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_parenthesis* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_index* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_variable* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_numeric_immediate* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_const* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_func* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_eval_str* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_immediate_str* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_oper* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_oper_ternary* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_oper_ternary_true* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_oper_ternary_false* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_contitional_node_helper* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_end_parenthesis* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_end_index* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_func_end_args* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_func_arg_separator* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_conditional_end_marker* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_dummy* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_ternary_true_end_marker* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_ternary_false_end_marker* left);
};

class c_sub_exp_const : public c_sub_exp
{
    string m_name;
    bool m_complete;

public:
    c_sub_exp_const(const string& name)
        : c_sub_exp()
        , m_name(name)
        , m_complete(false)
    { ; }

    virtual ~c_sub_exp_const() { ; }

    virtual string emit_human_desc(string::size_type indent);

    string get_const();

    virtual int get_prec();

    virtual bool emit_evaluation_order(list<c_sub_exp*>& st, list<c_sub_exp*>::iterator where);

    virtual void compile(c_tree_microcode& tm_callback);

    virtual string emit_human_microcode(string::size_type indent);

    virtual void emit_microcode(c_expression& expression, c_micro_symbol_variable_mapping& symbols);

    virtual bool complete(tree<c_sub_exp*>& UNUSEDARG(t)) { return m_complete; }
    virtual void set_complete() { m_complete=true; };

    virtual const c_exp_sanity* can_follow(c_sub_exp& right);

    virtual const c_exp_sanity* can_precede(c_sub_exp_root* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_root_end_dummy* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_parenthesis* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_index* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_variable* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_numeric_immediate* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_const* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_func* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_eval_str* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_immediate_str* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_oper* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_oper_ternary* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_oper_ternary_true* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_oper_ternary_false* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_contitional_node_helper* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_end_parenthesis* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_end_index* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_func_end_args* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_func_arg_separator* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_conditional_end_marker* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_dummy* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_ternary_true_end_marker* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_ternary_false_end_marker* left);
};

class c_sub_exp_func : public c_sub_exp
{
    string m_name;

    vector<c_sub_exp*> m_args;
    vector<pair<var_sym_index_t, var_sym_index_t> > m_array_args;

    bool m_complete;

public:
    c_sub_exp_func(const string& name)
        : c_sub_exp()
        , m_name(name)
        , m_complete(false)
    { ; }

    virtual ~c_sub_exp_func() { ; }

    virtual string emit_human_desc(string::size_type indent);

    string get_func_name();

    void add_arg(c_sub_exp* arg);

    virtual int get_prec();

    virtual bool complete(tree<c_sub_exp*>& UNUSEDARG(t));

    virtual void set_complete();

    virtual bool emit_evaluation_order(list<c_sub_exp*>& st, list<c_sub_exp*>::iterator where);

    virtual void compile(c_tree_microcode& tm_callback);

    virtual string emit_human_microcode(string::size_type indent);

    virtual void emit_microcode(c_expression& expression, c_micro_symbol_variable_mapping& symbols);

    virtual const c_exp_sanity* can_follow(c_sub_exp& right);

    virtual const c_exp_sanity* can_precede(c_sub_exp_root* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_root_end_dummy* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_parenthesis* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_index* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_variable* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_numeric_immediate* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_const* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_func* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_eval_str* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_immediate_str* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_oper* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_oper_ternary* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_oper_ternary_true* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_oper_ternary_false* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_contitional_node_helper* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_end_parenthesis* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_end_index* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_func_end_args* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_func_arg_separator* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_conditional_end_marker* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_dummy* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_ternary_true_end_marker* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_ternary_false_end_marker* left);
};

class c_sub_exp_eval_str : public c_sub_exp
{
    ostringstream m_str;
    bool m_complete;
    bool m_have_dollar;

public:
    c_sub_exp_eval_str()
        : c_sub_exp()
        , m_complete(false)
        , m_have_dollar(false)
    { ; }

    virtual ~c_sub_exp_eval_str() { ; }

    virtual string emit_human_desc(string::size_type indent);

    void append(char ch);

    string get_eval_str();

    virtual int get_prec();

    virtual bool emit_evaluation_order(list<c_sub_exp*>& st, list<c_sub_exp*>::iterator where);

    virtual void compile(c_tree_microcode& tm_callback);

    virtual string emit_human_microcode(string::size_type indent);

    virtual void emit_microcode(c_expression& expression, c_micro_symbol_variable_mapping& symbols);

    virtual bool complete(tree<c_sub_exp*>& UNUSEDARG(t)) { return m_complete; }
    virtual void set_complete() { m_complete=true; };
    void have_dollar() { m_have_dollar=true; }

    virtual const c_exp_sanity* can_follow(c_sub_exp& right);

    virtual const c_exp_sanity* can_precede(c_sub_exp_root* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_root_end_dummy* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_parenthesis* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_index* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_variable* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_numeric_immediate* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_const* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_func* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_eval_str* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_immediate_str* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_oper* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_oper_ternary* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_oper_ternary_true* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_oper_ternary_false* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_contitional_node_helper* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_end_parenthesis* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_end_index* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_func_end_args* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_func_arg_separator* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_conditional_end_marker* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_dummy* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_ternary_true_end_marker* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_ternary_false_end_marker* left);
};

class c_sub_exp_immediate_str : public c_sub_exp
{
    ostringstream m_str;
    bool m_complete;

public:
    c_sub_exp_immediate_str()
        : c_sub_exp()
        , m_complete(false)
    { ; }

    virtual ~c_sub_exp_immediate_str() { ; }

    virtual string emit_human_desc(string::size_type indent);

    void append(char ch);

    string get_imm_str();

    virtual int get_prec();

    virtual bool emit_evaluation_order(list<c_sub_exp*>& st, list<c_sub_exp*>::iterator where);

    virtual void compile(c_tree_microcode& tm_callback);

    virtual string emit_human_microcode(string::size_type indent);

    virtual void emit_microcode(c_expression& expression, c_micro_symbol_variable_mapping& symbols);

    virtual bool complete(tree<c_sub_exp*>& UNUSEDARG(t)) { return m_complete; }
    virtual void set_complete() { m_complete=true; };

    virtual const c_exp_sanity* can_follow(c_sub_exp& right);

    virtual const c_exp_sanity* can_precede(c_sub_exp_root* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_root_end_dummy* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_parenthesis* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_index* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_variable* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_numeric_immediate* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_const* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_func* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_eval_str* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_immediate_str* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_oper* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_oper_ternary* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_oper_ternary_true* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_oper_ternary_false* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_contitional_node_helper* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_end_parenthesis* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_end_index* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_func_end_args* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_func_arg_separator* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_conditional_end_marker* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_dummy* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_ternary_true_end_marker* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_ternary_false_end_marker* left);
};

class c_micro_log_or;
class c_micro_log_and;

class c_sub_exp_oper : public c_sub_exp
{
    bool m_postfix;

    vector<c_sub_exp*> m_same_prec_opers;

    vector<c_sub_exp_oper*> m_all_conditionals;

    vector<c_micro_log_or*> m_all_conditionals_micro_or;
    vector<c_micro_log_and*> m_all_conditionals_micro_and;

    bool m_return_complete;

protected:
    const c_oper_desc& m_oper;
    int m_oper_index;

    vector<c_sub_exp*> m_eval;

public:
    c_sub_exp_oper(int oper_index)
        : c_sub_exp()
        , m_return_complete(false)
        , m_oper(g_oper_precedence[oper_index])
        , m_oper_index(oper_index)
    { ; }

    virtual ~c_sub_exp_oper() { ; }

    virtual string emit_human_desc(string::size_type indent);

    virtual int get_prec();

    const c_oper_desc& get_oper() const;

    bool add_eval(c_sub_exp* el);

    virtual bool complete(tree<c_sub_exp*>& UNUSEDARG(t));

    virtual bool emit_evaluation_order(list<c_sub_exp*>& st, list<c_sub_exp*>::iterator where);

    virtual void compile(c_tree_microcode& tm_callback);

    virtual string emit_human_microcode(string::size_type indent);

    virtual void emit_microcode(c_expression& expression, c_micro_symbol_variable_mapping& symbols);

    void set_postfix(bool postfix)
    {
        m_postfix=postfix;
    }

    void add_same_prec_opers(c_sub_exp_oper* el)
    {
        m_same_prec_opers.push_back(el);
    }

    const vector<c_sub_exp*>& get_eval() const;

    void set_all_conditionals(const vector<c_sub_exp_oper*>& all_conditionals);
    vector<c_sub_exp_oper*>& get_all_conditionals();

    void add_conditional_micro_or(c_micro_log_or* m);
    void add_conditional_micro_and(c_micro_log_and* m);

    vector<c_micro_log_or*>& get_all_conditionals_micro_or();
    vector<c_micro_log_and*>& get_all_conditionals_micro_and();

    int get_oper_index() const;

    bool get_return_complete();
    void set_return_complete();

    const vector<c_sub_exp*>& get_eval();

    virtual const c_exp_sanity* can_follow(c_sub_exp& right);

    virtual const c_exp_sanity* can_precede(c_sub_exp_root* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_root_end_dummy* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_parenthesis* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_index* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_variable* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_numeric_immediate* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_const* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_func* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_eval_str* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_immediate_str* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_oper* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_oper_ternary* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_oper_ternary_true* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_oper_ternary_false* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_contitional_node_helper* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_end_parenthesis* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_end_index* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_func_end_args* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_func_arg_separator* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_conditional_end_marker* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_dummy* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_ternary_true_end_marker* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_ternary_false_end_marker* left);
};

class c_micro_ternary;
class c_micro_jump;

class c_sub_exp_oper_ternary : public c_sub_exp_oper
{
    c_micro_ternary* m_micro;
    c_micro_jump* m_true_jump;
    c_sub_exp* m_result;
    c_sub_exp_oper_ternary* m_other_ternary;

public:
    c_sub_exp_oper_ternary(int oper_index)
        : c_sub_exp_oper(oper_index)
        , m_micro(NULL)
        , m_true_jump(NULL)
        , m_result(NULL)
        , m_other_ternary(NULL)
    { ; }

    virtual ~c_sub_exp_oper_ternary() { ; }

    virtual string emit_human_desc(string::size_type indent);

    virtual bool emit_evaluation_order(list<c_sub_exp*>& st, list<c_sub_exp*>::iterator where);

    virtual void compile(c_tree_microcode& tm_callback);

    virtual string emit_human_microcode(string::size_type indent);

    virtual void emit_microcode(c_expression& expression, c_micro_symbol_variable_mapping& symbols);

    void add_condition(c_sub_exp* el);

    c_micro_ternary* get_micro();

    void set_true_jump(c_micro_jump* micro);
    c_micro_jump* get_true_jump();

    void set_result(c_sub_exp* result);
    c_sub_exp* get_result();

    void set_other_ternary(c_sub_exp_oper_ternary* other_ternary);

    virtual const c_exp_sanity* can_follow(c_sub_exp& right);

    virtual const c_exp_sanity* can_precede(c_sub_exp_root* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_root_end_dummy* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_parenthesis* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_index* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_variable* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_numeric_immediate* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_const* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_func* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_eval_str* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_immediate_str* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_oper* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_oper_ternary* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_oper_ternary_true* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_oper_ternary_false* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_contitional_node_helper* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_end_parenthesis* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_end_index* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_func_end_args* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_func_arg_separator* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_conditional_end_marker* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_dummy* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_ternary_true_end_marker* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_ternary_false_end_marker* left);
};

class c_sub_exp_oper_ternary_true : public c_sub_exp
{
    vector<c_sub_exp*> m_eval;
    c_sub_exp* m_result;
    c_sub_exp_oper_ternary* m_ternary;

public:
    c_sub_exp_oper_ternary_true()
        : c_sub_exp()
        , m_result(NULL)
        , m_ternary(NULL)
    { ; }

    virtual ~c_sub_exp_oper_ternary_true() { ; }

    virtual string emit_human_desc(string::size_type UNUSEDARG(indent)) { return (string)""; }
    virtual string emit_human_desc_no_recursion(string::size_type UNUSEDARG(indent)) { return (string)""; }

    void add_sub(c_sub_exp* sub);

    virtual int get_prec();

    virtual bool complete(tree<c_sub_exp*>& t);

    virtual bool emit_evaluation_order(list<c_sub_exp*>& st, list<c_sub_exp*>::iterator where);

    virtual void compile(c_tree_microcode& tm_callback);

    virtual string emit_human_microcode(string::size_type UNUSEDARG(indent)) { return (string)""; }
    virtual string emit_human_microcode_no_recursion(string::size_type UNUSEDARG(indent)) { return (string)""; }

    virtual void emit_microcode(c_expression& expression, c_micro_symbol_variable_mapping& symbols);

    const vector<c_sub_exp*>& get_eval() const
    {
        return m_eval;
    }

    void set_result(c_sub_exp* result);
    c_sub_exp* get_result();

    void set_ternary(c_sub_exp_oper_ternary* ternary);
    c_sub_exp_oper_ternary* get_ternary();

    virtual const c_exp_sanity* can_follow(c_sub_exp& right);

    virtual const c_exp_sanity* can_precede(c_sub_exp_root* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_root_end_dummy* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_parenthesis* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_index* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_variable* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_numeric_immediate* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_const* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_func* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_eval_str* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_immediate_str* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_oper* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_oper_ternary* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_oper_ternary_true* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_oper_ternary_false* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_contitional_node_helper* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_end_parenthesis* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_end_index* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_func_end_args* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_func_arg_separator* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_conditional_end_marker* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_dummy* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_ternary_true_end_marker* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_ternary_false_end_marker* left);
};

class c_sub_exp_oper_ternary_false : public c_sub_exp_oper
{
    c_sub_exp* m_result;
    c_sub_exp_oper_ternary* m_ternary;

    int m_microcode_emit_cnt;

public:
    c_sub_exp_oper_ternary_false(int oper_index)
        : c_sub_exp_oper(oper_index)
        , m_result(NULL)
        , m_ternary(NULL)
        , m_microcode_emit_cnt(0)
    { ; }

    virtual ~c_sub_exp_oper_ternary_false() { ; }

    virtual string emit_human_desc(string::size_type indent);
    virtual string emit_human_desc_no_recursion(string::size_type indent);

    virtual int get_prec();

    const c_oper_desc& get_oper() const;

    bool add_eval(c_sub_exp* el);

    virtual bool complete(tree<c_sub_exp*>& UNUSEDARG(t));

    virtual bool emit_evaluation_order(list<c_sub_exp*>& st, list<c_sub_exp*>::iterator where);

    virtual void compile(c_tree_microcode& tm_callback);

    virtual string emit_human_microcode(string::size_type indent);
    virtual string emit_human_microcode_no_recursion(string::size_type indent);

    virtual void emit_microcode(c_expression& expression, c_micro_symbol_variable_mapping& symbols);

    void set_result(c_sub_exp* result);
    c_sub_exp* get_result();

    void set_ternary(c_sub_exp_oper_ternary* ternary);
    c_sub_exp_oper_ternary* get_ternary();

    virtual const c_exp_sanity* can_follow(c_sub_exp& right);

    virtual const c_exp_sanity* can_precede(c_sub_exp_root* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_root_end_dummy* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_parenthesis* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_index* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_variable* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_numeric_immediate* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_const* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_func* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_eval_str* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_immediate_str* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_oper* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_oper_ternary* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_oper_ternary_true* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_oper_ternary_false* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_contitional_node_helper* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_end_parenthesis* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_end_index* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_func_end_args* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_func_arg_separator* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_conditional_end_marker* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_dummy* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_ternary_true_end_marker* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_ternary_false_end_marker* left);
};

class c_sub_exp_contitional_node_helper : public c_sub_exp
{
    vector<c_sub_exp*> m_eval;
    c_sub_exp_oper* m_conditional_oper;

public:
    c_sub_exp_contitional_node_helper(c_sub_exp_oper* conditional_oper)
        : m_conditional_oper(conditional_oper)
    { ; }

    virtual ~c_sub_exp_contitional_node_helper() { ; }

    virtual string emit_human_desc(string::size_type UNUSEDARG(indent)) { return (string)""; }
    virtual string emit_human_desc_no_recursion(string::size_type UNUSEDARG(indent)) { return (string)""; }

    void add_sub(c_sub_exp* sub);

    virtual int get_prec();

    virtual bool complete(tree<c_sub_exp*>& UNUSEDARG(t));

    virtual bool emit_evaluation_order(list<c_sub_exp*>& st, list<c_sub_exp*>::iterator where);

    virtual void compile(c_tree_microcode& tm_callback);

    virtual string emit_human_microcode(string::size_type indent);
    virtual string emit_human_microcode_no_recursion(string::size_type indent);

    virtual void emit_microcode(c_expression& expression, c_micro_symbol_variable_mapping& symbols);

    const vector<c_sub_exp*>& get_eval() const
    {
        return m_eval;
    }

    virtual const c_exp_sanity* can_follow(c_sub_exp& right);

    virtual const c_exp_sanity* can_precede(c_sub_exp_root* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_root_end_dummy* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_parenthesis* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_index* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_variable* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_numeric_immediate* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_const* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_func* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_eval_str* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_immediate_str* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_oper* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_oper_ternary* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_oper_ternary_true* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_oper_ternary_false* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_contitional_node_helper* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_end_parenthesis* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_end_index* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_func_end_args* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_func_arg_separator* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_conditional_end_marker* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_dummy* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_ternary_true_end_marker* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_ternary_false_end_marker* left);
};

// helpers follows

class c_sub_exp_end_parenthesis : public c_sub_exp
{
public:
    c_sub_exp_end_parenthesis()
        : c_sub_exp()
    { ; }

    virtual ~c_sub_exp_end_parenthesis() { ; }

    virtual string emit_human_desc(string::size_type indent);

    virtual int get_prec();

    virtual void compile(c_tree_microcode& UNUSEDARG(tm_callback)) { ; }

    virtual string emit_human_microcode(string::size_type indent);

    virtual void emit_microcode(c_expression& UNUSEDARG(expression), c_micro_symbol_variable_mapping& UNUSEDARG(symbols)) { ; }

    virtual bool complete(tree<c_sub_exp*>& UNUSEDARG(t)) { return true; }

    virtual const c_exp_sanity* can_follow(c_sub_exp& right);

    virtual const c_exp_sanity* can_precede(c_sub_exp_root* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_root_end_dummy* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_parenthesis* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_index* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_variable* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_numeric_immediate* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_const* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_func* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_eval_str* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_immediate_str* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_oper* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_oper_ternary* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_oper_ternary_true* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_oper_ternary_false* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_contitional_node_helper* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_end_parenthesis* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_end_index* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_func_end_args* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_func_arg_separator* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_conditional_end_marker* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_dummy* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_ternary_true_end_marker* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_ternary_false_end_marker* left);
};

class c_sub_exp_end_index : public c_sub_exp
{
public:
    c_sub_exp_end_index()
        : c_sub_exp()
    { ; }

    virtual ~c_sub_exp_end_index() { ; }

    virtual string emit_human_desc(string::size_type indent);

    virtual int get_prec();

    virtual void compile(c_tree_microcode& UNUSEDARG(tm_callback)) { ; }

    virtual string emit_human_microcode(string::size_type indent);

    virtual void emit_microcode(c_expression& UNUSEDARG(expression), c_micro_symbol_variable_mapping& UNUSEDARG(symbols)) { ; }

    virtual bool complete(tree<c_sub_exp*>& UNUSEDARG(t)) { return true; }

    virtual const c_exp_sanity* can_follow(c_sub_exp& right);

    virtual const c_exp_sanity* can_precede(c_sub_exp_root* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_root_end_dummy* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_parenthesis* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_index* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_variable* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_numeric_immediate* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_const* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_func* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_eval_str* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_immediate_str* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_oper* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_oper_ternary* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_oper_ternary_true* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_oper_ternary_false* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_contitional_node_helper* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_end_parenthesis* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_end_index* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_func_end_args* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_func_arg_separator* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_conditional_end_marker* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_dummy* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_ternary_true_end_marker* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_ternary_false_end_marker* left);
};

class c_sub_exp_func_end_args : public c_sub_exp
{
public:
    c_sub_exp_func_end_args()
        : c_sub_exp()
    { ; }

    virtual ~c_sub_exp_func_end_args() { ; }

    virtual string emit_human_desc(string::size_type indent);

    virtual int get_prec();

    virtual void compile(c_tree_microcode& UNUSEDARG(tm_callback)) { ; }

    virtual string emit_human_microcode(string::size_type indent);

    virtual void emit_microcode(c_expression& UNUSEDARG(expression), c_micro_symbol_variable_mapping& UNUSEDARG(symbols)) { ; }

    virtual bool complete(tree<c_sub_exp*>& UNUSEDARG(t)) { return true; }

    virtual const c_exp_sanity* can_follow(c_sub_exp& right);

    virtual const c_exp_sanity* can_precede(c_sub_exp_root* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_root_end_dummy* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_parenthesis* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_index* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_variable* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_numeric_immediate* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_const* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_func* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_eval_str* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_immediate_str* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_oper* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_oper_ternary* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_oper_ternary_true* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_oper_ternary_false* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_contitional_node_helper* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_end_parenthesis* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_end_index* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_func_end_args* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_func_arg_separator* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_conditional_end_marker* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_dummy* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_ternary_true_end_marker* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_ternary_false_end_marker* left);
};

class c_sub_exp_func_arg_separator : public c_sub_exp
{
public:
    c_sub_exp_func_arg_separator()
        : c_sub_exp()
    { ; }

    virtual ~c_sub_exp_func_arg_separator() { ; }

    virtual string emit_human_desc(string::size_type indent);

    virtual int get_prec();

    virtual void compile(c_tree_microcode& UNUSEDARG(tm_callback)) { ; }

    virtual string emit_human_microcode(string::size_type indent);

    virtual void emit_microcode(c_expression& UNUSEDARG(expression), c_micro_symbol_variable_mapping& UNUSEDARG(symbols)) { ; }

    virtual bool complete(tree<c_sub_exp*>& UNUSEDARG(t)) { return true; }

    virtual const c_exp_sanity* can_follow(c_sub_exp& right);

    virtual const c_exp_sanity* can_precede(c_sub_exp_root* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_root_end_dummy* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_parenthesis* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_index* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_variable* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_numeric_immediate* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_const* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_func* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_eval_str* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_immediate_str* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_oper* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_oper_ternary* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_oper_ternary_true* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_oper_ternary_false* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_contitional_node_helper* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_end_parenthesis* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_end_index* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_func_end_args* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_func_arg_separator* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_conditional_end_marker* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_dummy* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_ternary_true_end_marker* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_ternary_false_end_marker* left);
};

class c_tree_groupper;

class c_sub_exp_conditional_end_marker : public c_sub_exp
{
    vector<c_sub_exp_oper*> m_conditionals;
    c_sub_exp* m_return_to;

    bool m_complete;

public:
    c_sub_exp_conditional_end_marker()
        : m_return_to(NULL)
        , m_complete(false)
    { ; }

    virtual ~c_sub_exp_conditional_end_marker() { ; }

    virtual string emit_human_desc(string::size_type UNUSEDARG(indent)) { return (string)""; };

    virtual int get_prec();

    virtual bool emit_evaluation_order(list<c_sub_exp*>& st, list<c_sub_exp*>::iterator where);

    virtual void compile(c_tree_microcode& tm_callback);

    virtual string emit_human_microcode(string::size_type UNUSEDARG(indent)) { return (string)""; }

    virtual void emit_microcode(c_expression& expression, c_micro_symbol_variable_mapping& symbols);

    virtual bool complete(tree<c_sub_exp*>& UNUSEDARG(t));
    virtual void set_complete();
    bool all_complete(c_tree_groupper& tg, tree<c_sub_exp*>& t);

    void add_conditional(c_sub_exp_oper* el);

    void set_return_to(c_sub_exp* el);
    c_sub_exp* get_return_to();

    virtual const c_exp_sanity* can_follow(c_sub_exp& right);

    virtual const c_exp_sanity* can_precede(c_sub_exp_root* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_root_end_dummy* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_parenthesis* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_index* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_variable* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_numeric_immediate* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_const* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_func* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_eval_str* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_immediate_str* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_oper* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_oper_ternary* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_oper_ternary_true* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_oper_ternary_false* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_contitional_node_helper* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_end_parenthesis* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_end_index* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_func_end_args* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_func_arg_separator* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_conditional_end_marker* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_dummy* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_ternary_true_end_marker* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_ternary_false_end_marker* left);
};

class c_sub_exp_dummy : public c_sub_exp
{
    int m_prec;

public:
    c_sub_exp_dummy(int prec)
        : m_prec(prec)
    { ; }

    virtual ~c_sub_exp_dummy() { ; }

    virtual string emit_human_desc(string::size_type UNUSEDARG(indent)) { return (string)""; };

    virtual int get_prec() { return m_prec; }

    virtual bool emit_evaluation_order(list<c_sub_exp*>& UNUSEDARG(st), list<c_sub_exp*>::iterator UNUSEDARG(where)) { return true; }

    virtual void compile(c_tree_microcode& UNUSEDARG(tm_callback)) { ; }

    virtual string emit_human_microcode(string::size_type UNUSEDARG(indent)) { return (string)""; }

    virtual void emit_microcode(c_expression& UNUSEDARG(expression), c_micro_symbol_variable_mapping& UNUSEDARG(symbols)) { ; }

    virtual bool complete(tree<c_sub_exp*>& UNUSEDARG(t)) { return true; }
    virtual void set_complete() { ; }

    virtual const c_exp_sanity* can_follow(c_sub_exp& right);

    virtual const c_exp_sanity* can_precede(c_sub_exp_root* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_root_end_dummy* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_parenthesis* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_index* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_variable* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_numeric_immediate* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_const* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_func* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_eval_str* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_immediate_str* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_oper* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_oper_ternary* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_oper_ternary_true* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_oper_ternary_false* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_contitional_node_helper* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_end_parenthesis* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_end_index* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_func_end_args* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_func_arg_separator* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_conditional_end_marker* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_dummy* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_ternary_true_end_marker* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_ternary_false_end_marker* left);
};

class c_sub_exp_ternary_false_end_marker;

class c_sub_exp_ternary_true_end_marker : public c_sub_exp
{
    c_sub_exp_oper_ternary* m_ternary_true;
    c_sub_exp_ternary_false_end_marker* m_ternary_jump;

public:
    c_sub_exp_ternary_true_end_marker()
        : m_ternary_true(NULL)
        , m_ternary_jump(NULL)
    { ; }

    virtual ~c_sub_exp_ternary_true_end_marker() { ; }

    virtual string emit_human_desc(string::size_type UNUSEDARG(indent)) { return (string)""; };

    virtual int get_prec();

    virtual bool emit_evaluation_order(list<c_sub_exp*>& st, list<c_sub_exp*>::iterator where);

    virtual void compile(c_tree_microcode& tm_callback);

    virtual string emit_human_microcode(string::size_type UNUSEDARG(indent)) { return (string)""; }

    virtual void emit_microcode(c_expression& expression, c_micro_symbol_variable_mapping& symbols);

    virtual bool complete(tree<c_sub_exp*>& UNUSEDARG(t));

    void set_true(c_sub_exp_oper_ternary* ternary_true);
    void set_jump(c_sub_exp_ternary_false_end_marker* marker);

    virtual const c_exp_sanity* can_follow(c_sub_exp& right);

    virtual const c_exp_sanity* can_precede(c_sub_exp_root* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_root_end_dummy* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_parenthesis* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_index* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_variable* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_numeric_immediate* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_const* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_func* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_eval_str* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_immediate_str* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_oper* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_oper_ternary* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_oper_ternary_true* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_oper_ternary_false* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_contitional_node_helper* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_end_parenthesis* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_end_index* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_func_end_args* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_func_arg_separator* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_conditional_end_marker* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_dummy* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_ternary_true_end_marker* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_ternary_false_end_marker* left);
};

class c_sub_exp_ternary_false_end_marker : public c_sub_exp
{
    c_sub_exp_oper_ternary* m_ternary_true;

public:
    c_sub_exp_ternary_false_end_marker()
        : m_ternary_true(NULL)
    { ; }

    virtual ~c_sub_exp_ternary_false_end_marker() { ; }

    virtual string emit_human_desc(string::size_type UNUSEDARG(indent)) { return (string)""; };

    virtual int get_prec();

    virtual bool emit_evaluation_order(list<c_sub_exp*>& st, list<c_sub_exp*>::iterator where);

    virtual void compile(c_tree_microcode& tm_callback);

    virtual string emit_human_microcode(string::size_type UNUSEDARG(indent)) { return (string)""; }

    virtual void emit_microcode(c_expression& expression, c_micro_symbol_variable_mapping& symbols);

    virtual bool complete(tree<c_sub_exp*>& UNUSEDARG(t));

    void set_true(c_sub_exp_oper_ternary* ternary_true);

    virtual const c_exp_sanity* can_follow(c_sub_exp& right);

    virtual const c_exp_sanity* can_precede(c_sub_exp_root* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_root_end_dummy* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_parenthesis* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_index* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_variable* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_numeric_immediate* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_const* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_func* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_eval_str* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_immediate_str* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_oper* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_oper_ternary* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_oper_ternary_true* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_oper_ternary_false* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_contitional_node_helper* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_end_parenthesis* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_end_index* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_func_end_args* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_func_arg_separator* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_conditional_end_marker* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_dummy* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_ternary_true_end_marker* left);
    virtual const c_exp_sanity* can_precede(c_sub_exp_ternary_false_end_marker* left);
};

#if defined(_MSC_VER) && _MSC_VER>=1400
#   pragma warning(pop)
#endif

#endif

