/***************************************************************************
                          function.cpp  -  description
                             -------------------
    begin                : Fri Sep 30 2005
    copyright            : (C) 2005-2007 by Marian VooDooMan Meravy
    email                : ghostvoodooman -a.T= users (dot) sourceforge.net
 ***************************************************************************/

/* $Id: function.cpp,v 1.53 2008/08/23 15:10:39 ghostvoodooman Exp $ */

/***************************************************************************

VooDoo cIRCle - an IRC (ro)bot
Copyright (C) 2005-2008 by Marian VooDooMan Meravy
  (ghostvoodooman -a.T= users (dot) sourceforge.net)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************/

/*!
    \file
    \brief Functions implementation
    \version $Id: function.cpp,v 1.53 2008/08/23 15:10:39 ghostvoodooman Exp $
*/

#if defined(HAVE_CONFIG_H)
#   include <config.h>
#endif

#include <string>
#include <vector>
#include <stack>
#include <list>
#include <map>

#include <memory.h>
#include <string.h>
#include <stdlib.h>

using namespace std;

#include "utils/memleak.h"
#include "utils/assert.h"
#include "utils.h"
#include "params.h"
#include "compat.h"
#include "function.h"
#include "variable.h"
#include "shared.h"
#include "irc.h"
#include "sock_shared.h"
#include "logic.h"
#include "conf.h"
#include "botnet.h"
#include "dcc.h"
#include "log.h"
#include "identd.h"
#include "match.h"
#include "stats.h"
#include "modules.h"
#include "threads.h"
#include "proc.h"
#include "eval.h"
#include "casemap.h"
#include "var/c_var_array.h"
#include "var/c_var_bool.h"
#include "var/c_var_int.h"
#include "var/c_var_string.h"
#include "runtime/c_delay_link.h"
#include "runtime/c_delay_command.h"
#include "irc/c_irc_server.h"
#include "irc/c_online_channel.h"
#include "md5.h"
#include "microcode/c_micro_symbol_variable_mapping.h"
#include "func/c_fn_cmd_do.h"
#include "func/c_fn_cmd_case_default.h"
#include "func/c_fn_cmd_continue.h"
#include "func/c_fn_cmd_do_while.h"
#include "func/c_fn_cmd_end_of_func.h"
#include "func/c_fn_cmd_for_cond.h"
#include "func/c_fn_cmd_global.h"
#include "func/c_fn_cmd_goto.h"
#include "func/c_fn_cmd_label.h"
#include "func/c_fn_cmd_left_brace.h"
#include "func/c_fn_cmd_right_brace.h"
#include "func/c_fn_cmd_break.h"
#include "func/c_expression.h"
#include "func/c_command.h"
#include "func/c_fn_cmd_while.h"
#include "func/c_fn_cmd_jump.h"
#include "func/c_fn_cmd_foreach_next_iter.h"
#include "func/c_fn_cmd_dummy.h"
#include "utils/parse_flags.h"

#include "utils/precomp.h"

#if defined(_MSC_VER) || defined(__BORLANDC__)
#   pragma hdrstop
#endif

c_var_base* explode_function(c_module* this_module, const string& separator, const string& str, MY_LONG_LONG limit, const string& debug)
{
    if(separator.length()==0)
        return MY_NEW c_var_bool(false);
    c_var_array* res=MY_NEW c_var_array();
    if(limit<=1) {
        c_index_base* idx=res->get_new_int_index();
        c_var_string* s=MY_NEW c_var_string(str);
        res->set(this_module,idx,s,false,debug);
        return res;
    }
    string last_str;
    MY_LONG_LONG got=1;
    char* separ=(char*)my_malloc(separator.length()+16);
    if(!separ)
        return res;
    memset(separ,0,separator.length()+16);
    strcpy(separ,separator.c_str());
    char* buff=(char*)my_malloc(str.length()+separator.length()+16);
    if(!buff) {
        my_free(separ);
        return res;
    }
    memset(buff,0,str.length()+separator.length()+16);
    memcpy(buff,str.c_str(),str.length());
    char* start=buff;
    size_t len=str.length();
    char* eob=start+len;
    while(start<=eob) {
        char* end=NULL;
        if(start<eob)
            end=strstr(start,separ);
        if(NULL==end) {
            char* b=MY_NEW char[strlen(start)+1];
            memcpy(b,start,strlen(start));
            b[strlen(start)]='\0';
            last_str=b;
            MY_DELETE_A(b);
            b=NULL;
            break;
        }
        char* b=MY_NEW char[(size_t)(end-start)+1];
        memcpy(b,start,end-start);
        b[(size_t)(end-start)]='\0';
        last_str+=b;
        MY_DELETE_A(b);
        b=NULL;
        start=end+strlen(separ);
        if(got<limit) {
            got++;
            c_index_base* idx=res->get_new_int_index();
            c_var_string* s=MY_NEW c_var_string(last_str);
            res->set(this_module,idx,s,false,debug);
            last_str="";
            if(got==limit) {
                last_str=start;
                break;
            }
        }
    }
    if(last_str.length()) {
        c_index_base* idx=res->get_new_int_index();
        c_var_string* s=MY_NEW c_var_string(last_str);
        res->set(this_module,idx,s,false,debug);
    }
    my_free(separ);
    my_free(buff);
    return res;
}

bool eval_function(c_module* UNUSEDARG(this_module), const string& code_str, c_var_array* UNUSEDARG(vars), c_functions& UNUSEDARG(functions), const string& debug, stack<c_eval_iteration>& st)
{
    c_function* f=MY_NEW c_function();
    string error_msg;
    string code="function __eval__(){";
    code+=code_str;
    code+="}";

    bool okay=true;

    bool error=false;
    bool need_more=false;
    string v="#version 2";
    f->parse_command(st.top().this_module_,v,error,error_msg,need_more,debug,"");
    if(error || !need_more) {
        f->clear_microcode();
        f->destroy_symbols();
        f->destroy(true,true);
        MY_DELETE(f);
        f=NULL;
        okay=false;
    } else
        f->parse_command(st.top().this_module_,code,error,error_msg,need_more,debug,"");
    if(error || need_more || !f->is_okay()) {
        if(f) {
            f->clear_microcode();
            f->destroy_symbols();
            f->destroy(true,true);
            MY_DELETE(f);
            f=NULL;
        }
        okay=false;
    }
    if(!okay) {
        /*
        string error="WARNING: 'eval' built-in function parse error";
        var_oops(st.top().this_module_,__FILE__,__my_func__,error,st.top().proc_cmd.get_debug());
        */
        if(f) {
            f->clear_microcode();
            f->destroy_symbols();
            f->destroy(true,true);
            MY_DELETE(f);
            f=NULL;
        }
        error_msg=((string)"'eval' built-in function parse error: ")+error_msg;
        c_eval_runtime_error ee(st.top().this_module_,debug,error_msg,__FILE__,__my_func__);
        ee.report(false);

        return false;
    } else {
        c_eval_iteration ste(st.top().this_module_,st.top().vars_,debug);
        ste.fn_destroy=true;
        ste.fn_destroy_mc=true;
        ste.fn=f;
        //ste.fn->create_symbols();
        ste.fn_destroy=true;
        //ste.symbol_table_=ste.fn->make_symbol_table(ste.fn->symbol_table);
        ste.symbol_table_=ste.fn->symbol_table;
        ste.proc_eval=true;
        ste.state=1000;
        ste.return_to_state=st.top().state;
        ste.parent=&(st.top());
        st.push(ste);
        return true;
    }
}

#if defined(_MSC_VER)
#   pragma warning(push)
#   pragma warning(disable:4706)    // warning C4706: assignment within conditional expression
#endif

string get_s_arg(c_module* this_module, vector<c_func_arg>& args, size_t num, const string& UNUSEDARG(debug), string& error, bool& ok)
{
    if(!ok)
        return (string)"";

    MY_ASSERT(0!=num);

    c_var_base* val;
    if(args.size()<num || !(val=args[num-1].value)) {
        error="Cannot get argument (one-based index) #";
        error+=tos(num);
        ok=false;
        return (string)"";
    }

    if(val->get_var_type()!=VARTYPE_STRING) {
        if(val->get_var_type()!=VARTYPE_INT && val->get_var_type()!=VARTYPE_BOOL) {
            error="Argument #";
            error+=tos(num);
            error+=" is not STRING, it is ";
            error+=val->get_var_type_as_string();
            ok=false;
            return (string)"";
        }
        return val->to_string(this_module);
    }
    c_var_string* s=NULL;
    if(val && typeid(*val)==typeid(c_var_string))
        s=dynamic_cast<c_var_string*>(val);
    if(!s) {
        error="Argument #";
        error+=tos(num);
        error+=" C++ dynamic_cast<c_var_string*>(arg) returned NULL on variable of type ";
        error+=val->get_var_type_as_string();
        ok=false;
        return (string)"";
    }

    return s->get();
}

#if defined(_MSC_VER)
#   pragma warning(pop)
#endif

#if defined(_MSC_VER)
#   pragma warning(push)
#   pragma warning(disable:4706)    // warning C4706: assignment within conditional expression
#endif

MY_LONG_LONG get_i_arg(c_module* this_module, vector<c_func_arg>& args, size_t num, const string& debug, string& error, bool& ok)
{
    if(!ok)
        return 0;

    MY_ASSERT(0!=num);

    c_var_base* val;
    if(args.size()<num || !(val=args[num-1].value)) {
        error="Cannot get argument (one-based index) #";
        error+=tos(num);
        ok=false;
        return 0;
    }

    if(val->get_var_type()!=VARTYPE_INT) {
        switch(val->get_var_type()) {
            case VARTYPE_STRING:
                {
                    c_var_string* s=dynamic_cast<c_var_string*>(val);
                    if(!s) {
                        error="Argument #";
                        error+=tos(num);
                        error+=": C++ dynamic_cast<c_var_string*>((*i1).value) returned NULL on variable of type ";
                        error+=val->get_var_type_as_string();
                        ok=false;
                        return 0;
                    }
                    c_var_int* res=s->to_int(this_module,c_var_string::RET_NULL,debug);
                    if(!res) {
                        error="Argument #";
                        error+=tos(num);
                        error+=": Cannot get value of INT, variable is of type STRING and contains \"";
                        error+=s->get();
                        error+="\"";
                        ok=false;
                        return 0;
                    }
                    MY_LONG_LONG i=res->get();
                    MY_DELETE(res);
                    return i;
                }
                // NOTREACH
            case VARTYPE_BOOL:
                {
                    c_var_bool* b=dynamic_cast<c_var_bool*>(val);
                    if(!b) {
                        error="Argument #";
                        error+=tos(num);
                        error+=": C++ dynamic_cast<c_var_bool*>((*i1).value) returned NULL on variable of type ";
                        error+=val->get_var_type_as_string();
                        ok=false;
                        return 0;
                    }
                    return b->to_int();
                }
            default:
                break;
        }
        error="Argument #";
        error+=tos(num);
        error+=" is not INT, it is ";
        error+=val->get_var_type_as_string();
        ok=false;
        return 0;
    }
    c_var_int* i=NULL;
    if(val && typeid(*val)==typeid(c_var_int))
        i=dynamic_cast<c_var_int*>(val);
    if(!i) {
        error="Argument #";
        error+=tos(num);
        error+=" C++ dynamic_cast<c_var_int*>(arg) returned NULL on variable of type ";
        error+=val->get_var_type_as_string();
        ok=false;
        return 0;
    }

    return i->get();
}

#if defined(_MSC_VER)
#   pragma warning(pop)
#endif

c_var_base* get_b_arg(c_module* UNUSEDARG(this_module), vector<c_func_arg>& args, size_t num, const string& UNUSEDARG(debug), string& error, bool& ok)
{
    if(!ok)
        return NULL;

    MY_ASSERT(0!=num);

    if(args.size()<num) {
        error="Cannot get argument (one-based index) #";
        error+=tos(num);
        ok=false;
        return NULL;
    }

    /*
    if(!(*i1).value) {
        error="Internal error: Argument (one-based index) #";
        error+=tos(num);
        error+=" is NULL";
        ok=false;
        return 0;
    }
    */

    return args[num-1].value;
}

var_sym_index_t get_var_sym_index(c_module* UNUSEDARG(this_module), vector<c_func_arg>& args, size_t num, const string& UNUSEDARG(debug), string& error, bool& ok, bool need_a_reference_to_var)
{
    if(!ok)
        return VAR_SYM_INDEX_INVALID;

    MY_ASSERT(0!=num);

    if(args.size()<num) {
        error="Cannot get argument (one-based index) #";
        error+=tos(num);
        ok=false;
        return VAR_SYM_INDEX_INVALID;
    }

    if(need_a_reference_to_var && !args[num-1].reference) {
        error="Cannot get argument, it is needed to be a $variable (one-based index) #";
        error+=tos(num);
        ok=false;
        return VAR_SYM_INDEX_INVALID;
    }

    return args[num-1].value_sym;
}

#if defined(_MSC_VER)
#   pragma warning(push)
#   pragma warning(disable:4706)    // warning C4706: assignment within conditional expression
#endif

c_var_array* get_a_arg(c_module* UNUSEDARG(this_module), vector<c_func_arg>& args, size_t num, const string& UNUSEDARG(debug), string& error, bool& ok)
{
    if(!ok)
        return NULL;

    MY_ASSERT(0!=num);

    c_var_base* val;
    if(args.size()<num || !(val=args[num-1].value)) {
        error="Cannot get argument (one-based index) #";
        error+=tos(num);
        ok=false;
        return NULL;
    }

    if(val->get_var_type()!=VARTYPE_ARRAY) {
        error="Argument #";
        error+=tos(num);
        error+=" is not ARRAY, it is ";
        error+=val->get_var_type_as_string();
        ok=false;
        return NULL;
    }
    c_var_array* a=NULL;
    if(val && typeid(*val)==typeid(c_var_array))
        a=dynamic_cast<c_var_array*>(val);
    if(!a) {
        error="Argument #";
        error+=tos(num);
        error+=" C++ dynamic_cast<c_var_array*>(arg) returned NULL on variable of type ";
        error+=val->get_var_type_as_string();
        ok=false;
        return NULL;
    }

    return a;
}

#if defined(_MSC_VER)
#   pragma warning(pop)
#endif

bool set_arg(c_module* this_module, vector<c_func_arg>& args, c_var_base* to_set, size_t num, const string& debug, string& error, bool& ok)
{
    if(!ok)
        return false;

    MY_ASSERT(0!=num);

    if(args.size()<num) {
        MY_DELETE(to_set);
        error="Cannot set argument #";
        error+=tos(num);
        ok=false;
        return 0;
    }

    c_var_base*& val=args[num-1].value;

    if(val && val->get_read_only()) {
        MY_DELETE(to_set);
        if(this_module && this_module->get_logger()) {
            string log="OOPS: Cannot modify read-only variable; Additional info: ";
            log+=debug;
            this_module->get_logger()->log_print1(E_LOG_DEBUG,LOG_SCOPE_MODULE,"",GET_STR_DEBUG_INFO_EX(log));
        }
        return true;
    }

    if(!args[num-1].reference) {
        MY_DELETE(to_set);
        if(this_module && this_module->get_logger()) {
            string log="OOPS: Cannot modify argument variable - it should be a variable, not an expression; Additional info: ";
            log+=debug;
            this_module->get_logger()->log_print1(E_LOG_DEBUG,LOG_SCOPE_MODULE,"",GET_STR_DEBUG_INFO_EX(log));
        }
        return true;
    }

    args[num-1].value=to_set;

    return true;
}

bool bad_num_args(const string& UNUSEDARG(debug), string& error)
{
    error="Bad number of arguments";
    return false;
}

string parse_get_func(string& str, bool& error, bool& need_more)
{
    error=false;
    need_more=false;
    string res;
    int ctx=0;
    for(size_t i1=0; i1<str.length(); ++i1) {
        if(0x20==str[i1]) {
            continue;
        }
        if('{'==str[i1] || '}'==str[i1]) {
            error=true;
            return (string)"";
        }
        if('('==str[i1]) {
            ctx++;
            res+='(';
            continue;
        }
        if(')'==str[i1] && ctx!=1) {
            error=true;
            return (string)"";
        }
        if(')'==str[i1]) {
            res+=')';
            str.erase(0,i1+1);
            return res;
        }
        res+=str[i1];
    }
    need_more=true;
    return (string)"";
}

string explode_command(string& str, bool& error, bool& need_more, c_command::e_type& type, bool& semicolon, string& sub_exp)
{
    error=false;
    need_more=false;
    semicolon=false;
    sub_exp="";

    string res;

    bool in_esc=false;
    char in_str='\0';
    bool in_sub=false;
    string main_exp;
    stack<char> parenthesis;
    for(size_t i1=0; i1<str.length(); ++i1) {
        if(in_esc) {
            res+=str[i1];
            in_esc=false;
            continue;
        }
        if(0==in_str && '\''==str[i1]) {
            res+="'";
            in_str='\'';
            continue;
        }
        if('\''==in_str) {
            if('\\'==str[i1]) {
                res+=str[i1];
                in_esc=true;
                continue;
            }
            res+=str[i1];
            if('\''==str[i1]) {
                in_str='\0';
            }
            continue;
        }
        if('\\'==str[i1] && '\''!=in_str) {
            res+=str[i1];
            in_esc=true;
            continue;
        }
        if('"'==str[i1]) {
            res+=str[i1];
            if('"'==in_str)
                in_str='\0';
            else
                in_str='"';
            continue;
        }
        if(in_str) {
            res+=str[i1];
            continue;
        }
        if(('\x20'==str[i1] || '\t'==str[i1] || '\r'==str[i1] || '\n'==str[i1])) {
            if(res.empty() || '\x20'==res[res.length()-1])
                continue;
            res+=" ";
            continue;
        }
        if('\x20'==str[i1] || '\t'==str[i1]) {
            size_t pos=res.rfind("return");
            if(pos!=string::npos && pos==res.length()-strlen("return"))
                res+=" ";
            pos=res.rfind("goto");
            if(pos!=string::npos && pos==res.length()-strlen("goto"))
                res+=" ";
            pos=res.rfind("global");
            if(pos!=string::npos && pos==res.length()-strlen("global"))
                res+=" ";
            pos=res.rfind("case");
            if(pos!=string::npos && pos==res.length()-strlen("case"))
                res+=" ";
            pos=res.rfind("do");
            if(pos!=string::npos && pos==res.length()-strlen("do"))
                res+=" ";
            pos=res.rfind("and");
            if(pos!=string::npos && pos==res.length()-strlen("and")) {
                res.insert(pos," ");
                res+=" ";
            }
            pos=res.rfind("xor");
            if(pos!=string::npos && pos==res.length()-strlen("xor")) {
                res.insert(pos," ");
                res+=" ";
            }
            else {
                pos=res.rfind("or");
                if(pos!=string::npos && pos==res.length()-strlen("or")) {
                    res.insert(pos," ");
                    res+=" ";
                }
            }
            pos=res.rfind("as");
            if(pos!=string::npos && pos==res.length()-strlen("as"))
                res+=" ";
            if(str.length()>=i1+3 && 'a'==str[i1+1] && 's'==str[i1+2])
                res+=" ";
            continue;
        }
        if('('==str[i1]) {
            parenthesis.push('(');
            res+=str[i1];
            continue;
        }
        if('['==str[i1]) {
            parenthesis.push('[');
            res+=str[i1];
            continue;
        }
        if(')'==str[i1]) {
            if(parenthesis.empty() || '('!=parenthesis.top()) {
                error=true;
                return res;
            }
            parenthesis.pop();
            res+=str[i1];
            if(parenthesis.empty() && res.find("for(")==0) {
                str.erase(0,i1+1);
                type=c_command::CMD_FOR;
                return res;
            }
            if(parenthesis.empty() && res.find("while(")==0) {
                str.erase(0,i1+1);
                res+=";"; // will be removed later, it's just helper
                type=c_command::CMD_WHILE;
                return res;
            }
            if(parenthesis.empty() && res.find("foreach(")==0) {
                str.erase(0,i1+1);
                type=c_command::CMD_FOREACH_1;
                return res;
            }
            if(parenthesis.empty() && (res.find("timer_every(")==0 || res.find("timer_once(")==0)) {
                sub_exp="";
                main_exp=res;
                main_exp+=";"; // helper
                res="";
                in_sub=true;
            }
            continue;
        }
        if(']'==str[i1]) {
            if(parenthesis.empty() || '['!=parenthesis.top()) {
                error=true;
                return res;
            }
            parenthesis.pop();
            res+=str[i1];
            continue;
        }
        if(parenthesis.empty() && res.length()>=3 && 'i'==res[0] && 'f'==res[1] && '('==res[2]) {
            str.erase(0,i1+1-1); // exclude this one which is next!
            res+=";"; // will be removed later, it's just helper
            type=c_command::CMD_IF_EXP;
            return res;
        }
        if(parenthesis.empty() && (res.find("if_error(")==0 || res.find("wait_if_error(")==0)) {
            str.erase(0,i1+1-1); // exclude this one which is next!
            res+=";"; // will be removed later, it's just helper
            type=c_command::CMD_IF_EXP;
            return res;
        }
        if(parenthesis.empty() && res.find("switch(")==0) {
            str.erase(0,i1+1-1); // exclude this one which is next!
            res+=";"; // will be removed later, it's just helper
            type=c_command::CMD_SWITCH;
            return res;
        }
        if(!parenthesis.empty() && '='==str[i1] && res.find("foreach(")==0) {
            res+="=";
            continue;
        }
        if(!parenthesis.empty() && '>'==str[i1] && res.find("foreach(")==0) {
            res+=">";
            continue;
        }
        if(!parenthesis.empty() && '&'==str[i1] && res.find("foreach(")==0) {
            res+="&";
            continue;
        }
        if('{'==str[i1]) {
            if(!res.empty()) {
                str.erase(0,i1+1-1); // exclude this '{' !!!
                res+=";"; // will be removed later, it's just helper
                type=c_command::CMD_EXP;
                return res;
            }
            str.erase(0,i1+1);
            type=c_command::CMD_LEFT_BRACKET;
            return res;
        }
        if('}'==str[i1]) {
            str.erase(0,i1+1);
            type=c_command::CMD_RIGHT_BRACKET;
            return res;
        }
        if(';'==str[i1]) {
            if(!parenthesis.empty()) {
                res+=";";
                continue;
            }
            if(res.find("for(",0)!=string::npos) {
                res+=";";
                continue;
            }
            str.erase(0,i1+1);
            res+=';';
            type=c_command::CMD_EXP;
            if(in_sub) {
                sub_exp=res;
                res=main_exp;
                if(res.length()) {
                    //char last=res[res.length()-1];
                    res.erase(res.length()-1,1);
                    //sub_exp=((string)""+last)+sub_exp;
                }
                type=c_command::CMD_EXP_SUB;
            }
            semicolon=true;
            return res;
        }
        if(':'==str[i1] && parenthesis.empty()) {
            bool is_ternary=false;
            for(size_t i2=0; i2<i1; ++i2) {
                if('?'==str[i2]) {
                    is_ternary=true;
                    break;
                }
            }
            if(!is_ternary) {
                type=c_command::CMD_LABEL;
                if(res.find("case")==0)
                    type=c_command::CMD_CASE;
                if(res.find("default")==0)
                    type=c_command::CMD_CASE;
                str.erase(0,i1+1);
                return res;
            }
        }
        res+=str[i1];
        if(!res.compare("else")) {
            str.erase(0,i1+1);
            res+=';';
            type=c_command::CMD_ELSE;
            return res;
        }
    }
    if(parenthesis.empty() && res.length()>=3 && 'i'==res[0] && 'f'==res[1] && '('==res[2]) {
        str="";
        res+=";"; // will be removed later, it's just helper
        type=c_command::CMD_IF_EXP;
        return res;
    }
    if(in_sub) {
        sub_exp=res;
        res=main_exp;
        if(res.length()) {
            res.erase(res.length()-1,1);
        }
        if(sub_exp.empty() || sub_exp[sub_exp.length()-1]!=';') {
            need_more=true;
            str=res;
            str+=sub_exp;
            return (string)"";
        }
        type=c_command::CMD_EXP_SUB;
        return res;
    }
    if(!semicolon) {
        need_more=true;
        return (string)"";
    }
    str="";
    need_more=true;
    return (string)"";
}

string add_slashes(const string& src)
{
    string res;
    for(size_t i1=0; i1<src.length(); ++i1) {
        if('"'==src[i1]) {
            res+="\\\"";
            continue;
        }
        res+=src[i1];
    }
    return res;
}

