/***************************************************************************
                          identd.cpp  -  description
                             -------------------
    begin                : Thu Dec 9 2004
    copyright            : (C) 2004-2007 by Marian VooDooMan Meravy
    email                : ghostvoodooman -a.T= users (dot) sourceforge.net
 ***************************************************************************/

/* $Id: identd.cpp,v 1.49 2008/08/01 08:28:30 ghostvoodooman Exp $ */

/***************************************************************************

VooDoo cIRCle - an IRC (ro)bot
Copyright (C) 2004-2008 by Marian VooDooMan Meravy
  (ghostvoodooman -a.T= users (dot) sourceforge.net)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************/

/*!
    \file
    \brief Emulates Ident daemon
    \version $Id: identd.cpp,v 1.49 2008/08/01 08:28:30 ghostvoodooman Exp $
*/

#if defined(HAVE_CONFIG_H)
#   include <config.h>
#endif

#include <string>
#include <vector>
#include <stack>
#include <list>
#include <map>

#include <stdlib.h>
#include <time.h>

#include "utils/memleak.h"
#include "utils/assert.h"
#include "utils.h"

#include "sock.h"
#include "log.h"
#include "stats.h"
#include "match.h"

#include "identd.h"

#include "compat.h"

#include "params.h"

#include "utils/precomp.h"

#if defined(_MSC_VER) || defined(__BORLANDC__)
#   pragma hdrstop
#endif

//---------------------------------------------------------------------------
#ifdef __BORLANDC__
#pragma package(smart_init)
#endif

c_socket identd_sock;                   //!< Stores socket handle of IDENT daemon
c_socket identd_client;                 //!< Socket of IDENTD client
#ifdef IPv6
c_socket identd_sock_ipv6;              //!< Stores socket handle of IDENT daemon for IPv6 connections
c_socket identd_client_ipv6;            //!< Socket of IDENTD client for IPv6 connections
#endif
my_time_t ident_last_response;             //!< Timestamp of last response

/*!
    \brief Writes OS's socket I/O error to log file
    \author VooDooMan
    \version 2
    \date 2004, 2007-08-19
    \param io_error OS's socket I/O error code
*/
void identd_error_log(int io_error)
{
    string log="I/O error, OS error code #";
    log+=tos(io_error);
    log+=": ";
    char* errstr=NULL;
    log+=errstr=sock_error(io_error);
    MY_DELETE_A(errstr);
    log_identd(GET_STR_DEBUG_INFO_EX(log));
}

/*!
    \brief Starts IDENT daemon
    \author VooDooMan
    \version 2
    \date 2005
    \param bind IPv4 Address to bind
    \param port Port to bind
    \return Returns false for error
*/
bool identd_startup(in_addr bind, unsigned short port)
{
    if(identd_sock.cmp()) {
        sock_close(identd_sock,__FILE__,__LINE__);
        identd_sock.clear();
        if(identd_client.cmp())
            sock_close(identd_client,__FILE__,__LINE__);
        identd_client.clear();
        ident_last_response=0;
    }

    int error_code=0;
    identd_sock=sock_bind(bind,port,error_code,"IDENTd IPv4 server");
    if(!identd_sock.cmp()) {
        identd_error_log(error_code);
        return false;
    }
    log_identd(GET_STR_DEBUG_INFO_EX("Started up (IPv4)"));

    return true;
}

/*!
    \brief Starts IDENT daemon
    \author VooDooMan
    \version 1
    \date 2005
    \param bind_ipv6 IPv6 Address to bind
    \param port Port to bind
    \return Returns false for error
*/
bool identd_startup(char* UNUSEDARG_NO_IPv6(bind_ipv6), unsigned short UNUSEDARG_NO_IPv6(port))
{
#ifdef IPv6
    if(identd_sock_ipv6.cmp()) {
        sock_close(identd_sock_ipv6,__FILE__,__LINE__);
        identd_sock_ipv6.clear();
        if(identd_client_ipv6.cmp())
            sock_close(identd_client_ipv6,__FILE__,__LINE__);
        identd_client_ipv6.clear();
        ident_last_response=0;
    }

    int error_code=0;
    identd_sock_ipv6=sock_bind6(bind_ipv6,port,error_code,"IDENTd IPv6 server");
    if(!identd_sock_ipv6.cmp()) {
        identd_error_log(error_code);
        return false;
    }
    log_identd(GET_STR_DEBUG_INFO_EX("Started up (IPv6)"));

    return true;
#else
    return false;
#endif
}

/*!
    \brief Checks and handles IDENT request
    \author VooDooMan
    \version 2
    \date 2005
    \param ident_string IDENT string used as response to the request
    \param local_port Local socket name
    \param remote_port Remote socket name
    \return Returns false for error or nothing done, true for successful IDENT response
*/
bool identd_check(const string& ident_string, unsigned short local_port, unsigned short remote_port)
{
    for(int i=0; i<2; ++i) {
        c_socket* identd;
        c_socket* client;
        if(i==0) {
            identd=&identd_sock;
            client=&identd_client;
        } else {
#ifdef IPv6
            identd=&identd_sock_ipv6;
            client=&identd_client_ipv6;
#else
            return false;
#endif
        }

        if(identd->cmp()==false)
            continue;

        if(!client->cmp()) {
            ident_last_response=0;
            c_socket s;
            if(i==0)
                s=sock_accept(*identd,"IDENTd client",true);
            else
                s=sock_accept6(*identd,"IDENTd client",true);
            if(s.cmp()) {
                if(i==0)
                    identd_client=s;
#ifdef IPv6
                else
                    identd_client_ipv6=s;
#endif
                stats_identd_new_connection();
            }
        }

        enum
        {
            IDENTD_WAIT_REQUEST_SECS = 20,  //!< How many seconds for IdentD to wait before shutdown
        };

        if(client->cmp()) {
            if(ident_last_response) {
                if(ident_last_response+IDENTD_WAIT_REQUEST_SECS<my_time(NULL)) {
                    ident_last_response=0;
                    sock_close(*client,__FILE__,__LINE__);
                    client->clear();
                    ident_last_response=0;
                    return true;
                }
            }
            char buff[1024];
            int ec=0;
            bool closed;
            size_t size=sock_read(*client,buff,sizeof(buff)-1,ec,closed);
            stats_identd_bytes_received(size);
            if(ec) {
                sock_close(*client,__FILE__,__LINE__);
                client->clear();
                ident_last_response=0;
                identd_error_log(ec);
                return false;
            }
            buff[size]=0;
            char tmp[2*1024];
            if(size) {
                string log="Ident request ";
                if(i==0) {
                    log+="(IPv4) ";
                    int ec=0;
                    in_addr a=sock_get_remote_addr(*client,ec);
                    char tmp[512];
                    tmp[0]=0;
                    my_inet_ntop(AF_INET,&a,tmp,sizeof(tmp)-1);
                    log+=tmp;
                    log+=" ";
                } else {
                    log+="(IPv6) ";
                    //int ec=0;
                    in_addr6_ a=sock_get_remote_addr6(*client);
                    char tmp[512];
                    tmp[0]=0;
                    my_inet_ntop(AF_INET6,&a,tmp,sizeof(tmp)-1);
                    log+=tmp;
                    log+=" ";
                }

                log+=": ";
                log+=buff;
                if(log.length() && '\r'==log[log.length()-1])
                    log.erase(log.length()-1,1);
                if(log.length() && '\n'==log[log.length()-1])
                    log.erase(log.length()-1,1);
                if(log.length() && '\r'==log[log.length()-1])
                    log.erase(log.length()-1,1);
                if(log.length() && '\n'==log[log.length()-1])
                    log.erase(log.length()-1,1);
                log_identd(GET_STR_DEBUG_INFO_EX(log));

                //int i2=0;
                int cnt=0;
                unsigned int i1;
                char num1[64], num2[64];
                int num_c=0;
                for(i1=0; i1<strlen(buff); ++i1) {
                    if(cnt==0 && buff[i1]==',') {
                        cnt++;
                        num_c=0;
                        continue;
                    }
                    if(cnt==0 && (buff[i1]<'0' || buff[i1]>'9'))
                        continue;
                    if(cnt==0 && buff[i1]!=0x20 && buff[i1]!=',') {
                        num1[num_c++]=buff[i1];
                        num1[num_c]=0;
                        if(num_c>60)
                            break;
                        continue;
                    }
                    if(cnt==1 && (buff[i1]<'0' || buff[i1]>'9'))
                        continue;
                    if(cnt==1 && buff[i1]!=0x20 && buff[i1]!=',') {
                        num2[num_c++]=buff[i1];
                        num2[num_c]=0;
                        if(num_c>60)
                            break;
                        continue;
                    }
                }

                strcpy(tmp,num1);
                strcat(tmp," , ");
                strcat(tmp,num2);
                strcat(tmp," ");

                {
                    unsigned short l, r;
                    l=stous(num1);
                    r=stous(num2);
                    if(r!=remote_port && l!=local_port) {
                        strcat(tmp,": ERROR : NO-USER");
                        strcat(tmp,"\r\n");

                        size_t sent=sock_send(*client,tmp,strlen(tmp),ec);
                        stats_identd_bytes_sent(sent);
                    } else {
                        strcat(tmp,ident_string.c_str());
                        strcat(tmp,"\r\n");

                        size_t sent=sock_send(*client,tmp,strlen(tmp),ec);
                        stats_identd_bytes_sent(sent);
                    }
                }

                if(ec) {
                    sock_close(*client,__FILE__,__LINE__);
                    client->clear();
                    ident_last_response=0;
                    identd_error_log(ec);
                    return false;
                }

                log="Ident reply: ";
                log+=tmp;
                if(log.length() && '\r'==log[log.length()-1])
                    log.erase(log.length()-1,1);
                if(log.length() && '\n'==log[log.length()-1])
                    log.erase(log.length()-1,1);
                if(log.length() && '\r'==log[log.length()-1])
                    log.erase(log.length()-1,1);
                if(log.length() && '\n'==log[log.length()-1])
                    log.erase(log.length()-1,1);
                log_identd(GET_STR_DEBUG_INFO_EX(log));
                my_time(&ident_last_response);
                return true;
            } else
                return false;
        }
    }
    return false;
}

/*!
    \brief Shuts down IDENT daemon (both IPv4 and IPv6)
    \author VooDooMan
    \version 2
    \date 2005
*/
void identd_shutdown()
{
    if(identd_sock.cmp()) {
        log_identd(GET_STR_DEBUG_INFO_EX("Shut down (IPv4)"));
        sock_close(identd_sock,__FILE__,__LINE__);
        identd_sock.clear();
        if(identd_client.cmp())
            sock_close(identd_client,__FILE__,__LINE__);
        identd_client.clear();
        ident_last_response=0;
    }

#ifdef IPv6
    if(identd_sock_ipv6.cmp()) {
        log_identd(GET_STR_DEBUG_INFO_EX("Shut down (IPv6)"));
        sock_close(identd_sock_ipv6,__FILE__,__LINE__);
        identd_sock_ipv6.clear();
        if(identd_client.cmp())
            sock_close(identd_client_ipv6,__FILE__,__LINE__);
        identd_client_ipv6.clear();
        ident_last_response=0;
    }
#endif
}

/*!
    \brief Shuts down IDENT daemon (IPv4)
    \author VooDooMan
    \version 1
    \date 2005
*/
void identd4_shutdown()
{
    if(identd_sock.cmp()) {
        log_identd(GET_STR_DEBUG_INFO_EX("Shut down (IPv4)"));
        sock_close(identd_sock,__FILE__,__LINE__);
        identd_sock.clear();
        if(identd_client.cmp())
            sock_close(identd_client,__FILE__,__LINE__);
        identd_client.clear();
        ident_last_response=0;
    }
}

/*!
    \brief Shuts down IDENT daemon (IPv6)
    \author VooDooMan
    \version 1
    \date 2005
*/
void identd6_shutdown()
{
#ifdef IPv6
    if(identd_sock_ipv6.cmp()) {
        log_identd(GET_STR_DEBUG_INFO_EX("Shut down (IPv6)"));
        sock_close(identd_sock_ipv6,__FILE__,__LINE__);
        identd_sock_ipv6.clear();
        if(identd_client_ipv6.cmp())
            sock_close(identd_client_ipv6,__FILE__,__LINE__);
        identd_client_ipv6.clear();
        ident_last_response=0;
    }
#endif
}

/*!
    \brief Resets IDENT daemon (both IPv4 and IPv6)
    \author VooDooMan
    \version 1
    \date 2005
*/
void identd_renew()
{
    if(identd_client.cmp()) {
        sock_close(identd_client,__FILE__,__LINE__);
        identd_client.clear();
        ident_last_response=0;
    }

#ifdef IPv6
    if(identd_client_ipv6.cmp()) {
        sock_close(identd_client_ipv6,__FILE__,__LINE__);
        identd_client_ipv6.clear();
        ident_last_response=0;
    }
#endif
}

