/***************************************************************************
                          c_online_user.h  -  description
                             -------------------
    begin                : Thu May 10 2007
    copyright            : (C) 2004-2007 by Marian VooDooMan Meravy
    email                : ghostvoodooman -a.T= users (dot) sourceforge.net
 ***************************************************************************/

/* $Id: c_online_user.h,v 1.15 2008/03/25 12:04:27 ghostvoodooman Exp $ */

/***************************************************************************

VooDoo cIRCle - an IRC (ro)bot
Copyright (C) 2004-2008 by Marian VooDooMan Meravy
  (ghostvoodooman -a.T= users (dot) sourceforge.net)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************/

/*!
    \file
    \brief Class 'c_online_user'
    \version $Id: c_online_user.h,v 1.15 2008/03/25 12:04:27 ghostvoodooman Exp $
*/

#ifndef C_ONLINE_USER_H__
#define C_ONLINE_USER_H__

#if defined(HAVE_CONFIG_H)
#   include <config.h>
#endif

#include <string>
#include <vector>
#include <stack>
#include <list>
#include <map>
#include <algorithm>

#include "../utils/memleak.h"

#include "../params.h"

#include "c_flood_history.h"

/*!
    \brief Stores data about connected user of channel
    \author VooDooMan
    \version 1
    \date 2004
*/
struct c_online_user
{
    string nick;                                //!< User's nick name

    bool got_whois;                             //!< Have we received WHOIS command reply yet for this user?
    bool whois_sent;                            //!< Have we sent WHOIS command for this user yet?
    bool on_join_called;                        //!< Have we tried to call on_join event yet? If on_join isn't assigned, there can be true.

    string ident;                               //!< If got_whois == true, there is IDENT of user
    string host;                                //!< If got_whois == true, there is host of user
    string fullname;                            //!< If got_whois == true, there is full name of user

    bool got_mode;                              //!< Have we received mode of user? (+v and +o / '+' and '@' prefixes; and '~' on some irc networks)
    string mode;                                //!< If got_mode == true, there is prefix of mode (i.e. "+", "@", "@+")
    string old_mode;                            //!< Mode of previous WHOIS

    bool irc_op;                                //!< If got_whois == true: Is this user an irc operator?

    string in_logic_as;                         //!< Name of user as in logic.h::c_user structure

    // only for var tmp_users
    my_time_t last_whois;                          //!< Used only in tmp_users variable / for irc_access_to_partyline() function
    string dcc_msg;                             //!< Used only in tmp_users variable / for irc_access_to_partyline() function

    // for on_join after JOIN and WHOIS
    bool just_joined;                           //!< true if user just joined the channel (for on_join after JOIN and WHOIS)
    string just_joined_channel;                 //!< Name of channel that user just joined (for on_join after JOIN and WHOIS)

    bool nick_validated;                        //!< true if we already called logic_on_nick_validate(), should be cleared after nick change
    bool mod_disp_join_called;                  //!< true if we already called mod_dispatch() for "join@channel", should be cleared after nick change

    bool on_join_passive;
    bool on_join_called2;

    vector<c_flood_history> msg_flood_hist;     //!< History for msg_flood (PRIVMSG) detection
    vector<c_flood_history> repeat_flood_hist;  //!< History for repeat_flood detection
    vector<c_flood_history> notice_flood_hist;  //!< History for notice_flood detection
    vector<c_flood_history> nick_flood_hist;    //!< History for nick_flood detection
    vector<c_flood_history> join_flood_hist;    //!< History for join_flood detection
    vector<c_flood_history> mode_flood_hist;    //!< History for mode_flood detection
    vector<c_flood_history> ctcp_flood_hist;    //!< History for ctcp_flood detection
    my_time_t quit_time;                           //!< Time of user's quit/part/kick-ed (for gone_users variable only)

    string old_nick;                            //!< Workaround to multiple call logic_on_nick_validate

    string resolv_host;                         //!< Only for tmp_users: user's host to be resolved
    int ip_version;                             //!< Only for tmp_users: 4 for IPv4, 6 for IPv6

    c_online_user();
};

#endif

