/***************************************************************************
                          irc_bot.cpp  -  description
                             -------------------
    begin                : Thu Dec 9 2004
    copyright            : (C) 2004-2007 by Marian VooDooMan Meravy
    email                : ghostvoodooman -aT= users (dot) sourceforge.net
 ***************************************************************************/

/* $Id: irc_bot.cpp,v 1.96 2008/11/24 12:09:23 ghostvoodooman Exp $ */

/***************************************************************************

VooDoo cIRCle - an IRC (ro)bot
Copyright (C) 2004-2008 by Marian VooDooMan Meravy
  (ghostvoodooman -aT= users (dot) sourceforge.net)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************/

/*!
    \file
    \brief Main file containing entry point
    \version $Id: irc_bot.cpp,v 1.96 2008/11/24 12:09:23 ghostvoodooman Exp $
*/

#if defined(HAVE_CONFIG_H)
#   include <config.h>
#endif

#include "params.h"
#include "utils/memleak.h"
#include "utils/assert.h"
#include "compat.h"

#include "sock_shared.h"

#include <string>
#include <vector>
#include <stack>
#include <list>
#include <map>
#include <iostream>

#include <stdlib.h>
#include <stdio.h>

#if (defined(_WIN32) || defined(_WIN64)) && !defined(__CYGWIN__)
#   include <conio.h>
#   include <process.h>
#   include <io.h>
#   define _WINSOCKAPI_
#   include <windows.h>
#else
#   include <unistd.h>
#endif

#if defined(__CYGWIN__)
#   define _WINSOCKAPI_
#   include <windows.h>
#endif

#ifdef HAVE_SYS_TIME_H
#   include <sys/time.h>
#endif

#include <sys/stat.h>

#if (!defined(_WIN32) && !defined(_WIN64)) || defined(__CYGWIN__)
#   ifdef HAVE_SIGNAL
#       include <signal.h>
#   endif
#endif

#include <time.h>
#include <map>
#include <list>
#include <string.h>
#include <time.h>

#include "utils.h"
#include "log.h"
#include "sock.h"
#include "irc.h"
#include "logic.h"
#include "dcc.h"
#include "botnet.h"
#include "conf.h"
#include "stats.h"
#include "identd.h"
#include "ssl.h"
#include "exec.h"
#include "filesys.h"
#include "lang.h"
#include "match.h"
#include "variable.h"
#include "function.h"
#include "modules.h"
#include "compress.h"
#include "proc.h"
#include "threads.h"
#include "eval.h"
#include "debug1.h"
#include "debug2.h"
#include "mdump/mdump.h"
#include "log2.h"
#include "sock_garbage.h"
#include "var/c_var_array.h"
#include "var/c_var_string.h"
#include "memory/heap_debug2.h"
#include "thread/mutex.h"
#include "misc/shutdown.h"
#include "utils/my_getpid.h"
#include "memory/low_memory_condition.h"
#include "pcre/pattern_cache.h"
#include "pcre/c_pcre.h"
#include "internal_state.h"
#include "misc/c_script_timeout_protector_locker.h"

#include "irc_bot.h"

#include "utils/precomp.h"

#if defined(_MSC_VER) || defined(__BORLANDC__)
#   pragma hdrstop
#endif

using namespace std;

bool copy_me_to_upper_dir=false;                        //!< Contains true if this binary of us should be copied to upper dir (after .upgrade command)
string binary_name;                                     //!< Name of binary file (expected how to be the file called)
string this_instance_binary;                            //!< Contains name of this instance's binary file
bool spawned_by_service=false;                          //!< Contains true if the bot is spawned by service.

static c_exec_handle* s_restart_exec_handle=NULL;       //!< Exec handle for restarting bot

/*!
    \brief Describes return code of bot (used as an thrown exception to distinguish between this struct and other uncatched exceptions)
    \author VooDooMan
    \version 1
    \date 2008-03-09
*/
struct c_return_code
{
    int m_rc;                       //!< Return code
    bool m_suppress_debug_info;     //!< If to suppress debugging informations

    /*!
        \brief Constructor
        \author VooDooMan
        \version 1
        \date 2008-03-09
        \param rc Return code
        \param suppress_debug_info If to suppress debugging informations
    */
    c_return_code(int rc, bool suppress_debug_info = false)
        : m_rc(rc)
        , m_suppress_debug_info(suppress_debug_info)
    { ; }
};

#if (!defined(_WIN32) && !defined(_WIN64)) || defined(__CYGWIN__)

#ifdef HAVE_SIGNAL

#ifndef RETSIGTYPE
#   define RETSIGTYPE void
#endif

MY_STATIC RETSIGTYPE sig_handler(int sig);

/*!
    \brief Initializes signal handling
    \version 2
    \date ..., 2007-07-30
    \author VooDooMan
*/
void init_signals()
{
    { // totally ingnore
        struct sigaction act;
        memset(&act,0,sizeof(act));
        sigemptyset(&act.sa_mask);
        act.sa_flags=SA_RESTART; // for sure's sake
        act.sa_handler=SIG_IGN;

#ifdef HAVE_SIGPOLL
        sigaction(SIGPOLL,&act,NULL); // for System V :-)
#endif
        sigaction(SIGCHLD,&act,NULL); // just in case; however, default is ignore anyways
        sigaction(SIGTSTP,&act,NULL);
        sigaction(SIGWINCH,&act,NULL); // just in case; however, default is ignore anyways
    }

    { // ignores with SA_RESTART
        struct sigaction act;
        memset(&act,0,sizeof(act));
        sigemptyset(&act.sa_mask);
        act.sa_flags=SA_RESTART; // for sure's sake
        act.sa_handler=SIG_IGN;

        sigaction(SIGFPE,&act,NULL);
        sigaction(SIGPIPE,&act,NULL);
        sigaction(SIGURG,&act,NULL); // very importannt
#ifdef HAVE_SIGLOST
        sigaction(SIGLOST,&act,NULL); // e.g. NFS error
#endif
    }

    struct sigaction act_handled;

    { // handled ones with SA_RESTART
        struct sigaction& act=act_handled;
        memset(&act,0,sizeof(act));
        sigemptyset(&act.sa_mask);
        act.sa_flags=SA_RESTART;
        act.sa_handler=sig_handler;

        sigaction(SIGTERM,&act,NULL);
        sigaction(SIGINT,&act,NULL);

        sigaction(SIGXCPU,&act,NULL);
        sigaction(SIGXFSZ,&act,NULL);
    }

    { // lost terminal et al
        struct sigaction old;
        sigaction(SIGHUP,NULL,&old);
        if(SIG_IGN==old.sa_handler) {
            // respect parent's desired action of this signal
            struct sigaction ign;
            memset(&ign,0,sizeof(ign));
            sigemptyset(&ign.sa_mask);
            ign.sa_flags=SA_RESTART;
            ign.sa_handler=SIG_IGN;

            sigaction(SIGHUP,&ign,NULL);
        } else {
            sigaction(SIGHUP,&act_handled,NULL);
        }
    }

    { // handled, special case: mask this signal (to not allow recursion)
        struct sigaction new_act;
        memset(&new_act,0,sizeof(new_act));
        sigemptyset(&new_act.sa_mask);
        new_act.sa_flags=SA_RESTART;
        new_act.sa_handler=sig_handler;

        sigaddset(&new_act.sa_mask,SIGUSR1);

        sigaction(SIGUSR1,&new_act,NULL);
    }
}

/*!
    \brief UNIX handler for signals
    \author VooDooMan
    \version 2
    \date 2004, 2007-07-30
    \param sig Signal type
*/
MY_STATIC RETSIGTYPE sig_handler(int sig)
{
    // glibc's function 'strsignal()' is a GNU extension, even with
    //     full respect for glibc, I don't want to use it
    //     for portabiliby reasons...
    //     ...so there are strings:
    static const char* s_SIGTERM=   "SIGTERM";
    static const char* s_SIGINT=    "SIGINT";
    static const char* s_SIGHUP=    "SIGHUP";

    const char* str=NULL;

    switch(sig) {
        case SIGTERM:
            str=s_SIGTERM;
            // fall thru
        case SIGINT:
            if(NULL==str)
                str=s_SIGINT;
            // fall thru
        case SIGHUP:
            if(NULL==str)
                str=s_SIGHUP;

            { // ignore further signals
                struct sigaction act;
                memset(&act,0,sizeof(act));
                sigemptyset(&act.sa_mask);
                act.sa_flags=SA_RESTART; // for sure's sake
                act.sa_handler=SIG_IGN;

                sigaction(SIGTERM,&act,NULL);
                sigaction(SIGHUP,&act,NULL);

                // except for SIGINT (user may force by doubled C-c)
                act.sa_handler=SIG_DFL;
                sigaction(SIGINT,&act,NULL);
            }

            if(bot_shutting_down())
                return;
            bot_shutdown_action(E_SD_DIE,str);
            return;
        case SIGUSR1:
            // restore handler in case we are running System V
            init_signals();

            lang_rehash();
            logic_partyline_rehash("en","","SIGUSR1");
            return;
        case SIGXCPU:
            // restore handler in case we are running System V
            init_signals();

            log_bot("Got signal SIGXCPU");
            log_debug("Got signal SIGXCPU");
            return;
        case SIGXFSZ:
            // restore handler in case we are running System V
            init_signals();

            log_bot("Got signal SIGXFSZ");
            log_debug("Got signal SIGXFSZ");
            return;
    }

    // NOTREACH ; else: logical error
    MY_ASSERT_NOTREACH();

    // restore handler in case we are running System V
    init_signals();
}

#endif

#if !defined(_WIN32) && !defined(_WIN64)

/*!
    \brief Dummy function for non-windoze platforms
    \author VooDooMan
    \version 1
    \date 2006-05-08
*/
void ProcessMessages()
{
}

#endif

#else

static c_mutex s_win_signal_m;

/*!
    \brief Handler for WIN32 system signals
    \author VooDooMan
    \version 1
    \date 2004
    \param dwCtrlType Signal type
    \return Returns true if signal has been handled
*/
BOOL WINAPI HandlerRoutine(DWORD dwCtrlType)
{
    c_mutex_locker lock(s_win_signal_m); // windows signals can be recursive
                                   // ... also we have static variable there

    static bool user_break_already=false;
    if(user_break_already) {
        if(CTRL_C_EVENT==dwCtrlType || CTRL_BREAK_EVENT==dwCtrlType ||
            CTRL_CLOSE_EVENT==dwCtrlType) {
                // signal already signalled: exit immediately!
                ExitProcess(RC_BOT_DIE);
        }
    }

    if(CTRL_C_EVENT==dwCtrlType || CTRL_BREAK_EVENT==dwCtrlType ||
        CTRL_CLOSE_EVENT==dwCtrlType) {
            user_break_already=true;
    }

    string killing;

    switch(dwCtrlType) {
        case CTRL_C_EVENT:
            if(killing.empty())
                killing="CTRL_C_EVENT";
            // fall thru
        case CTRL_BREAK_EVENT:
            if(killing.empty())
                killing="CTRL_BREAK_EVENT";
            // fall thru
        case CTRL_SHUTDOWN_EVENT:
            if(killing.empty())
                killing="CTRL_SHUTDOWN_EVENT";
            // fall thru
        case CTRL_CLOSE_EVENT:
            if(killing.empty())
                killing="CTRL_CLOSE_EVENT";
            bot_shutdown_action(E_SD_DIE,killing);
            return true;
        case CTRL_LOGOFF_EVENT:
            // ignore: no guarantee which user is logging-off
            return true;
    }
    return true;
}

#endif

#if defined(_WIN32) || defined(_WIN64)

/*!
    \brief Precess the Windows(TM) message
    \author VooDooMan
    \version 2
    \date 2004, 2007-07-30
    \return Returns true if message was dispatched; false if none messages was in the queue
*/
bool ProcessMessage()
{
    tagMSG m;
    if(PeekMessage(&m,0,0,0,PM_REMOVE)) {
        TranslateMessage(&m);
        DispatchMessage(&m);
        return true;
    }
    return false;
}

/*!
    \brief Precess the Windows(TM) all pending messages
    \author VooDooMan
    \version 2
    \date 2004, 2007-07-30
*/
void ProcessMessages()
{
    for(;;) {
        while(ProcessMessage())
            ;
        if(!ProcessMessage())
            break;
    }
}

#endif

/*!
    \brief Gets wall clock in miliseconds
    \author VooDooMan
    \warning Can owerflow from time to time
    \version 1
    \date 2007-02-05
*/
long get_wall_clock()
{
#if !defined(HAVE_GETTIMEOFDAY)
    // windoze hase improper implementation of 'clock()', so okay to go there
    clock_t c1=clock();
    unsigned long int c2=(unsigned long int)c1;
    double c3=(double)c2/CLOCKS_PER_SEC*1000;
    return (unsigned long int)c3;
#else
    timeval tv;
    int rc=gettimeofday(&tv,NULL);
    if(!rc)
        return 0;
    return tv.tv_usec/1000;
#endif
}

/*!
    \brief Sleeps
    \author VooDooMan
    \version 3
    \date 2004, 2006-05-08, 2007-02-05
    \param ms Mili-seconds to sleep; or negative number for processing garbage of variables only, without sleeping
*/
void my_sleep(int ms)
{
    ProcessMessages();

    ssl_loop();
    sock_send_cache();

    ProcessMessages();

    ssl_loop();
    sock_send_cache();

    {
        static bool in=false;
        if(!in) {
            in=true;
            mod_loop();
            in=false;
        }
    }

    if(ms) {
        long start=get_wall_clock();

        bool got=false;
        static my_time_t last_garbage1=my_time(NULL);
        if(!got && last_garbage1+30<my_time(NULL)) {
            var_keeper->do_garbage(ms);
            last_garbage1=my_time(NULL);
            got=true;
        }
        static my_time_t last_garbage2=my_time(NULL);
        if(!got && last_garbage2+61<my_time(NULL)) {
            var_keeper->do_garbage(c_var_keeper::GARBAGE_MAX_TIME_INFINITE);
            last_garbage2=my_time(NULL);
            got=true;
        }

        if(ms>0) {
            ms-=get_wall_clock()-start;
            if(ms<=0)
                return;
            if(ms<1000) {
                sock_interruptible_sleep(ms);
            } else {
                my_time_t now=my_time(NULL);
                for(;;) {
                    sock_interruptible_sleep(100);
                    ProcessMessages();
                    e_shutdown_reason sr=bot_get_shutdown_state();
                    if(E_SD_DIE!=sr && E_SD_RESTART!=sr)
                        dcc_loop();
                    ssl_loop();
                    botnet_loop();
                    sock_send_cache();
                    threads_loop();
                    if(now+ms/1000<my_time(NULL))
                        break;
                }
            }
        }
    }
    ProcessMessages();
}

/*!
    \brief Does shutdown epilog
    \author VooDooMan
    \version 1
    \date 2005-11-07
*/
void backup()
{
    c_module* core_mod=get_module("core",true);
    if(core_mod)
        logic_partyline_backup(core_mod,bot_get_shutdown_reason());
}

/*!
    \brief Returns version information
    \author VooDooMan
    \version 1
    \date 2005-11-11
*/
string get_bot_version()
{
    string ver="VooDoo cIRCle " BOT_VERSION " (";
    ver+=ssl_get_ssl_version();
    ver+=", zlib ";
    ver+=compress_get_zlib_version();
    ver+=", PCRE ";
    ver+=c_pcre::get_pcre_version();
    ver+=")";
    return ver;
}

/*!
    \brief Displays license information and version
    \author VooDooMan
    \version 3
    \date 2004, 2005-10-20, 2006-09-29
*/
void disp_legals()
{
    char lic[]=
        "Copyright (C) 2004-2008 by Marian VooDooMan Meravy\n"
        "(ghostvoodooman -aT= users (dot) sourceforge.net).\n"
        "Licensed under GNU/GPL license version 2. For details see `license.html'.\n"
        "\n"
        "VooDoo cIRCle comes with ABSOLUTELY NO WARRANTY; for details see\n"
        "`license.html'. This is free software, and you are welcome to redistribute\n"
        "it under certain conditions; see the license for details.\n"
        "\n"
#ifdef USE_OPENSSL
        "This product includes software developed by the OpenSSL Project for use in the\n"
        "OpenSSL Toolkit. (http://www.openssl.org/)\n"
        "\"This product includes cryptographic software written by Eric Young\n"
        "(eay@cryptsoft.com).\"\n"
        "\"This product includes software written by Tim Hudson (tjh@cryptsoft.com).\"\n"
        "\n"
#endif
        "This product uses zlib library for compressing writen by\n"
        "Jean-loup Gailly and Mark Adler\n"
        "\n"
        ;
    cout << get_bot_version() << endl;
    cout << lic;
}

/*!
    \brief Writes to log file that bot was asked via DCC / telnet connection to die or restart
    \author VooDooMan
    \version 2
    \date 2004, 2007-08-20
    \return Returns log from kill, or an empty string if it is called subsequently second, or more times
*/
string write_kill_log()
{
    static bool s_once=false;
    if(s_once)
        return (string)"";
    s_once=true;

    string s;
    string msg;

    map<string,string> subst;
    pair<string,string> p;
    p.first="%uname;";
    p.second=bot_get_shutdown_reason();
    subst.insert(p);
    e_shutdown_reason sr=bot_get_shutdown_state();
    if(E_SD_DIE==sr) {
        s="DIE by ";
        msg=lang_subst(core_conf_getvar("irc_quit_msg_on_.die"),subst);
    } else if(E_SD_RESTART==sr) {
        s="RESTART by ";
        msg=lang_subst(core_conf_getvar("irc_quit_msg_on_.restart"),subst);
    } else if(E_SD_UPGRADE==sr) {
        s="UPGRADE by ";
        msg=lang_subst(core_conf_getvar("irc_quit_msg_on_.die"),subst); // use "die" message
    }

    s+=bot_get_shutdown_reason();
    irc_quit(msg.c_str(),false);
    string log=(string)"*** "+s;
    log_bot(log);
    return s;
}

/*!
    \brief From time to time check "logs/irc_bot.action" file
    \version 1
    \date 2007-08-19
    \author VooDooMan
*/
void check_action_file()
{
    static my_time_t last_action_check=0;
    if(last_action_check+5<my_time(NULL)) {
        last_action_check=my_time(NULL);

        FILE* f=fopen("logs" FILE_SLASH "irc_bot.action","r");
        if(f) {
            char tmp[512];
            tmp[0]=0;
            fgets(tmp,512-16,f);
            while(strlen(tmp) && (tmp[strlen(tmp)-1]=='\r' || tmp[strlen(tmp)-1]=='\n'))
                tmp[strlen(tmp)-1]='\0';

            fclose(f);
            f=NULL;
            bool understand=false;
            understand|=!strcmp(tmp,"die");
            understand|=!strcmp(tmp,"restart");
            understand|=!strcmp(tmp,"flushdns");
            understand|=!strcmp(tmp,"rehash");
            int rc;
            if(understand) {
                rc=unlink("logs" FILE_SLASH "irc_bot.action");
                errno=0;
                if(0==rc) {
                    if(!strcmp(tmp,"die")) {
                        string killing="logs/irc_bot.action";
                        bot_shutdown_action(E_SD_DIE,killing);
                        c_module* core_mod=get_module("core",true);
                        logic_partyline_backup(core_mod,killing);
                    }
                    if(!strcmp(tmp,"restart")) {
                        string killing="logs/irc_bot.action";
                        bot_shutdown_action(E_SD_RESTART,killing);
                        c_module* core_mod=get_module("core",true);
                        logic_partyline_backup(core_mod,killing);
                    }
                    if(!strcmp(tmp,"flushdns")) {
                        threads_flushdns();
                    }
                    if(!strcmp(tmp,"rehash")) {
                        lang_rehash();
                        logic_partyline_rehash("en","","logs/irc_bot.action");
                    }
                }
            }
        }
    }
}

/*!
    \author Andrew Gierth
    \version 1
    \date 1997
    \brief Close all FDs >= a specified value
*/
void closeall(int UNUSEDARG_WIN(fd))
{
#if (!defined(_WIN32) && !defined(_WIN64)) || defined(__CYGWIN__)
    int fdlimit=sysconf(_SC_OPEN_MAX);

    while(fd<fdlimit)
        close(fd++);
#endif
}

/*!
    \author Andrew Gierth, VooDooMan
    \version 4
    \date 1997, 2004, 2005, 2006
    \return Returns zero for no error
    \brief Detach process from user and disappear into the background
        returns -1 on failure, but you can't do much except exit in that case
        since we may already have forked. This is based on the BSD version,
        so the caller is responsible for things like the umask, etc.
    \note Believed to work on all Posix systems
    \retval 0 Okay
    \retval -1 Error
*/
int daemon(bool UNUSEDARG_WIN(nochdir), bool UNUSEDARG_WIN(noclose), bool UNUSEDARG_WIN(stdout_new_pid))
{
#if (!defined(_WIN32) && !defined(_WIN64)) || defined(__CYGWIN__)
    switch(fork()) {
        case 0:  break;
        case -1: _exit(RC_BOT_DIE); return -1;
        default:
            {
                my_time_t now=my_time(NULL);
                while(stdout_new_pid && now+2>my_time(NULL)) {
                    timeval tv;
                    tv.tv_sec=1;
                    tv.tv_usec=0;
                    select(1,NULL,NULL,NULL,&tv);
                }
                _exit(RC_BOT_DIE);          /* exit the original process */
            }
    }

    if(conf_umask!=-1)
        umask(conf_umask);

    if(setsid()<0) {               /* shoudn't fail */
        _exit(RC_BOT_DIE);
        return -1;
    }

    /* dyke out this switch if you want to acquire a control tty in */
    /* the future - not normally advisable for daemons */

    switch(fork()) {
        case 0:  break;
        case -1: _exit(RC_BOT_DIE); return -1;
        default:
            _exit(RC_BOT_DIE);
    }

    if(conf_umask!=-1)
        umask(conf_umask);

    if(setsid()<0) {              /* shoudn't fail */
        _exit(RC_BOT_DIE);
        return -1;
    }

    if(!nochdir)
        chdir("/");

    if(stdout_new_pid)
        cout << "Background process PID is " << my_getpid() << endl << endl;
    cout.flush();

    if(!noclose) {
        closeall(0);
        open("/dev/null",O_RDWR);
        dup(0); dup(0);
    }

#endif
    return 0;
}

/*!
    \brief Does a 'working loop' of bot, invoked by internal script
    \author VooDooMan
    \date 2004, ..., 2007-07-28
*/
void do_work()
{
    int error=0;
    for(;;) {
        e_shutdown_reason sr=bot_get_shutdown_state();
        if(E_SD_DIE==sr || E_SD_RESTART==sr)
            break;

        var_keeper->set_current_session(SESSION_WORK_ID);

        {
            c_heap_map_recreation_impossible impossible;

            eval_cleanup_timeout();

            logic_shift_backup(false);

            if(irc_loop_process_input()) {
                // error handled below
            }

            logic_loop();

            if(!error) {
                if(irc_loop_putserv())
                    error++;
            } else
                error++;

            // we want to process remaining lines from the IRC server (if any)
            //     so do NOT break here! but only after some iterations...
            if(error>=10) {
                break;
            }

            logic_loop_identd();

            botnet_loop();

            mod_loop();

            ssl_loop();
            sock_send_cache();

            ssl_loop();
            sock_send_cache();

            threads_loop();

            mod_loop2();

            sock_garbage_loop();

            check_action_file();
        }

        {
            if(msgs_from_server.empty())
                my_sleep(100);
            else
                my_sleep(1);
        }
        static my_time_t last_copy_attemp=0;
        static bool copy_attemp_successful=false;
        if(!copy_attemp_successful && copy_me_to_upper_dir && last_copy_attemp+10<my_time(NULL)) {
            last_copy_attemp=my_time(NULL);
            FILE* f1=fopen(this_instance_binary.c_str(),"rb");
            if(f1) {
                FILE* f2=fopen(binary_name.c_str(),"wb");
                if(f2) {
                    char tmp[1024];
                    while(!feof(f1)) {
                        size_t read=fread(tmp,1,sizeof(tmp),f1);
                        fwrite(tmp,1,read,f2);
                    }
                    fclose(f2);
#if (!defined(_WIN32) && !defined(_WIN64)) || defined(__CYGWIN__)
                    if(-1!=conf_umask)
                        chmod(binary_name.c_str(),0777 & ~conf_umask);
                    else
                        chmod(binary_name.c_str(),0700);
#endif
                    copy_attemp_successful=true;
                }
                fclose(f1);
            }
        }
#if defined(_WIN32) || defined(_WIN64)
        ProcessMessages();
#endif
    }

    e_shutdown_reason sr=bot_get_shutdown_state();
    if(E_SD_DIE==sr || E_SD_RESTART==sr) {
        string log=write_kill_log();
        if(!log.empty()) { //to avoid doubled call elsewhere
            dcc_broadcast(log);

            if(E_SD_DIE==sr)
                logic_on_internal_event("@die@",bot_get_shutdown_reason(),"","","",PRIVATE_EVENT_SEVERITY_CODE_INFORMATIONAL,"","");
            if(E_SD_RESTART==sr)
                logic_on_internal_event("@restart@",bot_get_shutdown_reason(),"","","",PRIVATE_EVENT_SEVERITY_CODE_INFORMATIONAL,"","");
        }

        dcc_flush_all();
    }

    if(E_SD_RESTART!=sr && E_SD_DIE!=sr) {
        // exec from work
    }

    if(error)
        irc_disconnect();
}

/*!
    \brief Deos an epilog on shutdown
    \author VooDooMan
    \version 2
    \date 2006-07-13, 2007-08-19
*/
MY_STATIC void do_epilog(bool self_test = false)
{
    dcc_flush_all();
    if(!self_test)
        ssl_loop();
    filesys_flush();
    dcc_flush_all();
    if(!self_test)
        ssl_loop();
    backup();
    dcc_flush_all();
    if(!self_test)
        ssl_loop();
    dcc_flush_all();
    if(!self_test)
        ssl_loop();
    irc_dummy_operation();
    dcc_flush_all();
    if(!self_test)
        ssl_loop();
    irc_disconnect();
    dcc_flush_all();
    if(!self_test)
        ssl_loop();
    botnet_loop(true);
    dcc_flush_all();
    if(!self_test)
        ssl_loop();
    mod_loop(true);
    dcc_flush_all();
    if(!self_test)
        ssl_loop();
    irc_cleanup();
    dcc_flush_all();
    if(!self_test)
        ssl_loop();
    threads_done();
    dcc_flush_all();
    if(!self_test)
        ssl_loop();
    dcc_close_all();
    botnet_loop(true);
#if defined(_WIN32) || defined(_WIN64)
    my_sleep(1000);
#endif
    ssl_done();
    sock_shutdown();
    log_free_locks();
}

/*!
    \brief Does an idle tasks
    \author VooDooMan
    \version 1
    \date 2006-07-13
*/
void do_idle()
{
    logic_shift_backup(false);
    sock_send_cache();
    logic_loop_identd();
    botnet_loop();
    ssl_loop();
    threads_loop();
    mod_loop2();
    dcc_loop();
    ssl_loop();
    sock_garbage_loop();
    check_action_file();
    my_sleep(-1); // for garbage collection when network is down and bot is in a loop
    g_pcre.loop();
}

MY_STATIC void seed_prng()
{
    BEGIN_DEBUG_TIME_MEASUREMENT(true,1,"Seeding PRNG");

    unsigned int seed=0;
    FILE* f;
    size_t size=0;

    string log;

    log=((string)"Seedeng PRNG with ")+tos(sizeof(seed))+" bytes from /dev/random";
    log_debug(log);
    f=fopen("/dev/random","rb");
    if(f) {
        while(size<sizeof(seed))
            size+=fread(&reinterpret_cast<char*>(&seed)[size],1,sizeof(seed)-size,f);
        srand(seed);
        log_debug("Done seeding");
        fclose(f);
        return;
    } else {
        log_debug("Not exists");
    }

    log=((string)"Seedeng PRNG with ")+tos(sizeof(seed))+" bytes from /dev/urandom";
    log_debug(log);
    f=fopen("/dev/urandom","rb");
    if(f) {
        while(size<sizeof(seed))
            size+=fread(&reinterpret_cast<char*>(&seed)[size],1,sizeof(seed)-size,f);
        srand(seed);
        log_debug("Done seeding");
        fclose(f);
        return;
    } else {
        log_debug("Not exists");
    }

    log="Seedeng PRNG with time() function";
    log_debug(log);
    srand(static_cast<unsigned>(my_time(NULL)));
    log_debug("Done seeding");

    END_DEBUG_TIME_MEASUREMENT(1,0);
}

/*!
    \brief Entry point
    \author VooDooMan
    \version 4
    \date 2004, 2007-07-28, 2007-08-03, 2007-09-06
    \param argc Argument count
    \param argv Array of arguments
    \return Returns result action
    \retval RC_BOT_DIE If bot should die (.die command)
    \retval RC_BOT_RESTART If bot wants to be restarted
    \retval RC_BOT_ERROR If there's an error
*/
int main(int argc, char* argv[])
{
    srand(static_cast<unsigned>(time(NULL))); // for sure's sake

#if defined(_MSC_VER) && defined(_DEBUG)
    _CrtMemState sa1;
    _CrtMemCheckpoint(&sa1);
#endif

    low_memory_condition_install_new_handler_callback();
    heap_debug_init();

    bool eval_test=false;
    bool prng_seeded=false;

    MY_NEW c_var_keeper();

    c_return_code ret_code(RC_BOT_DIE);

    try {
        botnet_loop(false);

        conf_init_pre_start();
        if(conf_umask!=-1)
            umask(conf_umask);

        {
            bool wait=false;
            for(int i1=1; i1<argc; ++i1) {
                if(!strcmp("--wait-for-released-lock",argv[i1])) {
                    wait=true;
                    break;
                }
            }
            if(wait)
                log_wait_for_releasing_lock();
        }

        // here we go!
        this_instance_binary=argv[0];
        binary_name=argv[0];
        bool match_test_1=false;

        bool core_dump=false;

        bool get_ssl_cert=false;
        string get_ssl_cert_bind;
        string get_ssl_cert_server;
        long get_ssl_cert_port=0;

#if (!defined(_WIN32) && !defined(_WIN64)) || defined(__CYGWIN__)
        bool b_interactive=false;
#endif
        for(int i1=1; i1<argc; ++i1) {
            if(!strcmp("--hush",argv[i1])) {
#if (!defined(_WIN32) && !defined(_WIN64)) || defined(__CYGWIN__)
                closeall(0);
                open("/dev/null",O_RDWR);
                dup(0); dup(0);
#endif
                continue;
            }
            if(!strcmp("-h",argv[i1]) ||
               !strcmp("--help",argv[i1])) {
            usage:
                disp_legals();
                cout <<
                    "Options:" << endl <<
                    "  -h, --help         : this message" << endl <<
                    "  -v, -V, --version  : prints out version only" << endl <<
                    "  -i, --interactive  : run in foreground, interactive waiting for Ctrl-c" << endl <<
                    "                       (default on Windoze)" << endl <<
                    "  -d, --daemon       : run in background (i.e. daemonize; default on POSIX)" << endl <<
                    "  --get-ssl-cert 0.0.0.0 ssl.example.org 7780" << endl <<
                    "  --get-ssl-cert :: ssl.example.org 7780" << endl <<
                    "                     : Echoes to console server certificate of SSL" << endl <<
                    "                       server ssl.example.org running on port 7780." << endl <<
                    "                       Binds to IPv4 or IPv6 address as local address" <<endl <<
                    "                       (first argument)." << endl <<
                    "                       RC is 0 for suceess, 1 for error." << endl <<
                    "                       Certificate if case of success is in PEM format," << endl <<
                    "                       so you can put it to \"./cert/ssl.example.org.crt\"" << endl <<
                    "  --eval-test-1      : Perform self-test of internal scripting." << endl <<
                    "                       (exits immediately when finished; RC is 0 when passed)" << endl <<
                    "  --match-test-1     : Perform self-test of hostmask matching." << endl <<
                    "                       (exits immediately when finished)" << endl <<
                    "  --core-dump        : Test to generate core dump" << endl <<
                    "  --legal            : prints out legal informations only" << endl;
                throw c_return_code(RC_BOT_DIE,true);
            }
            if(!strcmp("-i",argv[i1]) ||
            !strcmp("--interactive",argv[i1])) {
#if (!defined(_WIN32) && !defined(_WIN64)) || defined(__CYGWIN__)
                b_interactive=true;
#endif
                continue;
            }
            if(!strcmp("-v",argv[i1]) ||
                !strcmp("-V",argv[i1]) ||
                !strcmp("--version",argv[i1])) {
                    cout << get_bot_version() << endl;
                    throw c_return_code(RC_BOT_DIE,true);
            }
            if(!strcmp("--legal",argv[i1])) {
                disp_legals();
                throw c_return_code(RC_BOT_DIE,true);
            }
            if(!strcmp("-d",argv[i1]) ||
                !strcmp("--daemon",argv[i1])) {
#if (!defined(_WIN32) && !defined(_WIN64)) || defined(__CYGWIN__)
                    b_interactive=false;
#endif
                    continue;
            }
            if(!strncmp("--upgrade=",argv[i1],strlen("--upgrade="))) {
                copy_me_to_upper_dir=true;
                string tmp=argv[i1];
                if(tmp.find("=",0)!=string::npos)
                    tmp.erase(0,tmp.find("=",0)+1);
                binary_name=tmp;
                cout << binary_name << endl;
                continue;
            }
            if(!strcmp("--service",argv[i1])) {
                spawned_by_service=true;
#if (!defined(_WIN32) && !defined(_WIN64)) || defined(__CYGWIN__)
                spawned_by_service=false;
#endif
                continue;
            }
            if(!strcmp("--eval-test-1",argv[i1])) {
#if (!defined(_WIN32) && !defined(_WIN64)) || defined(__CYGWIN__)
                b_interactive=true;
#endif
                eval_test=true;
                continue;
            }
            if(!strcmp("--match-test-1",argv[i1])) {
#if (!defined(_WIN32) && !defined(_WIN64)) || defined(__CYGWIN__)
                b_interactive=true;
#endif
                match_test_1=true;
                continue;
            }
            if(!strcmp("--wait-for-released-lock",argv[i1])) {
                continue;
            }
            if(1==i1 && !strcmp("--get-ssl-cert",argv[i1])) {
                get_ssl_cert=true;
                log_supress_logging();
                util_debug_report_suppress();
                continue;
            }
            if(2==i1 && get_ssl_cert) {
                get_ssl_cert_bind=argv[i1];
                continue;
            }
            if(3==i1 && get_ssl_cert) {
                get_ssl_cert_server=argv[i1];
                continue;
            }
            if(4==i1 && get_ssl_cert) {
                get_ssl_cert_port=atol(argv[i1]);
                continue;
            }
            if(!strcmp("--core-dump",argv[i1])) {
                core_dump=true;
                continue;
            }
            goto usage;
        }

        if(get_ssl_cert) {
            // sanity checks
            if(get_ssl_cert_bind.empty()) {
                cout << "You need to specify local IP address to bind to (hint: use \"0.0.0.0\" for as wildcard bind for IPv4, or \"::\" as widlcard bind for IPv6)." << endl;
                throw c_return_code(RC_BOT_ERROR,true);
            }
            if(get_ssl_cert_server.empty()) {
                cout << "You need to specify server hostname." << endl;
                throw c_return_code(RC_BOT_ERROR,true);
            }
            if(get_ssl_cert_port<1 || get_ssl_cert_port>65534) {
                cout << "You need to specify server port in range from 1 to 65534 inclusive." << endl;
                throw c_return_code(RC_BOT_ERROR,true);
            }

            // init
            conf_init_pre_start();
            if(conf_umask!=-1)
                umask(conf_umask);
            sock_startup();

            // init SSL
            if(!prng_seeded) {
                prng_seeded=true;
                seed_prng();
                ssl_init();
                ssl_loop();
            }

            // here we go
            bool error;
            string rc=ssl_get_pem_cert(get_ssl_cert_bind,get_ssl_cert_server,static_cast<unsigned short>(get_ssl_cert_port),error);
            cout << rc << endl;
            throw c_return_code(error ? RC_BOT_ERROR : RC_BOT_DIE,true);
            // NOTREACH
        }

        printf("%s",log_get_ipc_mutex_error_log().c_str());

        disp_legals();

        if(match_test_1 || eval_test) {
            log_procetct_pid_file();
        }

        if(!match_test_1 && !eval_test) {

            if(!log_lock(true)) {
                cout << "Another instance is running! Exiting with code 0." << endl;
                throw c_return_code(RC_BOT_DIE,true);
            }

#if (!defined(_WIN32) && !defined(_WIN64)) || defined(__CYGWIN__)
            if(!b_interactive && !match_test_1) {
                if(daemon(true,false,true)!=0)
                    _exit(RC_BOT_ERROR);
            }
#endif
        }

        if(conf_umask!=-1)
            umask(conf_umask);

        string timestamp=util_debug_report_init();
#if defined(_MSC_VER)   // other compilers produce binary with incompatible symbols
        MiniDumper crash_dumper(timestamp.c_str());
#endif

        if(core_dump) {
            cout << "I am going to crash to test dumping of core." << endl;
            dump_core();
            return RC_BOT_DIE;
        }

        if(!match_test_1 && !eval_test) {
            if(!log_init()) {
                cout << "Another instance is running! Exiting with code 0." << endl;
                throw c_return_code(RC_BOT_DIE);
            }
        }

        log_can_overwrite_pid_file();

        {
            string ver=get_bot_version();
            log_bot(ver,true);
            log_debug(ver,true);
        }

        stats_init();

        sock_startup();

        modules_init();

        bool tests_already_performed=!eval_test;

    try_load_logic_again:

        c_module* core_mod1=NULL;
        if(!tests_already_performed)
            core_mod1=MY_NEW c_module();

        if(!eval_test) {
            if(conf_ssl_init()) {
                cout << "Cannot open file \"ssl.txt\"!" << endl;
                throw c_return_code(RC_BOT_ERROR,true);
            }
        }

        if(!prng_seeded) {
            prng_seeded=true;
            seed_prng();
            ssl_init();
            ssl_loop();
        }
        if(!tests_already_performed) {
            bool passed=debug_do_evaluate_tests();

            MY_DELETE(core_mod1);
            core_mod1=NULL;
            do_epilog(true);
            {
                while(modules_maintenance(true)) {
                    ;
                }
                modules_done();

                while(var_keeper->purge(SESSION_INVALID_ID)) {
                    ;
                }
                var_keeper->free_all();
            }
            // DEBUG: set breakpoint here
            int dummy=0;
            dummy++;

#ifdef _MSC_VER
#   if REPORT_LEAKS==1
            if(!_CrtCheckMemory()) {
                MY_RPT1(MY_RPT_INFO,"Memory corruption detected in function %s()",__my_func__);
                // DEBUG: set breakpoint here
                int dummy=0;
                dummy++;
            }
#   endif
#endif

            // DEBUG: set breakpoint here
            dummy++;

            throw c_return_code(passed ? RC_BOT_DIE : RC_BOT_SELF_TEST_ERROR);
        }
        //_CrtSetAllocHook(debug_alloc_hook);

        if(!tests_already_performed)
            MY_DELETE(core_mod1);

        tests_already_performed=true;

        if(!match_test_1 && !eval_test) {
            FILE* f=fopen("logs" FILE_SLASH "irc_bot.pid","w");
            if(f) {
                fprintf(f,"%s\n",tos(my_getpid()).c_str());
                fclose(f);
            }

            unlink("logs" FILE_SLASH "irc_bot.action");
        }

#if (defined(_WIN32) || defined(_WIN64)) && !defined(__CYGWIN__)
        SetConsoleCtrlHandler(HandlerRoutine,true);
#endif

#if (!defined(_WIN32) && !defined(_WIN64)) || defined(__CYGWIN__)
#   ifdef HAVE_SIGNAL
        init_signals();
#   endif
#endif

        {
delete_again:
            bool ok=false;

            c_module* core_mod2=get_module("core",false);
            if(!core_mod2) {
                core_mod2=MY_NEW c_module();
                modules.push_back(core_mod2);
                c_module::e_rc rc=core_mod2->load("core","");
                if(c_module::RC_OK==rc)
                    ok=true;
            } else {
                MY_DELETE(core_mod2);
                goto delete_again;
            }

            if(!prng_seeded) {
                prng_seeded=true;
                seed_prng();
                ssl_init();
                ssl_loop();
            }

            if(!match_test_1) {
                if(ok) {
                    ok=false;

                    super_global=var_keeper->get_super_global();

                    c_script_timeout_protector_locker stp_lock;

                    var_keeper->set_current_session(core_mod2->get_session());
                    string proc="_sys_startup($pid)";
                    c_var_array* vars=var_keeper->register_mod_global(NULL,__FILE__,__my_func__,__LINE__);
                    string pid=tos(my_getpid());
                    c_index_base* idx=MY_NEW c_index_str("$pid");
                    c_var_string* val=MY_NEW c_var_string(pid);
                    string debug="$pid";
                    vars->set(core_mod2,idx,val,false,debug);
                    debug="Entry point";
                    if(core_mod2)
                        ok=core_mod2->execute(proc,vars,debug,true);
                    if(vars) {
                        if(!ok)
                            MY_DELETE(vars);
                        vars=NULL;
                    }
                }
                if(!ok) {
                    bot_cancel_shutdown();
                    dcc_killing_user_eol="";
                    dcc_killing_user_lang="";
                    dcc_killing_user_socket=NULL;

                    vector<string> log;
                    bool rc=logic_rollback(log,"en");
                    for(vector<string>::iterator ii=log.begin(); ii!=log.end(); ++ii) {
                        log_bot(*ii);
                        log_debug(*ii);
                    }
                    if(!rc)
                        throw c_return_code(RC_BOT_DIE);

                    e_shutdown_reason sr=bot_get_shutdown_state();
                    if(E_SD_DIE!=sr && E_SD_RESTART!=sr) {
                        var_keeper->free_global(super_global,true);
                        super_global=NULL;
                        MY_DELETE(core_mod1);
                        MY_DELETE(core_mod2);
                        core_mod1=NULL;
                        core_mod2=NULL;
                        goto try_load_logic_again;
                    }
                }
                e_shutdown_reason sr=bot_get_shutdown_state();
                if(E_SD_DIE==sr || E_SD_RESTART==sr) {
                    string log=write_kill_log();
                    if(!log.empty()) { //to avoid doubled call elsewhere
                        dcc_broadcast(log);
                    }
                    dcc_flush_all();
                }
            } else {
                // match_test_1
                debug2_do_tests();
                throw c_return_code(RC_BOT_DIE);
            }
        }

        e_shutdown_reason sr=bot_get_shutdown_state();
        if(E_SD_RESTART==sr && NULL==s_restart_exec_handle) {
            dcc_close_servers();
            filesys_flush();
            dcc_flush_all();

            int ec=0;

            if(!spawned_by_service) {
                string tmp=binary_name;
                s_restart_exec_handle=exec_async_exec2(tmp,"--hush --wait-for-released-lock","");

                int exit_code;
                string disp_log, disp_dcc;
                if(s_restart_exec_handle->error) {
                    exit_code=RC_BOT_RESTART;

                    disp_log=disp_dcc="RESTART: Error spawning new instance. Something should run it.";
                    disp_dcc+=dcc_killing_user_eol;

                    log_bot(disp_log);

                    logic_on_internal_event("@restart_error@",bot_get_shutdown_reason(),"","","",PRIVATE_EVENT_SEVERITY_CODE_FATAL_ERROR,disp_log,"");

                    if(dcc_killing_user_socket)
                        dcc_send_msg(*dcc_killing_user_socket,disp_dcc,ec,dcc_killing_ssl_hnd);
                } else
                    exit_code=RC_BOT_DIE;

                string log=log_done(exit_code,dcc_killing_user_eol);
                ec=0;
                if(dcc_killing_user_socket)
                    dcc_send_msg(*dcc_killing_user_socket,log,ec,dcc_killing_ssl_hnd);
                sock_send_cache();
                if(dcc_killing_user_socket)
                    sock_flush_later(*dcc_killing_user_socket);

                if(!disp_log.empty() || !disp_dcc.empty()) {
                    log_bot(disp_log);
                    if(dcc_killing_user_socket)
                        dcc_send_msg(*dcc_killing_user_socket,disp_dcc,ec,dcc_killing_ssl_hnd);
                }

                if(s_restart_exec_handle) {
                    exec_detach_process(s_restart_exec_handle);
                    s_restart_exec_handle=NULL;
                }

                do_epilog();

                throw c_return_code(exit_code);
            }

            // it was spawned by service
            string log=log_done(RC_BOT_RESTART,dcc_killing_user_eol);
            ec=0;
            if(dcc_killing_user_socket)
                dcc_send_msg(*dcc_killing_user_socket,log,ec,dcc_killing_ssl_hnd);
            sock_send_cache();
            if(dcc_killing_user_socket)
                sock_flush_later(*dcc_killing_user_socket);

            do_epilog();

            throw c_return_code(RC_BOT_RESTART);
        }

        string log=log_done(RC_BOT_DIE,dcc_killing_user_eol);
        int ec=0;
        if(dcc_killing_user_socket)
            dcc_send_msg(*dcc_killing_user_socket,log,ec,dcc_killing_ssl_hnd);
        sock_send_cache();
        if(dcc_killing_user_socket)
            sock_flush_later(*dcc_killing_user_socket);

        do_epilog();

        {
            while(modules_maintenance(true)) {
                ;
            }
            modules_done();

            while(var_keeper->purge(SESSION_INVALID_ID)) {
                ;
            }
            var_keeper->free_all();
        }

#ifdef _MSC_VER
#if REPORT_LEAKS==1
        if(!_CrtCheckMemory()) {
            MY_RPT1(MY_RPT_INFO,"Memory corruption detected in function %s()",__my_func__);
            // DEBUG: set breakpoint here
            int dummy=0;
            dummy++;
        }
#endif
#endif

        throw c_return_code(RC_BOT_DIE);
    } catch(c_return_code& e) {
        botnet_loop(true);

        ret_code=e;
    }

    //

    g_pcre.loop(true);

    sock_shutdown();

    if(ret_code.m_suppress_debug_info)
        util_debug_report_suppress();

    MY_DELETE(var_keeper);
    var_keeper=NULL;

    if(!ret_code.m_suppress_debug_info) {
        string leaks;
        heap_debug_done(leaks);

        vector<string> lines;
        explode(leaks,'\n',lines,false);
        for(vector<string>::iterator i1=lines.begin(); i1!=lines.end(); ++i1) {
            // single line only is always "no mem leaks" message, so INFO, not ERROR
            MY_RPT0(lines.size()==1 ? MY_RPT_INFO : MY_RPT_ERROR,(*i1).c_str());
        }
    }

#if defined(_MSC_VER) && defined(_DEBUG)
    if(!_CrtCheckMemory()) {
        MY_RPT0(MY_RPT_INFO,"MSVC heap debugging functionality: detected heap corruption. Please, report this problem.");
        // DEBUG: set breakpoint here
        int dummy=0;
        dummy++;
    } else {
        MY_RPT0(MY_RPT_INFO,"MSVC heap debugging functionality: not detected any heap corruption");
    }
#endif

#if defined(_MSC_VER) && defined(_DEBUG)
    _CrtMemState sa2;
    _CrtMemCheckpoint(&sa2);
    _CrtMemState sa3;
    if(_CrtMemDifference(&sa3,&sa1,&sa2) && sa3.lCounts[_CLIENT_BLOCK]!=0) {
        MY_RPT0(MY_RPT_INFO,"MSVC heap debugging functionality: detected memory leaks, summary follows");
        _CrtMemDumpStatistics(&sa3);

        MY_RPT0(MY_RPT_INFO,"MSVC heap debugging functionality: detected memory leaks, dump of all objects follows");
        _CrtMemDumpAllObjectsSince(&sa1);
        MY_RPT0(MY_RPT_INFO,"MSVC heap debugging functionality: detected memory leaks, end of dump");

        // DEBUG: set breakpoint here
        int dummy=0;
        dummy++;
    } else {
        MY_RPT0(MY_RPT_INFO,"MSVC heap debugging functionality: no memory leaks detected (\"client\" blocks)");
    }
#endif

    if(eval_test) {
        if(!ret_code.m_suppress_debug_info)
            util_debug_report_done(ret_code.m_rc);

        ostringstream disp;
        int kind=MY_RPT_ERROR;
        switch(ret_code.m_rc) {
            case RC_BOT_DIE:
                kind=MY_RPT_INFO;
                disp << "*** EVAL_TEST_1: All self tests PASSED *** ";
                break;
            case RC_BOT_SELF_TEST_ERROR:
                kind=MY_RPT_ERROR;
                disp << "!!! EVAL_TEST_1 ERROR: Not all self tests completed sucessfully !!! ";
                break;
        }
        disp << "Exitcode " << ret_code.m_rc;
        MY_RPT1(kind,"%s\n",disp.str().c_str());
        cout << disp.str() << endl;

        return ret_code.m_rc;
    }

    if(!ret_code.m_suppress_debug_info)
        return util_debug_report_done(ret_code.m_rc);

    return ret_code.m_rc;
}

