/***************************************************************************
                          lang.cpp  -  description
                             -------------------
    begin                : Thu Dec 9 2004
    copyright            : (C) 2004-2007 by Marian VooDooMan Meravy
    email                : ghostvoodooman -a.T= users (dot) sourceforge.net
 ***************************************************************************/

/* $Id: lang.cpp,v 1.52 2008/06/29 17:37:50 ghostvoodooman Exp $ */

/***************************************************************************

VooDoo cIRCle - an IRC (ro)bot
Copyright (C) 2004-2008 by Marian VooDooMan Meravy
  (ghostvoodooman -a.T= users (dot) sourceforge.net)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************/

/*!
    \file
    \brief Supports multiple languages
    \version $Id: lang.cpp,v 1.52 2008/06/29 17:37:50 ghostvoodooman Exp $
*/

#if defined(HAVE_CONFIG_H)
#   include <config.h>
#endif

#include <string>
#include <vector>
#include <stack>
#include <list>
#include <map>

#include <stdlib.h>
#include <stdio.h>

#include "utils/memleak.h"
#include "utils/assert.h"
#include "utils.h"

#include "log.h"
#include "lang.h"
#include "compat.h"

#include "params.h"

#include "utils/precomp.h"

#if defined(_MSC_VER) || defined(__BORLANDC__)
#   pragma hdrstop
#endif

//---------------------------------------------------------------------------
#ifdef __BORLANDC__
#pragma package(smart_init)
#endif

using namespace std;

/*!
    \brief Stores language entry (string in specific language) from lang/lang\#\#.txt
    \author VooDooMan
    \version 1
    \date 2004
*/
class c_lang
{
public:
    int id;         //!< ID of message
    string lang;    //!< Language-identification string
    string text;    //!< Text of message in specific language

    c_lang()
    {
        id=0;
        lang="";
        text="";
    }
};

vector<c_lang> lang01;  //!< If lang01_init == true: stores language entries from file lang/lang01.txt
bool lang01_init=false; //!< Set to true if lang01 contains entries (had been initialized)

/*!
    \brief Initializes lang01 array from file lang/lang01.txt and sets lang01_init to true
    \author VooDooMan
    \version 1
    \date 2004
    \param file_name File name (should be "lang/lang01.txt")
    \return Returns an empty string
*/
string lang_01_init(string file_name)
{
    lang01.clear();

    FILE* f=fopen(file_name.c_str(),"r");
    if(!f)
        return (string)"";
    char ln[10240+1];
    while(!feof(f)) {
        if(fgets(ln,1024*10-1,f)==NULL)
            break;
        ln[10240]=0;
        if(ln[0] && ln[strlen(ln)-1]=='\r')
            ln[strlen(ln)-1]=0;
        if(ln[0] && ln[strlen(ln)-1]=='\n')
            ln[strlen(ln)-1]=0;
        if(ln[0] && ln[strlen(ln)-1]=='\r')
            ln[strlen(ln)-1]=0;
        if(ln[0] && ln[strlen(ln)-1]=='\n')
            ln[strlen(ln)-1]=0;
        string id_, lang_, text_;
        int pos=0;
        for(unsigned int i1=0; i1<strlen(ln); ++i1) {
            if(pos==0 && ln[i1]=='_') {
                pos++;
                continue;
            }
            if(pos==1 && ln[i1]=='=') {
                pos++;
                continue;
            }
            if(pos==0)
                id_+=ln[i1];
            if(pos==1)
                lang_+=ln[i1];
            if(pos==2)
                text_+=ln[i1];
        }
        c_lang s;
        s.id=stoi(id_);
        s.lang=lang_;
        s.text=text_;
        if(s.lang.empty() || s.text.empty())
            continue;

        vector<c_lang>::iterator i2;
        for(i2=lang01.begin(); i2!=lang01.end(); ++i2)
            if((*i2).id==s.id && !(*i2).lang.compare(s.lang)) {
                string d="OOPS: Occurred warning: duplicite language string entry for: file_id=1 lang=\""+s.lang+"\" id=";
                d+=tos(s.id);
                log_debug(GET_STR_DEBUG_INFO_EX(d));
            }

        lang01.push_back(s);
    }
    fclose(f);
    return (string)"";
}

/*!
    \brief Gets language-specific string, and initializes an array if needed
    \author VooDooMan
    \version 1
    \date 2004
    \param file_id ID of language file ("lang/langID.txt")
    \param lang Language identification string
    \param id ID of message
    \return Returns language-specific string, if not found, returns an empty string
*/
string lang_get_string(int file_id, string lang, int id)
{
    if(file_id==1 && !lang01_init) {
        lang_01_init("lang" FILE_SLASH "lang01.txt");
        lang01_init=true;
    }
    if(file_id==1) {
        vector<c_lang>::iterator i;
        for(i=lang01.begin(); i!=lang01.end(); ++i)
            if((*i).id==id && !(*i).lang.compare(lang))
                return (*i).text;
        string d="OOPS: Occurred error: cannot get language string: file_id=";
        d+=tos(file_id);
        d+=" lang=\""+lang+"\" id=";
        d+=tos(id);
        d+=" - trying default language \"en\"";
        log_debug(GET_STR_DEBUG_INFO_EX(d));
        // not found? try "en" as lang
        lang="en";
        for(i=lang01.begin(); i!=lang01.end(); ++i)
            if((*i).id==id && !(*i).lang.compare(lang))
                return (*i).text;
    }
    string d="OOPS: Occurred error: cannot get language string: file_id=";
    d+=tos(file_id);
    d+=" lang=\""+lang+"\" id=";
    d+=tos(id);
    log_debug(GET_STR_DEBUG_INFO_EX(d));

    string res="<";
    res+=tos(file_id);
    res+="|";
    res+=lang;
    res+="|";
    res+=tos(id);
    res+=">";
    return res;
}

/*!
    \brief Routine for "%s" substitution
    \author VooDooMan
    \version 2
    \date 2004, 2005
    \param str Message containing at least one "%s" to substitute
    \param substr Substring, wich will be inserted instead of FIRST "%s" found in str
    \return Returns resulting string
*/
string lang_subst(string str, string substr)
{
    string res;
    int esc=0;
    bool got=false;
    for(unsigned int i1=0; i1<str.length(); ++i1) {
        if(esc==0 && str[i1]=='%') {
            esc++;
            continue;
        }
        if(esc==1 && str[i1]=='s' /*&& substr.compare("")*/) {
            if(!got) {
                got=true;
                res+=substr;
                substr="";
            } else {
                res+="%s";
            }
            esc=0;
            continue;
        }
        if(esc==1 && str[i1]=='s') {
            res+="%s";
            esc=0;
            continue;
        }
        if((esc==1 || esc==2) && str[i1]=='%') {
            res+="%";
            continue;
        }
        res+=str[i1];
    }

    if(!got) {
        while(str.find("\r",0)!=string::npos)
            str.erase(str.find("\r",0),1);
        while(str.find("\n",0)!=string::npos)
            str.erase(str.find("\n",0),1);

        while(substr.find("\r",0)!=string::npos)
            substr.erase(substr.find("\r",0),1);
        while(substr.find("\n",0)!=string::npos)
            substr.erase(substr.find("\n",0),1);

        string log="OOPS: Warning: In function ";
        log+=__my_func__;
        log+=" I was unable to find \"%s\" sequence to substitute it with sub-string. (following strings will have \\r and \\n stripped off). String was: \"";
        log+=str;
        log+="\", and sub-string as substituent was: \"";
        log+=substr;
        log+="\"";
        log_debug(log);
    }

    return res;
}

/*!
    \brief Routine for "%xxx;" substitution
    \author VooDooMan
    \version 1
    \date 2005-07-13
    \param str Message containing "%xxx;" sequences to substitute
    \param subst Associative array for substitution
    \return Returns resulting string
*/
string lang_subst(string str, const map<string,string>& subst)
{
    string res;
    res=str;

    string error;

    {
        size_t pos=0;
        while(pos<res.length() && (pos=res.find("%%",pos))!=string::npos) {
            res.erase(pos,1);
            pos++;
        }
    }

    map<string,string>::const_iterator i1;
    for(i1=subst.begin(); i1!=subst.end(); ++i1) {
        bool got=false;
        size_t pos=0;
        while(pos<res.length() && (pos=res.find((*i1).first,pos))!=string::npos) {
            if(pos>0 && res[pos-1]=='%') {
                pos++;
                continue;
            }
            res.erase(pos,(*i1).first.length());
            res.insert(pos,(*i1).second);
            pos++;
            got=true;
        }
        if(!got) {
            if(!error.empty())
                error+=", ";
            error+="not found \"";
            error+=(*i1).first;
            error+="\"";
        }
    }

    if(!error.empty()) {
        while(str.find("\r",0)!=string::npos)
            str.erase(str.find("\r",0),1);
        while(str.find("\n",0)!=string::npos)
            str.erase(str.find("\n",0),1);

        while(error.find("\r",0)!=string::npos)
            error.erase(error.find("\r",0),1);
        while(error.find("\n",0)!=string::npos)
            error.erase(error.find("\n",0),1);

        string log="OOPS: Warning: In function ";
        log+=__my_func__;
        log+=" I was unable to find sequences to substitute them with sub-string. (following strings will have \\r and \\n stripped off). String was: \"";
        log+=str;
        log+="\", sequences which have failed: ";
        log+=error;
        log_debug(log);
    }

    return res;
}

string c_lang_file::subst(const string& lang, int id, const map<string,string>& substs)
{
    pair<int, string> id_p=make_pair(id,lang);
    map<pair<int /* id */, string /*lang_id*/>, c_lang_entry>::iterator e;
    e=entries.find(id_p);
    if(e==entries.end()) {
        string d="OOPS: Warning: Couldn't find language string: file=\"";
        d+=file_name_;
        d+="\", id=";
        d+=tos(id);
        d+=", lang=\""+id_p.second+"\" - trying to fall back to \"en\" string";
        log_debug(GET_STR_DEBUG_INFO_EX(d));

        id_p=make_pair(id,"en");
        e=entries.find(id_p);
        if(e==entries.end()) {
            string d="OOPS: Warning: Couldn't find language string: file=\"";
            d+=file_name_;
            d+="\", id=";
            d+=tos(id);
            d+=", lang=\""+id_p.second+"\"";
            log_debug(GET_STR_DEBUG_INFO_EX(d));

            return (string)"<?>";
        }
    }

    return (*e).second.get_msg(substs);
}

/*!
    \brief Enumerates available languages
    \author VooDooMan
    \version 1
    \date 2004
    \param file_id ID of language file ("lang/langID.txt")
    \param langs Returns array of available languages
*/
void lang_get_langs(int file_id, vector<string>& langs)
{
    langs.clear();

    if(file_id==1 && !lang01_init) {
        lang_01_init("lang" FILE_SLASH "lang01.txt");
        lang01_init=true;
    }
    if(file_id==1) {
        vector<c_lang>::iterator i;
        for(i=lang01.begin(); i!=lang01.end(); ++i) {
            vector<string>::iterator i1;
            bool got=false;
            for(i1=langs.begin(); i1!=langs.end(); ++i1)
                if(!(*i1).compare((*i).lang)) {
                    got=true;
                    break;
                }
            if(!got)
                langs.push_back((*i).lang);
        }
    }
}

/*!
    \brief Rehashses the language files
    \author VooDooMan
    \version 1
    \date 2004
*/
void lang_rehash()
{
    lang01_init=false;
    lang_01_init("lang" FILE_SLASH "lang01.txt");
    lang01_init=true;
}

string c_lang_entry::get_msg(const map<string,string>& replacers)
{
    return lang_subst(message_,replacers);
}

c_lang_file::c_lang_file(const string& file_name)
{
    file_name_=file_name;
    rehash();
}

void c_lang_file::rehash()
{
    entries.clear();

    FILE* f=fopen(file_name_.c_str(),"r");
    if(!f)
        return;
    char ln[10240+1];
    while(!feof(f)) {
        if(fgets(ln,1024*10-1,f)==NULL)
            break;
        ln[10240]=0;
        if(ln[0] && ln[strlen(ln)-1]=='\r')
            ln[strlen(ln)-1]=0;
        if(ln[0] && ln[strlen(ln)-1]=='\n')
            ln[strlen(ln)-1]=0;
        if(ln[0] && ln[strlen(ln)-1]=='\r')
            ln[strlen(ln)-1]=0;
        if(ln[0] && ln[strlen(ln)-1]=='\n')
            ln[strlen(ln)-1]=0;
        string id_, lang_, text_;
        int pos=0;
        for(unsigned int i1=0; i1<strlen(ln); ++i1) {
            if(pos==0 && ln[i1]=='_') {
                pos++;
                continue;
            }
            if(pos==1 && ln[i1]=='=') {
                pos++;
                continue;
            }
            if(pos==0)
                id_+=ln[i1];
            if(pos==1)
                lang_+=ln[i1];
            if(pos==2)
                text_+=ln[i1];
        }
        if(lang_.empty() || text_.empty())
            continue;
        int id__=stoi(id_);
        pair<int, string> id_p=make_pair(id__, lang_);
        if(entries.find(id_p)!=entries.end()) {
            string d="OOPS: Ignoring duplicite language string entry for: file=\"";
            d+=file_name_;
            d+="\", id=";
            d+=tos(id__);
            d+=", lang=\""+lang_+"\"";
            log_debug(GET_STR_DEBUG_INFO_EX(d));
            continue;
        }

        c_lang_entry e(text_);
        entries.insert(make_pair(id_p,e));
    }
    fclose(f);
    return;
}

/*!
    \brief Substitutes multiple
    \version 1
    \date 2006-01-13 (Friday)
    \param str String with multiple "%s" sequences that will be substitued
    \param subst Array of strings that will replace "%s"'s
    \return Returns resulting string
*/
string lang_subst_multi(const string& str, const vector<string>& subst)
{
    string res;
    size_t i1=0;
    vector<string>::const_iterator is=subst.begin();
    while(is!=subst.end()) {
        size_t pos=str.find("%s",i1);
        if(string::npos==i1)
            break;
        for(size_t i2=i1; i2<pos; ++i2)
            res.append(1,str[i2]);
        res+=*(is++);
        i1=pos+2; // '2' stands for strlen("%s")
    }
#if defined(_DEBUG)
    if(is!=subst.end() || str.find("%s",i1)!=string::npos) {
        // DEBUG: set breakpoint here!
        int x=0;
        x++;
    }
#endif
    return res;
}

