/***************************************************************************
                          lang.h  -  description
                             -------------------
    begin                : Thu Dec 9 2004
    copyright            : (C) 2004-2007 by Marian VooDooMan Meravy
    email                : ghostvoodooman -a.T= users (dot) sourceforge.net
 ***************************************************************************/

/* $Id: lang.h,v 1.43 2008/03/11 21:11:29 ghostvoodooman Exp $ */

/***************************************************************************

VooDoo cIRCle - an IRC (ro)bot
Copyright (C) 2004-2008 by Marian VooDooMan Meravy
  (ghostvoodooman -a.T= users (dot) sourceforge.net)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************/

/*!
    \file
    \brief Supports multiple languages
    \version $Id: lang.h,v 1.43 2008/03/11 21:11:29 ghostvoodooman Exp $
*/

#ifndef LANG_H__
#define LANG_H__

#if defined(HAVE_CONFIG_H)
#   include <config.h>
#endif

#include <string>
#include <vector>
#include <stack>
#include <list>
#include <map>

#include "utils/memleak.h"
#include "utils/assert.h"
#include "utils.h"
#include "params.h"

using namespace std;

/*!
    \brief Stores one message from language file
    \author VooDooMan
    \version 1
    \date 2005
*/
class c_lang_entry
{
    string message_;                    //!< Message
public:
    /*!
        \brief Constructor
        \author VooDooMan
        \version 1
        \date 2005
        \param message Message to set
    */
    c_lang_entry(const string& message)
        : message_(message)
    { ; }

    /*!
        \brief Returns substitued message
        \author VooDooMan
        \version 1
        \date 2005
        \param subst Substitutes
    */
    string get_msg(const map<string,string>& subst);
};

/*!
    \brief Stores language file
    \author VooDooMan
    \version 1
    \date 2005
*/
class c_lang_file
{
    string file_name_;                      //!< Name of file
    map<pair<int /* id */, string /*lang_id*/>, c_lang_entry> entries;  //!< Language file entries
public:

    /*!
        \brief Constructor
        \author VooDooMan
        \version 1
        \date 2005
        \param file_name Name of file to parse
    */
    c_lang_file(const string& file_name);

    /*!
        \brief Rehashes
        \author VooDooMan
        \version 1
        \date 2005
    */
    void rehash();

    /*!
        \brief Substitues message
        \author VooDooMan
        \version 1
        \date 2005
        \param lang Language code
        \param id ID of message
        \param substs Substitutes
    */
    string subst(const string& lang, int id, const map<string,string>& substs);
};

string lang_get_string(int file_id, string lang, int id);
string lang_subst(string str, string substr);
string lang_subst(string str, const map<string,string>& subst);
void lang_get_langs(int file_id, vector<string>& langs);

void lang_rehash();

string lang_subst_multi(const string& str, const vector<string>& subst);

#endif
