/***************************************************************************
                          log.cpp  -  description
                             -------------------
    begin                : Thu Dec 9 2004
    copyright            : (C) 2004-2007 by Marian VooDooMan Meravy
    email                : ghostvoodooman -aT= users (dot) sourceforge.net
 ***************************************************************************/

/* $Id: log.cpp,v 1.73 2008/11/24 12:09:23 ghostvoodooman Exp $ */

/***************************************************************************

VooDoo cIRCle - an IRC (ro)bot
Copyright (C) 2004-2008 by Marian VooDooMan Meravy
  (ghostvoodooman -aT= users (dot) sourceforge.net)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************/

/*!
    \file
    \brief Handles log files
    \version $Id: log.cpp,v 1.73 2008/11/24 12:09:23 ghostvoodooman Exp $
*/

#if defined(HAVE_CONFIG_H)
#   include <config.h>
#endif

#include <string>
#include <vector>
#include <stack>
#include <list>
#include <map>
#include <sstream>
#include <iostream>
#include <fstream>

#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>

#include <stdarg.h>

#ifndef _MSC_VER
#if defined(_WIN32) || defined(_WIN64)
#       include <dir.h>
#       include <io.h>
#   else
#       include <sys/types.h>
#       include <sys/stat.h>
#       include <dirent.h>
#       define ROT_LOG_POSIX
#   endif
#else
#   include <direct.h>
#   include <io.h>
#   define mkdir _mkdir
#endif

#include "utils/memleak.h"
#include "utils/assert.h"
#include "utils.h"

#include "conf.h"
#include "stats.h"
#include "shared.h"
#include "logic.h"
#include "compat.h"
#include "modules.h"
#include "utils.h"
#include "sock.h"
#include "params.h"
#include "log.h"
#include "conf/c_conf.h"
#include "thread/ipc_mutex.h"
#include "utils/my_getpid.h"
#include "misc/shutdown.h"
#include "misc/c_lock_releaser.h"
#include "internal_state.h"

#if defined(__CYGWIN__)
#   include <windows.h>     // for Sleep()
#endif

#include "utils/precomp.h"

#if defined(_MSC_VER) || defined(__BORLANDC__)
#   pragma hdrstop
#endif

//---------------------------------------------------------------------------

using namespace std;

bool log_initialized=false;             //!< Was echoes/logs initialized?

static c_lock_releaser s_lock_releaser; //!< Lock file releaser class

bool protect_pid=false;                 //!< To protect .pid and .lock files from beeing overwritten
bool overwrite_pid=false;               //!< To protect .pid files from beeing overwritten

static MY_U_LONG_LONG log_event_id=0;   //!< Unique log event, assigned to each log message, then incremented
bool log_cumul=true;                    //!< If to write to cumulative log file ("logs/cumul.log")
bool log_suppress_any_logs=false;       //!< 'true' to suppress "cumul.log" and "ssl.log" regardless of "log_anyway"

#define LOCK_1_FILE     "logs" FILE_SLASH "irc_bot.lock1"
#define LOCK_2_FILE     "logs" FILE_SLASH "irc_bot.lock2"

MY_STATIC c_ipc_mutex log_files_mutex(LOCK_1_FILE);        //!< Mutex for writing to log files
MY_STATIC c_ipc_mutex log_cumul_file_mutex(LOCK_2_FILE);   //!< Mutex for cumulative log

MY_STATIC void log_check_semaphore_timeout()
{
    unsigned long cnt1, cnt2;

    cnt1=log_files_mutex.get_timeout_cnt_and_reset();
    cnt2=log_cumul_file_mutex.get_timeout_cnt_and_reset();
    if(cnt1 || cnt2) {
        if(cnt1) {
            ostringstream log;
            log << LOCK_1_FILE << " semaphore condition timeout (1 second) " << cnt1 << " my_time(s)";
            log << "(another process killed?)";
            log_bot(log.str());
            log_debug(log.str());
        }
        if(cnt2) {
            ostringstream log;
            log << LOCK_2_FILE << " semaphore condition timeout (1 second) " << cnt1 << " my_time(s)";
            log << "(another process killed?)";
            log_bot(log.str());
            log_debug(log.str());
        }
    }
}

/*!
    \brief Returns an unique log event ID
    \author VooDooMan
    \version 2
    \date 2006-12-25, 2007-07-27
*/
MY_U_LONG_LONG log_get_next_event_id()
{
    static c_mutex mutex;
    c_mutex_locker lock(mutex);

    return log_event_id++;
}

/*!
    \brief Retuns prefix for log entries (pid)
    \version 2
    \date 2007-08-20
    \author VooDooMan
*/
MY_STATIC const string& log_get_log_prefix()
{
    static const string s_prefix=(string)"["+tos(my_getpid())+"] ";
    return s_prefix;
}

/*!
    \brief Initializes echoes to stdout (log_echo_XXX) form conf.txt
*/
void init_echoes()
{
    log_initialized=true;

#if (defined(_WIN32) || defined(_WIN64)) && !defined(__CYGWIN__)
    mkdir("logs");
#else
    mkdir("." FILE_SLASH "logs",0777);
#endif
}

/*!
    \brief Writes an entry to cumulative log
    \author VooDooMan
    \date 2006-12-25
    \version 1
*/
void log_cumulative(const MY_U_LONG_LONG event_id, const e_log_type type, const string& channel, const string& log)
{
    if(log_suppress_any_logs)
        return;

    if(!log_cumul)
        return;

    c_mutex_locker lock(log_cumul_file_mutex);
    lock.lock_file_if_needed(LOCK_2_FILE);
    ofstream f("logs" FILE_SLASH "cumul.log",ios::app);

    f << make_nice_time(my_time(NULL),true) << " ";

    switch(type) {
        case E_LOG_CUSTOM:
            f << "[CUSTOM       ]";
            break;
        case E_LOG_IDENTD:
            f << "[IDENTD       ]";
            break;
        case E_LOG_SSL:
            f << "[SSL          ]";
            break;
        case E_LOG_BOT:
            f << "[BOT          ]";
            break;
        case E_LOG_DEBUG:
            f << "[DEBUG        ]";
            break;
        case E_LOG_IRC:
            f << "[IRC          ]";
            break;
        case E_LOG_SOCKET:
            f << "[SOCKET       ]";
            break;
        case E_LOG_CHAN:
            f << "[CHANNEL      ]";
            break;
        case E_LOG_BOTNET_DEBUG:
            f << "[BOTNET_DEBUG ]";
            break;
        case E_LOG_BOTNET:
            f << "[BOTNET       ]";
            break;
        case E_LOG_IRC_BCAST:
            f << "[IRC_BROADCAST]";
            break;
        case E_LOG_CLIENTS:
            f << "[CLIENTS      ]";
            break;
        case E_LOG_DEBUG_RPT:
            f << "[DEBUG_RPT    ]";
            break;
        default:
            f << "[<OTHER>      ]";
            break;
    }
    f << " [" << event_id << "] " << log_get_log_prefix() << ":";
    if(E_LOG_CHAN==type)
        f << (!channel.empty()?channel:"<NULL> :") << " :";
    f << log << endl;
    f.flush();
    f.close();
}

/*!
    \brief Writes debug string to file logs/debug.log and to stdout if log_echo_debug == true
    \author VooDooMan
    \version 1
    \date 2004
    \param msg The log message
    \param log_anyway If to log even if it is disabled
*/
void log_debug(const string& msg, const bool log_anyway)
{
    log_check_semaphore_timeout();
    MY_U_LONG_LONG event_id=log_get_next_event_id();
    log_cumulative(event_id,E_LOG_DEBUG,"",msg);

    if(!log_initialized)
        init_echoes();

    c_conf* core_conf=get_core_conf();

    if(log_anyway || (core_conf && core_conf->log_log_debug())) {
        c_mutex_locker lock(log_files_mutex);
        lock.lock_file_if_needed(LOCK_1_FILE);
        ofstream f("logs" FILE_SLASH "debug.log",ios::app);
        f << make_nice_time(my_time(NULL),true) << " " << log_get_log_prefix() << ":" << msg << endl;
        f.flush();
        f.close();
    }

    if(log_anyway || (core_conf && core_conf->log_echo_debug())) {
        cout << "DEBUG   " << msg << endl;
        cout.flush();
    }

    {
        logic_on_internal_event("@log@","debug","","","",PRIVATE_EVENT_SEVERITY_CODE_INFORMATIONAL,msg,"");
    }
}

/*!
    \brief Writes line to file logs/irc.log and to stdout if log_echo_irc == true
    \author VooDooMan
    \version 2
    \date 2004, 2007-07-27
    \param from_server Set to true if message goes from server, false if it goes to server
    \param msg The log message
    \param log_anyway If to log even if it is disabled
*/
void log_irc(const bool from_server, const string& msg, const bool log_anyway)
{
    log_irc_ex(from_server ? E_LOG_IRC_EX_IN : E_LOG_IRC_EX_OUT,msg,log_anyway);
}

/*!
    \brief Writes debug string to file logs/socket.log and to stdout if log_echo_socket == true
    \author VooDooMan
    \version 2
    \date 2004, 2005-07-31
    \param sock_error OS's socket I/O error code
    \param os_msg Message from OS describing socket I/O error
    \param msg The log message
    \param socket Socket handle for comment (or NULL)
    \param log_anyway If to log even if it is disabled
*/
void log_socket(const int sock_error, const string& os_msg, const string& msg, const c_socket* socket, const bool log_anyway)
{
    log_check_semaphore_timeout();
    MY_U_LONG_LONG event_id=log_get_next_event_id();

    if(!log_initialized)
        init_echoes();

    string msg_=msg;
    if(socket && !socket->comment.empty()) {
        msg_+=" (";
        msg_+=socket->comment;
        msg_+=")";
    }

    c_conf* core_conf=get_core_conf();

    ostringstream log;
    log << "OS socket error code: " << sock_error << " \"" << os_msg << "\" (" << msg_ << ")";

    log_cumulative(event_id,E_LOG_SOCKET,"",log.str());

    if(log_anyway || (core_conf && core_conf->log_log_socket())) {
        c_mutex_locker lock(log_files_mutex);
        lock.lock_file_if_needed(LOCK_1_FILE);
        ofstream f("logs" FILE_SLASH "socket.log",ios::app);
        f << make_nice_time(my_time(NULL),true) << " " << log_get_log_prefix() << ":OS socket error code: " << sock_error << " " << os_msg << " (" << msg_ << ")" << endl;
        f.flush();
        f.close();
    }

    if(log_anyway || (core_conf && core_conf->log_echo_socket())) {
        cout << "SOCKET  \"" << os_msg << "\" :" << msg_ << endl;
        cout.flush();
    }

    logic_on_internal_event("@log@","socket",tos(sock_error).c_str(),"","",PRIVATE_EVENT_SEVERITY_CODE_INFORMATIONAL,os_msg,msg_.c_str());
}

/*!
    \brief Writes debug string to file logs/bot.log and to stdout if log_echo_bot == true
    \author VooDooMan
    \version 1
    \date 2004
    \param msg The log message
    \param log_anyway If to log even if it is disabled
*/
void log_bot(const string& msg, const bool log_anyway)
{
    log_check_semaphore_timeout();

    c_logger* core_logger=get_core_logger();
    if(core_logger)
        core_logger->log_print1(E_LOG_BOT,LOG_SCOPE_GLOBAL,"",msg,log_anyway);
    else {
        MY_U_LONG_LONG event_id=log_get_next_event_id();
        log_cumulative(event_id,E_LOG_BOT,"",msg);
        if(log_anyway) {
            c_mutex_locker lock(log_files_mutex);
            lock.lock_file_if_needed(LOCK_1_FILE);
            ofstream f("logs" FILE_SLASH "bot.log",ios::app);
            f << make_nice_time(my_time(NULL),true) << " " << log_get_log_prefix() << ":" << msg << endl;
            f.flush();
            f.close();
        }
    }
}

/*!
    \brief Writes debug string to file logs/identd.log and to stdout if log_echo_identd == true
    \author VooDooMan
    \version 1
    \date 2004
    \param msg The log message
    \param log_anyway If to log even if it is disabled
*/
void log_identd(const string& msg, const bool log_anyway)
{
    log_check_semaphore_timeout();

    c_logger* core_logger=get_core_logger();
    if(core_logger)
        core_logger->log_print1(E_LOG_IDENTD,LOG_SCOPE_GLOBAL,"",msg,log_anyway);
    else {
        MY_U_LONG_LONG event_id=log_get_next_event_id();
        log_cumulative(event_id,E_LOG_IDENTD,"",msg);
        if(log_anyway) {
            c_mutex_locker lock(log_files_mutex);
            lock.lock_file_if_needed(LOCK_1_FILE);
            ofstream f("logs" FILE_SLASH "identd.log",ios::app);
            f << make_nice_time(my_time(NULL),true) << " " << log_get_log_prefix() << ":" << msg << endl;
            f.flush();
            f.close();
        }
    }
}

/*!
    \brief Writes debug string to file logs/\#channel_name.log, or logs/status.log and to stdout if log_echo_channel == true
    \author VooDooMan
    \version 2
    \date 2004, 2005
    \param channel Channel name, or empty string for message to logs/status.log
    \param type Type of message: TYPE_PRIVMSG, TYPE_NOTICE, TYPE_JOIN, TYPE_PART, TYPE_QUIT, TYPE_KICK, TYPE_MODE, TYPE_NICK, TYPE_SCRIPT, TYPE_IRCOP_JOIN, TYPE_IRCOP_LEFT
    \param who For type == TYPE_PRIVMSG, or TYPE_NOTICE, or TYPE_JOIN, or TYPE_PART, or TYPE_QUIT, TYPE_KICK, TYPE_MODE host string ("nick!ident@host") of originator of message, and for type TYPE_NICK host string with old nick name
    \param whom For type == TYPE_NOTICE who is target (channel name or bot's nick), or for type == TYPE_KICK nick of target
    \param msg For type == TYPE_PRIVMSG, or TYPE_NOTICE, or TYPE_PART, or TYPE_KICK, or TYPE_QUIT message, or for type == TYPE_MODE MODE string (i.e. "+b *!*@*.gov), or for type == TYPE_SCRIPT message from script result, or for type == TYPE_TOPIC it is new topic
    \param user_modes Mode(s) of user (e.g. "v", "o", "h", "vo",...), or single character long string "?" for not-got-WHOIS-yet
*/
void log_channel(const string& channel, const int type, const string& who, const string& whom, const string& msg, const string& user_modes)
{
    log_check_semaphore_timeout();
    MY_U_LONG_LONG event_id=log_get_next_event_id();

    if(!log_initialized)
        init_echoes();

    if(channel.length()>128)
        return;

    string ch;
    if(!channel.empty())
        ch=c_logger::get_channel_log_name(channel);

    string fn="logs" FILE_SLASH;
    if(!ch.empty())
        fn+=ch;
    else
        fn+="status";
    fn+=".log";

    char modes[128];
    unsigned ii2=0;
    modes[ii2++]='[';
    bool first=true;
    for(unsigned ii=0; ii<128-16 && user_modes[ii]!=0; ++ii) {
        if(ii==0 && (type==TYPE_IRCOP_JOIN || type==TYPE_IRCOP_LEFT) && user_modes[ii]!='!')
            modes[ii2++]='!';
        if(first) {
            if(ii==0 && user_modes[ii]=='?') {
                first=false;
                continue;
            }
            if(ii==0 && user_modes[ii]=='!') {
                first=true;
                modes[ii2++]='!';
                continue;
            }
            modes[ii2++]='+';
            first=false;
        }
        modes[ii2++]=user_modes[ii];
    }
    while(ii2<8+2-1) // (8 chars) + (one * '[') + (one * ']') - (1 coz it's zero-based index)
        modes[ii2++]=0x20;
    modes[ii2++]=']';
    modes[ii2++]=0;

    ostringstream log;

    {
        c_mutex_locker lock(log_files_mutex);
        lock.lock_file_if_needed(LOCK_1_FILE);
        ofstream f(fn.c_str(),ios::app);
        f << make_nice_time(my_time(NULL),true) << " " << log_get_log_prefix();
        switch(type) {
            case TYPE_PRIVMSG:
                log << "PRIVMSG " << modes << " <" << who << "> :" << msg;
                log_cumulative(event_id,E_LOG_CHAN,channel,log.str());
                break;
            case TYPE_NOTICE:
                log << "NOTICE  " << modes << " <" << who << "> " << whom << " :" << msg;
                log_cumulative(event_id,E_LOG_CHAN,channel,log.str());
                break;
            case TYPE_JOIN:
                log << "JOIN    " << modes << " <" << who << "> :" << msg;
                log_cumulative(event_id,E_LOG_CHAN,channel,log.str());
                break;
            case TYPE_PART:
                log << "PART    " << modes << " <" << who << "> :" << msg;
                log_cumulative(event_id,E_LOG_CHAN,channel,log.str());
                break;
            case TYPE_QUIT:
                log << "QUIT    " << modes << " <" << who << "> :" << msg;
                log_cumulative(event_id,E_LOG_CHAN,channel,log.str());
                break;
            case TYPE_KICK:
                log << "KICK    " << modes << " <" << who << "> " << whom << " :" << msg;
                log_cumulative(event_id,E_LOG_CHAN,channel,log.str());
                break;
            case TYPE_MODE:
                log << "MODE    " << modes << " <" << who << "> :" << msg;
                log_cumulative(event_id,E_LOG_CHAN,channel,log.str());
                break;
            case TYPE_NICK:
                log << "NICK    " << modes << " <" << who << "> :" << msg;
                log_cumulative(event_id,E_LOG_CHAN,channel,log.str());
                break;
            case TYPE_SCRIPT:
                log << "SCRIPT  " << ":" << msg;
                log_cumulative(event_id,E_LOG_CHAN,channel,log.str());
                break;
            case TYPE_TOPIC:
                log << "TOPIC   " << modes << " <" << who << "> :" << msg;
                log_cumulative(event_id,E_LOG_CHAN,channel,log.str());
                break;
            case TYPE_IRCOP_JOIN:
                log << "IRCOP+  " << modes << " <" << who << "> :Is an IRC operator on channel " << channel;
                log_cumulative(event_id,E_LOG_CHAN,channel,log.str());
                break;
            case TYPE_IRCOP_LEFT:
                log << "IRCOP-  " << modes << " <" << who << "> :Was an IRC operator on channel " << channel;
                log_cumulative(event_id,E_LOG_CHAN,channel,log.str());
                break;
            default:
                break;
        }
        f << log.str() << endl;
        f.flush();
        f.close();
        cout << log.str() << endl;
        cout.flush();
    }

    string chnl=channel;
    if(chnl.empty())
        chnl="status";
    if(!log.str().empty())
        logic_on_internal_event("@log@","channel",chnl,"","",PRIVATE_EVENT_SEVERITY_CODE_INFORMATIONAL,log.str(),"");
}

/*!
    \brief Writes error message of origin botnet (botnet debug messages)
    \author VooDooMan
    \version 1
    \date 2004
    \param myname Name of this bot
    \param remote_name Name of remote bot
    \param packet_type Type of packet
    \param command Command number
    \param msg The log message
    \param log_anyway If to log even if it is disabled
*/
void log_botnet_debug(const string& myname, const string& remote_name, const int packet_type, const int command, const string& msg, const bool log_anyway)
{
    log_check_semaphore_timeout();
    MY_U_LONG_LONG event_id=log_get_next_event_id();

    if(!log_initialized)
        init_echoes();

    c_conf* core_conf=get_core_conf();

    ostringstream log;
    log << "MyName=" << myname << " RemoteName=" << remote_name << " PacketType=" << packet_type << " Command=" << command << " :" << msg;

    if(log_anyway || (core_conf && core_conf->log_log_botnet_debug())) {
        c_mutex_locker lock(log_files_mutex);
        lock.lock_file_if_needed(LOCK_1_FILE);
        ofstream f("logs" FILE_SLASH "botnet_debug.log",ios::app);
        f << make_nice_time(my_time(NULL),true) << " " << log_get_log_prefix() << ":" << log.str() << endl;
        f.flush();
        f.close();
    }

    log_cumulative(event_id,E_LOG_BOTNET_DEBUG,"",log.str());

    if(log_anyway || (core_conf && core_conf->log_echo_botnet_debug())) {
        cout << "BOTNET  " << log.str() << endl;
        cout.flush();
    }

    string final="BOTNET  ";
    final+=log.str();
    logic_on_internal_event("@log@","botnet_debug",remote_name,"","",PRIVATE_EVENT_SEVERITY_CODE_INFORMATIONAL,final,"");
}

/*!
    \brief Writes error message of origin botnet
    \author VooDooMan
    \version 1
    \date 2004
    \param myname Name of this bot
    \param remote_name Name of remote bot
    \param msg The log message
    \param log_anyway If to log even if it is disabled
*/
void log_botnet(const string& myname, const string& remote_name, const string& msg, const bool log_anyway)
{
    log_check_semaphore_timeout();
    MY_U_LONG_LONG event_id=log_get_next_event_id();

    if(!log_initialized)
        init_echoes();

    c_conf* core_conf=get_core_conf();

    ostringstream log;

    if(log_anyway || (core_conf && core_conf->log_log_botnet())) {
        c_mutex_locker lock(log_files_mutex);
        lock.lock_file_if_needed(LOCK_1_FILE);
        ofstream f("logs" FILE_SLASH "botnet.log",ios::app);
        f << make_nice_time(my_time(NULL),true) << " " << log_get_log_prefix() << ":";
        log << myname << "->" << remote_name << " :" << msg;
        f << log.str() << endl;
        f.flush();
        f.close();
    }

    log_cumulative(event_id,E_LOG_BOTNET,"",log.str());

    if(log_anyway || (core_conf && core_conf->log_echo_botnet())) {
        cout << "BOTNET  " << log.str() << endl;
        cout.flush();
    }

    logic_on_internal_event("@log@","botnet",remote_name,"","",PRIVATE_EVENT_SEVERITY_CODE_INFORMATIONAL,log.str(),"");
}

/*!
    \brief Writes debug string to file logs/ssl.log and to stdout if log_echo_ssl == true
    \author VooDooMan
    \version 1
    \date 2004
    \param msg The log message
    \param log_anyway If to log even if it is disabled
*/
void log_ssl(const string& msg, const bool log_anyway)
{
    if(log_suppress_any_logs)
        return;

    log_check_semaphore_timeout();

    c_logger* core_logger=get_core_logger();
    if(core_logger)
        core_logger->log_print1(E_LOG_SSL,LOG_SCOPE_GLOBAL,"",msg,log_anyway);
    else {
        MY_U_LONG_LONG event_id=log_get_next_event_id();
        log_cumulative(event_id,E_LOG_SSL,"",msg);
        if(log_anyway) {
            c_mutex_locker lock(log_files_mutex);
            lock.lock_file_if_needed(LOCK_1_FILE);
            ofstream f("logs" FILE_SLASH "ssl.log",ios::app);
            f << make_nice_time(my_time(NULL),true) << " " << log_get_log_prefix() << ":" << msg << endl;
            f.flush();
            f.close();
        }
    }
}

/*!
    \brief Inits the log files
    \author VooDooMan
    \version 1
    \date 2004
    \retval true If success
    \retval false If another instance is running (file "logs/irc_bot.pid" is blocked for unlink-ing)
*/
bool log_init()
{
    {
        FILE* f=fopen("conf.txt","r");
        if(f) {
            while(!feof(f)) {
                char tmp[128+1];
                if(!fgets(tmp,128,f))
                    break;
                if(strlen(tmp)) {
                    if('\n'==tmp[strlen(tmp)-1])
                        tmp[strlen(tmp)-1]='\0';
                    if(strlen(tmp)) {
                        vector<string> p;
                        explode(tmp,'=',p,true);
                        if(p.size()==2 && !(*p.begin()).compare("time_offset")) {
                            string tz=p[1];
                            string off=c_conf::convert_old_time_offset_if_needed(tz);

                            nice_time_set_time_offset(off);

                            if(off.compare(tz)) { // !=
                                // old obsolete format

                                string log="OOPS: Key \"time_offset\" from \"conf.txt\""
                                    " has an obsolete format. Please, use e.g. \"+0200\" for UTC+02, "
                                    "or \"-0630\" for UTC-06 and 30 minutes and so on. "
                                    "Temporarily converting value \"";
                                log+=tz;
                                log+="\" to \"";
                                log+=off;
                                log+="\"";
                                log_bot(log,true);
                                log_debug(log,true);
                            }
                            break;
                        }
                    }
                }
            }
            fclose(f);
        }
    }
#if (defined(_WIN32) || defined(_WIN64)) && !defined(__CYGWIN__)
    mkdir("logs");
#else
    mkdir("." FILE_SLASH "logs",0777);
#endif

    if(!log_lock(false))
        return false;

    string log="BOT STARTUP!";
    {
        log+=" (pid ";
        log+=tos(my_getpid());
        log+=")";
    }
    log_all_logs(log,true);

    sock_log_fd_array_size(true);

    {
        string log=utils_get_timezone();
        log=((string)"Using timezone ")+log;
        log_all_logs(log,true);
    }

    return true;
}

/*!
    \brief Frees the lock files
    \author VooDooMan
    \version 2
    \date 2005, 2007-08-19
*/
void log_free_locks()
{
    e_shutdown_reason sr=bot_get_shutdown_state();
    if(E_SD_UPGRADE==sr) {
        s_lock_releaser.release();
        return;
    }
    // else: Let releasing of lock/pid to be a destructor's job
}

/*!
    \brief Shuts down the log files
    \author VooDooMan
    \version 2
    \date 2004, 2005-07-31
    \param exit_code Exit code for bot used to write it to log
    \param eol EOL string for user who is killing the bot (only if it is a real user, not "root")
    \return Returns kill log using EOL as end-of-line marker for user who is killing the bot
*/
string log_done(int exit_code, string eol)
{
    string res;
    if(log_initialized) {
        log_bot("*** Statistics",true);

        res+="*** Statistics"+eol;
        list<string> lines;
        stats_display(lines,"");
        list<string>::iterator i1=lines.begin();
        for(; i1!=lines.end(); ++i1) {
            string ln;
            ln=(string)"*** "+*i1;
            res+=(string)"*** "+*i1+eol;
            log_bot(ln,true);
        }

        {
            string my_pid;
            my_pid="*** I was pid ";
            my_pid+=tos(my_getpid());

            log_bot(my_pid,true);
            res+=(string)my_pid+eol;
        }

        char msg[1024];
        sprintf(msg,"I will exit with exit code %s",tos(exit_code).c_str());
        log_bot(msg,true);
        res+=(string)msg+eol;
    }

    log_free_locks();

    return res;
}

/*!
    \brief Writes string to file logs/broadcast.log and to stdout if log_echo_broadcast == true
    \author VooDooMan
    \version 1
    \date 2005-04-22
    \param bcast_mask Broadcast mask used (something like "$$irc.nonexistingdomain.org")
    \param type Type of message: TYPE_PRIVMSG, TYPE_NOTICE; TODO: KILL
    \param who For type == TYPE_PRIVMSG, or TYPE_NOTICE ("nick!ident@host") of originator of message (most likely an irc op)
    \param msg Message
    \param server Server DNS host we are using now
    \param server_port Port of IRC server we are on
*/
void log_broadcast(const string& bcast_mask, const int type, const string& who, const string& msg, const string& server, const unsigned short server_port)
{
    log_check_semaphore_timeout();
    MY_U_LONG_LONG event_id=log_get_next_event_id();

    if(!log_initialized)
        init_echoes();

    ostringstream host_str;
    bool _ipv6=server.find(":",0)!=string::npos;
    if(_ipv6)
        host_str << "[";
    host_str << server;
    if(_ipv6)
        host_str << "]";
    host_str << ":";
    host_str << server_port;

    string fn="logs" FILE_SLASH "broadcast.log";

    ostringstream log;

    {
        c_mutex_locker lock(log_files_mutex);
        lock.lock_file_if_needed(LOCK_1_FILE);
        ofstream f(fn.c_str(),ios::app);
        f << make_nice_time(my_time(NULL),true) << " " << log_get_log_prefix();
        switch(type) {
            case TYPE_PRIVMSG:
                log << ":" << "<" << who << "> PRIVMSG " << bcast_mask << "(using server " << host_str << ") :" <<msg;
                f << log.str();
                log_cumulative(event_id,E_LOG_IRC_BCAST,"",log.str());
                break;
            case TYPE_NOTICE:
                log << ":" << "<" << who << "> NOTICE  " << bcast_mask << "(using server " << host_str << ") :" <<msg;
                f << log.str();
                log_cumulative(event_id,E_LOG_IRC_BCAST,"",log.str());
                break;
            // TODO: case KILL
            default:
                break;
        }
        f << endl;
        f.flush();
        f.close();
    }

    logic_on_internal_event("@log@","broadcast",who,"","",PRIVATE_EVENT_SEVERITY_CODE_INFORMATIONAL,log.str(),"");
}

string c_logger::get_channel_log_name(string channel)
{
    if(channel.empty())
        return channel;
    char prefix=channel[0];
    switch(prefix) {
        case '#':
        case '!':
        case '&':
        case '+':
            break;
        default:
            if(
                (prefix>='a' && prefix<='Z') ||
                (prefix>='A' && prefix<='Z') ||
                (prefix>='0' && prefix<='9')
                )
                prefix=0;
            break;
    }
    if(!prefix)
        return (string)""; // it is not a channel name
    for(size_t i1=0; i1<channel.length(); ++i1) {
        channel[i1]=tolowertab[(unsigned char)channel[i1]];
        unsigned char* tmp=(unsigned char*)&channel[i1];
        if(
            *tmp<=32 || *tmp>=127 || *tmp=='*' || *tmp=='/' || *tmp==':' ||
            *tmp=='<' || *tmp=='>' || *tmp=='?' || *tmp=='\\' || *tmp=='|' ||
            *tmp=='&' || *tmp==127 || *tmp=='~' || *tmp=='`' || *tmp=='"' ||
            *tmp=='\''
            )
            *tmp='_';
    }
    return channel;
}

void c_logger::log_print1(e_log_type type, int scope, const string& channel, const string& msg, bool log_anyway)
{
    log_check_semaphore_timeout();
    MY_U_LONG_LONG event_id=log_get_next_event_id();

    if(scope & LOG_SCOPE_DEFAULT)
        scope|=default_scope_;

    string file;

    bool log_mod=false, log_glob=false;

    switch(type) {
        case E_LOG_DEBUG:
            file="debug.log";
            log_mod=mod_conf && mod_conf->log_log_debug();
            log_glob=glob_conf && glob_conf->log_log_debug();
            break;
        case E_LOG_CHAN:
            file=get_channel_log_name(channel);
            file+=".log";
            log_mod=mod_conf && mod_conf->log_log_irc();
            log_glob=glob_conf && glob_conf->log_log_irc();
            break;
        case E_LOG_BOT:
            file="bot.log";
            log_mod=mod_conf && mod_conf->log_log_bot();
            log_glob=glob_conf && glob_conf->log_log_bot();
            break;
        case E_LOG_IDENTD:
            file="identd.log";
            log_mod=mod_conf && mod_conf->log_log_identd();
            log_glob=glob_conf && glob_conf->log_log_identd();
            break;
        case E_LOG_SSL:
            file="ssl.log";
            log_mod=mod_conf && mod_conf->log_log_ssl();
            log_glob=glob_conf && glob_conf->log_log_ssl();
            break;
        case E_LOG_IRC:
            file="irc.log";
            log_mod=mod_conf && mod_conf->log_log_irc();
            log_glob=glob_conf && glob_conf->log_log_irc();
            break;
        case E_LOG_CUSTOM:
            if(!custom_name.compare("core")) {
                file="debug.log";
                log_mod=mod_conf && mod_conf->log_log_debug();
                break;
            }
            // else: fall thru default
        default:
            log_mod=true;
            file=custom_name;
            file+=".log";
            break;
    }

    if(log_anyway)
        log_mod=log_glob=true;

    string now_time=make_nice_time(my_time(NULL),true);

    if(E_LOG_CHAN==type)
        log_cumulative(event_id,type,channel,msg);

    ostringstream log;

    if(type!=E_LOG_IDENTD && type!=E_LOG_SSL) {
        log << "MODULE(" << custom_name << ") :" << msg;
        if(E_LOG_CHAN!=type)
            log_cumulative(event_id,type,channel,log.str());
    } else {
        log << ":" << msg;
        if(E_LOG_CHAN!=type) // odd check, just for for code's clarity
            log_cumulative(event_id,type,channel,msg);
    }

    bool got=false;

    if(log_mod && (scope & LOG_SCOPE_MODULE)) {
        got=true;
        string mod=dir_module_logs_;
        mod+=file;

        c_mutex_locker lock(log_files_mutex);
        lock.lock_file_if_needed(LOCK_1_FILE);
        ofstream f(mod.c_str(),ios::app);
        f << now_time << " " << log_get_log_prefix() << log.str() << endl;
        f.flush();
        f.close();
    }

    if(!dir_module_logs_.compare(dir_global_logs_) && got && (scope & LOG_SCOPE_GLOBAL)) {
        // avoid doubled messages
        scope&=~LOG_SCOPE_GLOBAL;
    }

    if(log_glob && (scope & LOG_SCOPE_GLOBAL)) {
        got=true;
        string g=dir_global_logs_;
        g+=file;

        c_mutex_locker lock(log_files_mutex);
        lock.lock_file_if_needed(LOCK_1_FILE);
        ofstream f(g.c_str(),ios::app);
        f << now_time << " " << log_get_log_prefix() << log.str() << endl;
        f.flush();
        f.close();
    }

    if(got) {
        switch(type) {
            case E_LOG_DEBUG:
                if(glob_conf && glob_conf->log_echo_debug() || log_anyway || mod_conf && mod_conf->log_echo_debug()) {
                    cout << "DEBUG   " << log.str() << endl;
                    cout.flush();
                }
                logic_on_internal_event("@log@","debug","","","",PRIVATE_EVENT_SEVERITY_CODE_INFORMATIONAL,log.str(),"");
                break;
            case E_LOG_CHAN:
                if(glob_conf && glob_conf->log_echo_irc() || log_anyway || mod_conf && mod_conf->log_echo_irc()) {
                    cout << "CHANNEL " << log.str() << endl;
                    cout.flush();
                }
                logic_on_internal_event("@log@","channel",channel,"","",PRIVATE_EVENT_SEVERITY_CODE_INFORMATIONAL,log.str(),"");
                break;
            case E_LOG_BOT:
                if(glob_conf && glob_conf->log_echo_bot() || log_anyway || mod_conf && mod_conf->log_echo_bot()) {
                    cout << "BOT     " << log.str() << endl;
                    cout.flush();
                }
                logic_on_internal_event("@log@","bot","","","",PRIVATE_EVENT_SEVERITY_CODE_INFORMATIONAL,log.str(),"");
                break;
            case E_LOG_IDENTD:
                if(glob_conf && glob_conf->log_echo_identd() || log_anyway || mod_conf && mod_conf->log_echo_identd()) {
                    cout << "IDENTD  " << log.str() << endl;
                    cout.flush();
                }
                logic_on_internal_event("@log@","identd","","","",PRIVATE_EVENT_SEVERITY_CODE_INFORMATIONAL,log.str(),"");
                break;
            case E_LOG_SSL:
                if(glob_conf && glob_conf->log_echo_ssl() || log_anyway || mod_conf && mod_conf->log_echo_ssl()) {
                    cout << "SSL     " << log.str() << endl;
                    cout.flush();
                }
                logic_on_internal_event("@log@","ssl","","","",PRIVATE_EVENT_SEVERITY_CODE_INFORMATIONAL,log.str(),"");
                break;
            case E_LOG_IRC:
                if(glob_conf && glob_conf->log_echo_irc() || log_anyway || mod_conf && mod_conf->log_echo_irc()) {
                    cout << "IRC     " << log.str() << endl;
                    cout.flush();
                }
                logic_on_internal_event("@log@","irc","OUT","","",PRIVATE_EVENT_SEVERITY_CODE_INFORMATIONAL,log.str(),"");
                break;
            case E_LOG_CUSTOM:
            default:
                cout << "CUSTOM  " << log.str() << endl;
                cout.flush();
                logic_on_internal_event("@log@","custom",custom_name,"","",PRIVATE_EVENT_SEVERITY_CODE_INFORMATIONAL,log.str(),"");
                break;
        }
    }
}

void c_logger::loop()
{
    if(!mod_conf)
        return;
    int log_rot=stoi(mod_conf->getvar("log_rot"));
    if(log_rot!=1 && log_rot!=2)
        return;

    string log_rot_switch_time=mod_conf->getvar("log_rot_switch_time");
    if(log_rot_switch_time.empty())
        return;

    time_t now=static_cast<time_t>(my_time(NULL));
    tm* t=gmtime(&now);
    char tmp[128];
    strftime(tmp,sizeof(tmp)-1,"%H%M",t);

    if(!log_rot_switch_time.compare(tmp)) {
        if(was_rotation)
            return;
        was_rotation=true;

        log_print1(E_LOG_BOT,LOG_SCOPE_MODULE,"","Switching logs");

        util_debug_report_close(true);
        try {
            string dir=dir_module_logs_;
            if(dir.length()) {
                if(dir[dir.length()-1]!=FILE_SLASH_CH)
                    dir+=FILE_SLASH;

                long log_rot_keep_days=stol(mod_conf->getvar("log_rot_keep_days"));
                if(log_rot_keep_days) {
                    string del_dir=dir;
                    del_dir+="logs.";
                    del_dir+=tos(log_rot_keep_days);

                    vector<string> files;

#ifdef ROT_LOG_POSIX
                    DIR* dirp=opendir(del_dir.c_str());
                    while(dirp) {
                        errno=0;
                        dirent* dp=readdir(dirp);
                        if(dp!=NULL) {
                            if(!strcmp(dp->d_name,".") || !strcmp(dp->d_name,".."))
                                continue;
                            string name=del_dir;
                            name+=FILE_SLASH;
                            name+=dp->d_name;
                            errno=0;
                            struct stat s;
                            if(stat(name.c_str(),&s)==0) {
                                if(S_ISREG(s.st_mode)) {
                                    files.push_back(dp->d_name);
                                }
                            }
                            errno=0;
                        } else {
                            if(errno==0) {
                                closedir(dirp);
                                dirp=NULL;
                                break;
                            }
                            closedir(dirp);
                            dirp=NULL;
                            break;
                        }
                    }
#else
                    string find=del_dir;
                    find+=FILE_SLASH;
                    find+="*.*";
                    _finddata_t data;
#ifdef          __BORLANDC__
#define             intptr_t long
#endif
                    intptr_t hnd=_findfirst(const_cast<char*>(find.c_str()),&data);
                    if(hnd!=-1) {
                        int ret=0;
                        while(0==ret) {
                            if((data.attrib & _A_SUBDIR) || !strcmp(data.name,".") || !strcmp(data.name,"..")) {
                                // continue
                            } else {
                                files.push_back(data.name);
                            }

                            ret=_findnext(hnd,&data);
                        }
                        _findclose(hnd);
                    }
#endif
                    vector<string>::iterator ii;
                    for(ii=files.begin(); ii!=files.end(); ++ii) {
                        string del=del_dir;
                        del+=FILE_SLASH;
                        del+=*ii;
                        unlink(del.c_str());
                    }

                    rmdir(del_dir.c_str());
                }

                if(1==log_rot && log_rot_keep_days) {
                    for(int i1=log_rot_keep_days-1; i1>0; --i1) {
                        string old=dir;
                        old+="logs.";
                        old+=tos(i1);

                        string new_=dir;
                        new_+="logs.";
                        new_+=tos(i1+1);

                        rename(old.c_str(),new_.c_str());
                    }
                }

                string bak_dir=dir;
                if(2==log_rot) {
                    time_t now=static_cast<time_t>(my_time(NULL));
                    struct tm* t=gmtime(&now);
                    char tmp[64+1];
                    strftime(tmp,64,"%Y-%m-%d",t);
                    tmp[64]='\0';
                    bak_dir+=tmp;
                } else
                    bak_dir+="logs.1";

#if (defined(_WIN32) || defined(_WIN64)) && !defined(__CYGWIN__)
                mkdir(bak_dir.c_str());
#else
                mkdir(bak_dir.c_str(),0777);
#endif

                vector<string> files;

#ifdef ROT_LOG_POSIX
                DIR* dirp=opendir(dir.c_str());
                while(dirp) {
                    errno=0;
                    dirent* dp=readdir(dirp);
                    if(dp!=NULL) {
                        if(!strcmp(dp->d_name,".") || !strcmp(dp->d_name,".."))
                            continue;
                        if(!strcmp(dp->d_name,"irc_bot.pid")
                            || !strcmp(dp->d_name,"irc_bot.lock")
                            || !strcmp(dp->d_name,"irc_bot.lock1")
                            || !strcmp(dp->d_name,"irc_bot.lock2")
                            )
                            continue;
                        string name=dir;
                        name+=dp->d_name;
                        errno=0;
                        struct stat s;
                        if(stat(name.c_str(),&s)==0) {
                            if(S_ISREG(s.st_mode)) {
                                files.push_back(dp->d_name);
                            }
                        }
                        errno=0;
                    } else {
                        if(errno==0) {
                            closedir(dirp);
                            dirp=NULL;
                            break;
                        }
                        closedir(dirp);
                        dirp=NULL;
                        break;
                    }
                }
#else
                string find=dir;
                find+="*.*";
                _finddata_t data;
                intptr_t hnd=_findfirst(const_cast<char*>(find.c_str()),&data);
                if(hnd!=-1) {
                    int ret=0;
                    while(0==ret) {
                        if((data.attrib & _A_SUBDIR) || !strcmp(data.name,".") || !strcmp(data.name,"..")) {
                            // continue
                        } else {
                            if(!strcmp(data.name,"irc_bot.pid")
                                || !strcmp(data.name,"irc_bot.lock")
                                || !strcmp(data.name,"irc_bot.lock1")
                                || !strcmp(data.name,"irc_bot.lock2")
                                || !strcmp(data.name,"irc_bot.action")
                                ) {
                                //continue
                            } else
                                files.push_back(data.name);
                        }

                        ret=_findnext(hnd,&data);
                    }
                    _findclose(hnd);
                }
#endif

                vector<string>::iterator ii;
                for(ii=files.begin(); ii!=files.end(); ++ii) {
                    string old=dir;
                    old+=*ii;

                    string new_=bak_dir;
                    new_+=FILE_SLASH;
                    new_+=*ii;

                    rename(old.c_str(),new_.c_str());
                }
            }

            log_print1(E_LOG_BOT,LOG_SCOPE_MODULE,"","Logs were switched");
            util_debug_report_close(false);
        } catch(int UNUSEDARG(x)) {
            util_debug_report_close(false);
            return;
        }
    } else {
        was_rotation=false;
        return;
    }
}

/*!
    \brief Writes log string to file logs/clients.log and to stdout if log_echo_clients == true
    \author VooDooMan
    \version 1
    \date 2006-06-17
    \param id Client ID
    \param msg The log message
    \param log_anyway If to log even if it is disabled
*/
void log_clients(const log_client_id_t id, const string& msg, const bool log_anyway)
{
    log_check_semaphore_timeout();
    MY_U_LONG_LONG event_id=log_get_next_event_id();

    if(!log_initialized)
        init_echoes();

    c_conf* core_conf=get_core_conf();

    ostringstream log;
    log << "Client #" << id << ": " << msg;

    if(log_anyway || (core_conf && core_conf->log_log_clients())) {
        c_mutex_locker lock(log_files_mutex);
        lock.lock_file_if_needed(LOCK_1_FILE);
        ofstream f("logs" FILE_SLASH "clients.log",ios::app);
        f << make_nice_time(my_time(NULL),true) << " " << log_get_log_prefix() << ":" << log.str() << endl;
        f.flush();
        f.close();
    }

    log_cumulative(event_id,E_LOG_CLIENTS,"",log.str());

    if(log_anyway || (core_conf && core_conf->log_echo_clients())) {
        cout << log.str() << endl;
        cout.flush();
    }

    logic_on_internal_event("@log@","clients",tos(id),"","",PRIVATE_EVENT_SEVERITY_CODE_INFORMATIONAL,log.str(),"");
}

/*!
    \brief Opens lock file
    \version 2
    \author VooDooMan
    \retval true Okay
    \retval false Antoher instance is running
    \param probe Only test it
*/
bool log_lock(bool probe)
{
    if(protect_pid)
        return true;

    int posix_lock_file=-1;
    FILE* pid=NULL;

#if !defined(_WIN32) && !defined(_WIN64)
    int fd=open("logs" FILE_SLASH "irc_bot.lock",O_RDWR | O_CREAT,0666);
    if(fd<0) {
        fd=-1;
        char* tmp=MY_NEW char[strlen(strerror(errno))+1024];
        sprintf(tmp,"Error while open() #1 - LOCK file: %s",strerror(errno));
        log_bot(tmp);
        MY_DELETE_A(tmp);
    } else {
        struct flock lck;
        lck.l_type=F_RDLCK;
        lck.l_start=0;
        lck.l_whence=SEEK_SET;
        lck.l_len=0;
        lck.l_pid=0;
        int ec=fcntl(fd,F_GETLK,&lck);
        if(ec) {
            char* tmp=MY_NEW char[strlen(strerror(errno))+1024];
            sprintf(tmp,"Error while fcntl() - getting lock parameters on LOCK file: %s",strerror(errno));
            log_bot(tmp);
            MY_DELETE_A(tmp);
        }
        if(ec==0 && lck.l_type!=F_UNLCK) {
            close(fd);
            return false;
        }

        char tmp[1024];
        sprintf(tmp,"%s\n",tos(my_getpid()).c_str());
        write(fd,tmp,strlen(tmp));
        close(fd);
        fd=open("logs" FILE_SLASH "irc_bot.lock",O_RDWR,0666);
        if(fd<0) {
            fd=-1;
            char* tmp=MY_NEW char[strlen(strerror(errno))+1024];
            sprintf(tmp,"Error while open() #2 - LOCK file: %s",strerror(errno));
            log_bot(tmp);
            MY_DELETE_A(tmp);
        } else {
            struct flock lck;
            lck.l_type=F_WRLCK;
            lck.l_start=strlen(tmp)-1;
            lck.l_whence=SEEK_SET;
            lck.l_len=1;
            lck.l_pid=getpid();
            int ec=fcntl(fd,F_SETLK,&lck);
            if(ec) {
                char* tmp=MY_NEW char[strlen(strerror(errno))+1024];
                sprintf(tmp,"Error while fcntl() - setting lock parameters on LOCK file: %s",strerror(errno));
                log_bot(tmp);
                MY_DELETE_A(tmp);
            }
        }
    }
    if(probe)
        close(fd);
    else
        posix_lock_file=fd;
#else
    // erase test (works only on Win32!)
    int ec=unlink("logs" FILE_SLASH "irc_bot.pid");
    if(ec!=0 && errno==EACCES)
        return false;
#endif

    if(probe)
        return true;

    FILE* f=fopen("logs" FILE_SLASH "irc_bot.pid","w");
    if(f) {
        fprintf(f,"%s\n",tos(my_getpid()).c_str());
        fclose(f);
        pid=fopen("logs" FILE_SLASH "irc_bot.pid","r");
    }

    s_lock_releaser.assign(posix_lock_file,pid);

    return true;
}

/*!
    \brief Writes line to file logs/irc.log and to stdout if log_echo_irc == true
    \author VooDooMan
    \version 2
    \date 2004, 2007-07-27
    \param type Type of event
    \param msg The log message
    \param log_anyway If to log even if it is disabled
*/
void log_irc_ex(const e_log_irc_ex_type type, const string& msg, const bool log_anyway)
{
    log_check_semaphore_timeout();
    MY_U_LONG_LONG event_id=log_get_next_event_id();

    static const char* type_strings[] = {
        "IN ",
        "OUT",
        "INF",
    };

    const char* type_str=NULL;
    switch(type) {
        case E_LOG_IRC_EX_IN:
            type_str=type_strings[0];
            break;
        case E_LOG_IRC_EX_OUT:
            type_str=type_strings[1];
            break;
        case E_LOG_IRC_EX_INF:
            type_str=type_strings[2];
            break;
        default:
            MY_ASSERT_NOTREACH();
            return;
    }

    ostringstream log;
    log << type_str << " :" << msg;

    log_cumulative(event_id,E_LOG_IRC,"",log.str());

    if(!log_initialized)
        init_echoes();

    c_conf* core_conf=get_core_conf();

    if(log_anyway || (core_conf && core_conf->log_log_irc())) {
        c_mutex_locker lock(log_files_mutex);
        lock.lock_file_if_needed(LOCK_1_FILE);
        ofstream f("logs" FILE_SLASH "irc.log",ios::app);
        f << make_nice_time(my_time(NULL),true) << " " << log_get_log_prefix() << ":" << log.str() << endl;
        f.flush();
        f.close();
    }

    if(core_conf && core_conf->log_echo_irc()) {
        cout << "IRC     " << log.str() << endl;
        cout.flush();
    }

    string flags2;
    switch(type) {
        case E_LOG_IRC_EX_IN:
            flags2="IN";
            break;
        case E_LOG_IRC_EX_OUT:
            flags2="OUT";
            break;
        case E_LOG_IRC_EX_INF:
            flags2="INF";
            break;
        default:
            MY_ASSERT_NOTREACH();
            return;
    }

    logic_on_internal_event("@log@","irc",flags2,"","",PRIVATE_EVENT_SEVERITY_CODE_INFORMATIONAL,log.str(),"");
}

/*!
    \brief Logs message to (almost) all logs
    \version 1
    \author VooDooMan
    \date 2007-07-31
    \param msg Log message
    \param log_anyway If to log even if it is disabled
*/
void log_all_logs(const string& msg, const bool log_anyway)
{
    MY_RPT0(MY_RPT_INFO,msg.c_str());
    log_debug(msg,log_anyway);
    log_irc_ex(E_LOG_IRC_EX_INF,msg,log_anyway);
    log_socket(0,"",msg,NULL,log_anyway);
    log_bot(msg,log_anyway);
    log_botnet_debug("","",-1,-1,msg,log_anyway);
    log_botnet("","",msg,log_anyway);
    log_ssl(msg,log_anyway);
    log_identd(msg,log_anyway);
    log_clients(0,msg,log_anyway);
}

void log_procetct_pid_file()
{
    protect_pid=true;
}

void log_wait_for_releasing_lock()
{
    for(;;) {

        enum
        {
            E_WAIT = 1,
            E_EXIT,
            E_ERROR,
        };

        try {
#if !defined(_WIN32) && !defined(_WIN64)
            int fd=open("logs" FILE_SLASH "irc_bot.lock",O_RDWR | O_CREAT,0666);
            if(fd<0) {
                throw (int)E_ERROR;
            } else {
                struct flock lck;
                lck.l_type=F_RDLCK;
                lck.l_start=0;
                lck.l_whence=SEEK_SET;
                lck.l_len=0;
                lck.l_pid=0;
                int ec=fcntl(fd,F_GETLK,&lck);
                if(ec==0 && lck.l_type!=F_UNLCK) {
                    close(fd);
                    throw (int)E_WAIT;
                }

                // okay, it is not locked
                close(fd);
                unlink("logs" FILE_SLASH "irc_bot.lock");
                throw (int)E_EXIT;
            }
#else
            // erase test (works only on Win32!)
            int ec=unlink("logs" FILE_SLASH "irc_bot.pid");
            if(ec!=0 && errno==EACCES)
                throw (int)E_WAIT;

            throw (int)E_ERROR;
#endif
        } catch(int action) {
            switch(action) {
                case E_ERROR:
                    // fall thru
                case E_EXIT:
                    return;
                case E_WAIT:
                    // fall thru to wait
                    break;
            }
        }
        // wait
#if !defined(_WIN32) && !defined(_WIN64) && !defined(__CYGWIN__)
        timeval t;
        t.tv_sec=0;
        t.tv_usec=500*1000; // 500 ms
        select(1,NULL,NULL,NULL,&t);
#else
        Sleep((DWORD)500); // 500 ms
#endif
        // continue the loop
    }

    // NOTREACH
}

void log_can_overwrite_pid_file()
{
    overwrite_pid=true;
}

void log_supress_logging()
{
    log_suppress_any_logs=true;
}

/*!
    \brief Retrieves queued errors from IPC locking objects and then clears error queue
    \version 1
    \author VooDooMan
    \date 2008-06-06
*/
string log_get_ipc_mutex_error_log()
{
    string res;
    res=log_files_mutex.get_error_log();
    res+=log_cumul_file_mutex.get_error_log();

    return res;
}

