/***************************************************************************
                          log.h  -  description
                             -------------------
    begin                : Thu Dec 9 2004
    copyright            : (C) 2004-2007 by Marian VooDooMan Meravy
    email                : ghostvoodooman -a.T= users (dot) sourceforge.net
 ***************************************************************************/

/* $Id: log.h,v 1.53 2008/06/08 08:43:13 ghostvoodooman Exp $ */

/***************************************************************************

VooDoo cIRCle - an IRC (ro)bot
Copyright (C) 2004-2008 by Marian VooDooMan Meravy
  (ghostvoodooman -a.T= users (dot) sourceforge.net)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************/

/*!
    \file
    \brief Handles log files
    \version $Id: log.h,v 1.53 2008/06/08 08:43:13 ghostvoodooman Exp $
*/

#ifndef LOG_H__
#define LOG_H__

#if defined(HAVE_CONFIG_H)
#   include <config.h>
#endif

#include <string>
#include <vector>
#include <stack>
#include <list>
#include <map>

#include <stdarg.h>

#include "params.h"
#include "compat.h"
#include "utils.h"

#include "log2.h"
#include "internal_state.h"

#include "utils/assert.h"

using namespace std;

struct c_socket;

void log_debug(const string& msg, const bool log_anyway = false);
void log_irc(const bool from_server, const string& msg, const bool log_anyway = false);
void log_socket(const int sock_error, const string& os_msg, const string& msg, const c_socket* socket, const bool log_anyway = false);
void log_bot(const string& msg, const bool log_anyway = false);
void log_identd(const string& msg, const bool log_anyway = false);
void log_channel(const string& channel, const int type, const string& who, const string& whom, const string& msg, const string& user_modes);
void log_botnet_debug(const string& myname, const string& remote_name, const int packet_type, const int command, const string& msg, const bool log_anyway = false);
void log_botnet(const string& myname, const string& remote_name, const string& msg, const bool log_anyway = false);
void log_ssl(const string& msg, const bool log_anyway = false);
void log_broadcast(const string& bcast_mask, const int type, const string& who, const string& msg, const string& server, const unsigned short server_port);
void log_clients(const log_client_id_t id, const string& msg, const bool log_anyway = false);

enum e_log_irc_ex_type
{
    E_LOG_IRC_EX_IN,    //!< Line received from an IRC server
    E_LOG_IRC_EX_OUT,   //!< Line sent to an IRC server
    E_LOG_IRC_EX_INF,   //!< Informative line that will be written to log file/stdout
};                      //!< Type of line that will be logged/written to stdout

void log_irc_ex(const e_log_irc_ex_type type, const string& msg, const bool log_anyway = false);

bool log_init();
string log_done(int exit_code, string eol);

/*!
    \def TYPE_PRIVMSG
    \brief Type: PRIVMSG
*/
#define TYPE_PRIVMSG    1               //!< Type: PRIVMSG
/*!
    \def TYPE_NOTICE
    \brief Type: NOTICE
*/
#define TYPE_NOTICE     2               //!< Type: NOTICE
/*!
    \def TYPE_JOIN
    \brief Type: JOIN
*/
#define TYPE_JOIN       3               //!< Type: JOIN
/*!
    \def TYPE_PART
    \brief Type: PART
*/
#define TYPE_PART       4               //!< Type: PART
/*!
    \def TYPE_QUIT
    \brief Type: QUIT
*/
#define TYPE_QUIT       5               //!< Type: QUIT
/*!
    \def TYPE_KICK
    \brief Type: KICK
*/
#define TYPE_KICK       6               //!< Type: KICK
/*!
    \def TYPE_MODE
    \brief Type: MODE
*/
#define TYPE_MODE       7               //!< Type: MODE
/*!
    \def TYPE_NICK
    \brief Type: NICK
*/
#define TYPE_NICK       8               //!< Type: NICK
/*!
    \def TYPE_SCRIPT
    \brief Type: Message from script
*/
#define TYPE_SCRIPT     9               //!< Type: Message from script
/*!
    \def TYPE_TOPIC
    \brief Type TOPIC (not for flood!)
*/
#define TYPE_TOPIC      10              //!< Type TOPIC (not for flood!)
/*!
    \def TYPE_QUIT_PART_KICK
    \brief QUIT/PART/been KICK-ed (only for irc_check_flood())
*/
#define TYPE_QUIT_PART_KICK     128     //!< Type: QUIT/PART/been KICK-ed (only for irc_check_flood())
/*!
    \def TYPE_REPEAT
    \brief Type: repeat (only for irc_check_flood())
*/
#define TYPE_REPEAT             129     //!< Type: repeat (only for irc_check_flood())
/*!
    \def TYPE_CTCP
    \brief Type: CTCP (only for irc_check_flood())
*/
#define TYPE_CTCP               130     //!< Type: CTCP (only for irc_check_flood())
/*!
    \def TYPE_IRCOP_JOIN
    \brief Type: For logging, that user is an IRC operator (has joined)
*/
#define TYPE_IRCOP_JOIN         131     //!< Type: TYPE_IRCOP_JOIN
/*!
    \def TYPE_IRCOP_LEFT
    \brief Type: For logging, that user is an IRC operator (has quit, part, or has been kicked = wow! ;-)
*/
#define TYPE_IRCOP_LEFT         132     //!< Type: TYPE_IRCOP_LEFT

void log_free_locks();
bool log_lock(bool probe);

void log_all_logs(const string& msg, const bool log_anyway = false);

void log_procetct_pid_file();
void log_wait_for_releasing_lock();
void log_can_overwrite_pid_file();
void log_supress_logging();

string log_get_ipc_mutex_error_log();

#endif
