/***************************************************************************
                          c_logger.h  -  description
                             -------------------
    begin                : Thu May 10 2004
    copyright            : (C) 2004-2007 by Marian VooDooMan Meravy
    email                : ghostvoodooman -a.T= users (dot) sourceforge.net
 ***************************************************************************/

/* $Id: c_logger.h,v 1.14 2008/03/11 21:11:38 ghostvoodooman Exp $ */

/***************************************************************************

VooDoo cIRCle - an IRC (ro)bot
Copyright (C) 2004-2008 by Marian VooDooMan Meravy
  (ghostvoodooman -a.T= users (dot) sourceforge.net)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************/

/*!
    \file
    \brief Class 'c_logger'
    \version $Id: c_logger.h,v 1.14 2008/03/11 21:11:38 ghostvoodooman Exp $
*/

#ifndef C_LOGGER_H__
#define C_LOGGER_H__

#if defined(HAVE_CONFIG_H)
#   include <config.h>
#endif

#include <string>
#include <vector>
#include <stack>
#include <list>
#include <map>
#include <algorithm>

#include "../utils/memleak.h"
#include "../params.h"

#include "../conf/c_conf.h"
#include "../log2.h"

class c_conf;

/*!
    \brief Logger
    \author VooDooMan
    \version 1
    \date 2005
*/
class c_logger
{
    string dir_global_logs_,            //!< Directory for core module logs
        dir_module_logs_;               //!< Directory for module's logs
    string custom_name;                 //!< Name of custom log
    int default_scope_;                 //!< Flags for default scope: LOG_SCOPE_xxx (OR-ed)
    c_conf *glob_conf;                  //!< Core module's config
    c_conf *mod_conf;                   //!< Module's config

    bool was_rotation;                  //!< true if there was log rotation time reached

public:
    /*!
        \brief Constructor
        \author VooDooMan
        \version 1
        \date 2005
        \param dir_global_logs Directory for core module logs
        \param dir_module_logs Directory for module's logs
        \param custom_log_name Name of custom log
        \param default_scope Flags for default scope: LOG_SCOPE_xxx (OR-ed)
        \param global_conf Core module's config
        \param module_conf Module's config
    */
    c_logger(const string& dir_global_logs, const string& dir_module_logs, const string& custom_log_name, int default_scope, c_conf* global_conf, c_conf* module_conf);

    c_logger()
        : glob_conf(NULL), mod_conf(NULL), was_rotation(false)
    { ; }

    c_logger(const c_logger& orig);

    /*!
        \brief Re-inits
        \author VooDooMan
        \version 1
        \date 2005
        \param dir_global_logs Directory for core module logs
        \param dir_module_logs Directory for module's logs
        \param custom_log_name Name of custom log
        \param default_scope Flags for default scope: LOG_SCOPE_xxx (OR-ed)
        \param global_conf Core module's config
        \param module_conf Module's config
    */
    void init(const string& dir_global_logs, const string& dir_module_logs, const string& custom_log_name, int default_scope, c_conf* global_conf, c_conf* module_conf);

    /*!
        \brief Replaces some characters in channel name that are not suitable to store as file name
        \author VooDooMan
        \version 1
        \date 2005
        \param channel Channel name
        \return Returns name of file
    */
    static string get_channel_log_name(string channel);

/*!
    \def LOG_SCOPE_MODULE
    \brief OR flag for module's log
*/
#define LOG_SCOPE_MODULE    1L
/*!
    \def LOG_SCOPE_MODULE
    \brief OR flag for core module's log
*/
#define LOG_SCOPE_GLOBAL    2L
/*!
    \def LOG_SCOPE_MODULE
    \brief OR flag for default scope of log
*/
#define LOG_SCOPE_DEFAULT   4L

    /*!
        \brief Prints message to log
        \author VooDooMan
        \version 1
        \date 2005
        \param type Type of log
        \param scope OR-ed flags for scope (LOG_SCOPE_xxx)
        \param channel Channel name, if applicable
        \param msg Message
        \param log_anyway Set to true if to log even if it is disabled
    */
    void log_print1(e_log_type type, int scope, const string& channel, const string& msg, bool log_anyway = false);

    /*!
        \brief Checks for log rotation, and performs, if necessary (should be called in loop)
        \author VooDooMan
        \version 1
        \date 2006-01-03
    */
    void loop();
};

#endif

