/***************************************************************************
                          logic.cpp  -  description
                             -------------------
    begin                : Thu Dec 9 2004
    copyright            : (C) 2004-2007 by Marian VooDooMan Meravy
    email                : ghostvoodooman -a.T= users (dot) sourceforge.net
 ***************************************************************************/

/* $Id: logic.cpp,v 1.107 2008/08/23 15:10:40 ghostvoodooman Exp $ */

/***************************************************************************

VooDoo cIRCle - an IRC (ro)bot
Copyright (C) 2004-2008 by Marian VooDooMan Meravy
  (ghostvoodooman -a.T= users (dot) sourceforge.net)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
a    with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************/

/*!
    \file
    \brief Implements core logic stored in "logic.txt" file
    \version $Id: logic.cpp,v 1.107 2008/08/23 15:10:40 ghostvoodooman Exp $
*/

#if defined(HAVE_CONFIG_H)
#   include <config.h>
#endif

#include <string>
#include <vector>
#include <stack>
#include <list>
#include <map>
#include <algorithm>
#include <iostream>

#include <stdlib.h>
#include <stdio.h>
#include <string>
#if defined(_WIN32) || defined(_WIN64)
#   include <conio.h>
#   include <io.h>
#else
#   include <unistd.h>
#endif
#include <sys/stat.h>
#ifndef _MSC_VER
#if defined(_WIN32) || defined(_WIN64)
#       include <dir.h>
#   else
#       include <sys/types.h>
#       include <sys/stat.h>
#   endif
#else
#    include <direct.h>
#    define mkdir _mkdir
#endif

#include "utils/memleak.h"
#include "utils/assert.h"
#include "utils.h"
#include "compat.h"
#include "log.h"
#include "irc.h"
#include "sock.h"
#include "identd.h"
#include "dcc.h"
#include "lang.h"
#include "conf.h"
#include "exec.h"
#include "botnet.h"
#include "stats.h"
#include "filesys.h"
#include "compat.h"
#include "variable.h"
#include "proc.h"
#include "modules.h"
#include "threads.h"
#include "match.h"
#include "eval.h"
#include "params.h"
#include "match.h"
#include "logic.h"
#include "var/c_var_base.h"
#include "var/c_var_bool.h"
#include "var/c_var_int.h"
#include "var/c_var_string.h"
#include "var/c_var_array.h"
#include "proc/c_proc_exc.h"
#include "irc/c_online_channel.h"
#include "irc/c_online_user.h"
#include "botnet/c_replication.h"
#include "logic/logic_internal.h"
#include "misc/shutdown.h"

#include "utils/precomp.h"

#if defined(_MSC_VER) || defined(__BORLANDC__)
#   pragma hdrstop
#endif

//---------------------------------------------------------------------------

#include "shared.h"

using namespace std;

deque<c_user> r_users;                             //!< List of users for real usage (not temporary / compilation!)
deque<c_user> r_user_templates;                    //!< List of users templates for real usage (not temporary / compilation!)
deque<c_user> r_terminators;                       //!< List of terminated users for real usage (not temporary / compilation!)
vector<c_channel> r_channels;                       //!< List of channels for real usage (not temporary / compilation!)
vector<c_channel> r_channel_templates;              //!< List of channel templates for real usage (not temporary / compilation!)
vector<c_channel> r_channel_terminators;            //!< List of channel terminators for real usage (not temporary / compilation!)
vector<c_channel_def> r_channel_defs;               //!< List of real channel definitions (not temporary / compilation!)

vector<string> r_all_groups;                        //!< List of all groups for real usage (not temporary / compilation!)

c_private r_private;                                //!< Informations about procedures when someone sends query message (not temporary / compilation!)

int last_nick_idx;                                  //!< Contains index of last selected nick for bot, from logic.txt

string ident_string;                                //!< Contains IDENTD response for IDENT daemon

my_time_t last_autobackup=0;                           //!< Time of last automatic backup

bool was_whois_burst;                               //!< "true" if there was WHOIS burst
bool into_whois_burst;                              //!< "true" if we are into WHOIS burst

c_var_array* super_global;                          //!< Global variables for all modules, containing status of the bot

map<int,string> severity_codes;                     //!< Associative array to convers severity code to text representation

string event_id;                                    //!< Event ID

vector<c_tmp_file> tmp_files_garbage;               //!< There are tmp files (running scripts) after module unload, to be discarded

/*!
    \brief Checks number of arguments in call string
    \author VooDooMan
    \version 1
    \date 2005-12-03
    \param proc Call string
    \param mandatory Number of mandatory arguments
    \return Returns true if okay
*/
bool logic_check_num_agrs(string proc, const size_t mandatory)
{
    //int res=0;
    if(proc.find("(",0)!=string::npos)
        proc.erase(0,proc.find("(",0)+1);
    else
        return false;

    if(!proc.length())
        return false;

    if(proc[proc.length()-1]!=')')
        return false;
    else
        proc.erase(proc.length()-1,1);

    if(!proc.length())
        return false;

    size_t args=1;

    bool in_str=false;
    //bool in_esc=false;
    bool in_arg=false;
    for(size_t i1=0; i1<proc.length(); ++i1) {
        if('"'==proc[i1]) {
            if(in_arg)
                return false;
            in_str=!in_str;
            if(in_str && args<mandatory)
                return false;
            continue;
        }
        if(in_str)
            continue;
        if(0x20==proc[i1])
            return false;
        if('$'==proc[i1]) {
            if(in_arg)
                return false;
            in_arg=true;
            continue;
        }
        if(','==proc[i1]) {
            in_arg=false;
            args++;
        }
    }
    if(args<mandatory)
        return false;

    return true;
}

/*!
    \brief Executes procedure/function in core module
    \author VooDooMan
    \version 1
    \param proc Name of procedure/function
    \param args String arguments
    \param debug Debug string for case of error
    \param num_mandatory_args Number of mandatory arguments
    \param eid Name of event
*/
void logic_core_execute(const string& proc, vector<string>& args, const string& debug, size_t num_mandatory_args, const string& eid)
{
    if(proc.empty() || !proc.compare(" "))
        return;
    if(!logic_check_num_agrs(proc,num_mandatory_args)) {
        string log="OOPS: Number of mandatory arguments mismatch, or invalid call string, also, mandatory arguments cannot be overridden by immediates. Call string was: \"";
        log+=proc;
        log+="\"";
        log_debug(log.c_str());
        return;
    }
    event_id=eid;

    eval_cleanup_timeout();

    c_module* core=get_module("core",true);
    if(core)
        core->execute(proc,args,debug);

    event_id="";
}

/*!
    \brief Executes procedure/function in core module and cleans-up argumets
    \author VooDooMan
    \version 1
    \param proc Name of procedure/function
    \param args Arguments, will be MY_DELETE'd
    \param debug Debug string for case of error
    \param num_mandatory_args Number of mandatory arguments
    \param eid Name of event
*/
void logic_core_execute(const string& proc, vector<c_var_base*>& args, const string& debug, size_t num_mandatory_args, const string& eid)
{
    if(proc.empty() || !proc.compare(" "))
        return;
    if(!logic_check_num_agrs(proc,num_mandatory_args)) {
        string log="OOPS: Number of mandatory arguments mismatch, or invalid call string, also, mandatory arguments cannot be overridden by immediates. Call string was: \"";
        log+=proc;
        log+="\"";
        log_debug(log.c_str());
        return;
    }

    event_id=eid;

    eval_cleanup_timeout();
    c_module* core=get_module("core",true);
    if(core)
        core->execute(proc,args,debug);
    args.clear();

    event_id="";
}

/*!
    \brief Looks up and evalueates variable beginning with '$' sign, if it is possible
    \author VooDooMan
    \version 3
    \date 2004, 2005-10-06, 2007-11-05
    \param str Name of variable beginning with '$' sign
    \param vars Array of variables ant their values (can be NULL)
    \return Returns evaluated variable, if it wasn't found in vars array, it returns str value
    \param this_module Module
    \param debug Debug string
*/
string logic_eval(c_module* this_module, const string& str, c_var_array* vars, const string& debug)
{
    return var_eval_string_simple(this_module,str,vars,debug);
}

/*!
    \brief Evaluates string for hashing logic.txt
    \author VooDooMan
    \version 1
    \date 2007-05-02
    \param str String
    \return Returns evaluated variable, if it wasn't found in vars array, it returns str value
    \param this_module Module
    \param debug Debug string
*/
string logic_eval_str(c_module* this_module, const string& str, const string& debug)
{
    c_logger* logger=NULL;
    if(!this_module) {
        logger=get_core_logger();
    } else
        logger=this_module->get_logger();

    string log;

    if(str.length()==0)
        return (string)"";
    if(str[0]>='0' && str[0]<='9') {
        long tmp=stol(str);
        return tos(tmp);
    }
    if(str.length()<2) {
        if(logger) {
            log="OOPS: Bogus value; ";
            log+=debug;
            logger->log_print1(E_LOG_DEBUG,LOG_SCOPE_MODULE,"",log);
            logger->log_print1(E_LOG_BOT,LOG_SCOPE_MODULE,"",log);
        }
        return str;
    }
    if(str[0]!='"' || str[str.length()-1]!='"') {
        if(logger) {
            log="OOPS: Bogus value; ";
            log+=debug;
            logger->log_print1(E_LOG_DEBUG,LOG_SCOPE_MODULE,"",log);
            logger->log_print1(E_LOG_BOT,LOG_SCOPE_MODULE,"",log);
        }
        return str;
    }

    string wo_quotes=str;
    wo_quotes.erase(0,1);
    wo_quotes.erase(wo_quotes.length()-1,1);

    string res;

    bool in_esc=false;
    string esc;

    for(size_t i1=1; i1<str.length(); ++i1) { // ignore leading '"'
again:
        if(str.length()-1==i1) {
            if(!in_esc)
                break;
        }
        if(!in_esc) {
            if(str[i1]=='\\') {
                in_esc=true;
                esc="";
                continue;
            }
            res+=(string)""+str[i1];
            continue;
        } else {
            if(esc.length()==0) {
                if(str.length()-1==i1) {
                    if(logger) {
                        log="OOPS: Bogus value; ";
                        log+=debug;
                        logger->log_print1(E_LOG_DEBUG,LOG_SCOPE_MODULE,"",log);
                        logger->log_print1(E_LOG_BOT,LOG_SCOPE_MODULE,"",log);
                    }
                    return wo_quotes;
                }
                if('\\'==str[i1]) {
                    res+="\\";
                    in_esc=false;
                    continue;
                }
                if('x'==str[i1]) {
                    esc="x";
                    continue;
                }
                if(str[i1]>='0' && str[i1]<='9') {
                    esc+=(string)""+str[i1];
                    continue;
                }
                switch(str[i1]) {
                    case '"':
                        res+="\"";
                        in_esc=false;
                        continue;
                    case '\'':
                        res+="'";
                        in_esc=false;
                        continue;
                    case 'r':
                        res+="\r";
                        in_esc=false;
                        continue;
                    case 'n':
                        res+="\n";
                        in_esc=false;
                        continue;
                    case 't':
                        res+="\t";
                        in_esc=false;
                        continue;
                    case 'a':
                        res+="\x06";
                        in_esc=false;
                        continue;
                    case 'b':
                        res+="\b";
                        in_esc=false;
                        continue;
                    case 'f':
                        res+="\f";
                        in_esc=false;
                        continue;
                    case 'v':
                        res+="\v";
                        in_esc=false;
                        continue;
                    case '?':
                        res+="?";
                        in_esc=false;
                        continue;
                    case '$':
                        res+="$";
                        in_esc=false;
                        continue;
                }
                if(logger) {
                    log="OOPS: Bogus value; ";
                    log+=debug;
                    logger->log_print1(E_LOG_DEBUG,LOG_SCOPE_MODULE,"",log);
                    logger->log_print1(E_LOG_BOT,LOG_SCOPE_MODULE,"",log);
                }
                return wo_quotes;
            }
            // >0
            int x=0;
            if('x'==esc[0]) {
                if((str[i1]>='0' && str[i1]<='9') || (str[i1]>='a' && str[i1]<='f')
                    || (str[i1]>='A' && str[i1]<='F')) {
                        if(esc.length()!=3) {
                            esc+=(string)""+str[i1];
                            if(esc.length()<=3)
                                continue;
                        }
                    }
                // finish
                if(esc.length()==1 || esc.length()>3) {
                    if(logger) {
                        log="OOPS: Bogus value; ";
                        log+=debug;
                        logger->log_print1(E_LOG_DEBUG,LOG_SCOPE_MODULE,"",log);
                        logger->log_print1(E_LOG_BOT,LOG_SCOPE_MODULE,"",log);
                    }
                    return wo_quotes;
                }
                if(esc.length()==2)
                    esc[0]='0';
                else // length() == 3
                    esc.erase(0,1);
                esc=str_to_lower(esc);
                x=0;
                if(sscanf(esc.c_str(),"%2x",&x)!=1)
                    x=0;
            process:
                if(x<0 || x>255) {
                    if(logger) {
                        log="OOPS: Bogus value; ";
                        log+=debug;
                        logger->log_print1(E_LOG_DEBUG,LOG_SCOPE_MODULE,"",log);
                        logger->log_print1(E_LOG_BOT,LOG_SCOPE_MODULE,"",log);
                    }
                    return wo_quotes;
                }
                if(!x) {
                    // NULL terminator?! okay, bail out and return what we got so far
                    if(logger) {
                        log="OOPS: Bogus value; ";
                        log+=debug;
                        logger->log_print1(E_LOG_DEBUG,LOG_SCOPE_MODULE,"",log);
                        logger->log_print1(E_LOG_BOT,LOG_SCOPE_MODULE,"",log);
                    }
                    return res;
                }
                char xx=(char)((unsigned)x);
                res+=(string)""+xx;
                in_esc=false;
                goto again;
            }
            // octal
            if(str[i1]>='0' && str[i1]<='9' && esc.length()<3) {
                esc+=(string)""+str[i1];
                continue;
            }
            x=0;
            if(sscanf(esc.c_str(),"%o",&x)!=1)
                x=0;
            goto process;
        }
    }

    return res;
}

/*!
    \brief Fills-up "severity_codes" associative array
    \author VooDooMan
    \version 1
    \date 2005-07-08
*/
void logic_init_severities()
{
    severity_codes.clear();

    pair<int,string> p;

    p.first=PRIVATE_EVENT_SEVERITY_CODE_NA;
    p.second="N/A";
    severity_codes.insert(p);

    p.first=PRIVATE_EVENT_SEVERITY_CODE_INFORMATIONAL;
    p.second="informational";
    severity_codes.insert(p);

    p.first=PRIVATE_EVENT_SEVERITY_CODE_WARNING;
    p.second="warning";
    severity_codes.insert(p);

    p.first=PRIVATE_EVENT_SEVERITY_CODE_ERROR;
    p.second="error";
    severity_codes.insert(p);

    p.first=PRIVATE_EVENT_SEVERITY_CODE_CRITICAL_ERROR;
    p.second="critical error";
    severity_codes.insert(p);

    p.first=PRIVATE_EVENT_SEVERITY_CODE_FATAL_ERROR;
    p.second="fatal error";
    severity_codes.insert(p);
}

/*!
    \brief Gets text representation for severity code
    \author VooDooMan
    \version 1
    \date 2005-07-08
    \param code Severity code
    \return Returns text representation, or "?" string for unknown code
*/
string logic_get_severity_text(int code)
{
    if(severity_codes.empty())
        logic_init_severities();

    map<int,string>::iterator i;
    i=severity_codes.find(code);
    if(i==severity_codes.end())
        return (string)"?";
    return (*i).second;
}

/*!
    \brief Adds/remove new dynamic mode for user on channel, but only when group relationship and setting allows it
    \author VooDooMan
    \version 1
    \date 2004
    \param username Name of affected user
    \param from_user Name of origin user
    \param prefix '+' to set mode, '-' to clear mode
    \param mode Mode: '+' or '@'
*/
void logic_set_dynamic(const string& username, const string& UNUSEDARG(from_user), const char prefix, const char mode)
{
    vector<c_channel>::iterator i1;
    for(i1=r_channels.begin(); i1!=r_channels.end(); ++i1) {
        if(!(*i1).username.compare(username)) {
            vector<c_dynamic_rule>::iterator i2;
            for(i2=(*i1).allow_dynamic.begin(); i2!=(*i1).allow_dynamic.end(); ++i2) {
                deque<c_user>::iterator i3;
                for(i3=r_users.begin(); i3!=r_users.end(); ++i3) {
                    vector<string>::iterator i4;
                    for(i4=(*i3).groups.begin(); i4!=(*i3).groups.end(); ++i4) {
                        if(!(*i4).compare((*i2).group) && (*i2).plus_modes.find(mode,0)!=string::npos) {
                            string mp=(*i1).dynamic_plus_modes;
                            while(mp.find(mode,0)!=string::npos)
                                mp.erase(mp.find(mode,0),1);
                            if(prefix=='+')
                                mp+=mode;
                            (*i1).dynamic_plus_modes=mp;
                        }
                        if(!(*i4).compare((*i2).group) && (*i2).minus_modes.find(mode,0)!=string::npos) {
                            string mm=(*i1).dynamic_minus_modes;
                            while(mm.find(mode,0)!=string::npos)
                                mm.erase(mm.find(mode,0),1);
                            if(prefix=='-')
                                mm+=mode;
                            (*i1).dynamic_minus_modes=mm;
                        }
                        // do NOT break! cycle through all settings!
                    }
                }
            }
        }
    }
}

/*!
    \brief Gets user list to pull
    \author VooDooMan
    \version 1
    \date 2004
    \param botname Name of remote bot
    \param users Returns list of users to pull
*/
void logic_get_pull_users(const string& botname, deque<c_user>& users)
{
    users.clear();
    deque<c_user>::iterator i1;
    for(i1=r_users.begin(); i1!=r_users.end(); ++i1) {
        vector<c_replication>::iterator i2;
        for(i2=(*i1).replication.begin(); i2!=(*i1).replication.end(); ++i2) {
            if(!(*i2).partner.compare(botname) && ((*i2).type==REPL_PULL || (*i2).type==REPL_PUSHPULL)) {
                users.push_back(*i1);
            }
        }
    }
    for(i1=r_terminators.begin(); i1!=r_terminators.end(); ++i1) {
        vector<c_replication>::iterator i2;
        for(i2=(*i1).replication.begin(); i2!=(*i1).replication.end(); ++i2) {
            if(!(*i2).partner.compare(botname) && ((*i2).type==REPL_PULL || (*i2).type==REPL_PUSHPULL)) {
                users.push_back(*i1);
            }
        }
    }
    for(i1=r_user_templates.begin(); i1!=r_user_templates.end(); ++i1) {
        vector<c_replication>::iterator i2;
        for(i2=(*i1).replication.begin(); i2!=(*i1).replication.end(); ++i2) {
            if(!(*i2).partner.compare(botname) && ((*i2).type==REPL_PULL || (*i2).type==REPL_PUSHPULL)) {
                users.push_back(*i1);
            }
        }
    }
}

/*!
    \brief Gets user list to push
    \author VooDooMan
    \version 1
    \date 2004
    \param botname Name of remote bot
    \param users Returns list of users to push
*/
void logic_get_push_users(const string& botname, deque<c_user>& users)
{
    users.clear();
    deque<c_user>::iterator i1;
    for(i1=r_users.begin(); i1!=r_users.end(); ++i1) {
        vector<c_replication>::iterator i2;
        for(i2=(*i1).replication.begin(); i2!=(*i1).replication.end(); ++i2) {
            if(!(*i2).partner.compare(botname) && ((*i2).type==REPL_PUSH || (*i2).type==REPL_PUSHPULL)) {
                users.push_back(*i1);
            }
        }
    }
    for(i1=r_terminators.begin(); i1!=r_terminators.end(); ++i1) {
        vector<c_replication>::iterator i2;
        for(i2=(*i1).replication.begin(); i2!=(*i1).replication.end(); ++i2) {
            if(!(*i2).partner.compare(botname) && ((*i2).type==REPL_PUSH || (*i2).type==REPL_PUSHPULL)) {
                users.push_back(*i1);
            }
        }
    }
    for(i1=r_user_templates.begin(); i1!=r_user_templates.end(); ++i1) {
        vector<c_replication>::iterator i2;
        for(i2=(*i1).replication.begin(); i2!=(*i1).replication.end(); ++i2) {
            if(!(*i2).partner.compare(botname) && ((*i2).type==REPL_PUSH || (*i2).type==REPL_PUSHPULL)) {
                users.push_back(*i1);
            }
        }
    }
}

/*!
    \brief Gets procedure list to push
    \author VooDooMan
    \version 2
    \date 2004, 2005-10-05
    \param botname Name of remote bot
    \param procs Returns list of procedures to push
*/
void logic_get_push_procs(const string& botname, vector<c_procedure>& procs)
{
    procs.clear();
    c_module* core=get_module("core",true);
    if(core)
        core->get_replication_procs(procs,botname,REPL_PUSH);
}

/*!
    \brief Gets function list to push
    \author VooDooMan
    \version 1
    \date 2005-11-19
    \param botname Name of remote bot
    \param funcs Returns list of functions to push
*/
void logic_get_push_funcs(const string& botname, vector<c_function*>& funcs)
{
    funcs.clear();
    c_module* core=get_module("core",true);
    if(core)
        core->get_replication_funcs(funcs,botname,REPL_PUSH);
}

/*!
    \brief Gets channel definition list to pull
    \author VooDooMan
    \version 1
    \date 2004
    \param botname Name of remote bot
    \param chans Returns list of channels to pull
*/
void logic_get_pull_chan_defs(const string& botname, vector<c_channel_def>& chans)
{
    chans.clear();
    vector<c_channel_def>::iterator i1;
    for(i1=r_channel_defs.begin(); i1!=r_channel_defs.end(); ++i1) {
        vector<c_replication>::iterator i2;
        for(i2=(*i1).replication.begin(); i2!=(*i1).replication.end(); ++i2) {
            if(!(*i2).partner.compare(botname) && ((*i2).type==REPL_PULL || (*i2).type==REPL_PUSHPULL)) {
                chans.push_back(*i1);
            }
        }
    }
}

/*!
    \brief Gets channel definition list to push
    \author VooDooMan
    \version 1
    \date 2004
    \param botname Name of remote bot
    \param chans Returns list of channels to push
*/
void logic_get_push_chan_defs(const string& botname, vector<c_channel_def>& chans)
{
    chans.clear();
    vector<c_channel_def>::iterator i1;
    for(i1=r_channel_defs.begin(); i1!=r_channel_defs.end(); ++i1) {
        vector<c_replication>::iterator i2;
        for(i2=(*i1).replication.begin(); i2!=(*i1).replication.end(); ++i2) {
            if(!(*i2).partner.compare(botname) && ((*i2).type==REPL_PUSH || (*i2).type==REPL_PUSHPULL)) {
                chans.push_back(*i1);
            }
        }
    }
}

/*!
    \brief Get maximum nuber of logic.txt backups from conf.txt
    \author VooDooMan
    \version 1
    \date 2006-01-03
*/
int logic_get_max_logic_backups()
{
    long max=stol(core_conf_getvar("max_logic_backups"));
    if(max<=0 || max>1000)
        max=1000;
    return max;
}

/*!
    \brief Shifts backups of "logic.txt"
    \author VooDooMan
    \version 1
    \date 2005-10-19
    \param shift_all Set to true to shift all remaining
*/
void logic_shift_backup(const bool shift_all)
{
    {
        char tmp1[256], tmp2[256];
        static int i1=logic_get_max_logic_backups()-1;
        for( ; i1>0; --i1) {
            sprintf(tmp1,"%s%03d","logic.txt.bak",i1);
            sprintf(tmp2,"%s%03d","logic.txt.bak",i1-1);
            remove(tmp1);
            rename(tmp2,tmp1);
            if(1==i1) {
                remove(tmp2);
            }
            if(shift_all)
                continue;
            if(i1)
                i1--;
            break;
        }
        if(shift_all) {
            rename("logic.txt","logic.txt.bak000");
            i1=logic_get_max_logic_backups()-1;
        }
    }
}

/*!
    \brief Forces to writes all object to logic.txt file
    \author VooDooMan
    \version 2
    \date 2004, 2006-12-27
    \param who This should contain who caused rehasing (for log purposes - bot.log)
    \param this_module Module
*/
void logic_partyline_backup(c_module* this_module, string who)
{
    filesys_flush();

    string debug="Backup";

    e_shutdown_reason sr=bot_get_shutdown_state();
    if(E_SD_UPGRADE==sr) {
        string log2="*** UPGRADE state, ignoring backup request by "+who;
        log_bot(log2.c_str());

        logic_on_internal_event("@backup@",who,"","","",PRIVATE_EVENT_SEVERITY_CODE_WARNING,log2,"");
        return;
    }

    logic_on_internal_event("@backup@",who,"","","",PRIVATE_EVENT_SEVERITY_CODE_INFORMATIONAL,"","");

    who=(string)"*** Backup invoked by "+who;
    log_bot(who.c_str());
    logic_shift_backup(true);

    vector<c_replication>::iterator repl_i;

    FILE* f=fopen("logic.txt","w");
    if(!f)
        return;
    fprintf(f,"#version 3\n");
    int indent=0;
    deque<c_user>::iterator i1;
    string ln;
    int ii;
    int iii;
    for(iii=0; iii<3; ++iii) {
        deque<c_user>* users=NULL;
        if(iii==0)
            users=&r_users;
        else
            if(iii==1)
                users=&r_user_templates;
            else
                if(iii==2)
                    users=&r_terminators;
        for(i1=users->begin(); i1!=users->end(); ++i1) {
            if(!(*i1).is_template) {
                ln=""; for(ii=0; ii<indent; ++ii) ln+=" ";
                ln+=(string)"user \""+c_var_string::script_esc("logic.txt",(*i1).name)+"\"";
                fprintf(f,"%s\n",ln.c_str());
            } else {
                ln=""; for(ii=0; ii<indent; ++ii) ln+=" ";
                ln+=(string)"user_template \""+c_var_string::script_esc("logic.txt",(*i1).name)+"\"";
                fprintf(f,"%s\n",ln.c_str());
            }

            ln=""; for(ii=0; ii<indent; ++ii) ln+=" ";
            ln+="{";
            fprintf(f,"%s\n",ln.c_str());
            indent+=4;

            if((*i1).host_unknown) {
                ln=""; for(ii=0; ii<indent; ++ii) ln+=" ";
                ln+="host_unknown";
                fprintf(f,"%s\n",ln.c_str());
            }

            ln=""; for(ii=0; ii<indent; ++ii) ln+=" ";
            ln+="terminator ";
            fprintf(f,"%s%d\n",ln.c_str(),BOOL2INT((*i1).terminator));

            if((*i1).host_bot) {
                ln=""; for(ii=0; ii<indent; ++ii) ln+=" ";
                ln+="host_bot";
                fprintf(f,"%s\n",ln.c_str());
            }

            ln=""; for(ii=0; ii<indent; ++ii) ln+=" ";
            ln+="last_changed ";
            ln+=tos((*i1).last_changed);
            fprintf(f,"%s\n",ln.c_str());

            vector<string>::iterator i2;
            for(i2=(*i1).hostmask.begin(); i2!=(*i1).hostmask.end(); ++i2) {
                ln=""; for(ii=0; ii<indent; ++ii) ln+=" ";
                ln+=(string)"host \""+c_var_string::script_esc("logic.txt",*i2)+"\"";
                fprintf(f,"%s\n",ln.c_str());
            }

            for(i2=(*i1).fullname.begin(); i2!=(*i1).fullname.end(); ++i2) {
                ln=""; for(ii=0; ii<indent; ++ii) ln+=" ";
                ln+=(string)"full_name \""+c_var_string::script_esc("logic.txt",*i2)+"\"";
                fprintf(f,"%s\n",ln.c_str());
            }

            sort((*i1).replication.begin(),(*i1).replication.end());
            repl_i=unique((*i1).replication.begin(),(*i1).replication.end());
            (*i1).replication.erase(repl_i,(*i1).replication.end());

            for(vector<c_replication>::iterator i8=(*i1).replication.begin(); i8!=(*i1).replication.end(); ++i8) {
                ln=""; for(ii=0; ii<indent; ++ii) ln+=" ";
                string type;
                if((*i8).type==REPL_PUSH)
                    type="push";
                if((*i8).type==REPL_PULL)
                    type="pull";
                if((*i8).type==REPL_PUSHPULL)
                    type="pushpull";
                ln+=(string)"replication "+(*i8).partner+" "+type;
                fprintf(f,"%s\n",ln.c_str());
            }

            if((*i1).replication_partner) {
                ln=""; for(ii=0; ii<indent; ++ii) ln+=" ";
                ln+=(string)"replication_partner";
                fprintf(f,"%s\n",ln.c_str());
            }

            for(i2=(*i1).access_to_group.begin(); i2!=(*i1).access_to_group.end(); ++i2) {
                ln=""; for(ii=0; ii<indent; ++ii) ln+=" ";
                ln+=(string)"access_to_group \""+c_var_string::script_esc("logic.txt",*i2)+"\"";
                fprintf(f,"%s\n",ln.c_str());
            }

            for(i2=(*i1).access_usage_procedure.begin(); i2!=(*i1).access_usage_procedure.end(); ++i2) {
                ln=""; for(ii=0; ii<indent; ++ii) ln+=" ";
                ln+=(string)"access_usage_procedure \""+c_var_string::script_esc("logic.txt",*i2)+"\"";
                fprintf(f,"%s\n",ln.c_str());
            }

            for(i2=(*i1).access_grant_group.begin(); i2!=(*i1).access_grant_group.end(); ++i2) {
                ln=""; for(ii=0; ii<indent; ++ii) ln+=" ";
                ln+=(string)"access_grant_group \""+c_var_string::script_esc("logic.txt",*i2)+"\"";
                fprintf(f,"%s\n",ln.c_str());
            }

            vector<string>::iterator i3;
            for(i3=(*i1).groups.begin(); i3!=(*i1).groups.end(); ++i3) {
                ln=""; for(ii=0; ii<indent; ++ii) ln+=" ";
                ln+=(string)"member_of_group \""+c_var_string::script_esc("logic.txt",(*i3))+"\"";
                fprintf(f,"%s\n",ln.c_str());
            }

            vector<c_dcc_host>::iterator i4;
            for(i4=(*i1).dcc_hosts.begin(); i4!=(*i1).dcc_hosts.end(); ++i4) {
                ln=""; for(ii=0; ii<indent; ++ii) ln+=" ";
                ln+=(string)"dcc_host ";
                ln+=tos((*i4).group);
                ln+=" "+(*i4).host;
                fprintf(f,"%s\n",ln.c_str());
            }

            ln=""; for(ii=0; ii<indent; ++ii) ln+=" ";
            ln+=((*i1).access_to_partyline?"access_to_partyline 1":"access_to_partyline 0");
            fprintf(f,"%s\n",ln.c_str());

            ln=""; for(ii=0; ii<indent; ++ii) ln+=" ";
            ln+=((*i1).access_grant_partyline?"access_grant_partyline 1":"access_grant_partyline 0");
            fprintf(f,"%s\n",ln.c_str());

            //---
            ln=""; for(ii=0; ii<indent; ++ii) ln+=" ";
            ln+=((*i1).access_to_backup?"access_to_backup 1":"access_to_backup 0");
            fprintf(f,"%s\n",ln.c_str());

            ln=""; for(ii=0; ii<indent; ++ii) ln+=" ";
            ln+=((*i1).access_to_rehash?"access_to_rehash 1":"access_to_rehash 0");
            fprintf(f,"%s\n",ln.c_str());

            ln=""; for(ii=0; ii<indent; ++ii) ln+=" ";
            ln+=((*i1).access_grant_backup?"access_grant_backup 1":"access_grant_backup 0");
            fprintf(f,"%s\n",ln.c_str());

            ln=""; for(ii=0; ii<indent; ++ii) ln+=" ";
            ln+=((*i1).access_grant_rehash?"access_grant_rehash 1":"access_grant_rehash 0");
            fprintf(f,"%s\n",ln.c_str());


            //---
            ln=""; for(ii=0; ii<indent; ++ii) ln+=" ";
            ln+=((*i1).access_to_restart?"access_to_restart 1":"access_to_restart 0");
            fprintf(f,"%s\n",ln.c_str());

            ln=""; for(ii=0; ii<indent; ++ii) ln+=" ";
            ln+=((*i1).access_grant_restart?"access_grant_restart 1":"access_grant_restart 0");
            fprintf(f,"%s\n",ln.c_str());

            ln=""; for(ii=0; ii<indent; ++ii) ln+=" ";
            ln+=((*i1).access_to_die?"access_to_die 1":"access_to_die 0");
            fprintf(f,"%s\n",ln.c_str());

            ln=""; for(ii=0; ii<indent; ++ii) ln+=" ";
            ln+=((*i1).access_grant_die?"access_grant_die 1":"access_grant_die 0");
            fprintf(f,"%s\n",ln.c_str());
            //---

            ln=""; for(ii=0; ii<indent; ++ii) ln+=" ";
            ln+=((*i1).access_to_plususer?"access_to_+user 1":"access_to_+user 0");
            fprintf(f,"%s\n",ln.c_str());

            ln=""; for(ii=0; ii<indent; ++ii) ln+=" ";
            ln+=((*i1).access_grant_plususer?"access_grant_+user 1":"access_grant_+user 0");
            fprintf(f,"%s\n",ln.c_str());

            ln=""; for(ii=0; ii<indent; ++ii) ln+=" ";
            ln+=((*i1).access_to_plusproc?"access_to_+proc 1":"access_to_+proc 0");
            fprintf(f,"%s\n",ln.c_str());

            ln=""; for(ii=0; ii<indent; ++ii) ln+=" ";
            ln+=((*i1).access_grant_plusproc?"access_grant_+proc 1":"access_grant_+proc 0");
            fprintf(f,"%s\n",ln.c_str());

            ln=""; for(ii=0; ii<indent; ++ii) ln+=" ";
            ln+=((*i1).access_to_replication?"access_to_replication 1":"access_to_replication 0");
            fprintf(f,"%s\n",ln.c_str());

            ln=""; for(ii=0; ii<indent; ++ii) ln+=" ";
            ln+=((*i1).access_grant_replication?"access_grant_replication 1":"access_grant_replication 0");
            fprintf(f,"%s\n",ln.c_str());

            //---
            ln=""; for(ii=0; ii<indent; ++ii) ln+=" ";
            ln+=((*i1).access_to_filesystem?"access_to_filesystem 1":"access_to_filesystem 0");
            fprintf(f,"%s\n",ln.c_str());

            ln=""; for(ii=0; ii<indent; ++ii) ln+=" ";
            ln+=((*i1).access_grant_filesystem?"access_grant_filesystem 1":"access_grant_filesystem 0");
            fprintf(f,"%s\n",ln.c_str());
            //---

            //---
            ln=""; for(ii=0; ii<indent; ++ii) ln+=" ";
            ln+=((*i1).access_to_private?"access_to_private 1":"access_to_private 0");
            fprintf(f,"%s\n",ln.c_str());

            ln=""; for(ii=0; ii<indent; ++ii) ln+=" ";
            ln+=((*i1).access_grant_private?"access_grant_private 1":"access_grant_private 0");
            fprintf(f,"%s\n",ln.c_str());
            //---

            //---
            ln=""; for(ii=0; ii<indent; ++ii) ln+=" ";
            ln+=((*i1).access_to_upgrade?"access_to_upgrade 1":"access_to_upgrade 0");
            fprintf(f,"%s\n",ln.c_str());

            ln=""; for(ii=0; ii<indent; ++ii) ln+=" ";
            ln+=((*i1).access_grant_upgrade?"access_grant_upgrade 1":"access_grant_upgrade 0");
            fprintf(f,"%s\n",ln.c_str());
            //---

            //---
            ln=""; for(ii=0; ii<indent; ++ii) ln+=" ";
            ln+=((*i1).access_to_apply?"access_to_apply 1":"access_to_apply 0");
            fprintf(f,"%s\n",ln.c_str());

            ln=""; for(ii=0; ii<indent; ++ii) ln+=" ";
            ln+=((*i1).access_grant_apply?"access_grant_apply 1":"access_grant_apply 0");
            fprintf(f,"%s\n",ln.c_str());
            //---

            ln=""; for(ii=0; ii<indent; ++ii) ln+=" ";
            ln+=((*i1).access_to_can_send_all_users?"access_to_can_send_all_users 1":"access_to_can_send_all_users 0");
            fprintf(f,"%s\n",ln.c_str());

            ln=""; for(ii=0; ii<indent; ++ii) ln+=" ";
            ln+=((*i1).access_grant_can_send_all_users?"access_grant_can_send_all_users 1":"access_grant_can_send_all_users 0");
            fprintf(f,"%s\n",ln.c_str());
            //---

            ln=""; for(ii=0; ii<indent; ++ii) ln+=" ";
            ln+=((*i1).access_to_can_send_unknown_users?"access_to_can_send_unknown_users 1":"access_to_can_send_unknown_users 0");
            fprintf(f,"%s\n",ln.c_str());

            ln=""; for(ii=0; ii<indent; ++ii) ln+=" ";
            ln+=((*i1).access_grant_can_send_unknown_users?"access_grant_can_send_unknown_users 1":"access_grant_can_send_unknown_users 0");
            fprintf(f,"%s\n",ln.c_str());

            //---

            {
                ln=""; for(ii=0; ii<indent; ++ii) ln+=" ";
                ln+=(string)"partyline_msg_flood "+(*i1).partyline_msg_flood.make_string();
                fprintf(f,"%s\n",ln.c_str());
            }

            map<string,string>::iterator i6;
            for(i6=(*i1).meta.begin(); i6!=(*i1).meta.end(); ++i6) {
                ln=""; for(ii=0; ii<indent; ++ii) ln+=" ";
                ln+="meta \"";
                ln+=c_var_string::script_esc("logic.txt",(*i6).first);
                ln+="\" \"";
                ln+=c_var_string::script_esc("logic.txt",(*i6).second);
                ln+="\"";
                fprintf(f,"%s\n",ln.c_str());
            }

            vector<c_channel>* chans=NULL;
            if(iii==0)
                chans=&r_channels;
            else if(iii==1)
                chans=&r_channel_templates;
            else if(iii==2)
                chans=&r_channel_terminators;
            vector<c_channel>::iterator i5;
            for(i5=chans->begin(); i5!=chans->end(); ++i5) {
                if(!(*i5).username.compare((*i1).name)) {
                    ln=""; for(ii=0; ii<indent; ++ii) ln+=" ";
                    ln+=(string)"channel \""+c_var_string::script_esc("logic.txt",(*i5).channel_name)+"\"";
                    fprintf(f,"%s\n",ln.c_str());

                    ln=""; for(ii=0; ii<indent; ++ii) ln+=" ";
                    ln+="{";
                    fprintf(f,"%s\n",ln.c_str());
                    indent+=4;

                    for(i3=(*i5).groups.begin(); i3!=(*i5).groups.end(); ++i3) {
                        ln=""; for(ii=0; ii<indent; ++ii) ln+=" ";
                        ln+=(string)"member_of_group \""+c_var_string::script_esc("logic.txt",*i3)+"\"";
                        fprintf(f,"%s\n",ln.c_str());
                    }

                    ln=""; for(ii=0; ii<indent; ++ii) ln+=" ";
                    ln+="on_deop "+(*i5).on_deop;
                    fprintf(f,"%s\n",ln.c_str());

                    ln=""; for(ii=0; ii<indent; ++ii) ln+=" ";
                    ln+="on_ban "+(*i5).on_ban;
                    fprintf(f,"%s\n",ln.c_str());

                    ln=""; for(ii=0; ii<indent; ++ii) ln+=" ";
                    ln+="on_unban "+(*i5).on_unban;
                    fprintf(f,"%s\n",ln.c_str());

                    ln=""; for(ii=0; ii<indent; ++ii) ln+=" ";
                    ln+="on_kick "+(*i5).on_kick;
                    fprintf(f,"%s\n",ln.c_str());

                    ln=""; for(ii=0; ii<indent; ++ii) ln+=" ";
                    ln+="on_op "+(*i5).on_op;
                    fprintf(f,"%s\n",ln.c_str());

                    ln=""; for(ii=0; ii<indent; ++ii) ln+=" ";
                    ln+="on_voice "+(*i5).on_voice;
                    fprintf(f,"%s\n",ln.c_str());

                    ln=""; for(ii=0; ii<indent; ++ii) ln+=" ";
                    ln+="on_devoice "+(*i5).on_devoice;
                    fprintf(f,"%s\n",ln.c_str());

                    ln=""; for(ii=0; ii<indent; ++ii) ln+=" ";
                    ln+="on_join "+(*i5).on_join;
                    fprintf(f,"%s\n",ln.c_str());

                    ln=""; for(ii=0; ii<indent; ++ii) ln+=" ";
                    ln+="on_banned "+(*i5).on_banned;
                    fprintf(f,"%s\n",ln.c_str());

                    ln=""; for(ii=0; ii<indent; ++ii) ln+=" ";
                    ln+="on_flood "+(*i5).on_flood;
                    fprintf(f,"%s\n",ln.c_str());

                    ln=""; for(ii=0; ii<indent; ++ii) ln+=" ";
                    ln+="on_privmsg "+(*i5).on_privmsg;
                    fprintf(f,"%s\n",ln.c_str());

                    ln=""; for(ii=0; ii<indent; ++ii) ln+=" ";
                    ln+="on_notice "+(*i5).on_notice;
                    fprintf(f,"%s\n",ln.c_str());

                    ln=""; for(ii=0; ii<indent; ++ii) ln+=" ";
                    ln+="on_except "+(*i5).on_except;
                    fprintf(f,"%s\n",ln.c_str());

                    ln=""; for(ii=0; ii<indent; ++ii) ln+=" ";
                    ln+="on_unexcept "+(*i5).on_unexcept;
                    fprintf(f,"%s\n",ln.c_str());

                    ln=""; for(ii=0; ii<indent; ++ii) ln+=" ";
                    ln+="on_invite "+(*i5).on_invite;
                    fprintf(f,"%s\n",ln.c_str());

                    ln=""; for(ii=0; ii<indent; ++ii) ln+=" ";
                    ln+="on_uninvite "+(*i5).on_uninvite;
                    fprintf(f,"%s\n",ln.c_str());

                    ln=""; for(ii=0; ii<indent; ++ii) ln+=" ";
                    ln+="on_not_invited "+(*i5).on_not_invited;
                    fprintf(f,"%s\n",ln.c_str());

                    ln=""; for(ii=0; ii<indent; ++ii) ln+=" ";
                    ln+="on_not_in_reop "+(*i5).on_not_in_reop;
                    fprintf(f,"%s\n",ln.c_str());

                    ln=""; for(ii=0; ii<indent; ++ii) ln+=" ";
                    ln+="on_reop "+(*i5).on_reop;
                    fprintf(f,"%s\n",ln.c_str());

                    ln=""; for(ii=0; ii<indent; ++ii) ln+=" ";
                    ln+="on_other_mode "+(*i5).on_other_mode;
                    fprintf(f,"%s\n",ln.c_str());

                    ln=""; for(ii=0; ii<indent; ++ii) ln+=" ";
                    ln+="on_other_mode_mask "+(*i5).on_other_mode_mask;
                    fprintf(f,"%s\n",ln.c_str());

                    if((*i5).msg_flood.lines==0)
                        (*i5).msg_flood.seconds=0;
                    if((*i5).notice_flood.lines==0)
                        (*i5).notice_flood.seconds=0;
                    if((*i5).repeat_flood.lines==0)
                        (*i5).repeat_flood.seconds=0;
                    if((*i5).nick_flood.lines==0)
                        (*i5).nick_flood.seconds=0;
                    if((*i5).join_flood.lines==0)
                        (*i5).join_flood.seconds=0;
                    if((*i5).mode_flood.lines==0)
                        (*i5).mode_flood.seconds=0;
                    if((*i5).ctcp_flood.lines==0)
                        (*i5).ctcp_flood.seconds=0;

                    ln=""; for(ii=0; ii<indent; ++ii) ln+=" ";
                    ln+=(string)"msg_flood "+(*i5).msg_flood.make_string();
                    fprintf(f,"%s\n",ln.c_str());

                    ln=""; for(ii=0; ii<indent; ++ii) ln+=" ";
                    ln+=(string)"notice_flood "+(*i5).notice_flood.make_string();
                    fprintf(f,"%s\n",ln.c_str());

                    ln=""; for(ii=0; ii<indent; ++ii) ln+=" ";
                    ln+=(string)"repeat_flood "+(*i5).repeat_flood.make_string();
                    fprintf(f,"%s\n",ln.c_str());

                    ln=""; for(ii=0; ii<indent; ++ii) ln+=" ";
                    ln+=(string)"nick_flood "+(*i5).nick_flood.make_string();
                    fprintf(f,"%s\n",ln.c_str());

                    ln=""; for(ii=0; ii<indent; ++ii) ln+=" ";
                    ln+=(string)"join_flood "+(*i5).join_flood.make_string();
                    fprintf(f,"%s\n",ln.c_str());

                    ln=""; for(ii=0; ii<indent; ++ii) ln+=" ";
                    ln+=(string)"mode_flood "+(*i5).mode_flood.make_string();
                    fprintf(f,"%s\n",ln.c_str());

                    ln=""; for(ii=0; ii<indent; ++ii) ln+=" ";
                    ln+=(string)"ctcp_flood "+(*i5).ctcp_flood.make_string();
                    fprintf(f,"%s\n",ln.c_str());

                    vector<c_dynamic_rule>::iterator i8;
                    for(i8=(*i5).allow_dynamic.begin(); i8!=(*i5).allow_dynamic.end(); ++i8) {
                        ln=""; for(ii=0; ii<indent; ++ii) ln+=" ";
                        ln+="allow_dynamic \"";
                        ln+=(*i8).group;
                        ln+="\" \"";
                        ln+=c_var_string::script_esc("logic.txt",(*i8).plus_modes);
                        ln+="\" \"";
                        ln+=c_var_string::script_esc("logic.txt",(*i8).minus_modes);
                        ln+="\"";
                        fprintf(f,"%s\n",ln.c_str());
                    }

                    ln=""; for(ii=0; ii<indent; ++ii) ln+=" ";
                    ln+=(string)"dynamic1 \""+c_var_string::script_esc("logic.txt",(*i5).dynamic_plus_modes)+"\"";
                    fprintf(f,"%s\n",ln.c_str());

                    ln=""; for(ii=0; ii<indent; ++ii) ln+=" ";
                    ln+=(string)"dynamic2 \""+c_var_string::script_esc("logic.txt",(*i5).dynamic_minus_modes)+"\"";
                    fprintf(f,"%s\n",ln.c_str());

                    ln=""; for(ii=0; ii<indent; ++ii) ln+=" ";
                    ln+=(string)"can_send_unknown_users "+((*i5).can_send_unknown_users?"1":"0");
                    fprintf(f,"%s\n",ln.c_str());

                    indent-=4;
                    ln=""; for(ii=0; ii<indent; ++ii) ln+=" ";
                    ln+="}";
                    fprintf(f,"%s\n",ln.c_str());
                }
            }
            indent-=4;
            ln=""; for(ii=0; ii<indent; ++ii) ln+=" ";
            ln+="}";
            fprintf(f,"%s\n",ln.c_str());
        }
    }

    vector<c_channel_def>::iterator i3;
    for(i3=r_channel_defs.begin(); i3!=r_channel_defs.end(); ++i3) {

        ln=""; for(ii=0; ii<indent; ++ii) ln+=" ";
        ln+=(string)"channel \""+c_var_string::script_esc("logic.txt",(*i3).channel_name)+"\"";
        fprintf(f,"%s\n",ln.c_str());

        ln=""; for(ii=0; ii<indent; ++ii) ln+=" ";
        ln+="{";
        fprintf(f,"%s\n",ln.c_str());
        indent+=4;

        ln=""; for(ii=0; ii<indent; ++ii) ln+=" ";
        ln+="last_changed ";
        ln+=tos((*i3).last_changed);
        fprintf(f,"%s\n",ln.c_str());

        vector<string>::iterator i5;
        for(i5=(*i3).groups.begin(); i5!=(*i3).groups.end(); ++i5) {
            ln=""; for(ii=0; ii<indent; ++ii) ln+=" ";
            ln+=(string)"member_of_group \""+c_var_string::script_esc("logic.txt",(*i5))+"\"";
            fprintf(f,"%s\n",ln.c_str());
        }

        sort((*i3).replication.begin(),(*i3).replication.end());
        repl_i=unique((*i3).replication.begin(),(*i3).replication.end());
        (*i3).replication.erase(repl_i,(*i3).replication.end());

        vector<c_replication>::iterator i7;
        for(i7=(*i3).replication.begin(); i7!=(*i3).replication.end(); ++i7) {
            ln=""; for(ii=0; ii<indent; ++ii) ln+=" ";
            string type;
            if((*i7).type==REPL_PUSH)
                type="push";
            if((*i7).type==REPL_PULL)
                type="pull";
            if((*i7).type==REPL_PUSHPULL)
                type="pushpull";
            ln+=(string)"replication "+(*i7).partner+" "+type;
            fprintf(f,"%s\n",ln.c_str());
        }

        ln=""; for(ii=0; ii<indent; ++ii) ln+=" ";
        ln+="on_mode "+(*i3).on_mode;
        fprintf(f,"%s\n",ln.c_str());

        ln=""; for(ii=0; ii<indent; ++ii) ln+=" ";
        ln+="on_key "+(*i3).on_key;
        fprintf(f,"%s\n",ln.c_str());

        ln=""; for(ii=0; ii<indent; ++ii) ln+=" ";
        ln+="on_limit "+(*i3).on_limit;
        fprintf(f,"%s\n",ln.c_str());

        ln=""; for(ii=0; ii<indent; ++ii) ln+=" ";
        ln+="on_topic "+(*i3).on_topic;
        fprintf(f,"%s\n",ln.c_str());

        ln=""; for(ii=0; ii<indent; ++ii) ln+=" ";
        ln+="on_ircop "+(*i3).on_ircop;
        fprintf(f,"%s\n",ln.c_str());

        ln=""; for(ii=0; ii<indent; ++ii) ln+=" ";
        ln+="on_ctcp "+(*i3).on_ctcp;
        fprintf(f,"%s\n",ln.c_str());

        ln=""; for(ii=0; ii<indent; ++ii) ln+=" ";
        ln+="on_privmsg "+(*i3).on_privmsg;
        fprintf(f,"%s\n",ln.c_str());

        ln=""; for(ii=0; ii<indent; ++ii) ln+=" ";
        ln+="on_notice "+(*i3).on_notice;
        fprintf(f,"%s\n",ln.c_str());

        ln=""; for(ii=0; ii<indent; ++ii) ln+=" ";
        ln+="on_part "+(*i3).on_part;
        fprintf(f,"%s\n",ln.c_str());

        ln=""; for(ii=0; ii<indent; ++ii) ln+=" ";
        ln+="on_dynamic_ban "+(*i3).on_dynamic_ban;
        fprintf(f,"%s\n",ln.c_str());

        ln=""; for(ii=0; ii<indent; ++ii) ln+=" ";
        ln+="on_server_msg "+(*i3).on_server_msg;
        fprintf(f,"%s\n",ln.c_str());

        {
            vector<string>::iterator i1;
            for(i1=(*i3).keys.begin(); i1!=(*i3).keys.end(); ++i1) {
                ln=""; for(ii=0; ii<indent; ++ii) ln+=" ";
                ln+=(string)"key \""+c_var_string::script_esc("logic.txt",logic_eval(this_module,*i1,NULL,debug))+"\"";
                fprintf(f,"%s\n",ln.c_str());
            }
        }

        {
            map<string,string>::iterator i1;
            for(i1=(*i3).dynamic_bans.begin(); i1!=(*i3).dynamic_bans.end(); ++i1) {
                ln=""; for(ii=0; ii<indent; ++ii) ln+=" ";
                ln+=(string)"dynamic_ban \""+c_var_string::script_esc("logic.txt",(*i1).first)+"\" \""+c_var_string::script_esc("logic.txt",(*i1).second)+"\"";
                fprintf(f,"%s\n",ln.c_str());
            }
        }

        {
            vector<string>::iterator i1;
            for(i1=(*i3).dynbans_editors.begin(); i1!=(*i3).dynbans_editors.end(); ++i1) {
                ln=""; for(ii=0; ii<indent; ++ii) ln+=" ";
                ln+=(string)"partyline_dynbans_editors \""+c_var_string::script_esc("logic.txt",(*i1))+"\"";
                fprintf(f,"%s\n",ln.c_str());
            }
        }

        vector<c_dynamic_rule>::iterator i8;
        for(i8=(*i3).allow_dynamic.begin(); i8!=(*i3).allow_dynamic.end(); ++i8) {
            ln=""; for(ii=0; ii<indent; ++ii) ln+=" ";
            ln+="allow_dynamic \"";
            ln+=(*i8).group;
            ln+="\" \"";
            ln+=c_var_string::script_esc("logic.txt",(*i8).plus_modes);
            ln+="\" \"";
            ln+=c_var_string::script_esc("logic.txt",(*i8).minus_modes);
            ln+="\"";
            fprintf(f,"%s\n",ln.c_str());
        }

        ln=""; for(ii=0; ii<indent; ++ii) ln+=" ";
        ln+=(string)"dynamic1 \""+c_var_string::script_esc("logic.txt",(*i3).dynamic_plus_modes)+"\"";
        fprintf(f,"%s\n",ln.c_str());

        ln=""; for(ii=0; ii<indent; ++ii) ln+=" ";
        ln+=(string)"dynamic2 \""+c_var_string::script_esc("logic.txt",(*i3).dynamic_minus_modes)+"\"";
        fprintf(f,"%s\n",ln.c_str());

        indent-=4;
        ln=""; for(ii=0; ii<indent; ++ii) ln+=" ";
        ln+="}";
        fprintf(f,"%s\n",ln.c_str());
    }

    {
        ln=""; for(ii=0; ii<indent; ++ii) ln+=" ";
        ln+=(string)"private";
        fprintf(f,"%s\n",ln.c_str());

        ln=""; for(ii=0; ii<indent; ++ii) ln+=" ";
        ln+="{";
        fprintf(f,"%s\n",ln.c_str());
        indent+=4;

        ln=""; for(ii=0; ii<indent; ++ii) ln+=" ";
        ln+="last_changed ";
        ln+=tos(r_private.last_changed);
        fprintf(f,"%s\n",ln.c_str());

        vector<string>::iterator i5;
        for(i5=r_private.groups.begin(); i5!=r_private.groups.end(); ++i5) {
            ln=""; for(ii=0; ii<indent; ++ii) ln+=" ";
            ln+=(string)"member_of_group \""+c_var_string::script_esc("logic.txt",(*i5))+"\"";
            fprintf(f,"%s\n",ln.c_str());
        }

        sort(r_private.replication.begin(),r_private.replication.end());
        repl_i=unique(r_private.replication.begin(),r_private.replication.end());
        r_private.replication.erase(repl_i,r_private.replication.end());

        vector<c_replication>::iterator i7;
        for(i7=r_private.replication.begin(); i7!=r_private.replication.end(); ++i7) {
            ln=""; for(ii=0; ii<indent; ++ii) ln+=" ";
            string type;
            if((*i7).type==REPL_PUSH)
                type="push";
            if((*i7).type==REPL_PULL)
                type="pull";
            if((*i7).type==REPL_PUSHPULL)
                type="pushpull";
            ln+=(string)"replication "+(*i7).partner+" "+type;
            fprintf(f,"%s\n",ln.c_str());
        }

        ln=""; for(ii=0; ii<indent; ++ii) ln+=" ";
        ln+="on_privmsg "+r_private.on_privmsg;
        fprintf(f,"%s\n",ln.c_str());

        ln=""; for(ii=0; ii<indent; ++ii) ln+=" ";
        ln+="on_notice "+r_private.on_notice;
        fprintf(f,"%s\n",ln.c_str());

        ln=""; for(ii=0; ii<indent; ++ii) ln+=" ";
        ln+="on_ctcp "+r_private.on_ctcp;
        fprintf(f,"%s\n",ln.c_str());

        ln=""; for(ii=0; ii<indent; ++ii) ln+=" ";
        ln+="on_filesys_got_new "+r_private.on_filesys_got_new;
        fprintf(f,"%s\n",ln.c_str());

        ln=""; for(ii=0; ii<indent; ++ii) ln+=" ";
        ln+="on_fnc "+r_private.on_fnc;
        fprintf(f,"%s\n",ln.c_str());

        ln=""; for(ii=0; ii<indent; ++ii) ln+=" ";
        ln+="on_broadcast "+r_private.on_broadcast;
        fprintf(f,"%s\n",ln.c_str());

        ln=""; for(ii=0; ii<indent; ++ii) ln+=" ";
        ln+="on_server_msg "+r_private.on_server_msg;
        fprintf(f,"%s\n",ln.c_str());

        ln=""; for(ii=0; ii<indent; ++ii) ln+=" ";
        ln+="on_internal_event "+r_private.on_internal_event;
        fprintf(f,"%s\n",ln.c_str());

        indent-=4;
        ln=""; for(ii=0; ii<indent; ++ii) ln+=" ";
        ln+="}";
        fprintf(f,"%s\n",ln.c_str());
    }

    c_module* core_mod=get_module("core",true);
    if(core_mod) {
        vector<c_procedure>::iterator i2;
        for(i2=core_mod->r_procedures.begin(); i2!=core_mod->r_procedures.end(); ++i2) {

            ln=""; for(ii=0; ii<indent; ++ii) ln+=" ";
            ln+=(string)"procedure "+(*i2).name;
            fprintf(f,"%s\n",ln.c_str());

            ln=""; for(ii=0; ii<indent; ++ii) ln+=" ";
            ln+="{";
            fprintf(f,"%s\n",ln.c_str());
            indent+=4;

            ln=""; for(ii=0; ii<indent; ++ii) ln+=" ";
            ln+="last_changed ";
            ln+=tos((*i2).last_changed);
            fprintf(f,"%s\n",ln.c_str());

            vector<string>::iterator i5;
            for(i5=(*i2).groups.begin(); i5!=(*i2).groups.end(); ++i5) {
                ln=""; for(ii=0; ii<indent; ++ii) ln+=" ";
                ln+=(string)"member_of_group \""+c_var_string::script_esc("logic.txt",(*i5))+"\"";
                fprintf(f,"%s\n",ln.c_str());
            }

            sort((*i2).replication.begin(),(*i2).replication.end());
            repl_i=unique((*i2).replication.begin(),(*i2).replication.end());
            (*i2).replication.erase(repl_i,(*i2).replication.end());

            vector<c_replication>::iterator i7;
            for(i7=(*i2).replication.begin(); i7!=(*i2).replication.end(); ++i7) {
                ln=""; for(ii=0; ii<indent; ++ii) ln+=" ";
                string type;
                if((*i7).type==REPL_PUSH)
                    type="push";
                if((*i7).type==REPL_PULL)
                    type="pull";
                if((*i7).type==REPL_PUSHPULL)
                    type="pushpull";
                ln+=(string)"replication "+(*i7).partner+" "+type;
                fprintf(f,"%s\n",ln.c_str());
            }

            vector<c_rproc>::iterator i_rproc;
            for(i_rproc=(*i2).rproc.begin(); i_rproc!=(*i2).rproc.end(); ++i_rproc) {
                ln=""; for(ii=0; ii<indent; ++ii) ln+=" ";
                ln+=(string)"remote_bot_call "+(*i_rproc).remote_bot;
                fprintf(f,"%s\n",ln.c_str());
            }

            string name;
            for(unsigned int i7=0; i7<(*i2).name.length(); ++i7) {
                if((*i2).name[i7]==0x20)
                    break;
                name+=(*i2).name[i7];
            }

            list<string> commands;
            vector<string> groups;
            vector<c_replication> repls;
            vector<string> rproc;
            logic_list_proc(name,commands,(*i2).name,groups,repls,rproc);
            list<string>::iterator i6;
            for(i6=commands.begin(); i6!=commands.end(); ++i6)
                fprintf(f,"%s\n",(*i6).c_str());

            indent-=4;
        }

        core_mod->dump_funcs_to_file(f);
    }

    {
        fprintf(f,"groups\n");

        ln=""; for(ii=0; ii<indent; ++ii) ln+=" ";
        ln+="{";
        fprintf(f,"%s\n",ln.c_str());
        indent+=4;

        vector<string> got;

        vector<string>::iterator i1;
        for(i1=r_all_groups.begin(); i1!=r_all_groups.end(); ++i1) {
            if((*i1).empty())
                continue;
            vector<string>::iterator i2;
            for(i2=got.begin(); i2!=got.end(); ++i2)
                if(!(*i1).compare(*i2))
                    goto skip;

            got.push_back(*i1);

            ln=""; for(ii=0; ii<indent; ++ii) ln+=" ";
            ln+="\"";
            ln+=c_var_string::script_esc("logic.txt",(*i1));
            ln+="\"";
            fprintf(f,"%s\n",ln.c_str());
skip:
            ;
        }

        indent-=4;
        ln=""; for(ii=0; ii<indent; ++ii) ln+=" ";
        ln+="}";
        fprintf(f,"%s\n",ln.c_str());
    }

    fprintf(f,"#eof\n");

    fclose(f);

    conf_pass.backup();

    irc_last_keepalive_detection=my_time(NULL)-IRC_KEEPALIVE_PERIOD;    // do not check for timeout now - this operation could have been slow
    irc_keepalive_sent=false;
}

int logic_get_priority(const string& class_)
{
    if(!class_.compare(LOGIC_TXT_CONST_HIGH))
        return HIGH_PRIORITY;
    if(!class_.compare(LOGIC_TXT_CONST_LOW))
        return LOW_PRIORITY;
    if(!class_.compare(LOGIC_TXT_CONST_CRITICAL))
        return CRITICAL_PRIORITY;
    return -1;
}

string logic_set_priority(const int priority)
{
    switch(priority) {
        case LOW_PRIORITY:
            return LOGIC_TXT_CONST_LOW;
        case HIGH_PRIORITY:
            return LOGIC_TXT_CONST_HIGH;
        case CRITICAL_PRIORITY:
            return LOGIC_TXT_CONST_CRITICAL;
        default:
            return LOGIC_TXT_CONST_LOW;
    }
}

/*!
    \brief Generates debug string used in case of run-time error
    \author VooDooMan
    \version 1
    \date 2005-09-08
    \param line Line number
    \param ln Whole command line
*/
string logic_make_debug_string(int line, string ln)
{
    string res="logic.txt:";
    res+=tos(line);
    res+=" -> ";
    res+=ln;
    return res;
}

/*!
    \brief Checks for IDENT connection (should be called in a loop)
    \author VooDooMan
*/
void logic_loop_identd()
{
    unsigned short local=0, remote=0;
    if(irc_socket.cmp()) {
        local=sock_get_local_port(irc_socket);
        remote=sock_get_remote_port(irc_socket);
    }
    if(identd_check(ident_string,local,remote)) {
        //identd_shutdown();
        // do not shutdown, we will do it manually after successful auth (for case of identd conflict)
    }
}

/*!
    \brief Checks if host mask matches criteria
    \author VooDooMan
    \version 2
    \date 2004, 2005-12-28
    \param hostmask Host mask ("*!*@*.??.info")
    \param fullname_mask Mask for full name
    \param hostname Host name ("nick!ident@host") (does it match?)
    \param fullname Full name (does it match?)
    \return Returns true it it all match
    \retval true All match
    \retval false No match
*/
bool logic_match2(const string& hostmask, const string& fullname_mask, const string& hostname, const string& fullname)
{
    if(!match(hostmask.c_str(),hostname.c_str()) &&
       !match(fullname_mask.c_str(),fullname.c_str())) {
        return true;
    }
    return false;
}

/*!
    \brief Expands IPv6 "::" compression from host part, if needed
    \author VooDooMan
    \version 3
    \date 2005, 2005-12-28
    \param hostmask Hostmask
    \return Returns resulting hostmask
*/
string logic_hostmask_ipv6_decompress(const string& hostmask)
{
    if(hostmask.find(':')==string::npos)
        return hostmask;
    string first, host, cidr;
    int context=0;
    for(unsigned int i1=0; i1<hostmask.length(); ++i1) {
        if(hostmask[i1]=='@' && context==0) {
            context++;
            continue;
        }
        if(hostmask[i1]=='/' && context==1) {
            context++;
            continue;
        }
        if(context==0)
            first+=hostmask[i1];
        if(context==1)
            host+=hostmask[i1];
        if(context==2)
            cidr+=hostmask[i1];
    }
    char* h=ipv6_decompress((char*)host.c_str());
    host=h;
    MY_DELETE_A(h);
    first=first+"@"+host;
    if(!cidr.empty())
        first+=(string)"/"+cidr;
    return first;
}

/*!
    \brief Checks mask to match (advanced with DNS resolution)
    \author VooDooMan
    \version 2
    \date 2004, 2006-07-03
    \param mask1 Mask #1 to compare for match
    \param mask2 Mask #2 to compare for match
    \param fnm1 Full name mask #1 to compare for match
    \param fnm2 Full name mask #2 to compare for match
    \param was_dns Remains untouched if we have tried DNS, if not, returns false, so it should be called again later
    \retval true For full match
    \retval false For no match
*/
bool logic_check_mask(string mask1, string mask2, const string& fnm1, const string& fnm2, bool& was_dns)
{
    mask1=logic_hostmask_ipv6_decompress(mask1);
    mask2=logic_hostmask_ipv6_decompress(mask2);

    // match?
    if(logic_match2(mask2,fnm2,mask1,fnm1) || logic_match2(mask1,fnm1,mask2,fnm2))
        return true;

    // if full name doesn't match, it is all lost = no match
    if(match(fnm2.c_str(),fnm1.c_str()) &&
        match(fnm1.c_str(),fnm2.c_str()))
        return false;
    /*
    if(!logic_match2("*",fnm2,"*",fnm1) && !logic_match2("*",fnm1,"*",fnm2))
        return false;
    */

    // h_maskX <- host field only
    string h_mask1=mask1;
    string h_mask2=mask2;
    if(h_mask1.find("@",0)!=string::npos)
        h_mask1.erase(0,h_mask1.find("@",0)+1);
    if(h_mask2.find("@",0)!=string::npos)
        h_mask2.erase(0,h_mask2.find("@",0)+1);

    if(h_mask1.length()==0 || h_mask2.length()==0)
        return false;

    // unX <- user name only ("nick!ident")
    string un1=mask1;
    if(un1.find("@",0)!=string::npos)
        un1.erase(un1.find("@",0),un1.length()-un1.find("@",0));
    string un2=mask2;
    if(un2.find("@",0)!=string::npos)
        un2.erase(un2.find("@",0),un2.length()-un2.find("@",0));

    // ip_maskX <- try to parse IPv4 dotted, or zero, if zero, ip_maskX <- try to IPv4 resolve, or zero

    vector<in_addr> ip41;
    // we don't want to resolve if there are wildcards or CIDR
    if(h_mask1.find("?",0)==string::npos && h_mask1.find("*",0)==string::npos && h_mask1.find("#",0)==string::npos && h_mask1.find("/",0)==string::npos) {
        uint32 ip_mask1=sock_ipv4_from_string(h_mask1.c_str());
        if(ip_mask1==0) {
            e_resolv_result r=threads_resolve4(h_mask1.c_str(),ip41,RT_RESOLVE_SCHEDULE);
            if(RR_SCHEDULED==r)
                was_dns=false;
            /*
            if(RR_FAILURE==r)
                ;
            */
        } else {
            in_addr a;
            memcpy(&a,&ip_mask1,sizeof(a));
            ip41.push_back(a);
        }
    }
    if(ip41.empty()) {
        in_addr a;
        memset(&a,0,sizeof(a));
        ip41.push_back(a);
    }

    vector<in_addr> ip42;
    if(h_mask2.find("?",0)==string::npos && h_mask2.find("*",0)==string::npos && h_mask2.find("#",0)==string::npos && h_mask2.find("/",0)==string::npos) {
        uint32 ip_mask2=sock_ipv4_from_string(h_mask2.c_str());
        if(ip_mask2==0) {
            e_resolv_result r=threads_resolve4(h_mask2.c_str(),ip42,RT_RESOLVE_SCHEDULE);
            if(RR_SCHEDULED==r)
                was_dns=false;
            /*
            if(RR_FAILURE==r)
                ;
            */
        } else {
            in_addr a;
            memcpy(&a,&ip_mask2,sizeof(a));
            ip42.push_back(a);
        }
    }
    if(ip42.empty()) {
        in_addr a;
        memset(&a,0,sizeof(a));
        ip42.push_back(a);
    }

    string h_mask1_old=h_mask1;
    string h_mask2_old=h_mask2;

    for(vector<in_addr>::iterator ip41i=ip41.begin(); ip41i!=ip41.end(); ++ip41i) {
        for(vector<in_addr>::iterator ip42i=ip42.begin(); ip42i!=ip42.end(); ++ip42i) {
            uint32 ip_mask1;
            memcpy(&ip_mask1,&(*ip41i),sizeof(ip_mask1));
            uint32 ip_mask2;
            memcpy(&ip_mask2,&(*ip42i),sizeof(ip_mask2));
            h_mask1=h_mask1_old;
            h_mask2=h_mask2_old;
            {
                // h_maskX <- if it is not IPv6, convert it to IPv4-mapped IPv6
                if(h_mask1.find(":",0)==string::npos) {
                    uint32 x=sock_ipv4_from_string(h_mask1.c_str());
                    if(x) {
                        h_mask1="0:0:0:0:0:ffff:"+h_mask1;
                        string tmp=h_mask1;
                        if(h_mask1.find("/",0)!=string::npos) {
                            h_mask1.erase(h_mask1.find("/",0),h_mask1.length()-h_mask1.find("/",0));
                            tmp.erase(0,tmp.find("/",0)+1);
                            int j=stoi(tmp);
                            j+=(128-32);
                            h_mask1+="/";
                            h_mask1+=tos(j);
                        }
                    } else {
                        if(ip_mask1) {
                            ostringstream tmp;
                            tmp << "0:0:0:0:0:ffff:";
                            tmp << tos(ip_mask1 & 0xff) << "." << tos((ip_mask1>>8) & 0xff) << "." << tos((ip_mask1>>16) & 0xff) << "." << tos((ip_mask1>>24) & 0xff);
                            h_mask1=tmp.str();
                        } else
                            h_mask1=(string)"0:0:0:0:0:ffff:"+h_mask1;
                    }
                }

                if(h_mask2.find(":",0)==string::npos) {
                    uint32 x=sock_ipv4_from_string(h_mask2.c_str());
                    if(x) {
                        h_mask2="0:0:0:0:0:ffff:"+h_mask2;
                        string tmp=h_mask2;
                        if(h_mask2.find("/",0)!=string::npos) {
                            h_mask2.erase(h_mask2.find("/",0),h_mask2.length()-h_mask2.find("/",0));
                            tmp.erase(0,tmp.find("/",0)+1);
                            int j=stoi(tmp);
                            j+=(128-32);
                            h_mask2+="/";
                            h_mask2+=tos(j);
                        }
                    } else {
                        if(ip_mask2) {
                            ostringstream tmp;
                            tmp << "0:0:0:0:0:ffff:";
                            tmp << tos(ip_mask2 & 0xff) << "." << tos((ip_mask2>>8) & 0xff) << "." << tos((ip_mask2>>16) & 0xff) << "." << tos((ip_mask2>>24) & 0xff);
                            h_mask2=tmp.str();
                        } else
                            h_mask2=(string)"0:0:0:0:0:ffff:"+h_mask2;
                    }
                }
            }

            // maskX <- build "nick!ident@IPv6"
            mask1=un1+"@"+h_mask1;
            mask2=un2+"@"+h_mask2;

            // still no match?
            if(!match((un1+"@"+h_mask1).c_str(),(un2+"@"+h_mask2).c_str()))
                return true;
            if(!match((un2+"@"+h_mask2).c_str(),(un1+"@"+h_mask1).c_str()))
                return true;
        }
    }

    // ip6_maskX <- try to resolve via IPv6 DNS
    vector<in_addr6_> ip61;
    if(h_mask1.find("?",0)==string::npos && h_mask1.find("*",0)==string::npos && h_mask1.find("#",0)==string::npos && h_mask1.find("/",0)==string::npos) {
        e_resolv_result r=threads_resolve6(h_mask1.c_str(),ip61,RT_RESOLVE_SCHEDULE);
        if(RR_SCHEDULED==r)
            was_dns=false;
        /*
        if(RR_FAILURE==r)
            ;
        */
    }
    if(ip61.empty()) {
        in_addr6_ a;
        memset(&a,0,sizeof(a));
        ip61.push_back(a);
    }

    vector<in_addr6_> ip62;
    if(h_mask2.find("?",0)==string::npos && h_mask2.find("*",0)==string::npos && h_mask2.find("#",0)==string::npos && h_mask2.find("/",0)==string::npos) {
        e_resolv_result r=threads_resolve6(h_mask2.c_str(),ip62,RT_RESOLVE_SCHEDULE);
        if(RR_SCHEDULED==r)
            was_dns=false;
        /*
        if(RR_FAILURE==r)
            ;
        */
    }
    if(ip62.empty()) {
        in_addr6_ a;
        memset(&a,0,sizeof(a));
        ip62.push_back(a);
    }

    for(vector<in_addr6_>::iterator ip61i=ip61.begin(); ip61i!=ip61.end(); ++ip61i) {
        for(vector<in_addr6_>::iterator ip62i=ip62.begin(); ip62i!=ip62.end(); ++ip62i) {
            in_addr6_ ip6_mask1;
            memcpy(&ip6_mask1,&(*ip61i),sizeof(ip6_mask1));
            in_addr6_ ip6_mask2;
            memcpy(&ip6_mask2,&(*ip62i),sizeof(ip6_mask2));
            h_mask1=h_mask1_old;
            h_mask2=h_mask2_old;

            bool chance=false;

            // if we have at least one successful IPv6 DNS resolution... then continue
            if(ip6_mask1.bytes[0] || ip6_mask1.bytes[1] || ip6_mask1.bytes[2] || ip6_mask1.bytes[3] || ip6_mask1.bytes[4] ||
               ip6_mask1.bytes[5] || ip6_mask1.bytes[6] || ip6_mask1.bytes[7] || ip6_mask1.bytes[8] || ip6_mask1.bytes[9] ||
               ip6_mask1.bytes[10] || ip6_mask1.bytes[11] || ip6_mask1.bytes[12] || ip6_mask1.bytes[13] ||
               ip6_mask1.bytes[14] || ip6_mask1.bytes[15]) {
                char tmp[256];
                // we don't want compressed format
                sprintf(tmp,"%x:%x:%x:%x:%x:%x:%x:%x",  ((unsigned int)ip6_mask1.bytes[0])<<8 | ((unsigned int)ip6_mask1.bytes[1]),
                                                            ((unsigned int)ip6_mask1.bytes[2])<<8 | ((unsigned int)ip6_mask1.bytes[3]),
                                                            ((unsigned int)ip6_mask1.bytes[4])<<8 | ((unsigned int)ip6_mask1.bytes[5]),
                                                            ((unsigned int)ip6_mask1.bytes[6])<<8 | ((unsigned int)ip6_mask1.bytes[7]),
                                                            ((unsigned int)ip6_mask1.bytes[8])<<8 | ((unsigned int)ip6_mask1.bytes[9]),
                                                            ((unsigned int)ip6_mask1.bytes[10])<<8 | ((unsigned int)ip6_mask1.bytes[11]),
                                                            ((unsigned int)ip6_mask1.bytes[12])<<8 | ((unsigned int)ip6_mask1.bytes[13]),
                                                            ((unsigned int)ip6_mask1.bytes[14])<<8 | ((unsigned int)ip6_mask1.bytes[15]));
                    h_mask1=tmp;
                    chance=true;
            }
            if(ip6_mask2.bytes[0] || ip6_mask2.bytes[1] || ip6_mask2.bytes[2] || ip6_mask2.bytes[3] || ip6_mask2.bytes[4] ||
                ip6_mask2.bytes[5] || ip6_mask2.bytes[6] || ip6_mask2.bytes[7] || ip6_mask2.bytes[8] || ip6_mask2.bytes[9] ||
                ip6_mask2.bytes[10] || ip6_mask2.bytes[11] || ip6_mask2.bytes[12] || ip6_mask2.bytes[13] ||
                ip6_mask2.bytes[14] || ip6_mask2.bytes[15]) {
                char tmp[256];
                // we don't want compressed format
                sprintf(tmp,"%x:%x:%x:%x:%x:%x:%x:%x",  ((unsigned int)ip6_mask2.bytes[0])<<8 | ((unsigned int)ip6_mask2.bytes[1]),
                                                            ((unsigned int)ip6_mask2.bytes[2])<<8 | ((unsigned int)ip6_mask2.bytes[3]),
                                                            ((unsigned int)ip6_mask2.bytes[4])<<8 | ((unsigned int)ip6_mask2.bytes[5]),
                                                            ((unsigned int)ip6_mask2.bytes[6])<<8 | ((unsigned int)ip6_mask2.bytes[7]),
                                                            ((unsigned int)ip6_mask2.bytes[8])<<8 | ((unsigned int)ip6_mask2.bytes[9]),
                                                            ((unsigned int)ip6_mask2.bytes[10])<<8 | ((unsigned int)ip6_mask2.bytes[11]),
                                                            ((unsigned int)ip6_mask2.bytes[12])<<8 | ((unsigned int)ip6_mask2.bytes[13]),
                                                            ((unsigned int)ip6_mask2.bytes[14])<<8 | ((unsigned int)ip6_mask2.bytes[15]));
                h_mask2=tmp;
                chance=true;
            }

            // ... then continue, or not
            if(!chance)
                continue;

            // last chance to match
            if(!match_ipmask((char*)(un1+"@"+h_mask1).c_str(),(char*)un2.c_str(),ip6_mask2,true))
                return true;
            if(!match_ipmask((char*)(un2+"@"+h_mask2).c_str(),(char*)un1.c_str(),ip6_mask1,true))
                return true;
        }
    }

    return false;
}

/*!
    \brief Looks up and find user as specified in logic.txt
    \author VooDooMan
    \version 3
    \date 2004, 2005-11-29, 2007-11-05
    \param nick Nick name
    \param ident Ident
    \param host Host
    \param fullname Full name
    \param irc_op Pass there true if WHOIS indicates that this is an irc op
    \param was_dns Remains untouched if we have tried DNS, if not, returns false, so it should be called again later
    \return Returns name of user as in logic.txt, or an empty string if user was not found
*/
string logic_find_user(const string& nick, const string& ident, string host, const string& fullname, bool UNUSEDARG(irc_op), bool& was_dns)
{
    {
        bool not_got=false;
        vector<c_online_channel>::iterator i1;
        for(i1=irc_channels.begin(); i1!=irc_channels.end(); ++i1) {
            vector<c_online_user>::iterator i2;
            for(i2=(*i1).users.begin(); i2!=(*i1).users.end(); ++i2) {
                if(!irc_cmp_strings_case_insensitive((*i2).nick,nick)) {
                    if((*i2).in_logic_as.empty() || !(*i2).got_whois) {
                        not_got=true;
                        break;
                    }
                    return (*i2).in_logic_as;
                }
            }
            if(not_got)
                break;
        }
    }
    if(!irc_cmp_strings_case_insensitive(nick,irc_nick)) {
        deque<c_user>::iterator i;
        for(i=r_users.begin(); i!=r_users.end(); ++i) {
            if((*i).host_bot)
                return (*i).name;
        }
    }

    bool trying_reverse=false;
again:

    deque<c_user>::iterator i;
    for(i=r_users.begin(); i!=r_users.end(); ++i) {
        vector<string>::iterator is, ii;
        for(is=(*i).fullname.begin(); is!=(*i).fullname.end(); ++is) {
            for(ii=(*i).hostmask.begin(); ii!=(*i).hostmask.end(); ++ii) {
                if(!(*i).host_unknown && logic_check_mask(*ii,nick+"!"+ident+"@"+host,*is,fullname,was_dns))
                    return (*i).name;
            }
        }
    }

    if(!trying_reverse && core_conf_getvar("try_reverse_lookup").compare("0")) {
        if(host.length() && host[0]>='0' && host[0]<='9') {
            /*
            char name[1024];
            logic_reverse((char*)host.c_str(),name,sizeof(name));
            if(name[0]!=0) {
                host=name;
                trying_reverse=true;
                goto again;
            }
            */
            char name[1024];
            e_resolv_result r=threads_reverse4(host.c_str(),name,sizeof(name)-1,RT_RESOLVE_SCHEDULE);
            if(RR_OK==r) {
                host=name;
                trying_reverse=true;
                goto again;
            }
            if(RR_SCHEDULED==r)
                was_dns=false;
            in_addr6_ ip6;
            if(my_inet_pton(AF_INET6,host.c_str(),&ip6)==1) {
                e_resolv_result r=threads_reverse6(ip6,name,sizeof(name)-1,RT_RESOLVE_SCHEDULE);
                if(RR_OK==r) {
                    host=name;
                    trying_reverse=true;
                    goto again;
                }
                if(RR_SCHEDULED==r)
                    was_dns=false;
            }
        }
    }

    // try host_unknown
    for(i=r_users.begin(); i!=r_users.end(); ++i) {
        vector<string>::iterator is, ii;
        if((*i).host_unknown) {
            return (*i).name;
        }
    }

    return (string)"";
}

/*!
    \brief Looks up and find user as specified in logic.txt
    \author VooDooMan
    \version 1
    \date 2005-11-29
    \param nick Nick name
    \param ident Ident
    \param host Host
    \param fullname Full name
    \param irc_op Pass there true if WHOIS indicates that this is an irc op
    \return Returns name of user as in logic.txt, or an empty string if user was not found
*/
string logic_find_user(const string& nick, const string& ident, const string& host, const string& fullname, bool irc_op)
{
    bool was_dns=true;
    return logic_find_user(nick,ident,host,fullname,irc_op,was_dns);
}

/*!
    \brief Looks up and find user as specified in logic.txt, and checks if user has access to partyline
    \author VooDooMan
    \version 2
    \date 2004, 2007-02-11
    \param nick Nick name
    \param ident Ident
    \param host Host
    \param fullname Full name
    \param irc_op Pass there true if WHOIS indicates that this is an irc op
    \param user_name_as_in_logic Returns name of user as in logic.txt, or undefined if it was not found
    \return Returns true if user has access to partyline; and false if not or user was not found
*/
bool logic_access_to_partyline(const string& nick, const string& ident, const string& host, const string& fullname, const bool UNUSEDARG(irc_op), string& user_name_as_in_logic)
{
    if(!irc_cmp_strings_case_insensitive(nick,irc_nick)) {
        return false;
    }

    deque<c_user>::iterator i;
    for(i=r_users.begin(); i!=r_users.end(); ++i) {
        if((*i).host_unknown)
            continue;
        vector<string>::iterator is, ii;
        for(is=(*i).fullname.begin(); is!=(*i).fullname.end(); ++is) {
            for(ii=(*i).hostmask.begin(); ii!=(*i).hostmask.end(); ++ii) {
                bool was_dns=true;
                if(logic_check_mask(*ii,nick+"!"+ident+"@"+host,fullname,*is,was_dns)) {
                    user_name_as_in_logic=(*i).name;
                    return (*i).access_to_partyline;
                }
            }
        }
    }

    // try host_unknown
    for(i=r_users.begin(); i!=r_users.end(); ++i) {
        vector<string>::iterator is, ii;
        if((*i).host_unknown) {
            user_name_as_in_logic=(*i).name;
            return (*i).access_to_partyline;
        }
    }

    return false;
}

/*!
    \brief This should be called in a loop. Executes timers and processes script output with freeing resources
    \author VooDooMan
    \version 1
    \date 2004
*/
void logic_loop()
{
    my_time_t now;
    my_time(&now);

    static c_module* core_mod=NULL;
    if(!core_mod)
        core_mod=get_module("core",true);
    if(!core_mod)
        return;

    static my_time_t last_autobackup_check=0;
    if(last_autobackup_check+5<now) {
        last_autobackup_check=now;

        my_time_t auto_backup_interval=sto_time(core_conf_getvar("auto_backup_interval"));
        if(!last_autobackup)
            last_autobackup=now;
        if(auto_backup_interval!=0 && last_autobackup+auto_backup_interval<=now) {
            last_autobackup=now;
            logic_partyline_backup(core_mod,"SELF (autobackup)");
        }
    }

    static my_time_t last_module_process=0;
    bool process=false;
    if(last_module_process!=now) {
        process=true;
        last_module_process=now;
    }
    vector<c_module*>::iterator i_mod;
    for(i_mod=modules.begin(); process && i_mod!=modules.end(); ++i_mod) {
        if(!(*i_mod)->is_loaded())
            continue;
    l1:
        vector<c_timer*>::iterator i;
        for(i=(*i_mod)->r_timers.begin(); i!=(*i_mod)->r_timers.end(); ++i) {
            if((*i)->erase) {
                if((*i)->bot_unlink_timer)
                    botnet_kill_timer(*i);
                if((*i)->vars) {
                    var_keeper->free_global((*i)->vars,true);
                }
                (*i)->vars=NULL;
                MY_DELETE(*i);
                (*i_mod)->r_timers.erase(i);
                goto l1;
            }
            if(((*i)->again_at<=now && !(*i)->bot_unlink_timer) || (*i)->bot_unlink_trigger) {
                c_timer* t=*i;
                (*i)->set_at=now;
                (*i)->again_at=now+(*i)->duration;

                if(t->once || t->bot_unlink_trigger) {
                    t->erase=true;
                    t->bot_unlink_trigger=false;
                }

                session_id_t saved_session_id=var_keeper->get_current_session();

                var_keeper->set_current_session((*i_mod)->get_session());

                //c_var_array* vars=var_keeper->register_mod_global(t->vars);
                c_temp_local local(t->vars);
                c_var_array* vars=NULL;
                unsigned new_local_id=local.get_params(vars);
                string call_string=t->expression;

                c_function f;
                f.clear(false);
                string error_msg;
                string code="function __timer_pseudofunction__(){\n";
                code+=call_string;
                code+=";\n}";

                bool okay=true;

                stack<c_eval_iteration> st;

                bool error=false;
                bool need_more=false;
                string v="#version 2\n";
                f.parse_command(*i_mod,v,error,error_msg,need_more,t->debug,"");
                if(error || !need_more) {
                    f.clear_microcode();
                    f.destroy_symbols();
                    f.destroy(true);
                    okay=false;
                } else
                    f.parse_command(*i_mod,code,error,error_msg,need_more,t->debug,"");
                if(error || need_more || !f.is_okay()) {
                    f.clear_microcode();
                    f.destroy_symbols();
                    f.destroy(true);
                    okay=false;
                }
                if(!okay) {
                    var_keeper->free_local(new_local_id,true);
                    f.clear_microcode();
                    f.destroy_symbols();
                    f.destroy(true);
                    error_msg=((string)"'execute' command parse error: ")+error_msg;
                    c_eval_runtime_error ee(*i_mod,t->debug,error_msg,__FILE__,__my_func__);
                    ee.report(false);
                } else {
                    c_eval_iteration ste(*i_mod,vars,t->debug);
                    ste.fn_destroy=true;
                    ste.fn_destroy_mc=true;
                    ste.fn=MY_NEW c_function(f);
                    f.destroy(false);
                    ste.fn_destroy=true;
                    //ste.symbol_table_=ste.fn->make_symbol_table(ste.fn->symbol_table);
                    ste.symbol_table_=ste.fn->symbol_table;
                    ste.symbol_table_->update_symbols_from_vars(vars);
                    ste.proc_eval=true;
                    ste.state=1000;
                    ste.return_to_state=3001;
                    ste.local_id=new_local_id;
                    st.push(ste);

                    eval_loop(st);
                }

                var_keeper->set_current_session(saved_session_id);

                // because iterator could be invalid in case of deletion/addition of timer
                goto l1;
            }
        }

    l2:
        map<int,int> num_scripts;
        vector<c_tmp_file>::iterator i1;
        for(i1=(*i_mod)->tmp_files.begin(); i1!=(*i_mod)->tmp_files.end(); ++i1) {
            if(!(*i1).script)
                continue;
            bool active=(*i1).h && exec_process_active((*i1).h);
            if(!(*i1).has_been_run)
                active=false;
            if(!active)
                continue;
            bool got=false;
            map<int,int>::iterator i2;
            for(i2=num_scripts.begin(); i2!=num_scripts.end(); ++i2)
                if((*i2).first==(*i1).script_num) {
                    got=true;
                    (*i2).second++;
                }
            if(!got) {
                pair<int,int> p;
                p.first=(*i1).script_num;
                p.second=1;
                num_scripts.insert(p);
            }
        }
        for(i1=(*i_mod)->tmp_files.begin(); i1!=(*i_mod)->tmp_files.end(); ++i1) {
            if((*i1).has_been_run && (*i1).h && !exec_process_active((*i1).h)) {
                if((*i1).script)
                    (*i_mod)->process_script_output((*i1).script_type,(*i1).script_num,(*i1).h,(*i1).filename,(*i1).noremove);
                else {
                    int size;
                    char* buf=NULL;
                    exec_read_output((*i1).h,buf,size);
                    if(buf)
                        my_free(buf);
                    exec_process_end((*i1).h);
                    MY_DELETE((*i1).h);
                }
                (*i_mod)->tmp_files.erase(i1);
                goto l2;
            }
#if defined(_WIN32) || defined(_WIN64)
            // on *NIX following will not cause to kill the process,
            //     when process finishes later, there is zombie...
            //     (hopefully fixed those zombie processes)
            if((*i1).has_been_run && (*i1).start_up_time+10*60<my_time(NULL)) {
                exec_terminate_process((*i1).h,1);
                exec_process_end((*i1).h);
                MY_DELETE((*i1).h);
                (*i_mod)->tmp_files.erase(i1);
                goto l2;
            }
#endif
            if(!(*i1).script)
                continue;
            map<int,int>::iterator i2;
            int i=0;
            for(i2=num_scripts.begin(); i2!=num_scripts.end(); ++i2)
                if((*i2).first==(*i1).script_num) {
                    i=(*i2).second;
                    break;
                }
            if(!i && !(*i1).has_been_run) {
                cout << "Executing in " << (*i1).command_dir << ": " << (*i1).command_line << " " << (*i1).command_args << endl;
                (*i1).h=exec_async_exec((*i1).command_line,(*i1).command_args,(*i1).command_dir);
                (*i1).has_been_run=true;
                (*i1).start_up_time=my_time(NULL);
                break;
            }
        }

l3:
        // process tmp_files_garbage
        for(i1=tmp_files_garbage.begin(); i1!=tmp_files_garbage.end(); ++i1) {
            if((*i1).has_been_run && (*i1).h && !exec_process_active((*i1).h)) {
                int size;
                char* buf=NULL;
                exec_read_output((*i1).h,buf,size);
                if(buf)
                    my_free(buf);
                exec_process_end((*i1).h);
                MY_DELETE((*i1).h);

                tmp_files_garbage.erase(i1);
                goto l3;
            }
#if defined(_WIN32) || defined(_WIN64)
            // on *NIX following will not cause to kill the process,
            //     when process finishes later, there is zombie...
            //     (hopefully fixed those zombie processes)
            if((*i1).has_been_run && (*i1).start_up_time+10*60<my_time(NULL)) {
                exec_terminate_process((*i1).h,1);
                exec_process_end((*i1).h);
                MY_DELETE((*i1).h);
                tmp_files_garbage.erase(i1);
                goto l3;
            }
#endif
            if(!(*i1).script)
                continue;
            map<int,int>::iterator i2;
            int i=0;
            for(i2=num_scripts.begin(); i2!=num_scripts.end(); ++i2)
                if((*i2).first==(*i1).script_num) {
                    i=(*i2).second;
                    break;
                }
            if(!i && !(*i1).has_been_run) {
                tmp_files_garbage.erase(i1);
                goto l3;
            }
        }
    }
}

/*!
    \brief Retrieves host for DCC by DCC gorup and user's name
    \author VooDooMan
    \version 1
    \date 2004
    \param dcc_group DCC group
    \param user_name_as_in_logic User name as define in logic.txt
    \return Returns IPv4/IPv6 address of DCC host, or "0.0.0.0" if dcc_group is invalid for user_name_as_in_logic
*/
string logic_get_dcc_host(const int dcc_group, const string& user_name_as_in_logic)
{
    deque<c_user>::iterator i;
    for(i=r_users.begin(); i!=r_users.end(); ++i) {
        if(!(*i).name.compare(user_name_as_in_logic)) {
            vector<c_dcc_host>::iterator i2;
            for(i2=(*i).dcc_hosts.begin(); i2!=(*i).dcc_hosts.end(); ++i2) {
                if((*i2).group==dcc_group) {
                    return (*i2).host;
                }
            }
        }
    }
    return (string)"0.0.0.0";
}

/*!
    \brief Erases duplicated or empty strings in the single vector
    \author VooDooMan
    \version 2
    \date 2004, 2005-11-30
    \param v Vector of string to clean up
*/
void logic_erase_duplicate_or_empty_string_in_vector(vector<string>& v)
{
    sort(v.begin(),v.end());
    vector<string>::iterator i1=unique(v.begin(),v.end());
    v.erase(i1,v.end());
}

/*!
    \brief Retrieves access rights of user on partyline
    \author VooDooMan
    \version 1
    \date 2004
    \param user User name as defined in logic.txt
    \param access_to_users Array with list of users that can modify
    \param access_to_groups Array with list of groups that can modify
    \param access_grant_groups Array with list of groups that can add/exclude objects to/from
    \param access_to_proc_groups Array with list of procedure groups that can modify
    \param access_grant_proc_groups Array with list of procedure groups that can add/exclude procedure to/from
    \param access_grant_proc Array with list of procedures that can add/exclude procedure to/from
    \param access_to_channel Array with list of channels that can modify
    \param access_grant_channel Array with list of channels that can add/exclude users to/from
    \param access_grant_partyline Set to true if can grant/deny access to partyline to users
    \param access_to_procs Array with list of procedures that can modify
    \param access_to_backup Has the user access to .backup command?
    \param access_grant_backup Can the user grant access to .backup command to other user?
    \param access_to_rehash Has the user access to .rehash command?
    \param access_grant_rehash Can the user grant access to .rehash command to other user?
    \param access_to_plususer Has the user access to .+user command?
    \param access_grant_plususer Can the user grant access to .+user command to other user?
    \param access_to_plusproc Has the user access to .+proc command?
    \param access_grant_plusproc Can the user grant access to .+proc command to other user?
    \param access_to_replication Can the user change replication parameters?
    \param access_to_restart Has the user access to .restart command?
    \param access_grant_restart Can the user grant access to .restart command to other user?
    \param access_to_die Has the user access to .die command?
    \param access_grant_die Can the user grant access to .die command to other user?
    \param partyline_msg_flood Flood limit for partyline
    \param access_to_chan_defs List of channels that the user has access to (definitions)
    \param access_to_filesystem Has the user access to filesystem? (.filesystem)
    \param access_grant_filesystem Can the user grant access to filesystem (.filesystem) to other user?
    \param access_to_private Can user modify private query definitions?
    \param access_grant_private Can user grant access to other user modify private query definitions?
    \param access_to_can_send_all_users Can user send message / file to all users?
    \param access_grant_can_send_all_users Can user grant access to other user to send message / file to all users?
    \param access_to_can_send_unknown_users Can user send message / file to unknown users?
    \param access_grant_can_send_unknown_users Can user grant access to other user to send message / file to unknown users?
    \param access_usage_proc Array with list of procedures that can assign events to
    \param access_usage_proc_groups Array with list groups of procedures that can assign events to
    \param access_grant_replication Can this user grant right to change replication parameters?
    \param access_to_upgrade Has user access to .upgrade command?
    \param access_grant_upgrade Can this user grant accesst to .upgrade command to other user?
    \param access_to_apply Has user access to .apply command?
    \param access_grant_apply Can this user grant accesst to .apply command to other user?
    \param access_to_users_also_terminators Array with list of users that can modify, with terminators as well
*/
void logic_partyline_access(const string& user, vector<string>& access_to_users,
                            vector<string>& access_to_groups, vector<string>& access_grant_groups,
                            vector<string>& access_to_proc_groups, vector<string>& access_grant_proc_groups,
                            vector<string>& access_grant_proc, vector<string>& access_to_channel,
                            vector<string>& access_grant_channel, bool& access_grant_partyline,
                            vector<string>& access_to_procs, bool& access_to_backup,
                            bool& access_grant_backup, bool& access_to_rehash, bool& access_grant_rehash,
                            bool& access_to_plususer, bool& access_grant_plususer, bool& access_to_plusproc,
                            bool& access_grant_plusproc, c_flood& partyline_msg_flood,
                            bool& access_to_replication, bool& access_grant_replication,
                            vector<string>& access_to_chan_defs, bool& access_to_restart,
                            bool& access_grant_restart, bool& access_to_die, bool& access_grant_die,
                            bool& access_to_filesystem, bool& access_grant_filesystem, bool& access_to_private,
                            bool& access_grant_private, bool& access_to_can_send_all_users, bool& access_grant_can_send_all_users,
                            bool& access_to_can_send_unknown_users, bool& access_grant_can_send_unknown_users,
                            vector<string>& access_usage_proc, vector<string>& access_usage_proc_groups, bool& access_to_upgrade,
                            bool& access_grant_upgrade, bool& access_to_apply, bool& access_grant_apply,
                            vector<string>& access_to_users_also_terminators)
{
    access_to_users.clear();
    access_to_groups.clear();
    access_grant_groups.clear();
    access_to_proc_groups.clear();
    access_grant_proc_groups.clear();
    access_grant_proc.clear();
    access_to_channel.clear();
    access_grant_channel.clear();
    access_grant_partyline=false;
    access_to_procs.clear();
    access_to_backup=access_grant_backup=access_to_rehash=access_grant_rehash=
        access_to_plususer=access_grant_plususer=access_to_plusproc=
        access_grant_plusproc=false;
    partyline_msg_flood.lines=0;
    partyline_msg_flood.seconds=0;
    access_to_replication=access_grant_replication=false;
    access_to_chan_defs.clear();
    access_to_restart=access_grant_restart=access_to_die=access_grant_die=
        access_to_filesystem=access_grant_filesystem=access_to_private=
        access_grant_private=access_to_can_send_all_users=
        access_grant_can_send_all_users=access_to_can_send_unknown_users=
        access_grant_can_send_unknown_users=false;
    access_usage_proc.clear();
    access_usage_proc_groups.clear();
    access_to_upgrade=access_grant_upgrade=false;
    access_to_apply=access_grant_apply=false;
    access_to_users_also_terminators.clear();

    c_module* core_mod=get_module("core",true);

    deque<c_user>::iterator i1;
    for(i1=r_users.begin(); i1!=r_users.end(); ++i1) {
        if(!(*i1).name.compare(user)) {
            bool to_all_groups=false;
            bool grant_all_groups=false;
            bool usage_all_groups=false;
            if(core_mod) {
                // access_to_group
                {
                    access_to_groups.clear();
                    vector<string>::iterator i2;
                    for(i2=(*i1).access_to_group.begin(); i2!=(*i1).access_to_group.end(); ++i2) {
                        if(!(*i2).compare("*")) {
                            to_all_groups=true;
                        }
                        access_to_groups.push_back(*i2);
                    }
                    if(to_all_groups) {
                        vector<string>::iterator i2;
                        for(i2=r_all_groups.begin(); i2!=r_all_groups.end(); ++i2)
                            access_to_groups.push_back(*i2);
                    }
                }
                // access_grant_group
                {
                    access_grant_groups.clear();
                    vector<string>::iterator i2;
                    for(i2=(*i1).access_grant_group.begin(); i2!=(*i1).access_grant_group.end(); ++i2) {
                        if(!(*i2).compare("*")) {
                            grant_all_groups=true;
                        }
                        access_grant_groups.push_back(*i2);
                    }
                    if(grant_all_groups) {
                        vector<string>::iterator i2;
                        for(i2=r_all_groups.begin(); i2!=r_all_groups.end(); ++i2)
                            access_grant_groups.push_back(*i2);
                    }
                }
                // access_usage_group
                {
                    access_usage_proc_groups.clear();
                    vector<string>::iterator i2;
                    for(i2=(*i1).access_usage_procedure.begin(); i2!=(*i1).access_usage_procedure.end(); ++i2) {
                        if(!(*i2).compare("*")) {
                            usage_all_groups=true;
                        }
                        access_usage_proc_groups.push_back(*i2);
                    }
                    if(usage_all_groups) {
                        vector<string>::iterator i2;
                        for(i2=r_all_groups.begin(); i2!=r_all_groups.end(); ++i2)
                            access_usage_proc_groups.push_back(*i2);
                    }
                }
                {
                    access_to_users.clear();
                    // iterate users
                    vector<string>::iterator i2;
                    for(i2=access_to_groups.begin(); i2!=access_to_groups.end(); ++i2) {
                        deque<c_user>::iterator i3;
                        for(i3=r_users.begin(); i3!=r_users.end(); ++i3) {
                            if(to_all_groups) {
                                access_to_users.push_back((*i3).name);
                                continue;
                            }
                            vector<string>::iterator i4;
                            for(i4=(*i3).groups.begin(); i4!=(*i3).groups.end(); ++i4) {
                                if(!(*i4).compare(*i2)) {
                                    access_to_users.push_back((*i3).name);
                                    break;
                                }
                            }
                        }
                        if(to_all_groups)
                            break;
                    }
                }
                {
                    // terminators
                    access_to_users_also_terminators=access_to_users;
                    vector<string>::iterator i2;
                    for(i2=access_to_groups.begin(); i2!=access_to_groups.end(); ++i2) {
                        deque<c_user>::iterator i3;
                        for(i3=r_terminators.begin(); i3!=r_terminators.end(); ++i3) {
                            if(to_all_groups) {
                                access_to_users_also_terminators.push_back((*i3).name);
                                continue;
                            }
                            vector<string>::iterator i4;
                            for(i4=(*i3).groups.begin(); i4!=(*i3).groups.end(); ++i4) {
                                if(!(*i4).compare(*i2)) {
                                    access_to_users_also_terminators.push_back((*i3).name);
                                    break;
                                }
                            }
                        }
                        if(to_all_groups)
                            break;
                    }
                }
                // access_to_proc
                {
                    access_to_proc_groups=access_to_groups;
                    access_to_procs.clear();
                    vector<string>::iterator i2;
                    // iterate procedures
                    for(i2=access_to_proc_groups.begin(); i2!=access_to_proc_groups.end(); ++i2) {
                        vector<c_procedure>::iterator i3;
                        for(i3=core_mod->r_procedures.begin(); i3!=core_mod->r_procedures.end(); ++i3) {
                            if(to_all_groups) {
                                access_to_procs.push_back((*i3).name);
                                continue;
                            }
                            vector<string>::iterator i4;
                            for(i4=(*i3).groups.begin(); i4!=(*i3).groups.end(); ++i4) {
                                if(!(*i4).compare(*i2)) {
                                    access_to_procs.push_back((*i3).name);
                                    break;
                                }
                            }
                        }
                        if(to_all_groups)
                            break;
                    }
                    // iterate functions
                    for(i2=access_to_proc_groups.begin(); i2!=access_to_proc_groups.end(); ++i2) {
                        map<string,c_function*>::iterator i3;
                        for(i3=core_mod->r_functions.functions.begin(); i3!=core_mod->r_functions.functions.end(); ++i3) {
                            if(to_all_groups) {
                                access_to_procs.push_back((*i3).second->declaration);
                                continue;
                            }
                            vector<string>::iterator i4;
                            for(i4=(*i3).second->groups.begin(); i4!=(*i3).second->groups.end(); ++i4) {
                                if(!(*i4).compare(*i2)) {
                                    access_to_procs.push_back((*i3).second->declaration);
                                    break;
                                }
                            }
                        }
                        if(to_all_groups)
                            break;
                    }
                }
                // access_grant_proc
                {
                    access_grant_proc_groups=access_grant_groups;
                    access_grant_proc.clear();
                    // iterate procedures
                    vector<string>::iterator i2;
                    for(i2=access_grant_proc_groups.begin(); i2!=access_grant_proc_groups.end(); ++i2) {
                        vector<c_procedure>::iterator i3;
                        for(i3=core_mod->r_procedures.begin(); i3!=core_mod->r_procedures.end(); ++i3) {
                            if(grant_all_groups) {
                                access_grant_proc.push_back((*i3).name);
                                continue;
                            }
                            vector<string>::iterator i4;
                            for(i4=(*i3).groups.begin(); i4!=(*i3).groups.end(); ++i4) {
                                if(!(*i4).compare(*i2)) {
                                    access_grant_proc.push_back((*i3).name);
                                    break;
                                }
                            }
                        }
                        if(grant_all_groups)
                            break;
                    }
                    // iterate functions
                    for(i2=access_grant_proc_groups.begin(); i2!=access_grant_proc_groups.end(); ++i2) {
                        map<string,c_function*>::iterator i3;
                        for(i3=core_mod->r_functions.functions.begin(); i3!=core_mod->r_functions.functions.end(); ++i3) {
                            if(grant_all_groups) {
                                access_grant_proc.push_back((*i3).second->declaration);
                                continue;
                            }
                            vector<string>::iterator i4;
                            for(i4=(*i3).second->groups.begin(); i4!=(*i3).second->groups.end(); ++i4) {
                                if(!(*i4).compare(*i2)) {
                                    access_grant_proc.push_back((*i3).second->declaration);
                                    break;
                                }
                            }
                        }
                        if(grant_all_groups)
                            break;
                    }
                }
                // access_usage_proc
                {
                    access_usage_proc.clear();
                    // iterate procedures
                    vector<string>::iterator i2;
                    for(i2=access_usage_proc_groups.begin(); i2!=access_usage_proc_groups.end(); ++i2) {
                        vector<c_procedure>::iterator i3;
                        for(i3=core_mod->r_procedures.begin(); i3!=core_mod->r_procedures.end(); ++i3) {
                            if(usage_all_groups) {
                                access_usage_proc.push_back((*i3).name);
                                continue;
                            }
                            vector<string>::iterator i4;
                            for(i4=(*i3).groups.begin(); i4!=(*i3).groups.end(); ++i4) {
                                if(!(*i4).compare(*i2)) {
                                    access_usage_proc.push_back((*i3).name);
                                    break;
                                }
                            }
                        }
                        if(usage_all_groups)
                            break;
                    }
                    // iterate functions
                    for(i2=access_usage_proc_groups.begin(); i2!=access_usage_proc_groups.end(); ++i2) {
                        map<string,c_function*>::iterator i3;
                        for(i3=core_mod->r_functions.functions.begin(); i3!=core_mod->r_functions.functions.end(); ++i3) {
                            if(usage_all_groups) {
                                access_usage_proc.push_back((*i3).second->declaration);
                                continue;
                            }
                            vector<string>::iterator i4;
                            for(i4=(*i3).second->groups.begin(); i4!=(*i3).second->groups.end(); ++i4) {
                                if(!(*i4).compare(*i2)) {
                                    access_usage_proc.push_back((*i3).second->declaration);
                                    break;
                                }
                            }
                        }
                        if(usage_all_groups)
                            break;
                    }
                }
                // channel definitions
                {
                    access_to_chan_defs.clear();
                    access_to_channel.clear();
                    vector<string>::iterator i2;
                    // iterate channels
                    for(i2=access_to_groups.begin(); i2!=access_to_groups.end(); ++i2) {
                        vector<c_channel_def>::iterator i3;
                        for(i3=r_channel_defs.begin(); i3!=r_channel_defs.end(); ++i3) {
                            if(to_all_groups) {
                                access_to_chan_defs.push_back((*i3).channel_name);
                                access_to_channel.push_back((*i3).channel_name);
                                continue;
                            }
                            vector<string>::iterator i4;
                            for(i4=(*i3).groups.begin(); i4!=(*i3).groups.end(); ++i4) {
                                if(!(*i4).compare(*i2)) {
                                    access_to_chan_defs.push_back((*i3).channel_name);
                                    access_to_channel.push_back((*i3).channel_name);
                                    break;
                                }
                            }
                        }
                        if(to_all_groups)
                            break;
                    }
                    access_grant_channel.clear();
                    for(i2=access_grant_groups.begin(); i2!=access_grant_groups.end(); ++i2) {
                        vector<c_channel_def>::iterator i3;
                        for(i3=r_channel_defs.begin(); i3!=r_channel_defs.end(); ++i3) {
                            if(grant_all_groups) {
                                access_grant_channel.push_back((*i3).channel_name);
                                continue;
                            }
                            vector<string>::iterator i4;
                            for(i4=(*i3).groups.begin(); i4!=(*i3).groups.end(); ++i4) {
                                if(!(*i4).compare(*i2)) {
                                    access_grant_channel.push_back((*i3).channel_name);
                                    break;
                                }
                            }
                        }
                        if(grant_all_groups)
                            break;
                    }
                }
                access_grant_partyline=(*i1).access_grant_partyline;

                access_to_backup=(*i1).access_to_backup;
                access_grant_backup=(*i1).access_grant_backup;
                access_to_rehash=(*i1).access_to_rehash;
                access_grant_rehash=(*i1).access_grant_rehash;

                access_to_plususer=(*i1).access_to_plususer;
                access_grant_plususer=(*i1).access_grant_plususer;

                access_to_plusproc=(*i1).access_to_plusproc;
                access_grant_plusproc=(*i1).access_grant_plusproc;

                partyline_msg_flood=(*i1).partyline_msg_flood;

                access_to_replication=(*i1).access_to_replication;
                access_grant_replication=(*i1).access_grant_replication;

                access_to_restart=(*i1).access_to_restart;
                access_grant_restart=(*i1).access_grant_restart;
                access_to_die=(*i1).access_to_die;
                access_grant_die=(*i1).access_grant_die;

                access_to_filesystem=(*i1).access_to_filesystem;
                access_grant_filesystem=(*i1).access_grant_filesystem;

                access_to_upgrade=(*i1).access_to_upgrade;
                access_grant_upgrade=(*i1).access_grant_upgrade;

                access_to_apply=(*i1).access_to_apply;
                access_grant_apply=(*i1).access_grant_apply;

                access_to_private=(*i1).access_to_private;
                access_grant_private=(*i1).access_grant_private;

                access_to_can_send_all_users=(*i1).access_to_can_send_all_users;
                access_grant_can_send_all_users=(*i1).access_grant_can_send_all_users;

                access_to_can_send_unknown_users=(*i1).access_to_can_send_unknown_users;
                access_grant_can_send_unknown_users=(*i1).access_grant_can_send_unknown_users;

                break;
            }
            break;
        }
    }

    logic_erase_duplicate_or_empty_string_in_vector(access_to_users);
    logic_erase_duplicate_or_empty_string_in_vector(access_to_groups);
    logic_erase_duplicate_or_empty_string_in_vector(access_grant_groups);
    logic_erase_duplicate_or_empty_string_in_vector(access_to_proc_groups);
    logic_erase_duplicate_or_empty_string_in_vector(access_grant_proc_groups);
    logic_erase_duplicate_or_empty_string_in_vector(access_grant_proc);
    logic_erase_duplicate_or_empty_string_in_vector(access_to_channel);
    logic_erase_duplicate_or_empty_string_in_vector(access_grant_channel);
    logic_erase_duplicate_or_empty_string_in_vector(access_to_procs);
    logic_erase_duplicate_or_empty_string_in_vector(access_to_chan_defs);
    logic_erase_duplicate_or_empty_string_in_vector(access_usage_proc);
    logic_erase_duplicate_or_empty_string_in_vector(access_usage_proc_groups);
    logic_erase_duplicate_or_empty_string_in_vector(access_to_users_also_terminators);
}

/*!
    \brief Check wheter user has access to the filesystem
    \author VooDooMan
    \version 1
    \date 2004
    \param user User name from "logic.txt"
    \return True if user has access to the filesystem
*/
bool logic_access_to_filesystem(const string& user)
{
    deque<c_user>::iterator i1;
    for(i1=r_users.begin(); i1!=r_users.end(); ++i1)
        if(!(*i1).name.compare(user))
            return (*i1).access_to_filesystem;
    return false;
}

/*!
    \brief Retrieves full whois list for users of mask param1 (".whois" command on partyline)
    \author VooDooMan
    \version 2
    \date 2004, 2005-10-02
    \param lang Language-identification string
    \param param1 Mask (e.g. "Voo*Man") of user as defined in logic.txt, or "*!ident@host" mask
    \param lines Returns user-friendly whois listing
    \return Returns number of matching users
*/
int logic_partyline_whois(const string& lang, string param1, list<string>& lines)
{
    int cnt=0;
    param1=str_to_upper(param1);
    deque<c_user>::iterator i;
    for(i=r_users.begin(); i!=r_users.end(); ++i) {
        bool got=false;
        if(!match(param1.c_str(),str_to_upper((*i).name).c_str()))
            got=true;
        vector<string>::iterator i2;
        if(!got)
            for(i2=(*i).hostmask.begin(); i2!=(*i).hostmask.end(); ++i2)
                if(!match(param1.c_str(),str_to_upper(*i2).c_str()) || !match(str_to_upper(*i2).c_str(),param1.c_str()))
                    got=true;
        if(got) {
            string l=lang_get_string(1,lang,9)+" "+(*i).name+" "+lang_get_string(1,lang,10)+" ";
            if((*i).host_unknown)
                l+=lang_get_string(1,lang,11)+" ";
            if((*i).host_bot)
                l+=lang_get_string(1,lang,12)+" ";
            l+=lang_get_string(1,lang,13);
            lines.push_back(l);
            vector<string>::iterator i3;
            bool got=false;
            for(i3=(*i).hostmask.begin(); i3!=(*i).hostmask.end(); ++i3) {
                lines.push_back((string)"    \""+*i3+"\"");
                got=true;
            }
            if(!got)
                lines.push_back(lang_get_string(1,lang,14));
            l=lang_get_string(1,lang,15);
            lines.push_back(l);
            got=false;
            for(i3=(*i).fullname.begin(); i3!=(*i).fullname.end(); ++i3) {
                lines.push_back((string)"    \""+*i3+"\"");
                got=true;
            }
            if(!got)
                lines.push_back(lang_get_string(1,lang,14));
            l=lang_get_string(1,lang,18);
            lines.push_back(l);
            vector<string>::iterator i4;
            got=false;
            for(i4=(*i).groups.begin(); i4!=(*i).groups.end(); ++i4) {
                lines.push_back((string)"    \""+(*i4)+"\"");
                got=true;
            }
            if(!got)
                lines.push_back(lang_get_string(1,lang,14));
            l=lang_get_string(1,lang,19);
            lines.push_back(l);
            vector<c_dcc_host>::iterator i5;
            got=false;
            for(i5=(*i).dcc_hosts.begin(); i5!=(*i).dcc_hosts.end(); ++i5) {
                lines.push_back((string)"    "+lang_get_string(1,lang,20)+tos((*i5).group)+" "+lang_get_string(1,lang,21)+" "+(*i5).host);
                got=true;
            }
            if(!got)
                lines.push_back(lang_get_string(1,lang,14));
            l=lang_get_string(1,lang,22);
            lines.push_back(l);
            got=false;
            bool all=false;
            for(i3=(*i).access_to_group.begin(); i3!=(*i).access_to_group.end(); ++i3) {
                if(!(*i3).compare("*")) {
                    all=true;
                    got=true;
                    break;
                }
                lines.push_back((string)"    \""+*i3+"\"");
                got=true;
            }
            if(!got)
                lines.push_back(lang_get_string(1,lang,14));
            if(all)
                lines.push_back(lang_get_string(1,lang,166));
            l=lang_get_string(1,lang,23);
            lines.push_back(l);
            got=false;
            all=false;
            for(i3=(*i).access_grant_group.begin(); i3!=(*i).access_grant_group.end(); ++i3) {
                if(!(*i3).compare("*")) {
                    all=true;
                    got=true;
                    break;
                }
                lines.push_back((string)"    \""+*i3+"\"");
                got=true;
            }
            if(!got)
                lines.push_back(lang_get_string(1,lang,14));
            if(all)
                lines.push_back(lang_get_string(1,lang,166));
            l=lang_get_string(1,lang,24);
            lines.push_back(l);
            got=false;
            all=false;
            for(i3=(*i).access_to_proc.begin(); i3!=(*i).access_to_proc.end(); ++i3) {
                if(!(*i3).compare("*")) {
                    all=true;
                    got=true;
                    break;
                }
                lines.push_back((string)"    \""+*i3+"\"");
                got=true;
            }
            if(!got)
                lines.push_back(lang_get_string(1,lang,14));
            if(all)
                lines.push_back(lang_get_string(1,lang,166));
            l=lang_get_string(1,lang,25);
            lines.push_back(l);
            got=false;
            all=false;
            for(i3=(*i).access_grant_procedure.begin(); i3!=(*i).access_grant_procedure.end(); ++i3) {
                if(!(*i3).compare("*")) {
                    all=true;
                    got=true;
                    break;
                }
                lines.push_back((string)"    \""+*i3+"\"");
                got=true;
            }
            if(!got)
                lines.push_back(lang_get_string(1,lang,14));
            if(all)
                lines.push_back(lang_get_string(1,lang,166));
            l=lang_get_string(1,lang,589);
            lines.push_back(l);
            got=false;
            all=false;
            for(i3=(*i).access_usage_procedure.begin(); i3!=(*i).access_usage_procedure.end(); ++i3) {
                if(!(*i3).compare("*")) {
                    all=true;
                    got=true;
                    break;
                }
                lines.push_back((string)"    \""+*i3+"\"");
                got=true;
            }
            if(!got)
                lines.push_back(lang_get_string(1,lang,14));
            if(all)
                lines.push_back(lang_get_string(1,lang,166));
            l=lang_get_string(1,lang,30);
            lines.push_back(l);
            got=false;
            all=false;
            for(i3=(*i).access_to_channel.begin(); i3!=(*i).access_to_channel.end(); ++i3) {
                if(!(*i3).compare("*")) {
                    all=true;
                    break;
                }
                lines.push_back((string)"    \""+*i3+"\"");
                got=true;
            }
            if(!got)
                lines.push_back(lang_get_string(1,lang,14));
            if(all)
                lines.push_back(lang_get_string(1,lang,166));
            l=lang_get_string(1,lang,31);
            lines.push_back(l);
            got=false;
            all=false;
            for(i3=(*i).access_grant_channel.begin(); i3!=(*i).access_grant_channel.end(); ++i3) {
                if(!(*i3).compare("*")) {
                    all=true;
                    got=true;
                    break;
                }
                lines.push_back((string)"    \""+*i3+"\"");
                got=true;
            }
            if(!got)
                lines.push_back(lang_get_string(1,lang,14));
            if(all)
                lines.push_back(lang_get_string(1,lang,166));

            if((*i).access_to_partyline)
                l=lang_get_string(1,lang,16);
                else
                l=lang_get_string(1,lang,17);
            lines.push_back(l);

            if((*i).access_grant_partyline)
                l=lang_get_string(1,lang,41);
                else
                l=lang_get_string(1,lang,42);
            lines.push_back(l);

            l=lang_get_string(1,lang,202)+" "+(*i).partyline_msg_flood.make_string();
            lines.push_back(l);

            if((*i).access_grant_partyline)
                l=lang_get_string(1,lang,394);
                else
                l=lang_get_string(1,lang,395);
            lines.push_back(l);

            if((*i).access_to_backup)
                l=lang_get_string(1,lang,175);
                else
                l=lang_get_string(1,lang,176);
            lines.push_back(l);

            if((*i).access_grant_backup)
                l=lang_get_string(1,lang,177);
                else
                l=lang_get_string(1,lang,178);
            lines.push_back(l);

            if((*i).access_to_rehash)
                l=lang_get_string(1,lang,179);
                else
                l=lang_get_string(1,lang,180);
            lines.push_back(l);

            if((*i).access_grant_rehash)
                l=lang_get_string(1,lang,181);
                else
                l=lang_get_string(1,lang,182);
            lines.push_back(l);

            if((*i).access_to_restart)
                l=lang_get_string(1,lang,256);
                else
                l=lang_get_string(1,lang,257);
            lines.push_back(l);

            if((*i).access_grant_restart)
                l=lang_get_string(1,lang,258);
                else
                l=lang_get_string(1,lang,259);
            lines.push_back(l);

            if((*i).access_to_die)
                l=lang_get_string(1,lang,263);
                else
                l=lang_get_string(1,lang,264);
            lines.push_back(l);

            if((*i).access_grant_die)
                l=lang_get_string(1,lang,265);
                else
                l=lang_get_string(1,lang,266);
            lines.push_back(l);

            if((*i).access_to_plususer)
                l=lang_get_string(1,lang,185);
                else
                l=lang_get_string(1,lang,184);
            lines.push_back(l);

            if((*i).access_grant_plususer)
                l=lang_get_string(1,lang,186);
                else
                l=lang_get_string(1,lang,187);
            lines.push_back(l);

            if((*i).access_to_plusproc)
                l=lang_get_string(1,lang,194);
                else
                l=lang_get_string(1,lang,193);
            lines.push_back(l);

            if((*i).access_grant_plusproc)
                l=lang_get_string(1,lang,195);
                else
                l=lang_get_string(1,lang,196);
            lines.push_back(l);

            if((*i).access_to_replication)
                l=lang_get_string(1,lang,208);
                else
                l=lang_get_string(1,lang,209);
            lines.push_back(l);

            if((*i).access_grant_replication)
                l=lang_get_string(1,lang,598);
                else
                l=lang_get_string(1,lang,599);
            lines.push_back(l);

            if((*i).access_to_filesystem)
                l=lang_get_string(1,lang,390);
                else
                l=lang_get_string(1,lang,391);
            lines.push_back(l);

            if((*i).access_grant_filesystem)
                l=lang_get_string(1,lang,392);
                else
                l=lang_get_string(1,lang,393);
            lines.push_back(l);

            if((*i).access_to_private)
                l=lang_get_string(1,lang,490);
                else
                l=lang_get_string(1,lang,489);
            lines.push_back(l);

            if(!(*i).access_grant_private)
                l=lang_get_string(1,lang,491);
                else
                l=lang_get_string(1,lang,492);
            lines.push_back(l);

            if((*i).access_to_upgrade)
                l=lang_get_string(1,lang,621);
                else
                l=lang_get_string(1,lang,622);
            lines.push_back(l);

            if((*i).access_grant_upgrade)
                l=lang_get_string(1,lang,623);
                else
                l=lang_get_string(1,lang,624);
            lines.push_back(l);

            if((*i).access_to_apply)
                l=lang_get_string(1,lang,683);
                else
                l=lang_get_string(1,lang,684);
            lines.push_back(l);

            if((*i).access_grant_apply)
                l=lang_get_string(1,lang,685);
                else
                l=lang_get_string(1,lang,686);
            lines.push_back(l);

            if(!(*i).access_to_can_send_all_users)
                l=lang_get_string(1,lang,493);
                else
                l=lang_get_string(1,lang,494);
            lines.push_back(l);

            if(!(*i).access_grant_can_send_all_users)
                l=lang_get_string(1,lang,495);
                else
                l=lang_get_string(1,lang,496);
            lines.push_back(l);

            if(!(*i).access_to_can_send_unknown_users)
                l=lang_get_string(1,lang,497);
                else
                l=lang_get_string(1,lang,498);
            lines.push_back(l);

            if(!(*i).access_grant_can_send_unknown_users)
                l=lang_get_string(1,lang,499);
                else
                l=lang_get_string(1,lang,500);
            lines.push_back(l);

            vector<c_channel>::iterator i6;
            for(i6=r_channels.begin(); i6!=r_channels.end(); ++i6) {
                if(!(*i6).username.compare((*i).name)) {
                    l=lang_get_string(1,lang,26)+" "+(*i6).channel_name;
                    lines.push_back(l);
                    l=(string)"    "+lang_get_string(1,lang,27);
                    lines.push_back(l);
                    got=false;
                    for(i4=(*i6).groups.begin(); i4!=(*i6).groups.end(); ++i4) {
                        l="    \"";
                        l+=(*i4)+"\"";
                        lines.push_back(l);
                        got=true;
                    }
                    if(!got)
                        lines.push_back(lang_get_string(1,lang,14));
                    l=(string)"    "+lang_get_string(1,lang,28)+" on_op = "+(*i6).on_op;
                    lines.push_back(l);
                    l=(string)"    "+lang_get_string(1,lang,28)+" on_deop = "+(*i6).on_deop;
                    lines.push_back(l);
                    l=(string)"    "+lang_get_string(1,lang,28)+" on_ban = "+(*i6).on_ban;
                    lines.push_back(l);
                    l=(string)"    "+lang_get_string(1,lang,28)+" on_unban = "+(*i6).on_unban;
                    lines.push_back(l);
                    l=(string)"    "+lang_get_string(1,lang,28)+" on_kick = "+(*i6).on_kick;
                    lines.push_back(l);
                    l=(string)"    "+lang_get_string(1,lang,28)+" on_voice = "+(*i6).on_voice;
                    lines.push_back(l);
                    l=(string)"    "+lang_get_string(1,lang,28)+" on_devoice = "+(*i6).on_devoice;
                    lines.push_back(l);
                    l=(string)"    "+lang_get_string(1,lang,28)+" on_join = "+(*i6).on_join;
                    lines.push_back(l);
                    l=(string)"    "+lang_get_string(1,lang,28)+" on_banned = "+(*i6).on_banned;
                    lines.push_back(l);
                    l=(string)"    "+lang_get_string(1,lang,28)+" on_flood = "+(*i6).on_flood;
                    lines.push_back(l);
                    l=(string)"    "+lang_get_string(1,lang,28)+" on_privmsg = "+(*i6).on_privmsg;
                    lines.push_back(l);
                    l=(string)"    "+lang_get_string(1,lang,28)+" on_notice = "+(*i6).on_notice;
                    lines.push_back(l);
                    l=(string)"    "+lang_get_string(1,lang,28)+" on_except = "+(*i6).on_except;
                    lines.push_back(l);
                    l=(string)"    "+lang_get_string(1,lang,28)+" on_unexcept = "+(*i6).on_unexcept;
                    lines.push_back(l);
                    l=(string)"    "+lang_get_string(1,lang,28)+" on_invite = "+(*i6).on_invite;
                    lines.push_back(l);
                    l=(string)"    "+lang_get_string(1,lang,28)+" on_uninvite = "+(*i6).on_uninvite;
                    lines.push_back(l);
                    l=(string)"    "+lang_get_string(1,lang,28)+" on_not_invited = "+(*i6).on_not_invited;
                    lines.push_back(l);
                    l=(string)"    "+lang_get_string(1,lang,28)+" on_not_in_reop = "+(*i6).on_not_in_reop;
                    lines.push_back(l);
                    l=(string)"    "+lang_get_string(1,lang,28)+" on_reop = "+(*i6).on_reop;
                    lines.push_back(l);
                    l=(string)"    "+lang_get_string(1,lang,28)+" on_other_mode = "+(*i6).on_other_mode;
                    lines.push_back(l);
                    l=(string)"    "+lang_get_string(1,lang,28)+" on_other_mode_mask = "+(*i6).on_other_mode_mask;
                    lines.push_back(l);

                    l=(string)"    "+"msg_flood "+(*i6).msg_flood.make_string();
                    lines.push_back(l);

                    l=(string)"    "+"notice_flood "+(*i6).notice_flood.make_string();
                    lines.push_back(l);

                    l=(string)"    "+"repeat_flood "+(*i6).repeat_flood.make_string();
                    lines.push_back(l);

                    l=(string)"    "+"nick_flood "+(*i6).nick_flood.make_string();
                    lines.push_back(l);

                    l=(string)"    "+"join_flood "+(*i6).join_flood.make_string();
                    lines.push_back(l);

                    l=(string)"    "+"mode_flood "+(*i6).mode_flood.make_string();
                    lines.push_back(l);

                    l=(string)"    "+"ctcp_flood "+(*i6).ctcp_flood.make_string();
                    lines.push_back(l);

                    l=(string)"    "+"dynamic1 (plus modes)  "+(*i6).dynamic_plus_modes;
                    lines.push_back(l);
                    l=(string)"    "+"dynamic2 (minus modes) "+(*i6).dynamic_minus_modes;
                    lines.push_back(l);

                    if(!(*i6).can_send_unknown_users)
                        l=(string)"    "+lang_get_string(1,lang,501);
                    else
                        l=(string)"    "+lang_get_string(1,lang,502);
                    lines.push_back(l);
                }
            }

            lines.push_back("----");
            cnt++;
        }
    }
    return cnt;
}

/*!
    \brief Checks if such user defined in logic.txt exists
    \author VooDooMan
    \version 2
    \date 2004, 2005-10-02
    \param name User name as defined in logic.txt
    \return Returns true if such user exists
*/
bool logic_partyline_user_exists(string name)
{
    name=str_to_upper(name);
    bool res=false;
    deque<c_user>::iterator i;
    for(i=r_users.begin(); i!=r_users.end(); ++i) {
        if(!str_to_upper((*i).name).compare(name)) {
            res=true;
            break;
        }
    }
    return res;
}

/*!
    \brief Retrieves list of known channels (channel must reflect at least one time in logic.txt)
    \author VooDooMan
    \version 1
    \date 2004
    \param channels Returns channel list
*/
void logic_partyline_get_channels(vector<string>& channels)
{
    channels.clear();
    vector<c_channel>::iterator i;
    for(i=r_channels.begin(); i!=r_channels.end(); ++i)
        channels.push_back((*i).channel_name);
    for(i=r_channel_templates.begin(); i!=r_channel_templates.end(); ++i)
        channels.push_back((*i).channel_name);
}

/*!
    \brief Gets channel definition
    \author VooDooMan
    \version 1
    \date 2004
    \param channel_name Name of the channel
    \param chd Channel definition (case-insensitive)
    \return Returns true if channel was found in pre-loaded "logic.txt"
*/
bool logic_partyline_get_channel_def(const string& channel_name, c_channel_def& chd)
{
    vector<c_channel_def>::iterator i;
    for(i=r_channel_defs.begin(); i!=r_channel_defs.end(); ++i) {
        if(!irc_cmp_strings_case_insensitive((*i).channel_name,channel_name)) {
            chd=*i;
            return true;
        }
    }
    return false;
}

/*!
    \brief Deletes channel definition
    \author VooDooMan
    \version 1
    \date 2004
    \param channel_name Name of the channel
*/
void logic_partyline_remove_channel_def(const string& channel_name)
{
    vector<c_channel_def>::iterator i;
    for(i=r_channel_defs.begin(); i!=r_channel_defs.end(); ++i) {
        if(!(*i).channel_name.compare(channel_name)) {
            r_channel_defs.erase(i);
            break;
        }
    }
}

/*!
    \brief Adds a channel definition
    \author VooDooMan
    \version 1
    \date 2004
    \param chd Channel definition
    \return Returns always true
*/
bool logic_partyline_add_channel_def(c_channel_def& chd)
{
    r_channel_defs.push_back(chd);

    return true;
}

/*!
    \brief Adds an user with channels definition
    \author VooDooMan
    \version 1
    \date 2004
    \param user User definition
    \param channels User's channels
*/
void logic_partyline_add_user(c_user& user, vector<c_channel>& channels)
{
    if(user.is_template) {
        user.hostmask.clear();
        user.fullname.clear();
        vector<c_channel>::iterator i2;
l2:
        for(i2=r_channel_templates.begin(); i2!=r_channel_templates.end(); ++i2) {
            if(!(*i2).username.compare(user.name)) {
                r_channel_templates.erase(i2);
                goto l2;
            }
        }

        //my_time(&user.last_changed);
        r_user_templates.push_back(user);

        for(i2=channels.begin(); i2!=channels.end(); ++i2) {
            (*i2).is_template=true;
            /*if((*i2).channel_name.length()>0 && (*i2).channel_name[0]!='"')
                (*i2).channel_name="\""+(*i2).channel_name+"\"";*/
            r_channel_templates.push_back(*i2);
        }
    } else {
        vector<c_channel>::iterator i2;
l1:
        for(i2=r_channels.begin(); i2!=r_channels.end(); ++i2) {
            if(!(*i2).username.compare(user.name)) {
                r_channels.erase(i2);
                goto l1;
            }
        }

        //my_time(&user.last_changed);

        if(!user.terminator) {
            r_users.push_back(user);
            for(i2=channels.begin(); i2!=channels.end(); ++i2) {
                (*i2).is_template=false;
                (*i2).terminator=false;
                /*if((*i2).channel_name.length()>0 && (*i2).channel_name[0]!='"')
                    (*i2).channel_name="\""+(*i2).channel_name+"\"";*/
                r_channels.push_back(*i2);
            }
        } else {
            r_terminators.push_back(user);
            for(i2=channels.begin(); i2!=channels.end(); ++i2) {
                (*i2).is_template=false;
                (*i2).terminator=true;
                /*if((*i2).channel_name.length()>0 && (*i2).channel_name[0]!='"')
                    (*i2).channel_name="\""+(*i2).channel_name+"\"";*/
                r_channel_terminators.push_back(*i2);
            }
        }
    }
}

/*!
    \brief Removes user or user template defined in logic.txt
    \author VooDooMan
    \version 2
    \date 2004, 2007-05-16
    \param username Name of user as define in logic.txt
*/
void logic_partyline_remove_user(const string& username)
{
    deque<c_user>::iterator i1;
    bool template_=false;
l1:
    for(i1=r_users.begin(); i1!=r_users.end(); ++i1) {
        if(!(*i1).name.compare(username)) {
            if((*i1).is_template)
                template_=true;
            r_users.erase(i1);
            goto l1;
        }
    }
    if(template_) {
        vector<c_channel>::iterator i2;
l2:
        for(i2=r_channel_templates.begin(); i2!=r_channel_templates.end(); ++i2) {
            if(!(*i2).username.compare(username)) {
                r_channel_templates.erase(i2);
                goto l2;
            }
        }
    }
l3:
    for(i1=r_terminators.begin(); i1!=r_terminators.end(); ++i1) {
        if(!(*i1).name.compare(username)) {
            r_terminators.erase(i1);
            goto l3;
        }
    }
    vector<c_channel>::iterator i2;
l4:
    for(i2=r_channel_terminators.begin(); i2!=r_channel_terminators.end(); ++i2) {
        if(!(*i2).username.compare(username)) {
            r_channel_terminators.erase(i2);
            goto l4;
        }
    }
l5:
    for(i2=r_channels.begin(); i2!=r_channels.end(); ++i2) {
        if(!(*i2).username.compare(username)) {
            r_channels.erase(i2);
            goto l5;
        }
    }
}

/*!
    \brief Lists procedure's commands
    \author VooDooMan
    \version 1
    \date 2004
    \param proc_name Name of procedure with full declaration
    \param commands Returns commands
    \param declaration Returns full declaration
    \param groups Returns list of groups which is the procedure member of
    \param replications Returns replication rules
    \param remote_call Returns remote call rules
*/
void logic_list_proc(string proc_name, list<string>& commands, string& declaration, vector<string>& groups, vector<c_replication>& replications, vector<string>& remote_call)
{
    commands.clear();
    groups.clear();
    replications.clear();
    remote_call.clear();

    string ln;
    int ii;
    int indent=1;

    if(proc_name.find("(",0)!=string::npos)
        proc_name.erase(proc_name.find("(",0),proc_name.length()-proc_name.find("(",0));

    c_module* core_mod=get_module("core",true);
    if(core_mod) {
        vector<c_procedure>::iterator i2;
        for(i2=core_mod->r_procedures.begin(); i2!=core_mod->r_procedures.end(); ++i2) {

            string wo_decl=(*i2).name;
            if(wo_decl.find("(",0)!=string::npos)
                wo_decl.erase(wo_decl.find("(",0),wo_decl.length()-wo_decl.find("(",0));

            if(!wo_decl.compare(proc_name)) {
                vector<string>::iterator i4;
                for(i4=(*i2).groups.begin(); i4!=(*i2).groups.end(); ++i4)
                    groups.push_back(*i4);
                vector<c_replication>::iterator i5;
                for(i5=(*i2).replication.begin(); i5!=(*i2).replication.end(); ++i5)
                    replications.push_back(*i5);
                vector<c_rproc>::iterator i6;
                for(i6=(*i2).rproc.begin(); i6!=(*i2).rproc.end(); ++i6)
                    remote_call.push_back((*i6).remote_bot);
                declaration=(*i2).name;
                deque<c_proc_command>::iterator i3;
                for(i3=(*i2).commands.begin(); i3!=(*i2).commands.end(); ++i3) {
                    ln=(*i3).dump(core_mod,indent); // always core module
                    vector<string> lines;
                    explode(ln,"\n",lines);
                    vector<string>::iterator i4;
                    for(i4=lines.begin(); i4!=lines.end(); ++i4)
                        commands.push_back(*i4);
                }
                while(indent>0) {
                    indent--;
                    ln=""; for(ii=0; ii<indent; ++ii) ln+="    ";
                    ln+="}";
                    commands.push_back(ln);
                }
                break;
            }
        }
    }
}

/*!
    \brief Lists functions's commands
    \author VooDooMan
    \version 2
    \date 2004, 2005-10-29
    \param func_name Name of function with full declaration
    \param commands Returns commands
    \param declaration Returns full declaration
    \param groups Returns list of groups which is the procedure member of
    \param replications Returns replication rules
    \param remote_call Returns remote call rules
*/
void logic_list_func(string func_name, list<string>& commands, string& declaration, vector<string>& groups, vector<c_replication>& replications, vector<string>& remote_call)
{
    commands.clear();
    groups.clear();
    replications.clear();
    remote_call.clear();

    string ln;
    //int ii;
    //int indent=1;

    if(func_name.find("(",0)!=string::npos)
        func_name.erase(func_name.find("(",0),func_name.length()-func_name.find("(",0));

    c_module* core_mod=get_module("core",true);
    if(core_mod) {
        map<string,c_function*>::iterator i2;
        for(i2=core_mod->r_functions.functions.begin(); i2!=core_mod->r_functions.functions.end(); ++i2) {

            // .first is already w/o declaration

            if(!(*i2).first.compare(func_name)) {
                {
                    vector<string>::iterator i4;
                    for(i4=(*i2).second->groups.begin(); i4!=(*i2).second->groups.end(); ++i4)
                        groups.push_back(*i4);
                }
                {
                    vector<c_replication>::iterator i4;
                    for(i4=(*i2).second->replications.begin(); i4!=(*i2).second->replications.end(); ++i4)
                        replications.push_back(*i4);
                }
                {
                    vector<string>::iterator i6;
                    for(i6=(*i2).second->remote_bot_call.begin(); i6!=(*i2).second->remote_bot_call.end(); ++i6)
                        remote_call.push_back(*i6);
                }
                declaration=(*i2).second->declaration;
                string lines=(*i2).second->dump_to_string();
                // remove declaration and first bracket; and last bracket
                if(lines.find("{",0)!=string::npos)
                    lines.erase(0,lines.find("{",0)+1);
                if(lines.rfind("}",string::npos)!=string::npos)
                    lines.erase(lines.rfind("}",string::npos),lines.length()-lines.rfind("}",string::npos));
                vector<string> vlines;
                explode(lines,'\n',vlines,false);
                vector<string>::iterator i4;
                for(i4=vlines.begin(); i4!=vlines.end(); ++i4)
                    commands.push_back(*i4);
                break;
            }
        }
    }
}

/*!
    \brief Retrieves an user with channels
    \author VooDooMan
    \version 3
    \date 2005, 2006-01-13 (Friday)
    \param user_name Name of user as define in logic.txt
    \param user Returns user definition
    \param channels Returns list of channels
*/
void logic_partyline_get_user(const string& user_name, c_user& user, vector<c_channel>& channels)
{
    channels.clear();
    deque<c_user>::iterator i;
    deque<c_user>* usrs=NULL;
    for(int iii=0; iii<2; ++iii) {
        switch(iii) {
            case 0:
                usrs=&r_users;
                break;
            case 1:
                usrs=&r_terminators;
                break;
        }
        for(i=usrs->begin(); i!=usrs->end(); ++i) {
            if(!(*i).name.compare(user_name)) {
                user=*i;
                vector<c_string> all_channels;

                vector<c_channel>::iterator i3;
                if(iii==0) {
                    for(i3=r_channels.begin(); i3!=r_channels.end(); ++i3) {
                        all_channels.push_back((*i3).channel_name);

                        if(!(*i3).username.compare((*i).name)) {
                            c_channel ch=*i3;
                            ch.configured=true;
                            channels.push_back(ch);
                        }
                    }
                } else {
                    for(i3=r_channel_terminators.begin(); i3!=r_channel_terminators.end(); ++i3) {
                        all_channels.push_back((*i3).channel_name);

                        if(!(*i3).username.compare((*i).name)) {
                            c_channel ch=*i3;
                            ch.configured=true;
                            channels.push_back(ch);
                        }
                    }
                }

                {
                    sort(all_channels.begin(),all_channels.end());
                    vector<c_string>::iterator i1=unique(all_channels.begin(),all_channels.end());
                    all_channels.erase(i1,all_channels.end());
                }

            l1:
                if(iii==0) {
                    for(i3=r_channels.begin(); i3!=r_channels.end(); ++i3) {
                        if(!(*i3).username.compare((*i).name)) {
                            vector<c_string>::iterator i4;
                            for(i4=all_channels.begin(); i4!=all_channels.end(); ++i4) {
                                if(!(*i4).compare((*i3).channel_name)) {
                                    all_channels.erase(i4);
                                    goto l1;
                                }
                            }
                        }
                    }
                } else {
                    for(i3=r_channel_terminators.begin(); i3!=r_channel_terminators.end(); ++i3) {
                        if(!(*i3).username.compare((*i).name)) {
                            vector<c_string>::iterator i4;
                            for(i4=all_channels.begin(); i4!=all_channels.end(); ++i4) {
                                if(!(*i4).compare((*i3).channel_name)) {
                                    all_channels.erase(i4);
                                    goto l1;
                                }
                            }
                        }
                    }
                }
                vector<string>::iterator i1;
                vector<c_string>::iterator i4;
                for(i4=all_channels.begin(); i4!=all_channels.end(); ++i4) {
                    for(i1=(*i).access_to_channel.begin(); i1!=(*i).access_to_channel.end(); ++i1) {
                        if(!(*i4).compare(*i1)) {
                            c_channel ch;
                            ch.terminator=(1==iii);
                            ch.channel_name=(*i4).str;
                            channels.push_back(ch);
                        }
                    }
                }
                return;
            }
        }
    }
    user.clear();
    channels.clear();
}

/*!
    \brief Retrieves an user template with channels
    \author VooDooMan
    \version 2
    \date 2004, 2006-01-13 (Friday)
    \param user_name Name of user template as define in logic.txt
    \param user Returns user template definition
    \param channels Returns list of channels
*/
void logic_partyline_get_user_template(const string& user_name, c_user& user, vector<c_channel>& channels)
{
    channels.clear();
    deque<c_user>::iterator i;
    for(i=r_user_templates.begin(); i!=r_user_templates.end(); ++i) {
        if(!(*i).name.compare(user_name)) {
            user=*i;
            user.is_template=false;
            vector<c_string> all_channels;

            vector<c_channel>::iterator i3;
            for(i3=r_channel_templates.begin(); i3!=r_channel_templates.end(); ++i3) {
                all_channels.push_back((*i3).channel_name);

                if(!(*i3).username.compare((*i).name)) {
                    c_channel ch=*i3;
                    ch.configured=true;
                    ch.is_template=false;
                    channels.push_back(ch);
                }
            }

            {
                sort(all_channels.begin(),all_channels.end());
                vector<c_string>::iterator i1=unique(all_channels.begin(),all_channels.end());
                all_channels.erase(i1,all_channels.end());
            }

        l1:
            for(i3=r_channel_templates.begin(); i3!=r_channel_templates.end(); ++i3) {
                if(!(*i3).username.compare((*i).name)) {
                    vector<c_string>::iterator i4;
                    for(i4=all_channels.begin(); i4!=all_channels.end(); ++i4) {
                        if(!(*i4).compare((*i3).channel_name)) {
                            all_channels.erase(i4);
                            goto l1;
                        }
                    }
                }
            }
            vector<string>::iterator i1;
            vector<c_string>::iterator i4;
            for(i4=all_channels.begin(); i4!=all_channels.end(); ++i4) {
                for(i1=(*i).access_to_channel.begin(); i1!=(*i).access_to_channel.end(); ++i1) {
                    if(!(*i4).compare(*i1)) {
                        c_channel ch;
                        ch.is_template=true;
                        ch.channel_name=(*i4).str;
                        channels.push_back(ch);
                    }
                }
            }
            return;
        }
    }
    user.clear();
}

/*!
    \brief Retrieves flood limits
    \author VooDooMan
    \version 1
    \date 2004
    \param channel Name of channel
    \param user Name of user as defined in logic.txt
    \param msg_flood Returns PRIVMSG flood limit
    \param notice_flood Returns NOTICE flood limit
    \param repeat_flood Returns repeat flood limit
    \param nick_flood Returns NICK flood limit
    \param join_flood Returns JOIN flood limit
    \param mode_flood Returns MODE flood limit
    \param ctcp_flood Returns CTCP flood limit
*/
void logic_get_flood_limits(const string& channel, const string& user, c_flood& msg_flood, c_flood& notice_flood, c_flood& repeat_flood, c_flood& nick_flood, c_flood& join_flood, c_flood& mode_flood, c_flood& ctcp_flood)
{
    msg_flood.lines=0;
    notice_flood.lines=0;
    repeat_flood.lines=0;
    nick_flood.lines=0;
    join_flood.lines=0;
    mode_flood.lines=0;
    ctcp_flood.lines=0;

    vector<c_channel>::iterator i1;
    for(i1=r_channels.begin(); i1!=r_channels.end(); ++i1) {
        if(!user.compare((string)"") && !irc_cmp_strings_case_insensitive((*i1).channel_name,channel) && (*i1).host_unknown) {
            msg_flood=(*i1).msg_flood;
            notice_flood=(*i1).notice_flood;
            repeat_flood=(*i1).repeat_flood;
            nick_flood=(*i1).nick_flood;
            join_flood=(*i1).join_flood;
            mode_flood=(*i1).mode_flood;
            ctcp_flood=(*i1).ctcp_flood;
            return;
        }
        if(!irc_cmp_strings_case_insensitive((*i1).channel_name,channel) && !(*i1).username.compare(user)) {
            msg_flood=(*i1).msg_flood;
            notice_flood=(*i1).notice_flood;
            repeat_flood=(*i1).repeat_flood;
            nick_flood=(*i1).nick_flood;
            join_flood=(*i1).join_flood;
            mode_flood=(*i1).mode_flood;
            ctcp_flood=(*i1).ctcp_flood;
            return;
        }
    }

    // consider as unknown host
    for(i1=r_channels.begin(); i1!=r_channels.end(); ++i1) {
        if(!irc_cmp_strings_case_insensitive((*i1).channel_name,channel) && (*i1).host_unknown) {
            msg_flood=(*i1).msg_flood;
            notice_flood=(*i1).notice_flood;
            repeat_flood=(*i1).repeat_flood;
            nick_flood=(*i1).nick_flood;
            join_flood=(*i1).join_flood;
            mode_flood=(*i1).mode_flood;
            ctcp_flood=(*i1).ctcp_flood;
            return;
        }
    }
}

vector<c_nick_to_lag> nicks_to_lag;             //!< Array of CTCP requests statistics (for flood protection)

/*!
    \brief Checks wheter user (as known in logic.txt) has access to partyline
    \param user_name_as_in_logic User name as in logic.txt
    \return True, if has granted access
*/
bool logic_access_to_partyline(const string& user_name_as_in_logic)
{
    deque<c_user>::iterator i;
    for(i=r_users.begin(); i!=r_users.end(); ++i) {
        vector<string>::iterator is, ii;
        if(!(*i).host_unknown && !(*i).name.compare(user_name_as_in_logic)) {
            return (*i).access_to_partyline;
        }
    }

    // try host_unknown
    for(i=r_users.begin(); i!=r_users.end(); ++i) {
        vector<string>::iterator is, ii;
        if((*i).host_unknown) {
            return (*i).access_to_partyline;
        }
    }

    return false;
}

/*!
    \brief Checks wheter user (the bot) is replication partner
    \param user_name_as_in_logic User name as in logic.txt
    \return True, if user (the bot) is replication partner
*/
bool logic_is_replication_partner(const string& user_name_as_in_logic)
{
    deque<c_user>::iterator i;
    for(i=r_users.begin(); i!=r_users.end(); ++i) {
        vector<string>::iterator is, ii;
        if(!(*i).host_unknown && !(*i).name.compare(user_name_as_in_logic)) {
            return (*i).replication_partner;
        }
    }

    // try host_unknown
    for(i=r_users.begin(); i!=r_users.end(); ++i) {
        vector<string>::iterator is, ii;
        if((*i).host_unknown) {
            return (*i).replication_partner;
        }
    }

    return false;
}

/*!
    \brief Finds procedure by name (without declatarion)
    \param name Procedure name
    \return Returns full declaration of procedure
    \retval "" An empty string, if procedure was not found
*/
string logic_find_proc_by_name(const string& name)
{
    vector<c_procedure>::iterator i1;
    c_module* core_mod=get_module("core",true);
    if(core_mod) {
        for(i1=core_mod->r_procedures.begin(); i1!=core_mod->r_procedures.end(); ++i1) {
            string s=(*i1).name;
            unsigned int i2;
            for(i2=0; i2<s.length(); ++i2)
                if(s[i2]=='(')
                    break;
            s.erase(i2,s.length()-i2); // erase everything after '('
            if(!s.compare(name))
                return (*i1).name;
        }
    }
    return (string)"";
}

/*!
    \brief Gets user definition
    \author VooDooMan
    \version 1
    \date 2004
    \param name User name from "logic.txt"
    \param user Returns user
    \return Returns true if user was found
*/
bool logic_botnet_get_user(const string& name, c_user& user)
{
    deque<c_user>::iterator i1;
    for(i1=r_users.begin(); i1!=r_users.end(); ++i1) {
        if(!(*i1).name.compare(name)) {
            user=*i1;
            return true;
        }
    }
    for(i1=r_user_templates.begin(); i1!=r_user_templates.end(); ++i1) {
        if(!(*i1).name.compare(name)) {
            user=*i1;
            return true;
        }
    }
    for(i1=r_terminators.begin(); i1!=r_terminators.end(); ++i1) {
        if(!(*i1).name.compare(name)) {
            user=*i1;
            return true;
        }
    }
    return false;
}

/*!
    \brief Gets channel definition
    \author VooDooMan
    \version 2
    \date 2004, 2008-02-25
    \param channel_name Channel name
    \param chd Returns channel definition
    \return Returns true if channel was found
*/
bool logic_botnet_get_channel_def(const string& channel_name, c_channel_def& chd)
{
    vector<c_channel_def>::iterator i1;
    for(i1=r_channel_defs.begin(); i1!=r_channel_defs.end(); ++i1) {
        if(!cmp_strings_case_insensitive((*i1).channel_name,channel_name)) {
            chd=*i1;
            return true;
        }
    }
    return false;
}

/*!
    \brief Removes channel definition
    \author VooDooMan
    \version 2
    \date 2004, 2008-02-25
    \param channel_name Channel name
*/
void logic_botnet_remove_channel_def(const string& channel_name)
{
    vector<c_channel_def>::iterator i1;
    for(i1=r_channel_defs.begin(); i1!=r_channel_defs.end(); ++i1) {
        if(!cmp_strings_case_insensitive((*i1).channel_name,channel_name)) {
            r_channel_defs.erase(i1);
            break;
        }
    }
}

/*!
    \brief Gets procedure
    \author VooDooMan
    \version 1
    \date 2004
    \param name Procedure name
    \param proc Returns procedure
    \return Returns true if procedure was found
*/
bool logic_botnet_get_proc(string name, c_procedure& proc)
{
    if(name.find("(",0)!=string::npos)
        name.erase(name.find("(",0),name.length()-name.find("(",0));
    c_module* core_mod=get_module("core",true);
    if(core_mod) {
        vector<c_procedure>::iterator i1;
        for(i1=core_mod->r_procedures.begin(); i1!=core_mod->r_procedures.end(); ++i1) {
            string wo_decl=(*i1).name;
            if(wo_decl.find("(",0)!=string::npos)
                wo_decl.erase(wo_decl.find("(",0),wo_decl.length()-wo_decl.find("(",0));
            if(!wo_decl.compare(name)) {
                proc=*i1;
                return true;
            }
        }
    }
    return false;
}

/*!
    \brief Gets function
    \author VooDooMan
    \version 1
    \date 2005-11-17
    \param name Function name
    \param func Returns function
    \return Returns true if function was found
*/
bool logic_botnet_get_func(string name, c_function& func)
{
    if(name.find("(",0)!=string::npos)
        name.erase(name.find("(",0),name.length()-name.find("(",0));
    c_module* core_mod=get_module("core",true);
    if(core_mod) {
        map<string,c_function*>::iterator i1=core_mod->r_functions.functions.find(name);
        if(i1!=core_mod->r_functions.functions.end()) {
            func=c_function(*(*i1).second);
            return true;
        }
    }
    return false;
}

/*!
    \brief Gets channels that user has defined properties to
    \author VooDooMan
    \version 1
    \date 2004
    \param name Channel name
    \param channels Returns channels
    \return Returns true if user was found
*/
bool logic_botnet_get_user_channels(const string& name, vector<c_channel>& channels)
{
    channels.clear();

    deque<c_user>::iterator i1;

    for(i1=r_users.begin(); i1!=r_users.end(); ++i1) {
        if(!(*i1).name.compare(name)) {
            vector<c_channel>::iterator i2;
            for(i2=r_channels.begin(); i2!=r_channels.end(); ++i2) {
                if(!(*i2).username.compare((*i1).name))
                    channels.push_back(*i2);
            }
            return true;
        }
    }

    for(i1=r_user_templates.begin(); i1!=r_user_templates.end(); ++i1) {
        if(!(*i1).name.compare(name)) {
            vector<c_channel>::iterator i2;
            for(i2=r_channel_templates.begin(); i2!=r_channel_templates.end(); ++i2) {
                if(!(*i2).username.compare((*i1).name))
                    channels.push_back(*i2);
            }
            return true;
        }
    }

    for(i1=r_terminators.begin(); i1!=r_terminators.end(); ++i1) {
        if(!(*i1).name.compare(name)) {
            vector<c_channel>::iterator i2;
            for(i2=r_channels.begin(); i2!=r_channels.end(); ++i2) {
                if(!(*i2).username.compare((*i1).name))
                    channels.push_back(*i2);
            }
            for(i2=r_channel_terminators.begin(); i2!=r_channel_terminators.end(); ++i2) {
                if(!(*i2).username.compare((*i1).name))
                    channels.push_back(*i2);
            }
            return true;
        }
    }

    return false;
}

/*!
    \brief Writes new "logic.txt" file
    \author VooDooMan
    \version 1
    \date 2004
*/
void logic_botnet_backup(string who)
{
    c_module* core_mod=get_module("core",true);
    who+=" (BOTNET)";
    logic_partyline_backup(core_mod,who);
}

/*!
    \brief Removes user
    \author VooDooMan
    \version 2
    \date 2004, 2007-05-16
    \param username Name of user
*/
void logic_botnet_remove_user(const string& username)
{
    logic_partyline_remove_user(username);
}

/*!
    \brief Tries to rehash (reload and apply new configuration file logic.txt)
    \author VooDooMan
    \version 2
    \date 2004, 2005
    \param error_on_line If an error occurs, there will be returned line number of error in logic.txt, if succeeded, it is zero
    \param error_reason If an error occurs, there can be helping string with description on error; could be NULL, and is NULL for success
    \param who This should contain who caused rehasing (for log purposes - bot.log) - could contain also comment about name of user
    \param who_only_name_of_user Same as "who", but *only* plain name of user
    \return Returns true on error
    \retval true Error
    \retval false Success
    \warning For internal use only! Do not call it directly!
*/
// returns true if error
bool logic_rehash(int& error_on_line, string& error_reason, string UNUSEDARG(who), string who_only_name_of_user)
{
    log_bot("Trying to rehash...");
    error_on_line=1;
    error_reason="";
    logic_on_internal_event("@rehash@",who_only_name_of_user,"","","",PRIVATE_EVENT_SEVERITY_CODE_INFORMATIONAL,"Trying to rehash...","");
    bool res=true;
    c_module* core_mod=get_module("core",false);
    if(core_mod)
        res=core_mod->rehash(error_on_line,error_reason);
    if(res)
        logic_on_internal_event("@rehash_done@",who_only_name_of_user,"ERROR",tos(error_on_line).c_str(),"",PRIVATE_EVENT_SEVERITY_CODE_ERROR,error_reason,"");
    else
        logic_on_internal_event("@rehash_done@",who_only_name_of_user,"OK","","",PRIVATE_EVENT_SEVERITY_CODE_INFORMATIONAL,"Rehashing succeeded.","");
    return res;
}

/*!
    \brief Performs rollback to previous configuration from backup, if new logic.txt could not be successfully parsed
    \author VooDooMan
    \version 1
    \date 2004
    \param log Returns log from rollback operation
    \param lang Language code (e.g. "en") for DCC or telnet user (applies to log argument)
    \return Returns true if succeeded
    \retval true Success
    \retval false Fatal error: the bot should DIE, because there are no more backup files
*/
// returns true if okay
bool logic_rollback(vector<string>& log, const string& lang)
{
    log.clear();
    char tmp1[256];
    for(int i1=0; i1<=logic_get_max_logic_backups()-1; ++i1) {
        string private_event="";

        sprintf(tmp1,"%s%03d","logic.txt.bak",i1);
        FILE* test=fopen(tmp1,"r");
        if(test) {
            fclose(test);
            {
#if (defined(_WIN32) || defined(_WIN64)) && !defined(__CYGWIN__)
                mkdir("bad_conf");
#else
                mkdir("." FILE_SLASH "bad_conf",0777);
#endif
                my_time_t now=my_time(NULL);
                char tmp[128];
                strcpy(tmp,make_nice_time(now,false).c_str());
                for(int i1=0; tmp[i1]!=0; ++i1)
                    if(tmp[i1]==0x20 || tmp[i1]==':' || tmp[i1]=='/')
                        tmp[i1]='_';
                string bad="." FILE_SLASH "bad_conf" FILE_SLASH;
                bad+=tmp;
                bad+=".";
                char tmp2[64];
                unsigned int rnd=rand();
                sprintf(tmp2,"%u",rnd);
                bad+=tmp2;
                bad+=".logic.txt";
                rename("." FILE_SLASH "logic.txt",bad.c_str());
                string log_=lang_get_string(1,lang,543)+(string)" "+bad;
                log_=(string)"Bad logic.txt was stored to file: "+bad;

                private_event+=log_;

                log.push_back(log_);
                log_bot(log_.c_str());
                unlink("logic.txt"); // for sure's sake ;-)
            }
            rename(tmp1,"logic.txt");
            string tmp2=lang_get_string(1,lang,383)+tmp1;
            log.push_back(tmp2);
            tmp2=(string)"logic.txt rollback from backup file: "+tmp1;
            log_bot(tmp2);

            private_event+=(string)"; "+tmp2;

            {
                struct stat s;
                stat("logic.txt",&s);

                char tch[128];
                strcpy(tch,ctime(&s.st_mtime));
                if(strlen(tch) && tch[strlen(tch)-1]=='\n')
                    tch[strlen(tch)-1]=0;

                tmp2="   ";
                tmp2+=lang_get_string(1,lang,558);
                tmp2+=" ";
                tmp2+=tch;
                log.push_back(tmp2);
                tmp2="   Time of the file: ";
                tmp2+=tch;
                log_bot(tmp2);

                private_event+=(string)"; "+tmp2;
            }
            logic_on_internal_event("@rehash_rollback_ok@","","","","",PRIVATE_EVENT_SEVERITY_CODE_INFORMATIONAL,private_event,"");
            return true;
        }
    }
    logic_on_internal_event("@rehash_rollback_error@","","","","",PRIVATE_EVENT_SEVERITY_CODE_ERROR,"Don't have any backup files for roll-back.","");
    return false;
}

/*!
    \brief Tries to rehash (reload and apply new configuration file logic.txt)
    \author VooDooMan
    \version 1
    \date 2004
    \param log Returns log from operation
    \param lang Language code (e.g. "en") for log argument
    \param who_string String with name of user who caused rehashing (for log purposes - bot.log)
    \param who This should contain who caused rehasing (for log purposes - bot.log)
    \param eol End-of-line sequence (specific and different for telnet and DCC)
    \param who_only_name_of_user This should contain who caused rehasing (for log purposes - bot.log) - but only plain name of user, no decoration
    \param startup Set to true for initailization, false for common rehash
*/
void logic_rehash(string& log, string lang, string eol, string who_string, string who, string who_only_name_of_user, bool startup = false)
{
    log="";
    for(;;) {
        log+=lang_get_string(1,lang,378).c_str();
        log+=eol;

        int error_on_line=0;
        string error_reason;
        bool error=true;
        //if(!first_iteration)
        error=logic_rehash(error_on_line,error_reason,who_string,who_only_name_of_user);
        if(error) {
            if(startup) {
                log+=lang_get_string(1,lang,379);
                log+=tos(error_on_line);
                log+=lang_get_string(1,lang,380);
                log+=error_reason;
                log+=eol;
                string tmp2="Rehashing has failed, trying to perform rollback to previous configuration. Error on line #";
                tmp2+=tos(error_on_line);
                tmp2+=" error reason: ";
                tmp2+=error_reason;
                log_bot(tmp2);
                vector<string> log2;
                bool okay=logic_rollback(log2,lang);
                for(vector<string>::iterator ii=log2.begin(); ii!=log2.end(); ++ii)
                    log+=*ii+eol;
                if(okay==false) {
                    log+=lang_get_string(1,lang,381);
                    log+=eol;
                    log_bot("Fatal error: Rollback operation failed (no more backup files?) I am gonna DIE now!");

                    logic_on_internal_event("@rehash_rollback_error@","","","","",PRIVATE_EVENT_SEVERITY_CODE_FATAL_ERROR,tmp2,"");

                    bot_shutdown_action(E_SD_DIE,"SELF");
                    return;
                }
                continue;
            } else {
                log+=lang_get_string(1,lang,767);
                log+=tos(error_on_line);
                log+=lang_get_string(1,lang,380);
                log+=error_reason;
                log+=eol;
                string tmp2="Rehashing has failed, ignoring rehash request. Please, correct errors. Error on line #";
                tmp2+=tos(error_on_line);
                tmp2+=" error reason: ";
                tmp2+=error_reason;
                log_bot(tmp2);
                string log2;
                return;
            }
        } else {
            log+=lang_get_string(1,lang,382);
            log+=eol;
            log_bot("Rehashing succeeded.");
            break;
        }
    }

    if(!startup) {
        conf_pass.rehash();

        {
            conf_ssl_init();
            string log="Rehash by ";
            log+=who;
            log_ssl(log);
        }

        irc_rehashed();
        dcc_rehashed(who);
        botnet_logic_rehashed();

        irc_last_keepalive_detection=my_time(NULL)-IRC_KEEPALIVE_PERIOD;    // do not check for timeout now - this operation could have been slow
        irc_keepalive_sent=false;
    } else {
        conf_ssl_init();
    }
}

/*!
    \brief Tries to rehash (reload and apply new configuration file logic.txt)
    \author VooDooMan
    \version 1
    \date 2004
*/
void logic_rollback_startup()
{
    string log;
    string lang="en";
    string eol="\n";
    string who="SELF";
    string who_string="SELF";
    string who_only_name_of_user="SELF";
    logic_rehash(log, lang, eol, who_string, who, who_only_name_of_user, true);
}

/*!
    \brief Rehashes new backed up configuration (should be called from BOTNET) and performs rollback to old backed up file(s) if parse error occurs
    \param who This should contain which bot caused rehasing (for log purposes - bot.log)
    \author VooDooMan
    \version 1
    \date 2004
*/
void logic_botnet_rehash(string who)
{
    string log;
    string eol;
    string lang="en";
    string who_string=who;
    who_string+=" (BOTNET)";
    who+=" (BOTNET)";
    who_string=(string)"*** Rehash invoked by "+who_string;
    log_bot(who_string.c_str());
    logic_rehash(log,lang,eol,who_string,who,who);
}

/*!
    \brief Rehashes new backed up configuration (should be called from DCC / telnet) and performs rollback to old backed up file(s) if parse error occurs
    \author VooDooMan
    \version 1
    \date 2004
    \param lang Language code for user (e.g. "en") for returned log
    \param eol End-of-line sequence (specific and different for DCC and telnet)
    \param who This should contain who caused rehasing (for log purposes - bot.log)
    \return Returns log from operation in language provided by lang argument
*/
string logic_partyline_rehash(const string& lang, const string& eol, const string& who)
{
    string who_string=who;
    who_string=(string)"*** Rehash invoked by "+who_string;
    log_bot(who_string.c_str());
    string log;
    logic_rehash(log,lang,eol,who_string,who,who);
    return log;
}

/*!
    \brief Converts CR (0x0d) and LN (0x0a) characters to back-slash prefixed sequences as in C-style (for log purposes, not to contain these control characters)
    \author VooDooMan
    \version 1
    \date 2005-07-08
    \param str Input string
    \return Returns encoded string
*/
string logic_kill_EOLs(const string& str)
{
    string res;
    for(unsigned i1=0; i1<str.length(); ++i1) {
        switch(str[i1]) {
            case '\r':
                res+="\\r";
                break;
            case '\n':
                res+="\\n";
                break;
            default:
                res+=(string)""+str[i1];
                break;
        }
    }
    return res;
}

/*!
    \brief Returns string with channel definition properties (for partyline)
    \author VooDooMan
    \version 2
    \date 2005-07-14, 2005-10-02
    \param ch Mask of channel(s) (case-insensitive, wild card-powered)
    \param lang Language to use
    \param eol EOL sequence to use
    \param access_to_channel List of channels that user has access to (input)
    \return Returns listing for human, or an empty string for no match and/or access denied
*/
string logic_partyline_get_chandef_listing(string ch, const string& lang, const string& eol, vector<string>& access_to_channel)
{
    string res;

    ch=str_to_lower(ch);

    vector<c_channel_def>::iterator i1;
    for(i1=r_channel_defs.begin(); i1!=r_channel_defs.end(); ++i1) {
        string tmp=str_to_lower((*i1).channel_name);
        if(match(ch.c_str(),tmp.c_str()))
            continue;
        bool got_access=false;
        vector<string>::iterator i2;
        for(i2=access_to_channel.begin(); i2!=access_to_channel.end(); ++i2) {
            if(!(*i2).compare("*") || !irc_cmp_strings_case_insensitive((*i1).channel_name,*i2)) {
                got_access=true;
                break;
            }
        }
        if(!got_access)
            continue;

        res+=lang_subst(lang_get_string(1,lang,702),(*i1).channel_name)+eol;
        vector<string>::iterator i5;
        for(i5=(*i1).groups.begin(); i5!=(*i1).groups.end(); ++i5)
            res+=(string)"    member_of_group \""+(*i5)+"\""+eol;
        res+=(string)"    on_mode "+(*i1).on_mode+eol;
        res+=(string)"    on_key "+(*i1).on_key+eol;
        res+=(string)"    on_limit "+(*i1).on_limit+eol;
        res+=(string)"    on_topic "+(*i1).on_topic+eol;
        res+=(string)"    on_ircop "+(*i1).on_ircop+eol;
        res+=(string)"    on_ctcp "+(*i1).on_ctcp+eol;
        res+=(string)"    on_privmsg "+(*i1).on_privmsg+eol;
        res+=(string)"    on_notice "+(*i1).on_notice+eol;
        res+=(string)"    on_part "+(*i1).on_part+eol;
        res+=(string)"    on_dynamic_ban "+(*i1).on_dynamic_ban+eol;
        res+=(string)"    on_server_msg "+(*i1).on_server_msg+eol;
        map<string,string>::iterator i3;
        for(i3=(*i1).dynamic_bans.begin(); i3!=(*i1).dynamic_bans.end(); ++i3)
            res+=(string)"    dynamic_ban \""+(*i3).first+"\" \""+(*i3).second+"\""+eol;
        for(i2=(*i1).dynbans_editors.begin(); i2!=(*i1).dynbans_editors.end(); ++i2)
            res+=(string)"    partyline_dynbans_editors \""+(*i2)+"\""+eol;
        vector<c_dynamic_rule>::iterator i4;
        for(i4=(*i1).allow_dynamic.begin(); i4!=(*i1).allow_dynamic.end(); ++i4)
            res+=(string)"    allow_dynamic \""+(*i4).group+"\" \""+(*i4).plus_modes+"\" \""+(*i4).minus_modes+"\""+eol;
        res+=(string)"    dynamic1 \""+(*i1).dynamic_plus_modes+"\""+eol;
        res+=(string)"    dynamic2 \""+(*i1).dynamic_minus_modes+"\""+eol;
    }

    return res;
}

/*!
    \brief Adds privileges to user for new group (".+group" command)
    \author VooDooMan
    \version 1
    \date 2006-01-21
    \param username Name of user as define in logic.txt
    \param group Name of group
*/
void logic_partyline_plus_group(const string& username, const string& group)
{
    bool got=false;
    deque<c_user>::iterator i1;
    for(i1=r_users.begin(); i1!=r_users.end(); ++i1) {
        if(!(*i1).name.compare(username)) {
            if((*i1).is_template)
                return;
            got=true;
            break;
        }
    }
    if(got) {
        (*i1).access_grant_group.push_back(group);
        (*i1).access_to_group.push_back(group);
        (*i1).access_usage_procedure.push_back(group);
    }
}

/*!
    \brief Gets name of user marked as "host_bot"
    \version 1
    \date 2006-01-27
    \author VooDooMan
*/
string logic_get_host_bot_user_name()
{
    deque<c_user>::iterator i1;
    for(i1=r_users.begin(); i1!=r_users.end(); ++i1)
        if((*i1).host_bot) {
            return (*i1).name;
        }
    return (string)"?"; // huh?! we got no "host_bot"! hmmm, lame configuration...
}

