/***************************************************************************
                          logic.h  -  description
                             -------------------
    begin                : Thu Dec 9 2004
    copyright            : (C) 2004-2007 by Marian VooDooMan Meravy
    email                : ghostvoodooman -a.T= users (dot) sourceforge.net
 ***************************************************************************/

/* $Id: logic.h,v 1.55 2008/07/02 09:35:48 ghostvoodooman Exp $ */

/***************************************************************************

VooDoo cIRCle - an IRC (ro)bot
Copyright (C) 2004-2008 by Marian VooDooMan Meravy
  (ghostvoodooman -a.T= users (dot) sourceforge.net)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************/

/*!
    \file
    \brief Implements core logic stored in "logic.txt" file
    \version $Id: logic.h,v 1.55 2008/07/02 09:35:48 ghostvoodooman Exp $
*/

#ifndef LOGIC_H__
#define LOGIC_H__

#if defined(HAVE_CONFIG_H)
#   include <config.h>
#endif

#include <string>
#include <vector>
#include <stack>
#include <list>
#include <map>

#include "utils/memleak.h"
#include "utils/assert.h"
#include "utils.h"
#include "params.h"

#include "function.h"
#include "proc/c_lines.h"
#include "logic/c_smtp.h"
#include "logic/c_channel_def.h"
#include "logic/c_private.h"
#include "irc/c_flood.h"
#include "logic/c_user.h"
#include "logic/c_channel.h"

//---------------------------------------------------------------------------

/*!
    \def LOGIC_PROC_NOT_FOUND
    \brief Procedure not found
*/
#define LOGIC_PROC_NOT_FOUND    -1
/*!
    \def LOGIC_PROC_BAD_TEXT
    \brief Bad text
*/
#define LOGIC_PROC_BAD_TEXT     -2
/*!
    \def LOGIC_PROC_BAD_NUM
    \brief Bad number
*/
#define LOGIC_PROC_BAD_NUM      -3
/*!
    \def LOGIC_PROC_BAD_SYNTAX
    \brief Bad syntax
*/
#define LOGIC_PROC_BAD_SYNTAX   -4
/*!
    \def LOGIC_SOCKET_ERROR
    \brief Socket I/O error
*/
#define LOGIC_SOCKET_ERROR      -5
/*!
    \def LOGIC_RESTART
    \brief Bot wants to be restarted ("restart" command)
*/
#define LOGIC_RESTART           -6
/*!
    \def LOGIC_NO_SERVER
    \brief No irc server in the list
*/
#define LOGIC_NO_SERVER         -7
/*!
    \def LOGIC_OK
    \brief Ok, no error
*/
#define LOGIC_OK                0
int logic_call_proc(const char* call_string);

#include <vector>
#include <map>
#include <list>
#include <string>

using namespace std;

string logic_find_user(const string& nick, const string& ident, const string& host, const string& fullname, bool irc_op);
string logic_find_user(const string& nick, const string& ident, string host, const string& fullname, bool irc_op, bool& was_dns);
bool logic_access_to_partyline(const string& nick, const string& ident, const string& host, const string& fullname, const bool irc_op, string& user_name_as_in_logic);
string logic_get_dcc_host(const int dcc_group, const string& user_name_as_in_logic);
bool logic_access_to_partyline(const string& user_name_as_in_logic);

void logic_validate(const string& channel, const string& source_user, const string& target_user, const string& source_nick, const string& target_nick, const char prefix, const char action);
void logic_validate_by_mask(const string& channel, const string& source_user, const string& source_nick, const string& mask, const char prefix, const char action);
void logic_on_join(const string& nick, const string& ident, const string& host, const string& fullname, const string& channel, const bool irc_op);
void logic_on_nick_change(const string& nick, const string& ident, const string& host, const string& fullname, const string& channel, const bool irc_op, const string& old_nick);
void logic_on_nick_validate(const string& nick, const string& ident, const string& host, const string& fullname, const string& channel, const bool irc_op, const bool passive, bool& mod_disp_join_called);
void logic_on_banned(const string& channel, const string& ban_mask, const vector<string>& excepts, const vector<string>& invites, const vector<string>& bans);
void logic_validate_kick(const string& channel, const string& source_user, const string& target_user, const string& source_nick, const string& target_nick);
void logic_loop();

void logic_loop_identd();

int logic_partyline_whois(const string& lang, string param1, list<string>& lines);
bool logic_partyline_user_exists(string name);
void logic_partyline_get_channels(vector<string>& channels);
void logic_partyline_backup(c_module* this_module, string who);
void logic_partyline_remove_user(const string& username);

void logic_list_proc(string proc_name, list<string>& commands, string& declaration, vector<string>& groups, vector<c_replication>& replications, vector<string>& remote_call);
void logic_list_func(string func_name, list<string>& commands, string& declaration, vector<string>& groups, vector<c_replication>& replications, vector<string>& remote_call);

string logic_find_proc_by_name(const string& name);

void logic_parse_proc(string& error_reason, int& error_line, list<string>& commands, bool copy_to_r_procedures, string declaration, vector<string>& groups, vector<c_replication>& replications, vector<string>& remote_call);
bool logic_parse_func(string& error_reason, int& error_line, list<string>& commands, bool copy_to_r_procedures, string declaration, vector<string>& groups, vector<c_replication>& replications, vector<string>& remote_call);

void logic_on_flood(const string& channel, const string& user, const string& nick, const int type, const int stat_num, const my_time_t stat_sec, const my_time_t stat_sec_of, const vector<string>& flist);

void logic_on_privmsg(const string& channel, string user, const string& nick, string msg);
void logic_on_privmsg_query(string user, const string& nick, string msg);
void logic_on_notice(const string& channel, string user, const string& nick, const string& msg);

void logic_botnet_backup(string who);
void logic_botnet_remove_user(const string& username);

void logic_set_dynamic(const string& username, const string& from_user, const char prefix, const char mode);

void logic_validate_chan_mode(const string& channel, const string& source_user, const string& source_nick, const char prefix, const char mode);
void logic_validate_chan_key(const string& channel, const string& source_user, const string& source_nick, const char prefix, const string& key);
void logic_validate_chan_limit(const string& channel, const string& source_user, const string& source_nick, const char prefix, const string& limit);
void logic_validate_topic(const string& channel, const string& source_user, const string& source_nick, const string& topic);

void logic_on_not_invited(const string& channel, const string& mask, const string& user_name);
void logic_on_not_in_reop(const string& channel, const string& mask, const string& user_name);

bool logic_access_to_filesystem(const string& user);

void logic_botnet_rehash(string who);
string logic_partyline_rehash(const string& lang, const string& eol, const string& who);

void logic_on_ircop(const string& channel, const string& nick, const bool on);
void logic_on_part(const string& channel, const string& nick, const string& msg, const int type);

void logic_partyline_get_user(const string& user_name, c_user& user, vector<c_channel>& channels);
void logic_partyline_get_user_template(const string& user_name, c_user& user, vector<c_channel>& channels);

bool logic_check_mask(string mask1, string mask2, const string& fnm1, const string& fnm2, bool& was_dns);

void logic_filesys_got_new(const string& user_name, const string& nick, const string& ident, const string& host, const string& internal_name);
void logic_fnc(const string& old_nick, const string& new_nick);
void logic_on_broadcast(string username, const string& nick,
                        const string& ident, const string& host,
                        const string& bcast_mask, const int type,
                        string msg, const string& server,
                        const unsigned short port);
void logic_on_server_msg(const string& channel, const string& source, const int type, const string& msg, const string& server, const unsigned short port);
void logic_on_internal_event(string type, string flags1, string flags2, string flags3, string flags4, const int severity_numeric, string msg1, string msg2);

string logic_partyline_get_chandef_listing(string ch, const string& lang, const string& eol, vector<string>& access_to_channel);

#include "proc/c_procedure.h"

void logic_get_push_procs(const string& botname, vector<c_procedure>& procs);
void logic_get_push_funcs(const string& botname, vector<c_function*>& funcs);

int logic_get_priority(const string& class_);
string logic_set_priority(const int priority);

void logic_user_mode_changing(const string& target_nick, const string& channel, char prefix, char action);

string logic_eval(c_module* this_module, const string& str, c_var_array* vars, const string& debug);

void logic_shift_backup(const bool shift_all);

bool logic_process_line(c_module* this_module, string ln, string& error_reason, int& error_line, bool& b_user2, c_user& user, bool& b_channel, c_channel& channel, bool& b_user, int& i_brackets, bool& b_proc, c_procedure& procedure, deque<c_lines>& lines, bool& b_smtp, c_smtp& smtp, bool& b_smtp_data, vector<c_procedure>& procedures, deque<c_user>& users, vector<c_channel>& channels, c_channel_def& chan, vector<c_channel_def>& channel_defs, bool& b_chan, vector<string>& all_groups, bool& b_groups, bool& b_prv, c_private& prv, deque<c_user>& terminators, map<string,c_function*>& functions, const string& debug);

bool logic_botnet_get_proc(string name, c_procedure& proc);
bool logic_botnet_get_func(string name, c_function& func);
bool logic_botnet_get_user(const string& name, c_user& user);
void logic_get_pull_users(const string& botname, deque<c_user>& users);
void logic_get_push_users(const string& botname, deque<c_user>& users);
bool logic_botnet_get_user_channels(const string& name, vector<c_channel>& channels);
void logic_get_pull_chan_defs(const string& botname, vector<c_channel_def>& chans);
void logic_get_push_chan_defs(const string& botname, vector<c_channel_def>& chans);
bool logic_botnet_get_channel_def(const string& channel_name, c_channel_def& chd);
void logic_botnet_remove_channel_def(const string& channel_name);

void logic_partyline_add_user(c_user& user, vector<c_channel>& channels);
void logic_partyline_access(const string& user, vector<string>& access_to_users,
                            vector<string>& access_to_groups, vector<string>& access_grant_groups,
                            vector<string>& access_to_proc_groups, vector<string>& access_grant_proc_groups,
                            vector<string>& access_grant_proc, vector<string>& access_to_channel,
                            vector<string>& access_grant_channel, bool& access_grant_partyline,
                            vector<string>& access_to_procs, bool& access_to_backup,
                            bool& access_grant_backup, bool& access_to_rehash, bool& access_grant_rehash,
                            bool& access_to_plususer, bool& access_grant_plususer, bool& access_to_plusproc,
                            bool& access_grant_plusproc, c_flood& partyline_msg_flood,
                            bool& access_to_replication, bool& access_grant_replication,
                            vector<string>& access_to_chan_defs, bool& access_to_restart,
                            bool& access_grant_restart, bool& access_to_die, bool& access_grant_die,
                            bool& access_to_filesystem, bool& access_grant_filesystem, bool& access_to_private,
                            bool& access_grant_private, bool& access_to_can_send_all_users, bool& access_grant_can_send_all_users,
                            bool& access_to_can_send_unknown_users, bool& access_grant_can_send_unknown_users,
                            vector<string>& access_usage_proc, vector<string>& access_usage_proc_groups, bool& access_to_upgrade,
                            bool& access_grant_upgrade, bool& access_to_apply, bool& access_grant_apply,
                            vector<string>& access_to_users_also_terminators);
bool logic_partyline_get_channel_def(const string& channel_name, c_channel_def& chd);
void logic_partyline_remove_channel_def(const string& channel_name);
bool logic_partyline_add_channel_def(c_channel_def& chd);
void logic_partyline_plus_group(const string& username, const string& group);
bool logic_check_num_agrs(string proc, const size_t mandatory);
void logic_rollback_startup();
string logic_get_host_bot_user_name();
void logic_get_flood_limits(const string& channel, const string& user, c_flood& msg_flood, c_flood& notice_flood, c_flood& repeat_flood, c_flood& nick_flood, c_flood& join_flood, c_flood& mode_flood, c_flood& ctcp_flood);
bool logic_is_replication_partner(const string& user_name_as_in_logic);
bool logic_rollback(vector<string>& log, const string& lang);

#endif
