/***************************************************************************
                          c_channel.h  -  description
                             -------------------
    begin                : Thu May 10 2007
    copyright            : (C) 2004-2007 by Marian VooDooMan Meravy
    email                : ghostvoodooman -a.T= users (dot) sourceforge.net
 ***************************************************************************/

/* $Id: c_channel.h,v 1.14 2008/03/11 21:11:38 ghostvoodooman Exp $ */

/***************************************************************************

VooDoo cIRCle - an IRC (ro)bot
Copyright (C) 2004-2008 by Marian VooDooMan Meravy
  (ghostvoodooman -a.T= users (dot) sourceforge.net)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************/

/*!
    \file
    \brief Class 'c_channel'
    \version $Id: c_channel.h,v 1.14 2008/03/11 21:11:38 ghostvoodooman Exp $
*/

#ifndef C_CHANNEL_H__
#define C_CHANNEL_H__

#if defined(HAVE_CONFIG_H)
#   include <config.h>
#endif

#include <string>
#include <vector>
#include <stack>
#include <list>
#include <map>
#include <algorithm>

#include "../utils/memleak.h"

#include "../params.h"

#include "c_dynamic_rule.h"
#include "../irc/c_flood.h"

/*!
    \brief Stores definition of properties for user on specific channel
    \author VooDooMan
    \version 2
    \date 2005
*/
class c_channel
{
public:
    string username;                                //!< Name of user
    string channel_name;                            //!< Name of channel
    bool host_unknown;                              //!< Is this user "an unknown user"? (if true, this structure contains definition for user that is NOT recognized - something like "guest" on channel)
    bool is_template;                               //!< Is this an user template?
    bool terminator;                                //!< Is this user terminated?

    bool can_send_unknown_users;                    //!< Can user send messages / files to unknown users?

    vector<string> groups;                          //!< List of groups that this user is member of
    string on_deop;                                 //!< Procedure for on_deop event (-o)
    string on_ban;                                  //!< Procedure for on_ban event (+b)
    string on_unban;                                //!< Procedure for on_unban event (-b)
    string on_kick;                                 //!< Procedure for on_kick event
    string on_op;                                   //!< Procedure for on_op event (+o)
    string on_voice;                                //!< Procedure for on_voice event (+v)
    string on_devoice;                              //!< Procedure for on_devoice event (-v)
    string on_join;                                 //!< Procedure for on_join event
    string on_banned;                               //!< Procedure for on_banned event (if some ban mask on bot's JOIN match this user)
    string on_flood;                                //!< Procedure for on_flood event
    string on_privmsg;                              //!< Procedure for on_privmsg event (PRIVMSG)
    string on_notice;                               //!< Procedure for on_notice event (NOTICE)
    string on_except;                               //!< Procedure for on_except event
    string on_unexcept;                             //!< Procedure for on_unexcept event
    string on_invite;                               //!< Procedure for on_invite event
    string on_uninvite;                             //!< Procedure for on_uninvite event
    string on_not_invited;                          //!< Procedure for on_not_invited event
    string on_other_mode;                           //!< Procedure for on_other_mode event
    string on_not_in_reop;                          //!< Procedure for on_not_in_reop event
    string on_reop;                                 //!< Procedure for on_reop event

    string on_other_mode_mask;                      //!< Procedure for on_other_mode event

    c_flood msg_flood;                              //!< Contains msg_flood limitation for this user (PRIVMSG)
    c_flood notice_flood;                           //!< Contains notice_flood limitation for this user (NOTICE)
    c_flood repeat_flood;                           //!< Contains repeat_flood limitation for this user
    c_flood nick_flood;                             //!< Contains nick_flood limitation for this user (NICK)
    c_flood join_flood;                             //!< Contains join_flood limitation for this user (JOIN <-> PART/QUIT/benn KICK-ed)
    c_flood mode_flood;                             //!< Contains mode_flood limitation for this user (MODE)
    c_flood ctcp_flood;                             //!< Contains ctcp_flood limitation for this user (CTCP to all users in the channel)

    vector<c_dynamic_rule> allow_dynamic;           //!< List of dynamic rule that can be accepted

    string dynamic_plus_modes;                      //!< Wich modes accepted? ('+' for +v, '@' for +o)
    string dynamic_minus_modes;                     //!< Wich modes accepted? ('+' for -v, '@' for -o)

    bool configured;                                //!< Set to true if all has been configured (only valid for partyline)

    c_channel();

    void clear();
};

#endif

