/***************************************************************************
                          c_channel_def.h  -  description
                             -------------------
    begin                : Thu May 10 2007
    copyright            : (C) 2004-2007 by Marian VooDooMan Meravy
    email                : ghostvoodooman -a.T= users (dot) sourceforge.net
 ***************************************************************************/

/* $Id: c_channel_def.h,v 1.15 2008/03/25 12:04:27 ghostvoodooman Exp $ */

/***************************************************************************

VooDoo cIRCle - an IRC (ro)bot
Copyright (C) 2004-2008 by Marian VooDooMan Meravy
  (ghostvoodooman -a.T= users (dot) sourceforge.net)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************/

/*!
    \file
    \brief Class 'c_channel_def'
    \version $Id: c_channel_def.h,v 1.15 2008/03/25 12:04:27 ghostvoodooman Exp $
*/

#ifndef C_CHANNEL_DEF_H__
#define C_CHANNEL_DEF_H__

#if defined(HAVE_CONFIG_H)
#   include <config.h>
#endif

#include <string>
#include <vector>
#include <stack>
#include <list>
#include <map>
#include <algorithm>

#include "../utils/memleak.h"

#include "../params.h"

#include "c_dynamic_rule.h"
#include "../botnet/c_replication.h"

/*!
    \brief Stores informations about channel definitions
    \author VooDooMan
    \version 1
    \date 2004
*/
class c_channel_def
{
public:
    string channel_name;                            //!< Name of channel
    string on_mode;                                 //!< on_mode event
    string on_key;                                  //!< on_key event
    string on_limit;                                //!< on_limit event
    string on_topic;                                //!< on_topic event
    string on_ircop;                                //!< on_ircop event: An irc operator on channel event
    string on_privmsg;                              //!< on_privmsg event: Some has written PRIVMSG to channel
    string on_notice;                               //!< on_notice event
    string on_part;                                 //!< on_part event: Some on channel has left channel / quit IRC
    string on_dynamic_ban;                          //!< on_dynamic_ban event
    string on_ctcp;                                 //!< on_ctcp event
    string on_server_msg;                           //!< on_server_msg event
    vector<c_dynamic_rule> allow_dynamic;           //!< Dynamic rules
    string dynamic_plus_modes;                      //!< Which modes accepted? ('+' for +v, '@' for +o)
    string dynamic_minus_modes;                     //!< Which modes accepted? ('+' for -v, '@' for -o)
    map<string,string> dynamic_bans;                //!< List of dynamically banned users

    vector<string> keys;                            //!< History of +k key's

    vector<string> groups;                          //!< List of groups which this channel is member of

    vector<string> dynbans_editors;                 //!< List of groups that can use .dynbans command for this channel

    vector<c_replication> replication;              //!< Informations of replication

    my_time_t last_changed;                            //!< Timestamp for replication

    /*!
        \brief Cleans up
        \author VooDooMan
        \version 1
        \date 2005
    */
    void clear();

    c_channel_def();
};

#endif

