/***************************************************************************
                          c_functions.cpp  -  description
                             -------------------
    begin                : Thu May 10 2007
    copyright            : (C) 2004-2007 by Marian VooDooMan Meravy
    email                : ghostvoodooman -a.T= users (dot) sourceforge.net
 ***************************************************************************/

/* $Id: c_functions.cpp,v 1.25 2008/11/24 12:09:46 ghostvoodooman Exp $ */

/***************************************************************************

VooDoo cIRCle - an IRC (ro)bot
Copyright (C) 2004-2008 by Marian VooDooMan Meravy
  (ghostvoodooman -a.T= users (dot) sourceforge.net)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************/

/*!
    \file
    \brief Class 'c_functions'
    \version $Id: c_functions.cpp,v 1.25 2008/11/24 12:09:46 ghostvoodooman Exp $
*/

#if defined(HAVE_CONFIG_H)
#   include <config.h>
#endif

#include <iostream>

#include "../params.h"
#include "../utils/memleak.h"
#include "../utils/assert.h"
#include "../compat.h"

#include "c_functions.h"
#include "../var/c_var_bool.h"
#include "../var/c_var_int.h"
#include "../var/c_var_string.h"
#include "../var/c_var_unset.h"
#include "../irc.h"
#include "../modules.h"
#include "../logic.h"
#include "../irc/c_irc_server.h"
#include "../match.h"
#include "../dcc.h"
#include "../runtime/c_timer.h"
#include "../mod/c_module.h"
#include "../botnet.h"
#include "../md5.h"
#include "../utils/parse_flags.h"
#include "../stats.h"
#include "../runtime/c_delay_link.h"
#include "../runtime/c_delay_command.h"
#include "../irc/c_online_channel.h"
#include "../identd.h"
#include "../conf/c_conf.h"
#include "../log.h"
#include "../misc/shutdown.h"
#include "../pcre/c_pcre.h"
#include "../casemap.h"
#include "../internal_state.h"
#include "../misc/c_script_timeout_protector_locker.h"

extern deque<c_user> r_terminators;
extern deque<c_user> r_user_templates;
extern deque<c_user> r_users;

#include "../utils/precomp.h"

#if defined(_MSC_VER) || defined(__BORLANDC__)
#   pragma hdrstop
#endif

string get_s_arg(c_module* this_module, vector<c_func_arg>& args, size_t num, const string& UNUSEDARG(debug), string& error, bool& ok);
MY_LONG_LONG get_i_arg(c_module* this_module, vector<c_func_arg>& args, size_t num, const string& debug, string& error, bool& ok);
c_var_base* get_b_arg(c_module* UNUSEDARG(this_module), vector<c_func_arg>& args, size_t num, const string& UNUSEDARG(debug), string& error, bool& ok);
c_var_array* get_a_arg(c_module* UNUSEDARG(this_module), vector<c_func_arg>& args, size_t num, const string& UNUSEDARG(debug), string& error, bool& ok);
var_sym_index_t get_var_sym_index(c_module* UNUSEDARG(this_module), vector<c_func_arg>& args, size_t num, const string& UNUSEDARG(debug), string& error, bool& ok, bool need_a_reference_to_var);

bool set_arg(c_module* UNUSEDARG(this_module), vector<c_func_arg>& args, c_var_base* to_set, size_t num, const string& UNUSEDARG(debug), string& error, bool& ok);
bool bad_num_args(const string& UNUSEDARG(debug), string& error);

c_var_base* explode_function(c_module* this_module, const string& separator, const string& str, MY_LONG_LONG limit, const string& debug);

void eval_cleanup_timeout();
void microcode_perform_updates(stack<c_eval_iteration>& st);

MY_STATIC bool php_preg_grep(c_module* this_module, const string& debug, const string& pattern, c_var_array* input, MY_LONG_LONG flags, c_var_base*& retval, string& error);
MY_STATIC bool php_preg_match_all(c_module* this_module, const string& debug, const string& pattern, const string& subject, c_var_array* &matches_array, MY_LONG_LONG flags, MY_LONG_LONG offset, c_var_base*& retval, string& error);
MY_STATIC bool php_preg_match(c_module* this_module, const string& debug, const string& pattern, const string& subject, c_var_array* &matches_array, MY_LONG_LONG flags, MY_LONG_LONG offset, c_var_base*& retval, string& error);
MY_STATIC bool php_preg_split(c_module* this_module, const string& debug, const string& pattern, const string& subject, MY_LONG_LONG limit, MY_LONG_LONG flags, c_var_base*& retval, string& error);
MY_STATIC bool php_preg_replace(c_module* this_module, const string& debug, c_var_base* pattern, c_var_base* replacement, c_var_base* subject, size_t limit, c_var_base*& retval, c_var_int* &count, string& error);

bool php_do_preg_match_all(c_module* this_module, const string& debug, c_var_array* &matches_array, MY_LONG_LONG flags, string& error, const vector<vector<pair<int,string> > >& matches);
bool php_do_preg_match(c_module* this_module, const string& debug, c_var_array* &matches_array, MY_LONG_LONG flags, string& error, const vector<vector<pair<int,string> > >& matches);

extern deque<c_msg_to_server> msgs_to_server;

c_functions::c_functions()
{
    clear();
}

c_functions::c_functions(const c_functions& orig)
{
    functions=orig.functions;
}

void c_functions::clear()
{
    functions.clear();
}

void c_functions::destroy_all()
{
    map<string,c_function*>::iterator i1;
    for(i1=functions.begin(); i1!=functions.end(); ++i1) {
        (*i1).second->destroy(true,true);
        MY_DELETE((*i1).second);
    }
    functions.clear();
}

bool c_functions::is_built_in_function(const string& func_name)
{
    // handle special cases first
    if(func_name.empty())
        return false;
    switch(func_name[0]) {
        case 'a':
            if(!func_name.compare("array"))
                return true;
            break;
        case 'i':
            if(!func_name.compare("isset"))
                return true;
            break;
        case 'u':
            if(!func_name.compare("unset"))
                return true;
            break;
        case 'e':
            if(!func_name.compare("eval"))
                return true;
            break;
    }

    c_var_base* dummy1;
    vector<c_func_arg> dummy2;
    string dummy3;
    bool dummy4;
    stack<c_eval_iteration> dummy5;

    c_functions fns;

    return fns.call_built_in_func(NULL,func_name,dummy1,dummy2,"",dummy3,dummy4,"",NULL,dummy5);
}

bool c_functions::call_built_in_func(c_module* this_module, const string& func_name, c_var_base*& retval, vector<c_func_arg>& args, const string& debug, string& error, bool& skip_block, const string& next_expression, c_var_array* vars, stack<c_eval_iteration>& st)
{
    retval=NULL;
    skip_block=false;
    error="";
    if(func_name.empty())
        return false;

    bool got_func=false;
    bool ok=true;

    switch(func_name[0]) {
        case 'e':
            if(!func_name.compare("explode")) {
                if(NULL==this_module)
                    return true;

                if(args.size()<2 || args.size()>3)
                    return bad_num_args(debug,error);
                size_t arg=1;
                string separator=get_s_arg(this_module,args,arg++,debug,error,ok);
                string str=get_s_arg(this_module,args,arg++,debug,error,ok);
                MY_LONG_LONG limit=MAX_S_LONG_LONG;
                if(args.size()==3)
                    limit=get_i_arg(this_module,args,arg++,debug,error,ok);
                if(!ok)
                    return false;
                got_func=true;

                retval=explode_function(this_module,separator,str,limit,debug);

                break;
            }
            if(!func_name.compare("eval")) {
                if(NULL==this_module)
                    return true;

                if(args.size()!=1)
                    return bad_num_args(debug,error);
                size_t arg=1;
                string code_str=get_s_arg(this_module,args,arg++,debug,error,ok);
                if(!ok)
                    return false;
                got_func=true;

                if(!eval_function(this_module,code_str,vars,*this,debug,st)) {
                    retval=MY_NEW c_var_bool(false);
                }

                return true;
            }
            break;
        case 'o':
            if(!func_name.compare("op")) {
                if(NULL==this_module)
                    return true;

                if(args.size()!=3)
                    return bad_num_args(debug,error);
                size_t arg=1;
                int priority=static_cast<int>(get_i_arg(this_module,args,arg++,debug,error,ok));
                string channel=get_s_arg(this_module,args,arg++,debug,error,ok);
                string nick=get_s_arg(this_module,args,arg++,debug,error,ok);
                if(!ok)
                    return false;
                got_func=true;

                if(!irc_got_op(channel) || (stoi(core_conf_getvar("smart_mode"))!=0 && irc_get_mode(channel,nick).find((string)""+irc_isupport.get_prefix1('o'),0)!=string::npos))
                    break;
                irc_op(channel,nick,priority);

                logic_user_mode_changing(nick,channel,'+','o');

                break;
            }
            if(!func_name.compare("ord")) {
                if(NULL==this_module)
                    return true;

                if(args.size()!=1)
                    return bad_num_args(debug,error);
                size_t arg=1;
                string str=get_s_arg(this_module,args,arg++,debug,error,ok);
                if(!ok)
                    return false;
                got_func=true;

                char ch=0;
                if(str.length())
                    ch=str[0];

                retval=MY_NEW c_var_int((int)ch);

                break;
            }
            break;
        case 'd':
            if(!func_name.compare("deop")) {
                if(NULL==this_module)
                    return true;

                if(args.size()!=3)
                    return bad_num_args(debug,error);
                size_t arg=1;
                int priority=static_cast<int>(get_i_arg(this_module,args,arg++,debug,error,ok));
                string channel=get_s_arg(this_module,args,arg++,debug,error,ok);
                string nick=get_s_arg(this_module,args,arg++,debug,error,ok);
                if(!ok)
                    return false;
                got_func=true;

                if(!irc_got_op(channel) || (stoi(core_conf_getvar("smart_mode"))!=0 && irc_get_mode(channel,nick).find((string)""+irc_isupport.get_prefix1('o'),0)==string::npos))
                    break;
                irc_deop(channel,nick,priority);

                logic_user_mode_changing(nick,channel,'-','o');

                break;
            }
            if(!func_name.compare("devoice")) {
                if(NULL==this_module)
                    return true;

                if(args.size()!=3)
                    return bad_num_args(debug,error);
                size_t arg=1;
                int priority=static_cast<int>(get_i_arg(this_module,args,arg++,debug,error,ok));
                string channel=get_s_arg(this_module,args,arg++,debug,error,ok);
                string nick=get_s_arg(this_module,args,arg++,debug,error,ok);
                if(!ok)
                    return false;
                got_func=true;

                if(!irc_got_op(channel) || (stoi(core_conf_getvar("smart_mode"))!=0 && irc_get_mode(channel,nick).find('+',0)==string::npos))
                    break;
                irc_devoice(channel,nick,priority);

                logic_user_mode_changing(nick,channel,'-','v');

                break;
            }
            if(!func_name.compare("disconnect")) {
                if(NULL==this_module)
                    return true;

                if(args.size()!=0)
                    return bad_num_args(debug,error);
                //size_t arg=1;
                if(!ok)
                    return false;
                got_func=true;

                irc_disconnect();
                break;
            }
            if(!func_name.compare("delete_irc_servers")) {
                if(NULL==this_module)
                    return true;

                if(args.size()<1)
                    return bad_num_args(debug,error);
                size_t arg=1;
                vector<string> data;

                string datum=get_s_arg(this_module,args,arg++,debug,error,ok);
                data.push_back(datum);
                if(!ok)
                    return false;
                while(arg<args.size()+1) {
                    string additional=get_s_arg(this_module,args,arg++,debug,error,ok);
                    if(ok) {
                        data.push_back(additional);
                    } else
                        return false;
                }
                if(!ok)
                    return false;
                got_func=true;

                vector<string>::iterator i_d;
                for(i_d=data.begin(); i_d!=data.end(); ++i_d) {
                    vector<c_irc_server>::iterator ii;
                again_erase1:
                    for(ii=irc_servers.begin(); ii!=irc_servers.end(); ++ii) {
                        string id="[";
                        id+=(*ii).host;
                        id+="]:";
                        id+=tos((*ii).port);

                        if(!match((*i_d).c_str(),id.c_str())) {
                            irc_servers.erase(ii);
                            goto again_erase1;
                        }
                    }
                    irc_server_iterator=irc_servers.end();
                }
                break;
            }
            if(!func_name.compare("delete_nicks")) {
                if(NULL==this_module)
                    return true;

                if(args.size()<1)
                    return bad_num_args(debug,error);
                size_t arg=1;
                vector<string> data;

                string datum=get_s_arg(this_module,args,arg++,debug,error,ok);
                data.push_back(datum);
                if(!ok)
                    return false;
                while(arg<args.size()+1) {
                    string additional=get_s_arg(this_module,args,arg++,debug,error,ok);
                    if(ok) {
                        data.push_back(additional);
                    } else
                        return false;
                }
                if(!ok)
                    return false;
                got_func=true;

                vector<string>::iterator i_d;
                for(i_d=data.begin(); i_d!=data.end(); ++i_d) {
                    vector<string>::iterator ii;
            again_erase2:
                    for(ii=irc_nicks.begin(); ii!=irc_nicks.end(); ++ii) {
                        if(!match((*i_d).c_str(),(*ii).c_str())) {
                            irc_nicks.erase(ii);
                            goto again_erase2;
                        }
                    }
                }
                break;
            }
            if(!func_name.compare("dynamic_ban")) {
                if(NULL==this_module)
                    return true;

                if(args.size()<3)
                    return bad_num_args(debug,error);
                size_t arg=1;
                string channel=get_s_arg(this_module,args,arg++,debug,error,ok);
                string prefix=get_s_arg(this_module,args,arg++,debug,error,ok);
                string mask=get_s_arg(this_module,args,arg++,debug,error,ok);
                string reason;
                if(arg<args.size()+1)
                    reason=get_s_arg(this_module,args,arg++,debug,error,ok);
                if(!ok)
                    return false;
                got_func=true;

                vector<c_channel_def>::iterator i1;
                for(i1=r_channel_defs.begin(); i1!=r_channel_defs.end(); ++i1) {
                    if(!irc_cmp_strings_case_insensitive((*i1).channel_name,channel)) {
                        if(prefix.compare("+") && prefix.compare("-"))
                            break;

                        pair<string,string> p;
                        p.first=mask;
                        p.second=reason;

                        map<string,string>::iterator i2;
                    erase_dyn_ban:
                        for(i2=(*i1).dynamic_bans.begin(); i2!=(*i1).dynamic_bans.end(); ++i2) {
                            if(!p.first.compare((*i2).first)) {
                                (*i1).dynamic_bans.erase(i2);
                                goto erase_dyn_ban;
                            }
                        }

                        if(!prefix.compare("+")) {
                            (*i1).dynamic_bans.insert(p);
                        }
                        if(!prefix.compare("-")) {
                            // it is erased (see above)
                        }
                        break;
                    }
                }
                break;
            }
            if(!func_name.compare("dcc_server")) {
                if(NULL==this_module)
                    return true;

                if(args.size()!=4 && args.size()!=5)
                    return bad_num_args(debug,error);
                size_t arg=1;
                int ip_version=static_cast<int>(get_i_arg(this_module,args,arg++,debug,error,ok));
                int group=static_cast<int>(get_i_arg(this_module,args,arg++,debug,error,ok));
                string bind=get_s_arg(this_module,args,arg++,debug,error,ok);
                int port=static_cast<int>(get_i_arg(this_module,args,arg++,debug,error,ok));
                string type;
                if(args.size()==5)
                    type=get_s_arg(this_module,args,arg++,debug,error,ok);
                if(!ok)
                    return false;
                got_func=true;

                if(ip_version!=4 && ip_version!=6) {
                    error="Invalid IP protocol version";
                    return ok=false;
                }

                dcc_make_server(ip_version,group,bind,(unsigned short)port,type);
                break;
            }
            if(!func_name.compare("dispatch")) {
                if(NULL==this_module)
                    return true;

                if(args.size()!=5)
                    return bad_num_args(debug,error);
                size_t arg=1;
                string type=get_s_arg(this_module,args,arg++,debug,error,ok);
                string user=get_s_arg(this_module,args,arg++,debug,error,ok);
                string channel=get_s_arg(this_module,args,arg++,debug,error,ok);
                string line=get_s_arg(this_module,args,arg++,debug,error,ok);
                c_var_array* meta=get_a_arg(this_module,args,arg++,debug,error,ok);
                if(!ok)
                    return false;
                got_func=true;

                mod_dispatch2(type,user,channel,line,true,meta,0,NULL);
                break;
            }
            if(!func_name.compare("deop_ban_mask")) {
                if(NULL==this_module)
                    return true;

                if(args.size()!=4)
                    return bad_num_args(debug,error);
                size_t arg=1;
                int priority=static_cast<int>(get_i_arg(this_module,args,arg++,debug,error,ok));
                string channel=get_s_arg(this_module,args,arg++,debug,error,ok);
                string nick=get_s_arg(this_module,args,arg++,debug,error,ok);
                string mask=get_s_arg(this_module,args,arg++,debug,error,ok);
                if(!ok)
                    return false;
                got_func=true;

                if(!irc_got_op(channel))
                    break;
                irc_deop_ban_mask(channel,nick,mask,priority);

                logic_user_mode_changing(nick,channel,'-','o');

                break;
            }
            break;
        case 'v':
            if(!func_name.compare("voice")) {
                if(NULL==this_module)
                    return true;

                if(args.size()!=3)
                    return bad_num_args(debug,error);
                size_t arg=1;
                int priority=static_cast<int>(get_i_arg(this_module,args,arg++,debug,error,ok));
                string channel=get_s_arg(this_module,args,arg++,debug,error,ok);
                string nick=get_s_arg(this_module,args,arg++,debug,error,ok);
                if(!ok)
                    return false;
                got_func=true;

                if(!irc_got_op(channel) || (stoi(core_conf_getvar("smart_mode"))!=0 && irc_get_mode(channel,nick).find('+',0)!=string::npos))
                    break;
                irc_voice(channel,nick,priority);

                logic_user_mode_changing(nick,channel,'+','v');

                break;
            }
            if(!func_name.compare("var_dump")) {
                if(NULL==this_module)
                    return true;

                if(args.size()!=1)
                    return bad_num_args(debug,error);
                size_t arg=1;
                c_var_base* var=get_b_arg(this_module,args,arg++,debug,error,ok);
                if(!ok)
                    return false;
                got_func=true;

                string dump="NULL";
                if(var)
                    dump=var->dump(this_module);
                retval=MY_NEW c_var_string(dump);

                break;
            }
            break;
        case 'k':
            if(!func_name.compare("kick")) {
                if(NULL==this_module)
                    return true;

                if(args.size()!=4)
                    return bad_num_args(debug,error);
                size_t arg=1;
                int priority=static_cast<int>(get_i_arg(this_module,args,arg++,debug,error,ok));
                string channel=get_s_arg(this_module,args,arg++,debug,error,ok);
                string nick=get_s_arg(this_module,args,arg++,debug,error,ok);
                string reason=get_s_arg(this_module,args,arg++,debug,error,ok);
                if(!ok)
                    return false;
                got_func=true;

                if(!irc_got_op(channel))
                    break;
                irc_kick(channel,nick,reason,priority);
                break;
            }
            if(!func_name.compare("kill_timers")) {
                if(NULL==this_module)
                    return true;

                if(args.size()!=1)
                    return bad_num_args(debug,error);
                size_t arg=1;
                string name=get_s_arg(this_module,args,arg++,debug,error,ok);
                if(!ok)
                    return false;
                got_func=true;

                vector<c_timer*>::iterator i1;
            kt_l:
                for(i1=this_module->r_timers.begin(); i1!=this_module->r_timers.end(); ++i1) {
                    if(!match(name.c_str(),(*i1)->name.c_str())) {
                        if((*i1)->bot_unlink_timer)
                            botnet_kill_timer(*i1);
                        if((*i1)->vars) {
                            var_keeper->free_global((*i1)->vars,true);
                        }
                        (*i1)->vars=NULL;
                        MY_DELETE(*i1);
                        this_module->r_timers.erase(i1);
                        goto kt_l;
                    }
                }
                break;
            }
            break;
        case 'm':
            if(!func_name.compare("msg")) {
                if(NULL==this_module)
                    return true;

                if(args.size()!=3)
                    return bad_num_args(debug,error);
                size_t arg=1;
                int priority=static_cast<int>(get_i_arg(this_module,args,arg++,debug,error,ok));
                string whom=get_s_arg(this_module,args,arg++,debug,error,ok);
                string msg=get_s_arg(this_module,args,arg++,debug,error,ok);
                if(!ok)
                    return false;
                got_func=true;

                irc_privmsg(whom,msg,priority);
                break;
            }
            if(!func_name.compare("msgq")) {
                if(NULL==this_module)
                    return true;

                if(args.size()!=4)
                    return bad_num_args(debug,error);
                size_t arg=1;
                int priority=static_cast<int>(get_i_arg(this_module,args,arg++,debug,error,ok));
                string whom=get_s_arg(this_module,args,arg++,debug,error,ok);
                string text=get_s_arg(this_module,args,arg++,debug,error,ok);
                string quoted=get_s_arg(this_module,args,arg++,debug,error,ok);
                if(!ok)
                    return false;
                got_func=true;

                irc_privmsg(whom,text+"\x01"+quoted+"\x01",priority);
                break;
            }
            if(!func_name.compare("mask_match")) {
                if(NULL==this_module)
                    return true;

                if(args.size()!=2)
                    return bad_num_args(debug,error);
                size_t arg=1;
                string mask=get_s_arg(this_module,args,arg++,debug,error,ok);
                string what=get_s_arg(this_module,args,arg++,debug,error,ok);
                if(!ok)
                    return false;
                got_func=true;

                retval=MY_NEW c_var_bool(!match(mask.c_str(),what.c_str()));

                break;
            }
            if(!func_name.compare("make_valid_filename")) {
                if(NULL==this_module)
                    return true;

                if(args.size()!=1)
                    return bad_num_args(debug,error);
                size_t arg=1;
                string fn=get_s_arg(this_module,args,arg++,debug,error,ok);
                if(!ok)
                    return false;
                got_func=true;

                if(!fn.length()) {
                    retval=MY_NEW c_var_string("_empty_");
                    break;
                }
                string newfn;
                for(size_t i1=0; i1<fn.length(); ++i1) {
                    if(fn[i1]>='a' && fn[i1]<='z') {
                        newfn+=(string)""+fn[i1];
                        continue;
                    }
                    if(fn[i1]>='A' && fn[i1]<='Z') {
                        newfn+=(string)""+fn[i1];
                        continue;
                    }
                    if(fn[i1]>='0' && fn[i1]<='9') {
                        newfn+=(string)""+fn[i1];
                        continue;
                    }
                    if('_'==fn[i1]) {
                        newfn+="_-_";
                        continue;
                    }
                    char tmp[16];
                    sprintf(tmp,"_%02x_",(unsigned)(fn[i1]));
                    newfn+=tmp;
                }
                retval=MY_NEW c_var_string(newfn);
                break;
            }
            if(!func_name.compare("md5")) {
                if(NULL==this_module)
                    return true;

                if(args.size()!=1)
                    return bad_num_args(debug,error);
                size_t arg=1;
                string str=get_s_arg(this_module,args,arg++,debug,error,ok);
                if(!ok)
                    return false;
                got_func=true;

                char* buff=(char*)my_malloc(str.length()+1);
                if(!buff) {
                    string log="Error while allocating buffer for 'md5()' function; "+debug;
                    if(this_module->get_logger()) {
                        this_module->get_logger()->log_print1(E_LOG_DEBUG,LOG_SCOPE_GLOBAL | LOG_SCOPE_MODULE,"",log);
                        this_module->get_logger()->log_print1(E_LOG_BOT,LOG_SCOPE_GLOBAL | LOG_SCOPE_MODULE,"",log);
                    }
                    retval=MY_NEW c_var_bool(false);
                    got_func=false; // so terminate execution
                    break;
                }
                strcpy(buff,str.c_str());

                MD5Context ctx;
                MD5Init(&ctx);
                MD5Update(&ctx,(unsigned char*)buff,static_cast<unsigned>(str.length()));
                unsigned char digest[32];
                MD5Final(digest,&ctx);

                my_free(buff);

                char sig[128];
                sig[0]=0;
                for(int i1_=0; i1_<16; ++i1_) {
                    char tmp[16];
                    sprintf(tmp,"%02x",digest[i1_]);
                    strcat(sig,tmp);
                }

                retval=MY_NEW c_var_string(sig);

                break;
            }
            break;
        case 'i':
            if(!func_name.compare("if")) {
                if(NULL==this_module)
                    return true;

                if(args.size()<1)
                    return bad_num_args(debug,error);
                size_t arg=1;
                // there can be operator ',' (comma) used in if(...,...,...)
                //     expression, so ignore all, except of the last one
                while(arg<args.size())
                    arg++;
                c_var_base* last=get_b_arg(this_module,args,arg++,debug,error,ok);
                if(!ok)
                    return false;
                got_func=true;

                if(last->is_true())
                    skip_block=false;
                else
                    skip_block=true;

                return true;
            }
            if(!func_name.compare("isset")) {
                if(NULL==this_module)
                    return true;

                if(args.size()!=1)
                    return bad_num_args(debug,error);
                size_t arg=1;
                c_var_base* last=get_b_arg(this_module,args,arg++,debug,error,ok);

                bool isset=false;

                if(ok) {
                    if(last && last->get_var_type()!=VARTYPE_UNSET)
                        isset=true;
                }

                got_func=true;

                retval=MY_NEW c_var_bool(isset);

                return true;
            }
            if(!func_name.compare("in_group")) {
                if(NULL==this_module)
                    return true;

                if(args.size()!=2)
                    return bad_num_args(debug,error);
                size_t arg=1;
                string obj=get_s_arg(this_module,args,arg++,debug,error,ok);
                string grp=get_s_arg(this_module,args,arg++,debug,error,ok);
                if(!ok)
                    return false;
                got_func=true;

                bool skip=true;
                bool got=false;
                {
                    vector<c_channel>::iterator i1;
                    for(i1=r_channels.begin(); i1!=r_channels.end(); ++i1) {
                        if((*i1).username.compare(obj))
                            continue;
                        vector<string>::iterator i2;
                        for(i2=(*i1).groups.begin(); i2!=(*i1).groups.end(); ++i2) {
                            {
                                if(!(*i2).compare(grp)) {
                                    skip=false;
                                    got=true;
                                    break;
                                }
                            }
                        }
                        if(got)
                            break;
                    }
                }
                if(!got) {
                    deque<c_user>::iterator i1;
                    for(i1=r_users.begin(); i1!=r_users.end(); ++i1) {
                        if((*i1).name.compare(obj))
                            continue;
                        vector<string>::iterator i2;
                        for(i2=(*i1).groups.begin(); i2!=(*i1).groups.end(); ++i2) {
                            {
                                if(!(*i2).compare(grp)) {
                                    skip=false;
                                    got=true;
                                    break;
                                }
                            }
                        }
                        if(got)
                            break;
                    }
                }
                if(!got) {
                    vector<c_procedure>::iterator i1;
                    for(i1=this_module->r_procedures.begin(); i1!=this_module->r_procedures.end(); ++i1) {
                        if((*i1).name.compare(obj))
                            continue;
                        vector<string>::iterator i2;
                        for(i2=(*i1).groups.begin(); i2!=(*i1).groups.end(); ++i2) {
                            {
                                if(!(*i2).compare(grp)) {
                                    skip=false;
                                    got=true;
                                    break;
                                }
                            }
                        }
                        if(got)
                            break;
                    }
                }
                if(skip)
                    retval=MY_NEW c_var_bool(false);
                else
                    retval=MY_NEW c_var_bool(true);
                break;
            }
            if(!func_name.compare("if_error")) {
                if(NULL==this_module)
                    return true;

                if(args.size()!=0)
                    return bad_num_args(debug,error);
                //size_t arg=1;
                if(!ok)
                    return false;
                got_func=true;

                if(irc_socket.cmp()) {
                    skip_block=true;
                }
                break;
            }
            if(!func_name.compare("irc_server")) {
                if(NULL==this_module)
                    return true;

                if(args.size()!=2 && args.size()!=4)
                    return bad_num_args(debug,error);
                string flags;
                string myip="0.0.0.0";
                string host;
                unsigned port=0;

                if(2==args.size()) {
                    size_t arg=1;
                    host=get_s_arg(this_module,args,arg++,debug,error,ok);
                    port=static_cast<unsigned>(get_i_arg(this_module,args,arg++,debug,error,ok));
                }
                if(4==args.size()) {
                    size_t arg=1;
                    flags=get_s_arg(this_module,args,arg++,debug,error,ok);
                    myip=get_s_arg(this_module,args,arg++,debug,error,ok);
                    host=get_s_arg(this_module,args,arg++,debug,error,ok);
                    port=static_cast<unsigned>(get_i_arg(this_module,args,arg++,debug,error,ok));
                }
                if(!ok)
                    return false;
                got_func=true;

                c_irc_server s;
                s.ex=(4==args.size());
                s.host=host;
                s.port=static_cast<unsigned short>(port);
                if(s.port==0)
                    s.port=6667;
                s.flags.clear();

                if(4==args.size()) {
                    s.bind_ip=myip;
                    s.irc_type=flags;
                    s.flags=parse_flags(flags);
                }

                irc_servers.push_back(s);
                irc_server_iterator=irc_servers.begin();
                break;
            }
            if(!func_name.compare("ident")) {
                if(NULL==this_module)
                    return true;

                if(args.size()!=1)
                    return bad_num_args(debug,error);
                size_t arg=1;
                string nick=get_s_arg(this_module,args,arg++,debug,error,ok);
                if(!ok)
                    return false;
                got_func=true;

                retval=MY_NEW c_var_string(irc_get_ident(nick));
                break;
            }
            break;
        case 't':
            if(!func_name.compare("timer_once") || !func_name.compare("timer_every")) {
                if(NULL==this_module)
                    return true;

                if(args.size()!=2)
                    return bad_num_args(debug,error);
                size_t arg=1;
                string name=get_s_arg(this_module,args,arg++,debug,error,ok);
                string time=get_s_arg(this_module,args,arg++,debug,error,ok);
                if(!ok)
                    return false;
                got_func=true;

                microcode_perform_updates(st);

                int t=0;
                {
                    int mul=1;
                    int mul2=1;
                    for(size_t i4=time.length()-1; (signed)i4>=0; --i4) {
                        if(time[i4]==':') {
                            mul2=1;
                            if(mul==1) {
                                mul=60;
                                continue;
                            }
                            if(mul==60) {
                                mul*=60;
                                continue;
                            }
                            if(mul==60*60) {
                                mul*=24;
                                continue;
                            }
                            error="Invalid time (use syntax dd:hh:mm:ss)";
                            return false;
                        }
                        if(!(time[i4]>='0' && time[i4]<='9')) {
                            error="Invalid time (use syntax dd:hh:mm:ss)";
                            return false;
                        }
                        t+=mul*mul2*(time[i4]-'0');
                        mul2*=10;
                    }
                }

                c_timer* tm=MY_NEW c_timer();
                tm->once=!func_name.compare("timer_once");
                tm->name=name;
                ::my_time(&tm->set_at);
                ::my_time(&tm->again_at);
                tm->duration=t;
                if(tm->duration<=0)
                    tm->duration=1;
                tm->again_at+=tm->duration;

                tm->debug=debug;

                tm->expression=next_expression;
                {
                    session_id_t old_session=var_keeper->get_current_session();
                    var_keeper->set_current_session(SESSION_RESERVED_ID);
                    unsigned old_local=var_keeper->set_current_local(var_keeper->register_local(tm->vars,NULL,__FILE__,__my_func__,__LINE__));
                    st.top().symbol_table_->update_vars_from_symbols(this_module,tm->vars);
                    tm->vars->set_all_as_persistent();
                    var_keeper->set_current_session(old_session);
                    var_keeper->set_current_local(old_local);
                }
                this_module->r_timers.push_back(tm);
                break;
            }
            if(!func_name.compare("try_connect")) {
                if(NULL==this_module)
                    return true;

                if(args.size()!=0)
                    return bad_num_args(debug,error);
                //size_t arg=1;
                if(!ok)
                    return false;
                got_func=true;

                irc_disconnect();
                do_idle();
                if(!irc_socket.cmp()) {
                    if(0==irc_recommended_port) {
                        if(irc_server_iterator==irc_servers.end()) {
                            irc_server_iterator=irc_servers.begin();
                            if(irc_server_iterator==irc_servers.end()) {
                                string d="No server in the list! Fatal error.";
                                if(this_module->get_logger())
                                    this_module->get_logger()->log_print1(E_LOG_BOT,LOG_SCOPE_GLOBAL,"",d,true);
                                error=d;
                                return false;
                            }
                        }
                        irc_server_host=(*irc_server_iterator).host;
                        irc_server_port=(*irc_server_iterator).port;
                        string bind_ip=(*irc_server_iterator).ex?(*irc_server_iterator).bind_ip:((string)"0.0.0.0");

                        string ssl="";
                        vector<string>::iterator fi;
                        for(fi=(*irc_server_iterator).flags.begin(); fi!=(*irc_server_iterator).flags.end(); ++fi) {
                            if((*fi).find("@irc_ssl=",0)==0) {
                                ssl=*fi;
                                if(ssl.length() && ssl[0]=='@')
                                    ssl.erase(0,1);
                                if(ssl.length() && ssl[ssl.length()-1]=='@')
                                    ssl.erase(ssl.length()-1,1);
                            }
                        }
                        irc_ssl_section=ssl;

                        irc_connect(bind_ip.c_str(),(*irc_server_iterator).host.c_str(),(*irc_server_iterator).port,irc_ssl_section);
                        eval_cleanup_timeout();
                        irc_server_iterator++;
                        do_idle();
                    } else {
                        unsigned short p=irc_recommended_port;
                        irc_recommended_port=0;
                        irc_connect("0.0.0.0",irc_recommended_server.c_str(),p,irc_ssl_section);
                        eval_cleanup_timeout();
                        do_idle();
                    }
                }
                e_shutdown_reason sr=bot_get_shutdown_state();
                if(E_SD_DIE==sr || E_SD_RESTART==sr) {
                    error="Bot shutdown request";
                    return false;
                }
                break;
            }
            if(!func_name.compare("telnet_server")) {
                if(NULL==this_module)
                    return true;

                if(args.size()!=3)
                    return bad_num_args(debug,error);
                size_t arg=1;
                int ip_version=static_cast<int>(get_i_arg(this_module,args,arg++,debug,error,ok));
                string bind=get_s_arg(this_module,args,arg++,debug,error,ok);
                int port=static_cast<int>(get_i_arg(this_module,args,arg++,debug,error,ok));
                if(!ok)
                    return false;
                got_func=true;

                if(ip_version!=4 && ip_version!=6) {
                    error="Invalid IP protocol version";
                    return ok=false;
                }

                dcc_make_telnet_server(ip_version,bind,(unsigned short)port);
                break;
            }
            if(!func_name.compare("topic")) {
                if(NULL==this_module)
                    return true;

                if(args.size()!=3)
                    return bad_num_args(debug,error);
                size_t arg=1;
                int priority=static_cast<int>(get_i_arg(this_module,args,arg++,debug,error,ok));
                string channel=get_s_arg(this_module,args,arg++,debug,error,ok);
                string topic=get_s_arg(this_module,args,arg++,debug,error,ok);
                if(!ok)
                    return false;
                got_func=true;

                string msg="TOPIC ";
                msg+=channel;
                msg+=" ";
                msg+=topic;
                irc_putserv(c_tracked_msg_to_server(msg),true,priority);
                break;
            }
            if(!func_name.compare("time")) {
                if(NULL==this_module)
                    return true;

                if(args.size()!=0)
                    return bad_num_args(debug,error);
                //size_t arg=1;
                if(!ok)
                    return false;
                got_func=true;

                retval=MY_NEW c_var_int((long)my_time(NULL));
                break;
            }
            if(!func_name.compare("trim")) {
                if(NULL==this_module)
                    return true;

                if(args.size()!=1 && args.size()!=2)
                    return bad_num_args(debug,error);
                size_t arg=1;
                string str=get_s_arg(this_module,args,arg++,debug,error,ok);
                string char_list;
                bool got_char_list=false;
                if(args.size()>=2) {
                    char_list=get_s_arg(this_module,args,arg++,debug,error,ok);
                    got_char_list=true;
                }
                if(!ok)
                    return false;
                got_func=true;

                bool end=false;
                size_t cnt=0;
                signed i1;
                for(i1=0; static_cast<unsigned>(i1)<str.length(); ++i1) {
                    if(!got_char_list) {
                        switch(str[i1]) {
                            case ' ':
                            case '\t':
                            case '\n':
                            case '\r':
                            case '\0':
                            case '\x0B':    // vertical TAB
                                cnt++;
                                continue;
                            default:
                                end=true;
                                break;
                        }
                        if(end)
                            break;
                    } else {
                        bool blacklisted=false;
                        for(size_t i2=0; i2<char_list.length(); ++i2)
                            if(str[i1]==char_list[i2]) {
                                blacklisted=true;
                                break;
                            }
                        if(blacklisted) {
                            cnt++;
                            continue;
                        }
                        end=true;
                        break;
                    }
                }
                if(cnt)
                    str.erase(0,cnt);
                end=false;
                cnt=0;
                for(i1=static_cast<signed>(str.length())-1; i1>=0; --i1) {
                    if(i1<0)
                        break;
                    if(!got_char_list) {
                        switch(str[i1]) {
                            case ' ':
                            case '\t':
                            case '\n':
                            case '\r':
                            case '\0':
                            case '\x0B':    // vertical TAB
                                cnt++;
                                continue;
                            default:
                                end=true;
                                break;
                        }
                        if(end)
                            break;
                    } else {
                        bool blacklisted=false;
                        for(size_t i2=0; i2<char_list.length(); ++i2)
                            if(str[i1]==char_list[i2]) {
                                blacklisted=true;
                                break;
                            }
                        if(blacklisted) {
                            cnt++;
                            continue;
                        }
                        end=true;
                        break;
                    }
                }
                if(cnt)
                    str.erase(i1+1,cnt);

                retval=MY_NEW c_var_string(str);
                break;
            }
            break;
        case 'r':
            if(!func_name.compare("remote_execute")) {
                if(NULL==this_module)
                    return true;

                if(args.size()<2)
                    return bad_num_args(debug,error);
                size_t arg=1;
                string botname=get_s_arg(this_module,args,arg++,debug,error,ok);
                string name_only=get_s_arg(this_module,args,arg++,debug,error,ok);
                vector<c_var_base*> params;
                while(arg<args.size()+1) {
                    c_var_base* p=get_b_arg(this_module,args,arg++,debug,error,ok);
                    if(ok) {
                        p=p->copy();
                        params.push_back(p);
                    } else
                        return false;
                }
                if(!ok)
                    return false;
                got_func=true;

                string result=botnet_remote_procedure_call(botname,name_only,params,this_module);

                vector<c_var_base*>::iterator i_arg=params.begin();
                while(i_arg!=params.end()) {
                    if(*i_arg)
                        MY_DELETE(*i_arg);
                    params.erase(i_arg);
                    i_arg=params.begin();
                }

                retval=MY_NEW c_var_string(result);
                break;
            }
            if(!func_name.compare("rem")) {
                if(NULL==this_module)
                    return true;

                got_func=true;
                break;
            }
            if(!func_name.compare("restart")) {
                if(NULL==this_module)
                    return true;

                if(args.size()!=0)
                    return bad_num_args(debug,error);
                //size_t arg=1;
                if(!ok)
                    return false;
                got_func=true;

                irc_disconnect();

                bot_shutdown_action(E_SD_RESTART,"SELF");

                error="Bot shutdown request";
                return false;
            }
            if(!func_name.compare("raw")) {
                if(NULL==this_module)
                    return true;

                if(args.size()!=2)
                    return bad_num_args(debug,error);
                size_t arg=1;
                int priority=static_cast<int>(get_i_arg(this_module,args,arg++,debug,error,ok));
                string raw=get_s_arg(this_module,args,arg++,debug,error,ok);
                if(!ok)
                    return false;
                got_func=true;

                irc_put(raw,priority);
                break;
            }
            if(!func_name.compare("rand")) {
                if(NULL==this_module)
                    return true;

                if(args.size()>2)
                    return bad_num_args(debug,error);
                size_t arg=1;
                MY_LONG_LONG min=0;
                MY_LONG_LONG max=MAX_S_LONG_LONG;
                if(args.size()>0)
                    min=get_i_arg(this_module,args,arg++,debug,error,ok);
                if(args.size()>1)
                    max=get_i_arg(this_module,args,arg++,debug,error,ok);
                if(!ok)
                    return false;
                got_func=true;

                if(min>max) {
                    MY_LONG_LONG tmp=min;
                    min=max;
                    max=tmp;
                }

                MY_LONG_LONG rnd=0;
                unsigned char* ptr=(unsigned char*)&rnd;
                for(size_t i1=0; i1<sizeof(rnd); ++i1) {
                    ptr[i1]=(unsigned char)rand();
                }
                if(rnd<0)
                    rnd=-rnd;

                if(max-min+1!=0)
                    rnd%=max-min+1;
                else {
                    rnd%=max-min+1+1;
                }
                rnd+=min;
                if(rnd<min)
                    rnd=min;
                if(rnd>max)
                    rnd=max;

                retval=MY_NEW c_var_int(rnd);
                break;
            }
            break;
        case 's':
            if(!func_name.compare("smtp_raw1")) {
                if(NULL==this_module)
                    return true;

                if(args.size()<7)
                    return bad_num_args(debug,error);
                size_t arg=1;
                string myip=get_s_arg(this_module,args,arg++,debug,error,ok);
                string server=get_s_arg(this_module,args,arg++,debug,error,ok);
                int port=static_cast<int>(get_i_arg(this_module,args,arg++,debug,error,ok));
                string helo=get_s_arg(this_module,args,arg++,debug,error,ok);
                string mail_from=get_s_arg(this_module,args,arg++,debug,error,ok);
                string rcpt_to=get_s_arg(this_module,args,arg++,debug,error,ok);
                string data=get_s_arg(this_module,args,arg++,debug,error,ok);
                if(!ok)
                    return false;
                while(arg<args.size()+1) {
                    string additional=get_s_arg(this_module,args,arg++,debug,error,ok);
                    if(ok) {
                        data+="\r\n";
                        data+=additional;
                    } else
                        return false;
                }
                if(!ok)
                    return false;
                got_func=true;

                data+="\r\n";

                string fn="." FILE_SLASH "sendmail" FILE_SLASH;
                fn+=toshex(rand());
                fn+=toshex(rand());
                fn+=toshex(rand());
                fn+=toshex(rand());
                fn+=".sendmail";
                FILE* f=fopen(fn.c_str(),"wb");
                if(f) {
                    fprintf(f,"%s\r\n",myip.c_str());
                    fprintf(f,"%s\r\n",server.c_str());
                    fprintf(f,"%u\r\n",static_cast<unsigned>(port));
                    fprintf(f,"HELO %s\r\n",helo.c_str());
                    fprintf(f,"MAIL FROM: %s\r\n",mail_from.c_str());
                    fprintf(f,"RCPT TO: %s\r\n",rcpt_to.c_str());
                    fprintf(f,"DATA\r\n");
                    fprintf(f,"%s\r\n",data.c_str());
                    fclose(f);

                    string params=fn;
                    params=(string)"\""+params+"\"";

                    fn="";
                    c_conf* core_conf=get_core_conf();
                    if(core_conf)
                        fn=utils_convert_unix_path(core_conf->getvar("sendmail"));
                    if(fn.empty()) {
#if defined(_WIN32) || defined(_WIN64)
                        fn="." FILE_SLASH "sendmail" FILE_SLASH "vdc_sendmail.exe";
#else
                        fn="." FILE_SLASH "sendmail" FILE_SLASH "vdc_sendmail";
#endif
                    }
                    c_tmp_file t;
                    t.script=false;
                    cout << "Executing in CWD: " << fn << " " << params << endl;
                    t.h=exec_async_exec(fn.c_str(),params.c_str(),"");
                    t.start_up_time=my_time(NULL);
                    t.has_been_run=true;
                    this_module->tmp_files.push_back(t);

                    stats_mail_sent();
                }
                break;
            }
            if(!func_name.compare("script")) {
                if(NULL==this_module)
                    return true;

                if(args.size()<3)
                    return bad_num_args(debug,error);
                size_t arg=1;
                string type=get_s_arg(this_module,args,arg++,debug,error,ok);
                int num=static_cast<int>(get_i_arg(this_module,args,arg++,debug,error,ok));
                string channel=get_s_arg(this_module,args,arg++,debug,error,ok);
                vector<c_var_base*> params;
                while(arg<args.size()+1) {
                    c_var_base* param=get_b_arg(this_module,args,arg++,debug,error,ok);
                    if(ok) {
                        params.push_back(param);
                    } else
                        return false;
                }
                if(!ok)
                    return false;
                got_func=true;

                vector<string> flags;
                if(type.find("@")!=string::npos) {
                    string fl=type;
                    type.erase(type.find("@"),type.length()-type.find("@"));
                    fl.erase(0,fl.find("@")+1);
                    explode(fl,',',flags,true);
                }

                this_module->exec_script(type,num,channel,params,flags);
                break;
            }
            if(!func_name.compare("sleep")) {
                if(NULL==this_module)
                    return true;

                if(args.size()!=1)
                    return bad_num_args(debug,error);
                size_t arg=1;
                int secs=static_cast<int>(get_i_arg(this_module,args,arg++,debug,error,ok));
                if(!ok)
                    return false;
                got_func=true;

                string log="Sleeping for ";
                log+=tos(secs);
                log+=" second(s)...";
                log_irc_ex(E_LOG_IRC_EX_INF,log);

                utils_deep_sleep(secs);
                break;
            }
            if(!func_name.compare("strtoupper")) {
                if(NULL==this_module)
                    return true;

                if(args.size()>2)
                    return bad_num_args(debug,error);
                size_t arg=1;
                string str=get_s_arg(this_module,args,arg++,debug,error,ok);
                int casemap=MY_CASE_ASCII;
                if(args.size()>=2)
                    casemap=static_cast<int>(get_i_arg(this_module,args,arg++,debug,error,ok));
                if(!ok)
                    return false;
                got_func=true;

                switch(casemap) {
                    case MY_CASE_SERVER:
                        retval=MY_NEW c_var_string(strtoupper(str,irc_casemapping));
                        break;
                    case MY_CASE_RFC1459:
                        retval=MY_NEW c_var_string(strtoupper(str,CASE_RFC1459));
                        break;
                    case MY_CASE_ASCII:
                    default:
                        retval=MY_NEW c_var_string(str_to_upper(str));
                        break;
                }
                break;
            }
            if(!func_name.compare("strtolower")) {
                if(NULL==this_module)
                    return true;

                if(args.size()>2)
                    return bad_num_args(debug,error);
                size_t arg=1;
                string str=get_s_arg(this_module,args,arg++,debug,error,ok);
                int casemap=MY_CASE_ASCII;
                if(args.size()>=2)
                    casemap=static_cast<int>(get_i_arg(this_module,args,arg++,debug,error,ok));
                if(!ok)
                    return false;
                got_func=true;

                switch(casemap) {
                    case MY_CASE_SERVER:
                        retval=MY_NEW c_var_string(strtolower(str,irc_casemapping));
                        break;
                    case MY_CASE_RFC1459:
                        retval=MY_NEW c_var_string(strtolower(str,CASE_RFC1459));
                        break;
                    case MY_CASE_ASCII:
                    default:
                        retval=MY_NEW c_var_string(str_to_lower(str));
                        break;
                }
                break;
            }
            if(!func_name.compare("strpos")) {
                if(NULL==this_module)
                    return true;

                if(args.size()!=2)
                    return bad_num_args(debug,error);
                size_t arg=1;
                string haystack=get_s_arg(this_module,args,arg++,debug,error,ok);
                string needle=get_s_arg(this_module,args,arg++,debug,error,ok);
                got_func=true;

                size_t pos;
                if((pos=haystack.find(needle))==string::npos)
                    retval=MY_NEW c_var_bool(false);
                else
                    retval=MY_NEW c_var_int(static_cast<long>(pos));
                break;
            }
            if(!func_name.compare("strlen")) {
                if(NULL==this_module)
                    return true;

                if(args.size()!=1)
                    return bad_num_args(debug,error);
                size_t arg=1;
                string str=get_s_arg(this_module,args,arg++,debug,error,ok);
                got_func=true;

                retval=MY_NEW c_var_int(static_cast<long>(str.length()));
                break;
            }
            if(!func_name.compare("ssl_server")) {
                if(NULL==this_module)
                    return true;

                if(args.size()!=3)
                    return bad_num_args(debug,error);
                size_t arg=1;
                int ip_version=static_cast<int>(get_i_arg(this_module,args,arg++,debug,error,ok));
                string bind=get_s_arg(this_module,args,arg++,debug,error,ok);
                int port=static_cast<int>(get_i_arg(this_module,args,arg++,debug,error,ok));
                if(!ok)
                    return false;
                got_func=true;

                if(ip_version!=4 && ip_version!=6) {
                    error="Invalid IP protocol version";
                    return ok=false;
                }

                dcc_make_telnet_server(ip_version,bind,(unsigned short)port,true);
                break;
            }
            break;
        case 'n':
            if(!func_name.compare("net_send")) {
                if(NULL==this_module)
                    return true;

                if(args.size()!=2)
                    return bad_num_args(debug,error);
                size_t arg=1;
                string whom=get_s_arg(this_module,args,arg++,debug,error,ok);
                string what=get_s_arg(this_module,args,arg++,debug,error,ok);
                if(!ok)
                    return false;
                got_func=true;

#if defined(_WIN32) || defined(_WIN64)
                string fn="net";
                string param1=whom;
                string param2=what;

                while(param1.find("\"",0)!=string::npos)
                    param1.erase(param1.find("\"",0),1);
                while(param2.find("\"",0)!=string::npos)
                    param2.erase(param2.find("\"",0),1);

                while(param1.find("%",0)!=string::npos)
                    param1.erase(param1.find("%",0),1);
                while(param2.find("%",0)!=string::npos)
                    param2.erase(param2.find("%",0),1);

                string params=(string)"send \""+param1+"\" \""+param2+"\"";
                c_tmp_file t;
                t.script=false;
                cout << "Executing in CWD: " << fn << " " << params << endl;
                t.h=exec_async_exec(fn.c_str(),params.c_str(),"");
                t.start_up_time=my_time(NULL);
                t.has_been_run=true;
                this_module->tmp_files.push_back(t);
#else
                // TODO: oops: not supported
#endif
                break;
            }
            if(!func_name.compare("notice")) {
                if(NULL==this_module)
                    return true;

                if(args.size()!=3)
                    return bad_num_args(debug,error);
                size_t arg=1;
                int priority=static_cast<int>(get_i_arg(this_module,args,arg++,debug,error,ok));
                string whom=get_s_arg(this_module,args,arg++,debug,error,ok);
                string msg=get_s_arg(this_module,args,arg++,debug,error,ok);
                if(!ok)
                    return false;
                got_func=true;

                irc_notice(whom,msg,priority);
                break;
            }
            if(!func_name.compare("noticeq")) {
                if(NULL==this_module)
                    return true;

                if(args.size()!=4)
                    return bad_num_args(debug,error);
                size_t arg=1;
                int priority=static_cast<int>(get_i_arg(this_module,args,arg++,debug,error,ok));
                string whom=get_s_arg(this_module,args,arg++,debug,error,ok);
                string text=get_s_arg(this_module,args,arg++,debug,error,ok);
                string quoted=get_s_arg(this_module,args,arg++,debug,error,ok);
                if(!ok)
                    return false;
                got_func=true;

                irc_notice(whom,text+"\x01"+quoted+"\x01",priority);
                break;
            }
            break;
        case 'l':
            if(!func_name.compare("log_print")) {
                if(NULL==this_module)
                    return true;

                if(args.size()<1)
                    return bad_num_args(debug,error);
                size_t arg=1;
                string log;
                if(!(*(args.begin())).value) {
                    log="NULL";
                    arg++;
                } else {
                    c_var_base* b=get_b_arg(this_module,args,arg,debug,error,ok);
                    string additional;
                    if(ok && b && b->get_var_type()!=VARTYPE_STRING)
                        log=b->dump(this_module);
                    else
                        log=get_s_arg(this_module,args,arg,debug,error,ok);
                    arg++;
                }
                while(arg<args.size()+1) {
                    c_var_base* b=get_b_arg(this_module,args,arg,debug,error,ok);
                    string additional;
                    if(ok && b && b->get_var_type()!=VARTYPE_STRING)
                        additional=b->dump(this_module);
                    else
                        additional=get_s_arg(this_module,args,arg,debug,error,ok);
                    arg++;
                    if(ok) {
                        log+="\n";
                        log+=additional;
                    } else
                        return false;
                }
                if(!ok)
                    return false;
                got_func=true;

                vector<string> log_;
                explode(log,'\n',log_,true);

                vector<string>::iterator i1;
                if(this_module->get_logger())
                    for(i1=log_.begin(); i1!=log_.end(); ++i1)
                        this_module->get_logger()->log_print1(E_LOG_CUSTOM,LOG_SCOPE_MODULE,"",*i1);
                break;
            }
            if(!func_name.compare("link")) {
                if(NULL==this_module)
                    return true;

                if(args.size()!=7)
                    return bad_num_args(debug,error);
                size_t arg=1;
                int ip_version=static_cast<int>(get_i_arg(this_module,args,arg++,debug,error,ok));
                string type=get_s_arg(this_module,args,arg++,debug,error,ok);
                string bot_name=get_s_arg(this_module,args,arg++,debug,error,ok);
                string bind=get_s_arg(this_module,args,arg++,debug,error,ok);
                string host=get_s_arg(this_module,args,arg++,debug,error,ok);
                int port=static_cast<int>(get_i_arg(this_module,args,arg++,debug,error,ok));
                string cmd=get_s_arg(this_module,args,arg++,debug,error,ok);
                if(!ok)
                    return false;
                got_func=true;

                if(ip_version!=4 && ip_version!=6 && ip_version!=0) {
                    error="Invalid IP protocol version";
                    return ok=false;
                }

                if(0==ip_version) {
                    if(host.find(":",0)==string::npos) {
                        in_addr ip;
                        e_resolv_result res=threads_resolve4(host.c_str(),ip,RT_RESOLVE_SCHEDULE);
                        if(RR_SCHEDULED==res) {
                            c_delay_link dl;
                            dl.botname=bot_name;
                            dl.localip=bind;
                            dl.bot_host=host;
                            dl.botport=(unsigned short)port;
                            dl.remote_MD5_password=dcc_get_password(bot_name);
                            dl.bot_link_type=type;
                            dl.unlink_proc=cmd;
                            //dl.vars=MY_NEW c_var_array(*vars);
                            var_keeper->set_current_local(var_keeper->register_local(dl.vars,vars,__FILE__,__my_func__,__LINE__));
                            dl.this_module=this_module;
                            botnet_delay_link.push_back(dl);
                            break;
                        }
                        uint32 ip_;
                        memcpy(&ip_,&ip,sizeof(ip_));
                        if(ip_) {
                            char tmp[128+1];
                            tmp[0]=0;
                            my_inet_ntop(AF_INET,&ip_,tmp,128);
                            botnet_link(bot_name,bind,tmp,(unsigned short)port,dcc_get_password(bot_name),type,cmd,vars,this_module);
                        }
                    } else {
                        botnet_link(bot_name,bind,host,(unsigned short)port,dcc_get_password(bot_name),type,cmd,vars,this_module);
                    }
                } else if(4==ip_version || 6==ip_version) {
                    botnet_link(bot_name,bind,host,(unsigned short)port,dcc_get_password(bot_name),type,cmd,vars,this_module);
                }
                break;
            }
            if(!func_name.compare("load_module")) {
                if(NULL==this_module)
                    return true;

                if(args.size()!=1)
                    return bad_num_args(debug,error);
                size_t arg=1;

                c_delay_command c;
                c.type=c_delay_command::DELAY_LOAD;

                c.mod_name=get_s_arg(this_module,args,arg++,debug,error,ok);
                if(!ok)
                    return false;
                got_func=true;

                c.debug=debug;

                delay_commands.push_back(c);
                break;
            }
            if(!func_name.compare("load_bin_module")) {
                if(NULL==this_module)
                    return true;

                if(args.size()!=2)
                    return bad_num_args(debug,error);
                size_t arg=1;

                c_delay_command c;
                c.type=c_delay_command::DELAY_LOAD_BIN;

                c.mod_name=get_s_arg(this_module,args,arg++,debug,error,ok);
                c.so_name=get_s_arg(this_module,args,arg++,debug,error,ok);
                if(!ok)
                    return false;

                got_func=true;

                c.debug=debug;

                delay_commands.push_back(c);
                break;
            }
            break;
        case 'j':
            if(!func_name.compare("join")) {
                if(NULL==this_module)
                    return true;

                if(args.size()<1 || args.size()>2)
                    return bad_num_args(debug,error);
                size_t arg=1;
                string channel=get_s_arg(this_module,args,arg++,debug,error,ok);
                string key;
                if(2==args.size())
                    key=get_s_arg(this_module,args,arg++,debug,error,ok);
                if(!ok)
                    return false;
                got_func=true;

                /*
                bool on_channel=false;
                vector<c_online_channel>::iterator i2;
                for(i2=irc_channels.begin(); i2!=irc_channels.end(); ++i2)
                    if(!irc_cmp_strings_case_insensitive((*i2).name,channel)) {
                        on_channel=true;
                        break;
                    }

                if(!on_channel)
                    irc_join(channel,key);
                */
                irc_join(channel,key);
                break;
            }
            break;
        case 'p':
            if(!func_name.compare("part")) {
                if(NULL==this_module)
                    return true;

                if(args.size()<1 || args.size()>2)
                    return bad_num_args(debug,error);
                size_t arg=1;
                string channel=get_s_arg(this_module,args,arg++,debug,error,ok);
                string reason;
                if(2==args.size())
                    reason=get_s_arg(this_module,args,arg++,debug,error,ok);
                if(!ok)
                    return false;
                got_func=true;

                /*
                bool on_channel=false;
                vector<c_online_channel>::iterator i2;
                for(i2=irc_channels.begin(); i2!=irc_channels.end(); ++i2)
                    if(!irc_cmp_strings_case_insensitive((*i2).name,channel)) {
                        on_channel=true;
                        break;
                    }

                //if(on_channel)
                    irc_part(channel,reason);
                */
                irc_part(channel,reason);
                break;
            }
            if(!func_name.compare("process_on_banned")) {
                if(NULL==this_module)
                    return true;

                if(args.size()!=2)
                    return bad_num_args(debug,error);
                size_t arg=1;
                string chan=get_s_arg(this_module,args,arg++,debug,error,ok);
                string mask=get_s_arg(this_module,args,arg++,debug,error,ok);
                if(!ok)
                    return false;
                got_func=true;

                vector<c_online_channel>::iterator i;
                for(i=irc_channels.begin(); i!=irc_channels.end(); ++i) {
                    if(!irc_cmp_strings_case_insensitive((*i).name,chan)) {
                        logic_on_banned(chan, mask,(*i).excepts,(*i).invites,(*i).bans);
                        break;
                    }
                }
                break;
            }
            if(!func_name.compare("preg_grep")) {
                if(NULL==this_module)
                    return true;

                if(args.size()<2 || args.size()>3)
                    return bad_num_args(debug,error);
                size_t arg=1;
                string pattern=get_s_arg(this_module,args,arg++,debug,error,ok);
                c_var_array* input=get_a_arg(this_module,args,arg++,debug,error,ok);
                if(NULL==input) {
                    error="Second argument ($input) must be an array";
                    return false;
                }
                MY_LONG_LONG flags=0;
                if(3==args.size())
                    flags=get_i_arg(this_module,args,arg++,debug,error,ok);
                if(!ok)
                    return false;

                got_func=true;

                st.top().m_last_pcre_exec_return_code=PREG_NO_ERROR;
                if(!php_preg_grep(this_module,debug,pattern,input,flags,retval,error))
                    return false;
                break;
            }
            if(!func_name.compare("preg_last_error")) {
                if(NULL==this_module)
                    return true;

                if(args.size()>1)
                    return bad_num_args(debug,error);

                if(args.size()==1) {
                    c_var_base* error_string=MY_NEW c_var_string(get_strerr_pcre_exec(st.top().m_last_pcre_exec_return_code));
                    if(!set_arg(this_module,args,error_string,1,debug,error,ok)) {
                        MY_DELETE(error_string);
                        return false;
                    }
                }

                got_func=true;

                retval=MY_NEW c_var_int(st.top().m_last_pcre_exec_return_code);
                break;
            }
            if(!func_name.compare("preg_match_all")) {
                if(NULL==this_module)
                    return true;

                if(args.size()<3 || args.size()>5)
                    return bad_num_args(debug,error);
                size_t arg=1;
                string pattern=get_s_arg(this_module,args,arg++,debug,error,ok);
                string subject=get_s_arg(this_module,args,arg++,debug,error,ok);
                size_t matches_result_arg_num=0;
                if(ok) {
                    var_sym_index_t matches_result_var=get_var_sym_index(this_module,args,matches_result_arg_num = arg++,debug,error,ok,true);
                    if(!ok || !VAR_SYM_INDEX_IS_VALID(matches_result_var)) {
                        if(!error.empty())
                            error+="; ";
                        error+="Third argument ($matches) must be a variable got overwritten by array as result";
                        return false;
                    }
                }
                if(!ok)
                    return false;

                MY_LONG_LONG flags=0;
                MY_LONG_LONG offset=0;
                if(args.size()>=4)
                    flags=get_i_arg(this_module,args,arg++,debug,error,ok);

                if(args.size()==5)
                    offset=get_i_arg(this_module,args,arg++,debug,error,ok);

                if(!ok)
                    return false;

                got_func=true;

                st.top().m_last_pcre_exec_return_code=PREG_NO_ERROR;
                c_var_array* matches_array=NULL;
                ok=php_preg_match_all(this_module,debug,pattern,subject,matches_array,flags,offset,retval,error);

                if(0!=matches_result_arg_num) {
                    if(ok) {
                        if(!set_arg(this_module,args,matches_array,matches_result_arg_num,debug,error,ok)) {
                            MY_DELETE(matches_array);
                            return false;
                        }
                        if(!ok)
                            return false;
                    } else {
                        MY_DELETE(matches_array);
                        return false;
                    }
                } else
                    MY_DELETE(matches_array);

                if(!ok)
                    return false;

                break;
            }
            if(!func_name.compare("preg_match")) {
                if(NULL==this_module)
                    return true;

                if(args.size()<2 || args.size()>5)
                    return bad_num_args(debug,error);
                size_t arg=1;
                string pattern=get_s_arg(this_module,args,arg++,debug,error,ok);
                string subject=get_s_arg(this_module,args,arg++,debug,error,ok);
                size_t matches_result_arg_num=0;
                if(ok && args.size()>=3) {
                    var_sym_index_t matches_result_var=get_var_sym_index(this_module,args,matches_result_arg_num = arg++,debug,error,ok,true);
                    if(!ok || !VAR_SYM_INDEX_IS_VALID(matches_result_var)) {
                        if(!error.empty())
                            error+="; ";
                        error+="Third argument ($matches) must be a variable got overwritten by array as result";
                        return false;
                    }
                }
                if(!ok)
                    return false;

                MY_LONG_LONG flags=0;
                MY_LONG_LONG offset=0;
                if(args.size()>=4)
                    flags=get_i_arg(this_module,args,arg++,debug,error,ok);

                if(args.size()==5)
                    offset=get_i_arg(this_module,args,arg++,debug,error,ok);

                if(!ok)
                    return false;

                got_func=true;

                st.top().m_last_pcre_exec_return_code=PREG_NO_ERROR;
                c_var_array* matches_array=NULL;
                ok=php_preg_match(this_module,debug,pattern,subject,matches_array,flags,offset,retval,error);

                if(0!=matches_result_arg_num) {
                    if(ok) {
                        if(!set_arg(this_module,args,matches_array,matches_result_arg_num,debug,error,ok)) {
                            MY_DELETE(matches_array);
                            return false;
                        }
                        if(!ok)
                            return false;
                    } else {
                        MY_DELETE(matches_array);
                        return false;
                    }
                } else
                    MY_DELETE(matches_array);

                if(!ok)
                    return false;

                break;
            }
            if(!func_name.compare("preg_split")) {
                if(NULL==this_module)
                    return true;

                if(args.size()<2 || args.size()>4)
                    return bad_num_args(debug,error);
                size_t arg=1;
                string pattern=get_s_arg(this_module,args,arg++,debug,error,ok);
                string subject=get_s_arg(this_module,args,arg++,debug,error,ok);

                MY_LONG_LONG limit=MAX_SIGNED_INT;
                MY_LONG_LONG flags=0;
                if(args.size()>=3)
                    limit=get_i_arg(this_module,args,arg++,debug,error,ok);
                if(limit>MAX_SIGNED_INT || -1==limit || 0==limit)
                    limit=MAX_SIGNED_INT;

                if(args.size()==4)
                    flags=get_i_arg(this_module,args,arg++,debug,error,ok);

                if(!ok)
                    return false;

                got_func=true;

                st.top().m_last_pcre_exec_return_code=PREG_NO_ERROR;
                ok=php_preg_split(this_module,debug,pattern,subject,limit,flags,retval,error);

                if(!ok)
                    return false;

                break;
            }
            if(!func_name.compare("preg_replace")) {
                if(NULL==this_module)
                    return true;

                if(args.size()<3 || args.size()>5)
                    return bad_num_args(debug,error);
                size_t arg=1;
                c_var_base* pattern=get_b_arg(this_module,args,arg++,debug,error,ok);
                c_var_base* replacement=get_b_arg(this_module,args,arg++,debug,error,ok);
                c_var_base* subject=get_b_arg(this_module,args,arg++,debug,error,ok);

                if(!ok)
                    return false;

                MY_LONG_LONG limit=MAX_SIGNED_INT;
                if(args.size()>=4)
                    limit=get_i_arg(this_module,args,arg++,debug,error,ok);
                if(limit>MAX_SIGNED_INT || -1==limit || 0==limit)
                    limit=MAX_SIGNED_INT;

                if(!ok)
                    return false;

                size_t count_arg_num=0;
                if(args.size()==5) {
                    var_sym_index_t count_var=get_var_sym_index(this_module,args,count_arg_num = arg++,debug,error,ok,true);
                    if(!ok || !VAR_SYM_INDEX_IS_VALID(count_var)) {
                        if(!error.empty())
                            error+="; ";
                        error+="Fifth argument ($count) must be a variable got overwritten by INT as result count";
                        return false;
                    }
                }
                if(!ok)
                    return false;

                if(!ok)
                    return false;

                got_func=true;

                st.top().m_last_pcre_exec_return_code=PREG_NO_ERROR;
                c_var_int* count_result=NULL;
                ok=php_preg_replace(this_module,debug,pattern,replacement,subject,static_cast<size_t>(limit),retval,count_result,error);

                if(0!=count_arg_num) {
                    if(ok) {
                        if(!set_arg(this_module,args,count_result,count_arg_num,debug,error,ok)) {
                            MY_DELETE(count_result);
                            return false;
                        }
                        if(!ok)
                            return false;
                    } else {
                        MY_DELETE(count_result);
                        return false;
                    }
                } else
                    MY_DELETE(count_result);

                if(!ok)
                    return false;

                break;
            }
            break;
        case 'b':
            if(!func_name.compare("bot_nick")) {
                if(NULL==this_module)
                    return true;

                if(args.size()<1)
                    return bad_num_args(debug,error);
                size_t arg=1;
                vector<string> nicks;
                string nick=get_s_arg(this_module,args,arg++,debug,error,ok);
                if(!ok)
                    return false;

                nicks.push_back(nick);
                while(arg<args.size()+1) {
                    string additional=get_s_arg(this_module,args,arg++,debug,error,ok);
                    if(ok) {
                        nicks.push_back(additional);
                    } else
                        return false;
                }
                if(!ok)
                    return false;
                got_func=true;

                vector<string>::const_iterator i1=nicks.begin(), i2;
                while(i1!=nicks.end()) {
                    bool got=false;
                    i2=irc_nicks.begin();
                    while(i2!=irc_nicks.end()) {
                        if(!(*i2).compare(*i1)) {
                            got=true;
                            break;
                        }
                        i2++;
                    }
                    if(!got)
                        irc_nicks.push_back(*i1);
                    i1++;
                }
                break;
            }
            if(!func_name.compare("bot_ident")) {
                if(NULL==this_module)
                    return true;

                if(args.size()!=3)
                    return bad_num_args(debug,error);
                size_t arg=1;
                int ip_version=static_cast<int>(get_i_arg(this_module,args,arg++,debug,error,ok));
                string myip=get_s_arg(this_module,args,arg++,debug,error,ok);
                string bot_ident=get_s_arg(this_module,args,arg++,debug,error,ok);
                if(!ok)
                    return false;
                got_func=true;

                if(ip_version!=4 && ip_version!=6) {
                    error="Unsupported IP protocol version number";
                    ok=false;
                    return false;
                }

                if(4==ip_version) {
                    irc_ident=bot_ident;
                    if(stoi(core_conf_getvar("disable_ident_daemon"))!=0)
                        break;
                identd_fail:
                    in_addr a;
                    memset(&a,0,sizeof(a));
                    if(1!=my_inet_pton(AF_INET,myip.c_str(),&a)) {
                        error="Erroneous IPv4 address";
                        ok=false;
                        return false;
                    }
                    identd4_shutdown();
                    bool ok=identd_startup(a,113);
                    ident_string=": USERID : UNIX :";
                    ident_string+=bot_ident;
                    if(!ok && stoi(core_conf_getvar("important_ident_daemon"))!=0) {
                        do_idle();
                        e_shutdown_reason sr=bot_get_shutdown_state();
                        if(E_SD_DIE==sr || E_SD_RESTART==sr) {
                            error="Bot shutdown request";
                            return false;
                        }
                        my_sleep(500);
                        goto identd_fail;
                    }
                    break;
                }
                if(6==ip_version) {
                    irc_ident=bot_ident;
                    if(stoi(core_conf_getvar("disable_ident_daemon"))!=0)
                        break;
                identd_fail6:
                    identd6_shutdown();
                    bool ok=identd_startup(const_cast<char*>(myip.c_str()),113);
                    ident_string=": USERID : UNIX :";
                    ident_string+=bot_ident;
                    if(!ok && stoi(core_conf_getvar("important_ident_daemon"))!=0) {
                        do_idle();
                        e_shutdown_reason sr=bot_get_shutdown_state();
                        if(E_SD_DIE==sr || E_SD_RESTART==sr) {
                            error="Bot shutdown request";
                            return false;
                        }
                        my_sleep(500);
                        goto identd_fail6;
                    }
                    break;
                }
                break;
            }
            if(!func_name.compare("bot_fullname")) {
                if(NULL==this_module)
                    return true;

                if(args.size()!=1)
                    return bad_num_args(debug,error);
                size_t arg=1;
                string name=get_s_arg(this_module,args,arg++,debug,error,ok);
                if(!ok)
                    return false;
                got_func=true;

                irc_fullname=name;
                break;
            }
            if(!func_name.compare("bot_auth")) {
                if(NULL==this_module)
                    return true;

                if(args.size()!=0)
                    return bad_num_args(debug,error);
                //size_t arg=1;
                if(!ok)
                    return false;
                got_func=true;
                break;
            }
            if(!func_name.compare("ban_mask")) {
                if(NULL==this_module)
                    return true;

                if(args.size()!=3)
                    return bad_num_args(debug,error);
                size_t arg=1;
                int priority=static_cast<int>(get_i_arg(this_module,args,arg++,debug,error,ok));
                string channel=get_s_arg(this_module,args,arg++,debug,error,ok);
                string mask=get_s_arg(this_module,args,arg++,debug,error,ok);
                if(!ok)
                    return false;
                got_func=true;

                if(!irc_got_op(channel))
                    break;
                irc_ban(channel,mask,priority);
                break;
            }
            if(!func_name.compare("bind")) {
                if(NULL==this_module)
                    return true;

                if(args.size()!=6)
                    return bad_num_args(debug,error);
                size_t arg=1;

                c_delay_command c;
                c.type=c_delay_command::DELAY_BIND;

                c.mod_name=this_module->get_mod_name();
                c.type_mask=get_s_arg(this_module,args,arg++,debug,error,ok);
                c.user_mask=get_s_arg(this_module,args,arg++,debug,error,ok);
                c.channel_mask=get_s_arg(this_module,args,arg++,debug,error,ok);
                c.line_mask=get_s_arg(this_module,args,arg++,debug,error,ok);
                c.before=TO_BOOL(get_i_arg(this_module,args,arg++,debug,error,ok));
                c.func=get_s_arg(this_module,args,arg++,debug,error,ok);
                if(!ok)
                    return false;
                got_func=true;

                c.debug=debug;

                delay_commands.push_back(c);
                break;
            }
            if(!func_name.compare("bind_ex")) {
                if(NULL==this_module)
                    return true;

                if(args.size()!=6)
                    return bad_num_args(debug,error);
                size_t arg=1;

                c_delay_command c;
                c.type=c_delay_command::DELAY_BIND_EX;

                c.mod_name=this_module->get_mod_name();
                c.type_mask=get_s_arg(this_module,args,arg++,debug,error,ok);
                c.user_mask=get_s_arg(this_module,args,arg++,debug,error,ok);
                c.channel_mask=get_s_arg(this_module,args,arg++,debug,error,ok);
                c.line_mask=get_s_arg(this_module,args,arg++,debug,error,ok);
                c.before=TO_BOOL(get_i_arg(this_module,args,arg++,debug,error,ok));
                c.func=get_s_arg(this_module,args,arg++,debug,error,ok);
                if(!ok)
                    return false;
                got_func=true;

                c.debug=debug;

                delay_commands.push_back(c);
                break;
            }
        case 'c':
            if(!func_name.compare("change_nick")) {
                if(NULL==this_module)
                    return true;

                if(args.size()!=1)
                    return bad_num_args(debug,error);
                size_t arg=1;
                string nick=get_s_arg(this_module,args,arg++,debug,error,ok);
                if(!ok)
                    return false;
                got_func=true;

                if(!nick.compare(irc_nick))
                    break;

                string msg="NICK ";
                msg+=nick;
                irc_putserv(c_tracked_msg_to_server(msg),true,HIGH_PRIORITY);
                break;
            }
            if(!func_name.compare("check_dynamic_bans")) {
                if(NULL==this_module)
                    return true;

                if(args.size()<2)
                    return bad_num_args(debug,error);
                size_t arg=1;
                string channel_=get_s_arg(this_module,args,arg++,debug,error,ok);
                string nick_=get_s_arg(this_module,args,arg++,debug,error,ok);
                bool got_reason=false;
                if(arg<args.size()+1)
                    got_reason=true;
                if(!ok)
                    return false;
                got_func=true;

                string result="0";
                string reason="";

                bool was_dns=true;

                vector<c_channel_def>::iterator i1;
                for(i1=r_channel_defs.begin(); i1!=r_channel_defs.end(); ++i1) {
                    if(!irc_cmp_strings_case_insensitive((*i1).channel_name,channel_)) {
                        string nick=nick_;
                        string ident=irc_get_ident(nick);
                        string host=irc_get_host(nick);

                        map<string,string>::iterator i2;
                        for(i2=(*i1).dynamic_bans.begin(); i2!=(*i1).dynamic_bans.end(); ++i2) {
                            string mask=(*i2).first;
                            if(logic_check_mask(mask,nick+"!"+ident+"@"+host,"*","*",was_dns)) {
                                result="1";
                                reason=(*i2).second;
                                break;
                            }
                        }
                        break;
                    }
                }
                if(!was_dns)
                    result+="2";
                retval=MY_NEW c_var_string(result);
                if(got_reason) {
                    c_var_base* reason_=MY_NEW c_var_string(reason);
                    if(!set_arg(this_module,args,reason_,arg++,debug,error,ok)) {
                        MY_DELETE(reason_);
                        reason_=NULL;
                        return false;
                    }
                    if(!ok)
                        return false;
                }
                break;
            }
            if(!func_name.compare("chan_mode")) {
                if(NULL==this_module)
                    return true;

                if(args.size()!=3)
                    return bad_num_args(debug,error);
                size_t arg=1;
                int priority=static_cast<int>(get_i_arg(this_module,args,arg++,debug,error,ok));
                string channel=get_s_arg(this_module,args,arg++,debug,error,ok);
                string mode=get_s_arg(this_module,args,arg++,debug,error,ok);
                if(!ok)
                    return false;
                got_func=true;

                irc_chan_mode(channel,mode,priority);
                break;
            }
            if(!func_name.compare("count")) {
                if(NULL==this_module)
                    return true;

                if(args.size()!=1)
                    return bad_num_args(debug,error);
                size_t arg=1;
                c_var_base* b=get_b_arg(this_module,args,arg++,debug,error,ok);
                if(!ok)
                    return false;
                got_func=true;

                int res=1;
                if(b && b->get_var_type()==VARTYPE_ARRAY) {
                    c_var_array* a=dynamic_cast<c_var_array*>(b);
                    if(a) {
                        res=var_keeper->array_num_children(this_module,a);
                    }
                    a->clear_unset();
                }
                if(!b || b->get_var_type()==VARTYPE_UNSET)
                    res=0;

                retval=MY_NEW c_var_int(res);
                break;
            }
            if(!func_name.compare("chr")) {
                if(NULL==this_module)
                    return true;

                if(args.size()!=1)
                    return bad_num_args(debug,error);
                size_t arg=1;
                char ch=static_cast<char>(get_i_arg(this_module,args,arg++,debug,error,ok));
                if(!ok)
                    return false;
                got_func=true;

                string res=(string)""+ch;
                retval=MY_NEW c_var_string(res);
                break;
            }
            break;
        case 'a':
            if(!func_name.compare("allow_redirect")) {
                if(NULL==this_module)
                    return true;

                if(args.size()!=1)
                    return bad_num_args(debug,error);
                size_t arg=1;
                bool allow=TO_BOOL(get_i_arg(this_module,args,arg++,debug,error,ok));
                if(!ok)
                    return false;
                got_func=true;

                irc_set_redir(allow);
                break;
            }
            if(!func_name.compare("admin_msg")) {
                if(NULL==this_module)
                    return true;

                if(args.size()!=3)
                    return bad_num_args(debug,error);
                size_t arg=1;
                string type=get_s_arg(this_module,args,arg++,debug,error,ok);
                string mask=get_s_arg(this_module,args,arg++,debug,error,ok);
                string msg=get_s_arg(this_module,args,arg++,debug,error,ok);
                if(!ok)
                    return false;
                got_func=true;

                if(type.find("@telnet@",0)!=string::npos) {
                    if(mask.find("@",0)==string::npos) {
                        // local message
                        dcc_broadcast(mask,msg,"","");
                    } else {
                        // BOTNET broadcast message
                        string user, bot;

                        user=mask;
                        user.erase(user.find("@",0),user.length()-user.find("@",0));

                        bot=mask;
                        bot.erase(0,bot.find("@",0)+1);

                        botnet_admin_msg(user,bot,msg);
                    }
                }
                break;
            }
            break;
        case 'w':
            if(!func_name.compare("wait_if_error")) {
                if(NULL==this_module)
                    return true;

                try {

                    e_shutdown_reason sr=bot_get_shutdown_state(); // forward, because of goto

                    if(args.size()!=0)
                        return bad_num_args(debug,error);
                    if(!ok)
                        throw false;
                    got_func=true;

                    int got_nick=0;

                    my_time_t first_try=my_time(NULL);
                    int auth_state=0;
                    vector<string>::iterator ii;
                    bool error=false;

                l2:
                    eval_cleanup_timeout();
                    {
                        my_time_t now=my_time(NULL);
                        while(msgs_from_server.empty()) {
                            if(now+10<=my_time(NULL))
                                break;
                            my_sleep(10);
                            if(irc_loop_putserv()) {
                                int reason=irc_loop_process_input();
                                irc_disconnect();
                                eval_cleanup_timeout();
                                if(reason==2)
                                    goto throttled;
                                eval_cleanup_timeout();
                                break;
                            }
                            logic_loop_identd();
                            do_idle();
                        }
                    }

                    ii=irc_nicks.begin();
                    for(int i1=1; ii!=irc_nicks.end(); ++ii, ++i1) {
                        if(i1<=last_nick_idx)
                            continue;
                        got_nick=1;
                        irc_nick=*ii;
                        last_nick_idx++;
                        break;
                    }
                    if(got_nick==0) {
                        last_nick_idx=0;
                        got_nick=2;
                        goto l2;
                    }
                    if(got_nick==2) {
                        last_nick_idx=0;
                        string d="I don't have any nicks! I am dying now!";
                        if(this_module->get_logger()) {
                            this_module->get_logger()->log_print1(E_LOG_BOT,LOG_SCOPE_GLOBAL,"",d,true);
                            this_module->get_logger()->log_print1(E_LOG_IRC,LOG_SCOPE_GLOBAL,"",d,true);
                        }
                        irc_disconnect();

                        bot_shutdown_action(E_SD_DIE,"SELF");

                        error="Bot shutdown request";
                        throw false;
                    }

                    first_try=my_time(NULL);
                    auth_state=0;
                l1:
                    eval_cleanup_timeout();
                    my_time(&irc_last_keepalive_detection);    // do not check for timeout by this method
                    sr=bot_get_shutdown_state();
                    if(E_SD_DIE==sr || E_SD_RESTART==sr) {
                        error="Bot shutdown request";
                        throw false;
                    }
                    if(first_try+60*3<my_time(NULL)) {         // check for timeout here
                        skip_block=false;
                        throw true;
                    }
                    if(auth_state==0) {
                        //irc_put("PASS *",HIGH_PRIORITY);
                        auth_state++;
                    }
                    if(auth_state==1) {
                        string msg="USER ";
                        if(!irc_ident.compare("*")) {
                            irc_ident=str_to_lower(irc_nick);
                        erase_again:
                            for(unsigned int ii=0; ii<irc_ident.length(); ++ii)
                                if(irc_ident[ii]<'a' || irc_ident[ii]>'z') {
                                    irc_ident.erase(ii,1);
                                    goto erase_again;
                                }
                            if(irc_ident.empty())
                                irc_ident=str_to_lower(irc_nick);
                        }
                        msg+=irc_ident;
                        msg+=" 0 \"";
                        msg+=irc_server_host;
                        msg+="\" :";
                        msg+=irc_fullname;
                        irc_put(msg,HIGH_PRIORITY);

                        auth_state++;

                        if(irc_loop_putserv()) {
                            irc_disconnect();
                            skip_block=false;
                            throw true;
                        }
                    }
                    my_sleep(50);
                    error=false;
                    while(msgs_to_server.begin()!=msgs_to_server.end() /*&& last_msg.wait_for_response*/) {
                        eval_cleanup_timeout();

                        my_sleep(1);
                        if(irc_loop_putserv()) {
                            int reason=irc_loop_process_input();
                            irc_disconnect();
                            eval_cleanup_timeout();
                            if(reason==2)
                                goto throttled;
                            eval_cleanup_timeout();
                            break;
                        }
                        logic_loop_identd();
                        irc_loop_process_input();
                        do_idle();
                    }
                    if(error) {
                        skip_block=false;
                        throw true;
                    }
                    logic_loop_identd();
                    do_idle();
                    if(irc_loop_process_input()==2) {
                        irc_disconnect();
                throttled:
                        last_nick_idx=0;

                        string d="Reconnecting too fast: sleeping for 60 seconds...";
                        if(this_module->get_logger()) {
                            this_module->get_logger()->log_print1(E_LOG_BOT,LOG_SCOPE_GLOBAL,"",d,true);
                            this_module->get_logger()->log_print1(E_LOG_IRC,LOG_SCOPE_GLOBAL,"",d,true);
                        }

                        utils_deep_sleep(60);

                        skip_block=false;
                        throw true;
                    }

                    if(!irc_socket.cmp() && last_msg.response[0]==0) {
                        last_nick_idx=0;
                        skip_block=false;
                        throw true;
                    }
                    size_t i1_;
                    for(i1_=0; i1_<strlen(last_msg.response); ++i1_)
                        if(last_msg.response[i1_]==0x20)
                            break;
                    int resp=-1;
                    char redirect[1024*2];
                    if(last_msg.response[i1_]==0x20 &&
                        last_msg.response[i1_+1]>='0' && last_msg.response[i1_+1]<='9' &&
                        last_msg.response[i1_+2]>='0' && last_msg.response[i1_+2]<='9' &&
                        last_msg.response[i1_+3]>='0' && last_msg.response[i1_+3]<='9') {
                            resp=(last_msg.response[i1_+1]-'0')*100+(last_msg.response[i1_+2]-'0')*10+(last_msg.response[i1_+3]-'0');
                            redirect[0]=0;
                            if(strlen(last_msg.response)<sizeof(redirect))
                                strcpy(redirect,last_msg.response);
                            last_msg.response[0]=0;
                        }

                    if(resp==-1) {
                        if(auth_state==4) {
                            auth_state++;
                        }
                        if(auth_state==3) {
                            for(int i1=0; i1<100; ) {
                                do_idle();
                                i1+=50;
                                if(irc_loop_putserv()) {
                                    irc_disconnect();
                                    break;
                                }
                                irc_loop_process_input();
                            }
                            if(irc_loop_putserv()) {
                                irc_disconnect();
                                skip_block=false;
                                throw true;
                            }
                            string nick="NICK ";
                            nick+=irc_nick;
                            irc_put(nick,HIGH_PRIORITY);
                            auth_state++;
                        }
                        if(auth_state==2) {
                            auth_state++;
                            for(int i1=0; i1<1*1000; ) {
                                do_idle();
                                my_sleep(50);
                                i1+=50;
                                if(irc_loop_putserv()) {
                                    irc_disconnect();
                                    break;
                                }
                                irc_loop_process_input();
                            }
                        }
                        if(irc_loop_putserv()) {
                            irc_loop_process_input();
                            irc_disconnect();
                            last_nick_idx=0;
                            skip_block=false;
                            throw true;
                        }
                        logic_loop_identd();
                        do_idle();
                        goto l1;
                    }
                    if(resp>=400 && resp<=599) {
                        // error reply
                        if(resp==433 || resp==437) {
                            // nick already in use / temporarily unavailable
                            int got_nick=0;
                        l3:
                            vector<string>::iterator ii=irc_nicks.begin();
                            for(int i1=1; ii!=irc_nicks.end(); ++ii, ++i1) {
                                if(i1<=last_nick_idx)
                                    continue;
                                got_nick=1;
                                irc_nick=*ii;
                                last_nick_idx++;
                                break;
                            }
                            if(got_nick==0) {
                                last_nick_idx=0;
                                got_nick=2;
                                goto l3;
                            }
                            if(got_nick==2) {
                                last_nick_idx=0;
                                string d="I don't have any nicks! I am dying now!";
                                if(this_module->get_logger()) {
                                    this_module->get_logger()->log_print1(E_LOG_BOT,LOG_SCOPE_GLOBAL,"",d,true);
                                    this_module->get_logger()->log_print1(E_LOG_IRC,LOG_SCOPE_GLOBAL,"",d,true);
                                }
                                irc_disconnect();

                                bot_shutdown_action(E_SD_DIE,"SELF");

                                error="Bot shutdown request";
                                throw false;
                            }
                            if(got_nick==1) {
                                string nick="NICK ";
                                nick+=irc_nick;
                                irc_put(nick,HIGH_PRIORITY);
                                //auth_state=2;
                                if(irc_loop_putserv()) {
                                    irc_disconnect();
                                    skip_block=false;
                                    throw true;
                                }
                                irc_loop_process_input();
                                goto l1;
                            }
                            // give up
                            // TODO: do somethig clever!!!
                            bot_shutdown_action(E_SD_DIE,"SELF");

                            error="Bot shutdown request";
                            throw false;
                        }
                    } else {
                        if(resp==10) {
                            // redirect
                            string msg="Received redirect message: ";
                            msg+=redirect;
                            if(this_module->get_logger()) {
                                this_module->get_logger()->log_print1(E_LOG_BOT,LOG_SCOPE_GLOBAL,"",msg,true);
                                this_module->get_logger()->log_print1(E_LOG_IRC,LOG_SCOPE_GLOBAL,"",msg,true);
                            }
                            if(irc_follow_redirs) {
                                string d="Redirecting...";
                                if(this_module->get_logger()) {
                                    this_module->get_logger()->log_print1(E_LOG_BOT,LOG_SCOPE_GLOBAL,"",d,true);
                                    this_module->get_logger()->log_print1(E_LOG_IRC,LOG_SCOPE_GLOBAL,"",d,true);
                                }
                                string msg=redirect;
                                //:irc.server.com 010 BotNick0 irc2.server.com 6669 :Please use this Server/Port instead
                                if(msg.find(" ")!=string::npos)
                                    msg.erase(0,msg.find(" ")+1); // erase server name
                                if(msg.find(" ")!=string::npos)
                                    msg.erase(0,msg.find(" ")+1); // erase numeric reply
                                if(msg.find(" ")!=string::npos)
                                    msg.erase(0,msg.find(" ")+1); // erase my nick
                                irc_recommended_server="";
                                while(msg.length()>0 && msg[0]!=0x20) {
                                    irc_recommended_server+=msg[0];
                                    msg.erase(0,1);
                                }
                                msg.erase(0,1);
                                string port;
                                while(msg.length()>0 && msg[0]!=0x20) {
                                    port+=msg[0];
                                    msg.erase(0,1);
                                }
                                irc_recommended_port=stous(port);

                                last_nick_idx=0;

                                {
                                    irc_disconnect();
                                    irc_recommended_port=0;

                                    skip_block=false;
                                    throw true;
                                }
                            } else {
                                irc_disconnect(); // try another server, coz redirects are dissabled (better said, continue to wait_if_error { content }
                                last_nick_idx=0;
                                skip_block=false;
                                throw true;
                            }
                        }
                        if(first_try+60*3<my_time(NULL)) {
                            last_nick_idx=0;
                            // go to wait_if_error { content }
                            skip_block=false;
                            throw true;
                        }
                        bool got=false;
                        if(resp>=1 && resp<=5)
                            got=true;
                        if(resp>=251 && resp<=255)
                            got=true;
                        if(resp==372 || resp==375)
                            got=true;
                        if(got)
                            identd_shutdown();
                        if(resp==20) {
                            // wait
                            for(int i1=0; i1<500; ) {
                                do_idle();
                                my_sleep(50);
                                i1+=50;
                            }
                            goto l1;
                        }
                        if(!got) {
                            logic_loop_identd();
                            irc_loop_process_input();
                            goto l1;
                        }
                        skip_block=true;
                        throw true;
                    }
                    skip_block=true;
                    throw true;
                } catch(bool e) {
                    eval_cleanup_timeout();
                    return e;
                }
            }
            if(!func_name.compare("work")) {
                if(NULL==this_module)
                    return true;

                if(args.size()!=0)
                    return bad_num_args(debug,error);

                if(this_module->get_mod_name().compare("core")) { // !=
                    error="It is prohibited to use 'work' command, or 'work()' function in other module than 'core' module";
                    return false;
                }

                //size_t arg=1;
                if(!ok)
                    return false;
                got_func=true;

                {
                    c_script_timeout_protector_unlocker stp_unlock;

                    var_keeper->work_issued(this_module->get_session(),true);

                    do_work();

                    var_keeper->work_issued(this_module->get_session(),false);
                }

                e_shutdown_reason sr=bot_get_shutdown_state();
                if(E_SD_DIE==sr || E_SD_RESTART==sr)
                    return true;

                break;
            }
            break;
        case 'u':
            if(!func_name.compare("unban_mask")) {
                if(NULL==this_module)
                    return true;

                if(args.size()!=3)
                    return bad_num_args(debug,error);
                size_t arg=1;
                int priority=static_cast<int>(get_i_arg(this_module,args,arg++,debug,error,ok));
                string channel=get_s_arg(this_module,args,arg++,debug,error,ok);
                string mask=get_s_arg(this_module,args,arg++,debug,error,ok);
                if(!ok)
                    return false;
                got_func=true;

                if(!irc_got_op(channel))
                    break;

                vector<c_online_channel>::iterator i1;
                for(i1=irc_channels.begin(); i1!=irc_channels.end(); ++i1) {
                    if(irc_cmp_strings_case_insensitive((*i1).name,channel))
                        continue;
                    vector<string>::iterator i2;
                    for(i2=(*i1).bans.begin(); i2!=(*i1).bans.end(); ++i2)
                        if(!(*i2).compare(mask)) {
                            (*i1).bans.erase(i2);
                            break;
                        }
                    break;
                }

                irc_unban(channel,mask,priority);
                break;
            }
            if(!func_name.compare("unbind")) {
                if(NULL==this_module)
                    return true;

                if(args.size()!=6)
                    return bad_num_args(debug,error);
                size_t arg=1;

                c_delay_command c;
                c.type=c_delay_command::DELAY_UNBIND;

                c.mod_name=this_module->get_mod_name();
                c.type_mask=get_s_arg(this_module,args,arg++,debug,error,ok);
                c.user_mask=get_s_arg(this_module,args,arg++,debug,error,ok);
                c.channel_mask=get_s_arg(this_module,args,arg++,debug,error,ok);
                c.line_mask=get_s_arg(this_module,args,arg++,debug,error,ok);
                c.before=TO_BOOL(get_i_arg(this_module,args,arg++,debug,error,ok));
                c.func=get_s_arg(this_module,args,arg++,debug,error,ok);
                if(!ok)
                    return false;
                got_func=true;

                c.debug=debug;

                delay_commands.push_back(c);
                break;
            }
            if(!func_name.compare("unbind_ex")) {
                if(NULL==this_module)
                    return true;

                if(args.size()!=6)
                    return bad_num_args(debug,error);
                size_t arg=1;

                c_delay_command c;
                c.type=c_delay_command::DELAY_UNBIND_EX;

                c.mod_name=this_module->get_mod_name();
                c.type_mask=get_s_arg(this_module,args,arg++,debug,error,ok);
                c.user_mask=get_s_arg(this_module,args,arg++,debug,error,ok);
                c.channel_mask=get_s_arg(this_module,args,arg++,debug,error,ok);
                c.line_mask=get_s_arg(this_module,args,arg++,debug,error,ok);
                c.before=TO_BOOL(get_i_arg(this_module,args,arg++,debug,error,ok));
                c.func=get_s_arg(this_module,args,arg++,debug,error,ok);
                if(!ok)
                    return false;
                got_func=true;

                c.debug=debug;

                delay_commands.push_back(c);
                break;
            }
            if(!func_name.compare("unload_module")) {
                if(NULL==this_module)
                    return true;

                if(args.size()!=1)
                    return bad_num_args(debug,error);
                size_t arg=1;

                c_delay_command c;
                c.type=c_delay_command::DELAY_UNLOAD;

                c.mod_name=get_s_arg(this_module,args,arg++,debug,error,ok);
                if(!ok)
                    return false;
                got_func=true;

                c.debug=debug;

                delay_commands.push_back(c);
                break;
            }
            break;
        case 'h':
            if(!func_name.compare("host")) {
                if(NULL==this_module)
                    return true;

                if(args.size()!=1)
                    return bad_num_args(debug,error);
                size_t arg=1;
                string nick=get_s_arg(this_module,args,arg++,debug,error,ok);
                if(!ok)
                    return false;
                got_func=true;

                retval=MY_NEW c_var_string(irc_get_host(nick));
                break;
            }
            if(!func_name.compare("have_func")) {
                if(NULL==this_module)
                    return true;

                if(args.size()!=1)
                    return bad_num_args(debug,error);
                size_t arg=1;
                string name=get_s_arg(this_module,args,arg++,debug,error,ok);
                if(!ok)
                    return false;
                got_func=true;

                retval=MY_NEW c_var_bool(is_built_in_function(name));
                break;
            }
            break;
        case 'g':
            if(!func_name.compare("get_chan_mode")) {
                if(NULL==this_module)
                    return true;

                if(args.size()!=1)
                    return bad_num_args(debug,error);
                size_t arg=1;
                string channel=get_s_arg(this_module,args,arg++,debug,error,ok);
                if(!ok)
                    return false;
                got_func=true;

                string topic; // dummy
                retval=MY_NEW c_var_string(irc_get_chan_mode(channel,topic));
                break;
            }
            if(!func_name.compare("get_chan_topic")) {
                if(NULL==this_module)
                    return true;

                if(args.size()!=1)
                    return bad_num_args(debug,error);
                size_t arg=1;
                string channel=get_s_arg(this_module,args,arg++,debug,error,ok);
                if(!ok)
                    return false;
                got_func=true;

                string topic;
                irc_get_chan_mode(channel,topic);

                retval=MY_NEW c_var_string(topic);
                break;
            }
            if(!func_name.compare("get_nick_mode")) {
                if(NULL==this_module)
                    return true;

                if(args.size()!=2)
                    return bad_num_args(debug,error);
                size_t arg=1;
                string channel=get_s_arg(this_module,args,arg++,debug,error,ok);
                string nick=get_s_arg(this_module,args,arg++,debug,error,ok);
                if(!ok)
                    return false;
                got_func=true;

                retval=MY_NEW c_var_string(irc_get_mode(channel,nick));
                break;
            }
            if(!func_name.compare("get_bot_nick")) {
                if(NULL==this_module)
                    return true;

                if(args.size()!=0)
                    return bad_num_args(debug,error);
                //size_t arg=1;
                if(!ok)
                    return false;
                got_func=true;

                if(irc_nick_confirmed)
                    retval=MY_NEW c_var_string(irc_nick);
                else
                    retval=MY_NEW c_var_bool(false);
                break;
            }
            if(!func_name.compare("get_event_id")) {
                if(NULL==this_module)
                    return true;

                if(args.size()!=0)
                    return bad_num_args(debug,error);
                //size_t arg=1;
                if(!ok)
                    return false;
                got_func=true;

                retval=MY_NEW c_var_string(event_id);
                break;
            }
            if(!func_name.compare("get_bot_name")) {
                if(NULL==this_module)
                    return true;

                if(args.size()!=0)
                    return bad_num_args(debug,error);
                //size_t arg=1;
                if(!ok)
                    return false;
                got_func=true;

                extern string botnet_botname;
                retval=MY_NEW c_var_string(botnet_botname);
                break;
            }
            if(!func_name.compare("getrandmax")) {
                if(NULL==this_module)
                    return true;

                if(args.size()!=0)
                    return bad_num_args(debug,error);
                //size_t arg=1;
                if(!ok)
                    return false;
                got_func=true;

                retval=MY_NEW c_var_int(MAX_S_LONG_LONG);
                break;
            }
            if(!func_name.compare("get_current_irc_server")) {
                if(NULL==this_module)
                    return true;

                if(args.size()!=2)
                    return bad_num_args(debug,error);

                size_t arg=1;

                c_var_string* server=MY_NEW c_var_string(irc_server_host);
                if(!set_arg(this_module,args,server,arg++,debug,error,ok)) {
                    MY_DELETE(server);
                    return false;
                }
                if(!ok)
                    return false;

                c_var_int* port=MY_NEW c_var_int(irc_server_port);
                if(!set_arg(this_module,args,port,arg++,debug,error,ok)) {
                    MY_DELETE(port);
                    return false;
                }

                if(!ok)
                    return false;
                got_func=true;
                break;
            }
            break;
        case '_':
            if(!func_name.compare("_breakpoint")) {
                if(NULL==this_module)
                    return true;

                size_t arg=1;

                int number=-1;
                string comment;

                if(args.size()>0)
                    number=static_cast<int>(get_i_arg(this_module,args,arg++,debug,error,ok));

                if(args.size()>1)
                    comment=get_s_arg(this_module,args,arg++,debug,error,ok);

                /*
                if(!ok)
                    return false;
                */
                got_func=true;

                // to force not to generate Intel(TM)'s compiler remark
                number++;
                number--;

                {
#ifdef _DEBUG
                    // DEBUG: set breakpoint here!
                    int x=0;
                    x++;
#endif
                }
                break;
            }
            if(!func_name.compare("_dump_vars")) {
                if(NULL==this_module)
                    return true;

                if(args.size()!=0)
                    return bad_num_args(debug,error);

                got_func=true;
                var_keeper->_dump(this_module,__LINE__,"_dump_vars-function");
                {
#ifdef _DEBUG
                    // DEBUG: set breakpoint here!
                    int x=0;
                    x++;
#endif
                }
                break;
            }
            break;
        case 'f':
            if(!func_name.compare("fopen")) {
                if(NULL==this_module)
                    return true;

                if(args.size()!=2)
                    return bad_num_args(debug,error);
                size_t arg=1;
                string name=get_s_arg(this_module,args,arg++,debug,error,ok);
                string mode=get_s_arg(this_module,args,arg++,debug,error,ok);
                if(!ok)
                    return false;
                got_func=true;

                if(!this_module) {
                    retval=MY_NEW c_var_bool(false);
                    break;
                }
                retval=this_module->open_file(name,mode,debug);
                break;
            }
            if(!func_name.compare("fgets")) {
                if(NULL==this_module)
                    return true;

                if(args.size()!=2 && args.size()!=1)
                    return bad_num_args(debug,error);
                size_t arg=1;
                int hnd=static_cast<int>(get_i_arg(this_module,args,arg++,debug,error,ok));
                MY_LONG_LONG max_len_val=1024;
                if(args.size()==2)
                    max_len_val=get_i_arg(this_module,args,arg++,debug,error,ok);
                if(!ok)
                    return false;

                int max_len=static_cast<int>(max_len_val);
                if(max_len<0)
                    max_len=0;

                got_func=true;

                if(!this_module) {
                    retval=MY_NEW c_var_bool(false);
                    break;
                }
                FILE* f=this_module->get_file(hnd);
                if(!f) {
                    string log="Invalid resource handle passed to 'fgets()' function; "+debug;
                    if(this_module->get_logger())
                        this_module->get_logger()->log_print1(E_LOG_DEBUG,LOG_SCOPE_DEFAULT,"",log);
                    retval=MY_NEW c_var_bool(false);
                    break;
                }
                char* buff=NULL;
                buff=(char*)my_malloc(max_len+1+1);
                if(!buff) {
                    string log="Error while allocating buffer for 'fgets()' function; "+debug;
                    if(this_module->get_logger())
                        this_module->get_logger()->log_print1(E_LOG_DEBUG,LOG_SCOPE_DEFAULT,"",log);
                    retval=MY_NEW c_var_bool(false);
                    break;
                }
                buff[0]=buff[max_len]=0;
                char* res=fgets(buff,max_len,f);
                if(res) {
                    buff[max_len]=0;
                    retval=MY_NEW c_var_string(buff);
                } else
                    retval=MY_NEW c_var_bool(false);
                my_free(buff);

                break;
            }
            if(!func_name.compare("fread")) {
                if(NULL==this_module)
                    return true;

                if(args.size()!=2)
                    return bad_num_args(debug,error);
                size_t arg=1;
                int hnd=static_cast<int>(get_i_arg(this_module,args,arg++,debug,error,ok));
                MY_LONG_LONG max_len_val=get_i_arg(this_module,args,arg++,debug,error,ok);
                if(!ok)
                    return false;
                got_func=true;

                FILE* f=this_module->get_file(hnd);
                if(!f) {
                    string log="Invalid resource handle passed to 'fread()' function; "+debug;
                    if(this_module->get_logger())
                        this_module->get_logger()->log_print1(E_LOG_DEBUG,LOG_SCOPE_DEFAULT,"",log);
                    retval=MY_NEW c_var_bool(false);
                    break;
                }

                int my_errno;

                MY_LONG_LONG pos=my_ftell(f,my_errno);
                if(0!=my_errno) {
                    string log="Error while getting file size, errno=";
                    log+=tos(my_errno);
                    log+="; "+debug;
                    log=GET_STR_DEBUG_INFO_EX(log);
                    if(this_module->get_logger())
                        this_module->get_logger()->log_print1(E_LOG_DEBUG,LOG_SCOPE_DEFAULT,"",log);
                    retval=MY_NEW c_var_bool(false);
                    break;
                }

                MY_LONG_LONG size=my_filesize(f,my_errno);
                if(-1==size || 0!=my_errno) {
                    string log="Error while getting file size, errno=";
                    log+=tos(my_errno);
                    log+="; "+debug;
                    log=GET_STR_DEBUG_INFO_EX(log);
                    if(this_module->get_logger())
                        this_module->get_logger()->log_print1(E_LOG_DEBUG,LOG_SCOPE_DEFAULT,"",log);
                    retval=MY_NEW c_var_bool(false);
                    break;
                }

                if(size-pos<max_len_val) {
                    max_len_val=size-pos;
                    if(max_len_val<=0) { // beyond EOF
                        retval=MY_NEW c_var_string("");
                        break;
                    }
                }

                {
                    size_t tmp;
                    memset(&tmp,0xff,sizeof(tmp));
                    tmp/=2;
                    MY_LONG_LONG tmp2=tmp;
                    if(max_len_val>tmp2)
                        max_len_val=tmp2;
                }

                char* buff=NULL;
                buff=(char*)my_malloc(static_cast<size_t>(max_len_val)+1+1);
                if(!buff) {
                    string log="Error while allocating buffer for 'fread()' function; "+debug;
                    if(this_module->get_logger())
                        this_module->get_logger()->log_print1(E_LOG_DEBUG,LOG_SCOPE_DEFAULT,"",log);
                    retval=MY_NEW c_var_bool(false);
                    break;
                }
                buff[0]=buff[max_len_val]=0;
                size_t res=fread(buff,1,static_cast<size_t>(max_len_val),f);
                buff[res]=0;
                retval=MY_NEW c_var_string(buff);
                my_free(buff);

                break;
            }
            if(!func_name.compare("fputs")) {
                if(NULL==this_module)
                    return true;

                if(args.size()!=2 && args.size()!=3)
                    return bad_num_args(debug,error);
                size_t arg=1;
                int hnd=static_cast<int>(get_i_arg(this_module,args,arg++,debug,error,ok));
                string str=get_s_arg(this_module,args,arg++,debug,error,ok);
                MY_LONG_LONG max_len_val;
                if(args.size()>=3)
                    max_len_val=get_i_arg(this_module,args,arg++,debug,error,ok);
                else
                    max_len_val=str.length();
                if(!ok)
                    return false;
                got_func=true;

                int max_len=static_cast<int>(max_len_val);
                if(max_len<0)
                    max_len=0;

                if(!this_module) {
                    retval=MY_NEW c_var_bool(false);
                    break;
                }
                FILE* f=this_module->get_file(hnd);
                if(!f) {
                    string log="Invalid resource handle passed to 'fputs()' function; "+debug;
                    if(this_module->get_logger())
                        this_module->get_logger()->log_print1(E_LOG_DEBUG,LOG_SCOPE_DEFAULT,"",log);
                    retval=MY_NEW c_var_bool(false);
                    break;
                }

                if(static_cast<MY_LONG_LONG>(str.length())>max_len)
                    str.erase(max_len,str.length()-max_len);

                size_t res=fwrite(str.c_str(),1,max_len,f);

                retval=MY_NEW c_var_int(static_cast<MY_LONG_LONG>(res));

                break;
            }
            if(!func_name.compare("fclose")) {
                if(NULL==this_module)
                    return true;

                if(args.size()!=1)
                    return bad_num_args(debug,error);
                size_t arg=1;
                int hnd=static_cast<int>(get_i_arg(this_module,args,arg++,debug,error,ok));
                if(!ok)
                    return false;
                got_func=true;

                FILE* f=this_module->get_file(hnd);
                if(!f) {
                    string log="Invalid resource handle passed to 'fclose()' function; "+debug;
                    if(this_module->get_logger())
                        this_module->get_logger()->log_print1(E_LOG_DEBUG,LOG_SCOPE_DEFAULT,"",log);
                    retval=MY_NEW c_var_bool(false);
                    break;
                }

                retval=this_module->close_file(hnd);
                break;
            }
            if(!func_name.compare("fseek")) {
                if(NULL==this_module)
                    return true;

                if(args.size()!=2 && args.size()!=3)
                    return bad_num_args(debug,error);
                size_t arg=1;
                int hnd=static_cast<int>(get_i_arg(this_module,args,arg++,debug,error,ok));
                MY_LONG_LONG offset=get_i_arg(this_module,args,arg++,debug,error,ok);
                int whence=MY_SEEK_SET;
                if(args.size()>=3)
                    whence=static_cast<int>(get_i_arg(this_module,args,arg++,debug,error,ok));
                if(!ok)
                    return false;
                got_func=true;

                if(!this_module) {
                    retval=MY_NEW c_var_int(-1);
                    break;
                }
                FILE* f=this_module->get_file(hnd);
                if(!f) {
                    string log="Invalid resource handle passed to 'fseek()' function; "+debug;
                    if(this_module->get_logger())
                        this_module->get_logger()->log_print1(E_LOG_DEBUG,LOG_SCOPE_DEFAULT,"",log);
                    retval=MY_NEW c_var_int(-1);
                    break;
                }
                string off=tos(offset);
                int my_errno;
                switch(whence) {
                    case MY_SEEK_SET:
                        {
                            int res=my_fseek(f,offset,SEEK_SET,my_errno);
                            if(res!=0) {
                                res=-1;

                                string log=(string)"my_fseek() function returned non-zero result for offset "+off+", SEEK_SET; errno=";
                                log+=tos(my_errno);
                                log+="; ";
                                log+=debug;
                                if(this_module->get_logger())
                                    this_module->get_logger()->log_print1(E_LOG_DEBUG,LOG_SCOPE_DEFAULT,"",log);
                            }
                            retval=MY_NEW c_var_int(res);
                        }
                        break;
                    case MY_SEEK_CUR:
                        {
                            int res=my_fseek(f,offset,SEEK_CUR,my_errno);
                            if(res!=0) {
                                res=-1;

                                string log=(string)"my_fseek() function returned non-zero result for offset "+off+", SEEK_CUR; errno=";
                                log+=tos(my_errno);
                                log+="; ";
                                log+=debug;
                                if(this_module->get_logger())
                                    this_module->get_logger()->log_print1(E_LOG_DEBUG,LOG_SCOPE_DEFAULT,"",log);
                            }
                            retval=MY_NEW c_var_int(res);
                        }
                        break;
                    case MY_SEEK_END:
                        {
                            int res=my_fseek(f,offset,SEEK_END,my_errno);
                            if(res!=0) {
                                res=-1;

                                string log=(string)"my_fseek() function returned non-zero result for offset "+off+", SEEK_END; errno=";
                                log+=tos(my_errno);
                                log+="; ";
                                log+=debug;
                                if(this_module->get_logger())
                                    this_module->get_logger()->log_print1(E_LOG_DEBUG,LOG_SCOPE_DEFAULT,"",log);
                            }
                            retval=MY_NEW c_var_int(res);
                        }
                        break;
                    default:
                        {
                            string log="Invalid 'whence' value passed to 'fseek()' function (3rd argument); "+debug;
                            if(this_module->get_logger())
                                this_module->get_logger()->log_print1(E_LOG_DEBUG,LOG_SCOPE_DEFAULT,"",log);
                            retval=MY_NEW c_var_int(-1);
                        }
                        break;
                }
                break;
            }
            if(!func_name.compare("ftell")) {
                if(NULL==this_module)
                    return true;

                if(args.size()!=1)
                    return bad_num_args(debug,error);
                size_t arg=1;
                int hnd=static_cast<int>(get_i_arg(this_module,args,arg++,debug,error,ok));
                if(!ok)
                    return false;
                got_func=true;

                if(!this_module) {
                    retval=MY_NEW c_var_bool(false);
                    break;
                }
                FILE* f=this_module->get_file(hnd);
                if(!f) {
                    string log="Invalid resource handle passed to 'ftell()' function; "+debug;
                    if(this_module->get_logger())
                        this_module->get_logger()->log_print1(E_LOG_DEBUG,LOG_SCOPE_DEFAULT,"",log);
                    retval=MY_NEW c_var_bool(false);
                    break;
                }
                retval=MY_NEW c_var_int(ftell(f));
                break;
            }
            if(!func_name.compare("feof")) {
                if(NULL==this_module)
                    return true;

                if(args.size()!=1)
                    return bad_num_args(debug,error);
                size_t arg=1;
                int hnd=static_cast<int>(get_i_arg(this_module,args,arg++,debug,error,ok));
                if(!ok)
                    return false;
                got_func=true;

                if(!this_module) {
                    retval=MY_NEW c_var_bool(false);
                    break;
                }
                FILE* f=this_module->get_file(hnd);
                if(!f) {
                    string log="Invalid resource handle passed to 'feof()' function; "+debug;
                    if(this_module->get_logger())
                        this_module->get_logger()->log_print1(E_LOG_DEBUG,LOG_SCOPE_DEFAULT,"",log);
                    retval=MY_NEW c_var_bool(false);
                    break;
                }
                retval=MY_NEW c_var_bool(feof(f)?true:false);
                break;
            }
            if(!func_name.compare("filesize")) {
                if(NULL==this_module)
                    return true;

                if(args.size()!=1)
                    return bad_num_args(debug,error);
                size_t arg=1;
                c_var_base* fd=get_b_arg(this_module,args,arg++,debug,error,ok);
                if(!ok)
                    return false;
                got_func=true;

                if(!this_module) {
                    retval=MY_NEW c_var_bool(false);
                    break;
                }
                if(fd && fd->get_var_type()==VARTYPE_INT) {
                    MY_LONG_LONG hnd=dynamic_cast<c_var_int*>(fd)->get();
                    FILE* f=NULL;
                    if(hnd<=MAX_SIGNED_INT && hnd>0)
                        f=this_module->get_file(static_cast<int>(hnd));
                    if(!f) {
                        string log="Invalid resource handle passed to 'filesize()' function; "+debug;
                        if(this_module->get_logger())
                            this_module->get_logger()->log_print1(E_LOG_DEBUG,LOG_SCOPE_DEFAULT,"",log);
                        retval=MY_NEW c_var_bool(false);
                        break;
                    }
                    int my_errno;
                    MY_LONG_LONG size=my_filesize(f,my_errno);
                    if(size<0 || 0!=my_errno) {
                        string log="Error getting size of file in 'filesize()' function; "+debug;
                        if(this_module->get_logger())
                            this_module->get_logger()->log_print1(E_LOG_DEBUG,LOG_SCOPE_DEFAULT,"",log);
                        retval=MY_NEW c_var_bool(false);
                        break;
                    }
                    retval=MY_NEW c_var_int(size);
                    break;
                }
                if(fd &&  fd->get_var_type()==VARTYPE_STRING) {
                    string hnd=dynamic_cast<c_var_string*>(fd)->get();
                    FILE* f=fopen(hnd.c_str(),"rb");
                    if(!f) {
                        string log="Could not open file in 'filesize()' function; "+debug;
                        if(this_module->get_logger())
                            this_module->get_logger()->log_print1(E_LOG_DEBUG,LOG_SCOPE_DEFAULT,"",log);
                        retval=MY_NEW c_var_bool(false);
                        break;
                    }
                    int my_errno;
                    MY_LONG_LONG size=my_filesize(f,my_errno);
                    fclose(f);
                    if(size<0 || 0!=my_errno) {
                        string log="Error getting size of file in 'filesize()' function; "+debug;
                        if(this_module->get_logger())
                            this_module->get_logger()->log_print1(E_LOG_DEBUG,LOG_SCOPE_DEFAULT,"",log);
                        retval=MY_NEW c_var_bool(false);
                        break;
                    }
                    retval=MY_NEW c_var_int(size);
                    break;
                }
                string log="Invalid argument passed to function; can be file descriptor returned by 'fopen()', or STRING as name of the file; "+debug;
                if(this_module->get_logger())
                    this_module->get_logger()->log_print1(E_LOG_DEBUG,LOG_SCOPE_DEFAULT,"",log);
                retval=MY_NEW c_var_bool(false);
                break;
            }
            break;
    }

    if(NULL==this_module)
        return false;

    if(!got_func) {
        error="Internal error: Invalid built-in function '";
        error+=func_name;
        error+="'";
    }

    return got_func && ok;
}

bool c_functions::add_function(c_function* func, string& error)
{
    if(!func->is_okay()) {
        error="Bad function passed to ";
        error+=__my_func__;
        error+="()";
        return false;
    }

    map<string,c_function*>::iterator f=functions.find(func->get_name());
    if(f!=functions.end()) {
        error="Overloading of functions is not supported (or it is in conflict with built-in function) - name duplicate function is '";
        error+=func->get_name();
        error+="'";
        return false;
    }

    pair<string,c_function*> p;
    p.first=func->get_name();
    p.second=func;
    functions.insert(p);

    return true;
}

bool c_functions::parse_call(string declaration, vector<c_func_arg>& args, string& func_name, string& error)
{
    error="";
    args.clear();
    func_name="";

    int ctx=0;
    c_func_arg arg;
    int arg_num=1;
    bool has_arguments=false;

    stack<char> one_arg;

    for(size_t i1=0; i1<declaration.length(); ++i1) {
        if(0x20==declaration[i1]) {
            if(0==ctx || -1==ctx) {
                ctx=-1;
                continue;
            }
            if(1==ctx)
                continue;
            if(2==ctx || -2==ctx) {
                ctx=-2;
                continue;
            }
        }
        bool is_alphanum=false;
        if((declaration[i1]>='a' && declaration[i1]<='z') ||
           (declaration[i1]>='A' && declaration[i1]<='Z') ||
           (declaration[i1]>='0' && declaration[i1]<='1') ||
           '_'==declaration[i1])
           is_alphanum=true;
        if(0==ctx && is_alphanum) { // name of function
            func_name+=(string)""+declaration[i1];
            continue;
        }
        if((0==ctx && -1==ctx) && '('==declaration[i1]) {
            if(func_name.empty()) {
                error="Missing name of function";
                return false;
            }
            ctx++;
            continue;
        }
        if(1==ctx && '$'==declaration[i1]) { // argument '$' prefix
            arg.arg_expression="$";
            has_arguments=true;
            ctx++;
            continue;
        }
        if(1==ctx && is_alphanum && arg.arg_expression.empty()) {
            // it is function call
            arg.arg_expression+=declaration[i1];
            has_arguments=true;
            ctx++;
            continue;
        }
        if((2==ctx || -2==ctx) && ','==declaration[i1]) { // end of argument
            if(arg.arg_expression.length()<2) {
                // it is an empty string, or just an "$"
                error="In call of function '";
                error+=func_name;
                error+="' argument #";
                error+=tos(arg_num);
                error+=" is erroneous";
                return false;
            }
            args.push_back(arg);
            arg.clear();
            ctx=1;
            arg_num++;
            continue;
        }
        if(2==ctx) { // ...else: it is still argument
            if(!one_arg.empty() && '\\'==one_arg.top()) {
                one_arg.pop();
                arg.arg_expression+=declaration[i1];
                continue;
            }
            if('('==declaration[i1]) {
                one_arg.push('(');
                arg.arg_expression+='(';
                continue;
            }
            if('"'==declaration[i1]) {
                if(!one_arg.empty() && '"'==one_arg.top()) {
                    one_arg.pop();
                    arg.arg_expression+='"';
                    continue;
                }
                one_arg.push('"');
                arg.arg_expression+='"';
                continue;
            }
            if('\\'==declaration[i1]) {
                one_arg.push('\\');
                arg.arg_expression+='\\';
                continue;
            }
            if(')'==declaration[i1]) {
                if(!one_arg.empty() && '('==one_arg.top()) {
                    one_arg.pop();
                    arg.arg_expression+=')';
                }
                if(one_arg.empty()) {
                    if(has_arguments && arg.arg_expression.length()<2) {
                        // it is an empty string, or just an "$"
                        error="In call of function '";
                        error+=func_name;
                        error+="' argument #";
                        error+=tos(arg_num);
                        error+=" is erroneous";
                        return false;
                    }
                    if(has_arguments)
                        args.push_back(arg);
                    ctx++;
                    continue;
                }
                // else
                error="Syntax error";
                return false;
            }
        }
        if(3==ctx) {
            error="Invalid call of function (syntax error) '";
            error+=func_name;
            error+="' - some character(s) trail(s) after ending ')' character";
            return false;
        }
        error="Invalid function call (syntax error)";
        return false;
    }
    return true;
}

void c_functions::get_replication(vector<c_function*>& result, const string& partner, int type)
{
    result.clear();
    map<string,c_function*>::iterator i1;
    for(i1=functions.begin(); i1!=functions.end(); ++i1) {
        if((*i1).second->get_replication(partner,type)) {
            result.push_back((*i1).second);
        }
    }
}

void c_functions::dump_to_file(FILE* f)
{
    map<string,c_function*>::iterator i1;
    for(i1=functions.begin(); i1!=functions.end(); ++i1) {
        (*i1).second->dump_to_file(f);
    }
}

void c_functions::remove_function(const string& name)
{
    map<string,c_function*>::iterator f=functions.find(name);
    if(f!=functions.end()) {
        if((*f).second)
            (*f).second->destroy(true,true);
        MY_DELETE((*f).second);
        functions.erase(f);
    }
}

bool c_functions::get_function_prototype(string func_name, vector<c_func_arg>& args_)
{
    args_.clear();
    map<string,c_function*>::iterator f=functions.find(func_name);
    if(f!=functions.end()) {
        (*f).second->get_prototype(args_);
        return true;
    } else
        return false;
}

void report_pcre_error(c_module* this_module, const char* prefix, const c_pcre& re, const char* error_report, const string& debug, string& error)
{
    int eo;
    string es;
    re.get_compile_error_info(eo,es);

    if(this_module->get_logger()) {
        ostringstream log;
        log << prefix << ": "
            << error_report
            << "; PCRE library did not compiled the pattern; PCRE error offset: "
            << tos(eo)
            << "; PCRE error string: \""
            << es.c_str()
            << "\"; Additional info: "
            << debug.c_str();
        error=log.str();
    }
}

/*!
    \brief Implements "preg_grep" php function
    \version 1
    \date 2008-02-06
    \author VooDooMan
    \param this_module Module
    \param pattern $pattern (1st argument)
    \param input $input (2nd argument)
    \param flags $flags (3rd argument)
    \param retval Resulting array (left intact if return is false)
    \param debug Debug string
    \param error In case of error, there will be message
*/
MY_STATIC bool php_preg_grep(c_module* this_module, const string& debug, const string& pattern, c_var_array* input, MY_LONG_LONG flags, c_var_base*& retval, string& error)
{
    if(flags & ~PREG_GREP_INVERT) {
        string log="ERROR: Invalid value ($flags) passed as third argument; could be only numerical zero, or 'PREG_GREP_INVERT'; Additional info: ";
        log+=debug;
        if(this_module->get_logger()) {
            this_module->get_logger()->log_print1(E_LOG_DEBUG,LOG_SCOPE_MODULE,"",log);
        }
        error=log;
        return false;
    }

    c_pcre re(pattern);
    if(!re.is_okay()) {
        report_pcre_error(this_module,"FATAL",re,"Invalid value ($pattern) passed as first argument",debug,error);
        return false;
    }

    vector<pair<const c_index_base*,string> > data;
    if(!var_keeper->array_dump_to_string_vector_wo_unset_and_array(this_module,input,data)) {
        error="ERROR: Invalid value ($input) passed as second argument; bad array to process";
        return false;
    }

    c_var_array* res=MY_NEW c_var_array();

    vector<pair<const c_index_base*,string> >::iterator i1;
    for(i1=data.begin(); i1!=data.end(); ++i1) {
        vector<vector<pair<int,string> > > dummy;
        size_t matches=re.exec((*i1).second,0,0,dummy,1);
        if(re.get_last_error()) {
            MY_DELETE(res);
            error="PCRE returned an error: ";
            error+=re.get_last_error_string();
            return false;
        }
        if(
            ((flags & PREG_GREP_INVERT) && 0==matches)
            ||
            (!(flags & PREG_GREP_INVERT) && 0!=matches)
            ) {
            c_index_base* idx=(*i1).first->copy();
            c_var_string* val=MY_NEW c_var_string((*i1).second);
            bool ok=res->set(this_module,idx,val,false,debug);
            if(!ok) {
                MY_DELETE(res);
                error="Unable to create array member";
                return false;
            }
        }
    }

    retval=res;
    return true;
}

/*!
    \brief Implements "preg_match_all" php function
    \version 1
    \date 2008-02-08
    \author VooDooMan
    \param this_module Module
    \param pattern RegExp pattern
    \param subject Subject
    \param flags Flags
    \param offset Starting offset in bytes
    \param debug Debug string
    \param error In case of error, there will be message
    \param retval Resulting match count (left intact if return is false)
    \param matches_array Resulting array (left intact if return is false)
*/
MY_STATIC bool php_preg_match_all(c_module* this_module, const string& debug, const string& pattern, const string& subject, c_var_array* &matches_array, MY_LONG_LONG flags, MY_LONG_LONG offset, c_var_base*& retval, string& error)
{
    if(flags & ~(PREG_PATTERN_ORDER | PREG_SET_ORDER | PREG_OFFSET_CAPTURE)) {
        string log="ERROR: Invalid value ($flags) passed as fourth argument; could be only numerical zero ('PREG_PATTERN_ORDER' is assumed as the default), or PREG_PATTERN_ORDER, or PREG_SET_ORDER, or PREG_OFFSET_CAPTURE, or their bitwise OR combination (using '|' operator); Additional info: ";
        log+=debug;
        if(this_module->get_logger()) {
            this_module->get_logger()->log_print1(E_LOG_DEBUG,LOG_SCOPE_MODULE,"",log);
        }
        error=log;
        return false;
    }

    // sanity check
    if((flags & (PREG_PATTERN_ORDER | PREG_SET_ORDER)) == (PREG_PATTERN_ORDER | PREG_SET_ORDER)) {
        string log="WARNING: Value ($flags) passed as fourth argument: 'PREG_PATTERN_ORDER | PREG_SET_ORDER' is not sane, choose only one of them, not both; Additional info: ";
        log+=debug;
        if(this_module->get_logger()) {
            this_module->get_logger()->log_print1(E_LOG_DEBUG,LOG_SCOPE_MODULE,"",log);
        }
        retval=MY_NEW c_var_bool(false);
        return true;
    }

    if((flags & (PREG_PATTERN_ORDER | PREG_SET_ORDER)) == 0) {
        flags=PREG_PATTERN_ORDER; // default value is assumed
    }

    if(flags & PREG_PATTERN_ORDER) {
        string log="ERROR: Value ($flags) passed as fourth argument is either 'PREG_PATTERN_ORDER' or defaulted to; this is not yet implemented; Please, use 'PREG_SET_ORDER'; Additional info: ";
        log+=debug;
        if(this_module->get_logger()) {
            this_module->get_logger()->log_print1(E_LOG_DEBUG,LOG_SCOPE_MODULE,"",log);
        }
        retval=MY_NEW c_var_bool(false);
        return true;
    }

    c_pcre re(pattern);
    if(!re.is_okay()) {
        report_pcre_error(this_module,"FATAL",re,"Invalid value ($pattern) passed as first argument",debug,error);
        return false;
    }

    vector<vector<pair<int,string> > > matches;
    size_t cnt=re.exec(subject,offset,0,matches,MAX_SIGNED_INT);
    if(re.get_last_error()) {
        report_pcre_error(this_module,"FATAL",re,"PCRE error",debug,error);
        return false;
    }

    if(php_do_preg_match_all(this_module,debug,matches_array,flags,error,matches)) {
        retval=MY_NEW c_var_int(cnt);
        return true;
    }

    return false;
}

/*!
    \brief Implements "preg_match" php function
    \version 1
    \date 2008-02-08
    \author VooDooMan
    \param this_module Module
    \param pattern RegExp pattern
    \param subject Subject
    \param matches_array Array that will receive output
    \param flags Flags
    \param offset Starting offset in bytes
    \param retval Return value
    \param debug Debug string
    \param error In case of error, there will be message
    \return Returns false for error
*/
MY_STATIC bool php_preg_match(c_module* this_module, const string& debug, const string& pattern, const string& subject, c_var_array* &matches_array, MY_LONG_LONG flags, MY_LONG_LONG offset, c_var_base*& retval, string& error)
{
    if(flags & ~(PREG_OFFSET_CAPTURE)) {
        string log="ERROR: Invalid value ($flags) passed as fourth argument; could be only numerical zero or 'PREG_OFFSET_CAPTURE'; Additional info: ";
        log+=debug;
        if(this_module->get_logger()) {
            this_module->get_logger()->log_print1(E_LOG_DEBUG,LOG_SCOPE_MODULE,"",log);
        }
        error=log;
        return false;
    }

    c_pcre re(pattern);
    if(!re.is_okay()) {
        report_pcre_error(this_module,"FATAL",re,"Invalid value ($pattern) passed as first argument",debug,error);
        return false;
    }

    vector<vector<pair<int,string> > > matches;
    size_t cnt=re.exec(subject,offset,0,matches,1);
    if(re.get_last_error()) {
        report_pcre_error(this_module,"FATAL",re,"PCRE error",debug,error);
        return false;
    }

    if(php_do_preg_match(this_module,debug,matches_array,flags,error,matches)) {
        retval=MY_NEW c_var_int(cnt);
        return true;
    }

    return false;
}

MY_STATIC bool php_preg_split(c_module* this_module, const string& debug, const string& pattern, const string& subject, MY_LONG_LONG limit, MY_LONG_LONG flags, c_var_base*& retval, string& error)
{
    if(flags & ~(PREG_SPLIT_NO_EMPTY | PREG_SPLIT_DELIM_CAPTURE | PREG_SPLIT_OFFSET_CAPTURE)) {
        string log="ERROR: Invalid value ($flags) passed as fourth argument; could be either PREG_SPLIT_NO_EMPTY, or PREG_SPLIT_DELIM_CAPTURE, or PREG_SPLIT_OFFSET_CAPTURE, or their bitwise OR combination (using '|' operator); Additional info: ";
        log+=debug;
        if(this_module->get_logger()) {
            this_module->get_logger()->log_print1(E_LOG_DEBUG,LOG_SCOPE_MODULE,"",log);
        }
        error=log;
        return false;
    }

    if(flags & PREG_SPLIT_NO_EMPTY || flags & PREG_SPLIT_DELIM_CAPTURE) {
        string log="ERROR: Value ($flags) passed as fourth argument is either 'PREG_SPLIT_NO_EMPTY' and/or 'PREG_SPLIT_DELIM_CAPTURE'; this is not yet implemented; Additional info: ";
        log+=debug;
        if(this_module->get_logger()) {
            this_module->get_logger()->log_print1(E_LOG_DEBUG,LOG_SCOPE_MODULE,"",log);
        }
        retval=MY_NEW c_var_bool(false);
        return true;
    }

    c_pcre re(pattern);
    if(!re.is_okay()) {
        report_pcre_error(this_module,"FATAL",re,"Invalid value ($pattern) passed as first argument",debug,error);
        return false;
    }

    vector<vector<pair<int,string> > > matches;
    (void)re.exec(subject,0,0,matches,MAX_SIGNED_INT);
    if(re.get_last_error()) {
        report_pcre_error(this_module,"FATAL",re,"PCRE error",debug,error);
        return false;
    }

    c_var_array* res1=MY_NEW c_var_array();

    bool ok;

    MY_LONG_LONG idx1=0;

    MY_LONG_LONG start=0;
    MY_LONG_LONG limit_cnt=0;

    vector<vector<pair<int,string> > >::iterator i1;
    for(i1=matches.begin(); ; ++i1) {
        bool tail=false;
        if(matches.end()==i1) {
            tail=true;
            if(static_cast<MY_LONG_LONG>(subject.length())-start<0)
                break;
            vector<pair<int,string> > v;
            v.push_back(make_pair(static_cast<int>(start)+1,"")); // dummy
            matches.push_back(v);
            i1=matches.end();
            --i1;
        }
        vector<pair<int,string> >::iterator i2;
        for(i2=(*i1).begin(); i2!=(*i1).end(); ++i2) {
            if(start<static_cast<MY_LONG_LONG>((*i2).first) || !(*i2).second.empty()) {
                c_index_base* vi2=MY_NEW c_index_int(idx1++);

                c_var_string* val;

                if((0!=limit && -1!=limit) && ++limit_cnt>=limit) {
                    val=MY_NEW c_var_string(subject.substr(static_cast<string::size_type>(start),string::npos));
                } else {
                    if(!tail)
                        val=MY_NEW c_var_string(subject.substr(static_cast<string::size_type>(start),(*i2).first-static_cast<string::size_type>(start)));
                    else {
                        if(static_cast<MY_LONG_LONG>(subject.length())-start>0)
                            val=MY_NEW c_var_string(subject.substr(static_cast<string::size_type>(start),string::npos));
                        else
                            val=MY_NEW c_var_string();
                    }
                }

                if(flags & PREG_SPLIT_OFFSET_CAPTURE) {
                    c_index_base* i;
                    c_var_array* a=MY_NEW c_var_array();

                    i=MY_NEW c_index_int(0);
                    ok=a->set(this_module,i,val,false,debug);
                    if(!ok) {
                        MY_DELETE(val);
                        MY_DELETE(a);
                        MY_DELETE(res1);
                        error="Unable to create array member";
                        return false;
                    }

                    i=MY_NEW c_index_int(1);
                    c_var_int* val=MY_NEW c_var_int(start);
                    ok=a->set(this_module,i,val,false,debug);
                    if(!ok) {
                        MY_DELETE(a);
                        MY_DELETE(res1);
                        error="Unable to create array member";
                        return false;
                    }

                    ok=res1->set(this_module,vi2,a,false,debug);
                    if(!ok) {
                        MY_DELETE(a);
                        MY_DELETE(res1);
                        error="Unable to create array member";
                        return false;
                    }
                } else {
                    ok=res1->set(this_module,vi2,val,false,debug);
                    if(!ok) {
                        MY_DELETE(val);
                        MY_DELETE(res1);
                        error="Unable to create array member";
                        return false;
                    }
                }
            }

            if((0!=limit && -1!=limit) && limit_cnt>=limit)
                break;

            start=(*i2).first+(*i2).second.length();
        }

        if(tail)
            break;

        if((0!=limit && -1!=limit) && limit_cnt>=limit)
            break;
    }

    retval=res1;
    return true;
}

MY_STATIC bool make_vector_of_strings_from_var(c_module* this_module, c_var_array* var, vector<string>& result)
{
    vector<pair<const c_index_base*,string> > data;
    if(!var_keeper->array_dump_to_string_vector_wo_unset_and_array(this_module,var,data))
        return false;
    vector<pair<const c_index_base*,string> >::iterator i1;
    for(i1=data.begin(); i1!=data.end(); ++i1)
        result.push_back((*i1).second);

    return true;
}

/*!
    \brief Implements parser for $replacement string for PHP-like 'preg_replace()' function
    \author VooDooMan
    \date 2008-02-10
    \version 1
*/
class c_preg_replacement_parser
{
public:
    enum e_kind
    {
        NUM,        //!< It is a number, e.g. '$2', or '${32}'
        STR,        //!< Is is a string, e.g. 'foo', 'bar', or even 'foobar'
    };              //!< Kind of part of replacement string

private:

    /*!
        \brief Implements base class for type of part of replacement string
        \author VooDooMan
        \date 2008-02-10
        \version 1
    */
    class c_part
    {
    public:

        /*!
            \brief Dummy d-tor
            \author VooDooMan
            \date 2008-02-10
            \version 1
        */
        virtual ~c_part() { ; }
    };

    /*!
        \brief Implements e.g. '$2', or '${32}' of part of replacement string
        \author VooDooMan
        \date 2008-02-10
        \version 1
    */
    class c_part_num : public c_part
    {
        vector<long> m_val;

    public:

        /*!
            \brief C-tor
            \author VooDooMan
            \date 2008-02-10
            \version 1
            \param val Value
        */
        MY_WISH_INLINE c_part_num(long val)
        {
            m_val.push_back(val);
        }

        /*!
            \brief Appends next number to the vector
            \author VooDooMan
            \date 2008-02-10
            \version 1
            \param val Value
        */
        MY_WISH_INLINE void append(long val)
        {
            m_val.push_back(val);
        }

        /*!
            \brief Returns a value
            \author VooDooMan
            \date 2008-02-10
            \version 1
        */
        MY_WISH_INLINE const vector<long>& get() const
        {
            return m_val;
        }

        /*!
            \brief Dummy d-tor
            \author VooDooMan
            \date 2008-02-10
            \version 1
        */
        virtual ~c_part_num() { ; }
    };

    /*!
        \brief Implements e.g. 'foo', or 'bar', or even 'foobar' of part of replacement string
        \author VooDooMan
        \date 2008-02-10
        \version 1
    */
    class c_part_str : public c_part
    {
        string m_val;

    public:

        /*!
            \brief C-tor
            \author VooDooMan
            \date 2008-02-10
            \version 1
            \param val Value
        */
        MY_WISH_INLINE c_part_str(const string& val)
            : m_val(val)
        { ; }

        /*!
            \brief Returns a value
            \author VooDooMan
            \date 2008-02-10
            \version 1
        */
        MY_WISH_INLINE const string& get() const
        {
            return m_val;
        }

        /*!
            \brief Dummy d-tor
            \author VooDooMan
            \date 2008-02-10
            \version 1
        */
        virtual ~c_part_str() { ; }
    };

    vector<pair<e_kind,c_part*> > m_stream;     //!< Parsed stream
    bool m_error;                               //!< 'true' if parse error

public:

    /*!
        \brief C-tor
        \author VooDooMan
        \date 2008-02-10
        \version 1
        \param replacement Replacement string to parse
    */
    c_preg_replacement_parser(const string& replacement)
        : m_error(false)
    {
        enum
        {
            NOTHING,    //!< Nothing
            DOLLAR,     //!< We are after '$' character, expecting '{' or single number to follow
            BRACKET,    //!< We are inside '{' bracket
            BACKSLASH,  //!< We are after backslash (escape)
        };

        string str;
        c_part_num* last_num=NULL;

        int ctx=NOTHING;

        for(string::size_type i1=0; i1<replacement.length(); ++i1) {
            char x=replacement[i1];

            switch(ctx) {
                case NOTHING:
                    {
                        if('$'==x) {
                            if(!str.empty()) {
                                m_stream.push_back(make_pair(STR,MY_NEW c_part_str(str)));
                                str.clear();
                            }

                            ctx=DOLLAR;
                            continue;
                        }
                        if('\\'==x) {
                            str.append("\\");

                            ctx=BACKSLASH;
                            continue;
                        }

                        last_num=NULL;

                        str.append((string)""+x);
                        continue;
                    }
                case DOLLAR:
                    {
                        if(x>='0' && x<='9') {
                            if(NULL!=last_num)
                                last_num->append(x-'0');
                            else
                                m_stream.push_back(make_pair(NUM,last_num = MY_NEW c_part_num(x-'0')));
                            ctx=NOTHING;
                            continue;
                        }

                        if('{'==x) {
                            ctx=BRACKET;
                            continue;
                        }

                        m_error=true;
                        return;
                    }
                case BRACKET:
                    {
                        if(x>='0' && x<='9') {
                            str.append((string)""+x);
                            continue;
                        }

                        if('}'==x) {
                            if(str.empty()) {
                                m_error=true;
                                return;
                            }

                            long num;
                            int rc=sscanf(str.c_str(),"%ld",&num);
                            if(1!=rc) {
                                m_error=true;
                                return;
                            }

                            str.clear();

                            if(NULL!=last_num)
                                last_num->append(num);
                            else
                                m_stream.push_back(make_pair(NUM,last_num = MY_NEW c_part_num(num)));

                            ctx=NOTHING;
                            continue;
                        }

                        m_error=true;
                        return;
                    }
                case BACKSLASH:
                    {
                        str.append((string)""+x);
                        last_num=NULL;
                        ctx=NOTHING;
                    }
            }
        }

        if(NOTHING==ctx && !str.empty()) {
            m_stream.push_back(make_pair(STR,MY_NEW c_part_str(str)));
            str.clear();
        }

        m_error= NOTHING != ctx;
    }

    /*!
        \brief D-tor, cleans-up
        \author VooDooMan
        \date 2008-02-10
        \version 1
    */
    MY_WISH_INLINE ~c_preg_replacement_parser()
    {
        vector<pair<e_kind,c_part*> >::iterator i1;
        for(i1=m_stream.begin(); i1!=m_stream.end(); ++i1)
            MY_DELETE((*i1).second);
    }

    /*!
        \brief Reports parse error
        \author VooDooMan
        \date 2008-02-10
        \version 1
        \retval true Okay
        \retval false Parse error
    */
    MY_WISH_INLINE bool is_okay()
    {
        return !m_error;
    }

    /*!
        \brief Reports parse error
        \author VooDooMan
        \date 2008-02-10
        \version 1
    */
    e_kind get(size_t idx, string& str, vector<long>& num)
    {
        MY_ASSERT(false == m_error);
        if(idx>=m_stream.size()) {
            str.clear();
            return STR;
        }
        switch(m_stream[idx].first) {
            case NUM:
                num=dynamic_cast<c_part_num*>(m_stream[idx].second)->get();
                return NUM;
            case STR:
                str=dynamic_cast<c_part_str*>(m_stream[idx].second)->get();
                return STR;
        }
        MY_ASSERT_NOTREACH();
        num.clear();
        return NUM;
    }
};

/*!
    \brief Implements position remapping of string
    \author VooDooMan
    \version 1
    \date 2008-02-10
*/
class c_string_remapper
{
    /*!
        \brief Tree node
        \author VooDooMan
        \version 1
        \date 2008-02-11
    */
    struct c_node
    {
        char m_data;                    //!< Data, only valid if 'm_valid==true'
        string m_str;                   //!< Inserted string, only valid if 'm_valid==false'
        bool m_valid;                   //!< 'false' if it was invalidated

        /*!
            \brief C-tor
            \author VooDooMan
            \version 1
            \date 2008-02-11
            \param data Data
        */
        MY_WISH_INLINE c_node(char data)
            : m_data(data)
            , m_str()
            , m_valid(true)
        { ; }

        /*!
            \brief Appends data to 'out'
            \author VooDooMan
            \version 1
            \date 2008-02-11
            \param out Updated string
        */
        MY_WISH_INLINE void get(ostringstream& out) const
        {
            if(!m_valid) {
                out.write(m_str.c_str(),static_cast<streamsize>(m_str.length()));
                return;
            }

            out.put(m_data);
        }

        /*!
            \brief Replaces portion of string
            \author VooDooMan
            \version 1
            \date 2008-02-11
            \param str String to be put
        */
        MY_WISH_INLINE void replace(const string& str)
        {
            m_valid=false;
            m_str=str;
        }

        /*!
            \brief Invalidates the node
            \author VooDooMan
            \version 1
            \date 2008-02-11
        */
        MY_WISH_INLINE void invalidate()
        {
            m_valid=false;
        }

        /*!
            \brief Returns length
            \author VooDooMan
            \version 1
            \date 2008-02-11
        */
        MY_WISH_INLINE string::size_type length()
        {
            return m_valid ? 1 : m_str.length();
        }

        /*!
            \brief Cuts characters from beginning
            \author VooDooMan
            \version 1
            \date 2008-02-11
            \param cnt How many char's to cut from the beginning
            \return Returns how many char's were successfuly cut-off
        */
        MY_WISH_INLINE string::size_type cut(string::size_type cnt)
        {
            if(m_valid) {
                m_valid=false;
                return 1;
            }
            if(m_str.length()<cnt)
                cnt=m_str.length();
            m_str.erase(0,cnt);
            return cnt;
        }

        /*!
            \brief Replaces part of data
            \author VooDooMan
            \version 1
            \date 2008-02-11
            \param start Offset
            \param with Replacement
            \return Returns how many char's were successfuly replaced
        */
        MY_WISH_INLINE string::size_type replace(string::size_type start, const string& with)
        {
            if(m_valid) {
                m_valid=false;
                m_str=with;
                return with.length();
            }

            string::size_type cnt=0;
            string::size_type i2=0;
            for(string::size_type i1=start; i1<m_str.length() && i2<with.length(); ++i1, ++i2, ++cnt)
                m_str[i1]=with[i2];
            return cnt;
        }
    };

    map<string::size_type, c_node> m_str;       //!< String

    typedef map<string::size_type, c_node>::iterator it;
    typedef map<string::size_type, c_node>::const_iterator cit;

public:

    /*!
        \brief C-tor
        \author VooDooMan
        \version 1
        \date 2008-02-10
        \param str Initial string
    */
    MY_WISH_INLINE c_string_remapper(const string& str)
    {
        for(string::size_type i1=0; i1<str.length(); ++i1)
            m_str.insert(make_pair(i1,c_node(str[i1])));
    }

    /*!
        \brief Returns current state of string
        \author VooDooMan
        \version 1
        \date 2008-02-10
    */
    MY_WISH_INLINE string get() const
    {
        ostringstream os;
        for(cit i1=m_str.begin(); i1!=m_str.end(); ++i1)
            (*i1).second.get(os);
        return os.str();
    }

    /*!
        \brief Replaces portion of string
        \author VooDooMan
        \version 2
        \date 2008-02-10, 2008-02-11
        \param start Position to start of replacing
        \param with To replace region with
        \param cnt How many char's to cut-off
    */
    MY_WISH_INLINE void replace(string::size_type start, string::size_type cnt, string with)
    {
        it i1=m_str.find(start);
        it i2=i1;
        while(0!=cnt) {
            if(m_str.end()==i1)
                break;
            cnt-=(*(i1++)).second.cut(cnt);
        };
        (*i2).second.replace(with);
    }
};

/*!
    \brief Implements "preg_replace" php function
    \version 1
    \date 2008-02-08
    \author VooDooMan
    \param this_module Module
    \param pattern RegExp pattern
    \param subject Subject
    \param debug Debug string
    \param error In case of error, there will be message
    \param retval Resulting match count (left intact if return is false)
    \param replacement Replacement (user-provided)
    \param limit Limit (user-provided)
    \param count Returns count of replacements for user
*/
MY_STATIC bool php_preg_replace(c_module* this_module, const string& debug, c_var_base* pattern, c_var_base* replacement, c_var_base* subject, size_t limit, c_var_base*& retval, c_var_int* &count, string& error)
{
    bool ok=true;

    {
        // BEGIN: sanity checks
        if(NULL==pattern || NULL==replacement || NULL==subject) {
            error=GET_STR_DEBUG_INFO_EX("Internal error, please report this problem!");
            return false;
        }
        if(!ok)
            return false;
        // END  : sanity checks
    }

    // pattern
    vector<string> pattern_vec;
    switch(pattern->get_var_type()) {
        case VARTYPE_STRING:
            pattern_vec.push_back(pattern->to_string(this_module));
            break;
        case VARTYPE_ARRAY:
            {
                if(!make_vector_of_strings_from_var(this_module,dynamic_cast<c_var_array*>(pattern),pattern_vec)) {
                    // fall thru
                } else
                    break;
            }
        default:
            error="ERROR: Invalid value ($pattern) passed as first argument; bad array to process; it can be STRING, or ARRAY of STRINGs or INTs";
            return false;
    }

    // replacement
    vector<string> replacement_vec;
    switch(replacement->get_var_type()) {
        case VARTYPE_STRING:
            replacement_vec.push_back(replacement->to_string(this_module));
            break;
        case VARTYPE_ARRAY:
            {
                if(!make_vector_of_strings_from_var(this_module,dynamic_cast<c_var_array*>(replacement),replacement_vec)) {
                    // fall thru
                } else
                    break;
            }
        default:
            error="ERROR: Invalid value ($replacement) passed as second argument; bad array to process; it can be STRING, or ARRAY of STRINGs or INTs";
            return false;
    }

    bool return_array=false;

    // subject
    vector<string> subject_vec;
    switch(subject->get_var_type()) {
        case VARTYPE_STRING:
            subject_vec.push_back(subject->to_string(this_module));
            break;
        case VARTYPE_ARRAY:
            {
                return_array=true;
                if(!make_vector_of_strings_from_var(this_module,dynamic_cast<c_var_array*>(subject),subject_vec)) {
                    // fall thru
                } else
                    break;
            }
        default:
            error="ERROR: Invalid value ($subject) passed as third argument; bad array to process; it can be STRING, or ARRAY of STRINGs or INTs";
            return false;
    }

    string res_str;
    c_var_array* res_arr= return_array ? MY_NEW c_var_array() : NULL;

    MY_LONG_LONG idx=0;
    MY_LONG_LONG res_cnt=0;

    if(pattern_vec.empty()) {
        if(return_array) {
            vector<string>::iterator si1;
            for(si1=subject_vec.begin(); si1!=subject_vec.end(); ++si1) {
                c_index_base* i=MY_NEW c_index_int(idx++);
                c_var_string* val=MY_NEW c_var_string(*si1);
                ok=res_arr->set(this_module,i,val,false,debug);
                if(!ok) {
                    MY_DELETE(val);
                    MY_DELETE(res_arr);
                    error="Unable to create array member";
                    return false;
                }
            }
            retval=res_arr;
            count=MY_NEW c_var_int(0);
            return true;
        } else {
            string res_str2;
            if(!subject_vec.empty())
                res_str2=subject_vec.front();
            retval=MY_NEW c_var_string(res_str2);
            count=MY_NEW c_var_int(0);
            return true;
        }
    }

    vector<string>::iterator i_p;
    vector<string>::iterator i_r;
    vector<string>::iterator i_s;
    for(i_s=subject_vec.begin(); i_s!=subject_vec.end(); ++i_s) {

        c_string_remapper result(*i_s);

        for(i_p=pattern_vec.begin(), i_r=replacement_vec.begin(); i_p!=pattern_vec.end(); ++i_p) {
            string _repl;
            if(replacement_vec.end()!=i_r)
                _repl=*(i_r++);

            c_preg_replacement_parser repl(_repl);
            if(!repl.is_okay()) {
                MY_DELETE(res_arr);

                error="Invalid value ($replacement) passed as second argument; "
                    "It was (zero-based index) #";
                error+=tos(idx);
                error+=" member of $pattern array";

                return false;
            }

            c_pcre re(*i_p);
            if(!re.is_okay()) {
                MY_DELETE(res_arr);

                string log="Invalid value ($pattern) passed as first argument; ";
                log+="It was (zero-based index) #";
                log+=tos(idx);
                log+=" member of $pattern array";

                report_pcre_error(this_module,"FATAL",re,log.c_str(),debug,error);
                return false;
            }

            vector<vector<pair<int,string> > > matches;
            (void)re.exec(*i_s,0,0,matches,MAX_SIZE_T);
            if(re.get_last_error()) {
                MY_DELETE(res_arr);

                string log="PCRE error; ";
                log+="It was (zero-based index) #";
                log+=tos(idx);
                log+=" member of $pattern array";

                report_pcre_error(this_module,"FATAL",re,log.c_str(),debug,error);
                return false;
            }

            string str;
            vector<long> nums;

            vector<vector<pair<int,string> > >::iterator i1;
            for(i1=matches.begin(); i1!=matches.end(); ++i1) {
                size_t limit_cnt=0;
                size_t replacement_idx=0;

                vector<pair<int,string> >::iterator i2;
                for(i2=(*i1).begin(); i2!=(*i1).end(); ++i2) {
                    string::size_type length;
                    {
                        vector<pair<int,string> >::iterator ii2=i2;
                        ++ii2;
                        if((*i1).end()!=ii2 && (*i2).first==(*ii2).first)
                            continue;
                        if((*i1).end()!=ii2)
                            length=(*ii2).first-(*i2).first;
                        else
                            length=(*i2).second.length();
                    }

                    c_preg_replacement_parser::e_kind kind=repl.get(replacement_idx++,str,nums);
                    switch(kind) {
                        case c_preg_replacement_parser::STR:
                            {
                                result.replace((*i2).first,length,str);
                                ++res_cnt;
                                ++limit_cnt;
                            }
                            break;
                        case c_preg_replacement_parser::NUM:
                            {
                                vector<long>::const_iterator i3;
                                for(i3=nums.begin(); i3!=nums.end(); ++i3) {
                                    register long num=*i3;

                                    // "+1" two times: we need to ignore the first result, as it is full match
                                    if(num<=0 || num+1>static_cast<MY_LONG_LONG>((*i1).size()))
                                        continue;
                                    result.replace((*i2).first,length,(*i1)[num-1+1].second);
                                    if(++limit_cnt>=limit)
                                        break;
                                }
                                ++res_cnt;
                            }
                    }
                    if(limit_cnt>=limit)
                        break;
                }
            }
        }
        if(return_array) {
            c_index_base* i=MY_NEW c_index_int(idx++);
            c_var_string* val=MY_NEW c_var_string(result.get());
            ok=res_arr->set(this_module,i,val,false,debug);
            if(!ok) {
                MY_DELETE(val);
                MY_DELETE(res_arr);
                error="Unable to create array member";
                return false;
            }
        } else {
            res_str+=result.get();
        }
    }

    if(return_array)
        retval=res_arr;
    else
        retval=MY_NEW c_var_string(res_str);
    count=MY_NEW c_var_int(res_cnt);
    return true;
}

bool php_do_preg_match(c_module* this_module, const string& debug, c_var_array* &matches_array, MY_LONG_LONG flags, string& error, const vector<vector<pair<int,string> > >& matches)
{
    c_var_array* res1=MY_NEW c_var_array();

    bool ok;

    {
        MY_LONG_LONG idx1=0;

        vector<vector<pair<int,string> > >::const_iterator i1=matches.begin();
        if(matches.end()!=i1) {

            vector<pair<int,string> >::const_iterator i2;
            for(i2=(*i1).begin(); i2!=(*i1).end(); ++i2) {

                c_index_base* vi1=MY_NEW c_index_int(idx1++);

                c_var_string* val=MY_NEW c_var_string((*i2).second);

                if(flags & PREG_OFFSET_CAPTURE) {
                    c_index_base* i;
                    c_var_array* a=MY_NEW c_var_array();

                    i=MY_NEW c_index_int(0);
                    ok=a->set(this_module,i,val,false,debug);
                    if(!ok) {
                        MY_DELETE(val);
                        MY_DELETE(a);
                        MY_DELETE(res1);
                        error="Unable to create array member";
                        return false;
                    }

                    i=MY_NEW c_index_int(1);
                    c_var_int* val=MY_NEW c_var_int((*i2).first);
                    ok=a->set(this_module,i,val,false,debug);
                    if(!ok) {
                        MY_DELETE(a);
                        MY_DELETE(res1);
                        error="Unable to create array member";
                        return false;
                    }

                    ok=res1->set(this_module,vi1,a,false,debug);
                    if(!ok) {
                        MY_DELETE(a);
                        MY_DELETE(res1);
                        error="Unable to create array member";
                        return false;
                    }
                } else {
                    ok=res1->set(this_module,vi1,val,false,debug);
                    if(!ok) {
                        MY_DELETE(val);
                        MY_DELETE(res1);
                        error="Unable to create array member";
                        return false;
                    }
                }
            }
        }
    }

    matches_array=res1;

    return true;
}

bool php_do_preg_match_all(c_module* this_module, const string& debug, c_var_array* &matches_array, MY_LONG_LONG flags, string& error, const vector<vector<pair<int,string> > >& matches)
{
    c_var_array* res1=MY_NEW c_var_array();

    bool ok;

    if(flags & PREG_SET_ORDER) {
        MY_LONG_LONG idx1=0;

        vector<vector<pair<int,string> > >::const_iterator i1;
        for(i1=matches.begin(); i1!=matches.end(); ++i1) {

            c_index_base* vi1=MY_NEW c_index_int(idx1++);

            c_var_array* res2=MY_NEW c_var_array();

            ok=res1->set(this_module,vi1,res2,false,debug);
            if(!ok) {
                MY_DELETE(res2);
                MY_DELETE(res1);
                error="Unable to create array member";
                return false;
            }

            MY_LONG_LONG idx2=0;
            vector<pair<int,string> >::const_iterator i2;
            for(i2=(*i1).begin(); i2!=(*i1).end(); ++i2) {

                c_index_base* vi2=MY_NEW c_index_int(idx2++);

                c_var_string* val=MY_NEW c_var_string((*i2).second);

                if(flags & PREG_OFFSET_CAPTURE) {
                    c_index_base* i;
                    c_var_array* a=MY_NEW c_var_array();

                    i=MY_NEW c_index_int(0);
                    ok=a->set(this_module,i,val,false,debug);
                    if(!ok) {
                        MY_DELETE(val);
                        MY_DELETE(a);
                        MY_DELETE(res1);
                        error="Unable to create array member";
                        return false;
                    }

                    i=MY_NEW c_index_int(1);
                    c_var_int* val=MY_NEW c_var_int((*i2).first);
                    ok=a->set(this_module,i,val,false,debug);
                    if(!ok) {
                        MY_DELETE(a);
                        MY_DELETE(res1);
                        error="Unable to create array member";
                        return false;
                    }

                    ok=res2->set(this_module,vi2,a,false,debug);
                    if(!ok) {
                        MY_DELETE(a);
                        MY_DELETE(res1);
                        error="Unable to create array member";
                        return false;
                    }
                } else {
                    ok=res2->set(this_module,vi2,val,false,debug);
                    if(!ok) {
                        MY_DELETE(val);
                        MY_DELETE(res1);
                        error="Unable to create array member";
                        return false;
                    }
                }
            }
        }
    } else if(flags & PREG_PATTERN_ORDER) {
        // TODO
    }

    matches_array=res1;

    return true;
}

