/***************************************************************************
                          c_functions.h  -  description
                             -------------------
    begin                : Thu May 10 2007
    copyright            : (C) 2004-2007 by Marian VooDooMan Meravy
    email                : ghostvoodooman -a.T= users (dot) sourceforge.net
 ***************************************************************************/

/* $Id: c_functions.h,v 1.14 2008/03/11 21:11:39 ghostvoodooman Exp $ */

/***************************************************************************

VooDoo cIRCle - an IRC (ro)bot
Copyright (C) 2004-2008 by Marian VooDooMan Meravy
  (ghostvoodooman -a.T= users (dot) sourceforge.net)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************/

/*!
    \file
    \brief Class 'c_functions'
    \version $Id: c_functions.h,v 1.14 2008/03/11 21:11:39 ghostvoodooman Exp $
*/

#ifndef C_FUNCTIONS_H__
#define C_FUNCTIONS_H__

#if defined(HAVE_CONFIG_H)
#   include <config.h>
#endif

#include <string>
#include <vector>
#include <stack>
#include <list>
#include <map>
#include <algorithm>

#include "../utils/memleak.h"

#include "../params.h"

#include "../func/c_func_arg.h"
#include "../func/c_function.h"

/*!
    \brief Stores list of functions
    \author VooDooMan
    \version 1
    \date 2005
*/
class c_functions
{
private:

    /*!
        \brief Parses call
        \author VooDooMan
        \version 1
        \date 2005
        \param declaration Calling function declaration
        \param args Returns argument list
        \param func_name Returns name of function
        \param error Returns error string
        \return Returns true if ok
    */
    bool parse_call(string declaration, vector<c_func_arg>& args, string& func_name, string& error);

public:
    map<string,c_function*> functions;               //!< List of user-defined functions

    c_functions();

    c_functions(const c_functions& orig);

    /*!
        \brief Adds new function
        \author VooDooMan
        \version 1
        \date 2005
        \param func Function
        \param error returns error string
        \return Returns true if ok
    */
    bool add_function(c_function* func, string& error);

    /*!
        \brief Removes function
        \author VooDooMan
        \version 1
        \date 2005
        \param name Name of function
    */
    void remove_function(const string& name);

    /*!
        \brief Checks if it is name of built-in function
        \author VooDooMan
        \version 2
        \date 2005, 2007-08-01
        \param func_name Name of function
        \return Returns true if it is built-in function
    */
    static bool is_built_in_function(const string& func_name);

    /*!
        \brief Calls built-in function
        \author VooDooMan
        \version 2
        \date 2005, 2007-08-01
        \param this_module Module name; or NULL if only checking of presence of built-in function
        \param func_name Function name
        \param retval Returns return value
        \param args Arguments
        \param error Returns error string
        \param debug Debug string
        \param skip_block Returns true if next block should be skipped
        \param next_expression Pass here the next expression
        \param vars Caller's variables
        \param st Evaluation stack
        \return Returns true if ok; if 'this_module' is NULL, does not execute actual function, but only returns 'true' if internal function exists
    */
    bool call_built_in_func(c_module* this_module, const string& func_name, c_var_base*& retval, vector<c_func_arg>& args, const string& debug, string& error, bool& skip_block, const string& next_expression, c_var_array* vars, stack<c_eval_iteration>& st);

    /*!
        \brief Gets replication rules for specified partner and type
        \author VooDooMan
        \version 1
        \date 2005
        \param result Returns functions to replicate
        \param partner Name of remote bot
        \param type Flags for replication types
    */
    void get_replication(vector<c_function*>& result, const string& partner, int type);

    /*!
        \brief Dumps functions to file
        \author VooDooMan
        \version 1
        \date 2005
        \param f File descriptor
    */
    void dump_to_file(FILE* f);

    /*!
        \brief Cleans up
        \author VooDooMan
        \version 1
        \date 2005
    */
    void clear();

    /*!
        \brief Destroys all functions
        \author VooDooMan
        \version 1
        \date 2006-06-09
    */
    void destroy_all();

    /*!
        \brief Gets arguments for function
        \author VooDooMan
        \version 1
        \date 2005
        \param func_name Name of function
        \param args_ Returns argument list
        \return Returns true if ok
    */
    bool get_function_prototype(string func_name, vector<c_func_arg>& args_);
};

#endif

