/***************************************************************************
                          c_user.h  -  description
                             -------------------
    begin                : Thu May 10 2007
    copyright            : (C) 2004-2007 by Marian VooDooMan Meravy
    email                : ghostvoodooman -a.T= users (dot) sourceforge.net
 ***************************************************************************/

/* $Id: c_user.h,v 1.15 2008/03/25 12:04:27 ghostvoodooman Exp $ */

/***************************************************************************

VooDoo cIRCle - an IRC (ro)bot
Copyright (C) 2004-2008 by Marian VooDooMan Meravy
  (ghostvoodooman -a.T= users (dot) sourceforge.net)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************/

/*!
    \file
    \brief Class 'c_user'
    \version $Id: c_user.h,v 1.15 2008/03/25 12:04:27 ghostvoodooman Exp $
*/

#ifndef C_USER_H__
#define C_USER_H__

#if defined(HAVE_CONFIG_H)
#   include <config.h>
#endif

#include <string>
#include <vector>
#include <stack>
#include <list>
#include <map>
#include <algorithm>

#include "../utils/memleak.h"

#include "../params.h"

#include "../botnet/c_replication.h"
#include "../irc/c_flood.h"
#include "../dcc/c_dcc_host.h"

/*!
    \brief Stores user options
    \author VooDooMan
    \version 1
    \date 2004
*/
struct c_user
{
    string name;                                    //!< Name of user
    bool host_unknown;                              //!< Is this user "an unknown user"? (if true, this structure contains definition for user that is NOT rekognized - something like "guest" on channel)
    bool host_bot;                                  //!< Is this user a bot?
    bool is_template;                               //!< Is this user a template?
    bool terminator;                                //!< Is this user terminated?
    vector<string> hostmask;                        //!< Host hasks ("nick!ident@host")
    vector<string> fullname;                        //!< Full name masks
    vector<string> access_to_group;                 //!< This user can modify properties of users of these groups
    vector<string> access_to_proc;                  //!< This user can modify procedure commands
    vector<string> access_grant_procedure;          //!< This user can grant access to this procedure to other user
    vector<string> access_usage_procedure;          //!< This user can assign events of user to these procedures
    vector<string> access_grant_group;              //!< This user can add or exclude users to/from these groups
    vector<string> access_to_channel;               //!< This user can modify properties of these channels
    vector<string> access_grant_channel;            //!< This user can add or exclude users to/from these channels
    bool access_grant_partyline;                    //!< Can this user grant/deny access to partyline to other user?
    bool access_to_backup;                          //!< Can this user use .backup command?
    bool access_grant_backup;                       //!< Can this user grant/deny access to .backup command to other user?
    bool access_to_rehash;                          //!< Can this user use .rehash command?
    bool access_grant_rehash;                       //!< Can this user grant/deny access to .rehash command to other user?
    bool access_to_plususer;                        //!< Can this user (or bot) create new user?
    bool access_grant_plususer;                     //!< Can this user (or bot) grant creation of new user to other user?

    bool access_to_can_send_all_users;              //!< Can this user send message / file to all users?
    bool access_grant_can_send_all_users;           //!< Can this user grant right to send message / file to all users?

    bool access_to_plusproc;                        //!< Can this user (or bot) create new procedure?
    bool access_grant_plusproc;                     //!< Can this user (or bot) grant creation of new procedure to other user?

    bool access_to_replication;                     //!< Can this user change replication parameters?
    bool access_grant_replication;                  //!< Can this user (or bot) grant to other user right to change replication parameters?

    bool access_to_restart;                         //!< Has user access to .restart command?
    bool access_grant_restart;                      //!< Can this user grant accesst to .restart command to other user?
    bool access_to_die;                             //!< Has user access to .die command?
    bool access_grant_die;                          //!< Can this user grant accesst to .die command to other user?

    bool access_to_filesystem;                      //!< Has user access to .filesystem command?
    bool access_grant_filesystem;                   //!< Can this user grant accesst to .filesystem command to other user?

    bool access_to_can_send_unknown_users;          //!< Can this user send message / file to unknown users on all bot's channels?
    bool access_grant_can_send_unknown_users;       //!< Can this grant right to other user to send message / file to unknown users on all bot's channels?

    bool access_to_partyline;                       //!< Has this user access to partyline?
    vector<string> groups;                          //!< List of groups that this user is member of
    vector<c_dcc_host> dcc_hosts;                   //!< Hosts for DCC connection for this user

    bool access_to_private;                         //!< Has this user access to modufy private query definition?
    bool access_grant_private;                      //!< Can this user grant accesst to private definition to other user?

    bool access_to_upgrade;                         //!< Has user access to .upgrade command?
    bool access_grant_upgrade;                      //!< Can this user grant accesst to .upgrade command to other user?

    bool access_to_apply;                           //!< Has user access to .apply command?
    bool access_grant_apply;                        //!< Can this user grant accesst to .apply command to other user?

    vector<c_replication> replication;              //!< Informations of replication
    my_time_t last_changed;                            //!< Time of last change
    bool replication_partner;                       //!< Is this user a replication partner? (==bot?)

    c_flood partyline_msg_flood;                    //!< Message flood limit on partyline channels

    map<string,string> meta;

    void clear();

    c_user();
};

#endif

