/***************************************************************************
                          event_dispatch.cpp  -  description
                             -------------------
    begin                : Sun Mar 9 2008
    copyright            : (C) 2008 by Marian VooDooMan Meravy
    email                : ghostvoodooman -a.T= users (dot) sourceforge.net
 ***************************************************************************/

/* $Id: event_dispatch.cpp,v 1.2 2008/03/11 21:11:40 ghostvoodooman Exp $ */

/***************************************************************************

VooDoo cIRCle - an IRC (ro)bot
Copyright (C) 2004-2008 by Marian VooDooMan Meravy
  (ghostvoodooman -a.T= users (dot) sourceforge.net)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************/

/*!
    \file
    \brief Implements dispatching of bound events
    \version $Id: event_dispatch.cpp,v 1.2 2008/03/11 21:11:40 ghostvoodooman Exp $
*/

#include "../internal_state.h"
#include "../irc.h"
#include "../irc/c_online_user.h"
#include "../modules.h"

#include "event_dispatch.h"

#include "../utils/precomp.h"

#if defined(_MSC_VER) || defined(__BORLANDC__)
#   pragma hdrstop
#endif

/*!
    \brief Triggers dispatching of the event on user
    \author VooDooMan
    \version 2
    \date 2004, 2008-03-09
    \param channel Name of channel
    \param target_nick Nick name
    \param source_nick Which nick set the mode?
    \param prefix '-' or '+'
    \param mode '@' for +o, and '+' for +v, or '&', '%', etc.
    \param before Before flag
*/
void event_dispatch_mode_on_user(const string& channel, const string& target_nick, const string& source_nick, const char prefix, const char mode, bool before)
{
    string mm=(string)""+irc_isupport.get_prefix2(mode);

    ostringstream m;
    m << prefix << mm.c_str() << "@user";

    char pr[2];
    pr[0]=prefix;
    pr[1]='\0';

    char mo[2]="?";   // '?' because "get_prefix2(mode)" can return an empty string
    strncpy(mo,mm.c_str(),1);

    char sym[2];
    sym[0]=mode;
    sym[1]='\0';

    mod_dispatch(m.str(),irc_find_user(source_nick).in_logic_as,channel,"",before,0,
        "channel",channel.c_str(),"source",irc_find_user(source_nick).in_logic_as.c_str(),"source_nick",source_nick.c_str(),"target",irc_find_user(target_nick).in_logic_as.c_str(),"target_nick",target_nick.c_str(),"prefix",pr,"mode_char",mo,"mode",mo,"target_ident",irc_get_ident(target_nick).c_str(),"target_host",irc_get_host(target_nick).c_str(),"symbol",sym,
        NULL,
        "event_type","user",
        "nick",source_nick.c_str(),"origin","@irc@","ident",irc_get_ident(source_nick).c_str(),"host",irc_get_host(source_nick).c_str(),NULL);
}

/*!
    \brief Triggers dispatching of the event on channel mode
    \author VooDooMan
    \version 2
    \date 2004, 2008-03-09
    \param channel Name of channel
    \param source_nick Which nick set the mode?
    \param prefix '-' or '+'
    \param mode '@' for +o, and '+' for +v, or '&', '%', etc.
    \param arg Argument/mask (if applicable; if not, an empty string)
    \param before Before flag
*/
void event_dispatch_mode_on_channel(const string& channel, const string& source_nick, const char prefix, const char mode, const string& arg, bool before)
{
    string mm=(string)""+irc_isupport.get_prefix2(mode);

    ostringstream m;
    m << prefix << mm.c_str() << "@channel";

    char pr[2];
    pr[0]=prefix;
    pr[1]='\0';

    char mo[2];
    mo[0]=mode;
    mo[1]='\0';

    mod_dispatch(m.str(),irc_find_user(source_nick).in_logic_as,channel,"",before,0,
        "channel",channel.c_str(),"source",irc_find_user(source_nick).in_logic_as.c_str(),"source_nick",source_nick.c_str(),"prefix",pr,"mode",mo,"mask",arg.c_str(),
        NULL,
        "event_type","channel",
        "nick",source_nick.c_str(),"origin","@irc@","ident",irc_get_ident(source_nick).c_str(),"host",irc_get_host(source_nick).c_str(),NULL);
}

