/***************************************************************************
                          logic_compile.cpp  -  description
                             -------------------
    begin                : Thu May 10 2007
    copyright            : (C) 2004-2007 by Marian VooDooMan Meravy
    email                : ghostvoodooman -a.T= users (dot) sourceforge.net
 ***************************************************************************/

/* $Id: logic_compile.cpp,v 1.22 2008/08/23 15:10:44 ghostvoodooman Exp $ */

/***************************************************************************

VooDoo cIRCle - an IRC (ro)bot
Copyright (C) 2004-2008 by Marian VooDooMan Meravy
  (ghostvoodooman -a.T= users (dot) sourceforge.net)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************/

/*!
    \file
    \brief Implements compilation of 'logic.txt'
    \version $Id: logic_compile.cpp,v 1.22 2008/08/23 15:10:44 ghostvoodooman Exp $
*/

#if defined(HAVE_CONFIG_H)
#   include <config.h>
#endif

#include <string>
#include <vector>
#include <stack>
#include <list>
#include <map>
#include <algorithm>

#include "../params.h"
#include "../utils/memleak.h"
#include "../utils/assert.h"
#include "../compat.h"

#include "../sock_shared.h"

#include "../mod/c_module.h"
#include "c_user.h"
#include "c_channel.h"
#include "../log.h"
#include "../proc/c_lines.h"
#include "c_smtp.h"
#include "c_channel_def.h"
#include "c_private.h"
#include "../sock.h"
#include "../threads.h"
#include "../modules.h"
#include "../match.h"
#include "../proc/c_proc_exc.h"
#include "logic_internal.h"

#include "../utils/precomp.h"

#if defined(_MSC_VER) || defined(__BORLANDC__)
#   pragma hdrstop
#endif

/*
    \brief Gets a string from processed line
    \author VooDooMan
    \version 1
    \date 2004
    \param pos Pass and after return get position in scanned line
    \param ln Scanned line
    \return Returns substring
*/
string logic_process_line_get_string(unsigned int &pos, string ln)
{
    string res;
    bool in_string=false;
    for(; pos<ln.length(); ++pos) {
        if(ln[pos]=='"' && (pos==0 || (pos>0 && ln[pos-1]!='\\'))) {
            res+=ln[pos];
            in_string=!in_string;
            continue;
        }
        if(ln[pos]==0x20 && !in_string) {
            pos++;
            return res;
        }
        res+=ln[pos];
    }
    return res;
}

enum e_return
{
    ELC_RETURN_TRUE,
    ELC_RETURN_FALSE,
    ELC_CONTINUE,
    ELC_BREAK,
};

MY_STATIC e_return logic_compile_user_a(c_module* this_module, string& ln, string& token, unsigned int& i1, string& error_reason, c_user& user, c_channel& channel, bool& b_channel, vector<string>& all_groups, deque<c_user>& users, deque<c_user>& terminators, vector<c_channel>& channels, bool& b_user2, int& i_brackets, const string& debug);

MY_STATIC e_return logic_compile_proc_a(c_module* this_module, string& ln, string& token, unsigned int& i1, string& error_reason, c_procedure& procedure, vector<string>& all_groups, deque<c_lines>& lines, int& i_brackets, int& error_line, const string& debug);
MY_STATIC e_return logic_compile_proc_b(c_module* this_module, string& ln, string& token, unsigned int& i1, string& error_reason, c_procedure& procedure, vector<string>& all_groups, deque<c_lines>& lines, int& i_brackets, int& error_line, const string& debug, string& result);
MY_STATIC e_return logic_compile_proc_c(c_module* this_module, string& ln, string& token, unsigned int& i1, string& error_reason, c_procedure& procedure, vector<string>& all_groups, deque<c_lines>& lines, int& i_brackets, int& error_line, const string& debug, string& result, bool& b_chan, bool& b_proc, vector<c_procedure>& procedures);

MY_STATIC e_return logic_compile_chan_a(c_module* this_module, string& ln, string& token, unsigned int& i1, string& error_reason, c_channel& channel, c_channel_def& chan, vector<string>& all_groups, const string& debug);

MY_STATIC e_return logic_compile_prv_a(c_module* this_module, string& ln, string& token, unsigned int& i1, string& error_reason, c_private& prv, vector<string>& all_groups, const string& debug);

/*!
    \brief Parses and compiles one line from logic.txt
    \author VooDooMan
    \version 1
    \date 2004
    \param ln Line from file
    \param error_reason If result == true: Returns addres of string with error description
    \param error_line Line counter. This parameter MUST have the same value as returned on previous call! (for the first time it MUST be initialized to one)
    \param b_user2 This parameter MUST have the same value as returned on previous call! (for the first time it MUST be initialized to false)
    \param user This parameter MUST have the same value as returned on previous call!
    \param b_channel This parameter MUST have the same value as returned on previous call! (for the first time it MUST be initialized to false)
    \param channel This parameter MUST have the same value as returned on previous call!
    \param b_user This parameter MUST have the same value as returned on previous call! (for the first time it MUST be initialized to false)
    \param i_brackets This parameter MUST have the same value as returned on previous call! (for the first time it MUST be initialized to zero)
    \param b_proc This parameter MUST have the same value as returned on previous call! (for the first time it MUST be initialized to false)
    \param procedure This parameter MUST have the same value as returned on previous call!
    \param lines This parameter MUST have the same value as returned on previous call!
    \param b_smtp This parameter MUST have the same value as returned on previous call! (for the first time it MUST be initialized to false)
    \param smtp This parameter MUST have the same value as returned on previous call!
    \param b_smtp_data This parameter MUST have the same value as returned on previous call! (for the first time it MUST be initialized to false)
    \param procedures This parameter MUST have the same value as returned on previous call!
    \param users This parameter MUST have the same value as returned on previous call!
    \param channels This parameter MUST have the same value as returned on previous call!
    \param chan This parameter MUST have the same value as returned on previous call!
    \param channel_defs This parameter MUST have the same value as returned on previous call!
    \param b_chan This parameter MUST have the same value as returned on previous call!
    \param all_groups This parameter MUST have the same value as returned on previous call!
    \param b_groups This parameter MUST have the same value as returned on previous call!
    \param b_prv This parameter MUST have the same value as returned on previous call!
    \param prv This parameter MUST have the same value as returned on previous call!
    \param terminators This parameter MUST have the same value as returned on previous call!
    \param functions Array of functions (to check for same-named procedure)
    \result Returns true if error; error_line contains line number
    \warning Parameter "ln" MUST be without leading SPACEs and/or TABs!
    \param this_module Module
    \param debug Debug string
*/
// returns true if error
bool logic_process_line(c_module* this_module, string ln, string& error_reason, int& error_line, bool& b_user2, c_user& user, bool& b_channel, c_channel& channel, bool& b_user, int& i_brackets, bool& b_proc, c_procedure& procedure, deque<c_lines>& lines, bool& b_smtp, c_smtp& smtp, bool& b_smtp_data, vector<c_procedure>& procedures, deque<c_user>& users, vector<c_channel>& channels, c_channel_def& chan, vector<c_channel_def>& channel_defs, bool& b_chan, vector<string>& all_groups, bool& b_groups, bool& b_prv, c_private& prv, deque<c_user>& terminators, map<string,c_function*>& functions, const string& debug)
{
    if(!ln.compare((string)""))
        return false;

    e_return r;

    string mod_name;
    if(this_module)
        mod_name=this_module->get_mod_name();

    string token;
    string result;
    for(unsigned int i1=0; i1<ln.length()+1; ++i1) {
        unsigned int i2;
        if(i1<ln.length() && ln[i1]=='=') {
            result=token;
            token="";
            continue;
        }
        if(i1<ln.length() && ln[i1]!=0x20)
            token+=ln[i1];
            else {
                if((!token.compare((string)"user") || !token.compare((string)"user_template")) && !b_chan) {
                    b_user=true;
                    b_channel=false;

                    i1++;
                    string name;
                    for(i2=i1; i2<ln.length(); ++i2, ++i1) {
                        if(ln[i2]==0x20 || ln[i2]=='\t') {
                            error_reason="Name of user/group cannot contain space or TAB";
                            return true;
                        } else
                            name+=ln[i2];
                    }

                    {
                        // compatibility
                        bool changed=(name.length() && name[0]!='"') || name.empty();
                        if(name.length() && name[0]!='"')
                            name=(string)"\""+name+"\"";
                        if(name.empty())
                            name="\"\"";

                        string log="OOPS: Compatibility warning: forcing username "+name+" to be enclosed into double-quotes. This problem will be resolved on the next backup.";
                        if(changed)
                            log_debug(log);
                    }

                    name=logic_eval_str(this_module,name,debug);

                    user.name=name;
                    user.host_unknown=false;
                    user.host_bot=false;
                    user.terminator=false;
                    user.is_template=!token.compare((string)"user_template");
                    user.hostmask.clear();
                    user.fullname.clear();
                    user.access_to_group.clear();
                    user.access_to_proc.clear();
                    user.access_usage_procedure.clear();
                    user.access_grant_procedure.clear();
                    user.access_grant_group.clear();
                    user.access_to_channel.clear();
                    user.access_grant_channel.clear();
                    user.access_grant_partyline=false;
                    user.access_to_backup=false;
                    user.access_grant_backup=false;
                    user.access_to_rehash=false;
                    user.access_grant_rehash=false;
                    user.access_to_plususer=false;
                    user.access_grant_plususer=false;
                    user.access_to_plusproc=false;
                    user.access_grant_plusproc=false;
                    user.access_to_replication=false;
                    user.access_grant_replication=false;
                    user.access_to_restart=false;
                    user.access_grant_restart=false;
                    user.access_to_die=false;
                    user.access_grant_die=false;
                    user.access_to_filesystem=false;
                    user.access_to_can_send_all_users=false;
                    user.access_grant_filesystem=false;
                    user.access_to_partyline=false;
                    user.access_to_private=false;
                    user.access_grant_private=false;
                    user.access_to_can_send_all_users=false;
                    user.access_grant_can_send_all_users=false;
                    user.access_to_can_send_unknown_users=false;
                    user.access_grant_can_send_unknown_users=false;
                    user.access_to_upgrade=false;
                    user.access_grant_upgrade=false;
                    user.access_to_apply=false;
                    user.access_grant_apply=false;
                    user.dcc_hosts.clear();
                    user.groups.clear();
                    user.replication.clear();
                    user.last_changed=0;
                    user.replication_partner=false;
                    user.partyline_msg_flood.lines=0;
                    user.partyline_msg_flood.seconds=0;
                    user.meta.clear();
                    b_user=false;
                    b_user2=true;

                    return false;
                }
                if(!token.compare((string)"{") && b_user!=true && b_user2!=true && b_channel!=true && b_proc!=true && !b_chan && b_groups!=true && !b_prv) {
                    error_reason="Unexpected '{'";
                    return true;
                }

                if(!token.compare((string)"{")) {
                    i_brackets++;
                    return false;
                }

                if(b_user2 && !b_chan) {
                    r=logic_compile_user_a(this_module,ln,token,i1,error_reason,user,channel,b_channel,all_groups,users,terminators,channels,b_user2,i_brackets,debug);
                    switch(r) {
                        case ELC_RETURN_TRUE:
                            return true;
                        case ELC_RETURN_FALSE:
                            return false;
                        default:
                            break;
                    }
                    if(ELC_BREAK==r)
                        break;

                    if(!token.compare((string)"}")) {
                        if(i_brackets--<0) {
                            error_reason="Unexpected '}'";
                            return true;
                        }

                        if(b_user2 && !b_channel) {
                            if(!user.is_template) {
                                if((!user.name.compare((string)"") || user.hostmask.empty() || user.fullname.empty()) && !user.host_unknown) {
                                    error_reason="Username, or host mask, or full name was not specified";
                                    return true;
                                }
                            } else {
                                if(!user.name.compare((string)"")) {
                                    error_reason="Username was not specified";
                                    return true;
                                }
                            }
                            if(!user.terminator)
                                users.push_back(user);
                            else
                                terminators.push_back(user);
                        }

                        if(b_channel) {
                            b_channel=false;
                            channels.push_back(channel);
                        } else {
                            if(b_user2)
                                b_user2=false;
                        }

                        if(b_proc) {
                            c_proc_command c(this_module,_label,error_line,debug);

                            procedure.commands.push_back(c);

                            if(lines.size()) {
                                deque<c_proc_command>::iterator i2;
                                for(i2=procedure.commands.begin(); i2!=procedure.commands.end(); ++i2) {
                                    if(lines.back().line==(*i2).get_line()) {
                                        (*i2).set_goto_line(error_line);
                                    }
                                }
                            }

                            lines.pop_back();
                        }

                        return false;
                    }

                    if(!token.compare((string)"channel")) {
                        i1++;
                        string chan=logic_eval_str(this_module,logic_process_line_get_string(i1,ln),debug);
                        if(chan.length()>128) {
                            error_reason="Name of channel cannot be longer than 128 characters";
                            return true;
                        }

                        channel.username=user.name;
                        channel.is_template=user.is_template;
                        channel.terminator=user.terminator;
                        channel.channel_name=chan;
                        channel.host_unknown=user.host_unknown;
                        channel.groups.clear();
                        channel.on_deop="";
                        channel.on_ban="";
                        channel.on_unban="";
                        channel.on_kick="";
                        channel.on_op="";
                        channel.on_voice="";
                        channel.on_devoice="";
                        channel.on_join="";
                        channel.on_banned="";
                        channel.on_flood="";
                        channel.on_privmsg="";
                        channel.on_notice="";
                        channel.on_other_mode="";
                        channel.on_other_mode_mask="";

                        channel.on_except="";
                        channel.on_unexcept="";
                        channel.on_invite="";
                        channel.on_uninvite="";
                        channel.on_not_invited="";
                        channel.on_not_in_reop="";
                        channel.on_reop="";

                        channel.msg_flood.lines=0;
                        channel.notice_flood.lines=0;
                        channel.repeat_flood.lines=0;
                        channel.nick_flood.lines=0;
                        channel.join_flood.lines=0;
                        channel.mode_flood.lines=0;
                        channel.ctcp_flood.lines=0;

                        channel.allow_dynamic.clear();

                        channel.dynamic_plus_modes="";
                        channel.dynamic_minus_modes="";

                        channel.can_send_unknown_users=false;

                        b_channel=true;
                        break;
                    }
                }

                if(!token.compare((string)"procedure") && !b_chan) {
                    i1++;
                    string proc=logic_process_line_get_string(i1,ln);

                    {
                        string n1=proc;
                        if(n1.find("(",0)!=string::npos)
                            n1.erase(n1.find("(",0),n1.length()-n1.find("(",0));
                        {
                            vector<c_procedure>::iterator i1;
                            for(i1=procedures.begin(); i1!=procedures.end(); ++i1) {
                                string n2=(*i1).name;
                                if(n2.find("(",0)!=string::npos)
                                    n2.erase(n2.find("(",0),n2.length()-n2.find("(",0));
                                if(!n1.compare(n2)) {
                                    error_reason="Procedure already defined";
                                    return true;
                                }
                            }
                        }
                        {
                            map<string,c_function*>::iterator i1;
                            for(i1=functions.begin(); i1!=functions.end(); ++i1) {
                                string n2=(*i1).second->get_name();
                                if(n2.find("(",0)!=string::npos)
                                    n2.erase(n2.find("(",0),n2.length()-n2.find("(",0));
                                if(!n1.compare(n2)) {
                                    error_reason="Function with same name already defined";
                                    return true;
                                }
                            }
                        }
                    }

                    procedure.name=proc;
                    procedure.commands.clear();
                    procedure.groups.clear();
                    procedure.replication.clear();
                    procedure.last_changed=0;
                    procedure.minimal_botnet_version=1;
                    procedure.rproc.clear();

                    b_proc=true;
                    break;
                }

                if(b_proc && !b_smtp && !b_chan) {
                    r=logic_compile_proc_a(this_module,ln,token,i1,error_reason,procedure,all_groups,lines,i_brackets,error_line,debug);
                    switch(r) {
                        case ELC_RETURN_TRUE:
                            return true;
                        case ELC_RETURN_FALSE:
                            return false;
                        default:
                            break;
                    }
                    if(ELC_BREAK==r)
                        break;
                    r=logic_compile_proc_b(this_module,ln,token,i1,error_reason,procedure,all_groups,lines,i_brackets,error_line,debug,result);
                    switch(r) {
                        case ELC_RETURN_TRUE:
                            return true;
                        case ELC_RETURN_FALSE:
                            return false;
                        default:
                            break;
                    }
                    if(ELC_BREAK==r)
                        break;

                    if(!token.compare((string)"SMTP")) {
                        b_smtp=true;
                        b_smtp_data=false;

                        smtp.server="127.0.0.1";
                        smtp.port=25;
                        smtp.helo="";
                        smtp.mail_from="root";
                        smtp.rcpt_to="root";
                        smtp.data="";
                        smtp.data_lines.clear();

                        break;
                    }

                    r=logic_compile_proc_c(this_module,ln,token,i1,error_reason,procedure,all_groups,lines,i_brackets,error_line,debug,result,b_chan,b_proc,procedures);
                    switch(r) {
                        case ELC_RETURN_TRUE:
                            return true;
                        case ELC_RETURN_FALSE:
                            return false;
                        default:
                            break;
                    }
                    if(ELC_BREAK==r)
                        break;
                }

                if(b_smtp && !b_chan) {
                    if(b_smtp_data) {
                        if(!ln.compare((string)"-")) {
                            smtp.data+="\r\n";
                            smtp.data_lines.push_back("-");
                        } else {
                            smtp.data+=ln+"\r\n";
                            smtp.data_lines.push_back(ln);
                        }
                        if(!ln.compare((string)".")) {
                            b_smtp_data=false;
                        }

                        break;
                    }

                    if(!token.compare((string)"server")) {
                        i1++;
                        string data=logic_process_line_get_string(i1,ln);

                        smtp.server=data;

                        break;
                    }

                    if(!token.compare((string)"port")) {
                        i1++;
                        string data=logic_process_line_get_string(i1,ln);

                        smtp.port=stous(data);
                        if(smtp.port==0)
                            smtp.port=25;

                        break;
                    }

                    if(!token.compare((string)"HELO")) {
                        i1++;
                        string data=logic_process_line_get_string(i1,ln);

                        smtp.helo=data;

                        break;
                    }

                    if(!token.compare((string)"MAIL_FROM")) {
                        i1++;
                        string data=logic_process_line_get_string(i1,ln);

                        smtp.mail_from=data;

                        break;
                    }

                    if(!token.compare((string)"RCPT_TO")) {
                        i1++;
                        string data=logic_process_line_get_string(i1,ln);

                        smtp.rcpt_to=data;

                        break;
                    }

                    if(!token.compare((string)"DATA")) {
                        smtp.data="";

                        b_smtp_data=true;

                        break;
                    }
                }

                if(!token.compare((string)"channel") && !b_chan) {
                    i1++;
                    string chan_=logic_eval_str(this_module,logic_process_line_get_string(i1,ln),debug);

                    chan.channel_name=chan_;
                    chan.on_mode="";
                    chan.on_key="";
                    chan.on_limit="";
                    chan.on_topic="";
                    chan.on_ircop="";
                    chan.on_ctcp="";
                    chan.on_privmsg="";
                    chan.on_notice="";
                    chan.on_part="";
                    chan.on_dynamic_ban="";
                    chan.on_server_msg="";
                    chan.allow_dynamic.clear();
                    chan.last_changed=0;
                    chan.replication.clear();
                    chan.keys.clear();
                    chan.dynamic_minus_modes="";
                    chan.dynamic_plus_modes="";
                    chan.groups.clear();
                    chan.keys.clear();
                    chan.dynamic_bans.clear();
                    chan.dynbans_editors.clear();

                    b_chan=true;
                    break;
                }

                if(b_chan) {
                    r=logic_compile_chan_a(this_module,ln,token,i1,error_reason,channel,chan,all_groups,debug);
                    switch(r) {
                        case ELC_RETURN_TRUE:
                            return true;
                        case ELC_RETURN_FALSE:
                            return false;
                        default:
                            break;
                    }
                    if(ELC_BREAK==r)
                        break;
                }

                if(!token.compare((string)"groups") && !b_groups) {
                    b_groups=true;

                    return false;
                }

                if(b_groups && token.compare("}")) {
                    i1++;
                    string grp=logic_process_line_get_string(i1,ln);

                    all_groups.push_back(grp);

                    return false;
                }

                if(!token.compare((string)"private") && !b_prv) {
                    prv.on_privmsg="";
                    prv.on_notice="";
                    prv.on_ctcp="";
                    prv.on_filesys_got_new="";
                    prv.on_fnc="";
                    prv.on_broadcast="";
                    prv.on_server_msg="";
                    prv.last_changed=0;
                    prv.replication.clear();

                    b_prv=true;
                    return false;
                }

                if(b_prv) {
                    r=logic_compile_prv_a(this_module,ln,token,i1,error_reason,prv,all_groups,debug);
                    switch(r) {
                        case ELC_RETURN_TRUE:
                            return true;
                        case ELC_RETURN_FALSE:
                            return false;
                        default:
                            break;
                    }
                    if(ELC_BREAK==r)
                        break;
                }

                if(!token.compare((string)"}")) {
                    if(i_brackets--<0) {
                        error_reason="Unexpected '}'";
                        return true;
                    }

                    if(b_smtp) {
                        b_smtp=false;

                        c_proc_command c(this_module,_smtp,error_line,debug);

                        c.set_next_arg(smtp.server);
                        c.set_next_arg(smtp.port);
                        c.set_next_arg(smtp.helo);
                        c.set_next_arg(smtp.mail_from);
                        c.set_next_arg(smtp.rcpt_to);
                        c.set_next_arg(smtp.data);

                        procedure.commands.push_back(c);
                    }

                    if(b_chan) {
                        b_chan=false;
                        channel_defs.push_back(chan);
                    }

                    if(b_groups) {
                        b_groups=false;
                    }

                    if(b_prv) {
                        b_prv=false;
                    }

                    c_proc_command c(this_module,_label,error_line,debug);

                    procedure.commands.push_back(c);

                    if(lines.size()) {
                        deque<c_proc_command>::iterator i2;
                        for(i2=procedure.commands.begin(); i2!=procedure.commands.end(); ++i2) {
                            if(lines.back().line==(*i2).get_line()) {
                                (*i2).set_goto_line(error_line);
                                lines.pop_back();
                            }
                        }
                    }

                    break;
                }

                error_reason="Unknown token";
                return true; // unknown token
            }
    }

    if(b_user)
        return true;

    return false;
}

/*!
    \brief Parses and compiles function (for partyline)
    \author VooDooMan
    \version 2
    \date 2004, 2005-10-05
    \param error_reason If result == true: Returns string with error description
    \param error_line Line counter. This parameter MUST have the same value as returned on previous call! (for the first time it MUST be initialized to one)
    \param commands Array of lines to parse
    \param copy_to_r_procedures Set tu true if successfully parsed procedure need to be updated/added in/to r_procedures
    \param declaration Procedure name with full declaration (without "procedure" token!)
    \param groups List of groups which procedure is member of
    \return Returns "true" for error
    \param replications Replication rules
    \param remote_call Remote call rules
*/
bool logic_parse_func(string& error_reason, int& error_line, list<string>& commands, bool copy_to_r_procedures, string declaration, vector<string>& groups, vector<c_replication>& replications, vector<string>& remote_call)
{
    c_module* core_mod=get_module("core",true);
    if(!core_mod) {
        error_reason="Internal error: core module is not loaded";
        return true;
    }

    error_line=0;
    error_reason="";

    c_function* func=MY_NEW c_function();

    bool error=false;
    string error_msg;
    bool need_more=false;
    string debug;

    string dump;

    string ln="#version 2";

    func->parse_command(core_mod,ln,error,error_msg,need_more,debug,"");

    ln="function ";
    ln+=declaration;
    dump+=ln+"\n";

    func->parse_command(core_mod,ln,error,error_msg,need_more,debug,"");

    ln="{";
    dump+=ln+"\n";

    func->parse_command(core_mod,ln,error,error_msg,need_more,debug,"");

    vector<string>::iterator gi;
    for(gi=groups.begin(); gi!=groups.end(); ++gi) {
        ln="#member_of_group ";
        ln+=*gi;
        dump+=ln+"\n";

        func->parse_command(core_mod,ln,error,error_msg,need_more,debug,"");
    }

    vector<c_replication>::iterator ri;
    for(ri=replications.begin(); ri!=replications.end(); ++ri) {
        ln="#replication ";
        ln+=(*ri).partner;
        ln+=" ";
        if(REPL_PUSH==(*ri).type)
            ln+="push";
        if(REPL_PULL==(*ri).type)
            ln+="pull";
        if(REPL_PUSHPULL==(*ri).type)
            ln+="pushpull";
        dump+=ln+"\n";

        func->parse_command(core_mod,ln,error,error_msg,need_more,debug,"");
    }

    vector<string>::iterator rpi;
    for(rpi=remote_call.begin(); rpi!=remote_call.end(); ++rpi) {
        ln="#remote_bot_call ";
        ln+=*rpi;
        dump+=ln+"\n";

        func->parse_command(core_mod,ln,error,error_msg,need_more,debug,"");
    }

    {
        bool got_last_changed=false;
        for(list<string>::iterator ci=commands.begin(); ci!=commands.end(); ++ci) {
            if((*ci).find("#last_changed ")==0) {
                got_last_changed=true;
                break;
            }
        }

        if(!got_last_changed) {
            ln="#last_changed ";
            ln+=tos(static_cast<unsigned long>(my_time(NULL)));
            dump+=ln+"\n";
        }
    }

    func->parse_command(core_mod,ln,error,error_msg,need_more,debug,"");

    error_line=1;
    list<string>::iterator i1;
    for(i1=commands.begin(); i1!=commands.end(); ++i1, ++error_line) {
        debug="Line #";
        debug+=tos(error_line);
        dump+=*i1+"\n";
        string ln=*i1;
        func->parse_command(core_mod,ln,error,error_msg,need_more,debug,"");
        if(error) {
            error_reason=error_msg;
            return true;
        }
    }

    ln="}";
    dump+=ln+"\n";
    func->parse_command(core_mod,ln,error,error_msg,need_more,debug,"");

    if(!func->is_okay()) {
        error_reason="Mistake in brackets { }";
        MY_DELETE(func);
        return true;
    }

    func->set_dump(dump);

    if(!error && copy_to_r_procedures) {
        core_mod->r_functions.remove_function(func->get_name());
        if(!core_mod->r_functions.add_function(func,error_reason)) {
            MY_DELETE(func);
            return true;
        }

        error_line=0;

        vector<string>::iterator ig;
        for(ig=groups.begin(); ig!=groups.end(); ++ig) {
            r_all_groups.push_back(*ig);
        }

        return false;
    }

    error_line=0;

    return false;
}

/*!
    \brief Parses and compiles procedure (for partyline)
    \author VooDooMan
    \version 1
    \date 2004
    \param error_reason If result == true: Returns addres of string with error description
    \param error_line Line counter. This parameter MUST have the same value as returned on previous call! (for the first time it MUST be initialized to one)
    \param commands Array of lines to parse
    \param copy_to_r_procedures Set tu true if successfully parsed procedure need to be updated/added in/to r_procedures
    \param declaration Procedure name with full declaration (without "procedure" token!)
    \param groups List of groups which procedure is member of
    \param replications Replication rules
    \param remote_call Remote call rules
*/
void logic_parse_proc(string& error_reason, int& error_line, list<string>& commands, bool copy_to_r_procedures, string declaration, vector<string>& groups, vector<c_replication>& replications, vector<string>& remote_call)
{
    c_module* core_mod=get_module("core",true);
    if(!core_mod) {
        error_reason="Internal error: core module is not loaded";
        error_line=1;
        return;
    }

    error_line=0;
    error_reason="";

    bool b_user2=false; // user in bracket
    c_user user;
    bool b_channel=false; // user in bracket, channel in sub-bracket
    c_channel channel;
    bool b_user=false; // user token
    int i_brackets=0;
    bool b_proc=false; // in procedure
    c_procedure procedure;
    deque<c_lines> lines;
    bool b_smtp=false;
    c_smtp smtp;
    bool b_smtp_data=false;
    bool b_chan=false;
    bool b_groups=false;
    bool b_prv=false;

    c_channel_def chan;

    vector<c_procedure> procedures;
    deque<c_user> users;
    vector<c_channel> channels;
    vector<c_channel_def> channel_defs;
    deque<c_user> terminators;

    vector<string> all_groups;

    c_private prv;

    bool error=false;

    map<string,c_function*> fns;

    error=logic_process_line(core_mod,(string)"procedure "+declaration,error_reason,error_line,b_user2,user,b_channel,channel,b_user,i_brackets,b_proc,procedure,lines,b_smtp,smtp,b_smtp_data,procedures,users,channels,chan,channel_defs,b_chan,all_groups,b_groups,b_prv,prv,terminators,fns,"");
    error=logic_process_line(core_mod,"{",error_reason,error_line,b_user2,user,b_channel,channel,b_user,i_brackets,b_proc,procedure,lines,b_smtp,smtp,b_smtp_data,procedures,users,channels,chan,channel_defs,b_chan,all_groups,b_groups,b_prv,prv,terminators,fns,"");

    int i3=1;

    list<string>::const_iterator i1;
    for(i1=commands.begin(); i1!=commands.end(); ++i1, ++i3) {
        string s=*i1;
        size_t i2;
        for(i2=s.length()-1; (signed)i2>=0; --i2)
            if(s[i2]==0x20 || s[i2]=='\t')
                s=s.erase(i2,1);
                else
                break;
        while(s.length()>0 && (s[0]==0x20 || s[0]=='\t'))
            s=s.erase(0,1);
        error=logic_process_line(core_mod,s,error_reason,error_line,b_user2,user,b_channel,channel,b_user,i_brackets,b_proc,procedure,lines,b_smtp,smtp,b_smtp_data,procedures,users,channels,chan,channel_defs,b_chan,all_groups,b_groups,b_prv,prv,terminators,fns,s);
        if(error) {
            error_line=i3;
            return;
        }
    }
    error_line=0;

    if(i_brackets!=1 || procedure.name.compare(declaration)) {
        error_line=i3;
        error_reason="Mistake in brackets { }";
    }

    if(0==error_line) {
        // dump test
        string mod_name="dummy";
        int indent=0;
        bool ok=true;
        deque<c_proc_command>::iterator i1;
        for(i1=procedure.commands.begin(); i1!=procedure.commands.end(); ++i1, ++error_line) {
            try {
                (*i1).dump(core_mod,indent);
            } catch(c_proc_exc& e) {
                ok=false;
                error_reason="Dump recheck failed. Debug string: ";
                error_reason+=e.get_msg();
                break;
            }
        }
        if(ok) {
            error_line=0;
            error_reason="";
        }
    }

    if(error_line==0 && copy_to_r_procedures) {
        // always core module
        c_proc_command c(core_mod,_end_bracket,i3,"");
        procedure.commands.push_back(c);
        c=c_proc_command(core_mod,_label,i3,"");
        procedure.commands.push_back(c);

        bool got=false;
        vector<c_procedure>::iterator i1;
        for(i1=core_mod->r_procedures.begin(); i1!=core_mod->r_procedures.end(); ++i1) {
            if(!(*i1).name.compare(declaration)) {
                got=true;
                (*i1).groups.clear();
                vector<string>::iterator i2;
                for(i2=groups.begin(); i2!=groups.end(); ++i2) {
                    (*i1).groups.push_back(*i2);
                }

                (*i1).replication.clear();
                vector<c_replication>::iterator i3;
                for(i3=replications.begin(); i3!=replications.end(); ++i3)
                    (*i1).replication.push_back(*i3);

                (*i1).rproc.clear();
                vector<string>::iterator i4;
                for(i4=remote_call.begin(); i4!=remote_call.end(); ++i4) {
                    c_rproc r;
                    r.remote_bot=*i4;
                    (*i1).rproc.push_back(r);
                }

                (*i1).commands=procedure.commands;
                (*i1).last_changed=procedure.last_changed;
                break;
            }
        }
        if(!got) {
            procedure.groups.clear();
            vector<string>::iterator i2;
            for(i2=groups.begin(); i2!=groups.end(); ++i2) {
                procedure.groups.push_back(*i2);
            }

            vector<c_replication>::iterator i3;
            for(i3=replications.begin(); i3!=replications.end(); ++i3)
                procedure.replication.push_back(*i3);

            vector<string>::iterator i4;
            for(i4=remote_call.begin(); i4!=remote_call.end(); ++i4) {
                c_rproc r;
                r.remote_bot=*i4;
                procedure.rproc.push_back(r);
            }
            core_mod->r_procedures.push_back(procedure);
        }
        vector<string>::iterator ig;
        for(ig=groups.begin(); ig!=groups.end(); ++ig) {
            r_all_groups.push_back(*ig);
        }
    }
}

MY_STATIC e_return logic_compile_user_a(c_module* this_module, string& ln, string& token, unsigned int& i1, string& error_reason, c_user& user, c_channel& channel, bool& b_channel, vector<string>& all_groups, deque<c_user>& UNUSEDARG(users), deque<c_user>& UNUSEDARG(terminators), vector<c_channel>& UNUSEDARG(channels), bool& UNUSEDARG(b_user2), int& UNUSEDARG(i_brackets), const string& debug)
{
    if(!token.compare((string)"replication")) {
        i1++;
        string partner=logic_process_line_get_string(i1,ln);

        string type=logic_process_line_get_string(i1,ln);

        c_replication r;
        r.partner=partner;
        r.type=REPL_INVALID;
        if(!type.compare("push"))
            r.type=REPL_PUSH;
        if(!type.compare("pull"))
            r.type=REPL_PULL;
        if(!type.compare("pushpull"))
            r.type=REPL_PUSHPULL;
        if(r.type==REPL_INVALID) {
            error_reason="Invalid replication type";
            return ELC_RETURN_TRUE;
        }

        user.replication.push_back(r);

        return ELC_RETURN_FALSE;
    }

    if(!token.compare((string)"access_to_backup")) {
        i1++;
        string a=logic_process_line_get_string(i1,ln);

        user.access_to_backup=stoi(a)!=0;

        return ELC_RETURN_FALSE;
    }

    if(!token.compare((string)"access_grant_backup")) {
        i1++;
        string a=logic_process_line_get_string(i1,ln);

        user.access_grant_backup=stoi(a)!=0;

        return ELC_RETURN_FALSE;
    }

    if(!token.compare((string)"access_to_rehash")) {
        i1++;
        string a=logic_process_line_get_string(i1,ln);

        user.access_to_rehash=stoi(a)!=0;

        return ELC_RETURN_FALSE;
    }

    if(!token.compare((string)"access_grant_rehash")) {
        i1++;
        string a=logic_process_line_get_string(i1,ln);

        user.access_grant_rehash=stoi(a)!=0;

        return ELC_RETURN_FALSE;
    }

    if(!token.compare((string)"access_to_+user")) {
        i1++;
        string a=logic_process_line_get_string(i1,ln);

        user.access_to_plususer=stoi(a)!=0;

        return ELC_RETURN_FALSE;
    }

    if(!token.compare((string)"access_grant_+user")) {
        i1++;
        string a=logic_process_line_get_string(i1,ln);

        user.access_grant_plususer=stoi(a)!=0;

        return ELC_RETURN_FALSE;
    }

    if(!token.compare((string)"access_to_+proc")) {
        i1++;
        string a=logic_process_line_get_string(i1,ln);

        user.access_to_plusproc=stoi(a)!=0;

        return ELC_RETURN_FALSE;
    }

    if(!token.compare((string)"access_grant_+proc")) {
        i1++;
        string a=logic_process_line_get_string(i1,ln);

        user.access_grant_plusproc=stoi(a)!=0;

        return ELC_RETURN_FALSE;
    }

    if(!token.compare((string)"access_to_replication")) {
        i1++;
        string a=logic_process_line_get_string(i1,ln);

        user.access_to_replication=stoi(a)!=0;

        return ELC_RETURN_FALSE;
    }

    if(!token.compare((string)"access_grant_replication")) {
        i1++;
        string a=logic_process_line_get_string(i1,ln);

        user.access_grant_replication=stoi(a)!=0;

        return ELC_RETURN_FALSE;
    }

    if(!token.compare((string)"access_to_restart")) {
        i1++;
        string a=logic_process_line_get_string(i1,ln);

        user.access_to_restart=stoi(a)!=0;

        return ELC_RETURN_FALSE;
    }

    if(!token.compare((string)"access_grant_restart")) {
        i1++;
        string a=logic_process_line_get_string(i1,ln);

        user.access_grant_restart=stoi(a)!=0;

        return ELC_RETURN_FALSE;
    }

    if(!token.compare((string)"access_to_die")) {
        i1++;
        string a=logic_process_line_get_string(i1,ln);

        user.access_to_die=stoi(a)!=0;

        return ELC_RETURN_FALSE;
    }

    if(!token.compare((string)"access_grant_die")) {
        i1++;
        string a=logic_process_line_get_string(i1,ln);

        user.access_grant_die=stoi(a)!=0;

        return ELC_RETURN_FALSE;
    }

    if(!token.compare((string)"access_to_upgrade")) {
        i1++;
        string a=logic_process_line_get_string(i1,ln);

        user.access_to_upgrade=stoi(a)!=0;

        return ELC_RETURN_FALSE;
    }

    if(!token.compare((string)"access_grant_upgrade")) {
        i1++;
        string a=logic_process_line_get_string(i1,ln);

        user.access_grant_upgrade=stoi(a)!=0;

        return ELC_RETURN_FALSE;
    }

    if(!token.compare((string)"access_to_apply")) {
        i1++;
        string a=logic_process_line_get_string(i1,ln);

        user.access_to_apply=stoi(a)!=0;

        return ELC_RETURN_FALSE;
    }

    if(!token.compare((string)"access_grant_apply")) {
        i1++;
        string a=logic_process_line_get_string(i1,ln);

        user.access_grant_apply=stoi(a)!=0;

        return ELC_RETURN_FALSE;
    }

    if(!token.compare((string)"access_to_filesystem")) {
        i1++;
        string a=logic_process_line_get_string(i1,ln);

        user.access_to_filesystem=stoi(a)!=0;

        return ELC_RETURN_FALSE;
    }

    if(!token.compare((string)"access_to_can_send_all_users")) {
        i1++;
        string a=logic_process_line_get_string(i1,ln);

        user.access_to_can_send_all_users=stoi(a)!=0;

        return ELC_RETURN_FALSE;
    }

    if(!token.compare((string)"access_grant_can_send_all_users")) {
        i1++;
        string a=logic_process_line_get_string(i1,ln);

        user.access_grant_can_send_all_users=stoi(a)!=0;

        return ELC_RETURN_FALSE;
    }

    if(!token.compare((string)"access_to_can_send_unknown_users")) {
        i1++;
        string a=logic_process_line_get_string(i1,ln);

        user.access_to_can_send_unknown_users=stoi(a)!=0;

        return ELC_RETURN_FALSE;
    }

    if(!token.compare((string)"access_grant_can_send_unknown_users")) {
        i1++;
        string a=logic_process_line_get_string(i1,ln);

        user.access_grant_can_send_unknown_users=stoi(a)!=0;

        return ELC_RETURN_FALSE;
    }

    if(!token.compare((string)"access_grant_filesystem")) {
        i1++;
        string a=logic_process_line_get_string(i1,ln);

        user.access_grant_filesystem=stoi(a)!=0;

        return ELC_RETURN_FALSE;
    }

    if(!token.compare((string)"replication_partner")) {
        i1++;

        user.replication_partner=true;

        return ELC_RETURN_FALSE;
    }

    if(!token.compare((string)"last_changed")) {
        i1++;
        string time=logic_process_line_get_string(i1,ln);
        user.last_changed=sto_time(time);
        return ELC_RETURN_FALSE;
    }

    if(!token.compare((string)"meta")) {
        i1++;
        pair<string,string> p;
        p.first=logic_eval_str(this_module,logic_process_line_get_string(i1,ln),debug);
        p.second=logic_eval_str(this_module,logic_process_line_get_string(i1,ln),debug);
        user.meta.insert(p);
        return ELC_RETURN_FALSE;
    }

    if(!token.compare((string)"terminator")) {
        i1++;
        string value=logic_process_line_get_string(i1,ln);
        user.terminator=stoi(value)!=0;
        return ELC_RETURN_FALSE;
    }

    if(!token.compare((string)"host")) {
        if(user.is_template) {
            error_reason="User template cannot have a host mask";
            return ELC_RETURN_TRUE;
        }

        i1++;
        string hostmask;
        for(unsigned i2=i1; i2<ln.length(); ++i2, ++i1) {
            if(ln[i2]==0x20 || ln[i2]=='\t') {
                error_reason="Hostmask cannot contain space or TAB";
                return ELC_RETURN_TRUE;
            }
            hostmask+=ln[i2];
        }
        hostmask=logic_eval_str(this_module,hostmask,debug);
        user.hostmask.push_back(hostmask);

        {
            map<string,string> vars;
            string host=hostmask;
            if(host.find("@",0)!=string::npos)
                host.erase(0,host.find("@",0)+1);
            if(host.find("*",0)==string::npos && host.find("?",0)==string::npos && host.find("/",0)==string::npos && host.find("#",0)==string::npos) {
                in_addr addr4;
                threads_resolve4(host.c_str(),addr4,RT_RESOLVE_SCHEDULE);
                in_addr6_ addr6;
                threads_resolve6(host.c_str(),addr6,RT_RESOLVE_SCHEDULE);
                if(core_conf_getvar("try_reverse_lookup").compare("0")) {
                    char tmp[2048];
                    if(my_inet_pton(AF_INET,host.c_str(),&addr4)==1)
                        threads_reverse4(host.c_str(),tmp,sizeof(tmp)-1,RT_RESOLVE_SCHEDULE);
                    if(my_inet_pton(AF_INET6,host.c_str(),&addr6)==1)
                        threads_reverse6(addr6,tmp,sizeof(tmp)-1,RT_RESOLVE_SCHEDULE);
                }
            }
        }

        return ELC_RETURN_FALSE;
    }

    if(!token.compare((string)"dcc_host")) {
        i1++;
        string dcc_group=logic_process_line_get_string(i1,ln);

        string host=logic_eval_str(this_module,logic_process_line_get_string(i1,ln),debug);

        c_dcc_host h;
        h.group=stoi(dcc_group);
        h.host=host;

        user.dcc_hosts.push_back(h);

        return ELC_RETURN_FALSE;
    }

    if(!token.compare((string)"host_unknown")) {
        user.host_unknown=true;
        return ELC_RETURN_FALSE;
    }

    if(!token.compare((string)"host_bot")) {
        user.host_bot=true;
        if(user.name.length()>255-1) {
            error_reason="Name of bot (host_bot) cannot be longer than 254 characters!";
            return ELC_RETURN_TRUE;
        }
        botnet_botname=user.name;
        return ELC_RETURN_FALSE;
    }

    if(!token.compare((string)"full_name")) {
        if(user.is_template) {
            error_reason="User template cannot have a full name";
            return ELC_RETURN_TRUE;
        }
        i1++;
        string fullname=logic_eval_str(this_module,logic_process_line_get_string(i1,ln),debug);
        user.fullname.push_back(fullname);
        return ELC_RETURN_FALSE;
    }

    if(b_channel && !token.compare((string)"allow_dynamic")) {
        i1++;
        string group=logic_eval_str(this_module,logic_process_line_get_string(i1,ln),debug);

        string plus=logic_eval_str(this_module,logic_process_line_get_string(i1,ln),debug);

        string minus=logic_eval_str(this_module,logic_process_line_get_string(i1,ln),debug);

        c_dynamic_rule r;
        r.group=group;
        r.plus_modes=plus;
        r.minus_modes=minus;

        channel.allow_dynamic.push_back(r);

        return ELC_RETURN_FALSE;
    }

    if(b_channel && !token.compare((string)"dynamic1")) {
        i1++;
        string modes=logic_process_line_get_string(i1,ln);

        {
            // compatibility
            bool changed=(modes.length() && modes[0]!='"') || modes.empty();
            if(modes.length() && modes[0]!='"')
                modes=(string)"\""+modes+"\"";
            if(modes.empty())
                modes="\"\"";

            string log="OOPS: Compatibility warning: forcing dynamic1 modes "+modes+" to be enclosed into double-quotes. This problem will be resolved on the next backup.";
            if(changed)
                log_debug(log);
        }

        channel.dynamic_plus_modes=logic_eval_str(this_module,modes,debug);

        return ELC_RETURN_FALSE;
    }

    if(b_channel && !token.compare((string)"dynamic2")) {
        i1++;
        string modes=logic_process_line_get_string(i1,ln);

        {
            // compatibility
            bool changed=(modes.length() && modes[0]!='"') || modes.empty();
            if(modes.length() && modes[0]!='"')
                modes=(string)"\""+modes+"\"";
            if(modes.empty())
                modes="\"\"";

            string log="OOPS: Compatibility warning: forcing dynamic2 modes "+modes+" to be enclosed into double-quotes. This problem will be resolved on the next backup.";
            if(changed)
                log_debug(log);
        }

        channel.dynamic_minus_modes=logic_eval_str(this_module,modes,debug);

        return ELC_RETURN_FALSE;
    }

    if(b_channel && !token.compare((string)"member_of_group")) {
        i1++;
        string grp=logic_eval_str(this_module,logic_process_line_get_string(i1,ln),debug);

        channel.groups.push_back(grp);

        all_groups.push_back(grp);

        return ELC_BREAK;
    }

    if(!b_channel && !token.compare((string)"member_of_group")) {
        i1++;
        string grp=logic_eval_str(this_module,logic_process_line_get_string(i1,ln),debug);

        user.groups.push_back(grp);

        all_groups.push_back(grp);

        return ELC_BREAK;
    }

    if(!token.compare((string)"on_deop")) {
        if(channel.on_deop.compare("")) {
            error_reason="Duplicate event";
            return ELC_RETURN_TRUE;
        }

        i1++;
        string proc=logic_process_line_get_string(i1,ln);

        channel.on_deop=proc;

        return ELC_BREAK;
    }

    if(!token.compare((string)"on_ban")) {
        if(channel.on_ban.compare("")) {
            error_reason="Duplicate event";
            return ELC_RETURN_TRUE;
        }

        i1++;
        string proc=logic_process_line_get_string(i1,ln);

        channel.on_ban=proc;

        return ELC_BREAK;
    }

    if(!token.compare((string)"on_unban")) {
        if(channel.on_unban.compare("")) {
            error_reason="Duplicate event";
            return ELC_RETURN_TRUE;
        }

        i1++;
        string proc=logic_process_line_get_string(i1,ln);

        channel.on_unban=proc;

        return ELC_BREAK;
    }

    if(!token.compare((string)"on_kick")) {
        if(channel.on_kick.compare("")) {
            error_reason="Duplicate event";
            return ELC_RETURN_TRUE;
        }

        i1++;
        string proc=logic_process_line_get_string(i1,ln);

        channel.on_kick=proc;

        return ELC_BREAK;
    }

    if(!token.compare((string)"on_op")) {
        if(channel.on_op.compare("")) {
            error_reason="Duplicate event";
            return ELC_RETURN_TRUE;
        }

        i1++;
        string proc=logic_process_line_get_string(i1,ln);

        channel.on_op=proc;

        return ELC_BREAK;
    }

    if(!token.compare((string)"on_voice")) {
        if(channel.on_voice.compare("")) {
            error_reason="Duplicate event";
            return ELC_RETURN_TRUE;
        }

        i1++;
        string proc=logic_process_line_get_string(i1,ln);

        channel.on_voice=proc;

        return ELC_BREAK;
    }

    if(!token.compare((string)"on_devoice")) {
        if(channel.on_devoice.compare("")) {
            error_reason="Duplicate event";
            return ELC_RETURN_TRUE;
        }

        i1++;
        string proc=logic_process_line_get_string(i1,ln);

        channel.on_devoice=proc;

        return ELC_BREAK;
    }

    if(!token.compare((string)"on_creator")) {
        i1++;
        logic_process_line_get_string(i1,ln);

        string log="OOPS: Compatibility warning: This event is not used anymore. This problem will be resolved on the next backup. -> ";
        log+=ln;
        log_debug(log);

        return ELC_BREAK;
    }

    if(!token.compare((string)"on_decreator")) {
        i1++;
        logic_process_line_get_string(i1,ln);

        string log="OOPS: Compatibility warning: This event is not used anymore. This problem will be resolved on the next backup. -> ";
        log+=ln;
        log_debug(log);

        return ELC_BREAK;
    }

    if(!token.compare((string)"on_join")) {
        if(channel.on_join.compare("")) {
            error_reason="Duplicate event";
            return ELC_RETURN_TRUE;
        }

        i1++;
        string proc=logic_process_line_get_string(i1,ln);

        channel.on_join=proc;

        return ELC_BREAK;
    }

    if(!token.compare((string)"on_banned")) {
        if(channel.on_banned.compare("")) {
            error_reason="Duplicate event";
            return ELC_RETURN_TRUE;
        }

        i1++;
        string proc=logic_process_line_get_string(i1,ln);

        channel.on_banned=proc;

        return ELC_BREAK;
    }

    if(!token.compare((string)"on_flood")) {
        if(channel.on_flood.compare("")) {
            error_reason="Duplicate event";
            return ELC_RETURN_TRUE;
        }

        i1++;
        string proc=logic_process_line_get_string(i1,ln);

        channel.on_flood=proc;

        return ELC_BREAK;
    }

    if(!token.compare((string)"on_privmsg")) {
        if(channel.on_privmsg.compare("")) {
            error_reason="Duplicate event";
            return ELC_RETURN_TRUE;
        }

        i1++;
        string proc=logic_process_line_get_string(i1,ln);

        channel.on_privmsg=proc;

        return ELC_BREAK;
    }

    if(!token.compare((string)"on_notice")) {
        if(channel.on_notice.compare("")) {
            error_reason="Duplicate event";
            return ELC_RETURN_TRUE;
        }

        i1++;
        string proc=logic_process_line_get_string(i1,ln);

        channel.on_notice=proc;

        return ELC_BREAK;
    }

    if(!token.compare((string)"on_except")) {
        if(channel.on_except.compare("")) {
            error_reason="Duplicate event";
            return ELC_RETURN_TRUE;
        }

        i1++;
        string proc=logic_process_line_get_string(i1,ln);

        channel.on_except=proc;

        return ELC_BREAK;
    }

    if(!token.compare((string)"on_unexcept")) {
        if(channel.on_unexcept.compare("")) {
            error_reason="Duplicate event";
            return ELC_RETURN_TRUE;
        }

        i1++;
        string proc=logic_process_line_get_string(i1,ln);

        channel.on_unexcept=proc;

        return ELC_BREAK;
    }

    if(!token.compare((string)"on_invite")) {
        if(channel.on_invite.compare("")) {
            error_reason="Duplicate event";
            return ELC_RETURN_TRUE;
        }

        i1++;
        string proc=logic_process_line_get_string(i1,ln);

        channel.on_invite=proc;

        return ELC_BREAK;
    }

    if(!token.compare((string)"on_uninvite")) {
        if(channel.on_uninvite.compare("")) {
            error_reason="Duplicate event";
            return ELC_RETURN_TRUE;
        }

        i1++;
        string proc=logic_process_line_get_string(i1,ln);

        channel.on_uninvite=proc;

        return ELC_BREAK;
    }

    if(!token.compare((string)"on_not_invited")) {
        if(channel.on_not_invited.compare("")) {
            error_reason="Duplicate event";
            return ELC_RETURN_TRUE;
        }

        i1++;
        string proc=logic_process_line_get_string(i1,ln);

        channel.on_not_invited=proc;

        return ELC_BREAK;
    }

    if(!token.compare((string)"on_not_in_reop")) {
        if(channel.on_not_in_reop.compare("")) {
            error_reason="Duplicate event";
            return ELC_RETURN_TRUE;
        }

        i1++;
        string proc=logic_process_line_get_string(i1,ln);

        channel.on_not_in_reop=proc;

        return ELC_BREAK;
    }

    if(!token.compare((string)"on_reop")) {
        if(channel.on_reop.compare("")) {
            error_reason="Duplicate event";
            return ELC_RETURN_TRUE;
        }

        i1++;
        string proc=logic_process_line_get_string(i1,ln);

        channel.on_reop=proc;

        return ELC_BREAK;
    }

    if(!token.compare((string)"on_other_mode")) {
        if(channel.on_other_mode.compare("")) {
            error_reason="Duplicate event";
            return ELC_RETURN_TRUE;
        }

        i1++;
        string proc=logic_process_line_get_string(i1,ln);

        channel.on_other_mode=proc;

        return ELC_BREAK;
    }

    if(!token.compare((string)"on_other_mode_mask")) {
        if(channel.on_other_mode_mask.compare("")) {
            error_reason="Duplicate event";
            return ELC_RETURN_TRUE;
        }

        i1++;
        string proc=logic_process_line_get_string(i1,ln);

        channel.on_other_mode_mask=proc;

        return ELC_BREAK;
    }

    if(!token.compare((string)"msg_flood")) {
        i1++;
        string lines=logic_process_line_get_string(i1,ln);

        string seconds=logic_process_line_get_string(i1,ln);

        channel.msg_flood.lines=stoui(lines);
        channel.msg_flood.seconds=sto_time(seconds);

        return ELC_BREAK;
    }

    if(!token.compare((string)"notice_flood")) {
        i1++;
        string lines=logic_process_line_get_string(i1,ln);

        string seconds=logic_process_line_get_string(i1,ln);

        channel.notice_flood.lines=stoui(lines);
        channel.notice_flood.seconds=sto_time(seconds);

        return ELC_BREAK;
    }

    if(!token.compare((string)"repeat_flood")) {
        i1++;
        string lines=logic_process_line_get_string(i1,ln);

        string seconds=logic_process_line_get_string(i1,ln);

        channel.repeat_flood.lines=stoui(lines);
        channel.repeat_flood.seconds=sto_time(seconds);

        return ELC_BREAK;
    }

    if(!token.compare((string)"nick_flood")) {
        i1++;
        string lines=logic_process_line_get_string(i1,ln);

        string seconds=logic_process_line_get_string(i1,ln);

        channel.nick_flood.lines=stoui(lines);
        channel.nick_flood.seconds=sto_time(seconds);

        return ELC_BREAK;
    }

    if(!token.compare((string)"join_flood")) {
        i1++;
        string lines=logic_process_line_get_string(i1,ln);

        string seconds=logic_process_line_get_string(i1,ln);

        channel.join_flood.lines=stoui(lines);
        channel.join_flood.seconds=sto_time(seconds);

        return ELC_BREAK;
    }

    if(!token.compare((string)"mode_flood")) {
        i1++;
        string lines=logic_process_line_get_string(i1,ln);

        string seconds=logic_process_line_get_string(i1,ln);

        channel.mode_flood.lines=stoui(lines);
        channel.mode_flood.seconds=sto_time(seconds);

        return ELC_BREAK;
    }

    if(!token.compare((string)"ctcp_flood")) {
        i1++;
        string lines=logic_process_line_get_string(i1,ln);

        string seconds=logic_process_line_get_string(i1,ln);

        channel.ctcp_flood.lines=stoui(lines);
        channel.ctcp_flood.seconds=sto_time(seconds);

        return ELC_BREAK;
    }

    if(!token.compare((string)"can_send_unknown_users")) {
        i1++;
        string val=logic_process_line_get_string(i1,ln);

        channel.can_send_unknown_users=stoi(val)!=0;

        return ELC_BREAK;
    }

    if(!token.compare((string)"access_to_partyline") && !b_channel) {
        i1++;
        string s=logic_process_line_get_string(i1,ln);
        if(s.compare("0") && s.compare("1")) {
            error_reason="Token 'access_to_partyline' must be followed by '0' or '1'";
            return ELC_RETURN_TRUE;
        }
        user.access_to_partyline=!s.compare((string)"1");
        return ELC_BREAK;
    }

    if(!token.compare((string)"access_to_private") && !b_channel) {
        i1++;
        string s=logic_process_line_get_string(i1,ln);
        if(s.compare("0") && s.compare("1")) {
            error_reason="Token 'access_to_private' must be followed by '0' or '1'";
            return ELC_RETURN_TRUE;
        }
        user.access_to_private=!s.compare((string)"1");
        return ELC_BREAK;
    }

    if(!token.compare((string)"access_grant_private") && !b_channel) {
        i1++;
        string s=logic_process_line_get_string(i1,ln);
        if(s.compare("0") && s.compare("1")) {
            error_reason="Token 'access_grant_private' must be followed by '0' or '1'";
            return ELC_RETURN_TRUE;
        }
        user.access_grant_private=!s.compare((string)"1");
        return ELC_BREAK;
    }

    if(!token.compare((string)"partyline_msg_flood") && !b_channel) {
        i1++;
        string lines=logic_process_line_get_string(i1,ln);

        string seconds=logic_process_line_get_string(i1,ln);

        user.partyline_msg_flood.lines=stoui(lines);
        user.partyline_msg_flood.seconds=sto_time(seconds);

        return ELC_BREAK;
    }

    if(!token.compare((string)"access_to_group")) {
        i1++;
        string gr=logic_process_line_get_string(i1,ln);

        {
            string name=gr;
            // compatibility
            bool changed=(name.length() && name[0]!='"') || name.empty();
            if(name.length() && name[0]!='"')
                name=(string)"\""+name+"\"";
            if(name.empty())
                name="\"\"";
            gr=name;

            string log="OOPS: Compatibility warning: forcing string "+name+" to be enclosed into double-quotes. This problem will be resolved on the next backup.";
            if(changed)
                log_debug(log);
        }

        gr=logic_eval_str(this_module,gr,debug);

        user.access_to_group.push_back(gr);
        return ELC_RETURN_FALSE;
    }

    if(!token.compare((string)"access_grant_procedure")) {
        i1++;
        string proc=logic_process_line_get_string(i1,ln);

        {
            string name=proc;
            // compatibility
            bool changed=(name.length() && name[0]!='"') || name.empty();
            if(name.length() && name[0]!='"')
                name=(string)"\""+name+"\"";
            if(name.empty())
                name="\"\"";
            proc=name;

            string log="OOPS: Compatibility warning: forcing string "+name+" to be enclosed into double-quotes. This problem will be resolved on the next backup.";
            if(changed)
                log_debug(log);
        }

        proc=logic_eval_str(this_module,proc,debug);

        user.access_grant_procedure.push_back(proc);
        return ELC_RETURN_FALSE;
    }

    if(!token.compare((string)"access_usage_procedure")) {
        i1++;
        string proc=logic_process_line_get_string(i1,ln);

        {
            string name=proc;
            // compatibility
            bool changed=(name.length() && name[0]!='"') || name.empty();
            if(name.length() && name[0]!='"')
                name=(string)"\""+name+"\"";
            if(name.empty())
                name="\"\"";
            proc=name;

            string log="OOPS: Compatibility warning: forcing string "+name+" to be enclosed into double-quotes. This problem will be resolved on the next backup.";
            if(changed)
                log_debug(log);
        }

        proc=logic_eval_str(this_module,proc,debug);

        user.access_usage_procedure.push_back(proc);
        return ELC_RETURN_FALSE;
    }

    if(!token.compare((string)"access_grant_group")) {
        i1++;
        string gr=logic_process_line_get_string(i1,ln);

        {
            string name=gr;
            // compatibility
            bool changed=(name.length() && name[0]!='"') || name.empty();
            if(name.length() && name[0]!='"')
                name=(string)"\""+name+"\"";
            if(name.empty())
                name="\"\"";
            gr=name;

            string log="OOPS: Compatibility warning: forcing string "+name+" to be enclosed into double-quotes. This problem will be resolved on the next backup.";
            if(changed)
                log_debug(log);
        }

        gr=logic_eval_str(this_module,gr,debug);

        user.access_grant_group.push_back(gr);
        return ELC_RETURN_FALSE;
    }

    if(!token.compare((string)"access_to_procedure")) {
        i1++;
        string pr=logic_process_line_get_string(i1,ln);

        {
            string name=pr;
            // compatibility
            bool changed=(name.length() && name[0]!='"') || name.empty();
            if(name.length() && name[0]!='"')
                name=(string)"\""+name+"\"";
            if(name.empty())
                name="\"\"";
            pr=name;

            string log="OOPS: Compatibility warning: forcing string "+name+" to be enclosed into double-quotes. This problem will be resolved on the next backup.";
            if(changed)
                log_debug(log);
        }

        pr=logic_eval_str(this_module,pr,debug);

        user.access_to_proc.push_back(pr);
        return ELC_RETURN_FALSE;
    }

    if(!token.compare((string)"access_to_channel")) {
        i1++;
        string ch=logic_process_line_get_string(i1,ln);

        {
            string name=ch;
            // compatibility
            bool changed=(name.length() && name[0]!='"') || name.empty();
            if(name.length() && name[0]!='"')
                name=(string)"\""+name+"\"";
            if(name.empty())
                name="\"\"";
            ch=name;

            string log="OOPS: Compatibility warning: forcing string "+name+" to be enclosed into double-quotes. This problem will be resolved on the next backup.";
            if(changed)
                log_debug(log);
        }

        ch=logic_eval_str(this_module,ch,debug);

        user.access_to_channel.push_back(ch);
        return ELC_RETURN_FALSE;
    }

    if(!token.compare((string)"access_grant_channel")) {
        i1++;
        string ch=logic_process_line_get_string(i1,ln);

        {
            string name=ch;
            // compatibility
            bool changed=(name.length() && name[0]!='"') || name.empty();
            if(name.length() && name[0]!='"')
                name=(string)"\""+name+"\"";
            if(name.empty())
                name="\"\"";
            ch=name;

            string log="OOPS: Compatibility warning: forcing string "+name+" to be enclosed into double-quotes. This problem will be resolved on the next backup.";
            if(changed)
                log_debug(log);
        }

        ch=logic_eval_str(this_module,ch,debug);

        user.access_grant_channel.push_back(ch);
        return ELC_RETURN_FALSE;
    }

    if(!token.compare((string)"access_grant_partyline")) {
        i1++;
        string ch=logic_process_line_get_string(i1,ln);
        user.access_grant_partyline=!ch.compare((string)"1");
        return ELC_RETURN_FALSE;
    }

    return ELC_CONTINUE;
}

MY_STATIC e_return logic_compile_proc_a(c_module* this_module, string& ln, string& token, unsigned int& i1, string& error_reason, c_procedure& procedure, vector<string>& all_groups, deque<c_lines>& lines, int& UNUSEDARG(i_brackets), int& error_line, const string& debug)
{
    if(!token.compare((string)"replication")) {
        i1++;
        string partner=logic_process_line_get_string(i1,ln);

        string type=logic_process_line_get_string(i1,ln);

        c_replication r;
        r.partner=partner;
        r.type=REPL_INVALID;
        if(!type.compare("push"))
            r.type=REPL_PUSH;
        if(!type.compare("pull"))
            r.type=REPL_PULL;
        if(!type.compare("pushpull"))
            r.type=REPL_PUSHPULL;
        if(r.type==REPL_INVALID) {
            error_reason="Invalid replication type";
            return ELC_RETURN_TRUE;
        }

        procedure.replication.push_back(r);

        return ELC_RETURN_FALSE;
    }

    if(!token.compare((string)"remote_bot_call")) {
        i1++;
        string partner=logic_process_line_get_string(i1,ln);

        c_rproc r;
        r.remote_bot=partner;

        procedure.rproc.push_back(r);

        procedure.minimal_botnet_version=3;

        return ELC_RETURN_FALSE;
    }

    if(!token.compare((string)"last_changed")) {
        i1++;
        string time=logic_process_line_get_string(i1,ln);
        procedure.last_changed=sto_time(time);
        return ELC_RETURN_FALSE;
    }

    if(!token.compare((string)"link")) {
        i1++;
        c_proc_command c(this_module,_link,error_line,debug);

        string type=logic_process_line_get_string(i1,ln);
        bool set=false;
        if(!type.compare("telnet"))
            set=c.set_next_arg("telnet");
        if(!type.compare("telnet_ssl"))
            set=c.set_next_arg("telnet_ssl");
        if(!type.compare("ssl"))
            set=c.set_next_arg("ssl");

        if(!set) {
            error_reason="Invalid link type (\"telnet\", \"telnet_ssl\" and \"ssl\" are only supported yet)";
            return ELC_RETURN_TRUE;
        }

        string botname=logic_process_line_get_string(i1,ln);
        c.set_next_arg(botname);

        string bind=logic_process_line_get_string(i1,ln);
        c.set_next_arg(bind);

        string host=logic_process_line_get_string(i1,ln);
        c.set_next_arg(host);

        string port=logic_process_line_get_string(i1,ln);
        long port_=stol(port);
        if(0==port_)
            port_=4444;
        c.set_next_arg(port_);

        string unlink_proc=logic_process_line_get_string(i1,ln);
        c.set_next_arg(unlink_proc);

        procedure.commands.push_back(c);

        return ELC_BREAK;
    }

    if(!token.compare((string)"chan_mode")) {
        i1++;
        c_proc_command c(this_module,_chan_mode,error_line,debug);

        string priority=logic_process_line_get_string(i1,ln);
        int priority_=0;
        if((priority_=logic_get_priority(priority.c_str()))==0) {
            error_reason="Unknown priority class";
            return ELC_RETURN_TRUE;
        }
        c.set_next_arg(priority_);

        string chan=logic_process_line_get_string(i1,ln);
        c.set_next_arg(chan);

        string mode=logic_process_line_get_string(i1,ln);
        c.set_next_arg(mode);

        procedure.commands.push_back(c);

        return ELC_BREAK;
    }

    if(!token.compare((string)"member_of_group")) {
        i1++;
        string grp=logic_eval_str(this_module,logic_process_line_get_string(i1,ln),debug);

        procedure.groups.push_back(grp);

        all_groups.push_back(grp);

        return ELC_BREAK;
    }

    if(!token.compare((string)"op")) {
        i1++;
        string priority=logic_process_line_get_string(i1,ln);
        int priority_;
        if((priority_=logic_get_priority(priority.c_str()))==0) {
            error_reason="Unknown priority class";
            return ELC_RETURN_TRUE;
        }
        c_proc_command c(this_module,_op,error_line,debug);

        c.set_next_arg(priority_);

        string chan=logic_process_line_get_string(i1,ln);
        c.set_next_arg(chan);

        string nick=logic_process_line_get_string(i1,ln);
        c.set_next_arg(nick);

        procedure.commands.push_back(c);

        return ELC_BREAK;
    }

    if(!token.compare((string)"deop")) {
        i1++;
        c_proc_command c(this_module,_deop,error_line,debug);

        string priority=logic_process_line_get_string(i1,ln);
        int priority_;
        if((priority_=logic_get_priority(priority.c_str()))==0) {
            error_reason="Unknown priority class";
            return ELC_RETURN_TRUE;
        }
        c.set_next_arg(priority_);

        string chan=logic_process_line_get_string(i1,ln);
        c.set_next_arg(chan);

        string nick=logic_process_line_get_string(i1,ln);
        c.set_next_arg(nick);

        procedure.commands.push_back(c);

        return ELC_BREAK;
    }

    if(!token.compare((string)"ban_mask")) {
        i1++;
        c_proc_command c(this_module,_ban_mask,error_line,debug);

        string chan=logic_process_line_get_string(i1,ln);
        c.set_next_arg(chan);

        string mask=logic_process_line_get_string(i1,ln);
        c.set_next_arg(mask);

        procedure.commands.push_back(c);

        return ELC_BREAK;
    }

    if(!token.compare((string)"deop_ban_mask")) {
        i1++;
        c_proc_command c(this_module,_deop_ban_mask,error_line,debug);

        string priority=logic_process_line_get_string(i1,ln);
        int priority_;
        if((priority_=logic_get_priority(priority.c_str()))==0) {
            error_reason="Unknown priority class";
            return ELC_RETURN_TRUE;
        }
        c.set_next_arg(priority_);

        string chan=logic_process_line_get_string(i1,ln);
        c.set_next_arg(chan);

        string nick=logic_process_line_get_string(i1,ln);
        c.set_next_arg(nick);

        string mask=logic_process_line_get_string(i1,ln);
        c.set_next_arg(mask);

        procedure.commands.push_back(c);

        procedure.minimal_botnet_version=1002;

        return ELC_BREAK;
    }

    if(!token.compare((string)"unban_mask")) {
        i1++;
        c_proc_command c(this_module,_unban_mask,error_line,debug);

        string chan=logic_process_line_get_string(i1,ln);
        c.set_next_arg(chan);

        string mask=logic_process_line_get_string(i1,ln);
        c.set_next_arg(mask);

        procedure.commands.push_back(c);

        return ELC_BREAK;
    }

    if(!token.compare((string)"voice")) {
        i1++;
        c_proc_command c(this_module,_voice,error_line,debug);

        string priority=logic_process_line_get_string(i1,ln);
        int priority_;
        if((priority_=logic_get_priority(priority.c_str()))==0) {
            error_reason="Unknown priority class";
            return ELC_RETURN_TRUE;
        }
        c.set_next_arg(priority_);

        string chan=logic_process_line_get_string(i1,ln);
        c.set_next_arg(chan);

        string nick=logic_process_line_get_string(i1,ln);
        c.set_next_arg(nick);

        procedure.commands.push_back(c);

        return ELC_BREAK;
    }

    if(!token.compare((string)"devoice")) {
        i1++;
        c_proc_command c(this_module,_devoice,error_line,debug);

        string priority=logic_process_line_get_string(i1,ln);
        int priority_;
        if((priority_=logic_get_priority(priority.c_str()))==0) {
            error_reason="Unknown priority class";
            return ELC_RETURN_TRUE;
        }
        c.set_next_arg(priority_);

        string chan=logic_process_line_get_string(i1,ln);
        c.set_next_arg(chan);

        string nick=logic_process_line_get_string(i1,ln);
        c.set_next_arg(nick);

        procedure.commands.push_back(c);

        return ELC_BREAK;
    }

    if(!token.compare((string)"kick")) {
        i1++;
        c_proc_command c(this_module,_kick,error_line,debug);

        string chan=logic_process_line_get_string(i1,ln);
        c.set_next_arg(chan);

        string nick=logic_process_line_get_string(i1,ln);
        c.set_next_arg(nick);

        string msg=logic_process_line_get_string(i1,ln);
        c.set_next_arg(msg);

        procedure.commands.push_back(c);

        return ELC_BREAK;
    }

    if(!token.compare((string)"msg")) {
        i1++;
        c_proc_command c(this_module,_msg,error_line,debug);

        string priority=logic_process_line_get_string(i1,ln);
        int priority_;
        if((priority_=logic_get_priority(priority.c_str()))==0) {
            error_reason="Unknown priority class";
            return ELC_RETURN_TRUE;
        }
        c.set_next_arg(priority_);

        string target=logic_process_line_get_string(i1,ln);
        c.set_next_arg(target);

        string msg=logic_process_line_get_string(i1,ln);
        c.set_next_arg(msg);

        procedure.commands.push_back(c);

        return ELC_BREAK;
    }

    if(!token.compare((string)"msgq")) {
        i1++;
        c_proc_command c(this_module,_msgq,error_line,debug);

        string priority=logic_process_line_get_string(i1,ln);
        int priority_;
        if((priority_=logic_get_priority(priority.c_str()))==0) {
            error_reason="Unknown priority class";
            return ELC_RETURN_TRUE;
        }
        c.set_next_arg(priority_);

        string target=logic_process_line_get_string(i1,ln);
        c.set_next_arg(target);

        string msg=logic_process_line_get_string(i1,ln);
        c.set_next_arg(msg);

        string msgq=logic_process_line_get_string(i1,ln);
        c.set_next_arg(msgq);

        procedure.commands.push_back(c);

        return ELC_BREAK;
    }

    if(!token.compare((string)"notice")) {
        i1++;
        c_proc_command c(this_module,_notice,error_line,debug);

        string priority=logic_process_line_get_string(i1,ln);
        int priority_;
        if((priority_=logic_get_priority(priority.c_str()))==0) {
            error_reason="Unknown priority class";
            return ELC_RETURN_TRUE;
        }
        c.set_next_arg(priority_);

        string target=logic_process_line_get_string(i1,ln);
        c.set_next_arg(target);

        string msg=logic_process_line_get_string(i1,ln);
        c.set_next_arg(msg);

        procedure.commands.push_back(c);

        return ELC_BREAK;
    }

    if(!token.compare((string)"noticeq")) {
        i1++;
        c_proc_command c(this_module,_noticeq,error_line,debug);

        string priority=logic_process_line_get_string(i1,ln);
        int priority_;
        if((priority_=logic_get_priority(priority.c_str()))==0) {
            error_reason="Unknown priority class";
            return ELC_RETURN_TRUE;
        }
        c.set_next_arg(priority_);

        string target=logic_process_line_get_string(i1,ln);
        c.set_next_arg(target);

        string msg=logic_process_line_get_string(i1,ln);
        c.set_next_arg(msg);

        string msgq=logic_process_line_get_string(i1,ln);
        c.set_next_arg(msgq);

        procedure.commands.push_back(c);

        return ELC_BREAK;
    }

    if(!token.compare((string)"if_in")) {
        i1++;
        c_proc_command c(this_module,_if_in,error_line,debug);

        string left=logic_process_line_get_string(i1,ln);
        c.set_next_arg(left);

        string right=logic_process_line_get_string(i1,ln);
        c.set_next_arg(right);

        procedure.commands.push_back(c);

        c_lines l;
        l.proc_name=procedure.name;
        l.line=error_line;
        lines.push_back(l);

        return ELC_BREAK;
    }

    if(!token.compare((string)"!if_in")) {
        i1++;
        c_proc_command c(this_module,_if_n_in,error_line,debug);

        string left=logic_process_line_get_string(i1,ln);
        c.set_next_arg(left);

        string right=logic_process_line_get_string(i1,ln);
        c.set_next_arg(right);

        procedure.commands.push_back(c);

        c_lines l;
        l.proc_name=procedure.name;
        l.line=error_line;
        lines.push_back(l);

        return ELC_BREAK;
    }

    if(!token.compare((string)"if_match")) {
        i1++;
        c_proc_command c(this_module,_if_match,error_line,debug);

        string left=logic_process_line_get_string(i1,ln);
        c.set_next_arg(left);

        string right=logic_process_line_get_string(i1,ln);
        c.set_next_arg(right);

        procedure.commands.push_back(c);

        c_lines l;
        l.proc_name=procedure.name;
        l.line=error_line;
        lines.push_back(l);

        return ELC_BREAK;
    }

    if(!token.compare((string)"!if_match")) {
        i1++;
        c_proc_command c(this_module,_if_n_match,error_line,debug);

        string left=logic_process_line_get_string(i1,ln);
        c.set_next_arg(left);

        string right=logic_process_line_get_string(i1,ln);
        c.set_next_arg(right);

        procedure.commands.push_back(c);

        c_lines l;
        l.proc_name=procedure.name;
        l.line=error_line;
        lines.push_back(l);

        return ELC_BREAK;
    }

    if(!token.compare((string)"if_match_case_insensitive")) {
        i1++;
        c_proc_command c(this_module,_if_match_case_insensitive,error_line,debug);

        string left=logic_process_line_get_string(i1,ln);
        c.set_next_arg(left);

        string right=logic_process_line_get_string(i1,ln);
        c.set_next_arg(right);

        procedure.commands.push_back(c);

        c_lines l;
        l.proc_name=procedure.name;
        l.line=error_line;
        lines.push_back(l);

        return ELC_BREAK;
    }

    if(!token.compare((string)"!if_match_case_insensitive")) {
        i1++;
        c_proc_command c(this_module,_if_n_match_case_insensitive,error_line,debug);

        string left=logic_process_line_get_string(i1,ln);
        c.set_next_arg(left);

        string right=logic_process_line_get_string(i1,ln);
        c.set_next_arg(right);

        procedure.commands.push_back(c);

        c_lines l;
        l.proc_name=procedure.name;
        l.line=error_line;
        lines.push_back(l);

        return ELC_BREAK;
    }

    return ELC_CONTINUE;
}

MY_STATIC e_return logic_compile_proc_b(c_module* this_module, string& ln, string& token, unsigned int& i1, string& error_reason, c_procedure& procedure, vector<string>& UNUSEDARG(all_groups), deque<c_lines>& lines, int& UNUSEDARG(i_brackets), int& error_line, const string& debug, string& result)
{
    if(!token.compare((string)"if_group")) {
        i1++;
        c_proc_command c(this_module,_if_group,error_line,debug);

        string left=logic_process_line_get_string(i1,ln);
        c.set_next_arg(left);

        string right=logic_process_line_get_string(i1,ln);
        c.set_next_arg(right);

        procedure.commands.push_back(c);

        c_lines l;
        l.proc_name=procedure.name;
        l.line=error_line;
        lines.push_back(l);

        return ELC_BREAK;
    }

    if(!token.compare((string)"!if_group")) {
        i1++;
        c_proc_command c(this_module,_if_n_group,error_line,debug);

        string left=logic_process_line_get_string(i1,ln);
        c.set_next_arg(left);

        string right=logic_process_line_get_string(i1,ln);
        c.set_next_arg(right);

        procedure.commands.push_back(c);

        c_lines l;
        l.proc_name=procedure.name;
        l.line=error_line;
        lines.push_back(l);

        return ELC_BREAK;
    }

    if(!token.compare((string)"wait_if_error")) {
        c_proc_command c(this_module,_wait_if_error,error_line,debug);

        c_lines l;
        l.proc_name=procedure.name;
        l.line=error_line;
        lines.push_back(l);

        procedure.commands.push_back(c);

        return ELC_BREAK;
    }

    if(!token.compare((string)"if_error")) {
        c_proc_command c(this_module,_if_error,error_line,debug);

        c_lines l;
        l.proc_name=procedure.name;
        l.line=error_line;
        lines.push_back(l);

        procedure.commands.push_back(c);

        return ELC_BREAK;
    }

    if(!token.compare((string)"return")) {
        c_proc_command c(this_module,_return,error_line,debug);;

        procedure.commands.push_back(c);

        return ELC_BREAK;
    }

    if(!token.compare((string)"work")) {
        c_proc_command c(this_module,_work,error_line,debug);

        procedure.commands.push_back(c);

        return ELC_BREAK;
    }

    if(!token.compare((string)"timer_once") || !token.compare((string)"timer_every")) {
        i1++;
        c_proc_command c(this_module,(!token.compare((string)"timer_once"))?_timer_once:_timer_every,error_line,debug);

        string name=logic_process_line_get_string(i1,ln);
        c.set_next_arg(name);

        string time=logic_process_line_get_string(i1,ln);
        my_time_t t=0;
        int mul=1;
        int mul2=1;
        for(size_t i4=time.length()-1; (signed)i4>=0; --i4) {
            if(time[i4]==':') {
                mul2=1;
                if(mul==1) {
                    mul=60;
                    continue;
                }
                if(mul==60) {
                    mul*=60;
                    continue;
                }
                if(mul==60*60) {
                    mul*=24;
                    continue;
                }
                error_reason="Invalid time (use syntax dd:hh:mm:ss)";
                return ELC_RETURN_TRUE;
            }
            if(!(time[i4]>='0' && time[i4]<='9')) {
                error_reason="Invalid time (use syntax dd:hh:mm:ss)";
                return ELC_RETURN_TRUE;
            }
            t+=mul*mul2*(time[i4]-'0');
            mul2*=10;
        }

        c.set_next_arg((long)t);

        string proc=logic_process_line_get_string(i1,ln);
        c.set_next_arg(proc);

        procedure.commands.push_back(c);

        return ELC_BREAK;
    }

    if(!token.compare((string)"execute")) {
        i1++;

        c_proc_command c(this_module,_execute,error_line,debug);

        string exec=logic_process_line_get_string(i1,ln);
        c.set_next_arg(exec);

        procedure.commands.push_back(c);

        return ELC_BREAK;
    }

    if(!token.compare((string)"remote_execute")) {
        i1++;
        c_proc_command c(this_module,_remote_execute,error_line,debug);

        string botname=logic_process_line_get_string(i1,ln);
        c.set_next_arg(botname);

        string call_string=logic_process_line_get_string(i1,ln);
        c.set_next_arg(call_string);

        c.set_next_arg(result);

        procedure.commands.push_back(c);

        procedure.minimal_botnet_version=3;

        return ELC_BREAK;
    }

    if(!token.compare((string)"kill_timers")) {
        i1++;
        c_proc_command c(this_module,_kill_timers,error_line,debug);

        string mask=logic_process_line_get_string(i1,ln);
        c.set_next_arg(mask);

        procedure.commands.push_back(c);

        return ELC_BREAK;
    }

    if(!token.compare((string)"ident")) {
        i1++;
        c_proc_command c(this_module,_ident,error_line,debug);

        string nick=logic_process_line_get_string(i1,ln);
        c.set_next_arg(nick);

        c.set_next_arg(result);

        procedure.commands.push_back(c);

        return ELC_BREAK;
    }

    if(!token.compare((string)"host")) {
        i1++;
        c_proc_command c(this_module,_host,error_line,debug);

        string nick=logic_process_line_get_string(i1,ln);
        c.set_next_arg(nick);

        c.set_next_arg(result);

        procedure.commands.push_back(c);

        return ELC_BREAK;
    }

    if(!token.compare((string)"get_chan_mode")) {
        i1++;
        c_proc_command c(this_module,_get_chan_mode,error_line,debug);

        string chan=logic_process_line_get_string(i1,ln);
        c.set_next_arg(chan);

        c.set_next_arg(result);

        procedure.commands.push_back(c);

        return ELC_BREAK;
    }

    if(!token.compare((string)"get_chan_topic")) {
        i1++;
        c_proc_command c(this_module,_get_chan_topic,error_line,debug);

        string chan=logic_process_line_get_string(i1,ln);
        c.set_next_arg(chan);

        c.set_next_arg(result);

        procedure.commands.push_back(c);

        return ELC_BREAK;
    }

    if(!token.compare((string)"topic")) {
        i1++;
        c_proc_command c(this_module,_topic,error_line,debug);

        string chan=logic_process_line_get_string(i1,ln);
        c.set_next_arg(chan);

        string topic;
        for(unsigned i2=i1; i2<ln.length() && ln[i2]!=' '; ++i2, ++i1)
            topic+=ln[i2];
        c.set_next_arg(topic);

        procedure.commands.push_back(c);

        return ELC_BREAK;
    }

    if(!token.compare((string)"SCRIPT")) {
        i1++;
        c_proc_command c(this_module,_script,error_line,debug);

        string type=logic_process_line_get_string(i1,ln);
        c.set_next_arg(type);

        string num=logic_process_line_get_string(i1,ln);
        c.set_next_arg(stol(num));

        string chan=logic_process_line_get_string(i1,ln);
        c.set_next_arg(chan);

        string params=logic_process_line_get_string(i1,ln);
        c.set_next_arg(params);

        procedure.commands.push_back(c);

        return ELC_BREAK;
    }

    if(!token.compare((string)"NET_SEND")) {
        i1++;
        c_proc_command c(this_module,_net_send,error_line,debug);

        string target=logic_process_line_get_string(i1,ln);
        c.set_next_arg(target);

        string msg=logic_process_line_get_string(i1,ln);
        c.set_next_arg(msg);

        procedure.commands.push_back(c);

        return ELC_BREAK;
    }

    if(!token.compare((string)"LOG")) {
        i1++;
        c_proc_command c(this_module,_log,error_line,debug);

        string msg=logic_process_line_get_string(i1,ln);
        c.set_next_arg(msg);

        procedure.commands.push_back(c);

        return ELC_BREAK;
    }

    if(!token.compare((string)"join")) {
        i1++;
        string ch=logic_process_line_get_string(i1,ln);

        string key=logic_process_line_get_string(i1,ln);

        {
            // compatibility
            bool changed=(key.length() && key[0]!='"') || key.empty();
            if(key.length() && key[0]!='"')
                key=(string)"\""+key+"\"";
            if(key.empty())
                key="\"\"";

            string log="OOPS: Compatibility warning: forcing channel key "+key+" to be enclosed into double-quotes. This problem will be resolved on the next backup.";
            if(changed)
                log_debug(log);
        }

        c_proc_command c(this_module,_join,error_line,debug);

        c.set_next_arg(ch);

        c.set_next_arg(key);

        procedure.commands.push_back(c);

        return ELC_BREAK;
    }

    if(!token.compare((string)"part")) {
        i1++;
        c_proc_command c(this_module,_part,error_line,debug);

        string ch=logic_process_line_get_string(i1,ln);
        c.set_next_arg(ch);

        string msg=logic_process_line_get_string(i1,ln);
        c.set_next_arg(msg);

        procedure.commands.push_back(c);

        return ELC_BREAK;
    }

    if(!token.compare((string)"disconnect")) {
        c_proc_command c(this_module,_disconnect,error_line,debug);

        procedure.commands.push_back(c);

        return ELC_BREAK;
    }

    if(!token.compare((string)"irc_server")) {
        i1++;
        c_proc_command c(this_module,_irc_server,error_line,debug);

        string host=logic_process_line_get_string(i1,ln);
        c.set_next_arg(host);

        string port=logic_process_line_get_string(i1,ln);
        long port_=stol(port);
        if(0==port_)
            port_=6667;
        c.set_next_arg(port_);

        procedure.commands.push_back(c);

        return ELC_BREAK;
    }

    if(!token.compare((string)"irc_server_ex")) {
        i1++;
        c_proc_command c(this_module,_irc_server_ex,error_line,debug);

        string flags=logic_process_line_get_string(i1,ln);
        c.set_next_arg(flags);

        string bindip=logic_process_line_get_string(i1,ln);
        c.set_next_arg(bindip);

        string host=logic_process_line_get_string(i1,ln);
        c.set_next_arg(host);

        string port=logic_process_line_get_string(i1,ln);
        long port_=stol(port);
        if(port_==0)
            port_=6667;
        c.set_next_arg(port_);

        procedure.commands.push_back(c);

        procedure.minimal_botnet_version=13;

        return ELC_BREAK;
    }

    if(!token.compare((string)"try_connect")) {
        i1++;

        c_proc_command c(this_module,_try_connect,error_line,debug);

        procedure.commands.push_back(c);

        return ELC_BREAK;
    }

    if(!token.compare((string)"dcc_server")) {
        i1++;
        c_proc_command c(this_module,_dcc_server,error_line,debug);

        string group=logic_process_line_get_string(i1,ln);
        long group_=stol(group);
        c.set_next_arg(group_);

        string ip=logic_process_line_get_string(i1,ln);
        c.set_next_arg(ip);

        string port=logic_process_line_get_string(i1,ln);
        long port_=stol(port);
        if(0==port_)
            port_=2010;
        c.set_next_arg(port_);

        string server_type=logic_process_line_get_string(i1,ln);
        c.set_next_arg(server_type);

        procedure.commands.push_back(c);

        return ELC_BREAK;
    }

    if(!token.compare((string)"dcc_server_ipv6")) {
        i1++;
        c_proc_command c(this_module,_dcc_server_ipv6,error_line,debug);

        string group=logic_process_line_get_string(i1,ln);
        long group_=stol(group);
        c.set_next_arg(group_);

        string ip=logic_process_line_get_string(i1,ln);
        c.set_next_arg(ip);

        string port=logic_process_line_get_string(i1,ln);
        long port_=stol(port);
        if(0==port_)
            port_=2010;
        c.set_next_arg(port_);

        string server_type=logic_process_line_get_string(i1,ln);
        c.set_next_arg(server_type);

        procedure.commands.push_back(c);

        return ELC_BREAK;
    }

    if(!token.compare((string)"telnet_server")) {
        i1++;
        c_proc_command c(this_module,_telnet_server,error_line,debug);

        string ip=logic_process_line_get_string(i1,ln);
        c.set_next_arg(ip);

        string port=logic_process_line_get_string(i1,ln);
        long port_=stol(port);
        if(0==port_)
            port_=4444;
        c.set_next_arg(port_);

        procedure.commands.push_back(c);

        return ELC_BREAK;
    }

    return ELC_CONTINUE;
}

MY_STATIC e_return logic_compile_proc_c(c_module* this_module, string& ln, string& token, unsigned int& i1, string& error_reason, c_procedure& procedure, vector<string>& UNUSEDARG(all_groups), deque<c_lines>& lines, int& i_brackets, int& error_line, const string& debug, string& result, bool& b_chan, bool& b_proc, vector<c_procedure>& procedures)
{
    if(!token.compare((string)"telnet_server_ipv6")) {
        i1++;
        c_proc_command c(this_module,_telnet_server_ipv6,error_line,debug);

        string ip=logic_process_line_get_string(i1,ln);
        c.set_next_arg(ip);

        string port=logic_process_line_get_string(i1,ln);
        long port_=stol(port);
        if(0==port_)
            port_=4444;
        c.set_next_arg(port_);

        procedure.commands.push_back(c);

        return ELC_BREAK;
    }

    if(!token.compare((string)"ssl_server")) {
        i1++;
        c_proc_command c(this_module,_ssl_server,error_line,debug);

        string ipv=logic_process_line_get_string(i1,ln);
        long ipv_=stol(ipv);
        c.set_next_arg(ipv_);

        string ip=logic_process_line_get_string(i1,ln);
        c.set_next_arg(ip);

        string port=logic_process_line_get_string(i1,ln);
        long port_=stol(port);
        if(0==port_)
            port_=4444;
        c.set_next_arg(port_);

        procedure.commands.push_back(c);

        procedure.minimal_botnet_version=1006;

        return ELC_BREAK;
    }

    if(!token.compare((string)"bot_nick")) {
        c_proc_command c(this_module,_bot_nick,error_line,debug);

        i1++;
        unsigned int i2=i1;
        while(i2<ln.length()) {
            string nick;
            for(; i2<ln.length() && (ln[i2]!=' ' || ln[i2]!=','); ++i2, ++i1)
                nick+=ln[i2];
            i2++;
            if(!nick.empty())
                c.strings.push_back(nick);
        }

        procedure.commands.push_back(c);

        return ELC_BREAK;
    }

    if(!token.compare((string)"eval")) {
        i1++;
        c_proc_command c(this_module,_eval,error_line,debug);

        string exp=logic_process_line_get_string(i1,ln);
        c.set_next_arg(exp);

        c.set_next_arg(result);

        procedure.commands.push_back(c);

        procedure.minimal_botnet_version=1001;

        return ELC_BREAK;
    }

    if(!token.compare((string)"set")) {
        i1++;
        c_proc_command c(this_module,_set,error_line,debug);

        string exp=logic_process_line_get_string(i1,ln);
        c.set_next_arg(exp);

        c.set_next_arg(result);

        procedure.commands.push_back(c);

        procedure.minimal_botnet_version=1001;

        return ELC_BREAK;
    }

    if(!token.compare((string)"change_nick")) {
        c_proc_command c(this_module,_change_nick,error_line,debug);

        i1++;
        string nick=logic_process_line_get_string(i1,ln);
        c.set_next_arg(nick);

        procedure.commands.push_back(c);

        return ELC_BREAK;
    }

    if(!token.compare((string)"bot_ident")) {
        i1++;
        c_proc_command c(this_module,_bot_ident,error_line,debug);

        string ident=logic_process_line_get_string(i1,ln);

        for(unsigned i2=0; i2<ident.length(); ++i2) {
            if(ident[i2]==0x20) {
                error_reason="Ident cannot contain space";
                return ELC_RETURN_TRUE;
            }
        }

        c.set_next_arg(ident);

        procedure.commands.push_back(c);

        return ELC_BREAK;
    }

    if(!token.compare((string)"bot_ident_ipv6")) {
        i1++;
        c_proc_command c(this_module,_bot_ident_ipv6,error_line,debug);

        string ident=logic_process_line_get_string(i1,ln);

        for(unsigned i2=0; i2<ident.length(); ++i2) {
            if(ident[i2]==0x20) {
                error_reason="Ident cannot contain space";
                return ELC_RETURN_TRUE;
            }
        }

        c.set_next_arg(ident);

        procedure.commands.push_back(c);

        procedure.minimal_botnet_version=4;

        return ELC_BREAK;
    }

    if(!token.compare((string)"delete_irc_servers")) {
        i1++;
        c_proc_command c(this_module,_delete_irc_servers,error_line,debug);

        string mask=logic_process_line_get_string(i1,ln);
        c.set_next_arg(mask);

        procedure.commands.push_back(c);

        procedure.minimal_botnet_version=6;

        return ELC_BREAK;
    }

    if(!token.compare((string)"delete_nicks")) {
        i1++;
        c_proc_command c(this_module,_delete_nicks,error_line,debug);

        string mask=logic_process_line_get_string(i1,ln);
        c.set_next_arg(mask);

        procedure.commands.push_back(c);

        procedure.minimal_botnet_version=6;

        return ELC_BREAK;
    }

    if(!token.compare((string)"bot_fullname")) {
        i1++;
        c_proc_command c(this_module,_bot_fullname,error_line,debug);

        string fn=logic_process_line_get_string(i1,ln);
        c.set_next_arg(fn);

        procedure.commands.push_back(c);

        return ELC_BREAK;
    }

    if(!token.compare((string)"bot_auth")) {
        c_proc_command c(this_module,_bot_auth,error_line,debug);

        procedure.commands.push_back(c);

        return ELC_BREAK;
    }

    if(!token.compare((string)"label")) {
        i1++;
        c_proc_command c(this_module,_real_label,error_line,debug);

        string label=logic_process_line_get_string(i1,ln);
        c.set_next_arg(label);

        procedure.commands.push_back(c);

        return ELC_BREAK;
    }

    if(!token.compare((string)"goto")) {
        i1++;
        c_proc_command c(this_module,_real_goto,error_line,debug);

        c.set_goto_line(-1);

        string label=logic_process_line_get_string(i1,ln);
        c.set_next_arg(label);

        deque<c_proc_command>::iterator i2;
        for(i2=procedure.commands.begin(); i2!=procedure.commands.end(); ++i2)
            if((*i2).get_command()==_real_label && !(*i2).get_label().compare(label)) {
                c.set_goto_line((*i2).get_line());
                break;
            }

        if(c.get_goto_line()==-1) {
            error_reason="goto command to unknown label";
            return ELC_RETURN_TRUE;
        }

        procedure.commands.push_back(c);

        return ELC_BREAK;
    }

    if(!token.compare((string)"allow_redirect")) {
        i1++;
        c_proc_command c(this_module,_allow_redirect,error_line,debug);

        string param=logic_process_line_get_string(i1,ln);
        c.set_next_arg((!param.compare((string)"0"))?0:1);

        procedure.commands.push_back(c);

        return ELC_BREAK;
    }

    if(!token.compare((string)"sleep")) {
        i1++;
        string parm=logic_process_line_get_string(i1,ln);

        c_proc_command c(this_module,_sleep_,error_line,debug);
        c.set_next_arg(stol(parm));

        procedure.commands.push_back(c);

        return ELC_BREAK;
    }

    if(!token.compare((string)"dynamic_ban") && !b_chan) {
        i1++;
        c_proc_command c(this_module,_dynamic_ban,error_line,debug);

        string chan=logic_process_line_get_string(i1,ln);
        c.set_next_arg(chan);

        string prefix=logic_process_line_get_string(i1,ln);
        c.set_next_arg(prefix);

        string mask=logic_process_line_get_string(i1,ln);
        c.set_next_arg(mask);

        string reason=logic_process_line_get_string(i1,ln);
        c.set_next_arg(reason);

        procedure.commands.push_back(c);

        return ELC_BREAK;
    }

    if(!token.compare((string)"process_on_banned")) {
        i1++;
        c_proc_command c(this_module,_process_on_banned,error_line,debug);

        string chan=logic_process_line_get_string(i1,ln);
        c.set_next_arg(chan);

        string mask=logic_process_line_get_string(i1,ln);
        c.set_next_arg(mask);

        procedure.commands.push_back(c);

        return ELC_BREAK;
    }

    if(!token.compare((string)"check_dynamic_bans")) {
        i1++;
        c_proc_command c(this_module,_check_dynamic_bans,error_line,debug);

        string chan=logic_process_line_get_string(i1,ln);
        c.set_next_arg(chan);

        string nick=logic_process_line_get_string(i1,ln);
        c.set_next_arg(nick);

        string reason_var=logic_process_line_get_string(i1,ln);
        c.set_next_arg(reason_var);

        c.set_next_arg(result);

        procedure.commands.push_back(c);

        return ELC_BREAK;
    }

    if(!token.compare((string)"restart")) {
        c_proc_command c(this_module,_restart,error_line,debug);

        procedure.commands.push_back(c);

        return ELC_BREAK;
    }

    if(!token.compare((string)"raw")) {
        i1++;
        c_proc_command c(this_module,_raw,error_line,debug);

        string priority=logic_process_line_get_string(i1,ln);
        long priority_;
        if((priority_=logic_get_priority(priority.c_str()))==0) {
            error_reason="Unknown priority class";
            return ELC_RETURN_TRUE;
        }
        c.set_next_arg(priority_);

        string raw=logic_process_line_get_string(i1,ln);
        c.set_next_arg(raw);

        procedure.commands.push_back(c);

        procedure.minimal_botnet_version=2;

        return ELC_BREAK;
    }

    if(!token.compare((string)"rem")) {
        i1++;
        c_proc_command c(this_module,_rem,error_line,debug);

        string rem=logic_process_line_get_string(i1,ln);
        c.set_next_arg(rem);

        procedure.commands.push_back(c);

        procedure.minimal_botnet_version=12;

        return ELC_BREAK;
    }

    if(!token.compare((string)"admin_msg")) {
        i1++;
        c_proc_command c(this_module,_admin_msg,error_line,debug);

        string type=logic_process_line_get_string(i1,ln);
        c.set_next_arg(type);

        string mask=logic_process_line_get_string(i1,ln);
        c.set_next_arg(mask);

        string message=logic_process_line_get_string(i1,ln);
        c.set_next_arg(message);

        procedure.commands.push_back(c);

        procedure.minimal_botnet_version=10;

        return ELC_BREAK;
    }

    if(!token.compare((string)"load_module")) {
        i1++;
        c_proc_command c(this_module,_load_module,error_line,debug);

        string mod_name=logic_process_line_get_string(i1,ln);
        c.set_next_arg(mod_name);

        procedure.commands.push_back(c);

        procedure.minimal_botnet_version=1001;

        return ELC_BREAK;
    }

    if(!token.compare((string)"unload_module")) {
        i1++;
        c_proc_command c(this_module,_unload_module,error_line,debug);

        string mod_name=logic_process_line_get_string(i1,ln);
        c.set_next_arg(mod_name);

        procedure.commands.push_back(c);

        procedure.minimal_botnet_version=1001;

        return ELC_BREAK;
    }

    if(!token.compare((string)"bind")) {
        i1++;
        c_proc_command c(this_module,_bind,error_line,debug);

        string arg;

        // type mask
        arg=logic_process_line_get_string(i1,ln);
        c.set_next_arg(arg);

        // user mask
        arg=logic_process_line_get_string(i1,ln);
        c.set_next_arg(arg);

        // channel mask
        arg=logic_process_line_get_string(i1,ln);
        c.set_next_arg(arg);

        // line mask
        arg=logic_process_line_get_string(i1,ln);
        c.set_next_arg(arg);

        // before
        arg=logic_process_line_get_string(i1,ln);
        c.set_next_arg((!arg.compare((string)"0"))?0:1);

        // function/procedure
        arg=logic_process_line_get_string(i1,ln);
        c.set_next_arg(arg);

        procedure.commands.push_back(c);

        procedure.minimal_botnet_version=1001;

        return ELC_BREAK;
    }

    if(!token.compare((string)"unbind")) {
        i1++;
        c_proc_command c(this_module,_unbind,error_line,debug);

        string arg;

        // type mask
        arg=logic_process_line_get_string(i1,ln);
        c.set_next_arg(arg);

        // user mask
        arg=logic_process_line_get_string(i1,ln);
        c.set_next_arg(arg);

        // channel mask
        arg=logic_process_line_get_string(i1,ln);
        c.set_next_arg(arg);

        // line mask
        arg=logic_process_line_get_string(i1,ln);
        c.set_next_arg(arg);

        // before
        arg=logic_process_line_get_string(i1,ln);
        c.set_next_arg((!arg.compare((string)"0"))?0:1);

        // function/procedure
        arg=logic_process_line_get_string(i1,ln);
        c.set_next_arg(arg);

        procedure.commands.push_back(c);

        procedure.minimal_botnet_version=1001;

        return ELC_BREAK;
    }

    if(!token.compare((string)"bind_ex")) {
        i1++;
        c_proc_command c(this_module,_bind_ex,error_line,debug);

        string arg;

        // type mask
        arg=logic_process_line_get_string(i1,ln);
        c.set_next_arg(arg);

        // user mask
        arg=logic_process_line_get_string(i1,ln);
        c.set_next_arg(arg);

        // channel mask
        arg=logic_process_line_get_string(i1,ln);
        c.set_next_arg(arg);

        // line mask
        arg=logic_process_line_get_string(i1,ln);
        c.set_next_arg(arg);

        // before
        arg=logic_process_line_get_string(i1,ln);
        c.set_next_arg((!arg.compare((string)"0"))?0:1);

        // function/procedure
        arg=logic_process_line_get_string(i1,ln);
        c.set_next_arg(arg);

        procedure.commands.push_back(c);

        procedure.minimal_botnet_version=1009;

        return ELC_BREAK;
    }

    if(!token.compare((string)"unbind_ex")) {
        i1++;
        c_proc_command c(this_module,_unbind_ex,error_line,debug);

        string arg;

        // type mask
        arg=logic_process_line_get_string(i1,ln);
        c.set_next_arg(arg);

        // user mask
        arg=logic_process_line_get_string(i1,ln);
        c.set_next_arg(arg);

        // channel mask
        arg=logic_process_line_get_string(i1,ln);
        c.set_next_arg(arg);

        // line mask
        arg=logic_process_line_get_string(i1,ln);
        c.set_next_arg(arg);

        // before
        arg=logic_process_line_get_string(i1,ln);
        c.set_next_arg((!arg.compare((string)"0"))?0:1);

        // function/procedure
        arg=logic_process_line_get_string(i1,ln);
        c.set_next_arg(arg);

        procedure.commands.push_back(c);

        procedure.minimal_botnet_version=1009;

        return ELC_BREAK;
    }

    if(!token.compare((string)"load_bin_module")) {
        i1++;
        c_proc_command c(this_module,_load_bin_module,error_line,debug);

        string mod_name=logic_process_line_get_string(i1,ln);
        c.set_next_arg(mod_name);

        string so_name=logic_process_line_get_string(i1,ln);
        c.set_next_arg(so_name);

        procedure.commands.push_back(c);

        procedure.minimal_botnet_version=1007;

        return ELC_BREAK;
    }

    if(!token.compare((string)"get_current_irc_server")) {
        i1++;
        c_proc_command c(this_module,_get_current_irc_server,error_line,debug);

        string server_out_var=logic_process_line_get_string(i1,ln);
        c.set_next_arg(server_out_var);

        string port_out_var=logic_process_line_get_string(i1,ln);
        c.set_next_arg(port_out_var);

        procedure.commands.push_back(c);

        procedure.minimal_botnet_version=1008;

        return ELC_BREAK;
    }

    if(!token.compare((string)"get_bot_name")) {
        i1++;
        c_proc_command c(this_module,_get_bot_name,error_line,debug);

        c.set_next_arg(result);

        procedure.commands.push_back(c);

        procedure.minimal_botnet_version=1008;

        return ELC_BREAK;
    }

    if(!token.compare((string)"get_nick_mode")) {
        i1++;
        c_proc_command c(this_module,_get_nick_mode,error_line,debug);

        string chan=logic_process_line_get_string(i1,ln);
        c.set_next_arg(chan);

        string nick=logic_process_line_get_string(i1,ln);
        c.set_next_arg(nick);

        c.set_next_arg(result);

        procedure.commands.push_back(c);

        procedure.minimal_botnet_version=1001;

        return ELC_BREAK;
    }

    if(!token.compare((string)"get_bot_nick")) {
        i1++;
        c_proc_command c(this_module,_get_bot_nick,error_line,debug);

        c.set_next_arg(result);

        procedure.commands.push_back(c);

        procedure.minimal_botnet_version=1001;

        return ELC_BREAK;
    }

    if(!token.compare((string)"get_event_id")) {
        i1++;
        c_proc_command c(this_module,_get_event_id,error_line,debug);

        c.set_next_arg(result);

        procedure.commands.push_back(c);

        procedure.minimal_botnet_version=1001;

        return ELC_BREAK;
    }

    if(!token.compare((string)"}")) {
        if(i_brackets--<0) {
            error_reason="Unexpected '}'";
            return ELC_RETURN_TRUE;
        }

        if(i_brackets==0)
            b_proc=false;

        c_proc_command c(this_module,_end_bracket,error_line,debug);
        procedure.commands.push_back(c);

        c=c_proc_command(this_module,_label,error_line,debug);

        procedure.commands.push_back(c);

        bool got_jump=false;

        deque<c_lines>::reverse_iterator i;
        for(i=lines.rbegin(); i!=lines.rend(); ++i) {
            deque<c_proc_command>::reverse_iterator i2;
            for(i2=procedure.commands.rbegin(); i2!=procedure.commands.rend(); ++i2) {
                if((*i).line==(*i2).get_line()) {
                    (*i2).set_goto_line(error_line);
                    got_jump=true;
                    break;
                }
            }
            if(got_jump)
                break;
        }

        if(lines.rbegin()!=lines.rend())
            lines.pop_back();

        if(i_brackets==0) {
            procedures.push_back(procedure);
        }

        return ELC_BREAK;
    }

    return ELC_CONTINUE;
}

MY_STATIC e_return logic_compile_chan_a(c_module* this_module, string& ln, string& token, unsigned int& i1, string& error_reason, c_channel& UNUSEDARG(channel), c_channel_def& chan, vector<string>& all_groups, const string& debug)
{
    if(!token.compare((string)"last_changed")) {
        i1++;
        string time=logic_process_line_get_string(i1,ln);
        chan.last_changed=sto_time(time);
        return ELC_RETURN_FALSE;
    }

    if(!token.compare((string)"member_of_group")) {
        i1++;
        string grp=logic_eval_str(this_module,logic_process_line_get_string(i1,ln),debug);

        chan.groups.push_back(grp);

        all_groups.push_back(grp);

        return ELC_BREAK;
    }

    if(!token.compare((string)"replication")) {
        i1++;
        string partner=logic_process_line_get_string(i1,ln);

        string type=logic_process_line_get_string(i1,ln);

        c_replication r;
        r.partner=partner;
        r.type=REPL_INVALID;
        if(!type.compare("push"))
            r.type=REPL_PUSH;
        if(!type.compare("pull"))
            r.type=REPL_PULL;
        if(!type.compare("pushpull"))
            r.type=REPL_PUSHPULL;
        if(r.type==REPL_INVALID) {
            error_reason="Invalid replication type";
            return ELC_RETURN_TRUE;
        }

        chan.replication.push_back(r);

        return ELC_RETURN_FALSE;
    }

    if(!token.compare((string)"on_mode")) {
        if(chan.on_mode.compare("")) {
            error_reason="Duplicate event";
            return ELC_RETURN_TRUE;
        }

        i1++;
        string proc=logic_process_line_get_string(i1,ln);

        chan.on_mode=proc;

        return ELC_BREAK;
    }

    if(!token.compare((string)"on_key")) {
        if(chan.on_key.compare("")) {
            error_reason="Duplicate event";
            return ELC_RETURN_TRUE;
        }

        i1++;
        string proc=logic_process_line_get_string(i1,ln);

        chan.on_key=proc;

        return ELC_BREAK;
    }

    if(!token.compare((string)"on_limit")) {
        if(chan.on_limit.compare("")) {
            error_reason="Duplicate event";
            return ELC_RETURN_TRUE;
        }

        i1++;
        string proc=logic_process_line_get_string(i1,ln);

        chan.on_limit=proc;

        return ELC_BREAK;
    }

    if(!token.compare((string)"on_topic")) {
        if(chan.on_topic.compare("")) {
            error_reason="Duplicate event";
            return ELC_RETURN_TRUE;
        }

        i1++;
        string proc=logic_process_line_get_string(i1,ln);

        chan.on_topic=proc;

        return ELC_BREAK;
    }

    if(!token.compare((string)"on_ircop")) {
        if(chan.on_ircop.compare("")) {
            error_reason="Duplicate event";
            return ELC_RETURN_TRUE;
        }

        i1++;
        string proc=logic_process_line_get_string(i1,ln);

        chan.on_ircop=proc;

        return ELC_BREAK;
    }

    if(!token.compare((string)"on_ctcp")) {
        if(chan.on_ctcp.compare("")) {
            error_reason="Duplicate event";
            return ELC_RETURN_TRUE;
        }

        i1++;
        string proc=logic_process_line_get_string(i1,ln);

        chan.on_ctcp=proc;

        return ELC_BREAK;
    }

    if(!token.compare((string)"on_privmsg")) {
        if(chan.on_privmsg.compare("")) {
            error_reason="Duplicate event";
            return ELC_RETURN_TRUE;
        }

        i1++;
        string proc=logic_process_line_get_string(i1,ln);

        chan.on_privmsg=proc;

        return ELC_BREAK;
    }

    if(!token.compare((string)"on_notice")) {
        if(chan.on_notice.compare("")) {
            error_reason="Duplicate event";
            return ELC_RETURN_TRUE;
        }

        i1++;
        string proc=logic_process_line_get_string(i1,ln);

        chan.on_notice=proc;

        return ELC_BREAK;
    }

    if(!token.compare((string)"on_part")) {
        if(chan.on_part.compare("")) {
            error_reason="Duplicate event";
            return ELC_RETURN_TRUE;
        }

        i1++;
        string proc=logic_process_line_get_string(i1,ln);

        chan.on_part=proc;

        return ELC_BREAK;
    }

    if(!token.compare((string)"on_dynamic_ban")) {
        if(chan.on_dynamic_ban.compare("")) {
            error_reason="Duplicate event";
            return ELC_RETURN_TRUE;
        }

        i1++;
        string proc=logic_process_line_get_string(i1,ln);

        chan.on_dynamic_ban=proc;

        return ELC_BREAK;
    }

    if(!token.compare((string)"on_server_msg")) {
        if(chan.on_server_msg.compare("")) {
            error_reason="Duplicate event";
            return ELC_RETURN_TRUE;
        }

        i1++;
        string proc=logic_process_line_get_string(i1,ln);

        chan.on_server_msg=proc;

        return ELC_BREAK;
    }

    if(!token.compare((string)"allow_dynamic")) {
        i1++;
        string group=logic_eval_str(this_module,logic_process_line_get_string(i1,ln),debug);

        string plus=logic_eval_str(this_module,logic_process_line_get_string(i1,ln),debug);

        string minus=logic_eval_str(this_module,logic_process_line_get_string(i1,ln),debug);

        c_dynamic_rule r;
        r.group=group;
        r.plus_modes=plus;
        r.minus_modes=minus;

        chan.allow_dynamic.push_back(r);

        return ELC_RETURN_FALSE;
    }

    if(!token.compare((string)"dynamic1")) {
        i1++;
        string modes=logic_process_line_get_string(i1,ln);

        {
            // compatibility
            bool changed=(modes.length() && modes[0]!='"') || modes.empty();
            if(modes.length() && modes[0]!='"')
                modes=(string)"\""+modes+"\"";
            if(modes.empty())
                modes="\"\"";

            string log="OOPS: Compatibility warning: forcing dynamic1 modes "+modes+" to be enclosed into double-quotes. This problem will be resolved on the next backup.";
            if(changed)
                log_debug(log);
        }

        chan.dynamic_plus_modes=logic_eval_str(this_module,modes,debug);

        return ELC_RETURN_FALSE;
    }

    if(!token.compare((string)"dynamic2")) {
        i1++;
        string modes=logic_process_line_get_string(i1,ln);

        {
            // compatibility
            bool changed=(modes.length() && modes[0]!='"') || modes.empty();
            if(modes.length() && modes[0]!='"')
                modes=(string)"\""+modes+"\"";
            if(modes.empty())
                modes="\"\"";

            string log="OOPS: Compatibility warning: forcing dynamic2 modes "+modes+" to be enclosed into double-quotes. This problem will be resolved on the next backup.";
            if(changed)
                log_debug(log);
        }

        chan.dynamic_minus_modes=logic_eval_str(this_module,modes,debug);

        return ELC_RETURN_FALSE;
    }

    if(!token.compare((string)"key")) {
        i1++;
        string key=logic_process_line_get_string(i1,ln);

        {
            // compatibility
            bool changed=(key.length() && key[0]!='"') || key.empty();
            if(key.length() && key[0]!='"')
                key=(string)"\""+key+"\"";
            if(key.empty())
                key="\"\"";

            string log="OOPS: Compatibility warning: forcing channel key "+key+" to be enclosed into double-quotes. This problem will be resolved on the next backup.";
            if(changed)
                log_debug(log);
        }

        chan.keys.push_back(key);

        return ELC_BREAK;
    }

    if(!token.compare((string)"dynamic_ban")) {
        i1++;
        string mask=logic_eval_str(this_module,logic_process_line_get_string(i1,ln),debug);

        string reason=logic_eval_str(this_module,logic_process_line_get_string(i1,ln),debug);

        pair<string,string> p;
        p.first=mask;
        p.second=reason;
        chan.dynamic_bans.insert(p);

        return ELC_BREAK;
    }

    if(!token.compare((string)"partyline_dynbans_editors")) {
        i1++;
        string group=logic_eval_str(this_module,logic_process_line_get_string(i1,ln),debug);
        chan.dynbans_editors.push_back(group);

        return ELC_BREAK;
    }

    return ELC_CONTINUE;
}

MY_STATIC e_return logic_compile_prv_a(c_module* this_module, string& ln, string& token, unsigned int& i1, string& error_reason, c_private& prv, vector<string>& all_groups, const string& debug)
{
    if(!token.compare((string)"member_of_group")) {
        i1++;
        string grp=logic_eval(this_module,logic_process_line_get_string(i1,ln),NULL,debug);

        prv.groups.push_back(grp);

        all_groups.push_back(grp);

        return ELC_BREAK;
    }

    if(!token.compare((string)"on_privmsg")) {
        i1++;
        string proc=logic_process_line_get_string(i1,ln);

        prv.on_privmsg=proc;

        return ELC_RETURN_FALSE;
    }

    if(!token.compare((string)"on_notice")) {
        i1++;
        string proc=logic_process_line_get_string(i1,ln);

        prv.on_notice=proc;

        return ELC_RETURN_FALSE;
    }

    if(!token.compare((string)"on_ctcp")) {
        i1++;
        string proc=logic_process_line_get_string(i1,ln);

        prv.on_ctcp=proc;

        return ELC_RETURN_FALSE;
    }

    if(!token.compare((string)"on_filesys_got_new")) {
        i1++;
        string proc=logic_process_line_get_string(i1,ln);

        prv.on_filesys_got_new=proc;

        return ELC_RETURN_FALSE;
    }

    if(!token.compare((string)"on_internal_event")) {
        i1++;
        string proc=logic_process_line_get_string(i1,ln);

        prv.on_internal_event=proc;

        return ELC_RETURN_FALSE;
    }

    if(!token.compare((string)"on_fnc")) {
        i1++;
        string proc=logic_process_line_get_string(i1,ln);

        prv.on_fnc=proc;

        return ELC_RETURN_FALSE;
    }

    if(!token.compare((string)"on_broadcast")) {
        i1++;
        string proc=logic_process_line_get_string(i1,ln);

        prv.on_broadcast=proc;

        return ELC_RETURN_FALSE;
    }

    if(!token.compare((string)"on_server_msg")) {
        i1++;
        string proc=logic_process_line_get_string(i1,ln);

        prv.on_server_msg=proc;

        return ELC_RETURN_FALSE;
    }

    if(!token.compare((string)"replication")) {
        i1++;
        string partner=logic_process_line_get_string(i1,ln);

        string type=logic_process_line_get_string(i1,ln);

        c_replication r;
        r.partner=partner;
        r.type=REPL_INVALID;
        if(!type.compare("push"))
            r.type=REPL_PUSH;
        if(!type.compare("pull"))
            r.type=REPL_PULL;
        if(!type.compare("pushpull"))
            r.type=REPL_PUSHPULL;
        if(r.type==REPL_INVALID) {
            error_reason="Invalid replication type";
            return ELC_RETURN_TRUE;
        }

        prv.replication.push_back(r);

        return ELC_RETURN_FALSE;
    }

    if(!token.compare((string)"last_changed")) {
        i1++;
        string time=logic_process_line_get_string(i1,ln);
        prv.last_changed=sto_time(time);
        return ELC_RETURN_FALSE;
    }

    return ELC_CONTINUE;
}

