/***************************************************************************
                          logic_events.cpp  -  description
                             -------------------
    begin                : Thu May 10 2007
    copyright            : (C) 2004-2007 by Marian VooDooMan Meravy
    email                : ghostvoodooman -a.T= users (dot) sourceforge.net
 ***************************************************************************/

/* $Id: logic_events.cpp,v 1.18 2008/08/23 15:10:44 ghostvoodooman Exp $ */

/***************************************************************************

VooDoo cIRCle - an IRC (ro)bot
Copyright (C) 2004-2008 by Marian VooDooMan Meravy
  (ghostvoodooman -a.T= users (dot) sourceforge.net)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************/

/*!
    \file
    \brief Implements events for 'logic.txt'
    \version $Id: logic_events.cpp,v 1.18 2008/08/23 15:10:44 ghostvoodooman Exp $
*/

#if defined(HAVE_CONFIG_H)
#   include <config.h>
#endif

#include <string>
#include <vector>
#include <stack>
#include <list>
#include <map>
#include <algorithm>

#include "../params.h"
#include "../utils/memleak.h"
#include "../utils/assert.h"
#include "../compat.h"

#include "c_channel_def.h"
#include "c_user.h"
#include "c_channel.h"
#include "../modules.h"
#include "../irc.h"
#include "../logic.h"
#include "logic_internal.h"
#include "../dcc.h"
#include "../var/c_var_string.h"
#include "../mod/c_module.h"
#include "../log.h"
#include "../misc/shutdown.h"
#include "event_dispatch.h"
#include "../internal_state.h"
#include "../irc/c_online_channel.h"

extern deque<c_user> r_terminators;
extern deque<c_user> r_user_templates;
extern deque<c_user> r_users;

#include "../utils/precomp.h"

#if defined(_MSC_VER) || defined(__BORLANDC__)
#   pragma hdrstop
#endif

/*!
    \brief Validates the mode change of the channel and (if needed) calls events
    \author VooDooMan
    \version 2
    \date 2004, 2008-03-09
    \param channel Name of the channel
    \param source_user Originator user name from "logic.txt"
    \param source_nick Nick name of originator
    \param prefix '+' or '-'
    \param mode Mode character
*/
void logic_validate_chan_mode(const string& channel, const string& source_user, const string& source_nick, const char prefix, const char mode)
{
    if(irc_cmp_strings_case_insensitive(source_nick,irc_nick)) {
        vector<c_channel_def>::iterator i1;

        for(i1=r_channel_defs.begin(); i1!=r_channel_defs.end(); ++i1) {
            if(!irc_cmp_strings_case_insensitive((*i1).channel_name,channel)) {
                c_user u;
                vector<c_channel> chs;
                logic_partyline_get_user(source_user,u,chs);
                vector<c_dynamic_rule>::iterator i2;
                for(i2=(*i1).allow_dynamic.begin(); i2!=(*i1).allow_dynamic.end(); ++i2) {
                    vector<string>::iterator i3;
                    bool got=false;
                    for(i3=u.groups.begin(); i3!=u.groups.end(); ++i3) {
                        if(!(*i2).group.compare(*i3)) {
                            if((prefix=='-' && (*i2).minus_modes.find((string)""+mode,0)==string::npos))
                                continue;
                            if((prefix=='+' && (*i2).plus_modes.find((string)""+mode,0)==string::npos))
                                continue;
                            got=true;
                            while((*i1).dynamic_plus_modes.find((string)""+mode,0)!=string::npos)
                                (*i1).dynamic_plus_modes.erase((*i1).dynamic_plus_modes.find((string)""+mode,0),1);
                            while((*i1).dynamic_minus_modes.find((string)""+mode,0)!=string::npos)
                                (*i1).dynamic_minus_modes.erase((*i1).dynamic_minus_modes.find((string)""+mode,0),1);
                            if(prefix=='-')
                                (*i1).dynamic_minus_modes+=(string)""+mode;
                            else
                                (*i1).dynamic_plus_modes+=(string)""+mode;
                            break;
                        }
                    }
                    if(got)
                        break;
                }

                string proc=(*i1).on_mode;

                vector<string> params;
                params.push_back(channel); // $channel
                params.push_back(source_user); // $source
                params.push_back(source_nick); // $source_nick

                string str="";
                if(prefix=='+')
                    str=mode;

                params.push_back(str); // $plus_modes

                str="";
                if(prefix=='-')
                    str=mode;

                params.push_back(str); // $minus_modes

                logic_core_execute(proc,params,"on_mode",params.size(),"on_mode@channel");
            }
        }
    }

    event_dispatch_mode_on_channel(channel,source_nick,prefix,mode,"",false);
}

/*!
    \brief Validates the key change (+/-k) for the channel and (if needed) calls event, and adds a key to the history
    \author VooDooMan
    \version 1
    \date 2004
    \param channel Name of the channel
    \param source_user Originator user name from "logic.txt"
    \param source_nick Nick name of originator
    \param prefix '+' or '-'
    \param key Channel key
*/
void logic_validate_chan_key(const string& channel, const string& source_user, const string& source_nick, const char prefix, const string& key)
{
    if(irc_cmp_strings_case_insensitive(source_nick,irc_nick)) {
        vector<c_channel_def>::iterator i1;

        for(i1=r_channel_defs.begin(); i1!=r_channel_defs.end(); ++i1) {
            if(!irc_cmp_strings_case_insensitive((*i1).channel_name,channel)) {

                (*i1).keys.push_back(key);

                string proc=(*i1).on_key;

                vector<string> params;
                params.push_back(channel); // $channel
                params.push_back(source_user); // $source
                params.push_back(source_nick); // $source_nick

                params.push_back((string)""+prefix); // $prefix
                params.push_back(key); // $key

                logic_core_execute(proc,params,"on_key",params.size(),"on_key@channel");
            }
        }
    }

    event_dispatch_mode_on_channel(channel,source_nick,prefix,'k',key,false);
}

/*!
    \brief Validates the limit change (+/-l) for the channel and (if needed) calls event
    \author VooDooMan
    \version 1
    \date 2004
    \param channel Name of the channel
    \param source_user Originator user name from "logic.txt"
    \param source_nick Nick name of originator
    \param prefix '+' or '-'
    \param limit Channel limit
*/
void logic_validate_chan_limit(const string& channel, const string& source_user, const string& source_nick, const char prefix, const string& limit)
{
    if(irc_cmp_strings_case_insensitive(source_nick,irc_nick)) {
        vector<c_channel_def>::iterator i1;

        for(i1=r_channel_defs.begin(); i1!=r_channel_defs.end(); ++i1) {
            if(!irc_cmp_strings_case_insensitive((*i1).channel_name,channel)) {
                string proc=(*i1).on_limit;

                vector<string> params;
                params.push_back(channel); // $channel
                params.push_back(source_user); // $source
                params.push_back(source_nick); // $source_nick

                params.push_back((string)""+prefix); // $prefix
                params.push_back(limit); // $limit

                logic_core_execute(proc,params,"on_limit",params.size(),"on_limit@channel");
            }
        }
    }

    event_dispatch_mode_on_channel(channel,source_nick,prefix,'l',limit,false);
}

/*!
    \brief Validates the topic change and (if assigned) calls event
    \author VooDooMan
    \version 1
    \date 2004
    \param channel Name of the channel
    \param source_user Originator user name from "logic.txt"
    \param source_nick Nick name of originator
    \param topic New topic
*/
void logic_validate_topic(const string& channel, const string& source_user, const string& source_nick, const string& topic)
{
    if(irc_cmp_strings_case_insensitive(source_nick,irc_nick)) {
        vector<c_channel_def>::iterator i1;

        for(i1=r_channel_defs.begin(); i1!=r_channel_defs.end(); ++i1) {
            if(!irc_cmp_strings_case_insensitive((*i1).channel_name,channel)) {
                string proc=(*i1).on_topic;
                if(proc.empty())
                    continue;

                vector<string> params;
                params.push_back(channel); // $channel
                params.push_back(source_user); // $source
                params.push_back(source_nick); // $source_nick

                params.push_back(topic); // $topic

                logic_core_execute(proc,params,"on_topic",params.size(),"on_topic@channel");
            }
        }
    }

    mod_dispatch("topic@channel",source_user,channel,"",false,0,"channel",channel.c_str(),"source",source_user.c_str(),"source_nick",source_nick.c_str(),"topic",topic.c_str(),NULL,"event_type","channel","nick",source_nick.c_str(),"origin","@irc@","ident",irc_get_ident(source_nick).c_str(),"host",irc_get_host(source_nick).c_str(),NULL);
}

/*!
    \brief If we are .upgrade'ing then notifies about mode change for new bot
    \author VooDooMan
    \version 1
    \param target_nick Target nick
    \param channel Channel
    \param prefix '+', or '-'
    \param action '@' or similat
*/
void logic_user_mode_changing(const string& target_nick, const string& channel, char prefix, char action)
{
    e_shutdown_reason sr=bot_get_shutdown_state();
    if(E_SD_UPGRADE==sr) {
        string ident=irc_get_ident(target_nick);
        string host=irc_get_host(target_nick);
        string fullname=irc_get_fullname(target_nick);
        string user=logic_find_user(target_nick, ident, host, fullname, irc_is_ircop(target_nick));
        deque<c_user>::iterator i1;
        for(i1=r_users.begin(); i1!=r_users.end(); ++i1) {
            if((*i1).host_bot && !(*i1).name.compare(user) && target_nick.compare(irc_nick)) { // target_nick!=my nick
                string mode=(string)""+prefix;
                mode+=(string)""+action;
                dcc_upgrade_bot_mode(target_nick,ident,host,fullname,channel,mode,irc_nick);
                break;
            }
        }
    }
}

/*!
    \brief Validates MODE change and, if needed, executes on_XXX event
    \author VooDooMan
    \version 1
    \date 2004
    \param channel Name of channel
    \param source_user Name of user (originator) as in logic.txt
    \param target_user Name of user (victim) as in logic.txt
    \param source_nick Nick name of originator
    \param target_nick Nick name of victim
    \param prefix '+' or '-'
    \param action '@', '+', '&', '%' mode chars
*/
void logic_validate(const string& channel, const string& source_user, const string& target_user, const string& source_nick, const string& target_nick, const char prefix, const char action)
{
    if(irc_cmp_strings_case_insensitive(source_nick,irc_nick)) {
        e_shutdown_reason sr=bot_get_shutdown_state();
        if(E_SD_UPGRADE==sr) {
            deque<c_user>::iterator i1;
            for(i1=r_users.begin(); i1!=r_users.end(); ++i1) {
                if((*i1).host_bot && !(*i1).name.compare(target_user) && target_nick.compare(irc_nick)) { // target_nick!=my nick
                    string ident=irc_get_ident(target_nick);
                    string host=irc_get_host(target_nick);
                    string fullname=irc_get_fullname(target_nick);
                    string mode=(string)""+prefix;
                    mode+=(string)""+irc_isupport.get_prefix2(action);
                    string src=source_nick;
                    src+="!";
                    src+=irc_get_ident(source_nick);
                    src+="@";
                    src+=irc_get_host(source_nick);
                    dcc_upgrade_bot_mode(target_nick,ident,host,fullname,channel,mode,src);
                    break;
                }
            }
        }

        deque<c_user>::iterator host_bot;
        bool b_host_bot=false;

        bool got_user=false;

        vector<c_channel>::iterator i1;

        {
            for(host_bot=r_users.begin(); host_bot!=r_users.end(); ++host_bot) {
                if((*host_bot).host_bot && !(*host_bot).name.compare(target_user)) {
                    b_host_bot=true;
                    goto l1;
                }
            }
        }

        for(i1=r_channels.begin(); i1!=r_channels.end(); ++i1) {
            if(!(*i1).username.compare(target_user) && !irc_cmp_strings_case_insensitive((*i1).channel_name,channel)) {
                if((prefix=='-' || prefix=='+') /*&& (action=='@' || action=='+' || action=='O')*/) {
                    string proc="";
                    string event;
                    if(prefix=='-' && action=='@') {
                        proc=(*i1).on_deop;
                        event="on_deop";
                    }
                    if(prefix=='+' && action=='@') {
                        proc=(*i1).on_op;
                        event="on_op";
                    }
                    if(prefix=='-' && action=='+') {
                        proc=(*i1).on_devoice;
                        event="on_devoice";
                    }
                    if(prefix=='+' && action=='+') {
                        proc=(*i1).on_voice;
                        event="on_voice";
                    }
                    if(action!='+' && action!='@') {
                        proc=(*i1).on_other_mode;
                        event="on_other_mode";
                    }
                    if(!proc.compare((string)""))
                        continue;

                    event+="@user";

                    vector<string> params;
                    params.push_back(channel); // $channel
                    params.push_back(source_user); // $source
                    params.push_back(target_user); // $target
                    params.push_back(source_nick); // $source_nick
                    params.push_back(target_nick); // $target_nick

                    if(action!='+' && action!='@' && action!='O') {
                        params.push_back((string)""+prefix); // $prefix
                        params.push_back((string)""+action); // $mode_char
                    }

                    logic_core_execute(proc,params,"mode validation",params.size(),event);
                    got_user=true;
                    break;
                }
            }
        }

        if(!got_user) {
            // try host_unknown
        l1:
            deque<c_user>::iterator i;
            if(b_host_bot)
                i=host_bot;
                else {
                    for(i=r_users.begin(); i!=r_users.end(); ++i) {
                        if((*i).host_unknown) {
                            got_user=true;
                            break;
                        }
                    }
                    if(!got_user)
                        return;
                }

            vector<c_channel>::iterator i1;
            for(i1=r_channels.begin(); i1!=r_channels.end(); ++i1) {
                if(irc_cmp_strings_case_insensitive((*i1).channel_name,channel))
                    continue;
                vector<string>::iterator is, ii;
                if(!(*i1).username.compare((*i).name)) {
                    if((prefix=='-' || prefix=='+') /*&& (action=='@' || action=='+' || action=='O')*/) {
                        string proc="";
                        string event;
                        if(prefix=='-' && action=='@') {
                            proc=(*i1).on_deop;
                            event="on_deop";
                        }
                        if(prefix=='+' && action=='@') {
                            proc=(*i1).on_op;
                            event="on_op";
                        }
                        if(prefix=='-' && action=='+') {
                            proc=(*i1).on_devoice;
                            event="on_devoice";
                        }
                        if(prefix=='+' && action=='+') {
                            proc=(*i1).on_voice;
                            event="on_voice";
                        }
                        if(action!='+' && action!='@') {
                            proc=(*i1).on_other_mode;
                            event="on_other_mode";
                        }
                        if(!proc.compare((string)""))
                            continue;

                        event+="@user";

                        vector<string> params;
                        params.push_back(channel); // $channel
                        params.push_back(source_user); // $source
                        params.push_back(target_user); // $target
                        params.push_back(source_nick); // $source_nick
                        params.push_back(target_nick); // $target_nick

                        if(action!='+' && action!='@' && action!='O') {
                            params.push_back((string)""+prefix); // $prefix
                            params.push_back((string)""+action); // $mode_char
                        }

                        logic_core_execute(proc,params,"mode validation",params.size(),event);
                        break;
                    }
                }
            }
        }
    }

    event_dispatch_mode_on_user(channel,target_nick,source_nick,prefix,action,false);
}

/*!
    \brief Validates MODE change and, if needed, executes on_XXX event
    \author VooDooMan
    \version 2
    \date 2004, 2008-03-09
    \param channel Name of channel
    \param source_user Name of user (originator) as in logic.txt
    \param source_nick Nick name of originator
    \param mask Mask of victim
    \param prefix '+' or '-'
    \param action 'b', 'e', 'I', or 'R', or other mode (exotic IRC servers)
*/
void logic_validate_by_mask(const string& channel, const string& source_user, const string& source_nick, const string& mask, const char prefix, const char action)
{
    if(irc_cmp_strings_case_insensitive(source_nick,irc_nick)) {
        bool got_user=false;

        vector<c_channel>::iterator i1;
        for(i1=r_channels.begin(); i1!=r_channels.end(); ++i1) {
            if(irc_cmp_strings_case_insensitive((*i1).channel_name,channel)) // !=
                continue;
            deque<c_user>::iterator i3;
            for(i3=r_users.begin(); i3!=r_users.end(); ++i3) {
                if(!(*i3).name.compare((*i1).username)) {
                    vector<string>::iterator i2;
                    for(i2=(*i3).hostmask.begin(); i2!=(*i3).hostmask.end(); ++i2) {
                        bool was_dns=true;
                        if(!(*i3).host_unknown && logic_check_mask(mask,*i2,"*","*",was_dns)) {
                            got_user=true;

                            string proc="";
                            string event;
                            if(prefix=='-' && action=='b') {
                                proc=(*i1).on_unban;
                                event="on_unban";
                            } else if(prefix=='+' && action=='b') {
                                proc=(*i1).on_ban;
                                event="on_ban";
                            } else if(prefix=='-' && action=='e') {
                                proc=(*i1).on_unexcept;
                                event="on_unexcept";
                            } else if(prefix=='+' && action=='e') {
                                proc=(*i1).on_except;
                                event="on_except";
                            } else if(prefix=='-' && action=='I') {
                                proc=(*i1).on_uninvite;
                                event="on_uninvite";
                            } else if(prefix=='+' && action=='I') {
                                proc=(*i1).on_invite;
                                event="on_invite";
                            } else if(prefix=='-' && action=='R') {
                                proc=(*i1).on_reop;
                                event="on_reop";
                            } else if(prefix=='+' && action=='R') {
                                proc=(*i1).on_reop;
                                event="on_reop";
                            } else {
                                proc=(*i1).on_other_mode_mask;
                                event="on_other_mode_mask";
                            }
                            if(!proc.compare(""))
                                continue;

                            event+="@user";

                            vector<string> params;
                            params.push_back(channel); // $channel
                            params.push_back(source_user); // $source
                            params.push_back((*i3).name); // $target
                            params.push_back(source_nick); // $source_nick
                            params.push_back(irc_get_nick((*i3).name)); // $target_nick

                            if(action=='b') {
                                params.push_back(mask); // $ban_mask
                            } else if('e'==action || 'I'==action || 'R'==action) {
                                params.push_back((string)""+prefix); // $prefix
                                params.push_back(mask); // $mask
                            } else {
                                params.push_back((string)""+prefix); // $prefix
                                params.push_back((string)""+action); // $mode
                                params.push_back(mask); // $mask
                            }

                            logic_core_execute(proc,params,"mode validation with mask for user",params.size(),event);
                            break;
                        }
                    }
                    // do NOT break; coz we need to track EVERY user!
                }
            }
        }

        if(!got_user) {
            // try host_unknown
            deque<c_user>::iterator i;
            for(i=r_users.begin(); i!=r_users.end(); ++i) {
                if((*i).host_unknown) {
                    vector<c_channel>::iterator i1;
                    for(i1=r_channels.begin(); i1!=r_channels.end(); ++i1) {
                        if(!(*i1).username.compare((*i).name) && !irc_cmp_strings_case_insensitive((*i1).channel_name,channel)) {
                            got_user=true;
                            string proc="";
                            string event;
                            if(prefix=='-' && action=='b') {
                                proc=(*i1).on_unban;
                                event="on_unban";
                            } else if(prefix=='+' && action=='b') {
                                proc=(*i1).on_ban;
                                event="on_ban";
                            } else if(prefix=='-' && action=='e') {
                                proc=(*i1).on_unexcept;
                                event="on_unexcept";
                            } else if(prefix=='+' && action=='e') {
                                proc=(*i1).on_except;
                                event="on_except";
                            } else if(prefix=='-' && action=='I') {
                                proc=(*i1).on_uninvite;
                                event="on_uninvite";
                            } else if(prefix=='+' && action=='I') {
                                proc=(*i1).on_invite;
                                event="on_invite";
                            } else if(prefix=='-' && action=='R') {
                                proc=(*i1).on_reop;
                                event="on_reop";
                            } else if(prefix=='+' && action=='R') {
                                proc=(*i1).on_reop;
                                event="on_reop";
                            } else {
                                proc=(*i1).on_other_mode_mask;
                                event="on_other_mode_mask";
                            }
                            if(!proc.compare(""))
                                continue;

                            event+="@user";

                            vector<string> params;
                            params.push_back(channel); // $channel
                            params.push_back(source_user); // $source
                            params.push_back((*i).name); // $target
                            params.push_back(source_nick); // $source_nick
                            params.push_back(""); // $target_nick

                            if(action=='b') {
                                params.push_back(mask); // $ban_mask
                            } else if('e'==action || 'I'==action || 'R'==action) {
                                params.push_back((string)""+prefix); // $prefix
                                params.push_back(mask); // $mask
                            } else {
                                params.push_back((string)""+prefix); // $prefix
                                params.push_back((string)""+action); // $mode
                                params.push_back(mask); // $mask
                            }

                            logic_core_execute(proc,params,"mode validation with mask for an unknown user",params.size(),event);
                            break;
                        }
                    }
                    if(got_user)
                        break;
                }
            }
        }

        if(action=='b') {
            vector<c_channel_def>::iterator i1;
            for(i1=r_channel_defs.begin(); i1!=r_channel_defs.end(); ++i1) {
                if(!irc_cmp_strings_case_insensitive((*i1).channel_name,channel)) {
                    string proc=(*i1).on_dynamic_ban;
                    if(proc.empty())
                        break;
                    vector<string> params;
                    params.push_back(channel); // $channel
                    params.push_back(source_user); // $source
                    params.push_back(source_nick); // $source_nick
                    params.push_back((string)""+prefix); // $prefix
                    params.push_back(mask); // $ban_mask

                    logic_core_execute(proc,params,"mode validation - dynamic ban",params.size(),"on_dynamic_ban@channel");
                    break;
                }
            }
        }
    }

    event_dispatch_mode_on_channel(channel,source_nick,prefix,action,mask,false);
}

/*!
    \brief Executes on_join event (called when someone JOINs the channel)
    \author VooDooMan
    \version 1
    \date 2004
    \param nick Nick name
    \param ident Ident
    \param host Host
    \param fullname Full name
    \param channel Channel name
    \param irc_op Pass there true if WHOIS indicates that this is an irc op
*/
void logic_on_join(const string& nick, const string& ident, const string& host, const string& fullname, const string& channel, const bool irc_op)
{
    bool got_user=false;

    string user=logic_find_user(nick, ident, host, fullname, irc_op);

    e_shutdown_reason sr=bot_get_shutdown_state();
    if(E_SD_UPGRADE==sr) {
        deque<c_user>::iterator i1;
        for(i1=r_users.begin(); i1!=r_users.end(); ++i1) {
            if((*i1).host_bot && !(*i1).name.compare(user) && nick.compare(irc_nick)) { // target_nick!=my nick
                dcc_upgrade_bot_join(nick,ident,host,fullname,channel);
                break;
            }
        }
    }

    vector<c_channel>::iterator i1;
    for(i1=r_channels.begin(); i1!=r_channels.end(); ++i1) {
        if(!(*i1).username.compare(user) && !irc_cmp_strings_case_insensitive((*i1).channel_name,channel)) {

            string proc=(*i1).on_join;
            if(!proc.compare(""))
                break;

            vector<string> params;
            params.push_back(channel); // $channel
            params.push_back(user); // $user
            params.push_back(nick); // $nick
            params.push_back((*i1).dynamic_plus_modes); // $dynamic1
            params.push_back((*i1).dynamic_minus_modes); // $dynamic2
            params.push_back("0"); // $passive

            logic_core_execute(proc,params,"on_join",params.size(),"on_join@user");
            got_user=true;
            break;
        }
    }

    if(!got_user) {
        // try host_unknown
        deque<c_user>::iterator i;
        for(i=r_users.begin(); i!=r_users.end(); ++i) {
            if((*i).host_unknown) {
                got_user=true;
                break;
            }
        }
        if(!got_user)
            return;

        vector<c_channel>::iterator i1;
        for(i1=r_channels.begin(); i1!=r_channels.end(); ++i1) {
            if(!(*i1).username.compare((*i).name) && !irc_cmp_strings_case_insensitive((*i1).channel_name,channel)) {

                string proc=(*i1).on_join;
                if(!proc.compare(""))
                    break;

                vector<string> params;
                params.push_back(channel); // $channel
                params.push_back(user); // $user
                params.push_back(nick); // $nick
                params.push_back((*i1).dynamic_plus_modes); // $dynamic1
                params.push_back((*i1).dynamic_minus_modes); // $dynamic2
                params.push_back("0"); // $passive

                logic_core_execute(proc,params,"on_join",params.size(),"on_join@user");
            }
        }
    }

    mod_dispatch("join@channel",user,channel,"",false,0,"channel",channel.c_str(),"user",user.c_str(),"nick",nick.c_str(),"passive","0",NULL,"event_type","channel","nick",nick.c_str(),"origin","@irc@","ident",irc_get_ident(nick).c_str(),"host",irc_get_host(nick).c_str(),NULL);
}

/*!
    \brief Validates new nick
    \author VooDooMan
    \version 1
    \date 2004
    \param nick Nick name
    \param ident Ident
    \param host Host
    \param fullname Full name
    \param channel Channel name
    \param irc_op Pass there true if WHOIS indicates that this is an irc op
    \param old_nick Previous (old) nick name
*/
void logic_on_nick_change(const string& nick, const string& ident, const string& host, const string& fullname, const string& channel, const bool irc_op, const string& old_nick)
{
    string user=logic_find_user(nick, ident, host, fullname, irc_op);

    e_shutdown_reason sr=bot_get_shutdown_state();
    if(E_SD_UPGRADE==sr) {
        deque<c_user>::iterator i1;
        for(i1=r_users.begin(); i1!=r_users.end(); ++i1) {
            if((*i1).host_bot && !(*i1).name.compare(user) && old_nick.compare(irc_nick)) { // old_nick!=my nick
                dcc_upgrade_bot_nick(old_nick,ident,host,fullname,nick);
                break;
            }
        }
    }

    bool got_user=false;

    string old_user=logic_find_user(old_nick, ident, host, fullname, irc_op);

    vector<c_channel>::iterator i1;
    for(i1=r_channels.begin(); i1!=r_channels.end(); ++i1) {
        if(!(*i1).username.compare(user) && !irc_cmp_strings_case_insensitive((*i1).channel_name,channel) && old_user.compare(user)) {

            got_user=true;

            string proc=(*i1).on_join;
            if(!proc.compare(""))
                break;

            vector<string> params;
            params.push_back(channel); // $channel
            params.push_back(user); // $user
            params.push_back(nick); // $nick
            params.push_back((*i1).dynamic_plus_modes); // $dynamic1
            params.push_back((*i1).dynamic_minus_modes); // $dynamic2
            params.push_back((string)"1"+irc_get_mode(channel,nick)); // $passive

            logic_core_execute(proc,params,"on_join",params.size(),"on_join@user");
            break;
        }
    }

    if(!got_user) {
        // try host_unknown
        deque<c_user>::iterator i;
        for(i=r_users.begin(); i!=r_users.end(); ++i) {
            if((*i).host_unknown) {
                got_user=true;
                break;
            }
        }
        if(!got_user)
            return;

        vector<c_channel>::iterator i1;
        for(i1=r_channels.begin(); i1!=r_channels.end(); ++i1) {
            if(!(*i1).username.compare((*i).name) && !irc_cmp_strings_case_insensitive((*i1).channel_name,channel) && !old_user.compare(user)) {

                string proc=(*i1).on_join;
                if(!proc.compare(""))
                    break;

                vector<string> params;
                params.push_back(channel); // $channel
                params.push_back(user); // $user
                params.push_back(nick); // $nick
                params.push_back((*i1).dynamic_plus_modes); // $dynamic1
                params.push_back((*i1).dynamic_minus_modes); // $dynamic2
                params.push_back((string)"1"+irc_get_mode(channel,nick)); // $passive

                logic_core_execute(proc,params,"on_join",params.size(),"on_join@user");
            }
        }
    }

    mod_dispatch("nick@channel",user,channel,"",false,0,"channel",channel.c_str(),"user",user.c_str(),"nick",nick.c_str(),"old_nick",old_nick.c_str(),NULL,"event_type","channel","nick",nick.c_str(),"origin","@irc@","ident",irc_get_ident(nick).c_str(),"host",irc_get_host(nick).c_str(),NULL);
}

/*!
    \brief Validates user after bot has joined / got +o (op) / rehashed
    \author VooDooMan
    \version 1
    \date 2004
    \param nick Nick name
    \param ident Ident
    \param host Host
    \param fullname Full name
    \param channel Channel name
    \param irc_op Pass there true if WHOIS indicates that this is an irc op
    \param passive Set to true if this is called after .rehash, or after bot's join - it pass to possibly new user or new host to $passive it's mode as an argument to procedure on_join in "logic.txt" file
    \param mod_disp_join_called true if we have called mod_dispatch() for "join@channel", if false, it will automatically return true in this argument
*/
void logic_on_nick_validate(const string& nick, const string& ident, const string& host, const string& fullname, const string& channel, const bool irc_op, const bool passive, bool& mod_disp_join_called)
{
    bool got_user=false;

    string user=logic_find_user(nick, ident, host, fullname, irc_op);

    vector<c_channel>::iterator i1;
    for(i1=r_channels.begin(); i1!=r_channels.end(); ++i1) {
        if(!(*i1).username.compare(user) && !irc_cmp_strings_case_insensitive((*i1).channel_name,channel)) {
            got_user=true;

            string proc=(*i1).on_join;
            if(!proc.compare(""))
                break;

            vector<string> params;
            params.push_back(channel); // $channel
            params.push_back(user); // $user
            params.push_back(nick); // $nick
            params.push_back((*i1).dynamic_plus_modes); // $dynamic1
            params.push_back((*i1).dynamic_minus_modes); // $dynamic2
            if(passive)
                params.push_back((string)"1"+irc_get_mode(channel,nick)); // $passive
            else
                params.push_back("0"); // $passive

            BEGIN_DEBUG_TIME_MEASUREMENT(true,1,"#1");
            logic_core_execute(proc,params,"on_join",params.size(),"on_join@user");
            END_DEBUG_TIME_MEASUREMENT(1,0);

            break;
        }
    }

    if(!got_user) {
        // try host_unknown
        deque<c_user>::iterator i;
        for(i=r_users.begin(); i!=r_users.end(); ++i) {
            if((*i).host_unknown) {
                got_user=true;
                break;
            }
        }
        if(!got_user)
            goto end_func;

        vector<c_channel>::iterator i1;
        for(i1=r_channels.begin(); i1!=r_channels.end(); ++i1) {
            if(!(*i1).username.compare((*i).name) && !irc_cmp_strings_case_insensitive((*i1).channel_name,channel)) {

                string proc=(*i1).on_join;
                if(!proc.compare(""))
                    break;

                vector<string> params;
                params.push_back(channel); // $channel
                params.push_back(user); // $user
                params.push_back(nick); // $nick
                params.push_back((*i1).dynamic_plus_modes); // $dynamic1
                params.push_back((*i1).dynamic_minus_modes); // $dynamic2
                if(passive)
                    params.push_back((string)"1"+irc_get_mode(channel,nick)); // $passive
                else
                    params.push_back("0"); // $passive

                BEGIN_DEBUG_TIME_MEASUREMENT(true,2,"#2");
                logic_core_execute(proc,params,"on_join",params.size(),"on_join@user");
                END_DEBUG_TIME_MEASUREMENT(2,0);

                break;
            }
        }
    }
end_func:
    if(!mod_disp_join_called) {
        mod_disp_join_called=true;

        string psv;
        if(passive)
            psv=(string)"1"+irc_get_mode(channel,nick);
        else
            psv="0";
        BEGIN_DEBUG_TIME_MEASUREMENT(true,1,"mod_dispatch()");
        mod_dispatch("join@channel",user,channel,"",false,0,"channel",channel.c_str(),"user",user.c_str(),"nick",nick.c_str(),"passive",psv.c_str(),NULL,"event_type","channel","nick",nick.c_str(),"origin","@irc@","ident",irc_get_ident(nick).c_str(),"host",irc_get_host(nick).c_str(),NULL);
        END_DEBUG_TIME_MEASUREMENT(1,0);
    }
}

/*
    \brief Evaluates, and validates (or accidentally calls specific event) MODE to channel
    \author VooDooMan
    \version 1
    \date 2004
    \param channel Name of channel
    \param prefix Contains '+' or '-' prefix
    \param mode Mode char
    \param class_ 'A', 'B', 'C', 'D' (as in CHANMODES in 005 num reply), or 'x' for user mode
    \param user_prefix If class_ == 'x', there is user prefix ('@' for op, '+' for voice, etc.)
    \param nick Nick of originator of MODE change
    \param ident Ident of originator of MODE change
    \param host Host of originator of MODE change
    \param param Parameter
*]
void logic_mode_change_ex(string UNUSEDARG(channel), char UNUSEDARG(prefix), char UNUSEDARG(mode), char UNUSEDARG(class_), char UNUSEDARG(user_prefix), string UNUSEDARG(nick), string UNUSEDARG(ident), string UNUSEDARG(host), string UNUSEDARG(param))
{
}
*/

/*!
    \brief Executes on_banned event (called when bot JOINs the channel and finds out that some user is banned by ban mask) also checks whether user is not +e excepted or +I invited, which overrides the ban mask
    \author VooDooMan
    \version 1
    \date 2004
    \param channel Channel name
    \param ban_mask Ban mask
    \param excepts Array with list of excepts (+e)
    \param invites Array with list of invites (+I)
    \param bans Reference to always-up-to-date ban list
*/
void logic_on_banned(const string& channel, const string& ban_mask, const vector<string>& excepts, const vector<string>& invites, const vector<string>& bans)
{
    bool got_user=false;

    vector<c_channel>::iterator i1;
    for(i1=r_channels.begin(); i1!=r_channels.end(); ++i1) {
        if(irc_cmp_strings_case_insensitive((*i1).channel_name,channel))
            continue;
        deque<c_user>::iterator i3;
        for(i3=r_users.begin(); i3!=r_users.end(); ++i3) {
            if((*i3).name.compare((*i1).username))
                continue;
            vector<string>::iterator i2;
            for(i2=(*i3).hostmask.begin(); i2!=(*i3).hostmask.end(); ++i2) {
                bool was_dns=true;
                if(!(*i3).host_unknown && logic_check_mask(ban_mask,*i2,"*","*",was_dns) && !irc_cmp_strings_case_insensitive((*i1).channel_name,channel)) {
                    vector<string>::const_iterator i4;
                    bool in_except=false;
                    for(i4=excepts.begin(); i4!=excepts.end(); ++i4)
                        if(logic_check_mask(*i4,*i2,"*","*",was_dns))
                            in_except=true;
                    for(i4=invites.begin(); i4!=invites.end(); ++i4)
                        if(logic_check_mask(*i4,*i2,"*","*",was_dns))
                            in_except=true;
                    if(!in_except) {
                        got_user=true;
                        string proc=(*i1).on_banned;
                        if(!proc.compare(""))
                            continue; // do NOT break; coz we need to track EVERY user!

                        vector<string>::const_iterator i5;
                        bool got=false;
                        for(i5=bans.begin(); i5!=bans.end(); ++i5)
                            if(!(*i5).compare(ban_mask)) {
                                got=true;
                                break;
                            }
                        if(!got)
                            continue;

                        vector<string> params;
                        params.push_back(channel); // $channel
                        params.push_back((*i3).name); // $user
                        params.push_back(ban_mask); // $ban_mask

                        logic_core_execute(proc,params,"on_banned",params.size(),"on_banned@user");
                        // do NOT break; coz we need to track EVERY user!
                    }
                }
            }
        }
    }

    if(!got_user) {
        // try host_unknown
        vector<c_channel>::iterator i1;
        for(i1=r_channels.begin(); i1!=r_channels.end(); ++i1) {
            deque<c_user>::iterator i3;
            for(i3=r_users.begin(); i3!=r_users.end(); ++i3) {
                if(!(*i1).username.compare((*i3).name) && (*i3).host_unknown && !irc_cmp_strings_case_insensitive((*i1).channel_name,channel)) {
                    string proc=(*i1).on_banned;
                    if(!proc.compare(""))
                        continue;

                    vector<string>::const_iterator i5;
                    bool got=false;
                    for(i5=bans.begin(); i5!=bans.end(); ++i5)
                        if(!(*i5).compare(ban_mask)) {
                            got=true;
                            break;
                        }
                    if(!got)
                        continue;

                    vector<string> params;
                    params.push_back(channel); // $channel
                    params.push_back((*i3).name); // $user
                    params.push_back(ban_mask); // $ban_mask

                    logic_core_execute(proc,params,"on_banned",params.size(),"on_banned@user");
                }
            }
        }
    }
}

/*!
    \brief Executes on_not_invited event (called when bot JOINs the channel and finds out that some user is not invited by +I)
    \author VooDooMan
    \version 1
    \date 2004
    \param channel Channel name
    \param mask Host mask
    \param user_name Name of an user as in logic.txt
*/
void logic_on_not_invited(const string& channel, const string& mask, const string& user_name)
{
    bool got_user=false;

    vector<c_channel>::iterator i1;
    for(i1=r_channels.begin(); i1!=r_channels.end(); ++i1) {
        deque<c_user>::iterator i3;
        if(!irc_cmp_strings_case_insensitive((*i1).channel_name,channel)) {
            if(!(*i1).username.compare(user_name)) {
                got_user=true;

                vector<string>::iterator i2;

                string proc=(*i1).on_not_invited;
                if(!proc.compare(""))
                    break;

                vector<string> params;
                params.push_back(channel); // $channel
                params.push_back(user_name); // $user
                params.push_back(mask); // $mask

                logic_core_execute(proc,params,"on_not_invited",params.size(),"on_not_invited@user");
                break;
            }
        }
    }

    if(!got_user) {
        // try host_unknown
        vector<c_channel>::iterator i1;
        for(i1=r_channels.begin(); i1!=r_channels.end(); ++i1) {
            deque<c_user>::iterator i3;
            for(i3=r_users.begin(); i3!=r_users.end(); ++i3) {
                if(!(*i1).username.compare((*i3).name) && (*i3).host_unknown && !irc_cmp_strings_case_insensitive((*i1).channel_name,channel)) {
                    vector<string>::iterator i2;

                    string proc=(*i1).on_not_invited;
                    if(!proc.compare(""))
                        continue; // do NOT break; coz we need to track EVERY user!

                    vector<string> params;
                    params.push_back(channel); // $channel
                    params.push_back((*i1).username); // $user
                    params.push_back(mask); // $mask

                    logic_core_execute(proc,params,"on_not_invited",params.size(),"on_not_invited@user");
                }
            }
        }
    }
}

/*!
    \brief Executes on_not_in_reop event (called when bot JOINs the channel and finds out that some user is not in reop list by +R)
    \author VooDooMan
    \version 1
    \date 2004
    \param channel Channel name
    \param mask Host mask
    \param user_name Name of an user as in logic.txt
*/
void logic_on_not_in_reop(const string& channel, const string& mask, const string& user_name)
{
    bool got_user=false;

    vector<c_channel>::iterator i1;
    for(i1=r_channels.begin(); i1!=r_channels.end(); ++i1) {
        deque<c_user>::iterator i3;
        if(!irc_cmp_strings_case_insensitive((*i1).channel_name,channel)) {
            if(!(*i1).username.compare(user_name)) {
                got_user=true;

                vector<string>::iterator i2;

                string proc=(*i1).on_not_in_reop;
                if(!proc.compare(""))
                    break;

                vector<string> params;
                params.push_back(channel); // $channel
                params.push_back(user_name); // $user
                params.push_back(mask); // $mask

                logic_core_execute(proc,params,"on_not_in_reop",params.size(),"on_not_in_reop@user");
                break;
            }
        }
    }

    if(!got_user) {
        // try host_unknown
        vector<c_channel>::iterator i1;
        for(i1=r_channels.begin(); i1!=r_channels.end(); ++i1) {
            deque<c_user>::iterator i3;
            for(i3=r_users.begin(); i3!=r_users.end(); ++i3) {
                if(!(*i1).username.compare((*i3).name) && (*i3).host_unknown && !irc_cmp_strings_case_insensitive((*i1).channel_name,channel)) {
                    vector<string>::iterator i2;

                    string proc=(*i1).on_not_in_reop;
                    if(!proc.compare(""))
                        continue; // do NOT break; coz we need to track EVERY user!

                    vector<string> params;
                    params.push_back(channel); // $channel
                    params.push_back((*i1).username); // $user
                    params.push_back(mask); // $mask

                    logic_core_execute(proc,params,"on_not_in_reop",params.size(),"on_not_in_reop@user");
                }
            }
        }
    }
}

/*!
    \brief Validates KICK and, if needed, executes on_kick event
    \author VooDooMan
    \version 1
    \date 2004
    \param channel Name of channel
    \param source_user Name of user (originator) as in logic.txt
    \param target_user Name of user (victim) as in logic.txt
    \param source_nick Nick name of originator
    \param target_nick Nick name of victim
*/
void logic_validate_kick(const string& channel, const string& source_user, const string& target_user, const string& source_nick, const string& target_nick)
{
    if(!source_nick.compare(irc_nick) && target_nick.compare(irc_nick)) {
        // only if I AM killing, and only if I am not killing myself
    } else {
        deque<c_user>::iterator host_bot;
        bool b_host_bot=false;

        bool got_user=false;

        vector<c_channel>::iterator i1;

        {
            for(host_bot=r_users.begin(); host_bot!=r_users.end(); ++host_bot) {
                if((*host_bot).host_bot && !(*host_bot).name.compare(target_user)) {
                    b_host_bot=true;
                    goto l1;
                }
            }
        }

        for(i1=r_channels.begin(); i1!=r_channels.end(); ++i1) {
            if(!(*i1).username.compare(target_user) && !irc_cmp_strings_case_insensitive((*i1).channel_name,channel)) {
                    string proc=(*i1).on_kick;
                    if(!proc.compare(""))
                        continue;

                    vector<string> params;
                    params.push_back(channel); // $channel
                    params.push_back(source_user); // $source
                    params.push_back(target_user); // $target
                    params.push_back(source_nick); // $source_nick
                    params.push_back(target_nick); // $target_nick

                    logic_core_execute(proc,params,"on_kick",params.size(),"on_kick@user");
                    got_user=true;
                    break;
            }
        }

        if(!got_user) {
            // try host_unknown
        l1:
            deque<c_user>::iterator i;
            if(b_host_bot)
                i=host_bot;
                else {
                    for(i=r_users.begin(); i!=r_users.end(); ++i) {
                        if((*i).host_unknown) {
                            got_user=true;
                            break;
                        }
                    }
                    if(!got_user)
                        return;
                }

            vector<c_channel>::iterator i1;
            for(i1=r_channels.begin(); i1!=r_channels.end(); ++i1) {
                if(irc_cmp_strings_case_insensitive((*i1).channel_name,channel))
                    continue;
                vector<string>::iterator is, ii;
                if(!(*i1).username.compare((*i).name)) {
                        string proc=(*i1).on_kick;
                        if(!proc.compare(""))
                            continue;

                        vector<string> params;
                        params.push_back(channel); // $channel
                        params.push_back(source_user); // $source
                        params.push_back(target_user); // $target
                        params.push_back(source_nick); // $source_nick
                        params.push_back(target_nick); // $target_nick

                        logic_core_execute(proc,params,"on_kick",params.size(),"on_kick@user");
                        break;
                }
            }
        }
    }

    mod_dispatch("kick@channel",source_user,channel,"",false,0,"channel",channel.c_str(),"source",source_user.c_str(),"target",target_user.c_str(),"source_nick",source_nick.c_str(),"target_nick",target_nick.c_str(),"target_ident",irc_get_ident(target_nick).c_str(),"target_host",irc_get_host(target_nick).c_str(),NULL,"event_type","channel","nick",source_nick.c_str(),"origin","@irc@","ident",irc_get_ident(source_nick).c_str(),"host",irc_get_host(source_nick).c_str(),NULL);
}

/*!
    \brief Executes on_flood event procedure
    \author VooDooMan
    \version 1
    \date 2004
    \param channel Channel name
    \param user User name as define in logic.txt
    \param nick Nick name
    \param type Indicates type of flood - one of following: TYPE_PRIVMSG, TYPE_NOTICE, TYPE_NICK, TYPE_MODE, TYPE_JOIN, TYPE_REPEAT
    \param stat_num Number of occurency
    \param stat_sec Number of seconds that flood took
    \param stat_sec_of Maximum number of seconds (defined as flood)
    \param flist Listing of flood messages
*/
void logic_on_flood(const string& channel, const string& user, const string& nick, const int type, const int stat_num, const my_time_t stat_sec, const my_time_t stat_sec_of, const vector<string>& flist)
{
    bool got_user=false;

    vector<c_channel>::iterator i1;

    if(!user.compare((string)""))
        goto unknown;

    for(i1=r_channels.begin(); i1!=r_channels.end(); ++i1) {
        if(!(*i1).username.compare(user) && !irc_cmp_strings_case_insensitive((*i1).channel_name,channel)) {
            got_user=true;

            string proc=(*i1).on_flood;
            if(!proc.compare(""))
                break;

            c_module* core=get_module("core",true);

            if(core)
                var_keeper->set_current_session(core->get_session());

            c_temp_local local;
            c_var_array* vars=NULL;
            /*unsigned new_local_id=*/local.get_params(vars);

            vector<c_var_base*> params;
            params.push_back(MY_NEW c_var_string(channel)); // $channel
            params.push_back(MY_NEW c_var_string(user)); // $source
            params.push_back(MY_NEW c_var_string(nick)); // $source_nick
            switch(type) {
                case TYPE_PRIVMSG:
                    params.push_back(MY_NEW c_var_string("message"));
                    break;
                case TYPE_NOTICE:
                    params.push_back(MY_NEW c_var_string("notice"));
                    break;
                case TYPE_NICK:
                    params.push_back(MY_NEW c_var_string("nick"));
                    break;
                case TYPE_MODE:
                    params.push_back(MY_NEW c_var_string("mode"));
                    break;
                case TYPE_JOIN:
                    params.push_back(MY_NEW c_var_string("join"));
                    break;
                case TYPE_REPEAT:
                    params.push_back(MY_NEW c_var_string("repeat"));
                    break;
                case TYPE_CTCP:
                    params.push_back(MY_NEW c_var_string("ctcp"));
                    break;
                default:
                    params.push_back(MY_NEW c_var_string("other"));
                    break;
            }

            params.push_back(MY_NEW c_var_string(tos(stat_num)));
            params.push_back(MY_NEW c_var_string(tos(static_cast<unsigned long>(stat_sec))));
            params.push_back(MY_NEW c_var_string(tos(static_cast<unsigned long>(stat_sec_of))));

            if(core) {
                c_var_array* lst=MY_NEW c_var_array();
                vector<string>::const_iterator i2;
                for(i2=flist.begin(); i2!=flist.end(); ++i2) {
                    c_index_base* idx=lst->get_new_int_index();
                    lst->set(core,idx,MY_NEW c_var_string(*i2),false,"on_flood");
                }
                params.push_back(lst);
            }

            logic_core_execute(proc,params,"on_flood",params.size(),"on_flood@user");

            var_keeper->set_current_session(SESSION_RESERVED_ID);
            break;
        }
    }

    if(!got_user) {
        // try host_unknown
    unknown:
        vector<c_channel>::iterator i1;
        for(i1=r_channels.begin(); i1!=r_channels.end(); ++i1) {
            if((*i1).host_unknown && !irc_cmp_strings_case_insensitive((*i1).channel_name,channel)) {
                string proc=(*i1).on_flood;
                if(!proc.compare(""))
                    break;

                c_module* core=get_module("core",true);

                if(core)
                    var_keeper->set_current_session(core->get_session());

                c_temp_local local;
                c_var_array* vars=NULL;
                /*unsigned new_local_id=*/local.get_params(vars);

                vector<c_var_base*> params;
                params.push_back(MY_NEW c_var_string(channel)); // $channel
                params.push_back(MY_NEW c_var_string(user)); // $source
                params.push_back(MY_NEW c_var_string(nick)); // $source_nick
                switch(type) {
                    case TYPE_PRIVMSG:
                        params.push_back(MY_NEW c_var_string("message"));
                        break;
                    case TYPE_NOTICE:
                        params.push_back(MY_NEW c_var_string("notice"));
                        break;
                    case TYPE_NICK:
                        params.push_back(MY_NEW c_var_string("nick"));
                        break;
                    case TYPE_MODE:
                        params.push_back(MY_NEW c_var_string("mode"));
                        break;
                    case TYPE_JOIN:
                        params.push_back(MY_NEW c_var_string("join"));
                        break;
                    case TYPE_REPEAT:
                        params.push_back(MY_NEW c_var_string("repeat"));
                        break;
                    case TYPE_CTCP:
                        params.push_back(MY_NEW c_var_string("ctcp"));
                        break;
                    default:
                        params.push_back(MY_NEW c_var_string("other"));
                        break;
                }

                params.push_back(MY_NEW c_var_string(tos(stat_num)));
                params.push_back(MY_NEW c_var_string(tos(static_cast<unsigned long>(stat_sec))));
                params.push_back(MY_NEW c_var_string(tos(static_cast<unsigned long>(stat_sec_of))));

                if(core) {
                    c_var_array* lst=MY_NEW c_var_array();
                    vector<string>::const_iterator i2;
                    for(i2=flist.begin(); i2!=flist.end(); ++i2) {
                        c_index_base* idx=lst->get_new_int_index();
                        lst->set(core,idx,MY_NEW c_var_string(*i2),false,"on_flood");
                    }
                    params.push_back(lst);
                }

                logic_core_execute(proc,params,"on_flood",params.size(),"on_flood@user");

                var_keeper->set_current_session(SESSION_RESERVED_ID);
                break;
            }
        }
    }
    string type_str;
    switch(type) {
        case TYPE_PRIVMSG:
            type_str="message";
            break;
        case TYPE_NOTICE:
            type_str="notice";
            break;
        case TYPE_NICK:
            type_str="nick";
            break;
        case TYPE_MODE:
            type_str="mode";
            break;
        case TYPE_JOIN:
            type_str="join";
            break;
        case TYPE_REPEAT:
            type_str="repeat";
            break;
        case TYPE_CTCP:
            type_str="ctcp";
            break;
        default:
            type_str="other";
            break;
    }
    string num=tos(stat_num);
    string sec=tos(static_cast<unsigned long>(stat_sec));
    string sec_of=tos(static_cast<unsigned long>(stat_sec_of));
    c_var_array* meta=MY_NEW c_var_array();

    {
        c_index_base* index=MY_NEW c_index_str("channel");
        c_var_string* value=MY_NEW c_var_string(channel);
        meta->set(NULL,index,value,false,"...");
    }
    {
        c_index_base* index=MY_NEW c_index_str("source");
        c_var_string* value=MY_NEW c_var_string(user);
        meta->set(NULL,index,value,false,"...");
    }
    {
        c_index_base* index=MY_NEW c_index_str("source_nick");
        c_var_string* value=MY_NEW c_var_string(nick);
        meta->set(NULL,index,value,false,"...");
    }
    {
        c_index_base* index=MY_NEW c_index_str("type");
        c_var_string* value=MY_NEW c_var_string(type_str);
        meta->set(NULL,index,value,false,"...");
    }
    {
        c_index_base* index=MY_NEW c_index_str("num");
        c_var_string* value=MY_NEW c_var_string(num);
        meta->set(NULL,index,value,false,"...");
    }
    {
        c_index_base* index=MY_NEW c_index_str("sec");
        c_var_string* value=MY_NEW c_var_string(sec);
        meta->set(NULL,index,value,false,"...");
    }
    {
        c_index_base* index=MY_NEW c_index_str("sec_of");
        c_var_string* value=MY_NEW c_var_string(sec_of);
        meta->set(NULL,index,value,false,"...");
    }
    c_module* core=get_module("core",true);
    if(core) {
        c_var_array* lst=MY_NEW c_var_array();
        vector<string>::const_iterator i2;
        for(i2=flist.begin(); i2!=flist.end(); ++i2) {
            c_index_base* idx=lst->get_new_int_index();
            lst->set(core,idx,MY_NEW c_var_string(*i2),false,"on_flood");
        }

        c_index_base* index=MY_NEW c_index_str("list");
        meta->set(NULL,index,lst,false,"...");
    }
    mod_dispatch2("flood@channel",user,channel,"",false,meta,0,NULL,"event_type","channel","nick",nick.c_str(),"origin","@irc@","ident",irc_get_ident(nick).c_str(),"host",irc_get_host(nick).c_str(),NULL);
    MY_DELETE(meta);
}

/*!
    \brief Handles CTCP messages
    \author VooDooMan
    \version 1
    \date 2004
    \param user User name as defined in logic.txt
    \param nick Nick name
    \param channel Name of channel, or empty for private query
    \param msg Message (PRIVMSG parameter)
*/
void logic_ctcp(string user, string nick, string channel, string msg)
{
    string command=msg;
    string parameters;
    if(command.find(" ",0)!=string::npos) {
        command.erase(command.find(" ",0),command.length()-command.find(" ",0));
        parameters=msg;
        parameters.erase(0,parameters.find(" ",0)+1);
    }
    for(;;) {
        {
            for(size_t i1=0; i1<msg.length() && msg[i1]!=0x20; ++i1)
                if(msg[i1]>='a' && msg[i1]<='z')
                    msg[i1]-='a'-'A';
        }
        if(msg.find("DCC ")!=0 && msg.find("ACTION")!=0) {
            // DCC flood protection is handled separately elsewhere
            vector<c_nick_to_lag>::iterator i1;
    again:
            for(i1=nicks_to_lag.begin(); i1!=nicks_to_lag.end(); ++i1) {
                if((*i1).last_try+sto_time(core_conf_getvar("ctcp_lag_user_seconds"))<my_time(NULL)) {
                    nicks_to_lag.erase(i1);
                    goto again;
                }
            }
            bool got=false;
            for(i1=nicks_to_lag.begin(); i1!=nicks_to_lag.end(); ++i1) {
                if(!(*i1).nick.compare(nick)) {
                    (*i1).cnt++;
                    (*i1).last_try=my_time(NULL);
                    got=true;
                    break;
                }
            }
            if(got) {
                if((*i1).cnt>stoi(core_conf_getvar("ctcp_lag_user_msgs"))) {
                    my_time_t x=(*i1).last_try;
                    my_time_t now=my_time(NULL);
                    while(x>0 && (*i1).cnt>0 && my_difftime(now,x)>sto_time(core_conf_getvar("ctcp_lag_user_seconds"))) {
                        x-=sto_time(core_conf_getvar("ctcp_lag_user_seconds"));
                        (*i1).cnt--;
                    }
                    if((*i1).cnt>stoi(core_conf_getvar("ctcp_lag_user_msgs"))) {
                        break;
                    }
                    if((*i1).cnt==0) {
                        (*i1).cnt=1;
                    }
                }
            } else {
                c_nick_to_lag u;
                u.cnt=1;
                u.last_try=my_time(NULL);
                u.nick=nick;
                nicks_to_lag.push_back(u);
            }
        }

        if(msg.find("PING")==0 && (msg.length()==4 || (msg.length()>4 && msg[4]==0x20))) {
            if(msg.length()>64) {
                // PING of death - ignoring
            } else {
                string orig=msg;
                string m="PING";
                msg.erase(0,m.length());
                m=":\x01";
                m+="PING";
                m+=msg;
                m+="\x01";
                irc_putserv(c_tracked_msg_to_server("NOTICE").add_nick(nick).add_piece(m),false,LOW_PRIORITY);
                msg=orig;
            }
        }
        if(msg.find("TIME")==0) {
            string m=":\x01";
            m+="TIME ";
            time_t now=static_cast<time_t>(my_time(NULL));
            m+=ctime(&now);
            if(m.find('\n',0)!=string::npos)
                m.erase(m.find('\n',0),1);
            m+="\x01";
            irc_putserv(c_tracked_msg_to_server("NOTICE").add_nick(nick).add_piece(m),false,LOW_PRIORITY);
        }
        // other CTCP
        {
            string proc=r_private.on_ctcp;
            if(!proc.compare(""))
                break;

            vector<string> params;
            params.push_back(user); // $source
            params.push_back(nick); // $source_nick
            params.push_back(channel); // $channel
            params.push_back(command); // $command
            params.push_back(parameters); // $parameters

            logic_core_execute(proc,params,"on_ctcp",params.size(),"on_ctcp@private");
        }

        vector<c_channel_def>::iterator i1;
        for(i1=r_channel_defs.begin(); i1!=r_channel_defs.end(); ++i1) {
            if(!irc_cmp_strings_case_insensitive((*i1).channel_name,channel)) {

                string proc=(*i1).on_ctcp;
                if(proc.empty())
                    break;

                vector<string> params;
                params.push_back(user); // $source
                params.push_back(nick); // $source_nick
                params.push_back(channel); // $channel
                params.push_back(command); // $command
                params.push_back(parameters); // $parameters

                logic_core_execute(proc,params,"on_ctcp",params.size(),"on_ctcp@channel");
                break;
            }
        }
        break;
    }

    if(!channel.empty())
        mod_dispatch("ctcp@channel",user,channel,"",false,0,"channel",channel.c_str(),"source",user.c_str(),"source_nick",nick.c_str(),"command",command.c_str(),"parameters",parameters.c_str(),NULL,"event_type","channel","nick",nick.c_str(),"origin","@irc@","ident",irc_get_ident(nick).c_str(),"host",irc_get_host(nick).c_str(),NULL);
    else
        mod_dispatch("ctcp@private",user,channel,"",false,0,"source",user.c_str(),"source_nick",nick.c_str(),"command",command.c_str(),"parameters",parameters.c_str(),NULL,"event_type","private","nick",nick.c_str(),"origin","@irc@","ident",irc_get_ident(nick).c_str(),"host",irc_get_host(nick).c_str(),NULL);
}

/*!
    \brief Executes on_filesys_got_new event
    \author VooDooMan
    \version 1
    \date 2005
    \param user_name User name as in logic.txt
    \param nick Nick name
    \param ident Ident
    \param host Host
    \param internal_name Internal name of file
*/
void logic_filesys_got_new(const string& user_name, const string& nick, const string& ident, const string& host, const string& internal_name)
{
    string proc=r_private.on_filesys_got_new;
    if(!proc.compare(""))
        return;

    vector<string> params;
    params.push_back(user_name); // $user_name
    params.push_back(nick); // $nick
    params.push_back(ident); // $ident
    params.push_back(host); // $host
    params.push_back(internal_name); // $internal_name

    logic_core_execute(proc,params,"on_filesys_got_new",params.size(),"on_filesys_got_new@private");

    mod_dispatch("got_new@filesys",user_name,"","",false,0,"user_name",user_name.c_str(),"nick",nick.c_str(),"ident",ident.c_str(),"host",host.c_str(),"internal_name",internal_name.c_str(),NULL,"event_type","filesys","origin","@filesys@",NULL);
}

/*!
    \brief Executes on_fnc event
    \author VooDooMan
    \version 1
    \date 2005
    \param old_nick Old nick name
    \param new_nick New nick name
*/
void logic_fnc(const string& old_nick, const string& new_nick)
{
    string proc=r_private.on_fnc;
    if(!proc.compare(""))
        return;

    vector<string> params;
    params.push_back(old_nick); // $old_nick
    params.push_back(new_nick); // $new_nick

    logic_core_execute(proc,params,"on_fnc",params.size(),"on_fnc@private");

    mod_dispatch("fnc@private","","","",false,0,"old_nick",old_nick.c_str(),"new_nick",new_nick.c_str(),NULL,"event_type","private","origin","@fnc@",NULL);
}

/*!
    \brief Executes on_privmsg event procedure
    \author VooDooMan
    \version 1
    \date 2004
    \param channel Channel name
    \param user User name as define in logic.txt
    \param nick Nick name
    \param msg Message (PRIVMSG parameter)
*/
void logic_on_privmsg(const string& channel, string user, const string& nick, string msg)
{
    bool got_user=false;

    vector<c_channel>::iterator i1;

    if(!user.compare((string)""))
        goto unknown;

    for(i1=r_channels.begin(); i1!=r_channels.end(); ++i1) {
        if(!(*i1).username.compare(user) && !irc_cmp_strings_case_insensitive((*i1).channel_name,channel)) {
            got_user=true;
            {
                string proc=(*i1).on_privmsg;
                if(!proc.compare(""))
                    break;

                vector<string> params;
                params.push_back(channel); // $channel
                params.push_back(user); // $source
                params.push_back(nick); // $source_nick
                params.push_back(msg); // $msg

                logic_core_execute(proc,params,"on_privmsg",params.size(),"on_privmsg@user");
                break;
            }
        }
    }

    if(!got_user) {
        // try host_unknown
    unknown:

        {
            deque<c_user>::iterator i1;
            for(i1=r_users.begin(); i1!=r_users.end(); ++i1)
                if((*i1).host_unknown) {
                    user=(*i1).name;
                    break;
                }
        }

        vector<c_channel>::iterator i1;
        for(i1=r_channels.begin(); i1!=r_channels.end(); ++i1) {
            if((*i1).host_unknown && !irc_cmp_strings_case_insensitive((*i1).channel_name,channel)) {
                {
                    string proc=(*i1).on_privmsg;
                    if(!proc.compare(""))
                        break;

                    vector<string> params;
                    params.push_back(channel); // $channel
                    params.push_back(user); // $source
                    params.push_back(nick); // $source_nick
                    params.push_back(msg); // $msg

                    logic_core_execute(proc,params,"on_privmsg",params.size(),"on_privmsg@user");
                    break;
                }
            }
        }
    }

    vector<c_channel_def>::iterator i2;
    for(i2=r_channel_defs.begin(); i2!=r_channel_defs.end(); ++i2) {
        if(!irc_cmp_strings_case_insensitive(channel,(*i2).channel_name)) {
            string proc=(*i2).on_privmsg;
            if(!proc.compare(""))
                break;
            vector<string> params;
            params.push_back(channel); // $channel
            params.push_back(user); // $source
            params.push_back(nick); // $source_nick
            params.push_back(msg); // $msg

            logic_core_execute(proc,params,"on_privmsg",params.size(),"on_privmsg@channel");
            break;
        }
    }

    string command=msg;
    if(command.find(" ",0)!=string::npos)
        command.erase(command.find(" ",0),command.length()-command.find(" ",0));

    mod_dispatch("privmsg@channel",user,channel,msg,false,0,"channel",channel.c_str(),"source",user.c_str(),"source_nick",nick.c_str(),"msg",msg.c_str(),NULL,"event_type","channel","nick",nick.c_str(),"origin","@irc@","ident",irc_get_ident(nick).c_str(),"host",irc_get_host(nick).c_str(),NULL);

    if(msg.find("\001",0)!=string::npos) {
        if(msg.find("\x01",0)!=string::npos) {
            msg.erase(msg.find("\x01",0),1);
            if(msg.find("\x01",0)!=string::npos)
                msg.erase(msg.find("\x01",0),1);
        }
        logic_ctcp(user,nick,channel,msg);
    }
}

/*!
    \brief Executes on_notice event procedure
    \author VooDooMan
    \version 1
    \date 2004
    \param channel Channel name, or an empty string for query/private notice
    \param user User name as define in logic.txt
    \param nick Nick name
    \param msg Message (NOTICE parameter)
*/
void logic_on_notice(const string& channel, string user, const string& nick, const string& msg)
{
    bool got_user=false;

    vector<c_channel>::iterator i1;

    if(!user.compare((string)""))
        goto unknown;

    if(!channel.empty()) {
        for(i1=r_channels.begin(); i1!=r_channels.end(); ++i1) {
            if(!(*i1).username.compare(user) && !irc_cmp_strings_case_insensitive((*i1).channel_name,channel)) {
                got_user=true;
                {
                    string proc=(*i1).on_notice;
                    if(!proc.compare(""))
                        break;

                    vector<string> params;
                    params.push_back(channel); // $channel
                    params.push_back(user); // $source
                    params.push_back(nick); // $source_nick
                    params.push_back(msg); // $msg

                    logic_core_execute(proc,params,"on_notice",params.size(),"on_notice@user");
                    break;
                }
            }
        }

        if(!got_user) {
            // try host_unknown
        unknown:

            {
                deque<c_user>::iterator i1;
                for(i1=r_users.begin(); i1!=r_users.end(); ++i1)
                    if((*i1).host_unknown) {
                        user=(*i1).name;
                        break;
                    }
            }

            vector<c_channel>::iterator i1;
            for(i1=r_channels.begin(); i1!=r_channels.end(); ++i1) {
                if((*i1).host_unknown && !irc_cmp_strings_case_insensitive((*i1).channel_name,channel)) {
                    string proc=(*i1).on_notice;
                    if(!proc.compare(""))
                        break;

                    vector<string> params;
                    params.push_back(channel); // $channel
                    params.push_back(user); // $source
                    params.push_back(nick); // $source_nick
                    params.push_back(msg); // $msg

                    logic_core_execute(proc,params,"on_notice",params.size(),"on_notice@user");
                    break;
                }
            }
        }

        vector<c_channel_def>::iterator i2;
        for(i2=r_channel_defs.begin(); i2!=r_channel_defs.end(); ++i2)
            if(!irc_cmp_strings_case_insensitive(channel,(*i2).channel_name)) {
                string proc=(*i2).on_notice;
                if(!proc.compare(""))
                    break;
                vector<string> params;
                params.push_back(channel); // $channel
                params.push_back(user); // $source
                params.push_back(nick); // $source_nick
                params.push_back(msg); // $msg

                logic_core_execute(proc,params,"on_notice",params.size(),"on_notice@channel");
                break;
            }

        if(channel.empty())
            goto private_query;

    } else {
    private_query:
        string proc=r_private.on_notice;
        if(!proc.compare(""))
            return;

        vector<string> params;
        params.push_back(user); // $source
        params.push_back(nick); // $source_nick
        params.push_back(msg); // $msg

        logic_core_execute(proc,params,"on_notice",params.size(),"on_notice@private");
    }

    string command=msg;
    if(command.find(" ",0)!=string::npos)
        command.erase(command.find(" ",0),command.length()-command.find(" ",0));

    if(!channel.empty())
        mod_dispatch("notice@channel",user,channel,msg,false,0,"channel",channel.c_str(),"source",user.c_str(),"source_nick",nick.c_str(),"msg",msg.c_str(),NULL,"event_type","channel","nick",nick.c_str(),"origin","@irc@","ident",irc_get_ident(nick).c_str(),"host",irc_get_host(nick).c_str(),NULL);
    else
        mod_dispatch("notice@private",user,channel,msg,false,0,"source",user.c_str(),"source_nick",nick.c_str(),"msg",msg.c_str(),NULL,"event_type","private","nick",nick.c_str(),"origin","@irc@","ident",irc_get_ident(nick).c_str(),"host",irc_get_host(nick).c_str(),NULL);

    // commented out 2005-04-22 by Marian VooDooMan Meravy
    // 1) NOTICE's are NOT CTCP requests, but responses, so don't process them here!
    // 2) If we would respond to NOTICE, we would violate RFC saying that there
    //    MUST NOT be automatic response to NOTICE!
    /*
    if(msg.find("\001",0)!=string::npos) {
        if(msg.find("\x01",0)!=string::npos) {
            msg.erase(msg.find("\x01",0),1);
            if(msg.find("\x01",0)!=string::npos)
                msg.erase(msg.find("\x01",0),1);
        }
        logic_ctcp(user,nick,channel,msg);
    }
    */
}

/*!
    \brief Executes on_privmsg event procedure (for query only - not channel)
    \author VooDooMan
    \version 1
    \date 2004
    \param user User name as defined in logic.txt
    \param nick Nick name
    \param msg Message (PRIVMSG parameter)
*/
void logic_on_privmsg_query(string user, const string& nick, string msg)
{
    bool quoted=false;

    if(msg.find("\x01",0)!=string::npos)
        quoted=true;

    vector<c_channel>::iterator i1;

    if(user.empty()) {
        deque<c_user>::iterator i1;
        for(i1=r_users.begin(); i1!=r_users.end(); ++i1) {
            if((*i1).host_unknown) {
                user=(*i1).name;
                break;
            }
        }

        // we don't have "host_unknown" defined!
    }

    {
        string proc=r_private.on_privmsg;
        if(!proc.empty()) {
            vector<string> params;
            params.push_back(user); // $source
            params.push_back(nick); // $source_nick
            params.push_back(msg); // $msg

            logic_core_execute(proc,params,"on_privmsg",params.size(),"on_privmsg@private");
        }
    }

    mod_dispatch("privmsg@private",user,"",msg,false,0,"source",user.c_str(),"source_nick",nick.c_str(),"msg",msg.c_str(),NULL,"event_type","private","nick",nick.c_str(),"origin","@irc@","ident",irc_get_ident(nick).c_str(),"host",irc_get_host(nick).c_str(),NULL);

    if(quoted) {
        if(msg.find("\x01",0)!=string::npos) {
            msg.erase(msg.find("\x01",0),1);
            if(msg.find("\x01",0)!=string::npos)
                msg.erase(msg.find("\x01",0),1);
        }
        logic_ctcp(user,nick,"",msg);
    }
}

/*!
    \brief Executes procedure for event on_ircop (when an IRC operator joins / is on after bot's join / left channel)
    \author VooDooMan
    \version 1
    \date 2004
    \param channel Name of channel
    \param nick Nick name of IRC operator
    \param on Set to true if is on channel, to false if it has left the channel
*/
void logic_on_ircop(const string& channel, const string& nick, const bool on)
{
    string user_name=logic_find_user(nick,irc_get_ident(nick),irc_get_host(nick),irc_get_fullname(nick),irc_is_ircop(nick));

    {
        string host_mask=nick;
        host_mask+="!";
        host_mask+=irc_get_ident(nick);
        host_mask+="@";
        host_mask+=irc_get_host(nick);
        string _on=on==true?"ON":"OFF";
        logic_on_internal_event("@ircop@",nick,host_mask,channel,_on,PRIVATE_EVENT_SEVERITY_CODE_WARNING,"",irc_get_fullname(nick));
    }

    vector<c_channel_def>::iterator i1;
    for(i1=r_channel_defs.begin(); i1!=r_channel_defs.end(); ++i1) {
        if(!irc_cmp_strings_case_insensitive((*i1).channel_name,channel)) {
            string proc=(*i1).on_ircop;
            if(proc.empty())
                break;

            vector<string> params;
            params.push_back(channel); // $channel
            params.push_back(user_name); // $source
            params.push_back(nick); // $source_nick
            params.push_back(on?"1":"0"); // $on

            logic_core_execute(proc,params,"on_ircop",params.size(),"on_ircop@channel");
            break;
        }
    }

    char on_[2];
    on_[0]=on ? '1' : '0';
    on_[1]='\0';

    mod_dispatch("ircop@channel",user_name,channel,"",false,0,"channel",channel.c_str(),"source",user_name.c_str(),"source_nick",nick.c_str(),"on",on_,NULL,"event_type","channel","nick",nick.c_str(),"origin","@irc@","ident",irc_get_ident(nick).c_str(),"host",irc_get_host(nick).c_str(),NULL);
}

/*!
    \brief Executes on_part event procedure
    \author VooDooMan
    \version 1
    \date 2004
    \param channel Channel name
    \param nick Nick name
    \param msg Reason
    \param type TYPE_PART, TYPE_QUIT, or TYPE_KICK
*/
void logic_on_part(const string& channel, const string& nick, const string& msg, const int type)
{
    string user=logic_find_user(nick,irc_get_ident(nick),irc_get_host(nick),irc_get_fullname(nick),irc_is_ircop(nick));

    vector<c_channel_def>::iterator i2;
    for(i2=r_channel_defs.begin(); i2!=r_channel_defs.end(); ++i2)
        if(!irc_cmp_strings_case_insensitive(channel,(*i2).channel_name)) {
            string proc=(*i2).on_part;
            if(!proc.compare(""))
                break;
            vector<string> params;
            params.push_back(channel); // $channel
            params.push_back(user); // $source
            params.push_back(nick); // $source_nick
            params.push_back(msg); // $msg
            switch(type) {
                case TYPE_PART:
                    params.push_back("PART");
                    break;
                case TYPE_QUIT:
                    params.push_back("QUIT");
                    break;
                case TYPE_KICK:
                    params.push_back("KICK");
                    break;
                default:
                    params.push_back("");
                    break;
            }

            logic_core_execute(proc,params,"on_part",params.size(),"on_part@channel");
            break;
        }
    string t;
    switch(type) {
        case TYPE_PART:
            t="PART";
            break;
        case TYPE_QUIT:
            t="QUIT";
            break;
        case TYPE_KICK:
            t="KICK";
            break;
        default:
            t="";
            break;
    }

    mod_dispatch("part@channel",user,channel,msg,false,0,"channel",channel.c_str(),"source",user.c_str(),"source_nick",nick.c_str(),"msg",msg.c_str(),"type",t.c_str(),NULL,"event_type","channel","nick",nick.c_str(),"origin","@irc@","ident",irc_get_ident(nick).c_str(),"host",irc_get_host(nick).c_str(),NULL);
}

/*!
    \brief Raises on_broadcast event
    \author VooDooMan
    \version 1
    \date 2005-04-22
    \param username Name of user as in logic.txt, or an empty string for "host_unknown" user
    \param nick Nick name of originator
    \param ident Ident of originator
    \param host Host of originator
    \param bcast_mask Broadcast mask ("$$irc.someserver.net")
    \param type Type: TYPE_PRIVMSG, TYPE_NOTICE
    \param msg The message
    \param server IRC server host we are on
    \param port IRC server port we are on
*/
void logic_on_broadcast(string username, const string& nick,
                        const string& ident, const string& host,
                        const string& bcast_mask, const int type,
                        string msg, const string& server,
                        const unsigned short port)
{
    for(;;) {

        if(username.empty()) {
            deque<c_user>::iterator i1;
            for(i1=r_users.begin(); i1!=r_users.end(); ++i1)
                if((*i1).host_unknown) {
                    username=(*i1).name;
                    break;
                }
        }

        string proc=r_private.on_broadcast;
        if(!proc.compare(""))
            break;

        vector<string> params;
        params.push_back(username); // $source
        params.push_back(nick); // $source_nick
        params.push_back(ident); // $source_ident
        params.push_back(host); // $source_host
        params.push_back(bcast_mask); // $bcast_mask
        switch(type) {
            case TYPE_PRIVMSG:
                params.push_back("PRIVMSG"); // $type
                break;
            case TYPE_NOTICE:
                params.push_back("NOTICE"); // $type
                break;
            default:
                // should never occur, if even so, we have somewhere bug!
                params.push_back("?"); // $type
                break;
        }
        params.push_back(msg); // $msg
        params.push_back(server); // $server
        char tmp[512];
        sprintf(tmp,"%u",static_cast<unsigned>(port));
        params.push_back(tmp); // $port

        logic_core_execute(proc,params,"on_broadcast",params.size(),"on_broadcast@private");

        string t="?";
        switch(type) {
            case TYPE_PRIVMSG:
                t="PRIVMSG";
                break;
            case TYPE_NOTICE:
                t="NOTICE";
                break;
        }

        mod_dispatch("broadcast@private",username,"","",false,0,"source",username.c_str(),"source_nick",nick.c_str(),"ident",ident.c_str(),"host",host.c_str(),"bcast_mask",bcast_mask.c_str(),"type",t.c_str(),"msg",msg.c_str(),NULL,"event_type","private","nick",nick.c_str(),"origin","@irc@","ident",ident.c_str(),"host",host.c_str(),NULL);

        // only respond to PRIVMSG, NOTICE's are NOT CTCP requests, but rather responses!
        if(type==TYPE_PRIVMSG && msg.find("\001",0)!=string::npos) {
            if(msg.find("\x01",0)!=string::npos) {
                msg.erase(msg.find("\x01",0),1);
                if(msg.find("\x01",0)!=string::npos)
                    msg.erase(msg.find("\x01",0),1);
            }
            logic_ctcp(username,nick,"",msg);
        }
        break;
    }
}

/*!
    \brief Raises on_server_msg event
    \author VooDooMan
    \version 1
    \date 2005-04-25
    \param channel Name of channel, or an empty string for private message
    \param source Name of server
    \param type Type: TYPE_PRIVMSG, TYPE_NOTICE
    \param msg The message
    \param server IRC server host we are on
    \param port IRC server port we are on
*/
void logic_on_server_msg(const string& channel, const string& source, const int type, const string& msg, const string& server, const unsigned short port)
{
    for(;;) {

        string event;

        string proc;
        if(channel.empty()) {
            proc=r_private.on_server_msg;
            event="on_server_msg@private";
        } else {
            vector<c_channel_def>::iterator i2;
            for(i2=r_channel_defs.begin(); i2!=r_channel_defs.end(); ++i2)
                if(!irc_cmp_strings_case_insensitive(channel,(*i2).channel_name)) {
                    proc=(*i2).on_server_msg;
                    break;
                }
            event="on_server_msg@channel";
        }

        if(!proc.compare(""))
            break;

        vector<string> params;
        if(!channel.empty())
            params.push_back(channel); // $channel
        else
            params.push_back(""); // $channel
        params.push_back(source); // $source
        switch(type) {
            case TYPE_PRIVMSG:
                params.push_back("PRIVMSG"); // $type
                break;
            case TYPE_NOTICE:
                params.push_back("NOTICE"); // $type
                break;
            default:
                // should never occur, if even so, we have somewhere bug!
                params.push_back("?"); // $type
                break;
        }
        params.push_back(msg); // $msg
        params.push_back(server); // $server
        char tmp[512];
        sprintf(tmp,"%u",static_cast<unsigned>(port));
        params.push_back(tmp); // $port

        logic_core_execute(proc,params,"on_server_msg",params.size(),event);
        break;
    }

    string t="?";
    switch(type) {
        case TYPE_PRIVMSG:
            t="PRIVMSG";
            break;
        case TYPE_NOTICE:
            t="NOTICE";
            break;
    }
    if(!channel.empty())
        mod_dispatch("server_msg@channel","",channel,"",false,0,"channel",channel.c_str(),"source",source.c_str(),"msg",msg.c_str(),"type",t.c_str(),NULL,"event_type","channel","origin","@server@",NULL);
    else
        mod_dispatch("server_msg@private","","","",false,0,"source",source.c_str(),"msg",msg.c_str(),"type",t.c_str(),NULL,"event_type","private","origin","@server@",NULL);
}

/*!
    \brief Raises on_intenal_event event
    \author VooDooMan
    \version 1
    \date 2005-07-08
    \param type Type of event
    \param flags1 Flags #1
    \param flags2 Flags #2
    \param flags3 Flags #3
    \param flags4 Flags #4
    \param severity_numeric Number of severity
    \param msg1 Message #1
    \param msg2 Message #2
*/
void logic_on_internal_event(string type, string flags1, string flags2, string flags3, string flags4, const int severity_numeric, string msg1, string msg2)
{
    string proc=r_private.on_internal_event;

    if(proc.empty())
        return;

    static bool in_event=false;
    if(in_event)
        return;
    in_event=true;

    type=logic_kill_EOLs(type);
    flags1=logic_kill_EOLs(flags1);
    flags2=logic_kill_EOLs(flags2);
    flags3=logic_kill_EOLs(flags3);
    flags4=logic_kill_EOLs(flags4);
    msg1=logic_kill_EOLs(msg1);
    msg2=logic_kill_EOLs(msg2);

    vector<string> params;
    params.push_back(type); // $type

    my_time_t now=my_time(NULL);
    string time_string=make_nice_time(now,false,false);
    if(time_string.find("\n",0)!=string::npos)
        time_string.erase(time_string.find("\n",0),1);

    params.push_back(tos(static_cast<unsigned long>(now))); // $timestamp
    params.push_back(time_string); // $time_string
    params.push_back(flags1); // $flags1
    params.push_back(flags2); // $flags2
    params.push_back(flags3); // $flags3
    params.push_back(flags4); // $flags4
    params.push_back(tos(severity_numeric)); // $severity_numeric
    params.push_back(logic_get_severity_text(severity_numeric)); // $severity_string
    params.push_back(msg1); // $msg1
    params.push_back(msg2); // $msg2
    params.push_back(irc_server_host); // $server
    params.push_back(tos((unsigned)irc_server_port)); // $port

    session_id_t old_session=var_keeper->get_current_session();

    if(core_mod)
        var_keeper->internal_event_issued(core_mod->get_session(),true);
    logic_core_execute(proc,params,"on_internal_event",params.size(),"on_internal_event@private");
    if(core_mod)
        var_keeper->internal_event_issued(core_mod->get_session(),false);

    var_keeper->set_current_session(old_session);

    in_event=false;
}

