/***************************************************************************
                          match.cpp  -  description
                             -------------------
    begin                : Thu Dec 9 2004
    copyright            : (C) 1990 by Jarkko Oikarinen
    copyright            : (C) 2004-2007 by Marian VooDooMan Meravy
    email                : ghostvoodooman -a.T= users (dot) sourceforge.net
 ***************************************************************************/

/* $Id: match.cpp,v 1.53 2008/08/23 15:10:41 ghostvoodooman Exp $ */

/************************************************************************
 *   IRC - Internet Relay Chat, common/match.c
 *   Copyright (C) 1990 Jarkko Oikarinen
 *   Copyright (C) 2004-2008 by Marian VooDooMan Meravy
 *     (ghostvoodooman -a.T= users (dot) sourceforge.net)
 *     (re-coded from C to C++)
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 1, or (at your option)
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*!
    \file
    \brief Provides wildcard checks
    \version $Id: match.cpp,v 1.53 2008/08/23 15:10:41 ghostvoodooman Exp $
*/

#if defined(HAVE_CONFIG_H)
#   include <config.h>
#endif

#include "params.h"
#include "utils/memleak.h"
#include "utils/assert.h"
#include "compat.h"

#include "sock_shared.h"

#include <string>
#include <vector>
#include <stack>
#include <list>
#include <map>

#include <ctype.h>
#include <string.h>
#include <stdio.h>

#include "utils.h"

#include "match.h"

#include "utils/my_inttypes.h"
#include "internal_state.h"

#include "utils/precomp.h"

#if defined(_MSC_VER) || defined(__BORLANDC__)
#   pragma hdrstop
#endif

#if defined(_MSC_VER)
#   pragma warning(disable:4018)    // warning C4018: '>=' : signed/unsigned mismatch
#   pragma warning(disable:4244)    // warning C4244: '=' : conversion from '__w64 int' to 'int', possible loss of data
#   pragma warning(disable:4127)    // warning C4127: conditional expression is constant
#   pragma warning(disable:4701)    // warning C4701: potentially uninitialized local variable 'best' used
#   pragma warning(disable:4706)    // warning C4706: assignment within conditional expression
#endif

unsigned char tolowertab[] =
      { 0, 0x1, 0x2, 0x3, 0x4, 0x5, 0x6, 0x7, 0x8, 0x9, 0xa,
        0xb, 0xc, 0xd, 0xe, 0xf, 0x10, 0x11, 0x12, 0x13, 0x14,
        0x15, 0x16, 0x17, 0x18, 0x19, 0x1a, 0x1b, 0x1c, 0x1d,
        0x1e, 0x1f,
        ' ', '!', '"', '#', '$', '%', '&', 0x27, '(', ')',
        '*', '+', ',', '-', '.', '/',
        '0', '1', '2', '3', '4', '5', '6', '7', '8', '9',
        ':', ';', '<', '=', '>', '?',
        '@', 'a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i',
        'j', 'k', 'l', 'm', 'n', 'o', 'p', 'q', 'r', 's',
        't', 'u', 'v', 'w', 'x', 'y', 'z', '[', '\\', ']', '^',
        '_',
        '`', 'a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i',
        'j', 'k', 'l', 'm', 'n', 'o', 'p', 'q', 'r', 's',
        't', 'u', 'v', 'w', 'x', 'y', 'z', '{', '|', '}', '~',
        0x7f,
        0x80, 0x81, 0x82, 0x83, 0x84, 0x85, 0x86, 0x87, 0x88, 0x89,
        0x8a, 0x8b, 0x8c, 0x8d, 0x8e, 0x8f,
        0x90, 0x91, 0x92, 0x93, 0x94, 0x95, 0x96, 0x97, 0x98, 0x99,
        0x9a, 0x9b, 0x9c, 0x9d, 0x9e, 0x9f,
        0xa0, 0xa1, 0xa2, 0xa3, 0xa4, 0xa5, 0xa6, 0xa7, 0xa8, 0xa9,
        0xaa, 0xab, 0xac, 0xad, 0xae, 0xaf,
        0xb0, 0xb1, 0xb2, 0xb3, 0xb4, 0xb5, 0xb6, 0xb7, 0xb8, 0xb9,
        0xba, 0xbb, 0xbc, 0xbd, 0xbe, 0xbf,
        0xc0, 0xc1, 0xc2, 0xc3, 0xc4, 0xc5, 0xc6, 0xc7, 0xc8, 0xc9,
        0xca, 0xcb, 0xcc, 0xcd, 0xce, 0xcf,
        0xd0, 0xd1, 0xd2, 0xd3, 0xd4, 0xd5, 0xd6, 0xd7, 0xd8, 0xd9,
        0xda, 0xdb, 0xdc, 0xdd, 0xde, 0xdf,
        0xe0, 0xe1, 0xe2, 0xe3, 0xe4, 0xe5, 0xe6, 0xe7, 0xe8, 0xe9,
        0xea, 0xeb, 0xec, 0xed, 0xee, 0xef,
        0xf0, 0xf1, 0xf2, 0xf3, 0xf4, 0xf5, 0xf6, 0xf7, 0xf8, 0xf9,
        0xfa, 0xfb, 0xfc, 0xfd, 0xfe, 0xff };

unsigned char touppertab[] =
      { 0, 0x1, 0x2, 0x3, 0x4, 0x5, 0x6, 0x7, 0x8, 0x9, 0xa,
        0xb, 0xc, 0xd, 0xe, 0xf, 0x10, 0x11, 0x12, 0x13, 0x14,
        0x15, 0x16, 0x17, 0x18, 0x19, 0x1a, 0x1b, 0x1c, 0x1d,
        0x1e, 0x1f,
        ' ', '!', '"', '#', '$', '%', '&', 0x27, '(', ')',
        '*', '+', ',', '-', '.', '/',
        '0', '1', '2', '3', '4', '5', '6', '7', '8', '9',
        ':', ';', '<', '=', '>', '?',
        '@', 'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I',
        'J', 'K', 'L', 'M', 'N', 'O', 'P', 'Q', 'R', 'S',
        'T', 'U', 'V', 'W', 'X', 'Y', 'Z', '[', '\\', ']', '^',
        0x5f,
        '`', 'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I',
        'J', 'K', 'L', 'M', 'N', 'O', 'P', 'Q', 'R', 'S',
        'T', 'U', 'V', 'W', 'X', 'Y', 'Z', '{', '|', '}', '~',
        0x7f,
        0x80, 0x81, 0x82, 0x83, 0x84, 0x85, 0x86, 0x87, 0x88, 0x89,
        0x8a, 0x8b, 0x8c, 0x8d, 0x8e, 0x8f,
        0x90, 0x91, 0x92, 0x93, 0x94, 0x95, 0x96, 0x97, 0x98, 0x99,
        0x9a, 0x9b, 0x9c, 0x9d, 0x9e, 0x9f,
        0xa0, 0xa1, 0xa2, 0xa3, 0xa4, 0xa5, 0xa6, 0xa7, 0xa8, 0xa9,
        0xaa, 0xab, 0xac, 0xad, 0xae, 0xaf,
        0xb0, 0xb1, 0xb2, 0xb3, 0xb4, 0xb5, 0xb6, 0xb7, 0xb8, 0xb9,
        0xba, 0xbb, 0xbc, 0xbd, 0xbe, 0xbf,
        0xc0, 0xc1, 0xc2, 0xc3, 0xc4, 0xc5, 0xc6, 0xc7, 0xc8, 0xc9,
        0xca, 0xcb, 0xcc, 0xcd, 0xce, 0xcf,
        0xd0, 0xd1, 0xd2, 0xd3, 0xd4, 0xd5, 0xd6, 0xd7, 0xd8, 0xd9,
        0xda, 0xdb, 0xdc, 0xdd, 0xde, 0xdf,
        0xe0, 0xe1, 0xe2, 0xe3, 0xe4, 0xe5, 0xe6, 0xe7, 0xe8, 0xe9,
        0xea, 0xeb, 0xec, 0xed, 0xee, 0xef,
        0xf0, 0xf1, 0xf2, 0xf3, 0xf4, 0xf5, 0xf6, 0xf7, 0xf8, 0xf9,
        0xfa, 0xfb, 0xfc, 0xfd, 0xfe, 0xff };

static const char *inet_ntop4 (const u_char *src, char *dst, size_t size)
{
  char tmp [sizeof("255.255.255.255")+16];

  if ((size_t)sprintf(tmp,"%u.%u.%u.%u",
      static_cast<unsigned>(static_cast<unsigned char>(src[0])),
      static_cast<unsigned>(static_cast<unsigned char>(src[1])),
      static_cast<unsigned>(static_cast<unsigned char>(src[2])),
      static_cast<unsigned>(static_cast<unsigned char>(src[3]))) > size)
  {
    //SOCK_ERRNO (ENOSPC);
    return (NULL);
  }
  return strcpy (dst, tmp);
}

#ifdef SPRINTF_CHAR
# define SPRINTF(x) strlen(sprintfx)
#else
# define SPRINTF(x) ((size_t)sprintf x)
#endif


/* const char *
 * inet_ntop6(src, dst, size)
 *      convert IPv6 binary address into presentation (printable) format
 * author:
 *      Paul Vixie, 1996.
 */
const char * inet_ntop6(const u_char *src, char *dst, size_t size)
{
        /*
         * Note that int32_t and int16_t need only be "at least" large enough
         * to contain a value of the specified size.  On some systems, like
         * Crays, there is no such thing as an integer variable with 16 bits.
         * Keep this in mind if you think this function should have been coded
         * to use pointer overlays.  All the world's not a VAX.
         */
        char tmp[sizeof "ffff:ffff:ffff:ffff:ffff:ffff:255.255.255.255"], *tp;
        struct { int base, len; } best, cur;
        u_int words[IN6ADDRSZ / INT16SZ];
        int i;

        /*
         * Preprocess:
         *      Copy the input (bytewise) array into a wordwise array.

         *      Find the longest run of 0x00's in src[] for :: shorthanding.
         */
        memset(words, '\0', sizeof words);
        for (i = 0; i < IN6ADDRSZ; i++)
                words[i / 2] |= (src[i] << ((1 - (i % 2)) << 3));
        best.base = -1;
        cur.base = -1;
        for (i = 0; i < (IN6ADDRSZ / INT16SZ); i++) {
                if (words[i] == 0) {
                        if (cur.base == -1)
                                cur.base = i, cur.len = 1;
                        else
                                cur.len++;
                } else {
                        if (cur.base != -1) {
                                if (best.base == -1 || cur.len > best.len)
                                        best = cur;
                                cur.base = -1;
                        }
                }
        }
        if (cur.base != -1) {
                if (best.base == -1 || cur.len > best.len)
                        best = cur;
        }
        if (best.base != -1 && best.len < 2)
                best.base = -1;

        /*
         * Format the result.
         */
        tp = tmp;
        for (i = 0; i < (IN6ADDRSZ / INT16SZ); i++) {
                /* Are we inside the best run of 0x00's? */
                if (best.base != -1 && i >= best.base &&
                    i < (best.base + best.len)) {
                        if (i == best.base)
                                *tp++ = ':';
                        continue;
                }

                /* Are we following an initial run of 0x00s or any real hex? */
                if (i != 0)
                        *tp++ = ':';
                /* Is this address an encapsulated IPv4? */
                if (i == 6 && best.base == 0 &&
                    (best.len == 6 || (best.len == 5 && words[5] == 0xffff))) {
                        if (!inet_ntop4(src+12, tp, sizeof tmp - (tp - tmp)))
                                return (NULL);
                        tp += strlen(tp);
                        break;
                }
                tp += SPRINTF((tp, "%x", words[i]));
        }
        /* Was it a trailing run of 0x00's? */
        if (best.base != -1 && (best.base + best.len) == (IN6ADDRSZ / INT16SZ))
                *tp++ = ':';
        *tp++ = '\0';

        /*
         * Check for overflow, copy, and we're done.
         */
        if ((size_t)(tp - tmp) > size) {
                //errno = ENOSPC;
                return (NULL);
        }
        strcpy(dst, tmp);
        return (dst);
}


const char * my_inet_ntop (int af, const void *src, char *dst, size_t size)
{
    switch (af)
    {
        case AF_INET:
            return inet_ntop4 ((const u_char*)src, dst, size);
        case AF_INET6:
            return inet_ntop6 ((const u_char*)src, dst, size);
        default:
            return (NULL);
    }
}

/*
 * inetntop: return the : notation of a given IPv6 internet number.
 *       or the dotted-decimal notation for IPv4
 *           make sure the compressed representation (rfc 1884) isn't used.
 */
char  *inetntop(int af, const void *in, char *out, size_t the_size)
{
  static char local_dummy[MYDUMMY_SIZE];

  if (the_size > sizeof(local_dummy))
  {
    the_size = sizeof(local_dummy);
  }

  if (!my_inet_ntop(af, in, local_dummy, the_size))
  {
    /* good that every function calling this one
     * checks the return value ... NOT */
    return NULL;
  }
  /* quick and dirty hack to give ipv4 just ipv4 instead of
   * ::ffff:ipv4 - Q */
  if (af == AF_INET6 && IN6_IS_ADDR_V4MAPPED((const struct in_addr6_ *)in))
  {
    char  *p;

    if (!(p = strstr(local_dummy, ":ffff:")) &&
      !(p = strstr(local_dummy, ":FFFF:")))
    {
      return NULL;  /* crash and burn */
    }
    strcpy(out, p + 6);
    return out;
  }
  if (strstr(local_dummy, "::"))
      {
    char cnt = 0, *cp = local_dummy, *op = out;

    while (*cp)
        {
      if (*cp == ':')
        cnt += 1;
      if (*cp++ == '.')
          {
        cnt += 1;
        break;
          }
        }
    cp = local_dummy;
    while (*cp)
        {
      *op++ = *cp++;
      if (*(cp-1) == ':' && *cp == ':')
          {
        if ((cp-1) == local_dummy)
            {
          op--;
          *op++ = '0';
          *op++ = ':';
            }

        *op++ = '0';
        while (cnt++ < 7)
            {
          *op++ = ':';
          *op++ = '0';
            }
          }
        }
    if (*(op-1)==':') *op++ = '0';
    *op = '\0';
/*#ifndef  CLIENT_COMPILE
    Debug((DEBUG_DNS,"Expanding `%s' -> `%s'", local_dummy,
           out));
#endif*/
      }
  else
    bcopy(local_dummy, out,  the_size);

  return out;
}

/* int
 * inet_pton4(src, dst, pton)
 *      when last arg is 0: inet_aton(). with hexadecimal, octal and shorthand.
 *      when last arg is 1: my_inet_pton(). decimal dotted-quad only.
 * return:
 *      1 if `src' is a valid input, else 0.
 * notice:
 *      does not touch `dst' unless it's returning 1.
 * author:
 *      Paul Vixie, 1996.
 */
int
inet_pton4(const char *src, u_char *dst, int pton)
{
        u_int val;
        u_int digit;
        u_int base;
        int n;
        unsigned char c;
        u_int parts[4];
        register u_int *pp = parts;

        c = *src;
        for (;;) {
                /*
                 * Collect number up to ``.''.
                 * Values are specified as for C:
                 * 0x=hex, 0=octal, isdigit=decimal.
                 */
                if (!isdigit(c))
                        return (0);
                val = 0; base = 10;
                if (c == '0') {
                        c = *++src;
                        if (c == 'x' || c == 'X')
                                base = 16, c = *++src;
                        else if (isdigit(c) && c != '9')
                                base = 8;
                }
                /* my_inet_pton() takes decimal only */
                if (pton && base != 10)
                        return (0);
                for (;;) {
                        if (isdigit(c)) {
                                digit = c - '0';
                                if (digit >= base)
                                        break;
                                val = (val * base) + digit;
                                c = *++src;
                        } else if (base == 16 && isxdigit(c)) {
                                digit = c + 10 - (islower(c) ? 'a' : 'A');
                                if (digit >= 16)
                                        break;
                                val = (val << 4) | digit;
                                c = *++src;
                        } else
                                break;
                }
                if (c == '.') {
                        /*
                         * Internet format:
                         *      a.b.c.d
                         *      a.b.c   (with c treated as 16 bits)
                         *      a.b     (with b treated as 24 bits)
                         *      a       (with a treated as 32 bits)
                         */
                        if (pp >= parts + 3)
                                return (0);
                        *pp++ = val;
                        c = *++src;
                } else
                        break;
        }
        /*
         * Check for trailing characters.
         */
        if (c != '\0' && !isspace(c))
                return (0);
        /*
         * Concoct the address according to
         * the number of parts specified.
         */
        n = pp - parts + 1;
        /* my_inet_pton() takes dotted-quad only.  it does not take shorthand. */
        if (pton && n != 4)
                return (0);
        switch (n) {

        case 0:
                return (0);             /* initial nondigit */

        case 1:                         /* a -- 32 bits */
                break;

        case 2:                         /* a.b -- 8.24 bits */
                if (parts[0] > 0xff || val > 0xffffff)
                        return (0);
                val |= parts[0] << 24;
                break;

        case 3:                         /* a.b.c -- 8.8.16 bits */
                if ((parts[0] | parts[1]) > 0xff || val > 0xffff)
                        return (0);
                val |= (parts[0] << 24) | (parts[1] << 16);
                break;

        case 4:                         /* a.b.c.d -- 8.8.8.8 bits */
                if ((parts[0] | parts[1] | parts[2] | val) > 0xff)
                        return (0);
                val |= (parts[0] << 24) | (parts[1] << 16) | (parts[2] << 8);
                break;
        }
        if (dst) {
                val = htonl(val);
                memcpy(dst, &val, INADDRSZ);
        }
        return (1);
}

/* int
 * inet_pton6(src, dst)
 *      convert presentation level address to network order binary form.
 * return:
 *      1 if `src' is a valid [RFC1884 2.2] address, else 0.
 * notice:
 *      (1) does not touch `dst' unless it's returning 1.
 *      (2) :: in a full address is silently ignored.
 *          NB by Marian VooDooMan Meravy: 2005-06-29: Added decompress :: addresses first, so it should work
 * credit:
 *      inspired by Mark Andrews.
 * author:
 *      Paul Vixie, 1996.
 *      VooDooMan, 2005-06-29
 *      VooDooMan, 2005-10-02 (irrelevant changes)
 *      VooDooMan, 2005-11-29 Make thread-safe (because of ipv6_decompress() function has changed, MY_DELETE_A(src_saved))
 */
int
inet_pton6(const char *src, u_char *dst)
{
        char* src_saved=NULL;
        if(src)
                src=src_saved=ipv6_decompress(src);
        else
                return 0;

        static const char xdigits_l[] = "0123456789abcdef",
                          xdigits_u[] = "0123456789ABCDEF";
        u_char tmp[IN6ADDRSZ], *tp, *endp, *colonp;
        const char *xdigits, *curtok;
        int ch, saw_xdigit;
        u_int val;

        memset((tp = tmp), '\0', IN6ADDRSZ);
        endp = tp + IN6ADDRSZ;
        colonp = NULL;
        /* Leading :: requires some special handling. */
        if (*src == ':')
                if (*++src != ':') {
                        MY_DELETE_A(src_saved);
                        return (0);
                }
        curtok = src;
        saw_xdigit = 0;
        val = 0;
        while ((ch = *src++) != '\0') {
                const char *pch;

                if ((pch = strchr((xdigits = xdigits_l), ch)) == NULL)
                        pch = strchr((xdigits = xdigits_u), ch);
                if (pch != NULL) {
                        val <<= 4;
                        val |= (pch - xdigits);
                        if (val > 0xffff) {
                                MY_DELETE_A(src_saved);
                                return (0);
                        }
                        saw_xdigit = 1;
                        continue;
                }
                if (ch == ':') {
                        curtok = src;
                        if (!saw_xdigit) {
                                if (colonp) {
                                        MY_DELETE_A(src_saved);
                                        return (0);
                                }
                                colonp = tp;
                                continue;
                        } else if (*src == '\0') {
                                MY_DELETE_A(src_saved);
                                return (0);
                        }
                        if (tp + INT16SZ > endp) {
                                MY_DELETE_A(src_saved);
                                return (0);
                        }
                        *tp++ = (u_char) (val >> 8) & 0xff;
                        *tp++ = (u_char) val & 0xff;
                        saw_xdigit = 0;
                        val = 0;
                        continue;
                }
                if (ch == '.' && ((tp + INADDRSZ) <= endp) &&
                    inet_pton4(curtok, tp, 1) > 0) {
                        tp += INADDRSZ;
                        saw_xdigit = 0;
                        break;  /* '\0' was seen by inet_pton4(). */
                }
                MY_DELETE_A(src_saved);
                return (0);
        }
        if (saw_xdigit) {
                if (tp + INT16SZ > endp) {
                        MY_DELETE_A(src_saved);
                        return (0);
                }
                *tp++ = (u_char) (val >> 8) & 0xff;
                *tp++ = (u_char) val & 0xff;
        }
        if (colonp != NULL) {
                /*
                 * Since some memmove()'s erroneously fail to handle
                 * overlapping regions, we'll do the shift by hand.
                 */
                const int n = tp - colonp;
                int i;

                if (tp == endp) {
                        MY_DELETE_A(src_saved);
                        return (0);
                }
                for (i = 1; i <= n; i++) {
                        endp[- i] = colonp[n - i];
                        colonp[n - i] = 0;
                }
                tp = endp;
        }
        if (tp != endp) {
                MY_DELETE_A(src_saved);
                return (0);
        }
        memcpy(dst, tmp, IN6ADDRSZ);
        MY_DELETE_A(src_saved);
        return (1);
}

/* int
 * my_inet_pton(af, src, dst)
 *      convert from presentation format (which usually means ASCII printable)
 *      to network format (which is usually some kind of binary format).
 * return:
 *      1 if the address was valid for the specified address family
 *      0 if the address wasn't valid (`dst' is untouched in this case)
 *      -1 if some other error occurred (`dst' is untouched in this case, too)
 * author:
 *      Paul Vixie, 1996.
 */
int
my_inet_pton(int af, const char *src, void *dst)
{

        switch (af) {
        case AF_INET:
                return (inet_pton4(src, (unsigned char*)dst, 1));
        case AF_INET6:
                return (inet_pton6(src, (unsigned char*)dst));
        default:
                //errno = EAFNOSUPPORT;
                return (-1);
        }
        /* NOTREACHED */
}

/* inetpton(af, src, dst)
**
** This is a wrapper for my_inet_pton(), so we can use ipv4 addresses with an
** af of AF_INET6, and that it gets converted to ipv4 mapped ipv6.
*/
int  inetpton(int af, const char *src, void *dst)
{
  int  i;

  /* an empty string should listen to all */
  if (af == AF_INET6 && *src && !strchr(src, ':'))
      {
    i = my_inet_pton(AF_INET, src, dst);

    /* ugly hack */
    memcpy((char *)dst + 12, dst, 4);
    memset(dst, 0, 10);
    memset((char *)dst + 10, 0xff, 2);
    return i;
      }
  return my_inet_pton(af, src, dst);
}
//#endif

char* ipv6_decompress(const char* src)
{
    size_t res_l=1024;
    if(strlen(src)>res_l+1)
        res_l=strlen(src)+1;
    char *res=MY_NEW char[res_l+1];
    if(strstr(src,":")==NULL) {
        strcpy(res,src);
        return res;
    }
    if(strstr(src,"::")==NULL) {
        strcpy(res,src);
        return res;
    }
    int expected_num_frags=8;
    if(strstr(src,".")!=NULL)
        expected_num_frags=7;
    int num_frags=0;
    int num_size=0;
    bool got=false;
    unsigned int i1;
    for(i1=0; i1<strlen(src)+1; ++i1) { // +1 here is not typo! we need to process zero-terminator as well!
        if(src[i1]==':' || src[i1]==0) {
            if(/*num_size>4 ||*/ num_frags>expected_num_frags) {
                // error
                strcpy(res,src);
                return res;
            }
            num_size=0;
            if(i1>0 && src[i1-1]==':')
                got=true;
            else
                if((i1>0 && src[i1-1]!=':') && i1>0)
                    num_frags++;
        } else {
            num_size++;
        }
    }
    if(!got) {
        strcpy(res,src);
        return res;
    }
    if(num_frags>8) {
        strcpy(res,src);
        return res;
    }
    bool need_colon=false;
    size_t i2=0;
    for(i1=0; i1<strlen(src); ++i1) {
        if(i2>res_l-8) {
            strcpy(res,src);
            return res;
        }
        if(src[i1]==':' && i1>0 && src[i1-1]==':') {
            for(int i3=0; i3<expected_num_frags-num_frags; ++i3) {
                if(i2>res_l-8) {
                    strcpy(res,src);
                    return res;
                }
                if(i2>0 && res[i2-1]!=':')
                    res[i2++]=':';
                res[i2++]='0';
                need_colon=true;
            }
        } else {
            if(!(i1==0 && src[i1]==':')) {
                if(need_colon)
                    res[i2++]=':';
                if(!need_colon)
                    res[i2++]=src[i1];
                if(need_colon && src[i1]!=':')
                    res[i2++]=src[i1];
            }
            need_colon=false;
        }
    }
    res[i2++]=0;
    return res;
}

char* index(char* str, char sub)
{
    char tmp[2];
    tmp[0]=sub;
    tmp[1]=0;
    return strstr(str,tmp);
}

u_char my_tolower(u_char ch)
{
    switch(irc_casemapping) {
        case CASE_RFC1459:
            return tolowertab_rfc1459[ch];
        case CASE_ASCII:
        default:
            break;
    }
    return tolowertab_ascii[ch];
}

int  match(const char *mask, const char *name)
{
  Reg  u_char  *m = (u_char *)mask, *n = (u_char *)name;
  const char  *ma = mask, *na = name;
  int  wild = 0, q = 0, calls = 0;

  if (!*mask)
    return 1;

  if (mask[0]=='*' && mask[1]=='\0')
    return 0;

  while (1)
      {
#ifdef  MAX_ITERATIONS
          if (calls++ > MAX_ITERATIONS)
      break;
#endif

    if (*m == '*')
       {
      while (*m == '*')
        m++;
      wild = 1;
      ma = (char *)m;
      na = (char *)n;
        }

    if (!*m)
        {
        if (!*n)
        return 0;
        for (m--; (m > (u_char *)mask) && (*m == '?'); m--)
        ;
      if ((m > (u_char *)mask) && (*m == '*') &&
          (m[-1] != '\\'))
        return 0;
      if (!wild)
        return 1;
      m = (u_char *)ma;
      n = (u_char *)++na;
        }
    else if (!*n)
      return 1;
    if ((*m == '\\') &&
      ((m[1] == '*') || (m[1] == '?') || (m[1] == '#')))
        {
      m++;
      q = 1;
        }
    else
      q = 0;

    if ((my_tolower(*m) == my_tolower(*n))
      || (*m == '?' && !q)
      || (*m == '#' && !q && isdigit(*n)))
        {
      if (*m)
        m++;
      if (*n)
        n++;
        }
    else
        {
      if (!wild)
        return 1;
      m = (u_char *)ma;
      n = (u_char *)++na;
        }
      }

  return 1;
}

// BEGIN CHANGE - VooDooMan 2004/10/18 - changed
//int    match_ipmask(char *mask, aClient *cptr, int maskwithusername)
int    match_ipmask(char *mask, char* username, in_addr6_ u_addr, int maskwithusername)
{
  int  m;
  char  *p;
// BEGIN CHANGE - VooDooMan 2004/10/18 - changed
//  struct  IN_ADDR addr;
  struct  in_addr6_ addr;
// END   CHANGE - VooDooMan 2004/10/18 - changed
  char  dummy[128];
  //char  *omask;
  u_long  lmask;
#ifdef  INET6
  int  j;
#endif

  //omask = mask;
  strncpyzt(dummy, mask, sizeof(dummy));
  mask = dummy;
  if (maskwithusername && (p = index(mask, '@')))
  {
    *p = '\0';
// BEGIN CHANGE - VooDooMan 2004/10/18 - changed
//    if (match(mask, cptr->username))
    if (match(mask, username))
// END   CHANGE - VooDooMan 2004/10/18 - changed
      return 1;
    mask = p + 1;
  }
  if (!(p = index(mask, '/')))
    goto badmask;
  *p = '\0';

  if (sscanf(p + 1, "%d", &m) != 1)
  {
    goto badmask;
  }
  if (!m)
    return 0;       /* x.x.x.x/0 always matches */
#ifndef  INET6
  if (m < 0 || m > 32)
    goto badmask;
  lmask = htonl((u_long)0xffffffffL << (32 - m));
// BEGIN CHANGE - VooDooMan 2004/10/18 - changed
//  addr.s_addr = inetaddr(mask);
//  return ((addr.s_addr ^ cptr->ip.s_addr) & lmask) ? 1 : 0;
  addr.s_addr=inet_addr(mask);
  return ((addr.s_addr ^ u_addr) & lmask) ? 1 : 0;
// END   CHANGE - VooDooMan 2004/10/18 - changed
#else
  if (m < 0 || m > 128)
    goto badmask;
  if (my_inet_pton(AF_INET6, mask, (void *)addr.bytes) != 1)
  {
    return -1;
  }

  /* Make sure that the ipv4 notation still works. */
  if (IN6_IS_ADDR_V4MAPPED(&addr) && m < 96)
  {
    m += 96;
  }

  j = m & 0x1F;  /* number not mutliple of 32 bits */
  m >>= 5;  /* number of 32 bits */

  if (m && memcmp((void *)(addr.bytes),
      (void *)(/*cptr->ip.s6_addr*/&(u_addr.bytes)), m << 2))
    return 1;

  if (j)
  {
    lmask = htonl((u_long)0xffffffffL << (32 - j));
    if ((((u_int32_t *)(addr.bytes))[m] ^
        ((u_int32_t *)(/*cptr->ip.s6_addr*/&(u_addr.bytes)))[m]) & lmask)
      return 1;
  }

  return 0;
#endif
badmask:
// BEGIN CHANGE - VooDooMan 2004/10/18 - changed
/*  if (maskwithusername)
  sendto_flag(SCH_ERROR, "Ignoring bad mask: %s", omask);*/
// END   CHANGE - VooDooMan 2004/10/18 - changed
  return -1;
}

