/* $Id: mdump.cpp,v 1.31 2008/08/23 15:10:44 ghostvoodooman Exp $ */

/*

URL: http://www.codeproject.com/debug/postmortemdebug_standalone1.asp

Article content copyright Andy Pennell, 2002

This file was downloaded as part of article at above URL.
This file copyright Andy Pennell, 2002

Modified by Marian VooDooMan Meravy for purposes of VooDoo cIRCle IRC bot.

*/

#if defined(_MSC_VER)   // other compilers produce binary with incompatible symbols

#include <time.h>
#include <stdio.h>
#include <malloc.h>
#include <process.h>
#include <iostream>

#define _tcsrchr    strrchr
#define _tcscat     strcat
#define _tcscpy     strcpy

#include "mdump.h"

#include "../utils/my_getpid.h"

#if defined(_MSC_VER)
#   pragma warning(disable:4996)    // warning C4996: 'sprintf': This function or variable may be unsafe. Consider using sprintf_s instead. To disable deprecation, use _CRT_SECURE_NO_WARNINGS. See online help for details.
#endif

LPCSTR MiniDumper::m_szAppName;

MiniDumper::MiniDumper(const char* timestamp)
{
    // if this assert fires then you have two instances of MiniDumper
    // which is not allowed
    //assert( m_szAppName==NULL );

    m_szAppName = (LPCSTR)malloc(strlen(timestamp)+512);
    sprintf((char*)m_szAppName,"vdc.crash-dump.%s.pid%lu",timestamp,static_cast<unsigned long>(my_getpid()));

    ::SetUnhandledExceptionFilter( TopLevelFilter );
}

LONG WINAPI MiniDumper::TopLevelFilter( struct _EXCEPTION_POINTERS *pExceptionInfo )
{
    if(!m_szAppName)
        return EXCEPTION_EXECUTE_HANDLER; // not initialized

    LONG retval = EXCEPTION_CONTINUE_SEARCH;
    //HWND hParent = NULL;                        // find a better value for your app

    // firstly see if dbghelp.dll is around and has the function we need
    // look next to the EXE first, as the one in System32 might be old
    // (e.g. Windows 2000)
    HMODULE hDll = NULL;
    char szDbgHelpPath[_MAX_PATH];

    if (GetModuleFileName( NULL, szDbgHelpPath, _MAX_PATH ))
    {
        char *pSlash = _tcsrchr( szDbgHelpPath, '\\' );
        if (pSlash)
        {
            _tcscpy( pSlash+1, "DBGHELP.DLL" );
            hDll = ::LoadLibrary( szDbgHelpPath );
        }
    }

    if(NULL==hDll)
        hDll=::LoadLibrary(".\\DBGHELP.DLL");

    if (hDll==NULL)
    {
        // load any version we can
        hDll = ::LoadLibrary( "DBGHELP.DLL" );
    }

    LPCTSTR szResult = NULL;

    if (hDll)
    {
        MINIDUMPWRITEDUMP pDump = (MINIDUMPWRITEDUMP)::GetProcAddress( hDll, "MiniDumpWriteDump" );
        if (pDump)
        {
            char szDumpPath[_MAX_PATH];
            char szScratch [_MAX_PATH];

            // work out a good place for the dump file
            /*
            if (!GetTempPath( _MAX_PATH, szDumpPath ))
                _tcscpy( szDumpPath, ".\\" );
            */
            _tcscpy( szDumpPath, ".\\" );

            _tcscat( szDumpPath, m_szAppName );

            static long cnt=0;
            char tmp[64];
            _ltoa(cnt++,tmp,10);
            strcat(szDumpPath,".");
            strcat(szDumpPath,tmp);

            _tcscat( szDumpPath, ".dmp" );

            // ask the user if they want to save a dump file
            //if (::MessageBox( NULL, "Something bad happened in VooDoo cIRCle, would you like to save a diagnostic dump file for sending it to developer?", "VooDoo cIRCle", MB_YESNO )==IDYES)
            {
                // create the file
                HANDLE hFile = ::CreateFile( szDumpPath, GENERIC_WRITE, FILE_SHARE_WRITE, NULL, CREATE_ALWAYS,
                                            FILE_ATTRIBUTE_NORMAL, NULL );

                if (hFile!=INVALID_HANDLE_VALUE)
                {
                    _MINIDUMP_EXCEPTION_INFORMATION ExInfo;

                    ExInfo.ThreadId = ::GetCurrentThreadId();
                    ExInfo.ExceptionPointers = pExceptionInfo;
                    //ExInfo.ClientPointers = NULL;
                    ExInfo.ClientPointers = 0;

                    // write the dump
                    BOOL bOK = pDump( GetCurrentProcess(), GetCurrentProcessId(), hFile, MiniDumpWithFullMemory, (pExceptionInfo ? &ExInfo : NULL), NULL, NULL );
                    if (bOK)
                    {
                        sprintf( szScratch, "Saved dump file to '%s'", szDumpPath );
                        szResult = szScratch;
                        retval = EXCEPTION_EXECUTE_HANDLER;
                    }
                    else
                    {
                        sprintf( szScratch, "Failed to save dump file to '%s' (error %lu)", szDumpPath, GetLastError() );
                        szResult = szScratch;
                    }
                    ::CloseHandle(hFile);
                }
                else
                {
                    sprintf( szScratch, "Failed to create dump file '%s' (error %lu)", szDumpPath, GetLastError() );
                    szResult = szScratch;
                }
            }
        }
        else
        {
            szResult = "DBGHELP.DLL too old";
        }
    }
    else
    {
        szResult = "DBGHELP.DLL not found";
    }

    /*
    if (szResult)
        ::MessageBox( NULL, szResult, m_szAppName, MB_OK );
    */
    if(szResult)
        std::cout << szResult << std::endl;
    std::cout.flush();

    return retval;
}

MiniDumper::~MiniDumper()
{
    //assert(m_szAppName);
    free((void*)m_szAppName);
    m_szAppName=NULL;
}

#else  // defined(_MSC_VER)

void mdump_dummy_function()
{
    // get rid of linker warning about empty table of ".a" library
}

#endif

