/***************************************************************************
                          c_single_element_allocator.hh  -  description
                             -------------------
    begin                : Sun Jul 29 2007
    copyright            : (C) 2007 by Marian VooDooMan Meravy
    email                : ghostvoodooman -aT= users (dot) sourceforge.net
 ***************************************************************************/

/* $Id: c_single_element_allocator.hh,v 1.14 2008/11/24 12:09:46 ghostvoodooman Exp $ */

/***************************************************************************

VooDoo cIRCle - an IRC (ro)bot
Copyright (C) 2005-2008 by Marian VooDooMan Meravy
  (ghostvoodooman -aT= users (dot) sourceforge.net)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the

GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************/

/*!
    \file
    \brief Provides single element at once allocator for STL
    \version $Id: c_single_element_allocator.hh,v 1.14 2008/11/24 12:09:46 ghostvoodooman Exp $
*/

#ifndef C_SINGLE_ELEMENT_ALLOCATOR_HH__
#define C_SINGLE_ELEMENT_ALLOCATOR_HH__

#include "../utils/prologue.h"

#include <stddef.h>     // for size_t

#include "hp_std.hh"

template<class T>
class c_single_element_allocator;

template<class T>
class c_single_element_allocator
{
public:
    typedef size_t    size_type;
    typedef ptrdiff_t difference_type;
    typedef T*        pointer;
    typedef const T*  const_pointer;
    typedef T&        reference;
    typedef const T&  const_reference;
    typedef T         value_type;
    template<class U> struct rebind { typedef c_single_element_allocator<U> other; };

    MY_WISH_INLINE c_single_element_allocator() { ; }
    MY_WISH_INLINE c_single_element_allocator(const c_single_element_allocator&) { ; }
    template<class U> MY_WISH_INLINE c_single_element_allocator(const c_single_element_allocator<U>&) { ; }
    MY_WISH_INLINE ~c_single_element_allocator() { ; }

    MY_WISH_INLINE pointer address(reference x) const
    {
        return &x;
    }

    MY_WISH_INLINE const_pointer address(const_reference x) const
    {
        return &x;
    }

    MY_WISH_INLINE pointer allocate(size_type n, const_pointer hint = NULL)
    {
        MY_ASSERT(1==n);
        MY_ASSERT(NULL==hint);

        return reinterpret_cast<pointer>(::malloc(sizeof(T)*n));
    }

    MY_WISH_INLINE void deallocate(pointer p, size_type UNUSEDARG(n))
    {
        ::free(p);
    }

    MY_WISH_INLINE size_type max_size() const
    {
        return 1;
    }

    MY_WISH_INLINE void construct(pointer p, const T& val)
    {
        hp_std::constructor<T,T>(p,val);
    }

    void destroy(pointer p)
    {
        hp_std::destructor<T>(p);
    }
};

template<class T1, class T2>
MY_WISH_INLINE
bool operator==(const c_single_element_allocator<T1>& x1, const c_single_element_allocator<T2>& x2)
{
    return typeid(x1)==typeid(x2);
}

template<class T1, class T2>
MY_WISH_INLINE
bool operator!=(const c_single_element_allocator<T1>& x1, const c_single_element_allocator<T2>& x2)
{
    return typeid(x1)!=typeid(x2);
}

#endif

