/***************************************************************************
                          low_memory_condition.cpp  -  description
                             -------------------
    begin                : Fri Sep 07 2007
    copyright            : (C) 2007 by Marian VooDooMan Meravy
    email                : ghostvoodooman -aT= users (dot) sourceforge.net
 ***************************************************************************/

/* $Id: low_memory_condition.cpp,v 1.13 2008/11/24 12:09:46 ghostvoodooman Exp $ */

/***************************************************************************

VooDoo cIRCle - an IRC (ro)bot
Copyright (C) 2004-2008 by Marian VooDooMan Meravy
  (ghostvoodooman -aT= users (dot) sourceforge.net)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************/

/*!
    \file
    \brief Provides "handling" of low memory condition
    \version $Id: low_memory_condition.cpp,v 1.13 2008/11/24 12:09:46 ghostvoodooman Exp $
*/

#include "../utils/prologue.h"

#include <iostream>

#include <new>          // bad_alloc

#include <stdlib.h>     // for _exit()

#if defined(_WIN32) || defined(_WIN64) || defined(__CYGWIN__)
#   include <windows.h>    // for Sleep()
#endif

#include "../irc_bot.h"

using namespace std;

#include "../utils/precomp.h"

#if defined(_MSC_VER) || defined(__BORLANDC__)
#   pragma hdrstop
#endif

#ifdef _MSC_VER
void __cdecl
#else
void
#endif
low_memory_condition_handler()
{
    static bool in=false;
    if(in) {
        // another thread?
        enum
        {
            WAIT_SECONDS = 5,
            MILISECS_IN_SECOND = 1000,
        };
#if defined(_WIN32) || defined(_WIN64) || defined(__CYGWIN__)
        Sleep(WAIT_SECONDS * MILISECS_IN_SECOND);
#else
        struct timeval tv;
        tv.tv_sec=WAIT_SECONDS;
        tv.tv_usec=0;
        select(1,NULL,NULL,NULL,&tv);
#endif
        throw bad_alloc(
#ifdef _MSC_VER
            "PANIC: recursive call to " __my_func__
#endif
        );
        return;
    }
    in=true;
    static const char* msg="PANIC: Low memory condition. Going to dump the core via NULL pointer dereference.";
    try {
        cout << msg << endl;
    } catch(...) {
        // nevermind
    }
    cout.flush();

    try { // for sure's sake
        closeall(3); // close all FD's >= 3 (to force flushing buffers)
    } catch(...)
    {
        // nevermind
    }

    try { // for sure's sake
        dump_core();
    } catch(...) {
        // nevermind
    }
    // NON-MSVC: NOTREACH
    _exit(RC_BOT_RESTART);
    // NOTREACH
}

void low_memory_condition_install_new_handler_callback()
{
    (void)set_new_handler(low_memory_condition_handler);
}

