/***************************************************************************
                          memleak.h  -  description
                             -------------------
    begin                : Fri May 4 2007
    copyright            : (C) 2005-2007 by Marian VooDooMan Meravy
    email                : ghostvoodooman -aT= users (dot) sourceforge.net
 ***************************************************************************/

/* $Id: memleak.h,v 1.1 2008/11/24 12:09:46 ghostvoodooman Exp $ */

/***************************************************************************

VooDoo cIRCle - an IRC (ro)bot
Copyright (C) 2004-2008 by Marian VooDooMan Meravy
  (ghostvoodooman -aT= users (dot) sourceforge.net)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************/

/*!
    \file
    \brief Provides memory leak detection
    \author VooDooMan, Kasper Peeters, and possibly HP
    \version $Id: memleak.h,v 1.1 2008/11/24 12:09:46 ghostvoodooman Exp $
*/

#ifndef MEMLEAK_H__
#define MEMLEAK_H__

#include "../utils/prologue.h"

#include "../memory/heap_debug.h"

using namespace std;

#ifdef _MSC_VER
#   ifdef _DEBUG
#   include <crtdbg.h>
#       define SET_CRT_DEBUG_FIELD(a) \
            _CrtSetDbgFlag((a) | _CrtSetDbgFlag(_CRTDBG_REPORT_FLAG))
#       define CLEAR_CRT_DEBUG_FIELD(a) \
            _CrtSetDbgFlag(~(a) & _CrtSetDbgFlag(_CRTDBG_REPORT_FLAG))
#   else
#       define SET_CRT_DEBUG_FIELD(a) ((void) 0)
#       define CLEAR_CRT_DEBUG_FIELD(a) ((void) 0)
#   endif // _DEBUG
#endif

namespace my_mem_manager
{

    // Author: Kasper Peeters, VooDooMan, and possibly HP

#ifdef _MSC_VER
//  disable MSVC warning: warning C4100: 'p' : unreferenced formal parameter
//      (when class has no destructor)
#   pragma warning( push )
#   pragma warning( disable : 4100 )
#endif

    template<class T>
    MY_FORCE_INLINE
    void destruct(T *p)
    {
        p->~T();
    }

#ifdef _MSC_VER
//  disable MSVC warning: warning C4100: 'p' : unreferenced formal parameter
//      (when class has no destructor)
#   pragma warning( pop )
#endif

    template<class T>
    MY_FORCE_INLINE
    void deallocate(T* p, const char* fn, int ln)
    {
        if(NULL==p)
            return;
        destruct(p);
        ::operator delete(static_cast<void*>(p),fn,ln);
    }

}

#if REPORT_LEAKS==1
#   define MY_MALLOC__(x)       ::my_malloc__(__FILE__,__LINE__,(x))
#else
#   define MY_MALLOC__(x)       ::malloc((x))
#endif

#define my_malloc(x)        MY_MALLOC__( (x) ? (x) : 1 )

#define my_s_malloc(x)          ::my_ex_malloc__(__FILE__,__LINE__,(x),true,false)
#define my_ex_malloc(x,s,sp)    ::my_ex_malloc__(__FILE__,__LINE__,(x),(s),(sp))

#define MY_REALLOC__(p,s)   ::my_realloc__(__FILE__,__LINE__,(p),(s))
#define my_realloc(p,s)     MY_REALLOC__((p),(s))

#define MY_FREE__(p)        ::my_free__(__FILE__,__LINE__,(p))
#define my_free(p)          MY_FREE__((p))

#define MY_NEW              ::new(__FILE__,__LINE__)

#define MY_DELETE(x)        ::my_mem_manager::deallocate((x),__FILE__,__LINE__)
#define MY_DELETE_A(x)      ::operator delete[]((x),__FILE__,__LINE__)

#include "../auto_ptr/auto_ptr.h"
#include <memory>               // to not to collide with our auto_ptr

#define MY_AUTO_PTR__   ::my_auto_ptr::auto_ptr
#define auto_ptr        MY_AUTO_PTR__

#endif

