/***************************************************************************
                          microcode.cpp  -  description
                             -------------------
    begin                : Thu Mar 16 2006
    copyright            : (C) 2006-2007 by Marian VooDooMan Meravy
    email                : ghostvoodooman -a.T= users (dot) sourceforge.net
 ***************************************************************************/

/* $Id: microcode.cpp,v 1.37 2008/07/02 09:35:48 ghostvoodooman Exp $ */

/***************************************************************************

VooDoo cIRCle - an IRC (ro)bot
Copyright (C) 2006-2008 by Marian VooDooMan Meravy
  (ghostvoodooman -a.T= users (dot) sourceforge.net)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************/

/*!
    \file
    \brief Provides microcode execution for functions
    \version $Id: microcode.cpp,v 1.37 2008/07/02 09:35:48 ghostvoodooman Exp $
*/

#if defined(HAVE_CONFIG_H)
#   include <config.h>
#endif

#include <string>
#include <vector>
#include <stack>
#include <list>
#include <map>

#include <typeinfo>
#include <stdlib.h>

#include "utils/memleak.h"
#include "utils/assert.h"
#include "utils.h"
#include "params.h"

#include "utils.h"
#include "microcode.h"
#include "var_index.h"
#include "var_except.h"
#include "utils/assert.h"
#include "microcode/c_micro_symbol_variable_table.h"
#include "var/c_var_unset.h"
#include "microcode/c_micro_error_func_internal_2.h"

#include "utils/precomp.h"

#if defined(_MSC_VER) || defined(__BORLANDC__)
#   pragma hdrstop
#endif

/*!
    \brief Performs delayed updates
    \version 1
    \date 2006-04-14
    \author VooDooMan
    \param st Evaluation stack
*/
void microcode_perform_updates(stack<c_eval_iteration>& st)
{
    if(st.top().updates.empty())
        return;
    map<string,int> cnt;
    for(deque<pair<var_id_t, string> >::reverse_iterator i2=st.top().updates.rbegin(); i2!=st.top().updates.rend(); ++i2) {
        map<string,int>::iterator i3=cnt.find((*i2).second);
        if(cnt.end()==i3) {
            cnt.insert(make_pair((*i2).second,1));
        } else {
            (*i3).second++;
        }
    }
    for(deque<pair<var_id_t, string> >::iterator i1=st.top().updates.begin(); i1!=st.top().updates.end(); ++i1) {
        map<string,int>::iterator i3=cnt.find((*i1).second);
        if(cnt.end()!=i3) {
            (*i3).second--;
            if((*i3).second)
                continue;
        }
        c_var_base* output=NULL;
        bool dummy_b;
        c_micro_error* error=NULL;
        if(!st.top().symbol_table_->get_var_table()->get((*i1).first,output,dummy_b,error)) {
            MY_DELETE(error);
            continue;
        }
        if(!output)
            output=MY_NEW c_var_unset();
        else {
            c_var_base* tmp=st.top().vars_->get((*i1).second,false);
            if(output==tmp)
                continue;
            output=output->copy();
        }
        c_index_base* idx=MY_NEW c_index_str((*i1).second);
        st.top().vars_->set(st.top().this_module_,idx,output,false,"(updating variables)");
    }
    st.top().updates.clear();
}

/*!
    \brief Updates variable
    \version 1
    \date 2006-04-28
    \author VooDooMan
    \param st Evaluation stack
    \param id ID of variable
    \param var_name Name of variable (beginning with '$' sign)
*/
void microcode_update_variable(stack<c_eval_iteration>& st, var_sym_index_t id, const string& var_name)
{
    c_var_base* output=NULL;
    bool dummy_b=false;
    c_micro_error* error=NULL;
    if(!st.top().symbol_table_->get_var_table()->get(id,output,dummy_b,error,true)) {
        MY_DELETE(error);
        return;
    }
    if(!output)
        return;
    c_var_base* tmp=st.top().vars_->get(var_name,false);
    if(output==tmp)
        return;

    output=output->copy();

    c_index_base* idx=MY_NEW c_index_str(var_name);
    st.top().vars_->set(st.top().this_module_,idx,output,false,"(updating variable)");
}

void log_symbol_table_err_fn(c_var_keeper::e_log_severity severity, const string& msg, void* data)
{
    if(c_var_keeper::LS_WARNING==severity)
        return;
    c_micro_error** error=(c_micro_error**)data;
    *error=MY_NEW c_micro_error_keeper(msg);
}

void log_fn_c_micro_func_call(c_var_keeper::e_log_severity severity, const string& msg, void* data)
{
    c_micro_func_call* m=(c_micro_func_call*)data;
    m->log_error=MY_NEW c_micro_error_func_internal_2(m->func_name_,msg);
    m->log_severity=severity;
}

string get_oper_name(e_micro_type oper)
{
    switch(oper) {
        case MICRO_NA:
            return (string)"(N/A)";
        case MICRO_POST_INCREMENT:
            return (string)"x++";
        case MICRO_POST_DECREMENT:
            return (string)"x--";
        case MICRO_PRE_INCREMENT:
            return (string)"++x";
        case MICRO_PRE_DECREMENT:
            return (string)"--x";
        case MICRO_PLUS:
            return (string)"+";
        case MICRO_MINUS:
            return (string)"-";
        case MICRO_MUL:
            return (string)"*";
        case MICRO_DIV:
            return (string)"/";
        case MICRO_NOT:
            return (string)"!";
        case MICRO_EQUAL:
            return (string)"==";
        case MICRO_NOT_EQUAL:
            return (string)"!=";
        case MICRO_IDENTICAL:
            return (string)"===";
        case MICRO_NOT_IDENTICAL:
            return (string)"!==";
        case MICRO_CONCAT:
            return (string)".";
        case MICRO_MOD:
            return (string)"%";
        case MICRO_BIT_AND:
            return (string)"&";
        case MICRO_BIT_OR:
            return (string)"|";
        case MICRO_BIT_XOR:
            return (string)"^";
        case MICRO_SHIFT_LEFT:
            return (string)"<<";
        case MICRO_SHIFT_RIGHT:
            return (string)">>";
        case MICRO_GREATER:
            return (string)">";
        case MICRO_LESS:
            return (string)"<";
        case MICRO_GREATER_OR_EQUAL:
            return (string)">=";
        case MICRO_LESS_OR_EQUAL:
            return (string)"<=";
        case MICRO_ONE_COMPLEMENT:
            return (string)"~";
        case MICRO_CAST_INT:
            return (string)"(int)";
        case MICRO_CAST_STRING:
            return (string)"(string)";
        case MICRO_CAST_BOOL:
            return (string)"(bool)";
        case MICRO_LOG_OR:
            return (string)"||";
        case MICRO_LOG_AND:
            return (string)"&&";
        case MICRO_TERNARY:
            return (string)"?:";
        case MICRO_FUNC_CALL:
            return (string)"f(...)";
        case MICRO_ARRAY:
            return (string)"array(...)";
        case MICRO_ISSET:
            return (string)"isset(...)";
        case MICRO_RETURN:
            return (string)"return";
        case MICRO_ASSIGN:
            return (string)"=";
        case MICRO_INDEX:
            return (string)"[...]";
        case MICRO_JUMP:
            return (string)"<JUMP>";
        case MICRO_LVALUE:
            return (string)"<SET_LVALUE>";
        case MICRO_MAKE_INDEX_STREAM:
            return (string)"[...][...]...";
        case MICRO_STR_EVAL:
            return (string)"<evaluation_of_double-quoted_string_with_dollar_sign>";
        default:
            break;
    }
    return (string)"<invalid_operator>";
}

