/***************************************************************************
                          c_micro.h  -  description
                             -------------------
    begin                : Thu May 10 2007
    copyright            : (C) 2004-2007 by Marian VooDooMan Meravy
    email                : ghostvoodooman -a.T= users (dot) sourceforge.net
 ***************************************************************************/

/* $Id: c_micro.h,v 1.14 2008/03/11 21:11:40 ghostvoodooman Exp $ */

/***************************************************************************

VooDoo cIRCle - an IRC (ro)bot
Copyright (C) 2004-2008 by Marian VooDooMan Meravy
  (ghostvoodooman -a.T= users (dot) sourceforge.net)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************/

/*!
    \file
    \brief Class 'c_micro'
    \version $Id: c_micro.h,v 1.14 2008/03/11 21:11:40 ghostvoodooman Exp $
*/

#ifndef C_MICRO_H__
#define C_MICRO_H__

#if defined(HAVE_CONFIG_H)
#   include <config.h>
#endif

#include <string>
#include <vector>
#include <stack>
#include <list>
#include <map>
#include <algorithm>

#include "../utils/memleak.h"

#include "../params.h"

#include "../microcode_shared.h"
#include "c_micro_symbol_table.h"

class c_micro_error;

/*!
    \brief Microcode class: base class
    \author VooDooMan
    \version 1
    \date 2006-03-16
*/
class c_micro
{
protected:
    bool output_must_be_lvalue;     //!< If to set l-value flag on an output variable, and if output is not UNSET, report an "l-value needed" error if it is not l-value already; (if applicable; used in cases of assignment operators)

public:

    /*!
        \brief Constructor
        \author VooDooMan
        \version 1
        \date 2006-03-16
    */
    c_micro()
    {
        output_must_be_lvalue=false;
    }

    /*!
        \brief Copy-constructor
        \author VooDooMan
        \version 1
        \date 2006-08-18
    */
    c_micro(const c_micro& orig)
    {
        output_must_be_lvalue=orig.output_must_be_lvalue;
    }

    /*!
        \brief Copy
        \author VooDooMan
        \version 1
        \date 2006-08-18
    */
    virtual c_micro* copy() = 0;

    /*!
        \brief Destructor
        \author VooDooMan
        \version 1
        \date 2006-03-16
    */
    virtual ~c_micro() { ; }

    /*!
        \brief Gets type of class
        \author VooDooMan
        \version 1
        \date 2006-03-16
    */
    virtual e_micro_type get_type() = 0;

    /*!
        \brief Performs microcode
        \author VooDooMan
        \version 1
        \date 2006-03-16
        \retval true Success
        \retval false Error
        \param symbols Symbol table object
        \param error If returned 'false', there is error reporter
        \param this_module Module
        \param debug Debug string
    */
    virtual bool perform(c_micro_symbol_table* symbols, c_micro_error*& error, c_module* this_module, const string& debug) = 0;

    void set_output_must_be_lvalue()
    {
        output_must_be_lvalue=true;
    }

    virtual var_sym_index_t get_output() = 0;

    /*!
        \brief Adds all inputs to vector
        \author VooDooMan
        \version 1
        \date 2007-03-18
        \param inputs Vector to add to
    */
    virtual void add_inputs(vector<var_sym_index_t>& inputs) = 0;

    /*!
        \brief Optimizes output by removing result variables taht are not referenced as input
        \date 2007-03-18
        \author VooDooMan
        \version 1
        \param inputs All input variables
    */
    virtual void optimize_output(const vector<var_sym_index_t>& inputs) = 0;
};

#endif

