/***************************************************************************
                          c_micro_cast_string.cpp  -  description
                             -------------------
    begin                : Thu May 10 2007
    copyright            : (C) 2004-2007 by Marian VooDooMan Meravy
    email                : ghostvoodooman -a.T= users (dot) sourceforge.net
 ***************************************************************************/

/* $Id: c_micro_cast_string.cpp,v 1.14 2008/03/11 21:11:40 ghostvoodooman Exp $ */

/***************************************************************************

VooDoo cIRCle - an IRC (ro)bot
Copyright (C) 2004-2008 by Marian VooDooMan Meravy
  (ghostvoodooman -a.T= users (dot) sourceforge.net)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************/

/*!
    \file
    \brief Class 'c_micro_cast_string'
    \version $Id: c_micro_cast_string.cpp,v 1.14 2008/03/11 21:11:40 ghostvoodooman Exp $
*/

#if defined(HAVE_CONFIG_H)
#   include <config.h>
#endif

#include "../params.h"
#include "../utils/memleak.h"
#include "../utils/assert.h"
#include "../compat.h"

#include "c_micro_cast_string.h"
#include "../var/c_var_base.h"
#include "c_micro_symbol_variable_table.h"
#include "c_micro_error_unset.h"
#include "c_micro_symbol_variable_mapping.h"
#include "../var_except.h"
#include "c_micro_error_var_except.h"
#include "../var/c_var_int.h"
#include "../var/c_var_string.h"
#include "../var/c_var_bool.h"
#include "../utils/precomp.h"
#include "c_micro_error_notstring.h"

#if defined(_MSC_VER) || defined(__BORLANDC__)
#   pragma hdrstop
#endif

bool c_micro_cast_string::perform(c_micro_symbol_table* symbols, c_micro_error*& error, c_module* this_module, const string& debug)
{
    c_var_base* value=NULL;
    bool lvalue;
    if(!symbols->get_var_table()->get(input_,value,lvalue,error))
        return false;
    if(!value || value->get_var_type()==VARTYPE_UNSET) {
        error=MY_NEW c_micro_error_unset(input_,get_type(),symbols->get_variable_mapping()->get_var_name(input_));
        return false;
    }

    bool created=false;

    if(value->get_var_type()==VARTYPE_INT) {
        c_var_int* i=dynamic_cast<c_var_int*>(value);
        value=MY_NEW c_var_string(i->to_string(this_module));
        created=true;
    }
    if(value->get_var_type()==VARTYPE_BOOL) {
        c_var_bool* b=dynamic_cast<c_var_bool*>(value);
        value=MY_NEW c_var_string(b->get()?"TRUE":"FALSE");
        created=true;
    }

    if(value->get_var_type()!=VARTYPE_STRING) {
        error=MY_NEW c_micro_error_notstring(input_,get_type(),symbols->get_variable_mapping()->get_var_name(input_),value->get_var_type_as_string());
        if(created)
            MY_DELETE(value);
        return false;
    }

    /*
    c_var_int* i=dynamic_cast<c_var_int*>(value);

    c_var_int* io=MY_NEW c_var_int(i->get());

    if(created)
        MY_DELETE(value);
    */

    if(symbols->get_var_table()->is_ro(output_)) {
        var_oops(this_module,__FILE__,__my_func__,GET_STR_DEBUG_INFO_EX("Attemp to write to read-only variable"),debug);
    }

    if(!symbols->get_var_table()->set(output_,/*io*/value,output_must_be_lvalue,error,get_type(),false,this_module,"",debug)) {
        //MY_DELETE(io);
        if(created)
            MY_DELETE(value);
        return false;
    }
    return true;
}

c_micro_cast_string::c_micro_cast_string(const c_micro_cast_string& orig)
    : c_micro(orig)
{
    input_=orig.input_;
    output_=orig.output_;
    output_must_be_lvalue=orig.output_must_be_lvalue;
}

c_micro* c_micro_cast_string::copy()
{
    return MY_NEW c_micro_cast_string(*this);
}

var_sym_index_t c_micro_cast_string::get_output()
{
    return output_;
}

void c_micro_cast_string::add_inputs(vector<var_sym_index_t>& inputs)
{
    inputs.push_back(input_);
}

void c_micro_cast_string::optimize_output(const vector<var_sym_index_t>& UNUSEDARG(inputs))
{
}

