/***************************************************************************
                          c_micro_func_call.cpp  -  description
                             -------------------
    begin                : Thu May 10 2007
    copyright            : (C) 2004-2007 by Marian VooDooMan Meravy
    email                : ghostvoodooman -a.T= users (dot) sourceforge.net
 ***************************************************************************/

/* $Id: c_micro_func_call.cpp,v 1.15 2008/06/08 08:43:13 ghostvoodooman Exp $ */

/***************************************************************************

VooDoo cIRCle - an IRC (ro)bot
Copyright (C) 2004-2008 by Marian VooDooMan Meravy
  (ghostvoodooman -a.T= users (dot) sourceforge.net)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************/

/*!
    \file
    \brief Class 'c_micro_func_call'
    \version $Id: c_micro_func_call.cpp,v 1.15 2008/06/08 08:43:13 ghostvoodooman Exp $
*/

#if defined(HAVE_CONFIG_H)
#   include <config.h>
#endif

#include "../params.h"
#include "../utils/memleak.h"
#include "../utils/assert.h"
#include "../compat.h"

#include "c_micro_func_call.h"
#include "../var/c_var_base.h"
#include "c_micro_symbol_variable_table.h"
#include "c_micro_error_unset.h"
#include "c_micro_symbol_variable_mapping.h"
#include "../var_except.h"
#include "c_micro_error_var_except.h"
#include "c_micro_error_func_argc.h"
#include "c_micro_error_func_builtin.h"
#include "c_micro_error_func_internal_1.h"
#include "c_micro_error_func_internal_3.h"
#include "c_micro_error_func_unknown.h"
#include "c_micro_error_lvalue.h"
#include "c_micro_error_func_argc.h"
#include "c_micro_error_func_builtin.h"
#include "../var/c_var_unset.h"
#include "../utils/precomp.h"

#if defined(_MSC_VER) || defined(__BORLANDC__)
#   pragma hdrstop
#endif

void microcode_perform_updates(stack<c_eval_iteration>& st);

c_micro_func_call::c_micro_func_call(const c_micro_func_call& orig)
    : c_micro(orig)
{
    func_name_=orig.func_name_;
    args_=orig.args_;
    output_=orig.output_;

    log_error=orig.log_error;
    log_severity=orig.log_severity;

    next_expression=orig.next_expression;
}

c_micro* c_micro_func_call::copy()
{
    return MY_NEW c_micro_func_call(*this);
}

bool c_micro_func_call::make_iteration(stack<c_eval_iteration>& st, c_micro_symbol_table* symbols, c_micro_error*& error, c_module* this_module, const string& debug, c_var_array* vars, const string& next_expression)
{
    error=NULL;

    // first: check for built-in function
    if(this_module->r_functions.is_built_in_function(func_name_)) {
        vector<c_func_arg> args1, args2;
        vector<var_sym_index_t>::iterator ai;
        for(ai=args_.begin(); ai!=args_.end(); ++ai) {
            c_func_arg a;
            bool dummy_b;
            if(!symbols->get_var_table()->get(*ai,a.value,dummy_b,error)) {
                return false;
            }
            a.arg_expression=symbols->get_var_table()->get_var_name(*ai);
            a.value_sym=*ai;
            a.reference=true;
            if(a.arg_expression.length()==0 || a.arg_expression[0]!='$')
                a.reference=false;
            args1.push_back(a);
            args2.push_back(a);
        }
        c_var_base* retval=NULL;
        string err_str;
        bool skip_block=false;
        //string next_expression="";
        if(!func_name_.compare("eval"))
            microcode_perform_updates(st);

        int return_to=st.top().state;

        if(!this_module->r_functions.call_built_in_func(this_module,func_name_,retval,args1,debug,err_str,skip_block,next_expression,vars,st)) {
            error=MY_NEW c_micro_error_func_builtin(func_name_,err_str);
            return false;
        }
        if(func_name_.compare("eval")) { // !=
            // update references
            vector<c_func_arg>::iterator ai1, ai2;
            for(ai=args_.begin(), ai1=args1.begin(), ai2=args2.begin(); ai!=args_.end(), ai1!=args1.end() && ai2!=args2.end(); ++ai, ++ai1, ++ai2) {
                if((*ai1).value==(*ai2).value)
                    continue;

                /*c_index_base* idx=MY_NEW c_index_str((*ai).arg_expression);
                vars->set(this_module,idx,(*ai).value,false,debug);*/

                if(!symbols->get_var_table()->set(*ai,(*ai1).value,true,error,MICRO_NA,false,this_module,"",debug)) {
                    return false;
                }
            }
            if(!retval)
                retval=MY_NEW c_var_unset();
            if(!symbols->get_var_table()->set(output_,retval,false,error,MICRO_FUNC_CALL,false,this_module,"",debug)) {
                return false;
            }
            st.top().result=output_;
            st.top().skip_block=skip_block;
        } else {
            if(retval) {
                if(!symbols->get_var_table()->set(output_,retval,false,error,MICRO_FUNC_CALL,false,this_module,"",debug)) {
                    return false;
                }
                st.top().result=output_;
                st.top().skip_block=skip_block;
                return false;
            }
            // new iteration
            st.top().retval_store_to=output_;
            st.top().return_to_state=return_to;
            return true;
        }
        return false;
    }

    microcode_perform_updates(st);

    // second: check for user's procedure
    {
        if(this_module->is_procedure(func_name_)) {
            c_procedure p;
            bool got=false;
            vector<c_procedure>::iterator i1;
            for(i1=this_module->r_procedures.begin(); i1!=this_module->r_procedures.end(); ++i1) {
                if((*i1).name_only.empty())
                    (*i1).parse_call_string();
                if(!(*i1).name_only.compare(func_name_)) {
                    p=*i1;
                    got=true;
                    break;
                }
            }
            if(got) {
                if(p.args.size()!=args_.size()) {
                    error=MY_NEW c_micro_error_func_argc(func_name_);
                    return false;
                }

                // child instance
                c_var_array* new_vars;
                unsigned new_local;
                unsigned old_local=var_keeper->set_current_local(new_local=var_keeper->register_local(new_vars,NULL,__FILE__,__my_func__,__LINE__));

                vector<string>::iterator i1;
                vector<var_sym_index_t>::iterator ai;
                for(i1=p.args.begin(), ai=args_.begin(); i1!=p.args.end() && ai!=args_.end(); ++i1, ++ai) {
                    c_index_base* idx=MY_NEW c_index_str(*i1);
                    c_var_base* val;
                    bool dummy_b;
                    if(!symbols->get_var_table()->get(*ai,val,dummy_b,error)) {
                        MY_DELETE(new_vars);
                        var_keeper->free_local(new_local,true);
                        var_keeper->set_current_local(old_local);
                        return false;
                    }
                    if(!val) {
                        if(st.top().vars_) {
                            val=st.top().vars_->get(st.top().this_module_,idx,false,debug);
                        }
                    }
                    if(val) {
                        val=val->copy();
                        val->set_read_only(false);
                        val->set_persistent(false);
                    }
                    new_vars->set(this_module,idx,val,false,debug);
                }

                c_eval_iteration ste(this_module,new_vars,debug);
                ste.state=4000;
                ste.return_to_state=st.top().state;
                ste.proc=p;
                ste.proc_iter=p.commands.begin();
                ste.proc_cmd=*(ste.proc_iter);
                ste.goto_line=-1;
                ste.after_redirect=false;
                ste.redirect_checkpoint=false;
                ste.do_redirect=false;
                ste.r_checkpoint=false;
                ste.return_value=NULL;
                ste.local_id=new_local;
                ste.parent=&(st.top());

                st.push(ste);
                return true;
            }
        }
    }

    // third: check for user's function
    {
        vector<c_func_arg> fn_args;
        if(this_module->r_functions.get_function_prototype(func_name_,fn_args)) {
            if(fn_args.size()!=args_.size()) {
                error=MY_NEW c_micro_error_func_argc(func_name_);
                return false;
            }
            c_function* new_fn;
            bool dummy_b=false;
            if(!this_module->get_function(func_name_,dummy_b,new_fn) || dummy_b) {
                error=MY_NEW c_micro_error_func_internal_1(func_name_);
                return false;
            }

            // child instance
            c_var_array* new_vars;
            unsigned new_local;
            unsigned old_local=var_keeper->set_current_local(new_local=var_keeper->register_local(new_vars,NULL,__FILE__,__my_func__,__LINE__));
            c_micro_symbol_table* new_symbol_table=new_fn->symbol_table;

            vector<c_func_arg> args1, args2;
            vector<var_sym_index_t>::iterator ai;
            vector<c_func_arg>::iterator afi;
            long arg_c=0;
            for(ai=args_.begin(), afi=fn_args.begin(); ai!=args_.end() && afi!=fn_args.end(); ++afi, ++ai, ++arg_c) {
                c_func_arg a;
                a.arg_expression=(*afi).arg_expression;
                a.reference=(*afi).reference;
                c_var_base* val;
                bool lvalue;
                if(!symbols->get_var_table()->get(*ai,val,lvalue,error)) {
                    MY_DELETE(new_vars);

                    MY_DELETE(new_fn);

                    var_keeper->free_local(new_local,true);
                    var_keeper->set_current_local(old_local);
                    return false;
                }
                if(!val) {
                    val=MY_NEW c_var_unset();
                    if(!symbols->get_var_table()->set(*ai,val,true,error,get_type(),false,this_module,"",debug)) {
                        MY_DELETE(new_vars);

                        MY_DELETE(new_fn);

                        var_keeper->free_local(new_local,true);
                        var_keeper->set_current_local(old_local);
                        return false;
                    }
                    if(!symbols->get_var_table()->get(*ai,val,lvalue,error)) {
                        MY_DELETE(new_vars);

                        MY_DELETE(new_fn);

                        var_keeper->free_local(new_local,true);
                        var_keeper->set_current_local(old_local);
                        return false;
                    }
                }
                if((*afi).reference && !lvalue) {
                    string var_name=symbols->get_variable_mapping()->get_var_name(*ai);
                    error=MY_NEW c_micro_error_lvalue(*ai,get_type(),var_name);

                    MY_DELETE(new_fn);

                    var_keeper->free_local(new_local,true);
                    var_keeper->set_current_local(old_local);
                    return false;
                }
                if(!(*afi).reference) {
                    if(val) {
                        val=val->copy();
                        val->set_read_only(false);
                        val->set_persistent(false);
                    }
                } else {
                    c_var_base* new_val=NULL;
                    if(val)
                        new_val=MY_NEW c_var_unset();
                    log_error=NULL;
                    if(val && new_val)
                        var_keeper->add_reference(new_val->get_id(),val->get_id(),&log_fn_c_micro_func_call,this);
                    if(log_error) {
                        if(c_var_keeper::LS_WARNING==log_severity) {
                            if(this_module->get_logger())
                                this_module->get_logger()->log_print1(E_LOG_DEBUG,LOG_SCOPE_DEFAULT,"",log_error->get_msg());

                            MY_DELETE(log_error);
                            log_error=NULL;
                        } else {
                            MY_DELETE(new_val);

                            MY_DELETE(new_vars);

                            MY_DELETE(new_fn);

                            error=log_error;
                            var_keeper->free_local(new_local,true);
                            var_keeper->set_current_local(old_local);
                            return false;
                        }
                    }
                    val=new_val;
                }
                if(!new_symbol_table->get_var_table()->set((*afi).value_sym,val,true,error,MICRO_NA,(*afi).reference,this_module,"",debug)) {
                    MY_DELETE(val);

                    MY_DELETE(new_vars);

                    MY_DELETE(new_fn);

                    var_keeper->free_local(new_local,true);
                    var_keeper->set_current_local(old_local);
                    return false;
                }
                c_index_base* idx=MY_NEW c_index_str((*afi).arg_expression);
                if(!new_vars->set(this_module,idx,val,false,debug)) {
                    MY_DELETE(val);

                    MY_DELETE(new_vars);

                    MY_DELETE(new_fn);

                    string msg="Error setting argument (zero-based) #";
                    msg+=tos(arg_c);
                    msg+=" to new function's internal variables";

                    error=MY_NEW c_micro_error_func_internal_3(func_name_,msg);
                    var_keeper->free_local(new_local,true);
                    var_keeper->set_current_local(old_local);
                    return false;
                }
                a.value=val;
                a.value_sym=(*afi).value_sym;
                args1.push_back(a);
                args2.push_back(a);
            }
            {
                c_eval_iteration ste(this_module,new_vars,debug);
                ste.state=1000;
                ste.return_to_state=st.top().state;
                ste.fn=new_fn;
                ste.fn->create_symbols();
                ste.fn_destroy=true;
                ste.instruct_id=0;
                ste.symbol_table_=new_symbol_table;
                ste.return_value=NULL;

                ste.retval_store_to=output_;

                ste.local_id=new_local;
                ste.parent=&(st.top());

                st.push(ste);
            }
            return true;
        }
    }

    // fourth: unknown function call
    error=MY_NEW c_micro_error_func_unknown(func_name_);
    return false;
}

var_sym_index_t c_micro_func_call::get_output()
{
    return output_;
}

void c_micro_func_call::add_inputs(vector<var_sym_index_t>& inputs)
{
    inputs.insert(inputs.end(),args_.begin(),args_.end());
}

void c_micro_func_call::optimize_output(const vector<var_sym_index_t>& UNUSEDARG(inputs))
{
    /*
    if(output_!=VAR_SYM_INDEX_NONE && output_!=VAR_SYM_INDEX_DISCARD &&
        output_!=VAR_SYM_INDEX_INVALID) {

        vector<var_sym_index_t>::const_iterator i1;
        for(i1=inputs.begin(); i1!=inputs.end(); ++i1) {
            if(*i1==output_)
                return;
        }
    } else
        return;

    // no reference
    output_=VAR_SYM_INDEX_DISCARD;
    */
}

