/***************************************************************************
                          c_micro_func_call.h  -  description
                             -------------------
    begin                : Thu May 10 2007
    copyright            : (C) 2004-2007 by Marian VooDooMan Meravy
    email                : ghostvoodooman -aT= users (dot) sourceforge.net
 ***************************************************************************/

/* $Id: c_micro_func_call.h,v 1.14 2008/03/11 21:11:44 ghostvoodooman Exp $ */

/***************************************************************************

VooDoo cIRCle - an IRC (ro)bot
Copyright (C) 2004-2008 by Marian VooDooMan Meravy
  (ghostvoodooman -aT= users (dot) sourceforge.net)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************/

/*!
    \file
    \brief Class 'c_micro_func_call'
    \version $Id: c_micro_func_call.h,v 1.14 2008/03/11 21:11:44 ghostvoodooman Exp $
*/

#ifndef C_MICRO_FUNC_CALL_H__
#define C_MICRO_FUNC_CALL_H__

#if defined(HAVE_CONFIG_H)
#   include <config.h>
#endif

#include <string>
#include <vector>
#include <stack>
#include <list>
#include <map>
#include <algorithm>

#include "../utils/memleak.h"

#include "../params.h"

#include "c_micro.h"
#include "../var_keeper.h"
#include "../runtime/c_eval_iteration.h"

/*!
    \brief Microcode class: function call
    \author VooDooMan
    \version 1
    \date 2006-03-22
*/
class c_micro_func_call : public c_micro
{
    string func_name_;                  //!< Name of function
    vector<var_sym_index_t> args_;      //!< Arguments, variable indexes
    var_sym_index_t output_;            //!< Output variable index

    c_micro_error* log_error;           //!< Logged error (from call-back)
    c_var_keeper::e_log_severity log_severity;     //!< Logged severity (from call-back)

    string next_expression;             //!< Next expression, for timer_*()

    friend void log_fn_c_micro_func_call(c_var_keeper::e_log_severity severity, const string& msg, void* data);

public:

    /*!
        \brief Constructor
        \author VooDooMan
        \version 1
        \date 2006-03-22
    */
    c_micro_func_call(const string& func_name)
        : c_micro()
        , func_name_(func_name)
        , output_(VAR_SYM_INDEX_INVALID)
    {
        ;
    }

    /*!
        \brief Copy-constructor
        \author VooDooMan
        \version 1
        \date 2006-08-18
    */
    c_micro_func_call(const c_micro_func_call& orig);

    /*!
        \brief Copy
        \author VooDooMan
        \version 1
        \date 2006-08-18
    */
    c_micro* copy();

    /*!
        \brief Destructor
        \author VooDooMan
        \version 1
        \date 2006-03-22
    */
    virtual ~c_micro_func_call() { ; }

    e_micro_type get_type()
    {
        return MICRO_FUNC_CALL;
    }

    /*!
        \brief Adds next argument to function call
        \author VooDooMan
        \version 1
        \date 2006-03-22
        \param id Argument variable ID
    */
    void add_arg(var_sym_index_t id)
    {
        args_.push_back(id);
    }

    /*!
        \brief Sets output variable ID (return value)
        \author VooDooMan
        \version 1
        \date 2006-03-22
        \param id Variable ID
    */
    void set_output(var_sym_index_t id)
    {
        output_=id;
    }

    /*!
        \brief Gets argument list
        \author VooDooMan
        \version 1
        \date 2006-03-22
    */
    const vector<var_sym_index_t>& get_args()
    {
        return args_;
    }

    /*!
        \brief Gets function name
        \author VooDooMan
        \version 1
        \date 2006-03-22
    */
    string get_func_name()
    {
        return func_name_;
    }

    bool perform(c_micro_symbol_table* UNUSEDARG(symbols), c_micro_error*& UNUSEDARG(error), c_module* UNUSEDARG(this_module), const string& UNUSEDARG(debug))
    {
        // do nothing. this will be handled in a different way.
        return true;
    }

    /*!
        \brief Performs function call, calls internal one (except 'eval'), or for external puts new iteration on stack
        \param st Evaluation stack
        \param symbols Symbol table
        \param error Check this after call, it could contain error reporting class instance on error, or to be 'NULL' for success of call
        \param this_module Module
        \param debug Debug string
        \param vars Local variables
        \param next_expression Next expression
        \retval true New iteration is on the stack
        \retval false It was internal function, so continue with next microcode in chain
    */
    bool make_iteration(stack<c_eval_iteration>& st, c_micro_symbol_table* symbols, c_micro_error*& error, c_module* this_module, const string& debug, c_var_array* vars, const string& next_expression);

    void set_next_expression(const string& exp)
    {
        next_expression=exp;
    }

    virtual var_sym_index_t get_output();

    void add_inputs(vector<var_sym_index_t>& inputs);
    void optimize_output(const vector<var_sym_index_t>& inputs);
};

#endif

