/***************************************************************************
                          c_micro_index.cpp  -  description
                             -------------------
    begin                : Thu May 10 2007
    copyright            : (C) 2004-2007 by Marian VooDooMan Meravy
    email                : ghostvoodooman -a.T= users (dot) sourceforge.net
 ***************************************************************************/

/* $Id: c_micro_index.cpp,v 1.14 2008/03/11 21:11:44 ghostvoodooman Exp $ */

/***************************************************************************

VooDoo cIRCle - an IRC (ro)bot
Copyright (C) 2004-2008 by Marian VooDooMan Meravy
  (ghostvoodooman -a.T= users (dot) sourceforge.net)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************/

/*!
    \file
    \brief Class 'c_micro_index'
    \version $Id: c_micro_index.cpp,v 1.14 2008/03/11 21:11:44 ghostvoodooman Exp $
*/

#if defined(HAVE_CONFIG_H)
#   include <config.h>
#endif

#include "../params.h"
#include "../utils/memleak.h"
#include "../utils/assert.h"
#include "../compat.h"

#include "c_micro_index.h"
#include "../var/c_var_base.h"
#include "../var/c_var_int.h"
#include "c_micro_symbol_variable_table.h"
#include "c_micro_error_unset.h"
#include "c_micro_symbol_variable_mapping.h"
#include "../var_except.h"
#include "c_micro_error_var_except.h"
#include "c_micro_error_index_base_invalid.h"
#include "c_micro_error_index_unset.h"
#include "c_micro_error_index_is_invalid.h"
#include "../var/c_var_unset.h"
#include "../var/c_var_string.h"
#include "c_micro_error_index_not_int.h"
#include "../utils/precomp.h"

#if defined(_MSC_VER) || defined(__BORLANDC__)
#   pragma hdrstop
#endif

bool c_micro_index::perform(c_micro_symbol_table* symbols, c_micro_error*& error, c_module* this_module, const string& debug)
{
    c_var_base* value1=NULL, *value2=NULL;
    bool lvalue;
    if(!symbols->get_var_table()->get(input1_,value1,lvalue,error))
        return false;
    if(input2_!=VAR_SYM_INDEX_NONE) {
        if(!symbols->get_var_table()->get(input2_,value2,lvalue,error))
            return false;
    }
    if(!value1 || value1->get_var_type()==VARTYPE_UNSET) {
        // create an array
        if(symbols->get_var_table()->is_ro(input1_)) {
            var_oops(this_module,__FILE__,__my_func__,GET_STR_DEBUG_INFO_EX("Attemp to write to read-only variable"),debug);
        } else {
            value1=MY_NEW c_var_array();
            if(!symbols->get_var_table()->set(input1_,value1,true,error,get_type(),false,this_module,"",debug)) {
                MY_DELETE(value1);
                return false;
            }
        }
        /*
        error=MY_NEW c_micro_error_unset(input1_,get_type(),symbols->get_variable_mapping()->get_var_name(input1_));
        return false;
        */
    }
    if(value1->get_var_type()!=VARTYPE_ARRAY && value1->get_var_type()!=VARTYPE_STRING) {
        error=MY_NEW c_micro_error_index_base_invalid(input1_,input2_,symbols->get_variable_mapping()->get_var_name(input1_),symbols->get_variable_mapping()->get_var_name(input2_));
        return false;
    }
    if(input2_!=VAR_SYM_INDEX_NONE) {
        if(!value2 || value2->get_var_type()==VARTYPE_UNSET) {
            error=MY_NEW c_micro_error_index_unset(input1_,input2_,symbols->get_variable_mapping()->get_var_name(input1_),symbols->get_variable_mapping()->get_var_name(input2_));
            return false;
        }
        if(value2->get_var_type()!=VARTYPE_INT && value2->get_var_type()!=VARTYPE_STRING) {
            error=MY_NEW c_micro_error_index_is_invalid(input1_,input2_,symbols->get_variable_mapping()->get_var_name(input1_),symbols->get_variable_mapping()->get_var_name(input2_));
            return false;
        }
        if(value1->get_var_type()==VARTYPE_STRING && value2->get_var_type()!=VARTYPE_INT) {
            error=MY_NEW c_micro_error_index_not_int(input1_,input2_,symbols->get_variable_mapping()->get_var_name(input1_),symbols->get_variable_mapping()->get_var_name(input2_));
            return false;
        }
    }

    c_index_base* value2_=value2 ? value2->make_index() : NULL;

    c_var_base* res=NULL;
    c_var_base* orig_res=NULL;

    bool set_lvalue;

    if(value1->get_var_type()==VARTYPE_ARRAY) {
        set_lvalue=true;

        c_var_array* a=dynamic_cast<c_var_array*>(value1);
        if(VAR_SYM_INDEX_NONE==input2_)
            value2_=a->get_new_int_index();
        orig_res=res=a->get(this_module,value2_,insensitive_,debug);
        if(res) {
            //res=res->copy();
            res=MY_NEW c_var_unset();
            //res->set_parent(a,value2);
            /*var_id_t r=*/var_keeper->add_reference(res->get_id(),orig_res->get_id(),NULL,NULL);
        } else {
            res=MY_NEW c_var_unset();
            a->set(this_module,value2_,res,insensitive_,debug);
            //res->set_parent(a,value2);

            c_var_base* orig_res=res;
            //res=res->copy();
            res=MY_NEW c_var_unset();
            //res->set_parent(a,value2);
            /*var_id_t r=*/var_keeper->add_reference(res->get_id(),orig_res->get_id(),NULL,NULL);
        }
    } else {
        // VARTYPE_STRING
        set_lvalue=false;

        c_var_string* s=dynamic_cast<c_var_string*>(value1);
        if(VAR_SYM_INDEX_NONE==input2_)
            value2=MY_NEW c_var_int(0);
        c_var_int* idx=dynamic_cast<c_var_int*>(value2);
        string str=s->get();
        if(VAR_SYM_INDEX_NONE==input2_)
            idx->set((long)str.length());
        if(idx->get()>=static_cast<MY_LONG_LONG>(str.length()) || idx->get()<0)
            res=MY_NEW c_var_unset();
        else
            res=MY_NEW c_var_string((string)""+str[static_cast<size_t>(idx->get())]);
    }

    if(symbols->get_var_table()->is_ro(output_,true)) {
        var_oops(this_module,__FILE__,__my_func__,GET_STR_DEBUG_INFO_EX("Attemp to write to read-only variable"),debug);
    }

    if(!symbols->get_var_table()->set(output_,res,set_lvalue,error,get_type(),true,this_module,"",debug)) {
        MY_DELETE(res);
        if(value2_)
            MY_DELETE(value2_);
        return false;
    }

    if(value2_)
        MY_DELETE(value2_);

    return true;
}

c_micro_index::c_micro_index(const c_micro_index& orig)
    : c_micro(orig)
{
    input1_=orig.input1_;
    input2_=orig.input2_;
    output_=orig.output_;
    insensitive_=orig.insensitive_;
}

c_micro* c_micro_index::copy()
{
    return MY_NEW c_micro_index(*this);
}

var_sym_index_t c_micro_index::get_output()
{
    return output_;
}

void c_micro_index::add_inputs(vector<var_sym_index_t>& inputs)
{
    inputs.push_back(input1_);
    inputs.push_back(input2_);
}

void c_micro_index::optimize_output(const vector<var_sym_index_t>& UNUSEDARG(inputs))
{
}

