/***************************************************************************
                          c_micro_index.h  -  description
                             -------------------
    begin                : Thu May 10 2007
    copyright            : (C) 2004-2007 by Marian VooDooMan Meravy
    email                : ghostvoodooman -a.T= users (dot) sourceforge.net
 ***************************************************************************/

/* $Id: c_micro_index.h,v 1.14 2008/03/11 21:11:44 ghostvoodooman Exp $ */

/***************************************************************************

VooDoo cIRCle - an IRC (ro)bot
Copyright (C) 2004-2008 by Marian VooDooMan Meravy
  (ghostvoodooman -a.T= users (dot) sourceforge.net)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************/

/*!
    \file
    \brief Class 'c_micro_index'
    \version $Id: c_micro_index.h,v 1.14 2008/03/11 21:11:44 ghostvoodooman Exp $
*/

#ifndef C_MICRO_INDEX_H__
#define C_MICRO_INDEX_H__

#if defined(HAVE_CONFIG_H)
#   include <config.h>
#endif

#include <string>
#include <vector>
#include <stack>
#include <list>
#include <map>
#include <algorithm>

#include "../utils/memleak.h"

#include "../params.h"

#include "c_micro.h"

/*!
    \brief Microcode class: index operator class
    \author VooDooMan
    \version 1
    \date 2006-03-31
*/
class c_micro_index : public c_micro
{
public:
    var_sym_index_t input1_;            //!< Base variable index
    var_sym_index_t input2_;            //!< Index variable index (that one between '[' and ']'
    var_sym_index_t output_;            //!< Output variable index
    bool insensitive_;                  //!< Case-insensitivity flag for STRING indexes

public:

    /*!
        \brief Constructor
        \author VooDooMan
        \version 1
        \date 2006-03-31
        \param input1 Base variable index
        \param input2 Index variable index (that one between '[' and ']'
        \param output Output variable index
        \param insensitive Case-insensitivity flag for STRING indexes
    */
    c_micro_index(var_sym_index_t input1, var_sym_index_t input2, var_sym_index_t output, bool insensitive)
        : c_micro()
        , input1_(input1)
        , input2_(input2)
        , output_(output)
        , insensitive_(insensitive)
    {
        ;
    }

    /*!
        \brief Copy-constructor
        \author VooDooMan
        \version 1
        \date 2006-08-18
    */
    c_micro_index(const c_micro_index& orig);

    /*!
        \brief Copy
        \author VooDooMan
        \version 1
        \date 2006-08-18
    */
    c_micro* copy();

    /*!
        \brief Destructor
        \author VooDooMan
        \version 1
        \date 2006-03-31
    */
    virtual ~c_micro_index() { ; }

    e_micro_type get_type()
    {
        return MICRO_INDEX;
    }

    bool perform(c_micro_symbol_table* symbols, c_micro_error*& error, c_module* this_module, const string& debug);

    virtual var_sym_index_t get_output();

    void add_inputs(vector<var_sym_index_t>& inputs);
    void optimize_output(const vector<var_sym_index_t>& inputs);
};

#endif

