/***************************************************************************
                          c_micro_log_and.cpp  -  description
                             -------------------
    begin                : Thu May 10 2007
    copyright            : (C) 2004-2007 by Marian VooDooMan Meravy
    email                : ghostvoodooman -a.T= users (dot) sourceforge.net
 ***************************************************************************/

/* $Id: c_micro_log_and.cpp,v 1.14 2008/03/11 21:11:44 ghostvoodooman Exp $ */

/***************************************************************************

VooDoo cIRCle - an IRC (ro)bot
Copyright (C) 2004-2008 by Marian VooDooMan Meravy
  (ghostvoodooman -a.T= users (dot) sourceforge.net)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************/

/*!
    \file
    \brief Class 'c_micro_log_and'
    \version $Id: c_micro_log_and.cpp,v 1.14 2008/03/11 21:11:44 ghostvoodooman Exp $
*/

#if defined(HAVE_CONFIG_H)
#   include <config.h>
#endif

#include "../params.h"
#include "../utils/memleak.h"
#include "../utils/assert.h"
#include "../compat.h"

#include "c_micro_log_and.h"
#include "../var/c_var_base.h"
#include "../var/c_var_bool.h"
#include "c_micro_symbol_variable_table.h"
#include "c_micro_error_unset.h"
#include "c_micro_symbol_variable_mapping.h"
#include "../utils/precomp.h"

#if defined(_MSC_VER) || defined(__BORLANDC__)
#   pragma hdrstop
#endif

bool c_micro_log_and::perform(c_micro_symbol_table* symbols, c_micro_error*& error, c_module* this_module, const string& debug)
{
    c_var_base* value1=NULL, *value2=NULL;
    bool lvalue;
    if(!symbols->get_var_table()->get(input1_,value1,lvalue,error))
        return false;
    if(!value1 || value1->get_var_type()==VARTYPE_UNSET) {
        error=MY_NEW c_micro_error_unset(input1_,get_type(),symbols->get_variable_mapping()->get_var_name(input1_));
        return false;
    }

    c_var_bool* bo=NULL;

    if(!value1->is_true()) {
        bo=MY_NEW c_var_bool(false);
    } else {
        if(!symbols->get_var_table()->get(input2_,value2,lvalue,error))
            return false;
        if(!value2 || value2->get_var_type()==VARTYPE_UNSET) {
            error=MY_NEW c_micro_error_unset(input2_,get_type(),symbols->get_variable_mapping()->get_var_name(input2_));
            return false;
        }

        bo=MY_NEW c_var_bool(value1->is_true() && value2->is_true());
    }

    if(symbols->get_var_table()->is_ro(output_)) {
        var_oops(this_module,__FILE__,__my_func__,GET_STR_DEBUG_INFO_EX("Attemp to write to read-only variable"),debug);
    }

    if(!symbols->get_var_table()->set(output_,bo,output_must_be_lvalue,error,get_type(),false,this_module,"",debug)) {
        MY_DELETE(bo);
        return false;
    }

    if(VAR_SYM_INDEX_INVALID!=output2_) {
        if(symbols->get_var_table()->is_ro(output2_)) {
            var_oops(this_module,__FILE__,__my_func__,GET_STR_DEBUG_INFO_EX("Attemp to write to read-only variable"),debug);
        }
        //bo=MY_NEW c_var_bool(*bo);
        bo=dynamic_cast<c_var_bool*>(bo->copy());
        if(!symbols->get_var_table()->set(output2_,bo,output_must_be_lvalue,error,get_type(),false,this_module,"",debug)) {
            MY_DELETE(bo);
            return false;
        }
    }

    return true;
}

bool c_micro_log_and::get_conditional(c_micro_symbol_table* symbols, c_micro_error*& error, bool& skip)
{
    c_var_base* value=NULL;
    bool lvalue;
    if(!symbols->get_var_table()->get(output_,value,lvalue,error) || !value)
        return false;

    skip=!value->is_true();

    return true;
}

bool c_micro_log_and::check_first(c_micro_symbol_table* symbols, c_micro_error*& error, bool& skip, c_module* this_module, const string& debug)
{
    c_var_base* value1=NULL, *value2=NULL;
    bool lvalue;
    if(!symbols->get_var_table()->get(input1_,value1,lvalue,error))
        return false;
    if(!value1 || value1->get_var_type()==VARTYPE_UNSET) {
        error=MY_NEW c_micro_error_unset(input1_,get_type(),symbols->get_variable_mapping()->get_var_name(input1_));
        return false;
    }

    skip=!value1->is_true();

    if(skip) {
        if(symbols->get_var_table()->is_ro(output_)) {
            var_oops(this_module,__FILE__,__my_func__,GET_STR_DEBUG_INFO_EX("Attemp to write to read-only variable"),debug);
        }

        c_var_bool* bo=MY_NEW c_var_bool(false);

        if(!symbols->get_var_table()->set(output_,bo,output_must_be_lvalue,error,get_type(),false,this_module,"",debug)) {
            MY_DELETE(bo);
            return false;
        }

        if(VAR_SYM_INDEX_INVALID!=output2_) {
            if(symbols->get_var_table()->is_ro(output2_)) {
                var_oops(this_module,__FILE__,__my_func__,GET_STR_DEBUG_INFO_EX("Attemp to write to read-only variable"),debug);
            }
            //bo=MY_NEW c_var_bool(*bo);
            bo=dynamic_cast<c_var_bool*>(bo->copy());
            if(!symbols->get_var_table()->set(output2_,bo,output_must_be_lvalue,error,get_type(),false,this_module,"",debug)) {
                MY_DELETE(bo);
                return false;
            }
        }

        return true;
    } else {
        if(!symbols->get_var_table()->get(input2_,value2,lvalue,error))
            return false;
        if(!value2 || value2->get_var_type()==VARTYPE_UNSET) {
            // nevermind!
            return true;
        }

        if(symbols->get_var_table()->is_ro(output_)) {
            var_oops(this_module,__FILE__,__my_func__,GET_STR_DEBUG_INFO_EX("Attemp to write to read-only variable"),debug);
        }

        c_var_bool* bo=MY_NEW c_var_bool(value1->is_true() && value2->is_true());

        if(!symbols->get_var_table()->set(output_,bo,output_must_be_lvalue,error,get_type(),false,this_module,"",debug)) {
            MY_DELETE(bo);
            return false;
        }

        if(VAR_SYM_INDEX_INVALID!=output2_) {
            if(symbols->get_var_table()->is_ro(output2_)) {
                var_oops(this_module,__FILE__,__my_func__,GET_STR_DEBUG_INFO_EX("Attemp to write to read-only variable"),debug);
            }
            //bo=MY_NEW c_var_bool(*bo);
            bo=dynamic_cast<c_var_bool*>(bo->copy());
            if(!symbols->get_var_table()->set(output2_,bo,output_must_be_lvalue,error,get_type(),false,this_module,"",debug)) {
                MY_DELETE(bo);
                return false;
            }
        }
    }

    return true;
}

bool c_micro_log_and::is_just_check_first()
{
    return just_check_first;
}

c_micro_log_and::c_micro_log_and(const c_micro_log_and& orig)
    : c_micro(orig)
{
    input1_=orig.input1_;
    input2_=orig.input2_;
    output_=orig.output_;
    output_must_be_lvalue=orig.output_must_be_lvalue;
    cond_eval_jump=orig.cond_eval_jump;
    just_check_first=orig.just_check_first;
    output2_=orig.output2_;
    outputs_=orig.outputs_;
}

c_micro* c_micro_log_and::copy()
{
    return MY_NEW c_micro_log_and(*this);
}

var_sym_index_t c_micro_log_and::get_output()
{
    return output_;
}

void c_micro_log_and::add_inputs(vector<var_sym_index_t>& inputs)
{
    inputs.push_back(input1_);
    inputs.push_back(input2_);
}

void c_micro_log_and::optimize_output(const vector<var_sym_index_t>& UNUSEDARG(inputs))
{
}

const vector<var_sym_index_t>& c_micro_log_and::get_all_outputs()
{
    return outputs_;
}

void c_micro_log_and::set_all_outputs(const vector<var_sym_index_t>& outputs)
{
    outputs_=outputs;
}

