/***************************************************************************
                          c_micro_log_or.h  -  description
                             -------------------
    begin                : Thu May 10 2007
    copyright            : (C) 2004-2007 by Marian VooDooMan Meravy
    email                : ghostvoodooman -a.T= users (dot) sourceforge.net
 ***************************************************************************/

/* $Id: c_micro_log_or.h,v 1.14 2008/03/11 21:11:44 ghostvoodooman Exp $ */

/***************************************************************************

VooDoo cIRCle - an IRC (ro)bot
Copyright (C) 2004-2008 by Marian VooDooMan Meravy
  (ghostvoodooman -a.T= users (dot) sourceforge.net)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************/

/*!
    \file
    \brief Class 'c_micro_log_or'
    \version $Id: c_micro_log_or.h,v 1.14 2008/03/11 21:11:44 ghostvoodooman Exp $
*/

#ifndef C_MICRO_LOG_OR_H__
#define C_MICRO_LOG_OR_H__

#if defined(HAVE_CONFIG_H)
#   include <config.h>
#endif

#include <string>
#include <vector>
#include <stack>
#include <list>
#include <map>
#include <algorithm>

#include "../utils/memleak.h"

#include "../params.h"

#include "c_micro.h"

/*!
    \brief Microcode class: logical OR class
    \author VooDooMan
    \version 1
    \date 2006-03-16, 2007-10-23
*/
class c_micro_log_or : public c_micro
{
    var_sym_index_t input1_;            //!< Left-hand side variable index
    var_sym_index_t input2_;            //!< Right-hand side variable index
    var_sym_index_t output_;            //!< Output variable index
    var_sym_index_t output2_;           //!< Parallel output variable index
    bool just_check_first;              //!< If only to check first value

    vector<var_sym_index_t> outputs_;   //!< All conditionals' outputs (in the same context)

    instruct_id_t cond_eval_jump;       //!< ID where to jump in case of conditional evaluation (operator returned TRUE)
public:

    /*!
        \brief Constructor
        \author VooDooMan
        \version 1
        \date 2006-03-16
        \param input1 Left-hand side variable index
        \param input2 Right-hand side variable index
        \param output Output variable index
        \param first_only If only to check first value
    */
    c_micro_log_or(var_sym_index_t input1, var_sym_index_t input2, var_sym_index_t output, bool first_only)
        : c_micro()
        , input1_(input1)
        , input2_(input2)
        , output_(output)
        , output2_(VAR_SYM_INDEX_INVALID)
        , just_check_first(first_only)
    {
        cond_eval_jump=INSTRUCT_ID_INVALID;
    }

    /*!
        \brief Copy-constructor
        \author VooDooMan
        \version 1
        \date 2006-08-18
    */
    c_micro_log_or(const c_micro_log_or& orig);

    /*!
        \brief Copy
        \author VooDooMan
        \version 1
        \date 2006-08-18
    */
    c_micro* copy();

    /*!
        \brief Destructor
        \author VooDooMan
        \version 1
        \date 2006-03-16
    */
    virtual ~c_micro_log_or() { ; }

    e_micro_type get_type()
    {
        return MICRO_LOG_OR;
    }

    /*!
        \brief Sets ID where to jump in case of conditional evaluation (operator returned TRUE)
        \author VooDooMan
        \version 1
        \date 2006-03-16
    */
    void set_cond_eval_jump(instruct_id_t id)
    {
        cond_eval_jump=id;
    }

    /*!
        \brief Gets ID where to jump in case of conditional evaluation (operator returned TRUE)
        \author VooDooMan
        \version 1
        \date 2006-03-16
    */
    instruct_id_t get_cond_eval_jump()
    {
        return cond_eval_jump;
    }

    bool perform(c_micro_symbol_table* symbols, c_micro_error*& error, c_module* this_module, const string& debug);

    /*!
        \brief Checks for conditional evaluation
        \author VooDooMan
        \version 1
        \date 2006-03-16
        \retval true Success
        \retval false Error
        \param symbols Symbol table object
        \param error If returned 'false', there is error reporter
        \param skip If returned 'true', this returns if we should skip rest of expression
    */
    bool get_conditional(c_micro_symbol_table* symbols, c_micro_error*& error, bool& skip);

    /*!
        \brief Checks for first value only
        \author VooDooMan
        \version 1
        \date 2006-09-28
        \retval true Success
        \retval false Error
        \param symbols Symbol table object
        \param error If returned 'false', there is error reporter
        \param skip If returned 'true', this returns if we should skip rest of expression
        \param this_module Module
        \param debug Debug string
    */
    bool check_first(c_micro_symbol_table* symbols, c_micro_error*& error, bool& skip, c_module* this_module, const string& debug);

    /*!
        \brief Checks if this is only for checking of conditional evaluation
        \author VooDooMan
        \version 1
        \date 2006-09-28
        \retval true Yes
        \retval false No
    */
    bool is_just_check_first();

    /*!
        \brief Sets output variable
        \date 2006-09-28
        \version 1
        \author VooDooMan
        \param output Output variable ID
    */
    void set_output(var_sym_index_t output)
    {
        if(VAR_SYM_INDEX_INVALID==output_)
            output_=output;
        else
            output2_=output;
    }

    virtual var_sym_index_t get_output();

    void add_inputs(vector<var_sym_index_t>& inputs);
    void optimize_output(const vector<var_sym_index_t>& inputs);

    /*!
        \brief Returns outputs from all conditionals' output in the same context
        \date 2007-10-23
        \version 1
        \author VooDooMan
    */
    const vector<var_sym_index_t>& get_all_outputs();

    /*!
        \brief Adds outputs to all conditionals' output in the same context
        \date 2007-10-23
        \version 1
        \author VooDooMan
        \param outputs Output variable IDs
    */
    void set_all_outputs(const vector<var_sym_index_t>& outputs);
};

#endif

