/***************************************************************************
                          c_micro_make_index_stream.cpp  -  description
                             -------------------
    begin                : Thu May 10 2007
    copyright            : (C) 2004-2007 by Marian VooDooMan Meravy
    email                : ghostvoodooman -a.T= users (dot) sourceforge.net
 ***************************************************************************/

/* $Id: c_micro_make_index_stream.cpp,v 1.14 2008/03/11 21:11:44 ghostvoodooman Exp $ */

/***************************************************************************

VooDoo cIRCle - an IRC (ro)bot
Copyright (C) 2004-2008 by Marian VooDooMan Meravy
  (ghostvoodooman -a.T= users (dot) sourceforge.net)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************/

/*!
    \file
    \brief Class 'c_micro_make_index_stream'
    \version $Id: c_micro_make_index_stream.cpp,v 1.14 2008/03/11 21:11:44 ghostvoodooman Exp $
*/

#if defined(HAVE_CONFIG_H)
#   include <config.h>
#endif

#include "../params.h"
#include "../utils/memleak.h"
#include "../utils/assert.h"
#include "../compat.h"

#include "c_micro_make_index_stream.h"
#include "../var/c_var_base.h"
#include "../var/c_var_array.h"
#include "c_micro_symbol_variable_table.h"
#include "c_micro_error_index_base_invalid.h"
#include "c_micro_symbol_variable_mapping.h"
#include "c_micro_error_index_unset.h"
#include "../utils/precomp.h"

#if defined(_MSC_VER) || defined(__BORLANDC__)
#   pragma hdrstop
#endif

bool c_micro_make_index_stream::perform(c_micro_symbol_table* symbols, c_micro_error*& error, c_module* this_module, const string& debug)
{
    c_var_base* value=NULL;
    bool lvalue;

    if(!symbols->get_var_table()->get(base_,value,lvalue,error))
        return false;
    if(value && value->get_var_type()==VARTYPE_STRING)
        return true;
    if(!value || value->get_var_type()!=VARTYPE_ARRAY) {
        value=MY_NEW c_var_array();
        if(!symbols->get_var_table()->set(base_,value,true,error,get_type(),false,this_module,"",debug)) {
            MY_DELETE(value);
            return false;
        }
    }

    c_var_array* intermed_base=NULL;
    if(value && typeid(*value)==typeid(c_var_array))
        intermed_base=dynamic_cast<c_var_array*>(value);
    for(vector<pair<var_sym_index_t,bool> >::iterator i1=indexed_var_stream_.begin(); i1!=indexed_var_stream_.end(); ++i1) {
        c_var_base* value2=NULL;
        c_index_base* value2_=NULL;
        bool value2_created=false;
        bool lvalue;
        var_sym_index_t new_index=(*i1).first;
        if(VAR_SYM_INDEX_NONE==new_index) {
            value2_=intermed_base->get_new_int_index();
            new_index=symbols->get_var_table()->add_dynamic(error);
            if(error) {
                MY_DELETE(value2);
                MY_DELETE(value2_);
                return false;
            }
            value2_created=true;
        } else {
            if(!symbols->get_var_table()->get(new_index,value2,lvalue,error))
                return false;
            if(!value2 || value2->get_var_type()==VARTYPE_UNSET) {
                error=MY_NEW c_micro_error_index_unset(base_,new_index,symbols->get_variable_mapping()->get_var_name(base_),symbols->get_variable_mapping()->get_var_name(new_index));
                return false;
            }
            if(!(value2->get_var_type()==VARTYPE_INT || value2->get_var_type()==VARTYPE_STRING)) {
                error=MY_NEW c_micro_error_index_base_invalid(base_,new_index,symbols->get_variable_mapping()->get_var_name(base_),symbols->get_variable_mapping()->get_var_name(new_index));
                return false;
            }
            value2_=value2->make_index();
        }
        c_var_base* intermed=intermed_base->get(this_module,value2_,(*i1).second,debug);
        bool last_index=(i1+1)==indexed_var_stream_.end();
        if(!intermed) {
            if(last_index) {
                /*
                intermed_base->set(this_module,value2_created?value2:value2->copy(),MY_NEW c_var_unset(),(*i1).second,debug);
                value2_created=false;
                */
            } else
                intermed=MY_NEW c_var_array();
            /*
            if(!symbols->get_var_table()->set(new_index,intermed,true,error,get_type(),false,this_module,"",debug)) {
                MY_DELETE(intermed);
                return false;
            }
            */
            if(!last_index)
                intermed_base->set(this_module,value2_,intermed,(*i1).second,debug);
            else {
                if(value2_created) {
                    MY_DELETE(value2);
                    MY_DELETE(value2_);
                    value2=NULL;
                    value2_=NULL;
                }
            }
        } else {
            if(last_index)
                ; // okay
            else {
                if(intermed->get_var_type()!=VARTYPE_ARRAY) {
                    intermed=MY_NEW c_var_array();
                    /*
                    if(!symbols->get_var_table()->set(new_index,intermed,true,error,get_type(),false,this_module,"",debug)) {
                        MY_DELETE(intermed);
                        return false;
                    }
                    */
                    intermed_base->set(this_module,value2_,intermed,(*i1).second,debug);
                }
            }
        }
        if(!last_index) {
            intermed_base=NULL;
            if(intermed && typeid(*intermed)==typeid(c_var_array))
                intermed_base=dynamic_cast<c_var_array*>(intermed);
        }
        if(value2_)
            MY_DELETE(value2_);
    }

    /*
    if(!symbols->get_var_table()->set(base_,value,true,error,get_type(),false,this_module,"",debug)) {
        return false;
    }
    */
    return true;
}

c_micro_make_index_stream::c_micro_make_index_stream(const c_micro_make_index_stream& orig)
    : c_micro(orig)
{
    base_=orig.base_;
    indexed_var_stream_=orig.indexed_var_stream_;
}

c_micro* c_micro_make_index_stream::copy()
{
    return MY_NEW c_micro_make_index_stream(*this);
}

var_sym_index_t c_micro_make_index_stream::get_output()
{
    return VAR_SYM_INDEX_NONE;
}

void c_micro_make_index_stream::add_inputs(vector<var_sym_index_t>& inputs)
{
    inputs.push_back(base_);
    vector<pair<var_sym_index_t,bool> >::iterator i1;
    for(i1=indexed_var_stream_.begin(); i1!=indexed_var_stream_.end(); ++i1) {
        inputs.push_back((*i1).first);
    }
}

void c_micro_make_index_stream::optimize_output(const vector<var_sym_index_t>& UNUSEDARG(inputs))
{
}

