/***************************************************************************
                          c_micro_make_index_stream.h  -  description
                             -------------------
    begin                : Thu May 10 2007
    copyright            : (C) 2004-2007 by Marian VooDooMan Meravy
    email                : ghostvoodooman -a.T= users (dot) sourceforge.net
 ***************************************************************************/

/* $Id: c_micro_make_index_stream.h,v 1.14 2008/03/11 21:11:44 ghostvoodooman Exp $ */

/***************************************************************************

VooDoo cIRCle - an IRC (ro)bot
Copyright (C) 2004-2008 by Marian VooDooMan Meravy
  (ghostvoodooman -a.T= users (dot) sourceforge.net)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************/

/*!
    \file
    \brief Class 'c_micro_make_index_stream'
    \version $Id: c_micro_make_index_stream.h,v 1.14 2008/03/11 21:11:44 ghostvoodooman Exp $
*/

#ifndef C_MICRO_MAKE_INDEX_STREAM_H__
#define C_MICRO_MAKE_INDEX_STREAM_H__

#if defined(HAVE_CONFIG_H)
#   include <config.h>
#endif

#include <string>
#include <vector>
#include <stack>
#include <list>
#include <map>
#include <algorithm>

#include "../utils/memleak.h"

#include "../params.h"

#include "c_micro.h"

/*!
    \brief Microcode class: creates streamed indexed variable tree, if it deos not exist yet (if it parts of path are UNSET), e.g. for things like this: $foo[1][2][3][4][][][]["aBcDe"/i][]="bar"
    \author VooDooMan
    \version 1
    \date 2006-04-07
*/
class c_micro_make_index_stream : public c_micro
{
public:
    var_sym_index_t base_;              //!< Base variable index/output
    vector<pair<var_sym_index_t,bool> > indexed_var_stream_;    //!< Stream of multiple indexes (e.g. $foo[1][22][3][][][][]="bar"); first==ID, second==index case-insensitivity flag

public:

    /*!
        \brief Constructor
        \author VooDooMan
        \version 1
        \date 2006-04-37
        \param base Base variable index/output
        \param indexed_var_stream Stream of multiple indexes (e.g. $foo[1][22][3][][][][]="bar"); first==ID, second==index case-insensitivity flag
    */
    c_micro_make_index_stream(var_sym_index_t base, vector<pair<var_sym_index_t,bool> > indexed_var_stream)
        : c_micro(), base_(base), indexed_var_stream_(indexed_var_stream)
    {
        ;
    }

    /*!
        \brief Copy-constructor
        \author VooDooMan
        \version 1
        \date 2006-08-18
    */
    c_micro_make_index_stream(const c_micro_make_index_stream& orig);

    /*!
        \brief Copy
        \author VooDooMan
        \version 1
        \date 2006-08-18
    */
    c_micro* copy();

    /*!
        \brief Destructor
        \author VooDooMan
        \version 1
        \date 2006-04-07
    */
    virtual ~c_micro_make_index_stream() { ; }

    e_micro_type get_type()
    {
        return MICRO_MAKE_INDEX_STREAM;
    }

    bool perform(c_micro_symbol_table* symbols, c_micro_error*& error, c_module* this_module, const string& debug);

    virtual var_sym_index_t get_output();

    void add_inputs(vector<var_sym_index_t>& inputs);
    void optimize_output(const vector<var_sym_index_t>& inputs);
};

#endif

