/***************************************************************************
                          c_micro_str_eval.h  -  description
                             -------------------
    begin                : Thu May 10 2007
    copyright            : (C) 2004-2007 by Marian VooDooMan Meravy
    email                : ghostvoodooman -a.T= users (dot) sourceforge.net
 ***************************************************************************/

/* $Id: c_micro_str_eval.h,v 1.14 2008/03/11 21:11:45 ghostvoodooman Exp $ */

/***************************************************************************

VooDoo cIRCle - an IRC (ro)bot
Copyright (C) 2004-2008 by Marian VooDooMan Meravy
  (ghostvoodooman -a.T= users (dot) sourceforge.net)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************/

/*!
    \file
    \brief Class 'c_micro_str_eval'
    \version $Id: c_micro_str_eval.h,v 1.14 2008/03/11 21:11:45 ghostvoodooman Exp $
*/

#ifndef C_MICRO_STR_EVAL_H__
#define C_MICRO_STR_EVAL_H__

#if defined(HAVE_CONFIG_H)
#   include <config.h>
#endif

#include <string>
#include <vector>
#include <stack>
#include <list>
#include <map>
#include <algorithm>

#include "../utils/memleak.h"

#include "../params.h"

#include "c_micro.h"

/*!
    \brief Microcode class: evaluates string in double quotes
    \author VooDooMan
    \version 1
    \date 2006-04-28
*/
class c_micro_str_eval : public c_micro
{
    string input_;                      //!< Input string
    var_sym_index_t output_;            //!< Output variable index
public:

    /*!
        \brief Constructor
        \author VooDooMan
        \version 1
        \date 2006-04-28
        \param input Input string
        \param output Output variable ID
    */
    c_micro_str_eval(const string& input, var_sym_index_t output)
        : c_micro(), input_(input), output_(output)
    {
    }

    /*!
        \brief Copy-constructor
        \author VooDooMan
        \version 1
        \date 2006-08-18
    */
    c_micro_str_eval(const c_micro_str_eval& orig);

    /*!
        \brief Copy
        \author VooDooMan
        \version 1
        \date 2006-08-18
    */
    c_micro* copy();

    /*!
        \brief Destructor
        \author VooDooMan
        \version 1
        \date 2006-04-28
    */
    virtual ~c_micro_str_eval() { ; }

    e_micro_type get_type()
    {
        return MICRO_STR_EVAL;
    }

    bool perform(c_micro_symbol_table* symbols, c_micro_error*& error, c_module* this_module, const string& debug);

    /*!
        \brief Retrieves input string
        \version 1
        \date 2006-04-28
        \author VooDooMan
        \return Returns string that needs to be evaluated (e.g. "abc $bar def")
    */
    const string& get_input()
    {
        return input_;
    }

    virtual var_sym_index_t get_output();

    void add_inputs(vector<var_sym_index_t>& inputs);
    void optimize_output(const vector<var_sym_index_t>& inputs);
};

#endif

