/***************************************************************************
                          c_micro_symbol_table.cpp  -  description
                             -------------------
    begin                : Thu May 10 2007
    copyright            : (C) 2004-2007 by Marian VooDooMan Meravy
    email                : ghostvoodooman -a.T= users (dot) sourceforge.net
 ***************************************************************************/

/* $Id: c_micro_symbol_table.cpp,v 1.16 2008/06/08 08:43:13 ghostvoodooman Exp $ */

/***************************************************************************

VooDoo cIRCle - an IRC (ro)bot
Copyright (C) 2004-2008 by Marian VooDooMan Meravy
  (ghostvoodooman -a.T= users (dot) sourceforge.net)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************/

/*!
    \file
    \brief Class 'c_micro_symbol_table'
    \version $Id: c_micro_symbol_table.cpp,v 1.16 2008/06/08 08:43:13 ghostvoodooman Exp $
*/

#if defined(HAVE_CONFIG_H)
#   include <config.h>
#endif

#include "../params.h"
#include "../utils/memleak.h"
#include "../utils/assert.h"
#include "../compat.h"

#include "c_micro.h"
#include "c_micro_symbol_variable_mapping.h"
#include "../var/c_var_unset.h"
#include "../utils/precomp.h"

#if defined(_MSC_VER) || defined(__BORLANDC__)
#   pragma hdrstop
#endif

void c_micro_symbol_table::update_symbols_from_vars(c_var_array* vars, bool reference)
{
    map<var_sym_index_t,string>::const_iterator i1;
    for(i1=var_mapping_->get_var_mapping().begin(); i1!=var_mapping_->get_var_mapping().end(); ++i1) {
        if((*i1).second.empty() || (*i1).second[0]!='$')
            continue;
        c_var_base* val=vars->get((*i1).second,false);
        c_micro_error* error=NULL;
        if(val) {
            c_var_base* orig=val;
            if(reference)
                val=MY_NEW c_var_unset();
            else
                val=orig->copy();
            if(reference)
                var_keeper->add_reference(val->get_id(),orig->get_id(),NULL,NULL);
            var_table_->set((*i1).first,val,true,error,MICRO_ASSIGN,false,core_mod,"","");
            if(error)
                MY_DELETE(error);
        }
    }
}

void c_micro_symbol_table::update_vars_from_symbols(c_module* this_module, c_var_array* vars)
{
    map<var_sym_index_t,string>::const_iterator i1;
    for(i1=var_mapping_->get_var_mapping().begin(); i1!=var_mapping_->get_var_mapping().end(); ++i1) {
        if((*i1).second.empty() || (*i1).second[0]!='$')
            continue;
        c_var_base* val=NULL;
        bool dummy;
        c_micro_error* error=NULL;
        var_table_->get((*i1).first,val,dummy,error);
        if(error)
            MY_DELETE(error);
        if(val) {
            val=val->copy();
            c_index_base* idx=MY_NEW c_index_str((*i1).second);
            vars->set(this_module,idx,val,false,"");
        }
    }
}

c_micro_symbol_table::c_micro_symbol_table(c_micro_symbol_variable_mapping* var_mapping,
    c_micro_symbol_variable_table* var_table)
{
    var_mapping_=var_mapping;
    var_table_=var_table;
    var_table_->set_var_mapping(var_mapping_);
    c_micro_error* error=NULL;
    for(map<var_sym_index_t, c_var_base*>::iterator ci=var_mapping_->consts.begin(); ci!=var_mapping_->consts.end(); ++ci) {
        c_var_base* constant=(*ci).second;
        if(constant)
            constant=constant->copy();
        else
            constant=MY_NEW c_var_unset();
        constant->set_persistent(false);
        constant->set_read_only(false);
        var_table_->set((*ci).first,constant,false,error,MICRO_NA,false,core_mod,"","");
    }
}

c_micro_symbol_table::~c_micro_symbol_table()
{
    if(NULL!=var_mapping_)
        var_mapping_->mapping.clear();
    MY_DELETE(var_mapping_);
    MY_DELETE(var_table_);
}

c_micro_symbol_table::c_micro_symbol_table(const c_micro_symbol_table& orig)
{
    var_table_=MY_NEW c_micro_symbol_variable_table(*orig.var_table_);
    var_mapping_=NULL;
    if(orig.var_mapping_)
        var_mapping_=MY_NEW c_micro_symbol_variable_mapping(*orig.var_mapping_,var_table_);
}

c_micro_symbol_variable_mapping* c_micro_symbol_table::get_variable_mapping()
{
    return var_mapping_;
}

