/***************************************************************************
                          c_micro_symbol_table.h  -  description
                             -------------------
    begin                : Thu May 10 2007
    copyright            : (C) 2004-2007 by Marian VooDooMan Meravy
    email                : ghostvoodooman -a.T= users (dot) sourceforge.net
 ***************************************************************************/

/* $Id: c_micro_symbol_table.h,v 1.14 2008/03/11 21:11:45 ghostvoodooman Exp $ */

/***************************************************************************

VooDoo cIRCle - an IRC (ro)bot
Copyright (C) 2004-2008 by Marian VooDooMan Meravy
  (ghostvoodooman -a.T= users (dot) sourceforge.net)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************/

/*!
    \file
    \brief Class 'c_micro_symbol_table'
    \version $Id: c_micro_symbol_table.h,v 1.14 2008/03/11 21:11:45 ghostvoodooman Exp $
*/

#ifndef C_MICRO_SYMBOL_TABLE_H__
#define C_MICRO_SYMBOL_TABLE_H__

#if defined(HAVE_CONFIG_H)
#   include <config.h>
#endif

#include <string>
#include <vector>
#include <stack>
#include <list>
#include <map>
#include <algorithm>

#include "../utils/memleak.h"

#include "../params.h"

class c_micro_symbol_variable_mapping;
class c_micro_symbol_variable_table;
class c_var_array;
class c_module;

/*!
    \brief Stores full symbol table for microcode
    \author VooDooMan
    \version 1
    \date 2006-03-16
*/
class c_micro_symbol_table
{
    c_micro_symbol_variable_mapping* var_mapping_;  //!< Mapping of variables from mircocode compilation
    c_micro_symbol_variable_table* var_table_;      //!< Real variables

    friend class c_function;

public:

    /*!
        \brief Constructor
        \author VooDooMan
        \version 1
        \date 2006-03-16
        \param var_mapping Variable mapping
        \param var_table Variable table
    */
    c_micro_symbol_table(c_micro_symbol_variable_mapping* var_mapping,
        c_micro_symbol_variable_table* var_table);

    /*!
        \brief Destructor
        \author VooDooMan
        \version 1
        \date 2006-04-16
    */
    virtual ~c_micro_symbol_table();

    c_micro_symbol_table(const c_micro_symbol_table& orig);

    /*!
        \brief Gets variable mapping object
        \author VooDooMan
        \version 1
        \date 2006-03-16
    */
    c_micro_symbol_variable_mapping* get_variable_mapping();

    /*!
        \brief Gets variable variable table object
        \author VooDooMan
        \version 1
        \date 2006-03-16
    */
    MY_FORCE_INLINE c_micro_symbol_variable_table* get_var_table()
    {
        return var_table_;
    }

    /*!
        \brief Updates symbol values from variables (a.k.a. proxy)
        \author VooDooMan
        \version 1
        \date 2006-06-23
        \param vars Variables
        \param reference If to make reference form new variable to old
    */
    void update_symbols_from_vars(c_var_array* vars, bool reference = false);

    /*!
        \brief Updates variables from symbols (a.k.a. proxy)
        \author VooDooMan
        \version 1
        \date 2006-06-23
        \param this_module Module
        \param vars Variables
    */
    void update_vars_from_symbols(c_module* this_module, c_var_array* vars);
};

#endif

