/***************************************************************************
                          c_micro_symbol_variable_mapping.cpp  -  description
                             -------------------
    begin                : Thu May 10 2007
    copyright            : (C) 2004-2007 by Marian VooDooMan Meravy
    email                : ghostvoodooman -a.T= users (dot) sourceforge.net
 ***************************************************************************/

/* $Id: c_micro_symbol_variable_mapping.cpp,v 1.16 2008/11/24 12:09:46 ghostvoodooman Exp $ */

/***************************************************************************

VooDoo cIRCle - an IRC (ro)bot
Copyright (C) 2004-2008 by Marian VooDooMan Meravy
  (ghostvoodooman -a.T= users (dot) sourceforge.net)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************/

/*!
    \file
    \brief Class 'c_micro_symbol_variable_mapping'
    \version $Id: c_micro_symbol_variable_mapping.cpp,v 1.16 2008/11/24 12:09:46 ghostvoodooman Exp $
*/

#if defined(HAVE_CONFIG_H)
#   include <config.h>
#endif

#include "../params.h"
#include "../utils/memleak.h"
#include "../utils/assert.h"
#include "../compat.h"

#include "c_micro.h"
#include "c_micro_symbol_variable_mapping.h"
#include "../func/c_expression.h"
#include "c_micro_str_eval.h"
#include "c_micro_commit_vars.h"
#include "../script_consts/script_consts.h"

#include "../utils/precomp.h"

#if defined(_MSC_VER) || defined(__BORLANDC__)
#   pragma hdrstop
#endif

void c_micro_symbol_variable_mapping::reset()
{
    mapping.clear();
    consts.clear();
    next_index=1;
}

string c_micro_symbol_variable_mapping::get_var_name(var_sym_index_t index)
{
    if(index<=0 || index>=next_index)
        return (string)"<index_out_of_range>";
    map<var_sym_index_t, string>::iterator i1=mapping.find(index);
    if(mapping.end()==i1)
        return (string)"<non-existing_index>";
    if((*i1).second.empty())
        return (string)"<temporary_variable>";
    return (*i1).second;
}

var_sym_index_t c_micro_symbol_variable_mapping::get_var_index(const string& var_name)
{
    if(var_name.empty() || '$'!=var_name[0])
        return VAR_SYM_INDEX_INVALID;
    map<var_sym_index_t, string>::iterator i1;
    for(i1=mapping.begin(); i1!=mapping.end(); ++i1)
        if(!(*i1).second.compare(var_name))
            return (*i1).first;
    return VAR_SYM_INDEX_INVALID;
}

var_sym_index_t c_micro_symbol_variable_mapping::get_capacity()
{
    return next_index-1+1;
}

const map<var_sym_index_t, string>& c_micro_symbol_variable_mapping::get_var_mapping()
{
    return mapping;
}

/*!
    \brief Adds new mapping (if it does not exist yet)
    \author VooDooMan
    \version 4
    \date 2006-03-16, 2006-04-28, 2007-09-11, 2007-09-29
    \param var_name Variable name
    \param exp Expression used for 'add_micro()' method if there is evaluation needed
    \return Returns new index; or existing one, if it already exists
    \warning If an error occurs it will call argumet "exp"'s 'clear_microcode()', 'clear_compile_stack()' and 'clean_up()' methods, and throw an exception
    \warning If string evaluation needs to be evaluated first (e.g. $foo="abc $bar def") it will add microcode class 'c_micro_str_eval' for evaluation of this kind of string into 'exp' argument
    \warning Argument 'var_name' must not be declared as 'const string&' but just 'string' because we do clean up before potential exception is thrown
*/
var_sym_index_t c_micro_symbol_variable_mapping::add(string var_name, c_expression* exp)
{
    var_sym_index_t index=VAR_SYM_INDEX_INVALID;

    if(var_name.empty()) {
        MY_ASSERT_NOTREACH();
        return add();
    }

    static const string d="Invalid string passed to compiler of microcode";

    if(var_name.length()==1) {
        if('$'==var_name[0] || '"'==var_name[1])
            throw c_var_exception(NULL,__FILE__,__my_func__,c_var_exception::E_USER_ERROR,d,VARTYPE_STRING);
    }

    bool built_in_const=script_function_translate_constant(var_name);

    if((var_name[0]>='0' && var_name[0]<='9'))
        built_in_const=true;
    if(var_name.length()>1 && ('-'==var_name[0] || '+'==var_name[0])) {
        if(var_name[1]>='0' && var_name[1]<='9')
            built_in_const=true;
    }

    bool simple=true;

    if(!built_in_const) {
        if('\''!=var_name[0] && var_name[var_name.length()-1]!='\'') {
            if(var_name.length()<2
                || (var_name[0]!='$' && var_name[0]!='"')
                || ('"'==var_name[0] && var_name[var_name.length()-1]!='"')) {
                    exp->clear_microcode();
                    //exp->clear_compile_stack(st);
                    exp->clean_up();

                    throw c_var_exception(NULL,__FILE__,__my_func__,c_var_exception::E_USER_ERROR,d,VARTYPE_STRING);
            }
        }
    }
    for(size_t i1=0; i1<var_name.length(); ++i1) {
        if('\\'==var_name[i1]) {
            // skip next char
            i1++;
            continue;
        }
        if('$'==var_name[i1]) {
            simple=false;
            break;
        }
    }
    if(simple) {
        index=add(var_name);
        return index;
    }
    mapping.insert(make_pair(next_index,var_name));

    var_sym_index_t out=add();

    c_micro_commit_vars* m1=MY_NEW c_micro_commit_vars(); // TODO: optimize
    exp->add_micro(m1);

    c_micro_str_eval* m2=MY_NEW c_micro_str_eval(var_name,out);
    exp->add_micro(m2);

    return out;
}

/*!
    \brief Adds new mapping (new temporary symbol)
    \author VooDooMan
    \version 3
    \date 2006-03-31, 2006-04-28, 2007-09-11
    \return Returns new index
*/
var_sym_index_t c_micro_symbol_variable_mapping::add()
{
    mapping.insert(make_pair(next_index,(string)""));
    return next_index++;
}

/*!
    \brief Adds new mapping (if it does not exist yet)
    \author VooDooMan
    \version 3
    \date 2006-03-16, 2006-04-28, 2008-02-26
    \param var_name Variable name
    \return Returns new index; or existing one, if it already exists
    \warning Use this _only_ for variables ('var_name' argument) that begins with '$', in other case use other overload of this method
*/
var_sym_index_t c_micro_symbol_variable_mapping::add(const string& var_name)
{
    var_sym_index_t index=VAR_SYM_INDEX_INVALID;
    index=get_var_index(var_name);
    if(VAR_SYM_INDEX_INVALID!=index)
        return index;
    if(var_name.length() && var_name[0]!='$') {
        auto_ptr<c_var_array> tmp_vars(MY_NEW c_var_array);
        bool ref=true;
        c_var_base* data=var_eval_obj(NULL,var_name,tmp_vars.get(),var_name,ref);
        data->set_persistent();
        data->set_read_only();
        consts.insert(make_pair(next_index,data));
    }
    mapping.insert(make_pair(next_index++,var_name));
    return next_index-1;
}

c_micro_symbol_variable_mapping::c_micro_symbol_variable_mapping(const c_micro_symbol_variable_mapping& orig)
{
    mapping=orig.mapping;
    next_index=orig.next_index;
    map<var_sym_index_t, c_var_base*>::const_iterator i1;
    for(i1=orig.consts.begin(); i1!=orig.consts.end(); ++i1) {
        c_var_base* var=(*i1).second->copy();
        consts.insert(make_pair(var->get_id(),var));
    }
}

c_micro_symbol_variable_mapping::c_micro_symbol_variable_mapping(const c_micro_symbol_variable_mapping& orig, c_micro_symbol_variable_table* t)
{
    mapping=orig.mapping;
    next_index=orig.next_index;

    map<var_sym_index_t, c_var_base*>::const_iterator i1;
    for(i1=orig.consts.begin(); i1!=orig.consts.end(); ++i1) {
        c_var_base* var=(*i1).second->copy();
        var->set_persistent();
        var->set_read_only();
        consts.insert(make_pair((*i1).first,var));
    }

    map<var_sym_index_t, c_var_base*>::iterator i4=consts.begin();
    for(var_sym_index_t i2=0; i2<t->real_space; ++i2) {
        if(!t->table[i2].value)
            continue;
        if(consts.end()!=i4) {
            t->table[i2].value=(*i4).second->get_id();
            i4++;
        } else {
            // DEBUG: set breakpoint here!
            int dummy=0;
            dummy++;
        }
    }
    if(consts.end()!=i4) {
        // DEBUG: set breakpoint here!
        int dummy=0;
        dummy++;
    }
}

void c_micro_symbol_variable_mapping::instantiate_consts(c_micro_symbol_variable_table* t)
{
    map<var_sym_index_t, c_var_base*>::iterator i4;
    for(i4=consts.begin(); i4!=consts.end(); ++i4) {
        if(!t->sanitize_index((*i4).first)) {
            // out of memory, we can ignere it, later on execution it will fail
            //     with index-out-of-bounds error in logs
            return;
        }
        if(t->table[(*i4).first].value) // we should not overwrite it
            continue;
        t->table[(*i4).first].value=(*i4).second->get_id();
    }
}

c_micro_symbol_variable_mapping::~c_micro_symbol_variable_mapping()
{
    map<var_sym_index_t, c_var_base*>::const_iterator i1;
    for(i1=consts.begin(); i1!=consts.end(); ++i1) {
        //MY_DELETE((*i1).second);
        (*i1).second->set_read_only(false);
        (*i1).second->set_persistent(false);
    }
}

