/***************************************************************************
                          c_micro_symbol_variable_mapping.h  -  description
                             -------------------
    begin                : Thu May 10 2007
    copyright            : (C) 2004-2007 by Marian VooDooMan Meravy
    email                : ghostvoodooman -a.T= users (dot) sourceforge.net
 ***************************************************************************/

/* $Id: c_micro_symbol_variable_mapping.h,v 1.14 2008/03/11 21:11:45 ghostvoodooman Exp $ */

/***************************************************************************

VooDoo cIRCle - an IRC (ro)bot
Copyright (C) 2004-2008 by Marian VooDooMan Meravy
  (ghostvoodooman -a.T= users (dot) sourceforge.net)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************/

/*!
    \file
    \brief Class 'c_micro_symbol_variable_mapping'
    \version $Id: c_micro_symbol_variable_mapping.h,v 1.14 2008/03/11 21:11:45 ghostvoodooman Exp $
*/

#ifndef C_MICRO_SYMBOL_VARIABLE_MAPPING_H__
#define C_MICRO_SYMBOL_VARIABLE_MAPPING_H__

#if defined(HAVE_CONFIG_H)
#   include <config.h>
#endif

#include <string>
#include <vector>
#include <stack>
#include <list>
#include <map>
#include <algorithm>

#include "../utils/memleak.h"

#include "../params.h"

#include "../microcode_shared.h"
#include "../var/c_var_base.h"
#include "c_micro_symbol_variable_table.h"

/*!
    \brief Stores symbol mapping of variables to indexes (compile time)
    \author VooDooMan
    \version 1
    \date 2006-03-16
*/
class c_micro_symbol_variable_mapping
{
    map<var_sym_index_t, string> mapping;   //!< Mapping of index to variable name
    var_sym_index_t next_index;             //!< Next index
    map<var_sym_index_t, c_var_base*> consts;   //!< Constants

    friend class c_function;

    friend class c_micro_symbol_table;
public:

    /*!
        \brief Constructor
        \author VooDooMan
        \version 1
        \date 2006-03-16
    */
    c_micro_symbol_variable_mapping()
    {
        next_index=1;
    }

    c_micro_symbol_variable_mapping(const c_micro_symbol_variable_mapping& orig);
    c_micro_symbol_variable_mapping(const c_micro_symbol_variable_mapping& orig, c_micro_symbol_variable_table* t);

    ~c_micro_symbol_variable_mapping();

    /*!
        \brief Creates entries in variable table from table of constants
        \author VooDooMan
        \version 1
        \date 2006-12-23
        \param t Variable table
    */
    void instantiate_consts(c_micro_symbol_variable_table* t);

    /*!
        \brief Resets/cleans-up class
        \author VooDooMan
        \version 1
        \date 2006-04-02
    */
    void reset();

    // tags for doxygen are at implementation (microcode.cpp)
    // Warning: Argument 'var_name' must not be declared as 'const string&' but just 'string' because we do clean up before potential exception is thrown
    var_sym_index_t add(string var_name, c_expression* exp);

    // tags for doxygen are at implementation (microcode.cpp)
    var_sym_index_t add(const string& var_name);

    // tags for doxygen are at implementation (microcode.cpp)
    var_sym_index_t add();

    /*!
        \brief Gets variable name from mapping
        \author VooDooMan
        \version 1
        \date 2006-03-16
        \param index Index to variable
        \return Returns new index
    */
    string get_var_name(var_sym_index_t index);

    /*!
        \brief Gets variable index from its name
        \author VooDooMan
        \version 1
        \date 2006-03-16
        \param var_name Name of variable
        \return Returns index, or 'VAR_SYM_INDEX_INVALID' if variable has not been added to table
    */
    var_sym_index_t get_var_index(const string& var_name);

    /*!
        \brief Gets number of symbols
        \author VooDooMan
        \version 1
        \date 2006-04-03
    */
    var_sym_index_t get_capacity();

    /*!
        \brief Gets variable mapping map
        \author VooDooMan
        \version 1
        \date 2006-04-28
    */
    const map<var_sym_index_t, string>& get_var_mapping();
};

#endif

