/***************************************************************************
                          c_micro_symbol_variable_table.cpp  -  description
                             -------------------
    begin                : Thu May 10 2007
    copyright            : (C) 2004-2007 by Marian VooDooMan Meravy
    email                : ghostvoodooman -aT= users (dot) sourceforge.net
 ***************************************************************************/

/* $Id: c_micro_symbol_variable_table.cpp,v 1.15 2008/06/08 08:43:14 ghostvoodooman Exp $ */

/***************************************************************************

VooDoo cIRCle - an IRC (ro)bot
Copyright (C) 2004-2008 by Marian VooDooMan Meravy
  (ghostvoodooman -aT= users (dot) sourceforge.net)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************/

/*!
    \file
    \brief Class 'c_micro_symbol_variable_table'
    \version $Id: c_micro_symbol_variable_table.cpp,v 1.15 2008/06/08 08:43:14 ghostvoodooman Exp $
*/

#if defined(HAVE_CONFIG_H)
#   include <config.h>
#endif

#include "../params.h"
#include "../utils/memleak.h"
#include "../utils/assert.h"
#include "../compat.h"

#include "c_micro_symbol_variable_table.h"
#include "c_micro_symbol_variable_mapping.h"
#include "c_micro_error_memory.h"
#include "c_micro_error_symbol_index.h"
#include "../var/c_var_bool.h"
#include "../var/c_var_int.h"
#include "../var/c_var_string.h"
#include "../var/c_var_base.h"
#include "../utils/precomp.h"

#if defined(_MSC_VER) || defined(__BORLANDC__)
#   pragma hdrstop
#endif

enum
{
    SPARE_VAR_INCREMENT = 20,   //!< How much spare should be allocated
};

void c_micro_symbol_variable_table::set_var_mapping(c_micro_symbol_variable_mapping* var_mapping)
{
    var_mapping_=var_mapping;
}

c_micro_symbol_variable_table::c_micro_symbol_variable_table(var_sym_index_t entries)
    : entries_(entries)
{
    var_mapping_=NULL;

    first_dynamic=entries;

    real_space=entries+SPARE_VAR_INCREMENT;
    table=(c_table*)my_malloc(sizeof(c_table)*real_space);
    if(!table) {
        entries_=0;
        real_space=0;
        return;
    }
    memset(table,0,sizeof(c_table)*real_space);
}

c_micro_symbol_variable_table::c_micro_symbol_variable_table(const c_micro_symbol_variable_table& orig)
{
    var_mapping_=orig.var_mapping_;

    first_dynamic=orig.first_dynamic;

    entries_=orig.entries_;
    real_space=orig.real_space;

    table=(c_table*)my_malloc(sizeof(c_table)*real_space);
    memcpy(table,orig.table,sizeof(c_table)*real_space);
}

c_micro_symbol_variable_table::~c_micro_symbol_variable_table()
{
    if(!table)
        return;
    c_micro_error* error=NULL;
    for(var_sym_index_t i1=0; i1<entries_; ++i1) {
        if(table[i1].value) {
            c_var_base* x=var_keeper->get(table[i1].value,&log_symbol_table_err_fn,&error);
            c_var_ref* ref=NULL;
            if(x && typeid(*x)==typeid(c_var_ref))
                ref=dynamic_cast<c_var_ref*>(x);
            if(x) {
                if(!ref && (x->get_read_only() || x->get_persistent()))
                    continue;
            }
            if(error) {
                if(core_mod && core_mod->get_logger()) {
                    string msg=error->get_msg();
                    if(var_mapping_) {
                        msg+="; variable: ";
                        if(i1>=first_dynamic)
                            msg+="<dynamic_variable>";
                        else
                            msg+=var_mapping_->get_var_name(i1);
                    }
                    core_mod->get_logger()->log_print1(E_LOG_DEBUG,LOG_SCOPE_DEFAULT,"",msg);
                }
                MY_DELETE(error);
                error=NULL;
            }
            var_keeper->free(table[i1].value,&log_symbol_table_err_fn,&error,true);
        }
        MY_DELETE(error);
        error=NULL;
    }
    my_free(table);
}

var_sym_index_t c_micro_symbol_variable_table::add_dynamic(c_micro_error*& error)
{
    if(real_space==entries_) {
        var_sym_index_t increment=SPARE_VAR_INCREMENT;
        void* new_table=my_realloc(table,sizeof(c_table)*(real_space+increment));
        if(!new_table) {
            error=MY_NEW c_micro_error_memory();
            return VAR_SYM_INDEX_INVALID;
        }
        table=(c_table*)new_table;
        char* first_new=((char*)table)+sizeof(c_table)*real_space;
        memset(first_new,0,sizeof(c_table)*increment);
        real_space+=increment;
    }
    return ++entries_-1;
}

bool c_micro_symbol_variable_table::set_lvalue(var_sym_index_t index, bool lvalue, c_micro_error*& error, e_micro_type UNUSEDARG(oper))
{
    error=NULL;
    if(VAR_SYM_INDEX_DISCARD==index)
        return true;
    if(index<0 || index>=entries_) {
        error=MY_NEW c_micro_error_symbol_index(index);
        return false;
    }
    table[index].lvalue=lvalue;
    return true;
}

bool c_micro_symbol_variable_table::is_ro(var_sym_index_t index, bool no_deref)
{
    if(VAR_SYM_INDEX_DISCARD==index)
        return false;
    if(index<0 || index>=entries_) {
        return false;
    }
    if(table[index].value) {
        c_var_base* dummy=var_keeper->get(table[index].value,NULL,NULL);
        if(no_deref && dummy && dummy->get_id()!=table[index].value)
            return false;
        if(dummy) {
            return dummy->get_read_only();
        }
    }
    return false;
}

bool c_micro_symbol_variable_table::invalidate(var_sym_index_t index, c_micro_error*& error, e_micro_type UNUSEDARG(oper))
{
    error=NULL;
    if(VAR_SYM_INDEX_DISCARD==index)
        return true;
    if(index<0 || index>=entries_) {
        error=MY_NEW c_micro_error_symbol_index(index);
        return false;
    }
    table[index].value=VAR_ID_INVALID;
    table[index].lvalue=false;
    return true;
}

void c_micro_symbol_variable_table::update_id_change(const vector<pair<var_id_t, var_id_t> >& changes)
{
    vector<pair<var_id_t, var_id_t> >::const_iterator i1;
    for(i1=changes.begin(); i1!=changes.end(); ++i1) {
        for(var_sym_index_t i2=0; i2<entries_; ++i2) {
            if(table[i2].value==(*i1).first)
                table[i2].value=(*i1).second;
        }
    }
}

void c_micro_symbol_variable_table::commit_changes(c_micro_symbol_variable_table& to)
{
    vector<pair<var_id_t, var_id_t> >::const_iterator i1;
    for(i1=all_changes.begin(); i1!=all_changes.end(); ++i1) {
        to.all_changes.push_back(*i1);
        for(var_sym_index_t i2=0; i2<to.entries_; ++i2) {
            if(to.table[i2].value==(*i1).first)
                to.table[i2].value=(*i1).second;
        }
    }
}

bool c_micro_symbol_variable_table::set(var_sym_index_t index, c_var_base*& value, bool lvalue, c_micro_error*& error, e_micro_type UNUSEDARG(oper), bool no_deref, c_module* this_module, const string& debug_prefix, const string& debug)
{
    error=NULL;
    if(VAR_SYM_INDEX_DISCARD==index)
        return true;
    if(index<0 || index>=entries_) {
        error=MY_NEW c_micro_error_symbol_index(index);
        return false;
    }
    if(value && table[index].value==value->get_id()) {
        table[index].lvalue=lvalue;
        return true;
    }
    vector<pair<var_id_t, var_id_t> > changes;
    if(table[index].value) {
        table[index].value=var_keeper->set(table[index].value,&log_symbol_table_err_fn,&error,false,value,no_deref,changes,all_changes,this_module,debug_prefix,debug);
        var_keeper->adjust_ref_cnt(table[index].value,+1);
    } else {
        table[index].value=value?value->get_id():0;
        if(table[index].value) {
            table[index].value=var_keeper->set(table[index].value,&log_symbol_table_err_fn,&error,false,value,no_deref,changes,all_changes,this_module,debug_prefix,debug);
            var_keeper->adjust_ref_cnt(table[index].value,+1);
        }
    }

    update_id_change(changes);

    table[index].lvalue=lvalue;
    return true;
}

bool c_micro_symbol_variable_table::set(var_sym_index_t index, c_var_array*& value, bool lvalue, c_micro_error*& error, e_micro_type UNUSEDARG(oper), bool no_deref, c_module* this_module, const string& debug_prefix, const string& debug)
{
    error=NULL;
    if(VAR_SYM_INDEX_DISCARD==index)
        return true;
    if(index<0 || index>=entries_) {
        error=MY_NEW c_micro_error_symbol_index(index);
        return false;
    }
    c_var_base* val=value;
    if(value && table[index].value==value->get_id()) {
        table[index].lvalue=lvalue;
        return true;
    }
    vector<pair<var_id_t, var_id_t> > changes;
    if(table[index].value) {
        table[index].value=var_keeper->set(table[index].value,&log_symbol_table_err_fn,&error,false,val,no_deref,changes,all_changes,this_module,debug_prefix,debug);
        var_keeper->adjust_ref_cnt(table[index].value,+1);
    } else {
        table[index].value=val?val->get_id():0;
        if(table[index].value) {
            table[index].value=var_keeper->set(table[index].value,&log_symbol_table_err_fn,&error,false,val,no_deref,changes,all_changes,this_module,debug_prefix,debug);
            var_keeper->adjust_ref_cnt(table[index].value,+1);
        }
    }

    update_id_change(changes);

    value=NULL;
    if(val && typeid(*val)==typeid(c_var_array))
        value=dynamic_cast<c_var_array*>(val);

    table[index].lvalue=lvalue;
    return true;
}

bool c_micro_symbol_variable_table::set(var_sym_index_t index, c_var_bool*& value, bool lvalue, c_micro_error*& error, e_micro_type UNUSEDARG(oper), bool no_deref, c_module* this_module, const string& debug_prefix, const string& debug)
{
    error=NULL;
    if(VAR_SYM_INDEX_DISCARD==index)
        return true;
    if(index<0 || index>=entries_) {
        error=MY_NEW c_micro_error_symbol_index(index);
        return false;
    }
    c_var_base* val=value;
    if(value && table[index].value==value->get_id()) {
        table[index].lvalue=lvalue;
        return true;
    }
    vector<pair<var_id_t, var_id_t> > changes;
    if(table[index].value) {
        table[index].value=var_keeper->set(table[index].value,&log_symbol_table_err_fn,&error,false,val,no_deref,changes,all_changes,this_module,debug_prefix,debug);
        var_keeper->adjust_ref_cnt(table[index].value,+1);
    } else {
        table[index].value=val?val->get_id():0;
        if(table[index].value) {
            table[index].value=var_keeper->set(table[index].value,&log_symbol_table_err_fn,&error,false,val,no_deref,changes,all_changes,this_module,debug_prefix,debug);
            var_keeper->adjust_ref_cnt(table[index].value,+1);
        }
    }

    update_id_change(changes);

    value=NULL;
    if(val && typeid(*val)==typeid(c_var_bool))
        value=dynamic_cast<c_var_bool*>(val);

    table[index].lvalue=lvalue;
    return true;
}

bool c_micro_symbol_variable_table::set(var_sym_index_t index, c_var_int*& value, bool lvalue, c_micro_error*& error, e_micro_type UNUSEDARG(oper), bool no_deref, c_module* this_module, const string& debug_prefix, const string& debug)
{
    error=NULL;
    if(VAR_SYM_INDEX_DISCARD==index)
        return true;
    if(index<0 || index>=entries_) {
        error=MY_NEW c_micro_error_symbol_index(index);
        return false;
    }
    c_var_base* val=value;
    if(value && table[index].value==value->get_id()) {
        table[index].lvalue=lvalue;
        return true;
    }
    vector<pair<var_id_t, var_id_t> > changes;
    if(table[index].value) {
        table[index].value=var_keeper->set(table[index].value,&log_symbol_table_err_fn,&error,false,val,no_deref,changes,all_changes,this_module,debug_prefix,debug);
        var_keeper->adjust_ref_cnt(table[index].value,+1);
    } else {
        table[index].value=val?val->get_id():0;
        if(table[index].value) {
            table[index].value=var_keeper->set(table[index].value,&log_symbol_table_err_fn,&error,false,val,no_deref,changes,all_changes,this_module,debug_prefix,debug);
            var_keeper->adjust_ref_cnt(table[index].value,+1);
        }
    }

    update_id_change(changes);

    value=NULL;
    if(val && typeid(*val)==typeid(c_var_int))
        value=dynamic_cast<c_var_int*>(val);

    table[index].lvalue=lvalue;
    return true;
}

bool c_micro_symbol_variable_table::set(var_sym_index_t index, c_var_string*& value, bool lvalue, c_micro_error*& error, e_micro_type UNUSEDARG(oper), bool no_deref, c_module* this_module, const string& debug_prefix, const string& debug)
{
    error=NULL;
    if(VAR_SYM_INDEX_DISCARD==index)
        return true;
    if(index<0 || index>=entries_) {
        error=MY_NEW c_micro_error_symbol_index(index);
        return false;
    }
    c_var_base* val=value;
    if(value && table[index].value==value->get_id()) {
        table[index].lvalue=lvalue;
        return true;
    }
    vector<pair<var_id_t, var_id_t> > changes;
    if(table[index].value) {
        table[index].value=var_keeper->set(table[index].value,&log_symbol_table_err_fn,&error,false,val,no_deref,changes,all_changes,this_module,debug_prefix,debug);
        var_keeper->adjust_ref_cnt(table[index].value,+1);
    } else {
        table[index].value=val?val->get_id():0;
        if(table[index].value) {
            table[index].value=var_keeper->set(table[index].value,&log_symbol_table_err_fn,&error,false,val,no_deref,changes,all_changes,this_module,debug_prefix,debug);
            var_keeper->adjust_ref_cnt(table[index].value,+1);
        }
    }

    update_id_change(changes);

    value=NULL;
    if(val && typeid(*val)==typeid(c_var_string))
        value=dynamic_cast<c_var_string*>(val);

    table[index].lvalue=lvalue;
    return true;
}

bool c_micro_symbol_variable_table::get(var_sym_index_t index, c_var_base*& value, bool& lvalue, c_micro_error*& error, bool do_not_allocate)
{
    if(index<0 || index>=entries_) {
        error=MY_NEW c_micro_error_symbol_index(index);
        return false;
    }
    if(do_not_allocate && !table[index].value) {
        value=NULL;
        lvalue=false;
        return true;
    }
    if(!table[index].value) {
        value=NULL;
        lvalue=false;
        return true;
    }
    value=var_keeper->get(table[index].value,&log_symbol_table_err_fn,&error);
    lvalue=table[index].lvalue;
    return true;
}

bool c_micro_symbol_variable_table::sanitize_index(var_sym_index_t index)
{
    if(entries_+1<index)
        entries_=index;
    while(real_space<=entries_) {
        var_sym_index_t increment=SPARE_VAR_INCREMENT;
        void* new_table=my_realloc(table,sizeof(c_table)*(real_space+increment));
        if(!new_table) {
            return false;
        }
        table=(c_table*)new_table;
        char* first_new=((char*)table)+sizeof(c_table)*real_space;
        memset(first_new,0,sizeof(c_table)*increment);
        real_space+=increment;
    }
    return true;
}

string c_micro_symbol_variable_table::get_var_name(var_sym_index_t index)
{
    return var_mapping_->get_var_name(index);
}

