/***************************************************************************
                          c_micro_symbol_variable_table.h  -  description
                             -------------------
    begin                : Thu May 10 2007
    copyright            : (C) 2004-2007 by Marian VooDooMan Meravy
    email                : ghostvoodooman -a.T= users (dot) sourceforge.net
 ***************************************************************************/

/* $Id: c_micro_symbol_variable_table.h,v 1.16 2008/06/08 08:43:14 ghostvoodooman Exp $ */

/***************************************************************************

VooDoo cIRCle - an IRC (ro)bot
Copyright (C) 2004-2008 by Marian VooDooMan Meravy
  (ghostvoodooman -a.T= users (dot) sourceforge.net)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************/

/*!
    \file
    \brief Class 'c_micro_symbol_variable_table'
    \version $Id: c_micro_symbol_variable_table.h,v 1.16 2008/06/08 08:43:14 ghostvoodooman Exp $
*/

#ifndef C_MICRO_SYMBOL_VARIABLE_TABLE_H__
#define C_MICRO_SYMBOL_VARIABLE_TABLE_H__

#if defined(HAVE_CONFIG_H)
#   include <config.h>
#endif

#include <string>
#include <vector>
#include <stack>
#include <list>
#include <map>
#include <algorithm>

#include "../utils/memleak.h"

#include "../params.h"

#include "c_micro_error.h"
#include "../var_keeper.h"
#include "../var/c_var_ref.h"
#include "../mod/c_module.h"

void log_symbol_table_err_fn(c_var_keeper::e_log_severity severity, const string& msg, void* data);
void log_fn_c_micro_func_call(c_var_keeper::e_log_severity severity, const string& msg, void* data);

/*!
    \brief Stores symbol table of variable for microcode (runtime)
    \author VooDooMan
    \version 1
    \date 2006-03-16
*/
class c_micro_symbol_variable_table
{
    struct c_table
    {
        var_id_t value;             //!< ID of variable in variable keeper
        bool lvalue;                //!< Is it l-value?
    };

    c_table* table;                 //!< Table of variables

    var_sym_index_t entries_;       //!< Number of active entries in table
    var_sym_index_t real_space;     //!< Real space: number of entries in table
    var_sym_index_t first_dynamic;  //!< ID of first dynamic variable

    vector<pair<var_id_t, var_id_t> > all_changes;  //!< All changes of variables' ID
    
    friend class c_micro_symbol_variable_mapping;

    /*!
        \brief Updates change of IDs
        \param changes From -> to
        \author VooDooMan
        \version 1
        \date 2006-10-30
    */
    void update_id_change(const vector<pair<var_id_t, var_id_t> >& changes);

protected:

    c_micro_symbol_variable_mapping* var_mapping_;
    void set_var_mapping(c_micro_symbol_variable_mapping* var_mapping);
    friend class c_micro_symbol_table;

public:

    /*!
        \brief Constructor
        \author VooDooMan
        \version 1
        \date 2006-03-16
        \param entries Number of entries in symbol table
    */
    c_micro_symbol_variable_table(var_sym_index_t entries);

    c_micro_symbol_variable_table(const c_micro_symbol_variable_table& orig);

    /*!
        \brief Destructor
        \author VooDooMan
        \version 1
        \date 2006-03-16
    */
    ~c_micro_symbol_variable_table();

    /*!
        \brief Sets value for symbol
        \author VooDooMan
        \version 1
        \date 2006-03-16
        \retval true Success
        \retval false Error
        \param index Index in symbol table
        \param lvalue Is 'value' l-value? If value is not UNSET, is replaced variable required to be l-value? - if it is not, error will be reported
        \param error If returned 'false' there is error reporting class
        \param oper Operator
        \param value Value (could be changed after return!)
        \param no_deref Do not dereference
        \param this_module Module
        \param debug_prefix Debug prefix
        \param debug Debug string
    */
    bool set(var_sym_index_t index, c_var_base*& value, bool lvalue, c_micro_error*& error, e_micro_type UNUSEDARG(oper), bool no_deref, c_module* this_module, const string& debug_prefix, const string& debug);

    /*!
        \brief Sets value for symbol
        \author VooDooMan
        \version 1
        \date 2006-03-16
        \retval true Success
        \retval false Error
        \param index Index in symbol table
        \param lvalue Is 'value' l-value? If value is not UNSET, is replaced variable required to be l-value? - if it is not, error will be reported
        \param error If returned 'false' there is error reporting class
        \param oper Operator
        \param value Value (could be changed after return!)
        \param no_deref Do not dereference
        \param this_module Module
        \param debug_prefix Debug prefix
        \param debug Debug string
    */
    bool set(var_sym_index_t index, c_var_array*& value, bool lvalue, c_micro_error*& error, e_micro_type UNUSEDARG(oper), bool no_deref, c_module* this_module, const string& debug_prefix, const string& debug);

    /*!
        \brief Sets value for symbol
        \author VooDooMan
        \version 1
        \date 2006-03-16
        \retval true Success
        \retval false Error
        \param index Index in symbol table
        \param lvalue Is 'value' l-value? If value is not UNSET, is replaced variable required to be l-value? - if it is not, error will be reported
        \param error If returned 'false' there is error reporting class
        \param oper Operator
        \param value Value (could be changed after return!)
        \param no_deref Do not dereference
        \param this_module Module
        \param debug_prefix Debug prefix
        \param debug Debug string
    */
    bool set(var_sym_index_t index, c_var_bool*& value, bool lvalue, c_micro_error*& error, e_micro_type UNUSEDARG(oper), bool no_deref, c_module* this_module, const string& debug_prefix, const string& debug);

    /*!
        \brief Sets value for symbol
        \author VooDooMan
        \version 1
        \date 2006-03-16
        \retval true Success
        \retval false Error
        \param index Index in symbol table
        \param lvalue Is 'value' l-value? If value is not UNSET, is replaced variable required to be l-value? - if it is not, error will be reported
        \param error If returned 'false' there is error reporting class
        \param oper Operator
        \param value Value (could be changed after return!)
        \param no_deref Do not dereference
        \param this_module Module
        \param debug_prefix Debug prefix
        \param debug Debug string
    */
    bool set(var_sym_index_t index, c_var_int*& value, bool lvalue, c_micro_error*& error, e_micro_type UNUSEDARG(oper), bool no_deref, c_module* this_module, const string& debug_prefix, const string& debug);

    /*!
        \brief Sets value for symbol
        \author VooDooMan
        \version 1
        \date 2006-03-16
        \retval true Success
        \retval false Error
        \param index Index in symbol table
        \param lvalue Is 'value' l-value? If value is not UNSET, is replaced variable required to be l-value? - if it is not, error will be reported
        \param error If returned 'false' there is error reporting class
        \param oper Operator
        \param value Value (could be changed after return!)
        \param no_deref Do not dereference
        \param this_module Module
        \param debug_prefix Debug prefix
        \param debug Debug string
    */
    bool set(var_sym_index_t index, c_var_string*& value, bool lvalue, c_micro_error*& error, e_micro_type UNUSEDARG(oper), bool no_deref, c_module* this_module, const string& debug_prefix, const string& debug);

    /*!
        \brief Gets value for symbol
        \author VooDooMan
        \version 1
        \date 2006-03-16
        \retval true Success
        \retval false Error
        \param error If returned 'false' there is error reporting class
        \param do_not_allocate Set to 'true' if we want forced not creation/allocation of entry in variable table in class 'c_var_keeper'
        \param index Index in symbol table
        \param lvalue Is 'value' l-value?
        \param value Value (could be changed after return!)
    */
    bool get(var_sym_index_t index, c_var_base*& value, bool& lvalue, c_micro_error*& error, bool do_not_allocate = false);

    /*!
        \brief Invalidates variable ID for symbol
        \author VooDooMan
        \version 1
        \date 2006-05-14
        \retval true Success
        \retval false Error
        \param index Index in symbol table
        \param error If returned 'false' there is error reporting class
        \param oper Operator
    */
    bool invalidate(var_sym_index_t index, c_micro_error*& error, e_micro_type UNUSEDARG(oper));

    /*!
        \brief Adds a dynamic variable, allocates more space if needed
        \author VooDooMan
        \version 1
        \date 2006-04-03
        \retval VAR_SYM_INDEX_INVALID For error
        \param error If returned 'VAR_SYM_INDEX_INVALID' there is error reporting class
    */
    var_sym_index_t add_dynamic(c_micro_error*& error);

    /*!
        \brief Makes sure that index of variable fits into the table, or allocates more entries
        \author VooDooMan
        \version 1
        \date 2006-12-23
        \param index Index in symbol table
        \return Returs 'false' if out of memory
    */
    bool sanitize_index(var_sym_index_t index);

    /*!
        \brief Sets l-value flag for symbol
        \author VooDooMan
        \version 1
        \date 2006-04-05
        \retval true Success
        \retval false Error
        \param index Index in symbol table
        \param error If returned 'false' there is error reporting class
        \param oper Operator
        \param lvalue L-value flag to set
    */
    bool set_lvalue(var_sym_index_t index, bool lvalue, c_micro_error*& error, e_micro_type UNUSEDARG(oper));

    /*!
        \brief Checks whether variable is read-only
        \author VooDooMan
        \version 2
        \date 2006-06-07, 2007-04-26
        \param index Index in symbol table
        \param no_deref Do not dereference
    */
    bool is_ro(var_sym_index_t index, bool no_deref = false);

    /*!
        \brief Makes changes of IDs of variables (to parent scope)
        \date 2006-10-30
        \version 1
        \author VooDooMan
        \param to Parent scope
    */
    void commit_changes(c_micro_symbol_variable_table& to);

    /*!
        \brief Retrieves variable name from ID
        \author VooDooMan
        \version 1
        \date 2008-06-07
    */
    string get_var_name(var_sym_index_t index);
};

#endif

