/***************************************************************************
                          c_micro_ternary.h  -  description
                             -------------------
    begin                : Thu May 10 2007
    copyright            : (C) 2004-2007 by Marian VooDooMan Meravy
    email                : ghostvoodooman -a.T= users (dot) sourceforge.net
 ***************************************************************************/

/* $Id: c_micro_ternary.h,v 1.14 2008/03/11 21:11:45 ghostvoodooman Exp $ */

/***************************************************************************

VooDoo cIRCle - an IRC (ro)bot
Copyright (C) 2004-2008 by Marian VooDooMan Meravy
  (ghostvoodooman -a.T= users (dot) sourceforge.net)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************/

/*!
    \file
    \brief Class 'c_micro_ternary'
    \version $Id: c_micro_ternary.h,v 1.14 2008/03/11 21:11:45 ghostvoodooman Exp $
*/

#ifndef C_MICRO_TERNARY_H__
#define C_MICRO_TERNARY_H__

#if defined(HAVE_CONFIG_H)
#   include <config.h>
#endif

#include <string>
#include <vector>
#include <stack>
#include <list>
#include <map>
#include <algorithm>

#include "../utils/memleak.h"

#include "../params.h"

#include "c_micro.h"

/*!
    \brief Microcode class: ternary class
    \author VooDooMan
    \version 1
    \date 2006-03-16
*/
class c_micro_ternary : public c_micro
{
    var_sym_index_t input_;             //!< Input variable index
    var_sym_index_t output_;            //!< Output variable index (result of '?:' operator)
    var_sym_index_t shared_res;         //!< Shared result for both cases

    instruct_id_t false_jump;           //!< ID where to jump in case of operator returned FALSE
    vector<instruct_id_t> after_true_jump_inst_ids; //!< IDs of MICRO_JUMP instructions (after TRUE part of ternary operator to get over FALSE part)

    bool result_;                       //!< Run-time result

public:

    /*!
        \brief Constructor
        \author VooDooMan
        \version 1
        \date 2006-03-16
        \param input Left-hand side variable index
    */
    c_micro_ternary(var_sym_index_t input)
        : c_micro(), input_(input)
    {
        false_jump=INSTRUCT_ID_INVALID;
        output_=VAR_SYM_INDEX_INVALID;
        shared_res=VAR_SYM_INDEX_INVALID;
    }

    /*!
        \brief Copy-constructor
        \author VooDooMan
        \version 1
        \date 2006-08-18
    */
    c_micro_ternary(const c_micro_ternary& orig);

    /*!
        \brief Copy
        \author VooDooMan
        \version 1
        \date 2006-08-18
    */
    c_micro* copy();

    /*!
        \brief Destructor
        \author VooDooMan
        \version 1
        \date 2006-03-16
    */
    virtual ~c_micro_ternary() { ; }

    e_micro_type get_type()
    {
        return MICRO_TERNARY;
    }

    /*!
        \brief Sets ID where to jump in case of operator returned FALSE
        \author VooDooMan
        \version 1
        \date 2006-03-16
    */
    void set_false_jump(instruct_id_t id)
    {
        false_jump=id;
    }

    /*!
        \brief Gets ID where to jump in case of operator returned FALSE
        \author VooDooMan
        \version 1
        \date 2006-03-16
    */
    instruct_id_t get_false_jump()
    {
        return false_jump;
    }

    bool perform(c_micro_symbol_table* symbols, c_micro_error*& error, c_module* this_module, const string& debug);

    /*!
        \brief Checks if to jump over TRUE section of ternary operator
        \author VooDooMan
        \version 1
        \date 2006-03-16
        \retval true Success
        \retval false Error
        \param symbols Symbol table object
        \param error If returned 'false', there is error reporter
        \param jump_over If returned 'true', this returns if we should skip rest of expression (ternary operator returned FALSE)
    */
    bool get_jump_over(c_micro_symbol_table* symbols, c_micro_error*& error, bool& jump_over);

    /*!
        \brief Set another one MICRO_JUMP instruction ID, in case of TRUE evaluation, to skip FALSE part
        \date 2006-04-05
        \version 1
        \author VooDooMan
    */
    void send_another_jump_id(instruct_id_t id)
    {
        after_true_jump_inst_ids.push_back(id);
    }

    /*!
        \brief Gets list of MICRO_JUMP instruction IDs, in case of TRUE evaluation, to skip FALSE part
        \date 2006-04-05
        \version 1
        \author VooDooMan
    */
    vector<instruct_id_t> get_jump_ids()
    {
        return after_true_jump_inst_ids;
    }

    /*!
        \brief Sets output variable
        \date 2006-04-27
        \version 1
        \author VooDooMan
        \param output Output variable ID
    */
    void set_output(var_sym_index_t output)
    {
        output_=output;
    }

    virtual var_sym_index_t get_output();

    /*!
        \brief Sets shared result variable ID
        \date 2006-11-06
        \version 1
        \author VooDooMan
        \param id Output variable ID
    */
    void set_shared_result(var_sym_index_t id)
    {
        shared_res=id;
    }

    /*!
        \brief Gets shared result variable ID
        \date 2006-11-06
        \version 1
        \author VooDooMan
    */
    var_sym_index_t get_shared_result()
    {
        return shared_res;
    }

    void add_inputs(vector<var_sym_index_t>& inputs);
    void optimize_output(const vector<var_sym_index_t>& inputs);
};

#endif

