/***************************************************************************
                          c_micro_update.h  -  description
                             -------------------
    begin                : Thu May 10 2007
    copyright            : (C) 2004-2007 by Marian VooDooMan Meravy
    email                : ghostvoodooman -a.T= users (dot) sourceforge.net
 ***************************************************************************/

/* $Id: c_micro_update.h,v 1.14 2008/03/11 21:11:45 ghostvoodooman Exp $ */

/***************************************************************************

VooDoo cIRCle - an IRC (ro)bot
Copyright (C) 2004-2008 by Marian VooDooMan Meravy
  (ghostvoodooman -a.T= users (dot) sourceforge.net)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************/

/*!
    \file
    \brief Class 'c_micro_update'
    \version $Id: c_micro_update.h,v 1.14 2008/03/11 21:11:45 ghostvoodooman Exp $
*/

#ifndef C_MICRO_UPDATE_H__
#define C_MICRO_UPDATE_H__

#if defined(HAVE_CONFIG_H)
#   include <config.h>
#endif

#include <string>
#include <vector>
#include <stack>
#include <list>
#include <map>
#include <algorithm>

#include "../utils/memleak.h"

#include "../params.h"

#include "c_micro.h"
#include "../utils.h"
#include "../variable.h"

/*!
    \brief Microcode class: updates indexed variable
    \author VooDooMan
    \version 1
    \date 2006-04-13
*/
class c_micro_update : public c_micro
{
    var_sym_index_t input_;             //!< Input variable index
    string var_name_;                   //!< Variable name
public:

    /*!
        \brief Constructor
        \author VooDooMan
        \version 1
        \date 2006-04-13
        \param input Input variable index
        \param var_name Name of variable, or an empty string
    */
    c_micro_update(var_sym_index_t input, const string& var_name)
        : c_micro(), input_(input), var_name_(var_name)
    {
    }

    /*!
        \brief Copy-constructor
        \author VooDooMan
        \version 1
        \date 2006-08-18
    */
    c_micro_update(const c_micro_update& orig);

    /*!
        \brief Copy
        \author VooDooMan
        \version 1
        \date 2006-08-18
    */
    c_micro* copy();

    /*!
        \brief Destructor
        \author VooDooMan
        \version 1
        \date 2006-04-13
    */
    virtual ~c_micro_update() { ; }

    e_micro_type get_type()
    {
        return MICRO_UPDATE;
    }

    bool perform(c_micro_symbol_table* UNUSEDARG(symbols), c_micro_error*& UNUSEDARG(error), c_module* UNUSEDARG(this_module), const string& UNUSEDARG(debug))
    {
        // handled separately
        return true;
    }

    /*!
        \brief Returns variable name
        \author VooDooMan
        \version 1
        \date 2006-04-13
    */
    const string& get_var_name()
    {
        return var_name_;
    }

    /*!
        \brief Returns variable ID
        \author VooDooMan
        \version 1
        \date 2006-04-13
    */
    var_id_t get_var()
    {
        return input_;
    }

    virtual var_sym_index_t get_output();

    void add_inputs(vector<var_sym_index_t>& inputs);
    void optimize_output(const vector<var_sym_index_t>& inputs);
};

#endif

