/***************************************************************************
                          c_script_timeout_protector_locker.h  -  description
                             -------------------
    begin                : Sun Jun 15 2008
    copyright            : (C) 2008 by Marian VooDooMan Meravy
    email                : ghostvoodooman -aT= users (dot) sourceforge.net
 ***************************************************************************/

/* $Id: c_script_timeout_protector_locker.h,v 1.4 2008/06/16 06:04:50 ghostvoodooman Exp $ */

/***************************************************************************

VooDoo cIRCle - an IRC (ro)bot
Copyright (C) 2004-2008 by Marian VooDooMan Meravy
  (ghostvoodooman -aT= users (dot) sourceforge.net)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************/

/*!
    \file
    \brief Implements class for locking and unlocking protection against script timeout
    \version $Id: c_script_timeout_protector_locker.h,v 1.4 2008/06/16 06:04:50 ghostvoodooman Exp $
*/

#ifndef C_SCRIPT_TIMEOUT_PROTECTOR_LOCKER_H__
#define C_SCRIPT_TIMEOUT_PROTECTOR_LOCKER_H__

#if defined(HAVE_CONFIG_H)
#   include <config.h>
#endif

class c_script_timeout_protector_unlocker;

/*!
    \brief Class for locking protection against script timeout
    \version 1
    \date 2008-06-15
    \author VooDooMan
*/
class c_script_timeout_protector_locker
{
    static int m_lock_cnt;  //!< Lock count (recursive)
    friend class c_script_timeout_protector_unlocker;

public:

    /*!
        \brief C-tor, locks counter recursively
        \version 1
        \date 2008-06-15
        \author VooDooMan
    */
    c_script_timeout_protector_locker();

    /*!
        \brief D-tor, unlocks counter recursively
        \version 1
        \date 2008-06-15
        \author VooDooMan
    */
    ~c_script_timeout_protector_locker();
};

/*!
    \brief Class for unlocking locker protection against script timeout (for "work" execution)
    \version 1
    \date 2008-06-15
    \author VooDooMan
*/
class c_script_timeout_protector_unlocker
{
public:

    /*!
        \brief C-tor, unlocks counter recursively (begin of "work" execution)
        \version 1
        \date 2008-06-15
        \author VooDooMan
    */
    c_script_timeout_protector_unlocker();

    /*!
        \brief D-tor, locks counter recursively (end of "work" execution)
        \version 1
        \date 2008-06-15
        \author VooDooMan
    */
    ~c_script_timeout_protector_unlocker();
};

#endif

