/***************************************************************************
                          c_binder.cpp  -  description
                             -------------------
    begin                : Thu May 10 2007
    copyright            : (C) 2004-2007 by Marian VooDooMan Meravy
    email                : ghostvoodooman -a.T= users (dot) sourceforge.net
 ***************************************************************************/

/* $Id: c_binder.cpp,v 1.17 2008/11/24 12:09:46 ghostvoodooman Exp $ */

/***************************************************************************

VooDoo cIRCle - an IRC (ro)bot
Copyright (C) 2004-2008 by Marian VooDooMan Meravy
  (ghostvoodooman -a.T= users (dot) sourceforge.net)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************/

/*!
    \file
    \brief Class 'c_binder', 'c_binder_mask' and 'c_binder_data'
    \version $Id: c_binder.cpp,v 1.17 2008/11/24 12:09:46 ghostvoodooman Exp $
*/

#if defined(HAVE_CONFIG_H)
#   include <config.h>
#endif

#include "../params.h"
#include "../utils/memleak.h"
#include "../utils/assert.h"
#include "../compat.h"

#include "c_binder.h"
#include "../match.h"
#include "../mod/c_module.h"
#include "../modules.h"
#include "../var_keeper.h"
#include "../var_index.h"
#include "../binmod.h"
#include "../pcre/c_pcre.h"

#include "../utils/precomp.h"

#if defined(_MSC_VER) || defined(__BORLANDC__)
#   pragma hdrstop
#endif

static const char* s_err_empty_flags="No flags were specified";
static const char* s_err_exclusivity="You cannot mix @exact@, @mask@ and regexp flags";
static const char* s_err_case_exclusivity="You can specify only one case mapping in case-insensitive matching";
static const char* s_err_both_match_and_match_all="@match@ and @match_all@ flags cannot be both specified";
static const char* s_err_neither_match_and_match_all="Either @match@ or @match_all@ flag must be specified";

extern bool php_do_preg_match_all(c_module* m_module, const string& debug, c_var_array* &matches_array, MY_LONG_LONG flags, string& error, const vector<vector<pair<int,string> > >& matches);
extern bool php_do_preg_match(c_module* m_module, const string& debug, c_var_array* &matches_array, MY_LONG_LONG flags, string& error, const vector<vector<pair<int,string> > >& matches);
extern void report_pcre_error(c_module* m_module, const char* prefix, const c_pcre& re, const char* error_report, const string& debug, string& error);

bool c_binder::call_bound_dispatch_event(c_var_array* dispatch, const c_binder_data& internal_data)
{
    bool has_error;
    string error;

    {
        c_var_array* matches_array;
        m_type.get_match_array(m_debug,error,has_error,matches_array,internal_data.m_matches_type);
        if(has_error) {
            if(m_module && m_module->get_logger())
                m_module->get_logger()->log_print1(E_LOG_DEBUG,LOG_SCOPE_MODULE,"",error);
            return false;
        }
        if(matches_array) {
            c_index_base* index=MY_NEW c_index_str("type_matches");
            dispatch->set(m_module,index,matches_array,false,m_debug);
        }
    }

    {
        c_var_array* matches_array;
        m_user.get_match_array(m_debug,error,has_error,matches_array,internal_data.m_matches_user);
        if(has_error) {
            if(m_module && m_module->get_logger())
                m_module->get_logger()->log_print1(E_LOG_DEBUG,LOG_SCOPE_MODULE,"",error);
            return false;
        }
        if(matches_array) {
            c_index_base* index=MY_NEW c_index_str("user_matches");
            dispatch->set(m_module,index,matches_array,false,m_debug);
        }
    }

    {
        c_var_array* matches_array;
        m_channel.get_match_array(m_debug,error,has_error,matches_array,internal_data.m_matches_channel);
        if(has_error) {
            if(m_module && m_module->get_logger())
                m_module->get_logger()->log_print1(E_LOG_DEBUG,LOG_SCOPE_MODULE,"",error);
            return false;
        }
        if(matches_array) {
            c_index_base* index=MY_NEW c_index_str("channel_matches");
            dispatch->set(m_module,index,matches_array,false,m_debug);
        }
    }

    {
        c_var_array* matches_array;
        m_line.get_match_array(m_debug,error,has_error,matches_array,internal_data.m_matches_line);
        if(has_error) {
            if(m_module && m_module->get_logger())
                m_module->get_logger()->log_print1(E_LOG_DEBUG,LOG_SCOPE_MODULE,"",error);
            return false;
        }
        if(matches_array) {
            c_index_base* index=MY_NEW c_index_str("line_matches");
            dispatch->set(m_module,index,matches_array,false,m_debug);
        }
    }

    if(!m_bin_module) {
        vector<c_var_base*> args;
        args.push_back(dispatch);
        m_module->execute(m_function,args,m_debug);

        var_keeper->set_current_session(SESSION_RESERVED_ID);
        return true;
    } else {
        if(m_func) {
            c_binmodule* mod=dynamic_cast<c_binmodule*>(m_module);
            if(mod) {
// test suite
#ifdef _DEBUG
#   if 0
                int test=0;
                for(;;) {
                    test++;
//test=5;
                    int xx=0; // set breakpoint here; just modify value of 'test'
                    xx++;
                    switch(test) {
                        case 1:
                        {
                            c_var_array* a=MY_NEW c_var_array();
                            //
                            c_vdcmod_data_t* r=NULL;
                            mod->convert_vars(a,r);
                            // DEBUG: inspect structure hierarchy 'r'
                            mod->_converted_vars_dump("converted1.dump.txt",r);
                            int xx=0; // set breakpoint HERE
                            xx++;
                            break;
                        }
                        case 2:
                        {
                            c_var_array* a=MY_NEW c_var_array();
                            c_var_int* key=MY_NEW c_var_int(1);
                            c_var_string* val=MY_NEW c_var_string("A");
                            a->set(NULL,key,val,false,"");
                            //
                            c_vdcmod_data_t* r=NULL;
                            mod->convert_vars(a,r);
                            // DEBUG: inspect structure hierarchy 'r'
                            mod->_converted_vars_dump("converted2.dump.txt",r);
                            int xx=0; // set breakpoint HERE
                            xx++;
                            break;
                        }
                        case 3:
                        {
                            c_var_array* a=MY_NEW c_var_array();
                            c_var_int* key=MY_NEW c_var_int(1);
                            c_var_array* val=MY_NEW c_var_array();
                            a->set(NULL,key,val,false,"");

                            c_var_int* key2=MY_NEW c_var_int(2);
                            c_var_int* val2=MY_NEW c_var_int(22);
                            val->set(NULL,key2,val2,false,"");
                            //
                            var_keeper->_dump("suite.3.txt");
                            //
                            c_vdcmod_data_t* r=NULL;
                            mod->convert_vars(a,r);
                            mod->_converted_vars_dump("converted3.dump.txt",r);
                            // DEBUG: inspect structure hierarchy 'r'
                            int xx=0; // set breakpoint HERE
                            xx++;
                            break;
                        }
                        case 4:
                        {
                            c_var_array* a=MY_NEW c_var_array();
                            c_var_int* key=MY_NEW c_var_int(1);
                            c_var_array* val=MY_NEW c_var_array();
                            a->set(NULL,key,val,false,"");

                            c_var_int* key2=MY_NEW c_var_int(2);
                            c_var_array* val2=MY_NEW c_var_array();
                            val->set(NULL,key2,val2,false,"");

                            c_var_int* key3=MY_NEW c_var_int(3);
                            c_var_int* val3=MY_NEW c_var_int(33);
                            a->set(NULL,key3,val3,false,"");

                            c_var_string* s1=MY_NEW c_var_string("deep_1");
                            c_var_array* aa=MY_NEW c_var_array();
                            a->set(NULL,s1,aa,false,"");

                            c_var_string* s2=MY_NEW c_var_string("deep_2");
                            c_var_array* aaa=MY_NEW c_var_array();
                            aa->set(NULL,s2,aaa,false,"");

                            c_var_int* key4=MY_NEW c_var_int(4);
                            c_var_int* val4=MY_NEW c_var_int(44);
                            a->set(NULL,key4,val4,false,"");
                            //
                            //var_keeper->_dump("suite.4.txt");
                            //
                            c_vdcmod_data_t* r=NULL;
                            mod->convert_vars(a,r);
                            // DEBUG: inspect structure hierarchy 'r'
                            mod->_converted_vars_dump("converted4.dump.txt",r);
                            int xx=0; // set breakpoint HERE
                            xx++;
                            break;
                        }
                        case 5:
                        {
                            c_var_array* a=super_global;
                            //
                            c_vdcmod_data_t* r=NULL;
                            mod->convert_vars(a,r);
                            // DEBUG: inspect structure hierarchy 'r'
                            mod->_converted_vars_dump("superglobals.dump.txt",r);
                            int xx=0; // set breakpoint HERE
                            xx++;
                            break;
                        }
                        default:
                            test=0;
                    }
                    continue;
                }
#   endif
#endif

                c_vdcmod_data_t* disp=NULL;
                mod->convert_vars(dispatch,disp);
                (*m_func)(disp);
                mod->free_vars(disp);
                disp=NULL;
            }
        }
        return true;
    }

    // NOTREACH
}

bool c_binder::got_match(const string& type, const string& user, const string& channel, const string& line, bool before, c_binder_data& internal_data)
{
    if(m_before!=before)
        return false;

    bool match, has_error;
    string error;

    internal_data.m_matches_type.clear();
    match=m_type.match(type,m_debug,error,has_error,internal_data.m_matches_type,"Argument $type");
    if(has_error) {
        if(m_module && m_module->get_logger())
            m_module->get_logger()->log_print1(E_LOG_DEBUG,LOG_SCOPE_MODULE,"",error);
        return false;
    }
    if(!match)
        return false;

    internal_data.m_matches_user.clear();
    match=m_user.match(user,m_debug,error,has_error,internal_data.m_matches_user,"Argument $user");
    if(has_error) {
        if(m_module && m_module->get_logger())
            m_module->get_logger()->log_print1(E_LOG_DEBUG,LOG_SCOPE_MODULE,"",error);
        return false;
    }
    if(!match)
        return false;

    internal_data.m_matches_channel.clear();
    match=m_channel.match(channel,m_debug,error,has_error,internal_data.m_matches_channel,"Argument $channel");
    if(has_error) {
        if(m_module && m_module->get_logger())
            m_module->get_logger()->log_print1(E_LOG_DEBUG,LOG_SCOPE_MODULE,"",error);
        return false;
    }
    if(!match)
        return false;

    internal_data.m_matches_line.clear();
    match=m_line.match(line,m_debug,error,has_error,internal_data.m_matches_line,"Argument $line");
    if(has_error) {
        if(m_module && m_module->get_logger())
            m_module->get_logger()->log_print1(E_LOG_DEBUG,LOG_SCOPE_MODULE,"",error);
        return false;
    }
    if(!match)
        return false;
    
    return true;
}

c_binder::c_binder(bool ex, const string& type, const string& user, const string& channel, const string& line, bool before, const string& function, c_module* module, vector<string>& error_log, const string& debug)
    : m_type(module,ex,type,debug,error_log,"Argument $type")
    , m_user(module,ex,user,debug,error_log,"Argument $user")
    , m_channel(module,ex,channel,debug,error_log,"Argument $channel")
    , m_line(module,ex,line,debug,error_log,"Argument $line")
    , m_before(before)
    , m_function(function)
    , m_module(module)
    , m_debug(debug)
    , m_is_ex(ex)
    , m_bin_module(false)
    , m_func(NULL)
{
}

c_binder::c_binder(bool ex, const string& type, const string& user, const string& channel, const string& line, bool before, vdcmod_dispatcher_func function, c_module* module, vector<string>& error_log, const string& debug)
    : m_type(module,ex,type,debug,error_log,"Argument $type")
    , m_user(module,ex,user,debug,error_log,"Argument $user")
    , m_channel(module,ex,channel,debug,error_log,"Argument $channel")
    , m_line(module,ex,line,debug,error_log,"Argument $line")
    , m_before(before)
    , m_module(module)
    , m_debug(debug)
    , m_is_ex(ex)
    , m_bin_module(true)
    , m_func(function)
{
}

c_binder_mask::c_binder_mask(c_module* module, bool ex, const string& str, const string& debug, vector<string>& error_log, const char* error_prefix)
{
    m_str=str;
    m_type=0;
    m_error_msg=NULL;
    m_pattern=str;
    m_module=module;

    if(ex) {
        vector<string> flags;
        m_pattern=str;

        parse_string(flags,m_pattern);

        m_error_msg=compile_flags(flags);

        if(NULL!=m_error_msg) {
            error_log.push_back((string)"OOPS: un/bind_ex error: "+error_prefix+"; "+m_error_msg+"; Additional info: "+debug);
            return;
        }

        if((m_type & BM_WC_CI_ASCII) || (m_type & BM_EXACT_CI_ASCII))
            m_pattern=strtoupper(m_pattern,CASE_ASCII);
        if((m_type & BM_WC_CI_RFC) || (m_type & BM_EXACT_CI_RFC))
            m_pattern=strtoupper(m_pattern,CASE_RFC1459);

        if((m_type & BM_WC) || (m_type & BM_EXACT))
            return; // done okay

        c_pcre re(m_pattern);
        if(!re.is_okay()) {
            string error;
            report_pcre_error(m_module,"OOPS",re,error_prefix,debug,error);
            error_log.push_back((string)error_prefix+"; "+m_error_msg);
            return;
        }

        // done okay
    } else
        m_type=BM_WC;
}

void c_binder_mask::parse_string(vector<string>& flags, string& regexp)
{
    flags.clear();
    string flag;
    bool in=false;
    char last='\0';
    for(; regexp.length()>0; regexp.erase(0,1)) {
        if(!in && '@'==regexp[0]) {
            in=true;
            flag+="@";
            if('@'==last)
                break;
            last='@';
            continue;
        }
        if(in && '@'==regexp[0]) {
            in=false;
            flag+="@";
            if(!flag.compare("@@")) {
                regexp.erase(0,1);
                break;
            }
            flags.push_back(flag);
            flag.clear();
            last='\0';
            continue;
        }
        if(in)
            flag+=(string)""+regexp[0];
        else
            break;
    }
}

const char* c_binder_mask::compile_flags(const vector<string>& flags)
{
    m_type=0;

    if(flags.empty())
        return s_err_empty_flags;

    vector<string>::const_iterator i1;
    for(i1=flags.begin(); i1!=flags.end(); ++i1) {
        if(!(*i1).compare("@exact@")) {
            m_type|=BM_EXACT;
            continue;
        }
        if(!(*i1).compare("@exact_ci_ascii@")) {
            m_type|=BM_EXACT | BM_EXACT_CI_ASCII;
            continue;
        }
        if(!(*i1).compare("@exact_ci_rfc@")) {
            m_type|=BM_EXACT | BM_EXACT_CI_RFC;
            continue;
        }
        if(!(*i1).compare("@exact_ci_server@")) {
            m_type|=BM_EXACT | BM_EXACT_CI_SERVER;
            continue;
        }
        if(!(*i1).compare("@preg@")) {
            m_type|=BM_RE | BM_PREG;
            continue;
        }
        if(!(*i1).compare("@mask@")) {
            m_type|=BM_WC;
            continue;
        }
        if(!(*i1).compare("@mask_ci_ascii@")) {
            m_type|=BM_WC | BM_WC_CI_ASCII;
            continue;
        }
        if(!(*i1).compare("@mask_ci_rfc@")) {
            m_type|=BM_WC | BM_WC_CI_RFC;
            continue;
        }
        if(!(*i1).compare("@mask_ci_server@")) {
            m_type|=BM_WC | BM_WC_CI_SERVER;
            continue;
        }
        if(!(*i1).compare("@match@")) {
            m_type|=BM_MATCH | BM_RE | BM_PREG;
            continue;
        }
        if(!(*i1).compare("@match_all@")) {
            m_type|=BM_MATCH_ALL | BM_RE | BM_PREG | BM_PATTERN_ORDER;
            continue;
        }
        if(!(*i1).compare("@set_order@")) {
            m_type&=~BM_PATTERN_ORDER;
            m_type|=BM_SET_ORDER;
            continue;
        }
        if(!(*i1).compare("@pattern_order@")) {
            m_type&=~BM_SET_ORDER;
            m_type|=BM_PATTERN_ORDER;
            continue;
        }
        if(!(*i1).compare("@offset_capture@")) {
            m_type|=BM_OFFSET_CAPTURE | BM_RE | BM_PREG;
            continue;
        }
        if(!(*i1).compare("@@")) {
            break;
        }
    }

    if((m_type & BM_WC) && ( (m_type & (~BM_WC)) & (~(BM_WC_CI_ASCII | BM_WC_CI_RFC | BM_WC_CI_SERVER))))
        return s_err_case_exclusivity;

    if((m_type & BM_EXACT) && ( (m_type & (~BM_EXACT)) & (~(BM_EXACT_CI_ASCII | BM_EXACT_CI_RFC | BM_EXACT_CI_SERVER))))
        return s_err_case_exclusivity;

    if(m_type & (BM_WC & BM_EXACT))
        return s_err_exclusivity;

    if((m_type & BM_WC) && (m_type & BM_EXACT))
        return s_err_exclusivity;

    if((m_type & BM_WC) && (m_type & BM_RE))
        return s_err_exclusivity;

    if((m_type & BM_EXACT) && (m_type & BM_RE))
        return s_err_exclusivity;

    if(BM_WC & m_type)
        return NULL; // okay

    if(BM_EXACT & m_type)
        return NULL; // okay

    if((m_type & BM_MATCH) && (m_type & BM_MATCH_ALL))
        return s_err_both_match_and_match_all;

    if(!(m_type & BM_MATCH) && !(m_type & BM_MATCH_ALL))
        return s_err_neither_match_and_match_all;

    return NULL;
}

const char* c_binder_mask::get_error_str()
{
    return m_error_msg;
}

bool c_binder_mask::match(const string& subject, const string& debug, string& error, bool& has_error, vector<vector<pair<int,string> > >& matches, const char* error_prefix)
{
    MY_ASSERT(NULL==m_error_msg);

    matches.clear();
    has_error=false;

    if(m_type & BM_WC) {
        if(m_type & BM_WC_CI_ASCII) {
            string s=strtoupper(subject,CASE_ASCII);
            return !::match(m_pattern.c_str(),s.c_str());
        } else if(m_type & BM_WC_CI_RFC) {
            string s=strtoupper(subject,CASE_RFC1459);
            return !::match(m_pattern.c_str(),s.c_str());
        } else if(m_type & BM_WC_CI_SERVER) {
            string s1=strtoupper(subject,irc_casemapping);
            string s2=strtoupper(m_pattern,irc_casemapping);
            return !::match(s1.c_str(),s2.c_str());
        } else
            return !::match(m_pattern.c_str(),subject.c_str());
    }

    if(m_type & BM_EXACT) {
        if(m_type & BM_EXACT_CI_ASCII) {
            string s=strtoupper(subject,CASE_ASCII);
            return !m_pattern.compare(s);
        } else if(m_type & BM_EXACT_CI_RFC) {
            string s=strtoupper(subject,CASE_RFC1459);
            return !m_pattern.compare(s);
        } else if(m_type & BM_EXACT_CI_SERVER) {
            string s1=strtoupper(subject,irc_casemapping);
            string s2=strtoupper(m_pattern,irc_casemapping);
            return !s1.compare(s2);
        } else
            return !m_pattern.compare(subject);
    }

    c_pcre re(m_pattern);
    if(!re.is_okay()) {
        MY_ASSERT_NOTREACH();
        has_error=true;
        return false;
    }

    size_t max=1;
    if(m_type & BM_MATCH_ALL)
        max=MAX_SIGNED_INT;

    size_t cnt=re.exec(subject,0,0,matches,max);
    if(re.get_last_error()) {
        int eo;
        string es;
        re.get_compile_error_info(eo,es);

        ostringstream log;
        log << "OOPS: "
            << error_prefix
            << "; PCRE library failed to execute compiled pattern; PCRE error offset: "
            << tos(eo)
            << "; PCRE error string: \""
            << es.c_str()
            << "\"; "
            << "Pattern was: \""
            << m_pattern
            << "\"; Additional info: "
            << debug.c_str();
        error=log.str();
        has_error=true;
        return false;
    }

    return 0!=cnt;
}

bool c_binder_mask::get_match_array(const string& debug, string& error, bool& has_error, c_var_array* &matches_array, const vector<vector<pair<int,string> > >& matches)
{
    has_error=false;
    matches_array=NULL;

    if(m_type & BM_WC)
        return true;

    if(m_type & BM_EXACT)
        return true;

    MY_LONG_LONG flags=0;

    if(m_type & BM_SET_ORDER)
        flags|=PREG_SET_ORDER;

    if(m_type & BM_PATTERN_ORDER)
        flags|=PREG_PATTERN_ORDER;

    if(m_type & BM_OFFSET_CAPTURE)
        flags|=PREG_OFFSET_CAPTURE;

    if(m_type & BM_MATCH_ALL) {
        if(php_do_preg_match_all(m_module,debug,matches_array,flags,error,matches))
            return true;
    } else if(m_type & BM_MATCH) {
        if(php_do_preg_match(m_module,debug,matches_array,flags,error,matches))
            return true;
    } else {
        error="Internal error in " __FILE__"::";
        error+=__my_func__;
        error+=(string)":"+tos(__LINE__);
        error+=" - PLEASE REPORT THIS PROBLEM!";

        has_error=true;
        MY_ASSERT_NOTREACH();

        return false;
    }

    has_error=true;
    return false;
}

bool c_binder_mask::operator== (const c_binder_mask& x) const
{
    return m_type==x.m_type && !m_str.compare(x.m_str);
}

bool c_binder::operator== (const c_binder& x) const
{
    return
        m_is_ex==x.m_is_ex
        && m_before==x.m_before
        && !m_function.compare(x.m_function)
        && m_type==x.m_type
        && m_user==x.m_user
        && m_channel==x.m_channel
        && m_line==x.m_line
        ;
}

