/***************************************************************************
                          c_binder.h  -  description
                             -------------------
    begin                : Thu May 10 2007
    copyright            : (C) 2004-2007 by Marian VooDooMan Meravy
    email                : ghostvoodooman -a.T= users (dot) sourceforge.net
 ***************************************************************************/

/* $Id: c_binder.h,v 1.15 2008/08/23 15:10:45 ghostvoodooman Exp $ */

/***************************************************************************

VooDoo cIRCle - an IRC (ro)bot
Copyright (C) 2004-2008 by Marian VooDooMan Meravy
  (ghostvoodooman -a.T= users (dot) sourceforge.net)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************/

/*!
    \file
    \brief Class 'c_binder', 'c_binder_mask' and 'c_binder_data'
    \version $Id: c_binder.h,v 1.15 2008/08/23 15:10:45 ghostvoodooman Exp $
*/

#ifndef C_BINDER_H__
#define C_BINDER_H__

#if defined(HAVE_CONFIG_H)
#   include <config.h>
#endif

#include <string>
#include <vector>
#include <stack>
#include <list>
#include <map>
#include <algorithm>

#include "../utils/memleak.h"

#include "../params.h"

#include "../vdcmod.h"

class c_module;
class c_var_array;

/*!
    \brief Stores compiled data for mask/regexp
    \author VooDooMan
    \version 1
    \date 2008-08-15
*/
class c_binder_mask
{
    string m_str;                   //!< User-supplied mask/regexp (full string)
    string m_pattern;               //!< User-supplied wildcard mask/regexp pattern

    unsigned m_type;                //!< OR'ed flags of 'e_type' type

    c_module* m_module;             //!< Module descriptor

    const char* m_error_msg;        //!< If non-empty, there was and error in parsing string, and there is an error message for human

    /*!
        \brief Parses the string for flags and regexp
        \author VooDooMan
        \version 1
        \date 2008-08-15
        \param flags Returns list of flags
        \param regexp IN: string to parse, OUT: Regexp (tail after flags)
    */
    void parse_string(vector<string>& flags, string& regexp);

    /*!
        \brief Validates and compiles flags
        \author VooDooMan
        \version 1
        \date 2008-08-15
        \param flags To validate
        \return Ruturns 'NULL' for okay, for error there is an error message for user
    */
    const char* compile_flags(const vector<string>& flags);

    /*!
        \brief Type of this entry, OR'ed flags
        \author VooDooMan
        \version 1
        \date 2008-08-15
    */
    enum e_type {
        BM_EXACT = 1U << 0,             //!< Exact string required
        BM_WC = 1U << 1,                //!< Wild-carded mask
        BM_RE = 1U << 2,                //!< Regexp
        BM_PREG = 1U << 3,              //!< Whether this is prce regexp
        BM_MATCH = 1U << 4,             //!< Whether we want to use 'match'
        BM_MATCH_ALL = 1U << 5,         //!< Whether we mant to use 'match_all'
        BM_OFFSET_CAPTURE = 1U << 6,    //!< Whether we want to use capture offsets in matching
        BM_EXACT_CI_ASCII = 1U << 7,    //!< In addition to 'BM_EXACT', case in-sensitive (ASCII case-mapping) compare is required
        BM_WC_CI_ASCII = 1U << 8,       //!< In addition to 'BM_WC', case in-sensitive (ASCII case-mapping) compare is required
        BM_EXACT_CI_RFC = 1U << 9,      //!< In addition to 'BM_EXACT', case in-sensitive (IRC RFC case-mapping) compare is required
        BM_WC_CI_RFC = 1U << 10,        //!< In addition to 'BM_WC', case in-sensitive (IRC RFC case-mapping) compare is required
        BM_EXACT_CI_SERVER = 1U << 11,  //!< In addition to 'BM_EXACT', case in-sensitive (current IRC server case-mapping) compare is required
        BM_WC_CI_SERVER = 1U << 12,     //!< In addition to 'BM_WC', case in-sensitive (current IRC server case-mapping) compare is required
        BM_SET_ORDER = 1U << 13,        //!< PREG_SET_ORDER
        BM_PATTERN_ORDER = 1U << 14,    //!< PREG_PATTERN_ORDER
    };

public:
    /*!
        \brief C-tor
        \author VooDooMan
        \version 1
        \date 2008-08-15
        \param regexp 'true' when str is regexp, 'false' for wildcard mask
        \param str User-supplied mask/regexp
        \param module Current module
        \param debug Debug string for user
        \param error_log Returns log of errors (must be clean before call)
        \param error_prefix Prefix for error
    */
    c_binder_mask(c_module* module, bool regexp, const string& str, const string& debug, vector<string>& error_log, const char* error_prefix);

    /*!
        \brief Retrieves a human-readable error
        \author VooDooMan
        \version 1
        \date 2008-08-15
        \return Returns error string, or 'NULL' if okay
    */
    const char* get_error_str();

    /*!
        \brief Matches string agains mask/regexp
        \author VooDooMan
        \version 1
        \date 2008-08-15
        \param subject String to match
        \param debug Debug string
        \param error Error message if 'has_error' returns 'true', else undefined
        \param has_error Returns 'true' if an error occurred
        \param matches Returns internal state information to be passed to subsequent call to 'get_match_array()'
        \param error_prefix Prefix for error
        \return Returns 'true' if match was successful
    */
    bool match(const string& subject, const string& debug, string& error, bool& has_error, vector<vector<pair<int,string> > >& matches, const char* error_prefix);

    /*!
        \brief Matches string agains mask/regexp
        \author VooDooMan
        \version 1
        \date 2008-08-15
        \param debug Debug string
        \param error Error message if 'has_error' returns 'true'
        \param has_error Returns 'true' if an error occurred
        \param matches_array Returns an array with matches
        \param matches Internal state information generated by call to 'match()'
        \warning Should be called only when prior call to 'match()' returned 'true'
    */
    bool get_match_array(const string& debug, string& error, bool& has_error, c_var_array* &matches_array, const vector<vector<pair<int,string> > >& matches);

    /*!
        \brief Comparsion operator
        \author VooDooMan
        \version 1
        \date 2008-08-15
        \param x Right operand
    */
    bool operator== (const c_binder_mask& x) const;
};

/*!
    \brief Stores internal state information for one match
    \author VooDooMan
    \version 1
    \date 2008-08-15
*/
class c_binder_data
{
    friend class c_binder;

    vector<vector<pair<int,string> > > m_matches_type;
    vector<vector<pair<int,string> > > m_matches_user;
    vector<vector<pair<int,string> > > m_matches_channel;
    vector<vector<pair<int,string> > > m_matches_line;
};

/*!
    \brief Stores bound events
    \author VooDooMan
    \version 2
    \date 2005, 2008-08-15
*/
class c_binder
{
    friend class c_binder_data;

public:
    c_binder_mask m_type;               //!< Type mask
    c_binder_mask m_user;               //!< User mask
    c_binder_mask m_channel;            //!< Channel mask
    c_binder_mask m_line;               //!< Line mask
    bool m_before;                      //!< If to throw before processing
    string m_function;                  //!< Name of function as handler
    c_module* m_module;                 //!< Module
    string m_debug;                     //!< Debug string

    bool m_is_ex;                       //!< Whether this is "bind_ex"

    bool m_bin_module;                  //!< Is this binary module bind?
    vdcmod_dispatcher_func m_func;      //!< Binary module dispatcher

    /*!
        \brief Constructor
        \author VooDooMan
        \version 2
        \date 2005, 2008-08-15
        \param ex 'true' for extended bind
        \param type Type mask
        \param user User mask
        \param channel Channel mask
        \param line Line mask
        \param before If to throw before processing
        \param function Name of function as handler
        \param debug Debug string
        \param module Module
        \param error_log Returns log of errors (must be clean before call)
    */
    c_binder(bool ex, const string& type, const string& user, const string& channel, const string& line, bool before, const string& function, c_module* module, vector<string>& error_log, const string& debug);

    /*!
        \brief Constructor
        \author VooDooMan
        \version 2
        \date 2005, 2008-08-15
        \param ex 'true' for extended bind
        \param type Type mask
        \param user User mask
        \param channel Channel mask
        \param line Line mask
        \param before If to throw before processing
        \param function Binary module's dispatcher
        \param debug Debug string
        \param module Module
        \param error_log Returns log of errors (must be clean before call)
    */
    c_binder(bool ex, const string& type, const string& user, const string& channel, const string& line, bool before, vdcmod_dispatcher_func function, c_module* module, vector<string>& error_log, const string& debug);

    /*!
        \brief Matches this bind against masks
        \author VooDooMan
        \version 2
        \date 2005
        \param dispatch Dispatch structure
        \param internal_data Pass there internal state information from prior call to 'got_match()'
        \warning Should be called only if there was prior call to 'got_match()'
        \retval false An error occurred, and has been logged, bound event function was not called
        \retval true Bound event function was called
    */
    bool call_bound_dispatch_event(c_var_array* dispatch, const c_binder_data& internal_data);

    /*!
        \brief Checks whether at least one bound event matches
        \author VooDooMan
        \version 1
        \date 2005
        \param type Type of event
        \param user User of event
        \param channel Channel of event
        \param line Line of event
        \param before If to trigger before processing
        \param internal_data Returns internal state information for subsequent call to 'call_bound_dispatch_event()'
        \return Returns true if there is a match
    */
    bool got_match(const string& type, const string& user, const string& channel, const string& line, bool before, c_binder_data& internal_data);

    /*!
        \brief Comparsion operator
        \author VooDooMan
        \version 1
        \date 2008-08-15
        \param x Right operand
    */
    bool operator== (const c_binder& x) const;
};

#endif

