/***************************************************************************
                          c_module.cpp  -  description
                             -------------------
    begin                : Thu May 10 2007
    copyright            : (C) 2004-2007 by Marian VooDooMan Meravy
    email                : ghostvoodooman -aT= users (dot) sourceforge.net
 ***************************************************************************/

/* $Id: c_module.cpp,v 1.26 2008/08/23 15:10:45 ghostvoodooman Exp $ */

/***************************************************************************

VooDoo cIRCle - an IRC (ro)bot
Copyright (C) 2004-2008 by Marian VooDooMan Meravy
  (ghostvoodooman -aT= users (dot) sourceforge.net)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************/

/*!
    \file
    \brief Class 'c_module'
    \version $Id: c_module.cpp,v 1.26 2008/08/23 15:10:45 ghostvoodooman Exp $
*/

#if defined(HAVE_CONFIG_H)
#   include <config.h>
#endif

#include <iostream>

#include "../params.h"
#include "../utils/memleak.h"
#include "../utils/assert.h"
#include "../compat.h"

#include "../sock_shared.h"
#include "c_module.h"
#include "../microcode/c_micro_symbol_variable_table.h"
#include "../var_keeper.h"
#include "../modules.h"
#include "../botnet.h"
#include "../var/c_var_string.h"
#include "../var/c_var_unset.h"
#include "../logic.h"
#include "../microcode/c_micro_error_func_internal_1.h"
#include "../microcode/c_micro_error_func_internal_3.h"
#include "../eval.h"
#include "../log.h"
#include "../filesys.h"
#include "../irc.h"
#include "../stats.h"
#include "../irc/c_online_channel.h"
#include "../irc/c_online_user.h"
#include "../var/c_var_bool.h"
#include "../var/c_var_int.h"
#include "../internal_state.h"
#include "../log/c_logger.h"
#include "../misc/c_script_timeout_protector_locker.h"

extern deque<c_user> r_terminators;
extern deque<c_user> r_user_templates;
extern deque<c_user> r_users;

#include "../utils/precomp.h"

#if defined(_MSC_VER) || defined(__BORLANDC__)
#   pragma hdrstop
#endif

MY_STATIC void log_fn_error(c_var_keeper::e_log_severity UNUSEDARG(severity), const string& msg, void* data)
{
    c_module* m=(c_module*)data;
    if(m && m->get_logger())
        m->get_logger()->log_print1(E_LOG_DEBUG,LOG_SCOPE_DEFAULT,"",msg);
}

bool c_module::parse_logic(int& error_on_line, string& error_reason, vector<string>& ignoring_procs)
{
    string logic_txt=dir_prefix;
    if(logic_txt.length()==0)
        logic_txt="." FILE_SLASH;
    if(logic_txt.length() && logic_txt[logic_txt.length()-1]!=FILE_SLASH_CH)
        logic_txt+=FILE_SLASH;
    logic_txt+="logic.txt";
    error_on_line=1;
    error_reason="";
    return !load_logic(error_on_line,error_reason,ignoring_procs);
}

bool c_module::parse_conf(bool is_core)
{
    string conf_txt=dir_prefix;
    if(conf_txt.length()==0)
        conf_txt="." FILE_SLASH;
    if(conf_txt.length() && conf_txt[conf_txt.length()-1]!=FILE_SLASH_CH)
        conf_txt+=FILE_SLASH;
    conf_txt+="conf.txt";
    conf->clear();
    return conf->init(conf_txt,logger,is_core);
}

c_module::c_module(const c_module& orig)
{
    name=orig.name;
    loaded=orig.loaded;

    // do NOT copy-construct these 2 classes! just copy the pointers!
    conf=orig.conf;
    logger=orig.logger;

    if(orig.global) {
        global=var_keeper->register_mod_global(NULL,__FILE__,__my_func__,__LINE__);
        global=MY_NEW c_var_array(*orig.global);
    } else
        global=NULL;

    files=orig.files;

    session_=orig.session_;

    local_id=0;
}

c_module::~c_module()
{
    unload();

    {
        vector<c_procedure>::iterator i;
        while((i=r_procedures.begin())!=r_procedures.end())
            r_procedures.erase(i);
    }
    {
        map<string,c_function*>::iterator i;
        while((i=r_functions.functions.begin())!=r_functions.functions.end()) {
            (*i).second->clear_microcode();
            (*i).second->destroy(true,true);
            MY_DELETE((*i).second);
            r_functions.functions.erase(i);
        }
    }
}

c_module::e_rc c_module::load(const string& mod_name, const string& so_name)
{
    if(loaded)
        return RC_W_LOADED;

    name="";

    {
        vector<c_module*>::iterator i1;
        for(i1=modules.begin(); i1!=modules.end(); ++i1) {
            if(!cmp_strings_case_insensitive(mod_name,(*i1)->get_mod_name())) {
                return RC_E_DUPLICATE_MODULE;
            }
        }
    }

    {
        bool ok=!mod_name.empty();
        size_t i1;
        for(i1=0; i1<mod_name.length(); ++i1) {
            char ch=mod_name[i1];
            if( (ch>='a' && ch<='z') || (ch>='A' && ch<='Z') || (ch>='0' && ch<='9') || ('_'==ch) || ('-'==ch) ) {
                // ok
            } else {
                ok=false;
                break;
            }
        }
        for(i1=0; i1<so_name.length(); ++i1) {
            char ch=so_name[i1];
            if( (ch>='a' && ch<='z') || (ch>='A' && ch<='Z') || (ch>='0' && ch<='9') || ('_'==ch) || ('-'==ch) || ('.'==ch) ) {
                // ok
            } else {
                ok=false;
                break;
            }
        }
        if(!so_name.compare(".") || !so_name.compare(".."))
            ok=false;
        if(!ok)
            return RC_E_INV_MODNAME;
    }

    name=mod_name;

    bool is_core=false;
    if(!mod_name.compare("core")) {
        dir_prefix="." FILE_SLASH;
        is_core=true;

        core_mod=this;
    } else {
        dir_prefix="." FILE_SLASH "modules" FILE_SLASH;
        dir_prefix+=mod_name;
        dir_prefix+=FILE_SLASH;
    }

    if(conf)
        MY_DELETE(conf);
    conf=MY_NEW c_conf();

    if(logger)
        MY_DELETE(logger);
    logger=MY_NEW c_logger();

    if(!parse_conf(is_core)) {
        string log="Failed to load module \"";
        log+=name;
        log+="\"'s \"conf.txt\" file";
        c_logger* l=get_core_logger();
        if(l) {
            l->log_print1(E_LOG_BOT,LOG_SCOPE_GLOBAL,"",log,true);
            l->log_print1(E_LOG_DEBUG,LOG_SCOPE_GLOBAL,"",log,true);
        }
        return RC_E_CONF_ERROR;
    }

    session_id_t old_session=var_keeper->get_current_session();

    session_=++last_session;
    sessions.push_back(session_);

    var_keeper->set_current_session(session_);

    if(global) {
        global->set_session(session_);
        var_keeper->free_global(global,true);
        global=NULL;
    }
    global=var_keeper->register_mod_global(NULL,__FILE__,__my_func__,__LINE__);
    global->set_persistent();

    string mod_logs=dir_prefix;
    mod_logs+="logs";
#if (defined(_WIN32) || defined(_WIN64)) && !defined(__CYGWIN__)
    mkdir(mod_logs.c_str());
#else
    {
        string tmp="." FILE_SLASH;
        tmp+=mod_logs;
        mkdir(tmp.c_str(),0777);
    }
#endif

    loaded=false;
    logger->init("." FILE_SLASH "logs",mod_logs,mod_name,LOG_SCOPE_MODULE,get_core_conf(),conf);

    logger->log_print1(E_LOG_BOT,LOG_SCOPE_GLOBAL | LOG_SCOPE_MODULE,"","Loading module",true);
    logger->log_print1(E_LOG_DEBUG,LOG_SCOPE_GLOBAL | LOG_SCOPE_MODULE,"","Loading module",true);

    r_procedures.clear();
    r_functions.clear();

    int error_on_line;
    string error_reason;
    vector<string> ignoring_procs;

    c_var_array* base=NULL;
    unsigned l_id=local_id=var_keeper->register_local(base,NULL,__FILE__,__my_func__,__LINE__);
    unsigned old_local=var_keeper->set_current_local(l_id);

    if(!parse_logic(error_on_line,error_reason,ignoring_procs)) {
        var_keeper->free_local(l_id,true);
        var_keeper->set_current_local(old_local);
        var_keeper->set_current_session(old_session);
        string error="logic.txt:";
        error+=tos(error_on_line);
        error+=": ";
        error+=error_reason;
        logger->log_print1(E_LOG_BOT,LOG_SCOPE_MODULE,"",error,true);
        logger->log_print1(E_LOG_DEBUG,LOG_SCOPE_MODULE,"",error,true);
        return RC_E_LOGIC_ERROR;
    }
    var_keeper->set_current_local(old_local);
    var_keeper->set_current_session(old_session);

    {
        vector<string>::iterator i1;
        for(i1=ignoring_procs.begin(); i1!=ignoring_procs.end(); ++i1) {
            string msg="Ignoring procedure/function, it is locked: ";
            msg+=*i1;
            logger->log_print1(E_LOG_BOT,LOG_SCOPE_MODULE,"",msg);
        }
    }

    loaded=true;

    if(is_core) {
        var_keeper->set_core_module_working_session(session_);
        session_=SESSION_CORE_ENTRY_ID;
    }

    {
        bool got=false;
        for(vector<c_module*>::iterator i1=modules.begin(); i1!=modules.end(); ++i1)
            if((*i1)==this) {
                got=true;
                break;
            }
        if(!got)
            modules.push_back(this);
    }

    if(is_core) // again, beacuse we need to update core logger (peviously NULL)
        logger->init("." FILE_SLASH "logs",mod_logs,mod_name,LOG_SCOPE_MODULE,get_core_conf(),conf);

    logger->log_print1(E_LOG_BOT,LOG_SCOPE_GLOBAL | LOG_SCOPE_MODULE,"","Module loaded",true);
    logger->log_print1(E_LOG_DEBUG,LOG_SCOPE_GLOBAL | LOG_SCOPE_MODULE,"","Module loaded",true);

    update_groups();

    return RC_OK;
}

/*!
    \brief Calls startup routine of module
    \author VooDooMan
    \version 1
    \retval true OK
    \retval false Procedure/function not found
*/
bool c_module::call_startup()
{
    var_keeper->set_current_session(session_);

    c_script_timeout_protector_locker stp_lock;

    string call_string="_sys_startup($mod_name)";
    vector<string> string_vars;
    string_vars.push_back(name);
    bool res=execute(call_string,string_vars,call_string);

    var_keeper->set_current_session(SESSION_RESERVED_ID);

    return res;
}

void c_module::unload()
{
    if(!name.compare("core")) {
        core_mod=NULL;
    }

    if(loaded && logger) {
        logger->log_print1(E_LOG_BOT,LOG_SCOPE_MODULE | LOG_SCOPE_GLOBAL,"","Unloading module",true);
        logger->log_print1(E_LOG_DEBUG,LOG_SCOPE_MODULE | LOG_SCOPE_GLOBAL,"","Unloading module",true);
    }

    if(!name.compare("core")) { // important again!
        core_mod=NULL;
    }

    {
    erase_1:
        vector<c_timer*>::iterator i1=r_timers.begin();
        if(i1!=r_timers.end()) {
            if((*i1)->bot_unlink_timer)
                botnet_kill_timer(*i1);
            if((*i1)->vars) {
                var_keeper->free_global((*i1)->vars,true);
            }
            (*i1)->vars=NULL;
            MY_DELETE(*i1);
            r_timers.erase(i1);
            goto erase_1;
        }
    }
    {
        tmp_files_garbage.insert(tmp_files_garbage.end(),tmp_files.begin(),tmp_files.end());
        tmp_files.clear();
    }
    botnet_kill_module(this);
    if(logger)
        MY_DELETE(logger);
    logger=NULL;
    if(conf)
        MY_DELETE(conf);
    conf=NULL;

    if(global) {
        var_keeper->free_global(global,true);
        global=NULL;
    }

    map<int,FILE*>::iterator i1;
    for(i1=files.begin(); i1!=files.end(); ++i1)
        fclose((*i1).second);
    files.clear();

    {
        vector<c_module*>::iterator i1;
        for(i1=modules.begin(); i1!=modules.end(); ++i1)
            if(!(*i1)->get_mod_name().compare(get_mod_name())) {
                modules.erase(i1);
                break;
            }
    }

    if(!name.compare("core")) { // important again!
        core_mod=NULL;
    }

    if(local_id)
        var_keeper->free_local(local_id,true);

    if(!loaded)
        return;

    loaded=false;

    // TODO
}

bool c_module::execute(string func, vector<string>& string_vars, const string& debug)
{
    var_keeper->set_current_session(session_);

    string name_only=func;
    if(name_only.find("(",0)!=string::npos)
        name_only.erase(name_only.find("(",0),name_only.length()-name_only.find("(",0));

    vector<string> child_vars;
    {
        string f=func;
        if(f.find("(",0)!=string::npos)
            f.erase(0,f.find("(",0)+1);
        if(f.length() && ')'==f[f.length()-1])
            f.erase(f.length()-1,1);
        explode(f,',',child_vars);
    }

    {
        // try procedure first
        vector<c_procedure>::iterator i1;
        bool got=false;
        for(i1=r_procedures.begin(); i1!=r_procedures.end(); ++i1) {
            (*i1).parse_call_string();
            if(!name_only.compare((*i1).name_only)) {
                got=true;
                break;
            }
        }
        if(got) {
            c_var_array* vars;
            c_temp_local tl(NULL);
            unsigned new_local_id=tl.get_params(vars);
            vector<string>::iterator i2, i3, i5;
            bool exc=false;
            for(i2=string_vars.begin(), i3=(*i1).args.begin(), i5=child_vars.begin(); i2!=string_vars.end() && i3!=(*i1).args.end(); ++i2, ++i3) {
                c_index_base* idx=MY_NEW c_index_str(*i3);
                c_var_string* val=MY_NEW c_var_string(*i2);
                try {
                    vars->set(this,idx,val,false,debug);
                } catch(c_var_exception e) {
                    e.print_oops_to_log(debug);
                    MY_DELETE(val);
                    exc=true;
                    break;
                }
                if(i5!=child_vars.end())
                    i5++;
            }
            if(exc) {
                return true;
            }
            {
                auto_ptr<c_var_array> dummy_vars(MY_NEW c_var_array);
                for( ; i3!=(*i1).args.end(); ++i3) {
                    c_index_base* idx=MY_NEW c_index_str(*i3);
                    string val_;
                    if(i5!=child_vars.end())
                        val_=*i5;
                    c_var_string* val=MY_NEW c_var_string(logic_eval(this,val_,dummy_vars.get(),debug));
                    try {
                        vars->set(this,idx,val,false,debug);
                    } catch(c_var_exception e) {
                        e.print_oops_to_log(debug);
                        MY_DELETE(val);
                        exc=true;
                        break;
                    }
                    if(i5!=child_vars.end())
                        i5++;
                }
                if(exc) {
                    return false;
                }
            }
            // we need to make temporary copy, because stl do mess after rehash
            c_procedure p=*i1;
            // call it!
            exec(p,vars,new_local_id);
            return true;
        }
    }

    c_var_array* vars;
    c_temp_local tl(NULL);
    (void)tl.get_params(vars);
    {
        string func_name=func;
        size_t pos=func_name.find("(",0);
        if(pos!=string::npos)
            func_name.erase(pos,func_name.length()-pos);

        vector<c_func_arg> args;
        if(!r_functions.get_function_prototype(func_name,args)) {
            string d="Function/procedure '";
            d+=func_name;
            d+="' doesn't exist - ";
            d+=debug;
            if(logger)
                logger->log_print1(E_LOG_DEBUG,LOG_SCOPE_MODULE,"",d);
            return false;
        }

        if(args.size()!=string_vars.size()) {
            string d="Function '";
            d+=func_name;
            d+="' argument count mismatch - ";
            d+=debug;
            if(logger)
                logger->log_print1(E_LOG_DEBUG,LOG_SCOPE_MODULE,"",d);
            return false;
        }
        {
            c_module* this_module=this;
            string func_name_=name_only;
            c_micro_error* error=NULL;

            vector<c_func_arg> fn_args;
            if(this_module->r_functions.get_function_prototype(func_name_,fn_args)) {
                c_function* new_fn;
                bool dummy_b=false;
                if(!this_module->get_function(func_name_,dummy_b,new_fn) || dummy_b) {
                    error=MY_NEW c_micro_error_func_internal_1(func_name_);
                    if(get_logger())
                        get_logger()->log_print1(E_LOG_DEBUG,LOG_SCOPE_DEFAULT,"",error->get_msg());
                    MY_DELETE(error);
                    return false;
                }

                // child instance
                c_var_array* new_vars;
                c_temp_local tl(NULL);
                unsigned new_local_id=tl.get_params(new_vars);
                c_micro_symbol_table* new_symbol_table=new_fn->symbol_table;

                vector<c_func_arg> args1, args2;
                vector<string>::iterator ai;
                vector<c_func_arg>::iterator afi;
                long arg_c=0;
                for(ai=string_vars.begin(), afi=fn_args.begin(); ai!=string_vars.end() && afi!=fn_args.end(); ++afi, ++ai, ++arg_c) {
                    c_func_arg a;
                    a.arg_expression=(*afi).arg_expression;
                    a.reference=(*afi).reference;
                    c_var_base* val=MY_NEW c_var_string(*ai);
                    if(!new_symbol_table->get_var_table()->set((*afi).value_sym,val,false,error,MICRO_NA,false,this,"",debug)) {
                        if(get_logger())
                            get_logger()->log_print1(E_LOG_DEBUG,LOG_SCOPE_DEFAULT,"",error->get_msg());
                        MY_DELETE(error);

                        MY_DELETE(val);

                        MY_DELETE(new_symbol_table);

                        MY_DELETE(new_fn);

                        return false;
                    }
                    c_index_base* idx=MY_NEW c_index_str((*afi).arg_expression);
                    if(!new_vars->set(this_module,idx,val,false,debug)) {
                        MY_DELETE(val);

                        MY_DELETE(new_symbol_table);

                        MY_DELETE(new_fn);

                        string msg="Error setting argument (zero-based) #";
                        msg+=tos(arg_c);
                        msg+=" to new function's internal variables";

                        error=MY_NEW c_micro_error_func_internal_3(func_name_,msg);
                        if(get_logger())
                            get_logger()->log_print1(E_LOG_DEBUG,LOG_SCOPE_DEFAULT,"",error->get_msg());
                        MY_DELETE(error);

                        return false;
                    }
                    a.value=val;
                    a.value_sym=(*afi).value_sym;
                    args1.push_back(a);
                    args2.push_back(a);
                }
                c_eval_iteration ste(this_module,new_vars,debug);
                ste.state=1000;
                ste.return_to_state=0;
                ste.fn=new_fn;
                ste.fn->create_symbols();
                ste.fn_destroy=true;
                ste.instruct_id=0;
                ste.symbol_table_=new_symbol_table;
                ste.return_value=NULL;
                ste.symbol_table_->update_symbols_from_vars(new_vars);
                ste.local_id=new_local_id;

                stack<c_eval_iteration> st;
                st.push(ste);
                eval_loop(st);
            }
        }
    }
    return true;
}

bool c_module::execute(string func, c_var_array* caller_vars, const string& debug, bool core_module_entrypoint)
{
    var_keeper->set_current_session(session_);

    string name_only=func;
    if(name_only.find("(",0)!=string::npos)
        name_only.erase(name_only.find("(",0),name_only.length()-name_only.find("(",0));

    vector<string> child_vars;
    {
        string f=func;
        if(f.find("(",0)!=string::npos)
            f.erase(0,f.find("(",0)+1);
        if(f.length() && ')'==f[f.length()-1])
            f.erase(f.length()-1,1);
        explode(f,',',child_vars);
    }

    {
        // try procedure first
        vector<c_procedure>::iterator i1;
        bool got=false;
        for(i1=r_procedures.begin(); i1!=r_procedures.end(); ++i1) {
            (*i1).parse_call_string();
            if(!name_only.compare((*i1).name_only)) {
                got=true;
                break;
            }
        }
        if(got) {
            c_var_array* vars;
            c_temp_local tl(NULL);
            unsigned new_local_id=tl.get_params(vars);
            bool exc=false;
            vector<string>::iterator i2, i3;
            for(i3=(*i1).args.begin(), i2=child_vars.begin(); i3!=(*i1).args.end() && i2!=child_vars.end(); ++i2, ++i3) {
                c_index_base* idx=MY_NEW c_index_str(*i3);
                c_var_base* val=NULL;
                if((*i3).length() && '$'==(*i3)[0]) {
                    try {
                        val=caller_vars->get(this,idx,false,debug);
                    } catch(c_var_exception e) {
                        e.print_oops_to_log(debug);
                        exc=true;
                    }
                }
                if(!val)
                    val=MY_NEW c_var_string(logic_eval(this,*i2,caller_vars,debug));
                else
                    val=val->copy();
                if(!exc) try {
                    vars->set(this,idx,val,false,debug);
                } catch(c_var_exception e) {
                    e.print_oops_to_log(debug);
                    MY_DELETE(val);
                    exc=true;
                    break;
                }
            }
            if(exc) {
                return false;
            }
            for( ; i3!=(*i1).args.end(); ++i3) {
                c_index_base* idx=MY_NEW c_index_str(*i3);
                c_var_string* val=MY_NEW c_var_string("");
                try {
                    vars->set(this,idx,val,false,debug);
                } catch(c_var_exception e) {
                    e.print_oops_to_log(debug);
                    MY_DELETE(val);
                    exc=true;
                    break;
                }
            }
            if(exc) {
                return false;
            }
            // we need to make temporary copy, because stl do mess after rehash
            c_procedure p=*i1;
            // call it!
            exec(p,vars,new_local_id,core_module_entrypoint);
            return true;
        }
    }

    c_var_array* vars;
    c_temp_local tl(NULL);
    unsigned new_local_id=tl.get_params(vars);
    {
        string func_name=func;
        size_t pos=func_name.find("(",0);
        if(pos!=string::npos)
            func_name.erase(pos,func_name.length()-pos);

        vector<c_func_arg> args;
        if(!r_functions.get_function_prototype(func_name,args)) {
            string d="Function/procedure '";
            d+=func_name;
            d+="' doesn't exist - ";
            d+=debug;
            if(logger)
                logger->log_print1(E_LOG_DEBUG,LOG_SCOPE_MODULE,"",d);
            return false;
        }

        if(args.size()!=child_vars.size()) {
            string d="Function '";
            d+=func_name;
            d+="' argument count mismatch - ";
            d+=debug;
            if(logger)
                logger->log_print1(E_LOG_DEBUG,LOG_SCOPE_MODULE,"",d);
            return false;
        }

        {
            c_module* this_module=this;
            string func_name_=name_only;
            c_micro_error* error=NULL;

            vector<c_func_arg> fn_args;
            if(this_module->r_functions.get_function_prototype(func_name_,fn_args)) {
                c_function* new_fn;
                bool dummy_b=false;
                if(!this_module->get_function(func_name_,dummy_b,new_fn) || dummy_b) {
                    error=MY_NEW c_micro_error_func_internal_1(func_name_);
                    if(get_logger())
                        get_logger()->log_print1(E_LOG_DEBUG,LOG_SCOPE_DEFAULT,"",error->get_msg());
                    MY_DELETE(error);
                    return false;
                }

                // child instance
                c_var_array* new_vars=vars;
                c_micro_symbol_table* new_symbol_table=new_fn->symbol_table;

                vector<c_func_arg> args1, args2;
                vector<c_func_arg>::iterator afi;
                long arg_c=0;
                for(afi=fn_args.begin(); afi!=fn_args.end(); ++afi, ++arg_c) {
                    c_func_arg a;
                    a.arg_expression=(*afi).arg_expression;
                    a.reference=(*afi).reference;
                    c_index_base* idx=MY_NEW c_index_str((*afi).arg_expression);
                    c_var_base* val=NULL;
                    if((*afi).arg_expression.length() && '$'==(*afi).arg_expression[0]) {
                        try {
                            val=caller_vars->get(this,idx,false,debug);
                        } catch(c_var_exception e) {
                            MY_DELETE(new_fn);

                            e.print_oops_to_log(debug);
                            return false;
                        }
                    }
                    if(!(*afi).reference) {
                        if(val)
                            val=val->copy();
                    } else {
                        c_var_base* new_val=NULL;
                        if(val)
                            new_val=val->copy();
                        if(val && new_val)
                            var_keeper->add_reference(val->get_id(),new_val->get_id(),&log_fn_error,this);
                        val=new_val;
                    }
                    if(!new_symbol_table->get_var_table()->set((*afi).value_sym,val,false,error,MICRO_NA,false,this,"",debug)) {
                        if(get_logger())
                            get_logger()->log_print1(E_LOG_DEBUG,LOG_SCOPE_DEFAULT,"",error->get_msg());
                        MY_DELETE(error);

                        MY_DELETE(val);

                        MY_DELETE(new_symbol_table);

                        MY_DELETE(new_fn);

                        return false;
                    }
                    if(!new_vars->set(this_module,idx,val,false,debug)) {
                        MY_DELETE(val);

                        MY_DELETE(new_symbol_table);

                        MY_DELETE(new_fn);

                        string msg="Error setting argument (zero-based) #";
                        msg+=tos(arg_c);
                        msg+=" to new function's internal variables";

                        error=MY_NEW c_micro_error_func_internal_3(func_name_,msg);
                        if(get_logger())
                            get_logger()->log_print1(E_LOG_DEBUG,LOG_SCOPE_DEFAULT,"",error->get_msg());
                        MY_DELETE(error);

                        return false;
                    }
                    a.value=val;
                    a.value_sym=(*afi).value_sym;
                    args1.push_back(a);
                    args2.push_back(a);
                }
                c_eval_iteration ste(this_module,new_vars,debug);
                ste.state=1000;
                ste.return_to_state=0;
                ste.fn=new_fn;
                ste.fn->create_symbols();
                ste.fn_destroy=true;
                ste.instruct_id=0;
                ste.symbol_table_=new_symbol_table;
                ste.return_value=NULL;
                ste.symbol_table_->update_symbols_from_vars(new_vars);
                ste.local_id=new_local_id;

                ste.is_core_module_entrypoint=core_module_entrypoint;

                stack<c_eval_iteration> st;
                st.push(ste);

                {
                    eval_loop(st);
                }
            }
        }
    }
    vars=NULL;

    return true;
}

bool c_module::execute(string func, const vector<c_var_base*>& args_, const string& debug)
{
    var_keeper->set_current_session(session_);

    string name_only=func;
    if(name_only.find("(",0)!=string::npos)
        name_only.erase(name_only.find("(",0),name_only.length()-name_only.find("(",0));

    vector<string> child_vars;
    {
        string f=func;
        if(f.find("(",0)!=string::npos)
            f.erase(0,f.find("(",0)+1);
        if(f.length() && ')'==f[f.length()-1])
            f.erase(f.length()-1,1);
        explode(f,',',child_vars);
    }

    {
        // try procedure first
        vector<c_procedure>::iterator i1;
        bool got=false;
        for(i1=r_procedures.begin(); i1!=r_procedures.end(); ++i1) {
            (*i1).parse_call_string();
            if(!name_only.compare((*i1).name_only)) {
                got=true;
                break;
            }
        }
        if(got) {
            c_var_array* vars;
            c_temp_local tl(NULL);
            unsigned new_local_id=tl.get_params(vars);
            bool exc=false;
            vector<string>::iterator i2, i3;
            vector<c_var_base*>::const_iterator i_args;
            for(i3=(*i1).args.begin(), i2=child_vars.begin(), i_args=args_.begin(); i3!=(*i1).args.end() && i2!=child_vars.end() && i_args!=args_.end(); ++i2, ++i3, ++i_args) {
                c_index_base* idx=MY_NEW c_index_str(*i3);
                c_var_base* val=*i_args;
                if(!val)
                    val=MY_NEW c_var_unset();
                else
                    val=val->copy();
                if(!exc) try {
                    vars->set(this,idx,val,false,debug);
                } catch(c_var_exception e) {
                    e.print_oops_to_log(debug);
                    MY_DELETE(val);
                    exc=true;
                    break;
                }
            }
            if(exc) {
                return false;
            }
            for( ; i3!=(*i1).args.end(); ++i3) {
                c_index_base* idx=MY_NEW c_index_str(*i3);
                c_var_string* val=MY_NEW c_var_string("");
                try {
                    vars->set(this,idx,val,false,debug);
                } catch(c_var_exception e) {
                    e.print_oops_to_log(debug);
                    MY_DELETE(val);
                    exc=true;
                    break;
                }
            }
            if(exc) {
                return false;
            }
            // we need to make temporary copy, because stl do mess after rehash
            c_procedure p=*i1;
            // call it!
            exec(p,vars,new_local_id);
            return true;
        }
    }

    c_var_array* vars;
    c_temp_local tl(NULL);
    (void)tl.get_params(vars);
    {
        string func_name=func;
        size_t pos=func_name.find("(",0);
        if(pos!=string::npos)
            func_name.erase(pos,func_name.length()-pos);

        vector<c_func_arg> args;
        if(!r_functions.get_function_prototype(func_name,args)) {
            string d="Function/procedure '";
            d+=func_name;
            d+="' doesn't exist - ";
            d+=debug;
            if(logger)
                logger->log_print1(E_LOG_DEBUG,LOG_SCOPE_MODULE,"",d);
            return false;
        }

        if(args.size()!=child_vars.size() || args_.size()!=args.size()) {
            string d="Function '";
            d+=func_name;
            d+="' argument count mismatch - ";
            d+=debug;
            if(logger)
                logger->log_print1(E_LOG_DEBUG,LOG_SCOPE_MODULE,"",d);
            return false;
        }

        {
            c_module* this_module=this;
            string func_name_=name_only;
            c_micro_error* error=NULL;

            vector<c_func_arg> fn_args;
            if(this_module->r_functions.get_function_prototype(func_name_,fn_args)) {
                c_function* new_fn;
                bool dummy_b=false;
                if(!this_module->get_function(func_name_,dummy_b,new_fn) || dummy_b) {
                    error=MY_NEW c_micro_error_func_internal_1(func_name_);
                    if(get_logger())
                        get_logger()->log_print1(E_LOG_DEBUG,LOG_SCOPE_DEFAULT,"",error->get_msg());
                    MY_DELETE(error);
                    return false;
                }

                // child instance
                c_var_array* new_vars;
                c_temp_local tl(NULL);
                unsigned new_local_id=tl.get_params(new_vars);
                c_micro_symbol_table* new_symbol_table=new_fn->symbol_table;

                vector<c_func_arg> args1, args2;
                vector<c_var_base*>::const_iterator ai;
                vector<c_func_arg>::iterator afi;
                long arg_c=0;
                for(ai=args_.begin(), afi=fn_args.begin(); ai!=args_.end() && afi!=fn_args.end(); ++afi, ++ai, ++arg_c) {
                    c_func_arg a;
                    a.arg_expression=(*afi).arg_expression;
                    a.reference=(*afi).reference;
                    c_var_base* val=*ai;
                    if(!(*afi).reference) {
                        if(val)
                            val=val->copy();
                    } else {
                        c_var_base* new_val=NULL;
                        if(val)
                            new_val=val->copy();
                        if(val && new_val)
                            var_keeper->add_reference(val->get_id(),new_val->get_id(),&log_fn_error,this);
                        val=new_val;
                    }
                    if(!new_symbol_table->get_var_table()->set((*afi).value_sym,val,false,error,MICRO_NA,false,this,"",debug)) {
                        if(get_logger())
                            get_logger()->log_print1(E_LOG_DEBUG,LOG_SCOPE_DEFAULT,"",error->get_msg());
                        MY_DELETE(error);

                        MY_DELETE(val);

                        MY_DELETE(new_symbol_table);

                        MY_DELETE(new_fn);

                        return false;
                    }
                    c_index_base* idx=MY_NEW c_index_str((*afi).arg_expression);
                    if(!new_vars->set(this_module,idx,val,false,debug)) {
                        MY_DELETE(val);

                        MY_DELETE(new_symbol_table);

                        MY_DELETE(new_fn);

                        string msg="Error setting argument (zero-based) #";
                        msg+=tos(arg_c);
                        msg+=" to new function's internal variables";

                        error=MY_NEW c_micro_error_func_internal_3(func_name_,msg);
                        return false;
                    }
                    a.value=val;
                    a.value_sym=(*afi).value_sym;
                    args1.push_back(a);
                    args2.push_back(a);
                }
                c_eval_iteration ste(this_module,new_vars,debug);
                ste.state=1000;
                ste.return_to_state=0;
                ste.fn=new_fn;
                ste.fn->create_symbols();
                ste.fn_destroy=true;
                ste.instruct_id=0;
                ste.symbol_table_=new_symbol_table;
                ste.return_value=NULL;
                ste.symbol_table_->update_symbols_from_vars(new_vars);
                ste.local_id=new_local_id;

                stack<c_eval_iteration> st;
                st.push(ste);
                eval_loop(st);
            }
        }
    }
    return true;
}

bool c_module::execute_proc(const string& proc, vector<string>& string_vars, const string& debug)
{
    var_keeper->set_current_session(session_);

    c_temp_local local;
    c_var_array* vars=NULL;
    (void)local.get_params(vars);

    string name_only=proc;
    if(name_only.find("(",0)!=string::npos)
        name_only.erase(name_only.find("(",0),name_only.length()-name_only.find("(",0));

    {
        // try procedure first
        vector<c_procedure>::iterator i1;
        bool got=false;
        for(i1=r_procedures.begin(); i1!=r_procedures.end(); ++i1) {
            (*i1).parse_call_string();
            if(!name_only.compare((*i1).name_only)) {
                got=true;
                break;
            }
        }
        if(got) {
            c_var_array* vars;
            c_temp_local tl(NULL);
            unsigned new_local_id=tl.get_params(vars);
            vector<string>::iterator i2, i3;
            bool exc=false;
            for(i2=string_vars.begin(), i3=(*i1).args.begin(); i2!=string_vars.end() && i3!=(*i1).args.end(); ++i2, ++i3) {
                c_index_base* idx=MY_NEW c_index_str(*i3);
                c_var_string* val=MY_NEW c_var_string(*i2);
                try {
                    vars->set(this,idx,val,false,debug);
                } catch(c_var_exception e) {
                    e.print_oops_to_log(debug);
                    MY_DELETE(val);
                    exc=true;
                    break;
                }
            }
            if(exc) {
                var_keeper->set_current_session(SESSION_RESERVED_ID);
                return false;
            }
            for( ; i3!=(*i1).args.end(); ++i3) {
                c_index_base* idx=MY_NEW c_index_str(*i3);
                c_var_string* val=MY_NEW c_var_string("");
                try {
                    vars->set(this,idx,val,false,debug);
                } catch(c_var_exception e) {
                    e.print_oops_to_log(debug);
                    MY_DELETE(val);
                    exc=true;
                    break;
                }
            }
            if(exc) {
                var_keeper->set_current_session(SESSION_RESERVED_ID);
                return false;
            }
            // we need to make temporary copy, because stl do mess after rehash
            c_procedure p=*i1;
            // call it!
            exec(p,vars,new_local_id);
            var_keeper->set_current_session(SESSION_RESERVED_ID);
            return true;
        }
    }

    var_keeper->set_current_session(SESSION_RESERVED_ID);

    return false;
}

/*!
    \brief Loads logic.txt file and parses it
    \author VooDooMan
    \version 3
    \date 2004, 2005-10-23, 2006-12-27
    \param error_on_line If result == true, this contains number of line on which is error
    \param error_reason If result == true, this contains error string
    \param ignoring_procs Returns procedures/functions names that were ignored/locked
    \result Returns true if error occured
*/
// returns true if error
bool c_module::load_logic(int& error_on_line, string& error_reason, vector<string>& ignoring_procs)
{
    ignoring_procs.clear();

    bool is_core_mod=true;

    string filename;
    if(get_mod_name().compare("core")) {
        // != "core"
        is_core_mod=false;
        filename="modules" FILE_SLASH;
        filename+=get_mod_name();
        filename+=FILE_SLASH;
    }
    filename+="logic.txt";
    {
        FILE* f=fopen(filename.c_str(),"r");
        if(f==NULL) {
            error_on_line=1;
            string err="Cannot open file \"";
            err+=filename;
            err+="\"!";
            static char err_[1024];
            strcpy(err_,err.c_str());
            error_reason=err_;

            return true;
        }
        fclose(f);
    }

    error_on_line=1;
    error_reason="";

    bool b_user2=false; // user in bracket
    c_user user;
    bool b_channel=false; // user in bracket, channel in sub-bracket
    c_channel channel;
    bool b_user=false; // user token
    int i_brackets=0;
    bool b_proc=false; // in procedure
    c_procedure procedure;
    deque<c_lines> lines;
    bool b_chan=false;
    bool b_groups=false;
    bool b_prv=false;
    bool b_smtp=false;
    c_smtp smtp;
    bool b_smtp_data=false;

    c_channel_def chan;

    c_functions functions;
    c_function* func=NULL;

    vector<c_procedure> procedures;
    deque<c_user> users;
    vector<c_channel> channels;
    vector<c_channel_def> channel_defs;
    deque<c_user> terminators;

    all_groups.clear();

    c_private prv;

    long ver_num=1;

    string version="#version 2";
    bool very_first_line=true;
    bool in_function=false;

    string before;

    string func_dump;

    unsigned line_num=0;
    FILE* f=fopen(filename.c_str(),"r");
    static char ln[1024*10+1];
    bool echo_to_cout= stoi(conf->getvar("log_echo_debug"))!=0;
    while(f && !feof(f)) {
        if(fgets(ln,1024*10-1,f)==NULL)
            break;
        line_num++;

        ln[10240]=0;

        if(ln[0] && '\r'==ln[strlen(ln)-1])
            ln[strlen(ln)-1]='\0';
        if(ln[0] && '\n'==ln[strlen(ln)-1])
            ln[strlen(ln)-1]='\0';
        if(ln[0] && '\r'==ln[strlen(ln)-1])
            ln[strlen(ln)-1]='\0';
        if(ln[0] && '\n'==ln[strlen(ln)-1])
            ln[strlen(ln)-1]='\0';

        if(!in_function) {
            if('\0'==ln[0] || '\r'==ln[0] || '\n'==ln[0])
                continue;
        } else {
            if('\0'==ln[0] || '\r'==ln[0] || '\n'==ln[0]) {
                ln[0]='\n';
                ln[1]='\0';
            }
        }

        if(echo_to_cout)
            cout << ln << endl;

        if(3<=ver_num && !strcmp(ln,"#eof"))
            break;

        if(very_first_line) {
            if('#'==ln[0]) {
                version=ln;
                string v=version;
                if(v.find("#version ")==0) {
                    v.erase(0,strlen("#version "));
                    ver_num=stoi(v);
                    if(!ver_num)
                        ver_num=1;
                    if(3==ver_num)
                        version="#version 2";
                }
                very_first_line=false;
                error_on_line++;
                continue;
            } else
                very_first_line=false;
        }

        string s;
        if(!in_function) {
            bool space=true;
            for(int i1=0; ln[i1]!=0; ++i1) {
                if(ln[i1]==0x20 || ln[i1]=='\t')
                    if(space)
                        continue;
                        else
                        space=true;
                if(s.compare("") && space)
                    s.append("\x20");
                space=false;
                if(ln[i1]!=0x20 && ln[i1]!='\t')
                    s+=ln[i1];
            }
            for(size_t i2=s.length()-1; (signed)i2>=0; --i2)
                if(s[i2]==0x20 || s[i2]=='\t')
                    s=s.erase(i2,1);
                    else
                    break;
        } else
            s=ln;

        if(!in_function && s.find("function ",0)==0) {
            string n1=s;
            in_function=true;
            func=MY_NEW c_function();
            func->clear(false);
            string v=version;
            bool error=false;
            string error_msg;
            bool need_more=false;
            string debug;
            func->parse_command(this,v,error,error_msg,need_more,debug,"");
            before="";

            if(n1.find(" ",0)!=string::npos)
                n1.erase(0,n1.find(" ",0)+1);
            if(n1.find("(",0)!=string::npos)
                n1.erase(n1.find("(",0),n1.length()-n1.find("(",0));

            {
                vector<c_procedure>::iterator i1;
                for(i1=procedures.begin(); i1!=procedures.end(); ++i1) {
                    string n2=(*i1).name;
                    if(n2.find("(",0)!=string::npos)
                        n2.erase(n2.find("(",0),n2.length()-n2.find("(",0));
                    if(!n1.compare(n2)) {
                        error_reason="Procedure with same name already defined";
                        functions.destroy_all();
                        if(func)
                            func->destroy(true,true);
                        MY_DELETE(func);
                        func=NULL;
                        if(f)
                            fclose(f);
                        return true;
                    }
                }
            }
        }
        if(in_function) {
            func_dump+=ln;
            func_dump+="\n";

            bool error=false;
            string error_msg;
            bool need_more=false;

            string debug_prefix="logic.txt:";
            debug_prefix+=tos(line_num);
            debug_prefix+=" => ";
            string debug=debug_prefix+s;

            if(!before.empty())
                s=before+s;
            s+="\n";
            func->parse_command(this,s,error,error_msg,need_more,debug,debug_prefix);
            before=s;
            if(error) {
                functions.destroy_all();
                if(func)
                    func->destroy(true,true);
                MY_DELETE(func);
                func=NULL;
                error_reason=error_msg;
                if(f)
                    fclose(f);
                return true;
            }
            if(!need_more) {
                while(!func_dump.empty()
                    && (
                        '\r'==func_dump[func_dump.length()-1]
                        || '\n'==func_dump[func_dump.length()-1]
                        || '\t'==func_dump[func_dump.length()-1]
                        || '\x20'==func_dump[func_dump.length()-1]
                       )
                    ) {
                        func_dump.erase(func_dump.length()-1,1);
                }
                func_dump+="\n";
                func->set_dump(func_dump);
                func_dump="";
                if(!functions.add_function(func,error_reason)) {
                    functions.destroy_all();
                    if(func)
                        func->destroy(true,true);
                    MY_DELETE(func);
                    func=NULL;
                    if(f)
                        fclose(f);
                    return true;
                }
                func=NULL;
                in_function=false;
                before.clear();
            }
            error_on_line++;
            continue;
        }

        string debug="logic.txt:";
        debug+=tos(line_num);
        debug+=" => ";
        debug+=s;
        if(logic_process_line(this,s,error_reason,error_on_line,b_user2,user,b_channel,channel,b_user,i_brackets,b_proc,procedure,lines,b_smtp,smtp,b_smtp_data,procedures,users,channels,chan,channel_defs,b_chan,all_groups,b_groups,b_prv,prv,terminators,functions.functions,debug)) {
            if(f)
                fclose(f);
            functions.destroy_all();
            if(func)
                func->destroy(true,true);
            MY_DELETE(func);
            return true;
        }
        error_on_line++;
    }
    if(f)
        fclose(f);

    if(3<=ver_num && strcmp(ln,"#eof")) { // !=
        functions.destroy_all();
        if(func)
            func->destroy(true,true);
        MY_DELETE(func);

        error_on_line=line_num;
        error_reason="Missing mandatory '#eof' at the end of file for '#version 3' or later (corrupt file/incomplete?).";
        return true;
    }

    if(in_function) {
        functions.destroy_all();
        if(func)
            func->destroy(true,true);
        MY_DELETE(func);
        func=NULL;
        if(error_reason.empty())
            error_reason="Unteminated function past end of the file";
        return true;
    }

    {
        vector<string> errors;
        string msg;

#define MANDATORY_ERROR_1(USER,CHANNEL,EVENT,ARGS) \
    { \
        msg="OOPS: Warning: Security restriction in effect: Mandatory arguments for user '"; \
        msg+=USER; \
        msg+="' channel '"; \
        msg+=CHANNEL; \
        msg+="' event '"; \
        msg+=EVENT; \
        msg+="' mismatched. Should have "; \
        msg+=tos((unsigned)ARGS); \
        msg+=" arguments. Procedure/function will NOT be called."; \
        errors.push_back(msg); \
    }

        // check mandatory arguments
        {
            vector<c_channel>::iterator i1;
            for(i1=channels.begin(); i1!=channels.end(); ++i1) {

                if(!(*i1).on_ban.empty())
                    if(!logic_check_num_agrs((*i1).on_ban,6))
                        MANDATORY_ERROR_1((*i1).username,(*i1).channel_name,"on_ban",6);

                if(!(*i1).on_banned.empty())
                    if(!logic_check_num_agrs((*i1).on_banned,3))
                        MANDATORY_ERROR_1((*i1).username,(*i1).channel_name,"on_banned",3);

                if(!(*i1).on_deop.empty())
                    if(!logic_check_num_agrs((*i1).on_deop,5))
                        MANDATORY_ERROR_1((*i1).username,(*i1).channel_name,"on_deop",5);

                if(!(*i1).on_devoice.empty())
                    if(!logic_check_num_agrs((*i1).on_devoice,5))
                        MANDATORY_ERROR_1((*i1).username,(*i1).channel_name,"on_devoice",5);

                if(!(*i1).on_except.empty())
                    if(!logic_check_num_agrs((*i1).on_except,7))
                        MANDATORY_ERROR_1((*i1).username,(*i1).channel_name,"on_except",7);

                if(!(*i1).on_flood.empty())
                    if(!logic_check_num_agrs((*i1).on_flood,8))
                        MANDATORY_ERROR_1((*i1).username,(*i1).channel_name,"on_flood",8);

                if(!(*i1).on_invite.empty())
                    if(!logic_check_num_agrs((*i1).on_invite,7))
                        MANDATORY_ERROR_1((*i1).username,(*i1).channel_name,"on_invite",7);

                if(!(*i1).on_join.empty())
                    if(!logic_check_num_agrs((*i1).on_join,6))
                        MANDATORY_ERROR_1((*i1).username,(*i1).channel_name,"on_join",6);

                if(!(*i1).on_kick.empty())
                    if(!logic_check_num_agrs((*i1).on_kick,5))
                        MANDATORY_ERROR_1((*i1).username,(*i1).channel_name,"on_kick",5);

                if(!(*i1).on_not_in_reop.empty())
                    if(!logic_check_num_agrs((*i1).on_not_in_reop,3))
                        MANDATORY_ERROR_1((*i1).username,(*i1).channel_name,"on_not_in_reop",3);

                if(!(*i1).on_not_invited.empty())
                    if(!logic_check_num_agrs((*i1).on_not_invited,3))
                        MANDATORY_ERROR_1((*i1).username,(*i1).channel_name,"on_not_invited",3);

                if(!(*i1).on_notice.empty())
                    if(!logic_check_num_agrs((*i1).on_notice,4))
                        MANDATORY_ERROR_1((*i1).username,(*i1).channel_name,"on_notice",4);

                if(!(*i1).on_op.empty())
                    if(!logic_check_num_agrs((*i1).on_op,5))
                        MANDATORY_ERROR_1((*i1).username,(*i1).channel_name,"on_op",5);

                if(!(*i1).on_other_mode.empty())
                    if(!logic_check_num_agrs((*i1).on_other_mode,7))
                        MANDATORY_ERROR_1((*i1).username,(*i1).channel_name,"on_other_mode",7);

                if(!(*i1).on_other_mode_mask.empty())
                    if(!logic_check_num_agrs((*i1).on_other_mode_mask,8))
                        MANDATORY_ERROR_1((*i1).username,(*i1).channel_name,"on_other_mode_mask",8);

                if(!(*i1).on_privmsg.empty())
                    if(!logic_check_num_agrs((*i1).on_privmsg,4))
                        MANDATORY_ERROR_1((*i1).username,(*i1).channel_name,"on_privmsg",4);

                if(!(*i1).on_reop.empty())
                    if(!logic_check_num_agrs((*i1).on_reop,7))
                        MANDATORY_ERROR_1((*i1).username,(*i1).channel_name,"on_reop",7);

                if(!(*i1).on_unban.empty())
                    if(!logic_check_num_agrs((*i1).on_unban,6))
                        MANDATORY_ERROR_1((*i1).username,(*i1).channel_name,"on_unban",6);

                if(!(*i1).on_unexcept.empty())
                    if(!logic_check_num_agrs((*i1).on_unexcept,7))
                        MANDATORY_ERROR_1((*i1).username,(*i1).channel_name,"on_unexcept",7);

                if(!(*i1).on_uninvite.empty())
                    if(!logic_check_num_agrs((*i1).on_uninvite,7))
                        MANDATORY_ERROR_1((*i1).username,(*i1).channel_name,"on_uninvite",7);

                if(!(*i1).on_voice.empty())
                    if(!logic_check_num_agrs((*i1).on_voice,5))
                        MANDATORY_ERROR_1((*i1).username,(*i1).channel_name,"on_voice",5);

            }
        }
        {

#define MANDATORY_ERROR_2(CHANNEL,EVENT,ARGS) \
    { \
        msg="OOPS: Warning: Security restriction in effect: Mandatory arguments for channel definition '"; \
        msg+=CHANNEL; \
        msg+="' event '"; \
        msg+=EVENT; \
        msg+="' mismatched. Should have "; \
        msg+=tos((unsigned)ARGS); \
        msg+=" arguments. Procedure/function will NOT be called."; \
        errors.push_back(msg); \
    }

            vector<c_channel_def>::iterator i1;
            for(i1=channel_defs.begin(); i1!=channel_defs.end(); ++i1) {

                if(!(*i1).on_ctcp.empty())
                    if(!logic_check_num_agrs((*i1).on_ctcp,5))
                        MANDATORY_ERROR_2((*i1).channel_name,"on_ctcp",5);

                if(!(*i1).on_dynamic_ban.empty())
                    if(!logic_check_num_agrs((*i1).on_dynamic_ban,5))
                        MANDATORY_ERROR_2((*i1).channel_name,"on_dynamic_ban",5);

                if(!(*i1).on_ircop.empty())
                    if(!logic_check_num_agrs((*i1).on_ircop,4))
                        MANDATORY_ERROR_2((*i1).channel_name,"on_ircop",4);

                if(!(*i1).on_key.empty())
                    if(!logic_check_num_agrs((*i1).on_key,5))
                        MANDATORY_ERROR_2((*i1).channel_name,"on_key",5);

                if(!(*i1).on_limit.empty())
                    if(!logic_check_num_agrs((*i1).on_limit,5))
                        MANDATORY_ERROR_2((*i1).channel_name,"on_limit",5);

                if(!(*i1).on_mode.empty())
                    if(!logic_check_num_agrs((*i1).on_mode,5))
                        MANDATORY_ERROR_2((*i1).channel_name,"on_mode",5);

                if(!(*i1).on_notice.empty())
                    if(!logic_check_num_agrs((*i1).on_notice,4))
                        MANDATORY_ERROR_2((*i1).channel_name,"on_notice",4);

                if(!(*i1).on_part.empty())
                    if(!logic_check_num_agrs((*i1).on_part,5))
                        MANDATORY_ERROR_2((*i1).channel_name,"on_part",5);

                if(!(*i1).on_privmsg.empty())
                    if(!logic_check_num_agrs((*i1).on_privmsg,4))
                        MANDATORY_ERROR_2((*i1).channel_name,"on_privmsg",4);

                if(!(*i1).on_server_msg.empty())
                    if(!logic_check_num_agrs((*i1).on_server_msg,6))
                        MANDATORY_ERROR_2((*i1).channel_name,"on_server_msg",6);

                if(!(*i1).on_topic.empty())
                    if(!logic_check_num_agrs((*i1).on_topic,4))
                        MANDATORY_ERROR_2((*i1).channel_name,"on_topic",4);
            }
        }
        {
#define MANDATORY_ERROR_3(EVENT,ARGS) \
    { \
        msg="OOPS: Warning: Security restriction in effect: Mandatory arguments for private definition, event '"; \
        msg+=EVENT; \
        msg+="' mismatched. Should have "; \
        msg+=tos((unsigned)ARGS); \
        msg+=" arguments. Procedure/function will NOT be called."; \
        errors.push_back(msg); \
    }

            if(!prv.on_broadcast.empty())
                if(!logic_check_num_agrs(prv.on_broadcast,9))
                    MANDATORY_ERROR_3("on_broadcast",9);

            if(!prv.on_ctcp.empty())
                if(!logic_check_num_agrs(prv.on_ctcp,5))
                    MANDATORY_ERROR_3("on_ctcp",5);

            if(!prv.on_filesys_got_new.empty())
                if(!logic_check_num_agrs(prv.on_filesys_got_new,5))
                    MANDATORY_ERROR_3("on_filesys_got_new",5);

            if(!prv.on_fnc.empty())
                if(!logic_check_num_agrs(prv.on_fnc,2))
                    MANDATORY_ERROR_3("on_fnc",2);

            if(!prv.on_internal_event.empty())
                if(!logic_check_num_agrs(prv.on_internal_event,13))
                    MANDATORY_ERROR_3("on_internal_event",13);

            if(!prv.on_notice.empty())
                if(!logic_check_num_agrs(prv.on_notice,3))
                    MANDATORY_ERROR_3("on_notice",3);

            if(!prv.on_privmsg.empty())
                if(!logic_check_num_agrs(prv.on_privmsg,3))
                    MANDATORY_ERROR_3("on_privmsg",3);

            if(!prv.on_server_msg.empty())
                if(!logic_check_num_agrs(prv.on_server_msg,6))
                    MANDATORY_ERROR_3("on_server_msg",6);
        }

        vector<string>::iterator i1;
        for(i1=errors.begin(); i1!=errors.end(); ++i1) {
            log_bot(*i1,true);
            log_debug(*i1,true);
        }
    }

    // all okay, copy all objects to real variables
    {
        {
            vector<c_procedure>::iterator i;
            for(i=r_procedures.begin(); i!=r_procedures.end(); ++i)
                if((*i).lock_cnt) {
                    string name=(*i).name;
                    if(name.find('(')!=string::npos)
                        name.erase(name.find('('),name.length()-name.find('('));
                    ignoring_procs.push_back(name);
                }
        again_1:
            for(i=r_procedures.begin(); i!=r_procedures.end(); ++i) {
                if(!(*i).lock_cnt) {
                    r_procedures.erase(i);
                    goto again_1;
                }
            }
            // we resolved this by temporary copy
            /*
            for(i=procedures.begin(); i!=procedures.end(); ++i) {
                bool ignore=false;
                vector<c_procedure>::iterator ii;
                for(ii=r_procedures.begin(); ii!=r_procedures.end(); ++ii) {
                    string name=(*i).name;
                    if(name.find('(')!=string::npos)
                        name.erase(name.find('('),name.length()-name.find('('));

                    string name2=(*ii).name;
                    if(name2.find('(')!=string::npos)
                        name2.erase(name2.find('('),name2.length()-name2.find('('));

                    if(!name.compare(name2)) {
                        ignore=true;
                        break;
                    }
                }
                if(ignore)
                    continue;
                r_procedures.push_back(*i);
            }
            */
            r_procedures.insert(r_procedures.end(),procedures.begin(),procedures.end());
            procedures.clear();
        }

        {
            map<string,c_function*>::iterator i;
            /*
            for(i=r_functions.functions.begin(); i!=r_functions.functions.end(); ++i)
                if((*i).second.lock_cnt)
                    ignoring_procs.push_back((*i).second.get_name());
            */
            while((i=r_functions.functions.begin())!=r_functions.functions.end()) {
                /*if(!(*i).second.lock_cnt)*/ {
                    (*i).second->destroy(true,true);
                    MY_DELETE((*i).second);
                    r_functions.functions.erase(i);
                }
            }
            for(i=functions.functions.begin(); i!=functions.functions.end(); ++i) {
                // we resolved this by temporary copy
                /*
                bool ignore=false;
                map<string,c_function*>::iterator ii;
                for(ii=r_functions.functions.begin(); ii!=r_functions.functions.end(); ++ii) {
                    if(!(*ii).second->get_name().compare((*i).second->get_name())) {
                        ignore=true;
                        break;
                    }
                }
                if(ignore)
                    continue;
                */
                string error;
                r_functions.add_function((*i).second,error);
            }
            functions.clear();
        }

        if(is_core_mod) {
            {
                r_users.clear();
                r_user_templates.clear();
                deque<c_user>::iterator i;
                for(i=users.begin(); i!=users.end(); ++i) {
                    if(!(*i).is_template)
                        r_users.push_back(*i);
                    else
                        r_user_templates.push_back(*i);
                }
                users.clear();
            }

            {
                r_terminators.clear();
                deque<c_user>::iterator i;
                for(i=terminators.begin(); i!=terminators.end(); ++i) {
                    r_terminators.push_back(*i);
                }
                terminators.clear();
            }

            {
                r_channels.clear();
                r_channel_templates.clear();
                r_channel_terminators.clear();
                vector<c_channel>::iterator i;
                for(i=channels.begin(); i!=channels.end(); ++i) {
                    if(!(*i).is_template) {
                        if(!(*i).terminator)
                            r_channels.push_back(*i);
                        else
                            r_channel_terminators.push_back(*i);
                    } else
                        r_channel_templates.push_back(*i);
                }
                channels.clear();
            }

            {
                r_channel_defs.clear();
                vector<c_channel_def>::iterator i;
                for(i=channel_defs.begin(); i!=channel_defs.end(); ++i)
                    r_channel_defs.push_back(*i);
                channel_defs.clear();
            }

            {
                r_private.clear();
                r_private=prv;
            }
        }
    }

    if(!version.compare("#version 1")) {
        logic_partyline_backup(this,"self-upgrade");
        error_on_line=-1;
        error_reason="File logic.txt was rebuilt.";
        return true;
    }

    error_on_line=0;
    return false;
}

/*!
    \brief Processes output of script, and closes handle (and OS resources of handle)
    \author VooDooMan
    \version 1
    \date 2004
    \param h Handle to process
    \param script_type Type of the script (currently only "php" is supported)
    \param script_num Number of script (for logging purposes)
    \param filename Name of file for script
    \param noremove Do not remove temporary file
*/
void c_module::process_script_output(string script_type, int script_num, c_exec_handle* h, const string& filename, bool noremove)
{
    string orig_script_type=script_type;

    if(!script_type.compare("php_2"))
        script_type="php";

    char* line=MY_NEW char[1024*10+1];

    string fn="tmp_";
    fn+=toshex(rand());
    fn+=toshex(rand());
    fn+=toshex(rand());
    fn+=toshex(rand());
    if(!filename.empty())
        fn=filename;
    fn+=".tmp";

    string script_dir=dir_prefix;
    if(script_dir.length() && script_dir[script_dir.length()-1]!=FILE_SLASH_CH)
        script_dir+=FILE_SLASH;
    script_dir+=conf_getvar(script_type+(string)"_script_dir");

    fn=script_dir+fn;

    FILE* f=fopen(fn.c_str(),"wb");
    if(f) {
        int size;
        char* buf=NULL;
        exec_read_output(h,buf,size);
        fwrite(buf,size,1,f);
        my_free(buf);
        fclose(f);
    }

    exec_process_end(h);
    MY_DELETE(h);

    string text;

    f=fopen(fn.c_str(),"r");
    while(f && !feof(f)) {
        if(fgets(line,1024*10-1,f)==NULL)
            break;

        text+=line;
    }

    if(f)
        fclose(f);
    if(!noremove)
        remove(fn.c_str());

    MY_DELETE_A(line);

    process_script_output(script_type,script_num,text);
}

/*!
    \brief Processes output of script, and closes handle (and OS resources of handle)
    \author VooDooMan
    \version 2
    \date 2004, 2006-08-28
    \param script_num Number of script
    \param script_type Type of the script (currently only "php" is supported)
    \param text Script output
*/
void c_module::process_script_output(string script_type, int script_num, const string& text)
{
    string orig_script_type=script_type;

    if(!script_type.compare("php_2"))
        script_type="php";

    char* line=MY_NEW char[1024*10+1];

    bool commands=false;

    c_file fl;
    fl.clear();
    c_access fl_a;
    fl_a.clear();
    bool in_filesys=false;
    bool in_filesys_access=false;

    size_t ii=0;
    while(ii<text.length()) {
        size_t ii2=0;
        for(; ii<text.length(); ++ii, ++ii2) {
            if(ii2>=1024*10) {
                break;
            }
            if('\n'==text[ii] || '\r'==text[ii]) {
                while('\n'==text[ii] || '\r'==text[ii]) {
                    ii++;
                    if(ii>=text.length())
                        break;
                }
                break;
            }
            line[ii2]=text[ii];
        }
        line[ii2]='\0';
        if(!commands && !strcmp(line,"#####")) {
            commands=true;
            continue;
        }
        if(commands && !strcmp(line,"#####"))
            break;
        if(commands) {
            string orig=line;
            bool got=false;
            string cmd=line;
            if(!in_filesys) {
                if(cmd.length()>=8 && cmd[0]=='F' && cmd[1]=='I' && cmd[2]=='L' && cmd[3]=='E' && cmd[4]=='S' && cmd[5]=='Y' && cmd[6]=='S' && cmd[7]==0x20) {
                    got=true;
                    string internal_name=cmd;
                    if(internal_name.find(' ',0)!=string::npos)
                        internal_name.erase(0,internal_name.find(' ',0)+1);
                    string c=internal_name;
                    if(internal_name.find(' ',0)!=string::npos)
                        internal_name.erase(0,internal_name.find(' ',0)+1);
                    if(c.find(' ',0)!=string::npos)
                        c.erase(c.find(' ',0),c.length());
                    if(!c.compare("InternalName"))
                        if(filesys_logic_get_file(internal_name,fl))
                            in_filesys=true;
                }
                if(cmd.length()>=8 && cmd[0]=='P' && cmd[1]=='U' && cmd[2]=='T' && cmd[3]==0x20 && cmd[4]=='L' && cmd[5]=='O' && cmd[6]=='W' && cmd[7]==0x20) {
                    got=true;
                    cmd.erase(0,8);
                    irc_put(cmd,LOW_PRIORITY);
                    continue;
                }
                if(cmd.length()>=9 && cmd[0]=='P' && cmd[1]=='U' && cmd[2]=='T' && cmd[3]==0x20 && cmd[4]=='H' && cmd[5]=='I' && cmd[6]=='G' && cmd[7]=='H' && cmd[8]==0x20) {
                    got=true;
                    cmd.erase(0,9);
                    irc_put(cmd,HIGH_PRIORITY);
                    continue;
                }
                if(cmd.length()>=13 && cmd[0]=='P' && cmd[1]=='U' && cmd[2]=='T' && cmd[3]==0x20 && cmd[4]=='C' && cmd[5]=='R' && cmd[6]=='I' && cmd[7]=='T' && cmd[8]=='I' && cmd[9]=='C' && cmd[10]=='A' && cmd[11]=='L' && cmd[12]==0x20) {
                    got=true;
                    cmd.erase(0,13);
                    irc_put(cmd,CRITICAL_PRIORITY);
                    continue;
                }
                if(cmd.length()>=4 && cmd[0]=='L' && cmd[1]=='O' && cmd[2]=='G' && cmd[3]==0x20) {
                    got=true;
                    cmd.erase(0,4);
                    string logtype;
                    while(cmd.length()>0 && cmd[0]!=0x20) {
                        logtype+=cmd[0];
                        cmd.erase(0,1);
                    }
                    if(cmd.length()>0 && cmd[0]==0x20)
                        cmd.erase(0,1);
                    if(!logtype.compare("DEBUG")) {
                        cmd="SCRIPT: "+cmd;
                        log_debug(cmd);
                        continue;
                    }
                    log_channel(logtype,TYPE_SCRIPT,"","",cmd,"");
                    continue;
                }
                if(cmd.length()>=8 && cmd[0]=='E' && cmd[1]=='X' && cmd[2]=='E' && cmd[3]=='C' && cmd[4]=='U' && cmd[5]=='T' && cmd[6]=='E' && cmd[7]==0x20) {
                    string code_str=cmd;
                    code_str.erase(0,strlen("EXECUTE "));
                    var_keeper->set_current_session(session_);
                    //c_var_array* vars=MY_NEW c_var_array();
                    c_temp_local local;
                    c_var_array* vars=NULL;
                    unsigned new_local_id=local.get_params(vars);
                    string debug="<script_output>:1 -> ";
                    debug+=code_str;
                    stack<c_eval_iteration> st;

                    c_eval_iteration ste(this,vars,debug);
                    ste.proc_eval=true;
                    ste.state=0;
                    ste.return_to_state=0;
                    ste.local_id=new_local_id;
                    st.push(ste);

                    eval_function(this,code_str,vars,this->r_functions,debug,st);
                    eval_loop(st);
                    var_keeper->set_current_session(SESSION_RESERVED_ID);
                    continue;
                }
            } else {
                string cmd=line;
                string c1, c2, c3, c4;
                int ctx=0;
                for(unsigned int i1=0; i1<cmd.length(); ++i1) {
                    if(ctx<3 && cmd[i1]==0x20) {
                        ctx++;
                        continue;
                    }
                    if(ctx==0)
                        c1.push_back(cmd[i1]);
                    if(ctx==1)
                        c2.push_back(cmd[i1]);
                    if(ctx==2)
                        c3.push_back(cmd[i1]);
                    if(ctx==3)
                        c4.push_back(cmd[i1]);
                }
                if(!c1.compare("FILESYS")) {
                    if(!in_filesys_access && !c2.compare("SET")) {
                        if(!c3.compare("type")) {
                            if(!c4.compare("ft_message")) {
                                fl.file_type=ft_message;
                                got=true;
                            }
                            if(!c4.compare("ft_file")) {
                                fl.file_type=ft_file;
                                got=true;
                            }
                        }
                        if(!c3.compare("PublicName")) {
                            got=true;
                            fl.public_name=c4;
                        }
                        if(!c3.compare("Time")) {
                            got=true;
                            my_time_t t=sto_time(c4);
                            fl.ftime=t;
                        }
                        if(!c3.compare("Published")) {
                            got=true;
                            int b=stoi(c4);
                            fl.published=TO_BOOL(b);
                        }
                        if(!c3.compare("Complete")) {
                            got=true;
                            int b=stoi(c4);
                            fl.complete=TO_BOOL(b);
                        }
                        if(!c3.compare("Expiration")) {
                            got=true;
                            my_time_t t=sto_time(c4);
                            fl.expiration=t;
                        }
                    }
                    if(!in_filesys_access && !c2.compare("ACCESS_BEGIN")) {
                        got=true;
                        in_filesys_access=true;
                        fl.access.clear();
                        fl_a.clear();
                        continue;
                    }
                    if(in_filesys_access && !c2.compare("SET")) {
                        if(!c3.compare("AllUsers")) {
                            got=true;
                            int b=stoi(c4);
                            fl_a.all_users=TO_BOOL(b);
                        }
                        if(!c3.compare("UserName")) {
                            got=true;
                            fl_a.user_name=c4;
                        }
                        if(!c3.compare("Owner")) {
                            got=true;
                            int b=stoi(c4);
                            fl_a.owner=TO_BOOL(b);
                        }
                        if(!c3.compare("Read")) {
                            got=true;
                            int b=stoi(c4);
                            fl_a.read=TO_BOOL(b);
                        }
                        if(!c3.compare("Delete")) {
                            got=true;
                            int b=stoi(c4);
                            fl_a.del=TO_BOOL(b);
                        }
                        if(!c3.compare("NotifyOwner")) {
                            got=true;
                            int b=stoi(c4);
                            fl_a.notify_owner=TO_BOOL(b);
                        }
                        if(!c3.compare("NotifyUser")) {
                            got=true;
                            int b=stoi(c4);
                            fl_a.notify_user=TO_BOOL(b);
                        }
                        if(!c3.compare("Secure")) {
                            got=true;
                            int b=stoi(c4);
                            fl_a.secure=TO_BOOL(b);
                        }
                        if(!c3.compare("AllOnChannel")) {
                            got=true;
                            fl_a.all_on_channel=c4;
                        }
                        if(!c3.compare("AlsoUnknown")) {
                            got=true;
                            int b=stoi(c4);
                            fl_a.also_unknown=TO_BOOL(b);
                        }
                        if(!c3.compare("NotifyOwnerMessage") || !c3.compare("NotifyUserMessage")) {
                            string msg;
                            bool got_backslash=false;
                            for(unsigned int i1=0; i1<c4.length(); ++i1) {
                                if(!got_backslash && c4[i1]=='\\') {
                                    got_backslash=true;
                                    continue;
                                }
                                if(got_backslash) {
                                    if(i1+1>=c4.length()) {
                                        string d=GET_STR_DEBUG_INFO_EX("OOPS: Occurred error: erroneous escape sequence '\\' at the end of the string \"");
                                        d+=c4+"\"";
                                        log_debug(d);
                                    }
                                    char x=c4[i1];
                                    switch(x) {
                                        case '\\':
                                            msg+="\\";
                                            got_backslash=false;
                                            break;
                                        case '"':
                                            msg+="\"";
                                            got_backslash=false;
                                            break;
                                        case 'r':
                                            msg+="\r";
                                            got_backslash=false;
                                            break;
                                        case 'n':
                                            msg+="\n";
                                            got_backslash=false;
                                            break;
                                        case 't':
                                            msg+="\t";
                                            got_backslash=false;
                                            break;
                                        case 'a':
                                            msg+="\x07";
                                            got_backslash=false;
                                            break;
                                        case 'b':
                                            msg+="\b";
                                            got_backslash=false;
                                            break;
                                        case 'f':
                                            msg+="\f";
                                            got_backslash=false;
                                            break;
                                        case 'v':
                                            msg+="\v";
                                            got_backslash=false;
                                            break;
                                        case '\'':
                                            msg+="\'";
                                            got_backslash=false;
                                            break;
                                        case '?':
                                            msg+="\?";
                                            got_backslash=false;
                                            break;
                                        case 'x':
                                            {
                                                got_backslash=false;
                                                if(i1+1+2>=c4.length()) {
                                                    string d=GET_STR_DEBUG_INFO_EX("OOPS: Occurred error: erroneous escape character '\\x' at the end of the string \"");
                                                    d+=c4+"\"";
                                                    log_debug(d);
                                                } else {
                                                    unsigned char x1[8];
                                                    x1[0]=static_cast<unsigned char>(c4[++i1]);
                                                    x1[1]=c4[++i1];
                                                    unsigned char x2=0;
                                                    for(int i1=0; i1<2; ++i1) {
                                                        unsigned char nibble=0xff;
                                                        if(x1[i1]>='0' && x1[i1]<='9')
                                                            nibble=x1[i1]-'0';
                                                        if(x1[i1]>='a' && x1[i1]<='f')
                                                            nibble=x1[i1]+10-'a';
                                                        if(x1[i1]>='A' && x1[i1]<='F')
                                                            nibble=x1[i1]+10-'A';
                                                        if(nibble==0xff) {
                                                            nibble=0;
                                                            string d="OOPS: Occurred error: erroneous escape character '\\x' - character not in HEX family \"";
                                                            d+=c4+"\" ";
                                                            log_debug(GET_STR_DEBUG_INFO_EX(d));
                                                        }
                                                        x2|=nibble << (i1==0?4U:0U);
                                                    }
                                                    msg.push_back(x2);
                                                }
                                            }
                                            break;
                                        case '0':
                                        case '1':
                                        case '2':
                                        case '3':
                                        case '4':
                                        case '5':
                                        case '6':
                                        case '7':
                                            {
                                                got_backslash=false;
                                                if(i1+1+3>=c4.length()) {
                                                    string d="OOPS: Occurred error: erroneous escape character '\\<octal_num>' at the end of the string \"";
                                                    d+=c4+"\" ";
                                                    log_debug(GET_STR_DEBUG_INFO_EX(d));
                                                } else {
                                                    unsigned char x1[8];
                                                    x1[0]=x;
                                                    x1[1]=static_cast<unsigned char>(c4[++i1]);
                                                    x1[2]=static_cast<unsigned char>(c4[++i1]);
                                                    unsigned char x2=0;
                                                    for(int i1=0; i1<3; ++i1) {
                                                        unsigned char nibble=0xff;
                                                        if(x1[i1]>='0' && x1[i1]<='7')
                                                            nibble=x1[i1]-'0';
                                                        if(nibble==0xff) {
                                                            nibble=0;
                                                            string d="OOPS: Occurred error: erroneous escape character '\\<octal_num>' - character not in OCTAL family \"";
                                                            d+=c4+"\" ";
                                                            log_debug(GET_STR_DEBUG_INFO_EX(d));
                                                        }
                                                        if(i1==0)
                                                            nibble<<=6;
                                                        if(i1==1)
                                                            nibble<<=3;
                                                        /*if(i1==2)
                                                            nibble<<=0;*/
                                                        x2|=nibble;
                                                    }
                                                    msg.push_back(x2);
                                                }
                                            }
                                            break;
                                        default:
                                            {
                                                string d="OOPS: Occurred error: erroneous escape character '\\' - unknown next character \"";
                                                d+=c4+"\" ";
                                                log_debug(GET_STR_DEBUG_INFO_EX(d));
                                            }
                                            break;
                                    }
                                } else {
                                    msg.push_back(c4[i1]);
                                }
                            }
                            if(!c3.compare("NotifyOwnerMessage")) {
                                got=true;
                                fl_a.notify_owner_message=msg;
                            }
                            if(!c3.compare("NotifyUserMessage")) {
                                got=true;
                                fl_a.notify_user_message=msg;
                            }
                        }
                    }
                    if(in_filesys_access && !c2.compare("ACCESS_END")) {
                        got=true;
                        fl.access.push_back(fl_a);
                        fl_a.clear();
                        in_filesys_access=false;
                        continue;
                    }
                    if(!c2.compare("COMMIT")) {
                        got=true;
                        filesys_logic_set_file(fl.internal_name,fl);
                        fl.clear();
                        fl_a.clear();
                        in_filesys=in_filesys_access=false;
                        continue;
                    }
                }
            }
            if(!got) {
                string msg="Invalid result of script: type=\"";
                msg+=orig_script_type;
                msg+="\", number=";
                msg+=tos(script_num);
                msg+=": ";
                msg+=orig;
                log_debug(msg);
            }
        }
    }


    MY_DELETE_A(line);
}


/*!
    \brief Converts variable to script (printable) form
    \author VooDooMan
    \version 1
    \param script_type Type of script (e.g. "php", "php_2", "logic.txt")
    \param param Variable to convert
    \return Returns printable form
*/
string c_module::script_make_param(const string& script_type, c_var_base* param)
{
    if(NULL==param) {
        c_var_unset null;
        return null.get_script_string(this,script_type);
    }

    return param->get_script_string(this,script_type);
}

/*!
    \brief Executes a script
    \author VooDooMan
    \version 1
    \date 2004
    \param script_type Type of script ("php" and "php_2" supported yet)
    \param script_num Number of script (script_type+"_#=..." in conf.txt)
    \param channel Channel name to pass to the script
    \param params Parameters to pass to the script as array
    \param flags Flags
*/
void c_module::exec_script(string script_type, int script_num, string channel, vector<c_var_base*>& params, const vector<string>& flags)
{
    if(script_type.compare("php") && script_type.compare("php_2") && script_type.compare("perl")) {
        string d="OOPS: Occurred error: unsupported script type (";
        d+=script_type;
        d+=") ";
        if(logger)
            logger->log_print1(E_LOG_DEBUG,LOG_SCOPE_MODULE,"",GET_STR_DEBUG_INFO_EX(d));
        return;
    }

    bool noremove=false;
    {
        vector<string>::const_iterator i1;
        for(i1=flags.begin(); i1!=flags.end(); ++i1) {
            if((*i1).empty()) {
                continue;
            }
            if(!(*i1).compare("noremove")) {
                noremove=true;
                continue;
            }
            // else:
            {
                string d="OOPS: Occurred error: ignoring unsupported script flag (\"";
                d+=*i1;
                d+="\") ";
                if(logger)
                    logger->log_print1(E_LOG_DEBUG,LOG_SCOPE_MODULE,"",GET_STR_DEBUG_INFO_EX(d));
            }
        }
    }

    if(!script_type.compare("php")) {
        string tmp=script_type+(string)"_";
        tmp+=tos(script_num);

        string script_dir=utils_convert_unix_path(dir_prefix);
        if(script_dir.length() && script_dir[script_dir.length()-1]!=FILE_SLASH_CH)
            script_dir+=FILE_SLASH;
        script_dir+=conf_getvar(script_type+(string)"_script_dir");

        string script=utils_convert_unix_path(script_dir+conf_getvar(tmp));

        string fn="tmp_";
        fn+=toshex(rand());
        fn+=toshex(rand());
        fn+=toshex(rand());
        fn+=toshex(rand());
        string fn2=fn;
        fn+=(string)"."+script_type;
        fn2+=".tmp";

        string tmp_fn1=script_dir;
        string tmp_fn2=tmp_fn1;
        tmp_fn1+=fn;
        tmp_fn2+=fn2;

        FILE* out=fopen(tmp_fn1.c_str(),"w");

        char* line=MY_NEW char[1024*10+1];

        if(script.empty())
            return;

        FILE* in=fopen(script.c_str(),"r");
        line[0]=0;
        while(in && !feof(in)) {
            if(fgets(line,1024*10-1,in)==NULL)
                break;
            line[1024*10-1]=0;
            if(line[0] && line[strlen(line)-1]=='\r')
                line[strlen(line)-1]=0;
            if(line[0] && line[strlen(line)-1]=='\n')
                line[strlen(line)-1]=0;
            if(line[0] && line[strlen(line)-1]=='\r')
                line[strlen(line)-1]=0;
            if(line[0] && line[strlen(line)-1]=='\n')
                line[strlen(line)-1]=0;
            if(!strcmp(line,"// INIT"))
                break;
            if(strlen(line))
                if(out)
                    fprintf(out,"%s",line);
            string tmp="\n";
            if(out)
                fprintf(out,"%s",tmp.c_str());
            line[0]=0;
        }

        strcpy(line,"    $uptime=");
        if(out)
            fprintf(out,"%s",line);
        tmp=tos(static_cast<unsigned long>(stats_uptime()));
        tmp+=";\n";
        if(out)
            fprintf(out,"%s",tmp.c_str());

        strcpy(line,"    $using_irc_server_host=");
        if(out)
            fprintf(out,"%s",line);
        tmp=c_var_string::script_esc(script_type,irc_server_host);
        tmp+=";\n";
        if(out)
            fprintf(out,"%s",tmp.c_str());

        strcpy(line,"    $using_irc_server_port=");
        if(out)
            fprintf(out,"%s",line);
        tmp=tos(irc_server_port);
        tmp+=";\n";
        if(out)
            fprintf(out,"%s",tmp.c_str());
        strcpy(line,"\n");
        if(out)
            fprintf(out,"%s",line);

        strcpy(line,"    $bot_nick=");
        if(out)
            fprintf(out,"%s",line);
        tmp=c_var_string::script_esc(script_type,irc_nick);
        tmp+=";\n";
        if(out)
            fprintf(out,"%s",tmp.c_str());
        strcpy(line,"\n");
        if(out)
            fprintf(out,"%s",line);
        strcpy(line,"    $channel=");
        if(out)
            fprintf(out,"%s",line);
        tmp=c_var_string::script_esc(script_type,channel);
        tmp+=";\n";
        if(out)
            fprintf(out,"%s",tmp.c_str());
        strcpy(line,"\n");
        if(out)
            fprintf(out,"%s",line);
        strcpy(line,"    $params=array();\n");
        if(out)
            fprintf(out,"%s",line);

        {
            vector<c_var_base*>::iterator i1;
            unsigned i2=0;
            for(i1=params.begin(); i1!=params.end(); ++i1, ++i2) {
                sprintf(line,"    $params[%u]=",i2);
                if(out)
                    fprintf(out,"%s",line);
                string p=script_make_param(script_type,*i1);
                if(out)
                    fprintf(out,"%s",p.c_str());
                tmp=";\n";
                if(out)
                    fprintf(out,"%s",tmp.c_str());
            }
            strcpy(line,"\n");
            if(out)
                fprintf(out,"%s",line);
        }

        {
            strcpy(line,"    $users=array();");
            if(out)
                fprintf(out,"%s",line);
            vector<c_online_channel>::iterator i9;
            for(i9=irc_channels.begin(); i9!=irc_channels.end(); ++i9) {
                line[0]=0;
                if(i9==irc_channels.begin())
                    strcpy(line,"\n\n");
                strcat(line,"    // channel ");
                strcat(line,(*i9).name.c_str());
                if(out)
                    fprintf(out,"%s",line);

                vector<c_online_user>::iterator i1;
                for(i1=(*i9).users.begin(); i1!=(*i9).users.end(); ++i1) {
                    strcpy(line,"\n");
                    if(out)
                        fprintf(out,"%s",line);

                    string nick=c_var_string::script_esc(script_type,(*i1).nick);

                    strcpy(line,"    $users[");
                    if(out)
                        fprintf(out,"%s",line);
                    tmp=nick;
                    if(out)
                        fprintf(out,"%s",tmp.c_str());
                    tmp="]=Array(\"name\" => ";
                    if(out)
                        fprintf(out,"%s",tmp.c_str());
                    tmp=c_var_string::script_esc(script_type,(*i1).in_logic_as);
                    tmp+=", ";
                    if(out)
                        fprintf(out,"%s",tmp.c_str());

                    string ident=c_var_string::script_esc(script_type,(*i1).ident);
                    string host=c_var_string::script_esc(script_type,(*i1).host);
                    string fullname=c_var_string::script_esc(script_type,(*i1).fullname);

                    bool online=true;

                    sprintf(line,"\"online\" => %d, ",BOOL2INT(online));
                    if(out)
                        fprintf(out,"%s",line);

                    strcpy(line,"\"nick\" => ");
                    if(out)
                        fprintf(out,"%s",line);
                    if(out)
                        fprintf(out,"%s",nick.c_str());
                    tmp=", ";
                    if(out)
                        fprintf(out,"%s",tmp.c_str());

                    strcpy(line,"\"ident\" => ");
                    if(out)
                        fprintf(out,"%s",line);
                    if(out)
                        fprintf(out,"%s",ident.c_str());
                    tmp=", ";
                    if(out)
                        fprintf(out,"%s",tmp.c_str());

                    strcpy(line,"\"host\" => ");
                    if(out)
                        fprintf(out,"%s",line);
                    if(out)
                        fprintf(out,"%s",host.c_str());
                    tmp=", ";
                    if(out)
                        fprintf(out,"%s",tmp.c_str());

                    strcpy(line,"\"fullname\" => ");
                    if(out)
                        fprintf(out,"%s",line);
                    if(out)
                        fprintf(out,"%s",fullname.c_str());
                    tmp=", ";
                    if(out)
                        fprintf(out,"%s",tmp.c_str());

                    bool host_unknown=false;
                    bool host_bot=false;
                    bool got_i4=false;
                    deque<c_user>::iterator i4;
                    {
                        deque<c_user>::iterator i2;
                        for(i2=r_users.begin(); i2!=r_users.end(); ++i2) {
                            if((*i1).in_logic_as.compare((*i2).name))
                                continue;
                            i4=i2;
                            got_i4=true;
                            if((*i2).host_unknown && !(*i2).name.compare((*i1).in_logic_as)) {
                                host_unknown=true;
                                break;
                            }
                            if((*i2).host_bot && !(*i2).name.compare((*i1).in_logic_as)) {
                                host_bot=true;
                                break;
                            }
                        }
                    }
                    sprintf(line,"\"host_unknown\" => %d, ",BOOL2INT(host_unknown));
                    if(out)
                        fprintf(out,"%s",line);
                    sprintf(line,"\"host_bot\" => %d, ",BOOL2INT(host_bot));
                    if(out)
                        fprintf(out,"%s",line);
                    sprintf(line,"\"irc_op\" => %d, ",BOOL2INT((*i1).irc_op));
                    if(out)
                        fprintf(out,"%s",line);

                    string mode=(*i1).mode;
                    strcpy(line,"\"mode\" => \"");
                    if(out)
                        fprintf(out,"%s",line);
                    if(out)
                        fprintf(out,"%s",mode.c_str());
                    tmp="\");\n";
                    if(out)
                        fprintf(out,"%s",tmp.c_str());

                    auto_ptr<c_var_array> vars(MY_NEW c_var_array);
                    {
                        strcpy(line,"    $users[");
                        if(out)
                            fprintf(out,"%s",line);

                        tmp=c_var_string::script_esc(script_type,irc_get_nick((*i1).nick));
                        if(out)
                            fprintf(out,"%s",tmp.c_str());

                        tmp="][\"hostmasks\"]=Array(";
                        if(out)
                            fprintf(out,"%s",tmp.c_str());

                        if(got_i4) {
                            vector<string>::iterator i2;
                            unsigned i3=0;
                            for(i2=(*i4).hostmask.begin(); i2!=(*i4).hostmask.end(); ++i2, ++i3) {
                                if(i3) {
                                    strcpy(line,", ");
                                    if(out)
                                        fprintf(out,"%s",line);
                                }
                                sprintf(line,"%u => ",i3);
                                if(out)
                                    fprintf(out,"%s",line);
                                tmp=c_var_string::script_esc(script_type,*i2);
                                if(out)
                                    fprintf(out,"%s",tmp.c_str());
                            }
                        }
                    }
                    strcpy(line,");\n");
                    if(out)
                        fprintf(out,"%s",line);

                    //---

                    strcpy(line,"    $users[");
                    if(out)
                        fprintf(out,"%s",line);

                    tmp=nick;
                    if(out)
                        fprintf(out,"%s",tmp.c_str());

                    tmp="][\"fullname_masks\"]=Array(";
                    if(out)
                        fprintf(out,"%s",tmp.c_str());

                    if(got_i4) {
                        vector<string>::iterator i2;
                        unsigned i3=0;
                        for(i2=(*i4).fullname.begin(); i2!=(*i4).fullname.end(); ++i2, ++i3) {
                            if(i3) {
                                strcpy(line,", ");
                                if(out)
                                    fprintf(out,"%s",line);
                            }
                            sprintf(line,"%u => ",i3);
                            if(out)
                                fprintf(out,"%s",line);
                            tmp=c_var_string::script_esc(script_type,(*i2));
                            if(out)
                                fprintf(out,"%s",tmp.c_str());
                        }
                    }
                    strcpy(line,");\n");
                    if(out)
                        fprintf(out,"%s",line);

                    vector<c_online_channel>::iterator i2;
                    for(i2=irc_channels.begin(); i2!=irc_channels.end(); ++i2) {
                        vector<c_online_user>::iterator i3;
                        for(i3=(*i2).users.begin(); i3!=(*i2).users.end(); ++i3) {
                            if(!irc_cmp_strings_case_insensitive((*i1).nick,(*i3).nick)) {
                                tmp="    if(!isset($users[";
                                tmp+=nick;
                                tmp+="][\"channels\"])) $users[";
                                tmp+=nick;
                                tmp+="][\"channels\"]=array(); $tmp=array(); $tmp[\"channel\"]="+c_var_string::script_esc(script_type,(*i2).name);
                                tmp+="; $tmp[\"mode\"]="+c_var_string::script_esc(script_type,(*i3).mode);
                                tmp+="; array_push($users["+nick+"][\"channels\"],$tmp); unset($tmp);\n";
                                if(out)
                                    fprintf(out,"%s",tmp.c_str());
                            }
                        }
                    }

                    if(got_i4) {
                        if(out)
                            fprintf(out,"%s\n","    $groups=array();");
                        vector<string>::iterator i7;
                        for(i7=(*i4).groups.begin(); i7!=(*i4).groups.end(); ++i7)
                            if(out)
                                fprintf(out,"%s%s%s\n","    array_push($groups,",c_var_string::script_esc(script_type,*i7).c_str(),");");
                        if(out)
                            fprintf(out,"%s%s%s\n","    $users[",nick.c_str(),"][\"groups\"]=$groups; unset($groups);");
                    }
                }
                strcpy(line,"\n");
                if(out)
                    fprintf(out,"%s",line);
            }
        }

        {
            auto_ptr<c_var_array> vars(MY_NEW c_var_array);

            vector<c_online_channel>::iterator i9;
            for(i9=irc_channels.begin(); i9!=irc_channels.end(); ++i9) {
                vector<c_online_user>::iterator i1;
                for(i1=(*i9).users.begin(); i1!=(*i9).users.end(); ++i1) {
                    string nick=c_var_string::script_esc(script_type,(*i1).nick);

                    c_user u;
                    vector<c_channel> chs;
                    logic_partyline_get_user((*i1).in_logic_as,u,chs);
                    map<string,string>::iterator i2;
                    for(i2=u.meta.begin(); i2!=u.meta.end(); ++i2) {
                        strcpy(line,"    $users[");
                        if(out)
                            fprintf(out,"%s",line);

                        tmp=nick;
                        if(out)
                            fprintf(out,"%s",tmp.c_str());

                        tmp="][\"meta\"][";
                        tmp+=(string)c_var_string::script_esc(script_type,(*i2).first)+"]="+c_var_string::script_esc(script_type,(*i2).second)+";\n";
                        if(out)
                            fprintf(out,"%s",tmp.c_str());
                    }
                }
            }
        }
        strcpy(line,"\n");
        if(out)
            fprintf(out,"%s",line);

        // channels
        {
            auto_ptr<c_var_array> vars(MY_NEW c_var_array);

            tmp="";

            vector<c_online_channel>::iterator i9;
            for(i9=irc_channels.begin(); i9!=irc_channels.end(); ++i9) {
                string channel=c_var_string::script_esc(script_type,(*i9).name);

                {
                    tmp+="    $channels[";
                    tmp+=channel;
                    tmp+="][\"key\"]=";
                    tmp+=c_var_string::script_esc(script_type,(*i9).key);
                    tmp+=";\n";

                    tmp+="    $channels[";
                    tmp+=channel;
                    tmp+="][\"limit\"]=";
                    tmp+=c_var_string::script_esc(script_type,(*i9).limit);
                    tmp+=";\n";

                    tmp+="    $channels[";
                    tmp+=channel;
                    tmp+="][\"mode\"]=";
                    tmp+=c_var_string::script_esc(script_type,(*i9).modes);
                    tmp+=";\n";

                    tmp+="    $channels[";
                    tmp+=c_var_string::script_esc(script_type,channel);
                    tmp+="][\"topic\"]=";
                    tmp+=c_var_string::script_esc(script_type,(*i9).topic);
                    tmp+=";\n";
                }

                tmp+="    $channels[";
                tmp+=channel;
                tmp+="][\"bans\"]=array();\n";
                {
                    vector<string>::iterator i1;
                    for(i1=(*i9).bans.begin(); i1!=(*i9).bans.end(); ++i1) {
                        tmp+="    array_push($channels[";
                        tmp+=channel;
                        tmp+="][\"bans\"],";
                        tmp+=c_var_string::script_esc(script_type,*i1);
                        tmp+=");\n";
                    }
                }

                tmp+="    $channels[";
                tmp+=channel;
                tmp+="][\"exceptions\"]=array();\n";
                {
                    vector<string>::iterator i1;
                    for(i1=(*i9).excepts.begin(); i1!=(*i9).excepts.end(); ++i1) {
                        tmp+="    array_push($channels[";
                        tmp+=channel;
                        tmp+="][\"exceptions\"],";
                        tmp+=c_var_string::script_esc(script_type,*i1);
                        tmp+=");\n";
                    }
                }

                tmp+="    $channels[";
                tmp+=channel;
                tmp+="][\"invites\"]=array();\n";
                {
                    vector<string>::iterator i1;
                    for(i1=(*i9).invites.begin(); i1!=(*i9).invites.end(); ++i1) {
                        tmp+="    array_push($channels[";
                        tmp+=channel;
                        tmp+="][\"invites\"],";
                        tmp+=c_var_string::script_esc(script_type,*i1);
                        tmp+=");\n";
                    }
                }

                tmp+="    $channels[";
                tmp+=channel;
                tmp+="][\"reops\"]=array();\n";
                {
                    vector<string>::iterator i1;
                    for(i1=(*i9).reops.begin(); i1!=(*i9).reops.end(); ++i1) {
                        tmp+="    array_push($channels[";
                        tmp+=channel;
                        tmp+="][\"reops\"],";
                        tmp+=c_var_string::script_esc(script_type,*i1);
                        tmp+=");\n";
                    }
                }

                if(out)
                    fprintf(out,"%s",tmp.c_str());
                tmp="";
            }
        }

        // RPL_ISUPPORT (005 numeric)
        {
            map<string,string> raw;
            string chm_a, chm_b, chm_c, chm_d;
            map<char,char> prefix;
            irc_get_005(raw,prefix,chm_a,chm_b,chm_c,chm_d);
            map<string,string>::iterator i1;
            tmp="\n    $isupport=array();\n";
            for(i1=raw.begin(); i1!=raw.end(); ++i1) {
                tmp+=(string)"    $isupport[";
                tmp+=c_var_string::script_esc(script_type,(*i1).first);
                tmp+=(string)"]=";
                tmp+=c_var_string::script_esc(script_type,(*i1).second);
                tmp+=(string)";\n";
            }
            tmp+="\n";
            map<char,char>::iterator i2;
            tmp+="    $user_mode_prefix=array();\n";
            for(i2=prefix.begin(); i2!=prefix.end(); ++i2) {
                tmp+=(string)"    $user_mode_prefix[";
                tmp+=c_var_string::script_esc(script_type,(string)""+(*i2).first);
                tmp+=(string)"]=";
                tmp+=c_var_string::script_esc(script_type,(string)""+(*i2).second);
                tmp+=(string)";\n";
            }
            tmp+="\n";

            tmp+="    $chan_modes_class_a=";
            tmp+=c_var_string::script_esc(script_type,chm_a);
            tmp+=";\n";

            tmp+="    $chan_modes_class_b=";
            tmp+=c_var_string::script_esc(script_type,chm_b);
            tmp+=";\n";

            tmp+="    $chan_modes_class_c=";
            tmp+=c_var_string::script_esc(script_type,chm_c);
            tmp+=";\n";

            tmp+="    $chan_modes_class_d=";
            tmp+=c_var_string::script_esc(script_type,chm_d);
            tmp+=";\n";

            tmp+="\n";
            if(out)
                fprintf(out,"%s",tmp.c_str());
            tmp="";
        }

        // filesystem
        {
            tmp=filesys_get_script(script_type);
            if(out)
                fprintf(out,"%s",tmp.c_str());
            tmp="";
        }

        while(in && !feof(in)) {
            if(fgets(line,1024*10-1,in)==NULL)
                break;
            line[1024*10-1]=0;
            if(line[0] && line[strlen(line)-1]=='\r')
                line[strlen(line)-1]=0;
            if(line[0] && line[strlen(line)-1]=='\n')
                line[strlen(line)-1]=0;
            if(line[0] && line[strlen(line)-1]=='\r')
                line[strlen(line)-1]=0;
            if(line[0] && line[strlen(line)-1]=='\n')
                line[strlen(line)-1]=0;
            if(!strcmp(line,"// RESULT"))
                break;
            if(out)
                fprintf(out,"%s",line);
            tmp="\n";
            if(out)
                fprintf(out,"%s",tmp.c_str());
            line[0]=0;
        }

        if(!noremove)
            tmp="    unlink(\"./";
        else
            tmp="    // unlink(\"./";
        tmp+=fn;
        tmp+="\");\n\n";
        if(out)
            fprintf(out,"%s",tmp.c_str());

        strcpy(line,"    echo \"#####\\n\";\n");
        if(out)
            fprintf(out,"%s",line);
        strcpy(line,"    foreach($result as $element)\n");
        if(out)
            fprintf(out,"%s",line);
        strcpy(line,"        echo $element.\"\\n\";\n");
        if(out)
            fprintf(out,"%s",line);
        strcpy(line,"    echo \"#####\\n\";\n\n");
        if(out)
            fprintf(out,"%s",line);

        while(in && !feof(in)) {
            if(fgets(line,1024*10-1,in)==NULL)
                break;
            line[1024*10-1]=0;
            if(line[0] && line[strlen(line)-1]=='\r')
                line[strlen(line)-1]=0;
            if(line[0] && line[strlen(line)-1]=='\n')
                line[strlen(line)-1]=0;
            if(line[0] && line[strlen(line)-1]=='\r')
                line[strlen(line)-1]=0;
            if(line[0] && line[strlen(line)-1]=='\n')
                line[strlen(line)-1]=0;
            if(out)
                fprintf(out,"%s",line);
            tmp="\n";
            if(out)
                fprintf(out,"%s",tmp.c_str());
            line[0]=0;
        }

        if(out)
            fclose(out);
        if(in)
            fclose(in);

        c_tmp_file t;
        t.script=true;
        t.script_num=script_num;
        t.script_type=script_type;
        t.command_line=conf_getvar(script_type+(string)"_processor");

        t.command_args="\"." FILE_SLASH;
        t.command_args+=fn;
        t.command_args+="\"";
        if(!conf_getvar(script_type+(string)"_processor_args").empty()) {
            t.command_args+=" ";
            t.command_args+=conf_getvar(script_type+(string)"_processor_args");
        }
        t.command_dir=script_dir;

        t.has_been_run=false;
        t.start_up_time=0;
        t.noremove=noremove;
        t.filename=fn;
        tmp_files.push_back(t);

        MY_DELETE_A(line);
        line=NULL;
    }

    if(!script_type.compare("php_2")) {
        string tmp=script_type+(string)"_";
        tmp+=tos(script_num);

        string script_dir=utils_convert_unix_path(dir_prefix);
        if(script_dir.length() && script_dir[script_dir.length()-1]!=FILE_SLASH_CH)
            script_dir+=FILE_SLASH;
        script_dir+=utils_convert_unix_path(conf_getvar("php_script_dir"));

        string script=utils_convert_unix_path(script_dir);
        script+=utils_convert_unix_path(conf_getvar(tmp));

        string fn="tmp_";
        fn+=toshex(rand());
        fn+=toshex(rand());
        fn+=toshex(rand());
        fn+=toshex(rand());
        string fn2=fn;
        fn+=(string)".php";
        fn2+=".tmp";

        string tmp_fn1=script_dir;
        string tmp_fn2=tmp_fn1;
        tmp_fn1+=fn;
        tmp_fn2+=fn2;

        FILE* out=fopen(tmp_fn1.c_str(),"w");

        char* line=MY_NEW char[1024*10+1];

        if(script.empty())
            return;
        FILE* in=fopen(script.c_str(),"r");
        line[0]=0;
        while(in && !feof(in)) {
            if(fgets(line,1024*10-1,in)==NULL)
                break;
            line[1024*10]=0;
            if(line[0] && line[strlen(line)-1]=='\r')
                line[strlen(line)-1]=0;
            if(line[0] && line[strlen(line)-1]=='\n')
                line[strlen(line)-1]=0;
            if(line[0] && line[strlen(line)-1]=='\r')
                line[strlen(line)-1]=0;
            if(line[0] && line[strlen(line)-1]=='\n')
                line[strlen(line)-1]=0;
            if(!strcmp(line,"// INIT"))
                break;
            if(strlen(line))
                if(out)
                    fprintf(out,"%s",line);
            string tmp="\n";
            if(out)
                fprintf(out,"%s",tmp.c_str());
            line[0]=0;
        }


        string ln;

        ln="    $uptime=";
        ln.append(tos(static_cast<unsigned long>(stats_uptime())));
        ln.append(";\n");
        if(out)
            fprintf(out,"%s",ln.c_str());

        // c_server
        ln="    $server=new c_server; $server->init(";
        ln.append(c_var_string::script_esc(script_type,irc_server_host));
        ln.append(",");
        ln.append(tos(irc_server_port));
        ln.append(",");
        ln.append(c_var_string::script_esc(script_type,irc_nick));
        ln.append(");\n");
        if(out)
            fprintf(out,"%s",ln.c_str());

        // RPL_ISUPPORT (005 numeric)
        {
            map<string,string> raw;
            string chm_a, chm_b, chm_c, chm_d;
            map<char,char> prefix;
            irc_get_005(raw,prefix,chm_a,chm_b,chm_c,chm_d);
            map<string,string>::iterator i1;
            tmp="\n";
            for(i1=raw.begin(); i1!=raw.end(); ++i1) {
                tmp+=(string)"    $server->isupport[";
                tmp+=c_var_string::script_esc(script_type,(*i1).first);
                tmp+=(string)"]=";
                tmp+=c_var_string::script_esc(script_type,(*i1).second);
                tmp+=(string)";\n";
            }
            tmp+="\n";
            map<char,char>::iterator i2;
            for(i2=prefix.begin(); i2!=prefix.end(); ++i2) {
                tmp+=(string)"    $server->user_mode_prefix[";
                tmp+=c_var_string::script_esc(script_type,(string)""+(*i2).first);
                tmp+=(string)"]=";
                tmp+=c_var_string::script_esc(script_type,(string)""+(*i2).second);
                tmp+=(string)";\n";
            }
            tmp+="\n";

            tmp+="    $server->chan_modes_class_a=";
            tmp+=c_var_string::script_esc(script_type,chm_a);
            tmp+=";\n";

            tmp+="    $server->chan_modes_class_b=";
            tmp+=c_var_string::script_esc(script_type,chm_b);
            tmp+=";\n";

            tmp+="    $server->chan_modes_class_c=";
            tmp+=c_var_string::script_esc(script_type,chm_c);
            tmp+=";\n";

            tmp+="    $server->chan_modes_class_d=";
            tmp+=c_var_string::script_esc(script_type,chm_d);
            tmp+=";\n";

            tmp+="\n";
            if(out)
                fprintf(out,"%s",tmp.c_str());
            tmp="";
        }

        // c_input

        ln="\n    $input=new c_input; $input->init(";
        ln.append(c_var_string::script_esc(script_type,channel));
        ln.append(");\n");
        if(out)
            fprintf(out,"%s",ln.c_str());

        {
            vector<c_var_base*>::iterator i1;
            int i2=0;
            for(i1=params.begin(); i1!=params.end(); ++i1, ++i2) {
                ln="    $input->set_next_parameter(";
                ln.append(script_make_param(script_type,*i1));
                ln.append(");\n");
                if(out)
                    fprintf(out,"%s",ln.c_str());
            }
            strcpy(line,"\n");
            if(out)
                fprintf(out,"%s",line);
        }

        {
            ln="\n    $channels=array();\n\n";
            if(out)
                fprintf(out,"%s",ln.c_str());

            auto_ptr<c_var_array> vars(MY_NEW c_var_array);

            vector<c_online_channel>::iterator i9;
            for(i9=irc_channels.begin(); i9!=irc_channels.end(); ++i9) {
                vector<string>::iterator i1;

                ln="    $bans=array();\n";
                if(out)
                    fprintf(out,"%s",ln.c_str());
                for(i1=(*i9).bans.begin(); i1!=(*i9).bans.end(); ++i1) {
                    ln="    $x=";
                    ln.append(c_var_string::script_esc(script_type,*i1));
                    ln.append("; array_push($bans,$x); unset($x);\n");
                    if(out)
                        fprintf(out,"%s",ln.c_str());
                }

                ln="    $excepts=array();\n";
                if(out)
                    fprintf(out,"%s",ln.c_str());
                for(i1=(*i9).excepts.begin(); i1!=(*i9).excepts.end(); ++i1) {
                    ln="    $x=";
                    ln.append(c_var_string::script_esc(script_type,*i1));
                    ln.append("; array_push($excepts,$x); unset($x);\n");
                    if(out)
                        fprintf(out,"%s",ln.c_str());
                }

                ln="    $invites=array();\n";
                if(out)
                    fprintf(out,"%s",ln.c_str());
                for(i1=(*i9).invites.begin(); i1!=(*i9).invites.end(); ++i1) {
                    ln="    $x=";
                    ln.append(c_var_string::script_esc(script_type,*i1));
                    ln.append("; array_push($invites,$x); unset($x);\n");
                    if(out)
                        fprintf(out,"%s",ln.c_str());
                }

                ln="    $reops=array();\n";
                if(out)
                    fprintf(out,"%s",ln.c_str());
                for(i1=(*i9).reops.begin(); i1!=(*i9).reops.end(); ++i1) {
                    ln="    $x=";
                    ln.append(c_var_string::script_esc(script_type,*i1));
                    ln.append("; array_push($reops,$x); unset($x);\n");
                    if(out)
                        fprintf(out,"%s",ln.c_str());
                }

                ln="\n    $channel=new c_channel; $channel->init(";
                ln.append(c_var_string::script_esc(script_type,(*i9).name));
                ln.append(",$bans,$excepts,$invites,$reops);\n    unset($bans); unset($excepts); unset($invites); unset($reops);\n");
                if(out)
                    fprintf(out,"%s",ln.c_str());

                {
                    ln="";

                    ln+="    $channel->key=";
                    ln+=c_var_string::script_esc(script_type,(*i9).key);
                    ln+=";\n";

                    ln+="    $channel->limit=";
                    ln+=c_var_string::script_esc(script_type,(*i9).limit);
                    ln+=";\n";

                    ln+="    $channel->mode=";
                    ln+=c_var_string::script_esc(script_type,(*i9).modes);
                    ln+=";\n";

                    ln+="    $channel->topic=";
                    ln+=c_var_string::script_esc(script_type,(*i9).topic);
                    ln+=";\n";

                    if(out)
                        fprintf(out,"%s",ln.c_str());
                    ln="";
                }

                vector<c_channel>::iterator i2;
                for(i2=r_channels.begin(); i2!=r_channels.end(); ++i2) {
                    if(irc_cmp_strings_case_insensitive((*i9).name,(*i2).channel_name))
                        continue;

                    bool is_online=false;
                    vector<c_online_user>::iterator i3;
                    for(i3=(*i9).users.begin(); i3!=(*i9).users.end(); ++i3) {
                        if((*i3).in_logic_as.compare((*i2).username))
                            continue;
                        is_online=true;

                        deque<c_user>::iterator i4;
                        for(i4=r_users.begin(); i4!=r_users.end(); ++i4) {
                            if((*i4).name.compare((*i3).in_logic_as))
                                continue;
                            break;
                        }
                        if(i4==r_users.end())
                            break;

                        ln="\n    $host_masks=array();\n";
                        if(out)
                            fprintf(out,"%s",ln.c_str());
                        vector<string>::iterator i5;
                        for(i5=(*i4).hostmask.begin(); i5!=(*i4).hostmask.end(); ++i5) {
                            ln="    $x=";
                            ln.append(c_var_string::script_esc(script_type,*i5));
                            ln.append("; array_push($host_masks,$x);\n");
                            if(out)
                                fprintf(out,"%s",ln.c_str());
                        }
                        ln="    $fullname_masks=array();\n";
                        if(out)
                            fprintf(out,"%s",ln.c_str());
                        for(i5=(*i4).fullname.begin(); i5!=(*i4).fullname.end(); ++i5) {
                            ln="    $x=";
                            ln.append(c_var_string::script_esc(script_type,*i5));
                            ln.append("; array_push($fullname_masks,$x);\n");
                            if(out)
                                fprintf(out,"%s",ln.c_str());
                        }
                        ln="    unset($x);\n    $meta=array();\n";
                        if(out)
                            fprintf(out,"%s",ln.c_str());
                        map<string,string>::iterator i6;
                        for(i6=(*i4).meta.begin(); i6!=(*i4).meta.end(); ++i6) {
                            ln="    $meta[";
                            ln.append(c_var_string::script_esc(script_type,(*i6).first));
                            ln.append("]=");
                            ln.append(c_var_string::script_esc(script_type,(*i6).second));
                            ln.append(";\n");
                            if(out)
                                fprintf(out,"%s",ln.c_str());
                        }

                        if(out)
                            fprintf(out,"%s\n","    $groups=array();");
                        vector<string>::iterator i7;
                        for(i7=(*i4).groups.begin(); i7!=(*i4).groups.end(); ++i7)
                            if(out)
                                fprintf(out,"%s%s%s\n","    array_push($groups,",c_var_string::script_esc(script_type,*i7).c_str(),");");

                        ln="    $user=new c_user; $user->init(";
                        ln.append(c_var_string::script_esc(script_type,(*i3).nick));
                        ln.append(",");
                        ln.append(c_var_string::script_esc(script_type,(*i3).ident));
                        ln.append(",");
                        ln.append(c_var_string::script_esc(script_type,(*i3).host));
                        ln.append(",");
                        ln.append(c_var_string::script_esc(script_type,(*i3).fullname));
                        ln.append(",");
                        ln.append(c_var_string::script_esc(script_type,(*i3).in_logic_as));
                        ln.append(",TRUE,"); // $online
                        if((*i2).host_unknown)
                            ln.append("TRUE,");
                        else
                            ln.append("FALSE,");
                        if((*i4).host_bot)
                            ln.append("TRUE,");
                        else
                            ln.append("FALSE,");
                        if((*i3).irc_op)
                            ln.append("TRUE,");
                        else
                            ln.append("FALSE,");
                        ln.append(c_var_string::script_esc(script_type,(*i3).mode));
                        ln.append(",$host_masks,$fullname_masks,$meta,$groups);\n    unset($host_masks); unset($fullname_masks); unset($meta); unset($groups);\n");

                        if(out)
                            fprintf(out,"%s",ln.c_str());

                        ln="    $channel->add_user($user); unset($user);\n";
                        if(out)
                            fprintf(out,"%s",ln.c_str());

                        // do NOT break! There could be cloned users!
                        //break;
                    }

                    if(!is_online) {
                        deque<c_user>::iterator i4;
                        for(i4=r_users.begin(); i4!=r_users.end(); ++i4) {
                            if((*i4).name.compare((*i2).username))
                                continue;
                            break;
                        }
                        if(i4==r_users.end())
                            continue;

                        ln="\n    $host_masks=array();\n";
                        if(out)
                            fprintf(out,"%s",ln.c_str());
                        vector<string>::iterator i5;
                        for(i5=(*i4).hostmask.begin(); i5!=(*i4).hostmask.end(); ++i5) {
                            ln="    $x=";
                            ln.append(c_var_string::script_esc(script_type,*i5));
                            ln.append("; array_push($host_masks,$x);\n");
                            if(out)
                                fprintf(out,"%s",ln.c_str());
                        }
                        ln="    $fullname_masks=array();\n";
                        if(out)
                            fprintf(out,"%s",ln.c_str());
                        for(i5=(*i4).fullname.begin(); i5!=(*i4).fullname.end(); ++i5) {
                            ln="    $x=";
                            ln.append(c_var_string::script_esc(script_type,*i5));
                            ln.append("; array_push($fullname_masks,$x);\n");
                            if(out)
                                fprintf(out,"%s",ln.c_str());
                        }
                        ln="    unset($x);\n    $meta=array();\n";
                        if(out)
                            fprintf(out,"%s",ln.c_str());
                        map<string,string>::iterator i6;
                        for(i6=(*i4).meta.begin(); i6!=(*i4).meta.end(); ++i6) {
                            ln="    $meta[";
                            ln.append(c_var_string::script_esc(script_type,(*i6).first));
                            ln.append("]=");
                            ln.append(c_var_string::script_esc(script_type,(*i6).second));
                            ln.append(";\n");
                            if(out)
                                fprintf(out,"%s",ln.c_str());
                        }

                        if(out)
                            fprintf(out,"%s\n","    $groups=array();");
                        vector<string>::iterator i7;
                        for(i7=(*i4).groups.begin(); i7!=(*i4).groups.end(); ++i7)
                            if(out)
                                fprintf(out,"%s%s%s\n","    array_push($groups,",c_var_string::script_esc(script_type,*i7).c_str(),");");

                        ln="    $user=new c_user; $user->init(\"";
                        // empty nick
                        ln.append("\",\"");
                        // empty ident
                        ln.append("\",\"");
                        // empty host
                        ln.append("\",\"");
                        // empty fullname
                        ln.append("\",");
                        ln.append(c_var_string::script_esc(script_type,(*i4).name));
                        ln.append(",FALSE,"); // $online
                        if((*i2).host_unknown)
                            ln.append("TRUE,");
                        else
                            ln.append("FALSE,");
                        if((*i4).host_bot)
                            ln.append("TRUE,");
                        else
                            ln.append("FALSE,");
                        ln.append("FALSE,"); // irc_op
                        ln.append("\"");
                        ln.append("0"); // mode
                        ln.append("\",$host_masks,$fullname_masks,$meta,$groups);\n    unset($host_masks); unset($fullname_masks); unset($meta); unset($groups);\n");

                        if(out)
                            fprintf(out,"%s",ln.c_str());

                        ln="    $channel->add_user($user); unset($user);\n";
                        if(out)
                            fprintf(out,"%s",ln.c_str());
                    }
                }
                ln="    $channels[strtolower(\"";
                ln.append((*i9).name);
                ln.append("\")]=$channel; unset($channel);\n");
                if(out)
                    fprintf(out,"%s",ln.c_str());
            }
            ln="\n";
            if(out)
                fprintf(out,"%s",ln.c_str());
        }

        // filesystem
        {
            tmp=filesys_get_script(script_type);
            if(out)
                fprintf(out,"%s",tmp.c_str());
            tmp="";
        }

        line[0]=0;
        while(in && !feof(in)) {
            if(fgets(line,1024*10-1,in)==NULL)
                break;
            line[1024*10]=0;
            if(line[0] && line[strlen(line)-1]=='\r')
                line[strlen(line)-1]=0;
            if(line[0] && line[strlen(line)-1]=='\n')
                line[strlen(line)-1]=0;
            if(line[0] && line[strlen(line)-1]=='\r')
                line[strlen(line)-1]=0;
            if(line[0] && line[strlen(line)-1]=='\n')
                line[strlen(line)-1]=0;
            if(!strcmp(line,"// RESULT"))
                break;
            if(strlen(line))
                if(out)
                    fprintf(out,"%s",line);
            tmp="\n";
            if(out)
                fprintf(out,"%s",tmp.c_str());
            line[0]=0;
        }

        if(!noremove)
            tmp="    unlink(\"./";
        else
            tmp="    // unlink(\"./";
        tmp+=fn;
        tmp+="\");\n\n";
        if(out)
            fprintf(out,"%s",tmp.c_str());

        line[0]=0;
        while(in && !feof(in)) {
            if(fgets(line,1024*10-1,in)==NULL)
                break;
            line[1024*10]=0;
            if(line[0] && line[strlen(line)-1]=='\r')
                line[strlen(line)-1]=0;
            if(line[0] && line[strlen(line)-1]=='\n')
                line[strlen(line)-1]=0;
            if(line[0] && line[strlen(line)-1]=='\r')
                line[strlen(line)-1]=0;
            if(line[0] && line[strlen(line)-1]=='\n')
                line[strlen(line)-1]=0;
            if(strlen(line))
                if(out)
                    fprintf(out,"%s",line);
            tmp="\n";
            if(out)
                fprintf(out,"%s",tmp.c_str());
            line[0]=0;
        }

        if(out)
            fclose(out);
        if(in)
            fclose(in);

        c_tmp_file t;
        t.script=true;
        t.script_num=script_num;
        t.script_type=script_type;
        t.command_line=utils_convert_unix_path(conf_getvar("php_processor"));

        t.command_args="\"." FILE_SLASH;
        t.command_args+=fn.c_str();
        t.command_args+="\"";
        if(!conf_getvar("php_processor_args").empty()) {
            t.command_args+=" ";
            t.command_args+=conf_getvar("php_processor_args");
        }
        t.command_dir=script_dir;

        t.has_been_run=false;
        t.start_up_time=0;
        t.noremove=noremove;
        t.filename=fn;
        tmp_files.push_back(t);

        MY_DELETE_A(line);
        line=NULL;
    }
    if(!script_type.compare("perl")) {
        string tmp=script_type+(string)"_";
        tmp+=tos(script_num);

        string script_dir=utils_convert_unix_path(dir_prefix);
        if(script_dir.length() && script_dir[script_dir.length()-1]!=FILE_SLASH_CH)
            script_dir+=FILE_SLASH;
        script_dir+=utils_convert_unix_path(conf_getvar("perl_script_dir"));

        string script=utils_convert_unix_path(script_dir);
        script+=utils_convert_unix_path(conf_getvar(tmp));

        string fn="tmp_";
        fn+=toshex(rand());
        fn+=toshex(rand());
        fn+=toshex(rand());
        fn+=toshex(rand());
        string fn2=fn;
        fn+=(string)".pl";
        fn2+=".tmp";

        string tmp_fn1=script_dir;
        string tmp_fn2=tmp_fn1;
        tmp_fn1+=fn;
        tmp_fn2+=fn2;

        FILE* out=fopen(tmp_fn1.c_str(),"w");

        char* line=MY_NEW char[1024*10+1];

        if(script.empty())
            return;
        FILE* in=fopen(script.c_str(),"r");
        line[0]=0;
        while(in && !feof(in)) {
            if(fgets(line,1024*10-1,in)==NULL)
                break;
            line[1024*10]=0;
            if(line[0] && line[strlen(line)-1]=='\r')
                line[strlen(line)-1]=0;
            if(line[0] && line[strlen(line)-1]=='\n')
                line[strlen(line)-1]=0;
            if(line[0] && line[strlen(line)-1]=='\r')
                line[strlen(line)-1]=0;
            if(line[0] && line[strlen(line)-1]=='\n')
                line[strlen(line)-1]=0;
            if(!strcmp(line,"# INIT"))
                break;
            if(strlen(line))
                if(out)
                    fprintf(out,"%s",line);
            string tmp="\n";
            if(out)
                fprintf(out,"%s",tmp.c_str());
            line[0]=0;
        }

        string ln;

        ln="    my $x;\n\n";
        ln+="    my $uptime=";
        ln.append(tos(static_cast<unsigned long>(stats_uptime())));
        ln.append(";\n");
        if(out)
            fprintf(out,"%s",ln.c_str());

        // c_server
        ln="    my($server)=c_server->init(";
        ln.append(c_var_string::script_esc(script_type,irc_server_host));
        ln.append(",");
        ln.append(tos(irc_server_port));
        ln.append(",");
        ln.append(c_var_string::script_esc(script_type,irc_nick));
        ln.append(");\n");
        if(out)
            fprintf(out,"%s",ln.c_str());

        // RPL_ISUPPORT (005 numeric)
        {
            map<string,string> raw;
            string chm_a, chm_b, chm_c, chm_d;
            map<char,char> prefix;
            irc_get_005(raw,prefix,chm_a,chm_b,chm_c,chm_d);
            map<string,string>::iterator i1;
            tmp="\n";
            for(i1=raw.begin(); i1!=raw.end(); ++i1) {
                tmp+=(string)"    $server->{isupport}{";
                tmp+=c_var_string::script_esc(script_type,(*i1).first);
                tmp+=(string)"}=";
                tmp+=c_var_string::script_esc(script_type,(*i1).second);
                tmp+=(string)";\n";
            }
            tmp+="\n";
            map<char,char>::iterator i2;
            for(i2=prefix.begin(); i2!=prefix.end(); ++i2) {
                tmp+=(string)"    $server->{user_mode_prefix}{";
                tmp+=c_var_string::script_esc(script_type,(string)""+(*i2).first);
                tmp+=(string)"}=";
                tmp+=c_var_string::script_esc(script_type,(string)""+(*i2).second);
                tmp+=(string)";\n";
            }
            tmp+="\n";

            tmp+="    $server->{chan_modes_class_a}=";
            tmp+=c_var_string::script_esc(script_type,chm_a);
            tmp+=";\n";

            tmp+="    $server->{chan_modes_class_b}=";
            tmp+=c_var_string::script_esc(script_type,chm_b);
            tmp+=";\n";

            tmp+="    $server->{chan_modes_class_c}=";
            tmp+=c_var_string::script_esc(script_type,chm_c);
            tmp+=";\n";

            tmp+="    $server->{chan_modes_class_d}=";
            tmp+=c_var_string::script_esc(script_type,chm_d);
            tmp+=";\n";

            tmp+="\n";
            if(out)
                fprintf(out,"%s",tmp.c_str());
            tmp="";
        }

        // c_input

        ln="\n    my $input=c_input->init(";
        ln.append(c_var_string::script_esc(script_type,channel));
        ln.append(");\n");
        if(out)
            fprintf(out,"%s",ln.c_str());

        {
            vector<c_var_base*>::iterator i1;
            int i2=0;
            for(i1=params.begin(); i1!=params.end(); ++i1, ++i2) {
                ln="    $input->set_next_parameter(";
                ln.append(script_make_param(script_type,*i1));
                ln.append(");\n");
                if(out)
                    fprintf(out,"%s",ln.c_str());
            }
            strcpy(line,"\n");
            if(out)
                fprintf(out,"%s",line);
        }

        {
            ln="    my %channels;\n    my (@bans, @excepts, @invites, @reops);\n    my (@groups, @fullname_masks, @host_masks);\n    my %meta;\n    my $user;\n    my $channel;\n";
            if(out)
                fprintf(out,"%s",ln.c_str());
        }

        {
            ln="\n";
            if(out)
                fprintf(out,"%s",ln.c_str());

            auto_ptr<c_var_array> vars(MY_NEW c_var_array);

            vector<c_online_channel>::iterator i9;
            for(i9=irc_channels.begin(); i9!=irc_channels.end(); ++i9) {
                vector<string>::iterator i1;

                for(i1=(*i9).bans.begin(); i1!=(*i9).bans.end(); ++i1) {
                    ln="    $x=";
                    ln.append(c_var_string::script_esc(script_type,*i1));
                    ln.append("; push(@bans,$x); undef($x);\n");
                    if(out)
                        fprintf(out,"%s",ln.c_str());
                }

                for(i1=(*i9).excepts.begin(); i1!=(*i9).excepts.end(); ++i1) {
                    ln="    $x=";
                    ln.append(c_var_string::script_esc(script_type,*i1));
                    ln.append("; push(@excepts,$x); undef($x);\n");
                    if(out)
                        fprintf(out,"%s",ln.c_str());
                }

                for(i1=(*i9).invites.begin(); i1!=(*i9).invites.end(); ++i1) {
                    ln="    $x=";
                    ln.append(c_var_string::script_esc(script_type,*i1));
                    ln.append("; push(@invites,$x); undef($x);\n");
                    if(out)
                        fprintf(out,"%s",ln.c_str());
                }

                for(i1=(*i9).reops.begin(); i1!=(*i9).reops.end(); ++i1) {
                    ln="    $x=";
                    ln.append(c_var_string::script_esc(script_type,*i1));
                    ln.append("; push(@reops,$x); undef($x);\n");
                    if(out)
                        fprintf(out,"%s",ln.c_str());
                }

                ln="\n    $channel=c_channel->init(";
                ln.append(c_var_string::script_esc(script_type,(*i9).name));
                ln.append(",@bans,@excepts,@invites,@reops);\n    undef(@bans); undef(@excepts); undef(@invites); undef(@reops);\n");
                if(out)
                    fprintf(out,"%s",ln.c_str());

                {
                    ln="";

                    ln+="    $channel->{key}=";
                    ln+=c_var_string::script_esc(script_type,(*i9).key);
                    ln+=";\n";

                    ln+="    $channel->{limit}=";
                    ln+=c_var_string::script_esc(script_type,(*i9).limit);
                    ln+=";\n";

                    ln+="    $channel->{mode}=";
                    ln+=c_var_string::script_esc(script_type,(*i9).modes);
                    ln+=";\n";

                    ln+="    $channel->{topic}=";
                    ln+=c_var_string::script_esc(script_type,(*i9).topic);
                    ln+=";\n\n";

                    if(out)
                        fprintf(out,"%s",ln.c_str());
                    ln="";
                }

                vector<c_channel>::iterator i2;
                for(i2=r_channels.begin(); i2!=r_channels.end(); ++i2) {
                    if(irc_cmp_strings_case_insensitive((*i9).name,(*i2).channel_name))
                        continue;

                    bool is_online=false;
                    vector<c_online_user>::iterator i3;
                    for(i3=(*i9).users.begin(); i3!=(*i9).users.end(); ++i3) {
                        if((*i3).in_logic_as.compare((*i2).username))
                            continue;
                        is_online=true;

                        deque<c_user>::iterator i4;
                        for(i4=r_users.begin(); i4!=r_users.end(); ++i4) {
                            if((*i4).name.compare((*i3).in_logic_as))
                                continue;
                            break;
                        }
                        if(i4==r_users.end())
                            break;

                        vector<string>::iterator i5;
                        for(i5=(*i4).hostmask.begin(); i5!=(*i4).hostmask.end(); ++i5) {
                            ln="    $x=";
                            ln.append(c_var_string::script_esc(script_type,*i5));
                            ln.append("; push(@host_masks,$x); undef($x);\n");
                            if(out)
                                fprintf(out,"%s",ln.c_str());
                        }
                        for(i5=(*i4).fullname.begin(); i5!=(*i4).fullname.end(); ++i5) {
                            ln="    $x=";
                            ln.append(c_var_string::script_esc(script_type,*i5));
                            ln.append("; push(@fullname_masks,$x); undef($x);\n");
                            if(out)
                                fprintf(out,"%s",ln.c_str());
                        }
                        map<string,string>::iterator i6;
                        for(i6=(*i4).meta.begin(); i6!=(*i4).meta.end(); ++i6) {
                            ln="    $meta{";
                            ln.append(c_var_string::script_esc(script_type,(*i6).first));
                            ln.append("}=");
                            ln.append(c_var_string::script_esc(script_type,(*i6).second));
                            ln.append(";\n");
                            if(out)
                                fprintf(out,"%s",ln.c_str());
                        }

                        if(out)
                            fprintf(out,"%s\n","    undef(@groups);");
                        vector<string>::iterator i7;
                        for(i7=(*i4).groups.begin(); i7!=(*i4).groups.end(); ++i7)
                            if(out)
                                fprintf(out,"%s%s%s\n","    push(@groups,",c_var_string::script_esc(script_type,*i7).c_str(),");");

                        ln="    $user=c_user->init(";
                        ln.append(c_var_string::script_esc(script_type,(*i3).nick));
                        ln.append(",");
                        ln.append(c_var_string::script_esc(script_type,(*i3).ident));
                        ln.append(",");
                        ln.append(c_var_string::script_esc(script_type,(*i3).host));
                        ln.append(",");
                        ln.append(c_var_string::script_esc(script_type,(*i3).fullname));
                        ln.append(",");
                        ln.append(c_var_string::script_esc(script_type,(*i3).in_logic_as));
                        ln.append(",1,"); // $online
                        if((*i2).host_unknown)
                            ln.append("1,");
                        else
                            ln.append("0,");
                        if((*i4).host_bot)
                            ln.append("1,");
                        else
                            ln.append("0,");
                        if((*i3).irc_op)
                            ln.append("1,");
                        else
                            ln.append("0,");
                        ln.append(c_var_string::script_esc(script_type,(*i3).mode));
                        ln.append(",\\@host_masks,\\@fullname_masks,\\%meta,\\@groups);\n");

                        if(out)
                            fprintf(out,"%s",ln.c_str());

                        ln="    $channel->add_user($user); undef($user);\n";
                        if(out)
                            fprintf(out,"%s%s",ln.c_str(),"\n    undef(@host_masks); undef(@fullname_masks); undef(%meta); undef(@groups);\n\n");

                        // do NOT break! There could be cloned users!
                        //break;
                    }

                    if(!is_online) {
                        deque<c_user>::iterator i4;
                        for(i4=r_users.begin(); i4!=r_users.end(); ++i4) {
                            if((*i4).name.compare((*i2).username))
                                continue;
                            break;
                        }
                        if(i4==r_users.end())
                            continue;

                        vector<string>::iterator i5;
                        for(i5=(*i4).hostmask.begin(); i5!=(*i4).hostmask.end(); ++i5) {
                            ln="    $x=";
                            ln.append(c_var_string::script_esc(script_type,*i5));
                            ln.append("; push(@host_masks,$x); undef($x);\n");
                            if(out)
                                fprintf(out,"%s",ln.c_str());
                        }
                        for(i5=(*i4).fullname.begin(); i5!=(*i4).fullname.end(); ++i5) {
                            ln="    $x=";
                            ln.append(c_var_string::script_esc(script_type,*i5));
                            ln.append("; push(@fullname_masks,$x); undef($x);\n");
                            if(out)
                                fprintf(out,"%s",ln.c_str());
                        }
                        map<string,string>::iterator i6;
                        for(i6=(*i4).meta.begin(); i6!=(*i4).meta.end(); ++i6) {
                            ln="    $meta{";
                            ln.append(c_var_string::script_esc(script_type,(*i6).first));
                            ln.append("}=");
                            ln.append(c_var_string::script_esc(script_type,(*i6).second));
                            ln.append(";\n");
                            if(out)
                                fprintf(out,"%s",ln.c_str());
                        }

                        vector<string>::iterator i7;
                        for(i7=(*i4).groups.begin(); i7!=(*i4).groups.end(); ++i7)
                            if(out)
                                fprintf(out,"%s%s%s\n","    push(@groups,",c_var_string::script_esc(script_type,*i7).c_str(),");");

                        ln="    $user=c_user->init(\"";
                        // empty nick
                        ln.append("\",\"");
                        // empty ident
                        ln.append("\",\"");
                        // empty host
                        ln.append("\",\"");
                        // empty fullname
                        ln.append("\",");
                        ln.append(c_var_string::script_esc(script_type,(*i4).name));
                        ln.append(",0,"); // $online
                        if((*i2).host_unknown)
                            ln.append("1,");
                        else
                            ln.append("0,");
                        if((*i4).host_bot)
                            ln.append("1,");
                        else
                            ln.append("0,");
                        ln.append("0,"); // irc_op
                        ln.append("'");
                        ln.append("0"); // mode
                        ln.append("',\\@host_masks,\\@fullname_masks,\\%meta,\\@groups);\n    undef(@host_masks); undef(@fullname_masks); undef(%meta); undef(@groups);\n");

                        if(out)
                            fprintf(out,"%s",ln.c_str());

                        ln="    $channel->add_user($user); undef($user);\n\n";
                        if(out)
                            fprintf(out,"%s",ln.c_str());
                    }
                }
                ln="    $channels{lc(\"";
                ln.append((*i9).name);
                ln.append("\")}=$channel; undef($channel);\n");
                if(out)
                    fprintf(out,"%s",ln.c_str());
            }
            ln="\n";
            if(out)
                fprintf(out,"%s",ln.c_str());
        }

        // filesystem
        {
            tmp=filesys_get_script(script_type);
            if(out)
                fprintf(out,"%s",tmp.c_str());
            tmp="";
        }

        line[0]=0;
        while(in && !feof(in)) {
            if(fgets(line,1024*10-1,in)==NULL)
                break;
            line[1024*10]=0;
            if(line[0] && line[strlen(line)-1]=='\r')
                line[strlen(line)-1]=0;
            if(line[0] && line[strlen(line)-1]=='\n')
                line[strlen(line)-1]=0;
            if(line[0] && line[strlen(line)-1]=='\r')
                line[strlen(line)-1]=0;
            if(line[0] && line[strlen(line)-1]=='\n')
                line[strlen(line)-1]=0;
            if(!strcmp(line,"# RESULT"))
                break;
            if(strlen(line))
                if(out)
                    fprintf(out,"%s",line);
            tmp="\n";
            if(out)
                fprintf(out,"%s",tmp.c_str());
            line[0]=0;
        }

        if(!noremove)
            tmp="    unlink(\"./";
        else
            tmp="#    unlink(\"./";
        tmp+=fn;
        tmp+="\");\n\n";
        if(out)
            fprintf(out,"%s",tmp.c_str());

        line[0]=0;
        while(in && !feof(in)) {
            if(fgets(line,1024*10-1,in)==NULL)
                break;
            line[1024*10]=0;
            if(line[0] && line[strlen(line)-1]=='\r')
                line[strlen(line)-1]=0;
            if(line[0] && line[strlen(line)-1]=='\n')
                line[strlen(line)-1]=0;
            if(line[0] && line[strlen(line)-1]=='\r')
                line[strlen(line)-1]=0;
            if(line[0] && line[strlen(line)-1]=='\n')
                line[strlen(line)-1]=0;
            if(strlen(line))
                if(out)
                    fprintf(out,"%s",line);
            tmp="\n";
            if(out)
                fprintf(out,"%s",tmp.c_str());
            line[0]=0;
        }

        if(out)
            fclose(out);
        if(in)
            fclose(in);

        c_tmp_file t;
        t.script=true;
        t.script_num=script_num;
        t.script_type=script_type;
        t.command_line=utils_convert_unix_path(conf_getvar("perl_processor"));

        t.command_args="\"." FILE_SLASH;
        t.command_args+=fn.c_str();
        t.command_args+="\"";
        if(!conf_getvar("perl_processor_args").empty()) {
            t.command_args+=" ";
            t.command_args+=conf_getvar("perl_processor_args");
        }
        t.command_dir=script_dir;

        t.has_been_run=false;
        t.start_up_time=0;
        t.noremove=noremove;
        t.filename=fn;
        tmp_files.push_back(t);

        MY_DELETE_A(line);
        line=NULL;
    }
}

/*!
    \brief Executes an procedure
    \author VooDooMan
    \version 2
    \date 2004, 2005-10-06
    \param proc Procedure to execute
    \param vars Variables for parameters of procedure
    \param local_id New local ID
    \param core_module_entrypoint Set this to 'true' if this is core module's entrypoint
    \return Returns zero for no error, or LOGIC_RESTART for bot restart request (e.g. "restart" command)
    \retval 0 No error
    \retval LOGIC_RESTART Bot restart request (e.g. "restart" command)
    \retval LOGIC_NO_SERVER Don't have any server in the list, so die.
*/
int c_module::exec(c_procedure& proc, c_var_array* vars, unsigned local_id, bool core_module_entrypoint)
{
    if(proc.name_only.empty())
        proc.parse_call_string();
    c_eval_iteration ste(this,vars,proc.name_only);
    ste.state=4000;
    ste.return_to_state=0;
    ste.proc=proc;
    ste.proc_iter=proc.commands.begin();
    ste.proc_cmd=*(ste.proc_iter);
    ste.goto_line=-1;
    ste.after_redirect=false;
    ste.redirect_checkpoint=false;
    ste.do_redirect=false;
    ste.r_checkpoint=false;
    ste.return_value=NULL;
    ste.local_id=local_id;
    ste.is_core_module_entrypoint=core_module_entrypoint;
    stack<c_eval_iteration> st;
    st.push(ste);
    bool first=true;
    while(st.size()>1 || first) {
        first=false;
        eval_loop(st);
    }
    return 0;
}

string c_module::conf_getvar(const string& key)
{
    string res;
    if(conf)
        res=conf->getvar(key);
    if(res.empty()) {
        c_module* core=get_module("core",true);
        if(core)
            res=core->conf->getvar(key);
    }
    return res;
}

/*!
    \brief Rehashes module
    \author VooDooMan
    \version 1
    \param error_on_line Returns erroneous line number
    \param error_reason Returns error reason/description
    \return Returns true for error
*/
bool c_module::rehash(int& error_on_line, string& error_reason)
{
    if(conf->rehash()) {
        logger->log_print1(E_LOG_BOT,LOG_SCOPE_MODULE,"","Failed to load module's \"conf.txt\" file");
        return true;
    }

    session_id_t old_session=var_keeper->get_current_session();
    unsigned old_local=var_keeper->get_current_local();
    if(0==old_local)
        var_keeper->free_local(0,false);

    session_=++last_session;
    sessions.push_back(session_);
    var_keeper->set_current_session(session_);

    c_var_array* base=NULL;
    unsigned l_id=var_keeper->register_local(base,NULL,__FILE__,__my_func__,__LINE__);
    var_keeper->set_current_local(l_id);
    
    error_on_line=1;
    error_reason="";

    vector<string> ignoring_procs;

    if(!parse_logic(error_on_line,error_reason,ignoring_procs)) {
        string error="logic.txt:";
        error+=tos(error_on_line);
        error+=": ";
        error+=error_reason;
        logger->log_print1(E_LOG_BOT,LOG_SCOPE_MODULE,"",error);

        var_keeper->free_local(l_id,true);
        var_keeper->set_current_local(old_local);
        var_keeper->set_current_session(old_session);

        if(!name.compare("core"))
            session_=SESSION_CORE_ENTRY_ID;

        return true;
    }

    {
        vector<string>::iterator i1;
        for(i1=ignoring_procs.begin(); i1!=ignoring_procs.end(); ++i1) {
            string msg="Ignoring procedure/function, it is locked: ";
            msg+=*i1;
            logger->log_print1(E_LOG_BOT,LOG_SCOPE_MODULE,"",msg);
        }
    }

    if(!name.compare("core")) {
        string val=conf->getvar("time_offset");
        if(val.empty())
            val="+0000";
        nice_time_set_time_offset(val);
    }

    var_keeper->free_local(local_id,true);
    local_id=l_id;
    var_keeper->set_current_local(old_local);
    var_keeper->set_current_session(old_session);

    if(!name.compare("core")) {
        var_keeper->set_core_module_working_session(session_);
        session_=SESSION_CORE_ENTRY_ID;
    }

    return false;
}

void c_module::get_replication_procs(vector<c_procedure>& procs, const string& botname, int repl_type)
{
    procs.clear();
    vector<c_procedure>::iterator i1;
    for(i1=r_procedures.begin(); i1!=r_procedures.end(); ++i1) {
        vector<c_replication>::iterator i2;
        for(i2=(*i1).replication.begin(); i2!=(*i1).replication.end(); ++i2) {
            if(!(*i2).partner.compare(botname) && (
                ( (repl_type==REPL_PUSH || repl_type==REPL_PUSHPULL) && ( (*i2).type==REPL_PUSH || (*i2).type==REPL_PUSHPULL) )
                ||
                ( (repl_type==REPL_PULL || repl_type==REPL_PUSHPULL) && ( (*i2).type==REPL_PULL || (*i2).type==REPL_PUSHPULL) )
            )) {
                procs.push_back(*i1);
            }
        }
    }
}

void c_module::get_replication_funcs(vector<c_function*>& funcs, const string& botname, int repl_type)
{
    funcs.clear();
    map<string,c_function*>::iterator i1;
    for(i1=r_functions.functions.begin(); i1!=r_functions.functions.end(); ++i1) {
        vector<c_replication>::iterator i2;
        for(i2=(*i1).second->replications.begin(); i2!=(*i1).second->replications.end(); ++i2) {
            if(!(*i2).partner.compare(botname) && (
                ( (repl_type==REPL_PUSH || repl_type==REPL_PUSHPULL) && ( (*i2).type==REPL_PUSH || (*i2).type==REPL_PUSHPULL) )
                ||
                ( (repl_type==REPL_PULL || repl_type==REPL_PUSHPULL) && ( (*i2).type==REPL_PULL || (*i2).type==REPL_PUSHPULL) )
            )) {
                funcs.push_back((*i1).second);
            }
        }
    }
}

/*!
    \brief Does a match against bound events
    \author VooDooMan
    \version 1
    \param dispatch Dispatch structure
    \param matches List of matches as returnde by prior call to 'get_matches()'
    \return Returns true if at least one bound event was handled
    \todo TODO: Optimization: When 'matches.size()==1' do not call copy-constuctior of 'dispatch'
*/
bool c_module::call_bound_dispatch_events(c_var_array* dispatch, const vector<pair<vector<c_binder>::iterator,c_binder_data> >& matches)
{
    if(matches.size()==0) {
        // optimization
        return false;
    }

    if(matches.size()==1) {
        // optimization
        return (*(matches.front().first)).call_bound_dispatch_event(dispatch,matches.front().second);
    }

    bool handled=false;

    vector<pair<vector<c_binder>::iterator,c_binder_data> >::const_iterator i1;
    for(i1=matches.begin(); i1!=matches.end(); ++i1) {
        c_var_array* dispatch_tmp=dispatch->copy();
        handled|=(*((*i1).first)).call_bound_dispatch_event(dispatch_tmp,(*i1).second);
        var_keeper->free_global(dispatch_tmp,true);
    }
    return handled;
}

/*!
    \brief Checks whether at least one bound event matches
    \author VooDooMan
    \version 1
    \param type Type string
    \param user User
    \param channel Channel
    \param line Full line
    \param before "Before" flag
    \param matches Returns list of matches (actually appends to the vector)
    \return Returns true if at least one bound event is matched
*/
bool c_module::get_matches(const string& type, const string& user, const string& channel, const string& line, bool before, vector<pair<vector<c_binder>::iterator,c_binder_data> >& matches)
{
    c_binder_data id;

    vector<c_binder>::iterator i1;
    for(i1=binds.begin(); i1!=binds.end(); ++i1)
        if((*i1).got_match(type,user,channel,line,before,id))
            matches.push_back(make_pair(i1,id));
    return !matches.empty();
}

void c_module::update_variables()
{
    var_keeper->update_variables(session_);
}

session_id_t c_module::get_session()
{
    return session_;
}

bool c_module::get_function(const string& func_name, bool& builtin, c_function*& func)
{
    if(r_functions.is_built_in_function(func_name))
        return builtin=true;

    builtin=false;

    map<string,c_function*>::iterator i1=r_functions.functions.find(func_name);
    if(r_functions.functions.end()==i1)
        return false;

    func=MY_NEW c_function(*(*i1).second);
    return true;
}

c_module::c_module()
{
    session_=++last_session;

    sessions.push_back(session_);

    session_id_t old_session=var_keeper->get_current_session();

    var_keeper->set_current_session(session_);

    global=var_keeper->register_mod_global(NULL,__FILE__,__my_func__,__LINE__);

    var_keeper->set_current_session(old_session);

    loaded=false;
    logger=NULL;
    conf=NULL;
    files.clear();

    local_id=0;
}

string c_module::get_mod_name()
{
    return name;
}

bool c_module::is_loaded()
{
    return loaded;
}

void c_module::dump_funcs_to_file(FILE* f)
{
    r_functions.dump_to_file(f);
}

c_conf* c_module::get_conf()
{
    return conf;
}

c_logger* c_module::get_logger()
{
    return logger;
}

c_var_base* c_module::open_file(const string& name, const string& mode, const string& debug)
{
    if(name.find("\\")!=string::npos || name.find("/")!=string::npos ||
        !name.compare(".") || !name.compare("..") ||
        name.find("?")!=string::npos || name.find("*")!=string::npos
        ) {
            if(logger) {
                string log="Illegal file name to open: \""+name+"\"; => "+debug;
                logger->log_print1(E_LOG_BOT,LOG_SCOPE_DEFAULT,"",log);
            }
            return MY_NEW c_var_bool(false);
        }
    string path=utils_convert_unix_path(dir_prefix+"/data/"+name);
    FILE* f=fopen(path.c_str(),mode.c_str());
    if(!f) {
        string log="Failed to open file: \""+path+"\"; => "+debug;
        logger->log_print1(E_LOG_BOT,LOG_SCOPE_DEFAULT,"",log);
        return MY_NEW c_var_bool(false);
    }
    if(MODULE_MAXIMUM_OPEN_FILES<=files.size()) {
        string log="Maximum number of open file descriptors reached; failed to open file: \""+path+"\"; This can be tuned in 'params.h' preprocessor define 'MODULE_MAXIMUM_OPEN_FILES'; => "+debug;
        logger->log_print1(E_LOG_BOT,LOG_SCOPE_DEFAULT,"",log);
        return MY_NEW c_var_bool(false);
    }
    int fd;
    for(;;) {
        fd=rand() % 65536;
        if(fd<0)
            fd=-fd;
        fd*=65536;
        fd+=rand() % 65536;
        if(fd<=0) // we don't even want "0" (zero), because zero is like FALSE
            continue;
        map<int,FILE*>::iterator i1=files.find(fd);
        if(files.end()==i1)
            break;
    }
    files.insert(make_pair(fd,f));

    return MY_NEW c_var_int(fd);
}

FILE* c_module::get_file(int desc)
{
    map<int,FILE*>::iterator i=files.find(desc);
    if(files.end()==i)
        return NULL;
    return (*i).second;
}

c_var_bool* c_module::close_file(int desc)
{
    map<int,FILE*>::iterator i=files.find(desc);
    if(files.end()==i)
        return MY_NEW c_var_bool(false);
    fclose((*i).second);
    (*i).second=NULL;
    files.erase(i);
    return MY_NEW c_var_bool(true);
}

bool c_module::is_procedure(string proc_name)
{
    if(proc_name.find("(")!=string::npos)
        proc_name.erase(proc_name.find("("),proc_name.length()-proc_name.find("("));

    vector<c_procedure>::iterator i1;
    for(i1=r_procedures.begin(); i1!=r_procedures.end(); ++i1) {
        string no=(*i1).name;
        if(no.find("(")!=string::npos)
            no.erase(no.find("("),no.length()-no.find("("));
        if(!no.compare(proc_name))
            return true;
    }
    return false;
}

string c_module::get_str_error(c_module::e_rc rc)
{
    switch(rc) {
        case RC_OK:
            return (string)"OK";
        case RC_E_CONF_ERROR:
            return (string)"Error in 'conf.txt'";
        case RC_E_LOGIC_ERROR:
            return (string)"Error in 'logic.txt'";
        case RC_E_DUPLICATE_MODULE:
            return (string)"Module with same name already registered";
        case RC_W_LOADED:
            return (string)"Warning: Already loaded";
        case RC_E_INV_MODNAME:
            return (string)"Invalid module name";
        case RC_E_SO_ERROR:
            return (string)"Loading dynamic library failed";
    }
    return (string)"(other error)";
}

const string& c_module::get_dir_prefix() const
{
    return dir_prefix;
}

