/***************************************************************************
                          modules.cpp  -  description
                             -------------------
    begin                : Fri Oct 21 2005
    copyright            : (C) 2005-2007 by Marian VooDooMan Meravy
    email                : ghostvoodooman -a.T= users (dot) sourceforge.net
 ***************************************************************************/

/* $Id: modules.cpp,v 1.56 2008/08/23 15:10:41 ghostvoodooman Exp $ */

/***************************************************************************

VooDoo cIRCle - an IRC (ro)bot
Copyright (C) 2005-2008 by Marian VooDooMan Meravy
  (ghostvoodooman -a.T= users (dot) sourceforge.net)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************/

/*!
    \file
    \brief Provides messaging to/from/between modules and handles modules
    \version $Id: modules.cpp,v 1.56 2008/08/23 15:10:41 ghostvoodooman Exp $
*/

#if defined(HAVE_CONFIG_H)
#   include <config.h>
#endif

#include <string>
#include <vector>
#include <stack>
#include <list>
#include <map>
#include <typeinfo>

#include <stdarg.h>
#include <stdlib.h>

using namespace std;

#include "utils/memleak.h"
#include "utils/assert.h"
#include "utils.h"

#include "params.h"
#include "compat.h"
#include "utils.h"
#include "variable.h"
#include "logic.h"
#include "modules.h"
#include "filesys.h"
#include "shared.h"
#include "irc.h"
#include "match.h"
#include "proc.h"
#include "stats.h"
#include "botnet.h"
#include "eval.h"
#include "binmod.h"
#include "function.h"
#include "variable.h"
#include "shared.h"
#include "exec.h"
#include "irc/c_irc_server.h"
#include "irc/c_online_channel.h"
#include "runtime/c_delay_command.h"
#include "var/c_var_base.h"
#include "var/c_var_array.h"
#include "var/c_var_bool.h"
#include "var/c_var_int.h"
#include "var/c_var_ref.h"
#include "var/c_var_string.h"
#include "var/c_var_unset.h"
#include "conf.h"
#include "log.h"
#include "microcode/c_micro_symbol_variable_table.h"
#include "microcode/c_micro_error_func_internal_1.h"
#include "microcode/c_micro_error_func_internal_3.h"
#include "internal_state.h"

#include "utils/precomp.h"

vector<c_module*> modules;
vector<c_delay_command> delay_commands;

c_conf* get_core_conf();

c_module* core_mod=NULL;

session_id_t last_session=SESSION_FIRST_ID-1;   //!< Last session ID, next will be this +1
vector<session_id_t> sessions;                  //!< List of all used session

void update_groups()
{
    r_all_groups.clear();

    vector<c_module*>::iterator i1;
    for(i1=modules.begin(); i1!=modules.end(); ++i1) {
        string group_prefix;
        if((*i1)->get_mod_name().compare("core")) {
            // != "core"
            group_prefix=(*i1)->get_mod_name();
            group_prefix+="::";
        }
        {
            vector<string>::iterator i;
            for(i=(*i1)->all_groups.begin(); i!=(*i1)->all_groups.end(); ++i) {
                string group=group_prefix;
                group+=(*i);
                r_all_groups.push_back(group);
            }
        }
    }
}

/*!
    \brief Unloads module
    \author VooDooMan
*/
void unload_module(const string& mod_name)
{
    vector<c_module*>::iterator i1;
    for(i1=modules.begin(); i1!=modules.end(); ++i1) {
        if(!cmp_strings_case_insensitive(mod_name,(*i1)->get_mod_name())) {
            MY_DELETE(*i1);
            break;
        }
    }
}

/*!
    \brief Dispatches event
    \author VooDooMan
    \param type Type
    \param user User
    \param channel Channel
    \param line Line
    \param before "BEFORE" flag
    \param dummy Variable arguments start point
*/
bool mod_dispatch(const string& type, const string& user, const string& channel, const string& line, bool before, int dummy, ...)
{
    vector<pair<vector<c_binder>::iterator,c_binder_data> > matches;
    vector<c_module*>::iterator i1;
    for(i1=modules.begin(); i1!=modules.end(); ++i1)
        (*i1)->get_matches(type,user,channel,line,before,matches);
    if(matches.empty())
        return false;

    c_global_keeper::c_session_keeper sess_keeper;
    var_keeper->set_current_session(SESSION_RESERVED_ID);
    c_var_array* base=NULL;
    unsigned l_id=var_keeper->register_local(base,NULL,__FILE__,__my_func__,__LINE__);
    (void)var_keeper->set_current_local(l_id);

    c_var_array* dispatch=MY_NEW c_var_array();
    {
        c_index_base* index=MY_NEW c_index_str("type");
        c_var_string* value=MY_NEW c_var_string(type);
        dispatch->set(NULL,index,value,false,"...");
    }
    {
        c_index_base* index=MY_NEW c_index_str("user");
        c_var_string* value=MY_NEW c_var_string(user);
        dispatch->set(NULL,index,value,false,"...");
    }
    {
        c_index_base* index=MY_NEW c_index_str("channel");
        c_var_string* value=MY_NEW c_var_string(channel);
        dispatch->set(NULL,index,value,false,"...");
    }
    {
        c_index_base* index=MY_NEW c_index_str("line");
        c_var_string* value=MY_NEW c_var_string(line);
        dispatch->set(NULL,index,value,false,"...");
    }
    {
        c_index_base* index=MY_NEW c_index_str("before");
        c_var_bool* value=MY_NEW c_var_bool(before);
        dispatch->set(NULL,index,value,false,"...");
    }
    c_var_array* meta=MY_NEW c_var_array();

    va_list marker;
    va_start(marker,dummy);
    {
        for(;;) {
            char* f=va_arg(marker,char*);
            if(!f)
                break;
            char* s=va_arg(marker,char*);
            if(!s)
                break;
            string first=f;
            string second=s;
            if(first.empty())
                continue;

            if(!first.compare("channel") && second.empty()) {
                // skip
            } else {
                c_index_base* index=MY_NEW c_index_str(first);
                c_var_string* value=MY_NEW c_var_string(second);
                meta->set(NULL,index,value,false,"...");
            }
        }
    }
    {
        for(;;) {
            char* f=va_arg(marker,char*);
            if(!f)
                break;
            char* s=va_arg(marker,char*);
            if(!s)
                break;
            string first=f;
            string second=s;
            if(first.empty())
                continue;

            c_index_base* index=MY_NEW c_index_str(first);
            c_var_string* value=MY_NEW c_var_string(second);
            dispatch->set(NULL,index,value,false,"...");
        }
    }
    va_end(marker);

    {
        c_index_base* index=MY_NEW c_index_str("meta");
        dispatch->set(NULL,index,meta,false,"...");
    }

    bool handled=c_module::call_bound_dispatch_events(dispatch,matches);

    var_keeper->free_global(dispatch,true);

    return handled;
}

/*!
    \brief Dispatches event
    \author VooDooMan
    \param type Type
    \param user User
    \param channel Channel
    \param line Line
    \param before "BEFORE" flag
    \param meta Meta array
    \param dummy Variable arguments start point
*/
bool mod_dispatch2(const string& type, const string& user, const string& channel, const string& line, bool before, c_var_array* meta, int dummy, ...)
{
    vector<pair<vector<c_binder>::iterator,c_binder_data> > matches;
    vector<c_module*>::iterator i1;
    for(i1=modules.begin(); i1!=modules.end(); ++i1)
        (*i1)->get_matches(type,user,channel,line,before,matches);
    if(matches.empty())
        return false;

    c_global_keeper::c_session_keeper sess_keeper;
    var_keeper->set_current_session(SESSION_RESERVED_ID);
    c_var_array* base=NULL;
    unsigned l_id=var_keeper->register_local(base,NULL,__FILE__,__my_func__,__LINE__);
    (void)var_keeper->set_current_local(l_id);

    c_var_array* dispatch=MY_NEW c_var_array();
    {
        c_index_base* index=MY_NEW c_index_str("type");
        c_var_string* value=MY_NEW c_var_string(type);
        dispatch->set(NULL,index,value,false,"...");
    }
    {
        c_index_base* index=MY_NEW c_index_str("user");
        c_var_string* value=MY_NEW c_var_string(user);
        dispatch->set(NULL,index,value,false,"...");
    }
    {
        c_index_base* index=MY_NEW c_index_str("channel");
        c_var_string* value=MY_NEW c_var_string(channel);
        dispatch->set(NULL,index,value,false,"...");
    }
    {
        c_index_base* index=MY_NEW c_index_str("line");
        c_var_string* value=MY_NEW c_var_string(line);
        dispatch->set(NULL,index,value,false,"...");
    }
    {
        c_index_base* index=MY_NEW c_index_str("before");
        c_var_bool* value=MY_NEW c_var_bool(before);
        dispatch->set(NULL,index,value,false,"...");
    }
    c_var_array* meta_=meta->copy();

    {
        c_index_base* index=MY_NEW c_index_str("meta");
        dispatch->set(NULL,index,meta_,false,"...");
    }

    va_list marker;
    va_start(marker,dummy);
    {
        for(;;) {
            char* f=va_arg(marker,char*);
            if(!f)
                break;
            char* s=va_arg(marker,char*);
            if(!s)
                break;
            string first=f;
            string second=s;
            if(first.empty())
                continue;

            c_index_base* index=MY_NEW c_index_str(first);
            c_var_string* value=MY_NEW c_var_string(second);
            dispatch->set(NULL,index,value,false,"...");
        }
    }
    va_end(marker);

    bool handled=c_module::call_bound_dispatch_events(dispatch,matches);

    var_keeper->free_global(dispatch,true);

    return handled;
}

struct c_exception { };

/*!
    \brief Modules main loop (should be called in a loop)
    \author VooDooMan
    \date 2005
    \version 1
    \param shutdown Set to true if we are shutting down (to free internal memory)
*/
void mod_loop(bool shutdown)
{
    modules_maintenance();

    if(shutdown) {
erase_again:
        vector<c_module*>::iterator mod=modules.begin();
        if(mod!=modules.end()) {
            MY_DELETE(*mod);
            goto erase_again;
        }
        if(super_global)
            MY_DELETE(super_global);
        super_global=NULL;
        return;
    }

    mod_loop2();

again:

    try {
        vector<c_delay_command>::iterator i1;
        for(i1=delay_commands.begin(); i1!=delay_commands.end(); ++i1) {

            var_keeper->do_garbage(c_var_keeper::GARBAGE_MAX_TIME_INFINITE);

            switch((*i1).type) {
                case c_delay_command::DELAY_LOAD:
                    {
                        c_global_keeper::c_session_keeper sess_keeper;

                        string mod_name=(*i1).mod_name;
                        delay_commands.erase(i1);
                        c_module* mod=MY_NEW c_module();
                        c_module::e_rc rc=mod->load(mod_name,"");
                        if(c_module::RC_OK==rc)
                            mod->call_startup();
                        else {
                            if(c_module::RC_OK!=rc)
                                MY_DELETE(mod);
                        }
                        if(c_module::RC_OK!=rc) {
                            string log="Delayed load of module '";
                            log+=mod_name;
                            log+="' ended-up with: ";
                            log+=c_module::get_str_error(rc);
                            log_bot(log);
                            log_debug(log);
                        }
                        throw c_exception();
                    }
                    break;
                case c_delay_command::DELAY_UNLOAD:
                    {
                        vector<c_module*>::iterator i2;
                        for(i2=modules.begin(); i2!=modules.end(); ++i2) {
                            if(!cmp_strings_case_insensitive((*i1).mod_name,(*i2)->get_mod_name())) {
                                MY_DELETE(*i2);
                                break;
                            }
                        }
                        delay_commands.erase(i1);
                        throw c_exception();
                    }
                    break;
                case c_delay_command::DELAY_BIND:
                case c_delay_command::DELAY_BIND_EX:
                    {
                        vector<c_module*>::iterator i2;
                        for(i2=modules.begin(); i2!=modules.end(); ++i2) {
                            if(!cmp_strings_case_insensitive((*i1).mod_name,(*i2)->get_mod_name())) {
                                if((*i2)->is_loaded()) {
                                    vector<string> error_log;
                                    c_binder b(c_delay_command::DELAY_BIND_EX==(*i1).type,(*i1).type_mask,(*i1).user_mask,(*i1).channel_mask,(*i1).line_mask,(*i1).before,(*i1).func,*i2,error_log,(*i1).debug);
                                    if(error_log.empty())
                                        (*i2)->binds.push_back(b);
                                    else if((*i2)->get_logger()) {
                                        vector<string>::iterator i3;
                                        for(i3=error_log.begin(); i3!=error_log.end(); ++i3)
                                            (*i2)->get_logger()->log_print1(E_LOG_DEBUG,LOG_SCOPE_MODULE,"",*i3);
                                    }
                                }
                                break;
                            }
                        }
                        delay_commands.erase(i1);
                        throw c_exception();
                    }
                    break;
                case c_delay_command::DELAY_UNBIND:
                case c_delay_command::DELAY_UNBIND_EX:
                    {
                        vector<c_module*>::iterator i2;
                        for(i2=modules.begin(); i2!=modules.end(); ++i2) {
                            if(!cmp_strings_case_insensitive((*i1).mod_name,(*i2)->get_mod_name())) {
                                if((*i2)->is_loaded()) {
                                    vector<string> error_log;
                                    c_binder b(c_delay_command::DELAY_UNBIND_EX==(*i1).type,(*i1).type_mask,(*i1).user_mask,(*i1).channel_mask,(*i1).line_mask,(*i1).before,(*i1).func,*i2,error_log,(*i1).debug);
                                    if(error_log.empty()) {
                                        size_t unbound=0;
                                unbind_again:
                                        vector<c_binder>::iterator i3;
                                        for(i3=(*i2)->binds.begin(); i3!=(*i2)->binds.end(); ++i3) {
                                            if((*i3)==b) {
                                                (*i2)->binds.erase(i3);
                                                ++unbound;
                                                goto unbind_again;
                                            }
                                        }
                                        if((*i2)->get_logger()) {
                                            string log;
                                            if(0==unbound)
                                                log="WARNING: None buond event was unbound";
                                            if(1<unbound)
                                                log=(string)"WARNING: "+tos(unbound)+" buond event were unbound";
                                            if(!log.empty()) {
                                                log+="; Additional info: ";
                                                log+=(*i1).debug;
                                                (*i2)->get_logger()->log_print1(E_LOG_DEBUG,LOG_SCOPE_MODULE,"",log);
                                            }
                                        }
                                    } else if((*i2)->get_logger()) {
                                        vector<string>::iterator i3;
                                        for(i3=error_log.begin(); i3!=error_log.end(); ++i3)
                                            (*i2)->get_logger()->log_print1(E_LOG_DEBUG,LOG_SCOPE_MODULE,"",*i3);
                                    }
                                }
                                break;
                            }
                        }
                        delay_commands.erase(i1);
                        throw c_exception();
                    }
                    break;
                case c_delay_command::DELAY_LOAD_BIN:
                    {
                        string mod_name=(*i1).mod_name;
                        string so_name=(*i1).so_name;
                        delay_commands.erase(i1);
                        c_binmodule* mod=MY_NEW c_binmodule();
                        c_module::e_rc rc=mod->load(mod_name,so_name);
                        if(c_module::RC_OK==rc)
                            mod->call_startup();
                        else {
                            if(c_module::RC_OK!=rc)
                                MY_DELETE(mod);
                        }
                        if(c_module::RC_OK!=rc) {
                            string log="Delayed load of binary module '";
                            log+=mod_name;
                            log+="' with binary '";
                            log+=so_name;
                            log+="' ended-up with: ";
                            log+=c_module::get_str_error(rc);
                            log_bot(log);
                            log_debug(log);
                        }
                        throw c_exception();
                    }
                    break;
            }
        }

        var_keeper->do_garbage(c_var_keeper::GARBAGE_MAX_TIME_INFINITE);

    } catch(c_exception& UNUSEDARG(e)) {
        irc_loop_heavy_time_consuming_operation("Still performing delayed commands...");
        my_sleep(-1);
        goto again;
    }
}

/*!
    \brief Modules main loop # 2 (should be called in a loop)
    \author VooDooMan
    \date 2006-01-03
    \version 1
*/
void mod_loop2()
{
    static my_time_t last_logger_loop=0;
    if(last_logger_loop+10<my_time(NULL)) {
        vector<c_module*>::iterator mod;
        for(mod=modules.begin(); mod!=modules.end(); ++mod) {
            c_logger* l;
            if((*mod)->is_loaded()) {
                l=(*mod)->get_logger();
                if(l)
                    l->loop();
            }
        }
    }
}

/*!
    \brief Initializes modules and session
    \author VooDooMan
    \version 1
    \date 2006-05-15
    \warning Call only once at initialization, before any other sessions (modules) are allocated their ID
*/
void modules_init()
{
    sessions.push_back(SESSION_RESERVED_ID);
}

/*!
    \brief Periodically cleans-up dropped variables, removes them from memory
    \author VooDooMan
    \version 1
    \date 2006-05-15
    \return Returns number of objects purged
    \param force Forces purge now
*/
unsigned modules_maintenance(bool force)
{
    static my_time_t last=0;
    if(!force && last+60>my_time(NULL))
        return 0;

    if(!force && var_keeper->get_current_session()!=SESSION_RESERVED_ID)
        return 0;

    unsigned cnt=0;

    vector<session_id_t>::iterator i1;
    for(i1=sessions.begin(); i1!=sessions.end(); ++i1) {
        cnt+=var_keeper->purge(*i1);
    }
    last=my_time(NULL);

    return cnt;
}

/*!
    \brief Cleans-up modules' sessions
    \author VooDooMan
    \version 1
    \date 2006-05-15
*/
void modules_done()
{
    sessions.clear();
}

c_module* get_module(const string& name, bool must_be_loaded)
{
    vector<c_module*>::iterator i1;
    for(i1=modules.begin(); i1!=modules.end(); ++i1) {
        if( (must_be_loaded && (*i1)->is_loaded()) || !must_be_loaded)
            if(!name.compare((*i1)->get_mod_name()))
                return *i1;
    }
    return NULL;
}

c_conf* get_core_conf()
{
    if(!core_mod)
        core_mod=get_module("core",false);

    c_conf* core_conf=NULL;
    if(core_mod)
        core_conf=core_mod->get_conf();

    return core_conf;
}

c_logger* get_core_logger()
{
    if(!core_mod)
        core_mod=get_module("core",false);

    c_logger* core_logger=NULL;
    if(core_mod)
        core_logger=core_mod->get_logger();

    return core_logger;
}

