/***************************************************************************
                          c_tracked_msg_to_server.cpp  -  description
                             -------------------
    begin                : Sun Dec 02 2007
    copyright            : (C) 2007 by Marian VooDooMan Meravy
    email                : ghostvoodooman -a.T= users (dot) sourceforge.net
 ***************************************************************************/

/* $Id: c_tracked_msg_to_server.cpp,v 1.13 2008/03/25 12:04:28 ghostvoodooman Exp $ */

/***************************************************************************

VooDoo cIRCle - an IRC (ro)bot
Copyright (C) 2004-2008 by Marian VooDooMan Meravy
  (ghostvoodooman -a.T= users (dot) sourceforge.net)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************/

/*!
    \file
    \brief Class 'c_tracked_msg_to_server'
    \version $Id: c_tracked_msg_to_server.cpp,v 1.13 2008/03/25 12:04:28 ghostvoodooman Exp $
*/

#if defined(HAVE_CONFIG_H)
#   include <config.h>
#endif

#include "../utils/memleak.h"
#include "../utils/assert.h"
#include "../utils.h"
#include "../params.h"

#include "c_tracked_msg_to_server.h"

#include "../irc.h" // for c_005
#include "../casemap.h"

#include "../utils/precomp.h"

#if defined(_MSC_VER) || defined(__BORLANDC__)
#   pragma hdrstop
#endif

/*!
    \brief Adds first piece of message to the vector (not-adjustable, as it is always IRC protocol command)
    \author VooDooMan
    \version 1
    \date 2007-12-02
    \param first_piece The first piece of message (IRC command, not-adjustable)
*/
c_tracked_msg_to_server::c_tracked_msg_to_server(const string& first_piece)
{
    m_pieces.push_back(make_pair(PIECE_IGNORE,first_piece));
}

/*!
    \brief Dummy constructor to allow uninitialized construction
    \author VooDooMan
    \version 1
    \date 2007-12-02
*/
c_tracked_msg_to_server::c_tracked_msg_to_server()
{
}

/*!
    \brief Adds next piece of command (not-adjustable)
    \author VooDooMan
    \version 1
    \date 2007-12-02
    \param piece Appends next piece of message
    \return Returns itself
*/
c_tracked_msg_to_server& c_tracked_msg_to_server::add_piece(const string& piece)
{
    if(!piece.empty())
        m_pieces.push_back(make_pair(PIECE_IGNORE,piece));
    return *this;
}

/*!
    \brief Adds next piece of command (adjustable: should be nick)
    \author VooDooMan
    \version 1
    \date 2007-12-02
    \param nick Nick of user to be adjusted
    \return Returns itself
*/
c_tracked_msg_to_server& c_tracked_msg_to_server::add_nick(const string& nick)
{
    m_pieces.push_back(make_pair(PIECE_NICK,nick));
    return *this;
}

/*!
    \brief Adds next piece of command (adjustable: should be channel)
    \author VooDooMan
    \version 1
    \date 2007-12-03
    \param channel Channel to be adjusted
    \return Returns itself
*/
c_tracked_msg_to_server& c_tracked_msg_to_server::add_channel(const string& channel)
{
    m_pieces.push_back(make_pair(PIECE_CHANNEL,channel));
    return *this;
}

/*!
    \brief Called if someone has changed nick, and, possibly adjust it in the vector if particular nick is related to this message
    \author VooDooMan
    \version 1
    \date 2007-12-02
    \param old_nick Old nick
    \param new_nick New nick
    \return Returns 'true' if nick change in queue has occured
*/
bool c_tracked_msg_to_server::process_nick_change(const string& old_nick, const string new_nick)
{
    bool res=false;

    vector<pair<e_kind,string> >::iterator i1;
    for(i1=m_pieces.begin(); i1!=m_pieces.end(); ++i1) {
        if(PIECE_NICK==(*i1).first && !strtoupper((*i1).second,irc_casemapping).compare(strtoupper(old_nick,irc_casemapping))) {
            (*i1).second=new_nick;
            res=true;
        }
    }

    return res;
}

/*!
    \brief Called if someone has part/quit/got kicked from channel, and, possibly removing it in the vector if particular nick is related to this message
    \author VooDooMan
    \version 1
    \date 2007-12-12
    \param channel Channel name
    \param part_nick Nick that has part/quit/got kicked from channel
    \return Returns 'true' if nick removal in queue has occured
*/
bool c_tracked_msg_to_server::process_nick_change_chan(const string& channel, const string& part_nick)
{
    if(m_pieces.empty())
        return false;

    bool res=false;
    vector<pair<e_kind,string> >::iterator i1;

    // 3: MODE #channel +modes arg1 arg2 ...
    if(m_pieces.size()>3 && !m_pieces[0].second.compare("MODE") && !strtoupper(m_pieces[1].second,irc_casemapping).compare(strtoupper(channel,irc_casemapping))) {
        string m=m_pieces[2].second;
        // break it up
        vector<string> modes;
        string tmp;
        for(string::size_type i2=0; i2<m.size(); ++i2) {
            if('+'==m[i2] || '-'==m[i2]) {
                tmp=m[i2];
                continue;
            }
            tmp+=(string)""+m[i2];
            modes.push_back(tmp);
            // preserve +/- for next iteration
            if(tmp.size()!=0) {
                if('+'==tmp[0])
                    tmp="+";
                else if('-'==tmp[0])
                    tmp="-";
                else
                    tmp=""; // huh?!
            } else
                tmp=""; // huh?!
        }

    again:
        vector<string>::iterator i3=modes.begin();
        bool valid=modes.end()!=i3;
        for(i1=m_pieces.begin()+1+1+1; valid && i1!=m_pieces.end(); ++i1) { // skip "MODE #channel +modes"
            if(PIECE_NICK==(*i1).first && !strtoupper((*i1).second,irc_casemapping).compare(strtoupper(part_nick,irc_casemapping))) {
                modes.erase(i3);
                m_pieces.erase(i1);
                res=true;
                goto again;
            }
            for(;;) {
                if(valid) {
                    ++i3;
                    valid=modes.end()!=i3;
                }
                if(!valid)
                    break;

                tmp=*i3;

                if(tmp.size()==2) {
                    // check if mode has an argument
                    if(irc_isupport.chanmodes_a.find(tmp[1])!=string::npos) {
                        // class A: a mask, skip this
                        ++i1;
                        if(m_pieces.end()==i1)
                            break;
                        continue;
                    }
                    if(irc_isupport.chanmodes_b.find(tmp[1])!=string::npos) {
                        // class B: always has parameter, okay to process
                        break;
                    }
                    if(irc_isupport.chanmodes_c.find(tmp[1])!=string::npos) {
                        // class C: has parameter if set ('+')
                        if('-'==tmp[1]) { // skip
                            ++i1;
                            if(m_pieces.end()==i1)
                                break;
                            continue;
                        }
                        break;
                    }
                    if(irc_isupport.chanmodes_d.find(tmp[1])!=string::npos) {
                        // class D: never has parameter, skip this
                        continue;
                    }
                }
            }
            if(m_pieces.end()==i1)
                break;
        }

        if(res) {
            // rebuild
            if(!modes.empty()) {
                tmp="";
                for(i3=modes.begin(); i3!=modes.end(); ++i3)
                    tmp+=*i3;
                m_pieces[2].second=tmp;
            } else
                m_pieces.clear();
        }
    }

    // 3: KICK #channel nick [:optional reason]
    if(m_pieces.size()>=3 && !m_pieces[0].second.compare("KICK") && !strtoupper(m_pieces[1].second,irc_casemapping).compare(strtoupper(channel,irc_casemapping)) && !strtoupper(m_pieces[2].second,irc_casemapping).compare(strtoupper(part_nick,irc_casemapping))) {
        m_pieces.clear();
        res=true;
    }

    return res;
}

/*!
    \brief Returns message including EOL sequence
    \author VooDooMan
    \version 1
    \date 2007-12-02
*/
string c_tracked_msg_to_server::get_msg() const
{
    // MY_ASSERT(!m_pieces.empty());
    // ^^^ we cannot do this, because we got uninitialized instance for flood
    //     protection purposes (bot waits after some commands sent to IRC for
    //     response before issuing next one in the queue)

    if(m_pieces.empty())
        return (string)"";

    ostringstream s;
    vector<pair<e_kind,string> >::const_iterator i1;
    bool first=true;
    for(i1=m_pieces.begin(); i1!=m_pieces.end(); ++i1) {
        if(!first)
            s << " ";
        first=false;
        s << (*i1).second.c_str();
    }
    s << "\r\n";

    return s.str();
}

/*!
    \brief Checks whether argument is adjustable entry (nick)
    \author VooDooMan
    \version 1
    \date 2007-12-02
*/
bool c_tracked_msg_to_server::is_nick(const string& nick) const
{
    vector<pair<e_kind,string> >::const_iterator i1;
    for(i1=m_pieces.begin(); i1!=m_pieces.end(); ++i1)
        if(PIECE_NICK==(*i1).first && !strtoupper((*i1).second,irc_casemapping).compare(strtoupper(nick,irc_casemapping)))
            return true;

    return false;
}

/*!
    \brief Adds multiple pieces at once
    \author VooDooMan
    \version 1
    \date 2007-12-02
    \param pieces Multiple pieces
    \return Returns itself
*/
c_tracked_msg_to_server& c_tracked_msg_to_server::add_multi(const vector<pair<e_kind,string> >& pieces)
{
    m_pieces.insert(m_pieces.end(),pieces.begin(),pieces.end());
    return *this;
}

/*!
    \brief Check whether we have message
    \author VooDooMan
    \version 1
    \date 2007-12-02
*/
bool c_tracked_msg_to_server::empty()
{
    return m_pieces.empty();
}

