/***************************************************************************
                          params.h  -  description
                             -------------------
    begin                : Thu Dec 9 2004
    copyright            : (C) 2004-2007 by Marian VooDooMan Meravy
    email                : ghostvoodooman -a.T= users (dot) sourceforge.net
 ***************************************************************************/

/* $Id: params.h,v 1.113 2008/11/24 12:09:23 ghostvoodooman Exp $ */

/***************************************************************************

VooDoo cIRCle - an IRC (ro)bot
Copyright (C) 2004-2008 by Marian VooDooMan Meravy
  (ghostvoodooman -a.T= users (dot) sourceforge.net)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************/

/*!
    \file
    \brief Parameters for compilation
    \version $Id: params.h,v 1.113 2008/11/24 12:09:23 ghostvoodooman Exp $
*/

#ifndef PARAMS_H__
#define PARAMS_H__

#if defined(HAVE_CONFIG_H)
#   include <config.h>
#endif

#if !defined(WANT_DEBUG) && !defined(NDEBUG)
    // default behaviour to support better bug reports follows on the next line:
#   define WANT_DEBUG      1                // change to '0' instead of '1': don't touch this if you want to take over '_DEBUG' and 'NDEBUG' defines (supplied as environment variables to './configure', and you really want to get debugging support enabled) (really a _little_ overhead in CPU/RAM utilization, but your bug reports will be _much_ more verbose, which helps to figure out where the bug is, thus you will help in improving this software)
#endif

#if !defined(WANT_IPv6)
#   define WANT_IPv6       1                // Set to '0' instead of '1', to get rid of IPv6 support for your bot ('1' is harmless even it you don't have IPv6 connectivity; *recommended* even if your line is only IPv4-capable */
#endif

#if defined(_MSC_VER)
#   if defined(_CRT_SECURE_NO_WARNINGS)
#       undef _CRT_SECURE_NO_WARNINGS
#   endif
#   define _CRT_SECURE_NO_WARNINGS  1   /* MS Visual Stodio 8 ugly warnings against POSIX standard vs. ISO C++ standard functions */
#endif

#include "params_prologue.h"

/*!
    \def BOT_VERSION
    \brief Bot version
    \warning Cannot contain SPACE (0x20) character!
*/
#define BOT_VERSION "1.1.33"

/*!
    \def BOT_VERSION_COMMENT
    \brief Comment to version (will be shown in reply to ".broadcastping" command)
    \warning Cannot contain SPACE (0x20) character!
    \warning Will be enclosed in () brackets as defined in file "lang/lang01.txt"
*/
#define BOT_VERSION_COMMENT "licensed_under_GNU/GPLv2"

/*!
    \def IRC_SOCKET_TIMEOUT
    \brief Timeout of IRC connection socket (seconds)
*/
#define IRC_SOCKET_TIMEOUT          (60*3)

/*!
    \def IRC_KEEPALIVE_PERIOD
    \brief Ineterval of sending PING to IRC server (seconds)
    \warning Must not be bigger than IRC_SOCKET_TIMEOUT
*/
#define IRC_KEEPALIVE_PERIOD        (30)

/*!
    \def DNS_CACHE_VALIDITY
    \brief Validity of DNS chache entry (seconds)
*/
#define DNS_CACHE_VALIDITY          (60*60*2)

/*!
    \def DNS_MAX_RETRIES_FATAL
    \brief Maximum tries total (this value minus 1) if DNS lookup failed with error (e.g. no entry)
*/
#define DNS_MAX_RETRIES_FATAL       (2)

/*!
    \def DNS_MAX_RETRIES_TMP
    \brief Maximum tries total if DNS lookup failed with temporary error (e.g. try again)
*/
#define DNS_MAX_RETRIES_TMP         (9)

/*!
    \def NEW_INSTANCE_TIMEOUT_AFTER_LAST_JOIN
    \brief Number of seconds as timeout after .upgrade command, and after last new instance's join to kill myself
*/
#define NEW_INSTANCE_TIMEOUT_AFTER_LAST_JOIN    (8*60)

#define DH_BITS                                 (1024)  // Do NOT touch unless you _exactly_ know what you are doing
#define RSA_BITS                                (512)   // Do NOT touch unless you _exactly_ know what you are doing
#define DH_RSA_REGENERATION_NUM_SECS            (60*60*12)
#define DH_RSA_REGENERATION_NUM_TRANSACTIONS    (500)

/* BEGIN: Do NOT touch following unless you _exactly_ know what you are doing */
#define BOTNET_MSG_ID_LIFE                      (300)
#define BOTNET_KEEPALIVE_PING_PERIOD            (30)
#define BOTNET_KEEPALIVE_PING_DRIFT             (rand()%20)
#define BOTNET_LOGIN_GREETING_TIMEOUT           (60)
#define BOTNET_OLDER_OBJECT_DIFF_TIME           (60)
#define BOTNET_PUSH_PROC_EVERY                  (60)
#define BOTNET_PUSH_PROC_LAST_UPDATE            (120+40)
#define BOTNET_CHANS_LAST_UPDATE                (120+25)
#define BOTNET_PRV_PUSH_LAST_UPDATE             (120-10)
#define BOTNET_PING_TIMEOUT                     (120)
#define BOTNET_PRELINK_WAIT                     (10)
#define BOTNET_MAXIMUM_SOCKET_CACHE             (64*1024*16)
#define BOTNET_KILL_CONNECTION_SOCKET_CACHE     (64*1024*64)
#define BOTNET_ROUNDTRIP_DIFF_TIME_MAX          (10)
#define BOTNET_REPORT_DIFF_TIME_CHANGE_ABOVE    (5)
#define SSL_CLIENT_MAXIMUM_QUEUE                (1024*128)
#define DCC_AUTH_TIMEOUT                        (60*3)
/*   :END */

/*!
    \def DH_BITS
    \brief Number of Diffie && Hellman bits to use with GnuTLS
*/
/*!
    \def RSA_BITS
    \brief Number of RSA bits to use with GnuTLS
*/
/*!
    \def DH_RSA_REGENERATION_NUM_SECS
    \brief Number of seconds, period of regenerating DH and RSA bits, adjust after security consideration (GnuTLS)
*/
/*!
    \def DH_RSA_REGENERATION_NUM_TRANSACTIONS
    \brief Number of TLS transactions, period of regenerating DH and RSA bits, adjust after security consideration (GnuTLS)
*/
/*!
    \def BOTNET_KEEPALIVE_PING_PERIOD
    \brief Number of seconds, BOTNET keep-alive pings interval
*/
/*!
    \def BOTNET_KEEPALIVE_PING_DRIFT
    \brief Random number of seconds that will be added to next PING packet schedule in time (to lower network traffic)
*/
/*!
    \def BOTNET_LOGIN_GREETING_TIMEOUT
    \brief Number of seconds to wait for initial greeting message terminated by '\\n' from remote bot
*/
/*!
    \def BOTNET_OLDER_OBJECT_DIFF_TIME
    \brief Number of seconds, if remote object is older than this value, replication is done (note: there is time difference computing)
*/
/*!
    \def BOTNET_PUSH_PROC_EVERY
    \brief Number of seconds, period we try to push procedures
*/
/*!
    \def BOTNET_PUSH_PROC_LAST_UPDATE
    \brief Number of seconds, period we update procedures
*/
/*!
    \def BOTNET_CHANS_LAST_UPDATE
    \brief Number of seconds, period we try to push channel definitions
*/
/*!
    \def BOTNET_PRV_PUSH_LAST_UPDATE
    \brief Number of seconds, period we try to private definition
*/
/*!
    \def BOTNET_PING_TIMEOUT
    \brief Number of seconds, of none message from remote party that is considered as ping timeout
*/
/*!
    \def BOTNET_PRELINK_WAIT
    \brief Number of seconds, how many we wait for parties sync until we state "we are linked"
*/

/*!
    \def FD_SETSIZE_MAX
    \brief Maximum number of rotating sockets for idle sleeper, or FD_SETSIZE will be used, whichever has less value
*/
#define FD_SETSIZE_MAX          (64)

#if !defined(SOCKET_BIND_WILDCARD)
/*!
    \def SOCKET_BIND_WILDCARD
    \brief If set to 1, all sockets will be bound to wildcard IP address (0.0.0.0)
            and port 0, so if you are running Mac OS X which has broken
            implementation of BSD sockets, define this to 1; otherwise it is
            strongly recommended to set it to 0
*/
#   define SOCKET_BIND_WILDCARD     0
#endif

// === BEGIN DEBUG: debugging support follows

// NOTE: if 'NDEBUG' is defined, or '_DEBUG' is not defined, following has not effect and all of them is dissabled later in code

// Values of preprocessor define's can be 1 to enable, 0 to disable feature

/*!
    \def DEBUG_TIME_MEASUREMENT
    \brief Define to 1, if we want to echo to console execution time measurement in some functions (useful for profiling / optimisation experimental tests)
    \warning Will be ignored/disabled if '_DEBUG' preprocessor define is NOT set
*/
#ifndef DEBUG_TIME_MEASUREMENT
#   define DEBUG_TIME_MEASUREMENT      1
#endif


/*!
    \def DEBUG_TIME_MEASUREMENT_WAIT
    \brief If 'DEBUG_TIME_MEASUREMENT' is defined to '1', this is default number of seconds to wait after each measurement (to notice execution time's echo in console)
    \note Can be overriden by explicit argument to measurement function, this is just default
    \warning If not set to '(0)' (zero), this may significantly slow down your bot! Enable 'DEBUG_TIME_MEASUREMENT' ONLY for profiling purposes, not in production use!
    \warning Will be ignored/disabled if '_DEBUG' preprocessor define is NOT set
    \warning The value SHOULD be enclosed in curly braces (e.g. "(5)"), because it is used in mathematical expression
    \note Currently has no effect; unless you apply patch to sources, if you wish to wait these number of seconds after on each of your custom time measurement
*/
#if !defined(DEBUG_TIME_MEASUREMENT_WAIT)
#   define DEBUG_TIME_MEASUREMENT_WAIT     (5)
#endif


/*!
    \def RESOLV_APPEND_DOT
    \brief Define to 1, if we want to append dot at the end of hostname when getting its IP address (if there it is not yet); this is workaround for some broken configurations of DNS resolver (resolver returning local address because of automatic appending of local domain in case of non-existing domain)
*/
#if !defined(RESOLV_APPEND_DOT)
#   define RESOLV_APPEND_DOT                1
#endif

/*!
    \def HEAPDEBUG_NUM_TRANS_FOR_RECREATION
    \brief Number of transactions (re/allocation + deallocation events) after which is STL's map of allocated blocks re-created (experimental feature), to force better heap defragmentation (in theory)
    \warning Do not set it to very low number, or your bot will be _significantly_ slower!
    \warning Set to "(0)" (zero) to disable this feature
*/
#if !defined(HEAPDEBUG_NUM_TRANS_FOR_RECREATION)
#   define HEAPDEBUG_NUM_TRANS_FOR_RECREATION       (1000 * 1000)
#endif

/*!
    \def HARD_LOCKOUT_LIMIT_ADJUST
    \brief Number of unsuccessful logins from same IP above lockout limit defined in "conf.txt"
*/
#if !defined(HARD_LOCKOUT_LIMIT_ADJUST)
#   define HARD_LOCKOUT_LIMIT_ADJUST                (5)
#endif

/*!
    \def MAX_RESOLV_THREADS
    \brief Number of maximum threads created for DNS lookup
*/
#if !defined(MAX_RESOLV_THREADS)
#   define MAX_RESOLV_THREADS                       (10)
#endif

/*!
    \def PCRE_BACKTRACK_LIMIT
    \brief PCRE limit
*/
# if !defined(PCRE_BACKTRACK_LIMIT)
#   define PCRE_BACKTRACK_LIMIT                     (100000)
#endif

/*!
    \def PCRE_RECURSION_LIMIT
    \brief PCRE limit
*/
# if !defined(PCRE_RECURSION_LIMIT)
#   define PCRE_RECURSION_LIMIT                     (100000)
#endif

/*!
    \def PCRE_PATTERN_COMPILER_CACHE_TTL
    \brief RegExp compiled patterns with PCRE are cached,
        this defines how long compiled pattern lives in the cache,
        after last use (reference). This speeds-up the bot,
        so the same pattern need not to be compiled again
        (which is much CPU time-consuming)
*/
# if !defined(PCRE_PATTERN_COMPILER_CACHE_TTL)
#   define PCRE_PATTERN_COMPILER_CACHE_TTL          (60*60*4)
#endif

/*!
    \def BOTNET_BOTS_VECTOR_RECREATION
    \brief If defined to non-zero, vector of remote bots on BOTNET is recreated re-created on each new bot added to vector (experimental feature), to force better heap defragmentation (in theory)
    \warning Set to "0" (zero) to disable this feature
*/
#if !defined(BOTNET_BOTS_VECTOR_RECREATION)
#   define BOTNET_BOTS_VECTOR_RECREATION            1
#endif

/*!
    \def MODULE_MAXIMUM_OPEN_FILES
    \brief Maximum open files (via 'fopen()' built-in function) in module's user script
*/
#if !defined(MODULE_MAXIMUM_OPEN_FILES)
#   define MODULE_MAXIMUM_OPEN_FILES                (32)
#endif

/*!
    \def MY_STATIC
    \brief Define this to "static" (without quotes) to improve linking time, but then compiler eats much more memory
*/
#if !defined(MY_STATIC)
#   define MY_STATIC                                static
#endif

/*!
    \def MY_LISTEN_BACKLOG
    \brief This defines BSD sockets API call listen() backlog value; recommended: 5; if unsure, keep it intact/at default value of 5
*/
#if !defined(MY_LISTEN_BACKLOG)
#   define MY_LISTEN_BACKLOG                        (5)
#endif

/*!
    \def REPORT_LEAKS
    \brief If set to "1", heap debugging is turned on
    \warning Note that this will not debug STL allocations
*/
#if !defined(REPORT_LEAKS)
#   define REPORT_LEAKS                             1
#endif

/*!
    \def SANITIZE_FREED_MEMORY
    \brief If set to "1", it zeroes all freed memory
    \warning Only available in heap-debugging mode, i.e. REPORT_LEAKS==1
    \warning Note that this will not affect STL deallocations
*/
#if !defined(SANITIZE_FREED_MEMORY)
#   define SANITIZE_FREED_MEMORY                    1
#endif

#include "params_epilogue.h"

#endif

