/***************************************************************************
                          params_epilogue.h  -  description
                             -------------------
    begin                : Mon Aug 8 2007
    copyright            : (C) 2007 by Marian VooDooMan Meravy
    email                : ghostvoodooman -a.T= users (dot) sourceforge.net
 ***************************************************************************/

/* $Id: params_epilogue.h,v 1.16 2008/11/24 12:09:23 ghostvoodooman Exp $ */

/***************************************************************************

VooDoo cIRCle - an IRC (ro)bot
Copyright (C) 2004-2008 by Marian VooDooMan Meravy
  (ghostvoodooman -a.T= users (dot) sourceforge.net)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************/

/*!
    \file
    \brief Parameters for compilation - epilogue section
    \version $Id: params_epilogue.h,v 1.16 2008/11/24 12:09:23 ghostvoodooman Exp $
*/

#ifndef PARAMS_H__EPILOGUE_H__
#define PARAMS_H__EPILOGUE_H__

#include <typeinfo>             // globally: we need this

#ifdef _DEBUG
#   if !defined(REPORT_LEAKS)
#       define REPORT_LEAKS             1
#   endif
#else
#   ifndef REPORT_LEAKS
#       define REPORT_LEAKS             0
#   endif
#endif

#ifndef USE_OPENSSL
#   if defined(_MSC_VER) || defined(__BORLANDC__)
#       define USING_GNUTLS_DLL
#       ifndef GNUTLS_USEDLL
#           define GNUTLS_USEDLL
#       endif
#   endif
#endif

#ifdef IPv6
#   if defined(_WIN32) || defined(_WIN64)
#       ifdef __GNUC__
#           undef WINVER            // to make Dev-C++ happy
#           define WINVER 0x0501    // ^^ ...
#       endif
#   endif
#endif

// ===
// misc: do NOT change following unless you know what you are doing!
#ifndef _DEBUG
#   undef DEBUG_ALL_DUMPS_DISABLE
#   define DEBUG_ALL_DUMPS_DISABLE      1

#   undef DEBUG_TIME_MEASUREMENT
#   define DEBUG_TIME_MEASUREMENT       0

#   undef DEBUG_TIME_MEASUREMENT_WAIT
#   define DEBUG_TIME_MEASUREMENT_WAIT  (0)
#endif

#ifndef _DEBUG
#   undef DEBUG_WANT_ASSERTS
#   define DEBUG_WANT_ASSERTS           0
#endif

#if !defined(DEBUG_ALL_DUMPS_DISABLE)
#   define  DEBUG_ALL_DUMPS_DISABLE     1
#endif

#if 1==DEBUG_ALL_DUMPS_DISABLE
#   undef DEBUG_MICROCODE_ADD_ECHO
#   define DEBUG_MICROCODE_ADD_ECHO     0

#   undef DEBUG_MICROCODE_EXEC_ECHO
#   define DEBUG_MICROCODE_EXEC_ECHO    0

#   undef DEBUG_PROCFUNC_CMDS_ECHO
#   define DEBUG_PROCFUNC_CMDS_ECHO     0
#else
#   undef DEBUG_MICROCODE_ADD_ECHO
#   define DEBUG_MICROCODE_ADD_ECHO     1

#   undef DEBUG_MICROCODE_EXEC_ECHO
#   define DEBUG_MICROCODE_EXEC_ECHO    1

#   undef DEBUG_PROCFUNC_CMDS_ECHO
#   define DEBUG_PROCFUNC_CMDS_ECHO     1
#endif

// === END   DEBUG: debugging support

#ifdef __linux__
#   ifndef _REENTRANT
#       define _REENTRANT
#   endif
#   ifndef  _POSIX_SOURCE
#       define _POSIX_SOURCE
#   endif
#endif

#ifdef _MSC_VER
#   define MY_FORCE_INLINE      __forceinline
#   define MY_WISH_INLINE       __inline
#else
#   define MY_FORCE_INLINE      inline
#   define MY_WISH_INLINE       inline
#endif

/*!
    \def DEBUG_ALL_DUMPS_DISABLE
    \brief If set to 1, we will disable ALL following dumps at once
*/
#ifndef _DEBUG
#   if !defined(DEBUG_ALL_DUMPS_DISABLE)
#       define DEBUG_ALL_DUMPS_DISABLE      1
#   endif
#endif

/*!
    \def DEBUG_VAR_KEEPER_DEBUG_STRING_DISABLE
    \brief If set to 1, we will disable storing of debugging strings when allocating local IDs
*/
#ifndef _DEBUG
#   ifndef DEBUG_VAR_KEEPER_DEBUG_STRING_DISABLE
#       define DEBUG_VAR_KEEPER_DEBUG_STRING_DISABLE       0
#   endif
#endif

/*!
    \def DEBUG_MICROCODE_ADD_ECHO
    \brief Define to 1, if we want to echo to console on each microcode instruction add while compiling functions
    \warning If defined to 1, this will SIGNIFICANTLY slow down bot and generates HUGE amount of messages
*/
#ifndef _DEBUG
#   ifndef DEBUG_MICROCODE_ADD_ECHO
#       define DEBUG_MICROCODE_ADD_ECHO    1
#   endif
#endif

/*!
    \def DEBUG_MICROCODE_EXEC_ECHO
    \brief Define to 1, if we want to echo to console on each microcode instruction execution while evaluating/executing functions
    \warning If defined to 1, this will SIGNIFICANTLY slow down bot and generates HUGE amount of messages
*/
#ifndef _DEBUG
#   ifndef DEBUG_MICROCODE_EXEC_ECHO
#       define DEBUG_MICROCODE_EXEC_ECHO   1
#   endif
#endif

/*!
    \def DEBUG_PROCFUNC_CMDS_ECHO
    \brief Define to 1, if we want to echo to console on each procedure/function command beeing processed
    \warning If defined to 1, this will SIGNIFICANTLY slow down bot and generates HUGE amount of messages
*/
#ifndef _DEBUG
#   ifndef DEBUG_PROCFUNC_CMDS_ECHO
#       define DEBUG_PROCFUNC_CMDS_ECHO    1
#   endif
#endif

#ifdef _DEBUG
#   ifndef DEBUG_VAR_KEEPER_DEBUG_STRING_DISABLE
#       define DEBUG_VAR_KEEPER_DEBUG_STRING_DISABLE       0
#   endif
#else
#   ifndef DEBUG_VAR_KEEPER_DEBUG_STRING_DISABLE
#       define DEBUG_VAR_KEEPER_DEBUG_STRING_DISABLE       1
#   endif
#endif

/*!
    \def DEBUG_WANT_ASSERTS
    \brief If set to 1, if we want to throw assertion failure in case of strange things (i.e. script variable ID not found); on Win core is dumped, bot continues normally; *NIX: core may be dumped (depends on parent process limit settings), bot crashes on SEGV, i.e. NULL pointer derefecence in order to get core to be dumped
*/
#ifdef _DEBUG
#   if !defined(DEBUG_WANT_ASSERTS)
#       define DEBUG_WANT_ASSERTS       1
#   endif
#endif

// sensitive store
#define SENSITIVE_STORE_WIPE_CHAR           ((int)'w')
#define UNINITIALIZED_MEMORY_MAGICK_MARKER  ((unsigned char)0xcd)

// for binary modules interface header
#define VDCMOD_HOST                     1
#define VDCMOD_API_REQUEST_VERSION      4
#define VDCMOD_LONG_LONG                MY_LONG_LONG

#endif

