/***************************************************************************
                          c_pcre.cpp  -  description
                             -------------------
    begin                : Wed Feb 07 2008
    copyright            : (C) 2008 by Marian VooDooMan Meravy
    email                : ghostvoodooman -a.T= users (dot) sourceforge.net
 ***************************************************************************/

/* $Id: c_pcre.cpp,v 1.9 2008/11/24 12:09:46 ghostvoodooman Exp $ */

/***************************************************************************

VooDoo cIRCle - an IRC (ro)bot
Copyright (C) 2004-2008 by Marian VooDooMan Meravy
  (ghostvoodooman -a.T= users (dot) sourceforge.net)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************/

/*!
    \file
    \brief Implements wrapper class for PCRE cache and library
    \version $Id: c_pcre.cpp,v 1.9 2008/11/24 12:09:46 ghostvoodooman Exp $
*/

#if defined(HAVE_CONFIG_H)
#   include <config.h>
#endif

#include <string>

using namespace std;

#include "../utils/memleak.h"
#include "../params.h"

#include "c_pcre.h"

#include "pattern_cache.h"
#include "../internal_state.h"

#include "../utils/precomp.h"

#if defined(_MSC_VER) || defined(__BORLANDC__)
#   pragma hdrstop
#endif

#ifdef _MSC_VER
//  BEGIN: disable MSVC warning: warning C4355: 'this' : used in base member initializer list
#   pragma warning( push )
#   pragma warning( disable : 4355 )
#endif

c_pcre::c_pcre(const string& pattern, int options)
    : m_pcre(g_pcre.get_pcre(pattern,options))
    , m_last_error(0)
    , m_callout_registrar(this)
{
    m_pcre.get_error_info(m_last_error_offset,m_compile_error_str);
}

#ifdef _MSC_VER
//  END  : disable MSVC warning: warning C4355: 'this' : used in base member initializer list
#   pragma warning( pop )
#endif

c_pcre::~c_pcre()
{
}

bool c_pcre::is_okay() const
{
    return m_pcre.is_okay();
}

size_t c_pcre::exec(const string& subject, MY_LONG_LONG start_offset, int options, vector<vector<pair<int,string> > >& captured_matches, size_t max_matches)
{
    MY_ASSERT(is_okay());

    m_last_error=0;

    // BEGIN: sanity checks
    if(start_offset>MAX_SIGNED_INT) {
        return 0;
    }
    if(start_offset<0) {
        return 0;
    }
    if(subject.length()>MAX_SIGNED_INT) {
        return 0;
    }
    if(subject.length()<=static_cast<string::size_type>(start_offset)) {
        return 0;
    }

    size_t res=m_pcre.exec(subject,static_cast<size_t>(start_offset),options,max_matches,captured_matches,m_last_error,reinterpret_cast<void*>(this));

    return res;
}

string c_pcre::get_error_string(int rc) const
{
    string res;
    switch(rc) {
        case PCRE_ERROR_NOMATCH:
            res="PCRE_ERROR_NOMATCH";
            break;
        case PCRE_ERROR_NULL:
            res="PCRE_ERROR_NULL";
            break;
        case PCRE_ERROR_BADOPTION:
            res="PCRE_ERROR_BADOPTION";
            break;
        case PCRE_ERROR_BADMAGIC:
            res="PCRE_ERROR_BADMAGIC";
            break;
#if defined(PCRE_ERROR_UNKNOWN_OPCODE)
        case PCRE_ERROR_UNKNOWN_OPCODE:
            res="PCRE_ERROR_UNKNOWN_OPCODE or PCRE_ERROR_UNKNOWN_NODE";
            break;
#endif
#if defined(PCRE_ERROR_UNKNOWN_OPCODE) && (PCRE_ERROR_UNKNOWN_NODE != PCRE_ERROR_UNKNOWN_OPCODE)
            /* For backward compatibility */
        case PCRE_ERROR_UNKNOWN_NODE:
            res="PCRE_ERROR_UNKNOWN_NODE";
            break;
#endif
        case PCRE_ERROR_NOMEMORY:
            res="PCRE_ERROR_NOMEMORY";
            break;
        case PCRE_ERROR_NOSUBSTRING:
            res="PCRE_ERROR_NOSUBSTRING";
            break;
        case PCRE_ERROR_MATCHLIMIT:
            res="PCRE_ERROR_MATCHLIMIT";
            break;
        case PCRE_ERROR_CALLOUT:
            res="PCRE_ERROR_CALLOUT";
            break;
        case PCRE_ERROR_BADUTF8:
            res="PCRE_ERROR_BADUTF8";
            break;
        case PCRE_ERROR_BADUTF8_OFFSET:
            res="PCRE_ERROR_BADUTF8_OFFSET";
            break;
        case PCRE_ERROR_PARTIAL:
            res="PCRE_ERROR_PARTIAL";
            break;
        case PCRE_ERROR_BADPARTIAL:
            res="PCRE_ERROR_BADPARTIAL";
            break;
        case PCRE_ERROR_INTERNAL:
            res="PCRE_ERROR_INTERNAL";
            break;
        case PCRE_ERROR_BADCOUNT:
            res="PCRE_ERROR_BADCOUNT";
            break;
        case PCRE_ERROR_DFA_UITEM:
            res="PCRE_ERROR_DFA_UITEM";
            break;
        case PCRE_ERROR_DFA_UCOND:
            res="PCRE_ERROR_DFA_UCOND";
            break;
        case PCRE_ERROR_DFA_UMLIMIT:
            res="PCRE_ERROR_DFA_UMLIMIT";
            break;
        case PCRE_ERROR_DFA_WSSIZE:
            res="PCRE_ERROR_DFA_WSSIZE";
            break;
        case PCRE_ERROR_DFA_RECURSE:
            res="PCRE_ERROR_DFA_RECURSE";
            break;
#if defined(PCRE_ERROR_RECURSIONLIMIT)
        case PCRE_ERROR_RECURSIONLIMIT:
            res="PCRE_ERROR_RECURSIONLIMIT";
            break;
#endif
#if defined(PCRE_ERROR_NULLWSLIMIT)
        case PCRE_ERROR_NULLWSLIMIT:
            res="PCRE_ERROR_NULLWSLIMIT";
            break;
#endif
#if defined(PCRE_ERROR_BADNEWLINE)
        case PCRE_ERROR_BADNEWLINE:
            res="PCRE_ERROR_BADNEWLINE";
            break;
#endif
        default:
            res="<other undocumented error>";
            break;
    }

    res+=" (numeric error code: ";
    res+=tos(rc);
    res+=")";

    return res;
}

bool c_pcre::get_last_error() const
{
    return 0 != m_last_error;
}

string c_pcre::get_last_error_string() const
{
    return get_error_string(m_last_error);
}

void c_pcre::get_compile_error_info(int &offset, string& error_string) const
{
    offset=m_last_error_offset;
    error_string=m_compile_error_str;
}

string c_pcre::get_pcre_version()
{
    return pcre_version();
}

int c_pcre::callout(pcre_callout_block* UNUSEDARG(data))
{
    return PCRE_CALLOUT_OKAY;
}

