/***************************************************************************
                          c_pcre.h  -  description
                             -------------------
    begin                : Wed Feb 07 2008
    copyright            : (C) 2008 by Marian VooDooMan Meravy
    email                : ghostvoodooman -a.T= users (dot) sourceforge.net
 ***************************************************************************/

/* $Id: c_pcre.h,v 1.8 2008/03/11 21:11:46 ghostvoodooman Exp $ */

/***************************************************************************

VooDoo cIRCle - an IRC (ro)bot
Copyright (C) 2004-2008 by Marian VooDooMan Meravy
  (ghostvoodooman -a.T= users (dot) sourceforge.net)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************/

/*!
    \file
    \brief Implements wrapper class for PCRE cache and library
    \version $Id: c_pcre.h,v 1.8 2008/03/11 21:11:46 ghostvoodooman Exp $
*/

#ifndef C_PCRE_H__
#define C_PCRE_H__

#if defined(HAVE_CONFIG_H)
#   include <config.h>
#endif

#include <string>
#include <vector>

using namespace std;

#include "../utils/memleak.h"
#include "../params.h"

#include "pattern_cache.h"

#ifdef _MSC_VER
//  BEGIN: disable MSVC warning: warning C4512: 'c_pcre' : assignment operator could not be generated
#   pragma warning( push )
#   pragma warning( disable : 4512 )
#endif

class c_pcre_lib_callout;

/*!
    \brief Implements wrapper class for PCRE cache and library, and automatic deallocation of offset vector in d-tor
    \author VooDooMan
    \version 1
    \date 2008-02-07
*/
class c_pcre
{
    const c_pattern& m_pcre;    //!< PCRE object

    int m_last_error;           //!< Last PCRE error code
    int m_last_error_offset;    //!< Last PCRE pattern error offset
    string m_compile_error_str; //!< PCRE pattern compilation error string

    /*!
        \brief Retrieves error message for human
        \author VooDooMan
        \version 1
        \date 2008-02-07
    */
    string get_error_string(int rc) const;

    /*!
        \brief PCRE callout callback
        \author VooDooMan
        \version 1
        \date 2008-02-10
        \param data Callback data structure
        \retval Return value for callout callback
    */
    int callout(pcre_callout_block* data);

    friend class c_pcre_lib_callout;

    c_pcre_callout_registrar m_callout_registrar;   //!< PCRE callout callback registrar instance

public:

    /*!
        \brief C-tor
        \author VooDooMan
        \date 2008-02-07
        \version 1
        \param pattern Pattern
        \param options PCRE options
    */
    c_pcre(const string& pattern, int options = 0);

    /*!
        \brief D-tor (clean-ups temporary data)
        \version 1
        \author VooDooMan
        \date 2008-02-07
    */
    ~c_pcre();

    /*!
        \brief Checks whether pattern compilation succeeded
        \version 1
        \date 2008-02-07
        \author VooDooMan
        \retval true Okay
        \retval false Error
    */
    bool is_okay() const;

    typedef vector<vector<string> > captures_t;

    /*!
        \brief Executes a regexp pattern on subject
        \version 1
        \date 2008-02-07
        \author VooDooMan
        \param subject Subject
        \param start_offset Offset in 'subject' to start from
        \param options PCRE library options
        \return Returns number of offsets captured
        \param captured_matches Returns captured offsets and strings
        \param max_matches Maximum matches limit
    */
    size_t exec(const string& subject, MY_LONG_LONG start_offset, int options, vector<vector<pair<int,string> > >& captured_matches, size_t max_matches);

    /*!
        \brief Maps return code to human string
        \version 1
        \date 2008-02-08
        \author VooDooMan
        \retval true Error
        \retval false Okay
    */
    bool get_last_error() const;

    /*!
        \brief Retrieves last PCRE error
        \version 2
        \date 2008-02-07, 2008-02-08
        \author VooDooMan
        \return Returns an error message
    */
    string get_last_error_string() const;

    /*!
        \brief Retrieves last PCRE error
        \version 2
        \date 2008-02-07, 2008-02-08
        \author VooDooMan
        \return Returns an error message
    */
    void get_compile_error_info(int &offset, string& error_string) const;

    /*!
        \brief Retrieves PCRE library version
        \version 1
        \date 2008-02-09
        \author VooDooMan
    */
    static string get_pcre_version();
};

#ifdef _MSC_VER
//  END  : disable MSVC warning: warning C4512: 'c_pcre' : assignment operator could not be generated
#   pragma warning( pop )
#endif

#endif

