/***************************************************************************
                          pcre_wrapper.cpp  -  description
                             -------------------
    begin                : Wed Feb 06 2008
    copyright            : (C) 2008 by Marian VooDooMan Meravy
    email                : ghostvoodooman -a.T= users (dot) sourceforge.net
 ***************************************************************************/

/* $Id: pcre_wrapper.cpp,v 1.8 2008/03/11 21:11:46 ghostvoodooman Exp $ */

/***************************************************************************

VooDoo cIRCle - an IRC (ro)bot
Copyright (C) 2004-2008 by Marian VooDooMan Meravy
  (ghostvoodooman -a.T= users (dot) sourceforge.net)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************/

/*!
    \file
    \brief Implements communication wrapper with PCRE lib API
    \version $Id: pcre_wrapper.cpp,v 1.8 2008/03/11 21:11:46 ghostvoodooman Exp $
*/

#if defined(HAVE_CONFIG_H)
#   include <config.h>
#endif

#include "../utils/memleak.h"
#include "../params.h"

#include <pcre.h>

#include "../utils.h"

#include "pcre_wrapper.h"
#include "c_pcre.h"

#include "../utils/precomp.h"

#if defined(_MSC_VER) || defined(__BORLANDC__)
#   pragma hdrstop
#endif

/*!
    \brief PCRE memory management to allow memory leak detection
    \author VooDooMan
    \date 2008-02-06
    \version 1
*/
class c_pcre_lib_memory_management
{
    /*!
        \brief PCRE callback allowing to detect memory leaka on debug build)
        \author VooDooMan
        \version 1
        \date 2008-02-06
        \param sz Size of the buffer
    */
    static void* pcre_my_malloc(size_t sz)
    {
        return my_malloc(sz);
    }

    /*!
        \brief PCRE callback allowing to detect memory leaka on debug build)
        \author VooDooMan
        \version 1
        \date 2008-02-06
        \param p Pointer
    */
    static void pcre_my_free(void* p)
    {
        return my_free(p);
    }

    /*!
        \brief PCRE callback allowing to detect memory leaka on debug build)
        \author VooDooMan
        \version 1
        \date 2008-02-06
        \param sz Size of the buffer
    */
    static void* pcre_my_stack_malloc(size_t sz)
    {
        return my_malloc(sz);
    }

    /*!
        \brief PCRE callback allowing to detect memory leaka on debug build)
        \author VooDooMan
        \version 1
        \date 2008-02-06
        \param p Pointer
    */
    static void pcre_my_stack_free(void* p)
    {
        return my_free(p);
    }

public:

    c_pcre_lib_memory_management()
    {
        pcre_malloc=pcre_my_malloc;
        pcre_free=pcre_my_free;
        pcre_stack_malloc=pcre_my_stack_malloc;
        pcre_stack_free=pcre_my_stack_free;
    }
};

static c_pcre_lib_memory_management pcre_lib_memory_management;

class c_pcre;

/*!
    \brief PCRE callout implementation
    \author VooDooMan
    \date 2008-02-10
    \version 1
*/
class c_pcre_lib_callout
{
    static vector<const c_pcre*> m_pcre;

    /*!
        \brief Registers PCRE callback (callout)
        \author VooDooMan
        \version 1
        \date 2008-02-10
        \param pcre C_pcre class instance
        \warning Not thread-safe!
    */
    void register_callout(const c_pcre* pcre)
    {
        m_pcre.push_back(pcre);
    }

    /*!
        \brief Unregisters PCRE callback (callout)
        \author VooDooMan
        \version 1
        \date 2008-02-10
        \param pcre C_pcre class instance
        \warning Not thread-safe!
    */
    void unregister_callout(const c_pcre* pcre)
    {
        vector<const c_pcre*>::iterator i1;
        for(i1=m_pcre.begin(); i1!=m_pcre.end(); ++i1) {
            if(*i1==pcre) {
                m_pcre.erase(i1);
                return;
            }
        }
    }

    friend class c_pcre_callout_registrar;

public:

    /*!
        \brief PCRE callback (callout)
        \author VooDooMan
        \version 1
        \date 2008-02-10
        \param data Data
    */
    static int my_pcre_callout(pcre_callout_block *data)
    {
        if(NULL==data || NULL==data->callout_data)
            return PCRE_CALLOUT_OKAY;

        vector<const c_pcre*>::iterator i1;
        for(i1=m_pcre.begin(); i1!=m_pcre.end(); ++i1) {
            if(*i1==reinterpret_cast<c_pcre*>(data->callout_data)) {
                return const_cast<c_pcre*>(*i1)->callout(data);
            }
        }

        return PCRE_CALLOUT_OKAY;
    }

    c_pcre_lib_callout()
    {
        pcre_callout=my_pcre_callout;
    }
};

vector<const c_pcre*> c_pcre_lib_callout::m_pcre;

static c_pcre_lib_callout pcre_lib_callout;

bool pcre_my_compile(const string& pattern, int options, int& error_offset, string& error_string, pcre* &result1, pcre_extra* &result2)
{
    int ec=0;
    const char* estr=NULL;

    result1=pcre_compile2(pattern.c_str(),options,
            &ec,
            &estr,
            &error_offset,
            NULL
            );

    if(0!=ec) {
        if(NULL==estr) {
            error_string="<pcre libarary returned invalid error string (NULL pointer) as description of error>";
        } else {
            error_string=estr;
        }

        return false;
    }

    const char* err;
    result2=pcre_study(result1,0,&err);
    if(NULL==result2) {
        result2=reinterpret_cast<pcre_extra*>(my_malloc(sizeof(pcre_extra)));
        memset(result2,0,sizeof(pcre_extra));
    }

    return true;
}

c_pcre_callout_registrar::c_pcre_callout_registrar(const c_pcre* pcre)
    : m_pcre(pcre)
{
    pcre_lib_callout.register_callout(m_pcre);
}

c_pcre_callout_registrar::~c_pcre_callout_registrar()
{
    pcre_lib_callout.unregister_callout(m_pcre);
}

