/***************************************************************************
                          pcre_wrapper.h  -  description
                             -------------------
    begin                : Wed Feb 06 2008
    copyright            : (C) 2008 by Marian VooDooMan Meravy
    email                : ghostvoodooman -a.T= users (dot) sourceforge.net
 ***************************************************************************/

/* $Id: pcre_wrapper.h,v 1.8 2008/03/11 21:11:47 ghostvoodooman Exp $ */

/***************************************************************************

VooDoo cIRCle - an IRC (ro)bot
Copyright (C) 2004-2008 by Marian VooDooMan Meravy
  (ghostvoodooman -a.T= users (dot) sourceforge.net)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************/

/*!
    \file
    \brief Implements communication wrapper with PCRE lib API
    \version $Id: pcre_wrapper.h,v 1.8 2008/03/11 21:11:47 ghostvoodooman Exp $
*/

#ifndef PCRE_WRAPPER_H__
#define PCRE_WRAPPER_H__

#if defined(HAVE_CONFIG_H)
#   include <config.h>
#endif

#include "../utils/memleak.h"
#include "../params.h"

#include "../utils.h"

#include <pcre.h>

/*!
    \brief Compiles a pcre pattern
    \version 1
    \date 2008-02-06
    \author VooDooMan
    \retval true No error
    \retval false Error
    \param pattern Regexp pattern to compile
    \param options PCRE options
    \param error_offset If: retval==true: intact; retval==false: offset where an error occured
    \param error_string If: retval==true: intact; retval==false: error message from pcre library
    \param result1 If: retval==true: PCRE compiled pattern; retval==false: NULL
    \param result2 If: retval==true: PCRE compiled extra study pattern; retval==false: NULL
*/
bool pcre_my_compile(const string& pattern, int options, int& error_offset, string& error_string, pcre* &result1, pcre_extra* &result2);

class c_pcre;

enum
{
    PCRE_CALLOUT_OKAY = 0,       //!< Callout callback okay, continue
    PCRE_CALLOUT_ERROR = -1,     //!< Error, stop executing RegExp
};

/*!
    \brief PCRE callout implementation, automatic registering in c-tor and unregistering in d-tor
    \author VooDooMan
    \date 2008-02-10
    \version 1
*/
class c_pcre_callout_registrar
{
    const c_pcre* m_pcre;

public:

    /*!
        \brief C-tor, registers PCRE callout callback
        \author VooDooMan
        \date 2008-02-10
        \version 1
        \param pcre c_pcre class instance
    */
    c_pcre_callout_registrar(const c_pcre* pcre);

    /*!
        \brief D-tor, unregisters PCRE callout callback
        \author VooDooMan
        \date 2008-02-10
        \version 1
    */
    ~c_pcre_callout_registrar();
};

#endif

