/***************************************************************************
                          proc.cpp  -  description
                             -------------------
    begin                : Sat Oct 8 2005
    copyright            : (C) 2005-2007 by Marian VooDooMan Meravy
    email                : ghostvoodooman -a.T= users (dot) sourceforge.net
 ***************************************************************************/

/* $Id: proc.cpp,v 1.40 2008/06/08 08:43:13 ghostvoodooman Exp $ */

/***************************************************************************

VooDoo cIRCle - an IRC (ro)bot
Copyright (C) 2005-2008 by Marian VooDooMan Meravy
  (ghostvoodooman -a.T= users (dot) sourceforge.net)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************/

/*!
    \file
    \brief Provides command execution for procedures
    \version $Id: proc.cpp,v 1.40 2008/06/08 08:43:13 ghostvoodooman Exp $
*/

#if defined(HAVE_CONFIG_H)
#   include <config.h>
#endif

#include <string>
#include <vector>
#include <stack>
#include <list>
#include <map>

#include <stdlib.h>

#include "utils/memleak.h"
#include "utils/assert.h"
#include "utils.h"
#include "params.h"

#include "proc.h"

#include "irc.h"
#include "logic.h"
#include "conf.h"
#include "compat.h"
#include "utils.h"
#include "botnet.h"
#include "dcc.h"
#include "identd.h"
#include "match.h"
#include "stats.h"
#include "modules.h"
#include "threads.h"
#include "irc/c_irc_server.h"
#include "irc/c_online_channel.h"
#include "proc/c_proc_exc.h"
#include "internal_state.h"

#include "utils/precomp.h"

#if defined(_MSC_VER) || defined(__BORLANDC__)
#   pragma hdrstop
#endif

string get_timer_duration(long duration)
{
    string dur;

    long j=duration%60;
    char tmp2[2];
    tmp2[1]=0;
    if(j<10)
        tmp2[0]='0';
        else
        tmp2[0]=0;
    dur=(string)":"+(string)tmp2+tos(j)+dur;

    duration/=60;
    j=duration%60;
    if(j<10)
        tmp2[0]='0';
        else
        tmp2[0]=0;
    dur=(string)":"+(string)tmp2+tos(j)+dur;

    duration/=60;
    j=duration%24;
    if(j<10)
        tmp2[0]='0';
        else
        tmp2[0]=0;
    dur=(string)":"+(string)tmp2+tos(j)+dur;

    duration/=24;
    dur=tos(j)+dur;

    return dur;
}

string c_proc_command::dump(c_module* this_module, int& indent)
{
    string res;
    for(int i1=0; i1<indent; ++i1)
        res+="    ";
    switch(command_) {
        case _op:
            {
                if(num_args!=3)
                    THROW_C_PROC_EXC(this_module,command_,c_proc_exc::E_NUM_ARGS,debug_);

                res+="op ";
                res+=logic_set_priority(get_int_arg(0));
                res+=" ";
                res+=get_str_arg(1);
                res+=" ";
                res+=get_str_arg(2);
            }
            break;
        case _deop:
            {
                if(num_args!=3)
                    THROW_C_PROC_EXC(this_module,command_,c_proc_exc::E_NUM_ARGS,debug_);

                res+="deop ";
                res+=logic_set_priority(get_int_arg(0));
                res+=" ";
                res+=get_str_arg(1);
                res+=" ";
                res+=get_str_arg(2);
            }
            break;
        case _deop_ban_mask:
            {
                if(num_args!=4)
                    THROW_C_PROC_EXC(this_module,command_,c_proc_exc::E_NUM_ARGS,debug_);

                res+="deop_ban_mask ";
                res+=logic_set_priority(get_int_arg(0));
                res+=" ";
                res+=get_str_arg(1);
                res+=" ";
                res+=get_str_arg(2);
                res+=" ";
                res+=get_str_arg(3);
            }
            break;
        case _voice:
            {
                if(num_args!=3)
                    THROW_C_PROC_EXC(this_module,command_,c_proc_exc::E_NUM_ARGS,debug_);

                res+="voice ";
                res+=logic_set_priority(get_int_arg(0));
                res+=" ";
                res+=get_str_arg(1);
                res+=" ";
                res+=get_str_arg(2);
            }
            break;
        case _devoice:
            {
                if(num_args!=3)
                    THROW_C_PROC_EXC(this_module,command_,c_proc_exc::E_NUM_ARGS,debug_);

                res+="devoice ";
                res+=logic_set_priority(get_int_arg(0));
                res+=" ";
                res+=get_str_arg(1);
                res+=" ";
                res+=get_str_arg(2);
            }
            break;
        case _kick:
            if(num_args!=3)
                THROW_C_PROC_EXC(this_module,command_,c_proc_exc::E_NUM_ARGS,debug_);

            res+="kick ";
            res+=get_str_arg(0);
            res+=" ";
            res+=get_str_arg(1);
            res+=" ";
            res+=get_str_arg(2);
            break;
        case _msg:
            if(num_args!=3)
                THROW_C_PROC_EXC(this_module,command_,c_proc_exc::E_NUM_ARGS,debug_);

            res+="msg ";
            res+=logic_set_priority(get_int_arg(0));
            res+=" ";
            res+=get_str_arg(1);
            res+=" ";
            res+=get_str_arg(2);
            break;
        case _msgq:
            if(num_args!=4)
                THROW_C_PROC_EXC(this_module,command_,c_proc_exc::E_NUM_ARGS,debug_);

            res+="msgq ";
            res+=logic_set_priority(get_int_arg(0));
            res+=" ";
            res+=get_str_arg(1);
            res+=" ";
            res+=get_str_arg(2);
            res+=" ";
            res+=get_str_arg(3);
            break;
        case _if_in:
            if(num_args!=2)
                THROW_C_PROC_EXC(this_module,command_,c_proc_exc::E_NUM_ARGS,debug_);

            res+="if_in ";
            res+=get_str_arg(0);
            res+=" ";
            res+=get_str_arg(1);
            res+="\n";
            for(int i1=0; i1<indent; ++i1)
                res+="    ";
            res+="{";
            indent++;
            break;
        case _if_n_in:
            if(num_args!=2)
                THROW_C_PROC_EXC(this_module,command_,c_proc_exc::E_NUM_ARGS,debug_);

            res+="!if_in ";
            res+=get_str_arg(0);
            res+=" ";
            res+=get_str_arg(1);
            res+="\n";
            for(int i1=0; i1<indent; ++i1)
                res+="    ";
            res+="{";
            indent++;
            break;
        case _if_match:
            if(num_args!=2)
                THROW_C_PROC_EXC(this_module,command_,c_proc_exc::E_NUM_ARGS,debug_);

            res+="if_match ";
            res+=get_str_arg(0);
            res+=" ";
            res+=get_str_arg(1);
            res+="\n";
            for(int i1=0; i1<indent; ++i1)
                res+="    ";
            res+="{";
            indent++;
            break;
        case _if_match_case_insensitive:
            if(num_args!=2)
                THROW_C_PROC_EXC(this_module,command_,c_proc_exc::E_NUM_ARGS,debug_);

            res+="if_match_case_insensitive ";
            res+=get_str_arg(0);
            res+=" ";
            res+=get_str_arg(1);
            res+="\n";
            for(int i1=0; i1<indent; ++i1)
                res+="    ";
            res+="{";
            indent++;
            break;
        case _if_n_match:
            if(num_args!=2)
                THROW_C_PROC_EXC(this_module,command_,c_proc_exc::E_NUM_ARGS,debug_);

            res+="!if_match ";
            res+=get_str_arg(0);
            res+=" ";
            res+=get_str_arg(1);
            res+="\n";
            for(int i1=0; i1<indent; ++i1)
                res+="    ";
            res+="{";
            indent++;
            break;
        case _if_n_match_case_insensitive:
            if(num_args!=2)
                THROW_C_PROC_EXC(this_module,command_,c_proc_exc::E_NUM_ARGS,debug_);

            res+="!if_match_case_insensitive ";
            res+=get_str_arg(0);
            res+=" ";
            res+=get_str_arg(1);
            res+="\n";
            for(int i1=0; i1<indent; ++i1)
                res+="    ";
            res+="{";
            indent++;
            break;
        case _if_group:
            if(num_args!=2)
                THROW_C_PROC_EXC(this_module,command_,c_proc_exc::E_NUM_ARGS,debug_);

            res+="if_group ";
            res+=get_str_arg(0);
            res+=" ";
            res+=get_str_arg(1);
            res+="\n";
            for(int i1=0; i1<indent; ++i1)
                res+="    ";
            res+="{";
            indent++;
            break;
        case _if_n_group:
            if(num_args!=2)
                THROW_C_PROC_EXC(this_module,command_,c_proc_exc::E_NUM_ARGS,debug_);

            res+="!if_group ";
            res+=get_str_arg(0);
            res+=" ";
            res+=get_str_arg(1);
            res+="\n";
            for(int i1=0; i1<indent; ++i1)
                res+="    ";
            res+="{";
            indent++;
            break;
        case _return:
            if(num_args!=0)
                THROW_C_PROC_EXC(this_module,command_,c_proc_exc::E_NUM_ARGS,debug_);

            res+="return";
            break;
        case _end_bracket:
            if(num_args!=0)
                THROW_C_PROC_EXC(this_module,command_,c_proc_exc::E_NUM_ARGS,debug_);

            res="";
            --indent;
            for(int i1=0; i1<indent; ++i1)
                res+="    ";
            res+="}";
            break;
        case _work:
            if(num_args!=0)
                THROW_C_PROC_EXC(this_module,command_,c_proc_exc::E_NUM_ARGS,debug_);

            res+="work";
            break;
        case _timer_once:
            if(num_args!=3)
                THROW_C_PROC_EXC(this_module,command_,c_proc_exc::E_NUM_ARGS,debug_);

            res+="timer_once ";
            res+=get_str_arg(0);
            res+=" ";
            res+=get_timer_duration(get_int_arg(1));
            res+=" ";
            res+=get_str_arg(2);
            break;
        case _timer_every:
            if(num_args!=3)
                THROW_C_PROC_EXC(this_module,command_,c_proc_exc::E_NUM_ARGS,debug_);

            res+="timer_every ";
            res+=get_str_arg(0);
            res+=" ";
            res+=get_timer_duration(get_int_arg(1));
            res+=" ";
            res+=get_str_arg(2);
            break;
            break;
        case _kill_timers:
            if(num_args!=1)
                THROW_C_PROC_EXC(this_module,command_,c_proc_exc::E_NUM_ARGS,debug_);

            res+="kill_timers ";
            res+=get_str_arg(0);
            break;
        case _execute:
            if(num_args!=1)
                THROW_C_PROC_EXC(this_module,command_,c_proc_exc::E_NUM_ARGS,debug_);

            res+="execute ";
            res+=get_str_arg(0);
            break;
        case _remote_execute:
            if(num_args!=3)
                THROW_C_PROC_EXC(this_module,command_,c_proc_exc::E_NUM_ARGS,debug_);

            res+="remote_execute ";
            res+=get_str_arg(0);
            res+=" ";
            res+=get_str_arg(1);
            res+=" ";
            res+=get_str_arg(2);
            break;
        case _smtp:
            if(num_args!=6)
                THROW_C_PROC_EXC(this_module,command_,c_proc_exc::E_NUM_ARGS,debug_);

            res+="SMTP\n";
            for(int i1=0; i1<indent; ++i1)
                res+="    ";
            res+="{\n";
            indent++;

            for(int i1=0; i1<indent; ++i1)
                res+="    ";
            res+="server ";
            res+=get_str_arg(0);
            res+="\n";

            for(int i1=0; i1<indent; ++i1)
                res+="    ";
            res+="port ";
            res+=tos(get_int_arg(1));
            res+="\n";

            for(int i1=0; i1<indent; ++i1)
                res+="    ";
            res+="HELO ";
            res+=get_str_arg(2);
            res+="\n";

            for(int i1=0; i1<indent; ++i1)
                res+="    ";
            res+="MAIL_FROM ";
            res+=get_str_arg(3);
            res+="\n";

            for(int i1=0; i1<indent; ++i1)
                res+="    ";
            res+="RCPT_TO ";
            res+=get_str_arg(4);
            res+="\n";

            for(int i1=0; i1<indent; ++i1)
                res+="    ";
            res+="DATA\n";

            {
                string data=get_str_arg(5);
                string part;
                for(size_t i2=0; i2<data.length(); ++i2) {
                    if('\r'==data[i2]) {
                        for(int i1=0; i1<indent; ++i1)
                            res+="    ";
                        if(!part.empty())
                            res+=part;
                        else
                            res+="-";
                        res+="\n";
                        part="";
                        continue;
                    }
                    if('\n'==data[i2])
                        continue;
                    part+=data[i2];
                }
            }
            --indent;
            for(int i1=0; i1<indent; ++i1)
                res+="    ";
            res+="}";
            break;
        case _net_send:
            if(num_args!=2)
                THROW_C_PROC_EXC(this_module,command_,c_proc_exc::E_NUM_ARGS,debug_);

            res+="NET_SEND ";
            res+=get_str_arg(0);
            res+=" ";
            res+=get_str_arg(1);
            break;
        case _log:
            if(num_args!=1)
                THROW_C_PROC_EXC(this_module,command_,c_proc_exc::E_NUM_ARGS,debug_);

            res+="LOG ";
            res+=get_str_arg(0);
            break;
        case _join:
            if(num_args!=2)
                THROW_C_PROC_EXC(this_module,command_,c_proc_exc::E_NUM_ARGS,debug_);

            res+="join ";
            res+=get_str_arg(0);
            res+=" ";
            res+=get_str_arg(1);
            break;
        case _part:
            if(num_args!=2)
                THROW_C_PROC_EXC(this_module,command_,c_proc_exc::E_NUM_ARGS,debug_);

            res+="part ";
            res+=get_str_arg(0);
            res+=" ";
            res+=get_str_arg(1);
            break;
        case _disconnect:
            if(num_args!=0)
                THROW_C_PROC_EXC(this_module,command_,c_proc_exc::E_NUM_ARGS,debug_);

            res+="disconnect";
            break;
        case _try_connect:
            if(num_args!=0)
                THROW_C_PROC_EXC(this_module,command_,c_proc_exc::E_NUM_ARGS,debug_);

            res+="try_connect";
            break;
        case _irc_server:
            if(num_args!=2)
                THROW_C_PROC_EXC(this_module,command_,c_proc_exc::E_NUM_ARGS,debug_);

            res+="irc_server ";
            res+=get_str_arg(0);
            res+=" ";
            res+=tos(get_int_arg(1));
            break;
        case _irc_server_ex:
            if(num_args!=4)
                THROW_C_PROC_EXC(this_module,command_,c_proc_exc::E_NUM_ARGS,debug_);

            res+="irc_server_ex ";
            res+=get_str_arg(0);
            res+=" ";
            res+=get_str_arg(1);
            res+=" ";
            res+=get_str_arg(2);
            res+=" ";
            res+=tos(get_int_arg(3));
            break;
        case _bot_nick:
            if(num_args!=0)
                THROW_C_PROC_EXC(this_module,command_,c_proc_exc::E_NUM_ARGS,debug_);

            res+="bot_nick ";
            {
                vector<string>::const_iterator ii=strings.begin();
                while(ii!=strings.end()) {
                    res+=*ii;
                    ii++;
                    if(ii!=strings.end())
                        res+=" ";
                }
            }
            break;
        case _delete_irc_servers:
            if(num_args!=1)
                THROW_C_PROC_EXC(this_module,command_,c_proc_exc::E_NUM_ARGS,debug_);

            res+="delete_irc_servers ";
            res+=get_str_arg(0);
            break;
        case _delete_nicks:
            if(num_args!=1)
                THROW_C_PROC_EXC(this_module,command_,c_proc_exc::E_NUM_ARGS,debug_);

            res+="delete_nicks ";
            res+=get_str_arg(0);
            break;
        case _change_nick:
            if(num_args!=1)
                THROW_C_PROC_EXC(this_module,command_,c_proc_exc::E_NUM_ARGS,debug_);

            res+="change_nick ";
            res+=get_str_arg(0);
            break;
        case _bot_ident:
            if(num_args!=1)
                THROW_C_PROC_EXC(this_module,command_,c_proc_exc::E_NUM_ARGS,debug_);

            res+="bot_ident ";
            res+=get_str_arg(0);
            break;
        case _bot_ident_ipv6:
            if(num_args!=1)
                THROW_C_PROC_EXC(this_module,command_,c_proc_exc::E_NUM_ARGS,debug_);

            res+="bot_ident_ipv6 ";
            res+=get_str_arg(0);
            break;
        case _bot_fullname:
            if(num_args!=1)
                THROW_C_PROC_EXC(this_module,command_,c_proc_exc::E_NUM_ARGS,debug_);

            res+="bot_fullname ";
            res+=get_str_arg(0);
            break;
        case _bot_auth:
            if(num_args!=0)
                THROW_C_PROC_EXC(this_module,command_,c_proc_exc::E_NUM_ARGS,debug_);

            res+="bot_auth";
            break;
        case _allow_redirect:
            if(num_args!=1)
                THROW_C_PROC_EXC(this_module,command_,c_proc_exc::E_NUM_ARGS,debug_);

            res+="allow_redirect ";
            res+=tos(get_int_arg(0));
            break;
        case _if_error:
            if(num_args!=0)
                THROW_C_PROC_EXC(this_module,command_,c_proc_exc::E_NUM_ARGS,debug_);

            res+="if_error";
            res+="\n";
            for(int i1=0; i1<indent; ++i1)
                res+="    ";
            res+="{";
            indent++;
            break;
        case _wait_if_error:
            if(num_args!=0)
                THROW_C_PROC_EXC(this_module,command_,c_proc_exc::E_NUM_ARGS,debug_);

            res+="wait_if_error";
            res+="\n";
            for(int i1=0; i1<indent; ++i1)
                res+="    ";
            res+="{";
            indent++;
            break;
        case _label:
            if(num_args!=0)
                THROW_C_PROC_EXC(this_module,command_,c_proc_exc::E_NUM_ARGS,debug_);
            res="";
            break;
        case _real_label:
            if(num_args!=1)
                THROW_C_PROC_EXC(this_module,command_,c_proc_exc::E_NUM_ARGS,debug_);

            res+="label ";
            res+=get_str_arg(0);
            break;
        case _goto:
            if(num_args!=0)
                THROW_C_PROC_EXC(this_module,command_,c_proc_exc::E_NUM_ARGS,debug_);
            res="";
            break;
        case _real_goto:
            if(num_args!=1)
                THROW_C_PROC_EXC(this_module,command_,c_proc_exc::E_NUM_ARGS,debug_);

            res+="goto ";
            res+=get_str_arg(0);
            break;
        case _sleep_:
            if(num_args!=1)
                THROW_C_PROC_EXC(this_module,command_,c_proc_exc::E_NUM_ARGS,debug_);

            res+="sleep ";
            res+=tos(get_int_arg(0));
            break;
        case _restart:
            if(num_args!=0)
                THROW_C_PROC_EXC(this_module,command_,c_proc_exc::E_NUM_ARGS,debug_);

            res+="restart";
            break;
        case _raw:
            if(num_args!=2)
                THROW_C_PROC_EXC(this_module,command_,c_proc_exc::E_NUM_ARGS,debug_);

            res+="raw ";
            res+=logic_set_priority(get_int_arg(0));
            res+=" ";
            res+=get_str_arg(1);
            break;
        case _admin_msg:
            if(num_args!=3)
                THROW_C_PROC_EXC(this_module,command_,c_proc_exc::E_NUM_ARGS,debug_);

            res+="admin_msg ";
            res+=get_str_arg(0);
            res+=" ";
            res+=get_str_arg(1);
            res+=" ";
            res+=get_str_arg(2);
            break;
        case _notice:
            if(num_args!=3)
                THROW_C_PROC_EXC(this_module,command_,c_proc_exc::E_NUM_ARGS,debug_);

            res+="notice ";
            res+=logic_set_priority(get_int_arg(0));
            res+=" ";
            res+=get_str_arg(1);
            res+=" ";
            res+=get_str_arg(2);
            break;
        case _noticeq:
            if(num_args!=4)
                THROW_C_PROC_EXC(this_module,command_,c_proc_exc::E_NUM_ARGS,debug_);

            res+="noticeq ";
            res+=logic_set_priority(get_int_arg(0));
            res+=" ";
            res+=get_str_arg(1);
            res+=" ";
            res+=get_str_arg(2);
            res+=" ";
            res+=get_str_arg(3);
            break;
        case _unban_mask:
            if(num_args!=2)
                THROW_C_PROC_EXC(this_module,command_,c_proc_exc::E_NUM_ARGS,debug_);

            res+="unban_mask ";
            res+=get_str_arg(0);
            res+=" ";
            res+=get_str_arg(1);
            break;
        case _ban_mask:
            if(num_args!=2)
                THROW_C_PROC_EXC(this_module,command_,c_proc_exc::E_NUM_ARGS,debug_);

            res+="ban_mask ";
            res+=get_str_arg(0);
            res+=" ";
            res+=get_str_arg(1);
            break;
        case _dynamic_ban:
            if(num_args!=4)
                THROW_C_PROC_EXC(this_module,command_,c_proc_exc::E_NUM_ARGS,debug_);

            res+="dynamic_ban ";
            res+=get_str_arg(0);
            res+=" ";
            res+=get_str_arg(1);
            res+=" ";
            res+=get_str_arg(2);
            res+=" ";
            res+=get_str_arg(3);
            break;
        case _process_on_banned:
            if(num_args!=2)
                THROW_C_PROC_EXC(this_module,command_,c_proc_exc::E_NUM_ARGS,debug_);

            res+="process_on_banned ";
            res+=get_str_arg(0);
            res+=" ";
            res+=get_str_arg(1);
            break;
        case _check_dynamic_bans:
            if(num_args!=4)
                THROW_C_PROC_EXC(this_module,command_,c_proc_exc::E_NUM_ARGS,debug_);

            res+=get_str_arg(3);
            res+="=";
            res+="check_dynamic_bans ";
            res+=get_str_arg(0);
            res+=" ";
            res+=get_str_arg(1);
            res+=" ";
            res+=get_str_arg(2);
            break;
        case _dcc_server:
            if(num_args!=4)
                THROW_C_PROC_EXC(this_module,command_,c_proc_exc::E_NUM_ARGS,debug_);

            res+="dcc_server ";
            res+=tos(get_int_arg(0));
            res+=" ";
            res+=get_str_arg(1);
            res+=" ";
            res+=tos(get_int_arg(2));
            res+=" ";
            res+=get_str_arg(3);
            break;
        case _dcc_server_ipv6:
            if(num_args!=4)
                THROW_C_PROC_EXC(this_module,command_,c_proc_exc::E_NUM_ARGS,debug_);

            res+="dcc_server_ipv6 ";
            res+=tos(get_int_arg(0));
            res+=" ";
            res+=get_str_arg(1);
            res+=" ";
            res+=tos(get_int_arg(2));
            res+=" ";
            res+=get_str_arg(3);
            break;
        case _telnet_server:
            if(num_args!=2)
                THROW_C_PROC_EXC(this_module,command_,c_proc_exc::E_NUM_ARGS,debug_);

            res+="telnet_server ";
            res+=get_str_arg(0);
            res+=" ";
            res+=tos(get_int_arg(1));
            break;
        case _telnet_server_ipv6:
            if(num_args!=2)
                THROW_C_PROC_EXC(this_module,command_,c_proc_exc::E_NUM_ARGS,debug_);

            res+="telnet_server_ipv6 ";
            res+=get_str_arg(0);
            res+=" ";
            res+=tos(get_int_arg(1));
            break;
        case _ssl_server:
            if(num_args!=3)
                THROW_C_PROC_EXC(this_module,command_,c_proc_exc::E_NUM_ARGS,debug_);

            res+="ssl_server ";
            res+=tos(get_int_arg(0));
            res+=" ";
            res+=get_str_arg(1);
            res+=" ";
            res+=tos(get_int_arg(2));
            break;
        case _ident:
            if(num_args!=2)
                THROW_C_PROC_EXC(this_module,command_,c_proc_exc::E_NUM_ARGS,debug_);

            res+=get_str_arg(1);
            res+="=";
            res+="ident ";
            res+=get_str_arg(0);
            break;
        case _host:
            if(num_args!=2)
                THROW_C_PROC_EXC(this_module,command_,c_proc_exc::E_NUM_ARGS,debug_);

            res+=get_str_arg(1);
            res+="=";
            res+="host ";
            res+=get_str_arg(0);
            break;
        case _get_chan_mode:
            if(num_args!=2)
                THROW_C_PROC_EXC(this_module,command_,c_proc_exc::E_NUM_ARGS,debug_);

            res+=get_str_arg(1);
            res+="=";
            res+="get_chan_mode ";
            res+=get_str_arg(0);
            break;
        case _topic:
            if(num_args!=2)
                THROW_C_PROC_EXC(this_module,command_,c_proc_exc::E_NUM_ARGS,debug_);

            res+="topic ";
            res+=get_str_arg(0);
            res+=" ";
            res+=get_str_arg(1);
            break;
        case _get_chan_topic:
            if(num_args!=2)
                THROW_C_PROC_EXC(this_module,command_,c_proc_exc::E_NUM_ARGS,debug_);

            res+=get_str_arg(1);
            res+="=";
            res+="get_chan_topic ";
            res+=get_str_arg(0);
            break;
        case _script:
            if(num_args!=4)
                THROW_C_PROC_EXC(this_module,command_,c_proc_exc::E_NUM_ARGS,debug_);

            res+="SCRIPT ";
            res+=get_str_arg(0);
            res+=" ";
            res+=tos(get_int_arg(1));
            res+=" ";
            res+=get_str_arg(2);
            res+=" ";
            res+=get_str_arg(3);
            break;
        case _link:
            if(num_args!=6)
                THROW_C_PROC_EXC(this_module,command_,c_proc_exc::E_NUM_ARGS,debug_);

            res+="link ";
            res+=get_str_arg(0);
            res+=" ";
            res+=get_str_arg(1);
            res+=" ";
            res+=get_str_arg(2);
            res+=" ";
            res+=get_str_arg(3);
            res+=" ";
            res+=tos(get_int_arg(4));
            res+=" ";
            res+=get_str_arg(5);
            break;
        case _chan_mode:
            if(num_args!=3)
                THROW_C_PROC_EXC(this_module,command_,c_proc_exc::E_NUM_ARGS,debug_);

            res+="chan_mode ";
            res+=logic_set_priority(get_int_arg(0));
            res+=" ";
            res+=get_str_arg(1);
            res+=" ";
            res+=get_str_arg(2);
            break;
        case _rem:
            if(num_args!=1)
                THROW_C_PROC_EXC(this_module,command_,c_proc_exc::E_NUM_ARGS,debug_);

            res+="rem ";
            res+=get_str_arg(0);
            break;
        case _load_module:
            if(num_args!=1)
                THROW_C_PROC_EXC(this_module,command_,c_proc_exc::E_NUM_ARGS,debug_);

            {
                res+="load_module ";
                res+=get_str_arg(0);
            }
            break;
        case _unload_module:
            if(num_args!=1)
                THROW_C_PROC_EXC(this_module,command_,c_proc_exc::E_NUM_ARGS,debug_);

            {
                res+="unload_module ";
                res+=get_str_arg(0);
            }
            break;
        case _bind:
            if(num_args!=6)
                THROW_C_PROC_EXC(this_module,command_,c_proc_exc::E_NUM_ARGS,debug_);

            {
                res+="bind ";
                res+=get_str_arg(0);
                res+=" ";
                res+=get_str_arg(1);
                res+=" ";
                res+=get_str_arg(2);
                res+=" ";
                res+=get_str_arg(3);
                res+=" ";
                res+=tos(get_int_arg(4));
                res+=" ";
                res+=get_str_arg(5);
            }
            break;
        case _unbind:
            if(num_args!=6)
                THROW_C_PROC_EXC(this_module,command_,c_proc_exc::E_NUM_ARGS,debug_);

            {
                res+="unbind ";
                res+=get_str_arg(0);
                res+=" ";
                res+=get_str_arg(1);
                res+=" ";
                res+=get_str_arg(2);
                res+=" ";
                res+=get_str_arg(3);
                res+=" ";
                res+=tos(get_int_arg(4));
                res+=" ";
                res+=get_str_arg(5);
            }
            break;
        case _load_bin_module:
            if(num_args!=2)
                THROW_C_PROC_EXC(this_module,command_,c_proc_exc::E_NUM_ARGS,debug_);

            {
                res+="load_bin_module ";
                res+=get_str_arg(0);
                res+=" ";
                res+=get_str_arg(1);
            }
            break;
        case _eval:
            if(num_args!=2)
                THROW_C_PROC_EXC(this_module,command_,c_proc_exc::E_NUM_ARGS,debug_);

            res+=get_str_arg(1);
            res+="=eval ";
            res+=get_str_arg(0);
            break;
        case _set:
            if(num_args!=2)
                THROW_C_PROC_EXC(this_module,command_,c_proc_exc::E_NUM_ARGS,debug_);

            res+=get_str_arg(1);
            res+="=set ";
            res+=get_str_arg(0);
            break;
        case _get_nick_mode:
            if(num_args!=3)
                THROW_C_PROC_EXC(this_module,command_,c_proc_exc::E_NUM_ARGS,debug_);

            res+=get_str_arg(2);
            res+="=get_nick_mode ";
            res+=get_str_arg(0);
            res+=" ";
            res+=get_str_arg(1);
            break;
        case _get_bot_nick:
            if(num_args!=1)
                THROW_C_PROC_EXC(this_module,command_,c_proc_exc::E_NUM_ARGS,debug_);

            res+=get_str_arg(0);
            res+="=get_bot_nick";
            break;
        case _get_event_id:
            if(num_args!=1)
                THROW_C_PROC_EXC(this_module,command_,c_proc_exc::E_NUM_ARGS,debug_);

            res+=get_str_arg(0);
            res+="=get_event_id";
            break;
        case _get_current_irc_server:
            if(num_args!=2)
                THROW_C_PROC_EXC(this_module,command_,c_proc_exc::E_NUM_ARGS,debug_);

            res+="get_current_irc_server ";
            res+=get_str_arg(0);
            res+=" ";
            res+=get_str_arg(1);
            break;
        case _get_bot_name:
            if(num_args!=1)
                THROW_C_PROC_EXC(this_module,command_,c_proc_exc::E_NUM_ARGS,debug_);

            res+=get_str_arg(0);
            res+="=get_bot_name";
            break;
        default:
            THROW_C_PROC_EXC(this_module,command_,c_proc_exc::E_INV_COMMAND,debug_);
    }
    return res;
}

