/***************************************************************************
                          c_proc_command.cpp  -  description
                             -------------------
    begin                : Thu May 10 2007
    copyright            : (C) 2004-2007 by Marian VooDooMan Meravy
    email                : ghostvoodooman -a.T= users (dot) sourceforge.net
 ***************************************************************************/

/* $Id: c_proc_command.cpp,v 1.14 2008/03/11 21:11:47 ghostvoodooman Exp $ */

/***************************************************************************

VooDoo cIRCle - an IRC (ro)bot
Copyright (C) 2004-2008 by Marian VooDooMan Meravy
  (ghostvoodooman -a.T= users (dot) sourceforge.net)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************/

/*!
    \file
    \brief Class 'c_proc_command'
    \version $Id: c_proc_command.cpp,v 1.14 2008/03/11 21:11:47 ghostvoodooman Exp $
*/

#if defined(HAVE_CONFIG_H)
#   include <config.h>
#endif

#include "../utils/memleak.h"
#include "../utils/assert.h"
#include "../utils.h"
#include "../params.h"

#include "c_proc_command.h"
#include "../compat.h"
#include "c_proc_exc.h"
#include "../utils/precomp.h"

#if defined(_MSC_VER) || defined(__BORLANDC__)
#   pragma hdrstop
#endif

c_proc_command::c_proc_command(c_module* this_module, int command, int line, const string& debug)
{
    clear();
    command_=command;
    line_=line;
    debug_=debug;
    this_module_=this_module;
}

void c_proc_command::parse_irc_type_flags(vector<string>& flags, const string& irc_type)
{
    flags.clear();
    string current;
    bool in=false;
    for(size_t i1=0; i1<irc_type.length(); ++i1) {
        if(!in && irc_type[i1]=='@') {
            in=true;
            current="@";
            continue;
        }
        if(in && irc_type[i1]=='@') {
            in=false;
            current+="@";
            flags.push_back(current);
            current="";
            continue;
        }
        if(in)
            current+=(string)""+(irc_type[i1]);
    }
}

void c_proc_command::clear()
{
    line_=0;
    debug_="";
    command_=0;
    _goto_line=0;
    label="";
    strings.clear();
    for(int i1=0; i1<CMD_MAX_ARGS; ++i1)
        args[i1]=ARG_TYPE_NOT_SET;
    num_args=0;
    this_module_=NULL;
}

c_proc_command::c_proc_command(const c_proc_command& orig)
{
    line_=orig.line_;
    debug_=orig.debug_;
    command_=orig.command_;
    _goto_line=orig._goto_line;
    label=orig.label;
    strings=orig.strings;
    for(int i1=0; i1<CMD_MAX_ARGS; ++i1) {
        str_args[i1]=orig.str_args[i1];
        int_args[i1]=orig.int_args[i1];
        args[i1]=orig.args[i1];
    }
    num_args=orig.num_args;
    this_module_=orig.this_module_;
}

bool c_proc_command::set_next_arg(const string& arg, int arg_num)
{
    if(num_args>=CMD_MAX_ARGS)
        THROW_C_PROC_EXC(this_module_,command_,c_proc_exc::E_ARG_CNT_EXC,debug_);
    if(-1==arg_num) {
        num_args++;
        args[num_args-1]=ARG_TYPE_STR;
        str_args[num_args-1]=arg;
        return true;
    } else {
        if(arg_num>=CMD_MAX_ARGS)
            THROW_C_PROC_EXC(this_module_,command_,c_proc_exc::E_ARG_CNT_EXC,debug_);
        if(num_args<arg_num+1)
            num_args=arg_num+1;
        args[arg_num]=ARG_TYPE_STR;
        str_args[arg_num]=arg;
        return true;
    }
}

bool c_proc_command::set_next_arg(long arg, int arg_num)
{
    if(num_args>=CMD_MAX_ARGS)
        THROW_C_PROC_EXC(this_module_,command_,c_proc_exc::E_ARG_CNT_EXC,debug_);
    if(-1==arg_num) {
        num_args++;
        args[num_args-1]=ARG_TYPE_INT;
        int_args[num_args-1]=arg;
        return true;
    } else {
        if(arg_num>=CMD_MAX_ARGS)
            THROW_C_PROC_EXC(this_module_,command_,c_proc_exc::E_ARG_CNT_EXC,debug_);
        if(num_args<arg_num+1)
            num_args=arg_num+1;
        args[arg_num]=ARG_TYPE_INT;
        int_args[arg_num]=arg;
        return true;
    }
}

int c_proc_command::get_num_args()
{
    return num_args;
}

long c_proc_command::get_int_arg(int arg_num)
{
    if(arg_num>num_args)
        THROW_C_PROC_EXC(this_module_,command_,c_proc_exc::E_GET_NUM_ARGS,debug_);
    if(args[arg_num]!=ARG_TYPE_INT)
        THROW_C_PROC_EXC(this_module_,command_,c_proc_exc::E_INV_ARG_TYPE,debug_);
    return int_args[arg_num];
}

string c_proc_command::get_str_arg(int arg_num)
{
    if(arg_num>num_args)
        THROW_C_PROC_EXC(this_module_,command_,c_proc_exc::E_GET_NUM_ARGS,debug_);
    if(args[arg_num]!=ARG_TYPE_STR)
        THROW_C_PROC_EXC(this_module_,command_,c_proc_exc::E_INV_ARG_TYPE,debug_);
    return str_args[arg_num];
}

string c_proc_command::get_channel()
{
    if(_op==command_ || _deop==command_ || _voice==command_ || _devoice==command_)
        return get_str_arg(1);
    THROW_C_PROC_EXC(this_module_,command_,c_proc_exc::E_INV_OPERATION,debug_);
}

string c_proc_command::get_nick()
{
    if(_op==command_ || _deop==command_ || _voice==command_ || _devoice==command_)
        return get_str_arg(2);
    THROW_C_PROC_EXC(this_module_,command_,c_proc_exc::E_INV_OPERATION,debug_);
}

string c_proc_command::get_label()
{
    if(_real_label==command_)
        return get_str_arg(0);
    THROW_C_PROC_EXC(this_module_,command_,c_proc_exc::E_INV_OPERATION,debug_);
}

int c_proc_command::get_line()
{
    return line_;
}

void c_proc_command::set_goto_line(int goto_line)
{
    _goto_line=goto_line;
}

int c_proc_command::get_command()
{
    return command_;
}

const string& c_proc_command::get_debug()
{
    return debug_;
}

int c_proc_command::get_goto_line()
{
    return _goto_line;
}

