/***************************************************************************
                          c_proc_command.h  -  description
                             -------------------
    begin                : Thu May 10 2007
    copyright            : (C) 2004-2007 by Marian VooDooMan Meravy
    email                : ghostvoodooman -a.T= users (dot) sourceforge.net
 ***************************************************************************/

/* $Id: c_proc_command.h,v 1.16 2008/08/23 15:10:46 ghostvoodooman Exp $ */

/***************************************************************************

VooDoo cIRCle - an IRC (ro)bot
Copyright (C) 2004-2008 by Marian VooDooMan Meravy
  (ghostvoodooman -a.T= users (dot) sourceforge.net)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************/

/*!
    \file
    \brief Class 'c_proc_command'
    \version $Id: c_proc_command.h,v 1.16 2008/08/23 15:10:46 ghostvoodooman Exp $
*/

#ifndef C_PROC_COMMAND_H__
#define C_PROC_COMMAND_H__

#if defined(HAVE_CONFIG_H)
#   include <config.h>
#endif

#include <string>
#include <vector>
#include <stack>
#include <list>
#include <map>
#include <algorithm>

#include "../utils/memleak.h"

#include "../params.h"

struct c_eval_iteration;
class c_var_array;
class c_module;

/*!
    \brief Stores one command_ in the procedure
    \author VooDooMan
    \version 2
    \date 2004, 2005-10-08
*/
class c_proc_command
{
    int line_;                      //!< Line number

    string debug_;                  //!< Debug string

/*!
    \def _op
    \brief "op" command_, argument list: [int]priority, [str]channel, [str]nick
*/
#define _op             1
/*!
    \def _deop
    \brief "deop" command_, argument list: [int]priority, [str]channel, [str]nick
*/
#define _deop           2
/*!
    \def _voice
    \brief "voice" command_, argument list: [int]priority, [str]channel, [str]nick
*/
#define _voice          5
/*!
    \def _devoice
    \brief "devoice" command_, argument list: [int]priority, [str]channel, [str]nick
*/
#define _devoice        6
/*!
    \def _kick
    \brief "kick" command_, argument list: [str]channel, [str]nick, [str]message
*/
#define _kick           7
/*!
    \def _msg
    \brief "msg" command_, argument list: [int]priority, [str]target, [str]message
*/
#define _msg            8
/*!
    \def _if_match
    \brief "if_match" command_, argument list: [str]left, [str]right
*/
#define _if_match       9
/*!
    \def _if_group
    \brief "if_group" command_, argument list: [str]left, [str]right
*/
#define _if_group       10
/*!
    \def _return
    \brief "return" command_, argument list: (none)
*/
#define _return         11
/*!
    \def _timer_once
    \brief "timer_once" command_, argument list: [str]name, [int]seconds, [str]proc
*/
#define _timer_once     12
/*!
    \def _timer_every
    \brief "timer_every" command_, argument list: [str]name, [int]seconds, [str]proc
*/
#define _timer_every    13
/*!
    \def _execute
    \brief "execute" command_, argument list: [str]function call string
*/
#define _execute        14
/*!
    \def _smtp
    \brief "SMTP" command_, argument list: [str]smtp server, [int]smtp port, [str]HELO, [str]MAIL FROM, [str]RCPT TO, [str]DATA lines, delimitered by "\r\n", without ending "."
*/
#define _smtp           15
/*!
    \def _net_send
    \brief "NET_SEND" command_, argument list: [str]target, [str]message
*/
#define _net_send       16
/*!
    \def _log
    \brief "LOG" command_, argument list: [str]message
*/
#define _log            17
/*!
    \def _join
    \brief "join" command_, argument list: [str]channel, [str]key
*/
#define _join           18
/*!
    \def _part
    \brief "part" command_, argument list: [str]channel, [str]message
*/
#define _part           19
/*!
    \def _disconnect
    \brief "disconnect" command_, argument list: (none)
*/
#define _disconnect     20
/*!
    \def _try_connect
    \brief "try_connect" command_, argument list: (none)
*/
#define _try_connect    21
/*!
    \def _bot_nick
    \brief "bot_nick" command_, argument list: (none) = list of nicks is in "strings" vector
*/
#define _bot_nick       22
/*!
    \def _bot_ident
    \brief "bot_ident" command_, argument list: [str]ident
*/
#define _bot_ident      23
/*!
    \def _bot_fullname
    \brief "bot_fullname" command_, argument list: [str]full name
*/
#define _bot_fullname   24
/*!
    \def _bot_auth
    \brief "bot_auth" command_, argument list: (none)
*/
#define _bot_auth       25
/*!
    \def _allow_redirect
    \brief "allow_redirect" command_, argument list: [int](bool)yes?
*/
#define _allow_redirect      26
/*!
    \def _wait_if_error
    \brief "wait_if_error" command_, argument list: (none)
*/
#define _wait_if_error           27
/*!
    \def _sleep_
    \brief "sleep" command_, argument list: [int]seconds
*/
#define _sleep_         28
/*!
    \def _restart
    \brief "restart" command_, argument list: (none)
*/
#define _restart        29
/*!
    \def _notice
    \brief "notice" command_, argument list: [int]priority, [str]target, [str]message
*/
#define _notice         30
/*!
    \def _unban_mask
    \brief "unban_mask" command_, argument list: [str]channel, [str]mask
*/
#define _unban_mask     31
/*!
    \def _ban_mask
    \brief "ban_mask" command_, argument list: [str]channel, [str]mask
*/
#define _ban_mask       32
/*!
    \def _dcc_server
    \brief "dcc_server" command_, argument list: [int] group, [str]bind ip, [int]port, [str]server type
*/
#define _dcc_server     33
/*!
    \def _if_error
    \brief "if_error" command_, argument list: (none)
*/
#define _if_error       34
/*!
    \def _real_label
    \brief Working value (for _goto), argument list: [str]label
*/
#define _real_label     35
/*!
    \def _real_goto
    \brief "goto" command_, argument list: [str]label
*/
#define _real_goto      36
/*!
    \def _end_bracket
    \brief Working value (for converting function back to text), argument list: (none)
*/
#define _end_bracket    37
/*!
    \def _ident
    \brief "ident" command_, argument list: [str]nick, [str]result variable
*/
#define _ident          38
/*!
    \def _host
    \brief "host" command_, argument list: [str]nick, [str]result variable
*/
#define _host           39
/*!
    \def _script
    \brief "SCRIPT" command_, argument list: [str]type, [int]script number, [str]channel, [str]params
*/
#define _script         40
/*!
    \def _if_match_case_insensitive
    \brief "if_match_case_insensitive" command_, argument list: [str]left, [str]right
*/
#define _if_match_case_insensitive  41
/*!
    \def _telnet_server
    \brief "telnet_server" command_, argument list: [str]bind ip, [int]port
*/
#define _telnet_server  42
/*!
    \def _link
    \brief "link" command_, argument list: [str]type, [str]botname, [str]bind ip, [str]host, [int]port, [str]unlink proc
*/
#define _link           43
/*!
    \def _work
    \brief "work" command_, argument list: (none)
*/
#define _work           44
/*!
    \def _if_in
    \brief "if_in" command_, argument list: [str]left, [str]right
*/
#define _if_in          45
/*!
    \def _if_n_in
    \brief "!if_in" command_, argument list: [str]left, [str]right
*/
#define _if_n_in        46
/*!
    \def _if_n_match
    \brief "!if_match" command_, argument list: [str]left, [str]right
*/
#define _if_n_match     47
/*!
    \def _if_n_group
    \brief "!if_group" command_, argument list: [str]left, [str]right
*/
#define _if_n_group     48
/*!
    \def _if_n_match_case_insensitive
    \brief "!if_match_case_insensitive" command_, argument list: [str]left, [str]right
*/
#define _if_n_match_case_insensitive  49
/*!
    \def _chan_mode
    \brief "chan_mode" command_, argument list: [int]priority, [str]channel, [str]mode
*/
#define _chan_mode      50
/*!
    \def _msgq
    \brief "msgq" command_, argument list: [int]priority, [str]target, [str]message prefix, [str]quoted message
*/
#define _msgq           51
/*!
    \def _change_nick
    \brief "change_nick" command_, [str]new nick
*/
#define _change_nick    52
/*!
    \def _noticeq
    \brief "noticeq" command_, argument list: [int]priority, [str]target, [str]message prefix, [str]quoted message
*/
#define _noticeq        53
/*!
    \def _irc_server
    \brief "irc_server" command_, argument list: [str]host, [int]port
*/
#define _irc_server     54
/*!
    \def _kill_timers
    \brief "kill_timers" command_, argument list: [str]mask
*/
#define _kill_timers    55
/*!
    \def _get_chan_mode
    \brief "get_chan_mode" command_, argument list: [str]channel, [str]result variable
*/
#define _get_chan_mode  56
/*!
    \def _get_chan_topic
    \brief "get_chan_topic" command_, argument list: [str]channel, [str]result variable
*/
#define _get_chan_topic 57
/*!
    \def _topic
    \brief "topic" command_, argument list: [str]channel, [str]new topic
*/
#define _topic          58
/*!
    \def _dynamic_ban
    \brief "dynamic_ban" command_, argument list: [str]channel, [str]prefix, [str]mask, [str]reason
*/
#define _dynamic_ban    59
/*!
    \def _check_dynamic_bans
    \brief "check_dynamic_bans" command_, argument list: [str]channel, [str]nick, [str]reason variable, [str]result variable
*/
#define _check_dynamic_bans 60
/*!
    \def _process_on_banned
    \brief "process_on_banned" command_, argument list: [str]channel, [str]mask
*/
#define _process_on_banned  61
/*!
    \def _telnet_server_ipv6
    \brief "telnet_server_ipv6" command_, argument list: [str]ip, [int]port
*/
#define _telnet_server_ipv6 62
/*!
    \def _dcc_server_ipv6
    \brief "dcc_server_ipv6" command_, argument list: [int] group, [str]bind ip, [int]port, [str]server type
*/
#define _dcc_server_ipv6    63

// BOTNET version 2
/*!
    \def _raw
    \brief "raw" command_, argument list: [int]priority, [str]raw command_
    \warning Botnet protocol version 2 required for replication!
*/
#define _raw            64                          //!< "raw" command_

// BOTNET version 3
/*!
    \def _remote_execute
    \brief "remote_execute" command_, argument list: [str]bot name, [str]call string, [str]result variable
    \warning Botnet protocol version 3 required for replication!
*/
#define _remote_execute     65

// BOTNET version 4
/*!
    \def _bot_ident_ipv6
    \brief "bot_ident_ipv6" command_, argument list: [str]ident
    \warning Botnet protocol version 4 required for replication!
*/
#define _bot_ident_ipv6     66

// BOTNET version 6
/*!
    \def _delete_irc_servers
    \brief "delete_irc_servers" command_, argument list: [str]mask
    \warning Botnet protocol version 6 required for replication!
*/
#define _delete_irc_servers 67

// BOTNET version 6
/*!
    \def _delete_nicks
    \brief "delete_nicks" command_, argument list: [str]mask
    \warning Botnet protocol version 6 required for replication!
*/
#define _delete_nicks       68

// BOTNET version 10
/*!
    \def _admin_msg
    \brief "admin_msg" command_, argument list: [str]type, [str]mask, [str]message
    \warning Botnet protocol version 10 required for replication!
*/
#define _admin_msg          69

// BOTNET version 12
/*!
    \def _rem
    \brief "rem" command_, argument list: [str]remark
    \warning Botnet protocol version 12 required for replication!
*/
#define _rem                70

// BOTNET version 13
/*!
    \def _irc_server_ex
    \brief "irc_server_ex" command_, argument list: [str]flags, [str]bind ip, [str]host, [int]port
    \warning Botnet protocol version 13 required for replication!
*/
#define _irc_server_ex      71

// BOTNET version 1001
/*!
    \def _set
    \brief Expression evaluation, argument list: [str]expression, [str]result variable
    \warning Botnet protocol version 1001 required for replication!
*/
#define _set                72

// BOTNET version 1001
/*!
    \def _load_module
    \brief "load_module" command, argument list: [str]module name
    \warning Botnet protocol version 1001 required for replication!
*/
#define _load_module        73

// BOTNET version 1001
/*!
    \def _unload_module
    \brief "unload_module" command, argument list: [str]module name
    \warning Botnet protocol version 1001 required for replication!
*/
#define _unload_module      74

// BOTNET version 1001
/*!
    \def _bind
    \brief "bind" command, argument list: [str]type mask, [str]user mask, [str]channel mask, [str]command mask, [str]line mask, [int]before bool, [str]function/procedure name
    \warning Botnet protocol version 1001 required for replication!
*/
#define _bind               75

// BOTNET version 1001
/*!
    \def _unbind
    \brief "unbind" command, argument list: [str]type mask, [str]user mask, [str]channel mask, [str]command mask, [str]line mask, [int]before bool, [str]function/procedure name
    \warning Botnet protocol version 1001 required for replication!
*/
#define _unbind             76

// BOTNET version 1001
/*!
    \def _eval
    \brief "eval" command, argument list: [str]code_str, [str]result variable
    \warning Botnet protocol version 1001 required for replication!
*/
#define _eval               77

// BOTNET version 1001
/*!
    \def _get_nick_mode
    \brief "get_nick_mode" command, argument list: [str]channel, [str]nick, [str]result variable
    \warning Botnet protocol version 1001 required for replication!
*/
#define _get_nick_mode      78

// BOTNET version 1001
/*!
    \def _get_bot_nick
    \brief "get_bot_nick" command, argument list: [str]result variable
    \warning Botnet protocol version 1001 required for replication!
*/
#define _get_bot_nick       79

// BOTNET version 1001
/*!
    \def _get_event_id
    \brief "get_event_id" command, argument list: [str]result variable
    \warning Botnet protocol version 1001 required for replication!
*/
#define _get_event_id       80

// BOTNET version 1002
/*!
    \def _deop_ban_mask
    \brief "deop_ban_mask" command, argument list: [int]priority, [str]channel, [str]nick, [str]mask
    \warning Botnet protocol version 1002 required for replication!
*/
#define _deop_ban_mask      81

// BOTNET version 1006
/*!
    \def _ssl_server
    \brief "ssl_server" command_, argument list: [int]ip version, [str]bind ip, [int]port
    \warning Botnet protocol version 1006 required for replication!
*/
#define _ssl_server         82

// BOTNET version 1007
/*!
    \def _load_bin_module
    \brief "load_bin_module" command, argument list: [str]module name, [str]shared_object_name
    \warning Botnet protocol version 1007 required for replication!
*/
#define _load_bin_module        83

// BOTNET version 1008
/*!
    \def _get_current_irc_server
    \brief "get_current_irc_server" command, argument list: [str OUT]IRC server host, [int OUT]IRC server port
    \warning Botnet protocol version 1008 required for replication!
*/
#define _get_current_irc_server     84

// BOTNET version 1008
/*!
    \def _get_bot_name
    \brief "get_bot_name" command, argument list: [str OUT]result variable
    \warning Botnet protocol version 1008 required for replication!
*/
#define _get_bot_name               85

// BOTNET version 1009
/*!
    \def _bind_ex
    \brief "bind_ex" command, argument list: [str]type mask/regexp, [str]user mask/regexp, [str]channel mask/regexp, [str]command mask/regexp, [str]line mask/regexp, [int]before bool, [str]function/procedure name
    \warning Botnet protocol version 1009 required for replication!
*/

#define _unbind_ex              86
/*!
    \def _unbind_ex
    \brief "unbind_ex" command, argument list: [str]type mask/regexp, [str]user mask/regexp, [str]channel mask/regexp, [str]command mask/regexp, [str]line mask/regexp, [int]before bool, [str]function/procedure name
    \warning Botnet protocol version 1009 required for replication!
*/
#define _bind_ex                87

/*!
    \def _label
    \brief Working value (for _goto), argument list: (none)
*/
#define _label              1001
/*!
    \def _goto
    \brief Working value, argument list: (none)
*/
#define _goto               1002

    int command_;                                   //!< Command (\#define _xxx)

    int _goto_line;                                 //!< In some cases, this contains line number to jump to

    string label;                                   //!< In some cases, this contains text label (for _goto)

/*!
    \def CMD_MAX_ARGS
    \brief Maximum number of arguments
*/
#define CMD_MAX_ARGS    10
    string str_args[CMD_MAX_ARGS];                  //!< Arguments, used if they are strings
    long int_args[CMD_MAX_ARGS];                    //!< Arguments, used if they are ints

    enum arg_type                                   //!< Type of argument
    {
        ARG_TYPE_NOT_SET = 0,                       //!< Unset argument
        ARG_TYPE_STR,                               //!< Argument is string
        ARG_TYPE_INT,                               //!< Argument is int
    } args[CMD_MAX_ARGS];                           //!< Argumet types

    int num_args;                                   //!< Number of arguments

    c_module* this_module_;                         //!< Module

public:
    vector<string> strings;                         //!< In some cases, this contains list of bot's nicks, or list of globals

    /*!
        \brief Constructor
        \author VooDooMan
        \version 1
        \date 2005
        \param this_module Module
        \param command Command number
        \param line Line number
        \param debug Debug string
    */
    c_proc_command(c_module* this_module, int command, int line, const string& debug);

    /*!
        \brief Parses IRC type flags ("@something@@and_so_on@")
        \author VooDooMan
        \version 1
        \date 2005
        \param flags Returns parsed flags
        \param irc_type String to parse
    */
    static void parse_irc_type_flags(vector<string>& flags, const string& irc_type);

    /*!
        \brief Cleans up
        \author VooDooMan
        \version 1
        \date 2005
    */
    void clear();

    /*!
        \brief Copy constructor
        \author VooDooMan
        \version 1
        \date 2005
    */
    c_proc_command(const c_proc_command& orig);

    /*!
        \brief Sets next string argument
        \author VooDooMan
        \version 1
        \date 2005
        \param arg String
        \param arg_num Number of argument to set
        \return Returns true if ok
    */
    bool set_next_arg(const string& arg, int arg_num = -1);

    /*!
        \brief Sets next int argument
        \author VooDooMan
        \version 1
        \date 2005
        \param arg Integer
        \param arg_num Number of argument to set
        \return Returns true if ok
    */
    bool set_next_arg(long arg, int arg_num = -1);

    /*!
        \brief Gets number of  arguments
        \author VooDooMan
        \version 1
        \date 2006-02-16
        \return Returns number of arguments
    */
    int get_num_args();

    /*!
        \brief Gets int argument
        \author VooDooMan
        \version 1
        \date 2005
        \param arg_num Number of argument to get
        \return Returns argument
    */
    long get_int_arg(int arg_num);

    /*!
        \brief Gets string argument
        \author VooDooMan
        \version 1
        \date 2005
        \param arg_num Number of argument to get
        \return Returns argument
    */
    string get_str_arg(int arg_num);

    /*!
        \brief Gets name of channel
        \author VooDooMan
        \version 1
        \date 2005
        \return Returns name of channel
    */
    string get_channel();

    /*!
        \brief Gets nick name
        \author VooDooMan
        \version 1
        \date 2005
        \return Returns nick name
    */
    string get_nick();

    /*!
        \brief Gets goto label
        \author VooDooMan
        \version 1
        \date 2005
        \return Returns goto label
    */
    string get_label();

    /*!
        \brief Gets line number
        \author VooDooMan
        \version 1
        \date 2005
        \return Returns line number
    */
    int get_line();

    /*!
        \brief Sets line number
        \author VooDooMan
        \version 1
        \date 2005
        \param goto_line Line number
    */
    void set_goto_line(int goto_line);

    /*!
        \brief Gets command number
        \author VooDooMan
        \version 1
        \date 2005
        \return Returns command number
    */
    int get_command();

    /*!
        \brief Gets debug string
        \author VooDooMan
        \version 1
        \date 2005
        \return Returns debug string
    */
    const string& get_debug();

    /*!
        \brief Gets line number for goto
        \author VooDooMan
        \version 1
        \date 2005
        \return Returns line number for goto
    */
    int get_goto_line();

    /*!
        \brief Performs command
        \author VooDooMan
        \version 1
        \date 2005, 2006-02-16
        \param this_module Module name
        \param vars Variables
        \param goto_line Returns line number for goto
        \param is_return Returns true for return from procedure command
        \param skip_block Returns true to skip next block
        \param after_redirect Returns true if we are after redirection to another server
        \param redirect_checkpoint Returns true if we should remember position to jump to in case of redirect
        \param iteration_needed Returns 'true' if next iteration is needed (to avoid recursion)
        \param ste Returns state initialization informations for next iteration; undefined if 'iteration_needed' returns 'false'
    */
    void low_perform(c_module* this_module, c_var_array* vars, int& goto_line, bool& is_return, bool& skip_block, bool& after_redirect, bool& redirect_checkpoint, bool& iteration_needed, c_eval_iteration& ste);

    /*!
        \brief Dumps command
        \author VooDooMan
        \version 1
        \date 2005
        \param this_module Name of module
        \param indent Indent
    */
    string dump(c_module* this_module, int& indent);
};

#endif

