/***************************************************************************
                          c_proc_exc.cpp  -  description
                             -------------------
    begin                : Thu May 10 2004
    copyright            : (C) 2004-2007 by Marian VooDooMan Meravy
    email                : ghostvoodooman -a.T= users (dot) sourceforge.net
 ***************************************************************************/

/* $Id: c_proc_exc.cpp,v 1.14 2008/03/11 21:11:47 ghostvoodooman Exp $ */

/***************************************************************************

VooDoo cIRCle - an IRC (ro)bot
Copyright (C) 2004-2008 by Marian VooDooMan Meravy
  (ghostvoodooman -a.T= users (dot) sourceforge.net)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************/

/*!
    \file
    \brief Class 'c_proc_exc'
    \version $Id: c_proc_exc.cpp,v 1.14 2008/03/11 21:11:47 ghostvoodooman Exp $
*/

#if defined(HAVE_CONFIG_H)
#   include <config.h>
#endif

#include "../params.h"
#include "../utils/memleak.h"
#include "../utils/assert.h"
#include "../compat.h"

#include "c_proc_exc.h"
#include "../mod/c_module.h"
#include "../utils/precomp.h"

#if defined(_MSC_VER) || defined(__BORLANDC__)
#   pragma hdrstop
#endif

c_proc_exc::c_proc_exc(c_module* this_module, const string& file, const string& func, int line, int command_, e_type type, const string& debug)
{
    cmd=command_;
    type_=type;
    debug_=debug;
    this_module_=this_module;
    file_=file;
    func_=func;
    line_=line;
}

c_proc_exc::c_proc_exc(const c_proc_exc& orig)
{
    cmd=orig.cmd;
    type_=orig.type_;
    debug_=orig.debug_;
    this_module_=orig.this_module_;
    file_=orig.file_;
    func_=orig.func_;
    line_=orig.line_;
}

string c_proc_exc::get_msg()
{
    string res="Procedure error ";
    switch(type_) {
        case E_NUM_ARGS:
            res+="E_NUM_ARGS: Argument count mismatch";
            break;
        case E_INV_ARG_TYPE:
            res+="E_INV_ARG_TYPE: Argument type mismatch";
            break;
        case E_ARG_CNT_EXC:
            res+="E_ARG_CNT_EXC: Argument count exceeded";
            break;
        case E_GET_NUM_ARGS:
            res+="E_GET_NUM_ARGS: Argument count mismatch";
            break;
        case E_INV_COMMAND:
            res+="E_INV_COMMAND: Invalid command code";
            break;
        case E_INV_OPERATION:
            res+="E_INV_OPERATION: Invalid operation";
            break;
        case E_TIMEOUT:
            res+="E_TIMEOUT: Script execution timeout";
            break;
        case E_INV_WORK:
            res+="E_INV_WORK: It is prohibited to use 'work' command, or 'work()' function in other module than 'core' module";
            break;

        case E_VARIABLE_EVALUATION:
            res+="E_VARIABLE_EVALUATION: Variable evaluation error while processing 'execute' command";
            break;

        case E_NO_SERVER:
            res="Error E_NO_SERVER: No more servers";
            break;
        case E_RESTART:
            res="Error E_RESTART: Restart request by script";
            break;

        case E_BOT_SHUTDOWN_REQUEST:
            res="Bot shutdown request";
            break;

        default:
            res="Other internal error (Please report this problem)";
            break;
    }
    if(!debug_.empty()) {
        res+=" - ";
        res+=debug_;
    }
    return res;
}

void c_proc_exc::print_oops_to_log()
{
    string d="Procedure error in C++ function ";
    d+=file_;
    d+="::";
    d+=func_;
    d+=", line number ";
    d+=tos(line_);
    d+=": ";
    d+=get_msg();
    if(this_module_ && this_module_->get_logger()) {
        this_module_->get_logger()->log_print1(E_LOG_DEBUG,LOG_SCOPE_MODULE,"",d);
    }
}

