/***************************************************************************
                          c_proc_exc.h  -  description
                             -------------------
    begin                : Thu May 10 2004
    copyright            : (C) 2004-2007 by Marian VooDooMan Meravy
    email                : ghostvoodooman -a.T= users (dot) sourceforge.net
 ***************************************************************************/

/* $Id: c_proc_exc.h,v 1.14 2008/03/11 21:11:47 ghostvoodooman Exp $ */

/***************************************************************************

VooDoo cIRCle - an IRC (ro)bot
Copyright (C) 2004-2008 by Marian VooDooMan Meravy
  (ghostvoodooman -a.T= users (dot) sourceforge.net)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************/

/*!
    \file
    \brief Class 'c_proc_exc'
    \version $Id: c_proc_exc.h,v 1.14 2008/03/11 21:11:47 ghostvoodooman Exp $
*/

#ifndef C_PROC_EXC_H__
#define C_PROC_EXC_H__

#if defined(HAVE_CONFIG_H)
#   include <config.h>
#endif

#include <string>
#include <vector>
#include <stack>
#include <list>
#include <map>
#include <algorithm>

#include "../utils/memleak.h"

#include "../params.h"

class c_module;

/*!
    \brief Procedure exception
    \author VooDooMan
    \version 2
    \date 2005, 2008-02-27
*/
class c_proc_exc
{
public:
    enum e_type
    {
        E_NUM_ARGS,                     //!< Argument count mismatch
        E_INV_ARG_TYPE,                 //!< Argument type mismatch
        E_ARG_CNT_EXC,                  //!< Argument count exceeded
        E_GET_NUM_ARGS,                 //!< Argument count mismatch
        E_INV_COMMAND,                  //!< Invalid command_ code
        E_INV_OPERATION,                //!< Invalid operation
        E_TIMEOUT,                      //!< Script execution timeout
        E_VARIABLE_EVALUATION,          //!< Variable evaluation error (for "execute" command")
        E_INV_WORK,                     //!< Invalid 'work' command issued (in other module than 'core')

        E_NO_SERVER,                    //!< No more servers
        E_RESTART,                      //!< "restart" command

        E_BOT_SHUTDOWN_REQUEST,         //!< External shutdown has been requested (i.e. not via "restart" command)
    };
private:
    int cmd;                            //!< Erroneous command
    string debug_;                      //!< Debug string
    e_type type_;                       //!< Type of error
    c_module* this_module_;             //!< Module
    string file_;                       //!< C++ file
    string func_;                       //!< C++ function
    int line_;                          //!< C++ file line number

public:
    /*!
        \brief Constructor
        \author VooDooMan
        \version 1
        \date 2005
        \param this_module Name of module
        \param func C++ function where error occured
        \param file C++ file where error occured
        \param command_ Erroneous command
        \param type Type of error
        \param debug Debug string
        \param line C++ line number
    */
    c_proc_exc(c_module* this_module, const string& file, const string& func, int line, int command_, e_type type, const string& debug);

    /*!
        \brief Copy constructor
        \author VooDooMan
        \version 1
        \date 2005
    */
    c_proc_exc(const c_proc_exc& orig);

    /*!
        \brief Gets type of error
        \author VooDooMan
        \version 1
        \date 2005
    */
    e_type get_type() { return type_; }

    /*!
        \brief Gets error message for logs
        \author VooDooMan
        \version 1
        \date 2005
    */
    string get_msg();

    void print_oops_to_log();
};

#define THROW_C_PROC_EXC(this_module,command,type,debug)  \
    throw c_proc_exc(this_module,__FILE__,__my_func__,__LINE__,command,type,debug)

#endif

