/***************************************************************************
                          c_procedure.h  -  description
                             -------------------
    begin                : Thu May 10 2007
    copyright            : (C) 2004-2007 by Marian VooDooMan Meravy
    email                : ghostvoodooman -a.T= users (dot) sourceforge.net
 ***************************************************************************/

/* $Id: c_procedure.h,v 1.16 2008/07/02 09:35:50 ghostvoodooman Exp $ */

/***************************************************************************

VooDoo cIRCle - an IRC (ro)bot
Copyright (C) 2004-2008 by Marian VooDooMan Meravy
  (ghostvoodooman -a.T= users (dot) sourceforge.net)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************/

/*!
    \file
    \brief Class 'c_procedure'
    \version $Id: c_procedure.h,v 1.16 2008/07/02 09:35:50 ghostvoodooman Exp $
*/

#ifndef C_PROCEDURE_H__
#define C_PROCEDURE_H__

#if defined(HAVE_CONFIG_H)
#   include <config.h>
#endif

#include <string>
#include <vector>
#include <stack>
#include <list>
#include <map>
#include <algorithm>

#include "../utils/memleak.h"

#include "../params.h"

#include "c_proc_command.h"
#include "../logic/c_rproc.h"
#include "../botnet/c_replication.h"

/*!
    \brief Stores procedure implementation and its properties
    \author VooDooMan
    \version 1
    \date 2004
*/
class c_procedure
{
public:
    string name;                                    //!< Declaration string of procedure
    deque<c_proc_command> commands;                  //!< Commands of procedure
    vector<string> groups;                          //!< List of groups that procedure is member of
    vector<c_rproc> rproc;                          //!< List of bot which can call this procedure remotely (BOTNET)

    vector<c_replication> replication;              //!< Informations of replication
    my_time_t last_changed;                            //!< Time of last change

    int minimal_botnet_version;                     //!< Number of minimal BOTNET protocol version (for replication)

    bool parsed_call_string;                        //!< If we have parsed call string
    string name_only;                               //!< Only name of procedure
    vector<string> args;                            //!< List of arguments

    int lock_cnt;                                   //!< If we are executing this procedure

    /*!
        \brief Constructor
        \author VooDooMan
        \version 1
        \date 2005
    */
    c_procedure();

    /*!
        \brief Parses call string
        \author VooDooMan
        \version 1
        \date 2005
    */
    void parse_call_string();

    /*!
        \brief Returns last changed attribute
        \author VooDooMan
        \version 1
        \date 2005
    */
    my_time_t get_last_changed() { return last_changed; }

    /*!
        \brief Sets last changed attribute
        \author VooDooMan
        \version 1
        \date 2005
        \param when Last changed
    */
    void set_last_changed(my_time_t when) { last_changed=when; }

    /*!
        \brief Gets minimal botnet version (for replication)
        \author VooDooMan
        \version 1
        \date 2005
    */
    int get_minimal_botnet_version() { return minimal_botnet_version; }

    /*!
        \brief Gets name of procedure
        \author VooDooMan
        \version 1
        \date 2005
    */
    string get_name() { return name; }
};

#endif

