/***************************************************************************
                          script_consts.cpp  -  description
                             -------------------
    begin                : Wed Feb 06 2008
    copyright            : (C) 2008 by Marian VooDooMan Meravy
    email                : ghostvoodooman -a.T= users (dot) sourceforge.net
 ***************************************************************************/

/* $Id: script_consts.cpp,v 1.10 2008/11/24 12:09:46 ghostvoodooman Exp $ */

/***************************************************************************

VooDoo cIRCle - an IRC (ro)bot
Copyright (C) 2004-2008 by Marian VooDooMan Meravy
  (ghostvoodooman -a.T= users (dot) sourceforge.net)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************/

/*!
    \file
    \brief Defines constants for function scripting
    \version $Id: script_consts.cpp,v 1.10 2008/11/24 12:09:46 ghostvoodooman Exp $
*/

#if defined(HAVE_CONFIG_H)
#   include <config.h>
#endif

#include <string>
#include <map>

#include <pcre.h>

#include "../params.h"
#include "../utils/memleak.h"
#include "../utils/assert.h"

#include "script_consts.h"

#include "../var/c_var_base.h"
#include "../var/c_var_int.h"
#include "../var/c_var_string.h"
#include "../var/c_var_bool.h"
#include "../var/c_var_unset.h"
#include "../logic.h"

#include "../utils/precomp.h"

#if defined(_MSC_VER) || defined(__BORLANDC__)
#   pragma hdrstop
#endif

/*!
    \brief stores one constant
    \version 1
    \date 2008-02-06
    \author VooDooMan
*/
struct c_constant
{
    enum e_kind
    {
        KIND_INT,                           //!< It is INT
        KIND_STR,                           //!< It is STRING
        KIND_BOOL,                          //!< It is BOOL
        KIND_NULL,                          //!< It is NULL (unset)

        KIND_PCRE_ERROR_CODE_INT,           //!< Like 'KIND_INT' but this is an error code from PCRE (when compiling)
        KIND_PCRE_EXEC_ERROR_CODE_INT,      //!< Like 'KIND_INT' but this is an error code from PCRE (when executing)
    };

    e_kind m_kind;      //!< Kind

    MY_LONG_LONG m_int; //!< Value; valid only if m_kind==KIND_INT
    string m_str;       //!< Value; valid only if m_kind==KIND_STR
    bool m_bool;        //!< Value; valid only if m_kind==KIND_BOOL

    /*!
        \brief Stores KIND_INT value
        \version 1
        \date 2008-02-06
        \author VooDooMan
        \param x Value
    */
    c_constant(MY_LONG_LONG x)
        : m_kind(KIND_INT)
    {
        m_int=x;
    }

    /*!
        \brief Stores various-kind (KIND_INT) value
        \version 1
        \date 2008-02-07
        \author VooDooMan
        \param x Value
        \param kind Kind of a value
    */
    c_constant(MY_LONG_LONG x, e_kind kind)
        : m_kind(KIND_INT)
    {
        switch(kind) {
            case KIND_INT:
            case KIND_PCRE_ERROR_CODE_INT:
            case KIND_PCRE_EXEC_ERROR_CODE_INT:
                m_int=x;
                break;
            default:
                {
                    MY_ASSERT_NOTREACH();
                }
                break;
        }
    }

    /*!
        \brief Stores KIND_STR value
        \version 1
        \date 2008-02-06
        \author VooDooMan
        \param x Value
    */
    c_constant(const string& x)
        : m_kind(KIND_STR)
    {
        m_str=x;
    }

    /*!
        \brief Stores KIND_BOOL value
        \version 1
        \date 2008-02-06
        \author VooDooMan
        \param x Value
        \param dummy Ignored
    */
    c_constant(bool x, bool UNUSEDARG(dummy))
        : m_kind(KIND_BOOL)
    {
        m_bool=x;
    }

    /*!
        \brief Stores KIND_NULL value
        \version 1
        \date 2008-02-06
        \author VooDooMan
    */
    c_constant()
        : m_kind(KIND_NULL)
    { ; }
};

struct c_constants_lookup
{
    map<string,c_constant> m_constants;

public:
    c_constants_lookup()
    {
        m_constants.insert(make_pair("SEEK_SET",c_constant(MY_SEEK_SET)));
        m_constants.insert(make_pair("SEEK_CUR",c_constant(MY_SEEK_CUR)));
        m_constants.insert(make_pair("SEEK_END",c_constant(MY_SEEK_END)));

        m_constants.insert(make_pair("CASE_SERVER",c_constant(MY_CASE_SERVER )));
        m_constants.insert(make_pair("CASE_ASCII",c_constant(MY_CASE_ASCII)));
        m_constants.insert(make_pair("CASE_RFC1459",c_constant(MY_CASE_RFC1459)));

        m_constants.insert(make_pair("TRUE",c_constant(true,true)));
        m_constants.insert(make_pair("FALSE",c_constant(false,false)));

        m_constants.insert(make_pair("NULL",c_constant()));

        m_constants.insert(make_pair("LOW",c_constant(logic_get_priority(LOGIC_TXT_CONST_LOW))));
        m_constants.insert(make_pair("HIGH",c_constant(logic_get_priority(LOGIC_TXT_CONST_HIGH))));
        m_constants.insert(make_pair("CRITICAL",c_constant(logic_get_priority(LOGIC_TXT_CONST_CRITICAL))));

        m_constants.insert(make_pair("PCRE_ERROR_NOMATCH",c_constant(PCRE_ERROR_NOMATCH, c_constant::KIND_PCRE_ERROR_CODE_INT)));
        m_constants.insert(make_pair("PCRE_ERROR_NULL",c_constant(PCRE_ERROR_NULL, c_constant::KIND_PCRE_ERROR_CODE_INT)));
        m_constants.insert(make_pair("PCRE_ERROR_BADOPTION",c_constant(PCRE_ERROR_BADOPTION, c_constant::KIND_PCRE_ERROR_CODE_INT)));
        m_constants.insert(make_pair("PCRE_ERROR_BADMAGIC",c_constant(PCRE_ERROR_BADMAGIC, c_constant::KIND_PCRE_ERROR_CODE_INT)));
#if defined(PCRE_ERROR_UNKNOWN_OPCODE)
        m_constants.insert(make_pair("PCRE_ERROR_UNKNOWN_OPCODE",c_constant(PCRE_ERROR_UNKNOWN_OPCODE, c_constant::KIND_PCRE_ERROR_CODE_INT)));
#else
        m_constants.insert(make_pair("PCRE_ERROR_UNKNOWN_OPCODE",c_constant(-5, c_constant::KIND_PCRE_ERROR_CODE_INT)));
#endif
        m_constants.insert(make_pair("PCRE_ERROR_NOMEMORY",c_constant(PCRE_ERROR_NOMEMORY, c_constant::KIND_PCRE_ERROR_CODE_INT)));
        m_constants.insert(make_pair("PCRE_ERROR_NOSUBSTRING",c_constant(PCRE_ERROR_NOSUBSTRING, c_constant::KIND_PCRE_ERROR_CODE_INT)));
        m_constants.insert(make_pair("PCRE_ERROR_MATCHLIMIT",c_constant(PCRE_ERROR_MATCHLIMIT, c_constant::KIND_PCRE_ERROR_CODE_INT)));
        m_constants.insert(make_pair("PCRE_ERROR_CALLOUT",c_constant(PCRE_ERROR_CALLOUT, c_constant::KIND_PCRE_ERROR_CODE_INT)));
        m_constants.insert(make_pair("PCRE_ERROR_BADUTF8",c_constant(PCRE_ERROR_BADUTF8, c_constant::KIND_PCRE_ERROR_CODE_INT)));
        m_constants.insert(make_pair("PCRE_ERROR_BADUTF8_OFFSET",c_constant(PCRE_ERROR_BADUTF8_OFFSET, c_constant::KIND_PCRE_ERROR_CODE_INT)));
        m_constants.insert(make_pair("PCRE_ERROR_PARTIAL",c_constant(PCRE_ERROR_PARTIAL, c_constant::KIND_PCRE_ERROR_CODE_INT)));
        m_constants.insert(make_pair("PCRE_ERROR_BADPARTIAL",c_constant(PCRE_ERROR_BADPARTIAL, c_constant::KIND_PCRE_ERROR_CODE_INT)));
        m_constants.insert(make_pair("PCRE_ERROR_INTERNAL",c_constant(PCRE_ERROR_INTERNAL, c_constant::KIND_PCRE_ERROR_CODE_INT)));
        m_constants.insert(make_pair("PCRE_ERROR_BADCOUNT",c_constant(PCRE_ERROR_BADCOUNT, c_constant::KIND_PCRE_ERROR_CODE_INT)));
        m_constants.insert(make_pair("PCRE_ERROR_DFA_UITEM",c_constant(PCRE_ERROR_DFA_UITEM, c_constant::KIND_PCRE_ERROR_CODE_INT)));
        m_constants.insert(make_pair("PCRE_ERROR_DFA_UCOND",c_constant(PCRE_ERROR_DFA_UCOND, c_constant::KIND_PCRE_ERROR_CODE_INT)));
        m_constants.insert(make_pair("PCRE_ERROR_DFA_UMLIMIT",c_constant(PCRE_ERROR_DFA_UMLIMIT, c_constant::KIND_PCRE_ERROR_CODE_INT)));
        m_constants.insert(make_pair("PCRE_ERROR_DFA_WSSIZE",c_constant(PCRE_ERROR_DFA_WSSIZE, c_constant::KIND_PCRE_ERROR_CODE_INT)));
        m_constants.insert(make_pair("PCRE_ERROR_DFA_RECURSE",c_constant(PCRE_ERROR_DFA_RECURSE, c_constant::KIND_PCRE_ERROR_CODE_INT)));
#if defined(PCRE_ERROR_RECURSIONLIMIT)
        m_constants.insert(make_pair("PCRE_ERROR_RECURSIONLIMIT",c_constant(PCRE_ERROR_RECURSIONLIMIT, c_constant::KIND_PCRE_ERROR_CODE_INT)));
#else
        m_constants.insert(make_pair("PCRE_ERROR_RECURSIONLIMIT",c_constant(-21, c_constant::KIND_PCRE_ERROR_CODE_INT)));
#endif
#if defined(PCRE_ERROR_NULLWSLIMIT)
        m_constants.insert(make_pair("PCRE_ERROR_NULLWSLIMIT",c_constant(PCRE_ERROR_NULLWSLIMIT, c_constant::KIND_PCRE_ERROR_CODE_INT)));
#else
        m_constants.insert(make_pair("PCRE_ERROR_NULLWSLIMIT",c_constant(-22, c_constant::KIND_PCRE_ERROR_CODE_INT)));
#endif
#if defined(PCRE_ERROR_BADNEWLINE)
        m_constants.insert(make_pair("PCRE_ERROR_BADNEWLINE",c_constant(PCRE_ERROR_BADNEWLINE, c_constant::KIND_PCRE_ERROR_CODE_INT)));
#else
        m_constants.insert(make_pair("PCRE_ERROR_BADNEWLINE",c_constant(-23, c_constant::KIND_PCRE_ERROR_CODE_INT)));
#endif

        m_constants.insert(make_pair("PREG_NO_ERROR",c_constant(PREG_NO_ERROR, c_constant::KIND_PCRE_EXEC_ERROR_CODE_INT)));
        m_constants.insert(make_pair("PREG_INTERNAL_ERROR",c_constant(PCRE_ERROR_INTERNAL, c_constant::KIND_PCRE_EXEC_ERROR_CODE_INT)));
        m_constants.insert(make_pair("PREG_BACKTRACK_LIMIT_ERROR",c_constant(PCRE_ERROR_MATCHLIMIT, c_constant::KIND_PCRE_EXEC_ERROR_CODE_INT)));
#if defined(PCRE_ERROR_RECURSIONLIMIT)
        m_constants.insert(make_pair("PREG_RECURSION_LIMIT_ERROR",c_constant(PCRE_ERROR_RECURSIONLIMIT, c_constant::KIND_PCRE_EXEC_ERROR_CODE_INT)));
#else
        m_constants.insert(make_pair("PREG_RECURSION_LIMIT_ERROR",c_constant(-21, c_constant::KIND_PCRE_EXEC_ERROR_CODE_INT)));
#endif
        m_constants.insert(make_pair("PREG_BAD_UTF8_ERROR",c_constant(PCRE_ERROR_BADUTF8, c_constant::KIND_PCRE_EXEC_ERROR_CODE_INT)));

        m_constants.insert(make_pair("PREG_GREP_INVERT",c_constant(PREG_GREP_INVERT)));
        m_constants.insert(make_pair("PREG_PATTERN_ORDER",c_constant(PREG_PATTERN_ORDER)));
        m_constants.insert(make_pair("PREG_SET_ORDER",c_constant(PREG_SET_ORDER)));
        m_constants.insert(make_pair("PREG_OFFSET_CAPTURE",c_constant(PREG_OFFSET_CAPTURE)));
        m_constants.insert(make_pair("PREG_SPLIT_NO_EMPTY ",c_constant(PREG_SPLIT_NO_EMPTY )));
        m_constants.insert(make_pair("PREG_SPLIT_DELIM_CAPTURE",c_constant(PREG_SPLIT_DELIM_CAPTURE)));
        m_constants.insert(make_pair("PREG_SPLIT_OFFSET_CAPTURE",c_constant(PREG_SPLIT_OFFSET_CAPTURE)));
    }

    MY_WISH_INLINE bool translate(const string& str, c_var_base** translated = NULL) const
    {
        map<string,c_constant>::const_iterator i1=m_constants.find(str);
        if(m_constants.end()==i1) {
            if(NULL!=translated)
                *translated=NULL;
            return false;
        }

        if(NULL==translated)
            return true;

        const c_constant& c=(*i1).second;

        switch(c.m_kind) {
            case c_constant::KIND_INT:
            case c_constant::KIND_PCRE_ERROR_CODE_INT:
            case c_constant::KIND_PCRE_EXEC_ERROR_CODE_INT:
                *translated=MY_NEW c_var_int(c.m_int);
                break;
            case c_constant::KIND_STR:
                *translated=MY_NEW c_var_string(c.m_str);
                break;
            case c_constant::KIND_BOOL:
                *translated=MY_NEW c_var_bool(c.m_bool);
                break;
            case c_constant::KIND_NULL:
                *translated=MY_NEW c_var_unset();
                break;
            default:
                {
                    MY_ASSERT_NOTREACH();

                    translated=NULL;
                    return false;
                }
        }

        return true;
    }

    /*!
        \brief Stores KIND_BOOL value
        \version 1
        \date 2008-02-07
        \author VooDooMan
        \param rc Result from 'preg_*()' function
        \return Returns error string
    */
    string preg_strerr(int rc) const
    {
        if(rc>=0)
            return (string)"No error";

        string res;

        map<string,c_constant>::const_iterator i1;
        for(map<string,c_constant>::const_iterator i1=m_constants.begin(); i1!=m_constants.end(); ++i1) {
            if(c_constant::KIND_PCRE_EXEC_ERROR_CODE_INT!=(*i1).second.m_kind && c_constant::KIND_PCRE_ERROR_CODE_INT!=(*i1).second.m_kind)
                continue;
            if((*i1).second.m_int!=rc)
                continue;
            res=(*i1).first;
            break;
        }
        if(res.empty()) {
            res="Unknown PCRE error code";
            res+="(";
            res+=tos(rc);
            res+=")";
        }
        return res;
    }
};

static c_constants_lookup constants_lookup;

bool script_function_translate_constant(const string& str, c_var_base* &translated)
{
    return constants_lookup.translate(str,&translated);
}

bool script_function_translate_constant(const string& str)
{
    return constants_lookup.translate(str,NULL);
}

string get_strerr_pcre_exec(int rc)
{
    return constants_lookup.preg_strerr(rc);
}

