/***************************************************************************
                          shared.h  -  description
                             -------------------
    begin                : Thu Dec 9 2004
    copyright            : (C) 2004-2007 by Marian VooDooMan Meravy
    email                : ghostvoodooman -a.T= users (dot) sourceforge.net
 ***************************************************************************/

/* $Id: shared.h,v 1.80 2008/03/11 21:11:31 ghostvoodooman Exp $ */

/***************************************************************************

VooDoo cIRCle - an IRC (ro)bot
Copyright (C) 2004-2008 by Marian VooDooMan Meravy
  (ghostvoodooman -a.T= users (dot) sourceforge.net)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************/

/*!
    \file
    \brief Contains shared structures
    \version $Id: shared.h,v 1.80 2008/03/11 21:11:31 ghostvoodooman Exp $
*/

#ifndef SHARED_H__
#define SHARED_H__

#if defined(HAVE_CONFIG_H)
#   include <config.h>
#endif

#include <string>
#include <vector>
#include <stack>
#include <list>
#include <map>
#include <algorithm>

#include <stdlib.h>
#include <time.h>

using namespace std;

#include "utils/memleak.h"
#include "utils/assert.h"
#include "utils.h"
#include "params.h"

#include "variable.h"

/*!
    \def PRIVATE_EVENT_SEVERITY_CODE_NA
    \brief For event "on_internal_event": Severity code N/A
*/
#define PRIVATE_EVENT_SEVERITY_CODE_NA                      0

/*!
    \def PRIVATE_EVENT_SEVERITY_CODE_INFORMATIONAL
    \brief For event "on_internal_event": Severity code INFORMATIONAL
*/
#define PRIVATE_EVENT_SEVERITY_CODE_INFORMATIONAL           1

/*!
    \def PRIVATE_EVENT_SEVERITY_CODE_WARNING
    \brief For event "on_internal_event": Severity code WARNING
*/
#define PRIVATE_EVENT_SEVERITY_CODE_WARNING                 3

/*!
    \def PRIVATE_EVENT_SEVERITY_CODE_ERROR
    \brief For event "on_internal_event": Severity code ERROR
*/
#define PRIVATE_EVENT_SEVERITY_CODE_ERROR                   5

/*!
    \def PRIVATE_EVENT_SEVERITY_CODE_CRITICAL_ERROR
    \brief For event "on_internal_event": Severity code CRITICAL_ERROR
*/
#define PRIVATE_EVENT_SEVERITY_CODE_CRITICAL_ERROR          7

/*!
    \def PRIVATE_EVENT_SEVERITY_CODE_FATAL_ERROR
    \brief For event "on_internal_event": Severity code FATAL_ERROR
*/
#define PRIVATE_EVENT_SEVERITY_CODE_FATAL_ERROR             9

/*!
    \def JUMP_ID_INVALID
    \brief Represents invalid ID
*/
/*!
    \def JUMP_ID_NA
    \brief Represents ID which is not applicable
*/

#define JUMP_ID_INVALID         -1L
#define JUMP_ID_NA              -2L

/*!
    \def CHECK_ID
    \brief Checks ID for validity
    \param ID ID to check
*/

#define CHECK_ID(ID) \
    msg=""; \
    if(ID<=0) { \
        msg=(string)"Internal error, sanity check failed: instance of class '"+typeid(*this).name()+"' has invalid ID " \
            #ID + " == "+tos(ID)+"; "; \
        msg+=_debug; \
        msg=GET_STR_DEBUG_INFO_EX(msg); \
        return true; \
    }

/*!
    \def CHECK_ID_ALLOW_JUMP_ID_NA
    \brief Checks ID for validity, but allows ID to be == JUMP_ID_NA
    \param ID ID to check
*/
#define CHECK_ID_ALLOW_JUMP_ID_NA(ID) \
    msg=""; \
    if(JUMP_ID_NA!=ID) \
        CHECK_ID(ID)

/*!
    \def CHECK_EXP
    \brief Checks expression for validity
    \param EXP Expression (c_expression instance, pointer)
*/
#define CHECK_EXP(EXP) \
    msg=""; \
    if(EXP->got_error()) { \
        msg=(string)"Internal error, sanity check failed: instance of class '"+typeid(*this).name()+"' has invalid expression " \
            #EXP + "; "; \
        msg+=_debug; \
        msg=GET_STR_DEBUG_INFO_EX(msg); \
        return true; \
    }

enum e_cmd_type
{
    _INVALID = 0,
    CMD_FOR_COND,
    CMD_LEFT_BRACE,
    CMD_RIGHT_BRACE,
    CMD_EXP,
    CMD_JMP,
    CMD_NOP,
    CMD_IF,
    CMD_SWITCH,
    CMD_CASE,
    CMD_CASE_DEFAULT,
    CMD_LABEL,
    CMD_GOTO,
    CMD_GLOBAL,
    CMD_BREAK,
    CMD_CONTINUE,
    CMD_DO,
    CMD_WHILE,
    CMD_DO_WHILE,
    CMD_FOREACH,
    CMD_EOF,
    CMD_FOREACH_ITER,
};

enum e_func_type
{
    FUNC_OTHER,                     //!< User-defined function
    FUNC_NONE,                      //!< No function
    FUNC_ARRAY,                     //!< array() function
    FUNC_BUILT_IN,                  //!< Built-in function
    FUNC_ISSET,                     //!< isset() function
    FUNC_UNSET,                     //!< unset() function
    FUNC_EVAL,                      //!< eval() function
};                                  //!< Type of function

#include "func/expression/oper_assoc.h"

//c_var_base* var_evaluate_exp(c_module* this_module, c_functions& functions, vector<c_part>& parenthesis_parts, c_var_array* vars, e_oper& last_operator, const string& debug, c_var_array* rollback, bool& is_return, bool& skip_block, const string& next_expression, bool want_ref);

#endif

