/***************************************************************************
                          sock.cpp  -  description
                             -------------------
    begin                : Thu Dec 9 2004
    copyright            : (C) 2004-2007 by Marian VooDooMan Meravy
    email                : ghostvoodooman -aT= users (dot) sourceforge.net
 ***************************************************************************/

/* $Id: sock.cpp,v 1.76 2008/11/24 12:09:23 ghostvoodooman Exp $ */

/***************************************************************************

VooDoo cIRCle - an IRC (ro)bot
Copyright (C) 2004-2008 by Marian VooDooMan Meravy
  (ghostvoodooman -aT= users (dot) sourceforge.net)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************/

/*!
    \file
    \brief Provides low-level IP socket communication
    \version $Id: sock.cpp,v 1.76 2008/11/24 12:09:23 ghostvoodooman Exp $
*/

#if defined(HAVE_CONFIG_H)
#   include <config.h>
#endif

#include "params.h"
#include "utils/memleak.h"
#include "utils/assert.h"
#include "compat.h"

#include "sock_shared.h"

#include <string>
#include <vector>
#include <stack>
#include <list>
#include <map>

#include <stdio.h>
#include <stdlib.h>

#if defined(__CYGWIN__)
#    include <unistd.h>
#endif

#include <string.h>

#ifdef HAVE_SYS_TIME_H
#   include <sys/time.h>
#endif

#include "utils.h"      // because we want to enable mem leak detection

#include "params.h"

#include <stdlib.h>
#include <errno.h>

#include "sock.h"
#include "match.h"
#include "log.h"
#include "md5.h"            // for byteReverse()
#include "thread/mutex.h"

#if defined(_WIN32) || defined(_WIN64) || defined(__CYGWIN__)
#   define SOCK_ERRNO   WSAGetLastError()
#else
#   define SOCK_ERRNO   errno
#endif

#include "utils/precomp.h"

#if defined(_MSC_VER) || defined(__BORLANDC__)
#   pragma hdrstop
#endif

//---------------------------------------------------------------------------

using namespace std;

#if defined(_WIN32) || defined(_WIN64)
# ifndef WINSOCK_VERSION
#   define WINSOCK_VERSION 0x0002
# endif
#endif

#if defined(_WIN32) || defined(_WIN64)
WSADATA WSAData;                                    //!< Stores "WSAStartup()"'s data
#   define socklen_t int
#endif

short sock_tcp_proto_num;                           //!< Number of TPC protocol

vector<MY_SOCKET_FD> all_sockets;                         //!< List of all sockets

size_t sock_lo_get_input_amount(c_socket& socket, char* buff, size_t len, int& bsd_errno);
size_t sock_lo_recv(c_socket& socket, char* buff, size_t len, int& bsd_errno);
size_t sock_lo_send(c_socket& socket, char* buff, size_t len, int& bsd_errno);

static vector<pair<c_socket,my_time_t> > sock_linger;  //!< Lingering sockets

static map<unsigned,c_socket_hndl> sockets;
static c_mutex* m_sockets=NULL;

union sa_t
{
    sockaddr gen;
    sockaddr_in in;
#ifdef IPv6
    sockaddr_in6 in6;
#endif
};

/*!
    \brief Stores data to send/receive later
    \author VooDooMan
    \version 1
    \date 2004
*/
class c_data
{
public:
    bool closed;                        //!< Idicates that connection has been closed (if this is true, data && len are undefined)
    size_t len;                         //!< Length of data
    char* data;                         //!< Data

    c_data()
    {
        closed=false;
        len=0;
        data=NULL;
    }
};

/*!
    \brief Stores data to send later
    \author VooDooMan
    \version 2
    \date 2004, 2005
*/
class c_send_later
{
public:
    c_socket socket;                        //!< Socket handle
    list<c_data> data;                      //!< Array of data packets

    my_time_t last_send;                       //!< Timestamp of last send operation

    c_send_later()
    {
        socket.clear();
        data.clear();
        last_send=my_time(NULL);
    }
};

/*!
    \brief Stores data to receive later
    \author VooDooMan
    \version 2
    \date 2004, 2005
*/
class c_receive_later
{
public:
    bool has_error;                         //!< Set to true, if we have data, but there's an error on the socket, or socket has been closed (by remote host)
    int last_error;                         //!< Stores last error code
    int bsd_errno;                          //!< Low-level errno after BSD sockets API call
    c_socket socket;                        //!< Socket handle
    bool using_window;                      //!< Set to true if this socket is using message window callback to receive data
    list<c_data> data;                      //!< Array of data packets

    my_time_t last_recv;                       //!< Timestamp of last receive operation

    c_receive_later()
    {
        has_error=false;
        socket.clear();
        using_window=false;
        data.clear();
        last_error=0;
        last_recv=my_time(NULL);
        bsd_errno=0;
    }
};

vector<c_send_later> send_later;            //!< Packets to send later
vector<c_receive_later> receive_later;      //!< Packets to receive later

MY_STATIC unsigned sock_hndl_create(MY_SOCKET_FD h, c_socket& socket)
{
    c_mutex_locker lock(m_sockets);

    static unsigned last=0;
    for(;;) {
        last++;
        if(0==last || (unsigned)-1==last)
            continue;
        map<unsigned,c_socket_hndl>::iterator i1=sockets.find(last);
        if(sockets.end()==i1) {
            c_socket_hndl sh(h);
            sockets.insert(make_pair(last,sh));
            socket.hndl=last;
            socket.was_hndl=last;
            return last;
        }
    }
}

MY_STATIC map<unsigned,c_socket_hndl>::iterator sock_find(unsigned hndl, const char* fn, int ln)
{
    c_mutex_locker lock(m_sockets);

    map<unsigned,c_socket_hndl>::iterator i1=sockets.find(hndl);
    if(sockets.end()==i1) {
#ifdef _DEBUG
        // DEBUG: Set breakpoint here!
        int x=0;
        x++;
#endif

        string log="OOPS: BUG: Closing non-existing socket handle, ";
        log+=fn;
        log+=":";
        log+=tos(ln);
        log+=", PLEASE REPORT THIS PROBLEM! (old known bug, but have not been figured out)";
        log_debug(log,true);
        MY_ASSERT_NOTREACH();

        return sockets.end();
    }
    if((*i1).second.closed) {
#ifdef _DEBUG
        // DEBUG: Set breakpoint here!
        int x=0;
        x++;
#endif

        string log="OOPS: BUG: Double-close of socket handle, ";
        log+=fn;
        log+=":";
        log+=tos(ln);
        log+=", it was already closed at ";
        log+=(*i1).second.fn;
        log+=":";
        log+=tos((*i1).second.ln);
        log+=", PLEASE REPORT THIS PROBLEM! (old known bug, but have not been figured out)";
        log_debug(log,true);
        MY_ASSERT_NOTREACH();

        return sockets.end();
    }

    return i1;
}

void sock_hndl_delete(unsigned hndl, const char* fn, int ln)
{
    c_mutex_locker lock(m_sockets);

    map<unsigned,c_socket_hndl>::iterator i1=sock_find(hndl,fn,ln);
    if(sockets.end()==i1)
        return;

    (*i1).second.closed=my_time(NULL);
    (*i1).second.fn=const_cast<char*>(fn);
    (*i1).second.ln=ln;
#ifndef _DEBUG
    sockets.erase(i1);
#endif
}

#if defined(_WIN32) || defined(_WIN64)
#   define WM_TCPASYNCSELECT (WM_USER+123)     //!< Our own window message for receiving data (Windows(TM) only)

/*!
    \brief Callback windows procedure for receiving data (Windows(TM) only)
    \author VooDooMan
    \version 3
    \date 2005, 2008-03-13
    \param hwnd Handle of window
    \param uMsg Message
    \param wParam Parameter #1
    \param lParam Parameter #2
    \return Should return zero if message has been processed
    \warning Not thread-safe; we use static variable as a buffer and mutex, on Windoze it is recursive critical section! Beware of that!
*/
LRESULT CALLBACK WndProc(/*HWND*/void* UNUSEDARG(hwnd), /*UINT*/int uMsg, /*WPARAM*/int wParam, /*LPARAM*/int lParam)
{
    c_mutex_locker lock(m_sockets);

    if(uMsg==WM_NCCREATE)
        return true;

    enum {
        MAX_LEN = 1024U * 128U,
    };

    static char Buf[MAX_LEN];

    try {
        if(uMsg!=WM_TCPASYNCSELECT)
            return 0;
        int ErrorCode=WSAGETSELECTERROR(lParam);
        if(ErrorCode>WSABASEERR)
            return 0;

        int SelectEvent=WSAGETSELECTEVENT(lParam);
        switch (SelectEvent) {
            case FD_READ:
                {
                    for(;;) {
                        MY_SOCKET_FD h=wParam;
                        int Size=recvfrom(h,Buf,MAX_LEN,0,NULL,NULL);
                        int bsd_errno=errno;
                        errno=0;
                        if(Size==MY_SOCKET_ERROR) {
                            vector<c_receive_later>::iterator i1;
                            for(i1=receive_later.begin(); i1!=receive_later.end(); ++i1) {
                                if(((*i1).socket.deref(__FILE__,__LINE__))==h) {
                                    int la=SOCK_ERRNO;
                                    if(la!=MY_EAGAIN_WOULD_BLOCK && !(*i1).has_error) {
                                        (*i1).has_error=true;
                                        (*i1).last_error=la;
                                        if(!(*i1).bsd_errno && bsd_errno!=EAGAIN)
                                            (*i1).bsd_errno=bsd_errno;
                                    }
                                }
                            }
                            return 0;
                        }
                        vector<c_receive_later>::iterator i1;
                        for(i1=receive_later.begin(); Size>0 && i1!=receive_later.end(); ++i1) {
                            if(((*i1).socket.deref(__FILE__,__LINE__))==h) {
                                (*i1).socket.last_io=my_time(NULL);

                                if(Size>0)
                                    (*i1).last_recv=my_time(NULL);

                                c_data d;
                                char* b=MY_NEW char[Size];
                                memcpy(b,Buf,Size);
                                d.closed=false;
                                d.data=b;
                                d.len=Size;
                                (*i1).data.push_back(d);
                                break;
                            }
                        }
                        if(Size<=0)
                            break;
                    }
                }
                break;
            case FD_CLOSE:
                {
                    MY_SOCKET_FD h=wParam;
                    vector<c_receive_later>::iterator i1;
                    for(i1=receive_later.begin(); i1!=receive_later.end(); ++i1) {
                        if(((*i1).socket.deref(__FILE__,__LINE__))==h) {
                            c_data d;
                            d.closed=true;
                            (*i1).socket.closed=true;
                            d.data=NULL;
                            d.len=0;
                            (*i1).data.push_back(d);
                            break;
                        }
                    }
                }
                break;
        }
    } catch (...) { ; }
    return 0;
}

/*!
    \brief Allocates window handle (Windows(TM) only)
    \author VooDooMan
    \version 1
    \date 2004
    \param m Pointer to window callback message procedure
    \returns Window handle or NULL if something goes wrong
*/
HWND AllocateHWnd(void* m)
{
    WNDCLASSEXA c;
    memset(&c,0,sizeof(c));
    char tmp[1024];
    tmp[0]=0;
    for(int ii=0; ii<5; ++ii) {
        int i=rand();
        char tmp2[16];
        sprintf(tmp2,"%x",i);
        strcat(tmp,tmp2);
    }
    c.lpszClassName=tmp;
    c.lpfnWndProc=(WNDPROC)m;
    c.cbSize=sizeof(c);
    RegisterClassExA(&c);
    HWND h=CreateWindowExA(0,tmp,"",0,0,0,0,0,HWND_MESSAGE,NULL,NULL,NULL);
#if defined(_MSC_VER)
#   pragma warning(push)
#   pragma warning(disable:4244)    // warning C4244: 'argument' : conversion from 'LONG_PTR' to 'LONG', possible loss of data
#endif
#if defined(__INTEL_COMPILER)
#   pragma warning(push)
#   pragma warning(disable:1684)    // warning #1684: conversion from pointer to same-sized integral type (potential portability problem)
#endif
    SetWindowLongPtrA(h,GWLP_WNDPROC,reinterpret_cast<LONG_PTR>(m)); // okay on Win64, because "LONG_PTR" is 64 bit long as well as "HWND"
#if defined(__INTEL_COMPILER)
#   pragma warning(pop)
#endif
#if defined(_MSC_VER)
#   pragma warning(pop)
#endif

    return h;
}
#endif

/*!
    \brief Writes data to connected socket
    \author VooDooMan
    \version 2
    \date 2005
    \param socket Socket handle of connected socket
    \param data Buffer to write
    \param size Size of buffer
    \param error_code Returns OS's socket error
    \return Returns amount of sent data in bytes
*/
size_t sock_send(c_socket& socket, const char* data, size_t size, int& error_code)
{
    c_mutex_locker lock(m_sockets);

    {
        {
            {
            again:
                bool got=false;
                vector<c_send_later>::iterator i1;
                for(i1=send_later.begin(); i1!=send_later.end(); ++i1) {
                    if((*i1).socket==socket) {
                        if((*i1).socket.last_io<socket.last_io)
                            (*i1).socket.last_io=socket.last_io;
                        else
                            socket.last_io=(*i1).socket.last_io;
                        (*i1).last_send=my_time(NULL);
                        got=true;
                        break;
                    }
                }
                if(!got) {
                    c_send_later s;
                    s.socket=socket;
                    s.last_send=my_time(NULL);
                    send_later.push_back(s);
                    goto again;
                }
                c_data d;
                d.closed=false;
                d.len=size;
                d.data=MY_NEW char[size];
                memcpy(d.data,data,size);
                (*i1).data.push_back(d);

                error_code=0;
                return size;
            }
        }
    }
}

/*!
    \brief Reads data from connected socket
    \author VooDooMan
    \version 2
    \date 2005
    \param socket Socket handle of connected socket
    \param data Buffer to be filled up to "size" bytes
    \param size Size of buffer (maximum bytes to be read)
    \param error_code Returns OS's socket error
    \param connection_closed Returns true if data was successfully read, but remote host closed the connection (this is not error and read of data is working correctly)
    \return Returns amount of read data in bytes
*/
size_t sock_read(c_socket& socket, char* data, size_t size, int& error_code, bool& connection_closed)
{
    c_mutex_locker lock(m_sockets);

    error_code=0;
    connection_closed=false;
    {
        vector<c_receive_later>::iterator i1;
        for(i1=receive_later.begin(); i1!=receive_later.end(); ++i1) {
            if((*i1).socket==socket) {
                if((*i1).socket.last_io<socket.last_io)
                    (*i1).socket.last_io=socket.last_io;
                else
                    socket.last_io=(*i1).socket.last_io;

                if((*i1).data.begin()==(*i1).data.end() && (*i1).last_error) {
                    error_code=(*i1).last_error;
                    return 0;
                }

                list<c_data>::iterator i2;
                connection_closed=false;
again:
                for(i2=(*i1).data.begin(); i2!=(*i1).data.end(); ++i2) {
                    (*i1).last_recv=my_time(NULL);

                    error_code=0;
                    if((*i2).closed) {
                        connection_closed=true;
                        (*i1).socket.closed=true;
                        (*i1).data.pop_front();
                        goto again;
                    }
                    connection_closed=false;

                    size_t l=size;
                    if((*i2).len<=l) {
                        l=(*i2).len;
                        if((*i2).len)
                            memcpy(data,(*i2).data,l);
                        MY_DELETE_A((*i2).data);
                        (*i1).data.pop_front();
                        error_code=0;
                        return l;
                    }
                    if((*i2).len>l) {
                        memcpy(data,(*i2).data,l);
                        char* buf=MY_NEW char[(*i2).len-l];
                        memcpy(buf,&((*i2).data[l]),(*i2).len-l);
                        MY_DELETE_A((*i2).data);
                        (*i2).len-=l;
                        (*i2).data=buf;
                        error_code=0;
                        return l;
                    }
                }
                error_code=0;
                return 0;
            }
        }
    }

    return 0;
}

/*!
    \brief Puts buffer back to queue as if it weren't read
    \author VooDooMan
    \version 2
    \date 2005
    \param socket Socket handle of connected socket
    \param data Buffer to be filled up to "size" bytes
    \param size Size of buffer (maximum bytes to be read)
*/
void sock_unread(c_socket& socket, char* data, size_t size)
{
    c_mutex_locker lock(m_sockets);

    {
        vector<c_receive_later>::iterator i1;
        for(i1=receive_later.begin(); i1!=receive_later.end(); ++i1) {
            if((*i1).socket==socket) {
                if((*i1).socket.last_io<socket.last_io)
                    (*i1).socket.last_io=socket.last_io;
                else
                    socket.last_io=(*i1).socket.last_io;
                c_data d;
                d.closed=false;
                d.data=data;
                d.len=size;
                (*i1).data.push_front(d);
                return;
            }
        }
    }
}

/*!
    \brief Get size of input buffer on socket
    \author VooDooMan
    \version 1
    \date 2004
    \param socket Socket handle
    \param closed Returns true if connection was closed
    \return Number of bytes waiting to receive
*/
size_t sock_get_receive_size(c_socket& socket, bool& closed)
{
    c_mutex_locker lock(m_sockets);

    size_t result=0;
    closed=false;

    vector<c_receive_later>::iterator i1;
    for(i1=receive_later.begin(); i1!=receive_later.end(); ++i1) {
        if((*i1).socket==socket) {
            if((*i1).has_error)
                closed=true;
            if((*i1).socket.last_io<socket.last_io)
                (*i1).socket.last_io=socket.last_io;
            else
                socket.last_io=(*i1).socket.last_io;

            list<c_data>::iterator i2;
            for(i2=(*i1).data.begin(); i2!=(*i1).data.end(); ++i2) {
                if(!(*i2).closed)
                    result+=(*i2).len;
                else
                    closed=true;
            }
            break;
        }
    }
    if(0!=result)
        closed=false; // receive/send all, then report closed state
    return result;
}

void sock_erase_later(c_socket& socket);
void sock_flush_later(c_socket& socket);

#if defined(_MSC_VER)
#   pragma warning(push)
#   pragma warning(disable:4702)    // warning C4702: unreachable code
#   pragma warning(disable:4127)    // warning C4127: conditional expression is constant
#endif

/*!
    \brief Sends and receive all pending data (in the cache), should be called in loop
    \author VooDooMan
    \version 3
    \date 2005, 2008-03-13
    \warning Not thread-safe; we use static variable as a buffer and mutex, on Windoze it is recursive critical section! Beware of that!
*/
void sock_send_cache(bool force_garbaging)
{
    c_mutex_locker lock(m_sockets);

#ifdef _DEBUG
    {
        c_mutex_locker lock(m_sockets);

        map<unsigned,c_socket_hndl>::iterator i1;
        for(i1=sockets.begin(); i1!=sockets.end(); ++i1) {
            if((*i1).second.closed && (*i1).second.closed+10*60<my_time(NULL)) {
                sockets.erase(i1);
                break;
            }
        }
    }
#endif

    enum {
        MAX_LEN = 1024*64,
    };

    static char _tmp_[MAX_LEN];

    vector<c_receive_later>::iterator i2;
    for(i2=receive_later.begin(); i2!=receive_later.end(); ++i2) {
        if(!(*i2).socket.cmp())
            continue;

        {
            {
                // do a dummy operation
                int bsd_errno;
                size_t res=sock_lo_get_input_amount((*i2).socket,_tmp_,MAX_LEN,bsd_errno);

                int error_code=SOCK_ERRNO;
                if(res!=static_cast<size_t>(MY_SOCKET_ERROR))
                    error_code=0;
                if(error_code && error_code!=MY_EAGAIN_WOULD_BLOCK) {
                    if(!(*i2).has_error) {
                        (*i2).has_error=true;
                        (*i2).last_error=error_code;
                        if(!(*i2).bsd_errno && bsd_errno!=EAGAIN)
                            (*i2).bsd_errno=bsd_errno;
                    }
                }
            }
        }

        if((*i2).using_window)
            continue;

        char tmp[1024];
        socklen_t tmp2=1024;
        getsockopt((*i2).socket.deref(__FILE__,__LINE__),SOL_SOCKET,SO_ERROR,tmp,&tmp2);

        for(;;) {
            {
                // do a dummy operation
                int bsd_errno;
                size_t res=sock_lo_get_input_amount((*i2).socket,_tmp_,MAX_LEN,bsd_errno);

                int error_code=SOCK_ERRNO;
                if(res!=static_cast<size_t>(MY_SOCKET_ERROR))
                    error_code=0;
                if(error_code && error_code!=MY_EAGAIN_WOULD_BLOCK) {
                    if(!(*i2).has_error) {
                        (*i2).has_error=true;
                        (*i2).last_error=error_code;
                        if(!(*i2).bsd_errno && bsd_errno!=EAGAIN)
                            (*i2).bsd_errno=bsd_errno;
                    }
                }
            }

            fd_set a;
            timeval t;

            t.tv_sec=0;
            t.tv_usec=1;

            FD_ZERO(&a);
            FD_SET((*i2).socket.deref(__FILE__,__LINE__),&a);

            int i=select((int)(((*i2).socket).deref(__FILE__,__LINE__))+1,&a,NULL,NULL,&t);
            if(i==0)
                break;

            unsigned int max_size=1024*64;
            char* buf=MY_NEW char[max_size];
            int bsd_errno;
            size_t size_=sock_lo_recv((*i2).socket,buf,max_size,bsd_errno);
            if(size_==static_cast<size_t>(MY_SOCKET_ERROR)) {
                MY_DELETE_A(buf);
                int error_code=SOCK_ERRNO;
                if(error_code==MY_EAGAIN_WOULD_BLOCK)
                    break;
                if(!(*i2).has_error) {
                    (*i2).has_error=true;
                    (*i2).last_error=error_code;
                    if(!(*i2).bsd_errno && bsd_errno!=EAGAIN)
                        (*i2).bsd_errno=bsd_errno;
                    break;
                }
                break;
            }
            if(size_>0) {
                (*i2).socket.last_io=my_time(NULL);
                (*i2).last_recv=my_time(NULL);
            }
            if(size_==0) {
                MY_DELETE_A(buf);
                if(!(*i2).has_error) {
                    c_data d;
                    d.closed=true;
                    //d.error=0;
                    (*i2).data.push_back(d);
                    (*i2).has_error=true;
                }
                break;
            }
            if(size_) {
                c_data d;
                d.closed=false;
                d.len=size_;
                d.data=buf;
                (*i2).data.push_back(d);
            }
        }
    }

    vector<c_send_later>::iterator i1;
    for(i1=send_later.begin(); i1!=send_later.end(); ++i1) {
    again:
        {
            if(!(*i1).socket.cmp())
                continue;

            list<c_data>::iterator i2;
            for(i2=(*i1).data.begin(); i2!=(*i1).data.end(); ++i2) {
                size_t l=1024;
                if((*i2).len<l)
                    l=(*i2).len;
                int bsd_errno;
                size_t Sent=sock_lo_send((*i1).socket,(*i2).data,l,bsd_errno);
                if(Sent==static_cast<size_t>(MY_SOCKET_ERROR)) {
                    int la=SOCK_ERRNO;
                    if(la==MY_EAGAIN_WOULD_BLOCK)
                        break;
                    {
                        vector<c_receive_later>::iterator i2;
                        for(i2=receive_later.begin(); i2!=receive_later.end(); ++i2)
                            if((*i2).socket==(*i1).socket) {
                                (*i2).has_error=true;
                                if((*i2).last_error==0)
                                    (*i2).last_error=la;
                                if(!(*i2).bsd_errno && bsd_errno!=EAGAIN)
                                    (*i2).bsd_errno=bsd_errno;
                                break;
                            }
                    }
                    break;
                }
                if(Sent>0) {
                    (*i1).socket.last_io=my_time(NULL);
                    (*i1).last_send=my_time(NULL);
                }
                if(Sent<(*i2).len) {
                    char* buf=MY_NEW char[(*i2).len-Sent];
                    memcpy(buf,&(*i2).data[Sent],(*i2).len-Sent);
                    MY_DELETE_A((*i2).data);
                    (*i2).len-=Sent;
                    (*i2).data=buf;
                    if(Sent==0)
                        goto next;
                } else {
                    MY_DELETE_A((*i2).data);
                    (*i1).data.pop_front();
                }
                goto again;
            }
        }
    next:
        ;
    }

    {
        c_mutex_locker lock(m_sockets);

        my_time_t now=my_time(NULL);
        vector<pair<c_socket,my_time_t> >::iterator li;
    linger_again:
        for(li=sock_linger.begin(); li!=sock_linger.end(); ++li) {
            if(force_garbaging || (*li).second+10<now) {

                sock_erase_later((*li).first);
                errno=0;

                // to avoid recursion:
                c_socket s=(*li).first;
                sock_linger.erase(li);

                map<unsigned,c_socket_hndl>::iterator i2=sockets.find(s.hndl);
                if(sockets.end()==i2 || (*i2).second.closed) {
                    // probably system time skew caused this...
                    // or process is beeing traced in debugger, so we had
                    //     no chance to delay operation...
                    // don't call 'sock_close()', fix it:
                    goto linger_again;
                }

                sock_close(s,__FILE__,__LINE__);
                goto linger_again;
            }
        }
    }
}

#if defined(_MSC_VER)
#   pragma warning(pop)
#endif

/*!
    \brief Creates cache entries
    \author VooDooMan
    \version 1
    \date 2004
    \param socket Socket handle to create cache for
    \param using_window Are we using window message callback to receive? (Windows(TM) only)
*/
void sock_create_later(c_socket& socket, bool using_window)
{
    c_mutex_locker lock(m_sockets);

#if !defined(_WIN32) && !defined(_WIN64)
    using_window=false;
#endif
    c_receive_later r;
    r.has_error=false;
    r.socket=socket;
    r.socket.closed=false;
    r.using_window=using_window;
    r.last_error=0;
    r.last_recv=my_time(NULL);
    receive_later.push_back(r);

    c_send_later s;
    s.socket=socket;
    s.socket.closed=false;
    s.last_send=my_time(NULL);
    send_later.push_back(s);
}

/*!
    \brief Sets asynchronous mode for the socket
    \author VooDooMan
    \version 2
    \date 2005
    \param socket Socket handle
*/
void sock_async(c_socket& socket)
{
    c_mutex_locker lock(m_sockets);

#if defined(_WIN32) || defined(_WIN64)
#   ifndef __GNUC__
    HWND WindowHandle=AllocateHWnd(WndProc);
#   else
    HWND WindowHandle=AllocateHWnd((void*)WndProc);
#   endif
    socket.wnd=WindowHandle;
    WSAAsyncSelect(socket.deref(__FILE__,__LINE__),WindowHandle,WM_TCPASYNCSELECT,FD_READ | FD_CLOSE);
#endif
    sock_set_nonblocking(socket,true);
}

#if defined(_MSC_VER)
#   pragma warning(push)
#   pragma warning(disable:4702)    // warning C4702: unreachable code
#endif

/*!
    \brief Flushes cache data (receives and sends)
    \author VooDooMan
    \version 2
    \date 2005
    \param socket Socket handle to flush cache for
*/
void sock_flush_later(c_socket& socket)
{
    c_mutex_locker lock(m_sockets);

    vector<c_send_later>::iterator i1;
    for(i1=send_later.begin(); i1!=send_later.end(); ++i1) {
        if((*i1).socket==socket) {
            if((*i1).socket.last_io<socket.last_io)
                (*i1).socket.last_io=socket.last_io;
            else
                socket.last_io=(*i1).socket.last_io;

            (*i1).last_send=my_time(NULL);

            list<c_data>::iterator i2;
        again:
            for(i2=(*i1).data.begin(); i2!=(*i1).data.end(); ++i2) {
                if(!(*i2).closed) {
                    size_t l=1024;
                    if((*i2).len<l)
                        l=(*i2).len;
                    int bsd_errno;
                    size_t Sent=sock_lo_send((*i1).socket,(*i2).data,l,bsd_errno);
                    if(Sent==static_cast<size_t>(MY_SOCKET_ERROR)) {
                        int la=SOCK_ERRNO;
                        if(la!=MY_EAGAIN_WOULD_BLOCK) {
                            vector<c_receive_later>::iterator i3;
                            for(i3=receive_later.begin(); i3!=receive_later.end(); ++i3)
                                if((*i3).socket==(*i1).socket) {
                                    (*i3).has_error=true;
                                    if((*i3).last_error==0)
                                        (*i3).last_error=la;
                                    if(!(*i3).bsd_errno && bsd_errno!=EAGAIN)
                                        (*i3).bsd_errno=bsd_errno;
                                    break;
                                }
                            MY_DELETE_A((*i2).data);
                            (*i1).data.pop_front();
                            goto again;
                        } else {
                            break;
                        }
                    }
                    if(Sent>0)
                        (*i1).socket.last_io=my_time(NULL);
                    if(Sent<(*i2).len) {
                        char* buf=MY_NEW char[(*i2).len-Sent];
                        memcpy(buf,&(*i2).data[Sent],(*i2).len-Sent);
                        MY_DELETE_A((*i2).data);
                        (*i2).len-=Sent;
                        (*i2).data=buf;
                    } else {
                        MY_DELETE_A((*i2).data);
                        (*i1).data.pop_front();
                    }
                    goto again;
                }

                (*i1).data.pop_front();
                goto again;
            }
            break;
        }
    }
}

#if defined(_MSC_VER)
#   pragma warning(pop)
#endif

/*!
    \brief Frees cache data (receives and sends), without send / receive
    \author VooDooMan
    \version 1
    \date 2004
    \param socket Socket handle to free cache for
*/
void sock_erase_later(c_socket& socket)
{
    c_mutex_locker lock(m_sockets);

    {
        vector<c_send_later>::iterator i1;
    again_a:
        for(i1=send_later.begin(); i1!=send_later.end(); ++i1) {
            if((*i1).socket.was_it_you(socket)) {
                list<c_data>::iterator ii;
                while((ii=(*i1).data.begin())!=(*i1).data.end()) {
                    if(!(*ii).closed)
                        MY_DELETE_A((*ii).data);
                    (*i1).data.erase(ii);
                }
                send_later.erase(i1);
                goto again_a;
            }
        }
    }
    {
        vector<c_receive_later>::iterator i2;
    again_b:
        for(i2=receive_later.begin(); i2!=receive_later.end(); ++i2) {
            if((*i2).socket.was_it_you(socket)) {
                if((*i2).socket.last_io<socket.last_io)
                    (*i2).socket.last_io=socket.last_io;
                else
                    socket.last_io=(*i2).socket.last_io;

                list<c_data>::iterator ii;
                while((ii=(*i2).data.begin())!=(*i2).data.end()) {
                    if(!(*ii).closed)
                        MY_DELETE_A((*ii).data);
                    (*i2).data.erase(ii);
                }

                receive_later.erase(i2);
                goto again_b;
            }
        }
    }
}

/*!
    \brief Cancels caching
    \author VooDooMan
    \version 1
    \date 2004
    \param socket Socket handle
*/
void sock_cancel_caching(c_socket& socket)
{
    c_mutex_locker lock(m_sockets);

    sock_flush_later(socket);
    sock_erase_later(socket);
#if defined(_WIN32) || defined(_WIN64)
    WSAAsyncSelect(socket.deref(__FILE__,__LINE__),socket.wnd,0,0);
#endif
}

/*!
    \brief Returns last read/receive or send timestamp on socket
    \author VooDooMan
    \version 1
    \date 2005
    \param socket Socket handle
    \param mode Set to 'r' to get timestamp of last receive; set to 's' to get time of last send; set to 'a' (default) (stands for "all") to get most recent timestamp on both send and receive
    \return Timestamp of last send or receive on socket
    \warning If argument "mode" contain neither 'r', nor 's', nor 'a', the returned value is zero
*/
my_time_t sock_get_last_io(c_socket& socket, char mode)
{
    c_mutex_locker lock(m_sockets);

    my_time_t last_io=0;

    if(mode=='r') {
        vector<c_receive_later>::iterator i2;
        for(i2=receive_later.begin(); i2!=receive_later.end(); ++i2)
            if((*i2).socket==socket) {
                last_io=(*i2).last_recv;
                break;
            }
    }

    if(mode=='s') {
        vector<c_send_later>::iterator i2;
        for(i2=send_later.begin(); i2!=send_later.end(); ++i2)
            if((*i2).socket==socket) {
                last_io=(*i2).last_send;
                break;
            }
    }

    if(mode=='a') {
        last_io=socket.last_io;

        vector<c_receive_later>::iterator i2;
        for(i2=receive_later.begin(); i2!=receive_later.end(); ++i2)
            if((*i2).socket==socket) {
                if(last_io<(*i2).socket.last_io)
                    last_io=(*i2).socket.last_io;
                if(last_io<(*i2).last_recv)
                    last_io=(*i2).last_recv;
                break;
            }

        vector<c_send_later>::iterator i1;
        for(i1=send_later.begin(); i1!=send_later.end(); ++i1)
            if((*i1).socket==socket) {
                if(last_io<(*i1).socket.last_io)
                    last_io=(*i1).socket.last_io;
                if(last_io<(*i1).last_send)
                    last_io=(*i1).last_send;
                break;
            }
    }

    return last_io;
}

/*!
    \brief Sets last read/receive or send timestamp on socket
    \author VooDooMan
    \version 1
    \date 2006-07-16
    \param socket Socket handle
    \param last_io Set timestamp of last I/O
    \param mode Set to 'r' to set timestamp of last receive; set to 's' to set time of last send
    \return Timestamp of last send or receive on socket
*/
void sock_set_last_io(c_socket& socket, my_time_t last_io, char mode)
{
    c_mutex_locker lock(m_sockets);

    socket.last_io=last_io;
    if(mode=='r') {
        vector<c_receive_later>::iterator i2;
        for(i2=receive_later.begin(); i2!=receive_later.end(); ++i2)
            if((*i2).socket==socket) {
                (*i2).last_recv=last_io;
                break;
            }
    }

    if(mode=='s') {
        vector<c_send_later>::iterator i2;
        for(i2=send_later.begin(); i2!=send_later.end(); ++i2)
            if((*i2).socket==socket) {
                (*i2).last_send=last_io;
                break;
            }
    }
}

/*!
    \brief Discovers type of address
    \author VooDooMan
    \version 1
    \date 2005
    \param host IP of host name
    \retval 0 Given host is DNS name of host
    \retval 4 Given host is numeric IPv4 address
    \retval 6 Given host is numeric IPv6 address
*/
int sock_get_address_type(string host)
{
    if(!host.compare("0.0.0.0")) // special case
        return 4;
    uint32 x=sock_ipv4_from_string(host.c_str());
    if(x)
        return 4;
    if(host.find(":",0)!=string::npos) {
        in_addr6_ addr6;
        if(my_inet_pton(AF_INET6,host.c_str(),&addr6)==1)
            return 6;
    }
    return 0;
}

uint32 sock_ipv4_from_string(const char* str)
{
    uint32 ip=0;
    if(my_inet_pton(AF_INET,str,&ip)!=1)
        ip=0;
    return ip;
}

/*!
    \brief Retrieves statistics about sock I/O cache
    \author VooDooMan
    \version 1
    \date 2006-01-21
    \param socket Socket
    \param to_receive Returns number of bytes waiting in cache to receive them
    \param to_send Returns number of bytes waiting in cache to send them
*/
void sock_cache_stats(c_socket& socket, size_t& to_receive, size_t& to_send)
{
    c_mutex_locker lock(m_sockets);

    to_receive=to_send=0;

    {
        vector<c_receive_later>::iterator i2;
        for(i2=receive_later.begin(); i2!=receive_later.end(); ++i2) {
            if((*i2).socket==socket) {
                list<c_data>::iterator i1;
                for(i1=(*i2).data.begin(); i1!=(*i2).data.end(); ++i1)
                    if(!(*i1).closed)
                        to_receive+=(*i1).len;
                break;
            }
        }
    }

    {
        vector<c_send_later>::iterator i2;
        for(i2=send_later.begin(); i2!=send_later.end(); ++i2) {
            if((*i2).socket==socket) {
                list<c_data>::iterator i1;
                for(i1=(*i2).data.begin(); i1!=(*i2).data.end(); ++i1)
                    if(!(*i1).closed)
                        to_send+=(*i1).len;
                break;
            }
        }
    }
}

/*!
    \brief Returns OS's error string from socket I/O error
    \author VooDooMan
    \version 2
    \date 2004, 2005-05-19
    \param code Socket I/O error
    \param type Type of error code
    \return OS's error string (must be MY_DELETE_A'd)
    \warning Result must be MY_DELETE_A'd
*/
char* sock_error(int code, c_sock_log::e_type type)
{
    enum {
        MAX_ERR_LENGTH = 1024,
    };

    char* result=MY_NEW char[MAX_ERR_LENGTH+1];
    result[0]='\0';
    switch((unsigned)code) {
        case MY_SELECT_ERROR:
            strcpy(result,"(Unknown error, based on select() fuction)");
            return result;
        case MY_NO_IPV6:
            strcpy(result,"(Not compiled with IPv6 support)");
            return result;
        case MY_SSL_ERROR:
            strcpy(result,"(SSL error)");
            return result;
        case MY_UNKNOWN_ERROR:
            strcpy(result,"(Unknown error MY_UNKNOWN_ERROR)");
            return result;
        case MY_DNS_ERROR:
            strcpy(result,"(DNS lookup failed)");
            return result;
    }

    if(c_sock_log::TYPE_GAI==type) {
        strncpy(result,gai_strerror(code),MAX_ERR_LENGTH);
        result[MAX_ERR_LENGTH]='\0';
        return result;
    }

#if defined(_WIN32) || defined(_WIN64)
    switch(code) {
        case 0:
            strcpy(result,"No error, success");
            break;
#ifdef DNS_ERROR_RCODE_FORMAT_ERROR
        case DNS_ERROR_RCODE_FORMAT_ERROR:
            strcpy(result,"DNS server unable to interpret format");
            break;
#endif
#ifdef DNS_ERROR_RCODE_SERVER_FAILURE
        case DNS_ERROR_RCODE_SERVER_FAILURE:
            strcpy(result,"DNS server failure");
            break;
#endif
#ifdef DNS_ERROR_RCODE_NAME_ERROR
        case DNS_ERROR_RCODE_NAME_ERROR:
            strcpy(result,"DNS name does not exist");
            break;
#endif
#ifdef DNS_ERROR_RCODE_NOT_IMPLEMENTED
        case DNS_ERROR_RCODE_NOT_IMPLEMENTED:
            strcpy(result,"DNS request not supported by name server");
            break;
#endif
#ifdef DNS_ERROR_RCODE_REFUSED
        case DNS_ERROR_RCODE_REFUSED:
            strcpy(result,"DNS operation refused");
            break;
#endif
#ifdef DNS_ERROR_RCODE_YXDOMAIN
        case DNS_ERROR_RCODE_YXDOMAIN:
            strcpy(result,"DNS name that ought not exist, does exist");
            break;
#endif
#ifdef DNS_ERROR_RCODE_YXRRSET
        case DNS_ERROR_RCODE_YXRRSET:
            strcpy(result,"DNS RR set that ought not exist, does exist");
            break;
#endif
#ifdef DNS_ERROR_RCODE_NXRRSET
        case DNS_ERROR_RCODE_NXRRSET:
            strcpy(result,"DNS RR set that ought to exist, does not exist");
            break;
#endif
#ifdef DNS_ERROR_RCODE_NOTAUTH
        case DNS_ERROR_RCODE_NOTAUTH:
            strcpy(result,"DNS server not authoritative for zone");
            break;
#endif
#ifdef DNS_ERROR_RCODE_NOTZONE
        case DNS_ERROR_RCODE_NOTZONE:
            strcpy(result,"DNS name in update or prereq is not in zone");
            break;
#endif
#ifdef DNS_ERROR_RCODE_BADSIG
        case DNS_ERROR_RCODE_BADSIG:
            strcpy(result,"DNS signature failed to verify");
            break;
#endif
#ifdef DNS_ERROR_RCODE_BADKEY
        case DNS_ERROR_RCODE_BADKEY:
            strcpy(result,"DNS bad key");
            break;
#endif
#ifdef DNS_ERROR_RCODE_BADTIME
        case DNS_ERROR_RCODE_BADTIME:
            strcpy(result,"DNS signature validity expired");
            break;
#endif
#ifdef DNS_INFO_NO_RECORDS
        case DNS_INFO_NO_RECORDS:
            strcpy(result,"No records found for given DNS query");
            break;
#endif
#ifdef DNS_ERROR_BAD_PACKET
        case DNS_ERROR_BAD_PACKET:
            strcpy(result,"Bad DNS packet");
            break;
#endif
#ifdef DNS_ERROR_NO_PACKET
        case DNS_ERROR_NO_PACKET:
            strcpy(result,"No DNS packet");
            break;
#endif
#ifdef DNS_ERROR_RCODE
        case DNS_ERROR_RCODE:
            strcpy(result,"DNS error, check rcode");
            break;
#endif
#ifdef DNS_ERROR_UNSECURE_PACKET
        case DNS_ERROR_UNSECURE_PACKET:
            strcpy(result,"Unsecured DNS packet");
            break;
#endif
#ifdef DNS_ERROR_INVALID_TYPE
        case DNS_ERROR_INVALID_TYPE:
            strcpy(result,"Invalid DNS type");
            break;
#endif
#ifdef DNS_ERROR_INVALID_IP_ADDRESS
        case DNS_ERROR_INVALID_IP_ADDRESS:
            strcpy(result,"Invalid IP address");
            break;
#endif
#ifdef DNS_ERROR_INVALID_PROPERTY
        case DNS_ERROR_INVALID_PROPERTY:
            strcpy(result,"Invalid property");
            break;
#endif
#ifdef DNS_ERROR_TRY_AGAIN_LATER
        case DNS_ERROR_TRY_AGAIN_LATER:
            strcpy(result,"Try DNS operation again later");
            break;
#endif
#ifdef DNS_ERROR_NOT_UNIQUE
        case DNS_ERROR_NOT_UNIQUE:
            strcpy(result,"Record for given name and type is not unique");
            break;
#endif
#ifdef DNS_ERROR_NON_RFC_NAME
        case DNS_ERROR_NON_RFC_NAME:
            strcpy(result,"DNS name does not comply with RFC specifications");
            break;
#endif
#ifdef DNS_STATUS_FQDN
        case DNS_STATUS_FQDN:
            strcpy(result,"DNS name is a fully-qualified DNS name");
            break;
#endif
#ifdef DNS_STATUS_DOTTED_NAME
        case DNS_STATUS_DOTTED_NAME:
            strcpy(result,"DNS name is dotted (multi-label)");
            break;
#endif
#ifdef DNS_STATUS_SINGLE_PART_NAME
        case DNS_STATUS_SINGLE_PART_NAME:
            strcpy(result,"DNS name is a single-part name");
            break;
#endif
#ifdef DNS_ERROR_INVALID_NAME_CHAR
        case DNS_ERROR_INVALID_NAME_CHAR:
            strcpy(result,"DNS name contains an invalid character");
            break;
#endif
#ifdef DNS_ERROR_NUMERIC_NAME
        case DNS_ERROR_NUMERIC_NAME:
            strcpy(result,"DNS name is entirely numeric");
            break;
#endif
#ifdef DNS_ERROR_NOT_ALLOWED_ON_ROOT_SERVER
        case DNS_ERROR_NOT_ALLOWED_ON_ROOT_SERVER:
            strcpy(result,"The operation requested is not permitted on a DNS root server");
            break;
#endif
#ifdef DNS_ERROR_NOT_ALLOWED_UNDER_DELEGATION
        case DNS_ERROR_NOT_ALLOWED_UNDER_DELEGATION:
            strcpy(result,"The record could not be created because this part of the DNS namespace has been delegated to another server");
            break;
#endif
#ifdef DNS_ERROR_CANNOT_FIND_ROOT_HINTS
        case DNS_ERROR_CANNOT_FIND_ROOT_HINTS:
            strcpy(result,"The DNS server could not find a set of root hints");
            break;
#endif
#ifdef DNS_ERROR_INCONSISTENT_ROOT_HINTS
        case DNS_ERROR_INCONSISTENT_ROOT_HINTS:
            strcpy(result,"The DNS server found root hints but they were not consistent across all adapters");
            break;
#endif
#ifdef DNS_ERROR_DWORD_VALUE_TOO_SMALL
        case DNS_ERROR_DWORD_VALUE_TOO_SMALL:
            strcpy(result,"The specified value is too small for this parameter");
            break;
#endif
#ifdef DNS_ERROR_DWORD_VALUE_TOO_LARGE
        case DNS_ERROR_DWORD_VALUE_TOO_LARGE:
            strcpy(result,"The specified value is too large for this parameter");
            break;
#endif
#ifdef DNS_ERROR_BACKGROUND_LOADING
        case DNS_ERROR_BACKGROUND_LOADING:
            strcpy(result,"This operation is not allowed while the DNS server is loading zones in the background. Please try again later");
            break;
#endif
#ifdef DNS_ERROR_NOT_ALLOWED_ON_RODC
        case DNS_ERROR_NOT_ALLOWED_ON_RODC:
            strcpy(result,"The operation requested is not permitted on against a DNS server running on a read-only DC");
            break;
#endif
#ifdef DNS_ERROR_NOT_ALLOWED_UNDER_DNAME
        case DNS_ERROR_NOT_ALLOWED_UNDER_DNAME:
            strcpy(result,"No data is allowed to exist underneath a DNAME record");
            break;
#endif
#ifdef DNS_ERROR_DELEGATION_REQUIRED
        case DNS_ERROR_DELEGATION_REQUIRED:
            strcpy(result,"This operation requires credentials delegation");
            break;
#endif
#ifdef DNS_ERROR_ZONE_DOES_NOT_EXIST
        case DNS_ERROR_ZONE_DOES_NOT_EXIST:
            strcpy(result,"DNS zone does not exist");
            break;
#endif
#ifdef DNS_ERROR_NO_ZONE_INFO
        case DNS_ERROR_NO_ZONE_INFO:
            strcpy(result,"DNS zone information not available");
            break;
#endif
#ifdef DNS_ERROR_INVALID_ZONE_OPERATION
        case DNS_ERROR_INVALID_ZONE_OPERATION:
            strcpy(result,"Invalid operation for DNS zone");
            break;
#endif
#ifdef DNS_ERROR_ZONE_CONFIGURATION_ERROR
        case DNS_ERROR_ZONE_CONFIGURATION_ERROR:
            strcpy(result,"Invalid DNS zone configuration");
            break;
#endif
#ifdef DNS_ERROR_ZONE_HAS_NO_SOA_RECORD
        case DNS_ERROR_ZONE_HAS_NO_SOA_RECORD:
            strcpy(result,"DNS zone has no start of authority (SOA) record");
            break;
#endif
#ifdef DNS_ERROR_ZONE_HAS_NO_NS_RECORDS
        case DNS_ERROR_ZONE_HAS_NO_NS_RECORDS:
            strcpy(result,"DNS zone has no Name Server (NS) record");
            break;
#endif
#ifdef DNS_ERROR_ZONE_LOCKED
        case DNS_ERROR_ZONE_LOCKED:
            strcpy(result,"DNS zone is locked");
            break;
#endif
#ifdef DNS_ERROR_ZONE_CREATION_FAILED
        case DNS_ERROR_ZONE_CREATION_FAILED:
            strcpy(result,"DNS zone creation failed");
            break;
#endif
#ifdef DNS_ERROR_ZONE_ALREADY_EXISTS
        case DNS_ERROR_ZONE_ALREADY_EXISTS:
            strcpy(result,"DNS zone already exists");
            break;
#endif
#ifdef DNS_ERROR_AUTOZONE_ALREADY_EXISTS
        case DNS_ERROR_AUTOZONE_ALREADY_EXISTS:
            strcpy(result,"DNS automatic zone already exists");
            break;
#endif
#ifdef DNS_ERROR_INVALID_ZONE_TYPE
        case DNS_ERROR_INVALID_ZONE_TYPE:
            strcpy(result,"Invalid DNS zone type");
            break;
#endif
#ifdef DNS_ERROR_SECONDARY_REQUIRES_MASTER_IP
        case DNS_ERROR_SECONDARY_REQUIRES_MASTER_IP:
            strcpy(result,"Secondary DNS zone requires master IP address");
            break;
#endif
#ifdef DNS_ERROR_ZONE_NOT_SECONDARY
        case DNS_ERROR_ZONE_NOT_SECONDARY:
            strcpy(result,"DNS zone not secondary");
            break;
#endif
#ifdef DNS_ERROR_NEED_SECONDARY_ADDRESSES
        case DNS_ERROR_NEED_SECONDARY_ADDRESSES:
            strcpy(result,"Need secondary IP address");
            break;
#endif
#ifdef DNS_ERROR_WINS_INIT_FAILED
        case DNS_ERROR_WINS_INIT_FAILED:
            strcpy(result,"WINS initialization failed");
            break;
#endif
#ifdef DNS_ERROR_NEED_WINS_SERVERS
        case DNS_ERROR_NEED_WINS_SERVERS:
            strcpy(result,"Need WINS servers");
            break;
#endif
#ifdef DNS_ERROR_NBSTAT_INIT_FAILED
        case DNS_ERROR_NBSTAT_INIT_FAILED:
            strcpy(result,"NBTSTAT initialization call failed");
            break;
#endif
#ifdef DNS_ERROR_SOA_DELETE_INVALID
        case DNS_ERROR_SOA_DELETE_INVALID:
            strcpy(result,"Invalid delete of start of authority (SOA)");
            break;
#endif
#ifdef DNS_ERROR_FORWARDER_ALREADY_EXISTS
        case DNS_ERROR_FORWARDER_ALREADY_EXISTS:
            strcpy(result,"A conditional forwarding zone already exists for that name");
            break;
#endif
#ifdef DNS_ERROR_ZONE_REQUIRES_MASTER_IP
        case DNS_ERROR_ZONE_REQUIRES_MASTER_IP:
            strcpy(result,"This zone must be configured with one or more master DNS server IP addresses");
            break;
#endif
#ifdef DNS_ERROR_ZONE_IS_SHUTDOWN
        case DNS_ERROR_ZONE_IS_SHUTDOWN:
            strcpy(result,"The operation cannot be performed because this zone is shutdown");
            break;
#endif
#ifdef DNS_ERROR_PRIMARY_REQUIRES_DATAFILE
        case DNS_ERROR_PRIMARY_REQUIRES_DATAFILE:
            strcpy(result,"Primary DNS zone requires datafile");
            break;
#endif
#ifdef DNS_ERROR_INVALID_DATAFILE_NAME
        case DNS_ERROR_INVALID_DATAFILE_NAME:
            strcpy(result,"Invalid datafile name for DNS zone");
            break;
#endif
#ifdef DNS_ERROR_DATAFILE_OPEN_FAILURE
        case DNS_ERROR_DATAFILE_OPEN_FAILURE:
            strcpy(result,"Failed to open datafile for DNS zone");
            break;
#endif
#ifdef DNS_ERROR_FILE_WRITEBACK_FAILED
        case DNS_ERROR_FILE_WRITEBACK_FAILED:
            strcpy(result,"Failed to write datafile for DNS zone");
            break;
#endif
#ifdef DNS_ERROR_DATAFILE_PARSING
        case DNS_ERROR_DATAFILE_PARSING:
            strcpy(result,"Failure while reading datafile for DNS zone");
            break;
#endif
#ifdef DNS_ERROR_RECORD_DOES_NOT_EXIST
        case DNS_ERROR_RECORD_DOES_NOT_EXIST:
            strcpy(result,"DNS record does not exist");
            break;
#endif
#ifdef DNS_ERROR_RECORD_FORMAT
        case DNS_ERROR_RECORD_FORMAT:
            strcpy(result,"DNS record format error");
            break;
#endif
#ifdef DNS_ERROR_NODE_CREATION_FAILED
        case DNS_ERROR_NODE_CREATION_FAILED:
            strcpy(result,"Node creation failure in DNS");
            break;
#endif
#ifdef DNS_ERROR_UNKNOWN_RECORD_TYPE
        case DNS_ERROR_UNKNOWN_RECORD_TYPE:
            strcpy(result,"Unknown DNS record type");
            break;
#endif
#ifdef DNS_ERROR_RECORD_TIMED_OUT
        case DNS_ERROR_RECORD_TIMED_OUT:
            strcpy(result,"DNS record timed out");
            break;
#endif
#ifdef DNS_ERROR_NAME_NOT_IN_ZONE
        case DNS_ERROR_NAME_NOT_IN_ZONE:
            strcpy(result,"Name not in DNS zone");
            break;
#endif
#ifdef DNS_ERROR_CNAME_LOOP
        case DNS_ERROR_CNAME_LOOP:
            strcpy(result,"CNAME loop detected");
            break;
#endif
#ifdef DNS_ERROR_NODE_IS_CNAME
        case DNS_ERROR_NODE_IS_CNAME:
            strcpy(result,"Node is a CNAME DNS record");
            break;
#endif
#ifdef DNS_ERROR_CNAME_COLLISION
        case DNS_ERROR_CNAME_COLLISION:
            strcpy(result,"A CNAME record already exists for given name");
            break;
#endif
#ifdef DNS_ERROR_RECORD_ONLY_AT_ZONE_ROOT
        case DNS_ERROR_RECORD_ONLY_AT_ZONE_ROOT:
            strcpy(result,"Record only at DNS zone root");
            break;
#endif
#ifdef DNS_ERROR_RECORD_ALREADY_EXISTS
        case DNS_ERROR_RECORD_ALREADY_EXISTS:
            strcpy(result,"DNS record already exists");
            break;
#endif
#ifdef DNS_ERROR_SECONDARY_DATA
        case DNS_ERROR_SECONDARY_DATA:
            strcpy(result,"Secondary DNS zone data error");
            break;
#endif
#ifdef DNS_ERROR_NO_CREATE_CACHE_DATA
        case DNS_ERROR_NO_CREATE_CACHE_DATA:
            strcpy(result,"Could not create DNS cache data");
            break;
#endif
#ifdef DNS_ERROR_NAME_DOES_NOT_EXIST
        case DNS_ERROR_NAME_DOES_NOT_EXIST:
            strcpy(result,"DNS name does not exist");
            break;
#endif
#ifdef DNS_WARNING_PTR_CREATE_FAILED
        case DNS_WARNING_PTR_CREATE_FAILED:
            strcpy(result,"Could not create pointer (PTR) record");
            break;
#endif
#ifdef DNS_WARNING_DOMAIN_UNDELETED
        case DNS_WARNING_DOMAIN_UNDELETED:
            strcpy(result,"DNS domain was undeleted");
            break;
#endif
#ifdef DNS_ERROR_DS_UNAVAILABLE
        case DNS_ERROR_DS_UNAVAILABLE:
            strcpy(result,"The directory service is unavailable");
            break;
#endif
#ifdef DNS_ERROR_DS_ZONE_ALREADY_EXISTS
        case DNS_ERROR_DS_ZONE_ALREADY_EXISTS:
            strcpy(result,"DNS zone already exists in the directory service");
            break;
#endif
#ifdef DNS_ERROR_NO_BOOTFILE_IF_DS_ZONE
        case DNS_ERROR_NO_BOOTFILE_IF_DS_ZONE:
            strcpy(result,"DNS server not creating or reading the boot file for the directory service integrated DNS zone");
            break;
#endif
#ifdef DNS_ERROR_NODE_IS_DNAME
        case DNS_ERROR_NODE_IS_DNAME:
            strcpy(result,"Node is a DNAME DNS record");
            break;
#endif
#ifdef DNS_ERROR_DNAME_COLLISION
        case DNS_ERROR_DNAME_COLLISION:
            strcpy(result,"A DNAME record already exists for given name");
            break;
#endif
#ifdef DNS_ERROR_ALIAS_LOOP
        case DNS_ERROR_ALIAS_LOOP:
            strcpy(result,"An alias loop has been detected with either CNAME or DNAME records");
            break;
#endif
#ifdef DNS_INFO_AXFR_COMPLETE
        case DNS_INFO_AXFR_COMPLETE:
            strcpy(result,"DNS AXFR (zone transfer) complete");
            break;
#endif
#ifdef DNS_ERROR_AXFR
        case DNS_ERROR_AXFR:
            strcpy(result,"DNS zone transfer failed");
            break;
#endif
#ifdef DNS_INFO_ADDED_LOCAL_WINS
        case DNS_INFO_ADDED_LOCAL_WINS:
            strcpy(result,"Added local WINS server");
            break;
#endif
#ifdef DNS_STATUS_CONTINUE_NEEDED
        case DNS_STATUS_CONTINUE_NEEDED:
            strcpy(result,"Secure update call needs to continue update request");
            break;
#endif
#ifdef DNS_ERROR_NO_TCPIP
        case DNS_ERROR_NO_TCPIP:
            strcpy(result,"TCP/IP network protocol not installed");
            break;
#endif
#ifdef DNS_ERROR_NO_DNS_SERVERS
        case DNS_ERROR_NO_DNS_SERVERS:
            strcpy(result,"No DNS servers configured for local system");
            break;
#endif
#ifdef DNS_ERROR_DP_DOES_NOT_EXIST
        case DNS_ERROR_DP_DOES_NOT_EXIST:
            strcpy(result,"The specified directory partition does not exist");
            break;
#endif
#ifdef DNS_ERROR_DP_ALREADY_EXISTS
        case DNS_ERROR_DP_ALREADY_EXISTS:
            strcpy(result,"The specified directory partition already exists");
            break;
#endif
#ifdef DNS_ERROR_DP_NOT_ENLISTED
        case DNS_ERROR_DP_NOT_ENLISTED:
            strcpy(result,"This DNS server is not enlisted in the specified directory partition");
            break;
#endif
#ifdef DNS_ERROR_DP_ALREADY_ENLISTED
        case DNS_ERROR_DP_ALREADY_ENLISTED:
            strcpy(result,"This DNS server is already enlisted in the specified directory partition");
            break;
#endif
#ifdef DNS_ERROR_DP_NOT_AVAILABLE
        case DNS_ERROR_DP_NOT_AVAILABLE:
            strcpy(result,"The directory partition is not available at this time. Please wait a few minutes and try again");
            break;
#endif
#ifdef DNS_ERROR_DP_FSMO_ERROR
        case DNS_ERROR_DP_FSMO_ERROR:
            strcpy(result,"The application directory partition operation failed. The domain controller holding the domain naming master role is down or unable to service the request or is not running Windows Server 2003");
            break;
#endif
#ifdef WSAEINTR
        case WSAEINTR:
            strcpy(result,"A blocking operation was interrupted by a call to WSACancelBlockingCall");
            break;
#endif
#ifdef WSAEBADF
        case WSAEBADF:
            strcpy(result,"The file handle supplied is not valid");
            break;
#endif
#ifdef WSAEACCES
        case WSAEACCES:
            strcpy(result,"An attempt was made to access a socket in a way forbidden by its access permissions");
            break;
#endif
#ifdef WSAEFAULT
        case WSAEFAULT:
            strcpy(result,"The system detected an invalid pointer address in attempting to use a pointer argument in a call");
            break;
#endif
#ifdef WSAEINVAL
        case WSAEINVAL:
            strcpy(result,"An invalid argument was supplied");
            break;
#endif
#ifdef WSAEMFILE
        case WSAEMFILE:
            strcpy(result,"Too many open sockets");
            break;
#endif
#ifdef WSAEWOULDBLOCK
        case WSAEWOULDBLOCK:
            strcpy(result,"A non-blocking socket operation could not be completed immediately");
            break;
#endif
#ifdef WSAEINPROGRESS
        case WSAEINPROGRESS:
            strcpy(result,"A blocking operation is currently executing");
            break;
#endif
#ifdef WSAEALREADY
        case WSAEALREADY:
            strcpy(result,"An operation was attempted on a non-blocking socket that already had an operation in progress");
            break;
#endif
#ifdef WSAENOTSOCK
        case WSAENOTSOCK:
            strcpy(result,"An operation was attempted on something that is not a socket");
            break;
#endif
#ifdef WSAEDESTADDRREQ
        case WSAEDESTADDRREQ:
            strcpy(result,"A required address was omitted from an operation on a socket");
            break;
#endif
#ifdef WSAEMSGSIZE
        case WSAEMSGSIZE:
            strcpy(result,"A message sent on a datagram socket was larger than the internal message buffer or some other network limit, or the buffer used to receive a datagram into was smaller than the datagram itself");
            break;
#endif
#ifdef WSAEPROTOTYPE
        case WSAEPROTOTYPE:
            strcpy(result,"A protocol was specified in the socket function call that does not support the semantics of the socket type requested");
            break;
#endif
#ifdef WSAENOPROTOOPT
        case WSAENOPROTOOPT:
            strcpy(result,"An unknown, invalid, or unsupported option or level was specified in a getsockopt or setsockopt call");
            break;
#endif
#ifdef WSAEPROTONOSUPPORT
        case WSAEPROTONOSUPPORT:
            strcpy(result,"The requested protocol has not been configured into the system, or no implementation for it exists");
            break;
#endif
#ifdef WSAESOCKTNOSUPPORT
        case WSAESOCKTNOSUPPORT:
            strcpy(result,"The support for the specified socket type does not exist in this address family");
            break;
#endif
#ifdef WSAEOPNOTSUPP
        case WSAEOPNOTSUPP:
            strcpy(result,"The attempted operation is not supported for the type of object referenced");
            break;
#endif
#ifdef WSAEPFNOSUPPORT
        case WSAEPFNOSUPPORT:
            strcpy(result,"The protocol family has not been configured into the system or no implementation for it exists");
            break;
#endif
#ifdef WSAEAFNOSUPPORT
        case WSAEAFNOSUPPORT:
            strcpy(result,"An address incompatible with the requested protocol was used");
            break;
#endif
#ifdef WSAEADDRINUSE
        case WSAEADDRINUSE:
            strcpy(result,"Only one usage of each socket address (protocol/network address/port) is normally permitted");
            break;
#endif
#ifdef WSAEADDRNOTAVAIL
        case WSAEADDRNOTAVAIL:
            strcpy(result,"The requested address is not valid in its context");
            break;
#endif
#ifdef WSAENETDOWN
        case WSAENETDOWN:
            strcpy(result,"A socket operation encountered a dead network");
            break;
#endif
#ifdef WSAENETUNREACH
        case WSAENETUNREACH:
            strcpy(result,"A socket operation was attempted to an unreachable network");
            break;
#endif
#ifdef WSAENETRESET
        case WSAENETRESET:
            strcpy(result,"The connection has been broken due to keep-alive activity detecting a failure while the operation was in progress");
            break;
#endif
#ifdef WSAECONNABORTED
        case WSAECONNABORTED:
            strcpy(result,"An established connection was aborted by the software in your host machine");
            break;
#endif
#ifdef WSAECONNRESET
        case WSAECONNRESET:
            strcpy(result,"An existing connection was forcibly closed by the remote host");
            break;
#endif
#ifdef WSAENOBUFS
        case WSAENOBUFS:
            strcpy(result,"An operation on a socket could not be performed because the system lacked sufficient buffer space or because a queue was full");
            break;
#endif
#ifdef WSAEISCONN
        case WSAEISCONN:
            strcpy(result,"A connect request was made on an already connected socket");
            break;
#endif
#ifdef WSAENOTCONN
        case WSAENOTCONN:
            strcpy(result,"A request to send or receive data was disallowed because the socket is not connected and (when sending on a datagram socket using a sendto call) no address was supplied");
            break;
#endif
#ifdef WSAESHUTDOWN
        case WSAESHUTDOWN:
            strcpy(result,"A request to send or receive data was disallowed because the socket had already been shut down in that direction with a previous shutdown call");
            break;
#endif
#ifdef WSAETOOMANYREFS
        case WSAETOOMANYREFS:
            strcpy(result,"Too many references to some kernel object");
            break;
#endif
#ifdef WSAETIMEDOUT
        case WSAETIMEDOUT:
            strcpy(result,"A connection attempt failed because the connected party did not properly respond after a period of time, or established connection failed because connected host has failed to respond");
            break;
#endif
#ifdef WSAECONNREFUSED
        case WSAECONNREFUSED:
            strcpy(result,"No connection could be made because the target machine actively refused it");
            break;
#endif
#ifdef WSAELOOP
        case WSAELOOP:
            strcpy(result,"Cannot translate name");
            break;
#endif
#ifdef WSAENAMETOOLONG
        case WSAENAMETOOLONG:
            strcpy(result,"Name component or name was too long");
            break;
#endif
#ifdef WSAEHOSTDOWN
        case WSAEHOSTDOWN:
            strcpy(result,"A socket operation failed because the destination host was down");
            break;
#endif
#ifdef WSAEHOSTUNREACH
        case WSAEHOSTUNREACH:
            strcpy(result,"A socket operation was attempted to an unreachable host");
            break;
#endif
#ifdef WSAENOTEMPTY
        case WSAENOTEMPTY:
            strcpy(result,"Cannot remove a directory that is not empty");
            break;
#endif
#ifdef WSAEPROCLIM
        case WSAEPROCLIM:
            strcpy(result,"A Windows Sockets implementation may have a limit on the number of applications that may use it simultaneously");
            break;
#endif
#ifdef WSAEUSERS
        case WSAEUSERS:
            strcpy(result,"Ran out of quota");
            break;
#endif
#ifdef WSAEDQUOT
        case WSAEDQUOT:
            strcpy(result,"Ran out of disk quota");
            break;
#endif
#ifdef WSAESTALE
        case WSAESTALE:
            strcpy(result,"File handle reference is no longer available");
            break;
#endif
#ifdef WSAEREMOTE
        case WSAEREMOTE:
            strcpy(result,"Item is not available locally");
            break;
#endif
#ifdef WSASYSNOTREADY
        case WSASYSNOTREADY:
            strcpy(result,"WSAStartup cannot function at this time because the underlying system it uses to provide network services is currently unavailable");
            break;
#endif
#ifdef WSAVERNOTSUPPORTED
        case WSAVERNOTSUPPORTED:
            strcpy(result,"The Windows Sockets version requested is not supported");
            break;
#endif
#ifdef WSANOTINITIALISED
        case WSANOTINITIALISED:
            strcpy(result,"Either the application has not called WSAStartup, or WSAStartup failed");
            break;
#endif
#ifdef WSAEDISCON
        case WSAEDISCON:
            strcpy(result,"Returned by WSARecv or WSARecvFrom to indicate the remote party has initiated a graceful shutdown sequence");
            break;
#endif
#ifdef WSAENOMORE
        case WSAENOMORE:
            strcpy(result,"No more results can be returned by WSALookupServiceNext");
            break;
#endif
#ifdef WSAECANCELLED
        case WSAECANCELLED:
            strcpy(result,"A call to WSALookupServiceEnd was made while this call was still processing. The call has been canceled");
            break;
#endif
#ifdef WSAEINVALIDPROCTABLE
        case WSAEINVALIDPROCTABLE:
            strcpy(result,"The procedure call table is invalid");
            break;
#endif
#ifdef WSAEINVALIDPROVIDER
        case WSAEINVALIDPROVIDER:
            strcpy(result,"The requested service provider is invalid");
            break;
#endif
#ifdef WSAEPROVIDERFAILEDINIT
        case WSAEPROVIDERFAILEDINIT:
            strcpy(result,"The requested service provider could not be loaded or initialized");
            break;
#endif
#ifdef WSASYSCALLFAILURE
        case WSASYSCALLFAILURE:
            strcpy(result,"A system call that should never fail has failed");
            break;
#endif
#ifdef WSASERVICE_NOT_FOUND
        case WSASERVICE_NOT_FOUND:
            strcpy(result,"No such service is known. The service cannot be found in the specified name space");
            break;
#endif
#ifdef WSATYPE_NOT_FOUND
        case WSATYPE_NOT_FOUND:
            strcpy(result,"The specified class was not found");
            break;
#endif
#ifdef WSA_E_NO_MORE
        case WSA_E_NO_MORE:
            strcpy(result,"No more results can be returned by WSALookupServiceNext");
            break;
#endif
#ifdef WSA_E_CANCELLED
        case WSA_E_CANCELLED:
            strcpy(result,"A call to WSALookupServiceEnd was made while this call was still processing. The call has been canceled");
            break;
#endif
#ifdef WSAEREFUSED
        case WSAEREFUSED:
            strcpy(result,"A database query failed because it was actively refused");
            break;
#endif
#ifdef WSAHOST_NOT_FOUND
        case WSAHOST_NOT_FOUND:
            strcpy(result,"No such host is known");
            break;
#endif
#ifdef WSATRY_AGAIN
        case WSATRY_AGAIN:
            strcpy(result,"This is usually a temporary error during hostname resolution and means that the local server did not receive a response from an authoritative server");
            break;
#endif
#ifdef WSANO_RECOVERY
        case WSANO_RECOVERY:
            strcpy(result,"A non-recoverable error occurred during a database lookup");
            break;
#endif
#ifdef WSANO_DATA
        case WSANO_DATA:
            strcpy(result,"The requested name is valid, but no data of the requested type was found");
            break;
#endif
#ifdef WSA_QOS_RECEIVERS
        case WSA_QOS_RECEIVERS:
            strcpy(result,"At least one reserve has arrived");
            break;
#endif
#ifdef WSA_QOS_SENDERS
        case WSA_QOS_SENDERS:
            strcpy(result,"At least one path has arrived");
            break;
#endif
#ifdef WSA_QOS_NO_SENDERS
        case WSA_QOS_NO_SENDERS:
            strcpy(result,"There are no senders");
            break;
#endif
#ifdef WSA_QOS_NO_RECEIVERS
        case WSA_QOS_NO_RECEIVERS:
            strcpy(result,"There are no receivers");
            break;
#endif
#ifdef WSA_QOS_REQUEST_CONFIRMED
        case WSA_QOS_REQUEST_CONFIRMED:
            strcpy(result,"Reserve has been confirmed");
            break;
#endif
#ifdef WSA_QOS_ADMISSION_FAILURE
        case WSA_QOS_ADMISSION_FAILURE:
            strcpy(result,"Error due to lack of resources");
            break;
#endif
#ifdef WSA_QOS_POLICY_FAILURE
        case WSA_QOS_POLICY_FAILURE:
            strcpy(result,"Rejected for administrative reasons - bad credentials");
            break;
#endif
#ifdef WSA_QOS_BAD_STYLE
        case WSA_QOS_BAD_STYLE:
            strcpy(result,"Unknown or conflicting style");
            break;
#endif
#ifdef WSA_QOS_BAD_OBJECT
        case WSA_QOS_BAD_OBJECT:
            strcpy(result,"Problem with some part of the filterspec or providerspecific buffer in general");
            break;
#endif
#ifdef WSA_QOS_TRAFFIC_CTRL_ERROR
        case WSA_QOS_TRAFFIC_CTRL_ERROR:
            strcpy(result,"Problem with some part of the flowspec");
            break;
#endif
#ifdef WSA_QOS_GENERIC_ERROR
        case WSA_QOS_GENERIC_ERROR:
            strcpy(result,"General QOS error");
            break;
#endif
#ifdef WSA_QOS_ESERVICETYPE
        case WSA_QOS_ESERVICETYPE:
            strcpy(result,"An invalid or unrecognized service type was found in the flowspec");
            break;
#endif
#ifdef WSA_QOS_EFLOWSPEC
        case WSA_QOS_EFLOWSPEC:
            strcpy(result,"An invalid or inconsistent flowspec was found in the QOS structure");
            break;
#endif
#ifdef WSA_QOS_EPROVSPECBUF
        case WSA_QOS_EPROVSPECBUF:
            strcpy(result,"Invalid QOS provider-specific buffer");
            break;
#endif
#ifdef WSA_QOS_EFILTERSTYLE
        case WSA_QOS_EFILTERSTYLE:
            strcpy(result,"An invalid QOS filter style was used");
            break;
#endif
#ifdef WSA_QOS_EFILTERTYPE
        case WSA_QOS_EFILTERTYPE:
            strcpy(result,"An invalid QOS filter type was used");
            break;
#endif
#ifdef WSA_QOS_EFILTERCOUNT
        case WSA_QOS_EFILTERCOUNT:
            strcpy(result,"An incorrect number of QOS FILTERSPECs were specified in the FLOWDESCRIPTOR");
            break;
#endif
#ifdef WSA_QOS_EOBJLENGTH
        case WSA_QOS_EOBJLENGTH:
            strcpy(result,"An object with an invalid ObjectLength field was specified in the QOS provider-specific buffer");
            break;
#endif
#ifdef WSA_QOS_EFLOWCOUNT
        case WSA_QOS_EFLOWCOUNT:
            strcpy(result,"An incorrect number of flow descriptors was specified in the QOS structure");
            break;
#endif
#ifdef WSA_QOS_EUNKOWNPSOBJ
        case WSA_QOS_EUNKOWNPSOBJ:
            strcpy(result,"An unrecognized object was found in the QOS provider-specific buffer");
            break;
#endif
#ifdef WSA_QOS_EPOLICYOBJ
        case WSA_QOS_EPOLICYOBJ:
            strcpy(result,"An invalid policy object was found in the QOS provider-specific buffer");
            break;
#endif
#ifdef WSA_QOS_EFLOWDESC
        case WSA_QOS_EFLOWDESC:
            strcpy(result,"An invalid QOS flow descriptor was found in the flow descriptor list");
            break;
#endif
#ifdef WSA_QOS_EPSFLOWSPEC
        case WSA_QOS_EPSFLOWSPEC:
            strcpy(result,"An invalid or inconsistent flowspec was found in the QOS provider specific buffer");
            break;
#endif
#ifdef WSA_QOS_EPSFILTERSPEC
        case WSA_QOS_EPSFILTERSPEC:
            strcpy(result,"An invalid FILTERSPEC was found in the QOS provider-specific buffer");
            break;
#endif
#ifdef WSA_QOS_ESDMODEOBJ
        case WSA_QOS_ESDMODEOBJ:
            strcpy(result,"An invalid shape discard mode object was found in the QOS provider specific buffer");
            break;
#endif
#ifdef WSA_QOS_ESHAPERATEOBJ
        case WSA_QOS_ESHAPERATEOBJ:
            strcpy(result,"An invalid shaping rate object was found in the QOS provider-specific buffer");
            break;
#endif
#ifdef WSA_QOS_RESERVED_PETYPE
        case WSA_QOS_RESERVED_PETYPE:
            strcpy(result,"A reserved policy element was found in the QOS provider-specific buffer");
            break;
#endif
        default:
            sprintf(result,"(Unknown error code, code %d, that is %x)",code,code);
            break;
    }
#else
    switch(type) {
        case c_sock_log::TYPE_ERRNO:
            {
#ifndef HAVE_STRERROR_R
                strncpy(result,strerror(code),MAX_ERR_LENGTH);
                result[MAX_ERR_LENGTH]='\0'
#else
#   ifdef STRERROR_R_CHAR_P
                char buff[MAX_ERR_LENGTH+1];
                buff[0]='\0';
                strncpy(result,strerror_r(code,buff,MAX_ERR_LENGTH),MAX_ERR_LENGTH);
                result[MAX_ERR_LENGTH]='\0';
#   else
                strerror_r(code,result,MAX_ERR_LENGTH);
                result[MAX_ERR_LENGTH]='\0';
#   endif
                errno=0;
#endif
            }
            break;
        case c_sock_log::TYPE_GETHOSTBY:
            strncpy(result,hstrerror(code),MAX_ERR_LENGTH);
            result[MAX_ERR_LENGTH]='\0';
            break;
        default:
            sprintf(result,"(Unknown error code, code %d, that is %x)",code,code);
            break;
    }

#endif
    return(result);
}

/*!
    \brief Finds out if it is a temporary error
    \author VooDooMan
    \version 1
    \date 2006-05-19
    \param code Socket I/O error
    \param type Type of error code
    \retval true Temporary error
    \retval false Not temporary error
*/
bool sock_temporary_error(int code, c_sock_log::e_type UNUSEDARG_WIN(type))
{
#if defined(_WIN32) || defined(_WIN64)
    switch(code) {
        case WSAEACCES:
        case WSAEADDRINUSE:
        case WSAEADDRNOTAVAIL:
        case WSAEAFNOSUPPORT:
        case WSAEALREADY:
        case WSAECONNABORTED:
        case WSAECONNREFUSED:
        case WSAECONNRESET:
        case WSAEDESTADDRREQ:
        case WSAEFAULT:
        case WSAEHOSTDOWN:
        case WSAEHOSTUNREACH:
        case WSAEINPROGRESS:
        case WSAEINTR:
        case WSAEINVAL:
        case WSAEISCONN:
            return false;

        case WSAEMFILE:
            return true;

        case WSAEMSGSIZE:
            return false;

        case WSAENETDOWN:
            return true;

        case WSAENETRESET:
        case WSAENETUNREACH:
            return false;

        case WSAENOBUFS:
            return true;

        case WSAENOPROTOOPT:
        case WSAENOTCONN:
        case WSAENOTSOCK:
        case WSAEOPNOTSUPP:
        case WSAEPFNOSUPPORT:
            return false;

        case WSAEPROCLIM:
            return true;

        case WSAEPROTONOSUPPORT:
        case WSAEPROTOTYPE:
        case WSAESHUTDOWN:
        case WSAESOCKTNOSUPPORT:
        case WSAETIMEDOUT:
        case WSATYPE_NOT_FOUND:
            return false;

        case MY_EAGAIN_WOULD_BLOCK:
            return true;

        case WSAHOST_NOT_FOUND:
            return false;

        case WSANOTINITIALISED:
            return true;

        case WSANO_DATA:
        case WSANO_RECOVERY:
        case WSASYSCALLFAILURE:
        case WSASYSNOTREADY:
            return false;

        case WSATRY_AGAIN:
            return true;

        case WSAVERNOTSUPPORTED:
        case WSAEDISCON:
            return false;

        case MY_SELECT_ERROR:
            return true;
        case MY_NO_IPV6:
            return false;
        case MY_SSL_ERROR:
            return true;
        case MY_UNKNOWN_ERROR:
            return false;
        default:
            return true; // hm?
    }
#else
    if(MY_NO_IPV6==code)
        return false;
    else if(MY_UNKNOWN_ERROR==code)
        return false;

    switch(type) {
        case c_sock_log::TYPE_ERRNO:
            {
                if(EWOULDBLOCK==code) // to avoid confilict of multiple case's in case EWOULDBLOCK==EAGAIN==TRY_AGAIN
                    return true;

                if(TRY_AGAIN==code) // to avoid confilict of multiple case's in case EWOULDBLOCK==EAGAIN==TRY_AGAIN
                    return true;

                switch(code) {
                    case EAGAIN:
                    case EBUSY:
                    case EMFILE:
                    case EMLINK:
                    case ENETDOWN:
                    case ENFILE:
#if defined(ENODATA)
                    case ENODATA:
#endif
                    case ENOMEM:
                    case ENOMSG:
                    case ENOSPC:
                        return true;
                }
            }
            break;
        case c_sock_log::TYPE_GETHOSTBY:
            {
                if(NO_DATA==code) // to avoid confilict of multiple case's in case NO_ADDRESS==NO_DATA
                    return false;

                switch(code) {
                    case HOST_NOT_FOUND:
                    case NO_ADDRESS:
                    case NO_RECOVERY:
                        return false;
                    case TRY_AGAIN:
                        return true;
                    default:
                        return true;
                }
            }
            break;
        case c_sock_log::TYPE_GAI:
            return true;
        case c_sock_log::TYPE_MY:
            {
                switch(code) {
                    case MY_SELECT_ERROR:
                        return true;
                    case MY_NO_IPV6:
                        return false;
                    case MY_SSL_ERROR:
                        return true;
                    case MY_UNKNOWN_ERROR:
                        return false;
                    default:
                        return true;
                }
            }
            break;
        default:
            return true;
    }
    return false;
#endif
}

/*!
    \brief Initializes WinSock
    \author VooDooMan
    \version 1
    \date 2004
    \return OS's socket error
    \warning Only for Windows(TM)
*/
int sock_startup()
{
    m_sockets=MY_NEW c_mutex();

    errno=0;
    int la=0;
#if defined(_WIN32) || defined(_WIN64)
    int ec=WSAStartup(WINSOCK_VERSION,&WSAData);
    la=SOCK_ERRNO;
    if(ec==0)
        la=0;
    errno=0;
#endif

    protoent* pe=getprotobyname("tcp");
    if(pe)
        sock_tcp_proto_num=pe->p_proto;
    else {
        la=SOCK_ERRNO;
    }
    errno=0;

    return la;
}

/*!
    \brief Shuts down WinSock
    \author VooDooMan
    \version 1
    \date 2004
    \return OS's socket error
    \warning Only for Windows(TM)
*/
int sock_shutdown()
{
    if(NULL==m_sockets)
        return 0;

    sock_send_cache(true);

    errno=0;
#if defined(_WIN32) || defined(_WIN64)
    int ec=WSACleanup();
    int la=SOCK_ERRNO;
    if(ec==0)
        la=0;
    errno=0;

    MY_DELETE(m_sockets);
    m_sockets=NULL;

    return la;
#else

    MY_DELETE(m_sockets);
    m_sockets=NULL;

    return 0;
#endif

    // NOTREACH
}

/*!
    \brief Sets TCP keep-alive messages to be sent on socket
    \author VooDooMan
    \version 1
    \date 2006-05-19
    \param socket Socket
*/
void sock_set_tcp_keepalive(c_socket& socket)
{
    c_mutex_locker lock(m_sockets);

    int set=1;
    socklen_t size=sizeof(set);
    setsockopt(socket.deref(__FILE__,__LINE__),sock_tcp_proto_num,SO_KEEPALIVE,(char*)&set,size);
}

#if defined(_MSC_VER)
#   pragma warning(push)
#   pragma warning(disable:4267)    // warning C4267: 'argument' : conversion from 'size_t' to 'int', possible loss of data
#endif

#if defined(_MSC_VER)
#   pragma warning(push)
#   pragma warning(disable:4706)    // warning C4706: assignment within conditional expression
#endif

/*!
    \brief Connects to remote host
    \author VooDooMan
    \version 1
    \date 2004
    \param localIP Local IPv4 address as text to bind, if we are using IPv4 as destination, otherwise ignored
    \param remote Remote IPv4/IPv6 address as text, or IPv4/IPv6 host name
    \param remote_port Remote port
    \param error_code Returns OS's socket error
    \param async If true, it calls WSAAsyncSelect and creates cache buffers
    \param comment Comment for socket
    \return If error_code == zero, returns socket handle of connected socket, else undefined
*/
c_socket sock_connect(char* localIP, char* remote, unsigned short remote_port, int& error_code, bool async, const char* comment)
{
    errno=0;
    MY_SOCKET_FD s=MY_INVALID_SOCKET;
    c_socket Socket;
    Socket.clear();

#ifdef IPv6

    c_socket nul;
    nul.comment=comment;

    addrinfo Hints;
    addrinfo* AddrInfo=NULL;

    string rem=remote;

#if RESOLV_APPEND_DOT != 0
    if(rem.length() && rem[rem.length()-1]!='.') {
        unsigned tmp1, tmp2, tmp3, tmp4;
        if(rem.find(":",0)==string::npos && sscanf(rem.c_str(),"%u.%u.%u.%u",&tmp1,&tmp2,&tmp3,&tmp4)!=4)
            rem+=".";
    }
#endif

    int af;
    if(strstr(localIP,":"))
        af=AF_INET6;
    else
        af=AF_INET;

    memset(&Hints,0,sizeof(Hints));
    Hints.ai_family=af; //PF_UNSPEC;
    Hints.ai_socktype=SOCK_STREAM;
    Hints.ai_protocol=IPPROTO_TCP;
    char service[128];
    sprintf(service,"%u",static_cast<unsigned>(remote_port));
    errno=0;
    if((error_code=getaddrinfo(rem.c_str(),service,&Hints,&AddrInfo))) {
        string log="Cannot resolve address and port, while calling getaddrinfo() for [";
        log+=remote;
        log+=((string)"]:")+tos(remote_port);
        char* errstr=NULL;
        log_socket(error_code,errstr=sock_error(error_code,c_sock_log::TYPE_GAI),log,&nul);
        MY_DELETE_A(errstr);
        c_socket nul;
        errno=0;
        return nul;
    }
    errno=0;

    addrinfo* AI;
    for(AI=AddrInfo; AI; AI=AI->ai_next) {
        if(AI->ai_family!=af)
            continue;
        if(AI->ai_socktype!=SOCK_STREAM)
            continue;
        if(AI->ai_protocol!=IPPROTO_TCP)
            continue;

        errno=0;
        s=socket(AI->ai_family,AI->ai_socktype,AI->ai_protocol);
        if(s==MY_INVALID_SOCKET) {
            string log="Error opening socket, while calling socket() for [";
            log+=rem;
            log+=((string)"]:")+tos(remote_port);
            error_code=SOCK_ERRNO;
            char* errstr=NULL;
            if(error_code)
                log_socket(error_code,errstr=sock_error(error_code),log,&nul);
                MY_DELETE_A(errstr);
            errno=0;
            continue;
        }
        errno=0;

        if(strstr(localIP,":")!=NULL) {
            sockaddr_in6 LocalName;
            memset(&LocalName,0,sizeof(LocalName));
            LocalName.sin6_family=AF_INET6;
            LocalName.sin6_port=0;
            errno=0;
            my_inet_pton(AF_INET6,localIP,&LocalName.sin6_addr);
            errno=0;
#if SOCKET_BIND_WILDCARD==1
            memset(&LocalName.sin6_addr,0,sizeof(LocalName.sin6_addr));
            memset(&LocalName.sin6_port,0,sizeof(LocalName.sin6_port));
#endif
            int ErrorCode=bind(s,reinterpret_cast<sockaddr *>(&LocalName),sizeof(LocalName));
            if(ErrorCode!=0) {
                int la=error_code=SOCK_ERRNO;
                if(la!=0) {
                    c_socket nul;
                    nul.comment=comment;
                    errno=0;
                    freeaddrinfo(AddrInfo);
                    return nul;
                }
                errno=0;
            }
            errno=0;
        } else {
            sockaddr_in LocalName;
            memset(&LocalName,0,sizeof(LocalName));
            LocalName.sin_family=AF_INET;
            LocalName.sin_port=0;
            errno=0;
#if defined(_WIN32) || defined(_WIN64)
            my_inet_pton(AF_INET,localIP,&LocalName.sin_addr.S_un.S_addr);
#else
            my_inet_pton(AF_INET,localIP,&LocalName.sin_addr.s_addr);
#endif
            errno=0;

#if SOCKET_BIND_WILDCARD==1
            memset(&LocalName.sin_addr,0,sizeof(LocalName.sin_addr));
            memset(&LocalName.sin_port,0,sizeof(LocalName.sin_port));
#endif

            int ErrorCode=bind(s,reinterpret_cast<sockaddr *>(&LocalName),sizeof(LocalName));
            if(ErrorCode!=0) {
                int la=error_code=SOCK_ERRNO;
                if(la!=0) {
                    c_socket nul;
                    nul.comment=comment;
                    errno=0;
                    freeaddrinfo(AddrInfo);
                    return nul;
                }
                errno=0;
            }
            errno=0;
        }

        if(connect(s,AI->ai_addr,static_cast<int>(AI->ai_addrlen))==MY_SOCKET_ERROR) {
            if(Socket.cmp()) {
                sock_erase_later(Socket);
                sock_hndl_delete(Socket.hndl,__FILE__,__LINE__);
            }
            error_code=SOCK_ERRNO;
            my_closesocket(s);
            Socket.clear();
            string log="Error connecting socket, while calling connect() for [";
            log+=remote;
            log+=((string)"]:")+tos(remote_port);
            char* errstr=NULL;
            if(error_code)
                log_socket(error_code,errstr=sock_error(error_code),log,&nul);
            MY_DELETE_A(errstr);
            errno=0;
            continue;
        }

        errno=0;
        break;
    }

    errno=0;
    freeaddrinfo(AddrInfo);
    errno=0;

    if(!AI) {
        if(Socket.cmp()) {
            sock_erase_later(Socket);
            sock_hndl_delete(Socket.hndl,__FILE__,__LINE__);
        }
        errno=0;
        Socket.clear();
        c_socket nul;
        nul.comment=comment;
        errno=0;
        return nul;
    }

#else

    sockaddr_in SockAddrIn;
    SockAddrIn.sin_family=AF_INET;
    SockAddrIn.sin_addr=sock_resolve(remote);
    SockAddrIn.sin_port=htons(remote_port);

    uint32 a=0;
    memcpy(&a,&(SockAddrIn.sin_addr),sizeof(a));
    if(!a) {
        c_socket nul;
        nul.comment=comment;
        errno=0;
        error_code=MY_DNS_ERROR;
        return nul;
    }

    errno=0;

    s=socket(PF_INET,SOCK_STREAM,sock_tcp_proto_num);
    if(s==MY_SOCKET_ERROR) {
        int la=error_code=SOCK_ERRNO;
        if(la!=0) {
            c_socket nul;
            nul.comment=comment;
            errno=0;
            return nul;
        }
        errno=0;
    }
    errno=0;

    sockaddr_in LocalName;
    memset(&LocalName,0,sizeof(LocalName));
    LocalName.sin_family=AF_INET;
    LocalName.sin_port=0;
    errno=0;
#if defined(_WIN32) || defined(_WIN64)
    my_inet_pton(AF_INET,localIP,&LocalName.sin_addr.S_un.S_addr);
#else
    my_inet_pton(AF_INET,localIP,&LocalName.sin_addr.s_addr);
#endif
    errno=0;

#if SOCKET_BIND_WILDCARD==1
    memset(&LocalName.sin_addr,0,sizeof(LocalName.sin_addr));
    memset(&LocalName.sin_port,0,sizeof(LocalName.sin_port));
#endif

    int ErrorCode=bind(s,reinterpret_cast<sockaddr *>(&LocalName),sizeof(LocalName));
    if(ErrorCode!=0) {
        int la=error_code=SOCK_ERRNO;
        if(la!=0) {
            c_socket nul;
            nul.comment=comment;
            errno=0;
            return nul;
        }
        errno=0;
    }
    errno=0;

    ErrorCode=connect(s,reinterpret_cast<sockaddr *>(&SockAddrIn),sizeof(SockAddrIn));
    if(ErrorCode!=0) {
        int la=error_code=SOCK_ERRNO;
        if(la!=0) {
            if(Socket.cmp()) {
                sock_erase_later(Socket);
                sock_hndl_delete(Socket.hndl,__FILE__,__LINE__);
            }
            Socket.clear();
            my_closesocket(s);

            c_socket nul;
            nul.comment=comment;
            errno=0;
            return nul;
        }
        errno=0;
    }
    errno=0;

    if(!Socket.cmp())
        sock_hndl_create(s,Socket);
    errno=0;
    if(async)
        sock_async(Socket);
    errno=0;

#endif

    if(!Socket.cmp())
        sock_hndl_create(s,Socket);
    errno=0;

    sock_create_later(Socket,async);

    if(async)
        sock_async(Socket);
    errno=0;

    Socket.comment=comment;

    if(Socket.cmp()) {
        sock_set_tcp_keepalive(Socket);
        sock_register_socket(s);
    }

    Socket.closed=false;

    error_code=0;
    return Socket;
}

#if defined(_MSC_VER)
#   pragma warning(pop)
#endif

#if defined(_MSC_VER)
#   pragma warning(push)
#   pragma warning(disable:4267)    // warning C4267: 'argument' : conversion from 'size_t' to 'int', possible loss of data
#   pragma warning(disable:4706)    // warning C4706: assignment within conditional expression
#endif

/*!
    \brief Connects to remote host (for use in thread)
    \author VooDooMan
    \version 2
    \date 2004, 2005-12-07
    \param localIP Local IPv4 address as text to bind, if we are using IPv4 as destination, otherwise ignored
    \param remote Remote IPv4/IPv6 address as text, or IPv4/IPv6 host name
    \param remote_port Remote port
    \param error_code Returns OS's socket error
    \param comment Comment for socket
    \param log Returns data that should be logged
    \param errnum Returns error number (can be NULL)
    \param errtype Returns type of error code (can be NULL, but not when 'errnum' is not NULL)
    \return If error_code == zero, returns socket handle of connected socket, else undefined
*/
c_socket sock_thread_connect(const string& localIP, const string& remote, unsigned short remote_port, int& error_code, const string& comment, vector<c_sock_log>& log, int* errnum, c_sock_log::e_type* errtype)
{
    error_code=0;

    MY_SOCKET_FD s=0;

    string rem=remote;

#ifdef IPv6

    c_socket nul;
    nul.comment=comment;

    addrinfo Hints;
    addrinfo* AddrInfo=NULL;

    char port_name[64];
    sprintf(port_name,"%u",static_cast<unsigned>(remote_port));

#if RESOLV_APPEND_DOT != 0
    if(rem.length() && rem[rem.length()-1]!='.') {
        unsigned tmp1, tmp2, tmp3, tmp4;
        if(rem.find(":",0)==string::npos && sscanf(rem.c_str(),"%u.%u.%u.%u",&tmp1,&tmp2,&tmp3,&tmp4)!=4)
            rem+=".";
    }
#endif

    memset(&Hints,0,sizeof(Hints));
    Hints.ai_family=PF_UNSPEC;
    Hints.ai_socktype=SOCK_STREAM;
    int err=0;
    if((err=getaddrinfo(rem.c_str(),port_name,&Hints,&AddrInfo))) {
        error_code=err;
        string log_="Cannot resolve address and port, while calling getaddrinfo() for ";
        log_+=rem;
        log.push_back(c_sock_log(error_code,c_sock_log::TYPE_GAI,log_));
        c_socket nul;
        if(errnum) {
            *errnum=error_code;
            *errtype=c_sock_log::TYPE_GAI;
        }
        return nul;
    }

    addrinfo* AI;
    for(AI=AddrInfo; AI; AI=AI->ai_next) {
        s=socket(AI->ai_family,AI->ai_socktype,AI->ai_protocol);
        if(s==MY_INVALID_SOCKET) {
            string log_="Error opening socket, while calling socket() for ";
            log_+=rem;
            error_code=SOCK_ERRNO;
            log.push_back(c_sock_log(error_code,c_sock_log::TYPE_ERRNO,log_));
            if(errnum) {
                *errnum=error_code;
                *errtype=c_sock_log::TYPE_ERRNO;
            }
            continue;
        }

        if(localIP.find(':')!=string::npos) {
            sockaddr_in6 LocalName;
            memset(&LocalName,0,sizeof(LocalName));
            LocalName.sin6_family=AF_INET6;
            LocalName.sin6_port=0;
            my_inet_pton(AF_INET6,localIP.c_str(),&LocalName.sin6_addr);

#if SOCKET_BIND_WILDCARD==1
            memset(&LocalName.sin6_addr,0,sizeof(LocalName.sin6_addr));
            memset(&LocalName.sin6_port,0,sizeof(LocalName.sin6_port));
#endif

            int ErrorCode=bind(s,reinterpret_cast<sockaddr *>(&LocalName),sizeof(LocalName));
            if(ErrorCode!=0) {
                int la=error_code=SOCK_ERRNO;
                if(la!=0) {
                    c_socket nul;
                    nul.comment=comment;
                    if(errtype)
                        *errtype=c_sock_log::TYPE_ERRNO;
                    freeaddrinfo(AddrInfo);
                    return nul;
                }
            }
        } else {
            sockaddr_in LocalName;
            memset(&LocalName,0,sizeof(LocalName));
            LocalName.sin_family=AF_INET;
            LocalName.sin_port=0;
#if defined(_WIN32) || defined(_WIN64)
            my_inet_pton(AF_INET,localIP.c_str(),&LocalName.sin_addr.S_un.S_addr);
#else
            my_inet_pton(AF_INET,localIP.c_str(),&LocalName.sin_addr.s_addr);
#endif

#if SOCKET_BIND_WILDCARD==1
            memset(&LocalName.sin_addr,0,sizeof(LocalName.sin_addr));
            memset(&LocalName.sin_port,0,sizeof(LocalName.sin_port));
#endif

            int ErrorCode=bind(s,reinterpret_cast<sockaddr *>(&LocalName),sizeof(LocalName));
            if(ErrorCode!=0) {
                int la=error_code=SOCK_ERRNO;
                if(la!=0) {
                    c_socket nul;
                    nul.comment=comment;
                    if(errtype)
                        *errtype=c_sock_log::TYPE_ERRNO;
                    freeaddrinfo(AddrInfo);
                    return nul;
                }
            }
        }

        if(connect(s,AI->ai_addr,static_cast<int>(AI->ai_addrlen))!=0) {
            int la=error_code=SOCK_ERRNO;
            my_closesocket(s);
            string log_="Error connecting socket, while calling connect() for ";
            log_+=rem;
            if(la) {
                log_+=", ";
                char* e=NULL;
                log_+=e=sock_error(la,c_sock_log::TYPE_ERRNO);
                MY_DELETE_A(e);
                log.push_back(c_sock_log(la,c_sock_log::TYPE_ERRNO,log_));
            }
            continue;
        }
        break;
    }

    freeaddrinfo(AddrInfo);

    if(!AI) {
        c_socket nul;
        nul.comment=comment;
        return nul;
    }

#else

    vector<in_addr> a=sock_resolve_multi(remote.c_str(),log,errnum,errtype);
    if(a.size()==0) {
        c_socket nul;
        nul.comment=comment;
        return nul;
    }
    int r=rand()%static_cast<int>(a.size());

    sockaddr_in SockAddrIn;
    SockAddrIn.sin_family=AF_INET;

    memcpy(&(SockAddrIn.sin_addr),&(a.at(r)),sizeof(SockAddrIn.sin_addr));

    SockAddrIn.sin_port=htons(remote_port);

    s=socket(PF_INET,SOCK_STREAM,sock_tcp_proto_num);
    if(s==MY_SOCKET_ERROR) {
        int la=error_code=SOCK_ERRNO;
        if(la!=0) {
            c_socket nul;
            nul.comment=comment;
            if(errtype)
                *errtype=c_sock_log::TYPE_ERRNO;
            return nul;
        }
    }

    sockaddr_in LocalName;
    memset(&LocalName,0,sizeof(LocalName));
    LocalName.sin_family=AF_INET;
    LocalName.sin_port=0;
#if defined(_WIN32) || defined(_WIN64)
    my_inet_pton(AF_INET,localIP.c_str(),&LocalName.sin_addr.S_un.S_addr);
#else
    my_inet_pton(AF_INET,localIP.c_str(),&LocalName.sin_addr.s_addr);
#endif

#if SOCKET_BIND_WILDCARD==1
    memset(&LocalName.sin_addr,0,sizeof(LocalName.sin_addr));
    memset(&LocalName.sin_port,0,sizeof(LocalName.sin_port));
#endif

    int ErrorCode=bind(s,reinterpret_cast<sockaddr *>(&LocalName),sizeof(LocalName));
    if(ErrorCode!=0) {
        int la=error_code=SOCK_ERRNO;
        if(la!=0) {
            c_socket nul;
            nul.comment=comment;
            if(errtype)
                *errtype=c_sock_log::TYPE_ERRNO;
            return nul;
        }
    }

    ErrorCode=connect(s,reinterpret_cast<sockaddr *>(&SockAddrIn),sizeof(SockAddrIn));
    if(ErrorCode!=0) {
        int la=error_code=SOCK_ERRNO;
        if(la!=0) {
            my_closesocket(s);

            c_socket nul;
            nul.comment=comment;
            string log_="Error connecting socket, while calling connect() for ";
            log_+=rem;
            if(la) {
                log_+=", ";
                char* e=NULL;
                log_+=e=sock_error(la,c_sock_log::TYPE_ERRNO);
                MY_DELETE_A(e);
                log.push_back(c_sock_log(la,c_sock_log::TYPE_ERRNO,log_));
            }
            return nul;
        }
    }

#endif

    c_socket Socket;
    Socket.closed=false;
    Socket.comment=comment;
    if(!Socket.cmp())
        sock_hndl_create(s,Socket);

    error_code=0;
    return Socket;
}

#if defined(_MSC_VER)
#   pragma warning(pop)
#endif

/*!
    \brief Closes connected socket
    \author VooDooMan
    \version 2
    \date 2004, 2008-05-10
    \param socket Socket handle of connected socket
    \param fn File name of caller
    \param ln Line number of caller
*/
void sock_close(c_socket& socket, const char* fn, int ln)
{
    c_mutex_locker lock(m_sockets);

    vector<pair<c_socket,my_time_t> >::iterator li;
linger_again:
    for(li=sock_linger.begin(); li!=sock_linger.end(); ++li) {
        if((*li).first.was_it_you(socket)) {
            sock_erase_later((*li).first);
            errno=0;
            sock_linger.erase(li);
            goto linger_again;
        }
    }

    sock_unregister_socket(socket.deref(fn,ln));

    errno=0;
    sock_flush_later(socket);
    errno=0;
    sock_erase_later(socket);
    errno=0;
    if(socket.cmp()) {
        enum {
            MAX_TRIES = 100,
        };
        for(int i1=0; i1<MAX_TRIES; ++i1)
            if(0==my_closesocket(socket.deref(fn,ln)))
                break;
        errno=0;
    }

    sock_hndl_delete(socket.hndl,fn,ln);

    errno=0;
}

#if defined(_MSC_VER)
#   pragma warning(push)
#   pragma warning(disable:4706)    // warning C4706: assignment within conditional expression
#endif

/*!
    \brief Resolves DNS name or text IPv4 address to IPv4 address via DNS
    \author VooDooMan
    \version 1
    \date 2006-07-03
    \param domain DNS name or IPv4 address as text to resolve
    \param errnum Error number (could be NULL if don't care)
    \param errtype Returns type of error code (can be NULL, but not when 'errnum' is not NULL)
    \param log Returns data that should be logged
    \return Returns IPv4 address(es)
*/
vector<in_addr> sock_resolve_multi(const char* domain, vector<c_sock_log>& log, int* errnum, c_sock_log::e_type* errtype)
{
    int error_code=0;

    vector<in_addr> res;

    if(errnum) {
        *errnum=0;
        *errtype=c_sock_log::TYPE_ERRNO;
    }

    addrinfo Hints;
    addrinfo* AddrInfo=NULL;

    memset(&Hints,0,sizeof(Hints));
    Hints.ai_family=AF_INET;
    Hints.ai_socktype=SOCK_STREAM;

    int err=0;

    string rem=domain;

#if RESOLV_APPEND_DOT != 0
    if(rem.length() && rem[rem.length()-1]!='.') {
        unsigned tmp1, tmp2, tmp3, tmp4;
        if(rem.find(":",0)==string::npos && sscanf(rem.c_str(),"%u.%u.%u.%u",&tmp1,&tmp2,&tmp3,&tmp4)!=4)
            rem+=".";
    }
#endif

    if((err=getaddrinfo(rem.c_str(),NULL,&Hints,&AddrInfo))) {
        error_code=err;
        if(AddrInfo)
            freeaddrinfo(AddrInfo);

        string log_="Cannot resolve address and port, while calling getaddrinfo() [IPv4] for ";
        log_+=domain;
        log.push_back(c_sock_log(error_code,c_sock_log::TYPE_GAI,log_));
        if(errnum) {
            *errnum=error_code;
            *errtype=c_sock_log::TYPE_GAI;
        }
        return res;
    }

    addrinfo* AI;
    for(AI=AddrInfo; AI; AI=AI->ai_next) {
        if(AI->ai_family!=AF_INET)
            continue;
        in_addr a4=reinterpret_cast<sa_t*>(AI->ai_addr)->in.sin_addr;
        res.push_back(a4);
    }
    freeaddrinfo(AddrInfo);

    if(res.empty() && errnum) {
#if defined(_WIN32) || defined(_WIN64)
        *errnum=WSANO_DATA;
        *errtype=c_sock_log::TYPE_ERRNO;
#else
#   if defined(ENODATA)
        *errnum=ENODATA;
#   else
        *errnum=ENOENT;
#   endif
        *errtype=c_sock_log::TYPE_ERRNO;
#endif
    }

    return res;
}

#if defined(_MSC_VER)
#   pragma warning(pop)
#endif

/*!
    \brief Resolves IPv4 address to DNS name via DNS (reverse lookup)
    \author VooDooMan
    \version 1
    \date 2004
    \param address IPv4 address as text (dotted format)
    \param dns_name Buffer that receives DNS name (should be long enough)
    \param dns_max_size Length of dns_name buffer
    \param errnum Error number (could be NULL if don't care)
    \param errtype Returns type of error code (can be NULL, but not when 'errnum' is not NULL)
*/
void sock_reverse(const char* address, char* dns_name, int dns_max_size, int* errnum, c_sock_log::e_type* errtype)
{
    unsigned int addr=0;
    if(errnum) {
        *errnum=0;
        *errtype=c_sock_log::TYPE_ERRNO;
    }
    my_inet_pton(AF_INET,address,&addr);
#ifndef HAVE_GETHOSTBYADDR_R
    errno=0;
    hostent* remoteHost=gethostbyaddr((char *)&addr, 4, AF_INET);
    if(errnum) {
        *errnum=SOCK_ERRNO;
        *errtype=c_sock_log::TYPE_GETHOSTBY;
    }

    if(SOCK_ERRNO==0 && remoteHost) {
        errno=0;
        strncpy(dns_name,remoteHost->h_name,dns_max_size);
        dns_name[dns_max_size-1]=0;
    } else
        dns_name[0]=0;
    errno=0;
#else
    hostent remoteHost;
    size_t tmpl=1024*32;
    char* tmp=MY_NEW char[tmpl];
    hostent* tmp2=NULL;
    int errnop=0;
    if(gethostbyaddr_r((char*)&addr,4,AF_INET,&remoteHost,tmp,tmpl,&tmp2,&errnop)==0 && tmp2) {
        strncpy(dns_name,tmp2->h_name,dns_max_size);
        dns_name[dns_max_size-1]=0;
    } else
        dns_name[0]=0;

    if(errnum) {
        *errnum=errnop;
        *errtype=c_sock_log::TYPE_GETHOSTBY;
    }

    MY_DELETE_A(tmp);
#endif
}

/*!
    \brief Binds server socket and start listening
    \author VooDooMan
    \version 2
    \date 2004, 2005-08-01
    \param local Local IPv4 address to bind to
    \param local_port Local port to bind
    \param error_code Returns OS's socket error
    \param comment Comment for socket
    \return If error_code == zero, returns server socket handle, else undefined
*/
c_socket sock_bind(in_addr local, unsigned short local_port, int& error_code, const char* comment)
{
    errno=0;
    sockaddr_in SockAddrIn;

    SockAddrIn.sin_family=AF_INET;
    SockAddrIn.sin_addr=local;
    SockAddrIn.sin_port=htons(local_port);
    errno=0;

    MY_SOCKET_FD s;
    s=socket(PF_INET,SOCK_STREAM,sock_tcp_proto_num);
    if(MY_INVALID_SOCKET==s) {
        error_code=SOCK_ERRNO;
        c_socket nul;
        nul.comment=comment;
        errno=0;
        if(error_code!=0)
            return nul;
    }
    errno=0;
    int ErrorCode=0;
#if !defined(_WIN32) && !defined(_WIN64)
    // on Win32 this would be harmfull, only on POSIX it makes sense,
    // if socket is in TIME_WAIT state; on Win32 we would never get
    // "address already in use" error while socket is in TIME_WAIT,
    // but more dangerous on Win32/64 is that we will never get that error even
    // if address/port is really busy listening after setting socket option SO_REUSEADDR.
    int tmp=1;
    errno=0;
    ErrorCode=setsockopt(s,SOL_SOCKET,SO_REUSEADDR,(const char*)&tmp,sizeof(tmp));
    if(ErrorCode!=0) {
        error_code=SOCK_ERRNO;
        c_socket nul;
        nul.comment=comment;
        errno=0;
        if(error_code!=0)
            return nul;
    }
    errno=0;
#endif
    errno=0;

#if SOCKET_BIND_WILDCARD==1
    memset(&SockAddrIn.sin_addr,0,sizeof(SockAddrIn.sin_addr));
    // do NOT reset port to 0! tihs is server socket!
#endif

    ErrorCode=bind(s,reinterpret_cast<sockaddr*>(&SockAddrIn),sizeof(SockAddrIn));
    if(ErrorCode!=0) {
        error_code=SOCK_ERRNO;
        c_socket nul;
        nul.comment=comment;
        errno=0;
        if(error_code!=0)
            return nul;
    }
    errno=0;
    socklen_t i1=sizeof(SockAddrIn);
    ErrorCode=getsockname(s,reinterpret_cast<sockaddr*>(&SockAddrIn),&i1);
    if(ErrorCode!=0) {
        error_code=SOCK_ERRNO;
        c_socket nul;
        nul.comment=comment;
        errno=0;
        if(error_code!=0)
            return nul;
    }
    errno=0;
    ErrorCode=listen(s,MY_LISTEN_BACKLOG);
    if(ErrorCode!=0) {
        error_code=SOCK_ERRNO;
        c_socket nul;
        nul.comment=comment;
        errno=0;
        if(error_code!=0)
            return nul;
    }
    errno=0;
#if defined(_WIN32) || defined(_WIN64)
    errno=0;
    unsigned long non_blocking=1;
    ErrorCode=my_ioctlsocket(s,FIONBIO,&non_blocking);
#else
    errno=0;
    int flags=fcntl(s,F_GETFL,0);
    flags|=O_NONBLOCK;
    fcntl(s,F_SETFL,flags);
    ErrorCode=0;
#endif
    if(ErrorCode!=0) {
        error_code=SOCK_ERRNO;
        c_socket nul;
        nul.comment=comment;
        errno=0;
        if(error_code!=0)
            return nul;
    }
    errno=0;

    c_socket Socket;
    Socket.closed=false;
    if(!Socket.cmp())
        sock_hndl_create(s,Socket);
    Socket.comment=comment;
    return Socket;
}

#if defined(_MSC_VER)
#   pragma warning(push)
#   pragma warning(disable:4267)    // warning C4267: 'argument' : conversion from 'size_t' to 'int', possible loss of data
#   pragma warning(disable:4706)    // warning C4706: assignment within conditional expression
#endif

/*!
    \brief Binds server socket and start listening
    \author VooDooMan
    \version 1
    \date 2004
    \param local Local IPv6 address to bind to
    \param port Local port to bind
    \param error_code Returns OS's socket error
    \param comment Comment for socket
    \return If error_code == zero, returns server socket handle, else undefined
*/
c_socket sock_bind6(char* UNUSEDARG_NO_IPv6(local), unsigned short UNUSEDARG_NO_IPv6(port), int& error_code, const char* comment)
{
    errno=0;
#ifdef IPv6
    c_socket nul;
    nul.comment=comment;

    addrinfo Hints;
    addrinfo* AddrInfo=NULL;

    char port_name[64];
    sprintf(port_name,"%u",static_cast<unsigned>(port));

    memset(&Hints,0,sizeof(Hints));
    Hints.ai_family=PF_INET6;
    Hints.ai_socktype=SOCK_STREAM;
    Hints.ai_flags=AI_NUMERICHOST /*AI_PASSIVE*/;

    string rem=local;

#if RESOLV_APPEND_DOT != 0
    if(rem.length() && rem[rem.length()-1]!='.') {
        unsigned tmp1, tmp2, tmp3, tmp4;
        if(rem.find(":",0)==string::npos && sscanf(rem.c_str(),"%u.%u.%u.%u",&tmp1,&tmp2,&tmp3,&tmp4)!=4)
            rem+=".";
    }
#endif

    if((error_code=getaddrinfo(rem.c_str(),port_name,&Hints,&AddrInfo))) {
        string log="Cannot resolve address and port, while calling getaddrinfo() for [";
        log+=rem;
        log+=((string)"]:")+tos(port);
        error_code=SOCK_ERRNO;
        char* errstr=NULL;
        log_socket(error_code,errstr=sock_error(error_code,c_sock_log::TYPE_GAI),log.c_str(),&nul);
        MY_DELETE_A(errstr);
        c_socket nul;
        nul.comment=comment;
        errno=0;
        return nul;
    }
    errno=0;

    addrinfo* AI;
    for(AI=AddrInfo; AI; AI=AI->ai_next) {
        errno=0;
        if(AI->ai_family!=AF_INET6)
            continue;

        MY_SOCKET_FD s;

        s=socket(AI->ai_family,AI->ai_socktype,AI->ai_protocol);
        if(MY_INVALID_SOCKET==s)
            continue;

        errno=0;

        addrinfo AI2;
        memcpy(&AI2,AI,sizeof(AI2));

#if SOCKET_BIND_WILDCARD==1
        memset(&(AI2.ai_addr),0,sizeof(AI2.ai_addr));
        // do NOT reset port to 0! tihs is server socket!
#endif

        int ErrorCode=bind(s,AI2.ai_addr,static_cast<int>(AI2.ai_addrlen));
        if(ErrorCode!=0) {
            errno=0;
            my_closesocket(s);
            errno=0;
            continue;
        }
        errno=0;

        ErrorCode=listen(s,MY_LISTEN_BACKLOG);
        if(ErrorCode!=0) {
            errno=0;
            my_closesocket(s);
            errno=0;
            continue;
        }
        errno=0;

#if defined(_WIN32) || defined(_WIN64)
        unsigned long non_blocking=1;
        errno=0;
        ErrorCode=my_ioctlsocket(s,FIONBIO,&non_blocking);
        errno=0;
#else
        errno=0;
        int flags=fcntl(s,F_GETFL,0);
        errno=0;
        flags|=O_NONBLOCK;
        fcntl(s,F_SETFL,flags);
        ErrorCode=0;
        errno=0;
#endif
        if(ErrorCode!=0) {
            errno=0;
            my_closesocket(s);
            errno=0;
            continue;
        }
        errno=0;

        freeaddrinfo(AddrInfo);
        errno=0;

        c_socket Socket;
        Socket.closed=false;
        if(!Socket.cmp())
            sock_hndl_create(s,Socket);
        Socket.comment=comment;
        error_code=0;
        return Socket;
    }

    errno=0;
    freeaddrinfo(AddrInfo);
    errno=0;

    error_code=SOCK_ERRNO;
    nul.comment=comment;
    errno=0;
    return nul;

#else
    error_code=MY_NO_IPV6;
    c_socket nul;
    nul.comment=comment;
    errno=0;
    return nul;
#endif
}

#if defined(_MSC_VER)
#   pragma warning(pop)
#endif

/*!
    \brief Accepts client connection on listening server socket
    \author VooDooMan
    \version 1
    \date 2004
    \param socket Socket handle of listening server socket
    \param comment Comment for socket
    \param async true for asynchronous mode
    \return Returns socket handle of client connection
*/
c_socket sock_accept(c_socket& socket, const char* comment, bool async)
{
    errno=0;
    fd_set s_;
    FD_ZERO(&s_);
#if defined(_MSC_VER)
#   pragma warning(push)
#   pragma warning(disable:4127)    // warning C4127: conditional expression is constant
#endif
    FD_SET(socket.deref(__FILE__,__LINE__),&s_);
#if defined(_MSC_VER)
#   pragma warning(pop)
#endif
    timeval timeout;
    timeout.tv_sec=0;
    timeout.tv_usec=100;
    int num=select((int)(socket.deref(__FILE__,__LINE__))+1,&s_,NULL,NULL,&timeout);
    errno=0;

    c_socket nul;
    nul.comment=comment;

    if(num==0)
        return nul;
    sockaddr_in SockAddrIn;
    socklen_t i1=sizeof(SockAddrIn);
    errno=0;
    MY_SOCKET_FD ec=accept(socket.deref(__FILE__,__LINE__),reinterpret_cast<sockaddr *>(&SockAddrIn),&i1);

    if(ec!=MY_INVALID_SOCKET) {
        c_socket s;
        s.closed=false;
        if(!s.cmp())
            sock_hndl_create(ec,s);

        if(async) {
            sock_create_later(s,true);
            sock_async(s);
        }

        sock_set_tcp_keepalive(s);

        sock_register_socket(s.deref(__FILE__,__LINE__));

        s.comment=comment;
        errno=0;
        return s;
    }
    errno=0;

    return nul;
}

#if defined(_MSC_VER)
#   pragma warning(push)
#   pragma warning(disable:4127)    // warning C4127: conditional expression is constant
//#   pragma warning(disable:4244)    // warning C4244: 'argument' : conversion from 'MY_SOCKET_FD' to 'int', possible loss of data
#endif

/*!
    \brief Accepts client connection on listening server socket
    \author VooDooMan
    \version 1
    \date 2004
    \param socket Socket handle of listening server socket
    \param comment Comment for socket
    \param async true for asynchronous mode
    \return Returns socket handle of client connection
*/
c_socket sock_accept6(c_socket& UNUSEDARG_NO_IPv6(socket), const char* comment, bool UNUSEDARG_NO_IPv6(async))
{
    errno=0;
    c_socket nul;
    nul.comment=comment;
#ifdef IPv6
    fd_set s_;
    FD_ZERO(&s_);
    FD_SET(socket.deref(__FILE__,__LINE__),&s_);
    timeval timeout;
    timeout.tv_sec=0;
    timeout.tv_usec=100;
    int num=select(static_cast<int>((socket.deref(__FILE__,__LINE__)))+1,&s_,NULL,NULL,&timeout);
    errno=0;

    if(num==0)
        return nul;
    sockaddr_in6 SockAddrIn;
    socklen_t i1=sizeof(SockAddrIn);
    MY_SOCKET_FD ec=accept(socket.deref(__FILE__,__LINE__),reinterpret_cast<sockaddr *>(&SockAddrIn),&i1);
    errno=0;

    if(ec!=MY_INVALID_SOCKET) {
        c_socket s;
        s.closed=false;
        if(!s.cmp())
            sock_hndl_create(ec,s);

        errno=0;
        if(getnameinfo(reinterpret_cast<sockaddr *>(&SockAddrIn),sizeof(SockAddrIn),s.host,sizeof(s.host),NULL,0,NI_NUMERICHOST)!=0)
            s.host[0]=0;
        errno=0;

        if(async) {
            sock_create_later(s,true);
            sock_async(s);
        }

        sock_set_tcp_keepalive(s);

        sock_register_socket(s.deref(__FILE__,__LINE__));

        s.comment=comment;
        errno=0;
        return s;
    }

#endif
    errno=0;
    return nul;
}

#if defined(_MSC_VER)
#   pragma warning(pop)
#endif

/*!
    \brief Gets remote IPv4 address of remote host
    \author VooDooMan
    \version 1
    \date 2004
    \param socket Socket handle of connected socket
    \param error_code Returns OS's socket error
    \return If error_code == zero, returns IPv4 address of remote host, else undefined
*/
in_addr sock_get_remote_addr(c_socket& socket, int& error_code)
{
    c_mutex_locker lock(m_sockets);

    errno=0;
    sockaddr_in addr;
    socklen_t i2=sizeof(addr);
    error_code=getpeername(socket.deref(__FILE__,__LINE__),reinterpret_cast<sockaddr*>(&addr),&i2);
    errno=0;
    return addr.sin_addr;
}

/*!
    \brief Gets remote IPv6 address of remote host
    \author VooDooMan
    \version 1
    \date 2004
    \param socket Socket handle of connected socket
    \return Returns IPv6 address of remote host, or zero
*/
in_addr6_ sock_get_remote_addr6(c_socket& UNUSEDARG_NO_IPv6(socket))
{
    c_mutex_locker lock(m_sockets);

    errno=0;
#ifdef IPv6
    char tmp[2048];
    strcpy(tmp,socket.host);
    if(strstr(tmp,"%"))
        *strstr(tmp,"%")=0;
    in_addr6_ addr;
    errno=0;
    inetpton(AF_INET6,(const char*)tmp,&addr);
    errno=0;
    return addr;
#else
    in_addr6_ nul;
    memset(&nul,0,sizeof(nul));
    errno=0;
    return nul;
#endif
}

/*!
    \brief Low-level send to socket (Windows(TM) only)
    \author VooDooMan
    \version 1
    \date 2004
    \param socket Socket handle
    \param buff Buffer to send
    \param len Length of the buffer
    \param bsd_errno Receives low-level errno after BSD sockets API call
    \return Returns number of bytes actually sent
*/
size_t sock_lo_send(c_socket& socket, char* buff, size_t len, int& bsd_errno)
{
    c_mutex_locker lock(m_sockets);

    bsd_errno=EBADF;

    if(!socket.cmp())
        return static_cast<size_t>(MY_SOCKET_ERROR);

    errno=0;
    size_t sent=send(socket.deref(__FILE__,__LINE__),buff,(int)len,0);
    bsd_errno=errno;
    errno=0;
    if((signed)sent>0) {
        socket.last_io=my_time(NULL);
        socket.sent+=sent;
    }
    return sent;
}

/*!
    \brief Low-level: receives data from the socket input queue, but leaves it in it (can be used for check how mush data is waiting)
    \author VooDooMan
    \version 2
    \date 2005
    \param socket Socket handle
    \param buff Buffer to receive
    \param len Length of the buffer
    \param bsd_errno Receives low-level errno after BSD sockets API call
    \return Returns number of bytes waiting in queue
*/
size_t sock_lo_get_input_amount(c_socket& socket, char* buff, size_t len, int& bsd_errno)
{
    c_mutex_locker lock(m_sockets);

    bsd_errno=EBADF;

    if(!socket.cmp())
        return static_cast<size_t>(MY_SOCKET_ERROR);

    errno=0;
    fd_set s_;
    FD_ZERO(&s_);
#if defined(_MSC_VER)
#   pragma warning(push)
#   pragma warning(disable:4127)    // warning C4127: conditional expression is constant
#endif
    FD_SET(socket.deref(__FILE__,__LINE__),&s_);
#if defined(_MSC_VER)
#   pragma warning(pop)
#endif
    timeval timeout;
    timeout.tv_sec=0;
    timeout.tv_usec=100;
    int num=select((int)(socket.deref(__FILE__,__LINE__))+1,&s_,NULL,NULL,&timeout);
    bsd_errno=errno;
    if(SOCK_ERRNO)
        num=0;
    errno=0;
    ssize_t r=0;
    if(num) {
        r=recvfrom(socket.deref(__FILE__,__LINE__),buff,(int)len,MSG_PEEK,NULL,NULL);
        bsd_errno=errno;
    }
    if(r<0)
        r=0;
    errno=0;
    return static_cast<size_t>(r);
}

/*!
    \brief Low-level read from socket (Windows(TM) only)
    \author VooDooMan
    \version 2
    \date 2005
    \param socket Socket handle
    \param buff Buffer to store received data
    \param len Maximum length of the buffer
    \param bsd_errno Receives low-level errno after BSD sockets API call
    \return Returns number of bytes actually received
*/
size_t sock_lo_recv(c_socket& socket, char* buff, size_t len, int& bsd_errno)
{
    c_mutex_locker lock(m_sockets);

    bsd_errno=EBADF;

    if(!socket.cmp())
        return static_cast<size_t>(MY_SOCKET_ERROR);

    errno=0;
    fd_set s_;
    FD_ZERO(&s_);
#if defined(_MSC_VER)
#   pragma warning(push)
#   pragma warning(disable:4127)    // warning C4127: conditional expression is constant
#endif
    FD_SET(socket.deref(__FILE__,__LINE__),&s_);
#if defined(_MSC_VER)
#   pragma warning(pop)
#endif
    timeval timeout;
    timeout.tv_sec=0;
    timeout.tv_usec=100;
    int num=select((int)(socket.deref(__FILE__,__LINE__))+1,&s_,NULL,NULL,&timeout);
    bsd_errno=errno;
    if(SOCK_ERRNO)
        num=0;
    errno=0;
    ssize_t r=0;
    if(num) {
        r=recvfrom(socket.deref(__FILE__,__LINE__),buff,(int)len,0,NULL,NULL);
        bsd_errno=errno;
    }
    if(r<0)
        r=0;
    errno=0;
    if(r>0) {
        socket.last_io=my_time(NULL);
        socket.received+=r;
    }
    return static_cast<size_t>(r);
}

#if defined(_MSC_VER)
#   pragma warning(push)
#   pragma warning(disable:4706)    // warning C4706: assignment within conditional expression
#endif

/*!
    \brief Resolves DNS name or text IPv6 address to IPv6 address via DNS
    \author VooDooMan
    \version 1
    \date 2006-07-03
    \param domain DNS name or IPv6 address as text to resolve
    \param log Returns messages that should be logged
    \param errnum Returns error number (can be NULL)
    \param errtype Returns type of error code (can be NULL, but not when 'errnum' is not NULL)
    \return Returns IPv6 address(es)
*/
vector<in_addr6_> sock_resolve6_multi(const char* UNUSEDARG_NO_IPv6(domain), vector<c_sock_log>& UNUSEDARG_NO_IPv6(log), int* errnum, c_sock_log::e_type* errtype)
{
    vector<in_addr6_> res;

    if(errnum) {
        *errnum=0;
        *errtype=c_sock_log::TYPE_ERRNO;
    }
#ifdef IPv6
    int error_code=0;

    addrinfo Hints;
    addrinfo* AddrInfo=NULL;

    char port_name[64];
    sprintf(port_name,"%u",(unsigned)80); // something must be set as port...

    memset(&Hints,0,sizeof(Hints));
    Hints.ai_family=PF_INET;
    Hints.ai_socktype=SOCK_STREAM;
    Hints.ai_flags=AI_PASSIVE;

    int err=0;

    string rem=domain;

#if RESOLV_APPEND_DOT != 0
    if(rem.length() && rem[rem.length()-1]!='.') {
        unsigned tmp1, tmp2, tmp3, tmp4;
        if(rem.find(":",0)==string::npos && sscanf(rem.c_str(),"%u.%u.%u.%u",&tmp1,&tmp2,&tmp3,&tmp4)!=4)
            rem+=".";
    }
#endif

    if((err=getaddrinfo(rem.c_str(),port_name,&Hints,&AddrInfo))) {
        string log_="Cannot resolve address and port, while calling getaddrinfo() [IPv6] for ";
        log_+=domain;
        error_code=err;
        log.push_back(c_sock_log(error_code,c_sock_log::TYPE_GAI,log_));
        if(errnum) {
            *errnum=error_code;
            *errtype=c_sock_log::TYPE_GAI;
        }
        return res;
    }

    addrinfo* AI;
    for(AI=AddrInfo; AI; AI=AI->ai_next) {
        if(AI->ai_family!=AF_INET6)
            continue;
        in_addr6_ a6;
        memcpy(&a6,&(reinterpret_cast<sa_t*>(AI->ai_addr)->in6.sin6_addr),sizeof(a6));
        res.push_back(a6);
    }
    freeaddrinfo(AddrInfo);
#endif

    if(res.empty() && errnum) {
#if defined(_WIN32) || defined(_WIN64)
        *errnum=WSANO_DATA;
        *errtype=c_sock_log::TYPE_ERRNO;
#else
#   if defined(ENODATA)
        *errnum=ENODATA;
#   else
        *errnum=ENOENT;
#   endif
        *errtype=c_sock_log::TYPE_ERRNO;
#endif
    }

    return res;
}

#if defined(_MSC_VER)
#   pragma warning(pop)
#endif

/*!
    \brief Resolves IPv6 address to DNS name via DNS (reverse lookup)
    \author VooDooMan
    \version 1
    \date 2004
    \param address IPv6 address
    \param dns_name Buffer that receives DNS name (should be long enough)
    \param dns_max_size Length of dns_name buffer
    \param log Returns messages that should be logged
    \param errnum Error number (could be NULL if don't care)
    \param errtype Returns type of error code (can be NULL, but not when 'errnum' is not NULL)
    \bug Doesn't work :-(
*/
void sock_reverse(in_addr6_ UNUSEDARG_NO_IPv6(address), char* dns_name, int UNUSEDARG_NO_IPv6(dns_max_size), vector<c_sock_log>& UNUSEDARG_NO_IPv6(log), int* errnum, c_sock_log::e_type* errtype)
{
    if(errnum) {
        *errnum=0;
        *errtype=c_sock_log::TYPE_ERRNO;
    }
#ifdef IPv6
    int error_code=0;

    char host[2048];
    char servInfo[256];

    sockaddr_in6 a;
    memset(&a,0,sizeof(a));
    memcpy(&a.sin6_addr,&address,sizeof(a.sin6_addr));
    a.sin6_family=AF_INET6;
    a.sin6_port=80; // something for port...
    if(getnameinfo((sockaddr*)&a, sizeof(a), host, sizeof(host), servInfo, sizeof(servInfo), NI_NAMEREQD)!=0) {
        string log_=GET_STR_DEBUG_INFO_EX("Error while calling getnameinfo()");
        error_code=SOCK_ERRNO;
        if(error_code)
            log.push_back(c_sock_log(error_code,c_sock_log::TYPE_GAI,log_));
        if(errnum) {
            *errnum=error_code;
            *errtype=c_sock_log::TYPE_GAI;
        }
        host[0]=0;
    }
    host[sizeof(host)-1]=0;
    memset(dns_name,0,dns_max_size);
    strncpy(dns_name,host,dns_max_size);
    dns_name[dns_max_size-1]=0;
#else
    dns_name[0]=0;
#endif
}

/*!
    \brief Sets or unsets the non-blocking mode for the socket
    \author VooDooMan
    \version 1
    \date 2004
    \param socket Socket handle
    \param non_blocking true if to set non-blocking mode
*/
void sock_set_nonblocking(c_socket& socket, bool non_blocking)
{
    c_mutex_locker lock(m_sockets);

    if(!socket.cmp())
        return;

#if defined(_WIN32) || defined(_WIN64)
    unsigned long non_blocking_=non_blocking?1:0;
    errno=0;
    my_ioctlsocket(socket.deref(__FILE__,__LINE__),FIONBIO,&non_blocking_);
    errno=0;
#else
    errno=0;
    int flags=fcntl(socket.deref(__FILE__,__LINE__),F_GETFL,0);
    errno=0;
    flags&=~O_NONBLOCK;
    flags|=non_blocking?O_NONBLOCK:0;
    fcntl(socket.deref(__FILE__,__LINE__),F_SETFL,flags);
    errno=0;
#endif
}

/*!
    \brief Sets comment for the socket
    \author VooDooMan
    \version 1
    \date 2005-08-01
    \param socket Socket handle
    \param comment Comment for socket
*/
void sock_comment(c_socket& socket, const string& comment)
{
    socket.comment=comment;
}

/*!
    \brief Gets remote port number
    \author VooDooMan
    \version 1
    \date 2006-01-11
    \param socket Socket handle of connected socket
*/
unsigned short sock_get_remote_port(c_socket& socket)
{
    c_mutex_locker lock(m_sockets);

    sockaddr_in addr;
    socklen_t i2=sizeof(addr);
    errno=0;
    int error_code=getpeername(socket.deref(__FILE__,__LINE__),reinterpret_cast<sockaddr*>(&addr),&i2);
    errno=0;
    if(error_code)
        return 0;
    unsigned short r=ntohs(addr.sin_port);
    errno=0;
    return r;
}

/*!
    \brief Gets local port number
    \author VooDooMan
    \version 1
    \date 2006-01-11
    \param socket Socket handle of connected socket
*/
unsigned short sock_get_local_port(c_socket& socket)
{
    c_mutex_locker lock(m_sockets);

    sockaddr_in addr;
    socklen_t i2=sizeof(addr);
    errno=0;
    int error_code=getsockname(socket.deref(__FILE__,__LINE__),reinterpret_cast<sockaddr*>(&addr),&i2);
    errno=0;
    if(error_code)
        return 0;
    unsigned short r=ntohs(addr.sin_port);
    errno=0;
    return r;
}

void sock_register_socket(MY_SOCKET_FD s)
{
    c_mutex_locker lock(m_sockets);

    all_sockets.push_back(s);
}

void sock_unregister_socket(MY_SOCKET_FD s)
{
    c_mutex_locker lock(m_sockets);

    for(vector<MY_SOCKET_FD>::iterator i1=all_sockets.begin(); i1!=all_sockets.end(); ++i1) {
        if((*i1)==s) {
            all_sockets.erase(i1);
            break;
        }
    }
}

/*!
    \brief Checks if socket I/O queue is empty or not
    \author VooDooMan
    \version 1
    \date 2006-05-08
    \return Returns 'true' is it is empty, otherwise 'false'
*/
MY_STATIC bool sock_is_queue_empty()
{
    c_mutex_locker lock(m_sockets);

    {
        for(vector<c_receive_later>::iterator i1=receive_later.begin(); i1!=receive_later.end(); ++i1)
            if(!(*i1).data.empty() && !(*i1).socket.closed && !(*i1).has_error && !(*i1).using_window)
                return false;
    }
    {
        for(vector<c_send_later>::iterator i1=send_later.begin(); i1!=send_later.end(); ++i1)
            if(!(*i1).data.empty()
#if defined(_WIN32) || defined(_WIN64)
                && (*i1).socket.wnd
#endif
                && !(*i1).socket.closed
                )
                return false;
    }

    return true;
}

#ifndef XTELNET

void sock_interruptible_sleep(unsigned int ms)
{
    if(ms<50) {
    common_sleep:
#if !defined(_WIN32) && !defined(_WIN64) && !defined(__CYGWIN__)
        timeval t;
        t.tv_sec=0;
        t.tv_usec=ms*1000;
        select(1,NULL,NULL,NULL,&t);
#else
        SleepEx((DWORD)ms,true);
#endif
        return;
    }

    if(sock_is_queue_empty()) {
        goto common_sleep;
    }

    int one_more=1;
    fd_set s;
    FD_ZERO(&s);

    /*
    static unsigned last=0;
    if(last>all_sockets.size())
        last=0;
    */
    unsigned last=0;

    unsigned i2=0, i3=0, cnt=0;
    unsigned max=FD_SETSIZE;
    if(max>FD_SETSIZE_MAX)
        max=FD_SETSIZE_MAX;

    {
        c_mutex_locker lock(m_sockets);

        vector<MY_SOCKET_FD>::iterator i1=all_sockets.begin();

        while(i3<all_sockets.size()) {
            if(all_sockets.end()==i1) {
                i1=all_sockets.begin();
                i2=0;
                continue;
            }

            if(i3>=last) {
                last=i3;
                if(!FD_ISSET(*i1,&s)) {
#if defined(_MSC_VER)
#   pragma warning(push)
#   pragma warning(disable:4127)    // warning C4127: conditional expression is constant
#endif
                    FD_SET(*i1,&s);
#if defined(_MSC_VER)
#   pragma warning(pop)
#endif
                    cnt++;
                    if(one_more<=(int)*i1)
                        one_more=(int)*i1+1;
                }
                if(cnt==max)
                    break;
            }

            i1++;
            i2++;
            i3++;
        }
    }

    timeval t;
    t.tv_sec=ms/1000;
    t.tv_usec=(ms%1000)*1000;

    select(one_more,&s,NULL,NULL,&t);
    // <- ...
    // 1. we want NULL for write request because we want to sleep!
    // 2. we want NULL for exceptions, to avoid bug in case socket is already invalid
    //    and forgotten to be removed from array
}

/*!
    \brief Report FD_SETSIZE size to logs
    \version 1
    \author VooDooMan
    \date 2006-05-31
    \param anyway Argument 'anyway' to pass to logging function
*/
void sock_log_fd_array_size(bool anyway)
{
    unsigned max=FD_SETSIZE;
    if(max>FD_SETSIZE_MAX)
        max=FD_SETSIZE_MAX;

    c_socket nul;
    string log="Using ";
    log+=tos(max);
    log+=" rotating sockets as maximum for idle sleeper (FD_SETSIZE==";
    log+=tos(FD_SETSIZE);
    log+=")";
    log_socket(0,"",log.c_str(),&nul,anyway);
}

#endif

/*!
    \brief Resolves DNS name or text IPv4 address to IPv4 address via DNS
    \author VooDooMan
    \version 2
    \date 2004, 2006-07-03
    \param domain DNS name or IPv4 address as text to resolve
    \param errnum Error number (could be NULL if don't care)
    \param errtype Returns type of error code (can be NULL, but not when 'errnum' is not NULL)
    \return Returns IPv4 address (if there are more than one, returns randomly selected one)
*/
in_addr sock_resolve(const char* domain, int* errnum, c_sock_log::e_type* errtype)
{
    in_addr addr;

    vector<c_sock_log> log;
    vector<in_addr> a=sock_resolve_multi(domain,log,errnum,errtype);

    vector<c_sock_log>::iterator i1;
    c_socket nul;
    for(i1=log.begin(); i1!=log.end(); ++i1) {
#ifndef XTELNET
        char* errstr=NULL;
        string log_msg=(*i1).log;
#ifndef _MSC_VER
        log_msg+=((string)" [repeated ")+tos(log.size())+" times]";
#else
        // to get rid of warning on MSVC 2008 (9.0) about possible data lost by converting 'size_t' to 'unsigned int' (bug in STL << operator implementation)
        log_msg+=((string)" [repeated ")+tos(static_cast<unsigned long long>(log.size()))+" times]";
#endif
        log_socket((*i1).ec,errstr=sock_error((*i1).ec,(*i1).type),log_msg,&nul);
        MY_DELETE_A(errstr);
        break;
#endif
    }

    if(a.size()==0) {
        memset(&addr,0,sizeof(addr));
        return addr;
    }
    int r=rand()%static_cast<int>(a.size());
    memcpy(&addr,&(a.at(r)),sizeof(addr));
    return addr;
}

/*!
    \brief Resolves DNS name or text IPv6 address to IPv6 address via DNS
    \author VooDooMan
    \version 2
    \date 2004, 2006-07-03
    \param domain DNS name or IPv6 address as text to resolve
    \param log Returns messages that should be logged
    \param errnum Returns error number (can be NULL)
    \param errtype Returns type of error code (can be NULL, but not when 'errnum' is not NULL)
    \return Returns IPv6 address (if there are more than one, returns randomly selected one)
*/
in_addr6_ sock_resolve6(const char* domain, vector<c_sock_log>& log, int* errnum, c_sock_log::e_type* errtype)
{
    in_addr6_ addr;

    vector<in_addr6_> a=sock_resolve6_multi(domain,log,errnum,errtype);
    if(a.size()==0) {
        memset(&addr,0,sizeof(addr));
        return addr;
    }
    int r=rand()%static_cast<int>(a.size());
    memcpy(&addr,&(a.at(r)),sizeof(addr));
    return addr;
}

/*!
    \brief Check wether connection was closed
    \author VooDooMan
    \version 1
    \date 2006-07-10
    \param socket Socket handle of connected socket
    \return Return true for error, or connection was closed
*/
bool sock_closed(c_socket& socket)
{
    return socket.closed;
}

/*!
    \brief Pushes socket to linger emulation
    \author VooDooMan
    \version 1
    \date 2007-03-16
    \param s Socket
*/
void sock_linger_push(c_socket& s)
{
    if(!s.cmp())
        return;

    c_mutex_locker lock(m_sockets);

    vector<pair<c_socket,my_time_t> >::iterator i1;
    for(i1=sock_linger.begin(); i1!=sock_linger.end(); ++i1)
        if((*i1).first==s || (*i1).first.was_it_you(s))
            return;
    sock_linger.push_back(make_pair(s,my_time(NULL)));
}

c_socket::c_socket(const c_socket& orig)
{
    c_mutex_locker lock(m_sockets);

    hndl=orig.hndl;
    was_hndl=orig.was_hndl;
#if defined(_WIN32) || defined(_WIN64)
    wnd=orig.wnd;
#endif

    comment=orig.comment;

    received=orig.received;
    sent=orig.sent;

#ifndef USE_OPENSSL
    got_xcred=orig.got_xcred;
    xcred=orig.xcred;
    m_gnutls_session=orig.m_gnutls_session;
#endif

    last_io=orig.last_io;

    closed=orig.closed;

    memcpy(host,orig.host,sizeof(host));
}

/*!
    \brief Dereferences the socket descriptor
    \author VooDooMan
    \version 1
    \date 2007-07
    \param fn Caller's file name (__FILE__)
    \param ln Caller's line number (__LINE__)
    \return Returns OS's socket handle
*/
MY_SOCKET_FD c_socket::deref(const char* fn, int ln)
{
    c_mutex_locker lock(m_sockets);

    map<unsigned,c_socket_hndl>::iterator i1=sock_find(hndl,fn,ln);
    if(sockets.end()==i1)
        return MY_INVALID_SOCKET;
    return (*i1).second.h;
}

bool c_socket::was_it_you(const c_socket& other)
{
    if((unsigned)-1==was_hndl)
        return false;
    if((unsigned)-1==other.was_hndl)
        return false;
    return other.was_hndl==was_hndl;
}

/*!
    \brief Gets first non-zero errno (except for EAGAIN) from low-level BSD sockets API call
    \author VooDooMan
    \version 1
    \date 2007-08-19
    \param socket Socket handle of connected socket
    \return Returns errno
*/
int sock_get_errno(c_socket& socket)
{
    c_mutex_locker lock(m_sockets);

    vector<c_receive_later>::iterator i1;
    for(i1=receive_later.begin(); i1!=receive_later.end(); ++i1) {
        if((*i1).socket==socket) {
            return (*i1).bsd_errno;
        }
    }

    return EBADF;
}

c_socket::~c_socket()
{
    /*
    c_mutex_locker lock(m_sockets);

    hndl=0;

    {
        vector<c_receive_later>::iterator i1;
        for(i1=receive_later.begin(); i1!=receive_later.end(); ++i1) {
            if((*i1).socket.was_it_you(*this)) {
                while(!(*i1).data.empty()) {
                    MY_DELETE_A((*i1).data.front().data);
                    (*i1).data.erase((*i1).data.begin());
                }
                break;
            }
        }
    }

    {
        vector<c_send_later>::iterator i1;
        for(i1=send_later.begin(); i1!=send_later.end(); ++i1) {
            if((*i1).socket.was_it_you(*this)) {
                while(!(*i1).data.empty()) {
                    MY_DELETE_A((*i1).data.front().data);
                    (*i1).data.erase((*i1).data.begin());
                }
                break;
            }
        }
    }
    */
}

