/***************************************************************************
                          sock.h  -  description
                             -------------------
    begin                : Thu Dec 9 2004
    copyright            : (C) 2004-2007 by Marian VooDooMan Meravy
    email                : ghostvoodooman -a.T= users (dot) sourceforge.net
 ***************************************************************************/

/* $Id: sock.h,v 1.51 2008/08/23 15:10:42 ghostvoodooman Exp $ */

/***************************************************************************

VooDoo cIRCle - an IRC (ro)bot
Copyright (C) 2004-2008 by Marian VooDooMan Meravy
  (ghostvoodooman -a.T= users (dot) sourceforge.net)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************/

/*!
    \file
    \brief Provides low-level IP socket communication
    \version $Id: sock.h,v 1.51 2008/08/23 15:10:42 ghostvoodooman Exp $
*/

#ifndef SOCK_H__
#define SOCK_H__

#if defined(HAVE_CONFIG_H)
#   include <config.h>
#endif

#include <string>
#include <vector>
#include <stack>
#include <list>
#include <map>

#include <time.h>

using namespace std;

#include "utils/memleak.h"
#include "utils/assert.h"
#include "utils.h"
#include "params.h"

#include "compat.h"
#include "sock_shared.h"

/*!
    \brief Stores data that should be logged
*/
class c_sock_log
{
public:
    int ec;             //!< Error code
    enum e_type
    {
        TYPE_MY = 1,    //!< It is my own error code
        TYPE_ERRNO,     //!< Error code got from 'errno'
        TYPE_GAI,       //!< Message should be taken by call to 'gai_strerror()'
        TYPE_GETHOSTBY, //!< Message should be taken by call to 'hstrerror()'
    };                  //!< Type of error code
    e_type type;        //!< Type of error code
    string log;         //!< Message

    /*!
        \brief Constructor
        \param error_code OS error code
        \param message Message
        \param error_type Type of error code
    */
    c_sock_log(int error_code, e_type error_type, const string& message)
        : ec(error_code), type(error_type), log(message)
    { ; }
};

void sock_send_cache(bool force_garbaging = false);
void sock_flush_later(c_socket& socket);
void sock_cancel_caching(c_socket& socket);
my_time_t sock_get_last_io(c_socket& socket, char mode='a');
void sock_set_last_io(c_socket& socket, my_time_t last_io, char mode);
void sock_unread(c_socket& socket, char* data, size_t size);

void sock_create_later(c_socket& socket, bool using_window);
void sock_async(c_socket& socket);

int sock_get_address_type(string host);
uint32 sock_ipv4_from_string(const char* str);

void sock_cache_stats(c_socket& socket, size_t& to_receive, size_t& to_send);

char* sock_error(int code, c_sock_log::e_type type = c_sock_log::TYPE_ERRNO);
int sock_startup();
int sock_shutdown();
c_socket sock_connect(char* localIP, char* remote, unsigned short remote_port, int& error_code, bool async, const char* comment);
size_t sock_send(c_socket& socket, const char* data, size_t size, int& error_code);
void sock_close(c_socket& socket, const char* fn, int ln);
size_t sock_read(c_socket& socket, char* data, size_t size, int& error_code, bool& connection_closed);
void sock_reverse(const char* address, char* dns_name, int dns_max_size, int* errnum = NULL, c_sock_log::e_type* errtype = NULL);
size_t sock_get_receive_size(c_socket& socket, bool& closed);

// for server
c_socket sock_bind(in_addr local, unsigned short local_port, int& error_code, const char* comment);
c_socket sock_bind6(char* local, unsigned short port, int& error_code, const char* comment);
c_socket sock_accept(c_socket& socket, const char* comment, bool async);
c_socket sock_accept6(c_socket& socket, const char* comment, bool async);
in_addr sock_get_remote_addr(c_socket& socket, int& error_code);
in_addr6_ sock_get_remote_addr6(c_socket& socket);

void sock_reverse(in_addr6_ address, char* dns_name, int dns_max_size, vector<c_sock_log>& log, int* errnum = NULL, c_sock_log::e_type* errtype = NULL);

void sock_set_nonblocking(c_socket& socket, bool non_blocking);

void sock_comment(c_socket& socket, const string& comment);

c_socket sock_thread_connect(const string& localIP, const string& remote, unsigned short remote_port, int& error_code, const string& comment, vector<c_sock_log>& log, int* errnum = NULL, c_sock_log::e_type* errtype = NULL);

unsigned short sock_get_local_port(c_socket& socket);
unsigned short sock_get_remote_port(c_socket& socket);

void sock_register_socket(MY_SOCKET_FD s);
void sock_unregister_socket(MY_SOCKET_FD s);
void sock_interruptible_sleep(unsigned int ms);

bool sock_temporary_error(int code, c_sock_log::e_type type);

void sock_set_tcp_keepalive(c_socket& socket);

void sock_log_fd_array_size(bool anyway = false);

vector<in_addr> sock_resolve_multi(const char* domain, vector<c_sock_log>& log, int* errnum = NULL, c_sock_log::e_type* errtype = NULL);
vector<in_addr6_> sock_resolve6_multi(const char* domain, vector<c_sock_log>& log, int* errnum = NULL, c_sock_log::e_type* errtype = NULL);
in_addr sock_resolve(const char* domain, int* errnum = NULL, c_sock_log::e_type* errtype = NULL);
in_addr6_ sock_resolve6(const char* domain, vector<c_sock_log>& log, int* errnum = NULL, c_sock_log::e_type* errtype = NULL);

bool sock_closed(c_socket& socket);

int sock_get_errno(c_socket& socket);

void sock_linger_push(c_socket& s);

#define MY_SELECT_ERROR         (0x70000001U)
#define MY_NO_IPV6              (0x70000002U)
#define MY_SSL_ERROR            (0x70000003U)
#define MY_UNKNOWN_ERROR        (0x70000004U)
#define MY_DNS_ERROR            (0x70000005U)

#endif
