/***************************************************************************
                          sock_garbage.cpp  -  description
                             -------------------
    begin                : Thu Mar 22 2007
    copyright            : (C) 2007 by Marian VooDooMan Meravy
    email                : ghostvoodooman -a.T= users (dot) sourceforge.net
 ***************************************************************************/

/* $Id: sock_garbage.cpp,v 1.16 2008/03/11 21:11:31 ghostvoodooman Exp $ */

/***************************************************************************

VooDoo cIRCle - an IRC (ro)bot
Copyright (C) 2007-2008 by Marian VooDooMan Meravy
  (ghostvoodooman -a.T= users (dot) sourceforge.net)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************/

/*!
    \file
    \brief Provides garbaging of closed socket connections
    \version $Id: sock_garbage.cpp,v 1.16 2008/03/11 21:11:31 ghostvoodooman Exp $
*/

#if defined(HAVE_CONFIG_H)
#   include <config.h>
#endif

#include <vector>

#if defined(HAVE_MALLOC_H)
#include <malloc.h>
#endif

#include "utils/memleak.h"
#include "utils/assert.h"
#include "utils.h"

#include "params.h"

#include "sock_garbage.h"
#include "sock.h"

using namespace std;

#include "utils/precomp.h"

#if defined(_MSC_VER) || defined(__BORLANDC__)
#   pragma hdrstop
#endif

/*!
    \brief Socket garbage entry
    \date 2007-03-22
    \author VooDooMan
*/
class c_sock_garbage_entry
{
    threaded_connect_id _id;        //!< Threaded connect ID
    c_socket _s;                    //!< Socket
    c_ssl* _ssl;                    //!< SSL descriptor

public:

    c_sock_garbage_entry(threaded_connect_id id, const c_socket& s, c_ssl* ssl);

    bool process();
};

static vector<c_sock_garbage_entry> garbage;    //!< Garbage

/*!
    \brief Pushes socket to garbage
    \version 1
    \date 2007-03-22
    \param id Threaded connect ID (or TCID_INVALID if not applicable)
    \param s Socket (if applicable)
    \param ssl SSL descriptor (if applicable, or NULL)
*/
void sock_garbage_push(threaded_connect_id &id, c_socket& s, c_ssl* &ssl)
{
    garbage.push_back(c_sock_garbage_entry(id,s,ssl));

    id=TCID_INVALID;
    s.clear();
    ssl=NULL;
}

/*!
    \brief Constructor
    \version 1
    \date 2007-03-22
    \param id Threaded connect ID (or TCID_INVALID if not applicable)
    \param s Socket (if applicable)
    \param ssl SSL descriptor (if applicable, or NULL)
*/
c_sock_garbage_entry::c_sock_garbage_entry(threaded_connect_id id, const c_socket& s, c_ssl* ssl)
{
    _id=id;
    if(ssl) {
        c_socket* tmp=ssl->get_socket();
        // check whether 's' is associated with 'ssl' or independent socket
        if(tmp && *tmp==s)
            _s.clear();
        else
            _s=s;
    } else
        _s=s;
    _ssl=ssl;
}

/*!
    \brief Should be called in a loop
    \version 1
    \date 2007-03-22
    \retval true This class should be MY_DELETE'd
    \retval false Keep this class in memory, not finished yet, and call this again later
*/
bool c_sock_garbage_entry::process()
{
    if(TCID_INVALID!=_id) {
        int ec=0;
        c_socket tmp;
        bool done=threads_get_connect(_id,ec,tmp);
        if(done) {
            _id=TCID_INVALID;
            sock_linger_push(tmp);
        }
    }

    if(_ssl) {
        ssl_put_to_garbage(_ssl);
        _ssl=NULL;
    }

    if(_s.cmp()) {
        sock_linger_push(_s);
        _s.clear();
    }

    if(TCID_INVALID==_id && !_s.cmp() && !_ssl)
        return true;

    return false;
}

/*!
    \brief Should be called in a loop
    \version 1
    \date 2007-03-22
*/
void sock_garbage_loop()
{
again:
    vector<c_sock_garbage_entry>::iterator i1;
    for(i1=garbage.begin(); i1!=garbage.end(); ++i1) {
        if((*i1).process()) {
            garbage.erase(i1);
            goto again;
        }
    }
}

