/***************************************************************************
                          ssl.cpp  -  description
                             -------------------
    begin                : Thu Dec 9 2004
    copyright            : (C) 2004-2007 by Marian VooDooMan Meravy
    email                : ghostvoodooman -a.T= users (dot) sourceforge.net
 ***************************************************************************/

/* $Id: ssl.cpp,v 1.65 2008/09/04 05:23:49 ghostvoodooman Exp $ */

/***************************************************************************

VooDoo cIRCle - an IRC (ro)bot
Copyright (C) 2004-2008 by Marian VooDooMan Meravy
  (ghostvoodooman -a.T= users (dot) sourceforge.net)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************/

/*!
    \file
    \brief Provides interface to SSL library
    \version $Id: ssl.cpp,v 1.65 2008/09/04 05:23:49 ghostvoodooman Exp $
*/

#if defined(HAVE_CONFIG_H)
#   include <config.h>
#endif

#include "params.h"
#include "utils/memleak.h"
#include "utils/assert.h"
#include "compat.h"

#include "sock_shared.h"

#include <string>
#include <vector>
#include <stack>
#include <list>
#include <map>
#include <deque>
#include <iostream>

#include <stdlib.h>
#include <stdio.h>
#include <memory.h>
#include <errno.h>

#if defined(_WIN32) || defined(_WIN64) || defined(__CYGWIN__)
#   include <process.h>
#else
#    include <sys/types.h>
#    include <unistd.h>
#endif

#ifdef HAVE_SYS_TIME_H
#   include <sys/time.h>
#endif

#ifdef USE_OPENSSL
#   include <openssl/rsa.h>       /* SSLeay stuff */
#   include <openssl/crypto.h>
#   include <openssl/x509.h>
#   include <openssl/pem.h>
#   include <openssl/ssl.h>
#   include <openssl/err.h>
#   include <openssl/x509v3.h>
#   include <openssl/rand.h>
#   define GNUTLS_UNUSEDARG(x)  x
#else
#   define GNUTLS_UNUSEDARG(x)  /* unused */
#endif

#include "utils.h"      // because we want to enable mem leak detection

#include "sock_garbage.h"

#include "params.h"

#if !defined(_WIN32) && !defined(_WIN64)
#   ifdef USE_PTHREAD
#       ifdef HAVE_LIBPTHREAD
#           include <pthread.h>
#       endif
#   else
#       ifdef HAVE_LIBPTHREAD
#           undef HAVE_LIBPTHREAD
#       endif
#   endif
#endif

#ifndef USE_OPENSSL
#   include <gcrypt.h>
#   include <gnutls/gnutls.h>
#   include <gnutls/x509.h>
#endif

#include "compat.h"

using namespace std;

#include "utils/memleak.h"
#include "utils/assert.h"

#include "ssl.h"
#include "sock.h"
#include "log.h"
#include "compat.h"
#include "thread/mutex.h"
#include "utils/numconv.hh"

#include "utils/precomp.h"

#if defined(_MSC_VER) || defined(__BORLANDC__)
#   pragma hdrstop
#endif

int num_transactions=0;

#ifndef USE_OPENSSL

#define ERR_error_string XXXXXXXXXXX /* to prohibit using this function */

ssize_t pull_func(gnutls_transport_ptr handle, void* data, size_t size)
{
    c_socket* s=(c_socket*)handle;
    bool closed=false;
    int got=static_cast<int>(sock_get_receive_size(*s,closed));
    if(got<static_cast<signed>(size)) {
        if(closed) {
#if !defined(HAVE_WIN32_POSIX_EMULATION)
            gnutls_transport_set_errno(s->m_gnutls_session,0);
#else
            errno=0;
#endif
            return 0;
        }
        int ec=EAGAIN;
#if !defined(HAVE_WIN32_POSIX_EMULATION)
        gnutls_transport_set_errno(s->m_gnutls_session,ec);
#else
        errno=ec;
#endif
        return -1;
    }
    int ec=0;
    ssize_t res=0;
    while(res<static_cast<signed>(size)) {
        res+=static_cast<ssize_t>(sock_read(*s,&((char*)data)[res],size-res,ec,closed));
        if(ec || closed) {
#if !defined(HAVE_WIN32_POSIX_EMULATION)
            gnutls_transport_set_errno(s->m_gnutls_session,sock_get_errno(*s));
#else
            errno=sock_get_errno(*s);
#endif
            if(closed)
                return 0;
            return -1;
        }
    }
    return res;
}

ssize_t push_func(gnutls_transport_ptr handle, const void* data, size_t size)
{
    c_socket* s=(c_socket*)handle;
    int ec=0;
    ssize_t res=static_cast<ssize_t>(sock_send(*s,(const char*)data,size,ec));
    if(ec) {
#if !defined(HAVE_WIN32_POSIX_EMULATION)
        gnutls_transport_set_errno(s->m_gnutls_session,sock_get_errno(*s));
#else
        errno=sock_get_errno(*s);
#endif
        return -1;
    }
    return res;
}

/*!
    \brief Provides better error string
    \param rc Error code
    \return return error string
*/
static string my_ERR_error_string(int rc)
{
    string err=tos(rc);
    err+=", 0x";
    err+=toshex(rc);
    err+=": ";
    switch(rc) {
        case GNUTLS_E_SUCCESS:
            err+="GNUTLS_E_SUCCESS";
            break;
        case GNUTLS_E_UNKNOWN_COMPRESSION_ALGORITHM:
            err+="GNUTLS_E_UNKNOWN_COMPRESSION_ALGORITHM";
            break;
        case GNUTLS_E_UNKNOWN_CIPHER_TYPE:
            err+="GNUTLS_E_UNKNOWN_CIPHER_TYPE";
            break;
        case GNUTLS_E_LARGE_PACKET:
            err+="GNUTLS_E_LARGE_PACKET";
            break;
        case GNUTLS_E_UNSUPPORTED_VERSION_PACKET:
            err+="GNUTLS_E_UNSUPPORTED_VERSION_PACKET";
            break;
        case GNUTLS_E_UNEXPECTED_PACKET_LENGTH:
            err+="GNUTLS_E_UNEXPECTED_PACKET_LENGTH";
            break;
        case GNUTLS_E_INVALID_SESSION:
            err+="GNUTLS_E_INVALID_SESSION";
            break;
        case GNUTLS_E_FATAL_ALERT_RECEIVED:
            err+="GNUTLS_E_FATAL_ALERT_RECEIVED";
            break;
        case GNUTLS_E_UNEXPECTED_PACKET:
            err+="GNUTLS_E_UNEXPECTED_PACKET";
            break;
        case GNUTLS_E_WARNING_ALERT_RECEIVED:
            err+="GNUTLS_E_WARNING_ALERT_RECEIVED";
            break;
        case GNUTLS_E_ERROR_IN_FINISHED_PACKET:
            err+="GNUTLS_E_ERROR_IN_FINISHED_PACKET";
            break;
        case GNUTLS_E_UNEXPECTED_HANDSHAKE_PACKET:
            err+="GNUTLS_E_UNEXPECTED_HANDSHAKE_PACKET";
            break;
        case GNUTLS_E_UNKNOWN_CIPHER_SUITE:
            err+="GNUTLS_E_UNKNOWN_CIPHER_SUITE";
            break;
        case GNUTLS_E_UNWANTED_ALGORITHM:
            err+="GNUTLS_E_UNWANTED_ALGORITHM";
            break;
        case GNUTLS_E_MPI_SCAN_FAILED:
            err+="GNUTLS_E_MPI_SCAN_FAILED";
            break;
        case GNUTLS_E_DECRYPTION_FAILED:
            err+="GNUTLS_E_DECRYPTION_FAILED";
            break;
        case GNUTLS_E_MEMORY_ERROR:
            err+="GNUTLS_E_MEMORY_ERROR";
            break;
        case GNUTLS_E_DECOMPRESSION_FAILED:
            err+="GNUTLS_E_DECOMPRESSION_FAILED";
            break;
        case GNUTLS_E_COMPRESSION_FAILED:
            err+="GNUTLS_E_COMPRESSION_FAILED";
            break;
        case GNUTLS_E_AGAIN:
            err+="GNUTLS_E_AGAIN";
            break;
        case GNUTLS_E_EXPIRED:
            err+="GNUTLS_E_EXPIRED";
            break;
        case GNUTLS_E_DB_ERROR:
            err+="GNUTLS_E_DB_ERROR";
            break;
        case GNUTLS_E_SRP_PWD_ERROR:
            err+="GNUTLS_E_SRP_PWD_ERROR";
            break;
        case GNUTLS_E_INSUFFICIENT_CREDENTIALS:
        //case GNUTLS_E_INSUFICIENT_CREDENTIALS:
        //case GNUTLS_E_INSUFFICIENT_CRED:
        //case GNUTLS_E_INSUFICIENT_CRED:
            err+="GNUTLS_E_INSUFFICIENT_CREDENTIALS";
            break;
        case GNUTLS_E_HASH_FAILED:
            err+="GNUTLS_E_HASH_FAILED";
            break;
        case GNUTLS_E_BASE64_DECODING_ERROR:
            err+="GNUTLS_E_BASE64_DECODING_ERROR";
            break;
        case GNUTLS_E_MPI_PRINT_FAILED:
            err+="GNUTLS_E_MPI_PRINT_FAILED";
            break;
        case GNUTLS_E_REHANDSHAKE:
            err+="GNUTLS_E_REHANDSHAKE";
            break;
        case GNUTLS_E_GOT_APPLICATION_DATA:
            err+="GNUTLS_E_GOT_APPLICATION_DATA";
            break;
        case GNUTLS_E_RECORD_LIMIT_REACHED:
            err+="GNUTLS_E_RECORD_LIMIT_REACHED";
            break;
        case GNUTLS_E_ENCRYPTION_FAILED:
            err+="GNUTLS_E_ENCRYPTION_FAILED";
            break;
        case GNUTLS_E_PK_ENCRYPTION_FAILED:
            err+="GNUTLS_E_PK_ENCRYPTION_FAILED";
            break;
        case GNUTLS_E_PK_DECRYPTION_FAILED:
            err+="GNUTLS_E_PK_DECRYPTION_FAILED";
            break;
        case GNUTLS_E_PK_SIGN_FAILED:
            err+="GNUTLS_E_PK_SIGN_FAILED";
            break;
        case GNUTLS_E_X509_UNSUPPORTED_CRITICAL_EXTENSION:
            err+="GNUTLS_E_X509_UNSUPPORTED_CRITICAL_EXTENSION";
            break;
        case GNUTLS_E_KEY_USAGE_VIOLATION:
            err+="GNUTLS_E_KEY_USAGE_VIOLATION";
            break;
        case GNUTLS_E_NO_CERTIFICATE_FOUND:
            err+="GNUTLS_E_NO_CERTIFICATE_FOUND";
            break;
        case GNUTLS_E_INVALID_REQUEST:
            err+="GNUTLS_E_INVALID_REQUEST";
            break;
        case GNUTLS_E_SHORT_MEMORY_BUFFER:
            err+="GNUTLS_E_SHORT_MEMORY_BUFFER";
            break;
        case GNUTLS_E_INTERRUPTED:
            err+="GNUTLS_E_INTERRUPTED";
            break;
        case GNUTLS_E_PUSH_ERROR:
            err+="GNUTLS_E_PUSH_ERROR";
            break;
        case GNUTLS_E_PULL_ERROR:
            err+="GNUTLS_E_PULL_ERROR";
            break;
        case GNUTLS_E_RECEIVED_ILLEGAL_PARAMETER:
            err+="GNUTLS_E_RECEIVED_ILLEGAL_PARAMETER";
            break;
        case GNUTLS_E_REQUESTED_DATA_NOT_AVAILABLE:
            err+="GNUTLS_E_REQUESTED_DATA_NOT_AVAILABLE";
            break;
        case GNUTLS_E_PKCS1_WRONG_PAD:
            err+="GNUTLS_E_PKCS1_WRONG_PAD";
            break;
        case GNUTLS_E_RECEIVED_ILLEGAL_EXTENSION:
            err+="GNUTLS_E_RECEIVED_ILLEGAL_EXTENSION";
            break;
        case GNUTLS_E_INTERNAL_ERROR:
            err+="GNUTLS_E_INTERNAL_ERROR";
            break;
        case GNUTLS_E_DH_PRIME_UNACCEPTABLE:
            err+="GNUTLS_E_DH_PRIME_UNACCEPTABLE";
            break;
        case GNUTLS_E_FILE_ERROR:
            err+="GNUTLS_E_FILE_ERROR";
            break;
        case GNUTLS_E_TOO_MANY_EMPTY_PACKETS:
            err+="GNUTLS_E_TOO_MANY_EMPTY_PACKETS";
            break;
        case GNUTLS_E_UNKNOWN_PK_ALGORITHM:
            err+="GNUTLS_E_UNKNOWN_PK_ALGORITHM";
            break;
        case GNUTLS_E_INIT_LIBEXTRA:
            err+="GNUTLS_E_INIT_LIBEXTRA";
            break;
        case GNUTLS_E_LIBRARY_VERSION_MISMATCH:
            err+="GNUTLS_E_LIBRARY_VERSION_MISMATCH";
            break;
        case GNUTLS_E_NO_TEMPORARY_RSA_PARAMS:
            err+="GNUTLS_E_NO_TEMPORARY_RSA_PARAMS";
            break;
        case GNUTLS_E_LZO_INIT_FAILED:
            err+="GNUTLS_E_LZO_INIT_FAILED";
            break;
        case GNUTLS_E_NO_COMPRESSION_ALGORITHMS:
            err+="GNUTLS_E_NO_COMPRESSION_ALGORITHMS";
            break;
        case GNUTLS_E_NO_CIPHER_SUITES:
            err+="GNUTLS_E_NO_CIPHER_SUITES";
            break;
        case GNUTLS_E_OPENPGP_GETKEY_FAILED:
            err+="GNUTLS_E_OPENPGP_GETKEY_FAILED";
            break;
        case GNUTLS_E_PK_SIG_VERIFY_FAILED:
            err+="GNUTLS_E_PK_SIG_VERIFY_FAILED";
            break;
        case GNUTLS_E_ILLEGAL_SRP_USERNAME:
            err+="GNUTLS_E_ILLEGAL_SRP_USERNAME";
            break;
        case GNUTLS_E_SRP_PWD_PARSING_ERROR:
            err+="GNUTLS_E_SRP_PWD_PARSING_ERROR";
            break;
        case GNUTLS_E_NO_TEMPORARY_DH_PARAMS:
            err+="GNUTLS_E_NO_TEMPORARY_DH_PARAMS";
            break;
        case GNUTLS_E_ASN1_ELEMENT_NOT_FOUND:
            err+="GNUTLS_E_ASN1_ELEMENT_NOT_FOUND";
            break;
        case GNUTLS_E_ASN1_IDENTIFIER_NOT_FOUND:
            err+="GNUTLS_E_ASN1_IDENTIFIER_NOT_FOUND";
            break;
        case GNUTLS_E_ASN1_DER_ERROR:
            err+="GNUTLS_E_ASN1_DER_ERROR";
            break;
        case GNUTLS_E_ASN1_VALUE_NOT_FOUND:
            err+="GNUTLS_E_ASN1_VALUE_NOT_FOUND";
            break;
        case GNUTLS_E_ASN1_GENERIC_ERROR:
            err+="GNUTLS_E_ASN1_GENERIC_ERROR";
            break;
        case GNUTLS_E_ASN1_VALUE_NOT_VALID:
            err+="GNUTLS_E_ASN1_VALUE_NOT_VALID";
            break;
        case GNUTLS_E_ASN1_TAG_ERROR:
            err+="GNUTLS_E_ASN1_TAG_ERROR";
            break;
        case GNUTLS_E_ASN1_TAG_IMPLICIT:
            err+="GNUTLS_E_ASN1_TAG_IMPLICIT";
            break;
        case GNUTLS_E_ASN1_TYPE_ANY_ERROR:
            err+="GNUTLS_E_ASN1_TYPE_ANY_ERROR";
            break;
        case GNUTLS_E_ASN1_SYNTAX_ERROR:
            err+="GNUTLS_E_ASN1_SYNTAX_ERROR";
            break;
        case GNUTLS_E_ASN1_DER_OVERFLOW:
            err+="GNUTLS_E_ASN1_DER_OVERFLOW";
            break;
        /*
        case GNUTLS_E_OPENPGP_TRUSTDB_VERSION_UNSUPPORTED:
            err+="GNUTLS_E_OPENPGP_TRUSTDB_VERSION_UNSUPPORTED";
            break;
        */
        case GNUTLS_E_OPENPGP_UID_REVOKED:
            err+="GNUTLS_E_OPENPGP_UID_REVOKED";
            break;
        //case GNUTLS_E_X509_CERTIFICATE_ERROR:
        case GNUTLS_E_CERTIFICATE_ERROR:
            err+="GNUTLS_E_CERTIFICATE_ERROR";
            break;
        case GNUTLS_E_CERTIFICATE_KEY_MISMATCH:
            err+="GNUTLS_E_CERTIFICATE_KEY_MISMATCH";
            break;
        case GNUTLS_E_UNSUPPORTED_CERTIFICATE_TYPE:
            err+="GNUTLS_E_UNSUPPORTED_CERTIFICATE_TYPE";
            break;
        case GNUTLS_E_X509_UNKNOWN_SAN:
            err+="GNUTLS_E_X509_UNKNOWN_SAN";
            break;
        case GNUTLS_E_OPENPGP_FINGERPRINT_UNSUPPORTED:
            err+="GNUTLS_E_OPENPGP_FINGERPRINT_UNSUPPORTED";
            break;
        case GNUTLS_E_X509_UNSUPPORTED_ATTRIBUTE:
            err+="GNUTLS_E_X509_UNSUPPORTED_ATTRIBUTE";
            break;
        case GNUTLS_E_UNKNOWN_HASH_ALGORITHM:
            err+="GNUTLS_E_UNKNOWN_HASH_ALGORITHM";
            break;
        case GNUTLS_E_UNKNOWN_PKCS_CONTENT_TYPE:
            err+="GNUTLS_E_UNKNOWN_PKCS_CONTENT_TYPE";
            break;
        case GNUTLS_E_UNKNOWN_PKCS_BAG_TYPE:
            err+="GNUTLS_E_UNKNOWN_PKCS_BAG_TYPE";
            break;
        case GNUTLS_E_INVALID_PASSWORD:
            err+="GNUTLS_E_INVALID_PASSWORD";
            break;
        case GNUTLS_E_MAC_VERIFY_FAILED:
            err+="GNUTLS_E_MAC_VERIFY_FAILED";
            break;
        /*
        case GNUTLS_E_CONSTRAINT_ERROR:
            err+="GNUTLS_E_CONSTRAINT_ERROR";
            break;
        */
        case GNUTLS_E_BASE64_ENCODING_ERROR:
            err+="GNUTLS_E_BASE64_ENCODING_ERROR";
            break;
        case GNUTLS_E_INCOMPATIBLE_GCRYPT_LIBRARY:
        //case GNUTLS_E_INCOMPATIBLE_CRYPTO_LIBRARY:
            err+="GNUTLS_E_INCOMPATIBLE_CRYPTO_LIBRARY";
            break;
        case GNUTLS_E_INCOMPATIBLE_LIBTASN1_LIBRARY:
            err+="GNUTLS_E_INCOMPATIBLE_LIBTASN1_LIBRARY";
            break;
        case GNUTLS_E_OPENPGP_KEYRING_ERROR:
            err+="GNUTLS_E_OPENPGP_KEYRING_ERROR";
            break;
        case GNUTLS_E_X509_UNSUPPORTED_OID:
            err+="GNUTLS_E_X509_UNSUPPORTED_OID";
            break;
        /*case GNUTLS_E_RANDOM_FAILED:
            err+="GNUTLS_E_RANDOM_FAILED";
            break;*/
        case GNUTLS_E_UNIMPLEMENTED_FEATURE:
            err+="GNUTLS_E_UNIMPLEMENTED_FEATURE";
            break;
        default:
            err+="(other error)";
            break;
    }
    err+=": ";
    err+=gnutls_strerror(rc);
    return err;
}
#endif

/*!
    \def chain_verify_depth
    \brief Depth of verifying chained certificates
*/
enum {
    chain_verify_depth = 128,
};

/*!
    \brief Stores data for verification purposes
    \author VooDooMan
    \version 2
    \date 2004, 2005-11-07
*/
class ssl_meta_data
{
public:
    bool ok;                                        //!< If true, there's no verification error yet
    string error;                                   //!< Verification error message
    c_ssl* ssl;                                     //!< SSL descriptor

    ssl_meta_data(c_ssl* ssl_descriptor)
        : ok(true)
        , ssl(ssl_descriptor)
    { ; }
};

static c_mutex extra_data_mutex;                    //!< Certificate chain verification locking object

#ifdef USE_OPENSSL
map<SSL*,pair<int,ssl_meta_data*> > extra_data;     //!< SSL descriptor to extra data index
#endif

#ifndef USE_OPENSSL

#ifdef _MSC_VER
#   pragma warning( push )
#   pragma warning( disable : 4706 ) // warning C4706: assignment within conditional expression
#endif

/* Helper functions to load a certificate and key
 * files into memory.
 */
static gnutls_datum load_file(const char *file)
{
  FILE *f;
  gnutls_datum loaded_file = { NULL, 0 };
  long filelen;
  void *ptr;

  if (!(f = fopen(file, "r"))
      || fseek(f, 0, SEEK_END) != 0
      || (filelen = ftell(f)) < 0
      || fseek(f, 0, SEEK_SET) != 0
      || !(ptr = my_malloc((size_t)filelen))
      || fread(ptr, 1, (size_t)filelen, f) < (size_t)filelen)
    {
      return loaded_file;
    }

  loaded_file.data = (unsigned char*)ptr;
  loaded_file.size = (unsigned int)filelen;
  return loaded_file;
}

#ifdef _MSC_VER
#   pragma warning( pop )
#endif

static void unload_file(gnutls_datum data)
{
  my_free(data.data);
}
#endif

/*!
    \brief Loads a certificate from file
    \author VooDooMan
    \version 1
    \date 2005
    \param filename Name of file
*/
#ifdef USE_OPENSSL
X509 *ssl_read_cert(char* filename)
#else
gnutls_x509_crt *ssl_read_cert(char* filename)
#endif
{
#ifdef USE_OPENSSL
    c_ssl::clear_error_queue();

    X509* cert;
    BIO* s=BIO_new_file(filename,"r");

    if(!s) {
        c_ssl::clear_error_queue();
        return NULL;
    }

    BIO* f=BIO_new(BIO_f_base64());
    if(!f) {
        BIO_free(s);
        c_ssl::clear_error_queue();
        return NULL;
    }
    s=BIO_push(f,s);
    cert=d2i_X509_bio(s,NULL);
    BIO_free_all(s);

    c_ssl::clear_error_queue();
    return cert;
#else
    gnutls_x509_crt crt;

    gnutls_datum data=load_file(filename);
    gnutls_x509_crt_init(&crt);

    int ret=gnutls_x509_crt_import(crt,&data,GNUTLS_X509_FMT_PEM);
    if(ret<0)
    {
      return NULL;
    }

    unload_file(data);

    gnutls_x509_crt* c=MY_NEW gnutls_x509_crt(crt);

    return c;
#endif
}

#ifndef USE_OPENSSL
/* This function will try to verify the peer's certificate, and
 * also check the activation, expiration dates.
 */
static string verify_certificate(gnutls_session session, const char *UNUSEDARG(hostname), const char* expected_cert, string& remote_cert, bool only_get_peers_cert = false, bool disable_ca_verify = false, bool ignore_expired_peers_cert = false)
{
    string res;

    unsigned int status;
    const gnutls_datum *cert_list;
    int ret;
    unsigned cert_list_size;
    gnutls_x509_crt cert;

    /* This verification function uses the trusted CAs in the credentials
    * structure. So you must have installed one or more CA certificates.
    */
    ret = gnutls_certificate_verify_peers2(session, &status);

    if(disable_ca_verify) {
        status&=~(GNUTLS_CERT_SIGNER_NOT_FOUND | GNUTLS_CERT_SIGNER_NOT_CA);
    }

    if (!only_get_peers_cert && ret < 0)
        {
        //printf ("Error\n");
            if(!res.empty())
                res+=", ";
        res+="Certificate verifycation error";
        }

    if(!disable_ca_verify) {
        if (!only_get_peers_cert && (status & GNUTLS_CERT_INVALID)) {
            //printf ("The certificate is not trusted.\n");
            if(!res.empty())
                res+=", ";
            res+="The certificate is not trusted";
        }
    }

    if(!disable_ca_verify) {
        if (!only_get_peers_cert && (status & GNUTLS_CERT_SIGNER_NOT_FOUND)) {
            //printf ("The certificate hasn't got a known issuer.\n");
            if(!res.empty())
                res+=", ";
            res+="The certificate hasn't got a known issuer";
        }
    }

    if (!only_get_peers_cert && (status & GNUTLS_CERT_REVOKED)) {
        //printf ("The certificate has been revoked.\n");
        if(!res.empty())
            res+=", ";
        res+="The certificate has been revoked";
    }


    /* Up to here the process is the same for X.509 certificates and
    * OpenPGP keys. From now on X.509 certificates are assumed. This can
    * be easily extended to work with openpgp keys as well.
    */
    if (gnutls_certificate_type_get (session) != GNUTLS_CRT_X509) {
        if(!res.empty())
            res+=", ";
        res+="Not an X.509 certificate";
        return res;
    }

    if (gnutls_x509_crt_init (&cert) < 0)
        {
        //printf ("error in initialization\n");
            if(!res.empty())
                res+=", ";
            res+="Error in initialization";
            gnutls_x509_crt_deinit (cert);
        return res;
        }

    cert_list = gnutls_certificate_get_peers (session, &cert_list_size);
    if (cert_list == NULL)
        {
        //printf ("No certificate was found!\n");
            if(!res.empty())
                res+=", ";
            res+="No certificate found";
            gnutls_x509_crt_deinit (cert);
        return res;
        }

    /* This is not a real world example, since we only check the first
    * certificate in the given chain.
    */
    if (gnutls_x509_crt_import (cert, &cert_list[0], GNUTLS_X509_FMT_DER) < 0)
        {
        //printf ("error parsing certificate\n");
            if(!res.empty())
                res+=", ";
            res+="Error parsing certificate";
            gnutls_x509_crt_deinit (cert);
        return res;
        }

    my_time_t x;
    x=gnutls_x509_crt_get_expiration_time (cert);
    x&=0xFFFFFFFFUL; // MSVC has 64 bit my_time_t

    if(!only_get_peers_cert) {
        if(!ignore_expired_peers_cert) {
            /* Beware here we do not check for errors.
            */
            if (x < time (0))
                {
                //printf ("The certificate has expired\n");
                    if(!res.empty())
                        res+=", ";
                    res+="The certificate has expired";
                    gnutls_x509_crt_deinit (cert);
                return res;
                }
        }
    }

    x=gnutls_x509_crt_get_activation_time (cert);
    x&=0xFFFFFFFFUL; // MSVC has 64 bit my_time_t

    if(!only_get_peers_cert) {
        if(!ignore_expired_peers_cert) {
            if (x > time (0))
                {
                //printf ("The certificate is not yet activated\n");
                    if(!res.empty())
                        res+=", ";
                    res+="The certificate is not yet activated";
                    gnutls_x509_crt_deinit (cert);
                return res;
                }
        }
    }

        /*
    if (!gnutls_x509_crt_check_hostname (cert, hostname))
        {
        printf ("The certificate's owner does not match hostname '%s'\n",
            hostname);
        return;
        }
        */

    {
        enum {
            BUF_LEN = 1024U * 8U,
            BUF_SPARE = 16U,

            BUF_REAL = BUF_LEN + BUF_SPARE,
        };

        char* buf=MY_NEW char[BUF_REAL];
        memset(buf,0,BUF_REAL);

        size_t buflen=BUF_LEN;
        if(0!=gnutls_x509_crt_export(cert,GNUTLS_X509_FMT_PEM,buf,&buflen)) {
            gnutls_x509_crt_deinit(cert);
            MY_DELETE_A(buf);
            return "gnutls_x509_crt_export() failed";
        } else {
            remote_cert=buf;
        }
        MY_DELETE_A(buf);
    }

    if(expected_cert && expected_cert[0]!=0) {
        FILE* f=fopen(expected_cert,"r");
        if(!f) {
            gnutls_x509_crt_deinit (cert);
            if(!res.empty())
                res+=", ";
            res+="Certificate verifycation error";
            if(!res.empty())
                res+=", ";
            res+="Could not open file \"";
            res+=expected_cert;
            res+="\"";
            return res;
        }
        string read_cert;
        while(!feof(f)) {
            char tmp[1024+1];
            tmp[1024]=0;
            fgets(tmp,1024,f);
            read_cert+=tmp;
        }
        fclose(f);
        f=NULL;
        char* tmp=MY_NEW char[read_cert.length()+1];
        strcpy(tmp,read_cert.c_str());
        gnutls_datum read_cert2;
        read_cert2.data=(unsigned char*)tmp;
        read_cert2.size=static_cast<unsigned>(read_cert.length());

        gnutls_x509_crt r_cert;
        int ret=gnutls_x509_crt_init (&r_cert);
        if(ret!=0) {
            MY_DELETE_A(tmp);
            gnutls_x509_crt_deinit (cert);
            if(!res.empty())
                res+=", ";
            res+="Certificate verifycation error";
            if(!res.empty())
                res+=", ";
            res+="gnutls_x509_crt_init() failed: ";
            res+=my_ERR_error_string(ret);
            return res;
        }

        ret=gnutls_x509_crt_import(r_cert,&read_cert2,GNUTLS_X509_FMT_PEM);
        if(ret!=0) {
            gnutls_x509_crt_deinit (r_cert);
            MY_DELETE_A(tmp);
            gnutls_x509_crt_deinit (cert);
            if(!res.empty())
                res+=", ";
            res+="Certificate verifycation error";
            if(!res.empty())
                res+=", ";
            res+="gnutls_x509_crt_import() failed: ";
            res+=my_ERR_error_string(ret);
            return res;
        }
        char buff_1[1024*2];
        char buff_2[1024*2];
        size_t buff_1_size=sizeof(buff_1)-1;
        size_t buff_2_size=sizeof(buff_2)-1;

        ret=gnutls_x509_crt_get_fingerprint (cert,GNUTLS_DIG_SHA,buff_1,&buff_1_size);
        if(ret!=0) {
            gnutls_x509_crt_deinit (r_cert);
            MY_DELETE_A(tmp);
            gnutls_x509_crt_deinit (cert);
            if(!res.empty())
                res+=", ";
            res+="Certificate verifycation error";
            if(!res.empty())
                res+=", ";
            res+="gnutls_x509_crt_get_fingerprint() failed: ";
            res+=my_ERR_error_string(ret);
            return res;
        }

        ret=gnutls_x509_crt_get_fingerprint (r_cert,GNUTLS_DIG_SHA,buff_2,&buff_2_size);
        if(ret!=0) {
            gnutls_x509_crt_deinit (r_cert);
            MY_DELETE_A(tmp);
            gnutls_x509_crt_deinit (cert);
            if(!res.empty())
                res+=", ";
            res+="Certificate verifycation error";
            if(!res.empty())
                res+=", ";
            res+="gnutls_x509_crt_get_fingerprint() failed: ";
            res+=my_ERR_error_string(ret);
            return res;
        }

        bool ok=true;
        if(buff_1_size!=buff_2_size)
            ok=false;
        if(ok)
            for(size_t ii=0; ii<buff_1_size; ++ii)
                if(buff_1[ii]!=buff_2[ii]) {
                    ok=false;
                    break;
                }

        {
            gnutls_x509_crt_deinit (r_cert);
            MY_DELETE_A(tmp);
            gnutls_x509_crt_deinit (cert);

            if(!ok) {
                if(!res.empty())
                    res+=", ";
                res+="certificate verification failed: SHA1 hash doesn't match";
            }
            return res;
        }
        // NOTREACH
    }
    gnutls_x509_crt_deinit (cert);

    return res;
}
#endif

#ifdef USE_OPENSSL
/*!
    \brief SSL verification callback
    \author Sampo Kellomaki <sampo (at) iki.fi>; Wade Scholine <wades (at) mail.cybg.com>; VooDooMan
    \version 2
    \date 1996-1998, 2004, 2008-04-07
    \param preverify_ok I don't actually know what the SSL wants to say by this parameter ;-)
    \param ctx Context of stored certificates
    \return Returns 1 (one) to continue the verification process, 0 (zero) to terminate
*/
int c_ssl::my_verify_callback(int preverify_ok, X509_STORE_CTX *ctx)
{
    X509   *err_cert;
    int     err, depth;
    SSL    *ssl;
    c_ssl *mydata;

    clear_error_queue();

    err_cert = X509_STORE_CTX_get_current_cert(ctx);
    err = X509_STORE_CTX_get_error(ctx);

    if(X509_V_ERR_DEPTH_ZERO_SELF_SIGNED_CERT==err || X509_V_ERR_CERT_SIGNATURE_FAILURE==err) {
        err=0;
        preverify_ok=1;
    }

    depth = X509_STORE_CTX_get_error_depth(ctx);

    /*
     * Retrieve the pointer to the SSL of the connection currently treated
     * and the application specific data stored into the SSL object.
     */
    {
        c_mutex_locker lock(extra_data_mutex);

        ssl = (SSL*)X509_STORE_CTX_get_ex_data(ctx, SSL_get_ex_data_X509_STORE_CTX_idx());
        map<SSL*,pair<int,ssl_meta_data*> >::iterator i1=extra_data.find(ssl);
        if(extra_data.end()==i1) {
            clear_error_queue();
            return 0;
        }
        mydata=(*i1).second.second->ssl;
        if(!mydata) {
            clear_error_queue();
            return 0;
        }

        /*
         * Catch a too long certificate chain. The depth limit set using
         * SSL_CTX_set_verify_depth() is by purpose set to "limit+1" so
         * that whenever the "depth>verify_depth" condition is met, we
         * have violated the limit and want to log this error condition.
         * We must do it here, because the CHAIN_TOO_LONG error would not
         * be found explicitly; only errors introduced by cutting off the
         * additional certificates would be logged.
         */
        if (depth > chain_verify_depth) {
            preverify_ok = 0;
            err = X509_V_ERR_CERT_CHAIN_TOO_LONG;
            X509_STORE_CTX_set_error(ctx, err);
        }
        bool error=true;
        if(mydata->m_only_get_peers_cert &&
            (
                X509_V_ERR_UNABLE_TO_GET_ISSUER_CERT_LOCALLY==err ||
                X509_V_ERR_UNABLE_TO_VERIFY_LEAF_SIGNATURE==err ||
                X509_V_ERR_SELF_SIGNED_CERT_IN_CHAIN==err ||
                X509_V_ERR_CERT_HAS_EXPIRED==err
            )
            )
            error=false;
        if(mydata->m_disable_ca_verify &&
            (
                X509_V_ERR_UNABLE_TO_GET_ISSUER_CERT_LOCALLY==err ||
                X509_V_ERR_UNABLE_TO_VERIFY_LEAF_SIGNATURE==err ||
                X509_V_ERR_SELF_SIGNED_CERT_IN_CHAIN==err ||
                X509_V_ERR_CERT_UNTRUSTED==err
            )
            )
            error=false;
        if(mydata->m_ignore_expired_peers_cert &&
            (
                X509_V_ERR_CERT_HAS_EXPIRED==err ||
                X509_V_ERR_CERT_NOT_YET_VALID==err
            )
            )
            error=false;
        if(0!=err && !error) {
            preverify_ok=1;
            err=0;
        } else if(0!=err) {
            if (!preverify_ok) {
                enum {
                    BUF_LEN = 1024*8,
                    BUF_SPARE = 16,

                    BUF_LEN_REAL = BUF_LEN + BUF_SPARE,
                };
                char* buf=MY_NEW char[BUF_LEN_REAL];
                memset(buf,0,BUF_LEN_REAL);
                X509_NAME_oneline(X509_get_subject_name(err_cert),buf,BUF_LEN);

                mydata->cert_ok_so_far=false;
                string error="verify error:num=";
                error+=tos(err);
                error+=(string)":"+X509_verify_cert_error_string(err);
                error+=":depth=";
                error+=tos(depth);
                error+=(string)":"+buf;

                memset(buf,0,BUF_LEN_REAL);
                X509_NAME_oneline(X509_get_issuer_name(ctx->current_cert),buf,BUF_LEN);
                error+=" - remote certificate: ";
                error+=buf;
                mydata->set_error(error);

                MY_DELETE_A(buf);
            }
        }
    }

    clear_error_queue();

    return preverify_ok;
}
#endif

#ifdef USE_OPENSSL
SSL_METHOD *meth;                    //!< SSL method for context, unique for program's lifetime
#endif

vector<c_ssl*> ssl_descriptors;             //!< All SSL descriptors
vector<pair<c_ssl*,my_time_t> > ssl_garbage;   //!< All SSL descriptors taht should be closed after some time

#ifndef USE_OPENSSL

gnutls_dh_params dh_params;
gnutls_rsa_params rsa_params;

gnutls_dh_params thread_dh_params;              //!< DH params for thread
gnutls_rsa_params thread_rsa_params;            //!< RSA params for thread
static volatile bool cs_init;                   //! Has been "cs"/"mutex" variable initialized?
#if defined(_WIN32) || defined(_WIN64)
static CRITICAL_SECTION cs;                     //!< Critical section (thread locking)
#else
#   ifdef HAVE_LIBPTHREAD
static pthread_mutex_t mutex;                   //!< Mutex for phtread (thread locking)
#   endif
#endif

volatile bool ssl_dh_thread_running=false;

/*!
    \brief Thread function for DH and RSA params regeneration
    \author VooDooMan
    \version 1
    \date 2005-12-16
    \param d Data for thread
*/
#if defined(_WIN32) || defined(_WIN64)
void __cdecl
#else
void*
#endif
ssl_dh_thread(void* UNUSEDARG(d))
{
    gnutls_dh_params_init(&thread_dh_params);
    gnutls_rsa_params_init(&thread_rsa_params);
    gnutls_dh_params_generate2(thread_dh_params, DH_BITS);
    gnutls_rsa_params_generate2(thread_rsa_params, RSA_BITS);

#if defined(_WIN32) || defined(_WIN64)
    EnterCriticalSection(&cs);
#else
#   ifdef HAVE_LIBPTHREAD
    pthread_mutex_lock(&mutex);
#   endif
#endif

    ssl_dh_thread_running=false;

#if defined(_WIN32) || defined(_WIN64)
    LeaveCriticalSection(&cs);
#else
#   ifdef HAVE_LIBPTHREAD
    pthread_mutex_unlock(&mutex);
#   endif
#endif

#if !defined(_WIN32) && !defined(_WIN64)
    return NULL;
#endif
}

#endif

#if !defined(_MSC_VER) && !defined(__BORLANDC__)
#ifndef USE_OPENSSL
#ifdef USE_PTHREAD

static int gcry_pthread_mutex_init(void **priv)
{
    int err = 0;
    pthread_mutex_t *lock = (pthread_mutex_t *)malloc (sizeof (pthread_mutex_t));
    if (!lock)
        err = 12;
    if (!err) {
        err = pthread_mutex_init (lock, __null);
        if (err)
            free (lock);
        else
            *priv = lock;
    } return err;
}

static int gcry_pthread_mutex_destroy (void **lock)
{
    int err = pthread_mutex_destroy ((pthread_mutex_t *)*lock);
    free (*lock);
    return err;
}

static int gcry_pthread_mutex_lock (void **lock)
{
    return pthread_mutex_lock ((pthread_mutex_t *)*lock);
}

static int gcry_pthread_mutex_unlock (void **lock)
{
    return pthread_mutex_unlock ((pthread_mutex_t *)*lock);
}

static struct gcry_thread_cbs gcry_threads_pthread = { GCRY_THREAD_OPTION_PTHREAD, __null, gcry_pthread_mutex_init, gcry_pthread_mutex_destroy, gcry_pthread_mutex_lock, gcry_pthread_mutex_unlock };

#endif
#endif
#endif

/*!
    \brief Initializes SSL
    \author VooDooMan
    \version 2
    \date 2004, 2008-03-03
    \return true for error
*/
void ssl_init()
{
    static bool initialized=false;
    if(initialized)
        return;
    initialized=true;

#ifdef USE_OPENSSL
    /* SSL preliminaries. We keep the certificate and key with the context. */

    SSL_library_init();                      /* initialize library */

    SSL_load_error_strings();

    BEGIN_DEBUG_TIME_MEASUREMENT(true,1,"Seeding OpenSSL PRNG");

    do {
        bool file_ok=false;
        FILE* f=fopen("openssl.rnd","rb");
        if(f) {
            string log="Seedeng OpenSSL PRNG with bytes from 'openssl.rnd'";
            log_ssl(log);

            file_ok=true;
            int my_errno;
            MY_LONG_LONG size=my_filesize(f,my_errno);
            fclose(f);
            enum {
                MAX_ENTROPY_SIZE_BYTES = 1024 * 1024,
            };
            if(size>MAX_ENTROPY_SIZE_BYTES)
                size=MAX_ENTROPY_SIZE_BYTES;
            if(size>0 && 0==my_errno)
                RAND_load_file("openssl.rnd",static_cast<long>(size));
            log_ssl("Done seeding");
        }
        if(!file_ok || !RAND_status()) {
            do {
                enum {
                    SEED_BYTES = 1024U,
                };
                char seed[SEED_BYTES];

                FILE* f;
                size_t size=0;
                bool okay=false;

                string log;

                if(!okay) {
                    log=((string)"Seedeng OpenSSL PRNG with ")+tos(static_cast<unsigned>(SEED_BYTES))+" bytes from /dev/urandom";
                    log_ssl(log);
                    f=fopen("/dev/urandom","rb");
                    if(f) {
                        while(size<sizeof(seed))
                            size+=fread(&(seed[size]),1,SEED_BYTES-size,f);
                        log_ssl("Done seeding");
                        fclose(f);
                        okay=true;
                    } else {
                        log_ssl("Not exists");
                    }
                }

                if(!okay) {
                    log=((string)"Seedeng OpenSSL PRNG with ")+tos(static_cast<unsigned>(SEED_BYTES))+" bytes from /dev/random";
                    log_ssl(log);
                    f=fopen("/dev/random","rb");
                    if(f) {
                        while(size<sizeof(seed))
                            size+=fread(&(seed[size]),1,sizeof(seed)-size,f);
                        log_ssl("Done seeding");
                        fclose(f);
                        okay=true;
                    } else {
                        log_ssl("Not exists");
                    }
                }

                if(!okay) {
                    log=((string)"Seedeng OpenSSL PRNG with ")+tos(sizeof(seed))+" bytes of 'rand() ^ rand()' expression";
                    log_ssl(log);
                    while(size<sizeof(seed))
                        seed[size++]=static_cast<char>(rand() ^ rand());
                    log_ssl("Done seeding");
                    okay=true;
                }

                RAND_seed(seed,SEED_BYTES);
            } while(!RAND_status());
            RAND_write_file("openssl.rnd");
        }
    } while(!RAND_status());

    END_DEBUG_TIME_MEASUREMENT(1,0);

    SSLeay_add_ssl_algorithms();
    OpenSSL_add_all_algorithms();
    meth=SSLv3_method();
    c_ssl::clear_error_queue();
#else

#if !defined(_WIN32) && !defined(_WIN64)
#   ifdef USE_PTHREAD
    gcry_control(GCRYCTL_SET_THREAD_CBS, &gcry_threads_pthread);
#   endif
#else
#   if defined(_MSC_VER) || defined(__BORLANDC__)
    gnutls_init_msvc_thread();
#else
    gnutls_init_threads();
#   endif
#endif
    gnutls_global_init();

    gcry_control(GCRYCTL_ENABLE_QUICK_RANDOM,0); // we want to use '/dev/urandom' rather than '/dev/random'

    ssl_loop();
#endif
}

string ssl_get_ssl_version()
{
#ifdef USE_OPENSSL
    return OPENSSL_VERSION_TEXT;
#else
    string res="GNUTLS ";
    res+=LIBGNUTLS_VERSION;
    return res;
#endif
}

#if defined(_WIN32) || defined(_WIN64)
#ifndef USE_OPENSSL
static struct gcry_thread_cbs gcry_threads_other;   //!< Descriptor for thread locking

/*!
    \brief Initialization of thread library
    \author VooDooMan
    \version 1
*/
static int thread_init()
{
    return 0;
}

/*!
    \brief Initialization of mutex
    \author VooDooMan
    \version 1
*/
static int thread_mutex_init(void **priv)
{
    static HANDLE h=NULL;
    if(NULL==h) {
        h=CreateMutex(NULL,FALSE,NULL);
    }
    *priv=(void*)h;
    return 0;
}

/*!
    \brief Closing of mutex
    \author VooDooMan
    \version 1
*/
static int thread_mutex_destroy(void **priv)
{
    HANDLE h=(HANDLE)*priv;
    CloseHandle(h);
    return 0;
}

/*!
    \brief Obtaining a mutex exclusion
    \author VooDooMan
    \version 1
*/
static int thread_mutex_lock(void **priv)
{
    HANDLE h=(HANDLE)*priv;
    DWORD rc=WaitForSingleObject(h,INFINITE);
    MY_ASSERT(WAIT_OBJECT_0==rc);
    return 0;
}

/*!
    \brief Releasing of mutex exclusion
    \author VooDooMan
    \version 1
*/
static int thread_mutex_unlock(void **priv)
{
    HANDLE h=(HANDLE)*priv;
    ReleaseMutex(h);
    return 0;
}

/*!
    \brief Introducing threads into gcrypto library
    \author VooDooMan
    \version 1
*/
void gnutls_init_msvc_thread()
{
    memset(&gcry_threads_other,0,sizeof(gcry_threads_other));
    gcry_threads_other.option=GCRY_THREAD_OPTION_USER;
    gcry_threads_other.init=thread_init;
    gcry_threads_other.mutex_init=thread_mutex_init;
    gcry_threads_other.mutex_destroy=thread_mutex_destroy;
    gcry_threads_other.mutex_lock=thread_mutex_lock;
    gcry_threads_other.mutex_unlock=thread_mutex_unlock;
    gcry_control (GCRYCTL_SET_THREAD_CBS, &gcry_threads_other);
}
#endif
#endif

/*!
    \brief Retrieves statistics about sock I/O cache
    \author VooDooMan
    \version 2
    \date 2006-01-22, 2006-07-11
    \param to_receive Returns number of bytes waiting in cache to receive them
    \param to_send Returns number of bytes waiting in cache to send them
*/
void c_ssl::get_cache_stats(size_t& to_receive, size_t& to_send)
{
    to_receive=to_send=0;

    deque<c_ssl_data>::iterator i1;

    for(i1=to_read.begin(); i1!=to_read.end(); ++i1)
        to_receive+=(*i1).len;

    for(i1=to_write.begin(); i1!=to_write.end(); ++i1)
        to_send+=(*i1).len;
}

void c_ssl::set_error(const string& err_string)
{
#ifdef USE_OPENSSL
    if(handle) {
        c_mutex_locker lock(extra_data_mutex);

        map<SSL*,pair<int,ssl_meta_data*> >::iterator i1=extra_data.find(handle);
        if(extra_data.end()!=i1) {
            MY_DELETE((*i1).second.second);
            extra_data.erase(i1);
        }
    }
#endif

    error=true;
    if(!err_str.empty())
        err_str+="; ";
    err_str+=err_string;
    if(linger==state)
        destroy();
}

#if defined(_MSC_VER)
#   pragma warning(push)
#   pragma warning(disable:4706)    // warning C4706: assignment within conditional expression
#endif

c_ssl::c_ssl(c_socket& socket, c_socket* server, const string& cert, const string& key, bool only_get_peers_cert, bool disable_ca_verify, bool ignore_expired_peers_cert, const string& expected_remote_peer_cert_file)
{
    clear_error_queue();

    error=false;

    s=MY_NEW c_socket(socket);    // TODO: FIXME: is this really needed?

    state=disconnected;

    m_only_get_peers_cert=only_get_peers_cert;
    m_disable_ca_verify=disable_ca_verify;
    m_ignore_expired_peers_cert=ignore_expired_peers_cert;
    set_expected_cert_file(expected_remote_peer_cert_file);

#ifdef USE_OPENSSL
    rbio=wbio=NULL;
    handle=NULL;

    ctx=SSL_CTX_new(meth);

    if (!ctx) {
        err_str="SSL_CTX_new(): ";
        err_str+=get_openssl_error();
        set_error(err_str);
        destroy();
        return;
    }

    if(!m_only_get_peers_cert) {
        string ca_file=ssl_conf.get_cas();

        STACK_OF(X509_NAME)* cert_names=SSL_load_client_CA_file(ca_file.c_str());
        if(!cert_names) {
            err_str="SSL_load_client_CA_file(): ";
            err_str+=get_openssl_error();
            err_str+=" (using file \"";
            err_str+=ca_file;
            err_str+="\")";
            set_error(err_str);
            destroy();
            return;
        }
        SSL_CTX_set_client_CA_list(ctx,cert_names);

        if(!SSL_CTX_load_verify_locations(ctx,ca_file.c_str(),NULL)) {
            err_str="SSL_CTX_load_verify_locations(): ";
            err_str+=get_openssl_error();
            set_error(err_str);
            destroy();
            return;
        }
    } else {
        SSL_CTX_set_client_CA_list(ctx,NULL);
    }

    if(!expected_remote_peer_cert_file.empty()) {
        SSL_CTX_set_verify(ctx, SSL_VERIFY_PEER|SSL_VERIFY_CLIENT_ONCE|SSL_VERIFY_FAIL_IF_NO_PEER_CERT,c_ssl::my_verify_callback);
    } else {
        SSL_CTX_set_verify(ctx, SSL_VERIFY_PEER|SSL_VERIFY_FAIL_IF_NO_PEER_CERT,c_ssl::my_verify_callback);
    }
    SSL_CTX_set_verify_depth(ctx,chain_verify_depth + 1);

    clear_error_queue();
#endif

    start_awaiting_handshake=0;

    cert_ok_so_far=true;
    force_write=false;

    num_transactions++;
    ssl_descriptors.push_back(this);

    if(server) {
#ifdef USE_OPENSSL
        handle = SSL_new (ctx);
        if(!handle) {
            set_error("SSL_new() failed");
            destroy();
            return;
        }

        //SSL_set_mode(handle,SSL_MODE_AUTO_RETRY);

        rbio=BIO_new(BIO_s_mem());
        wbio=BIO_new(BIO_s_mem());
        if(!rbio || !wbio) {
            if(rbio)
                BIO_free(rbio);
            if(wbio)
                BIO_free(wbio);
            rbio= wbio= NULL;
            set_error("BIO_new(BIO_s_mem()) failed");
            destroy();
            return;
        }

        int err;

        if ((err=SSL_use_certificate_file(handle, cert.c_str(), SSL_FILETYPE_PEM)) <= 0) {
            string err_str="SSL_use_certificate_file(): ";
            err_str+=get_openssl_error();
            set_error(err_str);
            destroy();
            return;
        }
        if ((err=SSL_use_RSAPrivateKey_file(handle, key.c_str(), SSL_FILETYPE_PEM)) <= 0) {
            string err_str="SSL_use_RSAPrivateKey_file(): ";
            err_str+=get_openssl_error();
            set_error(err_str);
            destroy();
            return;
        }

        // NEXT LINE: warning okay
        if (!(err=SSL_check_private_key(handle))) {
            string err_str="SSL_check_private_key(): Private key does not match the certificate public key; ";
            err_str+=get_openssl_error();
            set_error(err_str);
            destroy();
            return;
        }

        int mydata_index = SSL_get_ex_new_index(0,NULL , NULL, NULL, NULL);
        ssl_meta_data* mydata=MY_NEW ssl_meta_data(this);
        {
            c_mutex_locker lock(extra_data_mutex);

            extra_data.insert(make_pair(handle,make_pair(mydata_index,mydata)));
        }

        SSL_set_ex_data(handle, mydata_index, mydata);

        /*
        SSL_set_fd (handle, (int)s->handle);
        */
        SSL_set_bio(handle,rbio,wbio);

        err=SSL_accept(handle);
        if(process_openssl_error(handle,err)) {
            //clear_error();
            shutdown_not_accepting_data(); // client probably using SSLv2, so signal them shutdown
            return;
        }

        if(err<0)
            state=want_accept;
        else if(1==err) {
            state=verify_cert;
        } else {
            // err == 0 || err > 1
            process_openssl_error(handle,err);
            destroy();
            return;
        }

        return;
#else
        handle=NULL;

        /* X509 stuff */
        int err=gnutls_certificate_allocate_credentials (&(s->xcred));
        if(err<0) {
            set_error("gnutls_certificate_allocate_credentials() failed");
            if(s) {
                sock_linger_push(*s);
                MY_DELETE(s);
                s=NULL;
            }
            return;
        }

        string ca_file=ssl_conf.get_cas();

        /* sets the trusted cas file
        */
        err=gnutls_certificate_set_x509_trust_file (s->xcred,ca_file.c_str(), GNUTLS_X509_FMT_PEM);
        if(err<0) {
            set_error("gnutls_certificate_set_x509_trust_file() failed");
            if(s) {
                sock_linger_push(*s);
                MY_DELETE(s);
                s=NULL;
            }
            gnutls_certificate_free_credentials (s->xcred);
            return;
        }

        /*
        gnutls_certificate_set_x509_crl_file (x509_cred, CRLFILE,
                        GNUTLS_X509_FMT_PEM);
        */

        err=gnutls_certificate_set_x509_key_file (s->xcred, cert.c_str(), key.c_str(),
                        GNUTLS_X509_FMT_PEM);
        if(err<0) {
            set_error("gnutls_certificate_set_x509_key_file() failed");
            destroy();
            return;
        }

        gnutls_certificate_set_dh_params (s->xcred, dh_params);
        gnutls_certificate_set_rsa_export_params (s->xcred, rsa_params);

        gnutls_session session;

        gnutls_init(&session, GNUTLS_SERVER);

        /*
        const int cert_type_priority[10] = { GNUTLS_CRT_X509, 0 };
        const int cipher_priority[10] = {GNUTLS_CIPHER_AES_256_CBC, GNUTLS_CIPHER_AES_128_CBC, GNUTLS_CIPHER_3DES_CBC, 0};
        const int protocol_priority[10] = {GNUTLS_SSL3, GNUTLS_TLS1_1, GNUTLS_TLS1_0, 0};
        const int kx_priority[10] = {GNUTLS_KX_RSA, 0};
        const int mac_priority[10] = {GNUTLS_MAC_SHA, GNUTLS_MAC_RMD160, GNUTLS_MAC_MD5, 0};
        const int compression_priority[10] = {GNUTLS_COMP_NULL, 0};
        */

        /* Use the default priorities, plus, export cipher suites.
        */
        //gnutls_set_default_export_priority (session);
        gnutls_set_default_priority (session);

        s->m_gnutls_session=session;

        /*
        gnutls_protocol_set_priority(ssl,protocol_priority);
        gnutls_kx_set_priority(ssl,kx_priority);
        gnutls_mac_set_priority(ssl,mac_priority);
        gnutls_cipher_set_priority(ssl,cipher_priority);
        gnutls_certificate_type_set_priority (ssl, cert_type_priority);
        gnutls_compression_set_priority(ssl,compression_priority);
        */

        gnutls_credentials_set (session, GNUTLS_CRD_CERTIFICATE, s->xcred);

        /* request client certificate if any.
        */
        gnutls_certificate_server_set_request (session, GNUTLS_CERT_REQUEST);

        gnutls_dh_set_prime_bits (session, DH_BITS);

        //gnutls_transport_set_ptr(session,(gnutls_transport_ptr)client->handle);
        gnutls_transport_set_pull_function(session,pull_func);
        gnutls_transport_set_push_function(session,push_func);
        gnutls_transport_set_ptr(session,(gnutls_transport_ptr)s);

        s->got_xcred=true;

        /*
        int ret=gnutls_handshake(session);

        gnutls_certificate_free_credentials(x509_cred);

        if(ret<0) {
            gnutls_deinit(session);
            session=NULL;
            string log="Handshake failed: ";
            log+=gnutls_strerror(ret);
            log_ssl(log);
            return true;
        }
        */

        state=want_handshake;

        handle=session;

        return;
#endif
    } else {
#ifdef USE_OPENSSL
        handle = SSL_new (ctx);
        if(!handle) {
            set_error("SSL_new() failed");
            if(s) {
                sock_linger_push(*s);
                MY_DELETE(s);
                s=NULL;
            }
            handle=NULL;
            clear_error_queue();
            return;
        }

        //SSL_set_mode(handle,SSL_MODE_AUTO_RETRY);

        rbio=BIO_new(BIO_s_mem());
        wbio=BIO_new(BIO_s_mem());
        if(!rbio || !wbio) {
            if(rbio)
                BIO_free(rbio);
            if(wbio)
                BIO_free(wbio);
            rbio= wbio= NULL;
            set_error("BIO_new(BIO_s_mem()) failed");
            destroy();
            return;
        }

        if(!m_only_get_peers_cert) {
            if(!expected_remote_peer_cert_file.empty()) {
                SSL_set_verify(handle, SSL_VERIFY_PEER|SSL_VERIFY_CLIENT_ONCE|SSL_VERIFY_FAIL_IF_NO_PEER_CERT,my_verify_callback);
            } else {
                SSL_set_verify(handle, SSL_VERIFY_PEER|SSL_VERIFY_FAIL_IF_NO_PEER_CERT,my_verify_callback);
            }
        } else {
            SSL_set_verify(handle, SSL_VERIFY_FAIL_IF_NO_PEER_CERT,my_verify_callback);
        }

        SSL_set_verify_depth(handle,chain_verify_depth + 1);

        int err;

        if(!cert.empty() || !key.empty()) { // do NOT "&&", but "||" rather (what if ket is empty, but cert is not? == error!)
            if ((err=SSL_use_certificate_file(handle, cert.c_str(), SSL_FILETYPE_PEM)) <= 0) {
                string err_str="SSL_use_certificate_file(): ";
                err_str+=get_openssl_error();
                set_error(err_str);
                destroy();
                return;
            }
            if ((err=SSL_use_RSAPrivateKey_file(handle, key.c_str(), SSL_FILETYPE_PEM)) <= 0) {
                string err_str="SSL_use_RSAPrivateKey_file(): ";
                err_str+=get_openssl_error();
                set_error(err_str);
                destroy();
                return;
            }
            // NEXT LINE: warning okay
            if (!(err=SSL_check_private_key(handle))) {
                string err_str="SSL_check_private_key(): Private key does not match the certificate public key; ";
                err_str+=get_openssl_error();
                set_error(err_str);
                destroy();
                return;
            }
        }

        int mydata_index = SSL_get_ex_new_index(0,NULL , NULL, NULL, NULL);
        ssl_meta_data* mydata=MY_NEW ssl_meta_data(this);
        {
            c_mutex_locker lock(extra_data_mutex);

            extra_data.insert(make_pair(handle,make_pair(mydata_index,mydata)));
        }

        SSL_set_ex_data(handle, mydata_index, mydata);

        /*
        SSL_set_fd (handle, (int)s->handle);
        */
        SSL_set_bio(handle,rbio,wbio);

        err=SSL_connect(handle);
        if(process_openssl_error(handle,err)) {
            destroy();
            return;
        }

        if(err<0)
            state=want_handshake;
        else if(1==err) {
            state=verify_cert;
        } else {
            // err == 0 || err > 1
            process_openssl_error(handle,err);
            destroy();
            return;
        }

        return;
#else
        handle=NULL;

        gnutls_certificate_credentials xcred;

        /* X509 stuff */
        int err=gnutls_certificate_allocate_credentials (&xcred);
        if(err<0) {
            set_error("gnutls_certificate_allocate_credentials() failed");
            destroy();
            return;
        }

        if(!m_only_get_peers_cert) {
            string ca_file=ssl_conf.get_cas();

            /* sets the trusted cas file
            */
            err=gnutls_certificate_set_x509_trust_file (xcred,ca_file.c_str(), GNUTLS_X509_FMT_PEM);
            if(err<0) {
                set_error("gnutls_certificate_set_x509_trust_file() failed");
                destroy();
                gnutls_certificate_free_credentials(xcred);
                clear_error_queue();
                return;
            }
        }

        /* Initialize TLS session
        */
        gnutls_init(&handle, GNUTLS_CLIENT);

        if(!cert.empty() || !key.empty()) { // do NOT "&&", but "||" rather (what if key is empty, but cert is not? == error!)
            err=gnutls_certificate_set_x509_key_file(xcred,cert.c_str(), key.c_str(), GNUTLS_X509_FMT_PEM);
            if(err<0) {
                set_error("gnutls_certificate_set_x509_key_file() failed");
                gnutls_certificate_free_credentials (xcred);
                destroy();
                return;
            }
        }

        gnutls_certificate_set_dh_params(xcred, dh_params);
        gnutls_certificate_set_rsa_export_params(xcred, rsa_params);

        /* put the x509 credentials to the current session
        */
        gnutls_credentials_set(handle, GNUTLS_CRD_CERTIFICATE, xcred);

        static const int cert_type_priority[10] = { GNUTLS_CRT_X509, 0 };
        static const int cipher_priority[10] = {GNUTLS_CIPHER_AES_256_CBC, GNUTLS_CIPHER_AES_128_CBC, GNUTLS_CIPHER_3DES_CBC, 0};
        static const int protocol_priority[10] = {GNUTLS_SSL3, GNUTLS_TLS1_2, GNUTLS_TLS1_1, GNUTLS_TLS1_0, 0};
        static const int kx_priority[10] = {GNUTLS_KX_RSA, GNUTLS_KX_DHE_DSS, GNUTLS_KX_DHE_PSK, 0};
        static const int mac_priority[10] = {GNUTLS_MAC_SHA, GNUTLS_MAC_RMD160, GNUTLS_MAC_MD5, GNUTLS_MAC_MD2, 0};
        static const int compression_priority[10] = {GNUTLS_COMP_NULL, GNUTLS_COMP_ZLIB, GNUTLS_COMP_DEFLATE, GNUTLS_COMP_LZO, 0};

        /* Use default priorities */
        //gnutls_set_default_priority(handle);

        /**/
        gnutls_protocol_set_priority(handle,protocol_priority);
        gnutls_kx_set_priority(handle,kx_priority);
        gnutls_mac_set_priority(handle,mac_priority);
        gnutls_cipher_set_priority(handle,cipher_priority);
        gnutls_certificate_type_set_priority (handle, cert_type_priority);
        gnutls_compression_set_priority(handle,compression_priority);
        /**/

        /* connect to the peer
        */
        //gnutls_transport_set_ptr(ssl,(gnutls_transport_ptr)server->handle);

        s->m_gnutls_session=handle;

        socket.m_gnutls_session=handle;

        socket.got_xcred=true;
        socket.xcred=xcred;

        s->got_xcred=true;
        s->xcred=xcred;

        gnutls_transport_set_pull_function(handle,pull_func);
        gnutls_transport_set_push_function(handle,push_func);
        gnutls_transport_set_ptr(handle,(gnutls_transport_ptr)s);

        state=want_handshake;

        return;
#endif
    }
    // NOTREACH
}

#if defined(_MSC_VER)
#   pragma warning(pop)
#endif

size_t c_ssl::read(void* buff, size_t len)
{
    if(state!=disconnected) { // to allow read something from the read queue, even if disconnected
        if(state!=session_ok)
            return 0;
    }
    if(to_read.empty())
        return 0;

    deque<c_ssl_data>::iterator i1=to_read.begin();
    if((*i1).len<=len) {
        memcpy(buff,(*i1).data,len=(*i1).len);
        my_free((*i1).data);
        to_read.pop_front();
        return len;
    }
    // resize
    c_ssl_data n;
    n.len=(*i1).len-len;
    n.data=my_malloc(n.len);
    if(!n.data)
        return 0;
    memcpy(n.data,&(((char*)((*i1).data))[len]),n.len);
    memcpy(buff,(*i1).data,len);
    my_free((*i1).data);
    to_read.pop_front();
    to_read.push_front(n);
    return len;
}

void c_ssl::write(const void* buff, size_t len)
{
    if(!len)
        return;
    c_ssl_data n;
    n.len=len;

    char buf[256];

    if(to_write.size()>=2) {
        deque<c_ssl_data>::reverse_iterator i1=to_write.rbegin();
        if((*i1).len+len<1024) {
            void* b=my_realloc((*i1).data,(*i1).len+len);
            if(b) {
                (*i1).data=b;
                memcpy(&((char*)b)[(*i1).len],buff,len);
                (*i1).len+=len;
                return;
            }
        }
    }

    n.data=my_malloc(n.len);
    if(!n.data) {
        sprintf(buf,"malloc(%s) failed, out of memory",tos(n.len).c_str());
        set_error(buf);
        destroy();
        return;
    }

    memcpy(n.data,buff,n.len);

    to_write.push_back(n);
}

bool c_ssl::get_error(string& str)
{
    if(!to_read.empty()) {
        // to allow read something from the read queue, even if disconnected
        str="";
        return false;
    }
    if(shutdown_reject_data==state || linger==state)
        return false;
    if(error)
        str=err_str;
    else
        str="";
    return error;
}

void c_ssl::set_expected_cert_file(const string& cert_file)
{
    expected_cert_file=cert_file;
    if(session_ok==state)
        state=verify_cert;
}

c_ssl::~c_ssl()
{
    clear_error_queue();

    destroy();

    while(!to_read.empty()) {
        my_free((*(to_read.begin())).data);
        to_read.pop_front();
    }
    while(!to_write.empty()) {
        my_free((*(to_write.begin())).data);
        to_write.pop_front();
    }

    vector<c_ssl*>::iterator i1;
    for(i1=ssl_descriptors.begin(); i1!=ssl_descriptors.end(); ++i1) {
        if(this==*i1) {
            ssl_descriptors.erase(i1);
            break;
        }
    }

    if(s)
        MY_DELETE(s);
}

void c_ssl::process()
{
    clear_error_queue();

    if(!handle) {
        destroy();
        return;
    }

    if(!s) {
        destroy();
        return;
    }

#ifdef USE_OPENSSL
    if(rbio && wbio && s) {
        char buff[1024*16];

        try {
            if(rbio && wbio && s) {
                // send data to socket

                errno=0;
                int ret=BIO_read(wbio,buff,sizeof(buff));
                if(ret!=-1 && process_openssl_error(handle,ret)) {
                    destroy();
                    throw (int)1;
                }
                if(ret>0) {
                    int ec=0;
                    int sent=static_cast<int>(sock_send(*s,buff,ret,ec));
                    if(ec || sent!=ret) {
                        destroy();

                        if(ec) {
                            char* errstr=sock_error(ec);
                            log_socket(ec,errstr,GET_STR_DEBUG_INFO_EX(errstr),s);
                            set_error(errstr);
                            MY_DELETE_A(errstr);
                        } else {
                            // closed
                            log_socket(0,"SSL socket EOF",GET_STR_DEBUG_INFO_EX("EOF"),s);
                            set_error("SSL socket EOF");
                        }
                        throw (int)1;
                    }
                } else if(0==ret || -1==ret) {
                    errno=0;
                    bool retry=TO_BOOL(BIO_should_retry(wbio));
                    if(!retry) {
                        process_openssl_error(handle,ret);
                        destroy();
                        throw (int)1;
                    }
                } else {
                    destroy();
                    throw (int)1;
                }
            }

            if(rbio && wbio && s) {
                // receive data from socket

                int ec=0;
                bool closed=false;
                int len=static_cast<int>(sock_read(*s,buff,sizeof(buff),ec,closed));
                if(len && !ec && !closed) {
                    errno=0;
                    int ret=BIO_write(rbio,buff,static_cast<int>(len));
                    if(ret!=-1 && process_openssl_error(handle,ret)) {
                        destroy();
                        throw (int)1;
                    }
                    if(ret>0 && ret!=len) {
                        sock_unread(*s,&(buff[ret]),len-ret);
                    } else if(ret==len) {
                        // okay
                    } else if(0==ret || -1==ret) {
                        errno=0;
                        if(!TO_BOOL(BIO_should_retry(rbio))) {
                            process_openssl_error(handle,ret);
                            destroy();
                            throw (int)1;
                        } else {
                            sock_unread(*s,buff,len);
                        }
                    } else {
                        process_openssl_error(handle,ret);
                        destroy();
                        throw (int)1;
                    }
                } else {
                    if(ec || closed) {
                        destroy();

                        if(ec) {
                            char* errstr=sock_error(ec);
                            log_socket(ec,errstr,GET_STR_DEBUG_INFO_EX(errstr),s);
                            set_error(errstr);
                            MY_DELETE_A(errstr);
                        } else {
                            // closed
                            log_socket(0,"SSL socket EOF",GET_STR_DEBUG_INFO_EX("EOF"),s);
                            set_error("SSL socket EOF");
                        }
                        throw (int)1;
                    }
                }
            }
        } catch(int UNUSEDARG(e)) {
            // nevermind
        }
    }
#endif

    switch(state) {
        case disconnected:
            return;
        case want_handshake:
            do_handshake();
            return;
        case want_accept:
            do_accept();
            return;
        case verify_cert:
            do_verify();
            return;
        case session_ok:
            do_transfer();
            return;
        case in_shutdown:
            {
                size_t in, out;
                get_cache_stats(in,out);
                if(in || out) {
                    do_transfer();
                    return;
                }
                do_shutdown();
                return;
            }
        case shutdown_reject_data:
            {
                do_shutdown();
                return;
            }
        case linger:
            {
                do_shutdown();
                return;
            }

    }
    // huh?!
    destroy();
}

void c_ssl::do_handshake()
{
    clear_error_queue();

#ifdef USE_OPENSSL
    int err=SSL_connect(handle);
    if(process_openssl_error(handle,err)) {
        destroy();
        return;
    }

    int err2=SSL_get_error(handle,err);

    if(err<0 && err2!=1)
        state=want_handshake;
    else if(1==err) {
        state=verify_cert;
    } else {
        // err == 0 || err > 1 || fatal || (err==-1 && err2==1)
        process_openssl_error(handle,err);
        destroy();
        return;
    }
    return;
#else
    int err=gnutls_handshake(handle);
    if(GNUTLS_E_INTERRUPTED==err || GNUTLS_E_AGAIN==err) {
        return;
    }
    if(err!=0 && gnutls_error_is_fatal(err)) {
        set_error(my_ERR_error_string(err));
        destroy();
        return;
    }
    if(0==err)
        state=verify_cert;
    return;
#endif
    // NOTREACH
}

void c_ssl::do_accept()
{
    clear_error_queue();

#ifdef USE_OPENSSL
    if(0==start_awaiting_handshake)
        my_time(&start_awaiting_handshake);

    int err;

    err=SSL_accept(handle);
    if(process_openssl_error(handle,err)) {
        //clear_error();
        shutdown_not_accepting_data(); // client probably using SSLv2, so signal them shutdown
        return;
    }

    if(err<0)
        state=want_accept;
    else if(1==err) {
        state=verify_cert;
    } else {
        // err == 0 || err > 1
        process_openssl_error(handle,err);
        destroy();
        return;
    }
    return;
#else
    int err=gnutls_handshake(handle);
    if(GNUTLS_E_INTERRUPTED==err || GNUTLS_E_AGAIN==err) {
        return;
    }
    if(err!=0 && gnutls_error_is_fatal(err)) {
        set_error(my_ERR_error_string(err));
        destroy();
        return;
    }
    if(0==err)
        state=verify_cert;
    return;
#endif
    // NOTREACH
}

void c_ssl::do_transfer()
{
    clear_error_queue();

    char buf[1024*10];
    int max_size=1024*10;
    int err;

    if(force_write)
        goto do_write;

    // read

#ifdef USE_OPENSSL
    err=(int)SSL_read(handle,buf,max_size);
    if(process_openssl_error(handle,err)) {
        destroy();
        return;
    }
    if(err>0) {
        c_ssl_data n;
        n.len=(size_t)err;
        n.data=my_malloc(n.len);
        if(!n.data) {
            sprintf(buf,"malloc(%s) failed, out of memory",tos(n.len).c_str());
            set_error(buf);
            destroy();
            return;
        }
        memcpy(n.data,buf,n.len);
        to_read.push_back(n);
    }
    if(err<0) {
        // ok, no problem
    }
    if(0==err) {
        process_openssl_error(handle,err);
        destroy();
        return;
    }
#else
    err=gnutls_record_recv(handle,buf,max_size);
    if(err<0 && gnutls_error_is_fatal(err)) {
        set_error(my_ERR_error_string(err));
        destroy();
        return;
    } else if(err<0) {
        if(GNUTLS_E_INTERRUPTED==err || GNUTLS_E_AGAIN==err) {
            // force_read
        }
        goto do_write;
    } else if(0==err) {
        set_error("EOF");
        destroy();
        return;
    }
    // err > 0
    {
        c_ssl_data n;
        n.len=(size_t)err;
        n.data=my_malloc(n.len);
        if(!n.data) {
            sprintf(buf,"malloc(%s) failed, out of memory",tos(n.len).c_str());
            set_error(buf);
            destroy();
            return;
        }
        memcpy(n.data,buf,n.len);
        to_read.push_back(n);
    }
#endif

do_write:
    // write
    clear_error_queue();
    force_write=false;
    if(to_write.empty())
        return;
    void* w=(*(to_write.begin())).data;
    size_t len=(*(to_write.begin())).len;
    if(!len) {
        to_write.pop_front();
        goto do_write;
    }

#ifdef USE_OPENSSL
    err=(int)SSL_write(handle,w,static_cast<int>(len));
    if(process_openssl_error(handle,err)) {
        destroy();
        return;
    }
    if(err>0) {
        if((signed)len==err) {
            my_free(w);
            to_write.pop_front();
            return;
        }
        c_ssl_data n;
        n.len=((signed)len)-err;
        n.data=my_malloc(n.len);
        if(!n.data) {
            sprintf(buf,"malloc(%s) failed, out of memory",tos(n.len).c_str());
            set_error(buf);
            destroy();
            return;
        }
        memcpy(n.data,&(((char*)w)[err]),n.len);
        my_free(w);
        to_write.pop_front();
        to_write.push_front(n);
        return;
    }
    if(err<0) {
        // okay, no problem
        force_write=true;
        return;
    }
    if(0==err) {
        process_openssl_error(handle,err);
        destroy();
        return;
    }
    return;
#else
    err=gnutls_record_send(handle,w,len);
    if(err<0 && gnutls_error_is_fatal(err)) {
        set_error(my_ERR_error_string(err));
        destroy();
        return;
    } else if(err<0 || 0==err) {
        if(GNUTLS_E_INTERRUPTED==err || GNUTLS_E_AGAIN==err) {
            force_write=true;
        }
        // okay, no problem
        return;
    }
    // err > 0
    {
        if((signed)len==err) {
            my_free(w);
            to_write.pop_front();
            return;
        }
        c_ssl_data n;
        n.len=((signed)len)-err;
        n.data=my_malloc(n.len);
        if(!n.data) {
            sprintf(buf,"malloc(%s) failed, out of memory",tos(n.len).c_str());
            set_error(buf);
            destroy();
            return;
        }
        memcpy(n.data,&(((char*)w)[err]),n.len);
        my_free(w);
        to_write.pop_front();
        to_write.push_front(n);
    }
    return;
#endif
    // NOTREACH
}

void c_ssl::do_verify()
{
    clear_error_queue();

    if(!cert_ok_so_far) {
        set_error("Certificate verification failed");
        destroy();
        return;
    }
#ifdef USE_OPENSSL
    bool ok=true;
    X509* client_cert=SSL_get_peer_certificate(handle);
    if(NULL!=client_cert) {
        {
            BIO* w=BIO_new(BIO_s_mem());
            if(NULL==w) {
                cert_ok_so_far=false;
                set_error("BIO_new(BIO_s_mem()) failed");
                X509_free(client_cert);
                destroy();
                return;
            }

            PEM_write_bio_X509(w,client_cert);

            enum {
                BUF_LEN = 1024U * 8U,
                BUF_SPARE = 16U,

                BUF_REAL = BUF_LEN + BUF_SPARE,
            };
            char* buff=MY_NEW char[BUF_REAL];
            memset(buff,0,BUF_REAL);
            if(BIO_read(w,buff,BUF_LEN)<=0) {
                MY_DELETE_A(buff);

                cert_ok_so_far=false;
                set_error("BIO_read() failed");
                X509_free(client_cert);
                destroy();
                return;
            }

            {
                c_mutex_locker lock(extra_data_mutex);

                remote_pkey=buff;
            }

            MY_DELETE_A(buff);
        }

        /*
        int rc=SSL_get_verify_result(handle);
        if(m_only_get_peers_cert &&
            (
                X509_V_ERR_UNABLE_TO_GET_ISSUER_CERT_LOCALLY==rc ||
                X509_V_ERR_UNABLE_TO_VERIFY_LEAF_SIGNATURE==rc
            )
            )
        {
            rc=X509_V_OK;
        }
        */

        /*if(X509_V_OK==rc)*/
        {
            // The client sent a certificate which verified OK
            if(!expected_cert_file.empty()) {
                X509* ecc=ssl_read_cert(const_cast<char*>(expected_cert_file.c_str()));
                if(ecc==NULL) {
                    ok=false;
                } else {
                    X509_check_purpose(ecc,-1,0); // compute SHA1 hash
                    char* hash1=(char*)&(ecc->sha1_hash);
                    char* hash2=(char*)&(client_cert->sha1_hash);
                    ok=true;
                    for(size_t i1=0; i1<sizeof(client_cert->sha1_hash); ++i1, ++hash1, ++hash2) {
                        if(*hash1!=*hash2) {
                            ok=false;
                            break;
                        }
                    }
                    X509_free(ecc);
                }
            } else
                ok=true;
            if(ok) {
                state=session_ok;
                X509_free(client_cert);
            } else {
                cert_ok_so_far=false;
                set_error("Certificate verification failed: SHA1 hash doesn't match");
                X509_free (client_cert);
                destroy();
                return;
            }
        }/* else {
            cert_ok_so_far=false;
            set_error("Certificate verification failed");
            X509_free(client_cert);
            destroy();
            return;
        }*/
    } else {
        cert_ok_so_far=false;
        set_error("Certificate verification failed - remote peer does not use certificate");
        destroy();
        return;
    }
    if(!cert_ok_so_far) {
        destroy();
        return;
    }
    state=session_ok;
    return;
#else   // -> !USE_OPENSSL
    string remote_cert;
    string res=verify_certificate(handle,"",expected_cert_file.c_str(),remote_cert,m_only_get_peers_cert,m_disable_ca_verify,m_ignore_expired_peers_cert);
    if(!res.empty()) {
        cert_ok_so_far=false;
        string err="Certificate verification failed: ";
        err+=res;
        set_error(err);
        destroy();
        return;
    }
    if(!cert_ok_so_far) {
        destroy();
        return;
    }

    {
        c_mutex_locker lock(extra_data_mutex);

        remote_pkey=remote_cert;
    }

    state=session_ok;
    return;
#endif

    // NOTREACH
}

void ssl_loop(bool force_garbaging)
{
    static bool in=false;
    if(in)
        return;
    in=true;

    {
        vector<c_ssl*>::iterator i1;
        for(i1=ssl_descriptors.begin(); i1!=ssl_descriptors.end(); ++i1) {
            c_ssl::clear_error_queue();
            (*i1)->process();
            c_ssl::clear_error_queue();
        }
    }
    {
        c_ssl::clear_error_queue();
        my_time_t now=my_time(NULL);
        vector<pair<c_ssl*,my_time_t> >::iterator i1;
again_a:
        for(i1=ssl_garbage.begin(); i1!=ssl_garbage.end(); ++i1) {
            if(force_garbaging || (*i1).second+10<now) {
                c_ssl::clear_error_queue();
                (*i1).first->shutdown();
            }
            if(force_garbaging) {
                for(unsigned i=0; i<10; ++i) {
                    sock_send_cache();
                    my_sleep(10);
                }
            }
            if(force_garbaging || (*i1).second+60<now) {
                c_ssl::clear_error_queue();
                MY_DELETE((*i1).first);
                ssl_garbage.erase(i1);
                c_ssl::clear_error_queue();
                goto again_a;
            }
        }
        c_ssl::clear_error_queue();
    }

    static my_time_t last_dh_rsa_gen=0;
    if(0==last_dh_rsa_gen) {
#ifndef USE_OPENSSL
        extern bool log_suppress_any_logs;

        string log="Generating DH and RSA parameters... (";
        log+=tos((unsigned)DH_BITS);
        log+=" DH bits, ";
        log+=tos((unsigned)RSA_BITS);
        log+=" RSA bits)";
        log_ssl(log,true);
        if(!log_suppress_any_logs) {
            cout << log << endl;
            cout.flush();
        }
        my_time_t start=my_time(NULL);

        gnutls_dh_params_init(&dh_params);
        gnutls_rsa_params_init(&rsa_params);
        /* Generate Diffie Hellman parameters - for use with DHE
         * kx algorithms. These should be discarded and regenerated
         * once a day, once a week or once a month. Depends on the
         * security requirements.
         */
        gnutls_dh_params_generate2(dh_params, DH_BITS);
        /* Generate RSA parameters - for use with RSA-export
         * cipher suites. These should be discarded and regenerated
         * once a day, once every 500 transactions etc. Depends on the
         * security requirements.
         */
        gnutls_rsa_params_generate2(rsa_params, RSA_BITS);
        my_time_t stop=my_time(NULL);
        log="Generation of DH and RSA parameters took ";
        log+=tos((long)my_difftime(stop,start));
        log+=" seconds";
        log_ssl(log);
        if(!log_suppress_any_logs) {
            cout << log << endl;
            cout.flush();
        }
#endif
        num_transactions=0;
        last_dh_rsa_gen=my_time(NULL);
    }
#ifndef USE_OPENSSL
    static my_time_t thread_start=0;

#if defined(HAVE_LIBPTHREAD)
    static pthread_t thread;
#endif

    if(last_dh_rsa_gen+DH_RSA_REGENERATION_NUM_SECS<my_time(NULL) || num_transactions>=DH_RSA_REGENERATION_NUM_TRANSACTIONS) {
        num_transactions=0;
        last_dh_rsa_gen=my_time(NULL);

        if(!cs_init) {
            thread_start=my_time(NULL);
            string log;
#if defined(HAVE_LIBPTHREAD) || defined(_WIN32) || defined(_WIN64)
            log="Regenerating DH and RSA parameters (in separate thread)... (";
#else
            log="Regenerating DH and RSA parameters... (";
#endif
            log+=tos((unsigned)DH_BITS);
            log+=" DH bits, ";
            log+=tos((unsigned)RSA_BITS);
            log+=" RSA bits)";
            log_ssl(log);
#if defined(_WIN32) || defined(_WIN64)
            InitializeCriticalSection(&cs);
            cs_init=true;
            if(cs_init) {
                ssl_dh_thread_running=true;
                uintptr_t res=_beginthread(ssl_dh_thread,0,(void*)NULL);
                if(-1L==res) {
                    cs_init=false;
                    ssl_dh_thread_running=false;
                }
            }
#else
#   ifdef HAVE_LIBPTHREAD
            if(pthread_mutex_init(&mutex,NULL)) {
                cs_init=false;
                ssl_dh_thread_running=false;
            } else {
                cs_init=true;
                int rc=pthread_create(&thread,NULL,ssl_dh_thread,NULL);
                if(rc) {
                    cs_init=false;
                    ssl_dh_thread_running=false;
                } else
                    ssl_dh_thread_running=true;
            }
#   else
            ssl_dh_thread(NULL);
            ssl_dh_thread_running=false;
            cs_init=true;
#endif
#endif
        }
    }
    if(cs_init) {
#if defined(_WIN32) || defined(_WIN64)
        EnterCriticalSection(&cs);
#else
#   ifdef HAVE_LIBPTHREAD
        pthread_mutex_lock(&mutex);
#   endif
#endif
        bool thread_running=ssl_dh_thread_running;
#if defined(_WIN32) || defined(_WIN64)
        LeaveCriticalSection(&cs);
#else
#   ifdef HAVE_LIBPTHREAD
        pthread_mutex_unlock(&mutex);
#   endif
#endif
        if(!thread_running) {
            my_time_t stop=my_time(NULL);
            string log;
#if defined(HAVE_LIBPTHREAD) || defined(_WIN32) || defined(_WIN64)
            log="Regeneration of DH and RSA parameters (in separate thread) took ";
#else
            log="Regeneration of DH and RSA parameters took ";
#endif
            log+=tos((long)my_difftime(stop,thread_start));
            log+=" seconds";
            log_ssl(log);

#       if defined(_WIN32) || defined(_WIN64)
            DeleteCriticalSection(&cs);
#       else
#         ifdef HAVE_LIBPTHREAD
            pthread_mutex_destroy(&mutex);
            void* code=NULL;
            pthread_join(thread,&code);
#         endif
#       endif

            cs_init=false;

            gnutls_dh_params_deinit(dh_params);
            gnutls_rsa_params_deinit(rsa_params);
            dh_params=thread_dh_params;
            rsa_params=thread_rsa_params;

#ifndef XTELNET
            irc_last_keepalive_detection=my_time(NULL)-IRC_KEEPALIVE_PERIOD;    // do not check for timeout now - this operation could have been slow
            irc_keepalive_sent=false;
#endif
        }
    }
#endif

#ifdef USE_OPENSSL
    static my_time_t last_rnd_save=0;
    if(last_rnd_save+60<my_time(NULL)) {
        last_rnd_save=my_time(NULL);
        RAND_write_file("openssl.rnd");
        c_ssl::clear_error_queue();
    }
#endif

    in=false;
}

#if defined(_MSC_VER)
#   pragma warning(push)
#   pragma warning(disable:4706)    // warning C4706: assignment within conditional expression
#endif

string c_ssl::get_openssl_error()
{
#ifdef USE_OPENSSL
    char buff[1024];
    string result;
    unsigned long err;
    // NEXT LINE: warning okay
    while((err=ERR_get_error())) {
        if(!result.empty())
            result+="; ";
        /*
        ERR_error_string(err,buff);
        result+=buff;
        if(ERR_lib_error_string(err)) {
            result+=", ";
            result+=ERR_lib_error_string(err);
        }
        if(ERR_func_error_string(err)) {
            result+=", ";
            result+=ERR_func_error_string(err);
        }
        if(ERR_reason_error_string(err)) {
            result+=", ";
            result+=ERR_reason_error_string(err);
        }
        */
        *buff=0;
        ERR_error_string_n(err,buff,sizeof(buff)-1);
        buff[sizeof(buff)-1]=0;
        if(*buff)
            result+=buff;
    }
    clear_error_queue();
    return result;
#else
    return (string)"";
#endif
}

#if defined(_MSC_VER)
#   pragma warning(pop)
#endif

#if defined(_MSC_VER)
#   pragma warning(push)
#   pragma warning(disable:4127)    // warning C4127: conditional expression is constant
#   pragma warning(disable:4706)    // warning C4706: assignment within conditional expression
#endif

bool c_ssl::process_openssl_error(ssl_handle GNUTLS_UNUSEDARG(handle), int GNUTLS_UNUSEDARG(rc))
{
#ifdef USE_OPENSSL
    if(rc>=0) {
        if(shutdown_reject_data==state || linger==state)
            return false;
        return error;
    }

    int err=SSL_get_error(handle,rc);
    switch(err) {
        case SSL_ERROR_NONE:
        case SSL_ERROR_WANT_READ:
        case SSL_ERROR_WANT_WRITE:
            return error;
    }

    string result;
    char buff[1024];
    unsigned long err2;
    // NEXT LINE: warning okay
    while((err2=ERR_get_error())) {
        if(!result.empty())
            result+="; ";
        *buff=0;
        ERR_error_string_n(err2,buff,sizeof(buff)-1);
        buff[sizeof(buff)-1]=0;
        if(*buff)
            result+=buff;
    }
    clear_error_queue();

    if(!err_str.empty())
        err_str+="; ";
    err_str+=result;
    error=true;
#endif

    return error;
}

#if defined(_MSC_VER)
#   pragma warning(pop)
#endif

void c_ssl::clear_error_queue()
{
#ifdef USE_OPENSSL
    ERR_clear_error();
#endif
}

void c_ssl::shutdown()
{
    if(disconnected==state || shutdown_reject_data==state)
        return;
    state=in_shutdown;
}

void c_ssl::do_shutdown()
{
    clear_error_queue();

#ifdef USE_OPENSSL
    int err=SSL_shutdown(handle);
    if(process_openssl_error(handle,err)) {
        destroy();
        return;
    }

    if(0==err) {
        // okay, no problem, we will call it again later
    } else if(1==err) {
        // ok
        state=linger;
        return;
    } else if(-1==err) {
        // error: nevermind
        destroy();
        return;
    } else {
        // huh? nevermind
        destroy();
        return;
    }
    return;
#else
    int err=gnutls_bye(handle,GNUTLS_SHUT_RDWR);
    if(GNUTLS_E_INTERRUPTED==err || GNUTLS_E_AGAIN==err) {
        return;
    }
    {
        destroy();
        return;
    }
    return;
#endif
    // NOTREACH
}

bool c_ssl::is_open()
{
    if(!to_read.empty()) {
        // to allow read something from the read queue, even if disconnected
        return true;
    }
    return state!=disconnected;
}

/*!
    \brief Closes all descriptors
    \version 2
    \date 2006-07-16, 2007-08-21
    \author VooDooMan
*/
void ssl_done()
{
    vector<c_ssl*>::iterator i1;
    while((i1=ssl_descriptors.begin())!=ssl_descriptors.end())
        MY_DELETE(*i1);
    ssl_garbage.clear();
}

/*!
    \brief Garbage collector
    \version 2
    \date 2006-07-17, 2008-06-23
    \author VooDooMan
    \param ssl Descriptor
*/
void ssl_put_to_garbage(c_ssl* ssl)
{
    if(!ssl)
        return;
    ssl_garbage.push_back(make_pair(ssl,my_time(NULL)));
}

void c_ssl::destroy()
{
#ifdef USE_OPENSSL
    if(handle) {
        c_mutex_locker lock(extra_data_mutex);

        map<SSL*,pair<int,ssl_meta_data*> >::iterator i1=extra_data.find(handle);
        if(extra_data.end()!=i1) {
            MY_DELETE((*i1).second.second);
            extra_data.erase(i1);
        }
    }
#endif

    if(s) {
        sock_linger_push(*s);
        MY_DELETE(s);
        s=NULL;
    }

    if(shutdown_reject_data==state)
        return;
#ifdef USE_OPENSSL

    rbio=NULL;
    wbio=NULL;

    if(handle) {
        SSL_free(handle);
        clear_error_queue();
    }
    handle=NULL;

    if(ctx) {
        SSL_CTX_free(ctx);
        clear_error_queue();
    }
    ctx=NULL;

    state=disconnected;

    clear_error_queue();
#else
    if(handle)
        gnutls_deinit(handle);
    handle=NULL;
    if(s && s->xcred) {
        gnutls_certificate_free_credentials(s->xcred);
        s->xcred=NULL;
    }
    if(s) {
        sock_linger_push(*s);
        s=NULL;
    }

    state=disconnected;
    clear_error_queue();
#endif
}

void c_ssl::clear_send_queue()
{
    deque<c_ssl_data>::iterator i1;
    while((i1=to_write.begin())!=to_write.end()) {
        if((*i1).data)
            my_free((*i1).data);
        (*i1).data=NULL;
        to_write.pop_front();
    }
}

c_socket* c_ssl::get_socket()
{
    return s;
}

string c_ssl::get_remote_pkey()
{
    c_mutex_locker lock(extra_data_mutex);

    return remote_pkey;
}

void c_ssl::shutdown_not_accepting_data()
{
    if(disconnected==state)
        return;
    state=shutdown_reject_data;
}

bool c_ssl_conf::rehash()
{
    c_mutex_locker lock(m_mutex);

    FILE* f=fopen("ssl.txt","r");
    if(f==NULL)
        return true;

    m_accept_ca_certs="";
    m_server_cert="";
    m_server_key="";
    m_client_cert="";
    m_client_key="";
    m_ssl_bots.clear();

    c_ssl_bot bot;
    bot.botname="";
    bot.cert="";
    bot.key="";

    bool in_general=false;
    string in_bot="";

    char ln[10240+1];
    while(!feof(f)) {
        if(fgets(ln,1024*10-1,f)==NULL)
            break;
        ln[10240]=0;
        if(ln[0] && ln[strlen(ln)-1]=='\r')
            ln[strlen(ln)-1]=0;
        if(ln[0] && ln[strlen(ln)-1]=='\n')
            ln[strlen(ln)-1]=0;
        if(ln[0] && ln[strlen(ln)-1]=='\r')
            ln[strlen(ln)-1]=0;
        if(ln[0] && ln[strlen(ln)-1]=='\n')
            ln[strlen(ln)-1]=0;
        if(ln[0] && ln[0]=='[') {
            if(!strcmp(ln,"[general]")) {
                in_general=true;
                in_bot="";
                bot.botname="";
                bot.cert="";
                bot.key="";
                continue;
            }
            if(!in_bot.empty()) {
                m_ssl_bots.push_back(bot);
                bot.botname="";
                bot.cert="";
                bot.key="";
            }
            in_bot=ln;
            if(in_bot.length() && in_bot[0]=='[')
                in_bot.erase(0,1);
            if(in_bot.length() && in_bot[in_bot.length()-1]==']')
                in_bot.erase(in_bot.length()-1,1);
            bot.botname=in_bot;
            in_general=false;
            continue;
        }
        string id_, val_;
        int pos=0;
        for(unsigned int i1=0; i1<strlen(ln); ++i1) {
            if(pos==0 && ln[i1]=='=') {
                pos++;
                continue;
            }
            if(pos==0)
                id_+=ln[i1];
            if(pos==1)
                val_+=ln[i1];
        }
        if(in_general && !id_.compare("accept_ca_certs")) {
            m_accept_ca_certs=utils_convert_unix_path(val_);
            continue;
        }
        if(in_general && !id_.compare("server_cert")) {
            m_server_cert=utils_convert_unix_path(val_);
            continue;
        }
        if(in_general && !id_.compare("server_key")) {
            m_server_key=utils_convert_unix_path(val_);
            continue;
        }
        if(in_general && !id_.compare("client_cert")) {
            m_client_cert=utils_convert_unix_path(val_);
            continue;
        }
        if(in_general && !id_.compare("client_key")) {
            m_client_key=utils_convert_unix_path(val_);
            continue;
        }
        if(!in_bot.empty() && !id_.compare("cert")) {
            bot.cert=utils_convert_unix_path(val_);
            continue;
        }
        if(!in_bot.empty() && !id_.compare("key")) {
            bot.key=utils_convert_unix_path(val_);
            continue;
        }
        if(!in_bot.empty() && !id_.compare("disable_ca_verification")) {
            if(!val_.compare("1"))
                bot.m_disable_ca_verify=true;
            continue;
        }
        if(!in_bot.empty() && !id_.compare("ignore_expired_cert")) {
            if(!val_.compare("1"))
                bot.m_ignore_expired_peers_cert=true;
            continue;
        }
        if(!in_bot.empty() && !id_.compare("server_cert")) {
            bot.m_server_cert=utils_convert_unix_path(val_);
            continue;
        }
    }
    fclose(f);
    if(!in_bot.empty()) {
        m_ssl_bots.push_back(bot);
        bot.botname="";
        bot.cert="";
        bot.key="";
    }

    return false;
}

string c_ssl_conf::get_cas()
{
    c_mutex_locker lock(m_mutex);

    return m_accept_ca_certs;
}

string c_ssl_conf::get_client_cert()
{
    c_mutex_locker lock(m_mutex);

    return m_client_cert;
}

string c_ssl_conf::get_client_key()
{
    c_mutex_locker lock(m_mutex);

    return m_client_key;
}

string c_ssl_conf::get_server_cert()
{
    c_mutex_locker lock(m_mutex);

    return m_server_cert;
}

string c_ssl_conf::get_server_key()
{
    c_mutex_locker lock(m_mutex);

    return m_server_key;
}

bool c_ssl_conf::get_ssl_bot(c_ssl_bot& bot)
{
    c_mutex_locker lock(m_mutex);

    vector<c_ssl_bot>::iterator i1;
    for(i1=m_ssl_bots.begin(); i1!=m_ssl_bots.end(); ++i1) {
        if(!bot.botname.compare((*i1).botname)) {
            bot=*i1;
            return true;
        }
    }

    return false;
}

c_ssl_bot::c_ssl_bot()
: m_only_get_peers_cert(false)
, m_disable_ca_verify(false)
, m_ignore_expired_peers_cert(false)
{
    ;
}

string ssl_get_pem_cert(const string& bind, const string& host, unsigned short port, bool& error)
{
    if(ssl_conf.rehash()) {
        error=true;
        return (string)"Error hashing \"ssl.txt\"";
    }

    error=false;
    string res;

    char* b=MY_NEW char[bind.length()+32];
    strcpy(b,bind.c_str());

    char* h=MY_NEW char[host.length()+32];
    strcpy(h,host.c_str());

    {
        int ec=0;
        c_socket s=sock_connect(b,h,port,ec,true,"SSL get remote server's certificate");
        if(0!=ec || !s.cmp()) {
            MY_DELETE_A(b);
            MY_DELETE_A(h);

            char* errstr=NULL;
            errstr=sock_error(ec);
            res=errstr;
            MY_DELETE_A(errstr);
            error=true;
            return res;
        }
        MY_DELETE_A(b);
        MY_DELETE_A(h);

        c_ssl* ssl=MY_NEW c_ssl(s,NULL,ssl_conf.get_client_cert(),ssl_conf.get_client_key(),true,true,true);

        static const char dummy[] = "BOO!";
        ssl->write(dummy,strlen(dummy));

        enum {
            TIMEOUT_SECS = 30U,
        };

        my_time_t start=my_time(NULL);
        while(start+TIMEOUT_SECS>my_time(NULL)) {
            res=ssl->get_remote_pkey();
            if(!res.empty())
                break;

            if(ssl->get_error(res)) {
                error=true;
                return res;
            }

            my_sleep(10);
        }

        if(res.empty()) {
            res="Unable to get remote peer's public key (timeout of ";
            res+=tos((unsigned)TIMEOUT_SECS)+" seconds). Is it SSL-enabled peer? Have you got an I:line?";
            error=true;
        }

        ssl_put_to_garbage(ssl);
    }
    ssl_loop(true);
    sock_send_cache(true);
    sock_garbage_loop();

    return res;
}

