/***************************************************************************
                          stats.cpp  -  description
                             -------------------
    begin                : Thu Dec 9 2004
    copyright            : (C) 2004-2007 by Marian VooDooMan Meravy
    email                : ghostvoodooman -a.T= users (dot) sourceforge.net
 ***************************************************************************/

/* $Id: stats.cpp,v 1.53 2008/03/25 12:04:23 ghostvoodooman Exp $ */

/***************************************************************************

VooDoo cIRCle - an IRC (ro)bot
Copyright (C) 2004-2008 by Marian VooDooMan Meravy
  (ghostvoodooman -a.T= users (dot) sourceforge.net)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************/

/*!
    \file
    \brief Provides some statistics
    \version $Id: stats.cpp,v 1.53 2008/03/25 12:04:23 ghostvoodooman Exp $
*/

#if defined(HAVE_CONFIG_H)
#   include <config.h>
#endif

#include "params.h"
#include "utils/memleak.h"
#include "utils/assert.h"
#include "compat.h"

#include "sock_shared.h"

#include <string>
#include <vector>
#include <stack>
#include <list>
#include <map>

#include <time.h>
#include <stdlib.h>

#if !defined(_WIN32) && !defined(_WIN64) && !defined(__CYGWIN__)
#   include <errno.h>
#   include <sys/time.h>
#   include <sys/resource.h>
#else
#   include <windows.h>
#endif

using namespace std;

#include "utils.h"

#include "stats.h"
#include "lang.h"
#include "utils.h"

#include "utils/my_getpid.h"
#include "memory/heap_debug2.h"
#include "var_keeper.h"

#include "utils/precomp.h"

#if defined(_MSC_VER) || defined(__BORLANDC__)
#   pragma hdrstop
#endif

my_time_t start_time=my_time(NULL);           //!< Timestamp of starup (used for UpTime statistics)

/*!
    \brief Stores statistics informations
    \author VooDooMan
    \version 1
    \date 2004
*/
class c_stats
{
public:
    size_t all_bytes_sent;                          //!< All TCP bytes sent
    size_t all_bytes_received;                      //!< All TCP bytes received

    size_t irc_all_bytes_sent;                      //!< All bytes sent to IRC server
    size_t irc_all_bytes_received;                  //!< All bytes received from IRC server

    size_t irc_last_bytes_sent;                     //!< All bytes sent to IRC server on last connection
    size_t irc_last_bytes_received;                 //!< All bytes received from IRC server on last connection

    unsigned long irc_connections;                  //!< Number of connects to IRC server

    size_t dcc_chat_bytes_sent;                     //!< DCC CHAT / telnet all bytes sent
    size_t dcc_chat_bytes_received;                 //!< DCC CHAT / telnet all bytes received

    unsigned long dcc_chat_connections;             //!< Number of DCC CHAT / telnet or BOTNET connections

    size_t dcc_send_bytes_sent;                     //!< Number of bytes sent via DCC SEND
    size_t dcc_send_bytes_received;                 //!< Number of bytes received via DCC SEND

    unsigned long dcc_send_connections;             //!< Number of DCC SEND connections

    size_t botnet_bytes_sent;                       //!< Number of bytes sent via BOTNET
    size_t botnet_bytes_received;                   //!< Number of bytes received via BOTNET

    unsigned long botnet_connections;               //!< Number of BOTNET connections
    unsigned long botnet_auth_ok;                   //!< Number of successful authentication on BOTNET
    unsigned long botnet_auth_error;                //!< Number of failed authentication on BOTNET
    unsigned long botnet_auth_proto_error;          //!< Number of protocol errors on BOTNET

    unsigned long botnet_objects_pushed;            //!< Number of object pushed (BOTNET)
    unsigned long botnet_objects_pulled;            //!< Number of object pulled (BOTNET)

    unsigned long botnet_proc_pushed;               //!< Number of procedures pushed (BOTNET)
    unsigned long botnet_proc_pulled;               //!< Number of procedures pulled (BOTNET)

    unsigned long botnet_user_pushed;               //!< Number of users pushed (BOTNET)
    unsigned long botnet_user_pulled;               //!< Number of users pulled (BOTNET)

    unsigned long botnet_private_pushed;            //!< Number of private definitions pushed (BOTNET)
    unsigned long botnet_private_pulled;            //!< Number of private definitions pulled (BOTNET)

    unsigned long botnet_chandef_pushed;            //!< Number of channel definitions pushed (BOTNET)
    unsigned long botnet_chandef_pulled;            //!< Number of channel definitions pulled (BOTNET)

    unsigned long botnet_partyline_events;          //!< Number of partyline events (BOTNET)

    unsigned long mail_sent;                        //!< Number of e-mails sent

    size_t identd_bytes_sent;                       //!< Number of IDENTD bytes sent
    size_t identd_bytes_received;                   //!< Number of IDENTD bytes received
    unsigned long identd_connections;               //!< Number of IDENTD connections

    string current_irc_server_host;                 //!< Name of host of the curret IRC server
    unsigned short  current_irc_server_port;        //!< Port of the curret IRC server

    /*!
        \brief Clean-ups the structure
        \author VooDooMan
        \version 1
        \date 2004
    */
    void clear()
    {
        all_bytes_sent=all_bytes_received=irc_all_bytes_sent=irc_all_bytes_received=
            irc_last_bytes_sent=irc_last_bytes_received=dcc_chat_bytes_sent=
            dcc_chat_bytes_received=dcc_send_bytes_sent=dcc_send_bytes_received=
            botnet_bytes_sent=botnet_bytes_received=identd_bytes_sent=
            identd_bytes_received=0;

        irc_connections=dcc_chat_connections=dcc_send_connections=botnet_connections=
            botnet_auth_ok=botnet_auth_error=botnet_auth_proto_error=botnet_objects_pushed=
            botnet_objects_pulled=botnet_proc_pushed=botnet_proc_pulled=
            botnet_user_pushed=botnet_user_pulled=botnet_private_pushed=
            botnet_private_pulled=botnet_chandef_pushed=botnet_chandef_pulled=
            botnet_partyline_events=mail_sent=identd_connections=0;

        current_irc_server_host="";
        current_irc_server_port=0;
    }

    c_stats()
    {
        clear();
    }
};

c_stats stats;                                  //!< Structure with statistics

/*!
    \brief Initializes statistics module
    \author VooDooMan
    \version 1
    \date 2004
*/
void stats_init()
{
    stats.clear();
}

/*!
    \brief Should be called when bytes were sent to IRC server
    \author VooDooMan
    \version 1
    \date 2004
    \param bytes Number of bytes
*/
void stats_irc_bytes_sent(size_t bytes)
{
    stats.all_bytes_sent+=bytes;
    stats.irc_all_bytes_sent+=bytes;
    stats.irc_last_bytes_sent+=bytes;
}

/*!
    \brief Should be called when bytes were received from IRC server
    \author VooDooMan
    \version 1
    \date 2004
    \param bytes Number of bytes
*/
void stats_irc_bytes_received(size_t bytes)
{
    stats.all_bytes_received+=bytes;
    stats.irc_all_bytes_received+=bytes;
    stats.irc_last_bytes_received+=bytes;
}

/*!
    \brief Should be called when new connection to IRC server was established
    \author VooDooMan
    \version 1
    \date 2004
    \param irc_server_host Host for new IRC server
    \param irc_server_port Port for new IRC server
*/
void stats_irc_new_connection(string irc_server_host, unsigned short irc_server_port)
{
    stats.irc_last_bytes_received=0;
    stats.irc_last_bytes_sent=0;
    stats.irc_connections++;
    stats.current_irc_server_host=irc_server_host;
    stats.current_irc_server_port=irc_server_port;
}

/*!
    \brief Should be called when bytes were sent to DCC CHAT connection
    \author VooDooMan
    \version 1
    \date 2004
    \param bytes Number of bytes
*/
void stats_dcc_chat_bytes_sent(size_t bytes)
{
    stats.all_bytes_sent+=bytes;
    stats.dcc_chat_bytes_sent+=bytes;
}

/*!
    \brief Should be called when bytes were received from DCC CHAT connection
    \author VooDooMan
    \version 1
    \date 2004
    \param bytes Number of bytes
*/
void stats_dcc_chat_bytes_received(size_t bytes)
{
    stats.all_bytes_received+=bytes;
    stats.dcc_chat_bytes_received+=bytes;
}

/*!
    \brief Should be called when bytes were sent to DCC SEND connection
    \author VooDooMan
    \version 1
    \date 2004
    \param bytes Number of bytes
*/
void stats_dcc_send_bytes_sent(size_t bytes)
{
    stats.all_bytes_sent+=bytes;
    stats.dcc_send_bytes_sent+=bytes;
}

/*!
    \brief Should be called when bytes were received from DCC SEND connection
    \author VooDooMan
    \version 1
    \date 2004
    \param bytes Number of bytes
*/
void stats_dcc_send_bytes_received(size_t bytes)
{
    stats.all_bytes_received+=bytes;
    stats.dcc_send_bytes_received+=bytes;
}

/*!
    \brief Should be called when new DCC CHAT connection was established
    \author VooDooMan
    \version 1
    \date 2004
*/
void stats_dcc_chat_new_connection()
{
    stats.dcc_chat_connections++;
}

/*!
    \brief Should be called when new DCC SEND connection was established
    \author VooDooMan
    \version 1
    \date 2004
*/
void stats_dcc_send_new_connection()
{
    stats.dcc_send_connections++;
}

/*!
    \brief Should be called when bytes were sent to BOTNET
    \author VooDooMan
    \version 1
    \date 2004
    \param bytes Number of bytes
*/
void stats_botnet_bytes_sent(size_t bytes)
{
    stats.all_bytes_sent+=bytes;
    stats.botnet_bytes_sent+=bytes;
}

/*!
    \brief Should be called when bytes were received from BOTNET
    \author VooDooMan
    \version 1
    \date 2004
    \param bytes Number of bytes
*/
void stats_botnet_bytes_received(size_t bytes)
{
    stats.all_bytes_received+=bytes;
    stats.botnet_bytes_received+=bytes;
}

/*!
    \brief Should be called when new IDENTD connection was establish
    \author VooDooMan
    \version 1
    \date 2004
*/
void stats_identd_new_connection()
{
    stats.identd_connections++;
}

/*!
    \brief Should be called when bytes were sent to IDENTD request
    \author VooDooMan
    \version 1
    \date 2004
    \param bytes Number of bytes
*/
void stats_identd_bytes_sent(size_t bytes)
{
    stats.all_bytes_sent+=bytes;
    stats.identd_bytes_sent+=bytes;
}

/*!
    \brief Should be called when bytes were received from IDENTD request
    \author VooDooMan
    \version 1
    \date 2004
    \param bytes Number of bytes
*/
void stats_identd_bytes_received(size_t bytes)
{
    stats.all_bytes_received+=bytes;
    stats.identd_bytes_received+=bytes;
}

/*!
    \brief Should be called when new BOTNET connection was established
    \author VooDooMan
    \version 1
    \date 2004
*/
void stats_botnet_new_connection()
{
    stats.botnet_connections++;
}

/*!
    \brief Should be called when a procedure was pushed (BOTNET)
    \author VooDooMan
    \version 1
    \date 2004
*/
void stats_botnet_proc_pushed()
{
    stats.botnet_objects_pushed++;
    stats.botnet_proc_pushed++;
}

/*!
    \brief Should be called when a procedure was pulled (BOTNET)
    \author VooDooMan
    \version 1
    \date 2004
*/
void stats_botnet_proc_pulled()
{
    stats.botnet_objects_pulled++;
    stats.botnet_proc_pulled++;
}

/*!
    \brief Should be called when an user was pushed (BOTNET)
    \author VooDooMan
    \version 1
    \date 2004
*/
void stats_botnet_user_pushed()
{
    stats.botnet_objects_pushed++;
    stats.botnet_user_pushed++;
}

/*!
    \brief Should be called when an user was pushed (BOTNET)
    \author VooDooMan
    \version 1
    \date 2004
*/
void stats_botnet_user_pulled()
{
    stats.botnet_objects_pulled++;
    stats.botnet_user_pulled++;
}

/*!
    \brief Should be called when a private definition was pushed (BOTNET)
    \author VooDooMan
    \version 1
    \date 2004
*/
void stats_botnet_private_pushed()
{
    stats.botnet_objects_pushed++;
    stats.botnet_private_pushed++;
}

/*!
    \brief Should be called when a private definition was pulled (BOTNET)
    \author VooDooMan
    \version 1
    \date 2004
*/
void stats_botnet_private_pulled()
{
    stats.botnet_objects_pulled++;
    stats.botnet_private_pulled++;
}

/*!
    \brief Should be called when a channel definition was pushed (BOTNET)
    \author VooDooMan
    \version 1
    \date 2004
*/
void stats_botnet_chandef_pushed()
{
    stats.botnet_objects_pushed++;
    stats.botnet_chandef_pushed++;
}

/*!
    \brief Should be called when a channel definition was pulled (BOTNET)
    \author VooDooMan
    \version 1
    \date 2004
*/
void stats_botnet_chandef_pulled()
{
    stats.botnet_objects_pulled++;
    stats.botnet_chandef_pulled++;
}

/*!
    \brief Should be called on a partyline event occured (BOTNET)
    \author VooDooMan
    \version 1
    \date 2004
*/
void stats_botnet_partyline_event()
{
    stats.botnet_partyline_events++;
}

/*!
    \brief Should be called when successful authentication with another bot on BOTNET
    \author VooDooMan
    \version 1
    \date 2004
*/
void stats_botnet_auth_ok()
{
    stats.botnet_auth_ok++;
}

/*!
    \brief Should be called when failed authentication with another bot on BOTNET
    \author VooDooMan
    \version 1
    \date 2004
*/
void stats_botnet_auth_error()
{
    stats.botnet_auth_error++;
}

/*!
    \brief Should be called when protocol error occured with another bot on BOTNET
    \author VooDooMan
    \version 1
    \date 2004
*/
void stats_botnet_auth_proto_error()
{
    stats.botnet_auth_proto_error++;
}

/*!
    \brief Should be called when an e-mail was sent
    \author VooDooMan
    \version 1
    \date 2004
*/
void stats_mail_sent()
{
    stats.mail_sent++;
}

/*!
    \brief Converts number to string in format: 1,234,567
    \author VooDooMan
    \version 1
    \date 2004
    \param number Number to convert
    \return Returns formatted string
*/
string stats_format_number(MY_U_LONG_LONG number)
{
    string str1=(string)"!"+tos(number); // add dummy helper "!"
    string str2;
    int i2=0;
    for(size_t i1=str1.length()-1; (signed)i1>0; --i1, ++i2) {
        if(i2==3) {
            str2=(string)","+str2;
            i2=0;
        }
        str2=(string)""+str1[i1]+str2;
    }
    return str2;
}

/*!
    \brief Puts statistics to an array
    \author VooDooMan
    \version 2
    \date 2004, 2007-11-01
    \param lines Returns array of strings (lines)
    \param lang Language code (from file "lang/lang01.txt") or an empty string to use default english (not from language file; for log purpose)
    \param only_heap_debug If set to 'true', result in 'lines' will contain only information about heap debug
*/
void stats_display(list<string>& lines, string lang, bool only_heap_debug)
{
    lines.clear();

    string line;

    if(!only_heap_debug) {

        {
            my_time_t now=my_time(NULL);
            if(!start_time)
                start_time=now;

            {
                string pid=tos(my_getpid());

                if(!lang.empty())
                    line=lang_subst(lang_get_string(1,lang,763),pid);
                else
                    line=lang_subst("I am pid %s",pid);
                lines.push_back(line);
                line="";
            }

            unsigned long up=(unsigned long)my_difftime(now,start_time);

            unsigned long days=up/(24*60*60);
            up%=24*60*60;
            unsigned long hrs=up/(60*60);
            up%=60*60;
            unsigned long mins=up/(60);
            up%=60;
            unsigned long secs=up;

            if(!lang.empty()) {
                line=lang_subst(lang_get_string(1,lang,752),make_nice_time(my_time(NULL),false));
                lines.push_back(line);
                line="";

                line+=lang_get_string(1,lang,552);
                line+=" ";

                line+=tos(days);
                line+=" ";
                line+=lang_get_string(1,lang,553);
                line+=", ";

                line+=tos(hrs);
                line+=" ";
                line+=lang_get_string(1,lang,554);
                line+=", ";

                line+=tos(mins);
                line+=" ";
                line+=lang_get_string(1,lang,555);
                line+=", ";

                line+=tos(secs);
                line+=" ";
                line+=lang_get_string(1,lang,556);
            } else {
                line="Local time: ";
                line+=make_nice_time(my_time(NULL),false);
                lines.push_back(line);
                line="";

                line+="Bot uptime: ";

                line+=tos(days);
                line+=" day(s), ";

                line+=tos(hrs);
                line+=" hour(s), ";

                line+=tos(mins);
                line+=" minute(s), ";

                line+=tos(secs);
                line+=" second(s)";
            }
            lines.push_back(line);
        }

        {
            if(!lang.empty()) {
                line=lang_subst(lang_get_string(1,lang,770),tos(stats_cpu_time()));
                lines.push_back(line);
                line="";

                line=lang_subst(lang_get_string(1,lang,770),make_nice_time_days_only(stats_cpu_time(),lang_get_string(1,lang,751).c_str()));
                lines.push_back(line);
                line="";
            } else {
                line="CPU time: ";
                line+=tos(stats_cpu_time());
                line+=" seconds";
                lines.push_back(line);
                line="";

                line="CPU time: ";
                line+=make_nice_time_days_only(stats_cpu_time()," days, ");
                lines.push_back(line);
                line="";
            }
        }

        line="";
        if(!lang.empty())
            line+=lang_get_string(1,lang,559);
        else
            line+="Using IRC server host/port:";
        line+=" ";
        if(stats.current_irc_server_host.find(":",0)!=string::npos)
            line+="[";
        line+=stats.current_irc_server_host;
        if(stats.current_irc_server_host.find(":",0)!=string::npos)
            line+="]";
        line+=":";
        line+=tos(stats.current_irc_server_port);
        lines.push_back(line);

        line="";
        if(!lang.empty())
            line+=lang_get_string(1,lang,503);
        else
            line+="All sent bytes:";
        line+=" ";
        line+=stats_format_number(stats.all_bytes_sent);
        lines.push_back(line);

        line="";
        if(!lang.empty())
            line+=lang_get_string(1,lang,504);
        else
            line+="All received bytes:";
        line+=" ";
        line+=stats_format_number(stats.all_bytes_received);
        lines.push_back(line);

        line="";
        if(!lang.empty())
            line+=lang_get_string(1,lang,505);
        else
            line+="All sent bytes to IRC server:";
        line+=" ";
        line+=stats_format_number(stats.irc_all_bytes_sent);
        lines.push_back(line);

        line="";
        if(!lang.empty())
            line+=lang_get_string(1,lang,506);
        else
            line+="All received bytes from IRC server:";
        line+=" ";
        line+=stats_format_number(stats.irc_all_bytes_received);
        lines.push_back(line);

        line="";
        if(!lang.empty())
            line+=lang_get_string(1,lang,507);
        else
            line+="All sent bytes to IRC server on last connection:";
        line+=" ";
        line+=stats_format_number(stats.irc_last_bytes_sent);
        lines.push_back(line);

        line="";
        if(!lang.empty())
            line+=lang_get_string(1,lang,508);
        else
            line+="All received bytes from IRC server on last connection:";
        line+=" ";
        line+=stats_format_number(stats.irc_last_bytes_received);
        lines.push_back(line);

        line="";
        if(!lang.empty())
            line+=lang_get_string(1,lang,509);
        else
            line+="Number of connections established to IRC server:";
        line+=" ";
        line+=stats_format_number(stats.irc_connections);
        lines.push_back(line);

        line="";
        if(!lang.empty())
            line+=lang_get_string(1,lang,510);
        else
            line+="All sent bytes to DCC CHAT / telnet:";
        line+=" ";
        line+=stats_format_number(stats.dcc_chat_bytes_sent);
        lines.push_back(line);

        line="";
        if(!lang.empty())
            line+=lang_get_string(1,lang,511);
        else
            line+="All received bytes from DCC CHAT / telnet:";
        line+=" ";
        line+=stats_format_number(stats.dcc_chat_bytes_received);
        lines.push_back(line);

        line="";
        if(!lang.empty())
            line+=lang_get_string(1,lang,512);
        else
            line+="Number of DCC CHAT / telnet, or incomming BOTNET connections established:";
        line+=" ";
        line+=stats_format_number(stats.dcc_chat_connections);
        lines.push_back(line);

        line="";
        if(!lang.empty())
            line+=lang_get_string(1,lang,513);
        else
            line+="All sent bytes to DCC SEND:";
        line+=" ";
        line+=stats_format_number(stats.dcc_send_bytes_sent);
        lines.push_back(line);

        line="";
        if(!lang.empty())
            line+=lang_get_string(1,lang,514);
        else
            line+="All received bytes from DCC SEND:";
        line+=" ";
        line+=stats_format_number(stats.dcc_send_bytes_received);
        lines.push_back(line);

        line="";
        if(!lang.empty())
            line+=lang_get_string(1,lang,515);
        else
            line+="Number of DCC SEND connections established:";
        line+=" ";
        line+=stats_format_number(stats.dcc_send_connections);
        lines.push_back(line);

        line="";
        if(!lang.empty())
            line+=lang_get_string(1,lang,536);
        else
            line+="All sent bytes to IDENT requests:";
        line+=" ";
        line+=stats_format_number(stats.identd_bytes_sent);
        lines.push_back(line);

        line="";
        if(!lang.empty())
            line+=lang_get_string(1,lang,537);
        else
            line+="All received bytes from IDENT requests:";
        line+=" ";
        line+=stats_format_number(stats.identd_bytes_received);
        lines.push_back(line);

        line="";
        if(!lang.empty())
            line+=lang_get_string(1,lang,538);
        else
            line+="Number of IDENT connections established:";
        line+=" ";
        line+=stats_format_number(stats.identd_connections);
        lines.push_back(line);

        line="";
        if(!lang.empty())
            line+=lang_get_string(1,lang,516);
        else
            line+="All sent bytes to BOTNET:";
        line+=" ";
        line+=stats_format_number(stats.botnet_bytes_sent);
        lines.push_back(line);

        line="";
        if(!lang.empty())
            line+=lang_get_string(1,lang,517);
        else
            line+="All received bytes from BOTNET:";
        line+=" ";
        line+=stats_format_number(stats.botnet_bytes_received);
        lines.push_back(line);

        line="";
        if(!lang.empty())
            line+=lang_get_string(1,lang,518);
        else
            line+="Number of BOTNET connections established:";
        line+=" ";
        line+=stats_format_number(stats.botnet_connections);
        lines.push_back(line);

        line="";
        if(!lang.empty())
            line+=lang_get_string(1,lang,519);
        else
            line+="Number of successful BOTNET authentications:";
        line+=" ";
        line+=stats_format_number(stats.botnet_auth_ok);
        lines.push_back(line);

        line="";
        if(!lang.empty())
            line+=lang_get_string(1,lang,520);
        else
            line+="Number of failed BOTNET authentications:";
        line+=" ";
        line+=stats_format_number(stats.botnet_auth_error);
        lines.push_back(line);

        line="";
        if(!lang.empty())
            line+=lang_get_string(1,lang,521);
        else
            line+="Number of BOTNET protocol errors:";
        line+=" ";
        line+=stats_format_number(stats.botnet_auth_proto_error);
        lines.push_back(line);

        line="";
        if(!lang.empty())
            line+=lang_get_string(1,lang,522);
        else
            line+="Number of all pushed BOTNET objects:";
        line+=" ";
        line+=stats_format_number(stats.botnet_objects_pushed);
        lines.push_back(line);

        line="";
        if(!lang.empty())
            line+=lang_get_string(1,lang,523);
        else
            line+="Number of all pulled BOTNET objects:";
        line+=" ";
        line+=stats_format_number(stats.botnet_objects_pulled);
        lines.push_back(line);

        line="";
        if(!lang.empty())
            line+=lang_get_string(1,lang,524);
        else
            line+="Number of pushed BOTNET procedures:";
        line+=" ";
        line+=stats_format_number(stats.botnet_proc_pushed);
        lines.push_back(line);

        line="";
        if(!lang.empty())
            line+=lang_get_string(1,lang,525);
        else
            line+="Number of pulled BOTNET procedures:";
        line+=" ";
        line+=stats_format_number(stats.botnet_proc_pulled);
        lines.push_back(line);

        line="";
        if(!lang.empty())
            line+=lang_get_string(1,lang,526);
        else
            line+="Number of pushed BOTNET users:";
        line+=" ";
        line+=stats_format_number(stats.botnet_user_pushed);
        lines.push_back(line);

        line="";
        if(!lang.empty())
            line+=lang_get_string(1,lang,527);
        else
            line+="Number of pulled BOTNET users:";
        line+=" ";
        line+=stats_format_number(stats.botnet_user_pulled);
        lines.push_back(line);

        line="";
        if(!lang.empty())
            line+=lang_get_string(1,lang,528);
        else
            line+="Number of pushed BOTNET private:";
        line+=" ";
        line+=stats_format_number(stats.botnet_private_pushed);
        lines.push_back(line);

        line="";
        if(!lang.empty())
            line+=lang_get_string(1,lang,529);
        else
            line+="Number of pulled BOTNET private:";
        line+=" ";
        line+=stats_format_number(stats.botnet_private_pulled);
        lines.push_back(line);

        line="";
        if(!lang.empty())
            line+=lang_get_string(1,lang,530);
        else
            line+="Number of pushed BOTNET channels:";
        line+=" ";
        line+=stats_format_number(stats.botnet_chandef_pushed);
        lines.push_back(line);

        line="";
        if(!lang.empty())
            line+=lang_get_string(1,lang,531);
        else
            line+="Number of pulled BOTNET channels:";
        line+=" ";
        line+=stats_format_number(stats.botnet_chandef_pulled);
        lines.push_back(line);

        line="";
        if(!lang.empty())
            line+=lang_get_string(1,lang,532);
        else
            line+="Number of BOTNET partyline events:";
        line+=" ";
        line+=stats_format_number(stats.botnet_partyline_events);
        lines.push_back(line);

        line="";
        if(!lang.empty())
            line+=lang_get_string(1,lang,533);
        else
            line+="Number of e-mail(s) sent:";
        line+=" ";
        line+=stats_format_number(stats.mail_sent);
        lines.push_back(line);
    }

    {
        MY_U_LONG_LONG tran_cnt, rp, errors, blocks_cnt, recreation_cnt;
        size_t largest, total_bytes;

        heap_debug_get_stats(tran_cnt,rp,errors,largest,total_bytes,blocks_cnt,recreation_cnt);

        string disabled;
        if(!lang.empty())
            disabled+=lang_get_string(1,lang,778);
        else
            disabled+="(disabled in file \"params.h\")";

        line="";
        if(!lang.empty())
            line+=lang_get_string(1,lang,774);
        else
            line+="HEAP DEBUG: Number of re/allocation + deallocation transactions/events:";
        line+=" ";
        line+=(tran_cnt==0) ? disabled : stats_format_number(tran_cnt);
        lines.push_back(line);

        line="";
        if(!lang.empty())
            line+=lang_get_string(1,lang,775);
        else
            line+="HEAP DEBUG: Number of heap block STL map's recreation events:";
        line+=" ";
        line+=(rp==0) ? disabled : stats_format_number(recreation_cnt);
        lines.push_back(line);

        line="";
        if(!lang.empty())
            line+=lang_get_string(1,lang,776);
        else
            line+="HEAP DEBUG: Number of heap transactions for STL map to recreate:";
        line+=" ";
        line+=(rp==0) ? disabled : stats_format_number(rp);
        lines.push_back(line);

        line="";
        if(!lang.empty())
            line+=lang_get_string(1,lang,780);
        else
            line+="HEAP DEBUG: Largest allocated heap block (not counting STL's allocations):";
        line+=" ";
#if REPORT_LEAKS==1
        line+=stats_format_number(largest);
#else
        line+=disabled;
#endif
        lines.push_back(line);

        line="";
        if(!lang.empty())
            line+=lang_get_string(1,lang,783);
        else
            line+="HEAP DEBUG: Number of allocated blocks on the heap (not counting STL's allocations):";
        line+=" ";
#if REPORT_LEAKS==1
        line+=stats_format_number(blocks_cnt);
#else
        line+=disabled;
#endif
        lines.push_back(line);

        line="";
        if(!lang.empty())
            line+=lang_get_string(1,lang,781);
        else
            line+="HEAP DEBUG: Total bytes allocated on the heap (not counting STL's allocations):";
        line+=" ";
#if REPORT_LEAKS==1
        line+=stats_format_number(total_bytes);
#else
        line+=disabled;
#endif
        lines.push_back(line);

        line="";
        if(!lang.empty())
            line+=lang_get_string(1,lang,779);
        else
            line+="HEAP DEBUG: Number of heap errors:";
        line+=" ";
#if REPORT_LEAKS==1
        line+=stats_format_number(errors);
#else
        line+=disabled;
#endif
        lines.push_back(line);

        line="";
        if(!lang.empty())
            line+=lang_get_string(1,lang,782);
        else
            line+="Number of variables in variable tree:";
        line+=" ";
        line+=stats_format_number(var_keeper->get_size());
        lines.push_back(line);
    }
}

/*!
    \brief Returns uptime
    \author VooDooMan
    \version 1
    \date 2005-11-26
*/
MY_U_LONG_LONG stats_uptime()
{
    my_time_t now=my_time(NULL);
    if(!start_time)
        start_time=now;

    return now-start_time;
}

/*!
    \brief Returns CPU time (number of seconds rounded to up);
        on POSIX: user CPU time + system CPU time, of process itself - not counting child processes;
        on Win: kernel CPU time + user CPU time of this process itself
    \author VooDooMan
    \version 2
    \date 2005-11-26, 2007-07-30
    \return Returns number of seconds
*/
MY_U_LONG_LONG stats_cpu_time()
{
    /*
    clock_t now=clock();
    return (unsigned long)(((double)(now-start_cpu_time))/CLOCKS_PER_SEC);
    */
#if !defined(_WIN32) && !defined(_WIN64)
    struct rusage ru;
    memset(&ru,0,sizeof(ru));
    if(0!=getrusage(RUSAGE_SELF,&ru)) {
        errno=0;
        return 0;
    }

    MY_U_LONG_LONG sec;
    sec=static_cast<MY_U_LONG_LONG>(ru.ru_utime.tv_sec);
    sec+=static_cast<MY_U_LONG_LONG>(ru.ru_stime.tv_sec);

    MY_U_LONG_LONG usec;
    usec=static_cast<MY_U_LONG_LONG>(ru.ru_utime.tv_sec);
    usec+=static_cast<MY_U_LONG_LONG>(ru.ru_stime.tv_sec);

    sec+=usec/1000000; // magic number: div by number of micro secs in 1 second
    if(usec%1000000)
        sec++; // round to up

    return sec;
#else
    FILETIME tmp1, tmp2;
    FILETIME s_time, u_time;
    if(0==GetProcessTimes(GetCurrentProcess(),&tmp1,&tmp2,&s_time,&u_time)) {
        return 0;
    }

    ULARGE_INTEGER s_time2;
    s_time2.LowPart=s_time.dwLowDateTime;
    s_time2.HighPart=s_time.dwHighDateTime;

    ULARGE_INTEGER u_time2;
    u_time2.LowPart=u_time.dwLowDateTime;
    u_time2.HighPart=u_time.dwHighDateTime;

    s_time2.QuadPart+=u_time2.QuadPart;
    s_time2.QuadPart/=10000000; // magic number: div by number of 100-nano secs in 1 second
    if(s_time2.QuadPart%10000000)
        s_time2.QuadPart++; // round to up

    return static_cast<MY_U_LONG_LONG>(s_time2.QuadPart);
#endif
}

