/***************************************************************************
                          ipc_mutex.h  -  description
                             -------------------
    begin                : Sun Sep 2 2007
    copyright            : (C) 2007 by Marian VooDooMan Meravy
    email                : ghostvoodooman -a.T= users (dot) sourceforge.net
 ***************************************************************************/

/* $Id: ipc_mutex.cpp,v 1.15 2008/06/08 08:43:14 ghostvoodooman Exp $ */

/***************************************************************************

VooDoo cIRCle - an IRC (ro)bot
Copyright (C) 2005-2008 by Marian VooDooMan Meravy
  (ghostvoodooman -a.T= users (dot) sourceforge.net)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************/

/*!
    \file
    \brief Implements interprocess mutex
    \version $Id: ipc_mutex.cpp,v 1.15 2008/06/08 08:43:14 ghostvoodooman Exp $
*/

#if defined(HAVE_CONFIG_H)
#   include <config.h>
#endif

#include <string>
#include <sstream>

using namespace std;

#ifdef HAVE_SIGNAL
#   include <signal.h>
#endif

#ifdef HAVE_UNISTD_H
#   include <unistd.h>
#endif

#include <errno.h>

#include "../params.h"
#include "../utils/assert.h"
#include "../compat.h"

#include "../conf.h"

#include "ipc_mutex.h"
#include "../internal_state.h"

#include "../utils/precomp.h"

c_ipc_mutex::c_ipc_mutex(const char* file_name)
    : c_mutex()
{
    // we need to set umask first
    conf_init_pre_start();
    if(conf_umask!=-1)
        umask(conf_umask);

#if defined(MY_USE_SEMAPHORE)
create_again:
    m_was_created=true;
    m_semaphore=semaphore_create(file_name,fd);
    if(!m_semaphore) {
        m_was_created=false;
        m_semaphore=semaphore_open(file_name,fd);

        if(!m_semaphore) {
            error_log+="OOPS: Cannot create/open shared semaphore in file '";
            error_log+=file_name;
            error_log+="'!\n";
            return;
        }

        if(test_stale_lock_file()) {
            if(m_semaphore)
                semaphore_close(m_semaphore,fd);
            if(!unlink(file_name)) {
                goto create_again;
            }
        }
    } else {
        semaphore_post(m_semaphore);
    }
    MY_ASSERT(m_semaphore);

    place_lock();
#endif

#if defined(_WIN32) || defined(_WIN64) || defined(__CYGWIN__)
    DWORD rc;
    char dir[MAX_PATH+1+1];
    dir[0]='\0';
    rc=GetCurrentDirectory(sizeof(dir)-1,(LPTSTR)dir);
    MY_ASSERT(rc);
    string kernel_name1=dir;
    kernel_name1+=FILE_SLASH;
    kernel_name1+=file_name;
    kernel_name1=str_to_lower(kernel_name1);

    ostringstream kernel_name2;
    kernel_name2 << "voodoo-circle_";
    for(size_t i1=0; i1<kernel_name1.length(); ++i1) {
        if(kernel_name1[i1]=='\\') {
            kernel_name2 << ".";
            continue;
        }
        if(kernel_name1[i1]=='.') {
            kernel_name2 << "_._";
            continue;
        }
        if(kernel_name1[i1]=='_') {
            kernel_name2 << "_-_";
            continue;
        }
        if(kernel_name1[i1]<'a' || kernel_name1[i1]>'z') {
            if(kernel_name1[i1]<'0' || kernel_name1[i1]>'9') {
                ostringstream tmp;
                tmp << "_0x";
                tmp << toshex((unsigned int)(kernel_name1[i1]));
                tmp << "_";
                kernel_name2 << tmp.str();
                continue;
            }
        }
        kernel_name2 << kernel_name1[i1];
    }

    SECURITY_ATTRIBUTES sa;
    memset(&sa,0,sizeof(sa));
    sa.nLength=sizeof(sa);
    sa.bInheritHandle=FALSE;

    m_mutex=CreateMutex(&sa,FALSE,(LPCTSTR)kernel_name2.str().c_str());
    //MY_ASSERT(m_mutex); // it is NULL when another instance is running (we check this condition later)
#endif
}

c_ipc_mutex::~c_ipc_mutex()
{
#if defined(MY_USE_SEMAPHORE)
    if(m_semaphore)
        semaphore_close(m_semaphore,fd);
#endif

#if defined(_WIN32) || defined(_WIN64) || defined(__CYGWIN__)
    if(m_mutex) {
        BOOL rc=CloseHandle(m_mutex);
        MY_ASSERT(rc);
    }
#endif
}

void c_ipc_mutex::lock()
{
#if defined(MY_USE_SEMAPHORE) && defined(HAVE_LIBPTHREAD)
    c_mutex::lock();
    pthread_mutex_lock(&m_cnt_mutex);
    if(m_semaphore && 1==m_lock_cnt) {
        pthread_mutex_unlock(&m_cnt_mutex);
        if(!semaphore_wait(m_semaphore)) {
            pthread_mutex_lock(&m_cnt_mutex);
            timeout_cnt++;
            pthread_mutex_unlock(&m_cnt_mutex);
        }
    } else {
        pthread_mutex_unlock(&m_cnt_mutex);
    }
#endif

#if defined(_WIN32) || defined(_WIN64) || defined(__CYGWIN__)
    if(m_mutex) {
        DWORD rc=WaitForSingleObject(m_mutex,INFINITE);
        MY_ASSERT(WAIT_OBJECT_0==rc || WAIT_ABANDONED==rc);
    }
#endif
}

void c_ipc_mutex::unlock()
{
#if defined(MY_USE_SEMAPHORE) && defined(HAVE_LIBPTHREAD)
    pthread_mutex_lock(&m_cnt_mutex);
    if(m_semaphore && 1==m_lock_cnt)
        semaphore_post(m_semaphore);
    pthread_mutex_unlock(&m_cnt_mutex);
    c_mutex::unlock();
#endif

#if defined(_WIN32) || defined(_WIN64) || defined(__CYGWIN__)
    if(m_mutex) {
        BOOL rc=ReleaseMutex(m_mutex);
        MY_ASSERT(rc);
    }
#endif
}

/*!
    \brief Retrieves our role in exclusivity
    \version 1
    \date 2007-09-02
    \retval true We are the first, i.e. created locking mmap'ed file with semaphore
    \retval false We just connected to already created locking file
*/
bool c_ipc_mutex::was_created()
{
#if defined(MY_USE_SEMAPHORE) && defined(HAVE_LIBPTHREAD)
    return m_was_created;
#endif

#if defined(_WIN32) || defined(_WIN64) || defined(__CYGWIN__)
    // TODO
    return true;
#endif
}

/*!
    \brief Retrieves contidion timeout count on semaphore (POSIX-only)
    \version 1
    \author VooDooMan
    \date 2007-09-06
    \retval 0 Zero times, always zero on Win
*/
unsigned long c_ipc_mutex::get_timeout_cnt_and_reset()
{
#if defined(MY_USE_SEMAPHORE) && defined(HAVE_LIBPTHREAD)
    unsigned long res;

    pthread_mutex_lock(&m_cnt_mutex);
    res=timeout_cnt;
    timeout_cnt=0;
    pthread_mutex_unlock(&m_cnt_mutex);

    return res;
#endif

    return 0;
}

#if !defined(_WIN32) && !defined(_WIN64) && !defined(__CYGWIN__)
#if defined(MY_USE_SEMAPHORE) && defined(HAVE_LIBPTHREAD)
/*!
    \brief Returns true if lock file is dead, and if it is not
    \version 1
    \author VooDooMan
    \date 2007-11-12
*/
bool c_ipc_mutex::test_stale_lock_file()
{
    if(fd<0)
        return true;

    struct flock lck;

    memset(&lck,0,sizeof(lck));
    lck.l_type=F_WRLCK;
    lck.l_start=sizeof(semaphore_t);
    lck.l_whence=SEEK_SET;
    lck.l_len=SEM_LOCKING_REGION_SIZE;
    lck.l_pid=0;
    int ec=fcntl(fd,F_GETLK,&lck);
    return (ec==0 && lck.l_type==F_UNLCK && 0==lck.l_pid);
}
#endif
#endif

#if !defined(_WIN32) && !defined(_WIN64) && !defined(__CYGWIN__)
#if defined(MY_USE_SEMAPHORE) && defined(HAVE_LIBPTHREAD)
/*!
    \brief Places shared (read) lock on locking region of lock file
    \version 1
    \author VooDooMan
    \date 2007-11-12
*/
void c_ipc_mutex::place_lock()
{
    if(fd<0)
        return;

    struct flock lck;

    memset(&lck,0,sizeof(lck));
    lck.l_type=F_RDLCK;
    lck.l_start=sizeof(semaphore_t);
    lck.l_whence=SEEK_SET;
    lck.l_len=SEM_LOCKING_REGION_SIZE;
    lck.l_pid=0;
    fcntl(fd,F_SETLK,&lck);
}
#endif
#endif

/*!
    \brief Retrieves whether inter-process locking (shared motex and condition) works on this platform
*/
bool c_ipc_mutex::ipc_works_on_this_platform()
{
#if defined(_WIN32) || defined(_WIN64) || defined(__CYGWIN__)
    return true;
#else
#   if defined(MY_USE_SEMAPHORE)
    return fd>=0;
#   else
    return false;
#   endif
#endif
}

/*!
    \brief Retrieves queued errors from IPC locking objects and then clears error queue
    \version 1
    \author VooDooMan
    \date 2008-06-06
*/
string c_ipc_mutex::get_error_log()
{
    string log=error_log;
    error_log.clear();
    return log;
}

