/***************************************************************************
                          ipc_mutex.h  -  description
                             -------------------
    begin                : Sun Sep 2 2007
    copyright            : (C) 2007 by Marian VooDooMan Meravy
    email                : ghostvoodooman -a.T= users (dot) sourceforge.net
 ***************************************************************************/

/* $Id: ipc_mutex.h,v 1.15 2008/06/08 08:43:14 ghostvoodooman Exp $ */

/***************************************************************************

VooDoo cIRCle - an IRC (ro)bot
Copyright (C) 2005-2008 by Marian VooDooMan Meravy
  (ghostvoodooman -a.T= users (dot) sourceforge.net)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************/

/*!
    \file
    \brief Implements interprocess mutex
    \version $Id: ipc_mutex.h,v 1.15 2008/06/08 08:43:14 ghostvoodooman Exp $
*/

#ifndef IPC_MUTEX_H__
#define IPC_MUTEX_H__

#if defined(HAVE_CONFIG_H)
#   include <config.h>
#endif

#include "../params.h"

#if !defined(_WIN32) && !defined(_WIN64) && !defined(__CYGWIN__)
#   ifdef USE_PTHREAD
#       ifdef HAVE_LIBPTHREAD
#           include <pthread.h>
#       endif
#   else
#       ifdef HAVE_LIBPTHREAD
#           undef HAVE_LIBPTHREAD
#       endif
#   endif
#else
#   include <windows.h>
#endif

using namespace std;

#include "sem.h"
#include "mutex.h"

/*!
    \brief Interprocess mutex class
    \author VooDooMan
    \version 1
    \date 2007-09-02
    \param file_name Name of shared mmap'ed file (POSIX only)
*/
class c_ipc_mutex : public c_mutex
{
#ifdef MY_USE_SEMAPHORE
    semaphore_t* m_semaphore;       //!< Semaphore structure
    bool m_was_created;             //!< 'true' if we are the first, i.e. created locking mmap'ed file with semaphore; 'false' if we just connected to already created locking file
    unsigned long timeout_cnt;      //!< Number of time-outs (protected by 'm_cnt_mutex')

    int fd;                         //!< File descriptor of lock file
    bool test_stale_lock_file();
    void place_lock();
#endif

#if defined(_WIN32) || defined(_WIN64) || defined(__CYGWIN__)
    HANDLE m_mutex;                 //!< Win* mutex
#endif

    string error_log;

public:
    c_ipc_mutex(const char* file_name);
    virtual ~c_ipc_mutex();

    string get_error_log();

    virtual void lock();
    virtual void unlock();

    bool was_created();

    unsigned long get_timeout_cnt_and_reset();

    bool ipc_works_on_this_platform();
};

#endif // IPC_MUTEX_H__

