/***************************************************************************
                          mutex.cpp  -  description
                             -------------------
    begin                : Sat Jul 28 2007
    copyright            : (C) 2007 by Marian VooDooMan Meravy
    email                : ghostvoodooman -a.T= users (dot) sourceforge.net
 ***************************************************************************/

/* $Id: mutex.cpp,v 1.13 2008/03/11 21:11:48 ghostvoodooman Exp $ */

/***************************************************************************

VooDoo cIRCle - an IRC (ro)bot
Copyright (C) 2005-2008 by Marian VooDooMan Meravy
  (ghostvoodooman -a.T= users (dot) sourceforge.net)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************/

/*!
    \file
    \brief Implements mutex stuff
    \version $Id: mutex.cpp,v 1.13 2008/03/11 21:11:48 ghostvoodooman Exp $
*/

#if defined(HAVE_CONFIG_H)
#   include <config.h>
#endif

#include "../params.h"
#include "../utils.h"

//#define _POSIX_C_SOURCE
#include <stdio.h>
#ifdef _MSC_VER
#   define fileno _fileno
#endif

#ifdef HAVE_FCNTL_H
#   include <fcntl.h>
#   include <errno.h>
#endif

#include "mutex.h"
#include "ipc_mutex.h"

#include "../utils/precomp.h"

c_mutex::c_mutex()
{
#if defined(_WIN32) || defined(_WIN64) || defined(__CYGWIN__)
    InitializeCriticalSection(&m_cs);
#else
#   ifdef HAVE_LIBPTHREAD
    m_lock_cnt=0;
    pthread_mutex_init(&m_mutex,NULL);
    pthread_mutex_init(&m_cnt_mutex,NULL);
#   endif
#endif
}

c_mutex::~c_mutex()
{
#if defined(_WIN32) || defined(_WIN64) || defined(__CYGWIN__)
    DeleteCriticalSection(&m_cs);
#else
#   ifdef HAVE_LIBPTHREAD
    pthread_mutex_destroy(&m_mutex);
    pthread_mutex_destroy(&m_cnt_mutex);
#   endif
#endif
}

void c_mutex::lock()
{
#if defined(_WIN32) || defined(_WIN64) || defined(__CYGWIN__)
    EnterCriticalSection(&m_cs);
#else
#   ifdef HAVE_LIBPTHREAD
lock_again:
    pthread_mutex_lock(&m_cnt_mutex);
    if(m_lock_cnt>0) {
        if(pthread_equal(pthread_self(),m_locking_thread)) {
            m_lock_cnt++;
            pthread_mutex_unlock(&m_cnt_mutex);
            return;
        } else {
            pthread_mutex_unlock(&m_cnt_mutex);
            pthread_mutex_lock(&m_mutex);

            // wait
            pthread_mutex_unlock(&m_mutex);
            goto lock_again;
        }
    } else {
        m_lock_cnt++;
        m_locking_thread=pthread_self();
        pthread_mutex_lock(&m_mutex);
        pthread_mutex_unlock(&m_cnt_mutex);
        return;
    }
#   endif
#endif
}

void c_mutex::unlock()
{
#if defined(_WIN32) || defined(_WIN64) || defined(__CYGWIN__)
    LeaveCriticalSection(&m_cs);
#else
#   ifdef HAVE_LIBPTHREAD
    pthread_mutex_lock(&m_cnt_mutex);
    m_lock_cnt--;
    pthread_mutex_unlock(&m_mutex);
    pthread_mutex_unlock(&m_cnt_mutex);
#   endif
#endif
}

c_mutex_locker::c_mutex_locker(c_mutex& m)
: m_m(&m)
, m_fd(-1)
{
    m_m->lock();
}

c_mutex_locker::c_mutex_locker(c_mutex* m)
: m_m(m)
, m_fd(-1)
{
    if(m_m)
        m_m->lock();
}

c_mutex_locker::~c_mutex_locker()
{
    if(m_m)
        m_m->unlock();
#ifdef HAVE_FCNTL_H
    if(m_fd>=0) {
        while(0!=close(m_fd)) {
            // NOOP
        }
    }
#endif
}

/*!
    \brief Detects if inter-process locking mechanism works on this platform, if not, do a locking on file descriptor as workaround
    \author VooDooMan
    \date 2007-11-12
    \version 1
*/
void c_mutex_locker::lock_file_if_needed(const char* UNUSEDARG_WIN(fn))
{
#ifdef HAVE_FCNTL_H
    if(typeid(*m_m)!=typeid(c_ipc_mutex))
        return; // not applicable

    if(dynamic_cast<c_ipc_mutex*>(m_m)->ipc_works_on_this_platform())
        return;

    m_fd=open(fn,O_RDWR | O_CREAT,0666); // VooDooMan: fmode 0666 is modified by umask in usual way

    if(m_fd>=0) {
        static char* buf="LOCKLOCKLOCKLOCK";
        static size_t count=strlen(buf);
        write(m_fd,buf,count);

        struct flock lck;

        for(;;) {
            memset(&lck,0,sizeof(lck));

            lck.l_type=F_WRLCK;
            lck.l_start=0;
            lck.l_whence=SEEK_SET;
            lck.l_len=0; // whole file
            lck.l_pid=0;
            int rc=fcntl(m_fd,F_SETLKW,&lck);

            if(rc!=0) {
                if(EINTR==errno)
                    continue; // interrupted by a signal

                // other errors: nevermid; what we can do?
            }
            break;
        }
    }
#endif
}

