/***************************************************************************
                          mutex.h  -  description
                             -------------------
    begin                : Sat Jul 28 2007
    copyright            : (C) 2007 by Marian VooDooMan Meravy
    email                : ghostvoodooman -a.T= users (dot) sourceforge.net
 ***************************************************************************/

/* $Id: mutex.h,v 1.13 2008/03/11 21:11:48 ghostvoodooman Exp $ */

/***************************************************************************

VooDoo cIRCle - an IRC (ro)bot
Copyright (C) 2005-2008 by Marian VooDooMan Meravy
  (ghostvoodooman -a.T= users (dot) sourceforge.net)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************/

/*!
    \file
    \brief Implements mutex stuff
    \version $Id: mutex.h,v 1.13 2008/03/11 21:11:48 ghostvoodooman Exp $
*/

#ifndef MUTEX_H__
#define MUTEX_H__

#if defined(HAVE_CONFIG_H)
#   include <config.h>
#endif

#include "../params.h"

#if !defined(_WIN32) && !defined(_WIN64) && !defined(__CYGWIN__)
#   ifdef USE_PTHREAD
#       ifdef HAVE_LIBPTHREAD
#           include <pthread.h>
#       endif
#   else
#       ifdef HAVE_LIBPTHREAD
#           undef HAVE_LIBPTHREAD
#       endif
#   endif
#else
#   include <windows.h>
#endif

/*!
    \brief Win's Critical Section / POSIX threads' mutex wrapper class, with recursive mutex emulation on POSIX
    \author VooDooMan
    \version 1
    \date 2007-07
*/
class c_mutex
{
protected:
#if defined(_WIN32) || defined(_WIN64) || defined(__CYGWIN__)
    CRITICAL_SECTION m_cs;          //!< Locking object, on Win, this is natively recursive
#else
#   ifdef HAVE_LIBPTHREAD
    pthread_mutex_t m_cnt_mutex;    //!< Lock count locking object
    pthread_mutex_t m_mutex;        //!< Locking object
    pthread_t m_locking_thread;     //!< Locking thread ID

    int m_lock_cnt;                 //!< Lock count
#   endif
#endif

public:
    c_mutex();
    virtual ~c_mutex();

    virtual void lock();
    virtual void unlock();
};

/*!
    \brief Helper class for automatic unlocking of locking object in destructor
    \author VooDooMan
    \version 1
    \date 2007-07
*/
class c_mutex_locker
{
    c_mutex* m_m;               //!< Locking object

    int m_fd;                   //!< File descriptor
public:
    c_mutex_locker(c_mutex& m);
    c_mutex_locker(c_mutex* m);
    ~c_mutex_locker();

    void lock_file_if_needed(const char* fn);
};

#endif // MUTEX_H__

