/***************************************************************************
                          sem.h  -  description
                             -------------------
    begin                : Sun Sep 2 2007
    copyright            : (C) 2007 by Marian VooDooMan Meravy
    email                : ghostvoodooman -a.T= users (dot) sourceforge.net
 ***************************************************************************/

/* $Id: sem.h,v 1.12 2008/03/11 21:11:48 ghostvoodooman Exp $ */

/***************************************************************************

VooDoo cIRCle - an IRC (ro)bot
Copyright (C) 2005-2008 by Marian VooDooMan Meravy
  (ghostvoodooman -a.T= users (dot) sourceforge.net)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************/

/*!
    \file
    \brief Implements mutex semaphore in mmap'ed file
    \version $Id: sem.h,v 1.12 2008/03/11 21:11:48 ghostvoodooman Exp $
*/

/*

CREDITS:

http://www.llnl.gov/computing/tutorials/pthreads/man/pthread_mutexattr_init.txt

COPYRIGHT
       Portions of this text are reprinted and reproduced in electronic  form  from  IEEE
       Std  1003.1, 2003 Edition, Standard for Information Technology -- Portable Operat-
       ing System Interface (POSIX), The Open Group Base Specifications  Issue  6,  Copy-
       right  (C) 2001-2003 by the Institute of Electrical and Electronics Engineers, Inc
       and The Open Group. In the event of any discrepancy between this version  and  the
       original  IEEE  and  The Open Group Standard, the original IEEE and The Open Group
       Standard is the referee document. The original Standard can be obtained online  at
       http://www.opengroup.org/unix/online.html .
*/

#ifndef SEM_H__
#define SEM_H__

#if defined(HAVE_CONFIG_H)
#   include <config.h>
#endif

#include "../params.h"

#if !defined(_WIN32) && !defined(_WIN64) && !defined(__CYGWIN__)
#   ifdef USE_PTHREAD
#       if defined(HAVE_LIBPTHREAD) && defined(HAVE_MMAP)
#           include <pthread.h>
#           define MY_USE_SEMAPHORE     1
#       endif
#   else
#       ifdef HAVE_LIBPTHREAD
#           undef HAVE_LIBPTHREAD
#       endif
#   endif
#endif

#ifdef MY_USE_SEMAPHORE

/* sem.h */
struct semaphore {
  pthread_mutex_t lock;
  pthread_cond_t nonzero;
  unsigned count;
};
typedef struct semaphore semaphore_t;

#define SEM_LOCKING_REGION_SIZE         (4)

semaphore_t *semaphore_create(const char *semaphore_name, int& fd);
semaphore_t *semaphore_open(const char *semaphore_name, int& fd);
void semaphore_post(semaphore_t *semap);
bool semaphore_wait(semaphore_t *semap, unsigned long timeout = 0);
void semaphore_close(semaphore_t *semap, int& fd);

#endif  // MY_USE_SEMAPHORE

#endif  // SEM_H__

