/***************************************************************************
                          threads.h  -  description
                             -------------------
    begin                : Sat Nov 28 2005
    copyright            : (C) 2005-2007 by Marian VooDooMan Meravy
    email                : ghostvoodooman -a.T= users (dot) sourceforge.net
 ***************************************************************************/

/* $Id: threads.h,v 1.39 2008/03/11 21:11:32 ghostvoodooman Exp $ */

/***************************************************************************

VooDoo cIRCle - an IRC (ro)bot
Copyright (C) 2005-2008 by Marian VooDooMan Meravy
  (ghostvoodooman -a.T= users (dot) sourceforge.net)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************/

/*!
    \file
    \brief Implements threaded DNS resolution
    \version $Id: threads.h,v 1.39 2008/03/11 21:11:32 ghostvoodooman Exp $
*/

#ifndef THREADS_H__
#define THREADS_H__

#if defined(HAVE_CONFIG_H)
#   include <config.h>
#endif

#include <string>
#include <vector>
#include <stack>
#include <list>
#include <map>

#include "params.h"
#include "utils.h"
#include "sock_shared.h"

#include "utils/assert.h"

enum e_resolv_time
{
    RT_RESOLVE_NOW,                 //!< Resolve immediately
    RT_RESOLVE_SCHEDULE,            //!< Could be scheduled
};                                  //!< Type of request for resolution

enum e_resolv_result
{
    RR_OK,                          //!< Resolved okay
    RR_FAILURE,                     //!< Could not resolve
    RR_SCHEDULED,                   //!< Resolve has been scheduled to thread
};                                  //!< Result of resolution

typedef long threaded_connect_id;   //!< ID for threaded connect
/*!
    \def TCID_INVALID
    \brief Value for 'threaded_connect_id' type, indicating invalid ID
*/
#define TCID_INVALID    (0L)

e_resolv_result threads_resolve4(const char* domain, vector<in_addr>& result, e_resolv_time request_type);
e_resolv_result threads_resolve4(const char* domain, in_addr& result, e_resolv_time request_type);
e_resolv_result threads_reverse4(const char* address, char* dns_name, size_t dns_max_size, e_resolv_time request_type);
e_resolv_result threads_resolve6(const char* domain, vector<in_addr6_>& result, e_resolv_time request_type);
e_resolv_result threads_resolve6(const char* domain, in_addr6_& result, e_resolv_time request_type);
e_resolv_result threads_reverse6(in_addr6_ address, char* dns_name, size_t dns_max_size, e_resolv_time request_type);

long threads_connect(const string& localIP, const string& remote, unsigned short remote_port, const string& comment, bool async);
bool threads_get_connect(long id, int& ec, c_socket& s);

void threads_loop();

void threads_done();

void threads_flushdns();

#endif

