/***************************************************************************
                          utils.cpp  -  description
                             -------------------
    begin                : Thu Dec 8 2005
    copyright            : (C) 2005-2007 by Marian VooDooMan Meravy
    email                : ghostvoodooman -aT= users (dot) sourceforge.net
 ***************************************************************************/

/* $Id: utils.cpp,v 1.49 2008/08/23 15:10:42 ghostvoodooman Exp $ */

/***************************************************************************

VooDoo cIRCle - an IRC (ro)bot
Copyright (C) 2005-2008 by Marian VooDooMan Meravy
  (ghostvoodooman -aT= users (dot) sourceforge.net)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************/

/*!
    \file
    \brief Provides basic utils
    \version $Id: utils.cpp,v 1.49 2008/08/23 15:10:42 ghostvoodooman Exp $
*/

#if defined(HAVE_CONFIG_H)
#   include <config.h>
#endif

#ifndef _GNU_SOURCE
#   define _GNU_SOURCE
#endif

#include <string>
#include <vector>
#include <stack>
#include <list>
#include <map>
#include <sstream>
#include <iostream>

#include <stdio.h>
#include <stdarg.h>
#include <time.h>
#include <stdlib.h>
#include <errno.h>

using namespace std;

#include "utils/memleak.h"
#include "utils/assert.h"
#include "utils.h"
#include "params.h"

#include "match.h"
#include "md5.h"
#include "stats.h"
#include "thread/mutex.h"
#include "log.h"
#include "utils/my_getpid.h"
#include "eval.h"
#include "misc/shutdown.h"
#include "irc.h"

#include "utils/precomp.h"

#if defined(_MSC_VER) || defined(__BORLANDC__)
#   pragma hdrstop
#endif

c_mutex nice_time_offset_m;
static string nice_time_offset_old="+0000";
static string nice_time_offset="+0000";
static int nice_time_offset_to_add_secs=0;

static bool debug_report_suppress=false;

#if defined(_MSC_VER) && defined(_DEBUG)
void __cdecl LeakDumper(void *ptr, size_t sz)
{
    int block = _CrtReportBlockType(ptr);
    MY_RPT4(MY_RPT_INFO, "LeakDumper found block at %p:"
                     " type %d, subtype %d, size %s\n", ptr,
          _BLOCK_TYPE(block), _BLOCK_SUBTYPE(block), tos(sz).c_str());
}
#endif

string escape_string_for_logs_1(string str)
{
    string res;
    for(size_t i1=0; i1<str.length(); ++i1) {
        if(str[i1]>=32 && str[i1]!='\\') {
            res+=str[i1];
            continue;
        }
        switch(str[i1]) {
            case '\n':
                res+="\\n";
                break;
            case '\r':
                res+="\\r";
                break;
            case '\f':
                res+="\\f";
                break;
            case '\t':
                res+="\\t";
                break;
            case '\b':
                res+="\\b";
                break;
            case 0x07:
                res+="\\a";
                break;
            case '\\':
                res+="\\\\";
                break;
            case '\'':
                res+="\\'";
                break;
            case '\0':
                res+="\\000";
                break;
            case '"':
                res+="\\\"";
                break;
            default:
                {
                    char tmp[16];
                    unsigned x=static_cast<unsigned>(str[i1]);
                    sprintf(tmp,"\\x%2x",x);
                    res+=tmp;
                }
                break;
        }
    }
    return res;
}

/*!
    \brief Splits string by delimiter
    \author VooDooMan
    \version 1
    \date 2005
    \param src Source string
    \param delim Delimiter
    \param res Returns pieces of string
    \param accept_empties If to accept two delimiter and nothing in between them (an empty piece), if false, it is like they were just one
*/
string explode(const string& src, char delim, vector<string>& res, bool accept_empties)
{
    res.clear();
    bool first_arg=true;
    string first;
    string arg;
    bool in_esc=false;
    bool in_str=false;
    for(size_t i1=0; i1<src.length(); ++i1) {
        if(in_esc) {
            arg+=src[i1];
            in_esc=false;
            continue;
        }
        if('\\'==src[i1]) {
            arg+=src[i1];
            in_esc=true;
            continue;
        }
        if('"'==src[i1]) {
            arg+=src[i1];
            in_str=!in_str;
            continue;
        }
        if(in_str) {
            arg+=src[i1];
            continue;
        }
        if(src[i1]==delim) {
            if(!arg.empty() || accept_empties) {
                res.push_back(arg);
                if(first_arg) {
                    first=arg;
                    first_arg=false;
                }
                arg="";
            }
            continue;
        }
        arg+=src[i1];
    }
    if(!arg.empty())
        res.push_back(arg);
    if(first_arg)
        return arg;
    return first;
}

/*!
    \brief Splits string by delimiter
    \author VooDooMan
    \version 1
    \date 2006-06-06
    \param str Source string
    \param separator Delimiter
    \param res Returns pieces of string
*/
void explode(const string& str, const string& separator, vector<string>& res)
{
    res.clear();
    if(separator.length()==0)
        return;
    string last_str;
    int got=1;
    char* separ=(char*)my_malloc(separator.length()+16);
    if(!separ)
        return;
    memset(separ,0,separator.length()+16);
    strcpy(separ,separator.c_str());
    char* buff=(char*)my_malloc(str.length()+separator.length()+16);
    if(!buff) {
        my_free(separ);
        return;
    }
    memset(buff,0,str.length()+separator.length()+16);
    memcpy(buff,str.c_str(),str.length());
    char* start=buff;
    size_t len=str.length();
    char* eob=start+len;
    while(start<=eob) {
        char* end=NULL;
        if(start<eob)
            end=strstr(start,separ);
        if(NULL==end) {
            char* b=MY_NEW char[strlen(start)+1];
            memcpy(b,start,strlen(start));
            b[strlen(start)]='\0';
            last_str=b;
            MY_DELETE_A(b);
            b=NULL;
            break;
        }
        char* b=MY_NEW char[(size_t)(end-start)+1];
        memcpy(b,start,end-start);
        b[(size_t)(end-start)]='\0';
        last_str+=b;
        MY_DELETE_A(b);
        b=NULL;
        start=end+strlen(separ);
        {
            got++;
            res.push_back(last_str);
            last_str="";
        }
    }
    if(last_str.length()) {
        res.push_back(last_str);
    }
    my_free(separ);
    my_free(buff);
    return;
}

/*!
    \brief Kills trailing EOL(s) in the string
    \param str String to process
    \warning Modifies "str" argument
*/
void str_kill_trailing_eol(char* str)
{
    while(str[0] && str[strlen(str)-1]=='\n')
        str[strlen(str)-1]=0;
    while(str[0] && str[strlen(str)-1]=='\r')
        str[strlen(str)-1]=0;
    while(str[0] && str[strlen(str)-1]=='\n')
        str[strlen(str)-1]=0;
    while(str[0] && str[strlen(str)-1]=='\r')
        str[strlen(str)-1]=0;
}

/*!
    \brief Returns text representation of time (without years and months, only number of days)
    \author VooDooMan
    \version 2
    \date 2005-11-26, 2007-07-30
    \param src Time to convert
    \param days_word The word "days, "
    \return Returns text, must be MY_DELETE_A'd
*/
string make_nice_time_days_only(my_time_t src, const char* days_word)
{
    MY_U_LONG_LONG days=src/(60*60*24);
    src-=days*(60*60*24);

    time_t src2=static_cast<time_t>(src);

    tm* t=gmtime(&src2);
    char tmp[128];
    strftime(tmp,sizeof(tmp)-1,"%H:%M:%S",t);
    char res[1024];
    if(snprintf(res,1024-1,"%s%s%s",tos(days).c_str(),days_word,tmp)<0)
        res[1024-1]='\0';
    return (string)res;
}

/*!
    \brief Sets time offset
    \author VooDooMan
    \version 2
    \date 2006-06-18, 2008-06-08
    \param timezone Time offset in format "+HHMM"
*/
void nice_time_set_time_offset(const string& timezone)
{
    MY_ASSERT(timezone.length()==sizeof("+HHMM")-1);
    if(timezone.length()!=sizeof("+HHMM")-1)
        return; // bail out

    c_mutex_locker lock(nice_time_offset_m);

    if(timezone.compare(nice_time_offset_old)) { // !=
        char tmp[128];

        sprintf(tmp,"Changing timezone to %s",timezone.c_str());
        log_all_logs(tmp);
        MY_RPT0(MY_RPT_INFO,tmp);

        nice_time_offset=timezone;

        nice_time_offset_old=nice_time_offset;

        // "+HHMM"
        //  01234
        nice_time_offset_to_add_secs=
            ( ((int)timezone[1] - (int)'0') * 10 + (int)timezone[2] - (int)'0') * 60 * 60 +
            ( ((int)timezone[3] - (int)'0') * 10 + (int)timezone[4] - (int)'0') * 60;
        if('-'==timezone[0])
            nice_time_offset_to_add_secs=-nice_time_offset_to_add_secs;

        sprintf(tmp,"Changed timezone to %s",timezone.c_str());
        log_all_logs(tmp);
        MY_RPT0(MY_RPT_INFO,tmp);
    }
}

/*!
    \brief Report time offset to logs, on startup
    \author VooDooMan
    \version 2
    \date 2007-07-31, 2008-06-08
*/
string utils_get_timezone()
{
    c_mutex_locker lock(nice_time_offset_m);

    return nice_time_offset;
}

/*!
    \brief Returns text representation of time
    \author VooDooMan
    \version 2
    \date 2005-11-19, 2006-11-17
    \param src Time to convert
    \param brackets If to enclose it into brackets
    \param year_substract1970 If to substract 1970 from year
    \param for_filename If set to 'true', space is replaced by '_', and ':' by '-'
*/
string make_nice_time(my_time_t src, bool brackets, bool year_substract1970, bool for_filename)
{
    c_mutex_locker lock(nice_time_offset_m);

    static char nice_time[128];
    my_time_t tt=src+nice_time_offset_to_add_secs;
    time_t ttt=static_cast<time_t>(tt);
    tm* t=gmtime(&ttt);
    if(!t) {
        ttt=0;
        t=gmtime(&ttt); // invalid time: before UNIX epoch (negative value)
    }
    if(year_substract1970)
        t->tm_year-=1970;
    static const char spc_brc[]="[%Y-%m-%dT%H:%M:%S";
    static const char spc_nobrc[]="%Y-%m-%dT%H:%M:%S";
    static const char ffn_brc[]="[%Y-%m-%dT%H-%M-%S";
    static const char ffn_nobrc[]="%Y-%m-%dT%H-%M-%S";
    const char* fmt=spc_brc;
    if(brackets && for_filename)
        fmt=ffn_brc;
    if(brackets && !for_filename)
        fmt=spc_brc;
    if(!brackets && for_filename)
        fmt=ffn_nobrc;
    if(!brackets && !for_filename)
        fmt=spc_nobrc;
    strftime(nice_time,sizeof(nice_time)-1,fmt,t);
    nice_time[sizeof(nice_time)-1]='\0';
    string res=nice_time;
    res+=nice_time_offset;
    if(brackets)
        res+="]";

    return res;
}

/*!
    \brief On win32, convert slashes to backslashes, on UNIX it's harmless. Also removes redundant slashes (like "abc\\\\def", or "./abc//def///")
    \author VooDooMan
    \version 2
    \date 2005-12-23, 2006-01-12
    \param path Source path
    \return Returns converted path
    \todo Should be able to convert '\\' as escape sequence
*/
string utils_convert_unix_path(const string &path)
{
    ostringstream new_path;
    for(size_t i=0; i<path.length(); ++i) {
        if('\\'==path[i] || '/'==path[i]) {
            new_path << "/";
            continue;
        }
        new_path.put(path[i]);
    }
    string new_path2=new_path.str();
#if defined(_WIN32) || defined(_WIN64)
    for(string::size_type i=0; i<new_path2.length(); ++i)
        if('/'==new_path2[i])
            new_path2[i]='\\';
#endif
    return new_path2;
}

/*!
    \brief Simple returns string for preprocessor macros
    \version 1
    \author VooDooMan
    \date 2006-01-19
    \param str String to return
*/
string return_string(string str)
{
    return str;
}

uint32 swap_bytes(uint32 data)
{
    unsigned char* x=(unsigned char*)&data;
    unsigned char tmp;
    tmp=x[0];
    x[0]=x[3];
    x[3]=tmp;
    tmp=x[2];
    x[2]=x[1];
    x[1]=tmp;
    return data;
}

/*!
    \brief Converts decimal-integer-encoded IPv4 address to string
    \param ip decimal-integer-encoded IPv4 address
    \version 1
    \author VooDooMan
    \date 2006-08-26
*/
string long2ip(uint32 ip)
{
    unsigned char* x=(unsigned char*)&ip;
    byteReverse(x,1);  // does nothing on BIG endian, swaps ONLY on LITTLE endian
    ip=swap_bytes(ip); // swaps ALWAYS

    char tmp[128];
    my_inet_ntop(AF_INET,&ip,tmp,sizeof(tmp)-1);
    return (string)tmp;
}

/*!
    \brief Converts IPv4 string to decimal-integer-encoded IPv4 address
    \param ip IPv4 string
    \version 1
    \author VooDooMan
    \date 2006-08-26
*/
uint32 ip2long(const string& ip)
{
    uint32 res=0;
    if(my_inet_pton(AF_INET,ip.c_str(),&res)!=1)
        return 0; // error

    unsigned char* x=(unsigned char*)&res;
    byteReverse(x,1);       // does nothing on BIG endian, swaps ONLY on LITTLE endian
    res=swap_bytes(res);    // swaps ALWAYS

    return res;
}

#if defined(_MSC_VER) && DEBUG_TIME_MEASUREMENT != 1
#   pragma warning( push )
#   pragma warning( disable : 4100 ) // warning C4100: 'sleep_ms' : unreferenced formal parameter
#endif

double debug_timer_end(clock_t diff, const char* file, const char* func, int begin_line, int end_line, const char* debug, unsigned sleep_ms)
{
#if 1==DEBUG_TIME_MEASUREMENT
    double x=diff;
    x/=CLOCKS_PER_SEC;
    char* tmp=MY_NEW char[strlen(file)+strlen(func)+strlen(debug)+128];
    sprintf(tmp,"%lf seconds (%s) for %s(%d-%d)::%s()",x,debug,file,begin_line,end_line,func);
    MY_RPT1(MY_RPT_INFO,"Debug time measurement: %s",tmp);
    MY_DELETE_A(tmp);
    if(sleep_ms)
        my_sleep(sleep_ms);
    return x;
#else
    return 0;
#endif
}

#if defined(_MSC_VER) && DEBUG_TIME_MEASUREMENT != 1
#   pragma warning( pop )
#endif

c_mutex util_debug_report_file_m;
FILE* util_debug_report_file=NULL;              //!< File handle for debug reports
string util_debug_file_timestamp;               //!< Timestamp for creating unique debug report log filename
#ifdef _MSC_VER
HANDLE util_debug_report_file_msc=NULL;         //!< File handle for MSVC's debug reports
int util_my_report_hook(int rpt_type, char *msg, int *ret_val);
#endif
bool utils_has_msvc_dbg_hook=false;

#ifdef _DEBUG

/*!
    \brief Constructs a filename for verbose debug log
    \author VooDooMan
    \version 1
    \date 2008-02-26
*/
MY_STATIC string utils_create_debug_report_filename()
{
    ostringstream fn;
    fn << "." FILE_SLASH "logs" FILE_SLASH "debug.";
    fn << tos(my_getpid()).c_str() << ".";
    fn << util_debug_file_timestamp.c_str();
    fn << ".log";

    return fn.str();
}

#endif  // _DEBUG

/*!
    \brief Opens file for debug reports
    \author VooDooMan
    \version 1
    \date 2006-11-15
    \return Returns timestamp as string
*/
string util_debug_report_init()
{
    util_debug_file_timestamp=make_nice_time(my_time(NULL),false,false,true);
#ifndef _DEBUG
    return util_debug_file_timestamp;
#else
    if(util_debug_report_file)
        return util_debug_file_timestamp;

#if (defined(_WIN32) || defined(_WIN64)) && !defined(__CYGWIN__)
    mkdir("." FILE_SLASH "logs"); // just to be sure; don't care about error returned
#else
    mkdir("." FILE_SLASH "logs",0777); // just to be sure; don't care about error returned
#endif

    util_debug_report_file=fopen(utils_create_debug_report_filename().c_str(),"a");

#ifndef SETVBUF_REVERSED
    if(util_debug_report_file)
        setvbuf(util_debug_report_file,NULL,_IONBF,0); // don't care about return code
#else
    // System V before release 3
    if(util_debug_report_file)
        setvbuf(util_debug_report_file,_IONBF,NULL,0); // don't care about return code
#endif

#ifdef _MSC_VER
    // install hook instead
    _CrtSetReportHook(util_my_report_hook);
    utils_has_msvc_dbg_hook=true;

    util_debug_report_sync(); // important!

    HANDLE hnd=util_debug_report_file_msc;
    if(!util_debug_report_file_msc)
        hnd=_CRTDBG_FILE_STDERR; // fallback to stderr

    _CrtSetReportFile(_CRT_WARN,hnd);
    _CrtSetReportFile(_CRT_ERROR,hnd);
    _CrtSetReportFile(_CRT_ASSERT,hnd);
    MSVC_SET_DEBUG_OUTPUTS(true);
#endif

    MY_RPT1(MY_RPT_INFO,"Time: %s",make_nice_time(my_time(NULL),false,false).c_str());
    MY_RPT1(MY_RPT_INFO,"BOT STARTUP, pid %s",tos(my_getpid()).c_str());

    return util_debug_file_timestamp;
#endif // _DEBUG
}

/*!
    \brief Closes file for debug reports
    \author VooDooMan
    \version 1
    \date 2006-11-15
    \param rc Return code
    \return Returns return code passed as argument 'rc'
    \bug Does NOT free debug files descriptors, because memory dump after exit may follow
*/
int util_debug_report_done(int rc)
{
    util_debug_report_sync(); // for sure's sake

    MY_RPT1(MY_RPT_INFO,"Time: %s",make_nice_time(my_time(NULL),false,false).c_str());

    MY_RPT1(MY_RPT_INFO,"CPU time: %s seconds",tos(stats_cpu_time()).c_str());

    string tmp=make_nice_time_days_only(stats_cpu_time()," day(s), ");
    MY_RPT1(MY_RPT_INFO,"CPU time: %s",tmp.c_str());

    MY_RPT1(MY_RPT_INFO,"Uptime: %s seconds",tos(stats_uptime()).c_str());

    tmp=make_nice_time_days_only(stats_uptime()," day(s), ");
    MY_RPT1(MY_RPT_INFO,"Uptime: %s",tmp.c_str());

    MY_RPT1(MY_RPT_INFO,"EXIT with return code %s",tos(rc).c_str());

    util_debug_report_sync(); // for sure's sake

    if(util_debug_report_file) {
        fclose(util_debug_report_file);
        util_debug_report_file=NULL;
    }

#ifdef _MSC_VER
    // invalidate (to be) closed handle

    // Do NOT free debug files descriptors, because memory dump after exit may follow
    /*
    _CrtSetReportFile(_CRT_WARN,_CRTDBG_FILE_STDERR);
    _CrtSetReportFile(_CRT_ERROR,_CRTDBG_FILE_STDERR);
    _CrtSetReportFile(_CRT_ASSERT,_CRTDBG_FILE_STDERR);

    if(util_debug_report_file_msc) {
        CloseHandle(util_debug_report_file_msc);
        util_debug_report_file_msc=NULL;
    }
    */
#endif

    return rc;
}

/*!
    \brief Flushes both file descriptor and Windows(TM) file handle (synchronisation)
    \author VooDooMan
    \version 1
    \date 2006-11-15
    \note SHOULD be called before and after any write!
*/
void util_debug_report_sync()
{
#ifndef _DEBUG
    return;
#else

#ifdef _MSC_VER
    if(util_debug_report_file && util_debug_report_file_msc) {
        // only if both of them are open

        SetFilePointer(util_debug_report_file_msc,0,NULL,FILE_END);
        FlushFileBuffers(util_debug_report_file_msc);

        int my_errno;
        my_fseek(util_debug_report_file,0,SEEK_END,my_errno);
        fflush(util_debug_report_file);
    }
#else
    if(util_debug_report_file)
        fflush(util_debug_report_file);
#endif  // _MSC_VER
#endif  // _DEBUG
}

/*!
    \brief Closes file for debug reports, or creates new one (for log rotation)
    \author VooDooMan
    \version 1
    \date 2006-12-25
    \param close 'true' for close log file, 'false' to open new one
*/
void util_debug_report_close(bool close)
{
    if(close) {
        MY_RPT1(MY_RPT_INFO,"Time: %s",make_nice_time(my_time(NULL),false,false).c_str());
        MY_RPT0(MY_RPT_INFO,"Switching logs");
        util_debug_report_sync(); // important!
        if(util_debug_report_file)
            fclose(util_debug_report_file);
        util_debug_report_file=NULL;
    } else {
        util_debug_report_file=NULL;
#ifdef _DEBUG
        util_debug_report_file=fopen(utils_create_debug_report_filename().c_str(),"a");

#ifndef SETVBUF_REVERSED
        if(util_debug_report_file)
            setvbuf(util_debug_report_file,NULL,_IONBF,0); // don't care about return code
#else
        // System V before release 3
        if(util_debug_report_file)
            setvbuf(util_debug_report_file,_IONBF,NULL,0); // don't care about return code
#endif

        util_debug_report_sync(); // important!

        MY_RPT1(MY_RPT_INFO,"Time: %s",make_nice_time(my_time(NULL),false,false).c_str());
        MY_RPT0(MY_RPT_INFO,"Logs were switched");
        util_debug_report_sync();
#endif
    }
}

int util_my_report_hook(int rpt_type, char *msg, int *ret_val)
{
    *ret_val=0;

    if(util_debug_report_file) {
        string nt=make_nice_time(my_time(NULL),true);
        fprintf(util_debug_report_file,"%s %s",nt.c_str(),msg);
    }

#if defined(_CRT_ASSERT)
    if(_CRT_ASSERT==rpt_type) {
        if(util_debug_report_file)
            fflush(util_debug_report_file);
        cout << "[ASSERT] " << msg << endl;

#ifdef _DEBUG
        // DEBUG: Set breakpoint here!
        int x=0;
        x++;
#endif

        *ret_val=1;
        return TRUE;
    }
#endif

    return 0;
}

string str_to_upper(string str)
{
    size_t s=str.length();

    for(register size_t i1=0; i1<s; ++i1) {
        if(((unsigned int)str[i1])<256)
            str[i1]=(char)touppertab[(unsigned char)str[i1]];
    }
    return str;
}

string str_to_lower(string str)
{
    size_t s=str.length();

    for(register size_t i1=0; i1<s; ++i1) {
        if(((unsigned int)str[i1])<256)
            str[i1]=(char)tolowertab[(unsigned char)str[i1]];
    }
    return str;
}

bool cmp_strings_case_insensitive(string s1, string s2)
{
    if(s1.length()!=s2.length())
        return true;
    s1=str_to_upper(s1);
    s2=str_to_upper(s2);
    return s1.compare(s2) ? true : false;
}

bool cmp_strings_case_insensitive(const char* s1, const char* s2, size_t size)
{
    if(!size)
        return false;
    while(size) {
        if(touppertab[static_cast<unsigned char>(*(s1++))]!=touppertab[static_cast<unsigned char>(*(s2++))])
            return true;
        size--;
    }
    return false;
}

/*!
    \brief Debugging reports
    \author VooDooMan
    \version 1
    \date 2007-07-30
    \param severity Severity ('MY_RPT_INFO' or 'MY_RPT_ERROR')
    \param fl File name of caller (__FILE__)
    \param fn Function name of caller (__my_func__)
    \param ln Line of caller (__LINE__)
    \param fmt Format string as for 'printf()', followed by optional arguments
*/
void utils_debug_report(int severity, const char* fl, const char* fn, int ln, const char* fmt, ...)
{
    va_list marker;
    va_start(marker,fmt);

    if(debug_report_suppress) {
        va_end(marker);
        return;
    }

    string file=make_nice_time(my_time(NULL),true);
    file+=" ";

    string con;
    if(MY_RPT_ERROR==severity)
        con="[ERRO] ";
    else
        con="[INFO] ";
    file+=con;

#ifdef _MSC_VER
#   define VSNPRINTF    _vsnprintf
#else
#   define VSNPRINTF    vsnprintf
#endif

    char tmp[1024];
    if(VSNPRINTF(tmp,1024-1,fmt,marker)<0)
        tmp[1024-1]='\0';

    string loc=" (in ";
    loc+=fl;
    loc+="::";
    loc+=fn;
    loc+="():";
    loc+=tos(ln);
    loc+=")";

    con+=tmp;
    file+=tmp;

    con+=loc;
    file+=loc;

    {
        c_mutex_locker lock(util_debug_report_file_m);

#if !defined(_MSC_VER) || !defined(_DEBUG)

        cout << con << endl;
#else
        string con2=con+"\n";
        _RPT0(MY_RPT_ERROR==severity ? _CRT_ERROR : _CRT_WARN, con2.c_str());
#endif

        if(!utils_has_msvc_dbg_hook && util_debug_report_file) {
            util_debug_report_sync();
            fprintf(util_debug_report_file,"%s\n",file.c_str());
            util_debug_report_sync();
        }
    }

    MY_U_LONG_LONG evt_id=log_get_next_event_id();
    log_cumulative(evt_id,E_LOG_DEBUG_RPT,"",file);

    va_end(marker);
}

/*!
    \brief Suppresses debug reporting from now on
    \author VooDooMan
    \version 1
    \date 2007-08-03
*/
void util_debug_report_suppress()
{
    debug_report_suppress=true;
}

/*!
    \brief Reverses the buffer at bytes
    \author VooDooMan
    \version 1
    \date 2007-11-16
    \param buf Buffer
    \param cnt Byte count (must be even number)
    \bug Only works when 'cnt' is even number
*/
void byte_stream_reverse(void* buf, size_t cnt)
{
    MY_ASSERT(cnt%2 == 0);

    for(size_t i1=0, i2=cnt-1; i1<cnt/2; ++i1, --i2) {
        char tmp=reinterpret_cast<char*>(buf)[i1];
        reinterpret_cast<char*>(buf)[i1]=reinterpret_cast<char*>(buf)[i2];
        reinterpret_cast<char*>(buf)[i2]=tmp;
    }
}

/*!
    \brief Causes bot to sleep, while giving CPU time to process tasks such as DCC loop, or checking for bot interrup event (Ctrl+C, or SIGINT signal, and so on)
    \version 1
    \author VooDooMan
    \date 2007-12-03
    \param seconds Number of seconds to sleep
*/
void utils_deep_sleep(int seconds)
{
    if(seconds<=0)
        return;

    my_time_t start=my_time(NULL);

    for(;;) {
        eval_cleanup_timeout();
        do_idle();
        irc_loop_heavy_time_consuming_operation("Still sleeping...");

        e_shutdown_reason sr=bot_get_shutdown_state();
        if(E_SD_NONE==sr)
            my_sleep(50);
        else
            break;

        if(start+seconds<my_time(NULL))
            break;
    }

    do_idle();
    eval_cleanup_timeout();
}

my_time_t my_time(my_time_t* ptr)
{
    my_time_t res=time(NULL);
    if(ptr)
        *ptr=res;
    return res;
}

