/***************************************************************************
                          utils.cpp  -  description
                             -------------------
    begin                : Thu Dec 8 2005
    copyright            : (C) 2005-2007 by Marian VooDooMan Meravy
    email                : ghostvoodooman -a.T= users (dot) sourceforge.net
 ***************************************************************************/

/* $Id: utils.h,v 1.47 2008/11/24 12:09:23 ghostvoodooman Exp $ */

/***************************************************************************

VooDoo cIRCle - an IRC (ro)bot
Copyright (C) 2005-2008 by Marian VooDooMan Meravy
  (ghostvoodooman -a.T= users (dot) sourceforge.net)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************/

/*!
    \file
    \brief Provides basic utils
    \version $Id: utils.h,v 1.47 2008/11/24 12:09:23 ghostvoodooman Exp $
*/

#ifndef UTILS_H__
#define UTILS_H__

#if defined(HAVE_CONFIG_H)
#   include <config.h>
#endif

#include <string>
#include <vector>
#include <stack>
#include <list>
#include <map>

#if defined(HAVE_CONFIG_H)
#   include <config.h>
#endif

#include <time.h>
#include <stdio.h>

#include "params.h"

#include "utils/assert.h"
#include "compat.h"

using namespace std;

#include "compat.h"
#include "jenkins_hash.h"
#include "irc_bot.h"

extern unsigned char tolowertab[];
extern unsigned char tolowertab_rfc1459[];
extern unsigned char touppertab[];

#if 1==DEBUG_TIME_MEASUREMENT

#   define BEGIN_DEBUG_TIME_MEASUREMENT(enabled,id,comment) \
        /* MSVC workaround: warning C4127: conditional expression is constant */ \
        static bool __##id##_enabled__=enabled; \
        /* ...so don't let 'enabled' be const */ \
        static const char* __##id##_comment__=comment; \
        static const int __##id##_begin_line__=__LINE__; \
        clock_t __##id##_timer_begin__=clock();

#   define END_DEBUG_TIME_MEASUREMENT(id,wait_ms) \
        clock_t __##id##_timer_end__=clock(); \
        if(__##id##_enabled__) \
            debug_timer_end(__##id##_timer_end__ - __##id##_timer_begin__,__FILE__,__my_func__,__##id##_begin_line__,__LINE__,__##id##_comment__,wait_ms);

#else

#   define BEGIN_DEBUG_TIME_MEASUREMENT(enabled,id,comment) /* disabled */
#   define END_DEBUG_TIME_MEASUREMENT(id,wait_ms) /* disabled */

#endif

string escape_string_for_logs_1(string str);
string explode(const string& src, char delim, vector<string>& res, bool accept_empties = false);
void explode(const string& str, const string& separator, vector<string>& res);

void str_kill_trailing_eol(char* str);

typedef MY_LONG_LONG my_time_t;
my_time_t my_time(my_time_t* ptr = NULL);

void nice_time_set_time_offset(const string& timezone);
string utils_get_timezone();
string make_nice_time_days_only(my_time_t src, const char* days_word);
string make_nice_time(my_time_t src, bool brackets = false, bool year_substract1970 = false, bool for_filename = false);

/*!
    \brief Provides execution time measuring for debugging (end of measuring)
    \author VooDooMan
    \param diff Difference between start of code and end of measured code (clock() at begin - clock() at end)
    \param func Debug string #1 to print (stdout)
    \param debug Debug string #2 to print (stdout)
    \param sleep_ms Miliseconds to wait (to notice message on screen)
    \param file File name
    \param begin_line Line number of beginning of measurement
    \param end_line Line number of end of measurement
    \return Returns real number of seconds elapsed in our measuring
*/
double debug_timer_end(clock_t diff, const char* file, const char* func, int begin_line, int end_line, const char* debug, unsigned sleep_ms = DEBUG_TIME_MEASUREMENT_WAIT*1000);

string utils_convert_unix_path(const string &path);

/*!
    \brief Converts string to UPPER case (using touppertab[] in match.cpp)
    \author VooDooMan
    \version 4
    \date 2004, 2005-10-02, 2006-01-13 (Friday), 2007-03-13
    \param str Source string
    \return Returns UPPERcase string (using touppertab[] in match.cpp)
*/
string str_to_upper(string str);

/*!
    \brief Converts string to lower case (using tolowertab[] in match.cpp)
    \author VooDooMan
    \version 3
    \date 2005-10-02, 2006-01-13 (Friday), 2007-03-13
    \param str Source string
    \return Returns lowercase string (using tolowertab[] in match.cpp)
*/
string str_to_lower(string str);

/*!
    \brief Compares two strings, case in-sensitive
    \author VooDooMan
    \version 2
    \date 2004, 2006-01-13 (Friday)
    \param s1 String 1
    \param s2 String 2
    \retval false Strings are the same (case in-sensitive)
    \retval true Strings are different (case in-sensitive)
*/
bool cmp_strings_case_insensitive(string s1, string s2);

/*!
    \brief Compares two strings, case in-sensitive
    \author VooDooMan
    \version 1
    \date 2006-10-24
    \param s1 String 1
    \param s2 String 2
    \param size Number of bytes to be compared
    \retval false Strings are the same (case in-sensitive)
    \retval true Strings are different (case in-sensitive)
*/
bool cmp_strings_case_insensitive(const char* s1, const char* s2, size_t size);

/*!
    \brief Class encapsulating STL string used for case-insensitive comparsion (sorting)
    \author VooDooMan
    \date 2006-01-13 (Friday)
    \version 1
*/
class c_string
{
public:
    string str;     //! String

    /*!
        \brief Comparsion operator (used by STL's algorithms to sort)
        \author VooDooMan
        \date 2006-01-13 (Friday)
        \version 1
    */
    friend bool operator<(const c_string& left, const c_string& right)
    {
        size_t left_len=left.str.length();
        size_t right_len=right.str.length();
        for(register size_t i1=0; i1<left_len && i1<right_len; ++i1) {
            register char ll=tolowertab[((unsigned)left.str[i1])];
            register char rr=tolowertab[((unsigned)right.str[i1])];
            if(ll<rr)
                return true;
            else
                if(ll>rr)
                    return false;
        }
        if(left_len<right_len)
            return true;
        else
            if(left_len>right_len)
                return false;
        // they are same...
        // The question is:
        //     Is myself LESS that other? NO! We are the same!
        // So return false
        return false;
    }

    /*!
        \brief Comparsion operator (used by STL's algorithms to unique)
        \author VooDooMan
        \date 2006-01-13 (Friday)
        \version 1
    */
    friend bool operator==(const c_string& left, const c_string& right)
    {
        if(left.str.length()!=right.str.length())
            return false;

        size_t left_len=left.str.length();
        size_t right_len=right.str.length();
        for(register size_t i1=0; i1<left_len && i1<right_len; ++i1) {
            if(
                tolowertab[((unsigned)left.str[i1])]
                !=
                tolowertab[((unsigned)right.str[i1])]
               )
                    return false;
        }
        // they are same...
        return true;
    }

    /*!
        \brief Constructor
        \author VooDooMan
        \date 2006-01-13 (Friday)
        \version 1
        \param set String to initialize class with
    */
    c_string(const string& set)
    {
        str=set;
    }

    /*!
        \brief Copy-constructor
        \author VooDooMan
        \date 2006-01-13 (Friday)
        \version 1
        \param orig Original value
    */
    c_string(const c_string& orig)
    {
        str=orig.str;
    }

    /*!
        \brief Comparsion of strings in STL style
        \author VooDooMan
        \date 2006-01-13 (Friday)
        \version 1
        \param other To compare with
    */
    int compare(const c_string& other)
    {
        return str.compare(other.str);
    }
};

string return_string(string str);

#define UNUSEDARG(arg)              /* unused argument */

#if defined(_WIN32) || defined(_WIN64)
#   define UNUSEDARG_WIN(arg)     /* unused argument on _WIN32 */
#else
#   define UNUSEDARG_WIN(arg)     arg
#endif

#ifndef IPv6
#   define UNUSEDARG_NO_IPv6(arg)   /* unused argument: IPv6 is not beeing compiled */
#else
#   define UNUSEDARG_NO_IPv6(arg)   arg
#endif

#ifndef WORDS_BIGENDIAN
#   define UNUSEDARG_NO_BIGENDIAN(arg)   /* unused argument: the machine is little-endian */
#else
#   define UNUSEDARG_NO_BIGENDIAN(arg)   arg
#endif

string long2ip(uint32 ip);
uint32 ip2long(const string& ip);
uint32 swap_bytes(uint32 data);

/*!
    \brief Stores IPv6 address
    \author VooDooMan
    \version 1
    \date 2004
*/
struct in_addr6_
{
    unsigned char bytes[16];            //!< 16 bytes of IPv6 address
    MY_FORCE_INLINE bool operator< (const in_addr6_& than) const
    {
        for(size_t i1=0; i1<16; ++i1) {
            if(bytes[i1]<than.bytes[i1])
                return true;
        }
        return false;
    }
    MY_FORCE_INLINE bool operator== (const in_addr6_& as) const
    {
        for(size_t i1=0; i1<16; ++i1) {
            if(bytes[i1]!=as.bytes[i1])
                return false;
        }
        return true;
    }
};

#if defined(HAVE_EXT_HASH_MAP) || defined(HAVE_GNUCXX_HASHMAP)
#   if defined(WANT_HASHMAP) && WANT_HASHMAP == 0
#   else
#       include <ext/hash_map>
using __gnu_cxx::hash_multimap;
using __gnu_cxx::hash_map;
#       define MY_USE_HASH_MAP      1
#   endif
#endif

#if defined(_MSC_VER)
#   include <hash_map>
using stdext::hash_multimap;
using stdext::hash_map;
#   if _MSC_VER>=1400
using namespace stdext;
#   endif
#   define MY_USE_HASH_MAP      1
#endif

#if defined(__BORLANDC__)
#   include <Stlport/hash_map>
#   define MY_USE_HASH_MAP      1
#endif

#if defined(MY_USE_HASH_MAP)
#   if defined(_MSC_VER)
class string_hasher : public hash_compare<string> {
#   else
class string_hasher {
#   endif
public:
    MY_FORCE_INLINE size_t operator() (const string& s) const
    {
        return (size_t)jenkins_hash(reinterpret_cast<const unsigned char*>(s.c_str()),s.length(),0);
    }

    MY_FORCE_INLINE bool operator() (const string& s1, const string& s2) const
    {
        return s1<s2;
    }
};

#   if defined(_MSC_VER)
class inet6_hasher : public hash_compare<in_addr6_> {
#   else
class inet6_hasher {
#   endif
public:
    MY_FORCE_INLINE size_t operator() (const in_addr6_& s) const
    {
        return (size_t)jenkins_hash(reinterpret_cast<const unsigned char*>(&s),sizeof(s),0);
    }

    MY_FORCE_INLINE bool operator() (const in_addr6_& s1, const in_addr6_& s2) const
    {
        return s1<s2;
    }
};
#endif

#define TO_BOOL(x) \
    ( ((x)!=0) ? true : false )

#define BOOL2INT(x) \
    ( (int) ( (x) ? 1 : 0 ) )

#define my_difftime(time1,time0)    ( static_cast<MY_LONG_LONG>(time1) - static_cast<MY_LONG_LONG>(time0) )

string util_debug_report_init();
int util_debug_report_done(int rc = 0);
void util_debug_report_sync();
void util_debug_report_lock(bool lock_);
void util_debug_report_close(bool close);
void util_debug_report_suppress();

#include "utils/numconv.hh"

void byte_stream_reverse(void* buf, size_t cnt);
void utils_deep_sleep(int seconds);

#define GET_STR_DEBUG_INFO__HELPER \
    ((return_string(__FILE__)+":"+tos(__LINE__)+":"+return_string(__my_func__)+"()").c_str())

#define GET_STR_DEBUG_INFO      GET_STR_DEBUG_INFO__HELPER

#define GET_STR_DEBUG_INFO_EX(msg_prefix__)   \
    ((((string)msg_prefix__)+"; Debugging info: "+GET_STR_DEBUG_INFO__HELPER).c_str())

#endif

