/***************************************************************************
                          assert.cpp  -  description
                             -------------------
    begin                : Thu Dec 28 2006
    copyright            : (C) 2006-2007 by Marian VooDooMan Meravy
    email                : ghostvoodooman -aT= users (dot) sourceforge.net
 ***************************************************************************/

/* $Id: assert.cpp,v 1.14 2008/03/11 21:11:48 ghostvoodooman Exp $ */

/***************************************************************************

VooDoo cIRCle - an IRC (ro)bot
Copyright (C) 2004-2008 by Marian VooDooMan Meravy
  (ghostvoodooman -aT= users (dot) sourceforge.net)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************/

/*!
    \file
    \brief Provides assert for debugging
    \version $Id: assert.cpp,v 1.14 2008/03/11 21:11:48 ghostvoodooman Exp $
*/

#if defined(HAVE_CONFIG_H)
#   include <config.h>
#endif

#include "memleak.h"    // fix for FreeBSD 6.2

#include <stddef.h>
#include <iostream>

#include "assert.h"

#include "../mdump/mdump.h"
#include "../log.h"

#include "precomp.h"

#if defined(_MSC_VER) || defined(__BORLANDC__)
#   pragma hdrstop
#endif

/*!
    \brief Forces to dump the core
    \version 2
    \author VooDooMan
*/
int dump_core()
{
#if !defined(_MSC_VER)
    char* x=NULL;
    *x=0;
#else
    __try {
        char* x=NULL;
        *x=0;
    }__except(MiniDumper::TopLevelFilter(GetExceptionInformation())) {
        ;
    }
#endif
    return 1;
}

/*!
    \brief Reports assertion failure to logs, when core dump via NULL pointer dereference is not enabled
    \version 1
    \date 2008-02-26
    \author VooDooMan
    \param file Name of C++ file
    \param func Name of C++ function
    \param line Line number in C++ file
    \param exp Failed expression
*/
int report_assertion_failure(const char* file, const char* func, int line, const char* exp)
{
    static MY_U_LONG_LONG cnt=0;
    ++cnt;

    ostringstream log;

    log << "FATAL: Assertion failure #" << tos(cnt).c_str()
        << " in file \"" << file << "\""
        << " in function \"" << func << "\" on line number "
        << tos(line).c_str() << ". Failed expression was \""
        << exp << "\". Please report this problem! "
        << "The bot may be in unstable state, restart is recommended.";

    log_bot(log.str(),true);
    log_debug(log.str(),true);

    cout << log.str() << endl;
    cout.flush();

    return 0;
}

