/***************************************************************************
                          memleak.h  -  description
                             -------------------
    begin                : Fri May 4 2007
    copyright            : (C) 2005-2007 by Marian VooDooMan Meravy
    email                : ghostvoodooman -aT= users (dot) sourceforge.net
 ***************************************************************************/

/* $Id: memleak.h,v 1.16 2008/11/24 12:09:46 ghostvoodooman Exp $ */

/***************************************************************************

VooDoo cIRCle - an IRC (ro)bot
Copyright (C) 2004-2008 by Marian VooDooMan Meravy
  (ghostvoodooman -aT= users (dot) sourceforge.net)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************/

/*!
    \file
    \brief Provides memory leak detection
    \version $Id: memleak.h,v 1.16 2008/11/24 12:09:46 ghostvoodooman Exp $
*/

#ifndef MEMLEAK_H__
#define MEMLEAK_H__

#if defined(HAVE_CONFIG_H)
#   include <config.h>
#endif

#include "../params.h"

#include "../memory/heap_debug.h"

#include "../utils/assert.h"

using namespace std;

#ifdef _MSC_VER
#   ifdef _DEBUG
#   include <crtdbg.h>
#       define SET_CRT_DEBUG_FIELD(a) \
            _CrtSetDbgFlag((a) | _CrtSetDbgFlag(_CRTDBG_REPORT_FLAG))
#       define CLEAR_CRT_DEBUG_FIELD(a) \
            _CrtSetDbgFlag(~(a) & _CrtSetDbgFlag(_CRTDBG_REPORT_FLAG))
#   else
#       define SET_CRT_DEBUG_FIELD(a) ((void) 0)
#       define CLEAR_CRT_DEBUG_FIELD(a) ((void) 0)
#   endif // _DEBUG
#endif

#if REPORT_LEAKS==1
#   define MY_NEW                           new(__FILE__,__LINE__)

#ifdef _MSC_VER
//  disable MSVC warning: warning C4100: 'p' : unreferenced formal parameter
//      (when class has no destructor)
#   pragma warning( push )
#   pragma warning( disable : 4100 )
#endif

namespace my_destructor
{
    template<class T>
    void destructor(T *p)
    {
        p->~T();
    }

    template<class T>
    void my_delete(T* p, const char* fn, int ln)
    {
        if(NULL==p)
            return;
        destructor(p);
        operator delete(static_cast<void*>(p),fn,ln);
    }
}

#ifdef _MSC_VER
#   pragma warning( pop )
#endif

#   define MY_DELETE(x)                     my_destructor::my_delete((x),__FILE__,__LINE__)
#   define MY_DELETE_A(x)                   operator delete[]((x),__FILE__,__LINE__)

#   include "../auto_ptr/auto_ptr.h"
#   include <memory>                        // to not to collide

#   define MY_AUTO_PTR                      my_auto_ptr::auto_ptr
#   define auto_ptr                         MY_AUTO_PTR

#   define MY_MALLOC__(x)                   my_malloc__(__FILE__,__LINE__,(x))
#   define my_malloc(x)                     MY_MALLOC__((x))

#   define MY_REALLOC__(p,s)                my_realloc__(__FILE__,__LINE__,(p),(s))
#   define my_realloc(p,s)                  MY_REALLOC__((p),(s))

#   define MY_FREE__(p)                     my_free__(__FILE__,__LINE__,(p))
#   define my_free(p)                       MY_FREE__((p))
#else
#   define MY_NEW                           new

#   define MY_DELETE(x)                     delete x
#   define MY_DELETE_A(x)                   delete[] x

#   include "../auto_ptr/auto_ptr.h"
#   include <memory>                        // to not to collide

#   define MY_AUTO_PTR                      my_auto_ptr::auto_ptr
#   define auto_ptr                         MY_AUTO_PTR

#   define MY_MALLOC__(x)                   malloc((x))
#   define my_malloc(x)                     MY_MALLOC__((x))

#   define MY_REALLOC__(p,s)                realloc((p),(s))
#   define my_realloc(p,s)                  MY_REALLOC__((p),(s))

#   define MY_FREE__(p)                     free((p))
#   define my_free(p)                       MY_FREE__((p))
#endif

#endif

