/***************************************************************************
                          numconv.hh  -  description
                             -------------------
    begin                : Sun Aug 19 2007
    copyright            : (C) 2007 by Marian VooDooMan Meravy
    email                : ghostvoodooman -a.T= users (dot) sourceforge.net
 ***************************************************************************/

/* $Id: numconv.hh,v 1.16 2008/11/24 12:09:46 ghostvoodooman Exp $ */

/***************************************************************************

VooDoo cIRCle - an IRC (ro)bot
Copyright (C) 2004-2008 by Marian VooDooMan Meravy
  (ghostvoodooman -a.T= users (dot) sourceforge.net)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************/

/*!
    \file
    \brief Implements templated number to string conversions
    \version $Id: numconv.hh,v 1.16 2008/11/24 12:09:46 ghostvoodooman Exp $
*/

#ifndef NUMCONV_HH__
#define NUMCONV_HH__

#include <string>
#include <sstream>

#include "../params.h"
#include "../compat.h"

template<typename T>
MY_WISH_INLINE std::string tos(const T& value)
{
    std::ostringstream s;
    s<<value;
    return s.str();
}

template<typename T>
MY_WISH_INLINE std::string toshex(const T& value)
{
    std::ostringstream s;
    s<<std::hex<<value;
    return s.str();
}

template<typename T>
MY_WISH_INLINE std::string tos(const T& value, int width, char pad)
{
    std::ostringstream s;
    s.width(width);
    s.fill(pad);
    s<<value;
    return s.str();
}

#if defined(_MSC_VER)
#   pragma warning(push)
#   pragma warning(disable:4996)    // warning C4996: 'sscanf': This function or variable may be unsafe. Consider using sprintf_s instead. To disable deprecation, use _CRT_SECURE_NO_WARNINGS. See online help for details.
#endif

MY_WISH_INLINE MY_LONG_LONG stoll(const std::string& str)
{
    MY_LONG_LONG res;
    if(sscanf(str.c_str(),"%" MY_LONG_LONG_STR "d",&res)!=1)
        res=0;
    return res;
}

MY_WISH_INLINE MY_LONG_LONG stoll(const char* str)
{
    MY_LONG_LONG res;
    if(sscanf(str,"%" MY_LONG_LONG_STR "d",&res)!=1)
        res=0;
    return res;
}

MY_WISH_INLINE MY_U_LONG_LONG stoull(const std::string& str)
{
    MY_U_LONG_LONG res;
    if(sscanf(str.c_str(),"%" MY_LONG_LONG_STR "u",&res)!=1)
        res=0;
    return res;
}

MY_WISH_INLINE MY_U_LONG_LONG stoull(const char* str)
{
    MY_U_LONG_LONG res;
    if(sscanf(str,"%" MY_LONG_LONG_STR "u",&res)!=1)
        res=0;
    return res;
}

MY_WISH_INLINE long stol(const std::string& str)
{
    long res;
    if(sscanf(str.c_str(),"%ld",&res)!=1)
        res=0;
    return res;
}

MY_WISH_INLINE unsigned long stoul(const std::string& str)
{
    unsigned long res;
    if(sscanf(str.c_str(),"%lu",&res)!=1)
        res=0;
    return res;
}

MY_WISH_INLINE int stoi(const std::string& str)
{
    int res;
    if(sscanf(str.c_str(),"%d",&res)!=1)
        res=0;
    return res;
}

MY_WISH_INLINE int stoi(const char* str)
{
    int res;
    if(sscanf(str,"%d",&res)!=1)
        res=0;
    return res;
}

MY_WISH_INLINE unsigned stoui(const std::string& str)
{
    unsigned res;
    if(sscanf(str.c_str(),"%u",&res)!=1)
        res=0;
    return res;
}

MY_WISH_INLINE unsigned stoui(const char* str)
{
    unsigned res;
    if(sscanf(str,"%u",&res)!=1)
        res=0;
    return res;
}

MY_WISH_INLINE size_t sto_size_t(const std::string& str)
{
    MY_U_LONG_LONG res=stoull(str);
    return static_cast<size_t>(res);
}

MY_WISH_INLINE size_t sto_size_t(const char* str)
{
    MY_U_LONG_LONG res=stoull(str);
    return static_cast<size_t>(res);
}

MY_WISH_INLINE my_time_t sto_time(const std::string& str)
{
    my_time_t res;
    if(sscanf(str.c_str(),"%" MY_LONG_LONG_STR "d",&res)!=1)
        res=0;
    return res;
}

MY_WISH_INLINE my_time_t sto_time(const char* str)
{
    my_time_t res;
    if(sscanf(str,"%" MY_LONG_LONG_STR "d",&res)!=1)
        res=0;
    return res;
}

MY_WISH_INLINE short stos(const std::string& str)
{
    int res;
    if(sscanf(str.c_str(),"%d",&res)!=1)
        res=0;
#if defined(__INTEL_COMPILER)
#   pragma warning(push)
#   pragma warning(disable:810)     // remark #810: conversion from "int" to "short" may lose significant bits
#endif
    return static_cast<short>(res);
#if defined(__INTEL_COMPILER)
#   pragma warning(pop)             // remark #810: conversion from "int" to "short" may lose significant bits
#endif
}

MY_WISH_INLINE short stos(const char* str)
{
    int res;
    if(sscanf(str,"%d",&res)!=1)
        res=0;
#if defined(__INTEL_COMPILER)
#   pragma warning(push)
#   pragma warning(disable:810)     // remark #810: conversion from "int" to "short" may lose significant bits
#endif
    return static_cast<short>(res);
#if defined(__INTEL_COMPILER)
#   pragma warning(pop)             // remark #810: conversion from "int" to "short" may lose significant bits
#endif
}

MY_WISH_INLINE unsigned short stous(const std::string& str)
{
    unsigned res;
    if(sscanf(str.c_str(),"%u",&res)!=1)
        res=0;

#if defined(__INTEL_COMPILER)
#   pragma warning(push)
#   pragma warning(disable:810)     // remark #810: conversion from "unsigned int" to "unsigned short" may lose significant bits
#endif
    return static_cast<unsigned short>(res);
#if defined(__INTEL_COMPILER)
#   pragma warning(pop)             // remark #810: conversion from "unsigned int" to "unsigned short" may lose significant bits
#endif
}

MY_WISH_INLINE short stous(const char* str)
{
    unsigned res;
    if(sscanf(str,"%u",&res)!=1)
        res=0;
#if defined(__INTEL_COMPILER)
#   pragma warning(push)
#   pragma warning(disable:810)     // remark #810: conversion from "unsigned int" to "unsigned short" may lose significant bits
#endif
    return static_cast<unsigned short>(res);
#if defined(__INTEL_COMPILER)
#   pragma warning(pop)             // remark #810: conversion from "unsigned int" to "unsigned short" may lose significant bits
#endif
}

#if defined(_MSC_VER)
#   pragma warning(pop)
#endif

#endif

