/***************************************************************************
                          c_var_array.cpp  -  description
                             -------------------
    begin                : Thu May 10 2007
    copyright            : (C) 2004-2007 by Marian VooDooMan Meravy
    email                : ghostvoodooman -a.T= users (dot) sourceforge.net
 ***************************************************************************/

/* $Id: c_var_array.cpp,v 1.16 2008/06/08 08:43:14 ghostvoodooman Exp $ */

/***************************************************************************

VooDoo cIRCle - an IRC (ro)bot
Copyright (C) 2004-2008 by Marian VooDooMan Meravy
  (ghostvoodooman -a.T= users (dot) sourceforge.net)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************/

/*!
    \file
    \brief Class 'c_var_array'
    \version $Id: c_var_array.cpp,v 1.16 2008/06/08 08:43:14 ghostvoodooman Exp $
*/

#if defined(HAVE_CONFIG_H)
#   include <config.h>
#endif

#include "../params.h"
#include "../utils/memleak.h"
#include "../utils/assert.h"
#include "../compat.h"

#include "c_var_array.h"
#include "c_var_bool.h"
#include "../var_keeper.h"

#include "../utils/precomp.h"

#if defined(_MSC_VER) || defined(__BORLANDC__)
#   pragma hdrstop
#endif

void array_log_fn(c_var_keeper::e_log_severity severity, const string& msg, void* user_data);

c_var_bool* c_var_array::to_bool()
{
    return MY_NEW c_var_bool(is_true());
}

c_var_array::c_var_array()
    : c_var_base()
{
    new_int_index=0;
    var_keeper->include_into_locals(this);
}

c_var_array::c_var_array(int)
    : c_var_base()
{
    new_int_index=0;
    // do NOT:
    //var_keeper->include_into_locals(this);
}

c_var_array::c_var_array(const c_var_array& orig)
    : c_var_base(orig)
{
    new_int_index=orig.new_int_index;
    // do NOT this:
    /*
    lookup_int=orig.lookup_int;
    lookup_str=orig.lookup_str;
    */
    //var_keeper->include_into_locals(this);
}

c_var_array::c_var_array(const c_var_array& orig, int)
    : c_var_base(orig)
{
    new_int_index=orig.new_int_index;
    // do NOT this:
    /*
    lookup_int=orig.lookup_int;
    lookup_str=orig.lookup_str;
    */
}

c_var_array::~c_var_array()
{
}

c_var_array* c_var_array::copy()
{
    c_var_array* res=MY_NEW c_var_array(*this,0);
    res->reset_id();
    res->id=var_keeper->alloc(res);
    var_keeper->include_into_locals(res);
    var_keeper->array_copy(res,this);
    return res;
}

c_var_base* c_var_array::copy(int)
{
    c_var_array* res=MY_NEW c_var_array(*this);

    res->reset_id();
    res->id=var_keeper->alloc(res);

    // do NOT call this!
    //var_keeper->include_into_locals(res);

    return res;
}

c_var_base* c_var_array::get(c_module* this_module, c_index_base* index, bool insensitive, const string& debug)
{
    c_var_base* val;
    if(insensitive)
        val=var_keeper->array_get(this_module,this,index,insensitive,"",debug);
    else
        val=do_lookup(index);
    if(!val)
        return NULL;
    // dereference
    return var_keeper->get(var_keeper->dereference(val->get_id()),array_log_fn,(void*)this_module);
}

bool c_var_array::set(c_module* this_module, c_index_base*& index, c_var_base* value, bool insensitive, const string& debug)
{
    if(index->type()==c_index_base::IDX_INT && ((c_index_int*)(index))->get()>=new_int_index)
        new_int_index=static_cast<c_index_int*>(index)->get()+1;

    update_lookup(index,value);

    var_keeper->array_set(this,index,value,this_module,"",debug,insensitive);
    return true;
}

bool c_var_array::replace(c_module* this_module, c_index_base*& index, c_var_base* value, bool insensitive, const string& debug)
{
    if(index->type()==c_index_base::IDX_INT && ((c_index_int*)(index))->get()>=new_int_index)
        new_int_index=((c_index_int*)(index))->get()+1;

    update_lookup(index,value);

    return var_keeper->array_replace(this_module,this,index,value,insensitive,"",debug);
}

e_vartype c_var_array::get_var_type()
{
    return VARTYPE_ARRAY;
}

e_vartype c_var_array::get_var_type() const
{
    return VARTYPE_ARRAY;
}

string c_var_array::get_var_type_as_string()
{
    return (string)"ARRAY";
}

void c_var_array::unset(c_module* this_module, c_index_base*& index, const string& debug)
{
    update_lookup(index,NULL);

    var_keeper->array_unset(this_module,this,index,"",debug);
    return;
}

void c_var_array::set_all_as_persistent()
{
    var_keeper->array_set_all_as_persistent(this);
}

string c_var_array::get_script_string(c_module* this_module, const string& script_type)
{
    return var_keeper->array_get_script_string(this_module,script_type,this);
}

c_index_int* c_var_array::get_new_int_index()
{
    return MY_NEW c_index_int(new_int_index);
}

void c_var_array::delete_var(c_module* this_module, c_index_base*& index, const string& debug)
{
    update_lookup(index,NULL);

    var_keeper->array_unset(this_module,this,index,"",debug);
}

void c_var_array::set_ro(const string& index_, c_var_base* value)
{
    value->set_persistent(true);
    value->set_read_only(true);

    c_index_base* idx=MY_NEW c_index_str(index_);
    update_lookup(idx,value);
    var_keeper->array_set(this,idx,value,core_mod,"",__my_func__);
}

void c_var_array::unset_ro(const string& index_, bool insensitive)
{
    if(!insensitive) {
        c_index_base* idx=MY_NEW c_index_str(index_);
        update_lookup(idx,NULL);
        var_keeper->array_unset(core_mod,this,idx,"",__my_func__);
    } else {
        c_var_base* val=get(index_,true);
        if(val)
            unset_ro_value(val);
    }
}

c_var_base* c_var_array::get(const string& index, bool insensitive)
{
    c_index_str idx(index);
    c_var_base* val;
    if(insensitive)
        val=var_keeper->array_get(core_mod,this,&idx,insensitive,"",__my_func__);
    else
        val=do_lookup(&idx);
    if(!val)
        return NULL;
    // dereference
    var_id_t id=var_keeper->dereference(val->get_id());
    if(val->get_id()==id)
        return val;
    return var_keeper->get(id,array_log_fn,(void*)core_mod);
}

void c_var_array::push_ro(c_module* this_module, c_var_base* element)
{
    element->set_persistent(true);
    element->set_read_only(true);

    string debug;
    c_index_base* idx=get_new_int_index();
    update_lookup(idx,element);
    set(this_module,idx,element,false,debug);
}

void c_var_array::unset_ro_value(c_var_base* value)
{
    if(!value)
        return;

    var_keeper->array_unset_ro_value(this,value);

    map<MY_LONG_LONG,c_var_base*>::iterator i1;
    for(i1=lookup_int.begin(); i1!=lookup_int.end(); ++i1)
        if((*i1).second==value) {
            lookup_int.erase(i1);
            break;
        }
    map<string,c_var_base*>::iterator i2;
    for(i2=lookup_str.begin(); i2!=lookup_str.end(); ++i2)
        if((*i2).second==value) {
            lookup_str.erase(i2);
            break;
        }
}

string c_var_array::dump(c_module* this_module)
{
    return var_keeper->array_dump(this_module,this);
}

void c_var_array::dump_for_botnet(c_module* this_module, char* &data, size_t& size)
{
    var_keeper->array_dump_for_botnet(this_module,this,data,size);
}

void c_var_array::clear_unset()
{
    //var_keeper->array_clear_unset(core_mod,this);
}

c_var_base* c_var_array::stream_read(vector<c_index_base*>& path, vector<bool>& insensitive, c_module* this_module, const string& debug)
{
    return var_keeper->array_stream_read(this,path,insensitive,this_module,debug);
}

void c_var_array::stream_write(vector<c_index_base*>& path, vector<bool>& insensitive, c_var_base* value, c_module* this_module, const string& debug)
{
    var_keeper->array_stream_write(this,path,insensitive,value,this_module,debug);
}

bool c_var_array::is_true()
{
    return var_keeper->array_num_nonunset_children(core_mod,this)!=0;
}

bool c_var_array::is_true() const
{
    return var_keeper->array_num_nonunset_children(core_mod,const_cast<c_var_array*>(this))!=0;
}

c_index_base* c_var_array::make_index()
{
    // invalid index
    return NULL;
}

c_index_int* c_var_array::make_next_index()
{
    return MY_NEW c_index_int(new_int_index++);
}

void c_var_array::update_lookup(c_index_base* idx, c_var_base* val)
{
    if(!idx)
        return; // huh?
    if(idx->type()==c_index_base::IDX_INT) {
        map<MY_LONG_LONG,c_var_base*>::iterator i1=lookup_int.find(static_cast<c_index_int*>(idx)->get());
        if(i1!=lookup_int.end()) {
            if(val)
                (*i1).second=val;
            else
                lookup_int.erase(i1);
        } else {
            if(val)
                lookup_int.insert(make_pair(static_cast<c_index_int*>(idx)->get(),val));
        }
    } else if(idx->type()==c_index_base::IDX_STR) {
        map<string,c_var_base*>::iterator i1=lookup_str.find(static_cast<c_index_str*>(idx)->get());
        if(i1!=lookup_str.end()) {
            if(val)
                (*i1).second=val;
            else
                lookup_str.erase(i1);
        } else {
            if(val)
                lookup_str.insert(make_pair(static_cast<c_index_str*>(idx)->get(),val));
        }
    }
}

c_var_base* c_var_array::do_lookup(c_index_base* idx)
{
    if(!idx)
        return NULL; // huh?
    if(idx->type()==c_index_base::IDX_INT) {
        map<MY_LONG_LONG,c_var_base*>::iterator i1=lookup_int.find(static_cast<c_index_int*>(idx)->get());
        if(i1!=lookup_int.end()) {
            return (*i1).second;
        } else {
            return NULL;
        }
    } else if(idx->type()==c_index_base::IDX_STR) {
        map<string,c_var_base*>::iterator i1=lookup_str.find(static_cast<c_index_str*>(idx)->get());
        if(i1!=lookup_str.end()) {
            return (*i1).second;
        } else {
            return NULL;
        }
    }
    // SHOULD NOT RREACH
    return NULL;
}

string c_var_array::to_string(c_module* UNUSEDARG(this_module))
{
    return (string)"(array)";
}

string c_var_array::to_string(c_module* UNUSEDARG(this_module)) const
{
    return (string)"(array)";
}

bool c_var_array::dump_to_string_vector_wo_unset_and_array(c_module* this_module, vector<pair<const c_index_base*,string> >& result)
{
    return var_keeper->array_dump_to_string_vector_wo_unset_and_array(this_module,this,result);
}

bool c_var_array::operator==(const string& UNUSEDARG(right)) const
{
    return false;
}

bool c_var_array::operator==(const MY_LONG_LONG& UNUSEDARG(right)) const
{
    return false;
}

