/***************************************************************************
                          c_var_array.h  -  description
                             -------------------
    begin                : Thu May 10 2007
    copyright            : (C) 2004-2007 by Marian VooDooMan Meravy
    email                : ghostvoodooman -a.T= users (dot) sourceforge.net
 ***************************************************************************/

/* $Id: c_var_array.h,v 1.14 2008/03/11 21:11:49 ghostvoodooman Exp $ */

/***************************************************************************

VooDoo cIRCle - an IRC (ro)bot
Copyright (C) 2004-2008 by Marian VooDooMan Meravy
  (ghostvoodooman -a.T= users (dot) sourceforge.net)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************/

/*!
    \file
    \brief Class 'c_var_array'
    \version $Id: c_var_array.h,v 1.14 2008/03/11 21:11:49 ghostvoodooman Exp $
*/

#ifndef C_VAR_ARRAY_H__
#define C_VAR_ARRAY_H__

#if defined(HAVE_CONFIG_H)
#   include <config.h>
#endif

#include <string>
#include <vector>
#include <stack>
#include <list>
#include <map>
#include <algorithm>

#include "../utils/memleak.h"

#include "../params.h"

#include "c_var_base.h"

/*!
    \author VooDooMan
    \version 3
    \date 2005, 2007-01-14, 2008-03-09
    \brief Class for array variable
*/
class c_var_array : public c_var_base
{
private:
    /*!
        \brief Formats oops for logs
        \param func Name of function
        \param msg Message
    */
    string format_oops(string func, string msg);

    /*!
        \brief Formats oops for logs
        \param func Name of function
        \param msg Message
        \param msg2 Additional message
    */
    string format_oops(string func, string msg, string msg2);

    MY_LONG_LONG new_int_index;                 //!< Int of next index (for "$x[]=123;")

    map<MY_LONG_LONG,c_var_base*> lookup_int;   //!< Fast lookup table for int indexes
    map<string,c_var_base*> lookup_str;         //!< Fast lookup table for string indexes

    /*!
        \brief Updates lookup tables
        \author VooDooMan
        \version 1
        \date 2006-11-27
        \param idx Index
        \param val Value, or NULL to invalidate/MY_DELETE(from table)s
    */
    void update_lookup(c_index_base* idx, c_var_base* val);

    /*!
        \brief Uses lookup tables to find a value
        \author VooDooMan
        \version 1
        \date 2006-11-27
        \param idx Index
        \return Returns Value, or NULL if not found
    */
    c_var_base* do_lookup(c_index_base* idx);

private:

    friend class c_var_keeper;

    friend class c_global_keeper;

public:

    c_var_array();                          //!< Constructor
    c_var_array(int);                       //!< Constructor
    c_var_array(const c_var_array& orig);   //!< Copy constructor
    virtual ~c_var_array();                 //!< Destructor

    /*!
        \brief Copy-constructor, but not calling c_var_keeper
        \param orig Original instance
    */
    c_var_array(const c_var_array& orig, int);

    virtual e_vartype get_var_type();           //!< Gets type of variable
    virtual e_vartype get_var_type() const;           //!< Gets type of variable
    virtual string get_var_type_as_string();    //!< Gets type of variable as string

    virtual c_var_array* copy();                 //!< Creates copy of itself

    /*!
        \brief Gets value of specified index
        \param this_module Module
        \param index Index
        \param insensitive Case insensitivity flag
        \param debug Debug string
        \return Returns value
        \warning Does a dereferencing
    */
    c_var_base* get(c_module* this_module, c_index_base* index, bool insensitive, const string& debug);

    /*!
        \brief Sets value of specified index
        \param this_module Module
        \param index Index
        \param value Value
        \param insensitive Case insensitivity flag
        \param debug Debug string
        \return Returns true if ok
        \warning Does not dereferences a reference
    */
    bool set(c_module* this_module, c_index_base*& index, c_var_base* value, bool insensitive, const string& debug);

    /*!
        \brief Replaces value of specified index
        \param this_module Module
        \param index Index
        \param value Value
        \param insensitive Case insensitivity flag
        \param debug Debug string
        \return Returns true if ok
    */
    bool replace(c_module* this_module, c_index_base*& index, c_var_base* value, bool insensitive, const string& debug);

    /*!
        \brief Unsets specified index
        \param this_module Module
        \param index Index
        \param debug Debug string
        \warning Does not dereferences a reference
    */
    void unset(c_module* this_module, c_index_base*& index, const string& debug);

    bool is_true();
    bool is_true() const;

    c_var_bool* to_bool();

    /*!
        \brief Sets all member (recursive) as persistent
    */
    void set_all_as_persistent();

    string to_string(c_module* UNUSEDARG(this_module));
    string to_string(c_module* UNUSEDARG(this_module)) const;

    /*!
        \brief Returns php string
        \param this_module Module
        \param script_type Type of script (e.g. "php", "php_2", "perl")
    */
    string get_script_string(c_module* this_module, const string& script_type);

    /*!
        \brief Gets new int index
    */
    c_index_int* get_new_int_index();

    /*!
        \brief Deletes variable
        \param this_module Module
        \param index Index
        \param debug Debug string
    */
    void delete_var(c_module* this_module, c_index_base*& index, const string& debug);

    /*!
        \brief Sets read-only variable
        \param index_ Index
        \param value Value
        \warning Does not dereferences a reference
    */
    void set_ro(const string& index_, c_var_base* value);

    /*!
        \brief Unsets read-only variable
        \param index_ Index
        \param insensitive Case-insensitivity flag
        \warning Does not dereferences a reference
    */
    void unset_ro(const string& index_, bool insensitive = false);

    /*!
        \brief Gets value
        \param index Index
        \param insensitive Case-insensitivity flag
        \return Returns value
        \warning Does a dereferencing
    */
    c_var_base* get(const string& index, bool insensitive);

    /*!
        \brief Pushes next read-only variable
        \param this_module Module
        \param element Value
    */
    void push_ro(c_module* this_module, c_var_base* element);

    /*!
        \brief Unsets read-only variable
        \param value Value
        \warning Does not dereferences a reference
    */
    void unset_ro_value(c_var_base* value);

    virtual string dump(c_module* this_module);

    virtual void dump_for_botnet(c_module* this_module, char* &data, size_t& size);

    /*!
        \brief Clears all unset variables
    */
    virtual void clear_unset();

    /*!
        \brief Reads value from streamed indexes (for multi-dimensional array)
        \version 1
        \date 2006-03-19
        \author VooDooMan
        \retval NULL Returns NULL for unset variable (e.g. stream index path doesn't exist)
        \param this_module Module
        \param debug Debug string
        \param path Stream of indexes
        \param insensitive Case-insensitive index lookup (stream associated with 'path' argument)
    */
    c_var_base* stream_read(vector<c_index_base*>& path, vector<bool>& insensitive, c_module* this_module, const string& debug);

    /*!
        \brief Writes value for streamed indexes (for multi-dimensional array), adds index(es) (INT zero) in path if that path deosn't exist yet, or overwrites value if it is not array (while walking thru array tree)
        \version 1
        \date 2006-03-19
        \author VooDooMan
        \param this_module Module
        \param debug Debug string
        \param path Stream of indexes
        \param value Value to set
        \param insensitive Case-insensitive index lookup (stream associated with 'path' argument)
    */
    void stream_write(vector<c_index_base*>& path, vector<bool>& insensitive, c_var_base* value, c_module* this_module, const string& debug);

    /*!
        \brief Deletes non-read only members, and removes read-only from list
        \version 1
        \author VooDooMan
        \date 2006-04-17
    */
    void delete_all_not_read_only();

    virtual c_var_base* copy(int);

    virtual c_index_base* make_index();

    /*!
        \brief Makes new index (for such expressions as '$a[]=1;')
        \author VooDooMan
        \version 1
        \date 2006-11-19
    */
    c_index_int* make_next_index();

    /*!
        \brief Dumps one-dimensional array to string vector (and make conversions from-non-strings to string; except for nested array and UNSET)
        \date 2008-02-06
        \version 1
        \author VooDooMan
        \retval true Okay
        \retval false There is nested sub-array (result of 'result' parameter is undefined)
        \param this_module Module
        \param result Returns result (should be an empty vector)
    */
    bool dump_to_string_vector_wo_unset_and_array(c_module* this_module, vector<pair<const c_index_base*,string> >& result);

    // BEGIN: Variable operations - prototypes follows
    c_var_base& operator++(int);

    c_var_base& operator--(int);

    c_var_base& operator++();

    c_var_base& operator--();

    void pre_inc();
    void pre_dec();
    void post_inc();
    void post_dec();

    c_var_base& operator+(c_var_base& right);
    c_var_base& plus(c_var_bool* left);
    c_var_base& plus(c_var_int* left);
    c_var_base& plus(c_var_unset* left);
    c_var_base& plus(c_var_string* left);
    c_var_base& plus(c_var_array* left);

    c_var_base& operator-(c_var_base& right);
    c_var_base& minus(c_var_bool* left);
    c_var_base& minus(c_var_int* left);
    c_var_base& minus(c_var_unset* left);
    c_var_base& minus(c_var_string* left);
    c_var_base& minus(c_var_array* left);

    c_var_base& operator*(c_var_base& right);
    c_var_base& mul(c_var_bool* left);
    c_var_base& mul(c_var_int* left);
    c_var_base& mul(c_var_unset* left);
    c_var_base& mul(c_var_string* left);
    c_var_base& mul(c_var_array* left);

    c_var_base& operator/(c_var_base& right);
    c_var_base& div(c_var_bool* left);
    c_var_base& div(c_var_int* left);
    c_var_base& div(c_var_unset* left);
    c_var_base& div(c_var_string* left);
    c_var_base& div(c_var_array* left);

    c_var_base& operator!();

    c_var_base& operator==(c_var_base& right);
    c_var_base& equal(c_var_bool* left);
    c_var_base& equal(c_var_int* left);
    c_var_base& equal(c_var_unset* left);
    c_var_base& equal(c_var_string* left);
    c_var_base& equal(c_var_array* left);

    c_var_base& operator!=(c_var_base& right);
    c_var_base& not_equal(c_var_bool* left);
    c_var_base& not_equal(c_var_int* left);
    c_var_base& not_equal(c_var_unset* left);
    c_var_base& not_equal(c_var_string* left);
    c_var_base& not_equal(c_var_array* left);

    c_var_base& concat(c_var_base& right);
    c_var_base& concat(c_var_bool* left);
    c_var_base& concat(c_var_int* left);
    c_var_base& concat(c_var_unset* left);
    c_var_base& concat(c_var_string* left);
    c_var_base& concat(c_var_array* left);

    c_var_base& operator%(c_var_base& right);
    c_var_base& mod(c_var_bool* left);
    c_var_base& mod(c_var_int* left);
    c_var_base& mod(c_var_unset* left);
    c_var_base& mod(c_var_string* left);
    c_var_base& mod(c_var_array* left);

    c_var_base& operator&(c_var_base& right);
    c_var_base& bit_and(c_var_bool* left);
    c_var_base& bit_and(c_var_int* left);
    c_var_base& bit_and(c_var_unset* left);
    c_var_base& bit_and(c_var_string* left);
    c_var_base& bit_and(c_var_array* left);

    c_var_base& operator|(c_var_base& right);
    c_var_base& bit_or(c_var_bool* left);
    c_var_base& bit_or(c_var_int* left);
    c_var_base& bit_or(c_var_unset* left);
    c_var_base& bit_or(c_var_string* left);
    c_var_base& bit_or(c_var_array* left);

    c_var_base& operator^(c_var_base& right);
    c_var_base& bit_xor(c_var_bool* left);
    c_var_base& bit_xor(c_var_int* left);
    c_var_base& bit_xor(c_var_unset* left);
    c_var_base& bit_xor(c_var_string* left);
    c_var_base& bit_xor(c_var_array* left);

    c_var_base& operator<<(c_var_base& right);
    c_var_base& bit_sh_left(c_var_bool* left);
    c_var_base& bit_sh_left(c_var_int* left);
    c_var_base& bit_sh_left(c_var_unset* left);
    c_var_base& bit_sh_left(c_var_string* left);
    c_var_base& bit_sh_left(c_var_array* left);

    c_var_base& operator>>(c_var_base& right);
    c_var_base& bit_sh_right(c_var_bool* left);
    c_var_base& bit_sh_right(c_var_int* left);
    c_var_base& bit_sh_right(c_var_unset* left);
    c_var_base& bit_sh_right(c_var_string* left);
    c_var_base& bit_sh_right(c_var_array* left);

    c_var_base& operator>(c_var_base& right);
    c_var_base& greater(c_var_bool* left);
    c_var_base& greater(c_var_int* left);
    c_var_base& greater(c_var_unset* left);
    c_var_base& greater(c_var_string* left);
    c_var_base& greater(c_var_array* left);

    c_var_base& operator<(c_var_base& right);
    c_var_base& less(c_var_bool* left);
    c_var_base& less(c_var_int* left);
    c_var_base& less(c_var_unset* left);
    c_var_base& less(c_var_string* left);
    c_var_base& less(c_var_array* left);

    c_var_base& operator>=(c_var_base& right);
    c_var_base& greater_or_eq(c_var_bool* left);
    c_var_base& greater_or_eq(c_var_int* left);
    c_var_base& greater_or_eq(c_var_unset* left);
    c_var_base& greater_or_eq(c_var_string* left);
    c_var_base& greater_or_eq(c_var_array* left);

    c_var_base& operator<=(c_var_base& right);
    c_var_base& less_or_eq(c_var_bool* left);
    c_var_base& less_or_eq(c_var_int* left);
    c_var_base& less_or_eq(c_var_unset* left);
    c_var_base& less_or_eq(c_var_string* left);
    c_var_base& less_or_eq(c_var_array* left);

    c_var_base& operator~();

    c_var_base& operator&&(c_var_base& right);
    c_var_base& log_and(c_var_bool* left);
    c_var_base& log_and(c_var_int* left);
    c_var_base& log_and(c_var_unset* left);
    c_var_base& log_and(c_var_string* left);
    c_var_base& log_and(c_var_array* left);

    c_var_base& operator||(c_var_base& right);
    c_var_base& log_or(c_var_bool* left);
    c_var_base& log_or(c_var_int* left);
    c_var_base& log_or(c_var_unset* left);
    c_var_base& log_or(c_var_string* left);
    c_var_base& log_or(c_var_array* left);
    // END

    bool operator==(const string& right) const;
    bool operator==(const MY_LONG_LONG& right) const;
};

#endif

