/***************************************************************************
                          c_var_base.cpp  -  description
                             -------------------
    begin                : Thu May 10 2007
    copyright            : (C) 2004-2007 by Marian VooDooMan Meravy
    email                : ghostvoodooman -aT= users (dot) sourceforge.net
 ***************************************************************************/

/* $Id: c_var_base.cpp,v 1.14 2008/03/11 21:11:49 ghostvoodooman Exp $ */

/***************************************************************************

VooDoo cIRCle - an IRC (ro)bot
Copyright (C) 2004-2008 by Marian VooDooMan Meravy
  (ghostvoodooman -aT= users (dot) sourceforge.net)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************/

/*!
    \file
    \brief Class 'c_var_base'
    \version $Id: c_var_base.cpp,v 1.14 2008/03/11 21:11:49 ghostvoodooman Exp $
*/

#if defined(HAVE_CONFIG_H)
#   include <config.h>
#endif

#include "../params.h"
#include "../utils/memleak.h"
#include "../utils/assert.h"
#include "../compat.h"

#include "c_var_base.h"
#include "../var_keeper.h"
#include "c_var_int.h"
#include "c_var_string.h"
#include "c_var_bool.h"
#include "c_var_array.h"
#include "c_var_unset.h"

#include "../utils/precomp.h"

#if defined(_MSC_VER) || defined(__BORLANDC__)
#   pragma hdrstop
#endif

void var_log_fn(c_var_keeper::e_log_severity severity, const string& msg, void* mod);
void array_log_fn(c_var_keeper::e_log_severity severity, const string& msg, void* user_data);

c_var_base::c_var_base()
{
    read_only=false;
    persistent_=false;

    in_deletion=false;

    session_=var_keeper->get_current_session();

    id=VAR_ID_INVALID;
    id=var_keeper->alloc(this);

    // do NOT call this, as it would mean pure virtual method call, since real class (not this
    //     base one) is not instantiated yet; instead, call it in each descendant
    //var_keeper->include_into_locals(this);
}

c_var_base::c_var_base(const c_var_base& orig)
{
    read_only=orig.read_only;
    persistent_=orig.persistent_;
    id=orig.id;
    session_=orig.session_;

    in_deletion=false;
}

c_var_base::~c_var_base()
{
    read_only=false;
    persistent_=false;
    if(id)
        var_keeper->free(id,var_log_fn,NULL);
    var_keeper->destroy(this);
}

void c_var_base::set_read_only(bool ro)
{
    read_only=ro;
}

bool c_var_base::get_read_only()
{
    return read_only;
}

void c_var_base::set_persistent(bool persistent)
{
    persistent_=persistent;
}

bool c_var_base::get_persistent()
{
    return persistent_;
}

bool c_var_base::is_identical(c_module* this_module, c_var_base* right, const string& debug_prefix, const string& debug)
{
    if(get_var_type()!=right->get_var_type()) {
        return false;
    }

    switch(right->get_var_type()) {
        case VARTYPE_UNSET:
            return true;
        case VARTYPE_INT:
            {
                c_var_int* l=dynamic_cast<c_var_int*>(this);
                if(!l) {
                    string d=debug_prefix;
                    d+="C++ dynamic_cast<c_var_int*>(this) returned NULL";
                    throw c_var_exception(this_module,__FILE__,__my_func__,c_var_exception::E_DYNCAST,GET_STR_DEBUG_INFO_EX(d),this);
                }
                c_var_int* r=dynamic_cast<c_var_int*>(right);
                if(!r) {
                    string d=debug_prefix;
                    d+="C++ dynamic_cast<c_var_int*>(right) returned NULL";
                    throw c_var_exception(this_module,__FILE__,__my_func__,c_var_exception::E_DYNCAST,GET_STR_DEBUG_INFO_EX(d),right);
                }
                return l->get()==r->get();
            }
        case VARTYPE_STRING:
            {
                c_var_string* l=dynamic_cast<c_var_string*>(this);
                if(!l) {
                    string d=debug_prefix;
                    d+="C++ dynamic_cast<c_var_string*>(this) returned NULL";
                    throw c_var_exception(this_module,__FILE__,__my_func__,c_var_exception::E_DYNCAST,GET_STR_DEBUG_INFO_EX(d),this);
                }
                c_var_string* r=dynamic_cast<c_var_string*>(right);
                if(!r) {
                    string d=debug_prefix;
                    d+="C++ dynamic_cast<c_var_string*>(right) returned NULL";
                    throw c_var_exception(this_module,__FILE__,__my_func__,c_var_exception::E_DYNCAST,GET_STR_DEBUG_INFO_EX(d),right);
                }
                return !l->get().compare(r->get());
            }
        case VARTYPE_BOOL:
            {
                c_var_bool* l=dynamic_cast<c_var_bool*>(this);
                if(!l) {
                    string d=debug_prefix;
                    d+="C++ dynamic_cast<c_var_bool*>(this) returned NULL";
                    throw c_var_exception(this_module,__FILE__,__my_func__,c_var_exception::E_DYNCAST,GET_STR_DEBUG_INFO_EX(d),this);
                }
                c_var_bool* r=dynamic_cast<c_var_bool*>(right);
                if(!r) {
                    string d=debug_prefix;
                    d+="C++ dynamic_cast<c_var_bool*>(right) returned NULL";
                    throw c_var_exception(this_module,__FILE__,__my_func__,c_var_exception::E_DYNCAST,GET_STR_DEBUG_INFO_EX(d),right);
                }
                return l->get()==r->get();
            }
        case VARTYPE_ARRAY:
            {
                return var_keeper->array_is_identical(this_module,this,right,debug_prefix,debug);
            }
        default:
            break;
    }
    {
        string d=debug_prefix;
        d+="Identity check for this variable type is not implemented";
        throw c_var_exception(this_module,__FILE__,__my_func__,c_var_exception::E_NIMPL,GET_STR_DEBUG_INFO_EX(d),this);
    }
}

bool c_var_base::is_equal(c_module* this_module, c_var_base* right, const string& debug_prefix, const string& debug)
{
    if(!right) {
        string d=debug_prefix;
        d+="Sanity check failed, value of 'right' argument is NULL";
        throw c_var_exception(this_module,__FILE__,__my_func__,c_var_exception::E_INTERNAL,GET_STR_DEBUG_INFO_EX(d),this);
    }
    if(this->get_var_type()!=right->get_var_type()) { // do we need to do implicit typecast?
        c_var_base* left=this;

        e_vartype lt=left->get_var_type();
        e_vartype rt=right->get_var_type();
        if(lt!=VARTYPE_INT && lt!=VARTYPE_STRING && lt!=VARTYPE_ARRAY && lt!=VARTYPE_UNSET && lt!=VARTYPE_BOOL) {
            string d=debug_prefix;
            d+="Equality check for this variable type is not implemented";
            throw c_var_exception(this_module,__FILE__,__my_func__,c_var_exception::E_NIMPL,GET_STR_DEBUG_INFO_EX(d),this);
        }
        if(rt!=VARTYPE_INT && rt!=VARTYPE_STRING && rt!=VARTYPE_ARRAY && rt!=VARTYPE_UNSET && lt!=VARTYPE_BOOL) {
            string d=debug_prefix;
            d+="Equality check for this variable type is not implemented";
            throw c_var_exception(this_module,__FILE__,__my_func__,c_var_exception::E_NIMPL,GET_STR_DEBUG_INFO_EX(d),right);
        }

        // BEGIN forward initialization
        // for left == this
        c_var_int* left_i=NULL;
        if(VARTYPE_INT==lt) {
            left_i=dynamic_cast<c_var_int*>(left);
            if(!left_i) {
                string d=debug_prefix;
                d+="C++ dynamic_cast<c_var_int*>(left) returned NULL";
                throw c_var_exception(this_module,__FILE__,__my_func__,c_var_exception::E_DYNCAST,GET_STR_DEBUG_INFO_EX(d),this);
            }
        }
        c_var_string* left_s=NULL;
        if(VARTYPE_STRING==lt) {
            left_s=dynamic_cast<c_var_string*>(left);
            if(!left_s) {
                string d=debug_prefix;
                d+="C++ dynamic_cast<c_var_string*>(left) returned NULL";
                throw c_var_exception(this_module,__FILE__,__my_func__,c_var_exception::E_DYNCAST,GET_STR_DEBUG_INFO_EX(d),this);
            }
        }
        c_var_array* left_a=NULL;
        if(VARTYPE_ARRAY==lt) {
            left_a=dynamic_cast<c_var_array*>(left);
            if(!left_a) {
                string d=debug_prefix;
                d+="C++ dynamic_cast<c_var_array*>(left) returned NULL";
                throw c_var_exception(this_module,__FILE__,__my_func__,c_var_exception::E_DYNCAST,GET_STR_DEBUG_INFO_EX(d),this);
            }
        }
        c_var_bool* left_b=NULL;
        if(VARTYPE_BOOL==lt) {
            left_b=dynamic_cast<c_var_bool*>(left);
            if(!left_b) {
                string d=debug_prefix;
                d+="C++ dynamic_cast<c_var_bool*>(left) returned NULL";
                throw c_var_exception(this_module,__FILE__,__my_func__,c_var_exception::E_DYNCAST,GET_STR_DEBUG_INFO_EX(d),this);
            }
        }
        c_var_unset* left_u=NULL;
        if(VARTYPE_UNSET==lt) {
            left_u=dynamic_cast<c_var_unset*>(left);
            if(!left_u) {
                string d=debug_prefix;
                d+="C++ dynamic_cast<c_var_unset*>(left) returned NULL";
                throw c_var_exception(this_module,__FILE__,__my_func__,c_var_exception::E_DYNCAST,GET_STR_DEBUG_INFO_EX(d),this);
            }
        }

        // for right
        c_var_int* right_i=NULL;
        if(VARTYPE_INT==rt) {
            right_i=dynamic_cast<c_var_int*>(right);
            if(!right_i) {
                string d=debug_prefix;
                d+="C++ dynamic_cast<c_var_int*>(right) returned NULL";
                throw c_var_exception(this_module,__FILE__,__my_func__,c_var_exception::E_DYNCAST,GET_STR_DEBUG_INFO_EX(d),right);
            }
        }
        c_var_string* right_s=NULL;
        if(VARTYPE_STRING==rt) {
            right_s=dynamic_cast<c_var_string*>(right);
            if(!right_s) {
                string d=debug_prefix;
                d+="C++ dynamic_cast<c_var_string*>(right) returned NULL";
                throw c_var_exception(this_module,__FILE__,__my_func__,c_var_exception::E_DYNCAST,GET_STR_DEBUG_INFO_EX(d),right);
            }
        }
        c_var_array* right_a=NULL;
        if(VARTYPE_ARRAY==rt) {
            right_a=dynamic_cast<c_var_array*>(right);
            if(!right_a) {
                string d=debug_prefix;
                d+="C++ dynamic_cast<c_var_array*>(right) returned NULL";
                throw c_var_exception(this_module,__FILE__,__my_func__,c_var_exception::E_DYNCAST,GET_STR_DEBUG_INFO_EX(d),right);
            }
        }
        c_var_bool* right_b=NULL;
        if(VARTYPE_BOOL==rt) {
            right_b=dynamic_cast<c_var_bool*>(right);
            if(!right_b) {
                string d=debug_prefix;
                d+="C++ dynamic_cast<c_var_bool*>(right) returned NULL";
                throw c_var_exception(this_module,__FILE__,__my_func__,c_var_exception::E_DYNCAST,GET_STR_DEBUG_INFO_EX(d),right);
            }
        }
        c_var_unset* right_u=NULL;
        if(VARTYPE_UNSET==rt) {
            right_u=dynamic_cast<c_var_unset*>(right);
            if(!right_u) {
                string d=debug_prefix;
                d+="C++ dynamic_cast<c_var_unset*>(right) returned NULL";
                throw c_var_exception(this_module,__FILE__,__my_func__,c_var_exception::E_DYNCAST,GET_STR_DEBUG_INFO_EX(d),right);
            }
        }
        // END   forward initialization

        // int == string, string == int
        if((VARTYPE_INT==lt && VARTYPE_STRING==rt) || (VARTYPE_STRING==lt && VARTYPE_INT==rt)) {
            bool l=false;
            bool r=false;

            if(left_i)
                l=left_i->is_true();
            if(right_i)
                r=right_i->is_true();

            if(left_s)
                l=left_s->is_true();
            if(right_s)
                r=right_s->is_true();

            return l == r;
        }

        // bool == string, string == bool
        if((VARTYPE_BOOL==lt && VARTYPE_STRING==rt) || (VARTYPE_STRING==lt && VARTYPE_BOOL==rt)) {
            bool l=false;
            bool r=false;

            if(left_b)
                l=left_b->is_true();
            if(right_b)
                r=right_b->is_true();

            if(left_s)
                l=left_s->is_true();
            if(right_s)
                r=right_s->is_true();

            return l == r;
        }

        // int == array, array == int
        if((VARTYPE_INT==lt && VARTYPE_ARRAY==rt) || (VARTYPE_ARRAY==lt && VARTYPE_INT==rt)) {
            bool l=false;
            bool r=false;

            if(left_i)
                l=left_i->is_true();
            if(right_i)
                r=right_i->is_true();

            if(left_a)
                l=left_a->is_true();
            if(right_a)
                r=right_a->is_true();

            return l == r;
        }

        // bool == array, array == bool
        if((VARTYPE_BOOL==lt && VARTYPE_ARRAY==rt) || (VARTYPE_ARRAY==lt && VARTYPE_BOOL==rt)) {
            bool l=false;
            bool r=false;

            if(left_b)
                l=left_b->is_true();
            if(right_b)
                r=right_b->is_true();

            if(left_a)
                l=left_a->is_true();
            if(right_a)
                r=right_a->is_true();

            return l == r;
        }

        // int == unset, unset == int
        if((VARTYPE_INT==lt && VARTYPE_UNSET==rt) || (VARTYPE_UNSET==lt && VARTYPE_INT==rt)) {
            bool l=false;
            bool r=false;

            if(left_i)
                l=left_i->is_true();
            if(right_i)
                r=right_i->is_true();

            if(left_u)
                l=left_u->is_true();
            if(right_u)
                r=right_u->is_true();

            return l == r;
        }

        // bool == unset, unset == bool
        if((VARTYPE_BOOL==lt && VARTYPE_UNSET==rt) || (VARTYPE_UNSET==lt && VARTYPE_BOOL==rt)) {
            bool l=false;
            bool r=false;

            if(left_b)
                l=left_b->is_true();
            if(right_b)
                r=right_b->is_true();

            if(left_u)
                l=left_u->is_true();
            if(right_u)
                r=right_u->is_true();

            return l == r;
        }

        // unset == array, array == unset
        if((VARTYPE_UNSET==lt && VARTYPE_ARRAY==rt) || (VARTYPE_ARRAY==lt && VARTYPE_UNSET==rt)) {
            bool l=false;
            bool r=false;

            if(left_u)
                l=left_u->is_true();
            if(right_u)
                r=right_u->is_true();

            if(left_a)
                l=left_a->is_true();
            if(right_a)
                r=right_a->is_true();

            return l == r;
        }

        // bool == int, int == bool
        if((VARTYPE_BOOL==lt && VARTYPE_INT==rt) || (VARTYPE_INT==lt && VARTYPE_BOOL==rt)) {
            bool l=false;
            bool r=false;

            if(left_b)
                l=left_b->is_true();
            if(right_b)
                r=right_b->is_true();

            if(left_i)
                l=left_i->is_true();
            if(right_i)
                r=right_i->is_true();

            return l == r;
        }

        // string == array, array == string
        if((VARTYPE_STRING==lt && VARTYPE_ARRAY==rt) || (VARTYPE_ARRAY==lt && VARTYPE_STRING==rt)) {
            bool l=false;
            bool r=false;

            if(left_s)
                l=left_s->is_true();
            if(right_s)
                r=right_s->is_true();

            if(left_a)
                l=left_a->is_true();
            if(right_a)
                r=right_a->is_true();

            return l == r;
        }

        // string == unset, unset == string
        if((VARTYPE_STRING==lt && VARTYPE_UNSET==rt) || (VARTYPE_UNSET==lt && VARTYPE_STRING==rt)) {
            bool l=false;
            bool r=false;

            if(left_s)
                l=left_s->is_true();
            if(right_s)
                r=right_s->is_true();

            if(left_u)
                l=left_u->is_true();
            if(right_u)
                r=right_u->is_true();

            return l == r;
        }

        // should never occur
        string d=debug_prefix;
        d+="Equality check for this variable type is not implemented (#1), left-handed is of type ";
        d+=this->get_var_type_as_string();
        d+=", right-handed is of type ";
        d+=right->get_var_type_as_string();
        throw c_var_exception(this_module,__FILE__,__my_func__,c_var_exception::E_NIMPL,GET_STR_DEBUG_INFO_EX(d),this);
    }
    // both are same types
    switch(this->get_var_type()) {
        case VARTYPE_UNSET:
            return true;
        case VARTYPE_INT:
            {
                c_var_int* l=dynamic_cast<c_var_int*>(this);
                if(!l) {
                    string d=debug_prefix;
                    d+="C++ dynamic_cast<c_var_int*>(this) returned NULL";
                    throw c_var_exception(this_module,__FILE__,__my_func__,c_var_exception::E_DYNCAST,GET_STR_DEBUG_INFO_EX(d),this);
                }
                c_var_int* r=dynamic_cast<c_var_int*>(right);
                if(!r) {
                    string d=debug_prefix;
                    d+="C++ dynamic_cast<c_var_int*>(right) returned NULL";
                    throw c_var_exception(this_module,__FILE__,__my_func__,c_var_exception::E_DYNCAST,GET_STR_DEBUG_INFO_EX(d),right);
                }
                return l->get()==r->get();
            }
        case VARTYPE_STRING:
            {
                c_var_string* l=dynamic_cast<c_var_string*>(this);
                if(!l) {
                    string d=debug_prefix;
                    d+="C++ dynamic_cast<c_var_string*>(this) returned NULL";
                    throw c_var_exception(this_module,__FILE__,__my_func__,c_var_exception::E_DYNCAST,GET_STR_DEBUG_INFO_EX(d),this);
                }
                c_var_string* r=dynamic_cast<c_var_string*>(right);
                if(!r) {
                    string d=debug_prefix;
                    d+="C++ dynamic_cast<c_var_string*>(right) returned NULL";
                    throw c_var_exception(this_module,__FILE__,__my_func__,c_var_exception::E_DYNCAST,GET_STR_DEBUG_INFO_EX(d),right);
                }
                return !l->get().compare(r->get());
            }
        case VARTYPE_ARRAY:
            {
                c_var_array* l=dynamic_cast<c_var_array*>(this);
                if(!l) {
                    string d=debug_prefix;
                    d+="C++ dynamic_cast<c_var_array*>(this) returned NULL";
                    throw c_var_exception(this_module,__FILE__,__my_func__,c_var_exception::E_DYNCAST,GET_STR_DEBUG_INFO_EX(d),this);
                }
                c_var_array* r=dynamic_cast<c_var_array*>(right);
                if(!r) {
                    string d=debug_prefix;
                    d+="C++ dynamic_cast<c_var_array*>(right) returned NULL";
                    throw c_var_exception(this_module,__FILE__,__my_func__,c_var_exception::E_DYNCAST,GET_STR_DEBUG_INFO_EX(d),right);
                }

                return var_keeper->array_is_equal(this_module,l,r,debug_prefix,debug);
            }
        case VARTYPE_BOOL:
            {
                c_var_bool* l=dynamic_cast<c_var_bool*>(this);
                if(!l) {
                    string d=debug_prefix;
                    d+="C++ dynamic_cast<c_var_bool*>(this) returned NULL";
                    throw c_var_exception(this_module,__FILE__,__my_func__,c_var_exception::E_DYNCAST,GET_STR_DEBUG_INFO_EX(d),this);
                }
                c_var_bool* r=dynamic_cast<c_var_bool*>(right);
                if(!r) {
                    string d=debug_prefix;
                    d+="C++ dynamic_cast<c_var_bool*>(right) returned NULL";
                    throw c_var_exception(this_module,__FILE__,__my_func__,c_var_exception::E_DYNCAST,GET_STR_DEBUG_INFO_EX(d),right);
                }
                return l->get()==r->get();
            }
        default:
            break;
    }
    string d=debug_prefix;
    d+="Equality check for this variable type is not implemented (#2), left-handed is of type ";
    d+=this->get_var_type_as_string();
    d+=", right-handed is of type ";
    d+=right->get_var_type_as_string();
    throw c_var_exception(this_module,__FILE__,__my_func__,c_var_exception::E_NIMPL,GET_STR_DEBUG_INFO_EX(d),this);
}

void c_var_base::reset_id()
{
    id=0;
}

void c_var_base::set_in_deletion()
{
    in_deletion=true;
}

bool c_var_base::get_in_deletion()
{
    return in_deletion;
}

