/***************************************************************************
                          c_var_bool.cpp  -  description
                             -------------------
    begin                : Thu May 10 2007
    copyright            : (C) 2004-2007 by Marian VooDooMan Meravy
    email                : ghostvoodooman -a.T= users (dot) sourceforge.net
 ***************************************************************************/

/* $Id: c_var_bool.cpp,v 1.16 2008/08/23 15:10:46 ghostvoodooman Exp $ */

/***************************************************************************

VooDoo cIRCle - an IRC (ro)bot
Copyright (C) 2004-2008 by Marian VooDooMan Meravy
  (ghostvoodooman -a.T= users (dot) sourceforge.net)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************/

/*!
    \file
    \brief Class 'c_var_bool'
    \version $Id: c_var_bool.cpp,v 1.16 2008/08/23 15:10:46 ghostvoodooman Exp $
*/

#if defined(HAVE_CONFIG_H)
#   include <config.h>
#endif

#include "../params.h"
#include "../utils/memleak.h"
#include "../utils/assert.h"
#include "../compat.h"

#include "c_var_bool.h"
#include "../var_keeper.h"

#include "../utils/precomp.h"

#if defined(_MSC_VER) || defined(__BORLANDC__)
#   pragma hdrstop
#endif

c_var_bool::c_var_bool(const c_var_bool& orig)
    : c_var_base(orig)
{
    value_=orig.value_;
    //var_keeper->include_into_locals(this);
}

c_var_bool::c_var_bool(const c_var_bool& orig, int)
    : c_var_base(orig)
{
    value_=orig.value_;
}

string c_var_bool::dump(c_module* UNUSEDARG(this_module))
{
    string res="(BOOL) ";
    res+=(value_?"TRUE":"FALSE");
    return res;
}

void c_var_bool::dump_for_botnet(c_module* UNUSEDARG(this_module), char* &data, size_t& size)
{
    size=1+1;
    char* d=MY_NEW char[size];
    d[0]='B'; // 'B' stands for bool
    d[1]=value_?(char)1:(char)0;
    data=d;
}

string c_var_bool::get_script_string(c_module* UNUSEDARG(this_module), const string& script_type)
{
    if(!script_type.compare("php") || !script_type.compare("php_2") || !script_type.compare("logic.txt")) {
        return value_?"TRUE":"FALSE";
    }
    else if(!script_type.compare("perl")) {
        return value_?"1":"0";
    }

    // error, not supported script
    return (string)"";
}

c_var_base* c_var_bool::copy()
{
    c_var_base* res=MY_NEW c_var_bool(*this,0);
    res->set_persistent(false);
    res->set_read_only(false);
    res->reset_id();
    res->id=var_keeper->alloc(res);
    res->set_session(var_keeper->get_current_session());
    var_keeper->include_into_locals(res);
    return res;
}

c_var_bool::c_var_bool()
    : c_var_base()
{
    value_=false;
    var_keeper->include_into_locals(this);
}

c_var_bool::c_var_bool(bool value)
    : c_var_base()
{
    value_=value;
    var_keeper->include_into_locals(this);
}

c_var_bool::c_var_bool(bool value, bool persistent)
    : c_var_base()
{
    value_=value;
    set_persistent(persistent);
    var_keeper->include_into_locals(this);
}

c_var_base* c_var_bool::copy(int)
{
    c_var_bool* res=MY_NEW c_var_bool(*this,0);
    res->set_persistent(false);
    res->set_read_only(false);
    res->reset_id();
    res->id=var_keeper->alloc(res);
    res->set_session(var_keeper->get_current_session());
    //var_keeper->include_into_locals(res);
    return res;
}

c_index_base* c_var_bool::make_index()
{
    // invalid index, BUT, implicit typecast to int
    return MY_NEW c_index_int(value_ ? 1 : 0);
}

bool c_var_bool::is_true()
{
    return value_;
}

bool c_var_bool::is_true() const
{
    return value_;
}

string c_var_bool::to_string(c_module* UNUSEDARG(this_module))
{
    return value_ ? "1" : "0";
}

string c_var_bool::to_string(c_module* UNUSEDARG(this_module)) const
{
    return value_ ? "1" : "0";
}

c_var_bool::~c_var_bool()
{
}

bool c_var_bool::operator==(const string& right) const
{
    if(!right.compare("TRUE")) { // ==
        return true==value_;
    } else if(!right.compare("FALSE")) { // ==
        return false==value_;
    }
    return false;
}

bool c_var_bool::operator==(const MY_LONG_LONG& right) const
{
    return TO_BOOL(right)==value_;
}

