/***************************************************************************
                          c_var_bool.h  -  description
                             -------------------
    begin                : Thu May 10 2007
    copyright            : (C) 2004-2007 by Marian VooDooMan Meravy
    email                : ghostvoodooman -a.T= users (dot) sourceforge.net
 ***************************************************************************/

/* $Id: c_var_bool.h,v 1.15 2008/03/25 12:04:30 ghostvoodooman Exp $ */

/***************************************************************************

VooDoo cIRCle - an IRC (ro)bot
Copyright (C) 2004-2008 by Marian VooDooMan Meravy
  (ghostvoodooman -a.T= users (dot) sourceforge.net)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************/

/*!
    \file
    \brief Class 'c_var_bool'
    \version $Id: c_var_bool.h,v 1.15 2008/03/25 12:04:30 ghostvoodooman Exp $
*/

#ifndef C_VAR_BOOL_H__
#define C_VAR_BOOL_H__

#if defined(HAVE_CONFIG_H)
#   include <config.h>
#endif

#include <string>
#include <vector>
#include <stack>
#include <list>
#include <map>
#include <algorithm>

#include "../utils/memleak.h"

#include "../params.h"

#include "c_var_base.h"

/*!
    \author VooDooMan
    \version 3
    \date 2005, 2007-01-14, 2008-03-09
    \brief Boolean variable implementation
*/
class c_var_bool : public c_var_base
{
protected:
    bool value_;                                    //!< Value of variable

public:
    /*!
        \brief Constructor
    */
    c_var_bool();

    /*!
        \brief Constructor
        \param value Initial value
    */
    c_var_bool(bool value);

    /*!
        \brief Constructor
        \param value Initial value
        \param persistent Persistency flag to set
    */
    c_var_bool(bool value, bool persistent);

    /*!
        \brief Copy-constructor
        \param orig Original instance
    */
    c_var_bool(const c_var_bool& orig);

    /*!
        \brief Copy-constructor, but not calling c_var_keeper
        \param orig Original instance
    */
    c_var_bool(const c_var_bool& orig, int);

    virtual ~c_var_bool();          //!< Destructor

    /*!
        \brief Gets type of variable
    */
    e_vartype get_var_type() {
        return VARTYPE_BOOL;
    }

    /*!
        \brief Gets type of variable
    */
    e_vartype get_var_type() const {
        return VARTYPE_BOOL;
    }

    /*!
        \brief Gets type of variable as string
    */
    string get_var_type_as_string() {
        return (string)"BOOL";
    }

    /*!
        \brief Copies itself and returns new variable
    */
    virtual c_var_base* copy();

    /*!
        \brief Gets content of variable
    */
    bool get() {
        return value_;
    }

    /*!
        \brief Sets content of variable
        \param value Value to set
    */
    void set(bool value) {
        value_=value;
    }

    /*!
        \brief Converts to int and returns result
    */
    int to_int() {
        return value_ ? 1 : 0;
    }

    /*!
        \brief Converts to long and returns result
    */
    long to_long() {
        return value_ ? 1L : 0L;
    }

    /*!
        \brief Converts to long and returns result
    */
    long to_long() const {
        return value_ ? 1L : 0L;
    }

    bool is_true();
    bool is_true() const;

    /*!
        \brief Converts to bool and returns result as c_var_bool
    */
    c_var_bool* to_bool()
    {
        return MY_NEW c_var_bool(is_true());
    }

    string to_string(c_module* UNUSEDARG(this_module));
    string to_string(c_module* UNUSEDARG(this_module)) const;

    virtual string dump(c_module* this_module);

    virtual void dump_for_botnet(c_module* this_module, char* &data, size_t& size);

    string get_script_string(c_module* this_module, const string& script_type);

    virtual c_var_base* copy(int);

    virtual c_index_base* make_index();

    // BEGIN: Variable operations - prototypes follows
    c_var_base& operator++(int);

    c_var_base& operator--(int);

    c_var_base& operator++();

    c_var_base& operator--();

    void pre_inc();
    void pre_dec();
    void post_inc();
    void post_dec();

    c_var_base& operator+(c_var_base& right);
    c_var_base& plus(c_var_bool* left);
    c_var_base& plus(c_var_int* left);
    c_var_base& plus(c_var_unset* left);
    c_var_base& plus(c_var_string* left);
    c_var_base& plus(c_var_array* left);

    c_var_base& operator-(c_var_base& right);
    c_var_base& minus(c_var_bool* left);
    c_var_base& minus(c_var_int* left);
    c_var_base& minus(c_var_unset* left);
    c_var_base& minus(c_var_string* left);
    c_var_base& minus(c_var_array* left);

    c_var_base& operator*(c_var_base& right);
    c_var_base& mul(c_var_bool* left);
    c_var_base& mul(c_var_int* left);
    c_var_base& mul(c_var_unset* left);
    c_var_base& mul(c_var_string* left);
    c_var_base& mul(c_var_array* left);

    c_var_base& operator/(c_var_base& right);
    c_var_base& div(c_var_bool* left);
    c_var_base& div(c_var_int* left);
    c_var_base& div(c_var_unset* left);
    c_var_base& div(c_var_string* left);
    c_var_base& div(c_var_array* left);

    c_var_base& operator!();

    c_var_base& operator==(c_var_base& right);
    c_var_base& equal(c_var_bool* left);
    c_var_base& equal(c_var_int* left);
    c_var_base& equal(c_var_unset* left);
    c_var_base& equal(c_var_string* left);
    c_var_base& equal(c_var_array* left);

    c_var_base& operator!=(c_var_base& right);
    c_var_base& not_equal(c_var_bool* left);
    c_var_base& not_equal(c_var_int* left);
    c_var_base& not_equal(c_var_unset* left);
    c_var_base& not_equal(c_var_string* left);
    c_var_base& not_equal(c_var_array* left);

    c_var_base& concat(c_var_base& right);
    c_var_base& concat(c_var_bool* left);
    c_var_base& concat(c_var_int* left);
    c_var_base& concat(c_var_unset* left);
    c_var_base& concat(c_var_string* left);
    c_var_base& concat(c_var_array* left);

    c_var_base& operator%(c_var_base& right);
    c_var_base& mod(c_var_bool* left);
    c_var_base& mod(c_var_int* left);
    c_var_base& mod(c_var_unset* left);
    c_var_base& mod(c_var_string* left);
    c_var_base& mod(c_var_array* left);

    c_var_base& operator&(c_var_base& right);
    c_var_base& bit_and(c_var_bool* left);
    c_var_base& bit_and(c_var_int* left);
    c_var_base& bit_and(c_var_unset* left);
    c_var_base& bit_and(c_var_string* left);
    c_var_base& bit_and(c_var_array* left);

    c_var_base& operator|(c_var_base& right);
    c_var_base& bit_or(c_var_bool* left);
    c_var_base& bit_or(c_var_int* left);
    c_var_base& bit_or(c_var_unset* left);
    c_var_base& bit_or(c_var_string* left);
    c_var_base& bit_or(c_var_array* left);

    c_var_base& operator^(c_var_base& right);
    c_var_base& bit_xor(c_var_bool* left);
    c_var_base& bit_xor(c_var_int* left);
    c_var_base& bit_xor(c_var_unset* left);
    c_var_base& bit_xor(c_var_string* left);
    c_var_base& bit_xor(c_var_array* left);

    c_var_base& operator<<(c_var_base& right);
    c_var_base& bit_sh_left(c_var_bool* left);
    c_var_base& bit_sh_left(c_var_int* left);
    c_var_base& bit_sh_left(c_var_unset* left);
    c_var_base& bit_sh_left(c_var_string* left);
    c_var_base& bit_sh_left(c_var_array* left);

    c_var_base& operator>>(c_var_base& right);
    c_var_base& bit_sh_right(c_var_bool* left);
    c_var_base& bit_sh_right(c_var_int* left);
    c_var_base& bit_sh_right(c_var_unset* left);
    c_var_base& bit_sh_right(c_var_string* left);
    c_var_base& bit_sh_right(c_var_array* left);

    c_var_base& operator>(c_var_base& right);
    c_var_base& greater(c_var_bool* left);
    c_var_base& greater(c_var_int* left);
    c_var_base& greater(c_var_unset* left);
    c_var_base& greater(c_var_string* left);
    c_var_base& greater(c_var_array* left);

    c_var_base& operator<(c_var_base& right);
    c_var_base& less(c_var_bool* left);
    c_var_base& less(c_var_int* left);
    c_var_base& less(c_var_unset* left);
    c_var_base& less(c_var_string* left);
    c_var_base& less(c_var_array* left);

    c_var_base& operator>=(c_var_base& right);
    c_var_base& greater_or_eq(c_var_bool* left);
    c_var_base& greater_or_eq(c_var_int* left);
    c_var_base& greater_or_eq(c_var_unset* left);
    c_var_base& greater_or_eq(c_var_string* left);
    c_var_base& greater_or_eq(c_var_array* left);

    c_var_base& operator<=(c_var_base& right);
    c_var_base& less_or_eq(c_var_bool* left);
    c_var_base& less_or_eq(c_var_int* left);
    c_var_base& less_or_eq(c_var_unset* left);
    c_var_base& less_or_eq(c_var_string* left);
    c_var_base& less_or_eq(c_var_array* left);

    c_var_base& operator~();

    c_var_base& operator&&(c_var_base& right);
    c_var_base& log_and(c_var_bool* left);
    c_var_base& log_and(c_var_int* left);
    c_var_base& log_and(c_var_unset* left);
    c_var_base& log_and(c_var_string* left);
    c_var_base& log_and(c_var_array* left);

    c_var_base& operator||(c_var_base& right);
    c_var_base& log_or(c_var_bool* left);
    c_var_base& log_or(c_var_int* left);
    c_var_base& log_or(c_var_unset* left);
    c_var_base& log_or(c_var_string* left);
    c_var_base& log_or(c_var_array* left);
    // END

    bool operator==(const string& right) const;
    bool operator==(const MY_LONG_LONG& right) const;
};

#endif

