/***************************************************************************
                          c_var_int.cpp  -  description
                             -------------------
    begin                : Thu May 10 2007
    copyright            : (C) 2004-2007 by Marian VooDooMan Meravy
    email                : ghostvoodooman -a.T= users (dot) sourceforge.net
 ***************************************************************************/

/* $Id: c_var_int.cpp,v 1.16 2008/11/24 12:09:46 ghostvoodooman Exp $ */

/***************************************************************************

VooDoo cIRCle - an IRC (ro)bot
Copyright (C) 2004-2008 by Marian VooDooMan Meravy
  (ghostvoodooman -a.T= users (dot) sourceforge.net)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************/

/*!
    \file
    \brief Class 'c_var_int'
    \version $Id: c_var_int.cpp,v 1.16 2008/11/24 12:09:46 ghostvoodooman Exp $
*/

#if defined(HAVE_CONFIG_H)
#   include <config.h>
#endif

#include "../params.h"
#include "../utils/memleak.h"
#include "../utils/assert.h"
#include "../compat.h"

#include "c_var_bool.h"
#include "c_var_int.h"
#include "../var_keeper.h"
#include "c_var_string.h"
#include "../md5.h"
#include "../utils.h"

#include "../utils/precomp.h"

#if defined(_MSC_VER) || defined(__BORLANDC__)
#   pragma hdrstop
#endif

c_var_bool* c_var_int::to_bool()
{
    return MY_NEW c_var_bool(is_true());
}

c_var_int::c_var_int()
    : c_var_base()
{
    value_=0;
    var_keeper->include_into_locals(this);
}

c_var_int::c_var_int(const c_var_int& orig)
    : c_var_base(orig)
{
    value_=orig.value_;
    //var_keeper->include_into_locals(this);
}

c_var_int::c_var_int(const c_var_int& orig, int)
    : c_var_base(orig)
{
    value_=orig.value_;
}

c_var_int::c_var_int(MY_LONG_LONG value)
    : c_var_base()
{
    value_=value;
    var_keeper->include_into_locals(this);
}

c_var_int::c_var_int(MY_LONG_LONG value, bool persistent)
    : c_var_base()
{
    value_=value;
    set_persistent(persistent);
    var_keeper->include_into_locals(this);
}

e_vartype c_var_int::get_var_type()
{
    return VARTYPE_INT;
}

e_vartype c_var_int::get_var_type() const
{
    return VARTYPE_INT;
}

string c_var_int::get_var_type_as_string()
{
    return (string)"INT";
}

c_var_int* c_var_int::copy()
{
    c_var_int* res=MY_NEW c_var_int(*this,0);
    res->set_persistent(false);
    res->set_read_only(false);
    res->reset_id();
    res->id=var_keeper->alloc(res);
    res->set_session(var_keeper->get_current_session());
    var_keeper->include_into_locals(res);
    return res;
}

c_var_int::~c_var_int()
{
}

MY_LONG_LONG c_var_int::get()
{
    return value_;
}

void c_var_int::set(MY_LONG_LONG value)
{
    value_=value;
}

c_var_string* c_var_int::to_string(c_module* this_module, c_var_int::e_return_on_error UNUSEDARG(return_on_error), c_var_int::e_num_system numeric_system, const string& debug)
{
    switch(numeric_system) {
        case NS_DECIMAL:
            {
                return MY_NEW c_var_string(tos(get()));
            }
    }

    var_oops(this_module,__FILE__,__my_func__,"Unsupported numeric system for use to convert INT to STRING - returning an empty string",debug);
    MY_ASSERT_NOTREACH();
    return MY_NEW c_var_string("");
}

string c_var_int::dump(c_module* this_module)
{
    string res="(INT) ";
    res+=to_string(this_module);
    return res;
}

void c_var_int::dump_for_botnet(c_module* UNUSEDARG(this_module), char* &data, size_t& size)
{
    size=1+1+1+sizeof(value_);
    char* d=MY_NEW char[size];
    d[0]='I'; // 'I' stands for int
    d[1]=static_cast<unsigned char>(sizeof(value_));
    memcpy(&(d[2]),&value_,sizeof(value_));
#ifdef WORDS_BIGENDIAN
    byte_stream_reverse(&(d[2]),sizeof(value_));
#endif
    d[2+sizeof(value_)]=0;
    data=d;
}

string c_var_int::get_script_string(c_module* this_module, const string& script_type)
{
    if(!script_type.compare("php") || !script_type.compare("php_2") || !script_type.compare("logic.txt")) {
        return to_string(this_module);
    }
    else if(!script_type.compare("perl")) {
        return to_string(this_module);
    }

    // error, not supported script
    return (string)"";
}

c_var_base* c_var_int::copy(int)
{
    c_var_int* res=MY_NEW c_var_int(*this,0);
    res->set_persistent(false);
    res->set_read_only(false);
    res->reset_id();
    res->id=var_keeper->alloc(res);
    res->set_session(var_keeper->get_current_session());
    //var_keeper->include_into_locals(res);
    return res;
}

c_index_base* c_var_int::make_index()
{
    return MY_NEW c_index_int(value_);
}

bool c_var_int::is_true()
{
    return value_ ? true : false;
}

bool c_var_int::is_true() const
{
    return value_ ? true : false;
}

string c_var_int::to_string(c_module* this_module)
{
    string debug=__my_func__;
    c_var_string* res=to_string(this_module?this_module:core_mod,c_var_int::RET_STR_EMPTY,c_var_int::NS_DECIMAL,debug);
    string r;
    if(res) {
        r=res->get();
        MY_DELETE(res);
    }
    return r;
}

string c_var_int::to_string(c_module* this_module) const
{
    string debug=__my_func__;
    c_var_string* res=const_cast<c_var_int*>(this)->to_string(this_module?this_module:core_mod,c_var_int::RET_STR_EMPTY,c_var_int::NS_DECIMAL,debug);
    string r;
    if(res) {
        r=res->get();
        MY_DELETE(res);
    }
    return r;
}

bool c_var_int::operator==(const string& right) const
{
    MY_LONG_LONG num;
    char tmp[32];
    if(sscanf(right.c_str(),"%" MY_LONG_LONG_STR "d%16s",&num,tmp)!=1) {
        // it's not number, or trailing string is followed by number
        return false;
    }
    return num==value_;
}

bool c_var_int::operator==(const MY_LONG_LONG& right) const
{
    return right==value_;
}

