/***************************************************************************
                          c_var_int.h  -  description
                             -------------------
    begin                : Thu May 10 2007
    copyright            : (C) 2004-2007 by Marian VooDooMan Meravy
    email                : ghostvoodooman -a.T= users (dot) sourceforge.net
 ***************************************************************************/

/* $Id: c_var_int.h,v 1.15 2008/03/25 12:04:30 ghostvoodooman Exp $ */

/***************************************************************************

VooDoo cIRCle - an IRC (ro)bot
Copyright (C) 2004-2008 by Marian VooDooMan Meravy
  (ghostvoodooman -a.T= users (dot) sourceforge.net)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************/

/*!
    \file
    \brief Class 'c_var_int'
    \version $Id: c_var_int.h,v 1.15 2008/03/25 12:04:30 ghostvoodooman Exp $
*/

#ifndef C_VAR_INT_H__
#define C_VAR_INT_H__

#if defined(HAVE_CONFIG_H)
#   include <config.h>
#endif

#include <string>
#include <vector>
#include <stack>
#include <list>
#include <map>
#include <algorithm>

#include "../utils/memleak.h"

#include "../params.h"

/*!
    \author VooDooMan
    \version 3
    \date 2005, 2007-01-14, 2008-03-09
    \brief Class for integer variable
*/
class c_var_int : public c_var_base
{
private:
    MY_LONG_LONG value_;                    //!< Content of int

public:
    c_var_int();                            //!< Constructor

    /*!
        \brief Copy-constructor
        \param orig Original instance
    */
    c_var_int(const c_var_int& orig);

    /*!
        \brief Copy-constructor, but not calling c_var_keeper
        \param orig Original instance
    */
    c_var_int(const c_var_int& orig, int);

    /*!
        \brief Constructor
        \param value Initial value
    */
    c_var_int(MY_LONG_LONG value);

    /*!
        \brief Constructor
        \param value Initial value
        \param persistent Persistency flag
    */
    c_var_int(MY_LONG_LONG value, bool persistent);

    virtual ~c_var_int();                       //!< Destructor

    virtual e_vartype get_var_type();           //!< Gets type of variable
    virtual e_vartype get_var_type() const;     //!< Gets type of variable
    virtual string get_var_type_as_string();    //!< Gets type of variable as string

    virtual c_var_int* copy();                  //!< Creates copy of itself

    MY_LONG_LONG get();                         //!< Gets variable's value

    /*!
        \brief Sets variable's value
        \param value Value to set
    */
    void set(MY_LONG_LONG value);

    /*!
        \brief Numeric system
    */
    enum e_num_system
    {
        NS_DECIMAL = 1,     //!< Decimal
    };

    /*!
        \brief Error-control type
    */
    enum e_return_on_error
    {
        RET_NULL,           //!< Return NULL on error
        RET_STR_ZERO,       //!< return (string)"0" string on error
        RET_STR_EMPTY,      //!< Return empty string on error
    };

    /*!
        \brief Convert itself to string
        \param return_on_error Error-control method
        \param numeric_system Use this numeric system
        \param debug Debug string in case of error
        \param this_module Module
    */
    c_var_string* to_string(c_module* this_module, c_var_int::e_return_on_error return_on_error, c_var_int::e_num_system numeric_system, const string& debug);

    bool is_true();
    bool is_true() const;

    /*!
        \brief Converts to bool and returns result as c_var_bool
    */
    c_var_bool* to_bool();

    string to_string(c_module* this_module = NULL);
    string to_string(c_module* this_module = NULL) const;

    virtual string dump(c_module* this_module);

    virtual void dump_for_botnet(c_module* this_module, char* &data, size_t& size);

    string get_script_string(c_module* this_module, const string& script_type);

    virtual c_var_base* copy(int);

    virtual c_index_base* make_index();

    // BEGIN: Variable operations - prototypes follows
    c_var_base& operator++(int);

    c_var_base& operator--(int);

    c_var_base& operator++();

    c_var_base& operator--();

    void pre_inc();
    void pre_dec();
    void post_inc();
    void post_dec();

    c_var_base& operator+(c_var_base& right);
    c_var_base& plus(c_var_bool* left);
    c_var_base& plus(c_var_int* left);
    c_var_base& plus(c_var_unset* left);
    c_var_base& plus(c_var_string* left);
    c_var_base& plus(c_var_array* left);

    c_var_base& operator-(c_var_base& right);
    c_var_base& minus(c_var_bool* left);
    c_var_base& minus(c_var_int* left);
    c_var_base& minus(c_var_unset* left);
    c_var_base& minus(c_var_string* left);
    c_var_base& minus(c_var_array* left);

    c_var_base& operator*(c_var_base& right);
    c_var_base& mul(c_var_bool* left);
    c_var_base& mul(c_var_int* left);
    c_var_base& mul(c_var_unset* left);
    c_var_base& mul(c_var_string* left);
    c_var_base& mul(c_var_array* left);

    c_var_base& operator/(c_var_base& right);
    c_var_base& div(c_var_bool* left);
    c_var_base& div(c_var_int* left);
    c_var_base& div(c_var_unset* left);
    c_var_base& div(c_var_string* left);
    c_var_base& div(c_var_array* left);

    c_var_base& operator!();

    c_var_base& operator==(c_var_base& right);
    c_var_base& equal(c_var_bool* left);
    c_var_base& equal(c_var_int* left);
    c_var_base& equal(c_var_unset* left);
    c_var_base& equal(c_var_string* left);
    c_var_base& equal(c_var_array* left);

    c_var_base& operator!=(c_var_base& right);
    c_var_base& not_equal(c_var_bool* left);
    c_var_base& not_equal(c_var_int* left);
    c_var_base& not_equal(c_var_unset* left);
    c_var_base& not_equal(c_var_string* left);
    c_var_base& not_equal(c_var_array* left);

    c_var_base& concat(c_var_base& right);
    c_var_base& concat(c_var_bool* left);
    c_var_base& concat(c_var_int* left);
    c_var_base& concat(c_var_unset* left);
    c_var_base& concat(c_var_string* left);
    c_var_base& concat(c_var_array* left);

    c_var_base& operator%(c_var_base& right);
    c_var_base& mod(c_var_bool* left);
    c_var_base& mod(c_var_int* left);
    c_var_base& mod(c_var_unset* left);
    c_var_base& mod(c_var_string* left);
    c_var_base& mod(c_var_array* left);

    c_var_base& operator&(c_var_base& right);
    c_var_base& bit_and(c_var_bool* left);
    c_var_base& bit_and(c_var_int* left);
    c_var_base& bit_and(c_var_unset* left);
    c_var_base& bit_and(c_var_string* left);
    c_var_base& bit_and(c_var_array* left);

    c_var_base& operator|(c_var_base& right);
    c_var_base& bit_or(c_var_bool* left);
    c_var_base& bit_or(c_var_int* left);
    c_var_base& bit_or(c_var_unset* left);
    c_var_base& bit_or(c_var_string* left);
    c_var_base& bit_or(c_var_array* left);

    c_var_base& operator^(c_var_base& right);
    c_var_base& bit_xor(c_var_bool* left);
    c_var_base& bit_xor(c_var_int* left);
    c_var_base& bit_xor(c_var_unset* left);
    c_var_base& bit_xor(c_var_string* left);
    c_var_base& bit_xor(c_var_array* left);

    c_var_base& operator<<(c_var_base& right);
    c_var_base& bit_sh_left(c_var_bool* left);
    c_var_base& bit_sh_left(c_var_int* left);
    c_var_base& bit_sh_left(c_var_unset* left);
    c_var_base& bit_sh_left(c_var_string* left);
    c_var_base& bit_sh_left(c_var_array* left);

    c_var_base& operator>>(c_var_base& right);
    c_var_base& bit_sh_right(c_var_bool* left);
    c_var_base& bit_sh_right(c_var_int* left);
    c_var_base& bit_sh_right(c_var_unset* left);
    c_var_base& bit_sh_right(c_var_string* left);
    c_var_base& bit_sh_right(c_var_array* left);

    c_var_base& operator>(c_var_base& right);
    c_var_base& greater(c_var_bool* left);
    c_var_base& greater(c_var_int* left);
    c_var_base& greater(c_var_unset* left);
    c_var_base& greater(c_var_string* left);
    c_var_base& greater(c_var_array* left);

    c_var_base& operator<(c_var_base& right);
    c_var_base& less(c_var_bool* left);
    c_var_base& less(c_var_int* left);
    c_var_base& less(c_var_unset* left);
    c_var_base& less(c_var_string* left);
    c_var_base& less(c_var_array* left);

    c_var_base& operator>=(c_var_base& right);
    c_var_base& greater_or_eq(c_var_bool* left);
    c_var_base& greater_or_eq(c_var_int* left);
    c_var_base& greater_or_eq(c_var_unset* left);
    c_var_base& greater_or_eq(c_var_string* left);
    c_var_base& greater_or_eq(c_var_array* left);

    c_var_base& operator<=(c_var_base& right);
    c_var_base& less_or_eq(c_var_bool* left);
    c_var_base& less_or_eq(c_var_int* left);
    c_var_base& less_or_eq(c_var_unset* left);
    c_var_base& less_or_eq(c_var_string* left);
    c_var_base& less_or_eq(c_var_array* left);

    c_var_base& operator~();

    c_var_base& operator&&(c_var_base& right);
    c_var_base& log_and(c_var_bool* left);
    c_var_base& log_and(c_var_int* left);
    c_var_base& log_and(c_var_unset* left);
    c_var_base& log_and(c_var_string* left);
    c_var_base& log_and(c_var_array* left);

    c_var_base& operator||(c_var_base& right);
    c_var_base& log_or(c_var_bool* left);
    c_var_base& log_or(c_var_int* left);
    c_var_base& log_or(c_var_unset* left);
    c_var_base& log_or(c_var_string* left);
    c_var_base& log_or(c_var_array* left);
    // END

    bool operator==(const string& right) const;
    bool operator==(const MY_LONG_LONG& right) const;
};

#endif

