/***************************************************************************
                          c_var_ref.cpp  -  description
                             -------------------
    begin                : Thu May 10 2007
    copyright            : (C) 2004-2007 by Marian VooDooMan Meravy
    email                : ghostvoodooman -a.T= users (dot) sourceforge.net
 ***************************************************************************/

/* $Id: c_var_ref.cpp,v 1.15 2008/03/25 12:04:30 ghostvoodooman Exp $ */

/***************************************************************************

VooDoo cIRCle - an IRC (ro)bot
Copyright (C) 2004-2008 by Marian VooDooMan Meravy
  (ghostvoodooman -a.T= users (dot) sourceforge.net)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************/

/*!
    \file
    \brief Class 'c_var_ref'
    \version $Id: c_var_ref.cpp,v 1.15 2008/03/25 12:04:30 ghostvoodooman Exp $
*/

#if defined(HAVE_CONFIG_H)
#   include <config.h>
#endif

#include "../params.h"
#include "../utils/memleak.h"
#include "../utils/assert.h"
#include "../compat.h"

#include "c_var_ref.h"
#include "../var_keeper.h"

#include "../utils/precomp.h"

#if defined(_MSC_VER) || defined(__BORLANDC__)
#   pragma hdrstop
#endif

c_var_ref::c_var_ref(var_id_t reference_from, var_id_t reference_to)
    : c_var_base()
{
    reference_from_=reference_from;
    reference_to_=reference_to;
    var_keeper->include_into_locals(this);
}

c_var_ref::c_var_ref(const c_var_ref& orig)
    : c_var_base(orig)
{
    reference_from_=orig.reference_from_;
    reference_to_=orig.reference_to_;
    //var_keeper->include_into_locals(this);
}

c_var_ref::c_var_ref(const c_var_ref& orig, int)
    : c_var_base(orig)
{
    reference_from_=orig.reference_from_;
    reference_to_=orig.reference_to_;
}

e_vartype c_var_ref::get_var_type()
{
    if(!reference_to_)
        return VARTYPE_NULL;

    var_id_t ref=reference_to_;
    for(;;) {
        c_var_base* res=var_keeper->get(ref,NULL,NULL);
        c_var_ref* ref2=NULL;
        if(res && typeid(*res)==typeid(c_var_ref))
            ref2=dynamic_cast<c_var_ref*>(res);
        if(ref2) {
            ref=ref2->reference_to_;
            continue;
        }
        return res?res->get_var_type():VARTYPE_UNSET;
    }
}

e_vartype c_var_ref::get_var_type() const
{
    if(!reference_to_)
        return VARTYPE_NULL;

    var_id_t ref=reference_to_;
    for(;;) {
        c_var_base* res=var_keeper->get(ref,NULL,NULL);
        c_var_ref* ref2=NULL;
        if(res && typeid(*res)==typeid(c_var_ref))
            ref2=dynamic_cast<c_var_ref*>(res);
        if(ref2) {
            ref=ref2->reference_to_;
            continue;
        }
        return res?res->get_var_type():VARTYPE_UNSET;
    }
}

string c_var_ref::get_var_type_as_string()
{
    if(!reference_to_)
        return (string)"NULL";

    var_id_t ref=reference_to_;
    for(;;) {
        c_var_base* res=var_keeper->get(ref,NULL,NULL);
        c_var_ref* ref2=NULL;
        if(res && typeid(*res)==typeid(c_var_ref))
            ref2=dynamic_cast<c_var_ref*>(res);
        if(ref2) {
            ref=ref2->reference_to_;
            continue;
        }
        return res?res->get_var_type_as_string():(string)"NULL";
    }
}

c_var_ref* c_var_ref::copy()
{
    c_var_ref* res=MY_NEW c_var_ref(*this,0);
    res->persistent_=false;
    res->read_only=false;
    res->reset_id();
    res->id=var_keeper->alloc(res);
    res->set_session(var_keeper->get_current_session());
    var_keeper->references.insert(make_pair(res->id,res->reference_to_));
    var_keeper->include_into_locals(res);
    return res;
}

bool c_var_ref::is_true()
{
    if(!reference_to_)
        return false;

    var_id_t ref=reference_to_;
    for(;;) {
        c_var_base* res=var_keeper->get(ref,NULL,NULL);
        c_var_ref* ref2=NULL;
        if(res && typeid(*res)==typeid(c_var_ref))
            ref2=dynamic_cast<c_var_ref*>(res);
        if(ref2) {
            ref=ref2->reference_to_;
            continue;
        }
        return res?res->is_true():false;
    }
}

bool c_var_ref::is_true() const
{
    if(!reference_to_)
        return false;

    var_id_t ref=reference_to_;
    for(;;) {
        c_var_base* res=var_keeper->get(ref,NULL,NULL);
        c_var_ref* ref2=NULL;
        if(res && typeid(*res)==typeid(c_var_ref))
            ref2=dynamic_cast<c_var_ref*>(res);
        if(ref2) {
            ref=ref2->reference_to_;
            continue;
        }
        return res?res->is_true():false;
    }
}

c_var_bool* c_var_ref::to_bool()
{
    if(!reference_to_)
        return NULL;

    var_id_t ref=reference_to_;
    for(;;) {
        c_var_base* res=var_keeper->get(ref,NULL,NULL);
        c_var_ref* ref2=NULL;
        if(res && typeid(*res)==typeid(c_var_ref))
            ref2=dynamic_cast<c_var_ref*>(res);
        if(ref2) {
            ref=ref2->reference_to_;
            continue;
        }
        return res?res->to_bool():NULL;
    }
}

string c_var_ref::dump(c_module* this_module)
{
    if(!reference_to_)
        return (string)"";

    var_id_t ref=reference_to_;
    for(;;) {
        c_var_base* res=var_keeper->get(ref,NULL,NULL);
        c_var_ref* ref2=NULL;
        if(res && typeid(*res)==typeid(c_var_ref))
            ref2=dynamic_cast<c_var_ref*>(res);
        if(ref2) {
            ref=ref2->reference_to_;
            continue;
        }
        return res?res->dump(this_module):(string)"";
    }
}

string c_var_ref::get_script_string(c_module* this_module, const string& script_type)
{
    if(!reference_to_)
        return (string)"";

    var_id_t ref=reference_to_;
    for(;;) {
        c_var_base* res=var_keeper->get(ref,NULL,NULL);
        c_var_ref* ref2=NULL;
        if(res && typeid(*res)==typeid(c_var_ref))
            ref2=dynamic_cast<c_var_ref*>(res);
        if(ref2) {
            ref=ref2->reference_to_;
            continue;
        }
        return res?res->get_script_string(this_module,script_type):(string)"";
    }
}

string c_var_ref::to_string(c_module* this_module)
{
    if(!reference_to_)
        return (string)"";

    var_id_t ref=reference_to_;
    for(;;) {
        c_var_base* res=var_keeper->get(ref,NULL,NULL);
        c_var_ref* ref2=NULL;
        if(res && typeid(*res)==typeid(c_var_ref))
            ref2=dynamic_cast<c_var_ref*>(res);
        if(ref2) {
            ref=ref2->reference_to_;
            continue;
        }
        return res?res->to_string(this_module?this_module:core_mod):(string)"";
    }
}

string c_var_ref::to_string(c_module* this_module) const
{
    if(!reference_to_)
        return (string)"";

    var_id_t ref=reference_to_;
    for(;;) {
        c_var_base* res=var_keeper->get(ref,NULL,NULL);
        c_var_ref* ref2=NULL;
        if(res && typeid(*res)==typeid(c_var_ref))
            ref2=dynamic_cast<c_var_ref*>(res);
        if(ref2) {
            ref=ref2->reference_to_;
            continue;
        }
        return res?res->to_string(this_module):(string)"";
    }
}

void c_var_ref::dump_for_botnet(c_module* this_module, char* &data, size_t& size)
{
    if(!reference_to_) {
        data=MY_NEW char[1];
        *data='\0';
        size=0;
        return;
    }

    var_id_t ref=reference_to_;
    for(;;) {
        c_var_base* res=var_keeper->get(ref,NULL,NULL);
        c_var_ref* ref2=NULL;
        if(res && typeid(*res)==typeid(c_var_ref))
            ref2=dynamic_cast<c_var_ref*>(res);
        if(ref2) {
            ref=ref2->reference_to_;
            continue;
        }
        if(!res) {
            data=MY_NEW char[1];
            *data='\0';
            size=0;
            return;
        }
        res->dump_for_botnet(this_module,data,size);
        return;
    }
}

void c_var_ref::set_read_only(bool ro)
{
    if(!reference_to_)
        return;

    var_id_t ref=reference_to_;
    for(;;) {
        c_var_base* res=var_keeper->get(ref,NULL,NULL);
        res->set_read_only(ro);
        c_var_ref* ref2=NULL;
        if(res && typeid(*res)==typeid(c_var_ref))
            ref2=dynamic_cast<c_var_ref*>(res);
        if(ref2) {
            ref=ref2->reference_to_;
            continue;
        }
        break;
    }
}

bool c_var_ref::get_read_only()
{
    if(!reference_to_)
        return false;

    var_id_t ref=reference_to_;
    for(;;) {
        c_var_base* res=var_keeper->get(ref,NULL,NULL);
        c_var_ref* ref2=NULL;
        if(res && typeid(*res)==typeid(c_var_ref))
            ref2=dynamic_cast<c_var_ref*>(res);
        if(ref2) {
            ref=ref2->reference_to_;
            continue;
        }
        return res?res->get_read_only():false;
    }
}

void c_var_ref::set_persistent(bool persistent)
{
    if(!reference_to_)
        return;

    var_id_t ref=reference_to_;
    for(;;) {
        c_var_base* res=var_keeper->get(ref,NULL,NULL);
        res->set_persistent(persistent);
        c_var_ref* ref2=NULL;
        if(res && typeid(*res)==typeid(c_var_ref))
            ref2=dynamic_cast<c_var_ref*>(res);
        if(ref2) {
            ref=ref2->reference_to_;
            continue;
        }
        break;
    }
}

bool c_var_ref::get_persistent()
{
    if(!reference_to_)
        return false;

    var_id_t ref=reference_to_;
    for(;;) {
        c_var_base* res=var_keeper->get(ref,NULL,NULL);
        c_var_ref* ref2=NULL;
        if(res && typeid(*res)==typeid(c_var_ref))
            ref2=dynamic_cast<c_var_ref*>(res);
        if(ref2) {
            ref=ref2->reference_to_;
            continue;
        }
        return res?res->get_persistent():false;
    }
}

var_id_t c_var_ref::get_id()
{
    return id;
}

c_var_ref::~c_var_ref()
{
    var_keeper->free_reference(id,NULL,NULL);
}

c_var_base* c_var_ref::copy(int)
{
    c_var_ref* res=MY_NEW c_var_ref(*this,0);
    res->persistent_=false;
    res->read_only=false;
    res->reset_id();
    res->id=var_keeper->alloc(res);
    res->set_session(var_keeper->get_current_session());
    var_keeper->references.insert(make_pair(res->id,res->reference_to_));
    //var_keeper->include_into_locals(res);
    return res;
}

c_index_base* c_var_ref::make_index()
{
    // invalid index
    return NULL;
}

bool c_var_ref::operator==(const string& right) const
{
    if(!reference_to_)
        return false;

    var_id_t ref=reference_to_;
    c_var_base* res;
    for(;;) {
        res=var_keeper->get(ref,NULL,NULL);
        c_var_ref* ref2=NULL;
        if(res && typeid(*res)==typeid(c_var_ref))
            ref2=dynamic_cast<c_var_ref*>(res);
        if(ref2) {
            ref=ref2->reference_to_;
            continue;
        }
        break;
    }
    return res ? (*res)==right : false;
}

bool c_var_ref::operator==(const MY_LONG_LONG& right) const
{
    if(!reference_to_)
        return false;

    var_id_t ref=reference_to_;
    c_var_base* res;
    for(;;) {
        res=var_keeper->get(ref,NULL,NULL);
        c_var_ref* ref2=NULL;
        if(res && typeid(*res)==typeid(c_var_ref))
            ref2=dynamic_cast<c_var_ref*>(res);
        if(ref2) {
            ref=ref2->reference_to_;
            continue;
        }
        break;
    }
    return res ? (*res)==right : false;
}

