/***************************************************************************
                          c_var_ref.h  -  description
                             -------------------
    begin                : Thu May 10 2007
    copyright            : (C) 2004-2007 by Marian VooDooMan Meravy
    email                : ghostvoodooman -a.T= users (dot) sourceforge.net
 ***************************************************************************/

/* $Id: c_var_ref.h,v 1.14 2008/03/11 21:11:49 ghostvoodooman Exp $ */

/***************************************************************************

VooDoo cIRCle - an IRC (ro)bot
Copyright (C) 2004-2008 by Marian VooDooMan Meravy
  (ghostvoodooman -a.T= users (dot) sourceforge.net)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************/

/*!
    \file
    \brief Class 'c_var_ref'
    \version $Id: c_var_ref.h,v 1.14 2008/03/11 21:11:49 ghostvoodooman Exp $
*/

#ifndef C_VAR_REF_H__
#define C_VAR_REF_H__

#if defined(HAVE_CONFIG_H)
#   include <config.h>
#endif

#include <string>
#include <vector>
#include <stack>
#include <list>
#include <map>
#include <algorithm>

#include "../utils/memleak.h"

#include "../params.h"

#include "c_var_base.h"

/*!
    \author VooDooMan
    \version 2
    \date 2006-05-13, 2008-03-09
    \brief Class for referenced variable (placeholder)
*/
class c_var_ref : public c_var_base
{
    var_id_t reference_from_;                   //!< Reference to ID of variable
    var_id_t reference_to_;                     //!< Reference to ID of variable

    friend class c_var_keeper;

public:
    c_var_ref(var_id_t reference_from, var_id_t reference_to);           //!< Constructor

    /*!
        \brief Resets reference pointer
        \author VooDooMan
        \date 2006-05-13
        \version 1
    */
    void reset()
    {
        reference_from_=VAR_ID_INVALID;
        reference_to_=VAR_ID_INVALID;
    }

    /*!
        \brief Gets ID of variable referenced to
        \author VooDooMan
        \date 2006-05-13
        \version 1
    */
    var_id_t get_referenced()
    {
        return reference_to_;
    }

    /*!
        \brief Gets ID of variable referencing from
        \author VooDooMan
        \date 2006-05-16
        \version 1
    */
    var_id_t get_referencing()
    {
        return reference_from_;
    }

    /*!
        \brief Copy-constructor
        \param orig Original instance
    */
    c_var_ref(const c_var_ref& orig);

    /*!
        \brief Copy-constructor, but not calling c_var_keeper
        \param orig Original instance
    */
    c_var_ref(const c_var_ref& orig, int);

    virtual e_vartype get_var_type();           //!< Gets type of variable
    virtual e_vartype get_var_type() const;     //!< Gets type of variable
    virtual string get_var_type_as_string();    //!< Gets type of variable as string

    virtual c_var_ref* copy();                  //!< Creates copy of itself

    bool is_true();
    bool is_true() const;

    c_var_bool* to_bool();

    string to_string(c_module* this_module = NULL);
    string to_string(c_module* this_module = NULL) const;

    virtual string dump(c_module* this_module);

    virtual void dump_for_botnet(c_module* this_module, char* &data, size_t& size);

    string get_script_string(c_module* this_module, const string& script_type);

    void set_read_only(bool ro = true);
    bool get_read_only();

    void set_persistent(bool persistent = true);
    bool get_persistent();

    var_id_t get_id();

    //void get_parent(c_var_array* &parent, c_var_base* &index);

    virtual ~c_var_ref();           //!< Destructor

    virtual c_var_base* copy(int);

    virtual c_index_base* make_index();

    // BEGIN: Variable operations - prototypes follows
    c_var_base& operator++(int);

    c_var_base& operator--(int);

    c_var_base& operator++();

    c_var_base& operator--();

    void pre_inc();
    void pre_dec();
    void post_inc();
    void post_dec();

    c_var_base& operator+(c_var_base& right);
    c_var_base& plus(c_var_bool* left);
    c_var_base& plus(c_var_int* left);
    c_var_base& plus(c_var_unset* left);
    c_var_base& plus(c_var_string* left);
    c_var_base& plus(c_var_array* left);

    c_var_base& operator-(c_var_base& right);
    c_var_base& minus(c_var_bool* left);
    c_var_base& minus(c_var_int* left);
    c_var_base& minus(c_var_unset* left);
    c_var_base& minus(c_var_string* left);
    c_var_base& minus(c_var_array* left);

    c_var_base& operator*(c_var_base& right);
    c_var_base& mul(c_var_bool* left);
    c_var_base& mul(c_var_int* left);
    c_var_base& mul(c_var_unset* left);
    c_var_base& mul(c_var_string* left);
    c_var_base& mul(c_var_array* left);

    c_var_base& operator/(c_var_base& right);
    c_var_base& div(c_var_bool* left);
    c_var_base& div(c_var_int* left);
    c_var_base& div(c_var_unset* left);
    c_var_base& div(c_var_string* left);
    c_var_base& div(c_var_array* left);

    c_var_base& operator!();

    c_var_base& operator==(c_var_base& right);
    c_var_base& equal(c_var_bool* left);
    c_var_base& equal(c_var_int* left);
    c_var_base& equal(c_var_unset* left);
    c_var_base& equal(c_var_string* left);
    c_var_base& equal(c_var_array* left);

    c_var_base& operator!=(c_var_base& right);
    c_var_base& not_equal(c_var_bool* left);
    c_var_base& not_equal(c_var_int* left);
    c_var_base& not_equal(c_var_unset* left);
    c_var_base& not_equal(c_var_string* left);
    c_var_base& not_equal(c_var_array* left);

    c_var_base& concat(c_var_base& right);
    c_var_base& concat(c_var_bool* left);
    c_var_base& concat(c_var_int* left);
    c_var_base& concat(c_var_unset* left);
    c_var_base& concat(c_var_string* left);
    c_var_base& concat(c_var_array* left);

    c_var_base& operator%(c_var_base& right);
    c_var_base& mod(c_var_bool* left);
    c_var_base& mod(c_var_int* left);
    c_var_base& mod(c_var_unset* left);
    c_var_base& mod(c_var_string* left);
    c_var_base& mod(c_var_array* left);

    c_var_base& operator&(c_var_base& right);
    c_var_base& bit_and(c_var_bool* left);
    c_var_base& bit_and(c_var_int* left);
    c_var_base& bit_and(c_var_unset* left);
    c_var_base& bit_and(c_var_string* left);
    c_var_base& bit_and(c_var_array* left);

    c_var_base& operator|(c_var_base& right);
    c_var_base& bit_or(c_var_bool* left);
    c_var_base& bit_or(c_var_int* left);
    c_var_base& bit_or(c_var_unset* left);
    c_var_base& bit_or(c_var_string* left);
    c_var_base& bit_or(c_var_array* left);

    c_var_base& operator^(c_var_base& right);
    c_var_base& bit_xor(c_var_bool* left);
    c_var_base& bit_xor(c_var_int* left);
    c_var_base& bit_xor(c_var_unset* left);
    c_var_base& bit_xor(c_var_string* left);
    c_var_base& bit_xor(c_var_array* left);

    c_var_base& operator<<(c_var_base& right);
    c_var_base& bit_sh_left(c_var_bool* left);
    c_var_base& bit_sh_left(c_var_int* left);
    c_var_base& bit_sh_left(c_var_unset* left);
    c_var_base& bit_sh_left(c_var_string* left);
    c_var_base& bit_sh_left(c_var_array* left);

    c_var_base& operator>>(c_var_base& right);
    c_var_base& bit_sh_right(c_var_bool* left);
    c_var_base& bit_sh_right(c_var_int* left);
    c_var_base& bit_sh_right(c_var_unset* left);
    c_var_base& bit_sh_right(c_var_string* left);
    c_var_base& bit_sh_right(c_var_array* left);

    c_var_base& operator>(c_var_base& right);
    c_var_base& greater(c_var_bool* left);
    c_var_base& greater(c_var_int* left);
    c_var_base& greater(c_var_unset* left);
    c_var_base& greater(c_var_string* left);
    c_var_base& greater(c_var_array* left);

    c_var_base& operator<(c_var_base& right);
    c_var_base& less(c_var_bool* left);
    c_var_base& less(c_var_int* left);
    c_var_base& less(c_var_unset* left);
    c_var_base& less(c_var_string* left);
    c_var_base& less(c_var_array* left);

    c_var_base& operator>=(c_var_base& right);
    c_var_base& greater_or_eq(c_var_bool* left);
    c_var_base& greater_or_eq(c_var_int* left);
    c_var_base& greater_or_eq(c_var_unset* left);
    c_var_base& greater_or_eq(c_var_string* left);
    c_var_base& greater_or_eq(c_var_array* left);

    c_var_base& operator<=(c_var_base& right);
    c_var_base& less_or_eq(c_var_bool* left);
    c_var_base& less_or_eq(c_var_int* left);
    c_var_base& less_or_eq(c_var_unset* left);
    c_var_base& less_or_eq(c_var_string* left);
    c_var_base& less_or_eq(c_var_array* left);

    c_var_base& operator~();

    c_var_base& operator&&(c_var_base& right);
    c_var_base& log_and(c_var_bool* left);
    c_var_base& log_and(c_var_int* left);
    c_var_base& log_and(c_var_unset* left);
    c_var_base& log_and(c_var_string* left);
    c_var_base& log_and(c_var_array* left);

    c_var_base& operator||(c_var_base& right);
    c_var_base& log_or(c_var_bool* left);
    c_var_base& log_or(c_var_int* left);
    c_var_base& log_or(c_var_unset* left);
    c_var_base& log_or(c_var_string* left);
    c_var_base& log_or(c_var_array* left);
    // END

    bool operator==(const string& right) const;
    bool operator==(const MY_LONG_LONG& right) const;
};

#endif

