/***************************************************************************
                          c_var_string.cpp  -  description
                             -------------------
    begin                : Thu May 10 2007
    copyright            : (C) 2004-2007 by Marian VooDooMan Meravy
    email                : ghostvoodooman -a.T= users (dot) sourceforge.net
 ***************************************************************************/

/* $Id: c_var_string.cpp,v 1.16 2008/11/24 12:09:46 ghostvoodooman Exp $ */

/***************************************************************************

VooDoo cIRCle - an IRC (ro)bot
Copyright (C) 2004-2008 by Marian VooDooMan Meravy
  (ghostvoodooman -a.T= users (dot) sourceforge.net)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************/

/*!
    \file
    \brief Class 'c_var_string'
    \version $Id: c_var_string.cpp,v 1.16 2008/11/24 12:09:46 ghostvoodooman Exp $
*/

#if defined(HAVE_CONFIG_H)
#   include <config.h>
#endif

#include "../params.h"
#include "../utils/memleak.h"
#include "../utils/assert.h"
#include "../compat.h"

#include "c_var_bool.h"
#include "c_var_string.h"
#include "../var_keeper.h"
#include "c_var_int.h"
#include "../md5.h"

#include "../utils/precomp.h"

#if defined(_MSC_VER) || defined(__BORLANDC__)
#   pragma hdrstop
#endif

c_var_bool* c_var_string::to_bool()
{
    return MY_NEW c_var_bool(is_true());
}

c_var_string::c_var_string()
    : c_var_base()
{
    value_="";
    var_keeper->include_into_locals(this);
}

c_var_string::c_var_string(const c_var_string& orig)
    : c_var_base(orig)
{
    value_=orig.value_;
    //var_keeper->include_into_locals(this);
}

c_var_string::c_var_string(const c_var_string& orig, int)
    : c_var_base(orig)
{
    value_=orig.value_;
}

c_var_string::c_var_string(const string& value)
    : c_var_base()
{
    value_=value;
    var_keeper->include_into_locals(this);
}

c_var_string::c_var_string(const string& value, bool persistent)
    : c_var_base()
{
    value_=value;
    set_persistent(persistent);
    var_keeper->include_into_locals(this);
}

c_var_string* c_var_string::copy()
{
    c_var_string* res=MY_NEW c_var_string(*this,0);
    res->set_persistent(false);
    res->set_read_only(false);
    res->reset_id();
    res->id=var_keeper->alloc(res);
    res->set_session(var_keeper->get_current_session());
    var_keeper->include_into_locals(res);
    return res;
}

c_var_base* c_var_string::copy(int)
{
    c_var_string* res=MY_NEW c_var_string(*this,0);
    res->set_persistent(false);
    res->set_read_only(false);
    res->reset_id();
    res->id=var_keeper->alloc(res);
    res->set_session(var_keeper->get_current_session());
    //var_keeper->include_into_locals(res);
    return res;
}

string c_var_string::get()
{
    return value_;
}

string c_var_string::get() const
{
    return value_;
}

void c_var_string::set(c_module* this_module, string value, const string& debug)
{
    if(!read_only)
        value_=value;
    else {
        var_oops(this_module,__FILE__,__my_func__,"ignoring write request to read-only variable",debug);
    }
}

string c_var_string::get(c_module* this_module, c_index_int*& index, const string& debug)
{
    MY_LONG_LONG idx=index->get();
    MY_DELETE(index);
    index=NULL;
    if(idx<0 || value_.length()<=static_cast<MY_U_LONG_LONG>(idx)) {
        string d="Index out of range, base variable (STRING) length is ";
        d+=tos((unsigned long)value_.length());
        d+=", while requested index (INT) was ";
        d+=tos(idx);
        d+=" => returning an empty string";
        var_oops(this_module,__FILE__,__my_func__,GET_STR_DEBUG_INFO_EX(d),debug);
        return (string)"";
    }
    return (string)""+value_[(long)idx];
}

void c_var_string::set(c_module* this_module, c_var_base* index, c_var_base* value, const string& debug)
{
    if(!index) {
        string d="argument \"index\" is NULL";
        throw c_var_exception(this_module,__FILE__,__my_func__,c_var_exception::E_INDEX_NULL,d);
    }

    if(get_read_only()) {
        var_oops(this_module,__FILE__,__my_func__,"ignoring write request to read-only variable",debug);
        return;
    }

    char add;
    {
        e_vartype value_vt=value->get_var_type();
        switch(value_vt) {
            case VARTYPE_INT:
                {
                    c_var_int* i=dynamic_cast<c_var_int*>(value);
                    if(!i)
                        throw c_var_exception(this_module,__FILE__,__my_func__,c_var_exception::E_DYNCAST,"dynamic_cast<c_var_int*>(value) returned NULL",value);
                    string tmp=tos(i->get());
                    if(tmp.length()!=1) {
                        string d="Value (INT) to set at specified index to variable (STRING) is not one character long => ignoring the rest, setting first character only, value was: ";
                        d+=tmp;
                        var_oops(this_module,__FILE__,__my_func__,GET_STR_DEBUG_INFO_EX(d),debug);
                    }
                    tmp+="0";
                    add=tmp[0];
                }
                break;
            case VARTYPE_STRING:
                {
                    c_var_string* s=dynamic_cast<c_var_string*>(value);
                    if(!s)
                        throw c_var_exception(this_module,__FILE__,__my_func__,c_var_exception::E_DYNCAST,"dynamic_cast<c_var_string*>(value) returned NULL",value);
                    string v=s->get();
                    if(v.empty()) {
                        string d="Value (STRING) to set at specified index to variable (STRING) is an empty string, ignoring";
                        var_oops(this_module,__FILE__,__my_func__,GET_STR_DEBUG_INFO_EX(d),debug);
                        return;
                    }
                    if(v.length()!=1) {
                        string d="Value (STRING) to set at specified index to variable (STRING) is not one character long => ignoring the rest, setting first character only, value was: \"";
                        d+=v;
                        d+="\"";
                        var_oops(this_module,__FILE__,__my_func__,GET_STR_DEBUG_INFO_EX(d),debug);
                    }
                    add=v[0];
                }
                break;
            default:
                {
                    throw c_var_exception(this_module,__FILE__,__my_func__,c_var_exception::E_INDEX_INV,"index is not INT, nor STRING",index);
                }
                break;
        }
    }

    e_vartype index_vt=index->get_var_type();
    if(index_vt!=VARTYPE_INT && index_vt!=VARTYPE_STRING) {
        string d="Argument \"index\" is NOT of type STRING, nor INT";
        throw c_var_exception(this_module,__FILE__,__my_func__,c_var_exception::E_INDEX_INV,GET_STR_DEBUG_INFO_EX(d),index);
    }

    c_var_int* index_i=NULL;
    c_var_string* index_s=NULL;
    switch(index_vt) {
        case VARTYPE_INT:
            {
                index_i=dynamic_cast<c_var_int*>(index);
                if(!index_i)
                    throw c_var_exception(this_module,__FILE__,__my_func__,c_var_exception::E_DYNCAST,"dynamic_cast<c_var_int*>(index) returned NULL",index);
            }
            break;
        case VARTYPE_STRING:
            {
                index_s=dynamic_cast<c_var_string*>(index);
                if(!index_s)
                    throw c_var_exception(this_module,__FILE__,__my_func__,c_var_exception::E_DYNCAST,"dynamic_cast<c_var_string*>(index) returned NULL",index);
            }
            break;
        default:
            {
                throw c_var_exception(this_module,__FILE__,__my_func__,c_var_exception::E_INDEX_INV,"index is not INT, nor STRING",index);
            }
            break;
    }

    // sanity check
    if(!index || !(index_i || index_s) || (!index_i && index_vt==VARTYPE_INT) || (!index_s && index_vt==VARTYPE_STRING))
        throw c_var_exception(this_module,__FILE__,__my_func__,c_var_exception::E_INTERNAL,"internal error - sanity check failed",index);

    if(index_vt==VARTYPE_INT) {
        MY_LONG_LONG idx=index_i->get();
        if(idx<0 || value_.length()<=static_cast<MY_U_LONG_LONG>(idx)) {
            string d="Index out of range, base variable (STRING) length is ";
            d+=tos(static_cast<unsigned long>(value_.length()));
            d+=", while requested index (INT) was ";
            d+=tos(idx);
            throw c_var_exception(this_module,__FILE__,__my_func__,c_var_exception::E_INDEX_RANGE,GET_STR_DEBUG_INFO_EX(d),index);
        }
        value_[(long)idx]=add;
        return;
    }

    if(index_vt==VARTYPE_STRING) {
        string idx_s=index_s->get();
        bool bad=idx_s.length()==0;
        for(unsigned i1=0; !bad && i1<idx_s.length(); ++i1)
            if(idx_s[i1]<'0' || idx_s[i1]>'9')
                bad=true;
        if(!bad) {
            MY_LONG_LONG idx=stoll(idx_s);
            if(idx<0 || value_.length()<=static_cast<MY_U_LONG_LONG>(idx))
                bad=true;
            else {
                value_[(long)idx]=add;
                return;
            }
        }
        string d="Index out of range, base variable (STRING) length is ";
        d+=tos(static_cast<unsigned long>(value_.length()));
        d+=", while requested index (STRING) was \"";
        d+=idx_s;
        d+="\"";
        throw c_var_exception(this_module,__FILE__,__my_func__,c_var_exception::E_INDEX_RANGE,GET_STR_DEBUG_INFO_EX(d),index);
    }
    throw c_var_exception(this_module,__FILE__,__my_func__,c_var_exception::E_INTERNAL,"internal error: the code should not go there");
}

c_var_string& c_var_string::operator+=(const c_var_string& to_add)
{
    value_+=to_add.value_;
    return *this;
}

c_var_string& c_var_string::operator+=(const string& to_add)
{
    value_+=to_add;
    return *this;
}

c_var_string& c_var_string::operator+=(const char& to_add)
{
    value_+=(string)""+to_add;
    return *this;
}

e_vartype c_var_string::get_var_type()
{
    return VARTYPE_STRING;
}

e_vartype c_var_string::get_var_type() const
{
    return VARTYPE_STRING;
}

string c_var_string::get_var_type_as_string()
{
    return (string)"STRING";
}

c_var_int* c_var_string::to_int_clever(c_module* this_module, c_var_string::e_return_on_error return_on_error, const string& debug)
{
    c_var_string* s=this;
    if(s->get().length()>128 || s->get().length()==0) {
        var_oops(this_module,__FILE__,__my_func__,"Invalid number, string was: ",get(),debug);
        return (return_on_error==RET_NULL)? NULL : MY_NEW c_var_int(0);
    } else {
        char tmp[512];
        strcpy(tmp,s->get().c_str());
        unsigned start=0;
        if('-'==tmp[start] || '+'==tmp[start])
            start++;
        if(tmp[start+0]=='0' && tmp[start+1]=='x') {
            // hex number
            unsigned i1=0;
            if('-'==tmp[i1] || '+'==tmp[i1])
                i1++;
            i1+=2;
            for( ; tmp[i1]!=0; ++i1) {
                if(!((tmp[i1]>='0' && tmp[i1]<='9') || (tmp[i1]>='a' && tmp[i1]<='f') || (tmp[i1]>='A' && tmp[i1]<='F'))) {
                    var_oops(this_module,__FILE__,__my_func__,"Invalid hexadecimal number, string was: ",get(),debug);
                    return (return_on_error==RET_NULL)? NULL : MY_NEW c_var_int(0);
                }
            }
            MY_LONG_LONG n;
            int num=sscanf(tmp,"%" MY_LONG_LONG_STR "x",&n);
            if(1!=num) {
                var_oops(this_module,__FILE__,__my_func__,"Invalid hexadecimal number, string was: ",get(),debug);
                return (return_on_error==RET_NULL)? NULL : MY_NEW c_var_int(0);
            }
            return MY_NEW c_var_int(n);
        } else if(tmp[start+0]=='0') {
            // octal number
            unsigned i1=0;
            if('-'==tmp[i1] || '+'==tmp[i1])
                i1++;
            for( ; tmp[i1]!=0; ++i1) {
                if(!(tmp[i1]>='0' && tmp[i1]<='7')) {
                    var_oops(this_module,__FILE__,__my_func__,"Invalid octal number, string was: ",get(),debug);
                    return (return_on_error==RET_NULL)? NULL : MY_NEW c_var_int(0);
                }
            }
            MY_LONG_LONG n;
            int num=sscanf(tmp,"%" MY_LONG_LONG_STR "o",&n);
            if(1!=num) {
                var_oops(this_module,__FILE__,__my_func__,"Invalid octal number, string was: ",get(),debug);
                return (return_on_error==RET_NULL)? NULL : MY_NEW c_var_int(0);
            }
            return MY_NEW c_var_int(n);
        } else {
            // decimal number
            unsigned i1=0;
            if('-'==tmp[i1] || '+'==tmp[i1])
                i1++;
            for( ; tmp[i1]!=0; ++i1) {
                if(!(tmp[i1]>='0' && tmp[i1]<='9')) {
                    var_oops(this_module,__FILE__,__my_func__,"Invalid decimal number, string was: ",get(),debug);
                    return (return_on_error==RET_NULL)? NULL : MY_NEW c_var_int(0);
                }
            }
            MY_LONG_LONG n;
            int num=sscanf(tmp,"%" MY_LONG_LONG_STR "d",&n);
            if(1!=num) {
                var_oops(this_module,__FILE__,__my_func__,"Invalid decimal number, string was: ",get(),debug);
                return (return_on_error==RET_NULL)? NULL : MY_NEW c_var_int(0);
            }
            return MY_NEW c_var_int(n);
        }
    }
}

c_var_int* c_var_string::to_int(c_module* this_module, c_var_string::e_return_on_error return_on_error, const string& debug)
{
    c_var_string* s=this;
    if(s->get().length()>128 || s->get().length()==0) {
        var_oops(this_module,__FILE__,__my_func__,"Invalid number, string was: ",get(),debug);
        return (return_on_error==RET_NULL)? NULL : MY_NEW c_var_int(0);
    } else {
        char tmp[512];
        strcpy(tmp,s->get().c_str());
        // decimal number
        unsigned i1=0;
        if('-'==tmp[i1] || '+'==tmp[i1])
            i1++;
        for( ; tmp[i1]!=0; ++i1) {
            if(!(tmp[i1]>='0' && tmp[i1]<='9')) {
                var_oops(this_module,__FILE__,__my_func__,"Invalid decimal number, string was: ",get(),debug);
                return (return_on_error==RET_NULL)? NULL : MY_NEW c_var_int(0);
            }
        }
        return MY_NEW c_var_int(stoll(tmp));
    }
}

c_var_int* c_var_string::to_int(c_var_string::e_return_on_error return_on_error)
{
    c_var_string* s=const_cast<c_var_string*>(this);
    if(s->get().length()>128 || s->get().length()==0) {
        return (return_on_error==RET_NULL)? NULL : MY_NEW c_var_int(0);
    } else {
        char tmp[512];
        strcpy(tmp,s->get().c_str());
        // decimal number
        unsigned i1=0;
        if('-'==tmp[i1] || '+'==tmp[i1])
            i1++;
        for( ; tmp[i1]!=0; ++i1) {
            if(!(tmp[i1]>='0' && tmp[i1]<='9')) {
                return (return_on_error==RET_NULL)? NULL : MY_NEW c_var_int(0);
            }
        }
        return MY_NEW c_var_int(stoll(tmp));
    }
}

void c_var_string::set_ro(const string& val)
{
    value_=val;
    set_read_only(true);
}

string c_var_string::dump(c_module* UNUSEDARG(this_module))
{
    string res="(STRING) ";
    res+=script_esc("php_2",value_);
    return res;
}

void c_var_string::dump_for_botnet(c_module* UNUSEDARG(this_module), char* &data, size_t& size)
{
    uint32 len=(uint32)value_.length();
    size=1+sizeof(len)+value_.length()+1;
    char* d=MY_NEW char[size];
    d[0]='S'; // 'S' stands for string
    byteReverse((unsigned char*)&len,1);
    memcpy(&(d[1]),&len,sizeof(len));
    memcpy(&(d[1+sizeof(len)]),value_.c_str(),value_.length());
    d[1+sizeof(len)+value_.length()]=0;
    data=d;
}

string c_var_string::get_script_string(c_module* UNUSEDARG(this_module), const string& script_type)
{
    return script_esc(script_type,value_);
}

/*!
    \brief Escapes a string
    \author VooDooMan
    \version 2
    \date 2004, 2006-01-27
    \param script_type Type of script ("php", "php_2", "logic.txt" supported yet)
    \param str String to escape
    \return Returns escaped string
*/
string c_var_string::script_esc(const string& script_type, const string& str)
{
    if(!script_type.compare("logic.txt")) {
        string res;
        for(unsigned int i1=0; i1<str.length(); ++i1) {
            if(str[i1]<32 || str[i1]=='$' || str[i1]=='\\' || str[i1]=='\'' || str[i1]=='"') {
                res+="\\x";
                char tmp[64];
                sprintf(tmp,"%02x",(unsigned int)((unsigned char)(str[i1])));
                res+=tmp;
                continue;
            }
            res+=str[i1];
        }
        return res;
    }
    if(!script_type.compare("php") || !script_type.compare("php_2")) {
        string res="\"";
        for(unsigned int i1=0; i1<str.length(); ++i1) {
            if(str[i1]<32 || str[i1]=='$' || str[i1]=='\\' || str[i1]=='\'' || str[i1]=='"' || str[i1]=='@' || str[i1]=='%') {
                res+="\\x";
                char tmp[64];
                sprintf(tmp,"%02x",(unsigned int)((unsigned char)(str[i1])));
                res+=tmp;
                continue;
            }
            res+=str[i1];
        }
        res+="\"";
        return res;
    }
    if(!script_type.compare("perl")) {
        string res="\"";
        for(unsigned int i1=0; i1<str.length(); ++i1) {
            if(str[i1]<32 || str[i1]=='$' || str[i1]=='\\' || str[i1]=='\'' || str[i1]=='"' || str[i1]=='@' || str[i1]=='%') {
                res+="\\x";
                char tmp[64];
                sprintf(tmp,"%02x",(unsigned int)((unsigned char)(str[i1])));
                res+=tmp;
                continue;
            }
            res+=str[i1];
        }
        res+="\"";
        return res;
    }
    return ((string)"\"")+str+"\"";
}

c_index_base* c_var_string::make_index()
{
    return MY_NEW c_index_str(value_);
}

bool c_index_str::operator==(const c_index_str& value)
{
    return !value_.compare(value.get());
}

bool c_var_string::is_true()
{
    if(!value_.compare("0") || value_.empty())
        return false;
    else
        return true;
}

bool c_var_string::is_true() const
{
    if(!value_.compare("0") || value_.empty())
        return false;
    else
        return true;
}

string c_var_string::to_string(c_module* UNUSEDARG(this_module))
{
    return value_;
}

string c_var_string::to_string(c_module* UNUSEDARG(this_module)) const
{
    return value_;
}

long c_var_string::to_long(bool& error)
{
    error=true;
    c_var_string* s=this;
    if(s->get().length()>128 || s->get().length()==0) {
        return 0;
    } else {
        char tmp[512];
        strcpy(tmp,s->get().c_str());
        // decimal number
        unsigned i1=0;
        if('-'==tmp[i1] || '+'==tmp[i1])
            i1++;
        for( ; tmp[i1]!=0; ++i1) {
            if(!(tmp[i1]>='0' && tmp[i1]<='9')) {
                return 0;
            }
        }
        error=false;
        return stol(tmp);
    }
}

long c_var_string::to_long(bool& error) const
{
    c_var_string* s=const_cast<c_var_string*>(this);
    return s->to_long(error);
}

c_var_string::~c_var_string()
{
}

bool c_var_string::operator==(const string& right) const
{
    return !value_.compare(right);
}

bool c_var_string::operator==(const MY_LONG_LONG& right) const
{
    MY_LONG_LONG num;
    char tmp[32];
    if(sscanf(value_.c_str(),"%" MY_LONG_LONG_STR "d%16s",&num,tmp)!=1) {
        // it's not number, or trailing string is followed by number
        return false;
    }
    return num==right;
}

