/***************************************************************************
                          var_except.cpp  -  description
                             -------------------
    begin                : Sun Jan 14 2007
    copyright            : (C) 2007 by Marian VooDooMan Meravy
    email                : ghostvoodooman -a.T= users (dot) sourceforge.net
 ***************************************************************************/

/* $Id: var_except.cpp,v 1.24 2008/11/24 12:09:23 ghostvoodooman Exp $ */

/***************************************************************************

VooDoo cIRCle - an IRC (ro)bot
Copyright (C) 2007-2008 by Marian VooDooMan Meravy
  (ghostvoodooman -a.T= users (dot) sourceforge.net)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************/

/*!
    \file
    \brief Provides classes used as exceptions which will be thrown on illegal operation with variable classes
    \version $Id: var_except.cpp,v 1.24 2008/11/24 12:09:23 ghostvoodooman Exp $
*/

#if defined(HAVE_CONFIG_H)
#   include <config.h>
#endif

#include "utils/memleak.h"
#include "utils/assert.h"
#include "utils.h"
#include "params.h"

#include "var_except.h"

#include "utils/precomp.h"

using namespace std;

/*!
    \brief Constructor
    \version 1
    \date 2007-01-14
    \author VooDooMan
    \param which Which operand caused the problem?
*/
c_var_except::c_var_except(e_which which)
{
    which_operand=which;
}

c_var_except::~c_var_except()
{
}

/*!
    \brief Returns which operand caused the problem
    \version 1
    \date 2007-01-14
    \author VooDooMan
*/
c_var_except::e_which c_var_except::get_which_operand()
{
    return which_operand;
}

/*!
    \brief Constructor
    \version 1
    \date 2007-01-14
    \author VooDooMan
    \param which Which operand caused the problem?
*/
c_var_except_unset::c_var_except_unset(c_var_except::e_which which)
    : c_var_except(which)
{
}

/*!
    \version 1
    \date 2007-01-15
*/
string c_var_except_unset::get_msg()
{
    return ((string)"Variable is UNSET")+get_bad_arg_name();
}

/*!
    \brief Constructor
    \version 1
    \date 2007-01-14
    \author VooDooMan
    \param which Which operand caused the problem?
*/
c_var_except_inv_num::c_var_except_inv_num(c_var_except::e_which which)
    : c_var_except(which)
{
}

/*!
    \version 1
    \date 2007-01-15
*/
string c_var_except_inv_num::get_msg()
{
    return ((string)"Variable of type STRING was unable to be cast to INT")+get_bad_arg_name();
}

/*!
    \brief Constructor
    \version 1
    \date 2007-01-14
    \author VooDooMan
    \param which Which operand caused the problem?
*/
c_var_except_not_num::c_var_except_not_num(c_var_except::e_which which)
    : c_var_except(which)
{
}

/*!
    \version 1
    \date 2007-01-15
*/
string c_var_except_not_num::get_msg()
{
    return ((string)"Variable is required to be INT or BOOL type for this operator")+get_bad_arg_name();
}

/*!
    \brief Constructor
    \version 1
    \date 2007-01-14
    \author VooDooMan
*/
c_var_except_div_zero::c_var_except_div_zero()
    : c_var_except(c_var_except::RIGHT)
{
}

/*!
    \version 1
    \date 2007-01-15
*/
string c_var_except_div_zero::get_msg()
{
    return ((string)"Division by zero")+get_bad_arg_name();
}

/*!
    \brief Constructor
    \version 1
    \date 2007-01-14
    \author VooDooMan
*/
c_var_except_n_impl_1::c_var_except_n_impl_1()
    : c_var_except(c_var_except::INCDEC)
{
}

/*!
    \version 1
    \date 2007-01-15
*/
string c_var_except_n_impl_1::get_msg()
{
    MY_ASSERT_NOTREACH();
    string msg("Internal error - not implemented, PLEASE REPORT this problem, it is BUG! (class '");
    msg+=typeid(*this).name();
    msg+="')"+get_bad_arg_name();
    return msg;
}

/*!
    \brief Constructor
    \version 1
    \date 2007-01-14
    \author VooDooMan
*/
c_var_except_n_impl_2::c_var_except_n_impl_2()
    : c_var_except(c_var_except::NA)
{
}

/*!
    \version 1
    \date 2007-01-15
*/
string c_var_except_n_impl_2::get_msg()
{
    MY_ASSERT_NOTREACH();
    string msg("Internal error - not implemented, PLEASE REPORT this problem, it is BUG! (class '");
    msg+=typeid(*this).name();
    msg+="')"+get_bad_arg_name();
    return msg;
}

/*!
    \brief Constructor
    \version 1
    \date 2007-01-15
    \author VooDooMan
*/
c_var_except_n_impl_3::c_var_except_n_impl_3()
    : c_var_except(c_var_except::NA)
{
}

/*!
    \version 1
    \date 2007-01-15
*/
string c_var_except_n_impl_3::get_msg()
{
    MY_ASSERT_NOTREACH();
    string msg("Internal error - not implemented, PLEASE REPORT this problem, it is BUG! (class '");
    msg+=typeid(*this).name();
    msg+="')"+get_bad_arg_name();
    return msg;
}

/*!
    \brief Returns number of erroneous operand
    \author VooDooMan
    \version 1
    \date 2007-01-15
    \retval -1 Returns zero for N/A
    \return Returns number of erroneous operand, counting from one
*/
int c_var_except::get_bad_arg_num()
{
    switch(which_operand) {
        case UNARY:
        case INCDEC:
        case LEFT:
            return 1;
        case RIGHT:
            return 2;
        case NA:
        default:
            return -1;
    }
    /* NOT REACH */
}

/*!
    \brief Returns human-readable name of erroneous operand
    \author VooDooMan
    \version 1
    \date 2007-01-15
    \return Returns an empty string if not applicable
*/
string c_var_except::get_bad_arg_name()
{
    switch(which_operand) {
        case LEFT:
            return (string)" (left-hand side operand)";
        case RIGHT:
            return (string)" (right-hand side operand)";
        case UNARY:
            return (string)" (unary operator operand)";
        case INCDEC:
            return (string)" (pre- or post-, inc- or decrement operator operand)";
        case NA:
        default:
            return (string)"";
    }
}

