/***************************************************************************
                          var_except.h  -  description
                             -------------------
    begin                : Sun Jan 14 2007
    copyright            : (C) 2007 by Marian VooDooMan Meravy
    email                : ghostvoodooman -a.T= users (dot) sourceforge.net
 ***************************************************************************/

/* $Id: var_except.h,v 1.22 2008/03/11 21:11:32 ghostvoodooman Exp $ */

/***************************************************************************

VooDoo cIRCle - an IRC (ro)bot
Copyright (C) 2007-2008 by Marian VooDooMan Meravy
  (ghostvoodooman -a.T= users (dot) sourceforge.net)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************/

/*!
    \file
    \brief Provides classes used as exceptions which will be thrown on illegal operation with variable classes
    \version $Id: var_except.h,v 1.22 2008/03/11 21:11:32 ghostvoodooman Exp $
*/

#ifndef VAR_EXCEPT_H__
#define VAR_EXCEPT_H__

#if defined(HAVE_CONFIG_H)
#   include <config.h>
#endif

#include <string>
#include <vector>

#include "utils/assert.h"

using namespace std;

#include "utils.h"      // because we want to enable mem leak detection

#include "params.h"
#include "compat.h"
#include "utils.h"

/*!
    \brief Base exception class
    \author VooDooMan
    \date 2007-01-14
    \version 1
*/
class c_var_except
{
public:
    enum e_which
    {
        NA = 0,         //!< N/A
        UNARY,          //!< Unary operator operand
        INCDEC,         //!< Post-/pre-/inc-/dec-rement operator operand
        LEFT,           //!< Left operand
        RIGHT,          //!< Right operand
    };                  //!< Which operand caused the problem?

    c_var_except(e_which which);
    virtual ~c_var_except() = 0;

    c_var_except::e_which get_which_operand();

    /*!
        \brief Returns human-readable error description
        \author VooDooMan
    */
    virtual string get_msg() = 0;

    int get_bad_arg_num();

    string get_bad_arg_name();

private:
    c_var_except::e_which which_operand;                //!< Which operand caused the problem?
};

/*!
    \brief Exception: UNSET variable
    \author VooDooMan
    \date 2007-01-14
    \version 1
*/
class c_var_except_unset : public c_var_except
{
public:
    c_var_except_unset(c_var_except::e_which which);

    string get_msg();
};

/*!
    \brief Exception: invalid number (converted from STRING to INT)
    \author VooDooMan
    \date 2007-01-14
    \version 1
*/
class c_var_except_inv_num : public c_var_except
{
public:
    c_var_except_inv_num(c_var_except::e_which which);

    string get_msg();
};

/*!
    \brief Exception: it is not a number (e.g. tried to convert from ARRAY to INT)
    \author VooDooMan
    \date 2007-01-14
    \version 1
*/
class c_var_except_not_num : public c_var_except
{
public:
    c_var_except_not_num(c_var_except::e_which which);

    string get_msg();
};

/*!
    \brief Exception: division by zero
    \author VooDooMan
    \date 2007-01-14
    \version 1
*/
class c_var_except_div_zero : public c_var_except
{
public:
    c_var_except_div_zero();

    string get_msg();
};

/*!
    \brief Exception: internal error, post-/pre-inc-/de-crement on STRING is handled separately
    \author VooDooMan
    \date 2007-01-14
    \version 1
*/
class c_var_except_n_impl_1 : public c_var_except
{
public:
    c_var_except_n_impl_1();

    string get_msg();
};

/*!
    \brief Exception: internal error, '==' or '!=' operator on two ARRAY's, implemented separately
    \author VooDooMan
    \date 2007-01-15
    \version 1
*/
class c_var_except_n_impl_2 : public c_var_except
{
public:
    c_var_except_n_impl_2();

    string get_msg();
};

/*!
    \brief Exception: internal error, variable should never be a reference
    \author VooDooMan
    \date 2007-01-15
    \version 1
*/
class c_var_except_n_impl_3 : public c_var_except
{
public:
    c_var_except_n_impl_3();

    string get_msg();
};

#endif

