/***************************************************************************
                          var_index.h  -  description
                             -------------------
    begin                : Sun Nov 6 2006
    copyright            : (C) 2006-2007 by Marian VooDooMan Meravy
    email                : ghostvoodooman -a.T= users (dot) sourceforge.net
 ***************************************************************************/

/* $Id: var_index.cpp,v 1.31 2008/03/25 12:04:23 ghostvoodooman Exp $ */

/***************************************************************************

VooDoo cIRCle - an IRC (ro)bot
Copyright (C) 2006-2008 by Marian VooDooMan Meravy
  (ghostvoodooman -a.T= users (dot) sourceforge.net)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************/

/*!
    \file
    \brief Implements classes for variables' indexes
    \version $Id: var_index.cpp,v 1.31 2008/03/25 12:04:23 ghostvoodooman Exp $
*/

#if defined(HAVE_CONFIG_H)
#   include <config.h>
#endif

#include <string>
#include <vector>
#include <stack>
#include <list>
#include <map>
#include <algorithm>

#include <stdlib.h>

using namespace std;

#include "utils/memleak.h"
#include "utils/assert.h"
#include "utils.h"

#include "params.h"

#include "var_index.h"
#include "md5.h"
#include "utils.h"

#include "utils/precomp.h"

c_index_base::e_type c_index_base::type()
{
    return IDX_NULL;
}

c_index_base::e_type c_index_base::type() const
{
    return IDX_NULL;
}

bool c_index_base::equal(const string& value, bool case_insensitive)
{
    if(type()!=IDX_STR)
        return false;
    if(!case_insensitive) {
        return !dynamic_cast<c_index_str*>(this)->get().compare(value);
    } else {
        return !cmp_strings_case_insensitive(dynamic_cast<c_index_str*>(this)->get(),value);
    }
}

string c_index_base::to_string()
{
    return (string)"NULL";
}

c_index_base::e_type c_index_int::type()
{
    return IDX_INT;
}

c_index_base::e_type c_index_int::type() const
{
    return IDX_INT;
}

MY_LONG_LONG c_index_int::get()
{
    return value_;
}

MY_LONG_LONG c_index_int::get() const
{
    return value_;
}

bool c_index_int::operator==(const MY_LONG_LONG& value)
{
    return value_==value;
}

bool c_index_int::operator==(const string& UNUSEDARG(value))
{
    return false;
}

bool c_index_int::operator!=(const MY_LONG_LONG& value)
{
    return value_!=value;
}

bool c_index_int::operator!=(const string& UNUSEDARG(value))
{
    return true;
}

bool c_index_int::equal(const string& UNUSEDARG(value), bool UNUSEDARG(case_insensitive))
{
    return false;
}

bool c_index_int::is_int()
{
    return true;
}

bool c_index_int::is_int() const
{
    return true;
}

bool c_index_int::is_str()
{
    return false;
}

bool c_index_int::is_str() const
{
    return false;
}

string c_index_int::to_string()
{
    return tos(value_);
}

c_index_base::e_type c_index_str::type()
{
    return IDX_STR;
}

c_index_base::e_type c_index_str::type() const
{
    return IDX_STR;
}

const string& c_index_str::get()
{
    return value_;
}

const string& c_index_str::get() const
{
    return value_;
}

bool c_index_str::operator==(const MY_LONG_LONG& UNUSEDARG(value))
{
    return false;
}

bool c_index_str::operator==(const string& value)
{
    return !value_.compare(value);
}

bool c_index_str::operator!=(const MY_LONG_LONG& UNUSEDARG(value))
{
    return true;
}

bool c_index_str::operator!=(const string& value)
{
    return value_.compare(value) != 0;
}

bool c_index_str::is_int()
{
    return false;
}

bool c_index_str::is_int() const
{
    return false;
}

bool c_index_str::is_str()
{
    return true;
}

bool c_index_str::is_str() const
{
    return true;
}

bool c_index_str::equal(const string& value, bool case_insensitive)
{
    if(case_insensitive)
        return !cmp_strings_case_insensitive(value_,value);
    else
        return !value_.compare(value);
    // NOTREACH
}

bool c_index_str::equal(c_index_base* value, bool case_insensitive)
{
    if(value->type()!=IDX_STR)
        return false;
    if(!case_insensitive)
        return !value_.compare(dynamic_cast<c_index_str*>(value)->get());
    else
        return !cmp_strings_case_insensitive(dynamic_cast<c_index_str*>(value)->get(),value_);
    // NOTREACH
}

bool c_index_str::equal(c_index_base* value, bool case_insensitive) const
{
    if(value->type()!=IDX_STR)
        return false;
    if(!case_insensitive)
        return !value_.compare(dynamic_cast<c_index_str*>(value)->get());
    else
        return !cmp_strings_case_insensitive(dynamic_cast<c_index_str*>(value)->get(),value_);
    // NOTREACH
}

bool c_index_str::equal(c_index_base& value, bool case_insensitive)
{
    if(value.type()!=IDX_STR)
        return false;
    if(!case_insensitive)
        return !value_.compare(dynamic_cast<c_index_str&>(value).get());
    else
        return !cmp_strings_case_insensitive(dynamic_cast<c_index_str&>(value).get(),value_);
    // NOTREACH
}

bool c_index_str::equal(c_index_base& value, bool case_insensitive) const
{
    if(value.type()!=IDX_STR)
        return false;
    if(!case_insensitive)
        return !value_.compare(dynamic_cast<c_index_str&>(value).get());
    else
        return !cmp_strings_case_insensitive(dynamic_cast<c_index_str&>(value).get(),value_);
    // NOTREACH
}

string c_index_str::to_string()
{
    return value_;
}

bool c_index_base::operator==(const c_index_int& UNUSEDARG(value))
{
    return false;
}

bool c_index_base::operator==(const c_index_str& UNUSEDARG(value))
{
    return false;
}

bool c_index_int::operator==(const c_index_str& UNUSEDARG(value))
{
    return false;
}

bool c_index_str::operator==(const c_index_int& UNUSEDARG(value))
{
    return false;
}

void c_index_int::dump_for_botnet(char*& data, size_t& size)
{
    size=1+1+1+sizeof(value_);
    char* d=MY_NEW char[size];
    d[0]='I'; // 'I' stands for int
    d[1]=((char)sizeof(value_));
    memcpy(&(d[2]),&value_,sizeof(value_));
    byteReverse((unsigned char*)(&(d[2])),sizeof(value_)/4);
    d[2+sizeof(value_)]=0;
    data=d;
}

void c_index_base::dump_for_botnet(char*& data, size_t& size)
{
    size=1+1;
    char* d=MY_NEW char[size];
    d[0]='U'; // 'U' stands for unset
    d[1]=0;
    data=d;
}

void c_index_str::dump_for_botnet(char*& data, size_t& size)
{
    uint32 len=(uint32)value_.length();
    size=1+sizeof(len)+value_.length()+1;
    char* d=MY_NEW char[size];
    d[0]='S'; // 'S' stands for string
    byteReverse((unsigned char*)&len,1);
    memcpy(&(d[1]),&len,sizeof(len));
    memcpy(&(d[1+sizeof(len)]),value_.c_str(),value_.length());
    d[1+sizeof(len)+value_.length()]=0;
    data=d;
}

c_index_base* c_index_int::copy() const
{
    return MY_NEW c_index_int(*this);
}

c_index_base* c_index_str::copy() const
{
    return MY_NEW c_index_str(*this);
}

bool c_index_str::operator==(const c_index_base& value)
{
    if(value.type()!=IDX_STR)
        return false;
    c_index_base& v=const_cast<c_index_base&>(value);
    if(!dynamic_cast<c_index_str&>(v).get().compare(value_))
        return true;
    return false;
}

bool c_index_str::operator!=(const c_index_base& value)
{
    if(value.type()!=IDX_STR)
        return true;
    c_index_base& v=const_cast<c_index_base&>(value);
    if(!dynamic_cast<c_index_str&>(v).get().compare(value_))
        return false;
    return true;
}

bool c_index_int::operator==(const c_index_base& value)
{
    if(value.type()!=IDX_INT)
        return false;
    c_index_base& v=const_cast<c_index_base&>(value);
    return dynamic_cast<c_index_int&>(v).get()==value_;
}

bool c_index_int::operator!=(const c_index_base& value)
{
    if(value.type()!=IDX_INT)
        return true;
    c_index_base& v=const_cast<c_index_base&>(value);
    return dynamic_cast<c_index_int&>(v).get()!=value_;
}

bool c_index_int::operator==(const c_index_int& value)
{
    return value_==value.get();
}

bool c_index_int::equal(c_index_base& value, bool UNUSEDARG(case_insensitive))
{
    if(value.type()!=IDX_INT)
        return false;
    return dynamic_cast<c_index_int&>(value).get()==value_;
}

bool c_index_int::equal(c_index_base* value, bool UNUSEDARG(case_insensitive))
{
    if(value->type()!=IDX_INT)
        return false;
    return dynamic_cast<c_index_int*>(value)->get()==value_;
}

bool c_index_int::equal(c_index_base& value, bool UNUSEDARG(case_insensitive)) const
{
    if(value.type()!=IDX_INT)
        return false;
    return dynamic_cast<c_index_int&>(value).get()==value_;
}

