/***************************************************************************
                          var_index.h  -  description
                             -------------------
    begin                : Sun Nov 6 2006
    copyright            : (C) 2006-2007 by Marian VooDooMan Meravy
    email                : ghostvoodooman -a.T= users (dot) sourceforge.net
 ***************************************************************************/

/* $Id: var_index.h,v 1.31 2008/03/11 21:11:32 ghostvoodooman Exp $ */

/***************************************************************************

VooDoo cIRCle - an IRC (ro)bot
Copyright (C) 2006-2008 by Marian VooDooMan Meravy
  (ghostvoodooman -a.T= users (dot) sourceforge.net)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************/

/*!
    \file
    \brief Implements classes for variables' indexes
    \version $Id: var_index.h,v 1.31 2008/03/11 21:11:32 ghostvoodooman Exp $
*/

#ifndef VAR_INDEX_H__
#define VAR_INDEX_H__

#if defined(HAVE_CONFIG_H)
#   include <config.h>
#endif

#include <string>
#include <vector>
#include <stack>
#include <list>
#include <map>
#include <algorithm>

using namespace std;

#include "utils.h"      // because we want to enable mem leak detection

#include "params.h"

#include "utils/assert.h"

class c_index_int;
class c_index_str;

/*!
    \brief Base class for index of an array entry
    \author VooDooMan
    \version 1
    \date 2006-11-18
*/
class c_index_base
{
public:
    enum e_type
    {
        IDX_NULL = 0,       //!< NULL index
        IDX_INT,            //!< Class is an integer index
        IDX_STR,            //!< Class is a string index
    };                      //!< Describes kind of class

    /*!
        \brief Constructor
        \version 1
        \date 2006-11-19
        \author VooDooMan
    */
    c_index_base() { ; }

    /*!
        \brief Destructor
        \version 1
        \date 2006-11-27
        \author VooDooMan
    */
    virtual ~c_index_base() { ; }

    /*!
        \brief Gets type of index
        \author VooDooMan
        \version 1
        \date 2006-11-18
        \return Returns type of index
    */
    virtual e_type type() = 0;

    /*!
        \brief Gets type of index
        \author VooDooMan
        \version 1
        \date 2006-11-18
        \return Returns type of index
    */
    virtual e_type type() const = 0;

    /*!
        \brief Equivalence operator (case (in)-sensitive)
        \param value Value to test
        \param case_insensitive 'true' for case in-sensitivity
        \retval true Equal
        \retval false Not equal
    */
    virtual bool equal(const string& value, bool case_insensitive);

    /*!
        \brief Equivalence operator (case (in)-sensitive)
        \param value Value to test
        \param case_insensitive 'true' for case in-sensitivity
        \retval true Equal
        \retval false Not equal
    */
    virtual bool equal(c_index_base& value, bool case_insensitive) = 0;

    /*!
        \brief Equivalence operator (case (in)-sensitive)
        \param value Value to test
        \param case_insensitive 'true' for case in-sensitivity
        \retval true Equal
        \retval false Not equal
    */
    virtual bool equal(c_index_base* value, bool case_insensitive) = 0;

    /*!
        \brief Equivalence operator (case (in)-sensitive)
        \param value Value to test
        \param case_insensitive 'true' for case in-sensitivity
        \retval true Equal
        \retval false Not equal
    */
    virtual bool equal(c_index_base& value, bool case_insensitive) const = 0;

    virtual string to_string() = 0;

    /*!
        \brief Equivalence operator
        \param value Value to test
        \retval true Equal
        \retval false Not equal
    */
    virtual bool operator==(const c_index_base& value) = 0;

    /*!
        \brief Equivalence operator
        \param value Value to test
        \retval true Equal
        \retval false Not equal
    */
    virtual bool operator==(const c_index_int& value) = 0;

    /*!
        \brief Equivalence operator
        \param value Value to test
        \retval true Equal
        \retval false Not equal
    */
    virtual bool operator==(const c_index_str& value) = 0;

    virtual void dump_for_botnet(char*& data, size_t& size) = 0;

    /*!
        \brief Creates a copy
        \author VooDooMan
        \version 1
        \date 2006-11-19
    */
    virtual c_index_base* copy() const = 0;

    virtual bool operator!=(const c_index_base& value) = 0;
};

/*!
    \brief Implements integer index of an array entry
    \author VooDooMan
    \version 1
    \date 2006-11-18
*/
class c_index_int : public c_index_base
{
protected:
    MY_LONG_LONG value_;                //!< Value of index

public:

    virtual e_type type();

    virtual e_type type() const;

    /*!
        \brief Constructor
        \version 1
        \date 2006-11-18
        \author VooDooMan
        \param value Value to initialize the class with
    */
    c_index_int(MY_LONG_LONG value)
        : c_index_base(), value_(value)
    {
    }

    /*!
        \brief Destructor
        \version 1
        \date 2006-11-27
        \author VooDooMan
    */
    virtual ~c_index_int()
    {
    }

    /*!
        \brief Copy-constructor
        \version 1
        \date 2006-11-18
        \author VooDooMan
        \param orig Value to initialize the class with
    */
    c_index_int(const c_index_int& orig)
        : c_index_base(), value_(orig.value_)
    {
    }

    /*!
        \brief Returns value of index
        \version 1
        \date 2006-11-18
        \author VooDooMan
    */
    MY_LONG_LONG get();

    /*!
        \brief Returns value of index
        \version 1
        \date 2006-11-18
        \author VooDooMan
    */
    MY_LONG_LONG get() const;

    bool operator==(const MY_LONG_LONG& value);

    bool operator==(const string& UNUSEDARG(value));

    bool operator!=(const MY_LONG_LONG& value);

    bool operator!=(const string& UNUSEDARG(value));

    bool equal(const string& UNUSEDARG(value), bool UNUSEDARG(case_insensitive));

    /*!
        \brief Returns 'true'
        \version 1
        \date 2006-11-19
        \author VooDooMan
    */
    bool is_int();

    /*!
        \brief Returns 'true'
        \version 1
        \date 2006-11-19
        \author VooDooMan
    */
    bool is_int() const;

    /*!
        \brief Returns 'false'
        \version 1
        \date 2006-11-19
        \author VooDooMan
    */
    bool is_str();

    /*!
        \brief Returns 'false'
        \version 1
        \date 2006-11-19
        \author VooDooMan
    */
    bool is_str() const;

    /*!
        \brief Equivalence operator (case (in)-sensitive)
        \param value Value to test
        \param case_insensitive (Ignored)
        \retval true Equal
        \retval false Not equal
    */
    bool equal(c_index_base* value, bool UNUSEDARG(case_insensitive));

    bool equal(c_index_base& value, bool UNUSEDARG(case_insensitive));

    string to_string();

    virtual bool operator!=(const c_index_base& value);

    virtual bool operator==(const c_index_base& value);

    /*!
        \brief Equivalence operator
        \param value Value to test
        \retval true Equal
        \retval false Not equal
    */
    bool operator==(const c_index_str& UNUSEDARG(value));

    virtual void dump_for_botnet(char*& data, size_t& size);

    virtual c_index_base* copy() const;

    virtual bool operator==(const c_index_int& value);

    virtual bool equal(c_index_base& value, bool UNUSEDARG(case_insensitive)) const;
};

/*!
    \brief Implements integer index of an array entry
    \author VooDooMan
    \version 1
    \date 2006-11-18
*/
class c_index_str : public c_index_base
{
protected:
    string value_;                      //!< Value of index

public:
    virtual e_type type();

    virtual e_type type() const;

    /*!
        \brief Constructor
        \version 1
        \date 2006-11-18
        \author VooDooMan
        \param value Value to initialize the class with
    */
    c_index_str(const string& value)
        : c_index_base(), value_(value)
    {
    }

    /*!
        \brief Destructor
        \version 1
        \date 2006-11-27
        \author VooDooMan
    */
    virtual ~c_index_str()
    {
    }

    /*!
        \brief Copy-constructor
        \version 1
        \date 2006-11-18
        \author VooDooMan
        \param orig Value to initialize the class with
    */
    c_index_str(const c_index_str& orig)
        : c_index_base(), value_(orig.value_)
    {
    }

    /*!
        \brief Returns value of index
        \version 1
        \date 2006-11-18
        \author VooDooMan
    */
    const string& get();

    /*!
        \brief Returns value of index
        \version 1
        \date 2006-11-18
        \author VooDooMan
    */
    const string& get() const;

    bool operator==(const MY_LONG_LONG& UNUSEDARG(value));

    bool operator==(const string& value);

    bool operator!=(const MY_LONG_LONG& UNUSEDARG(value));

    bool operator!=(const string& value);

    bool equal(const string& value, bool case_insensitive);

    /*!
        \brief Returns 'false'
        \version 1
        \date 2006-11-19
        \author VooDooMan
    */
    bool is_int();

    /*!
        \brief Returns 'false'
        \version 1
        \date 2006-11-19
        \author VooDooMan
    */
    bool is_int() const;

    /*!
        \brief Returns 'true'
        \version 1
        \date 2006-11-19
        \author VooDooMan
    */
    bool is_str();

    /*!
        \brief Returns 'true'
        \version 1
        \date 2006-11-19
        \author VooDooMan
    */
    bool is_str() const;

    virtual bool equal(c_index_base* value, bool case_insensitive);

    virtual bool equal(c_index_base* value, bool case_insensitive) const;

    string to_string();

    virtual bool operator!=(const c_index_base& value);

    /*!
        \brief Equivalence operator
        \param value Value to test
        \retval true Equal
        \retval false Not equal
    */
    bool operator==(const c_index_int& UNUSEDARG(value));

    virtual void dump_for_botnet(char*& data, size_t& size);

    virtual c_index_base* copy() const;

    virtual bool operator==(const c_index_str& value);

    virtual bool operator==(const c_index_base& value);

    bool equal(c_index_base& value, bool case_insensitive);

    bool equal(c_index_base& value, bool case_insensitive) const;
};

#endif

