/***************************************************************************
                          var_keeper.cpp  -  description
                             -------------------
    begin                : Sat May 13 2006
    copyright            : (C) 2006-2007 by Marian VooDooMan Meravy
    email                : ghostvoodooman -aT= users (dot) sourceforge.net
 ***************************************************************************/

/* $Id: var_keeper.cpp,v 1.47 2008/11/24 12:09:23 ghostvoodooman Exp $ */

/***************************************************************************

VooDoo cIRCle - an IRC (ro)bot
Copyright (C) 2006-2008 by Marian VooDooMan Meravy
  (ghostvoodooman -aT= users (dot) sourceforge.net)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************/

/*!
    \file
    \brief Provides variable keeper
    \version $Id: var_keeper.cpp,v 1.47 2008/11/24 12:09:23 ghostvoodooman Exp $
*/

#if defined(HAVE_CONFIG_H)
#   include <config.h>
#endif

#include "utils/assert.h"
#include "tree/tree.hh"

#include <string>
#include <vector>
#include <stack>
#include <list>
#include <map>
#include <deque>
#include <sstream>

#include <stdlib.h>

using namespace std;

#include "utils/memleak.h"

#include "utils/assert.h"
#include "utils.h"

#include "params.h"
#include "compat.h"
#include "variable.h"
#include "var_keeper.h"
#include "modules.h"
#include "var/c_var_bool.h"
#include "var/c_var_int.h"
#include "var/c_var_string.h"
#include "var/c_var_unset.h"
#include "var/c_var_array.h"
#include "var/c_var_ref.h"
#include "mod/c_module.h"
#include "utils/my_getpid.h"

#include "memory/c_single_element_allocator.hh"

#include "utils/precomp.h"

#if defined(_MSC_VER) || defined(__BORLANDC__)
#   pragma hdrstop
#endif

c_var_keeper* var_keeper=NULL;

#define DO_CHECK        0
#define WANT_DUMPS      1

#if (DO_CHECK!=0) && defined(_DEBUG)
#   define CHECK(ln,x)          _check((ln),(x))
#else
#   define CHECK(ln,x)          /* disabled */
#endif

#if (WANT_DUMPS!=0) && defined(_DEBUG)
#   define DUMP(ln,x)           _dump(NULL,(ln),(x))
#else
#   define DUMP(ln,x)           /* disabled */
#endif

#define USE_LOOKUP_FOR_LOCALS   0

static bool reparenting_nodes=false;        //!< To prevent copy construction/destruction

var_id_t c_var_keeper::dereference(var_id_t what)
{
    if(!what)
        return VAR_ID_INVALID;

    map<var_id_t, c_entry*>::iterator i1=storage.find(what);
    if(i1!=storage.end()) {
        if((*i1).second && (*i1).second->val) {
            c_var_ref* ref=NULL;
            if((*i1).second->val && typeid(*(*i1).second->val)==typeid(c_var_ref))
                ref=dynamic_cast<c_var_ref*>((*i1).second->val);
            if(ref)
                return ref->get_referenced();
        }
    }

    vector<var_id_t> chain;

    for(;;) {

        if(!chain.empty()) {
            for(vector<var_id_t>::iterator i1=chain.begin(); i1!=chain.end(); ++i1)
                if(*i1==what)
                    return VAR_ID_INVALID; // infinite reference
        }

        map<var_id_t, var_id_t>::iterator i1=references.find(what);
        if(references.end()==i1)
            return what;

        chain.push_back(what);

        what=(*i1).second;
    }
}

void c_var_keeper::dereference(var_id_t what, vector<var_id_t>& chain)
{
    chain.clear();

    if(!what)
        return;

    for(;;) {
        map<var_id_t, var_id_t>::iterator i1=references.find(what);
        if(references.end()==i1)
            return;
        chain.push_back(what);
        what=(*i1).second;

        for(vector<var_id_t>::iterator i2=chain.begin(); i2!=chain.end(); ++i2)
            if(*i2==what) {
                chain.clear();
                return; // infinite reference
            }
    }
}

c_entry* c_var_keeper::get_e(var_id_t id, log_function_ptr_t log_fn, void* user_data)
{
    if(!id)
        return NULL;
    map<var_id_t, c_entry*>::iterator i=storage.find(id);
    if(i!=storage.end()) {
        return (*i).second;
    } else {
        string d="Invalid ID #";
        d+=tos(id);
        if(log_fn)
            (*log_fn)(LS_FATAL,GET_STR_DEBUG_INFO_EX(d),user_data);
        MY_ASSERT_NOTREACH();
        return NULL;
    }
}

void c_var_keeper::dealloc(var_id_t id, log_function_ptr_t log_fn, void* user_data, bool no_deref)
{
    map<var_id_t, c_entry*>::iterator i=storage.find(id);
    if(i!=storage.end()) {
        var_id_t real=id;
        if(!no_deref)
            real=dereference(id);
        if(real!=id) {
            c_var_base* fake=MY_NEW c_var_unset();
            vector<pair<var_id_t, var_id_t> > dummy;
            set(real,log_fn,user_data,true,fake,true,dummy,dummy,core_mod,"","");
        }
        (*i).second->val->reset_id();
        MY_DELETE((*i).second->val);
        (*i).second->val=NULL;
        MY_DELETE((*i).second);
        storage.erase(i);

        used_var_ids.clear((unsigned long)id);
    } else {
        string d="No such ID #";
        d+=tos(id);
        if(log_fn)
            (*log_fn)(LS_FATAL,GET_STR_DEBUG_INFO_EX(d),user_data);
        MY_ASSERT_NOTREACH();
        return;
    }
}

c_var_base* c_var_keeper::get(var_id_t id, log_function_ptr_t log_fn, void* user_data)
{
    if(!id)
        return NULL;

    id=dereference(id);

    return get_no_deref(id,log_fn,user_data);
}

c_var_base* c_var_keeper::get_no_deref(var_id_t id, log_function_ptr_t log_fn, void* user_data)
{
    if(!id)
        return NULL;

    map<var_id_t, c_entry*>::iterator i=storage.find(id);
    if(i!=storage.end()) {
        return (*i).second->val;
    } else {
        string d="No such ID #";
        d+=tos(id);
        if(log_fn)
            (*log_fn)(LS_FATAL,GET_STR_DEBUG_INFO_EX(d),user_data);

        DUMP(__LINE__,"#1");

        MY_ASSERT_NOTREACH();
        return NULL;
    }
}

var_id_t c_var_keeper::get(c_var_base* value, log_function_ptr_t log_fn, void* user_data)
{
    if(!value)
        return VAR_ID_INVALID;

    c_var_base* res=get(value->get_id(),log_fn,user_data);
    return res?res->get_id():VAR_ID_INVALID;
}

var_id_t c_var_keeper::add_reference(var_id_t from, var_id_t to, log_function_ptr_t UNUSEDARG(log_fn), void* UNUSEDARG(user_data))
{
    if(from==to || !from || !to)
        return VAR_ID_INVALID;
    c_var_ref* ref=MY_NEW c_var_ref(from,to);
    references.insert(make_pair(from,to));
    adjust_ref_cnt(from,+1);
    return ref->get_id();
}

void c_var_keeper::free_all()
{
    BEGIN_DEBUG_TIME_MEASUREMENT(true,1,"Destroying all variables");

    do_garbage(GARBAGE_MAX_TIME_INFINITE);

    references.clear();

    {
        int cnt;
        do {
            cnt=0;
            tree<c_tree_node, c_single_element_allocator<tree_node_<c_tree_node> > >::iterator i2;
            i2=variables.begin();
            while(i2!=variables.end()) {
                if(variables.number_of_children(i2)) {
                    variables.flatten(i2);
                    cnt++;
                }
                ++i2;
            }
        } while(cnt);
    }

    CHECK(__LINE__,"#2");

    {
        tree<c_tree_node, c_single_element_allocator<tree_node_<c_tree_node> > >::iterator i2;
        vector<c_var_base*> all;
        for(i2=variables.begin(); i2!=variables.end(); ++i2) {
            if((*i2).data2.second) {
                all.push_back((*i2).data2.second);
                var_del_lookup((*i2).data2.second);
                var_destroy_lookup((*i2).data2.second);
            }
            (*i2).assign(NULL);
        }

        vector<c_var_base*>::iterator i3;
        sort(all.begin(),all.end());
        i3=unique(all.begin(),all.end());
        all.erase(i3,all.end());

        for(i3=all.begin(); i3!=all.end(); ++i3) {
            (*i3)->set_in_deletion();
            destroy(*i3);
            MY_DELETE(*i3);
        }
        variables.clear();
    }

    END_DEBUG_TIME_MEASUREMENT(1,0);
}

/*!
    \brief Generates unique filename for dumps
    \author VooDooMan
    \version 2
    \date 2008-02-23, 2008-03-08
    \param module This module (can be NULL -
        will return file for the core module instead)
    \param ln Line number
    \param name Name of dump
*/
MY_STATIC string get_dump_fn(const c_module* module, int ln, const string name)
{
    static MY_U_LONG_LONG cnt=0;

    ostringstream fn;
    fn << (module ? module->get_dir_prefix().c_str() : "");
    fn << "logs" << FILE_SLASH;
    fn << "pid." << tos(my_getpid());
    fn << ".#" << tos(cnt++,8,'0') << ".";
    fn << "var-dump." << make_nice_time(my_time(NULL),false,false,true);
    fn << "." << name;
    fn << ".line." << tos(ln) << ".txt";

    return fn.str();
}

void c_var_keeper::_dump(const c_module* module, int ln, const string& name)
{
    static bool in=false;
    if(in)
        return;
    in=true;

    FILE* f=fopen(get_dump_fn(module,ln,name).c_str(),"w");
    if(!f)
        return;

    fprintf(f,"\n\n=== tree dump ===\n\n");

    {
        static const char* nt_normal="nt_normal";
        static const char* nt_root_node_local="nt_root_node_local";
        static const char* nt_root_node_global="nt_root_node_global";
        static const char* nt_root_node_temp="nt_root_node_temp";
        static const char* nt_root_node_mod_global="nt_root_node_mod_global";
        static const char* nt_root_node_core_entry="nt_root_node_core_entry";
        static const char* nt__other="(other)";
        unsigned last_nid=(unsigned)-1;

        const char* last_nt=nt_normal;

        tree<c_tree_node, c_single_element_allocator<tree_node_<c_tree_node> > >::iterator i1;
        for(i1=variables.begin(); i1!=variables.end(); ++i1) {
            string indent;
            for(int i2=0; i2<variables.depth(i1); ++i2)
                indent+="    ";

            string d=indent;
            my_time_t t;
            switch((*i1).node_type) {
                case c_tree_node::nt_normal:
                    d+="nt_normal";
                    break;
                case c_tree_node::nt_root_node_local:
                    d+="nt_root_node_local ID=";
                    d+=tos((long)(*i1).root_node_id);

                    d+="\n";
                    d+=indent;
                    d+="Debug string: ";
                    d+=get_debug_string((*i1).root_node_id,t);
                    d+="\n";
                    d+=indent;
                    d+="Allocation: ";
                    {
                        time_t tt=static_cast<time_t>(t);
                        d+=ctime(&tt);
                    }
                    d.erase(d.length()-1,1);

                    last_nt=nt_root_node_local;
                    last_nid=(*i1).root_node_id;
                    break;
                case c_tree_node::nt_root_node_global:
                    d+="nt_root_node_global";
                    last_nt=nt_root_node_global;
                    last_nid=(unsigned)-1;
                    break;
                case c_tree_node::nt_root_node_temp:
                    d+="nt_root_node_temp";
                    last_nt=nt_root_node_temp;
                    last_nid=(unsigned)-1;
                    break;
                case c_tree_node::nt_root_node_core_entry:
                    d+="nt_root_node_core_entry";
                    last_nt=nt_root_node_core_entry;
                    last_nid=(unsigned)-1;
                    break;
                case c_tree_node::nt_root_node_mod_global:
                    d+="nt_root_node_mod_global ID=";
                    d+=tos((*i1).root_node_id);

                    d+="\n";
                    d+=indent;
                    d+="Debug string: ";
                    d+=get_debug_string((*i1).root_node_id,t);
                    d+="\n";
                    d+=indent;
                    d+="Allocation: ";
                    {
                        time_t tt=static_cast<time_t>(t);
                        d+=ctime(&tt);
                    }
                    d.erase(d.length()-1,1);

                    last_nt=nt_root_node_mod_global;
                    last_nid=(unsigned)-1;
                    break;
                default:
                    d+="(other == ";
                    d+=tos((*i1).node_type);
                    d+=")";
                    last_nt=nt__other;
                    last_nid=(unsigned)-1;
            }
            d+="\n";

            d+=indent;
            d+="root: ";
            d+=last_nt;
            d+="\n";
            if((unsigned)-1!=last_nid) {
                d+=indent;
                d+="root ID: ";
                d+=tos(last_nid);
                d+="\n";
            }

            d+=indent;
            d+="first:  ";
            d+=(*i1).data2.first ? (*i1).data2.first->to_string() : (string)"NULL";
            d+="\n";

            {
                d+=indent;
                d+="first:  ";
                c_index_base* in=(*i1).data2.first;
                try {
                    if(!in || in->type()==c_index_base::IDX_NULL) {
                        d+="(NULL) NULL";
                    } else if(in->type()==c_index_base::IDX_INT) {
                        d+="(INT) ";
                        d+=in->to_string();
                    } else if(in->type()==c_index_base::IDX_STR) {
                        d+="(STRING) >";
                        d+=in->to_string();
                        d+="<";
                    }
                } catch(...) {
                    d+="<SEGFAULT>";
                }
                d+="\n";
            }

            d+=indent;
            d+="second: ";
            d+=tos((*i1).value_id);
            d+="\n";

            d+=indent;
            d+="second: ";
            c_var_base* in=(*i1).data2.second;
            try {
                if(!in)
                    d+="(<NULL>) <NULL>";
                else if(typeid(*in)==typeid(c_var_unset))
                    d+="(UNSET) UNSET";
                else if(typeid(*in)==typeid(c_var_bool)) {
                    d+="(BOOL) ";
                    d+=(string)(in->is_true()?(string)"TRUE":(string)"FALSE");
                } else if(typeid(*in)==typeid(c_var_int)) {
                    MY_LONG_LONG l=dynamic_cast<c_var_int*>(in)->get();
                    d+="(INT) ";
                    d+=tos(l);
                } else if(typeid(*in)==typeid(c_var_ref)) {
                    long from=(long)dynamic_cast<c_var_ref*>(in)->get_referencing();
                    long to=(long)dynamic_cast<c_var_ref*>(in)->get_referenced();
                    d+="(REF) ";
                    d+=tos(from);
                    d+=" -> ";
                    d+=tos(to);
                } else if(typeid(*in)==typeid(c_var_array)) {
                    d+="(ARRAY)";
                } else if(typeid(*in)==typeid(c_var_string)) {
                    string s=dynamic_cast<c_var_string*>(in)->get();
                    d+="(STRING) >";
                    d+=s;
                    d+="<";
                }
            } catch(...) {
                d+="<SEGFAULT>";
            }
            d+="\n";
            try {
                if(in) {
                    d+=indent;
                    d+="Session ID: ";
                    d+=tos(in->get_session());
                    switch(in->get_session()) {
                        case SESSION_INVALID_ID:
                            d+=" (SESSION_INVALID_ID)";
                            break;
                        case SESSION_RESERVED_ID:
                            d+=" (SESSION_RESERVED_ID)";
                            break;
                        case SESSION_WORK_ID:
                            d+=" (SESSION_WORK_ID)";
                            break;
                        case SESSION_INTERNAL_ID:
                            d+=" (SESSION_INTERNAL_ID)";
                            break;
                        case SESSION_CORE_ENTRY_ID:
                            d+=" (SESSION_CORE_ENTRY_ID)";
                            break;
                        default:
                            d+=" (common session)";
                            break;
                    }
                }
            } catch(...) {
                d+="<SEGFAULT>";
            }
            d+="\n";
            if( ((*i1).value_id!=VAR_ID_INVALID && !(*i1).data2.second) ||
                ((*i1).value_id==VAR_ID_INVALID && (*i1).data2.second)  ||
                ( (*i1).data2.second ? (*i1).data2.second->get_id()!=(*i1).value_id : false ))
            {
                d+="=== ERROR ===\n";
            }
            fprintf(f,"%s\n",d.c_str());
        }
    }

    if(f)
        fclose(f);
    f=NULL;

    {
#ifdef _DEBUG
        // DEBUG: set breakpoint here!
        int x=0;
        x++;
#endif
    }

    in=false;
}

void c_var_keeper::update_variables(session_id_t session)
{
    map<var_id_t, c_entry*>::iterator i;
    for(i=storage.begin(); i!=storage.end(); ++i) {
        if(!(*i).second->val)
            continue;
        if((*i).second->val->get_session()==SESSION_INVALID_ID)
            (*i).second->val->set_session(session);
    }
}

void c_var_keeper::work_issued(session_id_t session, bool enter)
{
    session_id_t from, to;
    if(enter) {
        from=session;
        to=SESSION_WORK_ID;
    } else {
        from=SESSION_WORK_ID;
        to=session;
    }
    map<var_id_t, c_entry*>::iterator i;
    for(i=storage.begin(); i!=storage.end(); ++i) {
        if(!(*i).second->val)
            continue;
        if((*i).second->val->get_session()==from)
            (*i).second->val->set_session(to);
    }

    if(!enter)
        var_keeper->set_current_session(SESSION_CORE_ENTRY_ID);

    m_in_work=enter;
}

void c_var_keeper::internal_event_issued(session_id_t session, bool enter)
{
    session_id_t from, to;
    if(enter) {
        from=session;
        to=SESSION_INTERNAL_ID;
    } else {
        from=SESSION_INTERNAL_ID;
        to=session;
    }
    map<var_id_t, c_entry*>::iterator i;
    for(i=storage.begin(); i!=storage.end(); ++i) {
        if(!(*i).second->val)
            continue;
        if((*i).second->val->get_session()==from)
            (*i).second->val->set_session(to);
    }
}

void c_var_keeper::set_current_session(session_id_t session)
{
    if(m_in_work && SESSION_CORE_ENTRY_ID==session) {
        current_session=m_core_module_working_session;
        return;
    }
    current_session=session;
}

session_id_t c_var_keeper::get_current_session()
{
    return current_session;
}

void c_var_keeper::free_reference(var_id_t ref_id, log_function_ptr_t log_fn, void* user_data)
{
    if(VAR_ID_INVALID==ref_id)
        return;

    c_entry* e_f=get_e(ref_id,log_fn,user_data);
    if(!e_f) {
        string d="get_e(): no entry for ID #";
        d+=tos(ref_id);
        if(log_fn)
            (*log_fn)(LS_FATAL,GET_STR_DEBUG_INFO_EX(d),user_data);
        MY_ASSERT(e_f);
        return;
    }
    c_var_ref* ref=NULL;
    if(e_f->val && typeid(*e_f->val)==typeid(c_var_ref))
        ref=dynamic_cast<c_var_ref*>(e_f->val);
    if(!ref) {
        string d=__my_func__;
        d+=": it is not reference for ID #";
        d+=tos(ref_id);
        if(log_fn)
            (*log_fn)(LS_FATAL,GET_STR_DEBUG_INFO_EX(d),user_data);
        MY_ASSERT(ref);
        return;
    }
    var_id_t referencing=ref->get_referencing();
    map<var_id_t, var_id_t>::iterator i1=references.find(referencing);
    if(references.end()!=i1)
        references.erase(i1);
    ref->reset();

    if(!ref->get_in_deletion() && referencing!=VAR_ID_INVALID) {
        c_entry* f=get_e(referencing,log_fn,user_data);
        if(!f) {
            string d="get_e(): no entry for ID #";
            d+=tos(referencing);
            if(log_fn)
                (*log_fn)(LS_FATAL,GET_STR_DEBUG_INFO_EX(d),user_data);
            MY_ASSERT(e_f);
            return;
        }
        if(ref)
            MY_DELETE(ref);
    }
}

/*!
    \brief Constructor
    \author VooDooMan
    \version 1
    \date 2006-10-24
    \param key Key (index)
    \param val Value for index
*/
c_tree_node::c_tree_node(c_index_base* key, c_var_base* val)
{
    node_type=nt_normal;
    root_node_id=0;

    assign(key,val);
}

/*!
    \brief Constructor
    \author VooDooMan
    \version 1
    \date 2006-10-24
*/
c_tree_node::c_tree_node()
{
    value_id=VAR_ID_INVALID;
    data2.first=NULL;
    data2.second=NULL;

    node_type=nt_normal;
    root_node_id=0;
}

/*!
    \brief Asignment operator
    \author VooDooMan
    \version 1
    \date 2006-10-25
*/
c_tree_node& c_tree_node::operator=(const c_tree_node& orig)
{
    if(!reparenting_nodes) {
        if(orig.data2.first)
            assign(orig.data2.first,orig.data2.second);
        else {
            data2.first=NULL;
            data2.second=orig.data2.second;
        }
        value_id=orig.value_id;
    } else {
        value_id=orig.value_id;
        data2.first=orig.data2.first;
        data2.second=orig.data2.second;
        value_id=orig.value_id;
        node_type=orig.node_type;
        root_node_id=orig.root_node_id;
    }

    return *this;
}

/*!
    \brief Copy-constructor
    \author VooDooMan
    \version 1
    \date 2006-10-24
    \param orig Original instance
*/
c_tree_node::c_tree_node(const c_tree_node& orig)
{
    if(!reparenting_nodes) {
        if(orig.data2.first)
            assign(orig.data2.first,orig.data2.second);
        else {
            data2.first=NULL;
            data2.second=orig.data2.second;
        }
        value_id=orig.value_id;
        node_type=orig.node_type;
        root_node_id=orig.root_node_id;
    } else {
        value_id=orig.value_id;
        data2.first=orig.data2.first;
        data2.second=orig.data2.second;
        value_id=orig.value_id;
        node_type=orig.node_type;
        root_node_id=orig.root_node_id;
    }
}

/*!
    \brief Assignment operator
    \author VooDooMan
    \version 1
    \date 2006-10-24
    \param orig Original instance
*/
void c_tree_node::assign(const c_tree_node& orig)
{
    data2.first=orig.data2.first->copy();
    if(orig.data2.second) {
        data2.second=orig.data2.second->copy();
        value_id=data2.second->get_id();
    } else {
        value_id=VAR_ID_INVALID;
        data2.second=NULL;
    }
}

/*!
    \brief Destructor
    \author VooDooMan
    \version 1
    \date 2006-10-24
*/
c_tree_node::~c_tree_node()
{
    if(reparenting_nodes)
        return;
    var_keeper->var_del_lookup(data2.second);
    var_keeper->var_destroy_lookup(data2.second);

    if(data2.first) {
        MY_DELETE(data2.first);
        data2.first=NULL;
    }

    value_id=VAR_ID_INVALID;
    c_var_base* tmp=data2.second;
    data2.second=NULL;
    if(tmp) {
        tmp->set_in_deletion();
        MY_DELETE(tmp);
    }
}

void c_var_keeper::array_set(c_var_array* base, c_index_base*& key, c_var_base* val, c_module* UNUSEDARG(this_module), const string& UNUSEDARG(debug_prefix), const string& UNUSEDARG(debug), bool insensitive)
{
    if(!base || !val) {
        MY_DELETE(key);
        key=NULL;
        return; // huh?
    }

    tree<c_tree_node, c_single_element_allocator<tree_node_<c_tree_node> > >::iterator i1;
    tree<c_tree_node, c_single_element_allocator<tree_node_<c_tree_node> > >::sibling_iterator i2;

    if(var_lookup(base,i1))
        goto gotcha1;

    for(i1=variables.begin(); i1!=variables.end(); ++i1) {
        if((*i1).data2.second==base) {
gotcha1:

            tree<c_tree_node, c_single_element_allocator<tree_node_<c_tree_node> > >::iterator i3;

            if(var_lookup(val,i3))
                goto gotcha2;

            for(i3=variables.begin(); i3!=variables.end(); ++i3) {
                if((*i3).data2.second==val) {
                    break;
                }
            }

            if(variables.end()==i3 || (*i3).data2.second!=val) {
                MY_DELETE(key);
                key=NULL;
                return; // huh?
            }

gotcha2:

            c_var_base* old=NULL;
            for(i2=variables.begin(i1); i2!=variables.end(i1); ++i2) {
                //if((*i2).data2.first && (*i2).data2.first->is_identical(this_module,key,debug_prefix,debug)) {
                if((*i2).data2.first && (*i2).data2.first->equal(key,insensitive)) {
                    // old 'key' will be moved to temporary node store
                    //     to MY_DELETE(it later on discarding garbage collectio)n
                    //     NB: it is unsafe to 'MY_DELETE(old_key' here, since w)e
                    //         cannot predict how array_replace_tree() call
                    //         will deal with old 'old_key'
                    if(i2==i3) {
                        MY_DELETE(key);
                        key=NULL;
                        return;
                    }
                    CHECK(__LINE__,"#3-before");
                    array_replace_tree(i2,i3,key);
                    CHECK(__LINE__,"#4-after");
                    key=NULL;
                    return;
                }
            }
            //var_del_lookup(val);
            tree<c_tree_node, c_single_element_allocator<tree_node_<c_tree_node> > >::iterator x;
            if(val && typeid(*val)!=typeid(c_var_array)) {
                x=variables.append_child(i1);
                (*x).set_index(key);
                (*x).value_id=val->get_id();
                (*x).data2.second=val;
                var_add_tree_to_lookup(x);
                clear_subtree(i3);
                variables.erase(i3);
                key=NULL;

                CHECK(__LINE__,"#5");

                break;
            }

            x=variables.append_child(i1);
            (*x).set_index(key);
            (*x).value_id=val->get_id();
            (*x).data2.second=val;

            (*i3).assign(NULL);
            variables.reparent(x,i3);
            variables.erase(i3);

            var_add_tree_to_lookup(x);
            if(old)
                replace_ref(old,val);

            CHECK(__LINE__,"#6");
            key=NULL;
            break;
        }
    }
    MY_DELETE(key);
    key=NULL;
}

void c_var_keeper::array_del(c_var_array* base)
{
    tree<c_tree_node, c_single_element_allocator<tree_node_<c_tree_node> > >::iterator i1;
    if(var_lookup(base,i1))
        goto gotcha1;
    for(i1=variables.begin(); i1!=variables.end(); ++i1) {
        if((*i1).data2.second==base) {
gotcha1:
            variables.erase_children(i1);
            variables.erase(i1);
            break;
        }
    }
}

bool c_var_keeper::array_extract(c_var_array* base, c_index_base*& key, c_tree_node& result, c_module* UNUSEDARG(this_module), const string& UNUSEDARG(debug_prefix), const string& UNUSEDARG(debug))
{
    tree<c_tree_node, c_single_element_allocator<tree_node_<c_tree_node> > >::iterator i1;
    tree<c_tree_node, c_single_element_allocator<tree_node_<c_tree_node> > >::sibling_iterator i2;
    if(var_lookup(base,i1))
        goto gotcha1;
    for(i1=variables.begin(); i1!=variables.end(); ++i1) {
        if((*i1).data2.second==base) {
gotcha1:
            for(i2=variables.begin(i1); i2!=variables.end(i1); ++i2) {
                //if((*i2).data2.first && (*i2).data2.first->is_identical(this_module,key,debug_prefix,debug)) {
                if(*(*i2).data2.first==*key) {
                    MY_DELETE(key);
                    key=NULL;
                    result.assign(*i2);
                    return true;
                }
            }
            break;
        }
    }
    MY_DELETE(key);
    key=NULL;
    return false;
}

bool c_var_keeper::array_replace(c_module* UNUSEDARG(this_module), c_var_array* base, c_index_base*& key, c_var_base* value, bool insensitive, const string& UNUSEDARG(debug_prefix), const string& UNUSEDARG(debug))
{
    tree<c_tree_node, c_single_element_allocator<tree_node_<c_tree_node> > >::iterator i1;
    tree<c_tree_node, c_single_element_allocator<tree_node_<c_tree_node> > >::sibling_iterator i2;
    if(var_lookup(base,i1))
        goto gotcha1;
    for(i1=variables.begin(); i1!=variables.end(); ++i1) {
        if((*i1).data2.second==base) {
gotcha1:
            for(i2=variables.begin(i1); i2!=variables.end(i1); ++i2) {
                if((*i2).data2.first && (*i2).data2.first->equal(key,insensitive)) {
                    replace_ref((*i2).data2.second,value);
                    //var_del_lookup((*i2).data2.second);
                    add_local_var_no_reparent(get_current_local_node(),(*i2).data2.second);
                    (*i2).value_id=value ? value->get_id() : VAR_ID_INVALID;
                    (*i2).data2.second=value;
                    tree<c_tree_node, c_single_element_allocator<tree_node_<c_tree_node> > >::iterator i3=i2;
                    var_add_lookup(value,i3);
                    if(value)
                        (*i3).value_id=value->get_id();
                    MY_DELETE(key);
                    key=NULL;
                    return true;
                }
            }
            break;
        }
    }
    MY_DELETE(key);
    key=NULL;
    return false;
}

c_var_base* c_var_keeper::array_get(c_module* UNUSEDARG(this_module), c_var_array* base, c_index_base* key, bool insensitive, const string& UNUSEDARG(debug_prefix), const string& UNUSEDARG(debug))
{
    tree<c_tree_node, c_single_element_allocator<tree_node_<c_tree_node> > >::iterator i1;
    tree<c_tree_node, c_single_element_allocator<tree_node_<c_tree_node> > >::sibling_iterator i2;
    if(var_lookup(base,i1))
        goto gotcha1;
    for(i1=variables.begin(); i1!=variables.end(); ++i1) {
        if((*i1).data2.second==base) {
gotcha1:
            for(i2=variables.begin(i1); i2!=variables.end(i1); ++i2) {
                if((*i2).data2.first && (*i2).data2.first->equal(key,insensitive)) {
                    return (*i2).data2.second;
                }
            }
            break;
        }
    }
    return NULL;
}

bool c_var_keeper::array_is_identical(c_module* this_module, c_var_base* left, c_var_base* right, const string& debug_prefix, const string& debug)
{
    tree<c_tree_node, c_single_element_allocator<tree_node_<c_tree_node> > >::iterator i1;
    bool got=false;
    if(var_lookup(left,i1)) {
        got=true;
        goto gotcha1;
    }
    for(i1=variables.begin(); i1!=variables.end(); ++i1) {
        if((*i1).data2.second==left) {
            got=true;
            break;
        }
    }
gotcha1:
    if(!got)
        return false;
    int depth_l=variables.depth(i1);
    ++i1;

    tree<c_tree_node, c_single_element_allocator<tree_node_<c_tree_node> > >::iterator i2;
    got=false;
    if(var_lookup(right,i2)) {
        got=true;
        goto gotcha2;
    }
    for(i2=variables.begin(); i2!=variables.end(); ++i2) {
        if((*i2).data2.second==right) {
            got=true;
            break;
        }
    }
    if(!got)
        return false;
gotcha2:
    int depth_r=variables.depth(i2);
    ++i2;

    for(; i1!=variables.end() && i2!=variables.end(); ++i1, ++i2) {
        if(variables.depth(i1)<=depth_l && !(variables.depth(i2)<=depth_r))
            return false;
        if(!(variables.depth(i1)<=depth_l) && variables.depth(i2)<=depth_r)
            return false;
        if(variables.depth(i1)<=depth_l && variables.depth(i2)<=depth_r)
            return true;

        if(*(*i1).data2.first!=*(*i2).data2.first)
            return false;

        if((*i1).data2.second->get_var_type()==VARTYPE_ARRAY) {
            if((*i2).data2.second->get_var_type()==VARTYPE_ARRAY)
                continue;
            return false;
        }

        if(!(*i1).data2.second->is_identical(this_module,(*i2).data2.second,debug_prefix,debug))
            return false;
    }
    if(i1!=variables.end() && i2!=variables.end())
        return true;

    return false;
}

void c_var_keeper::array_unset(c_module* UNUSEDARG(this_module), c_var_array* base, c_index_base*& key, const string& UNUSEDARG(debug_prefix), const string& UNUSEDARG(debug))
{
    tree<c_tree_node, c_single_element_allocator<tree_node_<c_tree_node> > >::iterator i1;
    tree<c_tree_node, c_single_element_allocator<tree_node_<c_tree_node> > >::sibling_iterator i2;
    if(var_lookup(base,i1))
        goto gotcha1;
    for(i1=variables.begin(); i1!=variables.end(); ++i1) {
        if((*i1).data2.second==base) {
gotcha1:
            //bool first=true;
            for(i2=variables.begin(i1); i2!=variables.end(i1); ++i2) {
                //if((*i2).data2.first && (*i2).data2.first->is_identical(this_module,key,debug_prefix,debug)) {
                if((*i2).data2.first && *(*i2).data2.first==*key) {
                    tree<c_tree_node, c_single_element_allocator<tree_node_<c_tree_node> > >::iterator i3=i2;
                    move_to_temp(i3);
                    //(*i2).assign(NULL);
                    variables.erase(i3);
                    break;
                }
            }
            break;
        }
    }
    MY_DELETE(key);
    key=NULL;
}

bool c_var_keeper::array_is_equal(c_module* this_module, c_var_base* left, c_var_base* right, const string& debug_prefix, const string& debug)
{
    tree<c_tree_node, c_single_element_allocator<tree_node_<c_tree_node> > >::iterator l;
    bool got=false;
    if(var_lookup(left,l)) {
        got=true;
        goto gotcha1;
    }
    for(l=variables.begin(); l!=variables.end(); ++l) {
        if((*l).data2.second==left) {
            got=true;
            break;
        }
    }
    if(!got)
        return false;
gotcha1:
    int depth_l=variables.depth(l);
    unsigned children_l=variables.number_of_children(l);
    ++l;

    tree<c_tree_node, c_single_element_allocator<tree_node_<c_tree_node> > >::iterator r;
    got=false;
    if(var_lookup(right,r)) {
        got=true;
        goto gotcha2;
    }
    for(r=variables.begin(); r!=variables.end(); ++r) {
        if((*r).data2.second==right) {
            got=true;
            break;
        }
    }
    if(!got)
        return false;
gotcha2:
    int depth_r=variables.depth(r);
    unsigned children_r=variables.number_of_children(r);
    if(children_l!=children_r)
        return false;
    ++r;

    tree<c_tree_node, c_single_element_allocator<tree_node_<c_tree_node> > >::iterator i1;
    for(i1=l; i1!=variables.end(); ++i1) {
        if(variables.depth(i1)<=depth_l)
            return true;

        bool got=false;
        tree<c_tree_node, c_single_element_allocator<tree_node_<c_tree_node> > >::iterator i2;
        for(i2=r; i2!=variables.end(); ++i2) {
            if(variables.depth(i2)<=depth_r)
                return false;
            if(variables.depth(i1)-depth_l!=variables.depth(i2)-depth_r)
                continue;

            if((*i1).data2.first!=(*i2).data2.first)
                continue;

            got=true;
            break;
        }

        if(!got)
            return false;

        if((*i1).data2.second->get_var_type()==VARTYPE_ARRAY) {
            if((*i2).data2.second->get_var_type()==VARTYPE_ARRAY)
                continue;
            return false;
        }

        if(!(*i1).data2.second->is_identical(this_module,(*i2).data2.second,debug_prefix,debug))
            return false;
    }

    return true;
}

void c_var_keeper::array_set_all_as_persistent(c_var_array* base)
{
    tree<c_tree_node, c_single_element_allocator<tree_node_<c_tree_node> > >::iterator i1, i2;
    if(var_lookup(base,i1))
        goto gotcha1;
    for(i1=variables.begin(); i1!=variables.end(); ++i1) {
        if((*i1).data2.second==base) {
gotcha1:
            bool first=true;
            for(i2=i1; i2!=variables.end(); ++i2) {
                (*i2).data2.second->set_persistent();
                if(first) {
                    first=false;
                    continue;
                }
                if(variables.depth(i2)<=variables.depth(i1))
                    break;
            }
            break;
        }
    }
}

var_id_t c_var_keeper::set(var_id_t id, log_function_ptr_t log_fn, void* user_data, bool ro, c_var_base* &value, bool no_deref, vector<pair<var_id_t, var_id_t> >& changes1, vector<pair<var_id_t, var_id_t> >& changes2, c_module* this_module, const string& debug_prefix, const string& debug)
{
    var_id_t old_id=id;
    if(!no_deref)
        id=dereference(id);
    else {
#ifdef _DEBUG
        int x=0;
        x++;
#endif
    }

    CHECK(__LINE__,"#7");

    if(value && value->get_id()==id)
        return id;

    pair<var_id_t, var_id_t> last=make_pair(VAR_ID_INVALID,VAR_ID_INVALID);

    bool done=false;
    bool deleted=false;
    bool got_value=false;

    tree<c_tree_node, c_single_element_allocator<tree_node_<c_tree_node> > >::iterator i3;
    if(var_lookup(value,i3)) {
        got_value=true;
        goto gotcha1;
    }
    for(i3=variables.begin(); i3!=variables.end(); ++i3) {
        if((*i3).data2.second==value) {
            break;
        }
    }
    got_value=(variables.end()!=i3);
gotcha1:

    tree<c_tree_node, c_single_element_allocator<tree_node_<c_tree_node> > >::iterator i1, i2;
    for(i1=variables.begin(); i1!=variables.end(); ++i1) {
        if((*i1).value_id==id) {
            if(ro) {
                replace_ref((*i1).data2.second,value);
                //var_del_lookup((*i1).data2.second);
                add_local_var_no_reparent(get_current_local_node(),(*i1).data2.second);
                if((*i1).value_id!=VAR_ID_INVALID && value) {
                    if(last.first!=(*i1).value_id || last.second!=value->get_id()) {
                        last=make_pair((*i1).value_id,value->get_id());
                        changes1.push_back(last);
                        changes2.push_back(last);
                    }
                }

                if((*i1).data2.second) {
                    i2=variables.parent(i1);
                    c_var_array* a=dynamic_cast<c_var_array*>((*i2).data2.second);
                    if(a) {
                        c_index_base* idx=(*i1).data2.first->copy();
                        array_set(a,idx,value,this_module,debug_prefix,debug);
                        CHECK(__LINE__,"#8");
                        if(value && old_id!=id)
                            return old_id;
                        return value ? value->get_id() : VAR_ID_INVALID;
                    }
                }

                if(!got_value) {
                    remove_from_tree(value);
                    (*i1).data2.second=value;
                    (*i1).value_id=value ? value->get_id() : VAR_ID_INVALID;
                    var_add_lookup(value,i1);
                    CHECK(__LINE__,"#9");
                } else {
                    array_replace_tree(i1,i3,(*i1).data2.first);
                    CHECK(__LINE__,"#10");
                }
                return value->get_id();
            } else {
                if(!done && (*i1).data2.second && (*i1).data2.second->get_read_only()) {
                    if((*i1).data2.second!=value) {
                        if(log_fn)
                            (*log_fn)(LS_WARNING,"Cannot modify read-only variable",user_data);
                        value=(*i1).data2.second;
                    }
                    return value->get_id();
                } else {
                    c_var_base* to_del=NULL;
                    if(!no_deref && (*i1).data2.second!=value) {
                        if(dereference((*i1).data2.second->get_id())!=dereference(value->get_id())) {
                            if((*i1).value_id!=VAR_ID_INVALID && value) {
                                if(last.first!=(*i1).value_id || last.second!=value->get_id()) {
                                    last=make_pair((*i1).value_id,value->get_id());
                                    changes1.push_back(last);
                                    changes2.push_back(last);
                                }
                            }
                            if(!deleted) {
                                replace_ref((*i1).data2.second,value);
                                to_del=(*i1).data2.second;
                            } else {
                            }
                        }
                        if((*i1).data2.first) {
                            i2=variables.parent(i1);
                            c_var_array* a=dynamic_cast<c_var_array*>((*i2).data2.second);
                            if(a) {
                                c_index_base* idx=(*i1).data2.first->copy();
                                array_set(a,idx,value,this_module,debug_prefix,debug);
                                CHECK(__LINE__,"#11");
                                if(value && old_id!=id)
                                    return old_id;
                                return value ? value->get_id() : VAR_ID_INVALID;
                            }
                        }
                        if(!got_value) {
                            //var_del_lookup((*i1).data2.second);
                            add_local_var_no_reparent(get_current_local_node(),(*i1).data2.second);
                            remove_from_tree(value);
                            (*i1).value_id=value ? value->get_id() : VAR_ID_INVALID;
                            (*i1).data2.second=value;
                            //var_add_lookup((*i1).data2.second,i1);
                        } else {
                            c_index_base* key=(*i1).data2.first;
                            i1=array_replace_tree(i1,i3,key);
                            done=true;
                            deleted=true;
                            break;
                        }
                        deleted=true;
                    }
                    if(no_deref && (*i1).data2.second!=value) {
                        {
                            if((*i1).value_id!=VAR_ID_INVALID && value) {
                                if(last.first!=(*i1).value_id || last.second!=value->get_id()) {
                                    last=make_pair((*i1).value_id,value->get_id());
                                    changes1.push_back(last);
                                    changes2.push_back(last);
                                }
                            }
                            if(!deleted) {
                                replace_ref((*i1).data2.second,value);
                                to_del=(*i1).data2.second;
                                if(!got_value) {
                                    //var_del_lookup((*i1).data2.second);
                                    add_local_var_no_reparent(get_current_local_node(),(*i1).data2.second);
                                    remove_from_tree(value);
                                    (*i1).value_id=value ? value->get_id() : VAR_ID_INVALID;
                                    (*i1).data2.second=value;
                                    //var_add_lookup((*i1).data2.second,i1);
                                } else {
                                    i1=array_replace_tree(i1,i3,(*i1).data2.first);
                                    done=true;
                                    deleted=true;
                                    break;
                                }
                                deleted=true;
                            } else {
                                if(!got_value) {
                                    //var_del_lookup((*i1).data2.second);
                                    add_local_var_no_reparent(get_current_local_node(),(*i1).data2.second);
                                    remove_from_tree(value);
                                    (*i1).value_id=value ? value->get_id() : VAR_ID_INVALID;
                                    (*i1).data2.second=value;
                                    //var_add_lookup((*i1).data2.second,i1);
                                } else {
                                    i1=array_replace_tree(i1,i3,(*i1).data2.first);
                                    done=true;
                                    deleted=true;
                                    break;
                                }
                            }
                        }
                    }
                    done=true;
                    if(to_del) {
                        break; // important!
                    }
                    break; // important!
                }
            }
        }
    }

    if(done) {
        CHECK(__LINE__,"#12");
        if(value && old_id!=id)
            return old_id;
        return value ? value->get_id() : VAR_ID_INVALID;
    }

    CHECK(__LINE__,"#13");

    DUMP(__LINE__,"#14");

    string d="Could not find variable ID #";
    d+=tos(id);
    d+=" in variable tree";
    if(log_fn)
        (*log_fn)(LS_FATAL,GET_STR_DEBUG_INFO_EX(d),user_data);
    return VAR_ID_INVALID;
}

MY_STATIC void var_arr_log_fn(c_var_keeper::e_log_severity severity, const string& msg, void* mod)
{
    c_module* module=(c_module*)mod;
    if(!module) {
        module=core_mod;
    }
    if(module)
        if(module->get_logger()) {
            string log;
            switch(severity) {
                case c_var_keeper::LS_FATAL:
                    log="FATAL: ";
                    break;
                case c_var_keeper::LS_WARNING:
                    log="WARNING: ";
                    break;
                default:
                    log="<OTHER_SEVERITY>: ";
                    break;
            }
            log+=msg;
            module->get_logger()->log_print1(E_LOG_DEBUG,LOG_SCOPE_DEFAULT,"",log);
        }
}

string c_var_keeper::array_get_script_string(c_module* this_module, const string& script_type, c_var_array* base)
{
    string res;

    est st;

    if(!script_type.compare("php") || !script_type.compare("php_2") || !script_type.compare("logic.txt"))
        st=est_php;
    else if(!script_type.compare("perl"))
        st=est_perl;
    else
        return (string)""; // huh?

    switch(st) {
        case est_php:
            res="array(";
            break;
        case est_perl:
            res="{";
            break;
    }

    tree<c_tree_node, c_single_element_allocator<tree_node_<c_tree_node> > >::iterator i1;
    if(var_lookup(base,i1))
        goto gotcha1;
    for(i1=variables.begin(); i1!=variables.end(); ++i1) {
        if((*i1).data2.second==base) {
gotcha1:
            int depth=variables.depth(i1)+1;
            int last_depth=depth;
            ++i1;
            bool first=true;
            for(; i1!=variables.end(); ++i1) {
                if(variables.depth(i1)<depth) {
                    switch(st) {
                        case est_php:
                            res+=")";
                            break;
                        case est_perl:
                            res+="}";
                            break;
                    }
                    break;
                }

                if(variables.depth(i1)<last_depth) {
                    last_depth=variables.depth(i1);

                    switch(st) {
                        case est_php:
                            res+=")";
                            break;
                        case est_perl:
                            res+="}";
                            break;
                    }
                }

                if(!first) {
                    first=false;
                    switch(st) {
                        case est_php:
                            res+=", ";
                            break;
                        case est_perl:
                            res+=", ";
                            break;
                    }
                }

                res+=dump_script_simple(this_module,st,(*i1).data2.first);

                switch(st) {
                    case est_php:
                        res+=" => ";
                        break;
                    case est_perl:
                        res+=" => ";
                        break;
                }

                // process any reference:
                c_var_base* val=get((*i1).value_id,var_arr_log_fn,this_module);

                if(val && typeid(*val)==typeid(c_var_array)) {
                    last_depth=variables.depth(i1)+1;

                    switch(st) {
                        case est_php:
                            res+="array(";
                            break;
                        case est_perl:
                            res+="{";
                            break;
                    }

                    continue;
                }

                // ==

                res+=dump_script_simple(this_module,st,val);

                continue;
            }
            if(variables.end()==i1) {
                switch(st) {
                    case est_php:
                        res+=")";
                        break;
                    case est_perl:
                        res+="}";
                        break;
                }
            }
            break;
        }
    }

    return res;
}

string c_var_keeper::dump_script_simple(c_module* this_module, est st, c_var_base* base)
{
    if(!base)
        return (string)"";

    switch(st) {
        case est_php:
            if(!base)
                return (string)"NULL";
            if(typeid(*base)==typeid(c_var_unset)) {
                c_var_unset* u=dynamic_cast<c_var_unset*>(base);
                return u->get_script_string(this_module,"php");
            }
            if(typeid(*base)==typeid(c_var_bool)) {
                c_var_bool* b=dynamic_cast<c_var_bool*>(base);
                return b->get_script_string(this_module,"php");
            }
            if(typeid(*base)==typeid(c_var_int)) {
                c_var_int* i=dynamic_cast<c_var_int*>(base);
                return i->get_script_string(this_module,"php");
            }
            if(typeid(*base)==typeid(c_var_string)) {
                c_var_string* s=dynamic_cast<c_var_string*>(base);
                return s->get_script_string(this_module,"php");
            }
            return (string)"NULL";
        case est_perl:
            if(!base)
                return (string)"0";
            if(typeid(*base)==typeid(c_var_unset)) {
                c_var_unset* u=dynamic_cast<c_var_unset*>(base);
                return u->get_script_string(this_module,"perl");
            }
            if(typeid(*base)==typeid(c_var_bool)) {
                c_var_bool* b=dynamic_cast<c_var_bool*>(base);
                return b->get_script_string(this_module,"perl");
            }
            if(typeid(*base)==typeid(c_var_int)) {
                c_var_int* i=dynamic_cast<c_var_int*>(base);
                return i->get_script_string(this_module,"perl");
            }
            if(typeid(*base)==typeid(c_var_string)) {
                c_var_string* s=dynamic_cast<c_var_string*>(base);
                return s->get_script_string(this_module,"perl");
            }
            return (string)"0";
    }
    return (string)""; // huh?
}

string c_var_keeper::dump_script_simple(c_module* this_module, est st, c_index_base* value)
{
    switch(st) {
        case est_php:
            if(value->type()==c_index_base::IDX_NULL) {
                return (string)"NULL";
            }
            if(value->type()==c_index_base::IDX_INT) {
                c_var_int i=dynamic_cast<c_index_int*>(value)->get();
                return i.get_script_string(this_module,"php");
            }
            if(value->type()==c_index_base::IDX_STR) {
                c_var_string i=dynamic_cast<c_index_str*>(value)->get();
                return i.get_script_string(this_module,"php");
            }
            return (string)"NULL";
        case est_perl:
            if(value->type()==c_index_base::IDX_NULL) {
                return (string)"0";
            }
            if(value->type()==c_index_base::IDX_INT) {
                c_var_int i=dynamic_cast<c_index_int*>(value)->get();
                return i.get_script_string(this_module,"perl");
            }
            if(value->type()==c_index_base::IDX_STR) {
                c_var_string i=dynamic_cast<c_index_str*>(value)->get();
                return i.get_script_string(this_module,"perl");
            }
            return (string)"0";
    }
    return (string)""; // huh?
}

void c_var_keeper::array_unset_ro_value(c_var_array* base, c_var_base* value)
{
    if(!value)
        return;

    tree<c_tree_node, c_single_element_allocator<tree_node_<c_tree_node> > >::iterator i1;
    tree<c_tree_node, c_single_element_allocator<tree_node_<c_tree_node> > >::sibling_iterator i2;
    if(var_lookup(base,i1))
        goto gotcha1;
    for(i1=variables.begin(); i1!=variables.end(); ++i1) {
        if((*i1).data2.second==base) {
gotcha1:
            for(i2=variables.begin(i1); i2!=variables.end(i1); ++i2) {
                if((*i2).data2.second==value) {
                    // 'value' WILL BE DELETED by destructor!
                    tree<c_tree_node, c_single_element_allocator<tree_node_<c_tree_node> > >::iterator i3=i2;
                    move_to_temp(i3);
                    variables.erase(i3);
                    // so return!
                    return;
                }
            }
            break;
        }
    }

    MY_DELETE(value);
}

string c_var_keeper::array_dump(c_module* this_module, c_var_array* base)
{
    string res="(ARRAY)\n";

    tree<c_tree_node, c_single_element_allocator<tree_node_<c_tree_node> > >::iterator i1;
    if(var_lookup(base,i1))
        goto gotcha1;
    for(i1=variables.begin(); i1!=variables.end(); ++i1) {
        if((*i1).data2.second==base) {
gotcha1:
            int depth=variables.depth(i1)+1;
            ++i1;
            for(; i1!=variables.end(); ++i1) {
                if(variables.depth(i1)<depth)
                    break;
                int diff=variables.depth(i1)-depth+1;
                for(int i2=0; i2<diff; ++i2)
                    res+="    ";
                res+="[";
                if((*i1).data2.first && (*i1).data2.first->type()==c_index_int::IDX_INT) {
                    res+="(INT) ";
                    res+=(*i1).data2.first->to_string();
                }
                if((*i1).data2.first && (*i1).data2.first->type()==c_index_int::IDX_STR) {
                    res+="(STRING) ";
                    res+=dump_script_simple(this_module,est_php,(*i1).data2.first);
                }
                res+="] => ";

                // process any reference:
                c_var_base* val=get((*i1).value_id,var_arr_log_fn,this_module);

                if(val && typeid(*val)==typeid(c_var_array)) {
                    res+="(ARRAY)\n";
                    continue;
                }

                if(typeid(*val)==typeid(c_var_int))
                    res+="(INT) ";
                if(typeid(*val)==typeid(c_var_string))
                    res+="(STRING) ";
                if(typeid(*val)==typeid(c_var_bool))
                    res+="(BOOL) ";
                if(typeid(*val)==typeid(c_var_unset))
                    res+="(NULL) ";

                res+=dump_script_simple(this_module,est_php,val);
                res+="\n";
            }
            break;
        }
    }

    return res;
}

void c_var_keeper::array_dump_for_botnet(c_module* this_module, c_var_array* base, char* &data, size_t& size)
{
    char* d=MY_NEW char[1];
    size=1;
    d[0]='A'; // stands for 'array'

    tree<c_tree_node, c_single_element_allocator<tree_node_<c_tree_node> > >::iterator i1;
    if(var_lookup(base,i1))
        goto gotcha1;
    for(i1=variables.begin(); i1!=variables.end(); ++i1) {
        if((*i1).data2.second==base) {
gotcha1:
            int depth=variables.depth(i1)+1;
            int last_depth=depth;
            ++i1;
            for(; i1!=variables.end(); ++i1) {
                if(variables.depth(i1)<last_depth) {
                    //int diff=last_depth-variables.depth(i1)-1;
                    int diff=last_depth-variables.depth(i1);
                    char* nd=MY_NEW char[size+diff];
                    memcpy(nd,d,size);
                    for(int i2=0; i2<diff; ++i2)
                        nd[size+1-1+i2]=0; // end of array
                    size+=diff;
                    MY_DELETE_A(d);
                    d=nd;
                }
                if(variables.depth(i1)<depth) {
                    last_depth=depth;
                    break;
                }
                last_depth=variables.depth(i1);

                char* key_d=NULL;
                size_t key_s=0;
                if((*i1).data2.second)
                    (*i1).data2.first->dump_for_botnet(key_d,key_s);

                {
                    char* nd=MY_NEW char[size+key_s];
                    memcpy(nd,d,size);
                    memcpy(&(nd[size]),key_d,key_s);
                    MY_DELETE_A(key_d);
                    MY_DELETE_A(d);
                    d=nd;
                    size+=key_s;
                }

                // process any reference:
                c_var_base* val=get((*i1).value_id,var_arr_log_fn,this_module);

                char* val_d=NULL;
                size_t val_s=0;

                if(val && typeid(*val)==typeid(c_var_array)) {
                    val_s=1;
                    val_d=MY_NEW char[val_s];
                    val_d[0]='A';
                    last_depth=variables.depth(i1)+1;
                } else {
                    val->dump_for_botnet(this_module,val_d,val_s);
                }

                {
                    char* nd=MY_NEW char[size+val_s];
                    memcpy(nd,d,size);
                    memcpy(&(nd[size]),val_d,val_s);
                    MY_DELETE_A(val_d);
                    MY_DELETE_A(d);
                    d=nd;
                    size+=val_s;
                }
            }
            if(depth<last_depth) {
                int diff=last_depth-depth;
                char* nd=MY_NEW char[size+diff];
                memcpy(nd,d,size);
                for(int i2=0; i2<diff; ++i2)
                    nd[size+1-1+i2]=0; // end of array
                size+=diff;
                MY_DELETE_A(d);
                d=nd;
            }
            break;
        }
    }

    data=d;
}

void c_var_keeper::array_clear_unset(c_module* UNUSEDARG(this_module), c_var_array* UNUSEDARG(base))
{
    /*
    tree<c_tree_node, c_single_element_allocator<tree_node_<c_tree_node> > >::iterator i1, i2;
    if(var_lookup(base,i1))
        goto gotcha1;
    for(i1=variables.begin(); i1!=variables.end(); ++i1) {
        if((*i1).data2.second==base) {
gotcha1:
            int depth=variables.depth(i1)+1;
            ++i1;
            i2=i1;
        again:
            for(; i1!=variables.end(); i1.skip_children(), ++i1) {
                if(variables.depth(i1)<depth)
                    break;

                // process any reference:
                c_var_base* val=get((*i1).value_id,var_arr_log_fn,this_module);

                if(!val || typeid(*val)==typeid(c_var_unset)) {
                    if(i1==i2) {
                        variables.erase(i1);
                        break;
                    }
                    variables.erase(i1);
                    i1=i2;
                    goto again;
                }
            }
            break;
        }
    }
    */
}

c_var_base* c_var_keeper::array_stream_read(c_var_array* base, vector<c_index_base*>& path, vector<bool>& insensitive, c_module* UNUSEDARG(this_module), const string& UNUSEDARG(debug))
{
    vector<c_index_base*>::const_iterator pi=path.begin();
    vector<bool>::const_iterator ii=insensitive.begin();

    tree<c_tree_node, c_single_element_allocator<tree_node_<c_tree_node> > >::iterator i1;
    if(var_lookup(base,i1))
        goto gotcha1;
    for(i1=variables.begin(); i1!=variables.end(); ++i1) {
        if((*i1).data2.second==base) {
gotcha1:
            int depth=variables.depth(i1)+1;
            int got_depth=depth;
            bool got=false;
            ++i1;
            for(; i1!=variables.end(); ++i1) {
                if(variables.depth(i1)<depth)
                    return NULL;

                if(variables.depth(i1)>got_depth && got) {
                    got=false;
                    got_depth=variables.depth(i1);
                    // we continue in deeper scope
                } else if(variables.depth(i1)<got_depth) {
                    return NULL;
                } else if(variables.depth(i1)>got_depth && !got) {
                    // skip deeper scope
                    continue;
                } else if(variables.depth(i1)==got_depth && got) {
                    return NULL;
                }
                // else: same scope && !got

                if((*pi)->type()==c_index_int::IDX_INT && (*i1).data2.first->type()==c_index_int::IDX_INT) {
                    //if((*pi)->is_equal(this_module,(*i1).data2.first,"",debug)) {
                    if(*(*i1).data2.first==*(*pi)) {
                        // got it
                        ++pi;
                        ++ii;
                        if(path.end()==pi || insensitive.end()==ii) {
                            return (*i1).data2.second;
                        }
                        got=true;
                        continue;
                    }
                }
                if((*pi)->equal((*i1).data2.first,*ii)) {
                    // got it
                    ++pi;
                    ++ii;
                    if(path.end()==pi || insensitive.end()==ii) {
                        return (*i1).data2.second;
                    }
                    got=true;
                    continue;
                }
            }
            return NULL;
        }
    }

    return NULL;
}

void c_var_keeper::array_stream_write(c_var_array* base, vector<c_index_base*>& path, const vector<bool>& insensitive, c_var_base* value, c_module* UNUSEDARG(this_module), const string& UNUSEDARG(debug))
{
    vector<c_index_base*>::const_iterator pi=path.begin();
    vector<bool>::const_iterator ii=insensitive.begin();

    tree<c_tree_node, c_single_element_allocator<tree_node_<c_tree_node> > >::iterator i1;
    if(var_lookup(base,i1))
        goto gotcha1;
    for(i1=variables.begin(); i1!=variables.end(); ++i1) {
        if((*i1).data2.second==base) {
gotcha1:
            int depth=variables.depth(i1)+1;
            int got_depth=depth;
            bool got=false;
            tree<c_tree_node, c_single_element_allocator<tree_node_<c_tree_node> > >::iterator parent=i1;
            ++i1;
            bool build_new=true;
            for(; i1!=variables.end(); ++i1) {
                build_new=false;
                if(variables.depth(i1)<depth) {
                    build_new=true;
                    break;
                }

                if(variables.depth(i1)>got_depth && got) {
                    got=false;
                    got_depth=variables.depth(i1);
                    // we continue in deeper scope
                } else if(variables.depth(i1)<got_depth) {
                    build_new=true;
                    break;
                } else if(variables.depth(i1)>got_depth && !got) {
                    // skip deeper scope
                    continue;
                } else if(variables.depth(i1)==got_depth && got) {
                    build_new=true;
                    break;
                }
                // else: same scope && !got
                if((*pi)->equal((*i1).data2.first,*ii)) {
                    // got it
                    ++pi;
                    ++ii;
                    if(path.end()==pi || insensitive.end()==ii) {
                        replace_ref((*i1).data2.second,value);
                        //var_del_lookup((*i1).data2.second);
                        add_local_var_no_reparent(get_current_local_node(),(*i1).data2.second);
                        (*i1).data2.second=value;
                        (*i1).value_id=value ? value->get_id() : VAR_ID_INVALID;
                        //var_add_lookup(value,i1);
                        return;
                    }
                    got=true;
                    parent=i1;
                    continue;
                }
            }
            if(build_new) {
                for(; path.end()!=pi; ++pi) {
                    c_var_base* val;
                    if(pi+1==path.end()) {
                        val=value;
                    } else {
                        val=MY_NEW c_var_array();
                    }
                    {
                        c_tree_node n;
                        parent=variables.append_child(parent,n);
                    }
                    (*parent).assign((*pi),val);
                    var_add_lookup(val,parent);
                }
            }
            return;
        }
    }

    return;
}

void c_var_keeper::array_unset(c_module* UNUSEDARG(this_module), c_var_array* base, const string& key, bool insensitive, const string& UNUSEDARG(debug_prefix), const string& UNUSEDARG(debug))
{
    tree<c_tree_node, c_single_element_allocator<tree_node_<c_tree_node> > >::iterator i1;
    tree<c_tree_node, c_single_element_allocator<tree_node_<c_tree_node> > >::sibling_iterator i2;
    if(var_lookup(base,i1))
        goto gotcha1;
    for(i1=variables.begin(); i1!=variables.end(); ++i1) {
        if((*i1).data2.second==base) {
gotcha1:
            for(i2=variables.begin(i1); i2!=variables.end(i1); ++i2) {
                if(!(*i2).data2.first || (*i2).data2.first->type()!=c_index_int::IDX_STR)
                    continue;

                string k=(*i2).data2.first->to_string();
                if(!insensitive) {
                    if(!k.compare(key)) {
                        var_del_lookup((*i2).data2.second);
                        add_local_var_no_reparent(get_current_local_node(),(*i2).data2.second);
                        //MY_DELETE((*i2).data2.first);
                        clear_subtree(i2);
                        variables.erase(i2);
                        break;
                    }
                } else {
                    if(!cmp_strings_case_insensitive(k,key)) {
                        var_del_lookup((*i2).data2.second);
                        add_local_var_no_reparent(get_current_local_node(),(*i2).data2.second);
                        clear_subtree(i2);
                        variables.erase(i2);
                        break;
                    }
                }
            }
            break;
        }
    }
}

void c_var_keeper::array_replace_index(c_module* UNUSEDARG(this_module), c_var_array* base, const string& old_key, const string& new_key)
{
    tree<c_tree_node, c_single_element_allocator<tree_node_<c_tree_node> > >::iterator i1;
    tree<c_tree_node, c_single_element_allocator<tree_node_<c_tree_node> > >::sibling_iterator i2;
    if(var_lookup(base,i1))
        goto gotcha1;
    for(i1=variables.begin(); i1!=variables.end(); ++i1) {
        if((*i1).data2.second==base) {
gotcha1:
            for(i2=variables.begin(i1); i2!=variables.end(i1); ++i2) {
                if(!(*i2).data2.first || (*i2).data2.first->type()!=c_index_int::IDX_STR)
                    continue;
                if(*(*i2).data2.first==old_key) {
                    base->update_lookup((*i2).data2.first,NULL);
                    c_index_base* idx=MY_NEW c_index_str(new_key);
                    (*i2).set_index(idx);
                    base->update_lookup((*i2).data2.first,(*i2).data2.second);
                    break;
                }
            }
            break;
        }
    }
}

c_var_base* c_var_keeper::array_get(c_module* UNUSEDARG(this_module), c_var_array* base, const string& key, bool insensitive, const string& UNUSEDARG(debug_prefix), const string& UNUSEDARG(debug))
{
    string k;

    if(insensitive)
        k=str_to_lower(key);

    tree<c_tree_node, c_single_element_allocator<tree_node_<c_tree_node> > >::iterator i1;
    tree<c_tree_node, c_single_element_allocator<tree_node_<c_tree_node> > >::sibling_iterator i2;
    if(var_lookup(base,i1))
        goto gotcha1;
    for(i1=variables.begin(); i1!=variables.end(); ++i1) {
        if((*i1).data2.second==base) {
gotcha1:
            for(i2=variables.begin(i1); i2!=variables.end(i1); ++i2) {
                if((*i2).data2.first && (*i2).data2.first->type()==c_index_base::IDX_STR) {
                    if((*i2).data2.first && (*i2).data2.first->equal(key,insensitive))
                        return (*i2).data2.second;
                    continue;
                }
            }
            break;
        }
    }
    return NULL;
}

void c_var_keeper::array_remove_values(c_var_array* base, const string& val)
{
    tree<c_tree_node, c_single_element_allocator<tree_node_<c_tree_node> > >::iterator i1;
    tree<c_tree_node, c_single_element_allocator<tree_node_<c_tree_node> > >::sibling_iterator i2;
    if(var_lookup(base,i1))
        goto gotcha1;
    for(i1=variables.begin(); i1!=variables.end(); ++i1) {
        if((*i1).data2.second==base) {
gotcha1:
        again:
            for(i2=variables.begin(i1); i2!=variables.end(i1); ++i2) {
                if(typeid(*(*i2).data2.second)==typeid(c_var_string)) {
                    if((*(*i2).data2.second)==val) {
                        variables.erase(i2);
                        goto again;
                    }
                }
            }
            break;
        }
    }
}

void c_var_keeper::array_remove_values(c_var_array* base, const string& key, const string& val)
{
    tree<c_tree_node, c_single_element_allocator<tree_node_<c_tree_node> > >::iterator i1;
    tree<c_tree_node, c_single_element_allocator<tree_node_<c_tree_node> > >::sibling_iterator i2;
    if(var_lookup(base,i1))
        goto gotcha1;
    for(i1=variables.begin(); i1!=variables.end(); ++i1) {
        if((*i1).data2.second==base) {
gotcha1:
        again:
            for(i2=variables.begin(i1); i2!=variables.end(i1); ++i2) {
                if(typeid(*(*i2).data2.first)==typeid(c_index_str)) {
                    if(typeid(*(*i2).data2.second)==typeid(c_var_string)) {
                        if((*i2).data2.first->equal(key,false) && (*(*i2).data2.second)==val) {
                            variables.erase(i2);
                            goto again;
                        }
                    }
                }
            }
            break;
        }
    }
}

unsigned c_var_keeper::array_num_children(c_module* UNUSEDARG(this_module), c_var_array* base)
{
    tree<c_tree_node, c_single_element_allocator<tree_node_<c_tree_node> > >::iterator i1;
    if(var_lookup(base,i1))
        goto gotcha1;
    for(i1=variables.begin(); i1!=variables.end(); ++i1) {
        if((*i1).data2.second==base) {
gotcha1:
            return variables.number_of_children(i1);
        }
    }
    return 0;
}

map<string,c_var_base*> c_var_keeper::array_get_string_indexes(c_module* UNUSEDARG(this_module), c_var_array* base)
{
    map<string,c_var_base*> res;

    tree<c_tree_node, c_single_element_allocator<tree_node_<c_tree_node> > >::iterator i1;
    tree<c_tree_node, c_single_element_allocator<tree_node_<c_tree_node> > >::sibling_iterator i2;
    if(var_lookup(base,i1))
        goto gotcha1;
    for(i1=variables.begin(); i1!=variables.end(); ++i1) {
        if((*i1).data2.second==base) {
gotcha1:
            for(i2=variables.begin(i1); i2!=variables.end(i1); ++i2) {
                if((*i2).data2.first && (*i2).data2.first->type()==c_index_base::IDX_STR) {
                    res.insert(make_pair(
                        (*i2).data2.first->to_string(),
                        (*i2).data2.second
                        ));
                }
            }
            break;
        }
    }
    return res;
}

unsigned c_var_keeper::array_num_nonunset_children(c_module* this_module, c_var_array* base)
{
    unsigned res=0;

    tree<c_tree_node, c_single_element_allocator<tree_node_<c_tree_node> > >::iterator i1;
    tree<c_tree_node, c_single_element_allocator<tree_node_<c_tree_node> > >::sibling_iterator i2;
    if(var_lookup(base,i1))
        goto gotcha1;
    for(i1=variables.begin(); i1!=variables.end(); ++i1) {
        if((*i1).data2.second==base) {
gotcha1:
            for(i2=variables.begin(i1); i2!=variables.end(i1); ++i2) {
                // process any reference:
                c_var_base* val=get((*i2).value_id,var_arr_log_fn,this_module);

                if(val && typeid(*val)!=typeid(c_var_unset))
                    ++res;
            }
            break;
        }
    }
    return res;
}

c_tree_node* c_var_keeper::array_get_nonunset_child(c_module* UNUSEDARG(this_module), c_var_array* base, MY_U_LONG_LONG n)
{
    tree<c_tree_node, c_single_element_allocator<tree_node_<c_tree_node> > >::iterator i1;
    tree<c_tree_node, c_single_element_allocator<tree_node_<c_tree_node> > >::sibling_iterator i2;
    if(var_lookup(base,i1))
        goto gotcha1;
    for(i1=variables.begin(); i1!=variables.end(); ++i1) {
        if((*i1).data2.second==base) {
gotcha1:
            for(i2=variables.begin(i1); i2!=variables.end(i1); ++i2) {
                if(n--==0)
                    return &(*i2);
            }
            break;
        }
    }

    // index out of bounds
    return NULL;
}

void c_var_keeper::assign(var_id_t old_val, c_var_base* new_val)
{
    tree<c_tree_node, c_single_element_allocator<tree_node_<c_tree_node> > >::iterator i1;
    for(i1=variables.begin(); i1!=variables.end(); ++i1) {
        if((*i1).value_id==old_val) {
            replace_ref((*i1).data2.second,new_val);
            add_local_var_no_reparent(get_current_local_node(),(*i1).data2.second);
            (*i1).data2.second=new_val;
            (*i1).value_id=new_val ? new_val->get_id() : VAR_ID_INVALID;
            return;
        }
    }
}

unsigned c_var_keeper::get_unique_node_id(const char* file, const char* func, int line)
{
    unsigned new_local_id;
    for(;;) {
        new_local_id=++last_local;
        if(!new_local_id)
            continue; // overflow
        bool in_use=false;
        for(vector<unsigned>::iterator i1=locals.begin(); i1!=locals.end(); ++i1) {
            if(*i1==new_local_id) {
                in_use=true;
                break;
            }
        }
        if(in_use)
            continue;
        // okay now
        break;
    }
    locals.push_back(new_local_id);
    {
#if DEBUG_VAR_KEEPER_DEBUG_STRING_DISABLE==0
        locals_debug.insert(make_pair(new_local_id,c_locals_debug_info(file,line,func)));
#endif
    }
    return new_local_id;
}

unsigned c_var_keeper::register_local(c_var_array*& base, c_var_array* copy_from, const char* file, const char* func, int line)
{
    if(SESSION_CORE_ENTRY_ID!=current_session) {
        in_construct=true;
        unsigned local_id_save=current_local;
        unsigned new_local_id=get_unique_node_id(file,func,line);

        current_local=new_local_id;
        tree<c_tree_node, c_single_element_allocator<tree_node_<c_tree_node> > >::iterator x;
        {
            if(!copy_from)
                base=MY_NEW c_var_array(0);
            else
                base=dynamic_cast<c_var_array*>(copy_from->copy(0));

            {
                c_tree_node n;
                x=variables.insert(variables.begin(),n);
            }
            (*x).root_node_id=current_local;
            (*x).value_id=base->get_id();
            (*x).clear_index();
            (*x).data2.second=base;
            (*x).node_type=c_tree_node::nt_root_node_local;
        }
        var_add_lookup(base,x);

#if 1==USE_LOOKUP_FOR_LOCALS
        lookup_locals.insert(make_pair(new_local_id,x));
#endif

        if(copy_from)
            array_copy(base,copy_from);
        current_local=local_id_save;
        in_construct=false;

        return new_local_id;
    } else {
        in_construct=true;

        if(!copy_from)
            base=MY_NEW c_var_array(0);
        else
            base=dynamic_cast<c_var_array*>(copy_from->copy(0));

        tree<c_tree_node, c_single_element_allocator<tree_node_<c_tree_node> > >::iterator x;
        {
            c_tree_node n;
            x=variables.append_child(get_current_local_node(),n);
        }
        (*x).root_node_id=0;
        (*x).value_id=base->get_id();
        (*x).clear_index();
        (*x).data2.second=base;
        (*x).node_type=c_tree_node::nt_root_node_core_entry;

        var_add_lookup(base,x);

#if 1==USE_LOOKUP_FOR_LOCALS
        lookup_locals.insert(make_pair(new_local_id,x));
#endif

        if(copy_from)
            array_copy(base,copy_from);
        in_construct=false;

        return 0;
    }

    // NOTREACH
}

unsigned c_var_keeper::set_current_local(unsigned id)
{
    unsigned res=current_local;
    current_local=id;
    return res;
}

unsigned c_var_keeper::get_current_local()
{
    return current_local;
}

void c_var_keeper::include_into_locals(c_var_base* var)
{
    if(!var)
        return; // huh?

    if(in_construct)
        return;

    tree<c_tree_node, c_single_element_allocator<tree_node_<c_tree_node> > >::iterator i1=get_current_local_node();
    in_construct=true;
    tree<c_tree_node, c_single_element_allocator<tree_node_<c_tree_node> > >::iterator x=variables.append_child(i1);
    in_construct=false;
    (*x).assign(var);
    var_add_lookup(var,x);
    return;
}

c_var_keeper::c_var_keeper()
{
    in_construct=true;

    var_keeper=this;

    m_in_work=false;

    next_id=1;
    set_current_session(SESSION_RESERVED_ID);
    last_local=0;
    current_local=0;

    {
        super_global=MY_NEW c_var_array(0);
        tree<c_tree_node, c_single_element_allocator<tree_node_<c_tree_node> > >::iterator x;
        {
            c_tree_node n;
            x=variables.insert(variables.begin(),n);
        }
        (*x).value_id=super_global->get_id();
        (*x).clear_index();
        (*x).data2.second=super_global;
        (*x).node_type=c_tree_node::nt_root_node_global;
        var_add_lookup(super_global,x);
    }

    {
        core_entry_vars=MY_NEW c_var_array(0);
        tree<c_tree_node, c_single_element_allocator<tree_node_<c_tree_node> > >::iterator x;
        {
            c_tree_node n;
            x=variables.insert(variables.begin(),n);
        }
        (*x).value_id=core_entry_vars->get_id();
        (*x).clear_index();
        (*x).data2.second=core_entry_vars;
        (*x).node_type=c_tree_node::nt_root_node_core_entry;
        var_add_lookup(core_entry_vars,x);
    }

    {
        temp_node=MY_NEW c_var_array(0);
        tree<c_tree_node, c_single_element_allocator<tree_node_<c_tree_node> > >::iterator x;
        {
            c_tree_node n;
            x=variables.insert(variables.begin(),n);
        }
        (*x).value_id=temp_node->get_id();
        (*x).clear_index();
        (*x).data2.second=temp_node;
        (*x).node_type=c_tree_node::nt_root_node_temp;
        lookup_temp_node=x;
        var_add_lookup(temp_node,x);
    }

    in_construct=false;
}

void c_var_keeper::reparent(c_var_array* base, c_var_base* var)
{
    bool got1=false;
    tree<c_tree_node, c_single_element_allocator<tree_node_<c_tree_node> > >::iterator i1;
    if(var_lookup(base,i1)) {
        got1=true;
        goto gotcha1;
    }
    for(i1=variables.begin(); i1!=variables.end(); ++i1) {
        if((*i1).data2.second==base) {
            got1=true;
            break;
        }
    }
gotcha1:
    bool got2=false;
    tree<c_tree_node, c_single_element_allocator<tree_node_<c_tree_node> > >::iterator i2;
    if(var_lookup(var,i2)) {
        got2=true;
        goto gotcha2;
    }
    for(i2=variables.begin(); i2!=variables.end(); ++i2) {
        if((*i2).data2.second==var) {
            got2=true;
            break;
        }
    }
gotcha2:

    if(got1 && got2) {
        tree<c_tree_node, c_single_element_allocator<tree_node_<c_tree_node> > >::iterator i3=variables.append_child(i1);
        i3=variables.move_ontop(i3,i2);
        var_add_tree_to_lookup(i3);
        return;
    }

    // huh?
#ifdef _DEBUG
    // DEBUG: set breakpoint here!
    int x=0;
    x++;
#endif
    MY_ASSERT_NOTREACH();
}

c_var_array* c_var_keeper::get_super_global()
{
    return super_global;
}

void c_var_keeper::destroy(c_var_base* var)
{
    if(!var)
        return; // huh?

    c_var_unset* dummy=NULL;
    var_id_t old=var->get_id();

    map<var_id_t, c_entry*>::iterator i2=storage.find(old);
    if(storage.end()!=i2) {
        if((*i2).second && (*i2).second->val)
            (*i2).second->val->reset_id();
        MY_DELETE((*i2).second);

        used_var_ids.clear((unsigned long)(*i2).first);

        storage.erase(i2);
    }

    map<var_id_t, var_id_t>::iterator i1;
    while((i1=references.find(old))!=references.end()) {
        var_id_t second=(*i1).second;
        references.erase(i1);
        dummy=MY_NEW c_var_unset;
        references.insert(make_pair(dummy->get_id(),second));
    }
    for(i1=references.begin(); i1!=references.end(); ++i1) {
        if((*i1).second==old) {
            dummy=MY_NEW c_var_unset;
            (*i1).second=dummy->get_id();
        }
    }

    if(dummy)
        add_temp_var(dummy);

    if(!var->get_in_deletion()) {
        tree<c_tree_node, c_single_element_allocator<tree_node_<c_tree_node> > >::iterator i1;
        if(var_lookup(var,i1)) {
            var_del_lookup((*i1).data2.second);
            var_destroy_lookup((*i1).data2.second);

            //
            // because this gets MY_DELETE'd in ~c_tree_node()
            //(*i1).data1.first=VAR_ID_INVALID;
            (*i1).clear_index();
            //

            (*i1).value_id=VAR_ID_INVALID;
            (*i1).data2.second=NULL;
            var_flatten_children(i1);
            i1=variables.erase(i1);
        } else {
#ifdef _DEBUG
            // DEBUG: set breakpoint here!
            int x=0;
            x++;
#endif
        }
    }

    if(!var->get_in_deletion()) {
        CHECK(__LINE__,"#15");
    }
}

void c_var_keeper::add_temp_var(c_var_base* var)
{
    if(!var)
        return; // huh?

    reparent(temp_node,var);
}

tree<c_tree_node, c_single_element_allocator<tree_node_<c_tree_node> > >::iterator c_var_keeper::add_local_var_no_reparent(tree<c_tree_node, c_single_element_allocator<tree_node_<c_tree_node> > >::iterator node, c_var_base* var)
{
    if(!var)
        return variables.end(); // huh?

    {
        node=variables.append_child(node);
    }
    (*node).assign(var);

    var_add_lookup(var,node);

    return node;
}

c_var_array* c_var_keeper::register_mod_global(c_var_array* copy_from, const char* file, const char* func, int line)
{
    in_construct=true;

    unsigned new_local_id=get_unique_node_id(file,func,line);

    c_var_array* base;
    tree<c_tree_node, c_single_element_allocator<tree_node_<c_tree_node> > >::iterator x;
    {
        if(!copy_from)
            base=MY_NEW c_var_array(0);
        else
            base=dynamic_cast<c_var_array*>(copy_from->copy(0));

        {
            c_tree_node n;
            x=variables.insert(variables.begin(),n);
        }
        (*x).value_id=base->get_id();
        (*x).clear_index();
        (*x).data2.second=base;
        (*x).node_type=c_tree_node::nt_root_node_mod_global;
        (*x).root_node_id=new_local_id;
    }
    var_add_lookup(base,x);

#if 1==USE_LOOKUP_FOR_LOCALS
    lookup_locals.insert(make_pair(new_local_id,x));
#endif

    if(copy_from)
        array_copy(base,copy_from);
    in_construct=false;

    return base;
}

void c_tree_node::assign(c_index_base* key, c_var_base* val)
{
    value_id= val ? val->get_id() : VAR_ID_INVALID;
    if(data2.first && data2.first!=key)
        MY_DELETE(data2.first);
    data2.first=key;
    data2.second=val;
    if(SESSION_CORE_ENTRY_ID==var_keeper->current_session)
        node_type=nt_root_node_core_entry;
}

void c_tree_node::assign(c_var_base* val)
{
    value_id= val ? val->get_id() : VAR_ID_INVALID;
    if(data2.first)
        MY_DELETE(data2.first);
    data2.first=NULL;
    data2.second=val;
    if(SESSION_CORE_ENTRY_ID==var_keeper->current_session)
        node_type=nt_root_node_core_entry;
}

void c_tree_node::set_index(c_index_base* key)
{
    assign(key,data2.second);
}

void c_tree_node::clear_index()
{
    if(data2.first)
        MY_DELETE(data2.first);
    data2.first=NULL;
}

void c_var_keeper::replace_ref(c_var_base* old_val, c_var_base* new_val)
{
    if(!old_val)
        return;

    var_id_t old=old_val->get_id();
    replace_ref(old,new_val);

    tree<c_tree_node, c_single_element_allocator<tree_node_<c_tree_node> > >::iterator i1;
    if(var_lookup(old_val,i1))
        goto gotcha1;
    for(i1=variables.begin(); i1!=variables.end(); ++i1) {
        if((*i1).data2.second==old_val) {
gotcha1:
            c_index_base* idx=(*i1).data2.first;
            if(!idx)
                break;
            i1=variables.parent(i1);
            if(i1!=variables.end()) {
                c_var_array* a=dynamic_cast<c_var_array*>((*i1).data2.second);
                if(a)
                    a->update_lookup(idx,new_val);
            }
            break;
        }
    }
}

void c_var_keeper::replace_ref(var_id_t old_id, c_var_base* new_val)
{
    map<var_id_t, var_id_t>::iterator i1;
    while((i1=references.find(old_id))!=references.end()) {
        var_id_t second=(*i1).second;
        references.erase(i1);
        references.insert(make_pair(new_val->get_id(),second));
    }
//again:
    for(i1=references.begin(); i1!=references.end(); ++i1) {
        if((*i1).second==old_id) {
            if((*i1).first!=new_val->get_id()) {
                c_entry* e=get_e((*i1).second,NULL,NULL);
                (*i1).second=new_val->get_id();
                if(e && e->val && typeid(*e->val)==typeid(c_var_ref)) {
                    c_var_ref* r=dynamic_cast<c_var_ref*>(e->val);
                    if(r)
                        r->reference_to_=new_val->get_id();
                }
            } else {
            }
        }
    }
}

var_id_t c_var_keeper::alloc(c_var_base* value)
{
    var_id_t new_id;
    if(value && VAR_ID_INVALID!=value->id) {
        new_id=value->id;

        // TODO: OPTIMIZATION: Should we perform next statement?!
        used_var_ids.set((unsigned long)new_id);
    } else {
        new_id=(var_id_t)used_var_ids.find_first_zero_bit();
        if(!new_id) {
            used_var_ids.set(0); // we must ignore bit #0, because ID==0 has special meaning: INVALID
            new_id=(var_id_t)used_var_ids.find_first_zero_bit();
        }
        if(value)
            value->id=new_id;

        used_var_ids.set((unsigned long)new_id);
    }

    storage.insert(make_pair(new_id,MY_NEW c_entry(value)));
    return new_id;
}

tree<c_tree_node, c_single_element_allocator<tree_node_<c_tree_node> > >::iterator c_var_keeper::get_current_local_node()
{
    tree<c_tree_node, c_single_element_allocator<tree_node_<c_tree_node> > >::iterator i1;

    if(SESSION_RESERVED_ID==current_session || SESSION_WORK_ID==current_session || SESSION_INTERNAL_ID==current_session) {
        // super globals
        return lookup_temp_node;
    } else if(SESSION_CORE_ENTRY_ID==current_session) {
        tree<c_tree_node, c_single_element_allocator<tree_node_<c_tree_node> > >::iterator i1;
        for(i1=variables.begin(); i1!=variables.end(); i1.skip_children(), ++i1) {
            if(c_tree_node::nt_root_node_core_entry==(*i1).node_type) {
                return i1;
            }
        }
    } else {
#if 1==USE_LOOKUP_FOR_LOCALS
        map<unsigned,tree<c_tree_node, c_single_element_allocator<tree_node_<c_tree_node> > >::iterator>::iterator ii2=lookup_locals.find(current_local);
        if(lookup_locals.end()!=ii2) {
            return (*ii2).second;
        }
#endif
        tree<c_tree_node, c_single_element_allocator<tree_node_<c_tree_node> > >::iterator i1;
        for(i1=variables.begin(); i1!=variables.end(); i1.skip_children(), ++i1) {
            if(c_tree_node::nt_root_node_local==(*i1).node_type && (*i1).root_node_id==current_local) {
                return i1;
            }
        }
    }

    if(0==current_local) {
        return lookup_temp_node;
    }

    // huh?
#ifdef _DEBUG
    // DEBUG: Set breakpoint here!
    int x=0;
    x++;
#endif
    MY_ASSERT_NOTREACH();
    return variables.end();
}

void c_var_keeper::free_local(unsigned id, bool delayed)
{
    if(!delayed || 0==id) {
        tree<c_tree_node, c_single_element_allocator<tree_node_<c_tree_node> > >::iterator i1;
#if 1==USE_LOOKUP_FOR_LOCALS
        map<unsigned,tree<c_tree_node, c_single_element_allocator<tree_node_<c_tree_node> > >::iterator>::iterator ii2=lookup_locals.find(id);
        if(lookup_locals.end()!=ii2) {
            i1=(*ii2).second;
            lookup_locals.erase(ii2);
            goto gotcha1;
        }
#endif
        for(i1=variables.begin(); i1!=variables.end(); i1.skip_children(), ++i1) {
            if((c_tree_node::nt_root_node_local==(*i1).node_type || c_tree_node::nt_root_node_mod_global==(*i1).node_type) && (*i1).root_node_id==id) {
#if 1==USE_LOOKUP_FOR_LOCALS
        gotcha1:
#endif
                //variables.erase_children(i1);
                var_del_lookup((*i1).data2.second);
                var_destroy_lookup((*i1).data2.second);
                var_flatten_children(i1);
                variables.erase(i1);

                for(vector<unsigned>::iterator i1=locals.begin(); i1!=locals.end(); ++i1) {
                    if(*i1==id) {
                        locals.erase(i1);
                        break;
                    }
                }
#if DEBUG_VAR_KEEPER_DEBUG_STRING_DISABLE==0
                map<unsigned,c_locals_debug_info>::iterator ii=locals_debug.find(id);
                if(locals_debug.end()!=ii)
                    locals_debug.erase(ii);
#endif

                return;
            }
        }
    } else {
        delayed_ids.push_back(id);
        return;
    }

    // huh?
#ifdef _DEBUG
    // DEBUG: Set breakpoint here!
    int x=0;
    x++;
#endif
}

void c_var_keeper::free_global(c_var_array* vars, bool delayed)
{
    if(NULL==vars)
        return;

    tree<c_tree_node, c_single_element_allocator<tree_node_<c_tree_node> > >::iterator i1;
    for(i1=variables.begin(); i1!=variables.end(); i1.skip_children(), ++i1) {
        if((*i1).data2.second==vars) {
            free_local((*i1).root_node_id,delayed);
            return;
        }
    }
}

void c_var_keeper::var_add_lookup(c_var_base* UNUSEDARG(second), tree<c_tree_node, c_single_element_allocator<tree_node_<c_tree_node> > >::iterator& iter)
{
    var_add_tree_to_lookup(iter);
}

void c_var_keeper::var_del_lookup(c_var_base* second, bool simple)
{
    if(!second)
        return;
    map<c_var_base*,tree<c_tree_node, c_single_element_allocator<tree_node_<c_tree_node> > >::iterator*>::iterator i1=lookup1.find(second);
    if(lookup1.end()!=i1) {
        if(simple) {
            MY_DELETE((*i1).second);
            lookup1.erase(i1);
            return;
        }
    } else
        return;
    tree<c_tree_node, c_single_element_allocator<tree_node_<c_tree_node> > >::iterator start=*(*i1).second;

    bool first=true;
    int depth=variables.depth(start)+1;
    for(;; ++start) {
        if(first) {
            first=false;
        } else {
            if(variables.depth(start)<depth)
                break;
        }
        if((*start).data2.second) {
            map<c_var_base*,tree<c_tree_node, c_single_element_allocator<tree_node_<c_tree_node> > >::iterator*>::iterator i1=lookup1.find((*start).data2.second);
            if(lookup1.end()!=i1) {
                MY_DELETE((*i1).second);
                lookup1.erase(i1);
            }
        }
    }
}

void c_var_keeper::var_destroy_lookup(c_var_base* UNUSEDARG(second), bool UNUSEDARG(simple))
{
}

bool c_var_keeper::var_lookup(c_var_base* second, tree<c_tree_node, c_single_element_allocator<tree_node_<c_tree_node> > >::iterator& iter)
{
    map<c_var_base*,tree<c_tree_node, c_single_element_allocator<tree_node_<c_tree_node> > >::iterator*>::iterator i1=lookup1.find(second);
    if(lookup1.end()!=i1) {
        iter=*(*i1).second;
        return true;
    }

    char d[256];
    sprintf(d,"Lookup cache miss for variable pointer %p",(void*)second);
    c_logger* cl=get_core_logger();
    if(cl)
        cl->log_print1(E_LOG_DEBUG,LOG_SCOPE_GLOBAL,"",GET_STR_DEBUG_INFO_EX(d),true);
    MY_RPT0(MY_RPT_INFO,d);
#ifdef _DEBUG
    // DEBUG: set breakpoint here
    int x=0;
    x++;
#endif

    return false;
}

void c_var_keeper::var_add_tree_to_lookup(tree<c_tree_node, c_single_element_allocator<tree_node_<c_tree_node> > >::iterator start)
{
    start.skip_children(false);
    bool first=true;
    int depth=variables.depth(start)+1;
    for(;; ++start) {
        if(first) {
            first=false;
        } else {
            if(variables.depth(start)<depth)
                break;
        }
        if(!(*start).data2.second)
            continue;
        map<c_var_base*,tree<c_tree_node, c_single_element_allocator<tree_node_<c_tree_node> > >::iterator*>::iterator i1=lookup1.find((*start).data2.second);
        if(lookup1.end()!=i1) {
            MY_DELETE((*i1).second);
            lookup1.erase(i1);
        }
        tree<c_tree_node, c_single_element_allocator<tree_node_<c_tree_node> > >::iterator* iter=MY_NEW tree<c_tree_node, c_single_element_allocator<tree_node_<c_tree_node> > >::iterator(start);
        *(iter)=start;
        lookup1.insert(make_pair((*start).data2.second,iter));
    }
}

void c_var_keeper::_check(int ln, const string& name)
{
    vector<var_id_t> vars;
    vector<c_index_base*> ptrs;
    tree<c_tree_node, c_single_element_allocator<tree_node_<c_tree_node> > >::iterator i1;
    for(i1=variables.begin(); i1!=variables.end(); ++i1) {
        if((*i1).data2.first)
            ptrs.push_back((*i1).data2.first);

        {
            if( ((*i1).data2.first && (*i1).data2.first->type()==c_index_base::IDX_NULL && (*i1).data2.second) ||
                ((*i1).value_id!=VAR_ID_INVALID && !(*i1).data2.second) ||
                (VAR_ID_INVALID==(*i1).value_id && (*i1).data2.second) )
            {
                DUMP(ln,name);
#ifdef _DEBUG
                // DEBUG: set breakpoint here!
                int x=0;
                x++;
#endif
                break;
            }
            if((*i1).value_id)
                vars.push_back((*i1).value_id);
        }
    }
    size_t cnt=vars.size();

    sort(vars.begin(),vars.end());
    vector<var_id_t>::iterator i3=unique(vars.begin(),vars.end());
    vars.erase(i3,vars.end());

    if(cnt!=vars.size()) {
        // duplicate entries found

        DUMP(ln,name);

        FILE* f=fopen(get_dump_fn(NULL,ln,name).c_str(),"a");

        vars.clear();
        for(i1=variables.begin(); i1!=variables.end(); ++i1) {
            if(!(*i1).value_id)
                continue;
            vector<var_id_t>::iterator i2;
            for(i2=vars.begin(); i2!=vars.end(); ++i2) {
                if(*i2==(*i1).value_id) {
                    if(f) {
                        fprintf(f,"Duplicate variable ID #%s\n",tos((*i1).value_id).c_str());
                        break;
                    }
                }
            }
            if((*i1).value_id)
                vars.push_back((*i1).value_id);
        }

#ifdef _DEBUG
        // DEBUG: set breakpoint here!
        int x=0;
        x++;
#endif
    }

    /////

    cnt=ptrs.size();

    sort(ptrs.begin(),ptrs.end());
    vector<c_index_base*>::iterator i4=unique(ptrs.begin(),ptrs.end());
    ptrs.erase(i4,ptrs.end());

    if(cnt!=ptrs.size()) {
        // duplicate entries found

        DUMP(ln,name);

        FILE* f=fopen(get_dump_fn(NULL,ln,name).c_str(),"a");
        fprintf(f,"\n============ INDEXES ==============\n\n");

        ptrs.clear();
        for(i1=variables.begin(); i1!=variables.end(); ++i1) {
            if(!(*i1).data2.first)
                continue;
            vector<c_index_base*>::iterator i2;
            for(i2=ptrs.begin(); i2!=ptrs.end(); ++i2) {
                if(*i2==(*i1).data2.first) {
                    if(f) {
                        if((*i1).data2.first->type()==c_index_base::IDX_STR)
                            fprintf(f,"Duplicate index: %s\n",dynamic_cast<c_index_str*>(*i2)->get().c_str());
                        if((*i1).data2.first->type()==c_index_base::IDX_INT)
                            fprintf(f,"Duplicate index: %s\n",tos(dynamic_cast<c_index_int*>(*i2)->get()).c_str());
                        break;
                    }
                }
            }
            if((*i1).data2.first)
                ptrs.push_back((*i1).data2.first);
        }

        if(f)
            fclose(f);
        f=NULL;

#ifdef _DEBUG
        // DEBUG: set breakpoint here!
        int x=0;
        x++;
#endif
    }
}

void c_var_keeper::remove_from_tree(c_var_base* var)
{
    tree<c_tree_node, c_single_element_allocator<tree_node_<c_tree_node> > >::iterator i1;
    if(var_lookup(var,i1)) {
        goto gotcha1;
    }
    for(i1=variables.begin(); i1!=variables.end(); ++i1) {
        if((*i1).data2.second==var) {
            break;
        }
    }
    if((*i1).data2.second!=var)
        return; // huh?
gotcha1:

    var_del_lookup((*i1).data2.second);
    var_destroy_lookup((*i1).data2.second);

    (*i1).value_id=VAR_ID_INVALID;
    (*i1).data2.second=NULL;
    // destruct key (first)
    //c_var_base* tmp=(*i1).data2.first;
    //(*i1).data1.first=VAR_ID_INVALID;
    (*i1).clear_index();
    //MY_DELETE(tmp);
    variables.erase(i1);
}

void c_var_keeper::array_copy(c_var_array* target, c_var_array* source)
{
    CHECK(__LINE__,"#16-array-copy-before");

    //tree<c_tree_node, c_single_element_allocator<tree_node_<c_tree_node> > >::iterator tmp=variables.append_child(variables.begin());

    tree<c_tree_node, c_single_element_allocator<tree_node_<c_tree_node> > >::iterator i1;
    if(var_lookup(target,i1)) {
        goto gotcha1;
    }
    for(i1=variables.begin(); i1!=variables.end(); ++i1) {
        if((*i1).data2.second==target) {
            break;
        }
    }
    if((*i1).data2.second!=target) {
        //variables.erase(tmp);
        return; // huh?
    }
gotcha1:
    tree<c_tree_node, c_single_element_allocator<tree_node_<c_tree_node> > >::iterator i2;
    if(var_lookup(source,i2)) {
        goto gotcha2;
    }
    for(i2=variables.begin(); i2!=variables.end(); ++i2) {
        if((*i2).data2.second==source) {
            break;
        }
    }
    if((*i2).data2.second!=source) {
        //variables.erase(tmp);
        return; // huh?
    }
gotcha2:

    stack<tree<c_tree_node, c_single_element_allocator<tree_node_<c_tree_node> > >::iterator> st;
    tree<c_tree_node, c_single_element_allocator<tree_node_<c_tree_node> > >::iterator x;
    tree<c_tree_node, c_single_element_allocator<tree_node_<c_tree_node> > >::iterator res=i1;

    tree<c_tree_node, c_single_element_allocator<tree_node_<c_tree_node> > >::iterator dst=i1;
    {
        stack<c_var_array*> curr_target;
        curr_target.push(target);

        int depth=variables.depth(i2)+1;
        int last=depth;
        ++i2;
        for(; i2!=variables.end(); ++i2) {
            if(variables.depth(i2)<depth)
                break;
            if(variables.depth(i2)>last) {
                st.push(i1);
                i1=x;
                ++last;
            } else while(variables.depth(i2)<last) {
                i1=st.top();
                st.pop();
                --last;
                curr_target.pop();
            }

            x=variables.append_child(i1);

            if((*i2).value_id!=VAR_ID_INVALID && (*i2).data2.second) {
                if(typeid(*(*i2).data2.second)==typeid(c_var_array)) {
                    (*x).data2.second=MY_NEW c_var_array(0);
                    dynamic_cast<c_var_array*>((*x).data2.second)->new_int_index=
                        dynamic_cast<c_var_array*>((*i2).data2.second)->new_int_index;
                    (*x).value_id=(*x).data2.second->get_id();
                } else if(typeid(*(*i2).data2.second)==typeid(c_var_ref)) {
                    // not implemented!
                    (*x).data2.second=NULL;
                    (*x).value_id=VAR_ID_INVALID;
#ifdef _DEBUG
                    // DEBUG: set breakpoint here!
                    int xx=0;
                    xx++;
#endif
                } else {
                    (*x).data2.second=(*i2).data2.second->copy(0);
                    (*x).value_id=(*x).data2.second->get_id();
                }
                if(!(*x).data2.first && (*i2).data2.first) {
                    (*x).set_index((*i2).data2.first->copy());
                }
                curr_target.top()->update_lookup((*x).data2.first,(*x).data2.second);
                if(typeid(*(*i2).data2.second)==typeid(c_var_array)) {
                    curr_target.push(static_cast<c_var_array*>((*x).data2.second));
                }
            } else {
                (*x).data2.second=NULL;
                (*x).value_id=VAR_ID_INVALID;
            }
        }
    }

    var_add_tree_to_lookup(res);

    CHECK(__LINE__,"#17-array-copy-after");
}

tree<c_tree_node, c_single_element_allocator<tree_node_<c_tree_node> > >::iterator c_var_keeper::array_replace_tree(tree<c_tree_node, c_single_element_allocator<tree_node_<c_tree_node> > >::iterator dest, tree<c_tree_node, c_single_element_allocator<tree_node_<c_tree_node> > >::iterator src, c_index_base* key_save)
{
    if(src==dest)
        return dest;

    c_var_base* old=(*dest).data2.second;
    if((*dest).data2.first!=key_save) {
        (*dest).clear_index();
    } else {
        (*dest).data2.first=NULL;
    }
    move_to_temp(dest);

    {
        int depth=variables.depth(dest)+1;
        tree<c_tree_node, c_single_element_allocator<tree_node_<c_tree_node> > >::iterator i1=dest;

        // to get rid of accidental already-set skip_children flag:
        i1.skip_children(false);
        // ...done

        for(; i1!=variables.end(); ++i1) {
            if(variables.depth(i1)<depth && i1!=dest)
                break;
            // !!! We don't want to destroy old variables, since they may be still referenced
            (*i1).assign(NULL);
        }
    }

    // bug workaround
    tree<c_tree_node, c_single_element_allocator<tree_node_<c_tree_node> > >::iterator tmp1, tmp2;
    bool need_workaround=false;
    {
        tree<c_tree_node, c_single_element_allocator<tree_node_<c_tree_node> > >::iterator i1;
        i1=dest;
        ++i1;
        if(i1==src)
            need_workaround=true;
        if(!need_workaround) {
            i1=dest;
            --i1;
            if(i1==src)
                need_workaround=true;
        }
        if(!need_workaround) {
            i1=src;
            ++i1;
            if(i1==dest)
                need_workaround=true;
        }
        if(!need_workaround) {
            i1=src;
            --i1;
            if(i1==dest)
                need_workaround=true;
        }

        if(need_workaround) {
            tmp1=variables.append_child(variables.begin());
            tmp2=variables.append_child(variables.begin());

            reparenting_nodes=true;
            tmp2=variables.move_ontop(tmp2,src);
            reparenting_nodes=false;
            src=tmp2;
        }
    }

    reparenting_nodes=true;
    tree<c_tree_node, c_single_element_allocator<tree_node_<c_tree_node> > >::iterator x=variables.move_ontop(dest,src);
    reparenting_nodes=false;
    if(key_save && key_save->type()!=c_index_base::IDX_NULL)
        (*x).set_index(key_save);

    if(need_workaround)
        variables.erase(tmp1);

    var_add_tree_to_lookup(x);
    if(old)
        replace_ref(old,(*x).data2.second);

    return x;
}

void c_var_keeper::clear_subtree(tree<c_tree_node, c_single_element_allocator<tree_node_<c_tree_node> > >::iterator root)
{
    int depth=variables.depth(root)+1;
    tree<c_tree_node, c_single_element_allocator<tree_node_<c_tree_node> > >::iterator i1=root;

    // to get rid of accidental already-set skip_children flag:
    i1.skip_children(false);
    // ...done

    for(; i1!=variables.end(); ++i1) {
        if(variables.depth(i1)<depth && i1!=root)
            return;
        (*i1).assign(NULL);
    }
}

c_var_keeper::~c_var_keeper()
{
    // this is a MUST! because in other case we could get SEG FAULT,
    //     if there are some variables: on exit of program, after exit
    //     from this destructor, there all calls to this class' member
    //     functions, if "free_all()" have not been called!
    free_all();
}

size_t c_var_keeper::do_garbage(int max_time)
{
    long start=get_wall_clock();

    size_t cnt=0;

    // very important, because we can get SEGFAULT if tree is empty, because
    //     of invalidity of 'lookup_temp_node' (on program exit)
    if(variables.empty())
        return cnt;

    // temp first
    {
        tree<c_tree_node, c_single_element_allocator<tree_node_<c_tree_node> > >::iterator i2=lookup_temp_node;
        for(;;) {
            {
                // garbage it!

                var_flatten_children(i2);

                tree<c_tree_node, c_single_element_allocator<tree_node_<c_tree_node> > >::iterator i3;
                int depth=variables.depth(i2)+1;
                for(i3=i2, ++i3; i3!=variables.end(); ++i3) {
                    if(variables.depth(i3)<depth)
                        break;

                    MY_DELETE((*i3).data2.first);
                    (*i3).data2.first=NULL;
                    (*i3).value_id=0;

                    c_var_base* tmp;

                    if((*i3).data2.second) {
                        clear_reference((*i3).data2.second);
                        var_del_lookup((*i3).data2.second,true);
                        var_destroy_lookup((*i3).data2.second,true);

                        ++cnt; // some was found in temp

                        tmp=(*i3).data2.second;
                        (*i3).value_id=VAR_ID_INVALID;
                        (*i3).data2.second=NULL;
                        tmp->set_in_deletion();
                        if(tmp && tmp->get_id())
                            dealloc(tmp->get_id(),NULL,NULL,true);
                        else
                            MY_DELETE(tmp);
                    }

                    if(cnt%10==0 && max_time>=0) {
                        // check for time
                        long now=get_wall_clock();
                        if(now-start>=max_time || now-start<0)
                            return 1; // temp node is only one
                    }

                }
                variables.erase_children(i2);
                break;
            }
        }
    }

    if(cnt!=0)
        cnt=1; // temp node is only one

    if(max_time>=0) {
        // check for time
        long now=get_wall_clock();
        if(now-start>=max_time || now-start<0)
            return cnt+1;
    }

    vector<unsigned>::iterator i1;
    while((i1=delayed_ids.begin())!=delayed_ids.end()) {
        tree<c_tree_node, c_single_element_allocator<tree_node_<c_tree_node> > >::iterator i2;
#if 1==USE_LOOKUP_FOR_LOCALS
        map<unsigned,tree<c_tree_node, c_single_element_allocator<tree_node_<c_tree_node> > >::iterator>::iterator ii2=lookup_locals.find(*i1);
        if(lookup_locals.end()!=ii2) {
            i2=(*ii2).second;
            lookup_locals.erase(ii2);
            goto gotcha1;
        }
#endif
        for(i2=variables.begin(); i2!=variables.end(); i2.skip_children(), ++i2) {
            if((c_tree_node::nt_root_node_local==(*i2).node_type || c_tree_node::nt_root_node_global==(*i2).node_type || c_tree_node::nt_root_node_mod_global==(*i2).node_type) && *i1==(*i2).root_node_id) {
        //gotcha1:
                // garbage it!

                var_flatten_children(i2);

                tree<c_tree_node, c_single_element_allocator<tree_node_<c_tree_node> > >::iterator i3;
                int depth=variables.depth(i2)+1;
                for(i3=i2, ++i3; i3!=variables.end(); ++i3) {
                    if(variables.depth(i3)<depth)
                        break;

                    MY_DELETE((*i3).data2.first);
                    (*i3).data2.first=NULL;
                    (*i3).value_id=0;

                    c_var_base* tmp;

                    if((*i3).data2.second) {
                        clear_reference((*i3).data2.second);
                        var_del_lookup((*i3).data2.second,true);
                        var_destroy_lookup((*i3).data2.second,true);

                        tmp=(*i3).data2.second;
                        (*i3).value_id=VAR_ID_INVALID;
                        (*i3).data2.second=NULL;
                        tmp->set_in_deletion();
                        if(tmp && tmp->get_id())
                            dealloc(tmp->get_id(),NULL,NULL,true);
                        else
                            MY_DELETE(tmp);
                    }
                }
                variables.erase_children(i2);

                // root array as well
                // note: root node should not have index/key set, so don't care about it
                c_var_base* tmp=(*i2).data2.second;
                var_del_lookup((*i2).data2.second,true);
                var_destroy_lookup((*i2).data2.second,true);
                (*i2).value_id=VAR_ID_INVALID;
                (*i2).data2.second=NULL;
                tmp->set_in_deletion();
                if(tmp && tmp->get_id())
                    dealloc(tmp->get_id(),NULL,NULL,true);
                else {
                    MY_DELETE(tmp);
                    tmp=NULL;
                }

                variables.erase(i2);

                break;
            }
        }
        for(vector<unsigned>::iterator i3=locals.begin(); i3!=locals.end(); ++i3) {
#if DEBUG_VAR_KEEPER_DEBUG_STRING_DISABLE==0
            map<unsigned,c_locals_debug_info>::iterator ii=locals_debug.find(*i1);
            if(locals_debug.end()!=ii)
                locals_debug.erase(ii);
#endif
            if(*i3==*i1) {
                locals.erase(i3);
                break;
            }
        }
        delayed_ids.erase(i1);

        ++cnt;

        if(max_time>=0) {
            // check for time
            long now=get_wall_clock();
            if(now-start>=max_time || now-start<0)
                break;
        }
    }

    return cnt;
}

void c_var_keeper::var_flatten_children(tree<c_tree_node, c_single_element_allocator<tree_node_<c_tree_node> > >::iterator parent)
{
    int depth=variables.depth(parent)+1;

    ++parent;

    while(parent!=variables.end()) {
        if(variables.depth(parent)<depth)
            break;
        variables.flatten(parent++);
    }
}

void c_var_keeper::move_to_temp(tree<c_tree_node, c_single_element_allocator<tree_node_<c_tree_node> > >::iterator& what)
{
    tree<c_tree_node, c_single_element_allocator<tree_node_<c_tree_node> > >::iterator x=add_local_var_no_reparent(get_current_local_node(),(*what).data2.second);

    if(x!=variables.end()) {
        variables.reparent(x,what);
    }
    (*what).assign(NULL);
    var_add_tree_to_lookup(x);
}

c_temp_local::c_temp_local(c_var_array* copy_from)
{
    saved_local_id=var_keeper->get_current_local();
    local_vars=NULL;
    local_id=var_keeper->register_local(local_vars,copy_from,__FILE__,__my_func__,__LINE__);
    var_keeper->set_current_local(local_id);
}

c_temp_local::~c_temp_local()
{
    var_keeper->free_local(local_id,true);
    var_keeper->set_current_local(saved_local_id);
}

unsigned c_temp_local::get_params(c_var_array* &vars)
{
    vars=local_vars;
    return local_id;
}

void c_var_keeper::clear_reference(c_var_base* var)
{
    if(!var)
        return; // huh?!

    var_id_t id=var->get_id();

    map<var_id_t, var_id_t>::iterator i1;
again:
    for(i1=references.begin(); i1!=references.end(); ++i1) {
        if((*i1).first==id || (*i1).second==id) {
            references.erase(i1);
            goto again;
        }
    }
}

/*!
    \brief Gets debugging information from root node ID
    \version 1
    \author VooDooMan
    \date 2007-05-03
    \param local_id Root note ID
    \param allocation Returns time of allocation
*/
string c_var_keeper::get_debug_string(unsigned local_id, my_time_t& allocation)
{
#if DEBUG_VAR_KEEPER_DEBUG_STRING_DISABLE==0
    map<unsigned,c_locals_debug_info>::iterator ii=locals_debug.find(local_id);
    if(locals_debug.end()!=ii) {
        allocation=(*ii).second.m_time;
        string res=((string)(*ii).second.m_file)+"::"+(*ii).second.m_func+"():"+tos((*ii).second.m_line);
        return res;
    }
    allocation=0;
    return (string)"<internal error - not found, PLEASE, REPORT THIS PROBLEM!>";
#else
    allocation=0;
    return (string)"<feature disabled>";
#endif

    // NOTREACH
}

size_t c_var_keeper::get_size()
{
    return static_cast<size_t>(variables.size());
}

bool c_var_keeper::array_dump_to_string_vector_wo_unset_and_array(c_module* this_module, c_var_array* base, vector<pair<const c_index_base*,string> >& result)
{
    tree<c_tree_node, c_single_element_allocator<tree_node_<c_tree_node> > >::iterator i1;
    if(var_lookup(base,i1))
        goto gotcha1;
    for(i1=variables.begin(); i1!=variables.end(); ++i1) {
        if((*i1).data2.second==base) {
gotcha1:
            int depth=variables.depth(i1)+1;
            ++i1;
            for(; i1!=variables.end(); ++i1) {
                if(variables.depth(i1)<depth)
                    break;

                // process any reference:
                c_var_base* val=get((*i1).value_id,var_arr_log_fn,this_module);

                if(val && typeid(*val)==typeid(c_var_array))
                    return false;

                if(typeid(*val)!=typeid(c_var_unset))
                    result.push_back(make_pair((*i1).data2.first,val->to_string(this_module)));
                else
                    result.push_back(make_pair((*i1).data2.first,""));
            }
            break;
        }
    }

    return true;
}

void c_var_keeper::set_core_module_working_session(session_id_t session)
{
    m_core_module_working_session=session;
}

