/***************************************************************************
                          var_keeper.h  -  description
                             -------------------
    begin                : Sat May 13 2006
    copyright            : (C) 2006-2007 by Marian VooDooMan Meravy
    email                : ghostvoodooman -aT= users (dot) sourceforge.net
 ***************************************************************************/

/* $Id: var_keeper.h,v 1.36 2008/06/08 08:43:13 ghostvoodooman Exp $ */

/***************************************************************************

VooDoo cIRCle - an IRC (ro)bot
Copyright (C) 2006-2008 by Marian VooDooMan Meravy
  (ghostvoodooman -aT= users (dot) sourceforge.net)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************/

/*!
    \file
    \brief Provides variable keeper
    \version $Id: var_keeper.h,v 1.36 2008/06/08 08:43:13 ghostvoodooman Exp $
*/

//---------------------------------------------------------------------------

#ifndef VAR_KEEPER_H__
#define VAR_KEEPER_H__

#if defined(HAVE_CONFIG_H)
#   include <config.h>
#endif

#ifndef _MSC_VER
#   include "utils/assert.h"
#   include "tree/tree.hh"
#else
#   include "utils/assert.h"
#   include "tree/tree.hh"
#endif

#include <string>
#include <vector>
#include <stack>
#include <list>
#include <map>

using namespace std;

#include "utils.h"      // because we want to enable mem leak detection
#include "params.h"

#include "variable.h"
#include "utils/assert.h"
#include "bitarray.h"
#include "internal_state.h"

#include "memory/c_single_element_allocator.hh"

/*!
    \brief Variable entry
    \version 2
    \author VooDooMan
    \date 2006-03-19, 2006-04-18
*/
class c_entry
{
public:
    c_var_base* val;                    //!< Variable value

    /*!
        \brief Constructor
        \version 1
        \author VooDooMan
        \date 2006-03-19
        \param value Initial value to set
    */
    c_entry(c_var_base* value = NULL)
    {
        val=value;
    }
};

/*!
    \author VooDooMan
    \version 1
    \date 2006-10-24
    \brief Node of tree
*/
class c_tree_node
{
    enum e_node_type
    {
        nt_normal = 0,                      //!< Ordinary variable
        nt_root_node_local,                 //!< Root node for local variables
        nt_root_node_global,                //!< Root node for super global variables
        nt_root_node_temp,                  //!< Root node for temporary variables (when there is some reference to variable, but variable itsef has been destroyed, a placeholder is created under this node, until reference is freed)
        nt_root_node_mod_global,            //!< Root node for module's global variables
        nt_root_node_core_entry,            //!< "core" module's "_sys_startup" entrypoint node
    };                                      //!< Type of node

    e_node_type node_type;                  //!< Type of this node

    unsigned root_node_id;                  //!< ID, valid only for non-'nt_root_node_temp'

    friend class c_var_keeper;

public:
    var_id_t value_id;                      //!< Data
    pair<c_index_base*,c_var_base*> data2;  //!< Data

    c_tree_node();
    c_tree_node(c_index_base* key, c_var_base* val);
    c_tree_node(const c_tree_node& orig);
    c_tree_node& operator=(const c_tree_node& orig);
    ~c_tree_node();
    void assign(const c_tree_node& orig);
    void set_index(c_index_base* key);

private:
    void assign(c_index_base* key, c_var_base* val);
    void assign(c_var_base* val);
    void clear_index();
};

/*!
    \author VooDooMan
    \version 2
    \date 2006-03-19
    \brief Maintains variable/memory management
*/
class c_var_keeper
{
public:
    enum e_log_severity
    {
        LS_WARNING = 1,
        LS_FATAL,
    };

    typedef void (*log_function_ptr_t)(e_log_severity, const string&, void*);

private:

    friend class c_binmodule;

    map<var_id_t, c_entry*> storage;        //!< List of variables that are in use
    //vector<var_id_t> freed;               //!< List of freed ID's
    c_bitarray used_var_ids;                //!< Binary array, bit value '1' means used ID, '0' not, bit offset is variable ID; bit offset 0 is invalid: not used

    //map<c_var_base*, var_id_t> index;     //!< For fast lookup, translation of variable pointer to ID

    var_id_t next_id;                       //!< Next ID to use

    session_id_t current_session;           //!< Session ID that will be assigned to new variables

    map<var_id_t, var_id_t> references;     //!< Fast look-up of references; first==from, second==to
    friend class c_var_ref;

    tree<c_tree_node, c_single_element_allocator<tree_node_<c_tree_node> > > variables;     //!< Storage of variables

    vector<unsigned> locals;                //!< List of local root nodes (IDs)
    unsigned last_local;                    //!< Last ID used, okay to overflow
    unsigned current_local;                 //!< Current local ID

    /*!
        \brief debugging structure
        \version 1
        \date 2008-03-14
        \author VooDooMan
    */
    struct c_locals_debug_info
    {
        const char* m_file;     //<! File name of allocation
        int m_line;             //<! Line number of allocation
        const char* m_func;     //<! Function name of allocation
        my_time_t m_time;       //<! Time of allocation

        /*!
            \brief Constructor
            \version 1
            \date 2008-03-14
            \author VooDooMan
            \param file C++ __FILE__
            \param line C++ __LINE__
            \param func C++ __my_func__
        */
        MY_FORCE_INLINE c_locals_debug_info(const char* file, int line, const char* func)
            : m_file(file)
            , m_line(line)
            , m_func(func)
            , m_time(my_time(NULL))
        { ; }
    };

    map<unsigned,c_locals_debug_info> locals_debug;     //!< Associated local IDs -> debug information

    string get_debug_string(unsigned local_id, my_time_t& allocation);

    c_var_array* super_global;              //!< Superglobal root variable
    c_var_array* temp_node;                 //!< Temporary dummies (when there is some reference to variable, but variable itsef has been destroyed, a placeholder is created under this node, until reference is freed)
    c_var_array* core_entry_vars;           //!< "core" module's "_sys_startup" local variables
    vector<unsigned> delayed_ids;           //!< Local IDs queue to free (garbage collection)
    map<unsigned,tree<c_tree_node, c_single_element_allocator<tree_node_<c_tree_node> > >::iterator> lookup_locals;    //!< Fast lookup table for local root nodes
    tree<c_tree_node, c_single_element_allocator<tree_node_<c_tree_node> > >::iterator lookup_temp_node;   //!< Iterator for temporary root node

    bool in_construct;                      //!< If we are inside of constructor, or other initialisation

    /*!
        \brief Dereferences (possibly) referenced variable
        \author VooDooMan
        \date 2006-05-13
        \version 1
        \param what Variable ID to dereference
        \return Returns dereferenced variable, or same ID if it is not referenced
    */
    var_id_t dereference(var_id_t what);

    friend class c_micro_symbol_variable_table;

    /*!
        \brief Get reference chain
        \author VooDooMan
        \date 2006-05-13
        \version 1
        \param what Variable ID to dereference
        \param chain Returns chain in vector ('what' is excluded from chain)
    */
    void dereference(var_id_t what, vector<var_id_t>& chain);

    /*!
        \brief Deallocates (MY_DELETE's) variable associated with ID
        \version 1
        \author VooDooMan
        \date 2006-03-19
        \param id ID of variable
        \param log_fn Logger functiuon call-back
        \param user_data Data passed to call-back function
        \param no_deref Do not try to dereference (if there is reference to another variable)
    */
    void dealloc(var_id_t id, log_function_ptr_t log_fn, void* user_data, bool no_deref = false);

    /*!
        \brief Gets descriptor of variable
        \version 1
        \author VooDooMan
        \date 2006-03-19
        \param id ID of variable
        \return Returns descriptor
        \param log_fn Logger functiuon call-back
        \param user_data Data passed to call-back function
    */
    c_entry* get_e(var_id_t id, log_function_ptr_t log_fn, void* user_data);

    /*!
        \brief Generates unique node ID
        \version 1
        \date 2006-11-27
        \author VooDooMan
    */
    unsigned get_unique_node_id(const char* file, const char* func, int line);

    /*!
        \brief Includes variable into local variable store
        \param var Variable
    */
    void include_into_locals(c_var_base* var);

    /*!
        \brief Moves variable to another node (to be child of another parent)
        \param base New parent
        \param var Variable to reparent
    */
    void reparent(c_var_array* base, c_var_base* var);

    /*!
        \brief Called when destructing a variable (updates accidental references)
        \param var Variable to destroy
    */
    void destroy(c_var_base* var);

    friend class c_var_base;
    friend class c_var_unset;
    friend class c_var_string;
    friend class c_var_array;
    friend class c_var_int;
    friend class c_var_bool;

    /*!
        \brief Adds variable to temporary nodes
        \param var Variable
    */
    void add_temp_var(c_var_base* var);

    /*!
        \brief Returns root node for current local ID
    */
    tree<c_tree_node, c_single_element_allocator<tree_node_<c_tree_node> > >::iterator get_current_local_node();

    /*!
        \brief Makes new entry unser 'node' - useful for garbage collection
        \param node Node should be current local variables root
        \param var Variable to collect
        \return Returns new node's iterator
    */
    tree<c_tree_node, c_single_element_allocator<tree_node_<c_tree_node> > >::iterator add_local_var_no_reparent(tree<c_tree_node, c_single_element_allocator<tree_node_<c_tree_node> > >::iterator node, c_var_base* var);

    /*!
        \brief Moves whole subtree to temporary store (garbage collection)
        \version 1
        \date 2006-11-25
        \author VooDooMan
        \param what Root of subtree, should be erased after return
    */
    void move_to_temp(tree<c_tree_node, c_single_element_allocator<tree_node_<c_tree_node> > >::iterator& what);

    /*!
        \brief Checks if there is reference to/from variable, and clears that record (for use before variable deletion)
        \version 1
        \date 2007-02-07
        \author VooDooMan
        \param var Variable to check reference to/from it
    */
    void clear_reference(c_var_base* var);

public:

    /*!
        \brief Constructor
        \version 1
        \author VooDooMan
        \date 2006-03-19
    */
    c_var_keeper();

    /*!
        \brief Destructor
        \version 1
        \author VooDooMan
        \date 2006-11-15
    */
    ~c_var_keeper();

    /*!
        \brief Allocates new ID for new variable
        \version 2
        \author VooDooMan
        \date 2006-03-19, 2007-02-05
        \param value Value to assign to new ID
        \return Returns new allocated ID
    */
    var_id_t alloc(c_var_base* value = NULL);

    /*!
        \brief Gets value of variable (without possible dereferencing)
        \version 2
        \author VooDooMan
        \date 2006-05-13, 2006-10-24
        \param id ID of variable
        \return Returns variable
        \param log_fn Logger functiuon call-back
        \param user_data Data passed to call-back function
    */
    c_var_base* get_no_deref(var_id_t id, log_function_ptr_t log_fn, void* user_data);

    /*!
        \brief Gets ID of variable from pointer
        \version 1
        \author VooDooMan
        \date 2006-03-22
        \param value Pointer to variable
        \return Returns ID of variable
        \param log_fn Logger functiuon call-back
        \param user_data Data passed to call-back function
    */
    var_id_t get(c_var_base* value, log_function_ptr_t log_fn, void* user_data);

    /*!
        \brief Adds a new reference
        \version 1
        \author VooDooMan
        \date 2006-03-19
        \param from ID of variable that is referencing to another one
        \param to ID of variable to which the variable will be referencing
        \param log_fn Logger functiuon call-back
        \param user_data Data passed to call-back function
        \return Returns ID of newly created reference variable
    */
    var_id_t add_reference(var_id_t from, var_id_t to, log_function_ptr_t log_fn, void* user_data);

    /*!
        \brief Frees all variables
        \version 1
        \author VooDooMan
        \date 2006-04-17
    */
    void free_all();

    /*!
        \brief Purges objects with reference counter of zero
        \version 2
        \author VooDooMan
        \date 2006-04-18, 2006-10-24
        \return Returns number of object purged
        \warning Obsolete
    */
    unsigned purge(session_id_t) { return 0; }

    /*!
        \brief Dumps all variables to file (debug helper only)
        \version 2
        \author VooDooMan
        \date 2006-04-24, 2008-02-23
        \param name Name of dump
        \param module This module
        \param ln Line number in 'var_keeper.cpp'
    */
    void _dump(const c_module* module, int ln, const string& name);

    /*!
        \brief Performs consistency check agains variable tree (debug helper only)
        \version 2
        \author VooDooMan
        \date 2006-11-07, 2008-02-23
        \param name Name of dump of variables (in case of error)
        \param ln Line number in 'var_keeper.cpp'
    */
    void _check(int ln, const string& name);

    /*!
        \brief Updates variables' session ID (those that are unassigned), sets to 'session'
        \version 1
        \author VooDooMan
        \date 2006-04-24
        \param session Session ID
    */
    void update_variables(session_id_t session);

    /*!
        \brief Updates variables' session ID: remaps all variables in 'session' to 'SESSION_WORK_ID'. Call this before/after "work" command/function
        \version 1
        \author VooDooMan
        \date 2006-04-25
        \param session Session ID to remap
        \param enter 'true' if we are entering "work", false if we are leaving
    */
    void work_issued(session_id_t session, bool enter);

    /*!
        \brief Updates variables' session ID: remaps all variables in 'session' to 'SESSION_INTERNAL_ID'. Call this before/after on_internal_event
        \version 1
        \author VooDooMan
        \date 2006-06-06
        \param session Session ID to remap
        \param enter 'true' if we are entering on_internal_event, false if we are leaving
    */
    void internal_event_issued(session_id_t session, bool enter);

    /*!
        \brief Sets session ID that will be assigned to new variables
        \version 1
        \author VooDooMan
        \date 2006-04-27
        \param session Session ID to set
    */
    void set_current_session(session_id_t session);

    /*!
        \brief Sets session ID that will be assigned to new variables
            for the "core" module, in case into "work" command/function
        \version 1
        \author VooDooMan
        \date 2008-03-14
        \param session Session ID to set
    */
    void set_core_module_working_session(session_id_t session);

    /*!
        \brief Returns session ID that will be assigned to new variables
        \version 1
        \author VooDooMan
        \date 2006-04-27
    */
    session_id_t get_current_session();

    /*!
        \brief Removes reference (but does not allocate 'c_var_uset' placeholder; just remove the link, does not allocate/deallocate resources)
        \version 1
        \author VooDooMan
        \date 2006-05-15
        \param ref_id ID of variable that is referencing to another one (can be 'VAR_ID_INVALID' to ignore)
        \param log_fn Logger functiuon call-back
        \param user_data Data passed to call-back function
    */
    void free_reference(var_id_t ref_id, log_function_ptr_t log_fn, void* user_data);

    /*!
        \brief Sets a new child
        \author VooDooMan
        \version 1
        \date 2006-10-24
        \param base Base (parent) array
        \param key Index
        \param val Value
        \param this_module Module
        \param debug_prefix Debug prefix
        \param debug Debug string
        \param insensitive 'true' for case-insensitive index
    */
    void array_set(c_var_array* base, c_index_base*& key, c_var_base* val, c_module* this_module, const string& debug_prefix, const string& debug, bool insensitive = false);

    /*!
        \brief Deletes an array
        \author VooDooMan
        \version 1
        \date 2006-10-24
        \param base Base (parent) array
    */
    void array_del(c_var_array* base);

    /*!
        \brief Replaces value of specified index
        \param this_module Module
        \param index Index
        \param value Value
        \param insensitive Case insensitivity flag
        \param debug Debug string
        \return Returns true if ok
    */
    bool array_replace(c_module* this_module, c_index_base*& index, c_var_base* value, bool insensitive, const string& debug);

    /*!
        \brief Frees ID (without dereferencing)
        \version 2
        \author VooDooMan
        \date 2006-03-19, 2006-10-24
        \warning Obsolete
    */
    void free(var_id_t, log_function_ptr_t, void*, bool = false, bool = false) { ; }

    /*!
        \brief Gets value of specified index
        \param this_module Module
        \param base Base array
        \param key Index
        \param insensitive Case insensitivity flag
        \param debug_prefix Debug prefix
        \param debug Debug string
        \return Returns NULL if not found
    */
    c_var_base* array_get(c_module* this_module, c_var_array* base, c_index_base* key, bool insensitive, const string& debug_prefix, const string& debug);

    /*!
        \brief Gets value of specified index
        \param this_module Module
        \param base Base array
        \param key Index
        \param insensitive Case insensitivity flag
        \param debug_prefix Debug prefix
        \param debug Debug string
        \return Returns NULL if not found
    */
    c_var_base* array_get(c_module* this_module, c_var_array* base, const string& key, bool insensitive, const string& debug_prefix, const string& debug);

    /*!
        \brief Check for identity ("===")
        \param this_module Module
        \param left Check this against right
        \param right Check this against left
        \param debug_prefix Debug prefix
        \param debug Debug string
        \retval true Identical
        \retval false Not identical
    */
    bool array_is_identical(c_module* this_module, c_var_base* left, c_var_base* right, const string& debug_prefix, const string& debug);

    /*!
        \brief Unsets value of specified index
        \param this_module Module
        \param base Base array
        \param key Index
        \param debug_prefix Debug prefix
        \param debug Debug string
    */
    void array_unset(c_module* this_module, c_var_array* base, c_index_base*& key, const string& debug_prefix, const string& debug);

    /*!
        \brief Check for equality ("==")
        \param this_module Module
        \param left Check this against right
        \param right Check this against left
        \param debug_prefix Debug prefix
        \param debug Debug string
        \retval true Equal
        \retval false Not equal
    */
    bool array_is_equal(c_module* this_module, c_var_base* left, c_var_base* right, const string& debug_prefix, const string& debug);

    /*!
        \brief Sets all members as persistent
        \param base Base array
    */
    void array_set_all_as_persistent(c_var_array* base);

    /*!
        \brief Adjusts reference counter
        \version 2
        \author VooDooMan
        \date 2006-04-18, 2006-10-24
        \warning Obsolete
    */
    void adjust_ref_cnt(var_id_t, signed) { ; }

    /*!
        \brief Assigns new value of variable to ID
        \version 3
        \author VooDooMan
        \date 2006-03-19, 2006-10-24, 2006-11-07
        \param value Value to assign to ID, if 'ro'=='false' and variable is read-only, it deallocates this pointer and returns original read-only value here
        \param id ID of variable
        \param log_fn Logger functiuon call-back
        \param user_data Data passed to call-back function
        \param ro Set to 'true' if we want to overwrite read-only variables as well
        \param no_deref Set to 'true' to force no dereferencing referenced variables
        \param changes1 Adds changes of IDs to vector: from -> to
        \param changes2 Adds changes of IDs to vector: from -> to
        \param this_module Module
        \param debug_prefix Debug prefix
        \param debug Debug string
    */
    var_id_t set(var_id_t id, log_function_ptr_t log_fn, void* user_data, bool ro, c_var_base* &value, bool no_deref, vector<pair<var_id_t, var_id_t> >& changes1, vector<pair<var_id_t, var_id_t> >& changes2, c_module* this_module, const string& debug_prefix, const string& debug);

    /*!
        \brief Gets value of variable
        \version 2
        \author VooDooMan
        \date 2006-03-19, 2006-10-24
        \param id ID of variable
        \return Returns variable
        \param log_fn Logger functiuon call-back
        \param user_data Data passed to call-back function
    */
    c_var_base* get(var_id_t id, log_function_ptr_t log_fn, void* user_data);

    bool array_extract(c_var_array* base, c_index_base*& key, c_tree_node& result, c_module* this_module, const string& debug_prefix, const string& debug);

    bool array_replace(c_module* this_module, c_var_array* base, c_index_base*& key, c_var_base* value, bool insensitive, const string& debug_prefix, const string& debug);

    /*!
        \brief Dumps an array for script's usage
        \version 1
        \author VooDooMan
        \date 2006-10-25
        \param this_module Module
        \param script_type Type of script, e.g. "php", "php_2", "logic.txt", "perl"
        \param base Array
    */
    string array_get_script_string(c_module* this_module, const string& script_type, c_var_array* base);

    /*!
        \brief Updates references
        \param old_val Old value
        \param new_val New value
    */
    void replace_ref(c_var_base* old_val, c_var_base* new_val);

    /*!
        \brief Updates references
        \param old_id Old ID of value
        \param new_val New value
    */
    void replace_ref(var_id_t old_id, c_var_base* new_val);

private:

    enum est
    {
        est_php,
        est_perl,
    };

    /*!
        \brief Dumps non-array and non-reference variable for script's usage
        \version 1
        \author VooDooMan
        \date 2006-10-25
        \param this_module Module
        \param st Script type
        \param base Variable
    */
    string dump_script_simple(c_module* this_module, est st, c_var_base* base);

    /*!
        \brief Dumps 'c_index_int', 'c_index_str' and 'c_index_base' variable for script's usage
        \version 1
        \author VooDooMan
        \date 2006-11-19
        \param this_module Module
        \param st Script type
        \param value Variable
    */
    string dump_script_simple(c_module* this_module, est st, c_index_base* value);

public:

    /*!
        \brief Searches for value in an array and removes that entry (even if it is read-only)
        \version 1
        \author VooDooMan
        \date 2006-10-25
        \param base Array
        \param value Value
        \warning 'value' will be deleted!
    */
    void array_unset_ro_value(c_var_array* base, c_var_base* value);

    /*!
        \brief Dumps an array to string
        \version 1
        \author VooDooMan
        \date 2006-10-26
        \param this_module Module
        \param base Array
    */
    string array_dump(c_module* this_module, c_var_array* base);

    /*!
        \brief Dumps one-dimensional array to string vector (and make conversions from-non-strings to string; except for nested array and UNSET)
        \date 2008-02-06
        \version 1
        \author VooDooMan
        \retval true Okay
        \retval false There is nested sub-array
        \param this_module Module
        \param base Base array to dump
        \param result Returns result
    */
    bool array_dump_to_string_vector_wo_unset_and_array(c_module* this_module, c_var_array* base, vector<pair<const c_index_base*,string> >& result);

    /*!
        \brief Dumps an array for botnet (encode)
        \version 1
        \author VooDooMan
        \date 2006-10-26
        \param this_module Module
        \param base Array
        \param data Returns binary data, shuld be MY_DELETE_A'd
        \param size Returns size of 'data'
    */
    void array_dump_for_botnet(c_module* this_module, c_var_array* base, char* &data, size_t& size);

    /*!
        \brief Clears unset variables in array (gets them away from there)
        \version 1
        \author VooDooMan
        \date 2006-10-26
        \param this_module Module
        \param base Array
    */
    void array_clear_unset(c_module* this_module, c_var_array* base);

    /*!
        \brief Gets a value from stream of indexes
        \version 1
        \author VooDooMan
        \date 2006-10-26
        \param base Array
        \param path Stream of indexes
        \param insensitive Stream of "string index case insensitive" flags
        \param this_module Module
        \param debug Debug string
        \return Returns a value
    */
    c_var_base* array_stream_read(c_var_array* base, vector<c_index_base*>& path, vector<bool>& insensitive, c_module* this_module, const string& debug);

    /*!
        \brief Sets a value from stream of indexes
        \version 1
        \author VooDooMan
        \date 2006-10-27
        \param base Array
        \param path Stream of indexes
        \param insensitive Stream of "string index case insensitive" flags
        \param value Value to set
        \param this_module Module
        \param debug Debug string
    */
    void array_stream_write(c_var_array* base, vector<c_index_base*>& path, const vector<bool>& insensitive, c_var_base* value, c_module* this_module, const string& debug);

    /*!
        \brief Unsets value of specified index (also read-only)
        \param this_module Module
        \param base Base array
        \param key Index
        \param insensitive Case insensitivity flag
        \param debug_prefix Debug prefix
        \param debug Debug string
    */
    void array_unset(c_module* this_module, c_var_array* base, const string& key, bool insensitive, const string& debug_prefix, const string& debug);

    /*!
        \brief Replaces index in array (case insensitive)
        \param this_module Module
        \param base Base array
        \param old_key Old index
        \param new_key New index
    */
    void array_replace_index(c_module* this_module, c_var_array* base, const string& old_key, const string& new_key);

    /*!
        \brief Removes children if their values equal to specific value
        \param base Base array
        \param val Value to remove
        \version 2
        \date ?, 2008-03-08
    */
    void array_remove_values(c_var_array* base, const string& val);

    /*!
        \brief Removes children if their key and value equal to 'key' => 'value'
        \param base Base array
        \param val Value to match
        \param key Key to match
        \author VooDooMan
        \version 1
        \date 2008-03-08
    */
    void array_remove_values(c_var_array* base, const string& key, const string& val);

    /*!
        \brief Gets number of children (same depth, not recursive)
        \param this_module Module
        \param base Base array
    */
    unsigned array_num_children(c_module* this_module, c_var_array* base);

    /*!
        \brief Gets list of children, where indexes are strings
        \param this_module Module
        \param base Base array
    */
    map<string,c_var_base*> array_get_string_indexes(c_module* this_module, c_var_array* base);

    /*!
        \brief Gets number of children (same depth, not recursive), but counts only non-unset
        \param this_module Module
        \param base Base array
    */
    unsigned array_num_nonunset_children(c_module* this_module, c_var_array* base);

    /*!
        \brief Gets n-th child (not counting unset values)
        \param this_module Module
        \param base Base array
        \param n n-th
    */
    c_tree_node* array_get_nonunset_child(c_module* this_module, c_var_array* base, MY_U_LONG_LONG n);

    /*!
        \brief Assigns a value, replacing past value (even read-only value)
        \param old_val Old value
        \param new_val New value
    */
    void assign(var_id_t old_val, c_var_base* new_val);

    /*!
        \brief Registers new local root node
        \return Returns ID
        \param base Returns a new c_var_array (root of local variables)
        \param copy_from NULL, or array to clone
        \param file For generating debug string
        \param func For generating debug string
        \param line For generating debug string
    */
    unsigned register_local(c_var_array*& base, c_var_array* copy_from, const char* file, const char* func, int line);

    /*!
        \brief Fries local root node
        \param id ID
        \param delayed If to queue it
    */
    void free_local(unsigned id, bool delayed = false);

    /*!
        \brief Fries module's global root node
        \param vars Global variables array
        \param delayed If to queue it
    */
    void free_global(c_var_array* vars, bool delayed = false);

    /*!
        \brief Sets current local ID
        \param id New local ID
        \return Returns an old value
    */
    unsigned set_current_local(unsigned id);

    /*!
        \brief Gets current local ID
    */
    unsigned get_current_local();

    /*!
        \brief Returns super gloval variable
    */
    c_var_array* get_super_global();

    /*!
        \brief Returns new array variable for module's global variables
        \param copy_from Copy from there, can be NULL
        \param file For generating debug string
        \param func For generating debug string
        \param line For generating debug string
    */
    c_var_array* register_mod_global(c_var_array* copy_from, const char* file, const char* func, int line);

    /*!
        \brief Copies array members to target array
        \version 1
        \date 2006-11-07
        \author VooDooMan
        \param target Target array
        \param source Source array
    */
    void array_copy(c_var_array* target, c_var_array* source);

private:

    map<c_var_base*,tree<c_tree_node, c_single_element_allocator<tree_node_<c_tree_node> > >::iterator*> lookup1;      //!< Maps second value (first is key) to tree node iterator

    /*!
        \brief Looks up for variable
        \version 1
        \date 2006-11-04
        \author VooDooMan
        \param second Variable to lookup
        \param iter Returns an iterator, if return value is 'true'
        \return Returns 'true' if found
    */
    bool var_lookup(c_var_base* second, tree<c_tree_node, c_single_element_allocator<tree_node_<c_tree_node> > >::iterator& iter);

    /*!
        \brief Adds value to lookup table
        \version 1
        \date 2006-11-04
        \author VooDooMan
        \param second Value
        \param iter Iterator of value
    */
    void var_add_lookup(c_var_base* second, tree<c_tree_node, c_single_element_allocator<tree_node_<c_tree_node> > >::iterator& iter);

    /*!
        \brief Deletes value from lookup table
        \version 1
        \date 2006-11-04
        \author VooDooMan
        \param second Value
        \param simple If 'true', not childeren, stick on this node (it is faster then)
    */
    void var_del_lookup(c_var_base* second, bool simple = false);

    /*!
        \brief Deletes value from lookup table
        \version 1
        \date 2006-11-27
        \author VooDooMan
        \param second Value
        \param simple If 'true', not childeren, stick on this node (it is faster then)
    */
    void var_destroy_lookup(c_var_base* second, bool simple = false);

    friend class c_tree_node;

    /*!
        \brief Adds whole subtree to lookup table
        \version 1
        \date 2006-11-04
        \author VooDooMan
        \param start Where to start
    */
    void var_add_tree_to_lookup(tree<c_tree_node, c_single_element_allocator<tree_node_<c_tree_node> > >::iterator start);

    /*!
        \brief Removes variable from tree
        \version 1
        \date 2006-11-07
        \author VooDooMan
        \param var Variable
    */
    void remove_from_tree(c_var_base* var);

    /*!
        \brief Replaces entry or whole subtree by another
        \version 1
        \date 2006-11-08
        \author VooDooMan
        \param dest Point to be replaced
        \param src Point to replace with
        \param key_save This key will be assigned to 'dest' point, the old one will be MY_DELETE'd by destructor, but only if it is different than 'key_save'
    */
    tree<c_tree_node, c_single_element_allocator<tree_node_<c_tree_node> > >::iterator array_replace_tree(tree<c_tree_node, c_single_element_allocator<tree_node_<c_tree_node> > >::iterator dest, tree<c_tree_node, c_single_element_allocator<tree_node_<c_tree_node> > >::iterator src, c_index_base* key_save);

    /*!
        \brief Invalidates all key/val pointers (to NULL), to use after reparenting to
            current local store, and before whole subtree deletion
        \version 1
        \date 2006-11-08
        \author VooDooMan
        \param root Iterator of root of subtree
    */
    void clear_subtree(tree<c_tree_node, c_single_element_allocator<tree_node_<c_tree_node> > >::iterator root);

    /*!
        \brief Flattens all children and their children and etc, to be all children of parent node
        \version 1
        \date 2006-11-16
        \author VooDooMan
        \param parent Parent node
    */
    void var_flatten_children(tree<c_tree_node, c_single_element_allocator<tree_node_<c_tree_node> > >::iterator parent);

public:

    enum
    {
        GARBAGE_MAX_TIME_INFINITE = -1, //!< Infinite time for garbage collection
    };

    /*!
        \brief Performs garbaging of garbage collection
        \version 1
        \date 2006-11-16
        \author VooDooMan
        \param max_time 'GARBAGE_MAX_TIME_INFINITE' for unlimited, otherwise, maximum allowed time to spent in miliseconds; if reached, it will stop processing, but always processes at least one group (local ID)
        \return Returns number of root nodes garbaged
    */
    size_t do_garbage(int max_time);

    /*!
        \brief Retrieves variable count in the tree
        \version 1
        \date 2007-11-01
        \author VooDooMan
    */
    size_t get_size();

private:
    bool m_in_work;     //!< If we are in "work" command/function
    session_id_t m_core_module_working_session;     //!< Session ID for "core" module while into "work" commnad/function
};

/*!
    \brief Maintains creation of temporary local ID node, and automatic destruction of it in the destructor (when gets out of the scope)
    \version 1
    \date 2007-01-02
    \author VooDooMan
*/
class c_temp_local
{
    unsigned local_id;              //!< Current local ID
    unsigned saved_local_id;        //!< Old, saved local ID (before switching to temporary one)
    c_var_array* local_vars;        //!< Local variables

public:

    /*!
        \brief Constructor
        \param copy_from Variables to copy into new local ID scope
    */
    c_temp_local(c_var_array* copy_from = NULL);

    /*!
        \brief Destructor
    */
    ~c_temp_local();

    /*!
        \brief Returns new local ID
    */
    unsigned get_params(c_var_array* &vars);
};

#endif

