/***************************************************************************
                          var_opers.cpp  -  description
                             -------------------
    begin                : Sun Jan 14 2007
    copyright            : (C) 2007 by Marian VooDooMan Meravy
    email                : ghostvoodooman -a.T= users (dot) sourceforge.net
 ***************************************************************************/

/* $Id: var_opers.cpp,v 1.22 2008/03/11 21:11:32 ghostvoodooman Exp $ */

/***************************************************************************

VooDoo cIRCle - an IRC (ro)bot
Copyright (C) 2007-2008 by Marian VooDooMan Meravy
  (ghostvoodooman -a.T= users (dot) sourceforge.net)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************/

/*!
    \file
    \brief Provides implementation of variable operators
    \version $Id: var_opers.cpp,v 1.22 2008/03/11 21:11:32 ghostvoodooman Exp $
*/

#if defined(HAVE_CONFIG_H)
#   include <config.h>
#endif

#include <string>

#include "utils/assert.h"

#include <stdlib.h>

using namespace std;

#include "utils/memleak.h"
#include "utils/assert.h"
#include "utils.h"

#include "params.h"
#include "compat.h"

#include "variable.h"
#include "var_except.h"
#include "var/c_var_bool.h"
#include "var/c_var_int.h"
#include "var/c_var_string.h"
#include "var/c_var_unset.h"
#include "var/c_var_array.h"
#include "var/c_var_ref.h"
#include "utils/precomp.h"

#if defined(_MSC_VER) || defined(__BORLANDC__)
#   pragma hdrstop
#endif

// ---

// BEGIN: c_var_bool
c_var_base& c_var_bool::operator++(int)
{
    bool orig=value_;
    value_=!value_;
    return *(MY_NEW c_var_bool(orig));
}

c_var_base& c_var_bool::operator--(int)
{
    bool orig=value_;
    value_=!value_;
    return *(MY_NEW c_var_bool(orig));
}

c_var_base& c_var_bool::operator++()
{
    value_=!value_;
    return *(MY_NEW c_var_bool(value_));
}

c_var_base& c_var_bool::operator--()
{
    value_=!value_;
    return *(MY_NEW c_var_bool(value_));
}

void c_var_bool::pre_inc()
{
    value_=!value_;
}

void c_var_bool::pre_dec()
{
    value_=!value_;
}

void c_var_bool::post_inc()
{
    value_=!value_;
}

void c_var_bool::post_dec()
{
    value_=!value_;
}

c_var_base& c_var_bool::operator+(c_var_base& right)
{
    // dispatch it
    return right.plus(this);
}

c_var_base& c_var_bool::plus(c_var_bool* left)
{
    return *(MY_NEW c_var_int( (left->value_?1:0) + (value_?1:0) ));
}

c_var_base& c_var_bool::plus(c_var_int* left)
{
    return *(MY_NEW c_var_int( left->get() + (value_?1:0) ));
}

c_var_base& c_var_bool::plus(c_var_unset* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_unset(c_var_except::LEFT);
}

c_var_base& c_var_bool::plus(c_var_string* left)
{
    bool error;
    long left2=left->to_long(error);
    if(error) {
        throw MY_NEW c_var_except_inv_num(c_var_except::LEFT);
    }
    return *(MY_NEW c_var_int( left2 + (value_?1:0) ));
}

c_var_base& c_var_bool::plus(c_var_array* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_not_num(c_var_except::LEFT);
}

c_var_base& c_var_bool::operator-(c_var_base& right)
{
    // dispatch it
    return right.minus(this);
}

c_var_base& c_var_bool::minus(c_var_bool* left)
{
    return *(MY_NEW c_var_int( (left->value_?1:0) - (value_?1:0) ));
}

c_var_base& c_var_bool::minus(c_var_int* left)
{
    return *(MY_NEW c_var_int( left->get() - (value_?1:0) ));
}

c_var_base& c_var_bool::minus(c_var_unset* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_unset(c_var_except::LEFT);
}

c_var_base& c_var_bool::minus(c_var_string* left)
{
    bool error;
    long left2=left->to_long(error);
    if(error) {
        throw MY_NEW c_var_except_inv_num(c_var_except::LEFT);
    }
    return *(MY_NEW c_var_int( left2 - (value_?1:0) ));
}

c_var_base& c_var_bool::minus(c_var_array* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_not_num(c_var_except::LEFT);
}

c_var_base& c_var_bool::operator*(c_var_base& right)
{
    // dispatch it
    return right.mul(this);
}

c_var_base& c_var_bool::mul(c_var_bool* left)
{
    return *(MY_NEW c_var_int( (left->value_?1:0) * (value_?1:0) ));
}

c_var_base& c_var_bool::mul(c_var_int* left)
{
    return *(MY_NEW c_var_int( left->get() * (value_?1:0) ));
}

c_var_base& c_var_bool::mul(c_var_unset* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_unset(c_var_except::LEFT);
}

c_var_base& c_var_bool::mul(c_var_string* left)
{
    bool error;
    long left2=left->to_long(error);
    if(error) {
        throw MY_NEW c_var_except_inv_num(c_var_except::LEFT);
    }
    return *(MY_NEW c_var_int( left2 * (value_?1:0) ));
}

c_var_base& c_var_bool::mul(c_var_array* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_not_num(c_var_except::LEFT);
}

c_var_base& c_var_bool::operator/(c_var_base& right)
{
    // dispatch it
    return right.div(this);
}

c_var_base& c_var_bool::div(c_var_bool* left)
{
    if(!value_)
        throw MY_NEW c_var_except_div_zero();
    return *(MY_NEW c_var_int( (left->value_?1:0) ));
}

c_var_base& c_var_bool::div(c_var_int* left)
{
    if(!value_)
        throw MY_NEW c_var_except_div_zero();
    return *(MY_NEW c_var_int( left->get() ));
}

c_var_base& c_var_bool::div(c_var_unset* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_unset(c_var_except::LEFT);
}

c_var_base& c_var_bool::div(c_var_string* left)
{
    if(!value_)
        throw MY_NEW c_var_except_div_zero();
    bool error;
    long left2=left->to_long(error);
    if(error) {
        throw MY_NEW c_var_except_inv_num(c_var_except::LEFT);
    }
    return *(MY_NEW c_var_int( left2 /* /1 */ ));
}

c_var_base& c_var_bool::div(c_var_array* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_not_num(c_var_except::LEFT);
}

c_var_base& c_var_bool::operator!()
{
    return *(MY_NEW c_var_bool( !value_ ));
}

c_var_base& c_var_bool::operator==(c_var_base& right)
{
    // dispatch it
    return right.equal(this);
}

c_var_base& c_var_bool::equal(c_var_bool* left)
{
    return *(MY_NEW c_var_bool( left->value_ == value_ ));
}

c_var_base& c_var_bool::equal(c_var_int* left)
{
    return *(MY_NEW c_var_bool( left->is_true() == value_ ));
}

c_var_base& c_var_bool::equal(c_var_unset* UNUSEDARG(left))
{
    return *(MY_NEW c_var_bool( false ));
}

c_var_base& c_var_bool::equal(c_var_string* left)
{
    return *(MY_NEW c_var_bool( left->is_true() == value_ ));
}

c_var_base& c_var_bool::equal(c_var_array* left)
{
    return *(MY_NEW c_var_bool( left->is_true() == value_ ));
}

c_var_base& c_var_bool::operator!=(c_var_base& right)
{
    // dispatch it
    return right.not_equal(this);
}

c_var_base& c_var_bool::not_equal(c_var_bool* left)
{
    return *(MY_NEW c_var_bool( left->value_ != value_ ));
}

c_var_base& c_var_bool::not_equal(c_var_int* left)
{
    return *(MY_NEW c_var_bool( left->is_true() != value_ ));
}

c_var_base& c_var_bool::not_equal(c_var_unset* UNUSEDARG(left))
{
    return *(MY_NEW c_var_bool( true ));
}

c_var_base& c_var_bool::not_equal(c_var_string* left)
{
    return *(MY_NEW c_var_bool( left->is_true() != value_ ));
}

c_var_base& c_var_bool::not_equal(c_var_array* left)
{
    return *(MY_NEW c_var_bool( left->is_true() != value_ ));
}

c_var_base& c_var_bool::concat(c_var_base& right)
{
    // dispatch it
    return right.concat(this);
}

c_var_base& c_var_bool::concat(c_var_bool* left)
{
    return *(MY_NEW c_var_string( left->to_string(NULL)+this->to_string(NULL) ));
}

c_var_base& c_var_bool::concat(c_var_int* left)
{
    return *(MY_NEW c_var_string( left->to_string()+this->to_string(NULL) ));
}

c_var_base& c_var_bool::concat(c_var_unset* left)
{
    return *(MY_NEW c_var_string( left->to_string(NULL)+to_string(NULL) ));
}

c_var_base& c_var_bool::concat(c_var_string* left)
{
    return *(MY_NEW c_var_string( left->get()+to_string(NULL) ));
}

c_var_base& c_var_bool::concat(c_var_array* left)
{
    return *(MY_NEW c_var_string( left->to_string(NULL)+to_string(NULL) ));
}

c_var_base& c_var_bool::operator%(c_var_base& right)
{
    // dispatch it
    return right.mod(this);
}

c_var_base& c_var_bool::mod(c_var_bool* left)
{
    if(!value_)
        throw MY_NEW c_var_except_div_zero();
    return *(MY_NEW c_var_int( (left->value_?1:0) ));
}

c_var_base& c_var_bool::mod(c_var_int* UNUSEDARG(left))
{
    if(!value_)
        throw MY_NEW c_var_except_div_zero();
    return *(MY_NEW c_var_int( 0 ));
}

c_var_base& c_var_bool::mod(c_var_unset* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_unset(c_var_except::LEFT);
}

c_var_base& c_var_bool::mod(c_var_string* left)
{
    if(!value_)
        throw MY_NEW c_var_except_div_zero();
    bool error;
    long left2=left->to_long(error);
    if(error) {
        throw MY_NEW c_var_except_inv_num(c_var_except::LEFT);
    }
    return *(MY_NEW c_var_int( left2 /* %1 */ ));
}

c_var_base& c_var_bool::mod(c_var_array* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_not_num(c_var_except::LEFT);
}

c_var_base& c_var_bool::operator&(c_var_base& right)
{
    // dispatch it
    return right.bit_and(this);
}

c_var_base& c_var_bool::bit_and(c_var_bool* left)
{
    return *(MY_NEW c_var_bool( (value_ && left->value_)?1:0 ));
}

c_var_base& c_var_bool::bit_and(c_var_int* left)
{
    return *(MY_NEW c_var_int( left->get() & (value_?1:0) ));
}

c_var_base& c_var_bool::bit_and(c_var_unset* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_unset(c_var_except::LEFT);
}

c_var_base& c_var_bool::bit_and(c_var_string* left)
{
    bool error;
    long left2=left->to_long(error);
    if(error) {
        throw MY_NEW c_var_except_inv_num(c_var_except::LEFT);
    }
    return *(MY_NEW c_var_int( left2 & (value_?1:0) ));
}

c_var_base& c_var_bool::bit_and(c_var_array* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_not_num(c_var_except::LEFT);
}

c_var_base& c_var_bool::operator|(c_var_base& right)
{
    // dispatch it
    return right.bit_or(this);
}

c_var_base& c_var_bool::bit_or(c_var_bool* left)
{
    return *(MY_NEW c_var_bool( (value_ || left->value_)?1:0 ));
}

c_var_base& c_var_bool::bit_or(c_var_int* left)
{
    return *(MY_NEW c_var_int( left->get() | (value_?1:0) ));
}

c_var_base& c_var_bool::bit_or(c_var_unset* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_unset(c_var_except::LEFT);
}

c_var_base& c_var_bool::bit_or(c_var_string* left)
{
    bool error;
    long left2=left->to_long(error);
    if(error) {
        throw MY_NEW c_var_except_inv_num(c_var_except::LEFT);
    }
    return *(MY_NEW c_var_int( left2 | (value_?1:0) ));
}

c_var_base& c_var_bool::bit_or(c_var_array* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_not_num(c_var_except::LEFT);
}

c_var_base& c_var_bool::operator^(c_var_base& right)
{
    // dispatch it
    return right.bit_xor(this);
}

c_var_base& c_var_bool::bit_xor(c_var_bool* left)
{
    return *(MY_NEW c_var_bool( (left->value_? true : false) ^ (value_? true : false) ));
}

c_var_base& c_var_bool::bit_xor(c_var_int* left)
{
    return *(MY_NEW c_var_int( left->get() ^ (value_?1:0) ));
}

c_var_base& c_var_bool::bit_xor(c_var_unset* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_unset(c_var_except::LEFT);
}

c_var_base& c_var_bool::bit_xor(c_var_string* left)
{
    bool error;
    long left2=left->to_long(error);
    if(error) {
        throw MY_NEW c_var_except_inv_num(c_var_except::LEFT);
    }
    return *(MY_NEW c_var_int( left2 ^ (value_?1:0) ));
}

c_var_base& c_var_bool::bit_xor(c_var_array* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_not_num(c_var_except::LEFT);
}

c_var_base& c_var_bool::operator<<(c_var_base& right)
{
    // dispatch it
    return right.bit_sh_left(this);
}

c_var_base& c_var_bool::bit_sh_left(c_var_bool* left)
{
    return *(MY_NEW c_var_int( (left->value_?1:0) << (value_?1:0) ));
}

c_var_base& c_var_bool::bit_sh_left(c_var_int* left)
{
    return *(MY_NEW c_var_int( left->get() << (value_?1:0) ));
}

c_var_base& c_var_bool::bit_sh_left(c_var_unset* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_unset(c_var_except::LEFT);
}

c_var_base& c_var_bool::bit_sh_left(c_var_string* left)
{
    bool error;
    long left2=left->to_long(error);
    if(error) {
        throw MY_NEW c_var_except_inv_num(c_var_except::LEFT);
    }
    return *(MY_NEW c_var_int( left2 << (value_?1:0) ));
}

c_var_base& c_var_bool::bit_sh_left(c_var_array* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_not_num(c_var_except::LEFT);
}

c_var_base& c_var_bool::operator>>(c_var_base& right)
{
    // dispatch it
    return right.bit_sh_right(this);
}

c_var_base& c_var_bool::bit_sh_right(c_var_bool* left)
{
    return *(MY_NEW c_var_int( (left->value_?1:0) >> (value_?1:0) ));
}

c_var_base& c_var_bool::bit_sh_right(c_var_int* left)
{
    return *(MY_NEW c_var_int( left->get() >> (value_?1:0) ));
}

c_var_base& c_var_bool::bit_sh_right(c_var_unset* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_unset(c_var_except::LEFT);
}

c_var_base& c_var_bool::bit_sh_right(c_var_string* left)
{
    bool error;
    long left2=left->to_long(error);
    if(error) {
        throw MY_NEW c_var_except_inv_num(c_var_except::LEFT);
    }
    return *(MY_NEW c_var_int( left2 >> (value_?1:0) ));
}

c_var_base& c_var_bool::bit_sh_right(c_var_array* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_not_num(c_var_except::LEFT);
}

c_var_base& c_var_bool::operator>(c_var_base& right)
{
    // dispatch it
    return right.greater(this);
}

c_var_base& c_var_bool::greater(c_var_bool* left)
{
    return *(MY_NEW c_var_bool( (left->value_?1:0) > (value_?1:0) ));
}

c_var_base& c_var_bool::greater(c_var_int* left)
{
    return *(MY_NEW c_var_bool( left->get() > (value_?1:0) ));
}

c_var_base& c_var_bool::greater(c_var_unset* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_unset(c_var_except::LEFT);
}

c_var_base& c_var_bool::greater(c_var_string* left)
{
    bool error;
    long left2=left->to_long(error);
    if(error) {
        throw MY_NEW c_var_except_inv_num(c_var_except::LEFT);
    }
    return *(MY_NEW c_var_bool( left2 > (value_?1:0) ));
}

c_var_base& c_var_bool::greater(c_var_array* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_not_num(c_var_except::LEFT);
}

c_var_base& c_var_bool::operator<(c_var_base& right)
{
    // dispatch it
    return right.less(this);
}

c_var_base& c_var_bool::less(c_var_bool* left)
{
    return *(MY_NEW c_var_bool( (left->value_?1:0) < (value_?1:0) ));
}

c_var_base& c_var_bool::less(c_var_int* left)
{
    return *(MY_NEW c_var_bool( left->get() < (value_?1:0) ));
}

c_var_base& c_var_bool::less(c_var_unset* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_unset(c_var_except::LEFT);
}

c_var_base& c_var_bool::less(c_var_string* left)
{
    bool error;
    long left2=left->to_long(error);
    if(error) {
        throw MY_NEW c_var_except_inv_num(c_var_except::LEFT);
    }
    return *(MY_NEW c_var_bool( left2 < (value_?1:0) ));
}

c_var_base& c_var_bool::less(c_var_array* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_not_num(c_var_except::LEFT);
}

c_var_base& c_var_bool::operator>=(c_var_base& right)
{
    // dispatch it
    return right.greater_or_eq(this);
}

c_var_base& c_var_bool::greater_or_eq(c_var_bool* left)
{
    return *(MY_NEW c_var_bool( (left->value_?1:0) >= (value_?1:0) ));
}

c_var_base& c_var_bool::greater_or_eq(c_var_int* left)
{
    return *(MY_NEW c_var_bool( left->get() >= (value_?1:0) ));
}

c_var_base& c_var_bool::greater_or_eq(c_var_unset* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_unset(c_var_except::LEFT);
}

c_var_base& c_var_bool::greater_or_eq(c_var_string* left)
{
    bool error;
    long left2=left->to_long(error);
    if(error) {
        throw MY_NEW c_var_except_inv_num(c_var_except::LEFT);
    }
    return *(MY_NEW c_var_bool( left2 >= (value_?1:0) ));
}

c_var_base& c_var_bool::greater_or_eq(c_var_array* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_not_num(c_var_except::LEFT);
}

c_var_base& c_var_bool::operator<=(c_var_base& right)
{
    // dispatch it
    return right.less_or_eq(this);
}

c_var_base& c_var_bool::less_or_eq(c_var_bool* left)
{
    return *(MY_NEW c_var_bool( (left->value_?1:0) <= (value_?1:0) ));
}

c_var_base& c_var_bool::less_or_eq(c_var_int* left)
{
    return *(MY_NEW c_var_bool( left->get() <= (value_?1:0) ));
}

c_var_base& c_var_bool::less_or_eq(c_var_unset* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_unset(c_var_except::LEFT);
}

c_var_base& c_var_bool::less_or_eq(c_var_string* left)
{
    bool error;
    long left2=left->to_long(error);
    if(error) {
        throw MY_NEW c_var_except_inv_num(c_var_except::LEFT);
    }
    return *(MY_NEW c_var_bool( left2 <= (value_?1:0) ));
}

c_var_base& c_var_bool::less_or_eq(c_var_array* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_not_num(c_var_except::LEFT);
}

c_var_base& c_var_bool::operator~()
{
    return *(MY_NEW c_var_int( ~(value_?1:0) ));
}

c_var_base& c_var_bool::operator&&(c_var_base& right)
{
    // dispatch it
    return right.log_and(this);
}

c_var_base& c_var_bool::log_and(c_var_bool* left)
{
    return *(MY_NEW c_var_bool( left->is_true() && value_ ));
}

c_var_base& c_var_bool::log_and(c_var_int* left)
{
    return *(MY_NEW c_var_bool( left->is_true() && value_ ));
}

c_var_base& c_var_bool::log_and(c_var_unset* UNUSEDARG(left))
{
    return *(MY_NEW c_var_bool( false ));
}

c_var_base& c_var_bool::log_and(c_var_string* left)
{
    return *(MY_NEW c_var_bool( left->is_true() && value_ ));
}

c_var_base& c_var_bool::log_and(c_var_array* left)
{
    return *(MY_NEW c_var_bool( left->is_true() && value_ ));
}

c_var_base& c_var_bool::operator||(c_var_base& right)
{
    // dispatch it
    return right.log_or(this);
}

c_var_base& c_var_bool::log_or(c_var_bool* left)
{
    return *(MY_NEW c_var_bool( left->value_ || value_ ));
}

c_var_base& c_var_bool::log_or(c_var_int* left)
{
    return *(MY_NEW c_var_bool( left->is_true() || value_ ));
}

c_var_base& c_var_bool::log_or(c_var_unset* UNUSEDARG(left))
{
    return *(MY_NEW c_var_bool( is_true() ));
}

c_var_base& c_var_bool::log_or(c_var_string* left)
{
    return *(MY_NEW c_var_bool( left->is_true() || value_ ));
}

c_var_base& c_var_bool::log_or(c_var_array* left)
{
    return *(MY_NEW c_var_bool( left->is_true() || value_ ));
}

// BEGIN: c_var_int
c_var_base& c_var_int::operator++(int)
{
    MY_LONG_LONG orig=value_;
    ++value_;
    return *(MY_NEW c_var_int(orig));
}

c_var_base& c_var_int::operator--(int)
{
    MY_LONG_LONG orig=value_;
    --value_;
    return *(MY_NEW c_var_int(orig));
}

c_var_base& c_var_int::operator++()
{
    ++value_;
    return *(MY_NEW c_var_int(value_));
}

c_var_base& c_var_int::operator--()
{
    --value_;
    return *(MY_NEW c_var_int(value_));
}

void c_var_int::pre_inc()
{
    ++value_;
}

void c_var_int::pre_dec()
{
    --value_;
}

void c_var_int::post_inc()
{
    ++value_;
}

void c_var_int::post_dec()
{
    --value_;
}

c_var_base& c_var_int::operator+(c_var_base& right)
{
    // dispatch it
    return right.plus(this);
}

c_var_base& c_var_int::plus(c_var_bool* left)
{
    return *(MY_NEW c_var_int( left->to_long() + value_ ));
}

c_var_base& c_var_int::plus(c_var_int* left)
{
    return *(MY_NEW c_var_int( left->value_ + value_ ));
}

c_var_base& c_var_int::plus(c_var_unset* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_unset(c_var_except::LEFT);
}

c_var_base& c_var_int::plus(c_var_string* left)
{
    bool error;
    long left2=left->to_long(error);
    if(error) {
        throw MY_NEW c_var_except_inv_num(c_var_except::LEFT);
    }
    return *(MY_NEW c_var_int( left2 + value_ ));
}

c_var_base& c_var_int::plus(c_var_array* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_not_num(c_var_except::LEFT);
}

c_var_base& c_var_int::operator-(c_var_base& right)
{
    // dispatch it
    return right.minus(this);
}

c_var_base& c_var_int::minus(c_var_bool* left)
{
    return *(MY_NEW c_var_int( left->to_long() - value_ ));
}

c_var_base& c_var_int::minus(c_var_int* left)
{
    return *(MY_NEW c_var_int( left->value_ - value_ ));
}

c_var_base& c_var_int::minus(c_var_unset* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_unset(c_var_except::LEFT);
}

c_var_base& c_var_int::minus(c_var_string* left)
{
    bool error;
    long left2=left->to_long(error);
    if(error) {
        throw MY_NEW c_var_except_inv_num(c_var_except::LEFT);
    }
    return *(MY_NEW c_var_int( left2 - value_ ));
}

c_var_base& c_var_int::minus(c_var_array* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_not_num(c_var_except::LEFT);
}

c_var_base& c_var_int::operator*(c_var_base& right)
{
    // dispatch it
    return right.mul(this);
}

c_var_base& c_var_int::mul(c_var_bool* left)
{
    return *(MY_NEW c_var_int( left->to_long() * value_ ));
}

c_var_base& c_var_int::mul(c_var_int* left)
{
    return *(MY_NEW c_var_int( left->value_ * value_ ));
}

c_var_base& c_var_int::mul(c_var_unset* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_unset(c_var_except::LEFT);
}

c_var_base& c_var_int::mul(c_var_string* left)
{
    bool error;
    long left2=left->to_long(error);
    if(error) {
        throw MY_NEW c_var_except_inv_num(c_var_except::LEFT);
    }
    return *(MY_NEW c_var_int( left2 * value_ ));
}

c_var_base& c_var_int::mul(c_var_array* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_not_num(c_var_except::LEFT);
}

c_var_base& c_var_int::operator/(c_var_base& right)
{
    // dispatch it
    return right.div(this);
}

c_var_base& c_var_int::div(c_var_bool* left)
{
    if(!value_)
        throw MY_NEW c_var_except_div_zero();
    return *(MY_NEW c_var_int( left->to_long() / value_ ));
}

c_var_base& c_var_int::div(c_var_int* left)
{
    if(!value_)
        throw MY_NEW c_var_except_div_zero();
    return *(MY_NEW c_var_int( left->value_ / value_ ));
}

c_var_base& c_var_int::div(c_var_unset* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_unset(c_var_except::LEFT);
}

c_var_base& c_var_int::div(c_var_string* left)
{
    if(!value_)
        throw MY_NEW c_var_except_div_zero();
    bool error;
    long left2=left->to_long(error);
    if(error) {
        throw MY_NEW c_var_except_inv_num(c_var_except::LEFT);
    }
    return *(MY_NEW c_var_int( left2 / value_ ));
}

c_var_base& c_var_int::div(c_var_array* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_not_num(c_var_except::LEFT);
}

c_var_base& c_var_int::operator!()
{
    return *(MY_NEW c_var_bool( (!value_) ? true : false ));
}

c_var_base& c_var_int::operator==(c_var_base& right)
{
    // dispatch it
    return right.equal(this);
}

c_var_base& c_var_int::equal(c_var_bool* left)
{
    return *(MY_NEW c_var_bool( left->is_true() == is_true() ));
}

c_var_base& c_var_int::equal(c_var_int* left)
{
    return *(MY_NEW c_var_bool( left->value_ == value_ ));
}

c_var_base& c_var_int::equal(c_var_unset* UNUSEDARG(left))
{
    return *(MY_NEW c_var_bool( false ));
}

c_var_base& c_var_int::equal(c_var_string* left)
{
    return *(MY_NEW c_var_bool( left->is_true() == is_true() ));
}

c_var_base& c_var_int::equal(c_var_array* UNUSEDARG(left))
{
    return *(MY_NEW c_var_bool( false ));
}

c_var_base& c_var_int::operator!=(c_var_base& right)
{
    // dispatch it
    return right.not_equal(this);
}

c_var_base& c_var_int::not_equal(c_var_bool* left)
{
    return *(MY_NEW c_var_bool( left->is_true() != is_true() ));
}

c_var_base& c_var_int::not_equal(c_var_int* left)
{
    return *(MY_NEW c_var_bool( left->value_ != value_ ));
}

c_var_base& c_var_int::not_equal(c_var_unset* UNUSEDARG(left))
{
    return *(MY_NEW c_var_bool( true ));
}

c_var_base& c_var_int::not_equal(c_var_string* left)
{
    return *(MY_NEW c_var_bool( left->is_true() != is_true() ));
}

c_var_base& c_var_int::not_equal(c_var_array* UNUSEDARG(left))
{
    return *(MY_NEW c_var_bool( true ));
}

c_var_base& c_var_int::concat(c_var_base& right)
{
    // dispatch it
    return right.concat(this);
}

c_var_base& c_var_int::concat(c_var_bool* left)
{
    return *(MY_NEW c_var_string( left->to_string(NULL)+to_string() ));
}

c_var_base& c_var_int::concat(c_var_int* left)
{
    return *(MY_NEW c_var_string( left->to_string()+to_string() ));
}

c_var_base& c_var_int::concat(c_var_unset* left)
{
    return *(MY_NEW c_var_string( left->to_string(NULL)+to_string() ));
}

c_var_base& c_var_int::concat(c_var_string* left)
{
    return *(MY_NEW c_var_string( left->get()+to_string() ));
}

c_var_base& c_var_int::concat(c_var_array* left)
{
    return *(MY_NEW c_var_string( left->to_string(NULL)+to_string() ));
}

c_var_base& c_var_int::operator%(c_var_base& right)
{
    // dispatch it
    return right.mod(this);
}

c_var_base& c_var_int::mod(c_var_bool* left)
{
    if(!value_)
        throw MY_NEW c_var_except_div_zero();
    return *(MY_NEW c_var_int( left->to_long() % value_ ));
}

c_var_base& c_var_int::mod(c_var_int* left)
{
    if(!value_)
        throw MY_NEW c_var_except_div_zero();
    return *(MY_NEW c_var_int( left->value_ % value_ ));
}

c_var_base& c_var_int::mod(c_var_unset* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_unset(c_var_except::LEFT);
}

c_var_base& c_var_int::mod(c_var_string* left)
{
    if(!value_)
        throw MY_NEW c_var_except_div_zero();
    bool error;
    long left2=left->to_long(error);
    if(error) {
        throw MY_NEW c_var_except_inv_num(c_var_except::LEFT);
    }
    return *(MY_NEW c_var_int( left2 % value_ ));
}

c_var_base& c_var_int::mod(c_var_array* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_not_num(c_var_except::LEFT);
}

c_var_base& c_var_int::operator&(c_var_base& right)
{
    // dispatch it
    return right.bit_and(this);
}

c_var_base& c_var_int::bit_and(c_var_bool* left)
{
    return *(MY_NEW c_var_int( left->to_long() & value_ ));
}

c_var_base& c_var_int::bit_and(c_var_int* left)
{
    return *(MY_NEW c_var_int( left->value_ & value_ ));
}

c_var_base& c_var_int::bit_and(c_var_unset* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_unset(c_var_except::LEFT);
}

c_var_base& c_var_int::bit_and(c_var_string* left)
{
    bool error;
    long left2=left->to_long(error);
    if(error) {
        throw MY_NEW c_var_except_inv_num(c_var_except::LEFT);
    }
    return *(MY_NEW c_var_int( left2 & value_ ));
}

c_var_base& c_var_int::bit_and(c_var_array* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_not_num(c_var_except::LEFT);
}

c_var_base& c_var_int::operator|(c_var_base& right)
{
    // dispatch it
    return right.bit_or(this);
}

c_var_base& c_var_int::bit_or(c_var_bool* left)
{
    return *(MY_NEW c_var_int( left->to_long() | value_ ));
}

c_var_base& c_var_int::bit_or(c_var_int* left)
{
    return *(MY_NEW c_var_int( left->value_ | value_ ));
}

c_var_base& c_var_int::bit_or(c_var_unset* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_unset(c_var_except::LEFT);
}

c_var_base& c_var_int::bit_or(c_var_string* left)
{
    bool error;
    long left2=left->to_long(error);
    if(error) {
        throw MY_NEW c_var_except_inv_num(c_var_except::LEFT);
    }
    return *(MY_NEW c_var_int( left2 | value_ ));
}

c_var_base& c_var_int::bit_or(c_var_array* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_not_num(c_var_except::LEFT);
}

c_var_base& c_var_int::operator^(c_var_base& right)
{
    // dispatch it
    return right.bit_xor(this);
}

c_var_base& c_var_int::bit_xor(c_var_bool* left)
{
    return *(MY_NEW c_var_int( left->to_long() ^ value_ ));
}

c_var_base& c_var_int::bit_xor(c_var_int* left)
{
    return *(MY_NEW c_var_int( left->value_ ^ value_ ));
}

c_var_base& c_var_int::bit_xor(c_var_unset* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_unset(c_var_except::LEFT);
}

c_var_base& c_var_int::bit_xor(c_var_string* left)
{
    bool error;
    long left2=left->to_long(error);
    if(error) {
        throw MY_NEW c_var_except_inv_num(c_var_except::LEFT);
    }
    return *(MY_NEW c_var_int( left2 ^ value_ ));
}

c_var_base& c_var_int::bit_xor(c_var_array* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_not_num(c_var_except::LEFT);
}

c_var_base& c_var_int::operator<<(c_var_base& right)
{
    // dispatch it
    return right.bit_sh_left(this);
}

c_var_base& c_var_int::bit_sh_left(c_var_bool* left)
{
    return *(MY_NEW c_var_int( left->to_long() << value_ ));
}

c_var_base& c_var_int::bit_sh_left(c_var_int* left)
{
    return *(MY_NEW c_var_int( left->value_ << value_ ));
}

c_var_base& c_var_int::bit_sh_left(c_var_unset* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_unset(c_var_except::LEFT);
}

c_var_base& c_var_int::bit_sh_left(c_var_string* left)
{
    bool error;
    long left2=left->to_long(error);
    if(error) {
        throw MY_NEW c_var_except_inv_num(c_var_except::LEFT);
    }
    return *(MY_NEW c_var_int( left2 << value_ ));
}

c_var_base& c_var_int::bit_sh_left(c_var_array* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_not_num(c_var_except::LEFT);
}

c_var_base& c_var_int::operator>>(c_var_base& right)
{
    // dispatch it
    return right.bit_sh_right(this);
}

c_var_base& c_var_int::bit_sh_right(c_var_bool* left)
{
    return *(MY_NEW c_var_int( left->to_long() >> value_ ));
}

c_var_base& c_var_int::bit_sh_right(c_var_int* left)
{
    return *(MY_NEW c_var_int( left->value_ >> value_ ));
}

c_var_base& c_var_int::bit_sh_right(c_var_unset* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_unset(c_var_except::LEFT);
}

c_var_base& c_var_int::bit_sh_right(c_var_string* left)
{
    bool error;
    long left2=left->to_long(error);
    if(error) {
        throw MY_NEW c_var_except_inv_num(c_var_except::LEFT);
    }
    return *(MY_NEW c_var_int( left2 >> value_ ));
}

c_var_base& c_var_int::bit_sh_right(c_var_array* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_not_num(c_var_except::LEFT);
}

c_var_base& c_var_int::operator>(c_var_base& right)
{
    // dispatch it
    return right.greater(this);
}

c_var_base& c_var_int::greater(c_var_bool* left)
{
    return *(MY_NEW c_var_bool( left->to_long() > value_ ));
}

c_var_base& c_var_int::greater(c_var_int* left)
{
    return *(MY_NEW c_var_bool( left->value_ > value_ ));
}

c_var_base& c_var_int::greater(c_var_unset* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_unset(c_var_except::LEFT);
}

c_var_base& c_var_int::greater(c_var_string* left)
{
    bool error;
    long left2=left->to_long(error);
    if(error) {
        throw MY_NEW c_var_except_inv_num(c_var_except::LEFT);
    }
    return *(MY_NEW c_var_bool( left2 > value_ ));
}

c_var_base& c_var_int::greater(c_var_array* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_not_num(c_var_except::LEFT);
}

c_var_base& c_var_int::operator<(c_var_base& right)
{
    // dispatch it
    return right.less(this);
}

c_var_base& c_var_int::less(c_var_bool* left)
{
    return *(MY_NEW c_var_bool( left->to_long() < value_ ));
}

c_var_base& c_var_int::less(c_var_int* left)
{
    return *(MY_NEW c_var_bool( left->value_ < value_ ));
}

c_var_base& c_var_int::less(c_var_unset* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_unset(c_var_except::LEFT);
}

c_var_base& c_var_int::less(c_var_string* left)
{
    bool error;
    long left2=left->to_long(error);
    if(error) {
        throw MY_NEW c_var_except_inv_num(c_var_except::LEFT);
    }
    return *(MY_NEW c_var_bool( left2 < value_ ));
}

c_var_base& c_var_int::less(c_var_array* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_not_num(c_var_except::LEFT);
}

c_var_base& c_var_int::operator>=(c_var_base& right)
{
    // dispatch it
    return right.greater_or_eq(this);
}

c_var_base& c_var_int::greater_or_eq(c_var_bool* left)
{
    return *(MY_NEW c_var_bool( left->to_long() >= value_ ));
}

c_var_base& c_var_int::greater_or_eq(c_var_int* left)
{
    return *(MY_NEW c_var_bool( left->value_ >= value_ ));
}

c_var_base& c_var_int::greater_or_eq(c_var_unset* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_unset(c_var_except::LEFT);
}

c_var_base& c_var_int::greater_or_eq(c_var_string* left)
{
    bool error;
    long left2=left->to_long(error);
    if(error) {
        throw MY_NEW c_var_except_inv_num(c_var_except::LEFT);
    }
    return *(MY_NEW c_var_bool( left2 >= value_ ));
}

c_var_base& c_var_int::greater_or_eq(c_var_array* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_not_num(c_var_except::LEFT);
}

c_var_base& c_var_int::operator<=(c_var_base& right)
{
    // dispatch it
    return right.less_or_eq(this);
}

c_var_base& c_var_int::less_or_eq(c_var_bool* left)
{
    return *(MY_NEW c_var_bool( left->to_long() <= value_ ));
}

c_var_base& c_var_int::less_or_eq(c_var_int* left)
{
    return *(MY_NEW c_var_bool( left->value_ <= value_ ));
}

c_var_base& c_var_int::less_or_eq(c_var_unset* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_unset(c_var_except::LEFT);
}

c_var_base& c_var_int::less_or_eq(c_var_string* left)
{
    bool error;
    long left2=left->to_long(error);
    if(error) {
        throw MY_NEW c_var_except_inv_num(c_var_except::LEFT);
    }
    return *(MY_NEW c_var_bool( left2 <= value_ ));
}

c_var_base& c_var_int::less_or_eq(c_var_array* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_not_num(c_var_except::LEFT);
}

c_var_base& c_var_int::operator~()
{
    return *(MY_NEW c_var_int( ~value_ ));
}

c_var_base& c_var_int::operator&&(c_var_base& right)
{
    // dispatch it
    return right.log_and(this);
}

c_var_base& c_var_int::log_and(c_var_bool* left)
{
    return *(MY_NEW c_var_bool( left->is_true() && value_ ));
}

c_var_base& c_var_int::log_and(c_var_int* left)
{
    return *(MY_NEW c_var_bool( left->value_ && value_ ));
}

c_var_base& c_var_int::log_and(c_var_unset* UNUSEDARG(left))
{
    return *(MY_NEW c_var_bool( false ));
}

c_var_base& c_var_int::log_and(c_var_string* left)
{
    return *(MY_NEW c_var_bool( left->is_true() && value_ ));
}

c_var_base& c_var_int::log_and(c_var_array* left)
{
    return *(MY_NEW c_var_bool( left->is_true() && value_ ));
}

c_var_base& c_var_int::operator||(c_var_base& right)
{
    // dispatch it
    return right.log_or(this);
}

c_var_base& c_var_int::log_or(c_var_bool* left)
{
    return *(MY_NEW c_var_bool( left->is_true() || value_ ));
}

c_var_base& c_var_int::log_or(c_var_int* left)
{
    return *(MY_NEW c_var_bool( left->value_ || value_ ));
}

c_var_base& c_var_int::log_or(c_var_unset* UNUSEDARG(left))
{
    return *(MY_NEW c_var_bool( is_true() ));
}

c_var_base& c_var_int::log_or(c_var_string* left)
{
    return *(MY_NEW c_var_bool( left->is_true() || value_ ));
}

c_var_base& c_var_int::log_or(c_var_array* left)
{
    return *(MY_NEW c_var_bool( left->is_true() || value_ ));
}

// BEGIN: c_var_string
c_var_base& c_var_string::operator++(int)
{
    // implemented separetely
    throw MY_NEW c_var_except_n_impl_1();
}

c_var_base& c_var_string::operator--(int)
{
    // implemented separetely
    throw MY_NEW c_var_except_n_impl_1();
}

c_var_base& c_var_string::operator++()
{
    // implemented separetely
    throw MY_NEW c_var_except_n_impl_1();
}

c_var_base& c_var_string::operator--()
{
    // implemented separetely
    throw MY_NEW c_var_except_n_impl_1();
}

void c_var_string::pre_inc()
{
    // implemented separetely
    throw MY_NEW c_var_except_n_impl_1();
}

void c_var_string::pre_dec()
{
    // implemented separetely
    throw MY_NEW c_var_except_n_impl_1();
}

void c_var_string::post_inc()
{
    // implemented separetely
    throw MY_NEW c_var_except_n_impl_1();
}

void c_var_string::post_dec()
{
    // implemented separetely
    throw MY_NEW c_var_except_n_impl_1();
}

c_var_base& c_var_string::operator+(c_var_base& right)
{
    // dispatch it
    return right.plus(this);
}

c_var_base& c_var_string::plus(c_var_bool* left)
{
    bool error;
    long right2=to_long(error);
    if(error) {
        throw MY_NEW c_var_except_inv_num(c_var_except::RIGHT);
    }
    return *(MY_NEW c_var_int( left->to_long() + right2 ));
}

c_var_base& c_var_string::plus(c_var_int* left)
{
    bool error;
    long right2=to_long(error);
    if(error) {
        throw MY_NEW c_var_except_inv_num(c_var_except::RIGHT);
    }
    return *(MY_NEW c_var_int( left->get() + right2 ));
}

c_var_base& c_var_string::plus(c_var_unset* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_unset(c_var_except::LEFT);
}

c_var_base& c_var_string::plus(c_var_string* left)
{
    bool error;
    long left2=left->to_long(error);
    if(error) {
        throw MY_NEW c_var_except_inv_num(c_var_except::LEFT);
    }
    long right2=to_long(error);
    if(error) {
        throw MY_NEW c_var_except_inv_num(c_var_except::RIGHT);
    }
    return *(MY_NEW c_var_int( left2 + right2 ));
}

c_var_base& c_var_string::plus(c_var_array* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_not_num(c_var_except::LEFT);
}

c_var_base& c_var_string::operator-(c_var_base& right)
{
    // dispatch it
    return right.minus(this);
}

c_var_base& c_var_string::minus(c_var_bool* left)
{
    bool error;
    long right2=to_long(error);
    if(error) {
        throw MY_NEW c_var_except_inv_num(c_var_except::RIGHT);
    }
    return *(MY_NEW c_var_int( left->to_long() - right2 ));
}

c_var_base& c_var_string::minus(c_var_int* left)
{
    bool error;
    long right2=to_long(error);
    if(error) {
        throw MY_NEW c_var_except_inv_num(c_var_except::RIGHT);
    }
    return *(MY_NEW c_var_int( left->get() - right2 ));
}

c_var_base& c_var_string::minus(c_var_unset* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_unset(c_var_except::LEFT);
}

c_var_base& c_var_string::minus(c_var_string* left)
{
    bool error;
    long left2=left->to_long(error);
    if(error) {
        throw MY_NEW c_var_except_inv_num(c_var_except::LEFT);
    }
    long right2=to_long(error);
    if(error) {
        throw MY_NEW c_var_except_inv_num(c_var_except::RIGHT);
    }
    return *(MY_NEW c_var_int( left2 - right2 ));
}

c_var_base& c_var_string::minus(c_var_array* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_not_num(c_var_except::LEFT);
}

c_var_base& c_var_string::operator*(c_var_base& right)
{
    // dispatch it
    return right.mul(this);
}

c_var_base& c_var_string::mul(c_var_bool* left)
{
    bool error;
    long right2=to_long(error);
    if(error) {
        throw MY_NEW c_var_except_inv_num(c_var_except::RIGHT);
    }
    return *(MY_NEW c_var_int( left->to_long() * right2 ));
}

c_var_base& c_var_string::mul(c_var_int* left)
{
    bool error;
    long right2=to_long(error);
    if(error) {
        throw MY_NEW c_var_except_inv_num(c_var_except::RIGHT);
    }
    return *(MY_NEW c_var_int( left->get() * right2 ));
}

c_var_base& c_var_string::mul(c_var_unset* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_unset(c_var_except::LEFT);
}

c_var_base& c_var_string::mul(c_var_string* left)
{
    bool error;
    long left2=left->to_long(error);
    if(error) {
        throw MY_NEW c_var_except_inv_num(c_var_except::LEFT);
    }
    long right2=to_long(error);
    if(error) {
        throw MY_NEW c_var_except_inv_num(c_var_except::RIGHT);
    }
    return *(MY_NEW c_var_int( left2 * right2 ));
}

c_var_base& c_var_string::mul(c_var_array* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_not_num(c_var_except::LEFT);
}

c_var_base& c_var_string::operator/(c_var_base& right)
{
    // dispatch it
    return right.div(this);
}

c_var_base& c_var_string::div(c_var_bool* left)
{
    bool error;
    long right2=to_long(error);
    if(error) {
        throw MY_NEW c_var_except_inv_num(c_var_except::RIGHT);
    }
    if(!right2)
        throw MY_NEW c_var_except_div_zero();
    return *(MY_NEW c_var_int( left->to_long() / right2 ));
}

c_var_base& c_var_string::div(c_var_int* left)
{
    bool error;
    long right2=to_long(error);
    if(error) {
        throw MY_NEW c_var_except_inv_num(c_var_except::RIGHT);
    }
    if(!right2)
        throw MY_NEW c_var_except_div_zero();
    return *(MY_NEW c_var_int( left->get() / right2 ));
}

c_var_base& c_var_string::div(c_var_unset* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_unset(c_var_except::LEFT);
}

c_var_base& c_var_string::div(c_var_string* left)
{
    bool error;
    long right2=to_long(error);
    if(error) {
        throw MY_NEW c_var_except_inv_num(c_var_except::RIGHT);
    }
    if(!right2)
        throw MY_NEW c_var_except_div_zero();
    long left2=left->to_long(error);
    if(error) {
        throw MY_NEW c_var_except_inv_num(c_var_except::LEFT);
    }
    return *(MY_NEW c_var_int( left2 / right2 ));
}

c_var_base& c_var_string::div(c_var_array* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_not_num(c_var_except::LEFT);
}

c_var_base& c_var_string::operator!()
{
    return *(MY_NEW c_var_bool( !is_true() ));
}

c_var_base& c_var_string::operator==(c_var_base& right)
{
    // dispatch it
    return right.equal(this);
}

c_var_base& c_var_string::equal(c_var_bool* left)
{
    return *(MY_NEW c_var_bool( left->is_true() == is_true() ));
}

c_var_base& c_var_string::equal(c_var_int* left)
{
    return *(MY_NEW c_var_bool( left->is_true() == is_true() ));
}

c_var_base& c_var_string::equal(c_var_unset* UNUSEDARG(left))
{
    return *(MY_NEW c_var_bool( false ));
}

c_var_base& c_var_string::equal(c_var_string* left)
{
    return *(MY_NEW c_var_bool( (!left->value_.compare(value_)) ? true : false ));
}

c_var_base& c_var_string::equal(c_var_array* UNUSEDARG(left))
{
    return *(MY_NEW c_var_bool( false ));
}

c_var_base& c_var_string::operator!=(c_var_base& right)
{
    // dispatch it
    return right.not_equal(this);
}

c_var_base& c_var_string::not_equal(c_var_bool* left)
{
    return *(MY_NEW c_var_bool( left->is_true() != is_true() ));
}

c_var_base& c_var_string::not_equal(c_var_int* left)
{
    return *(MY_NEW c_var_bool( left->is_true() != is_true() ));
}

c_var_base& c_var_string::not_equal(c_var_unset* UNUSEDARG(left))
{
    return *(MY_NEW c_var_bool( true ));
}

c_var_base& c_var_string::not_equal(c_var_string* left)
{
    return *(MY_NEW c_var_bool( (left->value_.compare(value_)) ? true : false )); // !=
}

c_var_base& c_var_string::not_equal(c_var_array* UNUSEDARG(left))
{
    return *(MY_NEW c_var_bool( true ));
}

c_var_base& c_var_string::concat(c_var_base& right)
{
    // dispatch it
    return right.concat(this);
}

c_var_base& c_var_string::concat(c_var_bool* left)
{
    return *(MY_NEW c_var_string( left->to_string(NULL)+value_ ));
}

c_var_base& c_var_string::concat(c_var_int* left)
{
    return *(MY_NEW c_var_string( left->to_string()+value_ ));
}

c_var_base& c_var_string::concat(c_var_unset* left)
{
    return *(MY_NEW c_var_string( left->to_string(NULL)+value_ ));
}

c_var_base& c_var_string::concat(c_var_string* left)
{
    return *(MY_NEW c_var_string( left->value_ + value_ ));
}

c_var_base& c_var_string::concat(c_var_array* left)
{
    return *(MY_NEW c_var_string( left->to_string(NULL)+value_ ));
}

c_var_base& c_var_string::operator%(c_var_base& right)
{
    // dispatch it
    return right.mod(this);
}

c_var_base& c_var_string::mod(c_var_bool* left)
{
    bool error;
    long right2=to_long(error);
    if(error) {
        throw MY_NEW c_var_except_inv_num(c_var_except::RIGHT);
    }
    if(!right2)
        throw MY_NEW c_var_except_div_zero();
    return *(MY_NEW c_var_int( left->to_long() % right2 ));
}

c_var_base& c_var_string::mod(c_var_int* left)
{
    bool error;
    long right2=to_long(error);
    if(error) {
        throw MY_NEW c_var_except_inv_num(c_var_except::RIGHT);
    }
    if(!right2)
        throw MY_NEW c_var_except_div_zero();
    return *(MY_NEW c_var_int( left->get() % right2 ));
}

c_var_base& c_var_string::mod(c_var_unset* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_unset(c_var_except::LEFT);
}

c_var_base& c_var_string::mod(c_var_string* left)
{
    bool error;
    long right2=to_long(error);
    if(error) {
        throw MY_NEW c_var_except_inv_num(c_var_except::RIGHT);
    }
    if(!right2)
        throw MY_NEW c_var_except_div_zero();
    long left2=left->to_long(error);
    if(error) {
        throw MY_NEW c_var_except_inv_num(c_var_except::LEFT);
    }
    return *(MY_NEW c_var_int( left2 % right2 ));
}

c_var_base& c_var_string::mod(c_var_array* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_not_num(c_var_except::LEFT);
}

c_var_base& c_var_string::operator&(c_var_base& right)
{
    // dispatch it
    return right.bit_and(this);
}

c_var_base& c_var_string::bit_and(c_var_bool* left)
{
    bool error;
    long right2=to_long(error);
    if(error) {
        throw MY_NEW c_var_except_inv_num(c_var_except::RIGHT);
    }
    return *(MY_NEW c_var_int( left->to_long() & right2 ));
}

c_var_base& c_var_string::bit_and(c_var_int* left)
{
    bool error;
    long right2=to_long(error);
    if(error) {
        throw MY_NEW c_var_except_inv_num(c_var_except::RIGHT);
    }
    return *(MY_NEW c_var_int( left->get() & right2 ));
}

c_var_base& c_var_string::bit_and(c_var_unset* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_unset(c_var_except::LEFT);
}

c_var_base& c_var_string::bit_and(c_var_string* left)
{
    bool error;
    long left2=left->to_long(error);
    if(error) {
        throw MY_NEW c_var_except_inv_num(c_var_except::LEFT);
    }
    long right2=to_long(error);
    if(error) {
        throw MY_NEW c_var_except_inv_num(c_var_except::RIGHT);
    }
    return *(MY_NEW c_var_int( left2 & right2 ));
}

c_var_base& c_var_string::bit_and(c_var_array* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_not_num(c_var_except::LEFT);
}

c_var_base& c_var_string::operator|(c_var_base& right)
{
    // dispatch it
    return right.bit_or(this);
}

c_var_base& c_var_string::bit_or(c_var_bool* left)
{
    bool error;
    long right2=to_long(error);
    if(error) {
        throw MY_NEW c_var_except_inv_num(c_var_except::RIGHT);
    }
    return *(MY_NEW c_var_int( left->to_long() | right2 ));
}

c_var_base& c_var_string::bit_or(c_var_int* left)
{
    bool error;
    long right2=to_long(error);
    if(error) {
        throw MY_NEW c_var_except_inv_num(c_var_except::RIGHT);
    }
    return *(MY_NEW c_var_int( left->get() | right2 ));
}

c_var_base& c_var_string::bit_or(c_var_unset* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_unset(c_var_except::LEFT);
}

c_var_base& c_var_string::bit_or(c_var_string* left)
{
    bool error;
    long left2=left->to_long(error);
    if(error) {
        throw MY_NEW c_var_except_inv_num(c_var_except::LEFT);
    }
    long right2=to_long(error);
    if(error) {
        throw MY_NEW c_var_except_inv_num(c_var_except::RIGHT);
    }
    return *(MY_NEW c_var_int( left2 | right2 ));
}

c_var_base& c_var_string::bit_or(c_var_array* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_not_num(c_var_except::LEFT);
}

c_var_base& c_var_string::operator^(c_var_base& right)
{
    // dispatch it
    return right.bit_xor(this);
}

c_var_base& c_var_string::bit_xor(c_var_bool* left)
{
    bool error;
    long right2=to_long(error);
    if(error) {
        throw MY_NEW c_var_except_inv_num(c_var_except::RIGHT);
    }
    return *(MY_NEW c_var_int( left->to_long() ^ right2 ));
}

c_var_base& c_var_string::bit_xor(c_var_int* left)
{
    bool error;
    long right2=to_long(error);
    if(error) {
        throw MY_NEW c_var_except_inv_num(c_var_except::RIGHT);
    }
    return *(MY_NEW c_var_int( left->get() ^ right2 ));
}

c_var_base& c_var_string::bit_xor(c_var_unset* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_unset(c_var_except::LEFT);
}

c_var_base& c_var_string::bit_xor(c_var_string* left)
{
    bool error;
    long left2=left->to_long(error);
    if(error) {
        throw MY_NEW c_var_except_inv_num(c_var_except::LEFT);
    }
    long right2=to_long(error);
    if(error) {
        throw MY_NEW c_var_except_inv_num(c_var_except::RIGHT);
    }
    return *(MY_NEW c_var_int( left2 ^ right2 ));
}

c_var_base& c_var_string::bit_xor(c_var_array* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_not_num(c_var_except::LEFT);
}

c_var_base& c_var_string::operator<<(c_var_base& right)
{
    // dispatch it
    return right.bit_sh_left(this);
}

c_var_base& c_var_string::bit_sh_left(c_var_bool* left)
{
    bool error;
    long right2=to_long(error);
    if(error) {
        throw MY_NEW c_var_except_inv_num(c_var_except::RIGHT);
    }
    return *(MY_NEW c_var_int( left->to_long() << right2 ));
}

c_var_base& c_var_string::bit_sh_left(c_var_int* left)
{
    bool error;
    long right2=to_long(error);
    if(error) {
        throw MY_NEW c_var_except_inv_num(c_var_except::RIGHT);
    }
    return *(MY_NEW c_var_int( left->get() << right2 ));
}

c_var_base& c_var_string::bit_sh_left(c_var_unset* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_unset(c_var_except::LEFT);
}

c_var_base& c_var_string::bit_sh_left(c_var_string* left)
{
    bool error;
    long left2=left->to_long(error);
    if(error) {
        throw MY_NEW c_var_except_inv_num(c_var_except::LEFT);
    }
    long right2=to_long(error);
    if(error) {
        throw MY_NEW c_var_except_inv_num(c_var_except::RIGHT);
    }
    return *(MY_NEW c_var_int( left2 << right2 ));
}

c_var_base& c_var_string::bit_sh_left(c_var_array* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_not_num(c_var_except::LEFT);
}

c_var_base& c_var_string::operator>>(c_var_base& right)
{
    // dispatch it
    return right.bit_sh_right(this);
}

c_var_base& c_var_string::bit_sh_right(c_var_bool* left)
{
    bool error;
    long right2=to_long(error);
    if(error) {
        throw MY_NEW c_var_except_inv_num(c_var_except::RIGHT);
    }
    return *(MY_NEW c_var_int( left->to_long() >> right2 ));
}

c_var_base& c_var_string::bit_sh_right(c_var_int* left)
{
    bool error;
    long right2=to_long(error);
    if(error) {
        throw MY_NEW c_var_except_inv_num(c_var_except::RIGHT);
    }
    return *(MY_NEW c_var_int( left->get() >> right2 ));
}

c_var_base& c_var_string::bit_sh_right(c_var_unset* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_unset(c_var_except::LEFT);
}

c_var_base& c_var_string::bit_sh_right(c_var_string* left)
{
    bool error;
    long left2=left->to_long(error);
    if(error) {
        throw MY_NEW c_var_except_inv_num(c_var_except::LEFT);
    }
    long right2=to_long(error);
    if(error) {
        throw MY_NEW c_var_except_inv_num(c_var_except::RIGHT);
    }
    return *(MY_NEW c_var_int( left2 >> right2 ));
}

c_var_base& c_var_string::bit_sh_right(c_var_array* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_not_num(c_var_except::LEFT);
}

c_var_base& c_var_string::operator>(c_var_base& right)
{
    // dispatch it
    return right.greater(this);
}

c_var_base& c_var_string::greater(c_var_bool* left)
{
    bool error;
    long right2=to_long(error);
    if(error) {
        throw MY_NEW c_var_except_inv_num(c_var_except::RIGHT);
    }
    return *(MY_NEW c_var_bool( left->to_long() > right2 ));
}

c_var_base& c_var_string::greater(c_var_int* left)
{
    bool error;
    long right2=to_long(error);
    if(error) {
        throw MY_NEW c_var_except_inv_num(c_var_except::RIGHT);
    }
    return *(MY_NEW c_var_bool( left->get() > right2 ));
}

c_var_base& c_var_string::greater(c_var_unset* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_unset(c_var_except::LEFT);
}

c_var_base& c_var_string::greater(c_var_string* left)
{
    bool error;
    long left2=left->to_long(error);
    if(error) {
        throw MY_NEW c_var_except_inv_num(c_var_except::LEFT);
    }
    long right2=to_long(error);
    if(error) {
        throw MY_NEW c_var_except_inv_num(c_var_except::RIGHT);
    }
    return *(MY_NEW c_var_bool( left2 > right2 ));
}

c_var_base& c_var_string::greater(c_var_array* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_not_num(c_var_except::LEFT);
}

c_var_base& c_var_string::operator<(c_var_base& right)
{
    // dispatch it
    return right.less(this);
}

c_var_base& c_var_string::less(c_var_bool* left)
{
    bool error;
    long right2=to_long(error);
    if(error) {
        throw MY_NEW c_var_except_inv_num(c_var_except::RIGHT);
    }
    return *(MY_NEW c_var_bool( left->to_long() < right2 ));
}

c_var_base& c_var_string::less(c_var_int* left)
{
    bool error;
    long right2=to_long(error);
    if(error) {
        throw MY_NEW c_var_except_inv_num(c_var_except::RIGHT);
    }
    return *(MY_NEW c_var_bool( left->get() < right2 ));
}

c_var_base& c_var_string::less(c_var_unset* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_unset(c_var_except::LEFT);
}

c_var_base& c_var_string::less(c_var_string* left)
{
    bool error;
    long left2=left->to_long(error);
    if(error) {
        throw MY_NEW c_var_except_inv_num(c_var_except::LEFT);
    }
    long right2=to_long(error);
    if(error) {
        throw MY_NEW c_var_except_inv_num(c_var_except::RIGHT);
    }
    return *(MY_NEW c_var_bool( left2 < right2 ));
}

c_var_base& c_var_string::less(c_var_array* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_not_num(c_var_except::LEFT);
}

c_var_base& c_var_string::operator>=(c_var_base& right)
{
    // dispatch it
    return right.greater_or_eq(this);
}

c_var_base& c_var_string::greater_or_eq(c_var_bool* left)
{
    bool error;
    long right2=to_long(error);
    if(error) {
        throw MY_NEW c_var_except_inv_num(c_var_except::RIGHT);
    }
    return *(MY_NEW c_var_bool( left->to_long() >= right2 ));
}

c_var_base& c_var_string::greater_or_eq(c_var_int* left)
{
    bool error;
    long right2=to_long(error);
    if(error) {
        throw MY_NEW c_var_except_inv_num(c_var_except::RIGHT);
    }
    return *(MY_NEW c_var_bool( left->get() >= right2 ));
}

c_var_base& c_var_string::greater_or_eq(c_var_unset* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_unset(c_var_except::LEFT);
}

c_var_base& c_var_string::greater_or_eq(c_var_string* left)
{
    bool error;
    long left2=left->to_long(error);
    if(error) {
        throw MY_NEW c_var_except_inv_num(c_var_except::LEFT);
    }
    long right2=to_long(error);
    if(error) {
        throw MY_NEW c_var_except_inv_num(c_var_except::RIGHT);
    }
    return *(MY_NEW c_var_bool( left2 >= right2 ));
}

c_var_base& c_var_string::greater_or_eq(c_var_array* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_not_num(c_var_except::LEFT);
}

c_var_base& c_var_string::operator<=(c_var_base& right)
{
    // dispatch it
    return right.less_or_eq(this);
}

c_var_base& c_var_string::less_or_eq(c_var_bool* left)
{
    bool error;
    long right2=to_long(error);
    if(error) {
        throw MY_NEW c_var_except_inv_num(c_var_except::RIGHT);
    }
    return *(MY_NEW c_var_bool( left->to_long() <= right2 ));
}

c_var_base& c_var_string::less_or_eq(c_var_int* left)
{
    bool error;
    long right2=to_long(error);
    if(error) {
        throw MY_NEW c_var_except_inv_num(c_var_except::RIGHT);
    }
    return *(MY_NEW c_var_bool( left->get() <= right2 ));
}

c_var_base& c_var_string::less_or_eq(c_var_unset* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_unset(c_var_except::LEFT);
}

c_var_base& c_var_string::less_or_eq(c_var_string* left)
{
    bool error;
    long left2=left->to_long(error);
    if(error) {
        throw MY_NEW c_var_except_inv_num(c_var_except::LEFT);
    }
    long right2=to_long(error);
    if(error) {
        throw MY_NEW c_var_except_inv_num(c_var_except::RIGHT);
    }
    return *(MY_NEW c_var_bool( left2 <= right2 ));
}

c_var_base& c_var_string::less_or_eq(c_var_array* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_not_num(c_var_except::LEFT);
}

c_var_base& c_var_string::operator~()
{
    bool error;
    long right2=to_long(error);
    if(error) {
        throw MY_NEW c_var_except_inv_num(c_var_except::UNARY);
    }
    return *(MY_NEW c_var_int( ~right2 ));
}

c_var_base& c_var_string::operator&&(c_var_base& right)
{
    // dispatch it
    return right.log_and(this);
}

c_var_base& c_var_string::log_and(c_var_bool* left)
{
    return *(MY_NEW c_var_bool( left->is_true() && is_true() ));
}

c_var_base& c_var_string::log_and(c_var_int* left)
{
    return *(MY_NEW c_var_bool( left->get() && is_true() ));
}

c_var_base& c_var_string::log_and(c_var_unset* UNUSEDARG(left))
{
    return *(MY_NEW c_var_bool( false ));
}

c_var_base& c_var_string::log_and(c_var_string* left)
{
    return *(MY_NEW c_var_bool( left->is_true() && is_true() ));
}

c_var_base& c_var_string::log_and(c_var_array* left)
{
    return *(MY_NEW c_var_bool( left->is_true() && is_true() ));
}

c_var_base& c_var_string::operator||(c_var_base& right)
{
    // dispatch it
    return right.log_or(this);
}

c_var_base& c_var_string::log_or(c_var_bool* left)
{
    return *(MY_NEW c_var_bool( left->is_true() || is_true() ));
}

c_var_base& c_var_string::log_or(c_var_int* left)
{
    return *(MY_NEW c_var_bool( left->is_true() || is_true() ));
}

c_var_base& c_var_string::log_or(c_var_unset* UNUSEDARG(left))
{
    return *(MY_NEW c_var_bool( is_true() ));
}

c_var_base& c_var_string::log_or(c_var_string* left)
{
    return *(MY_NEW c_var_bool( left->is_true() || is_true() ));
}

c_var_base& c_var_string::log_or(c_var_array* left)
{
    return *(MY_NEW c_var_bool( left->is_true() || is_true() ));
}

// BEGIN: c_var_unset
c_var_base& c_var_unset::operator++(int)
{
    throw MY_NEW c_var_except_unset(c_var_except::INCDEC);
}

c_var_base& c_var_unset::operator--(int)
{
    throw MY_NEW c_var_except_unset(c_var_except::INCDEC);
}

c_var_base& c_var_unset::operator++()
{
    throw MY_NEW c_var_except_unset(c_var_except::INCDEC);
}

c_var_base& c_var_unset::operator--()
{
    throw MY_NEW c_var_except_unset(c_var_except::INCDEC);
}

void c_var_unset::pre_inc()
{
    throw MY_NEW c_var_except_unset(c_var_except::INCDEC);
}

void c_var_unset::pre_dec()
{
    throw MY_NEW c_var_except_unset(c_var_except::INCDEC);
}

void c_var_unset::post_inc()
{
    throw MY_NEW c_var_except_unset(c_var_except::INCDEC);
}

void c_var_unset::post_dec()
{
    throw MY_NEW c_var_except_unset(c_var_except::INCDEC);
}

c_var_base& c_var_unset::operator+(c_var_base& right)
{
    // dispatch it
    return right.plus(this);
}

c_var_base& c_var_unset::plus(c_var_bool* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_unset(c_var_except::RIGHT);
}

c_var_base& c_var_unset::plus(c_var_int* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_unset(c_var_except::RIGHT);
}

c_var_base& c_var_unset::plus(c_var_unset* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_unset(c_var_except::LEFT);
}

c_var_base& c_var_unset::plus(c_var_string* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_unset(c_var_except::RIGHT);
}

c_var_base& c_var_unset::plus(c_var_array* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_unset(c_var_except::RIGHT);
}

c_var_base& c_var_unset::operator-(c_var_base& right)
{
    // dispatch it
    return right.minus(this);
}

c_var_base& c_var_unset::minus(c_var_bool* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_unset(c_var_except::RIGHT);
}

c_var_base& c_var_unset::minus(c_var_int* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_unset(c_var_except::RIGHT);
}

c_var_base& c_var_unset::minus(c_var_unset* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_unset(c_var_except::LEFT);
}

c_var_base& c_var_unset::minus(c_var_string* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_unset(c_var_except::RIGHT);
}

c_var_base& c_var_unset::minus(c_var_array* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_unset(c_var_except::RIGHT);
}

c_var_base& c_var_unset::operator*(c_var_base& right)
{
    // dispatch it
    return right.mul(this);
}

c_var_base& c_var_unset::mul(c_var_bool* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_unset(c_var_except::RIGHT);
}

c_var_base& c_var_unset::mul(c_var_int* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_unset(c_var_except::RIGHT);
}

c_var_base& c_var_unset::mul(c_var_unset* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_unset(c_var_except::LEFT);
}

c_var_base& c_var_unset::mul(c_var_string* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_unset(c_var_except::RIGHT);
}

c_var_base& c_var_unset::mul(c_var_array* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_unset(c_var_except::RIGHT);
}

c_var_base& c_var_unset::operator/(c_var_base& right)
{
    // dispatch it
    return right.div(this);
}

c_var_base& c_var_unset::div(c_var_bool* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_unset(c_var_except::RIGHT);
}

c_var_base& c_var_unset::div(c_var_int* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_unset(c_var_except::RIGHT);
}

c_var_base& c_var_unset::div(c_var_unset* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_unset(c_var_except::LEFT);
}

c_var_base& c_var_unset::div(c_var_string* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_unset(c_var_except::RIGHT);
}

c_var_base& c_var_unset::div(c_var_array* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_unset(c_var_except::RIGHT);
}

c_var_base& c_var_unset::operator!()
{
    throw MY_NEW c_var_except_unset(c_var_except::UNARY);
}

c_var_base& c_var_unset::operator==(c_var_base& right)
{
    // dispatch it
    return right.equal(this);
}

c_var_base& c_var_unset::equal(c_var_bool* UNUSEDARG(left))
{
    return *(MY_NEW c_var_bool( false ));
}

c_var_base& c_var_unset::equal(c_var_int* UNUSEDARG(left))
{
    return *(MY_NEW c_var_bool( false ));
}

c_var_base& c_var_unset::equal(c_var_unset* UNUSEDARG(left))
{
    return *(MY_NEW c_var_bool( true ));
}

c_var_base& c_var_unset::equal(c_var_string* UNUSEDARG(left))
{
    return *(MY_NEW c_var_bool( false ));
}

c_var_base& c_var_unset::equal(c_var_array* UNUSEDARG(left))
{
    return *(MY_NEW c_var_bool( false ));
}

c_var_base& c_var_unset::operator!=(c_var_base& right)
{
    // dispatch it
    return right.not_equal(this);
}

c_var_base& c_var_unset::not_equal(c_var_bool* UNUSEDARG(left))
{
    return *(MY_NEW c_var_bool( true ));
}

c_var_base& c_var_unset::not_equal(c_var_int* UNUSEDARG(left))
{
    return *(MY_NEW c_var_bool( true ));
}

c_var_base& c_var_unset::not_equal(c_var_unset* UNUSEDARG(left))
{
    return *(MY_NEW c_var_bool( false ));
}

c_var_base& c_var_unset::not_equal(c_var_string* UNUSEDARG(left))
{
    return *(MY_NEW c_var_bool( true ));
}

c_var_base& c_var_unset::not_equal(c_var_array* UNUSEDARG(left))
{
    return *(MY_NEW c_var_bool( true ));
}

c_var_base& c_var_unset::concat(c_var_base& right)
{
    // dispatch it
    return right.concat(this);
}

c_var_base& c_var_unset::concat(c_var_bool* left)
{
    return *(MY_NEW c_var_string( left->to_string(NULL)+to_string(NULL) ));
}

c_var_base& c_var_unset::concat(c_var_int* left)
{
    return *(MY_NEW c_var_string( left->to_string(NULL)+to_string(NULL) ));
}

c_var_base& c_var_unset::concat(c_var_unset* left)
{
    return *(MY_NEW c_var_string( left->to_string(NULL)+to_string(NULL) ));
}

c_var_base& c_var_unset::concat(c_var_string* left)
{
    return *(MY_NEW c_var_string( left->to_string(NULL)+to_string(NULL) ));
}

c_var_base& c_var_unset::concat(c_var_array* left)
{
    return *(MY_NEW c_var_string( left->to_string(NULL)+to_string(NULL) ));
}

c_var_base& c_var_unset::operator%(c_var_base& right)
{
    // dispatch it
    return right.mod(this);
}

c_var_base& c_var_unset::mod(c_var_bool* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_unset(c_var_except::RIGHT);
}

c_var_base& c_var_unset::mod(c_var_int* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_unset(c_var_except::RIGHT);
}

c_var_base& c_var_unset::mod(c_var_unset* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_unset(c_var_except::LEFT);
}

c_var_base& c_var_unset::mod(c_var_string* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_unset(c_var_except::RIGHT);
}

c_var_base& c_var_unset::mod(c_var_array* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_unset(c_var_except::RIGHT);
}

c_var_base& c_var_unset::operator&(c_var_base& right)
{
    // dispatch it
    return right.bit_and(this);
}

c_var_base& c_var_unset::bit_and(c_var_bool* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_unset(c_var_except::RIGHT);
}

c_var_base& c_var_unset::bit_and(c_var_int* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_unset(c_var_except::RIGHT);
}

c_var_base& c_var_unset::bit_and(c_var_unset* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_unset(c_var_except::LEFT);
}

c_var_base& c_var_unset::bit_and(c_var_string* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_unset(c_var_except::RIGHT);
}

c_var_base& c_var_unset::bit_and(c_var_array* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_not_num(c_var_except::LEFT);
}

c_var_base& c_var_unset::operator|(c_var_base& right)
{
    // dispatch it
    return right.bit_or(this);
}

c_var_base& c_var_unset::bit_or(c_var_bool* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_unset(c_var_except::RIGHT);
}

c_var_base& c_var_unset::bit_or(c_var_int* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_unset(c_var_except::RIGHT);
}

c_var_base& c_var_unset::bit_or(c_var_unset* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_unset(c_var_except::LEFT);
}

c_var_base& c_var_unset::bit_or(c_var_string* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_unset(c_var_except::RIGHT);
}

c_var_base& c_var_unset::bit_or(c_var_array* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_unset(c_var_except::RIGHT);
}

c_var_base& c_var_unset::operator^(c_var_base& right)
{
    // dispatch it
    return right.bit_xor(this);
}

c_var_base& c_var_unset::bit_xor(c_var_bool* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_unset(c_var_except::RIGHT);
}

c_var_base& c_var_unset::bit_xor(c_var_int* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_unset(c_var_except::RIGHT);
}

c_var_base& c_var_unset::bit_xor(c_var_unset* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_unset(c_var_except::LEFT);
}

c_var_base& c_var_unset::bit_xor(c_var_string* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_unset(c_var_except::RIGHT);
}

c_var_base& c_var_unset::bit_xor(c_var_array* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_unset(c_var_except::RIGHT);
}

c_var_base& c_var_unset::operator<<(c_var_base& right)
{
    // dispatch it
    return right.bit_sh_left(this);
}

c_var_base& c_var_unset::bit_sh_left(c_var_bool* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_unset(c_var_except::RIGHT);
}

c_var_base& c_var_unset::bit_sh_left(c_var_int* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_unset(c_var_except::RIGHT);
}

c_var_base& c_var_unset::bit_sh_left(c_var_unset* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_unset(c_var_except::LEFT);
}

c_var_base& c_var_unset::bit_sh_left(c_var_string* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_unset(c_var_except::RIGHT);
}

c_var_base& c_var_unset::bit_sh_left(c_var_array* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_unset(c_var_except::RIGHT);
}

c_var_base& c_var_unset::operator>>(c_var_base& right)
{
    // dispatch it
    return right.bit_sh_right(this);
}

c_var_base& c_var_unset::bit_sh_right(c_var_bool* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_unset(c_var_except::RIGHT);
}

c_var_base& c_var_unset::bit_sh_right(c_var_int* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_unset(c_var_except::RIGHT);
}

c_var_base& c_var_unset::bit_sh_right(c_var_unset* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_unset(c_var_except::LEFT);
}

c_var_base& c_var_unset::bit_sh_right(c_var_string* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_unset(c_var_except::RIGHT);
}

c_var_base& c_var_unset::bit_sh_right(c_var_array* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_unset(c_var_except::RIGHT);
}

c_var_base& c_var_unset::operator>(c_var_base& right)
{
    // dispatch it
    return right.greater(this);
}

c_var_base& c_var_unset::greater(c_var_bool* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_unset(c_var_except::RIGHT);
}

c_var_base& c_var_unset::greater(c_var_int* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_unset(c_var_except::RIGHT);
}

c_var_base& c_var_unset::greater(c_var_unset* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_unset(c_var_except::LEFT);
}

c_var_base& c_var_unset::greater(c_var_string* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_unset(c_var_except::RIGHT);
}

c_var_base& c_var_unset::greater(c_var_array* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_unset(c_var_except::RIGHT);
}

c_var_base& c_var_unset::operator<(c_var_base& right)
{
    // dispatch it
    return right.less(this);
}

c_var_base& c_var_unset::less(c_var_bool* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_unset(c_var_except::RIGHT);
}

c_var_base& c_var_unset::less(c_var_int* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_unset(c_var_except::RIGHT);
}

c_var_base& c_var_unset::less(c_var_unset* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_unset(c_var_except::LEFT);
}

c_var_base& c_var_unset::less(c_var_string* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_unset(c_var_except::RIGHT);
}

c_var_base& c_var_unset::less(c_var_array* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_not_num(c_var_except::LEFT);
}

c_var_base& c_var_unset::operator>=(c_var_base& right)
{
    // dispatch it
    return right.greater_or_eq(this);
}

c_var_base& c_var_unset::greater_or_eq(c_var_bool* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_unset(c_var_except::RIGHT);
}

c_var_base& c_var_unset::greater_or_eq(c_var_int* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_unset(c_var_except::RIGHT);
}

c_var_base& c_var_unset::greater_or_eq(c_var_unset* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_unset(c_var_except::LEFT);
}

c_var_base& c_var_unset::greater_or_eq(c_var_string* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_unset(c_var_except::RIGHT);
}

c_var_base& c_var_unset::greater_or_eq(c_var_array* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_unset(c_var_except::RIGHT);
}

c_var_base& c_var_unset::operator<=(c_var_base& right)
{
    // dispatch it
    return right.less_or_eq(this);
}

c_var_base& c_var_unset::less_or_eq(c_var_bool* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_unset(c_var_except::RIGHT);
}

c_var_base& c_var_unset::less_or_eq(c_var_int* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_unset(c_var_except::RIGHT);
}

c_var_base& c_var_unset::less_or_eq(c_var_unset* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_unset(c_var_except::LEFT);
}

c_var_base& c_var_unset::less_or_eq(c_var_string* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_unset(c_var_except::RIGHT);
}

c_var_base& c_var_unset::less_or_eq(c_var_array* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_unset(c_var_except::RIGHT);
}

c_var_base& c_var_unset::operator~()
{
    throw MY_NEW c_var_except_unset(c_var_except::UNARY);
}

c_var_base& c_var_unset::operator&&(c_var_base& right)
{
    // dispatch it
    return right.log_and(this);
}

c_var_base& c_var_unset::log_and(c_var_bool* UNUSEDARG(left))
{
    return *(MY_NEW c_var_bool( false ));
}

c_var_base& c_var_unset::log_and(c_var_int* UNUSEDARG(left))
{
    return *(MY_NEW c_var_bool( false ));
}

c_var_base& c_var_unset::log_and(c_var_unset* UNUSEDARG(left))
{
    return *(MY_NEW c_var_bool( false ));
}

c_var_base& c_var_unset::log_and(c_var_string* UNUSEDARG(left))
{
    return *(MY_NEW c_var_bool( false ));
}

c_var_base& c_var_unset::log_and(c_var_array* UNUSEDARG(left))
{
    return *(MY_NEW c_var_bool( false ));
}

c_var_base& c_var_unset::operator||(c_var_base& right)
{
    // dispatch it
    return right.log_or(this);
}

c_var_base& c_var_unset::log_or(c_var_bool* left)
{
    return *(MY_NEW c_var_bool( left->is_true() ));
}

c_var_base& c_var_unset::log_or(c_var_int* left)
{
    return *(MY_NEW c_var_bool( left->is_true() ));
}

c_var_base& c_var_unset::log_or(c_var_unset* UNUSEDARG(left))
{
    return *(MY_NEW c_var_bool( false ));
}

c_var_base& c_var_unset::log_or(c_var_string* left)
{
    return *(MY_NEW c_var_bool( left->is_true() ));
}

c_var_base& c_var_unset::log_or(c_var_array* left)
{
    return *(MY_NEW c_var_bool( left->is_true() ));
}

// BEGIN: c_var_array
c_var_base& c_var_array::operator++(int)
{
    throw MY_NEW c_var_except_not_num(c_var_except::INCDEC);
}

c_var_base& c_var_array::operator--(int)
{
    throw MY_NEW c_var_except_not_num(c_var_except::INCDEC);
}

c_var_base& c_var_array::operator++()
{
    throw MY_NEW c_var_except_not_num(c_var_except::INCDEC);
}

c_var_base& c_var_array::operator--()
{
    throw MY_NEW c_var_except_not_num(c_var_except::INCDEC);
}

void c_var_array::pre_inc()
{
    throw MY_NEW c_var_except_not_num(c_var_except::INCDEC);
}

void c_var_array::pre_dec()
{
    throw MY_NEW c_var_except_not_num(c_var_except::INCDEC);
}

void c_var_array::post_inc()
{
    throw MY_NEW c_var_except_not_num(c_var_except::INCDEC);
}

void c_var_array::post_dec()
{
    throw MY_NEW c_var_except_not_num(c_var_except::INCDEC);
}

c_var_base& c_var_array::operator+(c_var_base& right)
{
    // dispatch it
    return right.plus(this);
}

c_var_base& c_var_array::plus(c_var_bool* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_not_num(c_var_except::RIGHT);
}

c_var_base& c_var_array::plus(c_var_int* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_not_num(c_var_except::RIGHT);
}

c_var_base& c_var_array::plus(c_var_unset* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_unset(c_var_except::LEFT);
}

c_var_base& c_var_array::plus(c_var_string* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_not_num(c_var_except::RIGHT);
}

c_var_base& c_var_array::plus(c_var_array* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_not_num(c_var_except::RIGHT);
}

c_var_base& c_var_array::operator-(c_var_base& right)
{
    // dispatch it
    return right.minus(this);
}

c_var_base& c_var_array::minus(c_var_bool* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_not_num(c_var_except::RIGHT);
}

c_var_base& c_var_array::minus(c_var_int* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_not_num(c_var_except::RIGHT);
}

c_var_base& c_var_array::minus(c_var_unset* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_unset(c_var_except::LEFT);
}

c_var_base& c_var_array::minus(c_var_string* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_not_num(c_var_except::RIGHT);
}

c_var_base& c_var_array::minus(c_var_array* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_not_num(c_var_except::RIGHT);
}

c_var_base& c_var_array::operator*(c_var_base& right)
{
    // dispatch it
    return right.mul(this);
}

c_var_base& c_var_array::mul(c_var_bool* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_not_num(c_var_except::RIGHT);
}

c_var_base& c_var_array::mul(c_var_int* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_not_num(c_var_except::RIGHT);
}

c_var_base& c_var_array::mul(c_var_unset* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_unset(c_var_except::LEFT);
}

c_var_base& c_var_array::mul(c_var_string* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_not_num(c_var_except::RIGHT);
}

c_var_base& c_var_array::mul(c_var_array* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_not_num(c_var_except::RIGHT);
}

c_var_base& c_var_array::operator/(c_var_base& right)
{
    // dispatch it
    return right.div(this);
}

c_var_base& c_var_array::div(c_var_bool* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_not_num(c_var_except::RIGHT);
}

c_var_base& c_var_array::div(c_var_int* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_not_num(c_var_except::RIGHT);
}

c_var_base& c_var_array::div(c_var_unset* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_unset(c_var_except::LEFT);
}

c_var_base& c_var_array::div(c_var_string* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_not_num(c_var_except::RIGHT);
}

c_var_base& c_var_array::div(c_var_array* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_not_num(c_var_except::RIGHT);
}

c_var_base& c_var_array::operator!()
{
    return *(MY_NEW c_var_bool( !is_true() ));
}

c_var_base& c_var_array::operator==(c_var_base& right)
{
    // dispatch it
    return right.equal(this);
}

c_var_base& c_var_array::equal(c_var_bool* left)
{
    return *(MY_NEW c_var_bool( left->is_true() == is_true() ));
}

c_var_base& c_var_array::equal(c_var_int* UNUSEDARG(left))
{
    return *(MY_NEW c_var_bool( false ));
}

c_var_base& c_var_array::equal(c_var_unset* UNUSEDARG(left))
{
    return *(MY_NEW c_var_bool( false ));
}

c_var_base& c_var_array::equal(c_var_string* UNUSEDARG(left))
{
    return *(MY_NEW c_var_bool( false ));
}

c_var_base& c_var_array::equal(c_var_array* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_n_impl_2();
}

c_var_base& c_var_array::operator!=(c_var_base& right)
{
    // dispatch it
    return right.not_equal(this);
}

c_var_base& c_var_array::not_equal(c_var_bool* left)
{
    return *(MY_NEW c_var_bool( left->is_true() != is_true() ));
}

c_var_base& c_var_array::not_equal(c_var_int* UNUSEDARG(left))
{
    return *(MY_NEW c_var_bool( true ));
}

c_var_base& c_var_array::not_equal(c_var_unset* UNUSEDARG(left))
{
    return *(MY_NEW c_var_bool( true ));
}

c_var_base& c_var_array::not_equal(c_var_string* UNUSEDARG(left))
{
    return *(MY_NEW c_var_bool( true ));
}

c_var_base& c_var_array::not_equal(c_var_array* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_n_impl_2();
}

c_var_base& c_var_array::concat(c_var_base& right)
{
    // dispatch it
    return right.concat(this);
}

c_var_base& c_var_array::concat(c_var_bool* left)
{
    return *(MY_NEW c_var_string( left->to_string(NULL)+to_string(NULL) ));
}

c_var_base& c_var_array::concat(c_var_int* left)
{
    return *(MY_NEW c_var_string( left->to_string(NULL)+to_string(NULL) ));
}

c_var_base& c_var_array::concat(c_var_unset* left)
{
    return *(MY_NEW c_var_string( left->to_string(NULL)+to_string(NULL) ));
}

c_var_base& c_var_array::concat(c_var_string* left)
{
    return *(MY_NEW c_var_string( left->to_string(NULL)+to_string(NULL) ));
}

c_var_base& c_var_array::concat(c_var_array* left)
{
    return *(MY_NEW c_var_string( left->to_string(NULL)+to_string(NULL) ));
}

c_var_base& c_var_array::operator%(c_var_base& right)
{
    // dispatch it
    return right.mod(this);
}

c_var_base& c_var_array::mod(c_var_bool* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_not_num(c_var_except::RIGHT);
}

c_var_base& c_var_array::mod(c_var_int* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_not_num(c_var_except::RIGHT);
}

c_var_base& c_var_array::mod(c_var_unset* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_unset(c_var_except::LEFT);
}

c_var_base& c_var_array::mod(c_var_string* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_not_num(c_var_except::RIGHT);
}

c_var_base& c_var_array::mod(c_var_array* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_not_num(c_var_except::RIGHT);
}

c_var_base& c_var_array::operator&(c_var_base& right)
{
    // dispatch it
    return right.bit_and(this);
}

c_var_base& c_var_array::bit_and(c_var_bool* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_not_num(c_var_except::RIGHT);
}

c_var_base& c_var_array::bit_and(c_var_int* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_not_num(c_var_except::RIGHT);
}

c_var_base& c_var_array::bit_and(c_var_unset* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_unset(c_var_except::LEFT);
}

c_var_base& c_var_array::bit_and(c_var_string* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_not_num(c_var_except::RIGHT);
}

c_var_base& c_var_array::bit_and(c_var_array* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_not_num(c_var_except::RIGHT);
}

c_var_base& c_var_array::operator|(c_var_base& right)
{
    // dispatch it
    return right.bit_or(this);
}

c_var_base& c_var_array::bit_or(c_var_bool* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_not_num(c_var_except::RIGHT);
}

c_var_base& c_var_array::bit_or(c_var_int* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_not_num(c_var_except::RIGHT);
}

c_var_base& c_var_array::bit_or(c_var_unset* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_unset(c_var_except::LEFT);
}

c_var_base& c_var_array::bit_or(c_var_string* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_not_num(c_var_except::RIGHT);
}

c_var_base& c_var_array::bit_or(c_var_array* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_not_num(c_var_except::RIGHT);
}

c_var_base& c_var_array::operator^(c_var_base& right)
{
    // dispatch it
    return right.bit_xor(this);
}

c_var_base& c_var_array::bit_xor(c_var_bool* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_not_num(c_var_except::RIGHT);
}

c_var_base& c_var_array::bit_xor(c_var_int* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_not_num(c_var_except::RIGHT);
}

c_var_base& c_var_array::bit_xor(c_var_unset* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_unset(c_var_except::LEFT);
}

c_var_base& c_var_array::bit_xor(c_var_string* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_not_num(c_var_except::RIGHT);
}

c_var_base& c_var_array::bit_xor(c_var_array* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_not_num(c_var_except::RIGHT);
}

c_var_base& c_var_array::operator<<(c_var_base& right)
{
    // dispatch it
    return right.bit_sh_left(this);
}

c_var_base& c_var_array::bit_sh_left(c_var_bool* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_not_num(c_var_except::RIGHT);
}

c_var_base& c_var_array::bit_sh_left(c_var_int* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_not_num(c_var_except::RIGHT);
}

c_var_base& c_var_array::bit_sh_left(c_var_unset* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_unset(c_var_except::LEFT);
}

c_var_base& c_var_array::bit_sh_left(c_var_string* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_not_num(c_var_except::RIGHT);
}

c_var_base& c_var_array::bit_sh_left(c_var_array* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_not_num(c_var_except::RIGHT);
}

c_var_base& c_var_array::operator>>(c_var_base& right)
{
    // dispatch it
    return right.bit_sh_right(this);
}

c_var_base& c_var_array::bit_sh_right(c_var_bool* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_not_num(c_var_except::RIGHT);
}

c_var_base& c_var_array::bit_sh_right(c_var_int* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_not_num(c_var_except::RIGHT);
}

c_var_base& c_var_array::bit_sh_right(c_var_unset* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_unset(c_var_except::LEFT);
}

c_var_base& c_var_array::bit_sh_right(c_var_string* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_not_num(c_var_except::RIGHT);
}

c_var_base& c_var_array::bit_sh_right(c_var_array* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_not_num(c_var_except::RIGHT);
}

c_var_base& c_var_array::operator>(c_var_base& right)
{
    // dispatch it
    return right.greater(this);
}

c_var_base& c_var_array::greater(c_var_bool* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_not_num(c_var_except::RIGHT);
}

c_var_base& c_var_array::greater(c_var_int* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_not_num(c_var_except::RIGHT);
}

c_var_base& c_var_array::greater(c_var_unset* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_unset(c_var_except::LEFT);
}

c_var_base& c_var_array::greater(c_var_string* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_not_num(c_var_except::RIGHT);
}

c_var_base& c_var_array::greater(c_var_array* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_not_num(c_var_except::RIGHT);
}

c_var_base& c_var_array::operator<(c_var_base& right)
{
    // dispatch it
    return right.less(this);
}

c_var_base& c_var_array::less(c_var_bool* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_not_num(c_var_except::RIGHT);
}

c_var_base& c_var_array::less(c_var_int* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_not_num(c_var_except::RIGHT);
}

c_var_base& c_var_array::less(c_var_unset* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_unset(c_var_except::LEFT);
}

c_var_base& c_var_array::less(c_var_string* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_not_num(c_var_except::RIGHT);
}

c_var_base& c_var_array::less(c_var_array* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_not_num(c_var_except::RIGHT);
}

c_var_base& c_var_array::operator>=(c_var_base& right)
{
    // dispatch it
    return right.greater_or_eq(this);
}

c_var_base& c_var_array::greater_or_eq(c_var_bool* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_not_num(c_var_except::RIGHT);
}

c_var_base& c_var_array::greater_or_eq(c_var_int* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_not_num(c_var_except::RIGHT);
}

c_var_base& c_var_array::greater_or_eq(c_var_unset* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_unset(c_var_except::LEFT);
}

c_var_base& c_var_array::greater_or_eq(c_var_string* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_not_num(c_var_except::RIGHT);
}

c_var_base& c_var_array::greater_or_eq(c_var_array* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_not_num(c_var_except::RIGHT);
}

c_var_base& c_var_array::operator<=(c_var_base& right)
{
    // dispatch it
    return right.less_or_eq(this);
}

c_var_base& c_var_array::less_or_eq(c_var_bool* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_not_num(c_var_except::RIGHT);
}

c_var_base& c_var_array::less_or_eq(c_var_int* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_not_num(c_var_except::RIGHT);
}

c_var_base& c_var_array::less_or_eq(c_var_unset* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_unset(c_var_except::LEFT);
}

c_var_base& c_var_array::less_or_eq(c_var_string* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_not_num(c_var_except::RIGHT);
}

c_var_base& c_var_array::less_or_eq(c_var_array* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_not_num(c_var_except::RIGHT);
}

c_var_base& c_var_array::operator~()
{
    throw MY_NEW c_var_except_not_num(c_var_except::UNARY);
}

c_var_base& c_var_array::operator&&(c_var_base& right)
{
    // dispatch it
    return right.log_and(this);
}

c_var_base& c_var_array::log_and(c_var_bool* left)
{
    return *(MY_NEW c_var_bool( left->is_true() && is_true() ));
}

c_var_base& c_var_array::log_and(c_var_int* left)
{
    return *(MY_NEW c_var_bool( left->is_true() && is_true() ));
}

c_var_base& c_var_array::log_and(c_var_unset* UNUSEDARG(left))
{
    return *(MY_NEW c_var_bool( false ));
}

c_var_base& c_var_array::log_and(c_var_string* left)
{
    return *(MY_NEW c_var_bool( left->is_true() && is_true() ));
}

c_var_base& c_var_array::log_and(c_var_array* left)
{
    return *(MY_NEW c_var_bool( left->is_true() && is_true() ));
}

c_var_base& c_var_array::operator||(c_var_base& right)
{
    // dispatch it
    return right.log_or(this);
}

c_var_base& c_var_array::log_or(c_var_bool* left)
{
    return *(MY_NEW c_var_bool( left->is_true() || is_true() ));
}

c_var_base& c_var_array::log_or(c_var_int* left)
{
    return *(MY_NEW c_var_bool( left->is_true() || is_true() ));
}

c_var_base& c_var_array::log_or(c_var_unset* UNUSEDARG(left))
{
    return *(MY_NEW c_var_bool( is_true() ));
}

c_var_base& c_var_array::log_or(c_var_string* left)
{
    return *(MY_NEW c_var_bool( left->is_true() || is_true() ));
}

c_var_base& c_var_array::log_or(c_var_array* left)
{
    return *(MY_NEW c_var_bool( left->is_true() || is_true() ));
}

// BEGIN: c_var_ref
c_var_base& c_var_ref::operator++(int)
{
    throw MY_NEW c_var_except_n_impl_3();
}

c_var_base& c_var_ref::operator--(int)
{
    throw MY_NEW c_var_except_n_impl_3();
}

c_var_base& c_var_ref::operator++()
{
    throw MY_NEW c_var_except_n_impl_3();
}

c_var_base& c_var_ref::operator--()
{
    throw MY_NEW c_var_except_n_impl_3();
}

void c_var_ref::pre_inc()
{
    throw MY_NEW c_var_except_n_impl_3();
}

void c_var_ref::pre_dec()
{
    throw MY_NEW c_var_except_n_impl_3();
}

void c_var_ref::post_inc()
{
    throw MY_NEW c_var_except_n_impl_3();
}

void c_var_ref::post_dec()
{
    throw MY_NEW c_var_except_n_impl_3();
}

c_var_base& c_var_ref::operator+(c_var_base& UNUSEDARG(right))
{
    throw MY_NEW c_var_except_n_impl_3();
}

c_var_base& c_var_ref::plus(c_var_bool* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_n_impl_3();
}

c_var_base& c_var_ref::plus(c_var_int* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_n_impl_3();
}

c_var_base& c_var_ref::plus(c_var_unset* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_n_impl_3();
}

c_var_base& c_var_ref::plus(c_var_string* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_n_impl_3();
}

c_var_base& c_var_ref::plus(c_var_array* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_n_impl_3();
}

c_var_base& c_var_ref::operator-(c_var_base& UNUSEDARG(right))
{
    throw MY_NEW c_var_except_n_impl_3();
}

c_var_base& c_var_ref::minus(c_var_bool* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_n_impl_3();
}

c_var_base& c_var_ref::minus(c_var_int* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_n_impl_3();
}

c_var_base& c_var_ref::minus(c_var_unset* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_n_impl_3();
}

c_var_base& c_var_ref::minus(c_var_string* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_n_impl_3();
}

c_var_base& c_var_ref::minus(c_var_array* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_n_impl_3();
}

c_var_base& c_var_ref::operator*(c_var_base& UNUSEDARG(right))
{
    throw MY_NEW c_var_except_n_impl_3();
}

c_var_base& c_var_ref::mul(c_var_bool* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_n_impl_3();
}

c_var_base& c_var_ref::mul(c_var_int* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_n_impl_3();
}

c_var_base& c_var_ref::mul(c_var_unset* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_n_impl_3();
}

c_var_base& c_var_ref::mul(c_var_string* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_n_impl_3();
}

c_var_base& c_var_ref::mul(c_var_array* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_n_impl_3();
}

c_var_base& c_var_ref::operator/(c_var_base& UNUSEDARG(right))
{
    throw MY_NEW c_var_except_n_impl_3();
}

c_var_base& c_var_ref::div(c_var_bool* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_n_impl_3();
}

c_var_base& c_var_ref::div(c_var_int* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_n_impl_3();
}

c_var_base& c_var_ref::div(c_var_unset* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_n_impl_3();
}

c_var_base& c_var_ref::div(c_var_string* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_n_impl_3();
}

c_var_base& c_var_ref::div(c_var_array* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_n_impl_3();
}

c_var_base& c_var_ref::operator!()
{
    throw MY_NEW c_var_except_n_impl_3();
}

c_var_base& c_var_ref::operator==(c_var_base& UNUSEDARG(right))
{
    throw MY_NEW c_var_except_n_impl_3();
}

c_var_base& c_var_ref::equal(c_var_bool* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_n_impl_3();
}

c_var_base& c_var_ref::equal(c_var_int* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_n_impl_3();
}

c_var_base& c_var_ref::equal(c_var_unset* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_n_impl_3();
}

c_var_base& c_var_ref::equal(c_var_string* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_n_impl_3();
}

c_var_base& c_var_ref::equal(c_var_array* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_n_impl_3();
}

c_var_base& c_var_ref::operator!=(c_var_base& UNUSEDARG(right))
{
    throw MY_NEW c_var_except_n_impl_3();
}

c_var_base& c_var_ref::not_equal(c_var_bool* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_n_impl_3();
}

c_var_base& c_var_ref::not_equal(c_var_int* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_n_impl_3();
}

c_var_base& c_var_ref::not_equal(c_var_unset* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_n_impl_3();
}

c_var_base& c_var_ref::not_equal(c_var_string* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_n_impl_3();
}

c_var_base& c_var_ref::not_equal(c_var_array* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_n_impl_3();
}

c_var_base& c_var_ref::concat(c_var_base& UNUSEDARG(right))
{
    throw MY_NEW c_var_except_n_impl_3();
}

c_var_base& c_var_ref::concat(c_var_bool* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_n_impl_3();
}

c_var_base& c_var_ref::concat(c_var_int* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_n_impl_3();
}

c_var_base& c_var_ref::concat(c_var_unset* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_n_impl_3();
}

c_var_base& c_var_ref::concat(c_var_string* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_n_impl_3();
}

c_var_base& c_var_ref::concat(c_var_array* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_n_impl_3();
}

c_var_base& c_var_ref::operator%(c_var_base& UNUSEDARG(right))
{
    throw MY_NEW c_var_except_n_impl_3();
}

c_var_base& c_var_ref::mod(c_var_bool* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_n_impl_3();
}

c_var_base& c_var_ref::mod(c_var_int* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_n_impl_3();
}

c_var_base& c_var_ref::mod(c_var_unset* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_n_impl_3();
}

c_var_base& c_var_ref::mod(c_var_string* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_n_impl_3();
}

c_var_base& c_var_ref::mod(c_var_array* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_n_impl_3();
}

c_var_base& c_var_ref::operator&(c_var_base& UNUSEDARG(right))
{
    throw MY_NEW c_var_except_n_impl_3();
}

c_var_base& c_var_ref::bit_and(c_var_bool* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_n_impl_3();
}

c_var_base& c_var_ref::bit_and(c_var_int* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_n_impl_3();
}

c_var_base& c_var_ref::bit_and(c_var_unset* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_n_impl_3();
}

c_var_base& c_var_ref::bit_and(c_var_string* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_n_impl_3();
}

c_var_base& c_var_ref::bit_and(c_var_array* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_not_num(c_var_except::LEFT);
}

c_var_base& c_var_ref::operator|(c_var_base& UNUSEDARG(right))
{
    throw MY_NEW c_var_except_n_impl_3();
}

c_var_base& c_var_ref::bit_or(c_var_bool* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_n_impl_3();
}

c_var_base& c_var_ref::bit_or(c_var_int* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_n_impl_3();
}

c_var_base& c_var_ref::bit_or(c_var_unset* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_n_impl_3();
}

c_var_base& c_var_ref::bit_or(c_var_string* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_n_impl_3();
}

c_var_base& c_var_ref::bit_or(c_var_array* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_n_impl_3();
}

c_var_base& c_var_ref::operator^(c_var_base& UNUSEDARG(right))
{
    throw MY_NEW c_var_except_n_impl_3();
}

c_var_base& c_var_ref::bit_xor(c_var_bool* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_n_impl_3();
}

c_var_base& c_var_ref::bit_xor(c_var_int* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_n_impl_3();
}

c_var_base& c_var_ref::bit_xor(c_var_unset* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_n_impl_3();
}

c_var_base& c_var_ref::bit_xor(c_var_string* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_n_impl_3();
}

c_var_base& c_var_ref::bit_xor(c_var_array* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_n_impl_3();
}

c_var_base& c_var_ref::operator<<(c_var_base& UNUSEDARG(right))
{
    throw MY_NEW c_var_except_n_impl_3();
}

c_var_base& c_var_ref::bit_sh_left(c_var_bool* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_n_impl_3();
}

c_var_base& c_var_ref::bit_sh_left(c_var_int* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_n_impl_3();
}

c_var_base& c_var_ref::bit_sh_left(c_var_unset* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_n_impl_3();
}

c_var_base& c_var_ref::bit_sh_left(c_var_string* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_n_impl_3();
}

c_var_base& c_var_ref::bit_sh_left(c_var_array* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_n_impl_3();
}

c_var_base& c_var_ref::operator>>(c_var_base& UNUSEDARG(right))
{
    throw MY_NEW c_var_except_n_impl_3();
}

c_var_base& c_var_ref::bit_sh_right(c_var_bool* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_n_impl_3();
}

c_var_base& c_var_ref::bit_sh_right(c_var_int* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_n_impl_3();
}

c_var_base& c_var_ref::bit_sh_right(c_var_unset* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_n_impl_3();
}

c_var_base& c_var_ref::bit_sh_right(c_var_string* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_n_impl_3();
}

c_var_base& c_var_ref::bit_sh_right(c_var_array* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_n_impl_3();
}

c_var_base& c_var_ref::operator>(c_var_base& UNUSEDARG(right))
{
    throw MY_NEW c_var_except_n_impl_3();
}

c_var_base& c_var_ref::greater(c_var_bool* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_n_impl_3();
}

c_var_base& c_var_ref::greater(c_var_int* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_n_impl_3();
}

c_var_base& c_var_ref::greater(c_var_unset* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_n_impl_3();
}

c_var_base& c_var_ref::greater(c_var_string* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_n_impl_3();
}

c_var_base& c_var_ref::greater(c_var_array* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_n_impl_3();
}

c_var_base& c_var_ref::operator<(c_var_base& UNUSEDARG(right))
{
    throw MY_NEW c_var_except_n_impl_3();
}

c_var_base& c_var_ref::less(c_var_bool* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_n_impl_3();
}

c_var_base& c_var_ref::less(c_var_int* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_n_impl_3();
}

c_var_base& c_var_ref::less(c_var_unset* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_n_impl_3();
}

c_var_base& c_var_ref::less(c_var_string* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_n_impl_3();
}

c_var_base& c_var_ref::less(c_var_array* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_not_num(c_var_except::LEFT);
}

c_var_base& c_var_ref::operator>=(c_var_base& UNUSEDARG(right))
{
    throw MY_NEW c_var_except_n_impl_3();
}

c_var_base& c_var_ref::greater_or_eq(c_var_bool* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_n_impl_3();
}

c_var_base& c_var_ref::greater_or_eq(c_var_int* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_n_impl_3();
}

c_var_base& c_var_ref::greater_or_eq(c_var_unset* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_n_impl_3();
}

c_var_base& c_var_ref::greater_or_eq(c_var_string* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_n_impl_3();
}

c_var_base& c_var_ref::greater_or_eq(c_var_array* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_n_impl_3();
}

c_var_base& c_var_ref::operator<=(c_var_base& UNUSEDARG(right))
{
    throw MY_NEW c_var_except_n_impl_3();
}

c_var_base& c_var_ref::less_or_eq(c_var_bool* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_n_impl_3();
}

c_var_base& c_var_ref::less_or_eq(c_var_int* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_n_impl_3();
}

c_var_base& c_var_ref::less_or_eq(c_var_unset* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_n_impl_3();
}

c_var_base& c_var_ref::less_or_eq(c_var_string* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_n_impl_3();
}

c_var_base& c_var_ref::less_or_eq(c_var_array* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_n_impl_3();
}

c_var_base& c_var_ref::operator~()
{
    throw MY_NEW c_var_except_n_impl_3();
}

c_var_base& c_var_ref::operator&&(c_var_base& UNUSEDARG(right))
{
    throw MY_NEW c_var_except_n_impl_3();
}

c_var_base& c_var_ref::log_and(c_var_bool* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_n_impl_3();
}

c_var_base& c_var_ref::log_and(c_var_int* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_n_impl_3();
}

c_var_base& c_var_ref::log_and(c_var_unset* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_n_impl_3();
}

c_var_base& c_var_ref::log_and(c_var_string* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_n_impl_3();
}

c_var_base& c_var_ref::log_and(c_var_array* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_n_impl_3();
}

c_var_base& c_var_ref::operator||(c_var_base& UNUSEDARG(right))
{
    throw MY_NEW c_var_except_n_impl_3();
}

c_var_base& c_var_ref::log_or(c_var_bool* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_n_impl_3();
}

c_var_base& c_var_ref::log_or(c_var_int* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_n_impl_3();
}

c_var_base& c_var_ref::log_or(c_var_unset* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_n_impl_3();
}

c_var_base& c_var_ref::log_or(c_var_string* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_n_impl_3();
}

c_var_base& c_var_ref::log_or(c_var_array* UNUSEDARG(left))
{
    throw MY_NEW c_var_except_n_impl_3();
}

