/***************************************************************************
                          variable.cpp  -  description
                             -------------------
    begin                : Tue Sep 6 2005
    copyright            : (C) 2005-2007 by Marian VooDooMan Meravy
    email                : ghostvoodooman -a.T= users (dot) sourceforge.net
 ***************************************************************************/

/* $Id: variable.cpp,v 1.47 2008/08/23 15:10:42 ghostvoodooman Exp $ */

/***************************************************************************

VooDoo cIRCle - an IRC (ro)bot
Copyright (C) 2005-2008 by Marian VooDooMan Meravy
  (ghostvoodooman -a.T= users (dot) sourceforge.net)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************/

/*!
    \file
    \brief Provides classes for variables and expression evaluation
    \version $Id: variable.cpp,v 1.47 2008/08/23 15:10:42 ghostvoodooman Exp $
*/

#if defined(HAVE_CONFIG_H)
#   include <config.h>
#endif

#include <string>
#include <vector>
#include <stack>
#include <list>
#include <map>
#include <deque>

#ifndef _MSC_VER
#   include "utils/assert.h"
#   include "tree/tree.hh"
#else
#   include "utils/assert.h"
#   include "tree/tree.hh"
#endif

#include <stdlib.h>

using namespace std;

#include "utils/memleak.h"
#include "utils/assert.h"
#include "utils.h"

#include "params.h"
#include "compat.h"

#include "variable.h"
#include "function.h"
#include "utils/assert.h"
#include "proc/c_lines.h"
#include "logic/c_smtp.h"
#include "logic/c_dynamic_rule.h"
#include "logic/c_user.h"
#include "logic/c_channel.h"
#include "mod/c_module.h"
#include "log.h"
#include "utils.h"
#include "shared.h"
#include "function.h"
#include "logic.h"
#include "md5.h"
#include "modules.h"
#include "eval.h"
#include "microcode.h"
#include "shared.h"
#include "var_keeper.h"
#include "log/c_logger.h"
#include "var/c_var_base.h"
#include "var/c_var_array.h"
#include "var/c_var_bool.h"
#include "var/c_var_int.h"
#include "var/c_var_ref.h"
#include "var/c_var_string.h"
#include "var/c_var_unset.h"
#include "func/c_opers.h"
#include "func/c_part.h"
#include "func/expression/oper_prec.h"
#include "func/c_expression.h"
#include "microcode/c_micro.h"
#include "microcode/c_micro_array.h"
#include "script_consts/script_consts.h"
#include "utils.h"

#include "utils/precomp.h"

#if defined(_MSC_VER) || defined(__BORLANDC__)
#   pragma hdrstop
#endif

void var_log_fn(c_var_keeper::e_log_severity severity, const string& msg, void* mod)
{
    c_module* module=(c_module*)mod;
    if(!module) {
        module=core_mod;
    }
    if(module)
        if(module->get_logger()) {
            string log;
            switch(severity) {
                case c_var_keeper::LS_FATAL:
                    log="FATAL: ";
                    break;
                case c_var_keeper::LS_WARNING:
                    log="WARNING: ";
                    break;
                default:
                    log="<OTHER_SEVERITY>: ";
                    break;
            }
            log+=msg;
            module->get_logger()->log_print1(E_LOG_DEBUG,LOG_SCOPE_DEFAULT,"",log);
        }
}

/*!
    \brief Prits OOPS to logs
    \author VooDooMan
    \version 2
    \date 2005, 2008-08-16
    \param this_module Module
    \param prefix String perfix
    \param func Function
    \param file File
    \param str Message
    \param debug Debug string
*/
void var_oops_(c_module* this_module, const string& prefix, const string& file, const string& func, const string& str, const string& debug)
{
    string d=prefix;
    d+=": in C++ function ";
    d+=file;
    d+="::";
    d+=func;
    d+=" occurred error: ";
    d+=str;
    if(!debug.empty()) {
        d+=" (";

        string tmp=debug;
        bool nonprintable=false;
        for(size_t i1=0; i1<tmp.length(); ++i1) {
            if(static_cast<unsigned char>(tmp[i1])<0x20 || static_cast<unsigned char>(tmp[i1])>126U) {
                tmp[i1]='?';
                nonprintable=true;
            }
        }
        if(nonprintable)
            tmp+=" <WARNING: Non-printable characters have been replaced by '?' character>";

        d+=tmp;
        d+=")";
    }
    if(this_module && this_module->get_logger()) {
        this_module->get_logger()->log_print1(E_LOG_DEBUG,LOG_SCOPE_MODULE,"",d);
    }
}

/*!
    \brief Prits OOPS to logs
    \author VooDooMan
    \version 1
    \date 2005
    \param this_module Module
    \param prefix String perfix
    \param func Function
    \param file File
    \param str Message
    \param str2 String that will be escaped
    \param debug Debug string
*/
void var_oops_(c_module* this_module, const string& prefix, const string& file, const string& func, const string& str, const string& str2, const string& debug)
{
    string d=prefix;
    d+=": in C++ function ";
    d+=file;
    d+="::";
    d+=func;
    d+=" occurred error: ";
    d+=str;
    d+="\"";
    d+=escape_string_for_logs_1(str2);
    d+="\"";
    if(!debug.empty()) {
        d+=" (";

        string tmp=debug;
        bool nonprintable=false;
        for(size_t i1=0; i1<tmp.length(); ++i1) {
            if(static_cast<unsigned char>(tmp[i1])<0x20 || static_cast<unsigned char>(tmp[i1])>126U) {
                tmp[i1]='?';
                nonprintable=true;
            }
        }
        if(nonprintable)
            tmp+=" <WARNING: Non-printable characters have been replaced by '?' character>";

        d+=tmp;
        d+=")";
    }
    if(this_module && this_module->get_logger()) {
        this_module->get_logger()->log_print1(E_LOG_DEBUG,LOG_SCOPE_MODULE,"",d);
    }
}

/*!
    \brief Prits warning to logs
    \author VooDooMan
    \version 1
    \date 2005
    \param this_module Module
    \param func Function
    \param file File
    \param str Message
    \param debug Debug string
*/
void var_oops(c_module* this_module, const string& file, const string& func, const string& str, const string& debug)
{
    var_oops_(this_module,"WARNING",file,func,str,debug);
}

/*!
    \brief Prits warning to logs (without using C++ file, func, a line number informations)
    \author VooDooMan
    \version 2
    \date 2008-08-19
    \param this_module Module
    \param str Message
    \param debug Debug string for user
    \param prefix String perfix
*/
void var_oops(c_module* this_module, const string& prefix, const string& str, const string& debug)
{
    string d=prefix;
    d+=": ";
    d+=str;
    if(!debug.empty()) {
        d+=" (";

        string tmp=debug;
        bool nonprintable=false;
        for(size_t i1=0; i1<tmp.length(); ++i1) {
            if(static_cast<unsigned char>(tmp[i1])<0x20 || static_cast<unsigned char>(tmp[i1])>126U) {
                tmp[i1]='?';
                nonprintable=true;
            }
        }
        if(nonprintable)
            tmp+=" <WARNING: Non-printable characters have been replaced by '?' character>";

        d+=tmp;
        d+=")";
    }
    if(this_module && this_module->get_logger()) {
        this_module->get_logger()->log_print1(E_LOG_DEBUG,LOG_SCOPE_MODULE,"",d);
    }
}

/*!
    \brief Prits warning to logs
    \author VooDooMan
    \version 1
    \date 2005
    \param this_module Module
    \param func Function
    \param file File
    \param str Message
    \param str2 String that will be escaped
    \param debug Debug string
*/
void var_oops(c_module* this_module, const string& file, const string& func, const string& str, const string& str2, const string& debug)
{
    var_oops_(this_module,"WARNING",file,func,str,str2,debug);
}

/*!
    \brief Prits fatal error to logs
    \author VooDooMan
    \version 1
    \date 2005
    \param this_module Module
    \param func Function
    \param file File
    \param str Message
    \param debug Debug string
*/
void var_fatal(c_module* this_module, const string& file, const string& func, const string& str, const string& debug)
{
    var_oops_(this_module,"FATAL",file,func,str,debug);
}

#if defined(_MSC_VER)
#   pragma warning(push)
#   pragma warning(disable:4702)    // warning C4702: unreachable code
#endif

/*!
    \brief Looks up and evalueates variable beginning with '$' sign, if it is possible
    \author VooDooMan
    \version 2
    \date 2005, 2007-11-05
    \param str Name of variable beginning with '$' sign
    \param vars Array of variables ant their values (can be NULL)
    \param debug Debug string in case of error
    \param return_ref Set to true to return reference (if possible), not copy of variable; receives true if reference is returned
    \return Returns evaluated variable, if it wasn't found in vars array, it returns str value
    \warning If copy of object is returned, the copy receives non-persistent flag
    \param this_module Module
*/
c_var_base* var_eval_obj(c_module* this_module, string str, c_var_array* vars, const string& debug, bool& return_ref)
{
    char unary=0;
    if(str.length() && ('-'==str[0] || '+'==str[0])) {
        unary=str[0];
        str.erase(0,1);
    }

    c_var_base* constant_var;
    if(script_function_translate_constant(str,constant_var)) {
        switch(unary) {
            case 0:
                return_ref=false;
                return constant_var;
            case '-':
                {
                    c_var_base* var=NULL;
                    switch(constant_var->get_var_type()) {
                        case VARTYPE_INT:
                            var=MY_NEW c_var_int(-dynamic_cast<c_var_int*>(constant_var)->get());
                            break;
                        case VARTYPE_BOOL:
                            var=MY_NEW c_var_bool(TO_BOOL(-dynamic_cast<c_var_bool*>(constant_var)->to_int()));
                            break;

                        case VARTYPE_NULL:
                        case VARTYPE_NA:
                        case VARTYPE_UNSET:
                        case VARTYPE_STRING:
                        case VARTYPE_ARRAY:
                            break;
                    }
                    if(var) {
                        MY_DELETE(constant_var);
                        return_ref=false;
                        return var;
                    } else {
                        // user's error
                        MY_DELETE(constant_var);

                        string d="Invalid usage of unary '-' operator";
                        if(!debug.empty()) {
                            d+=" (";
                            d+=debug;
                            d+=")";
                        }
                        throw c_var_exception(this_module,__FILE__,__my_func__,c_var_exception::E_INVALID_OPER,GET_STR_DEBUG_INFO_EX(d));
                    }
                }
                // NOTREACH
            case '+':
                {
                    bool okay=false;
                    switch(constant_var->get_var_type()) {
                        case VARTYPE_INT:
                            okay=true;
                            break;
                        case VARTYPE_BOOL:
                            okay=true;
                            break;
                        case VARTYPE_NULL:
                        case VARTYPE_NA:
                        case VARTYPE_UNSET:
                        case VARTYPE_STRING:
                        case VARTYPE_ARRAY:
                            break;
                    }
                    if(okay) {
                        return_ref=false;
                        return constant_var;
                    } else {
                        // user's error
                        MY_DELETE(constant_var);

                        string d="Invalid usage of unary '+' operator";
                        if(!debug.empty()) {
                            d+=" (";
                            d+=debug;
                            d+=")";
                        }
                        throw c_var_exception(this_module,__FILE__,__my_func__,c_var_exception::E_INVALID_OPER,GET_STR_DEBUG_INFO_EX(d));
                    }
                }
                // NOTREACH
        }
    }

    bool is_number=false;
    if(str.length() && str[0]>='0' && str[0]<='9') {
        str=(string)"\""+str+"\"";
        is_number=true;
    }

    {
        string ln=str;
        string new_ln;
        bool in_string=false;
        bool in_esc=false;
        bool in_apostrophe=false;
        for(unsigned i1=0; i1<ln.length(); ++i1) {
            if(!in_string && !in_esc && !in_apostrophe && ln[i1]==0x20) {
                continue;
            }
            if(in_apostrophe) {
                new_ln+=(string)""+ln[i1];
                continue;
            }
            if(!in_string && '\''==ln[i1]) {
                in_apostrophe=!in_apostrophe;
                new_ln+=(string)""+ln[i1];
                continue;
            }
            if(!in_esc && ln[i1]=='\\') {
                new_ln+=(string)""+ln[i1];
                in_esc=true;
                continue;
            }
            if(in_esc) {
                new_ln+=(string)""+ln[i1];
                in_esc=false;
                continue;
            }
            if(ln[i1]=='"') {
                in_string=!in_string;
                new_ln+=(string)""+ln[i1];
                continue;
            }
            new_ln+=(string)""+ln[i1];
        }
        str=new_ln;
    }

    c_var_base* res=MY_NEW c_var_unset();

    bool got_error=false;

    bool b_str=false;

    bool in_apostrophe=false;

    bool esc=false;

    for(unsigned int i1=0; i1<str.length(); ++i1) {
        if(in_apostrophe && i1+1==str.length())
            break;
        if(!b_str && '\''==str[i1]) {
            b_str=true;
            in_apostrophe=!in_apostrophe;
            if(res->get_var_type()==VARTYPE_UNSET) {
                MY_DELETE(res);
                res=MY_NEW c_var_string();
            }
            continue;
        }
        if(in_apostrophe) {
            c_var_string* r=dynamic_cast<c_var_string*>(res);
            if(!r) {
                if(!got_error) {
                    got_error=true;
                    e_vartype vartype=VARTYPE_NULL;
                    if(res) {
                        vartype=res->get_var_type();
                        MY_DELETE(res);
                        res=NULL;
                    }
                    string d="C++ dynamic_cast<c_var_string*>(res) returned NULL";
                    if(!debug.empty()) {
                        d+=" (";
                        d+=debug;
                        d+=")";
                    }
                    throw c_var_exception(this_module,__FILE__,__my_func__,c_var_exception::E_DYNCAST,GET_STR_DEBUG_INFO_EX(d),vartype);
                }
            } else {
                if(in_apostrophe && esc && '\\'==str[i1]) {
                    (*r)+='\\';
                    esc=false;
                    continue;
                }
                if(in_apostrophe && esc && '\''==str[i1]) {
                    (*r)+='\'';
                    esc=false;
                    continue;
                }
                if(in_apostrophe && esc) {
                    (*r)+='\\';
                    (*r)+=str[i1];
                    esc=false;
                    continue;
                }
                if(in_apostrophe && '\\'==str[i1]) {
                    esc=true;
                    continue;
                }
                esc=false;
                (*r)+=str[i1];
            }
            continue;
        }
        if(str[i1]=='"') {
            b_str=!b_str;
            if(res->get_var_type()==VARTYPE_UNSET) {
                MY_DELETE(res);
                res=MY_NEW c_var_string();
            }
            continue;
        }
        if(b_str && str[i1]=='\\') {
            if(i1+1>=str.length()) {
                string d="OOPS: Occurred error: erroneous escape character '\\' at the end of the string \"";
                d+=str+"\"";
                if(!debug.empty()) {
                    d+=" (";
                    d+=debug;
                    d+=")";
                }
                if(this_module && this_module->get_logger()) {
                    this_module->get_logger()->log_print1(E_LOG_DEBUG,LOG_SCOPE_MODULE,"",GET_STR_DEBUG_INFO_EX(d));
                }
            } else {
                if(i1+1>=str.length()) {
                    string d="OOPS: Occurred error: erroneous escape sequence '\\' at the end of the string \"";
                    d+=str+"\"";
                    if(!debug.empty()) {
                        d+=" (";
                        d+=debug;
                        d+=")";
                    }
                    if(this_module && this_module->get_logger()) {
                        this_module->get_logger()->log_print1(E_LOG_DEBUG,LOG_SCOPE_MODULE,"",GET_STR_DEBUG_INFO_EX(d));
                    }
                }
                char x=str[++i1];
                switch(x) {
                    case '\\':
                        {
                            //res+="\\";
                            char x2='\\';
                            if(res->get_var_type()==VARTYPE_UNSET) {
                                MY_DELETE(res);
                                res=MY_NEW c_var_string();
                            }
                            if(res->get_var_type()!=VARTYPE_STRING) {
                                if(!got_error) {
                                    got_error=true;
                                    string d="OOPS: Occurred error: resulting variable is of type ";
                                    d+=res->get_var_type_as_string();
                                    d+=" and was used for string concatenation";
                                    if(!debug.empty()) {
                                        d+=" (";
                                        d+=debug;
                                        d+=")";
                                    }
                                    if(this_module && this_module->get_logger()) {
                                        this_module->get_logger()->log_print1(E_LOG_DEBUG,LOG_SCOPE_MODULE,"",GET_STR_DEBUG_INFO_EX(d));
                                    }
                                }
                            } else {
                                c_var_string* r=dynamic_cast<c_var_string*>(res);
                                if(!r) {
                                    if(!got_error) {
                                        e_vartype vartype=VARTYPE_NULL;
                                        if(res) {
                                            vartype=res->get_var_type();
                                            MY_DELETE(res);
                                            res=NULL;
                                        }
                                        got_error=true;
                                        string d="C++ dynamic_cast<c_var_string*>(res) returned NULL";
                                        if(!debug.empty()) {
                                            d+=" (";
                                            d+=debug;
                                            d+=")";
                                        }
                                        throw c_var_exception(this_module,__FILE__,__my_func__,c_var_exception::E_DYNCAST,GET_STR_DEBUG_INFO_EX(d),vartype);
                                    }
                                } else
                                    (*r)+=x2;
                            }
                        }
                        break;
                    case '"':
                        {
                            //res+="\"";
                            char x2='"';
                            if(res->get_var_type()==VARTYPE_UNSET) {
                                MY_DELETE(res);
                                res=MY_NEW c_var_string();
                            }
                            if(res->get_var_type()!=VARTYPE_STRING) {
                                if(!got_error) {
                                    got_error=true;
                                    string d="OOPS: Occurred error: resulting variable is of type ";
                                    d+=res->get_var_type_as_string();
                                    d+=" and was used for string concatenation";
                                    if(!debug.empty()) {
                                        d+=" (";
                                        d+=debug;
                                        d+=")";
                                    }
                                    if(this_module && this_module->get_logger()) {
                                        this_module->get_logger()->log_print1(E_LOG_DEBUG,LOG_SCOPE_MODULE,"",GET_STR_DEBUG_INFO_EX(d));
                                    }
                                }
                            } else {
                                c_var_string* r=dynamic_cast<c_var_string*>(res);
                                if(!r) {
                                    if(!got_error) {
                                        got_error=true;
                                        e_vartype vartype=VARTYPE_NULL;
                                        if(res) {
                                            vartype=res->get_var_type();
                                            MY_DELETE(res);
                                            res=NULL;
                                        }
                                        string d="C++ dynamic_cast<c_var_string*>(res) returned NULL";
                                        if(!debug.empty()) {
                                            d+=" (";
                                            d+=debug;
                                            d+=")";
                                        }
                                        throw c_var_exception(this_module,__FILE__,__my_func__,c_var_exception::E_DYNCAST,GET_STR_DEBUG_INFO_EX(d),vartype);
                                    }
                                } else
                                    (*r)+=x2;
                            }
                        }
                        break;
                    case 'r':
                        {
                            //res+="\r";
                            char x2='\r';
                            if(res->get_var_type()==VARTYPE_UNSET) {
                                MY_DELETE(res);
                                res=MY_NEW c_var_string();
                            }
                            if(res->get_var_type()!=VARTYPE_STRING) {
                                if(!got_error) {
                                    got_error=true;
                                    string d="OOPS: Occurred error: resulting variable is of type ";
                                    d+=res->get_var_type_as_string();
                                    d+=" and was used for string concatenation";
                                    if(!debug.empty()) {
                                        d+=" (";
                                        d+=debug;
                                        d+=")";
                                    }
                                    if(this_module && this_module->get_logger()) {
                                        this_module->get_logger()->log_print1(E_LOG_DEBUG,LOG_SCOPE_MODULE,"",GET_STR_DEBUG_INFO_EX(d));
                                    }
                                }
                            } else {
                                c_var_string* r=dynamic_cast<c_var_string*>(res);
                                if(!r) {
                                    if(!got_error) {
                                        got_error=true;
                                        e_vartype vartype=VARTYPE_NULL;
                                        if(res) {
                                            vartype=res->get_var_type();
                                            MY_DELETE(res);
                                            res=NULL;
                                        }
                                        string d="C++ dynamic_cast<c_var_string*>(res) returned NULL";
                                        if(!debug.empty()) {
                                            d+=" (";
                                            d+=debug;
                                            d+=")";
                                        }
                                        throw c_var_exception(this_module,__FILE__,__my_func__,c_var_exception::E_DYNCAST,GET_STR_DEBUG_INFO_EX(d),vartype);
                                    }
                                } else
                                    (*r)+=x2;
                            }
                        }
                        break;
                    case 'n':
                        {
                            //res+="\n";
                            char x2='\n';
                            if(res->get_var_type()==VARTYPE_UNSET) {
                                MY_DELETE(res);
                                res=MY_NEW c_var_string();
                            }
                            if(res->get_var_type()!=VARTYPE_STRING) {
                                if(!got_error) {
                                    got_error=true;
                                    string d="OOPS: Occurred error: resulting variable is of type ";
                                    d+=res->get_var_type_as_string();
                                    d+=" and was used for string concatenation";
                                    if(!debug.empty()) {
                                        d+=" (";
                                        d+=debug;
                                        d+=")";
                                    }
                                    if(this_module && this_module->get_logger()) {
                                        this_module->get_logger()->log_print1(E_LOG_DEBUG,LOG_SCOPE_MODULE,"",GET_STR_DEBUG_INFO_EX(d));
                                    }
                                }
                            } else {
                                c_var_string* r=dynamic_cast<c_var_string*>(res);
                                if(!r) {
                                    if(!got_error) {
                                        got_error=true;
                                        e_vartype vartype=VARTYPE_NULL;
                                        if(res) {
                                            vartype=res->get_var_type();
                                            MY_DELETE(res);
                                            res=NULL;
                                        }
                                        string d="C++ dynamic_cast<c_var_string*>(res) returned NULL";
                                        if(!debug.empty()) {
                                            d+=" (";
                                            d+=debug;
                                            d+=")";
                                        }
                                        throw c_var_exception(this_module,__FILE__,__my_func__,c_var_exception::E_DYNCAST,GET_STR_DEBUG_INFO_EX(d),vartype);
                                    }
                                } else
                                    (*r)+=x2;
                            }
                        }
                        break;
                    case 't':
                        {
                            //res+="\t";
                            char x2='\t';
                            if(res->get_var_type()==VARTYPE_UNSET) {
                                MY_DELETE(res);
                                res=MY_NEW c_var_string();
                            }
                            if(res->get_var_type()!=VARTYPE_STRING) {
                                if(!got_error) {
                                    got_error=true;
                                    string d="OOPS: Occurred error: resulting variable is of type ";
                                    d+=res->get_var_type_as_string();
                                    d+=" and was used for string concatenation";
                                    if(!debug.empty()) {
                                        d+=" (";
                                        d+=debug;
                                        d+=")";
                                    }
                                    if(this_module && this_module->get_logger()) {
                                        this_module->get_logger()->log_print1(E_LOG_DEBUG,LOG_SCOPE_MODULE,"",GET_STR_DEBUG_INFO_EX(d));
                                    }
                                }
                            } else {
                                c_var_string* r=dynamic_cast<c_var_string*>(res);
                                if(!r) {
                                    if(!got_error) {
                                        got_error=true;
                                        e_vartype vartype=VARTYPE_NULL;
                                        if(res) {
                                            vartype=res->get_var_type();
                                            MY_DELETE(res);
                                            res=NULL;
                                        }
                                        string d="C++ dynamic_cast<c_var_string*>(res) returned NULL";
                                        if(!debug.empty()) {
                                            d+=" (";
                                            d+=debug;
                                            d+=")";
                                        }
                                        throw c_var_exception(this_module,__FILE__,__my_func__,c_var_exception::E_DYNCAST,GET_STR_DEBUG_INFO_EX(d),vartype);
                                    }
                                } else
                                    (*r)+=x2;
                            }
                        }
                        break;
                    case 'a':
                        {
                            char x2=0x07;
                            if(res->get_var_type()==VARTYPE_UNSET) {
                                MY_DELETE(res);
                                res=MY_NEW c_var_string();
                            }
                            if(res->get_var_type()!=VARTYPE_STRING) {
                                if(!got_error) {
                                    got_error=true;
                                    string d="OOPS: Occurred error: resulting variable is of type ";
                                    d+=res->get_var_type_as_string();
                                    d+=" and was used for string concatenation";
                                    if(!debug.empty()) {
                                        d+=" (";
                                        d+=debug;
                                        d+=")";
                                    }
                                    if(this_module && this_module->get_logger()) {
                                        this_module->get_logger()->log_print1(E_LOG_DEBUG,LOG_SCOPE_MODULE,"",GET_STR_DEBUG_INFO_EX(d));
                                    }
                                }
                            } else {
                                c_var_string* r=dynamic_cast<c_var_string*>(res);
                                if(!r) {
                                    if(!got_error) {
                                        got_error=true;
                                        e_vartype vartype=VARTYPE_NULL;
                                        if(res) {
                                            vartype=res->get_var_type();
                                            MY_DELETE(res);
                                            res=NULL;
                                        }
                                        string d="C++ dynamic_cast<c_var_string*>(res) returned NULL";
                                        if(!debug.empty()) {
                                            d+=" (";
                                            d+=debug;
                                            d+=")";
                                        }
                                        throw c_var_exception(this_module,__FILE__,__my_func__,c_var_exception::E_DYNCAST,GET_STR_DEBUG_INFO_EX(d),vartype);
                                    }
                                } else
                                    (*r)+=x2;
                            }
                        }
                        break;
                    case 'b':
                        {
                            //res+="\b";
                            char x2='\b';
                            if(res->get_var_type()==VARTYPE_UNSET) {
                                MY_DELETE(res);
                                res=MY_NEW c_var_string();
                            }
                            if(res->get_var_type()!=VARTYPE_STRING) {
                                if(!got_error) {
                                    got_error=true;
                                    string d="OOPS: Occurred error: resulting variable is of type ";
                                    d+=res->get_var_type_as_string();
                                    d+=" and was used for string concatenation";
                                    if(!debug.empty()) {
                                        d+=" (";
                                        d+=debug;
                                        d+=")";
                                    }
                                    if(this_module && this_module->get_logger()) {
                                        this_module->get_logger()->log_print1(E_LOG_DEBUG,LOG_SCOPE_MODULE,"",GET_STR_DEBUG_INFO_EX(d));
                                    }
                                }
                            } else {
                                c_var_string* r=dynamic_cast<c_var_string*>(res);
                                if(!r) {
                                    if(!got_error) {
                                        got_error=true;
                                        e_vartype vartype=VARTYPE_NULL;
                                        if(res) {
                                            vartype=res->get_var_type();
                                            MY_DELETE(res);
                                            res=NULL;
                                        }
                                        string d="C++ dynamic_cast<c_var_string*>(res) returned NULL";
                                        if(!debug.empty()) {
                                            d+=" (";
                                            d+=debug;
                                            d+=")";
                                        }
                                        throw c_var_exception(this_module,__FILE__,__my_func__,c_var_exception::E_DYNCAST,GET_STR_DEBUG_INFO_EX(d),vartype);
                                    }
                                } else
                                    (*r)+=x2;
                            }
                        }
                        break;
                    case 'f':
                        {
                            //res+="\f";
                            char x2='\f';
                            if(res->get_var_type()==VARTYPE_UNSET) {
                                MY_DELETE(res);
                                res=MY_NEW c_var_string();
                            }
                            if(res->get_var_type()!=VARTYPE_STRING) {
                                if(!got_error) {
                                    got_error=true;
                                    string d="OOPS: Occurred error: resulting variable is of type ";
                                    d+=res->get_var_type_as_string();
                                    d+=" and was used for string concatenation";
                                    if(!debug.empty()) {
                                        d+=" (";
                                        d+=debug;
                                        d+=")";
                                    }
                                    if(this_module && this_module->get_logger()) {
                                        this_module->get_logger()->log_print1(E_LOG_DEBUG,LOG_SCOPE_MODULE,"",GET_STR_DEBUG_INFO_EX(d));
                                    }
                                }
                            } else {
                                c_var_string* r=dynamic_cast<c_var_string*>(res);
                                if(!r) {
                                    if(!got_error) {
                                        got_error=true;
                                        e_vartype vartype=VARTYPE_NULL;
                                        if(res) {
                                            vartype=res->get_var_type();
                                            MY_DELETE(res);
                                            res=NULL;
                                        }
                                        string d="C++ dynamic_cast<c_var_string*>(res) returned NULL";
                                        if(!debug.empty()) {
                                            d+=" (";
                                            d+=debug;
                                            d+=")";
                                        }
                                        throw c_var_exception(this_module,__FILE__,__my_func__,c_var_exception::E_DYNCAST,GET_STR_DEBUG_INFO_EX(d),vartype);
                                    }
                                } else
                                    (*r)+=x2;
                            }
                        }
                        break;
                    case 'v':
                        {
                            //res+="\v";
                            char x2='\v';
                            if(res->get_var_type()==VARTYPE_UNSET) {
                                MY_DELETE(res);
                                res=MY_NEW c_var_string();
                            }
                            if(res->get_var_type()!=VARTYPE_STRING) {
                                if(!got_error) {
                                    got_error=true;
                                    string d="OOPS: Occurred error: resulting variable is of type ";
                                    d+=res->get_var_type_as_string();
                                    d+=" and was used for string concatenation";
                                    if(!debug.empty()) {
                                        d+=" (";
                                        d+=debug;
                                        d+=")";
                                    }
                                    if(this_module && this_module->get_logger()) {
                                        this_module->get_logger()->log_print1(E_LOG_DEBUG,LOG_SCOPE_MODULE,"",GET_STR_DEBUG_INFO_EX(d));
                                    }
                                }
                            } else {
                                c_var_string* r=dynamic_cast<c_var_string*>(res);
                                if(!r) {
                                    if(!got_error) {
                                        got_error=true;
                                        e_vartype vartype=VARTYPE_NULL;
                                        if(res) {
                                            vartype=res->get_var_type();
                                            MY_DELETE(res);
                                            res=NULL;
                                        }
                                        string d="C++ dynamic_cast<c_var_string*>(res) returned NULL";
                                        if(!debug.empty()) {
                                            d+=" (";
                                            d+=debug;
                                            d+=")";
                                        }
                                        throw c_var_exception(this_module,__FILE__,__my_func__,c_var_exception::E_DYNCAST,GET_STR_DEBUG_INFO_EX(d),vartype);
                                    }
                                } else
                                    (*r)+=x2;
                            }
                        }
                        break;
                    case '\'':
                        {
                            //res+="\'";
                            char x2='\'';
                            if(res->get_var_type()==VARTYPE_UNSET) {
                                MY_DELETE(res);
                                res=MY_NEW c_var_string();
                            }
                            if(res->get_var_type()!=VARTYPE_STRING) {
                                if(!got_error) {
                                    got_error=true;
                                    string d="OOPS: Occurred error: resulting variable is of type ";
                                    d+=res->get_var_type_as_string();
                                    d+=" and was used for string concatenation";
                                    if(!debug.empty()) {
                                        d+=" (";
                                        d+=debug;
                                        d+=")";
                                    }
                                    if(this_module && this_module->get_logger()) {
                                        this_module->get_logger()->log_print1(E_LOG_DEBUG,LOG_SCOPE_MODULE,"",GET_STR_DEBUG_INFO_EX(d));
                                    }
                                }
                            } else {
                                c_var_string* r=dynamic_cast<c_var_string*>(res);
                                if(!r) {
                                    if(!got_error) {
                                        got_error=true;
                                        e_vartype vartype=VARTYPE_NULL;
                                        if(res) {
                                            vartype=res->get_var_type();
                                            MY_DELETE(res);
                                            res=NULL;
                                        }
                                        string d="C++ dynamic_cast<c_var_string*>(res) returned NULL";
                                        if(!debug.empty()) {
                                            d+=" (";
                                            d+=debug;
                                            d+=")";
                                        }
                                        throw c_var_exception(this_module,__FILE__,__my_func__,c_var_exception::E_DYNCAST,GET_STR_DEBUG_INFO_EX(d),vartype);
                                    }
                                } else
                                    (*r)+=x2;
                            }
                        }
                        break;
                    case '?':
                        {
                            //res+="?";
                            char x2='?';
                            if(res->get_var_type()==VARTYPE_UNSET) {
                                MY_DELETE(res);
                                res=MY_NEW c_var_string();
                            }
                            if(res->get_var_type()!=VARTYPE_STRING) {
                                if(!got_error) {
                                    got_error=true;
                                    string d="OOPS: Occurred error: resulting variable is of type ";
                                    d+=res->get_var_type_as_string();
                                    d+=" and was used for string concatenation";
                                    if(!debug.empty()) {
                                        d+=" (";
                                        d+=debug;
                                        d+=")";
                                    }
                                    if(this_module && this_module->get_logger()) {
                                        this_module->get_logger()->log_print1(E_LOG_DEBUG,LOG_SCOPE_MODULE,"",GET_STR_DEBUG_INFO_EX(d));
                                    }
                                }
                            } else {
                                c_var_string* r=dynamic_cast<c_var_string*>(res);
                                if(!r) {
                                    if(!got_error) {
                                        got_error=true;
                                        e_vartype vartype=VARTYPE_NULL;
                                        if(res) {
                                            vartype=res->get_var_type();
                                            MY_DELETE(res);
                                            res=NULL;
                                        }
                                        string d="C++ dynamic_cast<c_var_string*>(res) returned NULL";
                                        if(!debug.empty()) {
                                            d+=" (";
                                            d+=debug;
                                            d+=")";
                                        }
                                        throw c_var_exception(this_module,__FILE__,__my_func__,c_var_exception::E_DYNCAST,GET_STR_DEBUG_INFO_EX(d),vartype);
                                    }
                                } else
                                    (*r)+=x2;
                            }
                        }
                        break;
                    case '$':
                        {
                            //res+="$";
                            char x2='$';
                            if(res->get_var_type()==VARTYPE_UNSET) {
                                MY_DELETE(res);
                                res=MY_NEW c_var_string();
                            }
                            if(res->get_var_type()!=VARTYPE_STRING) {
                                if(!got_error) {
                                    got_error=true;
                                    string d="OOPS: Occurred error: resulting variable is of type ";
                                    d+=res->get_var_type_as_string();
                                    d+=" and was used for string concatenation";
                                    if(!debug.empty()) {
                                        d+=" (";
                                        d+=debug;
                                        d+=")";
                                    }
                                    if(this_module && this_module->get_logger()) {
                                        this_module->get_logger()->log_print1(E_LOG_DEBUG,LOG_SCOPE_MODULE,"",GET_STR_DEBUG_INFO_EX(d));
                                    }
                                }
                            } else {
                                c_var_string* r=dynamic_cast<c_var_string*>(res);
                                if(!r) {
                                    if(!got_error) {
                                        got_error=true;
                                        e_vartype vartype=VARTYPE_NULL;
                                        if(res) {
                                            vartype=res->get_var_type();
                                            MY_DELETE(res);
                                            res=NULL;
                                        }
                                        string d="C++ dynamic_cast<c_var_string*>(res) returned NULL";
                                        if(!debug.empty()) {
                                            d+=" (";
                                            d+=debug;
                                            d+=")";
                                        }
                                        throw c_var_exception(this_module,__FILE__,__my_func__,c_var_exception::E_DYNCAST,GET_STR_DEBUG_INFO_EX(d),vartype);
                                    }
                                } else
                                    (*r)+=x2;
                            }
                        }
                        break;
                    case 'x':
                        {
                            if(i1+2>=str.length()) {
                                string d="OOPS: Occurred error: erroneous escape character '\\x' at the end of the string \"";
                                d+=str+"\"";
                                if(!debug.empty()) {
                                    d+=" (";
                                    d+=debug;
                                    d+=")";
                                }
                                if(this_module && this_module->get_logger()) {
                                    this_module->get_logger()->log_print1(E_LOG_DEBUG,LOG_SCOPE_MODULE,"",GET_STR_DEBUG_INFO_EX(d));
                                }
                            } else {
                                unsigned char x1[8];
                                x1[0]=static_cast<unsigned char>(str[++i1]);
                                x1[1]=static_cast<unsigned char>(str[++i1]);
                                unsigned char x2=0;
                                for(int i1=0; i1<2; ++i1) {
                                    unsigned char nibble=0xff;
                                    if(x1[i1]>='0' && x1[i1]<='9')
                                        nibble=x1[i1]-'0';
                                    if(x1[i1]>='a' && x1[i1]<='f')
                                        nibble=x1[i1]+10-'a';
                                    if(x1[i1]>='A' && x1[i1]<='F')
                                        nibble=x1[i1]+10-'A';
                                    if(nibble==0xff) {
                                        nibble=0;
                                        string d="OOPS: Occurred error: erroneous escape character '\\x' - character not in HEX family \"";
                                        d+=str+"\"";
                                        if(!debug.empty()) {
                                            d+=" (";
                                            d+=debug;
                                            d+=")";
                                        }
                                        if(this_module && this_module->get_logger()) {
                                            this_module->get_logger()->log_print1(E_LOG_DEBUG,LOG_SCOPE_MODULE,"",GET_STR_DEBUG_INFO_EX(d));
                                        }
                                    }
                                    x2|=nibble << (i1==0?4U:0U);
                                }
                                //res.push_back(x2);
                                if(res->get_var_type()==VARTYPE_UNSET) {
                                    MY_DELETE(res);
                                    res=MY_NEW c_var_string();
                                }
                                if(res->get_var_type()!=VARTYPE_STRING) {
                                    if(!got_error) {
                                        got_error=true;
                                        string d="OOPS: Occurred error: resulting variable is of type ";
                                        d+=res->get_var_type_as_string();
                                        d+=" and was used for string concatenation";
                                        if(!debug.empty()) {
                                            d+=" (";
                                            d+=debug;
                                            d+=")";
                                        }
                                        if(this_module && this_module->get_logger()) {
                                            this_module->get_logger()->log_print1(E_LOG_DEBUG,LOG_SCOPE_MODULE,"",GET_STR_DEBUG_INFO_EX(d));
                                        }
                                    }
                                } else {
                                    c_var_string* r=dynamic_cast<c_var_string*>(res);
                                    if(!r) {
                                        if(!got_error) {
                                            got_error=true;
                                            e_vartype vartype=VARTYPE_NULL;
                                            if(res) {
                                                vartype=res->get_var_type();
                                                MY_DELETE(res);
                                                res=NULL;
                                            }
                                            string d="C++ dynamic_cast<c_var_string*>(res) returned NULL";
                                            if(!debug.empty()) {
                                                d+=" (";
                                                d+=debug;
                                                d+=")";
                                            }
                                            throw c_var_exception(this_module,__FILE__,__my_func__,c_var_exception::E_DYNCAST,GET_STR_DEBUG_INFO_EX(d),vartype);
                                        }
                                    } else
                                        (*r)+=x2;
                                }
                            }
                        }
                        break;
                    case '0':
                        if(i1+1>=str.length() || !(str[i1+1]>='0' && str[i1+1]<='9')) {
                            unsigned char x2=0;
                            c_var_string* r=dynamic_cast<c_var_string*>(res);
                            if(!r) {
                                if(!got_error) {
                                    got_error=true;
                                    e_vartype vartype=VARTYPE_NULL;
                                    if(res) {
                                        vartype=res->get_var_type();
                                        MY_DELETE(res);
                                        res=NULL;
                                    }
                                    string d="C++ dynamic_cast<c_var_string*>(res) returned NULL";
                                    if(!debug.empty()) {
                                        d+=" (";
                                        d+=debug;
                                        d+=")";
                                    }
                                    throw c_var_exception(this_module,__FILE__,__my_func__,c_var_exception::E_DYNCAST,GET_STR_DEBUG_INFO_EX(d),vartype);
                                }
                            } else
                                (*r)+=x2;
                            break;
                        }
                    case '1':
                    case '2':
                    case '3':
                    case '4':
                    case '5':
                    case '6':
                    case '7':
                        {
                            if(i1+3>=str.length()) {
                                string d="OOPS: Occurred error: erroneous escape character '\\<octal_num>' at the end of the string \"";
                                d+=str+"\"";
                                if(!debug.empty()) {
                                    d+=" (";
                                    d+=debug;
                                    d+=")";
                                }
                                if(this_module && this_module->get_logger()) {
                                    this_module->get_logger()->log_print1(E_LOG_DEBUG,LOG_SCOPE_MODULE,"",GET_STR_DEBUG_INFO_EX(d));
                                }
                            } else {
                                unsigned char x1[8];
                                x1[0]=x;
                                x1[1]=static_cast<unsigned char>(str[++i1]);
                                x1[2]=static_cast<unsigned char>(str[++i1]);
                                unsigned char x2=0;
                                for(int i1=0; i1<3; ++i1) {
                                    unsigned char nibble=0xff;
                                    if(x1[i1]>='0' && x1[i1]<='7')
                                        nibble=x1[i1]-'0';
                                    if(nibble==0xff) {
                                        nibble=0;
                                        string d="OOPS: Occurred error: erroneous escape character '\\<octal_num>' - character not in OCTAL family \"";
                                        d+=str+"\"";
                                        if(!debug.empty()) {
                                            d+=" (";
                                            d+=debug;
                                            d+=")";
                                        }
                                        if(this_module && this_module->get_logger()) {
                                            this_module->get_logger()->log_print1(E_LOG_DEBUG,LOG_SCOPE_MODULE,"",GET_STR_DEBUG_INFO_EX(d));
                                        }
                                    }
                                    if(i1==0)
                                        nibble<<=6;
                                    if(i1==1)
                                        nibble<<=3;
                                    /*if(i1==2)
                                        nibble<<=0;*/
                                    x2|=nibble;
                                }
                                //res.push_back(x2);
                                if(res->get_var_type()==VARTYPE_UNSET) {
                                    MY_DELETE(res);
                                    res=MY_NEW c_var_string();
                                }
                                if(res->get_var_type()!=VARTYPE_STRING) {
                                    if(!got_error) {
                                        got_error=true;
                                        string d="OOPS: Occurred error: resulting variable is of type ";
                                        d+=res->get_var_type_as_string();
                                        d+=" and was used for string concatenation";
                                        if(!debug.empty()) {
                                            d+=" (";
                                            d+=debug;
                                            d+=")";
                                        }
                                        if(this_module && this_module->get_logger()) {
                                            this_module->get_logger()->log_print1(E_LOG_DEBUG,LOG_SCOPE_MODULE,"",GET_STR_DEBUG_INFO_EX(d));
                                        }
                                    }
                                } else {
                                    c_var_string* r=dynamic_cast<c_var_string*>(res);
                                    if(!r) {
                                        if(!got_error) {
                                            got_error=true;
                                            e_vartype vartype=VARTYPE_NULL;
                                            if(res) {
                                                vartype=res->get_var_type();
                                                MY_DELETE(res);
                                                res=NULL;
                                            }
                                            string d="C++ dynamic_cast<c_var_string*>(res) returned NULL";
                                            if(!debug.empty()) {
                                                d+=" (";
                                                d+=debug;
                                                d+=")";
                                            }
                                            throw c_var_exception(this_module,__FILE__,__my_func__,c_var_exception::E_DYNCAST,GET_STR_DEBUG_INFO_EX(d),vartype);
                                        }
                                    } else
                                        (*r)+=x2;
                                }
                            }
                        }
                        break;
                    default:
                        {
                            string d="OOPS: Occurred error: erroneous escape character '\\' - unknown next character \"";
                            d+=str+"\"";
                            if(!debug.empty()) {
                                d+=" (";
                                d+=debug;
                                d+=")";
                            }
                            if(this_module && this_module->get_logger()) {
                                this_module->get_logger()->log_print1(E_LOG_DEBUG,LOG_SCOPE_MODULE,"",GET_STR_DEBUG_INFO_EX(d));
                            }
                        }
                        break;
                }
            }
            continue;
        }
        string v="";
        if(str[i1]=='$') {
            v="$";
            for(i1++; i1<str.length(); ++i1) {
                if(!((str[i1]>='a' && str[i1]<='z') || (str[i1]>='A' && str[i1]<='Z') || (str[i1]>='0' && str[i1]<='9') || str[i1]=='_'))
                    break;
                v+=str[i1];
            }
            //i1--; // later!...
            c_var_base* var=NULL;
            try {
                c_index_str idx(v);
                if(NULL!=vars)
                    var=vars->get(this_module,&idx,false,debug);
                else
                    var=NULL;
                if(!var) {
                    string d="Error while evaluating variable \"";
                    d+=v;
                    d+="\" - variable is UNSET";
                    var_oops(this_module,__FILE__,__my_func__,GET_STR_DEBUG_INFO_EX(d),debug);
                    if(res)
                        MY_DELETE(res);
                    return_ref=false;
                    return MY_NEW c_var_unset();
                }
            } catch(c_var_exception e) {
                e.print_oops_to_log(debug);
                if(res)
                    MY_DELETE(res);
                return_ref=false;
                return MY_NEW c_var_unset();
            }
            if(var!=NULL) {
                if(i1==str.length() && (!res || res->get_var_type()==VARTYPE_UNSET)) {
                    if(res)
                        MY_DELETE(res);
                    res=NULL;
                    res=var->copy();
                    if('-'==unary && res->get_var_type()==VARTYPE_INT) {
                        c_var_int* i=dynamic_cast<c_var_int*>(res);
                        if(!i) {
                            e_vartype vartype=VARTYPE_NULL;
                            if(res) {
                                vartype=res->get_var_type();
                                MY_DELETE(res);
                                res=NULL;
                            }
                            return_ref=false;
                            string d="C++ dynamic_cast<c_var_int*>(res) returned NULL";
                            if(!debug.empty()) {
                                d+=" (";
                                d+=debug;
                                d+=")";
                            }
                            throw c_var_exception(this_module,__FILE__,__my_func__,c_var_exception::E_DYNCAST,GET_STR_DEBUG_INFO_EX(d),vartype);
                        } else {
                            i->set(-i->get());
                            return_ref=false;
                            return res;
                        }
                    } else if('-'==unary) {
                        var_oops(this_module,__FILE__,__my_func__,GET_STR_DEBUG_INFO_EX("Invalid usage of unary operator"),debug);
                        if(res)
                            MY_DELETE(res);
                        res=NULL;
                        return_ref=false;
                        return MY_NEW c_var_unset();
                    }
                    if(return_ref) {
                        if(res)
                            MY_DELETE(res);
                        res=NULL;
                        return_ref=true;
                        return var;
                    }
                    return_ref=false;
                    return res;
                }
                if(!res || res->get_var_type()==VARTYPE_UNSET) {
                    if(res)
                        MY_DELETE(res);
                    res=MY_NEW c_var_string("");
                }
                {
                    if(!res || res->get_var_type()!=VARTYPE_STRING) {
                        got_error=true;
                        string d="OOPS: Occurred error: invalid expression to evaluate";
                        if(!debug.empty()) {
                            d+=" (";
                            d+=debug;
                            d+=")";
                        }
                        if(this_module && this_module->get_logger()) {
                            this_module->get_logger()->log_print1(E_LOG_DEBUG,LOG_SCOPE_MODULE,"",GET_STR_DEBUG_INFO_EX(d));
                        }
                        return_ref=false;
                        return res;
                    }
                    c_var_string* r=dynamic_cast<c_var_string*>(res);
                    if(!r) {
                        if(!got_error) {
                            got_error=true;
                            e_vartype vartype=VARTYPE_NULL;
                            if(res) {
                                vartype=res->get_var_type();
                                MY_DELETE(res);
                                res=NULL;
                            }
                            string d="C++ dynamic_cast<c_var_string*>(res) returned NULL";
                            if(!debug.empty()) {
                                d+=" (";
                                d+=debug;
                                d+=")";
                            }
                            throw c_var_exception(this_module,__FILE__,__my_func__,c_var_exception::E_DYNCAST,GET_STR_DEBUG_INFO_EX(d),vartype);
                        }
                    } else
                        (*r)+=var->to_string(this_module);

                    if(i1==str.length()) {
                        if('-'==unary && res->get_var_type()==VARTYPE_INT) {
                            c_var_int* i=dynamic_cast<c_var_int*>(res);
                            if(!i) {
                                e_vartype vartype=VARTYPE_NULL;
                                if(res) {
                                    vartype=res->get_var_type();
                                    MY_DELETE(res);
                                    res=NULL;
                                }
                                return_ref=false;
                                string d="C++ dynamic_cast<c_var_int*>(res) returned NULL";
                                if(!debug.empty()) {
                                    d+=" (";
                                    d+=debug;
                                    d+=")";
                                }
                                throw c_var_exception(this_module,__FILE__,__my_func__,c_var_exception::E_DYNCAST,GET_STR_DEBUG_INFO_EX(d),vartype);
                            } else {
                                i->set(-i->get());
                                return_ref=false;
                                return res;
                            }
                        } else if('-'==unary) {
                            var_oops(this_module,__FILE__,__my_func__,GET_STR_DEBUG_INFO_EX("Invalid usage of unary operator"),debug);
                            if(res)
                                MY_DELETE(res);
                            res=NULL;
                            return_ref=false;
                            return MY_NEW c_var_unset();
                        }
                        return_ref=false;
                        return res;
                    }

                    i1--; // ...here!
                    continue;
                }
            } else {
                if(!got_error) {
                    got_error=true;
                    string d="OOPS: Occurred error: unknown variable ";
                    d+=v;
                    if(!debug.empty()) {
                        d+=" (";
                        d+=debug;
                        d+=")";
                    }
                    if(this_module && this_module->get_logger()) {
                        this_module->get_logger()->log_print1(E_LOG_DEBUG,LOG_SCOPE_MODULE,"",GET_STR_DEBUG_INFO_EX(d));
                    }
                }
                return_ref=false;
                return res;
            }
            // NOTREACH
        }
        if(!b_str && '.'==str[i1]) {
            if(res->get_var_type()==VARTYPE_UNSET) {
                MY_DELETE(res);
                res=MY_NEW c_var_string();
            }
            continue;
        }
        if(b_str) {
            //res+=str[i1];
            if(res->get_var_type()==VARTYPE_UNSET) {
                MY_DELETE(res);
                res=MY_NEW c_var_string();
            }
            if(res->get_var_type()!=VARTYPE_STRING) {
                if(!got_error) {
                    got_error=true;
                    string d="OOPS: Occurred error: resulting variable is of type ";
                    d+=res->get_var_type_as_string();
                    d+=" and was used for string concatenation";
                    if(!debug.empty()) {
                        d+=" (";
                        d+=debug;
                        d+=")";
                    }
                    if(this_module && this_module->get_logger()) {
                        this_module->get_logger()->log_print1(E_LOG_DEBUG,LOG_SCOPE_MODULE,"",GET_STR_DEBUG_INFO_EX(d));
                    }
                }
            } else {
                c_var_string* r=dynamic_cast<c_var_string*>(res);
                if(!r) {
                    if(!got_error) {
                        got_error=true;
                        e_vartype vartype=VARTYPE_NULL;
                        if(res) {
                            vartype=res->get_var_type();
                            MY_DELETE(res);
                            res=NULL;
                        }
                        string d="C++ dynamic_cast<c_var_string*>(res) returned NULL";
                        if(!debug.empty()) {
                            d+=" (";
                            d+=debug;
                            d+=")";
                        }
                        throw c_var_exception(this_module,__FILE__,__my_func__,c_var_exception::E_DYNCAST,GET_STR_DEBUG_INFO_EX(d),vartype);
                    }
                } else
                    (*r)+=str[i1];
            }
            continue;
        }
        if(!got_error) {
            if(res)
                MY_DELETE(res);
            res=NULL;
            got_error=true;
            string d="WARNING: Occurred error: Syntax error";
            if(!debug.empty()) {
                d+=" (";
                d+=debug;
                d+=")";
            }
            if(this_module && this_module->get_logger()) {
                this_module->get_logger()->log_print1(E_LOG_DEBUG,LOG_SCOPE_MODULE,"",GET_STR_DEBUG_INFO_EX(d));
            }
            return_ref=false;
            return NULL;
        }
    }

    if(is_number && res->get_var_type()==VARTYPE_STRING) {
        c_var_string* s=dynamic_cast<c_var_string*>(res);
        if(!s) {
            e_vartype vartype=VARTYPE_NULL;
            if(res) {
                vartype=res->get_var_type();
                MY_DELETE(res);
                res=NULL;
            }
            string d="C++ dynamic_cast<c_var_string*>(res) returned NULL";
            if(!debug.empty()) {
                d+=" (";
                d+=debug;
                d+=")";
            }
            throw c_var_exception(this_module,__FILE__,__my_func__,c_var_exception::E_DYNCAST,GET_STR_DEBUG_INFO_EX(d),vartype);
        } else {
            c_var_int* i=s->to_int_clever(this_module,c_var_string::RET_NULL,debug);
            if(res)
                MY_DELETE(res);
            res=i;
        }
    }

    if(res && unary && res->get_var_type()!=VARTYPE_INT) {
        var_oops(this_module,__FILE__,__my_func__,GET_STR_DEBUG_INFO_EX("Invalid usage of unary operator"),debug);
    }
    if(res && '-'==unary && res->get_var_type()==VARTYPE_INT) {
        c_var_int* i=dynamic_cast<c_var_int*>(res);
        if(!i) {
            e_vartype vartype=VARTYPE_NULL;
            if(res) {
                vartype=res->get_var_type();
                MY_DELETE(res);
                res=NULL;
            }
            string d="C++ dynamic_cast<c_var_int*>(res) returned NULL";
            if(!debug.empty()) {
                d+=" (";
                d+=debug;
                d+=")";
            }
            throw c_var_exception(this_module,__FILE__,__my_func__,c_var_exception::E_DYNCAST,GET_STR_DEBUG_INFO_EX(d),vartype);
        } else {
            i->set(-i->get());
        }
    }

    return_ref=false;
    return res;
}

#if defined(_MSC_VER)
#   pragma warning(pop)
#endif

void c_var_exception::print_oops_to_log(const string& debug)
{
    var_fatal(this_module_,file_,func_,this->get_msg(),debug);
}

c_var_exception::c_var_exception(const c_var_exception& orig)
{
    error_=orig.error_;
    reason_=orig.reason_;
    var_type_=orig.var_type_;
    func_=orig.func_;
    this_module_=orig.this_module_;
    file_=orig.file_;
}

c_var_exception::c_var_exception(c_module* this_module, const char* file, const char* func, e_error_type error, string reason)
{
    error_=error;
    reason_=reason;
    var_type_=VARTYPE_NA;
    func_=func;
    file_=file;
    this_module_=this_module;
}

c_var_exception::c_var_exception(c_module* this_module, const char* file, const char* func, e_error_type error, string reason, const c_var_base* var_type)
{
    error_=error;
    reason_=reason;
    var_type_=VARTYPE_NULL;
    if(var_type)
        var_type_=var_type->get_var_type();
    func_=func;
    file_=file;
    this_module_=this_module;
}

c_var_exception::c_var_exception(c_module* this_module, const char* file, const char* func, e_error_type error, string reason, e_vartype var_type)
{
    error_=error;
    reason_=reason;
    var_type_=var_type;
    func_=func;
    file_=file;
    this_module_=this_module;
}

void c_var_exception::prefix_reason(string prefix)
{
    reason_=prefix+reason_;
}

string c_var_exception::get_msg()
{
    string res="C++ Function=";
    res+=file_;
    res+="::";
    res+=func_;
    res+="(), variable type=";
    res+=vartype_to_string(var_type_);
    res+=", error type=";
    switch(error_) {
        case E_INDEX_NULL:
            res+="E_INDEX_NULL(=index is NULL)";
            break;
        case E_INDEX_INV:
            res+="E_INDEX_INV(=invalid index, not INT, nor STRING)";
            break;
        case E_DYNCAST:
            res+="E_DYNCAST(=dynamic_cast operator returned NULL)";
            break;
        case E_NIMPL:
            res+="E_NIMPL(=not implemented)";
            break;
        case E_INTERNAL:
            res+="E_INTERNAL(=sanity check failed)";
            break;
        case E_READONLY:
            res+="E_READONLY(=attemp to write to read-only variable)";
            break;
        case E_INDEX_RANGE:
            res+="E_INDEX_RANGE(=index out of range)";
            break;
        case E_INDEX_NOT_INT:
            res+="E_INDEX_NOT_INT(=index to STRING as base variable is not INT)";
            break;
        case E_INVALID_OPER:
            res+="E_INVALID_OPER(=invalid operator, or expecting operator)";
            break;
        case E_SYNTAX:
            res+="E_SYNTAX(=invalid syntax)";
            break;
        case E_USER_ERROR:
            res+="E_USER_ERROR(=error while evaluating expression)";
            break;
        case E_LVALUE:
            res+="E_LVALUE(=l-value required)";
            break;
        case E_FUNC_CALL:
            res+="E_FUNC_CALL(=function call)";
            break;
        case E_DIV_ZERO:
            res+="E_DIV_ZERO(=division by zero)";
            break;
        case E_TIMEOUT:
            res+="E_TIMEOUT(=script execution timeout)";
            break;
        default:
            res+="(other)";
            break;
    }
    res+=", reason=";
    res+=reason_;
    return res;
}

void array_log_fn(c_var_keeper::e_log_severity UNUSEDARG(severity), const string& msg, void* user_data)
{
    c_module* mod=(c_module*)user_data;
    if(!mod)
        mod=core_mod;
    if(mod)
        var_oops(mod,__FILE__,__my_func__,GET_STR_DEBUG_INFO_EX(msg),"");
}

/*!
    \brief Gets explanation of return code
    \author VooDooMan
    \version 1
    \date 2005
    \param e Return code
*/
string exp_result_to_string(e_exp_result e)
{
    switch(e) {
        case EXP_OK:
            return (string)"OK";
        case EXP_E_PARENTHESIS:
            return (string)"Unmatched parenthesis";
        case EXP_E_UNKNOWN_OPER:
            return (string)"Unknown operator";
        case EXP_E_INTERNAL_1:
            return (string)"Internal error #1";
        case EXP_E_INTERNAL_2:
            return (string)"Internal error #2";
        case EXP_E_INTERNAL_3:
            return (string)"Internal error #3";
        case EXP_E_INTERNAL_4:
            return (string)"Internal error #4";
        case EXP_E_INTERNAL_5:
            return (string)"Internal error #5";
        case EXP_E_SUBEXP:
            return (string)"Subexpression in parenthesis '(...)' evaluation, or 'x[...]' evaluation failed";
        case EXP_E_SYNTAX_CAST:
            return (string)"Error in syntax - missing operator before explicit type cast, e.g. \"123(int)456\"";
        case EXP_E_SYNTAX_2:
            return (string)"Error in syntax - missing operator before '!' or '~' operator, e.g. \"123!456\"";
        case EXP_E_INVALID_OPER_USAGE_1:
            return (string)"Invalid usage of x++ operator: need variable, not constant";
        case EXP_E_INVALID_OPER_USAGE_2:
            return (string)"Invalid usage of x-- operator: need variable, not constant";
        case EXP_E_TIMEOUT:
            return (string)"Script execution timeout";
        case EXP_E_STACK_OVERFLOW:
            return (string)"Maximum stack size reached (too many nested iterations)";
        default:
            break;
    }
    return (string)"Other error";
}

/*!
    \brief Evaluates index
    \author VooDooMan
    \version 1
    \date 2005
    \param this_module Module
    \param base Base variable
    \param index Index
    \param insensitive Set to true if we want case insensitivity
    \param debug Debug string
*/
// returns reference or NULL; if base is of type VARTYPE_STRING, it must be MY_DELETE'd
c_var_base* var_eval_index(c_module* this_module, c_var_base* base, c_index_int*& index, bool insensitive, const string& debug)
{
    if(!base) {
        var_oops(this_module,__FILE__,__my_func__,GET_STR_DEBUG_INFO_EX("argument \"base\" is NULL"),debug);
        MY_DELETE(index);
        index=NULL;
        return NULL;
    }
    switch(base->get_var_type()) {
        case VARTYPE_STRING:
            {
                c_var_string* base_s=dynamic_cast<c_var_string*>(base);
                if(!base_s) {
                    MY_DELETE(index);
                    index=NULL;
                    var_oops(this_module,__FILE__,__my_func__,GET_STR_DEBUG_INFO_EX("dynamic_cast<c_var_string*>(base) returned NULL"),debug);
                    return NULL;
                }
                try {
                    c_var_base* res=MY_NEW c_var_string(base_s->get(this_module,index,debug));
                    index=NULL;
                    return res;
                } catch(c_var_exception e) {
                    MY_DELETE(index);
                    index=NULL;
                    e.print_oops_to_log(debug);
                    return NULL;
                }
            }
            // NOTREACH
        case VARTYPE_ARRAY:
            {
                c_var_array* base_a=dynamic_cast<c_var_array*>(base);
                if(!base_a) {
                    var_oops(this_module,__FILE__,__my_func__,GET_STR_DEBUG_INFO_EX("dynamic_cast<c_var_array*>(base) returned NULL"),debug);
                    return NULL;
                }
                try {
                    c_index_base* i=(c_index_base*)index;
                    c_var_base* res=base_a->get(this_module,i,insensitive,debug);
                    index=NULL;
                    return res;
                } catch(c_var_exception e) {
                    MY_DELETE(index);
                    index=NULL;
                    e.print_oops_to_log(debug);
                    return NULL;
                }
            }
            // NOTREACH
        default:
            {
                string d="index validation error: base variable is not STRING, nor ARRAY, it is of type ";
                d+=base->get_var_type_as_string();
                var_oops(this_module,__FILE__,__my_func__,GET_STR_DEBUG_INFO_EX(d),debug);
                MY_DELETE(index);
                index=NULL;
                return NULL;
            }
            // NOTREACH
    }
    // NOTREACH
}

/*!
    \brief Checks if it is direct variable name, or indexed
    \author VooDooMan
    \version 1
    \date 2005-09-13
    \param var Variable name
    \retval true Variable is in "$var["a"]" style (indexed)
    \retval false Variable is not indexed: "$var" style
*/
bool var_is_indexed(string var)
{
    return var.find("[",0)!=string::npos;
}

/*!
    \brief Returns string type of variable
    \author VooDooMan
    \version 1
    \date 2005
    \param vartype Type of variable
*/
string vartype_to_string(e_vartype vartype)
{
    switch(vartype) {
        case VARTYPE_NULL:
            return (string)"(NULL)";
        case VARTYPE_NA:
            return (string)"(N/A => not relevant)";
        case VARTYPE_UNSET:
            return (string)"UNSET";
        case VARTYPE_STRING:
            return (string)"STRING";
        case VARTYPE_ARRAY:
            return (string)"ARRAY";
        case VARTYPE_INT:
            return (string)"INT";
        case VARTYPE_BOOL:
            return (string)"BOOL";
    }
    return (string)"(INVALID)";
}

/*!
    \brief Evaluates variable to string
    \author VooDooMan
    \version 2
    \date 2005, 2007-11-05
    \param str Name of variable beginning with '$' sign
    \param vars Array of variables ant their values (can be NULL)
    \param debug Debug string in case of error
    \param this_module Module
*/
string var_eval_string_simple(c_module* this_module, const string& str, c_var_array* vars, const string& debug)
{
    bool return_ref=true;
    string result;
    c_var_base* res=var_eval_obj(this_module,str,vars,debug,return_ref);
    if(!res) {
        string d="OOPS: Occurred error: resulting object is NULL - returning an empty string, expression was: \"";
        d+=str;
        d+="\"";
        if(!debug.empty()) {
            d+=" (";
            d+=debug;
            d+=")";
        }
        if(this_module && this_module->get_logger()) {
            this_module->get_logger()->log_print1(E_LOG_DEBUG,LOG_SCOPE_MODULE,"",GET_STR_DEBUG_INFO_EX(d));
        }
    } else {
        result=res->to_string(this_module);
        if(!return_ref) {
            MY_DELETE(res);
        }
        res=NULL;
    }
    return result;
}

c_var_array* var_decode_array_for_botnet(c_module* this_module, char* data, size_t &pos, size_t size);

/*!
    \author VooDooMan
    \version 1
    \date 2005
    \brief Decodes binary-encoded variable (for BOTNET remote call)
    \param data Buffer
    \param pos Position in buffer
    \param size Size of buffer
    \return Returns NULL in case of error
    \param this_module Module
*/
c_var_base* var_decode_entity_for_botnet(c_module* this_module, char* data, size_t &pos, size_t size)
{
    if(pos==size)
        return NULL;

    if(pos>size) {
error:
        return NULL;
    }

    switch(data[pos]) {
        case 'B':
            {
                if(pos+1+1>size)
                    goto error;
                char xx=data[++pos];
                pos++;
                c_var_bool* x=MY_NEW c_var_bool(xx==0?false:true);
                return x;
            }
            // NOTREACH
        case 'I':
            {
                if(pos+1+1>size)
                    goto error;
                unsigned char len=data[++pos];
                if(len>256-4) // future compatibility: for sure's sake
                    goto error;
                if(0==len || len%sizeof(uint32_t)!=0) // sanity check
                    goto error;
                if(pos+1+len>size)
                    goto error;
                pos++;
                unsigned char tmp[256];
                memcpy(tmp,&(data[pos]),len);
                MY_LONG_LONG xx=0;
                unsigned char our_len=static_cast<unsigned char>((len>sizeof(xx))?sizeof(xx):len);
                if(our_len<len) {
                    // check whether we accept larger number
                    // tmp is in little endian state event on big endian machines
                    bool bad=false;
                    if(!(tmp[len-1] & 0x80)) {
                        // positive number
                        for(size_t i1=our_len; i1<len; ++i1) {
                            if(0!=tmp[i1]) {
                                bad=true;
                                break;
                            }
                        }
                    } else {
                        // negative number
                        if( (tmp[our_len-1] & 0x80) == 0)
                            bad=true;
                        else {
                            for(size_t i1=our_len; i1<len; ++i1) {
                                if(0xff!=tmp[i1]) {
                                    bad=true;
                                    break;
                                }
                            }
                        }
                    }
                    if(bad) {
                        if(this_module && this_module->get_logger()) {
                            string log="Remote bot sent packet for remote procedure call, and INT variable is larger on remote side and couln't fit into our size of INT. Rejecting remote procedure call request.";
                            this_module->get_logger()->log_print1(E_LOG_BOTNET,LOG_SCOPE_MODULE,"",log);
                            this_module->get_logger()->log_print1(E_LOG_BOTNET_DEBUG,LOG_SCOPE_MODULE,"",log);
                            this_module->get_logger()->log_print1(E_LOG_BOTNET,LOG_SCOPE_MODULE,"",log);
                            this_module->get_logger()->log_print1(E_LOG_BOT,LOG_SCOPE_MODULE,"",log);
                            this_module->get_logger()->log_print1(E_LOG_CUSTOM,LOG_SCOPE_MODULE,"",log);
                        }
                        return NULL;
                    }
                }
                memcpy(&xx,tmp,our_len);
#ifdef WORDS_BIGENDIAN
                byte_stream_reverse(&xx,our_len);
#endif
                pos+=len;
                pos++;
                c_var_int* x=MY_NEW c_var_int(xx);
                return x;
            }
            // NOTREACH
        case 'U':
            {
                if(pos+1>size)
                    goto error;
                pos++;
                pos++;
                c_var_unset* x=MY_NEW c_var_unset();
                return x;
            }
            // NOTREACH
        case 'S':
            {
                if(pos+sizeof(uint32)+1>size)
                    goto error;

                uint32 len=0;
                pos++;
                memcpy(&len,&(data[pos]),sizeof(uint32));
                pos+=sizeof(uint32);
                byteReverse((unsigned char*)&len,sizeof(len)/4);

                if(pos+len+1>size)
                    goto error;
                if(len>16384)
                    goto error;

                if(len) {
                    char* str=MY_NEW char[len+1];
                    memcpy(str,&(data[pos]),len);
                    pos+=len;
                    str[len]=0;
                    c_var_string* x=MY_NEW c_var_string((string)str);
                    MY_DELETE_A(str);
                    pos++;
                    return x;
                }
                pos++;
                return MY_NEW c_var_string("");
            }
            // NOTREACH
        case 'A':
            {
                c_var_array* x=var_decode_array_for_botnet(this_module,data,pos,size);
                if(!x)
                    goto error;
                return x;
            }
            // NOTREACH
        default:
            goto error;
    }
    // NOTREACH
}

/*!
    \author VooDooMan
    \version 1
    \date 2005
    \brief Decodes binary-encoded list of arguments (for BOTNET remote call)
    \param data Buffer
    \param size Size of buffer
    \param args Returns list of arguments
    \return Returns true if ok
    \param this_module Module
    \param func_name Returns name of function
*/
bool var_decode_for_botnet(c_module* this_module, char* data, size_t size, vector<c_var_base*>& args, string& func_name)
{
    args.clear();
    func_name="";
    size_t pos=0;

    bool first=true;

again:
    if(pos==size)
        return true;

    if(pos>size) {
error:
        vector<c_var_base*>::iterator i1=args.begin();
        while(i1!=args.end()) {
            MY_DELETE(*i1);
            args.erase(i1);
            i1=args.begin();
        }
        return false;
    }

    if(first) {
        first=false;
        while(pos<size && data[pos]!=0)
            func_name+=(string)""+data[pos++];
        pos++;
        goto again;
    }

    if(0==data[pos] && pos+1==size)
        return true;

    c_var_base* x=var_decode_entity_for_botnet(this_module,data,pos,size);
    if(!x)
        goto error;

    args.push_back(x);

    goto again;
}

/*!
    \brief Used for iterative array operation
    \author VooDooMan
    \version 1
    \date ?
*/
struct c_array_var_decode_array_for_botnet
{
    c_var_base* val;
    c_var_array* x;
    c_index_base* idx;

    ~c_array_var_decode_array_for_botnet()
    {
        ;
    }
};

/*!
    \author VooDooMan
    \version 1
    \date 2005
    \brief Decodes binary-encoded array (for BOTNET remote call)
    \param data Buffer
    \param pos Position in buffer
    \param size Size of buffer
    \return Returns NULL in case of error
    \param this_module Module
*/
c_var_array* var_decode_array_for_botnet(c_module* this_module, char* data, size_t &pos, size_t size)
{
    stack<c_array_var_decode_array_for_botnet> st;
    c_array_var_decode_array_for_botnet ste;
    ste.val=NULL;
    ste.x=MY_NEW c_var_array();
    st.push(ste);

again_iter:

    if(pos>=size) {
        while(st.size()) {
            if(st.top().idx)
                MY_DELETE(st.top().idx);
            if(st.top().val)
                MY_DELETE(st.top().val);
            if(st.top().x)
                MY_DELETE(st.top().x);
            st.pop();
        }
        return NULL;
    }
    if(data[pos]!='A') {
        while(st.size()) {
            if(st.top().idx)
                MY_DELETE(st.top().idx);
            if(st.top().val)
                MY_DELETE(st.top().val);
            if(st.top().x)
                MY_DELETE(st.top().x);
            st.pop();
        }
        return NULL;
    }
    pos++;

again:

    st.top().val=NULL;

    if(!data[pos]) {
        // end of array
        goto end_of_array;
    }

    {
        c_var_base* idx=var_decode_entity_for_botnet(this_module,data,pos,size);
        if(!idx || idx->get_var_type()!=VARTYPE_STRING && idx->get_var_type()!=VARTYPE_INT) {
            MY_DELETE(idx);
            while(st.size()) {
                if(st.top().idx)
                    MY_DELETE(st.top().idx);
                if(st.top().val)
                    MY_DELETE(st.top().val);
                if(st.top().x)
                    MY_DELETE(st.top().x);
                st.pop();
            }
            return NULL;
        }
        if(idx->get_var_type()==VARTYPE_STRING)
            st.top().idx=MY_NEW c_index_str(dynamic_cast<c_var_string*>(idx)->get());
        if(idx->get_var_type()==VARTYPE_INT)
            st.top().idx=MY_NEW c_index_int(dynamic_cast<c_var_int*>(idx)->get());
        MY_DELETE(idx);
    }

    if('A'!=data[pos])
        st.top().val=var_decode_entity_for_botnet(this_module,data,pos,size);
    else {
        c_array_var_decode_array_for_botnet ste;
        ste.idx=NULL;
        ste.val=NULL;
        ste.x=MY_NEW c_var_array();
        st.push(ste);
        goto again_iter;
    }
    if(!st.top().val) {
        while(st.size()) {
            if(st.top().idx)
                MY_DELETE(st.top().idx);
            if(st.top().val)
                MY_DELETE(st.top().val);
            if(st.top().x)
                MY_DELETE(st.top().x);
            st.pop();
        }
        return NULL;
    }

again_iter2:
    st.top().x->set(this_module,st.top().idx,st.top().val,false,"BOTNET remote call");

end_of_array:

    if(pos>=size) {
        while(st.size()) {
            if(st.top().idx)
                MY_DELETE(st.top().idx);
            if(st.top().val)
                MY_DELETE(st.top().val);
            if(st.top().x)
                MY_DELETE(st.top().x);
            st.pop();
        }
        return NULL;
    }

    if(data[pos]!=0)
        goto again;

    if(st.size()>1) {
        c_var_base* x=st.top().x;
        st.pop();
        if(!st.empty()) {
            st.top().val=x;
            pos++;
            goto again_iter2;
        }
    }

    /*
    if(st.top().idx)
        MY_DELETE(st.top().idx);
    if(st.top().val)
        MY_DELETE(st.top().val);
    */

    pos++;
    return st.top().x;
}

/*!
    \author VooDooMan
    \version 1
    \date 2005
    \brief Encodes list of arguments to binary form (for BOTNET remote call)
    \param data Returns buffer (must be MY_DELETE_A'd)
    \param size Returns size of buffer
    \param args List of arguments
    \return Returns true if ok
    \param this_module Module
    \param func_name Name of function
*/
bool var_encode_for_botnet(c_module* this_module, char* &data, size_t &size, const vector<c_var_base*>& args, const string& func_name)
{
    size_t res_l=func_name.length()+1;
    char* res=MY_NEW char[res_l];
    strcpy(res,func_name.c_str());

    vector<c_var_base*>::const_iterator i1;

    for(i1=args.begin(); i1!=args.end(); ++i1) {
        char* ent=NULL;
        size_t ent_l=0;
        if(*i1) {
            if((*i1)->get_var_type()==VARTYPE_BOOL) {
                c_var_bool* x=dynamic_cast<c_var_bool*>(*i1);
                if(!x) {
                    var_fatal(this_module,__FILE__,__my_func__,GET_STR_DEBUG_INFO_EX("C++ dynamic_cast<c_var_bool*>(*i1) returned NULL"),"");
                    MY_DELETE_A(res);
                    data=NULL;
                    size=0;
                    return false;
                }
                x->dump_for_botnet(this_module,ent,ent_l);
            }
            if((*i1)->get_var_type()==VARTYPE_STRING) {
                c_var_string* x=dynamic_cast<c_var_string*>(*i1);
                if(!x) {
                    var_fatal(this_module,__FILE__,__my_func__,GET_STR_DEBUG_INFO_EX("C++ dynamic_cast<c_var_string*>(*i1) returned NULL"),"");
                    MY_DELETE_A(res);
                    data=NULL;
                    size=0;
                    return false;
                }
                x->dump_for_botnet(this_module,ent,ent_l);
            }
            if((*i1)->get_var_type()==VARTYPE_INT) {
                c_var_int* x=dynamic_cast<c_var_int*>(*i1);
                if(!x) {
                    var_fatal(this_module,__FILE__,__my_func__,GET_STR_DEBUG_INFO_EX("C++ dynamic_cast<c_var_int*>(*i1) returned NULL"),"");
                    MY_DELETE_A(res);
                    data=NULL;
                    size=0;
                    return false;
                }
                x->dump_for_botnet(this_module,ent,ent_l);
            }
            if((*i1)->get_var_type()==VARTYPE_UNSET) {
                c_var_unset* x=dynamic_cast<c_var_unset*>(*i1);
                if(!x) {
                    var_fatal(this_module,__FILE__,__my_func__,GET_STR_DEBUG_INFO_EX("C++ dynamic_cast<c_var_unset*>(*i1) returned NULL"),"");
                    MY_DELETE_A(res);
                    data=NULL;
                    size=0;
                    return false;
                }
                x->dump_for_botnet(this_module,ent,ent_l);
            }
            if((*i1)->get_var_type()==VARTYPE_ARRAY) {
                c_var_array* x=dynamic_cast<c_var_array*>(*i1);
                if(!x) {
                    var_fatal(this_module,__FILE__,__my_func__,GET_STR_DEBUG_INFO_EX("C++ dynamic_cast<c_var_array*>(*i1) returned NULL"),"");
                    MY_DELETE_A(res);
                    data=NULL;
                    size=0;
                    return false;
                }
                x->dump_for_botnet(this_module,ent,ent_l);
            }
        } else {
            c_var_unset* x=MY_NEW c_var_unset();
            x->dump_for_botnet(this_module,ent,ent_l);
            MY_DELETE(x);
        }
        if(!ent) {
            var_fatal(this_module,__FILE__,__my_func__,GET_STR_DEBUG_INFO_EX("Unsupported type of variable"),"");
            MY_DELETE_A(res);
            data=NULL;
            size=0;
            return false;
        }

        size_t new_res_l=res_l+ent_l;
        char* new_res=MY_NEW char[new_res_l];
        memcpy(new_res,res,res_l);
        memcpy(&(new_res[res_l]),ent,ent_l);
        MY_DELETE_A(res);
        res=new_res;
        res_l=new_res_l;
        MY_DELETE_A(ent);
    }
    size_t new_res_l=res_l+1;
    char* new_res=MY_NEW char[new_res_l];
    memcpy(new_res,res,res_l);
    new_res[res_l]=0;
    MY_DELETE_A(res);
    res=new_res;
    res_l=new_res_l;

    data=res;
    size=res_l;
    return true;
}

/*!
    \brief Sets module in which an error occured
    \version 1
    \date 2006-04-28
    \author VooDooMan
*/
void c_var_exception::set_module(c_module* module)
{
    this->this_module_=module;
}

c_var_base* make_new_unset()
{
    return MY_NEW c_var_unset();
}

