/***************************************************************************
                          botnet.cpp  -  description
                             -------------------
    begin                : Tue Sep 6 2005
    copyright            : (C) 2005-2007 by Marian VooDooMan Meravy
    email                : ghostvoodooman -a.T= users (dot) sourceforge.net
 ***************************************************************************/

/* $Id: variable.h,v 1.43 2008/08/23 15:10:42 ghostvoodooman Exp $ */

/***************************************************************************

VooDoo cIRCle - an IRC (ro)bot
Copyright (C) 2005-2008 by Marian VooDooMan Meravy
  (ghostvoodooman -a.T= users (dot) sourceforge.net)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************/

/*!
    \file
    \brief Provides classes for variables and expression evaluation
    \version $Id: variable.h,v 1.43 2008/08/23 15:10:42 ghostvoodooman Exp $
*/

#ifndef VARIABLE_H__
#define VARIABLE_H__

#if defined(HAVE_CONFIG_H)
#   include <config.h>
#endif

#include <string>
#include <vector>
#include <stack>
#include <list>
#include <map>
#include <algorithm>

using namespace std;

#include "utils.h"      // because we want to enable mem leak detection

#include "params.h"
#include "compat.h"
#include "utils.h"

#include "microcode_shared.h"

#include "utils/assert.h"

class c_expression;

class c_var_base;
class c_var_unset;
class c_var_int;
class c_var_string;
class c_var_array;
class c_var_bool;

#include "var_index.h"

class c_functions;
class c_part;

class c_module;

class c_micro;
class c_micro_symbol_variable_mapping;

/*!
    \author VooDooMan
    \version 1
    \date 2005
    \brief Gives type of variable
*/
enum e_vartype
{
    VARTYPE_NULL=-100,              //!< Used for exceptions when variable is NULL and should not be (not as real variable)
    VARTYPE_NA,                     //!< Used for exceptions when variable type is not relevant (not as real variable)
    VARTYPE_UNSET=1,                //!< Unset variable
    VARTYPE_STRING,                 //!< Variable is string
    VARTYPE_ARRAY,                  //!< Variable is array
    VARTYPE_INT,                    //!< Variable is integer
    VARTYPE_BOOL,                   //!< Variable is boolean
};

/*!
    \def SESSION_INVALID_ID
    \brief Invalid session ID
*/
#define SESSION_INVALID_ID      0UL

/*!
    \def SESSION_RESERVED_ID
    \brief Reserved session ID
*/
#define SESSION_RESERVED_ID     1UL

/*!
    \def SESSION_WORK_ID
    \brief Session ID used before command/function "work", and after exit from "work" (e.g. after IRC server disconnect)
*/
#define SESSION_WORK_ID         2UL

/*!
    \def SESSION_INTERNAL_ID
    \brief Session ID used before on_internal_event, and after exit from it
*/
#define SESSION_INTERNAL_ID     3UL

/*!
    \def SESSION_CORE_ENTRY_ID
    \brief Session ID used for "_sys_startup" procedure/function of core module
*/
#define SESSION_CORE_ENTRY_ID   4UL

/*!
    \def SESSION_FIRST_ID
    \brief First usable session ID
*/
#define SESSION_FIRST_ID        5UL

string vartype_to_string(e_vartype vartype);

/*!
    \brief Exception class for errors in evaluation
    \version 1
    \author VooDooMan
    \date 2005-10
*/
class c_var_exception
{
public:
    /*!
        \brief Type of error
    */
    enum e_error_type
    {
        E_INDEX_NULL,                   //!< "index" is NULL
        E_INDEX_INV,                    //!< invalid "index": not INT, nor STRING
        E_DYNCAST,                      //!< dynamic_cast<>() returned NULL
        E_NIMPL,                        //!< not implemented
        E_INTERNAL,                     //!< internal error (sanity check)
        E_READONLY,                     //!< trying to write to indexed variable that is read-only
        E_INDEX_RANGE,                  //!< "index" out of range
        E_INDEX_NOT_INT,                //!< "index" to base STRING is not INT
        E_INVALID_OPER,                 //!< Invalid operator (expecting operator)
        E_SYNTAX,                       //!< Syntax error
        E_USER_ERROR,                   //!< Error on user's side
        E_LVALUE,                       //!< L-value required
        E_FUNC_CALL,                    //!< Invalid function call
        E_DIV_ZERO,                     //!< Division by zero
        E_TIMEOUT,                      //!< Script execution timeout
    };

private:
    e_error_type error_;                //!< Type of error
    e_vartype var_type_;                //!< Type of erroneous variable
    string reason_;                     //!< Message for logs
    string func_;                       //!< C++ function where error has occured
    string file_;                       //!< C++ file where error has occured

    c_module* this_module_;

public:
    /*!
        \brief Copy-constructor
        \param orig Original instance
    */
    c_var_exception(const c_var_exception& orig);

    /*!
        \brief Constructor
        \param func C++ function where error occured
        \param file C++ file where error occured
        \param error Type of error
        \param reason Message for logs
        \param this_module Module
    */
    c_var_exception(c_module* this_module, const char* file, const char* func, e_error_type error, string reason);

    /*!
        \brief Constructor
        \param func C++ function where error occured
        \param file C++ file where error occured
        \param error Type of error
        \param reason Message for logs
        \param var_type Erroneous variable (will be used its type, can be NULL)
        \param this_module Module
    */
    c_var_exception(c_module* this_module, const char* file, const char* func, e_error_type error, string reason, const c_var_base* var_type);

    /*!
        \brief Constructor
        \param func C++ function where error occured
        \param file C++ file where error occured
        \param error Type of error
        \param reason Message for logs
        \param var_type Type of erroneous variable
        \param this_module Module
    */
    c_var_exception(c_module* this_module, const char* file, const char* func, e_error_type error, string reason, e_vartype var_type);

    /*!
        \brief Dummy constructor
        \author VooDooMan
        \date 2006-02-02
        \version 1
    */
    c_var_exception()
    {
        error_=E_NIMPL;
        var_type_=VARTYPE_NA;
        this_module_=NULL;
        file_="(N/A)";
    }

    /*!
        \brief Destructor
    */
    virtual ~c_var_exception()
    {
        ;
    }

    /*!
        \brief Prefix reason message with something
        \param prefix Prefix
    */
    void prefix_reason(string prefix);

    /*!
        \brief Gets message
        \return Returns message
    */
    string get_msg();

    /*!
        \brief Gets function name where exception was thrown
    */
    string get_func()
    {
        return func_;
    }

    /*!
        \brief Gets file name where exception was thrown
    */
    string get_file()
    {
        return file_;
    }

    /*!
        \brief Prints error to logs
        \param debug Debug helping string
    */
    void print_oops_to_log(const string& debug);

    // doxygen tags are in 'variable.cpp'
    void set_module(c_module* module);
};

class c_reference_keeper;
class c_var_bool;

typedef unsigned long var_id_t;         //!< ID of variable (a.k.a. its call sign)
/*!
    \def VAR_ID_INVALID
    \brief Invalid variable ID
*/
#define VAR_ID_INVALID      (0UL)

class c_index_base;
class c_index_int;
class c_index_str;

c_var_base* make_new_unset();

enum e_exp_result
{
    EXP_OK=0,                       //!< OK
    EXP_E_PARENTHESIS,              //!< Parenthesis mismatch
    EXP_E_UNKNOWN_OPER,             //!< Unknown operator
    EXP_E_INTERNAL_1,               //!< Internal error #1
    EXP_E_INTERNAL_2,               //!< Internal error #2
    EXP_E_INTERNAL_3,               //!< Internal error #3
    EXP_E_INTERNAL_4,               //!< Internal error #4
    EXP_E_INTERNAL_5,               //!< Internal error #5
    EXP_E_SUBEXP,                   //!< Subexpression evaluation failed
    EXP_E_SYNTAX_CAST,              //!< Syntax error, e.g. "123(int)456"
    EXP_E_SYNTAX_2,                 //!< Syntax error, e.g. "123!456"
    EXP_E_INVALID_OPER_USAGE_1,     //!< Invalid usage of ++ operator, e.g. "-1++"
    EXP_E_INVALID_OPER_USAGE_2,     //!< Invalid usage of ++ operator, e.g. "-1--"

    EXP_I_END_OF_ITERATION,         //!< Internal exception: End of sub-iteration
    EXP_I_NEXT_ITERATION,           //!< Internal exception: Next sub-iteration request
    EXP_I_OTHER_EXCEPTION,          //!< Interval exception: Other exception thrown
    EXP_I_CONTINUE_ITERATION,       //!< Internal exception: Continue iteration (a return from sub-iteration)

    EXP_E_TIMEOUT,                  //!< Script execution timeout
    EXP_E_STACK_OVERFLOW,           //!< Maximum stack size exhausted
};

string exp_result_to_string(e_exp_result e);
void var_clear_parts(vector<c_part>& parts, c_var_base* except);

string var_eval_string_simple(c_module* this_module, const string& str, c_var_array* vars, const string& debug);

bool var_encode_for_botnet(c_module* this_module, char* &data, size_t &size, const vector<c_var_base*>& args, const string& func_name);
bool var_decode_for_botnet(c_module* this_module, char* data, size_t size, vector<c_var_base*>& args, string& func_name);

c_var_base* var_eval_obj(c_module* this_module, string str, c_var_array* vars, const string& debug, bool& return_ref);

void var_eval_simple_cleanup(c_var_base*& left, c_var_base*& right);
void var_kill_subsequent(vector<c_part>& parts, vector<c_part>::iterator from, c_part insert);
void var_decrease_conditionals(vector<c_part>& parts);
void var_eval_simple_error(c_module* this_module, c_var_base*& left, c_var_base*& right, const string& reason, c_var_exception::e_error_type error_type, c_var_base* error_var=NULL);
void var_eval_simple_error(c_module* this_module, c_var_base*& left, c_var_base*& right, const string& reason, c_var_exception::e_error_type error_type, e_vartype error_var);
void var_oops_(c_module* this_module, const string& prefix, const string& func, const string& str, const string& debug);
void var_oops_(c_module* this_module, const string& prefix, const string& func, const string& str, const string& str2, const string& debug);
void var_oops(c_module* this_module, const string& file, const string& func, const string& str, const string& debug);
void var_oops(c_module* this_module, const string& file, const string& func, const string& str, const string& str2, const string& debug);
void var_oops(c_module* this_module, const string& prefix, const string& str, const string& debug);
void var_fatal(c_module* this_module, const string& file, const string& func, const string& str, const string& debug);
c_var_base* var_eval_index(c_module* this_module, c_var_base* base, c_index_int& index, bool insensitive, const string& debug);

#endif

