/***************************************************************************
                          vdcmod.h  -  description
                             -------------------
    begin                : Fri Aug 25 2006
    copyright            : (C) 2006-2007 by Marian VooDooMan Meravy
    email                : ghostvoodooman -a.T= users (dot) sourceforge.net
 ***************************************************************************/

/* $Id: vdcmod.h,v 1.32 2008/08/23 15:10:42 ghostvoodooman Exp $ */

/***************************************************************************

VooDoo cIRCle - an IRC (ro)bot
Copyright (C) 2006-2008 by Marian VooDooMan Meravy
  (ghostvoodooman -a.T= users (dot) sourceforge.net)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************/

/*!
    \file
    \brief Provides interface for binary modules
    \version $Id: vdcmod.h,v 1.32 2008/08/23 15:10:42 ghostvoodooman Exp $
*/

#ifndef VDCMOD_H__
#define VDCMOD_H__

#ifdef VDCMOD_HOST
#   if defined(HAVE_CONFIG_H)
#       include <config.h>
#   endif
#endif

#ifdef __cplusplus
extern "C" {
#endif

#ifndef VDCMOD_API_REQUEST_VERSION
#   define VDCMOD_API_REQUEST_VERSION   1
#endif

#if VDCMOD_API_REQUEST_VERSION < 3
#define s_vdcmod_data_tag           c_vdcmod_data_tag
#define s_unset                     c_unset
#define s_bool                      c_bool
#define s_int                       c_int
#define s_str                       c_str
#define s_array_entry               c_array_entry
#define s_array                     c_array
#define s_vdcmod_data_t             c_vdcmod_data_t
#define s_vdcmod_init               c_vdcmod_init
#define s_vdcmod_init_ret           c_vdcmod_init_ret
#endif

/*!
    \def VDCMOD_API_VERSION
    \brief Version of API used in communication structures
*/

#define VDCMOD_API_VERSION              5

/*!
    \brief Module ID
*/
typedef void* vdcmod_mod_id_t;

typedef enum
{
    DATA_TYPE_UNSET = 1,
    DATA_TYPE_BOOL,
    DATA_TYPE_INT,
    DATA_TYPE_STRING,
    DATA_TYPE_ARRAY,
    DATA_TYPE_ARRAY_ENTRY,
} e_data_type_t;

/*!
    \brief Data structure
*/
typedef struct c_vdcmod_data_t
{
    e_data_type_t type;                 /*!< Type of entity */
    union
    {
        struct
        {
        } c_unset;                      /*!< For variable of type UNSET */

        struct
        {
            char data;                  /*!< Value */
        } c_bool;                       /*!< For variable of type BOOL */

        struct
        {
            signed long data;           /*!< Value */
        } c_int;                        /*!< For variable of type INT (may be truncated if bot was compiled on system where sizeof(long long) > sizeof(long) */

#if (VDCMOD_API_REQUEST_VERSION >= 4) && (VDCMOD_API_VERSION >= 4) && defined(VDCMOD_LONG_LONG)
        /* NB: If you want to use this, then you need to define:
                #define VDCMOD_LONG_LONG    signed long long
           NB: Check value of "vdc_host_sizeof_long_long" in "c_vdcmod_init" structure before assuming size of this!
           Note: To avoid ABI compatibility problems, the best practice is to compile both VooDoo cIRCle bot (host) and your binary module with the same compiler with same settings.
        */
        struct
        {
            VDCMOD_LONG_LONG data;      /*!< Value */
        } c_big_int;                    /*!< For variable of type INT ("signed long long" version, not truncated; or just "signed long" if bot was compiled on system where sizeof(signed long long)==sizeof(signed long), or "signed long long" is not supported by compiler */
#endif

        struct
        {
            char* data;                 /*!< Value */
        } c_str;                        /*!< For variable of type STRING */

        struct
        {
            struct c_vdcmod_data_t* index;  /*!< Index, or NULL if this is the last entry */
            struct c_vdcmod_data_t* value;  /*!< Value, or NULL if this is the last entry */
            struct c_vdcmod_data_t* next;   /*!< Link to next entry, or NULL if this is the last entry */
        } c_array_entry;                /*!< Entry for variable of type ARRAY */

        struct
        {
            struct c_vdcmod_data_t* value;  /*!< Value */
        } c_array;                      /*!< For variable of type ARRAY */
    } u_data;                           /*!< Data */
} c_vdcmod_data_t;

/*!
    \brief Pointer to bound dispatcher function in module
    \param dispatch Dispatch structure, should not be modified by module, nor deallocated
*/
typedef void (*vdcmod_dispatcher_func)(const c_vdcmod_data_t* dispatch);

/*!
    \brief Pointer to bind function
    \param mod_id ID of module, saved from call from VDC of module's init function
    \param type_mask Mask of type of event
    \param user_mask Mask of user
    \param channel_mask Mask of channel
    \param line_mask Mask of command line
    \param before Zero if event should be triggered after modification of global variables, non-zero if before
    \param dispatcher Pointer to dispatcher function in module
*/
typedef void (*vdcmod_bind_func)(vdcmod_mod_id_t mod_id, const char* type_mask, const char* user_mask, const char* channel_mask, const char* line_mask, int before, vdcmod_dispatcher_func dispatcher);

/*!
    \brief Pointer to function which returns superglobal variables
    \param superglobals Receives pointer with superglobal variables, should not be modified, nor deallocated
*/
typedef void (*vdcmod_get_superglobals_func)(vdcmod_mod_id_t mod_id, c_vdcmod_data_t** superglobals);

/*!
    \brief Pointer to function which frees superglobal variables, allocated by "vdcmod_get_superglobals_func", or as result by "vdcmod_eval_func"
    \param superglobals Pointer with superglobal variables, allocated by "vdcmod_get_superglobals_func", or as result by "vdcmod_eval_func"
*/
typedef void (*vdcmod_free_structure_func)(vdcmod_mod_id_t mod_id, c_vdcmod_data_t* struc);

/*!
    \brief Pointer to function which evaluates expression (in same language as functions in logic.txt)
    \param exp Expression
    \param result If return value is non-zero, receives pointer to result of evaluation (as passed by "return $x" command). If there was no "return $x" command, or simply "return;", it will contain UNSET variable. Also, it can return a NULL pointer, which should be treated as UNSET. Should be deallocated by calling "vdcmod_free_structure_func".
    \retval Zero if there was error
    \retval Non-zero if there was no error
*/
typedef int (*vdcmod_eval_func)(vdcmod_mod_id_t mod_id, const char* exp, c_vdcmod_data_t** result);

/*!
    \brief Processes text like it were output from an external script
    \param text Text to process. Lines are ended by single 0xA (LF) character, do not forget magick '#' character sequences
*/
typedef void (*vdcmod_process_script_out_func)(vdcmod_mod_id_t mod_id, const char* text);

/*!
    \brief Structure passed from bot to binary module
    \date 2006-08-25, 2007-11-07
    \author VooDooMan
*/
typedef struct c_vdcmod_init
{
    int version;                /*!< Version of structure: equals to "VDCMOD_API_VERSION" */
    unsigned size;              /*!< Size of this structure */
    vdcmod_mod_id_t mod_id;     /*!< ID of module (save this value, needed for each call of VDC API from module) */

    vdcmod_bind_func bind;      /*!< bind function */
    vdcmod_bind_func unbind;    /*!< unbind function */

    vdcmod_get_superglobals_func get_superglobals;  /*!< Function to get superglobal variables */

    vdcmod_free_structure_func free_struct;   /*!< Function to free structure for superglobal variables and result of "eval_func" */

    vdcmod_eval_func eval_func; /*!< Eval function */

    vdcmod_process_script_out_func process_script_out;  /*!< Function to process script output */

    /* past this point, everything for "version==1" is undefined */

    /* "version==4" */
    int vdc_host_sizeof_long;       /*!< sizeof(long long) of host (the bot), in bytes */
    int vdc_host_sizeof_long_long;  /*!< sizeof(long long) of host (the bot), in bytes */

    /* past this point, everything for "version==4" is undefined */

    /* "version==5" */
    vdcmod_bind_func bind_ex;       /*!< bind_ex function */
    vdcmod_bind_func unbind_ex;     /*!< unbind_ex function */

    /* past this point, everything for "version==5" is undefined */
} c_vdcmod_init;

/*!
    \brief Structure returned from module to bot
    \date 2006-08-26
    \author VooDooMan
*/
typedef struct c_vdcmod_init_ret
{
    int version;                /*!< Version of structure */
    unsigned size;              /*!< Size of this structure */

    /* past this point, everything for "version==1" is undefined */
} c_vdcmod_init_ret;

typedef void (*vdcmod_init_func)(const char* mod_name, const char* dir, const c_vdcmod_init* init, c_vdcmod_init_ret* result);
typedef void (*vdcmod_done_func)();

#ifdef WIN32
#   define MY_EXPORT    __declspec(dllexport)
#else
#   define MY_EXPORT    /* */
#endif

#ifndef VDCMOD_HOST /* this is defined only inside of VDC, if undefined, we are including this file inside of binary module */

/*!
    \brief Module must provide this external function, it is called once at module load
    \param mod_name Name of module
    \param dir Full (not relative) path to module's root directory
    \param init Initial structure (module should maintain it own copy of this structure)
    \param result Store here the result returning from module to bot
*/
MY_EXPORT void vdcmod_init(const char* mod_name, const char* dir, const c_vdcmod_init* init, c_vdcmod_init_ret* result);

/*!
    \brief Module must provide this external function, it is called once before module unload
*/
MY_EXPORT void vdcmod_done();

#endif /* !VDCMOD_HOST */

#ifdef __cplusplus
} /* extern "C" */
#endif

#endif

