/***************************************************************************
                          match.h  -  description
                             -------------------
    begin                : Thu Dec 9 2004
    copyright            : (C) 1990 by Jarkko Oikarinen
    copyright            : (C) 2004-2007 by Marian VooDooMan Meravy
    email                : ghostvoodooman -a.T= users (dot) sourceforge.net
 ***************************************************************************/

/* $Id: match.h,v 1.28 2008/03/11 21:11:50 ghostvoodooman Exp $ */

/***************************************************************************

VooDoo cIRCle - an IRC (ro)bot
Copyright (C) 1990 Jarkko Oikarinen
Copyright (C) 2004-2008 by Marian VooDooMan Meravy
  (ghostvoodooman -a.T= users (dot) sourceforge.net)
  (re-coded from C to C++)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************/

/*!
    \file
    \brief Provides wildcard checks
    \version $Id: match.h,v 1.28 2008/03/11 21:11:50 ghostvoodooman Exp $
*/

#ifndef MATCH_H__
#define MATCH_H__

#if defined(HAVE_CONFIG_H)
#   include <config.h>
#endif

#include "params.h"

#include "sock_shared.h"

#include <string>
#include <vector>
#include <stack>
#include <list>
#include <map>

#include "compat.h"

const char *inet_ntop6 (const u_char *src, char *dst, size_t size);
int  inetpton(int af, const char *src, void *dst);
int inet_pton(int af, const char *src, void *dst);

const char * my_inet_ntop (int af, const void *src, char *dst, size_t size);

/*!
    \brief Decompresses "::" from IPv6 numeric address; if other address (or possibly hostname given) returns untouched
    \author VooDooMan
    \version 2
    \date 2005, 2005-10-02
    \param src Source address
    \return Returns uncompressed address (must be MY_DELETE_A'd)
*/
char* ipv6_decompress(const char* src);

//#ifdef INET6

// BEGIN by Marian VooDooMan Meravy
#if defined(IN6_IS_ADDR_V4MAPPED)
#   undef IN6_IS_ADDR_V4MAPPED
#endif
// END   by Marian VooDooMan Meravy

# define MYDUMMY_SIZE 128
# define bcopy(a,b,c) memmove((b),(a),(c))
#define IN6_IS_ADDR_V4MAPPED(a)         \
    ((((u_int32_t*)(a))[0] == 0) \
    && (((u_int32_t*)(a))[1] == 0) \
    && (((u_int32_t*)(a))[2] == htonl(0xffff)))

#define u_int32_t unsigned int
#define u_char unsigned char
#define IN6ADDRSZ       16
#define INADDRSZ        4
#define INT16SZ         2

// do NOT touch this define! It doesn't mean if to use IPv6, but rather if to use IPv6 matching!
#define INET6

#define Reg register
#define u_char unsigned char

// BEGIN ADD    - VooDooMan 2004/10/18 - changed
#define u_long unsigned long

char* index(char* str, char sub);
// END   ADD    - VooDooMan 2004/10/18 - changed

u_char my_tolower(u_char ch);

#define  MAX_ITERATIONS  512
/*
**  Compare if a given string (name) matches the given
**  mask (which can contain wild cards: '*' - match any
**  number of chars, '?' - match any single character.
**
**  return  0, if match
**    1, if no match
*/

/*

** match()
** Iterative matching function, rather than recursive.
** Written by Douglas A Lewis (dalewis@acsu.buffalo.edu)
*/
int  match(char *mask, char *name);

#define  strncpyzt(x, y, N) do{(void)strncpy(x,y,N);x[N-1]='\0';}while(0)

/*
 * Match address by #IP bitmask (10.11.12.128/27)
 * Now should work for IPv6 too.
 * returns -1 on error, 0 on match, 1 when NO match.
 */
// BEGIN CHANGE - VooDooMan 2004/10/18 - changed
//int    match_ipmask(char *mask, aClient *cptr, int maskwithusername)
int    match_ipmask(char *mask, char* username, in_addr6_ u_addr, int maskwithusername);
// END   CHANGE - VooDooMan 2004/10/18 - changed

#endif

