/***************************************************************************
                          sendmail.cpp  -  description
                             -------------------
    begin                : Thu Dec 9 2004
    copyright            : (C) 2004-2006 by VooDooMan
    email                : ghostvoodooman@users.sourceforge.net
 ***************************************************************************/

/* $Id: sendmail.cpp,v 1.30 2008/03/11 21:11:50 ghostvoodooman Exp $ */

/***************************************************************************

VooDoo cIRCle - an IRC (ro)bot
Copyright (C) 2004-2008  Marian VooDooMan Meravy
                         (ghostvoodooman@users.sourceforge.net)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************/

/*!
    \file
    \brief Helper console application for sending an e-mail
*/

#if defined(HAVE_CONFIG_H)
#   include <config.h>
#endif

#include "sock_shared.h"

#include "params.h"
#include "compat.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#if (!defined(_WIN32) && !defined(_WIN64)) || defined(__CYGWIN__)
#   include <unistd.h>
#endif
#ifdef __BORLANDC__
#   include <messages.hpp>
#endif
#if defined(_WIN32) || defined(_WIN64) || defined(__CYGWIN__)
#   include <windows.h>
#   include <io.h>
#endif

#include "sock.h"
#include "log.h"
#include "match.h"

#include "params.h"

#if defined(_MSC_VER) || defined(__BORLANDC__)
#   pragma hdrstop
#endif

#ifdef _WIN32
#   define FILE_SLASH "\\"
#else
#   define FILE_SLASH "/"
#endif

//---------------------------------------------------------------------------

void ReadLn(FILE *f, char* buff, int max)
{
    int buff_pos=0;
    for(;;)
    {
        int ch1=fgetc(f);
        if(ch1==EOF) break;
        if(ch1=='\n') break;
        if(ch1=='\r') continue;
        buff[buff_pos++]=ch1;

        if(buff_pos+1>max)
            break;
    }
    buff[buff_pos++]=0;
}

bool ProcessMessage()
{
#ifdef _WIN32
    tagMSG* m=new tagMSG;
    if(PeekMessage(m, 0, 0, 0, PM_REMOVE)) {
        if(m->message!=WM_QUIT) {
            TranslateMessage(m);
            DispatchMessage(m);
        }
        delete m;
        return true;
    }
    delete m;
#endif
    return false;
}

void ProcessMessages()
{
    while(ProcessMessage())
        ;
}

void wait(bool bWait, c_socket& socket, char* buff1, int max, int& ec)
{
    while(bWait) {
        ec=0;
        bool closed=false;
        size_t size=sock_read(socket,buff1,max,ec,closed);
        buff1[size]=0;
        if(ec || closed) {
            break;
        }

        if(size)
            break;

        extern void sock_send_cache();
        sock_send_cache();
#ifdef _WIN32
        Sleep(1);
#endif
        ProcessMessages();
    }
}

void log(const char* l1, const char* l2, const char* l3)
{
    FILE* f=fopen("." FILE_SLASH "logs" FILE_SLASH "sendmail.log","at");
    if(f) {
        fprintf(f,l1,l2,l3);
        fclose(f);
    }
    printf(l1,l2,l3);
}

int main(int argc, char* argv[])
{
    if(argc!=2)
        return 1;

    FILE* f=fopen(argv[1],"r");
    if(!f)
        return 1;

    time_t now=time(NULL);
    log("%s%s\n\n","! BEGIN on ",ctime(&now));

#define MAX 2*1024
    char buff1[MAX];
    char buff2[MAX];
    char buff3[MAX];

    ReadLn(f,buff1,MAX); // local IP
    ReadLn(f,buff2,MAX); // remote host
    ReadLn(f,buff3,MAX); // remote port

    sock_startup();

    int ec=0;
    char ip[1024*4];
    ip[0]=0;
    in_addr ip_addr=sock_resolve(buff2);
    my_inet_ntop(AF_INET,&ip_addr,ip,sizeof(ip)-1);
    if(!strcmp(ip,"0.0.0.0"))
        strcpy(ip,buff2);
    c_socket socket=sock_connect(buff1,ip,(unsigned short)atol(buff3),ec,true,"sendmail");
    if(ec)
        return 1;

    bool bWait=true;

    while(1) {
        buff1[0]=0;
        wait(bWait,socket,buff1,MAX,ec);
        if(buff2[0]=='D' && buff2[1]=='A' && buff2[2]=='T' && buff2[3]=='A')
            bWait=false;
        if(strlen(buff1)>0)
            log("%s%s\n","< ",buff1);
        if(buff2[0]=='.' && buff2[1]=='\r' && buff2[2]=='\n' && buff2[3]==0) {
            bWait=true;
            break;
        }
        ReadLn(f,buff1,MAX);
        log("%s%s\n","> ",buff1);
        strcat(buff1,"\r\n");
        sock_send(socket,buff1,strlen(buff1),ec);
        strcpy(buff2,buff1);
        if(ec) {
            return 1;
        }
    }

    buff1[0]=0;
    wait(bWait,socket,buff1,MAX,ec);
    if(strlen(buff1)>0)

        log("%s%s\n","< ",buff1);

    strcpy(buff1,"QUIT\r\n");
    log("%s%s\n","> ",buff1);
    sock_send(socket,buff1,strlen(buff1),ec);
    if(ec) {
        return 1;
    }

    buff1[0]=0;
    wait(bWait,socket,buff1,MAX,ec);
    if(strlen(buff1)>0)
        log("%s%s\n","< ",buff1);

    sock_close(socket,__FILE__,__LINE__);

    sock_shutdown();

    now=time(NULL);
    log("%s%s\n","! END on ",ctime(&now));

    log("%s%s\n","==","=====");

    fclose(f);
    unlink(argv[1]);

    return 0;
}
//---------------------------------------------------------------------------

