/***************************************************************************
                          sock_shared.h  -  description
                             -------------------
    begin                : Thu Dec 9 2004
    copyright            : (C) 2004-2007 by Marian VooDooMan Meravy
    email                : ghostvoodooman -a.T= users (dot) sourceforge.net
 ***************************************************************************/

/* $Id: sock_shared.h,v 1.29 2008/03/11 21:11:50 ghostvoodooman Exp $ */

/***************************************************************************

VooDoo cIRCle - an IRC (ro)bot
Copyright (C) 2004-2008 by Marian VooDooMan Meravy
  (ghostvoodooman -a.T= users (dot) sourceforge.net)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************/

/*!
    \file
    \brief Provides low-level implementation of TCP IP sockets
    \version $Id: sock_shared.h,v 1.29 2008/03/11 21:11:50 ghostvoodooman Exp $
*/

#ifndef SOCK_SHARED_H__
#define SOCK_SHARED_H__

#if defined(HAVE_CONFIG_H)
#   include <config.h>
#endif

#include <string>
#include <vector>
#include <stack>
#include <list>
#include <map>
#include <sstream>

#include <time.h>
#include <string>

using namespace std;

#if !defined(_WIN32) && !defined(_WIN64) && !defined(__CYGWIN__)
#   include <sys/types.h>
#   include <sys/socket.h>
#   include <netinet/in.h>
#   include <arpa/inet.h>
#   include <netdb.h>

#   include <errno.h>
#   include <fcntl.h>

#   define WSAGetLastError() errno
#   define WSAEWOULDBLOCK EAGAIN
#   define closesocket close
#   define ioctlsocket ioctl
#   define SOCKET int
#   define PROTOENT protoent
#   define SOCKET_ERROR -1
#   define INVALID_SOCKET -1
#endif

#if defined(_WIN32) || defined(_WIN64) || defined(__CYGWIN__)
#   define WINVER           0x0501
#   define _WIN32_WINNT     0x0501
#   if !defined(__CYGWIN__)
#       include <winsock2.h>
#   endif
#   include <Ws2tcpip.h>
#   include <windows.h>
#endif

#include "params.h"
#include "compat.h"

#ifndef USE_OPENSSL
#   include <gnutls/gnutls.h>
#endif

/*!
    \brief Stores informnations about socket handle
    \author VooDooMan
    \version 1
    \date 2007-07-17
*/
struct c_socket_hndl
{
    time_t closed;          //!< Zero if open, or timestamp of close
    SOCKET h;               //!< OS handle
    char* fn;               //!< If closed, filename
    int ln;                 //!< If closed, line number
};

/*!
    \brief Stores informnations about socket handle
    \author VooDooMan
    \version 1
    \date 2004
*/
struct c_socket
{
    unsigned hndl;                          //!< Internal handle
    unsigned was_hndl;                      //!< Backup of internal handle
#if defined(_WIN32) || defined(_WIN64)
    HWND wnd;                               //!< Window handle for socket event messages (Windows(TM) only)
#endif

    string comment;                         //!< Comment for socket

    size_t received;                        //!< Number of bytes received
    size_t sent;                            //!< Number of bytes sent

#ifndef USE_OPENSSL
    bool got_xcred;                         //!< If to free "xcred"
    gnutls_certificate_credentials xcred;   //!< GNUTLS credentials
    gnutls_session_t m_gnutls_session;      //!< GnuTLS session structure
#endif

    time_t last_io;                         //!< Timestamp of last receive, or send data (only for client sockets)

    bool closed;                            //!< true if it is closed

    char host[2048];                        //!< If it is IPv6 server, there is numeric host of remote peer

    /*!
        \brief Clears the structure
        \author VooDooMan
        \version 1
        \date 2004
    */
    void clear() {
        hndl=(unsigned)-1;
#if defined(_WIN32) || defined(_WIN64)
        wnd=NULL;
#endif
        comment="";
        received=sent=0;
#ifndef USE_OPENSSL
        got_xcred=false;
#endif
        last_io=time(NULL);
        closed=true;
        host[0]='\0';
    }

    /*!
        \brief Returns true if handle to socket is not -1
        \author VooDooMan
        \version 1
        \date 2004
    */
    bool cmp()
    {
        return hndl!=(unsigned)-1;
    }

    /*!
        \brief Compares two handles for identicality
        \author VooDooMan
        \version 2
        \date 2004, 2007-03-23
    */
    bool operator== (const c_socket& x) {
        return this->hndl==x.hndl;
    }

    c_socket() {
        clear();
    }

    c_socket(const c_socket& orig);

    SOCKET deref(const char* fn, int ln);

    /*!
        \brief Retrieves if copy of socket descriptor was this one
        \author VooDooMan
        \version 1
        \date 2007-08-03
        \param other Other socket
        \return Returns 'true' if this was the 'other'
    */
    bool was_it_you(const c_socket& other);
};

#endif
