#if defined(HAVE_CONFIG_H)
#   include <config.h>
#endif

#include <string>
#include <vector>
#include <stack>
#include <list>
#include <map>
#include <algorithm>

#include "../../utils/memleak.h"

#include "../../params.h"

#include "c_tree_shared.h"
#include "c_tree_microcode.h"

#include "../c_expression.h"
#include "../../microcode/c_micro_symbol_variable_mapping.h"

#include "../../microcode/c_micro_minus.h"
#include "../../microcode/c_micro_plus.h"
#include "../../microcode/c_micro_post_decrement.h"
#include "../../microcode/c_micro_post_increment.h"
#include "../../microcode/c_micro_pre_decrement.h"
#include "../../microcode/c_micro_pre_increment.h"
#include "../../microcode/c_micro_one_compl.h"
#include "../../microcode/c_micro_cast_int.h"
#include "../../microcode/c_micro_cast_string.h"
#include "../../microcode/c_micro_cast_bool.h"
#include "../../microcode/c_micro_not.h"
#include "../../microcode/c_micro_mul.h"
#include "../../microcode/c_micro_div.h"
#include "../../microcode/c_micro_mod.h"
#include "../../microcode/c_micro_concat.h"
#include "../../microcode/c_micro_less.h"
#include "../../microcode/c_micro_less_or_equal.h"
#include "../../microcode/c_micro_greater.h"
#include "../../microcode/c_micro_greater_or_equal.h"
#include "../../microcode/c_micro_equal.h"
#include "../../microcode/c_micro_identical.h"
#include "../../microcode/c_micro_bit_and.h"
#include "../../microcode/c_micro_bit_xor.h"
#include "../../microcode/c_micro_bit_or.h"
#include "../../microcode/c_micro_return.h"
#include "../../microcode/c_micro_assign.h"
#include "../../microcode/c_micro_shift_left.h"
#include "../../microcode/c_micro_shift_right.h"
#include "../../microcode/c_micro_func_call.h"
#include "../../microcode/c_micro_array.h"
#include "../../microcode/c_micro_log_and.h"
#include "../../microcode/c_micro_log_or.h"
#include "../../microcode/c_micro_index.h"
#include "../../microcode/c_micro_str_eval.h"
#include "../../microcode/c_micro_not_equal.h"
#include "../../microcode/c_micro_not_identical.h"
#include "../../microcode/c_micro_ternary.h"
#include "../../microcode/c_micro_jump.h"
#include "../../microcode/c_micro_isset.h"

#include "../../utils/precomp.h"

#if defined(_MSC_VER) || defined(__BORLANDC__)
#   pragma hdrstop
#endif

c_sub_exp::~c_sub_exp()
{
}

void c_sub_exp_index::compile(c_tree_microcode& tm_callback)
{
    if(VAR_SYM_INDEX_INVALID==m_var_id)
        m_var_id=tm_callback.add_temp_var();
    MY_ASSERT(NULL!=m_base_exp);
}

void c_sub_exp_variable::compile(c_tree_microcode& tm_callback)
{
    if(VAR_SYM_INDEX_INVALID==m_var_id)
        m_var_id=tm_callback.add_var(m_name);
}

void c_sub_exp_numeric_immediate::compile(c_tree_microcode& tm_callback)
{
    if(VAR_SYM_INDEX_INVALID==m_var_id)
        m_var_id=tm_callback.add_const_int_var(m_num);
}

void c_sub_exp_const::compile(c_tree_microcode& tm_callback)
{
    if(VAR_SYM_INDEX_INVALID==m_var_id)
        m_var_id=tm_callback.add_const_built_in(m_name);
}

void c_sub_exp_func::compile(c_tree_microcode& tm_callback)
{
    if(VAR_SYM_INDEX_INVALID==m_var_id)
        m_var_id=tm_callback.add_temp_var();

    if(!get_func_name().compare("array")) {
        vector<c_sub_exp*>::iterator i1=m_args.begin();
        while(i1!=m_args.end()) {
            (*i1)->compile(tm_callback);
            var_sym_index_t key=VAR_SYM_INDEX_INVALID;
            var_sym_index_t val=(*i1)->get_var_id();
            if((*i1)->get_prec()==23) { // "=>"
                c_sub_exp_oper* o=dynamic_cast<c_sub_exp_oper*>(*i1);
                MY_ASSERT(o->get_eval().size()==2);
                o->get_eval()[0]->compile(tm_callback);
                o->get_eval()[1]->compile(tm_callback);
                key=o->get_eval()[0]->get_var_id();
                val=o->get_eval()[1]->get_var_id();
            }
            m_array_args.push_back(make_pair(key,val));
            ++i1;
        }
        m_args.clear();
    }
}

void c_sub_exp_eval_str::compile(c_tree_microcode& tm_callback)
{
    if(VAR_SYM_INDEX_INVALID==m_var_id)
        m_var_id=tm_callback.add_eval_str_var(m_str.str());
}

void c_sub_exp_immediate_str::compile(c_tree_microcode& tm_callback)
{
    if(VAR_SYM_INDEX_INVALID==m_var_id)
        m_var_id=tm_callback.add_const_str_var(m_str.str());
}

void c_sub_exp_parenthesis::compile(c_tree_microcode& tm_callback)
{
    if(VAR_SYM_INDEX_INVALID==m_var_id) {
        MY_ASSERT(!m_eval.empty());
        m_var_id=tm_callback.add_temp_var();
        MY_ASSERT(VAR_SYM_INDEX_INVALID!=m_var_id);
    }
}

void c_sub_exp_oper::compile(c_tree_microcode& tm_callback)
{
    if(get_prec()==19) { // assignment operators
        MY_ASSERT(m_eval.size()==2);
        m_var_id=m_eval[0]->get_var_id();
        MY_ASSERT(VAR_SYM_INDEX_INVALID!=m_var_id);
        return;
    }
    if(VAR_SYM_INDEX_INVALID==m_var_id)
        m_var_id=tm_callback.add_temp_var();
}

string c_sub_exp_root::emit_human_desc(string::size_type indent)
{
    static const string m_desc="EXPRESSION_ROOT";
    ostringstream o;
    string::size_type ii;
    for(ii=0; ii<indent; ++ii)
        o << " ";
    o << m_desc;
    o << "\n";
    return o.str();
}

string c_sub_exp_root::emit_human_microcode(string::size_type indent)
{
    static const string m_desc="<ROOT>\n";
    ostringstream o;
    string::size_type ii;
    for(ii=0; ii<indent; ++ii)
        o << " ";
    o << m_desc;
    return o.str();
}

string c_sub_exp_parenthesis::emit_human_desc(string::size_type indent)
{
    static const string m_desc="PARENTHESIS (";
    ostringstream o;
    string::size_type ii;
    for(ii=0; ii<indent; ++ii)
        o << " ";
    o << m_desc;
    o << "\n";
    indent+=2;
    vector<c_sub_exp*>::iterator i1;
    for(i1=m_eval.begin(); i1!=m_eval.end(); ++i1)
        o << (*i1)->emit_human_desc(indent).c_str();
    o << "\n";
    return o.str();
}

string c_sub_exp_parenthesis::emit_human_microcode(string::size_type indent)
{
    static const string m_desc="PARENTHESIS ";
    ostringstream o;
    string::size_type ii;
    for(ii=0; ii<indent; ++ii)
        o << " ";
    o << m_desc.c_str() << m_var_id;
    o << "\n";
    return o.str();
}

string c_sub_exp_variable::emit_human_desc(string::size_type indent)
{
    static const string m_desc="VARIABLE ";
    ostringstream o;
    string::size_type ii;
    for(ii=0; ii<indent; ++ii)
        o << " ";
    o << m_desc << m_name.c_str();
    o << "\n";
    return o.str();
}

string c_sub_exp_variable::emit_human_microcode(string::size_type indent)
{
    static const string m_desc="VARIABLE ";
    ostringstream o;
    string::size_type ii;
    for(ii=0; ii<indent; ++ii)
        o << " ";
    o << m_desc << m_var_id << " " << m_name.c_str();
    o << "\n";
    return o.str();
}

string c_sub_exp_numeric_immediate::emit_human_desc(string::size_type indent)
{
    static const string m_desc="NUMERIC_IMMEDIATE ";
    ostringstream o;
    string::size_type ii;
    for(ii=0; ii<indent; ++ii)
        o << " ";
    o << m_desc << m_num;
    o << "\n";
    return o.str();
}

string c_sub_exp_numeric_immediate::emit_human_microcode(string::size_type indent)
{
    static const string m_desc="NUMERIC_IMMEDIATE ";
    ostringstream o;
    string::size_type ii;
    for(ii=0; ii<indent; ++ii)
        o << " ";
    o << m_desc << m_var_id << " " << m_num;
    o << "\n";
    return o.str();
}

string c_sub_exp_const::emit_human_desc(string::size_type indent)
{
    static const string m_desc="CONST ";
    ostringstream o;
    string::size_type ii;
    for(ii=0; ii<indent; ++ii)
        o << " ";
    o << m_desc << "\"" << m_name.c_str() << "\"";
    o << "\n";
    return o.str();
}

string c_sub_exp_const::emit_human_microcode(string::size_type indent)
{
    static const string m_desc="CONST ";
    ostringstream o;
    string::size_type ii;
    for(ii=0; ii<indent; ++ii)
        o << " ";
    o << m_desc << m_var_id << " " << m_name.c_str();
    o << "\n";
    return o.str();
}

string c_sub_exp_func::emit_human_desc(string::size_type indent)
{
    static const string m_desc="FUNC_CALL ";
    ostringstream o;
    string::size_type ii;
    for(ii=0; ii<indent; ++ii)
        o << " ";
    o << m_desc + " \"" << m_name.c_str() << "\"";
    o << "\n";
    indent+=2;
    vector<c_sub_exp*>::iterator i1;
    for(i1=m_args.begin(); i1!=m_args.end(); ++i1)
        o << (*i1)->emit_human_desc(indent).c_str();
    o << "\n";
    return o.str();
}

string c_sub_exp_eval_str::emit_human_desc(string::size_type indent)
{
    static const string m_desc="EVAL_STRING ";
    ostringstream o;
    string::size_type ii;
    for(ii=0; ii<indent; ++ii)
        o << " ";
    o << m_desc << "\"" << m_str.str() << "\"";
    o << "\n";
    return o.str();
}

string c_sub_exp_eval_str::emit_human_microcode(string::size_type indent)
{
    static const string m_desc="EVAL_STRING ";
    ostringstream o;
    string::size_type ii;
    for(ii=0; ii<indent; ++ii)
        o << " ";
    o << m_desc << m_var_id << " " << m_str.str();
    o << "\n";
    return o.str();
}

string c_sub_exp_immediate_str::emit_human_desc(string::size_type indent)
{
    static const string m_desc="IMMEDIATE_STRING ";
    ostringstream o;
    string::size_type ii;
    for(ii=0; ii<indent; ++ii)
        o << " ";
    o << m_desc << "'" << m_str.str() << "'";
    o << "\n";
    return o.str();
}

string c_sub_exp_immediate_str::emit_human_microcode(string::size_type indent)
{
    static const string m_desc="IMMEDIATE_STRING ";
    ostringstream o;
    string::size_type ii;
    for(ii=0; ii<indent; ++ii)
        o << " ";
    o << m_desc << m_var_id << " " << m_str.str();
    o << "\n";
    return o.str();
}

string c_sub_exp_oper::emit_human_desc(string::size_type indent)
{
    static const string m_desc="OPERATOR ";
    ostringstream o;
    string::size_type ii;
    for(ii=0; ii<indent; ++ii)
        o << " ";
    o << m_desc + " \"" << m_oper.oper << "\"";
    o << "\n";
    indent+=2;
    vector<c_sub_exp*>::iterator i1;
    for(i1=m_eval.begin(); i1!=m_eval.end(); ++i1) {
        if(typeid(**i1)==typeid(c_sub_exp_oper_ternary_false)) {
            // a special case
            o << dynamic_cast<c_sub_exp_oper_ternary_false*>(*i1)->emit_human_desc_no_recursion(indent).c_str();
            continue;
        }
        if(typeid(**i1)==typeid(c_sub_exp_parenthesis)) {
            // a special case
            o << dynamic_cast<c_sub_exp_parenthesis*>(*i1)->emit_human_desc_no_recursion(indent).c_str();
            continue;
        }
        if(this==*i1)
            continue;
        o << (*i1)->emit_human_desc(indent).c_str();
    }
    o << "\n";
    return o.str();
}

string c_sub_exp_oper::emit_human_microcode(string::size_type indent)
{
    static const string m_desc="OPERATOR ";
    ostringstream o;
    string::size_type ii;
    for(ii=0; ii<indent; ++ii)
        o << " ";
    o << m_desc << m_var_id << " \"" << m_oper.oper << "\"";
    o << "\n";
    indent+=2;
    vector<c_sub_exp*>::iterator i1;
    for(i1=m_eval.begin(); i1!=m_eval.end(); ++i1) {
        if(typeid(**i1)==typeid(c_sub_exp_oper_ternary_false)) {
            // a special case
            o << dynamic_cast<c_sub_exp_oper_ternary_false*>(*i1)->emit_human_microcode_no_recursion(indent).c_str();
            continue;
        }
        if(typeid(**i1)==typeid(c_sub_exp_parenthesis)) {
            // a special case
            o << dynamic_cast<c_sub_exp_parenthesis*>(*i1)->emit_human_microcode_no_recursion(indent).c_str();
            continue;
        }
        if(this==*i1)
            continue;
        o << (*i1)->emit_human_microcode(indent).c_str();
    }
    o << "\n";
    return o.str();
}

string c_sub_exp_end_parenthesis::emit_human_desc(string::size_type indent)
{
    static const string m_desc="END_PARENTHESIS )";
    ostringstream o;
    string::size_type ii;
    for(ii=0; ii<indent; ++ii)
        o << " ";
    o << m_desc;
    o << "\n";
    return o.str();
}

string c_sub_exp_end_parenthesis::emit_human_microcode(string::size_type indent)
{
    static const string m_desc="<END_PARENTHESIS>\n";
    ostringstream o;
    string::size_type ii;
    for(ii=0; ii<indent; ++ii)
        o << " ";
    o << m_desc;
    return o.str();
}

string c_sub_exp_end_index::emit_human_desc(string::size_type indent)
{
    static const string m_desc="END_INDEX ]";
    ostringstream o;
    string::size_type ii;
    for(ii=0; ii<indent; ++ii)
        o << " ";
    o << m_desc;
    o << "\n";
    return o.str();
}

string c_sub_exp_end_index::emit_human_microcode(string::size_type indent)
{
    static const string m_desc="<END_INDEX>\n";
    ostringstream o;
    string::size_type ii;
    for(ii=0; ii<indent; ++ii)
        o << " ";
    o << m_desc;
    return o.str();
}

string c_sub_exp_func_end_args::emit_human_desc(string::size_type indent)
{
    static const string m_desc="FUNC_END_ARGS )";
    ostringstream o;
    string::size_type ii;
    for(ii=0; ii<indent; ++ii)
        o << " ";
    o << m_desc;
    o << "\n";
    return o.str();
}

string c_sub_exp_func_end_args::emit_human_microcode(string::size_type indent)
{
    static const string m_desc="<FUNC_END_ARGS>\n";
    ostringstream o;
    string::size_type ii;
    for(ii=0; ii<indent; ++ii)
        o << " ";
    o << m_desc;
    return o.str();
}

string c_sub_exp_func::emit_human_microcode(string::size_type indent)
{
    static const string m_desc="FUNC_CALL ";
    ostringstream o;
    string::size_type ii;
    for(ii=0; ii<indent; ++ii)
        o << " ";
    o << m_desc.c_str() << m_var_id << " " << m_name.c_str() << "\"";
    o << "\n";
    indent+=2;
    vector<c_sub_exp*>::iterator i1;
    for(i1=m_args.begin(); i1!=m_args.end(); ++i1)
        o << (*i1)->emit_human_microcode(indent).c_str();
    o << "\n";
    return o.str();
}

string c_sub_exp_index::emit_human_desc(string::size_type indent)
{
    static const string m_desc="INDEX [ ";
    ostringstream o;
    string::size_type ii;
    for(ii=0; ii<indent; ++ii)
        o << " ";
    o << m_desc;
    o << "\n";
    indent+=2;
    vector<c_sub_exp*>::iterator i1;
    for(i1=m_idx.begin(); i1!=m_idx.end(); ++i1)
        o << (*i1)->emit_human_desc(indent).c_str();
    o << "\n";
    return o.str();
}

string c_sub_exp_index::emit_human_microcode(string::size_type indent)
{
    static const string m_desc="INDEX ";
    ostringstream o;
    string::size_type ii;
    for(ii=0; ii<indent; ++ii)
        o << " ";
    o << m_desc.c_str() << m_var_id;
    o << "\n";
    indent+=2;
    vector<c_sub_exp*>::iterator i1;
    for(i1=m_idx.begin(); i1!=m_idx.end(); ++i1)
        o << (*i1)->emit_human_microcode(indent).c_str();
    o << "\n";
    return o.str();
}

string c_sub_exp_oper_ternary::emit_human_desc(string::size_type indent)
{
    static const string m_desc="TERNARY ";
    ostringstream o;
    string::size_type ii;
    for(ii=0; ii<indent; ++ii)
        o << " ";
    o << m_desc;
    o << "\n";
    indent+=2;
    vector<c_sub_exp*>::iterator i1;
    for(i1=m_eval.begin(); i1!=m_eval.end(); ++i1)
        o << (*i1)->emit_human_desc(indent).c_str();
    o << "\n";
    return o.str();
}

string c_sub_exp_oper_ternary::emit_human_microcode(string::size_type indent)
{
    static const string m_desc="TERNARY ";
    ostringstream o;
    string::size_type ii;
    for(ii=0; ii<indent; ++ii)
        o << " ";
    o << m_desc.c_str() << m_var_id;
    o << "\n";
    indent+=2;
    vector<c_sub_exp*>::iterator i1;
    for(i1=m_eval.begin(); i1!=m_eval.end(); ++i1)
        o << (*i1)->emit_human_microcode(indent).c_str();
    o << "\n";
    return o.str();
}

string c_sub_exp_oper_ternary_false::emit_human_desc(string::size_type indent)
{
    static const string m_desc="TERNARY_FALSE ";
    ostringstream o;
    string::size_type ii;
    for(ii=0; ii<indent; ++ii)
        o << " ";
    o << m_desc;
    o << "\n";
    indent+=2;
    vector<c_sub_exp*>::iterator i1;
    for(i1=m_eval.begin(); i1!=m_eval.end(); ++i1)
        o << (*i1)->emit_human_desc(indent).c_str();
    o << "\n";
    return o.str();
}

string c_sub_exp_oper_ternary_false::emit_human_microcode(string::size_type indent)
{
    static const string m_desc="TERNARY_FALSE ";
    ostringstream o;
    string::size_type ii;
    for(ii=0; ii<indent; ++ii)
        o << " ";
    o << m_desc.c_str() << m_var_id;
    o << "\n";
    indent+=2;
    vector<c_sub_exp*>::iterator i1;
    for(i1=m_eval.begin(); i1!=m_eval.end(); ++i1)
        o << (*i1)->emit_human_microcode(indent).c_str();
    o << "\n";
    return o.str();
}

void c_sub_exp_oper_ternary::compile(c_tree_microcode& tm_callback)
{
    if(NULL!=m_other_ternary) {
        m_var_id=m_other_ternary->get_var_id();
        return;
    }
    if(VAR_SYM_INDEX_INVALID==m_var_id)
        m_var_id=tm_callback.add_temp_var();
}

void c_sub_exp_oper_ternary_false::compile(c_tree_microcode& UNUSEDARG(tm_callback))
{
    m_var_id=m_ternary->get_var_id();
}

string c_sub_exp_oper_ternary_false::emit_human_desc_no_recursion(string::size_type indent)
{
    static const string m_desc="TERNARY_FALSE ";
    ostringstream o;
    string::size_type ii;
    for(ii=0; ii<indent; ++ii)
        o << " ";
    o << m_desc;
    o << "\n";
    return o.str();
}

string c_sub_exp_oper_ternary_false::emit_human_microcode_no_recursion(string::size_type indent)
{
    static const string m_desc="TERNARY_FALSE ";
    ostringstream o;
    string::size_type ii;
    for(ii=0; ii<indent; ++ii)
        o << " ";
    o << m_desc.c_str() << m_var_id;
    o << "\n";
    return o.str();
}

string c_sub_exp_parenthesis::emit_human_desc_no_recursion(string::size_type indent)
{
    static const string m_desc="PARENTHESIS (";
    ostringstream o;
    string::size_type ii;
    for(ii=0; ii<indent; ++ii)
        o << " ";
    o << m_desc;
    o << "\n";
    return o.str();
}

string c_sub_exp_parenthesis::emit_human_microcode_no_recursion(string::size_type indent)
{
    static const string m_desc="PARENTHESIS ";
    ostringstream o;
    string::size_type ii;
    for(ii=0; ii<indent; ++ii)
        o << " ";
    o << m_desc.c_str() << m_var_id;
    o << "\n";
    return o.str();
}

bool c_sub_exp_parenthesis::emit_evaluation_order(list<c_sub_exp*>& UNUSEDARG(st), list<c_sub_exp*>::iterator UNUSEDARG(where))
{
    if(m_evaluation_order_emited)
        return true;

    m_evaluation_order_emited=true;

    return false;
}

bool c_sub_exp_index::emit_evaluation_order(list<c_sub_exp*>& st, list<c_sub_exp*>::iterator where)
{
    if(m_evaluation_order_emited)
        return true;

    if(typeid(*m_base_exp)!=typeid(c_sub_exp_variable))
        where=st.insert(where,m_base_exp);
    vector<c_sub_exp*>::iterator i1;
    for(i1=m_idx.begin(); i1!=m_idx.end(); ++i1)
        where=st.insert(where,*i1);

    m_evaluation_order_emited=true;
    return false;
}

bool c_sub_exp_variable::emit_evaluation_order(list<c_sub_exp*>& UNUSEDARG(st), list<c_sub_exp*>::iterator UNUSEDARG(where))
{
    if(m_evaluation_order_emited)
        return true;

    m_evaluation_order_emited=true;
    return false;
}

bool c_sub_exp_numeric_immediate::emit_evaluation_order(list<c_sub_exp*>& UNUSEDARG(st), list<c_sub_exp*>::iterator UNUSEDARG(where))
{
    if(m_evaluation_order_emited)
        return true;

    m_evaluation_order_emited=true;
    return false;
}

bool c_sub_exp_const::emit_evaluation_order(list<c_sub_exp*>& UNUSEDARG(st), list<c_sub_exp*>::iterator UNUSEDARG(where))
{
    if(m_evaluation_order_emited)
        return true;

    m_evaluation_order_emited=true;
    return false;
}

bool c_sub_exp_func::emit_evaluation_order(list<c_sub_exp*>& UNUSEDARG(st), list<c_sub_exp*>::iterator UNUSEDARG(where))
{
    if(m_evaluation_order_emited)
        return true;

    m_evaluation_order_emited=true;
    return false;
}

bool c_sub_exp_eval_str::emit_evaluation_order(list<c_sub_exp*>& UNUSEDARG(st), list<c_sub_exp*>::iterator UNUSEDARG(where))
{
    if(m_evaluation_order_emited)
        return true;

    m_evaluation_order_emited=true;
    return false;
}

bool c_sub_exp_immediate_str::emit_evaluation_order(list<c_sub_exp*>& UNUSEDARG(st), list<c_sub_exp*>::iterator UNUSEDARG(where))
{
    if(m_evaluation_order_emited)
        return true;

    m_evaluation_order_emited=true;
    return false;
}

bool c_sub_exp_oper::emit_evaluation_order(list<c_sub_exp*>& UNUSEDARG(st), list<c_sub_exp*>::iterator UNUSEDARG(where))
{
    if(m_evaluation_order_emited)
        return true;

    m_evaluation_order_emited=true;

    return false;
}

bool c_sub_exp_oper_ternary::emit_evaluation_order(list<c_sub_exp*>& UNUSEDARG(st), list<c_sub_exp*>::iterator UNUSEDARG(where))
{
    if(m_evaluation_order_emited)
        return true;

    m_evaluation_order_emited=true;

    return false;
}

bool c_sub_exp_oper_ternary_false::emit_evaluation_order(list<c_sub_exp*>& UNUSEDARG(st), list<c_sub_exp*>::iterator UNUSEDARG(where))
{
    if(m_evaluation_order_emited)
        return true;

    m_evaluation_order_emited=true;

    return false;
}

bool c_sub_exp::emit_evaluation_order(list<c_sub_exp*>& UNUSEDARG(st), list<c_sub_exp*>::iterator UNUSEDARG(where))
{
    if(m_evaluation_order_emited)
        return true;

    m_evaluation_order_emited=true;
    return false;
}

void c_sub_exp::reset_evaluation_order_emited()
{
    m_evaluation_order_emited=false;
}

bool c_sub_exp::compiled()
{
    return VAR_SYM_INDEX_INVALID!=m_var_id;
}

void c_sub_exp::set_depth(int depth)
{
    m_depth=depth;
}

int c_sub_exp::get_depth()
{
    return m_depth;
}

void c_sub_exp_root::add_sub(c_sub_exp* sub)
{
    m_eval.push_back(sub);
}

int c_sub_exp_root::get_prec()
{
    return 0;
}

void c_sub_exp_parenthesis::add_sub(c_sub_exp* sub)
{
    m_eval.push_back(sub);
}

int c_sub_exp_parenthesis::get_prec()
{
    return 2;
}

bool c_sub_exp_parenthesis::complete(tree<c_sub_exp*>& t)
{
    tree<c_sub_exp*>::iterator i1;
    for(i1=t.begin(); i1!=t.end(); ++i1)
        if(this==*i1)
            if(t.number_of_children(i1)!=0)
                return false;
    return m_eval.size()!=0;
}

void c_sub_exp_index::add_index(c_sub_exp* idx)
{
    m_idx.push_back(idx);
}

int c_sub_exp_index::get_prec()
{
    return 3;
}

bool c_sub_exp_index::complete(tree<c_sub_exp*>& UNUSEDARG(t))
{
    if(NULL==m_base_exp)
        return false;
    return true; // can be also []
}

void c_sub_exp_index::set_base(c_sub_exp* base)
{
    MY_ASSERT(NULL==m_base_exp || m_base_exp==base);
    if(NULL!=m_base_exp)
        return;
    m_base_exp=base;
}

string c_sub_exp_variable::get_varname()
{
    return m_name;
}

int c_sub_exp_variable::get_prec()
{
    return OPER_VAR;
}

MY_LONG_LONG c_sub_exp_numeric_immediate::get_num()
{
    return m_num;
}

int c_sub_exp_numeric_immediate::get_prec()
{
    return OPER_VAR;
}

string c_sub_exp_const::get_const()
{
    return m_name;
}

int c_sub_exp_const::get_prec()
{
    return OPER_VAR;
}

string c_sub_exp_func::get_func_name()
{
    return m_name;
}

void c_sub_exp_func::add_arg(c_sub_exp* arg)
{
    m_args.push_back(arg);
}

int c_sub_exp_func::get_prec()
{
    return 2; // 2 == "("
}

bool c_sub_exp_func::complete(tree<c_sub_exp*>& UNUSEDARG(t))
{
    return m_complete;
}

void c_sub_exp_func::set_complete()
{
    m_complete=true;
}

void c_sub_exp_eval_str::append(char ch)
{
    m_str << ch;
}

string c_sub_exp_eval_str::get_eval_str()
{
    return m_str.str();
}

int c_sub_exp_eval_str::get_prec()
{
    return OPER_VAR;
}

void c_sub_exp_immediate_str::append(char ch)
{
    m_str << ch;
}

string c_sub_exp_immediate_str::get_imm_str()
{
    return m_str.str();
}

int c_sub_exp_immediate_str::get_prec()
{
    return OPER_VAR;
}

int c_sub_exp_oper::get_prec()
{
    return m_oper.get_prec();
}

const c_oper_desc& c_sub_exp_oper::get_oper() const
{
    return m_oper;
}

bool c_sub_exp_oper::add_eval(c_sub_exp* el)
{
    vector<c_sub_exp*>::iterator i1;
    for(i1=m_eval.begin(); i1!=m_eval.end(); ++i1)
        if(*i1==el)
            return true;
    m_eval.push_back(el);
    return false;
}

bool c_sub_exp_oper::complete(tree<c_sub_exp*>& UNUSEDARG(t))
{
    bool unary_or_binary=false;
    if('-'==get_oper().oper[0] && 0==get_oper().oper[1])
        unary_or_binary=true;
    else if('+'==get_oper().oper[0] && 0==get_oper().oper[1])
        unary_or_binary=true;
    if(unary_or_binary)
        return (m_eval.size()>=1);

    switch(m_oper.get_type()) {
        case OPER_UNARY:
            if(m_oper.get_prec()==25) // "return (string)"
                return (m_eval.size()==1 || m_eval.size()==0);
            return (m_eval.size()==1);
        case OPER_BINARY:
            return (m_eval.size()==2);
        case OPER_TERNARY:
            return m_eval.size()==1;
        case OPER_TERNARY_FALSE:
            return (m_eval.size()>=1);
        default:
            break;
    }
    MY_ASSERT_NOTREACH();
    return false;
}

int c_sub_exp_oper_ternary_false::get_prec()
{
    return m_oper.get_prec();
}

const c_oper_desc& c_sub_exp_oper_ternary_false::get_oper() const
{
    return m_oper;
}

bool c_sub_exp_oper_ternary_false::add_eval(c_sub_exp* el)
{
    vector<c_sub_exp*>::iterator i1;
    for(i1=m_eval.begin(); i1!=m_eval.end(); ++i1)
        if(*i1==el)
            return true;
    m_eval.push_back(el);
    return false;
}

bool c_sub_exp_oper_ternary_false::complete(tree<c_sub_exp*>& UNUSEDARG(t))
{
    return m_eval.size()>0;
}

int c_sub_exp_end_parenthesis::get_prec()
{
    return OPER_PARENTHESIS;
}

int c_sub_exp_end_index::get_prec()
{
    return OPER_PARENTHESIS;
}

int c_sub_exp_func_end_args::get_prec()
{
    return OPER_PARENTHESIS;
}

void c_sub_exp_index::emit_microcode(c_expression& expression, c_micro_symbol_variable_mapping& UNUSEDARG(symbols))
{
    MY_ASSERT(NULL!=m_base_exp);
    MY_ASSERT(VAR_SYM_INDEX_INVALID!=m_base_exp->get_var_id());

    var_sym_index_t subscript=VAR_SYM_INDEX_NONE;   // "[]"
    var_sym_index_t output=this->get_var_id();
    var_sym_index_t base=m_base_exp->get_var_id();

    vector<c_sub_exp*>::iterator i1=m_idx.end();
    if(!m_idx.empty()) {
        if(typeid(**(--i1))==typeid(c_sub_exp_oper) && dynamic_cast<c_sub_exp_oper*>(*i1)->get_oper().precedence==1)
            m_idx.erase(i1); // "/i" operator
    }

    i1=m_idx.begin();
    if(m_idx.size()==2) {
        // first is base of subscript
        ++i1;
        if(i1!=m_idx.end()) {
            // second is evaluated subscript result
            subscript=(*i1)->get_var_id();
        }
    } else if(m_idx.size()==1) {
        subscript=(*i1)->get_var_id();
    } else if(m_idx.size()>2) {
        while(dynamic_cast<c_sub_exp_oper*>(*i1)==NULL)
            ++i1;
        subscript=(*i1)->get_var_id();
    }

    if(!m_idx.empty())
        subscript=get_result()->get_var_id();

    MY_ASSERT(VAR_SYM_INDEX_INVALID!=subscript);
    MY_ASSERT(VAR_SYM_INDEX_INVALID!=output);
    MY_ASSERT(VAR_SYM_INDEX_INVALID!=base);

    expression.add_micro(MY_NEW c_micro_index(base,subscript,output,m_case_insensitive));
}

void c_sub_exp_variable::emit_microcode(c_expression& UNUSEDARG(expression), c_micro_symbol_variable_mapping& UNUSEDARG(symbols))
{
}

void c_sub_exp_numeric_immediate::emit_microcode(c_expression& UNUSEDARG(expression), c_micro_symbol_variable_mapping& UNUSEDARG(symbols))
{
}

void c_sub_exp_const::emit_microcode(c_expression& UNUSEDARG(expression), c_micro_symbol_variable_mapping& UNUSEDARG(symbols))
{
}

void c_sub_exp_func::emit_microcode(c_expression& expression, c_micro_symbol_variable_mapping& UNUSEDARG(symbols))
{
    if(!m_name.compare("isset") && m_args.size()==1) {
        expression.add_micro(MY_NEW c_micro_isset(m_args.front()->get_var_id(),this->get_var_id()));
        return;
    }
    if(m_name.compare("array")) { // !=
        c_micro_func_call* fc;
        expression.add_micro(fc=MY_NEW c_micro_func_call(get_func_name()));
        vector<c_sub_exp*>::iterator i1;
        for(i1=m_args.begin(); i1!=m_args.end(); ++i1) {
            if(typeid(c_sub_exp_oper)==typeid(**i1)) {
                c_sub_exp_oper* o=dynamic_cast<c_sub_exp_oper*>(*i1);
                if(','==o->get_oper().oper[0]) {
                    fc->add_arg(o->get_eval()[0]->get_var_id());
                    fc->add_arg(o->get_eval()[1]->get_var_id());
                    continue;
                }
            }
            fc->add_arg((*i1)->get_var_id());
        }
        fc->set_output(this->get_var_id());
    } else {
        // array()
        c_micro_array* fa;
        expression.add_micro(fa=MY_NEW c_micro_array());
        vector<pair<var_sym_index_t, var_sym_index_t> >::iterator i1;
        for(i1=m_array_args.begin(); i1!=m_array_args.end(); ++i1) {
            fa->add_arg((*i1).first,(*i1).second);
        }
        fa->set_output(this->get_var_id());
    }
}

void c_sub_exp_eval_str::emit_microcode(c_expression& UNUSEDARG(expression), c_micro_symbol_variable_mapping& UNUSEDARG(symbols))
{
    // NOOP
}

void c_sub_exp_immediate_str::emit_microcode(c_expression& UNUSEDARG(expression), c_micro_symbol_variable_mapping& UNUSEDARG(symbols))
{
    // NOOP
}

void c_sub_exp_oper::emit_microcode(c_expression& expression, c_micro_symbol_variable_mapping& symbols)
{
    c_micro_assign* ma;

    switch(m_oper.oper[0]) {
        case '-':
            switch(m_oper.oper[1]) {
                case 0:
                    switch(m_eval.size()) {
                        case 1:
                            expression.add_micro(MY_NEW c_micro_minus(symbols.add("0",&expression),m_eval[0]->get_var_id(),this->get_var_id()));
                            return;
                        case 2:
                            expression.add_micro(MY_NEW c_micro_minus(m_eval[0]->get_var_id(),m_eval[1]->get_var_id(),this->get_var_id()));
                            return;
                    }
                    MY_ASSERT_NOTREACH();
                    break;
                case '-':
                    MY_ASSERT(m_eval.size()==1);
                    if(m_postfix)
                        expression.add_micro(MY_NEW c_micro_post_decrement(m_eval[0]->get_var_id(),this->get_var_id()));
                    else
                        expression.add_micro(MY_NEW c_micro_pre_decrement(m_eval[0]->get_var_id(),this->get_var_id()));
                    return;
                case '=':
                    MY_ASSERT(m_eval.size()==2);
                    expression.add_micro(MY_NEW c_micro_minus(m_eval[0]->get_var_id(),m_eval[1]->get_var_id(),this->get_var_id()));
                    expression.add_micro(ma=MY_NEW c_micro_assign(m_eval[0]->get_var_id(),this->get_var_id()));
                    ma->set_output_must_be_lvalue();
                    return;
            }
            MY_ASSERT_NOTREACH();
            break;
        case '+':
            switch(m_oper.oper[1]) {
                case 0:
                    switch(m_eval.size()) {
                        case 1:
                            expression.add_micro(MY_NEW c_micro_plus(symbols.add("0",&expression),m_eval[0]->get_var_id(),this->get_var_id()));
                            return;
                        case 2:
                            expression.add_micro(MY_NEW c_micro_plus(m_eval[0]->get_var_id(),m_eval[1]->get_var_id(),this->get_var_id()));
                            return;
                    }
                    MY_ASSERT_NOTREACH();
                    break;
                case '+':
                    MY_ASSERT(m_eval.size()==1);
                    if(m_postfix)
                        expression.add_micro(MY_NEW c_micro_post_increment(m_eval[0]->get_var_id(),this->get_var_id()));
                    else
                        expression.add_micro(MY_NEW c_micro_pre_increment(m_eval[0]->get_var_id(),this->get_var_id()));
                    return;
                case '=':
                    MY_ASSERT(m_eval.size()==2);
                    expression.add_micro(MY_NEW c_micro_plus(m_eval[0]->get_var_id(),m_eval[1]->get_var_id(),this->get_var_id()));
                    expression.add_micro(ma=MY_NEW c_micro_assign(m_eval[0]->get_var_id(),this->get_var_id()));
                    ma->set_output_must_be_lvalue();
                    return;
            }
            MY_ASSERT_NOTREACH();
            break;
        case '~':
            switch(m_oper.oper[1]) {
                case 0:
                    MY_ASSERT(m_eval.size()==1);
                    expression.add_micro(MY_NEW c_micro_one_compl(m_eval[0]->get_var_id(),this->get_var_id()));
                    return;
                case '=':
                    MY_ASSERT(m_eval.size()==2);
                    expression.add_micro(MY_NEW c_micro_one_compl(m_eval[1]->get_var_id(),this->get_var_id()));
                    expression.add_micro(ma=MY_NEW c_micro_assign(m_eval[0]->get_var_id(),this->get_var_id()));
                    ma->set_output_must_be_lvalue();
                    return;
            }
            MY_ASSERT_NOTREACH();
            break;
        case '(':
            switch(m_oper.oper[1]) {
                case 'i': // (int)
                    MY_ASSERT(m_eval.size()==1);
                    expression.add_micro(MY_NEW c_micro_cast_int(m_eval[0]->get_var_id(),this->get_var_id()));
                    return;
                case 's': // (string)
                    MY_ASSERT(m_eval.size()==1);
                    expression.add_micro(MY_NEW c_micro_cast_string(m_eval[0]->get_var_id(),this->get_var_id()));
                    return;
                case 'b': // (bool)
                    MY_ASSERT(m_eval.size()==1);
                    expression.add_micro(MY_NEW c_micro_cast_bool(m_eval[0]->get_var_id(),this->get_var_id()));
                    return;
            }
            MY_ASSERT_NOTREACH();
            break;
        case '!':
            switch(m_oper.oper[1]) {
                case 0:
                    MY_ASSERT(m_eval.size()==1);
                    expression.add_micro(MY_NEW c_micro_not(m_eval[0]->get_var_id(),this->get_var_id()));
                    return;
                case '=':
                    MY_ASSERT(m_eval.size()==2);
                    if(0==m_oper.oper[2]) {
                        expression.add_micro(MY_NEW c_micro_not_equal(m_eval[0]->get_var_id(),m_eval[1]->get_var_id(),this->get_var_id()));
                        return;
                    } else {
                        expression.add_micro(MY_NEW c_micro_not_identical(m_eval[0]->get_var_id(),m_eval[1]->get_var_id(),this->get_var_id()));
                        return;
                    }
                    break;
            }
            MY_ASSERT_NOTREACH();
            break;
        case '*':
            switch(m_oper.oper[1]) {
                case 0:
                    MY_ASSERT(m_eval.size()==2);
                    expression.add_micro(MY_NEW c_micro_mul(m_eval[0]->get_var_id(),m_eval[1]->get_var_id(),this->get_var_id()));
                    return;
                case '=':
                    MY_ASSERT(m_eval.size()==2);
                    expression.add_micro(MY_NEW c_micro_mul(m_eval[0]->get_var_id(),m_eval[1]->get_var_id(),this->get_var_id()));
                    expression.add_micro(ma=MY_NEW c_micro_assign(m_eval[0]->get_var_id(),this->get_var_id()));
                    ma->set_output_must_be_lvalue();
                    return;
            }
            MY_ASSERT_NOTREACH();
            break;
        case '/':
            switch(m_oper.oper[1]) {
                case 0:
                    MY_ASSERT(m_eval.size()==2);
                    expression.add_micro(MY_NEW c_micro_div(m_eval[0]->get_var_id(),m_eval[1]->get_var_id(),this->get_var_id()));
                    return;
                case '=':
                    MY_ASSERT(m_eval.size()==2);
                    expression.add_micro(MY_NEW c_micro_div(m_eval[0]->get_var_id(),m_eval[1]->get_var_id(),this->get_var_id()));
                    expression.add_micro(ma=MY_NEW c_micro_assign(m_eval[0]->get_var_id(),this->get_var_id()));
                    ma->set_output_must_be_lvalue();
                    return;
                case 'i':
                    // ignore this
                    return;
            }
            MY_ASSERT_NOTREACH();
            break;
        case '%':
            switch(m_oper.oper[1]) {
                case 0:
                    MY_ASSERT(m_eval.size()==2);
                    expression.add_micro(MY_NEW c_micro_mod(m_eval[0]->get_var_id(),m_eval[1]->get_var_id(),this->get_var_id()));
                    return;
                case '=':
                    MY_ASSERT(m_eval.size()==2);
                    expression.add_micro(MY_NEW c_micro_mod(m_eval[0]->get_var_id(),m_eval[1]->get_var_id(),this->get_var_id()));
                    expression.add_micro(ma=MY_NEW c_micro_assign(m_eval[0]->get_var_id(),this->get_var_id()));
                    ma->set_output_must_be_lvalue();
                    return;
            }
            MY_ASSERT_NOTREACH();
            break;
        case '.':
            switch(m_oper.oper[1]) {
                case 0:
                    MY_ASSERT(m_eval.size()==2);
                    expression.add_micro(MY_NEW c_micro_concat(m_eval[0]->get_var_id(),m_eval[1]->get_var_id(),this->get_var_id()));
                    return;
                case '=':
                    MY_ASSERT(m_eval.size()==2);
                    expression.add_micro(MY_NEW c_micro_concat(m_eval[0]->get_var_id(),m_eval[1]->get_var_id(),this->get_var_id()));
                    expression.add_micro(ma=MY_NEW c_micro_assign(m_eval[0]->get_var_id(),this->get_var_id()));
                    ma->set_output_must_be_lvalue();
                    return;
            }
            MY_ASSERT_NOTREACH();
            break;
        case '<':
            switch(m_oper.oper[1]) {
                case 0:
                    MY_ASSERT(m_eval.size()==2);
                    expression.add_micro(MY_NEW c_micro_less(m_eval[0]->get_var_id(),m_eval[1]->get_var_id(),this->get_var_id()));
                    return;
                case '<':
                    MY_ASSERT(m_eval.size()==2);
                    expression.add_micro(MY_NEW c_micro_shift_left(m_eval[0]->get_var_id(),m_eval[1]->get_var_id(),this->get_var_id()));
                    return;
                case '=':
                    if(0==m_oper.oper[2]) {
                        MY_ASSERT(m_eval.size()==2);
                        expression.add_micro(MY_NEW c_micro_less_or_equal(m_eval[0]->get_var_id(),m_eval[1]->get_var_id(),this->get_var_id()));
                        return;
                    } else {
                        MY_ASSERT(m_eval.size()==2);
                        expression.add_micro(MY_NEW c_micro_shift_left(m_eval[0]->get_var_id(),m_eval[1]->get_var_id(),this->get_var_id()));
                        expression.add_micro(ma=MY_NEW c_micro_assign(m_eval[0]->get_var_id(),this->get_var_id()));
                        ma->set_output_must_be_lvalue();
                        return;
                    }
                    break;
            }
            MY_ASSERT_NOTREACH();
            break;
        case '>':
            switch(m_oper.oper[1]) {
                case 0:
                    MY_ASSERT(m_eval.size()==2);
                    expression.add_micro(MY_NEW c_micro_greater(m_eval[0]->get_var_id(),m_eval[1]->get_var_id(),this->get_var_id()));
                    return;
                case '>':
                    MY_ASSERT(m_eval.size()==2);
                    expression.add_micro(MY_NEW c_micro_shift_right(m_eval[0]->get_var_id(),m_eval[1]->get_var_id(),this->get_var_id()));
                    return;
                case '=':
                    if(0==m_oper.oper[2]) {
                        MY_ASSERT(m_eval.size()==2);
                        expression.add_micro(MY_NEW c_micro_greater_or_equal(m_eval[0]->get_var_id(),m_eval[1]->get_var_id(),this->get_var_id()));
                        return;
                    } else {
                        MY_ASSERT(m_eval.size()==2);
                        expression.add_micro(MY_NEW c_micro_shift_right(m_eval[0]->get_var_id(),m_eval[1]->get_var_id(),this->get_var_id()));
                        expression.add_micro(ma=MY_NEW c_micro_assign(m_eval[0]->get_var_id(),this->get_var_id()));
                        ma->set_output_must_be_lvalue();
                        return;
                    }
                    break;
            }
            MY_ASSERT_NOTREACH();
            break;
        case '=':
            switch(m_oper.oper[1]) {
                case 0:
                    MY_ASSERT(m_eval.size()==2);
                    expression.add_micro(ma=MY_NEW c_micro_assign(m_eval[1]->get_var_id(),m_eval[0]->get_var_id()));
                    ma->set_output_must_be_lvalue();
                    return;
                case '=':
                    if(0==m_oper.oper[2]) {
                        MY_ASSERT(m_eval.size()==2);
                        expression.add_micro(MY_NEW c_micro_equal(m_eval[0]->get_var_id(),m_eval[1]->get_var_id(),this->get_var_id()));
                        return;
                    } else {
                        MY_ASSERT(m_eval.size()==2);
                        expression.add_micro(MY_NEW c_micro_identical(m_eval[0]->get_var_id(),m_eval[1]->get_var_id(),this->get_var_id()));
                        return;
                    }
                    break;
                case '>':
                    // TODO: do an user's error
                    break;
            }
            MY_ASSERT_NOTREACH();
            break;
        case '&':
            switch(m_oper.oper[1]) {
                case 0:
                    MY_ASSERT(m_eval.size()==2);
                    expression.add_micro(MY_NEW c_micro_bit_and(m_eval[0]->get_var_id(),m_eval[1]->get_var_id(),this->get_var_id()));
                    return;
                case '&':
                    {
                        MY_ASSERT(m_eval.size()==2);

                        bool is_last=true;
                        if(!m_same_prec_opers.empty() && m_same_prec_opers.back()!=this)
                            is_last=false;

                        c_micro_log_and* o1;

                        expression.add_micro(o1 = MY_NEW c_micro_log_and(m_eval[0]->get_var_id(),m_eval[1]->get_var_id(),this->get_var_id(),!is_last));

                        add_conditional_micro_and(o1);

                        vector<c_sub_exp_oper*>& all_conditionals=get_all_conditionals();
                        if(!all_conditionals.empty()) {
                            vector<c_sub_exp_oper*>::iterator i3;
                            for(i3=all_conditionals.begin(); i3!=all_conditionals.end(); ++i3) {
                                (*i3)->add_conditional_micro_and(o1);
                            }
                        }

                        return;
                    }
                    // NOTREACH
                case '=':
                    MY_ASSERT(m_eval.size()==2);
                    expression.add_micro(MY_NEW c_micro_bit_and(m_eval[0]->get_var_id(),m_eval[1]->get_var_id(),this->get_var_id()));
                    expression.add_micro(ma=MY_NEW c_micro_assign(m_eval[0]->get_var_id(),this->get_var_id()));
                    ma->set_output_must_be_lvalue();
                    return;
            }
            MY_ASSERT_NOTREACH();
            break;
        case '^':
            switch(m_oper.oper[1]) {
                case 0:
                    MY_ASSERT(m_eval.size()==2);
                    expression.add_micro(MY_NEW c_micro_bit_xor(m_eval[0]->get_var_id(),m_eval[1]->get_var_id(),this->get_var_id()));
                    return;
                case '=':
                    MY_ASSERT(m_eval.size()==2);
                    expression.add_micro(MY_NEW c_micro_bit_xor(m_eval[0]->get_var_id(),m_eval[1]->get_var_id(),this->get_var_id()));
                    expression.add_micro(ma=MY_NEW c_micro_assign(m_eval[0]->get_var_id(),this->get_var_id()));
                    ma->set_output_must_be_lvalue();
                    return;
            }
            MY_ASSERT_NOTREACH();
            break;
        case '|':
            switch(m_oper.oper[1]) {
                case 0:
                    MY_ASSERT(m_eval.size()==2);
                    expression.add_micro(MY_NEW c_micro_bit_or(m_eval[0]->get_var_id(),m_eval[1]->get_var_id(),this->get_var_id()));
                    return;
                case '|':
                    {
                        MY_ASSERT(m_eval.size()==2);

                        bool is_last=true;
                        if(!m_same_prec_opers.empty() && m_same_prec_opers.back()!=this)
                            is_last=false;

                        c_micro_log_or* o1;

                        expression.add_micro(o1 = MY_NEW c_micro_log_or(m_eval[0]->get_var_id(),m_eval[1]->get_var_id(),this->get_var_id(),!is_last));

                        add_conditional_micro_or(o1);

                        vector<c_sub_exp_oper*>& all_conditionals=get_all_conditionals();
                        if(!all_conditionals.empty()) {
                            vector<c_sub_exp_oper*>::iterator i3;
                            for(i3=all_conditionals.begin(); i3!=all_conditionals.end(); ++i3) {
                                (*i3)->add_conditional_micro_or(o1);
                            }
                        }

                        return;
                    }
                case '=':
                    MY_ASSERT(m_eval.size()==2);
                    expression.add_micro(MY_NEW c_micro_bit_or(m_eval[0]->get_var_id(),m_eval[1]->get_var_id(),this->get_var_id()));
                    expression.add_micro(ma=MY_NEW c_micro_assign(m_eval[0]->get_var_id(),this->get_var_id()));
                    ma->set_output_must_be_lvalue();
                    return;
            }
            MY_ASSERT_NOTREACH();
            break;
        case 'a': // and
            {
                MY_ASSERT(m_eval.size()==2);

                bool is_last=true;
                if(!m_same_prec_opers.empty() && m_same_prec_opers.back()!=this)
                    is_last=false;

                c_micro_log_and* o1;

                expression.add_micro(o1 = MY_NEW c_micro_log_and(m_eval[0]->get_var_id(),m_eval[1]->get_var_id(),this->get_var_id(),!is_last));

                add_conditional_micro_and(o1);

                vector<c_sub_exp_oper*>& all_conditionals=get_all_conditionals();
                if(!all_conditionals.empty()) {
                    vector<c_sub_exp_oper*>::iterator i3;
                    for(i3=all_conditionals.begin(); i3!=all_conditionals.end(); ++i3) {
                        (*i3)->add_conditional_micro_and(o1);
                    }
                }

                return;
            }
            break;
        case 'x': // xor
            {
                MY_ASSERT(m_eval.size()==2);
                expression.add_micro(MY_NEW c_micro_bit_xor(m_eval[0]->get_var_id(),m_eval[1]->get_var_id(),this->get_var_id()));
                return;
            }
            break;
        case 'o': // or
            {
                MY_ASSERT(m_eval.size()==2);

                bool is_last=true;
                if(!m_same_prec_opers.empty() && m_same_prec_opers.back()!=this)
                    is_last=false;

                c_micro_log_or* o1;

                expression.add_micro(o1 = MY_NEW c_micro_log_or(m_eval[0]->get_var_id(),m_eval[1]->get_var_id(),this->get_var_id(),!is_last));

                add_conditional_micro_or(o1);

                vector<c_sub_exp_oper*>& all_conditionals=get_all_conditionals();
                if(!all_conditionals.empty()) {
                    vector<c_sub_exp_oper*>::iterator i3;
                    for(i3=all_conditionals.begin(); i3!=all_conditionals.end(); ++i3) {
                        (*i3)->add_conditional_micro_or(o1);
                    }
                }

                return;
            }
            break;
        case 'r': // return
            if(!m_eval.empty()) {
                MY_ASSERT(m_eval.size()==1);
                expression.add_micro(MY_NEW c_micro_return(m_eval[0]->get_var_id()));
            } else
                expression.add_micro(MY_NEW c_micro_return(VAR_SYM_INDEX_INVALID));
            return;
        case ',':
            expression.add_micro(MY_NEW c_micro_assign(m_eval[1]->get_var_id(),this->get_var_id()));
            return;
    }
    MY_ASSERT_NOTREACH();
}

void c_sub_exp_oper_ternary::emit_microcode(c_expression& expression, c_micro_symbol_variable_mapping& UNUSEDARG(symbols))
{
    expression.add_micro(m_micro=MY_NEW c_micro_ternary(m_eval[0]->get_var_id()));
}

void c_sub_exp_oper_ternary_false::emit_microcode(c_expression& expression, c_micro_symbol_variable_mapping& UNUSEDARG(symbols))
{
    MY_ASSERT(m_eval.size()>0);
    MY_ASSERT(VAR_SYM_INDEX_INVALID!=get_result()->get_var_id());

    if(this->get_result()->get_var_id()!=this->get_var_id())
        expression.add_micro(MY_NEW c_micro_assign(this->get_result()->get_var_id(),this->get_var_id()));
}

void c_sub_exp_oper_ternary::add_condition(c_sub_exp* el)
{
    m_eval.insert(m_eval.begin(),el);
}

void c_sub_exp_index::set_case_insensitive()
{
    m_case_insensitive=true;
}

c_sub_exp* c_sub_exp_index::get_base()
{
    return m_base_exp;
}

void c_sub_exp_parenthesis::emit_microcode(c_expression& expression, c_micro_symbol_variable_mapping& UNUSEDARG(symbols))
{
    MY_ASSERT(!m_eval.empty());
    MY_ASSERT(VAR_SYM_INDEX_INVALID!=get_result()->get_var_id());
    MY_ASSERT(VAR_SYM_INDEX_INVALID!=this->get_var_id());

    expression.add_micro(MY_NEW c_micro_assign(get_result()->get_var_id(),this->get_var_id()));
}

void c_sub_exp_parenthesis::set_result(c_sub_exp* result)
{
    MY_ASSERT(NULL==m_result);

    m_result=result;
}

c_sub_exp* c_sub_exp_parenthesis::get_result()
{
    MY_ASSERT(m_result!=NULL);

    return m_result;
}

void c_sub_exp_index::set_result(c_sub_exp* result)
{
    MY_ASSERT(NULL==m_result);

    m_result=result;
}

c_sub_exp* c_sub_exp_index::get_result()
{
    MY_ASSERT(m_result!=NULL);

    return m_result;
}

void c_sub_exp_contitional_node_helper::add_sub(c_sub_exp* sub)
{
    m_eval.push_back(sub);
}

int c_sub_exp_contitional_node_helper::get_prec()
{
    return m_conditional_oper->get_prec();
}

bool c_sub_exp_contitional_node_helper::complete(tree<c_sub_exp*>& UNUSEDARG(t))
{
    return true;
}

bool c_sub_exp_contitional_node_helper::emit_evaluation_order(list<c_sub_exp*>& UNUSEDARG(st), list<c_sub_exp*>::iterator UNUSEDARG(where))
{
    if(m_evaluation_order_emited)
        return true;

    m_evaluation_order_emited=true;

    return false;
}

void c_sub_exp_contitional_node_helper::compile(c_tree_microcode& UNUSEDARG(tm_callback))
{
    m_var_id=VAR_SYM_INDEX_NONE;
    return;
}

string c_sub_exp_contitional_node_helper::emit_human_microcode(string::size_type UNUSEDARG(indent))
{
    return (string)"";
}

string c_sub_exp_contitional_node_helper::emit_human_microcode_no_recursion(string::size_type UNUSEDARG(indent))
{
    return (string)"";
}

void c_sub_exp_contitional_node_helper::emit_microcode(c_expression& UNUSEDARG(expression), c_micro_symbol_variable_mapping& UNUSEDARG(symbols))
{
    // NOOP
    return;
}

const vector<c_sub_exp*>& c_sub_exp_oper::get_eval() const
{
    return m_eval;
}

void c_sub_exp_oper::set_all_conditionals(const vector<c_sub_exp_oper*>& all_conditionals)
{
    m_all_conditionals=all_conditionals;
}

vector<c_sub_exp_oper*>& c_sub_exp_oper::get_all_conditionals()
{
    return m_all_conditionals;
}

class c_micro_log_or;
class c_micro_log_and;

void c_sub_exp_oper::add_conditional_micro_or(c_micro_log_or* m)
{
    if(m)
        m_all_conditionals_micro_or.push_back(m);
}

void c_sub_exp_oper::add_conditional_micro_and(c_micro_log_and* m)
{
    if(m)
        m_all_conditionals_micro_and.push_back(m);
}

vector<c_micro_log_or*>& c_sub_exp_oper::get_all_conditionals_micro_or()
{
    return m_all_conditionals_micro_or;
}

vector<c_micro_log_and*>& c_sub_exp_oper::get_all_conditionals_micro_and()
{
    return m_all_conditionals_micro_and;
}

int c_sub_exp_oper::get_oper_index() const
{
    return m_oper_index;
}

bool c_sub_exp_conditional_end_marker::emit_evaluation_order(list<c_sub_exp*>& UNUSEDARG(st), list<c_sub_exp*>::iterator UNUSEDARG(where))
{
    if(m_evaluation_order_emited)
        return true;

    m_evaluation_order_emited=true;

    return false;
}

void c_sub_exp_conditional_end_marker::compile(c_tree_microcode& UNUSEDARG(tm_callback))
{
    m_var_id=VAR_SYM_INDEX_NONE;
}

void c_sub_exp_conditional_end_marker::emit_microcode(c_expression& expression, c_micro_symbol_variable_mapping& UNUSEDARG(symbols))
{
    instruct_id_t next=expression.get_next_id();

    vector<c_sub_exp_oper*>::iterator i1;
    for(i1=m_conditionals.begin(); i1!=m_conditionals.end(); ++i1) {
        {
            vector<c_micro_log_and*>& ma=(*i1)->get_all_conditionals_micro_and();
            vector<c_micro_log_and*>::iterator i2;
            for(i2=ma.begin(); i2!=ma.end(); ++i2) {
                if((*i2)->get_cond_eval_jump()!=INSTRUCT_ID_INVALID)
                    continue;
                (*i2)->set_cond_eval_jump(next);
            }
        }

        {
            vector<c_micro_log_or*>& mo=(*i1)->get_all_conditionals_micro_or();
            vector<c_micro_log_or*>::iterator i2;
            for(i2=mo.begin(); i2!=mo.end(); ++i2) {
                if((*i2)->get_cond_eval_jump()!=INSTRUCT_ID_INVALID)
                    continue;
                (*i2)->set_cond_eval_jump(next);
            }
        }
    }
}

void c_sub_exp_conditional_end_marker::add_conditional(c_sub_exp_oper* el)
{
    m_conditionals.push_back(el);
}

int c_sub_exp_conditional_end_marker::get_prec()
{
    MY_ASSERT(!m_conditionals.empty());

    return m_conditionals[0]->get_prec();
}

bool c_sub_exp_conditional_end_marker::complete(tree<c_sub_exp*>& UNUSEDARG(t))
{
    return m_complete;
}

void c_sub_exp_conditional_end_marker::set_complete()
{
    m_complete=true;
}

bool c_sub_exp_conditional_end_marker::all_complete(c_tree_groupper& tg, tree<c_sub_exp*>& t)
{
    vector<c_sub_exp_oper*>::iterator i1;
    for(i1=m_conditionals.begin(); i1!=m_conditionals.end(); ++i1) {
        tree<c_sub_exp*>::iterator i2;
        for(i2=t.begin(); i2!=t.end(); ++i2)
            if(*i1==*i2)
                break;
        if(!tg.m_elim_tree.is_valid(i2))
            continue;
        if(!tg.completed(EVAL_RIGHT,i2,true))
            return false;
    }
    return true;
}

void c_sub_exp_conditional_end_marker::set_return_to(c_sub_exp* el)
{
    m_return_to=el;
}

c_sub_exp* c_sub_exp_conditional_end_marker::get_return_to()
{
    return m_return_to;
}

bool c_sub_exp_oper::get_return_complete()
{
    return m_return_complete;
}

void c_sub_exp_oper::set_return_complete()
{
    m_return_complete=true;
}

void c_sub_exp_oper_ternary_false::set_result(c_sub_exp* result)
{
    m_result=result;
}

c_sub_exp* c_sub_exp_oper_ternary_false::get_result()
{
    return m_result;
}

void c_sub_exp_oper_ternary_false::set_ternary(c_sub_exp_oper_ternary* ternary)
{
    m_ternary=ternary;
}

c_sub_exp_oper_ternary* c_sub_exp_oper_ternary_false::get_ternary()
{
    return m_ternary;
}

c_micro_ternary* c_sub_exp_oper_ternary::get_micro()
{
    return m_micro;
}

int c_sub_exp_ternary_true_end_marker::get_prec()
{
    return 18; // "?"
}

bool c_sub_exp_ternary_true_end_marker::emit_evaluation_order(list<c_sub_exp*>& UNUSEDARG(st), list<c_sub_exp*>::iterator UNUSEDARG(where))
{
    if(m_evaluation_order_emited)
        return true;

    m_evaluation_order_emited=true;

    return false;
}

void c_sub_exp_ternary_true_end_marker::compile(c_tree_microcode& UNUSEDARG(tm_callback))
{
    // NOOP
}

void c_sub_exp_ternary_true_end_marker::emit_microcode(c_expression& expression, c_micro_symbol_variable_mapping& UNUSEDARG(symbols))
{
    c_micro_jump* m=MY_NEW c_micro_jump();
    m_ternary_true->set_true_jump(m);
    expression.add_micro(m);

    m_ternary_true->get_micro()->set_false_jump(expression.get_next_id());
}

bool c_sub_exp_ternary_true_end_marker::complete(tree<c_sub_exp*>& UNUSEDARG(t))
{
    return true;
}

void c_sub_exp_ternary_true_end_marker::set_true(c_sub_exp_oper_ternary* ternary_true)
{
    m_ternary_true=ternary_true;
}

void c_sub_exp_ternary_true_end_marker::set_jump(c_sub_exp_ternary_false_end_marker* marker)
{
    m_ternary_jump=marker;
}

int c_sub_exp_ternary_false_end_marker::get_prec()
{
    return 18; // "?"
}

bool c_sub_exp_ternary_false_end_marker::emit_evaluation_order(list<c_sub_exp*>& UNUSEDARG(st), list<c_sub_exp*>::iterator UNUSEDARG(where))
{
    if(m_evaluation_order_emited)
        return true;

    m_evaluation_order_emited=true;

    return false;
}

void c_sub_exp_ternary_false_end_marker::compile(c_tree_microcode& UNUSEDARG(tm_callback))
{
    // NOOP
}

void c_sub_exp_ternary_false_end_marker::emit_microcode(c_expression& expression, c_micro_symbol_variable_mapping& UNUSEDARG(symbols))
{
    m_ternary_true->get_true_jump()->set_jump_id(expression.get_next_id());
}

bool c_sub_exp_ternary_false_end_marker::complete(tree<c_sub_exp*>& UNUSEDARG(t))
{
    return true;
}

void c_sub_exp_ternary_false_end_marker::set_true(c_sub_exp_oper_ternary* ternary_true)
{
    m_ternary_true=ternary_true;
}

void c_sub_exp_oper_ternary::set_true_jump(c_micro_jump* micro)
{
    m_true_jump=micro;
}

c_micro_jump* c_sub_exp_oper_ternary::get_true_jump()
{
    return m_true_jump;
}

int c_sub_exp_oper_ternary_true::get_prec()
{
    return 18; // "?:"
}

bool c_sub_exp_oper_ternary_true::emit_evaluation_order(list<c_sub_exp*>& UNUSEDARG(st), list<c_sub_exp*>::iterator UNUSEDARG(where))
{
    if(m_evaluation_order_emited)
        return true;

    m_evaluation_order_emited=true;

    return false;
}

void c_sub_exp_oper_ternary_true::compile(c_tree_microcode& UNUSEDARG(tm_callback))
{
    m_var_id=m_ternary->get_var_id();
}

void c_sub_exp_oper_ternary_true::emit_microcode(c_expression& expression, c_micro_symbol_variable_mapping& UNUSEDARG(symbols))
{
    MY_ASSERT(m_eval.size()>0);
    MY_ASSERT(VAR_SYM_INDEX_INVALID!=get_result()->get_var_id());

    expression.add_micro(MY_NEW c_micro_assign(this->get_result()->get_var_id(),this->get_var_id()));
}

void c_sub_exp_oper_ternary::set_result(c_sub_exp* result)
{
    m_result=result;
}

c_sub_exp* c_sub_exp_oper_ternary::get_result()
{
    return m_result;
}

void c_sub_exp_oper_ternary_true::set_result(c_sub_exp* result)
{
    m_result=result;
}

c_sub_exp* c_sub_exp_oper_ternary_true::get_result()
{
    return m_result;
}

bool c_sub_exp_oper_ternary_true::complete(tree<c_sub_exp*>& t)
{
    tree<c_sub_exp*>::iterator i1;
    for(i1=t.begin(); i1!=t.end(); ++i1)
        if(this==*i1)
            if(t.number_of_children(i1)!=0)
                return false;
    return m_eval.size()!=0;
}

void c_sub_exp_oper_ternary_true::add_sub(c_sub_exp* sub)
{
    vector<c_sub_exp*>::iterator i1;
    for(i1=m_eval.begin(); i1!=m_eval.end(); ++i1)
        if(*i1==sub)
            return;
    m_eval.push_back(sub);
    return;
}

void c_sub_exp_oper_ternary_true::set_ternary(c_sub_exp_oper_ternary* ternary)
{
    m_ternary=ternary;
}

c_sub_exp_oper_ternary* c_sub_exp_oper_ternary_true::get_ternary()
{
    return m_ternary;
}

void c_sub_exp_oper_ternary::set_other_ternary(c_sub_exp_oper_ternary* other_ternary)
{
    m_other_ternary=other_ternary;
}

const vector<c_sub_exp*>& c_sub_exp_oper::get_eval()
{
    return m_eval;
}

string c_sub_exp_func_arg_separator::emit_human_desc(string::size_type indent)
{
    static const string m_desc="ARGUMENT_SEPARATOR ";
    ostringstream o;
    string::size_type ii;
    for(ii=0; ii<indent; ++ii)
        o << " ";
    o << m_desc;
    o << "\n";
    return o.str();
}

int c_sub_exp_func_arg_separator::get_prec()
{
    return 24; // ","
}

string c_sub_exp_func_arg_separator::emit_human_microcode(string::size_type indent)
{
    static const string m_desc="ARGUMENT_SEPARATOR ";
    ostringstream o;
    string::size_type ii;
    for(ii=0; ii<indent; ++ii)
        o << " ";
    o << m_desc.c_str() << m_var_id;
    o << "\n";
    return o.str();
}

//

#define RETURN_SANITY_OKAY()  {         \
        static const c_exp_sanity rc(c_exp_sanity::CF_OKAY,NULL,__LINE__);      \
        return &rc;     \
    }

#define INTERNAL_ERROR_OOPS_BUG()  {    \
        static const c_exp_sanity rc(c_exp_sanity::CF_OOPS_BUG,NULL,__LINE__);  \
        return &rc;     \
    }

#define RETURN_EMIT_WARNING(msg__)  {   \
        static const c_exp_sanity rc(c_exp_sanity::CF_WARN,msg__,__LINE__);     \
        return &rc;     \
    }

#define RETURN_USER_ERROR(msg__)  {     \
        static const c_exp_sanity rc(c_exp_sanity::CF_USER_FATAL_ERROR,msg__,__LINE__);     \
        return &rc;     \
    }

const c_exp_sanity* c_sub_exp_root::can_precede(c_sub_exp_root* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_root::can_precede(c_sub_exp_root_end_dummy* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_root::can_precede(c_sub_exp_parenthesis* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_root::can_precede(c_sub_exp_index* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_root::can_precede(c_sub_exp_variable* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_root::can_precede(c_sub_exp_numeric_immediate* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_root::can_precede(c_sub_exp_const* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_root::can_precede(c_sub_exp_func* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_root::can_precede(c_sub_exp_eval_str* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_root::can_precede(c_sub_exp_immediate_str* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_root::can_precede(c_sub_exp_oper* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_root::can_precede(c_sub_exp_oper_ternary* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_root::can_precede(c_sub_exp_oper_ternary_true* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_root::can_precede(c_sub_exp_oper_ternary_false* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_root::can_precede(c_sub_exp_contitional_node_helper* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_root::can_precede(c_sub_exp_end_parenthesis* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_root::can_precede(c_sub_exp_end_index* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_root::can_precede(c_sub_exp_func_end_args* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_root::can_precede(c_sub_exp_func_arg_separator* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_root::can_precede(c_sub_exp_conditional_end_marker* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_root::can_precede(c_sub_exp_dummy* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_root::can_precede(c_sub_exp_ternary_true_end_marker* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_root::can_precede(c_sub_exp_ternary_false_end_marker* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

//

const c_exp_sanity* c_sub_exp_root_end_dummy::can_precede(c_sub_exp_root* UNUSEDARG(left))
{
    RETURN_EMIT_WARNING("An empty expression");
}

const c_exp_sanity* c_sub_exp_root_end_dummy::can_precede(c_sub_exp_root_end_dummy* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_root_end_dummy::can_precede(c_sub_exp_parenthesis* UNUSEDARG(left))
{
    RETURN_EMIT_WARNING("Unmatched parenthesis");
}

const c_exp_sanity* c_sub_exp_root_end_dummy::can_precede(c_sub_exp_index* UNUSEDARG(left))
{
    RETURN_USER_ERROR("Unmatched index closure");
}

const c_exp_sanity* c_sub_exp_root_end_dummy::can_precede(c_sub_exp_variable* UNUSEDARG(left))
{
    RETURN_SANITY_OKAY();
}

const c_exp_sanity* c_sub_exp_root_end_dummy::can_precede(c_sub_exp_numeric_immediate* UNUSEDARG(left))
{
    RETURN_SANITY_OKAY();
}

const c_exp_sanity* c_sub_exp_root_end_dummy::can_precede(c_sub_exp_const* UNUSEDARG(left))
{
    RETURN_SANITY_OKAY();
}

const c_exp_sanity* c_sub_exp_root_end_dummy::can_precede(c_sub_exp_func* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_root_end_dummy::can_precede(c_sub_exp_eval_str* UNUSEDARG(left))
{
    RETURN_SANITY_OKAY();
}

const c_exp_sanity* c_sub_exp_root_end_dummy::can_precede(c_sub_exp_immediate_str* UNUSEDARG(left))
{
    RETURN_SANITY_OKAY();
}

const c_exp_sanity* c_sub_exp_root_end_dummy::can_precede(c_sub_exp_oper* left)
{
    if(left->get_oper().get_prec()==1) // "/i"
        RETURN_SANITY_OKAY();

    if(left->get_oper().get_type()==OPER_TERNARY_FALSE)
        INTERNAL_ERROR_OOPS_BUG();

    if(left->get_oper().get_type()==OPER_TERNARY)
        INTERNAL_ERROR_OOPS_BUG();

    if(left->get_oper().get_type()==OPER_BINARY_OR_UNARY)
        RETURN_USER_ERROR("Trailing operator at the end of an expression");

    if(left->get_oper().get_type()==OPER_UNARY) {
        if(left->get_oper().get_prec()==25) // "return"
            RETURN_SANITY_OKAY();
        if(left->get_oper().get_prec()==4) // "++" or "--"
            RETURN_SANITY_OKAY();
        
        RETURN_USER_ERROR("Trailing operator at the end of an expression");
    }

    if(left->get_oper().get_type()==OPER_BINARY) {
        RETURN_USER_ERROR("Trailing operator at the end of an expression");
    }

    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_root_end_dummy::can_precede(c_sub_exp_oper_ternary* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_root_end_dummy::can_precede(c_sub_exp_oper_ternary_true* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_root_end_dummy::can_precede(c_sub_exp_oper_ternary_false* UNUSEDARG(left))
{
    RETURN_USER_ERROR("Trailing operator at the end of an expression");
}

const c_exp_sanity* c_sub_exp_root_end_dummy::can_precede(c_sub_exp_contitional_node_helper* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_root_end_dummy::can_precede(c_sub_exp_end_parenthesis* UNUSEDARG(left))
{
    RETURN_SANITY_OKAY();
}

const c_exp_sanity* c_sub_exp_root_end_dummy::can_precede(c_sub_exp_end_index* UNUSEDARG(left))
{
    RETURN_SANITY_OKAY();
}

const c_exp_sanity* c_sub_exp_root_end_dummy::can_precede(c_sub_exp_func_end_args* UNUSEDARG(left))
{
    RETURN_SANITY_OKAY();
}

const c_exp_sanity* c_sub_exp_root_end_dummy::can_precede(c_sub_exp_func_arg_separator* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_root_end_dummy::can_precede(c_sub_exp_conditional_end_marker* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_root_end_dummy::can_precede(c_sub_exp_dummy* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_root_end_dummy::can_precede(c_sub_exp_ternary_true_end_marker* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_root_end_dummy::can_precede(c_sub_exp_ternary_false_end_marker* UNUSEDARG(left))
{
    RETURN_SANITY_OKAY();
}

//

const c_exp_sanity* c_sub_exp_parenthesis::can_precede(c_sub_exp_root* UNUSEDARG(left))
{
    RETURN_SANITY_OKAY();
}

const c_exp_sanity* c_sub_exp_parenthesis::can_precede(c_sub_exp_root_end_dummy* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_parenthesis::can_precede(c_sub_exp_parenthesis* UNUSEDARG(left))
{
    RETURN_SANITY_OKAY();
}

const c_exp_sanity* c_sub_exp_parenthesis::can_precede(c_sub_exp_index* UNUSEDARG(left))
{
    RETURN_SANITY_OKAY();
}

const c_exp_sanity* c_sub_exp_parenthesis::can_precede(c_sub_exp_variable* UNUSEDARG(left))
{
    RETURN_USER_ERROR("Missing operator");
}

const c_exp_sanity* c_sub_exp_parenthesis::can_precede(c_sub_exp_numeric_immediate* UNUSEDARG(left))
{
    RETURN_USER_ERROR("Missing operator");
}

const c_exp_sanity* c_sub_exp_parenthesis::can_precede(c_sub_exp_const* UNUSEDARG(left))
{
    RETURN_USER_ERROR("Missing operator");
}

const c_exp_sanity* c_sub_exp_parenthesis::can_precede(c_sub_exp_func* UNUSEDARG(left))
{
    RETURN_SANITY_OKAY();
}

const c_exp_sanity* c_sub_exp_parenthesis::can_precede(c_sub_exp_eval_str* UNUSEDARG(left))
{
    RETURN_USER_ERROR("Missing operator");
}

const c_exp_sanity* c_sub_exp_parenthesis::can_precede(c_sub_exp_immediate_str* UNUSEDARG(left))
{
    RETURN_USER_ERROR("Missing operator");
}

const c_exp_sanity* c_sub_exp_parenthesis::can_precede(c_sub_exp_oper* left)
{
    if(left->get_oper().get_prec()==1) // "/i"
        RETURN_USER_ERROR("Invalid usage of case-insensitive index subscripting operator");

    if(left->get_oper().get_type()==OPER_BINARY)
        RETURN_SANITY_OKAY();

    if(left->get_oper().get_type()==OPER_TERNARY)
        RETURN_SANITY_OKAY();

    if(left->get_oper().get_type()==OPER_TERNARY_FALSE)
        RETURN_SANITY_OKAY();

    RETURN_SANITY_OKAY();
}

const c_exp_sanity* c_sub_exp_parenthesis::can_precede(c_sub_exp_oper_ternary* UNUSEDARG(left))
{
    RETURN_SANITY_OKAY();
}

const c_exp_sanity* c_sub_exp_parenthesis::can_precede(c_sub_exp_oper_ternary_true* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_parenthesis::can_precede(c_sub_exp_oper_ternary_false* UNUSEDARG(left))
{
    RETURN_SANITY_OKAY();
}

const c_exp_sanity* c_sub_exp_parenthesis::can_precede(c_sub_exp_contitional_node_helper* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_parenthesis::can_precede(c_sub_exp_end_parenthesis* UNUSEDARG(left))
{
    RETURN_USER_ERROR("An empty parenthesis");
}

const c_exp_sanity* c_sub_exp_parenthesis::can_precede(c_sub_exp_end_index* UNUSEDARG(left))
{
    RETURN_SANITY_OKAY();
}

const c_exp_sanity* c_sub_exp_parenthesis::can_precede(c_sub_exp_func_end_args* UNUSEDARG(left))
{
    RETURN_USER_ERROR("Missing operator");
}

const c_exp_sanity* c_sub_exp_parenthesis::can_precede(c_sub_exp_func_arg_separator* UNUSEDARG(left))
{
    RETURN_SANITY_OKAY();
}

const c_exp_sanity* c_sub_exp_parenthesis::can_precede(c_sub_exp_conditional_end_marker* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_parenthesis::can_precede(c_sub_exp_dummy* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_parenthesis::can_precede(c_sub_exp_ternary_true_end_marker* UNUSEDARG(left))
{
    RETURN_SANITY_OKAY();
}

const c_exp_sanity* c_sub_exp_parenthesis::can_precede(c_sub_exp_ternary_false_end_marker* UNUSEDARG(left))
{
    RETURN_USER_ERROR("Missing operator");
}

//

const c_exp_sanity* c_sub_exp_index::can_precede(c_sub_exp_root* UNUSEDARG(left))
{
    RETURN_USER_ERROR("Missing variable for subscript indexing");
}

const c_exp_sanity* c_sub_exp_index::can_precede(c_sub_exp_root_end_dummy* UNUSEDARG(left))
{
    RETURN_USER_ERROR("Trailing index subscript operator");
}

const c_exp_sanity* c_sub_exp_index::can_precede(c_sub_exp_parenthesis* UNUSEDARG(left))
{
    RETURN_USER_ERROR("Missing variable for subscript indexing");
}

const c_exp_sanity* c_sub_exp_index::can_precede(c_sub_exp_index* UNUSEDARG(left))
{
    RETURN_USER_ERROR("Missing variable for subscript indexing");
}

const c_exp_sanity* c_sub_exp_index::can_precede(c_sub_exp_variable* UNUSEDARG(left))
{
    RETURN_SANITY_OKAY();
}

const c_exp_sanity* c_sub_exp_index::can_precede(c_sub_exp_numeric_immediate* UNUSEDARG(left))
{
    RETURN_SANITY_OKAY();
}

const c_exp_sanity* c_sub_exp_index::can_precede(c_sub_exp_const* UNUSEDARG(left))
{
    RETURN_EMIT_WARNING("Subscript of a constant");
}

const c_exp_sanity* c_sub_exp_index::can_precede(c_sub_exp_func* UNUSEDARG(left))
{
    RETURN_SANITY_OKAY();
}

const c_exp_sanity* c_sub_exp_index::can_precede(c_sub_exp_eval_str* UNUSEDARG(left))
{
    RETURN_SANITY_OKAY();
}

const c_exp_sanity* c_sub_exp_index::can_precede(c_sub_exp_immediate_str* UNUSEDARG(left))
{
    RETURN_SANITY_OKAY();
}

const c_exp_sanity* c_sub_exp_index::can_precede(c_sub_exp_oper* UNUSEDARG(left))
{
    RETURN_USER_ERROR("Missing variable for subscript indexing");
}

const c_exp_sanity* c_sub_exp_index::can_precede(c_sub_exp_oper_ternary* UNUSEDARG(left))
{
    RETURN_USER_ERROR("Missing variable for subscript indexing");
}

const c_exp_sanity* c_sub_exp_index::can_precede(c_sub_exp_oper_ternary_true* UNUSEDARG(left))
{
    RETURN_USER_ERROR("Missing variable for subscript indexing");
}

const c_exp_sanity* c_sub_exp_index::can_precede(c_sub_exp_oper_ternary_false* UNUSEDARG(left))
{
    RETURN_SANITY_OKAY();
}

const c_exp_sanity* c_sub_exp_index::can_precede(c_sub_exp_contitional_node_helper* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_index::can_precede(c_sub_exp_end_parenthesis* UNUSEDARG(left))
{
    RETURN_SANITY_OKAY();
}

const c_exp_sanity* c_sub_exp_index::can_precede(c_sub_exp_end_index* UNUSEDARG(left))
{
    RETURN_SANITY_OKAY();
}

const c_exp_sanity* c_sub_exp_index::can_precede(c_sub_exp_func_end_args* UNUSEDARG(left))
{
    RETURN_SANITY_OKAY();
}

const c_exp_sanity* c_sub_exp_index::can_precede(c_sub_exp_func_arg_separator* UNUSEDARG(left))
{
    RETURN_USER_ERROR("Missing variable for subscript indexing");
}

const c_exp_sanity* c_sub_exp_index::can_precede(c_sub_exp_conditional_end_marker* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_index::can_precede(c_sub_exp_dummy* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_index::can_precede(c_sub_exp_ternary_true_end_marker* UNUSEDARG(left))
{
    RETURN_SANITY_OKAY();
}

const c_exp_sanity* c_sub_exp_index::can_precede(c_sub_exp_ternary_false_end_marker* UNUSEDARG(left))
{
    RETURN_SANITY_OKAY();
}

//

const c_exp_sanity* c_sub_exp_variable::can_precede(c_sub_exp_root* UNUSEDARG(left))
{
    RETURN_SANITY_OKAY();
}

const c_exp_sanity* c_sub_exp_variable::can_precede(c_sub_exp_root_end_dummy* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_variable::can_precede(c_sub_exp_parenthesis* UNUSEDARG(left))
{
    RETURN_SANITY_OKAY();
}

const c_exp_sanity* c_sub_exp_variable::can_precede(c_sub_exp_index* UNUSEDARG(left))
{
    RETURN_SANITY_OKAY();
}

const c_exp_sanity* c_sub_exp_variable::can_precede(c_sub_exp_variable* UNUSEDARG(left))
{
    RETURN_USER_ERROR("Missing operator");
}

const c_exp_sanity* c_sub_exp_variable::can_precede(c_sub_exp_numeric_immediate* UNUSEDARG(left))
{
    RETURN_USER_ERROR("Missing operator");
}

const c_exp_sanity* c_sub_exp_variable::can_precede(c_sub_exp_const* UNUSEDARG(left))
{
    RETURN_USER_ERROR("Missing operator");
}

const c_exp_sanity* c_sub_exp_variable::can_precede(c_sub_exp_func* UNUSEDARG(left))
{
    RETURN_SANITY_OKAY();
}

const c_exp_sanity* c_sub_exp_variable::can_precede(c_sub_exp_eval_str* UNUSEDARG(left))
{
    RETURN_USER_ERROR("Missing operator");
}

const c_exp_sanity* c_sub_exp_variable::can_precede(c_sub_exp_immediate_str* UNUSEDARG(left))
{
    RETURN_USER_ERROR("Missing operator");
}

const c_exp_sanity* c_sub_exp_variable::can_precede(c_sub_exp_oper* left)
{
    if(left->get_oper().get_prec()==1) // "/i"
        RETURN_USER_ERROR("Invalid usage of case-insensitive index subscripting operator");

    if(left->get_oper().get_type()==OPER_UNARY)
        RETURN_SANITY_OKAY();

    if(left->get_oper().get_type()==OPER_BINARY)
        RETURN_SANITY_OKAY();

    if(left->get_oper().get_type()==OPER_BINARY_OR_UNARY)
        RETURN_SANITY_OKAY();

    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_variable::can_precede(c_sub_exp_oper_ternary* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_variable::can_precede(c_sub_exp_oper_ternary_true* UNUSEDARG(left))
{
    RETURN_SANITY_OKAY();
}

const c_exp_sanity* c_sub_exp_variable::can_precede(c_sub_exp_oper_ternary_false* UNUSEDARG(left))
{
    RETURN_SANITY_OKAY();
}

const c_exp_sanity* c_sub_exp_variable::can_precede(c_sub_exp_contitional_node_helper* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_variable::can_precede(c_sub_exp_end_parenthesis* UNUSEDARG(left))
{
    RETURN_USER_ERROR("Missing operator");
}

const c_exp_sanity* c_sub_exp_variable::can_precede(c_sub_exp_end_index* UNUSEDARG(left))
{
    RETURN_USER_ERROR("Missing operator");
}

const c_exp_sanity* c_sub_exp_variable::can_precede(c_sub_exp_func_end_args* UNUSEDARG(left))
{
    RETURN_USER_ERROR("Missing operator");
}

const c_exp_sanity* c_sub_exp_variable::can_precede(c_sub_exp_func_arg_separator* UNUSEDARG(left))
{
    RETURN_SANITY_OKAY();
}

const c_exp_sanity* c_sub_exp_variable::can_precede(c_sub_exp_conditional_end_marker* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_variable::can_precede(c_sub_exp_dummy* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_variable::can_precede(c_sub_exp_ternary_true_end_marker* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_variable::can_precede(c_sub_exp_ternary_false_end_marker* UNUSEDARG(left))
{
    RETURN_USER_ERROR("Missing operator");
}

//

const c_exp_sanity* c_sub_exp_numeric_immediate::can_precede(c_sub_exp_root* UNUSEDARG(left))
{
    RETURN_SANITY_OKAY();
}

const c_exp_sanity* c_sub_exp_numeric_immediate::can_precede(c_sub_exp_root_end_dummy* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_numeric_immediate::can_precede(c_sub_exp_parenthesis* UNUSEDARG(left))
{
    RETURN_SANITY_OKAY();
}

const c_exp_sanity* c_sub_exp_numeric_immediate::can_precede(c_sub_exp_index* UNUSEDARG(left))
{
    RETURN_SANITY_OKAY();
}

const c_exp_sanity* c_sub_exp_numeric_immediate::can_precede(c_sub_exp_variable* UNUSEDARG(left))
{
    RETURN_USER_ERROR("Missing operator");
}

const c_exp_sanity* c_sub_exp_numeric_immediate::can_precede(c_sub_exp_numeric_immediate* UNUSEDARG(left))
{
    RETURN_USER_ERROR("Missing operator");
}

const c_exp_sanity* c_sub_exp_numeric_immediate::can_precede(c_sub_exp_const* UNUSEDARG(left))
{
    RETURN_USER_ERROR("Missing operator");
}

const c_exp_sanity* c_sub_exp_numeric_immediate::can_precede(c_sub_exp_func* UNUSEDARG(left))
{
    RETURN_SANITY_OKAY();
}

const c_exp_sanity* c_sub_exp_numeric_immediate::can_precede(c_sub_exp_eval_str* UNUSEDARG(left))
{
    RETURN_USER_ERROR("Missing operator");
}

const c_exp_sanity* c_sub_exp_numeric_immediate::can_precede(c_sub_exp_immediate_str* UNUSEDARG(left))
{
    RETURN_USER_ERROR("Missing operator");
}

const c_exp_sanity* c_sub_exp_numeric_immediate::can_precede(c_sub_exp_oper* left)
{
    if(left->get_oper().get_prec()==1) // "/i"
        RETURN_USER_ERROR("Invalid usage of case-insensitive index subscripting operator");

    if(left->get_oper().get_type()==OPER_UNARY) {
        if(left->get_oper().get_prec()==25) // "return"
            RETURN_SANITY_OKAY();
        if(left->get_oper().get_prec()==4) // "++" or "--"
            RETURN_USER_ERROR("'++' or '--' can be used only on an L-value");
        
        RETURN_EMIT_WARNING("Unary operator on numeric immediate used");
    }

    if(left->get_oper().get_type()==OPER_BINARY)
        RETURN_SANITY_OKAY();

    if(left->get_oper().get_type()==OPER_BINARY_OR_UNARY)
        RETURN_SANITY_OKAY();

    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_numeric_immediate::can_precede(c_sub_exp_oper_ternary* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_numeric_immediate::can_precede(c_sub_exp_oper_ternary_true* UNUSEDARG(left))
{
    RETURN_SANITY_OKAY();
}

const c_exp_sanity* c_sub_exp_numeric_immediate::can_precede(c_sub_exp_oper_ternary_false* UNUSEDARG(left))
{
    RETURN_SANITY_OKAY();
}

const c_exp_sanity* c_sub_exp_numeric_immediate::can_precede(c_sub_exp_contitional_node_helper* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_numeric_immediate::can_precede(c_sub_exp_end_parenthesis* UNUSEDARG(left))
{
    RETURN_USER_ERROR("Missing operator");
}

const c_exp_sanity* c_sub_exp_numeric_immediate::can_precede(c_sub_exp_end_index* UNUSEDARG(left))
{
    RETURN_SANITY_OKAY();
}

const c_exp_sanity* c_sub_exp_numeric_immediate::can_precede(c_sub_exp_func_end_args* UNUSEDARG(left))
{
    RETURN_USER_ERROR("Missing operator");
}

const c_exp_sanity* c_sub_exp_numeric_immediate::can_precede(c_sub_exp_func_arg_separator* UNUSEDARG(left))
{
    RETURN_SANITY_OKAY();
}

const c_exp_sanity* c_sub_exp_numeric_immediate::can_precede(c_sub_exp_conditional_end_marker* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_numeric_immediate::can_precede(c_sub_exp_dummy* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_numeric_immediate::can_precede(c_sub_exp_ternary_true_end_marker* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_numeric_immediate::can_precede(c_sub_exp_ternary_false_end_marker* UNUSEDARG(left))
{
    RETURN_USER_ERROR("Missing operator");
}

//

const c_exp_sanity* c_sub_exp_const::can_precede(c_sub_exp_root* UNUSEDARG(left))
{
    RETURN_SANITY_OKAY();
}

const c_exp_sanity* c_sub_exp_const::can_precede(c_sub_exp_root_end_dummy* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_const::can_precede(c_sub_exp_parenthesis* UNUSEDARG(left))
{
    RETURN_SANITY_OKAY();
}

const c_exp_sanity* c_sub_exp_const::can_precede(c_sub_exp_index* UNUSEDARG(left))
{
    RETURN_USER_ERROR("Missing operator");
}

const c_exp_sanity* c_sub_exp_const::can_precede(c_sub_exp_variable* UNUSEDARG(left))
{
    RETURN_USER_ERROR("Missing operator");
}

const c_exp_sanity* c_sub_exp_const::can_precede(c_sub_exp_numeric_immediate* UNUSEDARG(left))
{
    RETURN_USER_ERROR("Missing operator");
}

const c_exp_sanity* c_sub_exp_const::can_precede(c_sub_exp_const* UNUSEDARG(left))
{
    RETURN_USER_ERROR("Missing operator");
}

const c_exp_sanity* c_sub_exp_const::can_precede(c_sub_exp_func* UNUSEDARG(left))
{
    RETURN_SANITY_OKAY();
}

const c_exp_sanity* c_sub_exp_const::can_precede(c_sub_exp_eval_str* UNUSEDARG(left))
{
    RETURN_USER_ERROR("Missing operator");
}

const c_exp_sanity* c_sub_exp_const::can_precede(c_sub_exp_immediate_str* UNUSEDARG(left))
{
    RETURN_USER_ERROR("Missing operator");
}

const c_exp_sanity* c_sub_exp_const::can_precede(c_sub_exp_oper* left)
{
    if(left->get_oper().get_prec()==1) // "/i"
        RETURN_USER_ERROR("Invalid usage of case-insensitive index subscripting operator");

    if(left->get_oper().get_type()==OPER_UNARY) {
        if(left->get_oper().get_prec()==25) // "return"
            RETURN_SANITY_OKAY();
        if(left->get_oper().get_prec()==4) // "++" or "--"
            RETURN_USER_ERROR("'++' or '--' can be used only on an L-value");
        
        RETURN_EMIT_WARNING("An empty expression");
    }

    if(left->get_oper().get_type()==OPER_BINARY)
        RETURN_SANITY_OKAY();

    if(left->get_oper().get_type()==OPER_BINARY_OR_UNARY)
        RETURN_SANITY_OKAY();

    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_const::can_precede(c_sub_exp_oper_ternary* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_const::can_precede(c_sub_exp_oper_ternary_true* UNUSEDARG(left))
{
    RETURN_SANITY_OKAY();
}

const c_exp_sanity* c_sub_exp_const::can_precede(c_sub_exp_oper_ternary_false* UNUSEDARG(left))
{
    RETURN_SANITY_OKAY();
}

const c_exp_sanity* c_sub_exp_const::can_precede(c_sub_exp_contitional_node_helper* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_const::can_precede(c_sub_exp_end_parenthesis* UNUSEDARG(left))
{
    RETURN_USER_ERROR("Missing operator");
}

const c_exp_sanity* c_sub_exp_const::can_precede(c_sub_exp_end_index* UNUSEDARG(left))
{
    RETURN_USER_ERROR("Missing operator");
}

const c_exp_sanity* c_sub_exp_const::can_precede(c_sub_exp_func_end_args* UNUSEDARG(left))
{
    RETURN_USER_ERROR("Missing operator");
}

const c_exp_sanity* c_sub_exp_const::can_precede(c_sub_exp_func_arg_separator* UNUSEDARG(left))
{
    RETURN_SANITY_OKAY();
}

const c_exp_sanity* c_sub_exp_const::can_precede(c_sub_exp_conditional_end_marker* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_const::can_precede(c_sub_exp_dummy* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_const::can_precede(c_sub_exp_ternary_true_end_marker* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_const::can_precede(c_sub_exp_ternary_false_end_marker* UNUSEDARG(left))
{
    RETURN_USER_ERROR("Missing operator");
}

//

const c_exp_sanity* c_sub_exp_func::can_precede(c_sub_exp_root* UNUSEDARG(left))
{
    RETURN_SANITY_OKAY();
}

const c_exp_sanity* c_sub_exp_func::can_precede(c_sub_exp_root_end_dummy* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_func::can_precede(c_sub_exp_parenthesis* UNUSEDARG(left))
{
    RETURN_SANITY_OKAY();
}

const c_exp_sanity* c_sub_exp_func::can_precede(c_sub_exp_index* UNUSEDARG(left))
{
    RETURN_SANITY_OKAY();
}

const c_exp_sanity* c_sub_exp_func::can_precede(c_sub_exp_variable* UNUSEDARG(left))
{
    RETURN_USER_ERROR("Missing operator");
}

const c_exp_sanity* c_sub_exp_func::can_precede(c_sub_exp_numeric_immediate* UNUSEDARG(left))
{
    RETURN_USER_ERROR("Missing operator");
}

const c_exp_sanity* c_sub_exp_func::can_precede(c_sub_exp_const* UNUSEDARG(left))
{
    RETURN_USER_ERROR("Missing operator");
}

const c_exp_sanity* c_sub_exp_func::can_precede(c_sub_exp_func* UNUSEDARG(left))
{
    RETURN_SANITY_OKAY();
}

const c_exp_sanity* c_sub_exp_func::can_precede(c_sub_exp_eval_str* UNUSEDARG(left))
{
    RETURN_USER_ERROR("Missing operator");
}

const c_exp_sanity* c_sub_exp_func::can_precede(c_sub_exp_immediate_str* UNUSEDARG(left))
{
    RETURN_USER_ERROR("Missing operator");
}

const c_exp_sanity* c_sub_exp_func::can_precede(c_sub_exp_oper* left)
{
    if(left->get_oper().get_prec()==1) // "/i"
        RETURN_USER_ERROR("Invalid usage of case-insensitive index subscripting operator");

    if(left->get_oper().get_type()==OPER_TERNARY_FALSE)
        RETURN_SANITY_OKAY();

    if(left->get_oper().get_type()==OPER_TERNARY)
        RETURN_SANITY_OKAY();

    if(left->get_oper().get_type()==OPER_BINARY_OR_UNARY)
        RETURN_SANITY_OKAY();

    if(left->get_oper().get_type()==OPER_UNARY) {
        if(left->get_oper().get_prec()==25) // "return"
            RETURN_SANITY_OKAY();
        if(left->get_oper().get_prec()==4) // "++" or "--"
            RETURN_USER_ERROR("'++' or '--' can be used only on an L-value");
        RETURN_SANITY_OKAY();
    }

    if(left->get_oper().get_type()==OPER_BINARY)
        RETURN_SANITY_OKAY();

    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_func::can_precede(c_sub_exp_oper_ternary* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_func::can_precede(c_sub_exp_oper_ternary_true* UNUSEDARG(left))
{
    RETURN_SANITY_OKAY();
}

const c_exp_sanity* c_sub_exp_func::can_precede(c_sub_exp_oper_ternary_false* UNUSEDARG(left))
{
    RETURN_SANITY_OKAY();
}

const c_exp_sanity* c_sub_exp_func::can_precede(c_sub_exp_contitional_node_helper* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_func::can_precede(c_sub_exp_end_parenthesis* UNUSEDARG(left))
{
    RETURN_USER_ERROR("Missing operator");
}

const c_exp_sanity* c_sub_exp_func::can_precede(c_sub_exp_end_index* UNUSEDARG(left))
{
    RETURN_SANITY_OKAY();
}

const c_exp_sanity* c_sub_exp_func::can_precede(c_sub_exp_func_end_args* UNUSEDARG(left))
{
    RETURN_USER_ERROR("Missing operator");
}

const c_exp_sanity* c_sub_exp_func::can_precede(c_sub_exp_func_arg_separator* UNUSEDARG(left))
{
    RETURN_SANITY_OKAY();
}

const c_exp_sanity* c_sub_exp_func::can_precede(c_sub_exp_conditional_end_marker* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_func::can_precede(c_sub_exp_dummy* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_func::can_precede(c_sub_exp_ternary_true_end_marker* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_func::can_precede(c_sub_exp_ternary_false_end_marker* UNUSEDARG(left))
{
    RETURN_USER_ERROR("Missing operator");
}

//

const c_exp_sanity* c_sub_exp_eval_str::can_precede(c_sub_exp_root* UNUSEDARG(left))
{
    RETURN_SANITY_OKAY();
}

const c_exp_sanity* c_sub_exp_eval_str::can_precede(c_sub_exp_root_end_dummy* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_eval_str::can_precede(c_sub_exp_parenthesis* UNUSEDARG(left))
{
    RETURN_SANITY_OKAY();
}

const c_exp_sanity* c_sub_exp_eval_str::can_precede(c_sub_exp_index* UNUSEDARG(left))
{
    RETURN_SANITY_OKAY();
}

const c_exp_sanity* c_sub_exp_eval_str::can_precede(c_sub_exp_variable* UNUSEDARG(left))
{
    RETURN_USER_ERROR("Missing operator");
}

const c_exp_sanity* c_sub_exp_eval_str::can_precede(c_sub_exp_numeric_immediate* UNUSEDARG(left))
{
    RETURN_USER_ERROR("Missing operator");
}

const c_exp_sanity* c_sub_exp_eval_str::can_precede(c_sub_exp_const* UNUSEDARG(left))
{
    RETURN_USER_ERROR("Missing operator");
}

const c_exp_sanity* c_sub_exp_eval_str::can_precede(c_sub_exp_func* UNUSEDARG(left))
{
    RETURN_SANITY_OKAY();
}

const c_exp_sanity* c_sub_exp_eval_str::can_precede(c_sub_exp_eval_str* UNUSEDARG(left))
{
    RETURN_USER_ERROR("Missing operator");
}

const c_exp_sanity* c_sub_exp_eval_str::can_precede(c_sub_exp_immediate_str* UNUSEDARG(left))
{
    RETURN_USER_ERROR("Missing operator");
}

const c_exp_sanity* c_sub_exp_eval_str::can_precede(c_sub_exp_oper* left)
{
    if(left->get_oper().get_prec()==1) // "/i"
        RETURN_USER_ERROR("Invalid usage of case-insensitive index subscripting operator");

    if(left->get_oper().get_type()==OPER_UNARY) {
        if(left->get_oper().get_prec()==25) // "return"
            RETURN_SANITY_OKAY();
        if(left->get_oper().get_prec()==4) // "++" or "--"
            RETURN_USER_ERROR("'++' or '--' can be used only on an L-value");
        
        RETURN_EMIT_WARNING("Unary operator on evaluation string immediate used");
    }

    if(left->get_oper().get_type()==OPER_BINARY)
        RETURN_SANITY_OKAY();

    if(left->get_oper().get_type()==OPER_BINARY_OR_UNARY)
        RETURN_SANITY_OKAY();

    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_eval_str::can_precede(c_sub_exp_oper_ternary* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_eval_str::can_precede(c_sub_exp_oper_ternary_true* UNUSEDARG(left))
{
    RETURN_SANITY_OKAY();
}

const c_exp_sanity* c_sub_exp_eval_str::can_precede(c_sub_exp_oper_ternary_false* UNUSEDARG(left))
{
    RETURN_SANITY_OKAY();
}

const c_exp_sanity* c_sub_exp_eval_str::can_precede(c_sub_exp_contitional_node_helper* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_eval_str::can_precede(c_sub_exp_end_parenthesis* UNUSEDARG(left))
{
    RETURN_USER_ERROR("Missing operator");
}

const c_exp_sanity* c_sub_exp_eval_str::can_precede(c_sub_exp_end_index* UNUSEDARG(left))
{
    RETURN_USER_ERROR("Missing operator");
}

const c_exp_sanity* c_sub_exp_eval_str::can_precede(c_sub_exp_func_end_args* UNUSEDARG(left))
{
    RETURN_USER_ERROR("Missing operator");
}

const c_exp_sanity* c_sub_exp_eval_str::can_precede(c_sub_exp_func_arg_separator* UNUSEDARG(left))
{
    RETURN_SANITY_OKAY();
}

const c_exp_sanity* c_sub_exp_eval_str::can_precede(c_sub_exp_conditional_end_marker* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_eval_str::can_precede(c_sub_exp_dummy* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_eval_str::can_precede(c_sub_exp_ternary_true_end_marker* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_eval_str::can_precede(c_sub_exp_ternary_false_end_marker* UNUSEDARG(left))
{
    RETURN_USER_ERROR("Missing operator");
}

//

const c_exp_sanity* c_sub_exp_immediate_str::can_precede(c_sub_exp_root* UNUSEDARG(left))
{
    RETURN_SANITY_OKAY();
}

const c_exp_sanity* c_sub_exp_immediate_str::can_precede(c_sub_exp_root_end_dummy* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_immediate_str::can_precede(c_sub_exp_parenthesis* UNUSEDARG(left))
{
    RETURN_USER_ERROR("Missing operator");
}

const c_exp_sanity* c_sub_exp_immediate_str::can_precede(c_sub_exp_index* UNUSEDARG(left))
{
    RETURN_SANITY_OKAY();
}

const c_exp_sanity* c_sub_exp_immediate_str::can_precede(c_sub_exp_variable* UNUSEDARG(left))
{
    RETURN_USER_ERROR("Missing operator");
}

const c_exp_sanity* c_sub_exp_immediate_str::can_precede(c_sub_exp_numeric_immediate* UNUSEDARG(left))
{
    RETURN_USER_ERROR("Missing operator");
}

const c_exp_sanity* c_sub_exp_immediate_str::can_precede(c_sub_exp_const* UNUSEDARG(left))
{
    RETURN_USER_ERROR("Missing operator");
}

const c_exp_sanity* c_sub_exp_immediate_str::can_precede(c_sub_exp_func* UNUSEDARG(left))
{
    RETURN_SANITY_OKAY();
}

const c_exp_sanity* c_sub_exp_immediate_str::can_precede(c_sub_exp_eval_str* UNUSEDARG(left))
{
    RETURN_USER_ERROR("Missing operator");
}

const c_exp_sanity* c_sub_exp_immediate_str::can_precede(c_sub_exp_immediate_str* UNUSEDARG(left))
{
    RETURN_USER_ERROR("Missing operator");
}

const c_exp_sanity* c_sub_exp_immediate_str::can_precede(c_sub_exp_oper* left)
{
    if(left->get_oper().get_prec()==1) // "/i"
        RETURN_USER_ERROR("Invalid usage of case-insensitive index subscripting operator");

    if(left->get_oper().get_type()==OPER_UNARY) {
        if(left->get_oper().get_prec()==25) // "return"
            RETURN_SANITY_OKAY();
        if(left->get_oper().get_prec()==4) // "++" or "--"
            RETURN_USER_ERROR("'++' or '--' can be used only on an L-value");
        
        RETURN_EMIT_WARNING("Unary operator on constant string immediate used");
    }

    if(left->get_oper().get_type()==OPER_BINARY)
        RETURN_SANITY_OKAY();

    if(left->get_oper().get_type()==OPER_BINARY_OR_UNARY)
        RETURN_SANITY_OKAY();

    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_immediate_str::can_precede(c_sub_exp_oper_ternary* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_immediate_str::can_precede(c_sub_exp_oper_ternary_true* UNUSEDARG(left))
{
    RETURN_SANITY_OKAY();
}

const c_exp_sanity* c_sub_exp_immediate_str::can_precede(c_sub_exp_oper_ternary_false* UNUSEDARG(left))
{
    RETURN_SANITY_OKAY();
}

const c_exp_sanity* c_sub_exp_immediate_str::can_precede(c_sub_exp_contitional_node_helper* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_immediate_str::can_precede(c_sub_exp_end_parenthesis* UNUSEDARG(left))
{
    RETURN_USER_ERROR("Missing operator");
}

const c_exp_sanity* c_sub_exp_immediate_str::can_precede(c_sub_exp_end_index* UNUSEDARG(left))
{
    RETURN_SANITY_OKAY();
}

const c_exp_sanity* c_sub_exp_immediate_str::can_precede(c_sub_exp_func_end_args* UNUSEDARG(left))
{
    RETURN_USER_ERROR("Missing operator");
}

const c_exp_sanity* c_sub_exp_immediate_str::can_precede(c_sub_exp_func_arg_separator* UNUSEDARG(left))
{
    RETURN_SANITY_OKAY();
}

const c_exp_sanity* c_sub_exp_immediate_str::can_precede(c_sub_exp_conditional_end_marker* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_immediate_str::can_precede(c_sub_exp_dummy* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_immediate_str::can_precede(c_sub_exp_ternary_true_end_marker* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_immediate_str::can_precede(c_sub_exp_ternary_false_end_marker* UNUSEDARG(left))
{
    RETURN_USER_ERROR("Missing operator");
}

//

const c_exp_sanity* c_sub_exp_oper::can_precede(c_sub_exp_root* UNUSEDARG(left))
{
    if(this->get_oper().get_prec()==1) // "/i"
        RETURN_USER_ERROR("Invalid usage of case-insensitive index subscripting operator");

    if(this->get_oper().get_type()==OPER_UNARY)
        RETURN_SANITY_OKAY();

    if(this->get_oper().get_type()==OPER_BINARY)
        RETURN_USER_ERROR("Invalid usage of binary operator");

    if(this->get_oper().get_type()==OPER_BINARY_OR_UNARY)
        RETURN_SANITY_OKAY();

    if(this->get_oper().get_type()==OPER_TERNARY)
        INTERNAL_ERROR_OOPS_BUG();

    if(this->get_oper().get_type()==OPER_TERNARY_FALSE)
        INTERNAL_ERROR_OOPS_BUG();

    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_oper::can_precede(c_sub_exp_root_end_dummy* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_oper::can_precede(c_sub_exp_parenthesis* UNUSEDARG(left))
{
    if(this->get_oper().get_prec()==1) // "/i"
        RETURN_USER_ERROR("Invalid usage of case-insensitive index subscripting operator");

    if(this->get_oper().get_type()==OPER_UNARY)
        RETURN_SANITY_OKAY();

    if(this->get_oper().get_type()==OPER_BINARY)
        RETURN_USER_ERROR("Invalid usage of binary operator");

    if(this->get_oper().get_type()==OPER_BINARY_OR_UNARY)
        RETURN_SANITY_OKAY();

    if(this->get_oper().get_type()==OPER_TERNARY)
        RETURN_USER_ERROR("Invalid usage of ternary conditional operator");

    if(this->get_oper().get_type()==OPER_TERNARY_FALSE)
        INTERNAL_ERROR_OOPS_BUG();

    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_oper::can_precede(c_sub_exp_index* UNUSEDARG(left))
{
    if(this->get_oper().get_prec()==1) // "/i"
        RETURN_USER_ERROR("Invalid usage of case-insensitive index subscripting operator");

    if(this->get_oper().get_type()==OPER_UNARY)
        RETURN_SANITY_OKAY();

    if(this->get_oper().get_type()==OPER_BINARY)
        RETURN_USER_ERROR("Invalid usage of binary operator");

    if(this->get_oper().get_type()==OPER_BINARY_OR_UNARY)
        RETURN_SANITY_OKAY();

    if(this->get_oper().get_type()==OPER_TERNARY)
        RETURN_USER_ERROR("Invalid usage of ternary conditional operator");

    if(this->get_oper().get_type()==OPER_TERNARY_FALSE)
        INTERNAL_ERROR_OOPS_BUG();

    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_oper::can_precede(c_sub_exp_variable* UNUSEDARG(left))
{
    if(this->get_oper().get_prec()==1) // "/i"
        RETURN_SANITY_OKAY();

    if(this->get_oper().get_type()==OPER_UNARY) {
        if(this->get_oper().get_prec()==4) // "++" or "--"
            RETURN_SANITY_OKAY();
        
        RETURN_USER_ERROR("Invalid usage of unary operator");
    }

    if(this->get_oper().get_type()==OPER_BINARY)
        RETURN_SANITY_OKAY();

    if(this->get_oper().get_type()==OPER_BINARY_OR_UNARY)
        RETURN_SANITY_OKAY();

    if(this->get_oper().get_type()==OPER_TERNARY)
        RETURN_SANITY_OKAY();

    if(this->get_oper().get_type()==OPER_TERNARY_FALSE)
        INTERNAL_ERROR_OOPS_BUG();

    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_oper::can_precede(c_sub_exp_numeric_immediate* UNUSEDARG(left))
{
    if(this->get_oper().get_prec()==1) // "/i"
        RETURN_EMIT_WARNING("Case-insensitive indexing operator on numeric immediate used");

    if(this->get_oper().get_type()==OPER_UNARY)
        RETURN_USER_ERROR("Invalid usage of unary operator");

    if(this->get_oper().get_type()==OPER_BINARY)
        RETURN_SANITY_OKAY();

    if(this->get_oper().get_type()==OPER_BINARY_OR_UNARY)
        RETURN_SANITY_OKAY();

    if(this->get_oper().get_type()==OPER_TERNARY) {
        RETURN_EMIT_WARNING("Numeric immediate constant used as a condition for ternary operator");
    }

    if(this->get_oper().get_type()==OPER_TERNARY_FALSE)
        INTERNAL_ERROR_OOPS_BUG();

    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_oper::can_precede(c_sub_exp_const* UNUSEDARG(left))
{
    if(this->get_oper().get_prec()==1) // "/i"
        RETURN_EMIT_WARNING("Case-insensitive indexing operator on built-in constant used");

    if(this->get_oper().get_type()==OPER_UNARY)
        RETURN_USER_ERROR("Invalid usage of unary operator");

    if(this->get_oper().get_type()==OPER_BINARY)
        RETURN_SANITY_OKAY();

    if(this->get_oper().get_type()==OPER_BINARY_OR_UNARY)
        RETURN_SANITY_OKAY();

    if(this->get_oper().get_type()==OPER_TERNARY) {
        RETURN_EMIT_WARNING("Built-in constant used as a condition for ternary operator");
    }

    if(this->get_oper().get_type()==OPER_TERNARY_FALSE)
        INTERNAL_ERROR_OOPS_BUG();

    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_oper::can_precede(c_sub_exp_func* UNUSEDARG(left))
{
    if(this->get_oper().get_prec()==1) // "/i"
        RETURN_USER_ERROR("Invalid usage of case-insensitive index subscripting operator");

    if(this->get_oper().get_type()==OPER_UNARY)
        RETURN_SANITY_OKAY();

    if(this->get_oper().get_type()==OPER_BINARY)
        RETURN_USER_ERROR("Invalid usage of binary operator");

    if(this->get_oper().get_type()==OPER_BINARY_OR_UNARY)
        RETURN_SANITY_OKAY();

    if(this->get_oper().get_type()==OPER_TERNARY)
        RETURN_USER_ERROR("Invalid usage of ternary conditional operator");

    if(this->get_oper().get_type()==OPER_TERNARY_FALSE)
        INTERNAL_ERROR_OOPS_BUG();

    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_oper::can_precede(c_sub_exp_eval_str* UNUSEDARG(left))
{
    if(this->get_oper().get_prec()==1) // "/i"
        RETURN_SANITY_OKAY();

    if(this->get_oper().get_type()==OPER_UNARY)
        RETURN_USER_ERROR("Invalid usage of unary operator");

    if(this->get_oper().get_type()==OPER_BINARY)
        RETURN_SANITY_OKAY();

    if(this->get_oper().get_type()==OPER_BINARY_OR_UNARY)
        RETURN_SANITY_OKAY();

    if(this->get_oper().get_type()==OPER_TERNARY) {
        RETURN_EMIT_WARNING("Evaluation string (non-numeric) used as a condition for ternary operator");
    }

    if(this->get_oper().get_type()==OPER_TERNARY_FALSE)
        RETURN_SANITY_OKAY();

    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_oper::can_precede(c_sub_exp_immediate_str* UNUSEDARG(left))
{
    if(this->get_oper().get_prec()==1) // "/i"
        RETURN_SANITY_OKAY();

    if(this->get_oper().get_type()==OPER_UNARY)
        RETURN_USER_ERROR("Invalid usage of unary operator");

    if(this->get_oper().get_type()==OPER_BINARY)
        RETURN_SANITY_OKAY();

    if(this->get_oper().get_type()==OPER_BINARY_OR_UNARY)
        RETURN_SANITY_OKAY();

    if(this->get_oper().get_type()==OPER_TERNARY) {
        RETURN_EMIT_WARNING("Immediate string constant used as a condition for ternary operator");
    }

    if(this->get_oper().get_type()==OPER_TERNARY_FALSE)
        RETURN_SANITY_OKAY();

    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_oper::can_precede(c_sub_exp_oper* left)
{
    if(this->get_oper().get_prec()==1) // "/i"
        RETURN_SANITY_OKAY();

    if(this->get_oper().get_type()==OPER_BINARY_OR_UNARY)
        RETURN_SANITY_OKAY();

    if(left->get_oper().get_type()==OPER_BINARY_OR_UNARY)
        RETURN_SANITY_OKAY();

    if(this->get_oper().get_type()==OPER_UNARY &&
        left->get_oper().get_type()==OPER_UNARY &&
        left->get_oper().get_prec()==25 // "return"
        )
            RETURN_SANITY_OKAY();

    if(this->get_oper().get_type()==left->get_oper().get_type())
        RETURN_USER_ERROR("Invalid combination of two operators");

    if(this->get_oper().get_type()==OPER_UNARY && left->get_oper().get_type()==OPER_BINARY)
        RETURN_SANITY_OKAY();

    if(this->get_oper().get_type()!=OPER_UNARY &&
        left->get_oper().get_type()==OPER_UNARY &&
        left->get_oper().get_prec()==4 // "++" or "--"
        )
            RETURN_SANITY_OKAY();

    RETURN_USER_ERROR("Invalid combination of two operators");
}

const c_exp_sanity* c_sub_exp_oper::can_precede(c_sub_exp_oper_ternary* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_oper::can_precede(c_sub_exp_oper_ternary_true* UNUSEDARG(left))
{
    if(this->get_oper().get_prec()==1) // "/i"
        RETURN_USER_ERROR("Invalid usage of case-insensitive index subscripting operator");

    if(this->get_oper().get_type()==OPER_UNARY)
        RETURN_SANITY_OKAY();

    if(this->get_oper().get_type()==OPER_BINARY)
        RETURN_USER_ERROR("Invalid usage of binary operator");

    if(this->get_oper().get_type()==OPER_BINARY_OR_UNARY)
        RETURN_SANITY_OKAY();

    if(this->get_oper().get_type()==OPER_TERNARY)
        RETURN_USER_ERROR("Invalid usage of ternary conditional operator");

    if(this->get_oper().get_type()==OPER_TERNARY_FALSE)
        RETURN_USER_ERROR("Invalid usage of ternary conditional operator");

    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_oper::can_precede(c_sub_exp_oper_ternary_false* UNUSEDARG(left))
{
    if(this->get_oper().get_prec()==1) // "/i"
        RETURN_USER_ERROR("Invalid usage of case-insensitive index subscripting operator");

    if(this->get_oper().get_type()==OPER_UNARY)
        RETURN_SANITY_OKAY();

    if(this->get_oper().get_type()==OPER_BINARY)
        RETURN_USER_ERROR("Invalid usage of binary operator");

    if(this->get_oper().get_type()==OPER_BINARY_OR_UNARY)
        RETURN_SANITY_OKAY();

    if(this->get_oper().get_type()==OPER_TERNARY)
        RETURN_USER_ERROR("Invalid usage of ternary conditional operator");

    if(this->get_oper().get_type()==OPER_TERNARY_FALSE)
        RETURN_USER_ERROR("Invalid usage of ternary conditional operator");

    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_oper::can_precede(c_sub_exp_contitional_node_helper* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_oper::can_precede(c_sub_exp_end_parenthesis* UNUSEDARG(left))
{
    if(this->get_oper().get_prec()==1) // "/i"
        RETURN_USER_ERROR("Invalid usage of case-insensitive index subscripting operator");

    if(this->get_oper().get_type()==OPER_UNARY)
        RETURN_USER_ERROR("Invalid usage of unary operator");

    if(this->get_oper().get_type()==OPER_BINARY)
        RETURN_SANITY_OKAY();

    if(this->get_oper().get_type()==OPER_BINARY_OR_UNARY)
        RETURN_SANITY_OKAY();

    if(this->get_oper().get_type()==OPER_TERNARY)
        RETURN_SANITY_OKAY();

    if(this->get_oper().get_type()==OPER_TERNARY_FALSE)
        RETURN_USER_ERROR("Invalid usage of ternary conditional operator");

    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_oper::can_precede(c_sub_exp_end_index* UNUSEDARG(left))
{
    if(this->get_oper().get_prec()==1) // "/i"
        RETURN_SANITY_OKAY();

    if(this->get_oper().get_type()==OPER_UNARY)
        RETURN_SANITY_OKAY();

    if(this->get_oper().get_type()==OPER_BINARY)
        RETURN_SANITY_OKAY();

    if(this->get_oper().get_type()==OPER_BINARY_OR_UNARY)
        RETURN_SANITY_OKAY();

    if(this->get_oper().get_type()==OPER_TERNARY)
        RETURN_SANITY_OKAY();

    if(this->get_oper().get_type()==OPER_TERNARY_FALSE)
        RETURN_USER_ERROR("Invalid usage of ternary conditional operator");

    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_oper::can_precede(c_sub_exp_func_end_args* UNUSEDARG(left))
{
    if(this->get_oper().get_prec()==1) // "/i"
        RETURN_SANITY_OKAY();

    if(this->get_oper().get_type()==OPER_UNARY)
        RETURN_USER_ERROR("Invalid usage of unary operator");

    if(this->get_oper().get_type()==OPER_BINARY)
        RETURN_SANITY_OKAY();

    if(this->get_oper().get_type()==OPER_BINARY_OR_UNARY)
        RETURN_SANITY_OKAY();

    if(this->get_oper().get_type()==OPER_TERNARY)
        RETURN_SANITY_OKAY();

    if(this->get_oper().get_type()==OPER_TERNARY_FALSE)
        RETURN_USER_ERROR("Invalid usage of ternary conditional operator");

    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_oper::can_precede(c_sub_exp_func_arg_separator* UNUSEDARG(left))
{
    if(this->get_oper().get_prec()==1) // "/i"
        RETURN_USER_ERROR("Invalid usage of case-insensitive index subscripting operator");

    if(this->get_oper().get_type()==OPER_UNARY)
        RETURN_SANITY_OKAY();

    if(this->get_oper().get_type()==OPER_BINARY)
        RETURN_USER_ERROR("Invalid usage of binary operator");

    if(this->get_oper().get_type()==OPER_BINARY_OR_UNARY)
        RETURN_SANITY_OKAY();

    if(this->get_oper().get_type()==OPER_TERNARY)
        RETURN_USER_ERROR("Invalid usage of ternary conditional operator");

    if(this->get_oper().get_type()==OPER_TERNARY_FALSE)
        RETURN_USER_ERROR("Invalid usage of ternary conditional operator");

    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_oper::can_precede(c_sub_exp_conditional_end_marker* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_oper::can_precede(c_sub_exp_dummy* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_oper::can_precede(c_sub_exp_ternary_true_end_marker* UNUSEDARG(left))
{
    if(this->get_oper().get_prec()==1) // "/i"
        RETURN_USER_ERROR("Invalid usage of case-insensitive index subscripting operator");

    if(this->get_oper().get_type()==OPER_UNARY)
        RETURN_USER_ERROR("Invalid usage of unary operator");

    if(this->get_oper().get_type()==OPER_BINARY)
        RETURN_USER_ERROR("Invalid usage of binary operator");

    if(this->get_oper().get_type()==OPER_BINARY_OR_UNARY)
        RETURN_USER_ERROR("Invalid usage of binary or unary operator");

    if(this->get_oper().get_type()==OPER_TERNARY)
        RETURN_USER_ERROR("Invalid usage of ternary conditional operator");

    if(this->get_oper().get_type()==OPER_TERNARY_FALSE)
        RETURN_USER_ERROR("Invalid usage of ternary conditional operator");

    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_oper::can_precede(c_sub_exp_ternary_false_end_marker* UNUSEDARG(left))
{
    if(this->get_oper().get_prec()==1) // "/i"
        RETURN_SANITY_OKAY();

    if(this->get_oper().get_type()==OPER_UNARY)
        RETURN_USER_ERROR("Invalid usage of unary operator");

    if(this->get_oper().get_type()==OPER_BINARY)
        RETURN_SANITY_OKAY();

    if(this->get_oper().get_type()==OPER_BINARY_OR_UNARY)
        RETURN_SANITY_OKAY();

    if(this->get_oper().get_type()==OPER_TERNARY)
        RETURN_SANITY_OKAY();

    if(this->get_oper().get_type()==OPER_TERNARY_FALSE)
        RETURN_USER_ERROR("Invalid usage of ternary conditional operator");

    INTERNAL_ERROR_OOPS_BUG();
}

//

const c_exp_sanity* c_sub_exp_oper_ternary::can_precede(c_sub_exp_root* UNUSEDARG(left))
{
    RETURN_USER_ERROR("Missing conditional expression for ternary operator");
}

const c_exp_sanity* c_sub_exp_oper_ternary::can_precede(c_sub_exp_root_end_dummy* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_oper_ternary::can_precede(c_sub_exp_parenthesis* UNUSEDARG(left))
{
    RETURN_USER_ERROR("Missing conditional expression for ternary operator");
}

const c_exp_sanity* c_sub_exp_oper_ternary::can_precede(c_sub_exp_index* UNUSEDARG(left))
{
    RETURN_USER_ERROR("Missing conditional expression for ternary operator");
}

const c_exp_sanity* c_sub_exp_oper_ternary::can_precede(c_sub_exp_variable* UNUSEDARG(left))
{
    RETURN_SANITY_OKAY();
}

const c_exp_sanity* c_sub_exp_oper_ternary::can_precede(c_sub_exp_numeric_immediate* UNUSEDARG(left))
{
    RETURN_EMIT_WARNING("Numeric immediate constant used as a condition for ternary operator");
}

const c_exp_sanity* c_sub_exp_oper_ternary::can_precede(c_sub_exp_const* UNUSEDARG(left))
{
    RETURN_EMIT_WARNING("Built-in constant used as a condition for ternary operator");
}

const c_exp_sanity* c_sub_exp_oper_ternary::can_precede(c_sub_exp_func* UNUSEDARG(left))
{
    RETURN_USER_ERROR("Missing conditional expression for ternary operator");
}

const c_exp_sanity* c_sub_exp_oper_ternary::can_precede(c_sub_exp_eval_str* UNUSEDARG(left))
{
    RETURN_SANITY_OKAY();
}

const c_exp_sanity* c_sub_exp_oper_ternary::can_precede(c_sub_exp_immediate_str* UNUSEDARG(left))
{
    RETURN_EMIT_WARNING("Immediate string constant used as a condition for ternary operator");
}

const c_exp_sanity* c_sub_exp_oper_ternary::can_precede(c_sub_exp_oper* left)
{
    if(left->get_oper().get_prec()==1) // "/i"
        RETURN_USER_ERROR("Invalid usage of case-insensitive index subscripting operator");

    if(left->get_oper().get_type()==OPER_UNARY) {
        if(left->get_oper().get_prec()==25) // "return"
            RETURN_SANITY_OKAY();
        if(left->get_oper().get_prec()==4) // "++" or "--"
            RETURN_SANITY_OKAY();

        RETURN_USER_ERROR("Invalid usage of unary operator");
    }

    if(left->get_oper().get_type()==OPER_BINARY)
        RETURN_USER_ERROR("Invalid usage of binary operator");

    if(left->get_oper().get_type()==OPER_BINARY_OR_UNARY)
        RETURN_SANITY_OKAY();

    if(left->get_oper().get_type()==OPER_TERNARY)
        RETURN_USER_ERROR("Invalid usage of ternary conditional operator");

    if(left->get_oper().get_type()==OPER_TERNARY_FALSE)
        RETURN_USER_ERROR("Invalid usage of ternary conditional operator");

    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_oper_ternary::can_precede(c_sub_exp_oper_ternary* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_oper_ternary::can_precede(c_sub_exp_oper_ternary_true* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_oper_ternary::can_precede(c_sub_exp_oper_ternary_false* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_oper_ternary::can_precede(c_sub_exp_contitional_node_helper* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_oper_ternary::can_precede(c_sub_exp_end_parenthesis* UNUSEDARG(left))
{
    RETURN_SANITY_OKAY();
}

const c_exp_sanity* c_sub_exp_oper_ternary::can_precede(c_sub_exp_end_index* UNUSEDARG(left))
{
    RETURN_SANITY_OKAY();
}

const c_exp_sanity* c_sub_exp_oper_ternary::can_precede(c_sub_exp_func_end_args* UNUSEDARG(left))
{
    RETURN_SANITY_OKAY();
}

const c_exp_sanity* c_sub_exp_oper_ternary::can_precede(c_sub_exp_func_arg_separator* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_oper_ternary::can_precede(c_sub_exp_conditional_end_marker* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_oper_ternary::can_precede(c_sub_exp_dummy* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_oper_ternary::can_precede(c_sub_exp_ternary_true_end_marker* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_oper_ternary::can_precede(c_sub_exp_ternary_false_end_marker* UNUSEDARG(left))
{
    RETURN_SANITY_OKAY();
}

//

const c_exp_sanity* c_sub_exp_oper_ternary_true::can_precede(c_sub_exp_root* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_oper_ternary_true::can_precede(c_sub_exp_root_end_dummy* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_oper_ternary_true::can_precede(c_sub_exp_parenthesis* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_oper_ternary_true::can_precede(c_sub_exp_index* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_oper_ternary_true::can_precede(c_sub_exp_variable* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_oper_ternary_true::can_precede(c_sub_exp_numeric_immediate* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_oper_ternary_true::can_precede(c_sub_exp_const* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_oper_ternary_true::can_precede(c_sub_exp_func* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_oper_ternary_true::can_precede(c_sub_exp_eval_str* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_oper_ternary_true::can_precede(c_sub_exp_immediate_str* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_oper_ternary_true::can_precede(c_sub_exp_oper* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_oper_ternary_true::can_precede(c_sub_exp_oper_ternary* UNUSEDARG(left))
{
    RETURN_SANITY_OKAY();
}

const c_exp_sanity* c_sub_exp_oper_ternary_true::can_precede(c_sub_exp_oper_ternary_true* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_oper_ternary_true::can_precede(c_sub_exp_oper_ternary_false* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_oper_ternary_true::can_precede(c_sub_exp_contitional_node_helper* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_oper_ternary_true::can_precede(c_sub_exp_end_parenthesis* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_oper_ternary_true::can_precede(c_sub_exp_end_index* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_oper_ternary_true::can_precede(c_sub_exp_func_end_args* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_oper_ternary_true::can_precede(c_sub_exp_func_arg_separator* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_oper_ternary_true::can_precede(c_sub_exp_conditional_end_marker* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_oper_ternary_true::can_precede(c_sub_exp_dummy* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_oper_ternary_true::can_precede(c_sub_exp_ternary_true_end_marker* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_oper_ternary_true::can_precede(c_sub_exp_ternary_false_end_marker* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

//

const c_exp_sanity* c_sub_exp_oper_ternary_false::can_precede(c_sub_exp_root* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_oper_ternary_false::can_precede(c_sub_exp_root_end_dummy* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_oper_ternary_false::can_precede(c_sub_exp_parenthesis* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_oper_ternary_false::can_precede(c_sub_exp_index* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_oper_ternary_false::can_precede(c_sub_exp_variable* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_oper_ternary_false::can_precede(c_sub_exp_numeric_immediate* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_oper_ternary_false::can_precede(c_sub_exp_const* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_oper_ternary_false::can_precede(c_sub_exp_func* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_oper_ternary_false::can_precede(c_sub_exp_eval_str* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_oper_ternary_false::can_precede(c_sub_exp_immediate_str* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_oper_ternary_false::can_precede(c_sub_exp_oper* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_oper_ternary_false::can_precede(c_sub_exp_oper_ternary* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_oper_ternary_false::can_precede(c_sub_exp_oper_ternary_true* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_oper_ternary_false::can_precede(c_sub_exp_oper_ternary_false* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_oper_ternary_false::can_precede(c_sub_exp_contitional_node_helper* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_oper_ternary_false::can_precede(c_sub_exp_end_parenthesis* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_oper_ternary_false::can_precede(c_sub_exp_end_index* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_oper_ternary_false::can_precede(c_sub_exp_func_end_args* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_oper_ternary_false::can_precede(c_sub_exp_func_arg_separator* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_oper_ternary_false::can_precede(c_sub_exp_conditional_end_marker* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_oper_ternary_false::can_precede(c_sub_exp_dummy* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_oper_ternary_false::can_precede(c_sub_exp_ternary_true_end_marker* UNUSEDARG(left))
{
    RETURN_SANITY_OKAY();
}

const c_exp_sanity* c_sub_exp_oper_ternary_false::can_precede(c_sub_exp_ternary_false_end_marker* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

//

const c_exp_sanity* c_sub_exp_contitional_node_helper::can_precede(c_sub_exp_root* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_contitional_node_helper::can_precede(c_sub_exp_root_end_dummy* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_contitional_node_helper::can_precede(c_sub_exp_parenthesis* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_contitional_node_helper::can_precede(c_sub_exp_index* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_contitional_node_helper::can_precede(c_sub_exp_variable* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_contitional_node_helper::can_precede(c_sub_exp_numeric_immediate* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_contitional_node_helper::can_precede(c_sub_exp_const* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_contitional_node_helper::can_precede(c_sub_exp_func* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_contitional_node_helper::can_precede(c_sub_exp_eval_str* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_contitional_node_helper::can_precede(c_sub_exp_immediate_str* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_contitional_node_helper::can_precede(c_sub_exp_oper* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_contitional_node_helper::can_precede(c_sub_exp_oper_ternary* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_contitional_node_helper::can_precede(c_sub_exp_oper_ternary_true* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_contitional_node_helper::can_precede(c_sub_exp_oper_ternary_false* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_contitional_node_helper::can_precede(c_sub_exp_contitional_node_helper* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_contitional_node_helper::can_precede(c_sub_exp_end_parenthesis* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_contitional_node_helper::can_precede(c_sub_exp_end_index* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_contitional_node_helper::can_precede(c_sub_exp_func_end_args* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_contitional_node_helper::can_precede(c_sub_exp_func_arg_separator* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_contitional_node_helper::can_precede(c_sub_exp_conditional_end_marker* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_contitional_node_helper::can_precede(c_sub_exp_dummy* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_contitional_node_helper::can_precede(c_sub_exp_ternary_true_end_marker* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_contitional_node_helper::can_precede(c_sub_exp_ternary_false_end_marker* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

//

const c_exp_sanity* c_sub_exp_end_parenthesis::can_precede(c_sub_exp_root* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_end_parenthesis::can_precede(c_sub_exp_root_end_dummy* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_end_parenthesis::can_precede(c_sub_exp_parenthesis* UNUSEDARG(left))
{
    RETURN_USER_ERROR("Missing operator");
}

const c_exp_sanity* c_sub_exp_end_parenthesis::can_precede(c_sub_exp_index* UNUSEDARG(left))
{
    RETURN_USER_ERROR("Unmatched index closure");
}

const c_exp_sanity* c_sub_exp_end_parenthesis::can_precede(c_sub_exp_variable* UNUSEDARG(left))
{
    RETURN_SANITY_OKAY();
}

const c_exp_sanity* c_sub_exp_end_parenthesis::can_precede(c_sub_exp_numeric_immediate* UNUSEDARG(left))
{
    RETURN_SANITY_OKAY();
}

const c_exp_sanity* c_sub_exp_end_parenthesis::can_precede(c_sub_exp_const* UNUSEDARG(left))
{
    RETURN_SANITY_OKAY();
}

const c_exp_sanity* c_sub_exp_end_parenthesis::can_precede(c_sub_exp_func* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_end_parenthesis::can_precede(c_sub_exp_eval_str* UNUSEDARG(left))
{
    RETURN_SANITY_OKAY();
}

const c_exp_sanity* c_sub_exp_end_parenthesis::can_precede(c_sub_exp_immediate_str* UNUSEDARG(left))
{
    RETURN_SANITY_OKAY();
}

const c_exp_sanity* c_sub_exp_end_parenthesis::can_precede(c_sub_exp_oper* left)
{
    if(left->get_oper().get_prec()==1) // "/i"
        RETURN_USER_ERROR("Invalid usage of case-insensitive index subscripting operator");

    if(left->get_oper().get_type()==OPER_UNARY)
        RETURN_SANITY_OKAY();

    if(left->get_oper().get_type()==OPER_BINARY)
        RETURN_USER_ERROR("Invalid usage of binary operator");

    if(left->get_oper().get_type()==OPER_BINARY_OR_UNARY)
        RETURN_SANITY_OKAY();

    if(left->get_oper().get_type()==OPER_TERNARY)
        INTERNAL_ERROR_OOPS_BUG();

    if(left->get_oper().get_type()==OPER_TERNARY_FALSE)
        INTERNAL_ERROR_OOPS_BUG();

    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_end_parenthesis::can_precede(c_sub_exp_oper_ternary* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_end_parenthesis::can_precede(c_sub_exp_oper_ternary_true* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_end_parenthesis::can_precede(c_sub_exp_oper_ternary_false* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_end_parenthesis::can_precede(c_sub_exp_contitional_node_helper* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_end_parenthesis::can_precede(c_sub_exp_end_parenthesis* UNUSEDARG(left))
{
    RETURN_SANITY_OKAY();
}

const c_exp_sanity* c_sub_exp_end_parenthesis::can_precede(c_sub_exp_end_index* UNUSEDARG(left))
{
    RETURN_SANITY_OKAY();
}

const c_exp_sanity* c_sub_exp_end_parenthesis::can_precede(c_sub_exp_func_end_args* UNUSEDARG(left))
{
    RETURN_SANITY_OKAY();
}

const c_exp_sanity* c_sub_exp_end_parenthesis::can_precede(c_sub_exp_func_arg_separator* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_end_parenthesis::can_precede(c_sub_exp_conditional_end_marker* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_end_parenthesis::can_precede(c_sub_exp_dummy* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_end_parenthesis::can_precede(c_sub_exp_ternary_true_end_marker* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_end_parenthesis::can_precede(c_sub_exp_ternary_false_end_marker* UNUSEDARG(left))
{
    RETURN_SANITY_OKAY();
}

//

const c_exp_sanity* c_sub_exp_end_index::can_precede(c_sub_exp_root* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_end_index::can_precede(c_sub_exp_root_end_dummy* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_end_index::can_precede(c_sub_exp_parenthesis* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_end_index::can_precede(c_sub_exp_index* UNUSEDARG(left))
{
    RETURN_SANITY_OKAY();
}

const c_exp_sanity* c_sub_exp_end_index::can_precede(c_sub_exp_variable* UNUSEDARG(left))
{
    RETURN_SANITY_OKAY();
}

const c_exp_sanity* c_sub_exp_end_index::can_precede(c_sub_exp_numeric_immediate* UNUSEDARG(left))
{
    RETURN_SANITY_OKAY();
}

const c_exp_sanity* c_sub_exp_end_index::can_precede(c_sub_exp_const* UNUSEDARG(left))
{
    RETURN_EMIT_WARNING("Built-in constant used as a subscript");
}

const c_exp_sanity* c_sub_exp_end_index::can_precede(c_sub_exp_func* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_end_index::can_precede(c_sub_exp_eval_str* UNUSEDARG(left))
{
    RETURN_SANITY_OKAY();
}

const c_exp_sanity* c_sub_exp_end_index::can_precede(c_sub_exp_immediate_str* UNUSEDARG(left))
{
    RETURN_SANITY_OKAY();
}

const c_exp_sanity* c_sub_exp_end_index::can_precede(c_sub_exp_oper* left)
{
    if(left->get_oper().get_prec()==1) // "/i"
        RETURN_SANITY_OKAY();

    if(left->get_oper().get_type()==OPER_UNARY)
        RETURN_SANITY_OKAY();

    if(left->get_oper().get_type()==OPER_BINARY)
        RETURN_USER_ERROR("Invalid usage of binary operator");

    if(left->get_oper().get_type()==OPER_BINARY_OR_UNARY)
        RETURN_SANITY_OKAY();

    if(left->get_oper().get_type()==OPER_TERNARY)
        INTERNAL_ERROR_OOPS_BUG();

    if(left->get_oper().get_type()==OPER_TERNARY_FALSE)
        INTERNAL_ERROR_OOPS_BUG();

    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_end_index::can_precede(c_sub_exp_oper_ternary* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_end_index::can_precede(c_sub_exp_oper_ternary_true* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_end_index::can_precede(c_sub_exp_oper_ternary_false* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_end_index::can_precede(c_sub_exp_contitional_node_helper* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_end_index::can_precede(c_sub_exp_end_parenthesis* UNUSEDARG(left))
{
    RETURN_SANITY_OKAY();
}

const c_exp_sanity* c_sub_exp_end_index::can_precede(c_sub_exp_end_index* UNUSEDARG(left))
{
    RETURN_SANITY_OKAY();
}

const c_exp_sanity* c_sub_exp_end_index::can_precede(c_sub_exp_func_end_args* UNUSEDARG(left))
{
    RETURN_SANITY_OKAY();
}

const c_exp_sanity* c_sub_exp_end_index::can_precede(c_sub_exp_func_arg_separator* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_end_index::can_precede(c_sub_exp_conditional_end_marker* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_end_index::can_precede(c_sub_exp_dummy* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_end_index::can_precede(c_sub_exp_ternary_true_end_marker* UNUSEDARG(left))
{
    RETURN_SANITY_OKAY();
}

const c_exp_sanity* c_sub_exp_end_index::can_precede(c_sub_exp_ternary_false_end_marker* UNUSEDARG(left))
{
    RETURN_SANITY_OKAY();
}

//

const c_exp_sanity* c_sub_exp_func_end_args::can_precede(c_sub_exp_root* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_func_end_args::can_precede(c_sub_exp_root_end_dummy* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_func_end_args::can_precede(c_sub_exp_parenthesis* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_func_end_args::can_precede(c_sub_exp_index* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_func_end_args::can_precede(c_sub_exp_variable* UNUSEDARG(left))
{
    RETURN_SANITY_OKAY();
}

const c_exp_sanity* c_sub_exp_func_end_args::can_precede(c_sub_exp_numeric_immediate* UNUSEDARG(left))
{
    RETURN_SANITY_OKAY();
}

const c_exp_sanity* c_sub_exp_func_end_args::can_precede(c_sub_exp_const* UNUSEDARG(left))
{
    RETURN_SANITY_OKAY();
}

const c_exp_sanity* c_sub_exp_func_end_args::can_precede(c_sub_exp_func* UNUSEDARG(left))
{
    RETURN_SANITY_OKAY();
}

const c_exp_sanity* c_sub_exp_func_end_args::can_precede(c_sub_exp_eval_str* UNUSEDARG(left))
{
    RETURN_SANITY_OKAY();
}

const c_exp_sanity* c_sub_exp_func_end_args::can_precede(c_sub_exp_immediate_str* UNUSEDARG(left))
{
    RETURN_SANITY_OKAY();
}

const c_exp_sanity* c_sub_exp_func_end_args::can_precede(c_sub_exp_oper* left)
{
    if(left->get_oper().get_prec()==1) // "/i"
        RETURN_USER_ERROR("Invalid usage of case-insensitive index subscripting operator");

    if(left->get_oper().get_type()==OPER_UNARY)
        RETURN_SANITY_OKAY();

    if(left->get_oper().get_type()==OPER_BINARY)
        RETURN_USER_ERROR("Invalid usage of binary operator");

    if(left->get_oper().get_type()==OPER_BINARY_OR_UNARY)
        RETURN_SANITY_OKAY();

    if(left->get_oper().get_type()==OPER_TERNARY)
        INTERNAL_ERROR_OOPS_BUG();

    if(left->get_oper().get_type()==OPER_TERNARY_FALSE)
        INTERNAL_ERROR_OOPS_BUG();

    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_func_end_args::can_precede(c_sub_exp_oper_ternary* UNUSEDARG(left))
{
    RETURN_USER_ERROR("Invalid usage of ternary operator");
}

const c_exp_sanity* c_sub_exp_func_end_args::can_precede(c_sub_exp_oper_ternary_true* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_func_end_args::can_precede(c_sub_exp_oper_ternary_false* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_func_end_args::can_precede(c_sub_exp_contitional_node_helper* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_func_end_args::can_precede(c_sub_exp_end_parenthesis* UNUSEDARG(left))
{
    RETURN_SANITY_OKAY();
}

const c_exp_sanity* c_sub_exp_func_end_args::can_precede(c_sub_exp_end_index* UNUSEDARG(left))
{
    RETURN_SANITY_OKAY();
}

const c_exp_sanity* c_sub_exp_func_end_args::can_precede(c_sub_exp_func_end_args* UNUSEDARG(left))
{
    RETURN_SANITY_OKAY();
}

const c_exp_sanity* c_sub_exp_func_end_args::can_precede(c_sub_exp_func_arg_separator* UNUSEDARG(left))
{
    RETURN_USER_ERROR("Trailing comma as function argument separator at end of function arguments");
}

const c_exp_sanity* c_sub_exp_func_end_args::can_precede(c_sub_exp_conditional_end_marker* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_func_end_args::can_precede(c_sub_exp_dummy* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_func_end_args::can_precede(c_sub_exp_ternary_true_end_marker* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_func_end_args::can_precede(c_sub_exp_ternary_false_end_marker* UNUSEDARG(left))
{
    RETURN_SANITY_OKAY();
}

//

const c_exp_sanity* c_sub_exp_func_arg_separator::can_precede(c_sub_exp_root* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_func_arg_separator::can_precede(c_sub_exp_root_end_dummy* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_func_arg_separator::can_precede(c_sub_exp_parenthesis* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_func_arg_separator::can_precede(c_sub_exp_index* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_func_arg_separator::can_precede(c_sub_exp_variable* UNUSEDARG(left))
{
    RETURN_SANITY_OKAY();
}

const c_exp_sanity* c_sub_exp_func_arg_separator::can_precede(c_sub_exp_numeric_immediate* UNUSEDARG(left))
{
    RETURN_SANITY_OKAY();
}

const c_exp_sanity* c_sub_exp_func_arg_separator::can_precede(c_sub_exp_const* UNUSEDARG(left))
{
    RETURN_SANITY_OKAY();
}

const c_exp_sanity* c_sub_exp_func_arg_separator::can_precede(c_sub_exp_func* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_func_arg_separator::can_precede(c_sub_exp_eval_str* UNUSEDARG(left))
{
    RETURN_SANITY_OKAY();
}

const c_exp_sanity* c_sub_exp_func_arg_separator::can_precede(c_sub_exp_immediate_str* UNUSEDARG(left))
{
    RETURN_SANITY_OKAY();
}

const c_exp_sanity* c_sub_exp_func_arg_separator::can_precede(c_sub_exp_oper* left)
{
    if(left->get_oper().get_prec()==1) // "/i"
        RETURN_USER_ERROR("Invalid usage of case-insensitive index subscripting operator");

    if(left->get_oper().get_type()==OPER_UNARY)
        RETURN_SANITY_OKAY();

    if(left->get_oper().get_type()==OPER_BINARY)
        RETURN_USER_ERROR("Invalid usage of binary operator");

    if(left->get_oper().get_type()==OPER_BINARY_OR_UNARY)
        RETURN_SANITY_OKAY();

    if(left->get_oper().get_type()==OPER_TERNARY)
        INTERNAL_ERROR_OOPS_BUG();

    if(left->get_oper().get_type()==OPER_TERNARY_FALSE)
        INTERNAL_ERROR_OOPS_BUG();

    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_func_arg_separator::can_precede(c_sub_exp_oper_ternary* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_func_arg_separator::can_precede(c_sub_exp_oper_ternary_true* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_func_arg_separator::can_precede(c_sub_exp_oper_ternary_false* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_func_arg_separator::can_precede(c_sub_exp_contitional_node_helper* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_func_arg_separator::can_precede(c_sub_exp_end_parenthesis* UNUSEDARG(left))
{
    RETURN_SANITY_OKAY();
}

const c_exp_sanity* c_sub_exp_func_arg_separator::can_precede(c_sub_exp_end_index* UNUSEDARG(left))
{
    RETURN_SANITY_OKAY();
}

const c_exp_sanity* c_sub_exp_func_arg_separator::can_precede(c_sub_exp_func_end_args* UNUSEDARG(left))
{
    RETURN_SANITY_OKAY();
}

const c_exp_sanity* c_sub_exp_func_arg_separator::can_precede(c_sub_exp_func_arg_separator* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_func_arg_separator::can_precede(c_sub_exp_conditional_end_marker* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_func_arg_separator::can_precede(c_sub_exp_dummy* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_func_arg_separator::can_precede(c_sub_exp_ternary_true_end_marker* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_func_arg_separator::can_precede(c_sub_exp_ternary_false_end_marker* UNUSEDARG(left))
{
    RETURN_SANITY_OKAY();
}

//

const c_exp_sanity* c_sub_exp_conditional_end_marker::can_precede(c_sub_exp_root* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_conditional_end_marker::can_precede(c_sub_exp_root_end_dummy* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_conditional_end_marker::can_precede(c_sub_exp_parenthesis* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_conditional_end_marker::can_precede(c_sub_exp_index* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_conditional_end_marker::can_precede(c_sub_exp_variable* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_conditional_end_marker::can_precede(c_sub_exp_numeric_immediate* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_conditional_end_marker::can_precede(c_sub_exp_const* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_conditional_end_marker::can_precede(c_sub_exp_func* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_conditional_end_marker::can_precede(c_sub_exp_eval_str* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_conditional_end_marker::can_precede(c_sub_exp_immediate_str* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_conditional_end_marker::can_precede(c_sub_exp_oper* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_conditional_end_marker::can_precede(c_sub_exp_oper_ternary* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_conditional_end_marker::can_precede(c_sub_exp_oper_ternary_true* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_conditional_end_marker::can_precede(c_sub_exp_oper_ternary_false* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_conditional_end_marker::can_precede(c_sub_exp_contitional_node_helper* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_conditional_end_marker::can_precede(c_sub_exp_end_parenthesis* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_conditional_end_marker::can_precede(c_sub_exp_end_index* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_conditional_end_marker::can_precede(c_sub_exp_func_end_args* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_conditional_end_marker::can_precede(c_sub_exp_func_arg_separator* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_conditional_end_marker::can_precede(c_sub_exp_conditional_end_marker* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_conditional_end_marker::can_precede(c_sub_exp_dummy* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_conditional_end_marker::can_precede(c_sub_exp_ternary_true_end_marker* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_conditional_end_marker::can_precede(c_sub_exp_ternary_false_end_marker* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

//

const c_exp_sanity* c_sub_exp_dummy::can_precede(c_sub_exp_root* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_dummy::can_precede(c_sub_exp_root_end_dummy* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_dummy::can_precede(c_sub_exp_parenthesis* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_dummy::can_precede(c_sub_exp_index* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_dummy::can_precede(c_sub_exp_variable* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_dummy::can_precede(c_sub_exp_numeric_immediate* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_dummy::can_precede(c_sub_exp_const* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_dummy::can_precede(c_sub_exp_func* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_dummy::can_precede(c_sub_exp_eval_str* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_dummy::can_precede(c_sub_exp_immediate_str* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_dummy::can_precede(c_sub_exp_oper* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_dummy::can_precede(c_sub_exp_oper_ternary* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_dummy::can_precede(c_sub_exp_oper_ternary_true* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_dummy::can_precede(c_sub_exp_oper_ternary_false* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_dummy::can_precede(c_sub_exp_contitional_node_helper* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_dummy::can_precede(c_sub_exp_end_parenthesis* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_dummy::can_precede(c_sub_exp_end_index* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_dummy::can_precede(c_sub_exp_func_end_args* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_dummy::can_precede(c_sub_exp_func_arg_separator* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_dummy::can_precede(c_sub_exp_conditional_end_marker* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_dummy::can_precede(c_sub_exp_dummy* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_dummy::can_precede(c_sub_exp_ternary_true_end_marker* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_dummy::can_precede(c_sub_exp_ternary_false_end_marker* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

//

const c_exp_sanity* c_sub_exp_ternary_true_end_marker::can_precede(c_sub_exp_root* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_ternary_true_end_marker::can_precede(c_sub_exp_root_end_dummy* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_ternary_true_end_marker::can_precede(c_sub_exp_parenthesis* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_ternary_true_end_marker::can_precede(c_sub_exp_index* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_ternary_true_end_marker::can_precede(c_sub_exp_variable* UNUSEDARG(left))
{
    RETURN_SANITY_OKAY();
}

const c_exp_sanity* c_sub_exp_ternary_true_end_marker::can_precede(c_sub_exp_numeric_immediate* UNUSEDARG(left))
{
    RETURN_SANITY_OKAY();
}

const c_exp_sanity* c_sub_exp_ternary_true_end_marker::can_precede(c_sub_exp_const* UNUSEDARG(left))
{
    RETURN_SANITY_OKAY();
}

const c_exp_sanity* c_sub_exp_ternary_true_end_marker::can_precede(c_sub_exp_func* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_ternary_true_end_marker::can_precede(c_sub_exp_eval_str* UNUSEDARG(left))
{
    RETURN_SANITY_OKAY();
}

const c_exp_sanity* c_sub_exp_ternary_true_end_marker::can_precede(c_sub_exp_immediate_str* UNUSEDARG(left))
{
    RETURN_SANITY_OKAY();
}

const c_exp_sanity* c_sub_exp_ternary_true_end_marker::can_precede(c_sub_exp_oper* left)
{
    if(left->get_oper().get_prec()==1) // "/i"
        RETURN_USER_ERROR("Invalid usage of case-insensitive index subscripting operator");

    if(left->get_oper().get_type()==OPER_UNARY)
        RETURN_SANITY_OKAY();

    if(left->get_oper().get_type()==OPER_BINARY)
        RETURN_USER_ERROR("Invalid usage of binary operator");

    if(left->get_oper().get_type()==OPER_BINARY_OR_UNARY)
        RETURN_SANITY_OKAY();

    if(left->get_oper().get_type()==OPER_TERNARY)
        INTERNAL_ERROR_OOPS_BUG();

    if(left->get_oper().get_type()==OPER_TERNARY_FALSE)
        INTERNAL_ERROR_OOPS_BUG();

    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_ternary_true_end_marker::can_precede(c_sub_exp_oper_ternary* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_ternary_true_end_marker::can_precede(c_sub_exp_oper_ternary_true* UNUSEDARG(left))
{
    RETURN_USER_ERROR("Invalid usage of ternary operator");
}

const c_exp_sanity* c_sub_exp_ternary_true_end_marker::can_precede(c_sub_exp_oper_ternary_false* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_ternary_true_end_marker::can_precede(c_sub_exp_contitional_node_helper* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_ternary_true_end_marker::can_precede(c_sub_exp_end_parenthesis* UNUSEDARG(left))
{
    RETURN_SANITY_OKAY();
}

const c_exp_sanity* c_sub_exp_ternary_true_end_marker::can_precede(c_sub_exp_end_index* UNUSEDARG(left))
{
    RETURN_SANITY_OKAY();
}

const c_exp_sanity* c_sub_exp_ternary_true_end_marker::can_precede(c_sub_exp_func_end_args* UNUSEDARG(left))
{
    RETURN_SANITY_OKAY();
}

const c_exp_sanity* c_sub_exp_ternary_true_end_marker::can_precede(c_sub_exp_func_arg_separator* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_ternary_true_end_marker::can_precede(c_sub_exp_conditional_end_marker* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_ternary_true_end_marker::can_precede(c_sub_exp_dummy* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_ternary_true_end_marker::can_precede(c_sub_exp_ternary_true_end_marker* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_ternary_true_end_marker::can_precede(c_sub_exp_ternary_false_end_marker* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

//

const c_exp_sanity* c_sub_exp_ternary_false_end_marker::can_precede(c_sub_exp_root* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_ternary_false_end_marker::can_precede(c_sub_exp_root_end_dummy* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_ternary_false_end_marker::can_precede(c_sub_exp_parenthesis* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_ternary_false_end_marker::can_precede(c_sub_exp_index* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_ternary_false_end_marker::can_precede(c_sub_exp_variable* UNUSEDARG(left))
{
    RETURN_SANITY_OKAY();
}

const c_exp_sanity* c_sub_exp_ternary_false_end_marker::can_precede(c_sub_exp_numeric_immediate* UNUSEDARG(left))
{
    RETURN_SANITY_OKAY();
}

const c_exp_sanity* c_sub_exp_ternary_false_end_marker::can_precede(c_sub_exp_const* UNUSEDARG(left))
{
    RETURN_SANITY_OKAY();
}

const c_exp_sanity* c_sub_exp_ternary_false_end_marker::can_precede(c_sub_exp_func* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_ternary_false_end_marker::can_precede(c_sub_exp_eval_str* UNUSEDARG(left))
{
    RETURN_SANITY_OKAY();
}

const c_exp_sanity* c_sub_exp_ternary_false_end_marker::can_precede(c_sub_exp_immediate_str* UNUSEDARG(left))
{
    RETURN_SANITY_OKAY();
}

const c_exp_sanity* c_sub_exp_ternary_false_end_marker::can_precede(c_sub_exp_oper* left)
{
    if(left->get_oper().get_prec()==1) // "/i"
        RETURN_USER_ERROR("Invalid usage of case-insensitive index subscripting operator");

    if(left->get_oper().get_type()==OPER_UNARY)
        RETURN_SANITY_OKAY();

    if(left->get_oper().get_type()==OPER_BINARY)
        RETURN_USER_ERROR("Invalid usage of binary operator");

    if(left->get_oper().get_type()==OPER_BINARY_OR_UNARY)
        RETURN_SANITY_OKAY();

    if(left->get_oper().get_type()==OPER_TERNARY)
        INTERNAL_ERROR_OOPS_BUG();

    if(left->get_oper().get_type()==OPER_TERNARY_FALSE)
        INTERNAL_ERROR_OOPS_BUG();

    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_ternary_false_end_marker::can_precede(c_sub_exp_oper_ternary* UNUSEDARG(left))
{
    RETURN_SANITY_OKAY();
}

const c_exp_sanity* c_sub_exp_ternary_false_end_marker::can_precede(c_sub_exp_oper_ternary_true* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_ternary_false_end_marker::can_precede(c_sub_exp_oper_ternary_false* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_ternary_false_end_marker::can_precede(c_sub_exp_contitional_node_helper* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_ternary_false_end_marker::can_precede(c_sub_exp_end_parenthesis* UNUSEDARG(left))
{
    RETURN_SANITY_OKAY();
}

const c_exp_sanity* c_sub_exp_ternary_false_end_marker::can_precede(c_sub_exp_end_index* UNUSEDARG(left))
{
    RETURN_SANITY_OKAY();
}

const c_exp_sanity* c_sub_exp_ternary_false_end_marker::can_precede(c_sub_exp_func_end_args* UNUSEDARG(left))
{
    RETURN_SANITY_OKAY();
}

const c_exp_sanity* c_sub_exp_ternary_false_end_marker::can_precede(c_sub_exp_func_arg_separator* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_ternary_false_end_marker::can_precede(c_sub_exp_conditional_end_marker* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_ternary_false_end_marker::can_precede(c_sub_exp_dummy* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_ternary_false_end_marker::can_precede(c_sub_exp_ternary_true_end_marker* UNUSEDARG(left))
{
    INTERNAL_ERROR_OOPS_BUG();
}

const c_exp_sanity* c_sub_exp_ternary_false_end_marker::can_precede(c_sub_exp_ternary_false_end_marker* UNUSEDARG(left))
{
    RETURN_SANITY_OKAY();
}

// 2nd dispatchers

const c_exp_sanity* c_sub_exp_root::can_follow(c_sub_exp& right)
{
    return right.can_precede(this);
}

const c_exp_sanity* c_sub_exp_root_end_dummy::can_follow(c_sub_exp& right)
{
    return right.can_precede(this);
}

const c_exp_sanity* c_sub_exp_parenthesis::can_follow(c_sub_exp& right)
{
    return right.can_precede(this);
}

const c_exp_sanity* c_sub_exp_index::can_follow(c_sub_exp& right)
{
    return right.can_precede(this);
}

const c_exp_sanity* c_sub_exp_variable::can_follow(c_sub_exp& right)
{
    return right.can_precede(this);
}

const c_exp_sanity* c_sub_exp_numeric_immediate::can_follow(c_sub_exp& right)
{
    return right.can_precede(this);
}

const c_exp_sanity* c_sub_exp_const::can_follow(c_sub_exp& right)
{
    return right.can_precede(this);
}

const c_exp_sanity* c_sub_exp_func::can_follow(c_sub_exp& right)
{
    return right.can_precede(this);
}

const c_exp_sanity* c_sub_exp_eval_str::can_follow(c_sub_exp& right)
{
    return right.can_precede(this);
}

const c_exp_sanity* c_sub_exp_immediate_str::can_follow(c_sub_exp& right)
{
    return right.can_precede(this);
}

const c_exp_sanity* c_sub_exp_oper::can_follow(c_sub_exp& right)
{
    return right.can_precede(this);
}

const c_exp_sanity* c_sub_exp_oper_ternary::can_follow(c_sub_exp& right)
{
    return right.can_precede(this);
}

const c_exp_sanity* c_sub_exp_oper_ternary_true::can_follow(c_sub_exp& right)
{
    return right.can_precede(this);
}

const c_exp_sanity* c_sub_exp_oper_ternary_false::can_follow(c_sub_exp& right)
{
    return right.can_precede(this);
}

const c_exp_sanity* c_sub_exp_contitional_node_helper::can_follow(c_sub_exp& right)
{
    return right.can_precede(this);
}

const c_exp_sanity* c_sub_exp_end_parenthesis::can_follow(c_sub_exp& right)
{
    return right.can_precede(this);
}

const c_exp_sanity* c_sub_exp_end_index::can_follow(c_sub_exp& right)
{
    return right.can_precede(this);
}

const c_exp_sanity* c_sub_exp_func_end_args::can_follow(c_sub_exp& right)
{
    return right.can_precede(this);
}

const c_exp_sanity* c_sub_exp_func_arg_separator::can_follow(c_sub_exp& right)
{
    return right.can_precede(this);
}

const c_exp_sanity* c_sub_exp_conditional_end_marker::can_follow(c_sub_exp& right)
{
    return right.can_precede(this);
}

const c_exp_sanity* c_sub_exp_dummy::can_follow(c_sub_exp& right)
{
    return right.can_precede(this);
}

const c_exp_sanity* c_sub_exp_ternary_true_end_marker::can_follow(c_sub_exp& right)
{
    return right.can_precede(this);
}

const c_exp_sanity* c_sub_exp_ternary_false_end_marker::can_follow(c_sub_exp& right)
{
    return right.can_precede(this);
}

//

c_exp_sanity::c_exp_sanity(e_can_follow rc, const char* msg, int line)
    : m_rc(rc)
    , m_msg(msg)
    , m_line(line)
{
    static const char* oops="Internal error - PLEASE REPORT THIS PROBLEM - it is probably a bug";
    static const char* other="Internal error - PLEASE REPORT THIS PROBLEM - it is definitely a bug";

    if(NULL==m_msg)
        m_msg=other;

    switch(m_rc) {
        case CF_USER_FATAL_ERROR:
            break;
        case CF_OKAY:
            break;
        case CF_WARN:
            break;
        case CF_OOPS_BUG:
            m_msg=oops;
            break;
    }
}

c_exp_sanity::e_can_follow c_exp_sanity::get(const char* &msg, const char* &file, int& line) const
{
    msg=m_msg;
    file=__FILE__;
    line=m_line;
    return m_rc;
}

