/***************************************************************************
                          irc.cpp  -  description
                             -------------------
    begin                : Thu Dec 9 2004
    copyright            : (C) 2004-2007 by Marian VooDooMan Meravy
    email                : ghostvoodooman -a.T= users (dot) sourceforge.net
 ***************************************************************************/

/* $Id: irc.cpp,v 1.86 2008/07/02 09:35:46 ghostvoodooman Exp $ */

/***************************************************************************

VooDoo cIRCle - an IRC (ro)bot
Copyright (C) 2004-2008 by Marian VooDooMan Meravy
  (ghostvoodooman -a.T= users (dot) sourceforge.net)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************/

/*!
    \file
    \brief Handles core communication with IRC server
    \version $Id: irc.cpp,v 1.86 2008/07/02 09:35:46 ghostvoodooman Exp $
*/

#if defined(HAVE_CONFIG_H)
#   include <config.h>
#endif

#include <string>
#include <vector>
#include <stack>
#include <list>
#include <map>

#include <stdlib.h>
#include <time.h>

#include "utils/memleak.h"
#include "utils/assert.h"
#include "utils.h"
#include "irc.h"
#include "log.h"
#include "sock.h"
#include "logic.h"
#include "dcc.h"
#include "shared.h"
#include "match.h"
#include "filesys.h"
#include "lang.h"
#include "conf.h"
#include "stats.h"
#include "compat.h"
#include "ssl.h"
#include "utils.h"
#include "modules.h"
#include "threads.h"
#include "casemap.h"
#include "irc.h"
#include "params.h"
#include "var/c_var_base.h"
#include "var/c_var_bool.h"
#include "var/c_var_int.h"
#include "var/c_var_string.h"
#include "var/c_var_array.h"
#include "irc/c_online_channel.h"
#include "irc/c_online_user.h"
#include "irc/c_irc_server.h"
#include "dcc/c_dcc_request_whois.h"
#include "logic/event_dispatch.h"
#include "internal_state.h"

extern deque<c_user> r_terminators;
extern deque<c_user> r_user_templates;
extern deque<c_user> r_users;

#include "utils/precomp.h"

#if defined(_MSC_VER) || defined(__BORLANDC__)
#   pragma hdrstop
#endif

enum
{
    IRC_RESPONSE_TIMEOUT = 20,      //!< Response timeout in seconds
};

//---------------------------------------------------------------------------
#ifdef __BORLANDC__
#pragma package(smart_init)
#endif

using namespace std;

//-------------------

string irc_server_bind_ipv4;            //!< IPv4 bind address when connecting to IRC server
string irc_server_bind_ipv6;            //!< IPv6 bind address when connecting to IRC server

size_t irc_bot_flood_bytes;             //!< Number of bytes considered as flood from bot
my_time_t irc_bot_flood_seconds;           //!< Number of seconds to clear flood buffer (from bot)
unsigned int irc_bot_flood_lines;       //!< Number of lines maximum in irc_bot_flood_seconds
my_time_t irc_bot_flood_other_seconds;     //!< Number of seconds to clear flood buffer for non-PRIVMSG && non-NOTICE messages (from bot)

bool irc_follow_redirs=false;           //!< Stores "allow_redirect" state from logic.txt (true if bot should follow redirect hints from server)
string irc_recommended_server;          //!< Hint from server: recommended server (together with irc_recommended_port)
unsigned short irc_recommended_port;    //!< Hint from server: recommended port (together with irc_recommended_server)

c_socket irc_socket;                    //!< Handle of socket connected to server (zero for disconnected state)
bool irc_ssl;                           //!< "true" if socket connection to IRC server is via SSL
c_ssl* irc_ssl_connection=NULL;         //!< Descriptor of SSL connection to IRC server
my_time_t irc_last_keepalive_detection;    //!< Used for storing time of last command sent to server (for detection of lost connection)
bool irc_keepalive_sent;                //!< Was PING to IRC sent?
string irc_ssl_section;                 //!< Current section in file "ssl.txt" to use for IRC server connection

bool irc_nick_confirmed=false;          //!< 'true' if we are successfully connected, and value of 'irc_nick' is our real nick on IRC (not trying-to-be nick while connecting)
string irc_nick;                        //!< Bot's current nick
string irc_fullname;                    //!< Bot's current full name
string irc_unique_id;                   //!< Bot's unique ID
string irc_ident="*";                   //!< Bot's ident username
string irc_host;                        //!< Bot's host, as reported from server
string irc_ip4;                         //!< Bot's IPv4 address, resolved from irc_host
string irc_ip6;                         //!< Bot's IPv6 address, resolved from irc_host

string irc_server_host;                 //!< Current IRC server host
unsigned short irc_server_port;         //!< Current IRC server port

vector<string> irc_nicks;               //!< Bot's nicks available to try them if they are unused (from logic.txt)
vector<c_irc_server> irc_servers;       //!< List of irc servers
vector<c_irc_server>::iterator irc_server_iterator=irc_servers.end();   //!< Current irc server

my_time_t filesys_notify_interval;         //!< Interval for filesystem notifications
my_time_t filesys_last_notify;             //!< Timestam of last filesystem notification

multimap<string,string> irc_into_whois; //!< first==nick, second==parts of whois (lines)

bool irc_sent_quit=false;               //!< If we sent QUIT already
string irc_proto_buff;                  //!< Buffer for message that does not ends with EOL

struct c_dns_wait
{
    string m_nick;
    string m_ident;
    string m_host;
    string m_fullname;

    c_dns_wait(const string& nick, const string& ident, const string& host, const string& fullname)
        : m_nick(nick)
        , m_ident(ident)
        , m_host(host)
        , m_fullname(fullname)
    { ; }
};

static list<c_dns_wait> irc_waiting_for_dns;    //!< nick,host

/*!
    \brief Stores data about channel, which is +k, and we are trying all keys in whole history
    \author VooDooMan
    \version 1
    \date 2004
*/
struct c_key_chan
{
    string channel_name;                //!< Name of channel
    vector<string> keys;                //!< Keys from history to try them
};

vector<c_key_chan> irc_key_chans;       //!< List of channels which we are not on, because they are +k and we are trying all keys in the history

vector<c_online_channel> irc_channels;          //!< List of channels we are on

vector<c_online_user> tmp_users;                //!< Temporary list of users

vector<c_online_user> gone_users;               //!< Temporary list of users that have QUIT/PART/been KICK-ed

vector<c_online_channel> unknown_users_on_channels;     //!< Temporary list of user used ONLY for flood history

//-------------------

c_msg_to_server last_msg;                       //!< Last message sent to server

/*!
    \brief Stores messages to IRC server we are waiting response for
    \author VooDooMan
    \version 1
    \date 2005
*/
struct c_awaiting_response
{
    string message;                         //!< Sent message
    my_time_t time;                            //!< Timestamp of message
};

vector<c_awaiting_response> waiting_for_response;   //!< Array of last messages to server

deque<c_msg_to_server> msgs_to_server;           //!< Queue of messages that are waiting to be sent to server (for bot's flood control)
deque<c_msg_to_server> msgs_to_server2;          //!< Queue of messages that are waiting to be sent to server (for bot's flood control), but these has lower priority (notifies)
list<string> msgs_from_server;                  //!< Queue of messages from server to be processed
bool irc_socket_error;                          //!< If true, then we have socket error on connection to IRC server

list<c_flood_history> flood_history;            //!< History of messages sent to server (for bot's flood control)
list<c_flood_history> flood_history_other;      //!< History of other messages sent to server (non-PRIVMSG && non-NOTICE) (for bot's flood protection)

vector<string> irc_cannot_join_channels;        //!< List of channels that i am banned
my_time_t irc_last_join_try;                       //!< Timestamp of last try to join because of +b on myself, or +l, or +i

c_005 irc_isupport;

/*!
    \brief One entry of mode queue (used when "compress_mode_wait" form conf.txt is more than zero)
    \author VooDooMan
    \version 2
    \date 2004, 2007-12-02
*/
struct c_mode_queue_entry
{
    string channel;                             //!< Name of channel
    string mode;                                //!< One mode (+/- prefix + mode char)
    string param;                               //!< Parameter (argument) of mode
    my_time_t submit;                              //!< Time when mode was wished to post
    bool param_is_nick;                         //!< 'true' if 'param' is nick
};

list<c_mode_queue_entry> mode_queue;            //!< Mode queue used when "compress_mode_wait" form conf.txt is more than zero
list<c_mode_queue_entry> modes_from_server;     //!< Modes that server set

c_global_keeper global_keeper;

e_casemap irc_casemapping = CASE_ASCII;

static bool irc_was_disconnect=true;

MY_STATIC void process_tracked_nicks(const string& old_nick, const string& new_nick);
MY_STATIC void process_tracked_nicks_chan(const string& channel, const string& part_nick);
MY_STATIC void irc_process_nicks();

/*!
    \brief Compares two strings, case in-sensitive, according to casemapping from IRC server
    \author VooDooMan
    \version 1
    \date 2007-03-13
    \param s1 String 1
    \param s2 String 2
    \retval false Strings are the same (case in-sensitive)
    \retval true Strings are different (case in-sensitive)
*/
bool irc_cmp_strings_case_insensitive(string s1, string s2)
{
    s1=strtolower(s1,irc_casemapping);
    s2=strtolower(s2,irc_casemapping);
    return (s1.compare(s2)) ? true : false;
}

c_global_keeper::c_session_keeper::c_session_keeper()
{
    saved_session=var_keeper->get_current_session();
    saved_local_id=var_keeper->set_current_local(0);
    var_keeper->set_current_session(SESSION_RESERVED_ID);
}

c_global_keeper::c_session_keeper::~c_session_keeper()
{
    var_keeper->set_current_session(saved_session);
    var_keeper->set_current_local(saved_local_id);
}

/*!
    \brief Resets $_users array
    \author VooDooMan
    \version 1
    \date 2005
*/
void c_global_keeper::reset_users()
{
    c_session_keeper keeper;

    if(!super_global)
        return;
    super_global->unset_ro("$_users");
    users=MY_NEW c_var_array();
    super_global->set_ro("$_users",users);
}

/*!
    \brief Resets $_isupport array
    \author VooDooMan
    \version 1
    \date 2005
*/
void c_global_keeper::reset_isupport()
{
    c_session_keeper keeper;

    if(!super_global)
        return;
    super_global->unset_ro("$_isupport");
    isupport=MY_NEW c_var_array();
    super_global->set_ro("$_isupport",isupport);
}

/*!
    \brief Resets $_user_mode_prefix array
    \author VooDooMan
    \version 1
    \date 2005
*/
void c_global_keeper::reset_user_mode_prefix()
{
    c_session_keeper keeper;

    if(!super_global)
        return;
    super_global->unset_ro("$_user_mode_prefix");
    user_mode_prefix=MY_NEW c_var_array();
    super_global->set_ro("$_user_mode_prefix",user_mode_prefix);
}

/*!
    \brief Resets $_channels array
    \author VooDooMan
    \version 1
    \date 2005
*/
void c_global_keeper::reset_channels()
{
    c_session_keeper keeper;

    if(!super_global)
        return;
    super_global->unset_ro("$_channels");
    channels=MY_NEW c_var_array();
    super_global->set_ro("$_channels",channels);
}

/*!
    \brief Sets feature mapping
    \author VooDooMan
    \version 1
    \date 2005
    \param feature1 Name of feature
    \param feature2 Value of feature
*/
void c_global_keeper::add_isupport(const string& feature1, const string& feature2)
{
    c_session_keeper keeper;

    if(!isupport)
        return;
    isupport->set_ro(feature1,MY_NEW c_var_string(feature2));
}

/*!
    \brief Sets user mode mapping
    \author VooDooMan
    \version 1
    \date 2005
    \param ch1 Mode prefix
    \param ch2 Mode char
*/
void c_global_keeper::add_user_mode_prefix(char ch1, char ch2)
{
    c_session_keeper keeper;

    if(!user_mode_prefix)
        return;
    user_mode_prefix->set_ro((string)""+ch1,MY_NEW c_var_string((string)""+ch2));
}

/*!
    \brief Adds a new channel
    \author VooDooMan
    \version 1
    \date 2005
    \param name Name of channel
*/
void c_global_keeper::add_channel(const string& name)
{
    c_session_keeper keeper;

    if(!channels)
        return;
    c_var_array* chan=MY_NEW c_var_array();
    chan->set_ro("key",MY_NEW c_var_string(""));
    chan->set_ro("limit",MY_NEW c_var_string(""));
    chan->set_ro("mode",MY_NEW c_var_string(""));
    chan->set_ro("topic",MY_NEW c_var_string(""));
    chan->set_ro("bans",MY_NEW c_var_array());
    chan->set_ro("exceptions",MY_NEW c_var_array());
    chan->set_ro("invites",MY_NEW c_var_array());
    chan->set_ro("reops",MY_NEW c_var_array());
    channels->set_ro(name,chan);
}

/*!
    \brief Deletes a channel
    \author VooDooMan
    \version 1
    \date 2005
    \param name Name of channel
    \param channels $_channels array
*/
void c_global_keeper::delete_chan(c_var_array* channels, const string& name)
{
    c_session_keeper keeper;

    if(!channels)
        return;

    var_keeper->array_unset(core_mod,channels,name,true,"",__my_func__);
}

/*!
    \brief Adds an user
    \author VooDooMan
    \version 1
    \date 2005
    \param channel Name of channel
    \param nick Nick
    \param mode Mode
*/
void c_global_keeper::add_user(const string& channel, const string& nick, const string& mode)
{
    c_session_keeper keeper;

    c_var_array* user=dynamic_cast<c_var_array*>(users->get(nick,true));
    if(!user) {
        user=MY_NEW c_var_array();
        users->set_ro(nick,user);
    }

    c_var_string* s;

    s=dynamic_cast<c_var_string*>(user->get("name",false));
    if(!s)
        user->set_ro("name",MY_NEW c_var_string(""));

    s=dynamic_cast<c_var_string*>(user->get("nick",false));
    if(!s)
        user->set_ro("nick",MY_NEW c_var_string(nick));

    s=dynamic_cast<c_var_string*>(user->get("ident",false));
    if(!s)
        user->set_ro("ident",MY_NEW c_var_string(""));

    s=dynamic_cast<c_var_string*>(user->get("host",false));
    if(!s)
        user->set_ro("host",MY_NEW c_var_string(""));

    s=dynamic_cast<c_var_string*>(user->get("fullname",false));
    if(!s)
        user->set_ro("fullname",MY_NEW c_var_string(""));

    c_var_bool* b;

    b=dynamic_cast<c_var_bool*>(user->get("host_unknown",false));
    if(!b)
        user->set_ro("host_unknown",MY_NEW c_var_bool(true));

    b=dynamic_cast<c_var_bool*>(user->get("host_bot",false));
    if(!b)
        user->set_ro("host_bot",MY_NEW c_var_bool(false));

    b=dynamic_cast<c_var_bool*>(user->get("irc_op",false));
    if(!b)
        user->set_ro("irc_op",MY_NEW c_var_bool(false));

    c_var_array* a;

    a=dynamic_cast<c_var_array*>(user->get("hostmasks",false));
    if(!a)
        user->set_ro("hostmasks",MY_NEW c_var_array());

    a=dynamic_cast<c_var_array*>(user->get("fullname_masks",false));
    if(!a)
        user->set_ro("fullname_masks",MY_NEW c_var_array());

    a=dynamic_cast<c_var_array*>(user->get("channels",false));
    if(!a)
        user->set_ro("channels",a=MY_NEW c_var_array());
    c_var_array* chan=dynamic_cast<c_var_array*>(a->get(channel,true));
    if(!chan) {
        chan=MY_NEW c_var_array();
        a->set_ro(channel,chan);
    }
    chan->set_ro("channel",MY_NEW c_var_string(channel));
    chan->set_ro("mode",MY_NEW c_var_string(mode));
    chan->set_read_only(true);
    /*
    delete_chan(a,channel);
    a->push_ro(NULL,chan);
    */

    a=dynamic_cast<c_var_array*>(user->get("groups",false));
    if(!a)
        user->set_ro("groups",MY_NEW c_var_array());
}

/*!
    \brief Updates an user (after WHOIS)
    \author VooDooMan
    \version 2
    \date 2005, 2008-01-20
    \param nick Nick
    \param ident Ident
    \param host Host
    \param fullname Fullname
*/
void c_global_keeper::add_user_whois1(const string& nick, const string& ident, const string& host, const string& fullname)
{
    c_session_keeper keeper;

    c_var_array* user=dynamic_cast<c_var_array*>(users->get(nick,true));
    if(!user) {
        user=MY_NEW c_var_array();
        users->set_ro(nick,user);
    }

    c_var_string* s;

    user->unset_ro("ident");
    s=dynamic_cast<c_var_string*>(user->get("ident",false));
    if(!s)
        user->set_ro("ident",MY_NEW c_var_string(ident));

    user->unset_ro("host");
    s=dynamic_cast<c_var_string*>(user->get("host",false));
    if(!s)
        user->set_ro("host",MY_NEW c_var_string(host));

    user->unset_ro("fullname");
    s=dynamic_cast<c_var_string*>(user->get("fullname",false));
    if(!s)
        user->set_ro("fullname",MY_NEW c_var_string(fullname));
}

/*!
    \brief Updates an user (after WHOIS), sets ircop flag to true
    \author VooDooMan
    \version 1
    \date 2005
    \param nick Nick
*/
void c_global_keeper::add_user_ircop(const string& nick)
{
    c_session_keeper keeper;

    c_var_array* user=dynamic_cast<c_var_array*>(users->get(nick,true));
    if(!user) {
        user=MY_NEW c_var_array();
        users->set_ro(nick,user);
    }

    user->set_ro("irc_op",MY_NEW c_var_bool(true));
}

/*!
    \brief Updates an user (after WHOIS)
    \author VooDooMan
    \version 1
    \date 2005
    \param nick Nick
    \param channel Channel
    \param mode Mode
*/
void c_global_keeper::add_user_whois2(const string& nick, const string& channel, const string& mode)
{
    c_session_keeper keeper;

    c_var_array* user=dynamic_cast<c_var_array*>(users->get(nick,true));
    if(!user) {
        user=MY_NEW c_var_array();
        users->set_ro(nick,user);
    }

    c_var_array* a=dynamic_cast<c_var_array*>(user->get("channels",false));
    if(!a)
        user->set_ro("channels",a=MY_NEW c_var_array());

    c_var_array* chan=dynamic_cast<c_var_array*>(a->get(channel,true));
    if(!chan) {
        chan=MY_NEW c_var_array();
        a->set_ro(channel,chan);
    }
    chan->set_ro("channel",MY_NEW c_var_string(channel));
    chan->set_ro("mode",MY_NEW c_var_string(mode));
    chan->set_read_only(true);
}

/*!
    \brief Updates an user (after WHOIS)
    \author VooDooMan
    \version 1
    \date 2006-01-03
    \param nick Nick
    \param lines Raw lines with WHOIS
*/
void c_global_keeper::add_user_whois3(const string& nick, const vector<string>& lines)
{
    if(lines.size()<=1) {
        // empty lines, or, one record: end-of-whois
        // User just changed nickname or has just quit, and WHOIS request was in queue
        // to send to IRC server.
        // Normally, no-such-nick error message is preceeding in our case.
        // We just ignore this, and do not create an entry in $_users
        return;
    }

    c_session_keeper keeper;

    c_var_array* user=dynamic_cast<c_var_array*>(users->get(nick,true));
    if(!user) {
        user=MY_NEW c_var_array();
        users->set_ro(nick,user);
    }

    user->unset_ro("raw_whois");

    c_var_array* a=MY_NEW c_var_array();
    vector<string>::const_iterator i1;
    for(i1=lines.begin(); i1!=lines.end(); ++i1)
        a->push_ro(NULL,MY_NEW c_var_string(*i1,true));
    a->set_read_only(true);

    user->set_ro("raw_whois",a);
}

/*!
    \brief Updates an user (after DNS)
    \author VooDooMan
    \version 1
    \date 2008-01-20
    \param nick Nick
    \param in_logic_as Name as in "logic.txt"
*/
void c_global_keeper::add_user_whois4(const string& nick, const string& in_logic_as)
{
    c_session_keeper keeper;

    c_var_array* user=dynamic_cast<c_var_array*>(users->get(nick,true));
    if(!user) {
        user=MY_NEW c_var_array();
        users->set_ro(nick,user);
    }

    c_var_string* s;

    user->unset_ro("name");
    s=dynamic_cast<c_var_string*>(user->get("name",false));
    if(!s)
        user->set_ro("name",MY_NEW c_var_string(in_logic_as));

    c_var_array* a;

    if(in_logic_as.empty())
        return;

    deque<c_user>::iterator i2;
    for(i2=r_users.begin(); i2!=r_users.end(); ++i2) {
        if(in_logic_as.compare((*i2).name))
            continue;

        user->set_ro("host_unknown",MY_NEW c_var_bool((*i2).host_unknown));
        user->set_ro("host_bot",MY_NEW c_var_bool((*i2).host_bot));

        user->unset_ro("hostmasks");
        a=dynamic_cast<c_var_array*>(user->get("hostmasks",false));
        if(!a) {
            a=MY_NEW c_var_array();
            vector<string>::iterator i3;
            for(i3=(*i2).hostmask.begin(); i3!=(*i2).hostmask.end(); ++i3)
                a->push_ro(NULL,MY_NEW c_var_string(*i3));
            user->set_ro("hostmasks",a);
        }

        user->unset_ro("fullname_masks");
        a=dynamic_cast<c_var_array*>(user->get("fullname_masks",false));
        if(!a) {
            a=MY_NEW c_var_array();
            vector<string>::iterator i3;
            for(i3=(*i2).fullname.begin(); i3!=(*i2).fullname.end(); ++i3)
                a->push_ro(NULL,MY_NEW c_var_string(*i3));
            user->set_ro("fullname_masks",a);
        }

        user->unset_ro("groups");
        a=dynamic_cast<c_var_array*>(user->get("groups",false));
        if(!a) {
            a=MY_NEW c_var_array();
            vector<string>::iterator i3;
            for(i3=(*i2).groups.begin(); i3!=(*i2).groups.end(); ++i3)
                a->push_ro(NULL,MY_NEW c_var_string(*i3));
            user->set_ro("groups",a);
        }
    }
}

/*!
    \brief Gets channel
    \author VooDooMan
    \version 1
    \date 2005
    \param channel Channel
    \return Returns specified channel
*/
c_var_array* c_global_keeper::get_chan(const string& channel)
{
    c_session_keeper keeper;

    c_var_base* res=var_keeper->array_get(core_mod,channels,channel,true,"",__my_func__);
    return dynamic_cast<c_var_array*>(res);
}

/*!
    \brief Adds ban mask to channel
    \author VooDooMan
    \version 1
    \date 2005
    \param channel Channel
    \param mask Mask
*/
void c_global_keeper::add_banned(const string& channel, const string& mask)
{
    c_session_keeper keeper;

    c_var_array* chan=get_chan(channel);
    if(!chan)
        return;
    c_var_array* bans=dynamic_cast<c_var_array*>(chan->get("bans",false));
    if(!bans) {
        bans=MY_NEW c_var_array();
        chan->set_ro("bans",bans);
    }
    bans->push_ro(NULL,MY_NEW c_var_string(mask));
}

/*!
    \brief Adds exception mask to channel
    \author VooDooMan
    \version 1
    \date 2005
    \param channel Channel
    \param mask Mask
*/
void c_global_keeper::add_except(const string& channel, const string& mask)
{
    c_session_keeper keeper;

    c_var_array* chan=get_chan(channel);
    if(!chan)
        return;
    c_var_array* bans=dynamic_cast<c_var_array*>(chan->get("exceptions",false));
    if(!bans) {
        bans=MY_NEW c_var_array();
        chan->set_ro("exceptions",bans);
    }
    bans->push_ro(NULL,MY_NEW c_var_string(mask));
}

/*!
    \brief Adds invite mask to channel
    \author VooDooMan
    \version 1
    \date 2005
    \param channel Channel
    \param mask Mask
*/
void c_global_keeper::add_invite(const string& channel, const string& mask)
{
    c_session_keeper keeper;

    c_var_array* chan=get_chan(channel);
    if(!chan)
        return;
    c_var_array* bans=dynamic_cast<c_var_array*>(chan->get("invites",false));
    if(!bans) {
        bans=MY_NEW c_var_array();
        chan->set_ro("invites",bans);
    }
    bans->push_ro(NULL,MY_NEW c_var_string(mask));
}

/*!
    \brief Adds reop mask to channel
    \author VooDooMan
    \version 1
    \date 2005
    \param channel Channel
    \param mask Mask
*/
void c_global_keeper::add_reop(const string& channel, const string& mask)
{
    c_session_keeper keeper;

    c_var_array* chan=get_chan(channel);
    if(!chan)
        return;
    c_var_array* bans=dynamic_cast<c_var_array*>(chan->get("reops",false));
    if(!bans) {
        bans=MY_NEW c_var_array();
        chan->set_ro("reops",bans);
    }
    bans->push_ro(NULL,MY_NEW c_var_string(mask));
}

/*!
    \brief Changes user's mode on channel
    \author VooDooMan
    \version 1
    \date 2005
    \param channel Channel
    \param prefix '+', or '-'
    \param mode New mode
    \param nick Nick
*/
void c_global_keeper::change_user_mode(const string& channel, const string& nick, char prefix, char mode)
{
    c_session_keeper keeper;

    c_var_array* user=dynamic_cast<c_var_array*>(users->get(nick,true));
    if(!user) {
        user=MY_NEW c_var_array();
        users->set_ro(nick,user);
    }

    c_var_array* chans=dynamic_cast<c_var_array*>(user->get("channels",false));
    if(!chans) {
        chans=MY_NEW c_var_array();
        user->set_ro("channels",chans);
    }

    c_var_array* chan=dynamic_cast<c_var_array*>(chans->get(channel,true));
    if(!chan) {
        chan=MY_NEW c_var_array();
        chan->set_ro("channel",MY_NEW c_var_string(channel));
        chans->push_ro(NULL,chan);
    }
    c_var_string* old=dynamic_cast<c_var_string*>(chan->get("mode",false));
    if(!old)
        old=MY_NEW c_var_string("");
    string old_=old->get();
    while(old_.find(mode)!=string::npos)
        old_.erase(old_.find(mode),1);
    while(old_.find('0')!=string::npos)
        old_.erase(old_.find('0'),1);
    if('+'==prefix)
        old_+=(string)""+mode;
    if(old_.empty())
        old_="0";
    old->set_ro(old_);
    chan->set_ro("mode",old);
}

/*!
    \brief Changes channel's mode
    \author VooDooMan
    \version 1
    \date 2005
    \param channel Channel
    \param prefix '+', or '-'
    \param mode New mode
*/
void c_global_keeper::change_chan_mode(const string& channel, char prefix, char mode)
{
    c_session_keeper keeper;

    c_var_array* chan=this->get_chan(channel);
    if(!chan)
        return;

    c_var_string* old=dynamic_cast<c_var_string*>(chan->get("mode",false));
    if(!old)
        old=MY_NEW c_var_string("");
    string old_=old->get();
    while(old_.find(mode)!=string::npos)
        old_.erase(old_.find(mode),1);
    while(old_.find('0')!=string::npos)
        old_.erase(old_.find('0'),1);
    if('+'==prefix)
        old_+=(string)""+mode;
    if(old_.empty())
        old_="0";
    old->set_ro(old_);
    chan->set_ro("mode",old);
}

/*!
    \brief Changes channel's key
    \author VooDooMan
    \version 1
    \date 2005
    \param channel Channel
    \param prefix '+', or '-'
    \param key New key
*/
void c_global_keeper::change_chan_key(const string& channel, char prefix, const string& key)
{
    c_session_keeper keeper;

    c_var_array* chan=this->get_chan(channel);
    if(!chan)
        return;

    c_var_string* old=dynamic_cast<c_var_string*>(chan->get("key",false));
    if(!old)
        old=MY_NEW c_var_string("");
    if('+'==prefix)
        old->set_ro(key);
    else
        old->set_ro("");
    chan->set_ro("key",old);
}

/*!
    \brief Changes channel's limit
    \author VooDooMan
    \version 1
    \date 2005
    \param channel Channel
    \param prefix '+', or '-'
    \param limit New limit
*/
void c_global_keeper::change_chan_limit(const string& channel, char prefix, const string& limit)
{
    c_session_keeper keeper;

    c_var_array* chan=this->get_chan(channel);
    if(!chan)
        return;

    c_var_string* old=dynamic_cast<c_var_string*>(chan->get("limit",false));
    if(!old)
        old=MY_NEW c_var_string("");
    if('+'==prefix)
        old->set_ro(limit);
    else
        old->set_ro("");
    chan->set_ro("limit",old);
}

/*!
    \brief Erases channel's ban mask, and inserts it again (safe insertion, to avoid duplicates)
    \author VooDooMan
    \version 2
    \date 2005, 2008-03-09
    \param channel Channel
    \param prefix '+', or '-'
    \param mask Mask
*/
void c_global_keeper::change_ban(const string& channel, char prefix, const string& mask)
{
    c_session_keeper keeper;

    c_var_array* chan=get_chan(channel);
    if(!chan)
        return;
    c_var_array* x=dynamic_cast<c_var_array*>(chan->get("bans",false));
    if(!x) {
        x=MY_NEW c_var_array();
        chan->set_ro("bans",x);
    }

    var_keeper->array_remove_values(x,mask);

    if('+'==prefix)
        x->push_ro(NULL,MY_NEW c_var_string(mask));
}

/*!
    \brief Erases channel's exception mask, and inserts it again (safe insertion, to avoid duplicates)
    \author VooDooMan
    \version 2
    \date 2005, 2008-03-09
    \param channel Channel
    \param prefix '+', or '-'
    \param mask Mask
*/
void c_global_keeper::change_except(const string& channel, char prefix, const string& mask)
{
    c_session_keeper keeper;

    c_var_array* chan=get_chan(channel);
    if(!chan)
        return;
    c_var_array* x=dynamic_cast<c_var_array*>(chan->get("exceptions",false));
    if(!x) {
        x=MY_NEW c_var_array();
        chan->set_ro("exceptions",x);
    }

    var_keeper->array_remove_values(x,mask);

    if('+'==prefix)
        x->push_ro(NULL,MY_NEW c_var_string(mask));
}

/*!
    \brief Erases channel's invite mask, and inserts it again (safe insertion, to avoid duplicates)
    \author VooDooMan
    \version 2
    \date 2005, 2008-03-09
    \param channel Channel
    \param prefix '+', or '-'
    \param mask Mask
*/
void c_global_keeper::change_invite(const string& channel, char prefix, const string& mask)
{
    c_session_keeper keeper;

    c_var_array* chan=get_chan(channel);
    if(!chan)
        return;
    c_var_array* x=dynamic_cast<c_var_array*>(chan->get("invites",false));
    if(!x) {
        x=MY_NEW c_var_array();
        chan->set_ro("invites",x);
    }

    var_keeper->array_remove_values(x,mask);

    if('+'==prefix)
        x->push_ro(NULL,MY_NEW c_var_string(mask));
}

/*!
    \brief Erases channel's reop mask, and inserts it again (safe insertion, to avoid duplicates)
    \author VooDooMan
    \version 2
    \date 2005, 2008-03-09
    \param channel Channel
    \param prefix '+', or '-'
    \param mask Mask
*/
void c_global_keeper::change_reop(const string& channel, char prefix, const string& mask)
{
    c_session_keeper keeper;

    c_var_array* chan=get_chan(channel);
    if(!chan)
        return;
    c_var_array* x=dynamic_cast<c_var_array*>(chan->get("reops",false));
    if(!x) {
        x=MY_NEW c_var_array();
        chan->set_ro("reops",x);
    }

    var_keeper->array_remove_values(x,mask);

    if('+'==prefix)
        x->push_ro(NULL,MY_NEW c_var_string(mask));
}

/*!
    \brief Erases channel's other mode mask, and inserts it again (safe insertion, to avoid duplicates)
    \author VooDooMan
    \version 1
    \date 2008-03-09
    \param channel Channel
    \param prefix '+', or '-'
    \param mode Mode
    \param mask Mask
*/
void c_global_keeper::change_other_mode_mask(const string& channel, char prefix, char mode, const string& mask)
{
    c_session_keeper keeper;

    c_var_array* chan=get_chan(channel);
    if(!chan)
        return;
    c_var_array* x=dynamic_cast<c_var_array*>(chan->get("other_modes",false));
    if(!x) {
        x=MY_NEW c_var_array();
        chan->set_ro("other_modes",x);
    }

    string my_mode=(string)""+mode;

    var_keeper->array_remove_values(x,my_mode,mask);

    if('+'==prefix)
        x->set_ro(my_mode,MY_NEW c_var_string(mask));
}

/*!
    \brief Sets channel's topic
    \author VooDooMan
    \version 1
    \date 2005
    \param channel Channel
    \param topic Topic
*/
void c_global_keeper::change_topic(const string& channel, const string& topic)
{
    c_session_keeper keeper;

    c_var_array* chan=get_chan(channel);
    if(!chan)
        return;
    c_var_string* old=dynamic_cast<c_var_string*>(chan->get("topic",false));
    string debug;
    if(!old) {
        old=MY_NEW c_var_string(topic);
        chan->set_ro("topic",old);
    } else
        old->set_ro(topic);
}

/*!
    \brief User left the channel
    \author VooDooMan
    \version 1
    \date 2005
    \param channel Channel
    \param nick Nick
    \param me Is it me?
*/
void c_global_keeper::nick_left(const string& channel, const string& nick, bool me)
{
    c_session_keeper keeper;

    c_var_array* user=dynamic_cast<c_var_array*>(users->get(nick,true));
    if(!user) {
        user=MY_NEW c_var_array();
        users->set_ro(nick,user);
    }

    c_var_array* chans=dynamic_cast<c_var_array*>(user->get("channels",false));
    if(!chans) {
        chans=MY_NEW c_var_array();
        user->set_ro("channels",chans);
    }

    c_var_array* chan=dynamic_cast<c_var_array*>(chans->get(channel,true));
    if(chan) {
        chans->unset_ro(channel,true);
    }
    if(me) {
        //users->unset_ro_value(user->get("nick",true));
        users->unset_ro(nick,true);
        channels->unset_ro_value(get_chan(channel));
    }

    if(var_keeper->array_num_children(core_mod,chans)==0)
        users->unset_ro(nick,true);
}

/*!
    \brief After rehash
    \author VooDooMan
    \version 1
    \date 2005
*/
void c_global_keeper::rehashed()
{
    c_session_keeper keeper;

    map<string,c_var_base*> idx=var_keeper->array_get_string_indexes(core_mod,users);

    map<string,c_var_base*>::iterator i1;
    for(i1=idx.begin(); i1!=idx.end(); ++i1) {
        string nick=(*i1).first;

        c_var_array* user=dynamic_cast<c_var_array*>((*i1).second);
        if(!user) {
            user=MY_NEW c_var_array();
            users->set_ro(nick,user);
        }

        string in_logic_as;
        c_var_string* in=dynamic_cast<c_var_string*>(user->get("name",false));
        if(in)
            in_logic_as=in->get();

        c_var_array* a;

        deque<c_user>::iterator i2;
        for(i2=r_users.begin(); i2!=r_users.end(); ++i2) {
            if(in_logic_as.compare((*i2).name))
                continue;

            user->set_ro("host_unknown",MY_NEW c_var_bool((*i2).host_unknown));
            user->set_ro("host_bot",MY_NEW c_var_bool((*i2).host_bot));

            user->unset_ro("hostmasks");
            {
                a=MY_NEW c_var_array();
                vector<string>::iterator i3;
                for(i3=(*i2).hostmask.begin(); i3!=(*i2).hostmask.end(); ++i3)
                    a->push_ro(NULL,MY_NEW c_var_string(*i3));
                user->set_ro("hostmasks",a);
            }

            user->unset_ro("fullname_masks");
            {
                a=MY_NEW c_var_array();
                vector<string>::iterator i3;
                for(i3=(*i2).fullname.begin(); i3!=(*i2).fullname.end(); ++i3)
                    a->push_ro(NULL,MY_NEW c_var_string(*i3));
                user->set_ro("fullname_masks",a);
            }

            user->unset_ro("groups");
            {
                a=MY_NEW c_var_array();
                vector<string>::iterator i3;
                for(i3=(*i2).groups.begin(); i3!=(*i2).groups.end(); ++i3)
                    a->push_ro(NULL,MY_NEW c_var_string(*i3));
                user->set_ro("groups",a);
            }
        }
    }
}

/*!
    \brief User has changed nicks
    \author VooDooMan
    \version 1
    \date 2005
    \param old_nick Old nick
    \param new_nick New nick
*/
void c_global_keeper::nick_change(const string& old_nick, const string& new_nick)
{
    c_session_keeper keeper;

    c_var_array* user=dynamic_cast<c_var_array*>(users->get(old_nick,true));

    {
        // because there can be entry in case there was unsuccessful WHOIS on nick
        //     who just QUIT, so perform this test:
        c_var_array* new_user=NULL;
        c_var_base* new_user2=users->get(new_nick,true);
        if(new_user2)
            new_user=dynamic_cast<c_var_array*>(new_user2);
        if(new_user && new_user!=user) {
            // yes, it is our case, kill the entry
            users->unset_ro(new_nick,true);
        }
    }

    if(!user) {
        user=MY_NEW c_var_array();
        users->set_ro(new_nick,user);
    } else {
        var_keeper->array_replace_index(core_mod,users,old_nick,new_nick);
    }
    user->set_ro("nick",MY_NEW c_var_string(new_nick));
}

/*!
    \brief Updates $_uptime
    \author VooDooMan
    \version 1
    \date 2005
*/
void c_global_keeper::update_uptime()
{
    c_session_keeper keeper;

    if(!super_global)
        return;
    if(last_uptime_update!=my_time(NULL)) {
        last_uptime_update=my_time(NULL);

        super_global->unset_ro("$_uptime");
        c_var_int* up=MY_NEW c_var_int((long)stats_uptime(),true);
        super_global->set_ro("$_uptime",up);
    }
}

/*!
    \brief If "smart_mode" in conf.txt is non-zero, removes redundant modes from mode_queue
    \author VooDooMan
    \version 1
    \date 2004
    \param channel Channel name
    \param nick Nick that mode is applied to
    \param mode Mode (+o, -v, ...)
*/
MY_STATIC void irc_remove_redundant_mode(const string& channel, const string& nick, const string& mode)
{
    if(stoi(core_conf_getvar("smart_mode"))!=0) {
        if(mode.length()!=2)
            return;
        list<c_mode_queue_entry>::iterator i1;
    again:
        for(i1=mode_queue.begin(); i1!=mode_queue.end(); ++i1) {
            if(!irc_cmp_strings_case_insensitive((*i1).channel,channel) && !(*i1).mode.compare(mode) && !irc_cmp_strings_case_insensitive((*i1).param,nick)) {
                mode_queue.erase(i1);
                goto again;
            }
        }
    }
}

/*!
    \brief If "smart_mode" in conf.txt is non-zero, removes redundant modes from mode_queue
    \author VooDooMan
    \version 1
    \date 2004
*/
MY_STATIC void irc_remove_redundant_mode2()
{
    if(stoi(core_conf_getvar("smart_mode"))!=0) {
        list<c_mode_queue_entry>::iterator i1;
    again:
        for(i1=mode_queue.begin(); i1!=mode_queue.end(); ++i1) {
            list<c_mode_queue_entry>::iterator i2=i1;
            i2++;
            for(; i2!=mode_queue.end(); ++i2) {
                if(!irc_cmp_strings_case_insensitive((*i1).channel,(*i2).channel) &&
                    !(*i1).mode.compare((*i2).mode) &&
                    !(*i1).param.compare((*i2).param)) {
                        mode_queue.erase(i2);
                        goto again;
                    }
            }
        }
        int compress_mode_wait=stoi(core_conf_getvar("compress_mode_wait"));
    again2:
        // there is +5 seconds to wait, to ensure that mode from server will not be removed too soon
        for(i1=modes_from_server.begin(); i1!=modes_from_server.end(); ++i1) {
            if((*i1).submit+compress_mode_wait+5<my_time(NULL)) {
                modes_from_server.erase(i1);
                goto again2;
            }
        }
    again3:
        for(i1=mode_queue.begin(); i1!=mode_queue.end(); ++i1) {
            list<c_mode_queue_entry>::iterator i2;
            for(i2=modes_from_server.begin(); i2!=modes_from_server.end(); ++i2) {
                if(!irc_cmp_strings_case_insensitive((*i1).channel,(*i2).channel) &&
                    !(*i1).mode.compare((*i2).mode) &&
                    !(*i1).param.compare((*i2).param)) {
                        mode_queue.erase(i1);
                        goto again3;
                    }
            }
        }
    }
}

/*!
    \brief Returns reference to channel that we are on
    \author VooDooMan
    \version 1
    \date 2004
    \param channel_name Name of channel
    \param chan Returns pointer to channel we are on, or NULL if channel was not found
    \return Returns true if found
    \retval true Channel found
    \retval false Channel not found
*/
bool irc_get_online_channel(const string& channel_name, c_online_channel* &chan)
{
    chan=NULL;

    vector<c_online_channel>::iterator i1;
    for(i1=irc_channels.begin(); i1!=irc_channels.end(); ++i1) {
        if(!irc_cmp_strings_case_insensitive((*i1).name,channel_name)) {
            chan=&(*i1);
            return true;
        }
    }

    return false;
}

/*!
    \def IRC_MSG_TOO_LONG
    \brief ERROR: msg is longer than 512 bytes
*/
#define IRC_MSG_TOO_LONG    0x81234567
/*!
    \def SOCK_SEND_ERROR
    \brief ERROR: not all the msg was sent
*/
#define SOCK_SEND_ERROR     0x81234568

/*!
    \brief Imeddiately sends message to server
    \author VooDooMan
    \version 2
    \date 2005
    \param socket Socket handle of connection to server
    \param msg Message to be sent
    \param wait_for_response Set to true if we want to wait for response from IRC server to this command
    \return Returns zero for no error, IRC_MSG_TOO_LONG if msg is too long (>512 bytes), SOCK_SEND_ERROR is not all the msg was sent, or OS's socket I/O error
    \retval 0 No error
    \retval IRC_MSG_TOO_LONG msg is too long (>512 bytes)
    \retval SOCK_SEND_ERROR Not all the msg was sent
    \param deep_sleep If 'true' we are in a deep sleep, it IRC server connection error occurs, "msgs_from_server.clear();" is explicitly called
*/
MY_STATIC int irc_putserv_immediately(c_socket& socket, const char* msg, bool wait_for_response=false, bool deep_sleep=false)
{
    if(!socket.cmp())
        return SOCK_SEND_ERROR;

    if('\0'==msg[0])
        return 0;

    if(strlen(msg)>512) {
        log_debug(GET_STR_DEBUG_INFO_EX("OOPS: Occurred error IRC_MSG_TOO_LONG"));
        return IRC_MSG_TOO_LONG;
    }

    // BEGIN: sanity checks
    char tmp_msg[1024];
    strcpy(tmp_msg,msg);
    for(unsigned int ii=0; ii<strlen(tmp_msg); ++ii) {
        if(tmp_msg[ii]=='\r' && ii!=strlen(tmp_msg)-2) {
            tmp_msg[ii]=0;
            string log="OOPS: Warning: Message to IRC server contained \\r character, truncated message is: \"";
            log+=tmp_msg;
            log+="\"";
            log_debug(GET_STR_DEBUG_INFO_EX(log));
            strcat(tmp_msg,"\r\n");
            break;
        }
        if(tmp_msg[ii]=='\n' && ii!=strlen(tmp_msg)-1) {
            tmp_msg[ii]=0;
            string log="OOPS: Warning: Message to IRC server contained \\n character, truncated message is: \"";
            log+=tmp_msg;
            log+="\"";
            log_debug(GET_STR_DEBUG_INFO_EX(log));
            strcat(tmp_msg,"\r\n");
            break;
        }
    }
    if(strlen(tmp_msg)<3) {
        string log="OOPS: Warning: Message to IRC server is shorter than 3 caracters, appending \"\\r\\n\" (subsequent error message may follow)";
        log_debug(GET_STR_DEBUG_INFO_EX(log));
        strcat(tmp_msg,"\r\n");
    }
    if(tmp_msg[strlen(tmp_msg)-2]!='\r' || tmp_msg[strlen(tmp_msg)-1]!='\n') {
        while(strlen(tmp_msg) && (tmp_msg[strlen(tmp_msg)-1]=='\r' || tmp_msg[strlen(tmp_msg)-1]=='\n'))
            tmp_msg[strlen(tmp_msg)-1]=0;
        string log="OOPS: Warning: Message to IRC server doesn't end with \\r\\n sequence, correcting to: \"";
        log+=tmp_msg;
        log+="\"";
        log_debug(GET_STR_DEBUG_INFO_EX(log));
        strcat(tmp_msg,"\r\n");
    }
    if(!strcmp(tmp_msg,"\r\n")) {
        log_debug(GET_STR_DEBUG_INFO_EX("OOPS: Warning: Message to IRC server is empty!"));
        return 0;
    }
    msg=tmp_msg;
    // END:   sanity checks

    if(wait_for_response) {
        char tmp[128];
        unsigned int i1;
        for(i1=0; i1<64 && msg[i1]!=0 && msg[i1]!=0x20; ++i1)
            tmp[i1]=msg[i1];
        tmp[i1]=0;
        if(strcmp(tmp,"PRIVMSG") && strcmp(tmp,"NOTICE") && strcmp(tmp,"PONG") && strcmp(tmp,"PASS") && strcmp(tmp,"USER")) {
            c_awaiting_response r;
            r.message=msg;
            r.time=my_time(NULL);
            waiting_for_response.push_back(r);
        }
    }

    string tmp=msg;
    if(tmp.length()>=2)
        tmp.erase(tmp.length()-2,2); // remove CR/LF
    log_irc(false,tmp);

    int ec=0;
    size_t sent=0;
    if(!irc_ssl) {
        sent=sock_send(socket,msg,strlen(msg),ec);
    } else if(irc_ssl_connection) {
        string err_str;
        if(irc_ssl_connection->get_error(err_str)) {
            string log="Closing connection to IRC server due to SSL error: ";
            log+=err_str;
            log_bot(log);
            log_ssl(log);
            log_irc_ex(E_LOG_IRC_EX_INF,log);
            MY_DELETE(irc_ssl_connection);
            irc_ssl_connection=NULL;

            log_irc_ex(E_LOG_IRC_EX_INF,log);

            if(deep_sleep)
                msgs_from_server.clear();
            irc_disconnect();
            return SOCK_SEND_ERROR;
        } else
            irc_ssl_connection->write(msg,strlen(msg));
        sent=strlen(msg);
    }
    stats_irc_bytes_sent(sent);
    if(ec) {
        char* errstr=NULL;
        string log=((string)"Socket error: ")+(errstr=sock_error(ec));
        log=GET_STR_DEBUG_INFO_EX(log);
        log_socket(ec,errstr,log,&socket);
        MY_DELETE_A(errstr);

        log_bot(log);
        log_irc_ex(E_LOG_IRC_EX_INF,log);
        if(irc_ssl_connection) {
            MY_DELETE(irc_ssl_connection);
            irc_ssl_connection=NULL;
        }
        if(deep_sleep)
            msgs_from_server.clear();
        irc_disconnect();

        return ec;
    }
    if(sent!=strlen(msg)) {
        string log=GET_STR_DEBUG_INFO_EX("Socket error: Cannot send all buffer");
        log_socket(0,"(cannot send all buffer)",log,&socket);
        log_bot(log);
        log_irc_ex(E_LOG_IRC_EX_INF,log);
        if(irc_ssl_connection) {
            MY_DELETE(irc_ssl_connection);
            irc_ssl_connection=NULL;
        }
        if(deep_sleep)
            msgs_from_server.clear();
        irc_disconnect();

        return SOCK_SEND_ERROR;
    }

    return 0;
}

/*!
    \brief Puts message for server to queue to send
    \author VooDooMan
    \version 2
    \date 2004, 2007-12-02
    \param msg_to_send Message
    \param wait_for_response Set this to true if message has to be marked as wait for response and then send next message
    \param priority CRITICAL_PRIORITY, HIGH_PRIORITY, or LOW_PRIORITY
    \return Returns false if socket I/O error occured
*/
bool irc_putserv(const c_tracked_msg_to_server& msg_to_send, bool wait_for_response, int priority)
{
    string msg=msg_to_send.get_msg();
    if(msg.length()==0)
        return true;

    if(msg.length()>=512)
        return false;

    int compress_mode_wait=stoi(core_conf_getvar("compress_mode_wait"));
    if(compress_mode_wait>0 && msg.find("MODE ",0)==0 && priority==LOW_PRIORITY) {
        string chan;
        string mode;
        string param;
        int context=1;
        for(string::size_type i1=msg.find(' ',0)+1; i1<msg.length(); ++i1) {
            if('\r'==msg[i1])
                break;
            if(context<3 && ' '==msg[i1]) {
                ++context;
                continue;
            }
            if(context==1)
                chan+=msg[i1];
            if(context==2)
                mode+=msg[i1];
            if(context==3)
                param+=msg[i1];
        }
        if(context<=3 && mode.length()==2 && (mode[0]=='+' || mode[0]=='-')) {
            // process only single mode
            int process=0;
            if(irc_isupport.get_prefix1(mode[1])!='0') {
                process=1;
            }
            if(irc_isupport.chanmodes_a.find(mode[1])!=string::npos) {
                if(mode[1]=='b') {
                    // do not precess bans, they should be put immediately
                    process=0;
                } else {
                    process=1;
                }
            }
            if(irc_isupport.chanmodes_b.find(mode[1])!=string::npos || irc_isupport.chanmodes_c.find(mode[1])!=string::npos) {
                process=2;
            }
            if(irc_isupport.chanmodes_d.find(mode[1])!=string::npos) {
                process=3;
            }
            if(process==0) {
                // do nothing!
            }
            if((process==1 || process==2) && !param.empty()) {
                // with parameter
                c_mode_queue_entry e;
                e.channel=chan;
                e.mode=mode;
                e.param=param;
                my_time(&e.submit);
                e.param_is_nick=msg_to_send.is_nick(param);
                mode_queue.push_back(e);
                return true;
            }
            if(process==3 && param.empty()) {
                // without parameter
                c_mode_queue_entry e;
                e.channel=chan;
                e.mode=mode;
                my_time(&e.submit);
                e.param_is_nick=false;
                mode_queue.push_back(e);
                return true;
            }
        }
    }

    if(priority==CRITICAL_PRIORITY) {
        irc_putserv_immediately(irc_socket,msg.c_str(),wait_for_response);
        return true;
    }

    c_msg_to_server m(msg_to_send);
    m.flood_protection=msg.find("PRIVMSG ",0)==0 || msg.find("NOTICE ",0)==0;

    m.wait_for_response=wait_for_response;
    if(wait_for_response)
        m.response[0]=0;

    if(priority==HIGH_PRIORITY)
        msgs_to_server.push_back(m);
    else if(priority==LOW_PRIORITY)
        msgs_to_server2.push_back(m);

    return true;
}

void irc_check_for_filesystem(const string& user_name, const string& channel, const string& nick);
void irc_cleanup();

#if defined(_MSC_VER)
#   pragma warning(push)
#   pragma warning(disable:4702)    // warning C4702: unreachable code
#endif

/*!
    \brief This function should be called in a loop to process send and receive to/from server
    \author VooDooMan
    \version 2
    \date 2004, 2008-02-02
    \return Returns true if socket I/O error occured
    \param deep_sleep If 'true' we are in a deep sleep, it IRC server connection error occurs, "msgs_from_server.clear();" is explicitly called
*/
MY_STATIC bool irc_process_input_from_server(bool deep_sleep = false)
{
    enum
    {
        MAX_INPUT_FROM_SERVER = 1024*20,    //!< Maximum number of bytes from server (with NULL terminator)
    };
    static char buff[MAX_INPUT_FROM_SERVER];

    bool got_error=false;

    for(;;) {
        buff[0]='\0';
        int ec=0;
        bool closed=false;
        size_t size=0;
        if(!irc_ssl) {
            size=sock_read(irc_socket,buff,MAX_INPUT_FROM_SERVER-1,ec,closed);
        } else if(irc_ssl_connection) {
            string err_str;
            if(irc_ssl_connection->get_error(err_str)) {
                string log="Closing connection to IRC server due to SSL error: ";
                log+=err_str;
                log_bot(log);
                log_ssl(log);
                log_irc_ex(E_LOG_IRC_EX_INF,log);
                if(deep_sleep)
                    msgs_from_server.clear();
                irc_disconnect();
                closed=true;
            } else
                size=irc_ssl_connection->read(buff,MAX_INPUT_FROM_SERVER-1);
        }
        if(closed) {
            //if(msgs_from_server.empty()) {
            // wait for queue to empty

            //return true;
            got_error=true;
            break;
        }

        if(ec) {
            char* errstr=NULL;
            log_socket(ec,errstr=sock_error(ec),GET_STR_DEBUG_INFO_EX("Error while reading from the socket"),&irc_socket);
            MY_DELETE_A(errstr);

            // wait for queue to empty
            //return true;
            got_error=true;
            break;
        }

        if(!size)
            break;

        buff[size]='\0';
        irc_proto_buff+=buff;

        stats_irc_bytes_received(size);

        irc_last_keepalive_detection=my_time(NULL);
        irc_keepalive_sent=false;
    }

    if(irc_proto_buff.length()>=2) {
        if('\r'==irc_proto_buff[irc_proto_buff.length()-2] && '\n'==irc_proto_buff[irc_proto_buff.length()-1]) {
            vector<string> lines;
            explode(irc_proto_buff,"\r\n",lines);
            irc_proto_buff="";
            vector<string>::iterator i1;
            for(i1=lines.begin(); i1!=lines.end(); ++i1) {
                if(!(*i1).empty())
                    msgs_from_server.push_back(*i1);
            }
        }
    }
    if(got_error) {
        if(deep_sleep)
            msgs_from_server.clear();
        return true;
    }

    return false;
}

/*!
    \brief Check if there is pending errorn on SSL conncetion to IRC server
    \author VooDooMan
    \version 1
    \date 2004
    \return Returns true on error
    \param deep_sleep If 'true' we are in a deep sleep, it IRC server connection error occurs, "msgs_from_server.clear();" is explicitly called
*/
MY_STATIC bool irc_ssl_error(bool deep_sleep = false)
{
    if(irc_ssl_connection) {
        irc_ssl_connection->process();

        string err_str;
        if(irc_ssl_connection->get_error(err_str)) {
            string log="Closing connection to IRC server due to SSL error: ";
            log+=err_str;
            log_bot(log);
            log_ssl(log);
            log_irc_ex(E_LOG_IRC_EX_INF,log);
            if(deep_sleep)
                msgs_from_server.clear();
            irc_disconnect();
            return true;
        }
    }

    return false;
}

/*!
    \brief Places exclusive locks for "irc_loop_putserv()" execution
    \author VooDooMan
    \date 2008-05-27
    \version 1
*/
class c_in_irc_loop_locker
{
    static bool m_in;               //!< Exclusity indicator
public:

    /*!
        \brief Retrieves exclusity
        \author VooDooMan
        \date 2008-05-27
        \version 1
        \retval true It is in
        \retval false Safe to execute
    */
    static bool is_in()
    {
        if(m_in)
            return true;
        m_in=true;
        return false;
    }

    /*!
        \brief C-tor
        \author VooDooMan
        \date 2008-05-27
        \version 1
    */
    c_in_irc_loop_locker()
    {
        m_in=true;
    }

    /*!
        \brief D-tor
        \author VooDooMan
        \date 2008-05-27
        \version 1
    */
    ~c_in_irc_loop_locker()
    {
        m_in=false;
    }
};

bool c_in_irc_loop_locker::m_in=false;

/*!
    \brief This function should be called in a loop to process send and receive to/from server
    \author VooDooMan
    \version 2
    \date 2004, 2008-02-02
    \return Returns true if socket I/O error occured
*/
bool irc_loop_putserv()
{
    if(c_in_irc_loop_locker::is_in())
        return false;

    c_in_irc_loop_locker lock;

    global_keeper.update_uptime();

    if(!irc_socket.cmp())
        return true;

    if(into_whois_burst && msgs_to_server.begin()==msgs_to_server.end() && msgs_to_server2.begin()==msgs_to_server2.end()) {
        into_whois_burst=false;
        if(was_whois_burst) {
            logic_on_internal_event("@end_of_init_burst@","","","","",PRIVATE_EVENT_SEVERITY_CODE_INFORMATIONAL,"","");
            log_irc_ex(E_LOG_IRC_EX_INF,"End of init burst");
        }
        was_whois_burst=false;
    }

    if(irc_ssl_error())
        return true;

    if(irc_last_keepalive_detection+IRC_SOCKET_TIMEOUT<my_time(NULL) && irc_keepalive_sent) {
        irc_last_keepalive_detection=my_time(NULL);
        irc_keepalive_sent=false;
        irc_disconnect();
        //logic_exec_from_work();
        return false;
    }
    if(irc_last_join_try+15<my_time(NULL) && msgs_to_server.empty() && msgs_to_server2.empty()) {
        my_time(&irc_last_join_try);
        vector<string>::iterator i1;
    erase_again:
        for(i1=irc_cannot_join_channels.begin(); i1!=irc_cannot_join_channels.end(); ++i1) {
            vector<c_online_channel>::iterator i;
            for(i=irc_channels.begin(); i!=irc_channels.end(); ++i)
                if(!irc_cmp_strings_case_insensitive((*i).name,*i1)) {
                    irc_cannot_join_channels.erase(i1);
                    goto erase_again;
                }
        }
        for(i1=irc_cannot_join_channels.begin(); i1!=irc_cannot_join_channels.end(); ++i1) {
            string msg="JOIN ";
            msg+=*i1;
            irc_putserv(c_tracked_msg_to_server(msg),true,HIGH_PRIORITY);
            msg="MODE ";
            msg+=*i1;
            irc_putserv(c_tracked_msg_to_server(msg),true,HIGH_PRIORITY);
            msg="TOPIC ";
            msg+=*i1;
            irc_putserv(c_tracked_msg_to_server(msg),true,HIGH_PRIORITY);
            irc_putserv(c_tracked_msg_to_server("WHOIS").add_nick(irc_nick),true,LOW_PRIORITY);
        }
    }

    int compress_mode_wait=stoi(core_conf_getvar("compress_mode_wait"));
    if(compress_mode_wait>0) {
        my_time_t now=my_time(NULL);
        list<c_mode_queue_entry> wait_elapsed;
        wait_elapsed.clear();
    again:
        list<c_mode_queue_entry>::iterator i1;
        for(i1=mode_queue.begin(); i1!=mode_queue.end(); ++i1) {
            if((*i1).submit+compress_mode_wait<now) {
                wait_elapsed.push_back(*i1);
                mode_queue.erase(i1);
                goto again;
            }
        }
        string chan="";
        list<c_mode_queue_entry> single_channel;
        single_channel.clear();
    again2:
        for(i1=wait_elapsed.begin(); i1!=wait_elapsed.end(); ++i1) {
            if(chan.empty()) {
                chan=(*i1).channel;
                single_channel.clear();
                single_channel.push_back(*i1);
                wait_elapsed.erase(i1);
                goto again2;
            }
            if(!irc_cmp_strings_case_insensitive((*i1).channel,chan)) {
                single_channel.push_back(*i1);
                wait_elapsed.erase(i1);
                goto again2;
            }
        }
        {
            list<c_mode_queue_entry>::iterator i2;
            string cumulative1; // mode specifiers
            vector<pair<c_tracked_msg_to_server::e_kind,string> > cumulative2; // params (args)
            int cum_num=0;
        again3:
            if(cum_num==irc_isupport.max_modes) {
        jump_1:
                c_msg_to_server m(c_tracked_msg_to_server("MODE").add_channel(chan).add_piece(cumulative1).add_multi(cumulative2));

                m.flood_protection=false;

                m.wait_for_response=false;
                if(m.wait_for_response)
                    m.response[0]=0;

                msgs_to_server.push_back(m);

                cumulative1.clear();
                cumulative2.clear();

                cum_num=0;
            }
            for(i2=single_channel.begin(); i2!=single_channel.end(); ++i2) {
                if(cum_num<irc_isupport.max_modes) {
                    cum_num++;
                    cumulative1+=(*i2).mode;
                    if(!(*i2).param.empty()) {
                        cumulative2.push_back(make_pair((*i2).param_is_nick ? c_tracked_msg_to_server::PIECE_NICK : c_tracked_msg_to_server::PIECE_IGNORE,(*i2).param));
                    }
                    single_channel.erase(i2);
                    goto again3;
                } else
                    break;
            }
            if(cum_num>0)
                goto jump_1;
        }
        if(!wait_elapsed.empty()) {
            chan="";
            single_channel.clear();
            goto again2;
        }
    }

    irc_remove_redundant_mode2();

    if(irc_process_input_from_server())
        return true;

    if(irc_last_keepalive_detection+IRC_KEEPALIVE_PERIOD<my_time(NULL) && !irc_keepalive_sent) {
        irc_keepalive_sent=true;
        if(!irc_sent_quit) {
            string ping="PING :";
            ping+=tos(static_cast<unsigned long>(my_time(NULL)));
            irc_putserv(c_tracked_msg_to_server(ping),true,HIGH_PRIORITY);
        }
    }

    //---
    if(last_msg.timestamp+IRC_RESPONSE_TIMEOUT<my_time(NULL))
        last_msg.wait_for_response=false;

    if(last_msg.wait_for_response && last_msg.response[0]==0)
        return false; // still waiting

    last_msg.m_msg=c_tracked_msg_to_server();
    //---

    {
        vector<c_awaiting_response>::iterator i1;
_again:
        for(i1=waiting_for_response.begin(); i1!=waiting_for_response.end(); ++i1) {
            if((*i1).time+IRC_RESPONSE_TIMEOUT<my_time(NULL)) {
                string msg="Timeout while waiting for response to command: ";
                msg+=(*i1).message;

                while(msg.length() && msg[msg.length()-1]=='\r')
                    msg.erase(msg.length()-1,1);
                while(msg.length() && msg[msg.length()-1]=='\n')
                    msg.erase(msg.length()-1,1);
                while(msg.length() && msg[msg.length()-1]=='\r')
                    msg.erase(msg.length()-1,1);
                while(msg.length() && msg[msg.length()-1]=='\n')
                    msg.erase(msg.length()-1,1);

                log_debug(msg);

                waiting_for_response.erase(i1);
                goto _again;
            }
        }

        if(waiting_for_response.begin()!=waiting_for_response.end())
            return false; // still waiting
    }

    //---

    size_t bytes=0;
    list<c_flood_history>::iterator i;
    for(i=flood_history.begin(); i!=flood_history.end(); ++i)
        bytes+=(*i).bytes;

    size_t all_1=0, all_2=0;
    size_t cnt=0;
    size_t cnt2=0;
    {
        list<c_flood_history>::iterator i2;
fl_again1:
        for(i2=flood_history.begin(); i2!=flood_history.end(); ++i2) {
            if((*i2).time+irc_bot_flood_seconds<my_time(NULL)) {
                flood_history.erase(i2);
                goto fl_again1;
            }
            all_1+=(*i2).msg.length();
            cnt++;
        }
    }
    {
        list<c_flood_history>::iterator i2;
fl_again2:
        for(i2=flood_history_other.begin(); i2!=flood_history_other.end(); ++i2) {
            if((*i2).time+irc_bot_flood_other_seconds<my_time(NULL)) {
                flood_history_other.erase(i2);
                goto fl_again2;
            }
            all_2+=(*i2).msg.length();
            cnt2++;
        }
    }

    for(;;) {
        deque<c_msg_to_server>::iterator ii;

        int queue_num=1;

        ii=msgs_to_server.begin();
        if(ii==msgs_to_server.end()) {
            ii=msgs_to_server2.begin();
            if(ii==msgs_to_server2.end())
                break;
            queue_num=2;
        }

        if(((queue_num==1?all_1:all_1+all_2)<irc_bot_flood_bytes && ((queue_num==1)?cnt<irc_bot_flood_lines:cnt2<1)) && (*ii).flood_protection) {
            bool wait_for_response=(*ii).wait_for_response;
            if(wait_for_response) {
                last_msg=*ii;
                last_msg.timestamp=my_time(NULL);
            }

            // const char* msg=(*ii).m_msg.get_msg().c_str(); // produces wrong result because of temporary instance of string class destruction
            string tmp_msg=(*ii).m_msg.get_msg();
            const char* msg=tmp_msg.c_str();

            // no flood should occur
            int error=irc_putserv_immediately(irc_socket,msg,wait_for_response);

            if(error==static_cast<int>(IRC_MSG_TOO_LONG)) {
                log_debug(GET_STR_DEBUG_INFO_EX("OOPS: IRC_MSG_TOO_LONG error #1"));
            }

            if(error==static_cast<int>(SOCK_SEND_ERROR)) {
                irc_cleanup();
                flood_history.clear();

                irc_disconnect();

                return true;
            }

            bytes+=strlen(msg);

            if((*ii).flood_protection) {
                c_flood_history entry;
                entry.bytes=strlen(msg);
                entry.msg=msg;
                my_time(&entry.time);
                flood_history.push_back(entry);
            } else {
                c_flood_history entry;
                entry.bytes=strlen(msg);
                entry.msg=msg;
                my_time(&entry.time);
                flood_history_other.push_back(entry);
            }

            if(queue_num==1)
                msgs_to_server.pop_front();
            if(queue_num==2)
                msgs_to_server2.pop_front();

            if(wait_for_response)
                break;
        } else {
            if(((queue_num==1?all_1:all_1+all_2)<irc_bot_flood_bytes && ((queue_num==1)?cnt<irc_bot_flood_lines:cnt2<1)) && !(*ii).flood_protection) {
                bool wait_for_response=(*ii).wait_for_response;
                if(wait_for_response) {
                    last_msg=*ii;
                    last_msg.timestamp=my_time(NULL);
                }

                // const char* msg=(*ii).m_msg.get_msg().c_str(); // produces wrong result because of temporary instance of string class destruction
                string tmp_msg=(*ii).m_msg.get_msg();
                const char* msg=tmp_msg.c_str();

                // no flood should occur
                int error=irc_putserv_immediately(irc_socket,msg,wait_for_response);

                if(error==static_cast<int>(IRC_MSG_TOO_LONG)) {
                    log_debug(GET_STR_DEBUG_INFO_EX("OOPS: IRC_MSG_TOO_LONG error #2"));
                }

                if(error==static_cast<int>(SOCK_SEND_ERROR)) {
                    irc_cleanup();
                    flood_history.clear();

                    irc_disconnect();

                    return true;
                }

                bytes+=strlen(msg);

                if((*ii).flood_protection) {
                    c_flood_history entry;
                    entry.bytes=strlen(msg);
                    entry.msg=msg;
                    my_time(&entry.time);
                    flood_history.push_back(entry);
                } else {
                    c_flood_history entry;
                    entry.bytes=strlen(msg);
                    entry.msg=msg;
                    my_time(&entry.time);
                    flood_history_other.push_back(entry);
                }

                if(queue_num==1)
                    msgs_to_server.pop_front();
                if(queue_num==2)
                    msgs_to_server2.pop_front();

                if(wait_for_response)
                    break;
            }
        }
        break;
    }

    enum
    {
        FILESYS_NOTIFY_PERIOD_SECS = 20,        //!< Filesystem notify period
        DEFAULT_FILESYS_NOTIFY_INTERVAL = 5*60, //!< Default filesystem notify interval (if not in "confi.txt")
    };

    // notify for files on the filesystem
    if(filesys_notify_interval==0) {
        filesys_notify_interval=sto_time(core_conf_getvar("notify_interval"))*60;
        if(filesys_notify_interval==0)
            filesys_notify_interval=DEFAULT_FILESYS_NOTIFY_INTERVAL;
    }
    if(filesys_last_notify+filesys_notify_interval<=my_time(NULL)) {
        if(filesys_last_notify==0) {
            my_time(&filesys_last_notify);
            filesys_last_notify-=filesys_notify_interval;
            filesys_last_notify+=FILESYS_NOTIFY_PERIOD_SECS;
        } else {
            my_time(&filesys_last_notify);
            vector<c_online_channel>::iterator i1;
            for(i1=irc_channels.begin(); i1!=irc_channels.end(); ++i1) {
                vector<c_online_user>::iterator i2;
                for(i2=(*i1).users.begin(); i2!=(*i1).users.end(); ++i2)
                    irc_check_for_filesystem((*i2).in_logic_as,(*i1).name,(*i2).nick);
            }
        }
    }

    return false;
}

#if defined(_MSC_VER)
#   pragma warning(pop)
#endif

/*!
    \brief Parses the message (scans for new line character) and sends it to the server
    \author VooDooMan
    \version 1
    \date 2004
    \param command Command to use (for example "PRIVMSG")
    \param nick Whom to send the message?
    \param msg Message to parse
*/
void irc_put_multiple_lines(const string& command, const string& nick, const string& msg)
{
    list<string> cmds;
    string m;
    unsigned int cnt=0;;
    for(unsigned int i1=0; i1<msg.length(); ++i1) {
        if(msg[i1]=='\\') {
            if(!(i1+1<msg.length() && msg[i1+1]=='n')) {
                m+=msg[i1];
                cnt++;
            }
        } else
            if(msg[i1]!='\n') {
                m+=msg[i1];
                cnt++;
            }
        if((cnt>256 || i1==msg.length()-1) || (i1+1<msg.length() && msg[i1]=='\\' && msg[i1+1]=='n') || msg[i1]=='\n') {
            if((i1+1<msg.length() && msg[i1]=='\\' && msg[i1+1]=='n'))
                i1++;

            if(!m.empty()) {
                string cmd=command;
                cmd+=" ";
                cmd+=nick;
                cmd+=" :";

                // check to avoid bot loops
                while(m.find("\x01",0)!=string::npos)
                    m.erase(m.find("\x01",0),1);
                if(m.length()>0 && (m.find(".",0)==0 || m.find("!",0)==0 || m.find("?",0)==0 || m.find("dcc",0)==0 || m.find("getfile",0)==0))
                    m=(string)" "+m;

                cmd+=m;

                cmds.push_back(cmd);
            }
            m="";
            cnt=0;
        }
    }

    list<string>::iterator i2;
    for(i2=cmds.begin(); i2!=cmds.end(); ++i2)
        irc_putserv(c_tracked_msg_to_server(*i2),false,LOW_PRIORITY);
}

/*!
    \brief Check filesystem for new objects and notifies user if needed
    \author VooDooMan
    \version 1
    \date 2004
    \param user_name User name from "logic.txt"
    \param nick Nick name
    \param channel Name of channel that is the user on
*/
void irc_check_for_filesystem(const string& user_name, const string& channel, const string& nick)
{
    string lang="en"; // TODO: !!!

    vector<c_dcc_notify> notifies;
    c_user u;
    vector<c_channel> chs;
    logic_partyline_get_user(user_name,u,chs);
    filesys_dcc_check_for_notifies(user_name,u,chs,channel,notifies,false,lang,"\n");
    filesys_dcc_drop_notifies(user_name,true);
    vector<c_dcc_notify>::iterator i1;
    bool unpublished=false;
    bool notified=false;
    string incomplete_files;
    for(i1=notifies.begin(); i1!=notifies.end(); ++i1) {
        if(!unpublished && (*i1).unpublished) {
            string msg=lang_get_string(1,lang,331);
            irc_put_multiple_lines("PRIVMSG",nick,msg);
            unpublished=true;
        }
        if((*i1).notify_owner) {
            string msg=(*i1).message;
            irc_put_multiple_lines("PRIVMSG",nick,msg);
        }
        if((*i1).notify_user) {
            string msg=(*i1).message;
            irc_put_multiple_lines("PRIVMSG",nick,msg);
        }
        if(!notified && ((*i1).secure_notify_owner || (*i1).secure_notify_user)) {
            string msg=lang_get_string(1,lang,332);
            irc_put_multiple_lines("PRIVMSG",nick,msg);
            notified=true;
        }
        if((*i1).incomplete) {
            if(!incomplete_files.empty())
                incomplete_files+=", ";
            incomplete_files+="\"";
            incomplete_files+=(*i1).name;
            incomplete_files+="\"";
        }
    }
    if(!incomplete_files.empty()) {
        string msg=lang_get_string(1,lang,400)+"\n";
        msg+=incomplete_files;
        irc_put_multiple_lines("PRIVMSG",nick,msg);
    }
}

/*!
    \brief Disconnects from server
    \author VooDooMan
    \version 1
    \date 2004
    \return Returns zero for no error
*/
int irc_disconnect()
{
    if(irc_was_disconnect)
        return 0;
    irc_was_disconnect=true;

    sock_send_cache();
    sock_flush_later(irc_socket);
    if(irc_ssl_connection) {
        irc_socket.clear();
        MY_DELETE(irc_ssl_connection);
    }
    irc_ssl_connection=NULL;

    if(msgs_from_server.begin()!=msgs_from_server.end())
        irc_loop_process_input();

    {
        log_irc_ex(E_LOG_IRC_EX_INF,"Disconnecting from IRC server");
        log_bot("Disconnecting from IRC server");
    }

    global_keeper.reset_users();
    global_keeper.reset_isupport();
    global_keeper.reset_user_mode_prefix();
    global_keeper.reset_channels();
    global_keeper.update_uptime();

    irc_channels.clear();

    irc_keepalive_sent=false;

    irc_ssl=false;
    irc_ssl_connection=NULL;
    if(irc_socket.cmp())
        sock_close(irc_socket,__FILE__,__LINE__);
    irc_socket.clear();
    irc_cleanup();
    mode_queue.clear();
    last_msg.wait_for_response=false;
    waiting_for_response.clear();
    irc_ident="*";
    irc_into_whois.clear();
    irc_nick_confirmed=false;

    irc_waiting_for_dns.clear();

    return 0;
}

/*!
    \brief Connects to server
    \author VooDooMan
    \version 2
    \date 2004, 2005-08-12
    \param bind_ip IPv4 address to bind to (client address)
    \param host_ DNS host of server
    \param port Port of server
    \param ssl Empty string for normal connection, or string for section in file "ssl.txt" for SSL connection
    \return Returns zero for no error, or OS's socket I/O error code
*/
int irc_connect(const string& bind_ip, string host_, unsigned short port, const string& ssl)
{
    irc_disconnect(); // for sure's sake
    irc_was_disconnect=false;

    irc_sent_quit=false;
    irc_proto_buff="";
    global_keeper.reset_users();
    global_keeper.reset_isupport();
    global_keeper.reset_user_mode_prefix();
    global_keeper.reset_channels();
    global_keeper.update_uptime();

    string host=host_;

    was_whois_burst=into_whois_burst=false;

    irc_socket_error=false;

    irc_bot_flood_bytes=sto_size_t(core_conf_getvar("irc_bot_flood_bytes"));
    irc_bot_flood_seconds=sto_time(core_conf_getvar("irc_bot_flood_seconds"));
    irc_bot_flood_lines=stoui(core_conf_getvar("irc_bot_flood_lines"));
    irc_bot_flood_other_seconds=sto_time(core_conf_getvar("irc_bot_flood_other_seconds"));

    if(!irc_bot_flood_bytes)
        irc_bot_flood_bytes=64;
    if(!irc_bot_flood_seconds)
        irc_bot_flood_seconds=5;
    if(!irc_bot_flood_lines)
        irc_bot_flood_lines=2;
    if(!irc_bot_flood_other_seconds)
        irc_bot_flood_other_seconds=5;

    // do NOT re-set this anymore! - Fixed bug in 1.0.12
    //irc_follow_redirs=false;

    irc_recommended_server="";
    irc_recommended_port=0;

    irc_socket.clear();
    irc_last_keepalive_detection=0;
    irc_keepalive_sent=false;

    irc_nick="";
    irc_fullname="";
    irc_unique_id="";
    irc_host="";

    irc_nicks.clear();
    filesys_notify_interval=0;
    filesys_last_notify=0;

    irc_key_chans.clear();
    irc_channels.clear();

    tmp_users.clear();

    gone_users.clear();

    unknown_users_on_channels.clear();

    last_msg.wait_for_response=false;
    waiting_for_response.clear();

    irc_cleanup();

    mode_queue.clear();

    flood_history.clear();
    flood_history_other.clear();

    irc_cannot_join_channels.clear();
    irc_last_join_try=0;

    last_msg.m_msg=c_tracked_msg_to_server();
    last_msg.wait_for_response=false;
    last_msg.response[0]=0;

    irc_channels.clear();

    dcc_request_whois.clear();
    modes_from_server.clear();

    irc_isupport.clear();

    irc_into_whois.clear();

    string comment="IRC client connecting to ";

    if(!ssl.empty()) {
        comment="IRC SSL client (profile=\"";
        comment+=ssl;
        comment+="\") connecting to ";
    }

    {
        string msg="Connecting ";
        if(!ssl.empty()) {
            msg+="(SSL profile=\"";
            msg+=ssl;
            msg+="\") ";
        }
        msg+="to IRC server ";
        if(host.find(":",0)!=string::npos) {
            msg+="[";
            comment+="[";
        }
        msg+=host;
        comment+=host;
        if(host.find(":",0)!=string::npos) {
            msg+="]";
            comment+="]";
        }
        msg+=":";
        comment+=":";
        msg+=tos(port);
        comment+=tos(port);
        msg+="...";
        log_irc_ex(E_LOG_IRC_EX_INF,msg);
        log_bot(msg);
    }

    int ec=0;
    bool async=ssl.empty();
    async=true;

    irc_socket=sock_connect((char*)bind_ip.c_str(),(char*)host.c_str(),port,ec,async,comment.c_str());
    irc_last_keepalive_detection=my_time(NULL);
    if(ec) {
        char* errstr=NULL;
        log_socket(ec,errstr=sock_error(ec),GET_STR_DEBUG_INFO_EX("Error while calling sock_connect()"),&irc_socket);

        string log="Error connecting to IRC server: ";
        log+=errstr;
        log_irc_ex(E_LOG_IRC_EX_INF,GET_STR_DEBUG_INFO_EX(log));

        MY_DELETE_A(errstr);

        irc_disconnect();

        return ec;
    }

    stats_irc_new_connection(host,port);

    irc_ssl=!ssl.empty();

    if(irc_ssl) {
        if(conf_ssl_init()) {
            string log=GET_STR_DEBUG_INFO_EX("OOPS: Error while reading SSL configuration");
            log_bot(log);
            log_debug(log);

            log=GET_STR_DEBUG_INFO_EX("Error connecting to IRC server: Error while reading SSL configuration");
            log_irc_ex(E_LOG_IRC_EX_INF,log);

            irc_disconnect();

            return MY_SSL_ERROR;
        }
        c_ssl_bot irc_ssl_profile;
        irc_ssl_profile.botname=ssl;
        if(!ssl_conf.get_ssl_bot(irc_ssl_profile)) {
            string log="OOPS: Cannot find SSL profile \"";
            log+=ssl;
            log+="\" (section in file \"ssl.txt\") for connection to IRC server ";

            if(host.find(":",0)!=string::npos)
                log+="[";
            log+=host;
            if(host.find(":",0)!=string::npos)
                log+="]";
            log+=":";
            log+=tos(port);

            log=GET_STR_DEBUG_INFO_EX(log);

            log_bot(log);
            log_debug(log);
            log_ssl(log);
            irc_disconnect();

            log="Error connecting to IRC server: Cannot find SSL profile";
            log_irc_ex(E_LOG_IRC_EX_INF,log);

            return MY_SSL_ERROR;
        }
        string err_str;
        if(irc_ssl_connection)
            MY_DELETE(irc_ssl_connection);
        irc_ssl_connection=MY_NEW c_ssl(irc_socket,NULL,irc_ssl_profile.cert,irc_ssl_profile.key,irc_ssl_profile.m_only_get_peers_cert,irc_ssl_profile.m_disable_ca_verify,irc_ssl_profile.m_ignore_expired_peers_cert,irc_ssl_profile.m_server_cert);
        sock_set_nonblocking(irc_socket,true);
    }

    return 0;
}

/*!
    \brief Parses and evaluates RPL_ISUPPORT reply (005 numerics)
    \author VooDooMan
    \version 1
    \date 2004
    \param str Line of message from server
*/
void irc_RPL_ISUPPORT(string str)
{
    // BotNick RFC2812 PREFIX=(ov)@+ CHANTYPES=#&!+ MODES=3 CHANLIMIT=#&!+:31 NICKLEN=9 TOPICLEN=160 KICKLEN=160 MAXLIST=beIR:42 CHANNELLEN=50 IDCHAN=!:5 CHANMODES=beIR,k,l,imnpstaqr :are supported by this server

    unsigned int i1;

    for(i1=0; i1<str.length(); ++i1)
        if(str[i1]==0x20)
            break;
    str.erase(0,i1+1); // erase my nick

    global_keeper.reset_isupport();

    string feature1, feature2;
    while(str.length()>0) {
        feature1="";
        feature2="";
        int context=0;
        while(str.length()>0) {
            if(str[0]!=0x20) {
                if(str[0]=='=' && context==0) {
                    context++;
                    str.erase(0,1);
                    continue;
                }
                if(context==0)
                    feature1+=str[0];
                else
                    feature2+=str[0];
            } else {
                str.erase(0,1);
                break;
            }
            str.erase(0,1);
        }
        if(feature1.length()>0 && feature1[0]==':') {
            // end of list - comment found
            break;
        }
        pair<string,string> p;
        p.first=feature1;
        p.second=feature2;
        irc_isupport.raw.insert(p);
        global_keeper.add_isupport(feature1,feature2);
        if(!feature1.compare("CHANMODES")) {
            int context=0;
            string str=feature2;
            while(str.length()>0) {
                if(str[0]==',') {
                    context++;
                    str.erase(0,1);
                    continue;
                }
                if(context==0)
                    irc_isupport.chanmodes_a+=str[0];
                if(context==1)
                    irc_isupport.chanmodes_b+=str[0];
                if(context==2)
                    irc_isupport.chanmodes_c+=str[0];
                if(context==3)
                    irc_isupport.chanmodes_d+=str[0];
                str.erase(0,1);
            }
        }
        if(!feature1.compare("PREFIX")) {
            int context=0;
            string str=feature2;
            string first, second;
            while(str.length()>0) {
                if(str[0]=='(') {
                    str.erase(0,1);
                    continue;
                }
                if(str[0]==')') {
                    context++;
                    str.erase(0,1);
                    continue;
                }
                if(context==0)
                    first+=str[0];
                if(context==1)
                    second+=str[0];
                str.erase(0,1);
            }
            for(unsigned int i1=0; i1<first.length() && i1<second.length(); ++i1) {
                pair<char,char> p;
                p.first=first[i1];
                p.second=second[i1];
                irc_isupport.prefix.insert(p);
                global_keeper.add_user_mode_prefix(first[i1],second[i1]);
            }
        }
        if(!feature1.compare("MODES")) {
            long modes=stol(feature2);
            if(modes==0)
                modes=3;
            irc_isupport.max_modes=modes;
        }
        if(!feature1.compare("CASEMAPPING")) {
            if(!cmp_strings_case_insensitive(feature2,"ascii"))
                irc_casemapping=CASE_ASCII;
            else if(!cmp_strings_case_insensitive(feature2,"rfc1459"))
                irc_casemapping=CASE_RFC1459;
            else
                irc_casemapping=CASE_ASCII;
        }
    }
}

/*!
    \brief Parses and evaluates RPL_NAMREPLY reply (reply to NAMES command)
    \author VooDooMan
    \version 3
    \date 2004, 2005-07-13, 2005-10-21
    \param str Line of message from server
*/
void irc_RPL_NAMREPLY(string str)
{
    unsigned int i1;

    for(i1=0; i1<str.length(); ++i1)
        if(str[i1]==0x20)
            break;
    str.erase(0,i1+1); // erase numeric reply

    for(i1=0; i1<str.length(); ++i1)
        if(str[i1]==0x20)
            break;
    str.erase(0,i1+1); // erase my nick

    for(i1=0; i1<str.length(); ++i1)
        if(str[i1]==0x20)
            break;
    str.erase(0,i1+1); // erase '=' / '*' / '@'

    string channame;
    for(i1=0; i1<str.length(); ++i1)
        if(str[i1]==0x20)
            break;
            else
            channame+=str[i1];
    str.erase(0,i1+1); // erase channel name

    for(i1=0; i1<str.length(); ++i1)
        if(str[i1]==':')
            break;
    str.erase(0,i1+1); // erase everything before ':'

l3:
    vector<c_online_channel>::iterator i;
    bool got_channel=false;
    for(i=irc_channels.begin(); i!=irc_channels.end(); ++i) {
        if(!irc_cmp_strings_case_insensitive((*i).name,channame)) {
            got_channel=true;
            break;
        }
    }
    if(!got_channel) {
        // we are new on channel, right after JOIN
        c_online_channel chan(channame);
        irc_channels.push_back(chan);
        goto l3;
    }

    while(str.length()>0) {
        while(str.length()>0 && str[0]==0x20)
            str.erase(0,1);

        string nick;
        for(i1=0; i1<str.length(); ++i1)
            if(str[i1]==0x20)
                break;
                else
                nick+=str[i1];
        str.erase(0,i1); // erase nick

        while(str.length()>0 && str[0]==0x20)
            str.erase(0,1);
        while(nick.length()>0 && nick[0]==0x20)
            str.erase(0,1);

        char prefix='0';
        if(irc_isupport.get_prefix2(nick[0])!='0') {
            prefix=nick[0];
            nick.erase(0,1);
        }

        bool got_user=false;
        if(nick.length()>0) {
    l4:
            vector<c_online_user>::iterator ii;
            for(ii=(*i).users.begin(); ii!=(*i).users.end(); ++ii) {
                if(!irc_cmp_strings_case_insensitive((*ii).nick,nick)) {
                    got_user=true;
                    break;
                }
            }
            if(!got_user) {
                c_online_user user;
                user.nick=nick;
                user.got_whois=false;
                user.whois_sent=false;
                user.on_join_called=false;
                user.ident="";
                user.host="";
                user.fullname="";
                user.got_mode=true;
                if(prefix!='0')
                    user.mode=prefix;
                else
                    user.mode="0";
                user.irc_op=false;
                user.just_joined=false;
                user.old_mode="INVALID";
                user.on_join_passive=true;
                user.on_join_called2=false;
                (*i).users.push_back(user);
                global_keeper.add_user(channame,nick,(string)""+prefix);
                goto l4;
            }
            if(!(*ii).got_whois) {
                (*ii).whois_sent=true;
            } else {
                if(!nick.compare(irc_nick) && prefix!='0') {
                    string source_user=logic_find_user(nick,irc_get_ident(nick),irc_get_host(nick),irc_get_fullname(nick),irc_is_ircop(nick));
                    string target_user=source_user;
                    global_keeper.add_user(channame,nick,(string)""+prefix);
                    logic_validate(channame,source_user,target_user,nick,nick,'+',prefix);
                }
            }
        }
    }

    i1=0;
    vector<c_online_user>::iterator i2;
    c_online_channel* chan;
    if(irc_get_online_channel(channame,chan)) {
        i2=chan->users.begin();
        for(; i2!=chan->users.end(); ++i2)
            i1++;
        if(i1==1) {
            irc_putserv(c_tracked_msg_to_server("WHOIS").add_nick(irc_nick),true,LOW_PRIORITY);
        }

        if(!chan->b_invites_sent && irc_isupport.chanmodes_a.find("I")!=string::npos) {
            chan->b_invites_sent=true;
            string inv="MODE ";
            inv+=channame;
            inv+=" +I";
            irc_putserv(c_tracked_msg_to_server(inv),true,HIGH_PRIORITY);
        }

        if(!chan->b_reops_sent && irc_isupport.chanmodes_a.find("R")!=string::npos) {
            chan->b_reops_sent=true;
            string reop="MODE ";
            reop+=channame;
            reop+=" +R";
            irc_putserv(c_tracked_msg_to_server(reop),true,HIGH_PRIORITY);
        }
    }
}

/*!
    \brief Checks wheter user is found on channel
    \author VooDooMan
    \version 1
    \date 2004
    \param nick Nick name
    \param channel Channel name
    \retval true If user is on channel
    \retval false If user isn't on channel
*/
bool irc_check_nick_on_channel(const string& nick, const string& channel)
{
    vector<c_online_channel>::iterator i1;
    for(i1=irc_channels.begin(); i1!=irc_channels.end(); ++i1) {
        if(!irc_cmp_strings_case_insensitive((*i1).name,channel)) {
            vector<c_online_user>::iterator i2;
            for(i2=(*i1).users.begin(); i2!=(*i1).users.end(); ++i2) {
                if(!irc_cmp_strings_case_insensitive((*i2).nick,nick))
                    return true;
            }
        }
    }
    return false;
}

/*!
    \brief Parses and evaluates RPL_WHOISUSER reply (partial reply to WHOIS command)
    \author VooDooMan
    \version 1
    \date 2004
    \param str Line of message from server
    \param full_line Full line from server (not truncated)
*/
void irc_RPL_WHOISUSER(string str, const string& full_line)
{
    BEGIN_DEBUG_TIME_MEASUREMENT(true,1,"processing whois");

    size_t i1;

    // erase numeric reply
    if((i1=str.find(0x20))!=string::npos)
        str.erase(0,i1+1);

    string my_nick;
    for(i1=0; i1<str.length(); ++i1)
        if(str[i1]==0x20)
            break;
        else
            my_nick+=str[i1];
    irc_nick=my_nick;
    str.erase(0,i1+1); // erase my nick

    string nick;
    for(i1=0; i1<str.length(); ++i1)
        if(str[i1]==0x20)
            break;
        else
            nick+=str[i1];
    str.erase(0,i1+1); // erase nick name

    string ident;
    for(i1=0; i1<str.length(); ++i1)
        if(str[i1]==0x20)
            break;
        else
            ident+=str[i1];
    str.erase(0,i1+1); // erase ident

    string host;
    for(i1=0; i1<str.length(); ++i1)
        if(str[i1]==0x20)
            break;
        else
            host+=str[i1];
    str.erase(0,i1+1); // erase host

    {
    again_1:
        multimap<string,string>::iterator i1;
        for(i1=irc_into_whois.begin(); i1!=irc_into_whois.end(); ++i1)
            if(!irc_cmp_strings_case_insensitive((*i1).first,nick)) {
                irc_into_whois.erase(i1);
                goto again_1;
            }
        irc_into_whois.insert(make_pair(nick,full_line));
    }

    {
        in_addr addr4;
        threads_resolve4(host.c_str(),addr4,RT_RESOLVE_SCHEDULE);
        in_addr6_ addr6;
        threads_resolve6(host.c_str(),addr6,RT_RESOLVE_SCHEDULE);
        if(core_conf_getvar("try_reverse_lookup").compare("0")) {
            char tmp[2048];
            if(my_inet_pton(AF_INET,host.c_str(),&addr4)==1)
                threads_reverse4(host.c_str(),tmp,sizeof(tmp)-1,RT_RESOLVE_SCHEDULE);
            if(my_inet_pton(AF_INET6,host.c_str(),&addr6)==1)
                threads_reverse6(addr6,tmp,sizeof(tmp)-1,RT_RESOLVE_SCHEDULE);
        }
    }
    if(!irc_nick.compare(nick))
        irc_host=host;

    // erase everything before ':'
    if((i1=str.find(':'))!=string::npos)
        str.erase(0,i1+1);

    string fullname=str;

    irc_waiting_for_dns.push_back(c_dns_wait(nick,ident,host,fullname));
    {
        list<c_dns_wait>::iterator dns_i;
        for(dns_i=irc_waiting_for_dns.begin(); dns_i!=irc_waiting_for_dns.end(); ++dns_i) {
            if(!cmp_strings_case_insensitive((*dns_i).m_nick,nick))
                (*dns_i).m_fullname=fullname;
        }
    }

    vector<c_online_channel>::iterator i;
    for(i=irc_channels.begin(); i!=irc_channels.end(); ++i) {
        vector<c_online_user>::iterator ii;
        for(ii=(*i).users.begin(); ii!=(*i).users.end(); ++ii) {
            if(!irc_cmp_strings_case_insensitive((*ii).nick,nick)) {

                (*ii).got_whois=true;

                BEGIN_DEBUG_TIME_MEASUREMENT(true,2,"#2");
                global_keeper.add_user_whois1((*ii).nick,(*ii).ident,(*ii).host,(*ii).fullname);
                END_DEBUG_TIME_MEASUREMENT(2,0);

                break;
            }
        }
    }

    if(!irc_cmp_strings_case_insensitive(nick,irc_nick)) {
        // after re-JOIN to get op
        vector<c_online_channel>::iterator i1;
        for(i1=irc_channels.begin(); i1!=irc_channels.end(); ++i1) {
            vector<c_online_user>::iterator i2;
            for(i2=(*i1).users.begin(); i2!=(*i1).users.end(); ++i2) {
                if(!(*i2).nick.compare(irc_nick)) {
                    if(!(*i2).got_whois)
                        break;
                    if((*i2).nick_validated)
                        break;

                    BEGIN_DEBUG_TIME_MEASUREMENT(true,8,"#8");

                    (*i2).nick_validated=true;
                    logic_on_nick_validate(irc_nick,irc_get_ident(irc_nick),irc_get_host(irc_nick),irc_get_fullname(irc_nick),(*i1).name,irc_is_ircop(irc_nick),true,(*i2).mod_disp_join_called);
                    (*i2).on_join_called=true; // !!!

                    END_DEBUG_TIME_MEASUREMENT(8,0);

                    break;
                }
            }
        }
    }

    END_DEBUG_TIME_MEASUREMENT(1,0);
}

/*!
    \brief Parses and evaluates RPL_YOURID reply (your unique ID)
    \author VooDooMan
    \version 1
    \date 2004
    \param str Line of message from server
*/
void irc_RPL_YOURID(string str)
{
    unsigned int i1;

    for(i1=0; i1<str.length(); ++i1)
        if(str[i1]==0x20)
            break;
    str.erase(0,i1+1); // erase server name

    for(i1=0; i1<str.length(); ++i1)
        if(str[i1]==0x20)
            break;
    str.erase(0,i1+1); // erase numeric reply

    for(i1=0; i1<str.length(); ++i1)
        if(str[i1]==0x20)
            break;
    str.erase(0,i1+1); // erase my nick

    string id;
    for(i1=0; i1<str.length(); ++i1)
        if(str[i1]==0x20)
            break;
            else
            id+=str[i1];
    str.erase(0,i1+1); // erase ID

    irc_unique_id=id;
}

/*!
    \brief Parses and evaluates RPL_WHOISOPERATOR reply (partial reply to WHOIS command)
    \author VooDooMan
    \version 1
    \date 2004
    \param str Line of message from server
*/
void irc_RPL_WHOISOPERATOR(string str)
{
    unsigned int i1;

    for(i1=0; i1<str.length(); ++i1)
        if(str[i1]==0x20)
            break;
    str.erase(0,i1+1); // erase numeric reply

    for(i1=0; i1<str.length(); ++i1)
        if(str[i1]==0x20)
            break;
    str.erase(0,i1+1); // erase my nick

    string nick;
    for(i1=0; i1<str.length(); ++i1)
        if(str[i1]==0x20)
            break;
            else
            nick+=str[i1];
    str.erase(0,i1+1); // erase nick name

    for(i1=0; i1<str.length(); ++i1)
        if(str[i1]==':')
            break;
    str.erase(0,i1+1); // erase everything before ':'

    string chan;

    vector<c_online_channel>::iterator i;
    for(i=irc_channels.begin(); i!=irc_channels.end(); ++i) {
        vector<c_online_user>::iterator ii;
        for(ii=(*i).users.begin(); ii!=(*i).users.end(); ++ii) {
            if(!irc_cmp_strings_case_insensitive((*ii).nick,nick)) {
                if(!(*ii).irc_op) {
                    string host_mask=nick;
                    host_mask+="!";
                    host_mask+=(*ii).ident;
                    host_mask+="@";
                    host_mask+=(*ii).host;
                    log_channel((*i).name,TYPE_IRCOP_JOIN,host_mask,"","",irc_get_modes_for_log((*i).name,nick));
                }
                (*ii).irc_op=true;
                chan=(*i).name;
            }
        }
    }

    vector<c_online_user>::iterator ii;
    for(ii=tmp_users.begin(); ii!=tmp_users.end(); ++ii) {
        if(!irc_cmp_strings_case_insensitive((*ii).nick,nick)) {
            (*ii).irc_op=true;
        }
    }

    global_keeper.add_user_ircop(nick);

    logic_on_ircop(chan,nick,true);

    vector<c_dcc_request_whois>::iterator i2;
    for(i2=dcc_request_whois.begin(); i2!=dcc_request_whois.end(); ++i2) {
        if(!irc_cmp_strings_case_insensitive((*i2).nick,nick)) {
            (*i2).irc_op=true;
            break;
        }
    }
}

/*
    \brief Parses and evaluates RPL_WHOISCHANNELS reply (partial reply to WHOIS command)
    \author VooDooMan
    \version 1
    \date 2004
    \param str Line of message from server
*/
void irc_RPL_WHOISCHANNELS(string str)
{
    unsigned int i1;

    for(i1=0; i1<str.length(); ++i1)
        if(str[i1]==0x20)
            break;
    str.erase(0,i1+1); // erase numeric reply

    for(i1=0; i1<str.length(); ++i1)
        if(str[i1]==0x20)
            break;
    str.erase(0,i1+1); // erase my nick

    string nick;
    for(i1=0; i1<str.length(); ++i1)
        if(str[i1]==0x20)
            break;
            else
            nick+=str[i1];
    str.erase(0,i1+1); // erase nick name

    for(i1=0; i1<str.length(); ++i1)
        if(str[i1]==':')
            break;
    str.erase(0,i1+1); // erase everything before ':'

    while(str.length()>0) {
        string chan;
        for(i1=0; i1<str.length(); ++i1)
            if(str[i1]==0x20)
                break;
                else
                chan+=str[i1];
        str.erase(0,i1+1); // erase [@]#channame

        if(chan.length()==0)
            continue;

        char prefix='0';
        if(irc_isupport.get_prefix2(chan[0])!='0')
            prefix=chan[0];
        if(prefix=='+') {
            string chantypes;
            map<string,string>::iterator iii;
            for(iii=irc_isupport.raw.begin(); iii!=irc_isupport.raw.end(); ++iii)
                if(!(*iii).first.compare("CHANTYPES")) {
                    chantypes=(*iii).second;
                    break;
                }
            if(chan.length()>1 && chantypes.find(chan[1])==string::npos)
                // we got +channel = channel name is "+channel" of type '+'
                prefix='0';
            if(chan.length()>1 && chantypes.find(chan[1])!=string::npos)
                // we got +?channel = channel name is "#channel" or "!channel",..., but with mode of '+' - voice
                prefix='+';
            if(chan.length()==1)
                // we got "+" = channel name is "+" of type '+'
                prefix='0';
        }
        if(prefix!='0')
            chan.erase((unsigned int)0,(unsigned int)1);

        vector<c_online_channel>::iterator i;
        for(i=irc_channels.begin(); i!=irc_channels.end(); ++i) {
            vector<c_online_user>::iterator ii;
            for(ii=(*i).users.begin(); ii!=(*i).users.end(); ++ii) {
                if(!irc_cmp_strings_case_insensitive((*ii).nick,nick) && !irc_cmp_strings_case_insensitive((*i).name,chan)) {
                    (*ii).got_mode=true;
                    if(prefix!='0')
                        (*ii).mode=prefix;
                    else
                        (*ii).mode="0";

                    global_keeper.add_user_whois2((*ii).nick,chan,(*ii).mode);

                    if((*ii).got_whois && !(*ii).on_join_called) {
                        logic_on_join((*ii).nick, (*ii).ident, (*ii).host, (*ii).fullname, chan, (*ii).irc_op);
                        (*ii).mod_disp_join_called=true;

                        // this code was very slow... however, we got this check delayed on schedule
                        /*
                        irc_check_for_filesystem((*ii).in_logic_as,(*i).name,(*ii).nick);
                        */
                    }
                    if(!(*ii).nick.compare(irc_nick) && (*ii).mode.compare((*ii).old_mode)) {
                        // there it will go on channel cycle
                        if((*ii).mode.length() && (*ii).mode.compare("0"))
                            logic_validate((*i).name,"",(*ii).in_logic_as,"",(*ii).nick,'+',(*ii).mode[0]);
                        (*ii).old_mode=(*ii).mode;
                    }
                    (*ii).on_join_called=true;
                }
            }
        }
    }
}

/*!
    \brief Finds out if user has access to partyline; adds entry to tmp_users and performs WHOIS command if user is unknown yet
    \author VooDooMan
    \version 1
    \date 2004
    \param hostmask Host mask of user ("nick!ident@host")
    \param user_name_as_in_logic Returns user's name as defined in logic.txt
    \param msg Sets this string to dcc_msg member to tmp_users entry if user is not known yet
    \return Returns true if user has access to partyline
*/
bool irc_access_to_partyline(const char* hostmask, string& user_name_as_in_logic, const string& msg)
{
    string str=hostmask;

    string nick;
    unsigned int i1;
    for(i1=0; i1<str.length(); ++i1)
        if(str[i1]=='!')
            break;
            else
            nick+=str[i1];
    str.erase(0,i1+1); // erase nick name

    string ident;
    for(i1=0; i1<str.length(); ++i1)
        if(str[i1]=='@')
            break;
            else
            ident+=str[i1];
    str.erase(0,i1+1); // erase ident

    string host;
    for(i1=0; i1<str.length(); ++i1)
        if(str[i1]==0x20)
            break;
            else
            host+=str[i1];
    str.erase(0,i1+1); // erase host

    vector<c_online_channel>::iterator i;
    for(i=irc_channels.begin(); i!=irc_channels.end(); ++i) {
        vector<c_online_user>::iterator ii;
        for(ii=(*i).users.begin(); ii!=(*i).users.end(); ++ii) {
            if(!irc_cmp_strings_case_insensitive((*ii).nick,nick)) {
                return logic_access_to_partyline((*ii).nick,(*ii).ident,(*ii).host,(*ii).fullname,(*ii).irc_op,user_name_as_in_logic);
            }
        }
    }

    vector<c_online_user>::iterator ii;
    for(ii=tmp_users.begin(); ii!=tmp_users.end(); ++ii) {
        if(!irc_cmp_strings_case_insensitive((*ii).nick,nick)) {
            return logic_access_to_partyline((*ii).nick,(*ii).ident,(*ii).host,(*ii).fullname,(*ii).irc_op,user_name_as_in_logic);
        }
    }

    c_online_user u;
    u.nick=nick;
    u.got_whois=false;
    u.whois_sent=true;
    u.on_join_called=false;
    u.ident=ident;
    u.host=host;
    u.fullname="";
    u.got_mode=false;
    u.mode="0";
    u.irc_op=false;
    u.in_logic_as="";
    my_time(&u.last_whois);
    u.dcc_msg=msg;
    u.just_joined=false;
    u.old_mode="INVALID";
    u.on_join_passive=true;
    u.on_join_called2=false;
    tmp_users.push_back(u);

    irc_putserv(c_tracked_msg_to_server("WHOIS").add_nick(nick),true,LOW_PRIORITY);

    return false;
}

/*!
    \brief This must be called as RPL_ENDOFWHOIS reply from server - the end of WHOIS list for user
    \author VooDooMan
    \version 1
    \date 2004
    \param str Message line from server
*/
void irc_RPL_ENDOFWHOIS(string str)
{
    unsigned int i1;

    for(i1=0; i1<str.length(); ++i1)
        if(str[i1]==0x20)
            break;
    str.erase(0,i1+1); // erase numeric reply

    for(i1=0; i1<str.length(); ++i1)
        if(str[i1]==0x20)
            break;
    str.erase(0,i1+1); // erase my nick

    string nick;
    for(i1=0; i1<str.length(); ++i1)
        if(str[i1]==0x20)
            break;
            else
            nick+=str[i1];
    str.erase(0,i1+1); // erase nick name

    {
        vector<string> lines;
    again_1:
        multimap<string,string>::iterator i1;
        for(i1=irc_into_whois.begin(); i1!=irc_into_whois.end(); ++i1)
            if(!irc_cmp_strings_case_insensitive((*i1).first,nick)) {
                lines.push_back((*i1).second);
                irc_into_whois.erase(i1);
                goto again_1;
            }
        global_keeper.add_user_whois3(nick,lines);
    }

    mod_dispatch("endofwhois@nick",irc_find_user(nick).in_logic_as,"","",false,0,"user",irc_find_user(nick).in_logic_as.c_str(),"nick",nick.c_str(),NULL,"nick",nick.c_str(),"origin","@irc@","ident",irc_get_ident(nick).c_str(),"host",irc_get_host(nick).c_str(),NULL);
}

/*!
    \brief This is called on bot's JOIN, if the bot finds out that some user in logic.txt is banned; it forces on_banned event procedure to be called
    \author VooDooMan
    \version 1
    \date 2004
    \param str Partial message from server containing numeric reply, channel name and mask
    \warning Must be called after irc_excepted() was called to ensure exceptions (+e) to bans were evaluated
*/
void irc_banned(string str)
{
    unsigned int i1;

    for(i1=0; i1<str.length(); ++i1)
        if(str[i1]==0x20)
            break;
    str.erase(0,i1+1); // erase everything before SPACE

    for(i1=0; i1<str.length(); ++i1)
        if(str[i1]==0x20)
            break;
    str.erase(0,i1+1); // erase numeric reply

    for(i1=0; i1<str.length(); ++i1)
        if(str[i1]==0x20)
            break;
    str.erase(0,i1+1); // erase my nick

    string channel;
    for(i1=0; i1<str.length(); ++i1)
        if(str[i1]==0x20)
            break;
            else
            channel+=str[i1];
    str.erase(0,i1+1); // erase channel_name

    string mask;
    for(i1=0; i1<str.length(); ++i1)
        if(str[i1]==0x20)
            break;
            else
            mask+=str[i1];

    global_keeper.add_banned(channel,mask);

    vector<c_online_channel>::iterator i;
    for(i=irc_channels.begin(); i!=irc_channels.end(); ++i) {
        if(!irc_cmp_strings_case_insensitive((*i).name,channel)) {
            (*i).bans.push_back(mask);
            logic_on_banned(channel,mask,(*i).excepts,(*i).invites,(*i).bans);
            break;
        }
    }
}


/*!
    \brief Stores informations about user to invite (+I)
    \author VooDooMan
    \version 1
    \date 2004
*/
class c_user_to_invite
{
public:
    string mask;                    //!< Host mask
    string user_name;                //!< Name of user from "logic.txt"

    c_user_to_invite()
    {
        mask="";
        user_name="";
    }
};

/*!
    \brief This is called on bot's JOIN, checks wheter some user in logic.txt is not invited; it forces on_not_invited event procedure to be called
    \author VooDooMan
    \version 1
    \date 2004
    \warning Must be called after irc_invited() was called to ensure invitations (+I) to override +i mode are in vector
*/
MY_STATIC void irc_check_for_not_invited(const string& channel)
{
    vector<c_online_channel>::iterator i;
    for(i=irc_channels.begin(); i!=irc_channels.end(); ++i) {
        if(!irc_cmp_strings_case_insensitive((*i).name,channel)) {
            vector<c_user_to_invite> invites;
            vector<c_channel>::iterator i1;
            map<string,string> vars;
            for(i1=r_channels.begin(); i1!=r_channels.end(); ++i1) {
                if(!irc_cmp_strings_case_insensitive((*i1).channel_name,channel)) {
                    if(!(*i1).host_unknown) {
                        c_user user;
                        if(logic_botnet_get_user((*i1).username,user)) {
                            vector<string>::iterator i5;
                            for(i5=user.hostmask.begin(); i5!=user.hostmask.end(); ++i5) {
                                c_user_to_invite u;
                                u.mask=*i5;
                                if(u.mask.length()>0 && u.mask[0]=='"')
                                    u.mask.erase(0,1);
                                if(u.mask.length()>0 && u.mask[u.mask.length()-1]=='"')
                                    u.mask.erase(u.mask.length()-1,1);
                                u.user_name=user.name;
                                invites.push_back(u);
                            }
                        }
                    }
                }
            }
            vector<c_user_to_invite>::iterator i4;
            for(i4=invites.begin(); i4!=invites.end(); ++i4) {
                bool invited=false;
                vector<string>::iterator i3;
                for(i3=(*i).invites.begin(); i3!=(*i).invites.end(); ++i3) {
                    if(!match((*i3).c_str(),(*i4).mask.c_str())) {
                        invited=true;
                        break;
                    }
                }
                if(!invited) {
                    logic_on_not_invited(channel,(*i4).mask,(*i4).user_name);
                }
            }
        }
    }
}

/*!
    \brief This is called on bot's JOIN, checks wheter some user in logic.txt is not in +R list; it forces on_not_in_reopped procedure to be called
    \author VooDooMan
    \version 1
    \date 2004
*/
MY_STATIC void irc_check_for_not_reopped(const string& channel)
{
    vector<c_online_channel>::iterator i;
    for(i=irc_channels.begin(); i!=irc_channels.end(); ++i) {
        if(!irc_cmp_strings_case_insensitive((*i).name,channel)) {
            vector<c_user_to_invite> reops;
            vector<c_channel>::iterator i1;
            for(i1=r_channels.begin(); i1!=r_channels.end(); ++i1) {
                if(!irc_cmp_strings_case_insensitive((*i1).channel_name,channel)) {
                    if(!(*i1).host_unknown) {
                        c_user user;
                        if(logic_botnet_get_user((*i1).username,user)) {
                            vector<string>::iterator i5;
                            for(i5=user.hostmask.begin(); i5!=user.hostmask.end(); ++i5) {
                                c_user_to_invite u;
                                u.mask=*i5;
                                if(u.mask.length()>0 && u.mask[0]=='"')
                                    u.mask.erase(0,1);
                                if(u.mask.length()>0 && u.mask[u.mask.length()-1]=='"')
                                    u.mask.erase(u.mask.length()-1,1);
                                u.user_name=user.name;
                                reops.push_back(u);
                            }
                        }
                    }
                }
            }
            vector<c_user_to_invite>::iterator i4;
            for(i4=reops.begin(); i4!=reops.end(); ++i4) {
                bool reopped=false;
                vector<string>::iterator i3;
                for(i3=(*i).reops.begin(); i3!=(*i).reops.end(); ++i3) {
                    if(!match((*i3).c_str(),(*i4).mask.c_str())) {
                        reopped=true;
                        break;
                    }
                }
                if(!reopped) {
                    logic_on_not_in_reop(channel,(*i4).mask,(*i4).user_name);
                }
            }
        }
    }
}

/*!
    \brief Called as +e exception mask to ban is found out on channel after bot's JOIN
    \author VooDooMan
    \version 1
    \date 2004
    \param str Partial reply from server containing numeric reply, channel name and +e mask
*/
MY_STATIC void irc_excepted(string str)
{
    unsigned int i1;

    for(i1=0; i1<str.length(); ++i1)
        if(str[i1]==0x20)
            break;
    str.erase(0,i1+1); // erase everything before SPACE

    for(i1=0; i1<str.length(); ++i1)
        if(str[i1]==0x20)
            break;
    str.erase(0,i1+1); // erase numeric reply

    for(i1=0; i1<str.length(); ++i1)
        if(str[i1]==0x20)
            break;
    str.erase(0,i1+1); // erase my nick

    string channel;
    for(i1=0; i1<str.length(); ++i1)
        if(str[i1]==0x20)
            break;
            else
            channel+=str[i1];
    str.erase(0,i1+1); // erase channel_name

    string mask;
    for(i1=0; i1<str.length(); ++i1)
        if(str[i1]==0x20)
            break;
            else
            mask+=str[i1];

    global_keeper.add_except(channel,mask);

    vector<c_online_channel>::iterator i;
    for(i=irc_channels.begin(); i!=irc_channels.end(); ++i) {
        if(!irc_cmp_strings_case_insensitive((*i).name,channel)) {
            (*i).excepts.push_back(mask);
            (*i).got_excepts=false;
            break;
        }
    }
}

/*!
    \brief Called as +I invitation mask to +i is found out on channel after bot's JOIN
    \author VooDooMan
    \version 1
    \date 2004
    \param str Partial reply from server containing numeric reply, channel name and +I mask
*/
MY_STATIC void irc_invited(string str)
{
    unsigned int i1;

    for(i1=0; i1<str.length(); ++i1)
        if(str[i1]==0x20)
            break;
    str.erase(0,i1+1); // erase everything before SPACE

    for(i1=0; i1<str.length(); ++i1)
        if(str[i1]==0x20)
            break;
    str.erase(0,i1+1); // erase numeric reply

    for(i1=0; i1<str.length(); ++i1)
        if(str[i1]==0x20)
            break;
    str.erase(0,i1+1); // erase my nick

    string channel;
    for(i1=0; i1<str.length(); ++i1)
        if(str[i1]==0x20)
            break;
            else
            channel+=str[i1];
    str.erase(0,i1+1); // erase channel_name

    string mask;
    for(i1=0; i1<str.length(); ++i1)
        if(str[i1]==0x20)
            break;
            else
            mask+=str[i1];

    global_keeper.add_invite(channel,mask);

    vector<c_online_channel>::iterator i;
    for(i=irc_channels.begin(); i!=irc_channels.end(); ++i) {
        if(!irc_cmp_strings_case_insensitive((*i).name,channel)) {
            (*i).invites.push_back(mask);
            (*i).got_invites=false;
            break;
        }
    }
}

/*!
    \brief Called as +R reop mask is found out on channel after bot's JOIN
    \author VooDooMan
    \version 1
    \date 2004
    \param str Partial reply from server containing numeric reply, channel name and +R mask
*/
MY_STATIC void irc_in_reop(string str)
{
    unsigned int i1;

    for(i1=0; i1<str.length(); ++i1)
        if(str[i1]==0x20)
            break;
    str.erase(0,i1+1); // erase everything before SPACE

    for(i1=0; i1<str.length(); ++i1)
        if(str[i1]==0x20)
            break;
    str.erase(0,i1+1); // erase numeric reply

    for(i1=0; i1<str.length(); ++i1)
        if(str[i1]==0x20)
            break;
    str.erase(0,i1+1); // erase my nick

    string channel;
    for(i1=0; i1<str.length(); ++i1)
        if(str[i1]==0x20)
            break;
            else
            channel+=str[i1];
    str.erase(0,i1+1); // erase channel_name

    string mask;
    for(i1=0; i1<str.length(); ++i1)
        if(str[i1]==0x20)
            break;
            else
            mask+=str[i1];

    global_keeper.add_reop(channel,mask);

    vector<c_online_channel>::iterator i;
    for(i=irc_channels.begin(); i!=irc_channels.end(); ++i) {
        if(!irc_cmp_strings_case_insensitive((*i).name,channel)) {
            (*i).reops.push_back(mask);
            (*i).got_reops=false;
            break;
        }
    }
}

/*!
    \brief Checks ban list on channel after bot's JOIN by sending "MODE \#channel +b" command to server
    \author VooDooMan
    \version 1
    \date 2004
    \param channel Name of channel
*/
MY_STATIC void irc_check_bans(const string& channel)
{
    vector<c_online_channel>::iterator i;
    for(i=irc_channels.begin(); i!=irc_channels.end(); ++i) {
        if(!irc_cmp_strings_case_insensitive((*i).name,channel)) {
            (*i).got_excepts=true;
            c_online_channel* chan;
            if(irc_get_online_channel(channel,chan)) {
                if(!chan->b_bans_sent && irc_isupport.chanmodes_a.find("b")!=string::npos) {
                    chan->b_bans_sent=true;
                    string tmp="MODE ";
                    tmp+=channel;
                    tmp+=" +b";
                    irc_putserv(c_tracked_msg_to_server(tmp),true,HIGH_PRIORITY);
                }
            }
            break;
        }
    }
}

/*!
    \brief Checks invites list on channel after bot's JOIN by sending "MODE \#channel +I" command to server
    \author VooDooMan
    \version 1
    \date 2004
    \param channel Name of channel
*/
MY_STATIC void irc_check_invites(const string& channel)
{
    vector<c_online_channel>::iterator i;
    for(i=irc_channels.begin(); i!=irc_channels.end(); ++i) {
        if(!irc_cmp_strings_case_insensitive((*i).name,channel)) {
            (*i).got_invites=true;
            c_online_channel* chan;
            if(irc_get_online_channel(channel,chan)) {
                if(!chan->b_invites_sent && irc_isupport.chanmodes_a.find("I")!=string::npos) {
                    chan->b_invites_sent=true;
                    string tmp="MODE ";
                    tmp+=channel;
                    tmp+=" +I";
                    irc_putserv(c_tracked_msg_to_server(tmp),true,HIGH_PRIORITY);
                }
            }
            break;
        }
    }
}

/*!
    \brief Called when bot after JOIN has received message indicating end of +e exceptions / or error, that irc network doesn't support +e exceptions
    \author VooDooMan
    \version 1
    \date 2004
    \param str Partial reply from server, also containing channel name
*/
MY_STATIC void irc_end_of_excepted(string str)
{
    unsigned int i1;

    for(i1=0; i1<str.length(); ++i1)
        if(str[i1]==0x20)
            break;
    str.erase(0,i1+1); // erase everything before SPACE

    for(i1=0; i1<str.length(); ++i1)
        if(str[i1]==0x20)
            break;
    str.erase(0,i1+1); // erase numeric reply

    for(i1=0; i1<str.length(); ++i1)
        if(str[i1]==0x20)
            break;
    str.erase(0,i1+1); // erase my nick

    string channel;
    for(i1=0; i1<str.length(); ++i1)
        if(str[i1]==0x20)
            break;
            else
            channel+=str[i1];
    str.erase(0,i1+1); // erase channel_name

    irc_check_bans(channel);
}

/*!
    \brief Called when bot after JOIN has received message indicating end of +R reop hints
    \author VooDooMan
    \version 1
    \date 2005
    \param str Partial reply from server, also containing channel name
*/
MY_STATIC void irc_end_of_reop(string str)
{
    unsigned int i1;

    for(i1=0; i1<str.length(); ++i1)
        if(str[i1]==0x20)
            break;
    str.erase(0,i1+1); // erase everything before SPACE

    for(i1=0; i1<str.length(); ++i1)
        if(str[i1]==0x20)
            break;
    str.erase(0,i1+1); // erase numeric reply

    for(i1=0; i1<str.length(); ++i1)
        if(str[i1]==0x20)
            break;
    str.erase(0,i1+1); // erase my nick

    string channel;
    for(i1=0; i1<str.length(); ++i1)
        if(str[i1]==0x20)
            break;
            else
            channel+=str[i1];
    str.erase(0,i1+1); // erase channel_name

    vector<c_online_channel>::iterator i;
    for(i=irc_channels.begin(); i!=irc_channels.end(); ++i) {
        if(!irc_cmp_strings_case_insensitive((*i).name,channel)) {
            (*i).got_reops=true;
        }
    }

    irc_check_for_not_reopped(channel);
}

/*!
    \brief Finds c_online_user record in irc_channel list for user
    \author VooDooMan
    \version 1
    \date 2004
    \param nick Nick of user
    \return User entry
*/
c_online_user irc_find_user(const string& nick)
{
    vector<c_online_channel>::iterator i;
    for(i=irc_channels.begin(); i!=irc_channels.end(); ++i) {
        vector<c_online_user>::iterator ii;
        for(ii=(*i).users.begin(); ii!=(*i).users.end(); ++ii) {
            if(!irc_cmp_strings_case_insensitive((*ii).nick,nick)) {
                return *ii;
            }
        }
    }
    c_online_user u;
    u.nick=""; // do an OOPS!
    return u;
}

/*!
    \brief Adds/remove new mode for user on channel
    \author VooDooMan
    \version 2
    \date 2004, 2008-03-09
    \param channel Name of channel
    \param nick Nick name
    \param from_nick Which nick set the mode?
    \param prefix '-' or '+'
    \param mode '@' for +o, and '+' for +v, or '&', '%', etc.
*/
MY_STATIC void irc_user_mode_change(const string& channel, const string& nick, const string& from_nick, const char prefix, const char mode)
{
    event_dispatch_mode_on_user(channel,nick,from_nick,prefix,mode,true);

    global_keeper.change_user_mode(channel,nick,prefix,mode);

    if(!irc_cmp_strings_case_insensitive(nick,irc_nick)) {
        logic_validate(channel,irc_find_user(from_nick).in_logic_as,logic_find_user(irc_nick,irc_get_ident(irc_nick),irc_get_host(irc_nick),irc_get_fullname(irc_nick),irc_is_ircop(irc_nick)),from_nick,irc_nick,prefix,mode);

        vector<c_online_channel>::iterator i;
        for(i=irc_channels.begin(); i!=irc_channels.end(); ++i) {
            if(irc_cmp_strings_case_insensitive((*i).name,channel))
                continue;
            vector<c_online_user>::iterator ii;
            for(ii=(*i).users.begin(); ii!=(*i).users.end(); ++ii) {
                if((*ii).got_whois) {
                    if(!(*ii).nick.compare(irc_nick)) {
                        while((*ii).mode.find("0",0)!=string::npos)
                            (*ii).mode.erase((*ii).mode.find("0",0),1);
                        if(prefix=='+') {
                            while((*ii).mode.find(mode,0)!=string::npos)
                                (*ii).mode.erase((*ii).mode.find(mode,0),1);
                            (*ii).mode+=mode;
                        } else {
                            while((*ii).mode.find(mode,0)!=string::npos)
                                (*ii).mode.erase((*ii).mode.find(mode,0),1);
                        }
                        if((*ii).mode.empty())
                            (*ii).mode+="0";
                    }

                    logic_on_nick_validate((*ii).nick,(*ii).ident,(*ii).host,(*ii).fullname,channel,(*ii).irc_op,(*ii).on_join_passive,(*ii).mod_disp_join_called);
                    (*ii).on_join_called=true;
                    (*ii).old_mode=(*ii).mode;
                    (*ii).on_join_passive=true;
                    (*ii).on_join_called2=true;
                }
            }
        }
    }

    string from_user=irc_find_user(from_nick).in_logic_as;

    vector<c_online_channel>::iterator i1;
    for(i1=irc_channels.begin(); i1!=irc_channels.end(); ++i1) {
        if(irc_cmp_strings_case_insensitive((*i1).name,channel))
            continue;
        vector<c_online_user>::iterator i2;
        for(i2=(*i1).users.begin(); i2!=(*i1).users.end(); ++i2) {
            if(!irc_cmp_strings_case_insensitive((*i2).nick,nick)) {
                string m=(*i2).mode;
                while(m.find("0",0)!=string::npos)
                    m.erase(m.find("0",0),1);
                if(prefix=='+') {
                    while(m.find(mode,0)!=string::npos)
                        m.erase(m.find(mode,0),1);
                    m+=mode;
                } else {
                    while(m.find(mode,0)!=string::npos)
                        m.erase(m.find(mode,0),1);
                }
                if(m.empty())
                    m+="0";
                (*i2).mode=m;
                (*i2).got_mode=true;

                logic_set_dynamic((*i2).in_logic_as,from_user,prefix,mode);
            }
        }
    }
}

/*!
    \brief Evaluates, and validates (or accidentally calls specific event) MODE change +o/-o (op-erator) to user
    \author VooDooMan
    \version 2
    \date 2004, 2008-03-09
    \param channel Name of channel
    \param prefix Contains '+' or '-' prefix
    \param nick Nick of originator of MODE change
    \param ident Ident of originator of MODE change
    \param host Host of originator of MODE change
    \param whom Nick of target user
*/
// +-o nick
MY_STATIC void irc_mode_change_op(const string& channel, const char prefix, const string& nick, const string& ident, const string& host, const string& whom)
{
    irc_user_mode_change(channel,whom,nick,prefix,'@');

    if(!nick.compare(irc_nick))
        return;

    c_online_user source=irc_find_user(nick);
    c_online_user target=irc_find_user(whom);

    if(prefix=='+' && !irc_cmp_strings_case_insensitive(nick,irc_nick)) {
        vector<c_online_channel>::iterator i1;
        for(i1=irc_channels.begin(); i1!=irc_channels.end(); ++i1) {
            vector<c_online_user>::iterator i2;
            for(i2=(*i1).users.begin(); i2!=(*i1).users.end(); ++i2)
                if((*i2).got_whois) {
                    if((*i2).mode.compare((*i2).old_mode))
                        logic_on_nick_validate((*i2).nick,(*i2).ident,(*i2).host,(*i2).fullname,(*i1).name,(*i2).irc_op,(*i2).on_join_passive,(*i2).mod_disp_join_called);
                    (*i2).on_join_called=true;
                    (*i2).old_mode=(*i2).mode;
                    (*i2).on_join_passive=true;
                    (*i2).on_join_called2=true;
                }
        }
    }

    if(!source.nick.compare("") && !(ident.empty() && host.empty())) {
        string log;
        log="OOPS: Occurred OOPS #3: unknown nick \"";
        log+=nick;
        log+="!";
        log+=ident;
        log+="@";
        log+=host;
        log+="\"";
        log_debug(GET_STR_DEBUG_INFO_EX(log));
        return;
    }

    if(!target.nick.compare("")) {
        string log;
        log="OOPS: Occurred OOPS #4: unknown nick \"";
        log+=whom;
        log+="\"";
        log_debug(GET_STR_DEBUG_INFO_EX(log));
        return;
    }

    string source_user;
    if(ident.empty() && host.empty())
        source_user=nick; // it is a server
    else
        source_user=logic_find_user(source.nick,source.ident,source.host,source.fullname,source.irc_op);
    string target_user=logic_find_user(target.nick,target.ident,target.host,target.fullname,target.irc_op);

    if(irc_cmp_strings_case_insensitive(whom,irc_nick)) // we have already called it above
        logic_validate(channel,source_user,target_user,source.nick,target.nick,prefix,'@');

    if(prefix=='+') {
        if(!irc_cmp_strings_case_insensitive(target.nick,irc_nick)) {
            vector<c_online_channel>::iterator i;
            for(i=irc_channels.begin(); i!=irc_channels.end(); ++i) {
                if((*i).got_excepts) {
                    irc_check_bans((*i).name);
                }
                irc_check_invites((*i).name);

                if(!(*i).b_reops_sent && irc_isupport.chanmodes_a.find("R")!=string::npos) {
                    string reop="MODE ";
                    reop+=channel;
                    reop+=" +R";
                    irc_putserv(c_tracked_msg_to_server(reop),true,HIGH_PRIORITY);
                }
                if((*i).got_reops && irc_isupport.chanmodes_a.find("R")!=string::npos)
                    irc_check_for_not_reopped(channel);
                if(irc_isupport.chanmodes_a.find("R")!=string::npos) {
                    (*i).b_reops_sent=true;
                }
            }
        }
    }
}

/*!
    \brief Evaluates, and validates (or accidentally calls specific event) MODE change +v/-v (voice) to user
    \author VooDooMan
    \version 2
    \date 2004, 2008-03-09
    \param channel Name of channel
    \param prefix Contains '+' or '-' prefix
    \param nick Nick of originator of MODE change
    \param ident Ident of originator of MODE change
    \param host Host of originator of MODE change
    \param whom Nick of target user
*/
// +-v nick
MY_STATIC void irc_mode_change_voice(const string& channel, const char prefix, const string& nick, const string& ident, const string& host, const string& whom)
{
    irc_user_mode_change(channel,whom,nick,prefix,'+');

    c_online_user source=irc_find_user(nick);
    c_online_user target=irc_find_user(whom);

    if(!source.nick.compare("") && !(ident.empty() && host.empty())) {
        string log;
        log="OOPS: Occurred OOPS #5: unknown nick \"";
        log+=nick;
        log+="!";
        log+=ident;
        log+="@";
        log+=host;
        log+="\"";
        log_debug(GET_STR_DEBUG_INFO_EX(log));
        return;
    }

    if(!target.nick.compare("")) {
        string log;
        log="OOPS: Occurred OOPS #6: unknown nick \"";
        log+=whom;
        log+="\"";
        log_debug(GET_STR_DEBUG_INFO_EX(log));
        return;
    }

    string source_user;
    if(ident.empty() && host.empty())
        source_user=nick; // it is a server
    else
        source_user=logic_find_user(source.nick,source.ident,source.host,source.fullname,source.irc_op);
    string target_user=logic_find_user(target.nick,target.ident,target.host,target.fullname,target.irc_op);

    if(irc_cmp_strings_case_insensitive(whom,irc_nick)) // we have already called it above
        logic_validate(channel,source_user,target_user,source.nick,target.nick,prefix,'+');
}

/*!
    \brief Evaluates, and validates (or accidentally calls specific event) MODE change with other modes to user
    \author VooDooMan
    \version 2
    \date 2004, 2008-03-09
    \param channel Name of channel
    \param prefix Contains '+' or '-' prefix
    \param mode '~', '&', or '%'
    \param nick Nick of originator of MODE change
    \param ident Ident of originator of MODE change
    \param host Host of originator of MODE change
    \param whom Nick of target user
*/
MY_STATIC void irc_mode_change_other(const string& channel, const char prefix, const char mode, const string& nick, const string& ident, const string& host, const string& whom)
{
    irc_user_mode_change(channel,whom,nick,prefix,mode);

    c_online_user source=irc_find_user(nick);
    c_online_user target=irc_find_user(whom);

    if(!source.nick.compare("") && !(ident.empty() && host.empty())) {
        string log;
        log="OOPS: Occurred OOPS #7: unknown nick \"";
        log+=nick;
        log+="!";
        log+=ident;
        log+="@";
        log+=host;
        log+="\"";
        log_debug(GET_STR_DEBUG_INFO_EX(log));
        return;
    }

    if(!target.nick.compare("")) {
        string log;
        log="OOPS: Occurred OOPS #8: unknown nick \"";
        log+=whom;
        log+="\"";
        log_debug(GET_STR_DEBUG_INFO_EX(log));
        return;
    }

    string source_user;
    if(ident.empty() && host.empty())
        source_user=nick; // it is a server
    else
        source_user=logic_find_user(source.nick,source.ident,source.host,source.fullname,source.irc_op);
    string target_user=logic_find_user(target.nick,target.ident,target.host,target.fullname,target.irc_op);

    if(irc_cmp_strings_case_insensitive(whom,irc_nick)) // we have already called it above
        logic_validate(channel,source_user,target_user,source.nick,target.nick,prefix,mode);
}

/*!
    \brief Evaluates, and validates (or accidentally calls specific event) MODE change +/-a,i,m,n,q,p,s,r,t to channel
    \author VooDooMan
    \version 2
    \date 2004, 2008-03-09
    \param channel Name of channel
    \param prefix Contains '+' or '-' prefix
    \param mode One of followig: 'a', 'i', 'm', 'n', 'q', 'p', 's', 'r', 't'
    \param nick Nick of originator of MODE change
    \param ident Ident of originator of MODE change
    \param host Host of originator of MODE change
*/
// +-aimnqpsrt
MY_STATIC void irc_mode_change_chan_status(const string& channel, const char prefix, const char mode, const string& nick, const string& ident, const string& host)
{
    event_dispatch_mode_on_channel(channel,nick,prefix,mode,"",true);

    global_keeper.change_chan_mode(channel,prefix,mode);
    {
        vector<c_online_channel>::iterator i;
        for(i=irc_channels.begin(); i!=irc_channels.end(); ++i) {
            if(irc_cmp_strings_case_insensitive((*i).name,channel))
                continue;
            while((*i).modes.find(mode,0)!=string::npos)
                (*i).modes.erase((*i).modes.find(mode,0),1);
            if(prefix=='+')
                (*i).modes+=(string)""+mode;
            break;
        }
    }

    c_online_user source=irc_find_user(nick);

    if(!source.nick.compare("") && !(ident.empty() && host.empty())) {
        string log;
        log="OOPS: Occurred OOPS #9: unknown nick \"";
        log+=nick;
        log+="!";
        log+=ident;
        log+="@";
        log+=host;
        log+="\"";
        log_debug(GET_STR_DEBUG_INFO_EX(log));
        return;
    }

    string source_user;
    if(ident.empty() && host.empty())
        source_user=nick; // it is a server
    else
        source_user=logic_find_user(source.nick,source.ident,source.host,source.fullname,source.irc_op);

    logic_validate_chan_mode(channel,source_user,source.nick,prefix,mode);
}

/*!
    \brief Evaluates, and validates (or accidentally calls specific event) MODE change +/-k KEY to channel
    \author VooDooMan
    \version 2
    \date 2004, 2008-03-09
    \param channel Name of channel
    \param prefix Contains '+' or '-' prefix
    \param nick Nick of originator of MODE change
    \param ident Ident of originator of MODE change
    \param host Host of originator of MODE change
    \param key Parameter to +/-k mode (key)
*/
// +-k [key]
MY_STATIC void irc_mode_change_chan_key(const string& channel, const char prefix, const string& nick, const string& ident, const string& host, const string& key)
{
    event_dispatch_mode_on_channel(channel,nick,prefix,'k',"",true);

    global_keeper.change_chan_key(channel,prefix,key);
    {
        vector<c_online_channel>::iterator i;
        for(i=irc_channels.begin(); i!=irc_channels.end(); ++i) {
            if(irc_cmp_strings_case_insensitive((*i).name,channel))
                continue;
            if(prefix=='+')
                (*i).key=key;
            else
                (*i).key="";
            break;
        }
    }

    c_online_user source=irc_find_user(nick);

    if(!source.nick.compare("")&& !(ident.empty() && host.empty())) {
        string log;
        log="OOPS: Occurred OOPS #10: unknown nick \"";
        log+=nick;
        log+="!";
        log+=ident;
        log+="@";
        log+=host;
        log+="\"";
        log_debug(log);
        return;
    }

    string source_user;
    if(ident.empty() && host.empty())
        source_user=nick; // it is a server
    else
        source_user=logic_find_user(source.nick,source.ident,source.host,source.fullname,source.irc_op);

    logic_validate_chan_key(channel,source_user,source.nick,prefix,key);
}

/*!
    \brief Evaluates, and validates (or accidentally calls specific event) MODE change +/-l NUMBER to channel
    \author VooDooMan
    \version 2
    \date 2004, 2008-03-09
    \param channel Name of channel
    \param prefix Contains '+' or '-' prefix
    \param nick Nick of originator of MODE change
    \param ident Ident of originator of MODE change
    \param host Host of originator of MODE change
    \param limit Parameter to +/-l mode (limit - number)
*/
// +-l [limit]
MY_STATIC void irc_mode_change_chan_limit(const string& channel, const char prefix, const string& nick, const string& ident, const string& host, const string& limit)
{
    event_dispatch_mode_on_channel(channel,nick,prefix,'l',"",true);

    global_keeper.change_chan_limit(channel,prefix,limit);
    {
        vector<c_online_channel>::iterator i;
        for(i=irc_channels.begin(); i!=irc_channels.end(); ++i) {
            if(irc_cmp_strings_case_insensitive((*i).name,channel))
                continue;
            if(prefix=='+')
                (*i).limit=limit;
            else
                (*i).limit="";
            break;
        }
    }

    c_online_user source=irc_find_user(nick);

    if(!source.nick.compare("") && !(ident.empty() && host.empty())) {
        string log;
        log="OOPS: Occurred OOPS #11: unknown nick \"";
        log+=nick;
        log+="!";
        log+=ident;
        log+="@";
        log+=host;
        log+="\"";
        log_debug(GET_STR_DEBUG_INFO_EX(log));
        return;
    }

    string source_user;
    if(ident.empty() && host.empty())
        source_user=nick; // it is a server
    else
        source_user=logic_find_user(source.nick,source.ident,source.host,source.fullname,source.irc_op);

    logic_validate_chan_limit(channel,source_user,source.nick,prefix,limit);
}

/*!
    \brief Evaluates, and validates (or accidentally calls specific event) MODE change +/-b to channel
    \author VooDooMan
    \version 2
    \date 2004, 2008-03-09
    \param channel Name of channel
    \param prefix Contains '+' or '-' prefix
    \param nick Nick of originator of MODE change
    \param ident Ident of originator of MODE change
    \param host Host of originator of MODE change
    \param mask Parameter to +/-b mode (mask)
*/
// +-b mask
MY_STATIC void irc_mode_change_ban(const string& channel, const char prefix, const string& nick, const string& ident, const string& host, const string& mask)
{
    event_dispatch_mode_on_channel(channel,nick,prefix,'b',mask,true);

    global_keeper.change_ban(channel,prefix,mask);
    {
        vector<c_online_channel>::iterator i;
        for(i=irc_channels.begin(); i!=irc_channels.end(); ++i) {
            if(irc_cmp_strings_case_insensitive((*i).name,channel))
                continue;
            vector<string>::iterator i1;
        again1:
            for(i1=(*i).bans.begin(); i1!=(*i).bans.end(); ++i1)
                if(!mask.compare(*i1)) {
                    (*i).bans.erase(i1);
                    goto again1;
                }
            if(prefix=='+')
                (*i).bans.push_back(mask);
            break;
        }
    }

    c_online_user source=irc_find_user(nick);

    if(!source.nick.compare("") && !(ident.empty() && host.empty())) {
        string log;
        log="OOPS: Occurred OOPS #12: unknown nick \"";
        log+=nick;
        log+="!";
        log+=ident;
        log+="@";
        log+=host;
        log+="\"";
        log_debug(log);
        return;
    }

    string source_user;
    if(ident.empty() && host.empty())
        source_user=nick; // it is a server
    else
        source_user=logic_find_user(source.nick,source.ident,source.host,source.fullname,source.irc_op);

    logic_validate_by_mask(channel,source_user,nick,mask,prefix,'b');
}

/*!
    \brief Evaluates, and validates (or accidentally calls specific event) MODE change +/-e to channel
    \author VooDooMan
    \version 2
    \date 2004, 2008-03-09
    \param channel Name of channel
    \param prefix Contains '+' or '-' prefix
    \param nick Nick of originator of MODE change
    \param ident Ident of originator of MODE change
    \param host Host of originator of MODE change
    \param mask Parameter to +/-e mode (mask)
*/
// +-e mask
MY_STATIC void irc_mode_change_ban_exception(const string& channel, const char prefix, const string& nick, const string& ident, const string& host, const string& mask)
{
    event_dispatch_mode_on_channel(channel,nick,prefix,'e',mask,true);

    global_keeper.change_except(channel,prefix,mask);
    {
        vector<c_online_channel>::iterator i;
        for(i=irc_channels.begin(); i!=irc_channels.end(); ++i) {
            if(irc_cmp_strings_case_insensitive((*i).name,channel))
                continue;
            vector<string>::iterator i1;
        again1:
            for(i1=(*i).excepts.begin(); i1!=(*i).excepts.end(); ++i1)
                if(!mask.compare(*i1)) {
                    (*i).excepts.erase(i1);
                    goto again1;
                }
            if(prefix=='+')
                (*i).excepts.push_back(mask);
            break;
        }
    }

    c_online_user source=irc_find_user(nick);

    if(!source.nick.compare("") && !(ident.empty() && host.empty())) {
        string log;
        log="OOPS: Occurred OOPS #13: unknown nick \"";
        log+=nick;
        log+="!";
        log+=ident;
        log+="@";
        log+=host;
        log+="\"";
        log_debug(GET_STR_DEBUG_INFO_EX(log));
        return;
    }

    string source_user;
    if(ident.empty() && host.empty())
        source_user=nick; // it is a server
    else
        source_user=logic_find_user(source.nick,source.ident,source.host,source.fullname,source.irc_op);

    logic_validate_by_mask(channel,source_user,nick,mask,prefix,'e');
}

/*!
    \brief Evaluates, and validates (or accidentally calls specific event) MODE change +/-I to channel
    \author VooDooMan
    \version 2
    \date 2004, 2008-03-09
    \param channel Name of channel
    \param prefix Contains '+' or '-' prefix
    \param nick Nick of originator of MODE change
    \param ident Ident of originator of MODE change
    \param host Host of originator of MODE change
    \param mask Parameter to +/-I mode (mask)
*/
// +-I mask
MY_STATIC void irc_mode_change_invitation(const string& channel, const char prefix, const string& nick, const string& ident, const string& host, const string& mask)
{
    event_dispatch_mode_on_channel(channel,nick,prefix,'I',mask,true);

    global_keeper.change_invite(channel,prefix,mask);
    vector<c_online_channel>::iterator i1;
    for(i1=irc_channels.begin(); i1!=irc_channels.end(); ++i1) {
        if(!irc_cmp_strings_case_insensitive((*i1).name,channel)) {
            if(prefix=='+') {
                vector<string>::iterator i2;
                for(i2=(*i1).invites.begin(); i2!=(*i1).invites.end(); ++i2) {
                    if(!(*i2).compare(mask)) {
                        // we already got it
                        return;
                    }
                }
                (*i1).invites.push_back(mask);
                c_online_user source=irc_find_user(nick);

                string source_user, user_nick=nick;
                if(ident.empty() && host.empty()) {
                    source_user=nick; // it is a server
                    user_nick="";
                } else
                    source_user=logic_find_user(source.nick,source.ident,source.host,source.fullname,source.irc_op);
                logic_validate_by_mask(channel,source_user,user_nick,mask,prefix,'I');
            }
            if(prefix=='-') {
                vector<string>::iterator i2;
            l1:
                for(i2=(*i1).invites.begin(); i2!=(*i1).invites.end(); ++i2) {
                    if(!(*i2).compare(mask)) {
                        (*i1).invites.erase(i2);
                        goto l1;
                    }
                }
                c_online_user source=irc_find_user(nick);

                string source_user, user_nick=nick;
                if(ident.empty() && host.empty()) {
                    source_user=nick; // it is a server
                    user_nick="";
                } else
                    source_user=logic_find_user(source.nick,source.ident,source.host,source.fullname,source.irc_op);
                logic_validate_by_mask(channel,source_user,user_nick,mask,prefix,'I');
            }
            break;
        }
    }
}

/*!
    \brief Evaluates, and validates (or accidentally calls specific event) MODE change +/-R to channel
    \author VooDooMan
    \version 2
    \date 2004, 2008-03-09
    \param channel Name of channel
    \param prefix Contains '+' or '-' prefix
    \param nick Nick of originator of MODE change
    \param ident Ident of originator of MODE change
    \param host Host of originator of MODE change
    \param mask Parameter to +/-R mode (mask)
*/
// +-R mask
MY_STATIC void irc_mode_change_reop(const string& channel, const char prefix, const string& nick, const string& ident, const string& host, const string& mask)
{
    event_dispatch_mode_on_channel(channel,nick,prefix,'R',mask,true);

    global_keeper.change_reop(channel,prefix,mask);
    vector<c_online_channel>::iterator i1;
    for(i1=irc_channels.begin(); i1!=irc_channels.end(); ++i1) {
        if(!irc_cmp_strings_case_insensitive((*i1).name,channel)) {
            if(prefix=='+') {
                vector<string>::iterator i2;
                for(i2=(*i1).reops.begin(); i2!=(*i1).reops.end(); ++i2) {
                    if(!(*i2).compare(mask)) {
                        // we already got it
                        return;
                    }
                }
                (*i1).reops.push_back(mask);

                c_online_user source=irc_find_user(nick);

                string source_user, user_nick=nick;
                if(ident.empty() && host.empty()) {
                    source_user=nick; // it is a server
                    user_nick="";
                } else
                    source_user=logic_find_user(source.nick,source.ident,source.host,source.fullname,source.irc_op);
                logic_validate_by_mask(channel,source_user,user_nick,mask,prefix,'R');
            }
            if(prefix=='-') {
                vector<string>::iterator i2;
            l1:
                for(i2=(*i1).reops.begin(); i2!=(*i1).reops.end(); ++i2) {
                    if(!(*i2).compare(mask)) {
                        (*i1).reops.erase(i2);
                        goto l1;
                    }
                }
                c_online_user source=irc_find_user(nick);

                string source_user, user_nick=nick;
                if(ident.empty() && host.empty()) {
                    source_user=nick; // it is a server
                    user_nick="";
                } else
                    source_user=logic_find_user(source.nick,source.ident,source.host,source.fullname,source.irc_op);
                logic_validate_by_mask(channel,source_user,user_nick,mask,prefix,'R');
            }
            break;
        }
    }
}

/*!
    \brief Evaluates, and validates (or accidentally calls specific event) MODE change +/- other modes with mask (class A) to channel
    \author VooDooMan
    \version 2
    \date 2004, 2008-03-09
    \param channel Name of channel
    \param prefix Contains '+' or '-' prefix
    \param mode Mode
    \param nick Nick of originator of MODE change
    \param ident Ident of originator of MODE change
    \param host Host of originator of MODE change
    \param mask Parameter to mode (mask)
*/
MY_STATIC void irc_mode_change_other_mask(const string& channel, const char prefix, const char mode, const string& nick, const string& ident, const string host, const string& mask)
{
    event_dispatch_mode_on_channel(channel,nick,prefix,mode,mask,true);

    global_keeper.change_other_mode_mask(channel,prefix,mode,mask);
    vector<c_online_channel>::iterator i1;
    for(i1=irc_channels.begin(); i1!=irc_channels.end(); ++i1) {
        if(!irc_cmp_strings_case_insensitive((*i1).name,channel)) {
            if(prefix=='+') {
                map<char,string>::iterator i2;
                for(i2=(*i1).other_modes.begin(); i2!=(*i1).other_modes.end(); ++i2) {
                    if(mode==(*i2).first && !(*i2).second.compare(mask)) {
                        // we already got it
                        return;
                    }
                }
                (*i1).other_modes.insert(pair<char,string>(mode,mask));

                c_online_user source=irc_find_user(nick);

                string source_user, user_nick=nick;
                if(ident.empty() && host.empty()) {
                    source_user=nick; // it is a server
                    user_nick="";
                } else
                    source_user=logic_find_user(source.nick,source.ident,source.host,source.fullname,source.irc_op);
                logic_validate_by_mask(channel,source_user,user_nick,mask,prefix,mode);
            }
            if(prefix=='-') {
                map<char,string>::iterator i2;
            l1:
                for(i2=(*i1).other_modes.begin(); i2!=(*i1).other_modes.end(); ++i2) {
                    if(mode==(*i2).first && !(*i2).second.compare(mask)) {
                        (*i1).other_modes.erase(i2);
                        goto l1;
                    }
                }
                c_online_user source=irc_find_user(nick);

                string source_user, user_nick=nick;
                if(ident.empty() && host.empty()) {
                    source_user=nick; // it is a server
                    user_nick="";
                } else
                    source_user=logic_find_user(source.nick,source.ident,source.host,source.fullname,source.irc_op);
                logic_validate_by_mask(channel,source_user,user_nick,mask,prefix,mode);
            }
            break;
        }
    }
}

/*!
    \brief Called after receiving 324 numeric from server (channel modes after MODE \#channel), and calls on_key, on_limit functions after parsing the message
    \author VooDooMan
    \version 1
    \date 2004
    \param str Partial message from server
*/
MY_STATIC void irc_RPL_324(string str)
{
    // 324 my_nick #channel +ntlk 20 channel_key

    string log=str;
    string old_str=str;
    unsigned int i1;

    for(i1=0; i1<str.length(); ++i1)
        if(str[i1]==' ')
            break;
    str.erase(0,i1+1); // erase 324

    for(i1=0; i1<str.length(); ++i1)
        if(str[i1]==' ')
            break;
    str.erase(0,i1+1); // erase my nick

    string channel;
    for(i1=0; i1<str.length(); ++i1)
        if(str[i1]==0x20)
            break;
        else
            channel+=str[i1];
    str.erase(0,i1+1); // erase #channel

    string flags;
    for(i1=0; i1<str.length(); ++i1)
        if(str[i1]==0x20)
            break;
            else
            flags+=str[i1];
    str.erase(0,i1+1); // erase flags

    string parameters;
    for(i1=0; i1<str.length(); ++i1)
        parameters+=str[i1];

    multimap<string,string> modes;

    char prefix=0;
    unsigned int i2=0;
    for(i1=0; i1<flags.length(); ++i1) {
        if(flags[i1]=='+') {
            prefix='+';
            continue;
        }
        if(flags[i1]=='-') {
            prefix='-';
            continue;
        }
        string p;
        int type=0;

        type=0;
        if(irc_isupport.get_prefix1(flags[i1])!='0')
            type=1;
        if(irc_isupport.chanmodes_d.find(flags[i1])!=string::npos)
            type=2;
        if(irc_isupport.chanmodes_c.find(flags[i1])!=string::npos)
            type=3;
        if(irc_isupport.chanmodes_a.find(flags[i1])!=string::npos)
            type=4;
        if(irc_isupport.chanmodes_b.find(flags[i1])!=string::npos)
            type=5;

        if(type==1 || type==4) {
            for(; i2<parameters.length(); ++i2) {
                if(parameters[i2]==0x20) {
                    i2++;
                    break;
                }
                p+=parameters[i2];
            }
        }
        if(type==3 && prefix=='+') {
            for(; i2<parameters.length(); ++i2) {
                if(parameters[i2]==0x20) {
                    i2++;
                    break;
                }
                p+=parameters[i2];
            }
        }
        if(type==5) {
            for(; i2<parameters.length(); ++i2) {
                if(parameters[i2]==0x20) {
                    i2++;
                    break;
                }
                p+=parameters[i2];
            }
        }
        if(type==2)
            p="";
        if(type==0) {
            p="";
            log="OOPS: Occurred error: unknown mode ";
            log+=prefix;
            log+=flags[i1];
            log+=" = string: ";
            log+=old_str;
            log_debug(GET_STR_DEBUG_INFO_EX(log));
        }

        pair<string,string> pr;
        pr.first=prefix;
        pr.first+=flags[i1];
        pr.second=p;
        modes.insert(pr);
    }

    multimap<string,string>::iterator im;
    for(im=modes.begin(); im!=modes.end(); ++im) {
        char prefix=(*im).first[0];
        char mode=(*im).first[1];
        string param=(*im).second;

        string unknown=""; // name of host_unknown user
        {
            deque<c_user>::iterator i1;
            for(i1=r_users.begin(); i1!=r_users.end(); ++i1) {
                if((*i1).host_unknown) {
                    unknown=(*i1).name;
                    break;
                }
            }
        }

        bool got=false;

        vector<c_online_channel>::iterator ii;
        for(ii=irc_channels.begin(); ii!=irc_channels.end(); ++ii)
            if(!irc_cmp_strings_case_insensitive(channel,(*ii).name))
                break;
        if(ii==irc_channels.end())
            continue;

        if(!got && irc_isupport.chanmodes_d.find(mode)!=string::npos) {
            global_keeper.change_chan_mode(channel,prefix,mode);
            got=true;
            // these modes were handled elsewhere
            logic_validate_chan_mode(channel,unknown,"",prefix,mode);
        }
        if(!got && mode=='k' && irc_isupport.chanmodes_b.find(mode)!=string::npos) {
            global_keeper.change_chan_key(channel,prefix,param);
            got=true;
            (*ii).key=param;
            logic_validate_chan_key(channel,unknown,"",prefix,param);
        }
        if(!got && mode=='l' && irc_isupport.chanmodes_c.find(mode)!=string::npos) {
            global_keeper.change_chan_limit(channel,prefix,param);
            got=true;
            (*ii).limit=param;
            logic_validate_chan_limit(channel,unknown,"",prefix,param);
        }
        if(!got) {
            log="OOPS: Occurred OOPS: unsupported mode ";
            log+=prefix;
            log+=mode;
            log+=" ";
            log+=param;
            log+=" = string: ";
            log+=prefix;
            log+=mode;
            if(!param.empty()) {
                log+=" ";
                log+=param;
            }
            log_debug(GET_STR_DEBUG_INFO_EX(log));
        }
    }
}

MY_STATIC void irc_check_flood(int type, const string& channel, const string& full_host, const string& msg);

/*!
    \brief Called after receiving RPL_MODE from server (someone has changed MODE for channel), and calls irc_mode_change_XXX functions after parsing the message
    \author VooDooMan
    \version 1
    \date 2004
    \param str Partial message from server
*/
MY_STATIC void irc_RPL_MODE(string str)
{
    string log=str;
    string old_str=str;
    unsigned int i1;

    for(i1=0; i1<str.length(); ++i1)
        if(str[i1]==':')
            break;
    str.erase(0,i1+1); // erase everything before ':'

    string nick;
    for(i1=0; i1<str.length(); ++i1)
        if(str[i1]=='!' || str[i1]==0x20)
            break;
            else
            nick+=str[i1];
    bool by_server=str[i1]==0x20;
    str.erase(0,i1+1); // erase nick name

    string ident;
    string host;
    if(!by_server) {
        for(i1=0; i1<str.length(); ++i1)
            if(str[i1]=='@')
                break;
                else
                ident+=str[i1];
        str.erase(0,i1+1); // erase ident

        for(i1=0; i1<str.length(); ++i1)
            if(str[i1]==0x20)
                break;
                else
                host+=str[i1];
        str.erase(0,i1+1); // erase host name
    }

    for(i1=0; i1<str.length(); ++i1)
        if(str[i1]==0x20)
            break;
    str.erase(0,i1+1); // erase "MODE"

    string target;
    for(i1=0; i1<str.length(); ++i1)
        if(str[i1]==0x20)
            break;
            else
            target+=str[i1];
    str.erase(0,i1+1); // erase target

    if(!irc_cmp_strings_case_insensitive(target,irc_nick)) {
        // it is user MODE for me
        return;
    }

    string flags;
    for(i1=0; i1<str.length(); ++i1)
        if(str[i1]==0x20)
            break;
            else
            flags+=str[i1];
    str.erase(0,i1+1); // erase flags

    string parameters;
    for(i1=0; i1<str.length(); ++i1)
        parameters+=str[i1];

    multimap<string,string> modes;

    char prefix=0;
    unsigned int i2=0;
    for(i1=0; i1<flags.length(); ++i1) {
        if(flags[i1]=='+') {
            prefix='+';
            continue;
        }
        if(flags[i1]=='-') {
            prefix='-';
            continue;
        }
        string p;
        int type=0;
        if(irc_isupport.get_prefix1(flags[i1])!='0')
            type=1;
        if(irc_isupport.chanmodes_d.find(flags[i1])!=string::npos)
            type=2;
        if(irc_isupport.chanmodes_c.find(flags[i1])!=string::npos)
            type=3;
        if(irc_isupport.chanmodes_a.find(flags[i1])!=string::npos)
            type=4;
        if(irc_isupport.chanmodes_b.find(flags[i1])!=string::npos)
            type=5;

        if(type==1 || type==4) {
            for(; i2<parameters.length(); ++i2) {
                if(parameters[i2]==0x20) {
                    i2++;
                    break;
                }
                p+=parameters[i2];
            }
        }
        if(type==3 && prefix=='+') {
            for(; i2<parameters.length(); ++i2) {
                if(parameters[i2]==0x20) {
                    i2++;
                    break;
                }
                p+=parameters[i2];
            }
        }
        if(type==5) {
            for(; i2<parameters.length(); ++i2) {
                if(parameters[i2]==0x20) {
                    i2++;
                    break;
                }
                p+=parameters[i2];
            }
        }
        if(type==2)
            p="";
        if(type==0) {
            p="";
            log="OOPS: Occurred error: unknown mode ";
            log+=prefix;
            log+=flags[i1];
            log+=" = string: ";
            log+=old_str;
            log_debug(GET_STR_DEBUG_INFO_EX(log));
        }

        pair<string,string> pr;
        pr.first=prefix;
        pr.first+=flags[i1];
        pr.second=p;
        modes.insert(pr);
    }

    string full_host=nick;
    full_host+="!";
    full_host+=ident;
    full_host+="@";
    full_host+=host;

    multimap<string,string>::iterator im;
    for(im=modes.begin(); im!=modes.end(); ++im) {
        char prefix=(*im).first[0];
        char mode=(*im).first[1];
        string param=(*im).second;
        string msg=(string)""+prefix;
        msg+=(string)""+mode;
        if(!param.empty()) {
            msg+=" ";
            msg+=param;
        }
        irc_check_flood(TYPE_MODE,target,full_host,msg);

        bool got=false;
        if(mode=='o' && irc_isupport.get_prefix1(mode)!='0') {
            got=true;
            // +-o - channel operator
            irc_remove_redundant_mode(target,param,(string)""+prefix+mode);
            irc_mode_change_op(target,prefix,nick,ident,host,param);
            if(ident.empty() && host.empty()) {
                c_mode_queue_entry e;
                e.channel=target;
                e.mode=(string)""+prefix+mode;
                e.param=param;
                e.submit=my_time(NULL);
                e.param_is_nick=true;
                modes_from_server.push_back(e);
            }
        }
        if(!got && mode=='v' && irc_isupport.get_prefix1(mode)!='0') {
            got=true;
            // +-v - channel voice
            irc_remove_redundant_mode(target,param,(string)""+prefix+mode);
            irc_mode_change_voice(target,prefix,nick,ident,host,param);
            if(ident.empty() && host.empty()) {
                c_mode_queue_entry e;
                e.channel=target;
                e.mode=(string)""+prefix+mode;
                e.param=param;
                e.submit=my_time(NULL);
                e.param_is_nick=true;
                modes_from_server.push_back(e);
            }
        }
        if(!got && irc_isupport.get_prefix1(mode)!='0') {
            got=true;
            irc_remove_redundant_mode(target,param,(string)""+prefix+mode);
            irc_mode_change_other(target,prefix,irc_isupport.get_prefix1(mode),nick,ident,host,param);
            if(ident.empty() && host.empty()) {
                c_mode_queue_entry e;
                e.channel=target;
                e.mode=(string)""+prefix+mode;
                e.param=param;
                e.submit=my_time(NULL);
                e.param_is_nick=true;
                modes_from_server.push_back(e);
            }
        }
        if(!got && irc_isupport.chanmodes_d.find(mode)!=string::npos) {
            got=true;
            irc_mode_change_chan_status(target,prefix,mode,nick,ident,host);
        }
        if(!got && mode=='k' && irc_isupport.chanmodes_b.find(mode)!=string::npos) {
            got=true;
            irc_mode_change_chan_key(target,prefix,nick,ident,host,param);
        }
        if(!got && mode=='l' && irc_isupport.chanmodes_c.find(mode)!=string::npos) {
            got=true;
            irc_mode_change_chan_limit(target,prefix,nick,ident,host,param);
        }
        if(!got && mode=='b' && irc_isupport.chanmodes_a.find(mode)!=string::npos) {
            got=true;
            irc_mode_change_ban(target,prefix,nick,ident,host,param);
        }
        if(!got && mode=='e' && irc_isupport.chanmodes_a.find(mode)!=string::npos) {
            got=true;
            irc_mode_change_ban_exception(target,prefix,nick,ident,host,param);
        }
        if(!got && mode=='I' && irc_isupport.chanmodes_a.find(mode)!=string::npos) {
            got=true;
            irc_mode_change_invitation(target,prefix,nick,ident,host,param);
        }
        if(!got && mode=='R' && irc_isupport.chanmodes_a.find(mode)!=string::npos) {
            got=true;
            irc_mode_change_reop(target,prefix,nick,ident,host,param);
        }
        if(!got && irc_isupport.chanmodes_a.find(mode)!=string::npos) {
            got=true;
            irc_mode_change_other_mask(target,prefix,mode,nick,ident,host,param);
        }
        if(!got) {
            log="OOPS: Occurred error: unsupported mode ";
            log+=prefix;
            log+=mode;
            log+=" ";
            log+=param;
            log+=" = string: ";
            log+=str;
            log_debug(GET_STR_DEBUG_INFO_EX(log));
        }
    }
}

/*!
    \brief Called after receiving RPL_TOPIC from server (someone has changed TOPIC for channel), and calls logic_validate_topic function after parsing the message
    \author VooDooMan
    \version 1
    \date 2004
    \param str Partial message from server
*/
MY_STATIC void irc_RPL_TOPIC(string str)
{
    string log=str;
    unsigned int i1;

    for(i1=0; i1<str.length(); ++i1)
        if(str[i1]==':')
            break;
    str.erase(0,i1+1); // erase everything before ':'

    string nick;
    for(i1=0; i1<str.length(); ++i1)
        if(str[i1]=='!')
            break;
            else
            nick+=str[i1];
    str.erase(0,i1+1); // erase nick name

    string ident;
    for(i1=0; i1<str.length(); ++i1)
        if(str[i1]=='@')
            break;
            else
            ident+=str[i1];
    str.erase(0,i1+1); // erase ident

    string host;
    for(i1=0; i1<str.length(); ++i1)
        if(str[i1]==0x20)
            break;
            else
            host+=str[i1];
    str.erase(0,i1+1); // erase host name

    for(i1=0; i1<str.length(); ++i1)
        if(str[i1]==0x20)
            break;
    str.erase(0,i1+1); // erase "TOPIC"

    string channel;
    for(i1=0; i1<str.length(); ++i1)
        if(str[i1]==0x20)
            break;
            else
            channel+=str[i1];
    str.erase(0,i1+1); // erase channel

    for(i1=0; i1<str.length(); ++i1)
        if(str[i1]==':')
            break;
    str.erase(0,i1+1); // erase everything before ':'

    string topic;
    for(i1=0; i1<str.length(); ++i1)
        topic+=str[i1];

    {
        c_online_user source=irc_find_user(nick);

        string source_user, user_nick=nick;
        if(ident.empty() && host.empty()) {
            source_user=nick; // it is a server
            user_nick="";
        } else
            source_user=logic_find_user(source.nick,source.ident,source.host,source.fullname,source.irc_op);

        {
            mod_dispatch("topic@channel",source_user,channel,"",true,0,"channel",channel.c_str(),"source",source_user.c_str(),"source_nick",nick.c_str(),"topic",topic.c_str(),NULL,"nick",nick.c_str(),"origin","@irc@","ident",irc_get_ident(nick).c_str(),"host",irc_get_host(nick).c_str(),NULL);
        }

        global_keeper.change_topic(channel,topic);

        logic_validate_topic(channel,source_user,user_nick,topic);
    }
}

/*!
    \brief Called after receiving RPL_WHOREPLY from server
    \author VooDooMan
    \version 1
    \date 2008-01-20
    \param str Partial message from server
*/
MY_STATIC void irc_RPL_WHOREPLY(string str)
{
    // WHO reply: "<channel> <user> <host> <server> <nick> <H|G>[*][@|+] :<hopcount> <real name>"
    BEGIN_DEBUG_TIME_MEASUREMENT(true,1,"processing who");

    size_t i1;

    // erase numeric reply
    if((i1=str.find(0x20))!=string::npos)
        str.erase(0,i1+1);

    string my_nick;
    for(i1=0; i1<str.length(); ++i1)
        if(str[i1]==0x20)
            break;
        else
            my_nick+=str[i1];
    irc_nick=my_nick;
    str.erase(0,i1+1); // erase my nick

    string channel;
    for(i1=0; i1<str.length(); ++i1)
        if(str[i1]==0x20)
            break;
        else
            channel+=str[i1];
    str.erase(0,i1+1); // erase nick name

    string ident;
    for(i1=0; i1<str.length(); ++i1)
        if(str[i1]==0x20)
            break;
        else
            ident+=str[i1];
    str.erase(0,i1+1); // erase ident

    string host;
    for(i1=0; i1<str.length(); ++i1)
        if(str[i1]==0x20)
            break;
        else
            host+=str[i1];
    str.erase(0,i1+1); // erase host

    for(i1=0; i1<str.length(); ++i1)
        if(str[i1]==0x20)
            break;
    str.erase(0,i1+1); // erase server

    string nick;
    for(i1=0; i1<str.length(); ++i1)
        if(str[i1]==0x20)
            break;
        else
            nick+=str[i1];
    str.erase(0,i1+1); // erase host

    string mode;
    for(i1=0; i1<str.length(); ++i1)
        if(str[i1]==0x20)
            break;
        else
            mode+=str[i1];
    str.erase(0,i1+1); // erase mode

    for(i1=0; i1<str.length(); ++i1)
        if(str[i1]==0x20)
            break;
    str.erase(0,i1+1); // erase hop count

    string fullname=str;

    irc_waiting_for_dns.push_back(c_dns_wait(nick,ident,host,fullname));

    vector<c_online_channel>::iterator i;
    for(i=irc_channels.begin(); i!=irc_channels.end(); ++i) {
        vector<c_online_user>::iterator ii;
        for(ii=(*i).users.begin(); ii!=(*i).users.end(); ++ii) {
            if(!irc_cmp_strings_case_insensitive((*ii).nick,nick)) {

                (*ii).got_whois=true;

                BEGIN_DEBUG_TIME_MEASUREMENT(true,2,"#2");
                global_keeper.add_user_whois1(nick,ident,host,fullname);
                if(mode.find("*")!=string::npos)
                    global_keeper.add_user_ircop(nick);
                END_DEBUG_TIME_MEASUREMENT(2,0);

                break;
            }
        }
    }

    if(!irc_cmp_strings_case_insensitive(nick,irc_nick)) {
        // after re-JOIN to get op
        vector<c_online_channel>::iterator i1;
        for(i1=irc_channels.begin(); i1!=irc_channels.end(); ++i1) {
            vector<c_online_user>::iterator i2;
            for(i2=(*i1).users.begin(); i2!=(*i1).users.end(); ++i2) {
                if(!(*i2).nick.compare(irc_nick)) {
                    if(!(*i2).got_whois)
                        break;
                    if((*i2).nick_validated)
                        break;

                    BEGIN_DEBUG_TIME_MEASUREMENT(true,8,"#8");

                    (*i2).nick_validated=true;
                    logic_on_nick_validate(irc_nick,irc_get_ident(irc_nick),irc_get_host(irc_nick),irc_get_fullname(irc_nick),(*i1).name,irc_is_ircop(irc_nick),true,(*i2).mod_disp_join_called);
                    (*i2).on_join_called=true; // !!!

                    END_DEBUG_TIME_MEASUREMENT(8,0);

                    break;
                }
            }
        }
    }

    END_DEBUG_TIME_MEASUREMENT(1,0);
}

/*!
    \brief Called when user was KICKed and validates it (or calls on_kicked event)
    \author VooDooMan
    \version 1
    \date 2004
    \param str Partial message from server containing channel name and nick of kicked user
*/
MY_STATIC void irc_kicked(string str)
{
    string log=str;
    unsigned int i1;

    str.erase(0,1); // erase first ':'

    string nick;
    for(i1=0; i1<str.length(); ++i1)
        if(str[i1]=='!')
            break;
            else
            nick+=str[i1];
    str.erase(0,i1+1); // erase nick name

    string ident;
    for(i1=0; i1<str.length(); ++i1)
        if(str[i1]=='@')
            break;
            else
            ident+=str[i1];
    str.erase(0,i1+1); // erase ident

    string host;
    for(i1=0; i1<str.length(); ++i1)
        if(str[i1]==0x20)
            break;
            else
            host+=str[i1];
    str.erase(0,i1+1); // erase host name

    for(i1=0; i1<str.length(); ++i1)
        if(str[i1]==0x20)
            break;
    str.erase(0,i1+1); // erase "KICK"

    string channel;
    for(i1=0; i1<str.length(); ++i1)
        if(str[i1]==0x20)
            break;
            else
            channel+=str[i1];
    str.erase(0,i1+1); // erase channel

    string target;
    for(i1=0; i1<str.length(); ++i1)
        if(str[i1]==0x20)
            break;
            else
            target+=str[i1];
    str.erase(0,i1+1); // erase target

    c_online_user target_=irc_find_user(target);
    string target_user=logic_find_user(target_.nick,target_.ident,target_.host,target_.fullname,target_.irc_op);

    {
        c_online_user source=irc_find_user(nick);

        string source_user, user_nick=nick;
        if(ident.empty() && host.empty()) {
            source_user=nick; // it is a server
            user_nick="";
        } else
            source_user=logic_find_user(source.nick,source.ident,source.host,source.fullname,source.irc_op);

        mod_dispatch("kick@channel",source_user,channel,"",true,0,"channel",channel.c_str(),"source",source_user.c_str(),"target",target_user.c_str(),"source_nick",nick.c_str(),"target_nick",target.c_str(),"target_ident",irc_get_ident(target).c_str(),"target_host",irc_get_host(target).c_str(),NULL,"nick",nick.c_str(),"origin","@irc@","ident",irc_get_ident(nick).c_str(),"host",irc_get_host(nick).c_str(),NULL);

        global_keeper.nick_left(channel,target,!irc_nick.compare(target));

        logic_validate_kick(channel,source_user,target_user,user_nick,target);
    }
}

/*!
    \brief Compares the beginning of message
    \author VooDooMan
    \version 1
    \date 2004
    \param start The beginning of the message to be compared with last received message
    \return Returns true if it mathes
*/
MY_STATIC bool irc_check_last_msg(const char* start)
{
    return last_msg.m_msg.get_msg().find(start,0)==0;
}

/*!
    \brief Clears wait flag for sent message (reply has been received)
    \author VooDooMan
    \version 1
    \date 2004
    \param resp Received message from server
*/
MY_STATIC void irc_kill_last_msg_wait(const string& resp)
{
    size_t n=resp.length();
    if(n>sizeof(last_msg.response)-1)
        n=sizeof(last_msg.response)-1;
    strncpy(last_msg.response,resp.c_str(),n);
    last_msg.response[n]=0;
    last_msg.wait_for_response=false;
}

/*!
    \brief Clears wait flag for sent command (reply has been received)
    \author VooDooMan
    \version 1
    \date 2005
    \param command Command that should be cleared
*/
MY_STATIC void irc_kill_command_wait(const string& command)
{
    char tmp[128];
    unsigned int i1;
    for(i1=0; i1<64 && i1<command.length() && command[i1]!=0x20; ++i1)
        tmp[i1]=command[i1];
    tmp[i1]=0;

    vector<c_awaiting_response>::iterator i2;
    for(i2=waiting_for_response.begin(); i2!=waiting_for_response.end(); ++i2) {
        string cmd;
        for(unsigned int i3=0; i3<(*i2).message.length() && (*i2).message[i3]!=0x20; ++i3)
            cmd+=(*i2).message[i3];

        if(!cmd.compare(tmp)) {
            waiting_for_response.erase(i2);
            break;
        }
    }
}

/*!
    \brief Builds vector of string from flood history
    \author VooDooMan
    \version 1
    \date 2005-12-03
    \param hist Flood history
*/
MY_STATIC vector<string> irc_build_flood_list(const vector<c_flood_history>& hist)
{
    vector<string> flist;
    vector<c_flood_history>::const_iterator i1;
    for(i1=hist.begin(); i1!=hist.end(); ++i1)
        flist.push_back((*i1).msg);
    return flist;
}

/*!
    \brief Stores data for repeat flood detection
    \author VooDooMan
    \version 1
    \date 2004
*/
class c_log
{
public:
    string msg;                 //!< message
    unsigned int num;           //!< Number of its occurence
    my_time_t time;                //!< Time of occurence
};

/*!
    \brief Checks flood limits and history for just received message
    \author VooDooMan
    \version 1
    \date 2004
    \param type Type of message - one of following: TYPE_PRIVMSG, TYPE_NOTICE, TYPE_QUIT_PART_KICK, TYPE_JOIN, TYPE_NICK, TYPE_MODE, TYPE_CTCP
    \param channel Name of channel
    \param full_host Host of originator of message ("nick!ident@host")
    \param msg Additional "sub-message": for TYPE_PRIVMSG whole message, for TYPE_NOTICE notice message, for TYPE_QUIT_PART_KICK quit message/part message/kick reason (if any), for TYPE_NICK new nick name, and for TYPE_MODE mode string (parameter to MODE command), and for TYPE_MODE, when topic was changed, there is new topic, for TYPE_CTCP - the whole PRIVMSG
*/
MY_STATIC void irc_check_flood(int type, const string& channel, const string& full_host, const string& msg)
{
    string str=full_host;

    c_online_user* user=NULL;

    string nick;
    unsigned int i1;
    for(i1=0; i1<str.length(); ++i1)
        if(str[i1]=='!')
            break;
            else
            nick+=str[i1];
    str.erase(0,i1+1); // erase nick name

    if(!irc_cmp_strings_case_insensitive(nick,irc_nick))
        // it's me! the bot!
        return;

    if(type==TYPE_NICK && !irc_cmp_strings_case_insensitive(msg,irc_nick))
        // it's me while changing nick!
        return;

    string ident;
    for(i1=0; i1<str.length(); ++i1)
        if(str[i1]=='@')
            break;
            else
            ident+=str[i1];
    str.erase(0,i1+1); // erase ident

    string host;
    for(i1=0; i1<str.length(); ++i1)
        if(str[i1]==0x20)
            break;
            else
            host+=str[i1];
    str.erase(0,i1+1); // erase host

    bool got_user=false;
    vector<c_online_channel>::iterator i2;
    for(i2=irc_channels.begin(); i2!=irc_channels.end(); ++i2) {
        if(irc_cmp_strings_case_insensitive((*i2).name,channel))
            continue;
        vector<c_online_user>::iterator i3;
        for(i3=(*i2).users.begin(); i3!=(*i2).users.end(); ++i3) {
            if(!irc_cmp_strings_case_insensitive((*i3).nick,nick)) {
                user=&(*i3);
                got_user=true;
                break;
            }
        }
        if(got_user)
            break;
    }

    if(!got_user && TYPE_JOIN!=type) {
        log_irc_ex(E_LOG_IRC_EX_INF,"User is not in channel, ignoring possible flood rules violation");
        return;
    }

    bool user_gone=false;

    {
        // ensure nick is still in channel

        bool got_nick=false;
        for(vector<c_online_channel>::iterator i2=irc_channels.begin(); i2!=irc_channels.end(); ++i2) {
            vector<c_online_user>::iterator i3;
            for(i3=(*i2).users.begin(); i3!=(*i2).users.end(); ++i3) {
                if(!irc_cmp_strings_case_insensitive((*i3).nick,nick)) {
                    got_nick=true;
                }
            }
            if(got_nick)
                break;
        }
        if(!got_nick)
            return;

        bool rejoined=false;

        for(list<string>::const_iterator i3=msgs_from_server.begin(); i3!=msgs_from_server.end(); ++i3) {
            string str=*i3;
            if(str.empty())
                continue;
            if(':'!=str[0])
                continue;
            str.erase(0,1);

            string::size_type i1;

            string nick2;
            for(i1=0; i1<str.length(); ++i1)
                if(str[i1]=='!')
                    break;
                 else
                    nick2+=str[i1];
            str.erase(0,i1+1); // erase nick name

            for(i1=0; i1<str.length(); ++i1)
                if(str[i1]=='@')
                    break;
            str.erase(0,i1+1); // erase ident

            for(i1=0; i1<str.length(); ++i1)
                if(str[i1]==0x20)
                    break;
            str.erase(0,i1+1); // erase host name

            string cmd;
            for(i1=0; i1<str.length(); ++i1)
                if(str[i1]==0x20)
                    break;
                else
                    cmd+=str[i1];
            str.erase(0,i1+1); // erase "KICK"

            string chan;
            for(i1=0; i1<str.length(); ++i1)
                if(str[i1]==0x20)
                    break;
                else
                    chan+=str[i1];
            str.erase(0,i1+1); // erase channel

            if(!chan.empty() && ':'==chan[0])
                chan.erase(0,1); // make it universal for bothe "KICK/PART/QUIT" and "join"

            string param;
            for(i1=0; i1<str.length(); ++i1)
                if(str[i1]==0x20)
                    break;
                else
                    param+=str[i1];
            str.erase(0,i1+1); // erase nick

            if(!cmd.compare("KICK") || !cmd.compare("PART") || !cmd.compare("QUIT")) {
                if(!cmp_strings_case_insensitive(channel,chan) && !cmp_strings_case_insensitive(nick,param)) {
                    if(NULL!=user && !rejoined) {
                        user->msg_flood_hist.clear();
                        user->notice_flood_hist.clear();
                        user->repeat_flood_hist.clear();
                        user->nick_flood_hist.clear();
                        // user->join_flood_hist // do NOT clear join flood history!
                        user->mode_flood_hist.clear();
                        user->ctcp_flood_hist.clear();
                        user_gone=true;
                    }
                }
            }
            if(!cmd.compare("JOIN")) { // check if user rejoined the channel
                if(!cmp_strings_case_insensitive(channel,chan) && !cmp_strings_case_insensitive(nick,nick2)) {
                    rejoined=true;
                    break;
                }
            }
        }
    }

    if(user_gone)
        return;

    // check for myself
    {
        bool match=false;
        deque<c_user>::iterator i1;
        map<string,string> vars;
        for(i1=r_users.begin(); i1!=r_users.end(); ++i1) {
            if(!(*i1).host_bot)
                continue;
            vector<string>::iterator i2;
            for(i2=(*i1).hostmask.begin(); i2!=(*i1).hostmask.end(); ++i2) {
                bool was_dns=true;
                if(logic_check_mask(*i2,nick+"!"+ident+"@"+host,"*","*",was_dns)) {
                    match=true;
                    break;
                }
            }
            break;
        }
        if(match)
            return;
    }

    if(!got_user) {
        // type gone users (join/quit flood)
        vector<c_online_user>::iterator i3;
        for(i3=gone_users.begin(); i3!=gone_users.end(); ++i3) {
            if(!irc_cmp_strings_case_insensitive((*i3).nick,nick)) {
                user=&(*i3);
                got_user=true;
                break;
            }
        }
    }

    my_time_t now;
    my_time(&now);

    c_flood msg_flood;
    c_flood notice_flood;
    c_flood repeat_flood;
    c_flood nick_flood;
    c_flood join_flood;
    c_flood mode_flood;
    c_flood ctcp_flood;
    string username;
    if(!got_user || user==NULL || !user->got_whois) {
        // we don't got user (host_unknown not defined) or we don't have WHOIS yet
        return;
    } else {
        username=user->in_logic_as;
    }

    logic_get_flood_limits(channel,username,msg_flood,notice_flood,repeat_flood,nick_flood,join_flood,mode_flood,ctcp_flood);

    // clean up flood buffers
    vector<c_flood_history>::iterator i4;
l4:
    for(i4=user->msg_flood_hist.begin(); i4!=user->msg_flood_hist.end(); ++i4) {
        if(now-(*i4).time>(my_time_t)msg_flood.seconds) {
            user->msg_flood_hist.erase(i4);
            goto l4;
        }
    }
l5:
    for(i4=user->notice_flood_hist.begin(); i4!=user->notice_flood_hist.end(); ++i4) {
        if(now-(*i4).time>(my_time_t)notice_flood.seconds) {
            user->notice_flood_hist.erase(i4);
            goto l5;
        }
    }
l6:
    for(i4=user->repeat_flood_hist.begin(); i4!=user->repeat_flood_hist.end(); ++i4) {
        if(now-(*i4).time>(my_time_t)repeat_flood.seconds) {
            user->repeat_flood_hist.erase(i4);
            goto l6;
        }
    }
l7:
    for(i4=user->nick_flood_hist.begin(); i4!=user->nick_flood_hist.end(); ++i4) {
        if(now-(*i4).time>(my_time_t)nick_flood.seconds) {
            user->nick_flood_hist.erase(i4);
            goto l7;
        }
    }
l8:
    for(i4=user->join_flood_hist.begin(); i4!=user->join_flood_hist.end(); ++i4) {
        if(now-(*i4).time>(my_time_t)join_flood.seconds) {
            user->join_flood_hist.erase(i4);
            goto l8;
        }
    }
l9:
    for(i4=user->mode_flood_hist.begin(); i4!=user->mode_flood_hist.end(); ++i4) {
        if(now-(*i4).time>(my_time_t)mode_flood.seconds) {
            user->mode_flood_hist.erase(i4);
            goto l9;
        }
    }
l10:
    for(i4=user->ctcp_flood_hist.begin(); i4!=user->ctcp_flood_hist.end(); ++i4) {
        if(now-(*i4).time>(my_time_t)ctcp_flood.seconds) {
            user->ctcp_flood_hist.erase(i4);
            goto l10;
        }
    }

    c_flood_history entry;
    entry.time=now;
    entry.bytes=msg.length();
    entry.msg=msg;
    if(type==TYPE_JOIN)
        entry.msg=channel;

    bool got2=false;
    if(type==TYPE_PRIVMSG) {
        user->msg_flood_hist.push_back(entry);
        user->repeat_flood_hist.push_back(entry);
        unsigned int entries=0;
        my_time_t low_time=0;
        for(i4=user->msg_flood_hist.begin(); i4!=user->msg_flood_hist.end(); ++i4) {
            entries++;
            if(low_time==0 || low_time>(*i4).time)
                low_time=(*i4).time;
        }
        my_time_t diff=my_time(NULL)-low_time;
        if(msg_flood.lines!=0 && entries>=msg_flood.lines) {
            vector<string> flist=irc_build_flood_list(user->msg_flood_hist);
            logic_on_flood(channel,username,nick,TYPE_PRIVMSG,entries,diff,msg_flood.seconds,flist);
            got2=true;
        }
    }
    if(type==TYPE_NOTICE) {
        user->notice_flood_hist.push_back(entry);
        user->repeat_flood_hist.push_back(entry);
        unsigned int entries=0;
        my_time_t low_time=0;
        for(i4=user->notice_flood_hist.begin(); i4!=user->notice_flood_hist.end(); ++i4) {
            entries++;
            if(low_time==0 || low_time>(*i4).time)
                low_time=(*i4).time;
        }
        my_time_t diff=my_time(NULL)-low_time;
        if(!got2 && notice_flood.lines!=0 && entries>=notice_flood.lines) {
            vector<string> flist=irc_build_flood_list(user->notice_flood_hist);
            logic_on_flood(channel,username,nick,TYPE_NOTICE,entries,diff,notice_flood.seconds,flist);
            got2=true;
        }
    }
    if(type==TYPE_QUIT_PART_KICK) {
        user->join_flood_hist.push_back(entry);
        //user->repeat_flood_hist.push_back(entry);
    }
    if(type==TYPE_JOIN) {
        user->join_flood_hist.push_back(entry);
        unsigned int entries=0;
        my_time_t low_time=0;
        for(i4=user->join_flood_hist.begin(); i4!=user->join_flood_hist.end(); ++i4) {
            if(!irc_cmp_strings_case_insensitive((*i4).msg,channel)) {
                entries++;
                if(low_time==0 || low_time>(*i4).time)
                    low_time=(*i4).time;
            }
        }
        my_time_t diff=my_time(NULL)-low_time;
        if(!got2 && join_flood.lines!=0 && entries>=join_flood.lines) {
            vector<string> flist=irc_build_flood_list(user->join_flood_hist);
            logic_on_flood(channel,username,nick,TYPE_JOIN,entries,diff,join_flood.seconds,flist);
            got2=true;
        }
    }
    if(type==TYPE_NICK) {
        user->nick_flood_hist.push_back(entry);
        user->repeat_flood_hist.push_back(entry);
        unsigned int entries=0;
        my_time_t low_time=0;
        for(i4=user->nick_flood_hist.begin(); i4!=user->nick_flood_hist.end(); ++i4) {
            entries++;
            if(low_time==0 || low_time>(*i4).time)
                low_time=(*i4).time;
        }
        my_time_t diff=my_time(NULL)-low_time;
        if(!got2 && nick_flood.lines!=0 && entries>=nick_flood.lines) {
            vector<string> flist=irc_build_flood_list(user->nick_flood_hist);
            logic_on_flood(channel,username,nick,TYPE_NICK,entries,diff,nick_flood.seconds,flist);
            got2=true;
        }
    }
    if(type==TYPE_MODE) {
        user->mode_flood_hist.push_back(entry);
        user->repeat_flood_hist.push_back(entry);
        unsigned int entries=0;
        my_time_t low_time=0;
        for(i4=user->mode_flood_hist.begin(); i4!=user->mode_flood_hist.end(); ++i4) {
            entries++;
            if(low_time==0 || low_time>(*i4).time)
                low_time=(*i4).time;
        }
        my_time_t diff=my_time(NULL)-low_time;
        if(mode_flood.lines!=0 && entries>=mode_flood.lines) {
            vector<string> flist=irc_build_flood_list(user->mode_flood_hist);
            logic_on_flood(channel,username,nick,TYPE_MODE,entries,diff,mode_flood.seconds,flist);
            got2=true;
        }
    }
    if(type==TYPE_CTCP) {
        user->ctcp_flood_hist.push_back(entry);
        user->repeat_flood_hist.push_back(entry);
        unsigned int entries=0;
        my_time_t low_time=0;
        for(i4=user->ctcp_flood_hist.begin(); i4!=user->ctcp_flood_hist.end(); ++i4) {
            entries++;
            if(low_time==0 || low_time>(*i4).time)
                low_time=(*i4).time;
        }
        my_time_t diff=my_time(NULL)-low_time;
        if(ctcp_flood.lines!=0 && entries>=ctcp_flood.lines) {
            vector<string> flist=irc_build_flood_list(user->ctcp_flood_hist);
            logic_on_flood(channel,username,nick,TYPE_CTCP,entries,diff,ctcp_flood.seconds,flist);
            got2=true;
        }
    }
    // evaluate repeat
    if(type==TYPE_PRIVMSG || type==TYPE_NOTICE || type==TYPE_NICK || type==TYPE_CTCP) {
        vector<c_log> repeat_log;
        vector<c_log>::iterator i5;
        my_time_t low_time=0;
        for(i4=user->repeat_flood_hist.begin(); i4!=user->repeat_flood_hist.end(); ++i4) {
            bool got=false;
            for(i5=repeat_log.begin(); i5!=repeat_log.end(); ++i5)
                if(!(*i5).msg.compare((*i4).msg)) {
                    got=true;
                    (*i5).num++;
                    if(low_time==0 || low_time>(*i5).time)
                        low_time=(*i5).time;
                    break;
                }
            if(!got) {
                c_log e;
                e.msg=(*i4).msg;
                e.num=1;
                e.time=(*i4).time;
                repeat_log.push_back(e);
            }
        }
        my_time_t diff=my_time(NULL)-low_time;
        for(i5=repeat_log.begin(); i5!=repeat_log.end(); ++i5)
            if(!got2 && repeat_flood.lines!=0 && (*i5).num>=repeat_flood.lines) {
                vector<string> flist=irc_build_flood_list(user->repeat_flood_hist);
                logic_on_flood(channel,username,nick,TYPE_REPEAT,(*i5).num,diff,repeat_flood.seconds,flist);
                got2=true;
                break;
            }
    }

    // clean up gone_users
    vector<c_online_user>::iterator i3;
l1:
    for(i3=gone_users.begin(); i3!=gone_users.end(); ++i3) {
        if((*i3).quit_time+120<now) {
            // 120 seconds ought be enough!
            gone_users.erase(i3);
            goto l1;
        }
    }
    if(type==TYPE_QUIT_PART_KICK) {
        bool got=false;
        for(i3=gone_users.begin(); i3!=gone_users.end(); ++i3)
            if(!(*i3).nick.compare(nick)) {
                got=true;
                break;
            }
        if(!got) {
            my_time(&user->quit_time);
            gone_users.push_back(*user);
        }
    }
}

/*!
    \brief Raises on_privmsg event if defined for specific user and channel
    \author VooDooMan
    \version 1
    \date 2004
    \param channel Name of channel
    \param full_host Host of originator ("nick!ident@host")
    \param msg The message (PRIVMSG message)
*/
MY_STATIC void irc_on_privmsg(const string& channel, const string& full_host, const string& msg)
{
    string str=full_host;

    string nick;
    unsigned int i1;
    for(i1=0; i1<str.length(); ++i1)
        if(str[i1]=='!')
            break;
            else
            nick+=str[i1];
    str.erase(0,i1+1); // erase nick name

    if(!irc_cmp_strings_case_insensitive(nick,irc_nick)) // it's me! the bot!
        return;

    string ident;
    for(i1=0; i1<str.length(); ++i1)
        if(str[i1]=='@')
            break;
            else
            ident+=str[i1];
    str.erase(0,i1+1); // erase ident

    string host;
    for(i1=0; i1<str.length(); ++i1)
        if(str[i1]==0x20)
            break;
            else
            host+=str[i1];
    str.erase(0,i1+1); // erase host

    c_online_user* user=NULL;

    bool got_user=false;
    vector<c_online_channel>::iterator i2;
    for(i2=irc_channels.begin(); i2!=irc_channels.end(); ++i2) {
        vector<c_online_user>::iterator i3;
        for(i3=(*i2).users.begin(); i3!=(*i2).users.end(); ++i3) {
            if(!irc_cmp_strings_case_insensitive((*i3).nick,nick)) {
                user=&(*i3);
                got_user=true;
                break;
            }
        }
        if(got_user)
            break;
    }

    my_time_t now;
    my_time(&now);

    string username;
    if(!got_user) {
        // assume unknown user
        username="";

        vector<c_online_channel>::iterator i1;
    l3:
        bool got_channel=false;
        for(i1=unknown_users_on_channels.begin(); i1!=unknown_users_on_channels.end(); ++i1)
            if(!irc_cmp_strings_case_insensitive((*i1).name,channel)) {
                user=&(*(*i1).users.begin()); // just first record only
                got_channel=true;
                break;
            }
        if(!got_channel) {
            c_online_channel ch(channel);
            c_online_user u;
            u.old_mode="INVALID";
            u.on_join_passive=true;
            u.on_join_called2=false;
            ch.users.push_back(u);
            unknown_users_on_channels.push_back(ch);
            goto l3;
        }
    } else {
        username=user->in_logic_as;
    }

    if(!channel.empty())
        logic_on_privmsg(channel,username,nick,msg);
    else
        logic_on_privmsg_query(username,nick,msg);
}

/*!
    \brief Raises on_broadcast event
    \author VooDooMan
    \version 1
    \date 2005-04-22
    \param bcast_mask Broadcast mask ("$$irc.someserver.net")
    \param type Type: TYPE_PRIVMSG, TYPE_NOTICE
    \param full_host Host of originator ("nick!ident@host")
    \param msg The message
    \param server IRC server host we are on
    \param port IRC server port we are on
*/
MY_STATIC void irc_on_broadcast(const string& bcast_mask, int type, const string& full_host, const string& msg, const string& server, unsigned short port)
{
    string str=full_host;

    string nick;
    unsigned int i1;
    for(i1=0; i1<str.length(); ++i1)
        if(str[i1]=='!')
            break;
            else
            nick+=str[i1];
    str.erase(0,i1+1); // erase nick name

    string ident;
    for(i1=0; i1<str.length(); ++i1)
        if(str[i1]=='@')
            break;
            else
            ident+=str[i1];
    str.erase(0,i1+1); // erase ident

    string host;
    for(i1=0; i1<str.length(); ++i1)
        if(str[i1]==0x20)
            break;
            else
            host+=str[i1];
    str.erase(0,i1+1); // erase host

    c_online_user* user=NULL;

    bool got_user=false;
    vector<c_online_channel>::iterator i2;
    for(i2=irc_channels.begin(); i2!=irc_channels.end(); ++i2) {
        vector<c_online_user>::iterator i3;
        for(i3=(*i2).users.begin(); i3!=(*i2).users.end(); ++i3) {
            if(!irc_cmp_strings_case_insensitive((*i3).nick,nick)) {
                user=&(*i3);
                got_user=true;
                break;
            }
        }
        if(got_user)
            break;
    }

    my_time_t now;
    my_time(&now);

    string username;
    if(!got_user) {
        // assume unknown user
        username="";
    } else {
        username=user->in_logic_as;
    }

    logic_on_broadcast(username,nick,ident,host,bcast_mask,type,msg,server,port);
}

/*!
    \brief Raises on_notice event if defined for specific user and channel
    \author VooDooMan
    \version 1
    \date 2004
    \param channel Name of channel
    \param full_host Host of originator ("nick!ident@host")
    \param msg The message (NOTICE message)
*/
MY_STATIC void irc_on_notice(const string& channel, const string& full_host, const string& msg)
{
    string str=full_host;

    string nick;
    unsigned int i1;
    for(i1=0; i1<str.length(); ++i1)
        if(str[i1]=='!')
            break;
            else
            nick+=str[i1];
    str.erase(0,i1+1); // erase nick name

    if(!irc_cmp_strings_case_insensitive(nick,irc_nick)) // it's me! the bot!
        return;

    string ident;
    for(i1=0; i1<str.length(); ++i1)
        if(str[i1]=='@')
            break;
            else
            ident+=str[i1];
    str.erase(0,i1+1); // erase ident

    string host;
    for(i1=0; i1<str.length(); ++i1)
        if(str[i1]==0x20)
            break;
            else
            host+=str[i1];
    str.erase(0,i1+1); // erase host

    c_online_user* user=NULL;

    bool got_user=false;
    vector<c_online_channel>::iterator i2;
    for(i2=irc_channels.begin(); i2!=irc_channels.end(); ++i2) {
        vector<c_online_user>::iterator i3;
        for(i3=(*i2).users.begin(); i3!=(*i2).users.end(); ++i3) {
            if(!irc_cmp_strings_case_insensitive((*i3).nick,nick)) {
                user=&(*i3);
                got_user=true;
                break;
            }
        }
        if(got_user)
            break;
    }

    my_time_t now;
    my_time(&now);

    string username;
    if(!got_user) {
        // assume unknown user
        username="";

        vector<c_online_channel>::iterator i1;
    l3:
        bool got_channel=false;
        for(i1=unknown_users_on_channels.begin(); i1!=unknown_users_on_channels.end(); ++i1)
            if(!irc_cmp_strings_case_insensitive((*i1).name,channel)) {
                user=&(*(*i1).users.begin()); // just first record only
                got_channel=true;
                break;
            }
        if(!got_channel) {
            c_online_channel ch(channel);
            c_online_user u;
            u.old_mode="INVALID";
            u.on_join_passive=true;
            u.on_join_called2=false;
            ch.users.push_back(u);
            unknown_users_on_channels.push_back(ch);
            goto l3;
        }
    } else {
        username=user->in_logic_as;
    }

    logic_on_notice(channel,username,nick,msg);
}

/*!
    \brief This is called in loop to process messages from server
    \author VooDooMan
    \version 2
    \date 2004, 2007-07-27
    \retval 0 Okay
    \retval 1 Error occured and bot should reconnect to server
    \retval 2 Throttled - should be sleep for a couple of seconds
*/
int irc_loop_process_input()
{
    global_keeper.update_uptime();

    irc_process_nicks();

    if(!irc_host.empty())
        dcc_update_dynamic_servers(irc_host);

    my_time_t now;
    my_time(&now);
    vector<c_online_user>::iterator ii;
_label_1:
    for(ii=tmp_users.begin(); ii!=tmp_users.end(); ++ii) {
        if((*ii).last_whois+10<=now) {
            tmp_users.erase(ii);
            goto _label_1;
        }
        in_addr addr4;
        in_addr6_ addr6;
        memset(&addr4,0,sizeof(addr4));
        memset(&addr6,0,sizeof(addr6));
        if((*ii).got_whois) {
            if(!(*ii).resolv_host.empty()) {
                if(4==(*ii).ip_version) {
                    e_resolv_result res=threads_resolve4((*ii).resolv_host.c_str(),addr4,RT_RESOLVE_SCHEDULE);
                    if(RR_FAILURE==res) {
                        tmp_users.erase(ii);
                        goto _label_1;
                    }
                    if(RR_SCHEDULED==res) {
                        continue;
                    }
                }
                if(6==(*ii).ip_version) {
                    e_resolv_result res=threads_resolve6((*ii).resolv_host.c_str(),addr6,RT_RESOLVE_SCHEDULE);
                    if(RR_FAILURE==res) {
                        tmp_users.erase(ii);
                        goto _label_1;
                    }
                    if(RR_SCHEDULED==res) {
                        continue;
                    }
                }
            }
            string hm=(*ii).nick;
            hm+="!";
            hm+=(*ii).ident;
            hm+="@";
            hm+=(*ii).host;
            string user_name_as_in_logic;
            if(irc_access_to_partyline(hm.c_str(),user_name_as_in_logic,(*ii).dcc_msg)) {
                int dcc_group=0;
                int ip_ver;
                string dcc=dcc_get_string((*ii).dcc_msg,dcc_group,ip_ver);
                //string dcc_host=logic_get_dcc_host(dcc_group,user_name_as_in_logic);
                if(dcc.compare("")) {
                    int ip_version=ip_ver;//sock_get_address_type(dcc_host);
                    if(ip_version) {
                        if((*ii).resolv_host.empty()) {
                            (*ii).resolv_host=(*ii).host;
                            (*ii).ip_version=ip_version;
                            goto next_tmp;
                        }
                    } else {
                        tmp_users.erase(ii);
                        goto _label_1;
                    }

                    c_dcc_awaiting_user user;
                    string str="\x01";
                    str+=dcc;
                    str+="\x01";
                    irc_privmsg((*ii).nick.c_str(),str.c_str(),LOW_PRIORITY);

                    user.nick=(*ii).nick;
                    user.ident=irc_get_ident((*ii).nick);
                    user.host=irc_get_host((*ii).nick);
                    user.fullname=irc_get_fullname((*ii).nick);

                    size_t n=user_name_as_in_logic.length();
                    if(n>128-1)
                        n=128-1;
                    strncpy(user.user_name_as_in_logic,user_name_as_in_logic.c_str(),n);
                    user.user_name_as_in_logic[n]=0;
                    user.using_ipv6= ip_version==6;
                    user.addr4=addr4;
                    user.addr6=addr6;
                    my_time(&user.start_time);
                    dcc_awaiting_users.push_back(user);
                }
            }
            tmp_users.erase(ii);
            goto _label_1;
        }
    next_tmp:
        continue;
    }

    {
        vector<c_dcc_request_whois>::iterator i2;
        for(i2=dcc_request_whois.begin(); i2!=dcc_request_whois.end(); ++i2) {
            if((*i2).only_dns && !(*i2).dns_done) {
                bool ok=true;
                if(!(*i2).resolv4_1.empty()) {
                    in_addr ip;
                    e_resolv_result res1=threads_resolve4((*i2).resolv4_1.c_str(),ip,RT_RESOLVE_SCHEDULE);
                    if(RR_SCHEDULED==res1)
                        ok=false;
                }
                if(!(*i2).resolv4_2.empty()) {
                    in_addr ip;
                    e_resolv_result res1=threads_resolve4((*i2).resolv4_2.c_str(),ip,RT_RESOLVE_SCHEDULE);
                    if(RR_SCHEDULED==res1)
                        ok=false;
                }
                if(!(*i2).resolv6_1.empty()) {
                    in_addr6_ ip;
                    e_resolv_result res1=threads_resolve6((*i2).resolv6_1.c_str(),ip,RT_RESOLVE_SCHEDULE);
                    if(RR_SCHEDULED==res1)
                        ok=false;
                }
                if(!(*i2).resolv6_2.empty()) {
                    in_addr6_ ip;
                    e_resolv_result res1=threads_resolve6((*i2).resolv6_2.c_str(),ip,RT_RESOLVE_SCHEDULE);
                    if(RR_SCHEDULED==res1)
                        ok=false;
                }
                if(ok) {
                    (*i2).dns_done=true;
                    irc_quoted_callback((*i2).str.c_str(),(*i2).nick.c_str(),(*i2).host_mask.c_str(),(*i2).comment.c_str());
                }
                break;
            }
        }
    }

    list<string>::iterator i;

l1:
    i=msgs_from_server.begin();
    if(i!=msgs_from_server.end()) {
        {
            my_sleep(-1);

            enum {
                CNT_TO_PUT_TO_IRC_SERVER = 10U,
                CNT_HOW_MANY_TIMES = 100,
            };
            static unsigned cnt=0;
            if((cnt++) % CNT_TO_PUT_TO_IRC_SERVER == 0) {
                for(int i1=0; i1<CNT_HOW_MANY_TIMES; ++i1)
                    irc_loop_putserv();
                irc_loop_heavy_time_consuming_operation("Still executing script...");
                my_sleep(-1);

                goto l1;
            }
        }

        irc_last_keepalive_detection=my_time(NULL);

        log_irc(true,*i);

        bool implemented=false;

        string line=*i;
        string orig_line=*i;

        msgs_from_server.pop_front();

        {
            string resp=orig_line;

            mod_dispatch("irc_raw","","",line,true,0,NULL,"origin","@raw@",NULL);

            if(!irc_into_whois.empty()) {
                size_t i1;
                int i2=0;
                string nick;
                for(i1=0; i1<line.length(); ++i1) {
                    if(0x20==line[i1]) {
                        i2++;
                        if(i2>3)
                            break;
                        continue;
                    }
                    if(3==i2)
                        nick+=(string)""+line[i1];
                }
                multimap<string,string>::iterator i3;
                for(i3=irc_into_whois.begin(); i3!=irc_into_whois.end(); ++i3)
                    if(!irc_cmp_strings_case_insensitive((*i3).first,nick)) {
                        irc_into_whois.insert(make_pair(nick,line));
                        break;
                    }
            }

            if(resp.length()>=4) {
                if(resp[0]=='P' && resp[1]=='I' && resp[2]=='N' && resp[3]=='G') {
                    implemented=true;

                    string pong="PONG :";
                    for(unsigned int i1=6; i1<resp.length(); ++i1)
                        pong+=resp[i1];
                    pong+="\r\n";

                    irc_putserv_immediately(irc_socket,pong.c_str());

                    mod_dispatch("irc_raw","","",resp,false,0,NULL,"origin","@raw@",NULL);

                    goto l1;
                }

                if(resp[0]=='E' && resp[1]=='R' && resp[2]=='R' && resp[3]=='O' && resp[4]=='R') {
                    implemented=true;

                    if(strstr(resp.c_str(),"Throttled") ||
                        strstr(resp.c_str(),"throttled") ||
                        strstr(resp.c_str(),"too fast") ||
                        strstr(resp.c_str(),"Overridden") ||
                        strstr(resp.c_str(),"overridden")
                        ) {
                            irc_disconnect();

                            mod_dispatch("irc_raw","","",resp,false,0,NULL,"origin","@raw@",NULL);

                            return 2;
                        }
                    irc_disconnect();

                    mod_dispatch("irc_raw","","",resp,false,0,NULL,"origin","@raw@",NULL);

                    my_sleep(60);

                    return 1;
                }

                if(!orig_line.empty() && orig_line[0]!=':') {
                    string log="OOPS: Ignoring not-true origin message; message was: ";
                    log+=orig_line;
                    log_debug(log);

                    implemented=true;

                    goto l2;
                }

                unsigned int i1;
                for(i1=0; i1<resp.length(); ++i1)
                    if(resp[i1]==0x20)
                        break;
                i1++;

                size_t len=resp.length()-i1;
                if(len>4*1024-1) {
                    log_debug(GET_STR_DEBUG_INFO_EX("One line from IRC server exceeded about 4 KiB"));
                    implemented=true;
                    goto l2;
                }

                char tmp[4*1024];
                strcpy(tmp,(char*)&((resp.c_str())[i1]));
                if(tmp[strlen(tmp)-1]=='\r' || tmp[strlen(tmp)-1]=='\n')
                    tmp[strlen(tmp)-1]=0;
                if(tmp[strlen(tmp)-1]=='\r' || tmp[strlen(tmp)-1]=='\n')
                    tmp[strlen(tmp)-1]=0;

                string str=tmp;
                len=str.length();

                string hm;
                string nick, ident, host;
                string channel;
                string whom; // only for KICK
                string topic; // only for topic
                string notice; // only for NOTICE from server
                if(resp.length()>0 && resp[0]==':') {
                    unsigned int i1;
                    for(i1=1; i1<resp.length(); ++i1) {
                        if(resp[i1]==0x20)
                            break;
                        hm+=resp[i1];
                    }
                    if(hm.find("!")!=string::npos && hm.find("@")!=string::npos) {
                        i1++;
                        int pos=0;
                        for(i1=1; i1<resp.length(); ++i1) {
                            if(pos==0 && resp[i1]=='!') {
                                pos++;
                                continue;
                            }
                            if(pos==1 && resp[i1]=='@') {
                                pos++;
                                continue;
                            }
                            if(pos==2 && resp[i1]==0x20) {
                                pos++;
                                continue;
                            }
                            if(pos==3 && resp[i1]==0x20) {
                                pos++;
                                continue;
                            }
                            if(pos==4 && resp[i1]==0x20) {
                                pos++;
                                continue;
                            }
                            if(pos==5 && resp[i1]==':') {
                                pos++;
                                continue;
                            }
                            if(pos==0)
                                nick+=resp[i1];
                            if(pos==1)
                                ident+=resp[i1];
                            if(pos==2)
                                host+=resp[i1];
                            if(pos==3) {
                                // command
                            }
                            if(pos==4)
                                channel+=resp[i1];
                            if(pos==5)
                                whom+=resp[i1];
                            if(pos==6)
                                topic+=resp[i1];
                        }
                    } else {
                        // message from server
                        nick="";
                        ident="";
                        host="";
                        whom="";
                        channel="";
                        topic="";
                        notice="";
                        if(resp.length()>=i1+6 && resp[i1+1]=='M' && resp[i1+2]=='O' && resp[i1+3]=='D' &&
                            resp[i1+4]=='E' && resp[i1+5]==0x20) {
                                implemented=true;

                                channel=resp;
                                channel.erase(0,i1+1+strlen("MODE "));
                                whom=channel;
                                if(channel.find("\x20",0)!=string::npos)
                                    channel.erase(channel.find(0x20,0),channel.length()-channel.find(0x20,0));
                                if(whom.find(0x20,0)!=string::npos)
                                    whom.erase(0,whom.find(0x20,0)+1);
                            }
                        if(resp.length()>=i1+7 && resp[i1+1]=='T' && resp[i1+2]=='O' && resp[i1+3]=='P' &&
                            resp[i1+4]=='I' && resp[i1+5]=='C' && resp[i1+6]==0x20) {
                                implemented=true;

                                channel=resp;
                                channel.erase(0,i1+1+strlen("TOPIC "));
                                whom=channel;
                                if(channel.find("\x20",0)!=string::npos)
                                    channel.erase(channel.find(0x20,0),channel.length()-channel.find(0x20,0));
                                if(whom.find(0x20,0)!=string::npos)
                                    whom.erase(0,whom.find(0x20,0)+1);
                                whom.erase(0,1);
                                topic=whom;
                                whom="";
                            }
                        if(resp.length()>=i1+8 && resp[i1+1]=='N' && resp[i1+2]=='O' && resp[i1+3]=='T' &&
                            resp[i1+4]=='I' && resp[i1+5]=='C' && resp[i1+6]=='E' && resp[i1+7]==0x20) {
                                implemented=true;

                                channel=resp;
                                channel.erase(0,i1+1+strlen("NOTICE "));
                                whom=channel;
                                if(channel.find("\x20",0)!=string::npos)
                                    channel.erase(channel.find(0x20,0),channel.length()-channel.find(0x20,0));
                                if(whom.find(0x20,0)!=string::npos)
                                    whom.erase(0,whom.find(0x20,0)+1);
                                whom.erase(0,1);
                                notice=whom;
                                whom="";
                            }
                    }
                    while(whom.length()>0 && whom[whom.length()-1]==0x20)
                        whom.erase(whom.length()-1,1);
                }
                string quoted;
                string msg;
                bool sometihng=false;
                bool after_colon=false;
                for(i1=0; i1<len; ++i1) {
                    if(after_colon)
                        msg+=str[i1];
                    if(str[i1]==':')
                        after_colon=true;
                    if(sometihng && str[i1]==0x01)
                        break;
                    if(!sometihng && str[i1]==0x01) {
                        sometihng=true;
                        continue;
                    }
                    if(sometihng)
                        quoted+=str[i1];
                }
                if(channel.length()>0 && channel[0]==':')
                    channel.erase(0,1);

                if(len>=4 && str[0]=='P' && str[1]=='O' && str[2]=='N' && str[3]=='G') {
                    implemented=true;

                    irc_kill_command_wait("PING");
                }

                if(len>=7 && str[0]=='P' && str[1]=='R' && str[2]=='I' && str[3]=='V' && str[4]=='M' && str[5]=='S' && str[6]=='G' && str[7]==' ') {
                    implemented=true;

                    string tmp=nick;
                    tmp+="!";
                    tmp+=ident;
                    tmp+="@";
                    tmp+=host;
                    if(channel.find("$")!=string::npos || channel.find("%")!=string::npos) {
                        // we got broadcast PRIVMSG message
                        log_broadcast(channel,TYPE_PRIVMSG,tmp,msg,irc_server_host,irc_server_port);
                        irc_on_broadcast(channel,TYPE_PRIVMSG,tmp,msg,irc_server_host,irc_server_port);
                    } else if(!nick.empty()) {
                        if(irc_cmp_strings_case_insensitive(channel,irc_nick)) {
                            log_channel(channel,TYPE_PRIVMSG,tmp,"",msg,irc_get_modes_for_log(channel,nick));
                            if(msg.find('\x01',0)!=string::npos && msg.find("\x01" "ACTION ",0)==string::npos)
                                irc_check_flood(TYPE_CTCP,channel,tmp,msg);
                            else
                                irc_check_flood(TYPE_PRIVMSG,channel,tmp,msg);
                            irc_on_privmsg(channel,tmp,msg);
                        } else {
                            log_channel("private",TYPE_PRIVMSG,tmp,"",msg,"");
                            irc_on_privmsg("",tmp,msg);
                        }
                    } else {
                        // PRIVMSG from server
                        if(irc_cmp_strings_case_insensitive(channel,irc_nick)) {
                            // it is for channel
                            log_channel(channel,TYPE_PRIVMSG,hm,channel,msg,"");
                            // we really don't want to check for flood now, since it is not likely
                            //   when server begins to flood channel, and there is a fact that we
                            //   cannot ban server ;-)
                            //   -- quote of VooDooMan on 2005-04-22 @ 14:30 GMT +01:00 (zopiclonum 7.5 mg per oral)

                            logic_on_server_msg(channel,hm,TYPE_PRIVMSG,msg,irc_server_host,irc_server_port);
                        } else {
                            // it is privately for bot
                            log_channel("private",TYPE_PRIVMSG,hm,channel,msg,"");

                            logic_on_server_msg("",hm,TYPE_PRIVMSG,msg,irc_server_host,irc_server_port);
                        }
                    }

                    if(!irc_cmp_strings_case_insensitive(channel,irc_nick) &&
                        (
                        (msg.length()>=3 && msg[0]=='d' && msg[1]=='c' && msg[2]=='c')
                        ||
                        (msg.length()>=4 && msg[0]=='.' && msg[1]=='d' && msg[2]=='c' && msg[3]=='c')
                        )
                        ) {
                        bool lag=false;
                        {
                            vector<c_online_user>::iterator i1;
                            for(i1=tmp_users.begin(); i1!=tmp_users.end(); ++i1)
                                if(!(*i1).nick.compare(nick)) {
                                    lag=true;
                                    break;
                                }
                        }
                        if(!lag) {
                            string user_name_as_in_logic;
                            if(irc_access_to_partyline(hm.c_str(),user_name_as_in_logic,msg)) {
                                int dcc_group=0;
                                int ip_ver;
                                string dcc=dcc_get_string(msg,dcc_group,ip_ver);
                                {
                                    c_online_user u;
                                    u.nick=nick;
                                    u.got_whois=true;
                                    u.whois_sent=true;
                                    u.ident=ident;
                                    u.host=host;
                                    u.fullname=irc_get_fullname(nick);
                                    u.irc_op=irc_is_ircop(nick);
                                    u.in_logic_as=user_name_as_in_logic;
                                    u.last_whois=my_time(NULL);
                                    u.dcc_msg=msg;
                                    u.resolv_host=host;
                                    u.ip_version=ip_ver;

                                    if(4==u.ip_version || 6==u.ip_version)
                                        tmp_users.push_back(u);
                                }
                            }
                        }
                    }

                    if((msg.length()>=7 && msg[0]=='g' && msg[1]=='e' && msg[2]=='t' && msg[3]=='f' && msg[4]=='i' && msg[5]=='l' && msg[6]=='e' && msg[7]==' ')
                        ||
                        (msg.length()>=8 && msg[0]=='.' && msg[1]=='g' && msg[2]=='e' && msg[3]=='t' && msg[4]=='f' && msg[5]=='i' && msg[6]=='l' && msg[7]=='e' && msg[8]==' ')
                        ) {
                        string cmd=msg;
                        while(cmd[0]!=' ')
                            cmd.erase(0,1);
                        string fn;
                        string dcc_gr="0";
                        int pos=0;
                        while(cmd.length()>0) {
                            if(cmd[0]==' ') {
                                cmd.erase(0,1);
                                pos++;
                                continue;
                            }
                            if(pos==1) {
                                fn+=cmd[0];
                                cmd.erase(0,1);
                                continue;
                            }
                            if(pos==2) {
                                dcc_gr+=cmd[0];
                                cmd.erase(0,1);
                                continue;
                            }
                        }
                        int dcc_group=stoi(dcc_gr);
                        int res=dcc_send_file(false,fn,nick,dcc_group);
                        string disp;
                        switch(res) {
                            case DCC_NO_SUCH_FILE:
                                disp=lang_get_string(1,"en",334);
                                break;
                            case DCC_ACCESS_DENIED:
                                disp=lang_get_string(1,"en",335);
                                break;
                            case DCC_IO_ERROR:
                                disp=lang_get_string(1,"en",336);
                                break;
                        }

                        if(!disp.empty()) {
                            string m=":";
                            m+=disp;
                            irc_putserv(c_tracked_msg_to_server("PRIVMSG").add_nick(nick).add_piece(m),false,LOW_PRIORITY);
                        } else {
                            dcc_file_has_been_read(fn,nick);
                        }
                    }

                    if(!quoted.empty()) {
                        irc_quoted_callback(quoted,nick,hm,orig_line);
                    }

                    goto l1;
                }

                if(len>=6 && str[0]=='N' && str[1]=='O' && str[2]=='T' && str[3]=='I' && str[4]=='C' && str[5]=='E' && str[6]==' ') {
                    implemented=true;

                    string tmp=nick;
                    tmp+="!";
                    tmp+=ident;
                    tmp+="@";
                    tmp+=host;
                    if(channel.find("$")!=string::npos || channel.find("%")!=string::npos) {
                        // we got broadcast NOTICE message
                        log_broadcast(channel,TYPE_NOTICE,tmp,msg,irc_server_host,irc_server_port);
                        irc_on_broadcast(channel,TYPE_NOTICE,tmp,msg,irc_server_host,irc_server_port);
                    } else if(!nick.empty()) {
                        if(irc_cmp_strings_case_insensitive(channel,irc_nick)) {
                            log_channel(channel,TYPE_NOTICE,hm,channel,msg,irc_get_modes_for_log(channel,nick));
                            if(msg.find('\x01',0)!=string::npos && msg.find("\x01" "ACTION ",0)==string::npos)
                                irc_check_flood(TYPE_CTCP,channel,tmp,msg);
                            else
                                irc_check_flood(TYPE_NOTICE,channel,tmp,msg);
                            irc_on_notice(channel,tmp,msg);
                        } else {
                            log_channel("private",TYPE_NOTICE,hm,channel,msg,"");
                            irc_on_notice("",tmp,msg);
                        }
                    } else {
                        // NOTICE from server
                        if(irc_cmp_strings_case_insensitive(channel,irc_nick)) {
                            // it is for channel
                            log_channel(channel,TYPE_NOTICE,hm,channel,msg,"");
                            // we really don't want to check for flood now, since it is not likely
                            //   when server begins to flood channel, and there is a fact that we
                            //   cannot ban server ;-)
                            //   -- quote of VooDooMan on 2005-04-22 @ 14:30 GMT +01:00 (zopiclonum 7.5 mg per oral)

                            logic_on_server_msg(channel,hm,TYPE_NOTICE,msg,irc_server_host,irc_server_port);
                        } else {
                            // it is privately for bot
                            log_channel("private",TYPE_NOTICE,hm,channel,msg,"");

                            logic_on_server_msg("",hm,TYPE_NOTICE,msg,irc_server_host,irc_server_port);
                        }
                    }
                }

                if(len>4 && str[0]=='J' && str[1]=='O' && str[2]=='I' && str[3]=='N') {
                    implemented=true;

                    if(!nick.empty()) {
                        string tmp=nick;
                        tmp+="!";
                        tmp+=ident;
                        tmp+="@";
                        tmp+=host;

                        log_channel(channel,TYPE_JOIN,tmp,"","",irc_get_modes_for_log(channel,nick));

                        irc_check_flood(TYPE_JOIN,channel,tmp,""); // !!! we (maybe) don't know who "tmp" is here!
                    }

                    if(!nick.compare(irc_nick)) {
                        irc_kill_command_wait("JOIN");

                        global_keeper.add_channel(channel);

                        vector<c_key_chan>::iterator i2;
                    label_b:
                        for(i2=irc_key_chans.begin(); i2!=irc_key_chans.end(); ++i2) {
                            if(!irc_cmp_strings_case_insensitive((*i2).channel_name,channel)) {
                                irc_key_chans.erase(i2);
                                goto label_b;
                            }
                        }

                        vector<string>::iterator i1;
                    erase_again:
                        for(i1=irc_cannot_join_channels.begin(); i1!=irc_cannot_join_channels.end(); ++i1) {
                            if(!irc_cmp_strings_case_insensitive(*i1,channel)) {
                                irc_cannot_join_channels.erase(i1);
                                goto erase_again;
                            }
                        }

                        string msg="WHO ";
                        msg+=channel;
                        irc_putserv(c_tracked_msg_to_server(msg),true,HIGH_PRIORITY);
                    }

                    str.erase(0,5);
                    unsigned int i1;
                    for(i1=0; i1<str.length(); ++i1)
                        if(str[i1]==0x20)
                            break;
                    if(str.length()-i1>0)
                        str.erase(i1,str.length()-i1);

                    for(i1=0; i1<str.length(); ++i1)
                        if(str[i1]==':')
                            break;
                    if(str[i1]==':')
                        str.erase(0,i1+1); // erase everything before ':'

                    bool got_user=false;
                    bool got_channel=false;
                l5:
                    vector<c_online_channel>::iterator i2;
                    vector<c_online_user>::iterator ii;
                    for(i2=irc_channels.begin(); i2!=irc_channels.end(); ++i2) {
                        if(!irc_cmp_strings_case_insensitive((*i2).name,str)) {
                            got_channel=true;
                            for(ii=(*i2).users.begin(); ii!=(*i2).users.end(); ++ii) {
                                if(!irc_cmp_strings_case_insensitive((*ii).nick,nick)) {
                                    (*i2).users.erase(ii);
                                    got_user=false;
                                    goto l5;
                                }
                            }
                            if(got_user)
                                break;
                        }
                    }

                l10:
                    vector<c_online_channel>::iterator i;
                    got_channel=false;
                    for(i=irc_channels.begin(); i!=irc_channels.end(); ++i) {
                        if(!irc_cmp_strings_case_insensitive((*i).name,str)) {
                            got_channel=true;
                            break;
                        }
                    }
                    if(!got_channel) {
                        c_online_channel chan(str);
                        irc_channels.push_back(chan);

                        goto l10;
                    }

                    if(!nick.compare(irc_nick)) {
                        // we are new on channel, right after JOIN
                    }

                    if(got_channel && !irc_check_nick_on_channel(nick,channel)) {
                        for(i2=irc_channels.begin(); i2!=irc_channels.end(); ++i2)
                            if(!irc_cmp_strings_case_insensitive((*i2).name,str))
                                break;

                        c_online_user u;
                        u.nick=nick;
                        u.got_whois=false;
                        u.on_join_called=false;
                        u.old_mode="INVALID";
                        u.on_join_passive=false;
                        u.on_join_called2=false;

                        bool got_nick=false;
                        vector<c_online_channel>::iterator ii1;
                        vector<c_online_user>::iterator ii2;
                        for(ii1=irc_channels.begin(); ii1!=irc_channels.end(); ++ii1) {
                            for(ii2=(*ii1).users.begin(); ii2!=(*ii1).users.end(); ++ii2) {
                                if(!(*ii2).nick.compare(nick)) {
                                    got_nick=true;
                                    break;
                                }
                            }
                            if(got_nick)
                                break;
                        }

                        if(!got_nick) {
                    hard_whois:
                            irc_putserv(c_tracked_msg_to_server("WHOIS").add_nick(nick),true,LOW_PRIORITY);

                            u.whois_sent=true;

                            u.ident=ident;
                            u.host=host;
                            u.fullname="";
                            u.got_mode=false;
                            u.mode="0";
                            u.irc_op=false;
                            u.in_logic_as="";

                            u.just_joined=true;
                            u.just_joined_channel=channel;
                        } else {
                            if(!(*ii2).got_whois)
                                goto hard_whois;
                            u.fullname=(*ii2).fullname;
                            u.got_mode=true;
                            u.got_whois=true;
                            u.host=(*ii2).host;
                            u.ident=(*ii2).ident;
                            u.in_logic_as=(*ii2).in_logic_as;
                            u.irc_op=(*ii2).irc_op;
                            u.mode="0";
                            u.nick=(*ii2).nick;
                            u.old_nick=u.nick;
                            u.old_mode=u.mode;
                            u.on_join_called=true;
                            u.whois_sent=true;
                            u.just_joined_channel=channel;

                            u.just_joined=true;
                            if(nick.compare(irc_nick)) {
                                // only call if it is not me myself!
                                logic_on_join(u.nick,u.ident,u.host,u.fullname,u.just_joined_channel,u.irc_op);
                                u.mod_disp_join_called=true;
                            }
                            if(u.irc_op) {
                                logic_on_ircop(channel,u.nick,true);
                                string host_mask=u.nick;
                                host_mask+="!";
                                host_mask+=u.ident;
                                host_mask+="@";
                                host_mask+=u.host;
                                log_channel(channel,TYPE_IRCOP_JOIN,host_mask,"","",irc_get_modes_for_log(channel,nick));
                            }
                        }

                        global_keeper.add_user(channel,nick,"0");
                        global_keeper.add_user_whois1(nick,ident,host,u.fullname);
                        if(u.irc_op)
                            global_keeper.add_user_ircop(nick);
                        irc_waiting_for_dns.push_back(c_dns_wait(nick,ident,host,u.fullname));

                        (*i2).users.push_back(u);
                    }

                    if(irc_check_last_msg("JOIN"))
                        irc_kill_last_msg_wait(resp);

                    irc_kill_command_wait("JOIN");
                }
                str=tmp;

                if(len>4 && str[0]=='Q' && str[1]=='U' && str[2]=='I' && str[3]=='T') {
                    implemented=true;

                    string host_mask;
                    if(!nick.empty()) {
                        string tmp=nick;
                        tmp+="!";
                        tmp+=ident;
                        tmp+="@";
                        tmp+=host;
                        host_mask=tmp;
                        vector<c_online_channel>::iterator i1;
                        for(i1=irc_channels.begin(); i1!=irc_channels.end(); ++i1) {
                            vector<c_online_user>::iterator i2;
                            for(i2=(*i1).users.begin(); i2!=(*i1).users.end(); ++i2)
                                if(!irc_cmp_strings_case_insensitive((*i2).nick,nick)) {

                                    {
                                        mod_dispatch("part@channel",(*i2).in_logic_as,(*i1).name,"",true,0,"channel",(*i1).name.c_str(),"source",(*i2).in_logic_as.c_str(),"source_nick",(*i2).nick.c_str(),"msg",msg.c_str(),"type","QUIT",NULL,"nick",(*i2).nick.c_str(),"origin","@irc@","ident",irc_get_ident((*i2).nick).c_str(),"host",irc_get_host((*i2).nick).c_str(),NULL);
                                    }

                                    global_keeper.nick_left((*i1).name,nick,!irc_nick.compare(nick));

                                    log_channel((*i1).name,TYPE_QUIT,tmp,"",msg,irc_get_modes_for_log(channel,nick));
                                    if((*i2).irc_op) {
                                        logic_on_ircop((*i1).name,nick,false);
                                        log_channel((*i1).name,TYPE_IRCOP_LEFT,host_mask,"","",irc_get_modes_for_log((*i1).name,nick));
                                    }

                                    process_tracked_nicks_chan((*i1).name,nick);

                                    irc_check_flood(TYPE_QUIT_PART_KICK,(*i1).name,tmp,msg);

                                    logic_on_part((*i1).name,nick,msg,TYPE_QUIT);
                                }
                        }
                    }

                    vector<c_online_channel>::iterator i;
                l6:
                    for(i=irc_channels.begin(); i!=irc_channels.end(); ++i) {
                        vector<c_online_user>::iterator ii;
                        for(ii=(*i).users.begin(); ii!=(*i).users.end(); ++ii) {
                            if(!irc_cmp_strings_case_insensitive((*ii).nick,nick)) {
                                my_time(&(*ii).quit_time);
                                gone_users.push_back(*ii);
                                (*i).users.erase(ii);
                                goto l6;
                            }
                        }
                    }
                cycle1:
                    for(i=irc_channels.begin(); i!=irc_channels.end(); ++i) {
                        string my_modes=irc_get_mode((*i).name,irc_nick);
                        if(my_modes.find('@',0)==string::npos && irc_cmp_strings_case_insensitive(nick,irc_nick)) {
                            int users=0;
                            vector<c_online_user>::iterator ii;
                            for(ii=(*i).users.begin(); ii!=(*i).users.end(); ++ii, ++users)
                                ;
                            if(users==1) {
                                // try to get op
                                string m=(*i).name;
                                irc_kill_last_msg_wait(resp);
                                string msg="PART ";
                                msg+=m;
                                if(!core_conf_getvar("irc_part_msg_on_cycle").empty()) {
                                    msg+=" :";
                                    msg+=core_conf_getvar("irc_part_msg_on_cycle");
                                }
                                irc_putserv(c_tracked_msg_to_server(msg),false,HIGH_PRIORITY);
                                msg="JOIN ";
                                msg+=m;
                                irc_putserv(c_tracked_msg_to_server(msg),true,HIGH_PRIORITY);
                                msg="MODE ";
                                msg+=m;
                                irc_putserv(c_tracked_msg_to_server(msg),true,HIGH_PRIORITY);
                                msg="TOPIC ";
                                msg+=m;
                                irc_putserv(c_tracked_msg_to_server(msg),true,HIGH_PRIORITY);
                                irc_putserv(c_tracked_msg_to_server("WHOIS").add_nick(irc_nick),true,LOW_PRIORITY);

                                irc_channels.erase(i);
                                goto cycle1;
                            }
                        }
                    }
                }
                str=tmp;

                if(len>4 && str[0]=='P' && str[1]=='A' && str[2]=='R' && str[3]=='T') {
                    implemented=true;

                    {
                        mod_dispatch("part@channel",irc_find_user(nick).in_logic_as,channel.c_str(),"",true,0,"channel",channel.c_str(),"source",irc_find_user(nick).in_logic_as.c_str(),"source_nick",nick.c_str(),"msg",msg.c_str(),"type","PART",NULL,"nick",nick.c_str(),"origin","@irc@","ident",irc_get_ident(nick).c_str(),"host",irc_get_host(nick).c_str(),NULL);
                    }

                    global_keeper.nick_left(channel,nick,!irc_nick.compare(nick));

                    string host_mask;
                    if(!nick.empty()) {
                        string tmp=nick;
                        tmp+="!";
                        tmp+=ident;
                        tmp+="@";
                        tmp+=host;
                        host_mask=tmp;
                        log_channel(channel,TYPE_PART,tmp,"",msg,irc_get_modes_for_log(channel,nick));

                        if(!channel.empty())
                            process_tracked_nicks_chan(channel,nick);

                        irc_check_flood(TYPE_QUIT_PART_KICK,channel,tmp,msg);
                    }

                    str.erase(0,5);
                    unsigned int i1;
                    for(i1=0; i1<str.length(); ++i1)
                        if(str[i1]==0x20)
                            break;
                    if(str.length()-i1>0)
                        str.erase(i1,str.length()-i1);

                l8:
                    vector<c_online_channel>::iterator i;
                    bool got_channel=false;
                    for(i=irc_channels.begin(); i!=irc_channels.end(); ++i) {
                        if(!irc_cmp_strings_case_insensitive((*i).name,str)) {
                            got_channel=true;
                            break;
                        }
                    }
                    if(!got_channel) {
                        // we are new on channel, right after JOIN
                        // assume that irc server will send names reply
                        c_online_channel chan(str);
                        irc_channels.push_back(chan);

                        goto l8;
                    }

                l7:
                    vector<c_online_user>::iterator ii;
                    for(ii=(*i).users.begin(); ii!=(*i).users.end(); ++ii) {
                        if(!irc_cmp_strings_case_insensitive((*ii).nick,nick) && !irc_cmp_strings_case_insensitive((*i).name,str)) {
                            logic_on_part(channel,nick,msg,TYPE_PART);
                            if((*ii).irc_op) {
                                logic_on_ircop(channel,nick,false);
                                log_channel(channel,TYPE_IRCOP_LEFT,host_mask,"","",irc_get_modes_for_log((*i).name,nick));
                            }

                            if(irc_cmp_strings_case_insensitive((*ii).ident,ident) || irc_cmp_strings_case_insensitive((*ii).host,host)) {
                                my_time(&(*ii).quit_time);
                                gone_users.push_back(*ii);
                                (*i).users.erase(ii);

                                goto l7;
                            }
                            (*i).users.erase(ii);
                            goto l7;
                        }
                    }
                cycle2:
                    for(i=irc_channels.begin(); i!=irc_channels.end(); ++i) {
                        string my_modes=irc_get_mode((*i).name,irc_nick);
                        if(my_modes.find('@',0)==string::npos && irc_cmp_strings_case_insensitive(nick,irc_nick)) {
                            int users=0;
                            vector<c_online_user>::iterator ii;
                            for(ii=(*i).users.begin(); ii!=(*i).users.end(); ++ii, ++users)
                                ;
                            // if someone else has PART-ed, try to get op
                            if(users==1 && irc_cmp_strings_case_insensitive(nick,irc_nick)) {
                                // try to get op
                                string m=(*i).name;
                                irc_kill_last_msg_wait(resp);
                                string msg="PART ";
                                msg+=m;
                                if(!core_conf_getvar("irc_part_msg_on_cycle").empty()) {
                                    msg+=" :";
                                    msg+=core_conf_getvar("irc_part_msg_on_cycle");
                                }
                                irc_putserv(c_tracked_msg_to_server(msg),false,HIGH_PRIORITY);
                                msg="JOIN ";
                                msg+=m;
                                irc_putserv(c_tracked_msg_to_server(msg),true,HIGH_PRIORITY);
                                msg="MODE ";
                                msg+=m;
                                irc_putserv(c_tracked_msg_to_server(msg),true,HIGH_PRIORITY);
                                msg="TOPIC ";
                                msg+=m;
                                irc_putserv(c_tracked_msg_to_server(msg),true,HIGH_PRIORITY);
                                irc_putserv(c_tracked_msg_to_server("WHOIS").add_nick(irc_nick),true,LOW_PRIORITY);

                                irc_channels.erase(i);
                                goto cycle2;
                            }
                        }
                    }
                    if(!irc_cmp_strings_case_insensitive(nick,irc_nick)) {
                        // if I PART-ed
                        irc_kill_last_msg_wait(resp);

                        irc_kill_command_wait("PART");

                again_e_chan2:
                        vector<c_online_channel>::iterator i;
                        for(i=irc_channels.begin(); i!=irc_channels.end(); ++i) {
                            if(!irc_cmp_strings_case_insensitive((*i).name,channel)) {
                                irc_channels.erase(i);
                                goto again_e_chan2;
                            }
                        }
                    }
                }
                str=tmp;

                if(len>4 && str[0]=='K' && str[1]=='I' && str[2]=='C' && str[3]=='K') {
                    implemented=true;

                    {
                        mod_dispatch("part@channel",irc_find_user(nick).in_logic_as,channel.c_str(),"",true,0,"channel",channel.c_str(),"source",irc_find_user(nick).in_logic_as.c_str(),"source_nick",nick.c_str(),"msg",msg.c_str(),"type","KICK",NULL,"nick",nick.c_str(),"origin","@irc@","ident",irc_get_ident(nick).c_str(),"host",irc_get_host(nick).c_str(),NULL);
                    }

                    global_keeper.nick_left(channel,nick,!irc_nick.compare(nick));

                    irc_kill_command_wait("KICK");

                    string host_mask;
                    if(!nick.empty()) {
                        string tmp=nick;
                        tmp+="!";
                        tmp+=ident;
                        tmp+="@";
                        tmp+=host;
                        host_mask=tmp;
                        log_channel(channel,TYPE_KICK,tmp,whom,msg,irc_get_modes_for_log(channel,nick));

                        if(!channel.empty())
                            process_tracked_nicks_chan(channel,nick);

                        irc_check_flood(TYPE_QUIT_PART_KICK,channel,tmp,msg);

                        irc_kicked(resp);
                    }

                    vector<c_online_channel>::iterator i;
                    for(i=irc_channels.begin(); i!=irc_channels.end(); ++i) {
                        if(!irc_cmp_strings_case_insensitive((*i).name,channel)) {
                            vector<c_online_user>::iterator ii;
                            for(ii=(*i).users.begin(); ii!=(*i).users.end(); ++ii) {
                                if(!irc_cmp_strings_case_insensitive((*ii).nick,whom)) {
                                    logic_on_part(channel,whom,msg,TYPE_KICK);
                                    if((*ii).irc_op) {
                                        logic_on_ircop(channel,whom,false);
                                        log_channel(channel,TYPE_IRCOP_LEFT,host_mask,"","",irc_get_modes_for_log((*i).name,nick));
                                    }

                                    my_time(&(*ii).quit_time);
                                    gone_users.push_back(*ii);
                                    (*i).users.erase(ii);
                                    break;
                                }
                            }
                        }
                    }
                cycle3:
                    for(i=irc_channels.begin(); i!=irc_channels.end(); ++i) {
                        string my_modes=irc_get_mode((*i).name,irc_nick);
                        if(my_modes.find('@',0)==string::npos && irc_cmp_strings_case_insensitive(nick,irc_nick)) {
                            int users=0;
                            vector<c_online_user>::iterator ii;
                            for(ii=(*i).users.begin(); ii!=(*i).users.end(); ++ii, ++users)
                                ;
                            // if not KICK-ed me, but someone else, try to get op
                            if(users==1 && irc_cmp_strings_case_insensitive(whom,irc_nick)) {
                                // try to get op
                                string m=(*i).name;
                                irc_kill_last_msg_wait(resp);

                                irc_kill_command_wait("KICK");

                                string msg="PART ";
                                msg+=m;
                                if(!core_conf_getvar("irc_part_msg_on_cycle").empty()) {
                                    msg+=" :";
                                    msg+=core_conf_getvar("irc_part_msg_on_cycle");
                                }
                                irc_putserv(c_tracked_msg_to_server(msg),false,HIGH_PRIORITY);
                                msg="JOIN ";
                                msg+=m;
                                irc_putserv(c_tracked_msg_to_server(msg),true,HIGH_PRIORITY);
                                msg="MODE ";
                                msg+=m;
                                irc_putserv(c_tracked_msg_to_server(msg),true,HIGH_PRIORITY);
                                msg="TOPIC ";
                                msg+=m;
                                irc_putserv(c_tracked_msg_to_server(msg),true,HIGH_PRIORITY);
                                irc_putserv(c_tracked_msg_to_server("WHOIS").add_nick(irc_nick),true,LOW_PRIORITY);

                                irc_channels.erase(i);
                                goto cycle3;
                            }
                        }
                    }
                    if(!irc_cmp_strings_case_insensitive(whom,irc_nick)) {
                again_e_chan1:
                        vector<c_online_channel>::iterator i;
                        for(i=irc_channels.begin(); i!=irc_channels.end(); ++i) {
                            if(!irc_cmp_strings_case_insensitive((*i).name,channel)) {
                                irc_channels.erase(i);
                                goto again_e_chan1;
                            }
                        }
                    }
                    irc_kill_last_msg_wait(resp);
                }
                str=tmp;

                if(len>4 && str[0]=='N' && str[1]=='I' && str[2]=='C' && str[3]=='K') {
                    implemented=true;

                    irc_kill_command_wait("NICK");

            forced_nick_change:
                    if(!nick.empty()) {
                        string tmp=nick;
                        tmp+="!";
                        tmp+=ident;
                        tmp+="@";
                        tmp+=host;
                        vector<c_online_channel>::iterator i1;
                        for(i1=irc_channels.begin(); i1!=irc_channels.end(); ++i1) {
                            vector<c_online_user>::iterator ii;
                            for(ii=(*i1).users.begin(); ii!=(*i1).users.end(); ++ii) {
                                if(!irc_cmp_strings_case_insensitive((*ii).nick,nick)) {
                                    // msg contains new nick
                                    log_channel((*i1).name,TYPE_NICK,tmp,"",msg,irc_get_modes_for_log((*i1).name,nick));
                                    if(nick.compare(irc_nick)) {
                                        irc_check_flood(TYPE_NICK,(*i1).name,tmp,msg);
                                        // do NOT perform next line! we got it later
                                        //(*ii).nick=msg;
                                        irc_putserv(c_tracked_msg_to_server("WHOIS").add_nick(msg),true,LOW_PRIORITY);
                                    } else {
                                        {
                                            irc_putserv(c_tracked_msg_to_server("WHOIS").add_nick(msg),true,LOW_PRIORITY);
                                        }
                                    }
                                }
                            }
                        }
                        if(!irc_cmp_strings_case_insensitive(nick,irc_nick)) {
                            irc_nick=msg;
                        }
                    }

                    global_keeper.nick_change(nick,msg);
                    process_tracked_nicks(nick,msg);

                    vector<c_online_channel>::iterator i;
                    for(i=irc_channels.begin(); i!=irc_channels.end(); ++i) {
                        vector<c_online_user>::iterator ii;
                        for(ii=(*i).users.begin(); ii!=(*i).users.end(); ++ii) {
                            if(!irc_cmp_strings_case_insensitive((*ii).nick,nick)) {
                                (*ii).nick_validated=false;
                                (*ii).mod_disp_join_called=false;
                                string old_nick=(*ii).nick;
                                (*ii).nick=msg;

                                string channel=(*ii).just_joined_channel;
                                if(channel.empty())
                                    channel=(*i).name;

                                logic_on_nick_change((*ii).nick,(*ii).ident,(*ii).host,(*ii).fullname,channel,(*ii).irc_op,old_nick);
                            }
                        }
                    }
                }
                str=tmp;

                // 433: nickname already in use
                if(len>4 && str[0]=='4' && str[1]=='3' && str[2]=='3' && str[3]==' ') {
                    implemented=true;

                    if(irc_check_last_msg("NICK"))
                        irc_kill_last_msg_wait(resp);

                    irc_kill_command_wait("NICK");
                }

                if(len>4 && str[0]=='0' && str[1]=='0' && str[2]=='1' && str[3]==' ') {
                    implemented=true;

                    if(irc_check_last_msg("USER"))
                        irc_kill_last_msg_wait(resp);

                    irc_kill_command_wait("USER");

                    irc_nick_confirmed=true;
                }
                if(len>4 && str[0]=='0' && str[1]=='0' && str[2]=='2' && str[3]==' ') {
                    implemented=true;

                    if(irc_check_last_msg("USER"))
                        irc_kill_last_msg_wait(resp);

                    irc_kill_command_wait("USER");
                }
                if(len>4 && str[0]=='0' && str[1]=='0' && str[2]=='3' && str[3]==' ') {
                    implemented=true;

                    if(irc_check_last_msg("USER"))
                        irc_kill_last_msg_wait(resp);

                    irc_kill_command_wait("USER");
                }
                if(len>4 && str[0]=='0' && str[1]=='0' && str[2]=='4' && str[3]==' ') {
                    implemented=true;

                    if(irc_check_last_msg("USER"))
                        irc_kill_last_msg_wait(resp);

                    irc_kill_command_wait("USER");
                }
                if(len>4 && str[0]=='0' && str[1]=='0' && str[2]=='5' && str[3]==' ') {
                    implemented=true;

                    if(irc_check_last_msg("USER"))
                        irc_kill_last_msg_wait(resp);

                    irc_kill_command_wait("USER");

                    // check for string "Try server <server name>, port <port number>"
                    str.erase(0,4);
                    size_t len=str.length();
                    if(len>11 && str[1]=='r' && str[2]=='y' && str[3]==0x20 &&  str[4]=='s' && str[5]=='e' &&
                        str[6]=='r' &&  str[7]=='v' &&  str[8]=='e' &&  str[9]=='r' &&  str[10]==0x20) {
                            str.erase(0,11);
                            string p;
                            p=str;

                            while(p.length()>0 && p[0]!=',')
                                p.erase(0,1);
                            while(p.length()>0 && (p[0]<'0' || p[0]>'9'))
                                p.erase(0,1);

                            unsigned int i1;
                            for(i1=0; i1<str.length(); ++i1)
                                if(str[i1]==',' || str[i1]==0x20)
                                    break;
                            str.erase(0,i1+1);

                            unsigned short port=stous(p);
                            if(irc_follow_redirs) {
                                irc_recommended_server=str;
                                irc_recommended_port=port;
                            }
                        } else {
                            irc_RPL_ISUPPORT(str);
                        }
                }

                // NAMES reply
                if(len>4 && str[0]=='3' && str[1]=='5' && str[2]=='3' && str[3]==' ') {
                    implemented=true;

                    irc_RPL_NAMREPLY(str);
                }

                // RPL_ENDOFNAMES
                if(len>4 && str[0]=='3' && str[1]=='6' && str[2]=='6' && str[3]==' ') {
                    implemented=true;

                    if(irc_check_last_msg("NAMES"))
                        irc_kill_last_msg_wait(resp);

                    irc_kill_command_wait("NAMES");
                }

                // RPL_WHOISUSER
                // WHOIS reply
                if(len>4 && str[0]=='3' && str[1]=='1' && str[2]=='1' && str[3]==' ') {
                    implemented=true;

                    irc_RPL_WHOISUSER(str,line);
                }

                // RPL_WHOISOPERATOR
                // user is an IRC operator
                if(len>4 && str[0]=='3' && str[1]=='1' && str[2]=='3' && str[3]==' ') {
                    implemented=true;

                    irc_RPL_WHOISOPERATOR(str);
                }

                // RPL_ENDOFWHOIS
                if(len>4 && str[0]=='3' && str[1]=='1' && str[2]=='8' && str[3]==' ') {
                    implemented=true;

                    irc_RPL_ENDOFWHOIS(str);
                    if(irc_check_last_msg("WHOIS"))
                        irc_kill_last_msg_wait(resp);

                    irc_kill_command_wait("WHOIS");
                }

                // RPL_WHOISCHANNELS
                if(len>4 && str[0]=='3' && str[1]=='1' && str[2]=='9' && str[3]==' ') {
                    implemented=true;

                    irc_RPL_WHOISCHANNELS(str);
                }

                // 437: Channel/nick is temporarily unavailable (!!! also can be not channel, but nick!!!)
                if(len>4 && str[0]=='4' && str[1]=='3' && str[2]=='7' && str[3]==' ') {
                    implemented=true;

                    if(irc_check_last_msg("NICK"))
                        irc_kill_last_msg_wait(resp);
                    irc_kill_command_wait("NICK");

                    if(irc_check_last_msg("JOIN"))
                        irc_kill_last_msg_wait(resp);
                    irc_kill_command_wait("JOIN");
                }

                // 471: Cannot join channel (+l) - limit
                // 473: Cannot join channel (+i) - invite only
                // 474: Cannot join channel (+b) - banned
                // 437: Channel/nick is temporarily unavailable (!!! also can be not channel, but nick!!!)
                // 405: (Unreal IRCd, after "JOIN #channel" command) :vdm.no-ip.org 405 BotNick #channel :You have joined too many channels
                if(len>4 && (
                    (str[0]=='4' && str[1]=='7' && str[2]=='1' && str[3]==' ') ||
                    (str[0]=='4' && str[1]=='7' && str[2]=='3' && str[3]==' ') ||
                    (str[0]=='4' && str[1]=='7' && str[2]=='4' && str[3]==' ') ||
                    (str[0]=='4' && str[1]=='3' && str[2]=='7' && str[3]==' ') ||
                    (str[0]=='4' && str[1]=='0' && str[2]=='5' && str[3]==' ')
                    ))
                {
                    implemented=true;

                    irc_kill_command_wait("JOIN");
                    irc_kill_command_wait("MODE");
                    irc_kill_command_wait("TOPIC");

                    string s=str;
                    int i1=0;
                    while(i1<2 && s.length()>0) {
                        if(s[0]==0x20)
                            i1++;
                        s.erase(0,1);
                    }
                    string chan;
                    while(s.length()>0 && s[0]!=0x20) {
                        chan+=s[0];
                        s.erase(0,1);
                    }

                    bool got=false;
                    vector<string>::iterator i2;
                    for(i2=irc_cannot_join_channels.begin(); i2!=irc_cannot_join_channels.end(); ++i2) {
                        if(!irc_cmp_strings_case_insensitive(*i2,chan)) {
                            got=true;
                            break;
                        }
                    }
                    string chantypes;
                    map<string,string>::iterator iii;
                    for(iii=irc_isupport.raw.begin(); iii!=irc_isupport.raw.end(); ++iii)
                        if(!(*iii).first.compare("CHANTYPES")) {
                            chantypes=(*iii).second;
                            break;
                        }
                    if(!got && chan.length()>0 && chantypes.find(chan[0])==string::npos)
                        got=true; // it is just a "437 nick teporarily unavailable" (not channel)

                    if(!got)
                        irc_cannot_join_channels.push_back(chan);

                    vector<c_online_channel>::iterator i;
                    for(i=irc_channels.begin(); i!=irc_channels.end(); ++i)
                        if(!irc_cmp_strings_case_insensitive((*i).name,chan)) {
                            irc_channels.erase(i);
                            break;
                        }

                    irc_kill_last_msg_wait(resp);
                }

                //IRC IN  :irc.server.com 442 BOT #channel :You're not on that channel
                if(len>4 && str[0]=='4' && str[1]=='4' && str[2]=='2' && str[3]==' ') {
                    implemented=true;

                    irc_kill_command_wait("MODE");
                    irc_kill_command_wait("TOPIC");

                    irc_kill_last_msg_wait(resp);
                }

                // IRC IN  :irc.server.com 324 BOT #channel +nt
                // channel mode after JOIN
                if(len>4 && str[0]=='3' && str[1]=='2' && str[2]=='4' && str[3]==' ') {
                    implemented=true;

                    irc_RPL_324(str);

                    if(str.find(" ",0)!=string::npos)
                        str.erase(0,str.find(" ")+1); // erase everything before and inclusive trailing SPACE - numeric reply
                    if(str.find(" ",0)!=string::npos)
                        str.erase(0,str.find(" ")+1); // erase everything before SPACE inclusive - my nick
                    string chan;
                    if(str.find(" ",0)!=string::npos)
                        while(str.length() && str[0]!=0x20) {
                            chan+=str[0];
                            str.erase(0,1);
                        }
                    str.erase(0,1); // erase space
                    if(str.length() && str[0]=='+')
                        str.erase(0,1); // erase leading '+'
                    if(str.find(" ",0)!=string::npos)
                        str.erase(str.find(" ",0),str.length()-str.find(" ",0)); // erase everything after SAPCE inclusive (possibly the topic)
                    // in "str" we have channel modes now
                    vector<c_online_channel>::iterator i;
                    for(i=irc_channels.begin(); i!=irc_channels.end(); ++i) {
                        if(irc_cmp_strings_case_insensitive((*i).name,chan))
                            continue;
                        (*i).modes=str;
                        break;
                    }
                    irc_kill_last_msg_wait(resp);

                    irc_kill_command_wait("MODE");
                }
                str=tmp;

                // IRC IN  :irc.server.com 332 BOT #channel :topic
                // topic after join
                if(len>4 && str[0]=='3' && str[1]=='3' && str[2]=='2' && str[3]==' ') {
                    implemented=true;

                    irc_kill_command_wait("TOPIC");

                    if(str.find(" ",0)!=string::npos)
                        str.erase(0,str.find(" ")+1); // erase everything before and inclusive trailing SPACE - numeric reply
                    if(str.find(" ",0)!=string::npos)
                        str.erase(0,str.find(" ")+1); // erase everything before SPACE inclusive - my nick
                    string chan;
                    if(str.find(" ",0)!=string::npos)
                        while(str.length() && str[0]!=0x20) {
                            chan+=str[0];
                            str.erase(0,1);
                        }
                    str.erase(0,1); // erase space

                    if(str.find(":",0)!=string::npos)
                        str.erase(0,str.find(":")+1); // erase everything before ':' inclusive

                    // in str we have topic
                    vector<c_online_channel>::iterator i;
                    for(i=irc_channels.begin(); i!=irc_channels.end(); ++i) {
                        if(irc_cmp_strings_case_insensitive((*i).name,chan))
                            continue;
                        (*i).topic=str;
                        break;
                    }
                    irc_kill_last_msg_wait(resp);

                    irc_kill_command_wait("TOPIC");

                    global_keeper.change_topic(chan,str);
                }
                str=tmp;

                // IRC IN  :irc.server.com 331 BOT #channel :No topic is set.
                // NO topic after join
                if(len>4 && str[0]=='3' && str[1]=='3' && str[2]=='1' && str[3]==' ') {
                    implemented=true;

                    irc_kill_command_wait("TOPIC");

                    if(str.find(" ",0)!=string::npos)
                        str.erase(0,str.find(" ")+1); // erase everything before and inclusive trailing SPACE - numeric reply
                    if(str.find(" ",0)!=string::npos)
                        str.erase(0,str.find(" ")+1); // erase everything before SPACE inclusive - my nick
                    string chan;
                    if(str.find(" ",0)!=string::npos)
                        while(str.length() && str[0]!=0x20) {
                            chan+=str[0];
                            str.erase(0,1);
                        }
                    str.erase(0,1); // erase space

                    if(str.find(":",0)!=string::npos)
                        str.erase(0,str.find(":")+1); // erase everything before ':' inclusive

                    vector<c_online_channel>::iterator i;
                    for(i=irc_channels.begin(); i!=irc_channels.end(); ++i) {
                        if(irc_cmp_strings_case_insensitive((*i).name,chan))
                            continue;
                        (*i).topic="";
                        break;
                    }
                    irc_kill_last_msg_wait(resp);

                    irc_kill_command_wait("TOPIC");
                }
                str=tmp;

                if(len>4 && str[0]=='M' && str[1]=='O' && str[2]=='D' && str[3]=='E') {
                    implemented=true;

                    if(!channel.length() || !irc_cmp_strings_case_insensitive(channel,irc_nick)) {
                        whom=str;
                        if(whom.find(":",0)!=string::npos)
                            whom.erase(0,whom.find(":",0)+1);
                    }

                    if(irc_cmp_strings_case_insensitive(channel,irc_nick)) {
                        // it is for channel
                        log_channel(channel,TYPE_MODE,hm,"",whom,irc_get_modes_for_log(channel,nick));
                    } else {
                        // it is privately for bot
                        log_channel("status",TYPE_MODE,hm,"",whom,irc_get_modes_for_log(channel,nick));
                    }

                    if(irc_check_last_msg("MODE"))
                        irc_kill_last_msg_wait(resp);

                    irc_kill_command_wait("MODE");

                    irc_RPL_MODE(resp);
                }

                if(len>4 && str[0]=='T' && str[1]=='O' && str[2]=='P' && str[3]=='I' && str[4]=='C') {
                    implemented=true;

                    if(!nick.empty()) {
                        string tmp=nick;
                        tmp+="!";
                        tmp+=ident;
                        tmp+="@";
                        tmp+=host;
                        irc_check_flood(TYPE_MODE,channel,tmp,topic);
                    }

                    vector<c_online_channel>::iterator i;
                    for(i=irc_channels.begin(); i!=irc_channels.end(); ++i) {
                        if(irc_cmp_strings_case_insensitive((*i).name,channel))
                            continue;
                        (*i).topic=topic;
                        break;
                    }

                    log_channel(channel,TYPE_TOPIC,hm,"TOPIC",topic,irc_get_modes_for_log(channel,nick));

                    if(irc_check_last_msg("TOPIC"))
                        irc_kill_last_msg_wait(resp);

                    irc_kill_command_wait("TOPIC");

                    irc_RPL_TOPIC(resp);
                }

                if(str.find("INVITE")==0) {
                    implemented=true;

                    // dummy
                }

                // 472: unknown mode
                if(len>4 && str[0]=='4' && str[1]=='7' && str[2]=='2' && str[3]==' ') {
                    implemented=true;

                    if(irc_check_last_msg("MODE"))
                        irc_kill_last_msg_wait(resp);

                    irc_kill_command_wait("MODE");

                    // assume that error is on MODE #chan +e (some irc networks doesn't support +e exceptions to bans)
                    vector<c_online_channel>::iterator i;
                    for(i=irc_channels.begin(); i!=irc_channels.end(); ++i) {
                        (*i).got_excepts=true;
                        irc_check_bans((*i).name);
                    }

                    // assume that error is on MODE #chan +I (some irc networks doesn't support +I invitations)
                    for(i=irc_channels.begin(); i!=irc_channels.end(); ++i) {
                        (*i).got_invites=true;
                    }

                    for(i=irc_channels.begin(); i!=irc_channels.end(); ++i) {
                        (*i).got_reops=true;
                    }

                    c_online_channel* chan;
                    if(irc_get_online_channel(channel,chan)) {
                        if(!chan->b_excepts_sent && irc_isupport.chanmodes_a.find("e")!=string::npos) {
                            chan->b_excepts_sent=true;
                            string excepts="MODE ";
                            excepts+=channel;
                            excepts+=" +e";
                            irc_putserv(c_tracked_msg_to_server(excepts),true,HIGH_PRIORITY);
                        }

                        if(!chan->b_bans_sent && irc_isupport.chanmodes_a.find("b")!=string::npos) {
                            chan->b_bans_sent=true;
                            string bans="MODE ";
                            bans+=channel;
                            bans+=" +b";
                            irc_putserv(c_tracked_msg_to_server(bans),true,HIGH_PRIORITY);
                        }
                    }

                    irc_end_of_excepted(resp);
                }

                // 475: cannot join channel (+k)
                if(len>4 && str[0]=='4' && str[1]=='7' && str[2]=='5' && str[3]==' ') {
                    implemented=true;

                    irc_kill_command_wait("JOIN");
                    irc_kill_command_wait("MODE");
                    irc_kill_command_wait("TOPIC");

                    string s=str;
                    int i1=0;
                    while(i1<2 && s.length()>0) {
                        if(s[0]==0x20)
                            i1++;
                        s.erase(0,1);
                    }
                    string chan;
                    while(s.length()>0 && s[0]!=0x20) {
                        chan+=s[0];
                        s.erase(0,1);
                    }

                    bool got=false;
label_a:
                    vector<c_key_chan>::iterator i2;
                    for(i2=irc_key_chans.begin(); i2!=irc_key_chans.end(); ++i2) {
                        if(!irc_cmp_strings_case_insensitive((*i2).channel_name,chan)) {
                            string key;
                            if((*i2).keys.begin()==(*i2).keys.end()) {
                                // :-(
                                string log="Cannot join channel ";
                                log+=chan;
                                log+=" because it is +k, and I have tried all keys that I have in the history. Next retry is scheduled.";
                                log_bot(log);
                                irc_key_chans.erase(i2);

                                {
                                    bool got=false;
                                    vector<string>::iterator ii;
                                    for(ii=irc_cannot_join_channels.begin(); ii!=irc_cannot_join_channels.end(); ++ii)
                                        if(!chan.compare(*ii)) {
                                            got=true;
                                        }
                                    if(!got)
                                        irc_cannot_join_channels.push_back(chan);
                                }

                                got=true;
                                break;
                            } else {
                                key=*((*i2).keys.begin());
                                (*i2).keys.erase((*i2).keys.begin());

                                string msg="JOIN ";
                                msg+=chan;
                                msg+=" ";
                                msg+=key;
                                irc_putserv(c_tracked_msg_to_server(msg),true,HIGH_PRIORITY);
                                msg="MODE ";
                                msg+=chan;
                                irc_putserv(c_tracked_msg_to_server(msg),true,HIGH_PRIORITY);
                                msg="TOPIC ";
                                msg+=chan;
                                irc_putserv(c_tracked_msg_to_server(msg),true,HIGH_PRIORITY);
                                irc_putserv(c_tracked_msg_to_server("WHOIS").add_nick(irc_nick),true,LOW_PRIORITY);

                                irc_kill_last_msg_wait(resp);

                                got=true;
                            }
                        }
                    }
                    if(!got) {
                        c_key_chan kch;
                        kch.channel_name=chan;
                        kch.keys.clear();
                        bool got_chan_def=false;
                        vector<c_channel_def>::iterator i2;
                        for(i2=r_channel_defs.begin(); i2!=r_channel_defs.end(); ++i2)
                            if(!irc_cmp_strings_case_insensitive((*i2).channel_name,chan)) {
                                got_chan_def=true;
                                vector<string>::iterator i3;
                                for(i3=(*i2).keys.begin(); i3!=(*i2).keys.end(); ++i3)
                                    kch.keys.push_back(*i3);
                                irc_key_chans.push_back(kch);
                                break;
                            }
                        if(got_chan_def) {
                            got=true; // for case that we don't have a channel definition
                            goto label_a;
                        }
                    }
                }

                // 348: exceptions
                if(len>4 && str[0]=='3' && str[1]=='4' && str[2]=='8' && str[3]==' ') {
                    implemented=true;

                    if(irc_check_last_msg("MODE"))
                        irc_kill_last_msg_wait(resp);
                    irc_excepted(resp);
                }

                // 349: end of exceptions
                if(len>4 && str[0]=='3' && str[1]=='4' && str[2]=='9' && str[3]==' ') {
                    implemented=true;

                    if(irc_check_last_msg("MODE"))
                        irc_kill_last_msg_wait(resp);
                    irc_kill_command_wait("MODE");
                    irc_end_of_excepted(resp);
                }

                // 345: End of Channel Reop List
                if(len>4 && str[0]=='3' && str[1]=='4' && str[2]=='5' && str[3]==' ') {
                    implemented=true;

                    if(irc_check_last_msg("MODE"))
                        irc_kill_last_msg_wait(resp);
                    irc_kill_command_wait("MODE");
                    irc_end_of_reop(resp);
                }

                // 367: ban list
                if(len>4 && str[0]=='3' && str[1]=='6' && str[2]=='7' && str[3]==' ') {
                    implemented=true;

                    irc_banned(resp);
                }

                // 368: end of +b list
                if(len>4 && str[0]=='3' && str[1]=='6' && str[2]=='8' && str[3]==' ') {
                    implemented=true;

                    if(irc_check_last_msg("MODE"))
                        irc_kill_last_msg_wait(resp);
                    irc_kill_command_wait("MODE");
                }

                // 346: invited
                if(len>4 && str[0]=='3' && str[1]=='4' && str[2]=='6' && str[3]==' ') {
                    implemented=true;

                    irc_invited(resp);
                }

                // 347: end of +I list
                if(len>4 && str[0]=='3' && str[1]=='4' && str[2]=='7' && str[3]==' ') {
                    implemented=true;

                    if(irc_check_last_msg("MODE"))
                        irc_kill_last_msg_wait(resp);
                    irc_kill_command_wait("MODE");

                    string s=str;
                    string channel;
                    int pos=0;
                    for(unsigned int i1=0; i1<len; ++i1) {
                        if(str[i1]==0x20) {
                            pos++;
                            continue;
                        }
                        if(pos==2)
                            channel+=str[i1];
                        if(pos>2)
                            break;
                    }
                    irc_check_for_not_invited(channel);

                    c_online_channel* chan;
                    if(irc_get_online_channel(channel,chan)) {
                        if(!chan->b_excepts_sent && irc_isupport.chanmodes_a.find("e")!=string::npos) {
                            chan->b_excepts_sent=true;
                            string excepts="MODE ";
                            excepts+=channel;
                            excepts+=" +e";
                            irc_putserv(c_tracked_msg_to_server(excepts),true,HIGH_PRIORITY);
                        }
                    }
                }

                // 344: in reop list +R
                if(len>4 && str[0]=='3' && str[1]=='4' && str[2]=='4' && str[3]==' ') {
                    implemented=true;

                    irc_in_reop(resp);
                }

                // :%s 042 %s %s :your unique ID
                if(len>4 && str[0]=='0' && str[1]=='4' && str[2]=='2' && str[3]==' ') {
                    implemented=true;

                    irc_RPL_YOURID(resp);
                }

                // :%s 043 %s %s :nickname collision, forcing nick change to your unique ID.
                if(len>4 && str[0]=='0' && str[1]=='4' && str[2]=='3' && str[3]==' ') {
                    implemented=true;

                    static bool got=false;
                    if(!got) {
                        got=true;

                        nick=irc_nick;
                        ident=irc_get_ident(nick);
                        host=irc_get_host(nick);
                        msg=irc_unique_id;
                        logic_fnc(nick,irc_unique_id);
                        goto forced_nick_change;
                    } else
                        got=false;
                }

                // 482: you are not a channel operator
                if(len>4 && str[0]=='4' && str[1]=='8' && str[2]=='2' && str[3]==' ') {
                    implemented=true;

                    if(irc_check_last_msg("MODE"))
                        irc_kill_last_msg_wait(resp);
                    irc_kill_command_wait("MODE");

                    if(irc_check_last_msg("KICK"))
                        irc_kill_last_msg_wait(resp);
                    irc_kill_command_wait("KICK");
                }

                // 438: Nick change too fast. Please wait X seconds
                if(len>4 && str[0]=='4' && str[1]=='3' && str[2]=='8' && str[3]==' ') {
                    implemented=true;

                    if(irc_check_last_msg("NICK"))
                        irc_kill_last_msg_wait(resp);
                    irc_kill_command_wait("NICK");
                }

                // 251: There are X users and X invisible on X servers
                if(len>4 && str[0]=='2' && str[1]=='5' && str[2]=='1' && str[3]==' ') {
                    implemented=true;
                }

                // 252: X :operators online
                if(len>4 && str[0]=='2' && str[1]=='5' && str[2]=='2' && str[3]==' ') {
                    implemented=true;
                }

                // 253: X :unknown connections
                if(len>4 && str[0]=='2' && str[1]=='5' && str[2]=='3' && str[3]==' ') {
                    implemented=true;
                }

                // 254: X :channels formed
                if(len>4 && str[0]=='2' && str[1]=='5' && str[2]=='4' && str[3]==' ') {
                    implemented=true;
                }

                // 255: I have X clients and X servers
                if(len>4 && str[0]=='2' && str[1]=='5' && str[2]=='5' && str[3]==' ') {
                    implemented=true;
                }

                // 265: X X :Current local users X, max X
                if(len>4 && str[0]=='2' && str[1]=='6' && str[2]=='5' && str[3]==' ') {
                    implemented=true;
                }

                // 266: X X :Current global users X, max X
                if(len>4 && str[0]=='2' && str[1]=='6' && str[2]=='6' && str[3]==' ') {
                    implemented=true;
                }

                // 375: Message of the Day -
                if(len>4 && str[0]=='3' && str[1]=='7' && str[2]=='5' && str[3]==' ') {
                    implemented=true;
                }

                // 372: <MOTD content>
                if(len>4 && str[0]=='3' && str[1]=='7' && str[2]=='2' && str[3]==' ') {
                    implemented=true;
                }

                // 376: End of MOTD command.
                if(len>4 && str[0]=='3' && str[1]=='7' && str[2]=='6' && str[3]==' ') {
                    implemented=true;
                }

                // 422: MOTD File is missing
                if(len>4 && str[0]=='4' && str[1]=='2' && str[2]=='2' && str[3]==' ') {
                    implemented=true;
                }

                // 422: MOTD File is missing
                if(len>4 && str[0]=='4' && str[1]=='2' && str[2]=='2' && str[3]==' ') {
                    implemented=true;
                }

                // 329: (UnrealIRCd - I guess it is channel creation timestamp)
                if(len>4 && str[0]=='3' && str[1]=='2' && str[2]=='9' && str[3]==' ') {
                    implemented=true;
                }

                // 378: (UnrealIRCd) is connecting from X X
                if(len>4 && str[0]=='3' && str[1]=='7' && str[2]=='8' && str[3]==' ') {
                    implemented=true;
                }

                // 312: (server comment in WHOIS reply)
                if(len>4 && str[0]=='3' && str[1]=='1' && str[2]=='2' && str[3]==' ') {
                    implemented=true;
                }

                // 317: X X :seconds idle, signon time
                if(len>4 && str[0]=='3' && str[1]=='1' && str[2]=='7' && str[3]==' ') {
                    implemented=true;
                }

                // 401: No such nick/channel
                if(len>4 && str[0]=='4' && str[1]=='0' && str[2]=='1' && str[3]==' ') {
                    implemented=true;

                    if(irc_check_last_msg("WHOIS"))
                        irc_kill_last_msg_wait(resp);
                    irc_kill_command_wait("WHOIS");

                    if(irc_check_last_msg("MODE"))
                        irc_kill_last_msg_wait(resp);
                    irc_kill_command_wait("MODE");

                    if(irc_check_last_msg("TOPIC"))
                        irc_kill_last_msg_wait(resp);
                    irc_kill_command_wait("TOPIC");

                    if(irc_check_last_msg("NAMES"))
                        irc_kill_last_msg_wait(resp);
                    irc_kill_command_wait("NAMES");
                }

                // 403: No such channel
                if(len>4 && str[0]=='4' && str[1]=='0' && str[2]=='3' && str[3]==' ') {
                    implemented=true;

                    if(irc_check_last_msg("MODE"))
                        irc_kill_last_msg_wait(resp);
                    irc_kill_command_wait("MODE");

                    if(irc_check_last_msg("TOPIC"))
                        irc_kill_last_msg_wait(resp);
                    irc_kill_command_wait("TOPIC");

                    if(irc_check_last_msg("NAMES"))
                        irc_kill_last_msg_wait(resp);
                    irc_kill_command_wait("NAMES");
                }

                // 441: They aren't on that channel
                if(len>4 && str[0]=='4' && str[1]=='4' && str[2]=='1' && str[3]==' ') {
                    implemented=true;

                    if(irc_check_last_msg("MODE"))
                        irc_kill_last_msg_wait(resp);
                    irc_kill_command_wait("MODE");

                    if(irc_check_last_msg("TOPIC"))
                        irc_kill_last_msg_wait(resp);
                    irc_kill_command_wait("TOPIC");

                    if(irc_check_last_msg("NAMES"))
                        irc_kill_last_msg_wait(resp);
                    irc_kill_command_wait("NAMES");
                }

                // 301: Away
                if(len>4 && str[0]=='3' && str[1]=='0' && str[2]=='1' && str[3]==' ') {
                    implemented=true;
                }

                // 333: (IRCnet) sign-on time
                if(len>4 && str[0]=='3' && str[1]=='3' && str[2]=='3' && str[3]==' ') {
                    implemented=true;
                }

                // 020: (IRCnet) :Please wait while we process your connection.
                if(len>4 && str[0]=='0' && str[1]=='2' && str[2]=='0' && str[3]==' ') {
                    implemented=true;
                }

                // 671: (SSL-enabled IRC servers) :is using a Secure Connection
                if(len>4 && str[0]=='6' && str[1]=='7' && str[2]=='1' && str[3]==' ') {
                    implemented=true;
                }

                // 404 :irc.localhost 404 BotNick #channel :Cannot send to channel
                if(len>4 && str[0]=='4' && str[1]=='0' && str[2]=='4' && str[3]==' ') {
                    implemented=true;
                }

                // 010 :irc.localhost 010 BotNick irc.other.server.net 6669 :Please use this Server/Port instead
                if(len>4 && str[0]=='0' && str[1]=='1' && str[2]=='0' && str[3]==' ') {
                    implemented=true;
                }

                // 310 (Unreal IRCd) :vdm.no-ip.org 310 BotNick IrcOpNick :is available for help.
                if(len>4 && str[0]=='3' && str[1]=='1' && str[2]=='0' && str[3]==' ') {
                    implemented=true;
                }

                // WHO reply: "<channel> <user> <host> <server> <nick> <H|G>[*][@|+] :<hopcount> <real name>"
                if(len>4 && str[0]=='3' && str[1]=='5' && str[2]=='2' && str[3]==' ') {
                    implemented=true;

                    irc_RPL_WHOREPLY(str);
                }

                // <name> :End of /WHO list
                if(len>4 && str[0]=='3' && str[1]=='1' && str[2]=='5' && str[3]==' ') {
                    implemented=true;

                    irc_kill_command_wait("WHO");
                }

                // :example.org 335 nick nick :is a \001Bot\001 on ROXnet
                if(len>4 && str[0]=='3' && str[1]=='3' && str[2]=='5' && str[3]==' ') {
                    implemented=true;
                }

                // :example.org 330 my_nick nick login_name :is logged in as
                if(len>4 && str[0]=='3' && str[1]=='3' && str[2]=='0' && str[3]==' ') {
                    implemented=true;
                }

                // :irc.example.org 320 my_nick nick :is a Secure Connection (SSL)
                // (remote WHOIS reply, IRCnet)
                if(len>4 && str[0]=='3' && str[1]=='2' && str[2]=='0' && str[3]==' ') {
                    implemented=true;
                }

                if(last_msg.wait_for_response && last_msg.timestamp+IRC_RESPONSE_TIMEOUT<my_time(NULL))
                    irc_kill_last_msg_wait(resp);
            }
        }

        l2:

        implemented|=mod_dispatch("irc_raw","","",line,false,0,NULL,"origin","@raw@",NULL);

        if(!implemented) {
            string log="OOPS: Handler for message not implemented (please, report this problem); message was: ";
            log+=orig_line;
            log_debug(log);
        }

        memset(last_msg.response,0,sizeof(last_msg.response));
        strncpy(last_msg.response,orig_line.c_str(),sizeof(last_msg.response)-1);
    }

    dcc_loop();

    return 0;
}

/*!
    \brief Resets user's old_mode atribute
    \author VooDooMan
    \version 2
    \date 2005
    \param channel Name of channel
    \param nick Nick name
    \param prefix '-' or '+'
    \param mode '@' or '+'
    \warning Behaviour of this function changed in version 1.0.40: Now it just resets old_mode
*/
MY_STATIC void irc_internal_set_mode(const string& channel, const string& nick, char UNUSEDARG(prefix), char UNUSEDARG(mode))
{
    vector<c_online_channel>::iterator i1;
    for(i1=irc_channels.begin(); i1!=irc_channels.end(); ++i1) {
        if(irc_cmp_strings_case_insensitive((*i1).name,channel))
            continue;
        vector<c_online_user>::iterator i2;
        for(i2=(*i1).users.begin(); i2!=(*i1).users.end(); ++i2)
            if(!irc_cmp_strings_case_insensitive((*i2).nick,nick)) {
                /*while((*i2).mode.find("0",0)!=string::npos)
                    (*i2).mode.erase((*i2).mode.find("0",0),1);
                while((*i2).mode.find(mode,0)!=string::npos)
                    (*i2).mode.erase((*i2).mode.find(mode,0),1);
                if(prefix=='+')
                    (*i2).mode+=mode;
                if((*i2).mode.empty())
                    (*i2).mode="0";*/
                (*i2).old_mode=(*i2).mode;
            }
    }
}

/*!
    \brief Sets +o (chan op) status to nick on channel
    \author VooDooMan
    \version 1
    \date 2004
    \param channel Channel name
    \param nick Nick name
    \param priority Priority class (LOW_PRIORITY, HIGH_PRIORITY, or CRITICAL_PRIORITY)
    \return Returns false if socket I/O error occured
*/
bool irc_op(const string& channel, const string& nick, int priority)
{
    if(channel.length()>128)
        return false;
    if(nick.length()>128)
        return false;

    irc_internal_set_mode(channel,nick,'+','@');

    return irc_putserv(c_tracked_msg_to_server("MODE").add_channel(channel).add_piece("+o").add_nick(nick),true,priority);
}

/*!
    \brief Sets -o (chan op) status to nick on channel
    \author VooDooMan
    \version 1
    \date 2004
    \param channel Channel name
    \param nick Nick name
    \return Returns false if socket I/O error occured
    \param priority Priority class (LOW_PRIORITY, HIGH_PRIORITY, or CRITICAL_PRIORITY)
*/
bool irc_deop(const string& channel, const string& nick, int priority)
{
    if(channel.length()>128)
        return false;
    if(nick.length()>128)
        return false;

    irc_internal_set_mode(channel,nick,'-','@');

    return irc_putserv(c_tracked_msg_to_server("MODE").add_channel(channel).add_piece("-o").add_nick(nick),true,priority);
}

/*!
    \brief Sets -o (chan op) status to nick on channel, and sets +b ban mask on channel
    \author VooDooMan
    \version 1
    \date 2006-01-03
    \param channel Channel name
    \param nick Nick name
    \param mask Mask to ban
    \return Returns false if socket I/O error occured
    \param priority Priority class (LOW_PRIORITY, HIGH_PRIORITY, or CRITICAL_PRIORITY)
*/
bool irc_deop_ban_mask(const string& channel, const string& nick, const string& mask, int priority)
{
    if(channel.length()>128)
        return false;
    if(nick.length()>128)
        return false;
    if(mask.length()>512-1)
        return false;

    irc_internal_set_mode(channel,nick,'-','@');

    return irc_putserv(c_tracked_msg_to_server("MODE").add_channel(channel).add_piece("-o+b").add_nick(nick).add_piece(mask),true,priority);
}

/*!
    \brief Sets +v (voice) status to nick on channel
    \author VooDooMan
    \version 1
    \date 2004
    \param channel Channel name
    \param nick Nick name
    \return Returns false if socket I/O error occured
    \param priority Priority class (LOW_PRIORITY, HIGH_PRIORITY, or CRITICAL_PRIORITY)
*/
bool irc_voice(const string& channel, const string& nick, int priority)
{
    if(channel.length()>128)
        return false;
    if(nick.length()>128)
        return false;

    irc_internal_set_mode(channel,nick,'+','+');

    return irc_putserv(c_tracked_msg_to_server("MODE").add_channel(channel).add_piece("+v").add_nick(nick),true,priority);
}

/*!
    \brief Sets -v (voice) status to nick on channel
    \author VooDooMan
    \version 1
    \date 2004
    \param channel Channel name
    \param nick Nick name
    \param priority Priority class (LOW_PRIORITY, HIGH_PRIORITY, or CRITICAL_PRIORITY)
    \return Returns false if socket I/O error occured
*/
bool irc_devoice(const string& channel, const string& nick, int priority)
{
    if(channel.length()>128)
        return false;
    if(nick.length()>128)
        return false;

    irc_internal_set_mode(channel,nick,'-','+');

    return irc_putserv(c_tracked_msg_to_server("MODE").add_channel(channel).add_piece("-v").add_nick(nick),true,priority);
}

/*!
    \brief Kicks nick on channel
    \author VooDooMan
    \version 1
    \date 2004
    \param channel Channel name
    \param nick Nick name
    \param reason Kick message (reason of kick)
    \param priority Priority of message
    \return Returns false if socket I/O error occured
*/
bool irc_kick(const string& channel, const string& nick, const string& reason, int priority)
{
    if(channel.length()>128)
        return false;
    if(nick.length()>128)
        return false;
    if(reason.length()>128)
        return false;

    string msg;
    if(!reason.empty()) {
        msg=":";
        msg+=reason;
    }

    return irc_putserv(c_tracked_msg_to_server("KICK").add_channel(channel).add_nick(nick).add_piece(msg),false,priority);
}

/*!
    \brief JOINs channel
    \author VooDooMan
    \version 1
    \date 2004
    \param channel Channel name
    \param key Key to channel (can be empty)
    \return Returns false if socket I/O error occured
*/
bool irc_join(const string& channel, const string& key)
{
    if(channel.length()>128)
        return false;

    string msg="JOIN ";
    msg+=channel;
    if(!key.empty()) {
        msg+=" ";
        msg+=key;
    }
    bool ok=irc_putserv(c_tracked_msg_to_server(msg),true,HIGH_PRIORITY);
    msg="MODE ";
    msg+=channel;
    ok&=irc_putserv(c_tracked_msg_to_server(msg),true,HIGH_PRIORITY);
    msg="TOPIC ";
    msg+=channel;
    ok&=irc_putserv(c_tracked_msg_to_server(msg),true,HIGH_PRIORITY);
    ok&=irc_putserv(c_tracked_msg_to_server("WHOIS").add_nick(irc_nick),true,LOW_PRIORITY);

    if(ok)
        was_whois_burst=into_whois_burst=true;

    return ok;
}

/*!
    \brief PARTs channel
    \author VooDooMan
    \version 1
    \date 2004
    \param channel Channel name
    \param reason PART message (reason of PART)
    \return Returns false if socket I/O error occured
*/
bool irc_part(const string& channel, const string& reason)
{
    if(channel.length()>128)
        return false;
    if(reason.length()>128)
        return false;

    string msg="PART ";
    msg+=channel;
    if(!reason.empty()) {
        msg+=" :";
        msg+=reason;
    }

    return irc_putserv(c_tracked_msg_to_server(msg),false,HIGH_PRIORITY);
}

/*!
    \brief Guess (acording to ISUPPORT 005 numeric) if target is name of channel
    \version 1
    \date 2007-12-05
    \author VooDooMan
    \param target Target to check
    \return Returns 'true' if it is name of channel
*/
MY_STATIC bool irc_is_target_chan(const string& target)
{
    if(target.empty())
        return false; // hmm.. 3-state logic?

    map<string,string>::iterator i1;
    for(i1=irc_isupport.raw.begin(); i1!=irc_isupport.raw.end(); ++i1) {
        if(!(*i1).first.compare("CHANTYPES")) {
            return (*i1).second.find(target[0])!=string::npos;
        }
    }

    // huh?!, okay, guess it without ISUPPORT
    char x=target[0];
    return '#'==x || '&'==x || '!'==x || '+'==x;
}

/*!
    \brief Sends PRIVMSG to user/channel
    \author VooDooMan
    \version 1
    \date 2004
    \param target Nick name / channel name of target
    \param msg Whole message
    \param priority Priority class (LOW_PRIORITY, HIGH_PRIORITY, or CRITICAL_PRIORITY)
    \return Returns false if socket I/O error occured
*/
bool irc_privmsg(const string& target, const string& msg, int priority)
{
    if(target.length()>128)
        return false;
    if(msg.length()>512-1)
        return false;

    string tmp;
    if(irc_is_target_chan(target)) {
        tmp="PRIVMSG ";
        tmp+=target;
        tmp+=" :";
        tmp+=msg;

        return irc_putserv(c_tracked_msg_to_server(tmp),false,priority);
    } else {
        tmp=":";
        tmp+=msg;
        return irc_putserv(c_tracked_msg_to_server("PRIVMSG").add_nick(target).add_piece(tmp),false,priority);
    }
}

/*!
    \brief Sends NOTICE to user/channel
    \author VooDooMan
    \version 1
    \date 2004
    \param target Nick name / channel name of target
    \param msg Whole message
    \param priority Priority class (LOW_PRIORITY, HIGH_PRIORITY, or CRITICAL_PRIORITY)
    \return Returns false if socket I/O error occured
*/
bool irc_notice(const string& target, const string& msg, int priority)
{
    if(target.length()>128)
        return false;
    if(msg.length()>512-1)
        return false;

    string tmp;
    if(irc_is_target_chan(target)) {
        tmp="NOTICE ";
        tmp+=target;
        tmp+=" :";
        tmp+=msg;

        return irc_putserv(c_tracked_msg_to_server(tmp),false,priority);
    } else {
        tmp=":";
        tmp+=msg;
        return irc_putserv(c_tracked_msg_to_server("NOTICE").add_nick(target).add_piece(tmp),false,priority);
    }
}

/*!
    \brief Looks-up user from name in logic.txt to nick name (if online)
    \author VooDooMan
    \version 1
    \date 2004
    \param user User name as in logic.txt
    \return If user is online, returns its nick name, if not, returns user parameter
*/
string irc_get_nick(const string& user)
{
    vector<c_online_channel>::iterator i;
    for(i=irc_channels.begin(); i!=irc_channels.end(); ++i) {
        vector<c_online_user>::iterator ii;
        for(ii=(*i).users.begin(); ii!=(*i).users.end(); ++ii) {
            if(!cmp_strings_case_insensitive((*ii).in_logic_as,user)) {
                return (*ii).nick;
            }
        }
    }

    return user;
}

/*!
    \brief Looks-up user from name in logic.txt to nick name (if online)
    \author VooDooMan
    \version 1
    \date 2004
    \param user User name as in logic.txt
    \return If user is online, returns its nick name, if not, returns empty string
*/
string irc_is_online(const string& user)
{
    vector<c_online_channel>::iterator i;
    for(i=irc_channels.begin(); i!=irc_channels.end(); ++i) {
        vector<c_online_user>::iterator ii;
        for(ii=(*i).users.begin(); ii!=(*i).users.end(); ++ii) {
            if(!cmp_strings_case_insensitive((*ii).in_logic_as,user)) {
                return (*ii).nick;
            }
        }
    }

    return (string)"";
}

/*!
    \brief Looks-up user from name in logic.txt  and checks if it is an irc operator (if online)
    \author VooDooMan
    \version 1
    \date 2004
    \param nick Nick name
    \return If user is online, returns true for irc operator, if it is not online, returns false
*/
bool irc_is_ircop(const string& nick)
{
    vector<c_online_channel>::iterator i;
    for(i=irc_channels.begin(); i!=irc_channels.end(); ++i) {
        vector<c_online_user>::iterator ii;
        for(ii=(*i).users.begin(); ii!=(*i).users.end(); ++ii) {
            if(!irc_cmp_strings_case_insensitive((*ii).nick,nick)) {
                return (*ii).irc_op;
            }
        }
    }

    return false;
}

/*!
    \brief Looks-up user from nick name and checks its mode on channel
    \author VooDooMan
    \version 1
    \date 2004
    \param channel Channel name
    \param nick Nick name
    \return Returns mode (i.e. "+", "@", "~", "@+", "+@", or "0")
    \retval "+" User has voice
    \retval "@" User is channel operator
    \retval "0" Normal user
    \retval "@+" both +o and +v
*/
string irc_get_mode(const string& channel, const string& nick)
{
    vector<c_online_channel>::iterator i;
    for(i=irc_channels.begin(); i!=irc_channels.end(); ++i) {
        if(!irc_cmp_strings_case_insensitive((*i).name,channel)) {
            vector<c_online_user>::iterator ii;
            for(ii=(*i).users.begin(); ii!=(*i).users.end(); ++ii) {
                if(!irc_cmp_strings_case_insensitive((*ii).nick,nick)) {
                    if(!(*ii).got_mode)
                        return (string)"0";
                    return (*ii).mode;
                }
            }
        }
    }

    return (string)"0";
}

/*!
    \brief Looks-up user from nick name and returns their mode(s) for logging purposes to channel log
    \author VooDooMan
    \version 1
    \date 2005
    \param channel Channel name
    \param nick Nick name
    \return Returns mode(s) (i.e. "o", "v", "a", "ov", or "" - an empty string for none, or "?" for not-got-WHOIS-yet/unknown nick)
    \retval '?' if we don't have WHOIS yet, or nick is unknown
*/
string irc_get_modes_for_log(const string& channel, const string& nick)
{
    vector<c_online_channel>::iterator i;
    for(i=irc_channels.begin(); i!=irc_channels.end(); ++i) {
        if(!irc_cmp_strings_case_insensitive((*i).name,channel)) {
            vector<c_online_user>::iterator ii;
            for(ii=(*i).users.begin(); ii!=(*i).users.end(); ++ii) {
                if(!irc_cmp_strings_case_insensitive((*ii).nick,nick)) {
                    if(!(*ii).got_mode)
                        return (string)"?";
                    string res;
                    if((*ii).irc_op)
                        res+="!";
                    for(unsigned int i1=0; i1<(*ii).mode.length(); ++i1)
                        if((*ii).mode[i1]!='0')
                            res+=(string)""+irc_isupport.get_prefix2((*ii).mode[i1]);
                    return res;
                }
            }
        }
    }

    return (string)"?";
}

/*!
    \brief Looks-up user from nick name and checks its full name
    \author VooDooMan
    \version 1
    \date 2004
    \param nick Nick name
    \return Returns full name if online, if not, empty string
*/
string irc_get_fullname(const string& nick)
{
    vector<c_online_channel>::iterator i;
    for(i=irc_channels.begin(); i!=irc_channels.end(); ++i) {
        vector<c_online_user>::iterator ii;
        for(ii=(*i).users.begin(); ii!=(*i).users.end(); ++ii) {
            if(!irc_cmp_strings_case_insensitive((*ii).nick,nick)) {
                if(!(*ii).got_whois)
                    return (string)"";
                return (*ii).fullname;
            }
        }
    }

    return (string)"";
}

/*!
    \brief Unbans (MODE \#channel -b mask) host mask on channel
    \author VooDooMan
    \version 1
    \date 2004
    \param channel Channel name
    \param mask Ban mask
    \param priority Priority of message
    \return Returns false if socket I/O error occured, or channel is longer than 128 chars, or mask is longer than 511 bytes
*/
bool irc_unban(const string& channel, const string& mask, int priority)
{
    if(channel.length()>128)
        return false;
    if(mask.length()>512-1)
        return false;

    vector<c_online_channel>::iterator i;
    for(i=irc_channels.begin(); i!=irc_channels.end(); ++i) {
        if(!irc_cmp_strings_case_insensitive((*i).name,(string)channel)) {
            vector<string>::iterator i2;
            for(i2=(*i).bans.begin(); i2!=(*i).bans.end(); ++i2) {
                if(!irc_cmp_strings_case_insensitive((*i2),(string)mask)) {
                    (*i).bans.erase(i2);
                    break;
                }
            }
        }
    }

    string tmp="MODE ";
    tmp+=channel;
    tmp+=" -b ";
    tmp+=mask;

    return irc_putserv(c_tracked_msg_to_server(tmp),false,priority);
}

/*!
    \brief Bans (MODE \#channel +b mask) host mask on channel
    \author VooDooMan
    \version 1
    \date 2004
    \param channel Channel name
    \param mask Ban mask
    \param priority Priority of message
    \return Returns false if socket I/O error occured, or channel is longer than 128 chars, or mask is longer than 511 bytes
*/
bool irc_ban(const string& channel, const string& mask, int priority)
{
    if(channel.length()>128)
        return false;
    if(mask.length()>512-1)
        return false;

    if(irc_isupport.chanmodes_a.find("b")==string::npos)
         return true;

    string tmp="MODE ";
    tmp+=channel;
    tmp+=" +b ";
    tmp+=mask;

    return irc_putserv(c_tracked_msg_to_server(tmp),false,priority);
}

/*!
    \brief Sets value of irc_follow_redirs flag
    \author VooDooMan
    \version 1
    \date 2004
    \param follow_redirs flag to set
*/
void irc_set_redir(bool follow_redirs)
{
    irc_follow_redirs=follow_redirs;
}

/*!
    \brief Looks-up ident of online user
    \author VooDooMan
    \version 1
    \date 2004
    \param nick Nick name of user
    \return Returns ident of user, or empty string if there's no such nick online
*/
string irc_get_ident(const string& nick)
{
    vector<c_online_channel>::iterator i1;
    for(i1=irc_channels.begin(); i1!=irc_channels.end(); ++i1) {
        vector<c_online_user>::iterator i2;
        for(i2=(*i1).users.begin(); i2!=(*i1).users.end(); ++i2) {
            if(!irc_cmp_strings_case_insensitive((*i2).nick,nick)) {
                if(!(*i2).got_whois)
                    break;
                return (*i2).ident;
            }
        }
    }
    vector<c_online_user>::iterator i2;
    for(i2=gone_users.begin(); i2!=gone_users.end(); ++i2) {
        if(!irc_cmp_strings_case_insensitive((*i2).nick,nick)) {
            if(!(*i2).got_whois)
                break;
            return (*i2).ident;
        }
    }
    return (string)"";
}

/*!
    \brief Looks-up host of online user
    \author VooDooMan
    \version 1
    \date 2004
    \param nick Nick name of user
    \return Returns host of user, or empty string if there's no such nick online
*/
string irc_get_host(const string& nick)
{
    vector<c_online_channel>::iterator i1;
    for(i1=irc_channels.begin(); i1!=irc_channels.end(); ++i1) {
        vector<c_online_user>::iterator i2;
        for(i2=(*i1).users.begin(); i2!=(*i1).users.end(); ++i2) {
            if(!irc_cmp_strings_case_insensitive((*i2).nick,nick)) {
                if(!(*i2).got_whois)
                    break;
                return (*i2).host;
            }
        }
    }
    vector<c_online_user>::iterator i2;
    for(i2=gone_users.begin(); i2!=gone_users.end(); ++i2) {
        if(!irc_cmp_strings_case_insensitive((*i2).nick,nick)) {
            if(!(*i2).got_whois)
                break;
            return (*i2).host;
        }
    }
    return (string)"";
}

/*!
    \brief Raw-ly puts message to queue to send to server
    \author VooDooMan
    \version 1
    \date 2004
    \param data Raw message (i.e. "PRIVMSG nick :hello!")
    \param priority Priority class (LOW_PRIORITY, HIGH_PRIORITY, or CRITICAL_PRIORITY)
*/
void irc_put(const string& data, int priority)
{
    irc_putserv(c_tracked_msg_to_server(data),false,priority);
}

/*!
    \brief Sets mode of channel
    \author VooDooMan
    \version 1
    \date 2004
    \param channel Channel name
    \param mode Mode to set
    \param priority Priority class (LOW_PRIORITY, HIGH_PRIORITY, or CRITICAL_PRIORITY)
    \return Returns irc_putserv() result
*/
bool irc_chan_mode(const string& channel, const string& mode, int priority)
{
    string msg="MODE ";
    msg+=channel;
    msg+=" ";
    msg+=mode;
    return !irc_putserv(c_tracked_msg_to_server(msg),false,priority);
}

void irc_dummy_operation()
{
    my_time_t start=my_time(NULL);
    for(;;) {
        sock_send_cache();
        sock_flush_later(irc_socket);

        if(irc_ssl_connection)
            irc_ssl_connection->process();

        irc_loop_putserv();
        if(irc_loop_process_input()!=0)
            break; // finally! yeah!

        if(start+3<my_time(NULL))
            break;

        my_sleep(10);
    }
}

/*!
    \brief Quits the IRC
    \author VooDooMan
    \version 1
    \date 2004
    \param reason Quit message
    \param wait If to wait for response from server
*/
void irc_quit(const string& reason, bool wait)
{
    if(irc_sent_quit)
        return;
    irc_sent_quit=true;
    string tmp="QUIT :";
    tmp+=reason;
    tmp+=(string)"\r\n";
    irc_putserv_immediately(irc_socket,tmp.c_str(),true);

    if(wait)
        irc_dummy_operation();
}

/*!
    \brief Called to await DCC connection from user
    \author VooDooMan
    \version 1
    \date 2004
    \param nick Nick of user
    \param dcc_group DCC-group ID
    \param user_name_as_in_logic User name from "logic.txt"
*/
void irc_await_dcc_chat(const string& nick, const string& user_name_as_in_logic, int dcc_group)
{
    string dcc_host=logic_get_dcc_host(dcc_group,user_name_as_in_logic);

    c_dcc_awaiting_user user;

    user.nick=nick;
    user.ident=irc_get_ident(nick);
    user.host=irc_get_host(nick);
    user.fullname=irc_get_fullname(nick);

    size_t n=user_name_as_in_logic.length();
    if(n>128-1)
        n=128-1;
    strncpy(user.user_name_as_in_logic,user_name_as_in_logic.c_str(),n);
    user.user_name_as_in_logic[n]=0;

    in_addr6_ ip6;
    if(my_inet_pton(AF_INET6,dcc_host.c_str(),&ip6)!=1) {
        user.dcc_host=dcc_host;
        user.resolv_result=threads_resolve4(dcc_host.c_str(),user.addr4,RT_RESOLVE_SCHEDULE);
        user.using_ipv6=false;
    } else {
        memcpy(&user.addr6,&ip6,sizeof(user.addr6));
#if defined(_WIN32) || defined(_WIN64)
        user.addr4.S_un.S_addr=static_cast<unsigned int>(-1);
#else
        user.addr4.s_addr=static_cast<unsigned int>(-1);
#endif
        user.using_ipv6=true;
    }
    my_time(&user.start_time);
    dcc_awaiting_users.push_back(user);
}

/*!
    \brief Called from logic.cpp after rehashing. Deletes all online users and so applies new configuration
    \author VooDooMan
    \version 1
    \date 2004
*/
void irc_rehashed()
{
    irc_bot_flood_bytes=sto_size_t(core_conf_getvar("irc_bot_flood_bytes"));
    irc_bot_flood_seconds=sto_time(core_conf_getvar("irc_bot_flood_seconds"));
    irc_bot_flood_lines=stoui(core_conf_getvar("irc_bot_flood_lines"));

    global_keeper.rehashed();

    {
        vector<c_online_channel>::iterator i1;
        for(i1=irc_channels.begin(); i1!=irc_channels.end(); ++i1) {
            vector<c_online_user>::iterator i2;
            for(i2=(*i1).users.begin(); i2!=(*i1).users.end(); ++i2) {
                (*i2).in_logic_as="";
                if((*i2).got_whois)
                    (*i2).in_logic_as=logic_find_user((*i2).nick,(*i2).ident,(*i2).host,(*i2).fullname,(*i2).irc_op);
                /*
                logic_on_nick_validate((*i2).nick,(*i2).ident,(*i2).host,(*i2).fullname,(*i1).name,(*i2).irc_op,true,(*i2).mod_disp_join_called);
                (*i2).on_join_called=true; // !!!
                */

                logic_on_nick_validate((*i2).nick,(*i2).ident,(*i2).host,(*i2).fullname,(*i1).name,(*i2).irc_op,(*i2).on_join_passive,(*i2).mod_disp_join_called);
                (*i2).on_join_called=true; // !!!
                (*i2).on_join_passive=true;
                (*i2).on_join_called2=true;
            }
        }

        for(i1=irc_channels.begin(); i1!=irc_channels.end(); ++i1) {
            if((*i1).got_excepts && (*i1).got_invites && (*i1).got_reops) {
                vector<string>::iterator i3;

                irc_check_for_not_invited((*i1).name);
                irc_check_for_not_reopped((*i1).name);

                for(i3=(*i1).bans.begin(); i3!=(*i1).bans.end(); ++i3)
                    logic_on_banned((*i1).name,*i3,(*i1).excepts,(*i1).invites,(*i1).bans);
            }
        }
    }
}

/*!
    \brief Checks if bot has op (+o) on channel
    \author VooDooMan
    \version 1
    \date 2004
    \param channel Channel to check
    \retval true If bot got +o
    \retval false If bot got -o
*/
bool irc_got_op(const string& channel)
{
    string m=irc_get_mode(channel,irc_nick);
    return m.find(irc_isupport.get_prefix1('o'),0)!=string::npos;
}

/*!
    \brief Returns channel mode
    \author VooDooMan
    \version 1
    \date 2004
    \param channel Channel name
    \param topic Returns topic of channel
    \return Returns mode (i.e. "", "nt", "nts", "ntp", "n",..)
*/
string irc_get_chan_mode(const string& channel, string& topic)
{
    topic="";
    vector<c_online_channel>::iterator i;
    for(i=irc_channels.begin(); i!=irc_channels.end(); ++i) {
        if(!irc_cmp_strings_case_insensitive((*i).name,channel)) {
            topic=(*i).topic;
            return (*i).modes;
        }
    }

    return (string)"";
}

/*!
    \brief Gets information from parsed RPL_ISUPPORT message (005 numeric)
    \author VooDooMan
    \version 1
    \date 2004
    \param raw Assiciative array with all features
    \param prefix Associative array with prefixes (user modes: 'o'=='@', 'v'=='+', etc.)
    \param chm_a Channel modes class A - +b, +e, ...
    \param chm_b Channel modes class B - +k
    \param chm_c Channel modes class C - +l
    \param chm_d Channel modes class D - +n, +t, ...
*/
void irc_get_005(map<string,string>& raw, map<char,char>& prefix, string& chm_a, string& chm_b, string& chm_c, string& chm_d)
{
    raw=irc_isupport.raw;
    prefix=irc_isupport.prefix;
    chm_a=irc_isupport.chanmodes_a;
    chm_b=irc_isupport.chanmodes_b;
    chm_c=irc_isupport.chanmodes_c;
    chm_d=irc_isupport.chanmodes_d;
}

void irc_cleanup()
{
    msgs_to_server.clear();
    msgs_to_server2.clear();
    last_msg.m_msg=c_tracked_msg_to_server();
    msgs_from_server.clear();
}

/*!
    \brief Should be called when some user just changed their nick: Checks if old nick appears in send queues and adjusts it to new one
    \version 1
    \date 2007-12-03
    \author VooDooMan
    \param old_nick Old nick
    \param new_nick New nick
*/
MY_STATIC void process_tracked_nicks(const string& old_nick, const string& new_nick)
{
    bool changed=false;

    deque<c_msg_to_server>::iterator i1;

    for(i1=msgs_to_server.begin(); i1!=msgs_to_server.end(); ++i1)
        changed|=(*i1).m_msg.process_nick_change(old_nick,new_nick);

    for(i1=msgs_to_server2.begin(); i1!=msgs_to_server2.end(); ++i1)
        changed|=(*i1).m_msg.process_nick_change(old_nick,new_nick);

    if(changed) {
        string log="Adjusted nick change in the send queue for new nick: ";
        log+=new_nick;

        log_irc_ex(E_LOG_IRC_EX_INF,log);
    }
}

/*!
    \brief Should be called when some user has just part a channel: Checks if nick appears in send queues and removes it (from MODE)
    \version 1
    \date 2007-12-12
    \author VooDooMan
    \param channel Channel name
    \param part_nick Nick that has part/quit/got kicked from channel
*/
MY_STATIC void process_tracked_nicks_chan(const string& channel, const string& part_nick)
{
    bool changed=false;

    deque<c_msg_to_server>::iterator i1;

    for(i1=msgs_to_server.begin(); i1!=msgs_to_server.end(); ++i1)
        changed|=(*i1).m_msg.process_nick_change_chan(channel,part_nick);

    for(i1=msgs_to_server2.begin(); i1!=msgs_to_server2.end(); ++i1)
        changed|=(*i1).m_msg.process_nick_change_chan(channel,part_nick);

    if(changed) {
        string log="Removed nick in the send queue for nick: ";
        log+=part_nick;

        log_irc_ex(E_LOG_IRC_EX_INF,log);
    }
}

/*!
    \brief Called in a loop to process delayed WHO/WHOIS (waiting for DNS lookup)
    \author VooDooMan
    \date 2008-01-20
    \version 1
*/
MY_STATIC void irc_process_nicks()
{
    list<c_dns_wait>::iterator i1=irc_waiting_for_dns.begin();
    if(irc_waiting_for_dns.end()==i1)
        return;

    string nick=(*i1).m_nick;
    string ident=(*i1).m_ident;
    string host=(*i1).m_host;
    string fullname=(*i1).m_fullname;

    bool do_again=false;

    {
        in_addr addr4;
        if(threads_resolve4(host.c_str(),addr4,RT_RESOLVE_SCHEDULE)==RR_SCHEDULED)
            do_again=true;
        in_addr6_ addr6;
        if(threads_resolve6(host.c_str(),addr6,RT_RESOLVE_SCHEDULE)==RR_SCHEDULED)
            do_again=true;
        if(core_conf_getvar("try_reverse_lookup").compare("0")) {
            char tmp[2048];
            if(my_inet_pton(AF_INET,host.c_str(),&addr4)==1)
                if(threads_reverse4(host.c_str(),tmp,sizeof(tmp)-1,RT_RESOLVE_SCHEDULE)==RR_SCHEDULED)
                    do_again=true;
            if(my_inet_pton(AF_INET6,host.c_str(),&addr6)==1)
                if(threads_reverse6(addr6,tmp,sizeof(tmp)-1,RT_RESOLVE_SCHEDULE)==RR_SCHEDULED)
                    do_again=true;
        }
    }

    if(do_again)
        return;

    vector<c_online_channel>::iterator i;
    for(i=irc_channels.begin(); i!=irc_channels.end(); ++i) {
        vector<c_online_user>::iterator ii;
        for(ii=(*i).users.begin(); ii!=(*i).users.end(); ++ii) {
            if(!irc_cmp_strings_case_insensitive((*ii).nick,nick)) {

                if(!(*ii).got_whois) {
                    do_again=true;
                    break;
                }

                bool was_dns=true;

                BEGIN_DEBUG_TIME_MEASUREMENT(true,1,"#1");
                if(((*ii).in_logic_as.empty()) || !(*ii).nick_validated)
                    (*ii).in_logic_as=logic_find_user(nick,ident,host,fullname,(*ii).irc_op,was_dns);
                END_DEBUG_TIME_MEASUREMENT(1,0);

                if(!was_dns)
                    return;

                BEGIN_DEBUG_TIME_MEASUREMENT(true,2,"#2");
                global_keeper.add_user_whois4((*ii).nick,(*ii).in_logic_as);
                END_DEBUG_TIME_MEASUREMENT(2,0);

                (*ii).got_whois=true;
                (*ii).whois_sent=true;
                (*ii).ident=ident;
                (*ii).host=host;
                (*ii).fullname=fullname;

                if(!(*ii).on_join_passive) {
                    (*ii).just_joined=false;
                    BEGIN_DEBUG_TIME_MEASUREMENT(true,3,"#3");
                    logic_on_join((*ii).nick,(*ii).ident,(*ii).host,(*ii).fullname,(*ii).just_joined_channel,(*ii).irc_op);
                    END_DEBUG_TIME_MEASUREMENT(3,0);
                    (*ii).mod_disp_join_called=true;
                } else {
                    BEGIN_DEBUG_TIME_MEASUREMENT(true,4,"#4");
                    if(((*ii).mode.compare((*ii).old_mode) && (*ii).old_mode.compare("INVALID")) || (*ii).nick.compare((*ii).old_nick))
                        logic_on_nick_validate((*ii).nick,(*ii).ident,(*ii).host,(*ii).fullname,(*i).name,(*ii).irc_op,(*ii).on_join_passive,(*ii).mod_disp_join_called);
                    END_DEBUG_TIME_MEASUREMENT(4,0);
                }
                (*ii).on_join_passive=true;
                (*ii).on_join_called2=true;

                // this code was very slow... however, we got this check delayed on schedule
                /*
                BEGIN_DEBUG_TIME_MEASUREMENT(true,5,"#5");
                irc_check_for_filesystem((*ii).in_logic_as,(*i).name,(*ii).nick);
                END_DEBUG_TIME_MEASUREMENT(5,0);
                */

                (*ii).on_join_called=true;
                (*ii).just_joined=false;
                (*ii).old_mode=(*ii).mode;
                (*ii).old_nick=(*ii).nick;

                break; // because particular nick can be on one channel only once
            }
        }
    }

    if(do_again)
        return;

    irc_waiting_for_dns.pop_front();

    vector<c_online_user>::iterator ii;
    for(ii=tmp_users.begin(); ii!=tmp_users.end(); ++ii) {
        if(!irc_cmp_strings_case_insensitive((*ii).nick,nick)) {

            bool was_dns=true;

            BEGIN_DEBUG_TIME_MEASUREMENT(true,6,"#6");
            if((*ii).got_whois && (*ii).in_logic_as.empty())
                (*ii).in_logic_as=logic_find_user(nick,ident,host,fullname,(*ii).irc_op,was_dns);
            END_DEBUG_TIME_MEASUREMENT(6,0);

            (*ii).got_whois=true;
            (*ii).whois_sent=true;
            (*ii).ident=ident;
            (*ii).host=host;
            (*ii).fullname=fullname;
        }
    }

    vector<c_dcc_request_whois>::iterator i2;
    for(i2=dcc_request_whois.begin(); i2!=dcc_request_whois.end(); ++i2) {
        if(!(*i2).nick.compare(nick) && !(*i2).only_dns) {
            (*i2).full_name=fullname;
            (*i2).got_whois=true;
            (*i2).host=host;
            (*i2).ident=ident;

            BEGIN_DEBUG_TIME_MEASUREMENT(true,7,"#7");
            irc_quoted_callback((*i2).str.c_str(),(*i2).nick.c_str(),(*i2).host_mask.c_str(),(*i2).comment.c_str());
            END_DEBUG_TIME_MEASUREMENT(7,0);
            break;
        }
    }
}

/*!
    \brief This function should be called in a deep sleep loop ("sleep" command/function in user's script) to process receives from IRC server; respond to server's PINGs; put incomming messages (except for server PINGs) into queue to process them after deep sleep has gone away
    \author VooDooMan
    \version 1
    \date 2008-02-02
    \param msg Message for "irc.log" in case of response to PING from IRC server
*/
void irc_loop_heavy_time_consuming_operation(const string& msg)
{
    if(!irc_socket.cmp())
        return;

    if(irc_ssl_error(true))
        return;

    if(irc_process_input_from_server(true))
        return;
    irc_loop_putserv();

    static my_time_t last=my_time(NULL);
    if(last+10>my_time(NULL))
        return;
    my_time(&last);

    list<string>::iterator i1;
again:
    for(i1=msgs_from_server.begin(); i1!=msgs_from_server.end(); ++i1) {
        if((*i1).find("PING ")==0) {
            string input=*i1;
            string resp="PONG ";
            (*i1).erase(0,sizeof("PING ")-1);
            resp+=*i1;

            log_irc(true,input);

            irc_putserv_immediately(irc_socket,resp.c_str(),false,true);

            if(!msg.empty())
                log_irc_ex(E_LOG_IRC_EX_INF,msg);

            msgs_from_server.erase(i1);
            goto again;
        }
    }

    for(int ii=0; ii<10; ++ii) {
        sock_send_cache();
        ssl_loop();
    }
    irc_process_input_from_server(); // important!
}

